program ex1_bd_cmp2
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine BD_CMP2
!   in module SVD_Procedures .
!                                                                              
! LATEST REVISION : 06/06/2017
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c50, bd_cmp2,     &
                         norm, unit_matrix, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, m=8000, n=5000
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of bd_cmp2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err3, err, eps, elapsed_time
    real(stnd), allocatable, dimension(:)   :: d, e
    real(stnd), allocatable, dimension(:,:) :: a2, resid, bd, a, p
!
    integer(i4b) :: l
!
    integer      :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: failure, do_test
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : BIDIAGONAL REDUCTION OF A REAL m-by-n MATRIX (WITH m>=n)
!   USING THE Ralha-Barlow ONE_SIDED ALGORITHM.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
!
    err = zero
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n), p(n,n), d(n), e(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM DATA MATRIX .
!
    call random_number( a )
!
    if ( do_test ) then
!
        allocate( a2(m,n), bd(n,n), resid(n,n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SAVE THE DATA MATRIX.
!
        a2(:m,:n) = a(:m,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   CALL bd_cmp2 TO REDUCE THE MATRIX a TO BIDIAGONAL FORM
!
!               a = Q*BD*P**(t)
!
!   WHERE Q AND P ARE ORTHOGONAL AND BD IS AN UPPER BIDIAGONAL MATRIX.
!   bd_cmp2 USES THE ONE_SIDED RHALA-BARLOW ALGORITM AND IS FASTER THAN
!   THE STANDARD BIDIAGONALIZATION ALGORITHM USED IN bd_cmp. HOWEVER, A
!   A SLIGHT LOSS OF ORTHOGONALITY CAN BE EXPECTED FOR Q SINCE Q IS
!   COMPUTED FROM A RECURRENCE RELATIONSHIP.
!
    call bd_cmp2( a, d, e, p, failure=failure )
!
!   ON OUTPUT OF bd_cmp2:
!
!       a CONTAINS THE FIRST min(n,m) COLUMNS OF Q
!       AND p CONTAINS THE ORTHOGONAL MATRIX P.
!
!       d AND e CONTAINS RESPECTIVELY, THE DIAGONAL AND
!       SUBDIAGONAL OF THE BIDIAGONAL MATRIX BD.
!
!       failure= false :  INDICATES SUCCESSFUL EXIT.
!       failure= true  :  INDICATES THAT a IS NEARLY SINGULAR AND SOME LOSS
!                         OF ORTHOGONALITY FOR Q CAN BE EXPECTED.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION Q**(t)*a - BD*P**(t),
!
        bd(:n,:n) = zero
!
        do l = 1_i4b, n-1_i4b
            bd(l,l)       = d(l)
            bd(l,l+1_i4b) = e(l+1_i4b)
        end do
!
        bd(n,n) = d(n)
!
        resid(:n,:n) = matmul( transpose(a(:m,:n)), a2(:m,:n) )    &
                        - matmul( bd(:n,:n), transpose(p(:n,:n)) )
!
        bd(:n,1_i4b) = norm( resid(:n,:n), dim=2_i4b )
        err1 =  maxval( bd(:n,1_i4b) )/( norm( a2 )*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - Q**(t)*Q.
!
        call unit_matrix( a2(:n,:n) )
!
        resid(:n,:n) = abs( a2(:n,:n) - matmul( transpose( a ), a ) )
        err2 = maxval( resid(:n,:n) )/real(m,stnd)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - P**(t)*P.
!
        resid(:n,:n) = abs( a2(:n,:n) - matmul( transpose( p ), p ) )
        err3 = maxval( resid(:n,:n) )/real(n,stnd)
!
        err = max( err1, err2, err3 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a2, bd, resid )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, p, d, e )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps ) then
!    if ( err<=eps .and. .not.failure ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed bidiagonal decomposition a = Q*BD*P**(t) = ', err1
        write (prtunit,*) 'Orthogonality of the computed Q orthogonal matrix                 = ', err2
        write (prtunit,*) 'Orthogonality of the computed P orthogonal matrix                 = ', err3
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing the bidiagonal reduction of a ', &
       m, ' by ', n,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_bd_cmp2
! ==========================
!
end program ex1_bd_cmp2
