program ex1_bd_cmp3
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine BD_CMP3
!   in module SVD_Procedures.
!                                                                              
! LATEST REVISION : 10/09/2019
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c50, bd_cmp3, &
                         bd_svd, norm, unit_matrix, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, m AND n ARE THE DIMENSIONS OF THE MATRIX WHOSE
! SVD MUST BE COMPUTED, m MUST BE GREATER THAN n, OTHERWISE
! bd_cmp3 WILL STOP WITH AN ERROR MESSAGE.
!
    integer(i4b), parameter :: prtunit=6, m=3000, n=1000
!   
    real(stnd), parameter  :: fudge=c50
!
    character, parameter :: sort='d'
!
    character(len=*), parameter :: name_proc='Example 1 of bd_cmp3'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err, eps, elapsed_time
    real(stnd), allocatable, dimension(:)   :: d, e
    real(stnd), allocatable, dimension(:,:) :: a, at, ata, resid
!
    integer :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: do_test, failure
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : EIGENDECOMPOSITION OF A REAL SYMMETRIC MATRIX PRODUCT,
!               a**(t)*a, USING THE ONE-SIDED RALHA BIDIAGONALIZATION
!               METHOD APPLIED TO MATRIX a .
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n), d(n), e(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A m-by-n RANDOM DATA MATRIX a .
!
    call random_number( a )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( ata(n,n), at(n,m), resid(n,n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        at(:n,:m) = transpose( a(:m,:n) )
!
!       COMPUTE THE SYMMETRIC MATRIX CROSS-PRODUCT a**(t)*a .
!
        ata(:n,:n) = matmul( at(:n,:m), a(:m,:n) )
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   CALL bd_cmp3 TO REDUCE THE MATRIX a TO BIDIAGONAL FORM
!
!                      a = Q*BD*P**(t)
!
!   WHERE P AND Q ARE ORTHOGONAL AND BD IS AN UPPER BIDIAGONAL MATRIX.
!
!   ON ENTRY OF bd_cmp3, a MUST CONTAINS THE INITIAL m-by-n MATRIX.
!   THE ORTHOGONAL MATRIX P IS COMPUTED IF THE LOGICAL ARGUMENT gen_p IS SET TO true.
!   THE ORTHOGONAL MATRIX P IS STORED IN FACTORED FORM IF THE LOGICAL ARGUMENT gen_p IS
!   SET TO false.
!
    call bd_cmp3( a(:m,:n), d(:n), e(:n), gen_p=true, failure=failure )
!
!   ON EXIT OF bd_cmp3:
!
!         ARGUMENTS d AND e CONTAIN, RESPECTIVELY, THE DIAGONAL AND
!         OFF-DIAGONAL ELEMENTS OF THE BIDIAGONAL MATRIX BD.
!
!         IF THE LOGICAL ARGUMENT gen_p IS SET TO false ON ENTRY,
!         THE LEADING n-BY-n LOWER TRIANGULAR PART OF a IS OVERWRITTEN
!         BY THE MATRIX P AS A PRODUCT OF ELEMENTARY REFLECTORS.
!
!         IF THE LOGICAL ARGUMENT gen_p IS SET TO false ON ENTRY,
!         THE LEADING n-BY-n PART OF a IS OVERWRITTEN
!         BY THE MATRIX P.
!
!         Q IS NOT COMPUTED BY bd_cmp3.
!
!   COMPUTE SVD DECOMPOSITION OF MATRIX a WITH SUBROUTINE bd_svd:
!
!                            a = V*D*U**(t)
!
!   WHERE V AND U ARE THE LEFT AND RIGHT SINGULAR VECTORS, RESPECTIVELY, AND
!   D IS A DIAGONAL MATRIX, WITH SINGULAR VALUES ON THE DIAGONAL.
!
    call bd_svd( false, d(:n), e(:n), failure, a(:n,:n), sort=sort )
!
!   ON EXIT OF bd_svd :
!
!         ARGUMENTS d AND a(:n:n) CONTAIN, RESPECTIVELY, THE SINGULAR VALUES AND
!         RIGHT SINGULAR VECTORS OF MATRIX a. LEFT SINGULAR VECTORS ARE NOT COMPUTED
!         WITH THIS CALL OF bd_svd.
!
!   COMPUTE EIGENVALUES OF a**(t)*a FROM THE SINGULAR VALUES OF a .
!
    d(:n) = d(:n)*d(:n)
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION (a**(t)*a)*U - U*D
!
        resid(:n,:n) = matmul( ata(:n,:n), a(:n,:n)  )    &
                       - a(:n,:n)*spread( d(:n), 1, n )
!
        e(:n) = norm( resid(:n,:n), dim=2_i4b )
        err1 =  maxval( e(:n) )/( norm( ata )*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U**(t)*U.
!
        call unit_matrix( ata(:n,:n) )
!
        at(:n,:n) = transpose( a(:n,:n) )
!
        resid(:n,:n) = abs( ata(:n,:n) - matmul( at(:n,:n), a(:n,:n) ) )
        err2 = maxval( resid(:n,:n) )/real(n,stnd)
!
        err = max( err1, err2 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( ata, resid, at )
!
    endif
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, d, e )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed eigendecomposition a**(t)*a = U*D*U**(t) = ', err1
        write (prtunit,*) 'Orthogonality of the computed eigenvectors U**(t)*U - I           = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing the eigendecomposition of a ',       &
       n, ' by ', n,' real matrix cross-product is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_bd_cmp3
! ==========================
!
end program ex1_bd_cmp3
