program ex1_chol_cmp2
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine CHOL_CMP2
!   in module Lin_Procedures . 
!                                                                              
! LATEST REVISION : 11/06/2017
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, zero, one, c10, true, false,  &
                         chol_cmp2, norm, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n=4000, m=n+10
!
    real(stnd), parameter :: fudge=c10
!
    character(len=*), parameter :: name_proc='Example 1 of chol_cmp2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, d1, elapsed_time
    real(stnd), dimension(:,:), allocatable :: a, a2, ainv, c
    real(stnd), dimension(:),   allocatable :: invdiag
!
    integer(i4b) :: j
    integer      :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: do_test, upper
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : COMPUTE THE INVERSE OF A REAL n-BY-n SYMMETRIC POSITIVE DEFINITE
!               MATRIX a BY USING THE CHOLESKY DECOMPOSITION.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*sqrt( epsilon( err ) )
    err = zero
!
    do_test = false
    upper   = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( c(m,n), a(n,n), ainv(n,n), invdiag(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-by-n RANDOM SYMMETRIC POSITIVE DEFINITE MATRIX a .
!
    call random_number( c )
!
    a = matmul( transpose(c), c )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( a2(n,n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF THE SYMMETRIC POSITIVE DEFINITE MATRIX a .
!
        a2(:n,:n) = a(:n,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE THE INVERSE OF A SYMMETRIC POSITIVE DEFINITE
!   MATRIX a BY USING THE CHOLESKY DECOMPOSITION OF a.
!
!   IF ON OUTPUT OF chol_cmp2 d1 IS DIFFERENT FROM ZERO
!   THEN THE SYMMETRIC MATRIX IS NOT SINGULAR AND THE
!   SYMMETRIC INVERSE OF a HAS BEEN COMPUTED.
!
    call chol_cmp2( a, invdiag, d1, matinv=ainv, upper=upper, fill=true )
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( d1==zero ) then
!
!       ANORMAL EXIT IN chol_cmp2 SUBROUTINE, PRINT A WARNING.
!
        write (prtunit,*) 'Error in the call to CHOL_CMP2 subroutine, d1=', d1
!
    else if ( do_test ) then
!
!       COMPUTE a TIMES ITS INVERSE - IDENTITY.
!
        c(:n,:n) = matmul( a2, ainv )
!
        do j = 1_i4b, n
!
            c(j,j) = c(j,j) - one
!
        end do
!
!       CHECK THE RESULTS FOR SMALL RESIDUALS.
!
        err = norm( c(:n,:n) ) /( real(n,stnd)*norm(a2(:n,:n)) )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
        deallocate( a, ainv, c, invdiag, a2 )
    else
        deallocate( a, ainv, c, invdiag )
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. d1/=zero ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing the inverse of a positive definite symmetric matrix of size ', &
       n, ' by ', n, ' is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_chol_cmp2
! ============================
!
end program ex1_chol_cmp2
