program ex1_comp_ginv
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine COMP_GINV
!   in module SVD_Procedures. 
!                                                                              
! LATEST REVISION : 18/06/2017
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, comp_ginv, norm, &
                         c10, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, m=4000, n=2000, k=min(m,n)
!   
    real(stnd), parameter  :: fudge=c10
!
    character(len=*), parameter :: name_proc='Example 1 of comp_ginv'
!   
!   
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: eps, err, err1, err2, err3, err4, elapsed_time
    real(stnd), dimension(:,:), allocatable :: a, a2, ainv, ainv2, a_by_ainv, ainv_by_a
!
    integer :: iok, istart, iend, irate, imax, itime
!
    logical(lgl)   :: failure, do_test
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : COMPUTING THE GENERALIZED INVERSE OF A m-BY-n REAL MATRIX USING
!               THE SINGULAR VALUE DECOMPOSITION (SVD) OF THE MATRIX. THE SVD IS COMPUTED
!               BY THE BIDIAGONAL QR IMPLICIT METHOD WITH A WAVE-FRONT ALGORITHM FOR
!               APPLYING GIVENS ROTATIONS IN THE BIDIAGONAL QR ALGORITHM AND,
!               OPTIONALLY, A PERFECT SHIFT FOR THE SINGULAR VECTORS.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*sqrt( epsilon(eps) )
    err = zero
!
    do_test = false
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n), ainv(n,m), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM DATA MATRIX.
!
    call random_number( a )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(m,n), ainv2(n,m), a_by_ainv(m,m),   &
                  ainv_by_a(n,n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SAVE THE DATA MATRIX.
!
        a2(:m,:n) = a(:m,:n)
!
     end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE GENERALIZED INVERSE OF a(:m,:n) WITH SUBROUTINE comp_ginv.
!   THE GENERALIZED INVERSE IS COMPUTED WITH THE HELP OF THE SINGULAR
!   VALUE DECOMPOSITION (SVD) OF a(:m,:n).
!
    call comp_ginv( a, failure, ainv )
!
!   THE ROUTINE RETURNS THE GENERALIZED INVERSE OF a(:m,:n).
!
!   ON EXIT OF comp_ginv :
!
!       failure= false :  INDICATES SUCCESSFUL EXIT.
!       failure= true  :  INDICATES THAT THE ALGORITHM DID NOT CONVERGE AND
!                         THAT FULL ACCURACY WAS NOT ATTAINED IN THE BIDIAGONAL
!                         SVD OF AN INTERMEDIATE BIDIAGONAL FORM B OF a.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       COMPUTE ainv*a AND a*ainv*a .
!
        ainv_by_a = matmul( ainv, a2 )
        a = matmul( a2, ainv_by_a )
!
!       COMPUTE a*ainv AND ainv*a*ainv .
!
        a_by_ainv = matmul( a2, ainv )
        ainv2 = matmul( ainv, a_by_ainv )
!
!       CHECK THE Moore-Penrose EQUATIONS :
!
!             a*ainv*a = a            (1)
!          ainv*a*ainv = ainv         (2)
!            (a*ainv)' = a*ainv       (3)
!            (ainv*a)' = ainv*a       (4)
!
        err1 = norm( a - a2 )
        err2 = norm( ainv - ainv2 )
        err3 = norm( a_by_ainv - transpose(a_by_ainv) )
        err4 = norm( ainv_by_a - transpose(ainv_by_a) )
!
        err = max( err1, err2, err3, err4 )/ ( real(k,stnd)*norm(a) )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a2, ainv2, a_by_ainv, ainv_by_a )        
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, ainv )
!
    if ( err<=eps .and. .not.failure ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing the generalized inverse of a ', &
       m, ' by ', n,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_comp_ginv
! ============================
!
end program ex1_comp_ginv
