program ex1_comp_inv
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine COMP_INV
!   in module Lin_Procedures . 
!                                                                              
! LATEST REVISION : 18/06/2017
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    USE Statpack, ONLY : lgl, i4b, stnd, true, false, zero, one, comp_inv,   &
                         norm, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
!
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n=4000
!
    character(len=*), parameter :: name_proc='Example 1 of comp_inv'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, elapsed_time
    real(stnd), dimension(:,:), allocatable :: a, ainv, res
!
    integer(i4b) :: j
    integer      :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: do_test, failure
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : COMPUTE IN PLACE THE INVERSE OF A REAL n-BY-n MATRIX.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = sqrt( epsilon( err ) )
    err = zero
!
    do_test = true
!
!   ALLOCATE WORK ARRAY.
!
    allocate( ainv(n,n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
! 
!   GENERATE A RANDOM REAL MATRIX.
!
    call random_number( ainv )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( a(n,n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
! 
!       SAVE THE RANDOM REAL MATRIX.
!
        a(:n,:n) = ainv(:n,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE THE MATRIX INVERSE WITH SUBROUTINE comp_inv.
!   INPUT ARGUMENT OVERWRITTEN.
!
    call comp_inv( ainv, failure )
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( failure ) then
!
!       ANORMAL EXIT FROM comp_inv SUBROUTINE, PRINT A WARNING.
!
        write (prtunit,*) 'Error in exit of COMP_INV subroutine, failure=', failure
        write (prtunit,*)
!
    else if ( do_test ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( res(n,n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE a TIMES ITS INVERSE - IDENTITY.
!
        res = matmul( a, ainv )
!
        do j = 1_i4b, n
            res(j,j) = res(j,j) - one
        end do
!
        err = norm( res(:n,:n) ) /( real(n,stnd)*norm(a(:n,:n)) )
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( res )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( allocated(a) ) then
        deallocate( ainv, a )
    else
        deallocate( ainv )
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*)
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing in place the inverse of a real matrix of size ', &
       n, ' by ', n, ' is', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_comp_inv
! ===========================
!
end program ex1_comp_inv
