program ex1_comp_mca
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines COMP_MCA
!   and COMP_PC_MCA in module Mul_Stat_Procedures .
!                                                                              
! LATEST REVISION : 26/06/2006
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, lgl, true, false, one, comp_mca, comp_pc_mca
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, mx=20, my=10, m=min(mx,my), p=50, mm=(m*(m+1))/2
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                         :: err_xpc, err_ypc, xyn
    real(stnd), dimension(mx,my)       :: xsingvec
    real(stnd), dimension(my,m)        :: ysingvec
    real(stnd), dimension(mx,2)        :: xstat
    real(stnd), dimension(my,2)        :: ystat
    real(stnd), dimension(m)           :: xysingval, xysingvar
    real(stnd), dimension(mm)          :: pccorp_x, pccorp_y
    real(stnd), dimension(mx,m)        :: xpccor
    real(stnd), dimension(my,m)        :: ypccor
    real(stnd), dimension(p,m)         :: xpc, ypc
    real(stnd), dimension(mx,p)        :: x, x_std
    real(stnd), dimension(my,p)        :: y, y_std
!
    logical(lgl) :: first, last, failure
!
    character    :: sort = 'd'
!   
!   
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='Example 1 of comp_mca'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   GENERATE A RANDOM OBSERVATION MATRIX x(:mx,:p) WITH mx VARIABLES
!   AND p OBSERVATIONS.
!
    call random_number( x )
!
!   GENERATE A RANDOM OBSERVATION MATRIX y(:my,:p) WITH my VARIABLES
!   AND p OBSERVATIONS.
!
    call random_number( y )
!
!   COMPUTE LEFT AND RIGHT SINGULAR VECTORS OF THE CORRELATION MATRIX .
!
    first = true
    last  = true
!
    call comp_mca( x, y, first, last, xstat, ystat, xysingval, xsingvec, failure,    &
                   sort=sort, ysingvec=ysingvec, xysingvar=xysingvar  )
!
!   ON EXIT OF COMP_MCA WHEN last=true :
!
!   xstat CONTAINS THE FOLLOWING STATISTICS ON ALL VARIABLES FROM THE x MATRIX:
!            
!       xstat(:,1) CONTAINS THE MEAN VALUES OF THE "LEFT" DATA MATRIX x.
!       xstat(:,2) CONTAINS THE STANDARD-DEVIATIONS OF THE "LEFT" DATA MATRIX x.
!
!   ystat CONTAINS THE FOLLOWING STATISTICS ON ALL VARIABLES FROM THE y MATRIX:
!            
!       ystat(:,1) CONTAINS THE MEAN VALUES OF THE "RIGHT" DATA MATRIX y.
!       ystat(:,2) CONTAINS THE STANDARD-DEVIATIONS OF THE "RIGHT" DATA MATRIX y.
!            
!   xysingval CONTAINS THE m SINGULAR VALUES OF THE CORRELATION
!   (OR COVARIANCE) MATRIX BETWEEN THE DATA MATRICES x AND y.
!
!   xsingvec IS OVERWRITTEN WITH THE FIRST m LEFT SINGULAR VECTORS
!   OF THE CORRELATION (OR COVARIANCE) MATRIX BETWEEN x AND y.
!
!   failure = FALSE :  INDICATES SUCCESSFUL EXIT.
!   failure = TRUE  :  INDICATES THAT FEWER THAN TWO VALID OBSERVATIONS
!                      WERE PRESENT OR THAT MAXIMUM ACCURACY WAS NOT
!                      ACHIEVED WHEN COMPUTING THE SVD OF THE COVARIANCE
!                      (OR CORRELATION) MATRIX BETWEEN THE DATA MATRICES x AND y .
!
!   ysingvec CONTAINS THE FIRST m RIGHT SINGULAR VECTORS OF THE CORRELATION
!            (OR COVARIANCE) MATRIX BETWEEN x AND y.
!
!   xysingvar CONTAINS THE PERCENTAGES OF TOTAL SQUARED COVARIANCE ASSOCIATED 
!             WITH THE LEFT AND RIGHT SINGULAR VECTORS IN ORDER OF THE 
!             SINGULAR VALUES STORED IN xysingval.
!
!
!   NOW, COMPUTE THE LEFT SINGULAR VARIABLES FROM THE DATA AND THE LEFT SINGULAR VECTORS.
!
!   THE OPTIONAL PARAMETERS xmean AND xstd ARE REQUIRED IF THE SINGULAR VECTORS
!   HAVE BEEN COMPUTED FROM THE CORRRELATION MATRIX.
!
    first = true
    last  = true
!
    call comp_pc_mca( x(:mx,:p), xsingvec(:mx,:m), first, last,            &
                      xpccor(:mx,:m), pccorp_x(:mm), xpc(:p,:m), xyn,      &
                      xmean=xstat(:mx,1), xstd=xstat(:mx,2)                )
!
!   ON EXIT OF COMP_PC_MCA WHEN last=true :
!
!     xpccor  CONTAINS :
!                 - THE CORRELATIONS BETWEEN THE DATA MATRIX x
!                   AND THE SINGULAR VARIABLES IF THE OPTIONAL
!                   ARGUMENTS xmean AND xstd ARE PRESENT.
!                 - THE COVARIANCES BETWEEN THE DATA MATRIX x
!                   AND THE NORMALIZED SINGULAR VARIABLES IF ONLY
!                   THE OPTIONAL ARGUMENT xmean IS PRESENT.
!
!     pccorp_x CONTAINS THE CORRELATION MATRIX COR BETWEEN THE SINGULAR VARIABLES
!              STORED IN ARGUMENT xpc. COR IS STORED IN SYMMETRIC STORAGE MODE.
!              MORE PRECISELY, THE J-TH COLUMN OF THIS MATRIX COR IS STORED IN THE 
!              ARRAY pccorp_x AS FOLLOWS:
!
!                         pccorp_x(i + (j-1)*j/2) = COR(i,j) for 1<=i<=j;
!
!     xpc      CONTAINS THE UNNORMALIZED SINGULAR VARIABLES DERIVED
!              FROM x AND xsingvec.
!
!   NOW, COMPUTE THE RIGHT SINGULAR VARIABLES FROM THE DATA AND THE RIGHT SINGULAR VECTORS.
!
    call comp_pc_mca( y, ysingvec(:my,:m), first, last, ypccor, pccorp_y, ypc(:p,:m), xyn,  &
                      xmean=ystat(:my,1), xstd=ystat(:my,2)                                 )
!
!   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION x_std**t*xsingvec - xpc
!   WHERE xsingvec ARE THE LEFT SINGULAR VECTORS, x_std THE STANDARDIZED LEFT DATA
!   AND xpc THE UNNORMALIZED LEFT SINGULAR VARIABLES.
!
    x_std(:mx,:p) = x(:mx,:p) - spread( xstat(:mx,1) ,   dim=2, ncopies=p )
    x_std(:mx,:p) = x_std(:mx,:p)*spread( one/xstat(:mx,2) , dim=2, ncopies=p ) 
!
    err_xpc = sum( abs(matmul(transpose(x_std),xsingvec)-xpc) )/sum( abs(x_std) )
!
!   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION y_std**t*ysingvec - ypc
!   WHERE ysingvec ARE THE RIGHT SINGULAR VECTORS, y_std THE STANDARDIZED RIGHT DATA
!   AND ypc THE UNNORMALIZED RIGHT SINGULAR VARIABLES.
!
    y_std(:my,:p) = y(:my,:p) - spread( ystat(:my,1) ,   dim=2, ncopies=p )
    y_std(:my,:p) = y_std(:my,:p)*spread( one/ystat(:my,2) , dim=2, ncopies=p ) 
!
    err_ypc = sum( abs(matmul(transpose(y_std),ysingvec)-ypc) )/sum( abs(y_std) )
!
!   PRINT THE RESULT OF THE TESTS.
!
    if ( max(err_xpc, err_ypc)<=sqrt( epsilon(err_ypc) ) ) then
        write (prtunit,*) 'Example 1 of COMP_MCA is correct'
    else
        write (prtunit,*) 'Example 1 of COMP_MCA is incorrect'
    end if
!
!
! END OF PROGRAM ex1_comp_mca
! ===========================
!
end program ex1_comp_mca
