program ex1_comp_mca2
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines COMP_MCA2
!   and COMP_PC_MCA in module Mul_Stat_Procedures .
!                                                                              
! LATEST REVISION : 26/06/2006
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, lgl, true, false, one, comp_mca2, comp_pc_mca
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, mx=20, my=10, m=min(mx,my), mxy=mx+my, p=50,   &
                               nsvd=3, nsingp=(nsvd*(nsvd+1))/2
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                         :: err_xpc, err_ypc, xyn
    real(stnd), dimension(mx,my)       :: xycor
    real(stnd), dimension(mxy,nsvd)    :: xysingvec
    real(stnd), dimension(mx,2)        :: xstat
    real(stnd), dimension(my,2)        :: ystat
    real(stnd), dimension(m)           :: xysingval, xysingvar
    real(stnd), dimension(nsingp)      :: pccorp_x, pccorp_y
    real(stnd), dimension(mx,nsvd)     :: xpccor
    real(stnd), dimension(my,nsvd)     :: ypccor
    real(stnd), dimension(p,nsvd)      :: xpc, ypc
    real(stnd), dimension(mx,p)        :: x, x_std
    real(stnd), dimension(my,p)        :: y, y_std
!
    integer(i4b) :: maxiter=3
!
    logical(lgl) :: first, last, failure
!   
!   
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='Example 1 of comp_mca2'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   GENERATE A RANDOM OBSERVATION MATRIX x(:mx,:p) WITH mx VARIABLES
!   AND p OBSERVATIONS.
!
    call random_number( x )
!
!   GENERATE A RANDOM OBSERVATION MATRIX y(:my,:p) WITH my VARIABLES
!   AND p OBSERVATIONS.
!
    call random_number( y )
!
!   COMPUTE the first nsvd LEFT AND RIGHT SINGULAR VECTORS OF THE CORRELATION MATRIX .
!
    first = true
    last  = true
!
    call comp_mca2( x, y, first, last, xstat, ystat, xysingval, xycor, failure,    &
                    maxiter=maxiter, xysingvec=xysingvec, xysingvar=xysingvar  )
!
!   ON EXIT OF COMP_MCA2 WHEN last=true :
!
!   xstat CONTAINS THE FOLLOWING STATISTICS ON ALL VARIABLES FROM THE x MATRIX:
!            
!       xstat(:,1) CONTAINS THE MEAN VALUES OF THE "LEFT" DATA MATRIX x.
!       xstat(:,2) CONTAINS THE STANDARD-DEVIATIONS OF THE "LEFT" DATA MATRIX x.
!
!   ystat CONTAINS THE FOLLOWING STATISTICS ON ALL VARIABLES FROM THE y MATRIX:
!            
!       ystat(:,1) CONTAINS THE MEAN VALUES OF THE "RIGHT" DATA MATRIX y.
!       ystat(:,2) CONTAINS THE STANDARD-DEVIATIONS OF THE "RIGHT" DATA MATRIX y.
!            
!   xysingval CONTAINS THE m SINGULAR VALUES OF THE CORRELATION
!   (OR COVARIANCE) MATRIX BETWEEN THE DATA MATRICES x AND y.
!
!   WHEN OPTIONAL ARGUMENT savecor IS PRESENT AND savecor=TRUE, xycor CONTAINS
!
!       THE CORRELATION OR VARIANCE-COVARIANCE MATRIX AS CONTROLLED BY THE COV ARGUMENT.
!       IN THIS CASE xycor(i,j) CONTAINS THE CORRELATION (OR COVARIANCE) COEFFICIENT
!       BETWEEN x(i,:) AND y(j,:) ( x(:,i) AND y(:,j) IF dimvarx=2 AND
!       dimvary=2 ).
!
!   IF savecor=FALSE OR IS ABSENT, THE CORRELATION (OR COVARIANCE) MATRIX IS NOT SAVED ON EXIT.
!   IN THIS CASE, xycor DOES NOT CONTAIN USEFUL INFORMATION.
!
!   failure = FALSE :  INDICATES SUCCESSFUL EXIT.
!   failure = TRUE  :  INDICATES THAT FEWER THAN TWO VALID OBSERVATIONS
!                      WERE PRESENT OR THAT MAXIMUM ACCURACY WAS NOT
!                      ACHIEVED WHEN COMPUTING THE SVD OF THE COVARIANCE
!                      (OR CORRELATION) MATRIX BETWEEN THE DATA MATRICES x AND y .
!
!   xysingvec CONTAINS THE FIRST nsvd RIGHT SINGULAR VECTORS OF THE CORRELATION
!             (OR COVARIANCE) MATRIX BETWEEN x AND y IN xysingvec(1:mx,:nsvd)
!             AND THE FIRST nsvd LEFT SINGULAR VECTORS IN xysingvec(mx+1:mxy,:nsvd)
!
!   xysingvar CONTAINS THE PERCENTAGES OF TOTAL SQUARED COVARIANCE ASSOCIATED 
!             WITH THE LEFT AND RIGHT SINGULAR VECTORS IN ORDER OF THE 
!             SINGULAR VALUES STORED IN xysingval.
!
!
!   NOW, COMPUTE THE FIRST nsvd LEFT SINGULAR VARIABLES FROM THE DATA AND THE LEFT SINGULAR VECTORS.
!
!   THE OPTIONAL PARAMETERS xmean AND xstd ARE REQUIRED IF THE SINGULAR VECTORS
!   HAVE BEEN COMPUTED FROM THE CORRRELATION MATRIX.
!
    first = true
    last  = true
!
    call comp_pc_mca( x(:mx,:p), xysingvec(:mx,:nsvd), first, last,                  &
                      xpccor(:mx,:nsvd), pccorp_x(:nsingp), xpc(:p,:nsvd), xyn,      &
                      xmean=xstat(:mx,1), xstd=xstat(:mx,2)                          )
!
!   ON EXIT OF COMP_PC_MCA WHEN last=true :
!
!     xpccor  CONTAINS :
!                 - THE CORRELATIONS BETWEEN THE DATA MATRIX x
!                   AND THE SINGULAR VARIABLES IF THE OPTIONAL
!                   ARGUMENTS xmean AND xstd ARE PRESENT.
!                 - THE COVARIANCES BETWEEN THE DATA MATRIX x
!                   AND THE NORMALIZED SINGULAR VARIABLES IF ONLY
!                   THE OPTIONAL ARGUMENT xmean IS PRESENT.
!
!     pccorp_x CONTAINS THE CORRELATION MATRIX COR BETWEEN THE SINGULAR VARIABLES
!              STORED IN ARGUMENT xpc. COR IS STORED IN SYMMETRIC STORAGE MODE.
!              MORE PRECISELY, THE J-TH COLUMN OF THIS MATRIX COR IS STORED IN THE 
!              ARRAY pccorp_x AS FOLLOWS:
!
!                         pccorp_x(i + (j-1)*j/2) = COR(i,j) for 1<=i<=j;
!
!     xpc      CONTAINS THE UNNORMALIZED SINGULAR VARIABLES DERIVED
!              FROM x AND xsingvec.
!
!   NOW, COMPUTE THE FIRST nsvd RIGHT SINGULAR VARIABLES FROM THE DATA AND THE RIGHT SINGULAR VECTORS.
!
    call comp_pc_mca( y(:my,:p), xysingvec(mx+1:mxy,:nsvd), first, last,             &
                      ypccor(:my,:nsvd), pccorp_y(:nsingp), ypc(:p,:nsvd), xyn,      &
                      xmean=ystat(:my,1), xstd=ystat(:my,2)                          )
!
!   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION x_std**t*xsingvec - xpc
!   WHERE xsingvec ARE THE LEFT SINGULAR VECTORS, x_std THE STANDARDIZED LEFT DATA
!   AND xpc THE UNNORMALIZED LEFT SINGULAR VARIABLES.
!
    x_std(:mx,:p) = x(:mx,:p) - spread( xstat(:mx,1) , dim=2, ncopies=p )
    x_std(:mx,:p) = x_std(:mx,:p)*spread( one/xstat(:mx,2) , dim=2, ncopies=p ) 
!
    err_xpc = sum( abs(matmul(transpose(x_std),xysingvec(:mx,:nsvd))-xpc) )/sum( abs(x_std) )
!
!   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION y_std**t*ysingvec - ypc
!   WHERE ysingvec ARE THE RIGHT SINGULAR VECTORS, y_std THE STANDARDIZED RIGHT DATA
!   AND ypc THE UNNORMALIZED RIGHT SINGULAR VARIABLES.
!
    y_std(:my,:p) = y(:my,:p) - spread( ystat(:my,1) , dim=2, ncopies=p )
    y_std(:my,:p) = y_std(:my,:p)*spread( one/ystat(:my,2) , dim=2, ncopies=p ) 
!
    err_ypc = sum( abs(matmul(transpose(y_std),xysingvec(mx+1:mxy,:nsvd))-ypc) )/sum( abs(y_std) )
!
!   PRINT THE RESULT OF THE TESTS.
!
    if ( max(err_xpc, err_ypc)<=sqrt( epsilon(err_xpc) ) ) then
        write (prtunit,*) 'Example 1 of COMP_MCA2 is correct'
    else
        write (prtunit,*) 'Example 1 of COMP_MCA2 is incorrect'
    end if
!
!
! END OF PROGRAM ex1_comp_mca2
! ============================
!
end program ex1_comp_mca2
