program ex1_freq_func
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of functions/subroutines BD_COEF
!   and FREQ_FUNC in module Time_Series_Procedures.
!                                                                              
! LATEST REVISION : 28/11/2007
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, zero, half, one, true, merror, allocate_error,   &
                         bd_coef, freq_func, init_fft, fft, end_fft
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, n IS THE LENGTH OF THE TIME SERIES .
!
    integer(i4b), parameter :: prtunit=6, n=200
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                            :: err, fch, fcl
    real(stnd), dimension(n)              :: freqr, freqr2, coef2
    real(stnd), dimension(:), allocatable :: coef
!
    complex(stnd), dimension(n)           :: coefc
!
    integer(i4b) :: k, k1, k2, pl, ph, khalf, kmid
!   
    integer :: iok
!   
!   
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='Example 1 of freq_func'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   pl IS THE MINIMUM PERIOD OF OSCILLATION OF DESIRED COMPONENT.
!   ph IS THE MAXIMUM PERIOD OF OSCILLATION OF DESIRED COMPONENT.
!   pl AND ph ARE EXPRESSED IN NUMBER OF OBSERVATIONS, I.E. pl=6(18) and ph=32(96) SELECT
!   PERIODS BETWEEN 1.5 YRS AND 8 YRS FOR QUARTERLY (MONTHLY) DATA.
!
    pl  = 18
    ph  = 96
!
!   COMPUTE THE CORRESPONDING CUTOFF FREQUENCIES.
!
    fch = one/real( ph, stnd )
    fcl = one/real( pl, stnd )
!
!   NOW SELECT THE OPTIMAL NUMBER OF FILTER COEFFICIENTS k FOR THE LANCZOS FILTER.
!
    k1  = ceiling( one/(half-fcl) )
    k2  = ceiling(  2.6/(fcl-fch) )
    k   = max( k1, k2, ph+1 )
!
!   CHECK IF k IS ODD.
!
    if ( (k/2)*2==k  ) k = k + 1
!
!   ALLOCATE WORK ARRAY FOR THE FILTER COEFFICIENTS.
!
    allocate( coef(k), stat=iok )
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   FUNCTION bd_coef COMPUTES THE K-TERM LEAST SQUARES APPROXIMATION TO AN
!   -IDEAL- BAND PASS FILTER WITH CUTOFF PERIODS PL AND PH (EG CUTOFF FREQUENCIES 1/PL AND 1/PH).
!
    coef(:k) = bd_coef( PL=pl, PH=ph, K=k )
!
!   NOW COMPUTE THE TRANSFERT FUNCTION freqr(:) OF THE LANCZOS FILTER AT THE FOURIER FREQUENCIES.
!
    call freq_func( NFREQ=n, COEF=coef(:k), FREQR=freqr(:n), FOUR_FREQ=true )
!
!   NOW, COMPUTE THE TRANSFERT FUNCTION DIRECTLY WITH THE FFT.
!
    kmid  = ( k + 1 )/2
    khalf = ( k - 1 )/2
!
    coef2(:n)          = zero
    coef2(:kmid)       = coef(kmid:k)
    coef2(n-khalf+1:n) = coef(1:khalf)
!
!   FIRST INITIALIZE THE FFT SUBROUTINE.
!
    call init_fft( n )
!
!   TRANSFORM THE TIME SERIES.
!
    coefc(:n) = cmplx( coef2(:n), zero, kind=stnd )
!
    call fft( coefc(:n), FORWARD=true  )
!
    freqr2(:n) = real( coefc(:n),  kind=stnd )
!
!   DEALLOCATE WORK ARRAYS.
!
    call end_fft()
!
    deallocate( coef )
!
!   TEST THE ACCURACY OF THE RESULTS.
!
    err = maxval(abs(freqr(:n)-freqr2(:n)))/maxval(abs(freqr(:n)))
!
    if ( err<=sqrt(epsilon(err))  ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
!
! END OF PROGRAM ex1_freq_func
! ============================
!
end program ex1_freq_func
