program ex1_gchol_cmp2
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine GCHOL_CMP2
!   in module Lin_Procedures . 
!                                                                              
! LATEST REVISION : 18/06/2017
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, zero, c10, true, false, gchol_cmp2,  &
                         norm, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n=4000, m=n-10_i4b
!
    real(stnd), parameter :: fudge=c10
!
    character(len=*), parameter :: name_proc='Example 1 of gchol_cmp2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, eps, tol, d1, elapsed_time
    real(stnd), dimension(:,:), allocatable :: a, a2, ainv, c, res
    real(stnd), dimension(:),   allocatable :: invdiag
!
    integer(i4b) :: j, krank
    integer      :: iok, istart, iend, irate, imax, itime
!
    logical(lgl)   :: do_test, upper
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : COMPUTING A SYMMETRIC GENERALIZED INVERSE OF A REAL n-BY-n SYMMETRIC SEMI-DEFINITE
!               POSITIVE MATRIX WITH THE CHOLESKY DECOMPOSITION.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    tol = sqrt( epsilon( err ) )
    eps = fudge*tol
    err = zero
!
    do_test = true
    upper   = false
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( c(m,n), a(n,n), ainv(n,n), invdiag(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-by-n RANDOM SYMMETRIC POSITIVE SEMIDEFINITE MATRIX a .
!
    call random_number( c )
!
    a = matmul( transpose(c), c )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(n,n), res(n,n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF THE SYMMETRIC POSITIVE SEMIDEFINITE MATRIX a .
!
        a2(:n,:n) = a(:n,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE A GENERALIZED INVERSE OF A SYMMETRIC POSITIVE SEMIDEFINITE
!   MATRIX a BY USING THE CHOLESKY DECOMPOSITION OF a .
!
!   IF ON OUTPUT OF gchol_cmp2 d1 IS GREATER OR EQUAL TO ZERO
!   THEN THE SYMMETRIC MATRIX IS POSITIVE SEMIDEFINITE AND A
!   SYMMETRIC GENERALIZED INVERSE OF a HAS BEEN COMPUTED.
!
    call gchol_cmp2( a, invdiag, krank, d1, tol=tol, matinv=ainv, upper=upper, fill=true )
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( d1<zero ) then
!
!       ANORMAL EXIT FROM gchol_cmp2 SUBROUTINE, PRINT A WARNING.
!
        write (prtunit,*) 'Error in exit of GCHOL_CMP2 subroutine, d1=', d1
        write (prtunit,*)
!
    else if ( do_test ) then
!
!       CHECK THE IDENTITIES a*ainv*a = a AND  ainv*a*ainv = ainv ,
!       WHICH DEFINE THE GENERALIZED INVERSE OF a.
!
        res  = matmul(a2, matmul(ainv,a2)) - a2
        err1 = norm(res) / ( real(n,stnd)*norm(a2) )
!
        res = matmul(ainv, matmul(a2,ainv)) - ainv
        err2 = norm(res) / ( real(n,stnd)*norm(ainv) )
!
        err = max( err1, err2 )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
        deallocate( a, ainv, c, invdiag, a2, res )
    else
        deallocate( a, ainv, c, invdiag )
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. d1>=zero ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*)
    write (*,'(a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing a (generalized) inverse of a positive semi-definite symmetric matrix of size ', &
       n, ' is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_gchol_cmp2
! =============================
!
end program ex1_gchol_cmp2
