program ex1_lapack_gebrd
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines GEBRD and ORGBR
!   in LAPACK software . 
!                                                                              
! LATEST REVISION : 21/12/2015
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Lapack_interfaces, only : gebrd, orgbr
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c50, norm,    &
                         unit_matrix, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer, parameter :: prtunit = 6, m=3000, n=3000, k=min(m,n)
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of gebrd'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err3, err, eps, elapsed_time
    real(stnd)                              :: work2(1)
    real(stnd), allocatable, dimension(:)   :: d, e, tauq, taup, work, resid2
    real(stnd), allocatable, dimension(:,:) :: a, a2, resid, bd, pt, q
!
    integer      :: info_bd, info_q, info_pt, lwork, lwork_bd, lwork_q, lwork_pt,    &
                    iok, istart, iend, irate, imax, itime
    integer(i4b) :: l
!
    logical(lgl) :: do_test
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : BIDIAGONAL REDUCTION OF A REAL MATRIX.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    if ( m>=n ) then
        allocate( a(m,n), d(k), e(k-1_i4b), tauq(k), taup(k), pt(k,k), stat=iok )
    else
        allocate( a(m,n), d(k), e(k-1_i4b), tauq(k), taup(k), q(k,k), stat=iok )
    end if
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A m-by-n REAL RANDOM DATA MATRIX .
!
    call random_number( a )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(m,n), bd(k,k), resid(k,n), resid2(n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SAVE THE DATA MATRIX.
!
        a2(:m,:n) = a(:m,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE FIRST OPTIMAL WORKSPACE FOR gebrd AND orgbr SUBROUTINES.
!
    lwork = -1
!
    call gebrd( m, n, a, m, d, e, tauq, taup, work2, lwork, info=info_bd )
!
    lwork_bd = int(work2(1))
!
    if ( m>=n ) then
!
        call orgbr( 'Q', m, n, n, a, m, tauq, work2, lwork, info=info_q )
!
        lwork_q = int(work2(1))
!
        call orgbr( 'P', n, n, m, pt, n, taup, work2, lwork, info=info_pt )
!
        lwork_pt = int(work2(1))
!
    else
!
        call orgbr( 'Q', m, m, n, q, m, tauq, work2, lwork, info=info_q )
!
        lwork_q = int(work2(1))
!
        call orgbr( 'P', m, n, m, a, m, taup, work2, lwork, info=info_pt )
!
        lwork_pt = int(work2(1))
!
    end if
!
    if ( min(info_bd,info_q,info_pt)==0 ) then
!
        lwork = max( lwork_bd, lwork_q, lwork_pt )
!
!       ALLOCATE OPTIMAL WORK VARIABLE NEEDED BY gebrd AND orgbr SUBROUTINES.
!
        allocate( work(lwork), stat = iok    )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE BIDIAGONAL REDUCTION OF RANDOM DATA MATRIX
!
!                        a = Q*BD*P**(t) .
!
!       WHERE Q AND P ARE ORTHOGONAL MATRICES AND BD IS A REAL BIDIAGONAL MATRIX.
!
        call gebrd( m, n, a, m, d, e, tauq, taup, work(:lwork_bd), lwork_bd, info=info_bd )
!
!       ON EXIT OF gebrd :
!
!         info= 0     :  INDICATES SUCCESSFUL EXIT
!         info=-i < 0 :  INDICATES THAT THE iTH ARGUMENT HAS AN ILLEGAL VALUE
!
!         d AND e ARE THE DIAGONAL AND OFF-DIAGONAL ELEMENTS OF THE BIDAGONAL MATRIX BD, RESPECTIVELY.
!         a CONTAINS THE ELEMENTARY REFLECTORS WHICH DEFINE Q AND P.
!         taup AND tauq ARE THE SCALAR FACTORS OF THE ELEMENTARY REFLECTORS WHICH DEFINE Q AND P, RESPECTIVELY.
!
        if ( m>=n ) then
!
            pt(:n,:n) = a(:n,:n)
!
            call orgbr( 'Q', m, n, n, a, m, tauq, work(:lwork_q), lwork_q, info=info_q )
!
            call orgbr( 'P', n, n, m, pt, n, taup, work(:lwork_pt), lwork_pt, info=info_pt )
!
!           ON EXIT OF THESE orgbr CALLS:
!
!           info= 0     :  INDICATES SUCCESSFUL EXIT
!           info=-i < 0 :  INDICATES THAT THE iTH ARGUMENT HAS AN ILLEGAL VALUE
!
!           IF m>=n,    a IS OVERWRITTEN WITH THE FIRST min(m,n)
!                       COLUMNS OF Q;
!                       pt CONTAINS THE n-BY-n ORTHOGONAL MATRIX P**(t).
!
        else
!
            q(:m,:m) = a(:m,:m)
!
            call orgbr( 'Q', m, m, n, q, m, tauq, work(:lwork_q), lwork_q, info=info_q )
!
            call orgbr( 'P', m, n, m, a, m, taup, work(:lwork_pt), lwork_pt, info=info_pt )
!
!           ON EXIT OF orgbr CALLS:
!
!           info= 0     :  INDICATES SUCCESSFUL EXIT
!           info=-i < 0 :  INDICATES THAT THE iTH ARGUMENT HAS AN ILLEGAL VALUE
!
!           IF m<n,     a IS OVERWRITTEN WITH THE FIRST min(m,n)
!                       ROWS OF P**(t);
!                       q CONTAINS THE m-BY-m ORTHOGONAL MATRIX Q.
!
        end if
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate ( work )
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( min(info_bd,info_q,info_pt)/=0 ) then
!
!       ANORMAL EXIT FROM gebrd OR orgbr SUBROUTINES, PRINT A WARNING.
!
        write (prtunit,*) 'Error in the calls to GEBRD or ORGBR subroutines, Info=', min(info_bd,info_q,info_pt)
!                
    else if ( do_test ) then
!
        bd(:k,:k) = zero
!
        if ( m>=n ) then
!
!           BD IS UPPER BIDIAGONAL.
!
            do l = 1_i4b, n-1_i4b
                bd(l,l)       = d(l)
                bd(l,l+1_i4b) = e(l)
            end do
!
            bd(n,n) = d(n)
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION Q**(t)*a - BD*P**(t),
!
            resid(:n,:n) = matmul( transpose(a(:m,:n)), a2(:m,:n) )   &
                           - matmul( bd(:n,:n), pt(:n,:n) )
!
            resid2(:n) = norm( resid(:n,:n), dim=2_i4b )
            err1 = norm( resid2(:n) )/(norm( a2 )*real(n,stnd) )
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - Q**(t)*Q.
!
            call unit_matrix( a2(:n,:n) )
!
            resid(:n,:n) = abs( a2(:n,:n) - matmul( transpose(a(:m,:n)), a(:m,:n) ) )
            err2 = maxval( resid(:n,:n) )/real(m,stnd)
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - P**(t)*P.
!
            resid(:n,:n) = abs( a2(:n,:n) - matmul( pt(:n,:n), transpose(pt(:n,:n)) ) )
            err3 = maxval( resid(:n,:n) )/real(n,stnd)
!
        else
!
!           BD IS LOWER BIDIAGONAL.
!
            do l = 1_i4b, m-1_i4b
                bd(l,l)       = d(l)
                bd(l+1_i4b,l) = e(l)
            end do
!
            bd(m,m) = d(m)
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION Q**(t)*a - BD*P**(t),
!
            resid(:m,:n) = matmul( transpose(q(:m,:m)), a2(:m,:n) )  &
                           - matmul( bd(:m,:m), a(:m,:n) )
!
            resid2(:n) = norm( resid(:m,:n), dim=2_i4b )
            err1 = maxval( resid2(:n) )/(norm( a2 )*real(m,stnd) )
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - Q**(t)*Q.
!
            call unit_matrix( a2(:m,:m) )
!
            resid(:m,:m) = abs( a2(:m,:m) - matmul( transpose(q(:m,:m )), q(:m,:m ) ) )
            err2 = maxval( resid(:m,:m) )/real(m,stnd)
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - P**(t)*P.
!
            resid(:m,:m) = abs( a2(:m,:m) - matmul( a(:m,:n), transpose(a(:m,:n)) ) )
            err3 = maxval( resid(:m,:m) )/real(n,stnd)
!
        endif
!
        err = max( err1, err2, err3 )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( m>=n ) then
        deallocate( a, pt, d, e, tauq, taup )
    else
        deallocate( a, q, d, e, tauq, taup )
    end if
!
    if ( do_test ) then
        deallocate( a2, bd, resid, resid2 )
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. min(info_bd,info_q,info_pt)==0 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test .and. min(info_bd,info_q,info_pt)==0 ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed bidiagonal decomposition a = Q*BD*P**(t) = ', err1
        write (prtunit,*) 'Orthogonality of the computed Q orthogonal matrix                 = ', err2
        write (prtunit,*) 'Orthogonality of the computed P orthogonal matrix                 = ', err3
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing the bidiagonal reduction of a ', &
       m, ' by ', n,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_lapack_gebrd
! ===============================
!
end program ex1_lapack_gebrd
