program ex1_lapack_orgbr
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines GEBRD and ORGBR
!   in LAPACK software . 
!                                                                              
!                                                     
! Further Details
! ===============
!
!    The program also shows the use of subroutines BD_SVD2 in module SVD_Procedures.
!                                                                              
! LATEST REVISION : 20/03/2017
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Lapack_interfaces, only : gebrd, orgbr
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c50, norm,    &
                         unit_matrix, bd_svd2, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer, parameter :: prtunit = 6, m=3000, n=3000, k=min(m,n)
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of orgbr'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err3, err, eps, elapsed_time
    real(stnd)                              :: work2(1)
    real(stnd), allocatable, dimension(:)   :: d, e, tauq, taup, work, resid2
    real(stnd), allocatable, dimension(:,:) :: a, a2, resid, pt, q
!
    integer      :: info_bd, info_q, info_pt, lwork, lwork_bd, lwork_q, lwork_pt,    &
                    iok, istart, iend, irate, imax, itime
    integer(i4b) :: l
!
    logical(lgl) :: bd_is_upper, failure, do_test
!   
    character    :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : BIDIAGONAL REDUCTION AND SVD DECOMPOSITION OF A REAL MATRIX.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
    do_test = true
!
    bd_is_upper = m>=n
!
!   ALLOCATE WORK ARRAYS.
!
    if ( bd_is_upper ) then
        allocate( a(m,n), d(k), e(k), tauq(k), taup(k), pt(k,k), stat=iok )
    else
        allocate( a(m,n), d(k), e(k), tauq(k), taup(k), q(k,k), stat=iok )
    end if
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A m-by-n REAL RANDOM DATA MATRIX .
!
    call random_number( a )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(m,n), resid(m,k), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SAVE THE DATA MATRIX.
!
        a2(:m,:n) = a(:m,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE FIRST OPTIMAL WORKSPACE FOR gebrd AND orgbr SUBROUTINES.
!
    lwork = -1
!
    call gebrd( m, n, a, m, d, e(2_i4b:k), tauq, taup, work2, lwork, info=info_bd )
!
    lwork_bd = int(work2(1))
!
    if ( bd_is_upper ) then
!
        call orgbr( 'Q', m, n, n, a, m, tauq, work2, lwork, info=info_q )
!
        lwork_q = int(work2(1))
!
        call orgbr( 'P', n, n, m, pt, n, taup, work2, lwork, info=info_pt )
!
        lwork_pt = int(work2(1))
!
    else
!
        call orgbr( 'Q', m, m, n, q, m, tauq, work2, lwork, info=info_q )
!
        lwork_q = int(work2(1))
!
        call orgbr( 'P', m, n, m, a, m, taup, work2, lwork, info=info_pt )
!
        lwork_pt = int(work2(1))
!
    end if
!
    if ( min(info_bd,info_q,info_pt)==0 ) then
!
        lwork = max( lwork_bd, lwork_q, lwork_pt )
!
!       ALLOCATE OPTIMAL WORK VARIABLE NEEDED BY gebrd AND orgbr SUBROUTINES.
!
        allocate( work(lwork), stat = iok    )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE BIDIAGONAL REDUCTION OF RANDOM DATA MATRIX
!
!                        a = Q*BD*P**(t) .
!
!       WHERE Q AND P ARE ORTHOGONAL MATRICES AND BD IS A REAL BIDIAGONAL MATRIX.
!
        call gebrd( m, n, a, m, d, e(2_i4b:k), tauq, taup, work(:lwork_bd), lwork_bd, info=info_bd )
!
!       ON EXIT OF gebrd :
!
!         info= 0     :  INDICATES SUCCESSFUL EXIT
!         info=-i < 0 :  INDICATES THAT THE iTH ARGUMENT HAS AN ILLEGAL VALUE
!
!         d AND e ARE THE DIAGONAL AND OFF-DIAGONAL ELEMENTS OF THE BIDAGONAL MATRIX BD, RESPECTIVELY. BD IS UPPER
!         BIDIAGONAL IF m>=n AND LOWER-BIDIAGONAL IF m<n .
!         a CONTAINS THE ELEMENTARY REFLECTORS WHICH DEFINE Q AND P.
!         taup AND tauq ARE THE SCALAR FACTORS OF THE ELEMENTARY REFLECTORS WHICH DEFINE Q AND P, RESPECTIVELY.
!
!       NOW COMPUTE THE SINGULAR VALUE DECOMPOSITION (SVD) OF THE REAL
!       m-BY-n MATRIX a. THE SVD IS WRITTEN
!
!                       a = U * S * V**(t)
!
!       WHERE S IS AN m-BY-n MATRIX WHICH IS ZERO EXCEPT FOR ITS
!       min(m,n) DIAGONAL ELEMENTS, U IS AN m-BY-m ORTHOGONAL MATRIX, AND
!       V IS AN n-BY-n ORTHOGONAL MATRIX.  THE DIAGONAL ELEMENTS OF S
!       ARE THE SINGULAR VALUES OF a; THEY ARE REAL AND NON-NEGATIVE.
!       THE COLUMNS OF U AND V ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a.
!
        if ( bd_is_upper ) then
!
            pt(:n,:n) = a(:n,:n)
!
            call orgbr( 'Q', m, n, n, a, m, tauq, work(:lwork_q), lwork_q, info=info_q )
!
            call orgbr( 'P', n, n, m, pt, n, taup, work(:lwork_pt), lwork_pt, info=info_pt )
!
!           ON EXIT OF THESE orgbr CALLS:
!
!           info= 0     :  INDICATES SUCCESSFUL EXIT
!           info=-i < 0 :  INDICATES THAT THE iTH ARGUMENT HAS AN ILLEGAL VALUE
!
!           IF m>=n,    a IS OVERWRITTEN WITH THE FIRST min(m,n)
!                       COLUMNS OF Q;
!                       pt CONTAINS THE n-BY-n ORTHOGONAL MATRIX P**(t).
!
!           COMPUTE ALL SINGULAR VALUES AND VECTORS OF THE REAL MATRIX a. THE SINGULAR
!           VALUES ARE ALSO THE SINGULAR VALUES OF THE UPPER BIDIAGONAL MATRIX BD.
!
            call bd_svd2( bd_is_upper, d, e, failure, a, pt, sort=sort )
!
!           THE ROUTINE RETURNS THE SINGULAR VALUES, THE LEFT AND RIGHT
!           SINGULAR VECTORS. THE RIGHT SINGULAR VECTORS ARE RETURNED ROWWISE:
!
!           IF m>=n,  a IS OVERWRITTEN WITH THE FIRST min(m,n)
!                     COLUMNS OF U (THE LEFT SINGULAR VECTORS,
!                     STORED COLUMNWISE);
!                     pt CONTAINS THE n-BY-n ORTHOGONAL MATRIX V**(t).
!
!           d IS OVERWRITTEN WITH THE SINGULAR VALUES OF a IN DECREASING ORDER.
!
        else
!
            q(:m,:m) = a(:m,:m)
!
            call orgbr( 'Q', m, m, n, q, m, tauq, work(:lwork_q), lwork_q, info=info_q )
!
            call orgbr( 'P', m, n, m, a, m, taup, work(:lwork_pt), lwork_pt, info=info_pt )
!
!           ON EXIT OF orgbr CALLS:
!
!           info= 0     :  INDICATES SUCCESSFUL EXIT
!           info=-i < 0 :  INDICATES THAT THE iTH ARGUMENT HAS AN ILLEGAL VALUE
!
!           IF m<n,     a IS OVERWRITTEN WITH THE FIRST min(m,n)
!                       ROWS OF P**(t);
!                       q CONTAINS THE m-BY-m ORTHOGONAL MATRIX Q.
!
!           COMPUTE ALL SINGULAR VALUES AND VECTORS OF THE REAL LOWER BIDIAGONAL MATRIX BD,
!           WHICH ARE ALSO THE SINGULAR VALUES AND VECTORS OF a .
!
            call bd_svd2( bd_is_upper, d, e, failure, q, a, sort=sort )
!
!           THE ROUTINE RETURNS THE SINGULAR VALUES, THE LEFT AND RIGHT
!           SINGULAR VECTORS. THE RIGHT SINGULAR VECTORS ARE RETURNED ROWWISE:
!
!           IF m<n,   a IS OVERWRITTEN WITH THE FIRST min(m,n)
!                     ROWS OF V**(t) (THE RIGHT SINGULAR VECTORS,
!                     STORED ROWWISE);
!                     q CONTAINS THE m-BY-m ORTHOGONAL MATRIX U.
!
!           d IS OVERWRITTEN WITH THE SINGULAR VALUES OF a IN DECREASING ORDER.
!
        end if
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate ( work )
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( min(info_bd,info_q,info_pt)/=0 ) then
!
!       ANORMAL EXIT FROM gebrd OR orgbr SUBROUTINES, PRINT A WARNING.
!
        write (prtunit,*) 'Error in the calls to GEBRD or ORGBR subroutines, Info=', min(info_bd,info_q,info_pt)
!                
    else if ( do_test ) then
!
        if ( bd_is_upper ) then
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a(:m,:n)*V(:n,:k) - U(:m,:k)*S(:k,:k).
!
            resid(:m,:k) = matmul(a2(:m,:k),transpose(pt(:k,:k))) - a(:m,:k)*spread(d,dim=1,ncopies=m)
            a2(:k,1_i4b) = norm( resid(:m,:k), dim=2_i4b )
            err1 =  maxval( a2(:k,1_i4b) )/( sum( abs(d) )*real(m,stnd) )
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U**(t)*U.
!
            call unit_matrix( a2(:n,:n) )
!
            resid(:n,:n) = abs( a2(:n,:n) - matmul( transpose(a(:m,:n)), a(:m,:n) ) )
            err2 = maxval( resid(:n,:n) )/real(m,stnd)
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - V**(t)*V.
!
            resid(:n,:n) = abs( a2(:n,:n) - matmul( pt(:n,:n), transpose(pt(:n,:n)) ) )
            err3 = maxval( resid(:n,:n) )/real(n,stnd)
!
        else
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a(:m,:n)*V(:n,:k) - U(:m,:k)*S(:k,:k).
!
            resid(:m,:k) =  matmul(a2(:m,:n),transpose(a(:k,:n))) - q(:k,:k)*spread(d,dim=1,ncopies=k) 
            a2(:k,1_i4b) = norm( resid(:m,:k), dim=2_i4b )
            err1 =  maxval( a2(:k,1_i4b) )/( sum( abs(d) )*real(m,stnd) )
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U**(t)*U.
!
            call unit_matrix( a2(:m,:m) )
!
            resid(:m,:m) = abs( a2(:m,:m) - matmul( transpose(q(:m,:m )), q(:m,:m ) ) )
            err2 = maxval( resid(:m,:m) )/real(m,stnd)
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - V**(t)*V.
!
            resid(:m,:m) = abs( a2(:m,:m) - matmul( a(:m,:n), transpose(a(:m,:n)) ) )
            err3 = maxval( resid(:m,:m) )/real(n,stnd)
!
        endif
!
        err = max( err1, err2, err3 )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( m>=n ) then
        deallocate( a, pt, d, e, tauq, taup )
    else
        deallocate( a, q, d, e, tauq, taup )
    end if
!
    if ( do_test ) then
        deallocate( a2, resid )
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. min(info_bd,info_q,info_pt)==0 .and. .not.failure ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test .and. min(info_bd,info_q,info_pt)==0 ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed singular triplets           = ', err1
        write (prtunit,*) 'Orthogonality of the computed left singular vectors  = ', err2
        write (prtunit,*) 'Orthogonality of the computed right singular vectors = ', err3
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing all singular values and vectors of a ', &
       m, ' by ', n,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_lapack_orgbr
! ===============================
!
end program ex1_lapack_orgbr
