program ex1_lapack_ormtr
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines (x)SYTRD and (x)ORMTR
!   in LAPACK software . 
!                                                                              
!                                                     
! Further Details
! ===============
!
!    The program also shows the use of subroutines SYMTRID_QRI and TRID_INVITER
!    in module EIG_Procedures.
!                                                                              
! LATEST REVISION : 29/12/2015
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Lapack_interfaces, only : sytrd, ormtr
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c50, norm, unit_matrix,     &
                         symtrid_qri, trid_inviter, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer, parameter :: prtunit = 6, n=8000, nvec=8000
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of ormtr'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err, eps, elapsed_time
    real(stnd)                              :: work2(1)
    real(stnd), allocatable, dimension(:)   :: d, e, tau, eigval, e2, work, resid
    real(stnd), allocatable, dimension(:,:) :: a, eigvec, a2
!
    integer      :: info_trd, info_q, lwork, lwork_trd, lwork_q, iok,   &
                    istart, iend, irate, imax, itime
    integer(i4b) :: l, maxiter=2
!
    logical(lgl) :: failure, failure2, do_test
!   
    character    :: uplo, sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : TRIDIAGONAL REDUCTION OF A REAL SYMMETRIC MATRIX AND
!               EIGENVALUES AND, OPTIONALLY, EIGENVECTORS OF A REAL
!               SYMMETRIC MATRIX USING THE INVERSE ITERATION METHOD.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
    do_test = false
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,n), eigvec(n,nvec), d(n), e(n), tau(n-1_i4b),  &
              eigval(n), e2(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-by-n RANDOM DATA MATRIX AND
!   FROM IT A SELF-ADJOINT MATRIX a .
!
    call random_number( a )
    a = a + transpose( a ) 
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(n,n), resid(nvec), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SAVE THE SYMMETRIC MATRIX.
!
        a2(:n,:n) = a(:n,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   IF uplo='U' UPPER TRIANGLE OF a IS STORED,
!   IF uplo='L' LOWER TRIANGLE OF a IS STORED.
!
    uplo = 'U'
!
!   COMPUTE FIRST OPTIMAL WORKSPACE FOR sytrd AND ormtr SUBROUTINES.
!
    lwork = -1
!
    call sytrd( uplo, n, a, n, d, e, tau, work2, lwork, info=info_trd )
!
    lwork_trd = int(work2(1))
!
    call ormtr( 'L', uplo, 'N', n, nvec, a, n, tau, eigvec, n, work2, lwork, info=info_q )
!
    lwork_q = int(work2(1))
!
    if ( info_trd==0 .and. info_q==0 ) then
!
        lwork = max( lwork_trd, lwork_q )
!
!       ALLOCATE OPTIMAL WORK VARIABLE NEEDED BY sytrd AND ormtr SUBROUTINES.
!
        allocate( work(lwork), stat = iok    )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       CALL sytrd TO REDUCE THE MATRIX a TO TRIDIAGONAL FORM
!
!                      a = Q*TRID*Q**(t)
!
!       WHERE Q IS ORTHOGONAL AND TRID IS A SYMMETRIC TRIDIAGONAL MATRIX.
!
        call sytrd( uplo, n, a, n, d, e, tau, work(:lwork_trd), lwork_trd, info=info_trd )
!
!       ON EXIT OF sytrd :
!
!         info= 0     :  INDICATES SUCCESSFUL EXIT
!         info=-i < 0 :  INDICATES THAT THE iTH ARGUMENT HAS AN ILLEGAL VALUE
!
!         d AND e ARE THE DIAGONAL AND OFF-DIAGONAL ELEMENTS OF THE SYMMETRIC TRIDIAGONAL MATRIX, RESPECTIVELY.
!         a CONTAINS THE ELEMENTARY REFLECTORS WHICH DEFINE Q IN ITS UPPER TRIANGLE IF uplo='U'.
!         tau CONTAINS THE SCALAR FACTORS OF THE ELEMENTARY REFLECTORS WHICH DEFINE Q.
!
!       MAKE A COPY OF THE TRIDIAGONAL MATRIX.
!
        eigval(:n) = d(:n)
        e2(:n)     = e(:n)
!
!       COMPUTE ALL THE EIGENVALUES OF THE SYMMETRIC TRIDIAGONAL MATRIX WITH SUBROUTINE symtrid_qri.
!
        call symtrid_qri( eigval, e2, failure, sort=sort )
!
        if ( .not.failure ) then
!
!           COMPUTE THE FIRST nvec EIGENVECTORS OF THE SELF-ADJOINT MATRIX a BY
!           maxiter INVERSE ITERATIONS ON THE INTERMEDIATE TRIDIAGONAL MATRIX
!           AND BACK-TRANSFORMATION WITH LAPACK SUBROUTINE ormtr.
!
            call trid_inviter( d(:n), e(:n), eigval(:nvec), eigvec(:n,:nvec), failure, maxiter=maxiter )
!
            call ormtr( 'L', uplo, 'N', n, nvec, a, n, tau, eigvec, n, work(:lwork_q), lwork_q, info=info_q )
!
!           ON EXIT OF THIS ormtr CALL:
!
!             info= 0     :  INDICATES SUCCESSFUL EXIT
!             info=-i < 0 :  INDICATES THAT THE iTH ARGUMENT HAS AN ILLEGAL VALUE
!
!             eigvec IS OVERWRITTEN WITH THE FIRST nvec EIGENVECTORS OF a STORED COLUMNWISE.
!
        end if
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate ( work )
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    failure2 = info_trd/=0 .or. info_q/=0
!
    if ( failure2 ) then
!
!       ANORMAL EXIT FROM sytrd OR ormtr SUBROUTINES, PRINT A WARNING.
!
        write (prtunit,*) 'Errors in the calls to SYTRD or ORMTR subroutines, Info=',  &
                          info_trd, info_q
!
        if ( info_trd<0 ) then
!
            write (prtunit,'(''Argument '',i3,'' has an illegal value in SYTRD'')') - info_trd
!
        end if
!
        if ( info_q<0 ) then
!
            write (prtunit,'(''Argument '',i3,'' has an illegal value in ORMTR'')') - info_q
!
        end if
!
    else if ( do_test .and. .not.failure ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*eigvec - eigvec*diag(eigval(:nvec))
!       WHERE eigval ARE THE EIGENVALUES AND eigvec THE EIGENVECTORS OF a.
!
        a(:n,:nvec) = matmul( a2, eigvec ) - eigvec*spread( eigval(:nvec), dim=1, ncopies=n )
!
        resid(:nvec) = norm( a(:n,:nvec), dim=2_i4b )
        err1 = maxval( resid(:nvec) )/( norm(a2)*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - eigvec**(t)*eigvec
!       WHERE eigvec are THE EIGENVECTORS OF THE MATRIX a.
!
        call unit_matrix( a(:nvec,:nvec) )
!
        a2(:nvec,:nvec) = abs( a(:nvec,:nvec) - matmul( transpose( eigvec ), eigvec ) )
        err2 = maxval( a2(:nvec,:nvec) )/real(n,stnd)
!
        err = max( err1, err2 )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
        deallocate( a, a2, eigvec, d, e, tau, eigval, e2, resid )
    else
        deallocate( a, eigvec, d, e, tau, eigval, e2 )
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure .and. .not.failure2 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test .and. .not.failure .and. .not.failure2 ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed eigenvalues and eigenvectors = ', err1
        write (prtunit,*) 'Orthogonality of the computed eigenvectors vectors    = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing all eigenvalues and', nvec, ' eigenvectors of a', &
       n, ' by', n,' real symmetric matrix is', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_lapack_ormtr
! ===============================
!
end program ex1_lapack_ormtr
