program ex1_lapack_posv
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine POSV
!   in LAPACK software . 
!                                                                              
! LATEST REVISION : 21/12/2015
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Lapack_interfaces, only : posv
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, norm, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer, parameter :: prtunit = 6, n=4000, m=n+10, nrhs=4000
!
    character(len=*), parameter :: name_proc='Example 1 of posv'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, elapsed_time
    real(stnd), dimension(:,:), allocatable :: a, b, c, x, res
!
    integer :: info, iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: do_test
!
    character :: uplo
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : SOLUTION OF A LINEAR SYSTEM WITH A REAL SYMMETRIC DEFINITE POSITIVE COEFFICIENT
!               MATRIX AND SEVERAL RIGHT HAND-SIDES.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = sqrt( epsilon( err ) )
!
    err = zero
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( c(m,n), a(n,n), b(n,nrhs), x(n,nrhs), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-by-n RANDOM REAL SYMMETRIC POSITIVE DEFINITE MATRIX a .
!
    call random_number( c )
!
    a = matmul( transpose(c), c )
!
!   GENERATE A n-by-nrhs RANDOM SOLUTION MATRIX x .
!
    call random_number( x )
!
!   COMPUTE THE MATRIX-MATRIX PRODUCT b = a*x .
!
    b = matmul( a, x )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE THE SOLUTION MATRIX FOR SYMMETRIC POSITIVE DEFINITE SYSTEM
!
!            a*x = b .
!
!   FIRST SPECIFY IF UPPER OR LOWER TRIANGLE OF a IS STORED .
!
    uplo = 'U'
!
    call posv( uplo, n, nrhs, a, n, b, n, info )
!
!   ON EXIT OF posv :
!
!      info    = 0 :  INDICATES SUCCESSFUL EXIT
!      info=-i < 0 :  INDICATES THAT the iTH ARGUMENT HAS AN ILLEGAL VALUE
!      info=i  > 0 :  INDICATES THAT THE SOLUTION CAN NOT BE COMPUTED BECAUSE
!                     THE MATRIX IS SINGULAR
!
!   THE ROUTINE RETURNS THE SOLUTION VECTORS IN b.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
!   CHECK THAT THE LAPACK ROUTINE HAS BEEN SUCCESSFUL.
!
    if ( info/=0 ) then
!
!       ANORMAL EXIT FROM posv SUBROUTINE, PRINT A WARNING.
!
        write (prtunit,*) 'Error in the call to POSV subroutine, Info=', info
        write (prtunit,*)
!
        if ( info<0 ) then
!
            write (prtunit,'(''Argument '',i3,'' has an illegal value'')') - info
!
        else if (info>0) then
!
            write (prtunit,'(''Zero diagonal value detected in upper ''//   &
                             ''triangular factor at position '',i7)') info
!
        end if
!
    else if ( do_test ) then
!
!       ALLOCATE WORK ARRAY .
!
        allocate( res(n,nrhs), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       CHECK THE RESULTS FOR SMALL RESIDUALS.
!
        res(:n,:nrhs) = b(:n,:nrhs) - x(:n,:nrhs)
        err = maxval( norm(res, dim=2_i4b ) /    &
                      norm(x,  dim=2_i4b  ) )/real(n,stnd)
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( res )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, b, c, x )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. info==0 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test .and. info==0 ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed solutions = ', err
    end if
!
    write (prtunit,*)
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for solving a positive definite symmetric linear system of size ', &
       n, ' with', nrhs, ' right hand side vectors is', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_lapack_posv
! ==============================
!
end program ex1_lapack_posv
