program ex1_lapack_stemr
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines (x)SYTRD, (x)STEMR
!   and (x)ORMTR in LAPACK software . 
!                                                                              
!                                                     
! Further Details
! ===============
!
!   The (x)STEMR subroutines will work properly only with IEEE arithmetic.                                                                        
!                                                     
! LATEST REVISION : 31/12/2015
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Lapack_interfaces, only : sytrd, ormtr, stemr
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c50, norm, unit_matrix,     &
                         merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer, parameter :: prtunit = 6, n=3000, neig=1000
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of stemr'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err, eps, vl, vu, elapsed_time
    real(stnd)                              :: work2(1)
    real(stnd), allocatable, dimension(:)   :: d, e, tau, w, work, resid
    real(stnd), allocatable, dimension(:,:) :: a, z, a2
!
    integer      :: info_trd, info_q, info_mrrr, lwork, liwork, lwork_trd, lwork_mrrr, lwork_q,   &
                    iok, istart, iend, irate, imax, itime, il, iu, m, nzc
    integer                            :: iwork2(1)
    integer, dimension(:), allocatable :: isuppz, iwork
!
    logical      :: tryrac
    logical(lgl) :: do_test, failure
!   
    character    :: uplo, jobz, range
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : TRIDIAGONAL REDUCTION OF A REAL SYMMETRIC MATRIX AND,
!               OPTIONALLY, neig EIGENVALUES AND EIGENVECTORS OF A REAL
!               SYMMETRIC MATRIX USING THE MRRR METHOD.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,n), z(n,neig), d(n), e(n), w(n),    &
              isuppz(2*n), tau(n-1_i4b), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-by-n RANDOM DATA MATRIX AND
!   FROM IT A SELF-ADJOINT MATRIX a .
!
    call random_number( a )
    a = a + transpose( a ) 
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(n,n), resid(neig), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SAVE THE SYMMETRIC MATRIX.
!
        a2(:n,:n) = a(:n,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   IF uplo='U' UPPER TRIANGLE OF a IS STORED,
!   IF uplo='L' LOWER TRIANGLE OF a IS STORED.
!
    uplo = 'U'
!
!   FIRST DETERMINE THE WORK TO DO FOR stemr SUBROUTINE.
!
!   IF jobz='N' COMPUTE EIGENVALUES ONLY,
!   IF jobz='V' COMPUTE EIGENVALUES AND EIGENVECTORS.
!
    jobz  = 'V'
!
!   IF range='A' ALL EIGENVALUES WILL BE FOUND,
!   IF range='V' ALL EIGENVALUES IN THE HALF OPEN INTERVAL (vl,vu) WILL BE FOUND,
!   IF range='I' THE il-TH THROUGH iu-TH EIGENVALUES IN ASCENDING ORDER WILL BE FOUND.
!
    range = 'I'
!
    vl = zero
    vu = zero
!
    il = n - neig + 1
    iu = n
!
!   tryrac = .true. INDICATES THAT THE CODE WILL TRY TO COMPUTE THE EIGENVALUES OF THE TRIDIAGONAL MATRIX
!                   TO HIGH RELATIVE ACCURACY IF POSSIBLE. IF tryrac IS SET TO .false. THE CODE IS NOT REQUIRED
!                   TO GARANTEE RELATIVELY ACCURATE EIGENVALUES AND CAN USE A FASTEST METHOD.
!
    tryrac = .true.
!
!   nzc IS AN UPPER BOUND FOR THE NUMBER OF EIGENVECTORS TO BE FOUND. nzc DEPENDS ON THE VALUE OF RANGE, IF:
!
!       range='A' THEN nzc>=n;
!       range='V' THEN nzc>=THE NUMBER OF EIGENVALUES IN (VL,VU],
!       range='I' THEN nzc>=iu-il+1 .
!
!   IF nzc=-1, THEN THE stemr SUBROUTINE WILL RETURN THE ESTIMATE OF nzc IN z(1,1) WITHOUT DOING ANY COMPUTATION.
!
    nzc = neig
!
!   NOW COMPUTE OPTIMAL WORKSPACE FOR sytrd, stemr AND ormtr SUBROUTINES.
!
    lwork = -1
!
    call sytrd( uplo, n, a, n, d, e, tau, work2, lwork, info=info_trd )
!
    lwork_trd = int(work2(1))
!
    call ormtr( 'L', uplo, 'N', n, neig, a, n, tau, z, n, work2, lwork, info=info_q )
!
    lwork_q = int(work2(1))
!
    liwork = -1
!
    call stemr( jobz, range, n, d, e, vl, vu, il, iu, m, w, z, n, nzc,    &
                isuppz, tryrac, work2, lwork, iwork2, liwork, info=info_mrrr    )
!
    lwork_mrrr = int(work2(1))
!
    if ( info_trd==0 .and. info_q==0 .and. info_mrrr==0 ) then
!
        lwork = max( lwork_trd, lwork_q, lwork_mrrr )
        liwork = iwork2(1)
!
!       ALLOCATE OPTIMAL WORK VARIABLES NEEDED BY sytrd, stemr AND ormtr SUBROUTINES.
!
        allocate( work(lwork), iwork(liwork), stat = iok    )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       CALL sytrd TO REDUCE THE MATRIX a TO TRIDIAGONAL FORM
!
!                      a = Q*TRID*Q**(t)
!
!       WHERE Q IS ORTHOGONAL AND TRID IS A SYMMETRIC TRIDIAGONAL MATRIX.
!
        call sytrd( uplo, n, a, n, d, e, tau, work(:lwork_trd), lwork_trd, info=info_trd )
!
!       ON EXIT OF sytrd :
!
!         info= 0     :  INDICATES SUCCESSFUL EXIT
!         info=-i < 0 :  INDICATES THAT THE iTH ARGUMENT HAS AN ILLEGAL VALUE
!
!         d AND e ARE THE DIAGONAL AND OFF-DIAGONAL ELEMENTS OF THE SYMMETRIC TRIDIAGONAL MATRIX, RESPECTIVELY.
!         a CONTAINS THE ELEMENTARY REFLECTORS WHICH DEFINE Q IN ITS UPPER TRIANGLE IF uplo='U'.
!         tau CONTAINS THE SCALAR FACTORS OF THE ELEMENTARY REFLECTORS WHICH DEFINE Q.
!
!
!       COMPUTE neig EIGENVALUES AND EIGENVECTORS OF THE SYMMETRIC TRIDIAGONAL MATRIX WITH SUBROUTINE stemr.
!       THE EIGENVALUE DECOMPOSITION OF A REAL n-BY-n SYMMETRIC TRIDIAGONAL MATRIX TRID
!       IS WRITTEN
!
!                       TRID = U * D * U**(t)
!
!       WHERE D IS AN n-BY-n DIAGONAL MATRIX AND U IS AN n-BY-n ORTHOGONAL MATRIX.
!       THE DIAGONAL ELEMENTS OF D ARE THE EIGENVALUES OF TRID; THEY ARE REAL.
!       THE COLUMNS OF U ARE THE EIGENVECTORS OF TRID.
!
        call stemr( jobz, range, n, d, e, vl, vu, il, iu, m, w, z, n, nzc, isuppz, tryrac,     &
                    work(:lwork_mrrr), lwork_mrrr, iwork, liwork, info=info_mrrr )
!
!       ON EXIT OF stemr :
!
!         info= 0     :  INDICATES SUCCESSFUL EXIT.
!         info=-i < 0 :  INDICATES THAT THE iTH ARGUMENT HAS AN ILLEGAL VALUE.
!         info=i  > 0 :  INDICATES THAT THE ALGORITHM FAILED TO CONVERGE.
!
!       ON EXIT OF stemr :
!
!         m GIVES THE TOTAL NUMBER OF EIGENVALUES/EIGENVECTORS FOUND.
!
!         z IS OVERWRITTEN WITH THE FIRST m EIGENVECTORS
!         OF THE TRIDIAGONAL MATRIX (THE EIGENVECTORS ARE STORED COLUMNWISE).
!
!         w IS OVERWRITTEN WITH THE EIGENVALUES OF THE TRIDIAGONAL MATRIX IN ASCENDING ORDER,
!         CORRESPONDING TO THE EIGENVECTORS IN z .
!
!       COMPUTE THE FIRST m EIGENVECTORS OF THE SELF-ADJOINT MATRIX a BY
!       BACK-TRANSFORMATION WITH LAPACK SUBROUTINE ormtr.
!
        call ormtr( 'L', uplo, 'N', n, m, a, n, tau, z, n, work(:lwork_q), lwork_q, info=info_q )
!
!       ON EXIT OF THIS ormtr CALL:
!
!           info= 0     :  INDICATES SUCCESSFUL EXIT
!           info=-i < 0 :  INDICATES THAT THE iTH ARGUMENT HAS AN ILLEGAL VALUE
!
!           z IS OVERWRITTEN WITH THE FIRST m EIGENVECTORS OF a STORED COLUMNWISE.
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate ( work, iwork )
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    failure = info_trd/=0 .or. info_q/=0 .or. info_mrrr/=0
!
    if ( failure ) then
!
!       ANORMAL EXIT FROM sytrd, stemr OR ormtr SUBROUTINES, PRINT A WARNING.
!
        write (prtunit,*) 'Errors in the calls to SYTRD, STEMR or ORMTR subroutines, Info=', &
                           info_trd, info_mrrr, info_q
!
        if ( info_trd<0 ) then
!
            write (prtunit,'(''Argument '',i3,'' has an illegal value in SYTRD'')') - info_trd
!
        end if
!
        if ( info_mrrr<0 ) then
!
            write (prtunit,'(''Argument '',i3,'' has an illegal value in STEMR'')') - info_mrrr
!
        end if
!
        if ( info_q<0 ) then
!
            write (prtunit,'(''Argument '',i3,'' has an illegal value in ORMTR'')') - info_q
!
        end if
!                
    else if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*eigvec - eigvec*diag(eigval(:m))
!       WHERE eigval ARE THE EIGENVALUES AND eigvec THE EIGENVECTORS OF a.
!
        a(:n,:m) = matmul( a2, z(:n,:m) ) - z(:n,:m)*spread( w(:m), dim=1, ncopies=n )
!
        resid(:m) = norm( a(:n,:m), dim=2_i4b )
        err1 = maxval( resid(:m) )/( norm(a2)*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - eigvec**(t)*eigvec
!       WHERE eigvec are THE EIGENVECTORS OF THE MATRIX a.
!
        call unit_matrix( a(:m,:m) )
!
        a2(:m,:m) = abs( a(:m,:m) - matmul( transpose( z(:n,:m) ), z(:n,:m) ) )
        err2 = maxval( a2(:m,:m) )/real(n,stnd)
!
        err = max( err1, err2 )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
        deallocate( a, z, d, e, tau, w, isuppz, a2, resid )
    else
        deallocate( a, z, d, e, tau, w, isuppz )
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test .and. .not.failure ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed eigenvalues and eigenvectors = ', err1
        write (prtunit,*) 'Orthogonality of the computed eigenvectors vectors    = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing ', m,' eigenvalues and eigenvectors of a', &
       n, ' by', n,' real symmetric matrix is', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_lapack_stemr
! ===============================
!
end program ex1_lapack_stemr
