program ex1_lapack_syevr
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine SYEVR
!   in LAPACK software . 
!                                                                              
! LATEST REVISION : 21/12/2015
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Lapack_interfaces, only : syevr
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, two, c50, norm,   &
                         merror, allocate_error, lamch, unit_matrix
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer, parameter :: prtunit = 6, n=3000
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of syevr'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, eps, abstol, vl, vu, elapsed_time
    real(stnd)                              :: work2(1)
    real(stnd), dimension(:),   allocatable :: w, work, resid2
    real(stnd), dimension(:,:), allocatable :: a, a2, z
!
    integer                            :: info, lwork, liwork, iok, il, iu, m,    &
                                          istart, iend, irate, imax, itime
    integer                            :: iwork2(1)
    integer, dimension(:), allocatable :: isuppz, iwork
!
    logical(lgl)   :: do_test
!
    character      :: jobz, uplo, range
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : EIGENVALUES AND, OPTIONALLY, EIGENVECTORS OF A REAL SYMMETRIC
!               MATRIX USING THE MRRR ALGORITHM.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    abstol = two*lamch('s')
    eps = fudge*epsilon( err )
    err = zero
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,n), z(n,n), w(n), isuppz(2*n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-by-n RANDOM DATA MATRIX AND FROM IT
!   A SELF-ADJOINT MATRIX a .
!
    call random_number( z )
    a = z + transpose( z ) 
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(n,n), resid2(n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF THE RANDOM SELF-ADJOINT MATRIX.
!
        a2(:n,:n) = a(:n,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   FIRST DETERMINE THE WORK TO DO.
!
!   IF jobz='N' COMPUTE EIGENVALUES ONLY,
!   IF jobz='V' COMPUTE EIGENVALUES AND EIGENVECTORS.
!
    jobz  = 'V'
!
!   IF range='A' ALL EIGENVALUES WILL BE FOUND,
!   IF range='V' ALL EIGENVALUES IN THE HALF OPEN INTERVAL (vl,vu) WILL BE FOUND,
!   IF range='I' THE il-TH THROUGH iu-TH EIGENVALUES IN ASCENDING ORDER WILL BE FOUND.
!
    range = 'A'
!
!   IF uplo='U' UPPER TRIANGLE OF a IS STORED,
!   IF uplo='L' LOWER TRIANGLE OF a IS STORED.
!
    uplo  = 'U'
!
    vl = zero
    vu = zero
!
    il = 1
    iu = n
!
!   COMPUTE FIRST OPTIMAL WORKSPACE FOR syevr SUBROUTINE.
!
    lwork  = -1
    liwork = -1
!
    call syevr( jobz, range, uplo, n, a, n, vl, vu, il, iu, abstol,        &
                m, w, z, n, isuppz, work2, lwork, iwork2, liwork, info     )
!
    if ( info==0 ) then
!
        lwork  = int(work2(1)) + 10
        liwork = iwork2(1)
!
!       ALLOCATE WORK VARIABLES NEEDED BY syevr SUBROUTINE.
!
        allocate( work(lwork), iwork(liwork), stat = iok    )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE EIGENVALUES AND EIGENVECTORS OF THE SELF-ADJOINT MATRIX a
!       WITH SUBROUTINE syevr FROM LAPACK.
!       THE EIGENVALUE DECOMPOSITION OF A REAL n-BY-n SYMMETRIC MATRIX a
!       IS WRITTEN
!
!                       a = U * D * U**(t)
!
!       WHERE D IS AN n-BY-n DIAGONAL MATRIX AND U IS AN n-BY-n ORTHOGONAL MATRIX.
!       THE DIAGONAL ELEMENTS OF D ARE THE EIGENVALUES OF a; THEY ARE REAL.
!       THE COLUMNS OF U ARE THE EIGENVECTORS OF a.
!
        call syevr( jobz, range, uplo, n, a, n, vl, vu, il, iu, abstol,      &
                    m, w, z, n, isuppz, work, lwork, iwork, liwork, info     )
!
!       ON EXIT OF syevr :
!
!         info= 0     :  INDICATES SUCCESSFUL EXIT.
!         info=-i < 0 :  INDICATES THAT THE iTH ARGUMENT HAS AN ILLEGAL VALUE.
!         info=i  > 0 :  INDICATES THAT THE ALGORITHM FAILED TO CONVERGE.
!
!       ON EXIT OF syevr :
!
!         THE LOWER (IF uplo='L') OR UPPER (IF uplo='U') TRIANGLE
!         OF a IS DESTROYED, INCLUDING THE DIAGONAL.
!
!         m GIVES THE TOTAL NUMBER OF EIGENVALUES/EIGENVECTORS FOUND.
!
!         z IS OVERWRITTEN WITH THE FIRST m EIGENVECTORS
!         OF a (THE EIGENVECTORS ARE STORED COLUMNWISE).
!
!         w IS OVERWRITTEN WITH THE EIGENVALUES OF a IN ASCENDING ORDER,
!         CORRESPONDING TO THE EIGENVECTORS IN z .
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate ( work, iwork )
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( info/=0 ) then
!
!       ANORMAL EXIT FROM syevr SUBROUTINE, PRINT A WARNING.
!
        write (prtunit,*) 'Error in the call to SYEVR subroutine, Info=', info
!
        if ( info<0 ) then
!
            write (prtunit,'(''Argument '',i3,'' has an illegal value'')') - info
!
        end if
!                
    else if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*U - U*D
!
        a(:n,:m) = matmul(a2(:n,:n),z(:n,:m)) - z(:n,:m)*spread(w(:m),dim=1,ncopies=n)
        resid2(:m) = norm( a(:n,:m), dim=2_i4b )
        err1 =  maxval( resid2(:m) )/( norm( a2 )*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U**(t)*U.
!
        call unit_matrix( a2(:m,:m) )
!
        a(:m,:m) = abs( a2(:m,:m) - matmul( transpose(z(:n,:m)), z(:n,:m) ) )
        err2 = maxval( a(:m,:m) )/real(n,stnd)
!
        err = max( err1, err2 )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
        deallocate( a, z, w, isuppz, a2, resid2 )
    else
       deallocate( a, z, w, isuppz )
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. info==0 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed eigenvalues and eigenvectors = ', err1
        write (prtunit,*) 'Orthogonality of the computed eigenvectors            = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing ', m, ' eigenvalues and eigenvectors of a ', &
       n, ' by ', n,' real symmetric matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_lapack_syevr
! ===============================
!
end program ex1_lapack_syevr
