program ex1_llsq_svd_solve
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine LLSQ_SVD_SOLVE
!   in module LLSQ_Procedures. 
!                                               
! LATEST REVISION : 04/09/2014
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c50, lamch, norm,        &
                         print_array, llsq_svd_solve, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n=2000, m=4000, p=min(m,n)
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of llsq_svd_solve'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, tol, rnorm, bnorm, cond, sfmin
    real(stnd), dimension(m)                :: b
    real(stnd), dimension(n)                :: x, sing_values
    real(stnd), allocatable, dimension(:)   :: b2, res
    real(stnd), allocatable, dimension(:,:) :: a, a2
!
    integer(i4b) :: krank
!
    integer      :: iok
!
    logical(lgl) :: failure, do_test, do_print
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : LINEAR LEAST SQUARES SYSTEM AND ONE RIGHT HAND-SIDE.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
    err      = zero
    do_test  = true
    do_print = false
!
!   ALLOCATE WORK ARRAY.
!
    allocate( a(m,n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A m-by-n REAL RANDOM COEFFICIENT MATRIX a .
!
    call random_number( a )
!
!   GENERATE A RANDOM RIGHT HAND-SIDE VECTOR b .
!
    call random_number( b )
!
    if ( do_print ) then
!
!       COMPUTE THE NORM OF DEPENDENT VARIABLE b .
!
        bnorm = norm( b )
!
    end if
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(m,n), b2(m), res(m), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SAVE DATA MATRIX .
!
        a2(:m,:n) = a(:m,:n)
!
!       SAVE RIGHT HAND SIDE VECTOR .
!
        b2(:m) = b(:m)
!
    end if
!
!   llsq_svd_solve COMPUTES THE MINIMUM NORM SOLUTION TO A REAL LINEAR LEAST
!   SQUARES PROBLEM :
!
!                       Minimize 2-norm(| b - A*x |)
!
!   USING THE SINGULAR VALUE DECOMPOSITION (SVD) OF A. A IS AN m-BY-n MATRIX
!   WHICH MAY BE RANK-DEFICIENT.
!
!   THE EFFECTIVE RANK OF A, krank,IS DETERMINED BY TREATING AS ZERO THOSE
!   SINGULAR VALUES WHICH ARE LESS THAN tol TIMES THE LARGEST SINGULAR VALUE.
!
    tol = 0.0000001_stnd
!
!   COMPUTE THE LEAST-SQUARES SOLUTION MATRIX OF a*x=b .
!
    call llsq_svd_solve( a, b, failure, x,                                            &
                         singvalues=sing_values, krank=krank, rnorm=rnorm, tol=tol )
!
    if ( do_test ) then
!
!       CHECK THAT THE RESIDUALS ARE ORTHOGONAL TO THE COLUMN VECTORS OF a .
!
        res = b2 - matmul( a2, x )
        err = sum(abs(matmul(res,a2)) )/ sum( abs(a2) )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a2, b2, res )
!
    end if
!
!   DEALLOCATE WORK ARRAY.
!
    deallocate( a )
!
    if ( err<=eps .and. .not.failure ) then
!
        write (prtunit,*) name_proc//' is correct'
!
        if ( do_print ) then
!
!           GET MACHINE CONSTANT sfmin, SUCH THAT 1/sfmin DOES NOT OVERFLOW.
!
            sfmin = lamch( 's' )
!
!           COMPUTE THE CONDITION NUMBER OF A IN THE 2-NORM
!
!                   singvalues(1)/singvalues(min(m,n)) .
!
            if ( sing_values(p)/sing_values(1)<=sfmin ) then
                cond = huge( a )
            else
                cond = sing_values(1)/sing_values(p)
            end if
!
!           PRINT RESULTS .
!
            write (prtunit,*)
            write (prtunit,*)
            write (prtunit,*) 'LEAST SQUARES SOLUTION VIA SINGULAR VALUE DECOMPOSITION'
            write (prtunit,*) '           MIN OF IIA*x-bII**2 FOR x                   '
            write (prtunit,*)
!
            call print_array( sing_values, title=' SINGULAR VALUES ASSOCIATED WITH MATRIX A ' )
!
            write (prtunit,*)
            write (prtunit,*) 'TOLERANCE FOR ZERO SINGULAR VALUE (tol*sing_values(1)):',tol*sing_values(1)
            write (prtunit,*)
            write (prtunit,*) 'CONDITION NUMBER OF A :',cond
            write (prtunit,*) 'RANK OF A             :',krank
            write (prtunit,*)
            write (prtunit,*) 'RESIDUAL SUM OF SQUARES IIA*x-bII**2               :',rnorm**2
            write (prtunit,*) 'RESIDUAL SUM OF SQUARES (%) (IIA*x-bII**2/IIbII**2):',(rnorm/bnorm)**2
!
            call print_array( x, title=' ASSOCIATED LEAST SQUARES SOLUTION x ' )
!
        end if
!
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
!
! END OF PROGRAM ex1_llsq_svd_solve
! =================================
!
end program ex1_llsq_svd_solve
