program ex1_lq_cmp
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines LQ_CMP and ORTHO_GEN_LQ
!   in modules QR_Procedures.
!                                                                              
! LATEST REVISION : 15/06/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, zero, one, c50, true, false, lq_cmp, ortho_gen_lq, norm, &
                         merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n=1000, m=1500
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of lq_cmp'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd), dimension(:,:), allocatable :: a, q, l, res
    real(stnd), dimension(:),   allocatable :: diagl, tau, res2, norma
    real(stnd)                              :: err, err1, err2, err3, eps, ulp, elapsed_time
!
    integer(i4b) :: k, j, p
    integer      :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: do_test
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : COMPUTING THE QL DECOMPOSITION OF A MATRIX.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( err )
    eps = fudge*ulp
    err = zero
!
    do_test = true
!
    k = min( m, n )
    p = max( m, n )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( q(p,m), l(n,k), diagl(k), tau(k), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM COEFFICIENT MATRIX a(:n,:m).
!
    call random_number( q(:n,:m) )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a(n,m), res(p,m), res2(n), norma(n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF THE RANDOM DATA MATRIX.
!
        a(:n,:m) = q(:n,:m)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE LQ DECOMPOSITION OF RANDOM DATA MATRIX.
!
    call lq_cmp( q(:n,:m), diagl(:k), tau(:k) )
!
!   lq_cmp COMPUTES AN ORTHOGONAL FACTORIZATION OF A REAL n-BY-m MATRIX
!   BY THE HOUSEOLDER METHOD. THE MATRIX IS ASSUMED OF FULL RANK. THE ROUTINE
!   COMPUTES A LQ FACTORIZATION OF a AS:
!
!                     a = l * q
!
!   ON EXIT, a HAS BEEN OVERWRITTEN BY DETAILS OF ITS ORTHOGONAL
!   FACTORIZATION. 
!
!   THE MATRIX q IS REPRESENTED AS A PRODUCT OF ELEMENTARY REFLECTORS
!
!            q = h(k)*h(k-1)* ... *h(1), WHERE k = min( size(a,1) , size(a,2) )
!
!   EACH h(i) HAS THE FORM
!
!            h(i) = I + BETA * ( V * V' ) ,
!                      
!   WHERE BETA IS A REAL SCALAR AND V IS A REAL M-ELEMENT VECTOR WITH V(1:i-1) = 0.
!   V(i:m) IS STORED ON EXIT IN a(i,i:m) AND BETA IN beta(i).
!                      
!   THE ELEMENTS BELOW THE DIAGONAL OF THE ARRAY a CONTAIN THE
!   CORRESPONDING ELEMENTS OF THE TRIANGULAR MATRIX l. THE ELEMENTS
!   OF THE DIAGONAL OF l ARE STORED IN THE ARRAY diagl. 
!
!   RESTORE TRIANGULAR FACTOR L OF LQ DECOMPOSITION OF RANDOM DATA MATRIX a
!   IN MATRIX l(:n,:k).
!
    do j = 1, k
!
        l(1:j-1,j) = zero
        l(j,j)     = diagl(j)
        l(j+1:n,j) = q(j+1:n,j)
!
    end do
!
!   GENERATE ORTHOGONAL MATRIX Q OF LQ DECOMPOSITION OF RANDOM DATA MATRIX a
!   AND AN ORTHOGONAL BASIS FOR THE ORTHOGONAL COMPLEMENT TO THE RANGE OF a'.
!   a IS ASSUMED OF FULL RANK.
!
    call ortho_gen_lq( q(:m,:m), tau(:k) )
!
!   ortho_gen_lq GENERATES AN m-BY-m REAL MATRIX WITH ORTHONORMAL ROWS, WHICH IS
!   DEFINED AS THE FIRST m ROWS OF A PRODUCT OF k ELEMENTARY REFLECTORS OF ORDER m
!
!            q = h(k)*h(k-1)* ... *h(1)
!
!   AS RETURNED BY lq_cmp.
!
!   THE SIZE OF beta DETERMINES THE NUMBER k OF ELEMENTARY REFLECTORS
!   WHOSE PRODUCT DEFINES THE MATRIX q.
!
!   NOW, THE ROWS OF q(:k,:m) ARE AN ORTHOGONAL BASIS FOR THE RANGE OF a(:n,:m)'
!   AND THE ORTHOGONAL BASIS FOR THE ORTHOGONAL COMPLEMENT TO THE RANGE OF a(:n,:m)'
!   ARE THR ROWS OF q(k+1:m,:m).
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!                
    if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a(:n,:m) - l(:n,:n)*q(:n,:m).
!
        res(:n,:m) = a(:n,:m) - matmul( l(:n,:k), q(:k,:m) )
        res2(:n)   = norm( res(:n,:m), dim=1_i4b )
        norma(:n)  = norm( a(:n,:m), dim=1_i4b )
        err1       = maxval( res2(:n) / norma(:n) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - Q*Q'.
!
        res(:m,:m) = matmul( q(:m,:m), transpose(q(:m,:m)) )
!
        do j = 1, m
            res(j,j) =  res(j,j) - one
        end do
!
        err2 = maxval( abs(res(:m,:m)) )/real(m,stnd)
!
!       CHECK ORTHOGONALITY OF a(:n,:m) AND ITS ORTHOGONAL COMPLEMENT q(n+1:m,:m).
!
        if ( m>n ) then
!
            res(:n,n+1_i4b:m) = matmul( a(:n,:m), transpose(q(n+1_i4b:m,:m) ) )
            err3 = maxval( abs( res(:n,n+1_i4b:m) ) )/real(m,stnd)
!
        else
!
            err3 = zero
!
        end if
!
        err = max( err1, err2, err3 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a, res, res2, norma )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( l, q, diagl, tau )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
!
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the LQ decomposition                                       = ', err1
        write (prtunit,*) 'Orthogonality of the Q matrix                                          = ', err2
!
        if ( m>n ) then
            write (prtunit,*) 'Orthogonality of the row-space of the matrix and its orthogonal complement = ', err3
        end if
!
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing LQ decomposition of a ', &
       n, ' by ', m,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_lq_cmp
! =========================
!
end program ex1_lq_cmp
