program ex1_matmul2
!
!
! Purpose
! =======
!
!   This program illustrates the use of function MATMUL2
!   in module Module_Utilities and compares its efficiency with the intrinsic MATMUL function. 
!                                                                              
! LATEST REVISION : 08/05/2015
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, norm, matmul2, merror, allocate_error
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n=4000, p=2000, m=2000
!
    character(len=*), parameter :: name_proc='Example 1 of matmul2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, elapsed_time1, elapsed_time2
    real(stnd), dimension(:,:), allocatable :: a, b, c, c2, resid
!
    integer :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: do_test, failure
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : MULTIPLICATION OF TWO REAL MATRICES.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = epsilon( err )
    err = zero
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,p), b(p,m), c(n,m), c2(n,m), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE THE RANDOM REAL MATRICES a AND b.
!
    call random_number( a(:n,:p) )
    call random_number( b(:p,:m) )
!
!   MULTIPLY THE TWO MATRICES WITH matmul2 FUNCTION.
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
    c2(:n,:m) = matmul2( a(:n,:p), b(:p,:m) )
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time1 = real( itime, stnd )/real( irate, stnd )
!
!   NOW RECOMPUTE THE MATRIX PRODUCT WITH matmul INTRINSIC FUNCTION.
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count=istart )
!
    c(:n,:m) = matmul( a(:n,:p), b(:p,:m) )
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time2 = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( resid(n,m), stat=iok )
!
        if ( iok/=0 ) then
           call merror( name_proc//allocate_error )
        end if
!
        resid(:n,:m) = abs( c2(:n,:m) - c(:n,:m) )
!
!       CHECK THE RESULTS.
!
        err = maxval( resid(:n,:m) )/norm( c )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a, b, c, c2, resid )
!
    else
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a, b, c, c2 )
!
    end if
!
!   CHECK AND PRINT THE RESULTS.
!
    if ( err<=eps ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for multiplying two real matrices of sizes ', n, ' by ', p, ' and ', p, ' by ', m,  &
       ' with matmul2() function is ', elapsed_time1, ' seconds'
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for multiplying two real matrices of sizes ', n, ' by ', p, ' and ', p, ' by ', m,  &
       ' with the intrinsic matmul() function is ', elapsed_time2, ' seconds'
!
!
! END OF PROGRAM ex1_matmul2
! ==========================
!
end program ex1_matmul2
