program ex1_pk_coef
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines/functions PK_COEF,
!   FREQ_FUNC and SYMLIN_FILTER in module Time_Series_Procedures.
!                                                                              
! LATEST REVISION : 12/11/2007
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, zero, half, one, two, pi, true, false, merror, allocate_error,   &
                         pk_coef, freq_func, symlin_filter, init_fft, fft, end_fft
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, n IS THE LENGTH OF THE TIME SERIES .
!
    integer(i4b), parameter :: prtunit=6, n=2000
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                            :: err, tmp, fc
    real(stnd), dimension(n)              :: y, y2, y3, freqr
    real(stnd), dimension(:), allocatable :: coef
!
    complex(stnd), dimension(n)           :: yc
!
    integer(i4b) :: i, k, k1, k2, pc, nfilt, khalf, kmid
!   
    integer :: iok
!   
!   
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='Example 1 of pk_coef'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   GENERATE A RANDOM REAL NUMBER SEQUENCE OF LENGTH n .
!
    call random_number( y(:n) )
!
!   SAVE THE REAL RANDOM NUMBER ARRAY.
!
    y2(:n) = y(:n)
!
!   pc IS THE PERIOD OF OSCILLATION WITH A PEAK RESPONSE NEAR ONE.
!   pc IS EXPRESSED IN NUMBER OF OBSERVATIONS, I.E. pc==32(96) SELECT
!   A PERIOD OF 8 YRS FOR QUARTERLY (MONTHLY) DATA.
!
    pc  = 32
!
!   COMPUTE THE CORRESPONDING CUTOFF FREQUENCY.
!
    fc = one/real( pc, stnd )
!
!   NOW SELECT THE OPTIMAL NUMBER OF FILTER COEFFICIENTS k FOR THE LANCZOS FILTER.
!
    tmp = 2.3*max( real( pc, stnd ), one/(half-fc) )
    k = ceiling( tmp, i4b )
!
!   CHECK IF k IS ODD.
!
    if ( (k/2)*2==k  ) k = k + 1
!
!   ALLOCATE WORK ARRAY FOR THE FILTER COEFFICIENTS.
!
    allocate( coef(k), stat=iok )
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   FUNCTION pk_coef COMPUTES THE K-TERM LEAST SQUARES APPROXIMATION TO AN
!   -IDEAL- BAND PASS FILTER WITH A PEAK RESPONSE NEAR ONE AT THE PERIOD pc (EG FREQUENCY FREQ=1/pc).
!
    coef(:k) = pk_coef( FREQ=fc, K=k )
!
!   SUBROUTINE symlin_filter PERFORMS A SYMMETRIC FILTERING OPERATION ON AN IMPUT
!   TIME SERIES (THE ARGUMENT VEC).
!
!   HERE symlin_filter FILTERS THE TIME SERIES .
!   NOTE THAT (size(COEF)-1)/2 DATA POINTS WILL BE LOST FROM EACH END OF THE SERIES SO THAT
!   NFILT (NFILT= size(VEC) - size(COEF) + 1) TIME OBSERVATIONS ARE RETURNED IN VEC(:NFILT)
!   AND THE REMAINING OBSERVATIONS ARE SET TO ZERO.
!
    call symlin_filter( VEC=y2(:n), COEF=coef(:k),  NFILT=nfilt )
!
!   NOW COMPUTE THE TRANSFERT FUNCTION freqr(:) OF THE LANCZOS FILTER AT THE FOURIER FREQUENCIES.
!
    call freq_func( NFREQ=n, COEF=coef(:k), FREQR=freqr(:n), FOUR_FREQ=true )
!
!   APPLY THE FILTER USING THE FFT AND THE CONVOLUTION THEOREM.
!
!   INITIALIZE THE FFT SUBROUTINE.
!
    call init_fft( n )
!
!   TRANSFORM THE TIME SERIES.
!
    yc(:n) = cmplx( y(1:n), zero, kind=stnd )
!
    call fft( yc(:n), forward=true  )
!
!   MULTIPLY THE FOURIER TRANSFORM OF THE TIME SERIES
!   BY THE TRANSFERT FUNCTION OF THE FILTER.
!
    yc(:n) = yc(:n)*freqr(:n)
!
!   INVERT THE SEQUENCE BACK TO GET THE FILTERED TIME SERIES.
!
    call fft( yc(:n), forward=false )
!
    y3(:n) = real( yc(:n),  kind=stnd )
!
!   DEALLOCATE WORK ARRAYS.
!
    call end_fft()
!
    deallocate( coef )
!
!   TEST THE ACCURACY OF THE FILTERING OPERATION.
!
    kmid  = ( k + 1 )/2
    khalf = ( k - 1 )/2
    k1    = kmid
    k2    = n - khalf
!
    err = maxval(abs(y3(k1:k2)-y2(:nfilt)))/maxval(abs(y(k1:k2)))
!
    if ( err<=sqrt(epsilon(err))  ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
!
! END OF PROGRAM ex1_pk_coef
! ==========================
!
end program ex1_pk_coef
