program ex1_power_spectrum
!
!
! Purpose
! =======
                                                                                                                                      !
!   This program is intended to demonstrate the use of subroutine POWER_SPECTRUM
!   in module Time_Series_Procedures .
!                                                                              
! LATEST REVISION : 19/03/2008
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, lgl, true, false, one, power_spectrum, comp_mvs, print_array
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, n IS THE LENGTH OF THE TIME SERIES AND MUST BE EVEN.
!
    integer(i4b), parameter :: prtunit=6, n=100, psn=(n/2)+1
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                   :: err_mean, err_var, xmean, xmean2, xvar, xvar2, xstd, eps, tmp
    real(stnd), dimension(n)     :: x
    real(stnd), dimension(psn,2) :: psx
!
    integer(i4b) :: trend, win, i
!
    logical(lgl) :: smooth, normpsd
!   
!   
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='Example 1 of power_spectrum'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
    eps = sqrt( epsilon(err_mean) )
!
!   GENERATE A RANDOM REAL NUMBER SEQUENCE OF LENGTH n .
!
    call random_number( x(:n) )
!
!   COMPUTE THE POWER SPECTRUM OF THE TIME SERIES x(:n).
!
!   BY DEFAULT, BARTLETT WINDOW IS USED IN THE COMPUTATION OF THE POWER SPECTRUM (i.e. WIN=1).
!   SET WIN=2 FOR RECTANGULAR WINDOW, WIN=3 FOR WELCH WINDOW, WIN=4 FOR HANNING WINDOW
!   OR WIN=5 FOR HAMMING WINDOW.
!   IN ANY CASE, WIN MUST BE GREATER OR EQUAL TO 1 AND LESS OR EQUAL TO 5.
!
    win = 2
!
!   BY DEFAULT, THE MEAN OF THE TIME SERIES IS REMOVED BEFORE THE COMPUTATION
!   OF THE POWER SPECTRUM (i.e. TREND=1).
!   SET TREND=2 FOR REMOVING THE DRIFT OR TREND=3 FOR REMOVING THE LEAST SQUARES LINE
!   FROM THE TIME SERIES BEFORE ESTIMATING THE POWER SPECTRUM.
!   FOR OTHER VALUES OF TREND NOTHING IS DONE BEFORE ESTIMATING THE SPECTRUM.
!
    trend = 0
!
!   ON ENTRY, IF NORMPSD IS SET TO TRUE, THE PSD ESTIMATES ARE NORMALIZED IN SUCH
!   A WAY THAT THE TOTAL AREA UNDER THE POWER SPECTRUM IS EQUAL TO THE VARIANCE OF
!   THE TIME SERIES VEC. IF NORMPSD IS SET TO FALSE, THE SUM OF THE PSD ESTIMATES
!   (e.g. sum( PSVEC(2:) ) IS EQUAL TO THE VARIANCE OF THE TIME SERIES.
!   THE DEFAULT IS NORMPSD=true .
!
    normpsd = false
!
!   ON EXIT, PSVEC CONTAINS THE POWER SPECTRAL DENSITY (PSD) ESTIMATES OF VEC
!   AT THE psn FOURIER FREQUENCIES.
!
    call power_spectrum( VEC=x(:n), PSVEC=psx(:psn,2), NORMPSD=normpsd,   &
                         WIN=win, TREND=trend )
!
!   BUILD UP PERIOD AXIS.
!
    psx(1,1) = -one
    tmp = real( n, stnd )
!
    do i = 1, psn-1
        psx(i+1,1) = tmp/real( i, stnd )
    end do
!
!   PRINT POWER SPECTRUM OF x(:n).
!
    call print_array( psx, title='POWER SPECTRUM', namcol=(/ "PERIOD", "PSD   "/) )
!
!   ESTIMATE THE MEAN AND VARIANCE OF THE SIGNAL THROUGH THE POWER SPECTRUM.
!
    xmean2 = sqrt( psx(1,2) )
    xvar2  = sum( psx(2:psn,2) )
!
!   COMPUTE THE MEAN AND THE VARIANCE WITH SUBROUTINE comp_mvs .
!
    call comp_mvs( X=x(:n), FIRST=true, LAST=true, XMEAN=xmean, XVAR=xvar, XSTD=xstd )
!
!   COMPARE THE TWO SETS OF STATISTICS.
!
    err_mean = abs( (xmean-xmean2)/xmean )
    err_var  = abs( (xvar-xvar2)/xvar )
!
!   TEST THE ACCURACY OF THE STATISTICS.
!
    if ( max(err_mean,err_var)<=eps  ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
!
! END OF PROGRAM ex1_power_spectrum
! =================================
!
end program ex1_power_spectrum
