program ex1_qr_cmp2
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines QR_CMP2 and 
!   ORTHO_GEN_QR in modules QR_Procedures.
!                                                                              
! LATEST REVISION : 02/06/2017
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, one, c50, qr_cmp2, ortho_gen_qr,   &
                         norm, unit_matrix, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!    
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit = 6, m=3000, n=300
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of qr_cmp2'
!
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, err3, eps, ulp, tol, elapsed_time
    real(stnd), allocatable, dimension(:)   :: diagr, beta, resid2, norma
    real(stnd), allocatable, dimension(:,:) :: a, q, r, resid
!
    integer(i4b)                            :: k, j, l, krank
    integer(i4b), allocatable, dimension(:) :: ip
    integer                                 :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: do_test
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : COMPUTE QR DECOMPOSITION WITH COLUMN PIVOTING
!               OF RANDOM DATA MATRIX.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( err )
    eps = fudge*ulp
    err = zero
!
!   SET TOLERANCE .
!
    tol = 0.000001_stnd
!
    krank = 0
!
    do_test = true
!
    k = min( m, n )
    l = max( m, n )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( r(k,n), q(m,l), diagr(k), beta(k), ip(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM COEFFICIENT MATRIX a(:m,:n) .
!
    call random_number( q(:m,:n) )
!
!   GENERATE A DEFICIENT COEFFICIENT MATRIX a(:m,:n) .
!
    j = min( n, 5_i4b ) 
    q(:m,j) = q(:m,1_i4b) + q(:m,2_i4b)
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a(m,n), resid(m,l), resid2(n), norma(n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF THE RANDOM DATA MATRIX.
!
        a(:m,:n) = q(:m,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE QR DECOMPOSITION WITH COLUMN PIVOTING OF RANDOM DATA MATRIX
!   WITH SUBROUTINE qr_cmp2.
!
    call qr_cmp2( q(:m,:n), diagr(:k), beta(:k), ip(:n), krank, tol=tol )
!
!   qr_cmp2 COMPUTES A (COMPLETE) ORTHOGONAL FACTORIZATION OF A REAL m-BY-n MATRIX.
!   THE MATRIX MAY BE RANK-DEFICIENT. THE ROUTINE FIRST COMPUTES A QR FACTORIZATION
!   WITH COLUMN PIVOTING OF a AS:
!
!                     a * p = q * r = q * [ r11 r12 ]
!                                         [  0  r22 ]
!
!   WITH r11 DEFINED AS THE LARGEST LEADING SUBMATRIX WHOSE ESTIMATED CONDITION
!   NUMBER, IN THE 1-NORM, IS LESS THAN 1/tol OR SUCH THAT ABS(r11[j,j])>0 IF 
!   tol IS ABSENT. THE ORDER OF r11, krank, IS THE EFFECTIVE RANK OF a. 
!
!   ON INPUT, krank=k, IMPLIES THAT THE FIRST k COLUMNS OF A ARE
!   TO BE FORCED INTO THE BASIS. PIVOTING IS PERFORMED ON THE LAST n-k
!   COLUMNS OF a.
!   WHEN krank>=min(m,n) IS USED, PIVOTING IS NOT PERFORMED. THIS IS
!   APPROPRIATE WHEN a IS KNOWN TO BE FULL RANK. krank<=0 CAN BE USED
!   WHEN PIVOTING IS DESIRED ON ALL COLUMNS OF a.
!   
!   IF THE OPTIONAL PARAMETER tau IS PRESENT, THEN r22 IS CONSIDERED TO BE NEGLIGIBLE
!   AND r12 IS ANNIHILATED BY ORTHOGONAL TRANSFORMATIONS FROM THE RIGHT,
!   ARRIVING AT THE COMPLETE ORTHOGONAL FACTORIZATION:
!  
!                     a * p = q * [ t11 0 ] * z
!                                 [  0  0 ]
!   
!   p IS A n-BY-n PERMUTATION MATRIX, q IS A m-BY-m ORTHOGONAL MATRIX,
!   r IS A m-BY-n UPPER TRIANGULAR MATRIX, t11 IS A krank-BY-krank UPPER
!   TRIANGULAR MATRIX AND z IS A n-BY-n ORTHOGONAL MATRIX.
!
!   ON EXIT, a HAS BEEN OVERWRITTEN BY DETAILS OF ITS (COMPLETE) 
!   ORTHOGONAL FACTORIZATION. 
!
!   THE MATRIX q IS REPRESENTED AS A PRODUCT OF ELEMENTARY REFLECTORS
!
!            q = h(1)*h(2)* ... *h(k), WHERE k = min( size(a,1) , size(a,2) )
!
!   EACH h(i) HAS THE FORM
!
!            h(i) = I + BETA * ( V * V' ) ,
!                      
!   WHERE BETA IS A REAL SCALAR AND V IS A REAL M-ELEMENT VECTOR WITH V(1:i-1) = 0.
!   V(i:m) IS STORED ON EXIT IN a(i:m,i) AND BETA IN beta(i).
!                      
!   IF THE OPTIONAL PARAMETER tau IS ABSENT :         
!                      
!      qr_cmp2 COMPUTES ONLY A QR FACTORIZATION WITH COLUMN PIVOTING OF a.
!                      
!      THE ELEMENTS ABOVE THE DIAGONAL OF THE ARRAY a CONTAIN THE
!      CORRESPONDING ELEMENTS OF THE TRIANGULAR MATRIX r. THE ELEMENTS
!      OF THE DIAGONAL OF r ARE STORED IN THE ARRAY diagr. 
!
!      ON EXIT, IF ip(j)=k, THEN THE jTH COLUMN OF a*p WAS 
!      THE kTH COLUMN OF a. THE MATRIX p IS REPRESENTED IN THE ARRAY 
!      ip AS FOLLOWS: IF ip(j) = i THEN THE jTH COLUMN OF p IS 
!      THE iTH CANONICAL UNIT VECTOR.
!                      
!      ON EXIT, krank CONTAINS THE EFFECTIVE RANK OF a, I.E., THE ORDER OF
!      THE SUBMATRIX r11.
!                      
!   IF THE OPTIONAL PARAMETER tau IS PRESENT :         
!                      
!      qr_cmp2 COMPUTES A COMPLETE ORTHOGONAL FACTORIZATION OF a FROM THE QR 
!      FACTORIZATION WITH COLUMN PIVOTING OF a .
!                      
!      THE FACTORIZATION IS OBTAINED BY HOUSEHOLDER'S METHOD. THE kTH TRANSFORMATION
!      MATRIX, z(k), WHICH IS USED TO INTRODUCE ZEROS INTO THE kTH ROW OF r,
!      IS GIVEN IN THE FORM
!
!             z(k) = ( I    0  ),
!                    ( 0  T(k) )
!
!      WHERE
!
!             T(k) = I + TAU * ( U(k) * U(k)' ) , U(k) = (   1  )
!                                                        (   0  )
!                                                        ( L(k) )
!
!      TAU IS A SCALAR AND L(k) IS AN (n-krank) ELEMENT VECTOR. TAU and L(k) ARE CHOSEN
!      TO ANNIHILATE THE ELEMENTS OF THE kTH ROW OF r12.
!
!      ON EXIT, THE SCALAR TAU IS RETURNED IN THE kTH ELEMENT OF tau AND THE VECTOR U(K)
!      IN THE kTH ROW OF a, SUCH THAT THE ELEMENTS OF L(k) ARE IN a(k,krank+1:n).
!
!      THE z n-BY-n ORTHOGONAL MATRIX WHICH IS APPLIED FROM THE RIGHT TO R IS 
!      GIVEN BY THE PRODUCT
!
!             z = z(1) * z(2) * ... * z(krank)
!
!      ON EXIT, THE ELEMENTS ABOVE THE DIAGONAL OF THE ARRAY SECTION a(1:krank,1:krank)
!      CONTAIN THE CORRESPONDING ELEMENTS OF THE TRIANGULAR MATRIX t11. THE ELEMENTS OF
!      THE DIAGONAL OF t11 ARE STORED IN THE ARRAY SECTION diagr(1:krank).
!                      
!      ON EXIT, IF ip(j)=k, THEN THE jTH COLUMN OF a*p WAS 
!      THE kTH COLUMN OF a. THE MATRIX p IS REPRESENTED IN THE ARRAY 
!      ip AS FOLLOWS: IF ip(j) = i THEN THE jTH COLUMN OF p IS 
!      THE iTH CANONICAL UNIT VECTOR.
!                      
!      ON EXIT, krank CONTAINS THE EFFECTIVE RANK OF a, I.E., THE ORDER OF
!      THE SUBMATRIX t11.
!                      
!   IF tol IS PRESENT AND IS IN [0,1[, THEN :
!       ON ENTRY, THE CALCULATIONS TO DETERMINE THE CONDITION NUMBER OF a
!       ARE PERFORMED. THEN, tol IS USED TO DETERMINE THE EFFECTIVE RANK
!       OF a, WHICH IS DEFINED AS THE ORDER OF THE LARGEST LEADING
!       TRIANGULAR SUBMATRIX r11 IN THE QR FACTORIZATION WITH PIVOTING OF a,
!       WHOSE ESTIMATED CONDITION NUMBER IS LESS THAN 1/tol. IF tol=0 IS SPECIFIED
!       THE NUMERICAL RANK OF a IS DETERMINED.
!       ON EXIT, THE RECIPROCAL OF THE CONDITION NUMBER IS RETURNED IN tol.
!
!   IF tol IS NOT SPECIFIED OR IS OUTSIDE [0,1[ :
!       THE CALCULATIONS TO DETERMINE THE CONDITION NUMBER OF a ARE NOT
!       PERFORMED AND CRUDE TESTS ON r(j,j) ARE DONE TO DETERMINE
!       THE NUMERICAL RANK OF a.
!
!   IF IT IS POSSIBLE THAT a MAY NOT BE FULL RANK (I.E., CERTAIN COLUMNS OF a ARE
!   LINEAR COMBINATIONS OF OTHER COLUMNS), THEN THE LINEARLY DEPENDENT COLUMNS CAN
!   USUALLY BE DETERMINED BY USING krank=0  AND tol=RELATIVE PRECISION OF THE ELEMENTS
!   IN a. IF EACH ELEMENT IS CORRECT TO, SAY, 5 DIGITS THEN tol=0.00001 SHOULD
!   BE USED. ALSO, IT MAY BE HELPFUL TO SCALE THE COLUMNS OF a SO THAT ALL ELEMENTS 
!   ARE ABOUT THE SAME ORDER OF MAGNITUDE.
!
!   NOW, RESTORE TRIANGULAR FACTOR r OF QR DECOMPOSITION OF RANDOM DATA MATRIX a
!   IN MATRIX r(:k,:n) .
!
    do j = 1_i4b, k
!
        r(1_i4b:j-1_i4b,j) = q(1_i4b:j-1_i4b,j)
        r(j,j)             = diagr(j)
        r(j+1_i4b:k,j)     = zero
!
    end do
!
    do j = k+1_i4b, n
!
        r(1_i4b:k,j) = q(1_i4b:k,j)
!
    end do
!
!   GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX a
!   AND AN ORTHOGONAL BASIS FOR THE ORTHOGONAL COMPLEMENT TO THE RANGE OF a .
!   a IS NOT ASSUMED OF FULL RANK.
!
    call ortho_gen_qr( q(:m,:m), beta(:krank) )
!
!   ortho_gen_qr GENERATES AN m-BY-m REAL MATRIX WITH ORTHONORMAL COLUMNS, WHICH IS
!   DEFINED AS THE FIRST m COLUMNS OF A PRODUCT OF k ELEMENTARY REFLECTORS OF ORDER m
!
!            q = h(1)*h(2)* ... *h(k)
!
!   AS RETURNED BY qr_cmp OR qr_cmp2.
!
!   THE SIZE OF beta DETERMINES THE NUMBER k OF ELEMENTARY REFLECTORS
!   WHOSE PRODUCT DEFINES THE MATRIX q.
!
!   NOW q IS IN q(:m,:krank) AND THE ORTHOGONAL BASIS FOR THE ORTHOGONAL COMPLEMENT
!   TO THE RANGE OF a IS IN q(:m,krank+1:m).
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!                
    if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a(:m,:n)*p(:n,:n) - q(:m,:krank)*r(:krank,:n).
!
        resid(:m,:n) = a(:m,ip(:n)) - matmul( q(:m,:krank), r(:krank,:n) )
        resid2(:n)   = norm( resid(:m,:n), dim=2_i4b )
        norma(:n)    = norm( a(:m,:n), dim=2_i4b )
        err1         = maxval( resid2(:n) / norma(:n) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - Q**(t)*Q.
!
        call unit_matrix( resid(:m,:m) )
!
        resid(:m,:m) = abs( resid(:m,:m) - matmul( transpose(q(:m,:m)), q(:m,:m) ) )
        err2 = maxval( resid(:m,:m) )/real(m,stnd)
!
!       CHECK ORTHOGONALITY OF a(:m,:krank) AND ITS ORTHOGONAL COMPLEMENT q(:m,krank+1:n).
!
        if ( m>krank ) then
!
            resid(:krank,krank+1:m) = matmul( transpose(a(:m,:krank)), q(:m,krank+1:m) )
            err3 = maxval( abs( resid(:krank,krank+1:m) ) )/real(m,stnd)
!
        else
!
            err3 = zero
!
        end if
!
        err = max( err1, err2, err3 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a, resid, resid2, norma )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( r, q, diagr, beta, ip )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
!
        write (prtunit,*) 
        write (prtunit,*) 'Rank of the matrix               = ', krank
        write (prtunit,*) 'Accuracy of the QR decomposition = ', err1
        write (prtunit,*) 'Orthogonality of the Q matrix    = ', err2
!
        if ( m>n ) then
            write (prtunit,*) 'Orthogonality of the range of the matrix&
                              & and its orthogonal complement = ', err3
        end if
!
    end if
!
    write (prtunit,*)
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing QR decomposition with column pivoting of a ', &
       m, ' by ', n,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_qr_cmp2
! ==========================
!
end program ex1_qr_cmp2
