program ex1_random_eig_pos_with_blas
!
!
! Purpose
! =======
!
!   This program illustrates how to compute an approximate eigenvalue decomposition
!   of a symmetric positive (semi-)definite matrix with randomized subspace iterations and/or
!   the Nystrom method using STATPACK and BLAS subroutines.
!                                                                              
!                                                     
! Further Details
! ===============
!
!   The program shows the use of subroutines SVD_CMP in module SVD_Procedures,
!   EIG_CMP in module Eig_Procedures, QR_CMP, ORTHO_GEN_QR in module QR_Procedures,
!   CHOL_CMP in module Lin_procedures, and NORMAL_RANDOM_NUMBER3_ and GEN_RANDOM_SYM_MAT
!   in module Random and gemm, symm, syrk generic interfaces in module BLAS_interfaces.
!                                                                              
! LATEST REVISION : 23/10/2020
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, one, seven, c30, merror, allocate_error, &
                         norm, unit_matrix, random_seed_, normal_random_number3_, chol_cmp,          &
                         eigval_sort, eig_abs_sort, qr_cmp, ortho_gen_qr, svd_cmp, eig_cmp,          &
                         gen_random_sym_mat
!
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
#ifdef _BLAS
!    use BLAS_interfaces, only : gemm
    use BLAS_interfaces, only : symm, syrk, gemm
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, neig IS THE TARGET RANK OF THE APPROXIMATE PARTIAL EVD DECOMPOSITION,
! neig0 IS THE RANK OF THE GENERATED MATRIX,
! n IS THE DIMENSION OF THE GENERATED SYMMETRIC POSITIVE SEMI-DEFINITE MATRIX.
!
    integer(i4b), parameter :: prtunit=6, n=2000, neig=5, neig0=1000
!
    character(len=*), parameter :: name_proc='Example 1 of random_eig_pos_with_blas'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd) :: d1, err, err1, err2, eps, elapsed_time, norma, tmp, relerr, relerr2
#ifdef _BLAS
    real(stnd), dimension(:,:), allocatable :: a, q, b, bt, v, eigvec
#else
    real(stnd), dimension(:,:), allocatable :: a, q, qt, b, bt, v, eigvec
#endif
    real(stnd), dimension(:),   allocatable :: diagr, beta, eigval, eigval0
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: p, np, i, j, niter, mat_type
!
    logical(lgl) :: failure, do_test, ortho, use_nystrom
!   
    character    :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : PARTIAL EIGENVALUE DECOMPOSITION OF A n-BY-n REAL SYMMETRIC POSITIVE
!               SEMI-DEFINITE MATRIX USING RANDOMIZED POWER SUBSPACE OR BLOCK KRYLOV
!               ITERATIONS AND THE NYSTROM METHOD.
!
!   SPECIFY THE TYPE OF INPUT MATRIX:
!
!   mat_type = 1  -> SLOW DECAY OF EIGENVALUES
!   mat_type = 2  -> FAST DECAY OF EIGENVALUES
!   mat_type = 3  -> S-SHAPED DECAY OF EIGENVALUES
!   mat_type > 3  -> VERY SLOW DECAY OF EIGENVALUES
!
    mat_type = 2_i4b
!
!   SET TOLERANCE FOR THE ERROR OF THE RELATIVE ERROR IN FROBENIUS NORM
!   OF THE COMPUTED PARTIAL EVD.
!
    eps = 0.001_stnd
!
!   SPECIFY IF NUMERICAL TESTS ARE PERFORMED TO TEST THE ACCURACY OF THE EIGEN COUPLETS.
!
    do_test = true
!
!   CHOOSE TUNING PARAMETERS FOR THE RANDOMIZED EVD ALGORITHM.
!
!   DETERMINE THE NUMBER OF POWER ITERATIONS niter TO BE PERFORMED.
!
    niter = 6_i4b
!
!   DETERMINE THE OVERSAMPLING SIZE p .
!
    p  = 20_i4b
!
!   CHECK VALIDITY OF THE OVERSAMPLING SIZE PARAMETER.
!
    np = min( p + neig, n )
!
!   SPECIFY IF ORTHONORMALIZATION IS CARRIED OUT BETWEEN EACH STEP
!   OF THE (POWER) SUBSPACE ITERATIONS, TO AVOID LOSS OF ACCURACY DUE
!   TO ROUNDING ERRORS. THIS IS NORMALLY NOT NEEDED FOR POSITIVE MATRIX.
!
    ortho = false
!
!   SPECIFY IF LAST STEP OF THE ALGORITHM IS PERFORMED WITH THE NYSTROM METHOD (AND A SVD)
!   OR AN EIGENVALUE DECOMPOSITION.
!
    use_nystrom = true
!
    d1 = one
!
!   ALLOCATE WORK ARRAYS.
!
#ifdef _BLAS
    allocate( a(n,n), q(n,np), b(n,np), v(np,np), diagr(np),                    &
              beta(np),  eigval0(neig0), eigval(neig), eigvec(n,neig), stat=iok )
#else
    allocate( a(n,n), q(n,np), qt(np,n), b(n,np), v(np,np), diagr(np),          &
              beta(np),  eigval0(neig0), eigval(neig), eigvec(n,neig), stat=iok )
#endif
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-BY-n RANDOM REAL SYMMETRIC POSITIVE SEMI-DEFINITE MATRIX a WITH A
!   SPECIFIED DISTRIBUTION OF EIGENVALUES.
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
!   GENERATE EIGENVALUES.
!
    select case( mat_type )
!
        case( 1_i4b )
!
!           SLOW DECAY OF EIGENVALUES.
!
            do i = 1_i4b, neig0
!
                tmp = real( i, stnd )
!
                eigval0(i) = one/( tmp*tmp )
!        
            end do
!
        case( 2_i4b )
!
!           FAST DECAY OF EIGENVALUES.
!
            do i = 1_i4b, neig0
!
                tmp = real( i, stnd )
!
                eigval0(i) = exp( -tmp/seven )
!        
            end do
!
        case( 3_i4b )
!
!           S-SHAPED DECAY OF EIGENVALUES.
!
            do i = 1_i4b, neig0
!
                tmp = real( i, stnd )
!
                eigval0(i) = 0.0001_stnd + one/( one + exp( tmp -c30 ) )
!        
            end do
!
        case default
!
!           VERY SLOW DECAY OF EIGENVALUES.
!
            norma = real( neig0 - 1_i4b, stnd )
!
            do i = 1_i4b, neig0
!
                tmp = real( i - 1_i4b, stnd )
!
                eigval0(i) = exp( -tmp/norma )
!        
            end do
!
    end select
!
!   SORT THE EIGENVALUES BY DECREASING MAGNITUDE.
!
    call eigval_sort( sort, eigval0(:neig0) )
!
!   GENERATE A SYMMETRIC POSITIVE SEMI-DEFINITE MATRIX a WITH THE SPECIFIED EIGENVALUES
!   AND RANK neig0.
!
    call gen_random_sym_mat( eigval0(:neig0), a )
!
!   COMPUTE THE FROBENIUS NORM OF THE SYMMETRIC POSITIVE SEMI-DEFINITE MATRIX.
!
!    norma = norm( a(:n,:n) )
    norma = sqrt(sum( eigval0(:neig0)**2 ) )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   GENERATE A n-BY-np IID GAUSSIAN MATRIX .
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING MATRIX
!   FORM OF SUBROUTINE normal_random_number3_.
!
!    call normal_random_number3_( b(:n,:np) )
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING VECTOR
!   FORM OF SUBROUTINE normal_random_number_.
!
    do i = 1_i4b, np
!
        call normal_random_number3_( b(:n,i) )
!
    end do
!
!   COMPUTE RANDOM SAMPLE MATRIX.
!
#ifdef _BLAS
!    call gemm( 'N', 'N', n, np, n, one, a(1_i4b:n,1_i4b:n), n,       &
!               b(1_i4b:n,1_i4b:np), n, zero, q(1_i4b:n,1_i4b:np), n  )
    call symm( 'L', 'U', n, np, one, a(1_i4b:n,1_i4b:n), n,          &
               b(1_i4b:n,1_i4b:np), n, zero, q(1_i4b:n,1_i4b:np), n  )
#else
    q(:n,:np) = matmul( a(:n,:n), b(:n,:np) )
#endif
!
!   DO POWER ITERATIONS.
!
    do i = 1_i4b, niter
!
        if ( ortho ) then
!
!           COMPUTE QR DECOMPOSITION OF RANDOM SAMPLE MATRIX TO OBTAIN AN
!           ORTHONORMAL BASIS OF RANDOM SUBSPACE.
!
            call qr_cmp( q(:n,:np), diagr(:np), beta(:np) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM SAMPLE MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( q(:n,:np), beta(:np) )
!
        end if
!
!       COMPUTE RANDOM SUBSPACE PROJECTION.
!
#ifdef _BLAS
!        call gemm( 'N', 'N', n, np, n, one, a(1_i4b:n,1_i4b:n), n,       &
!                   q(1_i4b:n,1_i4b:np), n, zero, b(1_i4b:n,1_i4b:np), n  )
        call symm( 'L', 'U', n, np, one, a(1_i4b:n,1_i4b:n), n,          &
                   q(1_i4b:n,1_i4b:np), n, zero, b(1_i4b:n,1_i4b:np), n  )
#else
        b(:n,:np) = matmul( a(:n,:n), q(:n,:np)  )
#endif
!
        q(:n,:np) = b(:n,:np)
!
    end do
!
!   COMPUTE QR DECOMPOSITION OF RANDOM SAMPLE MATRIX TO OBTAIN AN
!   ORTHONORMAL BASIS OF RANDOM SUBSPACE.
!
    call qr_cmp( q(:n,:np), diagr(:np), beta(:np) )
!
!   GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM SAMPLE MATRIX.
!   q IS ASSUMED OF FULL RANK.
!
    call ortho_gen_qr( q(:n,:np), beta(:np) )
!
!   COMPUTE b = a*q AND v = q**(t)*b .
!
#ifdef _BLAS
!    call gemm( 'N', 'N', n, np, n, one, a(1_i4b:n,1_i4b:n), n,       &
!               q(1_i4b:n,1_i4b:np), n, zero, b(1_i4b:n,1_i4b:np), n  )
    call symm( 'L', 'U', n, np, one, a(1_i4b:n,1_i4b:n), n,          &
               q(1_i4b:n,1_i4b:np), n, zero, b(1_i4b:n,1_i4b:np), n  )
!
     call gemm( 'T', 'N', np, np, n, one, q(1_i4b:n,1_i4b:np), n,       &
                b(1_i4b:n,1_i4b:np), n, zero, v(1_i4b:np,1_i4b:np), np  )
#else
    b(:n,:np) = matmul( a(:n,:n), q(:n,:np)  )
!
    qt(:np,:n) = transpose( q(:n,:np) )
!
    v(:np,:np) = matmul( qt(:np,:n), b(:n,:np) )
#endif
!
    if ( use_nystrom ) then
!
!       USE THE NYSTROM METHOD.
!
!       ALLOCATE WORK ARRAY.
!
        allocate( bt(np,n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SAVE v IF CHOLESKY FACTORIZATION FAILS.
!
        bt(:np,:np) = v(:np,:np)   
!
!       COMPUTE CHOLESKY FACTORIZATION OF v = c**(t)*c .
!
        call chol_cmp( v(:np,:np), diagr(:np), d1 )
!
    else
!
!       USE AN EVD DECOMPOSITION AS A BACKUP.
!
        d1 = zero
!
    end if
!
    if ( d1/=zero ) then
!
!       USE THE NYSTROM METHOD.
!
!       COMPUTE F = b*c**(-1) USING A TRIANGULAR SOLVE.
!
        bt(:np,:n) = transpose( b(:n,:np) )    
!
        do i = 1_i4b, n
!
            do j = 1_i4b, np
!
                bt(j,i) = ( bt(j,i) - dot_product( v(1_i4b:j-1_i4b,j), bt(1_i4b:j-1_i4b,i) ) )*diagr(j)
!
            end do
!
        end do
!
        b(:n,:np) = transpose( bt(:np,:n) )    
!
!       COMPUTE SVD OF THE CHOLESKY FACTOR.
!
        call svd_cmp( b(:n,:np), beta(:np), failure, v=v(:np,:np), sort=sort, max_francis_steps=10_i4b )
!
!       EXTRACT THE APPROXIMATE TOP neig EIGENVECTORS OF a .
!
        eigvec(:n,:neig) = b(:n,:neig)
!
!       COMPUTE THE APPROXIMATE TOP neig EIGENVALUES OF a .
!
        eigval(:neig) = beta(:neig)*beta(:neig)
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( bt )
!
    else
!
        if ( use_nystrom ) then
!
!           RESTORE v AS CHOLESKY FACTORIZATION FAILED.
!
            v(:np,:np) = bt(:np,:np)    
!
!           DEALLOCATE WORK ARRAY.
!
            deallocate( bt )
!
        end if
!
!       USE A STANDARD SPECTRAL DECOMPOSITION.
!
        call eig_cmp( v(:np,:np), beta(:np), failure, sort=sort, maxiter=30_i4b )
!
        call eig_abs_sort( sort, beta(:np), v(:np,:np) )
!
!       COMPUTE THE APPROXIMATE TOP neig EIGENVECTORS OF a .
!!
#ifdef _BLAS
        call gemm( 'N', 'N', n, neig, np, one, q(1_i4b:n,1_i4b:np), n,              &
                   v(1_i4b:np,1_i4b:neig), np, zero, eigvec(1_i4b:n,1_i4b:neig), n  )
#else
        eigvec(:n,:neig) = matmul( q(:n,:np), v(:np,:neig) )
#endif
!
!       EXTRACT THE APPROXIMATE TOP neig EIGENVALUES OF a .
!
        eigval(:neig) = beta(:neig)
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
!   COMPUTE THE ESTIMATED RELATIVE ERROR.
!
    tmp = one - sum( (eigval(1_i4b:neig)/norma)**2 )
    relerr = sqrt( max( tmp, zero ) )
!
!   COMPUTE THE TRUE RELATIVE ERROR.
!
    relerr2 = sqrt( sum( (eigval0(neig+1_i4b:neig0)/norma)**2 ) )
!
!   COMPUTE ERROR BETWEEN THE TRUE AND ESTIMATED RELATIVE ERRORS.
!
    err = abs( relerr - relerr2 )
!
!   TEST ACCURACY OF THE EIGEN COUPLETS IF REQUIRED.
!
    if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*u(:n,:neig) - u(:n,:neig)*diag(eigval(:neig)),
!       WHERE u ARE THE EIGENVECTORS OF a.
!
        q(:n,:neig) = eigvec(:n,:neig)*spread(eigval(:neig),dim=1,ncopies=n)
!
#ifdef _BLAS
!        call gemm( 'N', 'N', n, neig, n, one, a(1_i4b:n,1_i4b:n), n,              &
!                   eigvec(1_i4b:n,1_i4b:neig), n, -one, q(1_i4b:n,1_i4b:neig), n  )
       call symm( 'L', 'U', n, neig, one, a(1_i4b:n,1_i4b:n), n,                  &
                   eigvec(1_i4b:n,1_i4b:neig), n, -one, q(1_i4b:n,1_i4b:neig), n  )
#else
        q(:n,:neig) = matmul( a(:n,:n), eigvec(:n,:neig)  ) - q(:n,:neig)
#endif
!
        beta(:neig) = norm( q(:n,:neig), dim=2_i4b )
!
        err1 = maxval( beta(:neig) )/( norma*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u(:n,:neig)**(t)*u(:n,:neig).
!
        call unit_matrix( q(:neig,:neig) )
!
        v(:neig,:neig) = abs( q(:neig,:neig) - matmul( transpose(eigvec), eigvec ) )
!
        err2 = maxval( v(:neig,:neig) )/real(n,stnd)
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
#ifdef _BLAS
    deallocate( a, q, b, v, diagr, beta, eigval0, eigval, eigvec )
#else
    deallocate( a, q, qt, b, v, diagr, beta, eigval0, eigval, eigvec )
#endif
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed eigenvalues and eigenvectors = ', err1
        write (prtunit,*) 'Orthogonality of the computed eigenvectors            = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing approximate ', neig, ' eigenvalues and eigenvectors of a ', &
       n, ' by ', n,' real symmetric positive semi-definite matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_random_eig_pos_with_blas
! ===========================================
!
end program ex1_random_eig_pos_with_blas
