program ex1_random_eig_with_blas
!
!
! Purpose
! =======
!
!   This program illustrates how to compute an approximate partial EigenValue
!   Decomposition (EVD) of a symmetric matrix with randomized power subspace iterations
!   using STATPACK and BLAS subroutines.
!                                                                              
!                                                     
! Further Details
! ===============
!
!   The program also shows the use of subroutines QR_CMP, ORTHO_GEN_QR in module QR_Procedures,
!   EIG_CMP in module Eig_procedures, NORMAL_RANDOM_NUMBER3_ and GEN_RANDOM_SYM_MAT in module
!   Random, and gemm, symm, syrk generic interfaces in module BLAS_interfaces.
!                                                                              
! LATEST REVISION : 23/10/2020
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, half, one, seven, c30, merror, allocate_error, &
                         norm, unit_matrix, random_seed_, random_number_, normal_random_number3_,          &
                         eig_abs_sort, qr_cmp, ortho_gen_qr, eig_cmp, gen_random_sym_mat
!
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
#ifdef _BLAS
!    use BLAS_interfaces, only : gemm
    use BLAS_interfaces, only : symm, syrk, gemm
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, neig IS THE TARGET RANK OF THE APPROXIMATE PARTIAL EVD DECOMPOSITION,
! neig0 IS THE RANK OF THE GENERATED MATRIX,
! n IS THE DIMENSION OF THE GENERATED SYMMETRIC MATRIX.
!
    integer(i4b), parameter :: prtunit=6, n=2000, neig=5, neig0=1000
!
    character(len=*), parameter :: name_proc='Example 1 of random_eig_with_blas'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd) :: err, err1, err2, eps, elapsed_time, tmp, norma, relerr, relerr2
#ifdef _BLAS
    real(stnd), dimension(:,:), allocatable :: a, q, b, eigvec, v
#else
    real(stnd), dimension(:,:), allocatable :: a, q, qt, b, eigvec, v
#endif
    real(stnd), dimension(:),   allocatable :: diagr, beta, eigval0, eigval
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: p, np, i, niter, mat_type
!
    logical(lgl) :: failure, do_test, ortho
!   
    character    :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : PARTIAL EIGENVALUE DECOMPOSITION OF A n-BY-n REAL SYMMETRIC
!               MATRIX USING RANDOMIZED POWER SUBSPACE OR BLOCK KRYLOV ITERATIONS.
!
!   SPECIFY THE TYPE OF INPUT MATRIX:
!
!   mat_type = 1  -> SLOW DECAY OF EIGENVALUES      (IN ABSOLUTE MAGNITUDE)
!   mat_type = 2  -> FAST DECAY OF EIGENVALUES      (IN ABSOLUTE MAGNITUDE)
!   mat_type = 3  -> S-SHAPED DECAY OF EIGENVALUES  (IN ABSOLUTE MAGNITUDE)
!   mat_type > 3  -> VERY SLOW DECAY OF EIGENVALUES (IN ABSOLUTE MAGNITUDE)
!
    mat_type = 2_i4b
!
!   SET TOLERANCE FOR THE ERROR OF THE RELATIVE ERROR IN FROBENIUS NORM
!   OF THE COMPUTED PARTIAL EVD.
!
    eps = 0.001_stnd
!
!   SPECIFY IF NUMERICAL TESTS ARE PERFORMED TO TEST THE ACCURACY OF THE EIGEN COUPLETS.
!
    do_test = true
!
!   CHOOSE TUNING PARAMETERS FOR THE RANDOMIZED EVD ALGORITHM.
!
!   DETERMINE THE NUMBER OF POWER ITERATIONS niter TO BE PERFORMED.
!
    niter = 6_i4b
!
!   DETERMINE THE OVERSAMPLING SIZE p .
!
    p  = 20_i4b
!
!   CHECK VALIDITY OF THE OVERSAMPLING SIZE PARAMETER.
!
    np = min( p + neig, n )
!
!   SPECIFY IF ORTHONORMALIZATION IS CARRIED OUT BETWEEN EACH STEP
!   OF THE (POWER) SUBSPACE ITERATIONS, TO AVOID LOSS OF ACCURACY DUE
!   TO ROUNDING ERRORS. THIS IS NORMALLY NOT NEEDED FOR POSITIVE MATRIX.
!
    ortho = true
!
!   ALLOCATE WORK ARRAYS.
!
    i = max( np, neig0 )
!
#ifdef _BLAS
    allocate( a(n,n), q(n,np), b(n,np), v(np,np), eigvec(n,neig),         &
              diagr(np), beta(i), eigval0(neig0), eigval(neig), stat=iok  )
#else
    allocate( a(n,n), q(n,np), qt(np,n), b(n,np), v(np,np), eigvec(n,neig), &
              diagr(np), beta(i), eigval0(neig0), eigval(neig), stat=iok    )
#endif
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-BY-n RANDOM REAL SYMMETRIC MATRIX a WITH
!   A SPECIFIED DISTRIBUTION OF EIGENVALUES.
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
!   GENERATE EIGENVALUES.
!
    select case( mat_type )
!
        case( 1_i4b )
!
!           SLOW DECAY OF EIGENVALUES.
!
            do i = 1_i4b, neig0
!
                tmp = real( i, stnd )
!
                eigval0(i) = one/( tmp*tmp )
!        
            end do
!
        case( 2_i4b )
!
!           FAST DECAY OF EIGENVALUES.
!
            do i = 1_i4b, neig0
!
                tmp = real( i, stnd )
!
                eigval0(i) = exp( -tmp/seven )
!        
            end do
!
        case( 3_i4b )
!
!           S-SHAPED DECAY OF EIGENVALUES.
!
            do i = 1_i4b, neig0
!
                tmp = real( i, stnd )
!
                eigval0(i) = 0.0001_stnd + one/( one + exp( tmp -c30 ) )
!        
            end do
!
        case default
!
!           VERY SLOW DECAY OF EIGENVALUES.
!
            norma = real( neig0 - 1_i4b, stnd )
!
            do i = 1_i4b, neig0
!
                tmp = real( i - 1_i4b, stnd )
!
                eigval0(i) = exp( -tmp/norma )
!        
            end do
!
    end select
!  
!   CHANGE SIGN OF HALF OF THE EIGENVALUES.
!
    call random_number_( beta(:neig0) )
!
    where ( beta(:neig0)>half ) eigval0(:neig0) = -eigval0(:neig0)
!
!   SORT THE EIGENVALUES BY DECREASING ABSOLUTE MAGNITUDE.
!
    call eig_abs_sort( sort, eigval0(:neig0) )
!
!   GENERATE A SYMMETRIC MATRIX a WITH THE SPECIFIED EIGENVALUES
!   AND RANK neig0.
!
    call gen_random_sym_mat( eigval0(:neig0), a )
!
!   COMPUTE THE FROBENIUS NORM OF THE SYMMETRIC MATRIX.
!
!    norma = norm( a(:n,:n) )
    norma = sqrt(sum( eigval0(:neig0)**2 ) )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   GENERATE A n-BY-np IID GAUSSIAN MATRIX .
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING MATRIX
!   FORM OF SUBROUTINE normal_random_number3_.
!
!    call normal_random_number3_( b(:n,:np) )
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING VECTOR
!   FORM OF SUBROUTINE normal_random_number_.
!
    do i = 1_i4b, np
!
        call normal_random_number3_( b(:n,i) )
!
    end do
!
!   COMPUTE RANDOM SAMPLE MATRIX.
!
#ifdef _BLAS
!    call gemm( 'N', 'N', n, np, n, one, a(1_i4b:n,1_i4b:n), n,       &
!               b(1_i4b:n,1_i4b:np), n, zero, q(1_i4b:n,1_i4b:np), n  )
    call symm( 'L', 'U', n, np, one, a(1_i4b:n,1_i4b:n), n,          &
               b(1_i4b:n,1_i4b:np), n, zero, q(1_i4b:n,1_i4b:np), n  )
#else
    q(:n,:np) = matmul( a(:n,:n), b(:n,:np) )
#endif
!
!   DO POWER ITERATIONS.
!
    do i = 1_i4b, niter
!
        if ( ortho ) then
!
!           COMPUTE QR DECOMPOSITION OF RANDOM SAMPLE MATRIX TO OBTAIN AN
!           ORTHONORMAL BASIS OF RANDOM SUBSPACE.
!
            call qr_cmp( q(:n,:np), diagr(:np), beta(:np) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM SAMPLE MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( q(:n,:np), beta(:np) )
!
        end if
!
!       COMPUTE RANDOM SUBSPACE PROJECTION.
!
#ifdef _BLAS
!        call gemm( 'N', 'N', n, np, n, one, a(1_i4b:n,1_i4b:n), n,       &
!                   q(1_i4b:n,1_i4b:np), n, zero, b(1_i4b:n,1_i4b:np), n  )
        call symm( 'L', 'U', n, np, one, a(1_i4b:n,1_i4b:n), n,          &
                   q(1_i4b:n,1_i4b:np), n, zero, b(1_i4b:n,1_i4b:np), n  )
#else
        b(:n,:np) = matmul( a(:n,:n), q(:n,:np)  )
#endif
!
        q(:n,:np) = b(:n,:np)
!
    end do
!
!   COMPUTE QR DECOMPOSITION OF RANDOM SAMPLE MATRIX TO OBTAIN AN
!   ORTHONORMAL BASIS OF RANDOM SUBSPACE.
!
    call qr_cmp( q(:n,:np), diagr(:np), beta(:np) )
!
!   GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM SAMPLE MATRIX.
!   q IS ASSUMED OF FULL RANK.
!
    call ortho_gen_qr( q(:n,:np), beta(:np) )
!
!   COMPUTE FINAL RANDOM SUBSPACE PROJECTION.
!
#ifdef _BLAS
!    call gemm( 'N', 'N', n, np, n, one, a(1_i4b:n,1_i4b:n), n,       &
!               q(1_i4b:n,1_i4b:np), n, zero, b(1_i4b:n,1_i4b:np), n  )
    call symm( 'L', 'U', n, np, one, a(1_i4b:n,1_i4b:n), n,          &
               q(1_i4b:n,1_i4b:np), n, zero, b(1_i4b:n,1_i4b:np), n  )
!
!   COMPUTE v = q**(t)*b = q**(t)*a*q .
!
    call gemm( 'T', 'N', np, np, n, one, q(1_i4b:n,1_i4b:np), n,       &
               b(1_i4b:n,1_i4b:np), n, zero, v(1_i4b:np,1_i4b:np), np  )
#else
    b(:n,:np) = matmul( a(:n,:n), q(:n,:np)  )
!
!   COMPUTE v = q**(t)*b = q**(t)*a*q .
!
    qt(:np,:n) = transpose( q(:n,:np) )
!
    v(:np,:np) = matmul( qt(:np,:n), b(:n,:np) )
#endif
!
!   USE A SPECTRAL DECOMPOSITION.
!
    call eig_cmp( v(:np,:np), beta(:np), failure, maxiter=30_i4b )
!
    call eig_abs_sort( sort, beta(:np), v(:np,:np) )
!
!   COMPUTE THE APPROXIMATE TOP neig EIGENVECTORS OF a .
!
#ifdef _BLAS
    call gemm( 'N', 'N', n, neig, np, one, q(1_i4b:n,1_i4b:np), n,              &
               v(1_i4b:np,1_i4b:neig), np, zero, eigvec(1_i4b:n,1_i4b:neig), n  )
#else
    eigvec(:n,:neig) = matmul( q(:n,:np), v(:np,:neig) )
#endif
!
!   EXTRACT THE APPROXIMATE TOP neig EIGENVALUES (IN ABSOLUTE MAGNITUDE) OF a .
!
    eigval(:neig) = beta(:neig)
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
!   COMPUTE THE ESTIMATED RELATIVE ERROR.
!
    tmp = one - sum( (eigval(1_i4b:neig)/norma)**2 )
    relerr = sqrt( max( tmp, zero ) )
!
!   COMPUTE THE TRUE RELATIVE ERROR.
!
    relerr2 = sqrt( sum( (eigval0(neig+1_i4b:neig0)/norma)**2 ) )
!
!   COMPUTE ERROR BETWEEN THE TRUE AND ESTIMATED RELATIVE ERRORS.
!
    err = abs( relerr - relerr2 )
!
!   TEST ACCURACY OF THE EIGEN COUPLETS IF REQUIRED.
!
    if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*u(:n,:neig) - u(:n,:neig)*diag(eigval(:neig)),
!       WHERE u ARE THE EIGENVECTORS OF a.
!
        q(:n,:neig) = eigvec(:n,:neig)*spread(eigval(:neig),dim=1,ncopies=n)
!
#ifdef _BLAS
!        call gemm( 'N', 'N', n, neig, n, one, a(1_i4b:n,1_i4b:n), n,              &
!                   eigvec(1_i4b:n,1_i4b:neig), n, -one, q(1_i4b:n,1_i4b:neig), n  )
       call symm( 'L', 'U', n, neig, one, a(1_i4b:n,1_i4b:n), n,                  &
                   eigvec(1_i4b:n,1_i4b:neig), n, -one, q(1_i4b:n,1_i4b:neig), n  )
#else
        q(:n,:neig) = matmul( a(:n,:n), eigvec(:n,:neig)  ) - q(:n,:neig)
#endif
!
        beta(:neig) = norm( q(:n,:neig), dim=2_i4b )
!
        err1 = maxval( beta(:neig) )/( norma*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u(:n,:neig)**(t)*u(:n,:neig).
!
        call unit_matrix( q(:neig,:neig) )
!
        v(:neig,:neig) = abs( q(:neig,:neig) - matmul( transpose(eigvec), eigvec ) )
!
        err2 = maxval( v(:neig,:neig) )/real(n,stnd)
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
#ifdef _BLAS
    deallocate( a, q, b, v, diagr, beta, eigval0, eigval, eigvec )
#else
    deallocate( a, q, qt, b, v, diagr, beta, eigval0, eigval, eigvec )
#endif
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed eigenvalues and eigenvectors = ', err1
        write (prtunit,*) 'Orthogonality of the computed eigenvectors            = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing approximate ', neig, ' eigenvalues and eigenvectors of a ', &
       n, ' by ', n,' real symmetric matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_random_eig_with_blas
! =======================================
!
end program ex1_random_eig_with_blas
