program ex1_random_svd_with_blas
!
!
! Purpose
! =======
!
!   This program illustrates how to compute an approximate partial SVD with randomized
!   subspace iterations using STATPACK and BLAS subroutines.
!                                                                              
!                                                     
! Further Details
! ===============
!
!   The program shows the use of subroutines SVD_CMP in module SVD_Procedures,
!   QR_CMP and ORTHO_GEN_QR in module QR_Procedures, NORMAL_RANDOM_NUMBER3_ and
!   GEN_RANDOM_MAT in module Random and gemm generic interface in module BLAS_interfaces.
!    
!                                                                              
! LATEST REVISION : 23/10/2020
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, one, seven, c30, merror, allocate_error, &
                         norm, unit_matrix, random_seed_, normal_random_number3_, singval_sort,      &
                         qr_cmp, ortho_gen_qr, svd_cmp, gen_random_mat
!
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
#ifdef _BLAS
    use BLAS_interfaces, only : gemm
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, m AND n ARE THE DIMENSIONS OF THE GENERATED MATRIX,
! nsvd0 IS THE RANK OF THE GENERATED MATRIX,
! nsvd IS THE TARGET RANK OF THE APPROXIMATE PARTIAL SVD, WHICH IS SOUGHT
!
    integer(i4b), parameter :: prtunit=6, m=2000, n=2000, mn=min(m,n), nsvd0=1000, nsvd=10
!
    character(len=*), parameter :: name_proc='Example 1 of random_svd_with_blas'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd) :: err, err1, err2, err3, eps, elapsed_time, norma, tmp, relerr, relerr2
#ifdef _BLAS
    real(stnd), dimension(:,:), allocatable :: a, q, b, v, leftvec, rightvec
#else
    real(stnd), dimension(:,:), allocatable :: a, q, qt, b, bt, leftvec, rightvec
#endif
    real(stnd), dimension(:),   allocatable :: diagr, beta, singval, singval0
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: p, np, niter, i, mat_type
!
    logical(lgl) :: failure, do_test, ortho
!   
    character    :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : PARTIAL SVD OF A m-BY-n REAL MATRIX USING RANDOMIZED
!               POWER SUBSPACE ITERATIONS.
!
!   SPECIFY THE TYPE OF INPUT MATRIX:
!
!   mat_type = 1  -> SLOW DECAY OF SINGULAR VALUES
!   mat_type = 2  -> FAST DECAY OF SINGULAR VALUES
!   mat_type = 3  -> S-SHAPED DECAY OF SINGULAR VALUES
!   mat_type > 3  -> VERY SLOW DECAY OF SINGULAR VALUES
!
    mat_type = 2_i4b
!
!   SET TOLERANCE FOR THE ERROR OF THE RELATIVE ERROR IN FROBENIUS NORM
!   OF THE COMPUTED PARTIAL SVD.
!
    eps = 0.001_stnd
!
!   SPECIFY IF NUMERICAL TESTS ARE PERFORMED TO TEST THE ACCURACY OF THE SINGULAR TRIPLETS.
!
    do_test = true
!
!   CHOOSE TUNING PARAMETERS FOR THE RANDOMIZED ALGORITHM.
!
!   DETERMINE THE NUMBER OF POWER ITERATIONS niter TO BE PERFORMED.
!
    niter = 4_i4b
!
!   DETERMINE THE OVERSAMPLING SIZE p .
!
    p  = 10_i4b
!
!   CHECK VALIDITY OF THE OVERSAMPLING SIZE PARAMETER.
!
    np = min( p + nsvd, mn )
!
!   SPECIFY IF ORTHONORMALIZATION IS CARRIED OUT BETWEEN EACH STEP
!   OF THE (POWER) SUBSPACE ITERATIONS, TO AVOID LOSS OF ACCURACY DUE
!   TO ROUNDING ERRORS.
!
    ortho = true
!
!   ALLOCATE WORK ARRAYS.
!
#ifdef _BLAS
    allocate( a(m,n), q(m,np), b(n,np), v(np,np),                        &
              diagr(nsvd0), beta(nsvd0), singval0(nsvd0), singval(nsvd), &
              leftvec(m,nsvd), rightvec(n,nsvd),  stat=iok )
#else
    allocate( a(m,n), q(m,np), qt(np,m), b(n,np), bt(np,n),              &
              diagr(nsvd0), beta(nsvd0), singval0(nsvd0), singval(nsvd), &
              leftvec(m,nsvd), rightvec(n,nsvd),  stat=iok )
#endif
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A m-BY-n RANDOM REAL MATRIX a WITH A GEOMETRIC DISTRIBUTION
!   OF EIGENVALUES WITH APPROXIMATE CONDITION NUMBER conda AND RANK nsvd0 .
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
!   GENERATE SINGULAR VALUES.
!
    select case( mat_type )
!
        case( 1_i4b )
!
!           SLOW DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = one/( tmp*tmp )
!        
            end do
!
        case( 2_i4b )
!
!           FAST DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = exp( -tmp/seven )
!        
            end do
!
        case( 3_i4b )
!
!           S-SHAPED DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = 0.0001_stnd + one/( one + exp( tmp -c30 ) )
!        
            end do
!
        case default
!
!           VERY SLOW DECAY OF SINGULAR VALUES.
!
            norma = real( nsvd0 - 1_i4b, stnd )
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i - 1_i4b, stnd )
!
                singval0(i) = exp( -tmp/norma )
!        
            end do
!
    end select
!
!   SORT THE SINGULAR VALUES.
!
    call singval_sort( sort, singval0(:nsvd0) )
!
!   GENERATE A m-BY-n MATRIX a OF RANK nsvd0 .
!
    call gen_random_mat( singval0(:nsvd0), a )
!
!   COMPUTE THE FROBENIUS NORM OF THE MATRIX.
!
!    norma = norm( a(:m,:n) )
    norma = sqrt(sum( singval0(:nsvd0)**2 ) )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   GENERATE A n-BY-np IID GAUSSIAN MATRIX .
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING MATRIX
!   FORM OF SUBROUTINE normal_random_number3_.
!
    call normal_random_number3_( b(:n,:np) )
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING VECTOR
!   FORM OF SUBROUTINE normal_random_number_.
!
!    do i = 1_i4b, np
!
!        call normal_random_number3_( b(:n,i) )
!
!    end do
!
!   COMPUTE RANDOM SAMPLE MATRIX.
!
#ifdef _BLAS
    call gemm( 'N', 'N', m, np, n, one, a(1_i4b:m,1_i4b:n), m,       &
               b(1_i4b:n,1_i4b:np), n, zero, q(1_i4b:m,1_i4b:np), m  )
#else
    q(:m,:np) = matmul( a(:m,:n), b(:n,:np) )
#endif
!
!   DO POWER ITERATIONS.
!
    do i = 1_i4b, niter
!
        if ( ortho ) then
!
!           COMPUTE QR DECOMPOSITION OF RANDOM MATRIX TO OBTAIN AN
!           ORTHONORMAL BASIS OF RANDOM SUBSPACE.
!
            call qr_cmp( q(:m,:np), diagr(:np), beta(:np) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( q(:m,:np), beta(:np) )
!
        end if
!
!       COMPUTE RANDOM SUBSPACE PROJECTION.
!
#ifdef _BLAS
        call gemm( 'T', 'N', n, np, m, one, a(1_i4b:m,1_i4b:n), m,       &
                   q(1_i4b:m,1_i4b:np), m, zero, b(1_i4b:n,1_i4b:np), n  )
#else
        qt(:np,:m) = transpose( q(:m,:np) )
!
        bt(:np,:n) = matmul( qt(:np,:m), a(:m,:n) )
!
        b(:n,:np) = transpose( bt(:np,:n) )
#endif
!
        if ( ortho ) then
!
!           COMPUTE QR DECOMPOSITION OF RANDOM MATRIX TO OBTAIN AN
!           ORTHONORMAL BASIS OF RANDOM SUBSPACE.
!
            call qr_cmp( b(:n,:np), diagr(:np), beta(:np) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( b(:n,:np), beta(:np) )
!
        end if
!
!       COMPUTE RANDOM SUBSPACE PROJECTION.
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, np, n, one, a(1_i4b:m,1_i4b:n), m,       &
                   b(1_i4b:n,1_i4b:np), n, zero, q(1_i4b:m,1_i4b:np), m  )
#else
        q(:m,:np) = matmul( a(:m,:n), b(:n,:np) )
#endif
!
    end do
!
!   COMPUTE QR DECOMPOSITION OF RANDOM MATRIX TO OBTAIN AN
!   ORTHONORMAL BASIS OF RANDOM SUBSPACE.
!
    call qr_cmp( q(:m,:np), diagr(:np), beta(:np) )
!
!   GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM MATRIX.
!   q IS ASSUMED OF FULL RANK.
!
    call ortho_gen_qr( q(:m,:np), beta(:np) )
!
#ifdef _BLAS
    call gemm( 'T', 'N', n, np, m, one, a(1_i4b:m,1_i4b:n), m,       &
               q(1_i4b:m,1_i4b:np), m, zero, b(1_i4b:n,1_i4b:np), n  )
!
!   COMPUTE SVD OF THE FINAL RANDOM MATRIX PROJECTION.
!
    call svd_cmp( b(:n,:np), beta(:np), failure, v=v(:np,:np), sort=sort, max_francis_steps=10_i4b )
!
!   COMPUTE THE APPROXIMATE TOP nsvd LEFT SINGULAR VECTORS OF a .
!
    call gemm( 'N', 'N', m, nsvd, np, one, q(1_i4b:m,1_i4b:np), m,             &
               v(1_i4b:np,1_i4b:np), np, zero, leftvec(1_i4b:m,1_i4b:nsvd), m  )
#else
    qt(:np,:m) = transpose( q(:m,:np) )
!
    bt(:np,:n) = matmul( qt(:np,:m), a(:m,:n) )
!
!   COMPUTE SVD OF THE FINAL RANDOM MATRIX PROJECTION.
!
    call svd_cmp( bt(:np,:n), beta(:np), failure, v=b(:n,:np), sort=sort, max_francis_steps=10_i4b )
!
!   COMPUTE THE APPROXIMATE TOP nsvd LEFT SINGULAR VECTORS OF a .
!
    leftvec(:m,:nsvd) = matmul( q(:m,:np), bt(:np,:nsvd) )
#endif
!
!   EXTRACT THE APPROXIMATE TOP nsvd RIGHT SINGULAR VECTORS OF a .
!
    rightvec(:n,:nsvd) = b(:n,:nsvd)
!
!   EXTRACT THE APPROXIMATE TOP nsvd SINGULAR VALUES OF a .
!
    singval(:nsvd) = beta(:nsvd)
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
!   COMPUTE THE ESTIMATED RELATIVE ERROR.
!
    tmp = one - sum( (singval(:nsvd)/norma)**2 )
    relerr = sqrt( max( tmp, zero ) )
!
!   COMPUTE THE TRUE RELATIVE ERROR.
!
    relerr2 = sqrt(sum( (singval0(nsvd+1_i4b:nsvd0)/norma)**2 ))
!
!   COMPUTE ERROR BETWEEN THE TRUE AND ESTIMATED RELATIVE ERRORS.
!
    err = abs( relerr - relerr2 )
!
!   TEST ACCURACY OF THE SINGULAR TRIPLETS IF REQUIRED.
!
    if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*v(:m,:nsvd) - u(:n,:nsvd)*diag(singval(:nsvd)),
!       WHERE u AND v ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a.
!
        q(:m,:nsvd) = leftvec(:m,:nsvd)*spread( singval(:nsvd), dim=1, ncopies=m)
!
#ifdef _BLAS
        call gemm( 'N', 'N', m, nsvd, n, one, a(1_i4b:m,1_i4b:n), m,                &
                   rightvec(1_i4b:n,1_i4b:nsvd), n, -one, q(1_i4b:m,1_i4b:nsvd), m  )
#else
        q(:m,:nsvd) = matmul( a(:m,:n), rightvec(:n,:nsvd) ) - q(:m,:nsvd)
#endif
!
        beta(:nsvd) = norm( q(:m,:nsvd), dim=2_i4b )
!
        err1 = maxval( beta(:nsvd) )/( norma*real(mn,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u(:n,:nsvd)**(t)*u(:n,:nsvd).
!
        call unit_matrix( q(:nsvd,:nsvd) )
!
        b(:nsvd,:nsvd) = abs( q(:nsvd,:nsvd) - matmul( transpose(leftvec), leftvec ) )
!
        err2 = maxval( b(:nsvd,:nsvd) )/real(m,stnd)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - v(:m,:nsvd)**(t)*v(:m,:nsvd).
!
        b(:nsvd,:nsvd) = abs( q(:nsvd,:nsvd) - matmul( transpose(rightvec), rightvec ) )
!
        err3 = maxval( b(:nsvd,:nsvd) )/real(n,stnd)
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
#ifdef _BLAS
    deallocate( a, q, b, v, diagr, beta, singval0, singval, leftvec, rightvec )
#else
    deallocate( a, q, qt, b, bt, diagr, beta, singval0, singval, leftvec, rightvec )
#endif
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed singular triplets           = ', err1
        write (prtunit,*) 'Orthogonality of the computed left singular vectors  = ', err2
        write (prtunit,*) 'Orthogonality of the computed right singular vectors = ', err3
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing approximate ', nsvd, ' singular values and vectors of a ', &
       m, ' by ', n,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_random_svd_with_blas
! =======================================
!
end program ex1_random_svd_with_blas
