program ex1_reig_cmp
!
!
! Purpose
! =======
!
!   This program illustrates how to compute an approximate partial EigenValue Decomposition (EVD)
!   of a symmetric matrix with randomized power, subspace or block Krylov iterations using
!   subroutine REIG_CMP in module Eig_Procedures.
!                                                                              
! LATEST REVISION : 23/10/2020
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, half, one, seven, c30, merror, allocate_error, &
                         norm, unit_matrix, random_seed_, random_number_, normal_random_number3_,          &
                         eigval_abs_sort, reig_cmp, gen_random_sym_mat
!
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, neig IS THE TARGET RANK OF THE APPROXIMATE PARTIAL EVD DECOMPOSITION,
! neig0 IS THE RANK OF THE GENERATED MATRIX,
! n IS THE DIMENSION OF THE GENERATED SYMMETRIC MATRIX.
!
    integer(i4b), parameter :: prtunit=6, n=2000, neig=5, neig0=1000
!
    character(len=*), parameter :: name_proc='Example 1 of reig_cmp'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd) :: err, err1, err2, eps, elapsed_time, tmp, norma, relerr, relerr2
    real(stnd), dimension(:,:), allocatable :: a, eigvec, res, id
    real(stnd), dimension(:),   allocatable :: eigval0, eigval, beta
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: i, niter, nover, mat_type
!
    logical(lgl) :: failure, extd_samp, do_test, ortho
!   
    character    :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : PARTIAL EIGENVALUE DECOMPOSITION OF A n-BY-n REAL SYMMETRIC
!               MATRIX USING RANDOMIZED POWER SUBSPACE OR BLOCK KRYLOV ITERATIONS.
!
!   SPECIFY THE TYPE OF INPUT MATRIX:
!
!   mat_type = 1  -> SLOW DECAY OF EIGENVALUES      (IN ABSOLUTE MAGNITUDE)
!   mat_type = 2  -> FAST DECAY OF EIGENVALUES      (IN ABSOLUTE MAGNITUDE)
!   mat_type = 3  -> S-SHAPED DECAY OF EIGENVALUES  (IN ABSOLUTE MAGNITUDE)
!   mat_type > 3  -> VERY SLOW DECAY OF EIGENVALUES (IN ABSOLUTE MAGNITUDE)
!
    mat_type = 2_i4b
!
!   SET TOLERANCE FOR THE ERROR OF THE RELATIVE ERROR IN FROBENIUS NORM
!   OF THE COMPUTED PARTIAL EVD.
!
    eps = 0.001_stnd
!
!   SPECIFY IF NUMERICAL TESTS ARE PERFORMED TO TEST THE ACCURACY OF THE EIGEN COUPLETS.
!
    do_test = true
!
!   CHOOSE TUNING PARAMETERS FOR THE RANDOMIZED EVD ALGORITHM.
!
!   DETERMINE THE NUMBER OF POWER SUBSPACE OR BLOCK KRYLOV ITERATIONS niter .
!
    niter = 6_i4b
!
!   DETERMINE THE OVERSAMPLING SIZE nover .
!
    nover = 10_i4b
!
!   SPECIFY IF POWER SUBSPACE OR BLOCK KRYLOV ITERATIONS ARE USED.
!
    extd_samp = false
!
!   DETERMINE IF ORTHONORMALIZATION IS CARRIED OUT BETWEEN EACH STEP
!   OF THE (POWER) SUBSPACE OR BLOCK KRYLOV ITERATIONS, TO AVOID LOSS
!   OF ACCURACY DUE TO ROUNDING ERRORS. THIS IS NOT NEEDED NORMALLY
!   FOR SEMI-DEFINITE POSITIVE MATRIX.
!
    ortho = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,n), eigvec(n,neig), eigval(neig),  &
              eigval0(neig0), beta(neig0), stat=iok  )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-BY-n RANDOM REAL SYMMETRIC MATRIX a WITH
!   A SPECIFIED DISTRIBUTION OF EIGENVALUES.
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
!   GENERATE EIGENVALUES.
!
    select case( mat_type )
!
        case( 1_i4b )
!
!           SLOW DECAY OF EIGENVALUES.
!
            do i = 1_i4b, neig0
!
                tmp = real( i, stnd )
!
                eigval0(i) = one/( tmp*tmp )
!        
            end do
!
        case( 2_i4b )
!
!           FAST DECAY OF EIGENVALUES.
!
            do i = 1_i4b, neig0
!
                tmp = real( i, stnd )
!
                eigval0(i) = exp( -tmp/seven )
!        
            end do
!
        case( 3_i4b )
!
!           S-SHAPED DECAY OF EIGENVALUES.
!
            do i = 1_i4b, neig0
!
                tmp = real( i, stnd )
!
                eigval0(i) = 0.0001_stnd + one/( one + exp( tmp -c30 ) )
!        
            end do
!
        case default
!
!           VERY SLOW DECAY OF EIGENVALUES.
!
            norma = real( neig0 - 1_i4b, stnd )
!
            do i = 1_i4b, neig0
!
                tmp = real( i - 1_i4b, stnd )
!
                eigval0(i) = exp( -tmp/norma )
!        
            end do
!
    end select
!  
!   CHANGE SIGN OF HALF OF THE EIGENVALUES.
!
    call random_number_( beta(:neig0) )
!
    where ( beta(:neig0)>half ) eigval0(:neig0) = -eigval0(:neig0)
!
!   SORT THE EIGENVALUES BY DECREASING ABSOLUTE MAGNITUDE.
!
    call eigval_abs_sort( sort, eigval0(:neig0) )
!
!   GENERATE A SYMMETRIC MATRIX a WITH THE SPECIFIED EIGENVALUES
!   AND RANK neig0.
!
    call gen_random_sym_mat( eigval0(:neig0), a )
!
!   COMPUTE THE FROBENIUS NORM OF THE SYMMETRIC MATRIX.
!
!    norma = norm( a(:n,:n) )
    norma = sqrt(sum( eigval0(:neig0)**2 ) )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   reig_cmp COMPUTES A PARTIAL EIGENVALUE DECOMPOSITION (EVD) OF A REAL
!   n-BY-n SYMMETRIC MATRIX a. THE PARTIAL EVD IS WRITTEN
!
!                       U * S * U**(t)
!
!   WHERE S IS AN neig-BY-neig MATRIX WHICH IS ZERO EXCEPT FOR ITS
!   DIAGONAL ELEMENTS, U IS AN n-BY-neig ORTHONORMAL MATRIX. THE DIAGONAL
!   ELEMENTS OF S ARE THE neig LARGEST EIGENVALUES OF a IN DECREASING ABSOLUTE
!   MAGNITUDE ORDER. THE COLUMNS OF U ARE THE ASSOCIATED EIGENVECTORS OF a.
!
    call reig_cmp( a(:n,:n), eigval(:neig), eigvec(:n,:neig), failure=failure, &
                   niter=niter, nover=nover, ortho=ortho, extd_samp=extd_samp    )
!
!   THE ROUTINE RETURNS THE neig LARGEST EIGENVALUES (IN ABSOLUTE MAGNITUDE) AND THE
!   ASSOCIATED EIGENVECTORS.
!
!   ON EXIT OF reig_cmp :
!
!       eigval CONTAINS THE neig LARGEST EIGENVALUES OF a IN DECREASING ORDER OF
!       ABSOLUTE MAGNITUDE.
!
!       eigvec CONTAINS THE ASSOCIATED neig EIGENVECTORS,
!       STORED COLUMNWISE;
!
!       failure= false :  INDICATES SUCCESSFUL EXIT.
!       failure= true  :  INDICATES THAT THE ALGORITHM DID NOT CONVERGE AND
!                         THAT FULL ACCURACY WAS NOT ATTAINED IN THE RANDOMIZED
!                         POWER SUBSPACE OR BLOCK KRYLOV ITERATIONS. THE RESULTS
!                         CAN BE STILL USEFUL, BUT THE APPROXIMATIONS OF THE neig
!                         TOP EIGEN COUPLETS CAN BE POOR.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
!   COMPUTE THE ESTIMATED RELATIVE ERROR.
!
    tmp = one - sum( (eigval(1_i4b:neig)/norma)**2 )
    relerr = sqrt( max( tmp, zero ) )
!
!   COMPUTE THE TRUE RELATIVE ERROR.
!
    relerr2 = sqrt( sum( (eigval0(neig+1_i4b:neig0)/norma)**2 ) )
!
!   COMPUTE ERROR BETWEEN THE TRUE AND ESTIMATED RELATIVE ERRORS.
!
    err = abs( relerr - relerr2 )
!
!   TEST ACCURACY OF THE EIGEN COUPLETS IF REQUIRED.
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( res(n,neig), id(neig,neig), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*u(:n,:neig) - u(:n,:neig)*diag(eigval(:neig)),
!       WHERE u ARE THE EIGENVECTORS OF a.
!
        res(:n,:neig)   = matmul(a,eigvec) - eigvec*spread(eigval(:neig),dim=1,ncopies=n)
        id(:neig,1_i4b) = norm( res(:n,:neig), dim=2_i4b )
!
        err1 = maxval( id(:neig,1_i4b) )/( norma*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u(:n,:neig)**(t)*u(:n,:neig).
!
        call unit_matrix( id(:neig,:neig) )
!
        res(:neig,:neig) = abs( id(:neig,:neig) - matmul( transpose(eigvec), eigvec ) )
!
        err2 = maxval( res(:neig,:neig) )/real(n,stnd)
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( res, id )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, eigval0, eigval, eigvec, beta )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed eigenvalues and eigenvectors = ', err1
        write (prtunit,*) 'Orthogonality of the computed eigenvectors            = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing approximate ', neig, ' eigenvalues and eigenvectors of a ', &
       n, ' by ', n,' real symmetric matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_reig_cmp
! ===========================
!
end program ex1_reig_cmp
