program ex1_rsvd_cmp_fixed_precision
!
!
! Purpose
! =======
!
!   This program illustrates how to compute an approximate reduced SVD with randomized
!   subspace iterations, which fullfills a given relative error in Frobenius norm using
!   subroutine RSVD_CMP_FIXED_PRECISION in module SVD_Procedures.
!                                                                              
!                                                                              
! LATEST REVISION : 26/11/2020
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, one, two, three, seven, c30, allocate_error,  &
                         merror, norm, unit_matrix, random_seed_, normal_random_number3_, singval_sort,   &
                         gen_random_mat, rsvd_cmp_fixed_precision
!
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, m AND n ARE THE DIMENSIONS OF THE GENERATED MATRIX,
! nsvd0 IS THE RANK OF THE GENERATED MATRIX,
! relerr0 IS THE REQUESTED RELATIVE ERROR OF THE RESTRICTED SVD IN FROBENIUS NORM.
!
    integer(i4b), parameter :: prtunit=6, m=10000, n=5000, mn=min(m,n), nsvd0=1000
!   
    real(stnd), parameter  :: relerr0=0.5_stnd
!
    character(len=*), parameter :: name_proc='Example 1 of rsvd_cmp_fixed_precision'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd) :: err, err1, err2, err3, eps, elapsed_time, norma, relerr, relerr2, tmp
    real(stnd), dimension(:,:), allocatable :: a, res, id
    real(stnd), dimension(:),   allocatable :: singval0
!
    real(stnd), dimension(:,:), pointer :: leftvec, rightvec
    real(stnd), dimension(:),   pointer :: singval
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: blk_size, niter, niter_qb, maxiter_qb, i, nsvd, mat_type
!
    logical(lgl) :: failure_relerr, failure, do_test, ortho, reortho
!   
    character    :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : PARTIAL SVD OF A m-BY-n REAL MATRIX USING RANDOMIZED
!               POWER SUBSPACE ITERATIONS.
!
!   SPECIFY THE TYPE OF INPUT MATRIX:
!
!   mat_type = 1  -> SLOW DECAY OF SINGULAR VALUES
!   mat_type = 2  -> FAST DECAY OF SINGULAR VALUES
!   mat_type = 3  -> S-SHAPED DECAY OF SINGULAR VALUES
!   mat_type > 3  -> VERY SLOW DECAY OF SINGULAR VALUES
!
    mat_type = 2_i4b
!
!   SET TOLERANCE FOR THE ERROR OF THE RELATIVE ERROR IN FROBENIUS NORM
!   OF THE COMPUTED PARTIAL SVD.
!
    eps = 0.01_stnd
!
!   SPECIFY IF NUMERICAL TESTS ARE PERFORMED TO TEST THE ACCURACY OF THE SINGULAR TRIPLETS.
!
    do_test = true
!
!   CHOOSE TUNING PARAMETERS FOR THE RANDOMIZED SVD ALGORITHM.
!   blk_size*maxiter_qb IS THE MAXIMUM ALLOWABLE RANK OF THE
!   PARTIAL SVD, WHICH IS SOUGHT.
!
    blk_size   = 10_i4b
    maxiter_qb = 20_i4b
!
!   DETERMINE THE NUMBER OF POWER OR SUBSPACE ITERATIONS niter TO BE PERFORMED
!   IN THE FIRST STEP OF THE QB FACTORIZATION.
!
    niter = 5_i4b
!
!   SPECIFY IF ORTHONORMALIZATION IS CARRIED OUT BETWEEN EACH STEP
!   OF THE (POWER) SUBSPACE ITERATIONS, TO AVOID LOSS OF ACCURACY DUE
!   TO ROUNDING ERRORS.
!
    ortho = true
!
!   SPECIFY IF REORTHONORMALIZATION IS CARRIED OUT TO AVOID LOSS OF ORTHOGONALITY
!   IN THE BLOCK GRAM-SCHMIDT ORTHOGONALISATION SCHEME USED TO BUILD THE ORTHOGONAL
!   MATRIX OF THE QB DECOMPOSITION OF THE INPUT MATRIX.
!
    reortho = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n), singval0(nsvd0), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   DETERMINE THE NUMBER OF SUBSPACE ITERATIONS niter_qb TO BE PERFORMED
!   IN THE LAST STEP OF THE QB FACTORIZATION.
!
    niter_qb = 4_i4b
!
!   GENERATE A m-BY-n RANDOM REAL MATRIX a WITH A SPECIFIED DISTRIBUTION
!   OF SINGULAR VALUES.
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
!   GENERATE SINGULAR VALUES.
!
    select case( mat_type )
!
        case( 1_i4b )
!
!           SLOW DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = one/( tmp*tmp )
!        
            end do
!
        case( 2_i4b )
!
!           FAST DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = exp( -tmp/seven )
!        
            end do
!
        case( 3_i4b )
!
!           S-SHAPED DECAY OF SINGULAR VALUES.
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i, stnd )
!
                singval0(i) = 0.0001_stnd + one/( one + exp( tmp -c30 ) )
!        
            end do
!
        case default
!
!           VERY SLOW DECAY OF SINGULAR VALUES.
!
            norma = real( nsvd0 - 1_i4b, stnd )
!
            do i = 1_i4b, nsvd0
!
                tmp = real( i - 1_i4b, stnd )
!
                singval0(i) = exp( -tmp/norma )
!        
            end do
!
    end select
!
!   SORT THE SINGULAR VALUES.
!
    call singval_sort( sort, singval0(:nsvd0) )
!
!   GENERATE A m-BY-n MATRIX a OF RANK nsvd0 .
!
    call gen_random_mat( singval0(:nsvd0), a )
!
!   COMPUTE THE FROBENIUS NORM OF THE MATRIX.
!
!    norma = norm( a(:m,:n) )
    norma = sqrt(sum( singval0(:nsvd0)**2 ) )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   rsvd_cmp_fixed_precision COMPUTES A PARTIAL SINGULAR VALUE DECOMPOSITION (SVD)
!   OF A REAL m-BY-n MATRIX a, WITH A RANK AS SMALL AS POSSIBLE, BUT WHICH FULFILLS A PRESET
!   TOLERANCE FOR ITS RELATIVE ERROR IN THE FROBENIUS NORM:
!
!            ||A-rSVD||_F <= ||A||_F * relerr
!
!   , WHERE rSVD IS THE COMPUTED PARTIAL SVD APPROXIMATION, || ||_F IS THE FROBENIUS NORM AND
!   relerr IS A PRESCRIBED ACCURACY TOLERANCE FOR THE RELATIVE ERROR OF THE COMPUTED PARTIAL SVD
!   APPROXIMATION, SPECIFIED IN THE INPUT ARGUMENT relerr.
!
!   THE PARTIAL SVD IS WRITTEN
!
!                       U * S * V**(t)
!
!   WHERE S IS AN nsvd-BY-nsvd MATRIX WHICH IS ZERO EXCEPT FOR ITS
!   DIAGONAL ELEMENTS, U IS AN m-BY-nsvd ORTHONORMAL MATRIX, AND
!   V IS AN n-BY-nsvd ORTHONORMAL MATRIX.  THE DIAGONAL ELEMENTS OF S
!   ARE THE nsvd LARGEST SINGULAR VALUES OF a; THEY ARE REAL AND NON-NEGATIVE.
!   THE COLUMNS OF U AND V ARE THE ASSOCIATED LEFT AND RIGHT SINGULAR VECTORS OF a.
!
!   HERE THE RANK OF THE PARTIAL SVD IS NOT KNOWN IN ADVANCE AND THE ARGUMENTS singval, leftvec
!   AND rightvec, WHICH WILL CONTAIN THE SINGULAR TRIPLETS MUST BE SPECIFIED AS REAL ARRAY
!   POINTERS. On EXIT, nsvd = size( singval ) IS THE RANK OF THE COMPUTED PARTIAL SVD.
!
!   SET THE TOLERANCE FOR THE RELATIVE ERROR IN FROBENIUS NORM.
!
    relerr = relerr0
!
!   NULLIFY THE POINTERS singval, leftvec AND rightvec SO THAT THEIR STATUT CAN BE CHECKED INSIDE
!   rsvd_cmp_fixed_precision SUBROUTINE.
!
    nullify( singval, leftvec, rightvec )
!
    call rsvd_cmp_fixed_precision( a(:m,:n), relerr, singval, leftvec, rightvec,          &
                                   failure_relerr=failure_relerr, failure=failure,        &
                                   niter=niter, blk_size=blk_size, maxiter_qb=maxiter_qb, &
                                   ortho=ortho, reortho=reortho, niter_qb=niter_qb,       &
                                   max_francis_steps=10_i4b )
!
!   THE ROUTINE RETURNS THE nsvd LARGEST SINGULAR VALUES AND THE ASSOCIATED LEFT
!   AND RIGHT SINGULAR VECTORS, WHICH FULFILLS THE PRESET TOLERANCE SPECIFIED IN ARGUMENT
!   relerr.
!
!   ON EXIT OF rsvd_cmp_fixed_precision :
!
!       relerr CONTAINS THE RELATIVE ERROR IN FROBENIUS NORM OF THE COMPUTED PARTIAL SVD.
!
!       POINTER singval CONTAINS THE nsvd = size(singval) LARGEST SINGULAR VALUES OF a
!       IN DECREASING ORDER.
!
!       POINTER leftvec CONTAINS THE ASSOCIATED nsvd LEFT SINGULAR VECTORS
!       STORED COLUMNWISE.
!   
!       POINTER rightvec CONTAINS THE ASSOCIATED nsvd RIGHT SINGULAR VECTORS,
!       STORED COLUMNWISE.
!
!       failure_relerr = false : INDICATES SUCCESSFUL EXIT AND THE COMPUTED PARTIAL
!       SVD FULFILLS THE REQUESTED RELATIVE ERROR SPECIFIED ON ENTRY IN THE ARGUMENT relerr.
!
!       failure_relerr = true  : INDICATES THAT THE COMPUTED PARTIAL SVD HAS A RELATIVE ERROR
!       LARGER THAN THE REQUESTED RELATIVE ERROR. THIS MEANS THAT THE REQUESTED ACCURACY TOLERANCE
!       FOR THE RELATIVE ERROR IS TOO SMALL (I.E., relerr < 2 * sqrt( epsilon( relerr )/relerr )
!       OR THAT THE INPUT PARAMETERS blk_size AND/OR maxiter_qb HAVE A TOO SMALL VALUE (E.G., THE
!       PRODUCT blk_size*maxiter_qb SETS THE MAXIMUM ALLOWABLE RANK FOR THE PARTIAL SVD, WHICH IS SOUGHT),
!       GIVEN THE DISTRIBUTION OF THE SINGULAR VALUES OF mat, AND MUST BE INCREASED TO FULLFILL THE PRESET
!       ACCURACY TOLERANCE FOR THE RELATIVE ERROR OF THE PARTIAL SVD APPROXIMATION.
!
!       failure = false : INDICATES SUCCESSFUL EXIT AND AN ACCURATE APPROXIMATION OF THE SINGULAR TRIPLETS
!
!       failure = true  : INDICATES THAT THE ALGORITHM DID NOT CONVERGE AND
!                         THAT FULL ACCURACY WAS NOT ATTAINED IN THE RANDOMIZED
!                         SUBSPACE ITERATIONS. THE RESULTS CAN BE STILL USEFUL, ESPECIALLY
!                         IF failure_relerr = false IS FALSE ON EXIT, BUT THE
!                         APPROXIMATIONS OF THE TOP nsvd SINGULAR TRIPLETS
!                         CAN BE POOR.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
!   GET THE RANK OF THE COMPUTED PARTIAL SVD.
!
    nsvd = size( singval )
!
!   COMPUTE THE TRUE RELATIVE ERROR.
!
    relerr2 = sqrt(sum( (singval0(nsvd+1_i4b:nsvd0)/norma)**2 ))
!
!   COMPUTE ERROR BETWEEN THE TRUE AND ESTIMATED RELATIVE ERRORS.
!
    err = abs( relerr - relerr2 )
!
!   TEST ACCURACY OF THE SINGULAR TRIPLETS IF REQUIRED.
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( res(m,nsvd), id(nsvd,nsvd), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*v(:m,:nsvd) - u(:n,:nsvd)*diag(singval(:nsvd)),
!       WHERE u AND v ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a.
!
        res(:m,:nsvd) = matmul(a,rightvec) - leftvec*spread(singval(:nsvd),dim=1,ncopies=m)
        id(:nsvd,1_i4b) = norm( res(:m,:nsvd), dim=2_i4b )
!
        if ( norma==zero ) then
            norma = one
        end if
!
        err1 = maxval( id(:nsvd,1_i4b) )/( norma*real(mn,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u(:n,:nsvd)**(t)*u(:n,:nsvd).
!
        call unit_matrix( id(:nsvd,:nsvd) )
!
        res(:nsvd,:nsvd) = abs( id(:nsvd,:nsvd) - matmul( transpose(leftvec), leftvec ) )
!
        err2 = maxval( res(:nsvd,:nsvd) )/real(m,stnd)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - v(:m,:nsvd)**(t)*v(:m,:nsvd).
!
        res(:nsvd,:nsvd) = abs( id(:nsvd,:nsvd) - matmul( transpose(rightvec), rightvec ) )
!
        err3 = maxval( res(:nsvd,:nsvd) )/real(n,stnd)
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( res, id )
!
    end if
!
!   DEALLOCATE WORK ARRAYS AND POINTERS.
!
    deallocate( a, singval0, singval, leftvec, rightvec )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( failure_relerr ) then
!
        write (prtunit,*) 
        write (prtunit,*) 'Fail to converge within ', maxiter_qb,   &
                          ' iterations! ||A-rSVD||_F / ||A||_F = ', relerr, ' >= ', relerr0
        write (prtunit,*) 
!
    else
!
        write (prtunit,*) 
        write (prtunit,*) 'Converge within less than ', maxiter_qb ,' iterations! ||A-rSVD||_F / ||A||_F = ', &
                          relerr, ' < ', relerr0
        write (prtunit,*) 
!
    end if
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed singular triplets           = ', err1
        write (prtunit,*) 'Orthogonality of the computed left singular vectors  = ', err2
        write (prtunit,*) 'Orthogonality of the computed right singular vectors = ', err3
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing approximate ', nsvd, ' singular values and vectors of a ', &
       m, ' by ', n,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_rsvd_cmp_fixed_precision
! ===========================================
!
end program ex1_rsvd_cmp_fixed_precision
