program ex1_select_singval_cmp2
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine SELECT_SINGVAL_CMP2
!   in module SVD_Procedures.
!                                                                              
!                                                     
! Further Details
! ===============
!
!    The program also shows the use of subroutine BD_INVITER2 in module SVD_Procedures.
!                                                                              
! LATEST REVISION : 25/10/2019
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c0_9, c1_5, c50, lamch, bd_inviter2,   &
                         select_singval_cmp2, merror, allocate_error, norm, unit_matrix
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n=3000, m=3000, mn=min(m,n), ls=10
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of select_singval_cmp2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, err3, eps, abstol, elapsed_time
    real(stnd), dimension(:,:), allocatable :: a, a2, leftvec, rightvec, resid, rla
    real(stnd), dimension(:),   allocatable :: s, d, e, tauo, tauq, taup, resid2
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: maxiter=2, nsing, mnthr, mnthr_nsing
!
    logical(lgl) :: failure1, failure2, gen_q, do_test, two_stage
!   
    character    :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : PARTIAL SVD OF A n-BY-m REAL MATRIX USING THE CHAN-GOLUB-REINSCH
!               BIDIAGONAL REDUCTION ALGORITHM, A BISECTION ALGORITHM FOR
!               SINGULAR VALUES AND THE INVERSE ITERATION TECHNIQUE FOR
!               SINGULAR VECTORS.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps    = fudge*epsilon( err )
    abstol = sqrt( lamch('S') )
    err    = zero
!
!   SPECIFY IF NUMERICAL TESTS ARE PERFORMED.
!
    do_test = false
!
!   CHOOSE TUNING PARAMETERS FOR THE ALGORITHM.
!
!   DETERMINE CROSSOVER POINT FOR THE SVD WHEN REDUCING
!   a TO BIDIAGONAL FORM. IF max(n,m) EXCEEDS mnthr ,
!   A QR OR LQ FACTORIZATION IS USED FIRST TO REDUCE THE
!   n-BY-m MATRIX a TO A TRIANGULAR FORM.
!    
    mnthr = int( real( mn, stnd )*c1_5, i4b )
!
    two_stage = max( n, m )>=mnthr
!    two_stage = true
!    two_stage = false
!
!   DETERMINE CROSSOVER POINT FOR THE BACK-TRANSFORMATION
!   ALGORITHM WHEN COMPUTING SINGULAR VECTORS. IF max(n,m)
!   EXCEEDS mnthr (E.G. A PRELIMINARY QR OR QL IS DONE) AND ls
!   EXCEEDS mnthr_nsing, THE ORTHOGONAL MATRIX q OF THE QR OR
!   QL FACTORIZATION IS GENERATED AND THE SINGULAR VECTORS
!   ARE COMPUTED BY A MATRIX MULTIPLICATION.
!
    mnthr_nsing = int( real( mn, stnd )*c0_9, i4b )
!
    gen_q = ls>=mnthr_nsing .and. two_stage
!    gen_q = true
!    gen_q = false
!
!   ALLOCATE WORK ARRAYS.
!
    if ( two_stage ) then
        if ( gen_q ) then
            allocate( a(n,m), rla(mn,mn), s(mn), d(mn), e(mn), &
                      tauq(mn), taup(mn), stat=iok )
        else
            allocate( a(n,m), rla(mn,mn), s(mn), d(mn), e(mn), &
                      tauo(mn), tauq(mn), taup(mn), stat=iok )
        end if
    else
        allocate( a(n,m), s(mn), d(mn), e(mn), tauq(mn), taup(mn), stat=iok )
    end if
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-BY-m RANDOM REAL MATRIX a.
!
    call random_number( a )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(n,m), resid(n,mn), resid2(mn), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF THE RANDOM DATA MATRIX .
!
        a2(:n,:m) = a(:n,:m)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   THE SINGULAR VALUE DECOMPOSITION (SVD) OF A REAL n-BY-m MATRIX a
!   IS WRITTEN
!
!                       a = U * S * V**(t)
!
!   WHERE S IS AN n-BY-m MATRIX WHICH IS ZERO EXCEPT FOR ITS
!   min(m,n) DIAGONAL ELEMENTS, U IS AN n-BY-n ORTHOGONAL MATRIX, AND
!   V IS AN m-BY-m ORTHOGONAL MATRIX. THE DIAGONAL ELEMENTS OF S
!   ARE THE SINGULAR VALUES OF a; THEY ARE REAL AND NON-NEGATIVE.
!   THE COLUMNS OF U AND V ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a.
!
!   THIS PROGRAM ILLUSTRATES HOW TO COMPUTE THE ls LARGEST SINGULAR VALUES OF a AND THE
!   ASSOCIATED LEFT AND RIGHT SINGULAR VECTORS OF a (EG A PARTIAL SVD DECOMPOSITION OF a)
!   IN TWO STEPS:
!
!   STEP1 : COMPUTE THE ls LARGEST SINGULAR VALUES OF a BY THE GOLUB AND REINSCH
!   BIDIAGONAL REDUCTION ALGORITHM APPLIED TO a, A BISECTION METHOD APPLIED TO THE
!   RESULTING BIDIAGONAL MATRIX AND STORE INTERMEDIATE RESULTS WITH SUBROUTINE
!   select_singval_cmp2.
!
!   THE OPTIONAL ARGUMENT ls OF select_singval_cmp2 MAY BE USED TO DETERMINE HOW MANY
!   SINGULAR VALUES MUST BE COMPUTED.
!
!   THE OPTIONAL ARGUMENT abstol OF select_singval_cmp2 MAY BE USED TO SET THE REQUIRED
!   PRECISION FOR THE SINGULAR VALUES. A SINGULAR VALUE IS CONSIDERED TO BE LOCATED
!   IF IT HAS BEEN DETERMINED TO LIE IN AN INTERVAL WHOSE WIDTH IS EQUAL OR LESS
!   THAN abstol. SINGULAR VALUES WILL BE COMPUTED MOST ACCURATELY WHEN abstol IS SET
!   TO THE SQUARE ROOT OF THE UNDERFLOW THRESHOLD (EG sqrt(LAMCH('S')) ).
!
!   select_singval_cmp2 IS FASTER THAN select_singval_cmp, BUT MAY BE LESS ACCURATE FOR SOME
!   MATRICES.
!
    if ( two_stage ) then
!
!       STEP1A : REDUCE THE MATRIX a TO BIDIAGONAL FORM BY A TWO-STAGE ALGORITHM.
!       THE MATRIX a IS FIRST REDUCED TO TRIANGULAR FORM BY A QR OR LQ FACTORIZATION
!       IN A FIRST STAGE. THE TRIANGULAR MATRIX IS REDUCED TO BIDIAGONAL FORM IN A
!       SECOND STAGE.
!       THE INTERMEDIATE BIDIAGONAL FORM OF a IS STORED IN d AND e.
!       THE ORTHOGONAL MATRICES USED TO REDUCE a TO BIDIAGONAL FORM IN THE TWO-STAGE 
!       ALGORITHM ARE STORED IN a, rla, tauo, tauq AND taup.
!
        if ( gen_q ) then
!
            call select_singval_cmp2( a, rla, nsing, s, failure=failure1, sort=sort, d=d, e=e, &
                                      ls=ls, abstol=abstol, tauq=tauq, taup=taup )
!
        else
!
            call select_singval_cmp2( a, rla, nsing, s, failure=failure1, sort=sort, d=d, e=e, &
                                      ls=ls, abstol=abstol, tauo=tauo, tauq=tauq, taup=taup )
!
        end if
!
    else
!
!       STEP1B : REDUCE THE MATRIX a TO BIDIAGONAL FORM WITHOUT A QR OR LQ FACTORIZATION.
!       THE INTERMEDIATE BIDIAGONAL FORM OF a IS STORED IN d AND e.
!       THE ORTHOGONAL MATRICES USED TO REDUCE a TO BIDIAGONAL FORM
!       ARE STORED IN a, tauq, taup.
!
        call select_singval_cmp2( a, nsing, s, failure=failure1, sort=sort, d=d, e=e, &
                                  ls=ls, abstol=abstol, tauq=tauq, taup=taup )
!
    end if
!
!   ON EXIT OF select_singval_cmp2 :
!
!   failure= false :  INDICATES SUCCESSFUL EXIT
!   failure= true  :  INDICATES THAT THE ALGORITHM DID NOT CONVERGE AND
!                      THAT FULL ACCURACY WAS NOT ATTAINED IN THE
!                      COMPUTATION OF THE SINGULAR VALUES OF a.
!
!   THE SINGULAR VALUES ARE STORED IN s IN DECREASING ORDER (IF sort='d').
!
!   THE INTERMEDIATE BIDIAGONAL FORM OF a IS STORED IN d AND e.
!   THE ORTHOGONAL MATRICES USED TO REDUCE a TO BIDIAGONAL FORM d_e
!   ARE STORED IN PACKED FORM IN a, tauq, taup AND ALSO rla AND tauo IF A
!   TWO-STAGE ALGORITHM HAS BEEN USED.
!
!   nsing GIVES THE NUMBER OF SINGULAR VALUES REALLY COMPUTED BY select_singval_cmp2.
!   nsing MAY BE GREATER THAN ls IF CLUSTERS OF SINGULAR VALUES ARE PRESENT.
!   THE nsing LARGEST SINGULAR VALUES ARE STORED IN THE FIRST nsing POSITIONS OF s
!   IN DECREASING ORDER IF sort='d'.
!
!   STEP2 : COMPUTE THE nsing ASSOCIATED SINGULAR VALUES OF a BY INVERSE ITERATION WITH SUBROUTINE
!   bd_inviter2 :
!
    if ( nsing>0 ) then
!
!       ALLOCATE WORK VARIABLES NEEDED TO STORE THE SINGULAR VECTORS.
!
        allocate( leftvec(n,nsing),    &
                  rightvec(m,nsing),   &
                  stat = iok    )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE THE FIRST nsing LEFT AND RIGHT SINGULAR VECTORS OF a BY maxiter
!       INVERSE ITERATIONS ON THE INTERMEDIATE BIDIAGONAL MATRIX
!       d_e AND BACK-TRANSFORMATION WITH SUBROUTINE bd_inviter2 .
!
!       ON ENTRY, PARAMETER s CONTAINS SELECTED SINGULAR VALUES OF THE BIDIAGONAL MATRIX d_e (OR a).
!       THE SINGULAR VALUES MUST BE GIVEN IN DECREASING ORDER.
!
        if ( two_stage ) then
!
            if ( gen_q ) then
!
                call bd_inviter2( a, tauq, taup, rla, d, e, s(:nsing), leftvec, rightvec, &
                                 failure=failure2, maxiter=maxiter )
!
            else
!
                call bd_inviter2( a, tauq, taup, rla, d, e, s(:nsing), leftvec, rightvec,  &
                                  failure=failure2, tauo=tauo, maxiter=maxiter )
!
            end if
!
        else
!
            call bd_inviter2( a, tauq, taup, d, e, s(:nsing), leftvec, rightvec,   &
                              failure=failure2, maxiter=maxiter )
!
        end if
!
!       ON EXIT OF bd_inviter2 :
!
!       failure= false :  INDICATES SUCCESSFUL EXIT
!       failure= true  :  INDICATES THAT THE ALGORITHM FAILS TO CONVERGE
!
!       THE SINGULAR VECTORS OF THE BIDIAGONAL FORM d_e ARE COMPUTED USING maxiter INVERSE ITERATIONS
!       FOR ALL THE SINGULAR VALUES AT ONE STEP. THE SINGULAR VECTORS OF d_e ARE THEN ORTHOGONALIZED BY 
!       THE MODIFIED GRAM-SCHMIDT ALGORITHM IF THE SINGULAR VALUES ARE NOT WELL SEPARATED.
!       THE SINGULAR VECTORS OF a ARE FINALLY COMPUTED BY BACK TRANSFORMATIONS.
!       ON EXIT, leftvec AND rightvec CONTAIN THE FIRST nsing SINGULAR VECTORS OF a .
!
!       bd_inviter2 MAY FAIL IF SOME THE SINGULAR VALUES SPECIFIED IN PARAMETER s ARE NEARLY
!       IDENTICAL FOR SOME PATHOLOGICAL MATRICES.
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test .and. nsing>0 ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*V(:m,:nsing) - U(:n,:nsing)*S(:nsing,:nsing),
!       WHERE leftvec AND rightvec ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a.
!
        resid(:n,:nsing) = matmul(a2,rightvec) - leftvec*spread(s(:nsing),dim=1,ncopies=n)
        resid2(:nsing) = norm( resid(:n,:nsing), dim=2_i4b )
!
        err1 = maxval( resid2(:nsing) )/ ( norm( a2 )*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U(:n,:nsing)**(t)*U(:n,:nsing).
!
        call unit_matrix( a2(:nsing,:nsing) )
!
        resid(:nsing,:nsing) = abs( a2(:nsing,:nsing) - matmul( transpose(leftvec), leftvec ) )
!
        err2 = maxval( resid(:nsing,:nsing) )/real(n,stnd)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - V(:m,:nsing)**(t)*V(:m,:nsing).
!
        resid(:nsing,:nsing) = abs( a2(:nsing,:nsing) - matmul( transpose(rightvec), rightvec ) )
        err3 = maxval( resid(:nsing,:nsing) )/real(m,stnd)
!
        err = max( err1, err2, err3 )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
!
        deallocate( a2, resid, resid2 )
!
    end if
!
    if ( nsing>0 ) then
!
        deallocate( leftvec, rightvec )
!
    end if
!
!
    if ( two_stage ) then
!
        if ( gen_q ) then
            deallocate( a, rla, s, d, e, tauq, taup )
        else
            deallocate( a, rla, s, d, e, tauo, tauq, taup )
        end if
!
    else
!
        deallocate( a, s, d, e, tauq, taup )
!
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure1 .and. .not.failure2 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test .and. nsing>0 ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed singular triplets           = ', err1
        write (prtunit,*) 'Orthogonality of the computed left singular vectors  = ', err2
        write (prtunit,*) 'Orthogonality of the computed right singular vectors = ', err3
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing ', nsing, ' singular values and vectors of a ', &
       n, ' by ', m,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_select_singval_cmp2
! ======================================
!
end program ex1_select_singval_cmp2
