program ex1_symtrid_bisect
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine SYMTRID_BISECT
!   in module Eig_Procedures .
!                                                                              
!                                                     
! Further Details
! ===============
!
!    The program also shows the use of subroutine TRID_DEFLATE in module Eig_Procedures.
!
! LATEST REVISION : 08/05/2016
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, one, two, c50,     &
                         trid_deflate, symtrid_bisect, norm, lamch, merror,    &
                         allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n=3000, neig=3000
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of symtrid_bisect'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err, eps, abstol, normt, &
                                               elapsed_time
    real(stnd), allocatable, dimension(:)   :: d, e, eigval, temp, temp2
    real(stnd), allocatable, dimension(:,:) :: resid, eigvec
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: max_qr_steps, neig2, j
!
    logical(lgl)  :: failure1, failure2, do_test, ortho
!   
    character    :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps    = fudge*epsilon( err )
    abstol = sqrt( lamch('s') )
    err    = zero
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( d(n), e(n), eigval(n), eigvec(n,neig), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A (1-2-1) TRIDIAGONAL MATRIX.
!   THE DIAGONAL ELEMENTS ARE STORED IN d .
!   THE OFF-DIAGONAL ELEMENTS ARE STORED IN e .
!
!    d(:n) = two
!    e(:n) = one
!
!    d(:n) = 0.5
!    e(:n) = 0.5
!
    call random_number( d(:n) )
    call random_number( e(:n) )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE neig EIGENVALUES OF THE TRIDIAGONAL MATRIX BY A BISECTION METHOD WITH HIGH ACCURACY.
!   THE EIGENVALUES ARE COMPUTED WITH MAXIMUM ACCURACY WHEN OPTIONAL PARAMETER abstol IS SET TO
!   sqrt( lamch('s') ), WHICH IS THE SQUARE ROOT OF THE UNDERFLOW THRESHOLD.
!
    call symtrid_bisect( d, e, neig2, eigval, failure1, sort=sort, le=neig, abstol=abstol )
!
!   NEXT, COMPUTE THE FIRST nvec EIGENVECTORS OF THE SELF-ADJOINT MATRIX a BY A
!   DEFLATION TECHNIQUE APPLIED ON THE TRIDIAGONAL MATRIX d_e.
!
!   ON ENTRY OF SUBROUTINE trid_deflate, PARAMETER eigval CONTAINS SELECTED
!   EIGENVALUES OF THE TRIDIAGONAL MATRIX.
!   THE EIGENVALUES VALUES MUST BE GIVEN IN DECREASING ORDER.
!
    ortho = false
    max_qr_steps = 4_i4b
!
    call trid_deflate( d(:n), e(:n), eigval(:neig), eigvec(:n,:neig), failure=failure2,     &
                       ortho=ortho, max_qr_steps=max_qr_steps )
!
!   ON EXIT OF trid_deflate :
!
!       failure= false :  INDICATES SUCCESSFUL EXIT.
!       failure= true  :  INDICATES THAT MAXIMUM ACCURACY WAS NOT OBTAIN IN THE ALGORITHM.

!       eigvec CONTAINS THE nvec EIGENVECTORS OF THE TRIDIAGONAL MATRIX ASSOCIATED WITH THE
!       EIGENVALUES eigval(:nvec).
!
!   trid_deflate MAY FAIL IF SOME THE EIGENVALUES SPECIFIED IN PARAMETER eigval ARE NEARLY
!   IDENTICAL AND IF THESE EIGENVALUES ARE NOT COMPUTED WITH HIGH ACCURACY.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( resid(n,neig), temp(n), temp2(n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*u - u*s
!       WHERE s ARE THE EIGENVALUES AND u THE EIGENVECTORS OF THE TRIDIAGONAL MATRIX.
!
        do j=1_i4b, neig
!
            temp(1_i4b:n) = eigvec(1_i4b:n,j)
!
            temp2(1_i4b)         = d(1_i4b)*temp(1_i4b) + e(1_i4b)*temp(2_i4b)
            temp2(2_i4b:n-1_i4b) = e(1_i4b:n-2_i4b)*temp(1_i4b:n-2_i4b)           +   &
                                   d(2_i4b:n-1_i4b)*temp(2_i4b:n-1_i4b)           +   &
                                   e(2_i4b:n-1_i4b)*temp(3_i4b:n)
            temp2(n)             = e(n-1_i4b)*temp(n-1_i4b) + d(n)*temp(n)
!
            resid(1_i4b:n,j) = temp2(1_i4b:n) - eigval(j)*temp(1_i4b:n)
!
        end do
!
        temp(:neig) = norm( resid(1_i4b:n,1_i4b:neig), dim=2_i4b )
        normt = sqrt( sum( d(1_i4b:n)**2 ) + two*(sum( e(1_i4b:n-1_i4b)**2) ) )
!
        err1  = maxval( temp(:neig) )/( normt*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!       WHERE u are THE EIGENVECTORS OF THE TRIDIAGONAL MATRIX a.
!
        resid(:neig,:neig) = matmul( transpose( eigvec(:n,:neig) ), eigvec(:n,:neig) )
!
        do j=1_i4b, neig
            resid(j,j) = resid(j,j) - one
        end do
!
        err2 = maxval( abs( resid(:neig,:neig) ) )/real(n,stnd)
!
        err = max( err1, err2 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( resid, temp, temp2, d, e, eigval, eigvec )
!
    else
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( d, e, eigval, eigvec )
!
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure1 .and. .not.failure2 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed eigenvalues and eigenvectors = ', err1
        write (prtunit,*) 'Orthogonality of the computed eigenvectors            = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing ', neig, ' eigenvalues and eigenvectors of a ', &
       n, ' by ', n,' real symmetric tridiagonal matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_symtrid_bisect
! =================================
!
end program ex1_symtrid_bisect
