program ex1_symtrid_qri2
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine SYMTRID_QRI2
!   in module Eig_Procedures .
!                                                                              
!
! LATEST REVISION : 24/05/2017
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, one, c50,            &
                         allocate_error, merror, symtrid_qri2, norm, unit_matrix
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n=3000
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of symtrid_qri2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err, eps, normt, elapsed_time
    real(stnd), allocatable, dimension(:)   :: d, e, d2, e2, temp, temp2, resid2
    real(stnd), allocatable, dimension(:,:) :: eigvec, resid, id
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: j
!
    logical(lgl) :: failure, do_test
!   
    character    :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : EIGENVALUES AND EIGENVECTORS OF A REAL SYMMETRIC
!               TRIDIAGONAL MATRIX USING THE QR METHOD WITH A
!               PERFECT SHIFT STRATEGY AND A WAVE FRONT ALGORITHM
!               FOR APPLYING GIVENS ROTATIONS.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( eigvec(n,n), d(n), e(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM TRIDIAGONAL MATRIX.
!   THE DIAGONAL ELEMENTS ARE STORED IN d .
!   THE OFF-DIAGONAL ELEMENTS ARE STORED IN e .
!
    call random_number( d(:n) )
    call random_number( e(:n) )

!    d(:n) = 0.5_stnd
!    e(:n) = 0.5_stnd

!    d(:n) = 1._stnd
!    e(:n) = 2._stnd
!
    if ( do_test ) then
!
        allocate( d2(n), e2(n), resid(n,n), resid2(n), id(n,n),  &
                  temp(n), temp2(n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SAVE THE TRIDIAGONAL MATRIX .
!
        d2(:n) = d(:n)
        e2(:n) = e(:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE EIGENVALUES AND EIGENVECTORS OF THE TRIDIAGONAL MATRIX TRID
!   WITH SUBROUTINE symtrid_qri2.
!
!   THE EIGENVALUE DECOMPOSITION OF A REAL n-BY-n SYMMETRIC TRIDIAGONAL MATRIX TRID
!   IS WRITTEN
!
!                       TRID = U * D * U**(t)
!
!   WHERE D IS AN n-BY-n DIAGONAL MATRIX AND U IS AN n-BY-n ORTHOGONAL MATRIX.
!   THE DIAGONAL ELEMENTS OF D ARE THE EIGENVALUES OF TRID; THEY ARE REAL.
!   THE COLUMNS OF U ARE THE EIGENVECTORS OF TRID.
!
!   ON ENTRY OF symtrid_qri2 d AND e CONTAIN, RESPECTIVELY, THE DIAGONAL AND OFF-DIAGONAL
!   OF THE SYMMETRIC TRIDIAGONAL MATRIX TRID. IF THE OPTIONAL ARGUMENT init_mat IS SET TO
!   TRUE, THE MATRIX ARGUMENT eigvec IS INITIALIZED TO THE IDENTITY MATRIX AND THE SUBROUTINE
!   WILL COMPUTE THE EIGENVECTORS OF TRID. IF init_mat IS NOT USED OR SET TO FALSE, THE QR ITERATIONS
!   ARE DIRECTLY APPLIED TO eigvec WITHOUT ANY INITIAIZATION. THIS IS USEFUL FOR COMPUTING THE
!   EIGENVECTORS OF A SYMMETRIC MATRIX.
!
    call symtrid_qri2( d(:n), e(:n), failure, eigvec(:n,:n), sort=sort, init_mat=true )
!
!   THE ROUTINE RETURNS THE EIGENVALUES AND THE EIGENVECTORS OF TRID.
!
!   ON EXIT OF symtrid_qri2 :
!
!       failure= false :  INDICATES SUCCESSFUL EXIT.
!       failure= true  :  INDICATES THAT THE ALGORITHM DID NOT CONVERGE AND
!                         THAT FULL ACCURACY WAS NOT ATTAINED IN THE EIGENDECOMPOSITION 
!                         OF THE TRIDIAGONAL MATRIX TRID .
!
!       eigvec IS OVERWRITTEN WITH THE EIGENVECTORS, STORED COLUMNWISE;
!
!       d IS OVERWRITTEN WITH THE EIGENVALUES OF TRID.
!
!   IF SORT = 'a' OR 'A', THE EIGENVALUES ARE SORTED INTO ASCENDING ORDER.
!   IF SORT = 'd' OR 'D', THE EIGENVALUES ARE SORTED INTO DESCENDING ORDER.
!   THE EIGENVECTORS ARE REARRANGED ACCORDINGLY.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION TRID*U - U*D
!       WHERE D ARE THE EIGENVALUES AND U THE EIGENVECTORS OF THE TRIDIAGONAL MATRIX.
!
        do j=1_i4b, n
!
            temp(1_i4b:n) = eigvec(1_i4b:n,j)
!
            temp2(1_i4b)         = d2(1_i4b)*temp(1_i4b) + e2(1_i4b)*temp(2_i4b)
            temp2(2_i4b:n-1_i4b) = e2(1_i4b:n-2_i4b)*temp(1_i4b:n-2_i4b)           +   &
                                   d2(2_i4b:n-1_i4b)*temp(2_i4b:n-1_i4b)           +   &
                                   e2(2_i4b:n-1_i4b)*temp(3_i4b:n)
            temp2(n)             = e2(n-1_i4b)*temp(n-1_i4b) + d2(n)*temp(n)
!
            resid(1_i4b:n,j) = temp2(1_i4b:n) - d(j)*temp(1_i4b:n)
!
        end do
!
        resid2(:n) = norm( resid(:n,:n), dim=2_i4b )
        normt      = sqrt( sum( d2(1_i4b:n)**2 ) + sum( e2(1_i4b:n-1_i4b)**2 ) )
!
        err1 =  maxval( resid2(:n) )/( normt*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U**(t)*U
!       WHERE U ARE THE EIGENVECTORS OF THE TRIDIAGONAL MATRIX TRID.
!
        call unit_matrix( id(:n,:n) )
!
        resid(:n,:n) = abs( id(:n,:n) - matmul( transpose(eigvec(:n,:n)), eigvec(:n,:n) ) )
!
        err2 = maxval( resid(:n,:n) )/real(n,stnd)
!
        err = max( err1, err2 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( resid, resid2, id, temp, temp2, d2, e2, eigvec, d, e )
!
    else
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( eigvec, d, e )
!
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed eigenvalues and eigenvectors = ', err1
        write (prtunit,*) 'Orthogonality of the computed eigenvectors            = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing ', n, ' eigenvalues and eigenvectors of a ', &
       n, ' by ', n,' real tridiagonal matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_symtrid_qri2
! ===============================
!
end program ex1_symtrid_qri2
