program ex1_trid_deflate
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine TRID_DEFLATE
!   in module Eig_Procedures .
!                                                                              
!                                                     
! Further Details
! ===============
!
!    The program also shows the use of subroutines SYMTRID_CMP and SYMTRID_BISECT
!    in module EIG_Procedures.
!                                                                              
! LATEST REVISION : 23/05/2017
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c50, lamch, symtrid_cmp, trid_deflate,  &
                         symtrid_bisect, norm, unit_matrix, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n=3000, nvec=3000
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 1 of trid_deflate'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err, eps, safmin, abstol, elapsed_time
    real(stnd), allocatable, dimension(:)   :: eigval, resid2, d, e
    real(stnd), allocatable, dimension(:,:) :: a, eigvec, a2
!
    integer       :: iok, istart, iend, irate, imax, itime
    integer(i4b)  :: max_qr_steps, neig
!
    logical(lgl)  :: failure, failure2, do_test, ortho
!   
    character     :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : EIGENVALUES AND, OPTIONALLY, SELECTED EIGENVECTORS OF
!               A REAL SYMMETRIC MATRIX USING BISECTION FOR EIGENVALUES
!               AND A DEFLATION METHOD FOR THE EIGENVECTORS.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
    safmin = lamch( 'S' )
    abstol = sqrt( safmin )
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,n), eigvec(n,nvec), eigval(n), d(n), e(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM DATA MATRIX AND FROM IT
!   A SELF-ADJOINT MATRIX a .
!
    call random_number( a )
    a = a + transpose( a ) 
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(n,n), resid2(nvec), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SAVE RANDOM SELF-ADJOINT MATRIX a .
!
        a2(:n,:n) = a(:n,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   THE EIGENVALUE DECOMPOSITION OF A REAL n-BY-n SYMMETRIC MATRIX a
!   IS WRITTEN
!
!                       a = U * D * U**(t)
!
!   WHERE D IS AN n-BY-n DIAGONAL MATRIX AND U IS AN n-BY-n ORTHOGONAL MATRIX.
!   THE DIAGONAL ELEMENTS OF D ARE THE EIGENVALUES OF a; THEY ARE REAL.
!   THE COLUMNS OF U ARE THE EIGENVECTORS OF a.
!
!   FIRST REDUCE THE SYMMETRIC MATRIX TO SYMMETRIC TRIDIAGONAL FORM BY ORTHOGONAL
!   TRANSFORMATIONS WITH SUBROUTINE symtrid_cmp. THE ORTHOGONAL TRANSFORMATIONS
!   ARE SAVED IF THE OPTIONAL PARAMETER store_q IS SET TO TRUE.
!
    call symtrid_cmp( a(:n,:n), d(:n), e(:n), store_q=true )
!
!   ON EXIT OF symtrid_cmp:
!
!   a IS OVERWRITTEN WITH THE HOUSEHOLDER TRANSFORMATIONS USED TO REDUCE a
!   TO TRIDIAGONAL FORM IF THE OPTIONAL PARAMETER store_q IS SET TO TRUE,
!   OTHERWISE a IS DESTROYED.
!   ARGUMENTS d and e CONTAIN, RESPECTIVELY THE DIAGONAL AND OFF-DIAGONAL
!   ELEMENTS OF THE TRIDIAGONAL MATRIX.
!
!   SECOND, COMPUTE ALL EIGENVALUES OF THE SELF-ADJOINT MATRIX a TO HIGH
!   ACCURACY WITH SUBROUTINE symtrid_bisect.
!
    call symtrid_bisect( d(:n), e(:n), neig, eigval(:n), failure, &
                         sort=sort, abstol=abstol )
!
!   ON EXIT OF symtrid_bisect:
!
!       failure= false :  INDICATES SUCCESSFUL EXIT.
!       failure= true  :  INDICATES THAT THE ALGORITHM DID NOT CONVERGE AND
!                         THAT FULL ACCURACY WAS NOT ATTAINED IN THE EIGENDECOMPOSITION
!                         OF AN INTERMEDIATE TRIDIAGONAL FORM TRID OF a.
!
!       eigval IS OVERWRITTEN WITH THE EIGENVALUES OF THE TRIDIAGONAL MATRIX.
!       IF SORT = 'a' OR 'A', THE EIGENVALUES ARE SORTED INTO ASCENDING ORDER.
!       IF SORT = 'd' OR 'D', THE EIGENVALUES ARE SORTED INTO DESCENDING ORDER.
!
!   NEXT, COMPUTE THE FIRST nvec EIGENVECTORS OF THE SELF-ADJOINT MATRIX a BY A
!   DEFLATION TECHNIQUE APPLIED ON THE INTERMEDIATE TRIDIAGONAL MATRIX
!   AND BACK-TRANSFORMATION.
!
!   ON ENTRY OF SUBROUTINE trid_deflate, PARAMETER eigval CONTAINS SELECTED
!   EIGENVALUES OF THE TRIDIAGONAL MATRIX.
!   THE EIGENVALUES VALUES MUST BE GIVEN IN DECREASING ORDER.
!
    ortho = false
    max_qr_steps = 4_i4b
!
    call trid_deflate( d(:n), e(:n), eigval(:nvec), eigvec(:n,:nvec), failure=failure2,     &
                       mat=a, ortho=ortho, max_qr_steps=max_qr_steps )
!
!   ON EXIT OF trid_deflate :
!
!       failure= false :  INDICATES SUCCESSFUL EXIT.
!       failure= true  :  INDICATES THAT MAXIMUM ACCURACY WAS NOT OBTAIN IN THE ALGORITHM.
!
!       eigvec CONTAINS THE nvec EIGENVECTORS OF a ASSOCIATED WITH THE EIGENVALUES eigval(:nvec).
!
!   trid_deflate MAY FAIL IF SOME THE EIGENVALUES SPECIFIED IN PARAMETER eigval ARE NEARLY
!   IDENTICAL AND IF THESE EIGENVALUES ARE NOT COMPUTED WITH HIGH ACCURACY.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*U - U*D
!
        a(:n,:nvec) = matmul( a2, eigvec ) - eigvec*spread( eigval(:nvec), 1, n )
        resid2(:nvec) = norm( a(:n,:nvec), dim=2_i4b )
!
        err1 =  maxval( resid2(:nvec) )/( norm( a2 )*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U**(t)*U.
!
        call unit_matrix( a(:nvec,:nvec) )
!
        a2(:nvec,:nvec) = abs( a(:nvec,:nvec) - matmul( transpose( eigvec ), eigvec ) )
!
        err2 = maxval(a2(:nvec,:nvec))/real(n,stnd)
!
        err = max( err1, err2 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a, a2, eigvec, eigval, d, e, resid2 )
!
    else
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a, eigvec, eigval, d, e )
!
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure .and. .not.failure2 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed eigenvalues and eigenvectors = ', err1
        write (prtunit,*) 'Orthogonality of the computed eigenvectors            = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing all eigenvalues and ', nvec, ' eigenvectors of a ', &
       n, ' by ', n,' real symmetric matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex1_trid_deflate
! ===============================
!
end program ex1_trid_deflate
