program ex2_lapack_gelsy
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine GELSY
!   in LAPACK software . 
!                                                                              
! LATEST REVISION : 24/12/2015
!                                                                              
! ================================================================================================
!
!
! USED MODULES
! ============
!
    use Lapack_interfaces, only : gelsy
!
    use Statpack, only : lgl, stnd, true, false, zero, c50, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer, parameter :: prtunit = 6, m=4000, n=2000, p=min(m,n), k=max(m,n), nrhs=400
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 2 of gelsy'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)              :: err, eps, rcond, elapsed_time
    real(stnd), allocatable :: ra(:,:), ia(:,:), rb(:,:), ib(:,:), rwork(:)
!
    complex(stnd)              :: work2(1)
    complex(stnd), allocatable :: work(:), a(:,:), a2(:,:), b(:,:), b2(:,:), res(:,:)
!
    integer, allocatable :: jpvt(:)
    integer              :: info, lwork, rank, iok, istart, iend, irate, imax, itime
!
    logical(lgl)   :: do_test
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 2 : MINIMUM-NORM SOLUTION OF A COMPLEX LINEAR LEAST SQUARES PROBLEM
!               WITH SEVERAL RIGHT HAND-SIDES USING A COMPLETE ORTHOGONAL
!               FACTORIZATION OF THE COEFFICIENT MATRIX.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
!
    err = zero
    do_test = false
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n),    ra(m,n),    ia(m,n),     &
              b(k,nrhs), rb(m,nrhs), ib(m,nrhs),  &
              jpvt(n), rwork(2*n), stat=iok       )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A m-by-n COMPLEX RANDOM DATA MATRIX a .
!
    call random_number( ra(:m,:n) )
    call random_number( ia(:m,:n) )
!
    a(:m,:n) = cmplx( ra(:m,:n), ia(:m,:n), stnd )
!
!   GENERATE A m-by-nrhs COMPLEX RANDOM RIGHT HAND-SIDE MATRIX b .
!
    call random_number( rb(:m,:nrhs) )
    call random_number( ib(:m,:nrhs) )
!
    b(:m,:nrhs) = cmplx( rb(:m,:nrhs), ib(:m,:nrhs), stnd )
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate ( ra, ia, rb, ib )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(m,n), b2(m,nrhs), res(m,nrhs), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SAVE DATA MATRIX .
!
        a2(:m,:n) = a(:m,:n)
!
!       SAVE RIGHT HAND SIDE MATRIX .
!
        b2(:m,:nrhs) = b(:m,:nrhs)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   INITIALIZE PERMUTATION VECTOR FOR COMPLETE PIVOTING OF THE
!   COLUMNS OF a .
!
    jpvt(:n) = 0
!
!   rcond IS USED TO DETERMINE THE EFFECTIVE RANK OF a, WHICH IS DEFINED
!   AS THE ORDER OF THE LARGEST LEADING TRIANGULAR SUBMATRIX R11 IN THE 
!   QR FACTORIZATION WITH PIVOTING OF A, WHOSE ESTIMATED CONDITION NUMBER
!   IS < 1/rcond .
!
    rcond = 0.0000001_stnd
!
!   COMPUTE FIRST OPTIMAL WORKSPACE FOR gelsy SUBROUTINE.
!
    lwork = -1
!
    call gelsy( m, n, nrhs, a, m, b, k, jpvt, rcond, rank, work2, lwork, rwork, info )
!
    if ( info==0 ) then
!
        lwork  = int(work2(1))
!
!       ALLOCATE WORK VARIABLE NEEDED BY gelsy SUBROUTINE.
!
        allocate( work(lwork), stat = iok    )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE MINIMAL NORM SOLUTION TO THE LINEAR LEAST SQUARES PROBLEM .
!
        call gelsy( m, n, nrhs, a, m, b, k, jpvt, rcond, rank, work, lwork, rwork, info )
!
!       THE ROUTINE RETURNS THE SOLUTION OF THE LINEAR
!       LEAST SQUARE PROBLEM.
!
!       ON EXIT OF gelsy :
!
!         info=     0 :  INDICATES SUCCESSFUL EXIT
!         info=-i < 0 :  INDICATES THAT the iTH ARGUMENT HAS AN ILLEGAL VALUE
!
!       ON EXIT OF gelsy :
!
!         a IS OVERWRITTEN BY DETAILS OF ITS COMPLETE ORTHOGONAL
!           FACTORIZATION.
!
!         b IS OVERWRITTEN BY THE n-by-nrhs MINIMUM-NORM SOLUTION.
!
!         rank IS OVERWRITTEN BY THE EFFECTIVE RANK OF a.
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate ( work )
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( info/=0 ) then
!
!       ANORMAL EXIT FROM gelsy SUBROUTINE, PRINT A WARNING.
!
        write (prtunit,*) 'Error in the call to GELSY subroutine, Info=', info
!
        if ( info<0 ) then
!
            write (prtunit,'(''Argument '',i3,'' has an illegal value'')') - info
!
        end if
!                
    else if ( do_test ) then
!
!       CHECK THAT THE RESIDUALS ARE ORTHOGONAL TO THE COLUMN VECTORS OF a .
!
        res = b2 - matmul( a2, b(:n,:nrhs) )
        err = maxval( sum(abs(matmul(transpose(conjg(a2)),res)),   dim=1) )/ sum( abs(a2) )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
        deallocate( a, b, jpvt, rwork, a2, b2, res )
    else
        deallocate( a, b, jpvt, rwork )
    end if
!
    if ( err<=eps .and. info==0 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test .and. info==0 ) then
        write (prtunit,*) 
        write (prtunit,*)   &
          'Orthogonality of the residuals against the columns of the coefficient matrix = ', err
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing the solution of a linear least squares complex problem of size ', &
       m, ' by ', n, ' with ', nrhs,' right hand sides is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex2_lapack_gelsy
! ===============================
!
end program ex2_lapack_gelsy
