program ex2_lapack_gesdd
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of the driver subroutine GESDD
!   in LAPACK software . 
!                                                                              
! LATEST REVISION : 27/04/2015
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Lapack_interfaces, only : gesdd
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c50, norm, unit_matrix, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer, parameter :: prtunit = 6, m=2000, n=2000, k=min(m,n)
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 2 of gesdd'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, err3, eps, elapsed_time
    real(stnd)                              :: rwork2(1)
    real(stnd), allocatable, dimension(:)   :: s, rwork
    real(stnd), allocatable, dimension(:,:) :: ra, ia, resid_ortho
!
    complex(stnd)                              :: work2(1)
    complex(stnd), allocatable, dimension(:)   :: work
    complex(stnd), allocatable, dimension(:,:) :: a, a2, c, resid
!
    integer                            :: info, lwork, lrwork, iok, istart, iend, irate, imax, itime
    integer, allocatable, dimension(:) :: iwork
!
    logical(lgl)   :: do_test
!
    character      :: jobz
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 2 : FULL SVD OF A COMPLEX MATRIX USING THE DIVIDE AND CONQUER METHOD.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
!
    err = zero
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n), ra(m,n), ia(m,n), c(k,k), s(k), iwork(8*k), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A m-by-n COMPLEX RANDOM DATA MATRIX .
!
    call random_number( ra(:m,:n) )
    call random_number( ia(:m,:n) )
!
    a(:m,:n) = cmplx( ra(:m,:n), ia(:m,:n), stnd )
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate ( ra, ia )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(m,n), resid(m,k), resid_ortho(k,k), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF THE COMPLEX RANDOM DATA MATRIX .
!
        a2(:m,:n) = a(:m,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   DETERMINE THE WORK TO DO AND HOW THE SINGULAR VECTORS WILL BE STORED.
!   WITH THE ARGUMENTS BELOW, THE SINGULAR VECTORS ARE COMPUTED AND
!   STORED AS IN SUBROUTINE svd_cmp2 IN STATPACK.
!
    jobz  = 'O'
!
!   COMPUTE FIRST OPTIMAL WORKSPACE FOR gesdd SUBROUTINE.
!
    lwork = -1
!
    call gesdd( jobz, m, n, a, m, s, c, k, c, k, work2, lwork, rwork2, iwork, info )
!
    if ( info==0 ) then
!
        lwork  = int(work2(1)) + 10
        lrwork = k*max(5*k+7,2*max(m,n)+2*k+1)
!
!       ALLOCATE WORK VARIABLES NEEDED BY gesdd SUBROUTINE.
!
        allocate( work(lwork), rwork(lrwork), stat = iok    )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       gesdd COMPUTES THE SINGULAR VALUE DECOMPOSITION (SVD) OF A COMPLEX
!       m-BY-n MATRIX a. THE SVD IS WRITTEN
!
!                       a = U * S * V**(h)
!
!       WHERE S IS AN m-BY-n MATRIX WHICH IS ZERO EXCEPT FOR ITS
!       min(m,n) DIAGONAL ELEMENTS, U IS AN m-BY-m UNITARY MATRIX, AND
!       V IS AN n-BY-n UNITARY MATRIX.  THE DIAGONAL ELEMENTS OF S
!       ARE THE SINGULAR VALUES OF a; THEY ARE REAL AND NON-NEGATIVE.
!       THE COLUMNS OF U AND V ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a.
!
        call gesdd( jobz, m, n, a, m, s, c, k, c, k, work, lwork, rwork, iwork, info )
!
!       THE ROUTINE RETURNS THE SINGULAR VALUES, THE LEFT AND RIGHT
!       SINGULAR VECTORS. THE RIGHT SINGULAR VECTORS ARE RETURNED ROWWISE.
!       MORE PRECISELY, THE ROUTINE RETURNS THE CONJUGATE-TRANSPOSE OF
!       THE RIGHT SINGULAR VECTORS.
!
!       ON EXIT OF gesdd :
!
!         info= 0     :  INDICATES SUCCESSFUL EXIT.
!         info=-i < 0 :  INDICATES THAT THE iTH ARGUMENT HAS AN ILLEGAL VALUE.
!         info=i  > 0 :  INDICATES THAT THE ALGORITHM FAILED TO CONVERGE.
!
!       ON EXIT OF gesdd IF JOBZ='O':
!
!         IF m>=n,    a IS OVERWRITTEN WITH THE FIRST min(m,n)
!                     COLUMNS OF U (THE LEFT SINGULAR VECTORS,
!                     STORED COLUMNWISE);
!                     c CONTAINS THE n-BY-n UNITARY MATRIX V**(h).
!
!         IF m<n,     a IS OVERWRITTEN WITH THE FIRST min(m,n)
!                     ROWS OF V**(h) (THE RIGHT SINGULAR VECTORS,
!                     STORED ROWWISE);
!                     c CONTAINS THE m-BY-m UNITARY MATRIX U.
!
!         s IS OVERWRITTEN WITH THE SINGULAR VALUES OF a IN DECREASING ORDER.
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate ( work, rwork )
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( info/=0 ) then
!
!       ANORMAL EXIT FROM gesdd SUBROUTINE, PRINT A WARNING.
!
        write (prtunit,*) 'Error in the call to GESDD subroutine, Info=', info
!                
    else if ( do_test ) then
!
        if ( m>=n ) then
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a(:m,:n)*V(:n,:k) - U(:m,:k)*S(:k,:k).
!
            resid(:m,:k) = matmul(a2(:m,:k),transpose(conjg(c(:k,:k)))) - a(:m,:k)*spread(s,dim=1,ncopies=m)
            resid_ortho(:k,1_i4b) = norm( resid(:m,:k), dim=2_i4b )
            err1 =  maxval( resid_ortho(:k,1_i4b) )/( sum( abs(s) )*real(m,stnd) )
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U**(t)*U.
!
            call unit_matrix( a2(:n,:n) )
!
            resid_ortho(:n,:n) = abs( a2(:n,:n) - matmul( transpose(conjg(a(:m,:n))), a(:m,:n) ) )
            err2 = maxval( resid_ortho(:n,:n) )/real(m,stnd)
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - V**(t)*V.
!
            resid_ortho(:n,:n) = abs( a2(:n,:n) - matmul( c(:n,:n), transpose(conjg(c(:n,:n))) ) )
            err3 = maxval( resid_ortho(:n,:n) )/real(n,stnd)
!
        else
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a(:m,:n)*V(:n,:k) - U(:m,:k)*S(:k,:k).
!
            resid(:m,:k) =  matmul(a2(:m,:n),transpose(conjg(a(:k,:n)))) - c(:k,:k)*spread(s,dim=1,ncopies=k) 
            resid_ortho(:k,1_i4b) = norm( resid(:m,:k), dim=2_i4b )
            err1 =  maxval( resid_ortho(:k,1_i4b) )/( sum( abs(s) )*real(m,stnd) )
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U**(t)*U.
!
            call unit_matrix( a2(:m,:m) )
!
            resid_ortho(:m,:m) = abs( a2(:m,:m) - matmul( transpose(conjg(c(:m,:m ))), c(:m,:m ) ) )
            err2 = maxval( resid_ortho(:m,:m) )/real(m,stnd)
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - V**(t)*V.
!
            resid_ortho(:m,:m) = abs( a2(:m,:m) - matmul( a(:m,:n), transpose(conjg(a(:m,:n))) ) )
            err3 = maxval( resid_ortho(:m,:m) )/real(n,stnd)
!
        end if
!
        err = max( err1, err2, err3 )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
        deallocate( a, c, s, iwork, a2, resid, resid_ortho )
    else
        deallocate( a, c, s, iwork )
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. info==0 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test .and. info==0 ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed singular triplets           = ', err1
        write (prtunit,*) 'Orthogonality of the computed left singular vectors  = ', err2
        write (prtunit,*) 'Orthogonality of the computed right singular vectors = ', err3
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing all singular values and vectors of a', &
       m, ' by', n,' complex matrix is', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex2_lapack_gesdd
! ===============================
!
end program ex2_lapack_gesdd
