program ex2_lapack_syevx
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine SYEVX
!   in LAPACK software . 
!                                                                              
! LATEST REVISION : 08/05/2015
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Lapack_interfaces, only : syevx
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, two, c50, norm,   &
                         merror, allocate_error, lamch, unit_matrix
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer, parameter :: prtunit = 6, n=1000
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 2 of syevx'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, eps, abstol, vl, vu, elapsed_time
    real(stnd), dimension(:),   allocatable :: w, rwork
    real(stnd), dimension(:,:), allocatable :: rz, iz, resid_ortho
!
    complex(stnd)                              :: work2(1)
    complex(stnd), dimension(:),   allocatable :: work
    complex(stnd), dimension(:,:), allocatable :: a2, a, z
!
    integer                            :: info, lwork, iok, il, iu, m,       &
                                          istart, iend, irate, imax, itime
    integer, dimension(:), allocatable :: iwork, ifail
!
    logical(lgl)   :: do_test
!
    character  :: jobz, uplo, range
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 2 : EIGENVALUES AND, OPTIONALLY, EIGENVECTORS OF A COMPLEX HERMITIAN MATRIX
!               MATRIX USING THE BISECTION AND INVERSE ITERATION METHODS.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    abstol = two*lamch('s')
    eps = fudge*epsilon( err )
!
    err = zero
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,n), rz(n,n), iz(n,n), z(n,n), w(n),    &
              rwork(7*n), iwork(5*n), ifail(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-by-n RANDOM COMPLEX DATA MATRIX AND FROM IT
!   A HERMITIAN MATRIX a .
!
    call random_number( rz(:n,:n) )
    call random_number( iz(:n,:n) )
!
    z(:n,:n) = cmplx( rz(:n,:n), iz(:n,:n), stnd )
!
    a = z + transpose( conjg( z ) ) 
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate ( rz, iz )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( a2(n,n), resid_ortho(n,n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF THE RANDOM DATA MATRIX .
!
        a2(:n,:n) = a(:n,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   FIRST DETERMINE THE WORK TO DO.
!
!   IF jobz='N' COMPUTE EIGENVALUES ONLY,
!   IF jobz='V' COMPUTE EIGENVALUES AND EIGENVECTORS.
!
    jobz  = 'V'
!
!   IF range='A' ALL EIGENVALUES WILL BE FOUND,
!   IF range='V' ALL EIGENVALUES IN THE HALF OPEN INTERVAL (vl,vu) WILL BE FOUND,
!   IF range='I' THE il-TH THROUGH iu-TH EIGENVALUES IN ASCENDING ORDER WILL BE FOUND.
!
    range = 'A'
!
!   IF uplo='U' UPPER TRIANGLE OF a IS STORED,
!   IF uplo='L' LOWER TRIANGLE OF a IS STORED.
!
    uplo  = 'U'
!
    vl = zero
    vu = zero
!
    il = 1
    iu = n
!
!   COMPUTE FIRST OPTIMAL WORKSPACE FOR syevx SUBROUTINE.
!
    lwork  = -1
!
    call syevx( jobz, range, uplo, n, a, n, vl, vu, il, iu, abstol,   &
                m, w, z, n, work2, lwork, rwork, iwork, ifail, info   )
!
    if ( info==0 ) then
!
        lwork  = int(work2(1))
!
!       ALLOCATE WORK VARIABLE NEEDED BY syevx SUBROUTINE.
!
        allocate( work(lwork), stat = iok    )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE EIGENVALUES AND EIGENVECTORS OF THE HERMITIAN MATRIX a
!       WITH SUBROUTINE syevx FROM LAPACK.
!       THE EIGENVALUE DECOMPOSITION OF A COMPLEX n-BY-n HERMITIAN MATRIX a
!       IS WRITTEN
!
!                       a = U * D * U**(h)
!
!       WHERE D IS AN n-BY-n DIAGONAL MATRIX AND U IS AN n-BY-n UNITARY MATRIX.
!       THE DIAGONAL ELEMENTS OF D ARE THE EIGENVALUES OF a; THEY ARE REAL.
!       THE COLUMNS OF U ARE THE EIGENVECTORS OF a.
!
        call syevx( jobz, range, uplo, n, a, n, vl, vu, il, iu, abstol,   &
                    m, w, z, n, work, lwork, rwork, iwork, ifail, info    )
!
!       ON EXIT OF syevx :
!
!         info= 0     :  INDICATES SUCCESSFUL EXIT
!         info=-i < 0 :  INDICATES THAT THE iTH ARGUMENT HAS AN ILLEGAL VALUE
!         info=i  > 0 :  INDICATES THAT i EIGENVECTORS FAILED TO CONVERGE AND
!                        THEIR INDICES ARE STORED IN ARRAY ifail.
!
!       ON EXIT OF syevx :
!
!         THE LOWER (IF uplo='L') OR UPPER (IF uplo='U') TRIANGLE
!         OF a IS DESTROYED, INCLUDING THE DIAGONAL.
!
!         m GIVES THE TOTAL NUMBER OF EIGENVALUES/EIGENVECTORS FOUND.
!
!         z IS OVERWRITTEN WITH THE FIRST m EIGENVECTORS
!         OF a (THE EIGENVECTORS ARE STORED COLUMNWISE).
!
!         w IS OVERWRITTEN WITH THE EIGENVALUES OF a IN ASCENDING ORDER,
!         CORRESPONDING TO THE EIGENVECTORS IN z .
!
        deallocate ( work )
!
   end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
   if ( info/=0 ) then
!
!       ANORMAL EXIT FROM syevx SUBROUTINE, PRINT A WARNING.
!
       write (prtunit,*) 'Error in the call to SYEVX subroutine, Info=', info
!                
    else if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*U - U*D
!
        a(:n,:m) = matmul(a2(:n,:n),z(:n,:m)) - z(:n,:m)*spread(w(:m),dim=1,ncopies=n)
        resid_ortho(:m,1_i4b) = norm( a(:n,:m), dim=2_i4b )
        err1 =  maxval( resid_ortho(:m,1_i4b) )/( norm( a2 )*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U**(h)*U.
!
        call unit_matrix( a2(:m,:m) )
!
        resid_ortho(:m,:m) = abs( a2(:m,:m) - matmul( transpose(conjg(z(:n,:m))), z(:n,:m) ) )
        err2 = maxval( resid_ortho(:m,:m) )/real(n,stnd)
!
        err = max( err1, err2 )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
        deallocate( a, z, w, rwork, iwork, ifail, a2, resid_ortho )
    else
       deallocate( a, z, w, rwork, iwork, ifail )
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. info==0 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed eigenvalues and eigenvectors = ', err1
        write (prtunit,*) 'Orthogonality of the computed eigenvectors            = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing ', m, ' eigenvalues and eigenvectors of a', &
       n, ' by', n,' complex hermitian matrix is', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex2_lapack_syevx
! ===============================
!
end program ex2_lapack_syevx
