program ex2_lapack_sytrd
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines SYTRD and ORGTR
!   in LAPACK software . 
!                                                                              
! LATEST REVISION : 27/04/2015
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Lapack_interfaces, only : sytrd, orgtr
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c50, norm,    &
                         unit_matrix, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer, parameter :: prtunit = 6, n=3000
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 2 of sytrd'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err, eps, elapsed_time
    real(stnd), allocatable, dimension(:)   :: d, e
    real(stnd), allocatable, dimension(:,:) :: ra, ia, resid_ortho
!
    complex(stnd)                              :: work2(1)
    complex(stnd), allocatable, dimension(:)   :: tau, work
    complex(stnd), allocatable, dimension(:,:) :: a, a2, resid, trid
!
    integer      :: info_trd, info_q, lwork, lwork_trd, lwork_q,     &
                    iok, istart, iend, irate, imax, itime
    integer(i4b) :: l
!
    logical(lgl) :: do_test
!   
    character    :: uplo
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : TRIDIAGONAL REDUCTION OF A COMPLEX HERMITIAN MATRIX.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,n), ra(n,n), ia(n,n), d(n), e(n-1_i4b), tau(n-1_i4b), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM COMPLEX DATA MATRIX AND FROM IT
!   A HERMITIAN MATRIX a .
!
    call random_number( ra(:n,:n) )
    call random_number( ia(:n,:n) )
!
    a(:n,:n) = cmplx( ra(:n,:n), ia(:n,:n), stnd )
!
    a = a + transpose( conjg( a ) ) 
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate ( ra, ia )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(n,n), trid(n,n), resid(n,n), resid_ortho(n,n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SAVE THE HERMITIAN MATRIX.
!
        a2(:n,:n) = a(:n,:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   IF uplo='U' UPPER TRIANGLE OF a IS STORED,
!   IF uplo='L' LOWER TRIANGLE OF a IS STORED.
!
    uplo = 'U'
!
!   COMPUTE FIRST OPTIMAL WORKSPACE FOR sytrd AND orgtr SUBROUTINES.
!
    lwork = -1
!
    call sytrd( uplo, n, a, n, d, e, tau, work2, lwork, info=info_trd )
!
    lwork_trd = int(work2(1))
!
    call orgtr( uplo, n, a, n, tau, work2, lwork, info=info_q )
!
    lwork_q = int(work2(1))
!
    if ( min(info_trd,info_q)==0 ) then
!
        lwork = max( lwork_trd, lwork_q )
!
!       ALLOCATE OPTIMAL WORK VARIABLE NEEDED BY sytrd AND orgtr SUBROUTINES.
!
        allocate( work(lwork), stat = iok    )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       CALL sytrd AND orgtr TO REDUCE THE MATRIX a TO TRIDIAGONAL FORM
!
!                      a = Q*TRID*Q**(h)
!
!       WHERE Q IS UNITARY AND TRID IS A REAL SYMMETRIC TRIDIAGONAL MATRIX.
!
        call sytrd( uplo, n, a, n, d, e, tau, work(:lwork_trd), lwork_trd, info=info_trd )
!
!       ON EXIT OF sytrd :
!
!         info= 0     :  INDICATES SUCCESSFUL EXIT
!         info=-i < 0 :  INDICATES THAT THE iTH ARGUMENT HAS AN ILLEGAL VALUE
!
!         d AND e ARE THE DIAGONAL AND OFF-DIAGONAL ELEMENTS OF THE REAL SYMMETRIC TRIDIAGONAL MATRIX, RESPECTIVELY.
!         a CONTAINS THE ELEMENTARY REFLECTORS WHICH DEFINE Q IN ITS UPPER TRIANGLE IF uplo='U'.
!         tau CONTAINS THE SCALAR FACTORS OF THE ELEMENTARY REFLECTORS WHICH DEFINE Q.
!
        call orgtr( uplo, n, a, n, tau, work(:lwork_q), lwork_q, info=info_q )
!
!       ON EXIT OF THIS orgtr CALL:
!
!         info= 0     :  INDICATES SUCCESSFUL EXIT
!         info=-i < 0 :  INDICATES THAT THE iTH ARGUMENT HAS AN ILLEGAL VALUE
!
!         a IS OVERWRITTEN WITH THE n-BY-n UNITARY MATRIX Q.
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate ( work )
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( min(info_trd,info_q)/=0 ) then
!
!       ANORMAL EXIT FROM sytrd AND orgtr SUBROUTINES, PRINT A WARNING.
!
        write (prtunit,*) 'Error in the calls to SYTRD or ORGTR subroutines, Info=', min(info_trd,info_q)
!                
    else if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*Q - Q*TRID
!
        trid(:n,:n) = cmplx( zero, zero, stnd )
!
        do l = 1_i4b, n-1_i4b
            trid(l,l)       = cmplx( d(l), zero, stnd )
            trid(l,l+1_i4b) = cmplx( e(l), zero, stnd )
            trid(l+1_i4b,l) = cmplx( e(l), zero, stnd )
        end do
!
        trid(n,n) = cmplx( d(n), zero, stnd )
!
        resid(:n,:n) = matmul( a2(:n,:n), a(:n,:n)  )           &
                       - matmul( a(:n,:n), trid(:n,:n) )
!
        resid_ortho(:n,1_i4b) = norm( resid(:n,:n), dim=2_i4b )
        err1 =  maxval( resid_ortho(:n,1_i4b) )/( norm( a2 )*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - Q**(h)*Q.
!
        call unit_matrix( a2(:n,:n) )
!
        resid_ortho(:n,:n) = abs( a2(:n,:n) - matmul( transpose(conjg(a(:n,:n ))), a(:n,:n) ) )
        err2 = maxval( resid_ortho(:n,:n) )/real(n,stnd)
!
        err = max( err1, err2 )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
        deallocate( a, a2, trid, resid, resid_ortho, d, e, tau )
    else
        deallocate( a, d, e, tau )
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. min(info_trd,info_q)==0  ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test .and. min(info_trd,info_q)==0 ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed tridiagonal decomposition a = Q*TRD*Q**(h) = ', err1
        write (prtunit,*) 'Orthogonality of the computed Q unitary matrix                      = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing the tridiagonal reduction of a ', &
       n, ' by ', n,' complex hermitian matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex2_lapack_sytrd
! ===============================
!
end program ex2_lapack_sytrd
