program ex2_llsq_qr_solve2
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine LLSQ_QR_SOLVE2
!   in module LLSQ_Procedures.
!                                                                              
! LATEST REVISION : 20/01/2021
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, lgl, zero, false, true, llsq_qr_solve2
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, m=4000, n=2000, nrhs=100
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, tol
    real(stnd), allocatable, dimension(:,:) :: x, b, a, a2
!
    integer(i4b) :: krank
!
    integer      :: iok
!
    logical(lgl) :: comp_resid, do_test
!   
!   
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='Example 2 of llsq_qr_solve2'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
    err      = zero
    do_test  = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n), b(m,nrhs), x(n,nrhs), stat=iok )
!
    if ( iok/=0 ) then
        write (prtunit,*)  'Problem in attempt to allocate memory !'
        stop
    end if
!
!   GENERATE A RANDOM COEFFICIENT MATRIX a .
!
    call random_number( a )
!
!   GENERATE A RANDOM RIGHT HAND-SIDE MATRIX b .
!
    call random_number( b )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( a2(m,n), stat=iok )
!
        if ( iok/=0 ) then
            write (prtunit,*)  'Problem in attempt to allocate memory !'
            stop
        end if
!
!       SAVE DATA MATRIX .
!
        a2(:m,:n) = a(:m,:n)
!
    end if
!
!   EXAMPLE 2 : LINEAR LEAST SQUARES SYSTEM.
!
!   COMPUTE SOLUTION MATRIX FOR LINEAR LEAST SQUARES SYSTEM
!
!           a(:m,:n)*x(:n,:nrhs)=b(:m,:nrhs) .
!
!   SET TOLERANCE .
!
    tol = 0.00001_stnd
!
    krank      = 0_i4b
    comp_resid = true
!
    call llsq_qr_solve2( a(:m,:n), b(:m,:nrhs), x(:n,:nrhs),               &
                         comp_resid=comp_resid, krank=krank, tol=tol )
!
!   llsq_qr_solve2 COMPUTES SOLUTION(S) TO REAL LINEAR LEAST SQUARES PROBLEMS 
!   OF THE FORM:
!
!                       Minimize 2-norm(| b - a*x |)
!
!   USING AN ORTHOGONAL FACTORIZATION WITH COLUMNS PIVOTING OF a. a IS A m VECTOR
!   OR AN m-BY-n MATRIX WHICH MAY BE RANK-DEFICIENT. IF a IS A MATRIX, m>=n OR
!   n>m IS PERMITTED.
!
!   SEVERAL RIGHT HAND SIDE VECTORS b AND SOLUTION VECTORS x CAN BE
!   HANDLED IN A SINGLE CALL; THEY ARE STORED AS THE COLUMNS OF THE
!   m-BY-nrhs RIGHT HAND SIDE MATRIX b AND THE n-BY-nrhs SOLUTION
!   MATRIX x, RESPECTIVELY. IN OTHER WORDS, THE ARGUMENTS b AND x MAY
!   BE VECTORS OR MATRICES.
!
!   a AND b ARE OVERWRITTEN BY llsq_qr_solve2. THE OPTIONAL ARGUMENTS krank,
!   tol AND min_norm MUST NOT BE PRESENT IF a IS A m-VECTOR.
!
!   ON INPUT, IF THE OPTIONAL PARAMETER krank IS SPECIFIED AND IF krank=k,
!   THIS IMPLIES THAT THE FIRST k COLUMNS OF a ARE TO BE FORCED INTO THE BASIS.
!   PIVOTING IS PERFORMED ON THE LAST n-k COLUMNS OF a.
!   
!   WHEN krank>=min(m,n) IS USED, PIVOTING IS NOT PERFORMED. THIS IS
!   APPROPRIATE WHEN a IS KNOWN TO BE FULL RANK. krank<=0 CAN BE USED
!   WHEN PIVOTING IS DESIRED ON ALL COLUMNS OF a.
!   
!   BY DEFAULT, PIVOTING IS DONE ON ALL THE COLUMNS OF a.
!   
!   ON EXIT, IF THE OPTIONAL PARAMETER krank IS SPECIFIED,
!   krank CONTAINS THE EFFECTIVE RANK OF a. I.E., THE NUMBER
!   OF INDEPENDENT COLUMNS IN MATRIX a.
!                      
!   IF tol IS PRESENT AND IS IN [0,1[, THEN :
!       ON ENTRY, SPECIFIC CALCULATIONS TO DETERMINE THE EFFECTIVE RANK a
!       ARE PERFORMED. tol IS USED TO DETERMINE THE EFFECTIVE RANK
!       OF a, WHICH IS THEN DEFINED AS THE ORDER OF THE LARGEST LEADING
!       TRIANGULAR SUBMATRIX r11 IN THE QR FACTORIZATION WITH PIVOTING OF a,
!       WHOSE ESTIMATED CONDITION NUMBER IN THE 1-NORM IS LESS THAN 1/tol.
!       IF tol=0 IS SPECIFIED THE NUMERICAL RANK OF a IS DETERMINED.
!       ON EXIT, THE RECIPROCAL OF THE CONDITION NUMBER IS RETURNED IN tol.
!
!   IF tol IS NOT SPECIFIED OR IS OUTSIDE [0,1[ :
!       THE CALCULATIONS TO DETERMINE THE EFFECTIVE RANK OF a ARE NOT
!       PERFORMED AND CRUDE TESTS ON r(j,j) ARE DONE TO DETERMINE
!       THE NUMERICAL RANK OF a.
!
!   IF IT IS POSSIBLE THAT a MAY NOT BE FULL RANK (I.E., CERTAIN COLUMNS OF a ARE
!   LINEAR COMBINATIONS OF OTHER COLUMNS), THEN THE LINEARLY DEPENDENT COLUMNS CAN
!   USUALLY BE DETERMINED BY USING krank=0  AND tol=RELATIVE PRECISION OF THE ELEMENTS
!   IN a. IF EACH ELEMENT IS CORRECT TO, SAY, 5 DIGITS THEN tol=0.00001 SHOULD
!   BE USED. ALSO, IT MAY BE HELPFUL TO SCALE THE COLUMNS OF a SO THAT ALL ELEMENTS 
!   ARE ABOUT THE SAME ORDER OF MAGNITUDE.
!   
!   ON EXIT, IF THE OPTIONAL LOGICAL PARAMETER comp_resid IS PRESENT AND IS SET TO true,
!   THE RESIDUAL VECTOR OR MATRIX b - a*x is COMPUTED AND IS OUTPUT IN b .
!
!   THE 2-NORM OF THE RESIDUAL VECTOR(S) FOR THE SOLUTION VECTOR(S) x, CAN ALSO BE COMPUTED
!   DIRECTLY BY SPECIFYING THE OPTIONAL PARAMETER rnorm IN THE CALL OF llsq_qr_solve2 .
!
!   THE MINIMUN 2-NORM SOLUTION(S) ARE COMPUTED IF THE OPTIONAL LOGICAL PARAMETER min_norm IS
!   PRESENT AND IS SET TO true. OTHERWISE, SOLUTION(S) ARE COMPUTED SUCH THAT IF THE jTH COLUMN
!   OF a IS OMITTED FROM THE BASIS, x[j] OR x[j,:nrhs] IS SET TO ZERO.
!
    if ( do_test ) then
!
!       CHECK THAT THE RESIDUALS ARE ORTHOGONAL TO THE COLUMN VECTORS OF a .
!
        err = maxval( sum( abs( matmul( transpose(b), a2 ) ), dim=2 ) )/ sum( abs(a2) )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a2 )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, b, x )
!
!   PRINT THE RESULT OF THE TEST.
!
    if ( err<=sqrt(epsilon(err)) ) then
        write (prtunit,*) 'Example 2 of LLSQ_QR_SOLVE2 is correct'
    else
        write (prtunit,*) 'Example 2 of LLSQ_QR_SOLVE2 is incorrect'
    end if
!
!
! END OF PROGRAM ex2_llsq_qr_solve2
! =================================
!
end program ex2_llsq_qr_solve2
