program ex2_select_eigval_cmp2
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine SELECT_EIGVAL_CMP2
!   in module Eig_Procedures.
!                                                                              
! LATEST REVISION : 20/01/2020
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, four, c50, allocate_error,    &
                         triangle, merror, get_diag, select_eigval_cmp2, trid_inviter,    &
                         norm, unit_matrix
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, n IS THE DIMENSION OF THE SYMMETRIX MATRIX 
!
    integer(i4b), parameter :: prtunit=6, n=1000, p=n*(n+1)/2
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 2 of select_eigval_cmp2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, eps, val, elapsed_time
    real(stnd), dimension(:),   pointer     :: d
    real(stnd), dimension(:),   allocatable :: res2, vec
    real(stnd), dimension(:,:), allocatable :: a, res, eigvec, d_e
!
    integer(i4b) :: m, maxiter=4
    integer      :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: failure, failure2, do_test, small, upper=true
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : PARTIAL EIGENVALUES/EIGENVECTORS DECOMPOSITION OF A n-BY-n REAL
!               SYMMETRIC MATRIX IN PACKED FORM USING A RATIONAL QR ALGORITHM FOR THE EIGENVALUES
!               AND THE INVERSE ITERATION TECHNIQUE FOR THE EIGENVECTORS.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
!
    err = zero
    do_test = true
!
!   DETERMINE IF YOU WANT TO COMPUTE THE m SMALLEST OR LARGEST EIGENVALUES.
!
    small = false
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,n), vec(p), d_e(n,2), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM DATA MATRIX AND FROM IT
!   A SEMI-POSITIVE MATRIX a .
!
    call random_number( a )
    a = matmul( a, transpose( a ) )
!
!   DETERMINE TRESHOLD FOR THE SUM OF THE EIGENVALUES.
!
    val = sum( get_diag(a) )/four
!
!   MAKE A COPY OF THE UPPER TRIANGLE OF SELF-ADJOINT MATRIX IN PACKED FORM.
!
    vec(:p) = pack( a, mask=triangle( upper, n, n, extra=1_i4b) )
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE THE LARGEST m EIGENVALUES OF THE SELF-ADJOINT MATRIX a (IN PACKED FORM)
!   IN ALEBRAIC VALUE WHOSE SUM EXCEEDS val AND SAVE THE INTERMEDIATE TRIDIAGONAL
!   MATRIX IN PARAMETER d_e .
!
    if ( upper ) then
!
!       USE A FAST ALGORITHM.
!
        call select_eigval_cmp2( vec, d, small, val, failure, d_e=d_e )
!
    else
!
!       USE A SLOW ALGORITHM.
!
        call select_eigval_cmp2( vec, d, small, val, failure, upper=upper, d_e=d_e )
!
    end if
!
!   DETERMINE THE NUMBER OF EIGENVALUES AND ALLOCATE WORK ARRAY FOR
!   COMPUTING THE ASSOCIATED EIGENVECTORS.
!
    m = size( d )
!
    allocate( eigvec(n,m), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    if ( .not. failure ) then
!
!       COMPUTE THE ASSOCIATED m EIGENVECTORS OF THE SELF-ADJOINT MATRIX a BY
!       maxiter INVERSE ITERATIONS ON THE INTERMEDIATE TRIDIAGONAL MATRIX
!       d_e AND BACK-TRANSFORMATION.
!
        call trid_inviter( d_e(:n,1), d_e(:n,2), d(:m), eigvec(:n,:m), failure2,     &
                           matp=vec(:p), maxiter=maxiter )
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test .and. .not.failure ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( res(n,m), res2(m), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*eigvec - eigvec*d
!       WHERE d ARE THE EIGENVALUES AND eigvec THE EIGENVECTORS OF a.
!
        res(:n,:m) = matmul( a(:n,:n), eigvec(:n,:m)) - eigvec(:n,:m)*spread( d(:m), dim=1, ncopies=n)
        res2(:m) = norm( res(:n,:m), dim=2_i4b )
!
        err1 = maxval( res2(:m) )/( norm( a )*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - eigvec**(t)*eigvec.
!
        call unit_matrix( a(:m,:m) )
!
        res(:m,:m) = abs( a(:m,:m) - matmul( transpose(eigvec(:n,:m)), eigvec(:n,:m) ) )
!
        err2 = maxval( res(:m,:m) )/real(n,stnd)
!
        err = max( err1, err2 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( res, res2 )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, vec, eigvec, d_e, d )
!
    if ( err<=eps .and. .not.failure  .and. .not.failure2 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test  .and. .not.failure ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed eigenvalues/eigenvectors couplets = ', err1
        write (prtunit,*) 'Orthogonality of the computed eigenvectors                 = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing ', m, ' eigenvalues and eigenvectors of a ', &
       n, ' by ', n,' real symmetric matrix in packed form is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex2_select_eigval_cmp2
! =====================================
!
end program ex2_select_eigval_cmp2
