program ex2_select_singval_cmp4
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine SELECT_SINGVAL_CMP4
!   in module SVD_Procedures.
!                                                                              
!                                                     
! Further Details
! ===============
!
!    The program also shows the use of subroutine BD_DEFLATE2 in module SVD_Procedures.
!                                                                              
! LATEST REVISION : 25/10/2019
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c0_9, c50, lamch, bd_deflate2,  &
                         select_singval_cmp4, merror, allocate_error, norm, unit_matrix
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n=2000, m=1000, ls=400
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 2 of select_singval_cmp4'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, err3, eps, abstol, elapsed_time
    real(stnd), dimension(:,:), allocatable :: a, a2, p, leftvec, rightvec, resid, ra
    real(stnd), dimension(:),   allocatable :: s, d, e, tauo, resid2
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: max_qr_steps, nsing, mnthr_nsing
!
    logical(lgl) :: failure1, failure2, ortho, gen_p, do_test
!   
    character    :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 2 : PARTIAL SVD OF A n-BY-m REAL MATRIX WITH n>=m USING THE RHALA-BARLOW
!               ONE-SIDED BIDIAGONAL REDUCTION ALGORITHM, A BISECTION ALGORITHM FOR
!               SINGULAR VALUES AND THE GODUNOV DEFLATION TECHNIQUE FOR SINGULAR VECTORS.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps    = fudge*epsilon( err )
    abstol = sqrt( lamch('S') )
    err    = zero
!
!   SPECIFY IF NUMERICAL TESTS ARE PERFORMED.
!
    do_test = true
!
!   CHOOSE TUNING PARAMETERS FOR THE ALGORITHM.
!
!   DETERMINE CROSSOVER POINT FOR THE BACK-TRANSFORMATION
!   ALGORITHM WHEN COMPUTING RIGHT SINGULAR VECTORS. IF
!   ls EXCEEDS mnthr_nsing, THE RIGHT ORTHOGONAL MATRIX p
!   OF THE ONE-SIDED BIDIAGONAL REDUCTION ALGORITHM IS GENERATED
!   AND BACK-TRANSFORMATION IS DONE BY A MATRIX MULTIPLICATION.
!
    mnthr_nsing = int( real( m, stnd )*c0_9, i4b )
!
    gen_p = ls>=mnthr_nsing
!    gen_p = true
!    gen_p = false
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,m), s(m), d(m), e(m), &
              p(m,m), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-BY-m RANDOM REAL MATRIX a.
!
    call random_number( a )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a2(n,m), resid(n,m), resid2(n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF THE RANDOM DATA MATRIX .
!
        a2(:n,:m) = a(:n,:m)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   THE SINGULAR VALUE DECOMPOSITION (SVD) OF A REAL n-BY-m MATRIX a
!   IS WRITTEN
!
!                       a = U * S * V**(t)
!
!   WHERE S IS AN n-BY-m MATRIX WHICH IS ZERO EXCEPT FOR ITS
!   min(m,n) DIAGONAL ELEMENTS, U IS AN n-BY-n ORTHOGONAL MATRIX, AND
!   V IS AN m-BY-m ORTHOGONAL MATRIX. THE DIAGONAL ELEMENTS OF S
!   ARE THE SINGULAR VALUES OF a; THEY ARE REAL AND NON-NEGATIVE.
!   THE COLUMNS OF U AND V ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a.
!
!   THIS PROGRAM ILLUSTRATES HOW TO COMPUTE THE ls LARGEST SINGULAR VALUES OF a AND THE
!   ASSOCIATED LEFT AND RIGHT SINGULAR VECTORS OF a (EG A PARTIAL SVD DECOMPOSITION OF a)
!   IN TWO STEPS:
!
!   STEP1 : COMPUTE THE ls LARGEST SINGULAR VALUES OF a BY THE RHALA-BARLOW ONE-SIDED
!   BIDIAGONAL REDUCTION ALGORITHM APPLIED TO a, A BISECTION METHOD APPLIED TO THE
!   THE RESULTING BIDIAGONAL MATRIX AND STORE INTERMEDIATE RESULTS WITH SUBROUTINE
!   select_singval_cmp4.
!
!   THE OPTIONAL ARGUMENT ls OF select_singval_cmp4 MAY BE USED TO DETERMINE HOW MANY
!   SINGULAR VALUES MUST BE COMPUTED.
!
!   THE OPTIONAL ARGUMENT abstol OF select_singval_cmp4 MAY BE USED TO SET THE REQUIRED
!   PRECISION FOR THE SINGULAR VALUES. A SINGULAR VALUE IS CONSIDERED TO BE LOCATED
!   IF IT HAS BEEN DETERMINED TO LIE IN AN INTERVAL WHOSE WIDTH IS EQUAL OR LESS
!   THAN abstol. SINGULAR VALUES WILL BE COMPUTED MOST ACCURATELY WHEN abstol IS SET
!   TO THE SQUARE ROOT OF THE UNDERFLOW THRESHOLD (EG sqrt(LAMCH('S')) ).
!
!   select_singval_cmp4 IS FASTER THAN select_singval_cmp3, BUT MAY BE LESS ACCURATE
!   FOR SOME MATRICES.
!
    call select_singval_cmp4( a, nsing, s, failure=failure1, sort=sort, d=d, e=e, &
                              ls=ls, abstol=abstol, p=p, gen_p=gen_p )
!
!   ON EXIT OF select_singval_cmp4 :
!
!   failure= false :  INDICATES SUCCESSFUL EXIT
!   failure= true  :  INDICATES THAT THE ALGORITHM DID NOT CONVERGE AND
!                     THAT FULL ACCURACY WAS NOT ATTAINED IN THE
!                     COMPUTATION OF THE SINGULAR VALUES OF a.
!
!   THE INTERMEDIATE BIDIAGONAL FORM OF a IS STORED IN d AND e.
!   THE ORTHOGONAL MATRICES USED TO REDUCE a TO BIDIAGONAL FORM ARE STORED IN a AND p .
!
!   THE SINGULAR VALUES ARE STORED IN s IN DECREASING ORDER (IF sort='d').
!
!   nsing GIVES THE NUMBER OF SINGULAR VALUES REALLY COMPUTED BY select_singval_cmp4.
!   nsing MAY BE GREATER THAN ls IF CLUSTERS OF SINGULAR VALUES ARE PRESENT.
!   THE nsing LARGEST SINGULAR VALUES ARE STORED IN THE FIRST nsing POSITIONS OF s
!   IN DECREASING ORDER IF sort='d'.
!
!   STEP2 : COMPUTE nsing LEFT AND RIGHT SINGULAR VECTORS OF a BY A DEFLATION TECHNIQUE
!   APPLIED TO THE INTERMEDIATE BIDIAGONAL MATRIX AND BACK-TRANSFORMATION WITH SUBROUTINE
!   bd_deflate2.
!
    if ( nsing>0 ) then
!
!       ALLOCATE WORK VARIABLES NEEDED TO STORE THE SINGULAR VECTORS.
!
        allocate( leftvec(n,nsing),    &
                  rightvec(m,nsing),   &
                  stat = iok    )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE THE FIRST nsing LEFT AND RIGHT SINGULAR VECTORS OF a BY 
!       A DEFLATION TECHNIQUE APPLIED ON THE INTERMEDIATE BIDIAGONAL MATRIX
!       d_e AND BACK-TRANSFORMATION WITH SUBROUTINE bd_deflate2 .
!
!       ON ENTRY OF bd_deflate2: PARAMETER s CONTAINS SELECTED SINGULAR VALUES OF THE BIDIAGONAL
!       MATRIX (OR OF a). THE SINGULAR VALUES MUST BE GIVEN IN DECREASING ORDER.
!
!       OPTIONAL PARAMETER ortho DETERMINES IF DEFLATION IS APPLIED SEQUENTIALLY TO ALL
!       COMPUTED SINGULAR VECTORS (ortho=true) OR ONLY IN EACH CLUSTER OF SINGULAR VALUES
!       (ortho=false). IN THE LATTER CASE, ORTHOGONALITY OF THE SINGULAR VECTORS MAY BE
!       SLIGHTLY DEGRADED, BUT COMPUTATIONS WILL BE MUCH FASTER.
!
        ortho = false
        max_qr_steps = 4_i4b
!
        call bd_deflate2( a(:n,:m), p(:m,:m), d(:m), e(:m), s(:nsing),  &
                          leftvec(:n,:nsing), rightvec(:m,:nsing),      &
                          failure=failure2, ortho=ortho,                &
                          max_qr_steps=max_qr_steps                     )
!
!       ON EXIT OF bd_deflate2 :
!
!           failure= false :  INDICATES SUCCESSFUL EXIT.
!           failure= true  :  INDICATES THAT MAXIMUM ACCURACY WAS NOT OBTAIN IN THE
!                             DEFLATION ALGORITHM.

!           leftvec AND rightvec CONTAIN, RESPECTIVELY, THE FIRST nsing LEFT AND RIGHT
!           SINGULAR VECTORS OF a .
!
!       bd_deflate2 MAY FAIL IF SOME THE SINGULAR VALUES SPECIFIED IN PARAMETER s ARE NEARLY
!       IDENTICAL FOR SOME PATHOLOGICAL MATRICES.
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test .and. nsing>0 ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*v(:m,:nsing) - u(:n,:nsing)*s(:nsing,:nsing),
!       WHERE u AND v ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a.
!
        resid(:n,:nsing) = matmul(a2,rightvec) - leftvec*spread(s(:nsing),dim=1,ncopies=n)
        resid2(:nsing) = norm( resid(:n,:nsing), dim=2_i4b )
!
        err1 = maxval( resid2(:nsing) )/( norm( a2 )*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u(:n,:nsing)**(t)*u(:n,:nsing).
!
        call unit_matrix( a2(:nsing,:nsing) )
!
        resid(:nsing,:nsing) = abs( a2(:nsing,:nsing) - matmul( transpose(leftvec), leftvec ) )
!
        err2 = maxval( resid(:nsing,:nsing) )/real(n,stnd)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - v(:m,:nsing)**(t)*v(:m,:nsing).
!
        resid(:nsing,:nsing) = abs( a2(:nsing,:nsing) - matmul( transpose(rightvec), rightvec ) )
!
        err3 = maxval( resid(:nsing,:nsing) )/real(m,stnd)
!
        err = max( err1, err2, err3 )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
!
        deallocate( a2, resid, resid2 )
!
    end if
!
    if ( nsing>0 ) then
!
        deallocate( a, s, d, e, p, leftvec, rightvec )
!
    else
!
        deallocate( a, s, d, e, p )
!
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure1 .and. .not.failure2 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test .and. nsing>0 ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed singular triplets           = ', err1
        write (prtunit,*) 'Orthogonality of the computed left singular vectors  = ', err2
        write (prtunit,*) 'Orthogonality of the computed right singular vectors = ', err3
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing ', nsing, ' singular values and vectors of a ', &
       n, ' by ', m,' real matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex2_select_singval_cmp4
! ======================================
!
end program ex2_select_singval_cmp4
