program ex2_svd_cmp
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine SVD_CMP
!   in module SVD_Procedures .
!                                                                              
!                                                     
! Further Details
! ===============
!
!    The program also shows the use of subroutine BD_INVITER2 in module SVD_Procedures.
!                                                                              
! LATEST REVISION : 28/04/2015
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, zero, true, false, bd_inviter2, svd_cmp, &
                         norm, unit_matrix, c50, c100, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n=1000, m=1000, mn=min(m,n)
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 2 of svd_cmp'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, err3, eps, tmp, tmp2, &
                                               min_explnorm, elapsed_time
    real(stnd), allocatable, dimension(:)   :: s, d, e, tauq, taup
    real(stnd), allocatable, dimension(:,:) :: a, a2, leftvec, rightvec, resid
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: maxiter=2, nsing
!
    logical(lgl) :: failure1, failure2, do_test
!   
    character    :: sort='d'
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : ALL SINGULAR VALUES AND nsing SINGULAR VECTORS OF A n-BY-m REAL MATRIX
!               BY THE INVERSE ITERATION METHOD (eg PARTIAL SVD DECOMPOSITION). nsing
!               IS DETERMINED AS THE NUMBER OF SINGULAR TRIPLETS WHICH DESCRIBED AT LEAST
!               90% OF THE FROBENIUS NORM OF THE REAL MATRIX.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(n,m), s(mn), d(mn), e(mn), tauq(mn), taup(mn), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-BY-m RANDOM REAL MATRIX a.
!
    call random_number( a )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( a2(n,m), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A COPY OF THE RANDOM DATA MATRIX .
!
        a2(:n,:m) = a(:n,:m)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   THE SINGULAR VALUE DECOMPOSITION (SVD) OF A REAL n-BY-m MATRIX a
!   IS WRITTEN
!
!                       a = U * S * V**(t)
!
!   WHERE s IS AN n-BY-m MATRIX WHICH IS ZERO EXCEPT FOR ITS
!   min(m,n) DIAGONAL ELEMENTS, u IS AN n-BY-n ORTHOGONAL MATRIX, AND
!   v IS AN m-BY-m ORTHOGONAL MATRIX.  THE DIAGONAL ELEMENTS OF s
!   ARE THE SINGULAR VALUES OF a; THEY ARE REAL AND NON-NEGATIVE.
!   THE COLUMNS OF U AND V ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a.
!
!   THIS PROGRAM ILLUSTRATES HOW TO COMPUTE ALL THE SINGULAR VALUES OF a AND
!   ONLY nsing LEFT AND RIGHT SINGULAR VECTORS OF a (EG A PARTIAL SVD DECOMPOSITION OF a)
!   IN TWO STEPS. nsing IS DETERMINED AS THE NUMBER OF SINGULAR TRIPLETS WHICH DESCRIBED
!   AT LEAST 90% OF THE FROBENIUS NORM OF a.
!
!   SEE EXAMPLE ex1_svd_cmp.f90, IF YOU WANT TO COMPUTE A FULL SVD OF a.  
!   THIS PROGRAM SHOWS HOW TO COMPUTE A PARTIAL SVD OF a WITH THE SUBROUTINES
!   svd_cmp AND bd_inviter2.
!
!   STEP1 : COMPUTE ALL SINGULAR VALUES OF a AND STORE INTERMEDIATE RESULTS WITH SUBROUTINE svd_cmp.
!
    call svd_cmp( a, s, failure=failure1, sort=sort, d=d, e=e, tauq=tauq, taup=taup  )
!
!   ON EXIT OF svd_cmp :
!
!   failure= false :  INDICATES SUCCESSFUL EXIT
!   failure= true  :  INDICATES THAT THE ALGORITHM DID NOT CONVERGE AND
!                     THAT FULL ACCURACY WAS NOT ATTAINED IN THE BIDIAGONAL
!                     SVD OF AN INTERMEDIATE BIDIAGONAL FORM B OF a.
!
!   s IS OVERWRITTEN WITH THE SINGULAR VALUES OF a.
!
!   IF SORT = 'a' OR 'A', THE SINGULAR VALUES ARE SORTED INTO ASCENDING ORDER.
!   IF SORT = 'd' OR 'D', THE SINGULAR VALUES ARE SORTED INTO DESCENDING ORDER.
!   HERE, SORT = 'd' IS USED THIS IS REQUIRED FOR THE USE OF bd_inviter2 .
!                
!   IF THE PARAMETER v IS ABSENT, svd_cmp COMPUTES ONLY THE SINGULAR VALUES OF a             
!   AND, OPTIONALLY, STORES THE INTERMEDIATE BIDIAGONAL FORM OF a AND THE ORTHOGONAL
!   MATRICES USED TO REDUCE a TO BIDIAGONAL FORM.
!
!   THE INTERMEDIATE BIDIAGONAL FORM OF a IS STORED IN THE OPTIONAL ARGUMENTS d AND e.
!   THE ORTHOGONAL MATRICES USED TO REDUCE a TO BIDIAGONAL FORM d_e ARE STORED 
!   IN mat, tauq, taup, WHEN THE OPTIONAL ARGUMENTS tauq AND taup ARE PRESENT.
!
!   NOW SELECT THE NUMBER OF SINGULAR TRIPLETS SUCH THAT THEY DESCRIBE AT LEAST 90%
!   OF THE NORM OF THE REAL MATRIX a.
!
    min_explnorm = 90._stnd
!
    tmp  = zero
    tmp2 = c100/sum( s(:mn)**2 )
!
    do nsing= 1_i4b, mn
!
        tmp = tmp + tmp2*s(nsing)**2
        if ( tmp>=min_explnorm ) exit
!
    end do
!
!   CHECK THE nsing VALUE.
!
    nsing = min( nsing, mn )
!
!   ALLOCATE WORK ARRAYS FOR THE SINGULAR VECTORS.
!
    allocate( leftvec(n,nsing), rightvec(m,nsing), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   STEP2 : COMPUTE THE FIRST nsing LEFT AND RIGHT SINGULAR VECTORS OF a BY maxiter
!   INVERSE ITERATIONS ON THE INTERMEDIATE BIDIAGONAL MATRIX
!   d_e AND BACK-TRANSFORMATION WITH SUBROUTINE bd_inviter2 .
!
    call bd_inviter2( a, tauq, taup, d, e, s(:nsing), leftvec, rightvec, failure=failure2, maxiter=maxiter )
!
!   ON ENTRY, PARAMETER s CONTAINS SELECTED SINGULAR VALUES OF THE BIDIAGONAL MATRIX d_e (OR a).
!   THE SINGULAR VALUES MUST BE GIVEN IN DECREASING ORDER.
!
!   ON EXIT OF bd_inviter2 :
!
!   failure= false :  INDICATES SUCCESSFUL EXIT.
!   failure= true  :  INDICATES THAT THE ALGORITHM FAILS TO CONVERGE.
!
!   THE SINGULAR VECTORS OF THE BIDIAGONAL FORM d_e ARE COMPUTED USING maxiter INVERSE ITERATIONS
!   FOR THE nsing SINGULAR VALUES AT ONE STEP. THE SINGULAR VECTORS OF d_e ARE THEN ORTHOGONALIZED BY 
!   THE MODIFIED GRAM-SCHMIDT ALGORITHM IF NECESSARY. THE SINGULAR VECTORS OF a ARE FINALLY COMPUTED 
!   BY BACK TRANSFORMATIONS. ON EXIT, leftvec AND rightvec CONTAIN THE FIRST nsing LEFT AND RIGHT
!   SINGULAR VECTORS OF a, RESPECTIVELY.
!
!   NOTE THAT bd_inviter2 MAY FAIL IF SOME THE SINGULAR VALUES SPECIFIED IN PARAMETER s ARE NEARLY
!   IDENTICAL.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( resid(n,nsing), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*V(:m,:nsing) - U(:n,:nsing)*S(:nsing,:nsing),
!       WHERE leftvec AND rightvec ARE THE LEFT AND RIGHT SINGULAR VECTORS OF a.
!
        resid(:n,:nsing) = matmul(a2,rightvec) - leftvec*spread(s(:nsing),dim=1,ncopies=n)
        a2(:nsing,1_i4b) = norm( resid(:n,:nsing), dim=2_i4b )
        err1 = maxval( a2(:nsing,1_i4b) )/ ( sum( abs(s(:mn)) )*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U(:n,:nsing)**(t)*U(:n,:nsing).
!
        call unit_matrix( a2(:nsing,:nsing) )
!
        resid(:nsing,:nsing) = abs( a2(:nsing,:nsing) - matmul( transpose(leftvec), leftvec ) )
        err2 = maxval( resid(:nsing,:nsing) )/real(n,stnd)
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - V(:m,:nsing)**(t)*V(:m,:nsing).
!
        resid(:nsing,:nsing) = abs( a2(:nsing,:nsing) - matmul( transpose(rightvec), rightvec ) )
        err3 = maxval( resid(:nsing,:nsing) )/real(m,stnd)
!
        err = max( err1, err2, err3 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a, a2, resid, leftvec, rightvec, s, d, e, tauq, taup )
!
    else
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a, leftvec, rightvec, s, d, e, tauq, taup )
!
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure1 .and. .not.failure2 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed singular triplets           = ', err1
        write (prtunit,*) 'Orthogonality of the computed left singular vectors  = ', err2
        write (prtunit,*) 'Orthogonality of the computed right singular vectors = ', err3
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing all singular values and ', nsing, ' singular vectors of a', &
       n, ' by', m,' real matrix is', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex2_svd_cmp
! ==========================
!
end program ex2_svd_cmp
