program ex2_symtrid_cmp2
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutines SYMTRID_CMP2 and
!   ORTHO_GEN_SYMTRID in module EIG_Procedures.
!                                                                              
! LATEST REVISION : 10/09/2019
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c50, symtrid_cmp2, &
                         ortho_gen_symtrid, symtrid_qri2, norm, unit_matrix,   &
                         merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, m AND n ARE THE DIMENSIONS OF THE MATRIX USED
! TO COMPUTE THE MATRIX CROSS-PRODUCT, m MUST BE GREATER THAN n, OTHERWISE
! symtrid_cmp2 WILL STOP WITH AN ERROR MESSAGE.
!
    integer(i4b), parameter :: prtunit=6, m=3000, n=1000
!   
    real(stnd), parameter  :: fudge=c50
!
    character, parameter :: sort='d'
!
    character(len=*), parameter :: name_proc='Example 2 of symtrid_cmp2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err, eps, elapsed_time
    real(stnd), allocatable, dimension(:)   :: d, e
    real(stnd), allocatable, dimension(:,:) :: a, at, ata, resid
!
    integer :: iok, istart, iend, irate, imax, itime
!
    logical(lgl) :: do_test, failure
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 1 : EIGEN DECOMPOSITION OF A REAL SYMMETRIC MATRIX PRODUCT,
!               a**(t)*a, USING THE ONE-SIDED RALHA METHOD.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(m,n), d(n), e(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A m-by-n RANDOM DATA MATRIX a .
!
    call random_number( a )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( ata(n,n), at(n,m), resid(n,n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        at(:n,:m) = transpose( a(:m,:n) )
!
!       COMPUTE THE SYMMETRIC MATRIX CROSS-PRODUCT.
!
        ata(:n,:n) = matmul( at(:n,:m), a(:m,:n) )
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   CALL symtrid_cmp2 AND ortho_gen_symtrid TO REDUCE THE MATRIX CROSS-PRODUCT TO TRIDIAGONAL FORM
!
!                      a**(t)*a = Q*TRID*Q**(t)
!
!   WHERE Q IS ORTHOGONAL AND TRID IS A SYMMETRIC TRIDIAGONAL MATRIX.
!
!   ON ENTRY OF symtrid_cmp2, a MUST CONTAINS THE INITIAL m-by-n MATRIX USED
!   FOR COMPUTING THE MATRIX CROSS-PRODUCT. THE ORTHOGONAL MATRIX Q IS STORED
!   IN FACTORED FORM IF THE LOGICAL ARGUMENT store_q IS SET TO true.
!
    call symtrid_cmp2( a(:m,:n), d, e, store_q=true )
!
!   ON EXIT OF symtrid_cmp2:
!
!         ARGUMENTS d AND e CONTAIN, RESPECTIVELY, THE DIAGONAL AND
!         OFF-DIAGONAL ELEMENTS OF THE TRIDIAGONAL MATRIX TRID.
!
!         IF THE LOGICAL ARGUMENT store_q IS SET TO TRUE ON ENTRY,
!         THE LEADING n-BY-n LOWER TRIANGULAR PART OF a IS OVERWRITTEN
!         BY THE MATRIX Q AS A PRODUCT OF ELEMENTARY REFLECTORS.
!
!   ortho_gen_symtrid GENERATES THE MATRIX Q STORED AS A PRODUCT OF
!   ELEMENTARY REFLECTORS AFTER A CALL TO symtrid_cmp2 WITH store_q=true.
!
    call ortho_gen_symtrid( a(:n,:n), false )
!
!   COMPUTE EIGENDECOMPOSITION OF a**(t)*a WITH SUBROUTINE symtrid_qri2:
!
!                            a**(t)*a = U*D*U**(t)
!
!   WHERE U ARE THE EIGENVECTORS AND D IS THE DIAGONAL MATRIX, WITH
!   EIGENVALUES ON THE DIAGONAL.
!
    call symtrid_qri2( d(:n), e(:n), failure, a(:n,:n), sort=sort )
!
!   ON EXIT OF symtrid_qri2:
!
!         ARGUMENTS d AND a(:n:n) CONTAIN, RESPECTIVELY, THE EIGENVALUES AND
!         EIGENVECTORS OF a**(t)*a.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*U - U*D
!
        resid(:n,:n) = matmul( ata(:n,:n), a(:n,:n)  )    &
                       - a(:n,:n)*spread( d(:n), 1, n )
!
        e(:n) = norm( resid(:n,:n), dim=2_i4b )
        err1 =  maxval( e(:n) )/( norm( ata )*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U**(t)*U.
!
        call unit_matrix( ata(:n,:n) )
!
        at(:n,:n) = transpose( a(:n,:n) )
!
        resid(:n,:n) = abs( ata(:n,:n) - matmul( at(:n,:n), a(:n,:n) ) )
        err2 = maxval( resid(:n,:n) )/real(n,stnd)
!
        err = max( err1, err2 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( ata, resid, at )
!
    endif
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, d, e )
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed eigendecomposition a**(t)*a = U*D*U**(t) = ', err1
        write (prtunit,*) 'Orthogonality of the computed eigenvectors U**(t)*U - I           = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing the eigendecomposition of a ',       &
       n, ' by ', n,' real symmetric matrix product is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex2_symtrid_cmp2
! ===============================
!
end program ex2_symtrid_cmp2
