program ex2_trid_inviter
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine TRID_INVITER
!   in module Eig_Procedures .
!                                                     
! Further Details
! ===============
!
!    The program also shows the use of subroutine EIGVAL_CMP in module EIG_Procedures.
!                                                                              
! LATEST REVISION : 08/05/2016
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, c50, triangle, trid_inviter,   &
                         eigval_cmp, norm, unit_matrix, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n=3000, p=(n*(n+1))/2, nvec=3000
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 2 of trid_inviter'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err, eps, elapsed_time
    real(stnd), allocatable, dimension(:,:) :: a, resid, eigvec, d_e
    real(stnd), allocatable, dimension(:)   :: a_packed, eigval, resid2
!
    integer       :: iok, istart, iend, irate, imax, itime
    integer(i4b)  :: maxiter=2, i
!
    logical(lgl)  :: failure1, failure2, do_test, upper=true
!   
    character     :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE EXAMPLE.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 2 : EIGENVALUES AND, OPTIONALLY, SELECTED EIGENVECTORS OF
!               A REAL SYMMETRIC MATRIX STORED IN PACKED FORM USING
!               THE INVERSE ITERATION METHOD.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    err = zero
!
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a_packed(p), eigvec(n,nvec), eigval(n), d_e(n,2), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM SELF-ADJOINT MATRIX a IN PACKED FORM.
!
    call random_number( a_packed(:p) )
!
!   MAKE a POSITIVE DEFINITE ASSUMING THAT THE UPPER TRIANGLE OF
!   THE SELF-ADJOINT MATRIX IS STORED IN PACKED FORM.
!
    do i = 1_i4b, n
        a_packed(i+((i-1_i4b)*i/2_i4b)) = real( n, stnd )
    end do
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( a(n,n), resid(n,nvec), resid2(nvec), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       UNPACKED AND SAVE RANDOM SELF-ADJOINT MATRIX a 
!       ASSUMING THAT THE UPPER TRIANGLE OF THE SELF-ADJOINT
!       MATRIX IS STORED IN PACKED FORM.
!
        a(:n,:n) = unpack( a_packed(:p), mask=triangle( true, n, n, extra=1_i4b), field=zero )
!
        do i = 1_i4b, n-1_i4b
             a(i+1_i4b:n,i) = a(i,i+1_i4b:n)
        end do
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   THE EIGENVALUE DECOMPOSITION OF A REAL n-BY-n SYMMETRIC MATRIX a
!   IS WRITTEN
!
!                       a = U * D * U**(t)
!
!   WHERE D IS AN n-BY-n DIAGONAL MATRIX AND U IS AN n-BY-n ORTHOGONAL MATRIX.
!   THE DIAGONAL ELEMENTS OF D ARE THE EIGENVALUES OF a; THEY ARE REAL.
!   THE COLUMNS OF U ARE THE EIGENVECTORS OF a.
!
!   FIRST, COMPUTE ALL EIGENVALUES OF THE SELF-ADJOINT MATRIX a IN PACKED FORM
!   AND SAVE THE INTERMEDIATE TRIDIAGONAL MATRIX IN PARAMETER d_e WITH
!   SUBROUTINE eigval_cmp.
!
    call eigval_cmp( a_packed, eigval, failure=failure1, sort=sort, d_e=d_e )
!
!   ON EXIT OF eigval_cmp:
!
!       failure= false :  INDICATES SUCCESSFUL EXIT.
!       failure= true  :  INDICATES THAT THE ALGORITHM DID NOT CONVERGE AND
!                         THAT FULL ACCURACY WAS NOT ATTAINED IN THE EIGENDECOMPOSITION 
!                         OF AN INTERMEDIATE TRIDIAGONAL FORM TRID OF a.
!
!       a_packed IS OVERWRITTEN WITH THE HOUSEHOLDER TRANSFORMATIONS USED TO REDUCE a
!       TO TRIDIAGONAL FORM IF THE OPTIONAL PARAMETER d_e IS PRESENT, OTHERWISE
!       a_packed IS DESTROYED.
!
!       eigval IS OVERWRITTEN WITH THE EIGENVALUES OF a.
!       IF SORT = 'a' OR 'A', THE EIGENVALUES ARE SORTED INTO ASCENDING ORDER.
!       IF SORT = 'd' OR 'D', THE EIGENVALUES ARE SORTED INTO DESCENDING ORDER.
!
!       d_e IS AN OPTIONAL ARGUMENT TO SAVE THE INTERMEDIATE TRIDIAGONAL FORM OF a.
!
!   NEXT COMPUTE THE FIRST nvec EIGENVECTORS OF THE SELF-ADJOINT MATRIX a BY
!   maxiter INVERSE ITERATIONS ON THE INTERMEDIATE TRIDIAGONAL MATRIX
!   d_e AND BACK-TRANSFORMATION.
!
    call trid_inviter( d_e(:n,1), d_e(:n,2), eigval(:nvec), eigvec(:n,:nvec), failure=failure2,   &
                       matp=a_packed, maxiter=maxiter )
!
!   ON EXIT OF trid_inviter :
!
!       failure= false :  INDICATES SUCCESSFUL EXIT
!       failure= true  :  INDICATES THAT MAXIMUM ACCURACY WAS NOT OBTAIN IN THE ALGORITHM

!       eigvec CONTAINS THE nvec EIGENVECTORS OF a ASSOCIATED WITH THE EIGENVALUES eigval(:nvec).
!
!   trid_inviter MAY FAIL IF SOME THE EIGENVALUES SPECIFIED IN PARAMETER eigval ARE NEARLY
!   IDENTICAL.
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*U - U*D
!
        resid(:n,:nvec) = matmul( a, eigvec ) - eigvec*spread( eigval(:nvec), 1, n )
        resid2(:nvec) = norm( resid(:n,:nvec), dim=2_i4b )
!
        err1 =  maxval( resid2(:nvec) )/( norm( a )*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - U**(t)*U.
!
        call unit_matrix( a(:nvec,:nvec) )
!
        resid(:nvec,:nvec) = abs( a(:nvec,:nvec) - matmul( transpose( eigvec ), eigvec ) )
!
        err2 = maxval(resid(:nvec,:nvec))/real(n,stnd)
!
        err = max( err1, err2 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a_packed, eigvec, eigval, d_e, a, resid, resid2 )
!
    else
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a_packed, eigvec, eigval, d_e )
!
    end if
!
!   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
    if ( err<=eps .and. .not.failure1 .and. .not.failure2 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed eigenvalues and eigenvectors = ', err1
        write (prtunit,*) 'Orthogonality of the computed eigenvectors            = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing all eigenvalues and ', nvec, ' eigenvectors of a ', &
       n, ' by ', n,' real symmetric matrix in packed form is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex2_trid_inviter
! ===============================
!
end program ex2_trid_inviter
