program ex3_lapack_stemr
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine (x)STEMR
!   in LAPACK software . 
!                                                                              
! Further Details
! ===============
!
!   The (x)STEMR subroutines will work properly only with IEEE arithmetic.                                                                        
!                                                     
!
! LATEST REVISION : 31/12/2015
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Lapack_interfaces, only : stemr
!
    use Statpack, only : lgl, stnd, true, false, zero, half, one, two, c50,     &
                         unit_matrix, norm, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer, parameter :: prtunit=6, p=5000, n=2*p+1, le=1000
!   
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 3 of stemr'
!
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err, eps, vl, vu, tmp, &
                                               normr, normt, elapsed_time
    real(stnd)                              :: work2(1), z2(1,1)
    real(stnd), allocatable, dimension(:)   :: d, e, w, work, diag, sup, sup2
    real(stnd), allocatable, dimension(:,:) :: id, resid, z
!
    integer                            :: info, lwork, liwork, il, iu, m, nzc,        &
                                          iok, istart, iend, irate, imax, itime, j, l
    integer                            :: iwork2(1)
    integer, dimension(:), allocatable :: isuppz, iwork
!
    logical      :: tryrac
    logical(lgl) :: do_test
!   
    character    :: jobz, range
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 3 : FIRST le EIGENVALUES AND, OPTIONALLY, EIGENVECTORS OF A REAL TRIDIAGONAL
!               MATRIX USING THE MRRR ALGORITHM.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
!
    err = zero
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( d(n), e(n), w(n), isuppz(2*n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE THE WILKINSON TRIDIAGONAL MATRIX OF ORDER n.
!
    j = p + 1
    d(j) = zero
!
    do l = 1, p
        tmp = real( l, stnd)
        d(j+l) = tmp
        d(j-l) = tmp
    end do
!
    e(:n) = one
!
!   GENERATE A (1-2-1) TRIDIAGONAL MATRIX.
!   THE DIAGONAL ELEMENTS ARE STORED IN d .
!   THE OFF-DIAGONAL ELEMENTS ARE STORED IN e .
!
!    d(:n) = two
!    e(:n)  = one
!
!    d(:n) = half
!    e(:n)  = half
!
!    call random_number( d(:n) )
!    call random_number( e(:n) )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( id(le,le), resid(n,le), diag(n), sup(n), sup2(n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SAVE THE TRIDIAGONAL MATRIX.
!
        diag(:n) = d(:n)
        sup(:n)  = e(:n)
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   FIRST DETERMINE THE WORK TO DO.
!
!   IF jobz='N' COMPUTE EIGENVALUES ONLY,
!   IF jobz='V' COMPUTE EIGENVALUES AND EIGENVECTORS.
!
    jobz  = 'V'
!
!   IF range='A' ALL EIGENVALUES WILL BE FOUND,
!   IF range='V' ALL EIGENVALUES IN THE HALF OPEN INTERVAL (vl,vu) WILL BE FOUND,
!   IF range='I' THE il-TH THROUGH iu-TH EIGENVALUES IN ASCENDING ORDER WILL BE FOUND.
!
    range = 'I'
!
    vl = zero
    vu = zero
!
    il = n - le + 1
    iu = n
!
!   tryrac = .true. INDICATES THAT THE CODE WILL TRY TO COMPUTE THE EIGENVALUES OF THE TRIDIAGONAL MATRIX
!                   TO HIGH RELATIVE ACCURACY IF POSSIBLE. IF tryrac IS SET TO .false. THE CODE IS NOT REQUIRED
!                   TO GARANTEE RELATIVELY ACCURATE EIGENVALUES AND CAN USE A FASTEST METHOD.
!
    tryrac = .true.
!
!   COMPUTE FIRST OPTIMAL WORKSPACE FOR stemr SUBROUTINE.
!
    lwork  = -1
    liwork = -1
!
!   nzc IS AN UPPER BOUND FOR THE NUMBER OF EIGENVECTORS TO BE FOUND. nzc DEPENDS ON THE VALUE OF RANGE, IF:
!
!       range='A' THEN nzc>=n;
!       range='V' THEN nzc>=THE NUMBER OF EIGENVALUES IN (VL,VU],
!       range='I' THEN nzc>=iu-il+1 .
!
!   IF nzc=-1, THEN THE SUBROUTINE WILL RETURN THE ESTIMATE OF nzc IN z2(1,1)
!
    nzc = -1
!
    call stemr( jobz, range, n, d, e, vl, vu, il, iu, m, w, z2, n, nzc,   &
                isuppz, tryrac, work2, lwork, iwork2, liwork, info        )
!
    if ( info==0 ) then
!
        lwork  = int(work2(1)) + 10
        liwork = iwork2(1)
        nzc    = z2(1,1)
!
!       ALLOCATE WORK VARIABLES NEEDED BY stemr SUBROUTINE.
!
        allocate( work(lwork), iwork(liwork), z(n,nzc), stat = iok    )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       COMPUTE SELECTED EIGENVALUES AND EIGENVECTORS OF THE SYMMETRIC TRIDIAGONAL MATRIX
!       WITH SUBROUTINE stemr FROM LAPACK.
!       THE EIGENVALUE DECOMPOSITION OF A REAL n-BY-n SYMMETRIC TRIDIAGONAL MATRIX a
!       IS WRITTEN
!
!                       a = U * D * U**(t)
!
!       WHERE D IS AN n-BY-n DIAGONAL MATRIX AND U IS AN n-BY-n ORTHOGONAL MATRIX.
!       THE DIAGONAL ELEMENTS OF D ARE THE EIGENVALUES OF a; THEY ARE REAL.
!       THE COLUMNS OF U ARE THE EIGENVECTORS OF a .
!
        call stemr( jobz, range, n, d, e, vl, vu, il, iu, m, w, z, n, nzc,    &
                    isuppz, tryrac, work, lwork, iwork, liwork, info          )
!
!       ON EXIT OF stemr :
!
!         info= 0     :  INDICATES SUCCESSFUL EXIT.
!         info=-i < 0 :  INDICATES THAT THE iTH ARGUMENT HAS AN ILLEGAL VALUE.
!         info=i  > 0 :  INDICATES THAT THE ALGORITHM FAILED TO CONVERGE.
!
!       ON EXIT OF stemr :
!
!         m GIVES THE TOTAL NUMBER OF EIGENVALUES/EIGENVECTORS FOUND.
!
!         z IS OVERWRITTEN WITH THE FIRST m EIGENVECTORS
!         OF THE TRIDIAGONAL MATRIX (THE EIGENVECTORS ARE STORED COLUMNWISE).
!
!         w IS OVERWRITTEN WITH THE EIGENVALUES OF THE TRIDIAGONAL MATRIX IN ASCENDING ORDER,
!         CORRESPONDING TO THE EIGENVECTORS IN z .
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate ( work, iwork )
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( info/=0 ) then
!
!       ANORMAL EXIT FROM stemr SUBROUTINE, PRINT A WARNING.
!
        write (prtunit,*) 'Error in the call to STEMR subroutine, Info=', info
!
        if ( info<0 ) then
!
            write (prtunit,'(''Argument '',i3,'' has an illegal value'')') - info
!
        end if
!                
    else if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*u - u*s,
!       WHERE s ARE THE EIGENVALUES AND u THE EIGENVECTORS OF THE TRIDIAGONAL MATRIX a .
!
        sup2(:n) = eoshift( sup(:n), -1 )
        resid(:n,:le) = spread( diag(:n),         dim=2, ncopies=le )*z(:n,:le)                              + &
                        spread( sup2(:n),         dim=2, ncopies=le )*eoshift( z(:n,:le), shift=-1, dim=1 )  + &
                        eoshift( spread(sup2(:n), dim=2, ncopies=le )*z(:n,:le), shift=1 )                   - &
                        spread( w(:le),           dim=1, ncopies=n  )*z(:n,:le)
!
        normr = norm( resid(:n,:le) )
        normt = sqrt( sum( diag(:n)**2 ) + sum( sup(:n-1)**2 ) )
!
        err1  = normr/( normt*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!       WHERE u are THE EIGENVECTORS OF THE TRIDIAGONAL MATRIX .
!
        call unit_matrix( id(:le,:le) )
!
        resid(:le,:le) = id(:le,:le) - matmul( transpose( z(:n,:le) ), z(:n,:le) )
        err2 = norm( resid(:le,:le) )/real(n,stnd)
!
        err = max( err1, err2 )
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    if ( do_test ) then
!
        deallocate( d, e, w, isuppz, z, diag, sup, sup2, id, resid )
!
    else
!
        deallocate( d, e, w, isuppz, z )
!
    end if
!
    if ( err<=eps .and. info==0 ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test .and. info==0 ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed eigenvalues and eigenvectors = ', err1
        write (prtunit,*) 'Orthogonality of the computed eigenvectors            = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing ', le,' eigenvalues and eigenvectors of a ', &
       n, ' by ', n,' real symmetric tridiagonal matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex3_lapack_stemr
! ===============================
!
end program ex3_lapack_stemr
