program ex3_trid_deflate
!
!
! Purpose
! =======
!
!   This program is intended to demonstrate the use of subroutine TRID_DEFLATE
!   in module Eig_Procedures .
!                                                                              
!                                                     
! Further Details
! ===============
!
!    The program also shows the use of subroutine SYMTRID_BISECT in module Eig_Procedures.
!
! LATEST REVISION : 23/05/2017
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, true, false, zero, one, two, c50, unit_matrix,          &
                         lamch, trid_deflate, symtrid_bisect, norm, merror, allocate_error
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
#ifdef _TRANSPOSE
    use Statpack, only : transpose=>transpose2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n=3000, neig=3000
!
    real(stnd), parameter  :: fudge=c50
!
    character(len=*), parameter :: name_proc='Example 3 of trid_deflate'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err, eps, abstol, &
                                               normr, normt, elapsed_time
    real(stnd), allocatable, dimension(:)   :: d, e, eigval, temp, temp2
    real(stnd), allocatable, dimension(:,:) :: resid, eigvec
!
    integer      :: iok, istart, iend, irate, imax, itime
    integer(i4b) :: neig2, j, max_qr_steps
!
    logical(lgl) :: failure, do_test, ortho
!   
    character    :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   EXAMPLE 3 : SELECTED EIGENVALUES AND EIGENVECTORS OF
!               A REAL SYMMETRIC TRIDIAGONAL MATRIX USING
!               BISECTION AND DEFLATION METHODS.
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*epsilon( err )
    abstol = sqrt( lamch('s') )
!
    err = zero
    do_test = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( d(n), e(n), eigval(n), eigvec(n,neig), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A (1-2-1) TRIDIAGONAL MATRIX.
!   THE DIAGONAL ELEMENTS ARE STORED IN d .
!   THE OFF-DIAGONAL ELEMENTS ARE STORED IN e .
!
!    d(:n) = two
!    e(:n) = one
!
!    d(:n) = 0.5_stnd
!    e(:n) = 0.5_stnd
!
!   GENERATE A RANDOM TRIDIAGONAL MATRIX.
!
    call random_number( d(:n) )
    call random_number( e(:n) )
!
    if ( do_test ) then
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( resid(n,neig), temp(n), temp2(n), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
    end if
!
!   START TIMING THE COMPUTATIONS.
!
    call system_clock( count_rate=irate, count_max=imax  )
    call system_clock( count=istart )
!
!   COMPUTE THE FIRST neig EIGENVALUES OF THE TRIDIAGONAL MATRIX TO HIGH ACCURACY BY BISECTION
!   WITH SUBROUTINE symtrid_bisect.
!
    call symtrid_bisect( d, e, neig2, eigval, failure, sort=sort, le=neig, abstol=abstol )
!
!   ON EXIT OF symtrid_bisect:
!
!       failure= false :  INDICATES SUCCESSFUL EXIT.
!       failure= true  :  INDICATES THAT THE ALGORITHM DID NOT CONVERGE AND
!                         THAT FULL ACCURACY WAS NOT ATTAINED IN THE EIGENDECOMPOSITION
!                         OF THE TRIDIAGONAL MATRIX.
!
!       eigval IS OVERWRITTEN WITH THE EIGENVALUES OF THE TRIDIAGONAL MATRIX.
!       IF SORT = 'a' OR 'A', THE EIGENVALUES ARE SORTED INTO ASCENDING ORDER.
!       IF SORT = 'd' OR 'D', THE EIGENVALUES ARE SORTED INTO DESCENDING ORDER.
!
    if ( .not. failure ) then
!
!       COMPUTE THE FIRST neig EIGENVECTORS OF THE TRIDIAGONAL MATRIX BY A
!       DEFLATION TECHNIQUE WITH SUBROUTINE trid_deflate.
!
!       ON ENTRY, PARAMETER eigval CONTAINS SELECTED EIGENVALUES OF THE TRIDIAGONAL MATRIX.
!       THE EIGENVALUES VALUES MUST BE GIVEN IN DECREASING ORDER.
!
        ortho = false
        max_qr_steps = 4_i4b
!
        call trid_deflate( d(:n), e(:n), eigval(:neig), eigvec(:n,:neig), failure,   &
                           ortho=ortho, max_qr_steps=max_qr_steps )
!
!       ON EXIT OF trid_deflate :
!
!       failure= false :  INDICATES SUCCESSFUL EXIT
!       failure= true  :  INDICATES THAT MAXIMUM ACCURACY WAS NOT OBTAIN IN THE ALGORITHM

!       eigvec CONTAINS THE neig EIGENVECTORS ASSOCIATED WITH THE EIGENVALUES eigval(:neig).
!
!       trid_deflate MAY FAIL IF SOME THE EIGENVALUES SPECIFIED IN PARAMETER eigval ARE NEARLY
!       IDENTICAL AND IF THESE EIGENVALUES ARE NOT COMPUTED WITH HIGH ACCURACY.
!
    end if
!
!   STOP THE TIMER.
!
    call system_clock( count=iend )
!
    itime = iend - istart
    if ( iend<istart ) then
        itime = itime + imax   
    end if
!
    elapsed_time = real( itime, stnd )/real( irate, stnd )
!
    if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*u - u*s
!       WHERE s ARE THE EIGENVALUES AND u THE EIGENVECTORS OF THE TRIDIAGONAL MATRIX.
!
        do j=1_i4b, neig
!
            temp(1_i4b:n) = eigvec(1_i4b:n,j)
!
            temp2(1_i4b)         = d(1_i4b)*temp(1_i4b) + e(1_i4b)*temp(2_i4b)
            temp2(2_i4b:n-1_i4b) = e(1_i4b:n-2_i4b)*temp(1_i4b:n-2_i4b)           +   &
                                   d(2_i4b:n-1_i4b)*temp(2_i4b:n-1_i4b)           +   &
                                   e(2_i4b:n-1_i4b)*temp(3_i4b:n)
            temp2(n)             = e(n-1_i4b)*temp(n-1_i4b) + d(n)*temp(n)
!
            resid(1_i4b:n,j) = temp2(1_i4b:n) - eigval(j)*temp(1_i4b:n)
!
        end do
!
        temp(:neig) = norm( resid(:n,:neig), dim=2_i4b )
!
        normr = maxval( temp(:neig) )
        normt = sqrt( sum( d(1_i4b:n)**2 ) + sum( e(1_i4b:n-1_i4b)**2 ) )
!
        err1  = normr/( normt*real(n,stnd) )
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!       WHERE u are THE EIGENVECTORS OF THE TRIDIAGONAL MATRIX a.
!
        call unit_matrix( resid(:neig,:neig) )
!
        resid(:neig,:neig) = abs( resid(:neig,:neig) - matmul( transpose( eigvec ), eigvec ) )
!
        normr = maxval( resid(:neig,:neig) )
!
        err2  = normr/real(n,stnd)
!
        err = max( err1, err2 )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( d, e, eigval, eigvec, resid, temp, temp2 )
!
    else
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( d, e, eigval, eigvec )
!
    end if
!
    if ( err<=eps .and. .not.failure ) then
        write (prtunit,*) name_proc//' is correct'
    else
        write (prtunit,*) name_proc//' is incorrect'
    end if
!
    if ( do_test ) then
        write (prtunit,*) 
        write (prtunit,*) 'Accuracy of the computed eigenvalues and eigenvectors = ', err1
        write (prtunit,*) 'Orthogonality of the computed eigenvectors            = ', err2
    end if
!
    write (prtunit,*) 
    write (*,'(a,i5,a,i5,a,i5,a,0pd12.4,a)')    &
      'The elapsed time for computing ', neig,' eigenvalues and eigenvectors of a ', &
       n, ' by ', n,' real symmetric tridiagonal matrix is ', elapsed_time, ' seconds'
!
!
! END OF PROGRAM ex3_trid_deflate
! ===============================
!
end program ex3_trid_deflate
