module BLAS_interfaces
! ####################################################################################
!
! Copyright 2020 IRD
!
! This file is part of statpack.
!
! statpack is free software: you can redistribute it and/or modify
! it under the terms of the GNU Lesser General Public License as
! published by the Free Software Foundation, either version 3 of 
! the License, or (at your option) any later version.
!
! statpack is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU Lesser General Public License for more details.
!
! You can find a copy of the GNU Lesser General Public License
! in the statpack/doc directory.
!
! ####################################################################################
!                                                                                    *
! ************************************************************************************
! MODULE EXPORTING GENERIC INTERFACES FOR SELECTED SUBROUTINES AND FUNCTIONS IN THE  *
! BLAS LIBRARY.                                                                      *
!                                                                                    *
! THIS INTERFACE MODULE ENSURES THAT CALLS TO BLAS ROUTINES ARE CORRECT, WHEN USED   *
! WITH STATPACK.                                                                     *
!                                                                                    *
! GENERIC INTERFACES ARE PRESENTLY PROVIDED FOR THE FOLLOWING BLAS ROUTINES:         *
!                                                                                    *
!    Xaxpy, Xcopy, Xdot, Xdotu, Xrot, Xswap, Xscal, Xnrm2, Xgemv, Xger, Xgeru,       *
!    Xtrsv, Xher2, Xsyr2, Xhpr2, Xspr2, Xgemm, Xsymm, Xhemm, Xtrmm, Xtrsm,           *
!    Xsyrk, Xherk,  Xsyr2k, Xher2k                                                   *
!                                                                                    *
! WHERE X CAN BE s, d, c AND z. THE GENERIC INTERFACES HAVE THE FORM:                *
!                                                                                    *
!    axpy, copy, dot, dotu, rot, swap, scal, nrm2, gemv, ger, geru, trsv, her2,      *
!    syr2, hpr2, spr2, gemm, symm, hemm, trmm, trsm, syrk, herk, syr2k, her2k        * 
!                                                                                    * 
! LATEST REVISION : 30/09/2020                                                       *
!                                                                                    *
! ####################################################################################
!                                                                                    *
! ************************************************************************************
! 
!
! PUBLIC ENTITIES 
! ===============
!    
! ALL SUBROUTINES, FUNCTIONS, VARIABLES AND PARAMETERS ARE PRIVATE BY DEFAULT.
!
    private
    public :: axpy, copy, dot, dotu, rot, swap, scal, nrm2, gemv, ger,  &
              geru, her2, syr2, hpr2, spr2, trsv, gemm, symm, hemm,     &
              trmm, trsm, syrk, herk, syr2k, her2k
!
! GENERIC INTERFACES FOR BLAS ROUTINES WITH OVERLOADED VERSIONS
! =============================================================
!
!   Interface for AXPY subroutines in BLAS
!
    interface axpy
!
        subroutine saxpy( N, SA, SX, INCX, SY, INCY )
!     ..
!     .. Scalar Arguments ..
           REAL             SA
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           REAL             SX( * ), SY( * )
        end subroutine
!
        subroutine daxpy( N, DA, DX, INCX, DY, INCY )
!     ..
!     .. Scalar Arguments ..
           DOUBLE PRECISION DA
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION DX( * ), DY( * )
        end subroutine
!
        subroutine caxpy( N, CA, CX, INCX, CY, INCY )
!     ..
!     .. Scalar Arguments ..
           COMPLEX          CA
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           COMPLEX          CX( * ), CY( * )
        end subroutine
!
        subroutine zaxpy( N, ZA, ZX, INCX, ZY, INCY )
!     ..
!     .. Scalar Arguments ..
           DOUBLE COMPLEX   ZA
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           DOUBLE COMPLEX   ZX( * ), ZY( * )
        end subroutine
!
    end interface
!
!   Interface for COPY subroutines in BLAS
!
    interface copy
!
        subroutine scopy( N, SX, INCX, SY, INCY )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           REAL             SX( * ), SY( * )
        end subroutine
!
        subroutine dcopy( N, DX, INCX, DY, INCY )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION DX( * ), DY( * )
        end subroutine
!
        subroutine ccopy( N, CX, INCX, CY, INCY )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           COMPLEX          CX( * ), CY( * )
        end subroutine
!
        subroutine zcopy( N, ZX, INCX, ZY, INCY )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           DOUBLE COMPLEX   ZX( * ), ZY( * )
        end subroutine
!
    end interface
!
!   Interface for DOT functions in BLAS
!
    interface dot
!
        REAL function sdot( N, SX, INCX, SY, INCY )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           REAL             SX( * ), SY( * )
        end function
!
        DOUBLE PRECISION function ddot( N, DX, INCX, DY, INCY )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION DX( * ), DY( * )
        end function
!
        COMPLEX function cdotc( N, CX, INCX, CY, INCY )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           COMPLEX          CX( * ), CY( * )
        end function
!
        DOUBLE COMPLEX function zdotc( N, ZX, INCX, ZY, INCY )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           DOUBLE COMPLEX   ZX( * ), ZY( * )
        end function
!
    end interface
!
!   Interface for DOTU functions in BLAS
!
    interface dotu
!
        COMPLEX function cdotu( N, CX, INCX, CY, INCY )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           COMPLEX          CX( * ), CY( * )
        end function
!
        DOUBLE COMPLEX function zdotu( N, ZX, INCX, ZY, INCY )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           DOUBLE COMPLEX   ZX( * ), ZY( * )
        end function
!
    end interface
!
!   Interface for ROT subroutines in BLAS
!
    interface rot
!
        subroutine srot( N, SX, INCX, SY, INCY, C, S )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
           REAL             C, S
!     ..
!     .. Array Arguments ..
           REAL             SX( * ), SY( * )
        end subroutine
!
        subroutine drot( N, DX, INCX, DY, INCY, C, S )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
           DOUBLE PRECISION C, S
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION DX( * ), DY( * )
        end subroutine
!
        subroutine csrot( N, CX, INCX, CY, INCY, C, S )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
           REAL             C, S
!     ..
!     .. Array Arguments ..
           COMPLEX          CX( * ), CY( * )
        end subroutine
!
        subroutine zdrot( N, ZX, INCX, ZY, INCY, C, S )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
           DOUBLE PRECISION C, S
!     ..
!     .. Array Arguments ..
           DOUBLE COMPLEX   ZX( * ), ZY( * )
        end subroutine
!
    end interface
!
!   Interface for SWAP subroutines in BLAS
!
    interface swap
!
        subroutine sswap( N, SX, INCX, SY, INCY )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           REAL             SX( * ), SY( * )
        end subroutine
!
        subroutine dswap( N, DX, INCX, DY, INCY )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION DX( * ), DY( * )
        end subroutine
!
        subroutine cswap( N, CX, INCX, CY, INCY )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           COMPLEX          CX( * ), CY( * )
        end subroutine
!
        subroutine zswap( N, ZX, INCX, ZY, INCY )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, INCY, N
!     ..
!     .. Array Arguments ..
           DOUBLE COMPLEX   ZX( * ), ZY( * )
        end subroutine
!
    end interface
!
!   Interface for SCAL subroutines in BLAS
!
    interface scal
!
        subroutine sscal( N, SA, SX, INCX )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, N
           REAL             SA
!     ..
!     .. Array Arguments ..
           REAL             SX( * )
        end subroutine
!
        subroutine dscal( N, DA, DX, INCX )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, N
           DOUBLE PRECISION DA
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION DX( * )
        end subroutine
!
        subroutine cscal( N, CA, CX, INCX )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, N
           COMPLEX          CA
!     ..
!     .. Array Arguments ..
           COMPLEX          CX( * )
        end subroutine
!
        subroutine zscal( N, ZA, ZX, INCX )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, N
           DOUBLE COMPLEX   ZA
!     ..
!     .. Array Arguments ..
           DOUBLE COMPLEX   ZX( * )
        end subroutine
!
        subroutine csscal( N, SA, CX, INCX )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, N
           REAL             SA
!     ..
!     .. Array Arguments ..
           COMPLEX          CX( * )
        end subroutine
!
        subroutine zdscal( N, DA, ZX, INCX )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, N
           DOUBLE PRECISION DA
!     ..
!     .. Array Arguments ..
           DOUBLE COMPLEX   ZX( * )
        end subroutine
!
    end interface
!
!   Interface for NRM2 subroutines in BLAS
!
    interface nrm2
!
        REAL function snrm2( N, X, INCX )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, N
!     ..
!     .. Array Arguments ..
           REAL             X( * )
        end function
!
        DOUBLE PRECISION function dnrm2( N, X, INCX )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, N
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION X( * )
        end function
!
        REAL function scnrm2( N, X, INCX )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, N
!     ..
!     .. Array Arguments ..
           COMPLEX          X( * )
        end function
!
        DOUBLE PRECISION function dznrm2( N, X, INCX )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, N
!     ..
!     .. Array Arguments ..
           DOUBLE COMPLEX   X( * )
        end function
!
    end interface
!
!   Interface for GEMV subroutines in BLAS
!
    interface gemv
!
        subroutine sgemv( TRANS, M, N, ALPHA, A, LDA, X, INCX, BETA, Y, INCY )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        TRANS
           INTEGER          M, N, LDA, INCX, INCY
           REAL             ALPHA, BETA
!     ..
!     .. Array Arguments ..
           REAL             A( LDA, * ), X( * ), Y( * )
        end subroutine
!
        subroutine dgemv( TRANS, M, N, ALPHA, A, LDA, X, INCX, BETA, Y, INCY )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        TRANS
           INTEGER          M, N, LDA, INCX, INCY
           DOUBLE PRECISION ALPHA, BETA
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION A( LDA, * ), X( * ), Y( * )
        end subroutine
!
        subroutine cgemv( TRANS, M, N, ALPHA, A, LDA, X, INCX, BETA, Y, INCY )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        TRANS
           INTEGER          M, N, LDA, INCX, INCY
           COMPLEX          ALPHA, BETA
!     ..
!     .. Array Arguments ..
           COMPLEX          A( LDA, * ), X( * ), Y( * )
        end subroutine
!
        subroutine zgemv( TRANS, M, N, ALPHA, A, LDA, X, INCX, BETA, Y, INCY )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        TRANS
           INTEGER          M, N, LDA, INCX, INCY
           DOUBLE COMPLEX   ALPHA, BETA
!     ..
!     .. Array Arguments ..
           DOUBLE COMPLEX   A( LDA, * ), X( * ), Y( * )
        end subroutine
!
    end interface
!
!   Interface for GER subroutines in BLAS
!
    interface ger
!
        subroutine sger( M, N, ALPHA, X, INCX, Y, INCY, A, LDA )
!     ..
!     .. Scalar Arguments ..
           INTEGER          M, N, LDA, INCX, INCY
           REAL             ALPHA
!     ..
!     .. Array Arguments ..
           REAL             A( LDA, * ), X( * ), Y( * )
        end subroutine
!
        subroutine dger( M, N, ALPHA, X, INCX, Y, INCY, A, LDA )
!     ..
!     .. Scalar Arguments ..
           INTEGER          M, N, LDA, INCX, INCY
           DOUBLE PRECISION ALPHA
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION A( LDA, * ), X( * ), Y( * )
        end subroutine
!
        subroutine cgerc( M, N, ALPHA, X, INCX, Y, INCY, A, LDA )
!     ..
!     .. Scalar Arguments ..
           INTEGER          M, N, LDA, INCX, INCY
           COMPLEX          ALPHA
!     ..
!     .. Array Arguments ..
           COMPLEX          A( LDA, * ), X( * ), Y( * )
        end subroutine
!
        subroutine zgerc( M, N, ALPHA, X, INCX, Y, INCY, A, LDA )
!     ..
!     .. Scalar Arguments ..
           INTEGER          M, N, LDA, INCX, INCY
           DOUBLE COMPLEX   ALPHA
!     ..
!     .. Array Arguments ..
           DOUBLE COMPLEX   A( LDA, * ), X( * ), Y( * )
        end subroutine
!
    end interface
!
!   Interface for GERU subroutines in BLAS
!
    interface geru
!
        subroutine cgeru( M, N, ALPHA, X, INCX, Y, INCY, A, LDA )
!     ..
!     .. Scalar Arguments ..
           INTEGER          M, N, LDA, INCX, INCY
           COMPLEX          ALPHA
!     ..
!     .. Array Arguments ..
           COMPLEX          A( LDA, * ), X( * ), Y( * )
        end subroutine
!
        subroutine zgeru( M, N, ALPHA, X, INCX, Y, INCY, A, LDA )
!     ..
!     .. Scalar Arguments ..
           INTEGER          M, N, LDA, INCX, INCY
           DOUBLE COMPLEX   ALPHA
!     ..
!     .. Array Arguments ..
           DOUBLE COMPLEX   A( LDA, * ), X( * ), Y( * )
        end subroutine
!
    end interface
!
!   Interface for TRSV subroutines in BLAS
!
    interface trsv
!
        subroutine strsv( UPLO, TRANS, DIAG, N, A, LDA, X, INCX )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, LDA, N
           CHARACTER        DIAG, TRANS, UPLO
!     ..
!     .. Array Arguments ..
           REAL             A( LDA, * ), X( * )
        end subroutine
!
        subroutine dtrsv( UPLO, TRANS, DIAG, N, A, LDA, X, INCX )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, LDA, N
           CHARACTER        DIAG, TRANS, UPLO
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION A( LDA, * ), X( * )
        end subroutine
!
        subroutine ctrsv( UPLO, TRANS, DIAG, N, A, LDA, X, INCX )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, LDA, N
           CHARACTER        DIAG, TRANS, UPLO
!     ..
!     .. Array Arguments ..
           COMPLEX          A( LDA, * ), X( * )
        end subroutine
!
        subroutine ztrsv( UPLO, TRANS, DIAG, N, A, LDA, X, INCX )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INCX, LDA, N
           CHARACTER        DIAG, TRANS, UPLO
!     ..
!     .. Array Arguments ..
           DOUBLE COMPLEX   A( LDA, * ), X( * )
        end subroutine
!
    end interface
!
!   Interface for SYR2 subroutines in BLAS
!
    interface syr2
!
        subroutine ssyr2( UPLO, N, ALPHA, X, INCX, Y, INCY, A, LDA )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        UPLO
           INTEGER          N, LDA, INCX, INCY
           REAL             ALPHA
!     ..
!     .. Array Arguments ..
           REAL             A( LDA, * ), X( * ), Y( * )
        end subroutine
!
        subroutine dsyr2( UPLO, N, ALPHA, X, INCX, Y, INCY, A, LDA )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        UPLO
           INTEGER          N, LDA, INCX, INCY
           DOUBLE PRECISION ALPHA
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION A( LDA, * ), X( * ), Y( * )
        end subroutine
!
    end interface
!
!   Interface for SPR2 subroutines in BLAS
!
    interface spr2
!
        subroutine sspr2( UPLO, N, ALPHA, X, INCX, Y, INCY, AP )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        UPLO
           INTEGER          N, INCX, INCY
           REAL             ALPHA
!     ..
!     .. Array Arguments ..
           REAL             AP( * ), X( * ), Y( * )
        end subroutine
!
        subroutine dspr2( UPLO, N, ALPHA, X, INCX, Y, INCY, AP )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        UPLO
           INTEGER          N, INCX, INCY
           DOUBLE PRECISION ALPHA
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION AP( * ), X( * ), Y( * )
        end subroutine
!
    end interface
!
!   Interface for HER2 subroutines in BLAS
!
    interface her2
!
        subroutine cher2( UPLO, N, ALPHA, X, INCX, Y, INCY, A, LDA )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        UPLO
           INTEGER          N, LDA, INCX, INCY
           COMPLEX          ALPHA
!     ..
!     .. Array Arguments ..
           COMPLEX          A( LDA, * ), X( * ), Y( * )
        end subroutine
!
        subroutine zher2( UPLO, N, ALPHA, X, INCX, Y, INCY, A, LDA )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        UPLO
           INTEGER          N, LDA, INCX, INCY
           DOUBLE COMPLEX   ALPHA
!     ..
!     .. Array Arguments ..
           DOUBLE COMPLEX   A( LDA, * ), X( * ), Y( * )
        end subroutine
!
    end interface
!
!   Interface for HPR2 subroutines in BLAS
!
    interface hpr2
!
        subroutine chpr2( UPLO, N, ALPHA, X, INCX, Y, INCY, AP )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        UPLO
           INTEGER          N, LDA, INCX, INCY
           COMPLEX          ALPHA
!     ..
!     .. Array Arguments ..
           COMPLEX          AP( * ), X( * ), Y( * )
        end subroutine
!
        subroutine zhpr2( UPLO, N, ALPHA, X, INCX, Y, INCY, AP )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        UPLO
           INTEGER          N, LDA, INCX, INCY
           DOUBLE COMPLEX   ALPHA
!     ..
!     .. Array Arguments ..
           DOUBLE COMPLEX   AP( * ), X( * ), Y( * )
        end subroutine
!
    end interface
!
!   Interface for GEMM subroutines in BLAS
!
    interface gemm
!
        subroutine sgemm( TRANSA, TRANSB, M, N, K, ALPHA, A, LDA, B, LDB, BETA, C, LDC )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        TRANSA, TRANSB
           INTEGER          M, N, K, LDA, LDB, LDC
           REAL             ALPHA, BETA
!     ..
!     .. Array Arguments ..
           REAL             A( LDA, * ), B( LDB, * ), C( LDC, * )
        end subroutine
!
        subroutine dgemm( TRANSA, TRANSB, M, N, K, ALPHA, A, LDA, B, LDB, BETA, C, LDC )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        TRANSA, TRANSB
           INTEGER          M, N, K, LDA, LDB, LDC
           DOUBLE PRECISION ALPHA, BETA
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION A( LDA, * ), B( LDB, * ), C( LDC, * )
        end subroutine
!
        subroutine cgemm( TRANSA, TRANSB, M, N, K, ALPHA, A, LDA, B, LDB, BETA, C, LDC )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        TRANSA, TRANSB
           INTEGER          M, N, K, LDA, LDB, LDC
           COMPLEX          ALPHA, BETA
!     ..
!     .. Array Arguments ..
           COMPLEX          A( LDA, * ), B( LDB, * ), C( LDC, * )
        end subroutine
!
        subroutine zgemm( TRANSA, TRANSB, M, N, K, ALPHA, A, LDA, B, LDB, BETA, C, LDC )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        TRANSA, TRANSB
           INTEGER          M, N, K, LDA, LDB, LDC
           DOUBLE COMPLEX   ALPHA, BETA
!     ..
!     .. Array Arguments ..
           DOUBLE COMPLEX   A( LDA, * ), B( LDB, * ), C( LDC, * )
        end subroutine
!
    end interface
!
!   Interface for SYMM subroutines in BLAS
!
    interface symm
!
        subroutine ssymm( SIDE, UPLO, M, N, ALPHA, A, LDA, B, LDB, BETA, C, LDC )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        SIDE, UPLO
           INTEGER          M, N, LDA, LDB, LDC
           REAL             ALPHA, BETA
!     ..
!     .. Array Arguments ..
           REAL             A( LDA, * ), B( LDB, * ), C( LDC, * )
        end subroutine
!
        subroutine dsymm( SIDE, UPLO, M, N, ALPHA, A, LDA, B, LDB, BETA, C, LDC )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        SIDE, UPLO
           INTEGER          M, N, LDA, LDB, LDC
           DOUBLE PRECISION ALPHA, BETA
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION A( LDA, * ), B( LDB, * ), C( LDC, * )
        end subroutine
!
        subroutine csymm( SIDE, UPLO, M, N, ALPHA, A, LDA, B, LDB, BETA, C, LDC )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        SIDE, UPLO
           INTEGER          M, N, LDA, LDB, LDC
           COMPLEX          ALPHA, BETA
!     ..
!     .. Array Arguments ..
           COMPLEX          A( LDA, * ), B( LDB, * ), C( LDC, * )
        end subroutine
!
        subroutine zsymm( SIDE, UPLO, M, N, ALPHA, A, LDA, B, LDB, BETA, C, LDC )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        SIDE, UPLO
           INTEGER          M, N, LDA, LDB, LDC
           DOUBLE COMPLEX   ALPHA, BETA
!     ..
!     .. Array Arguments ..
           DOUBLE COMPLEX   A( LDA, * ), B( LDB, * ), C( LDC, * )
        end subroutine
!
    end interface
!
!   Interface for HEMM subroutines in BLAS
!
    interface hemm
!
        subroutine chemm( SIDE, UPLO, M, N, ALPHA, A, LDA, B, LDB, BETA, C, LDC )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        SIDE, UPLO
           INTEGER          M, N, LDA, LDB, LDC
           COMPLEX          ALPHA, BETA
!     ..
!     .. Array Arguments ..
           COMPLEX          A( LDA, * ), B( LDB, * ), C( LDC, * )
        end subroutine
!
        subroutine zhemm( SIDE, UPLO, M, N, ALPHA, A, LDA, B, LDB, BETA, C, LDC )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        SIDE, UPLO
           INTEGER          M, N, LDA, LDB, LDC
           DOUBLE COMPLEX   ALPHA, BETA
!     ..
!     .. Array Arguments ..
           DOUBLE COMPLEX   A( LDA, * ), B( LDB, * ), C( LDC, * )
        end subroutine
!
    end interface
!
!   Interface for TRMM subroutines in BLAS
!
    interface trmm
!
        subroutine strmm( SIDE, UPLO, TRANSA, DIAG, M, N, ALPHA, A, LDA, B, LDB )
!     ..
!     .. Scalar Arguments ..
           INTEGER          LDA, LDB, M, N
           REAL             ALPHA
           CHARACTER        DIAG, SIDE, TRANSA, UPLO
!     ..
!     .. Array Arguments ..
           REAL             A( LDA, * ), B( LDB, * )
        end subroutine
!
        subroutine dtrmm( SIDE, UPLO, TRANSA, DIAG, M, N, ALPHA, A, LDA, B, LDB )
!     ..
!     .. Scalar Arguments ..
           INTEGER          LDA, LDB, M, N
           DOUBLE PRECISION ALPHA
           CHARACTER        DIAG, SIDE, TRANSA, UPLO
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION A( LDA, * ), B( LDB, * )
        end subroutine
!
        subroutine ctrmm( SIDE, UPLO, TRANSA, DIAG, M, N, ALPHA, A, LDA, B, LDB )
!     ..
!     .. Scalar Arguments ..
           INTEGER          LDA, LDB, M, N
           COMPLEX          ALPHA
           CHARACTER        DIAG, SIDE, TRANSA, UPLO
!     ..
!     .. Array Arguments ..
           COMPLEX          A( LDA, * ), B( LDB, * )
        end subroutine
!
        subroutine ztrmm( SIDE, UPLO, TRANSA, DIAG, M, N, ALPHA, A, LDA, B, LDB )
!     ..
!     .. Scalar Arguments ..
           INTEGER          LDA, LDB, M, N
           DOUBLE COMPLEX   ALPHA
           CHARACTER        DIAG, SIDE, TRANSA, UPLO
!     ..
!     .. Array Arguments ..
           DOUBLE COMPLEX   A( LDA, * ), B( LDB, * )
        end subroutine
!
    end interface
!
!   Interface for TRSM subroutines in BLAS
!
    interface trsm
!
        subroutine strsm( SIDE, UPLO, TRANSA, DIAG, M, N, ALPHA, A, LDA, B, LDB )
!     ..
!     .. Scalar Arguments ..
           INTEGER          LDA, LDB, M, N
           REAL             ALPHA
           CHARACTER        DIAG, SIDE, TRANSA, UPLO
!     ..
!     .. Array Arguments ..
           REAL             A( LDA, * ), B( LDB, * )
        end subroutine
!
        subroutine dtrsm( SIDE, UPLO, TRANSA, DIAG, M, N, ALPHA, A, LDA, B, LDB )
!     ..
!     .. Scalar Arguments ..
           INTEGER          LDA, LDB, M, N
           DOUBLE PRECISION ALPHA
           CHARACTER        DIAG, SIDE, TRANSA, UPLO
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION A( LDA, * ), B( LDB, * )
        end subroutine
!
        subroutine ctrsm( SIDE, UPLO, TRANSA, DIAG, M, N, ALPHA, A, LDA, B, LDB )
!     ..
!     .. Scalar Arguments ..
           INTEGER          LDA, LDB, M, N
           COMPLEX          ALPHA
           CHARACTER        DIAG, SIDE, TRANSA, UPLO
!     ..
!     .. Array Arguments ..
           COMPLEX          A( LDA, * ), B( LDB, * )
        end subroutine
!
        subroutine ztrsm( SIDE, UPLO, TRANSA, DIAG, M, N, ALPHA, A, LDA, B, LDB )
!     ..
!     .. Scalar Arguments ..
           INTEGER          LDA, LDB, M, N
           DOUBLE COMPLEX   ALPHA
           CHARACTER        DIAG, SIDE, TRANSA, UPLO
!     ..
!     .. Array Arguments ..
           DOUBLE COMPLEX   A( LDA, * ), B( LDB, * )
        end subroutine
!
    end interface
!
!   Interface for SYR2K subroutines in BLAS
!
    interface syr2k
!
        subroutine ssyr2k( UPLO, TRANS, N, K, ALPHA, A, LDA, B, LDB, BETA, C, LDC )
!     ..
!     .. Scalar Arguments ..
           INTEGER          LDA, LDB, LDC, N, K
           REAL             ALPHA, BETA
           CHARACTER        TRANS, UPLO
!     ..
!     .. Array Arguments ..
           REAL             A( LDA, * ), B( LDB, * ), C( LDC, * )
        end subroutine
!
        subroutine dsyr2k( UPLO, TRANS, N, K, ALPHA, A, LDA, B, LDB, BETA, C, LDC )
!     ..
!     .. Scalar Arguments ..
           INTEGER          LDA, LDB, LDC, N, K
           DOUBLE PRECISION ALPHA, BETA
           CHARACTER        TRANS, UPLO
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION A( LDA, * ), B( LDB, * ), C( LDC, * )
        end subroutine
!
        subroutine csyr2k( UPLO, TRANS, N, K, ALPHA, A, LDA, B, LDB, BETA, C, LDC )
!     ..
!     .. Scalar Arguments ..
           INTEGER          LDA, LDB, LDC, N, K
           COMPLEX          ALPHA, BETA
           CHARACTER        TRANS, UPLO
!     ..
!     .. Array Arguments ..
           COMPLEX          A( LDA, * ), B( LDB, * ), C( LDC, * )
        end subroutine
!
        subroutine zsyr2k( UPLO, TRANS, N, K, ALPHA, A, LDA, B, LDB, BETA, C, LDC )
!     ..
!     .. Scalar Arguments ..
           INTEGER          LDA, LDB, LDC, N, K
           DOUBLE COMPLEX   ALPHA, BETA
           CHARACTER        TRANS, UPLO
!     ..
!     .. Array Arguments ..
           DOUBLE COMPLEX   A( LDA, * ), B( LDB, * ), C( LDC, * )
        end subroutine
!
    end interface
!
!   Interface for HER2K subroutines in BLAS
!
    interface her2k
!
        subroutine cher2k( UPLO, TRANS, N, K, ALPHA, A, LDA, B, LDB, BETA, C, LDC )
!     ..
!     .. Scalar Arguments ..
           INTEGER          LDA, LDB, LDC, N, K
           COMPLEX          ALPHA
           REAL             BETA
           CHARACTER        TRANS, UPLO
!     ..
!     .. Array Arguments ..
           COMPLEX          A( LDA, * ), B( LDB, * ), C( LDC, * )
        end subroutine
!
        subroutine zher2k( UPLO, TRANS, N, K, ALPHA, A, LDA, B, LDB, BETA, C, LDC )
!     ..
!     .. Scalar Arguments ..
           INTEGER          LDA, LDB, LDC, N, K
           DOUBLE COMPLEX   ALPHA
           DOUBLE PRECISION BETA
           CHARACTER        TRANS, UPLO
!     ..
!     .. Array Arguments ..
           DOUBLE COMPLEX   A( LDA, * ), B( LDB, * ), C( LDC, * )
        end subroutine
!
    end interface
!
!   Interface for SYRK subroutines in BLAS
!
    interface syrk
!
        subroutine ssyrk( UPLO, TRANS, N, K, ALPHA, A, LDA, BETA, C, LDC )
!     ..
!     .. Scalar Arguments ..
           INTEGER          LDA, LDC, N, K
           REAL             ALPHA, BETA
           CHARACTER        TRANS, UPLO
!     ..
!     .. Array Arguments ..
           REAL             A( LDA, * ), C( LDC, * )
        end subroutine
!
        subroutine dsyrk( UPLO, TRANS, N, K, ALPHA, A, LDA, BETA, C, LDC )
!     ..
!     .. Scalar Arguments ..
           INTEGER          LDA, LDC, N, K
           DOUBLE PRECISION ALPHA, BETA
           CHARACTER        TRANS, UPLO
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION A( LDA, * ), C( LDC, * )
        end subroutine
!
        subroutine csyrk( UPLO, TRANS, N, K, ALPHA, A, LDA, BETA, C, LDC )
!     ..
!     .. Scalar Arguments ..
           INTEGER          LDA, LDC, N, K
           COMPLEX          ALPHA, BETA
           CHARACTER        TRANS, UPLO
!     ..
!     .. Array Arguments ..
           COMPLEX          A( LDA, * ), C( LDC, * )
        end subroutine
!
        subroutine zsyrk( UPLO, TRANS, N, K, ALPHA, A, LDA, BETA, C, LDC )
!     ..
!     .. Scalar Arguments ..
           INTEGER          LDA, LDC, N, K
           DOUBLE COMPLEX   ALPHA, BETA
           CHARACTER        TRANS, UPLO
!     ..
!     .. Array Arguments ..
           DOUBLE COMPLEX   A( LDA, * ), C( LDC, * )
        end subroutine
!
    end interface
!
!   Interface for HERK subroutines in BLAS
!
    interface herk
!
        subroutine cherk( UPLO, TRANS, N, K, ALPHA, A, LDA, BETA, C, LDC )
!     ..
!     .. Scalar Arguments ..
           INTEGER          LDA, LDC, N, K
           REAL             ALPHA, BETA
           CHARACTER        TRANS, UPLO
!     ..
!     .. Array Arguments ..
           COMPLEX          A( LDA, * ), C( LDC, * )
        end subroutine
!
        subroutine zherk( UPLO, TRANS, N, K, ALPHA, A, LDA, BETA, C, LDC )
!     ..
!     .. Scalar Arguments ..
           INTEGER          LDA, LDC, N, K
           DOUBLE PRECISION ALPHA, BETA
           CHARACTER        TRANS, UPLO
!     ..
!     .. Array Arguments ..
           DOUBLE COMPLEX   A( LDA, * ), C( LDC, * )
        end subroutine
!
    end interface
!
!
! *******************************
! END OF MODULE BLAS_interfaces *
! *******************************
!
end module BLAS_interfaces
