module Giv_Procedures
! ##############################################################################
!
! Copyright 2021 IRD
!
! This file is part of statpack.
!
! statpack is free software: you can redistribute it and/or modify
! it under the terms of the GNU Lesser General Public License as
! published by the Free Software Foundation, either version 3 of 
! the License, or (at your option) any later version.
!
! statpack is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU Lesser General Public License for more details.
!
! You can find a copy of the GNU Lesser General Public License
! in the statpack/doc directory.
!
! ##############################################################################
!                                                                              *
! ******************************************************************************
! MODULE EXPORTING GIVENS TOOLS (REFLECTIONS AND ROTATIONS).                   *
!                                                                              *
! LATEST REVISION : 12/02/2021                                                 *
!                                                                              *
! ##############################################################################
!                                                                              *
! ******************************************************************************
!                                                                              *
! THE C PROCESSOR MACROS USED IN THIS MODULE ARE:                              *
!                                                                              *
!  _BLAS        FOR USING BLAS WHEN POSSIBLE                                   *
!  _TRANSPOSE   FOR REPLACING THE transpose INTRINSIC FUNCTION WITH STATPACK   *
!               transpose2 FUNCTION                                            *
!                                                                              *
! ******************************************************************************
!
!
! USED MODULES 
! ============
!
    use Select_Parameters,  only : i2b, i4b, stnd
#ifdef _TRANSPOSE
    use Utilities,          only : transpose=>transpose2
#endif
!
! USED MODULES IN SPECIFIC PROCEDURES
! ===================================
!
!   use Logical_Constants
!   use Reals_Constants
!   use Char_Constants
!   use Num_Constants
!   use Utilities
!   use BLAS_interfaces
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!
! PUBLIC ENTITIES 
! ===============
!    
! ALL SUBROUTINES, FUNCTIONS, VARIABLES AND PARAMETERS ARE PRIVATE BY DEFAULT.
!
    private
    public ::                                                                 &
              define_rot_givens, apply_rot_givens, rot_givens,                &
              givens_vec, givens_mat_left, givens_mat_right,                  &
              givens_vec_mat_left, givens_vec_mat_right,                      &
              define_rot_fastgivens, apply_rot_fastgivens,                    &
              fastgivens_vec, fastgivens_mat_left, fastgivens_mat_right,      &
              fastgivens_vec_mat_left, fastgivens_vec_mat_right,              &
              define_rot_fastgivens2, fastgivens2_vec
!
! GENERIC INTERFACES FOR ROUTINES WITH OVERLOADED VERSIONS
! ========================================================
!
    interface rot_givens
        module procedure    rot_givens_r, rot_givens_r_cs,                &
                            rot_givens_rv, rot_givens_rv_cs
    end interface
!
    interface apply_rot_givens
        module procedure    apply_rot_givens_r, apply_rot_givens_r_cs,    &
                            apply_rot_givens_rv, apply_rot_givens_rv_cs
    end interface
!
    interface givens_vec
        module procedure    givens_rv, givens_rv_cs
    end interface
!
    interface givens_mat_left
        module procedure    givens_rm_left
    end interface
!
    interface givens_mat_right
        module procedure    givens_rm_right
    end interface
!
    interface givens_vec_mat_left
        module procedure    givens_rv_rm_left
    end interface
!
    interface givens_vec_mat_right
        module procedure    givens_rv_rm_right
    end interface
!
    interface apply_rot_fastgivens
        module procedure    apply_rot_fastgivens_r, apply_rot_fastgivens_rv
    end interface
!
    interface fastgivens_vec
        module procedure    fastgivens_rv, fastgivens_rv_t
    end interface
!
    interface fastgivens2_vec
        module procedure    fastgivens2_rv, fastgivens2_rv_t
    end interface
!
    interface fastgivens_mat_left
        module procedure    fastgivens_rm_left
    end interface
!
    interface fastgivens_mat_right
        module procedure    fastgivens_rm_right
    end interface
!
    interface fastgivens_vec_mat_left
        module procedure    fastgivens_rv_rm_left
    end interface
!
    interface fastgivens_vec_mat_right
        module procedure    fastgivens_rv_rm_right
    end interface
!
!
! =========================================================================================
!
                                  contains
!                                 ========
!
! =========================================================================================
!                                GIVENS ROTATIONS
! =========================================================================================
!
!
    subroutine define_rot_givens( a, b, cs, sn )
!
! Purpose
! _______
!
!   DEFINE_ROT_GIVENS generates the cosine and sine of a Givens plane rotation, ROT, so that
!
!       ( A  B ) ROT  = ( R  0 )
!
!   where R >= 0 and ROT is 2-by-2 matrix defined by
!
!       ( +CS  -SN )
!
!       ( +SN  +CS ) 
!
!   with CS**(2) + SN**(2) = 1.
!
!
! Arguments
! _________
!
!   A        (INPUT) real(stnd)
!            The first component of vector to be rotated.
!
!   B        (INPUT) real(stnd)
!            The second component of vector to be rotated.
!
!   CS       (OUTPUT) real(stnd)
!            The cosine of the rotation.
!
!   SN       (OUTPUT) real(stnd)
!            The sine of the rotation.
!
!
! Further Details
! _______________
!
!   A and B are unchanged on return.
!
!   Normally, the subprogram APPLY_ROT_GIVENS( VECA, VECB, CS, SN ) will next
!   be called to apply the rotation to a n-by-2 matrix [ VECA  VECB ].
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)  :: a, b
    real(stnd), intent(out) :: cs, sn
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: tn
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    if ( abs( a )>abs( b ) ) then
!
        tn = b/a
        sn = sqrt( one + tn*tn )
        cs = sign( one/sn, a )
        sn = cs*tn
!
    else if ( b/=zero ) then
!
        tn = a/b
        cs = sqrt( one + tn*tn )
        sn = sign( one/cs, b )
        cs = sn*tn
!
    else
!
        cs = one
        sn = zero
!
    end if
!
!
! END OF SUBROUTINE define_rot_givens
! ___________________________________
!
    end subroutine define_rot_givens
!
! =========================================================================================
!
    subroutine rot_givens_r( a, b )
!
! Purpose
! _______
!
!   ROT_GIVENS applies a Givens plane rotation, ROT, so that
!
!       ( A  B ) ROT  = ( R  0 )
!
!   where ROT is 2-by-2 matrix defined by
!
!       ( +CS  -SN )
!
!       ( +SN  +CS ) 
!
!   with CS**(2) + SN**(2) = 1.
!
!   On output, the rotation is also stored in compact form in B.
!
!
! Arguments
! _________
!
!   A        (INPUT/OUTPUT) real(stnd)
!            The first component of vector to be rotated.
!
!            On output, R = (+/-)sqrt( A**(2) + B**(2) ) overwrites A.
!
!   B        (INPUT/OUTPUT) real(stnd)
!            The second component of vector to be rotated.
!
!            On output, Z overwrites B. Z allows CS and SN
!            to be recovered by the following algorithm:
!
!            - If Z = 1  set CS = 0 and SN = 1
!            - If abs( Z ) < 1  set SN = Z   and CS = sqrt( 1 - SN**(2) )
!            - If abs( Z ) > 1  set CS = 1/Z and SN = sqrt( 1 - CS**(2) )
!
!
! Further Details
! _______________
!
!   Normally, the subprogram APPLY_ROT_GIVENS( VECA, VECB, B ) will
!   next be called to apply the rotation to a n-by-2 matrix [ VECA  VECB ].
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
    use Num_Constants,   only : lamch
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout)  :: a, b
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: x, tn, safmin
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( b==zero ) return
!
!   LET  o  BE THE ANGLE OF THE GIVENS ROTATION.
!
    if ( abs( a )>=abs( b ) ) then
!
!       HERE :  0         <  abs(o)      <= PI/4, 
!               0         <  abs(tan(o)) <= 1,
!               1/sqrt(2) <= cos(o)      <  1,
!               0         <  abs(sin(o)) <= 1/sqrt(2) .
!
        tn = b/a
        x = sqrt( one + tn*tn )
!
!       OVERWRITE  a  AND  b  WITH  r  AND  z , RESPECTIVELY.
!
        a = x*a
        b = tn/x
!
    else
!
!       HERE :  PI/4      <  abs(o)        <= PI/2,
!               0         <= abs(cotan(o)) <  1,
!               1/sqrt(2) <  sin(o)        <= 1,
!               0         <= abs(cos(o))   <  1/sqrt(2) .
!
        tn = a/b
        x = sqrt( one + tn*tn )
!
!       OVERWRITE  a  WITH  r .
!
        a = x*b
!
!       OVERWRITE  b  WITH  z .
!
        b = tn/x
!
!       FIND safmin, THE SAFE MINIMUM, SUCH 
!       THAT  1/safmin  DOES NOT OVERFLOW.
!
        safmin = lamch( 's' )
!
        if ( abs(b)<safmin ) then
!
!           HERE IF  b = zero  THEN  o = PI/2.
!
            b = merge( one, one/safmin, b==zero )
!
        else
!
!           HERE  PI/4 <= abs(o) < PI/2.
!
            b = one/b
!
        end if
!
    end if
!
!
! END OF SUBROUTINE rot_givens_r
! ______________________________
!
    end subroutine rot_givens_r
!
! =========================================================================================
!
    subroutine rot_givens_r_cs( a, b, cs, sn )
!
! Purpose
! _______
!
!   ROT_GIVENS generates and applies a Givens plane rotation, ROT, so that
!
!       ( A  B ) ROT  = ( R  0 )
!
!   where ROT is 2-by-2 matrix defined by
!
!       ( +CS  -SN )
!
!       ( +SN  +CS ) 
!
!   with CS**(2) + SN**(2) = 1.
!
!   On output, the rotation is also stored in compact form in B.
!
!
! Arguments
! _________
!
!   A        (INPUT/OUTPUT) real(stnd)
!            The first component of vector to be rotated.
!
!            On output, R = (+/-)sqrt( A**(2) + B**(2) ) overwrites A.
!
!   B        (INPUT/OUTPUT) real(stnd)
!            The second component of vector to be rotated.
!
!            On output, Z overwrites B. Z allows CS and SN
!            to be recovered by the following algorithm:
!
!            - If Z = 1  set CS = 0 and SN = 1
!            - If abs( Z ) < 1  set SN = Z   and CS = sqrt( 1 - SN**(2) )
!            - If abs( Z ) > 1  set CS = 1/Z and SN = sqrt( 1 - CS**(2) )
!
!   CS       (OUTPUT) real(stnd)
!            The cosine of the rotation.
!
!   SN       (OUTPUT) real(stnd)
!            The sine of the rotation.
!
!
! Further Details
! _______________
!
!   Normally, the subprograms APPLY_ROT_GIVENS( VECA, VECB, CS, SN ) or
!   APPLY_ROT_GIVENS( VECA, VECB, B ) will next be called to apply the rotation 
!   to a n-by-2 matrix [ VECA  VECB ].
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
    use Num_Constants,   only : lamch
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout)  :: a, b
    real(stnd), intent(out)    :: cs, sn
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: r, tn, safmin
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!    LET  o  BE THE ANGLE OF THE GIVENS ROTATION.
!
    if ( b==zero ) then
!
!       HERE  b = 0  AND  o = zero .
!
        cs = one
        sn = zero
!
    else if ( abs( a )>=abs( b ) ) then
!
!       HERE :  0         <  abs(o)      <= PI/4, 
!               0         <  abs(tan(o)) <= 1,
!               1/sqrt(2) <= cos(o)      <  1,
!               0         <  abs(sin(o)) <= 1/sqrt(2) .
!
        tn = b/a
        r = sqrt( one + tn*tn )*a
!
!       COMPUTE COSINE AND SINE OF THE ROTATION.
!
        cs = a/r
        sn = cs*tn
!
!       OVERWRITE  a  AND  b  WITH  r  AND  z , RESPECTIVELY.
!
        a = r
        b = sn
!
    else
!
!       HERE :  PI/4      <  abs(o)        <= PI/2,
!               0         <= abs(cotan(o)) <  1,
!               1/sqrt(2) <  sin(o)        <= 1,
!               0         <= abs(cos(o))   <  1/sqrt(2) .
!
        tn = a/b
!
!       OVERWRITE  a  WITH  r .
!
        a = sqrt( one + tn*tn )*b
!
!       COMPUTE COSINE AND SINE OF THE ROTATION.
!
        sn = b/a
        cs = sn*tn
!
!       OVERWRITE  b  WITH  z .
!
!
!       FIND safmin, THE SAFE MINIMUM, SUCH 
!       THAT  1/safmin  DOES NOT OVERFLOW.
!
        safmin = lamch( 's' )
!
        if ( abs(cs)<safmin ) then
!
!           HERE IF  cs = zero  THEN  o = PI/2.
!
            b = merge( one, one/safmin, cs==zero )
!
        else
!
!           HERE  PI/4 <= abs(o) < PI/2.
!
            b = one/cs
!
        end if
!
    end if
!
!
! END OF SUBROUTINE rot_givens_r_cs
! _________________________________
!
    end subroutine rot_givens_r_cs
!
! =========================================================================================
!
    subroutine rot_givens_rv( veca, vecb )
!
! Purpose
! _______
!
!   ROT_GIVENS applies a Givens plane rotation, ROT, to the n-by-2 matrix [ VECA VECB ].
!   The rotation is designed to annilhate the first element of VECB ( e.g. VECB(1)).
!   That is,
!
!       [ VECA  VECB ] ROT = [ (CS*VECA + SN*VECB)  (-SN*VECA + CS*VECB) ]
!
!   where
!
!   - CS**(2) + SN**(2) = 1,
!   - -SN*VECA(1) + CS*VECB(1) = 0
!   - and ROT is a 2-by-2 matrix defined by
!
!       ( +CS  -SN )
!
!       ( +SN  +CS ) 
!
!   On output, the rotation is also stored in compact form in VECB(1).
!
!
! Arguments
! _________
!
!   VECA    (INPUT/OUTPUT) real(stnd), dimension(:)
!           The first vector to be rotated.
!
!           On output, CS*VECA + SN*VECB overwrites VECA.
!
!   VECB    (INPUT/OUTPUT) real(stnd), dimension(:)
!           The second vector to be rotated.
!
!           On output, -SN*VECA(2:) + CS*VECB(2:) overwrites VECB(2:) and
!           Z overwrites VECB(1). Z allows CS and SN to be recovered by
!           the following algorithm:
!
!           - If Z = 1  set CS = 0 and SN = 1
!           - If abs( Z ) < 1  set SN = Z   and CS = sqrt( 1 - SN**(2) )
!           - If abs( Z ) > 1  set CS = 1/Z and SN = sqrt( 1 - CS**(2) )
!
!
! Further Details
! _______________
!
!   It is assumed that VECA and VECB have the same size.
!
!   The subprograms APPLY_ROT_GIVENS( VECC, VECD, VECB(1) ) may next be called 
!   to apply the rotation to another n-by-2 matrix [ VECC VECD ].
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
    use Num_Constants,   only : lamch
#ifdef _BLAS
    use BLAS_interfaces, only : rot
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:)  :: veca, vecb
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
#ifdef _BLAS
    real(stnd)   :: cs, sn, tn, safmin, tmp
#else
    real(stnd)   :: cs, sn, tn, safmin, tmp(size( vecb ))
#endif
    integer(i4b) :: n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   QUICK RETURN IF POSSIBLE.
!
    n = size( vecb )
!
    if ( n<=0_i4b ) return
!
    if ( vecb(1)==zero ) return
!
#ifdef _BLAS
    tmp = vecb(1_i4b)
#else
    tmp(:n) = vecb(:n)
#endif
!
!   LET  o  BE THE ANGLE OF THE GIVENS ROTATION.
!
    if ( abs( veca(1_i4b) )>=abs( vecb(1_i4b) ) ) then
!
!       HERE :  0         <  abs(o)      <= PI/4, 
!               0         <  abs(tan(o)) <= 1,
!               1/sqrt(2) <= cos(o)      <  1,
!               0         <  abs(sin(o)) <= 1/sqrt(2) .
!
        tn = vecb(1_i4b)/veca(1_i4b)
!
!       COMPUTE COSINE AND SINE OF THE ROTATION.
!
        cs = one/sqrt( one + tn*tn )
        sn = cs*tn
!
!       OVERWRITE  vecb(1)  WITH  z.
!
        vecb(1_i4b) = sn
!
    else
!
!       HERE :  PI/4      <  abs(o)        <= PI/2,
!               0         <= abs(cotan(o)) <  1,
!               1/sqrt(2) <  sin(o)        <= 1,
!               0         <= abs(cos(o))   <  1/sqrt(2) .
!
        tn = veca(1_i4b)/vecb(1_i4b)
!
!       COMPUTE COSINE AND SINE OF THE ROTATION.
!
        sn = one/sqrt( one + tn*tn )
        cs = sn*tn
!
!       OVERWRITE  vecb(1)  WITH z .
!
!
!       FIND safmin, THE SAFE MINIMUM, SUCH 
!       THAT  1/safmin  DOES NOT OVERFLOW.
!
        safmin = lamch( 's' )
!
        if ( abs(cs)<safmin ) then
!
!           HERE IF  cs = zero  THEN  o = PI/2.
!
!            vecb(1) = merge( one, one/safmin, cs==zero )
!
!BUG: For compiling STATPACK with NAG compiler
!     replace the preceding line with
!
            if ( cs==zero ) then
                vecb(1_i4b) = one
            else
                vecb(1_i4b) = one/safmin
            end if
!
        else
!
!           HERE  PI/4 <= abs(o) < PI/2.
!
            vecb(1) = one/cs
!
        end if
!
    end if
!
!   APPLY ROTATION.
!
#ifdef _BLAS
    veca(1_i4b) = cs*veca(1_i4b) + sn*tmp
    call rot( n-1_i4b, veca(2_i4b:n), 1_i4b, vecb(2_i4b:n), 1_i4b, cs, sn )
#else
    vecb(2_i4b:n) = cs*tmp(2_i4b:n) - sn*veca(2_i4b:n)
    veca(:n)      = cs*veca(:n)     + sn*tmp(:n)
#endif
!
!
! END OF SUBROUTINE rot_givens_rv
! _______________________________
!
    end subroutine rot_givens_rv
!
! =========================================================================================
!
    subroutine rot_givens_rv_cs( veca, vecb, cs, sn )
!
! Purpose
! _______
!
!   ROT_GIVENS defines and applies a Givens plane rotation, ROT, to the n-by-2 matrix
!   [ VECA  VECB ]. The rotation is designed to annilhate the first element of VECB (e.g. VECB(1)).
!   That is,
!
!       [ VECA  VECB ] ROT = [ (CS*VECA + SN*VECB)  (-SN*VECA + CS*VECB) ]
!
!   where
!
!   - CS**(2) + SN**(2) = 1,
!   - -SN*VECA(1) + CS*VECB(1) = 0
!   - and ROT is a 2-by-2 matrix defined by
!
!       ( +CS  -SN )
!
!       ( +SN  +CS ) 
!
!   On output, the rotation is also stored in compact form in VECB(1).
!
!
! Arguments
! _________
!
!   VECA    (INPUT/OUTPUT) real(stnd), dimension(:)
!           The first vector to be rotated.
!
!           On output, CS*VECA + SN*VECB overwrites VECA.
!
!   VECB    (INPUT/OUTPUT) real(stnd), dimension(:)
!           The second vector to be rotated.
!
!           On output, -SN*VECA(2:) + CS*VECB(2:) overwrites VECB(2:) and
!           Z overwrites VECB(1). Z allows CS and SN to be recovered by
!           the following algorithm:
!
!           - If Z = 1  set CS = 0 and SN = 1
!           - If abs( Z ) < 1  set SN = Z   and CS = sqrt( 1 - SN**(2) )
!           - If abs( Z ) > 1  set CS = 1/Z and SN = sqrt( 1 - CS**(2) )
!
!   CS      (OUTPUT) real(stnd)
!           The cosine of the rotation.
!
!   SN      (OUTPUT) real(stnd)
!           The sine of the rotation.
!
!
! Further Details
! _______________
!
!   It is assumed that VECA and VECB have the same size.
!
!   Normally, the subprograms APPLY_ROT_GIVENS( VECC, VECD, CS, SN ) or
!   APPLY_ROT_GIVENS( VECC, VECD, VECB(1) ) will next be called to apply 
!   the rotation to a n-by-2 matrix [ VECC  VECD ].
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
    use Num_Constants,   only : lamch
#ifdef _BLAS
    use BLAS_interfaces, only : rot
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:)  :: veca, vecb
    real(stnd), intent(out)                  :: cs, sn
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
#ifdef _BLAS
    real(stnd) :: tn, safmin, tmp
#else
    real(stnd) :: tn, safmin, tmp(size( vecb ))
#endif
!
    integer(i4b) :: n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   QUICK RETURN IF POSSIBLE.
!
    n = size( vecb )
!
    if ( n<=0_i4b ) then
!
        cs = one
        sn = zero
!
        return
!
    end if
!
!   LET  o  BE THE ANGLE OF THE GIVENS ROTATION.
!
    if ( vecb(1_i4b)==zero ) then
!
!       HERE  vecb(1) = zero  AND  o = 0.
!
        cs = one
        sn = zero
!
    else
!
#ifdef _BLAS
        tmp = vecb(1_i4b)
#else
        tmp(:n) = vecb(:n)
#endif
!
        if ( abs( veca(1_i4b) )>=abs( vecb(1_i4b) ) ) then
!
!           HERE :  0         <  abs(o)      <= PI/4, 
!                   0         <  abs(tan(o)) <= 1,
!                   1/sqrt(2) <= cos(o)      <  1,
!                   0         <  abs(sin(o)) <= 1/sqrt(2) .
!
            tn = vecb(1_i4b)/veca(1_i4b)
!
!           COMPUTE COSINE AND SINE OF THE ROTATION.
!
            cs = one/sqrt( one + tn*tn )
            sn = cs*tn
!
!           OVERWRITE  vecb(1)  WITH  z.
!
            vecb(1_i4b) = sn
!
        else
!
!           HERE :  PI/4      <  abs(o)        <= PI/2,
!                   0         <= abs(cotan(o)) <  1,
!                   1/sqrt(2) <  sin(o)        <= 1,
!                   0         <= abs(cos(o))   <  1/sqrt(2) .
!
            tn = veca(1_i4b)/vecb(1_i4b)
!
!           COMPUTE COSINE AND SINE OF THE ROTATION.
!
            sn = one/sqrt( one + tn*tn )
            cs = sn*tn
!
!           OVERWRITE  vecb(1)  WITH z .
!
!
!           FIND safmin, THE SAFE MINIMUM, SUCH 
!           THAT  1/safmin  DOES NOT OVERFLOW.
!
            safmin = lamch( 's' )
!
            if ( abs(cs)<safmin ) then
!
!               HERE IF  cs = zero  THEN  o = PI/2.
!
!                vecb(1) = merge( one, one/safmin, cs==zero )
!
!BUG: For compiling STATPACK with NAG compiler
!     replace the preceding line with
!
                if ( cs==zero ) then
                    vecb(1_i4b) = one
                else
                    vecb(1_i4b) = one/safmin
                end if
!
            else
!
!               HERE  PI/4 <= abs(o) < PI/2.
!
                vecb(1_i4b) = one/cs
!
            end if
!
        end if
!
!       APPLY ROTATION.
!
#ifdef _BLAS
        veca(1_i4b) = cs*veca(1_i4b) + sn*tmp
        call rot( n-1_i4b, veca(2_i4b:n), 1_i4b, vecb(2_i4b:n), 1_i4b, cs, sn )
#else
        vecb(2_i4b:n) = cs*tmp(2_i4b:n) - sn*veca(2_i4b:n)
        veca(:n) = cs*veca(:n) + sn*tmp(:n)
#endif
!
   end if
!
!
! END OF SUBROUTINE rot_givens_rv_cs
! __________________________________
!
    end subroutine rot_givens_rv_cs
!
! =========================================================================================
!
    subroutine apply_rot_givens_r( c, d, b )
!
! Purpose
! _______
!
!   APPLY_ROT_GIVENS reconstructs and applies a Givens plane rotation, ROT, stored
!   in compact form in B, to the vector ( C D ).
!
!   That is, the value B allows the cosine and sine of the Givens plane rotation
!   to be recovered by the following algorithm:
!
!   - If B = 1  set CS = 0 and SN = 1
!   - If abs( B ) < 1  set SN = B   and CS = sqrt( 1 - SN**(2) )
!   - If abs( B ) > 1  set CS = 1/B and SN = sqrt( 1 - CS**(2) )
!
!   Next, the Givens plane rotation, ROT, is applied to the vector ( C D ) :
!
!       ( C  D ) ROT  = ( (CS*C + SN*D)  (-SN*C + CS*D) )
!
!   where ROT is a 2-by-2 matrix defined by
!
!       ( +CS  -SN )
!
!       ( +SN  +CS ) 
!
!
! Arguments
! _________
!
!   C        (INPUT/OUTPUT) real(stnd)
!            The first element of vector to be rotated.
!
!            On output, CS*C + SN*D overwrites C.
!
!   D        (INPUT/OUTPUT) real(stnd)
!            The second element of vector to be rotated.
!
!            On output, -SN*C + CS*D overwrites D.
!
!   B        (INPUT) real(stnd)
!            The real number, which allows the cosine and sine
!            of the Givens plane rotation to be recovered.
!
!
! Further Details
! _______________
!
!   Normally:
!
!   - the subprogram APPLY_ROT_GIVENS( C, D, B ) is called to apply
!     a Givens rotation to the vector ( C D ) after a call to ROT_GIVENS( A, B, CS, SN )
!     or ROT_GIVENS( A, B ).
!   - the subprogram APPLY_ROT_GIVENS( C, D, VECB(1) ) is called to apply
!     a Givens rotation to the vector ( C D ) after a call to ROT_GIVENS( VECA, VECB, CS, SN )
!     or ROT_GIVENS( VECA, VECB ) where VECA and VECB are two vectors of the same length.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout) :: c, d
    real(stnd), intent(in)    :: b
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: cs, sn, tmp
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( b==zero ) return
!
    tmp = c
!
    if ( b==one ) then
!
!       APPLY PERMUTATION.
!
        c = d
        d = tmp
!
    else
!
!       RECONSTRUCT ROTATION.
!
        if ( abs( b )<one ) then
            sn = b
            cs = sqrt( one - b*b )
        else
            cs = one/b
            sn = sqrt( one - cs*cs )
        end if
!
!       APPLY ROTATION.
!
        c = cs*c + sn*d
        d = cs*d - sn*tmp
!
    end if
!
!
! END OF SUBROUTINE apply_rot_givens_r
! ____________________________________
!
    end subroutine apply_rot_givens_r
!
! =========================================================================================
!
    subroutine apply_rot_givens_r_cs( c, d, cs, sn )
!
! Purpose
! _______
!
!   APPLY_ROT_GIVENS applies a Givens plane rotation, ROT, to to the vector ( C D ).
!   That is,
!
!          ( C  D ) ROT  = ( (CS*C + SN*D)  (-SN*C + CS*D) )
!
!   where ROT is a 2-by-2 matrix defined by
!
!       ( +CS  -SN )
!
!       ( +SN  +CS ) 
!
!
! Arguments
! _________
!
!   C        (INPUT/OUTPUT) real(stnd)
!            The first element of vector to be rotated.
!
!            On output, CS*C + SN*D overwrites C.
!
!   D        (INPUT/OUTPUT) real(stnd)
!            The second element of vector to be rotated.
!
!            On output, -SN*C + CS*D overwrites D.
!
!   CS       (INPUT) real(stnd)
!            The cosine of the rotation.
!
!   SN       (INPUT) real(stnd)
!            The sine of the rotation.
!
!
! Further Details
! _______________
!
!   Normally, the subprogram APPLY_ROT_GIVENS( C, D, CS, SN ) is called to apply 
!   a Givens rotation to the vector ( C D) after a call to DEFINE_ROT_GIVENS( A, B,
!   CS, SN ), ROT_GIVENS( A, B, CS, SN ) or ROT_GIVENS( VECA, VECB, CS, SN ).
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout)  :: c, d
    real(stnd), intent(in)     :: cs, sn
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: tmp
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( sn==zero ) return
!
!   APPLY ROTATION.
!
    tmp = c
    c = cs*c + sn*d
    d = cs*d - sn*tmp
!
!
! END OF SUBROUTINE apply_rot_givens_r_cs
! _______________________________________
!
    end subroutine apply_rot_givens_r_cs
!
! =========================================================================================
!
    subroutine apply_rot_givens_rv( vecc, vecd, b )
!
! Purpose
! _______
!
!   APPLY_ROT_GIVENS reconstructs and applies a Givens plane rotation, ROT, stored in 
!   compact form in B, to the n-by-2 matrix [ VECC VECD ].
!
!   That is, the value B allows the cosine and sine of the Givens plane rotation to be 
!   recovered by the following algorithm:
!
!   - If B = 1  set CS = 0 and SN = 1
!   - If abs( B ) < 1  set SN = B   and CS = sqrt( 1 - SN**(2) )
!   - If abs( B ) > 1  set CS = 1/B and SN = sqrt( 1 - CS**(2) )
!
!   Next, the Givens plane rotation, ROT, is applied to the n-by-2 matrix [ VECC  VECD ]:
!
!       [ VECC  VECD ] ROT  = [ (CS*VECC + SN*VECD)  (-SN*VECC + CS*VECD) ]
!
!   where ROT is a 2-by-2 matrix defined by
!
!       ( +CS  -SN )
!
!       ( +SN  +CS ) 
!
!
! Arguments
! _________
!
!   VECC    (INPUT/OUTPUT) real(stnd), dimension(:)
!           The first vector to be rotated.
!
!           On output, CS*VECC + SN*VECD overwrites VECC.
!
!   VECD    (INPUT/OUTPUT) real(stnd), dimension(:)
!           The second vector to be rotated.
!
!           On output, -SN*VECC + CS*VECD overwrites VECD.
!
!   B       (INPUT) real(stnd)
!           The real number which allows the cosine and sine
!           of the Givens plane rotation to be recovered.
!
!
! Further Details
! _______________
!
!   Normally, the subprogram APPLY_ROT_GIVENS( VECC, VECD, B ) is called to apply
!   a Givens rotation to the n-by-2 matrix [ VECC  VECD ] after a call to 
!   ROT_GIVENS( A, B, CS, SN ) or ROT_GIVENS( A, B ).
!
!   Normally, the subprogram APPLY_ROT_GIVENS( VECC, VECD, VECB(1) ) is called
!   to apply a Givens rotation to the n-by-2 matrix [ VECC  VECD ] after a call to 
!   ROT_GIVENS( VECA, VECB, CS,SN ) or ROT_GIVENS( VECA, VECB ) where VECA and VECB
!   are two vectors of the same length.
!
!   It is assumed that VECC and VECD have the same size.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
#ifdef _BLAS
    use BLAS_interfaces, only : rot, swap
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:)  :: vecc, vecd
    real(stnd), intent(in)                   :: b
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
#ifdef _BLAS
    real(stnd) :: cs, sn
#else
    real(stnd) :: cs, sn, tmp(size( vecc ))
#endif
!
    integer(i4b) :: n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   QUICK RETURN IF POSSIBLE.
!
    n = size( vecc )
!
    if ( b==zero .or. n<=0_i4b ) return
!
    if ( b==one ) then
!
!       APPLY PERMUTATION.
!
#ifdef _BLAS
        call swap( n, vecc(:n), 1_i4b, vecd(:n), 1_i4b )
#else
        tmp(:n)  = vecc(:n)
        vecc(:n) = vecd(:n)
        vecd(:n) = tmp(:n)
#endif
!
    else
!
!       RECONSTRUCT THE ROTATION.
!
        if ( abs( b ) < one ) then
            sn = b
            cs = sqrt( one - b*b )
        else
            cs = one/b
            sn = sqrt( one - cs*cs )
        end if
!
!       APPLY ROTATION.
!
#ifdef _BLAS
        call rot( n, vecc(1_i4b:n), 1_i4b, vecd(1_i4b:n), 1_i4b, cs, sn )
#else
        tmp(:n)  = vecc(:n)
        vecc(:n) = cs*tmp(:n)  + sn*vecd(:n)
        vecd(:n) = cs*vecd(:n) - sn*tmp(:n)
#endif
!
    end if
!
!
! END OF SUBROUTINE apply_rot_givens_rv
! _____________________________________
!
    end subroutine apply_rot_givens_rv
!
! =========================================================================================
!
    subroutine apply_rot_givens_rv_cs( vecc, vecd, cs, sn )
!
! Purpose
! _______
!
!   APPLY_ROT_GIVENS applies a Givens plane rotation, ROT, to the n-by-2 matrix [ VECC VECD ].
!   That is,
!
!       [ VECC VECD ]  ROT  = [ (CS*VECC + SN*VECD)  (-SN*VECC + CS*VECD) ]
!
!   where ROT is a 2-by-2 matrix defined by
!
!       ( +CS  -SN )
!
!       ( +SN  +CS ) 
!
!
! Arguments
! _________
!
!   VECC    (INPUT/OUTPUT) real(stnd), dimension(:)
!           The first vector to be rotated.
!
!           On output, CS*VECC + SN*VECD overwrites VECC.
!
!   VECD    (INPUT/OUTPUT) real(stnd), dimension(:)
!           The second vector to be rotated.
!
!           On output, -SN*VECC + CS*VECD overwrites VECD.
!
!   CS      (INPUT) real(stnd)
!           The cosine of the rotation.
!
!   SN      (INPUT) real(stnd)
!           The sine of the rotation.
!
!
! Further Details
! _______________
!
!   Normally, the subprogram APPLY_ROT_GIVENS( VECC, VECD, CS, SN ) is called
!   to apply a Givens rotation to the n-by-2 matrix [ VECC VECD ] after a call
!   to DEFINE_ROT_GIVENS( A, B, CS, SN ), ROT_GIVENS( A, B, CS, SN )  or 
!   ROT_GIVENS( VECA, VECB, CS, SN ).
!
!   It is assumed that VECC and VECD have the same size.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero
#ifdef _BLAS
    use BLAS_interfaces, only : rot
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:)  :: vecc, vecd
    real(stnd), intent(in)                   :: cs, sn
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
#ifndef _BLAS
    real(stnd)   :: tmp(size( vecc ))
#endif
!
    integer(i4b) :: n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   QUICK RETURN IF POSSIBLE.
!
    n = size( vecc )
!
    if ( sn==zero .or. n<=0_i4b ) return
!
!   APPLY ROTATION.
!
#ifdef _BLAS
    call rot( n, vecc(1_i4b:n), 1_i4b, vecd(1_i4b:n), 1_i4b, cs, sn )
#else
    tmp(:n) = vecc(:n)
    vecc(:n) = cs*tmp(:n) + sn*vecd(:n)
    vecd(:n) = cs*vecd(:n) - sn*tmp(:n)
#endif
!
!
! END OF SUBROUTINE apply_rot_givens_rv_cs
! ________________________________________
!
    end subroutine apply_rot_givens_rv_cs
!
! =========================================================================================
!
    subroutine givens_rv( veca, vecb )
!
! Purpose
! _______
!
!   GIVENS defines and applies a Givens plane rotation, ROT, to the n-by-2 matrix [ VECA VECB ].
!   The rotation is designed to annilhate the first element of VECB (e.g. VECB(1)). That is,
!
!      [ VECA  VECB ] ROT  = [ (CS*VECA + SN*VECB)  (-SN*VECA + CS*VECB) ]
!
!   where:
!
!   - CS**(2) + SN**(2) = 1,
!   - -SN*VECA(1) + CS*VECB(1) =  0,
!   - CS*VECA(1) + SN*VECB(1)  >= 0.
!
!   and ROT is a 2-by-2 matrix defined by
!
!       ( +CS  -SN )
!
!       ( +SN  +CS ) 
!
!
! Arguments
! _________
!
!   VECA    (INPUT/OUTPUT) real(stnd), dimension(:)
!           The first vector to be rotated.
!
!           On output, CS*VECA + SN*VECB overwrites VECA.
!
!   VECB    (INPUT/OUTPUT) real(stnd), dimension(:)
!           The second vector to be rotated.
!
!           On output, -SN*VECA + CS*VECB overwrites VECB.
!
!
! Further Details
! _______________
!
!   It is assumed that VECA and VECB have the same size.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
#ifdef _BLAS
    use BLAS_interfaces, only : rot
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:)  :: veca, vecb
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
#ifdef _BLAS
    real(stnd) :: tn, cs, sn
#else
    real(stnd) :: tn, cs, sn, tmp(size(vecb))
#endif
!
    integer(i4b) :: n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   QUICK RETURN IF POSSIBLE.
!
    n = size( vecb )
!
    if ( n<=0_i4b ) return
!
!   DEFINE THE GIVENS ROTATION.
!
    if ( abs( veca(1_i4b) )>abs( vecb(1_i4b) ) ) then
!
        tn = vecb(1_i4b)/veca(1_i4b)
        sn = sqrt( one + tn*tn )
        cs = sign( one/sn, veca(1_i4b) )
        sn = cs*tn
!
    else if ( vecb(1_i4b)/=zero ) then
!
        tn = veca(1_i4b)/vecb(1_i4b)
        cs = sqrt( one + tn*tn )
        sn = sign( one/cs, vecb(1_i4b) )
        cs = sn*tn
!
    else
!
        return
!
    end if
!
!   APPLY ROTATION.
!
#ifdef _BLAS
    call rot( n, veca(1_i4b:n), 1_i4b, vecb(1_i4b:n), 1_i4b, cs, sn )
#else
    tmp(:n)  = veca(:n)
    veca(:n) = cs*tmp(:n)  + sn*vecb(:n)
    vecb(:n) = cs*vecb(:n) - sn*tmp(:n)
#endif
!
!
! END OF SUBROUTINE givens_rv
! ___________________________
!
    end subroutine givens_rv
!
! =========================================================================================
!
    subroutine givens_rv_cs( veca, vecb, cs, sn )
!
! Purpose
! _______
!
!   GIVENS defines and applies a Givens plane rotation, ROT, to the n-by-2 matrix 
!   [ VECA VECB ]. The rotation is designed to annilhate the first element of VECB (e.g. VECB(1)).
!   That is,
!
!      [ VECA VECB ] ROT  = [ (CS*VECA + SN*VECB)  (-SN*VECA + CS*VECB) ]
!
!   where:
!
!   - CS**(2) + SN**(2) = 1,
!   - -SN*VECA(1) + CS*VECB(1) =  0,
!   - CS*VECA(1) + SN*VECB(1)  >= 0.
!
!   and ROT is a 2-by-2 matrix defined by
!
!       ( +CS  -SN )
!
!       ( +SN  +CS ) 
!
!
! Arguments
! _________
!
!   VECA    (INPUT/OUTPUT) real(stnd), dimension(:)
!           The first vector to be rotated.
!
!           On output, CS*VECA + SN*VECB overwrites VECA.
!
!   VECB    (INPUT/OUTPUT) real(stnd), dimension(:)
!           The second vector to be rotated.
!
!           On output, -SN*VECA + CS*VECB overwrites VECB.
!
!   CS      (OUTPUT) real(stnd)
!           The cosine of the rotation.
!
!   SN      (OUTPUT) real(stnd)
!           The sine of the rotation.
!
!
! Further Details
! _______________
!
!   It is assumed that VECA and VECB have the same size.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
#ifdef _BLAS
    use BLAS_interfaces, only : rot
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:)  :: veca, vecb
    real(stnd), intent(out)                  :: cs, sn
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
#ifdef _BLAS
    real(stnd) :: tn
#else
    real(stnd) :: tn, tmp(size(vecb))
#endif
!
    integer(i4b) :: n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   QUICK RETURN IF POSSIBLE.
!
    n = size( vecb )
!
    if ( n<=0_i4b ) return
!
!   DEFINE THE GIVENS ROTATION.
!
    if ( abs( veca(1_i4b) )>abs( vecb(1_i4b) ) ) then
!
        tn = vecb(1_i4b)/veca(1_i4b)
        sn = sqrt( one + tn*tn )
        cs = sign( one/sn, veca(1_i4b) )
        sn = cs*tn
!
    else if ( vecb(1_i4b)/=zero ) then
!
        tn = veca(1_i4b)/vecb(1_i4b)
        cs = sqrt( one + tn*tn )
        sn = sign( one/cs, vecb(1_i4b) )
        cs = sn*tn
!
    else
!
        cs = one
        sn = zero
!
        return
!
    end if
!
!   APPLY ROTATION.
!
#ifdef _BLAS
    call rot( n, veca(1_i4b:n), 1_i4b, vecb(1_i4b:n), 1_i4b, cs, sn )
#else
    tmp(:n)  = veca(:n)
    veca(:n) = cs*tmp(:n)  + sn*vecb(:n)
    vecb(:n) = cs*vecb(:n) - sn*tmp(:n)
#endif
!
!
! END OF SUBROUTINE givens_rv_cs
! ______________________________
!
    end subroutine givens_rv_cs
!
! =========================================================================================
!
    subroutine givens_rm_left( mat )
!
! Purpose
! _______
!
!   GIVENS_MAT_LEFT transforms the matrix MAT to upper trapezoidal form by applying 
!   a serie of Givens plane rotations on the rows of MAT.
!
!
! Arguments
! _________
!
!   MAT    (INPUT/OUTPUT) real(stnd), dimension(:,:)
!          The matrix to be transformed.
!
!          On output, the transformed matrix overwrites MAT.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
    use Char_Constants,  only : allocate_error
    use Utilities,       only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout) :: mat(:,:)
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                              :: tn, cs, sn
    real(stnd), dimension(:),   allocatable :: tmp 
    real(stnd), dimension(:,:), allocatable :: matt
!
    integer(i4b) :: i, j, m, n
    integer      :: iok
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='givens_mat_left'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( mat, 1 )
    n = size( mat, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( m<=0_i4b .or. n<=0_i4b ) return
!
!   ALLOCATE WORK VARIABLE FOR A FAST ALGORITHM IF POSSIBLE.
!
    allocate( matt(n,m), stat = iok )
!
    if ( iok==0 ) then
!
!       TRANSPOSE THE MATRIX AND USE A FAST ALGORITHM.
!
        matt(:n,:m) = transpose( mat(:m,:n) )
!
        call givens_rm_right( matt(:n,:m) )
!
        mat(:m,:n) = transpose( matt(:n,:m) )
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( matt )
!
    else
!
!       ALLOCATE WORK ARRAY.
!
        allocate( tmp(n), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        do j = 1_i4b, min( n, m )
!
            tmp(j:n) = mat(m,j:n)
!
!           ZERO COMPONENTS OF  mat(j+1:,j) .
!
            do i = m-1_i4b, j, -1_i4b
!
!               ZERO  mat(i+1,j) .
!
!               DEFINE THE GIVENS ROTATION.
!
                if ( abs( mat(i,j) )>abs( tmp(j) ) ) then
!
                    tn = tmp(j)/mat(i,j)
                    sn = sqrt( one + tn*tn )
                    cs = sign( one/sn , mat(i,j) )
                    sn = cs*tn
!
                else if ( tmp(j)/=zero ) then
!
                    tn = mat(i,j)/tmp(j)
                    cs = sqrt( one + tn*tn )
                    sn = sign(one/cs , tmp(j) )
                    cs = sn*tn
!
                else
!
                    mat(i+1_i4b,j:n) = tmp(j:n)
                    tmp(j:n) = mat(i,j:n)
!
                    cycle
!
                end if
!
!               APPLY ROTATION.
!
                mat(i+1_i4b,j:n) = cs*tmp(j:n) - sn*mat(i,j:n)
                tmp(j:n) = cs*mat(i,j:n) + sn*tmp(j:n)
!
            end do
!
            mat(j,j:n) = tmp(j:n)
!
        end do
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( tmp )
!
    end if
!
!
! END OF SUBROUTINE givens_rm_left
! ________________________________
!
    end subroutine givens_rm_left
!
! =========================================================================================
!
    subroutine givens_rm_right( mat )
!
! Purpose
! _______
!
!   GIVENS_MAT_RIGHT transforms the matrix MAT to lower trapezoidal form by applying 
!   a serie of Givens plane rotations on the columns of MAT.
!
!
! Arguments
! _________
!
!   MAT    (INPUT/OUTPUT) real(stnd), dimension(:,:)
!          The matrix to be transformed.
!
!          On output, the transformed matrix overwrites MAT.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout) :: mat(:,:)
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: tn, cs, sn, tmp(size(mat,1))
!
    integer(i4b) :: i, j, m, n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( mat, 1 )
    n = size( mat, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( m<=0_i4b .or. n<=0_i4b ) return
!
    do j = 1_i4b, min( n, m )
!
        tmp(j:m) = mat(j:m,n)
!
!       ZERO COMPONENTS OF  mat(j,j+1:) .
!
        do i = n-1_i4b, j, -1_i4b
!
!           ZERO  mat(j,i+1) .
!
!           DEFINE THE GIVENS ROTATION.
!
            if ( abs( mat(j,i) )>abs( tmp(j) ) ) then
!
                tn = tmp(j)/mat(j,i)
                sn = sqrt( one + tn*tn )
                cs = sign( one/sn , mat(j,i) )
                sn = cs*tn
!
            else if ( tmp(j)/=zero ) then
!
                tn = mat(j,i)/tmp(j)
                cs = sqrt( one + tn*tn )
                sn = sign(one/cs , tmp(j) )
                cs = sn*tn
!
            else
!
                mat(j:m,i+1_i4b) = tmp(j:m)
                tmp(j:m) = mat(j:m,i)
!
                cycle
!
            end if
!
!           APPLY ROTATION.
!
            mat(j:m,i+1_i4b) = cs*tmp(j:m) - sn*mat(j:m,i)
            tmp(j:m) = cs*mat(j:m,i) + sn*tmp(j:m)
!
        end do
!
        mat(j:m,j) = tmp(j:m)
!
    end do
!
!
! END OF SUBROUTINE givens_rm_right
! _________________________________
!
    end subroutine givens_rm_right
!
! =========================================================================================
!
    subroutine givens_rv_rm_left( vec, mat )
!
! Purpose
! _______
!
!   GIVENS_VEC_MAT_LEFT defines and applies a serie of Givens rotations on a n-vector 
!   VEC and on the rows of a p-by-n matrix MAT. The rotations are designed to annilhate
!   all the elements of the first column of MAT.
!                     
!
! Arguments
! _________
!
!   VEC     (INPUT/OUTPUT) real(stnd), dimension(:)
!           On input, the n-vector to rotate. VEC(1) is used to define the rotations.
!
!           On output, the transformed vector overwrites VEC.
!
!   MAT     (INPUT/OUTPUT) real(stnd), dimension(:,:)
!           On input, the matrix to be transformed. MAT(:,1) is used to define
!           the rotations.
!
!           On output, the transformed matrix overwrites MAT and MAT(:,1) is
!           equal to zero.
!
!
! Further Details
! _______________
!
!   It is assumed that size( VEC ) = size( MAT, 2 ) .
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
    use Char_Constants,  only : allocate_error
    use Utilities,       only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout) :: vec(:), mat(:,:)
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                              :: tn, cs, sn
    real(stnd), dimension(:),   allocatable :: tmp 
    real(stnd), dimension(:,:), allocatable :: matt
!
    integer(i4b) :: i, m, n
    integer      :: iok
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='givens_vec_mat_left'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( mat, 1 )
    n = size( vec )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( m<=0_i4b .or. n<=0_i4b ) return
!
!   ALLOCATE WORK VARIABLE FOR A FAST ALGORITHM IF POSSIBLE.
!
    allocate( matt(n,m), stat = iok )
!
    if ( iok==0 ) then
!
!       TRANSPOSE THE MATRIX AND USE A FAST ALGORITHM.
!
        matt(:n,:m) = transpose( mat(:m,:n) )
!
        call givens_rv_rm_right( vec(:n), matt(:n,:m) )
!
        mat(:m,:n) = transpose( matt(:n,:m) )
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( matt )
!
    else
!
!       ALLOCATE WORK ARRAY.
!
        allocate( tmp(n), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        tmp(:n) = mat(m,:n)
!
!       ZERO COMPONENTS OF  mat(:,1) .
!
        do i = m-1_i4b, 1_i4b, -1_i4b
!
!           ZERO  mat(i+1,1) .
!
!           DEFINE THE GIVENS ROTATION.
!
            if ( abs( mat(i,1_i4b) )>abs( tmp(1_i4b) ) ) then
!
                tn = tmp(1_i4b)/mat(i,1_i4b)
                sn = sqrt( one + tn*tn )
                cs = sign( one/sn , mat(i,1_i4b) )
                sn = cs*tn
!
            else if ( tmp(1_i4b)/=zero ) then
!
                tn = mat(i,1_i4b)/tmp(1_i4b)
                cs = sqrt( one + tn*tn )
                sn = sign(one/cs , tmp(1_i4b) )
                cs = sn*tn
!
            else
!
                mat(i+1_i4b,:n) = tmp(:n)
                tmp(:n) = mat(i,:n)
!
                cycle
!
            end if
!
!           APPLY ROTATION.
!
            mat(i+1_i4b,:n) = cs*tmp(:n) - sn*mat(i,:n)
            tmp(:n) = cs*mat(i,:n) + sn*tmp(:n)
!
        end do
!
!       ZERO  mat(1,1) .
!
!       DEFINE THE GIVENS ROTATION.
!
        if ( abs( vec(1_i4b) )>abs( tmp(1_i4b) ) ) then
!
            tn = tmp(1_i4b)/vec(1_i4b)
            sn = sqrt( one + tn*tn )
            cs = sign( one/sn , vec(1_i4b) )
            sn = cs*tn
!
        else if ( tmp(1_i4b)/=zero ) then
!
            tn = vec(1_i4b)/tmp(1_i4b)
            cs = sqrt( one + tn*tn )
            sn = sign(one/cs , tmp(1_i4b) )
            cs = sn*tn
!
        else
!
            mat(1_i4b,:n) = tmp(:n)
!
            return
!
        end if
!
!       APPLY ROTATION.
!
        mat(1_i4b,:n) = cs*tmp(:n) - sn*vec(:n)
        vec(:n) = cs*vec(:n) + sn*tmp(:n)
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( tmp )
!
    end if
!
!
! END OF SUBROUTINE givens_rv_rm_left
! ___________________________________
!
    end subroutine givens_rv_rm_left
!
! =========================================================================================
!
    subroutine givens_rv_rm_right( vec, mat )
!
! Purpose
! _______
!
!   GIVENS_VEC_MAT_RIGHT defines and applies a serie of Givens rotations on a n-vector 
!   VEC and on the columns of a n-by-p matrix MAT. The rotations are designed to annilhate
!   all the elements of the first row of MAT.
!                     
!
! Arguments
! _________
!
!   VEC     (INPUT/OUTPUT) real(stnd), dimension(:)
!           On input, the n-vector to rotate. VEC(1) is used to define the rotations.
!
!           On output, the transformed vector overwrites VEC.
!
!   MAT     (INPUT/OUTPUT) real(stnd), dimension(:,:)
!           On input, the matrix to be transformed. MAT(1,:) is used to define
!           the rotations.
!
!           On output, the transformed matrix overwrites MAT and MAT(1,:) is
!           equal to zero.
!
!
! Further Details
! _______________
!
!   It is assumed that size( VEC ) = size( MAT, 1 ) .
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout) :: vec(:), mat(:,:)
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: tn, cs, sn, tmp(size(vec))
!
    integer(i4b) :: i, m, n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( mat, 2 )
    n = size( vec )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( m<=0_i4b .or. n<=0_i4b ) return
!
    tmp(:n) = mat(:n,m)
!
!   ZERO COMPONENTS OF  mat(1,:) .
!
    do i = m-1_i4b, 1_i4b, -1_i4b
!
!       ZERO  mat(1,i+1) .
!
!       DEFINE THE GIVENS ROTATION.
!
        if ( abs( mat(1_i4b,i) )>abs( tmp(1_i4b) ) ) then
!
            tn = tmp(1_i4b)/mat(1_i4b,i)
            sn = sqrt( one + tn*tn )
            cs = sign( one/sn , mat(1_i4b,i) )
            sn = cs*tn
!
        else if ( tmp(1_i4b)/=zero ) then
!
            tn = mat(1_i4b,i)/tmp(1_i4b)
            cs = sqrt( one + tn*tn )
            sn = sign(one/cs , tmp(1_i4b) )
            cs = sn*tn
!
        else
!
            mat(:n,i+1_i4b) = tmp(:n)
            tmp(:n) = mat(:n,i)
!
            cycle
!
        end if
!
!       APPLY ROTATION.
!
        mat(:n,i+1_i4b) = cs*tmp(:n) - sn*mat(:n,i)
        tmp(:n) = cs*mat(:n,i) + sn*tmp(:n)
!
    end do
!
!   ZERO  mat(1,1) .
!
!   DEFINE THE GIVENS ROTATION.
!
    if ( abs( vec(1_i4b) )>abs( tmp(1_i4b) ) ) then
!
        tn = tmp(1_i4b)/vec(1_i4b)
        sn = sqrt( one + tn*tn )
        cs = sign( one/sn , vec(1_i4b) )
        sn = cs*tn
!
    else if ( tmp(1_i4b)/=zero ) then
!
        tn = vec(1_i4b)/tmp(1_i4b)
        cs = sqrt( one + tn*tn )
        sn = sign(one/cs , tmp(1_i4b) )
        cs = sn*tn
!
    else
!
        mat(:n,1_i4b) = tmp(:n)
!
        return
!
    end if
!
!   APPLY ROTATION.
!
    mat(:n,1_i4b) = cs*tmp(:n) - sn*vec(:n)
    vec(:n) = cs*vec(:n) + sn*tmp(:n)
!
!
! END OF SUBROUTINE givens_rv_rm_right
! ____________________________________
!
    end subroutine givens_rv_rm_right
!
!
! =========================================================================================
!                                FASTGIVENS ROTATIONS
! =========================================================================================
!
!
    subroutine define_rot_fastgivens( x1, x2, d1, d2, beta, alpha, type_rot )
!
! Purpose
! _______
!
!   DEFINE_ROT_FASTGIVENS generates a fast Givens plane rotation H (defined by BETA, ALPHA, 
!   and TYPE_ROT on output) and updated scale factors (D1 and D2), which zero X2. That is,
!
!              ( X1  X2 )  H  = ( R  0 )
!
!   , where H is equal to
!        
!     - (1     0)             ,if TYPE_ROT = 0.
!
!       (0     1)
!
!     - (1     0) (1   A)     ,if TYPE_ROT = 1.
!
!       (B     1) (0   1)
!
!     - (1     A) (1   0)     ,if TYPE_ROT = 2.
!
!       (0     1) (B   1) 
!
!     - (0    -1) (1   0)      ,if TYPE_ROT = 3.
!
!       (1     A) (-B  1) 
!
!     - (B     1) (1   A)     ,if TYPE_ROT = 4.
!
!       (1     0) (0  -1) 
!        
!
!   Furtermore, if on input, Y1 = X1*SQRT(D1) and Y2 = X2*SQRT(D2), then on output
!
!              ( X1  X2 )  H  diag( SQRT(D1)  SQRT(D2) ) = ( (X1*H11 + X2*H21)*SQRT(D1)  0 )
!
!   is equal to
!
!              ( Y1  Y2 )  ROT  = ( (Y1*CS + Y2*SN)  0 )
!
!   where CS and SN define a standard Givens plane rotation, ROT, which zeros Y2.
!   Thus, ROT is equal to
!
!              ( +CS  -SN  )
!
!              ( +SN   +CS ) 
!
!   with CS**(2) + SN**(2) = 1.
!
!
! Arguments
! _________
!
!   X1       (INPUT) real(stnd)
!            First component of vector to be transformed.
!
!   X2       (INPUT) real(stnd)
!            Second component of vector to be transformed.
!
!   D1       (INPUT/OUTPUT) real(stnd)
!            On input, first scale factor.
!
!            On output, D1 is replaced with the update scale factor.
!
!   D2       (INPUT/OUTPUT) real(stnd)
!            On input, second scale factor.
!
!            On output, D2 is replaced with the update scale factor.
!
!   BETA     (OUTPUT) real(stnd)
!            The real scalar B which defines the transformation matrix H.
!
!   ALPHA    (OUTPUT) real(stnd)
!            The real scalar A which defines the transformation matrix H.
!
!   TYPE_ROT (OUTPUT) integer(i2b)
!            Integer which defines the transformation matrix H.
!
!
! Further Details
! _______________
!
!   X1 and X2 are unchanged on return.
!
!   It is assumed that D1 and D2 are positive scalars.
!
!   IF D1>=D2, D1 is diminished and D2 is augmented. IF D1<D2, D2 is diminished 
!   and D1 is augmented. The decrease or increase in magnitude of D1 and D2 
!   are bounded by 1/2 and 2, respectively.
!
!   Normally, the subprogram APPLY_ROT_FASTGIVENS( VECX1, VECX2, BETA, ALPHA, TYPE_ROT )  
!   will next be called to apply the rotation to a n-by-2 matrix [ VECX1  VECX2 ].
!
!   This subroutine is a square root free implementation of the two-way branch algorithm 
!   (fast plane rotations with dynamic scaling to avoid overflow/underflow) described in
!   reference (1).
!
!   For further details, see:
!
!   (1) Anda, A.A. and Park, H., 1994:
!            Fast plane rotations with dynamic scaling.
!            Siam J. Matrix Anal. Appl., 15, 162-174.
!
!    
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)     :: x1, x2
    real(stnd), intent(inout)  :: d1, d2
    real(stnd), intent(out)    :: beta, alpha
!
    integer(i2b), intent(out)  :: type_rot
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: tmp, gam
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   LET o BE THE ANGLE OF THE GIVENS ROTATION.
!
    beta = x2*d2
!
    if ( beta/=zero ) then
!
        alpha = x1*d1
!
        if ( alpha*x1>=beta*x2 ) then
!
!           HERE :  0         <  abs(o)      <= PI/4, 
!                   1/sqrt(2) <= abs(cos(o)) <  1.
!
            beta = beta/alpha
            alpha = -x2/x1
!
!           gam = 1/cos(o)**2 IS AT LEAST one AT MOST two.
!
            gam = one - alpha*beta
!
            if ( d1>=d2 ) then
!
!               HERE  d1 >= d2 , DECREASE  d1  AND INCREASE  d2.
!
                type_rot = 1_i2b
                alpha = alpha/gam
                d1 = d1/gam
                d2 = d2*gam
            else
!
!               HERE  d2 > d1 , DECREASE  d2  AND INCREASE  d1.
!
                type_rot = 2_i2b
                beta = beta/gam
                d1 = d1*gam
                d2 = d2/gam
            end if
!
        else
!
!           HERE :  PI/4      < abs(o)      <= PI/2,
!                   1/sqrt(2) < abs(sin(o)) <= 1.
!
            beta = alpha/beta
            alpha = x1/x2
!
!           gam = 1/sin(o)**2 IS AT LEAST one AT MOST two.
!
            gam = one + alpha*beta
            tmp = d1
!
            if ( d1>=d2 ) then
!
!               HERE  d1 >= d2 , DECREASE  d1  AND INCREASE  d2.
!
                type_rot = 3_i2b
                beta = beta/gam
                d1 = d2*gam
                d2 = tmp/gam
            else
!
!               HERE  d2 > d1 , DECREASE  d2  AND INCREASE  d1.
!
                type_rot = 4_i2b
                alpha = alpha/gam
                d1 = d2/gam
                d2 = tmp*gam
            end if
!
        end if
!
    else
!
!       HERE o = zero, H IS THE IDENTITY MATRIX.
!
        type_rot = 0_i2b
        alpha = zero
!
    end if
!
!
! END OF SUBROUTINE define_rot_fastgivens
! _______________________________________
!
    end subroutine define_rot_fastgivens
!
! =========================================================================================
!
    subroutine define_rot_fastgivens2( x1, x2, d1, d2, beta, alpha, type_rot )
!
! Purpose
! _______
!
!   DEFINE_ROT_FASTGIVENS2 generates a fast Givens plane rotation H (defined by BETA, ALPHA,
!   and TYPE_ROT on output) and updated scale factors (D1 and D2), which zero X2. That is,
!
!              ( X1  X2 )  H  = ( R  0 )
!
!   , where H is equal to
!        
!     - (1     0)             ,if TYPE_ROT = 0.
!
!       (0     1)
!
!     - (1     0) (1   A)     ,if TYPE_ROT = 1.
!
!       (B     1) (0   1)
!
!     - (1     A) (1   0)     ,if TYPE_ROT = 2.
!
!       (0     1) (B   1) 
!
!     - (0    -1) (1   0)      ,if TYPE_ROT = 3.
!
!       (1     A) (-B  1) 
!
!     - (B     1) (1   A)     ,if TYPE_ROT = 4.
!
!       (1     0) (0  -1) 
!
!
!   Furtermore, if on input, Y1 = X1*D1 and Y2 = X2*D2, then on output
!
!              ( X1  X2 )  H  diag( D1  D2 ) = ( (X1*H11 + X2*H21)*D1  0 )
!
!   is equal to
!
!              ( Y1  Y2 )  ROT  = ( (Y1*CS + Y2*SN)  0 )
!
!   where CS and SN define a standard Givens plane rotation, ROT, which zeros Y2.
!   Thus, ROT is equal to
!
!              ( +CS  -SN  )
!
!              ( +SN   +CS ) 
!
!   with CS**(2) + SN**(2) = 1.
!
!
! Arguments
! _________
!
!   X1       (INPUT) real(stnd)
!            First component of vector to be transformed.
!
!   X2       (INPUT) real(stnd)
!            Second component of vector to be transformed.
!
!   D1       (INPUT/OUTPUT) real(stnd)
!            On input, first scale factor.
!
!            On output, D1 is replaced with the update scale factor.
!
!   D2       (INPUT/OUTPUT) real(stnd)
!            On input, second scale factor.
!
!            On output, D2 is replaced with the update scale factor.
!
!   BETA     (OUTPUT) real(stnd)
!            The real scalar B which defines the transformation matrix H.
!
!   ALPHA    (OUTPUT) real(stnd)
!            The real scalar A which defines the transformation matrix H.
!
!   TYPE_ROT (OUTPUT) integer(i2b)
!            Integer which defines the transformation matrix H.
!
!
! Further Details
! _______________
!
!   X1 and X2 are unchanged on return.
!
!   It is assumed that D1 and D2 are positive scalars.
!
!   IF D1>=D2, D1 is diminished and D2 is augmented. IF D1<D2, D2 is diminished 
!   and D1 is augmented. The decrease or increase in magnitude of D1 and D2 
!   are bounded by 1/sqrt(2) and sqrt(2), respectively.
!
!   Normally, the subprogram APPLY_ROT_FASTGIVENS( VECX1, VECX2, BETA, ALPHA, TYPE_ROT )  
!   will next be called to apply the rotation to a n-by-2 matrix [ VECX1  VECX2 ].
!
!   This subroutine is an implementation of the two-way branch algorithm (fast plane 
!   rotations with dynamic scaling to avoid overflow/underflow) described in reference (1).
!
!   For further details, see:
!
!   (1) Anda, A.A. and Park, H., 1994:
!            Fast plane rotations with dynamic scaling.
!            Siam J. Matrix Anal. Appl., 15, 162-174.
!
!    
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)     :: x1, x2
    real(stnd), intent(inout)  :: d1, d2
    real(stnd), intent(out)    :: beta, alpha
!
    integer(i2b), intent(out)  :: type_rot
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: t, cs, sn
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   LET o BE THE ANGLE OF THE GIVENS ROTATION.
!
    beta = x2*d2
!
    if ( beta/=zero ) then
!
        alpha = x1*d1
!
        if ( abs( alpha )>=abs( beta ) ) then
!
!           HERE :  0         <  abs(o)      <= PI/4, 
!                   0         <  abs(tan(o)) <= 1,
!                   1/sqrt(2) <= abs(cos(o)) <  1.
!
            t = beta/alpha
            cs = one/sqrt( one + t*t )
!
            if ( d1>=d2 ) then
!
!               HERE  d1 >= d2 , SO DECREASE  d1  AND INCREASE  d2.
!
                type_rot = 1_i2b
                beta = t*( d2/d1 )
                d1 = d1*cs
                d2 = d2/cs
                alpha = -t*( d1/d2 )
!
            else
!
!               HERE  d2 > d1 , SO DECREASE  d2  AND INCREASE  d1.
!
                type_rot = 2_i2b
                alpha = -t*( d1/d2 )
                d1 = d1/cs
                d2 = d2*cs
                beta = t*( d2/d1 )
!
            end if
!
        else
!
!           HERE :  PI/4      <  abs(o)        <= PI/2,
!                   0         <= abs(cotan(o)) <   1,
!                   1/sqrt(2) <  abs(sin(o))   <=  1.
!
            t = alpha/beta
            sn = one/sqrt( one + t*t )
            cs = d2
!
            if ( d1>=d2 ) then
!
!               HERE  d1 >= d2 , SO DECREASE  d1  AND INCREASE  d2.
!
                type_rot = 3_i2b
                alpha = t*( d2/d1 )
                d2 = d1*sn
                d1 = cs/sn
                beta = t*( d2/d1 )
!
            else
!
!               HERE  d2 >  d1, SO DECREASE  d2  AND INCREASE  d1.
!
                type_rot = 4_i2b
                beta = t*( d1/d2 )
                d2 = d1/sn
                d1 = cs*sn
                alpha = t*( d1/d2 )
!
            end if
!
        end if
!
    else
!
!       HERE  o = zero , H IS THE IDENTITY MATRIX.
!
        type_rot = 0_i2b
        alpha = zero
!
    end if
!
!
! END OF SUBROUTINE define_rot_fastgivens2
! ________________________________________
!
    end subroutine define_rot_fastgivens2
!
! =========================================================================================
!
    subroutine apply_rot_fastgivens_r( y1, y2, beta, alpha, type_rot )
!
! Purpose
! _______
!
!   APPLY_ROT_FASTGIVENS applies a fast Givens plane rotation H (defined by BETA, ALPHA,
!   and TYPE_ROT on input) to the vector ( Y1 Y2 ). That is,
!
!              ( Y1  Y2 )  H  = ( (Y1*H11 + Y2*H21)  (Y1*H12 + Y2*H22) )
!
!   where H is a 2-by-2 matrix defined as
!
!              ( H11  H12 )
!
!              ( H21  H22 ) 
!
!   More precisely, H takes one of the following forms:
!        
!     - (1     0)             ,if TYPE_ROT = 0.
!
!       (0     1)
!
!     - (1     0) (1   A)     ,if TYPE_ROT = 1.
!
!       (B     1) (0   1)
!
!     - (1     A) (1   0)     ,if TYPE_ROT = 2.
!
!       (0     1) (B   1) 
!
!     - (0    -1) (1   0)      ,if TYPE_ROT = 3.
!
!       (1     A) (-B  1) 
!
!     - (B     1) (1   A)     ,if TYPE_ROT = 4.
!
!       (1     0) (0  -1) 
!
!
! Arguments
! _________
!
!   Y1       (INPUT/OUTPUT) real(stnd)
!            The first component of vector to be transformed.
!
!            On output, Y1*H11 + Y2*H21 overwrites Y1.
!
!   Y2       (INPUT/OUTPUT) real(stnd)
!            The second component of vector to be transformed.
!
!            On output, Y1*H12 + Y2*H22 overwrites Y2.
!
!   BETA     (OUTPUT) real(stnd)
!            The real scalar B which defines the transformation matrix H.
!
!   ALPHA    (OUTPUT) real(stnd)
!            The real scalar A which defines the transformation matrix H.
!
!   TYPE_ROT (INPUT) integer(i2b)
!            Integer which defines the transformation matrix H.
!
!
! Further Details
! _______________
!
!   Normally, the subprogram APPLY_ROT_FASTGIVENS( Y1, Y2, BETA, ALPHA, TYPE_ROT )
!   will be called to apply the transformation to the vector ( Y1 Y2 ) after a 
!   call to DEFINE_ROT_FASTGIVENS( X1, X2, BETA, ALPHA, TYPE_ROT ) or
!   DEFINE_ROT_FASTGIVENS2( X1, X2, BETA, ALPHA, TYPE_ROT ).
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout)  :: y1, y2
    real(stnd),   intent(in)     :: beta, alpha
!
    integer(i2b), intent(in)  :: type_rot
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: tmp
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   APPLY ROTATION.
!
    select case( type_rot )
!
        case( 1_i2b )
!
            y1 = y1 + beta*y2
            y2 = y2 + alpha*y1
!
        case( 2_i2b )
!
            y2 = y2 + alpha*y1
            y1 = y1 + beta*y2
!
        case( 3_i2b )
!
            tmp = y2
            y2 = alpha*tmp - y1
            y1 = tmp - beta*y2
!
        case( 4_i2b )
!
            tmp = y1
            y1 = y2 + beta*tmp
            y2 = alpha*y1 - tmp
!
    end select
!
!
! END OF SUBROUTINE apply_rot_fastgivens_r
! ________________________________________
!
    end subroutine apply_rot_fastgivens_r
!
! =========================================================================================
!
    subroutine apply_rot_fastgivens_rv( vecy1, vecy2, beta, alpha, type_rot )
!
! Purpose
! _______
!
!   APPLY_ROT_FASTGIVENS applies a fast Givens plane rotation H (defined by BETA, ALPHA,
!   and TYPE_ROT on input) to the n-by-2 matrix [ VECY1 VECY2 ]. That is,
!
!      [ VECY1  VECY2 ]  H  = [ (VECY1*H11 + VECY2*H21)  (VECY1*H12 + VECY2*H22) ]
!
!   where H is a 2-by-2 matrix defined as
!
!              ( H11  H12 )
!
!              ( H21  H22 ) 
!
!   More precisely, H takes one of the following forms:
!        
!     - (1     0)             ,if TYPE_ROT = 0.
!
!       (0     1)
!
!     - (1     0) (1   A)     ,if TYPE_ROT = 1.
!
!       (B     1) (0   1)
!
!     - (1     A) (1   0)     ,if TYPE_ROT = 2.
!
!       (0     1) (B   1) 
!
!     - (0    -1) (1   0)      ,if TYPE_ROT = 3.
!
!       (1     A) (-B  1) 
!
!     - (B     1) (1   A)     ,if TYPE_ROT = 4.
!
!       (1     0) (0  -1) 
!
!
! Arguments
! _________
!
!   VECY1    (INPUT/OUTPUT) real(stnd), dimension(:)
!            The first vector to be transformed.
!
!            On output, VECY1*H11 + VECY2*H21 overwrites VECY1.
!
!   VECY2    (INPUT/OUTPUT) real(stnd), dimension(:)
!            The second vector to be transformed.
!
!            On output, VECY1*H12 + VECY2*H22 overwrites VECY2.
!
!   BETA     (OUTPUT) real(stnd)
!            The real scalar B which defines the transformation matrix H.
!
!   ALPHA    (OUTPUT) real(stnd)
!            The real scalar A which defines the transformation matrix H.
!
!   TYPE_ROT (INPUT) integer(i2b)
!            Integer which defines the transformation matrix H.
!
!
! Further Details
! _______________
!
!   Normally, the subprogram APPLY_ROT_FASTGIVENS( VECY1, VECY2, BETA, ALPHA, TYPE_ROT )
!   will be called to apply the transformation to the n-by-2 matrix [ VECY1 VECY2 ] after a 
!   call to DEFINE_ROT_FASTGIVENS( VECY1(1), VECY2(1), BETA, ALPHA, TYPE_ROT ) or 
!   DEFINE_ROT_FASTGIVENS2( VECY1(1), VECY2(1), BETA, ALPHA, TYPE_ROT ).
!
!   It is assumed that VECY1 and VECY2 have the same size.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
#ifdef _BLAS
    use BLAS_interfaces, only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:)  :: vecy1, vecy2
    real(stnd), intent(in)                   :: beta, alpha
!
    integer(i2b), intent(in)  :: type_rot
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: tmp(size( vecy1 ))
!
    integer(i4b) :: n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   QUICK RETURN IF POSSIBLE.
!
    n = size( vecy1 )
!
    if ( n<=0_i4b ) return
!
!   APPLY ROTATION.
!
    select case( type_rot )
!
        case( 1_i2b )
!
#ifdef _BLAS
            call axpy( n, beta, vecy2(:n), 1_i4b, vecy1(:n), 1_i4b )
            call axpy( n, alpha, vecy1(:n), 1_i4b, vecy2(:n), 1_i4b )
#else
            vecy1(:n) = vecy1(:n) + beta*vecy2(:n)
            vecy2(:n) = vecy2(:n) + alpha*vecy1(:n)
#endif
!
        case( 2_i2b )
!
#ifdef _BLAS
            call axpy( n, alpha, vecy1(:n), 1_i4b, vecy2(:n), 1_i4b )
            call axpy( n, beta, vecy2(:n), 1_i4b, vecy1(:n), 1_i4b )
#else
            vecy2(:n) = vecy2(:n) + alpha*vecy1(:n)
            vecy1(:n) = vecy1(:n) + beta*vecy2(:n)
#endif
!
        case( 3_i2b )
!
            tmp(:n) = vecy2(:n)
            vecy2(:n) = alpha*tmp(:n) - vecy1(:n)
            vecy1(:n) = tmp(:n) - beta*vecy2(:n)
!
        case( 4_i2b )
!
            tmp(:n) = vecy1(:n)
            vecy1(:n) = vecy2(:n) + beta*tmp(:n)
            vecy2(:n) = alpha*vecy1(:n) - tmp(:n)
!
    end select
!
!
! END OF SUBROUTINE apply_rot_fastgivens_rv
! _________________________________________
!
    end subroutine apply_rot_fastgivens_rv
!
! =========================================================================================
!
    subroutine fastgivens_rv( vecx1, vecx2, d1, d2 )
!
! Purpose
! _______
!
!   FASTGIVENS generates and applies a fast Givens plane rotation H to 
!   the n-by-2 matrix [ VECX1  VECX2 ]. The rotation is designed to zero VECX2(1).
!   That is,
!
!      [ VECX1  VECX2 ]  H  = [ (VECX1*H11 + VECX2*H21)   (VECX1*H12 + VECX2*H22) ]
!
!   , where VECX1(1)*H12 + VECX2(1)*H22 = 0 and H is the 2-by-2 matrix:
!
!      ( H11  H12 )
!        
!      ( H21  H22 ) 
!        
!   Furthermore, the scale factors (D1 and D2) are updated accordingly. That is, if on input:
!
!          [ Y1  Y2 ] = [ VECX1  VECX2 ]  diag( SQRT(D1)  SQRT(D2) )
!
!   then on output:
!
!          [ VECX1  VECX2 ]  diag( SQRT(D1)  SQRT(D2) ) = [ Y1  Y2 ] ROT = [ (Y1*CS + Y2*SN)  (-SN*Y1 + CS*Y2) ]
!
!   where CS and SN define a standard Givens 2-by-2 plane rotation, ROT, which zeros -SN*Y1(1) + CS*Y2(1).
!   Thus, ROT has the following structure:
!
!       ( +CS  -SN )
!
!       ( +SN  +CS ) 
!
!   with CS**(2) + SN**(2) = 1.
!
!
! Arguments
! _________
!
!   VECX1    (INPUT/OUTPUT) real(stnd), dimension(:)
!            The first vector to be transformed.
!
!            On output, VECX1*H11 + VECX2*H21 overwrites VECX1.
!
!   VECX2    (INPUT/OUTPUT) real(stnd), dimension(:)
!            The second vector to be transformed.
!
!            On output, VECX1*H12 + VECX2*H22 overwrites VECX2.
!
!   D1       (INPUT/OUTPUT) real(stnd)
!            On input, first scale factor.
!
!            On output, D1 is replaced with the update scale factor.
!
!   D2       (INPUT/OUTPUT) real(stnd)
!            On input, second scale factor.
!
!            On output, D2 is replaced with the update scale factor.
!
!
! Further Details
! _______________
!
!   It is assumed that D1 and D2 are positive scalars and that VECX1 and VECX2 have 
!   the same size.
!
!   IF D1>=D2, D1 is diminished and D2 is augmented. IF D1<D2, D2 is diminished 
!   and D1 is augmented. The decrease or increase in magnitude of D1 and D2 
!   are bounded by 1/2 and 2, respectively.
!
!   This subroutine is a square root free implementation of the two-way branch algorithm 
!   (e.g. fast plane rotations with dynamic scaling to avoid overflow/underflow) described in
!   reference (1).
!
!   For further details, see:
!
!   (1) Anda, A.A. and Park, H., 1994:
!            Fast plane rotations with dynamic scaling.
!            Siam J. Matrix Anal. Appl., 15, 162-174.
!
!    
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
#ifdef _BLAS
    use BLAS_interfaces, only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:)  :: vecx1, vecx2
    real(stnd), intent(inout)                :: d1, d2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: tmp, tmp_vec(size(vecx1)), alpha, beta, gam
!
    integer(i4b) :: n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   QUICK RETURN IF POSSIBLE.
!
    n = size(vecx1)
!
    if ( n<=0_i4b ) return
!
    beta = vecx2(1_i4b)*d2
!
    if ( beta==zero ) return
!
!   DEFINE AND APPLY THE GIVENS ROTATION.
!   LET o BE THE ANGLE OF THE GIVENS ROTATION.
!
    alpha = vecx1(1_i4b)*d1
!
    if ( alpha*vecx1(1_i4b)>=beta*vecx2(1_i4b) ) then
!
!       HERE :  0         <  abs(o)      <= PI/4, 
!               1/sqrt(2) <= abs(cos(o)) <  1.
!
        beta = beta/alpha
        alpha = -vecx2(1_i4b)/vecx1(1_i4b)
!
!       gam = 1/cos(o)**2 IS AT LEAST one AT MOST two.
!
        gam = one - alpha*beta
!
        if ( d1>=d2 ) then
!
!           HERE d1 >= d2, DECREASE d1 AND INCREASE d2.
!
            d1 = d1/gam
            d2 = d2*gam
!
!           APPLY ROTATION.
!
            alpha = alpha/gam
!
#ifdef _BLAS
            call axpy( n, beta, vecx2(:n), 1_i4b, vecx1(:n), 1_i4b )
            call axpy( n, alpha, vecx1(:n), 1_i4b, vecx2(:n), 1_i4b )
#else
            vecx1(:n) = vecx1(:n) + beta*vecx2(:n)
            vecx2(:n) = vecx2(:n) + alpha*vecx1(:n)
#endif
!
        else
!
!           HERE d2 > d1, DECREASE d2 AND INCREASE d1.
!
            d1 = d1*gam
            d2 = d2/gam
!
!           APPLY ROTATION.
!
            beta = beta/gam
!
#ifdef _BLAS
            call axpy( n, alpha, vecx1(:n), 1_i4b, vecx2(:n), 1_i4b )
            call axpy( n, beta, vecx2(:n), 1_i4b, vecx1(:n), 1_i4b )
#else
            vecx2(:n) = vecx2(:n) + alpha*vecx1(:n)
            vecx1(:n) = vecx1(:n) + beta*vecx2(:n)
#endif
!
        end if
!
    else
!
!       HERE :  PI/4      < abs(o)      <= PI/2,
!               1/sqrt(2) < abs(sin(o)) <= 1.
!
        beta = alpha/beta
        alpha = vecx1(1_i4b)/vecx2(1_i4b)
!
!       gam = 1/sin(o)**2 IS AT LEAST one AT MOST two.
!
        gam = one + alpha*beta
        tmp = d1
!
        if ( d1>=d2 ) then
!
!           HERE d1 >= d2, DECREASE d1 AND INCREASE d2.
!
            d1 = d2*gam
            d2 = tmp/gam
!
!           APPLY ROTATION.
!
            beta = beta/gam
            tmp_vec(:n) = vecx2(:n)
!
            vecx2(:n) = alpha*tmp_vec(:n) - vecx1(:n)
            vecx1(:n) = tmp_vec(:n) - beta*vecx2(:n)
!
        else
!
!           HERE d2 > d1, DECREASE d2 AND INCREASE d1.
!
            d1 = d2/gam
            d2 = tmp*gam
!
!           APPLY ROTATION
!
            alpha = alpha/gam
            tmp_vec(:n) = vecx1(:n)
!
            vecx1(:n) = vecx2(:n) + beta*tmp_vec(:n)
            vecx2(:n) = alpha*vecx1(:n) - tmp_vec(:n)
!
        end if
!
    end if
!
!
! END OF SUBROUTINE fastgivens_rv
! _______________________________
!
    end subroutine fastgivens_rv
!
! =========================================================================================
!
    subroutine fastgivens_rv_t( vecx1, vecx2, d1, d2, beta, alpha, type_rot )
!
! Purpose
! _______
!
!   FASTGIVENS generates and applies a fast Givens plane rotation H (defined by 
!   BETA, ALPHA and TYPE_ROT on output) to the n-by-2 matrix [ VECX1  VECX2 ]. The rotation
!   is designed to zero VECX2(1). That is,
!
!      [ VECX1  VECX2 ]  H  = [ (VECX1*H11 + VECX2*H21)   (VECX1*H12 + VECX2*H22) ]
!
!   , where VECX1(1)*H12 + VECX2(1)*H22 = 0 and H is the 2-by-2 matrix:
!
!      ( H11  H12 )
!        
!      ( H21  H22 ) 
!        
!   and H takes one of the following forms:
!        
!     - (1     0)             ,if TYPE_ROT = 0.
!
!       (0     1)
!
!     - (1     0) (1   A)     ,if TYPE_ROT = 1.
!
!       (B     1) (0   1)
!
!     - (1     A) (1   0)     ,if TYPE_ROT = 2.
!
!       (0     1) (B   1) 
!
!     - (0    -1) (1   0)      ,if TYPE_ROT = 3.
!
!       (1     A) (-B  1) 
!
!     - (B     1) (1   A)     ,if TYPE_ROT = 4.
!
!       (1     0) (0  -1) 
!
!
!   Furthermore, the scale factors (D1 and D2) are updated accordingly. That is, if on input:
!
!          [ Y1  Y2 ] = [ VECX1  VECX2 ]  diag( SQRT(D1)  SQRT(D2) )
!
!   then on output:
!
!          [ VECX1  VECX2 ]  diag( SQRT(D1)  SQRT(D2) ) = [ Y1  Y2 ] ROT = [ (Y1*CS + Y2*SN)  (-SN*Y1 + CS*Y2) ]
!
!   where CS and SN define a standard Givens 2-by-2 plane rotation, ROT, which zeros -SN*Y1(1) + CS*Y2(1).
!   Thus, ROT has the following structure:
!
!       ( +CS  -SN )
!
!       ( +SN  +CS ) 
!
!   with CS**(2) + SN**(2) = 1.
!
!
! Arguments
! _________
!
!   VECX1    (INPUT/OUTPUT) real(stnd), dimension(:)
!            The first vector to be transformed.
!
!            On output, VECX1*H11 + VECX2*H21 overwrites VECX1.
!
!   VECX2    (INPUT/OUTPUT) real(stnd), dimension(:)
!            The second vector to be transformed.
!
!            On output, VECX1*H12 + VECX2*H22 overwrites VECX2.
!
!   D1       (INPUT/OUTPUT) real(stnd)
!            On input, first scale factor.
!
!            On output, D1 is replaced with the update scale factor.
!
!   D2       (INPUT/OUTPUT) real(stnd)
!            On input, second scale factor.
!
!            On output, D2 is replaced with the update scale factor.
!
!   BETA     (OUTPUT) real(stnd)
!            The real scalar B which defines the transformation matrix H.
!
!   ALPHA    (OUTPUT) real(stnd)
!            The real scalar A which defines the transformation matrix H.
!
!   TYPE_ROT (OUTPUT) integer(i2b)
!            Integer which defines the transformation matrix H.
!
!
! Further Details
! _______________
!
!   It is assumed that D1 and D2 are positive scalars and that VECX1 and VECX2 have 
!   the same size.
!
!   IF D1>=D2, D1 is diminished and D2 is augmented. IF D1<D2, D2 is diminished 
!   and D1 is augmented. The decrease or increase in magnitude of D1 and D2 
!   are bounded by 1/2 and 2, respectively.
!
!   This subroutine is a square root free implementation of the two-way branch algorithm 
!   (e.g. fast plane rotations with dynamic scaling to avoid overflow/underflow) described in
!   reference (1).
!
!   For further details, see:
!
!   (1) Anda, A.A. and Park, H., 1994:
!            Fast plane rotations with dynamic scaling.
!            Siam J. Matrix Anal. Appl., 15, 162-174.
!
!    
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
#ifdef _BLAS
    use BLAS_interfaces, only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:)  :: vecx1, vecx2
    real(stnd), intent(inout)                :: d1, d2
    real(stnd), intent(out)                  :: beta, alpha
!
    integer(i2b), intent(out)  :: type_rot
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: tmp, tmp_vec(size(vecx1)), gam
!
    integer(i4b) :: n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   QUICK RETURN IF POSSIBLE.
!
    n = size(vecx1)
    if ( n<=0_i4b ) return
!
    beta = vecx2(1_i4b)*d2
!
!   DEFINE AND APPLY THE GIVENS ROTATION.
!   LET o BE THE ANGLE OF THE GIVENS ROTATION.
!
    if ( beta/=zero ) then
!
        alpha = vecx1(1_i4b)*d1
!
        if ( alpha*vecx1(1_i4b)>=beta*vecx2(1_i4b) ) then
!
!           HERE :  0         <  abs(o)      <= PI/4, 
!                   1/sqrt(2) <= abs(cos(o)) <  1.
!
            beta = beta/alpha
            alpha = -vecx2(1_i4b)/vecx1(1_i4b)
!
!           gam = 1/cos(o)**2 IS AT LEAST one AT MOST two.
!
            gam = one - alpha*beta
!
            if ( d1>=d2 ) then
!
!               HERE d1 >= d2, DECREASE d1 AND INCREASE d2.
!
                d1 = d1/gam
                d2 = d2*gam
!
!               DEFINE ROTATION.
!
                type_rot = 1_i2b
                alpha = alpha/gam
!
!               APPLY ROTATION.
!
#ifdef _BLAS
                call axpy( n, beta, vecx2(:n), 1_i4b, vecx1(:n), 1_i4b )
                call axpy( n, alpha, vecx1(:n), 1_i4b, vecx2(:n), 1_i4b )
#else
                vecx1(:n) = vecx1(:n) + beta*vecx2(:n)
                vecx2(:n) = vecx2(:n) + alpha*vecx1(:n)
#endif
!
            else
!
!               HERE d2 > d1, DECREASE d2 AND INCREASE d1.
!
                d1 = d1*gam
                d2 = d2/gam
!
!               DEFINE ROTATION.
!
                type_rot = 2_i2b
                beta = beta/gam
!
!               APPLY ROTATION.
!
#ifdef _BLAS
                call axpy( n, alpha, vecx1(:n), 1_i4b, vecx2(:n), 1_i4b )
                call axpy( n, beta, vecx2(:n), 1_i4b, vecx1(:n), 1_i4b )
#else
                vecx2(:n) = vecx2(:n) + alpha*vecx1(:n)
                vecx1(:n) = vecx1(:n) + beta*vecx2(:n)
#endif
!
            end if
!
        else
!
!           HERE :  PI/4      < abs(o)      <= PI/2,
!                   1/sqrt(2) < abs(sin(o)) <= 1.
!
            beta = alpha/beta
            alpha = vecx1(1_i4b)/vecx2(1_i4b)
!
!           gam = 1/sin(o)**2 IS AT LEAST one AT MOST two.
!
            gam = one + alpha*beta
            tmp = d1
!
            if ( d1>=d2 ) then
!
!               HERE d1 >= d2, DECREASE d1 AND INCREASE d2.
!
                d1 = d2*gam
                d2 = tmp/gam
!
!               DEFINE ROTATION.
!
                type_rot = 3_i2b
                beta = beta/gam
!
!               APPLY ROTATION.
!
                tmp_vec(:n) = vecx2(:n)
                vecx2(:n) = alpha*tmp_vec(:n) - vecx1(:n)
                vecx1(:n) = tmp_vec(:n) - beta*vecx2(:n)
!
            else
!
!               HERE d2 > d1, DECREASE d2 AND INCREASE d1.
!
                d1 = d2/gam
                d2 = tmp*gam
!
!               DEFINE ROTATION.
!
                type_rot = 4_i2b
                alpha = alpha/gam
!
!               APPLY ROTATION.
!
                tmp_vec(:n) = vecx1(:n)
                vecx1(:n) = vecx2(:n) + beta*tmp_vec(:n)
                vecx2(:n) = alpha*vecx1(:n) - tmp_vec(:n)
!
            end if
!
        end if
!
    else
!
!       HERE o = zero, H IS THE IDENTITY MATRIX.
!
        type_rot = 0_i2b
        alpha = zero
!
    end if
!
!
! END OF SUBROUTINE fastgivens_rv_t
! _________________________________
!
    end subroutine fastgivens_rv_t
!
! =========================================================================================
!
    subroutine fastgivens2_rv( vecx1, vecx2, d1, d2 )
!
! Purpose
! _______
!
!   FASTGIVENS2 generates and applies a fast Givens plane rotation H to 
!   the n-by-2 matrix [ VECX1  VECX2 ]. The rotation is designed to zero VECX2(1).
!   That is,
!
!      [ VECX1  VECX2 ]  H  = [ (VECX1*H11 + VECX2*H21)   (VECX1*H12 + VECX2*H22) ]
!
!   , where VECX1(1)*H12 + VECX2(1)*H22 = 0 and H is the 2-by-2 matrix:
!
!      ( H11  H12 )
!        
!      ( H21  H22 ) 
!
!
!   , where VECX1(1)*H12 + VECX2(1)*H22 = 0.
!        
!   Furthermore, the scale factors (D1 and D2) are updated accordingly. That is, if on input:
!
!          [ Y1  Y2 ] = [ VECX1  VECX2 ]  diag( D1  D2 )
!
!   then on output:
!
!          [ VECX1  VECX2 ]  diag( D1  D2 ) = [ Y1  Y2 ] ROT = [ (Y1*CS + Y2*SN)  (-SN*Y1 + CS*Y2) ]
!
!   where CS and SN define a standard Givens 2-by-2 plane rotation, ROT, which zeros -SN*Y1(1) + CS*Y2(1).
!   Thus, ROT has the following structure:
!
!       ( +CS  -SN )
!
!       ( +SN  +CS ) 
!
!   with CS**(2) + SN**(2) = 1.
!
!
! Arguments
! _________
!
!   VECX1    (INPUT/OUTPUT) real(stnd), dimension(:)
!            The first vector to be transformed.
!
!            On output, VECX1*H11 + VECX2*H21 overwrites VECX1.
!
!   VECX2    (INPUT/OUTPUT) real(stnd), dimension(:)
!            The second vector to be transformed.
!
!            On output, VECX1*H12 + VECX2*H22 overwrites VECX2.
!
!   D1       (INPUT/OUTPUT) real(stnd)
!            On input, first scale factor.
!
!            On output, D1 is replaced with the update scale factor.
!
!   D2       (INPUT/OUTPUT) real(stnd)
!            On input, second scale factor.
!
!            On output, D2 is replaced with the update scale factor.
!
!
! Further Details
! _______________
!
!   It is assumed that D1 and D2 are positive scalars and that VECX1 and VECX2 have 
!   the same size.
!
!   IF D1>=D2, D1 is diminished and D2 is augmented. IF D1<D2, D2 is diminished 
!   and D1 is augmented. The decrease or increase in magnitude of D1 and D2 
!   are bounded by 1/sqrt(2) and sqrt(2), respectively.
!
!   This subroutine is an implementation of the two-way branch algorithm (e.g. fast plane 
!   rotations with dynamic scaling to avoid overflow/underflow) described in reference (1).
!
!   For further details, see:
!
!   (1) Anda, A.A. and Park, H., 1994:
!            Fast plane rotations with dynamic scaling.
!            Siam J. Matrix Anal. Appl., 15, 162-174.
!
!    
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
#ifdef _BLAS
    use BLAS_interfaces, only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:)  :: vecx1, vecx2
    real(stnd), intent(inout)                :: d1, d2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: tmp_vec(size(vecx1)), t, cs, sn, beta, alpha
!
    integer(i4b) :: n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   QUICK RETURN IF POSSIBLE.
!
    n = size(vecx1)
!
    if ( n<=0_i4b ) return
!
    beta = vecx2(1_i4b)*d2
!
    if ( beta==zero ) return
!
!   DEFINE AND APPLY THE GIVENS ROTATION.
!   LET o BE THE ANGLE OF THE GIVENS ROTATION.
!
    alpha = vecx1(1_i4b)*d1
!
    if ( abs( alpha )>=abs( beta ) ) then
!
!       HERE :  0         <  abs(o)      <= PI/4, 
!               0         <  abs(tan(o)) <= 1,
!               1/sqrt(2) <= abs(cos(o)) <  1.
!
        t = beta/alpha
        cs = one/sqrt( one + t*t )
!
        if ( d1>=d2 ) then
!
!           HERE  d1 >= d2 , SO DECREASE  d1  AND INCREASE  d2.
!
            beta = t*( d2/d1 )
            d1 = d1*cs
            d2 = d2/cs
            alpha = -t*( d1/d2 )
!
!           APPLY ROTATION.
!
#ifdef _BLAS
            call axpy( n, beta, vecx2(:n), 1_i4b, vecx1(:n), 1_i4b )
            call axpy( n, alpha, vecx1(:n), 1_i4b, vecx2(:n), 1_i4b )
#else
            vecx1(:n) = vecx1(:n) + beta*vecx2(:n)
            vecx2(:n) = vecx2(:n) + alpha*vecx1(:n)
#endif
!
        else
!
!           HERE  d2 > d1 , SO DECREASE  d2  AND INCREASE  d1.
!
            alpha = -t*( d1/d2 )
            d1 = d1/cs
            d2 = d2*cs
            beta = t*( d2/d1 )
!
!           APPLY ROTATION.
!
#ifdef _BLAS
            call axpy( n, alpha, vecx1(:n), 1_i4b, vecx2(:n), 1_i4b )
            call axpy( n, beta, vecx2(:n), 1_i4b, vecx1(:n), 1_i4b )
#else
            vecx2(:n) = vecx2(:n) + alpha*vecx1(:n)
            vecx1(:n) = vecx1(:n) + beta*vecx2(:n)
#endif
!
        end if
!
    else
!
!       HERE :  PI/4      <  abs(o)        <= PI/2,
!               0         <= abs(cotan(o)) <   1,
!               1/sqrt(2) <  abs(sin(o))   <=  1.
!
        t = alpha/beta
        sn = one/sqrt( one + t*t )
        cs = d2
!
        if ( d1>=d2 ) then
!
!           HERE  d1 >= d2 , SO DECREASE  d1  AND INCREASE  d2.
!
            alpha = t*( d2/d1 )
            d2 = d1*sn
            d1 = cs/sn
            beta = t*( d2/d1 )
!
!           APPLY ROTATION.
!
            tmp_vec(:n) = vecx2(:n)
            vecx2(:n) = alpha*tmp_vec(:n) - vecx1(:n)
            vecx1(:n) = tmp_vec(:n) - beta*vecx2(:n)
!
        else
!
!           HERE  d2 >  d1, SO DECREASE  d2  AND INCREASE  d1.
!
            beta = t*( d1/d2 )
            d2 = d1/sn
            d1 = cs*sn
            alpha = t*( d1/d2 )
!
!           APPLY ROTATION.
!
            tmp_vec(:n) = vecx1(:n)
            vecx1(:n) = vecx2(:n) + beta*tmp_vec(:n)
            vecx2(:n) = alpha*vecx1(:n) - tmp_vec(:n)
!
        end if
!
    end if
!
!
! END OF SUBROUTINE fastgivens2_rv
! ________________________________
!
    end subroutine fastgivens2_rv
!
! =========================================================================================
!
    subroutine fastgivens2_rv_t( vecx1, vecx2, d1, d2, beta, alpha, type_rot )
!
! Purpose
! _______
!
!   FASTGIVENS2 generates and applies a fast Givens plane rotation H (defined by 
!   BETA, ALPHA and TYPE_ROT on output) to the n-by-2 matrix [ VECX1  VECX2 ]. The rotation
!   is designed to zero VECX2(1). That is,
!
!      [ VECX1  VECX2 ]  H  = [ (VECX1*H11 + VECX2*H21)   (VECX1*H12 + VECX2*H22) ]
!
!   , where VECX1(1)*H12 + VECX2(1)*H22 = 0 and H is the 2-by-2 matrix:
!
!      ( H11  H12 )
!        
!      ( H21  H22 ) 
!        
!   and takes one of the following forms:
!        
!     - (1     0)             ,if TYPE_ROT = 0.
!
!       (0     1)
!
!     - (1     0) (1   A)     ,if TYPE_ROT = 1.
!
!       (B     1) (0   1)
!
!     - (1     A) (1   0)     ,if TYPE_ROT = 2.
!
!       (0     1) (B   1) 
!
!     - (0    -1) (1   0)      ,if TYPE_ROT = 3.
!
!       (1     A) (-B  1) 
!
!     - (B     1) (1   A)     ,if TYPE_ROT = 4.
!
!       (1     0) (0  -1) 
!
!
!   Furthermore, the scale factors (D1 and D2) are updated accordingly. That is, if on input:
!
!          [ Y1  Y2 ] = [ VECX1  VECX2 ]  diag( D1  D2 )
!
!   then on output:
!
!          [ VECX1  VECX2 ]  diag( D1  D2 ) = [ Y1  Y2 ] ROT = [ (Y1*CS + Y2*SN)  (-SN*Y1 + CS*Y2) ]
!
!   where CS and SN define a standard Givens 2-by-2 plane rotation, ROT, which zeros -SN*Y1(1) + CS*Y2(1).
!   Thus, ROT has the following structure:
!
!       ( +CS  -SN )
!
!       ( +SN  +CS ) 
!
!   with CS**(2) + SN**(2) = 1.
!
!
! Arguments
! _________
!
!   VECX1    (INPUT/OUTPUT) real(stnd), dimension(:)
!            The first vector to be transformed.
!
!            On output, VECX1*H11 + VECX2*H21 overwrites VECX1.
!
!   VECX2    (INPUT/OUTPUT) real(stnd), dimension(:)
!            The second vector to be transformed.
!
!            On output, VECX1*H12 + VECX2*H22 overwrites VECX2.
!
!   D1       (INPUT/OUTPUT) real(stnd)
!            On input, first scale factor.
!
!            On output, D1 is replaced with the update scale factor.
!
!   D2       (INPUT/OUTPUT) real(stnd)
!            On input, second scale factor.
!
!            On output, D2 is replaced with the update scale factor.
!
!   BETA     (OUTPUT) real(stnd)
!            The real scalar B which defines the transformation matrix H.
!
!   ALPHA    (OUTPUT) real(stnd)
!            The real scalar A which defines the transformation matrix H.
!
!   TYPE_ROT (OUTPUT) integer(i2b)
!            Integer which defines the transformation matrix H.
!
!
! Further Details
! _______________
!
!   It is assumed that D1 and D2 are positive scalars and that VECX1 and VECX2 have 
!   the same size.
!
!   IF D1>=D2, D1 is diminished and D2 is augmented. IF D1<D2, D2 is diminished 
!   and D1 is augmented. The decrease or increase in magnitude of D1 and D2 
!   are bounded by 1/sqrt(2) and sqrt(2), respectively.
!
!   This subroutine is an implementation of the two-way branch algorithm (e.g. fast plane 
!   rotations with dynamic scaling to avoid overflow/underflow) described in reference (1).
!
!   For further details, see:
!
!   (1) Anda, A.A. and Park, H., 1994:
!            Fast plane rotations with dynamic scaling.
!            Siam J. Matrix Anal. Appl., 15, 162-174.
!
!    
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
#ifdef _BLAS
    use BLAS_interfaces, only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:)  :: vecx1, vecx2
    real(stnd), intent(inout)                :: d1, d2
    real(stnd), intent(out)                  :: beta, alpha
!
    integer(i2b), intent(out)  :: type_rot
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: tmp_vec(size(vecx1)), t, cs, sn
!
    integer(i4b) :: n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   QUICK RETURN IF POSSIBLE.
!
    n = size(vecx1)
!
    if ( n<=0_i4b ) return
!
    beta = vecx2(1_i4b)*d2
!
!   DEFINE AND APPLY THE GIVENS ROTATION.
!   LET o BE THE ANGLE OF THE GIVENS ROTATION.
!
    if ( beta/=zero ) then
!
        alpha = vecx1(1_i4b)*d1
!
        if ( abs( alpha )>=abs( beta ) ) then
!
!           HERE :  0         <  abs(o)      <= PI/4, 
!                   0         <  abs(tan(o)) <= 1,
!                   1/sqrt(2) <= abs(cos(o)) <  1.
!
            t = beta/alpha
            cs = one/sqrt( one + t*t )
!
            if ( d1>=d2 ) then
!
!               HERE  d1 >= d2 , SO DECREASE  d1  AND INCREASE  d2.
!
                type_rot = 1_i2b
                beta = t*( d2/d1 )
                d1 = d1*cs
                d2 = d2/cs
                alpha = -t*( d1/d2 )
!
!               APPLY ROTATION.
!
#ifdef _BLAS
                call axpy( n, beta, vecx2(:n), 1_i4b, vecx1(:n), 1_i4b )
                call axpy( n, alpha, vecx1(:n), 1_i4b, vecx2(:n), 1_i4b )
#else
                vecx1(:n) = vecx1(:n) + beta*vecx2(:n)
                vecx2(:n) = vecx2(:n) + alpha*vecx1(:n)
#endif
!
            else
!
!               HERE  d2 > d1 , SO DECREASE  d2  AND INCREASE  d1.
!
                type_rot = 2_i2b
                alpha = -t*( d1/d2 )
                d1 = d1/cs
                d2 = d2*cs
                beta = t*( d2/d1 )
!
!               APPLY ROTATION.
!
#ifdef _BLAS
                call axpy( n, alpha, vecx1(:n), 1_i4b, vecx2(:n), 1_i4b )
                call axpy( n, beta, vecx2(:n), 1_i4b, vecx1(:n), 1_i4b )
#else
                vecx2(:n) = vecx2(:n) + alpha*vecx1(:n)
                vecx1(:n) = vecx1(:n) + beta*vecx2(:n)
#endif
!
            end if
!
        else
!
!           HERE :  PI/4      <  abs(o)        <= PI/2,
!                   0         <= abs(cotan(o)) <   1,
!                   1/sqrt(2) <  abs(sin(o))   <=  1.
!
            t = alpha/beta
            sn = one/sqrt( one + t*t )
            cs = d2
!
            if ( d1>=d2 ) then
!
!               HERE  d1 >= d2 , SO DECREASE  d1  AND INCREASE  d2.
!
                type_rot = 3_i2b
                alpha = t*( d2/d1 )
                d2 = d1*sn
                d1 = cs/sn
                beta = t*( d2/d1 )
!
!               APPLY ROTATION.
!
                tmp_vec(:n) = vecx2(:n)
                vecx2(:n) = alpha*tmp_vec(:n) - vecx1(:n)
                vecx1(:n) = tmp_vec(:n) - beta*vecx2(:n)
!
            else
!
!               HERE  d2 >  d1, SO DECREASE  d2  AND INCREASE  d1.
!
                type_rot = 4_i2b
                beta = t*( d1/d2 )
                d2 = d1/sn
                d1 = cs*sn
                alpha = t*( d1/d2 )
!
!               APPLY ROTATION.
!
                tmp_vec(:n) = vecx1(:n)
                vecx1(:n) = vecx2(:n) + beta*tmp_vec(:n)
                vecx2(:n) = alpha*vecx1(:n) - tmp_vec(:n)
!
            end if
!
        end if
!
    else
!
!       HERE o = zero, H IS THE IDENTITY MATRIX.
!
        type_rot = 0_i2b
        alpha = zero
!
    end if
!
!
! END OF SUBROUTINE fastgivens2_rv_t
! __________________________________
!
    end subroutine fastgivens2_rv_t
!
! =========================================================================================
!
    subroutine fastgivens_rm_left( mat, matd )
!
! Purpose
! _______
!
!   FASTGIVENS_MAT_LEFT reduces the matrix MAT to upper trapezoidal form by applying 
!   a serie of fast Givens plane rotations on the rows of MAT.
!
!   The (row) scale factors (MATD) are updated accordingly.
!
!
! Arguments
! _________
!
!   MAT    (INPUT/OUTPUT) real(stnd), dimension(:,:)
!          The matrix to be transformed.
!
!          On output, the transformed matrix overwrites MAT.
!
!   MATD   (INPUT/OUTPUT) real(stnd), dimension(:)
!          On input, scale factors associated with the rows of MAT.
!
!          On output, MATD is replaced with the update scale factors.
!
!
! Further Details
! _______________
!
!   It is assumed that size( MATD ) = size( MAT, 1 ) and that MATD is a positive vector.
!
!   See description of FASTGIVENS for further details.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
    use Char_Constants,  only : allocate_error
    use Utilities,       only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout) :: mat(:,:), matd(:)
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                              :: tmp, alpha, beta, gam, pivd
    real(stnd), dimension(:),   allocatable :: tmp_vec
    real(stnd), dimension(:,:), allocatable :: matt
!
    integer(i4b) :: i, j, m, n
    integer      :: iok
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='fastgivens_mat_left'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( mat, 1 )
    n = size( mat, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( m<=0_i4b .or. n<=0_i4b ) return
!
!   ALLOCATE WORK VARIABLE FOR A FAST ALGORITHM IF POSSIBLE.
!
    allocate( matt(n,m), stat = iok )
!
    if ( iok==0 ) then
!
!       TRANSPOSE THE MATRIX AND USE A FAST ALGORITHM.
!
        matt(:n,:m) = transpose( mat(:m,:n) )
!
        call fastgivens_rm_right( matt(:n,:m), matd(:m) )
!
        mat(:m,:n) = transpose( matt(:n,:m) )
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( matt )
!
    else
!
!       ALLOCATE WORK ARRAY.
!
        allocate( tmp_vec(n), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        do j = 1_i4b, min( n, m )
!
!           ZERO COMPONENTS OF  mat(j+1:,j) .
!
            pivd = matd(j)
!
            do i = j+1_i4b, m
!
!               ZERO  mat(i,j) .
!
                beta = mat(i,j)*matd(i)
!
                if ( beta==zero ) cycle
!
!               LET o BE THE ANGLE OF THE GIVENS ROTATION.
!
                alpha = mat(j,j)*pivd
!
                if ( alpha*mat(j,j)>=beta*mat(i,j) ) then
!
!                   HERE :  0         <  abs(o)      <= PI/4, 
!                           1/sqrt(2) <= abs(cos(o)) <  1.
!
                    beta = beta/alpha
                    alpha = -mat(i,j)/mat(j,j)
!
!                   gam = 1/cos(o)**2 IS AT LEAST one AT MOST two.
!
                    gam = one - alpha*beta
!
                    if ( pivd>=matd(i) ) then
!
!                       HERE matd(j) >= matd(i), DECREASE matd(j) AND INCREASE matd(i).
!
                        pivd = pivd/gam
                        matd(i) = matd(i)*gam
!
!                       APPLY ROTATION.
!
                        alpha = alpha/gam
                        mat(j,j:n) = mat(j,j:n) + beta*mat(i,j:n)
                        mat(i,j:n) = mat(i,j:n) + alpha*mat(j,j:n)
!
                    else
!
!                       HERE matd(i) > matd(j), DECREASE matd(i) AND INCREASE matd(j).
!
                        pivd = pivd*gam
                        matd(i) = matd(i)/gam
!
!                       APPLY ROTATION.
!
                        beta = beta/gam
                        mat(i,j:n) = mat(i,j:n) + alpha*mat(j,j:n)
                        mat(j,j:n) = mat(j,j:n) + beta*mat(i,j:n)
!
                    end if
!
                else
!
!                   HERE :  PI/4      < abs(o)      <= PI/2,
!                           1/sqrt(2) < abs(sin(o)) <= 1.
!
                    beta = alpha/beta
                    alpha = mat(j,j)/mat(i,j)
!
!                   gam = 1/sin(o)**2 IS AT LEAST one AT MOST two.
!
                    gam = one + alpha*beta
                    tmp = pivd
!
                    if ( pivd>=matd(i) ) then
!
!                       HERE matd(j) >= matd(i), DECREASE matd(j) AND INCREASE matd(i).
!
                        pivd = matd(i)*gam
                        matd(i) = tmp/gam
!
!                       APPLY ROTATION.
!
                        beta = beta/gam
                        tmp_vec(j:n) = mat(i,j:n)
                        mat(i,j:n) = alpha*tmp_vec(j:n) - mat(j,j:n)
                        mat(j,j:n) = tmp_vec(j:n) - beta*mat(i,j:n)
!
                    else
!
!                       HERE matd(i) > matd(j), DECREASE matd(i) AND INCREASE matd(j).
!
                        pivd = matd(i)/gam
                        matd(i) = tmp*gam
!
!                       APPLY ROTATION.
!
                        alpha = alpha/gam
                        tmp_vec(j:n) = mat(j,j:n)
                        mat(j,j:n) = mat(i,j:n) + beta*tmp_vec(j:n)
                        mat(i,j:n) = alpha*mat(j,j:n) - tmp_vec(j:n)
!
                    end if
!
                end if
!
            end do
!
            matd(j) = pivd
!
        end do
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( tmp_vec )
!
    end if
!
!
! END OF SUBROUTINE fastgivens_rm_left
! ____________________________________
!
    end subroutine fastgivens_rm_left
!
! =========================================================================================
!
    subroutine fastgivens_rm_right( mat, matd )
!
! Purpose
! _______
!
!   FASTGIVENS_MAT_RIGHT reduces the matrix MAT to lower trapezoidal form by applying 
!   a serie of fast Givens plane rotations on the columns of MAT.
!
!   The (column) scale factors (MATD) are updated accordingly.
!
!
! Arguments
! _________
!
!   MAT    (INPUT/OUTPUT) real(stnd), dimension(:,:)
!          The matrix to be transformed.
!
!          On output, the transformed matrix overwrites MAT.
!
!   MATD   (INPUT/OUTPUT) real(stnd), dimension(:)
!          On input, scale factors associated with the columns of MAT.
!
!          On output, MATD is replaced with the update scale factors.
!
!
! Further Details
! _______________
!
!   It is assumed that size( MATD ) = size( MAT, 2 ) and that MATD is a positive vector.
!
!   See description of FASTGIVENS for further details.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
#ifdef _BLAS
    use BLAS_interfaces, only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout) :: mat(:,:), matd(:)
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: tmp, tmp_vec(size(mat,1)), alpha, beta, gam, pivd
!
    integer(i4b) :: i, j, m, n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( mat, 1 )
    n = size( mat, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( m<=0_i4b .or. n<=0_i4b ) return
!
    do j = 1_i4b, min( n, m )
!
!       ZERO COMPONENTS OF  mat(j,j+1:) .
!
        pivd = matd(j)
!
        do i = j+1_i4b, n
!
!           ZERO  mat(j,i) .
!
            beta = mat(j,i)*matd(i)
!
            if ( beta==zero ) cycle
!
!           LET o BE THE ANGLE OF THE GIVENS ROTATION.
!
            alpha = mat(j,j)*pivd
!
            if ( alpha*mat(j,j)>=beta*mat(j,i) ) then
!
!               HERE :  0         <  abs(o)      <= PI/4, 
!                       1/sqrt(2) <= abs(cos(o)) <  1.
!
                beta = beta/alpha
                alpha = -mat(j,i)/mat(j,j)
!
!               gam = 1/cos(o)**2 IS AT LEAST one AT MOST two.
!
                gam = one - alpha*beta
!
                if ( pivd>=matd(i) ) then
!
!                   HERE matd(j) >= matd(i), DECREASE matd(j) AND INCREASE matd(i).
!
                    pivd = pivd/gam
                    matd(i) = matd(i)*gam
!
!                   APPLY ROTATION.
!
                    alpha = alpha/gam
#ifdef _BLAS
                    call axpy( m-j+1_i4b, beta, mat(j:m,i), 1_i4b, mat(j:m,j), 1_i4b )
                    call axpy( m-j+1_i4b, alpha, mat(j:m,j), 1_i4b, mat(j:m,i), 1_i4b )
#else
                    mat(j:m,j) = mat(j:m,j) + beta*mat(j:m,i)
                    mat(j:m,i) = mat(j:m,i) + alpha*mat(j:m,j)
#endif
!
                else
!
!                   HERE matd(i) > matd(j), DECREASE matd(i) AND INCREASE matd(j).
!
                    pivd = pivd*gam
                    matd(i) = matd(i)/gam
!
!                   APPLY ROTATION.
!
                    beta = beta/gam
#ifdef _BLAS
                    call axpy( m-j+1_i4b, alpha, mat(j:m,j), 1_i4b, mat(j:m,i), 1_i4b )
                    call axpy( m-j+1_i4b, beta, mat(j:m,i), 1_i4b, mat(j:m,j), 1_i4b )
#else
                    mat(j:m,i) = mat(j:m,i) + alpha*mat(j:m,j)
                    mat(j:m,j) = mat(j:m,j) + beta*mat(j:m,i)
#endif
!
                end if
!
            else
!
!               HERE :  PI/4      < abs(o)      <= PI/2,
!                       1/sqrt(2) < abs(sin(o)) <= 1.
!
                beta = alpha/beta
                alpha = mat(j,j)/mat(j,i)
!
!               gam = 1/sin(o)**2 IS AT LEAST one AT MOST two.
!
                gam = one + alpha*beta
                tmp = pivd
!
                if ( pivd>=matd(i) ) then
!
!                   HERE matd(j) >= matd(i), DECREASE matd(j) AND INCREASE matd(i).
!
                    pivd = matd(i)*gam
                    matd(i) = tmp/gam
!
!                   APPLY ROTATION.
!
                    beta = beta/gam
                    tmp_vec(j:m) = mat(j:m,i)
                    mat(j:m,i) = alpha*tmp_vec(j:m) - mat(j:m,j)
                    mat(j:m,j) = tmp_vec(j:m) - beta*mat(j:m,i)
!
                else
!
!                   HERE matd(i) > matd(j), DECREASE matd(i) AND INCREASE matd(j).
!
                    pivd = matd(i)/gam
                    matd(i) = tmp*gam
!
!                   APPLY ROTATION.
!
                    alpha = alpha/gam
                    tmp_vec(j:m) = mat(j:m,j)
                    mat(j:m,j) = mat(j:m,i) + beta*tmp_vec(j:m)
                    mat(j:m,i) = alpha*mat(j:m,j) - tmp_vec(j:m)
!
                end if
!
            end if
!
        end do
!
        matd(j) = pivd
!
    end do
!
!
! END OF SUBROUTINE fastgivens_rm_right
! _____________________________________
!
    end subroutine fastgivens_rm_right
!
! =========================================================================================
!
    subroutine fastgivens_rv_rm_left( vec, mat, vecd, matd )
!
! Purpose
! _______
!
!   FASTGIVENS_VEC_MAT_LEFT defines and applies a serie of fast Givens plane rotations on 
!   the n-vector VEC and on the rows of a m-by-n matrix MAT. The rotations are designed
!   to annilhate all the elements of the first column of MAT.
!
!   The (row) scale factors (VECD and MATD) are updated accordingly.
!
!
! Arguments
! _________
!
!   VEC    (INPUT/OUTPUT) real(stnd), dimension(:)
!          On input, the n-vector to rotate. VEC(1) is used to define the rotations.
!
!          On output, the transformed vector overwrites VEC.
!
!   MAT    (INPUT/OUTPUT) real(stnd), dimension(:,:)
!          On input, the matrix to be transformed. MAT(:,1) is used to define
!          the rotations.
!
!          On output, the transformed matrix overwrites MAT and MAT(:,1) is
!          equal to zero (within numerical accuracy).
!
!   VECD   (INPUT/OUTPUT) real(stnd)
!          On input, scale factor associated with the n-vector VEC.
!
!          On output, VECD is replaced with the update scale factor.
!
!   MATD   (INPUT/OUTPUT) real(stnd), dimension(:)
!          On input, scale factors associated with the rows of MAT.
!
!          On output, MATD is replaced with the update scale factors.
!
!
! Further Details
! _______________
!
!   It is assumed that:
!
!       - size( VEC )  = size( MAT, 2 );
!       - VECD is a positive scalar;
!       - size(MATD) = size(MAT,1) and that MATD is a positive vector.
!
!   See description of FASTGIVENS for further details.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
    use Char_Constants,  only : allocate_error
    use Utilities,       only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout) :: vec(:), mat(:,:), vecd, matd(:)
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                              :: tmp, alpha, beta, gam
    real(stnd), dimension(:),   allocatable :: tmp_vec
    real(stnd), dimension(:,:), allocatable :: matt
!
    integer(i4b) :: i, j, m, n
    integer      :: iok
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='fastgivens_vec_mat_left'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( mat, 1 )
    n = size( vec )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( m<=0_i4b .or. n<=0_i4b ) return
!
!   ALLOCATE WORK VARIABLE FOR A FAST ALGORITHM IF POSSIBLE.
!
    allocate( matt(n,m), stat = iok )
!
    if ( iok==0 ) then
!
!       TRANSPOSE THE MATRIX AND USE A FAST ALGORITHM.
!
        matt(:n,:m) = transpose( mat(:m,:n) )
!
        call fastgivens_rv_rm_right( vec(:n), matt(:n,:m), vecd, matd(:m) )
!
        mat(:m,:n) = transpose( matt(:n,:m) )
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( matt )
!
    else
!
!       ALLOCATE WORK ARRAY.
!
        allocate( tmp_vec(n), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       ZERO COMPONENTS OF  mat(:,1) .
!
        do i = 1_i4b, m
!
!           ZERO  mat(i,1) .
!
            beta = mat(i,1_i4b)*matd(i)
!
            if ( beta==zero ) cycle
!
!           LET o BE THE ANGLE OF THE GIVENS ROTATION.
!
            alpha = vec(1_i4b)*vecd
!
            if ( alpha*vec(1_i4b)>=beta*mat(i,1_i4b) ) then
!
!               HERE :  0         <  abs(o)      <= PI/4, 
!                       1/sqrt(2) <= abs(cos(o)) <  1.
!
                beta = beta/alpha
                alpha = -mat(i,1_i4b)/vec(1_i4b)
!
!               gam = 1/cos(o)**2 IS AT LEAST one AT MOST two.
!
                gam = one - alpha*beta
!
                if ( vecd>=matd(i) ) then
!
!                   HERE vecd >= matd(i), DECREASE vecd AND INCREASE matd(i).
!
                    vecd = vecd/gam
                    matd(i) = matd(i)*gam
!
!                   APPLY ROTATION.
!
                    alpha = alpha/gam
                    vec(:n) = vec(:n) + beta*mat(i,:n)
                    mat(i,:n) = mat(i,:n) + alpha*vec(:n)
!
                else
!
!                   HERE matd(i) > vecd, DECREASE matd(i) AND INCREASE vecd.
!
                    vecd = vecd*gam
                    matd(i) = matd(i)/gam
!
!                   APPLY ROTATION.
!
                    beta = beta/gam
                    mat(i,:n) = mat(i,:n) + alpha*vec(:n)
                    vec(:n) = vec(:n) + beta*mat(i,:n)
!
                end if
!
            else
!
!               HERE :  PI/4      < abs(o)      <= PI/2,
!                       1/sqrt(2) < abs(sin(o)) <= 1.
!
                beta = alpha/beta
                alpha = vec(1_i4b)/mat(i,1_i4b)
!
!               gam = 1/sin(o)**2 IS AT LEAST one AT MOST two.
!
                gam = one + alpha*beta
                tmp = vecd
!
                if ( vecd>=matd(i) ) then
!
!                   HERE vecd >= matd(i), DECREASE vecd AND INCREASE matd(i).
!
                    vecd = matd(i)*gam
                    matd(i) = tmp/gam
!
!                   APPLY ROTATION.
!
                    beta = beta/gam
                    tmp_vec(:n) = mat(i,:n)
                    mat(i,:n) = alpha*tmp_vec(:n) - vec(:n)
                    vec(:n) = tmp_vec(:n) - beta*mat(i,:n)
!
                else
!
!                   HERE matd(i) > vecd, DECREASE matd(i) AND INCREASE vecd.
!
                    vecd = matd(i)/gam
                    matd(i) = tmp*gam
!
!                   APPLY ROTATION.
!
                    alpha = alpha/gam
                    tmp_vec(:n) = vec(:n)
                    vec(:n) = mat(i,:n) + beta*tmp_vec(:n)
                    mat(i,:n) = alpha*vec(:n) - tmp_vec(:n)
!
                end if
!
            end if
!
        end do
!
!        DEALLOCATE WORK ARRAY.
!
        deallocate( tmp_vec )
!
    end if
!
!
! END OF SUBROUTINE fastgivens_rv_rm_left
! _______________________________________
!
    end subroutine fastgivens_rv_rm_left
!
! =========================================================================================
!
    subroutine fastgivens_rv_rm_right( vec, mat, vecd, matd )
!
! Purpose
! _______
!
!   FASTGIVENS_VEC_MAT_RIGHT defines and applies a serie of fast Givens plane rotations on 
!   the m-vector VEC and on the columns of a m-by-n matrix MAT. The rotations are designed
!   to annilhate all the elements of the first row of MAT.
!
!   The (column) scale factors (VECD and MATD) are updated accordingly.
!
!
! Arguments
! _________
!
!   VEC    (INPUT/OUTPUT) real(stnd), dimension(:)
!          On input, the m-vector to rotate. VEC(1) is used to define the rotations.
!
!          On output, the transformed vector overwrites VEC.
!
!   MAT    (INPUT/OUTPUT) real(stnd), dimension(:,:)
!          On input, the matrix to be transformed. MAT(1,:) is used to define
!          the rotations.
!
!          On output, the transformed matrix overwrites MAT and MAT(1,:) is
!          equal to zero (within numerical accuracy).
!
!   VECD   (INPUT/OUTPUT) real(stnd)
!          On input, scale factor associated with the m-vector VEC.
!
!          On output, VECD is replaced with the update scale factor.
!
!   MATD   (INPUT/OUTPUT) real(stnd), dimension(:)
!          On input, scale factors associated with the columns of MAT.
!
!          On output, MATD is replaced with the update scale factors.
!
!
! Further Details
! _______________
!
!   It is assumed that:
!
!       - size( VEC )  = size( MAT, 1 );
!       - VECD is a positive scalar;
!       - size( MATD ) = size (MAT, 2 ) and that MATD is a positive vector.
!
!   See description of FASTGIVENS for further details.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, one
#ifdef _BLAS
    use BLAS_interfaces, only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout) :: vec(:), mat(:,:), vecd, matd(:)
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: tmp, tmp_vec(size(vec)), alpha, beta, gam
!
    integer(i4b) :: i, m, n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    m = size( vec )
    n = size( mat, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( m<=0_i4b .or. n<=0_i4b ) return
!
!   ZERO COMPONENTS OF  mat(1,:) .
!
    do i = 1_i4b, n
!
!       ZERO  mat(1,i) .
!
        beta = mat(1_i4b,i)*matd(i)
!
        if ( beta==zero ) cycle
!
!       LET o BE THE ANGLE OF THE GIVENS ROTATION.
!
        alpha = vec(1_i4b)*vecd
!
        if ( alpha*vec(1_i4b)>=beta*mat(1_i4b,i) ) then
!
!           HERE :  0         <  abs(o)      <= PI/4, 
!                   1/sqrt(2) <= abs(cos(o)) <  1.
!
            beta = beta/alpha
            alpha = -mat(1_i4b,i)/vec(1_i4b)
!
!           gam = 1/cos(o)**2 IS AT LEAST one AT MOST two.
!
            gam = one - alpha*beta
!
            if ( vecd>=matd(i) ) then
!
!               HERE vecd >= matd(i), DECREASE vecd AND INCREASE matd(i).
!
                vecd = vecd/gam
                matd(i) = matd(i)*gam
!
!               APPLY ROTATION.
!
                alpha = alpha/gam
#ifdef _BLAS
                call axpy( m, beta, mat(:m,i), 1_i4b, vec(:m), 1_i4b )
                call axpy( m, alpha, vec(:m), 1_i4b, mat(:m,i), 1_i4b )
#else
                vec(:m) = vec(:m) + beta*mat(:m,i)
                mat(:m,i) = mat(:m,i) + alpha*vec(:m)
#endif
!
            else
!
!               HERE matd(i) > vecd, DECREASE matd(i) AND INCREASE vecd.
!
                vecd = vecd*gam
                matd(i) = matd(i)/gam
!
!               APPLY ROTATION.
!
                beta = beta/gam
#ifdef _BLAS
                call axpy( m, alpha, vec(:m), 1_i4b, mat(:m,i), 1_i4b )
                call axpy( m, beta, mat(:m,i), 1_i4b, vec(:m), 1_i4b )
#else
                mat(:m,i) = mat(:m,i) + alpha*vec(:m)
                vec(:m) = vec(:m) + beta*mat(:m,i)
#endif
!
            end if
!
        else
!
!           HERE :  PI/4      < abs(o)      <= PI/2,
!                   1/sqrt(2) < abs(sin(o)) <= 1.
!
            beta = alpha/beta
            alpha = vec(1_i4b)/mat(1_i4b,i)
!
!           gam = 1/sin(o)**2 IS AT LEAST one AT MOST two.
!
            gam = one + alpha*beta
            tmp = vecd
!
            if ( vecd>=matd(i) ) then
!
!               HERE vecd >= matd(i), DECREASE vecd AND INCREASE matd(i).
!
                vecd = matd(i)*gam
                matd(i) = tmp/gam
!
!               APPLY ROTATION.
!
                beta = beta/gam
                tmp_vec(:m) = mat(:m,i)
                mat(:m,i) = alpha*tmp_vec(:m) - vec(:m)
                vec(:m) = tmp_vec(:m) - beta*mat(:m,i)
!
            else
!
!               HERE matd(i) > vecd, DECREASE matd(i) AND INCREASE vecd.
!
                vecd = matd(i)/gam
                matd(i) = tmp*gam
!
!               APPLY ROTATION.
!
                alpha = alpha/gam
                tmp_vec(:m) = vec(:m)
                vec(:m) = mat(:m,i) + beta*tmp_vec(:m)
                mat(:m,i) = alpha*vec(:m) - tmp_vec(:m)
!
            end if
!
        end if
!
    end do
!
!
! END OF SUBROUTINE fastgivens_rv_rm_right
! ________________________________________
!
    end subroutine fastgivens_rv_rm_right
!
!
! =========================================================================================
!
!
! ******************************
! END OF MODULE Giv_Procedures *
! ******************************
!    
end module Giv_Procedures
