module Lapack_interfaces
! ####################################################################################
!
! Copyright 2018 IRD
!
! This file is part of statpack.
!
! statpack is free software: you can redistribute it and/or modify
! it under the terms of the GNU Lesser General Public License as
! published by the Free Software Foundation, either version 3 of 
! the License, or (at your option) any later version.
!
! statpack is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU Lesser General Public License for more details.
!
! You can find a copy of the GNU Lesser General Public License
! in the statpack/doc directory.
!
! ####################################################################################
!                                                                                    *
! ************************************************************************************
! MODULE EXPORTING GENERIC INTERFACES FOR SELECTED SUBROUTINES AND FUNCTIONS IN THE  *
! LAPACK LIBRARY.                                                                    *
!                                                                                    *
! THIS INTERFACE MODULE ENSURES THAT CALLS TO LAPACK ROUTINES ARE CORRECT, WHEN USED *
! WITH STATPACK.                                                                     *
!                                                                                    *
! GENERIC INTERFACES ARE PRESENTLY PROVIDED FOR THE FOLLOWING LAPACK ROUTINES AND    *
! DRIVERS :                                                                          *
!                                                                                    *
!    Xsytrd, Xorgtr, Xormtr, Xsyev, Xsyevd, Xsyevr, Xsyevx,                          *
!    Xspev, Xspevd, Xspevx, Xsygv, Xsygvd, Xsygvx,                                   *
!    Xsteqr, Xstedc, Xstemr, Xstev, Xstevd, Xstevr, Xstevx,                          *
!    Xgeev, Xgeevx,                                                                  *
!    Xgebrd, Xorgbr, Xormbr, Xgesvd, Xgesdd, Xgesvdx,                                *
!    Xbdsqr, Xbdsdc, Xbdsvdx,                                                        *
!    Xgesv, Xsysv, Xposv,                                                            *
!    Xgelsd, Xgelss, Xgelsy, Xgels                                                   *
!                                                                                    *
! WHERE X CAN BE s, d, c AND z. THE GENERIC INTERFACES HAVE THE FORM:                *
!                                                                                    *
!     sytrd,  orgtr,  ormtr,  syev,  syevd,  syevr,  syevx,                          *
!     spev,  spevd,  spevx,  sygv,  sygvd,  sygvx,                                   *
!     steqr,  stedc,  stemr,  stev,  stevd,  stevr,  stevx,                          *
!     geev,  geevx,                                                                  *
!     gebrd,  orgbr,  ormbr,  gesvd,  gesdd,  gesvdx,                                *
!     bdsqr,  bdsdc,  bdsvdx,                                                        *
!     gesv,  sysv,  posv,                                                            *
!     gelsd,  gelss,  gelsy,  gels                                                   *
!                                                                                    * 
! LATEST REVISION : 21/03/2018                                                       *
!                                                                                    *
! ####################################################################################
!                                                                                    *
! ************************************************************************************
!
!
! PUBLIC ENTITIES 
! ===============
!    
! ALL SUBROUTINES, FUNCTIONS, VARIABLES AND PARAMETERS ARE PRIVATE BY DEFAULT.
!
    private
    public :: sytrd, orgtr, ormtr, syev, syevd, syevr, syevx,    &
              spev, spevd, spevx, sygv, sygvd, sygvx,            &
              steqr, stedc, stemr, stev, stevd, stevr, stevx,    &
              geev, geevx,                                       &
              gebrd, orgbr, ormbr, gesvd, gesdd, gesvdx,         &
              bdsqr, bdsdc, bdsvdx,                              &
              gesv, sysv, posv,                                  &
              gelsd, gelss, gelsy, gels
!
! GENERIC INTERFACES FOR LAPACK ROUTINES WITH OVERLOADED VERSIONS
! ===============================================================
!
!   -----------------------------------------------------------------------!
!   Tridiagonal reduction of a real symmetric or complex hermitian matrix  !
!   -----------------------------------------------------------------------!
!
!   Generic interface for SSYTRD, DSYTRD, CHETRD and ZHETRD subroutines in LAPACK
!
    interface sytrd
!
        subroutine ssytrd( UPLO, N, A, LDA, D, E, TAU, WORK, LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        UPLO
           INTEGER          INFO, LDA, LWORK, N
!     ..
!     .. Array Arguments ..
           REAL             A( LDA, * ), D( * ), E( * ), TAU( * ), WORK( * )
        end subroutine
!
        subroutine dsytrd( UPLO, N, A, LDA, D, E, TAU, WORK, LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        UPLO
           INTEGER          INFO, LDA, LWORK, N
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION A( LDA, * ), D( * ), E( * ), TAU( * ), WORK( * )
        end subroutine
!
        subroutine chetrd( UPLO, N, A, LDA, D, E, TAU, WORK, LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        UPLO
           INTEGER          INFO, LDA, LWORK, N
!     ..
!     .. Array Arguments ..
           REAL             D( * ), E( * )
           COMPLEX          A( LDA, * ), TAU( * ), WORK( * )
        end subroutine
!
        subroutine zhetrd( UPLO, N, A, LDA, D, E, TAU, WORK, LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        UPLO
           INTEGER          INFO, LDA, LWORK, N
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION D( * ), E( * )
           COMPLEX*16       A( LDA, * ), TAU( * ), WORK( * )
        end subroutine
!
    end interface
!
!   Generic interface for SORGTR, CORGTR, CUNGTR and ZUNGTR subroutines in LAPACK
!
    interface orgtr
!
        subroutine sorgtr( UPLO, N, A, LDA, TAU, WORK, LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        UPLO
           INTEGER          INFO, LDA, LWORK, N
!     ..
!     .. Array Arguments ..
           REAL             A( LDA, * ), TAU( * ), WORK( * )
        end subroutine
!
        subroutine dorgtr( UPLO, N, A, LDA, TAU, WORK, LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        UPLO
           INTEGER          INFO, LDA, LWORK, N
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION A( LDA, * ), TAU( * ), WORK( * )
        end subroutine
!
        subroutine cungtr( UPLO, N, A, LDA, TAU, WORK, LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        UPLO
           INTEGER          INFO, LDA, LWORK, N
!     ..
!     .. Array Arguments ..
           COMPLEX          A( LDA, * ), TAU( * ), WORK( * )
        end subroutine
!
        subroutine zungtr( UPLO, N, A, LDA, TAU, WORK, LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        UPLO
           INTEGER          INFO, LDA, LWORK, N
!     ..
!     .. Array Arguments ..
           COMPLEX*16       A( LDA, * ), TAU( * ), WORK( * )
        end subroutine
!
    end interface
!
!   Generic interface for SORMTR, CORMTR, CUNMTR and ZUNMTR subroutines in LAPACK
!
    interface ormtr
!
        subroutine sormtr( SIDE, UPLO, TRANS, M, N, A, LDA, TAU, C, LDC, WORK, LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        SIDE, UPLO, TRANS
           INTEGER          INFO, LDA, LDC, LWORK, M, N
!     ..
!     .. Array Arguments ..
           REAL             A( LDA, * ), C( LDC, * ), TAU( * ), WORK( * )
        end subroutine
!
        subroutine dormtr( SIDE, UPLO, TRANS, M, N, A, LDA, TAU, C, LDC, WORK, LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        SIDE, UPLO, TRANS
           INTEGER          INFO, LDA, LDC, LWORK, M, N
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION A( LDA, * ), C( LDC, * ), TAU( * ), WORK( * )
        end subroutine
!
        subroutine cunmtr( SIDE, UPLO, TRANS, M, N, A, LDA, TAU, C, LDC, WORK, LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        SIDE, UPLO, TRANS
           INTEGER          INFO, LDA, LDC, LWORK, M, N
!     ..
!     .. Array Arguments ..
           COMPLEX          A( LDA, * ), C( LDC, * ), TAU( * ), WORK( * )
        end subroutine
!
        subroutine zunmtr( SIDE, UPLO, TRANS, M, N, A, LDA, TAU, C, LDC, WORK, LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        SIDE, UPLO, TRANS
           INTEGER          INFO, LDA, LDC, LWORK, M, N
!     ..
!     .. Array Arguments ..
           COMPLEX*16       A( LDA, * ), C( LDC, * ), TAU( * ), WORK( * )
        end subroutine
!
    end interface
!
!   --------------------------------------------------------------------------!
!   Eigenvalues and eigenvectors decomposition of a real symmetric or complex !
!   hermitian matrix                                                          !
!   --------------------------------------------------------------------------!
!
!   Generic interface for SSYEV, DSYEV, CHEEV and ZHEEV  subroutines in LAPACK
!
    interface syev
!
        subroutine ssyev( JOBZ, UPLO, N, A, LDA, W, WORK, LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, UPLO
           INTEGER          INFO, LDA, LWORK, N
!     ..
!     .. Array Arguments ..
           REAL             A( LDA, * ), W( * ), WORK( * )
        end subroutine
!
        subroutine dsyev( JOBZ, UPLO, N, A, LDA, W, WORK, LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, UPLO
           INTEGER          INFO, LDA, LWORK, N
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION A( LDA, * ), W( * ), WORK( * )
        end subroutine
!
        subroutine cheev( JOBZ, UPLO, N, A, LDA, W, WORK, LWORK, RWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, UPLO
           INTEGER          INFO, LDA, LWORK, N
!     ..
!     .. Array Arguments ..
           COMPLEX          A( LDA, * ), WORK( * )
           REAL             W( * ), RWORK( * )
        end subroutine
!
        subroutine zheev( JOBZ, UPLO, N, A, LDA, W, WORK, LWORK, RWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, UPLO
           INTEGER          INFO, LDA, LWORK, N
!     ..
!     .. Array Arguments ..
           COMPLEX*16       A( LDA, * ), WORK( * )
           DOUBLE PRECISION W( * ), RWORK( * )
        end subroutine
!
    end interface
!
!   Generic interface for SSYEVD, DSYEVD, CHEEVD and ZHEEVD  subroutines in LAPACK
!
    interface syevd
!
        subroutine ssyevd( JOBZ, UPLO, N, A, LDA, W, WORK, LWORK, IWORK, LIWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, UPLO
           INTEGER          INFO, LDA, LIWORK, LWORK, N
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           REAL             A( LDA, * ), W( * ), WORK( * )
        end subroutine
!
        subroutine dsyevd( JOBZ, UPLO, N, A, LDA, W, WORK, LWORK, IWORK, LIWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, UPLO
           INTEGER          INFO, LDA, LIWORK, LWORK, N
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           DOUBLE PRECISION A( LDA, * ), W( * ), WORK( * )
        end subroutine
!
        subroutine cheevd( JOBZ, UPLO, N, A, LDA, W, WORK, LWORK, RWORK, LRWORK, IWORK, LIWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, UPLO
           INTEGER          INFO, LDA, LIWORK, LRWORK, LWORK, N
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           REAL             W( * ), RWORK( * )
           COMPLEX          A( LDA, * ), WORK( * )
        end subroutine
!
        subroutine zheevd( JOBZ, UPLO, N, A, LDA, W, WORK, LWORK, RWORK, LRWORK, IWORK, LIWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, UPLO
           INTEGER          INFO, LDA, LIWORK, LRWORK, LWORK, N
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           DOUBLE PRECISION W( * ), RWORK( * )
           COMPLEX*16       A( LDA, * ), WORK( * )
        end subroutine
!
    end interface
!
!   Generic interface for SSYEVR, DSYEVR, CHEEVR and ZHEEVR  subroutines in LAPACK
!
    interface syevr
!
        subroutine ssyevr( JOBZ, RANGE, UPLO, N, A, LDA, VL, VU, IL, IU,    &
                           ABSTOL, M, W, Z, LDZ, ISUPPZ, WORK, LWORK,       &
                           IWORK, LIWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, RANGE, UPLO
           INTEGER          INFO, IL, IU, LDA, LDZ, LIWORK, LWORK, M, N
           REAL             ABSTOL, VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER          ISUPPZ( * ), IWORK( * )
           REAL             A( LDA, * ), W( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
        subroutine dsyevr( JOBZ, RANGE, UPLO, N, A, LDA, VL, VU, IL, IU,    &
                           ABSTOL, M, W, Z, LDZ, ISUPPZ, WORK, LWORK,       &
                           IWORK, LIWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, RANGE, UPLO
           INTEGER          INFO, IL, IU, LDA, LDZ, LIWORK, LWORK, M, N
           DOUBLE PRECISION ABSTOL, VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER          ISUPPZ( * ), IWORK( * )
           DOUBLE PRECISION A( LDA, * ), W( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
        subroutine cheevr( JOBZ, RANGE, UPLO, N, A, LDA, VL, VU, IL, IU,    &
                           ABSTOL, M, W, Z, LDZ, ISUPPZ, WORK, LWORK,       &
                           RWORK, LRWORK, IWORK, LIWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, RANGE, UPLO
           INTEGER          INFO, IL, IU, LDA, LDZ, LRWORK, LIWORK, LWORK, M, N
           REAL             ABSTOL, VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER          ISUPPZ( * ), IWORK( * )
           REAL             W( * ), RWORK( * )
           COMPLEX          A( LDA, * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
        subroutine zheevr( JOBZ, RANGE, UPLO, N, A, LDA, VL, VU, IL, IU,    &
                           ABSTOL, M, W, Z, LDZ, ISUPPZ, WORK, LWORK,       &
                           RWORK, LRWORK, IWORK, LIWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, RANGE, UPLO
           INTEGER          INFO, IL, IU, LDA, LDZ, LRWORK, LIWORK, LWORK, M, N
           DOUBLE PRECISION ABSTOL, VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER          ISUPPZ( * ), IWORK( * )
           DOUBLE PRECISION W( * ), RWORK( * )
           COMPLEX*16       A( LDA, * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
    end interface
!
!   Generic interface for SSYEVX, DSYEVX, CHEEVX and ZHEEVX  subroutines in LAPACK
!
    interface syevx
!
        subroutine ssyevx( JOBZ, RANGE, UPLO, N, A, LDA, VL, VU, IL, IU,    &
                           ABSTOL, M, W, Z, LDZ, WORK, LWORK,               &
                           IWORK, IFAIL, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, RANGE, UPLO
           INTEGER          INFO, IL, IU, LDA, LDZ, LWORK, M, N
           REAL             ABSTOL, VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER          IFAIL( * ), IWORK( * )
           REAL             A( LDA, * ), W( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
        subroutine dsyevx( JOBZ, RANGE, UPLO, N, A, LDA, VL, VU, IL, IU,    &
                           ABSTOL, M, W, Z, LDZ, WORK, LWORK,               &
                           IWORK, IFAIL, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, RANGE, UPLO
           INTEGER          INFO, IL, IU, LDA, LDZ, LWORK, M, N
           DOUBLE PRECISION ABSTOL, VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER          IFAIL( * ), IWORK( * )
           DOUBLE PRECISION A( LDA, * ), W( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
        subroutine cheevx( JOBZ, RANGE, UPLO, N, A, LDA, VL, VU, IL, IU,    &
                           ABSTOL, M, W, Z, LDZ, WORK, LWORK, RWORK,        &
                           IWORK, IFAIL, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, RANGE, UPLO
           INTEGER          INFO, IL, IU, LDA, LDZ, LWORK, M, N
           REAL             ABSTOL, VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER          IFAIL( * ), IWORK( * )
           REAL             W( * ), RWORK( * )
           COMPLEX          A( LDA, * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
        subroutine zheevx( JOBZ, RANGE, UPLO, N, A, LDA, VL, VU, IL, IU,    &
                           ABSTOL, M, W, Z, LDZ, WORK, LWORK, RWORK,        &
                           IWORK, IFAIL, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, RANGE, UPLO
           INTEGER          INFO, IL, IU, LDA, LDZ, LWORK, M, N
           DOUBLE PRECISION ABSTOL, VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER          IFAIL( * ), IWORK( * )
           DOUBLE PRECISION W( * ), RWORK( * )
           COMPLEX*16       A( LDA, * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
    end interface
!
!   --------------------------------------------------------------------------!
!   Eigenvalues and eigenvectors decomposition of a real symmetric or complex !
!   hermitian matrix in packed storage                                        !
!   --------------------------------------------------------------------------!
!
!   Generic interface for SSPEV, DSPEV, CHPEV and ZHPEV subroutines in LAPACK
!
    interface spev
!
        subroutine sspev( JOBZ, UPLO, N, AP, W, Z, LDZ, WORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, UPLO
           INTEGER          INFO, LDZ, N
!     ..
!     .. Array Arguments ..
           REAL             AP( * ), W( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
        subroutine dspev( JOBZ, UPLO, N, AP, W, Z, LDZ, WORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, UPLO
           INTEGER          INFO, LDZ, N
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION AP( * ), W( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
        subroutine chpev( JOBZ, UPLO, N, AP, W, Z, LDZ, WORK, RWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, UPLO
           INTEGER          INFO, LDZ, N
!     ..
!     .. Array Arguments ..
           REAL             RWORK( * ), W( * )
           COMPLEX          AP( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
        subroutine zhpev( JOBZ, UPLO, N, AP, W, Z, LDZ, WORK, RWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, UPLO
           INTEGER          INFO, LDZ, N
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION RWORK( * ), W( * )
           COMPLEX*16       AP( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
    end interface
!
!   Generic interface for SSPEVD, DSPEVD, CHPEVD and ZHPEVD subroutines in LAPACK
!
    interface spevd
!
        subroutine sspevd( JOBZ, UPLO, N, AP, W, Z, LDZ, WORK, LWORK,  &
                           IWORK, LIWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, UPLO
           INTEGER          INFO, LDZ, LIWORK, LWORK, N
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           REAL             AP( * ), W( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
        subroutine dspevd( JOBZ, UPLO, N, AP, W, Z, LDZ, WORK, LWORK,  &
                           IWORK, LIWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, UPLO
           INTEGER          INFO, LDZ, LIWORK, LWORK, N
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           DOUBLE PRECISION AP( * ), W( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
        subroutine chpevd( JOBZ, UPLO, N, AP, W, Z, LDZ, WORK, LWORK,   &
                           RWORK, LRWORK, IWORK, LIWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, UPLO
           INTEGER          INFO, LDZ, LIWORK, LRWORK, LWORK, N
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           REAL             RWORK( * ), W( * )
           COMPLEX          AP( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
        subroutine zhpevd( JOBZ, UPLO, N, AP, W, Z, LDZ, WORK, LWORK,   &
                           RWORK, LRWORK, IWORK, LIWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, UPLO
           INTEGER          INFO, LDZ, LIWORK, LRWORK, LWORK, N
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           DOUBLE PRECISION RWORK( * ), W( * )
           COMPLEX*16       AP( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
    end interface
!
!   Generic interface for SSPEVX, DSPEVX, CHPEVX and ZHPEVX subroutines in LAPACK
!
    interface spevx
!
        subroutine sspevx( JOBZ, RANGE, UPLO, N, AP, VL, VU, IL, IU,    &
                           ABSTOL, M, W, Z, LDZ, WORK, IWORK, IFAIL,    &
                           INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, RANGE, UPLO
           INTEGER          IL, INFO, IU, LDZ, M, N
           REAL             ABSTOL, VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER          IFAIL( * ), IWORK( * )
           REAL             AP( * ), W( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
        subroutine dspevx( JOBZ, RANGE, UPLO, N, AP, VL, VU, IL, IU,    &
                           ABSTOL, M, W, Z, LDZ, WORK, IWORK, IFAIL,    &
                           INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, RANGE, UPLO
           INTEGER          IL, INFO, IU, LDZ, M, N
           DOUBLE PRECISION ABSTOL, VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER          IFAIL( * ), IWORK( * )
           DOUBLE PRECISION AP( * ), W( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
        subroutine chpevx( JOBZ, RANGE, UPLO, N, AP, VL, VU, IL, IU,    &
                           ABSTOL, M, W, Z, LDZ, WORK, RWORK, IWORK,    &
                           IFAIL, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, RANGE, UPLO
           INTEGER          IL, INFO, IU, LDZ, M, N
           REAL             ABSTOL, VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER          IFAIL( * ), IWORK( * )
           REAL             RWORK( * ), W( * )
           COMPLEX          AP( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
        subroutine zhpevx( JOBZ, RANGE, UPLO, N, AP, VL, VU, IL, IU,    &
                           ABSTOL, M, W, Z, LDZ, WORK, RWORK, IWORK,    &
                           IFAIL, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, RANGE, UPLO
           INTEGER          IL, INFO, IU, LDZ, M, N
           DOUBLE PRECISION ABSTOL, VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER          IFAIL( * ), IWORK( * )
           DOUBLE PRECISION RWORK( * ), W( * )
           COMPLEX*16       AP( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
    end interface
!
!   ---------------------------------------------------------------!
!   Eigenvalues and eigenvectors decomposition of a real symmetric !
!   tridiagonal matrix                                             !
!   ---------------------------------------------------------------!
!
!   Generic interface for SSTEQR, DSTEQR, CSTEQR and ZSTEQR subroutines in LAPACK
!
    interface steqr
!
        subroutine ssteqr( COMPZ, N, D, E, Z, LDZ, WORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        COMPZ
           INTEGER          INFO, LDZ, N
!     ..
!     .. Array Arguments ..
           REAL             D( * ), E( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
        subroutine dsteqr( COMPZ, N, D, E, Z, LDZ, WORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        COMPZ
           INTEGER          INFO, LDZ, N
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION D( * ), E( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
        subroutine csteqr( COMPZ, N, D, E, Z, LDZ, WORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        COMPZ
           INTEGER          INFO, LDZ, N
!     ..
!     .. Array Arguments ..
           REAL             D( * ), E( * ), WORK( * )
           COMPLEX          Z( LDZ, * )
        end subroutine
!
        subroutine zsteqr( COMPZ, N, D, E, Z, LDZ, WORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        COMPZ
           INTEGER          INFO, LDZ, N
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION D( * ), E( * ), WORK( * )
           COMPLEX*16       Z( LDZ, * )
        end subroutine
!
    end interface
!
!   Generic interface for SSTEDC, DSTEDC, CSTEDC and ZSTEDC subroutines in LAPACK
!
    interface stedc
!
        subroutine sstedc( COMPZ, N, D, E, Z, LDZ, WORK, LWORK, IWORK,    &
                           LIWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        COMPZ
           INTEGER          INFO, LDZ, LIWORK, LWORK, N
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           REAL             D( * ), E( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
        subroutine dstedc( COMPZ, N, D, E, Z, LDZ, WORK, LWORK, IWORK,    &
                           LIWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        COMPZ
           INTEGER          INFO, LDZ, LIWORK, LWORK, N
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           DOUBLE PRECISION D( * ), E( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
        subroutine cstedc( COMPZ, N, D, E, Z, LDZ, WORK, LWORK, RWORK,    &
                           LRWORK, IWORK, LIWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        COMPZ
           INTEGER          INFO, LDZ, LIWORK, LRWORK, LWORK, N
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           REAL             D( * ), E( * ), RWORK( * )
           COMPLEX          WORK( * ), Z( LDZ, * )
        end subroutine
!
        subroutine zstedc( COMPZ, N, D, E, Z, LDZ, WORK, LWORK, RWORK,    &
                           LRWORK, IWORK, LIWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        COMPZ
           INTEGER          INFO, LDZ, LIWORK, LRWORK, LWORK, N
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           DOUBLE PRECISION D( * ), E( * ), RWORK( * )
           COMPLEX*16       WORK( * ), Z( LDZ, * )
        end subroutine
!
    end interface
!
!   Generic interface for SSTEMR, DSTEMR, CSTEMR and ZSTEMR subroutines in LAPACK
!
    interface stemr
!
        subroutine sstemr( JOBZ, RANGE, N, D, E, VL, VU, IL, IU,            &
                           M, W, Z, LDZ, NZC, ISUPPZ, TRYRAC, WORK, LWORK,  &
                           IWORK, LIWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, RANGE
           LOGICAL          TRYRAC
           INTEGER          IL, INFO, IU, LDZ, NZC, LIWORK, LWORK, M, N
           REAL             VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER          ISUPPZ( * ), IWORK( * )
           REAL             D( * ), E( * ), W( * ), WORK( * )
           REAL             Z( LDZ, * )
        end subroutine
!
        subroutine dstemr( JOBZ, RANGE, N, D, E, VL, VU, IL, IU,            &
                           M, W, Z, LDZ, NZC, ISUPPZ, TRYRAC, WORK, LWORK,  &
                           IWORK, LIWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, RANGE
           LOGICAL          TRYRAC
           INTEGER          IL, INFO, IU, LDZ, NZC, LIWORK, LWORK, M, N
           DOUBLE PRECISION VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER          ISUPPZ( * ), IWORK( * )
           DOUBLE PRECISION D( * ), E( * ), W( * ), WORK( * )
           DOUBLE PRECISION Z( LDZ, * )
        end subroutine
!
        subroutine cstemr( JOBZ, RANGE, N, D, E, VL, VU, IL, IU,             &
                           M, W, Z, LDZ, NZC, ISUPPZ, TRYRAC, WORK, LWORK,   &
                           IWORK, LIWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, RANGE
           LOGICAL          TRYRAC
           INTEGER          IL, INFO, IU, LDZ, NZC, LIWORK, LWORK, M, N
           REAL             VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER          ISUPPZ( * ), IWORK( * )
           REAL             D( * ), E( * ), W( * ), WORK( * )
           COMPLEX          Z( LDZ, * )
        end subroutine
!
        subroutine zstemr( JOBZ, RANGE, N, D, E, VL, VU, IL, IU,             &
                           M, W, Z, LDZ, NZC, ISUPPZ, TRYRAC, WORK, LWORK,   &
                           IWORK, LIWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, RANGE
           LOGICAL          TRYRAC
           INTEGER          IL, INFO, IU, LDZ, NZC, LIWORK, LWORK, M, N
           DOUBLE PRECISION VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER          ISUPPZ( * ), IWORK( * )
           DOUBLE PRECISION D( * ), E( * ), W( * ), WORK( * )
           COMPLEX*16       Z( LDZ, * )
        end subroutine
!
    end interface
!
!   Generic interface for SSTEV and DSTEV subroutines in LAPACK
!
    interface stev
!
        subroutine sstev( JOBZ, N, D, E, Z, LDZ, WORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ
           INTEGER          INFO, LDZ, N
!     ..
!     .. Array Arguments ..
           REAL             D( * ), E( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
        subroutine dstev( JOBZ, N, D, E, Z, LDZ, WORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ
           INTEGER          INFO, LDZ, N
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION D( * ), E( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
    end interface
!
!   Generic interface for SSTEVD and DSTEVD subroutines in LAPACK
!
    interface stevd
!
        subroutine sstevd( JOBZ, N, D, E, Z, LDZ, WORK, LWORK, IWORK,   &
                           LIWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ
           INTEGER          INFO, LDZ, LIWORK, LWORK, N
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           REAL             D( * ), E( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
        subroutine dstevd( JOBZ, N, D, E, Z, LDZ, WORK, LWORK, IWORK,   &
                           LIWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ
           INTEGER          INFO, LDZ, LIWORK, LWORK, N
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           DOUBLE PRECISION D( * ), E( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
    end interface
!
!   Generic interface for SSTEVR and DSTEVR subroutines in LAPACK
!
    interface stevr
!
        subroutine sstevr( JOBZ, RANGE, N, D, E, VL, VU, IL, IU, ABSTOL,    &
                           M, W, Z, LDZ, ISUPPZ, WORK, LWORK, IWORK,        &
                           LIWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, RANGE
           INTEGER          IL, INFO, IU, LDZ, LIWORK, LWORK, M, N
           REAL             ABSTOL, VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER          ISUPPZ( * ), IWORK( * )
           REAL             D( * ), E( * ), W( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
        subroutine dstevr( JOBZ, RANGE, N, D, E, VL, VU, IL, IU, ABSTOL,    &
                           M, W, Z, LDZ, ISUPPZ, WORK, LWORK, IWORK,        &
                           LIWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, RANGE
           INTEGER          IL, INFO, IU, LDZ, LIWORK, LWORK, M, N
           DOUBLE PRECISION ABSTOL, VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER          ISUPPZ( * ), IWORK( * )
           DOUBLE PRECISION D( * ), E( * ), W( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
    end interface
!
!   Generic interface for SSTEVX and DSTEVX subroutines in LAPACK
!
    interface stevx
!
        subroutine sstevx( JOBZ, RANGE, N, D, E, VL, VU, IL, IU, ABSTOL,   &
                           M, W, Z, LDZ, WORK, IWORK, IFAIL, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, RANGE
           INTEGER          IL, INFO, IU, LDZ, M, N
           REAL             ABSTOL, VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER          IFAIL( * ), IWORK( * )
           REAL             D( * ), E( * ), W( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
        subroutine dstevx( JOBZ, RANGE, N, D, E, VL, VU, IL, IU, ABSTOL,   &
                           M, W, Z, LDZ, WORK, IWORK, IFAIL, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, RANGE
           INTEGER          IL, INFO, IU, LDZ, M, N
           DOUBLE PRECISION ABSTOL, VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER          IFAIL( * ), IWORK( * )
           DOUBLE PRECISION D( * ), E( * ), W( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
    end interface
!
!   --------------------------------------------------------------------------!
!   Eigenvalues and eigenvectors of a real generalized symmetric-definite or  !
!   complex generalized hermitian-definite problem                            !
!   --------------------------------------------------------------------------!
!
!   Generic interface for SSYGV, DSYGV, CHEGV and ZHEGV  subroutines in LAPACK
!
    interface sygv
!
        subroutine ssygv( ITYPE, JOBZ, UPLO, N, A, LDA, B, LDB, W, WORK, &
                          LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, UPLO
           INTEGER          INFO, ITYPE, LDA, LDB, LWORK, N
!     ..
!     .. Array Arguments ..
           REAL             A( LDA, * ), B( LDB, * ), W( * ), WORK( * )
        end subroutine
!
        subroutine dsygv( ITYPE, JOBZ, UPLO, N, A, LDA, B, LDB, W, WORK, &
                          LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, UPLO
           INTEGER          INFO, ITYPE, LDA, LDB, LWORK, N
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION A( LDA, * ), B( LDB, * ), W( * ), WORK( * )
        end subroutine
!
        subroutine chegv( ITYPE, JOBZ, UPLO, N, A, LDA, B, LDB, W, WORK,   &
                          LWORK, RWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, UPLO
           INTEGER          INFO, ITYPE, LDA, LDB, LWORK, N
!     ..
!     .. Array Arguments ..
           REAL             W( * ), RWORK( * )
           COMPLEX          A( LDA, * ), B( LDB, * ), WORK( * )
        end subroutine
!
        subroutine zhegv( ITYPE, JOBZ, UPLO, N, A, LDA, B, LDB, W, WORK,   &
                          LWORK, RWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, UPLO
           INTEGER          INFO, ITYPE, LDA, LDB, LWORK, N
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION W( * ), RWORK( * )
           COMPLEX*16       A( LDA, * ), B( LDB, * ), WORK( * )
        end subroutine
!
    end interface
!
!   Generic interface for SSYGVD, DSYGVD, CHEGVD and ZHEGVD  subroutines in LAPACK
!
    interface sygvd
!
        subroutine ssygvd( ITYPE, JOBZ, UPLO, N, A, LDA, B, LDB, W, WORK, &
                           LWORK, IWORK, LIWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, UPLO
           INTEGER          INFO, ITYPE, LDA, LDB, LIWORK, LWORK, N
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           REAL             A( LDA, * ), B( LDB, * ), W( * ), WORK( * )
        end subroutine
!
        subroutine dsygvd( ITYPE, JOBZ, UPLO, N, A, LDA, B, LDB, W, WORK, &
                           LWORK, IWORK, LIWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, UPLO
           INTEGER          INFO, ITYPE, LDA, LDB, LIWORK, LWORK, N
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           DOUBLE PRECISION A( LDA, * ), B( LDB, * ), W( * ), WORK( * )
        end subroutine
!
        subroutine chegvd( ITYPE, JOBZ, UPLO, N, A, LDA, B, LDB, W, WORK,   &
                           LWORK, RWORK, LRWORK, IWORK, LIWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, UPLO
           INTEGER          INFO, ITYPE, LDA, LDB, LIWORK, LRWORK, LWORK, N
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           REAL             W( * ), RWORK( * )
           COMPLEX          A( LDA, * ), B( LDB, * ), WORK( * )
        end subroutine
!
        subroutine zhegvd( ITYPE, JOBZ, UPLO, N, A, LDA, B, LDB, W, WORK,   &
                           LWORK, RWORK, LRWORK, IWORK, LIWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, UPLO
           INTEGER          INFO, ITYPE, LDA, LDB, LIWORK, LRWORK, LWORK, N
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           DOUBLE PRECISION W( * ), RWORK( * )
           COMPLEX*16       A( LDA, * ), B( LDB, * ), WORK( * )
        end subroutine
!
    end interface
!
!   Generic interface for SSYGVX, DSYGVX, CHEGVX and ZHEGVX  subroutines in LAPACK
!
    interface sygvx
!
        subroutine ssygvx( ITYPE, JOBZ, RANGE, UPLO, N, A, LDA, B, LDB,  &
                           VL, VU, IL, IU, ABSTOL, M, W, Z, LDZ, WORK,   &
                           LWORK, IWORK, IFAIL, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, RANGE, UPLO
           INTEGER          IL, INFO, ITYPE, IU, LDA, LDB, LDZ, LWORK, M, N
           REAL             ABSTOL, VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER          IFAIL( * ), IWORK( * )
           REAL             A( LDA, * ), B( LDB, * ), W( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
        subroutine dsygvx( ITYPE, JOBZ, RANGE, UPLO, N, A, LDA, B, LDB,  &
                           VL, VU, IL, IU, ABSTOL, M, W, Z, LDZ, WORK,   &
                           LWORK, IWORK, IFAIL, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, RANGE, UPLO
           INTEGER          IL, INFO, ITYPE, IU, LDA, LDB, LDZ, LWORK, M, N
           DOUBLE PRECISION ABSTOL, VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER          IFAIL( * ), IWORK( * )
           DOUBLE PRECISION A( LDA, * ), B( LDB, * ), W( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
        subroutine chegvx( ITYPE, JOBZ, RANGE, UPLO, N, A, LDA, B, LDB,  &
                           VL, VU, IL, IU, ABSTOL, M, W, Z, LDZ, WORK,   &
                           LWORK, RWORK, IWORK, IFAIL, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, RANGE, UPLO
           INTEGER          IL, INFO, ITYPE, IU, LDA, LDB, LDZ, LWORK, M, N
           REAL             ABSTOL, VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER          IFAIL( * ), IWORK( * )
           REAL             W( * ), RWORK( * )
           COMPLEX          A( LDA, * ), B( LDB, * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
        subroutine zhegvx( ITYPE, JOBZ, RANGE, UPLO, N, A, LDA, B, LDB,  &
                           VL, VU, IL, IU, ABSTOL, M, W, Z, LDZ, WORK,   &
                           LWORK, RWORK, IWORK, IFAIL, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ, RANGE, UPLO
           INTEGER          IL, INFO, ITYPE, IU, LDA, LDB, LDZ, LWORK, M, N
           DOUBLE PRECISION ABSTOL, VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER          IFAIL( * ), IWORK( * )
           DOUBLE PRECISION W( * ), RWORK( * )
           COMPLEX*16       A( LDA, * ), B( LDB, * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
    end interface
!
!   -----------------------------------------------------------------!
!   Eigenvalues and eigenvectors of a real or complex general matrix !
!   -----------------------------------------------------------------!
!
!   Generic interface for SGEEV, DGEEV, CGEEV and ZGEEV  subroutines in LAPACK
!
    interface geev
!
        subroutine sgeev( JOBVL, JOBVR, N, A, LDA, WR, WI, VL, LDVL, VR, & 
                          LDVR, WORK, LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBVL, JOBVR
           INTEGER          INFO, LDA, LDVL, LDVR, LWORK, N
!     ..
!     .. Array Arguments ..
           REAL             A( LDA, * ), VL( LDVL, * ), VR( LDVR, * ),  &
                            WI( * ), WORK( * ), WR( * )
        end subroutine
!
        subroutine dgeev( JOBVL, JOBVR, N, A, LDA, WR, WI, VL, LDVL, VR, & 
                          LDVR, WORK, LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBVL, JOBVR
           INTEGER          INFO, LDA, LDVL, LDVR, LWORK, N
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION A( LDA, * ), VL( LDVL, * ), VR( LDVR, * ),  &
                            WI( * ), WORK( * ), WR( * )
        end subroutine
!
        subroutine cgeev( JOBVL, JOBVR, N, A, LDA, W, VL, LDVL, VR, LDVR, & 
                          WORK, LWORK, RWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBVL, JOBVR
           INTEGER          INFO, LDA, LDVL, LDVR, LWORK, N
!     ..
!     .. Array Arguments ..
           REAL             RWORK( * )
           COMPLEX          A( LDA, * ), VL( LDVL, * ), VR( LDVR, * ),  &
                            W( * ), WORK( * )
        end subroutine
!
        subroutine zgeev( JOBVL, JOBVR, N, A, LDA, W, VL, LDVL, VR, LDVR, & 
                          WORK, LWORK, RWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBVL, JOBVR
           INTEGER          INFO, LDA, LDVL, LDVR, LWORK, N
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION RWORK( * )
           COMPLEX*16       A( LDA, * ), VL( LDVL, * ), VR( LDVR, * ),  &
                            W( * ), WORK( * )
        end subroutine
!
    end interface
!
!   Generic interface for SGEEVX, DGEEVX, CGEEVX and ZGEEVX  subroutines in LAPACK
!
    interface geevx
!
        subroutine sgeevx( BALANC, JOBVL, JOBVR, SENSE, N, A, LDA, WR, WI,   &
                           VL, LDVL, VR, LDVR, ILO, IHI, SCALE, ABNRM, &
                           RCONDE, RCONDV, WORK, LWORK, IWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER          BALANC, JOBVL, JOBVR, SENSE
           INTEGER            IHI, ILO, INFO, LDA, LDVL, LDVR, LWORK, N
           REAL               ABNRM
!     ..
!     .. Array Arguments ..
           INTEGER            IWORK( * )
           REAL               A( LDA, * ), RCONDE( * ), RCONDV( * ),     &
                              SCALE( * ), VL( LDVL, * ), VR( LDVR, * ),  &
                              WI( * ), WORK( * ), WR( * )
        end subroutine
!
        subroutine dgeevx( BALANC, JOBVL, JOBVR, SENSE, N, A, LDA, WR, WI,   &
                           VL, LDVL, VR, LDVR, ILO, IHI, SCALE, ABNRM, &
                           RCONDE, RCONDV, WORK, LWORK, IWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER          BALANC, JOBVL, JOBVR, SENSE
           INTEGER            IHI, ILO, INFO, LDA, LDVL, LDVR, LWORK, N
           DOUBLE PRECISION   ABNRM
!     ..
!     .. Array Arguments ..
           INTEGER            IWORK( * )
           DOUBLE PRECISION   A( LDA, * ), RCONDE( * ), RCONDV( * ),     &
                              SCALE( * ), VL( LDVL, * ), VR( LDVR, * ),  &
                              WI( * ), WORK( * ), WR( * )
        end subroutine
!
        subroutine cgeevx( BALANC, JOBVL, JOBVR, SENSE, N, A, LDA, W, VL,  &
                           LDVL, VR, LDVR, ILO, IHI, SCALE, ABNRM, RCONDE, &
                           RCONDV, WORK, LWORK, RWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER          BALANC, JOBVL, JOBVR, SENSE
           INTEGER            IHI, ILO, INFO, LDA, LDVL, LDVR, LWORK, N
           REAL               ABNRM
!     ..
!     .. Array Arguments ..
           REAL               RCONDE( * ), RCONDV( * ), RWORK( * ),      &
                              SCALE( * )
           COMPLEX            A( LDA, * ), VL( LDVL, * ), VR( LDVR, * ), &
                              W( * ), WORK( * )
        end subroutine
!
        subroutine zgeevx( BALANC, JOBVL, JOBVR, SENSE, N, A, LDA, W, VL,  &
                           LDVL, VR, LDVR, ILO, IHI, SCALE, ABNRM, RCONDE, &
                           RCONDV, WORK, LWORK, RWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER          BALANC, JOBVL, JOBVR, SENSE
           INTEGER            IHI, ILO, INFO, LDA, LDVL, LDVR, LWORK, N
           DOUBLE PRECISION   ABNRM
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION   RCONDE( * ), RCONDV( * ), RWORK( * ),      &
                              SCALE( * )
           COMPLEX*16         A( LDA, * ), VL( LDVL, * ), VR( LDVR, * ), &
                              W( * ), WORK( * )
        end subroutine
!
    end interface
!
!   ----------------------------------------------------------!
!   Bidiagonal reduction of a real or complex general matrix  !
!   ----------------------------------------------------------!
!
!   Generic interface for SGEBRD, DGEBRD, CGEBRD and ZGEBRD  subroutines in LAPACK
!
    interface gebrd
!
        subroutine sgebrd( M, N, A, LDA, D, E, TAUQ, TAUP, WORK, LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INFO, LDA, LWORK, M, N
!     ..
!     .. Array Arguments ..
           REAL             A( LDA, * ), D( * ), E( * ), TAUQ( * ), TAUP( * ), WORK( * )
        end subroutine
!
        subroutine dgebrd( M, N, A, LDA, D, E, TAUQ, TAUP, WORK, LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INFO, LDA, LWORK, M, N
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION A( LDA, * ), D( * ), E( * ), TAUQ( * ), TAUP( * ), WORK( * )
        end subroutine
!
        subroutine cgebrd( M, N, A, LDA, D, E, TAUQ, TAUP, WORK, LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INFO, LDA, LWORK, M, N
!     ..
!     .. Array Arguments ..
           REAL             D( * ), E( * )
           COMPLEX          A( LDA, * ), TAUQ( * ), TAUP( * ), WORK( * )
        end subroutine
!
        subroutine zgebrd( M, N, A, LDA, D, E, TAUQ, TAUP, WORK, LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           INTEGER          INFO, LDA, LWORK, M, N
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION D( * ), E( * )
           COMPLEX*16       A( LDA, * ), TAUQ( * ), TAUP( * ), WORK( * )
        end subroutine
!
    end interface
!
!   Generic interface for SORGBR, DORGBR, CUNGBR and ZUNGBR subroutines in LAPACK
!
    interface orgbr
!
        subroutine sorgbr( VECT, M, N, K, A, LDA, TAU, WORK, LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        VECT
           INTEGER          INFO, LDA, LWORK, M, N, K
!     ..
!     .. Array Arguments ..
           REAL             A( LDA, * ), TAU( * ), WORK( * )
        end subroutine
!
        subroutine dorgbr( VECT, M, N, K, A, LDA, TAU, WORK, LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        VECT
           INTEGER          INFO, LDA, LWORK, M, N, K
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION A( LDA, * ), TAU( * ), WORK( * )
        end subroutine
!
        subroutine cungbr( VECT, M, N, K, A, LDA, TAU, WORK, LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        VECT
           INTEGER          INFO, LDA, LWORK, M, N, K
!     ..
!     .. Array Arguments ..
           COMPLEX          A( LDA, * ), TAU( * ), WORK( * )
        end subroutine
!
        subroutine zungbr( VECT, M, N, K, A, LDA, TAU, WORK, LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        VECT
           INTEGER          INFO, LDA, LWORK, M, N, K
!     ..
!     .. Array Arguments ..
           COMPLEX*16       A( LDA, * ), TAU( * ), WORK( * )
        end subroutine
!
    end interface
!
!   Generic interface for SORMBR, CORMBR, CUNMBR and ZUNMBR subroutines in LAPACK
!
    interface ormbr
!
        subroutine sormbr( VECT, SIDE, TRANS, M, N, K, A, LDA, TAU, C, LDC, WORK, LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        VECT, SIDE, TRANS
           INTEGER          INFO, LDA, LDC, LWORK, M, N, K
!     ..
!     .. Array Arguments ..
           REAL             A( LDA, * ), C( LDC, * ), TAU( * ), WORK( * )
        end subroutine
!
        subroutine dormbr( VECT, SIDE, TRANS, M, N, K, A, LDA, TAU, C, LDC, WORK, LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        VECT, SIDE, TRANS
           INTEGER          INFO, LDA, LDC, LWORK, M, N, K
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION A( LDA, * ), C( LDC, * ), TAU( * ), WORK( * )
        end subroutine
!
        subroutine cunmbr( VECT, SIDE, TRANS, M, N, K, A, LDA, TAU, C, LDC, WORK, LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        VECT, SIDE, TRANS
           INTEGER          INFO, LDA, LDC, LWORK, M, N, K
!     ..
!     .. Array Arguments ..
           COMPLEX          A( LDA, * ), C( LDC, * ), TAU( * ), WORK( * )
        end subroutine
!
        subroutine zunmbr( VECT, SIDE, TRANS, M, N, K, A, LDA, TAU, C, LDC, WORK, LWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER        VECT, SIDE, TRANS
           INTEGER          INFO, LDA, LDC, LWORK, M, N, K
!     ..
!     .. Array Arguments ..
           COMPLEX*16       A( LDA, * ), C( LDC, * ), TAU( * ), WORK( * )
        end subroutine
!
    end interface
!
!   ------------------------------------------------------------------!
!   Singular value decomposition of a real or complex general matrix  !
!   ------------------------------------------------------------------!
!
!   Generic interface for SGESVD, DGESVD, CGESVD and ZGESVD subroutines in LAPACK
!
    interface gesvd
!
        subroutine sgesvd( JOBU, JOBVT, M, N, A, LDA, S, U, LDU, VT, LDVT, WORK, LWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBU, JOBVT
           INTEGER          INFO, LDA, LDU, LDVT, LWORK, M, N
!     ..
!     .. Array Arguments ..
           REAL             A( LDA, * ), S( * ), U( LDU, * ), VT( LDVT, * ), WORK( * )
        end subroutine
!
        subroutine dgesvd( JOBU, JOBVT, M, N, A, LDA, S, U, LDU, VT, LDVT, WORK, LWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBU, JOBVT
           INTEGER          INFO, LDA, LDU, LDVT, LWORK, M, N
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION A( LDA, * ), S( * ), U( LDU, * ), VT( LDVT, * ), WORK( * )
        end subroutine
!
        subroutine cgesvd( JOBU, JOBVT, M, N, A, LDA, S, U, LDU, VT, LDVT, WORK, LWORK, RWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBU, JOBVT
           INTEGER          INFO, LDA, LDU, LDVT, LWORK, M, N
!     ..
!     .. Array Arguments ..
           COMPLEX          A( LDA, * ), U( LDU, * ), VT( LDVT, * ), WORK( * )
           REAL             S( * ), RWORK( * )
        end subroutine
!
        subroutine zgesvd( JOBU, JOBVT, M, N, A, LDA, S, U, LDU, VT, LDVT, WORK, LWORK, RWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBU, JOBVT
           INTEGER          INFO, LDA, LDU, LDVT, LWORK, M, N
!     ..
!     .. Array Arguments ..
           COMPLEX*16       A( LDA, * ), U( LDU, * ), VT( LDVT, * ), WORK( * )
           DOUBLE PRECISION S( * ), RWORK( * )
        end subroutine
!
    end interface
!
!   Generic interface for SGESDD, DGESDD, CGESDD and ZGESDD subroutines in LAPACK
!
    interface gesdd
!
        subroutine sgesdd( JOBZ, M, N, A, LDA, S, U, LDU, VT, LDVT, WORK, LWORK, IWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ
           INTEGER          INFO, LDA, LDU, LDVT, LWORK, M, N
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           REAL             A( LDA, * ), S( * ), U( LDU, * ), VT( LDVT, * ), WORK( * )
        end subroutine
!
        subroutine dgesdd( JOBZ, M, N, A, LDA, S, U, LDU, VT, LDVT, WORK, LWORK, IWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ
           INTEGER          INFO, LDA, LDU, LDVT, LWORK, M, N
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           DOUBLE PRECISION A( LDA, * ), S( * ), U( LDU, * ), VT( LDVT, * ), WORK( * )
        end subroutine
!
        subroutine cgesdd( JOBZ, M, N, A, LDA, S, U, LDU, VT, LDVT, WORK, LWORK, RWORK, IWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ
           INTEGER          INFO, LDA, LDU, LDVT, LWORK, M, N
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           COMPLEX          A( LDA, * ), U( LDU, * ), VT( LDVT, * ), WORK( * )
           REAL             S( * ), RWORK( * )
        end subroutine
!
        subroutine zgesdd( JOBZ, M, N, A, LDA, S, U, LDU, VT, LDVT, WORK, LWORK, RWORK, IWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBZ
           INTEGER          INFO, LDA, LDU, LDVT, LWORK, M, N
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           COMPLEX*16       A( LDA, * ), U( LDU, * ), VT( LDVT, * ), WORK( * )
           DOUBLE PRECISION S( * ), RWORK( * )
        end subroutine
!
    end interface
!
!   Generic interface for SGESVDX, DGESVDX, CGESVDX and ZGESVDX subroutines in LAPACK
!
    interface gesvdx
!
        subroutine sgesvdx( JOBU, JOBVT, RANGE, M, N, A, LDA, VL, VU, IL, IU,    &
                            NS, S, U, LDU, VT, LDVT, WORK, LWORK, IWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBU, JOBVT, RANGE
           INTEGER          INFO, IL, IU, LDA, LDU, LDVT, LWORK, M, N, NS
           REAL             VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           REAL             A( LDA, * ), S( * ), U( LDU, * ), VT( LDVT, * ), WORK( * )
        end subroutine
!
        subroutine dgesvdx( JOBU, JOBVT, RANGE, M, N, A, LDA, VL, VU, IL, IU,    &
                            NS, S, U, LDU, VT, LDVT, WORK, LWORK, IWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBU, JOBVT, RANGE
           INTEGER          INFO, IL, IU, LDA, LDU, LDVT, LWORK, M, N, NS
           DOUBLE PRECISION VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           DOUBLE PRECISION A( LDA, * ), S( * ), U( LDU, * ), VT( LDVT, * ), WORK( * )
        end subroutine
!
        subroutine cgesvdx( JOBU, JOBVT, RANGE, M, N, A, LDA, VL, VU, IL, IU,       &
                            NS, S, U, LDU, VT, LDVT, WORK, LWORK, RWORK, IWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBU, JOBVT, RANGE
           INTEGER          INFO, IL, IU, LDA, LDU, LDVT, LWORK, M, N, NS
           REAL             VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           REAL             S( * ), RWORK( * )
           COMPLEX          A( LDA, * ), U( LDU, * ), VT( LDVT, * ), WORK( * )
        end subroutine
!
        subroutine zgesvdx( JOBU, JOBVT, RANGE, M, N, A, LDA, VL, VU, IL, IU,       &
                            NS, S, U, LDU, VT, LDVT, WORK, LWORK, RWORK, IWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           CHARACTER        JOBU, JOBVT, RANGE
           INTEGER          INFO, IL, IU, LDA, LDU, LDVT, LWORK, M, N, NS
           DOUBLE PRECISION VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           DOUBLE PRECISION S( * ), RWORK( * )
           COMPLEX*16       A( LDA, * ), U( LDU, * ), VT( LDVT, * ), WORK( * )
        end subroutine
!
    end interface
!
!   ---------------------------------------------------------!
!   Singular value decomposition of a real bidiagonal matrix !
!   ---------------------------------------------------------!
!
!   Generic interface for SBDSQR, DBDSQR, CBDSQR and ZBDSQR subroutines in LAPACK
!
    interface bdsqr
!
        subroutine sbdsqr( UPLO, N, NCVT, NRU, NCC, D, E, VT, LDVT, U,   &
                           LDU, C, LDC, WORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER          UPLO
           INTEGER            INFO, LDC, LDU, LDVT, N, NCC, NCVT, NRU
!     ..
!     .. Array Arguments ..
           REAL               C( LDC, * ), D( * ), E( * ), U( LDU, * ),   &
                              VT( LDVT, * ), WORK( * )
        end subroutine
!
        subroutine dbdsqr( UPLO, N, NCVT, NRU, NCC, D, E, VT, LDVT, U,   &
                           LDU, C, LDC, WORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER          UPLO
           INTEGER            INFO, LDC, LDU, LDVT, N, NCC, NCVT, NRU
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION   C( LDC, * ), D( * ), E( * ), U( LDU, * ),   &
                              VT( LDVT, * ), WORK( * )
        end subroutine
!
        subroutine cbdsqr( UPLO, N, NCVT, NRU, NCC, D, E, VT, LDVT, U,   &
                           LDU, C, LDC, RWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER          UPLO
           INTEGER            INFO, LDC, LDU, LDVT, N, NCC, NCVT, NRU
!     ..
!     .. Array Arguments ..
           REAL               D( * ), E( * ), RWORK( * )
           COMPLEX            C( LDC, * ), U( LDU, * ), VT( LDVT, * )
        end subroutine
!
        subroutine zbdsqr( UPLO, N, NCVT, NRU, NCC, D, E, VT, LDVT, U,   &
                           LDU, C, LDC, RWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER          UPLO
           INTEGER            INFO, LDC, LDU, LDVT, N, NCC, NCVT, NRU
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION   D( * ), E( * ), RWORK( * )
           COMPLEX*16         C( LDC, * ), U( LDU, * ), VT( LDVT, * )
        end subroutine
!
    end interface
!
!   Generic interface for SBDSDC and DBDSDC subroutines in LAPACK
!
    interface bdsdc
!
        subroutine sbdsdc( UPLO, COMPQ, N, D, E, U, LDU, VT, LDVT, Q, IQ,   &
                           WORK, IWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER          COMPQ, UPLO
           INTEGER            INFO, LDU, LDVT, N
!     ..
!     .. Array Arguments ..
           INTEGER            IQ( * ), IWORK( * )
           REAL               D( * ), E( * ), Q( * ), U( LDU, * ),   &
                              VT( LDVT, * ), WORK( * )
        end subroutine
!
        subroutine dbdsdc( UPLO, COMPQ, N, D, E, U, LDU, VT, LDVT, Q, IQ,   &
                           WORK, IWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER          COMPQ, UPLO
           INTEGER            INFO, LDU, LDVT, N
!     ..
!     .. Array Arguments ..
           INTEGER            IQ( * ), IWORK( * )
           DOUBLE PRECISION   D( * ), E( * ), Q( * ), U( LDU, * ),   &
                              VT( LDVT, * ), WORK( * )
        end subroutine
!
    end interface
!
!   Generic interface for SBDSVDX and DBDSVDX subroutines in LAPACK
!
    interface bdsvdx
!
        subroutine sbdsvdx( UPLO, JOBZ, RANGE, N, D, E, VL, VU, IL, IU,   &
                            NS, S, Z, LDZ, WORK, IWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER          JOBZ, RANGE, UPLO
           INTEGER            IL, INFO, IU, LDZ, N, NS
           REAL               VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER            IWORK( * )
           REAL               D( * ), E( * ), S( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
        subroutine dbdsvdx( UPLO, JOBZ, RANGE, N, D, E, VL, VU, IL, IU,   &
                            NS, S, Z, LDZ, WORK, IWORK, INFO )
!     ..
!     .. Scalar Arguments ..
           CHARACTER          JOBZ, RANGE, UPLO
           INTEGER            IL, INFO, IU, LDZ, N, NS
           DOUBLE PRECISION   VL, VU
!     ..
!     .. Array Arguments ..
           INTEGER            IWORK( * )
           DOUBLE PRECISION   D( * ), E( * ), S( * ), WORK( * ), Z( LDZ, * )
        end subroutine
!
    end interface
!
!   --------------------------------------------------------------------------!
!   Solution of a real or complex system of linear equations with a general   !
!   matrix and several right hand side vectors                                !
!   --------------------------------------------------------------------------!
!
!   Generic interface for SGESV, DGESV, CGESV and ZGESV subroutines in LAPACK
!
    interface gesv
!
        subroutine sgesv( N, NRHS, A, LDA, IPIV, B, LDB, INFO)
!     ..
!     .. Scalar Arguments ..
           INTEGER            INFO, LDA, LDB, N, NRHS
!     ..
!     .. Array Arguments ..
           INTEGER            IPIV( * )
           REAL               A( LDA, * ), B( LDB, * )
        end subroutine
!
        subroutine dgesv( N, NRHS, A, LDA, IPIV, B, LDB, INFO)
!     ..
!     .. Scalar Arguments ..
           INTEGER            INFO, LDA, LDB, N, NRHS
!     ..
!     .. Array Arguments ..
           INTEGER            IPIV( * )
           DOUBLE PRECISION   A( LDA, * ), B( LDB, * )
        end subroutine
!
        subroutine cgesv( N, NRHS, A, LDA, IPIV, B, LDB, INFO)
!     ..
!     .. Scalar Arguments ..
           INTEGER            INFO, LDA, LDB, N, NRHS
!     ..
!     .. Array Arguments ..
           INTEGER            IPIV( * )
           COMPLEX            A( LDA, * ), B( LDB, * )
        end subroutine
!
        subroutine zgesv( N, NRHS, A, LDA, IPIV, B, LDB, INFO)
!     ..
!     .. Scalar Arguments ..
           INTEGER            INFO, LDA, LDB, N, NRHS
!     ..
!     .. Array Arguments ..
           INTEGER            IPIV( * )
           COMPLEX*16         A( LDA, * ), B( LDB, * )
        end subroutine
!
    end interface
!
!   --------------------------------------------------------------------------!
!   Solution of a real or complex system of linear equations with a symmetric !
!   matrix and several right hand side vectors                                !
!   --------------------------------------------------------------------------!
!
!   Generic interface for SSYSV, DSYSV, CSYSV and ZSYSV subroutines in LAPACK
!
    interface sysv
!
        subroutine ssysv( UPLO, N, NRHS, A, LDA, IPIV, B, LDB, WORK, LWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           CHARACTER        UPLO
           INTEGER          INFO, LDA, LDB, N, LWORK, NRHS
!     ..
!     .. Array Arguments ..
           INTEGER          IPIV( * )
           REAL             A( LDA, * ), B( LDB, * ), WORK( * )
        end subroutine
!
        subroutine dsysv( UPLO, N, NRHS, A, LDA, IPIV, B, LDB, WORK, LWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           CHARACTER        UPLO
           INTEGER          INFO, LDA, LDB, N, LWORK, NRHS
!     ..
!     .. Array Arguments ..
           INTEGER          IPIV( * )
           DOUBLE PRECISION A( LDA, * ), B( LDB, * ), WORK( * )
        end subroutine
!
        subroutine csysv( UPLO, N, NRHS, A, LDA, IPIV, B, LDB, WORK, LWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           CHARACTER        UPLO
           INTEGER          INFO, LDA, LDB, N, LWORK, NRHS
!     ..
!     .. Array Arguments ..
           INTEGER          IPIV( * )
           COMPLEX          A( LDA, * ), B( LDB, * ), WORK( * )
        end subroutine
!
        subroutine zsysv( UPLO, N, NRHS, A, LDA, IPIV, B, LDB, WORK, LWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           CHARACTER        UPLO
           INTEGER          INFO, LDA, LDB, N, LWORK, NRHS
!     ..
!     .. Array Arguments ..
           INTEGER          IPIV( * )
           COMPLEX*16       A( LDA, * ), B( LDB, * ), WORK( * )
        end subroutine
!
    end interface
!
!   --------------------------------------------------------------------------!
!   Solution of a real or complex system of linear equations with a symmetric !
!   or hermitian positive definite matrix and several right hand side vectors !
!   --------------------------------------------------------------------------!
!
!   Generic interface for SPOSV, DPOSV, CPOSV and ZPOSV subroutines in LAPACK
!
    interface posv
!
        subroutine sposv( UPLO, N, NRHS, A, LDA, B, LDB, INFO)
!     ..
!     .. Scalar Arguments ..
           CHARACTER        UPLO
           INTEGER          INFO, LDA, LDB, N, NRHS
!     ..
!     .. Array Arguments ..
           REAL             A( LDA, * ), B( LDB, * )
        end subroutine
!
        subroutine dposv( UPLO, N, NRHS, A, LDA, B, LDB, INFO)
!     ..
!     .. Scalar Arguments ..
           CHARACTER        UPLO
           INTEGER          INFO, LDA, LDB, N, NRHS
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION A( LDA, * ), B( LDB, * )
        end subroutine
!
        subroutine cposv( UPLO, N, NRHS, A, LDA, B, LDB, INFO)
!     ..
!     .. Scalar Arguments ..
           CHARACTER        UPLO
           INTEGER          INFO, LDA, LDB, N, NRHS
!     ..
!     .. Array Arguments ..
           COMPLEX          A( LDA, * ), B( LDB, * )
        end subroutine
!
        subroutine zposv( UPLO, N, NRHS, A, LDA, B, LDB, INFO)
!     ..
!     .. Scalar Arguments ..
           CHARACTER        UPLO
           INTEGER          INFO, LDA, LDB, N, NRHS
!     ..
!     .. Array Arguments ..
           COMPLEX*16       A( LDA, * ), B( LDB, * )
        end subroutine
!
    end interface
!
!   ------------------------------------------------------------------------!
!   Minimum-norm solution of a real or complex linear least square problem  !
!   with several right hand side vectors                                    !
!   ------------------------------------------------------------------------!
!
!   Generic interface for SGELSS, DGELSS, CGELSS and ZGELSS subroutines in LAPACK
!
    interface gelss
!
        subroutine sgelss( M, N, NRHS, A, LDA, B, LDB, S, RCOND, RANK, WORK, LWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           INTEGER          INFO, LDA, LDB, LWORK, M, N, NRHS, RANK
           REAL             RCOND
!     ..
!     .. Array Arguments ..
           REAL             S( * ), A( LDA, * ), B( LDB, * ), WORK( * )
        end subroutine
!
        subroutine dgelss( M, N, NRHS, A, LDA, B, LDB, S, RCOND, RANK, WORK, LWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           INTEGER          INFO, LDA, LDB, LWORK, M, N, NRHS, RANK
           DOUBLE PRECISION RCOND
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION S( * ), A( LDA, * ), B( LDB, * ), WORK( * )
        end subroutine
!
        subroutine cgelss( M, N, NRHS, A, LDA, B, LDB, S, RCOND, RANK, WORK, LWORK, RWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           INTEGER          INFO, LDA, LDB, LWORK, M, N, NRHS, RANK
           REAL             RCOND
!     ..
!     .. Array Arguments ..
           REAL             S( * ), RWORK( * )
           COMPLEX          A( LDA, * ), B( LDB, * ), WORK( * )
        end subroutine
!
        subroutine zgelss( M, N, NRHS, A, LDA, B, LDB, S, RCOND, RANK, WORK, LWORK, RWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           INTEGER          INFO, LDA, LDB, LWORK, M, N, NRHS, RANK
           DOUBLE PRECISION RCOND
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION S( * ), RWORK( * )
           COMPLEX*16       A( LDA, * ), B( LDB, * ), WORK( * )
        end subroutine
!
    end interface
!
!   Generic interface for SGELSD, DGELSD, CGELSD and ZGELSD subroutines in LAPACK
!
    interface gelsd
!
        subroutine sgelsd( M, N, NRHS, A, LDA, B, LDB, S, RCOND, RANK, WORK, LWORK, IWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           INTEGER          INFO, LDA, LDB, LWORK, M, N, NRHS, RANK
           REAL             RCOND
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           REAL             S( * ), A( LDA, * ), B( LDB, * ), WORK( * )
        end subroutine
!
        subroutine dgelsd( M, N, NRHS, A, LDA, B, LDB, S, RCOND, RANK, WORK, LWORK, IWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           INTEGER          INFO, LDA, LDB, LWORK, M, N, NRHS, RANK
           DOUBLE PRECISION RCOND
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           DOUBLE PRECISION S( * ), A( LDA, * ), B( LDB, * ), WORK( * )
        end subroutine
!
        subroutine cgelsd( M, N, NRHS, A, LDA, B, LDB, S, RCOND, RANK, WORK, LWORK, RWORK, IWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           INTEGER          INFO, LDA, LDB, LWORK, M, N, NRHS, RANK
           REAL             RCOND
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           REAL             S( * ), RWORK( * )
           COMPLEX          A( LDA, * ), B( LDB, * ), WORK( * )
        end subroutine
!
        subroutine zgelsd( M, N, NRHS, A, LDA, B, LDB, S, RCOND, RANK, WORK, LWORK, RWORK, IWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           INTEGER          INFO, LDA, LDB, LWORK, M, N, NRHS, RANK
           DOUBLE PRECISION RCOND
!     ..
!     .. Array Arguments ..
           INTEGER          IWORK( * )
           DOUBLE PRECISION S( * ), RWORK( * )
           COMPLEX*16       A( LDA, * ), B( LDB, * ), WORK( * )
        end subroutine
!
    end interface
!
!   Generic interface for SGELSY, DGELSY, CGELSY and ZGELSY subroutines in LAPACK
!
    interface gelsy
!
        subroutine sgelsy( M, N, NRHS, A, LDA, B, LDB, JPVT, RCOND, RANK, WORK, LWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           INTEGER          INFO, LDA, LDB, LWORK, M, N, NRHS, RANK
           REAL             RCOND
!     ..
!     .. Array Arguments ..
           INTEGER          JPVT( * )
           REAL             A( LDA, * ), B( LDB, * ), WORK( * )
        end subroutine
!
        subroutine dgelsy( M, N, NRHS, A, LDA, B, LDB, JPVT, RCOND, RANK, WORK, LWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           INTEGER          INFO, LDA, LDB, LWORK, M, N, NRHS, RANK
           DOUBLE PRECISION RCOND
!     ..
!     .. Array Arguments ..
           INTEGER          JPVT( * )
           DOUBLE PRECISION A( LDA, * ), B( LDB, * ), WORK( * )
        end subroutine
!
        subroutine cgelsy( M, N, NRHS, A, LDA, B, LDB, JPVT, RCOND, RANK, WORK, LWORK, RWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           INTEGER          INFO, LDA, LDB, LWORK, M, N, NRHS, RANK
           REAL             RCOND
!     ..
!     .. Array Arguments ..
           INTEGER          JPVT( * )
           REAL             RWORK( * )
           COMPLEX          A( LDA, * ), B( LDB, * ), WORK( * )
        end subroutine
!
        subroutine zgelsy( M, N, NRHS, A, LDA, B, LDB, JPVT, RCOND, RANK, WORK, LWORK, RWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           INTEGER          INFO, LDA, LDB, LWORK, M, N, NRHS, RANK
           DOUBLE PRECISION RCOND
!     ..
!     .. Array Arguments ..
           INTEGER          JPVT( * )
           DOUBLE PRECISION RWORK( * )
           COMPLEX*16       A( LDA, * ), B( LDB, * ), WORK( * )
        end subroutine
!
    end interface
!
!   -----------------------------------------------------------------------!
!   Solution of real or complex, overdetermined or underdetermined, linear !
!   systems with a coefficient matrix of full rank and several right hand  !
!   side vectors                                                           !
!   -----------------------------------------------------------------------!
!
!   Generic interface for SGELS, DGELS, CGELS and ZGELS subroutines in LAPACK
!
    interface gels
!
        subroutine sgels( TRANS, M, N, NRHS, A, LDA, B, LDB, WORK, LWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           CHARACTER        TRANS
           INTEGER          INFO, LDA, LDB, LWORK, M, N, NRHS
!     ..
!     .. Array Arguments ..
           REAL             A( LDA, * ), B( LDB, * ), WORK( * )
        end subroutine
!
        subroutine dgels( TRANS, M, N, NRHS, A, LDA, B, LDB, WORK, LWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           CHARACTER        TRANS
           INTEGER          INFO, LDA, LDB, LWORK, M, N, NRHS
!     ..
!     .. Array Arguments ..
           DOUBLE PRECISION A( LDA, * ), B( LDB, * ), WORK( * )
        end subroutine
!
        subroutine cgels( TRANS, M, N, NRHS, A, LDA, B, LDB, WORK, LWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           CHARACTER        TRANS
           INTEGER          INFO, LDA, LDB, LWORK, M, N, NRHS
!     ..
!     .. Array Arguments ..
           COMPLEX          A( LDA, * ), B( LDB, * ), WORK( * )
        end subroutine
!
        subroutine zgels( TRANS, M, N, NRHS, A, LDA, B, LDB, WORK, LWORK, INFO)
!     ..
!     .. Scalar Arguments ..
           CHARACTER        TRANS
           INTEGER          INFO, LDA, LDB, LWORK, M, N, NRHS
!     ..
!     .. Array Arguments ..
           COMPLEX*16       A( LDA, * ), B( LDB, * ), WORK( * )
        end subroutine
!
    end interface
!
!
! *********************************
! END OF MODULE Lapack_interfaces *
! *********************************
!
end module Lapack_interfaces
