module Print_Procedures
! ####################################################################################
!
! Copyright 2018 IRD
!
! This file is part of statpack.
!
! statpack is free software: you can redistribute it and/or modify
! it under the terms of the GNU Lesser General Public License as
! published by the Free Software Foundation, either version 3 of 
! the License, or (at your option) any later version.
!
! statpack is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU Lesser General Public License for more details.
!
! You can find a copy of the GNU Lesser General Public License
! in the statpack/doc directory.
!
! ####################################################################################
!                                                                                    *
! ************************************************************************************
! MODULE EXPORTING PRINTING UTILITIES.                                               *
!                                                                                    * 
! LATEST REVISION : 24/06/2018                                                       *
!                                                                                    *
! ####################################################################################
!                                                                                    *
! ************************************************************************************
!
!
! USED MODULES 
! ============
!
    use Select_Parameters, only : lgl, i1b, i4b, stnd, defunit
!
! OTHER MODULES USED IN SPECIFIC SUBROUTINES
! ==========================================
!
!   use Char_Constants
!   use Utilities
!   use String_Procedures
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!
! PUBLIC ENTITIES
! ===============
!    
    private
    public :: defline, defindent, defw, defd, defs,    &
              enter_proc, leave_proc,                  &
              entering, leaving, indent,               &
              write_array, print_array,                &
              print_prinfac, print_stat
!
! GENERIC INTERFACES FOR ROUTINES WITH OVERLOADED VERSIONS
! ========================================================
!
    interface write_array
        module procedure  write_rm, write_im, write_rv, write_iv
    end interface
!
    interface print_array
        module procedure  print_rm, print_im, print_rv, print_iv
    end interface
!
    interface print_prinfac
        module procedure  print_prinfac_rv, print_prinfac_rm
    end interface
!
    interface print_stat
        module procedure  print_stat_missing, print_stat_weight
    end interface
!   
! MODULE PARAMETERS 
! =================
!
! FORMAT SPECIFICATIONS.
!
    character , parameter ::  d_maj=achar(68),  d_min=achar(100),       &
                              f_maj=achar(70),  f_min=achar(102),       &
                              g_maj=achar(71),  g_min=achar(103),       &
                              p_maj=achar(80),  p_min=achar(112),       &
                              s_maj=achar(83),  s_min=achar(115)
!
!
! DEFAULT VALUES FOR PRINTING.
!
    integer(i4b),   parameter   :: defline   = 80 ,    & ! DEFAULT LINE SIZE
                                   defindent = 0  ,    & ! DEFAULT INDENTATION
                                   defw      = 12 ,    & ! DEFAULT WIDTH FOR EDIT DESCRIPTOR
                                   defd      = 6  ,    & ! DEFAULT NUMBER OF SIGNIFICANT DIGITS
                                   defs      = 3         ! DEFAULT NUMBER OF SPACES BETWEEN ENTRIES
!
    character(len=*), parameter :: deff       = f_min,       &   ! DEFAULT EDIT DESCRIPTOR
                                   defsign_ed = s_min//s_min     ! DEFAULT SIGN EDIT DESCRIPTOR
!
!
! =========================================================================================
!
                                  contains
!                                 ========
!
! =========================================================================================
!
    subroutine enter_proc( string, level, prt_unit )
!
! Purpose
! _______
!
!   Upon entering a procedure, this subroutine will be called.
!   It would skips two lines and outputs a message that the routine
!   identified by STRING was entered. If LEVEL is present, the message
!   is prepended by LEVEL blanks.
!
!
! Arguments
! _________
!
!   STRING      (INPUT) character(len=*)
!               The string which identifies the routine.
!
!   LEVEL       (INPUT, OPTIONAL) integer(i4b)
!               The number of blanks to use for indentation.
!
!   PRT_UNIT    (INPUT, OPTIONAL) integer(i4b)
!               The printing unit.
!
!
! Further Details
! _______________
!
!   Leading and trailing blanks in STRING are removed.
!   If PRT_UNIT is absent, then all output is on the unit DEFUNIT.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,   only : blank
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character(len=*), intent(in)           :: string
!
    integer(i4b),     intent(in), optional :: level, prt_unit
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: lev, actunit
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
    if ( present(prt_unit) ) then
        actunit = merge( prt_unit , defunit , prt_unit>=0_i4b )
    else
        actunit = defunit
    end if
!
    if ( present(level) ) then
        lev = level
    else
        lev = 0_i4b
    end if
!
    write( actunit, '(//,a)' )  repeat(blank,lev)//'[Into '//trim(adjustl(string))//']'
!
!
! END OF SUBROUTINE enter_proc
! ____________________________
!
    end subroutine enter_proc
!
! =========================================================================================
!
    subroutine leave_proc( string, level, prt_unit )
!
! Purpose
! _______
!
!   This is the 'opposite' to ENTER_PROC. It should be called
!   just before leaving a routine. An exit message is output
!   on PRT_UNIT and two lines are skipped.
!
!
! Arguments
! _________
!
!   STRING      (INPUT) character(len=*)
!               The string which identifies the routine.
!
!   LEVEL       (INPUT, OPTIONAL) integer(i4b)
!               The number of blanks to use for indentation.
!
!   PRT_UNIT    (INPUT, OPTIONAL) integer(i4b)
!               The printing unit.
!
!
! Further Details
! _______________
!
!   Leading and trailing blanks in STRING are removed.
!   If PRT_UNIT is absent, then all output is on the unit DEFUNIT.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,   only : blank
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character(len=*), intent(in)           :: string
!
    integer(i4b),     intent(in), optional :: level, prt_unit
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: lev, actunit
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    if ( present(prt_unit) ) then
        actunit = merge( prt_unit , defunit , prt_unit>=0_i4b )
    else
        actunit = defunit
    end if
!
    if ( present(level) ) then
        lev = level
    else
        lev = 0_i4b
    end if
!
    write( actunit, '(a,//)' ) repeat(blank,lev)//'[Done '//trim(adjustl(string))//']'
!
!
! END OF SUBROUTINE leave_proc
! ____________________________
!
    end subroutine leave_proc
!
! =========================================================================================
!
    function entering( string, level, trace, prt_unit )
!
! Purpose
! _______
!
!   Upon entering a procedure, this function will be called.
!   It will return a prefix string suitable for indenting
!   output lines from the procedure. It takes the given STRING
!   and prepends LEVEL blanks, followed by a '[', and appends
!   the character ']'. For example, if STRING were 'hi' and LEVEL
!   were 7, it would return "       [hi]".  LEVEL is then also
!   incremented by 2.
!
!   If TRACE is true, it also outputs a message that the
!   routine identified by STRING was entered.
!
!
! Arguments
! _________
!
!   STRING      (INPUT) character(len=*)
!               The string which identifies the routine.
!
!   LEVEL       (INPUT/OUTPUT) integer(i4b)
!               The number of blanks to use for indentation.
!               LEVEL is incremented by 2 on output.
!
!   TRACE       (INPUT) logical(lgl)
!               Logical indicator for printing the message.
!
!   PRT_UNIT    (INPUT, OPTIONAL) integer(i4b)
!               The printing unit.
!
!
! Further Details
! _______________
!
!   Trailing blanks in STRING are removed.
!   If PRT_UNIT is absent, then all output is on the unit DEFUNIT.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,   only : blank
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character(len=*), intent(in)              :: string
!
    logical(lgl),     intent(in)              :: trace
!
    integer(i4b),     intent(inout)           :: level
    integer(i4b),     intent(in), optional    :: prt_unit
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: actunit
!
    character(len=len_trim(string)+2+level) :: entering
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    entering(1_i4b:level) = blank
!
    if ( trace ) then
        if ( present(prt_unit) ) then
            actunit = merge( prt_unit , defunit , prt_unit>=0_i4b )
        else
            actunit = defunit
        end if
        write( actunit ,'(a)' ) entering(1_i4b:level)//'[into '//trim(string)//']'
    end if
!
    entering(level+1_i4b:) = '[' // trim(string) // ']'
!
    level = level + 2_i4b
!
!
! END OF FUNCTION entering
! ________________________
!
    end function entering
!
! =========================================================================================
!
    subroutine leaving( string, level, trace, prt_unit )
!
! Purpose
! _______
!
!   This is the 'opposite' to ENTERING. It should be called
!   just before leaving a routine. LEVEL is reduced by
!   2 and if TRACE is true, an exit message is output.
!
!
! Arguments
! _________
!
!   STRING      (INPUT) character(len=*)
!               The string which identifies the routine.
!
!   LEVEL       (INPUT/OUTPUT) integer(i4b)
!               The number of blanks to use for indentation.
!               LEVEL is reduced by 2 on output.
!
!   TRACE       (INPUT) logical(lgl)
!               Logical indicator for printing the message.
!
!   PRT_UNIT    (INPUT, OPTIONAL) integer(i4b)
!               The printing unit.
!
!
! Further Details
! _______________
!
!   Trailing blanks in STRING are removed.
!   If PRT_UNIT is absent, then all output is on the unit DEFUNIT.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,   only : blank
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character(len=*), intent(in)            :: string
!
    logical(lgl),     intent(in)            :: trace
!
    integer(i4b),     intent(inout)         :: level
    integer(i4b),     intent(in), optional  :: prt_unit
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: actunit
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    if ( present(prt_unit) ) then
        actunit = merge( prt_unit , defunit , prt_unit>=0_i4b )
    else
        actunit = defunit
    end if
!
    level = level - 2_i4b
    if ( trace ) write( actunit , '(a)' ) repeat(blank,level)//'[Done '//trim(string)//']'
!
!
! END OF SUBROUTINE leaving
! _________________________
!
    end subroutine leaving
!
! =========================================================================================
!
    subroutine indent( id, level, prt_unit )
!
! Purpose
! _______
!
!   This is also used to indent output, albeit in a manner
!   different from ENTERING and LEAVING. It simply writes
!   out LEVEL blanks followed by the string ID in [], and
!   leaves the output file marker where it is. It uses
!   nonadvancing output.  If LEVEL is not present, just
!   the ID part is output; i.e. LEVEL is treated as zero.
!
!
! Arguments
! _________
!
!   ID          (INPUT) character(len=*)
!               The string to print.
!
!   LEVEL       (INPUT, OPTIONAL) integer(i4b)
!               The number of blanks to use for indentation.
!
!   PRT_UNIT    (INPUT, OPTIONAL) integer(i4b)
!               The printing unit.
!
!
! Further Details
! _______________
!
!   Leading and trailing blanks in ID are removed.
!   If PRT_UNIT is absent, then all output is on the unit DEFUNIT.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,   only : blank
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character(len=*), intent(in)           :: id
!
    integer(i4b),     intent(in), optional :: level, prt_unit
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: lev, actunit
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    if ( present(prt_unit) ) then
        actunit = merge( prt_unit , defunit , prt_unit>=0_i4b )
    else
        actunit = defunit
    end if
!
    if ( present(level) ) then
        lev = level
    else
        lev = 0_i4b
    end if
!
    write( actunit, '(a)', advance='no' ) repeat(blank,lev)//'['//trim(adjustl(id))//']'
!
!
! END OF SUBROUTINE indent
! ________________________
!
    end subroutine indent
!
! =========================================================================================
!
    subroutine write_rm( x, f, w, d, s, name, indent, line, prt_unit )
!
! Purpose
! _______
!
!   Print out a real matrix with given format, as below.
!   The matrix is printed row by row.
!
!   Print also a title for the matrix: NAME
!
!
! Arguments
! _________
!
!   X           (INPUT) real(stnd), dimension(:,:)
!               The matrix to be output.
!
!   F           (INPUT, OPTIONAL) character
!
!   W, D        (INPUT, OPTIONAL) integer(i4b)
!               Selects the edit descriptor: fw.d .
!               Print each entry in format  FW.D  .
!               F is a character 'f', 'g', 'e' or 'd',
!               regardless of case.
!               W and D are integers .
!
!   S           (INPUT, OPTIONAL) integer(i4b)
!               The number of spaces between each entry.
!
!   NAME        (INPUT, OPTIONAL) character(len=*)
!               Prints the string NAME as a title for the matrix.
!
!   IDENT       (INPUT, OPTIONAL) integer(i4b)
!               The number of blanks to use for indentation.
!
!   LINE        (INPUT, OPTIONAL) integer(i4b)
!               The number of characters per line.
!
!   PRT_UNIT    (INPUT, OPTIONAL) integer(i4b)
!               Selects the output unit.
!
!
! Further Details
! _______________
!
!   If the value of the argument  INDENT  is positive, then each output
!   line is preceded by  INDENT  blank characters.  If  INDENT  is negative,
!   then only the matrix output will be indented (not the TITLE).
!
!
!   Each output line from this subroutine will have at most max(LINE,W+abs(INDENT))
!   characters plus one additional leading character for Fortran "carriage control".
!   The carriage control character will always be a blank.
!
!   If LINE is absent, then line width is at most max(80,W+abs(INDENT)) characters 
!   plus one for carriage control.
!
!   All output is on the unit PRT_UNIT.
!   If PRT_UNIT is absent, then all output is on the unit DEFUNIT.
!
!   Defaults are defined for all optional arguments.  See start of module.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,   only : blank
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),       intent(in)           :: x(:,:)
!
    integer(i4b),     intent(in), optional :: w, d, s, indent, line, prt_unit
!
    character,        intent(in), optional :: f
    character(len=*), intent(in), optional :: name
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: actunit, actw, actd, acts, actindent, actline, i
!                 
    character(len=34) :: form, actf*1
!
!
! PARAMETERS
! __________
!
    character(len=34), parameter :: defform ='((**x,**(* **.**, ** x), * **.**))'
!                             e.g.            ((02x,05(f 09.04, 03 x), f 09.04))
!                                             1234567890123456789012345678901234
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( size(x)==0 ) return
!
!   FIRST SET APPROPRIATE DEFAULT VALUES.
!
    form = defform
!
    if ( present(prt_unit) ) then
        actunit = merge( prt_unit , defunit , prt_unit>=0_i4b )
    else
        actunit = defunit
    end if
!
    if ( present(line) ) then
        actline = line
    else
        actline = defline
    end if
!
    if ( present(indent) ) then
        actindent = sign( min(98_i4b, abs(indent)), indent )
    else
        actindent = defindent
    end if
!
    if ( present(f) ) then
        select case( f )
            case( d_maj:g_maj, d_min:g_min )
                actf = f
            case default
                actf = deff
        end select
    else
        actf = deff
    end if
!
    if ( present(w) ) then
        actw = merge( w , defw , w>0_i4b .and. w<=99_i4b )
    else
        actw = defw
    end if
!
    if ( present(d) ) then
        actd = merge( d , defd , d>=0_i4b .and. d<=99_i4b )
    else
        actd = defd
    end if
    select case( actf )
        case( f_maj, f_min )
            actd =  merge( actd , max(0_i4b,actw-2_i4b) , actd<=actw-2_i4b )
        case default
            actd =  merge( actd , max(0_i4b,actw-6_i4b) , actd<=actw-6_i4b )
    end select
!
    if ( present(s) ) then
        acts = merge( s, defs, s>=0_i4b .and. s<=99_i4b )
    else
        acts = defs
    end if
!
!   CONSTRUCT THE FORMAT.
!
    write( form(03:04), 99 ) abs(actindent) + 1_i4b
!
    write( form(07:08), 99 )         &
        min( max( (actline-abs(actindent)-actw)/(actw+acts), 0_i4b ), 99_i4b )
    write( form(10:10), 97 ) actf
    write( form(12:13), 99 ) actw
    write( form(15:16), 99 ) actd
!
    if ( acts/=0_i4b ) then
        write( form(19:20), 99 ) acts
    else
        write( form(17:22), 98 ) blank
    end if
!
    write( form(26:26), 97 ) actf
    write( form(28:29), 99 ) actw
    write( form(31:32), 99 ) actd
!
!   SKIP TWO LINES.
!
    write( actunit, "(/)" )
!
!   WRITE TITLE.
!
    if ( present(name) ) then
        if ( actindent>0_i4b ) then
            i = min( int(len_trim(name),i4b) , max(actline-actindent,actw) )
            write( actunit, 98 )  repeat( blank, actindent+1_i4b )//name(1_i4b:i)
        else
            i = min( int(len_trim(name),i4b) , max(actline,actw-actindent) )
            write( actunit, 98 ) blank//name(1_i4b:i)
        end if
    end if
!
!   WRITE MATRIX.
!
    do i = 1_i4b, size( x, dim=1 )
        write( actunit, form )  x(i,:)
    end do
!
! FORMATS
! _______
!
    97   format (a1)
    98   format (a)
    99   format (i2.2)
!
!
! END OF SUBROUTINE write_rm
! __________________________
!
    end subroutine write_rm
!
! =========================================================================================
!
    subroutine write_im( x, w, s, name, indent, line, prt_unit )
!
! Purpose
! _______
!
!   Print out an integer matrix X with given format, as below.
!   The matrix is printed row by row.
!
!   Print also a title for the matrix: NAME
!
!
! Arguments
! _________
!
!   X           (INPUT) integer(i4b), dimension(:,:)
!               The matrix to be output.
!
!   W           (INPUT, OPTIONAL) integer(i4b)
!               Selects the width of each entry.
!               Print each entry in format  iW .
!
!   S           (INPUT, OPTIONAL) integer(i4b)
!               The number of spaces between each entry.
!
!   NAME        (INPUT, OPTIONAL) character(len=*)
!               Prints the string NAME as a title for the matrix.
!
!   IDENT       (INPUT, OPTIONAL) integer(i4b)
!               The number of blanks to use for indentation.
!
!   LINE        (INPUT, OPTIONAL) integer(i4b)
!               The number of characters per line.
!
!   PRT_UNIT    (INPUT, OPTIONAL) integer(i4b)
!               Selects the output unit.
!
!
! Further Details
! _______________
!
!   If W is absent, then the routine determines the best width w needed to edit 
!   the array X without excess blanks.
!
!   If the value of the argument  INDENT  is positive, then each output
!   line is preceded by  INDENT  blank characters.  If  INDENT  is negative,
!   then only the matrix output will be indented (not the TITLE).
!
!
!   Each output line from this subroutine will have at most max(LINE,W+abs(INDENT))
!   characters plus one additional leading character for Fortran "carriage control".
!   The carriage control character will always be a blank.
!
!   If LINE is absent, then line width is at most max(80,W+abs(INDENT)) characters 
!   plus one for carriage control.
!
!   All output is on the unit PRT_UNIT.
!   If PRT_UNIT is absent, then all output is on the unit DEFUNIT.
!
!   Defaults are defined for all optional arguments.  See start of module.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,    only : blank
    use String_Procedures, only : nbrchf
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b),     intent(in)           :: x(:,:)
    integer(i4b),     intent(in), optional :: w, s, indent, line, prt_unit
!
    character(len=*), intent(in), optional :: name
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: actunit, actw, acts, actindent, actline, i
!                 
    character(len=28) :: form
!
!
! PARAMETERS
! __________
!
    character(len=28), parameter :: defform ='((**x,**(i **, ** x), i **))'
!                           e.g.              ((02x,05(i 09, 03 x), i 09))
!                                             1234567890123456789012345678
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( size(x)==0 ) return
!
!   FIRST SET APPROPRIATE DEFAULT VALUES.
!
    form = defform
!
    if ( present(prt_unit) ) then
        actunit = merge( prt_unit , defunit , prt_unit>=0_i4b )
    else
        actunit = defunit
    end if
!
    if ( present(line) ) then
        actline = line
    else
        actline = defline
    end if
!
    if ( present(indent) ) then
        actindent = sign( min(98_i4b, abs(indent)), indent )
    else
        actindent = defindent
    end if
!
    if ( present(w) ) then
        if ( w>0_i4b .and. w<=99_i4b ) then
            actw = w
        else
            actw = min( nbrchf(maxval(abs(x)))+1_i4b , 99_i4b )
        end if
    else
        actw = min( nbrchf(maxval(abs(x)))+1_i4b , 99_i4b )
    end if
!
    if ( present(s) ) then
        acts = merge( s, defs, s>=0_i4b .and. s<=99_i4b )
    else
        acts = defs
    end if
!
!   CONSTRUCT THE FORMAT.
!
    write( form(03:04), 99 ) abs(actindent) + 1_i4b
!
    write( form(07:08), 99 )         &
        min( max( (actline-abs(actindent)-actw)/(actw+acts), 0_i4b ), 99_i4b )
    write( form(12:13), 99 ) actw
!
    if ( acts/=0_i4b ) then
        write( form(16:17), 99 ) acts
    else
        write( form(14:19), 98 ) blank
    end if
!
    write( form(25:26), 99 ) actw
!
!   SKIP TWO LINES.
!
    write( actunit, "(/)" )
!
!   WRITE TITLE.
!
    if ( present(name) ) then
        if ( actindent>0_i4b ) then
            i = min( int(len_trim(name),i4b) , max(actline-actindent,actw) )
            write( actunit, 98 )  repeat( blank, actindent+1_i4b )//name(1_i4b:i)
        else
            i = min( int(len_trim(name),i4b) , max(actline,actw-actindent) )
            write( actunit, 98 ) blank//name(1_i4b:i)
        end if
    end if
!
!   WRITE MATRIX.
!
    do i = 1_i4b, size( x, dim=1 )
        write( actunit, form )  x(i,:)
    end do
!
! FORMATS
! _______
!
    97   format (a1)
    98   format (a)
    99   format (i2.2)
!
!
! END OF SUBROUTINE write_im
! __________________________
!
    end subroutine write_im
!
! =========================================================================================
!
    subroutine write_rv( x, f, w, d, s, name, indent, line, prt_unit )
!
! Purpose
! _______
!
!   This subroutine prints out a real vector with a given format and a title,
!   as given in the input arguments.
!
!
! Arguments
! _________
!
!   X           (INPUT) real(stnd), dimension(:)
!               The vector to be output.
!
!   F           (INPUT, OPTIONAL) character
!
!   W, D        (INPUT, OPTIONAL) integer(i4b)
!               Selects the edit descriptor: fw.d  .
!               Print each entry in format  FW.D .
!               F is a character 'f', 'g', 'e' or 'd',
!               regardless of case.
!               W and D are integers .
!
!   S           (INPUT, OPTIONAL) integer(i4b)
!               The number of spaces between each entry.
!
!   NAME        (INPUT, OPTIONAL) character(len=*)
!               Prints the string NAME as a title for the vector.
!
!   IDENT       (INPUT, OPTIONAL) integer(i4b)
!               The number of blanks to use for indentation.
!
!   LINE        (INPUT, OPTIONAL) integer(i4b)
!               The number of characters per line.
!
!   PRT_UNIT    (INPUT, OPTIONAL) integer(i4b)
!               Selects the output unit.
!
!
! Further Details
! _______________
!
!   If the value of the argument  INDENT  is positive, then each output
!   line is preceded by  INDENT  blank characters.  If  INDENT  is negative,
!   then only the vector output will be indented (not the TITLE).
!
!   Each output line from this subroutine will have at most max(LINE,W+abs(INDENT))
!   characters plus one additional leading character for Fortran "carriage control".
!   The carriage control character will always be a blank.
!
!   If LINE is absent, then line width is at most max(80,W+abs(INDENT)) characters 
!   plus one for carriage control.
!
!   All output is on the unit PRT_UNIT.
!   If PRT_UNIT is absent, then all output is on the unit DEFUNIT.
!
!   Defaults are defined for all optional arguments.  See start of module.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,   only : blank
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),       intent(in)           :: x(:)
!
    integer(i4b),     intent(in), optional :: w, d, s, indent, line, prt_unit
!
    character,        intent(in), optional :: f
    character(len=*), intent(in), optional :: name
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: actunit, actw, actd, acts, actindent, actline, i
!                 
    character(len=34) :: form, actf*1
!
!
! PARAMETERS
! __________
!
    character(len=34), parameter :: defform ='((**x,**(* **.**, ** x), * **.**))'
!                           e.g.              ((02x,05(f 09.04, 03 x), f 09.04))
!                                             1234567890123456789012345678901234
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( size(x)==0 ) return
!
!   FIRST SET APPROPRIATE DEFAULT VALUES.
!
    form = defform
!
    if ( present(prt_unit) ) then
        actunit = merge( prt_unit , defunit , prt_unit>=0_i4b )
    else
        actunit = defunit
    end if
!
    if ( present(line) ) then
        actline = line
    else
        actline = defline
    end if
!
    if ( present(indent) ) then
        actindent = sign( min(98_i4b, abs(indent)), indent )
    else
        actindent = defindent
    end if
!
    if ( present(f) ) then
        select case( f )
            case( d_maj:g_maj, d_min:g_min )
                actf = f
            case default
                actf = deff
        end select
    else
        actf = deff
    end if
!
    if ( present(w) ) then
        actw = merge( w , defw , w>0_i4b .and. w<=99_i4b )
    else
        actw = defw
    end if
!
    if ( present(d) ) then
        actd = merge( d , defd , d>=0_i4b .and. d<=99_i4b )
    else
        actd = defd
    end if
    select case( actf )
        case( f_maj, f_min )
            actd =  merge( actd , max(0_i4b,actw-2_i4b) , actd<=actw-2_i4b )
        case default
            actd =  merge( actd , max(0_i4b,actw-6_i4b) , actd<=actw-6_i4b )
    end select
!
    if ( present(s) ) then
        acts = merge( s , defs , s>=0_i4b .and. s<=99_i4b )
    else
        acts = defs
    end if
!
!   CONSTRUCT THE FORMAT.
!
    write( form(03:04), 99 ) abs(actindent) + 1_i4b
!
    write( form(07:08), 99 )         &
            min( max( (actline-abs(actindent)-actw)/(actw+acts), 0_i4b ), 99_i4b )
    write( form(10:10), 97 ) actf
    write( form(12:13), 99 ) actw
    write( form(15:16), 99 ) actd
!
    if ( acts/=0_i4b ) then
        write( form(19:20), 99 ) acts
    else
        write( form(17:22), 98 ) blank
    end if
!
    write( form(26:26), 97 ) actf
    write( form(28:29), 99 ) actw
    write( form(31:32), 99 ) actd
!
!   SKIP TWO LINES.
!
    write( actunit, "(/)" )
!
!   WRITE TITLE.
!
    if ( present(name) ) then
        if ( actindent>0_i4b ) then
            i = min( int(len_trim(name),i4b) , max(actline-actindent,actw) )
            write( actunit, 98 )  repeat( blank, actindent+1_i4b )//name(1_i4b:i)
        else
            i = min( int(len_trim(name),i4b) , max(actline,actw-actindent) )
            write( actunit, 98 ) blank//name(1_i4b:i)
        end if
    end if
!
!   WRITE VECTOR.
!
    write( actunit, form ) x
!
! FORMATS
! _______
!
    97   format (a1)
    98   format (a)
    99   format (i2.2)
!
!
! END OF SUBROUTINE write_rv
! __________________________
!
    end subroutine write_rv
!
! =========================================================================================
!
    subroutine write_iv( x, w, s, name, indent, line, prt_unit )
!
! Purpose
! _______
!
!   This subroutine prints out an integer vector with a given format and a title,
!   as given in the input arguments.
!
!
! Arguments
! _________
!
!   X           (INPUT) integer(i4b), dimension(:)
!               The vector to be output.
!
!   W           (INPUT, OPTIONAL) integer(i4b)
!               Selects the width of each entry.
!               Print each entry in format  iW .
!
!   S           (INPUT, OPTIONAL) integer(i4b)
!               The number of spaces between each entry.
!
!   NAME        (INPUT, OPTIONAL) character(len=*)
!               Prints the string NAME as a title for the matrix.
!
!   IDENT       (INPUT, OPTIONAL) integer(i4b)
!               The number of blanks to use for indentation.
!
!   LINE        (INPUT, OPTIONAL) integer(i4b)
!               The number of characters per line.
!
!   PRT_UNIT    (INPUT, OPTIONAL) integer(i4b)
!               Selects the output unit.
!
!
! Further Details
! _______________
!
!   If W is absent, then the routine determines the best width w needed to edit 
!   the array X without excess blanks.
!
!   If the value of the argument  INDENT  is positive, then each output
!   line is preceded by  INDENT  blank characters.  If  INDENT  is negative,
!   then only the vector output will be indented (not the TITLE).
!
!   Each output line from this subroutine will have at most max(line,W+abs(INDENT))
!   characters plus one additional leading character for Fortran "carriage control".
!   The carriage control character will always be a blank.
!
!   If LINE is absent, then line width is at most max(80,W+abs(INDENT)) characters 
!   plus one for carriage control.
!
!   All output is on the unit PRT_UNIT.
!   If PRT_UNIT is absent, then all output is on the unit DEFUNIT.
!
!   Defaults are defined for all optional arguments.  See start of module.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,    only : blank
    use String_Procedures, only : nbrchf
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b),     intent(in)           :: x(:)
    integer(i4b),     intent(in), optional :: w, s, indent, line, prt_unit
!
    character(len=*), intent(in), optional :: name
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: actunit, actw, acts, actindent, actline, i
!                 
    character(len=28) :: form
!
!
! PARAMETERS
! __________
!
    character(len=28), parameter :: defform ='((**x,**(i **, ** x), i **))'
!                           e.g.              ((02x,05(i 09, 03 x), i 09))
!                                             1234567890123456789012345678
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( size(x)==0 ) return
!
!   FIRST SET APPROPRIATE DEFAULT VALUES.
!
    form = defform
!
    if ( present(prt_unit) ) then
        actunit = merge( prt_unit , defunit , prt_unit>=0_i4b )
    else
        actunit = defunit
    end if
!
    if ( present(line) ) then
        actline = line
    else
        actline = defline
    end if
!
    if ( present(indent) ) then
        actindent = sign( min(98_i4b, abs(indent)), indent )
    else
        actindent = defindent
    end if
!
    if ( present(w) ) then
        if ( w>0_i4b .and. w<=99_i4b ) then
            actw = w
        else
            actw = min( nbrchf(maxval(abs(x)))+1_i4b , 99_i4b )
        end if
    else
        actw = min( nbrchf(maxval(abs(x)))+1_i4b , 99_i4b )
    end if
!
    if ( present(s) ) then
        acts = merge( s , defs , s>=0_i4b .and. s<=99_i4b )
    else
        acts = defs
    end if
!
!   CONSTRUCT THE FORMAT.
!
    write( form(03:04), 99 ) abs(actindent) + 1_i4b
!
    write( form(07:08), 99 )         &
            min( max( (actline-abs(actindent)-actw)/(actw+acts), 0_i4b ), 99_i4b )
    write( form(12:13), 99 ) actw
!
    if ( acts/=0_i4b ) then
        write( form(16:17), 99 ) acts
    else
        write( form(14:19), 98 ) blank
    end if
!
    write( form(25:26), 99 ) actw
!
!   SKIP TWO LINES.
!
    write( actunit,"(/)" )
!
!   WRITE TITLE.
!
    if ( present(name) ) then
        if ( actindent>0_i4b ) then
            i = min( int(len_trim(name),i4b) , max(actline-actindent,actw) )
            write( actunit, 98 )  repeat( blank, actindent+1_i4b )//name(1_i4b:i)
        else
            i = min( int(len_trim(name),i4b) , max(actline,actw-actindent) )
            write( actunit, 98 ) blank//name(1_i4b:i)
        end if
    end if
!
!   WRITE VECTOR.
!
    write ( actunit, form ) x
!
! FORMATS
! _______
!
    97   format (a1)
    98   format (a)
    99   format (i2.2)
!
!
! END OF SUBROUTINE write_iv
! __________________________
!
    end subroutine write_iv
!
! =========================================================================================
!
    subroutine print_rm( x, f, w, d, sign_ed, s, title, namlig, namcol, indent, line, prt_unit )
!
! Purpose
! _______
!
!   Routine for labeled real matrix output with given format, as below.
!   The matrix is printed columns block by columns block.
!   Print also a title for the matrix: TITLE
!
!
! Arguments
! _________
!
!   X           (INPUT) real(stnd), dimension(:,:)
!               The matrix to be output.
!
!   F           (INPUT, OPTIONAL) character
!
!   W, D        (INPUT, OPTIONAL) integer(i4b)
!               Selects the edit descriptor: fw.d .
!               Print each entry in format  FW.D .
!               F is a character 'f', 'g', 'e' or 'd',
!               regardless of case.
!               W and D are integers .
!
!   SIGN_ED     (INPUT, OPTIONAL) character(len=2)
!               Selects the sign edit descriptor.
!               sign_ed is 'ss' or 'sp',
!               regardless of case.
!
!   S           (INPUT, OPTIONAL) integer(i4b)
!               the number of spaces between each entry.
!
!   TITLE       (INPUT, OPTIONAL) character(len=*)
!               Prints a title for the matrix.
!
!   NAMLIG      (INPUT, OPTIONAL) character(len=*), dimension(:)
!               Labels for the rows of the matrix. The size of
!               NAMLIG must match the number of ligns of X.
!
!   NAMCOL      (INPUT, OPTIONAL) character(len=*), dimension(:)
!               Labels for the columns of the matrix. The size of
!               NAMCOL must match the number of columns of X.
!
!   IDENT       (INPUT, OPTIONAL) integer(i4b)
!               The number of blanks to use for indentation.
!
!   LINE        (INPUT, OPTIONAL) integer(i4b)
!               The number of characters per line.
!
!   PRT_UNIT    (INPUT, OPTIONAL) integer(i4b)
!               Selects the output unit.
!
!
! Further Details
! _______________
!
!   If NAMLIG is absent, then the rows of X are labeled by row numbers.
!
!   If NAMCOL is absent, then the columns of X are labeled by column numbers.
!   Column labels are truncated to W characters.
!
!   If the value of the argument  INDENT  is positive, then each output
!   line is preceded by  INDENT  blank characters.  If  INDENT  is negative,
!   then only the matrix output will be indented (not the TITLE).
!
!   Each output line from this subroutine will have at most 
!
!           max( LINE, W+abs(INDENT)+6+len_trim(NAMLIG) )
!
!   characters plus one additional leading character for Fortran "carriage control".
!   The carriage control character will always be a blank.
!
!   If LINE is absent, then line width is at most max( 80, W+abs(INDENT)+6+len_trim(NAMLIG) )
!   characters plus one for carriage control.
!
!   All output is on the unit PRT_UNIT.
!   If PRT_UNIT is absent, then all output is on the unit DEFUNIT.
!
!   Defaults are defined for all optional arguments.  See start of module.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,   only : blank
    use Utilities,        only : assert
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),       intent(in)                         :: x(:,:)
!
    integer(i4b),     intent(in), optional               :: prt_unit, w, d, s, indent, line
!
    character,        intent(in), optional               :: f, sign_ed*2
    character(len=*), intent(in), optional               :: title
    character(len=*), intent(in), optional, dimension(:) :: namlig, namcol
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: actunit, actw, actd, acts, actindent, actline, namsize, &
                    m, n, maxcol, nblock, kblock, j1, j2, i
!                 
    CHARACTER(len=35)    :: chr*1, actf*1, actsign_ed*2, form*48, fmt_header
!                 
    logical(lgl) :: test_col, test_lig
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter ::  name_proc ='print_array'
!
    character(len=*),  parameter ::                                  &
       defform ='(**x,i4,1x,a**,1x,sp,**(* **.**, ** x), * **.**)'
!       e.g.     (01x,i4,1x,a10,1x,sp,05(f 09.04, 03 x), f 09.04)
!                123456789012345678901234567890123456789012345678
!
    character(len=*),  parameter ::                                  &
       deffmt_header ='(/**x,6x, **x,**(* **, ** x), * **)'
!       e.g.           (/01x,6x, 10x,05(a 09, 03 x), a 09)
!                      12345678901234567890123456789012345
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   VERIFY INPUT MATRICES.
!
    n = size( x, dim=2 )
    m = size( x, dim=1 )
    if ( n==0_i4b .or. m==0_i4b ) return
!
    test_lig = present( namlig )
    if ( test_lig ) call assert( logical( m==size(namlig), lgl ), name_proc )
!
    test_col = present( namcol )
    if ( test_col ) call assert( logical( n==size(namcol), lgl ), name_proc )
!
!   SET APPROPRIATE DEFAULT VALUES.
!
    if ( test_col ) then
        chr = "a"
    else
        chr = "i"
    end if
!
    form = defform
    fmt_header = deffmt_header
!
    if ( present(prt_unit) ) then
        actunit = merge( prt_unit , defunit , prt_unit>=0_i4b )
    else
        actunit = defunit
    end if
!
    if ( present(line) ) then
        actline = line
    else
        actline = defline
    end if
!
    if ( present(indent) ) then
        actindent = sign( min(98_i4b , abs(indent)) , indent )
    else
        actindent = defindent
    end if
!
    if ( present(f) ) then
        select case( f )
            case( d_maj:g_maj, d_min:g_min )
                actf = f
            case default
                actf = deff
        end select
    else
        actf = deff
    end if
!
    if ( present(w) ) then
        actw = merge( w , defw , w>0_i4b .and. w<=99_i4b )
    else
        actw = defw
    end if
!
    if ( present(d) ) then
        actd = merge( d , defd , d>=0_i4b .and. d<=99_i4b )
    else
        actd = defd
    end if
    select case( actf )
        case( f_maj, f_min )
            actd =  merge( actd , max(0_i4b,actw-2_i4b) , actd<=actw-2_i4b )
        case default
            actd =  merge( actd , max(0_i4b,actw-6_i4b) , actd<=actw-6_i4b )
    end select
!
    if ( present(sign_ed) ) then
        select case( sign_ed(1:1) )
            case( s_maj, s_min )
                select case( sign_ed(2:2) )
                    case( s_maj, s_min, p_maj, p_min )
                        actsign_ed = sign_ed
                    case default
                        actsign_ed = defsign_ed
                end select
            case default
                actsign_ed = defsign_ed
        end select
    else
        actsign_ed = defsign_ed
    end if
!
    if ( present(s) ) then
        acts = merge( s , defs, s>=0_i4b .and. s<=99_i4b )
    else
        acts = defs
    end if
!
!   START CONSTRUCTING THE FORMATS.
!
    write( form(02:03),       99 ) abs(actindent) + 1_i4b
    write( fmt_header(03:04), 99 ) abs(actindent) + 1_i4b
!
    if (test_lig) then
        namsize = min( 99, maxval(len_trim(namlig)) )
        write( form(13:14),       99 ) namsize 
        write( fmt_header(11:12), 99 ) namsize
    else
        namsize = 0_i4b
        write( form(12:15),       98 ) blank
        write( fmt_header(11:14), 98 ) blank
    end if
!
!   DETERMINE THE NUMBER OF COLUMNS OF x TO PRINT ON THE PAGE.
!
    maxcol =                    &
    min( max((actline-(6_i4b+namsize+actw+abs(actindent)))/(actw+acts),0_i4b)+1_i4b , 100_i4b )
!
!   FINISH THE FORMATS.
!
    write( form(19:20), 98 ) actsign_ed
    write( form(22:23), 99 ) maxcol - 1_i4b
    write( form(25:25), 97 ) actf
    write( form(27:28), 99 ) actw
    write( form(30:31), 99 ) actd
!
    write( fmt_header(15:16), 99 ) maxcol - 1_i4b
    write( fmt_header(18:18), 97 ) chr
    write( fmt_header(20:21), 99 ) actw
!
    if ( acts/=0_i4b ) then
        write( form(34:35),       99 ) acts
        write( fmt_header(24:25), 99 ) acts
    else
        write( form(32:37),       98 ) blank
        write( fmt_header(22:27), 98 ) blank
    end if
!
    write( form(41:41), 97 ) actf
    write( form(43:44), 99 ) actw
    write( form(46:47), 99 ) actd
!
    write( fmt_header(31:31), 97 ) chr
    write( fmt_header(33:34), 99 ) actw
!
!   SKIP TWO LINES.
!
    write( actunit,"(/)")
!
!   WRITE TITLE.
!
    if ( present(title) ) then
        if ( actindent>0_i4b ) then
            i = min( int(len_trim(title),i4b) , max(actline-actindent,actw+6_i4b+namsize) )
            write( actunit, 98 )  repeat( blank, actindent+1_i4b )// title(1_i4b:i)
        else
            i = min( int(len_trim(title),i4b) , max(actline,actw+6_i4b+namsize-actindent) )
            write( actunit, 98 ) blank // title(1_i4b:i)
        end if
    end if
!
!   WRITE MATRIX.
!
    nblock = (n+maxcol-1_i4b)/maxcol
    j2 = 0_i4b
    do kblock = 1_i4b, nblock
        j1 = j2 + 1_i4b
        j2 = min( n, j2 + maxcol )
!
!       PRINT HEADER.
!
        if ( test_col ) then
            write( actunit, fmt_header ) namcol(j1:j2)
        else
            write( actunit, fmt_header ) ( i, i=j1, j2)
        end if
!
!       SKIP A LINE.
!
        write( actunit, "()" )
!
!       PRINT DATA.
!
        do i = 1_i4b, m
            if ( test_lig ) then
                write( actunit, form ) i, namlig(i), x(i,j1:j2)
            else
                write( actunit, form ) i, x(i,j1:j2)
            end if
        end do
    end do
!
! FORMATS
! _______
!
    97   format (a1)
    98   format (a)
    99   format (i2.2)
!
!
! END OF SUBROUTINE print_rm
! __________________________
!
    end subroutine print_rm
!
! =========================================================================================
!
    subroutine print_im(x, w, sign_ed, s, title, namlig, namcol, indent, line, prt_unit)
!
! Purpose
! _______
!
!   Routine for labeled integer matrix output with given format, as below.
!   The matrix is printed columns block by columns block.
!   Print also a title for the matrix: TITLE
!
!
! Arguments
! _________
!
!   X           (INPUT) integer(i4b), dimension(:,:)
!               The matrix to be output.
!
!   W           (INPUT, OPTIONAL) integer(i4b)
!               Selects the width of each entry.
!               Print each entry in format  iW .
!
!   SIGN_ED     (INPUT, OPTIONAL) character(len=2)
!               Selects the sign edit descriptor.
!               sign_ed is 'ss' or 'sp',
!               regardless of case.
!
!   S           (INPUT, OPTIONAL) integer(i4b)
!               The number of spaces between each entry.
!
!   TITLE       (INPUT, OPTIONAL) character(len=*)
!               Prints a title for the matrix.
!
!   NAMLIG      (INPUT, OPTIONAL) character(len=*), dimension(:)
!               Labels for the rows of the matrix. The size of
!               NAMLIG must match the number of ligns of X.
!
!   NAMCOL      (INPUT, OPTIONAL) character(len=*), dimension(:)
!               Labels for the columns of the matrix. The size of
!               NAMCOL must match the number of columns of X.
!
!   IDENT       (INPUT, OPTIONAL) integer(i4b)
!               The number of blanks to use for indentation.
!
!   LINE        (INPUT, OPTIONAL) integer(i4b)
!               The number of characters per line.
!
!   PRT_UNIT    (INPUT, OPTIONAL) integer(i4b)
!               Selects the output unit.
!
!
! Further Details
! _______________
!
!   If W is absent, then the routine determines the best width w needed to edit 
!   the array X without excess blanks.
!
!   If NAMLIG is absent, then the rows of X are labeled by row numbers.
!
!   If NAMCOL is absent, then the columns of X are labeled by column numbers.
!   Column labels are truncated to W characters.
!
!   If the value of the argument  INDENT  is positive, then each output
!   line is preceded by  INDENT  blank characters.  If  INDENT  is negative,
!   then only the matrix output will be indented (not the TITLE).
!
!   Each output line from this subroutine will have at most 
!
!           max( LINE, W+abs(INDENT)+6+len_trim(NAMLIG) )
!
!   characters plus one additional leading character for Fortran "carriage control".
!   The carriage control character will always be a blank.
!
!   If LINE is absent, then line width is at most max( 80, W+abs(INDENT)+6+len_trim(NAMLIG) )
!   characters plus one for carriage control.
!
!   All output is on the unit PRT_UNIT.
!   If PRT_UNIT is absent, then all output is on the unit DEFUNIT.
!
!   Defaults are defined for all optional arguments.  See start of module.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,    only : blank
    use Utilities,         only : assert
    use String_Procedures, only : nbrchf
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b),     intent(in)                         :: x(:,:)
    integer(i4b),     intent(in), optional               :: prt_unit, w, s, indent, line
!
    character(len=2), intent(in), optional               :: sign_ed
    character(len=*), intent(in), optional               :: title
    character(len=*), intent(in), optional, dimension(:) :: namlig, namcol
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: actunit, actw, acts, actindent, actline, namsize, &
                    m, n, maxcol, nblock, kblock, j1, j2, i
!                 
    character(len=35) :: chr*1, actsign_ed*2, form*42, fmt_header
!                 
    logical(lgl) :: test_col, test_lig
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter ::    name_proc ='print_array'
!
    character(len=*),  parameter ::                                &
           defform ='(**x,i4,1x,a**,1x,sp,**(i **, ** x), i **)'
!   e.g.             (01x,i4,1x,a10,1x,sp,05(i 09, 03 x), i 09)
!                    123456789012345678901234567890123456789012
!
    character(len=*),  parameter ::                                &
           deffmt_header ='(/**x,6x, **x,**(* **, ** x), * **)'
!   e.g.                   (/01x,6x, 10x,05(a 09, 03 x), a 09)
!                          12345678901234567890123456789012345
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   VERIFY INPUT MATRICES.
!
    n = size( x, dim=2 )
    m = size( x, dim=1 )
    if ( n==0_i4b .or. m==0_i4b ) return
!
    test_lig = present( namlig )
    if ( test_lig ) call assert( logical( m==size(namlig), lgl ), name_proc )
!
    test_col = present( namcol )
    if ( test_col ) call assert( logical( n==size(namcol), lgl ), name_proc )
!
!   SET APPROPRIATE DEFAULT VALUES.
!
    if ( test_col ) then
        chr = "a"
    else
        chr = "i"
    end if
!
    form = defform
    fmt_header = deffmt_header
!
    if ( present(prt_unit) ) then
        actunit = merge( prt_unit , defunit , prt_unit>=0_i4b )
    else
        actunit = defunit
    end if
!
    if ( present(line) ) then
        actline = line
    else
        actline = defline
    end if
!
    if ( present(indent) ) then
        actindent = sign( min(98_i4b , abs(indent)) , indent )
    else
        actindent = defindent
    end if
!
    if ( present(w) ) then
        if ( w>0_i4b .and. w<=99_i4b ) then
            actw = w
        else
            actw = min( nbrchf(maxval(abs(x)))+1_i4b , 99_i4b )
        end if
    else
        actw = min( nbrchf(maxval(abs(x)))+1_i4b , 99_i4b )
    end if
!
    if ( present(sign_ed) ) then
        select case( sign_ed(1:1) )
            case( s_maj, s_min )
                select case( sign_ed(2:2) )
                    case( s_maj, s_min, p_maj, p_min )
                        actsign_ed = sign_ed
                    case default
                        actsign_ed = defsign_ed
                end select
            case default
                actsign_ed = defsign_ed
        end select
    else
        actsign_ed = defsign_ed
    end if
!
    if ( present(s) ) then
        acts = merge( s , defs, s>=0_i4b .and. s<=99_i4b )
    else
        acts = defs
    end if
!
!   START CONSTRUCTING THE FORMATS.
!
    write( form(02:03),       99 ) abs(actindent) + 1_i4b
    write( fmt_header(03:04), 99 ) abs(actindent) + 1_i4b
!
    if ( test_lig ) then
        namsize = min( 99, maxval(len_trim(namlig)) )
        write( form(13:14),       99 ) namsize 
        write( fmt_header(11:12), 99 ) namsize
    else
        namsize = 0_i4b
        write( form(12:15),       98 ) blank
        write( fmt_header(11:14), 98 ) blank
    end if
!
!   DETERMINE THE NUMBER OF COLUMNS OF x TO PRINT ON THE PAGE.
!
    maxcol =                    &
    min( max((actline-(6_i4b+namsize+actw+abs(actindent)))/(actw+acts),0_i4b)+1_i4b , 100_i4b )
!
!   FINISH THE FORMATS.
!
    write( form(19:20), 98 ) actsign_ed
    write( form(22:23), 99 ) maxcol - 1_i4b
    write( form(27:28), 99 ) actw
!
    write( fmt_header(15:16), 99 ) maxcol - 1_i4b
    write( fmt_header(18:18), 97 ) chr
    write( fmt_header(20:21), 99 ) actw
!
    if ( acts/=0_i4b ) then
        write( form(31:32),       99 ) acts
        write( fmt_header(24:25), 99 ) acts
    else
        write( form(29:34),       98 ) blank
        write( fmt_header(22:27), 98 ) blank
    end if
!
    write( form(40:41), 99 ) actw
!
    write( fmt_header(31:31), 97 ) chr
    write( fmt_header(33:34), 99 ) actw
!
!   SKIP TWO LINES.
!
    write( actunit, "(/)" )
!
!   WRITE TITLE.
!
    if ( present(title) ) then
        if ( actindent>0_i4b ) then
            i = min( int(len_trim(title),i4b) , max(actline-actindent,actw+6_i4b+namsize) )
            write( actunit, 98 )  repeat( blank, actindent+1_i4b )// title(1_i4b:i)
        else
            i = min( int(len_trim(title),i4b) , max(actline,actw+6_i4b+namsize-actindent) )
            write( actunit, 98 ) blank // title(1_i4b:i)
        end if
    end if
!
!   WRITE MATRIX.
!
    nblock = (n+maxcol-1_i4b)/maxcol
    j2 = 0_i4b
    do kblock = 1_i4b, nblock
        j1 = j2 + 1_i4b
        j2 = min( n, j2 + maxcol )
!
!       PRINT HEADER.
!
        if ( test_col ) then
            write( actunit, fmt_header ) namcol(j1:j2)
        else
            write( actunit, fmt_header ) ( i, i=j1, j2)
        end if
!
!       SKIP A LINE.
!
        write( actunit, "()" )
!
!       PRINT DATA.
!
        do i = 1_i4b, m
            if ( test_lig ) then
                write( actunit, form ) i, namlig(i), x(i,j1:j2)
            else
                write( actunit, form ) i, x(i,j1:j2)
            end if
        end do
    end do
!
! FORMATS
! _______
!
    97   format (a1)
    98   format (a)
    99   format (i2.2)
!
!
! END OF SUBROUTINE print_im
! __________________________
!
    end subroutine print_im
!
! =========================================================================================
!
    subroutine print_rv( x, f, w, d, sign_ed, title, namlig, indent, prt_unit )
!
! Purpose
! _______
!
!   Routine for labeled real vector output with given format, as below.
!
!   Print also a title for the vector: TITLE
!
!
! Arguments
! _________
!
!   X           (INPUT) real(stnd), dimension(:)
!               The vector to be output.
!
!   F           (INPUT, OPTIONAL) character
!
!   W, D        (INPUT, OPTIONAL) integer(i4b)
!               Selects the edit descriptor: fw.d .
!               Print each entry in format  FW.D .
!               F is a character 'f', 'g', 'e' or 'd',
!               regardless of case.
!               W and D are integers .
!
!   SIGN_ED     (INPUT, OPTIONAL) character(len=2)
!               Selects the sign edit descriptor.
!               sign_ed is 'ss' or 'sp'.
!
!   TITLE       (INPUT, OPTIONAL) character(len=*)
!               Prints a title for the vector.
!
!   NAMLIG      (INPUT, OPTIONAL) character(len=*), dimension(:)
!               Labels for the elements of the vector. The size of
!               NAMLIG must match the size of X.
!
!   IDENT       (INPUT, OPTIONAL) integer(i4b)
!               The number of blanks to use for indentation.
!
!   PRT_UNIT    (INPUT, OPTIONAL) integer(i4b)
!               Selects the output unit.
!
!
! Further Details
! _______________
!
!   If NAMLIG is absent, then the rows of X are labeled by row numbers.
!
!   If the value of the argument  INDENT  is positive, then each output
!   line is preceded by  INDENT  blank characters.  If  INDENT  is negative,
!   then only the vector output will be indented (not the TITLE).
!
!   Each output line from this subroutine will have at most 
!
!           abs(INDENT) + max( len(TITLE), W+6+len_trim(NAMLIG) )
!
!   characters plus one additional leading character for Fortran "carriage control".
!   The carriage control character will always be a blank.
!
!   All output is on the unit PRT_UNIT.
!   If PRT_UNIT is absent, then all output is on the unit DEFUNIT.
!
!   Defaults are defined for all optional arguments.  See start of module.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,   only : blank
    use Utilities,        only : assert
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),       intent(in)           :: x(:)
!
    integer(i4b),     intent(in), optional :: w, d, indent, prt_unit
!
    character,        intent(in), optional :: f, sign_ed*2
    character(len=*), intent(in), optional :: namlig(:), title
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: actunit, actw, actd, actindent, m, i
!                 
    character(len=32) :: actf*1, actsign_ed*2, form
!                 
    logical(lgl) :: test_lig
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter ::                     &
        defform ='(**x,i4,1x,a**,1x,**,* **.**,ss)'
!   e.g.          (02x,i4,1x,a10,1x,sp,f 09.04,ss)
!                 12345678901234567890123456789012
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   VERIFY INPUT VECTORS.
!
    m = size( x )
    if ( m==0_i4b ) return
!
    test_lig = present( namlig )
    if ( test_lig ) call assert( logical( m==size(namlig), lgl ), 'print_array' )
!
!   SET APPROPRIATE DEFAULT VALUES.
!
    form = defform
!
    if ( present(prt_unit) ) then
        actunit = merge( prt_unit , defunit, prt_unit>=0_i4b )
    else
        actunit = defunit
    end if
!
    if ( present(indent) ) then
        actindent = sign( min(98_i4b,abs(indent)) , indent )
    else
        actindent = defindent
    end if
!
    if ( present(f) ) then
        select case( f )
            case( d_maj:g_maj, d_min:g_min )
                actf = f
            case default
                actf = deff
        end select
    else
        actf = deff
    end if
!
    if ( present(w) ) then
        actw = merge( w , defw , w>0_i4b .and. w<=99_i4b )
    else
        actw = defw
    end if
!
    if ( present(d) ) then
        actd = merge( d , defd , d>=0_i4b .and. d<=99_i4b )
    else
        actd = defd
    end if
    select case( actf )
        case( f_maj, f_min )
            actd =  merge( actd , max(0_i4b,actw-2_i4b) , actd<=actw-2_i4b )
        case default
            actd =  merge( actd , max(0_i4b,actw-6_i4b) , actd<=actw-6_i4b )
    end select
!
    if ( present(sign_ed) ) then
        select case( sign_ed(1:1) )
            case( s_maj, s_min )
                select case( sign_ed(2:2) )
                    case( s_maj, s_min, p_maj, p_min )
                        actsign_ed = sign_ed
                    case default
                        actsign_ed = defsign_ed
                end select
            case default
                actsign_ed = defsign_ed
        end select
    else
        actsign_ed = defsign_ed
    end if
!
!   CONSTRUCT THE FORMAT.
!
    write ( form(02:03), 99 ) abs(actindent) + 1_i4b
!
    if ( test_lig ) then
        write( form(13:14), 99 ) min( 99, maxval(len_trim(namlig)) ) 
    else
        write( form(12:15), 98 ) blank
    end if
!
    write ( form(19:20), 98 ) actsign_ed
    write ( form(22:22), 97 ) actf
    write ( form(24:25), 99 ) actw
    write ( form(27:28), 99 ) actd
!
!   SKIP TWO LINES.
!
    write( actunit, "(/)" )
!
!   WRITE TITLE.
!
    if ( present(title) ) then
        if ( actindent>0_i4b )                    &
        write( actunit, 98, advance='no' )  repeat( blank, actindent )
        write( actunit, 98 ) blank // title
        write( actunit, "()" )
    end if
!
!   WRITE VECTOR.
!
    if ( test_lig ) then
        write( actunit, form ) ( i, namlig(i), x(i), i=1_i4b, m )
    else
        write( actunit, form ) ( i, x(i), i=1_i4b, m )
    end if
!
! FORMATS
! _______
!
    97   format (a1)
    98   format (a)
    99   format (i2.2)
!
!
! END OF SUBROUTINE print_rv
! __________________________
!
    end subroutine print_rv
!
! =========================================================================================
!
    subroutine print_iv( x, w, sign_ed, title, namlig, indent, prt_unit )
!
! Purpose
! _______
!
!   Routine for labeled integer vector output with given format, as below.
!
!   Print also a title for the vector: TITLE
!
!
! Arguments
! _________
!
!   X           (INPUT) integer(i4b), dimension(:)
!               The vector to be output.
!
!   W           (INPUT, OPTIONAL) integer(i4b)
!               Selects the width of each entry.
!               Print each entry in format  iW .
!
!   SIGN_ED     (INPUT, OPTIONAL) character(len=2)
!               Selects the sign edit descriptor.
!               sign_ed is 'ss' or 'sp',
!               regardless of case.
!
!   TITLE       (INPUT, OPTIONAL) character(len=*)
!               Prints a title for the vector.
!
!   NAMLIG      (INPUT, OPTIONAL) character(len=*), dimension(:)
!               Labels for the elements of the vector. The size of
!               NAMLIG must match the size of X.
!
!   IDENT       (INPUT, OPTIONAL) integer(i4b)
!               The number of blanks to use for indentation.
!
!   PRT_UNIT    (INPUT, OPTIONAL) integer(i4b)
!               Selects the output unit.
!
!
! Further Details
! _______________
!
!   If W is absent, then the routine determines the best width w needed to edit 
!   the array X without excess blanks.
!
!   If NAMLIG is absent, then the rows of X are labeled by row numbers.
!
!   If the value of the argument  INDENT  is positive, then each output
!   line is preceded by  INDENT  blank characters.  If  INDENT  is negative,
!   then only the vector output will be indented (not the TITLE).
!
!   Each output line from this subroutine will have at most 
!
!           abs(INDENT) + max( len(TITLE), W+6+len_trim(NAMLIG) )
!
!   characters plus one additional leading character for Fortran "carriage control".
!   The carriage control character will always be a blank.
!
!   All output is on the unit PRT_UNIT.
!   If PRT_UNIT is absent, then all output is on the unit DEFUNIT.
!
!   Defaults are defined for all optional arguments.  See start of module.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,    only : blank
    use Utilities,         only : assert
    use String_Procedures, only : nbrchf
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b),     intent(in)           :: x(:)
    integer(i4b),     intent(in), optional :: w, indent, prt_unit
!
    character(len=2), intent(in), optional :: sign_ed
    character(len=*), intent(in), optional :: namlig(:), title
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: actunit, actw, actindent, m, i
!                 
    character(len=2) :: actsign_ed, form*29
!                 
    logical(lgl) :: test_lig
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter ::                    &
          defform ='(**x,i4,1x,a**,1x,**,i **,ss)'
!   e.g.            (02x,i4,1x,a10,1x,sp,i 09,ss)
!                   12345678901234567890123456789
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   VERIFY INPUT VECTORS.
!
    m=size( x )
    if ( m==0_i4b ) return
!
    test_lig = present( namlig )
    if ( test_lig ) call assert( logical( m==size(namlig), lgl ), 'print_array' )
!
!   SET APPROPRIATE DEFAULT VALUES.
!
    form = defform
!
    if ( present(prt_unit) ) then
        actunit = merge( prt_unit , defunit, prt_unit>=0_i4b )
    else
        actunit = defunit
    end if
!
    if ( present(indent) ) then
        actindent = sign( min(98_i4b,abs(indent)) , indent)
    else
        actindent = defindent
    end if
!
    if ( present(w) ) then
        if ( w>0_i4b .and. w<=99_i4b ) then
            actw = w
        else
            actw = min( nbrchf(maxval(abs(x)))+1_i4b , 99_i4b )
        end if
    else
        actw = min( nbrchf(maxval(abs(x)))+1_i4b , 99_i4b )
    end if
!
    if ( present(sign_ed) ) then
        select case( sign_ed(1:1) )
            case( s_maj, s_min )
                select case( sign_ed(2:2) )
                    case( s_maj, s_min, p_maj, p_min )
                        actsign_ed = sign_ed
                    case default
                        actsign_ed = defsign_ed
                end select
            case default
                actsign_ed = defsign_ed
        end select
    else
        actsign_ed = defsign_ed
    end if
!
!   CONSTRUCT THE FORMAT.
!
    write( form(02:03), 99 ) abs(actindent) + 1_i4b
!
    if ( test_lig ) then
        write( form(13:14), 99 ) min( 99, maxval(len_trim(namlig)) ) 
    else
        write( form(12:15), 98 ) blank
    end if
!
    write ( form(19:20), 98 ) actsign_ed
    write ( form(24:25), 99 ) actw
!
!   SKIP TWO LINES.
!
    write( actunit, "(/)" )
!
!   WRITE TITLE.
!
    if ( present(title) ) then
        if ( actindent>0_i4b )                    &
        write( actunit, 98, advance='no' )  repeat( blank, actindent )
        write( actunit, 98 ) blank // title
        write( actunit, "()" )
    end if
!
!   WRITE VECTOR.
!
    if ( test_lig ) then
        write( actunit, form ) ( i, namlig(i), x(i), i=1_i4b, m )
    else
        write( actunit, form ) ( i, x(i), i=1_i4b, m )
    end if
!
! FORMATS
! _______
!
    97   format (a1)
    98   format (a)
    99   format (i2.2)
!
!
! END OF SUBROUTINE print_iv
! __________________________
!
    end subroutine print_iv
!
! =========================================================================================
!
    subroutine print_prinfac_rm( mode, a, f, names, line, prt_unit )
!
! Purpose
! _______
!
!               
!   Routine for labeled matrix output after an EOF or SVD analysis.
!
!   Print an EOF model (MODE=1) or the associated principal components (MODE=2) and
!   an SVD model (MODE=3) or the associated singular variables (MODE=4)
!
!
! Arguments
! _________
!
!   MODE        (INPUT) integer(i1b)
!               integer indicator for printing.
!
!   A           (INPUT) real(stnd), dimension(:,:)
!               The matrix to be output.
!
!   F           (INPUT, OPTIONAL) character
!               Selects the edit descriptor.
!               f is a character 'f', 'g', 'e' or 'd',
!               regardless of case.
!               Print each entry in format  f14.6 .
!
!   NAMES       (INPUT, OPTIONAL) character(len=*), dimension(:)
!               Labels for the rows of the matrix. The size of NAMES
!               must match the number of rows of A.
!
!   LINE        (INPUT, OPTIONAL) integer(i4b)
!               The number of characters per line.
!
!   PRT_UNIT    (INPUT, OPTIONAL) integer(i4b)
!               Selects the output unit.
!
!
! Further Details
! _______________
!
!   If F is absent, then the default edit descriptor is DEFF.
!
!   If NAMES is absent, then the rows of a are labeled by row numbers.
!
!   Each output line from this subroutine will have at least 20+len_trim(NAMES)
!   characters (print at least one column of a) and at most 118+len_trim(NAMES)  
!   characters (print at most eight columns of a) plus one additional leading
!   character for Fortran "carriage control". The carriage control character
!   will always be a blank.
!
!   If LINE is absent, then line width is 80 characters.
!
!   If PRT_UNIT is absent, then all output is on the unit DEFUNIT.
!
!   Defaults are defined for all optional arguments.  See start of module.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,   only : blank
    use Utilities,        only : assert
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),                 intent(in) :: a(:,:)
!
    integer(i1b),               intent(in) :: mode
    integer(i4b),     optional, intent(in) :: line, prt_unit
!
    character,        optional, intent(in) :: f 
    character(len=*), optional, intent(in) :: names(:) 
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: m, n, namsize, maxcol, nblock, kblock, j1, j2, i,                   &
                    actline, actunit, mode_tmp
!
    logical(lgl) :: test_names
!
    character(len=27) ::  fmt_header, form, actf*1
    character(len=29) ::  title2
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter ::                                                      &
       deffmt_header='(/7x,**x,8(5x,a4,i3.3,2x)/)'
!      e.g.           (/7x,10x,8(5x,a4,i3.3,2x)/)
!                     123456789012345678901234567
!      
    character(len=*),  parameter ::                                                      &
       defform =     '(1x,i4,1x,a**,1x,sp,8*14.6)'
!      e.g.           (1x,i4,1x,a10,1x,sp,8f14.6)
!                     123456789012345678901234567
!      
    character(len=*),  parameter ::                                                       & 
      head(4)=(/" fac","prin","svec","svar"/),                                            &
      title(4)=(/ "(//5x,'FACTORS             ')" , "(//5x,'PRINCIPAL COMPONENTS')" ,     &
                  "(//5x,'SINGULAR VECTORS    ')" , "(//5x,'SINGULAR VARIABLES  ')"      /)
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   VERIFY INPUT ARGUMENTS.
!
    n = min( size( a, dim=2 ), 999_i4b )
    m = size( a, dim=1 )
    if ( n==0_i4b .or. m==0_i4b ) return
!
    test_names = present( names )
    if ( test_names ) call assert( logical( m==size(names), lgl ), 'print_prinfac' )
!
    mode_tmp = merge( mode, 1_i1b, mode>=1_i1b .or. mode<=4_i1b )
!
    if ( present(f) ) then
        select case( f )
            case( d_maj:g_maj, d_min:g_min )
                actf = f
            case default
                actf = deff
        end select
    else
        actf = deff
    end if
!
    if ( present(line) ) then
        actline = line
    else
        actline = defline
    end if
!
    if ( present(prt_unit) ) then
        actunit = merge( prt_unit, defunit, prt_unit>=0_i4b )
    else
        actunit = defunit
    end if
!
!   CONSTRUCT FORMATS FOR THE HEADER AND THE DATA.
!
    form        = defform
    fmt_header  = deffmt_header
!
    if ( test_names ) then
        namsize = min( 99, maxval(len_trim(names)) )
        write( form(12:13),     99 ) namsize 
        write( fmt_header(6:7), 99 ) namsize
    else
        namsize = 0_i4b
        write( form(11:14),     98 ) blank
        write( fmt_header(6:9), 98 ) blank
    end if
!
    write( form(22:22), 97 ) actf
!
!   WRITE TITLE.
!
    title2 = title(mode_tmp)
    write( actunit, title2 )
!
!   DETERMINE THE NUMBER OF COLUMNS OF A TO PRINT ON THE PAGE.
!
    maxcol = max( 1_i4b , min( 8_i4b, (actline-6_i4b-namsize)/14_i4b ) )
!
!   PRINT THE MATRIX.
!
    nblock = (n+maxcol-1_i4b)/maxcol
    j2 = 0_i4b
    do kblock = 1_i4b, nblock
        j1 = j2 + 1_i4b
        j2 = min( n, j2+maxcol )
        write( actunit, fmt_header ) ( head(mode_tmp), i, i=j1, j2 )
        write( actunit, '(1x)' )
        do i = 1_i4b, m
            if ( test_names ) then
                write( actunit, form ) i, names(i), a(i,j1:j2)
            else
                write( actunit, form ) i, a(i,j1:j2)
            end if
        end do
    end do
!
! FORMATS
! _______
!
    97   format (a1)
    98   format (a)
    99   format (i2.2)
!
!
! END OF SUBROUTINE print_prinfac_rm
! __________________________________
!
    end subroutine print_prinfac_rm
!
! =========================================================================================
!
    subroutine print_prinfac_rv( mode, a, f, names, prt_unit )
!
! Purpose
! _______
!
!               
!   Routine for labeled matrix output after an EOF or SVD analysis.
!
!   Print an EOF vector (MODE=1) or the associated principal component (MODE=2) and
!   a singular vector (MODE=3) or the associated singular variable (MODE=4)
!
!
! Arguments
! _________
!
!   MODE        (INPUT) integer(i1b)
!               integer indicator for printing.
!
!   A           (INPUT) real(stnd), dimension(:)
!               The vector to be output.
!
!   F           (INPUT, OPTIONAL) character
!               Selects the edit descriptor.
!               F is a character 'f', 'g', 'e' or 'd',
!               regardless of case.
!               Print each entry in format  F14.6 .
!
!   NAMES       (INPUT, OPTIONAL) character(len=*), dimension(:)
!               Labels for the elements of the vector. The size of NAMES
!               must match the size of A.
!
!   PRT_UNIT    (INPUT, OPTIONAL) integer(i4b)
!               Selects the output unit.
!
!
! Further Details
! _______________
!
!   If F is absent, then the default edit descriptor is DEFF.
!
!   If NAMES is absent, then the elements of A are labeled by element numbers.
!
!   Each output line from this subroutine will have 20+len_trim(NAMES) characters
!   plus one additional leading  character for Fortran "carriage control".
!   The carriage control character will always be a blank.
!
!   If PRT_UNIT is absent, then all output is on the unit DEFUNIT.
!
!   Defaults are defined for all optional arguments.  See start of module.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,   only : blank
    use Utilities,        only : assert
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),                 intent(in) :: a(:)
!
    integer(i1b),               intent(in) :: mode
    integer(i4b),     optional, intent(in) :: prt_unit
!
    character,        optional, intent(in) :: f 
    character(len=*), optional, intent(in) :: names(:) 
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: actunit, mode_tmp, n, i
!
    logical(lgl) :: test_names
!
    character(len=29) :: form, fmt_header*14, actf*1
    character(len=28) :: title2
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter ::                                                     &
       deffmt_header='(/12x,**x,a5/)'
!      e.g.           (/12x,10x,a5/)
!                     12345678901234
!
    character(len=*),  parameter ::                                                     &
       defform =     '(1x,i4,1x,a**,1x,sp,*14.6,ss)'
!      e.g.           (1x,i4,1x,a10,1x,sp,f14.6,ss)
!                     12345678901234567890123456789
!      
    character(len=*),  parameter ::                                                     &
      title(4)=(/ "(//5x,'FACTOR             ')" , "(//5x,'PRINCIPAL COMPONENT')" ,     &
                  "(//5x,'SINGULAR VECTOR    ')" , "(//5x,'SINGULAR VARIABLE  ')"   /), &
      head(4)=(/" fac1","prin1","svec1","svar1"/)
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   VERIFY INPUT ARGUMENTS.
!
    n = size( a )
    if ( n==0_i4b ) return
!
    test_names = present( names )
    if ( test_names ) call assert( logical( n==size(names), lgl ), 'print_prinfac' )
!
    mode_tmp = merge( mode, 1_i1b, mode>=1_i1b .or. mode<=4_i1b )
!
    if ( present(f) ) then
        select case( f )
            case( d_maj:g_maj, d_min:g_min )
                actf = f
            case default
                actf = deff
        end select
    else
        actf = deff
    end if
!
    if ( present(prt_unit) ) then
        actunit = merge( prt_unit, defunit, prt_unit>=0_i4b )
    else
        actunit = defunit
    end if
!
!   CONSTRUCT FORMATS FOR THE HEADER AND THE DATA.
!
    form        = defform
    fmt_header  = deffmt_header
!
    if ( test_names ) then
        i = min( 99, maxval(len_trim(names)) )
        write( form(12:13),     99 ) i 
        write( fmt_header(7:8), 99 ) i
    else
        write( form(11:14),     98 ) blank
        write( fmt_header(6:9), 98 ) blank
    end if
!
    write( form(21:21), 97 ) actf
!
!   WRITE TITLE.
!
    title2 = title(mode_tmp)
    write( actunit, title2 )
!
!   WRITE HEADER.
!
    write( actunit, fmt_header )  head(mode_tmp)
!
!   WRITE VECTOR.
!
    if ( test_names ) then
        write( actunit, form ) ( i, names(i), a(i), i=1_i4b, n )
    else
        write( actunit, form ) ( i, a(i), i=1_i4b, n )
    end if
!
! FORMATS
! _______
!
    97   format (a1)
    98   format (a)
    99   format (i2.2)
!
!
! END OF SUBROUTINE print_prinfac_rv
! __________________________________
!
    end subroutine print_prinfac_rv
!
! =========================================================================================
!
    subroutine print_stat_missing( mode, nomiss, var, inr, qlt, names, prt_unit )
!
! Purpose
! _______
!               
!   Print statistics for an EOF "missing" analysis for
!
!       Variables    (MODE=1)
!       Observations (MODE=2)
!
!
! Arguments
! _________
!
!   MODE            (INPUT) integer(i1b)
!                   integer indicator for printing.
!
!   NOMISS          (INPUT) integer(i4b), dimension(:)
!                   Vector giving the number of non-missing elements
!                   for each variable (MODE=1) or observation (MODE=2).
!
!   VAR, INR, QLT   (INPUT) real(stnd), dimension(:)
!                   The statistics to be output for each variable (MODE=1)
!                   or observation (MODE=2). The size of these vectors must
!                   match the size of NOMISS.
!
!   NAMES           (INPUT, OPTIONAL) character(len=*), dimension(:)
!                   Labels for the  variables(MODE=1) or observations (MODE=2).
!                   The size of this vector must match the size of NOMISS.
!
!   PRT_UNIT        (INPUT, OPTIONAL) integer(i4b)
!                   Selects the output unit.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,   only : blank
    use Utilities,        only : assert_eq, assert
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
   integer(i1b),               intent(in) :: mode
   integer(i4b),               intent(in) :: nomiss(:)
   integer(i4b), optional,     intent(in) :: prt_unit
!
   real(stnd), dimension(:),   intent(in) :: var, inr, qlt
!
   character(len=*), optional, intent(in) :: names(:) 
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
   integer(i4b)  :: actunit, m, namsize, i
   integer(i1b)  :: mode_tmp
!
   character(42) :: fmt='(1x,i4,1x,a00,1x,i6,1x,g15.9,2(1x,2pf8.3))'     
   character(53) :: fmt_header="(/7x,00x,'Nomiss    Variance       Inr(%)   Qlt(%)'/)"
   character(37) :: title2
!
   logical(lgl)  :: test_names
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter ::                                                   &
    name_proc ='print_stat',                                                          &
    title(2)=(/"(//20x,'STATISTICS FOR VARIABLES   ')","(//20x,'STATISTICS FOR OBSERVATIONS')"/)
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   VERIFY INPUT ARGUMENTS.
!
    m = assert_eq( int(size(nomiss),i4b), int(size(var),i4b),              &
                   int(size(inr),i4b),    int(size(qlt),i4b), name_proc )
    if ( m == 0_i4b ) return
!
    test_names = present( names )
    if ( test_names ) call assert( logical( m==size(names), lgl ), name_proc )
!
    mode_tmp = merge( mode, 1_i1b, mode==1_i1b .or. mode==2_i1b )
!
    if ( present(prt_unit) ) then
        actunit = merge( prt_unit, defunit, prt_unit>=0_i4b )
    else
        actunit = defunit
    end if
!
!   FINISH TO CONSTRUCT THE FORMATS FOR HEADER AND DATA.
!
    if ( test_names ) then
        namsize = maxval( len_trim(names) )
    else
        namsize = 1_i4b
    end if
    write( fmt_header(6:7),'(i2.2)' ) namsize
    write( fmt(12:13),     '(i2.2)' ) namsize
!
!   WRITE TITLE AND HEADER.
!
    title2 = title(mode_tmp)
    write( actunit, title2 )
    write( actunit, fmt_header      )
!
!   PRINT THE STATISTICS.
!
    do i = 1_i4b, m
        if ( test_names ) then
            write( actunit, fmt ) i, names(i), nomiss(i), var(i), inr(i), qlt(i)
        else
            write( actunit, fmt ) i, blank   , nomiss(i), var(i), inr(i), qlt(i)
        end if
    end do
!
!
! END OF SUBROUTINE print_stat_missing
! ____________________________________
!
    end subroutine print_stat_missing
!
! =========================================================================================
!
    subroutine print_stat_weight( mode, weight, var, inr, qlt, names, prt_unit ) 
!
! Purpose
! _______
!
!               
!   Print statistics for an EOF "weighted" analysis for
!
!       Variables    (MODE=1)
!       Observations (MODE=2)
!
!
! Arguments
! _________
!
!   MODE            (INPUT) integer(i1b)
!                   integer indicator for printing.
!
!   WEIGHT          (INPUT) real(stnd), dimension(:)
!                   Vector giving the weights
!                   for each variable (MODE=1) or observation (MODE=2).
!
!   VAR, INR, QLT   (INPUT) real(stnd), dimension(:)
!                   The statistics to be output for each variable (MODE=1)
!                   or observation (MODE=2). The size of these vectors must
!                   match the size of WEIGHT.
!
!   NAMES           (INPUT, OPTIONAL) character(len=*), dimension(:)
!                   Labels for the  variables(MODE=1) or observations (MODE=2).
!                   The size of this vector must match the size of WEIGHT.
!
!   PRT_UNIT        (INPUT, OPTIONAL) integer(i4b)
!                   Selects the output unit.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,   only : blank
    use Utilities,        only : assert, assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i1b),               intent(in) :: mode
    integer(i4b), optional,     intent(in) :: prt_unit
!
    real(stnd),                 intent(in) :: weight(:)
    real(stnd), dimension(:),   intent(in) :: var, inr, qlt
!
    character(len=*), optional, intent(in) :: names(:) 
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b)  :: actunit, m, namsize, i
    integer(i1b)  :: mode_tmp
!
    character(58) :: fmt_header="(/7x,00x,'  Weight       Variance       Inr(%)   Qlt(%)'/)"
    character(45) :: fmt='(1x,i4,1x,a00,1x,g12.6,1x,g14.9,2(1x,2pf8.3))'
    character(37) :: title2
!
    logical(lgl)  :: test_names
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter ::                                                   &
    name_proc ='print_stat',                                                          &
    title(2)=(/"(//20x,'STATISTICS FOR VARIABLES   ')","(//20x,'STATISTICS FOR OBSERVATIONS')"/)
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   VERIFY INPUT ARGUMENTS.
!
    m = assert_eq( int(size(weight),i4b), int(size(var),i4b),              &
                   int(size(inr),i4b),    int(size(qlt),i4b), name_proc )
    if ( m==0_i4b ) return
!
    test_names = present( names )
    if ( test_names ) call assert( logical( m==size(names), lgl ), name_proc )
!
    mode_tmp = merge( mode, 1_i1b, mode==1_i1b .or. mode==2_i1b )
!
    if ( present(prt_unit) ) then
        actunit = merge( prt_unit, defunit, prt_unit>=0_i4b )
    else
        actunit = defunit
    end if
!
!   FINISH TO CONSTRUCT THE FORMATS FOR HEADER AND DATA.
!
    if ( test_names ) then
        namsize = maxval( len_trim(names) )
    else
        namsize = 1_i4b
    end if
    write( fmt_header(6:7), '(i2.2)' ) namsize
    write( fmt(12:13),      '(i2.2)' ) namsize
!
!   WRITE TITLE AND HEADER.
!
    title2 = title(mode_tmp)
    write( actunit, title2 )
    write( actunit, fmt_header      )
!
!   PRINT THE STATISTICS.
!
    do i = 1_i4b, m
        if ( test_names ) then
            write( actunit, fmt ) i, names(i), weight(i), var(i), inr(i), qlt(i)
        else
            write( actunit, fmt ) i, blank,    weight(i), var(i), inr(i), qlt(i)
        end if
    end do
!
!
! END OF SUBROUTINE print_stat_weight
! ___________________________________
!
    end subroutine print_stat_weight
!
! =========================================================================================
!
!
! ********************************
! END OF MODULE Print_Procedures *
! ********************************
!    
end module Print_Procedures
