module QR_Procedures
! ###################################################################################
!
! Copyright 2021 IRD
!
! This file is part of statpack.
!
! statpack is free software: you can redistribute it and/or modify
! it under the terms of the GNU Lesser General Public License as
! published by the Free Software Foundation, either version 3 of 
! the License, or (at your option) any later version.
!
! statpack is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU Lesser General Public License for more details.
!
! You can find a copy of the GNU Lesser General Public License
! in the statpack/doc directory.
!
! ####################################################################################
!                                                                                    *
! ************************************************************************************
! MODULE EXPORTING SUBROUTINES AND FUNCTIONS FOR COMPUTING QR, LQ DECOMPOSITIONS AND *
! RELATED DECOMPOSITIONS.                                                            *
!                                                                                    *
! LATEST REVISION : 20/12/2021                                                       *
!                                                                                    *
! ####################################################################################
!                                                                                    *
! ************************************************************************************
!                                                                                    *
! THE C PROCESSOR MACROS USED IN THIS MODULE ARE:                                    *
!                                                                                    *
!  _OPENMP      FOR ACTIVATING OPENMP PARALLELIZATION                                *
!  _BLAS        FOR USING BLAS WHEN POSSIBLE                                         *
!  _DOT_PRODUCT FOR REPLACING THE dot_product INTRINSIC FUNCTION WITH STATPACK       *
!               dot_product2 FUNCTION                                                *
!  _MATMUL      FOR REPLACING THE matmul INTRINSIC FUNCTION WITH STATPACK            *
!               matmul2 FUNCTION                                                     *
!  _TRANSPOSE   FOR REPLACING THE transpose INTRINSIC FUNCTION WITH STATPACK         *
!               transpose2 FUNCTION                                                  *
!  _WHERE       FOR REPLACING SOME where CONSTRUCTS BY do LOOPS INSIDE SOME OPENMP   *
!               DIRECTIVES (USEFUL FOR SOME COMPILERS LIKE IFORT WHEN OPENMP IS      *
!               USED)                                                                *
!  _USE_INTEL   FOR SIGNALING THAT THE INTEL IFORT COMPILER IS USED.                 *
!               HERE, THIS INCLUDES ONLY THE ACTIVATION OF THE _WHERE MACRO          *
!                                                                                    *
! ************************************************************************************
!
#ifdef _USE_INTEL
#define _WHERE
#endif
!
!
! USED MODULES
! ============
!
    use Select_Parameters,     only : lgl, i4b, stnd, blksz_qr
    use Reals_Constants,       only : zero, one, ten
#ifdef _MATMUL
    use Utilities,             only : matmul=>matmul2
#endif
#ifdef _DOT_PRODUCT
    use Utilities,             only : dot_product=>dot_product2
#endif
#ifdef _TRANSPOSE
    use Utilities,             only : transpose=>transpose2
#endif
#ifdef _OPENMP
    use Select_Parameters,     only : omp_limit, omp_limit2
    use omp_lib,               only : omp_get_max_threads, omp_in_parallel, omp_get_nested,  &
                                      omp_get_num_threads, omp_get_thread_num
#endif
#ifdef _BLAS
    use BLAS_interfaces,       only : axpy, dot, trsv, gemv, ger, trsm, gemm
#endif
!
! USED MODULES IN SPECIFIC PROCEDURES
! ===================================
!   
!   use Logical_Constants
!   use Num_Constants
!   use Char_Constants
!   use Utilities
!   use Hous_Procedures
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!
! PUBLIC ENTITIES 
! ===============
!    
!   ALL SUBROUTINES, FUNCTIONS, VARIABLES AND PARAMETERS ARE PRIVATE BY DEFAULT.
!
    private
    public ::   lq_cmp, qr_cmp, qrfac, qr_cmp2,                     &
                ortho_gen_lq, ortho_gen_qr, apply_q_lq, apply_q_qr
!
! GENERIC INTERFACES FOR ROUTINES WITH OVERLOADED VERSIONS
! ========================================================
!
!
    interface apply_q_qr
        module procedure    apply_q_qr_rv, apply_q_qr_rm
    end interface
!
    interface apply_q_lq
        module procedure    apply_q_lq_rv, apply_q_lq_rm
    end interface
!
!   
! MODULE PARAMETERS 
! =================
!
!   BLOCK SIZE FOR APPLYING HOUSEHOLDER TRANSFORMATIONS
!   IN lq_cmp, qr_cmp, apply_q_lq, apply_q_qr, ortho_gen_lq AND ortho_gen_qr SUBROUTINES.
!
    integer(i4b), parameter :: blksz=blksz_qr
!
!
! =========================================================================================
!
                                  contains
!                                 ========
!
!
! =========================================================================================
!                              LQ DECOMPOSITION
! =========================================================================================
!
!
    subroutine lq_cmp( mat, diagl, tau, use_qr )
!
! Purpose
! _______
!
!   LQ_CMP computes a LQ factorization of a real m-by-n matrix MAT :
!
!                       MAT = L * Q
!
!   where Q is orthogonal and L is lower trapezoidal (lower triangular if m<=n).
!
!
! Arguments
! _________
!
!   MAT      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, the real matrix to be decomposed.
!
!            On exit, the elements below the diagonal of the array contain  
!            the corresponding elements of L. The elements on and above the 
!            diagonal, with the array TAU, represent the orthogonal matrix Q
!            as a product of elementary reflectors, see Further Details.
!
!   DIAGL    (OUTPUT) real(stnd), dimension(:) 
!            On exit, the diagonal elements of the matrix L.
!
!            The size of DIAGL must be min( size(MAT,1) , size(MAT,2) ).
!
!   TAU      (OUTPUT) real(stnd), dimension(:)
!            On exit, the scalars factors of the elementary reflectors.
!            see Further Details.
!
!            The size of TAU must be min( size(MAT,1) , size(MAT,2) ).
!
!   USE_QR   (INPUT, OPTIONAL) logical(lgl)
!            If the optional argument USE_QR is set to true, the input matrix is
!            transposed, a fast QR decomposition is used for computing the LQ
!            decomposition and the results are transposed again in the input matrix.
!
!            The default is true.
!
!
! Further Details
! _______________
!
!   The matrix Q is represented as a product of elementary reflectors
!
!           Q = H(k) * . . . * H(2) * H(1), where k = min( size(MAT,1) , size(MAT,2) )
!
!   Each H(i) has the form
!
!            H(i) = I + tau * ( v * v' ) ,
!                      
!   where tau is a real scalar and v is a real n-element vector with v(1:i-1) = 0.
!   v(i:n) is stored on exit in MAT(i,i:n) and tau in TAU(i).

!   A blocked algorithm is used for computing the LQ factorization. Furthermore,
!   the computations are parallelized if OPENMP is used.
!
!   On exit of LQ_CMP, the orthogonal matrix Q (or its first rows) can be computed
!   explicitly by a call to suboutine ORTHO_GEN_LQ with arguments MAT and TAU.
!
!   For further details on the LQ factorization and its use or the blocked
!   algorithm used here, see:
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (2) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (3) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989:
!           Block reduction of matrices to condensed form for eigenvalue computations. 
!           J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!   (4) Walker, H.F., 1988:
!           Implementation of the GMRES method using Householder transformations.
!           Siam J. Sci. Stat. Comput., Vol. 9, No 1, pp. 152-163.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Char_Constants,   only : allocate_error
    use Utilities,        only : assert_eq, merror
    use Hous_Procedures,  only : hous1
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:) :: mat
    real(stnd),   intent(out),   dimension(:)   :: diagl, tau
!
    logical(lgl), intent(in), optional :: use_qr
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                               :: prod, tau1
    real(stnd), dimension(:,:), allocatable  :: lmat, umat, x
!
    integer(i4b) :: m, n, k, k2, i, i1, i2, j, l, l1, p, q, nblk, rem
    integer      :: iok
!
    logical(lgl) :: useqr
!
#ifdef _OPENMP
!
    real(stnd), dimension(:),   allocatable :: x2
    real(stnd), dimension(:,:), allocatable :: umat2, mat2
!
    integer(i4b) :: iblk, nblk2, j0, j1, j2, q2, num_threads, thread_num
!
    logical      :: test_par, ompparallel, ompnested
!
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='lq_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
    n = size( mat, 2 )
!
    k = assert_eq( min(m,n),              &
                   int(size(tau),i4b),    &
                   int(size(diagl),i4b),  &
                   name_proc              )
!
    useqr = true
!
    if ( present(use_qr) ) then
        useqr = use_qr
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) return
!
    if ( useqr ) then
!
!       ALLOCATE WORK VARIABLE FOR A FAST QR ALGORITHM IF POSSIBLE.
!
        allocate( x(n,m), stat = iok )
!
    else
!
!       USE AN LQ ALGORITHM AUTOMATICALLY.
!
        iok = -1
!
    end if
!
    if ( iok==0 ) then
!
!       TRANSPOSE THE MATRIX AND USE A FAST QR DECOMPOSITION.
!
        x(:n,:m) = transpose( mat(:m,:n) )
!
        call qr_cmp( x(:n,:m), diagl, tau )
!
        mat(:m,:n) = transpose( x(:n,:m) )
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( x )
!
    else
!
!       USE A SLOWER LQ ALGORITHM IF THE INPUT MATRIX IS TOO BIG.
!
!       COMPUTE THE NUMBER OF BLOCKS.
!
        nblk = k/blksz
        rem  = mod( k, blksz )
!
        if ( rem/=0_i4b ) then
!
            nblk = nblk + 1_i4b
            k2   = k - rem
!
        else
!
            k2   = k - blksz
!
        end if
!
#ifdef _OPENMP
!
!       COMPUTE THE NUMBER OF ROW BLOCKS.
!
        nblk2 = m/blksz
        rem  = mod( m, blksz )
!
        if ( rem/=0_i4b ) nblk2 = nblk2 + 1_i4b
!
        i           = omp_get_max_threads()
        ompparallel = omp_in_parallel()
        ompnested   = omp_get_nested() .and. ompparallel
        test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                      n*m>=omp_limit                      .and.      &
                      k>blksz                             .and.      &
                      nblk2>1_i4b                         .and.      &
                      i>1_i4b
!
        if ( test_par ) then
!
!           ALLOCATE SHARED WORK VARIABLES.
!
            allocate( umat(n,blksz), umat2(n,blksz), lmat(blksz,blksz), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           INITIALIZE THE UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!           USED IN THE BLOCKED ALGORITHM.
!
            lmat(1_i4b:blksz,1_i4b:blksz) = zero
!
            do i = 1_i4b, blksz
                lmat(i,i) = one
            end do
!
!           PARALLEL AND BLOCKED LQ DECOMPOSITION OF mat.
!
!$OMP PARALLEL PRIVATE(iok,i,i1,i2,iblk,j,j0,j1,j2,l,l1,p,q,q2,num_threads,thread_num,tau1,prod,x2,mat2)
!
!           ALLOCATE PRIVATE WORK VARIABLES.
!
            allocate( mat2(n,blksz), x2(blksz), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
            num_threads = omp_get_num_threads()
            thread_num  = omp_get_thread_num()
!
            if ( thread_num==0_i4b ) then
!
!               COMPUTE LQ FACTORIZATION OF THE FIRST BLOCK IN PLACE.
!
                umat(1_i4b:n,1_i4b:blksz) = transpose( mat(1_i4b:blksz,1_i4b:n) )
!
                do l = 1_i4b, blksz
!    
!                   GENERATE ELEMENTARY REFLECTOR H(l) TO ANNIHILATE mat(l,l+1:n) .
!    
                    call hous1( umat(l:n,l), tau(l), diagl(l) )
!    
!                   APPLY H(l) TO mat(l+1:blksz,l:n) FROM THE RIGHT.
!
                    tau1 = tau(l)
!
                    do j = l + 1_i4b, blksz
!
                        prod = tau1*dot_product( umat(l:n,l), umat(l:n,j) )
#ifdef _BLAS
                        call axpy( n-l+1_i4b, prod, umat(l:n,l), 1_i4b, umat(l:n,j), 1_i4b )
#else
                        umat(l:n,j) = umat(l:n,j) + prod*umat(l:n,l)
#endif
!    
                    end do
!    
                end do
!
                mat(1_i4b:blksz,1_i4b:n) = transpose( umat(1_i4b:n,1_i4b:blksz) )
!
            end if
!
!$OMP BARRIER
!
!           START THE BLOCKED LQ FACTORIZATION.
!
            do i = 0_i4b, k2, blksz
!    
                i2 = i + blksz
!
                if ( m>i2 ) then
!    
                    i1 = i + 1_i4b
                    q  = min( i2, k ) - i
!    
!                   APPLY H(i+1)* ...*H(i+blksz) TO mat(i+blksz+1:m,i+1:n) FROM THE RIGHT
!                   WITH A BLOCKED ALGORITHM AS
!
!                mat(i+blksz+1:m,i+1:n)*( I + U(i+1:n,:blksz)*inv( lmat(:blksz,:blksz) )*U(i+1:n,:blksz)' ) .
!
!                   First, COMPUTE THE blksz BY blksz UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz) .
!
!$OMP DO SCHEDULE(STATIC,1)
!
                    do l = 1_i4b, q
!
                        l1 = i + l
!
                        umat2(l1:n,l) = umat(l1:n,l)                      
!
                        do j = l + 1_i4b, q
!
                            l1 = i + j
!
                            lmat(j,l) = -tau(l1)*dot_product( umat(l1:n,l), umat(l1:n,j) )
!
                        end do
!                    
                    end do
!
!$OMP END DO
!
!                   COMPUTE THE PREVIOUS BLOCK NUMBER.
!
                    iblk = i/blksz + 1_i4b
!
!                   EACH THREAD WILL TOUCH EXACTLY THE SAME ROWS BLOCKS DURING THE ITERATIONS.
!
                    p = ( (iblk/num_threads)*num_threads ) + thread_num
                    if ( thread_num<mod(iblk,num_threads) ) p = p + num_threads
!
                    do j = p, nblk2-1_i4b, num_threads
!
!                       COMPUTE INDICES FOR THE FIRST AND LAST ROWS OF THE j+1 BLOCK.
!
                        j0 = j*blksz
                        j1 = j0 + 1_i4b
                        j2 = min( j0+blksz, m )
!
                        q2 = j2 - j0
!
                        mat2(i1:n,1_i4b:q2) = transpose( mat(j1:j2,i1:n) )
!
                        do p = 1_i4b, q2
!
!                           COMPUTE mat(p,i+1:n)*U(i+1:n,:blksz).
!
                            do l = 1_i4b, q
!
                                l1 = i + l
!
                                x2(l) = tau(l1)*dot_product( umat2(l1:n,l), mat2(l1:n,p) )
!                    
                            end do
!
!                           SOLVE x2(:blksz)*lmat(:blksz,:blksz) = mat(p,i+1:n)*U(i+1:n,:blksz) AND
!                           COMPUTE mat(p,i+1:n) = mat(p,i+1:n) + x2(:blksz)*U(i+1:n,:blksz)' .
!
                            do l = 1_i4b, q
!
                                x2(l+1_i4b:q) = x2(l+1_i4b:q) - x2(l)*lmat(l+1_i4b:q,l)
!
                                l1 = i + l
!
#ifdef _BLAS
                                call axpy( n-l1+1_i4b, x2(l), umat2(l1:n,l), 1_i4b, mat2(l1:n,p), 1_i4b )
#else
                                mat2(l1:n,p) = mat2(l1:n,p) + x2(l)*umat2(l1:n,l)
#endif
!
                            end do
!
                        end do
!
                        if ( j==iblk .and. j1<=k ) then
!
!                           COMPUTE THE LQ FACTORIZATION OF THE iblk+1 BLOCK, EG OF mat(i+blksz+1:i+2*blksz,i+blksz+1:n), IN PLACE.
!
                            do l = 1_i4b, min(q2,k-j0)
!    
                                l1 = j0 + l
!    
!                               GENERATE ELEMENTARY REFLECTOR H(l1) TO ANNIHILATE mat(l1,l1+1:n) .
!                    
                                call hous1( mat2(l1:n,l), tau(l1), diagl(l1) )
!
                                umat(l1:n,l) = mat2(l1:n,l)
!    
!                               APPLY H(l1) TO mat(l1+1:i+2*blksz,l1:n) FROM THE RIGHT.
!
                                tau1 = tau(l1)
!
                                do p = l + 1_i4b, q2
!
                                    prod = tau1*dot_product( mat2(l1:n,l), mat2(l1:n,p) )
#ifdef _BLAS
                                    call axpy( n-l1+1_i4b, prod, mat2(l1:n,l), 1_i4b, mat2(l1:n,p), 1_i4b )
#else
                                    mat2(l1:n,p) = mat2(l1:n,p) + prod*mat2(l1:n,l)
#endif
!
                                end do
!
                            end do
!
                        end if
!
                        mat(j1:j2,i1:n) = transpose( mat2(i1:n,1_i4b:q2) )
!
                    end do
!
                end if
!
!$OMP BARRIER
!
            end do
!
!           DEALLOCATE PRIVATE WORK VARIABLES.
!
            deallocate( x2, mat2 )
!
!$OMP END PARALLEL
!
!           DEALLOCATE SHARED WORK VARIABLES.
!
            deallocate( lmat, umat, umat2 )
!
        else
!
#endif
!
!           ALLOCATE WORK VARIABLES.
!
            allocate( lmat(blksz,blksz), umat(n,blksz), x(m,blksz), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           INITIALIZE THE UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!           USED IN THE BLOCKED ALGORITHM.
!
            lmat(1_i4b:blksz,1_i4b:blksz) = zero
!
            do i = 1_i4b, blksz
                lmat(i,i) = one
            end do
!
!           BLOCKED LQ DECOMPOSITION OF mat.
!
            do i = 0_i4b, k2, blksz
!
                i1 = i + 1_i4b
                i2 = min( i + blksz, k )
!
                q = i2 - i
!
                umat(i1:n,1_i4b:q) = transpose( mat(i1:i2,i1:n) )
!
                do l = 1_i4b, q
!    
                    l1 = i + l
!    
!                   GENERATE ELEMENTARY REFLECTOR H(i+l) TO ANNIHILATE mat(i+l,i+l+1:n) .
!    
                    call hous1( umat(l1:n,l), tau(l1), diagl(l1) )
!    
!                   APPLY H(i+l) TO mat(i+l+1:i+blksz,i+l:n) FROM THE RIGHT.
!
                    tau1 = tau(l1)
!
                    do j = l + 1_i4b, q
!
                        prod  = tau1*dot_product( umat(l1:n,l), umat(l1:n,j) )
#ifdef _BLAS
                        call axpy( n-l1+1_i4b, prod, umat(l1:n,l), 1_i4b, umat(l1:n,j), 1_i4b )
#else
                        umat(l1:n,j) = umat(l1:n,j) + prod*umat(l1:n,l)
#endif
!    
                    end do
!    
                end do
!
                mat(i1:i2,i1:n) = transpose( umat(i1:n,1_i4b:q) )
!    
                if ( m>i2 ) then
!    
!                   APPLY H(i+1)* ...*H(i+blksz) TO mat(i+blksz+1:m,i+1:n) FROM THE RIGHT
!                   WITH A BLOCKED ALGORITHM AS
!
!              mat(i+blksz+1:m,i+1:n)*( I + U(i+1:n,:blksz)*inv( lmat(:blksz,:blksz) )*U(i+1:n,:blksz)' ) .
!
!                   FIRST, COMPUTE THE blksz BY blksz UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz) .
!
                    do l = 1_i4b, q - 1_i4b
!
                        do j = l + 1_i4b, q
!
                            l1 = i + j
!
                            lmat(j,l) = -tau(l1)*dot_product( umat(l1:n,l), umat(l1:n,j) )
!
                        end do
!                    
                    end do
!
                    j = i2 + 1_i4b
!
!                   COMPUTE mat(j:m,i+1:n)*U(i+1:n,:blksz).
!
                    do l = 1_i4b, q
!
                        l1 = i + l
!
                        x(j:m,l) = tau(l1)*matmul( mat(j:m,l1:n), umat(l1:n,l) )
!                    
                    end do
!
!                   SOLVE x(j:m,:blksz)*lmat(:blksz,:blksz) = mat(j:m,i+1:n)*U(i+1:n,:blksz) AND
!                   COMPUTE mat(j:m,i+1:n) = mat(j:m,i+1:n) + x(j:m,:blksz)*U(i+1:n,:blksz)' .
!
                    do l = 1_i4b, q
!
                        do p = l + 1_i4b, q
!
#ifdef _BLAS
                            call axpy( m-j+1_i4b, -lmat(p,l), x(j:m,l), 1_i4b, x(j:m,p), 1_i4b )
#else
                            x(j:m,p) = x(j:m,p) - x(j:m,l)*lmat(p,l)
#endif
!
                            l1 = i + p - 1_i4b
!
#ifdef _BLAS
                            call axpy( m-j+1_i4b, umat(l1,l), x(j:m,l), 1_i4b, mat(j:m,l1), 1_i4b )
#else
                            mat(j:m,l1) = mat(j:m,l1) + x(j:m,l)*umat(l1,l)
#endif
!
                        end do
!
                    end do
!
                    do p = i + q, n
!
                        do l = 1_i4b, q
!
#ifdef _BLAS
                            call axpy( m-j+1_i4b, umat(p,l), x(j:m,l), 1_i4b, mat(j:m,p), 1_i4b )
#else
                            mat(j:m,p) = mat(j:m,p) + x(j:m,l)*umat(p,l)
#endif
!        
                        end do
!                    
                    end do
!
                end if
!
            end do
!
!           DEALLOCATE WORK VARIABLES.
!
            deallocate( lmat, umat, x )
!
#ifdef _OPENMP
        end if
#endif
!
    end if
!
!
! END OF SUBROUTINE lq_cmp
! ________________________
!
    end subroutine lq_cmp
!
! =========================================================================================
!
    subroutine ortho_gen_lq( mat, tau, use_qr )
!
! Purpose
! _______
!
!   ORTHO_GEN_LQ generates an m-by-n real matrix with orthonormal rows, which is
!   defined as the first m rows of a product of k elementary reflectors of order n
!
!           Q  =  H(k) * . . . * H(2) * H(1)
!
!   as returned by LQ_CMP.
!
!
! Arguments
! _________
!
!   MAT      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, the i-th row must contain the vector which
!            defines the elementary reflector H(i), for i = 1,2,...,k,
!            as returned by LQ_CMP in the first k rows of its array
!            argument MAT.
!
!            On exit, the first m rows of Q.
!
!            The shape of MAT must verify: size( MAT, 1 ) <= size( MAT, 2 ).
!
!   TAU      (INPUT) real(stnd), dimension(:)
!            TAU(i) must contain the scalar factor of the elementary
!            reflector H(i), as returned by LQ_CMP.
!            The size of TAU determines the number k of elementary reflectors
!            whose product defines the matrix Q.
!
!            The size of TAU must verify: size( TAU )  <= size( MAT, 1 ) .
!
!   USE_QR   (INPUT, OPTIONAL) logical(lgl)
!            If the optional argument USE_QR is set to true, the input matrix is
!            transposed, a fast QR type algorithm is used for computing the first m
!            rows of Q and the results are transposed again in the input matrix.
!
!            The default is true.
!
!
! Further Details
! _______________
!
!   This subroutine used a blocked algorithm for agregating the Householder
!   transformations (e.g. the elementary reflectors) stored in the upper triangle
!   of MAT and generating the orthogonal matrix Q of the LQ factorization.
!
!   Furthermore, the computations are parallelized if OPENMP is used.
!
!   For further details on the LQ factorization and its use or the blocked
!   algorithm, see:
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (2) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (3) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989:
!           Block reduction of matrices to condensed form for eigenvalue computations. 
!           J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!   (4) Walker, H.F., 1988:
!           Implementation of the GMRES method using Householder transformations.
!           Siam J. Sci. Stat. Comput., Vol. 9, No 1, pp. 152-163.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Char_Constants,   only : allocate_error
    use Utilities,        only : assert, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:) :: mat
    real(stnd),   intent(in),    dimension(:)   :: tau
!
    logical(lgl), intent(in), optional :: use_qr
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                              :: prod, tau1
    real(stnd), dimension(:),   allocatable :: x2
    real(stnd), dimension(:,:), allocatable :: lmat, umat, x, tmp
!
    integer(i4b) :: m, n, k, i, i0, i1, i2, l, l2, j, blksz2, p
    integer      :: iok
!
    logical(lgl) :: useqr
!
#ifdef _OPENMP
!
    integer(i4b) :: j1, j2, blksz3
!
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='ortho_gen_lq'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
    n = size( mat, 2 )
    k = size( tau )
!
    call assert( logical( (k<=m) .and. ( m<=n), lgl ), name_proc )
!
    useqr = true
!
    if ( present(use_qr) ) then
        useqr = use_qr
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) return
!
    if ( useqr ) then
!
!       ALLOCATE WORK VARIABLE FOR A FAST QR-TYPE ALGORITHM IF POSSIBLE.
!
        allocate( x(n,m), stat = iok )
!
    else
!
!       USE AN LQ-TYPE ALGORITHM AUTOMATICALLY.
!
        iok = -1
!
    end if
!
    if ( iok==0 ) then
!
!       TRANSPOSE THE MATRIX AND USE A FAST QR-TYPE ALGORITHM.
!
        x(:n,:m) = transpose( mat(:m,:n) )
!
        call ortho_gen_qr( x(:n,:m), tau )
!
        mat(:m,:n) = transpose( x(:n,:m) )
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( x )
!
    else
!
!       USE A SLOWER LQ-TYPE ALGORITHM IF THE INPUT MATRIX IS TOO BIG.
!
!       INITIALIZE ROWS k+1:m OF mat TO ROWS OF THE UNIT MATRIX.
!
        mat(k+1_i4b:m,1_i4b:n) = zero
!
        do i = k+1_i4b, min( n, m )
            mat(i,i) = one
        end do
!
#ifdef _OPENMP
        i = omp_get_max_threads()
        ompparallel = omp_in_parallel()
        ompnested   = omp_get_nested() .and. ompparallel
        test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                      n*m>=omp_limit                      .and.      &
                      m/i>=omp_limit2                     .and.      &
                      i>1_i4b
!
        if ( test_par ) then
!
!           ALLOCATE SHARED WORK VARIABLES.
!
            allocate( umat(n,blksz), lmat(blksz,blksz), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           INITIALIZE THE UNIT LOWER OR UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!           USED IN THE BLOCKED ALGORITHM.
!
            lmat(1_i4b:blksz,1_i4b:blksz) = zero
!
            do i = 1_i4b, blksz
                lmat(i,i) = one
            end do
!
!           PARALLEL AND BLOCKED LQ-TYPE ALGORITHM.
!
!$OMP PARALLEL PRIVATE(iok,i,i0,i1,i2,l,l2,j,j1,j2,p,blksz2,blksz3,tau1,prod,x2,tmp)
!
!           ALLOCATE PRIVATE WORK VARIABLES.
!
            allocate( tmp(n,blksz), x2(blksz), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
            do i = k - blksz, -blksz+1_i4b, -blksz
!    
!               APPLY H(i+blksz)* ...*H(i+1) TO mat(i+blksz+1:m,i+1:n) FROM THE RIGHT
!               WITH A BLOCKED ALGORITHM AS
!
!           mat(i+blksz+1:m,i+1:n)*( I + U(i+1:n,:blksz)*inv( lmat(:blksz,:blksz) )*U(i+1:n,:blksz)' ) .
!
                blksz2 = min( blksz, i+blksz )
                i0     = max( 0_i4b, i )
!
!               SAVE H(i+1) ...H(i+blksz) FOR LATER USE.
!    
!$OMP SINGLE
!
                umat(i0+1_i4b:n,1_i4b:blksz2) = transpose( mat(i0+1_i4b:i0+blksz2,i0+1_i4b:n) )
!
!$OMP END SINGLE
!
!               COMPUTE THE blksz BY blksz UNIT UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz) .
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do l = 2_i4b, blksz2
!
                    i1 = i0 + l
!
                    do l2 = 1_i4b, l - 1_i4b
!
                        i2 = i0 + l2
!
                        lmat(l2,l) = -tau(i2)*dot_product( umat(i1:n,l), umat(i1:n,l2) )
!
                    end do
!                    
                end do
!
!$OMP END DO
!
! $OMP DO SCHEDULE(STATIC,1)
!
!                do l = 1_i4b, blksz2
!
!                    i1 = i0 + l
!
!                    umat(i1:n,l) = mat(i1,i1:n)
!
!                    do l2 = 1_i4b, l - 1_i4b
!
!                        i2 = i0 + l2
!
!                        lmat(l2,l) = -tau(i2)*dot_product( umat(i1:n,l), umat(i1:n,l2) )
!
!                    end do
!                    
!                end do
!
! $OMP END DO
!
!$OMP DO SCHEDULE(STATIC)
!
                do j = i0+blksz2, m-1_i4b, blksz
!
                    j1     = j + 1_i4b
                    j2     = min( j + blksz, m)
                    blksz3 = j2-j1+1_i4b
!
                    tmp(i0+1_i4b:n,1_i4b:blksz3) = transpose( mat(j1:j2,i0+1_i4b:n) )
!
                    do p = 1_i4b, blksz3
!
!                       COMPUTE mat(j+p,i+1:n)*U(i+1:n,:blksz) .
!
                        do l = 1_i4b, blksz2
!
                            i1 = i0 + l
!
                            x2(l) = tau(i1)*dot_product( umat(i1:n,l), tmp(i1:n,p) )
!                    
                        end do
!
!                       SOLVE x2(:blksz)*lmat(:blksz,:blksz)' = mat(j+p,i+1:n)*U(i+1:n,:blksz) AND
!                       COMPUTE mat(j+p,i+1:n) = mat(j+p,i+1:n) + x2(:blksz)*U(i+1:n,:blksz)' .
!
#ifdef _BLAS
                        call trsv( 'U', 'N', 'U', blksz2, lmat, blksz, x2, 1_i4b )
!
                        do l = blksz2, 1_i4b, -1_i4b
!
                            i1 = i0 + l
!
                            call axpy( n-i1+1_i4b, x2(l), umat(i1:n,l), 1_i4b, tmp(i1:n,p), 1_i4b )
!
                        end do
#else
                        do l = blksz2, 1_i4b, -1_i4b
!
                            x2(1_i4b:l-1_i4b) = x2(1_i4b:l-1_i4b) - x2(l)*lmat(1_i4b:l-1_i4b,l)
!
                            i1 = i0 + l
!
                            tmp(i1:n,p) = tmp(i1:n,p) + x2(l)*umat(i1:n,l)
!
                        end do
#endif
!
                    end do
!
                    mat(j1:j2,i0+1_i4b:n) = transpose( tmp(i0+1_i4b:n,1_i4b:blksz3) )
!                    
                end do
!
!$OMP END DO NOWAIT
!    
!               APPLY H(i+blksz)* ...*H(i+1) TO mat(i+1:i+blksz,i+1:n) FROM THE RIGHT
!               WITH A SERIAL ALGORITHM.
!    
!$OMP SINGLE
!
                do l = blksz2, 1_i4b, -1_i4b
!
                    i1 = i0 + l
!
                    tau1 = tau(i1)
!
!                   APPLY H(i1) TO mat(i1+1:i+blksz,i1:n) FROM THE RIGHT.
!
                    do l2 = l + 1_i4b, blksz2
!
                        prod = tau1*dot_product( umat(i1+1_i4b:n,l), tmp(i1+1_i4b:n,l2) )
#ifdef _BLAS
                        call axpy( n-i1+1_i4b, prod, umat(i1:n,l), 1_i4b, tmp(i1:n,l2), 1_i4b )
#else
                        tmp(i1:n,l2) = tmp(i1:n,l2) + prod*umat(i1:n,l)
#endif
!    
                    end do
!
!                   APPLY H(i1) TO mat(i1,i1:n) FROM THE RIGHT.
!
                    tmp(1_i4b:i1-1_i4b,l) = zero
                    tmp(i1,l)             = one + tau1
                    tmp(i1+1_i4b:n,l)     = tau1*umat(i1+1_i4b:n,l)
!    
                end do
!    
                mat(i0+1_i4b:i0+blksz2,1_i4b:n) = transpose( tmp(1_i4b:n,1_i4b:blksz2) )
!    
!$OMP END SINGLE
!
            end do
!
!           DEALLOCATE PRIVATE WORK VARIABLES.
!
            deallocate( x2, tmp )
!
!$OMP END PARALLEL
!
!           DEALLOCATE SHARED WORK VARIABLES.
!
            deallocate( lmat, umat )
!
        else
#endif
!
!           ALLOCATE WORK VARIABLES.
!
            allocate( umat(n,blksz), tmp(n,blksz), lmat(blksz,blksz), x(m,blksz), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           INITIALIZE THE UNIT LOWER OR UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!           USED IN THE BLOCKED ALGORITHM.
!
            lmat(1_i4b:blksz,1_i4b:blksz) = zero
!
            do i = 1_i4b, blksz
                lmat(i,i) = one
            end do
!
            tmp(1_i4b:n,1_i4b:blksz) = zero
!
!           BLOCKED LQ-TYPE ALGORITHM.
!
            do i = k - blksz, -blksz+1_i4b, -blksz
!    
!               APPLY H(i+blksz)* ...*H(i+1) TO mat(i+blksz+1:m,i+1:n) FROM THE RIGHT
!               WITH A BLOCKED ALGORITHM AS
!
!           mat(i+blksz+1:m,i+1:n)*( I + U(i+1:n,:blksz)*inv( lmat(:blksz,:blksz) )*U(i+1:n,:blksz)' ) .
!
                blksz2 = min( blksz, i+blksz )
                i0     = max( 0_i4b, i )
!
!               SAVE H(i+1) ...H(i+blksz) FOR LATER USE.
!
                umat(i0+1_i4b:n,1_i4b:blksz2) = transpose( mat(i0+1_i4b:i0+blksz2,i0+1_i4b:n) )
!
!               COMPUTE THE blksz BY blksz UNIT UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz) .
!
                do l = 2_i4b, blksz2
!
                    i1 = i0 + l
!
                    do l2 = 1_i4b, l - 1_i4b
!
                        i2 = i0 + l2
!
                        lmat(l2,l) = -tau(i2)*dot_product( umat(i1:n,l), umat(i1:n,l2) )
!
                    end do
!                    
                end do
!
                p = i0 + blksz2 + 1_i4b
!
!               COMPUTE mat(p:m,i+1:n)*U(i+1:n,:blksz) .
!
                do l = 1_i4b, blksz2
!
                    i1 = i0 + l
!
                    x(p:m,l) = tau(i1)*matmul( mat(p:m,i1:n), umat(i1:n,l) )
!                    
                end do
!
!               SOLVE x(p:m,:blksz)*lmat(:blksz,:blksz)' = mat(p:m,i+1:n)*U(i+1:n,:blksz) AND
!               COMPUTE mat(p:m,i+blksz-1:n) = mat(p:m,i+blksz-1:n) + x(p:m,:blksz)*U(i+blksz-1:n,:blksz)' .
!
                do l = blksz2, 1_i4b, -1_i4b
!
                    do l2 = 1_i4b, l-1_i4b
!
#ifdef _BLAS
                        call axpy( m-p+1_i4b, -lmat(l2,l), x(p:m,l), 1_i4b, x(p:m,l2), 1_i4b )
#else
                        x(p:m,l2) = x(p:m,l2) - x(p:m,l)*lmat(l2,l)
#endif
!
                    end do
!
                    do l2 = l, blksz2-1_i4b
!
                        i2 = i0 + l2
!
#ifdef _BLAS
                        call axpy( m-p+1_i4b, umat(i2,l), x(p:m,l), 1_i4b, mat(p:m,i2), 1_i4b )
#else
                        mat(p:m,i2) = mat(p:m,i2) + x(p:m,l)*umat(i2,l)
#endif
!
                    end do
!
                end do
!
!               COMPUTE mat(p:m,i+blksz:n) = mat(p:m,i+blksz:n) + x(p:m,:blksz)*U(i+blksz:n,:blksz)' .
!
                do j = i0 + blksz2, n
!
                    do l = 1_i4b, blksz2
!
#ifdef _BLAS
                        call axpy( m-p+1_i4b, umat(j,l), x(p:m,l), 1_i4b, mat(p:m,j), 1_i4b )
#else
                        mat(p:m,j) = mat(p:m,j) + x(p:m,l)*umat(j,l)
#endif
!                    
                    end do
!                    
                end do
!    
!               APPLY H(i+blksz)* ...*H(i+1) TO mat(i+1:i+blksz,i+1:n) FROM THE RIGHT
!               WITH A SERIAL ALGORITHM.
!
                do l = blksz2, 1_i4b, -1_i4b
!
                    i1 = i0 + l
!
                    tau1 = tau(i1)
!
!                   APPLY H(i1) TO mat(i1+1:i+blksz,i1:n) FROM THE RIGHT.
!
                    do l2 = l + 1_i4b, blksz2
!
                        prod = tau1*dot_product( umat(i1+1_i4b:n,l), tmp(i1+1_i4b:n,l2) )
#ifdef _BLAS
                        call axpy( n-i1+1_i4b, prod, umat(i1:n,l), 1_i4b, tmp(i1:n,l2), 1_i4b )
#else
                        tmp(i1:n,l2) = tmp(i1:n,l2) + prod*umat(i1:n,l)
#endif
!    
                    end do
!
!                   APPLY H(i1) TO mat(i1,i1:n) FROM THE RIGHT.
!
!                    tmp(1_i4b:i1-1_i4b,l) = zero
                    tmp(i1,l)         = one + tau1
                    tmp(i1+1_i4b:n,l) = tau1*umat(i1+1_i4b:n,l)
!    
                end do
!    
                mat(i0+1_i4b:i0+blksz2,1_i4b:n) = transpose( tmp(1_i4b:n,1_i4b:blksz2) )
!
            end do
!
!           ALLOCATE WORK VARIABLES.
!
            deallocate( umat, tmp, lmat, x, stat = iok )
!
#ifdef _OPENMP
        end if
#endif
!
    end if
!
!
! END OF SUBROUTINE ortho_gen_lq
! ______________________________
!
    end subroutine ortho_gen_lq
!
! =========================================================================================
!
    subroutine apply_q_lq_rm( mat, tau, c, left, trans )
!
! Purpose
! _______
!
!   APPLY_Q_LQ overwrites the general real m-by-n matrix C with
!
!           Q * C   if LEFT = true  and TRANS = false, or
!
!           Q' * C  if LEFT = true  and TRANS = true,  or
!
!           C * Q   if LEFT = false and TRANS = false, or
!
!           C * Q'  if LEFT = false and TRANS = true,
!
!   where Q is a real orthogonal matrix defined as the product of k
!   elementary reflectors
!
!                   Q = H(k) * . . . * H(2) * H(1)
!
!   as returned by LQ_CMP. Q is of order m if LEFT = true and of order n
!   if LEFT = false.
!
!
! Arguments
! _________
!
!   MAT          (INPUT) real(stnd), dimension(:,:)
!                On entry, the i-th row must contain the vector which
!                defines the elementary reflector H(i), for i = 1,2,...,k,
!                as returned by LQ_CMP in the first k rows of its array
!                argument MAT. MAT is not modified by the routine.
!
!   TAU          (INPUT) real(stnd), dimension(:)
!                TAU(i) must contain the scalar factor of the elementary
!                reflector H(i), as returned by LQ_CMP.
!                The size of TAU determines the number k of elementary
!                reflectors whose product defines the matrix Q.
!
!                The size of TAU must verify: size( TAU ) <= min( size(MAT,1) , size(MAT,2) ) .
!
!   C            (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the m by n matrix C.
!
!                On exit, C is overwritten by Q * C or Q' * C or C * Q' or C * Q.
!
!                The shape of C must verify:
!
!                -  if LEFT = true,  size( C, 1 ) = size( MAT, 2 )
!                -  if LEFT = false, size( C, 2 ) = size( MAT, 2 ) .
!
!   LEFT         (INPUT) logical(lgl)
!                If:
!
!                - LEFT = true  : apply Q or Q' from the left
!                - LEFT = false : apply Q or Q' from the right
!
!   TRANS        (INPUT) logical(lgl)
!                If:
!
!                - TRANS = false : apply Q  (no transpose)
!                - TRANS = true  : apply Q' (transpose)
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the routine  DORML2 in LAPACK.
!
!   This subroutine used a blocked algorithm for agregating the Householder
!   transformations (e.g. the elementary reflectors) stored in the upper triangle
!   of MAT and applying the orthogonal matrix Q of the LQ factorization to the
!   real m-by-n matrix C.
!
!   Furthermore, the computations are parallelized if OPENMP is used.
!
!   For further details on the LQ factorization and its use or the blocked
!   algorithm used here, see:
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (2) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (3) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989:
!           Block reduction of matrices to condensed form for eigenvalue computations. 
!           J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!   (4) Walker, H.F., 1988:
!           Implementation of the GMRES method using Householder transformations.
!           Siam J. Sci. Stat. Comput., Vol. 9, No 1, pp. 152-163.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Char_Constants,   only : allocate_error
    use Utilities,        only : assert, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in),    dimension(:,:) :: mat
    real(stnd),   intent(in),    dimension(:)   :: tau
    real(stnd),   intent(inout), dimension(:,:) :: c
!
    logical(lgl), intent(in)    :: left, trans
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                           :: tau1, prod
    real(stnd), dimension( blksz )                       :: x
    real(stnd), dimension( max( size(c,1), size(c,2) ) ) :: tmp
    real(stnd), dimension( blksz, blksz )                :: lmat
    real(stnd), dimension( size(mat,2), blksz )          :: umat
    real(stnd), dimension( :, : ), allocatable           :: x2
!
    integer(i4b) :: m, n, nm, nq, k, i, i1, i2, j, j1, j2, j3, l, l2, p
    integer      :: iok
!
    logical(lgl) :: descending
!
#ifdef _OPENMP
!
    integer(i4b) :: num_threads, thread_num, jinc, jrem, joffset
!
    logical      :: test_par, ompparallel, ompnested
!
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='apply_q_lq'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( c, 1 )
    n = size( c, 2 )
    nm = n*m
!
    if ( left ) then
        nq = m
        p  = n
    else
        nq = n
        p  = m
    end if
    k = size( tau )
!
    call assert( logical( nq==size(mat,2), lgl ),             &
                 logical( k<=min(nq,size(mat,1)), lgl ),      &
                 name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b .or.  nm<=0_i4b ) return
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  nm>=omp_limit                       .and.      &
                  (nm/nq)/i>=omp_limit2               .and.      &
                  i>1_i4b
#endif
!
    descending = (left.and.trans) .or.  .not.(left.or.trans)
!
!$OMP PARALLEL IF(test_par)                                               &
!$OMP         ,PRIVATE(iok,i,i1,i2,l,l2,j,j1,j2,j3,jinc,jrem,joffset,     &
!$OMP                  num_threads,thread_num,tau1,prod,x,umat,tmp,lmat,x2)
!
#ifdef _OPENMP
    if ( test_par ) then
!
!       DISTRIBUTE THE WORK AMONG THE THREADS.
!
        num_threads = omp_get_num_threads()
        thread_num  = omp_get_thread_num()
!
        jrem        = mod( p, num_threads )
        jinc        = ( p - jrem )/num_threads
!
        if ( thread_num<jrem ) then
            jinc    = jinc + 1_i4b
            joffset = 1_i4b
        else
            joffset = jrem + 1_i4b
        end if
!
        j1 = thread_num*jinc + joffset
        j2 = j1 + jinc - 1_i4b
!
    else
#endif
        j1 = 1_i4b
        j2 = p
#ifdef _OPENMP
    end if
#endif
!
!   INITIALIZE THE UNIT LOWER OR UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!   USED IN THE BLOCKED ALGORITHM.
!
    lmat(1_i4b:blksz,1_i4b:blksz) = zero
!
    do i = 1_i4b, blksz
        lmat(i,i) = one
    end do
!
    if ( left ) then
!
        if ( descending ) then
!
!           COMPUTE H(1)*H(2)* . . .*H(k)*c(:m,:n) .
!
            do i = k - blksz, 0_i4b, -blksz
!    
!               APPLY H(i+1)* ...*H(i+blksz) TO c(i+1:m,1:n) FROM THE LEFT
!               WITH A BLOCKED ALGORITHM AS
!
!          ( I + U(i+1:m,:blksz)*inv( lmat(:blksz,:blksz) )*U(i+1:m,:blksz)' )*c(i+1:m,1:n) .
!
                i1 = i + 1_i4b
                i2 = i + blksz
!
                umat(i1:m,1_i4b:blksz) = transpose( mat(i1:i2,i1:m) )
!
!               COMPUTE THE blksz BY blksz UNIT UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz) .
!
                do l = 2_i4b, blksz
!
                    i1 = i + l
!
                    do l2 = 1_i4b, l - 1_i4b
!
                        i2 = i + l2
!
                        lmat(l2,l) = -tau(i2)*dot_product( umat(i1:m,l), umat(i1:m,l2) )
!
                    end do
!                    
                end do
!
                do j = j1, j2
!
!                   COMPUTE U(i+1:m,:blksz)'*c(i+1:m,j) .
!
                    tmp(i+1_i4b:m) = c(i+1_i4b:m,j)
!
                    do l = 1_i4b, blksz
!
                        i1 = i + l
!
                        x(l) = tau(i1)*dot_product( umat(i1:m,l), tmp(i1:m) )
!                    
                    end do
!
!                   SOLVE lmat(:blksz,:blksz)*x(:blksz) = U(i+1:m,:blksz)'*c(i+1:m,j)  AND
!                   COMPUTE c(i+1:m,j) = c(i+1:m,j) + U(i+1:m,:blksz)*x(:blksz) .
!
#ifdef _BLAS
                    call trsv( 'U', 'N', 'U', blksz, lmat, blksz, x, 1_i4b )
!
                    do l = blksz, 1_i4b, -1_i4b
!
                        i1 = i + l
!
                        call axpy( m-i1+1_i4b, x(l), umat(i1:m,l), 1_i4b, tmp(i1:m), 1_i4b )
!
                    end do
#else
                    do l = blksz, 1_i4b, -1_i4b
!
                        x(1_i4b:l-1_i4b) = x(1_i4b:l-1_i4b) - x(l)*lmat(1_i4b:l-1_i4b,l)
!
                        i1 = i + l
!
                        tmp(i1:m) = tmp(i1:m) + x(l)*umat(i1:m,l)
!
                    end do
#endif
!
                    c(i+1_i4b:m,j) = tmp(i+1_i4b:m)
!
                end do
!
            end do
!
!           APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
            i1 = i + blksz
!
            do i = i1, 1_i4b, -1_i4b
!
                tau1 = tau(i)
!
                if ( tau1==zero ) cycle
!
!               APPLY H(i) TO c(i:m,:n) FROM THE LEFT.
!
                tmp(i:m) = mat(i,i:m)
!
                do j = j1, j2
!
                    prod = dot_product( tmp(i:m), c(i:m,j) )*tau1
#ifdef _BLAS
                    call axpy( m-i+1_i4b, prod, tmp(i:m), 1_i4b, c(i:m,j), 1_i4b )
#else
                    c(i:m,j) = c(i:m,j) + prod*tmp(i:m)               
#endif
!
                end do
!
            end do            
!
        else
!
!           COMPUTE H(k)* . . .*H(2)*H(1)*c(:m,:n) .
!
            do i = 0_i4b, k - blksz, blksz
!
!               APPLY H(i+blksz)* ...*H(i+1) TO c(i+1:m,1:n) FROM THE LEFT
!               WITH A BLOCKED ALGORITHM AS
!
!          ( I + U(i+1:m,:blksz)*inv( lmat(:blksz,:blksz) )*U(i+1:m,:blksz)' )*c(i+1:m,1:n) .
!
                i1 = i + 1_i4b
                i2 = i + blksz
!
                umat(i1:m,1_i4b:blksz) = transpose( mat(i1:i2,i1:m) )
!
!               COMPUTE THE blksz BY blksz UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz) .
!
                do l = 1_i4b, blksz - 1_i4b
!
                    i1 = i + l
!
                    do l2 = l + 1_i4b, blksz
!
                        i2 = i + l2
!
                        lmat(l2,l) = -tau(i2)*dot_product( umat(i2:m,l), umat(i2:m,l2) )
!
                    end do
!                    
                end do
!
                do j = j1, j2
!
!                   COMPUTE U(i+1:m,:blksz)'*c(i+1:m,j) .
!
                    tmp(i+1_i4b:m) = c(i+1_i4b:m,j)
!
                    do l = 1_i4b, blksz
!
                        i1 = i + l
!
                        x(l) = tau(i1)*dot_product( umat(i1:m,l), tmp(i1:m) )
!                    
                    end do
!
!                   SOLVE lmat(:blksz,:blksz)*x(:blksz) = U(i+1:m,:blksz)'*c(i+1:m,j)  AND
!                   COMPUTE c(i+1:m,j) = c(i+1:m,j) + U(i+1:m,:blksz)*x(:blksz) .
!
#ifdef _BLAS
                    call trsv( 'L', 'N', 'U', blksz, lmat, blksz, x, 1_i4b )
!
                    do l = 1_i4b, blksz
!
                        i1 = i + l
!
                        call axpy( m-i1+1_i4b, x(l), umat(i1:m,l), 1_i4b, tmp(i1:m), 1_i4b )
!
                    end do
#else
                    do l = 1_i4b, blksz
!
                        x(l+1_i4b:blksz) = x(l+1_i4b:blksz) - x(l)*lmat(l+1_i4b:blksz,l)
!
                        i1 = i + l
!
                        tmp(i1:m) = tmp(i1:m) + x(l)*umat(i1:m,l)
!
                    end do
#endif
!
                    c(i+1_i4b:m,j) = tmp(i+1_i4b:m)
!                    
                end do
!
            end do
!
!           APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
            i1 = i + 1_i4b
!
            do i = i1, k
!
                tau1 = tau(i)
!
                if ( tau1==zero ) cycle
!
!               APPLY H(i) TO c(i:m,:n) FROM THE LEFT.
!
                tmp(i:m) = mat(i,i:m)
!
                do j = j1, j2
!
                    prod = dot_product( tmp(i:m), c(i:m,j) )*tau1
#ifdef _BLAS
                    call axpy( m-i+1_i4b, prod, tmp(i:m), 1_i4b, c(i:m,j), 1_i4b )
#else
                    c(i:m,j) = c(i:m,j) + prod*tmp(i:m)               
#endif
!
                end do            
!
            end do
!
        end if
!
    else
!
        j3 = j2 - j1 + 1_i4b
!
!       ALLOCATE WORK VARIABLE.
!
        allocate( x2(j3,blksz), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        if ( descending ) then
!
!           COMPUTE c(:m,:n)*H(k)* . . .*H(2)*H(1) .
!
            do i = k - blksz, 0_i4b, -blksz
!
!               APPLY H(i+blksz)* . . .*H(i+1) TO c(1:m,i+1:n) FROM THE RIGHT AS
!
!             c(1:m,i+1:n)*( I + U(i+1:n,:blksz)*inv( lmat(:blksz,:blksz) )*U(i+1:n,:blksz)' ) .
!
                i1 = i + 1_i4b
                i2 = i + blksz
!
                umat(i1:n,1_i4b:blksz) = transpose( mat(i1:i2,i1:n) )
!
!               COMPUTE THE blksz BY blksz UNIT UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz) .
!
                do l = 2_i4b, blksz
!
                    i1 = i + l
!
                    do l2 = 1_i4b, l - 1_i4b
!
                        i2 = i + l2
!
                        lmat(l2,l) = -tau(i2)*dot_product( umat(i1:n,l), umat(i1:n,l2) )
!
                    end do
!                    
                end do
!
!               COMPUTE c(1:m,i+1:n)*U(i+1:n,:blksz) .
!
                do l = 1_i4b, blksz
!
                    i1 = i + l
!
                    x2(1_i4b:j3,l) = umat(i1,l)*c(j1:j2,i1)
!
                    do l2 = i1 + 1_i4b, n
!
                        x2(1_i4b:j3,l) = x2(1_i4b:j3,l) +  umat(l2,l)*c(j1:j2,l2)
!                    
                    end do
!
                    x2(1_i4b:j3,l) = tau(i1)*x2(1_i4b:j3,l)
!                    
                end do
!
!                do l = 1_i4b, blksz
!
!                    i1 = i + l
!
!                    x2(1_i4b:j3,l) = tau(i1)*matmul( c(j1:j2,i1:n), umat(i1:n,l) )
!                    
!                end do
!
!               SOLVE x(1:m,:blksz)*lmat(:blksz,:blksz)' = c(1:m,i+1:n)*U(i+1:n,:blksz) AND
!               COMPUTE c(1:m,i+1:i+blksz-1) = c(1:m,i+1:i+blksz-1) + x(1:m,:blksz)*U(i+1:i+blksz-1,:blksz)' .
!
#ifdef _BLAS
                call trsm( 'R', 'U', 'T', 'U', j3, blksz, one, lmat, blksz, x2, j3 )
!
                do l = blksz, 1_i4b, -1_i4b
!
!                    tmp(1_i4b:j3) = x2(1_i4b:j3,l)
!
                    do l2 = l, blksz-1_i4b
!
                        i2 = i + l2
!
                        call axpy( j3, umat(i2,l), x2(1_i4b:j3,l), 1_i4b, c(j1:j2,i2), 1_i4b )
!                        call axpy( j3, umat(i2,l), tmp(1_i4b:j3), 1_i4b, c(j1:j2,i2), 1_i4b )
!
                    end do
!
                end do
#else
                do l = blksz, 1_i4b, -1_i4b
!
!                    tmp(1_i4b:j3) = x2(1_i4b:j3,l)
!
                    do l2 = 1_i4b, l-1_i4b
!
                        x2(1_i4b:j3,l2) = x2(1_i4b:j3,l2) - x2(1_i4b:j3,l)*lmat(l2,l)
!                        x2(1_i4b:j3,l2) = x2(1_i4b:j3,l2) - tmp(1_i4b:j3)*lmat(l2,l)
!
                    end do
!
                    do l2 = l, blksz-1_i4b
!
                        i2 = i + l2
!
                        c(j1:j2,i2) = c(j1:j2,i2) + x2(1_i4b:j3,l)*umat(i2,l)
!                        c(j1:j2,i2) = c(j1:j2,i2) + tmp(1_i4b:j3)*umat(i2,l)
!
                    end do
!
                end do
#endif
!
!               COMPUTE c(1:m,i+blksz:n) = c(1:m,i+blksz:n) + x(1:m,:blksz)*U(i+blksz:n,:blksz)' .
!
                do j = i + blksz, n
!
                    tmp(1_i4b:j3) = c(j1:j2,j)
!
                    do l = 1_i4b, blksz
!
#ifdef _BLAS
                        call axpy( j3, umat(j,l), x2(1_i4b:j3,l), 1_i4b, tmp(1_i4b:j3), 1_i4b )
#else
                        tmp(1_i4b:j3) = tmp(1_i4b:j3) + x2(1_i4b:j3,l)*umat(j,l)
#endif
!                    
                    end do
!
                    c(j1:j2,j) = tmp(1_i4b:j3)
!                    
                end do
!
            end do
!
!           APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
            i1 = i + blksz
!
            do i = i1, 1_i4b, -1_i4b
!
                tau1 = tau(i)
!
                if ( tau1==zero ) cycle
!
!               APPLY H(i) TO c(:m,i:n) FROM THE RIGHT.
!
                tmp(i:n) = mat(i,i:n)
!
                do j = j1, j2
!
                    prod     = dot_product( tmp(i:n), c(j,i:n) )*tau1
                    c(j,i:n) = c(j,i:n) + prod*tmp(i:n)               
!
                end do            
!
            end do            
!
        else
!
!           COMPUTE c(:m,:n)*H(1)*H(2)* . . .*H(k) .
!
            do i = 0_i4b, k - blksz, blksz
!
!               APPLY H(i+1)* . . .*H(i+blksz) TO c(:m,i+1:n) FROM THE RIGHT AS
!
!             c(:m,i+1:n)*( I + U(i+1i+1:n,:blksz)*inv( lmat(:blksz,:blksz) )*U(i+1:n,:blksz)' ) .
!
                i1 = i + 1_i4b
                i2 = i + blksz
!
                umat(i1:n,1_i4b:blksz) = transpose( mat(i1:i2,i1:n) )
!
!               COMPUTE THE blksz BY blksz UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz) .
!
                do l = 1_i4b, blksz - 1_i4b
!
                    i1 = i + l
!
                    do l2 = l + 1_i4b, blksz
!
                        i2 = i + l2
!
                        lmat(l2,l) = -tau(i2)*dot_product( umat(i2:n,l), umat(i2:n,l2) )
!
                    end do
!                    
                end do
!
!               COMPUTE c(1:m,i+1:n)*U(i+1:n,:blksz) .
!
                do l = 1_i4b, blksz
!
                    i1 = i + l
!
                    x2(1_i4b:j3,l) = umat(i1,l)*c(j1:j2,i1)
!
                    do l2 = i1 + 1_i4b, n
!
                        x2(1_i4b:j3,l) = x2(1_i4b:j3,l) +  umat(l2,l)*c(j1:j2,l2)
!                    
                    end do
!
                    x2(1_i4b:j3,l) = tau(i1)*x2(1_i4b:j3,l)
!                    
                end do
!
!                do l = 1_i4b, blksz
!
!                    i1 = i + l
!
!                    x2(1_i4b:j3,l) = tau(i1)*matmul( c(j1:j2,i1:n), umat(i1:n,l) )
!                    
!                end do
!
!               SOLVE x(1:m,:blksz)*lmat(:blksz,:blksz)' = c(1:m,i+1:n)*U(i+1:n,:blksz) AND
!               COMPUTE c(1:m,i+1:i+blksz-1) = c(1:m,i+1:i+blksz-1) + x(1:m,:blksz)*U(i+1:i+blksz-1,:blksz)' .
!
#ifdef _BLAS
                call trsm( 'R', 'L', 'T', 'U', j3, blksz, one, lmat, blksz, x2, j3 )
!
                do l = 1_i4b, blksz
!
!                    tmp(1_i4b:j3) = x2(1_i4b:j3,l)
!
                    do l2 = l + 1_i4b, blksz
!
                        i2 = i + l2 - 1_i4b
!
                        call axpy( j3, umat(i2,l), x2(1_i4b:j3,l), 1_i4b, c(j1:j2,i2), 1_i4b )
!                        call axpy( j3, umat(i2,l), tmp(1_i4b:j3), 1_i4b, c(j1:j2,i2), 1_i4b )
!
                    end do
!
                end do
#else
                do l = 1_i4b, blksz
!
!                    tmp(1_i4b:j3) = x2(1_i4b:j3,l)
!
                    do l2 = l + 1_i4b, blksz
!
                        x2(1_i4b:j3,l2) = x2(1_i4b:j3,l2) - x2(1_i4b:j3,l)*lmat(l2,l)
!                        x2(1_i4b:j3,l2) = x2(1_i4b:j3,l2) - tmp(1_i4b:j3)*lmat(l2,l)
!
                        i2 = i + l2 - 1_i4b
!
                        c(j1:j2,i2) = c(j1:j2,i2) + x2(1_i4b:j3,l)*umat(i2,l)
!                        c(j1:j2,i2) = c(j1:j2,i2) + tmp(1_i4b:j3)*umat(i2,l)
!
                    end do
!
                end do
#endif
!
!               COMPUTE c(1:m,i+blksz:n) = c(1:m,i+blksz:n) + x(1:m,:blksz)*U(i+blksz:n,:blksz)' .
!
                do j = i + blksz, n
!
                    tmp(1_i4b:j3) = c(j1:j2,j)
!
                    do l = 1_i4b, blksz
!
#ifdef _BLAS
                        call axpy( j3, umat(j,l), x2(1_i4b:j3,l), 1_i4b, tmp(1_i4b:j3), 1_i4b )
#else
                        tmp(1_i4b:j3) = tmp(1_i4b:j3) + x2(1_i4b:j3,l)*umat(j,l)
#endif
!                    
                    end do
!
                    c(j1:j2,j) = tmp(1_i4b:j3)
!                    
                end do
!
            end do
!
!           APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
            i1 = i + 1_i4b
!
            do i = i1, k
!
                tau1 = tau(i)
!
                if ( tau1==zero ) cycle
!
!               APPLY H(i) TO c(:m,i:n) FROM THE RIGHT.
!
                tmp(i:n) = mat(i,i:n)
!
                do j = j1, j2
!
                    prod     = dot_product( tmp(i:n), c(j,i:n) )*tau1
                    c(j,i:n) = c(j,i:n) + prod*tmp(i:n)     
!
                end do            
!
            end do
!
        end if
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( x2 )
!
    end if
!
!$OMP END PARALLEL
!
!
! END OF SUBROUTINE apply_q_lq_rm
! _______________________________
!
    end subroutine apply_q_lq_rm
!
! =========================================================================================
!
    subroutine apply_q_lq_rv( mat, tau, c, trans )
!
! Purpose
! _______
!
!   APPLY_Q_LQ overwrites the general real m vector C with
!
!           Q * C   if TRANS = false, or
!
!           Q' * C  if TRANS = true,
!
!   where Q is a real orthogonal matrix defined as the product of k
!   elementary reflectors
!
!                   Q = H(k) * . . . * H(2) * H(1)
!
!   as returned by LQ_CMP. Q is of order m.
!
!
! Arguments
! _________
!
!   MAT          (INPUT) real(stnd), dimension(:,:)
!                On entry, the i-th row must contain the vector which
!                defines the elementary reflector H(i), for i = 1,2,...,k,
!                as returned by LQ_CMP in the first k rows of its array
!                argument MAT. MAT is not modified by the routine.
!
!   TAU          (INPUT) real(stnd), dimension(:)
!                TAU(i) must contain the scalar factor of the elementary
!                reflector H(i), as returned by LQ_CMP.
!                The size of TAU determines the number k of elementary
!                reflectors whose product defines the matrix Q.
!
!                The size of TAU must verify: size( TAU ) <= min( size(MAT,1) , size(MAT,2) ) .
!
!   C            (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the m vector C.
!
!                On exit, C is overwritten by Q * C or Q' * C.
!
!                The shape of C must verify: size( C ) = size( MAT, 2 ).
!
!   TRANS        (INPUT) logical(lgl)
!                If:
!
!                - TRANS = false : apply Q  (no transpose)
!                - TRANS = true  : apply Q' (transpose)
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the routine  DORML2 in LAPACK.
!
!   For further details on the LQ factorization and its use or the blocked
!   algorithm used here, see
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (2) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in),    dimension(:,:) :: mat
    real(stnd),   intent(in),    dimension(:)   :: tau
    real(stnd),   intent(inout), dimension(:)   :: c
!
    logical(lgl), intent(in)    :: trans
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                       :: tau1, prod
    real(stnd), dimension( size(c) ) :: tmp
!
    integer(i4b) :: m, k, i
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='apply_q_lq'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( c )
    k = size( tau )
!
    call assert( logical( m==size(mat,2), lgl ),            &
                 logical( k<=min(m,size(mat,1)), lgl ),     &
                 name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b .or.  m<=0_i4b ) return
!
    if ( trans ) then
!
!       COMPUTE H(1)*H(2)* . . .*H(k)*c(:m,:n) .
!
        do i = k, 1_i4b, -1_i4b
!
            tau1 = tau(i)
!
            if ( tau1==zero ) cycle
!
!           APPLY H(i) TO c(i:m) FROM THE LEFT.
!
            tmp(i:m) = mat(i,i:m)
!
            prod   = dot_product( tmp(i:m), c(i:m) )*tau1
#ifdef _BLAS
            call axpy( m-i+1_i4b, prod, tmp(i:m), 1_i4b, c(i:m), 1_i4b )
#else
            c(i:m) = c(i:m) + prod*tmp(i:m)
#endif
!
        end do            
!
    else
!
!       COMPUTE H(k)* . . .*H(2)*H(1)*c(:m) .
!
        do i = 1_i4b, k
!
            tau1 = tau(i)
!
            if ( tau1==zero ) cycle
!
!           APPLY H(i) TO c(i:m) FROM THE LEFT.
!
            tmp(i:m) = mat(i,i:m)
!
            prod   = dot_product( tmp(i:m), c(i:m) )*tau1
#ifdef _BLAS
            call axpy( m-i+1_i4b, prod, tmp(i:m), 1_i4b, c(i:m), 1_i4b )
#else
            c(i:m) = c(i:m) + prod*tmp(i:m)     
#endif
!
        end do
!
    end if
!
!
! END OF SUBROUTINE apply_q_lq_rv
! _______________________________
!
    end subroutine apply_q_lq_rv
!
!
! =========================================================================================
!                            QR DECOMPOSITIONS
! =========================================================================================
!
!
    subroutine qr_cmp( mat, diagr, beta )
!
! Purpose
! _______
!
!   QR_CMP computes a QR factorization of a real m-by-n matrix MAT :
!
!                        MAT = Q * R
!
!   where Q is orthogonal and R is upper trapezoidal (upper triangular if m>=n).
!
!
! Arguments
! _________
!
!   MAT      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, the real matrix to be decomposed.
!
!            On exit, the elements above the diagonal of the array contain  
!            the corresponding elements of R. The elements on and below the 
!            diagonal, with the array BETA, represent the orthogonal matrix Q
!            as a product of elementary reflectors, see Further Details.
!
!   DIAGR    (OUTPUT) real(stnd), dimension(:) 
!            On exit, the diagonal elements of the matrix R.
!
!            The size of DIAGR must verify: size( DIAGR ) <= min( size(MAT,1) , size(MAT,2) )
!
!   BETA     (OUTPUT) real(stnd), dimension(:)
!            On exit, the scalars factors of the elementary reflectors.
!
!            The size of BETA must verify:
!
!            - size( BETA ) = size( DIAGR ) <= min( size(MAT,1) , size(MAT,2) )
!
!            See Further Details.
!
!
! Further Details
! _______________
!
!   The matrix Q is represented as a product of elementary reflectors
!
!            Q = H(1) * H(2) * ... * H(k), where k = size( BETA ) = size( DIAGR )
!
!   Each H(i) has the form
!
!            H(i) = I + beta * ( v * v' ) ,
!                      
!   where beta is a real scalar and v is a real m-element vector with v(1:i-1) = 0.
!   v(i:m) is stored on exit in MAT(i:m,i) and beta in BETA(i).
!
!   A blocked algorithm is used for computing the QR factorization. Furthermore,
!   the computations are parallelized if OPENMP is used.
!
!   On exit of QR_CMP, the orthogonal matrix Q (or its first columns) can be computed
!   explicitly by a call to suboutine ORTHO_GEN_QR with arguments MAT and BETA.
!
!   For further details on the QR factorization and its use or the blocked
!   algorithm used here, see:
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (2) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (3) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989:
!           Block reduction of matrices to condensed form for eigenvalue computations. 
!           J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!   (4) Walker, H.F., 1988:
!           Implementation of the GMRES method using Householder transformations.
!           Siam J. Sci. Stat. Comput., Vol. 9, No 1, pp. 152-163.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Utilities,          only : assert_eq, assert
    use Hous_Procedures,    only : hous1
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:) :: mat
    real(stnd),   intent(out),   dimension(:)   :: diagr, beta
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                  :: prod, beta1
    real(stnd), dimension( size(mat,1), blksz ) :: umat, mat2
    real(stnd), dimension( blksz, blksz )       :: lmat
#ifdef _BLAS
    real(stnd), dimension( blksz, blksz ) :: x
#else
    real(stnd), dimension( blksz )        :: x
#endif
!
    integer(i4b) :: m, n, k, k2, i, i1, i2, j, j1, j2, j3, l, l1, l2, p, nblk, rem, blksz2
!
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='qr_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
    n = size( mat, 2 )
!
    k = assert_eq( int(size(beta),i4b),   &
                   int(size(diagr),i4b),  &
                   name_proc              )
!
    call assert( logical( k<=min(m,n), lgl ), name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) return
!
!   COMPUTE THE NUMBER OF BLOCKS.
!
    nblk = k/blksz
    rem  = mod( k, blksz )
!
    if ( rem/=0_i4b ) then
!
        nblk = nblk + 1_i4b
        k2   = k - rem
!
    else
!
        k2   = k - blksz
!
    end if
!
!   INITIALIZE THE UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!   USED IN THE BLOCKED ALGORITHM.
!
    lmat(1_i4b:blksz,1_i4b:blksz) = zero
!
    do i = 1_i4b, blksz
        lmat(i,i) = one
    end do
!
!   INITIALIZE ARRAY FOR STORING HOUSEHOLDER VECTORS.
!
    umat(:m,:blksz) = zero
!
#ifdef _OPENMP
!
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  nblk>1_i4b                          .and.      &
                  i>1_i4b
!
#endif
!
!$OMP PARALLEL IF(test_par)                                   &
!$OMP         ,PRIVATE(i,i1,i2,j,j1,j2,j3,l,l1,l2,p,blksz2,beta1,prod,x,mat2)
!
!       BLOCKED QR DECOMPOSITION OF mat.
!
        do i = 0_i4b, k2, blksz
!
            i1 = i + 1_i4b
            i2 = min( i + blksz, k )
!    
            blksz2 = i2 - i
            p      = m - i
!    
!$OMP SINGLE
!
!           COMPUTE H(i+1)* ...*H(i+blksz) AND DO QR DECOMPOSITION 
!           OF THE i+1 BLOCK .
!
            do j = i1, i2
!    
!               GENERATE ELEMENTARY REFLECTOR H(j) TO ANNIHILATE mat(j+1:m,j) .
!    
                call hous1( mat(j:m,j), beta(j), diagr(j) )
!    
!               APPLY H(j) TO mat(j:m,j+1:i+blksz) FROM THE LEFT.
!
                beta1 = beta(j)
!
                do l = j + 1_i4b, i2
!
                    prod = beta1*dot_product( mat(j:m,j), mat(j:m,l) )
#ifdef _BLAS
                    call axpy( m-j+1_i4b, prod, mat(j:m,j), 1_i4b, mat(j:m,l), 1_i4b )
#else
                    mat(j:m,l) = mat(j:m,l) + prod*mat(j:m,j)
#endif
!    
                end do
!
            end do
!    
!$OMP END SINGLE
!    
            if ( i2<n ) then
!    
!               APPLY H(i+blksz)* ...*H(i+1) TO mat(i+1:m,i+blksz+1:n) FROM THE LEFT
!               WITH A BLOCKED ALGORITHM AS
!
!        ( I + U(i+1:m,:blksz)*inv( lmat(:blksz,:blksz) )*U(i+1:m,:blksz)' )*mat(i+1:m,i+blksz+1:n) .
!
!               FIRST SAVE H(i+1) ...H(i+blksz) FOR LATER USE AND COMPUTE THE blksz BY blksz UNIT
!               LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz) USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do l = 1_i4b, blksz2
!
                    l1 = i + l
!
!                   SAVE SHIFTED HOUSEHOLDER VECTOR mat(i+l:m,i+l) .
!
                    umat(l:p,l) = mat(l1:m,l1)
!
                    do j = l + 1_i4b, blksz2
!
                        l2 = i + j
!
                        lmat(j,l) = -beta(l2)*dot_product( umat(j:p,l), mat(l2:m,l2) )
!
                    end do
!                    
                end do
!
!$OMP END DO
!
!               BLOCK UPDATE OF THE TRAILING MATRIX mat(i+1:m,i+blksz+1:n) .
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do j = i2, n-1_i4b, blksz2
!
                    j1 = j + 1_i4b
                    j2 = min( j+blksz2, n )
                    j3 = j2 - j
!
!                   MAKE A LOCAL COPY OF mat(i+1:m,j1:j2) .
!
                    mat2(:p,:j3) = mat(i1:m,j1:j2)
!
#ifdef _BLAS
!
!                   COMPUTE U(i+1:m,:blksz)'*mat(i+1:m,j1:j2) .
!
                    call gemm( 'T', 'N', blksz2, j3, p, one, umat, m, mat2, m, zero, x, blksz  )
!
                    do l = 1_i4b, j3
!
                        x(:blksz2,l) = beta(i1:i2)*x(:blksz2,l)
!                    
                    end do
!
!                   SOLVE lmat(:blksz,:blksz)*x(:blksz,:j3) = U(i+1:m,:blksz)'*mat(i+1:m,j1:j2) .
!
                    call trsm( 'L', 'L', 'N', 'U', blksz2, j3, one, lmat, blksz, x, blksz )
!
!                   COMPUTE mat(i+1:m,j1:j2) = mat(i+1:m,j1:j2) + U(i+1:m,:blksz)*x(:blksz,:j3) .
!
!                    call gemm( 'N', 'N', p, j3, blksz2, one, umat(:p,1_i4b:blksz2), m,   &
!                               x(:blksz2,1_i4b:j3), blksz, one, mat2(:p,:j3), m   )
!
                    call gemm( 'N', 'N', p, j3, blksz2, one, umat, m, x, blksz, one, mat2, m   )
#else
                    do l = 1_i4b, j3
!
!                       COMPUTE U(i+1:m,:blksz)'*mat(i+1:m,j1:j2) .
!
                        do l2 = 1_i4b, blksz2
!
                            x(l2) = beta(i+l2)*dot_product( umat(l2:p,l2), mat2(l2:p,l) )
!                    
                        end do
!
!                       SOLVE lmat(:blksz,:blksz)*x(:blksz,:j3) = U(i+1:m,:blksz)'*mat(i+1:m,j1:j2) .
!                       COMPUTE mat(i+1:m,j1:j2) = mat(i+1:m,j1:j2) + U(i+1:m,:blksz)*x(:blksz,:j3) .
!
                        do l2 = 1_i4b, blksz2
!
                            x(l2+1_i4b:blksz2) = x(l2+1_i4b:blksz2) - x(l2)*lmat(l2+1_i4b:blksz2,l2)
!
                            mat2(l2:p,l) = mat2(l2:p,l) + x(l2)*umat(l2:p,l2)
!
                        end do
!
                    end do
#endif
!
!                   SAVE UPDATE OF mat(i+1:m,j1:j2) .
!
                    mat(i1:m,j1:j2) = mat2(:p,:j3)           
!                    
                end do
!
!$OMP END DO
!
            end if
!
        end do
!
!$OMP END PARALLEL
!
!
! END OF SUBROUTINE qr_cmp
! ________________________
!
    end subroutine qr_cmp
!
! =========================================================================================
!
    subroutine qrfac( name_proc, syst, kfix, krank, min_norm, diagr, beta, h, tol, ip )
!
! Purpose
! _______
!
!   QRFAC is a low-level subroutine for computing a QR or complete orthogonal factorization
!   of the array section SYST(:m,:n) where m=size(SYST,1) and n<=size(SYST,2).
!
!   The routine first computes a QR factorization with column pivoting of SYST(:m,:n):
!
!                     SYST(:m,:n) * P = Q * R
!
!   , where P is an n-by-n permutation matrix, R is an upper triangular or trapezoidal
!   (i.e., if n>m) matrix and Q is a m-by-m orthogonal matrix.
!
!   The orthogonal transformation Q is then applied to the array section SYST(:m,n+1:):
!
!                     SYST(:m,n+1:) = Q * B
!
!   Then, the rank of SYST(:m,:n) is determined by finding the submatrix R11 of R
!   which is defined as the largest leading submatrix whose estimated condition
!   number, in the 1-norm, is less than 1/TOL or such that abs(R11[j,j])>0 if TOL
!   is absent. The order of R11, krank, is an estimate of the rank of SYST(:m,:n).
!
!   This leads implicitly to the following partition of Q:
!
!   [  Q1  Q2  ]
!
!   where Q1 is an m-by-krank orthonormal matrix and Q2 is m-by-(m-krank) orthonormal matrix
!   orthogonal to Q1, and to the following corresponding partition of R:
!
!   [  R11  R12  ]
!
!   [  R21  R22  ]
!
!   where R11 is a krank-by-krank triangular matrix, R21 is zero by construction,
!   R12 is a full krank-by-(n-krank) matrix and R22 is a full (m-krank)-by-(n-krank)
!   matrix.
!
!   Q1 and Q2 are, respectively, orthonormal bases of the range and null space of SYST(:m,:n).
!
!   If MIN_NORM=true, R22 is considered to be negligible and R12 is annihilated by orthogonal
!   transformations from the right, arriving at the complete orthogonal factorization:
!
!                       SYST(:m,:n) * P = Q * T * Z
!
!   , where P is a n-by-n permutation matrix, Q is a m-by-m orthogonal matrix,
!   Z is a n-by-n orthogonal matrix and T is a m-by-n matrix, which has the form:
!
!   [  T11  T12  ]
!
!   [  T21  T22  ]
!
!   Here T21 (=R21) and T12 are all zero, T22 (=R22) is considered to be negligible and
!   T11 is a krank-by-krank upper triangular matrix.
!   
!   On exit, P is stored compactly in the vector argument IP, krank is stored in the scalar
!   argument KRANK and if:
!   
!   - MIN_NORM=false, QRFAC computes Q and submatrices R11 and R12. Submatrices R11,
!     and R12 are stored in the array section SYST(:krank,:n) and the array argument DIAGR.
!     Q is stored compactly in factored form in the array section SYST(:m,:n) (in its lower
!     triangle) and the array argument BETA.
!   
!   - MIN_NORM=true, QRFAC computes Q, Z and submatrice T11. Submatrice T11 is stored
!     in the array section SYST(:krank,:krank) and the array argument DIAGR.
!     Q is stored compactly in factored form in the array section SYST(:m,:n) (in its lower
!     triangle) and the array argument BETA.
!     Z is stored compactly in factored form in the array sections SYST(:krank,krank+1:n)
!     and H(:krank).
!   
!   See Further Details for more information.
!   
!
! Arguments
! _________
!
!   NAME_PROC    (INPUT) character(len=*)
!                Name of the subroutine calling QRFAC.
!
!   SYST         (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the real m-by-n matrix to be decomposed and, eventually, the
!                the right hand side matrix of an associated least squares problem in
!                the array section SYST(:m,n+1:).
!
!                On exit, SYST(:m,1:n) has been overwritten by details of its QR or (complete) 
!                orthogonal factorization and B is stored in SYST(:m,n+1:).
!
!                See Further Details.
!
!   KFIX         (INPUT) integer(i4b)
!                On entry:
!
!                - KFIX=k, implies that the first k columns of SYST(:m,:n) are
!                  to be forced into the basis. Pivoting is performed only on the
!                  last n-k columns of SYST(:m,:n) if k<min(m,n).
!                - KFIX<=0 can be used when pivoting is desired on all columns
!                  of SYST(:m,:n).
!                - If KFIX<min(m,n) then the optional argument IP must be present
!                  to store the permutation matrix P.
!                - When KFIX>=min(m,n) is used, pivoting is not performed. This is
!                  appropriate when SYST(:m,:n) is known to be of full rank.
!            
!   KRANK        (INPUT/OUTPUT) integer(i4b)
!                On entry, KRANK=n . KRANK must verify:  KRANK <= size( SYST, 2 ).
!            
!                On exit, KRANK contains the effective rank of SYST(:m,:n), i.e.,
!                the order of the submatrix R11, krank. This is the same as the order of
!                the submatrix T11 in the complete orthogonal factorization of SYST(:m,:n).
!
!   MIN_NORM     (INPUT) logical(lgl)
!                On entry:
!
!                - MIN_NORM=true indicates that a complete orthogonal factorization
!                  of SYST(:m,:n) must be computed.
!                - MIN_NORM=false indicates that only a QR factorization (eventually
!                  with column pivoting if KFIX<min(m,n)) of SYST(:m,:n) must be computed.
!
!   DIAGR        (OUTPUT) real(stnd), dimension(:) 
!                On exit, the diagonal elements of the matrix R if MIN_NORM=false or
!                the diagonal elements of the matrix T11 if MIN_NORM=true.
!
!                The diagonal elements of R are stored in DIAGR(:min(m,n)) and the
!                diagonal elements of T11 are stored in DIAGR(:krank) and
!                krank is stored in the real argument KRANK on exit.
!
!                See Further Details.
!
!                The size of DIAGR must verify:
!
!                - size( DIAGR ) >= min( m , n ) =  min( size(SYST,1) , KRANK ).
!
!   BETA         (OUTPUT) real(stnd), dimension(:)
!                On exit, the scalars factors of the elementary reflectors defining Q.
!
!                See Further Details.
!
!                The size of BETA must verify:
!
!                - size( BETA ) >= min( m , n ) = min( size(SYST,1) , KRANK ).
!
!   H            (OUTPUT) real(stnd), dimension(:)
!                On exit, if MIN_NORM=true, the scalars factors of the elementary
!                reflectors defining Z in the complete orthogonal factorization of
!                MAT are stored in H(:krank). On exit, krank is output in the real
!                argument KRANK.
!
!                See Further Details.
!
!                The size of H must verify: size( H ) >= n = KRANK  .
!
!   TOL          (INPUT/OUTPUT, OPTIONAL) real(stnd)
!                If TOL is present and is in [0,1[, then:
!
!                - On entry, the calculations to determine the condition number of
!                  of SYST(:m,:n) in the 1-norm are performed. Then, TOL is used to determine
!                  the effective rank of SYST(:m,:n), which is defined as the order of 
!                  the largest leading triangular submatrix R11 in the QR factorization
!                  with pivoting of SYST(:m,:n), whose estimated condition number is less than 1/TOL.
!                  If TOL=0 is specified the numerical rank of SYST(:m,:n) is determined.
!                - On exit, the reciprocal of the condition number is returned in TOL.
!
!                If TOL is not specified or is outside [0,1[ :
!
!                - The calculations to determine the condition number of MAT are not
!                  performed and crude tests on R(j,j) are done to determine
!                  the rank of MAT. If TOL is present, it is not changed.
!
!   IP           (OUTPUT, OPTIONAL) integer(i4b), dimension(:) 
!                On exit, if IP(j)=k, then the j-th column of SYST(:m,:n) * P was 
!                the k-th column of SYST(:m,:n). 
!
!                IP must be present if KFIX < min( m , n ) = min( SYST,1), KRANK ).
!
!                The size of IP must verify: size( IP ) >= n = KRANK.
!
!
! Further Details
! _______________
!
!   QRFAC is called by the subroutines QR_CMP2, LLSQ_QR_SOLVE, LLSQ_QR_SOLVE2 and
!   SOLVE_LLSQ in modules QR_Procedures and LLSQ_Procedures.
!
!   Since QRFAC is a low level subroutine, no checking of the correctness of the
!   dimensions of the array arguments is performed inside of the subroutine and
!   such checking must be done before calling QRFAC.
!
!   The matrix Q is represented as a product of elementary reflectors
!
!            Q = H(1) * H(2) * ... * H(k), where k = min( m , n ).
!
!   Each H(i) has the form
!
!            H(i) = I + beta * ( v * v' ) ,
!                      
!   where beta is a real scalar and v is a real m-element vector with v(1:i-1) = 0.
!   v(i:m) is stored on exit in SYST(i:m,i) and beta in BETA(i).
!
!   On exit of QRFAC, the orthonormal matrix Q (or its first n columns) can be computed
!   explicitly by a call to suboutine ORTHO_GEN_QR with arguments MAT and BETA.
!
!   The matrix P is represented in the array IP (if present) as follows:
!   If IP(j) = i then the jth column of P is the ith canonical unit vector.
!
!   If MIN_NORM=false, a QR factorization with column pivoting of SYST(:m,:n) is computed
!   and, on exit:
!
!   - The elements above the diagonal of the array SYST(:krank,:krank) contain the 
!     corresponding elements of the triangular matrix R11.
!   - The elements of the diagonal of R11 are stored in the array DIAGR.
!   - The submatrix R12 is stored in SYST(:krank,krank+1:n).
!   - krank is stored in the real argument KRANK.
!
!   If MIN_NORM=true, a complete orthogonal factorization of SYST(:m,:n) is computed.
!   The factorization is obtained by Householder's method. The kth transformation
!   matrix, Z(k), which is used to introduce zeros into the kth row of R,
!   is given in the form
!
!   [  I  0    ]
!
!   [  0  T(k) ]
!
!   where
!
!         T(k) = I + tau * ( u(k) * u(k)' ) and u(k)' = (  1  0  z(k)  )
!
!   tau is a scalar, u(k) is a n-k+1 vector and z(k) is an (n-krank) element vector.
!   tau and z(k) are chosen to annihilate the elements of the kth row of R12.
!
!   The Z n-by-n orthogonal matrix is given by
!
!         Z = Z(1) * Z(2) * ... * Z(krank)
!
!   On exit, if MIN_NORM=true:
!
!   - The scalar tau defining T(k) is returned in the kth element of H and the vector u(k)
!     in the kth row of SYST, such that the elements of z(k) are in SYST(k,krank+1:n).
!     The other elements of u(k) are not stored.
!   - The elements above the diagonal of the array section SYST(:krank,:krank)
!     contain the corresponding elements of the triangular matrix T11. The elements of
!     the diagonal of T11 are stored in the array DIAGR.
!   - krank is stored in the real argument KRANK.
!
!   The computations are parallelized if OPENMP is used. QRFAC uses a blocked "BLAS3" algorithm
!   to compute the QR factorization of the columns of SYST(:m,:n), which are forced to be included
!   in the basis as specified with the KFIX argument. However, if column pivoting is requested, QRFAC
!   uses a standard "BLAS2" algorithm  without any blocking on the columns which must be pivoted and is
!   thus not optimized for computing a full QR factorization with column pivoting of very large matrices.
!
!   QRFAC is an updated version of a subroutine with the same name provided in the
!   reference (1) with improvements suggested in the reference (3).
!
!   For further details, see:
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (2) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (3) Drmac, Z., and Bujanovic, Z., 2006:
!           On the failure of rank revealing QR factorization software - a case study
!           LAPACK Working Note 176.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Num_Constants,   only : lamch
    use Char_Constants,  only : qr_error1, qr_error3
    use Utilities,       only : merror, imaxloc, arth, norm
    use Hous_Procedures, only : hous1, hous2, apply_hous2
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:) :: syst
    real(stnd),   intent(out),   dimension(:)   :: diagr, beta, h
    real(stnd),   intent(inout), optional       :: tol
!
    integer(i4b), intent(in)                          :: kfix
    integer(i4b), intent(inout)                       :: krank
    integer(i4b), intent(out), dimension(:), optional :: ip
!
    logical(lgl), intent(in) :: min_norm
!
    character(len=*), intent(in) :: name_proc
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                          :: xnorm, anorm, rcond, rloss, safmin, &
                                           cl, tau, eps
    real(stnd), dimension(krank)        :: h0
    real(stnd), dimension(size(syst,1)) :: tmp
!
    integer(i4b) :: n, m, ndiag, i, j, jmax, ncol, k, kp1
!
    logical(lgl) :: test
!
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   INITIALIZE WORK VARIABLES FOR THE QR FACTORIZATION.
!    
    n     = krank
    m     = size( syst, 1 )
    ndiag = min( n, m )
    ncol  = size( syst, 2 )
    i     = min( ndiag, kfix )
    k     = max( i, 0_i4b )
    kp1   = k + 1_i4b
!
    if ( present(ip) ) then
!
!       INITIALIZE PERMUTATION VECTOR FOR COLUMN PIVOTING IF REQUIRED.
!
        ip(:n) = arth( 1_i4b, 1_i4b, n )
!
    else
!
        if ( k<ndiag )  call merror( name_proc//qr_error3 )
!
    end if
!    
!   COMPUTE QR FACTORIZATION OF syst(:m,:n) .
!    
    if ( k>0_i4b ) then
!    
!       FIRST COMPUTE QR FACTORIZATION WITHOUT COLUMN PIVOTING OF syst(:m,:kfix)
!       AND APPLY THE CORRESPONDING ORTHOGONAL TRANSFORMATIONS TO syst(:m,kfix+1:ncol) .
!    
        call qr_cmp( syst(:m,:ncol), diagr(:k), beta(:k) )
!    
    end if
!    
    if ( k<ndiag ) then
!    
!       NEXT COMPUTE QR FACTORIZATION WITH COLUMN PIVOTING OF syst(kfix+1:m,kfix+1:n)
!       AND APPLY THE CORRESPONDING ORTHOGONAL TRANSFORMATIONS TO syst(kfix+1:m,n+1:ncol) .
!    
        eps = sqrt( epsilon( eps ) )
!    
!       FIRST COMPUTE COLUMN NORMS OF syst(kfix+1:m,kfix+1:n) .
!
        h0(kp1:n) = norm( syst(kp1:m,kp1:n), dim=2_i4b )
!
!       INITIALIZE PARTIAL COLUMN NORMS.
!
        h(kp1:n) = h0(kp1:n)
!
#ifdef _OPENMP
        i = omp_get_max_threads()
        ompparallel = omp_in_parallel()
        ompnested   = omp_get_nested() .and. ompparallel
        test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                      m*(ncol-k)>=omp_limit               .and.      &
                      (ncol-k)/i>=omp_limit2              .and.      &
                      i>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)        &
!$OMP         ,PRIVATE(i,j,cl,rloss)
!
        do j = kp1, ndiag
!    
!$OMP SINGLE
!    
!           DO COLUMN PIVOTING OF syst(j:m,j:n) .
!    
!           FIND jmax .                
!    
            jmax = imaxloc( h(j:n) ) + j - 1_i4b
!
            tmp(:m) = syst(:m,jmax)
!    
!           DO COLUMN INTERCHANGES IF NEEDED.
!    
            if ( jmax/=j ) then
!    
                i        = ip(j)
                ip(j)    = ip(jmax)
                ip(jmax) = i
!    
                syst(:m,jmax) = syst(:m,j)
!    
                h(jmax) = h(j)
!    
            end if
!    
!           COMPUTE THE J-TH TRANSFORMATION.
!    
            call hous1( tmp(j:m), tau, diagr(j) )
!
!$OMP END SINGLE
!    
!           APPLY THE J-TH TRANSFORMATION TO syst(j:m,j+1:ncol) FROM THE LEFT.
!
!$OMP DO SCHEDULE(STATIC,1)
!
            do i = j+1_i4b, ncol
!
#ifdef _BLAS
                cl = dot( m-j+1_i4b, tmp(j:m), 1_i4b, syst(j:m,i), 1_i4b )*tau
                call axpy( m-j+1_i4b, cl, tmp(j:m), 1_i4b, syst(j:m,i), 1_i4b )
#else
                cl = dot_product( tmp(j:m), syst(j:m,i) )*tau
                syst(j:m,i) = syst(j:m,i) + cl*tmp(j:m)
#endif
!    
!               UPDATE COLUMN NORMS OF syst(j+1:m,i:n) .
!
                if ( i<=n .and. h(i)/=zero ) then
!
                    cl    = abs( syst(j,i) )/h(i)
                    rloss = max( (one-cl)*(one+cl), zero )
                    cl    = rloss*( h(i)/h0(i) )**2
!
                    if ( cl<=eps ) then
!
                        h(i)  = norm( syst(j+1_i4b:m,i) )
                        h0(i) = h(i)
!
                    else
!
                        h(i)  = h(i)*sqrt( rloss )
!
                    end if
!
                end if
!
            end do
!
!$OMP END DO NOWAIT
!    
!           SAVE THE J-TH TRANSFORMATION.
!
!$OMP SINGLE
!
            syst(:m,j) = tmp(:m)
            beta(j)    = tau
!
!$OMP END SINGLE
!
        end do
!
!$OMP END PARALLEL
!
    end if
!    
!   TEST TOLERANCE.
!    
    if ( present(tol) ) then
        test = zero<=tol .and. tol<one
    else
        test = false
    end if
!    
    tmp(:ndiag) = abs( diagr(:ndiag) )
!
!   FIND safmin, THE SAFE MINIMUM, SUCH
!   THAT  1/safmin  DOES NOT OVERFLOW.
!
    safmin = lamch( 's' )
!    
    if ( test ) then
!    
!       DETERMINE THE PSEUDO-RANK USING THE TOLERANCE, tol, AND
!       THE CONDITION NUMBER IN THE 1-NORM OF THE TRIANGULAR FACTOR R
!       OF THE QR FACTORIZATION.
!    
        anorm = zero
        xnorm = zero
        rcond = zero
!
        do j = 1_i4b, ndiag
!
!           CHECK IF abs( R(j,j) ) IS SMALL AND IF R(:j,:j) IS SINGULAR.
!           IF R(:j,:j) IS SINGULAR, EXIT AND THE PSEUDO-RANK IS j-1 .
!
            if ( tmp(j)<safmin ) exit
!
!           COMPUTE 1-NORM OF R(:j,:j).
!
            anorm  = max( anorm, sum(abs(syst(1_i4b:j-1_i4b,j)))+tmp(j) )
!
!           COMPUTE 1-NORM OF inv( R(:j,:j) ).
!
            tmp(j)             = one
            tmp(1_i4b:j-1_i4b) = zero
!
#ifdef _BLAS
            syst(j,j) = diagr(j)
!
            call trsv( 'U', 'N', 'N', j, syst, m, tmp(1_i4b:j), 1_i4b )    
#else
            do i = j, 1_i4b, -1_i4b
                tmp(i)             = tmp(i)/diagr(i)
                tmp(1_i4b:i-1_i4b) = tmp(1_i4b:i-1_i4b) - tmp(i)*syst(1_i4b:i-1_i4b,i)
            end do
#endif
!
            xnorm = max( xnorm, sum(abs(tmp(1_i4b:j))) )
!
!           COMPUTE INVERSE OF CONDITION NUMBER OF R(:j,:j).
!
            cl = one/( anorm*xnorm )
!
!           EXIT IF INVERSE OF CONDITION NUMBER IS LESS THAN THE TOLERANCE.
!
            if ( tol>=cl ) exit
!
            rcond = cl
!
        end do
!
        tol = rcond
!
#ifdef _BLAS
!
!       RESTORE FIRST ELEMENT OF HOUSEHOLDER VECTORS IF BLAS IS USED.
!
        do i = 1_i4b, min(j,ndiag)
            syst(i,i) = one
        end do
#endif
!
    else
!    
!       DETERMINE THE NUMERICAL RANK.
!    
        do j = 1_i4b, ndiag
            if ( tmp(j)<safmin ) exit
        end do
!
    end if
!    
!   OUTPUT RANK OF syst(:m,:n) .
!    
    krank = j - 1_i4b
!    
!   STOP THE PROGRAM IF syst(:m,:n) IS A DEFICIENT MATRIX WITH krank < min(kfix,ndiag) .
!    
    if ( krank<min(kfix,ndiag) ) call merror( name_proc//qr_error1 )
!    
!   IF THE PSEUDORANK krank IS LESS THAN n AND THE MINIMUM_NORM SOLUTION IS REQUIRED,
!   COMPUTE HOUSEHOLDER DECOMPOSITION OF syst(1:krank,:n) .
!    
    if ( krank/=n .and. min_norm ) then
!    
        h(j:n) = zero
!    
        do i = krank, 1_i4b, -1_i4b
!    
            h0(j:n) = syst(i,j:n)
!    
!           COMPUTE THE I-TH HOUSEHOLDER TRANSFORMATION.
!    
            call hous2( diagr(i), h0(j:n), h(i) )
!    
!           APPLY THE I-TH HOUSEHOLDER TRANSFORMATION FROM THE RIGHT.
!    
            call apply_hous2( h0(j:n), h(i),                                        &
                              syst(1_i4b:i-1_i4b,i), syst(1_i4b:i-1_i4b,j:n), false )
!    
            syst(i,j:n) = h0(j:n)
!    
        end do
!    
    end if
!
!
! END OF SUBROUTINE qrfac
! _______________________
!
    end subroutine qrfac
!
! =========================================================================================
!
    subroutine qr_cmp2( mat, diagr, beta, ip, krank, tol, tau )
!
! Purpose
! _______
!
!   QR_CMP2 computes a (complete) orthogonal factorization of a real m-by-n matrix
!   MAT. MAT may be rank-deficient. The routine first computes a QR factorization
!   with column pivoting of MAT:
!
!                     MAT * P = Q * R
!
!   , here P is an n-by-n permutation matrix, R is an upper triangular or trapezoidal
!   (if n>m) matrix and Q is a m-by-m orthogonal matrix.
!
!   R can then be partioned by defining R11 as the largest leading submatrix of R whose
!   estimated condition number, in the 1-norm, is less than 1/TOL (or such that abs(R[j,j])>0
!   if TOL is absent). The order of R11, krank, is the effective rank of MAT.
!
!   This leads implicitly to the following partition of Q:
!
!   [  Q1  Q2  ]
!
!   where Q1 is an m-by-krank orthonormal matrix and Q2 is m-by-(m-krank) orthonormal matrix
!   orthogonal to Q1, and to the following corresponding partition of R:
!
!   [  R11  R12  ]
!
!   [  R21  R22  ]
!
!   where R11 is a krank-by-krank triangular matrix, R21 is zero by construction,
!   R12 is a full krank-by-(n-krank) matrix and R22 is a full (m-krank)-by-(n-krank)
!   matrix.
!
!   Q1 and Q2 are, respectively, orthonormal bases of the range and null space of MAT.
!
!   In a final step, if TAU is present, R22 is considered to be negligible and R12
!   is annihilated  by orthogonal transformations from the right, arriving at the
!   complete orthogonal factorization:
!
!                       MAT * P = Q * T * Z
!
!   , where P is a n-by-n permutation matrix, Q is a m-by-m orthogonal matrix,
!   Z is a n-by-n orthogonal matrix and T is a m-by-n matrix and has the form:
!
!   [  T11  T12  ]
!
!   [  T21  T22  ]
!
!   Here T21 (=R21) and T12 are all zero, T22 (=R22) is considered to be negligible and
!   T11 is a krank-by-krank upper triangular matrix.
!   
!   On exit, P is stored compactly in the vector argument IP and if:
!   
!   - TAU is absent, QR_CMP2 computes Q and submatrices R11, R12. Submatrices R11 and
!     R12 are stored in the array arguments MAT and DIAGR.
!     Q is stored compactly in factored form in the array arguments MAT and BETA.
!   
!   - TAU is present, QR_CMP2 computes Q, Z and submatrice T11. Submatrice T11
!     is stored in the array arguments MAT and DIAGR.
!     Q is stored compactly in factored form in the array arguments MAT and BETA.
!     Z is stored compactly in factored form in the array arguments MAT and TAU.
!   
!   See Further Details for more information on how the partial QR or orthogonal decomposition
!   is stored in MAT on exit.
!   
!
! Arguments
! _________
!
!   MAT      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, the real m-by-n matrix to be decomposed.
!
!            On exit, MAT has been overwritten by details of its (complete) 
!            orthogonal factorization.
!
!            See Further Details.
!
!   DIAGR    (OUTPUT) real(stnd), dimension(:) 
!            On exit, the diagonal elements of the matrix R if TAU is absent or
!            the diagonal elements of the matrix T11 if TAU is present.
!
!            The diagonal elements of R or T11 are stored in DIAGR(1:krank) and
!            krank is stored in the real argument KRANK on exit.
!
!            See Further Details.
!
!            The size of DIAGR must be equal to min( size(MAT,1) , size(MAT,2) ).
!
!   BETA     (OUTPUT) real(stnd), dimension(:)
!            On exit, the scalars factors of the elementary reflectors defining Q.
!
!            See Further Details.
!
!            The size of BETA must be equal to min( size(MAT,1) , size(MAT,2) ).
!
!   IP       (OUTPUT) integer(i4b), dimension(:) 
!            On exit, if IP(j)=k, then the j-th column of MAT*P was 
!            the k-th column of MAT.
!
!            See Further Details.
!
!            The size of IP must be equal to size( MAT, 2 ) = n.
!
!   KRANK    (INPUT/OUTPUT) integer(i4b)
!            On entry:
!
!            - KRANK=k, implies that the first k columns of MAT are
!              to be forced into the basis. Pivoting is performed on the last n-k
!              columns of MAT.
!            - KRANK<=0 can be used when pivoting is desired
!              on all columns of MAT.
!            - When KRANK>=min(m,n) is used, pivoting is not performed at all.
!              This is appropriate when MAT is known to be full rank.
!
!            On exit, KRANK contains the effective rank of MAT, i.e., the order of
!            the submatrix R11, krank. This is the same as the order of the submatrix
!            T11 in the complete orthogonal factorization of MAT.
!
!   TOL      (INPUT/OUTPUT, OPTIONAL) real(stnd)
!            If TOL is present and is in [0,1[, then :
!
!            - On entry, the calculations to determine the condition number of MAT
!              are performed. Then, TOL is used to determine the effective rank
!              of MAT, which is defined as the order of the largest leading
!              triangular submatrix R11 in the QR factorization with pivoting of MAT,
!              whose estimated condition number < 1/TOL. If TOL=0 is specified the
!              numerical rank of MAT is determined.
!            - On exit, the reciprocal of the condition number is returned in TOL.
!
!            If TOL is not specified or is outside [0,1[ :
!
!            - The calculations to determine the condition number of MAT are not
!              performed and crude tests on R(j,j) are done to determine
!              the numerical rank of MAT. If TOL is present, it is not changed.
!
!   TAU      (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            On entry, if TAU is present, a complete orthogonal factorization of
!            MAT is computed. Otherwise, a QR factorization with column pivoting
!            of MAT is computed.
!
!            On exit, the scalars factors of the elementary reflectors defining the
!            orthogonal matrix Z in the complete orthogonal factorization of MAT.
!
!            See Further Details.
!
!            The size of TAU must be equal to min( size(MAT,1) , size(MAT,2) ).
!
!
! Further Details
! _______________
!
!   1) If it is possible that MAT may not be of full rank (i.e., certain columns of MAT are
!      linear combinations of other columns), then the linearly dependent columns can
!      usually be determined by using KRANK=0  and TOL=relative precision of the elements
!      in MAT. If each element is correct to, say, 5 digits then TOL=0.00001 should
!      be used. Also, it may be helpful to scale the columns of MAT so that all elements 
!      are about the same order of magnitude.
!
!
!   2) The matrix Q is represented as a product of elementary reflectors
!
!            Q = H(1) * H(2) * ... * H(k), where k = min( size(MAT,1) , size(MAT,2) )
!
!      Each H(i) has the form
!
!            H(i) = I + beta * ( v * v' ) ,
!                      
!      where beta is a real scalar and v is a real m-element vector with v(1:i-1) = 0.
!      v(i:m) is stored on exit in MAT(i:m,i) and beta in BETA(i).
!
!      On exit of QR_CMP2, the orthonormal matrix Q (or its first n columns) can be computed
!      explicitly by a call to suboutine ORTHO_GEN_QR with arguments MAT and BETA.
!
!
!   3) The matrix P is represented in the array IP as follows:
!      If IP(j) = i then the jth column of P is the ith canonical unit vector.
!
!
!   4) If TAU is absent, a QR factorization with column pivoting of MAT is computed
!      and, on exit:
!
!     - The elements above the diagonal of the array MAT(:krank,:krank) contain the 
!       corresponding elements of the triangular matrix R11.
!     - The elements of the diagonal of R11 are stored in the array DIAGR.
!     - The submatrix R12 is stored in MAT(:krank,krank+1:n).
!     - krank is stored in the real argument KRANK.
!
!
!   5) If TAU is present, a complete orthogonal factorization of MAT is computed.
!      The factorization is obtained by Householder's method. The kth transformation
!      matrix, Z(k), which is used to introduce zeros into the kth row of R (e.g., in
!      R12), is given in the form
!
!      [  I  0    ]
!
!      [  0  T(k) ]
!
!      where
!
!           T(k) = I + tau * ( u(k) * u(k)' ) and u(k)' = (  1  0  z(k)  )
!
!      tau is a scalar, u(k) is a n-k+1 vector and z(k) is an (n-krank) element vector.
!      tau and z(k) are chosen to annihilate the elements of the kth row of R12.
!
!      The Z n-by-n orthogonal matrix is given by
!
!            Z = Z(1) * Z(2) * ... * Z(krank)
!
!      On exit, the scalar tau is returned in the kth element of TAU and the vector u(k)
!      in the kth row of MAT, such that the elements of z(k) are in MAT(k,krank+1:n).
!      On exit:
!
!     - The scalar tau defining T(k) is returned in the kth element of TAU and the vector u(k)
!       in the kth row of MAT, such that the elements of z(k) are in MAT(k,krank+1:n).
!       The other elements of u(k) are not stored.
!     - The elements above the diagonal of the array section MAT(:krank,:krank)
!       contain the corresponding elements of the triangular matrix T11. The elements of
!       the diagonal of T11 are stored in the array DIAGR.
!     - krank is stored in the real argument KRANK.
!
!   The computations are parallelized if OPENMP is used. QR_CMP2 uses a blocked "BLAS3" algorithm
!   to compute the QR factorization of the columns of MAT, which are forced to be included
!   in the basis as specified with the KRANK argument on entry. However, if column pivoting is requested,
!   QR_CMP2 uses a standard "BLAS2" algorithm  without any blocking on the columns which must be pivoted and is
!   thus not optimized for computing a full QR factorization with column pivoting of very large matrices.
!
!   For further details, see:
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (2) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (3) Drmac, Z., and Bujanovic, Z., 2006:
!           On the failure of rank revealing QR factorization software - a case study
!           LAPACK Working Note 176.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Num_Constants, only : machhuge
    use Utilities,     only : assert_eq, arth
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:) :: mat
    real(stnd),   intent(out),   dimension(:)   :: diagr, beta
    real(stnd),   intent(inout),                 optional   :: tol
    real(stnd),   intent(out),   dimension(:),   optional   :: tau
!
    integer(i4b), intent(inout)               :: krank
    integer(i4b), intent(out),   dimension(:) :: ip
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: tau_tmp(size(ip))
!
    integer(i4b) :: nlig, ncol, ndiag, kfix 
!
    logical(lgl) :: comp_min_norm
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='qr_cmp2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    nlig = size( mat, 1 )
!
    ncol =  assert_eq( int(size(mat,2),i4b) ,   &
                       int(size(ip),i4b)    ,   &
                       name_proc )
!
    if ( present(tau) ) then
!
        ndiag =  assert_eq( min(ncol,nlig)       ,   &
                            int(size(beta),i4b)  ,   &
                            int(size(diagr),i4b) ,   &
                            int(size(tau),i4b)   ,   &
                            name_proc )
!
        comp_min_norm = true
        tau(:ndiag) = zero
!
    else
!
        ndiag =  assert_eq( min(ncol,nlig)       ,   &
                            int(size(beta),i4b)  ,   &
                            int(size(diagr),i4b) ,   &
                            name_proc )
!
        comp_min_norm = false
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( ndiag<=0_i4b .or. all(mat(:nlig,:ncol)==zero) ) then
!
        krank = 0_i4b
        diagr(:ndiag) = zero
        beta(:ndiag)  = zero
        ip(:ncol) = arth( 1_i4b, 1_i4b, ncol )
!
        if ( present(tol) ) tol = machhuge
!
        return
!
    end if
!    
!   COMPUTE QR FACTORIZATION WITH COLUMN PIVOTING OF MAT .
!    
    kfix = max( krank, 0_i4b )
    krank = ncol
!    
    call qrfac( name_proc, mat, kfix, krank, comp_min_norm, diagr, beta, tau_tmp,     &
                tol=tol, ip=ip )
!
!   OUTPUT OPTIONAL ARGUMENT tau ON EXIT, IF REQUIRED.
!
    if ( comp_min_norm ) tau(:krank) = tau_tmp(:krank)
!
!
! END OF SUBROUTINE qr_cmp2
! _________________________
!
    end subroutine qr_cmp2
!
! =========================================================================================
!
    subroutine ortho_gen_qr( mat, beta )
!
! Purpose
! _______
!
!   ORTHO_GEN_QR generates an m-by-n real matrix with orthonormal columns, which is
!   defined as the first n columns of a product of k elementary reflectors of order m
!
!            Q = H(1) * H(2) * ... * H(k)
!
!   as returned by QR_CMP or QR_CMP2.
!
!
! Arguments
! _________
!
!   MAT      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, the i-th column must contain the vector which
!            defines the elementary reflector H(i), for i = 1,2,...,k,
!            as returned by QR_CMP (or QR_CMP2) in the first k columns
!            of its array argument MAT.
!
!            On exit, the first n columns of Q.
!
!            The shape of MAT must verify: size( MAT, 2 ) <= size( MAT, 1 ).
!
!   BETA     (INPUT) real(stnd), dimension(:)
!            BETA(i) must contain the scalar factor of the elementary
!            reflector H(i), as returned by QR_CMP (or QR_CMP2).
!            The size of BETA determines the number k of elementary reflectors
!            whose product defines the matrix Q.
!
!            The size of BETA must verify: size( BETA ) <= ( MAT, 2 ) .
!
!
! Further Details
! _______________
!
!   This subroutine used a blocked algorithm for agregating the Householder
!   transformations (e.g. the elementary reflectors) stored in the lower triangle
!   of MAT and generating the orthogonal matrix Q of the QR factorization.
!
!   Furthermore, the computations are parallelized if OPENMP is used.
!
!   For further details on the QR factorization and its use or the blocked
!   algorithm, see:
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (2) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (3) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989:
!           Block reduction of matrices to condensed form for eigenvalue computations. 
!           J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!   (4) Walker, H.F., 1988:
!           Implementation of the GMRES method using Householder transformations.
!           Siam J. Sci. Stat. Comput., Vol. 9, No 1, pp. 152-163.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:) :: mat
    real(stnd),   intent(in),    dimension(:)   :: beta
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                  :: prod, beta1
    real(stnd), dimension( size(mat,1) )        :: tmp
    real(stnd), dimension( size(mat,1), blksz ) :: umat, mat2
    real(stnd), dimension( blksz, blksz )       :: lmat
#ifdef _BLAS
    real(stnd), dimension( blksz, blksz ) :: x
#else
    real(stnd), dimension( blksz )        :: x
#endif
!
    integer(i4b) :: m, n, k, i, i0, i1, i2, l, l2, j, j1, j2, j3, p, blksz2
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='ortho_gen_qr'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
    n = size( mat, 2 )
    k = size( beta )
!
    call assert( logical( (k<=n) .and. (m>=n), lgl ), name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) return
!
!   INITIALIZE COLUMNS k+1:n OF mat TO COLUMNS OF THE UNIT MATRIX.
!
    mat(:m,k+1_i4b:n) = zero
!
    do i = k+1_i4b, n
        mat(i,i) = one
    end do
!
!   INITIALIZE THE UNIT LOWER OR UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!   USED IN THE BLOCKED ALGORITHM.
!
    lmat(1_i4b:blksz,1_i4b:blksz) = zero
!
    do i = 1_i4b, blksz
        lmat(i,i) = one
    end do
!
!   INITIALIZE ARRAY FOR STORING HOUSEHOLDER VECTORS.
!
    umat(:m,:blksz) = zero
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*m>=omp_limit                      .and.      &
                  n/i>=omp_limit2                     .and.      &
                  i>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)                               &
!$OMP         ,PRIVATE(i,i0,i1,i2,l,l2,j,j1,j2,j3,p,blksz2,beta1,prod,x,tmp,mat2)
!
    do i = k - blksz, -blksz+1_i4b, -blksz
!    
!       APPLY H(i+1)* ...*H(i+blksz) TO mat(i+1:m,i+blksz+1:n) FROM THE LEFT
!       WITH A BLOCKED ALGORITHM AS
!
!    ( I + U(i+1:m,:blksz)*inv( lmat(:blksz,:blksz) )*U(i+1:m,:blksz)' )*mat(i+1:m,i+blksz+1:n) .
!
        blksz2 = min( blksz, i+blksz )
        i0     = max( 0_i4b, i )
!
        i1 = i0 + 1_i4b
        i2 = i0 + blksz2
        p  = m - i0
!
!       FIRST SAVE H(i+1) ...H(i+blksz) FOR LATER USE AND COMPUTE THE blksz BY blksz UNIT UPPER
!       TRIANGULAR MATRIX lmat(:blksz,:blksz) USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
        do l = 1_i4b, blksz2
!
            j1 = i0 + l
!
            umat(l:p,l) = mat(j1:m,j1)
!
            do l2 = 1_i4b, l - 1_i4b
!
                j2 = i0 + l2
!
                lmat(l2,l) = -beta(j2)*dot_product( umat(l:p,l), mat(j1:m,j2) )
!
            end do
!                    
        end do
!
!$OMP END DO
!
!       BLOCK UPDATE OF THE TRAILING MATRIX mat(i+1:m,i+blksz+1:n) .
!
!$OMP DO SCHEDULE(STATIC,1)
!
        do j = i2, n-1_i4b, blksz2
!
            j1 = j + 1_i4b
            j2 = min( j+blksz2, n )
            j3 = j2 - j
!
!           MAKE A LOCAL COPY OF mat(i0+1_i4b:m,j1:j2) .
!
            mat2(:p,:j3) = mat(i1:m,j1:j2)
!
#ifdef _BLAS
!
!           COMPUTE U(i+1:m,:blksz)'*mat(i+1:m,j1:j2) .
!
            call gemm( 'T', 'N', blksz2, j3, p, one, umat, m, mat2, m, zero, x, blksz  )
!
            do l = 1_i4b, j3
!
                x(:blksz2,l) = beta(i1:i2)*x(:blksz2,l)
!                    
            end do
!
!           SOLVE lmat(:blksz,:blksz)*x(:blksz,:j3) = U(i+1:m,:blksz)'*mat(i+1:m,j1:j2) .
!
            call trsm( 'L', 'U', 'N', 'U', blksz2, j3, one, lmat, blksz, x, blksz )
!
!           COMPUTE mat(i+1:m,j1:j2) = mat(i+1:m,j1:j2) + U(i+1:m,:blksz)*x(:blksz,:j3) .
!
!            call gemm( 'N', 'N', p, j3, blksz2, one, umat(:p,1_i4b:blksz2), m,   &
!                       x(:blksz2,1_i4b:j3), blksz, one, mat2(:p,:j3), m   )
!
            call gemm( 'N', 'N', p, j3, blksz2, one, umat, m, x, blksz, one, mat2, m   )
#else
!
            do l = 1_i4b, j3
!
!               COMPUTE U(i+1:m,:blksz)'*mat(i+1:m,j1:j2) .
!
                do l2 = 1_i4b, blksz2
!
                    x(l2) = beta(i0+l2)*dot_product( umat(l2:p,l2), mat2(l2:p,l) )
!                    
                end do
!
!               SOLVE lmat(:blksz,:blksz)*x(:blksz,:j3) = U(i+1:m,:blksz)'*mat(i+1:m,j1:j2) .
!               COMPUTE mat(i+1:m,j1:j2) = mat(i+1:m,j1:j2) + U(i+1:m,:blksz)*x(:blksz,:j3) .
!
                do l2 = blksz2, 1_i4b, -1_i4b
!
                    x(1_i4b:l2-1_i4b) = x(1_i4b:l2-1_i4b) - x(l2)*lmat(1_i4b:l2-1_i4b,l2)
!
                    mat2(l2:p,l) = mat2(l2:p,l) + x(l2)*umat(l2:p,l2)
!
                end do
!
            end do
#endif
!
!           SAVE UPDATE OF mat(i0+1_i4b:m,j1:j2) .
!
            mat(i1:m,j1:j2) = mat2(:p,:j3)           
!                    
        end do
!
!$OMP END DO NOWAIT
!    
!       APPLY H(i+1)* ...*H(i+blksz) TO mat(i+1:m,i+1:i+blksz) FROM THE LEFT
!       WITH A SERIAL ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
        do l = 1_i4b, blksz2
!
            i1 = i0 + l
!
            beta1 = beta(i1)
!
!           APPLY H(i1) TO mat(i1:m,i1) FROM THE LEFT.
!
            tmp(1_i4b:i1-1_i4b) = zero
            tmp(i1)             = one + beta1
            tmp(i1+1_i4b:m)     = beta1*umat(l+1_i4b:p,l)
!
            do l2 = l - 1_i4b, 1_i4b, -1_i4b
!
                i2 = i0 + l2
!
!               APPLY H(i2) TO mat(i2:m,i1).
!
                beta1 = beta(i2)
!
                prod = beta1*dot_product( umat(l2+1_i4b:p,l2), tmp(i2+1_i4b:m) )
#ifdef _BLAS
                call axpy( m-i2+1_i4b, prod, umat(l2:p,l2), 1_i4b, tmp(i2:m), 1_i4b )
#else
                tmp(i2:m) = tmp(i2:m) + prod*umat(l2:p,l2)
#endif
!    
            end do
!    
            mat(1_i4b:m,i1) = tmp(1_i4b:m)
!    
        end do
!    
!$OMP END DO
!
    end do
!
!$OMP END PARALLEL
!
!
! END OF SUBROUTINE ortho_gen_qr
! ______________________________
!
    end subroutine ortho_gen_qr
!
! =========================================================================================
!
    subroutine apply_q_qr_rm( mat, beta, c, left, trans  )
!
! Purpose
! _______
!
!   APPLY_Q_QR overwrites the general real m-by-n matrix C with
!
!            Q * C   if LEFT = true  and TRANS = false, or
!
!            Q' * C  if LEFT = true  and TRANS = true,  or
!
!            C * Q   if LEFT = false and TRANS = false, or
!
!            C * Q'  if LEFT = false and TRANS = true,
!
!   where Q is a real orthogonal matrix defined as the product of k
!   elementary reflectors
!
!                    Q = H(1) * H(2) * . . . * H(k)
!
!   as returned by QR_CMP or QR_CMP2. Q is of order m if LEFT = true and of order n
!   if LEFT = false.
!
!
! Arguments
! _________
!
!   MAT          (INPUT) real(stnd), dimension(:,:)
!                On entry, the i-th column must contain the vector which
!                defines the elementary reflector H(i), for i = 1,2,...,k,
!                as returned by QR_CMP (or QR_CMP2) in the first k columns of its array
!                argument MAT. MAT is not modified by the routine.
!
!   BETA         (INPUT) real(stnd), dimension(:)
!                BETA(i) must contain the scalar factor of the elementary
!                reflector H(i), as returned by QR_CMP (or QR_CMP2).
!                The size of BETA determines the number k of elementary
!                reflectors whose product defines the matrix Q.
!
!                The size of BETA must verify:
!
!                - size( BETA ) <= min( size(MAT,1) , size(MAT,2) ) .
!
!   C            (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the m by n matrix C.
!
!                On exit, C is overwritten by Q * C or Q' * C or C * Q' or C * Q.
!
!                The shape of C must verify:
!
!                - if LEFT = true,  size( C, 1 ) = size( MAT, 1 )
!                - if LEFT = false, size( C, 2 ) = size( MAT, 1 ) .
!
!   LEFT         (INPUT) logical(lgl)
!                If:
!
!                - LEFT = true  : apply Q or Q' from the left
!                - LEFT = false : apply Q or Q' from the right
!
!   TRANS        (INPUT) logical(lgl)
!                If:
!
!                - TRANS = false : apply Q  (no transpose)
!                - TRANS = true  : apply Q' (transpose)
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the routine DORM2R in LAPACK.
!
!   This subroutine used a blocked algorithm for agregating the Householder
!   transformations (e.g. the elementary reflectors) stored in the lower triangle
!   of MAT and applying the orthogonal matrix Q of the QR factorization to the
!   real m-by-n matrix C.
!
!   Furthermore, the computations are parallelized if OPENMP is used.
!
!   For further details on the QR factorization and its use or the blocked
!   algorithm used here, see:
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (2) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (3) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989:
!           Block reduction of matrices to condensed form for eigenvalue computations. 
!           J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!   (4) Walker, H.F., 1988:
!           Implementation of the GMRES method using Householder transformations.
!           Siam J. Sci. Stat. Comput., Vol. 9, No 1, pp. 152-163.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Char_Constants,   only : allocate_error
    use Utilities,        only : assert, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in),    dimension(:,:) :: mat
    real(stnd),   intent(in),    dimension(:)   :: beta
    real(stnd),   intent(inout), dimension(:,:) :: c
!
    logical(lgl), intent(in)    :: left, trans
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                           :: beta1, prod
    real(stnd), dimension( max( size(c,1), size(c,2) ) ) :: tmp
    real(stnd), dimension( blksz, blksz )                :: lmat
    real(stnd), dimension( :, : ), allocatable           :: umat, x, c2
!
    integer(i4b) :: m, n, nm, nq, k, i, i1, i2, j, j1, j2, j3, l, l2, p, p2, q, nblk, rem
    integer      :: iok
!
    logical(lgl) :: ascending
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='apply_q_qr'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( c, 1 )
    n = size( c, 2 )
    nm = n*m
!
    if ( left ) then
        nq = m
        p  = n
    else
        nq = n
        p  = m
    end if
!
    k = size( beta )
!
    call assert( logical( nq==size(mat,1), lgl ),             &
                 logical( k<=min(nq,size(mat,2)), lgl ),      &
                 name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b .or.  nm<=0_i4b ) return
!
    ascending = (left.and.trans) .or. .not.(left.or.trans)
!
!   COMPUTE THE NUMBER OF BLOCKS FOR THE INPUT MATRIX.
!
    nblk = p/blksz
    rem  = mod( p, blksz )
!
    if ( rem/=0_i4b ) then
!
        nblk = nblk + 1_i4b
        p2   = p - rem
!
    else
!
        p2   = p - blksz
!
    end if
!
!   INITIALIZE THE UNIT LOWER OR UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!   USED IN THE BLOCKED ALGORITHM.
!
    lmat(1_i4b:blksz,1_i4b:blksz) = zero
!
    do i = 1_i4b, blksz
        lmat(i,i) = one
    end do
!
!   ALLOCATE AND INITIALIZE (SHARED) WORK ARRAY FOR STORING HOUSEHOLDER
!   VECTORS IF left = true .
!
    allocate( umat(nq,blksz), stat = iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    umat(:nq,:blksz) = zero
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  nm>=omp_limit                       .and.      &
                  nblk>1_i4b                          .and.      &
                  i>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)                                             &
!$OMP         ,PRIVATE(iok,i,i1,i2,l,l2,j,j1,j2,j3,q,beta1,prod,x,c2,tmp)
!
    if ( left ) then
#ifdef _BLAS
!
!       ALLOCATE (PRIVATE) WORK ARRAYS.
!
        allocate( x(blksz,blksz), c2(m,blksz), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
#endif
!
        if ( ascending ) then
!
!           COMPUTE H(k)* . . .*H(2)*H(1)*c(:m,:n) .
!
            do i = 0_i4b, k - blksz, blksz
!
!               APPLY H(i+blksz)* . . . *H(i+1) TO c(i+1:m,:n) FROM THE LEFT AS
!
!             ( I + U(i+1:m,:blksz)*inv( lmat(:blksz,:blksz) )*U(i+1:m,:blksz)' )*c(i+1:m,:n) .
!
                q = m - i
!
!               COMPUTE THE blksz BY blksz UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz) .
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do l = 1_i4b, blksz
!
                    i1 = i + l
!
!                   SAVE SHIFTED HOUSEHOLDER VECTOR mat(i1:m,i1) .
!
                    umat(l:q,l) = mat(i1:m,i1)
!
                    do l2 = l + 1_i4b, blksz
!
                        i2 = i + l2
!
                        lmat(l2,l) = -beta(i2)*dot_product( mat(i2:m,i1), mat(i2:m,i2) )
!
                    end do
!                    
                end do
!
!$OMP END DO
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do j = 0_i4b, p2, blksz
!
                    j1 = j + 1_i4b
                    j2 = min( j + blksz, p )
!
#ifdef _BLAS
                    j3 = j2 - j
!
                    i1 = i + 1_i4b
                    i2 = i + blksz
!
                    c2(:q,:j3) = c(i1:m,j1:j2)
!
!                   COMPUTE U(i+1:m,:blksz)'*c(i+1:m,j1:j2) .
!
                    call gemm( 'T', 'N', blksz, j3, q, one, umat, m, c2, m, zero, x, blksz  )
!
                    do l = 1_i4b, j3
!
                        x(:blksz,l) = beta(i1:i2)*x(:blksz,l)
!                    
                    end do
!
!                   SOLVE lmat(:blksz,:blksz)*x(:blksz,:j3) = U(i+1:m,:blksz)'*c(i+1:m,j1:j2) .
!
                    call trsm( 'L', 'L', 'N', 'U', blksz, j3, one, lmat, blksz, x, blksz )
!
!                   COMPUTE c(i+1:m,j1:j2) = c(i+1_:m,j1:j2) + U(i+1:m,:blksz)*x(:blksz,:j3) .
!
                    call gemm( 'N', 'N', q, j3, blksz, one, umat, m, x, blksz, one, c2, m   )
!
                    c(i1:m,j1:j2) = c2(:q,:j3)
!
#else
!                   COMPUTE U(i+1:m,:blksz)'*c(i+1:m,j1:j2) .
!
                    do l = j1, j2
!
!                       COMPUTE U(i+1:m,:blksz)'*c(i+1:m,j1:j2) .
!
                        do l2 = 1_i4b, blksz
!
                            i1 = i + l2
!
                            tmp(l2) = beta(i1)*dot_product( umat(l2:q,l2), c(i1:m,l) )
!                    
                        end do
!
!                       SOLVE lmat(:blksz,:blksz)*x(:blksz,:j3) = U(i+1:m,:blksz)'*c(i+1:m,j1:j2) .
!                       COMPUTE c(i+1:m,j1:j2) = c(i+1:m,j1:j2) + U(i+1:m,:blksz)*x(:blksz,:j3) .
!
                        do l2 = 1_i4b, blksz
!
                            tmp(l2+1_i4b:blksz) = tmp(l2+1_i4b:blksz) - tmp(l2)*lmat(l2+1_i4b:blksz,l2)
!
                            i1 = i + l2
!
                            c(i1:m,l) = c(i1:m,l) + tmp(l2)*umat(l2:q,l2)
!
                        end do
!
                    end do
#endif
!
                end do
!
!$OMP END DO
!
            end do
!
!           APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
            i1 = i + 1_i4b
!
            do i = i1, k
!
                beta1 = beta(i)
!
                if ( beta1==zero ) cycle
!
!               APPLY H(i) TO c(i:m,:n) FROM THE LEFT.
!
                q = m - i + 1_i4b
!
                tmp(1_i4b:q) = mat(i:m,i)
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do j = 0_i4b, p2, blksz
!
                    j1 = j + 1_i4b
                    j2 = min( j + blksz, p )
!
#ifdef _BLAS
                    j3 = j2 - j
!
                    c2(:q,:j3) = c(i:m,j1:j2)
!
                    call gemv( 'T', q, j3, beta1, c2, m, tmp(1_i4b:q), 1_i4b, zero, x(:j3,1_i4b), 1_i4b )
!
                    call ger( q, j3, one, tmp(1_i4b:q), 1_i4b, x(:j3,1_i4b), 1_i4b, c2, m )
!
                    c(i:m,j1:j2) = c2(:q,:j3)
#else
                    do l = j1, j2
!
                        prod = dot_product( tmp(1_i4b:q), c(i:m,l) )*beta1
!
                        c(i:m,l) = c(i:m,l) + prod*tmp(1_i4b:q)        
!
                    end do            
#endif
!
                end do
!
!$OMP END DO
!
            end do
!
        else
!
!           COMPUTE H(1)*H(2)* . . .*H(k)*c(:m,:n) .
!
            do i = k - blksz, 0_i4b, -blksz
!
!               APPLY H(i+1)* . . . *H(i+blksz) TO c(i+1:m,:n) FROM THE LEFT AS
!
!             ( I + U(i+1:m,:blksz)*inv( lmat(:blksz,:blksz) )*U(i+1:m,:blksz)' )*c(i+1:m,:n) .
!
                q = m - i
!
!               COMPUTE THE blksz BY blksz UNIT UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz) .
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do l = 1_i4b, blksz
!
                    i1 = i + l
!
!                   SAVE SHIFTED HOUSEHOLDER VECTOR mat(i1:m,i1) .
!
                    umat(l:q,l) = mat(i1:m,i1)
!
                    do l2 = 1_i4b, l - 1_i4b
!
                        i2 = i + l2
!
                        lmat(l2,l) = -beta(i2)*dot_product( mat(i1:m,i1), mat(i1:m,i2) )
!
                    end do
!                    
                end do
!
!$OMP END DO
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do j = 0_i4b, p2, blksz
!
                    j1 = j + 1_i4b
                    j2 = min( j + blksz, p )
!
#ifdef _BLAS
                    j3 = j2 - j
!
                    i1 = i + 1_i4b
                    i2 = i + blksz
!
                    c2(:q,:j3) = c(i1:m,j1:j2)
!
!                   COMPUTE U(i+1:m,:blksz)'*c(i+1:m,j1:j2) .
!
                    call gemm( 'T', 'N', blksz, j3, q, one, umat, m, c2, m, zero, x, blksz  )
!
                    do l = 1_i4b, j3
!
                        x(:blksz,l) = beta(i1:i2)*x(:blksz,l)
!                    
                    end do
!
!                   SOLVE lmat(:blksz,:blksz)*x(:blksz,:j3) = U(i+1:m,:blksz)'*c(i+1:m,j1:j2) .
!
                    call trsm( 'L', 'U', 'N', 'U', blksz, j3, one, lmat, blksz, x, blksz )
!
!                   COMPUTE c(i+1:m,j1:j2) = c(i+1_:m,j1:j2) + U(i+1:m,:blksz)*x(:blksz,:j3) .
!
                    call gemm( 'N', 'N', q, j3, blksz, one, umat, m, x, blksz, one, c2, m   )
!    
                    c(i1:m,j1:j2) = c2(:q,:j3)
#else
                    do l = j1, j2
!
!                       COMPUTE U(i+1:m,:blksz)'*c(i+1:m,j1:j2) .
!
                        do l2 = 1_i4b, blksz
!
                            i1 = i + l2
!
                            tmp(l2) = beta(i1)*dot_product( umat(l2:q,l2), c(i1:m,l) )
!                    
                        end do
!
!                       SOLVE lmat(:blksz,:blksz)*x(:blksz,:j3) = U(i+1:m,:blksz)'*c(i+1:m,j1:j2) .
!                       COMPUTE c(i+1:m,j1:j2) = c(i+1:m,j1:j2) + U(i+1:m,:blksz)*x(:blksz,:j3) .
!
                        do l2 = blksz, 1_i4b, -1_i4b
!
                            tmp(1_i4b:l2-1_i4b) = tmp(1_i4b:l2-1_i4b) - tmp(l2)*lmat(1_i4b:l2-1_i4b,l2)
!
                            i1 = i + l2
!
                            c(i1:m,l) = c(i1:m,l) + tmp(l2)*umat(l2:q,l2)
!
                        end do
!
                    end do
#endif
!
                end do
!
!$OMP END DO
!
            end do
!
!           APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
            i1 = i + blksz
!
            do i = i1, 1_i4b, -1_i4b
!
                beta1 = beta(i)
!
                if ( beta1==zero ) cycle
!
!               APPLY H(i) TO c(i:m,:n) FROM THE LEFT.
!
                q = m - i + 1_i4b
!
                tmp(1_i4b:q) = mat(i:m,i)
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do j = 0_i4b, p2, blksz
!
                    j1 = j + 1_i4b
                    j2 = min( j + blksz, p )
!
#ifdef _BLAS
                    j3 = j2 - j
!
                    c2(:q,:j3) = c(i:m,j1:j2)
!
                    call gemv( 'T', q, j3, beta1, c2, m, tmp(1_i4b:q), 1_i4b, zero, x(:j3,1_i4b), 1_i4b )
!
                    call ger( q, j3, one, tmp(1_i4b:q), 1_i4b, x(:j3,1_i4b), 1_i4b, c2, m )
!
                    c(i:m,j1:j2) = c2(:q,:j3)
#else
                    do l = j1, j2
!
                        prod = dot_product( tmp(1_i4b:q), c(i:m,l) )*beta1
!
                        c(i:m,l) = c(i:m,l) + prod*tmp(1_i4b:q)         
!
                    end do            
#endif
!
                end do
!
!$OMP END DO
!
            end do
!
        end if
#ifdef _BLAS
!
!       DEALLOCATE (PRIVATE) WORK ARRAYS.
!
        deallocate( x, c2 )
#endif
!
    else
!
!       ALLOCATE (PRIVATE) WORK ARRAYS.
!
        allocate( x(blksz,blksz), c2(blksz,n), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        if ( ascending ) then
!
!           COMPUTE c(:m,:n)*H(1)*H(2)* . . .*H(k) .
!
            do i = 0_i4b, k - blksz, blksz
!
!               APPLY H(i+1)* . . .*H(i+blksz) TO c(:m,i+1:n) FROM THE RIGHT AS
!
!             c(:m,i+1:n)*( I + U(i+1:n,:blksz)*inv( lmat(:blksz,:blksz) )*U(i+1:n,:blksz)' ) .
!
                q = n - i
!
!               COMPUTE THE blksz BY blksz UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz) .
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do l = 1_i4b, blksz
!
                    i1 = i + l
!
!                   SAVE SHIFTED HOUSEHOLDER VECTOR mat(i1:m,i1) .
!
                    umat(l:q,l) = mat(i1:n,i1)
!
                    do l2 = l + 1_i4b, blksz
!
                        i2 = i + l2
!
                        lmat(l2,l) = -beta(i2)*dot_product( mat(i2:n,i1), mat(i2:n,i2) )
!
                    end do
!                    
                end do
!
!$OMP END DO
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do j = 0_i4b, p2, blksz
!
                    j1 = j + 1_i4b
                    j2 = min( j + blksz, p )
                    j3 = j2 - j
!
                    c2(:j3,:q) = c(j1:j2,i+1_i4b:n)
!
#ifdef _BLAS
!                   COMPUTE c(j1:j2,i+1:n)*U(i+1:n,:blksz) .
!
                    call gemm( 'N', 'N', j3, blksz, q, one, c2, blksz, umat, n, zero, x, blksz  )
!
                    do l = 1_i4b, blksz
!
                        l2 = i + l
!
                        prod = beta(l2)
!
                        x(1_i4b:j3,l) = prod*x(1_i4b:j3,l)
!                    
                    end do
!
!                   SOLVE x(j1:j2,:blksz)*lmat(:blksz,:blksz)' = c(j1:j2,i+1:n)*U(i+1:n,:blksz) .
!
                    call trsm( 'R', 'L', 'T', 'U', j3, blksz, one, lmat, blksz, x, blksz )
!
!                   COMPUTE c(j1:j2,i+1:n) = c(j1:j2,i+1:n) + x(j1:j2,:blksz)*U(i+1:n,:blksz)' .
!
                    call gemm( 'N', 'T', j3, q,  blksz, one, x, blksz, umat, n, one, c2, blksz )
#else
!                   COMPUTE c(j1:j2,i+1:n)*U(i+1:n,:blksz) .
!
                    do l = 1_i4b, blksz
!
                        i1 = i + l
!
                        x(1_i4b:j3,l) = umat(l,l)*c2(1_i4b:j3,l)
!
                        do l2 = l + 1_i4b, q
!
                            x(1_i4b:j3,l) = x(1_i4b:j3,l) +  umat(l2,l)*c2(1_i4b:j3,l2)
!                    
                        end do
!
                        x(1_i4b:j3,l) = beta(i1)*x(1_i4b:j3,l)
!                    
                    end do
!
!                   SOLVE x(j1:j2,:blksz)*lmat(:blksz,:blksz)' = c(j1:j2,i+1:n)*U(i+1:n,:blksz) AND
!                   COMPUTE c(j1:j2,i+1:i+blksz-1) = c(j1:j2,i+1:i+blksz-1) + x(j1:j2,:blksz)*U(i+1:i+blksz-1,:blksz)' .
!
                    do l = 1_i4b, blksz - 1_i4b
!
                        do l2 = l + 1_i4b, blksz
!
                            x(1_i4b:j3,l2) = x(1_i4b:j3,l2) - x(1_i4b:j3,l)*lmat(l2,l)
!
                            i1 = l2 - 1_i4b
!
                            c2(1_i4b:j3,i1) = c2(1_i4b:j3,i1) + x(1_i4b:j3,l)*umat(i1,l)
!
                        end do
!
                    end do
!
!                   COMPUTE c(j1:j2,i+blksz:n) = c(j1:j2,i+blksz:n) + x(j1:j2,:blksz)*U(i+blksz:n,:blksz)' .
!
                    do l = blksz, q
!
                        do l2 = 1_i4b, blksz
!
                            c2(1_i4b:j3,l) = c2(1_i4b:j3,l) + x(1_i4b:j3,l2)*umat(l,l2)
!                    
                        end do
!                    
                    end do
#endif
!
                    c(j1:j2,i+1_i4b:n) = c2(:j3,:q)
!
                end do
!
!$OMP END DO
!
            end do
!
!           APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
            i1 = i + 1_i4b
!
            do i = i1, k
!
                beta1 = beta(i)
!
                if ( beta1==zero ) cycle
!
!               APPLY H(i) TO c(j1:j2,i:n) FROM THE RIGHT.
!
                q = n - i + 1_i4b
!
                tmp(1_i4b:q) = mat(i:n,i)
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do j = 0_i4b, p2, blksz
!
                    j1 = j + 1_i4b
                    j2 = min( j + blksz, p )
                    j3 = j2 - j
!
                    c2(:j3,:q) = c(j1:j2,i:n)
!
#ifdef _BLAS
                    call gemv( 'N', j3, q, beta1, c2, blksz, tmp(1_i4b:q), 1_i4b, zero, x(:j3,1_i4b), 1_i4b )
!
                    call ger( j3, q, one, x(:j3,1_i4b), 1_i4b, tmp(1_i4b:q), 1_i4b, c2, blksz )
#else
                    x(:j3,1_i4b) = matmul( c2(:j3,:q), tmp(1_i4b:q) )

                    do l = 1_i4b, q
!
                        prod     = tmp(l)*beta1
                        c2(:j3,l) = c2(:j3,l) + prod*x(:j3,1_i4b)     
!
                    end do            
#endif
!
                    c(j1:j2,i:n) = c2(:j3,:q)
!
                end do
!
!$OMP END DO
!
            end do
!
        else
!
!           COMPUTE c(:m,:n)*H(k)* . . .*H(2)*H(1) .
!
            do i = k - blksz, 0_i4b, -blksz
!
!               APPLY H(i+blksz)* . . .*H(i+1) TO c(1:m,i+1:n) FROM THE RIGHT AS
!
!             c(1:m,i+1:n)*( I + U(i+1:n,:blksz)*inv( lmat(:blksz,:blksz) )*U(i+1:n,:blksz)' ) .
!
                q = n - i
!
!               COMPUTE THE blksz BY blksz UNIT UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz) .
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do l = 1_i4b, blksz
!
                    i1 = i + l
!
!                   SAVE SHIFTED HOUSEHOLDER VECTOR mat(i1:m,i1) .
!
                    umat(l:q,l) = mat(i1:n,i1)
!
                    do l2 = 1_i4b, l - 1_i4b
!
                        i2 = i + l2
!
                        lmat(l2,l) = -beta(i2)*dot_product( mat(i1:n,i1), mat(i1:n,i2) )
!
                    end do
!                    
                end do
!
!$OMP END DO
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do j = 0_i4b, p2, blksz
!
                    j1 = j + 1_i4b
                    j2 = min( j + blksz, p )
                    j3 = j2 - j
!
                    c2(:j3,:q) = c(j1:j2,i+1_i4b:n)
!
#ifdef _BLAS
!                   COMPUTE c(j1:j2,i+1:n)*U(i+1:n,:blksz) .
!
                    call gemm( 'N', 'N', j3, blksz, q, one, c2, blksz, umat, n, zero, x, blksz  )
!
                    do l = 1_i4b, blksz
!
                        l2 = i + l
!
                        prod = beta(l2)
!
                        x(1_i4b:j3,l) = prod*x(1_i4b:j3,l)
!                    
                    end do
!
!                   SOLVE x(j1:j2,:blksz)*lmat(:blksz,:blksz)' = c(j1:j2,i+1:n)*U(i+1:n,:blksz) .
!
                    call trsm( 'R', 'U', 'T', 'U', j3, blksz, one, lmat, blksz, x, blksz )
!
!                   COMPUTE c(j1:j2,i+1:n) = c(j1:j2,i+1:n) + x(j1:j2,:blksz)*U(i+1:n,:blksz)' .
!
                    call gemm( 'N', 'T', j3, q,  blksz, one, x, blksz, umat, n, one, c2, blksz )
#else
!                   COMPUTE c(j1:j2,i+1:n)*U(i+1:n,:blksz) .
!
                    do l = 1_i4b, blksz
!
                        i1 = i + l
!
                        x(1_i4b:j3,l) = umat(l,l)*c2(1_i4b:j3,l)
!
                        do l2 = l + 1_i4b, q
!
                            x(1_i4b:j3,l) = x(1_i4b:j3,l) +  umat(l2,l)*c2(1_i4b:j3,l2)
!                    
                        end do
!
                        x(1_i4b:j3,l) = beta(i1)*x(1_i4b:j3,l)
!                    
                    end do
!
!                   SOLVE x(j1:j2,:blksz)*lmat(:blksz,:blksz)' = c(j1:j2,i+1:n)*U(i+1:n,:blksz) AND
!                   COMPUTE c(j1:j2,i+1:i+blksz-1) = c(j1:j2,i+1:i+blksz-1) + x(j1:j2,:blksz)*U(i+1:i+blksz-1,:blksz)' .
!
                    do l = blksz, 1_i4b, -1_i4b
!
                        do l2 = 1_i4b, l-1_i4b
!
                            x(1_i4b:j3,l2) = x(1_i4b:j3,l2) - x(1_i4b:j3,l)*lmat(l2,l)
!
                        end do
!
                        do l2 = l, blksz-1_i4b
!
                            c2(1_i4b:j3,l2) = c2(1_i4b:j3,l2) + x(1_i4b:j3,l)*umat(l2,l)
!
                        end do
!
                    end do
!
!                   COMPUTE c(j1:j2,i+blksz:n) = c(j1:j2,i+blksz:n) + x(j1:j2,:blksz)*U(i+blksz:n,:blksz)' .
!
                    do l = blksz, q
!
                        do l2 = 1_i4b, blksz
!
                            c2(1_i4b:j3,l) = c2(1_i4b:j3,l) + x(1_i4b:j3,l2)*umat(l,l2)
!                    
                        end do
!                    
                    end do
#endif
!
                    c(j1:j2,i+1_i4b:n) = c2(:j3,:q)
!
                end do
!
!$OMP END DO
!
            end do
!
!           APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
            i1 = i + blksz
!
            do i = i1, 1_i4b, -1_i4b
!
                beta1 = beta(i)
!
                if ( beta1==zero ) cycle
!
!               APPLY H(i) TO c(j1:j2,i:n) FROM THE RIGHT.
!
                q = n - i + 1_i4b
!
                tmp(1_i4b:q) = mat(i:n,i)
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do j = 0_i4b, p2, blksz
!
                    j1 = j + 1_i4b
                    j2 = min( j + blksz, p )
                    j3 = j2 - j
!
                    c2(:j3,:q) = c(j1:j2,i:n)
!
#ifdef _BLAS
                    call gemv( 'N', j3, q, beta1, c2, blksz, tmp(1_i4b:q), 1_i4b, zero, x(:j3,1_i4b), 1_i4b )
!
                    call ger( j3, q, one, x(:j3,1_i4b), 1_i4b, tmp(1_i4b:q), 1_i4b, c2, blksz )
#else
                    x(:j3,1_i4b) = matmul( c2(:j3,:q), tmp(1_i4b:q) )

                    do l = 1_i4b, q
!
                        prod     = tmp(l)*beta1
                        c2(:j3,l) = c2(:j3,l) + prod*x(:j3,1_i4b)     
!
                    end do            
#endif
!
                    c(j1:j2,i:n) = c2(:j3,:q)
!
                end do
!
!$OMP END DO
!
            end do
!
        end if
!
!       DEALLOCATE (PRIVATE) WORK ARRAYS.
!
        deallocate( x, c2 )
!
    end if
!
!$OMP END PARALLEL
!
!   DEALLOCATE (SHARED) WORK ARRAY.
!
    deallocate( umat )
!
!
! END OF SUBROUTINE apply_q_qr_rm
! _______________________________
!
    end subroutine apply_q_qr_rm
!
! =========================================================================================
!
    subroutine apply_q_qr_rv( mat, beta, c, trans  )
!
! Purpose
! _______
!
!   APPLY_Q_QR overwrites the real m vector C with
!
!            Q * C   if TRANS = false, or
!
!            Q' * C  if TRANS = true
!
!   where Q is a real orthogonal matrix defined as the product of k
!   elementary reflectors
!
!                    Q = H(1) * H(2) * . . . * H(k)
!
!   as returned by QR_CMP or QR_CMP2. Q is of order m .
!
!
! Arguments
! _________
!
!   MAT          (INPUT) real(stnd), dimension(:,:)
!                On entry, the i-th column must contain the vector which
!                defines the elementary reflector H(i), for i = 1,2,...,k,
!                as returned by QR_CMP (or QR_CMP2) in the first k columns of its array
!                argument MAT. MAT is not modified by the routine.
!
!   BETA         (INPUT) real(stnd), dimension(:)
!                BETA(i) must contain the scalar factor of the elementary
!                reflector H(i), as returned by QR_CMP (or QR_CMP2).
!                The size of BETA determines the number k of elementary
!                reflectors whose product defines the matrix Q.
!
!                The size of BETA must verify: size( BETA ) <= min( size(MAT,1) , size(MAT,2) ) .
!
!   C            (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the m vector C.
!
!                On exit, C is overwritten by Q * C or Q' * C .
!
!                The shape of C must verify: size( C ) = size( MAT, 1 ) .
!
!   TRANS        (INPUT) logical(lgl)
!                If:
!
!                - TRANS = false : apply Q  (no transpose)
!                - TRANS = true  : apply Q' (transpose)
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the routine DORM2R in LAPACK.
!
!   For further details on the QR factorization and its use or the blocked
!   algorithm used here, see:
!
!   (1) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (2) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in),    dimension(:,:) :: mat
    real(stnd),   intent(in),    dimension(:)   :: beta
    real(stnd),   intent(inout), dimension(:)   :: c
!
    logical(lgl), intent(in) :: trans
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                        :: beta1, prod
    real(stnd), dimension( size(c) )  :: tmp
!
    integer(i4b) :: m, k, i
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='apply_q_qr'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( c )
    k = size( beta )
!
    call assert( logical( m==size(mat,1), lgl ),            &
                 logical( k<=min(m,size(mat,2)), lgl ),     &
                 name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b .or.  m<=0_i4b ) return
!
    if ( trans ) then
!
!       COMPUTE H(k)* . . .*H(2)*H(1)*c(:m) .
!
        do i = 1_i4b, k
!
            beta1 = beta(i)
!
            if ( beta1==zero ) cycle
!
!           APPLY H(i) TO c(i:m) FROM THE LEFT.
!
            tmp(i:m) = mat(i:m,i)
!
            prod   = dot_product( tmp(i:m), c(i:m) )*beta1
#ifdef _BLAS
            call axpy( m-i+1_i4b, prod, tmp(i:m), 1_i4b, c(i:m), 1_i4b )
#else
            c(i:m) = c(i:m) + prod*tmp(i:m)          
#endif
!
        end do
!
    else
!
!       COMPUTE H(1)*H(2)* . . .*H(k)*c(:m,:n) .
!
        do i = k, 1_i4b, -1_i4b
!
            beta1 = beta(i)
!
            if ( beta1==zero ) cycle
!
!           APPLY H(i) TO c(i:m) FROM THE LEFT.
!
            tmp(i:m) = mat(i:m,i)
!
            prod   = dot_product( tmp(i:m), c(i:m) )*beta1
#ifdef _BLAS
            call axpy( m-i+1_i4b, prod, tmp(i:m), 1_i4b, c(i:m), 1_i4b )
#else
            c(i:m) = c(i:m) + prod*tmp(i:m)             
#endif
!
        end do
!
    end if
!
!
! END OF SUBROUTINE apply_q_qr_rv
! _______________________________
!
    end subroutine apply_q_qr_rv
!
!
! =========================================================================================
!
!
! *****************************
! END OF MODULE QR_Procedures *
! *****************************
!    
end module QR_Procedures
