module String_Procedures 
! ####################################################################################
!
! Copyright 2020 IRD
!
! This file is part of statpack.
!
! statpack is free software: you can redistribute it and/or modify
! it under the terms of the GNU Lesser General Public License as
! published by the Free Software Foundation, either version 3 of 
! the License, or (at your option) any later version.
!
! statpack is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU Lesser General Public License for more details.
!
! You can find a copy of the GNU Lesser General Public License
! in the statpack/doc directory.
!
! ####################################################################################
!                                                                                    *
! ************************************************************************************
! MODULE EXPORTING ROUTINES AND PARAMETERS FOR STRING PROCESSING AND PRINTING.       *
!                                                                                    * 
! LATEST REVISION : 22/09/2020                                                       *
!                                                                                    *
! ####################################################################################
!                                                                                    *
! ************************************************************************************
!   
!
! USED MODULES 
! ============
!
    use Select_Parameters,  only : lgl, i1b, i4b, stnd
!
! OTHER MODULES USED IN SPECIFIC SUBROUTINES
! ==========================================
!
!   use Logical_Constants
!   use Char_Constants
!   use Reals_Constants
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!
! PUBLIC ENTITIES 
! ===============
!    
! ALL SUBROUTINES, FUNCTIONS, VARIABLES AND PARAMETERS ARE PRIVATE BY DEFAULT.
!
    private
    public ::  toupper, tolower, capitalize, kchr, kint, kfix, kexp,     &
               ascii_is_upper, is_upper, ascii_is_lower, is_lower,       &
               ascii_is_alpha, is_alpha, ascii_is_same, is_same,         &
               ascii_is_digit, is_digit, is_space, is_num,               &
               string_count, ascii_string_eq, string_eq,                 &
               ascii_string_index, string_index, ascii_string_comp,      &
               string_comp, ebc2asc, asc2ebc, ascii_to_upper, to_upper,  &
               ascii_to_lower, to_lower, ascii_case_change, case_change, &
               mid_shift, center, find_field, nbrchf, obt_fmt,           &
               val_to_string, string_to_val
!    
!    public
!    private :: lgl, i1b, i4b, stnd,                                 &
!               a_maj, a_min, i_maj, i_min, j_maj, j_min,            &
!               r_maj, r_min, s_maj, s_min, z_maj, z_min,            &
!               shift, ido, down_map_ascii, up_map_ascii
!
! GENERIC INTERFACES FOR ROUTINES WITH OVERLOADED VERSIONS
! ========================================================
!
    interface nbrchf
        module procedure  nbrchf_int, nbrchf_real
    end interface
!
    interface obt_fmt
        module procedure  fmt_int, fmt_real
    end interface
!
    interface val_to_string
        module procedure  int_to_string, real_to_string
    end interface
!
! MODULE VARIABLES 
! ================
!
    integer :: ido
!   
! MODULE PARAMETERS 
! =================
!
! CODES FOR CASE CONVERSIONS.
!
    integer(i1b), parameter ::     &
        toupper       = 1,         &
        tolower       = 2,         &
        capitalize    = 3
!
! CODES FOR NUMERICAL DATA TYPES STORED IN A STRING.
!
    integer(i1b), parameter  ::    &
        kchr = 0,                  & ! NON-NUMERICAL STRING
        kint = 1,                  & ! INTEGER
        kfix = 2,                  & ! FIXED REAL
        kexp = 3                     ! REAL WITH EXPONENT
!
! CODES FOR FOR UPPER OR LOWER CASE LETTERS. PRIVATE ENTITIES.
!
    character , parameter ::  a_min=achar(97),  i_min=achar(105),       &
                              j_min=achar(106), r_min=achar(114),       &
                              s_min=achar(115), z_min=achar(122),       &
                              a_maj=achar(65), i_maj=achar(73),         &
                              j_maj=achar(74), r_maj=achar(82),         &
                              s_maj=achar(83), z_maj=achar(90)
!
! CODE FOR  CONVERSION SHIFTS. THIS ASSUMES THAT THE RELATION
! BETWEEN A LOWER CASE LETTER AND ITS CORRESPONDING UPPER CASE
! LETTER IS THE SAME FOR EVERY LETTER. PRIVATE ENTITY.
!
    integer, parameter ::                   &
        shift = ichar( a_min ) - ichar( a_maj )   ! shift = 32 FOR ASCII OR - 64 FOR EBCDIC
!
! ASCII MAPS FOR  CONVERSION SHIFTS. PRIVATE ENTITIES.
!
    integer, save ::                                                                     &
    down_map_ascii(0:127)=(/ (ido, ido=0,64), (ido+32, ido=65,90), (ido, ido=91,127) /), &
    up_map_ascii(0:127)=(/ (ido, ido=0,96), (ido-32, ido=97,122), (ido, ido=123,127) /)
!
!
! =========================================================================================
!
!
                                  contains
!                                 ========
!
! =========================================================================================
!                          TEST SUBROUTINES AND FUNCTIONS
! =========================================================================================
!
    logical(lgl) function ascii_is_upper( c )
!
! Purpose
! _______
!
!   This function tests if the character  C  is an upper case letter.
!
!
! Arguments
! _________
!
!   C   (INPUT) character
!       The character to test
!
!
! Further Details
! _______________
!
!   It uses the ASCII collating sequence.
!
!
! ___________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character , intent(in) :: c
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    ascii_is_upper = lge( c, a_maj ) .and. lle( c, z_maj )
!
!
! END OF FUNCTION ascii_is_upper
! ______________________________
!
    end function ascii_is_upper
!
! =========================================================================================
!
    logical(lgl) function is_upper( c )
!
! Purpose
! _______
!
!   This function tests if the character  C  is an upper case letter.
!
!
! Arguments
! _________
!
!   C   (INPUT) character
!       The character to test
!
!
! Further Details
! _______________
!
!   It uses the underlying machine collating sequence.
!
!
! ___________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character, intent(in) :: c
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    select case( c )
!
        case( a_maj:i_maj , j_maj:r_maj , s_maj:z_maj )
!
            is_upper = true
!
        case default
!
            is_upper = false
!
    end select
!
!
! END OF FUNCTION is_upper
! ________________________
!
    end function is_upper
!
! =========================================================================================
!
    logical(lgl) function ascii_is_lower( c )
!
! Purpose
! _______
!
!   This function tests if the character  C  is a lower case letter.
!
!
! Arguments
! _________
!
!   C   (INPUT) character
!       The character to test
!
!
! Further Details
! _______________
!
!   It uses the ASCII collating sequence.
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character, intent(in) :: c
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    ascii_is_lower = lge( c , a_min ) .and. lle( c, z_min )
!
!
! END OF FUNCTION ascii_is_lower
! ______________________________
!
    end function ascii_is_lower
!
! =========================================================================================
!
    logical(lgl) function is_lower( c )
!
! Purpose
! _______
!
!   This function tests if the character  C  is a lower case letter.
!
!
! Arguments
! _________
!
!   C   (INPUT) character
!       The character to test
!
!
! Further Details
! _______________
!
!   It uses the underlying machine collating sequence.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character, intent(in) :: c
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    select case( c )
!
        case( a_min:i_min , j_min:r_min , s_min:z_min )
!
            is_lower = true
!
        case default
!
            is_lower = false
!
    end select
!
!   
! END OF FUNCTION is_lower
! ________________________
!
    end function is_lower
!
! =========================================================================================
!
    logical(lgl) function ascii_is_alpha( c )
!
! Purpose
! _______
!
!   This function tests if the character  C  is a letter.
!
!
! Arguments
! _________
!
!   C   (INPUT) character
!       The character to test
!
!
! Further Details
! _______________
!
!   It uses the ASCII collating sequence.
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character, intent(in) :: c
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    ascii_is_alpha = ascii_is_upper( c ) .or. ascii_is_lower( c )
!
!
! END OF FUNCTION ascii_is_alpha
! ______________________________
!
    end function ascii_is_alpha
!
! =========================================================================================
!
    logical(lgl) function is_alpha( c )
!
! Purpose
! _______
!
!   This function tests if the character  C  is a letter.
!
!
! Arguments
! _________
!
!   C   (INPUT) character
!       The character to test
!
!
! Further Details
! _______________
!
!   It uses the underlying machine collating sequence.
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character, intent(in) :: c
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    is_alpha = is_upper( c ) .or. is_lower( c )
!
!
! END OF FUNCTION is_alpha
! ________________________
!
    end function is_alpha
!
! =========================================================================================
!
    logical(lgl) function ascii_is_same( c1, c2 )
!
! Purpose
! _______
!
!   ascii_is_same tests if C1 is the same character as C2 regardless of case.
!
!
! Arguments
! _________
!
!   C1, C2  (INPUT) character
!           The characters to test
!
!
! Further Details
! _______________
!
!   It uses the ASCII collating sequence.
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character, intent(in) :: c1, c2
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST IF THE CHARACTERS ARE EQUAL.
!
    ascii_is_same = down_map_ascii(iachar(c1))==down_map_ascii(iachar(c2))
!
!
! END OF FUNCTION ascii_is_same
! _____________________________
!
    end function ascii_is_same
!
! =========================================================================================
!
    logical(lgl) function is_same( c1, c2 )
!
! Purpose
! _______
!
!   is_same tests if C1 is the same character as C2 regardless of case.
!
!
! Arguments
! _________
!
!   C1, C2  (INPUT) character
!           The characters to test
!
!
! Further Details
! _______________
!
!   It uses the underlying machine collating sequence.
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character, intent(in) :: c1, c2
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST IF THE CHARACTERS ARE EQUAL.
!
    is_same = c1==c2 
!
    if ( is_same ) return
!
!   NOW TEST FOR EQUIVALENCE IF BOTH CHARACTERS ARE ALPHABETIC.
!
    select case( c1 )
!
        case( a_min:i_min , j_min:r_min , s_min:z_min )
!
            if ( is_upper(c2) ) is_same = c1==char( ichar(c2) + shift )
!
        case( a_maj:i_maj , j_maj:r_maj , s_maj:z_maj )
!
            if ( is_lower(c2) ) is_same = c1==char( ichar(c2) - shift )
!
    end select
!
!
! END OF FUNCTION is_same
! _______________________
!
    end function is_same
!
! =========================================================================================
!
    logical(lgl) function ascii_is_digit( c )
!
! Purpose
! _______
!
!   This function tests if the character  C  is a digit.
!
!
! Arguments
! _________
!
!   C   (INPUT) character
!       The character to test
!
!
! Further Details
! _______________
!
!   It uses the ASCII collating sequence.
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character, intent(in) :: c
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    ascii_is_digit = lge( c , '0' ) .and. lle( c, '9' )
!
!
! END OF FUNCTION ascii_is_digit
! ______________________________
!
    end function ascii_is_digit
!
! =========================================================================================
!
    logical(lgl) function is_digit( c )
!
! Purpose
! _______
!
!   This function tests if the character  C  is a digit.
!
!
! Arguments
! _________
!
!   C   (INPUT) character
!       The character to test
!
!
! Further Details
! _______________
!
!   It uses the underlying machine collating sequence.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character, intent(in) :: c
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    select case( c )
!
        case( '0',  '1', '2', '3', '4', '5', '6', '7', '8', '9' )
!
            is_digit = true
!
        case default
!
            is_digit = false
!
    end select
!
!
! END OF FUNCTION is_digit
! ________________________
!
    end function is_digit
!
! =========================================================================================
!
    logical(lgl) function is_space( c )
!
! Purpose
! _______
!
!   This function tests if the character  C  is a space or a tabulation.
!
!
! Arguments
! _________
!
!   C   (INPUT) character
!       The character to test
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,   only : ht, blank
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character, intent(in) :: c
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    is_space = c==blank .or. c==ht
!
!
! END OF FUNCTION is_space
! ________________________
!
    end function is_space
!
! =========================================================================================
!
    integer(i1b) function is_num( string )
!
! Purpose
! _______
!
!   This function tests if the character argument STRING contains a numerical value.
!
!
! Arguments
! _________
!
!   STRING  (INPUT) character(len=*)
!           The string to analyze.
!
!
! Further Details
! _______________
!
!   Specifically, IS_NUM returns:
!
!   - KCHR = 0_i1b    if STRING is a non-numerical string
!   - KINT = 1_i1b    if STRING is an integer
!   - KFIX = 2_i1b    if STRING is a fixed real
!   - KEXP = 3_i1b    if STRING is a real with exponent
!       
!   Definitions of KCHR, KINT, KFIX and KEXP may be obtained from
!   the host module Strings.
!
!   This function is adapted from:
!
!   (1) Olagnon, M., 1996:
!            Traitement de donnees numeriques avec Fortran 90.
!            Masson, 264 pages, Chapter 5.3.2,  ISBN 2-225-85259-6.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,   only : blank, plus, minus, period
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character(len=*), intent(in) :: string
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i1b) :: num
    integer(i4b) :: ichr, slen, nmts, nexp, kmts, ifexp
!
    character(len=len_trim(string)) :: str
!
! PARAMETERS
! __________
!
    character , parameter ::  e_min=achar(101), d_min=achar(100),       &
                              e_maj=achar(69),  d_maj=achar(68)
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   INITIALIZATIONS.
!
    num = kchr
    nmts = 0_i4b
    nexp = 0_i4b
    kmts = 0_i4b
    ifexp = 0_i4b
!
!   MAKE A LEFT ADJUSTED COPY OF string.
!
    str = adjustl( string )
    slen = len_trim( str )
!
!   LOOP OVER THE CHARACTERS IN string WITHOUT LEADING AND TRAILING BLANKS.
!
    ichr = 0_i4b
!
    do
!
        if ( ichr>=slen ) then
!
!           LAST CONTROLS.
!
            if ( nmts==0_i4b .or. (num>=kexp .and. nexp==0_i4b) ) exit
!
            is_num = num
!
            return
!
        end if
!
        ichr = ichr + 1_i4b
!
        select case ( str(ichr:ichr) )
!
!           CASE BLANKS.
!
            case ( blank )
!
                continue
!
!           CASE DIGITS.
!
            case ( '0', '1', '2', '3', '4', '5', '6', '7', '8', '9' )
!
                select case ( num )
                    case ( kchr )
                        num = kint
                        nmts = nmts + 1_i4b
                    case ( kint, kfix )
                        nmts = nmts + 1_i4b
                    case ( kexp )
                        if ( nexp==0_i4b ) ifexp = 1_i4b
                        nexp = nexp + 1_i4b
                end select
!
!           CASE SIGN.
!
            case ( plus, minus )
!
                select case ( num )
                    case ( kchr )
                        if ( kmts>0_i4b ) exit
                        kmts = 1_i4b
                    case ( kint, kfix )
                        if ( nmts==0_i4b ) exit
                        num = kexp
                        ifexp = 1_i4b
                    case ( kexp )
                        if ( ifexp>0_i4b ) exit
                        ifexp = 1_i4b
                end select
!
!           CASE DECIMAL POINT.
!
            case ( period )
!
                if ( num>kint ) exit
                num = kfix
!
!           CASE EXPONENT.
!
            case ( e_min, e_maj, d_min, d_maj )
!
                if ( num>=kexp .or. nmts==0_i4b ) exit
                num = kexp
!
!           CASE NON-NUMERICAL.
!
            case default
!
                exit
!
        end select
!
    end do
!
!   IF WE GO HERE, string IS NON-NUMERICAL.
!
    is_num = kchr
!
!
! END OF FUNCTION is_num
! ______________________
!
    end function is_num
!
! =========================================================================================
!
    integer(i4b) function string_count( string, letter )
!
! Purpose
! _______
!
!   STRING_COUNT returns the number of occurences of the letter LETTER
!   in the string STRING.
!
!   Comparison is case-sensitive and trailing blanks are ignored.
!
!
! Arguments
! _________
!
!   STRING  (INPUT) character(len=*)
!           The string input.
!
!   LETTER  (INPUT) character(len=1)
!           The letter to compare against.
!
!
! Further Details
! _______________
!
!   The result is an integer of kind i4b.
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character(len=*), intent(in) :: string
    character(len=1), intent(in) :: letter
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: i
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    string_count = 0_i4b
!
    do i = 1_i4b , len_trim(string)
!
      if ( string(i:i)==letter ) string_count = string_count + 1_i4b
!
    end do
!
!
! END OF FUNCTION string_count
! ____________________________
!
    end function string_count
!
! =========================================================================================
!
    logical(lgl) function ascii_string_eq( string1, string2 )
!
! Purpose
! _______
!
!   ASCII_STRING_EQ tests if 2 strings are equal, ignoring case and trailing blanks.
!
!
! Arguments
! _________
!
!   STRING1, STRING2    (INPUT) character(len=*)
!                       The strings to test.
!
!
! Further Details
! _______________
!
!   It uses the ASCII collating sequence.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : false
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character(len=*), intent(in) :: string1, string2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: i, min_len
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    ascii_string_eq = false
!
!   COMPARE THE COMMON PORTION.
!
    min_len = min( len(string1), len(string2) )
!
    do i = 1_i4b , min_len
!
      if ( down_map_ascii(iachar(string1(i:i))) /= &
           down_map_ascii(iachar(string2(i:i))) ) return
!
    end do
!
!   THEN CHECK THE REMAINDERS, ONE OF WHICH IS EMPTY.
!
    ascii_string_eq = string1(min_len+1_i4b:)==string2(min_len+1_i4b:)
!
!
! END OF FUNCTION ascii_string_eq
! _______________________________
!
    end function ascii_string_eq
!
! =========================================================================================
!
    logical(lgl) function string_eq( string1, string2 )
!
! Purpose
! _______
!
!   STRING_EQ tests if 2 strings are equal, ignoring case and trailing blanks.
!
!
! Arguments
! _________
!
!   STRING1, STRING2    (INPUT) character(len=*)
!                       The strings to test.
!
!
! Further Details
! _______________
!
!   It uses the underlying machine collating sequence.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : false
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character(len=*), intent(in) :: string1, string2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: i, min_len
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    string_eq = false
!
!   COMPARE THE COMMON PORTION.
!
    min_len = min( len(string1), len(string2) )
!
    do i = 1_i4b , min_len
!
      if ( .not.is_same( string1(i:i) , string2(i:i) ) ) return
!
    end do
!
!   THEN CHECK THE REMAINDERS, ONE OF WHICH IS EMPTY.
!
    string_eq = string1(min_len+1_i4b:)==string2(min_len+1_i4b:)
!
!
! END OF FUNCTION string_eq
! _________________________
!
    end function string_eq
!
! =========================================================================================
!
    integer(i4b) function ascii_string_index( string, list )
!
! Purpose
! _______
!
!   ASCII_STRING_INDEX returns index of a string in a list of strings, or 0 if no match.
!   Comparison is case-insensitive and trailing blanks are ignored.
!
!
! Arguments
! _________
!
!   STRING  (INPUT) character(len=*)
!           The string input.
!
!   LIST    (INPUT) character(len=*), dimension(:)
!           The list to compare against.
!
!
! Further Details
! _______________
!
!   It uses the ASCII collating sequence.
!
!   The result is an integer of kind i4b.
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character(len=*), intent(in) :: string, list(:)
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: i, nlist
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    nlist = size( list )
!
    do i = 1_i4b , nlist
!
      if ( ascii_string_eq( string, list(i) ) ) exit
!
    end do
!
    ascii_string_index = merge( 0_i4b, i, i > nlist )
!
!
! END OF FUNCTION ascii_string_index
! __________________________________
!
    end function ascii_string_index
!
! =========================================================================================
!
    integer(i4b) function string_index( string, list )
!
! Purpose
! _______
!
!   STRING_INDEX returns index of a string in a list of strings, or 0 if no match.
!   Comparison is case-insensitive and trailing blanks are ignored.
!
!
! Arguments
! _________
!
!   STRING  (INPUT) character(len=*)
!           The string input.
!
!   LIST    (INPUT) character(len=*), dimension(:)
!           The list to compare against.
!
!
! Further Details
! _______________
!
!   It uses the underlying machine collating sequence.
!
!   The result is an integer of kind i4b.
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character(len=*), intent(in) :: string, list(:)
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: i, nlist
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    nlist = size( list )
!
    do i = 1_i4b , nlist
!
      if ( string_eq( string, list(i) ) ) exit
!
    end do
!
    string_index = merge( 0_i4b, i, i > nlist )
!
!
! END OF FUNCTION string_index
! ____________________________
!
    end function string_index
!
! =========================================================================================
!
    integer(i1b) function ascii_string_comp( string1, string2 )
!
! Purpose
! _______
!
!   ASCII_STRING_COMP compares 2 strings, ignoring case and trailing blanks.
!   It returns:
!
!   - 0  if strings are equal,
!   - -1 if STRING1 < STRING2,
!   - +1 if STRING1 > STRING2.
!
!
! Arguments
! _________
!
!   STRING1, STRING2    (INPUT) character(len=*)
!                       The strings to test.
!
!
! Further Details
! _______________
!
!   It uses the ASCII collating sequence.
!
!   The result is an integer of kind i1b.
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character(len=*), intent(in) :: string1, string2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: i, min_len
    integer(i1b) :: ichar1, ichar2
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   COMPARE THE COMMON PORTION.
!
    min_len = min( len(string1), len(string2) )
!
    do i = 1_i4b , min_len
!
        ichar1 = down_map_ascii(iachar(string1(i:i)))
        ichar2 = down_map_ascii(iachar(string2(i:i)))
!
        if ( ichar1==ichar2 ) cycle
!
        ascii_string_comp = merge( -1_i1b, 1_i1b, ichar1<ichar2 )
!
        return
!
    end do
!
!   THEN CHECK THE REMAINDERS, ONE OF WHICH IS EMPTY.
!
    if ( string1(min_len+1_i4b:)==string2(min_len+1_i4b:) ) then
        ascii_string_comp = 0_i1b
    else if ( len(string1)>min_len ) then
        ascii_string_comp = 1_i1b
    else
        ascii_string_comp = -1_i1b
    end if
!
!
! END OF FUNCTION ascii_string_comp
! _________________________________
!
    end function ascii_string_comp
!
! =========================================================================================
!
    integer(i1b) function string_comp( string1, string2 )
!
! Purpose
! _______
!
!   STRING_COMP compares 2 strings, ignoring case and trailing blanks.
!   It returns:
!
!   - 0  if strings are equal,
!   - -1 if STRING1 < STRING2,
!   - +1 if STRING1 > STRING2.
!
!
! Arguments
! _________
!
!   STRING1, STRING2    (INPUT) character(len=*)
!                       The strings to test.
!
!
! Further Details
! _______________
!
!   It uses the underlying machine collating sequence.
!
!   The result is an integer of kind i1b.
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character(len=*), intent(in) :: string1, string2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: i, min_len
!
    character    :: char1, char2
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   COMPARE THE COMMON PORTION.
!
    min_len = min( len(string1), len(string2) )
!
    do i = 1_i4b , min_len
!
        char1 = to_lower( string1(i:i) )
        char2 = to_lower( string2(i:i) )
!
        if ( char1==char2 ) cycle
!
        string_comp = merge( -1_i1b, 1_i1b, char1<char2 )
!
        return
!
    end do
!
!   THEN CHECK THE REMAINDERS, ONE OF WHICH IS EMPTY.
!
    if ( string1(min_len+1_i4b:)==string2(min_len+1_i4b:) ) then
        string_comp = 0_i1b
    else if ( len(string1)>min_len ) then
        string_comp = 1_i1b
    else
        string_comp = -1_i1b
    end if
!
!
! END OF FUNCTION string_comp
! ___________________________
!
    end function string_comp
!
! =========================================================================================
!                           CONVERSION SUBROUTINES AND FUNCTIONS
! =========================================================================================
!
    subroutine ebc2asc( ebc_str, asc_str, nchr )
!
! Purpose
! _______
!
!   EBC2ASC translates the EBCDIC string EBC_STR into a ASCII string ASC_STR.
!
!
! Arguments
! _________
!
!   EBC_STR     (INPUT) character(len=*)
!               The EBCDIC string to translate.
!
!   ASC_STR     (OUTPUT) character(len=*)
!               The translated string (ASCII).
!
!   NCHR        (INPUT) integer(i4b)
!               Number of characters to convert.
!
!
! Further Details
! _______________
!
!   EBC2ASC assumes that the storage unit for default characters is one byte and
!   that the storage unit for integers is a given number of bytes.
!
!   Non ASCII characters are translated as the "null" character (achar(0)).
!
!   This subroutine is adapted from:
!
!   (1) Olagnon, M., 1996:
!            Traitement de donnees numeriques avec Fortran 90.
!            Masson, 264 pages, Chapter 5.2,  ISBN 2-225-85259-6.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Char_Constants,    only : blank
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character(len=*), intent(in)  :: ebc_str
    character(len=*), intent(out) :: asc_str
!
    integer(i4b),     intent(in)  :: nchr
!
!
! PARAMETERS
! __________
!
    integer(i4b), parameter, dimension (0:255) :: itabt = (/ &
      0,  1,  2,  3,  0,  9,  0,127,  0,  0,  0, 11, 12, 13, 14, 15, &
     16, 17, 18, 19,  0,  0,  8,  0, 24, 25,  0,  0, 28, 29, 30, 31, &
      0,  0,  0,  0,  0, 10, 23, 27,  0,  0,  0,  0,  0,  5,  6,  7, &
      0,  0, 22,  0,  0,  0,  0,  4,  0,  0,  0,  0, 20, 21,  0, 26, &
     32,  0,  0,  0,  0,  0,  0,  0,  0,  0, 91, 46, 60, 40, 43, 33, &
     38,  0,  0,  0,  0,  0,  0,  0,  0,  0, 93, 36, 42, 41, 59, 94, &
     45, 47,  0,  0,  0,  0,  0,  0,  0,  0,124, 44, 37, 95, 62, 63, &
      0,  0,  0,  0,  0,  0,  0,  0,  0, 96, 58, 35, 64, 39, 61, 34, &
      0, 97, 98, 99,100,101,102,103,104,105,  0,  0,  0,  0,  0,  0, &
      0,106,107,108,109,110,111,112,113,114,  0,  0,  0,  0,  0,  0, &
      0,126,115,116,117,118,119,120,121,122,  0,  0,  0,  0,  0,  0, &
      0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, &
    123, 65, 66, 67, 68, 69, 70, 71, 72, 73,  0,  0,  0,  0,  0,  0, &
    125, 74, 75, 76, 77, 78, 79, 80, 81, 82,  0,  0,  0,  0,  0,  0, &
     92,  0, 83, 84, 85, 86, 87, 88, 89, 90,  0,  0,  0,  0,  0,  0, &
     48, 49, 50, 51, 52, 53, 54, 55, 56, 57,  0,  0,  0,  0,  0,  0 /)
!    
    integer(i4b), parameter :: lwrd=bit_size( itabt )/8_i4b
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    character(len=lwrd) :: zwrk
!
    integer(i4b)       :: ichr, nchre, nchr1, iwrk
    integer(i4b), save :: iput
!
    logical(lgl) :: first=true
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!$OMP CRITICAL
!
    if ( first ) then
!
        first = false
!
!       BIG-ENDIAN OR LITTLE-ENDIAN ?
!
        zwrk  = transfer( 0_i4b, zwrk )
        zwrk(lwrd:lwrd) = char( 1_i4b )
        iwrk = transfer( zwrk, iwrk )
        iput = merge( lwrd, 1_i4b, iwrk==1_i4b )
!
    end if
!
!$OMP END CRITICAL
!
!   CONVERT CHARACTERS.
!
    iwrk  = 0_i4b
    zwrk  = transfer( iwrk, zwrk )
    nchre = min( len( asc_str ), int(nchr) )
    nchr1 = min( len( ebc_str ), int(nchre) )
!
    do ichr = 1_i4b, nchr1
!
        zwrk(iput:iput) = ebc_str(ichr:ichr)
        iwrk = transfer( zwrk, iwrk )
        asc_str(ichr:ichr) = achar( itabt(iwrk) )
!
    end do
!
!  COMPLETE WITH BLANKS IF NEEDED (ASCII).
!
    do ichr = nchr1 + 1_i4b, nchre
!
        asc_str(ichr:ichr) = blank
!
    end do
!
!
! END OF SUBROUTINE ebc2asc
! _________________________
!
    end subroutine ebc2asc
!
! =========================================================================================
!
    subroutine asc2ebc( asc_str, ebc_str, nchr )
!
! Purpose
! _______
!
!   ASC2EBC translates the ASCII string ASC_STR into a EBCDIC string EBC_STR .
!
!
! Arguments
! _________
!
!   ASC_STR     (INPUT) character(len=*)
!               The ASCII string to translate.
!
!   EBC_STR     (OUTPUT) character(len=*)
!               The translated string (EBCDIC).
!
!   NCHR        (INPUT) integer(i4b)
!               Number of characters to convert.
!
!
! Further Details
! _______________
!
!   ASC2EBC assumes that the storage unit for default characters is one byte and
!   that the storage unit for integers is a given number of bytes.
!
!   For machines with extended ASCII characters set or machines returning values greater than
!   127 for intrinsic function IACHAR(), ASC2EBC assumes that characters "c" with IACHAR(c)>127
!   are the same as ACHAR(IACHAR(c)-128).
!
!   This subroutine is adapted from:
!
!   (1) Olagnon, M., 1996:
!            Traitement de donnees numeriques avec Fortran 90.
!            Masson, 264 pages, Chapter 5.2,  ISBN 2-225-85259-6.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character(len=*), intent(in)  :: asc_str
    character(len=*), intent(out) :: ebc_str
!
    integer(i4b),     intent(in)  :: nchr
!
!
! PARAMETERS
! __________
!
    integer(i4b), parameter, dimension (0:255) :: itabt = (/ &
      0,  1,  2,  3, 55, 45, 46, 47, 22,  5, 37, 11, 12, 13, 14, 15, &
     16, 17, 18, 19, 60, 61, 50, 38, 24, 25, 63, 39, 28, 29, 30, 31, &
     64, 79,127,123, 91,108, 80,125, 77, 93, 92, 78,107, 96, 75, 97, &
    240,241,242,243,244,245,246,247,248,249,122, 94, 76,126,110,111, &
    124,193,194,195,196,197,198,199,200,201,209,210,211,212,213,214, &
    215,216,217,226,227,228,229,230,231,232,233, 74,224, 90, 95,109, &
    121,129,130,131,132,133,134,135,136,137,145,146,147,148,149,150, &
    151,152,153,162,163,164,165,166,167,168,169,192,106,208,161,  7, &
      0,  1,  2,  3, 55, 45, 46, 47, 22,  5, 37, 11, 12, 13, 14, 15, &
     16, 17, 18, 19, 60, 61, 50, 38, 24, 25, 63, 39, 28, 29, 30, 31, &
     64, 79,127,123, 91,108, 80,125, 77, 93, 92, 78,107, 96, 75, 97, &
    240,241,242,243,244,245,246,247,248,249,122, 94, 76,126,110,111, &
    124,193,194,195,196,197,198,199,200,201,209,210,211,212,213,214, &
    215,216,217,226,227,228,229,230,231,232,233, 74,224, 90, 95,109, &
    121,129,130,131,132,133,134,135,136,137,145,146,147,148,149,150, &
    151,152,153,162,163,164,165,166,167,168,169,192,106,208,161,  7 /)
!    
    integer(i4b), parameter :: lwrd=bit_size( itabt )/8_i4b
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    character(len=lwrd) :: zwrk
!
    integer(i4b)       :: ichr, nchre, nchr1
    integer(i4b), save :: iget
!
    logical(lgl) :: first=true
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!$OMP CRITICAL
!
    if ( first ) then
!
        first = false
!
!       BIG-ENDIAN OR LITTLE-ENDIAN ?
!
        zwrk = transfer( 1_i4b, zwrk )
        ichr = ichar( zwrk(lwrd:lwrd) )
        iget = merge( lwrd, 1_i4b, ichr==1_i4b )
!
    end if
!
!$OMP END CRITICAL
!
!   CONVERT CHARACTERS.
!
    nchre = min( len( ebc_str ), int(nchr) )
    nchr1 = min( len( asc_str ), int(nchre) )
!
    do ichr = 1_i4b, nchr1
!
        zwrk = transfer( itabt( iachar( asc_str(ichr:ichr) ) ), zwrk )
        ebc_str(ichr:ichr) = zwrk(iget:iget)
!
    end do
!
!  COMPLETE WITH BLANKS IF NEEDED (EBCDIC).
!
    do ichr = nchr1 + 1_i4b, nchre
        ebc_str(ichr:ichr) = char( 64_i4b )
    end do
!
!
! END OF SUBROUTINE asc2ebc
! _________________________
!
    end subroutine asc2ebc
!
! =========================================================================================
!
    character function ascii_to_upper( c )
!
! Purpose
! _______
!
!   This function converts the character  C  to upper case.
!
!
! Arguments
! _________
!
!   C    (INPUT) character
!
!
! Further Details
! _______________
!
!   All non-alphabetic characters are left unchanged.
!   It uses the ASCII collating sequence.
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character, intent(in) :: c
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    ascii_to_upper = achar(up_map_ascii(iachar(c)))
!
! END OF FUNCTION ascii_to_upper
! ______________________________
!
    end function ascii_to_upper
!
! =========================================================================================
!
    character function to_upper( c )
!
! Purpose
! _______
!
!   This function converts the character  C  to upper case.
!
!
! Arguments
! _________
!
!   C    (INPUT) character
!
!
! Further Details
! _______________
!
!   All non-alphabetic characters are left unchanged.
!   It uses the underlying machine collating sequence.
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character, intent(in) :: c
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    select case( c )
!
        case( a_min:i_min , j_min:r_min , s_min:z_min )
!
            to_upper = char( ichar(c) - shift )
!
        case default
!
            to_upper = c
!
    end select
!
!
! END OF FUNCTION to_upper
! ________________________
!
    end function to_upper
!
! =========================================================================================
!
    character function ascii_to_lower( c )
!
! Purpose
! _______
!
!   This function converts the character  C  to lower case.
!
!
! Arguments
! _________
!
!   C    (INPUT) character
!
!
! Further Details
! _______________
!
!   All non-alphabetic characters are left unchanged.
!   It uses the ASCII collating sequence.
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character, intent(in) :: c
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    ascii_to_lower = achar(down_map_ascii(iachar(c)))
!
!
! END OF FUNCTION ascii_to_lower
! ______________________________
!
    end function ascii_to_lower
!
! =========================================================================================
!
    character function to_lower( c )
!
! Purpose
! _______
!
!   This function converts the character  C  to lower case.
!
!
! Arguments
! _________
!
!   C    (INPUT) character
!
!
! Further Details
! _______________
!
!   All non-alphabetic characters are left unchanged.
!   It uses the underlying machine collating sequence.
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character, intent(in) :: c
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    select case( c )
!
        case( a_maj:i_maj , j_maj:r_maj , s_maj:z_maj )
!
            to_lower = char( ichar(c) + shift )
!
        case default
!
            to_lower = c
!
    end select
!
!
! END OF FUNCTION to_lower
! ________________________
!
    end function to_lower
!
! =========================================================================================
!
    subroutine ascii_case_change( string, type )
!
! Purpose
! _______
!
!   This converts each lower case alphabetic letter in STRING to upper
!   case, or vice versa.
!
!
! Arguments
! _________
!
!   STRING  (INPUT/OUTPUT) character(len=*)
!           The string to convert
!
!   TYPE    (INPUT/OUTPUT) integer(i1b)
!           Define the conversion. Specifically, if:
!
!           - TYPE = 1_i1b = TOUPPER,    conversion is lower to upper
!           - TYPE = 2_i1b = TOLOWER,    conversion is upper to lower
!           - TYPE = 3_i1b = CAPITALIZE, use upper for first letter; lower for rest
!
!           Definitions of TOUPPER, TOLOWER and CAPITALIZE may be obtained from
!           the host module Strings.
!
!
! Further Details
! _______________
!
!   All non-alphabetic characters are left unchanged.
!   It uses the ASCII collating sequence.
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character(len=*), intent(inout) :: string
!
    integer(i1b),     intent(in)    :: type
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
   integer(i4b) :: i
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    if ( type==toupper ) then
!
        do i = 1_i4b, len( string )
            string(i:i) = achar(up_map_ascii(iachar(string(i:i))))
        end do
!
    else
!
        do i = 1_i4b, len( string )
            string(i:i) = achar(down_map_ascii(iachar(string(i:i))))
        end do
!
        if ( type==capitalize ) string(1:1) = achar(up_map_ascii(iachar(string(1:1))))
!
    end if
!
!   
! END OF SUBROUTINE ascii_case_change
! ___________________________________
!
    end subroutine ascii_case_change
!
! =========================================================================================
!
    subroutine case_change( string, type )
!
! Purpose
! _______
!
!   This converts each lower case alphabetic letter in STRING to upper
!   case, or vice versa.
!
!
! Arguments
! _________
!
!   string  (INPUT/OUTPUT) character(len=*)
!           The string to convert
!
!   TYPE    (INPUT/OUTPUT) integer(i1b)
!           Define the conversion. Specifically, if:
!
!           - TYPE = 1_i1b = TOUPPER,    conversion is lower to upper
!           - TYPE = 2_i1b = TOLOWER,    conversion is upper to lower
!           - TYPE = 3_i1b = CAPITALIZE, use upper for first letter; lower for rest
!
!           Definitions of TOUPPER, TOLOWER and CAPITALIZE may be obtained from
!           the host module Strings.
!
!
! Further Details
! _______________
!
!   All non-alphabetic characters are left unchanged.
!   It uses the underlying machine collating sequence.
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character(len=*), intent(inout) :: string
!
    integer(i1b),     intent(in)    :: type
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
   integer(i4b) :: i
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    if ( type==toupper ) then
!
        do i = 1_i4b, len( string )
!
            select case( string(i:i) )
                case( a_min:i_min , j_min:r_min , s_min:z_min )
                    string(i:i) = char( ichar(string(i:i)) - shift  )
            end select
!
        end do
!
    else
!
        do i = 1_i4b, len( string )
!
            select case( string(i:i) )
                case( a_maj:i_maj , j_maj:r_maj , s_maj:z_maj )
                    string(i:i) = char( ichar(string(i:i)) + shift  )
            end select
!
        end do
!
        if ( type==capitalize ) then
!
            select case( string(1:1) )
                case( a_min:i_min , j_min:r_min , s_min:z_min )
                    string(1:1) = char( ichar(string(1:1)) - shift  )
            end select
!
        end if
!
    end if
!
!   
! END OF SUBROUTINE case_change
! _____________________________
!
    end subroutine case_change
!
! =========================================================================================
!
    subroutine mid_shift( string, from, to, number )
!
! Purpose
! _______
!
!   This routine performs a shift of characters within STRING. The
!   number of characters shifted is NUMBER and they are shifted so
!   that the character in position FROM is moved to position TO.
!   Characters in the TO position are overwritten. Blanks replace
!   characters in the FROM position. Shifting may be left or right,
!   and the FROM and TO positions may overlap.  Care is taken not to
!   alter or use any characters beyond the defined limits of  STRING.
!
!
! Arguments
! _________
!
!   STRING              (INPUT/OUTPUT) character(len=*)
!                       The string to modify.
!
!   FROM, TO, NUMBER    (INPUT/OUTPUT) integer(i4b)
!                       Parameters defining the shift.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,   only : blank
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character(len=*), intent(inout) :: string
!
    integer(i4b),     intent(in)    :: from, to, number
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: end, end1, n, shorten, slen
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    slen = len( string )
!
    if ( from/=to ) then
!
!       end1 IS THE INITIAL POSITION OF THE LAST CHARACTER IN 
!       THE SUBSTRING SPECIFIED TO BE SHIFTED.
!
        end1 = from + number - 1_i4b
!
!       n IS THE NUMBER OF CHARACTERS THAT WILL BE SHIFTED AND THAT WILL
!       REMAIN WITHIN THE CONFINES OF STRING. n MAY HAVE TO BE REDUCED
!       FOR A SHIFT TO THE RIGHT TO ENSURE THAT CHARACTERS BEYOND THE
!       END OF THE STRING ARE NOT DEALT WITH (THIS IS NOT NECESSARY FOR LEFT
!       SHIFTS SINCE THE SUBSTRING MOVED TO THE LEFT CAN NEVER EXTEND
!       BEYOND THE END OF STRING.

        if ( end1>slen ) then
            shorten = end1 - slen
            n = number - shorten
        else
            n = number
        end if
!
        if ( from<to )  n = min( slen, to+n-1_i4b ) - to + 1_i4b
!
!       end IS THE LAST CHARACTER THAT WILL ACTUALLY BE SHIFTED (AND STILL
!       REMAIN WITHIN THE LIMITS OF string).
!
        end = from + n - 1_i4b
!
        string( to:min(slen,to+n-1_i4b) ) = string(from:end)
!
        if ( from<to ) then
!
!           SHIFT TO THE RIGHT.
!
            string( from:min(to-1_i4b,end1) ) = blank
!
        else
!
!           SHIFT TO THE LEFT.
!
            string( max(to+n,from):end ) = blank
!
        end if
!
     end if
!
!   
! END OF SUBROUTINE mid_shift
! ___________________________
!
    end subroutine mid_shift
!
! =========================================================================================
!
    subroutine center( string )
!
! Purpose
! _______
!
!   This routine shifts the nonblank characters of STRING so that
!   there is a balance of blanks on left and right.
!
!
! Arguments
! _________
!
!   STRING      (INPUT/OUTPUT) character(len=*)
!               The string to center.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,   only : blank
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character(len=*), intent(inout) :: string
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: start, endch, end, clen, slen, left
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   FIND LAST NON-BLANK CHARACTER.
!
    endch = len_trim( string )
!
    if ( endch/=0_i4b ) then
!
!   FIND FIRST NON-BLANK CHARACTER.
!
         start = verify( string, blank )
!
!   COMPUTE SHIFT AND DO IT.
!
        clen = endch - start + 1_i4b
        slen = len( string )
        left = 1_i4b + (slen - clen)/2_i4b
!
        if ( start > left ) then
!
!           MOVE LEFT.
!
            string(left:) = adjustl( string(left:) )
!
        else
!
!           MOVE RIGHT.
!
            end = min( slen , endch + left -start )
            string(start:end) = adjustr( string(start:end) )
!
        end if
!
    end if
!
!   
! END OF SUBROUTINE center
! ________________________
!
    end subroutine center
!
! =========================================================================================
!
!   
    subroutine find_field( string, istart, iend, delims, isearch )
!
! Purpose
! _______
!
!   This routine returns the starting and ending positions of a delimitted field in
!   STRING taking into account the set of delimitters stored in the string DELIMS.
!
!
! Arguments
! _________
!
!   STRING      (INPUT) character(len=*)
!               The string to analyze.
!
!   ISTART      (OUTPUT) integer(i4b)
!               The starting position of the field.
!
!   IEND        (OUTPUT) integer(i4b)
!               The ending position of the field.
!
!   DELIMS      (INPUT, OPTIONAL) character(len=*)
!               The string containing the characters to be accepted as delimitters.
!
!   ISEARCH     (INPUT, OPTIONAL) integer(i4b)
!               The starting position for searching for the field.
!
!
! Further Details
! _______________
!
!   If the optional argument DELIMS  is absent, the default delimitter set is a blank.
!   If the optional argument ISEARCH is absent, the starting position for searching for 
!   the field is the first character of STRING.
!
!   On return,  if:
!
!   - ISTART=0,  STRING is empty, i.e contains only delimitters or is the null string ;
!   - ISTART/=0, the delimitted field is STRING(ISTART:IEND).
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,   only : blank
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character(len=*), intent(in)           :: string
    character(len=*), intent(in), optional :: delims
!
    integer(i4b),     intent (out)         :: istart, iend
    integer(i4b),     intent(in), optional :: isearch
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    logical(lgl) :: test_delims
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST OPTIONAL ARGUMENTS.
!
    test_delims = present( delims )
!
    if ( present( isearch ) ) then
        istart = isearch
    else
        istart = 1_i4b
    end if
!
!   FIND LEFTMOST CHARACTER OF string(istart:) NOT IN THE SET OF DELIMITTERS.
!
    if ( test_delims ) then
        iend = verify( string(istart:), delims )
    else
        iend = verify( string(istart:), blank )
    end if
!
    if ( iend==0_i4b ) then
!
!       RETURN, string(istart:) IS EMPTY.
!
        istart = 0_i4b
!
    else
!
!       FIND FIRST POSITION OF FIELD.
!
        istart = istart + iend - 1_i4b
!
!       FIND THE LEFTMOST CHARACTER OF string(istart:) IN THE SET OF DELIMITTERS
!       AND ADJUST iend ACCORDINGLY.
!
        if ( test_delims ) then
            iend = scan( string(istart:), delims )
        else
            iend = scan( string(istart:), blank )
        end if
!
        if ( iend==0_i4b ) then
            iend = len( string )
        else
            iend = istart + iend - 2_i4b
        end if
!
    end if
!
!
! END OF SUBROUTINE find_field
! ____________________________
!
    end subroutine find_field
!
! =========================================================================================
!
    integer(i4b) function nbrchf_int( jval )
!
! Purpose
! _______
!
!   This function determines the number of characters (digits and sign)
!   needed to represent the integer JVAL.
!
!
! Arguments
! _________
!
!   JVAL    (INPUT) integer(i4b)
!           The integer to edit.
!
!
! Further Details
! _______________
!
!   The result is an integer of kind i4b.
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in) :: jval
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: jvala, nchf
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   THE RESULT IS THE INTEGER PART OF log10( abs(jval) ), PLUS ONE IF jval 
!   IS NEGATIVE, HOWEVER IT IS SAFER TO DETERMINE nbrchf_int BY DIVIDING.
!
    if ( jval<0_i4b ) then
        nchf = 2_i4b
        jvala = - jval
    else
        nchf = 1_i4b
        jvala = jval
    end if
!
!   PLUS ONE BY POWER OF TEN.
!
    do
!
        if ( jvala<10_i4b ) exit
!
        jvala = jvala/10_i4b
        nchf = nchf + 1_i4b
!
    end do
!
    nbrchf_int = nchf
!
!   
! END OF FUNCTION nbrchf_int
! __________________________
!
    end function nbrchf_int
!
! =========================================================================================
!
    integer(i4b) function nbrchf_real( rval )
!
! Purpose
! _______
!
!   This function determines the number of characters (digits and sign)
!   needed to represent the integer part of RVAL.
!
!
! Arguments
! _________
!
!   RVAL    (INPUT) real(stnd)
!           The real to edit.
!
!
! Further Details
! _______________
!
!   The result is an integer of kind i4b.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants, only : zero, ten
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in) :: rval
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: nchf
!
    real(stnd)   :: rvala
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   THE RESULT IS THE INTEGER PART OF log10( abs(rval) ), PLUS ONE IF rval
!   IS NEGATIVE, HOWEVER IT IS SAFER TO DETERMINE nbrchf_real BY DIVIDING.
!
    if ( rval<zero ) then
        nchf = 2_i4b
        rvala = -rval
    else
        nchf = 1_i4b
        rvala = rval
    end if
!
!   PLUS ONE BY POWER OF TEN.
!
    do
!
        if ( rvala<ten ) exit
!
        rvala = rvala/ten
        nchf = nchf + 1_i4b
!
    end do
!
    nbrchf_real = nchf
!
!   
! END OF FUNCTION nbrchf_real
! ___________________________
!
    end function nbrchf_real
!
! =========================================================================================
!
    character(len=12) function fmt_int( jval )
!
! Purpose
! _______
!
!   This function determines the "(Iw)" format needed to edit the integer JVAL
!   without excess blanks. The format is returned as a fixed length blank-padded 
!   string of 12 characters.
!
!
! Arguments
! _________
!
!   JVAL    (INPUT) integer(i4b)
!           The integer to edit.
!
!
! Further Details
! _______________
!
!   If there is an error, FMT_INT returns a "blank string".
!
!   The result is character string of length 12.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,   only : blank
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in) :: jval
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: nchf, nchrw
!
    character(len=13) :: form = "('(i',i?,')')"
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   DETERMINE NUMBER OF CHARACTERS NEEDED TO EDIT jval.
!
    nchf = nbrchf_int( jval )
!
!   DETERMINE NUMBER OF CHARACTERS NEEDED TO EDIT nchf.
!
    nchrw = nbrchf_int( nchf )
!   
    if ( nchrw>0_i4b .and. nchrw<10_i4b ) then
!
!       COMPLETE FORMAT TO WRITE THE FORMAT.
!
        write( form(8:8), "(i1)") nchrw
!
!       RETURNS THE FORMAT.
!
        write( fmt_int, form ) nchf
!
    else
!
!       ERROR, RETURNS BLANK STRING.
!
        fmt_int = blank
!
    end if
!   
!   
! END OF FUNCTION fmt_int
! _______________________
!
    end function fmt_int
!
! =========================================================================================
!
    character(len=22) function fmt_real( rval, d )
!
! Purpose
! _______
!
!   This function determines the "(Fw.d)" format needed to edit the real RVAL
!   without excess blanks. The format is returned as a fixed length blank-padded 
!   string of 22 characters.
!
!
! Arguments
! _________
!
!   RVAL    (INPUT) real(stnd)
!           The real to edit.
!
!   D       (INPUT, OPTIONAL) integer(i4b)
!           The desired number of decimal digits after the decimal point.
!
!
! Further Details
! _______________
!
!   If the optional parameter D is absent, the format returns by this function will
!   edit the real RVAL with five decimals digits after the decimal point.
!
!   If there is an error, FMT_REAL returns a "blank" string.
!
!   The result is character string of length 22.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,   only : blank
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in)           :: rval
!
    integer(i4b), intent(in), optional :: d
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: nd, nw, nchrd, nchrw
!
    character(len=20) :: form = "('(f',i?,'.',i?,')')"
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   FIRST SET APPROPRIATE DEFAULT VALUE FOR nd AND
!   DETERMINE NUMBER OF CHARACTERS NEEDED TO EDIT nd.
!
    if ( present( d ) ) then
        nd = d
        nchrd = nbrchf_int( nd )
    else
        nd = 5_i4b
        nchrd = 1_i4b
    end if
!
!   COMPUTE THE WIDTH, nw.
!
    nw = nbrchf_real( rval) + nd + 1_i4b
!
!   DETERMINE NUMBER OF CHARACTERS NEEDED TO EDIT nw.
!
    nchrw = nbrchf_int( nw )
!   
    if ( nchrw>0_i4b .and. nchrw<10_i4b ) then
!
!       COMPLETE FORMAT TO WRITE THE FORMAT.
!
        write( form(8:8),   "(i1)" )  nchrw
        write( form(15:15), "(i1)" )  nchrd
!
!       RETURNS THE FORMAT.
!
        write( fmt_real, form ) nw, nd
!
    else
!
!       ERROR, RETURNS BLANK STRING.
!
        fmt_real = blank
!
    end if
!
!   
! END OF FUNCTION fmt_real
! ________________________
!
    end function fmt_real
!
! =========================================================================================
!
    subroutine int_to_string( jval, string, nchar )
!
! Purpose
! _______
!
!   This function converts an integer to a string. The value is returned left adjusted
!   in the string. 
!
!
! Arguments
! _________
!
!   JVAL    (INPUT) integer(i4b)
!           The integer to convert.
!
!   STRING  (OUTPUT) character(len=*)
!           The string.
!
!   NCHAR   (OUTPUT) integer(i4b)
!           Number of characters used or needed to edit JVAL. The value is in STRING(1:NCHAR).
!
!
! Further Details
! _______________
!
!   If there is an error, INT_TO_STRING returns a string filled with '*' and 
!   the right length of string needed to edit JVAL in NCHAR.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Char_Constants,    only : blank, star
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b),     intent(in)   :: jval
    integer(i4b),     intent(out)  :: nchar
!
    character(len=*), intent(out)  :: string
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer :: iostat
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   WRITE THE STRING.
!
    iostat = 0
    string = repeat( star, len(string) )
!
    write( string, *, iostat=iostat ) jval
!
    if ( iostat/=0 .or. string(1:1)==star ) then
!
!       DETERMINE NUMBER OF CHARACTERS NEEDED TO EDIT jval.
!
        nchar = nbrchf_int( jval )
!
    else
!
!       ADJUST LEFT.
!
        string = adjustl( string )
!
!       DETERMINE NUMBER OF CHARACTERS USED TO EDIT jval.
!
        nchar = verify( string, blank, true )
!
    end if
!
!   
! END OF SUBROUTINE int_to_string
! _______________________________
!
    end subroutine int_to_string
!
! =========================================================================================
!
    subroutine real_to_string( rval, string, nchar, fmt, d )
!
! Purpose
! _______
!
!   This function converts the real RVAL to a string with a given "Fw.d" or "Gw.d" format.
!   The value is returned left adjusted in the string. 
!
!
! Arguments
! _________
!
!   RVAL    (INPUT) real(stnd)
!           The real to convert.
!
!   STRING  (OUTPUT) character(len=*)
!           The string.
!
!   NCHAR   (OUTPUT) integer(i4b)
!           Number of characters used or needed to edit RVAL. The value is in STRING(1:NCHAR).
!
!   FMT     (INPUT,OPTIONAL) character
!           "G" or "g" to use an G edit descriptor, an F edit descriptor is used
!           for other values of FMT.
!
!   D       (INPUT, OPTIONAL) integer(i4b)
!           Number of digits to appear after the decimal point in the output field
!           for an F edit descriptor or number of significant digits to print for
!           an G edit descriptor.
!
!
! Further Details
! _______________
!
!   If the optional parameter FMT is absent, an F edit descriptor is used by default.
!
!   If the optional parameter D is absent, five decimals digits will appear
!   after the decimal point in the string for an F edit descriptor or five
!   significant digits will be printed if an G edit descriptor is used.
!
!   If there is an error, REAL_TO_STRING returns a string filled with "*" and 
!   the right length of string needed to edit RVAL in NCHAR.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero
    use Char_Constants,    only : blank, star
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),       intent(in)           :: rval
!
    integer(i4b),     intent(in), optional :: d
    integer(i4b),     intent(out)          :: nchar
!
    character,        intent(in), optional :: fmt
    character(len=*), intent(out)          :: string
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: slen, nd, nchrw
    integer      :: iostat
!
    character(len=22) :: form = "(f?????????.?????????)"
!
    logical(lgl) :: test_f
!
!
! PARAMETERS
! __________
!
    character , parameter ::  f_maj=achar(70), f_min=achar(102)
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   FIRST SET APPROPRIATE VALUE FOR nd.
!
    if ( present( d ) ) then
        nd = d
    else
        nd = 5_i4b
    end if
!
!   DETERMINE AND SET EDIT DESCRIPTOR.
!
    if ( present( fmt ) ) then
        test_f = fmt==f_maj .or. fmt==f_min
    else
        test_f = true
    end if
!
    if ( test_f ) then
        nchrw= nd + merge( 1_i4b, 2_i4b, rval>=zero )
        write( form(2:2), "(a1)" ) 'f'
    else
        nchrw= nd + merge( 5_i4b, 6_i4b, rval>=zero )
        write( form(2:2), "(a1)" ) 'g'
    end if
!
    iostat = 0
    slen = len( string )
    string = repeat( star, slen )
!
    if ( nchrw<=slen ) then
!
!       DETERMINE NUMBER OF CHARACTERS NEEDED TO EDIT slen.
!
        nchrw = nbrchf_int( slen )
!   
        if ( nchrw>0_i4b .and. nchrw<10_i4b ) then
!
!           WRITE THE FORMAT.
!
            write( form(3:11),   "(i9)" )  slen
            write( form(13:21),  "(i9)" )  nd
!
!           WRITE THE STRING.
!
            write( string, form, iostat=iostat ) rval
!
        end if
!
    end if
!
!   DETERMINE nchar THE NUMBER OF CHARACTERS NEEDED TO EDIT rval.
!
    if ( string(1:1)==star .or. iostat/=0 ) then
!
        if ( test_f ) then
            nchar = nbrchf_real( rval ) + nd + 1_i4b
        else
            nchar = nd + 6_i4b
        end if
!
    else
!
        string = adjustl( string )
        nchar = verify( string, blank, true )
!
    end if
!
!   
! END OF SUBROUTINE real_to_string
! ________________________________
!
    end subroutine real_to_string
!
! =========================================================================================
!
    subroutine string_to_val( string, kcode, fmt )
!
! Purpose
! _______
!
!   This routine tests if the character argument STRING contains a numerical value
!   and returns a format to read the string.
!
!
! Arguments
! _________
!
!   STRING  (INPUT) character(len=*)
!           The string to analyze.
!
!   KCODE   (OUTPUT) integer(i1b)
!           KCODE is equal to:
!
!           - KCHR = 0_i1b    if STRING is a non-numerical string
!           - KINT = 1_i1b    if STRING is an integer
!           - KFIX = 2_i1b    if STRING is a fixed real
!           - KEXP = 3_i1b    if STRING is a real with exponent
!       
!           Definitions of KCHR, KINT, KFIX and KEXP may be obtained from
!           the host module Strings.
!
!   FMT     (OUTPUT) character(len=14)
!           The format to read the string.
!
!
! Further Details
! _______________
!
!   If there is an error, STRING_TO_VAL returns a format filled with blanks.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,   only : blank
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character(len=*),  intent(in)  :: string
    character(len=14), intent(out) :: fmt
!
    integer(i1b),      intent(out) :: kcode
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: slen, nchr
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    kcode = is_num( string )
!
    select case( kcode )
!
        case( kchr )
!
            slen = len( string )
            fmt(1:2) = '(a'
            fmt(12:) = ')'
!
        case( kint )
!
            slen = len_trim( string )
            fmt(1:2) = '(i'
            fmt(12:) = ')'
!
        case( kfix, kexp )
!
            slen = len_trim( string )
            fmt(1:2) = '(f'
            fmt(12:) = '.0)'
!
    end select
!
    nchr = nbrchf_int( slen )
!
    if ( nchr>0_i4b .and. nchr<10_i4b ) then
        write(fmt(3:11), "(i9)" ) slen
    else
        fmt = blank
    end if
!
!
! END OF SUBROUTINE string_to_val
! _______________________________
!
    end subroutine string_to_val
!
! =========================================================================================
!
!
! *********************************
! END OF MODULE String_Procedures *
! *********************************
!    
end module String_Procedures
