module Time_Procedures
! ####################################################################################
!
! Copyright 2020 IRD
!
! This file is part of statpack.
!
! statpack is free software: you can redistribute it and/or modify
! it under the terms of the GNU Lesser General Public License as
! published by the Free Software Foundation, either version 3 of 
! the License, or (at your option) any later version.
!
! statpack is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU Lesser General Public License for more details.
!
! You can find a copy of the GNU Lesser General Public License
! in the statpack/doc directory.
!
! ####################################################################################
!                                                                                    *
! ************************************************************************************
! MODULE EXPORTING TIME AND DATE UTILITIES.                                          *
!                                                                                    * 
! LATEST REVISION : 17/09/2020                                                       *
!                                                                                    *
! ####################################################################################
!                                                                                    *
! ************************************************************************************
!
!
! USED MODULES 
! ============
!
    use Select_Parameters,  only : lgl, i4b, i8b, extd
!
! OTHER MODULES USED IN SPECIFIC SUBROUTINES
! ==========================================
!
!   use Logical_Constants
!   use Reals_Constants
!   use Char_Constants
!   use Utilities
!   use String_Procedures
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!
! PUBLIC ENTITIES
! ===============
!    
    private
    public :: leapyr, daynum, day_of_week,                                         &
              daynum_to_ymd, ymd_to_daynum, ymd_to_dayweek, daynum_to_dayweek,     &
              rtsw, elapsed_time, cpusecs,                                         &
              time_to_hmsms, time_to_string,                                       &
              get_date, get_date_time,                                             &
              system_date_time, my_date_time,                                      &
              months, days
!   
! MODULE PARAMETERS 
! =================
!
    character(len=9), parameter :: months(12) = (/   'January  ', 'February ',  &
                                                     'March    ', 'April    ',  &
                                                     'May      ', 'June     ',  &
                                                     'July     ', 'August   ',  &
                                                     'September', 'October  ',  &
                                                     'November ', 'December '/),&
                                   days(1:7)  = (/   'Monday   ',               &
                                                     'Tuesday  ', 'Wednesday',  &
                                                     'Thursday ', 'Friday   ',  &
                                                     'Saturday ', 'Sunday   ' /)
!
!
! =========================================================================================
!
                                  contains
!                                 ========
!
! =========================================================================================
!
    logical(lgl) function leapyr( iyr )
!
! Purpose
! _______
!
!   Check for a leap year.  LEAPYR is returned as "true"
!   if IYR is a leap year, and "false" otherwise.
!
!
! Arguments
! _________
!
!   IYR     (INPUT) integer(i4b)
!           The year to test.
!
!
! Further Details
! _______________
!
!   This function uses the Gregorian calendar adopted the Oct. 15, 1582.
!
!   Leap years are years that are evenly divisible by 4, except years
!   that are evenly divisible by 100 must be divisible by 400.
!
!   The result is a logical of kind lgl.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in) :: iyr
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    if ( mod(iyr,400_i4b)==0_i4b ) then
!
!       YEARS DIVISIBLE BY 400 ARE ALSO DIVISIBLE BY 4 AND
!       100... THEY ARE LEAP YEARS.
!
        leapyr = true 
!
    else if ( mod(iyr,100_i4b)==0_i4b ) then
!
        leapyr = false 
!
    else if ( mod(iyr,4_i4b)==0_i4b ) then
!
!       LEAP YEAR.
!
        leapyr = true
!
    else
!
        leapyr = false
!
    end if
!
!
! END OF FUNCTION leapyr
! ______________________
!
    end function leapyr
!
! =========================================================================================
!
    integer(i4b) function daynum( iyr, imon, iday )
!
! Purpose
! _______
!
!   Compute a day number.  One of the more useful applications for this routine
!   is to compute the number of days between two dates.
!
!
! Arguments
! _________
!
!   IYR, IMON , IDAY    (INPUT) integer(i4b)
!                       Year (iyr), month (imon), and day (iday).
!
!
! Further Details
! _______________
!
!   This function uses the Gregorian calendar adopted the Oct. 15, 1582.
!
!   In other words, Oct. 15, 1582 will return a day number of unity and hence
!   this algorithm will not work properly for dates early than 10-15-1582.
!
!   The result is an integer of kind i4b.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : c365_25_extd
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in) :: iyr, imon, iday
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: nday
!
!
! PARAMETERS
! __________
!
! CUMULATIVE DAYS PER YEAR BEGINNING ON THE FIRST DAY OF THE MONTH.
!
    integer(i4b), parameter ::      &
        cdpy(0:11) = (/ 0,31,59,90,120,151,181,212,243,273,304,334 /)
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   THE INTEGER 577747 IS SIMPLY AN OFFSET SO THAT
!   OCT 15, 1582 RETURNS A DAY NUMBER OF ONE.
!
    nday = (iyr-1_i4b)*c365_25_extd + cdpy(imon-1_i4b) + iday - 577747_i4b
!
    if ( leapyr(iyr) ) then
!
!       ADD ONE DAY FOR THE CURRENT LEAP YEAR.
!
        if ( imon>2_i4b ) nday = nday + 1_i4b
!
    endif
!
    daynum = nday
!
!
! END OF FUNCTION daynum
! ______________________
!
    end function daynum
!
! =========================================================================================
!
    integer(i4b) function day_of_week( iyr, imon, iday )
!
! Purpose
! _______
!
!   This function returns the day of the week (e.g., Mon, Tue,...)
!   as an index (Mon=1 to Sun=7) for a given year, month, and
!   day.
!
!
! Arguments
! _________
!
!   IYR, IMON , IDAY    (INPUT) integer(i4b)
!                       Year (IYR), month (IMON), and day (IDAY).
!
!
! Further Details
! _______________
!
!   This routine assumes a valid day, month and year are input.
!
!   The result is an integer of kind i4b.
!
!   The algorithm is adapted from:
!
!   (1) Larson, K., 1995:
!          Computing the Day of the Week. Dr Dobb's
!          Journal,  p. 125-126, April.
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in) :: iyr, imon, iday
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: y, m
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    if ( imon==1_i4b .or. imon==2_i4b ) then
!
!       TREAT JAN. AND FEB. LIKE MONTH 13 AND 14 FROM THE
!       PREVIOUS YEAR.
!
        m = imon + 12_i4b
        y = iyr - 1_i4b
!
      else
!
        m = imon
        y = iyr
!
      endif
!
      day_of_week =     &
    mod((iday+2_i4b*m+3_i4b*(m+1_i4b)/5_i4b+y+y/4_i4b-y/100_i4b+y/400_i4b),7_i4b) + 1_i4b
!
!
! END OF FUNCTION day_of_week
! ___________________________
!
    end function day_of_week
!
! =========================================================================================
!
    subroutine daynum_to_ymd( jdaynum, iyr, imon, iday )
!
! Purpose
! _______
!
!   Converts a Julian Day Number (JDAYNUM) to Gregorian year (IYR), month (IMON)
!   and day (IDAY).
!
!
! Arguments
! _________
!
!   JDAYNUM             (INPUT) integer(i4b)
!                       The Julian day number to convert. See further details.
!
!   IYR, IMON , IDAY    (OUTPUT) integer(i4b)
!                       The year (IYR), month (IMON), and day (IDAY) in the 
!                       Gregorian calendar corresponding to the Julian day
!                       number JDAYNUM.
!
!
! Further Details
! _______________
!
!   This subroutine converts the integer JDAYNUM to three integers IYR, IMON and
!   IDAY standing for year, month, day in the Gregorian calendar promulgated by 
!   Gregory XIII, starting with JDAYNUM=1 on Friday, 15 October 1582.
!
!   To keep Pope Gregory's calendar synchronized with the seasons for the next 
!   16000 years or so, a small correction has been introduced; millennial years
!   divisible by 4000 are not considered leap-years.
!
!   Note the Gregorian calendar was adopted in Oct. 15, 1582, and hence
!   this algorithm will not work properly for dates early than 10-15-1582,
!   e.g. if JDAYNUM < 1.
!   
!   Note that England and its possessions remained 10 or 11 days behind this 
!   Gregorian calendar until 14 September 1752, e.g. if JDAYNUM < 62062.
!
!   Note that no consensus has been reached yet about dates beyond Monday, 
!   28 February, 4000 (e.g. if  JDAYNUM > 882928), and dates after 18000 A.D.
!   are extremely speculative at best.
!
!   This subroutine is adapted from a MATLAB M-file written by W. Kahan available
!   on the WEB.
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in)  :: jdaynum
    integer(i4b), intent(out) :: iyr, imon, iday
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: j, j2
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TRY JULIAN CALENDAR FOR 1501-1582.
!
    j = jdaynum + merge( 578031_i4b, 578041_i4b, jdaynum < 1_i4b )
!
!   iyr IS A VERNAL YEAR, USUALLY.
!
    iyr = (4000_i8b*j + 1608_i8b)/1460969_i8b
    j2 = (iyr/4_i4b) - (iyr/100_i4b) + (iyr/400_i4b) - (iyr/4000_i4b) + 365_i4b*iyr
!
!   iday IS A DAY OF VERNAL YEAR iyr IF > ZERO.
!
    iday = j - j2
!
    if ( iday<=0_i4b ) then
!
!       ADJUST iyr AND iday .
!
        iyr = iyr - 1_i4b
        j2 = (iyr/4_i4b) - (iyr/100_i4b) + (iyr/400_i4b) - (iyr/4000_i4b) + 365_i4b*iyr
        iday = j - j2
!
    end if
!
    imon = (10_i4b*iday - 5_i4b)/306_i4b
!
!   DAY iday OF VERNAL MONTH imon.
!
    iday = iday - (306_i4b*imon + 5_i4b)/10_i4b
!
!   CONVERT TO CALENDAR DAY AND MONTH.
!
    imon = imon + 3_i4b
!
    if ( imon>12_i4b ) then
!
        imon = imon - 12_i4b
        iyr = iyr + 1_i4b
!
    end if
!
!
! END OF SUBROUTINE daynum_to_ymd
! _______________________________
!
    end subroutine daynum_to_ymd
!
! =========================================================================================
!
    integer(i4b) function ymd_to_daynum( iyr, imon, iday )
!
! Purpose
! _______
!
!   Converts Gregorian year (IYR), month (IMON) and day (IDAY) to Julian day
!   Number. See further details.
!
!
! Arguments
! _________
!
!   IYR, IMON , IDAY        (INPUT) integer(i4b)
!                           The year (IYR), month (IMON), and day (IDAY) in the 
!                           Gregorian calendar to convert.
!
!
! Further Details
! _______________
!
!   This function converts the three integers IYR, IMON and IDAY standing for
!   year, month, day in the Gregorian calendar promulgated by Gregory XIII on
!   Friday, 15 October 1582, in the corresponding Julian day number starting
!   with YMD_TO_DAYNUM=1 on Friday, 15 October 1582.
!
!   Note the Gregorian calendar was adopted in Oct. 15, 1582, and hence
!   this algorithm will not work properly for dates early than 10-15-1582.
!   Dates are checked for validity by using DAYNUM_TO_YMD subroutine.
!   
!   The number of days between two dates is the difference between their Julian day.
!
!   The result is an integer of kind i4b.
!
!   This subroutine is adapted from a MATLAB M-file written by W. Kahan available
!   on the WEB.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,   only : time_error1
    use Utilities,        only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in)  :: iyr, imon, iday
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: d, m, y, l, jdaynum
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='ymd_to_daynum'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   COMPUTE VERNAL MONTH AND YEAR.
!
    m = imon - 3_i4b
!
    if ( m<0_i4b ) then
!
        m = m + 12_i4b
        y = iyr - 1_i4b
!
    else
!
        y = iyr
!
    end if
!
!   COMPUTE DAY OF VERNAL YEAR.
!
    d = (306_i4b*m + 5_i4b)/10_i4b
!
!   THE LAST TERM CORRECTS FOR LEAP-YEARS UP TO  18000 A.D.
!
    l = (y/4_i4b) - (y/100_i4b) + (y/400_i4b) - (y/4000_i4b)
!
!   JULIAN DAY NUMBER, UNLESS DATE IS INVALID.
!
    jdaynum = l + d + 365_i4b*y + iday - 578041_i4b
!
!   CHECK THE DATE.
!
    call daynum_to_ymd( jdaynum, y, m, d )
!
    if ( y/=iyr .or. m/=imon .or. d/=iday )  call merror( name_proc//time_error1 )
!
    ymd_to_daynum = jdaynum
!
!
! END OF FUNCTION ymd_to_daynum
! _____________________________
!
    end function ymd_to_daynum
!
! =========================================================================================
!
    integer(i4b) function ymd_to_dayweek( iyr, imon, iday )
!
! Purpose
! _______
!
!   Computes the day of the week from Gregorian year (IYR), month (IMON) and day (IDAY).
!   See further details.
!
!
! Arguments
! _________
!
!   IYR, IMON , IDAY        (INPUT) integer(i4b)
!                           The year (IYR), month (IMON), and day (IDAY) in the 
!                           Gregorian calendar to convert.
!
!
! Further Details
! _______________
!
!   This function returns the day of the week (e.g., Mon, Tue,...) as an integer 
!   index (Mon=1 to Sun=7) for the given year, month, and day in the Gregorian
!   calendar promulgated by Gregory XIII on Friday, 15 October 1582.
!
!   Note that the Gregorian calendar was adopted in Oct. 15, 1582, and hence
!   this algorithm will not work properly for dates early than 10-15-1582.
!   Dates are checked for validity by using DAYNUM_TO_YMD subroutine.
!
!   The result is an integer of kind i4b.
!   
!   This subroutine is adapted from a MATLAB M-file written by W. Kahan available
!   on the WEB.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,   only : time_error1
    use Utilities,        only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in)  :: iyr, imon, iday
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: d, m, y, l, jdaynum
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='ymd_to_dayweek'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   COMPUTE VERNAL MONTH AND YEAR.
!
    m = imon - 3_i4b
!
    if ( m<0_i4b ) then
!
        m = m + 12_i4b
        y = iyr - 1_i4b
!
    else
!
        y = iyr
!
    end if
!
!   COMPUTE DAY OF VERNAL YEAR.
!
    d = (306_i4b*m + 5_i4b)/10_i4b
!
!   THE LAST TERM CORRECTS FOR LEAP-YEARS UP TO  18000 A.D.
!
    l = (y/4_i4b) - (y/100_i4b) + (y/400_i4b) - (y/4000_i4b)
!
!   JULIAN DAY NUMBER, UNLESS DATE IS INVALID.
!
    jdaynum = l + d + 365_i4b*y + iday - 578041_i4b
!
!   CHECK THE DATE.
!
    call daynum_to_ymd( jdaynum, y, m, d )
!
    if ( y/=iyr .or. m/=imon .or. d/=iday ) call merror( name_proc//time_error1 )
!
!   COMPUTE DAY OF THE WEEK.
!
    ymd_to_dayweek = mod( jdaynum+578036_i4b, 7_i4b ) + 1_i4b
!
!
! END OF FUNCTION ymd_to_dayweek
! ______________________________
!
    end function ymd_to_dayweek
!
! =========================================================================================
!
    integer(i4b) function daynum_to_dayweek( jdaynum )
!
! Purpose
! _______
!
!   Computes the day of the week from Julian day number JDAYNUM.
!   See further details.
!
!
! Arguments
! _________
!
!   JDAYNUM     (INPUT) integer(i4b)
!               The Julian day number to convert.
!
!
! Further Details
! _______________
!
!   This function returns the day of the week (e.g., Mon, Tue,...) as an integer 
!   index (Mon=1 to Sun=7) for the given Julian day number JDAYNUM starting with
!   JDAYNUM=1 on Friday, 15 October 1582.
!
!   The result is an integer of kind i4b.
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in)  :: jdaynum
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   COMPUTE DAY OF THE WEEK.
!
    daynum_to_dayweek = mod( jdaynum+578036_i4b, 7_i4b ) + 1_i4b
!
!
! END OF FUNCTION daynum_to_dayweek
! _________________________________
!
    end function daynum_to_dayweek
!
! =========================================================================================
!
    real(extd) function rtsw()
!
! Purpose
! _______
!
!   RTSW is a Real-Time Stop Watch.
!
!   This routine can be used to compute the time lapse between
!   functions calls according to the system (wall) clock.
!
!
! Arguments
! _________
!
!   None.
!
!
! Further Details
! _______________
!
!   RTSW is a unique numeric identifier derived from 
!   the current system time and date.
!
!   This function works across month and year boundaries, but is not thread-safe
!   and cannot be called in parallel by different threads.
!
!   Since this routine uses the system clock, the elapsed time computed
!   with this routine may not (probably won't be in a multi-tasking OS)
!   an accurate reflection of the number of cpu cycles required to
!   perform a calculation. Therefore care should be exercised when
!   using this to profile a code.
!
!   The result is a real of kind extd.
!
!   The calling procedure for this function is as follow:
!
!       tim1 = rtsw()

!           [perform calculations]

!       tim2 = rtsw()

!       write(*,*) 'Elapsed Time (s): ',tim2-tim1
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : c1000_extd
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer            :: idate(8)
    integer(i4b)       :: ndays, nhrs, nmins, y, m, d
    integer(i4b), save :: ihr, imin, iday
!
    real(extd)   :: sec
!
    logical(lgl) :: first=true
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   RETRIEVE THE CURRENT SYSTEM DATE AND TIME FROM
!   INTRINSIC FUNCTION date_and_time.
!
    call date_and_time( values=idate(:) )
!
    y = idate(1)
    m = idate(2)
    d = idate(3)
!
    sec = real( idate(7), extd ) + real( idate(8), extd )/c1000_extd
!
    if ( first ) then
!
!       START THE WATCH.
!
        first = false
!
        iday = daynum( y, m, d )
        ihr = idate(5)
        imin = idate(6)
        rtsw = sec
!
    else
!
!       STOP THE WATCH.
!
        first = true
!
        ndays = daynum( y, m, d ) - iday
        nhrs = idate(5) - ihr
        nmins = idate(6) - imin
        rtsw = real( 60_i4b*nmins + 3600_i4b*nhrs + 86400_i4b*ndays, extd ) + sec
!
    end if
!
!
! END OF FUNCTION rtsw
! ____________________
!
    end function rtsw
!
! =========================================================================================
!
    integer(i4b) function elapsed_time( t1, t0 )
!
! Purpose
! _______
!
!   Computes elapsed time between two invocations of the intrinsic function
!   DATE_AND_TIME. ELAPSED_TIME( T1, T0 ) returns the time in seconds that has elapsed
!   between the vectors T0 and T1. Each vector must have at least seven elements
!   in the format returned by DATE_AND_TIME for the optional argument VALUES; namely
!
!           T = (/ year, month, day, x, hour, minute, second /)
!
!   This routine can be used to compute the elapsed time between
!   DATE_AND_TIME calls according to the system (wall) clock.
!
!
! Arguments
! _________
!
!   T1, T0      (INPUT) integer, dimension(:)
!               The two vectors which give the starting and ending
!               dates and times as returned by DATE_AND_TIME.
!
!
! Further Details
! _______________
!
!   Since this routine uses the system clock, the elapsed time computed
!   with this routine may not (probably won't be in a multi-tasking OS)
!   an accurate reflection of the number of cpu cycles required to
!   perform a calculation. Therefore care should be exercised when
!   using this to profile a code.
!
!   This function works across month and year boundaries but does not check
!   the validity of its arguments, which are expected to be obtained as in
!   the following example that shows how to time some operation by using ELAPSED_TIME.
!
!   The result is an integer of kind i4b.
!
!   The calling procedure for this subroutine is as follow:
!
!       call date_and_time( values=t0(:) )
!
!           [perform calculations]
!
!       call date_and_time( values=t1(:) )
!
!       write(*,*) 'Elapsed Time (s): ', elapsed_time( t1(:), t0(:) )
!
!   This subroutine is adapted from a MATLAB M-file written by W. Kahan available
!   on the WEB.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : time_error2
    use Utilities,       only : merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer, intent(in), dimension(:) :: t1, t0
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b), dimension(2) :: iyr, imon, iday, l(5)
!
    logical(lgl), dimension(2) :: test
!
!
! PARAMETERS
! __________
!
    integer(i4b), parameter ::      &
        ts(5) = (/ 86400_i4b, 0_i4b, 3600_i4b, 60_i4b, 1_i4b /)
!
    character(len=*),  parameter :: name_proc='elapsed_time'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   CHECK THE SIZE OF THE ARGUMENTS.
!
    if ( min( size(t1), size(t0) ) < 7 )        &
    call merror( name_proc//time_error2 )
!
    l(:) = t1(3:7) - t0(3:7)
!
!   O.K. FOR THE SAME MONTH.
!
    elapsed_time = dot_product( l(:), ts(:) )
!
    if ( any(t1(1:2)/=t0(1:2)) ) then
!
!       COMPUTE DAYS BETWEEN MONTHS.
!
        iyr(1) = t1(1)
        iyr(2) = t0(1)
        imon(1) = t1(2) - 3
        imon(2) = t0(2) - 3
!
!       COMPUTE VERNAL MONTH AND YEAR.
!
        test(:) = imon(:)<0_i4b
!
        where ( test(:) ) 
             imon(:) = imon(:) + 12_i4b
             iyr(:) = iyr(:) - 1_i4b
        end where
!
!       COMPUTE DAY OF VERNAL YEAR.
!
        iday(:) = (306_i4b*imon(:) + 5_i4b)/10_i4b
!
!       THE LAST TERM CORRECTS FOR LEAP-YEARS UP TO  18000 A.D.
!
        l(1:2) = (iyr(:)/4_i4b) - (iyr(:)/100_i4b) + (iyr(:)/400_i4b) - (iyr(:)/4000_i4b)
!
        iday(:) = l(1:2) + iday(:) + 365_i4b*iyr(:)
        elapsed_time = ( iday(1) - iday(2) )*86400_i4b + elapsed_time
!
    end if
!
!
! END OF FUNCTION elapsed_time
! ____________________________
!
    end function elapsed_time
!
! =========================================================================================
!
    real(extd) function cpusecs()
!
! Purpose
! _______
!
!   This function obtains, from the intrinsic routine SYSTEM_CLOCK,
!   the current value of the system CPU usage clock. This value
!   is then converted to seconds and returned as an extended precision
!   real value.
!
!
! Arguments
! _________
!
!   None.
!
!
! Further Details
! _______________
!
!   This functions assumes that the number of CPU cycles (clock counts) between
!   two calls is less than COUNT_MAX, the maximum possible value of clock counts
!   as returned by the intrinsic routine SYSTEM_CLOCK.
!
!   The result is a real of kind extd.
!
!   The calling procedure for this function is as follow:
!
!       tim1 = cpusecs()
!
!           [perform calculations]
!
!       tim2 = cpusecs()
!
!       write(*,*) 'CPU Time (s): ',tim2-tim1
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : zero_extd
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer      :: count, count_rate, count_max, lastcount = 0
!
    real(extd)   :: secs, nsecs=zero_extd, ct=zero_extd, ct_rate=zero_extd
!
    logical(lgl) :: first=true
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    if ( first ) then
!
        first = false
!
        call system_clock( count_rate=count_rate, count_max=count_max )
!
        ct_rate = real( count_rate, extd )
        ct = real( count_max, extd )/ct_rate
!
    end if
!
    call system_clock( count )
!
!   WRAPAROUND OF CLOCK TICKER.
!
    if ( count<lastcount ) nsecs = nsecs + ct
!
    lastcount = count
    secs  = real( count, extd )/ct_rate
    cpusecs  = secs + nsecs
!
!
! END OF FUNCTION cpusecs
! _______________________
!
    end function cpusecs
!
! =========================================================================================
!
    subroutine time_to_hmsms( time, hmsms )
!
! Purpose
! _______
!
!   Convert time to hours, minutes, seconds, milliseconds format.
!
!
! Arguments
! _________
!
!   TIME    (INPUT) real(extd)
!           The time in seconds.
!
!   HMSMS   (OUTPUT) integer(i4b), dimension(4)
!           On exit, an integer array with:
!
!           - HMSMS(1) = the hours as an integer number
!           - HMSMS(2) = the minutes as an integer number from 0 to 59
!           - HMSMS(3) = the seconds as an integer number from 0 to 59
!           - HMSMS(4) = the milliseconds as an integer number from 0 to 999
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Reals_Constants,   only : c1000_extd, c0_0004_extd
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(extd),   intent(in)  :: time
    integer(i4b), intent(out) :: hmsms(4)
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: i_time
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    i_time = int( time + c0_0004_extd, i4b )
    hmsms(1) = i_time/3600_i4b
    hmsms(2) = mod( i_time/60_i4b, 60_i4b )
    hmsms(3) = mod( i_time, 60_i4b )
    hmsms(4) = min( 999_i4b, nint( c1000_extd*(time - real( i_time, extd) ) , i4b ) )
!
!
! END OF SUBROUTINE time_to_hmsms
! _______________________________
!
    end subroutine time_to_hmsms
!
! =========================================================================================
!
    character(len=13) function time_to_string( time )
!
! Purpose
! _______
!
!   Convert TIME to a string format for printing as
!
!           'milliseconds.seconds.minutes.hours'
!
!
! Arguments
! _________
!
!   TIME    (INPUT) real(extd)
!           The time in seconds.
!
!
! Further Details
! _______________
!
!   The result is a string of (at least) 13 characters.
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(extd), intent(in)  :: time
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer          :: i, j
!
    integer(i4b)     :: hmsms(4), divide, i_digit
!
    character(len=3) :: hmsms_strings(4)
!
!
! PARAMETERS
! __________
!
    character, parameter :: digits(0:9) = &
         (/ '0','1','2','3','4','5','6','7','8','9'/)
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   RETRIEVE THE TIME AS AN INTEGER ARRAY hmsms.
!
    call time_to_hmsms( time, hmsms )
!
!   THE REST OF THIS ROUTINE IS A LONG-WINDED EQUIVALENT OF THE
!   FOLLOWING COMMENTED-OUT INTERNAL WRITE.  THE LONG WAY IS
!   USED SO THAT THIS FUNCTION CAN BE REFERENCED IN AN OUTPUT
!   IOLIST WITHOUT VIOLATING THE ANSI RESTRICTION ON THE USE OF
!   IO STATEMENTS IN SUCH FUNCTIONS.
!       write( time_to_string, '(i3.3,2(1h.,i2.2),1h.,i3.3)' ) hmsms
!
    do i = 1, 4
!
        divide = 1000_i4b
!
        do j = 1, 3
!
            i_digit = abs( mod( hmsms(i) ,divide ) )
            divide = divide/10_i4b
            hmsms_strings(i)(j:j) = digits( i_digit/divide )
!
        end do
!
    end do
!
    if (hmsms(1) > 999_i4b .or. hmsms(1) < 0_i4b)    hmsms_strings(1)='***'
!
    time_to_string = hmsms_strings(1) // '.' // hmsms_strings(2)(2:3)    &
        // '.' // hmsms_strings(3)(2:3) // '.' // hmsms_strings(4)
!
!
! END OF FUNCTION time_to_string
! ______________________________
!
    end function time_to_string
!
! =========================================================================================
!
    subroutine get_date( iyr, imon, iday, date )
!
! Purpose
! _______
!
!   Get a given date in a "nice" format.
!
!
! Arguments
! _________
!
!   IYR, IMON , IDAY    (INPUT) integer(i4b)
!                       Year (IYR), month (IMON), and day (IDAY).
!
!
!   DATE                (OUTPUT) character(len=*)
!                       The date in the form dd-mmm-yyyy as in 18-Mar-1992.
!
!                       Should be at least 11 chars long to hold the full string.
!
!
! Further Details
! _______________
!
!   If DATE is more than 11 characters in length, DATE is padded
!   with blanks. If it is less than 11 characters in length,
!   only the leftmost characters of the date will be returned.
!
!   If there is an error, GET_DATE returns a string filled with '*'.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,    only : star
    use String_Procedures, only : val_to_string
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    integer(i4b), intent(in)      :: iyr, imon, iday
!
    character(len=*), intent(out) :: date
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
!
    character(len=3)  :: cday, month
    character(len=10) :: cyr
!
    integer(i4b) :: ncyr, ncday
!
    logical(lgl) :: wrong_date
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    wrong_date = imon<=0_i4b .or. imon>12_i4b .or.     &
                 iday<=0_i4b .or. iday>31_i4b 
!
    if ( wrong_date ) then
!
!       iyr, imon, iday IS NOT A VALID DATE .
!
       date = repeat( star, len(date) )
!
    else
!
!       RETRIEVE THE DATE.
!
        call val_to_string( iyr, cyr, ncyr )
        call val_to_string( iday, cday, ncday )
        month = months(imon)(1:3)
!
        if ( ncday==1_i4b ) then
            date = '0' // cday(1:ncday) // '-' // month // '-' // cyr(1:ncyr)
        else
            date = cday(1:ncday) // '-' // month // '-' // cyr(1:ncyr)
        end if
!
    end if
!
!
!
! END OF SUBROUTINE get_date
! __________________________
!
    end subroutine get_date
!
! =========================================================================================
!
    subroutine get_date_time( date, time )
!
! Purpose
! _______
!
!   Get system date and time in "nice" formats. This routine just
!   reformats the output from the standard DATE_AND_TIME intrinsic.
!
!
! Arguments
! _________
!
!   DATE    (OUTPUT,OPTIONAL) character(len=*)
!           The date in the form dd-mmm-yyyy as in 18-Mar-1992.
!
!           Should be at least 11 chars long to hold full string.
!
!   TIME    (OUTPUT,OPTIONAL) character(len=*)
!           The time  in the form hh:mm:ss.
!
!           Should be at least 8 chars long to hold full string.
!
!
! Further Details
! _______________
!
!   If DATE is more than 11 characters in length, DATE is padded
!   with blanks. If it is less than 11 characters in length,
!   only the leftmost characters of the date will be returned.
!
!   If TIME is more than 8 characters in length, TIME is padded
!   with blanks. If it is less than 8 characters in length,
!   only the leftmost characters of the time will be returned.
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character(len=*), intent(out), optional :: date, time
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    character(len=8)  :: sys_date
    character(len=10) :: sys_time
    character(len=3)  :: month
!
    integer :: values(8)
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   RETRIEVE THE CURRENT SYSTEM DATE AND TIME FROM
!   INTRINSIC FUNCTION date_and_time.
!
    call date_and_time(sys_date, sys_time, values=values)
!
    if ( present(date) ) then
!
        month = '   '
!
        if ( values(2)>0 ) month = months(values(2))(1:3)
!
        date = sys_date(7:8) // '-' // month // '-' // sys_date(1:4)
!
    end if
!
    if ( present(time) )   &
        time = sys_time(1:2) // ':' // sys_time(3:4) // ':' // sys_time(5:6)
!
!
! END OF SUBROUTINE get_date_time
! _______________________________
!
    end subroutine get_date_time
!
! =========================================================================================
!
    subroutine system_date_time( chdate )
!
! Purpose
! _______
!
!     Retrieve the current system time and date and transfer
!     them to CHDATE in a "pretty" format, i.e.,
!
!       "DATE: DD-MMM-YYYY  TIME: HH:MM:SS"
!
!
! Arguments
! _________
!
!   CHDATE  (OUTPUT) character(len=*)
!           The string to hold the time and the date.
!
!           Should be at least 33 chars long to hold full string.
!
!
! Further Details
! _______________
!
!   If CHDATE is more than 33 characters in length, CHDATE is padded
!   with blanks. If it is less than 33 in length, only the leftmost
!   characters of the date will be returned.
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character(len=*), intent(out) :: chdate
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
   character(len=8)  :: time
   character(len=11) :: date
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    call get_date_time( date, time )
!
    chdate = "DATE: "//date//"  TIME: "//time
!
!
! END OF SUBROUTINE system_date_time
! __________________________________
!
    end subroutine system_date_time
!
! =========================================================================================
!
    subroutine my_date_time( chdate )
!
! Purpose
! _______
!
!   This routine returns in CHDATE a 41-character date of the form given
!   in model (below). It uses the time and date as obtained from the
!   intrinsic routine DATE_AND_TIME and converts them to the form of the
!   model given below:
!
!       '00:00 a.m., Wednesday, September 00, 1999'
!
!
! Arguments
! _________
!
!   CHDATE  (OUTPUT) character(len=*)
!           The string to hold the time and the date.
!           Should be at least 41 chars long to hold full string.
!
!
! Further Details
! _______________
!
!   Note that excess blanks in the date are eliminated.
!   If CHDATE is more than 41 characters in length, CHDATE
!   is padded with blanks.  If it is less than 41 in length,
!   only the leftmost characters of the date will be returned.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,    only : blank
    use String_Procedures, only : mid_shift
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    character(len=*), intent(out) :: chdate
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
   integer(i4b) :: dayno, kmon, to, k
!  
   character(len=10) :: time
   character(len=8)  :: date
   character(len=41) :: tdate
!
!
! PARAMETERS
! __________
!
   character(len=*), parameter  ::                         &
      model ='00:00 a.m., Wednesday, September 00, 1999'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    tdate  = model
!
    call date_and_time( date, time )
!
!   RETRIEVE YEAR, MONTH AND DAY NUMBER.
!
    read( date(1:4), '(i4)' )  k
    read( date(5:6), '(i2)' )  kmon
    read( date(7:8), '(i2)' )  dayno
!
!   WRITE YEAR, MONTH AND DAY NUMBER.
!
    tdate(38:41) = date(1:4)
!
    if ( date(7:7)=='0' ) date(7:7) = blank
!
    tdate(34:35) = date(7:8)
!
    tdate(24:32) = months(kmon)
    to = len_trim( months(kmon) )
!
    if ( to/=9_i4b ) call mid_shift( tdate, 33_i4b, to+24_i4b, 41_i4b )
!
!   DETERMINE DAY OF THE WEEK.
!
    dayno = day_of_week( k, kmon, dayno )
!
!   RETRIEVE HOURS.
!
    read( time(1:2), '(i2)' ) k
!
    if ( k>=13_i4b ) then
!
        k = k-12_i4b
        tdate(7:7) = 'p'
!
    else if ( k==12_i4b ) then
!
        tdate(7:7) = 'p'
!
    else if ( k==0_i4b ) then
!
        k = k + 12_i4b
        tdate(7:7) = 'a'
!
    else
!
        tdate(7:7) = 'a'
!
    end if
!
!   WRITE MINUTES, HOURS AND DAY OF THE WEEK.
!
    tdate(4:5) = time(3:4)
!
    write ( tdate(1:2), '(i2)' ) k
!
    tdate(13:21) = days(dayno)
    k = len_trim( days(dayno) )
!
!   SHIFT OVER BLANKS.
!
    if ( k/=9_i4b ) call mid_shift( tdate, 22_i4b, k+13_i4b, 41_i4b )
!
!   TRANSFER THE STRING.
!
    chdate = tdate
!
!
! END OF SUBROUTINE my_date_time
! ______________________________
!
    end subroutine my_date_time
!
! =========================================================================================
!
!
! *******************************
! END OF MODULE Time_Procedures *
! *******************************
!    
end module Time_Procedures
