program test2_comp_ginv
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine COMP_GINV in module SVD_Procedures.
!                                                                              
! LATEST REVISION : 19/11/2018
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    Random singular matrices are generated for
!    checking the accuracy of subroutine COMP_GINV.
!
!                                                                              
! ================================================================================================
!
!
! USED MODULES
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,  &
                         zero, half, one, comp_ginv, norm,        &
                         lamch
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def
!   
    real(stnd), parameter  :: fudge=one
!
    character(len=*), parameter :: name_proc='comp_ginv'
!   
!   
! SPECIFICATIONS FOR VARIABLES
! ============================
!
!
    real(stnd)                              :: err1, err2, err3, err4, err, eps,  &
                                               ulp, tol, safmin, tmp, norma,      &
                                               normainv
    real(stnd), allocatable, dimension(:,:) :: a, a2, res, ainv, ainv2,     &
                                               a_by_ainv, ainv_by_a
!
    integer      :: iok
    integer(i4b) :: i, j, k, m, n, nr, nc, krank0, krank, success, notdone
    logical(lgl) :: failure, details
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE TESTS.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    eps    = fudge*sqrt( ulp )
    safmin = lamch( 'S' )
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
        do k = 1_i4b, 10_i4b
!
            select case( k )
!
                case( 1_i4b, 6_i4b )
!
!                   TESTS 1 AND 6: COMPUTE GENERALIZED INVERSE OF a(:nr,:nc) WITH nr=nc .
!
                    nr = n
                    nc = n
!
                    krank0 = n - 1_i4b
!
                case( 2_i4b, 7_i4b )
!
!                   TESTS 2 AND 7 : COMPUTE GENERALIZED INVERSE OF a(:nr,:nc) WITH nr>=nc
!                                   AND nr NOT MUCH LARGER THAN nc .
!
                    nr = n + 2_i4b
                    nc = n
!
                    krank0 = merge( n - 1_i4b, n, k==2_i4b )
!
                case( 3_i4b, 8_i4b )
!
!                   TESTS 3 AND 8 : COMPUTE GENERALIZED INVERSE OF a(:nr,:nc) WITH nr>=nc
!                                   AND nr MUCH LARGER THAN nc .
!
                    nr = n + n
                    nc = n
!
                    krank0 = merge( n - 1_i4b, n, k==3_i4b )
!
                case( 4_i4b, 9_i4b )
!
!                   TESTS 4 AND 9 : COMPUTE GENERALIZED INVERSE OF a(:nr,:nc) WITH nc>=nr
!                                   AND nc NOT MUCH LARGER THAN nr .
!
                    nr = n
                    nc = n + 2_i4b
!
                    krank0 = merge( n, n - 1_i4b, k==4_i4b )
!
                case( 5_i4b, 10_i4b )
!
!                   TESTS 5 AND 10 : COMPUTE GENERALIZED INVERSE OF a(:nr,:nc) WITH nc>=nr
!                                    AND nc MUCH LARGER THAN nr .
!
                    nr = n
                    nc = n + n
!
                    krank0 = merge( n, n - 1_i4b, k==5_i4b )
!
            end select
!
!           ALLOCATE WORK ARRAYS.
!
            j = max( nr, nc)
!
            allocate( a(nr,nc), a2(nr,nc), res(j,j),           &
                      ainv(nc,nr), ainv2(nc,nr),               &
                      a_by_ainv(nr,nr), ainv_by_a(nc,nc),      &
                      stat=iok  )
!
            if ( iok/=0 ) then
!
                if ( allocated(a) )         deallocate( a )
                if ( allocated(a2) )        deallocate( a2 )
                if ( allocated(res) )       deallocate( res )
                if ( allocated(ainv) )      deallocate( ainv )
                if ( allocated(ainv2) )     deallocate( ainv2 )
                if ( allocated(a_by_ainv) ) deallocate( a_by_ainv )
                if ( allocated(ainv_by_a) ) deallocate( ainv_by_a )
!
                cycle
!
            end if
!
!           GENERATE A RANDOM COEFFICIENT MATRIX a(:nr,:nc) .
!
            call random_number( a(:nr,:nc) )
!
            a(:nr,:nc) = a(:nr,:nc) - half
!
            call random_number( tmp )
!
            if ( tmp>safmin ) then
                a(:nr,:nc) = a(:nr,:nc)/tmp
            end if
!
!           MAKE DATA MATRIX SINGULAR.
!
            if ( k<=5_i4b ) then
!            
                j = nc/2_i4b
!
                a(:nr,j) = a(:nr,1_i4b) + a(:nr,nc)
!
            else
!
                j = nr/2_i4b
!
                a(j,:nc) = a(1_i4b,:nc) + a(nr,:nc)
!
            end if
!
!           COMPUTE THE NORM OF THE MATRIX.
!
            norma = norm( a(:nr,:nc) )
!
            if ( norma==zero ) cycle
!
            tol = real( n, stnd )*eps
!
            do i = 1_i4b, 2_i4b
!
!               SAVE DATA MATRIX.
!
                a2(:nr,:nc) = a(:nr,:nc)
!
!               COMPUTE GENERALIZED INVERSE OF a(:m,:n).
!
                if ( i==1_i4b ) then
!
                    call comp_ginv( a2, failure, ainv, tol=tol, krank=krank )
!
                else
!
                    call comp_ginv( a2, failure, tol=tol, krank=krank )
!
                    ainv = transpose( a2 )
!
                end if
!
                if ( failure ) then
!
                    notdone = notdone + 1_i4b
!
                    cycle
!
                end if
!
                m = m + 1_i4b
!
!               COMPUTE THE NORM OF THE GENERALIZED INVERSE.
!
                normainv = norm( ainv )
!
!               COMPUTE ainv*a AND a*ainv*a .
!
                ainv_by_a = matmul( ainv, a )
                a2 = matmul( a, ainv_by_a )
!
!               COMPUTE a*ainv AND ainv*a*ainv .
!
                a_by_ainv = matmul( a, ainv )
                ainv2 = matmul( ainv, a_by_ainv )
!
!               CHECK THE Moore-Penrose EQUATIONS :
!
!                        a*ainv*a = a            (1)
!                     ainv*a*ainv = ainv         (2)
!                       (a*ainv)' = a*ainv       (3)
!                       (ainv*a)' = ainv*a       (4)
!
                res(:nr,:nc) = a - a2
                err1 = norm( res(:nr,:nc) )
!
                res(:nc,:nr)  = ainv - ainv2
                err2 = norm( res(:nc,:nr) )
!
                res(:nr,:nr) =  a_by_ainv - transpose( a_by_ainv )
                err3 = norm( res(:nr,:nr) )
!
                res(:nc,:nc) = ainv_by_a - transpose( ainv_by_a )
                err4 = norm( res(:nc,:nc) )
!
!               CHECK THE RESULTS FOR SMALL RESIDUALS.
!
                tmp = max( norma, normainv )
!
                err  = max( err1, err2, err3, err4 )/tmp
!
!                if ( err<=tol .and. krank==krank0 ) then
                if ( err<=tol ) then
!
                    success = success + 1_i4b
!
                    if ( details ) then
!
                        write (prtunit,*) 'Test ',m,' is correct'
!
                        write (prtunit,*) 'nr= ', nr,' , nc= ', nc,' , case= ', k,    &
                                          ' , krank0= ', krank0, ' , krank= ', krank, &
                                          ' , in_place= ', i==2_i4b
!
!                       PRINT ERRORS FOR THE Moore-Penrose EQUATIONS.
!
                        write (prtunit,*) 'Relative error for the first Moore-Penrose equation = ', &
                                          err1/tmp
!
                        write (prtunit,*) 'Relative error for the second Moore-Penrose equation = ', &
                                          err2/tmp
!
                        write (prtunit,*) 'Relative error for the third Moore-Penrose equation = ', &
                                          err3/tmp
!
                        write (prtunit,*) 'Relative error for the fourth Moore-Penrose equation = ', &
                                          err4/tmp
!
                        write (prtunit,*)
!
                    end if
!
                else
!
                    write (prtunit,*) 'Test ',m,' is incorrect'
!
!                    write (prtunit,*) '=> ',err,' > ', tol,' or ', krank0,' /= ', krank,  &
                    write (prtunit,*) '=> ',err,' > ', tol,                               &
                                      ' , nr= ', nr,' , nc= ', nc,' , case= ', k,         &
                                      ' , krank0= ', krank0, ' , krank= ', krank,         &
                                      ' , in_place= ', i==2_i4b
!
!                   PRINT ERRORS FOR THE Moore-Penrose EQUATIONS.
!
                    write (prtunit,*) 'Relative error for the first Moore-Penrose equation = ', &
                                      err1/tmp
!
                    write (prtunit,*) 'Relative error for the second Moore-Penrose equation = ', &
                                      err2/tmp
!
                    write (prtunit,*) 'Relative error for the third Moore-Penrose equation = ', &
                                      err3/tmp
!
                    write (prtunit,*) 'Relative error for the fourth Moore-Penrose equation = ', &
                                      err4/tmp
!
                    write (prtunit,*)
!
                end if
!
            end do
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( a, a2, ainv, ainv2, a_by_ainv, ainv_by_a, res )
!
        end do
!
    end do
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test2_comp_ginv
! ==============================
!
end program test2_comp_ginv
