program test2_gchol_cmp
!
!
! Purpose
! =======
!
!   This program is intended to test subroutines GCHOL_CMP and CHOL_SOLVE
!   in module Lin_Procedures.
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    Random positive symmetric matrices with given eigenvalue distributions are generated
!    for checking the accuracy of subroutines GCHOL_CMP and CHOL_SOLVE.
!
! LATEST REVISION : 01/06/2021
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,       &
                         true, zero, half, one, two, five, ten, c100,  &
                         minexp, maxexp, machhuge,  machbase,          &      
                         qr_cmp, ortho_gen_qr, gchol_cmp, chol_solve,  &
                         arth, lamch, norm, merror, allocate_error
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED
! =====================
!    
    implicit none
!
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def, nrhs=10
!   
    real(stnd), parameter  :: fudge=c100, max_eigval=1000._stnd
!
    real(stnd), parameter :: invmachbase=one/machbase
!
    character(len=*), parameter :: name_proc='gchol_cmp'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, eps2, ulp, safmin, tmp, shift,   &
                                               d1, dd1, conda, tol, adet, adet2
    real(stnd), allocatable, dimension(:)   :: eigval, diagr, beta, invdiag, normx, condax, &
                                               normb, condab
    real(stnd), allocatable, dimension(:,:) :: a, a2, eigvec, res, b, x, x2
!
    integer(i4b) :: i, j, k, l, n, m, success, krank0, krank, d2, dd2, notdone
    integer      :: iok
!
    logical(lgl) :: details, upper
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    eps    = fudge*ulp
!
    safmin = lamch( 'S' )
!
!   SPECIFY EVENTUALLY A SHIFT TO AVOID ZERO EIGENVALUE.
!
    shift = 0.05_stnd
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
    upper = true
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( normx(nrhs), condax(nrhs),  &
              normb(nrhs), condab(nrhs),  &
              stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    do n = n1, n2, n3
!
        upper = .not.upper
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( eigval(n), diagr(n), beta(n),    &
                  a(n,n), a2(n,n), res(n,nrhs),    &
                  eigvec(n,n), invdiag(n),         &
                  x(n,nrhs), x2(n,nrhs),           &
                  b(n,nrhs), stat=iok  )
!
        if ( iok/=0 ) then
!
            if ( allocated(eigval) )   deallocate( eigval )
            if ( allocated(diagr) )    deallocate( diagr )
            if ( allocated(beta) )     deallocate( beta )
            if ( allocated(a) )        deallocate( a )
            if ( allocated(a2) )       deallocate( a2 )
            if ( allocated(res) )      deallocate( res )
            if ( allocated(eigvec) )   deallocate( eigvec )
            if ( allocated(invdiag) )  deallocate( invdiag )
            if ( allocated(x) )        deallocate( x )
            if ( allocated(x2) )       deallocate( x2 )
            if ( allocated(b) )        deallocate( b )
!
            cycle
!
        end if
!
!       GENERATE A RANDOM COEFFICIENT MATRIX eigvec(:n,:n) .
!
        call random_number( eigvec(:n,:n) )
!
        eigvec(:n,:n) = eigvec(:n,:n) - half
!
        call random_number( tmp )
!
        if ( tmp>safmin ) then
            eigvec(:n,:n) = eigvec(:n,:n)/tmp
        end if
!
!       COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
        call qr_cmp( eigvec(:n,:n), diagr(:n), beta(:n) )
!
!       GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!       q IS ASSUMED OF FULL RANK.
!
        call ortho_gen_qr( eigvec(:n,:n), beta(:n) )
!
!       GENERATE A RANDOM SOLUTION MATRIX x .
!
        call random_number( x(:n,:nrhs) )
!
        call random_number( tmp )
!
        if ( tmp>safmin ) then
            x(:n,:nrhs) = x(:n,:nrhs)/tmp
        end if
!
!       COMPUTE THE NORMS OF THE COLUMNS OF RANDOM SOLUTION MATRIX x .
!
        normx(:nrhs) = norm( x(:n,:nrhs), dim=2_i4b )
!
!       COMPUTE FIRST THE EIGENVALUES OF THE MATRIX.
!
        do j = 1_i4b, 16_i4b
!
            select case( j )
!
            case( 1_i4b )
!
!               TEST 1: STRONGLY CLUSTERED EIGENVALUES.
!
                eigval(1_i4b)   = one
                eigval(2_i4b:n) = ulp
!
            case( 2_i4b )
!
!               TEST 2:  STRONGLY CLUSTERED EIGENVALUES.
!
                eigval(1_i4b:n-1_i4b) = one
                eigval(n)             = ulp
!
            case( 3_i4b )
!
!               TEST 3: WEAKLY CLUSTERED EIGENVALUES.
!
                eigval(1_i4b:n-1_i4b) = arth( ulp, ulp, n-1_i4b )
                eigval(n)             = one
!
            case( 4_i4b )
!
!               TEST 4: INTERNAL CLUSTER OF EIGENVALUES.
!
                eigval(1_i4b)         = ulp
                eigval(2_i4b:n-1_i4b) = one + sqrt(ulp)*arth( one, one, n-2_i4b )
                eigval(n)             = two
!
            case( 5_i4b )
!
!               TEST 5: WEAKLY CLUSTERED EIGENVALUES.
!
                eigval(1_i4b:n) = one + (c100*ulp)*arth( zero, one, n )
!
            case( 6_i4b )
!
!               TEST 6: GEOMETRIC DISTRIBUTION OF EIGENVALUES.
!
                err = real( n-1_i4b, stnd )
                do l = 1_i4b, n
                    eigval(l) = ulp**(real(l-1_i4b,stnd)/err)
                end do
!
            case( 7_i4b )
!
!               TEST 7: ARITHMETIC DISTRIBUTION OF EIGENVALUES.
!
                eigval(1_i4b:n) = ulp + ((one-ulp)/real( n-1_i4b, stnd ))*arth( zero, one, n )
!
            case( 8_i4b )
!
!               TEST 8: TWO CLUSTERS OF EIGENVALUES.
!
                i = n/2_i4b
!
                tmp = sqrt( ulp )
!
                eigval(1_i4b:i) = one + tmp*arth( one, one, i )
!
                eigval(i+1_i4b:n) = ten + tmp*arth( one, one, n-i )
!
            case( 9_i4b )
!
!               TEST 9: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR ZERO.
!
                i = n/2_i4b
!
                tmp = sqrt( ulp )
!
                eigval(1_i4b:i) = arth( one, one, i )
!
                eigval(i+1_i4b:n) = tmp*arth( one, one, n-i )
!
            case( 10_i4b )
!
!               TEST 10: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR n/2.
!
                i = n/2_i4b
!
                tmp = sqrt( ulp )
!
                eigval(1_i4b:i) = arth( one, one, i )
!
                eigval(i+1_i4b:n) = real( i, stnd) + tmp*arth( one, one, n-i )
!
            case( 11_i4b )
!
!               TEST 11: CLUSTERED DISTRIBUTION OF EIGENVALUES BETWEEN 0.9 and 1.
!
                call random_number( eigval(:n) )
!
                eigval(1_i4b:n) = 0.9_stnd + 0.1_stnd*eigval(1_i4b:n)
!
            case( 12_i4b )
!
!               TEST 12: EIGENVALUES EQUAL TO sinh(10*i/n) FOR i=1 TO n.
!
                tmp = ten/real( n, stnd)
!
                eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) )
!
            case( 13_i4b )
!
!               TEST 13: EIGENVALUES EQUAL TO sinh((10*i/n)-5) FOR i=1 TO n.
!
                tmp = ten/real( n, stnd)
!
                eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - five )
!
!               SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                where( eigval(1_i4b:n)<zero )
                    beta(:n) = -eigval(1_i4b:n)
                elsewhere
                    beta(:n) = zero
                end where
!
                tmp = maxval( beta(:n) ) + sqrt( ulp )
!
                eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
            case( 14_i4b )
!
!               TEST 14: EIGENVALUES EQUAL TO sinh((10*i/n)-10) FOR i=1 TO n.
!
                tmp = ten/real( n, stnd)
!
                eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - ten )
!
!               SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                where( eigval(1_i4b:n)<zero )
                    beta(:n) = -eigval(1_i4b:n)
                elsewhere
                    beta(:n) = zero
                end where
!
                tmp = maxval( beta(:n) ) + sqrt( ulp )
!
                eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
            case( 15_i4b )
!
!               TEST 15: EIGENVALUES EQUAL TO tanh((10*i/n)-5) FOR i=1 TO n.
!
                tmp = ten/real( n, stnd)
!
                eigval(1_i4b:n) = tanh( tmp*arth( one, one, n ) - five )
!
!               SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                where( eigval(1_i4b:n)<zero )
                    beta(:n) = -eigval(1_i4b:n)
                elsewhere
                    beta(:n) = zero
                end where
!
                tmp = maxval( beta(:n) ) + sqrt( ulp )
!
                eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
            case default
!
!               DEFAULT TEST : RANDOMLY UNIFORM DISTRIBUTION OF EIGENVALUES.
!
                call random_number( eigval(:n) )
!
                call random_number( tmp )
!
                if ( tmp>safmin ) then
                    eigval(1_i4b:n) = eigval(1_i4b:n)/tmp
                end if
!
            end select
!
#ifdef _F2003
            if ( ieee_support_datatype( eigval ) ) then
                if ( .not.all( ieee_is_normal( eigval(1_i4b:n) ) ) ) cycle
            end if
#endif
!
!           APPLY SHIFT TO THE EIGENVALUES.
!
            eigval(:n) = eigval(:n) + shift
!
            tol = real( n, stnd)*sqrt( ulp )
!
            where( eigval(:n)<=tol ) eigval(:n) = ulp
!
            krank0 = count( eigval(:n)>ulp )
!
!            where( eigval(:n)<=tol ) eigval(:n) = zero
!
!            krank0 = count( eigval(:n)/=zero )
!
!           SKIP ZERO MATRICES.
!
            if ( krank0==0_i4b ) cycle
!
!           FORM A MATRIX FROM THE SPECIFIED EIGENVALUE DISTRIBUTION AND RANDOM EIGENVECTORS.
!
            a2(:n,:n) = transpose( eigvec(:n,:n) )
!
            do i = 1_i4b, n
!
                a2(:n,i) = a2(:n,i)*eigval(:n)
!
            end do
!
            a(:n,:n) = matmul( eigvec(:n,:n), a2(:n,:n) )
!
!           SAVE THE COEFFICIENT MATRIX FOR LATER USE.
!
            a2(:n,:n) = a(:n,:n)
!
!           COMPUTE CONDITION NUMBER OF RANDOM MATRIX.
!
            tmp = maxval( eigval(:n) )
!
            if ( krank0==n ) then
!
                conda = tmp/minval( eigval(:n) )
!
!                conda = ( sum( eigval(:n) )*sum( one/eigval(:n) ) )/real( n, stnd)
!
                tmp = min( tmp, max_eigval )
!
                condax(:nrhs) = tmp*normx(:nrhs)
!
            else
!
                beta(:n) = eigval(:n)
                where( beta(:n)==zero ) beta(:n) = tmp
!
                conda = tmp/minval( beta(:n) )             
!
            end if
!
!           COMPUTE THE MATRIX-MATRIX PRODUCT b = a*x .
!
            b(:n,:nrhs) = matmul( a(:n,:n), x(:n,:nrhs) )
!
!           COMPUTE THE NORMS OF THE COLUMNS OF THE RIGHT HAND SIDE MATRIX b .
!
            normb(:nrhs) = norm( b(:n,:nrhs), dim=2_i4b )
!
            tmp = min( conda, max_eigval )
!
            condab(:nrhs) = ( one + two*tmp )*normb(:nrhs)
!
!           COMPUTE CHOLESKY FACTORIZATION OF a .
!
            tol = real( n, stnd)*ulp
!
            call gchol_cmp( a(:n,:n), invdiag(:n), krank, d1, d2=d2, upper=upper, tol=tol )
!
            if ( d1<zero ) then
!
                notdone = notdone + 1_i4b
!
                cycle
!
            end if
!
            do i = 1_i4b, 3_i4b
!
                select case( i )
!
                case( 1_i4b )
!
!                   TEST 1 : SEVERAL RIGHT HAND-SIDES.
!
!                   MAKE A COPY OF RIGHT HAND-SIDE MATRIX b .
!
                    x2(:n,:nrhs) = b(:n,:nrhs)
!
!                   COMPUTE THE SOLUTION MATRIX FOR LINEAR SYSTEM
!
!                               a*x = b .
!
                    call chol_solve( a(:n,:n), invdiag(:n), x2(:n,:nrhs), upper=upper )
!
!                   CHECK THE RESULTS FOR THE SOLUTION OR RESIDUAL MATRIX.
!
                    if ( krank0==n .and. krank==n ) then
!                    if ( krank0==n ) then
!
                        x2(:n,:nrhs) = x2(:n,:nrhs) - x(:n,:nrhs)
!
                        err = maxval( norm( x2(:n,:nrhs), dim=2_i4b )/condax(:nrhs) )
!
                    else
!
                        res(:n,:nrhs) = b(:n,:nrhs) - matmul( a2(:n,:n), x2(:n,:nrhs) )
!
                        err = maxval( norm( res(:n,:nrhs), dim=2_i4b )/condab(:nrhs) )
!
                    end if
!
                    k = nrhs
!
                case( 2_i4b )
!
!                   TEST 2 : ONE RIGHT HAND-SIDE.
!
!                   MAKE A COPY OF RIGHT HAND-SIDE VECTOR b(:,1) .
!
                    x2(:n,1_i4b) = b(:n,1_i4b)
!
!                   COMPUTE THE SOLUTION VECTOR FOR LINEAR SYSTEM
!
!                               a*x(:,1)=b(:,1) .
!
                    call chol_solve( a(:n,:n), invdiag(:n), x2(:n,1_i4b), upper=upper )
!
!                   CHECK THE RESULTS FOR THE SOLUTION OR RESIDUAL MATRIX.
!
                    if ( krank0==n .and. krank==n ) then
!                    if ( krank0==n ) then
!
                        x2(:n,1_i4b) = x2(:n,1_i4b) - x(:n,1_i4b)
!
                        err = norm( x2(:n,1_i4b) )/condax(1_i4b)
!
                    else
!
                        res(:n,1_i4b) = b(:n,1_i4b) - matmul( a2(:n,:n), x2(:n,1_i4b) )
!
                        err = norm( res(:n,1_i4b) )/condab(1_i4b)
!
                    end if
!
                    k = 1_i4b
!
                case( 3_i4b )
!
!                   TEST 3 : CHECK THE DETERMINANT OF a.
!
!                   COMPUTE DETERMINANT OF a .
!
                    select case( d2 )
!
                        case ( :minexp-1_i4b )
!
                            adet = sign( zero, d1 )
!
                        case ( maxexp+1_i4b: )
!
                            adet = sign( machhuge, d1 )
!
                        case default
!
!                           adet = scale( d1, d2 )
                            adet = d1*( machbase**d2 )
!
                    end select
!
!                   RECOMPUTE THE DETERMINANT OF a FROM THE EIGENVALUES.
!
                    if ( krank0==n ) then
!
                        dd1 = one
                        dd2 = zero
!
                        do l=1_i4b, n
!
                            dd1 = dd1*eigval(l)
!
                            do
                                if ( abs(dd1)<= one ) exit
                                dd1 = dd1*invmachbase
                                dd2 = dd2 + 1_i4b
                            end do
!
                            do
                                if ( abs(dd1)>= invmachbase ) exit
                                dd1 = dd1*machbase
                                dd2 = dd2 - 1_i4b
                            end do
!
                        end do
!
                        select case( dd2 )
!
                            case ( :minexp-1_i4b )
!
                                adet2 = sign( zero, dd1 )
!
                            case ( maxexp+1_i4b: )
!
                                adet2 = sign( machhuge, dd1 )
!
                            case default
!
!                               adet2 = scale( dd1, dd2 )
                                adet2 = dd1*( machbase**dd2 )
!
                        end select
!
                    else
!
                        adet2 = zero
!
                    end if
!
!                   NOW CHECK THE RESULTS.
!
#ifdef _F2003
                    if ( ieee_support_datatype( adet2 ) ) then
!
                        if ( ieee_is_normal( adet2 ) .and. ieee_is_normal( adet ) ) then
!
                            if ( min( abs(adet), abs(adet2) )<=safmin ) then
                                tmp = one
                            else
                                tmp = max( abs(adet), abs(adet2) )
                            end if
!
                            err = abs(adet - adet2) / tmp
!
                        else
!
                            err = zero
!
                        end if
!
                    else
!
#else
                        if ( abs( adet2 )<=machhuge .and. abs( adet )<=machhuge  ) then
!
                            if ( min( abs(adet), abs(adet2) )<=safmin ) then
                                tmp = one
                            else
                                tmp = max( abs(adet), abs(adet2) )
                            end if
!
                            err = abs(adet - adet2) / tmp
!
                        else
!
                            err = zero
!
                        end if
!
#endif
!
#ifdef _F2003
                    end if
#endif
!
                    k = 0_i4b
!
                end select
!
                m = m + 1_i4b
!
                if ( i<=2_i4b ) then
                    eps2 = real( n, stnd)*eps
                else
                    eps2 = real( n, stnd)*fudge*eps
                end if
!
                if ( err<=eps2 ) then
!
                    success = success + 1_i4b
!
                    if ( details ) then
!
                        write (prtunit,*) 'Test ',m,' is correct'
!
                        if ( i<=2_i4b ) then
!
                            write (prtunit,*) 'n= ', n,' , case= ', j,' , nrhs= ', k,          &
                                              ' , krank0= ', krank0, ' , krank= ', krank,      &
                                              ' , d1= ', d1,' , d2= ', d2,' , upper= ', upper
!
!                           PRINT CONDITION NUMBER OF COEFFICIENT MATRIX.
!
                            write (prtunit,*) 'Condition number of the coefficient matrix         = ', conda
!
!                           PRINT SCALED ERRORS OF SOLUTIONS OR RESIDUALS.
!
                            if ( krank0==n .and. krank==n ) then
!
                                write (prtunit,*) 'Maximum scaled error of computed solution vectors  = ', err
!
                            else
!
                                write (prtunit,*) 'Maximum scaled error of computed residual vectors  = ', err
!
                            end if
!
                        else
!
                            write (prtunit,*) 'n= ', n,' , case= ', j,                        &
                                              ' , krank0= ', krank0, ' , krank= ', krank,     &
                                              ' , d1= ', d1,' , d2= ', d2,' , upper= ', upper
!
!                           PRINT CONDITION NUMBER OF COEFFICIENT MATRIX.
!
                            write (prtunit,*) 'Condition number of the coefficient matrix = ', conda
!
!                           PRINT ABSOLUTE AND RELATIVE ERRORS OF DETERMINANT.
!
                            write (prtunit,*) 'Absolute error of the computed determinant = ', abs(adet - adet2)
                            write (prtunit,*) 'Relative error of the computed determinant = ', err
!
                        end if
!
                        write (prtunit,*)
!
                    end if
!
                else
!
                    write (prtunit,*) 'Test ',m,' is incorrect'
!
                    if ( i<=2_i4b ) then
!
                        write (prtunit,*) '=> ',err,' > ', eps2,                                &
                                          ' , n= ', n,' , case= ', j, ' , nrhs= ', k,           &
                                          ' , krank0= ', krank0, ' , krank= ', krank,           &
                                          ' , nrhs= ', k,' , d1= ', d1,' , d2= ', d2,           &
                                          ' , upper= ', upper
!
!                       PRINT CONDITION NUMBER OF COEFFICIENT MATRIX.
!
                        write (prtunit,*) 'Condition number of the coefficient matrix         = ', conda
!
!                       PRINT SCALED ERRORS OF SOLUTIONS OR RESIDUALS.
!
                        if ( krank0==n .and. krank==n ) then
!
                            write (prtunit,*) 'Maximum scaled error of computed solution vectors  = ', err
!
                        else
!
                            write (prtunit,*) 'Maximum scaled error of computed residual vectors  = ', err
!
                        end if
!
                    else
!
                        write (prtunit,*) '=> ',err,' > ', eps2,                                &
                                          ' , n= ', n,' , case= ', j,                           &
                                          ' , krank0= ', krank0, ' , krank= ', krank,           &
                                          ' , nrhs= ', k,' , d1= ', d1,' , d2= ', d2,           &
                                          ' , upper= ', upper
!
!                       PRINT CONDITION NUMBER OF COEFFICIENT MATRIX.
!
                        write (prtunit,*) 'Condition number of the coefficient matrix = ', conda
!
!                       PRINT ABSOLUTE AND RELATIVE ERRORS OF DETERMINANT.
!
                        write (prtunit,*) 'Absolute error of the computed determinant = ', abs(adet - adet2)
                        write (prtunit,*) 'Relative error of the computed determinant = ', err
!
                    end if
!
                    write (prtunit,*)
!
                end if
!
            end do
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( eigval, diagr, beta, a, a2, eigvec, &
                    res, b, x, x2, invdiag  )
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( normx, condax, normb, condab )
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test2_gchol_cmp
! ==============================
!
end program test2_gchol_cmp
