program test2_pinvt
!
!
! Purpose
! =======
!
!   This program is intended to test functions PINVT and PINVSTUDENT in module Prob_Procedures.
!                                                                              
!                                                                              
! LATEST REVISION : 23/01/2018
!
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, n1_def, n2_def, n3_def,    &
                         zero, half, one, two, allocate_error, &
                         merror, pinvt, pinvstudent
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def,   &
                               ndf_min=20, ndf_max=200, ndf_inc=7
!
    real(stnd), parameter :: eps = 1.e-4_stnd 
!
    character(len=*), parameter :: name_proc='pinvt'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                                :: err, df, ulp, q
    real(stnd), allocatable, dimension(:,:)   :: p, t2, diff
    real(stnd), allocatable, dimension(:,:,:) :: t, t3
!
    integer(i4b) :: i, j, k, l, m, n, ndf, ndfn, success
    integer      :: iok
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
    ulp = epsilon( ulp )
!
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        ndfn = ( ( ndf_max - ndf_min + 1_i4b )/ndf_inc ) + 1_i4b
!
        allocate( p(n,n), t2(n,n), diff(n,n), t(n,n,ndfn),  &
                  t3(n,n,ndfn), stat=iok  )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       GENERATE A RANDOM PROBABILITY MATRIX p .
!
        call random_number( p(:n,:n) )
!
        where( p(:n,:n)<ulp .or. p(:n,:n)>one-ulp )  p(:n,:n) = half
!
        l = 0_i4b
!
        do ndf = ndf_min, ndf_max, ndf_inc
!
            df = real( ndf, stnd )
!
            l = l + 1_i4b
!
            do j = 1_i4b, n
!
                do k = 1_i4b, n
!
!                   COMPUTE THE QUANTILE t(k,j,ndf) OF STUDENT T-DISTRIBUTION WITH ndf DEGREES
!                   OF FREEDOM CORRESPONDING TO LOWER TAIL AREA OF p(k,j) WITH pinvstudent
!                   FUNCTION.
!
                    if ( p(k,j)<=half ) then
!
                        q = two*p(k,j)
                        t(k,j,l) =  -pinvstudent( q, df )
!
                    else
!
                        q = two*( one - p(k,j) )
                        t(k,j,l) =  pinvstudent( q, df )
!
                    end if
!
                end do
!
            end do
!
        end do
!
        where ( t(:n,:n,:ndfn)==zero ) 
            t3(:n,:n,:ndfn) = one
        elsewhere
            t3(:n,:n,:ndfn) = t(:n,:n,:ndfn)
        end where
!
!       RECOMPUTE THE QUANTILES WITH pinvt FUNCTION.
!
        do i = 1_i4b, 3_i4b
!
            m = m + 1_i4b
!
            err = zero
!
            select case( i )
!
            case( 1_i4b )
!
!               TEST 1 : MATRIX FORM OF pinvt GENERIC FUNCTION.
!
                l = 0_i4b
!
                do ndf = ndf_min, ndf_max, ndf_inc
!
                    l = l + 1_i4b
!
!                   COMPUTE THE QUANTILES t2(:n,:n) OF STUDENT T-DISTRIBUTION WITH ndf DEGREES OF FREEDOM
!                   CORRESPONDING TO LOWER TAIL AREAS OF p(:n,:n) .
!
                    t2(:n,:n) = pinvt( p(:n,:n), ndf ) 
!
!                   CHECK THAT t2(:n,:n) AND t(:n,:n,ndf) AGREE.
!
                    diff(:n,:n) = abs( (t2(:n,:n) - t(:n,:n,l))/t3(:n,:n,l) )
!
                    err = max( err, maxval( diff(:n,:n) ) )
!
                end do
!
            case( 2_i4b )
!
!               TEST 2 : VECTOR FORM OF pinvt GENERIC FUNCTION.
!
                l = 0_i4b
!
                do ndf = ndf_min, ndf_max, ndf_inc
!
                    l = l + 1_i4b
!
                    do j = 1_i4b, n
!
!                       COMPUTE THE QUANTILES t2(:n,j) OF STUDENT T-DISTRIBUTION WITH ndf DEGREES
!                       OF FREEDOM CORRESPONDING TO LOWER TAIL AREAS OF p(:n,j) .
!
                        t2(:n,j) = pinvt( p(:n,j), ndf )
!
                    end do
!
!                   CHECK THAT t2(:n,:n) AND t(:n,:n,ndf) AGREE.
!
                    diff(:n,:n) = abs( (t2(:n,:n) - t(:n,:n,l))/t3(:n,:n,l) )
!
                    err = max( err, maxval( diff(:n,:n) ) )
!
                end do
!
            case( 3_i4b )
!
!               TEST 3 : SCALAR FORM OF pinvt GENERIC FUNCTION.
!
                l = 0_i4b
!
                do ndf = ndf_min, ndf_max, ndf_inc
!
                    l = l + 1_i4b
!
                    do j = 1_i4b, n
!
                        do k = 1_i4b, n
!
!                           COMPUTE THE QUANTILE t2(k,j) OF STUDENT T-DISTRIBUTION WITH ndf DEGREES
!                           OF FREEDOM CORRESPONDING TO LOWER TAIL AREA OF p(k,j) .
!
                            t2(k,j) = pinvt( p(k,j), ndf )
!
                        end do
!
                    end do
!
!                   CHECK THAT t2(:n,:n) AND t(:n,:n,ndf) AGREE.
!
                    diff(:n,:n) = abs( (t2(:n,:n) - t(:n,:n,l))/t3(:n,:n,l) )
!
                    err = max( err, maxval( diff(:n,:n) ) )
!
                end do
!
            end select
!
!           PRINT THE RESULTS OF THE TESTS.
!
            if ( err<=eps ) then
!
                success = success + 1_i4b
!
            else
!
                write (prtunit,*) 'Test ',m,' is incorrect'
!
                write (prtunit,*) '=> ',err,' > ',eps,' , n= ', n,' , case= ', i
!
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( p, t, t2, t3, diff )
!
    end do
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test2_pinvt
! ==========================
!
end program test2_pinvt
