program test_reig_pos_cmp
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine REIG_POS_CMP in module SVD_Procedures.
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    Random symmetric semi-positive matrices with given eigenvalue value distributions are
!    generated for checking the accuracy of subroutine REIG_POS_CMP.
!
! LATEST REVISION : 01/06/2021
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def, true, false,              &
                         zero, half, one, two, five, seven, ten, c30, c100,                &
                         qr_cmp, ortho_gen_qr, eigval_sort, reig_pos_cmp, norm,            &
                         arth, lamch, random_seed_, random_number_, normal_random_number3_
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, neig IS THE TARGET RANK OF THE APPROXIMATE EVD,
! niter IS THE NUMBER OF RANDOM POWER, SUBSPACE OR BLOCK KRYLOV ITERATIONS PERFORMED,
! nover IS THE OVERSAMPLING SIZE USED AND def_conda IS THE APPROXIMATE CONDITION NUMBER
! OF THE GENERATED SYMMETRIC SEMI-POSITIVE MATRIX FOR SOME CASES.
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def, neig=5, niter=40, nover=10
!   
    real(stnd), parameter  :: fudge=c100, def_conda=50000._stnd
!
    character(len=*), parameter :: name_proc='reig_pos_cmp'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err, eps, ulp, safmin, &
                                               tmp, norma, noise_level
    real(stnd), allocatable, dimension(:)   :: eigval, eigval2, diagr, beta
    real(stnd), allocatable, dimension(:,:) :: a, eigvec, eigvec2, eigvect, res
!
    integer      :: iok
    integer(i4b) :: i, j, q, l, n, nc, m, neig0, success, notdone
    logical(lgl) :: failure, details, add_noise, extd_samp, use_nystrom, ortho
!   
    character    :: sort='d'
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
!   SPECIFY IF GAUSSIAN NOISE IS ADDED TO THE GENERATED MATRIX.
!
    add_noise = true
!
!   SPECIFY THE LEVEL OF GAUSSIAN NOISE AS A SCALING FACTOR
!   OF THE LAST NON-ZERO SINGULAR VALUE.
!
    noise_level = 0.002_stnd
!
!   SPECIFY IF POWER SUBSPACE OR BLOCK KRYLOV ITERATIONS ARE USED.
!
    extd_samp = false
!
!   DETERMINE IF ORTHONORMALIZATION IS CARRIED OUT BETWEEN EACH STEP
!   OF THE (POWER) SUBSPACE OR BLOCK KRYLOV ITERATIONS, TO AVOID LOSS
!   OF ACCURACY DUE TO ROUNDING ERRORS. THIS IS NOT NEEDED NORMALLY
!   FOR POSITIVE MATRIX.
!
    ortho = true
!
!   SPECIFY IF LAST STEP OF THE ALGORITHM IS PERFORMED WITH THE NYSTROM
!   METHOD (AND A SVD) OR AN EIGENVALUE DECOMPOSITION.
!
    use_nystrom = true
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    safmin = lamch( 'S' )
!
!   SET THE TOLERANCE FOR THE ERROR OF THE RELATIVE ERROR IN FROBENIUS NORM.
!
    eps = 0.0001_stnd
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
        add_noise = .not.add_noise
!
        if ( add_noise ) then
!
            neig0 = min( neig + nover/2_i4b, n )
!
            nc = n
!
        else
!
            neig0 = n
!
            nc = neig
!
        end if
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( eigval(neig0), eigval2(neig), diagr(neig0), beta(neig0),  &
                  a(n,n), res(n,nc), eigvec(n,neig0), eigvec2(n,neig),      &
                  eigvect(neig0,n), stat=iok )
!
        if ( iok/=0 ) then
!
            if ( allocated(eigval) )    deallocate( eigval )
            if ( allocated(eigval2) )   deallocate( eigval2 )
            if ( allocated(diagr) )     deallocate( diagr )
            if ( allocated(beta) )      deallocate( beta )
            if ( allocated(a) )         deallocate( a )
            if ( allocated(res) )       deallocate( res )
            if ( allocated(eigvec) )    deallocate( eigvec )
            if ( allocated(eigvec2) )   deallocate( eigvec2 )
            if ( allocated(eigvect) )   deallocate( eigvect )
!
            cycle
!
        end if
!
!       GENERATE A RANDOM COEFFICIENT MATRIX eigvec(:n,:neig0) .
!
        call random_number_( eigvec(:n,:neig0) )
!
        eigvec(:n,:neig0) = eigvec(:n,:neig0) - half
!
        call random_number_( tmp )
!
        if ( tmp>safmin ) then
            eigvec(:n,:neig0) = eigvec(:n,:neig0)/tmp
        end if
!
!       COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
        call qr_cmp( eigvec(:n,:neig0), diagr(:neig0), beta(:neig0) )
!
!       GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!       q IS ASSUMED OF FULL RANK.
!
        call ortho_gen_qr( eigvec(:n,:neig0), beta(:neig0) )
!
        do j = 1_i4b, 24_i4b
!
            select case( j )
!
                case( 1_i4b )
!
!                   TEST 1: STRONGLY CLUSTERED EIGENVALUES.
!
                    eigval(1_i4b)       = one
                    eigval(2_i4b:neig0) = ulp
!
                case( 2_i4b )
!
!                   TEST 2:  STRONGLY CLUSTERED EIGENVALUES.
!
                    eigval(1_i4b:neig0-1_i4b) = one
                    eigval(neig0)             = ulp
!
                case( 3_i4b )
!
!                   TEST 3: WEAKLY CLUSTERED EIGENVALUES.
!
                    eigval(1_i4b:neig0-1_i4b) = arth( ulp, ulp, neig0-1_i4b )
                    eigval(neig0)             = one
!
                case( 4_i4b )
!
!                   TEST 4: INTERNAL CLUSTER OF EIGENVALUES.
!
                    eigval(1_i4b)             = ulp
                    eigval(2_i4b:neig0-1_i4b) = one + sqrt(ulp)*arth( one, one, neig0-2_i4b )
                    eigval(neig0)             = two
!
                case( 5_i4b )
!
!                   TEST 5: WEAKLY CLUSTERED EIGENVALUES.
!
                    eigval(1_i4b:neig0) = one + (c100*ulp)*arth( zero, one, neig0 )
!
                case( 6_i4b )
!
!                   TEST 6: GEOMETRIC DISTRIBUTION OF EIGENVALUES.
!
                    err = real( neig0-1_i4b, stnd )
!
                    do l = 1_i4b, neig0
                        eigval(l) = ulp**(real(l-1_i4b,stnd)/err)
                    end do
!
                case( 7_i4b )
!
!                   TEST 7: ARITHMETIC DISTRIBUTION OF EIGENVALUES.
!
                    eigval(1_i4b:neig0) = ulp + ((one-ulp)/real( neig0-1_i4b, stnd ))*arth( zero, one, neig0 )
!
                case( 8_i4b )
!
!                   TEST 8: TWO CLUSTERS OF EIGENVALUES.
!
                    i = neig0/2_i4b
!
                    tmp = sqrt( ulp )
!
                    eigval(1_i4b:i) = one + tmp*arth( one, one, i )
!
                    eigval(i+1_i4b:neig0) = ten + tmp*arth( one, one, neig0-i )
!
                case( 9_i4b )
!
!                   TEST 9: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR ZERO.
!
                    i = neig0/2_i4b
!
                    tmp = sqrt( ulp )
!
                    eigval(1_i4b:i) = arth( one, one, i )
!
                    eigval(i+1_i4b:neig0) = tmp*arth( one, one, neig0-i )
!
                case( 10_i4b )
!
!                   TEST 10: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR neig0/2.
!
                    i = neig0/2_i4b
!
                    tmp = sqrt( ulp )
!
                    eigval(1_i4b:i) = arth( one, one, i )
!
                    eigval(i+1_i4b:neig0) = real( i, stnd) + tmp*arth( one, one, neig0-i )
!
                case( 11_i4b )
!
!                   TEST 11: CLUSTERED DISTRIBUTION OF EIGENVALUES BETWEEN 0.9 and 1.
!
                    call random_number_( eigval(:neig0) )
!
                    eigval(1_i4b:neig0) = 0.9_stnd + 0.1_stnd*eigval(1_i4b:neig0)
!
                case( 12_i4b )
!
!                   TEST 12: EIGENVALUES EQUAL TO sinh(10*i/neig0) FOR i=1 TO neig0.
!
                    tmp = ten/real( neig0, stnd)
!
                    eigval(1_i4b:neig0) = sinh( tmp*arth( one, one, neig0 ) )
!
                case( 13_i4b )
!
!                   TEST 13: EIGENVALUES EQUAL TO sinh((10*i/neig0)-5) FOR i=1 TO neig0.
!
                    tmp = ten/real( neig0, stnd)
!
                    eigval(1_i4b:neig0) = sinh( tmp*arth( one, one, neig0 ) - five )
!
!                   SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                    where( eigval(1_i4b:neig0)<zero )
                         beta(:neig0) = -eigval(1_i4b:neig0)
                    elsewhere
                         beta(:neig0) = zero
                    end where
!
                    tmp = maxval( beta(:neig0) )
!
                    eigval(1_i4b:neig0) = tmp + eigval(1_i4b:neig0)
!
                case( 14_i4b )
!
!                   TEST 14: EIGENVALUES EQUAL TO sinh((10*i/neig0)-10) FOR i=1 TO neig0.
!
                    tmp = ten/real( neig0, stnd)
!
                    eigval(1_i4b:neig0) = sinh( tmp*arth( one, one, neig0 ) - ten )
!
!                   SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                    where( eigval(1_i4b:neig0)<zero )
                         beta(:neig0) = -eigval(1_i4b:neig0)
                    elsewhere
                         beta(:neig0) = zero
                    end where
!
                    tmp = maxval( beta(:neig0) )
!
                    eigval(1_i4b:neig0) = tmp + eigval(1_i4b:neig0)
!
                case( 15_i4b )
!
!                   TEST 15: EIGENVALUES EQUAL TO tanh((10*i/neig0)-5) FOR i=1 TO neig0.
!
                    tmp = ten/real( neig0, stnd)
!
                    eigval(1_i4b:neig0) = tanh( tmp*arth( one, one, neig0 ) - five )
!
!                   SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                    where( eigval(1_i4b:neig0)<zero )
                         beta(:neig0) = -eigval(1_i4b:neig0)
                    elsewhere
                         beta(:neig0) = zero
                    end where
!
                    tmp = maxval( beta(:neig0) )
!
                    eigval(1_i4b:neig0) = tmp + eigval(1_i4b:neig0)
!
                case( 16_i4b )
!
!                   TEST 16: CLUSTER OF EIGENVALUES AT 1 WITH CONDITION NUMBER def_conda.
!
                    eigval(1_i4b)       = one/def_conda
                    eigval(2_i4b:neig0) = one
!
               case( 17_i4b )
!
!                   TEST 17: CLUSTER OF EIGENVALUES AT 1/conda WITH CONDITION NUMBER def_conda.
!
                    eigval(1_i4b)       = one
                    eigval(2_i4b:neig0) = one/def_conda
!
                case( 18_i4b )
!
!                   TEST 18: ARITHMETIC DISTRIBUTION OF EIGENVALUES WITH CONDITION NUMBER def_conda.
!
                    tmp = real( neig0-1_i4b, stnd )
                    err = one - one/def_conda
!
                    do l = 1_i4b, neig0
                        eigval(l) = one - ( real(l-1_i4b,stnd)/tmp )*err
                    end do
!
                case( 19_i4b )
!
!                   TEST 19: GEOMETRIC DISTRIBUTION OF EIGENVALUES WITH CONDITION NUMBER def_conda.
!
                    tmp = real( neig0-1_i4b, stnd )
!
                    do l = 1_i4b, neig0
                        eigval(l) = def_conda**( -real(l-1_i4b,stnd)/tmp )
                    end do
!
                case( 20_i4b )
!
!                   TEST 20: RANDOM DISTRIBUTION OF EIGENVALUES IN THE INTERVAL 1/def_conda TO 1
!                            WITH CONDITION NUMBER def_conda.
!
                    call random_number_( beta(:neig0) )
!
                    tmp = log( one/def_conda )

                    beta(:neig0)   = tmp + beta(:neig0)*abs( tmp )
                    eigval(:neig0) = exp( beta(:neig0) )
!
                case( 21_i4b )
!
!                   TEST 21: VERY SLOW DECAY OF EIGENVALUES.
!
                    err = real( neig0 - 1_i4b, stnd )
!
                    do l = 1_i4b, neig0
!
                        tmp = real( l-1_i4b, stnd )
!
                        eigval(l) = exp( -tmp/err )
!        
                    end do
!
                case( 22_i4b )
!
!                   TEST 22: SLOW DECAY OF EIGENVALUES.
!
                    do l = 1_i4b, neig0
!
                        tmp = real( l, stnd )
!
                        eigval(l) = one/( tmp*tmp )
!        
                    end do
!
                case( 23_i4b )
!
!                   TEST 23: FAST DECAY OF EIGENVALUES.
!
                    do l = 1_i4b, neig0
!
                        tmp = real( l, stnd )
!
                        eigval(l) = exp( -tmp/seven )
!        
                    end do
!
                case( 24_i4b )
!
!                   TEST 24: S-SHAPE DECAY OF EIGENVALUES.
!
                    do l = 1_i4b, neig0
!
                        tmp = real( l, stnd )
!
                        eigval(l) = 0.0001_stnd + one/( one + exp( tmp -c30 ) )
!        
                    end do
!
                case default
!
!                   DEFAULT TEST : RANDOMLY UNIFORM DISTRIBUTION OF EIGENVALUES.
!
                    call random_number_( eigval(:neig0) )
!
                    call random_number_( tmp )
!
                    if ( tmp>safmin ) then
                        eigval(1_i4b:neig0) = eigval(1_i4b:neig0)/tmp
                    end if
!
            end select
!
#ifdef _F2003
            if ( ieee_support_datatype( eigval ) ) then
                if ( .not.all( ieee_is_normal( eigval(1_i4b:neig0) ) ) ) cycle
            end if
#endif
!
!           SORT THE EIGENVALUES BY DECREASING ABSOLUTE MAGNITUDE.
!
            call eigval_sort( sort, eigval(:neig0) )
!
!           FORM A SYMMETRIC MATRIX FROM THE SPECIFIED EIGENVALUE DISTRIBUTION AND RANDOM EIGENVECTORS.
!
            eigvect(:neig0,:n) = transpose( eigvec(:n,:neig0) )
!
            do i = 1_i4b, n
!
                eigvect(:neig0,i) = eigvect(:neig0,i)*eigval(:neig0)
!
            end do
!
            a(:n,:n) = matmul( eigvec(:n,:neig0), eigvect(:neig0,:n) )
!
            if ( add_noise ) then
!
!               ADD GAUSSIAN NOISE TO THE DEFICIENT SYMMETRIC MATRIX a .
!
                tmp = noise_level*abs(eigval(neig0))
!
!               call normal_random_number3_( res(:n,:n) )
!
                do i = 1_i4b, n
!
                    call normal_random_number3_( res(:n,i) )
!
                end do
!
                a(:n,:n) = a(:n,:n) + tmp*matmul( transpose( res(:n,:n) ), res(:n,:n) )
!
!               COMPUTE THE FROBENIUS NORM OF THE MATRIX.
!
                norma = norm( a(:n,:n) )
!
            else
!
!               COMPUTE THE FROBENIUS NORM OF THE MATRIX.
!
                norma = sqrt(sum( eigval(:neig0)**2 ) )
!
            end if
!
!           COMPUTE APPROXIMATE PARTIAL EVD OF RANDOM DATA MATRIX.
!
            extd_samp = .not.extd_samp
!
            use_nystrom = .not.use_nystrom
!
            call reig_pos_cmp( a(:n,:n), eigval2(:neig), eigvec2(:n,:neig),      &
                               failure=failure, niter=niter, nover=nover,        &
                               ortho=ortho, extd_samp=extd_samp,                 &
                               use_nystrom=use_nystrom )
!
!            if ( failure ) then
!
!                notdone = notdone + 1_i4b
!
!                cycle
!
!            end if
!
            m = m + 1_i4b
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*eigvec(:n,:neig) - eigvec(:n,:neig)*diag(eigval(:neig)),
!           WHERE eigval AND u ARE THE EIGENVALUES AND EIGENVECTORS OF a.
!
            res(:n,:neig) = matmul( a(:n,:n), eigvec2(:n,:neig) )    &
                            - eigvec2(:n,:neig)*spread(eigval2(:neig),1,n)
!
            if ( norma/=zero ) then
                err1 = maxval( norm(res(:n,:neig), dim=2_i4b) )/( norma*real(n,stnd) )
            else
                err1 = maxval( norm(res(:n,:neig), dim=2_i4b) )/real(n,stnd)
            end if
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - eigvec**(t)*eigvec
!           WHERE eigvec are THE EIGENVECTORS OF THE MATRIX.
!
            res(:neig,:neig) = matmul( transpose( eigvec2(:n,:neig) ), eigvec2(:n,:neig) )
!
            do l = 1_i4b, neig
                res(l,l) = one - res(l,l)
            end do
!
            err2 = maxval( norm(res(:neig,:neig), dim=2_i4b) )/real(n,stnd)
!
            err = max( err1, err2 )
!
            if ( err<=eps ) then
!
                success = success + 1_i4b
!
                if ( details ) then
!
                    write (prtunit,*) 'Test ',m,' is correct'
!
                    write (prtunit,*) 'n= ', n,' , case= ', j,' , add_noise= ', add_noise,        &
                                      ' , neig= ', neig,' , niter= ', niter,' , nover= ', nover,  &
                                      ' , ortho= ', ortho,' , extd_samp= ', extd_samp,            &
                                      ' , use_nystrom= ', use_nystrom
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION
!                           a*eigvec(:n,:neig) - eigvec(:n,:neig)*diag(eigval(:neig)),
!                   WHERE eigval AND u ARE EIGENVALUES AND EIGENVECTORS OF a.
!
                    write (prtunit,*) 'Accuracy of the computed eigen couplets       = ', &
                                      err1, ' (must be less than :', eps, ' )'
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - eigvec**(t)*eigvec
!                   WHERE eigvec are EIGENVECTORS OF THE MATRIX.
!
                    write (prtunit,*) 'Orthogonality of the computed eigenvectors    = ', &
                                      err2/ulp, ' (must be less than :', fudge, ' )'
!
                    write (prtunit,*)
!
                end if
!
            else
!
                write (prtunit,*) 'Test ',m,' is incorrect'
!
                write (prtunit,*) '=> ',err,' > ',eps,' , n= ', n,' , case= ', j,' , add_noise= ', add_noise,     &
                                  ' , neig= ', neig,' , niter= ', niter,' , nover= ', nover,' , ortho= ', ortho,  &
                                  ' , extd_samp= ', extd_samp,' , use_nystrom= ', use_nystrom,                    &
                                  ' , failure= ', failure
!
!               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION
!                         a*eigvec(:n,:neig) - eigvec(:n,:neig)*diag(eigval(:neig)),
!               WHERE eigval AND u ARE EIGENVALUES AND EIGENVECTORS OF a.
!
                write (prtunit,*) 'Accuracy of the computed eigen couplets       = ', &
                                  err1, ' (must be less than :', eps, ' )'
!
!               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - eigvec**(t)*eigvec
!               WHERE eigvec are EIGENVECTORS OF THE MATRIX.
!
                write (prtunit,*) 'Orthogonality of the computed eigenvectors   = ', &
                                  err2/ulp, ' (must be less than :', fudge, ' )'
!
                write (prtunit,*)
!
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( eigval, eigval2, diagr, beta, a, res, eigvec, eigvec2, eigvect )
!
    end do
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_reig_pos_cmp
! ================================
!
end program test_reig_pos_cmp
