program test2_symtrid_qri3
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine SYMTRID_QRI3 in module Eig_Procedures.
!                                                                              
!
! LATEST REVISION : 27/09/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,   &
                         true, zero, one, c50, gen_symtrid_mat,    &
                         symtrid_qri3, norm
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def
!
    character(len=*), parameter :: name_proc='symtrid_qri3'
!   
    real(stnd), parameter  :: fudge=c50
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                                :: err1, err2, err3, err, eps, ulp, norma
    real(stnd), allocatable, dimension(:)     :: d, d2, e2, e
    real(stnd), allocatable, dimension(:,:)   :: a, eigvec, res
!
    integer(i4b)     :: i, l, m, n, success, notdone
    integer          :: iok
!
    logical(lgl)     :: failure, failure2, details
!   
    character        :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( err )
    eps = fudge*ulp
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( d(n), e(n), d2(n), e2(n), a(n,n),   &
                  res(n,n), eigvec(n,n), stat=iok   )
!
        if ( iok/=0 ) then
!
            if ( allocated(d) )        deallocate( d )
            if ( allocated(e) )        deallocate( e )
            if ( allocated(d2) )       deallocate( d2 )
            if ( allocated(e2) )       deallocate( e2 )
            if ( allocated(eigvec) )   deallocate( eigvec )
            if ( allocated(a) )        deallocate( a )
            if ( allocated(res) )      deallocate( res )
!
            cycle
!
        end if
!
        a(:n,:n) = zero
!
        do i = 1_i4b, 50_i4b
!
!           SKIP GRADED MATRICES WITH A GEOMETRIC PROGRESSION TO AVOID
!           OVERFLOW PROBLEMS IN SINGLE PRECISION.
!
            if ( i==35_i4b .or. i==36_i4b ) cycle
!
!           GENERATE A TRIDIAGONAL MATRIX.
!           THE DIAGONAL ELEMENTS ARE STORED IN d.
!           THE OFF-DIAGONAL ELEMENTS ARE STORED IN e.
!
            call gen_symtrid_mat( i, d(:n), e(:n) )
!
#ifdef _F2003
            if ( ieee_support_datatype( d ) ) then
                if ( .not.all( ieee_is_normal( d(:n) ) ) ) cycle
                if ( .not.all( ieee_is_normal( e(:n) ) ) ) cycle
            end if
#endif
!
!           SAVE THE TRIDIAGONAL FORM FOR LATER USE.
!
            d2(:n) = d(:n)
            e2(:n) = e(:n)
!
!           FORM THE TRIDIAGONAL MATRIX.
!
            do l = 1_i4b, n-1_i4b
                a(l,l)       = d(l)
                a(l+1_i4b,l) = e(l)
                a(l,l+1_i4b) = e(l)
            end do
!
            a(n,n) = d(n)
!
!           COMPUTE EIGENVALUES AND EIGENVECTORS OF THE TRIDIAGONAL MATRIX.
!
            call symtrid_qri3( d(:n), e(:n), failure, eigvec(:n,:n), init_mat=true, sort=sort )
!
            if ( failure ) then
!
                notdone = notdone + 1_i4b
!
                cycle
!
            end if
!
            m = m + 1_i4b
!
!           COMPUTE THE NORM OF THE TRIDIAGONAL MATRIX.
!
            norma = norm( d(:n) )
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*u - u*diag(s)
!           WHERE s ARE THE EIGENVALUES AND u THE EIGENVECTORS OF THE TRIDIAGONAL MATRIX.
!
            res(:n,:n) = matmul( a(:n,:n), eigvec(:n,:n) ) - eigvec(:n,:n)*spread(d(:n),1,n)
!
            if ( norma/=zero ) then
                err1 = maxval( norm(res(:n,:n), dim=2_i4b) )/( norma*real(n,stnd) )
            else
                err1 = maxval( norm(res(:n,:n), dim=2_i4b) )/real(n,stnd)
            end if
!
!           RECOMPUTE EIGENVALUES OF THE TRIDIAGONAL MATRIX.
!
            call symtrid_qri3( d2(:n), e2(:n), failure2, sort=sort )
!
            if ( .not.failure2 ) then
!
!               RECOMPUTE THE RESIDUALS WITH THE SECOND ESTIMATES OF THE EIGENVALUES.
!
                res(:n,:n) = matmul( a(:n,:n), eigvec(:n,:n) ) - eigvec(:n,:n)*spread(d2(:n),1,n)
!
                if ( norma/=zero ) then
                    err2 = maxval( norm(res(:n,:n), dim=2_i4b) )/( norma*real(n,stnd) )
                else
                    err2 = maxval( norm(res(:n,:n), dim=2_i4b) )/real(n,stnd)
                end if
!
            else
!
                err2 = zero
!
            end if
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!           WHERE u ARE THE EIGENVECTORS OF THE TRIDIAGONAL MATRIX.
!
            res(:n,:n) = matmul( transpose( eigvec(:n,:n) ), eigvec(:n,:n) )
!
            do l = 1_i4b, n
                res(l,l) = one - res(l,l)
            end do
!
            err3 = maxval( norm(res(:n,:n), dim=2_i4b) )/real(n,stnd)
!
            err = max( err1, err2, err3 )
!
            if ( err<=eps .and. .not.failure2 ) then
!
                success = success + 1_i4b
!
                if ( details ) then
!
                    write (prtunit,*) 'Test ',m,' is correct'
!
                    write (prtunit,*) 'n= ', n,' , case= ', i
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*u - u*diag(s)
!                   WHERE s ARE THE EIGENVALUES AND u THE EIGENVECTORS OF THE
!                   TRIDIAGONAL MATRIX.
!
                    write (prtunit,*) 'Accuracy of the computed eigenvalues and eigenvectors = ', &
                                      max(err1,err2)/ulp, ' (must be less than :', fudge, ' )'
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!                   WHERE u are EIGENVECTORS OF THE TRIDIAGONAL MATRIX.
!
                    write (prtunit,*) 'Orthogonality of the computed eigenvectors            = ', &
                                      err3/ulp, ' (must be less than :', fudge, ' )'
!
                    write (prtunit,*)
!
                end if
!
            else
!
                write (prtunit,*) 'Test ',m,' is incorrect'
!
                write (prtunit,*) '=> ',err,' > ',eps,' , n= ', n,' , case= ', i,          &
                                  ' , failure1= ', failure, ' , failure2= ', failure2
!
!               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*u - u*diag(s)
!               WHERE s ARE THE EIGENVALUES AND u THE EIGENVECTORS OF THE
!               TRIDIAGONAL MATRIX.
!
                write (prtunit,*) 'Accuracy of the computed eigenvalues and eigenvectors = ', &
                                  max(err1,err2)/ulp, ' (must be less than :', fudge, ' )'
!
!               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!               WHERE u are THE EIGENVECTORS OF THE TRIDIAGONAL MATRIX.
!
                write (prtunit,*) 'Orthogonality of the computed eigenvectors            = ', &
                                  err3/ulp, ' (must be less than :', fudge, ' )'
!
                write (prtunit,*)
!
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( d, e, d2, e2, a, res, eigvec )
!
    end do
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test2_symtrid_qri3
! =================================
!
end program test2_symtrid_qri3
