program test_apply_q_symtrid
!
!
! Purpose
! =======
!
!   This program is intended to test subroutines SYMTRID_CMP and APPLY_Q_SYMTRID
!   in module Eig_Procedures.
!                                                                              
!                                                                              
! Further Details
! ===============
!
!   The program assumes that subroutines ORTHO_GEN_SYMTRID and SYMTRID_QRI
!   in module Eig_Procedures are correct.
!
! LATEST REVISION : 17/11/2019
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,            &
                         true, false, zero, half, one, triangle, lamch,     &
                         norm, symtrid_cmp, ortho_gen_symtrid, symtrid_qri, &
                         apply_q_symtrid
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def
!   
    real(stnd), parameter :: fudge=one
!
    character(len=*), parameter :: name_proc='apply_q_symtrid'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, ulp, safmin, tmp, norma
    real(stnd), allocatable, dimension(:)   :: vec, d, d2, e, e2
    real(stnd), allocatable, dimension(:,:) :: a, a2, u, v, work, proj
!
    integer      :: iok
    integer(i4b) :: i, j, m, n, p, success, notdone
!
    logical(lgl) :: failure, details, upper, packed, left, trans
!   
    character :: sort='d'
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TESTS.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    eps    = fudge*ulp
    safmin = lamch( 'S' )
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        p = n*(n+1_i4b)/2_i4b
!
        allocate( vec(p), d(n), e(n), d2(n), e2(n),  &
             a(n,n), a2(n,n), work(n,n), proj(n,n),  &
             u(n,n), v(n,n), stat=iok                )
!
        if ( iok/=0 ) cycle
!
!       GENERATE A RANDOM DATA MATRIX AND FROM IT
!       A SELF-ADJOINT MATRIX a.
!
        call random_number( work(:n,:n) )
!
        work(:n,:n) = work(:n,:n) - half
!
        call random_number( tmp )
!
        if ( tmp>safmin ) then
            work(:n,:n) = work(:n,:n)/tmp
        end if
!
        a(:n,:n) = work(:n,:n) + transpose( work(:n,:n) ) 
!
!       COMPUTE THE NORM OF THE SYMMETRIC MATRIX.
!
        norma = norm( a(:n,:n) )
!
        upper = false
!
        do j = 1_i4b, 6_i4b
!
            upper = .not.upper
!
!           FIRST TRIDIAGONALIZE THE RANDOM SELF-ADJOINT MATRIX.
!
            select case ( j )
!
            case ( 1_i4b, 2_i4b )
!
                packed = false
!
!               MAKE A COPY OF THE RANDOM SELF-ADJOINT MATRIX.
!
                a2(:n,:n) = a(:n,:n)
!
                call symtrid_cmp( a2(:n,:n), d(:n), e(:n), store_q=true, upper=upper )
!
            case ( 3_i4b, 4_i4b )
!
                packed = true
!
!               MAKE A COPY OF THE MATRIX IN PACKED FORM.
!
                vec(:p) = pack( a(:n,:n), mask=triangle( upper, n, n, extra=1_i4b) )
!
                call symtrid_cmp( vec(:p), d(:n), e(:n), store_q=true, upper=upper )
!
                a2(:n,:n) = unpack( vec(:p), mask=triangle( upper, n, n, extra=1_i4b), field=zero )
!
            case ( 5_i4b )
!
                packed = false
!
                upper = true
!
!               MAKE A COPY OF THE RANDOM SELF-ADJOINT MATRIX.
!
                a2(:n,:n) = a(:n,:n)
!
                call symtrid_cmp( a2(:n,:n), d(:n), e(:n), store_q=true )
!
            case ( 6_i4b )
!
                packed = true
!
                upper = true
!
!               MAKE A COPY OF THE MATRIX IN PACKED FORM.
!
                vec(:p) = pack( a(:n,:n), mask=triangle( true, n, n, extra=1_i4b) )
!
                call symtrid_cmp( vec(:p), d(:n), e(:n), store_q=true )
!
                a2(:n,:n) = unpack( vec(:p), mask=triangle( true, n, n, extra=1_i4b), field=zero )
!
            end select
!
!           CALL ortho_gen_symtrid TO GENERATE THE ORTHOGONAL MATRIX.
!
            u(:n,:n) = a2(:n,:n)
!
            call ortho_gen_symtrid( u(:n,:n), upper )
!
!           COMPUTE EIGENVALUES AND EIGENVECTORS OF THE SELF-ADJOINT MATRIX.
!
            d2(:n) = d(:n)
            e2(:n) = e(:n)
!
            call symtrid_qri( d2(:n), e2(:n), failure, u(:n,:n), sort=sort )
!
            if ( failure ) then
!
                notdone = notdone + 1_i4b
!
                cycle
!
            end if
!
!           MULTIPLY EIGENVECTORS BY EIGENVALUES.
!
            u(:n,:n) = u(:n,:n)*spread(d2(:n),1,n)
!
!           COMPUTE EIGENVALUES AND EIGENVECTORS OF THE TRIDIAGONAL MATRIX .
!
            call symtrid_qri( d(:n), e(:n), failure, mat=v(:n,:n), init_mat=true, sort=sort )
!
            if ( failure ) then
!
                notdone = notdone + 1_i4b
!
                cycle
!
            end if
!
            do i = 1_i4b, 4_i4b
!
                m = m + 1_i4b
!
                select case( i )
!
                case( 1_i4b )
!
!                   TESTING apply_q_symtrid SUBROUTINES WITH left=true AND trans=false.
!
                    left  = true
                    trans = false
!
                    work(:n,:n) = v(:n,:n)
!
                    select case ( j )
!
                    case ( 1_i4b, 2_i4b )
!
                        call apply_q_symtrid( a2,  work, left=true, trans=false, upper=upper )
!
                    case ( 3_i4b, 4_i4b )
!
                        call apply_q_symtrid( vec, work, left=true, trans=false, upper=upper )
!
                    case ( 5_i4b )
!
                        call apply_q_symtrid( a2,  work, left=true, trans=false )
!
                    case ( 6_i4b )
!
                        call apply_q_symtrid( vec, work, left=true, trans=false )
!
                    end select
!
                    proj(:n,:n) = matmul( a(:n,:n), work(:n,:n) )
!
                case( 2_i4b )
!
!                   TESTING apply_q_symtrid SUBROUTINES WITH left=false AND trans=false.
!
                    left  = false
                    trans = false
!
                    work(:n,:n) = a(:n,:n)
!
                    select case ( j )
!
                    case ( 1_i4b, 2_i4b )
!
                        call apply_q_symtrid( a2,  work, left=false, trans=false, upper=upper )
!
                    case ( 3_i4b, 4_i4b )
!
                        call apply_q_symtrid( vec, work, left=false, trans=false, upper=upper )
!
                    case ( 5_i4b )
!
                        call apply_q_symtrid( a2,  work, left=false, trans=false )
!
                    case ( 6_i4b )
!
                        call apply_q_symtrid( vec, work, left=false, trans=false )
!
                    end select
!
                    proj(:n,:n) = matmul( work(:n,:n), v(:n,:n) )
!
                case( 3_i4b )
!
!                   TESTING apply_q_symtrid SUBROUTINES WITH left=true AND trans=true.
!
                    left  = true
                    trans = true
!
                    work(:n,:n) = a(:n,:n)
!
                    select case ( j )
!
                    case ( 1_i4b, 2_i4b )
!
                        call apply_q_symtrid( a2,  work, left=true, trans=true, upper=upper )
!
                    case ( 3_i4b, 4_i4b )
!
                        call apply_q_symtrid( vec, work, left=true, trans=true, upper=upper )
!
                    case ( 5_i4b )
!
                        call apply_q_symtrid( a2,  work, left=true, trans=true )
!
                    case ( 6_i4b )
!
                        call apply_q_symtrid( vec, work, left=true, trans=true )
!
                    end select
!
                    proj(:n,:n) = matmul( transpose(work(:n,:n)), v(:n,:n) )
!
                case( 4_i4b )
!
!                   TESTING apply_q_symtrid SUBROUTINES WITH left=false AND trans=true.
!
                    left  = false
                    trans = true
!
                    work(:n,:n) = transpose( v(:n,:n) )
!
                    select case ( j )
!
                    case ( 1_i4b, 2_i4b )
!
                        call apply_q_symtrid( a2,  work, left=false, trans=true, upper=upper )
!
                    case ( 3_i4b, 4_i4b )
!
                        call apply_q_symtrid( vec, work, left=false, trans=true, upper=upper )
!
                    case ( 5_i4b )
!
                        call apply_q_symtrid( a2,  work, left=false, trans=true )
!
                    case ( 6_i4b )
!
                        call apply_q_symtrid( vec, work, left=false, trans=true )
!
                    end select
!
                    proj(:n,:n) = matmul( a(:n,:n), transpose(work(:n,:n)) )
!
                end select
!
!               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*u - u*d
!               WHERE d ARE THE EIGENVALUES AND u THE EIGENVECTORS OF
!               THE SELF-ADJOINT MATRIX a.
!
                work(:n,:n) = proj(:n,:n) - u(:n,:n)
!
                if ( norma/=zero ) then
                    err = maxval( norm(work(:n,:n), dim=2_i4b) )/( norma*real(n,stnd) )
                else
                    err = maxval( norm(work(:n,:n), dim=2_i4b) )/real(n,stnd)
                end if
!
                if ( err<=eps ) then
!
                    success = success + 1_i4b
!
                    if ( details ) then
!
                        write (prtunit,*) 'Test ',m,' is correct'
!
                        if ( j<5_i4b ) then
!
                            write (prtunit,*) 'n= ', n,' , case= ', j,' , upper= ', upper, &
                                 ' , packed= ', packed,' , left= ', left,                  &
                                 ' , trans= ', trans
!
                        else
!
                            write (prtunit,*) 'n= ', n,' , case= ', j,     &
                                 ' , packed= ', packed,' , left= ', left,  &
                                 ' , trans= ', trans
!
                        end if
!
!                       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*u - u*s
!                       WHERE s ARE THE EIGENVALUES AND u THE EIGENVECTORS OF THE MATRIX.
!
                        write (prtunit,*) 'Accuracy of the computed eigenpairs = ', &
                             err/ulp, ' (must be less than :', fudge, ' )'
!
                        write (prtunit,*)
!
                    end if
!
                else
!
                    write (prtunit,*) 'Test ',m,' is incorrect'
!
                    if ( j<5_i4b ) then
!
                        write (prtunit,*) '=> ',err,' > ', eps,' , n= ', n,' , case= ', j, &
                                          ' , upper= ', upper,' , packed= ', packed,       &
                                          ' , left= ', left,' , trans= ', trans
!
                    else
!
                        write (prtunit,*) '=> ',err,' > ', eps,' , n= ', n,' , case= ', j, &
                                          ' , packed= ', packed,' , left= ', left,         &
                                          ' , trans= ', trans
!
                    end if
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*u - u*s
!                   WHERE s ARE THE EIGENVALUES AND u THE EIGENVECTORS OF THE MATRIX.
!
                    write (prtunit,*) 'Accuracy of the computed eigenpairs = ',     &
                                      err/ulp, ' (must be less than :', fudge, ' )'
!
                    write (prtunit,*)
!
                end if
!
            end do
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( vec, d, e, d2, e2, a, a2, work, proj, u, v )
!
    end do
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_apply_q_symtrid
! ===================================
!
end program test_apply_q_symtrid
