program test_bd_inviter2
!
!
! Purpose
! =======
!
!   This program is intended to test subroutines SELECT_SINGVAL_CMP4, BD_INVITER and BD_INVITER2
!   in module SVD_Procedures.
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    Random matrices with given singular value distributions are generated for checking the accuracy
!    of subroutines SELECT_SINGVAL_CMP4, BD_INVITER and BD_INVITER2.
!
! LATEST REVISION : 31/05/2021
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def, false,           &
                         true, zero, half, one, two, five, ten, c100, c1000,      &
                         qr_cmp, ortho_gen_qr, eigval_sort, select_singval_cmp4,  &
                         bd_inviter2, norm, arth, lamch
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def, nsing_def=5
!   
    real(stnd), parameter  :: fudge=c1000
!
    character(len=*), parameter :: name_proc='bd_inviter2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err3, err, eps, eps0, eps2, ulp, safmin, &
                                               abstol, tmp, norma
    real(stnd), allocatable, dimension(:)   :: eigval, singval, singval2, d, e, diagr, beta, tauo
    real(stnd), allocatable, dimension(:,:) :: a, a2, leftvec, rightvec, leftvec2, rightvec2, res, p, ra
!
    integer      :: iok
    integer(i4b) :: i, j, k, l, n, nr, m, o, q, nsing, success, notdone, maxiter=2
    logical(lgl) :: gen_p, gen_o, two_stage, failure, failure2, details
!
    character    :: sort='d'
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    eps    = fudge*ulp
    eps0   = sqrt( ulp )
!    eps0   = eps
    safmin = lamch('S')
!
    abstol = sqrt( safmin )
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
    gen_o = false
    gen_p = true
!
    do n = n1, n2, n3
!
        gen_o = .not.gen_o
        gen_p = .not.gen_p
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( ra(n,n), eigval(n), singval(n), singval2(n), d(n), e(n),    &
                  diagr(n), beta(n), tauo(n), rightvec(n,n), rightvec2(n,n),  &
                  p(n,n), stat=iok  )
!
        if ( iok/=0 ) then
!
            if ( allocated(ra) )        deallocate( ra )
            if ( allocated(eigval) )    deallocate( eigval )
            if ( allocated(singval) )   deallocate( singval )
            if ( allocated(singval2) )  deallocate( singval2 )
            if ( allocated(d) )         deallocate( d )
            if ( allocated(e) )         deallocate( e )
            if ( allocated(diagr) )     deallocate( diagr )
            if ( allocated(beta) )      deallocate( beta )
            if ( allocated(tauo) )      deallocate( tauo )
            if ( allocated(rightvec) )  deallocate( rightvec )
            if ( allocated(rightvec2) ) deallocate( rightvec2 )
            if ( allocated(p) )         deallocate( p )
!
            cycle
!
        end if
!
        do k = 1_i4b, 3_i4b
!
            select case( k )
!
                case( 1_i4b )
!
!                   TEST 1: SVD OF a(:nr,:n) WITH nr=n .
!
                    nr = n
!
                case( 2_i4b )
!
!                   TEST 2: SVD OF a(:nr,:n) WITH nr>=n AND nr NOT MUCH LARGER THAN n .
!
                    nr = n + 1_i4b
!
                case( 3_i4b )
!
!                   TEST 3: SVD OF a(:nr,:n) WITH nr>=n AND nr MUCH LARGER THAN n .
!
                    nr = n + n
!
            end select
!
!           ALLOCATE WORK ARRAYS.
!
            allocate( a(nr,n), a2(nr,n), res(nr,n),            &
                      leftvec(nr,n), leftvec2(nr,n), stat=iok  )
!
            if ( iok/=0 ) then
!
                if ( allocated(a) )         deallocate( a )
                if ( allocated(a2) )        deallocate( a2 )
                if ( allocated(res) )       deallocate( res )
                if ( allocated(leftvec) )   deallocate( leftvec )
                if ( allocated(leftvec2) )  deallocate( leftvec2 )
!
                cycle
!
            end if
!
!           GENERATE A RANDOM COEFFICIENT MATRIX leftvec(:nr,:n) .
!
            call random_number( leftvec(:nr,:n) )
!
            leftvec(:nr,:n) = leftvec(:nr,:n) - half
!
            call random_number( tmp )
!
            if ( tmp>safmin ) then
                leftvec(:nr,:n) = leftvec(:nr,:n)/tmp
            end if
!
!           COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
            call qr_cmp( leftvec(:nr,:n), diagr(:n), beta(:n) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( leftvec(:nr,:n), beta(:n) )
!
!           GENERATE A RANDOM COEFFICIENT MATRIX rightvec(:n,:n) .
!
            call random_number( rightvec(:n,:n) )
!
            rightvec(:n,:n) = rightvec(:n,:n) - half
!
            call random_number( tmp )
!
            if ( tmp>safmin ) then
                rightvec(:n,:n) = rightvec(:n,:n)/tmp
            end if
!
!           COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
            call qr_cmp( rightvec(:n,:n), diagr(:n), beta(:n) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( rightvec(:n,:n), beta(:n) )
!
            do j = 1_i4b, 16_i4b
!
                select case( j )
!
                    case( 1_i4b )
!
!                       TEST 1: STRONGLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b)   = one
                        eigval(2_i4b:n) = ulp
!
                    case( 2_i4b )
!
!                       TEST 2:  STRONGLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n-1_i4b) = one
                        eigval(n)             = ulp
!
                    case( 3_i4b )
!
!                       TEST 3: WEAKLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n-1_i4b) = arth( ulp, ulp, n-1_i4b )
                        eigval(n)             = one
!
                    case( 4_i4b )
!
!                       TEST 4: INTERNAL CLUSTER OF EIGENVALUES.
!
                        eigval(1_i4b)         = ulp
                        eigval(2_i4b:n-1_i4b) = one + sqrt(ulp)*arth( one, one, n-2_i4b )
                        eigval(n)             = two
!
                    case( 5_i4b )
!
!                       TEST 5: WEAKLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n) = one + (c100*ulp)*arth( zero, one, n )
!
                    case( 6_i4b )
!
!                       TEST 6: GEOMETRIC DISTRIBUTION OF EIGENVALUES.
!
                        err = real( n-1_i4b, stnd )
!
                        do l = 1_i4b, n
                            eigval(l) = ulp**(real(l-1_i4b,stnd)/err)
                        end do
!
                    case( 7_i4b )
!
!                       TEST 7: ARITHMETIC DISTRIBUTION OF EIGENVALUES.
!
                        eigval(1_i4b:n) = ulp + ((one-ulp)/real( n-1_i4b, stnd ))*arth( zero, one, n )
!
                    case( 8_i4b )
!
!                       TEST 8: TWO CLUSTERS OF EIGENVALUES.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = one + tmp*arth( one, one, i )
!
                        eigval(i+1_i4b:n) = ten + tmp*arth( one, one, n-i )
!
                    case( 9_i4b )
!
!                       TEST 9: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR ZERO.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = arth( one, one, i )
!
                        eigval(i+1_i4b:n) = tmp*arth( one, one, n-i )
!
                    case( 10_i4b )
!
!                       TEST 10: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR n/2.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = arth( one, one, i )
!
                        eigval(i+1_i4b:n) = real( i, stnd) + tmp*arth( one, one, n-i )
!
                    case( 11_i4b )
!
!                       TEST 11: CLUSTERED DISTRIBUTION OF EIGENVALUES BETWEEN 0.9 and 1.
!
                        call random_number( eigval(:n) )
!
                        eigval(1_i4b:n) = 0.9_stnd + 0.1_stnd*eigval(1_i4b:n)
!
                    case( 12_i4b )
!
!                       TEST 12: EIGENVALUES EQUAL TO sinh(10*i/n) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) )
!
                    case( 13_i4b )
!
!                       TEST 13: EIGENVALUES EQUAL TO sinh((10*i/n)-5) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - five )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case( 14_i4b )
!
!                       TEST 14: EIGENVALUES EQUAL TO sinh((10*i/n)-10) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - ten )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case( 15_i4b )
!
!                       TEST 15: EIGENVALUES EQUAL TO tanh((10*i/n)-5) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = tanh( tmp*arth( one, one, n ) - five )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case default
!
!                       DEFAULT TEST : RANDOMLY UNIFORM DISTRIBUTION OF EIGENVALUES.
!
                        call random_number( eigval(:n) )
!
                        call random_number( tmp )
!
                        if ( tmp>safmin ) then
                             eigval(1_i4b:n) = eigval(1_i4b:n)/tmp
                        end if
!
                end select
!
#ifdef _F2003
                if ( ieee_support_datatype( eigval ) ) then
                    if ( .not.all( ieee_is_normal( eigval(1_i4b:n) ) ) ) cycle
                end if
#endif
!
!               COMPUTE THE SINGULAR VALUES FROM THE EIGENVALUES.
!
                singval(:n) = sqrt( eigval(:n) )
!
!               SORT THE SINGULAR VALUES.
!
                call eigval_sort( sort, singval(:n) )
!
                where ( singval(:n)==zero )
!
                    eigval(:n) = one
!
                elsewhere
!
                    eigval(:n) = singval(:n)
!
                end where
!
!               FORM A MATRIX FROM THE SPECIFIED SINGULAR VALUE DISTRIBUTION AND RANDOM SINGULAR VECTORS.
!
                p(:n,:n) = transpose( rightvec(:n,:n) )
!
                do i = 1_i4b, n
!
                    p(:n,i) = singval(:n)*p(:n,i)
!
                end do
!
                a(:nr,:n) = matmul( leftvec(:nr,:n), p(:n,:n) )
!
!               COMPUTE THE NORM OF THE MATRIX.
!
!                norma = norm( a(:nr,:n) )
                norma = norm( singval(:n) )
!
                do q = 1_i4b, 2_i4b
!
!                   SAVE THE ORIGINAL MATRIX.
!
                    a2(:nr,:n) = a(:nr,:n)
!
!                   REDUCE THE MATRIX a TO BIDIAGONAL FORM AND
!                   COMPUTE ALL SINGULAR VALUES OF THE MATRIX a TO HIGH ACCURACY.
!
!                   DETERMINE IF A TWO-STAGE TRANSFORMATION IS USED FOR REDUCTION.
!
                    two_stage = q==2_i4b
!
!                   REDUCE THE MATRIX a TO BIDIAGONAL FORM AND
!                   COMPUTE ALL SINGULAR VALUES OF THE MATRIX a TO HIGH ACCURACY.
!
                    if ( two_stage ) then
!
                        if ( gen_o ) then
!
                            call select_singval_cmp4( a2, ra, nsing, singval2, failure=failure, sort=sort, &
                                                      d=d, e=e, abstol=abstol, p=p, gen_p=gen_p )
!
                        else
!
                            call select_singval_cmp4( a2, ra, nsing, singval2, failure=failure, sort=sort, &
                                                      d=d, e=e, abstol=abstol, tauo=tauo, p=p, gen_p=gen_p )
!
                        end if
!
                    else
!
                        call select_singval_cmp4( a2, nsing, singval2, failure=failure, sort=sort, &
                                                  d=d, e=e, abstol=abstol, p=p, gen_p=gen_p        )
!
                    end if
!
                    if ( failure ) cycle
!
!                   COMPUTE SINGULAR VECTORS OF THE MATRIX.
!
                    do o = 1_i4b, 2_i4b
!
                        if ( o==1_i4b ) then
                            nsing = min( n, nsing_def )
                        else
                            nsing = n
                        end if
!
                        do i = 1_i4b, 2_i4b
!
!                           COMPUTE SINGULAR VECTORS OF THE MATRIX BY AN INVERSE ITERATION METHOD.
!
                            if ( i==1_i4b ) then
!
                                eps2 = eps
!
                                if ( two_stage ) then
!
                                    if ( gen_o ) then
!
                                        call bd_inviter2( a2(:nr,:n), ra(:n,:n), p(:n,:n), d(:n), e(:n), &
                                                          singval2(:nsing), leftvec2(:nr,:nsing),        &
                                                          rightvec2(:n,:nsing), failure2,                &
                                                          maxiter=maxiter, ortho=true                    )
!
                                    else
!
                                        call bd_inviter2( a2(:nr,:n), ra(:n,:n), p(:n,:n), d(:n), e(:n), &
                                                          singval2(:nsing), leftvec2(:nr,:nsing),        &
                                                          rightvec2(:n,:nsing), failure2,                &
                                                          tauo=tauo(:n), maxiter=maxiter, ortho=true     )
!
                                    end if
!
                                else
!
                                    call bd_inviter2( a2(:nr,:n), p(:n,:n), d(:n), e(:n), singval2(:nsing), &
                                                      leftvec2(:nr,:nsing), rightvec2(:n,:nsing), failure2, &
                                                      maxiter=maxiter, ortho=true                           )
!
                                end if
!
                            else
!
                                eps2 = eps0
!
                                if ( two_stage ) then
!
                                    if ( gen_o ) then
!
                                        call bd_inviter2( a2(:nr,:n), ra(:n,:n), p(:n,:n), d(:n), e(:n), &
                                                          singval2(:nsing), leftvec2(:nr,:nsing),        &
                                                          rightvec2(:n,:nsing), failure2,                &
                                                          maxiter=maxiter                                )
!
                                    else
                                        call bd_inviter2( a2(:nr,:n), ra(:n,:n), p(:n,:n), d(:n), e(:n), &
                                                          singval2(:nsing), leftvec2(:nr,:nsing),        &
                                                          rightvec2(:n,:nsing), failure2,                &
                                                          tauo=tauo(:n), maxiter=maxiter                 )
!
                                    end if
!
                                else
!
                                    call bd_inviter2( a2(:nr,:n), p(:n,:n), d(:n), e(:n), singval2(:nsing), &
                                                      leftvec2(:nr,:nsing), rightvec2(:n,:nsing), failure2, &
                                                      maxiter=maxiter                                       )
!
                                end if
!
                            end if
!
                            if ( failure2 ) then
!
                                notdone = notdone + 1_i4b
!
                                cycle
!
                            end if
!
                            m = m + 1_i4b
!
!                           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*v - u*s
!                           WHERE s ARE THE SINGULAR VALUES, u THE LEFT SINGULAR VECTORS AND
!                           v THE RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                            res(:nr,:nsing) = matmul( a(:nr,:n), rightvec2(:n,:nsing) )         &
                                             - leftvec2(:nr,:nsing)*spread(singval2(:nsing),1,nr)
!
                            if ( norma/=zero ) then
                                err1 = maxval( norm(res(:nr,:nsing), dim=2_i4b) )/( norma*real(nr,stnd) )
                            else
                                err1 = maxval( norm(res(:nr,:nsing), dim=2_i4b) )/real(nr,stnd)
                            end if
!
!                           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!                           WHERE u ARE LEFT SINGULAR VECTORS OF THE MATRIX.
!
                            res(:nsing,:nsing) = matmul( transpose( leftvec2(:nr,:nsing) ), leftvec2(:nr,:nsing) )
!
                            do l = 1_i4b, nsing
                                res(l,l) = one - res(l,l)
                            end do
!
                            err2 = maxval( norm(res(:nsing,:nsing), dim=2_i4b) )/real(nr,stnd)
!
!                           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - v**(t)*v
!                           WHERE v ARE RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                            res(:nsing,:nsing) = matmul( transpose( rightvec2(:n,:nsing) ), rightvec2(:n,:nsing) )
!
                            do l = 1_i4b, nsing
                                res(l,l) = one - res(l,l)
                            end do
!
                            err3 = maxval( norm(res(:nsing,:nsing), dim=2_i4b) )/real(n,stnd)
!
                            if ( err1<=eps .and. err2<=eps0 .and. err3<=eps2 ) then
!
                                success = success + 1_i4b
!
                                if ( details ) then
!
                                    write (prtunit,*) 'Test ',m,' is correct'
!
                                    if ( i==1_i4b ) then
!
                                        write (prtunit,*) 'nr= ', nr,' , nc= ', n,' , nsing= ', nsing,' , case= ', j, &
                                                          ' , two_stage= ', two_stage,' , gen_o= ', gen_o,            &
                                                          ' , gen_p= ', gen_p, ' , ortho= ', true
!
                                    else
!
                                        write (prtunit,*) 'nr= ', nr,' , nc= ', n,' , nsing= ', nsing,' , case= ', j, &
                                                          ' , two_stage= ', two_stage,' , gen_o= ', gen_o,            &
                                                          ' , gen_p= ', gen_p
!
                                    end if
!
!                                   COMPUTE ABSOLUTE ERRORS OF SINGULAR VALUES.
!
                                    write (prtunit,*) 'Maximum absolute error of the computed singular values = ', &
                                                      maxval( abs( singval2(:nsing)-singval(:nsing) ) )
!
!                                   COMPUTE RELATIVE ERRORS OF SINGULAR VALUES.
!
                                    write (prtunit,*) 'Maximum relative error of the computed singular values = ', &
                                                maxval( abs( ( singval2(:nsing)-singval(:nsing) )/eigval(:nsing) ) )
!
!                                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*v - u*s
!                                   WHERE s ARE THE SINGULAR VALUES, u THE LEFT SINGULAR VECTORS AND
!                                   v THE RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                                    write (prtunit,*) 'Accuracy of the computed singular triplets             = ', &
                                                      err1/ulp, ' (must be less than :', fudge, ' )'
!
!                                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!                                   WHERE u ARE LEFT SINGULAR VECTORS OF THE MATRIX.
!
                                    write (prtunit,*) 'Orthogonality of the computed left singular vectors    = ', &
                                                      err2/ulp, ' (must be less than :', fudge, ' )'
!
!                                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - v**(t)*v
!                                   WHERE v ARE RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                                    write (prtunit,*) 'Orthogonality of the computed right singular vectors   = ', &
                                                      err3/ulp, ' (must be less than :', fudge, ' )'
!
                                    write (prtunit,*)
!
                                end if
!
                            else
!
                                write (prtunit,*) 'Test ',m,' is incorrect'
!
                                if ( i==1_i4b ) then
!
                                    err = max( err1, err3 )
!
                                    write (prtunit,*) '=> ',err,' > ',eps,' or ', err2,' > ',eps0,     &
                                                      ' , nr= ', nr,' , nc= ', n,                      &
                                                      ' , nsing= ', nsing,' , case= ', j,              &
                                                      ' , two_stage= ', two_stage,' , gen_o= ', gen_o, &
                                                      ' , gen_p= ', gen_p, ' , ortho= ', true,         &
                                                      ' , failure= ', failure2
!
                                else
!
                                    err = max( err2, err3 )
!
                                    write (prtunit,*) '=> ',err1,' > ',eps,' or ', err,' > ',eps0,     &
                                                      ' , nr= ', nr,' , nc= ', n,' , nsing= ', nsing,  &
                                                      ' , case= ', j, ' , two_stage= ', two_stage,     &
                                                      ' , gen_o= ', gen_o, ' , gen_p= ', gen_p,        &
                                                      ' , failure= ', failure2
!
                                end if
!
!                               COMPUTE ABSOLUTE ERRORS OF SINGULAR VALUES.
!
                                write (prtunit,*) 'Maximum absolute error of the computed singular values = ', &
                                                  maxval( abs( singval2(:nsing)-singval(:nsing) ) )
!
!                               COMPUTE RELATIVE ERRORS OF SINGULAR VALUES.
!
                                write (prtunit,*) 'Maximum relative error of the computed singular values = ', &
                                               maxval( abs( ( singval2(:nsing)-singval(:nsing) )/eigval(:nsing) ) )
!
!                               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*v - u*s
!                               WHERE s ARE THE SINGULAR VALUES, u THE LEFT SINGULAR VECTORS AND
!                               v THE RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                                write (prtunit,*) 'Accuracy of the computed singular triplets             = ', &
                                                  err1/ulp, ' (must be less than :', fudge, ' )'
!
!                               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!                               WHERE u ARE LEFT SINGULAR VECTORS OF THE MATRIX.
!
                                write (prtunit,*) 'Orthogonality of the computed left singular vectors    = ', &
                                              err2/ulp, ' (must be less than :', fudge, ' )'
!
!                               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - v**(t)*v
!                               WHERE v ARE RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                                write (prtunit,*) 'Orthogonality of the computed right singular vectors   = ', &
                                                  err3/ulp, ' (must be less than :', fudge, ' )'
!
                                write (prtunit,*)
!
                            end if
!
                        end do
!
                    end do
!
                end do
!
            end do
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( a, a2, leftvec, leftvec2, res )
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( ra, eigval, singval, singval2, d, e, diagr, &
                    beta, tauo, rightvec, rightvec2, p          )
!
    end do
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_bd_inviter2
! ===============================
!
end program test_bd_inviter2
