program test_bd_svd
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine BD_SVD in module SVD_Procedures.
!                                                                              
!                                                                              
!
! LATEST REVISION : 27/09/2020
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,       &
                         true, zero, one, c50,                         &
                         bd_svd, gen_bd_mat
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def
!
    character(len=*), parameter :: name_proc='bd_svd'
!   
    real(stnd), parameter :: fudge=c50
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                            :: err1, err2, eps, eps1, ulp
    real(stnd), allocatable, dimension(:) :: d, e, singval, temp
!
    integer(i4b) :: k, m, n, success, notdone
    integer      :: iok
!
    logical(lgl) :: failure, upper, details
!   
    character    :: sort='d'
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( err1 )
    eps = fudge*ulp
!
    upper = true
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( d(n), e(n), singval(n), temp(n), stat=iok )
!
        if ( iok/=0 ) then
!
            if ( allocated(d) )         deallocate( d )
            if ( allocated(e) )         deallocate( e )
            if ( allocated(singval) )   deallocate( singval )
            if ( allocated(temp) )      deallocate( temp )
!
            cycle
!
        end if
!
        do k = 1_i4b, 57_i4b
!
!           SKIP GRADED MATRICES WITH A GEOMETRIC PROGRESSION TO AVOID
!           OVERFLOW PROBLEMS IN SINGLE PRECISION.
!
            if ( (k==29_i4b .or. k==31_i4b .or. k==32_i4b .or. k==33_i4b) .and. (n>60_i4b) ) cycle
!
!           GENERATE A BIDIAGONAL MATRIX.
!           THE DIAGONAL ELEMENTS ARE STORED IN d.
!           THE OFF-DIAGONAL ELEMENTS ARE STORED IN e.
!           THE SINGULAR VALUES ARE STORED IN singval.
!
            call gen_bd_mat( k, d(:n), e(:n), singval=singval(:n), sort=sort, failure=failure )
!
#ifdef _F2003
            if ( ieee_support_datatype( d ) ) then
!
                if ( .not.all( ieee_is_normal( d(:n) ) ) ) cycle
                if ( .not.all( ieee_is_normal( e(:n) ) ) ) cycle
!
            end if
#endif
!
            if ( failure ) cycle
!
!           COMPUTE SINGULAR VALUES OF THE BIDIAGONAL MATRIX.
!
            call bd_svd( upper, d(:n), e(:n), failure, sort=sort )
!
            if ( failure ) then
!
                notdone = notdone + 1_i4b
!
                cycle
!
            end if
!
            m = m + 1_i4b
!
!           NOW, TEST ACCURACY OF THE RESULTS.
!
            where( singval(:n)/=zero )
                temp(:n) = singval(:n)
            elsewhere
                temp(:n) = one
            end where
!
!           ABSOLUTE ERRORS OF SINGULAR VALUES.
!
            err1 = maxval( abs( d(:n) - singval(:n) ) )
!
!           RELATIVE ERRORS OF SINGULAR VALUES.
!
            err2 = maxval( abs( (d(:n) - singval(:n))/temp(:n) ) )
!
            eps1 = real( n, stnd)*eps*maxval( singval(:n) )
!
            if ( err1<=eps1 ) then
!
                success = success + 1_i4b
!
                if ( details ) then
!
                    write (prtunit,*) 'Test ',m,' is correct'
                    write (prtunit,*) 'n= ', n,' , case= ', k
!
                    write (prtunit,*) 'Absolute accuracy of the computed singular values    = ', err1
                    write (prtunit,*) 'Relative accuracy of the computed singular values    = ', err2
!
                    write (prtunit,*)
!
                end if
!
            else
!
                write (prtunit,*) 'Test ',m,' is incorrect'
                write (prtunit,*) '=> ',err1,' > ',eps1,' , n= ', n,' , case= ', k,   &
                     ' , failure= ', failure
!
                write (prtunit,*) 'Absolute accuracy of the computed singular values    = ', err1
                write (prtunit,*) 'Relative accuracy of the computed singular values    = ', err2
!
                write (prtunit,*)
!
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( d, e, singval, temp )
!
    end do
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_bd_svd
! ==========================
!
end program test_bd_svd
