program test_chol_cmp2
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine CHOL_CMP2
!   in module Lin_Procedures.
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    Random symmetric matrices with given eigenvalue value distributions are generated
!    for checking the accuracy of subroutine CHOL_CMP2.
!
! LATEST REVISION : 25/05/2021
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,             &
                         true, zero, half, one, two, five, ten, c100,        &
                         minexp, maxexp, machhuge,  machbase,                &      
                         qr_cmp, ortho_gen_qr, chol_cmp2, arth, lamch, norm
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def
!   
    real(stnd), parameter  :: fudge=c100, max_eigval=1000._stnd
!
    real(stnd), parameter :: invmachbase=one/machbase
!
    character(len=*), parameter :: name_proc='chol_cmp2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, err3, eps, eps2, ulp, &
                                               safmin, tmp, shift, d1, dd1, conda,    &
                                               condax, normx, adet, adet2
    real(stnd), allocatable, dimension(:)   :: eigval, diagr, beta, invdiag, x, b
    real(stnd), allocatable, dimension(:,:) :: a, ainv0, ainv, eigvec
!
    integer(i4b) :: i, j, l, n, m, krank0, success, notdone, d2, dd2
    integer      :: iok
!
    logical(lgl) :: details, upper
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    eps    = fudge*ulp
!
    safmin = lamch( 'S' )
!
!   SPECIFY A SHIFT TO AVOID ZERO EIGENVALUE.
!
    shift = half
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
    upper = true
!
    do n = n1, n2, n3
!
        upper = .not.upper
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( eigval(n), diagr(n), beta(n),              &
                  a(n,n), ainv0(n,n), ainv(n,n), invdiag(n), &
                  eigvec(n,n), x(n), b(n), stat=iok          )
!
        if ( iok/=0 ) then
!
            if ( allocated(eigval) )    deallocate( eigval )
            if ( allocated(diagr) )     deallocate( diagr )
            if ( allocated(beta) )      deallocate( beta )
            if ( allocated(a) )         deallocate( a )
            if ( allocated(ainv0) )     deallocate( ainv0 )
            if ( allocated(ainv) )      deallocate( ainv )
            if ( allocated(invdiag) )   deallocate( invdiag )
            if ( allocated(eigvec) )    deallocate( eigvec )
            if ( allocated(x) )         deallocate( x )
            if ( allocated(b) )         deallocate( b )
!
            cycle
!
        end if
!
!       GENERATE A RANDOM COEFFICIENT MATRIX eigvec(:n,:n) .
!
        call random_number( eigvec(:n,:n) )
!
        eigvec(:n,:n) = eigvec(:n,:n) - half
!
        call random_number( tmp )
!
        if ( tmp>safmin ) then
            eigvec(:n,:n) = eigvec(:n,:n)/tmp
        end if
!
!       COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
        call qr_cmp( eigvec(:n,:n), diagr(:n), beta(:n) )
!
!       GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!       q IS ASSUMED OF FULL RANK.
!
        call ortho_gen_qr( eigvec(:n,:n), beta(:n) )
!
!       GENERATE A RANDOM SOLUTION VECTOR x .
!
        call random_number( x(:n) )
!
        call random_number( tmp )
!
        if ( tmp>safmin ) then
            x(:n) = x(:n)/tmp
        end if
!
!       COMPUTE THE NORM OF RANDOM SOLUTION VECTOR x .
!
        normx = norm( x(:n) )
!
!       COMPUTE FIRST THE EIGENVALUES OF THE MATRIX PRODUCT
!                     a(:n,:n)**(t)*a(:n,:n).
!
        do j = 1_i4b, 16_i4b
!
            select case( j )
!
            case( 1_i4b )
!
!               TEST 1: STRONGLY CLUSTERED EIGENVALUES.
!
                eigval(1_i4b)   = one
                eigval(2_i4b:n) = ulp
!
            case( 2_i4b )
!
!               TEST 2:  STRONGLY CLUSTERED EIGENVALUES.
!
                eigval(1_i4b:n-1_i4b) = one
                eigval(n)             = ulp
!
            case( 3_i4b )
!
!               TEST 3: WEAKLY CLUSTERED EIGENVALUES.
!
                eigval(1_i4b:n-1_i4b) = arth( ulp, ulp, n-1_i4b )
                eigval(n)             = one
!
            case( 4_i4b )
!
!               TEST 4: INTERNAL CLUSTER OF EIGENVALUES.
!
                eigval(1_i4b)         = ulp
                eigval(2_i4b:n-1_i4b) = one + sqrt(ulp)*arth( one, one, n-2_i4b )
                eigval(n)             = two
!
            case( 5_i4b )
!
!               TEST 5: WEAKLY CLUSTERED EIGENVALUES.
!
                eigval(1_i4b:n) = one + (c100*ulp)*arth( zero, one, n )
!
            case( 6_i4b )
!
!               TEST 6: GEOMETRIC DISTRIBUTION OF EIGENVALUES.
!
                err = real( n-1_i4b, stnd )
                do l = 1_i4b, n
                    eigval(l) = ulp**(real(l-1_i4b,stnd)/err)
                end do
!
            case( 7_i4b )
!
!               TEST 7: ARITHMETIC DISTRIBUTION OF EIGENVALUES.
!
                eigval(1_i4b:n) = ulp + ((one-ulp)/real( n-1_i4b, stnd ))*arth( zero, one, n )
!
            case( 8_i4b )
!
!               TEST 8: TWO CLUSTERS OF EIGENVALUES.
!
                i = n/2_i4b
!
                tmp = sqrt( ulp )
!
                eigval(1_i4b:i) = one + tmp*arth( one, one, i )
!
                eigval(i+1_i4b:n) = ten + tmp*arth( one, one, n-i )
!
            case( 9_i4b )
!
!               TEST 9: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR ZERO.
!
                i = n/2_i4b
!
                tmp = sqrt( ulp )
!
                eigval(1_i4b:i) = arth( one, one, i )
!
                eigval(i+1_i4b:n) = tmp*arth( one, one, n-i )
!
            case( 10_i4b )
!
!               TEST 10: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR n/2.
!
                i = n/2_i4b
!
                tmp = sqrt( ulp )
!
                eigval(1_i4b:i) = arth( one, one, i )
!
                eigval(i+1_i4b:n) = real( i, stnd) + tmp*arth( one, one, n-i )
!
            case( 11_i4b )
!
!               TEST 11: CLUSTERED DISTRIBUTION OF EIGENVALUES BETWEEN 0.9 and 1.
!
                call random_number( eigval(:n) )
!
                eigval(1_i4b:n) = 0.9_stnd + 0.1_stnd*eigval(1_i4b:n)
!
            case( 12_i4b )
!
!               TEST 12: EIGENVALUES EQUAL TO sinh(10*i/n) FOR i=1 TO n.
!
                tmp = ten/real( n, stnd)
!
                eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) )
!
            case( 13_i4b )
!
!               TEST 13: EIGENVALUES EQUAL TO sinh((10*i/n)-5) FOR i=1 TO n.
!
                tmp = ten/real( n, stnd)
!
                eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - five )
!
!               SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                where( eigval(1_i4b:n)<zero )
                    beta(:n) = -eigval(1_i4b:n)
                elsewhere
                    beta(:n) = zero
                end where
!
                tmp = maxval( beta(:n) ) + one
!
!                tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
            case( 14_i4b )
!
!               TEST 14: EIGENVALUES EQUAL TO sinh((10*i/n)-10) FOR i=1 TO n.
!
                tmp = ten/real( n, stnd)
!
                eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - ten )
!
!               SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                where( eigval(1_i4b:n)<zero )
                    beta(:n) = -eigval(1_i4b:n)
                elsewhere
                    beta(:n) = zero
                end where
!
                tmp = maxval( beta(:n) ) + one
!
!                tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
            case( 15_i4b )
!
!               TEST 15: EIGENVALUES EQUAL TO tanh((10*i/n)-5) FOR i=1 TO n.
!
                tmp = ten/real( n, stnd)
!
                eigval(1_i4b:n) = tanh( tmp*arth( one, one, n ) - five )
!
!               SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                where( eigval(1_i4b:n)<zero )
                    beta(:n) = -eigval(1_i4b:n)
                elsewhere
                    beta(:n) = zero
                end where
!
                tmp = maxval( beta(:n) ) + one
!
!                tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
            case default
!
!               DEFAULT TEST : RANDOMLY UNIFORM DISTRIBUTION OF EIGENVALUES.
!
                call random_number( eigval(:n) )
!
                call random_number( tmp )
!
                if ( tmp>safmin ) then
                    eigval(1_i4b:n) = eigval(1_i4b:n)/tmp
                end if
!
            end select
!
#ifdef _F2003
            if ( ieee_support_datatype( eigval ) ) then
                if ( .not.all( ieee_is_normal( eigval(1_i4b:n) ) ) ) cycle
            end if
#endif
!
            eigval(:n) = eigval(:n) + shift
!
            krank0 = count( eigval(:n)>zero )
!
!           SKIP DEFICIENT MATRICES.
!
            if ( krank0/=n ) cycle
!
!           FORM A MATRIX FROM THE SPECIFIED EIGENVALUE DISTRIBUTION AND RANDOM EIGENVECTORS.
!
            ainv(:n,:n) = transpose( eigvec(:n,:n) )
!
            do i = 1_i4b, n
!
                ainv0(:n,i) = eigvec(:n,i)*eigval(i)
!
            end do
!
            a(:n,:n) = matmul( ainv0(:n,:n), ainv(:n,:n) )
!
!           COMPUTE CONDITION NUMBER OF RANDOM MATRIX.
!
            tmp = maxval( eigval(:n) )
!
            conda = tmp/minval( eigval(:n) )
!
!            conda = ( sum( eigval(:n) )*sum( one/eigval(:n) ) )/real( n, stnd)
!
            tmp = min( tmp, max_eigval )
!
            condax = tmp*normx
!
!           COMPUTE THE MATRIX-VECTOR PRODUCT b = a*x .
!
            b(:n) = matmul( a(:n,:n), x(:n) )
!
!           COMPUTE THE INVERSE OF RANDOM MATRIX a .
!
            beta(:n) = one/eigval(:n)
!
            do i = 1_i4b, n
!
                ainv(:n,i) = ainv(:n,i)*beta(:n)
!
            end do
!
            ainv0(:n,:n) = matmul( eigvec(:n,:n), ainv(:n,:n) )
!
            beta(:n) = norm( ainv0(:n,:n), dim=2_i4b )*tmp
!
!           COMPUTE CHOLESKY FACTORIZATION OF a, DETERMINANT OF a, INVERSE OF a AND SOLVE a*x = b .
!
            call chol_cmp2( a(:n,:n), invdiag(:n), d1, d2=d2, b=b(:n), matinv=ainv(:n,:n),     &
                            upper=upper, fill=true )
!
            if ( d1==zero ) then
!
                notdone = notdone + 1_i4b
!
                cycle
!
            end if
!
!           TEST 1 : CHECK THE SOLUTION VECTOR FOR LINEAR SYSTEM
!
!                               a*x(:n)=b(:n) .
!
            diagr(:n) = b(:n) - x(:n)
!
            err1 = norm( diagr(:n) )/condax
!
!           TEST 2 : CHECK THE INVERSE OF THE COEFFICIENT MATRIX.
!
!                               a*ainv = I .
!
            ainv(:n,:n) = ainv(:n,:n) - ainv0(:n,:n)
!
            diagr(:n) = norm( ainv(:n,:n), dim=2_i4b )
!
            err2 = maxval( diagr(:n)/beta(:n)  )
!
!           TEST 3 : CHECK THE DETERMINANT OF a.
!
!           COMPUTE DETERMINANT OF a .
!
            select case( d2 )
!
                case ( :minexp-1_i4b )
!
                    adet = sign( zero, d1 )
!
                case ( maxexp+1_i4b: )
!
                    adet = sign( machhuge, d1 )
!
                case default
!
!                   adet = scale( d1, d2 )
                    adet = d1*( machbase**d2 )
!
            end select
!
!           RECOMPUTE THE DETERMINANT OF a FROM THE EIGENVALUES.
!
            dd1 = one
            dd2 = zero
!
            do l=1_i4b, n
!
                dd1 = dd1*eigval(l)
!
                do
                    if ( abs(dd1)<= one ) exit
                    dd1 = dd1*invmachbase
                    dd2 = dd2 + 1_i4b
                end do
!
                do
                    if ( abs(dd1)>= invmachbase ) exit
                    dd1 = dd1*machbase
                    dd2 = dd2 - 1_i4b
                end do
!
            end do
!
            select case( dd2 )
!
                case ( :minexp-1_i4b )
!
                    adet2 = sign( zero, dd1 )
!
                case ( maxexp+1_i4b: )
!
                    adet2 = sign( machhuge, dd1 )
!
                case default
!
!                    adet2 = scale( dd1, dd2 )
                    adet2 = dd1*( machbase**dd2 )
!
            end select
!
!           NOW CHECK THE RESULTS.
!
#ifdef _F2003
            if ( ieee_support_datatype( adet2 ) ) then
!
                if ( ieee_is_normal( adet2 ) .and. ieee_is_normal( adet ) ) then
!
                    if ( min( abs(adet), abs(adet2) )==zero ) then
                        tmp = one
                    else
                        tmp = max( abs(adet), abs(adet2) )
                    end if
!
                    err3 = abs(adet - adet2) / tmp
!
                else
!
                    err3 = zero
!
                end if
!
            else
!
#else
                if ( abs( adet2 )<=machhuge .and. abs( adet )<=machhuge  ) then
!
                    if ( min( abs(adet), abs(adet2) )==zero ) then
                        tmp = one
                    else
                        tmp = max( abs(adet), abs(adet2) )
                    end if
!
                    err3 = abs(adet - adet2) / tmp
!
                else
!
                    err3 = zero
!
                end if
#endif
!
#ifdef _F2003
            end if
#endif
!
            m = m + 1_i4b
!
            eps2 = real( n, stnd)*eps
!
            err = max( err1, err2, err3 )
!
            if ( err<=eps2 ) then
!
                success = success + 1_i4b
!
                if ( details ) then
!
                    write (prtunit,*) 'Test ',m,' is correct'
!
                    write (prtunit,*) 'n= ', n,' , case= ', j,' , d1= ', d1, ' , d2= ', d2
!
!                   PRINT CONDITION NUMBER OF COEFFICIENT MATRIX.
!
                    write (prtunit,*) 'Condition number of coefficient matrix              = ', conda
!
!                   PRINT SCALED ERROR OF SOLUTION.
!
                    write (prtunit,*) 'Scaled error of computed solution vector            = ', err1
!
!                   PRINT SCALED ERRORS OF INVERSE.
!
                    write (prtunit,*) 'Maximum scaled error of columns of computed inverse = ', err2
!
!                   PRINT ABSOLUTE ERROR OF COMPUTED DETERMINANT.
!
                    write (prtunit,*) 'Absolute error of computed determinant              = ', abs(adet - adet2)
!
!                   PRINT RELATIVE ERROR OF COMPUTED DETERMINANT.
!
                    write (prtunit,*) 'Relative error of computed determinant              = ', err3
!
                    write (prtunit,*)
!
                end if
!
            else
!
                write (prtunit,*) 'Test ',m,' is incorrect'
!
                write (prtunit,*) '=> ',err,' > ',eps2,' , n= ', n,' , case= ', j,     &
                                  ' , d1= ', d1, ' , d2= ', d2
!
!               PRINT CONDITION NUMBER OF COEFFICIENT MATRIX.
!
                write (prtunit,*) 'Condition number of coefficient matrix              = ', conda
!
!               PRINT RELATIVE ERROR OF SOLUTION.
!
                write (prtunit,*) 'Relative error of computed solution                 = ', err1
!
!               PRINT SCALED ERRORS OF INVERSE.
!
                write (prtunit,*) 'Maximum scaled error of columns of computed inverse = ', err2
!
!               PRINT ABSOLUTE ERROR OF DETERMINANT.
!
                write (prtunit,*) 'Absolute error of computed determinant              = ', abs(adet - adet2)
!
!               PRINT RELATIVE ERROR OF COMPUTED DETERMINANT.
!
                write (prtunit,*) 'Relative error of computed determinant              = ', err3
!
                write (prtunit,*)
!
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( eigval, diagr, beta, invdiag,  &
                    a, ainv0, ainv, eigvec, b, x   )
!
    end do
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_chol_cmp2
! =============================
!
end program test_chol_cmp2
