program test_comp_mca2
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine comp_mca2 in module Mul_Stat_Procedures .
!                                                                              
! LATEST REVISION : 28/08/2020
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, lgl, false, one, comp_mca2
#ifdef _MATMUL
    use Statpack, only : matmul=>matmul2
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, mx=20, my=10, m=min(mx,my), mxy = mx+my, p=100, nsvd=4
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                         :: err_xstat, err_ystat, err_svd, err, eps
    real(stnd), dimension(mx,my)       :: xycor, xycov, xycor2
    real(stnd), dimension(mx,nsvd)     :: res
    real(stnd), dimension(mxy,nsvd)    :: xysingvec
    real(stnd), dimension(mx,2)        :: xstat1, xstat2
    real(stnd), dimension(my,2)        :: ystat1, ystat2
    real(stnd), dimension(m)           :: xysingval
    real(stnd), dimension(mx,p)        :: x, xt
    real(stnd), dimension(my,p)        :: y, yt
!
    integer(i4b) :: i, k, k2, inc, maxiter=4
!
    logical(lgl) :: first, last, cov, failure
!   
!   
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='comp_mca2'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = sqrt( epsilon(eps) )
!
!   GENERATE RANDOM OBSERVATION ARRAY x .
!
    call random_number( x )
!
!   COMPUTE THE MEANS AND STANDARD-DEVIATIONS OF x  .
!
    xstat1(:,1_i4b) = sum( x, dim=2_i4b )/p
    xt = x - spread( xstat1(:,1_i4b), dim=2_i4b, ncopies=p )
    xstat1(:,2_i4b) = sqrt( sum( xt**2, dim=2_i4b )/p )
!
!   GENERATE A RANDOM OBSERVATION ARRAY y .
!
    call random_number( y )
!
!   COMPUTE THE MEANS AND STANDARD-DEVIATIONS OF y  .
!
    ystat1(:,1_i4b) = sum( y, dim=2_i4b )/p
    yt = y - spread( ystat1(:,1_i4b), dim=2_i4b, ncopies=p )
    ystat1(:,2_i4b) = sqrt( sum( yt**2, dim=2_i4b )/p )
!
!   COMPUTE THE COVARIANCE MATRIX BETWEEN x AND y .
!
    xycov = matmul( xt, transpose( yt ) )/p
!
!   COMPUTE THE CORRELATION MATRIX BETWEEN x AND y .
!
    xycor = xycov*spread( one/xstat1(:,2_i4b), dim=2_i4b, ncopies=my )
    xycor = xycor*spread( one/ystat1(:,2_i4b), dim=1_i4b, ncopies=mx )
!
    cov = false
!
    do i = 1_i4b, 4_i4b
!
        cov = .not.cov
        inc = merge( 1_i4b, p, i<=2_i4b )
!
!       COMPUTE SINGULAR VECTORS OF THE COVARIANCE OR CORRELATION MATRIX .
!
        k2 = 0_i4b
!
        do k = 1_i4b, p, inc
!
            k2 = min( k2+inc, p )
!
            first = k==1_i4b
            last = k2==p
!
            call comp_mca2( x(:,k:k2), y(:,k:k2), first, last, xstat2, ystat2, xysingval, xycor2, failure,    &
                           cov=cov, maxiter=maxiter, xysingvec=xysingvec )
        end do
!
        if ( cov ) then
!
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION:
!                xycov*ysingvec(:,:nsvd) - xsingvec(:,:nsvd)*diag(xysingval(:nsvd)).
!
            res(:mx,:nsvd) = matmul(xycov,xysingvec(mx+1:mxy,:nsvd)) - xysingvec(:mx,:nsvd)*spread(xysingval(:nsvd),1,mx)
!
        else
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION:
!                 xycor*ysingvec(:,:nsvd) - xsingvec(:,:nsvd)*diag(xysingval(:nsvd)).
!
            res(:mx,:nsvd) = matmul(xycor,xysingvec(mx+1:mxy,:nsvd)) - xysingvec(:mx,:nsvd)*spread(xysingval(:nsvd),1,mx)
!
        end if
!
        err_svd = sum( abs( res(:mx,:nsvd) ) )/sum( xysingval(:nsvd) )
!
!       CHECK THAT THE TWO SETS OF UNIVARIATE STATISTICS AGREE.
!
        err_xstat  = maxval( abs( (xstat1-xstat2)/xstat2 ) )
        err_ystat  = maxval( abs( (ystat1-ystat2)/ystat2 ) )
!
!       PRINT THE RESULTS OF THE TESTS.
!
        err = max(err_xstat, err_ystat, err_svd )
!
        if ( err<=eps ) then
            write (prtunit,*) 'Test ',i,' is correct'
        else
            write (prtunit,*) 'Test ',i,' is incorrect'
            write (prtunit,*) '=> ',err,' > ',eps
        end if
!
    end do
!
!
! END OF PROGRAM test_comp_mca2
! =============================
!
end program test_comp_mca2
