program test_comp_mvs
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine comp_mvs in module Stat_Procedures.
!                                                                              
! LATEST REVISION : 08/04/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, lgl, comp_mvs, merror, allocate_error, n1_def, n2_def
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, p IS THE NUMBER OF OBSERVATIONS, xmiss IS THE FLAG FOR MISSING VALUES.
!
    integer(i4b), parameter :: prtunit=6, n=n1_def, m=n2_def, p=100
!
    real(stnd), parameter :: xmiss=-999.99_stnd
!
    character(len=*), parameter :: name_proc='comp_mvs'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                                :: err_mean, err_std, err_var, err, eps
    real(stnd), allocatable, dimension(:,:)   :: xmean, xmean1, xmean2, xstd, xstd1,   &
                                                 xstd2, xvar, xvar1, xvar2, xn
    real(stnd), allocatable, dimension(:,:,:) :: x1, x2
!
    integer      :: iok
    integer(i4b) :: i, j, k, l, k2, inc, success
!
    logical(lgl)                                :: first, last
    logical(lgl), allocatable, dimension(:,:,:) :: mask
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = sqrt( epsilon(err_mean) )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( xmean(n,m), xmean1(n,m), xmean2(n,m),        &
              xstd(n,m), xstd1(n,m), xstd2(n,m),           &
              xvar(n,m), xvar1(n,m), xvar2(n,m), xn(n,m),  &
              x1(n,m,p), x2(n,m,p), mask(n,m,p), stat=iok  )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM OBSERVATION ARRAY x1.
!
    call random_number( x1 )
!
!   COMPUTE THE MEANS, VARIANCES AND STANDARD-DEVIATIONS OF x1.
!
    xmean1 = sum( x1, dim=3_i4b )/p
    xvar1 = sum( (x1 - spread( xmean1, dim=3_i4b, ncopies=p ) )**2, dim=3_i4b )/p
    xstd1  = sqrt( xvar1 )
!
!   GENERATE A RANDOM OBSERVATION ARRAY x2 WITH MISSING VALUES.
!
    mask = x1<=0.25_stnd
    where( mask )
       x2 = xmiss
    elsewhere
       x2 = x1
    end where
!
!   COMPUTE THE MEANS, VARIANCES AND STANDARD-DEVIATIONS OF x2.
!
    mask = .not.mask
    xn = count( mask, dim=3_i4b )
    xmean2 = sum( x2, dim=3_i4b, mask=mask )/xn
    xvar2  = sum( (x2 - spread( xmean2, dim=3_i4b, ncopies=p ))**2, dim=3_i4b, mask=mask )/xn
    xstd2  = sqrt( xvar2 )
!
!   RECOMPUTE THE STATISTICS WITH comp_mvs SUBROUTINE.
!
    success = 0_i4b
!
    do i = 1_i4b, 12_i4b
!
        inc = merge( 1_i4b, p, i<=6_i4b )
!
        select case( i )
!
            case( 1_i4b, 7_i4b )
!
!               TESTS 1 AND 7 :
!
!               COMPUTE THE STATISTICS OF x1 WITH comp_mvs_rt.
!
                k2 = 0_i4b
!
                do k = 1_i4b, p, inc
!
                    k2 = min( k2+inc, p )
!
                    first = k==1_i4b
                    last = k2==p
!
                    call comp_mvs( x1(:,:,k:k2), first, last, xmean, xvar, xstd )
!
                end do
!
            case( 2_i4b, 8_i4b )
!
!               TESTS 2 AND 8 :
!
!               COMPUTE THE STATISTICS OF x2 WITH comp_mvs_miss_rt .
!
                k2 = 0_i4b
!
                do k = 1_i4b, p, inc
!
                    k2 = min( k2+inc, p )
!
                    first = k==1_i4b
                    last = k2==p
!
                    call comp_mvs( x2(:,:,k:k2), first, last, xmean, xvar, xstd, xmiss=xmiss )
!
                end do
!
            case( 3_i4b, 9_i4b )
!
!               TESTS 3 AND 9 :
!
!               COMPUTE THE STATISTICS OF x1 WITH comp_mvs_rm .
!
                do j = 1_i4b, m
!
                    k2 = 0_i4b
!
                    do k = 1_i4b, p, inc
!
                        k2 = min( k2+inc, p )
!
                        first = k==1_i4b
                        last = k2==p
!
                        call comp_mvs( x1(:,j,k:k2), first, last, xmean(:,j), xvar(:,j), xstd(:,j) )
!
                    end do
!
                end do
!
            case( 4_i4b, 10_i4b )
!
!               TESTS 4 AND 10 :
!
!               COMPUTE THE STATISTICS OF x2 WITH comp_mvs_miss_rm .
!
                do j = 1_i4b, m
!
                    k2 = 0_i4b
!
                    do k = 1_i4b, p, inc
!
                        k2 = min( k2+inc, p )
!
                        first = k==1_i4b
                        last = k2==p
!
                        call comp_mvs( x2(:,j,k:k2), first, last, xmean(:,j), xvar(:,j), xstd(:,j), xmiss=xmiss )
!
                    end do
!
                end do
!
            case( 5_i4b, 11_i4b )
!
!               TESTS 5 AND 11 :
!
!               COMPUTE THE STATISTICS OF x1 WITH comp_mvs_rv .
!
                do l = 1_i4b, n
!
                    do j = 1_i4b, m
!
                        k2 = 0_i4b
!
                        do k = 1_i4b, p, inc
!
                            k2 = min( k2+inc, p )
!
                            first = k==1_i4b
                            last = k2==p
!
                            call comp_mvs( x1(l,j,k:k2), first, last, xmean(l,j), xvar(l,j), xstd(l,j) )
!
                        end do
!
                    end do
!
                end do
!
            case( 6_i4b, 12_i4b )
!
!               TESTS 6 AND 12 :
!
!               COMPUTE THE STATISTICS OF x2 WITH comp_mvs_miss_rv .
!
                do l = 1_i4b, n
!
                    do j = 1_i4b, m
!
                        k2 = 0_i4b
!
                        do k = 1_i4b, p, inc
!
                            k2 = min( k2+inc, p )
!
                            first = k==1_i4b
                            last = k2==p
!
                            call comp_mvs( x2(l,j,k:k2), first, last, xmean(l,j), xvar(l,j), xstd(l,j), xmiss=xmiss )
!
                        end do
!
                    end do
!
                end do
!
        end select
!
!       CHECK THAT THE TWO SETS OF STATISTICS AGREE.
!
        if ( mod(i,2_i4b)==1_i4b ) then
!
            err_mean = maxval( abs( (xmean-xmean1)/xmean ) )
            err_var  = maxval( abs( (xvar-xvar1)/xvar ) )
            err_std  = maxval( abs( (xstd-xstd1)/xstd ) )
!
        else
!
            err_mean = maxval( abs( (xmean-xmean2)/xmean ) )
            err_var  = maxval( abs( (xvar-xvar2)/xvar ) )
            err_std  = maxval( abs( (xstd-xstd2)/xstd ) )
!
        end if
!
!       PRINT THE RESULTS OF THE TESTS.
!
        err = max( err_mean, err_std, err_var )
!
        if ( err<=eps ) then
!
            success = success + 1_i4b
!
!            write (prtunit,*) 'Test ',i,' is correct'
!
        else
!
            write (prtunit,*) 'Test ',i,' is incorrect'
            write (prtunit,*) '=> ', err,' > ', eps
!
            write (prtunit,*)
!
        end if
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( xmean, xmean1, xmean2, xstd, xstd1, xstd2,  &
                xvar, xvar1, xvar2, xn, x1, x2, mask        )
!
    if ( success/=12_i4b ) then
        write (prtunit,*) 12_i4b - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_comp_mvs
! ============================
!
end program test_comp_mvs
