program test_comp_sym_ginv
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine COMP_SYM_GINV
!   in module Lin_Procedures.
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    Random positive symmetric matrices with given eigenvalue distributions are generated for
!    checking the accuracy of subroutine COMP_SYM_GINV.
!
! LATEST REVISION : 30/09/2020
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,                  &
                         true, false, zero, half, one, two, five, ten, c100,      &
                         qr_cmp, ortho_gen_qr, comp_sym_ginv, arth, lamch, norm
!
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def
!   
    real(stnd), parameter  :: fudge=c100, max_eigval=1000._stnd
!
    character(len=*), parameter :: name_proc='comp_sym_ginv'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, eps, eps2, ulp, safmin,   &
                                               tmp, shift0, shift, tol, conda
    real(stnd), allocatable, dimension(:)   :: eigval, diagr, beta
    real(stnd), allocatable, dimension(:,:) :: a, a2, ainv0, ainv, eigvec
!
    integer(i4b) :: i, j, l, n, m, krank, krank0, success, notdone
    integer      :: iok
!
    logical(lgl) :: failure, in_place, upper, fill, details
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    eps    = fudge*ulp
!
    safmin = lamch( 'S' )
!
!   SPECIFY EVENTUALLY A SHIFT TO AVOID ZERO EIGENVALUE.
!
    shift0 = zero
    shift  = 0.1_stnd
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
    fill = true
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( eigval(n), diagr(n), beta(n), eigvec(n,n), &
                  a(n,n), a2(n,n), ainv0(n,n), ainv(n,n),    &
                  stat=iok  )
!
        if ( iok/=0 ) then
!
            if ( allocated(eigval) )   deallocate( eigval )
            if ( allocated(diagr) )    deallocate( diagr )
            if ( allocated(beta) )     deallocate( beta )
            if ( allocated(a) )        deallocate( a )
            if ( allocated(a2) )       deallocate( a2 )
            if ( allocated(ainv0) )    deallocate( ainv0 )
            if ( allocated(ainv) )     deallocate( ainv )
            if ( allocated(eigvec) )   deallocate( eigvec )
!
            cycle
!
        end if
!
!       GENERATE A RANDOM COEFFICIENT MATRIX eigvec(:n,:n) .
!
        call random_number( eigvec(:n,:n) )
!
        eigvec(:n,:n) = eigvec(:n,:n) - half
!
        call random_number( tmp )
!
        if ( tmp>safmin ) then
            eigvec(:n,:n) = eigvec(:n,:n)/tmp
        end if
!
!       COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
        call qr_cmp( eigvec(:n,:n), diagr(:n), beta(:n) )
!
!       GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!       q IS ASSUMED OF FULL RANK.
!
        call ortho_gen_qr( eigvec(:n,:n), beta(:n) )
!
!       COMPUTE FIRST THE EIGENVALUES OF THE MATRIX.
!
        do j = 1_i4b, 16_i4b
!
            select case( j )
!
            case( 1_i4b )
!
!               TEST 1: STRONGLY CLUSTERED EIGENVALUES.
!
                eigval(1_i4b)   = one
                eigval(2_i4b:n) = ulp
!
            case( 2_i4b )
!
!               TEST 2:  STRONGLY CLUSTERED EIGENVALUES.
!
                eigval(1_i4b:n-1_i4b) = one
                eigval(n)             = ulp
!
            case( 3_i4b )
!
!               TEST 3: WEAKLY CLUSTERED EIGENVALUES.
!
                eigval(1_i4b:n-1_i4b) = arth( ulp, ulp, n-1_i4b )
                eigval(n)             = one
!
            case( 4_i4b )
!
!               TEST 4: INTERNAL CLUSTER OF EIGENVALUES.
!
                eigval(1_i4b)         = ulp
                eigval(2_i4b:n-1_i4b) = one + sqrt(ulp)*arth( one, one, n-2_i4b )
                eigval(n)             = two
!
            case( 5_i4b )
!
!               TEST 5: WEAKLY CLUSTERED EIGENVALUES.
!
                eigval(1_i4b:n) = one + (c100*ulp)*arth( zero, one, n )
!
            case( 6_i4b )
!
!               TEST 6: GEOMETRIC DISTRIBUTION OF EIGENVALUES.
!
                err = real( n-1_i4b, stnd )
                do l = 1_i4b, n
                    eigval(l) = ulp**(real(l-1_i4b,stnd)/err)
                end do
!
            case( 7_i4b )
!
!               TEST 7: ARITHMETIC DISTRIBUTION OF EIGENVALUES.
!
                eigval(1_i4b:n) = ulp + ((one-ulp)/real( n-1_i4b, stnd ))*arth( zero, one, n )
!
            case( 8_i4b )
!
!               TEST 8: TWO CLUSTERS OF EIGENVALUES.
!
                i = n/2_i4b
!
                tmp = sqrt( ulp )
!
                eigval(1_i4b:i) = one + tmp*arth( one, one, i )
!
                eigval(i+1_i4b:n) = ten + tmp*arth( one, one, n-i )
!
            case( 9_i4b )
!
!               TEST 9: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR ZERO.
!
                i = n/2_i4b
!
                tmp = sqrt( ulp )
!
                eigval(1_i4b:i) = arth( one, one, i )
!
                eigval(i+1_i4b:n) = tmp*arth( one, one, n-i )
!
            case( 10_i4b )
!
!               TEST 10: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR n/2.
!
                i = n/2_i4b
!
                tmp = sqrt( ulp )
!
                eigval(1_i4b:i) = arth( one, one, i )
!
                eigval(i+1_i4b:n) = real( i, stnd) + tmp*arth( one, one, n-i )
!
            case( 11_i4b )
!
!               TEST 11: CLUSTERED DISTRIBUTION OF EIGENVALUES BETWEEN 0.9 and 1.
!
                call random_number( eigval(:n) )
!
                eigval(1_i4b:n) = 0.9_stnd + 0.1_stnd*eigval(1_i4b:n)
!
            case( 12_i4b )
!
!               TEST 12: EIGENVALUES EQUAL TO sinh(10*i/n) FOR i=1 TO n.
!
                tmp = ten/real( n, stnd)
!
                eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) )
!
            case( 13_i4b )
!
!               TEST 13: EIGENVALUES EQUAL TO sinh((10*i/n)-5) FOR i=1 TO n.
!
                tmp = ten/real( n, stnd)
!
                eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - five )
!
!               SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                where( eigval(1_i4b:n)<zero )
                    beta(:n) = -eigval(1_i4b:n)
                elsewhere
                    beta(:n) = zero
                end where
!
                tmp = maxval( beta(:n) ) + sqrt( ulp )
!
!                tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
            case( 14_i4b )
!
!               TEST 14: EIGENVALUES EQUAL TO sinh((10*i/n)-10) FOR i=1 TO n.
!
                tmp = ten/real( n, stnd)
!
                eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - ten )
!
!               SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                where( eigval(1_i4b:n)<zero )
                    beta(:n) = -eigval(1_i4b:n)
                elsewhere
                    beta(:n) = zero
                end where
!
                tmp = maxval( beta(:n) ) + sqrt( ulp )
!
!                tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
            case( 15_i4b )
!
!               TEST 15: EIGENVALUES EQUAL TO tanh((10*i/n)-5) FOR i=1 TO n.
!
                tmp = ten/real( n, stnd)
!
                eigval(1_i4b:n) = tanh( tmp*arth( one, one, n ) - five )
!
!               SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                where( eigval(1_i4b:n)<zero )
                    beta(:n) = -eigval(1_i4b:n)
                elsewhere
                    beta(:n) = zero
                end where
!
                tmp = maxval( beta(:n) ) + sqrt( ulp )
!
!                tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
            case default
!
!               DEFAULT TEST : RANDOMLY UNIFORM DISTRIBUTION OF EIGENVALUES.
!
                call random_number( eigval(:n) )
!
                call random_number( tmp )
!
                if ( tmp>safmin ) then
                    eigval(1_i4b:n) = eigval(1_i4b:n)/tmp
                end if
!
            end select
!
#ifdef _F2003
            if ( ieee_support_datatype( eigval ) ) then
                if ( .not.all( ieee_is_normal( eigval(1_i4b:n) ) ) ) cycle
            end if
#endif
!
            tol = real( n, stnd)*sqrt( ulp )
!
!           SHIFT EVENTUALLY THE EIGENVALUES.
!
            eigval(:n) = eigval(:n) + shift0
!
!            where( abs(eigval(:n))<=tol ) eigval(:n) = ulp
!
!            krank0 = count( eigval(:n)> ulp )
!
!
!            where( eigval(:n)<=tol ) eigval(:n) = zero
!
!            krank0 = count( eigval(:n)/=zero )
!
!
            where( eigval(:n)<=tol )
                eigval(:n) = ulp
            elsewhere
                eigval(:n) = eigval(:n) + shift
            end where
!
            krank0 = count( eigval(:n)>ulp )
!
!           SKIP ZERO MATRICES.
!
            if ( krank0==0_i4b ) cycle
!
!           FORM A MATRIX FROM THE SPECIFIED EIGENVALUE DISTRIBUTION AND RANDOM EIGENVECTORS.
!
            ainv(:n,:n) = transpose( eigvec(:n,:n) )
!
            do i = 1_i4b, n
!
                ainv(:n,i) = eigval(:n)*ainv(:n,i)
!
            end do
!
            a(:n,:n) = matmul( eigvec(:n,:n), ainv(:n,:n) )
!
!           SAVE THE COEFFICIENT MATRIX FOR LATER USE.
!
            a2(:n,:n) = a(:n,:n)
!
!           COMPUTE CONDITION NUMBER OF RANDOM MATRIX.
!
            tmp = maxval( eigval(:n) )
!
            if ( krank0==n ) then
!
                conda = tmp/minval( eigval(:n) )
!
!                conda = ( sum( eigval(:n) )*sum( one/eigval(:n) ) )/real( n, stnd)
!
!               COMPUTE THE INVERSE OF RANDOM MATRIX a .
!
                beta(:n) = one/eigval(:n)
!
                ainv(:n,:n) = transpose( eigvec(:n,:n) )
!
                do i = 1_i4b, n
!
                    ainv(:n,i) = beta(:n)*ainv(:n,i)
!
                end do
!
                ainv0(:n,:n) = matmul( eigvec(:n,:n), ainv(:n,:n) )
!
                tmp = min( tmp, max_eigval )
!
                beta(:n) = norm( ainv0(:n,:n), dim=2_i4b )*tmp
!
            else
!
                beta(:n) = eigval(:n)
                where( beta(:n)==zero ) beta(:n) = tmp
!
                conda = tmp/minval( beta(:n) )             
!
                ainv0(:n,:n) = zero
!
            end if
!
            upper = true
!
            tol = real( n, stnd)*ulp
!
            do i = 1_i4b, 4_i4b
!
                upper = .not.upper
!
                select case( i )
!
                    case( 1_i4b, 2_i4b )
!
!                       TESTS 1 AND 2: IN-PLACE INVERSION.
!
                        ainv(:n,:n) = a(:n,:n)
!
                        call comp_sym_ginv( ainv(:n,:n), failure, krank, upper=upper, fill=fill, tol=tol )
!
                        in_place = true
!
                    case( 3_i4b, 4_i4b )
!
!                       TESTS 3 AND 4 : OUT-OF-PLACE INVERSION.
!
                        call comp_sym_ginv( a(:n,:n), failure, krank, ainv(:n,:n), upper=upper, &
                                            fill=fill, tol=tol )
!
                        in_place = false
!
                end select
!
                if ( failure ) then
!
                    notdone = notdone + 1_i4b
!
                    cycle
!
                end if
!
!               TEST : CHECK INVERSE OR GENERALIZED INVERSE OF a.
!
                if ( krank0==n .and. krank==n ) then
!                if ( krank0==n ) then
!
                    ainv(:n,:n) = ainv(:n,:n) - ainv0(:n,:n)
!
                    diagr(:n) = norm( ainv(:n,:n), dim=2_i4b )
!
                    err1 = maxval( diagr(:n)/beta(:n)  )
!
                    err2 = zero
!
                else if ( krank0==krank ) then
!
                    tmp = max( norm( a2 ), norm( ainv ) )
!
                    ainv0(:n,:n) = matmul( a2, matmul(ainv,a2) ) - a2
!
                    err1 = norm( ainv0 ) / tmp
!
                    ainv0(:n,:n) = matmul( ainv, matmul(a2,ainv) ) - ainv
!
                    err2 = norm( ainv0 ) / tmp
!
                else
!
!                    err1 = zero
!                    err2 = zero
!
                    notdone = notdone + 1_i4b
!
                    cycle
!
                end if
!
                m = m + 1_i4b
!
                err = max( err1, err2 )
!
                eps2 = real( n, stnd)*eps
!
                if ( err<=eps2 ) then
!
                    success = success + 1_i4b
!
                    if ( details ) then
!
                        write (prtunit,*) 'Test ',m,' is correct'
!
                        write (prtunit,*) 'n= ', n, ' , case= ', j, ' , krank0= ', krank0, ' , krank= ', krank,  &
                                          ' , in_place= ', in_place, ' , upper= ', upper
!
!                       PRINT CONDITION NUMBER OF COEFFICIENT MATRIX.
!
                        write (prtunit,*) 'Condition number of coefficient matrix              = ', conda
!
                        if ( krank0==krank ) then
!
!                           PRINT SCALED ERRORS OF INVERSE OR GENERALIZED INVERSE.
!
                            if ( krank0==n ) then
                                write (prtunit,*) 'Maximum scaled error of columns of computed inverse = ', err
                            else
                                write (prtunit,*) 'Relative error of computed (generalized) inverse    = ', err
                            end if
!
                        end if
!
                        write (prtunit,*)
!
                    end if
!
                else
!
                    write (prtunit,*) 'Test ',m,' is incorrect'
!
                    write (prtunit,*) '=> ', err,' > ', tol, ' , n= ', n, ' , case= ', j,    &
                                      ' , krank0= ', krank0, ' , krank= ', krank,            &
                                      ' , in_place= ', in_place, ' , upper= ', upper
!
!                   PRINT CONDITION NUMBER OF COEFFICIENT MATRIX.
!
                    write (prtunit,*) 'Condition number of coefficient matrix              = ', conda
!
                    if ( krank0==krank ) then
!
!                       PRINT SCALED ERRORS OF INVERSE OR GENERALIZED INVERSE.
!
                        if ( krank0==n ) then
                            write (prtunit,*) 'Maximum scaled error of columns of computed inverse = ', err
                        else
                            write (prtunit,*) 'Relative error of computed (generalized) inverse    = ', err
                        end if
!
                    end if
!
                    write (prtunit,*)
!
                end if
!
            end do
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( eigval, diagr, beta,        &
                    a, a2, ainv0, ainv, eigvec  )
!
    end do
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_comp_sym_ginv
! =================================
!
end program test_comp_sym_ginv
