program test_comp_triang_inv
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine COMP_TRIANG_INV in module Lin_Procedures. 
!                                                                              
! LATEST REVISION : 28/09/2020
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,     &
                         false, zero, one, two, triangle, is_nan,    &
                         comp_triang_inv, norm, lamch, singvalues,   &
                         merror, allocate_error
!   
!
! STRONG TYPING IMPOSED
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def
!   
    real(stnd), parameter  :: fudge=one, max_conda=1000._stnd
!
    character(len=*), parameter :: name_proc='comp_triang_inv'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, ulp, eps, eps2, safmin, &
                                               conda, condar, tmp
    real(stnd), allocatable, dimension(:)   :: ap, singval
    real(stnd), allocatable, dimension(:,:) :: a, ainv, res
!
    integer(i4b) :: i, j, l, m, n, krank0, success
    integer      :: iok
!
    logical(lgl) :: upper, details
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   ALLOCATE WORK ARRAYS.
!
    n = max( n1, n2 )
!
    l = n*(n+1_i4b)/2_i4b
!
    allocate( a(n,n), ainv(n,n), res(n,n), ap(l), &
              singval(n), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    eps    = fudge*ulp
    safmin = lamch( 'S' )
!
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
        l = n*(n+1_i4b)/2_i4b
!
!       GENERATE TRIANGULAR MATRIX IN PACKED FORM ap .
!
        call random_number( ap(:l) )
!
!       MAKE SURE THAT TRIANGULAR MATRIX IS NOT SINGULAR.
!
        ap(:l) = ap(:l) + two
!
        call random_number( tmp )
!
        if ( tmp>safmin ) then
            ap(:l) = ap(:l)/tmp
        end if
!
        upper = false
!
        do i = 1_i4b, 2_i4b
!
!           UNPACK TRIANGULAR MATRIX a .
!
            upper = .not.upper
!
            a(:n,:n) = unpack( ap(:l), mask=triangle(upper,n,n,extra=1_i4b), field=zero )
!
!           COMPUTE THE CONDITION NUMBER OF MATRIX a.
!
            singval(:n) = singvalues( a(:n,:n) )
!
            if ( is_nan( singval(:n) ) ) cycle
!
            krank0 = count( singval(:n)/=zero )
!
!           SKIP DEFICIENT MATRICES.
!
            if ( krank0/=n ) cycle
!
!           COMPUTE CONDITION NUMBER OF RANDOM MATRIX.
!
            conda = maxval( singval(:n) )/minval( singval(:n) )
!
            tmp =  min( conda, max_conda )
!
            condar = one + two*tmp
!
            do j = 1_i4b, 2_i4b
!
                m = m + 1_i4b
!
                if ( j==1_i4b ) then
!
!                   COMPUTE THE MATRIX INVERSE OF a WITH SUBROUTINE comp_triang_inv.
!                   INPUT ARGUMENT OVERWRITTEN.
!
                    ainv(:n,:n) = a(:n,:n)
                    call comp_triang_inv( ainv(:n,:n), upper=upper )
!
                else
!
!                   COMPUTE THE MATRIX INVERSE OF a WITH SUBROUTINE comp_triang_inv.
!                   INPUT ARGUMENT NOT OVERWRITTEN.
!
                    ainv(:n,:n) = zero
                    call comp_triang_inv( a(:n,:n), ainv(:n,:n), upper=upper )
!
                end if
!
!               COMPUTE a TIMES INVERSE - IDENTITY.
!
                res(:n,:n) = matmul( a(:n,:n), ainv(:n,:n) )
!
                do l = 1_i4b, n
                    res(l,l) = res(l,l) - one
                end do
!
                err1 = maxval( norm(res(:n,:n), dim=2_i4b ) )
!
!               COMPUTE INVERSE TIMES a - IDENTITY.
!
                res(:n,:n) = matmul( ainv(:n,:n), a(:n,:n) )
!
                do l = 1_i4b, n
                    res(l,l) = res(l,l) - one
                end do
!
                err2 = maxval( norm(res(:n,:n), dim=2_i4b ) )
!
!               CHECK THE RESULTS FOR SMALL RESIDUALS.
!
                err = max( err1, err2 )/condar
!
                eps2 = real( n, stnd)*eps
!
                if ( err<=eps2 ) then
!
                    success = success + 1_i4b
!
                    if ( details ) then
!
                        write (prtunit,*) 'Test ',m,' is correct'
!
                        write (prtunit,*) 'n= ', n, ' , upper= ', upper, ' , case= ', j
!
!                       PRINT CONDITION NUMBER OF COEFFICIENT MATRIX.
!
                        write (prtunit,*) 'Condition number of the triangular matrix       = ', conda
!
!                       PRINT ABSOLUTE ERRORS OF SOLUTIONS.
!
                        write (prtunit,*) 'Maximum error of computed columns of a*ainv - I = ', err1
!
                        write (prtunit,*) 'Maximum error of computed columns of ainv*a - I = ', err2
!
                        write (prtunit,*)
!
                    end if
!
                else
!
                    write (prtunit,*) 'Test ',m,' is incorrect'
!
                    write (prtunit,*) '=> ',err, ' > ', eps2, ' , n= ', n, ' , upper= ', upper,   &
                                      ' , case= ', j
!
!                   PRINT CONDITION NUMBER OF COEFFICIENT MATRIX.
!
                    write (prtunit,*) 'Condition number of the triangular matrix       = ', conda
!
!                   PRINT ABSOLUTE ERRORS OF SOLUTIONS.
!
                    write (prtunit,*) 'Maximum error of computed columns of a*ainv - I = ', err1
!
                    write (prtunit,*) 'Maximum error of computed columns of ainv*a - I = ', err2
!
                    write (prtunit,*)
!
                end if
!
            end do
!
        end do
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( a, ainv, res, ap, singval )
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_comp_triang_inv
! ===================================
!
end program test_comp_triang_inv
