program test_comp_unistat
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine comp_unistat in module Stat_Procedures.
!                                                                              
! LATEST REVISION : 05/11/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, lgl, half, one, three, c10, c100, comp_unistat, merror,   &
                         allocate_error, lamch, n1_def, n2_def
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT, p IS THE NUMBER OF OBSERVATIONS, xmiss IS THE FLAG FOR
! MISSING VALUES.
!
    integer(i4b), parameter :: prtunit=6, n=max(n1_def,n2_def), m=min(n1_def,n2_def), p=200
!
    real(stnd), parameter :: xmiss=-999.99_stnd
!   
    real(stnd), parameter  :: fudge=c10
!
    character(len=*), parameter :: name_proc='comp_unistat'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                                :: err, eps, safmin, tmp
    real(stnd), dimension(7)                  :: err_stat
    real(stnd), allocatable, dimension(:,:)   :: xn
    real(stnd), allocatable, dimension(:,:,:) :: xstat, xstat1, xstat2, x1, x2, tmp1, tmp2
!
    integer      :: iok
    integer(i4b) :: i, j, k, l, k2, inc, success
!
    logical(lgl)                                :: first, last, details
    logical(lgl), allocatable, dimension(:,:,:) :: mask
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TESTS.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = fudge*sqrt( epsilon(eps) )
    safmin = lamch( 'S' )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( xn(n,m), xstat(n,m,7), xstat1(n,m,7), xstat2(n,m,7),  &
              x1(n,m,p), x2(n,m,p), tmp1(n,m,p), tmp2(n,m,p),       &
              mask(n,m,p), stat=iok  )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A RANDOM OBSERVATION ARRAY x1.
!
    call random_number( x1 )
!
    x1(:n,:m,:p) = x1(:n,:m,:p) - half
!
    call random_number( tmp )
!
    if ( tmp>safmin ) then
        x1(:n,:m,:p) = x1(:n,:m,:p)/tmp
    end if
!
!   SHIFT THE MEANS OF THE DISTRIBUTIONS.
!
    call random_number( xn )
!
    xn(:n,:m) = c100*xn(:n,:m)
    x1(:n,:m,:p) = x1(:n,:m,:p) + spread( xn(:n,:m), dim=3_i4b, ncopies=p )
!
!   COMPUTE THE STATISTICS OF x1.
!
    xstat1(:,:,1_i4b) = sum( x1, dim=3_i4b )/real( p, stnd )
!
    tmp1 = x1 - spread( xstat1(:,:,1_i4b), dim=3_i4b, ncopies=p )
    tmp2 = tmp1*tmp1
    xstat1(:,:,2_i4b) = sum( tmp2, dim=3_i4b )/real( p - 1_i4b, stnd )
    xstat1(:,:,3_i4b)  = sqrt( xstat1(:,:,2_i4b) )
!
    tmp2 = tmp2*tmp1
    xstat1(:,:,4_i4b)  = sum( tmp2, dim=3_i4b )/( real(p,stnd)*xstat1(:,:,3_i4b)**3 )
!
    tmp2 = tmp2*tmp1
    xstat1(:,:,5_i4b)  = sum( tmp2, dim=3_i4b )/( real(p,stnd)*xstat1(:,:,2_i4b)**2 ) - three
!
    xstat1(:,:,6_i4b)  = minval( x1, dim=3_i4b )
    xstat1(:,:,7_i4b)  = maxval( x1, dim=3_i4b )
!
!   GENERATE A RANDOM OBSERVATION ARRAY x2 WITH MISSING VALUES FROM x1.
!
    call random_number( x2 )
!
    mask = x2<=0.10_stnd
!
    where( mask )
       x2 = xmiss
    elsewhere
       x2 = x1
    end where
!
    mask = .not.mask
    xn = count( mask, dim=3_i4b )
!
!   COMPUTE THE STATISTICS OF x2.
!
    xstat2(:,:,1_i4b) = sum( x2, dim=3_i4b, mask=mask )/xn
!
    tmp1 = x2 - spread( xstat2(:,:,1_i4b), dim=3_i4b, ncopies=p )
    tmp2 = tmp1*tmp1

    xstat2(:,:,2_i4b) = sum( tmp2, dim=3_i4b, mask=mask )/( xn - one )
    xstat2(:,:,3_i4b) = sqrt( xstat2(:,:,2_i4b) )
!
    tmp2 = tmp2*tmp1
    xstat2(:,:,4_i4b) = sum( tmp2, dim=3_i4b, mask=mask )/( xn*(xstat2(:,:,3_i4b)**3) )
!
    tmp2 = tmp2*tmp1
    xstat2(:,:,5_i4b) = sum( tmp2, dim=3_i4b, mask=mask )/( xn*(xstat2(:,:,2_i4b)**2) ) - three
!
    xstat2(:,:,6_i4b)  = minval( x2, dim=3_i4b, mask=mask )
    xstat2(:,:,7_i4b)  = maxval( x2, dim=3_i4b, mask=mask )
!
!   RECOMPUTE THE STATISTICS OF x1 AND x2 WITH comp_unistat SUBROUTINE.
!
    success = 0_i4b
!
    do i = 1_i4b, 12_i4b
!
        inc = merge( 1_i4b, p, i<=6_i4b )
!
        select case( i )
!
            case( 1_i4b, 7_i4b )
!
!               TESTS 1 AND 7 :
!
!               COMPUTE THE STATISTICS OF x1 WITH comp_unistat_rt.
!
                k2 = 0_i4b
!
                do k = 1_i4b, p, inc
!
                    k2 = min( k2+inc, p )
!
                    first = k==1_i4b
                    last = k2==p
!
                    call comp_unistat( x1(:,:,k:k2), first, last, xstat )
!
                end do
!
            case( 2_i4b, 8_i4b )
!
!               TESTS 2 AND 8 :
!
!               COMPUTE THE STATISTICS OF x2 WITH comp_unistat_miss_rt.
!
                k2 = 0_i4b
!
                do k = 1_i4b, p, inc
!
                    k2 = min( k2+inc, p )
!
                    first = k==1_i4b
                    last = k2==p
!
                    call comp_unistat( x2(:,:,k:k2), first, last, xstat, xmiss=xmiss )
!
                end do
!
            case( 3_i4b, 9_i4b )
!
!               TESTS 3 AND 9 :
!
!               COMPUTE THE STATISTICS OF x1 WITH comp_unistat_rm.
!
                do j = 1_i4b, m
!
                    k2 = 0_i4b
!
                    do k = 1_i4b, p, inc
!
                        k2 = min( k2+inc, p )
!
                        first = k==1_i4b
                        last = k2==p
!
                        call comp_unistat( x1(:,j,k:k2), first, last, xstat(:,j,:) )
!
                    end do
!
                end do
!
            case( 4_i4b, 10_i4b )
!
!               TESTS 4 AND 10 :
!
!               COMPUTE THE STATISTICS OF x2 WITH comp_unistat_miss_rm.
!
                do j = 1_i4b, m
!
                    k2 = 0_i4b
!
                    do k = 1_i4b, p, inc
!
                        k2 = min( k2+inc, p )
!
                        first = k==1_i4b
                        last = k2==p
!
                        call comp_unistat( x2(:,j,k:k2), first, last, xstat(:,j,:), xmiss=xmiss )
!
                    end do
!
                end do
!
            case( 5_i4b, 11_i4b )
!
!               TESTS 5 AND 11 :
!
!               COMPUTE THE STATISTICS OF x1 WITH comp_unistat_rv.
!
                do l = 1_i4b, n
!
                    do j = 1_i4b, m
!
                        k2 = 0_i4b
!
                        do k = 1_i4b, p, inc
!
                            k2 = min( k2+inc, p )
!
                            first = k==1_i4b
                            last = k2==p
!
                            call comp_unistat( x1(l,j,k:k2), first, last, xstat(l,j,:) )
!
                        end do
!
                    end do
!
                end do
!
            case( 6_i4b, 12_i4b )
!
!               TESTS 6 AND 12 :
!
!               COMPUTE THE STATISTICS OF x2 WITH comp_unistat_miss_rv.
!
                do l = 1_i4b, n
!
                    do j = 1_i4b, m
!
                        k2 = 0_i4b
!
                        do k = 1_i4b, p, inc
!
                            k2 = min( k2+inc, p )
!
                            first = k==1_i4b
                            last = k2==p
!
                            call comp_unistat( x2(l,j,k:k2), first, last, xstat(l,j,:), xmiss=xmiss )
!
                        end do
!
                    end do
!
                end do
!
        end select
!
!       CHECK THAT THE TWO SETS OF STATISTICS AGREE.
!
        if ( mod(i,2_i4b)==1_i4b ) then
!
            do j = 1_i4b, 7_i4b
!
                if ( j<=3_i4b ) then
                    err_stat(j) = maxval( abs( (xstat(:,:,j)-xstat1(:,:,j))/xstat(:,:,j) ) )
                else
                    err_stat(j) = maxval( abs( xstat(:,:,j)-xstat1(:,:,j) ) )
                end if
!
            end do
!
        else
!
            do j = 1_i4b, 7_i4b
!
                if ( j<=3_i4b ) then
                    err_stat(j) = maxval( abs( (xstat(:,:,j)-xstat2(:,:,j))/xstat(:,:,j) ) )
                else
                    err_stat(j) = maxval( abs( xstat(:,:,j)-xstat2(:,:,j) ) )
                end if
!
            end do
!
        end if
!
!       PRINT THE RESULTS OF THE TESTS.
!
        err = maxval( err_stat(:7_i4b) )
!
        if ( err<=eps ) then
!
            success = success + 1_i4b
!
            if ( details ) then
!
                write (prtunit,*) 'Test ',i,' is correct'
!
!               PRINT ABSOLUTE OR RELATIVE ERRORS OF STATISTICS.
!
                write (prtunit,*) 'Relative error of the computed means               = ', err_stat(1_i4b)
                write (prtunit,*) 'Relative error of the computed variances           = ', err_stat(2_i4b)
                write (prtunit,*) 'Relative error of the computed standard-deviations = ', err_stat(3_i4b)
                write (prtunit,*) 'Absolute error of the computed skewness            = ', err_stat(4_i4b)
                write (prtunit,*) 'Absolute error of the computed kurtosis            = ', err_stat(5_i4b)
                write (prtunit,*) 'Absolute error of the computed minima              = ', err_stat(6_i4b)
                write (prtunit,*) 'Absolute error of the computed maxima              = ', err_stat(7_i4b)
!
                write (prtunit,*)
!
            end if
!
        else
!
            write (prtunit,*) 'Test ',i,' is incorrect'
            write (prtunit,*) '=> ', err,' > ', eps
!
!           PRINT ABSOLUTE OR RELATIVE ERRORS OF STATISTICS.
!
            write (prtunit,*) 'Relative error of the computed means               = ', err_stat(1_i4b)
            write (prtunit,*) 'Relative error of the computed variances           = ', err_stat(2_i4b)
            write (prtunit,*) 'Relative error of the computed standard-deviations = ', err_stat(3_i4b)
            write (prtunit,*) 'Absolute error of the computed skewness            = ', err_stat(4_i4b)
            write (prtunit,*) 'Absolute error of the computed kurtosis            = ', err_stat(5_i4b)
            write (prtunit,*) 'Absolute error of the computed minima              = ', err_stat(6_i4b)
            write (prtunit,*) 'Absolute error of the computed maxima              = ', err_stat(7_i4b)
!
            write (prtunit,*)
!
        end if
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( xn, xstat, xstat1, xstat2, x1, x2, tmp1, tmp2, mask )
!
    if ( success/=12_i4b ) then
        write (prtunit,*) 12_i4b - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_comp_unistat
! ================================
!
end program test_comp_unistat
