program test_fastgivens2_vec
!
!
! Purpose
! =======
!
!   This program is intended to test subroutines FASTGIVENS2_VEC and APPLY_ROT_FASTGIVENS
!   in module Giv_Procedures. 
!                                                                              
! LATEST REVISION : 30/10/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
! 
    use Statpack, only : lgl, i2b, i4b, stnd, n1_def, n2_def, n3_def,   &
                         half, one, c400, fastgivens2_vec,              &
                         apply_rot_fastgivens, norm, lamch
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT,
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def
!   
    real(stnd), parameter  :: fudge=c400
!
    character(len=*), parameter :: name_proc='fastgivens2_vec'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, eps2, safmin, beta, alpha, tmp
    real(stnd), dimension(:),   allocatable :: x, x2, res, d
    real(stnd), dimension(:,:), allocatable :: a, syst
!
    integer(i2b) :: type_rot
    integer(i4b) :: i, j, k, n, np1, m, success
    integer      :: iok
!
    logical(lgl) :: details
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE TESTS.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps    = fudge*epsilon( err )
    safmin = lamch( 'S' )
!
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
        np1 = n + 1_i4b
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( x(n), x2(n), d(n), res(n), a(n,n), syst(n,np1), stat=iok  )
!
        if ( iok/=0 ) cycle
!
!       GENERATE A RANDOM COEFFICIENT MATRIX a .
!
        call random_number( a )
!
        a = a - half
!
        call random_number( tmp )
!
        if ( tmp>safmin ) then
            a = a/tmp
        end if
!
!       GENERATE A RANDOM SOLUTION VECTOR x .
!
        call random_number( x )
!
        x = x - half
!
        call random_number( tmp )
!
        if ( tmp>safmin ) then
            x = x/tmp
        end if
!
        do j = 1_i4b, 2_i4b
!
            d(:n) = one
!
!           MAKE A COPY OF COEFFICIENT MATRIX a.
!    
            syst(:n,:n) = a(:n,:n)
!
!           COMPUTE RIGHT HAND-SIDE VECTOR b .
!
            syst(:n,np1) = matmul( a, x )
!
            select case( j )
!
                case( 1_i4b )
!
!                   TEST 1 :
!
!                   COMPUTE THE SOLUTION VECTOR FOR LINEAR SYSTEM
!
!                               a(:n,:n)*x(:n) = b(:n) .
!
!    
!                   TRANSFORM THE MATRIX a TO UPPER TRAPEZOIDAL FORM BY APPLYING 
!                   A SERIE OF GIVENS PLANE ROTATIONS ON THE ROWS OF a AND APPLY
!                   THE ROTATIONS TO b DIRECTLY.
!
                    do k = 1_i4b, n
!    
                        do i = k+1_i4b, n
!    
                            call fastgivens2_vec( syst(k,k:np1), syst(i,k:np1), d(k), d(i)  )
!    
                        end do
!    
                    end do
!
                case( 2_i4b )
!
!                   TEST 2 :
!
!                   COMPUTE THE SOLUTION VECTOR FOR LINEAR SYSTEM
!
!                               a(:n,:n)*x(:n) = b(:n) .
!
!    
!                   TRANSFORM THE MATRIX a TO UPPER TRAPEZOIDAL FORM BY APPLYING 
!                   A SERIE OF GIVENS PLANE ROTATIONS ON THE ROWS OF a.
!                   STORE THE ROTATIONS AND APPLY THEM TO b .
!
                    do k = 1_i4b, n
!    
                        do i = k+1_i4b, n
!    
                            call fastgivens2_vec( syst(k,k:n), syst(i,k:n), d(k), d(i), beta, alpha, type_rot   )
                            call apply_rot_fastgivens( syst(k,np1), syst(i,np1), beta, alpha, type_rot   )
!    
                        end do
!    
                    end do
!
            end select
!    
!           SOLVE THE n BY n TRIANGULAR SYSTEM.
!    
            do i = n, 1_i4b, -1_i4b
!    
                x2(i) = syst(i,np1)/syst(i,i)
                syst(1_i4b:i-1_i4b,np1) = syst(1_i4b:i-1_i4b,np1) - x2(i)*syst(1_i4b:i-1_i4b,i)
!    
            end do
!
            m = m + 1_i4b
!
            eps2 = real( n, stnd)*eps
!
!           CHECK THE ERROR FOR SOLUTION.
!
            res = x2 - x
!
!            err = norm(res) / ( norm(a) + norm(x) )
            err = norm(res) / norm(x)
!
            if ( err<=eps2 ) then
!
                success = success + 1_i4b
!
                if ( details ) then
!
                    write (prtunit,*) 'Test ',m,' is correct'
!
                    write (prtunit,*) 'n= ', n, ' , case= ', j
!
!                   PRINT RELATIVE ERROR OF COMPUTED SOLUTION.
!
                    write (prtunit,*) 'Relative error of the computed solution = ', err
!
                    write (prtunit,*)
!
                end if
!
            else
!
                write (prtunit,*) 'Test ',m,' is incorrect'
!
                write (prtunit,*) '=> ',err,' > ', eps2, ' , n= ', n, ' , case= ', j
!
!               PRINT RELATIVE ERROR OF COMPUTED SOLUTION.
!
                write (prtunit,*) 'Relative error of the computed solution = ', err
!
                write (prtunit,*)
!
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a, syst, res, x, x2, d )
!
    end do
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_fastgivens2_vec
! ===================================
!
end program test_fastgivens2_vec
