program test_freq_func
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine FREQ_FUNC in module Time_Series_Procedures.
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    The program assumes that subroutine FFT in module FFT_Procedures is correct.
!                                                                              
! LATEST REVISION : 09/09/2015
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, zero, half, one, true,                &
                         bd_coef, freq_func, init_fft, fft, end_fft
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT .
!
    integer(i4b), parameter :: prtunit=6, n=200
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                            :: err, eps, fch, fcl
    real(stnd), dimension(n)              :: freqr, freqr2, coef2
    real(stnd), dimension(:), allocatable :: coef
!
    complex(stnd), dimension(n)           :: coefc
!
    integer(i4b) :: k, k1, k2, pl, ph, khalf, kmid
!   
    integer :: iok
!   
!   
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='freq_func'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = sqrt( epsilon(eps) )
!
!   pl IS THE MINIMUM PERIOD OF OSCILLATION OF DESIRED COMPONENT.
!   ph IS THE MAXIMUM PERIOD OF OSCILLATION OF DESIRED COMPONENT.
!   pl AND ph ARE EXPRESSED IN NUMBER OF OBSERVATIONS.
!
    pl  = 18_i4b
    ph  = 96_i4b
!
!   COMPUTE THE CORRESPONDING CUTOFF FREQUENCIES.
!
    fch = one/real( ph, stnd )
    fcl = one/real( pl, stnd )
!
!   NOW SELECT THE OPTIMAL NUMBER OF FILTER COEFFICIENTS k FOR THE LANCZOS FILTER.
!
    k1  = ceiling( one/(half-fcl) )
    k2  = ceiling(  2.6_stnd/(fcl-fch) )
    k   = max( k1, k2, ph+1_i4b )
!
!   CHECK IF k IS ODD.
!
    if ( (k/2_i4b)*2_i4b==k  ) k = k + 1_i4b
!
!   ALLOCATE WORK ARRAY FOR THE FILTER COEFFICIENTS.
!
    allocate( coef(k), stat=iok )
!
    if ( iok==0 ) then
!
!       FUNCTION bd_coef COMPUTES THE K-TERM LEAST SQUARES APPROXIMATION TO AN
!       -IDEAL- BAND PASS FILTER WITH CUTOFF PERIODS PL AND PH (EG CUTOFF FREQUENCIES 1/PL AND 1/PH).
!
        coef(:k) = bd_coef( pl, ph, k )
!
!       NOW COMPUTE THE TRANSFERT FUNCTION freqr(:) OF THE LANCZOS FILTER AT THE FOURIER FREQUENCIES.
!
        call freq_func( n, coef(:k), freqr(:n), four_freq=true )
!
!       NOW, COMPUTE THE TRANSFERT FUNCTION DIRECTLY WITH THE FFT.
!
        kmid  = ( k + 1_i4b )/2_i4b
        khalf = ( k - 1_i4b )/2_i4b
!
        coef2(:n)          = zero
        coef2(:kmid)       = coef(kmid:k)
        coef2(n-khalf+1:n) = coef(1:khalf)
!
!       FIRST INITIALIZE THE FFT SUBROUTINE.
!
        call init_fft( n )
!
!       TRANSFORM THE TIME SERIES.
!
        coefc(:n) = cmplx( coef2(:n), zero, kind=stnd )
!
        call fft( coefc(:n), forward=true  )
!
        freqr2(:n) = real( coefc(:n),  kind=stnd )
!
!       DEALLOCATE WORK ARRAYS.
!
        call end_fft()
!
        deallocate( coef )
!
!       TEST THE ACCURACY OF THE RESULTS.
!
        err = maxval(abs(freqr(:n)-freqr2(:n)))/maxval(abs(freqr(:n)))
!
        if ( err<=eps  ) then
            write (prtunit,*) 'Test ',1_i4b,' is correct'
        else
            write (prtunit,*) 'Test ',1_i4b,' is incorrect'
            write (prtunit,*) '=> ',err,' > ',eps
        end if
!
    else
!
        write (prtunit,*) 'Problem in attempt to allocate memory for test 1'
!
    end if
!
!
! END OF PROGRAM test_freq_func
! =============================
!
end program test_freq_func
