program test_h2
!
!
! Purpose
! =======
!
!   This program is intended to test subroutines H2 and APPLY_H2 in module Hous_Procedures. 
!                                                                              
! LATEST REVISION : 01/12/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
! 
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def, true, false,     &
                         zero, half, c50, h2, apply_h2, norm, lamch, singvalues
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT,
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def
!   
    real(stnd), parameter :: fudge=c50, max_conda=1000._stnd
!
    character(len=*), parameter :: name_proc='h2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, eps2, safmin, tau, up, tmp, &
                                               conda, normx
    real(stnd), dimension(:),   allocatable :: x, b, res, singval
    real(stnd), dimension(:,:), allocatable :: a, a2
!
    integer(i4b) :: i, j, n, m, success
    integer      :: iok
!
    logical(lgl) :: details, left
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE TESTS.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps    = fudge*epsilon( err )
    safmin = lamch( 'S' )
!
    success = 0_i4b
    m       = 0_i4b
!
    left = false
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( b(n), x(n), res(n), singval(n),   &
                  a(n,n), a2(n,n), stat=iok  )
!
        if ( iok/=0 ) then
!
            if ( allocated(b) )         deallocate( b )
            if ( allocated(x) )         deallocate( x )
            if ( allocated(res) )       deallocate( res )
            if ( allocated(singval) )   deallocate( singval )
            if ( allocated(a) )         deallocate( a )
            if ( allocated(a2) )        deallocate( a2 )
!
            cycle
!
        end if
!
!       GENERATE A RANDOM COEFFICIENT MATRIX a .
!
        call random_number( a )
!
        a = a - half
!
        call random_number( tmp )
!
        if ( tmp>safmin ) then
            a = a/tmp
        end if
!
        singval(:n) = singvalues( a(:n,:n) )
!
#ifdef _F2003
        if ( ieee_support_datatype( singval ) ) then
            if ( .not.all( ieee_is_normal( singval(1_i4b:n) ) ) ) cycle
        end if
#endif
!
        j = count( singval(:n)/=zero )
!
!       SKIP DEFICIENT MATRICES.
!
        if ( j/=n ) cycle
!
!       COMPUTE CONDITION NUMBER OF RANDOM MATRIX.
!
        conda = maxval( singval(:n) )/minval( singval(:n) )
!
        tmp = min( conda, max_conda )
!
        eps2 = real( n, stnd)*eps*tmp
!
!       GENERATE A RANDOM SOLUTION VECTOR x .
!
        call random_number( x )
!
        x = x - half
!
        call random_number( tmp )
!
        if ( tmp>safmin ) then
            x = x/tmp
        end if
!
        normx = norm( x )
!
        do j = 1_i4b, 6_i4b
!
            left = .not. left
!
!           MAKE A COPY OF COEFFICIENT MATRIX a.
!    
            a2 = a
!
!           COMPUTE RIGHT HAND-SIDE VECTOR b .
!
            if ( left ) then
                b = matmul( a, x )
            else
                b = matmul( x, a )
            end if
!
            select case( j )
!
                case( 1_i4b )
!
!                   TEST 1 :
!
!                   COMPUTE THE SOLUTION VECTOR FOR LINEAR SYSTEM
!
!                               a(:n,:n)*x(:n) = b(:n) .
!    
!                   TRANSFORM THE MATRIX a TO UPPER TRIANGULAR FORM BY APPLYING 
!                   A SERIE OF HOUSEHOLDER REFLECTIONS ON THE COLUMNS OF a AND APPLY
!                   THE TRANSFORMATIONS TO b .
!
                    do i = 1_i4b, n
!
                        call h2( a2(i,i), a2(i+1_i4b:n,i), up, tau )
                        call apply_h2( a2(i+1_i4b:n,i), up, tau,    &
                                       a2(i,i+1_i4b:n), a2(i+1_i4b:n,i+1_i4b:n), left=true )
                        call apply_h2( a2(i+1_i4b:n,i), up, tau, b(i), b(i+1_i4b:n) )
!
                    end do
!
                case( 2_i4b )
!
!                   TEST 2 :
!
!                   COMPUTE THE SOLUTION VECTOR FOR LINEAR SYSTEM
!
!                               x(:n)*a(:n,:n) = b(:n) .
!
!    
!                   TRANSFORM THE MATRIX a TO LOWER TRIANGULAR FORM BY APPLYING 
!                   A SERIE OF HOUSEHOLDER REFLECTIONS ON THE ROWS OF a AND APPLY
!                   THE TRANSFORMATIONS TO b .
!
                    do i = 1_i4b, n
!
                        call h2( a2(i,i), a2(i,i+1_i4b:n), up, tau )
                        call apply_h2( a2(i,i+1_i4b:n), up, tau,    &
                                       a2(i+1_i4b:n,i), a2(i+1_i4b:n,i+1_i4b:n), left=false )
                        call apply_h2( a2(i,i+1_i4b:n), up, tau, b(i), b(i+1_i4b:n) )
!
                    end do
!
                case( 3_i4b )
!
!                   TEST 3 :
!
!                   COMPUTE THE SOLUTION VECTOR FOR LINEAR SYSTEM
!
!                               a(:n,:n)*x(:n) = b(:n) .
!    
!                   TRANSFORM THE MATRIX a TO UPPER TRIANGULAR FORM BY APPLYING 
!                   A SERIE OF HOUSEHOLDER REFLECTIONS ON THE COLUMNS OF a AND APPLY
!                   THE TRANSFORMATIONS TO b .
!
                    do i = 1_i4b, n
!
                        call h2( a2(i,i), a2(i+1_i4b:n,i), up, tau, b(i), b(i+1_i4b:n) )
                        call apply_h2( a2(i+1_i4b:n,i), up, tau,    &
                                       a2(i,i+1_i4b:n), a2(i+1_i4b:n,i+1_i4b:n), left=true )
!
                    end do
!
                case( 4_i4b )
!
!                   TEST 4 :
!
!                   COMPUTE THE SOLUTION VECTOR FOR LINEAR SYSTEM
!
!                               x(:n)*a(:n,:n) = b(:n) .
!
!    
!                   TRANSFORM THE MATRIX a TO LOWER TRIANGULAR FORM BY APPLYING 
!                   A SERIE OF HOUSEHOLDER REFLECTIONS ON THE ROWS OF a AND APPLY
!                   THE TRANSFORMATIONS TO b .
!
                    do i = 1_i4b, n
!
                        call h2( a2(i,i), a2(i,i+1_i4b:n), up, tau, b(i), b(i+1_i4b:n) )
                        call apply_h2( a2(i,i+1_i4b:n), up, tau,    &
                                       a2(i+1_i4b:n,i), a2(i+1_i4b:n,i+1_i4b:n), left=false )
!
                    end do
!
                case( 5_i4b )
!
!                   TEST 5 :
!
!                   COMPUTE THE SOLUTION VECTOR FOR LINEAR SYSTEM
!
!                               a(:n,:n)*x(:n) = b(:n) .
!    
!                   TRANSFORM THE MATRIX a TO UPPER TRIANGULAR FORM BY APPLYING 
!                   A SERIE OF HOUSEHOLDER REFLECTIONS ON THE COLUMNS OF a AND APPLY
!                   THE TRANSFORMATIONS TO b .
!
                    do i = 1_i4b, n
!
                        call h2( a2(i,i), a2(i+1_i4b:n,i), up, tau,     &
                                 a2(i,i+1_i4b:n), a2(i+1_i4b:n,i+1_i4b:n), left=true )
                        call apply_h2( a2(i+1_i4b:n,i), up, tau, b(i), b(i+1_i4b:n) )
!
                    end do
!
                case( 6_i4b )
!
!                   TEST 6 :
!
!                   COMPUTE THE SOLUTION VECTOR FOR LINEAR SYSTEM
!
!                               x(:n)*a(:n,:n) = b(:n) .
!
!    
!                   TRANSFORM THE MATRIX a TO LOWER TRIANGULAR FORM BY APPLYING 
!                   A SERIE OF HOUSEHOLDER REFLECTIONS ON THE ROWS OF a AND APPLY
!                   THE TRANSFORMATIONS TO b .
!
                    do i = 1_i4b, n
!
                        call h2( a2(i,i), a2(i,i+1_i4b:n), up, tau,    &
                                 a2(i+1_i4b:n,i), a2(i+1_i4b:n,i+1_i4b:n), left=false )
                        call apply_h2( a2(i,i+1_i4b:n), up, tau, b(i), b(i+1_i4b:n) )
!
                    end do
!
            end select
!
            if ( left ) then
!    
!               SOLVE THE n BY n UPPER TRIANGULAR SYSTEM.
!    
                do i = n, 1_i4b, -1_i4b
!    
                    b(i) = b(i)/a2(i,i)
                    b(1_i4b:i-1_i4b) = b(1_i4b:i-1_i4b) - b(i)*a2(1_i4b:i-1_i4b,i)
!    
                end do
!
            else
!    
!               SOLVE THE n BY n LOWER TRIANGULAR SYSTEM.
!    
                do i = n, 1_i4b, -1_i4b
!    
                    b(i) = ( b(i) - dot_product(a2(i+1_i4b:n,i),b(i+1_i4b:n)) )/a2(i,i)
!    
                end do
!
            end if
!
            m = m + 1_i4b
!
!           CHECK THE ERROR FOR SOLUTION.
!
            res = b - x
!
            err = norm( res ) / normx
!
            if ( err<=eps2 ) then
!
                success = success + 1_i4b
!
                if ( details ) then
!
                    write (prtunit,*) 'Test ',m,' is correct'
!
                    write (prtunit,*) 'n= ', n, ' , case= ', j
!
!                   PRINT RELATIVE ERROR OF COMPUTED SOLUTION.
!
                    write (prtunit,*) 'Relative error of the computed solution = ', err
!
                    write (prtunit,*)
!
                end if
!
            else
!
                write (prtunit,*) 'Test ',m,' is incorrect'
!
                write (prtunit,*) '=> ',err,' > ', eps2, ' , n= ', n, ' , case= ', j
!
!               PRINT RELATIVE ERROR OF COMPUTED SOLUTION.
!
                write (prtunit,*) 'Relative error of the computed solution = ', err
!
                write (prtunit,*)
!
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( a, a2, res, singval, x, b )
!
    end do
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_h2
! ======================
!
end program test_h2
