program test_llsq_svd_solve
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine LLSQ_SVD_SOLVE
!   in module LLSQ_Procedures.
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    Random matrices with given singular value distributions are generated for checking
!    the accuracy of subroutine LLSQ_SVD_SOLVE.
!
! LATEST REVISION : 25/05/2021
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def, allocate_error,    &
                         false, zero, half, one, two, five, ten, c100,              &
                         qr_cmp, ortho_gen_qr, llsq_svd_solve, arth, lamch, norm,   &
                         merror
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def, nrhs=10
!   
    real(stnd), parameter  :: fudge=ten, max_conda=10000._stnd
!
    character(len=*), parameter :: name_proc='llsq_svd_solve'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, err3, eps, eps2, ulp,   &
                                               safmin, tmp, shift, tol, conda
    real(stnd), allocatable, dimension(:)   :: eigval, singval, diagr, beta,   &
                                               bnorm, resnorm, res0norm,       &
                                               rnorm, x0norm, condax, condar
    real(stnd), allocatable, dimension(:,:) :: a, a2, leftvec, rightvec, rightvec2,    &
                                               rightvect, b0, b, b2, x, x0, res0, res2
!
    integer(i4b) :: i, j, k, l, n, nr, nc, m, p, krank0, krank, success, notdone
    integer      :: iok
!
    logical(lgl) :: details, failure, perfect_shift
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    eps    = fudge*ulp
!
    safmin = lamch( 'S' )
!
!   SPECIFY EVENTUALLY A SHIFT TO AVOID ZERO SINGULAR VALUE
!   OR SHIFT THE SINGULAR VALUES DISTRIBUTION.
!
    shift = 0.01_stnd
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( resnorm(nrhs), res0norm(nrhs),  &
              bnorm(nrhs),   condax(nrhs),    &
              condar(nrhs),  rnorm(nrhs),     &
              x0norm(nrhs),  stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    perfect_shift = false
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( eigval(n), singval(n),       &
                  diagr(n), beta(n), stat=iok  )
!
        if ( iok/=0 ) then
!
            if ( allocated(eigval) )    deallocate( eigval )
            if ( allocated(singval) )   deallocate( singval )
            if ( allocated(diagr) )     deallocate( diagr )
            if ( allocated(beta) )      deallocate( beta )
!
            cycle
!
        end if
!
        do k = 1_i4b, 5_i4b
!
            select case( k )
!
                case( 1_i4b )
!
!                   TEST 1: SVD OF a(:nr,:nc) WITH nr=nc .
!
                    nr = n
                    nc = n
!
                case( 2_i4b )
!
!                   TEST 2: SVD OF a(:nr,:nc) WITH nr>=nc AND nr NOT MUCH LARGER THAN nc .
!
                    nr = n + 1_i4b
                    nc = n
!
                case( 3_i4b )
!
!                   TEST 3: SVD OF a(:nr,:nc) WITH nr>=nc AND nr MUCH LARGER THAN nc .
!
                    nr = n + n
                    nc = n
!
                case( 4_i4b )
!
!                   TEST 4: SVD OF a(:nr,:nc) WITH nc>=nr AND nc NOT MUCH LARGER THAN nr .
!
                    nr = n
                    nc = n + 1_i4b
!
                case( 5_i4b )
!
!                   TEST 5: SVD OF a(:nr,:nc) WITH nc>=nr AND nc MUCH LARGER THAN nr .
!
                    nr = n
                    nc = n + n
!
            end select
!
!           ALLOCATE WORK ARRAYS.
!
            allocate( a(nr,nc), a2(nr,nc), res0(nr,nrhs), res2(nc,nrhs),  &
                      leftvec(nr,n), rightvec(nc,n), rightvec2(nc,n),     &
                      x0(nc,nrhs), x(nc,nrhs), b(nr,nrhs), b2(nr,nrhs),   &
                      b0(nr,nrhs), rightvect(n,nc), stat=iok  )
!
            if ( iok/=0 ) then
!
                if ( allocated(a) )         deallocate( a )
                if ( allocated(a2) )        deallocate( a2 )
                if ( allocated(res0) )      deallocate( res0 )
                if ( allocated(res2) )      deallocate( res2 )
                if ( allocated(leftvec) )   deallocate( leftvec )
                if ( allocated(rightvec) )  deallocate( rightvec )
                if ( allocated(rightvec2) ) deallocate( rightvec2 )
                if ( allocated(rightvect) ) deallocate( rightvect )
                if ( allocated(x) )         deallocate( x )
                if ( allocated(x0) )        deallocate( x0 )
                if ( allocated(b0) )        deallocate( b0 )
                if ( allocated(b) )         deallocate( b )
                if ( allocated(b2) )        deallocate( b2 )
!
                cycle
!
            end if
!
!           GENERATE A RANDOM COEFFICIENT MATRIX leftvec(:nr,:n) .
!
            call random_number( leftvec(:nr,:n) )
!
            leftvec(:nr,:n) = leftvec(:nr,:n) - half
!
            call random_number( tmp )
!
            if ( tmp>safmin ) then
                leftvec(:nr,:n) = leftvec(:nr,:n)/tmp
            end if
!
!           COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
            call qr_cmp( leftvec(:nr,:n), diagr(:n), beta(:n) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( leftvec(:nr,:n), beta(:n) )
!
!           GENERATE A RANDOM COEFFICIENT MATRIX rightvec(:nc,:n).
!
            call random_number( rightvec(:nc,:n) )
!
            rightvec(:nc,:n) = rightvec(:nc,:n) - half
!
            call random_number( tmp )
!
            if ( tmp>safmin ) then
                rightvec(:nc,:n) = rightvec(:nc,:n)/tmp
            end if
!
!           COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
            call qr_cmp( rightvec(:nc,:n), diagr(:n), beta(:n) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( rightvec(:nc,:n), beta(:n) )
!
!           GENERATE A RANDOM RIGHT HAND SIDE MATRIX b0.
!
            call random_number( b0(:nr,:nrhs) )
!
            b0(:nr,:nrhs) = b0(:nr,:nrhs) - half
!
            call random_number( tmp )
!
            if ( tmp>safmin ) then
                b0(:nr,:nrhs) = b0(:nr,:nrhs)/tmp
            end if
!
!           SPECIFY THRESHOLD FOR THE SINGULAR VALUES TO BE USED IN SUBROUTINE llsq_svd_solve.
!
!            tol = real( n, stnd)*ulp
            tol = ulp
!
!           COMPUTE FIRST THE EIGENVALUES OF THE MATRIX PRODUCT
!                     a(:nr,:nc)**(t)*a(:nr,:nc).
!
            do j = 1_i4b, 16_i4b
!
                select case( j )
!
                    case( 1_i4b )
!
!                       TEST 1: STRONGLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b)   = one
                        eigval(2_i4b:n) = ulp
!
                    case( 2_i4b )
!
!                       TEST 2:  STRONGLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n-1_i4b) = one
                        eigval(n)             = ulp
!
                    case( 3_i4b )
!
!                       TEST 3: WEAKLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n-1_i4b) = arth( ulp, ulp, n-1_i4b )
                        eigval(n)             = one
!
                    case( 4_i4b )
!
!                       TEST 4: INTERNAL CLUSTER OF EIGENVALUES.
!
                        eigval(1_i4b)         = ulp
                        eigval(2_i4b:n-1_i4b) = one + sqrt(ulp)*arth( one, one, n-2_i4b )
                        eigval(n)             = two
!
                    case( 5_i4b )
!
!                       TEST 5: WEAKLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n) = one + (c100*ulp)*arth( zero, one, n )
!
                    case( 6_i4b )
!
!                       TEST 6: GEOMETRIC DISTRIBUTION OF EIGENVALUES.
!
                        err = real( n-1_i4b, stnd )
                        do l = 1_i4b, n
                            eigval(l) = ulp**(real(l-1_i4b,stnd)/err)
                        end do
!
                    case( 7_i4b )
!
!                       TEST 7: ARITHMETIC DISTRIBUTION OF EIGENVALUES.
!
                        eigval(1_i4b:n) = ulp + ((one-ulp)/real( n-1_i4b, stnd ))*arth( zero, one, n )
!
                    case( 8_i4b )
!
!                       TEST 8: TWO CLUSTERS OF EIGENVALUES.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = one + tmp*arth( one, one, i )
!
                        eigval(i+1_i4b:n) = ten + tmp*arth( one, one, n-i )
!
                    case( 9_i4b )
!
!                       TEST 9: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR ZERO.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = arth( one, one, i )
!
                        eigval(i+1_i4b:n) = tmp*arth( one, one, n-i )
!
                    case( 10_i4b )
!
!                       TEST 10: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR n/2.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = arth( one, one, i )
!
                        eigval(i+1_i4b:n) = real( i, stnd) + tmp*arth( one, one, n-i )
!
                    case( 11_i4b )
!
!                       TEST 11: CLUSTERED DISTRIBUTION OF EIGENVALUES BETWEEN 0.9 and 1.
!
                        call random_number( eigval(:n) )
!
                        eigval(1_i4b:n) = 0.9_stnd + 0.1_stnd*eigval(1_i4b:n)
!
                    case( 12_i4b )
!
!                       TEST 12: EIGENVALUES EQUAL TO sinh(10*i/n) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) )
!
                    case( 13_i4b )
!
!                       TEST 13: EIGENVALUES EQUAL TO sinh((10*i/n)-5) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - five )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        where( eigval(1_i4b:n)<zero )
                            beta(:n) = -eigval(1_i4b:n)
                        elsewhere
                            beta(:n) = zero
                        end where
!
                        tmp = maxval( beta(:n) ) + one
!
!                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case( 14_i4b )
!
!                       TEST 14: EIGENVALUES EQUAL TO sinh((10*i/n)-10) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - ten )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        where( eigval(1_i4b:n)<zero )
                            beta(:n) = -eigval(1_i4b:n)
                        elsewhere
                            beta(:n) = zero
                        end where
!
                        tmp = maxval( beta(:n) ) + one
!
!                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case( 15_i4b )
!
!                       TEST 15: EIGENVALUES EQUAL TO tanh((10*i/n)-5) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = tanh( tmp*arth( one, one, n ) - five )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        where( eigval(1_i4b:n)<zero )
                            beta(:n) = -eigval(1_i4b:n)
                        elsewhere
                            beta(:n) = zero
                        end where
!
                        tmp = maxval( beta(:n) ) + one
!
!                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case default
!
!                       DEFAULT TEST : RANDOMLY UNIFORM DISTRIBUTION OF EIGENVALUES.
!
                        call random_number( eigval(:n) )
!
                        call random_number( tmp )
!
                        if ( tmp>safmin ) then
                            eigval(1_i4b:n) = eigval(1_i4b:n)/tmp
                        end if
!
                end select
!
#ifdef _F2003
                if ( ieee_support_datatype( eigval ) ) then
                    if ( .not.all( ieee_is_normal( eigval(1_i4b:n) ) ) ) cycle
                end if
#endif
!
!               COMPUTE THE SINGULAR VALUES FROM THE EIGENVALUES.
!
                singval(:n) = sqrt( eigval(:n) )
!
                singval(:n) = singval(:n) + shift
!
                krank0 = count( singval(:n)/=zero )
!
!               SKIP DEFICIENT MATRICES.
!
                if ( krank0/=n ) cycle
!
!               FORM A MATRIX FROM THE SPECIFIED SINGULAR VALUE DISTRIBUTION AND RANDOM SINGULAR VECTORS.
!
                rightvect(:n,:nc) = transpose( rightvec(:nc,:n) )
!
                do i = 1_i4b, nc
!
                    rightvect(:n,i) = rightvect(:n,i)*singval(:n)
!
                end do
!
                a2(:nr,:nc) = matmul( leftvec(:nr,:n), rightvect(:n,:nc) )
!
!               EXCLUDE TESTS OF MATRICES WITH DENORMALIZED REAL NUMBERS TO AVOID PROBLEMS IN EXTENDED
!               PRECISION.
!
#ifdef _F2003
                if ( ieee_support_datatype( a2 ) ) then
                    if ( .not.all( ieee_is_normal( a2 ) ) ) cycle
                end if
#else
                if ( any( abs(a2)<safmin .and. a2/=zero ) ) cycle
#endif
!
!               COMPUTE THE CONDITION NUMBER OF THE MATRIX.
!
                conda = maxval( singval(:n) )/minval( singval(:n) )
!
!               COMPUTE THE RIGHT HAND SIDE MATRIX b.
!
                call random_number( x(:nc,:nrhs) )
!
                b(:nr,:nrhs) = b0(:nr,:nrhs) + matmul( a2(:nr,:nc), x(:nc,:nrhs) )
!
!               COMPUTE THE NORMS OF THE COLUMNS OR RANDOM RIGHT HAND SIDE MATRIX b .
!
                bnorm(:nrhs) = norm( b(:nr,:nrhs), dim=2_i4b )
!
!               COMPUTE THE MINIMUM 2-NORM SOLUTIONS x0 OF THE LINEAR LEAST SQUARE PROBLEMS.
!
                do i = 1_i4b, n
!
                    x(i,:nrhs)       = matmul( leftvec(:nr,i), b(:nr,:nrhs) )
                    rightvec2(:nc,i) = rightvec(:nc,i)/singval(i)
!
                end do
!
                x0(:nc,:nrhs) = matmul( rightvec2(:nc,:n), x(:n,:nrhs) ) 
!
                x0norm(:nrhs) = norm( x0(:nc,:nrhs), dim=2_i4b )
!
                where( x0norm(:nrhs)<=safmin ) x0norm(:nrhs) = one
!
!               COMPUTE THE RESIDUALS res0 OF THE LINEAR LEAST SQUARE PROBLEMS
!               AND THE SCALING FACTORS FOR THE TESTS.
!
                tmp = min( conda, max_conda )
!
                if ( krank0==nr ) then
!
                    res0(:nr,:nrhs) = zero
!
                    res0norm(:nrhs) = zero
!
                    condax(:nrhs) = two*tmp*x0norm(:nrhs)
!
                else
!
                    res0(:nr,:nrhs) = b(:nr,:nrhs) - matmul( leftvec(:nr,:n), x(:n,:nrhs) )
!
                    res0norm(:nrhs) = norm( res0(:nr,:nrhs), dim=2_i4b )
!
                    condax(:nrhs) = bnorm(:nrhs)**2 - res0norm(:nrhs)**2
!
                    where( condax(:nrhs)<=safmin ) condax(:nrhs) = one
!
                    resnorm(:nrhs) = res0norm(:nrhs)/sqrt( condax(:nrhs) )
!
                    condax(:nrhs) = ( two*tmp + (tmp*tmp)*resnorm(:nrhs) )*x0norm(:nrhs)
!
                end if
!
                condar(:nrhs) = ( one + two*tmp )*bnorm(:nrhs)
!
!               RECOMPUTE THE SOLUTIONS AND RESIDUALS OF THE LEAST SQUARE PROBLEMS WITH
!               WITH SUBROUTINE llsq_svd_solve.
!
                do i = 1_i4b, 2_i4b
!
                    perfect_shift = .not.perfect_shift
!
                    if ( i==1_i4b ) then
                        p = nrhs
                    else
                        p = 1_i4b
                    end if
!
!                   MAKE A COPY OF RIGHT HAND-SIDE MATRIX b.
!
                    b2(:nr,:p) = b(:nr,:p)
!
!                   RESTORE RANDOM MATRIX.
!
                    a(:nr,:nc) = a2(:nr,:nc)
!
                    select case( i )
!
                        case( 1_i4b )
!
!                           TEST 1 : SEVERAL RIGHT HAND-SIDES.
!
!                           COMPUTE SVD FACTORIZATION OF a.
!                           COMPUTE THE SOLUTION MATRIX FOR LINEAR LEAST SQUARE SYSTEM
!
!                                        a*x = b .
!
                            call llsq_svd_solve( a, b2, failure, x, rnorm=rnorm,                   &
                                                 krank=krank, tol=tol, perfect_shift=perfect_shift )
!
!                           CHECK THE RESULTS FOR SMALL RELATIVE ERRORS FOR THE SOLUTIONS.
!
                            res2(:nc,:nrhs) = x0(:nc,:nrhs) - x(:nc,:nrhs)
!
                            resnorm(:nrhs) = norm( res2(:nc,:nrhs), dim=2_i4b )
!
                            err1 = maxval( resnorm(:nrhs)/condax(:nrhs) )
!
!                           CHECK THE RESULTS FOR SMALL RELATIVE ERRORS FOR THE RESIDUALS.
!
                            resnorm(:nrhs) = abs( rnorm(:nrhs) - res0norm(:nrhs) )
!
                            err2 = maxval( resnorm(:nrhs)/condar(:nrhs) )
!
                            b2(:nr,:nrhs) = b(:nr,:nrhs) - matmul( a2(:nr,:nc), x(:nc,:nrhs) )
!
                            if ( krank0/=nr ) then
!
                                b2(:nr,:nrhs) = b2(:nr,:nrhs) - res0(:nr,:nrhs)
!
                            end if
!
                            resnorm(:nrhs) = norm( b2(:nr,:nrhs), dim=2_i4b )
!
                            err3 = maxval( resnorm(:nrhs)/condar(:nrhs) )
!
                        case( 2_i4b  )
!
!                           TEST 2 : ONE RIGHT HAND-SIDE.
!
!                           COMPUTE SVD FACTORIZATION OF a.
!                           COMPUTE THE SOLUTION VECTOR FOR LINEAR LEAST SQUARE SYSTEM
!
!                                        a*x = b .
!
                            call llsq_svd_solve( a, b2(:nr,1_i4b), failure, x(:nc,1_i4b),  &
                                                 rnorm=rnorm(1_i4b), krank=krank, tol=tol, &
                                                 perfect_shift=perfect_shift )
!
!                           CHECK THE RESULTS FOR SMALL RELATIVE ERRORS FOR THE SOLUTIONS.
!
                            res2(:nc,1_i4b) = x0(:nc,1_i4b) - x(:nc,1_i4b)
!
                            resnorm(1_i4b) = norm( res2(:nc,1_i4b) )
!
                            err1 = resnorm(1_i4b)/condax(1_i4b)
!
!                           CHECK THE RESULTS FOR SMALL RELATIVE ERRORS FOR THE RESIDUALS.
!
                            resnorm(1_i4b) = abs( rnorm(1_i4b) - res0norm(1_i4b) )
!
                            err2 = resnorm(1_i4b)/condar(1_i4b)
!
                            b2(:nr,1_i4b) = b(:nr,1_i4b) - matmul( a2(:nr,:nc), x(:nc,1_i4b) )
!
                            if ( krank0/=nr ) then
!
                                b2(:nr,1_i4b) = b2(:nr,1_i4b) - res0(:nr,1_i4b)
!
                            end if
!
                            resnorm(1_i4b) = norm( b2(:nr,1_i4b) )
!
                            err3 = resnorm(1_i4b)/condar(1_i4b)
!
                    end select
!
                    if ( failure ) then
!
                        notdone = notdone + 1_i4b
!
                        cycle
!
                    end if
!
                    m = m + 1_i4b
!
                    err =  max( err1, err2, err3 )
!
                    eps2 = real( n, stnd)*eps
!                    eps2 = sqrt( ulp )
!
!                    if ( err<=eps2 ) then
                    if ( err<=eps2 .and. krank==krank0 ) then
!
                        success = success + 1_i4b
!
                        if ( details ) then
!
                            write (prtunit,*) 'Test ',m,' is correct'
!
                            write (prtunit,*) 'nr= ', nr, ' , nc= ', nc,' , case= ', j,                     &
                                              ' , perfect_shift= ', perfect_shift, ' , failure= ', failure, &
                                              ' , krank0= ', krank0, ' , krank= ', krank,                   &
                                              ' , condition number= ', conda, ' , nrhs= ', p
!
!                           PRINT SCALED ERRORS OF SOLUTIONS AND RESIDUALS.
!
                            write (prtunit,*) 'Maximum scaled error of the computed solutions          = ', err1
                            write (prtunit,*) 'Maximum scaled error of the norms of computed residuals = ', err2
                            write (prtunit,*) 'Maximum scaled error of the computed residuals          = ', err3
!
                            write (prtunit,*)
!
                        end if
!
                    else
!
                        write (prtunit,*) 'Test ',m,' is incorrect'
!
                        write (prtunit,*) '=> ', err, ' > ', eps2,' , nr= ', nr, ' , nc= ', nc,' , case= ', j,  &
                                          ' , perfect_shift= ', perfect_shift, ' , failure= ', failure,         &
                                          ' , krank0= ', krank0,' , krank= ', krank,                            &
                                          ' , condition number= ', conda, ' , nrhs= ', p
!
!                       PRINT SCALED ERRORS OF SOLUTIONS AND RESIDUALS.
!
                        write (prtunit,*) 'Maximum scaled error of the computed solutions          = ', err1
                        write (prtunit,*) 'Maximum scaled error of the norms of computed residuals = ', err2
                        write (prtunit,*) 'Maximum scaled error of the computed residuals          = ', err3
!
                        write (prtunit,*)
!
                    end if
!
                end do
!
            end do
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( a, a2, res0, res2, leftvec, rightvec,  &
                        rightvec2, rightvect, x, x0, b0, b, b2 )
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( eigval, singval, diagr, beta  )
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( rnorm, resnorm, res0norm, x0norm, bnorm, condax, condar )
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_llsq_svd_solve
! ==================================
!
end program test_llsq_svd_solve
