program test_moddan_filter
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine MODDAN_FILTER in module Time_Series_Procedures.
!                                                                              
! LATEST REVISION : 03/02/2010
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, moddan_coef, moddan_filter, symlin_filter2
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT .
!
    integer(i4b), parameter :: prtunit=6, nparam=3, m=10
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, tmp
    real(stnd), dimension(:,:), allocatable :: matb, matc
    real(stnd), dimension(:),   allocatable :: coef, vec, vec2, vec3
!
    integer(i4b) :: n, i, j, k, kmid, khalf, k1, k2
    integer(i4b), dimension(nparam) :: smooth_param
!   
    integer :: iok
!   
!   
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='moddan_filter'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TESTS.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   SET THE MODIFIED DANIELL FILTERS.
!
    smooth_param(:nparam) = (/ 2_i4b, 2_i4b, 3_i4b /)
!
!   COMPUTE THE NUMBER OF FILTER COEFFICIENTS.
!
    k = 2_i4b*( 2_i4b + sum( smooth_param(:nparam) )) - 1_i4b
!
!   SET THE NUMBER OF OBSERVATIONS
!
!    n = 100_i4b*k
    n = 1700_i4b
!
!   ALLOCATE WORK ARRAYS .
!
    allocate( coef(k), vec(n), vec2(n), vec3(n),    &
              matb(m,n), matc(m,n), stat=iok        )
!
    if ( iok==0 ) then
!
!       FUNCTION moddan_coef COMPUTES THE IMPULSE RESPONSE FUNCTION CORRESPONDING TO A NUMBER
!       OF APPLICATIONS MODIFIED DANIELL FILTERS .
!
        coef(:k) = moddan_coef( k, smooth_param(:nparam) )
!
!       GENERATE A RANDOM REAL NUMBER SEQUENCE OF LENGTH n .
!
        call random_number( vec(:n) )
!
        kmid  = ( k + 1_i4b )/2_i4b
        khalf = ( k - 1_i4b )/2_i4b
!
!        do i = 1_i4b, khalf
!            j = k + 1_i4b - i
!            coef(j) = coef(i)
!        end do
!
        k1    = kmid
        k2    = n - khalf
!
        tmp     =  maxval(abs(vec(k1:k2)))
!
        do i = 1_i4b, 3_i4b
!
            select case( i )
!
                case( 1_i4b)
!
!                   TEST 1 : TEST 1D MODIFIED DANIELL FILTERING.
!
!                   SAVE THE REAL RANDOM NUMBER ARRAY.
!
                    vec2(:n) = vec(:n)
                    vec3(:n) = vec(:n)
!
!                   FILTER THE TIME SERIES .
!                   NOTE THAT (size(COEF)-1)/2 DATA POINTS WILL BE AFFECTED
!                   BY END EFFECTS  FROM EACH END OF THE SERIES.
!
                    call symlin_filter2( vec2(:n), coef(1_i4b:k) )
!
!                   FILTER AGAIN THE TIME SERIES, BUT COMPUTE THE VALUES AT THE ENDS OF THE OUTPUT
!                   BY ASSUMING EVEN SYMMETRY OF THE INPUT TIME SERIES .
!
                    call moddan_filter( vec3(:n), smooth_param(:nparam) )
!
!                   TEST THE ACCURACY OF THE FILTERING OPERATION.
!
                    err = maxval(abs(vec3(k1:k2)-vec2(k1:k2)))/tmp
!
                case( 2_i4b)
!
!                   TEST 2 : TEST 1D MODIFIED DANIELL FILTERING.
!
!                   SAVE THE REAL RANDOM NUMBER ARRAY.
!
                    matb(1_i4b,:n) = vec(:n)
                    matc(1_i4b,:n) = vec(:n)
!
!                   FILTER THE TIME SERIES .
!                   NOTE THAT (size(COEF)-1)/2 DATA POINTS WILL BE AFFECTED
!                   BY END EFFECTS  FROM EACH END OF THE SERIES.
!
                    call symlin_filter2( matb(1_i4b,:n), coef(1_i4b:k) )
!
!                   FILTER AGAIN THE TIME SERIES, BUT COMPUTE THE VALUES AT THE ENDS OF THE OUTPUT
!                   BY ASSUMING EVEN SYMMETRY OF THE INPUT TIME SERIES .
!
                    call moddan_filter( matc(1_i4b,:n), smooth_param(:nparam) )
!
!                   TEST THE ACCURACY OF THE FILTERING OPERATION.
!
                    err = maxval(abs(matc(1_i4b,k1:k2)-matb(1_i4b,k1:k2)))/tmp
!
                case( 3_i4b)
!
!                   TEST 3 : TEST 2D MODIFIED DANIELL FILTERING.
!
!                   SAVE THE REAL RANDOM NUMBER ARRAY.
!
                    do j = 1_i4b, m
                        matb(j,:n) = vec(:n)
                        matc(j,:n) = vec(:n)
                    end do
!
!                   FILTER THE TIME SERIES .
!                   NOTE THAT (size(COEF)-1)/2 DATA POINTS WILL BE AFFECTED
!                   BY END EFFECTS  FROM EACH END OF THE SERIES.
!
                    call symlin_filter2( matb(:m,:n), coef(1_i4b:k) )
!
!                   FILTER AGAIN THE TIME SERIES, BUT COMPUTE THE VALUES AT THE ENDS OF THE OUTPUT
!                   BY ASSUMING EVEN SYMMETRY OF THE INPUT TIME SERIES .
!
                    call moddan_filter( matc(:m,:n), smooth_param(:nparam) )
!
!                   TEST THE ACCURACY OF THE FILTERING OPERATION.
!
                    err = maxval(abs(matc(:m,k1:k2)-matb(:m,k1:k2)))/tmp
!
            end select
!
!           PRINT THE RESULTS OF THE TEST.
!
            if ( err<=sqrt(epsilon(err)) ) then
                write (prtunit,*) 'Test ',i,' is correct'
            else
                write (prtunit,*) 'Test ',i,' is incorrect'
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( coef, vec, vec2, vec3, matb, matc )
!
    else
!
        write (prtunit,*) 'Problem in attempt to allocate memory for tests 1, 2 and 3'
!
    end if
!
!
! END OF PROGRAM test_moddan_filter
! =================================
!
end program test_moddan_filter
