program test_permute_cor
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine PERMUTE_COR in module Mul_Stat_Procedures .
!                                                                              
! LATEST REVISION : 05/07/2010
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, lgl, true, comp_cor, permute_cor
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
! p       IS THE NUMBER OF OBSERVATIONS OF THE RANDOM VECTORS
! nrep    IS THE NUMBER OF SHUFFLES FOR THE PERMUTATION TEST
! nsample IS THE NUMBER OF SAMPLES TO EVALUATE THE REJECTION RATE OF THE TEST
!
    integer(i4b), parameter :: prtunit=6, p=50, nrep=99, nsample=3000
!
! sign_level IS THE SIGNIFICANCE LEVEL OF THE PERMUTATION TEST
! eps        IS THE ALLOWED RELATIVE ERROR FOR THE REJECTION RATE
!
    real(stnd),  parameter :: sign_level=0.05, eps=0.2
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                       :: xyn, err_prob, prob_rej_rate
    real(stnd), dimension(nsample)   :: xycor, prob
    real(stnd), dimension(nsample,2) :: xstat
    real(stnd), dimension(2)         :: ystat
    real(stnd), dimension(p,nsample) :: x
    real(stnd), dimension(nsample,p) :: x2
    real(stnd), dimension(p)         :: y
!
    integer(i4b) :: i, j, rej_rate
!
    logical(lgl) :: first, last
!   
!   
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='permute_cor'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   INITIALIZE THE RANDOM GENERATOR.
!
    call random_seed()
!
!   GENERATE A RANDOM UNIFORM OBSERVATION VECTOR y .
!
    call random_number( y(:p) )
!
    first = true
    last  = true
!
    do i = 1_i4b, 3_i4b
!
!       GENERATE A RANDOM UNIFORM OBSERVATION ARRAY x .
!
        call random_number( x(:p,:nsample) )
!
!       COMPUTE THE CORRELATIONS BETWEEN x AND y
!       FOR THE p OBSERVATIONS .
!
        call comp_cor( x(:p,:nsample), y(:p), first, last, xstat(:nsample,:2_i4b), ystat(:2_i4b),    &
                       xycor(:nsample), xyn, dimvar=2_i4b )
!
!       NOW COMPUTE A PERMUTATION TEST OF THE CORRELATION BETWEEN x AND y WITH
!       SUBROUTINE permute_cor WITH nrep SHUFFLES .
!
        select case( i )
!
            case( 1_i4b )
!
!               TEST 1  :
!
!               COMPUTE THE PERMUTATION TEST WITH permute_cor_rv .
!
                do j = 1_i4b, nsample
                    call permute_cor( x(:p,j), y(:p), xstat(j,:2_i4b), ystat(:2_i4b),       &
                                      xycor(j), prob(j), nrep=nrep )
                end do
!
            case( 2_i4b )
!
!               TEST 2  :
!
!               COMPUTE THE PERMUTATION TEST WITH permute_cor_rm AND dimvar=2 .
!
                call permute_cor( x(:p,:nsample), y(:p), xstat(:nsample,:2_i4b), ystat(:2_i4b),       &
                                  xycor(:nsample), prob(:nsample), nrep=nrep, dimvar=2_i4b )
!
            case( 3_i4b )
!
!               TEST 3  :
!
!               COMPUTE THE PERMUTATION TEST WITH permute_cor_rm AND dimvar=1 .
!
                x2(:nsample,:p) = transpose( x(:p,:nsample) )
                call permute_cor( x2(:nsample,:p), y(:p), xstat(:nsample,:2_i4b), ystat(:2_i4b),       &
                                  xycor(:nsample), prob(:nsample), nrep=nrep )
!
        end select
!
!       NOW COMPUTE THE REJECTION RATE OF THE TEST. IDEALLY, FOR THE sign_level
!       SIGNIFICANCE LEVEL, WE WOULD REJECT ONLY THE NULL HYPOTHESIS sign_level %
!       OF THE TIME. IN OTHER WORDS, THE REJECTION RATE MUST BE APPROXIMATELY EQUAL
!       TO THE SIGNIFICANCE LEVEL sign_level .
!
        rej_rate      = count( prob(:nsample)<=sign_level )
        prob_rej_rate = real( rej_rate, stnd )/real( nsample, stnd )
!
!       COMPUTE THE RELATIVE ERROR OF THE REJECTION RATE.
!
        err_prob = abs( (prob_rej_rate-sign_level)/sign_level )
!
!       CHECK THAT THE RELATIVE ERROR OF THE REJECTION RATE IS LESS THAN eps .
!
        if ( err_prob<=eps ) then
            write (prtunit,*) 'Test ',i,' is correct'
        else
            write (prtunit,*) 'Test ',i,' is incorrect'
            write (prtunit,*) '=> ',err_prob,' > ',eps
        end if
!
    end do
!
!
! END OF PROGRAM test_permute_cor
! ===============================
!
end program test_permute_cor
