program test_pinvq
!
!
! Purpose
! =======
!
!   This program is intended to test functions PROBQ and PINVQ in module Prob_Procedures.
!                                                                              
!
! Further Details
! ===============
!                                                                              
! LATEST REVISION : 20/09/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, lgl, n1_def, n2_def, n3_def,    &
                         true, false, zero, half, one, c20,         &
                         allocate_error, merror, probq, pinvq
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def, ndf=10, ndf_max=100
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
!    real(stnd),  parameter :: eps = 1.0e-4_stnd 
!
    character(len=*), parameter :: name_proc='pinvq'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, ulp
    real(stnd), allocatable, dimension(:)   :: df_vec
    real(stnd), allocatable, dimension(:,:) :: prob, p, p2, x, df_mat
!
    integer(i4b), allocatable, dimension(:)   :: ndf_vec
    integer(i4b), allocatable, dimension(:,:) :: ndf_mat
    integer(i4b) :: i, j, k, l, m, n, success
    integer      :: iok
!
    logical(lgl) :: upper
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
    ulp = epsilon( ulp )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( df_vec(ndf), ndf_vec(ndf), stat=iok  )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( p(n,n), p2(n,n), prob(n,n), x(n,n),  &
                  ndf_mat(n,n), df_mat(n,n), stat=iok  )
!
        if ( iok/=0 ) then
!
            if ( allocated(p) )         deallocate( p )
            if ( allocated(p2) )        deallocate( p2 )
            if ( allocated(prob) )      deallocate( prob )
            if ( allocated(x) )         deallocate( x )
            if ( allocated(ndf_mat) )   deallocate( ndf_mat )
            if ( allocated(df_mat) )    deallocate( df_mat )
!
            cycle
!
        end if
!
!       GENERATE A RANDOM PROBABILITY MATRIX p .
!
        call random_number( p(:n,:n) )
!
        where( p(:n,:n)<=ulp .or. p(:n,:n)>=one-ulp )  p(:n,:n) = half
!
        do i = 1_i4b, 20_i4b
!
            m = m + 1_i4b
!
            err = zero
!
            select case( i )
!
            case( 1_i4b )
!
!               SET THE REQUIRED PRECISION OF THE RESULTS FOR TESTS 1 TO 10.
!
                eps = 1.0e-5_stnd 
!
!               GENERATE A VECTOR OF DEGREES OF FREEDOM OF CHI-SQUARED DISTRIBUTION .
!
                call random_number( df_vec(:ndf) )
!
                df_vec(:ndf) = real( ndf_max, stnd)*df_vec(:ndf)
!
!               WE TEST ONLY FOR NUMBER OF DEGREES OF FREEDOM GREATER THAN 20 .
!
                where( df_vec(:ndf)<c20 )  df_vec(:ndf) = df_vec(:ndf) + c20
!
                ndf_vec(:ndf) = int( df_vec(:ndf), i4b )
!
            case( 4_i4b )
!
!               GENERATE A MATRIX OF DEGREES OF FREEDOM OF CHI-SQUARED DISTRIBUTION .
!
                call random_number( df_mat(:n,:n) )
!
                df_mat(:n,:n) = real( ndf_max, stnd)*df_mat(:n,:n)
!
!               WE TEST ONLY FOR NUMBER OF DEGREES OF FREEDOM GREATER THAN 20 .
!
                where( df_mat(:n,:n)<c20 )  df_mat(:n,:n) = df_mat(:n,:n) + c20
!
                ndf_mat(:n,:n) = int( df_mat(:n,:n), i4b )
!
            case( 11_i4b )
!
!               SET THE REQUIRED PRECISION OF THE RESULTS FOR TESTS 11 TO 20.
!
                eps = 1.0e-4_stnd 
!
!               ADJUST THE NUMBERS OF DEGREES OF FREEDOM TO TEST NORMAL
!               APPROXIMATION OF CHI-SQUARED DISTRIBUTION .
!
                ndf_vec(:ndf) = ndf_vec(:ndf) + ndf_max
!
            case( 14_i4b )
!
!               ADJUST THE NUMBERS OF DEGREES OF FREEDOM TO TEST NORMAL
!               APPROXIMATION OF CHI-SQUARED DISTRIBUTION .
!
                ndf_mat(:n,:n) = ndf_mat(:n,:n) + ndf_max
!
            end select
!
            select case( i )
!
            case( 1_i4b, 6_i4b, 11_i4b, 16_i4b )
!
                if ( i==1_i4b .or. i==11_i4b ) then
!
!                   TESTING LOWER TAIL FOR TESTS 1 TO 5 AND 11 TO 15.
!
                    upper       = false
                    prob(:n,:n) = p(:n,:n)
!
                else
!
!                   TESTING UPPER TAIL FOR TESTS 6 TO 10 AND 16 TO 20.
!
                    upper       = true
!
!                   ADJUST THE PROBABILITIES.
!
                    prob(:n,:n) = one - p(:n,:n)
!
                end if
!
!               TESTS 1, 6, 11 AND 16 : MATRIX FORM OF FUNCTIONS pinvq AND probq.
!
                do l = 1_i4b, ndf
!
!                   COMPUTE THE QUANTILES X OF CHI-SQUARED DISTRIBUTION WITH ndf_vec(l) DEGREES OF FREEDOM
!                   CORRESPONDING TO LOWER TAIL AREAS OF prob(:n,:n) .
!
                    x(:n,:n) = pinvq( prob(:n,:n), ndf_vec(l) ) 
!
!                   RECOMPUTE THE PROBABILITIES FROM THE QUANTILES .
!
                    p2(:n,:n) = probq( x(:n,:n), ndf_vec(l), upper=upper, ndf_max=ndf_max )
!
!                   CHECK THAT p AND p2 AGREE.
!
                    err = max( err, maxval( abs( p(:n,:n) - p2(:n,:n) ) ) )
!
                end do
!
            case( 2_i4b, 7_i4b, 12_i4b, 17_i4b )
!
!               TESTS 2, 7, 12 AND 17 : VECTOR FORM OF FUNCTIONS pinvq AND probq.
!
                do l = 1_i4b, ndf
!
                    do j = 1_i4b, n
!
!                       COMPUTE THE QUANTILES X OF CHI-SQUARED DISTRIBUTION WITH ndf_vec(l) DEGREES
!                       OF FREEDOM CORRESPONDING TO LOWER TAIL AREAS OF prob(:n,j) .
!
                        x(:n,j) = pinvq( prob(:n,j), ndf_vec(l) )
!
!                       RECOMPUTE THE PROBABILITIES FROM THE QUANTILES .
!
                        p2(:n,j) = probq( x(:n,j), ndf_vec(l), upper=upper, ndf_max=ndf_max )
!
                    end do
!
!                   CHECK THAT p AND p2 AGREE.
!
                    err = max( err, maxval( abs( p(:n,:n) - p2(:n,:n) ) ) )
!
                end do
!
            case( 3_i4b, 8_i4b, 13_i4b, 18_i4b )
!
!               TESTS 3, 8, 13 AND 18 : SCALAR FORM OF FUNCTIONS pinvq AND probq.
!
                do l = 1_i4b, ndf
!
                    do j = 1_i4b, n
!
                        do k = 1_i4b, n
!
!                           COMPUTE THE QUANTILE X OF CHI-SQUARED DISTRIBUTION WITH ndf_vec(l) DEGREES
!                           OF FREEDOM CORRESPONDING TO LOWER TAIL AREA OF prob(k,j) .
!
                            x(k,j) = pinvq( prob(k,j), ndf_vec(l) )
!
!                           RECOMPUTE THE PROBABILITY FROM THE QUANTILE .
!
                            p2(k,j) = probq( x(k,j), ndf_vec(l), upper=upper, ndf_max=ndf_max )
!
                        end do
!
                    end do
!
!                   CHECK THAT p AND p2 AGREE.
!
                    err = max( err, maxval( abs( p(:n,:n) - p2(:n,:n) ) ) )
!
                end do
!
            case( 4_i4b, 9_i4b, 14_i4b, 19_i4b )
!
!               TESTS 4, 9, 14 AND 19 : MATRIX FORM OF FUNCTIONS pinvq AND probq.
!
!               COMPUTE THE QUANTILES X OF CHI-SQUARED DISTRIBUTION WITH ndf_mat(:n,:n) DEGREES
!               OF FREEDOM CORRESPONDING TO LOWER TAIL AREAS OF prob(:n,:n) .
!
                x(:n,:n) = pinvq( prob(:n,:n), ndf_mat(:n,:n) ) 
!
!               RECOMPUTE THE PROBABILITIES FROM THE QUANTILES .
!
                p2(:n,:n) = probq( x(:n,:n), ndf_mat(:n,:n), upper=upper, ndf_max=ndf_max )
!
!               CHECK THAT p AND p2 AGREE.
!
                err = maxval( abs( p(:n,:n) - p2(:n,:n) ) )
!
            case( 5_i4b, 10_i4b, 15_i4b, 20_i4b )
!
!               TESTS 5, 10, 15 AND 20 : VECTOR FORM OF FUNCTIONS pinvq AND probq.
!
                do j = 1_i4b, n
!
!                   COMPUTE THE QUANTILES X OF CHI-SQUARED DISTRIBUTION WITH ndf_mat(:n,j) DEGREES
!                   OF FREEDOM CORRESPONDING TO LOWER TAIL AREAS OF prob(:n,j) .
!
                    x(:n,j) = pinvq( prob(:n,j), ndf_mat(:n,j) )
!
!                   RECOMPUTE THE PROBABILITIES FROM THE QUANTILES .
!
                    p2(:n,j) = probq( x(:n,j), ndf_mat(:n,j), upper=upper, ndf_max=ndf_max )
!
                end do
!
!               CHECK THAT p AND p2 AGREE.
!
                err = maxval( abs( p(:n,:n) - p2(:n,:n) ) )
!
            end select
!
            if ( err<=eps ) then
!
                success = success + 1_i4b
!
            else
!
                write (prtunit,*) 'Test ',m,' is incorrect'
!
                write (prtunit,*) '=> ',err,' > ',eps,' , n= ', n,' , case= ', i,  &
                                  ' , upper= ', upper
!
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( p, p2, prob, x, df_mat, ndf_mat )
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( df_vec, ndf_vec )
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_pinvq
! =========================
!
end program test_pinvq
