program test_pinvq3
!
!
! Purpose
! =======
!
!   This program is intended to test functions PROBQ3 and PINVQ3 in module Prob_Procedures.
!                                                                              
!
! Further Details
! ===============
!                                                                              
! LATEST REVISION : 08/02/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, lgl, n1_def, n2_def, n3_def,    &
                         true, false, zero, half, one, two, c100,   &
                         allocate_error, merror, probq3, pinvq3
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def, ndf=10
!
    real(stnd), parameter :: df_max=c100
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
!    real(stnd),  parameter :: eps = 1.0e-4_stnd 
!
    character(len=*), parameter :: name_proc='pinvq3'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, ulp
    real(stnd), allocatable, dimension(:)   :: df_vec
    real(stnd), allocatable, dimension(:,:) :: prob, p, p2, x, df_mat
!
    integer(i4b) :: i, j, k, l, m, n, success
    integer      :: iok
!
    logical(lgl) :: upper
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
    ulp = epsilon( ulp )
!
!   ALLOCATE WORK ARRAY.
!
    allocate( df_vec(ndf), stat=iok  )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( p(n,n), p2(n,n), prob(n,n), x(n,n),  &
                  df_mat(n,n), stat=iok  )
!
        if ( iok/=0 ) cycle
!
!       GENERATE A RANDOM PROBABILITY MATRIX p .
!
        call random_number( p(:n,:n) )
!
        where( p(:n,:n)<=ulp .or. p(:n,:n)>=one-ulp )  p(:n,:n) = half
!
        do i = 1_i4b, 20_i4b
!
            m = m + 1_i4b
!
            err = zero
!
            select case( i )
!
            case( 1_i4b )
!
!               SET THE REQUIRED PRECISION OF THE RESULTS FOR TESTS 1 TO 10.
!
                eps = 1.0e-4_stnd 
!                eps = 1.0e-6_stnd 
!
!               GENERATE A VECTOR OF DEGREES OF FREEDOM OF CHI-SQUARED DISTRIBUTION .
!
                call random_number( df_vec(:ndf) )
!
                df_vec(:ndf) = (df_max-one)*df_vec(:ndf)
!
!               WE TEST ONLY FOR NUMBER OF DEGREES OF FREEDOM GREATER THAN 1 .
!
                where( df_vec(:ndf)<one )  df_vec(:ndf) = df_vec(:ndf) + one
!
            case( 4_i4b )
!
!               GENERATE A MATRIX OF DEGREES OF FREEDOM OF CHI-SQUARED DISTRIBUTION .
!
                call random_number( df_mat(:n,:n) )
!
                df_mat(:n,:n) = (df_max-one)*df_mat(:n,:n)
!
!               WE TEST ONLY FOR NUMBER OF DEGREES OF FREEDOM GREATER THAN 1 .
!
                where( df_mat(:n,:n)<one )  df_mat(:n,:n) = df_mat(:n,:n) + one
!
            case( 11_i4b )
!
!               SET THE REQUIRED PRECISION OF THE RESULTS FOR TESTS 11 TO 20.
!
                eps = 1.0e-4_stnd 
!
!               ADJUST THE NUMBERS OF DEGREES OF FREEDOM TO TEST NORMAL
!               APPROXIMATION OF CHI-SQUARED DISTRIBUTION .
!
                df_vec(:ndf) = df_vec(:ndf) + two*df_max
!
            case( 14_i4b )
!
!               ADJUST THE NUMBERS OF DEGREES OF FREEDOM TO TEST NORMAL
!               APPROXIMATION OF CHI-SQUARED DISTRIBUTION .
!
                df_mat(:n,:n) = df_mat(:n,:n) + two*df_max
!
            end select
!
            select case( i )
!
            case( 1_i4b, 6_i4b, 11_i4b, 16_i4b )
!
                if ( i==1_i4b .or. i==11_i4b ) then
!
!                   TESTING LOWER TAIL FOR TESTS 1 TO 5 AND 11 TO 15.
!
                    upper       = false
                    prob(:n,:n) = p(:n,:n)
!
                else
!
!                   TESTING UPPER TAIL FOR TESTS 6 TO 10 AND 16 TO 20.
!
                    upper       = true
!
!                   ADJUST THE PROBABILITIES.
!
                    prob(:n,:n) = one - p(:n,:n)
!
                end if
!
!               TESTS 1, 6, 11 AND 16 : MATRIX FORM OF FUNCTIONS pinvq3 AND probq3.
!
                do l = 1_i4b, ndf
!
!                   COMPUTE THE QUANTILES X OF CHI-SQUARED DISTRIBUTION WITH df_vec(l) DEGREES OF FREEDOM
!                   CORRESPONDING TO LOWER TAIL AREAS OF prob(:n,:n) .
!
                    x(:n,:n) = pinvq3( prob(:n,:n), df_vec(l), prec=0.5e-10_stnd ) 
!
!                   RECOMPUTE THE PROBABILITIES FROM THE QUANTILES .
!
                    p2(:n,:n) = probq3( x(:n,:n), df_vec(l), upper=upper, df_max=df_max )
!
!                   CHECK THAT p AND p2 AGREE.
!
                    err = max( err, maxval( abs( p(:n,:n) - p2(:n,:n) ) ) )
!
                end do
!
            case( 2_i4b, 7_i4b, 12_i4b, 17_i4b )
!
!               TESTS 2, 7, 12 AND 17 : VECTOR FORM OF FUNCTIONS pinvq3 AND probq3.
!
                do l = 1_i4b, ndf
!
                    do j = 1_i4b, n
!
!                       COMPUTE THE QUANTILES X OF CHI-SQUARED DISTRIBUTION WITH df_vec(l) DEGREES
!                       OF FREEDOM CORRESPONDING TO LOWER TAIL AREAS OF prob(:n,j) .
!
                        x(:n,j) = pinvq3( prob(:n,j), df_vec(l), prec=0.5e-10_stnd )
!
!                       RECOMPUTE THE PROBABILITIES FROM THE QUANTILES .
!
                        p2(:n,j) = probq3( x(:n,j), df_vec(l), upper=upper, df_max=df_max )
!
                    end do
!
!                   CHECK THAT p AND p2 AGREE.
!
                    err = max( err, maxval( abs( p(:n,:n) - p2(:n,:n) ) ) )
!
                end do
!
            case( 3_i4b, 8_i4b, 13_i4b, 18_i4b )
!
!               TESTS 3, 8, 13 AND 18 : SCALAR FORM OF FUNCTIONS pinvq3 AND probq3.
!
                do l = 1_i4b, ndf
!
                    do j = 1_i4b, n
!
                        do k = 1_i4b, n
!
!                           COMPUTE THE QUANTILE X OF CHI-SQUARED DISTRIBUTION WITH df_vec(l) DEGREES
!                           OF FREEDOM CORRESPONDING TO LOWER TAIL AREA OF prob(k,j) .
!
                            x(k,j) = pinvq3( prob(k,j), df_vec(l), prec=0.5e-10_stnd )
!
!                           RECOMPUTE THE PROBABILITY FROM THE QUANTILE .
!
                            p2(k,j) = probq3( x(k,j), df_vec(l), upper=upper, df_max=df_max )
!
                        end do
!
                    end do
!
!                   CHECK THAT p AND p2 AGREE.
!
                    err = max( err, maxval( abs( p(:n,:n) - p2(:n,:n) ) ) )
!
                end do
!
            case( 4_i4b, 9_i4b, 14_i4b, 19_i4b )
!
!               TESTS 4, 9, 14 AND 19 : MATRIX FORM OF FUNCTIONS pinvq3 AND probq3.
!
!               COMPUTE THE QUANTILES X OF CHI-SQUARED DISTRIBUTION WITH df_mat(:n,:n) DEGREES
!               OF FREEDOM CORRESPONDING TO LOWER TAIL AREAS OF prob(:n,:n) .
!
                x(:n,:n) = pinvq3( prob(:n,:n), df_mat(:n,:n), prec=0.5e-10_stnd ) 
!
!               RECOMPUTE THE PROBABILITIES FROM THE QUANTILES .
!
                p2(:n,:n) = probq3( x(:n,:n), df_mat(:n,:n), upper=upper, df_max=df_max )
!
!               CHECK THAT p AND p2 AGREE.
!
                err = maxval( abs( p(:n,:n) - p2(:n,:n) ) )
!
            case( 5_i4b, 10_i4b, 15_i4b, 20_i4b )
!
!               TESTS 5, 10, 15 AND 20 : VECTOR FORM OF FUNCTIONS pinvq3 AND probq3.
!
                do j = 1_i4b, n
!
!                   COMPUTE THE QUANTILES X OF CHI-SQUARED DISTRIBUTION WITH df_mat(:n,j) DEGREES
!                   OF FREEDOM CORRESPONDING TO LOWER TAIL AREAS OF prob(:n,j) .
!
                    x(:n,j) = pinvq3( prob(:n,j), df_mat(:n,j), prec=0.5e-10_stnd )
!
!                   RECOMPUTE THE PROBABILITIES FROM THE QUANTILES .
!
                    p2(:n,j) = probq3( x(:n,j), df_mat(:n,j), upper=upper, df_max=df_max )
!
                end do
!
!               CHECK THAT p AND p2 AGREE.
!
                err = maxval( abs( p(:n,:n) - p2(:n,:n) ) )
!
            end select
!
            if ( err<=eps ) then
!
                success = success + 1_i4b
!
            else
!
                write (prtunit,*) 'Test ',m,' is incorrect'
!
                write (prtunit,*) '=> ',err,' > ',eps,' , n= ', n,' , case= ', i,  &
                                  ' , upper= ', upper
!
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( p, p2, prob, x, df_mat )
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( df_vec )
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_pinvq3
! ==========================
!
end program test_pinvq3
