program test_pinvstudent
!
!
! Purpose
! =======
!
!   This program is intended to test functions PROBSTUDENT and PINVSTUDENT
!   in module Prob_Procedures.
!                                                                              
!
! Further Details
! ===============
!                                                                              
! LATEST REVISION : 08/02/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, n1_def, n2_def, n3_def,              &
                         zero, half, one, c99, allocate_error, merror,   &
                         probstudent, pinvstudent
                         
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def, ndf_max=50
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    real(stnd),  parameter :: eps = 1.0e-3_stnd 
!
    character(len=*), parameter :: name_proc='pinvstudent'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err
    real(stnd), allocatable, dimension(:)   :: df_vec
    real(stnd), allocatable, dimension(:,:) :: p, p2, t, df_mat
!
    integer(i4b) :: i, j, k, l, m, n, success
    integer      :: iok
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( df_vec(ndf_max), stat=iok  )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( p(n,n), p2(n,n), t(n,n), df_mat(n,n), stat=iok  )
!
        if ( iok/=0 ) cycle
!
!       GENERATE A RANDOM PROBABILITY MATRIX p .
!
        call random_number( p(:n,:n) )
!
        where( p(:n,:n)==zero .or. p(:n,:n)==one )  p(:n,:n) = half
!
        do i = 1_i4b, 5_i4b
!
            m = m + 1_i4b
!
            err = zero
!
            if ( i==1_i4b ) then
!
!               GENERATE A VECTOR OF DEGREES OF FREEDOM OF STUDENT T-DISTRIBUTION .
!
                call random_number( df_vec(:ndf_max) )
!
                df_vec(:ndf_max) = c99*df_vec(:ndf_max)
!
                where( df_vec(:ndf_max)<one )  df_vec(:ndf_max) = df_vec(:ndf_max) + one
!
                df_vec(:ndf_max) = int( df_vec(:ndf_max), i4b )
!
            else if ( i==4_i4b ) then
!
!               GENERATE A MATRIX OF DEGREES OF FREEDOM OF STUDENT T-DISTRIBUTION .
!
                call random_number( df_mat(:n,:n) )
!
                df_mat(:n,:n) = c99*df_mat(:n,:n)
!
                where( df_mat(:n,:n)<one )  df_mat(:n,:n) = df_mat(:n,:n) + one
!
                df_mat(:n,:n) = int( df_mat(:n,:n), i4b )
!
            end if
!
            select case( i )
!
            case( 1_i4b )
!
!               TEST 1 : MATRIX FORM OF FUNCTIONS pinvstudent AND probstudent .
!
                do l = 1_i4b, ndf_max
!
!                   COMPUTE THE TWO-TAIL QUANTILES T OF STUDENT T-DISTRIBUTION 
!                   WITH df_vec(l) DEGREES OF FREEDOM CORRESPONDING TO AREAS OF p(:n,:n) .
!
                    t(:n,:n) = pinvstudent( p(:n,:n), df_vec(l) ) 
!
!                   RECOMPUTE THE PROBABILITIES FROM THE QUANTILES .
!
                    p2(:n,:n) = probstudent( t(:n,:n), df_vec(l) )
!
!                   CHECK THAT p AND p2 AGREE.
!
                    err = max( err, maxval( abs( p(:n,:n) - p2(:n,:n) ) ) )
!
                end do
!
            case( 2_i4b )
!
!               TEST 2 : VECTOR FORM OF FUNCTIONS pinvstudent AND probstudent .
!
                do l = 1_i4b, ndf_max
!
                    do j = 1_i4b, n
!
!                       COMPUTE THE TWO-TAIL QUANTILES t(:n,j) OF STUDENT T-DISTRIBUTION 
!                       WITH df_vec(l) DEGREES OF FREEDOM CORRESPONDING TO AREAS OF p(:n,j) .
!
                        t(:n,j) = pinvstudent( p(:n,j), df_vec(l) )
!
!                       RECOMPUTE THE PROBABILITIES FROM THE QUANTILES .
!
                        p2(:n,j) = probstudent( t(:n,j), df_vec(l) )
!
                    end do
!
!                   CHECK THAT p AND p2 AGREE.
!
                    err = max( err, maxval( abs( p(:n,:n) - p2(:n,:n) ) ) )
!
                end do
!
            case( 3_i4b )
!
!               TEST 3 : SCALAR FORM OF FUNCTIONS pinvstudent AND probstudent .
!
                do l = 1_i4b, ndf_max
!
                    do j = 1_i4b, n
!
                        do k = 1_i4b, n
!
!                           COMPUTE THE TWO-TAIL QUANTILE t(k,j) OF STUDENT T-DISTRIBUTION 
!                           WITH df_vec(l) DEGREES OF FREEDOM CORRESPONDING TO AREA OF p(k,j) .
!
                            t(k,j) = pinvstudent( p(k,j), df_vec(l) )
!
!                           RECOMPUTE THE PROBABILITY FROM THE QUANTILE .
!
                            p2(k,j) = probstudent( t(k,j), df_vec(l) )
!
                        end do
!
                    end do
!
!                   CHECK THAT p AND p2 AGREE.
!
                    err = max( err, maxval( abs( p(:n,:n) - p2(:n,:n) ) ) )
!
                end do
!
            case( 4_i4b )
!
!               TEST 4 : MATRIX FORM OF FUNCTIONS pinvstudent AND probstudent .
!
!               COMPUTE THE TWO-TAIL QUANTILES T OF STUDENT T-DISTRIBUTION 
!               WITH df_mat(:n,:n) DEGREES OF FREEDOM CORRESPONDING TO AREAS OF p(:n,:n) .
!
                t(:n,:n) = pinvstudent( p(:n,:n), df_mat(:n,:n) ) 
!
!               RECOMPUTE THE PROBABILITIES FROM THE QUANTILES .
!
                p2(:n,:n) = probstudent( t(:n,:n), df_mat(:n,:n) )
!
!               CHECK THAT p AND p2 AGREE.
!
                err = maxval( abs( p(:n,:n) - p2(:n,:n) ) )
!
            case( 5_i4b )
!
!               TEST 5 : VECTOR FORM OF FUNCTIONS pinvstudent AND probstudent .
!
                do j = 1_i4b, n
!
!                   COMPUTE THE TWO-TAIL QUANTILES t(:n,j) OF STUDENT T-DISTRIBUTION 
!                   WITH df_mat(:n,j) DEGREES OF FREEDOM CORRESPONDING TO AREAS OF p(:n,j) .
!
                    t(:n,j) = pinvstudent( p(:n,j), df_mat(:n,j) )
!
!                   RECOMPUTE THE PROBABILITIES FROM THE QUANTILES .
!
                    p2(:n,j) = probstudent( t(:n,j), df_mat(:n,j) )
!
                end do
!
!               CHECK THAT p AND p2 AGREE.
!
                err = maxval( abs( p(:n,:n) - p2(:n,:n) ) )
!
            end select
!
            if ( err<=eps ) then
!
                success = success + 1_i4b
!
            else
!
                write (prtunit,*) 'Test ',m,' is incorrect'
!
                write (prtunit,*) '=> ',err,' > ',eps,' , n= ', n,' , case= ', i
!
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( p, p2, t, df_mat )
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( df_vec )
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_pinvstudent
! ===============================
!
end program test_pinvstudent
