program test_probbeta
!
!
! Purpose
! =======
!
!   This program is intended to test functions PROBBETA and  PINVBETA in module Prob_Procedures.
!                                                                              
!                                                                              
! LATEST REVISION : 21/05/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, n1_def, n2_def, n3_def,        &
                         zero, half, c20, allocate_error, merror,  &
                         probbeta, pinvbeta
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def,    &
                               na=20, nb=20, nprob=50
!
    real(stnd),  parameter :: eps = 1.0e-4_stnd 
!
    character(len=*), parameter :: name_proc='probbeta'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                                :: err
    real(stnd), allocatable, dimension(:)     :: a, b, prob
    real(stnd), allocatable, dimension(:,:)   :: p, x
    real(stnd), allocatable, dimension(:,:,:) :: xx
!
    integer(i4b) :: i, j, k, l, m, n, success
    integer      :: iok
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( a(na), b(nb), xx(nprob,na,nb), prob(nprob), stat=iok  )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( p(nprob,n), x(nprob,n), stat=iok  )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       GENERATE THE a AND b PARAMETERS OF THE BETA DISTRIBUTION .
!
        call random_number( a(:na) )
!
        a(:na) = c20*a(:na)
!
        call random_number( b(:nb) )
!
        b(:nb) = c20*b(:nb)
!
!       TEST ONLY FOR VALUES OF a AND b GREATER THAN 0.5 SINCE
!       pinvbeta IS INACCURATE FOR SMALL VALUES of a AND b.
!
        where( a(:na)<=half )  a(:na) = a(:na) + half
        where( b(:nb)<=half )  b(:nb) = b(:nb) + half
!
!       GENERATE A RANDOM PROBABILITY VECTOR prob(:nprob) .
!
        call random_number( prob(:nprob) )
!
!       COMPUTE THE REQUIRED LOWER TAILS QUANTILES.
!
        do i = 1_i4b, nb
!
            do j = 1_i4b, na
!
                do l = 1_i4b, nprob
!
                    xx(l,j,i) = pinvbeta( prob(l), a(j), b(i) )
!
                end do
!
            end do
!
        end do
!
!       RECOMPUTE THE PROBABILITIES FROM THE BETA DEVIATES.
!
        do k = 1_i4b, 3_i4b
!
            m = m + 1_i4b
!
            err = zero
!
            select case( k )
!
            case( 1_i4b )
!
!               TEST 1 : SCALAR FORM OF probbeta GENERIC FUNCTION.
!
                do i = 1_i4b, nb
!
                    do j = 1_i4b, na
!
                        do l = 1_i4b, nprob
!
                            p(1_i4b,1_i4b) = probbeta( xx(l,j,i), a(j), b(i) )
!
!                           CHECK THAT p(1,1) AND prob(l) AGREE.
!
                            err = max( abs( p(1_i4b,1_i4b) - prob(l) ), err )
!
                        end do
!
                    end do
!
                end do
!
            case( 2_i4b )
!
!               TEST 2 : VECTOR FORM OF probbeta GENERIC FUNCTION.
!
                do i = 1_i4b, nb
!
                    do j = 1_i4b, na
!
                        p(:nprob,1_i4b) = probbeta( xx(:nprob,j,i), a(j), b(i) )
!
!                       CHECK THAT p(:nprob,1) AND prob(:nprob) AGREE.
!
                        err = max( maxval( abs(  p(:nprob,1_i4b) - prob(:nprob) ) ), err )
!
                    end do
!
                end do
!
            case( 3_i4b )
!
!               TEST 3 : MATRIX FORM OF probbeta GENERIC FUNCTION.
!
                do i = 1_i4b, nb
!
                    do j = 1_i4b, na
!
                        x(:nprob,:n) = spread( xx(:nprob,j,i), dim=2, ncopies=n )
                        p(:nprob,:n) = probbeta( x(:nprob,:n), a(j), b(i) )
!
!                       CHECK THAT EACH COLUMN OF p(:nprob,:n) AGREES WITH prob(:nprob).
!
                        x(:nprob,:n) = spread( prob(:nprob), dim=2, ncopies=n )
!
                        err = max( maxval( abs(  p(:nprob,:n) - x(:nprob,:n) ) ), err )
!
                    end do
!
                end do
!
            end select
!
!           PRINT THE RESULTS OF THE TESTS.
!
            if ( err<=eps ) then
!
                success = success + 1_i4b
!
            else
!
                write (prtunit,*) 'Test ',m,' is incorrect'
!
                write (prtunit,*) '=> ',err,' > ',eps,' , n= ', n,' , case= ', k
!
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( p, x )
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( xx, a, b, prob )
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_probbeta
! ============================
!
end program test_probbeta
