program test_probf2
!
!
! Purpose
! =======
!
!   This program is intended to test functions PROBF2 and PINVF2 in module Prob_Procedures.
!                                                                              
!                                                                              
! LATEST REVISION : 08/02/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, lgl, n1_def, n2_def, n3_def,  &
                         true, false, zero, half, one, c100,      &
                         allocate_error, merror, probf2, pinvf2
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def,    &
                               ndf1=20, ndf2=20, nprob=50
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    real(stnd),  parameter :: eps = 1.0e-3_stnd 
!
    character(len=*), parameter :: name_proc='probf2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                                :: err, ulp
    real(stnd), allocatable, dimension(:)     :: df1, df2, prob
    real(stnd), allocatable, dimension(:,:)   :: p, x
    real(stnd), allocatable, dimension(:,:,:) :: xx
!
    integer(i4b) :: i, j, k, l, m, n, success
    integer      :: iok
!   
    logical(lgl) :: upper
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
    ulp = epsilon( ulp )
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( df1(ndf1), df2(ndf2), xx(nprob,ndf1,ndf2), prob(nprob), stat=iok  )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( p(nprob,n), x(nprob,n), stat=iok  )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       GENERATE THE NUMBERS OF DEGREES OF FREEDOM OF THE F DISTRIBUTION.
!
        call random_number( df1(:ndf1) )
!
        df1(:ndf1) = c100*df1(:ndf1)
!
        call random_number( df2(:ndf2) )
!
        df2(:ndf2) = c100*df2(:ndf2)
!
!       TEST ONLY FOR VALUES OF NUMBERS OF DEGREES OF FREEDOM GREATER THAN ONE SINCE
!       pinvbeta IS INACCURATE FOR SMALL VALUES of a AND b LESS THAN 0.5.
!
        where( df1(:ndf1)<=one )  df1(:ndf1) = df1(:ndf1) + one
        where( df2(:ndf2)<=one )  df2(:ndf2) = df2(:ndf2) + one
!
!       GENERATE A RANDOM PROBABILITY VECTOR prob(:nprob) .
!
        call random_number( prob(:nprob) )
!
        where( prob(:nprob)<ulp .or. prob(:nprob)>one-ulp )  prob(:nprob) = half
!
!       COMPUTE THE REQUIRED TAILS QUANTILES.
!
        do i = 1_i4b, ndf2
!
            do j = 1_i4b, ndf1
!
                do l = 1_i4b, nprob
!
                    xx(l,j,i) = pinvf2( prob(l), df1(j), df2(i) )
!
                end do
!
            end do
!
        end do
!
!       RECOMPUTE THE PROBABILITIES FROM THE BETA DEVIATES.
!
        do k = 1_i4b, 6_i4b
!
            m = m + 1_i4b
!
            err = zero
!
            select case( k )
!
            case( 1_i4b, 4_i4b )
!
                if ( k==1_i4b ) then
!
!                   TESTING LOWER TAIL FOR TESTS 1 TO 3.
!
                    upper = false
!
                else
!
!                   TESTING UPPER TAIL FOR TESTS 4 TO 6.
!
                    upper = true
!
!                   ADJUST THE PROBABILITIES.
!
                    prob(:nprob) = one - prob(:nprob)
!
                end if
!
!               TESTS 1 AND 4 : SCALAR FORM OF probf2 GENERIC FUNCTION.
!
                do i = 1_i4b, ndf2
!
                    do j = 1_i4b, ndf1
!
                        do l = 1_i4b, nprob
!
                            p(1_i4b,1_i4b) = probf2( xx(l,j,i), df1(j), df2(i), upper )
!
!                           CHECK THAT p(1,1) AND prob(l) AGREE.
!
                            err = max( abs( p(1_i4b,1_i4b) - prob(l) ), err )
!
                        end do
!
                    end do
!
                end do
!
            case( 2_i4b, 5_i4b )
!
!               TESTS 2 AND 5 : VECTOR FORM OF probf2 GENERIC FUNCTION.
!
                do i = 1_i4b, ndf2
!
                    do j = 1_i4b, ndf1
!
                        p(:nprob,1_i4b) = probf2( xx(:nprob,j,i), df1(j), df2(i), upper )
!
!                       CHECK THAT p(:nprob,1) AND prob(:nprob) AGREE.
!
                        err = max( maxval( abs(  p(:nprob,1_i4b) - prob(:nprob) ) ), err )
!
                    end do
!
                end do
!
            case( 3_i4b, 6_i4b )
!
!               TESTS 3 AND 6 : MATRIX FORM OF probf2 GENERIC FUNCTION.
!
                do i = 1_i4b, ndf2
!
                    do j = 1_i4b, ndf1
!
                        x(:nprob,:n) = spread( xx(:nprob,j,i), dim=2, ncopies=n )
                        p(:nprob,:n) = probf2( x(:nprob,:n), df1(j), df2(i), upper )
!
!                       CHECK THAT EACH COLUMN OF p(:nprob,:n) AGREES WITH prob(:nprob).
!
                        x(:nprob,:n) = spread( prob(:nprob), dim=2, ncopies=n )
!
                        err = max( maxval( abs(  p(:nprob,:n) - x(:nprob,:n) ) ), err )
!
                    end do
!
                end do
!
            end select
!
!           PRINT THE RESULTS OF THE TESTS.
!
            if ( err<=eps ) then
!
                success = success + 1_i4b
!
            else
!
                write (prtunit,*) 'Test ',m,' is incorrect'
!
                write (prtunit,*) '=> ',err,' > ',eps,' , n= ', n,' , case= ', k,  &
                                  ' , upper= ', upper
!
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( p, x )
!
    end do
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( xx, df1, df2, prob )
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_probf2
! ==========================
!
end program test_probf2
