program test_product_svd_cmp
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine PRODUCT_SVD_CMP in module SVD_Procedures. 
!                                                                              
! LATEST REVISION : 28/05/2021
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,   &
                         zero, half, one, c100, product_svd_cmp, norm, lamch
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit = 6, n1=n1_def, n2=n2_def, n3=n3_def
!   
    real(stnd), parameter  :: fudge=c100
!
    character(len=*), parameter :: name_proc='product_svd_cmp'
!   
!   
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, err1, err2, err3, eps, ulp, safmin, &
                                               tmp, norma
    real(stnd), allocatable, dimension(:)   :: singval
    real(stnd), allocatable, dimension(:,:) :: a, leftvec, rightvec, singvect, res
!
    integer(i4b) :: k, l, n, nr, nc, krank, m, success, notdone
    integer      :: iok
!
    logical(lgl) :: failure, details
!   
    character :: sort='a'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    eps    = fudge*ulp
    safmin = lamch( 'S' )
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
        do k = 1_i4b, 5_i4b
!
            select case( k )
!
                case( 1_i4b )
!
!                   TEST 1: SVD OF a(:nr,:nc) WITH nr=nc .
!
                    nr = n
                    nc = n
!
                case( 2_i4b )
!
!                   TEST 2: SVD OF a(:nr,:nc) WITH nr>=nc AND nr NOT MUCH LARGER THAN nc .
!
                    nr = n + 1_i4b
                    nc = n
!
                case( 3_i4b )
!
!                   TEST 3: SVD OF a(:nr,:nc) WITH nr>=nc AND nr MUCH LARGER THAN nc .
!
                    nr = n + n
                    nc = n
!
                case( 4_i4b )
!
!                   TEST 4: SVD OF a(:nr,:nc) WITH nc>=nr AND nc NOT MUCH LARGER THAN nr .
!
                    nr = n
                    nc = n + 1_i4b
!
                case( 5_i4b )
!
!                   TEST 5: SVD OF a(:nr,:nc) WITH nc>=nr AND nc MUCH LARGER THAN nr .
!
                    nr = n
                    nc = n + n
!
            end select
!
            call random_number( tmp )
!
            krank = real( n/2_i4b, stnd)*tmp
!
            if ( krank==0_i4b ) krank = 1_i4b
!
!           ALLOCATE WORK ARRAYS.
!
            l = max( nr, nc )
!
            allocate( a(nr,nc), res(nr,krank),                 &
                      leftvec(nr,krank), rightvec(nc,krank),   &
                      singvect(krank,l), singval(krank),       &
                      stat=iok                                 )
!
            if ( iok/=0 ) cycle
!
!           GENERATE A RIGHT RANDOM COEFFICIENT MATRIX leftvec(:nr,:krank) .
!
            call random_number( leftvec(:nr,:krank) )
!
            leftvec(:nr,:krank) = leftvec(:nr,:krank) - half
!
            call random_number( tmp )
!
            if ( tmp>safmin ) then
                leftvec(:nr,:krank) = leftvec(:nr,:krank)/tmp
            end if
!
!           GENERATE A LEFT RANDOM COEFFICIENT MATRIX rightvec(:nc,:krank) .
!
            call random_number( rightvec(:nc,:krank) )
!
            rightvec(:nc,:krank) = rightvec(:nc,:krank) - half
!
            call random_number( tmp )
!
            if ( tmp>safmin ) then
                rightvec(:nc,:krank) = rightvec(:nc,:krank)/tmp
            end if
!
!           COMPUTE THE MATRIX PRODUCT.
!
            singvect(:krank,:nc) = transpose( rightvec(:nc,:krank) )
!
            a(:nr,:nc) = matmul( leftvec(:nr,:krank), singvect(:krank,:nc) )
!
!           COMPUTE SVD OF PRODUCT OF RANDOM MATRICES .
!
            call product_svd_cmp( leftvec(:nr,:krank), rightvec(:nc,:krank), singval(:krank), failure, sort )
!
            if ( failure ) then
!
                notdone = notdone + 1_i4b
!
!               DEALLOCATE WORK ARRAYS.
!
                deallocate( a, leftvec, rightvec, singvect, singval, res )
!
                cycle
!
            end if
!
            m = m + 1_i4b
!
!           COMPUTE THE NORM OF THE MATRIX PRODUCT.
!
            norma = norm( a(:nr,:nc) )
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*v - u*s
!           WHERE s ARE THE SINGULAR VALUES, u THE LEFT SINGULAR VECTORS AND
!           v THE RIGHT SINGULAR VECTORS OF THE MATRIX PRODUCT.
!
            res(:nr,:krank) = matmul( a(:nr,:nc), rightvec(:nc,:krank) )
!
            do l = 1_i4b, krank
                res(:nr,l) = res(:nr,l) - singval(l)*leftvec(:nr,l)
            end do
!
!            res(:nr,:krank) = matmul( a(:nr,:nc), rightvec(:nc,:krank) )         &
!                              - leftvec(:nr,:krank)*spread(singval(:krank),1,nr)
!
            if ( norma/=zero ) then
                err1 = maxval( norm(res(:nr,:krank), dim=2_i4b) )/( norma*real(krank,stnd) )
            else
                err1 = maxval( norm(res(:nr,:krank), dim=2_i4b) )/real(krank,stnd)
            end if
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!           WHERE u are LEFT SINGULAR VECTORS OF THE MATRIX PRODUCT.
!
            singvect(:krank,:nr) = transpose( leftvec(:nr,:krank) )
!
            res(:krank,:krank) = matmul( singvect(:krank,:nr), leftvec(:nr,:krank) )
!
            do l = 1_i4b, krank
                res(l,l) = one - res(l,l)
            end do
!
            err2 = maxval( norm(res(:krank,:krank), dim=2_i4b) )/real(krank,stnd)
!
!           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - v**(t)*v
!           WHERE v are RIGHT SINGULAR VECTORS OF THE MATRIX PRODUCT.
!
            singvect(:krank,:nc) = transpose( rightvec(:nc,:krank) )
!
            res(:krank,:krank) = matmul( singvect(:krank,:nc), rightvec(:nc,:krank) )
!
            do l = 1_i4b, krank
                res(l,l) = one - res(l,l)
            end do
!
            err3 = maxval( norm(res(:krank,:krank), dim=2_i4b) )/real(krank,stnd)
!
            err = max( err1, err2, err3 )
!
            if ( err<=eps ) then
!
                success = success + 1_i4b
!
                if ( details ) then
!
                    write (prtunit,*) 'Test ',m,' is correct'
!
                    write (prtunit,*) 'nr= ', nr,' , nc= ', nc,' , krank= ', krank,' , case= ', k
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*v - u*s
!                   WHERE s ARE THE SINGULAR VALUES, u THE LEFT SINGULAR VECTORS AND
!                   v THE RIGHT SINGULAR VECTORS OF THE MATRIX PRODUCT.
!
                    write (prtunit,*) 'Accuracy of the computed singular triplets             = ', &
                                      err1/ulp, ' (must be less than :', fudge, ' )'
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!                   WHERE u are LEFT SINGULAR VECTORS OF THE MATRIX PRODUCT.
!
                    write (prtunit,*) 'Orthogonality of the computed left singular vectors    = ', &
                                      err2/ulp, ' (must be less than :', fudge, ' )'
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - v**(t)*v
!                   WHERE v are RIGHT SINGULAR VECTORS OF THE MATRIX PRODUCT.
!
                    write (prtunit,*) 'Orthogonality of the computed right singular vectors   = ', &
                                      err3/ulp, ' (must be less than :', fudge, ' )'
!
                    write (prtunit,*)
!
                end if
!
            else
!
                write (prtunit,*) 'Test ',m,' is incorrect'
!
                write (prtunit,*) '=> ',err,' > ',eps,' , nr= ', nr,' , nc= ', nc,   &
                                  ' , krank= ', krank,' , case= ', k
!
!               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*v - u*s
!               WHERE s ARE THE SINGULAR VALUES, u THE LEFT SINGULAR VECTORS AND
!               v THE RIGHT SINGULAR VECTORS OF THE MATRIX PRODUCT.
!
                write (prtunit,*) 'Accuracy of the computed singular triplets             = ', &
                                  err1/ulp, ' (must be less than :', fudge, ' )'
!
!               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!               WHERE u are LEFT SINGULAR VECTORS OF THE MATRIX PRODUCT.
!
                write (prtunit,*) 'Orthogonality of the computed left singular vectors    = ', &
                                  err2/ulp, ' (must be less than :', fudge, ' )'
!
!               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - v**(t)*v
!               WHERE v are RIGHT SINGULAR VECTORS OF THE MATRIX PRODUCT.
!
                write (prtunit,*) 'Orthogonality of the computed right singular vectors   = ', &
                                  err3/ulp, ' (must be less than :', fudge, ' )'
!
                write (prtunit,*)
!
            end if
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( a, leftvec, rightvec, singvect, singval, res )
!
        end do
!
    end do
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_product_svd_cmp
! ===================================
!
end program test_product_svd_cmp
