program test_rsvd_cmp_fixed_precision
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine RSVD_CMP_FIXED_PRECISION in module SVD_Procedures.
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    Random matrices with given singular value distributions are generated for checking the accuracy
!    of subroutine RSVD_CMP_FIXED_PRECISION.
!
! LATEST REVISION : 25/05/2021
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def, true, false,          &
                         zero, half, one, two, five, seven, ten, c30, c100,            &
                         qr_cmp, ortho_gen_qr, singval_sort, rsvd_cmp_fixed_precision, &
                         norm, arth, lamch, random_seed_, random_number_,              &
                         normal_random_number3_
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT,
! relerr0 IS THE REQUESTED RELATIVE ERROR OF THE RESTRICTED SVD IN FROBENIUS NORM,
! niter IS THE NUMBER OF RANDOM POWER OR SUBSPACE PERFORMED IN THE FIRST PHASE OF THE QB FACTORIZATION,
! niter_qb IS THE NUMBER OF SUBSPACE PERFORMED IN THE LAST PHASE OF THE QB FACTORIZATION,
! def_conda IS THE APPROXIMATE CONDITION NUMBER OF THE GENERATED MATRIX IN SOME CASES.
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def, niter=5, niter_qb=20 
!   
    real(stnd), parameter  :: fudge=ten, def_conda=50000._stnd, relerr0=0.5_stnd
!
    character(len=*), parameter :: name_proc='rsvd_cmp_fixed_precision'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err3, err, eps, ulp, sqrt_ulp,     &
                                               safmin, tmp, norma, conda, relerr, relerr2
    real(stnd), allocatable, dimension(:)   :: eigval, singval, diagr, beta
    real(stnd), allocatable, dimension(:,:) :: a, leftvec, rightvec, rightvect, res
!
    real(stnd), dimension(:),   pointer :: singval2
    real(stnd), dimension(:,:), pointer :: leftvec2, rightvec2
!
    integer      :: iok
    integer(i4b) :: i, j, k, l, n, nr, nc, m, nsvd0, success, notdone,   &
                    nsvd, blk_size, maxiter_qb
    logical(lgl) :: failure_relerr, failure, details, ortho, reortho
!   
    character    :: sort='d'
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
    call random_seed_( alg=3 )
!
!   RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
    call random_seed_()
!
!   NULLIFY THE POINTERS singval2, leftvec2 AND rightvec2 SO THAT THEIR STATUT CAN
!   BE CHECKED INSIDE rsvd_cmp_fixed_precision SUBROUTINE.
!
    nullify( singval2, leftvec2, rightvec2 )
!
!   CHOOSE TUNING PARAMETERS FOR THE RANDOMIZED SVD ALGORITHM.
!   nsvd0 = blk_size*maxiter_qb IS THE MAXIMUM ALLOWABLE RANK OF THE
!   PARTIAL SVD, WHICH IS SOUGHT.
!
    blk_size   = 5_i4b
    maxiter_qb = 10_i4b
!
    nsvd0 = blk_size*maxiter_qb
!
!   SPECIFY IF ORTHONORMALIZATION IS CARRIED OUT BETWEEN EACH STEP
!   OF THE (POWER) SUBSPACE ITERATIONS, TO AVOID LOSS OF ACCURACY DUE
!   TO ROUNDING ERRORS.
!
    ortho = true
!
!   SPECIFY IF REORTHONORMALIZATION IS CARRIED OUT TO AVOID LOSS OF ORTHOGONALITY
!   IN THE BLOCK GRAM-SCHMIDT ORTHOGONALISATION SCHEME USED TO BUILD THE ORTHOGONAL
!   MATRIX OF THE QB DECOMPOSITION OF THE INPUT MATRIX.
!
    reortho = true
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp      = epsilon( err )
    sqrt_ulp = sqrt ( ulp )
    safmin   = lamch( 'S' )
!
!   SET THE TOLERANCE FOR THE ERROR OF THE RELATIVE ERROR IN FROBENIUS NORM.
!
    eps = 0.005_stnd
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( eigval(n), singval(n), diagr(n), beta(n), stat=iok  )
!
        if ( iok/=0 ) then
!
            if ( allocated(eigval) )    deallocate( eigval )
            if ( allocated(singval) )   deallocate( singval )
            if ( allocated(diagr) )     deallocate( diagr )
            if ( allocated(beta) )      deallocate( beta )
!
            cycle
!
        end if
!
        do k = 1_i4b, 5_i4b
!
            select case( k )
!
                case( 1_i4b )
!
!                   TEST 1: SVD OF a(:nr,:nc) WITH nr=nc .
!
                    nr = n
                    nc = n
!
                case( 2_i4b )
!
!                   TEST 2: SVD OF a(:nr,:nc) WITH nr>=nc AND nr NOT MUCH LARGER THAN nc .
!
                    nr = n + 1_i4b
                    nc = n
!
                case( 3_i4b )
!
!                   TEST 3: SVD OF a(:nr,:nc) WITH nr>=nc AND nr MUCH LARGER THAN nc .
!
                    nr = n + n
                    nc = n
!
                case( 4_i4b )
!
!                   TEST 4: SVD OF a(:nr,:nc) WITH nc>=nr AND nc NOT MUCH LARGER THAN nr .
!
                    nr = n
                    nc = n + 1_i4b
!
                case( 5_i4b )
!
!                   TEST 5: SVD OF a(:nr,:nc) WITH nc>=nr AND nc MUCH LARGER THAN nr .
!
                    nr = n
                    nc = n + n
!
            end select
!
!           ALLOCATE WORK ARRAYS.
!
            allocate( a(nr,nc), res(nr,nsvd0),       &
                      leftvec(nr,n), rightvec(nc,n), &
                      rightvect(n,nc), stat=iok  )
!
            if ( iok/=0 ) then
!
                if ( allocated(a) )         deallocate( a )
                if ( allocated(res) )       deallocate( res )
                if ( allocated(leftvec) )   deallocate( leftvec )
                if ( allocated(rightvec) )  deallocate( rightvec )
                if ( allocated(rightvect) ) deallocate( rightvect )
!
                cycle
!
            end if
!
!           GENERATE A RANDOM COEFFICIENT MATRIX leftvec(:nr,:n) .
!
            call random_number_( leftvec(:nr,:n) )
!
            leftvec(:nr,:n) = leftvec(:nr,:n) - half
!
            call random_number_( tmp )
!
            if ( tmp>safmin ) then
                leftvec(:nr,:n) = leftvec(:nr,:n)/tmp
            end if
!
!           COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
            call qr_cmp( leftvec(:nr,:n), diagr(:n), beta(:n) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( leftvec(:nr,:n), beta(:n) )
!
!           GENERATE A RANDOM COEFFICIENT MATRIX rightvec(:nc,:n) .
!
            call random_number_( rightvec(:nc,:n) )
!
            rightvec(:nc,:n) = rightvec(:nc,:n) - half
!
            call random_number_( tmp )
!
            if ( tmp>safmin ) then
                rightvec(:nc,:n) = rightvec(:nc,:n)/tmp
            end if
!
!           COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
            call qr_cmp( rightvec(:nc,:n), diagr(:n), beta(:n) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( rightvec(:nc,:n), beta(:n) )
!
            do j = 1_i4b, 24_i4b
!
                select case( j )
!
                    case( 1_i4b )
!
!                       TEST 1: STRONGLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b)   = one
                        eigval(2_i4b:n) = ulp
!
                    case( 2_i4b )
!
!                       TEST 2:  STRONGLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n-1_i4b) = one
                        eigval(n)             = ulp
!
                    case( 3_i4b )
!
!                       TEST 3: WEAKLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n-1_i4b) = arth( ulp, ulp, n-1_i4b )
                        eigval(n)             = one
!
                    case( 4_i4b )
!
!                       TEST 4: INTERNAL CLUSTER OF EIGENVALUES.
!
                        eigval(1_i4b)         = ulp
                        eigval(2_i4b:n-1_i4b) = one + sqrt(ulp)*arth( one, one, n-2_i4b )
                        eigval(n)             = two
!
                    case( 5_i4b )
!
!                       TEST 5: WEAKLY CLUSTERED EIGENVALUES.
!
                        eigval(1_i4b:n) = one + (c100*ulp)*arth( zero, one, n )
!
                    case( 6_i4b )
!
!                       TEST 6: GEOMETRIC DISTRIBUTION OF EIGENVALUES.
!
                        err = real( n-1_i4b, stnd )
!
                        do l = 1_i4b, n
                            eigval(l) = ulp**(real(l-1_i4b,stnd)/err)
                        end do
!
                    case( 7_i4b )
!
!                       TEST 7: ARITHMETIC DISTRIBUTION OF EIGENVALUES.
!
                        eigval(1_i4b:n) = ulp + ((one-ulp)/real( n-1_i4b, stnd ))*arth( zero, one, n )
!
                    case( 8_i4b )
!
!                       TEST 8: TWO CLUSTERS OF EIGENVALUES.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = one + tmp*arth( one, one, i )
!
                        eigval(i+1_i4b:n) = ten + tmp*arth( one, one, n-i )
!
                    case( 9_i4b )
!
!                       TEST 9: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR ZERO.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = arth( one, one, i )
!
                        eigval(i+1_i4b:n) = tmp*arth( one, one, n-i )
!
                    case( 10_i4b )
!
!                       TEST 10: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR n/2.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        eigval(1_i4b:i) = arth( one, one, i )
!
                        eigval(i+1_i4b:n) = real( i, stnd) + tmp*arth( one, one, n-i )
!
                    case( 11_i4b )
!
!                       TEST 11: CLUSTERED DISTRIBUTION OF EIGENVALUES BETWEEN 0.9 and 1.
!
                        call random_number_( eigval(:n) )
!
                        eigval(1_i4b:n) = 0.9_stnd + 0.1_stnd*eigval(1_i4b:n)
!
                    case( 12_i4b )
!
!                       TEST 12: EIGENVALUES EQUAL TO sinh(10*i/n) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) )
!
                    case( 13_i4b )
!
!                       TEST 13: EIGENVALUES EQUAL TO sinh((10*i/n)-5) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - five )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        where( eigval(1_i4b:n)<zero )
                            beta(:n) = -eigval(1_i4b:n)
                        elsewhere
                            beta(:n) = zero
                        end where
!
                        tmp = maxval( beta(:n) ) + one
!
!                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case( 14_i4b )
!
!                       TEST 14: EIGENVALUES EQUAL TO sinh((10*i/n)-10) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - ten )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        where( eigval(1_i4b:n)<zero )
                            beta(:n) = -eigval(1_i4b:n)
                        elsewhere
                            beta(:n) = zero
                        end where
!
                        tmp = maxval( beta(:n) ) + one
!
!                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case( 15_i4b )
!
!                       TEST 15: EIGENVALUES EQUAL TO tanh((10*i/n)-5) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        eigval(1_i4b:n) = tanh( tmp*arth( one, one, n ) - five )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        where( eigval(1_i4b:n)<zero )
                            beta(:n) = -eigval(1_i4b:n)
                        elsewhere
                            beta(:n) = zero
                        end where
!
                        tmp = maxval( beta(:n) ) + one
!
!                        tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                        eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
                    case( 16_i4b )
!
!                       TEST 16: CLUSTER OF EIGENVALUES AT 1 WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        eigval(1_i4b)   = one/conda
                        eigval(2_i4b:n) = one
!
                    case( 17_i4b )
!
!                       TEST 17: CLUSTER OF EIGENVALUES AT 1/conda WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        eigval(1_i4b)   = one
                        eigval(2_i4b:n) = one/conda
!
                    case( 18_i4b )
!
!                       TEST 18: ARITHMETIC DISTRIBUTION OF EIGENVALUES WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        tmp = real( n-1_i4b, stnd )
                        err = one - one/conda
!
                        do l = 1_i4b, n
                            eigval(l) = one - ( real(l-1_i4b,stnd)/tmp )*err
                        end do
!
                    case( 19_i4b )
!
!                       TEST 19: GEOMETRIC DISTRIBUTION OF EIGENVALUES WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        tmp = real( n-1_i4b, stnd )
!
                        do l = 1_i4b, n
                            eigval(l) = conda**( -real(l-1_i4b,stnd)/tmp )
                        end do
!
                    case( 20_i4b )
!
!                       TEST 20: RANDOM DISTRIBUTION OF EIGENVALUES IN THE INTERVAL 1/conda TO 1
!                                WITH CONDITION NUMBER conda.
!
                        conda = def_conda*def_conda
!
                        call random_number_( beta(:n) )
!
                        tmp = log( one/conda )

                        beta(:n)   = tmp + beta(:n)*abs( tmp )
                        eigval(:n) = exp( beta(:n) )
!
                    case( 21_i4b )
!
!                       TEST 21: VERY SLOW DECAY OF EIGENVALUES.
!
                        conda = real( n - 1_i4b, stnd )
!
                        do l = 1_i4b, n
!
                            tmp = real( l-1_i4b, stnd )
!
                            eigval(l) = exp( -tmp/conda )
!        
                        end do
!
                    case( 22_i4b )
!
!                       TEST 22: SLOW DECAY OF EIGENVALUES.
!
                        do l = 1_i4b, n
!
                            tmp = real( l, stnd )
!
                            eigval(l) = one/( tmp*tmp )
!        
                        end do
!
                    case( 23_i4b )
!
!                       TEST 23: FAST DECAY OF EIGENVALUES.
!
                        do l = 1_i4b, n
!
                            tmp = real( l, stnd )
!
                            eigval(l) = exp( -tmp/seven )
!        
                        end do
!
                    case( 24_i4b )
!
!                       TEST 24: S-SHAPE DECAY OF EIGENVALUES.
!
                        do l = 1_i4b, n
!
                            tmp = real( l, stnd )
!
                            eigval(l) = 0.0001_stnd + one/( one + exp( tmp -c30 ) )
!        
                        end do
!
                    case default
!
!                       DEFAULT TEST : RANDOMLY UNIFORM DISTRIBUTION OF EIGENVALUES.
!
                        call random_number_( eigval(:n) )
!
                        call random_number_( tmp )
!
                        if ( tmp>safmin ) then
                             eigval(1_i4b:n) = eigval(1_i4b:n)/tmp
                        end if
!
                end select
!
#ifdef _F2003
                if ( ieee_support_datatype( eigval ) ) then
                    if ( .not.all( ieee_is_normal( eigval(1_i4b:n) ) ) ) cycle
                end if
#endif
!
!               COMPUTE THE SINGULAR VALUES FROM THE EIGENVALUES.
!
                singval(:n) = sqrt( eigval(:n) )
!
!               SORT THE SINGULAR VALUES.
!
                call singval_sort( sort, singval(:n) )
!
!               FORM A MATRIX FROM THE SPECIFIED SINGULAR VALUE DISTRIBUTION AND RANDOM SINGULAR VECTORS.
!
                rightvect(:n,:nc) = transpose( rightvec(:nc,:n) )
!
                do i = 1_i4b, nc
!
                    rightvect(:n,i) = rightvect(:n,i)*singval(:n)
!
                end do
!
                a(:nr,:nc) = matmul( leftvec(:nr,:n), rightvect(:n,:nc) )
!
!               COMPUTE THE FROBENIUS NORM OF THE MATRIX.
!
!                norma = norm( a(:nr,:nc) )
                norma = norm( singval(:n) )
!
!               DEALLOCATE POINTERS FOR STORING THE COMPUTED PARTIAL SVD.
!
                if ( associated(singval2) )  deallocate( singval2 )
                if ( associated(leftvec2) )  deallocate( leftvec2 )
                if ( associated(rightvec2) ) deallocate( rightvec2 )
!
!               RESET THE TOLERANCE.
!
                relerr2 = relerr0
!
!               COMPUTE APPROXIMATE PARTIAL SVD OF RANDOM DATA MATRIX, WHICH SATISFIES THE PRESET
!               TOLERANCE (E.G., relerr0) FOR ITS RELATIVE ERROR IN FROBENIUS NORM.
!
                call rsvd_cmp_fixed_precision( a(:nr,:nc), relerr2, singval2, leftvec2, rightvec2,    &
                                               failure_relerr=failure_relerr, failure=failure,        &
                                               niter=niter, blk_size=blk_size, maxiter_qb=maxiter_qb, &
                                               ortho=ortho, reortho=reortho, niter_qb=niter_qb        )
!
!                if ( failure_relerr ) then
!
!                    notdone = notdone + 1_i4b
!
!                    cycle
!
!                end if
!
!               GET THE RANK OF THE COMPUTED PARTIAL SVD APPROXIMATION.
!
                nsvd = size( singval2 )
!
!               COMPUTE THE TRUE RELATIVE ERROR.
!
                relerr = sqrt(sum( (singval(nsvd+1_i4b:n)/norma)**2 ))
!
!               COMPUTE ERROR BETWEEN THE TRUE AND ESTIMATED RELATIVE ERRORS.
!
                err = abs( relerr - relerr2 )
!
                m = m + 1_i4b
!
                if ( details .or. err>eps ) then
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*v - u*s
!                   WHERE s ARE THE SINGULAR VALUES, u THE LEFT SINGULAR VECTORS AND
!                   v THE RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                    res(:nr,:nsvd) = matmul( a(:nr,:nc), rightvec2(:nc,:nsvd) )         &
                                     - leftvec2(:nr,:nsvd)*spread(singval2(:nsvd),1,nr)
!
                    if ( norma/=zero ) then
                        err1 = maxval( norm(res(:nr,:nsvd), dim=2_i4b) )/( norma*real(nr,stnd) )
                    else
                        err1 = maxval( norm(res(:nr,:nsvd), dim=2_i4b) )/real(nr,stnd)
                    end if
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!                   WHERE u are LEFT SINGULAR VECTORS OF THE MATRIX.
!
                    res(:nsvd,:nsvd) = matmul( transpose( leftvec2(:nr,:nsvd) ), leftvec2(:nr,:nsvd) )
!
                    do l = 1_i4b, nsvd
                        res(l,l) = one - res(l,l)
                    end do
!
                    err2 = maxval( norm(res(:nsvd,:nsvd), dim=2_i4b) )/real(nr,stnd)
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - v**(t)*v
!                   WHERE v are RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                    res(:nsvd,:nsvd) = matmul( transpose( rightvec2(:nc,:nsvd) ), rightvec2(:nc,:nsvd) )
!
                    do l = 1_i4b, nsvd
                        res(l,l) = one - res(l,l)
                    end do
!
                    err3 = maxval( norm(res(:nsvd,:nsvd), dim=2_i4b) )/real(nc,stnd)
!
                end if
!
                if ( err<=eps ) then
!
                    success = success + 1_i4b
!
                    if ( details ) then
!
                        write (prtunit,*) 'Test ',m,' is correct'
!
                        write (prtunit,*) 'nr= ', nr,' , nc= ', nc,' , case= ', j,        &
                                          ' , nsvd= ', nsvd,  &
                                          ' , niter= ', niter,' , blk_size= ', blk_size,  &
                                          ' , maxiter_qb= ', maxiter_qb,                  &
                                          ' , ortho= ', ortho,' , reortho= ', reortho,    &
                                          ' , niter_qb= ', niter_qb,                      &
                                          ' , failure_relerr= ', failure_relerr,          &
                                          ' , failure= ', failure
!
!                       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*v - u*s
!                       WHERE s ARE THE SINGULAR VALUES, u THE LEFT SINGULAR VECTORS AND
!                       v THE RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                        write (prtunit,*) 'Accuracy of the computed singular triplets             = ', &
                                           err1/sqrt_ulp, ' (must be less than :', fudge, ' )'
!
!                       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!                       WHERE u are LEFT SINGULAR VECTORS OF THE MATRIX.
!
                        write (prtunit,*) 'Orthogonality of the computed left singular vectors    = ', &
                                          err2/ulp, ' (must be less than :', fudge, ' )'
!
!                       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - v**(t)*v
!                       WHERE v are RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                        write (prtunit,*) 'Orthogonality of the computed right singular vectors   = ', &
                                          err3/ulp, ' (must be less than :', fudge, ' )'
!
                        write (prtunit,*)
!
                    end if
!
                else
!
                    write (prtunit,*) 'Test ',m,' is incorrect'
!
                    write (prtunit,*) '=> ',err,' > ',eps,' , nr= ', nr,' , nc= ', nc,' , case= ', j,        &
                                      ' , nsvd= ', nsvd,' , niter= ', niter,                                 &
                                      ' , blk_size= ', blk_size,' , maxiter_qb= ', maxiter_qb,               &
                                      ' , ortho= ', ortho,' , reortho= ', reortho,' , niter_qb= ', niter_qb, &
                                      ' , failure_relerr= ', failure_relerr,' , failure= ', failure
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*v - u*s
!                   WHERE s ARE THE SINGULAR VALUES, u THE LEFT SINGULAR VECTORS AND
!                   v THE RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                    write (prtunit,*) 'Accuracy of the computed singular triplets             = ', &
                                      err1/sqrt_ulp, ' (must be less than :', fudge, ' )'
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!                   WHERE u are LEFT SINGULAR VECTORS OF THE MATRIX.
!
                    write (prtunit,*) 'Orthogonality of the computed left singular vectors    = ', &
                                      err2/ulp, ' (must be less than :', fudge, ' )'
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - v**(t)*v
!                   WHERE v are RIGHT SINGULAR VECTORS OF THE MATRIX.
!
                    write (prtunit,*) 'Orthogonality of the computed right singular vectors   = ', &
                                      err3/ulp, ' (must be less than :', fudge, ' )'
!
                    write (prtunit,*)
!
                end if
!
            end do
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( a, leftvec, rightvec, rightvect, res )
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( eigval, singval, diagr, beta )
!
    end do
!
!   DEALLOCATE POINTERS IF REQUIRED.
!
    if ( associated(singval2) )  deallocate( singval2 )
    if ( associated(leftvec2) )  deallocate( leftvec2 )
    if ( associated(rightvec2) ) deallocate( rightvec2 )
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test__rsvd_cmp_fixed_precision
! =============================================
!
end program test_rsvd_cmp_fixed_precision
