program test_solve_lin
!
!
! Purpose
! =======
!
!   This program is intended to test function SOLVE_LIN
!   in module Lin_Procedures.
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    Random matrices with given singular value distributions are generated for
!    checking the accuracy of function SOLVE_LIN.
!
! LATEST REVISION : 25/05/2021
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,        &
                         zero, half, one, two, five, ten, c100,         &
                         qr_cmp, ortho_gen_qr, solve_lin, arth, lamch,  &
                         norm, merror, allocate_error
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def, nrhs=20
!   
    real(stnd), parameter  :: fudge=ten, max_conda=1000._stnd
!
    character(len=*), parameter :: name_proc='solve_lin'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, eps, eps2, ulp, safmin, tmp,   &
                                               shift, conda
    real(stnd), allocatable, dimension(:)   :: eigval, singval, diagr, beta, normx, condax
    real(stnd), allocatable, dimension(:,:) :: a, leftvec, rightvec, res, b, x, x2
!
    integer(i4b)  :: i, j, k, l, n, m, krank0, success, notdone
    integer :: iok
!
    logical(lgl) :: details
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( normx(nrhs), condax(nrhs), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    eps    = fudge*ulp
!
    safmin = lamch( 'S' )
!
!   SPECIFY EVENTUALLY A SHIFT TO AVOID ZERO SINGULAR VALUE
!   OR SHIFT THE SINGULAR VALUES DISTRIBUTION.
!
    shift = 0.05_stnd
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        j = max( n, nrhs )
!
        allocate( eigval(n), singval(n), diagr(n),  &
                  beta(n),  a(n,n), res(n,j),       &
                  leftvec(n,n),  rightvec(n,n),     &
                  x(n,nrhs), x2(n,nrhs),            &
                  b(n,nrhs), stat=iok  )
!
        if ( iok/=0 ) then
!
            if ( allocated(eigval) )    deallocate( eigval )
            if ( allocated(singval) )   deallocate( singval )
            if ( allocated(diagr) )     deallocate( diagr )
            if ( allocated(beta) )      deallocate( beta )
            if ( allocated(a) )         deallocate( a )
            if ( allocated(res) )       deallocate( res )
            if ( allocated(leftvec) )   deallocate( leftvec )
            if ( allocated(rightvec) )  deallocate( rightvec )
            if ( allocated(x) )         deallocate( x )
            if ( allocated(x2) )        deallocate( x2 )
            if ( allocated(b) )         deallocate( b )
!
            cycle
!
        end if
!
!       GENERATE A RANDOM COEFFICIENT MATRIX leftvec(:n,:n) .
!
        call random_number( leftvec(:n,:n) )
!
        leftvec(:n,:n) = leftvec(:n,:n) - half
!
        call random_number( tmp )
!
        if ( tmp>safmin ) then
            leftvec(:n,:n) = leftvec(:n,:n)/tmp
        end if
!
!       COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
        call qr_cmp( leftvec(:n,:n), diagr(:n), beta(:n) )
!
!       GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!       q IS ASSUMED OF FULL RANK.
!
        call ortho_gen_qr( leftvec(:n,:n), beta(:n) )
!
!       GENERATE A RANDOM COEFFICIENT MATRIX rightvec(:n,:n) .
!
        call random_number( rightvec(:n,:n) )
!
        rightvec(:n,:n) = rightvec(:n,:n) - half
!
        call random_number( tmp )
!
        if ( tmp>safmin ) then
            rightvec(:n,:n) = rightvec(:n,:n)/tmp
        end if
!
!       COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
        call qr_cmp( rightvec(:n,:n), diagr(:n), beta(:n) )
!
!       GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!       q IS ASSUMED OF FULL RANK.
!
        call ortho_gen_qr( rightvec(:n,:n), beta(:n) )
!
!       GENERATE A RANDOM SOLUTION MATRIX x .
!
        call random_number( x(:n,:nrhs) )
!
        call random_number( tmp )
!
        if ( tmp>safmin ) then
            x(:n,:nrhs) = x(:n,:nrhs)/tmp
        end if
!
!       COMPUTE THE NORMS OF THE COLUMNS OF RANDOM SOLUTION MATRIX x .
!
        normx(:nrhs) = norm( x(:n,:nrhs), dim=2_i4b )
!
!       COMPUTE FIRST THE EIGENVALUES OF THE MATRIX PRODUCT
!                     a(:n,:n)**(t)*a(:n,:n).
!
        do j = 1_i4b, 16_i4b
!
            select case( j )
!
            case( 1_i4b )
!
!               TEST 1: STRONGLY CLUSTERED EIGENVALUES.
!
                eigval(1_i4b)   = one
                eigval(2_i4b:n) = ulp
!
            case( 2_i4b )
!
!               TEST 2:  STRONGLY CLUSTERED EIGENVALUES.
!
                eigval(1_i4b:n-1_i4b) = one
                eigval(n)             = ulp
!
            case( 3_i4b )
!
!               TEST 3: WEAKLY CLUSTERED EIGENVALUES.
!
                eigval(1_i4b:n-1_i4b) = arth( ulp, ulp, n-1_i4b )
                eigval(n)             = one
!
            case( 4_i4b )
!
!               TEST 4: INTERNAL CLUSTER OF EIGENVALUES.
!
                eigval(1_i4b)         = ulp
                eigval(2_i4b:n-1_i4b) = one + sqrt(ulp)*arth( one, one, n-2_i4b )
                eigval(n)             = two
!
            case( 5_i4b )
!
!               TEST 5: WEAKLY CLUSTERED EIGENVALUES.
!
                eigval(1_i4b:n) = one + (c100*ulp)*arth( zero, one, n )
!
            case( 6_i4b )
!
!               TEST 6: GEOMETRIC DISTRIBUTION OF EIGENVALUES.
!
                err = real( n-1_i4b, stnd )
                do l = 1_i4b, n
                    eigval(l) = ulp**(real(l-1_i4b,stnd)/err)
                end do
!
            case( 7_i4b )
!
!               TEST 7: ARITHMETIC DISTRIBUTION OF EIGENVALUES.
!
                eigval(1_i4b:n) = ulp + ((one-ulp)/real( n-1_i4b, stnd ))*arth( zero, one, n )
!
            case( 8_i4b )
!
!               TEST 8: TWO CLUSTERS OF EIGENVALUES.
!
                i = n/2_i4b
!
                tmp = sqrt( ulp )
!
                eigval(1_i4b:i) = one + tmp*arth( one, one, i )
!
                eigval(i+1_i4b:n) = ten + tmp*arth( one, one, n-i )
!
            case( 9_i4b )
!
!               TEST 9: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR ZERO.
!
                i = n/2_i4b
!
                tmp = sqrt( ulp )
!
                eigval(1_i4b:i) = arth( one, one, i )
!
                eigval(i+1_i4b:n) = tmp*arth( one, one, n-i )
!
            case( 10_i4b )
!
!               TEST 10: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR n/2.
!
                i = n/2_i4b
!
                tmp = sqrt( ulp )
!
                eigval(1_i4b:i) = arth( one, one, i )
!
                eigval(i+1_i4b:n) = real( i, stnd) + tmp*arth( one, one, n-i )
!
            case( 11_i4b )
!
!               TEST 11: CLUSTERED DISTRIBUTION OF EIGENVALUES BETWEEN 0.9 and 1.
!
                call random_number( eigval(:n) )
!
                eigval(1_i4b:n) = 0.9_stnd + 0.1_stnd*eigval(1_i4b:n)
!
            case( 12_i4b )
!
!               TEST 12: EIGENVALUES EQUAL TO sinh(10*i/n) FOR i=1 TO n.
!
                tmp = ten/real( n, stnd)
!
                eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) )
!
            case( 13_i4b )
!
!               TEST 13: EIGENVALUES EQUAL TO sinh((10*i/n)-5) FOR i=1 TO n.
!
                tmp = ten/real( n, stnd)
!
                eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - five )
!
!               SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                where( eigval(1_i4b:n)<zero )
                    beta(:n) = -eigval(1_i4b:n)
                elsewhere
                    beta(:n) = zero
                end where
!
                tmp = maxval( beta(:n) ) + one
!
!                tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
            case( 14_i4b )
!
!               TEST 14: EIGENVALUES EQUAL TO sinh((10*i/n)-10) FOR i=1 TO n.
!
                tmp = ten/real( n, stnd)
!
                eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - ten )
!
!               SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                where( eigval(1_i4b:n)<zero )
                    beta(:n) = -eigval(1_i4b:n)
                elsewhere
                    beta(:n) = zero
                end where
!
                tmp = maxval( beta(:n) ) + one
!
!                tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
            case( 15_i4b )
!
!               TEST 15: EIGENVALUES EQUAL TO tanh((10*i/n)-5) FOR i=1 TO n.
!
                tmp = ten/real( n, stnd)
!
                eigval(1_i4b:n) = tanh( tmp*arth( one, one, n ) - five )
!
!               SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                where( eigval(1_i4b:n)<zero )
                    beta(:n) = -eigval(1_i4b:n)
                elsewhere
                    beta(:n) = zero
                end where
!
                tmp = maxval( beta(:n) ) + one
!
!                tmp = maxval( abs( eigval(1_i4b:n) ) )
!
                eigval(1_i4b:n) = tmp + eigval(1_i4b:n)
!
            case default
!
!               DEFAULT TEST : RANDOMLY UNIFORM DISTRIBUTION OF EIGENVALUES.
!
                call random_number( eigval(:n) )
!
                call random_number( tmp )
!
                if ( tmp>safmin ) then
                    eigval(1_i4b:n) = eigval(1_i4b:n)/tmp
                end if
!
            end select
!
#ifdef _F2003
            if ( ieee_support_datatype( eigval ) ) then
                if ( .not.all( ieee_is_normal( eigval(1_i4b:n) ) ) ) cycle
            end if
#endif
!
!           COMPUTE THE SINGULAR VALUES FROM THE EIGENVALUES.
!
            singval(:n) = sqrt( eigval(:n) )
!
            singval(:n) = singval(:n) + shift
!
            krank0 = count( singval(:n)/=zero )
!
!           SKIP DEFICIENT MATRICES.
!
            if ( krank0/=n ) cycle
!
!           FORM A MATRIX FROM THE SPECIFIED SINGULAR VALUE DISTRIBUTION AND RANDOM SINGULAR VECTORS.
!
            res(:n,:n) = transpose( rightvec(:n,:n) )
!
            do i = 1_i4b, n
!
                res(:n,i) = res(:n,i)*singval(:n)
!
            end do
!
            a(:n,:n) = matmul( leftvec(:n,:n), res(:n,:n) )
!
!           COMPUTE CONDITION NUMBER OF RANDOM MATRIX.
!
            conda = maxval( singval(:n) )/minval( singval(:n) )
!
!            conda = ( sum( singval(:n) )*sum( one/singval(:n) ) )/real( n, stnd)
!
            tmp =  min( conda, max_conda )
!
            condax(:nrhs) = normx(:nrhs)*tmp
!
!           COMPUTE THE MATRIX-MATRIX PRODUCT b = a*x .
!
            b(:n,:nrhs) = matmul( a(:n,:n), x(:n,:nrhs) )
!
            do i = 1_i4b, 2_i4b
!
                select case( i )
!
                case( 1_i4b )
!
!                   TEST 1 : SEVERAL RIGHT HAND-SIDES.
!
!                   COMPUTE THE SOLUTION MATRIX FOR LINEAR SYSTEM
!
!                               a*x = b .
!
                    x2(:n,:nrhs) = solve_lin( a(:n,:n), b(:n,:nrhs) )
!
!                   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
                    res(:n,:nrhs) = x2(:n,:nrhs) - x(:n,:nrhs)
!
                    err = maxval( norm(res(:n,:nrhs), dim=2_i4b ) /    &
                                  condax(:nrhs)  )
!
                    k = nrhs
!
                case( 2_i4b )
!
!                   TEST 2 : ONE RIGHT HAND-SIDE.
!
!                   COMPUTE THE SOLUTION VECTOR FOR LINEAR SYSTEM
!
!                               a*x(:,1)=b(:,1) .
!
                    x2(:,1_i4b) = solve_lin( a(:n,:n), b(:,1_i4b) )
!
!                   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
                    res(:,1_i4b) = x2(:,1_i4b) - x(:,1_i4b)
!
                    err = norm( res(:,1_i4b) )/condax(1_i4b)
!
                    k = 1_i4b
!
                end select
!
                m = m + 1_i4b
!
                eps2 = real( n, stnd)*eps
!
                if ( err<=eps2 ) then
!
                    success = success + 1_i4b
!
                    if ( details ) then
!
                        write (prtunit,*) 'Test ',m,' is correct'
!
                        write (prtunit,*) 'n= ', n,' , case= ', j,' , nrhs= ', k
!
!                       PRINT CONDITION NUMBER OF COEFFICIENT MATRIX.
!
                        write (prtunit,*) 'Condition number of the coefficient matrix         = ', conda
!
!                       PRINT SCALED ERRORS OF SOLUTIONS.
!
                        write (prtunit,*) 'Maximum of scaled errors of the computed solutions = ', err
!
                        write (prtunit,*)
!
                    end if
!
                else
!
                    write (prtunit,*) 'Test ',m,' is incorrect'
!
                    write (prtunit,*) '=> ',err,' > ',eps2,' , n= ', n,' , case= ', j,   &
                                      ' , nrhs= ', k
!
!                   PRINT CONDITION NUMBER OF COEFFICIENT MATRIX.
!
                    write (prtunit,*) 'Condition number of the coefficient matrix         = ', conda
!
!                   PRINT SCALED ERRORS OF SOLUTIONS.
!
                    write (prtunit,*) 'Maximum of scaled errors of the computed solutions = ', err
!
                    write (prtunit,*)
!
                end if
!
            end do
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( eigval, singval, diagr, beta,       &
                    a, leftvec, rightvec, res, b, x, x2 )
!
    end do
!
!   ALLOCATE WORK ARRAYS.
!
    deallocate( normx, condax )
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_solve_lin
! =============================
!
end program test_solve_lin
