program test_symlin_filter
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine SYMLIN_FILTER in module Time_Series_Procedures.
!                                                                              
! LATEST REVISION : 20/04/2010
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : i4b, stnd, true, bd_coef2, symlin_filter, symlin_filter2
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT .
!
    integer(i4b), parameter :: prtunit=6, m=20
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, tmp, eps
    real(stnd), dimension(:,:), allocatable :: mat2, mat3
    real(stnd), dimension(:),   allocatable :: coef, vec, vec2, vec3
!
    integer(i4b) :: n, i, j, k, k1, k2, pl, ph, khalf, kmid, nfilt
!   
    integer :: iok
!   
!   
! PARAMETERS 
! ==========
!
    character(len=*), parameter :: name_proc='symlin_filter'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TESTS.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    eps = sqrt( epsilon(err) )
!
!   pl IS THE MINIMUM PERIOD OF OSCILLATION OF DESIRED COMPONENT.
!   ph IS THE MAXIMUM PERIOD OF OSCILLATION OF DESIRED COMPONENT.
!   pl AND ph ARE EXPRESSED IN NUMBER OF OBSERVATIONS.
!
    pl  = 35_i4b
    ph  = 96_i4b
!
!   NOW SELECT THE NUMBER OF FILTER COEFFICIENTS k FOR THE HAMMING FILTER.
!
    k = ph + 1_i4b
!
!   CHECK IF k IS ODD.
!
    if ( (k/2_i4b)*2_i4b==k )  k = k + 1_i4b
!
!   SET THE NUMBER OF OBSERVATIONS
!
    n = 3_i4b*k
!
!   ALLOCATE WORK ARRAYS .
!
    allocate( coef(k), vec(n), vec2(n), vec3(n),    &
              mat2(m,n), mat3(m,n), stat=iok        )
!
    if ( iok==0 ) then
!
!       FUNCTION bd_coef2 COMPUTES THE K-TERM LEAST SQUARES APPROXIMATION TO AN
!       -IDEAL- BAND PASS FILTER WITH CUTOFF PERIODS PL AND PH .
!
        coef(:k) = bd_coef2( pl, ph, k )
!
!       GENERATE A RANDOM REAL NUMBER SEQUENCE OF LENGTH n .
!
        call random_number( vec(:n) )
!
        kmid  = ( k + 1_i4b )/2_i4b
        khalf = ( k - 1_i4b )/2_i4b
!
        k1    = kmid
        k2    = n - khalf
!
        tmp     =  maxval(abs(vec(k1:k2)))
!
        do i = 1_i4b, 3_i4b
!
            select case( i )
!
                case( 1_i4b)
!
!                   TEST 1 : TEST 1D SYMMETRIC LINEAR FILTERING.
!
!                   SAVE THE REAL RANDOM NUMBER ARRAY.
!
                    vec2(:n) = vec(:n)
                    vec3(:n) = vec(:n)
!
!                   FILTER THE TIME SERIES, KEEPING THE PERIODS BETWEEN pl AND ph .
!                   NOTE THAT (size(COEF)-1)/2 DATA POINTS WILL BE LOST FROM EACH END OF THE TIME SERIES,
!
                    call symlin_filter( vec2(:n), coef(:k), nfilt=nfilt )
!
!                   FILTER AGAIN THE TIME SERIES, BUT COMPUTE THE VALUES AT THE ENDS OF THE OUTPUT
!                   BY ASSUMING THAT THE INPUT IS PART OF A PERIODIC SEQUENCE OF PERIOD n .
!                   NOTE THAT (size(COEF)-1)/2 DATA POINTS WILL BE AFFECTED
!                   BY END EFFECTS  FROM EACH END OF THE SERIES.
!
                    call symlin_filter2( vec3(:n), coef(:k), usefft=true )
!
!                   TEST THE ACCURACY OF THE FILTERING OPERATION.
!
                    err = maxval(abs(vec3(k1:k2)-vec2(:nfilt)))/tmp
!
                case( 2_i4b)
!
!                   TEST 2 : TEST 1D SYMMETRIC LINEAR FILTERING.
!
!                   SAVE THE REAL RANDOM NUMBER ARRAY.
!
                    mat2(1_i4b,:n) = vec(:n)
                    mat3(1_i4b,:n) = vec(:n)
!
!                   FILTER THE TIME SERIES, KEEPING THE PERIODS BETWEEN pl AND ph .
!                   NOTE THAT (size(COEF)-1)/2 DATA POINTS WILL BE LOST FROM EACH END OF THE TIME SERIES,
!
                    call symlin_filter( mat2(1_i4b,:n), coef(:k), nfilt=nfilt )
!
!                   FILTER AGAIN THE TIME SERIES, BUT COMPUTE THE VALUES AT THE ENDS OF THE OUTPUT
!                   BY ASSUMING THAT THE INPUT IS PART OF A PERIODIC SEQUENCE OF PERIOD n .
!                   NOTE THAT (size(COEF)-1)/2 DATA POINTS WILL BE AFFECTED
!                   BY END EFFECTS  FROM EACH END OF THE SERIES.
!
                    call symlin_filter2( mat3(1_i4b,:n), coef(:k), usefft=true )
!
!                   TEST THE ACCURACY OF THE FILTERING OPERATION.
!
                    err = maxval(abs(mat3(1_i4b,k1:k2)-mat2(1_i4b,:nfilt)))/tmp
!
                case( 3_i4b)
!
!                   TEST 3 : TEST 2D SYMMETRIC LINEAR FILTERING.
!
!                   SAVE THE REAL RANDOM NUMBER ARRAY.
!
                    do j = 1_i4b, m
                        mat2(j,:n) = vec(:n)
                        mat3(j,:n) = vec(:n)
                    end do
!
!                   FILTER THE TIME SERIES, KEEPING THE PERIODS BETWEEN pl AND ph .
!                   NOTE THAT (size(COEF)-1)/2 DATA POINTS WILL BE LOST FROM EACH END OF THE TIME SERIES,
!
                    call symlin_filter( mat2(:m,:n), coef(:k), nfilt=nfilt )
!
!                   FILTER AGAIN THE TIME SERIES, BUT COMPUTE THE VALUES AT THE ENDS OF THE OUTPUT
!                   BY ASSUMING THAT THE INPUT IS PART OF A PERIODIC SEQUENCE OF PERIOD n .
!                   NOTE THAT (size(COEF)-1)/2 DATA POINTS WILL BE AFFECTED
!                   BY END EFFECTS  FROM EACH END OF THE SERIES.
!
                    call symlin_filter2( mat3(:m,:n), coef(:k), usefft=true )
!
!                   TEST THE ACCURACY OF THE FILTERING OPERATION.
!
                    err = maxval(abs(mat3(:m,k1:k2)-mat2(:m,:nfilt)))/tmp
!
            end select
!
!           PRINT THE RESULTS OF THE TEST.
!
            if ( err<=eps ) then
                write (prtunit,*) 'Test ',i,' is correct'
            else
                write (prtunit,*) 'Test ',i,' is incorrect'
                write (prtunit,*) '=> ',err,' > ',eps
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( coef, vec, vec2, vec3, mat2, mat3 )
!
    else
!
        write (prtunit,*) 'Problem in attempt to allocate memory for tests 1, 2 and 3'
!
    end if
!
!
! END OF PROGRAM test_symlin_filter
! =================================
!
end program test_symlin_filter
