program test_symtrid_cmp2
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine SYMTRID_CMP2
!   in module EIG_Procedures.
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    Random matrices with given singular value distributions are generated for checking
!    the accuracy of subroutine SYMTRID_CMP2.
!
! LATEST REVISION : 30/05/2021
!                                                                              
! ================================================================================================
!
!
! USED MODULES
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def, true, false,   &
                         zero, half, one, two, five, ten, c100,                 &
                         qr_cmp, ortho_gen_qr, symtrid_cmp2, ortho_gen_symtrid, &
                         norm, arth, lamch
!
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def
!
    real(stnd), parameter  :: fudge=c100
!
    character(len=*), parameter :: name_proc='symtrid_cmp2'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err, eps, ulp,   &
                                               safmin, tmp, norma, shift
    real(stnd), allocatable, dimension(:)   :: singval, diagr, beta, d, e
    real(stnd), allocatable, dimension(:,:) :: a, ata, trid, leftvec, rightvec, res
!
    integer      :: iok
    integer(i4b) :: i, j, k, l, n, nr, m, o, success
!
    logical(lgl) :: details
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    eps    = fudge*ulp
    safmin = lamch( 'S' )
!
    shift = 0.01_stnd
!
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( singval(n), diagr(n), beta(n),     &
                  d(n), e(n), ata(n,n), trid(n,n),   &
                  res(n,n), rightvec(n,n), stat=iok  )
!
        if ( iok/=0 ) then
!
            if ( allocated(singval) )   deallocate( singval )
            if ( allocated(d) )         deallocate( d )
            if ( allocated(e) )         deallocate( e )
            if ( allocated(diagr) )     deallocate( diagr )
            if ( allocated(beta) )      deallocate( beta )
            if ( allocated(ata) )       deallocate( ata )
            if ( allocated(trid) )      deallocate( trid )
            if ( allocated(res) )       deallocate( res )
            if ( allocated(rightvec) )  deallocate( rightvec )
!
            cycle
!
        end if
!
        do k = 1_i4b, 3_i4b
!
            select case( k )
!
                case( 1_i4b )
!
!                   TEST 1: TRIDIAGONALIZATION OF a(:nr,:n)**(t).a(:nr,:n) WITH nr=n .
!
                    nr = n
!
                case( 2_i4b )
!
!                   TEST 2: TRIDIAGONALIZATION OF a(:nr,:n)**(t).a(:nr,:n) WITH nr>=n
!                           AND nr NOT MUCH LARGER THAN n .
!
                    nr = n + 1_i4b
!
                case( 3_i4b )
!
!                   TEST 3: TRIDIAGONALIZATION OF a(:nr,:n)**(t).a(:nr,:n) WITH nr>=n
!                           AND nr MUCH LARGER THAN n .
!
                    nr = n + n
!
            end select
!
!           ALLOCATE WORK ARRAYS.
!
            allocate( a(nr,n), leftvec(nr,n), stat=iok )
!
            if ( iok/=0 ) then
!
                if ( allocated(a) )       deallocate( a )
                if ( allocated(leftvec) ) deallocate( leftvec )
!
                cycle
!
            end if
!
!           GENERATE A RANDOM COEFFICIENT MATRIX leftvec(:nr,:n) .
!
            call random_number( leftvec(:nr,:n) )
!
            leftvec(:nr,:n) = leftvec(:nr,:n) - half
!
            call random_number( tmp )
!
            if ( tmp>safmin ) then
                leftvec(:nr,:n) = leftvec(:nr,:n)/tmp
            end if
!
!           COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
            call qr_cmp( leftvec(:nr,:n), diagr(:n), beta(:n) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( leftvec(:nr,:n), beta(:n) )
!
!           GENERATE A RANDOM COEFFICIENT MATRIX rightvec(:n,:n) .
!
            call random_number( rightvec(:n,:n) )
!
            rightvec(:n,:n) = rightvec(:n,:n) - half
!
            call random_number( tmp )
!
            if ( tmp>safmin ) then
                rightvec(:n,:n) = rightvec(:n,:n)/tmp
            end if
!
!           COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
            call qr_cmp( rightvec(:n,:n), diagr(:n), beta(:n) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( rightvec(:n,:n), beta(:n) )
!
            trid(:n,:n) = zero
!
            do j = 1_i4b, 16_i4b
!
                select case( j )
!
                    case( 1_i4b )
!
!                       TEST 1: STRONGLY CLUSTERED EIGENVALUES.
!
                        d(1_i4b)   = one
                        d(2_i4b:n) = ulp
!
                    case( 2_i4b )
!
!                       TEST 2:  STRONGLY CLUSTERED EIGENVALUES.
!
                        d(1_i4b:n-1_i4b) = one
                        d(n)             = ulp
!
                    case( 3_i4b )
!
!                       TEST 3: WEAKLY CLUSTERED EIGENVALUES.
!
                        d(1_i4b:n-1_i4b) = arth( ulp, ulp, n-1_i4b )
                        d(n)             = one
!
                    case( 4_i4b )
!
!                       TEST 4: INTERNAL CLUSTER OF EIGENVALUES.
!
                        d(1_i4b)         = ulp
                        d(2_i4b:n-1_i4b) = one + sqrt(ulp)*arth( one, one, n-2_i4b )
                        d(n)             = two
!
                    case( 5_i4b )
!
!                       TEST 5: WEAKLY CLUSTERED EIGENVALUES.
!
                        d(1_i4b:n) = one + (c100*ulp)*arth( zero, one, n )
!
                    case( 6_i4b )
!
!                       TEST 6: GEOMETRIC DISTRIBUTION OF EIGENVALUES.
!
                        err = real( n-1_i4b, stnd )
                        do l = 1_i4b, n
                            d(l) = ulp**(real(l-1_i4b,stnd)/err)
                        end do
!
                    case( 7_i4b )
!
!                       TEST 7: ARITHMETIC DISTRIBUTION OF EIGENVALUES.
!
                        d(1_i4b:n) = ulp + ((one-ulp)/real( n-1_i4b, stnd ))*arth( zero, one, n )
!
                    case( 8_i4b )
!
!                       TEST 8: TWO CLUSTERS OF EIGENVALUES.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        d(1_i4b:i) = one + tmp*arth( one, one, i )
!
                        d(i+1_i4b:n) = ten + tmp*arth( one, one, n-i )
!
                    case( 9_i4b )
!
!                       TEST 9: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR ZERO.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        d(1_i4b:i) = arth( one, one, i )
!
                        d(i+1_i4b:n) = tmp*arth( one, one, n-i )
!
                    case( 10_i4b )
!
!                       TEST 10: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR n/2.
!
                        i = n/2_i4b
!
                        tmp = sqrt( ulp )
!
                        d(1_i4b:i) = arth( one, one, i )
!
                        d(i+1_i4b:n) = real( i, stnd) + tmp*arth( one, one, n-i )
!
                    case( 11_i4b )
!
!                       TEST 11: CLUSTERED DISTRIBUTION OF EIGENVALUES BETWEEN 0.9 and 1.
!
                        call random_number( d(:n) )
!
                        d(1_i4b:n) = 0.9_stnd + 0.1_stnd*d(1_i4b:n)
!
                    case( 12_i4b )
!
!                       TEST 12: EIGENVALUES EQUAL TO sinh(10*i/n) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        d(1_i4b:n) = sinh( tmp*arth( one, one, n ) )
!
                    case( 13_i4b )
!
!                       TEST 13: EIGENVALUES EQUAL TO sinh((10*i/n)-5) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        d(1_i4b:n) = sinh( tmp*arth( one, one, n ) - five )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        tmp = maxval( abs( d(1_i4b:n) ) )
!
                        d(1_i4b:n) = tmp + d(1_i4b:n)
!
                    case( 14_i4b )
!
!                       TEST 14: EIGENVALUES EQUAL TO sinh((10*i/n)-10) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        d(1_i4b:n) = sinh( tmp*arth( one, one, n ) - ten )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        tmp = maxval( abs( d(1_i4b:n) ) )
!
                        d(1_i4b:n) = tmp + d(1_i4b:n)
!
                    case( 15_i4b )
!
!                       TEST 15: EIGENVALUES EQUAL TO tanh((10*i/n)-5) FOR i=1 TO n.
!
                        tmp = ten/real( n, stnd)
!
                        d(1_i4b:n) = tanh( tmp*arth( one, one, n ) - five )
!
!                       SHIFT EIGENVALUES TO MAKE THEM POSITIVE.
!
                        tmp = maxval( abs( d(1_i4b:n) ) )
!
                        d(1_i4b:n) = tmp + d(1_i4b:n)
!
                    case default
!
!                       DEFAULT TEST: RANDOMLY UNIFORM DISTRIBUTION OF EIGENVALUES.
!
                        call random_number( d(:n) )
!
                        call random_number( tmp )
!
                        if ( tmp>safmin ) then
                            d(1_i4b:n) = d(1_i4b:n)/tmp
                        end if
!
                end select
!
#ifdef _F2003
                if ( ieee_support_datatype( d ) ) then
                    if ( .not.all( ieee_is_normal( d(:n) ) ) ) cycle
                end if
#endif
!
!               COMPUTE THE SINGULAR VALUES FROM THE EIGENVALUES.
!
                singval(:n) = sqrt( d(:n) )
!
                do o=1_i4b, 2_i4b
!
                    m = m + 1_i4b
!
                    if ( o==2_i4b ) then
!
                        singval(:n) = singval(:n) + shift
!
                    end if
!
!                   FORM A MATRIX FROM THE SPECIFIED SINGULAR VALUE DISTRIBUTION AND RANDOM SINGULAR VECTORS.
!
                    ata(:n,:n) = transpose( rightvec(:n,:n) )
!
                    do i = 1_i4b, n
                        ata(:n,i) = singval(:n)*ata(:n,i)
                    end do
!
                    a(:nr,:n) = matmul( leftvec(:nr,:n), ata(:n,:n) )
!
!                   COMPUTE THE MATRIX PRODUCT.
!
                    ata(:n,:n) = matmul( transpose( a(:nr,:n) ), a(:nr,:n) )
!
!                   COMPUTE THE NORM OF THE MATRIX PRODUCT.
!
                    norma = norm( ata(:n,:n) )
!
!                   CALL trid_cmp2 AND ortho_gen_symtrid TO REDUCE THE MATRIX ata TO TRIDIAGONAL FORM
!
!                                 ata = Q*TRID*Q**(t)
!
!                   WHERE Q IS ORTHOGONAL AND TRD IS A TRIDIAGONAL MATRIX.
!
                    call symtrid_cmp2( a(:nr,:n), d(:n), e(:n), store_q=true )
!
                    call ortho_gen_symtrid( a(:n,:n), upper=false )
!
!                   TRD IS TRIDIAGONAL.
!
                    do l = 1_i4b, n-1_i4b
                        trid(l,l)       = d(l)
                        trid(l,l+1_i4b) = e(l)
                        trid(l+1_i4b,l) = e(l)
                    end do
!
                    trid(n,n) = d(n)
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION ata*Q - Q*TRID.
!
                    res(:n,:n) = matmul( ata(:n,:n), a(:n,:n) )   &
                                 - matmul( a(:n,:n), trid(:n,:n) )
!
                    if ( norma/=zero ) then
                        err1 = maxval( norm(res(:n,:n), dim=2_i4b) )/( norma*real(n,stnd) )
                    else
                        err1 = maxval( norm(res(:n,:n), dim=2_i4b) )/real(n,stnd)
                    end if
!
!                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - Q**(t)*Q
!                   WHERE Q ARE THE ORTHOGONAL VECTORS OF THE TRIDIAGONAL DECOMPOSITION.
!
                    res(:n,:n) = matmul( transpose( a(:n,:n) ), a(:n,:n) )
!
                    do l = 1_i4b, n
                        res(l,l) = one - res(l,l)
                    end do
!
                    err2 = maxval( norm(res(:n,:n), dim=2_i4b) )/real(n,stnd)
!
                    err  = max( err1, err2 )
!
                    if ( err<=eps ) then
!
                        success = success + 1_i4b
!
                        if ( details ) then
!
                            write (prtunit,*) 'Test ',m,' is correct'
!
                            write (prtunit,*) 'nr= ', nr,' , nc= ', n,' , case= ', j,' , type= ', o
!
!                           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION ata*Q - Q*TRID.
!
                            write (prtunit,*) 'Accuracy of the computed tridiagonal decomposition = ', &
                                              err1/ulp, ' (must be less than :', fudge, ' )'
!
!                           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - Q**(t)*Q
!                           WHERE Q ARE RIGHT ORTHOGONAL VECTORS OF THE MATRIX.
!
                            write (prtunit,*) 'Orthogonality of the computed orthogonal vectors   = ', &
                                              err2/ulp, ' (must be less than :', fudge, ' )'
!
                            write (prtunit,*)
!
                        end if
!
                    else
!
                        write (prtunit,*) 'Test ',m,' is incorrect'
!
                        write (prtunit,*) '=> ',err,' > ', eps,' , nr= ', nr,         &
                                          ' , nc= ', n,' , case= ', j,' , type= ', o
!
!                       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION ata*Q - Q*TRID.
!
                        write (prtunit,*) 'Accuracy of the computed tridiagonal decomposition = ', &
                                          err1/ulp, ' (must be less than :', fudge, ' )'
!
!                       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - Q**(t)*Q
!                       WHERE Q ARE RIGHT ORTHOGONAL VECTORS OF THE MATRIX.
!
                        write (prtunit,*) 'Orthogonality of the computed orthogonal vectors   = ', &
                                          err2/ulp, ' (must be less than :', fudge, ' )'
!
                        write (prtunit,*)
!
                    end if
!
                end do
!
            end do
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( a, leftvec )
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( singval, diagr, beta, d, e, ata, trid, rightvec, res  )
!
    end do
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_symtrid_cmp2
! ================================
!
end program test_symtrid_cmp2
