program test_symtrid_qri2
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine SYMTRID_QRI2 in module Eig_Procedures.
!                                                                              
!
! LATEST REVISION : 27/09/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,   &
                         zero, one, c50, gen_symtrid_mat,          &
                         symtrid_qri2
!
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def
!
    character(len=*), parameter :: name_proc='symtrid_qri2'
!   
    real(stnd), parameter  :: fudge=c50
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                            :: err1, err2, eps, eps1, ulp
    real(stnd), allocatable, dimension(:) :: d, e, eigval, temp
!
    integer(i4b)     :: i, m, n, success, notdone
    integer          :: iok
!
    logical(lgl)     :: failure, details
!   
    character        :: sort='d'
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp = epsilon( err1 )
    eps = fudge*ulp
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        allocate( d(n), e(n), eigval(n), temp(n), stat=iok   )
!
        if ( iok/=0 ) then
!
            if ( allocated(d) )         deallocate( d )
            if ( allocated(e) )         deallocate( e )
            if ( allocated(eigval) )    deallocate( eigval )
            if ( allocated(temp) )      deallocate( temp )
!
            cycle
!
        end if
!
        do i = 1_i4b, 50_i4b
!
!           SKIP GRADED MATRICES WITH A GEOMETRIC PROGRESSION TO AVOID
!           OVERFLOW PROBLEMS IN SINGLE PRECISION.
!
            if ( i==35_i4b .or. i==36_i4b ) cycle
!
!           GENERATE A TRIDIAGONAL MATRIX.
!           THE DIAGONAL ELEMENTS ARE STORED IN d.
!           THE OFF-DIAGONAL ELEMENTS ARE STORED IN e.
!           THE EIGENVALUES ARE STORED IN eigval.
!
            call gen_symtrid_mat( i, d(:n), e(:n), eigval=eigval(:n), sort=sort, failure=failure )
!
#ifdef _F2003
            if ( ieee_support_datatype( d ) ) then
                if ( .not.all( ieee_is_normal( d(:n) ) ) ) cycle
                if ( .not.all( ieee_is_normal( e(:n) ) ) ) cycle
            end if
#endif
!
            if ( failure ) cycle
!
!           COMPUTE EIGENVALUES OF THE TRIDIAGONAL MATRIX.
!
            call symtrid_qri2( d(:n), e(:n), failure, sort=sort )
!
            if ( failure ) then
!
                notdone = notdone + 1_i4b
!
                cycle
!
            end if
!
            m = m + 1_i4b
!
!           NOW, TEST ACCURACY OF THE RESULTS.
!
            where( eigval(:n)/=zero )
                temp(:n) = eigval(:n)
            elsewhere
                temp(:n) = one
            end where
!
!           ABSOLUTE ERRORS OF EIGENVALUES.
!
            err1 = maxval( abs( d(:n) - eigval(:n) ) )
!
!           RELATIVE ERRORS OF EIGENVALUES.
!
            err2 = maxval( abs( (d(:n) - eigval(:n))/temp(:n) ) )
!
            eps1 = real( n, stnd)*eps*maxval( abs(eigval(:n)) )
!
            if ( err1<=eps1 ) then
!
                success = success + 1_i4b
!
                if ( details ) then
!
                    write (prtunit,*) 'Test ',m,' is correct'
                    write (prtunit,*) 'n= ', n,' , case= ', i
!
                    write (prtunit,*) 'Absolute accuracy of the computed eigenvalues    = ', err1
                    write (prtunit,*) 'Relative accuracy of the computed eigenvalues    = ', err2
!
                    write (prtunit,*)
!
                end if
!
            else
!
                write (prtunit,*) 'Test ',m,' is incorrect'
                write (prtunit,*) '=> ',err1,' > ',eps,' , n= ', n,         &
                                  ' , case= ', i,' , failure= ', failure
!
                write (prtunit,*) 'Absolute accuracy of the computed eigenvalues    = ', err1
                write (prtunit,*) 'Relative accuracy of the computed eigenvalues    = ', err2
!
                write (prtunit,*)
!
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( d, e, eigval, temp )
!
    end do
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_symtrid_qri2
! ================================
!
end program test_symtrid_qri2
