program test_triang_solve
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine TRIANG_SOLVE in module Lin_Procedures. 
!                                                                              
! LATEST REVISION : 28/09/2020
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,        &
                         true, false, zero, half, two, c600, triangle,  &
                         triang_solve, norm, merror, allocate_error,    &
                         lamch
!   
!
! STRONG TYPING IMPOSED
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def, nrhs=10
!   
    real(stnd), parameter  :: fudge=c600
!
    character(len=*), parameter :: name_proc='triang_solve'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err, ulp, eps, eps2, safmin, tmp
    real(stnd), allocatable, dimension(:)   :: ap, scal, normx
    real(stnd), allocatable, dimension(:,:) :: a, res, b, x, x2
!
    integer(i4b) :: i, j, k, l, m, n, nrhs2, success
    integer      :: iok
!
    logical(lgl) :: upper, trans, details
!   
!
! EXECUTABLE STATEMENTS
! =====================
!
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   ALLOCATE WORK ARRAYS.
!
    n = max( n1, n2 )
!
    l = n*(n+1_i4b)/2_i4b
!
    allocate( normx(nrhs), scal(nrhs), a(n,n), res(n,nrhs),      &
              b(n,nrhs), x(n,nrhs), x2(n,nrhs), ap(l), stat=iok  )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    eps    = fudge*ulp
    safmin = lamch( 'S' )
!
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
        l = n*(n+1_i4b)/2_i4b
!
!       GENERATE TRIANGULAR MATRIX IN PACKED FORM ap .
!
        call random_number( ap(:l) )
!
!       MAKE SURE THAT TRIANGULAR MATRIX IS NOT SINGULAR.
!
        ap(:l) = ap(:l) + two
!
        call random_number( tmp )
!
        if ( tmp>safmin ) then
            ap(:l) = ap(:l)/tmp
        end if
!
!       GENERATE RANDOM SOLUTION MATRIX x .
!
        call random_number( x )
!
        x(:n,:nrhs) = x(:n,:nrhs) - half
!
        call random_number( tmp )
!
        if ( tmp>safmin ) then
            x(:n,:nrhs) = x(:n,:nrhs)/tmp
        end if
!
        normx(:nrhs) = norm( x(:n,:nrhs), dim=2_i4b )
!
        upper = false
        trans = true
!
        do i = 1_i4b, 2_i4b
!
!           UNPACK TRIANGULAR MATRIX a .
!
            upper = .not.upper
!
            a(:n,:n) = unpack( ap(:l), mask=triangle(upper,n,n,extra=1_i4b), field=zero )
!
            do j = 1_i4b, 2_i4b
!
                trans = .not.trans
!
!               COMPUTE RIGHT HAND SIDE MATRIX b.
!
                if ( trans ) then
                    b(:n,:nrhs) = matmul( transpose(a(:n,:n)), x(:n,:nrhs) )
                else
                    b(:n,:nrhs) = matmul( a(:n,:n), x(:n,:nrhs) )
                end if
!
                do k = 1_i4b, 4_i4b
!
                    select case( k )
!
                        case( 1_i4b )
!
!                           TEST 1 : SEVERAL RIGHT HAND-SIDES.
!
!                           MAKE A COPY OF RIGHT HAND-SIDE MATRIX b .
!
                            x2(:n,:nrhs) = b(:n,:nrhs)
!
!                           COMPUTE THE SOLUTION MATRIX FOR LINEAR TRIANGULAR SYSTEM
!
!                                            a*x = b  OR a**t*x = b .
!
                             call triang_solve( a(:n,:n), x2(:n,:nrhs), upper=upper, trans=trans )
!
                        case( 2_i4b )
!
!                            TEST 2 : SEVERAL RIGHT HAND-SIDES AND SCALING.
!
!                            MAKE A COPY OF RIGHT HAND-SIDE MATRIX b .
!
                             x2(:n,:nrhs) = b(:n,:nrhs)
!
!                            COMPUTE THE SCALED SOLUTION MATRIX FOR LINEAR TRIANGULAR SYSTEM
!
!                                            a*x = b  OR a**t*x = b .
!
                             call triang_solve( a(:n,:n), x2(:n,:nrhs), scal=scal, upper=upper, trans=trans )
!
!                            UNSCALE SOLUTION MATRIX.
!
                             x2(:n,:nrhs) = x2(:n,:nrhs)/spread( scal, dim=1, ncopies=n)
!
                        case( 3_i4b )
!
!                           TEST 3 : ONE RIGHT HAND-SIDE.
!
!                           MAKE A COPY OF RIGHT HAND-SIDE VECTOR b(:,1) .
!
                            x2(:n,1) = b(:n,1)
!
!                           COMPUTE THE SOLUTION VECTOR FOR LINEAR TRIANGULAR SYSTEM
!
!                                    a*x(:,1)=b(:,1) OR a**t*x(:,1)=b(:,1) .
!
                            call triang_solve( a(:n,:n), x2(:n,1_i4b), upper=upper, trans=trans )
!
                        case( 4_i4b )
!
!                           TEST 4 : ONE RIGHT HAND-SIDE AND SCALING.
!
!                           MAKE A COPY OF RIGHT HAND-SIDE VECTOR b(:,1) .
!
                            x2(:n,1_i4b) = b(:n,1_i4b)
!
!                           COMPUTE THE SCALED SOLUTION VECTOR FOR LINEAR TRIANGULAR SYSTEM
!
!                                          a*x(:,1)=b(:,1) OR a**t*x(:,1)=b(:,1) .
!
                            call triang_solve( a(:n,:n), x2(:n,1_i4b), scal=scal(1_i4b), upper=upper, trans=trans )
!
!                           UNSCALE SOLUTION VECTOR.
!
                            x2(:n,1_i4b) = x2(:n,1_i4b)/scal(1_i4b)
!
                    end select
!
!                   CHECK THE RESULTS FOR SMALL RESIDUALS.
!
                    if ( k<=2_i4b ) then
!
                         res(:n,:nrhs) = x(:n,:nrhs) - x2(:n,:nrhs)
!
                         err = maxval( norm( res, dim=2_i4b ) / normx )
!
                         nrhs2 = nrhs
!
                    else
!
                         res(:n,1_i4b) = x(:n,1_i4b) - x2(:n,1_i4b)
!
                         err = norm( res(:,1_i4b) ) / normx(1_i4b)
!
                         nrhs2 = 1_i4b
!
                    end if
!
                    m = m + 1_i4b
!
                    eps2 = real( n, stnd)*eps
!
                    if ( err<=eps2 ) then
!
                        success = success + 1_i4b
!
                        if ( details ) then
!
                            write (prtunit,*) 'Test ',m,' is correct'
!
                            write (prtunit,*) 'n= ', n,' , upper= ', upper,' , trans= ', trans,  &
                                              ' , case= ', k,' , nrhs= ', nrhs2
!
!                           PRINT RELATIVE ERRORS OF SOLUTIONS.
!
                            write (prtunit,*) 'Maximum relative error of the computed solutions = ', err
!
                            write (prtunit,*)
!
                        end if
!
                    else
!
                        write (prtunit,*) 'Test ',m,' is incorrect'
!
                        write (prtunit,*) '=> ',err,' > ',eps2,' , n= ', n,' , upper= ', upper,   &
                                         ' , trans= ', trans,' , case= ', k, ' , nrhs= ', nrhs2
!
!                       PRINT RELATIVE ERRORS OF SOLUTIONS.
!
                        write (prtunit,*) 'Maximum relative error of the computed solutions = ', err
!
                        write (prtunit,*)
!
                    end if
!
                end do
!
            end do
!
        end do
!
    end do
!
!   ALLOCATE WORK ARRAYS.
!
    deallocate( normx, scal, a, res, b, x, x2, ap )
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_triang_solve
! ================================
!
end program test_triang_solve
