program test_trid_deflate
!
!
! Purpose
! =======
!
!   This program is intended to test subroutine TRID_DEFLATE
!   in module EIG_Procedures.
!                                                                              
!                                                                              
! Further Details
! ===============
!
!    Random matrices with given eigenvalue value distributions are generated for checking the accuracy
!    of subroutine TRID_DEFLATE.
!
!    The program assumes that subroutine SYMTRID_BISECT in module EIG_Procedures is correct.
!
! LATEST REVISION : 26/09/2018
!                                                                              
! ================================================================================================
!
!
! USED MODULES 
! ============
!
    use Statpack, only : lgl, i4b, stnd, n1_def, n2_def, n3_def,                       &
                         true, false, zero, half, one, two, five, ten, c100,           &
                         qr_cmp, ortho_gen_qr, eigval_sort, symtrid_cmp, trid_deflate, &
                         symtrid_bisect, norm, arth, lamch, triangle
!   
#ifdef _F2003
    use ieee_arithmetic, only : ieee_support_datatype, ieee_is_normal
#endif
!   
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!   
!   
! PARAMETERS 
! ==========
!
! prtunit IS THE PRINTING UNIT
!
    integer(i4b), parameter :: prtunit=6, n1=n1_def, n2=n2_def, n3=n3_def, neig_def=5
!   
    real(stnd), parameter  :: fudge=c100
!
    character(len=*), parameter :: name_proc='trid_deflate'
!   
!
! SPECIFICATIONS FOR VARIABLES
! ============================
!
    real(stnd)                              :: err1, err2, err, eps, eps0, eps2, ulp, safmin,    &
                                               abstol, tmp, norma
    real(stnd), allocatable, dimension(:)   :: eigval, eigval2, d, e, diagr, beta, temp, vec
    real(stnd), allocatable, dimension(:,:) :: a, a2, eigvec, eigvec2, res
!
    integer      :: iok
    integer(i4b) :: i, j, k, l, n, m, o, p, q, neig, success, notdone, max_qr_steps=12_i4b
!
    logical(lgl) :: failure, failure2, details, ortho, upper, packed
!   
    character    :: sort='d'
!   
!   
! EXECUTABLE STATEMENTS
! =====================
!
!   PRINT LABEL OF THE TEST.
!
    write (prtunit,*)
    write (prtunit,*) name_proc,' :'
    write (prtunit,*) repeat('*', len(name_proc) )
    write (prtunit,*)
!
#ifdef _TEST_DETAILS
    details = .true.
#else
    details = .false.
#endif
!
!   SET THE REQUIRED PRECISION OF THE RESULTS.
!
    ulp    = epsilon( err )
    eps    = fudge*ulp
    eps0   = sqrt( ulp )
!    eps0   = eps
    safmin = lamch( 'S' )
!
    abstol = sqrt( safmin )
!
    notdone = 0_i4b
    success = 0_i4b
    m       = 0_i4b
!
    do n = n1, n2, n3
!
!       ALLOCATE WORK ARRAYS.
!
        p = n*(n+1_i4b)/2_i4b
!
        allocate( eigval(n), eigval2(n), d(n), e(n), diagr(n), beta(n), vec(p),  &
                  temp(n), a(n,n), a2(n,n), res(n,n), eigvec(n,n), eigvec2(n,n), &
                  stat=iok     )
!
        if ( iok/=0 ) then
!
            if ( allocated(d) )         deallocate( d )
            if ( allocated(e) )         deallocate( e )
            if ( allocated(diagr) )     deallocate( diagr )
            if ( allocated(beta) )      deallocate( beta )
            if ( allocated(eigval) )    deallocate( eigval )
            if ( allocated(eigvec) )    deallocate( eigvec )
            if ( allocated(eigval2) )   deallocate( eigval2 )
            if ( allocated(eigvec2) )   deallocate( eigvec2 )
            if ( allocated(a) )         deallocate( a )
            if ( allocated(a2) )        deallocate( a2 )
            if ( allocated(res) )       deallocate( res )
            if ( allocated(vec) )       deallocate( vec )
            if ( allocated(temp) )      deallocate( temp )
!
            cycle
!
        end if
!
!       GENERATE A RANDOM COEFFICIENT MATRIX eigvec(:n,:n) .
!
        call random_number( eigvec(:n,:n) )
!
        eigvec(:n,:n) = eigvec(:n,:n) - half
!
        call random_number( tmp )
!
        if ( tmp>safmin ) then
            eigvec(:n,:n) = eigvec(:n,:n)/tmp
        end if
!
!       COMPUTE QR DECOMPOSITION OF RANDOM DATA MATRIX.
!
        call qr_cmp( eigvec(:n,:n), diagr(:n), beta(:n) )
!
!       GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM DATA MATRIX.
!       q IS ASSUMED OF FULL RANK.
!
        call ortho_gen_qr( eigvec(:n,:n), beta(:n) )
!
        do j = 1_i4b, 16_i4b
!
            select case( j )
!
                case( 1_i4b )
!
!                   TEST 1: STRONGLY CLUSTERED EIGENVALUES.
!
                    eigval(1_i4b)   = one
                    eigval(2_i4b:n) = ulp
!
                case( 2_i4b )
!
!                   TEST 2:  STRONGLY CLUSTERED EIGENVALUES.
!
                    eigval(1_i4b:n-1_i4b) = one
                    eigval(n)             = ulp
!
                case( 3_i4b )
!
!                   TEST 3: WEAKLY CLUSTERED EIGENVALUES.
!
                    eigval(1_i4b:n-1_i4b) = arth( ulp, ulp, n-1_i4b )
                    eigval(n)             = one
!
                case( 4_i4b )
!
!                   TEST 4: INTERNAL CLUSTER OF EIGENVALUES.
!
                    eigval(1_i4b)         = ulp
                    eigval(2_i4b:n-1_i4b) = one + sqrt(ulp)*arth( one, one, n-2_i4b )
                    eigval(n)             = two
!
                case( 5_i4b )
!
!                   TEST 5: WEAKLY CLUSTERED EIGENVALUES.
!
                    eigval(1_i4b:n) = one + (c100*ulp)*arth( zero, one, n )
!
                case( 6_i4b )
!
!                   TEST 6: GEOMETRIC DISTRIBUTION OF EIGENVALUES.
!
                    err = real( n-1_i4b, stnd )
!
                    do l = 1_i4b, n
                        eigval(l) = ulp**(real(l-1_i4b,stnd)/err)
                    end do
!
                case( 7_i4b )
!
!                   TEST 7: ARITHMETIC DISTRIBUTION OF EIGENVALUES.
!
                    eigval(1_i4b:n) = ulp + ((one-ulp)/real( n-1_i4b, stnd ))*arth( zero, one, n )
!
                case( 8_i4b )
!
!                   TEST 8: TWO CLUSTERS OF EIGENVALUES.
!
                    i = n/2_i4b
!
                    tmp = sqrt( ulp )
!
                    eigval(1_i4b:i) = one + tmp*arth( one, one, i )
!
                    eigval(i+1_i4b:n) = ten + tmp*arth( one, one, n-i )
!
                case( 9_i4b )
!
!                   TEST 9: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR ZERO.
!
                    i = n/2_i4b
!
                    tmp = sqrt( ulp )
!
                    eigval(1_i4b:i) = arth( one, one, i )
!
                    eigval(i+1_i4b:n) = tmp*arth( one, one, n-i )
!
                case( 10_i4b )
!
!                   TEST 10: ARITHMETIC DISTRIBUTION OF EIGENVALUES AND ONE BIG CLUSTER NEAR n/2.
!
                    i = n/2_i4b
!
                    tmp = sqrt( ulp )
!
                    eigval(1_i4b:i) = arth( one, one, i )
!
                    eigval(i+1_i4b:n) = real( i, stnd) + tmp*arth( one, one, n-i )
!
                case( 11_i4b )
!
!                   TEST 11: CLUSTERED DISTRIBUTION OF EIGENVALUES BETWEEN 0.9 and 1.
!
                    call random_number( eigval(:n) )
!
                    eigval(1_i4b:n) = 0.9_stnd + 0.1_stnd*eigval(1_i4b:n)
!
                case( 12_i4b )
!
!                   TEST 12: EIGENVALUES EQUAL TO sinh(10*i/n) FOR i=1 TO n.
!
                    tmp = ten/real( n, stnd)
!
                    eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) )
!
                case( 13_i4b )
!
!                   TEST 13: EIGENVALUES EQUAL TO sinh((10*i/n)-5) FOR i=1 TO n.
!
                    tmp = ten/real( n, stnd)
!
                    eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - five )
!
                case( 14_i4b )
!
!                   TEST 14: EIGENVALUES EQUAL TO sinh((10*i/n)-10) FOR i=1 TO n.
!
                    tmp = ten/real( n, stnd)
!
                    eigval(1_i4b:n) = sinh( tmp*arth( one, one, n ) - ten )
!
                case( 15_i4b )
!
!                   TEST 15: EIGENVALUES EQUAL TO tanh((10*i/n)-5) FOR i=1 TO n.
!
                    tmp = ten/real( n, stnd)
!
                    eigval(1_i4b:n) = tanh( tmp*arth( one, one, n ) - five )
!
                case default
!
!                   DEFAULT TEST : RANDOMLY UNIFORM DISTRIBUTION OF EIGENVALUES.
!
                    call random_number( eigval(:n) )
!
                    call random_number( tmp )
!
                    if ( tmp>safmin ) then
                        eigval(1_i4b:n) = eigval(1_i4b:n)/tmp
                    end if
!
            end select
!
#ifdef _F2003
            if ( ieee_support_datatype( eigval ) ) then
                if ( .not.all( ieee_is_normal( eigval(1_i4b:n) ) ) ) cycle
            end if
#endif
!
            do q = 1_i4b, 2_i4b
!
                if ( q==2_i4b ) then
!
!                   EIGENVALUES ARE RANDOMLY NEGATED.
!
                    call random_number( temp(:n) )
!
                    where ( temp(:n)>half ) eigval(:n) = -eigval(:n)
!
                end if
!
!               SORT THE EIGENVALUES.
!
                call eigval_sort( sort, eigval(:n) )
!
                where ( eigval(:n)==zero )
!
                    temp(:n) = one
!
                elsewhere
!
                    temp(:n) = eigval(:n)
!
                end where
!
!               FORM A SYMMETRIC MATRIX FROM THE SPECIFIED EIGENVALUE DISTRIBUTION AND RANDOM EIGENVECTORS.
!
                a(:n,:n) = matmul( eigvec(:n,:n)*spread(eigval,1,n), transpose( eigvec(:n,:n) ) )
!
!               COMPUTE THE NORM OF THE SYMMETRIC MATRIX.
!
                norma = norm( eigval(:n) )
!
                upper  = true
                packed = false
!
                do o = 1_i4b, 6_i4b
!
!                   REDUCE THE SYMMETRIC MATRIX TO SYMMETRIC TRIDIAGONAL FORM.
!
                    select case( o )
!
                        case( 1_i4b )
!
!                           MAKE A COPY OF THE ORIGINAL SYMMETRIC MATRIX.
!
                            a2(:n,:n) = a(:n,:n)
!
                            call symtrid_cmp( a2(:n,:n), d(:n), e(:n), store_q=true )
!
                        case( 2_i4b, 3_i4b )
!
                            upper = .not.upper
!
!                           MAKE A COPY OF THE ORIGINAL SYMMETRIC MATRIX.
!
                            a2(:n,:n) = a(:n,:n)
!
                            call symtrid_cmp( a2(:n,:n), d(:n), e(:n), upper=upper, store_q=true )
!
                        case( 4_i4b )
!
                            packed = true
!
!                           MAKE A COPY OF THE SELF-ADJOINT MATRIX IN PACKED FORM.
!
                            vec(:p) = pack( a(:n,:n), mask=triangle( true, n, n, extra=1_i4b) )
!
                            call symtrid_cmp( vec(:p), d(:n), e(:n), store_q=true )
!
                        case( 5_i4b, 6_i4b )
!
                            upper = .not.upper
!
!                           MAKE A COPY OF THE SELF-ADJOINT MATRIX IN PACKED FORM.
!
                            vec(:p) = pack( a(:n,:n), mask=triangle( upper, n, n, extra=1_i4b) )
!
                            call symtrid_cmp( vec(:p), d(:n), e(:n), upper=upper, store_q=true )
!
                    end select
!
!                   COMPUTE ALL EIGENVALUES OF THE SYMMETRIC MATRIX a TO HIGH ACCURACY.
!
                    call symtrid_bisect( d(:n), e(:n), neig, eigval2(:n), failure, &
                                         sort=sort, abstol=abstol )
!
                    if ( failure ) cycle
!
!                   COMPUTE EIGENVECTORS OF THE SYMMETRIC MATRIX.
!
                    do k = 1_i4b, 2_i4b
!
                        if ( k==1_i4b ) then
                            neig = min( n, neig_def )
                        else
                            neig = n
                        end if
!
                        do i = 1_i4b, 2_i4b
!
!                           COMPUTE EIGENVECTORS OF THE SYMMETRIC MATRIX BY A DEFLATION METHOD.
!
                            if ( i==1_i4b ) then
!
                                eps2 = eps
!
                                ortho = true
!
                            else
!
                                eps2 = eps0
!
                                ortho = false
!
                            end if
!
                            if ( o<=3_i4b ) then
!
                                call trid_deflate( d(:n), e(:n), eigval2(:neig), eigvec2(:n,:neig), failure2,   &
                                                   mat=a2(:n,:n), max_qr_steps=max_qr_steps, ortho=ortho )
!
                            else
!
                                call trid_deflate( d(:n), e(:n), eigval2(:neig), eigvec2(:n,:neig), failure2,   &
                                                   matp=vec(:p), max_qr_steps=max_qr_steps, ortho=ortho )
!
                            end if
!
                            if ( failure2 ) then
!
                                notdone = notdone + 1_i4b
!
                                cycle
!
                            end if
!
                            m = m + 1_i4b
!
!                           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*v - u*s
!                           WHERE s ARE THE EIGENVALUES AND u THE EIGENVECTORS OF THE MATRIX.
!
                            res(:n,:neig) = matmul( a(:n,:n), eigvec2(:n,:neig) )        &
                                            - eigvec2(:n,:neig)*spread(eigval2(:neig),1,n)
!
                            if ( norma/=zero ) then
                                err1 = maxval( norm(res(:n,:neig), dim=2_i4b) )/( norma*real(n,stnd) )
                            else
                                err1 = maxval( norm(res(:n,:neig), dim=2_i4b) )/real(n,stnd)
                            end if
!
!                           CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!                           WHERE u are THE EIGENVECTORS OF THE MATRIX.
!
                            res(:neig,:neig) = matmul( transpose( eigvec2(:n,:neig) ), eigvec2(:n,:neig) )
!
                            do l = 1_i4b, neig
                                res(l,l) = one - res(l,l)
                            end do
!
                            err2 = maxval( norm(res(:neig,:neig), dim=2_i4b) )/real(n,stnd)
!
                            if ( err1<=eps .and. err2<=eps2 ) then
!
                                success = success + 1_i4b
!
                                if ( details ) then
!
                                    write (prtunit,*) 'Test ',m,' is correct'
!
                                    if ( o==1_i4b .or. o==4_i4b ) then
!
                                        write (prtunit,*) 'n= ', n,' , neig= ', neig,' , case= ', j,      &
                                                          ' , packed= ', packed,' , ortho= ', ortho
!
                                    else
!
                                        write (prtunit,*) 'n= ', n,' , neig= ', neig,' , case= ', j,      &
                                                          ' , packed= ', packed,' , upper= ', upper,      &
                                                          ' , ortho= ', ortho
!
                                    end if
!
!                                   COMPUTE ABSOLUE ERRORS OF EIGENVALUES.
!
                                    write (prtunit,*) 'Maximum absolute error of the computed eigenvalues = ', &
                                                      maxval( abs( eigval2(:neig)-eigval(:neig) ) )
!
!                                   COMPUTE RELATIVE ERRORS OF EIGENVALUES.
!
                                    write (prtunit,*) 'Maximum relative error of the computed eigenvalues = ', &
                                                      maxval( abs( (eigval2(:neig)-eigval(:neig) )/temp(:neig) ) )
!
!                                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*u - u*s
!                                   WHERE s ARE THE EIGENVALUES AND u THE EIGENVECTORS OF THE MATRIX.
!
                                    write (prtunit,*) 'Accuracy of the computed eigenpairs                = ', &
                                                      err1/ulp, ' (must be less than :', fudge, ' )'
!
!                                   CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!                                   WHERE u are THE EIGENVECTORS OF THE MATRIX.
!
                                    write (prtunit,*) 'Orthogonality of the computed eigenvectors         = ', &
                                                      err2/ulp, ' (must be less than :', fudge, ' )'
!
                                    write (prtunit,*)
!
                                end if
!
                            else
!
                                write (prtunit,*) 'Test ',m,' is incorrect'
!
                                if ( i==1_i4b ) then
!
                                    err = max( err1, err2 )
!
                                    if ( o==1_i4b .or. o==4_i4b ) then
!
                                        write (prtunit,*) '=> ',err,' > ', eps,' , n= ', n,' , neig= ', neig,          &
                                                          ' , case= ', j,' , packed= ', packed,' , ortho= ', ortho,    &
                                                          ' , failure= ', failure2
!
                                    else
!
                                        write (prtunit,*) '=> ',err,' > ', eps,' , n= ', n,' , neig= ', neig,         &
                                                          ' , case= ', j,' , packed= ', packed,' , upper= ', upper,   &
                                                          ' , ortho= ', ortho,' , failure= ', failure2
!
                                    end if
!
                                else
!
                                    if ( o==1_i4b .or. o==4_i4b ) then
!
                                        write (prtunit,*) '=> ',err1,' > ', eps,' or ', err2,' > ',eps2, ' , n= ', n,  &
                                                          ' , neig= ', neig,' , case= ', j,' , packed= ', packed,      &
                                                          ' , ortho= ', ortho,' , failure= ', failure2
!
                                    else
!
                                        write (prtunit,*) '=> ',err1,' > ', eps,' or ', err2,' > ',eps2, ' , n= ', n,      &
                                                          ' , neig= ', neig,' , case= ', j,' , packed= ', packed,          &
                                                          ' , upper= ', upper,' , ortho= ', ortho,' , failure= ', failure2
!
                                    end if
!
                                end if
!
!                               COMPUTE ABSOLUE ERRORS OF EIGENVALUES.
!
                                write (prtunit,*) 'Maximum absolute error of the computed eigenvalues = ', &
                                                  maxval( abs( eigval2(:neig)-eigval(:neig) ) )
!
!                               COMPUTE RELATIVE ERRORS OF EIGENVALUES.
!
                                write (prtunit,*) 'Maximum relative error of the computed eigenvalues = ', &
                                               maxval( abs( (eigval2(:neig)-eigval(:neig) )/temp(:neig) ) )
!
!                               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION a*u - u*s
!                               WHERE s ARE THE EIGENVALUES AND u THE EIGENVECTORS OF THE MATRIX.
!
                                write (prtunit,*) 'Accuracy of the computed eigenpairs                = ', &
                                                  err1/ulp, ' (must be less than :', fudge, ' )'
!
!                               CHECK FOR SMALL RESIDUALS OF THE EXPRESSION I - u**(t)*u
!                               WHERE u are THE EIGENVECTORS OF THE MATRIX.
!
                                write (prtunit,*) 'Orthogonality of the computed eigenvectors         = ', &
                                                  err2/ulp, ' (must be less than :', fudge, ' )'
!
                                write (prtunit,*)
!
                            end if
!
                        end do
!
                    end do
!
                end do
!
            end do
!
        end do
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( eigval, eigval2, d, e, temp, vec,        &
                    diagr, beta, a, a2, eigvec, eigvec2, res )
!
    end do
!
    if ( notdone/=0_i4b ) then
        write (prtunit,*) notdone, 'tests not done'
    end if
!
    if ( m/=success ) then
        write (prtunit,*) m - success, 'tests are incorrect'
    end if
!
    write (prtunit,*) success, 'tests are correct'
!
!
! END OF PROGRAM test_trid_deflate
! ================================
!
end program test_trid_deflate
