module Eig_Procedures
! ##############################################################################
!
! Copyright 2020 IRD
!
! This file is part of statpack.
!
! statpack is free software: you can redistribute it and/or modify
! it under the terms of the GNU Lesser General Public License as
! published by the Free Software Foundation, either version 3 of 
! the License, or (at your option) any later version.
!
! statpack is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU Lesser General Public License for more details.
!
! You can find a copy of the GNU Lesser General Public License
! in the statpack/doc directory.
!
! ##############################################################################
!                                                                              *
! ******************************************************************************
! MODULE EXPORTING PROCEDURES FOR COMPUTING (SELECTED) EIGENVALUES AND/OR      *
! (SELECTED) EIGENVECTORS OF A SYMMETRIC (TRIDIAGONAL) MATRIX.                 *
!                                                                              *
! SUBROUTINES FOR COMPUTING PARTIAL EIGENVALUE DECOMPOSITION OF SYMMETRIC      *
! MATRICES BASED ON RANDOMIZED ALGORITHMS ARE ALSO PROVIDED.                   *
!                                                                              *
! LATEST REVISION : 25/10/2020                                                 *
!                                                                              *
! ##############################################################################
!                                                                              *
! ******************************************************************************
!                                                                              *
! SOME OPENMP DIRECTIVES IN THIS MODULE ARE ACTIVATED WITH THE C PROCESSOR     *
! MACRO _OPENMP2 AND, BY DEFAULT, THESE DIRECTIVES ARE ACTIVATED IF OPENMP IS  *
! USED. THESE OPENMP DIRECTIVES ACTIVATE OPENMP PARALLELIZATION IN THE         *
! GENERIC TRIDIAGONAL QR SUBROUTINES symtrid_qri, symtrid_qri2, symtrid_qri3.  *
! HOWEVER, IN CASE OF PROBLEMS, THESE OPENMP DIRECTIVES MAY BE DESACTIVATED    *
! WITH THE C PROCESSOR MACRO _NOOPENMP2 DURING COMPILATION.                    *
!                                                                              *
! MORE GENERALLY, THE C PROCESSOR MACROS USED IN THIS MODULE ARE:              *
!                                                                              *
!  _OPENMP      FOR ACTIVATING OPENMP PARALLELIZATION                          *
!  _OPENMP2     FOR ACTIVATING OPENMP PARALLELIZATION IN symtrid_qri,          *
!               symtrid_qri2 AND symtrid_qri3 SUBROUTINES                      *
!  _F2003       FOR ACTIVATING FORTRAN 2003 CONSTRUCTS                         *
!  _BLAS        FOR USING BLAS WHEN POSSIBLE                                   *
!  _DOT_PRODUCT FOR REPLACING THE dot_product INTRINSIC FUNCTION WITH STATPACK *
!               dot_product2 FUNCTION                                          *
!  _MATMUL      FOR REPLACING THE matmul INTRINSIC FUNCTION WITH STATPACK      *
!               matmul2 FUNCTION                                               *
!                                                                              *
! ******************************************************************************
!
#ifdef _OPENMP
!
#ifndef _NOOPENMP2
#define _OPENMP2
#endif
!
#else
!
#undef _OPENMP2
!
#endif
!
! USED MODULES
! ============
!   
    use Select_Parameters,  only : lgl, i4b, stnd, extd, blksz_eig, blksz2_eig,  &
                                   max_francis_steps_eig
    use Reals_Constants,    only : zero, quarter, half, one, two, three, four,   &
                                   five, six, eight, ten, c0_6, c20, c50, c400
    use Logical_Constants
#ifdef _MATMUL
    use Utilities,          only : matmul=>matmul2
#endif
#ifdef _DOT_PRODUCT
    use Utilities,          only : dot_product=>dot_product2
#endif
#ifdef _OPENMP
    use Select_Parameters,  only : omp_limit, omp_limit2
    use omp_lib,            only : omp_get_max_threads, omp_in_parallel, omp_get_nested,  &
                                   omp_get_num_threads, omp_get_thread_num
#endif
!
! USED MODULES IN SPECIFIC PROCEDURES
! ===================================
! 
!   use Num_Constants
!   use Char_Constants
!   use Utilities
!   use Utilities_With_Pnter
!   use Sort_Procedures
!   use Random
!   use Giv_Procedures
!   use QR_Procedures
!   use BLAS_interfaces
!   use ieee_arithmetic  (if the CPP macro _F2003 is activated)
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!
! PUBLIC ENTITIES 
! ===============
!    
!   ALL SUBROUTINES, FUNCTIONS, VARIABLES AND PARAMETERS ARE PRIVATE BY DEFAULT.
!
    private
    public ::   symtrid_cmp, symtrid_cmp2, ortho_gen_symtrid, apply_q_symtrid,    &
                symtrid_qri, symtrid_qri2, symtrid_qri3,                          &
                symtrid_ratqri, symtrid_ratqri2, symtrid_bisect,                  &
                dflgen, dflgen2, dflapp, qrstep, prodgiv, prodgiv_eigvec,         &
                symtrid_deflate, trid_deflate,                                    &
                eig_cmp, eig_cmp2, eig_cmp3, laev2, eig_sort, eig_abs_sort,       &
                eigvalues, eigval_cmp, eigval_cmp2, eigval_cmp3, lae2,            &
                eigval_sort, eigval_abs_sort, reig_cmp,                           &
                select_eigval_cmp, select_eigval_cmp2, select_eigval_cmp3,        &
                maxdiag_tinv_qr, maxdiag_tinv_ldu, trid_qr_cmp, trid_qr_solve,    &
                trid_cmp, trid_cmp2, trid_solve, trid_inviter,                    &
                gen_symtrid_mat
!
! GENERIC INTERFACES FOR ROUTINES WITH OVERLOADED VERSIONS
! ========================================================
!
    interface symtrid_cmp
        module procedure    symtrid_cmp_a, symtrid_cmp_b, symtrid_cmp_c, symtrid_cmp_d
    end interface
!
    interface ortho_gen_symtrid
        module procedure    ortho_gen_symtrid_a, ortho_gen_symtrid_b
    end interface
!
    interface apply_q_symtrid
        module procedure    apply_q_symtrid_a, apply_q_symtrid_b, apply_q_symtrid_c, apply_q_symtrid_d
    end interface
!
    interface symtrid_qri
        module procedure    symtrid_qri_a, symtrid_qri_b
    end interface
!
    interface symtrid_qri2
        module procedure    symtrid_qri2_a, symtrid_qri2_b
    end interface
!
    interface symtrid_qri3
        module procedure    symtrid_qri3_a, symtrid_qri3_b
    end interface
!
    interface symtrid_deflate
        module procedure    symtrid_deflate_rv, symtrid_deflate_rm
    end interface
!
    interface trid_deflate
        module procedure    trid_deflate_a, trid_deflate_b, trid_deflate_c
    end interface
!
    interface eig_cmp
        module procedure    eig_cmp_a, eig_cmp_b
    end interface
!
    interface eig_cmp2
        module procedure    eig_cmp2_a, eig_cmp2_b
    end interface
!
    interface eig_cmp3
        module procedure    eig_cmp3_a, eig_cmp3_b
    end interface
!
    interface eigvalues
        module procedure    eigvalues_c, eigvalues_d
    end interface
!
    interface eigval_cmp
        module procedure    eigval_cmp_a, eigval_cmp_b, eigval_cmp_c, eigval_cmp_d
    end interface
!
    interface eigval_cmp2
        module procedure    eigval_cmp2_a, eigval_cmp2_b, eigval_cmp2_c, eigval_cmp2_d
    end interface
!
    interface eigval_cmp3
        module procedure    eigval_cmp3_a, eigval_cmp3_b, eigval_cmp3_c, eigval_cmp3_d
    end interface
!
    interface eig_sort
        module procedure    eig_sort_a, eigval_sort
    end interface
!
    interface eig_abs_sort
        module procedure    eig_abs_sort_a, eigval_abs_sort
    end interface
!
    interface select_eigval_cmp
        module procedure    select_eigval_cmp_a, select_eigval_cmp_b,    &
                            select_eigval_cmp_c, select_eigval_cmp_d
    end interface
!
    interface select_eigval_cmp2
        module procedure    select_eigval_cmp2_a, select_eigval_cmp2_b,   &
                            select_eigval_cmp2_c, select_eigval_cmp2_d
    end interface
!
    interface select_eigval_cmp3
        module procedure    select_eigval_cmp3_a, select_eigval_cmp3_b,   &
                            select_eigval_cmp3_c, select_eigval_cmp3_d
    end interface
!
    interface trid_qr_cmp
        module procedure    trid_qr_cmp_rv
    end interface
!
    interface trid_qr_solve
        module procedure    trid_qr_solve_rv
    end interface
!
    interface trid_cmp
        module procedure    trid_cmp_rm, trid_cmp_rv
    end interface
!
    interface trid_cmp2
        module procedure    trid_cmp2_rm, trid_cmp2_rv
    end interface
!
    interface trid_solve
        module procedure    trid_solve_rm, trid_solve_rv
    end interface
!
    interface trid_inviter
        module procedure    trid_inviter_rv, trid_inviter_a, trid_inviter_b, trid_inviter_c
    end interface
!
!   
! MODULE PARAMETERS 
! =================
!
!   BLOCK SIZE FOR REDUCTION OF A SQUARE MATRIX TO TRIDIAGONAL FORM
!   IN symtrid_cmp_b AND symtrid_cmp_d SUBROUTINES, FOR AGREGATING AND APPLYING
!   HOUSEHOLDER TRANSFORMATIONS IN apply_q_symtrid_a, apply_q_symtrid_b, apply_q_symtrid_c,
!   ortho_gen_symtrid_a, ortho_gen_symtrid_b, trid_deflate_b, trid_deflate_c, trid_inviter_b
!   AND trid_inviter_c SUBROUTINES.
!
    integer(i4b), parameter :: blksz = blksz_eig
!
!   BLOCK SIZE FOR REDUCTION OF A MATRIX PRODUCT TO TRIDIAGONAL FORM IN symtrid_cmp2.
!
    integer(i4b), parameter :: blksz2 = blksz2_eig
!
!   DEFAULT BLOCK SIZE FOR THE WAVE FRONT ALGORITHM USED TO APPLY GIVENS ROTATIONS
!   IN THE TRIDIAGONAL IMPLICIT QR ALGORITHM IN symtrid_qri2_a AND symtrid_qri3_a SUBROUTINES.
!
    integer(i4b), parameter :: max_francis_steps_default = max_francis_steps_eig
!
!   A "FUDGE FACTOR" TO DETERMINE WHEN A DEFLATION STEP IS ACCEPTED IN dflgen2 AND
!   dflapp SUBROUTINES. IDEALLY, A VALUE OF 1 SHOULD WORK, BUT A LARGER OR LOWER VALUE
!   CAN BE USED. NOTE THAT THIS PARAMETER HAS EFFECT ON ACCURACY OF THE SOLUTION AND SPEED
!   OF COMPUTATION. A SMALLER VALUE WILL GIVE HIGHER ACCURACY AT THE EXPENSE OF THE SPEED
!   OF THE COMPUTATION.
!
    real(stnd),  parameter :: fudge_deflate = one
!
!   A "FUDGE FACTOR" TO DETERMINE WHEN ORTHOGONALIZATION MUST BE PERFORMED IN trid_inviter
!   SUBROUTINES OR WHEN SUCCESSIVE DEFLATION MUST BE PERFORMED IN trid_deflate SUBROUTINES.
!   IDEALLY, A VALUE OF 1 SHOULD WORK, BUT A LARGER VALUE CAN BE USED FOR IMPROVED 
!   ORTHOGONALITY AT THE EXPENSE OF SPEED OF COMPUTATION. NOTE THAT THIS PARAMETER HAS 
!   EFFECT ON ACCURACY AND ORTHOGONALITY OF EIGENVECTORS PRODUCED BY trid_inviter AND
!   trid_deflate SUBROUTINES.
!
    real(stnd),  parameter :: fudge_ortho = c50
!
!   A "FUDGE FACTOR" TO DETERMINE WHEN REORTHOGONALIZATION OF EIGENVECTORS MUST
!   BE PERFORMED IN trid_inviter SUBROUTINES. A VALUE BETWEEN 0 AND 1 MUST BE SELECTED.
!   IDEALLY, A VERY SMALL VALUE SHOULD WORK, BUT A LARGER VALUE CAN BE USED FOR IMPROVED
!   ORTHOGONALITY AT THE EXPENSE OF SPEED OF COMPUTATION. NOTE THAT THIS PARAMETER HAS
!   EFFECT ON ACCURACY AND ORTHOGONALITY OF THE EIGENVECTORS PRODUCED BY trid_inviter
!   SUBROUTINES.
!
    real(stnd),  parameter :: fudge_ortho2 = c0_6
!
!   A "FUDGE FACTOR" FOR THE TEST OF ACCURACY OF EIGENVECTORS COMPUTED BY trid_inviter AND
!   trid_deflate SUBROUTINES. A LARGE VALUE RELAXES THE TESTS OF ACCURACY IN THESE
!   SUBROUTINES. THIS FUDGE FACTOR ALSO MODIFIES THE AMPLITUDE OF THE RELATIVE PERTURBATION ADDED
!   TO THE EIGENVALUES IF THESE EIGENVALUES ARE TOO CLOSED IN trid_inviter SUBROUTINES.
!
    real(stnd),  parameter :: fudge_prec = c20
!
!   A "FUDGE FACTOR" FOR THE TOLERANCE IN THE symtrid_qri, symtrid_qri2 AND symtrid_qri3
!   TRIDIAGONAL EIGENSOLVER SUBROUTINES. A LARGE VALUE RELAXES THE TESTS OF ACCURACY IN
!   THESE SUBROUTINES.
!
    real(stnd),  parameter :: fudge_eps = one
!
#ifdef _OPENMP
!
!   A THRESHOLD TO DETERMINE SMALL AND BIG CLUSTERS OF EIGENVALUES IN trid_inviter AND
!   trid_deflate SUBROUTINES. NOTE THAT THIS PARAMETER HAS EFFECT ON SPEED AND
!   PARALLELIZATION WHEN OPENMP IS USED, BUT NOT ON ACCURACY OF EIGENVECTORS PRODUCED BY
!   trid_inviter AND trid_deflate SUBROUTINES. neig_cluster MUST BE GREATER OR EQUAL TO 1.
!
    integer(i4b), parameter :: neig_cluster = 20_i4b
!
#endif
!
!
! =========================================================================================
!
                                  contains
!                                 ========
!
! =========================================================================================
!                         SYMMETRIC TRIDIAGONAL DECOMPOSITIONS
! =========================================================================================
!
!
    subroutine symtrid_cmp_a( mat, d, e, store_q, upper )
!
! Purpose
! _______
!
!   SYMTRID_CMP reduces a real n-by-n symmetric matrix MAT to symmetric
!   tridiagonal form T by an orthogonal similarity transformation:
!
!                        Q' * MAT * Q = T
!
!
! Arguments
! _________
!
!   MAT      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry:
!
!            - If UPPER = true :
!              The leading n-by-n upper triangular part of MAT contains the
!              upper triangular part of the symmetric matrix MAT, and the 
!              strictly lower triangular part of MAT is not referenced.
!            - If UPPER = false :
!              The leading n-by-n lower triangular part of MAT contains the
!              lower triangular part of the symmetric matrix MAT, and the
!              strictly upper triangular part of MAT is not referenced.
!
!            On exit:
!
!            - If UPPER = true and STORE_Q = true :
!              The leading n-by-n upper triangular part of MAT is overwritten
!              by the matrix Q as a product of elementary reflectors.
!            - If UPPER = false and STORE_Q = true :
!              The leading n-by-n lower triangular part of MAT is overwritten
!              by the matrix Q as a product of elementary reflectors.
!            - If UPPER = true and STORE_Q = false :
!              The leading n-by-n upper triangular part of MAT is destroyed.
!            - If UPPER = false and STORE_Q = false :
!              The leading n-by-n lower triangular part of MAT is destroyed.
!
!   D        (OUTPUT) real(stnd), dimension(:)
!            The diagonal elements of the tridiagonal matrix T:
!
!            - D(i) = T(i,i).
!
!            The size of D must verify:  size( D ) = size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   E        (OUTPUT) real(stnd), dimension(:)
!            The off-diagonal elements of the tridiagonal matrix T:
!
!            -  E(i) = T(i,i+1) = T(i+1,i)
!            -  E(n) is arbitrary.
!
!            The size of E must verify:  size( E ) = size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   STORE_Q  (INPUT) logical(lgl)
!            On exit:
!
!            - If UPPER = true and STORE_Q = true :
!              The leading n-by-n upper triangular part of MAT is overwritten
!              by the matrix Q as a product of elementary reflectors.
!            - If UPPER = false and STORE_Q = true :
!              The leading n-by-n lower triangular part of MAT is overwritten
!              by the matrix Q as a product of elementary reflectors.
!            - If UPPER = true and STORE_Q = false :
!              The leading n-by-n upper triangular part of MAT is destroyed.
!            - If UPPER = false and STORE_Q = false :
!              The leading n-by-n lower triangular part of MAT is destroyed.
!
!   UPPER    (INPUT) logical(lgl)
!            Specifies whether the upper or lower triangular part of the
!            symmetric matrix MAT is stored. If:
!
!            - UPPER= true :  Upper triangular is stored ;
!            - UPPER= false:  Lower triangular is stored .
!
!
! Further Details
! _______________
!
!   If UPPER = true and STORE_Q = true, the matrix Q is represented
!   as a product of elementary reflectors
!
!        Q = H(n-1) * . . . * H(2) * H(1).
!
!   Each H(i) has the form
!
!        H(i) = I + tau * v * v'
!
!   where tau is a real scalar, and v is a real vector with
!   v(i+1:n) = 0 ; v(1:i) is stored on exit in MAT(1:i,i+1),
!   and tau in MAT(i+1,i+1).
!
!   If UPPER = false and STORE_Q = true, the matrix Q is represented
!   as a product of elementary reflectors
!
!        Q = H(1) * H(2) * . . . * H(n-1).
!
!   Each H(i) has the form
!
!        H(i) = I + tau * v * v'
!
!   where tau is a real scalar, and v is a real vector with
!   v(1:i) = 0 ; v(i+1:n) is stored on exit in MAT(i+1:n,i),
!   and tau in MAT(i,i).
!
!   The contents of MAT on exit are illustrated by the following examples
!   with n = 5:
!
!   if UPPER = true and STORE_Q = true :
!
!    (  xx  v1  v2  v3  v4  )
!
!    (  yy  t1  v2  v3  v4  )
!
!    (  yy  yy  t2  v3  v4  )
!
!    (  yy  yy  yy  t3  v4  )
!
!    (  yy  yy  yy  yy  t4  )
!
!
!   if UPPER = false and STORE_Q = true :
!
!    (  t1  yy  yy  yy  yy  )
!
!    (  v1  t2  yy  yy  yy  )
!
!    (  v1  v2  t3  yy  yy  )
!
!    (  v1  v2  v3  t4  yy  )
!
!    (  v1  v2  v3  v4  xx  )
!
!
!   where vi and ti denote an element of the vector v and the scalar tau
!   defining H(i), respectively. xx = machhuge and is used by the subroutine
!   ORTHO_GEN_SYMTRID in order to verify that SYMTRID_CMP has been called
!   before ORTHO_GEN_SYMTRID. Elements yy are not modified by the subroutine.
!
!   This subroutine is adapted from the routine  DSYTD2 in LAPACK. Note that
!   this subroutine is not blocked and not parallelized.
!
!   For more details on the reduction algorithm used in SYMTRID_CMP, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : machhuge
    use Utilities,       only : assert_eq
#ifdef _BLAS
    use BLAS_interfaces, only : axpy, syr2
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:,:) :: mat
    real(stnd), intent(out),   dimension(:)   :: d, e
!
    logical(lgl), intent(in) :: store_q, upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: alpha, gam, tau, scal
    real(stnd), dimension(size(d)) :: v, w
!
    integer(i4b) :: n, k, i, j
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='symtrid_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(d),i4b)     ,   &
                    int(size(e),i4b)     ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!
    if ( upper ) then
!
!       REDUCE THE UPPER TRIANGLE OF mat .
!
        k = n
!
        do i = n-1_i4b, 1_i4b, -1_i4b
!
            d(k) = mat(k,k)
            scal = maxval( abs( mat(1_i4b:i-1_i4b,k) ) )
!
            if ( scal>zero ) then
!
!               GENERATE ELEMENTARY REFLECTOR H(i) = I + tau * v * v'
!               TO ANNIHILATE mat(1:i-1,i+1) STORING v IN v(1:i) .
!
                scal = max( scal, abs(mat(i,k)) )
                v(:i) = mat(:i,k)/scal
                alpha = -sign( sqrt( dot_product( v(:i), v(:i) ) ), v(i) )
                v(i) = v(i) - alpha
                tau = one/( alpha*v(i) )
!
!               APPLY H(i) FROM BOTH SIDES TO mat(1:i,1:i) :
!
!               COMPUTE  x := tau * mat * v  STORING x IN w(1:i) .
!
                w(:i) = zero
!
                do j = 1_i4b, i
!
#ifdef _BLAS
                    call axpy( j-1_i4b, v(j), mat(:j-1_i4b,j), 1_i4b, w(:j-1_i4b), 1_i4b )
#else
                    w(:j-1_i4b) = w(:j-1_i4b) + v(j)*mat(:j-1_i4b,j)
#endif
                    w(j) = w(j) + dot_product( mat(:j,j), v(:j) )
!
                end do
!
                w(:i) = w(:i)*tau
!
!               COMPUTE  w := x + 1/2 * tau * (x'*v) * v STORING w IN w(1:i) .
!
                gam = half*tau*dot_product(w(:i), v(:i) )
#ifdef _BLAS
                call axpy( i, gam, v(:i), 1_i4b, w(:i), 1_i4b )
#else
                w(:i) = w(:i) + gam*v(:i)
#endif
!
!               APPLY THE TRANSFORMATION AS A RANK-2 UPDATE :
!                     mat := mat + v * w' + w * v'
!
#ifdef _BLAS
                call syr2( 'U', i, one, w, 1_i4b, v, 1_i4b, mat, n )
#else
                do j = 1_i4b, i
!
                    mat(:j,j) = mat(:j,j) + v(j)*w(:j) + w(j)*v(:j)
!
                end do
#endif
!
                if ( store_q ) then
                    mat(:i,k) = v(:i)
                    mat(k,k) = tau
                end if
!
                e(i) = alpha*scal
!
            else
!    
!               H IS THE IDENTITY MATRIX.
!    
                e(i) = mat(i,k)
                mat(k,k) = zero
!
            end if
!
            k = i
!
        end do
!
        d(1_i4b) = mat(1_i4b,1_i4b)
        if ( store_q ) mat(1_i4b,1_i4b) = machhuge
!
    else
!
!       REDUCE THE LOWER TRIANGLE OF mat .
!
        do i = 1_i4b, n-1_i4b
!
            d(i) = mat(i,i)
            k = i + 1_i4b
            scal = maxval( abs( mat(k+1_i4b:n,i) ) )
!
            if ( scal>zero ) then
!
!               GENERATE ELEMENTARY REFLECTOR H(i) = I + tau * v * v'
!               TO ANNIHILATE mat(i+2:n,i) STORING v IN v(i+1:) .
!
                scal = max( scal, abs(mat(k,i)) )
                v(k:n) = mat(k:n,i)/scal
                alpha = -sign( sqrt( dot_product( v(k:n), v(k:n) ) ), v(k) )
                v(k) = v(k) - alpha
                tau = one/( alpha*v(k) )
!
!               APPLY H(i) FROM BOTH SIDES TO mat(i+1:,i+1:) :
!
!               COMPUTE  x := tau * mat * v  STORING x IN w(i+1:) .
!
                w(k:n) = zero
!
                do j = n, k, -1_i4b
!
#ifdef _BLAS
                    call axpy( n-j, v(j), mat(j+1_i4b:n,j), 1_i4b, w(j+1_i4b:n), 1_i4b )
#else
                    w(j+1_i4b:n) = w(j+1_i4b:n) + v(j)*mat(j+1_i4b:n,j)
#endif
                    w(j) = w(j) + dot_product( mat(j:n,j), v(j:n) )
!
                end do
!
                w(k:n) = w(k:n)*tau
!
!               COMPUTE  w := x + 1/2 * tau * (x'*v) * v STORING w IN w(i+1:) .
!
                gam = half*tau*dot_product(w(k:n), v(k:n) )
#ifdef _BLAS
                call axpy( n-k+1_i4b, gam, v(k:n), 1_i4b, w(k:n), 1_i4b )
#else
                w(k:n) = w(k:n) + gam*v(k:n)
#endif
!
!               APPLY THE TRANSFORMATION AS A RANK-2 UPDATE :
!                     mat := mat + v * w' + w * v'
!
                do j = n, k, -1_i4b
!
#ifdef _BLAS
                    call axpy( n-j+1_i4b, v(j), w(j:n), 1_i4b, mat(j:n,j), 1_i4b )
                    call axpy( n-j+1_i4b, w(j), v(j:n), 1_i4b, mat(j:n,j), 1_i4b )
#else
                    mat(j:n,j) = mat(j:n,j) + v(j)*w(j:n) + w(j)*v(j:n)
#endif
!
                end do
!
                if ( store_q ) then
                    mat(k:n,i) = v(k:n)
                    mat(i,i) = tau
                end if
!
                e(i) = alpha*scal
!
            else
!    
!               H IS THE IDENTITY MATRIX.
!    
                e(i) = mat(k,i)
                mat(i,i) = zero
!
            end if
!
        end do
!
        d(n) = mat(n,n)
        if ( store_q ) mat(n,n) = machhuge
!
    end if
!
    e(n) = zero
!
!
! END OF SUBROUTINE symtrid_cmp_a
! _______________________________
!
    end subroutine symtrid_cmp_a
!
! =========================================================================================
!
    subroutine symtrid_cmp_b( mat, d, e, store_q )
!
! Purpose
! _______
!
!   SYMTRID_CMP reduces a real n-by-n symmetric matrix MAT to symmetric
!   tridiagonal form T by an orthogonal similarity transformation:
!
!                        Q' * MAT * Q = T
!
!
! Arguments
! _________
!
!   MAT      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry:
!
!            - The leading n-by-n upper triangular part of MAT contains the
!              upper triangular part of the symmetric matrix MAT, and the 
!              strictly lower triangular part of MAT is not referenced.
!
!            On exit:
!
!            - If STORE_Q = true :
!              The leading n-by-n upper triangular part of MAT is overwritten
!              by the matrix Q as a product of elementary reflectors.
!            - If STORE_Q = false :
!              The leading n-by-n upper triangular part of MAT is destroyed.
!
!   D        (OUTPUT) real(stnd), dimension(:)
!            The diagonal elements of the tridiagonal matrix T:
!
!            - D(i) = T(i,i).
!
!            The size of D must verify:  size( D ) = size( MAT, 1 ) = size( MAT, 2 ) = n.
!
!   E        (OUTPUT) real(stnd), dimension(:)
!            The off-diagonal elements of the tridiagonal matrix T:
!
!            -  E(i) = T(i,i+1) = T(i+1,i)
!            -  E(n) is arbitrary.
!
!            The size of E must verify:  size( E ) = size( MAT, 1 ) = size( MAT, 2 ) = n.
!
!   STORE_Q  (INPUT) logical(lgl)
!            On exit:
!
!            - If STORE_Q = true :
!              The leading n-by-n upper triangular part of MAT is overwritten
!              by the matrix Q as a product of elementary reflectors and
!              the lower triangular part of MAT is destroyed. See Further Details.
!            - If STORE_Q = false :
!              The symmetric matrix MAT is destroyed.
!
!
! Further Details
! _______________
!
!   If STORE_Q = true, the matrix Q is represented as a product of
!   elementary reflectors
!
!        Q = H(n-1) * . . . * H(2) * H(1).
!
!   Each H(i) has the form
!
!        H(i) = I + tau * v * v'
!
!   where tau is a real scalar, and v is a real vector with
!   v(i+1:n) = 0 ; v(1:i) is stored on exit in MAT(1:i,i+1),
!   and tau in MAT(i+1,i+1).
!
!   The contents of MAT on exit are illustrated by the following example
!   with n = 5:
!
!    (  xx  v1  v2  v3  v4  )
!
!    (  yy  t1  v2  v3  v4  )
!
!    (  yy  yy  t2  v3  v4  )
!
!    (  yy  yy  yy  t3  v4  )
!
!    (  yy  yy  yy  yy  t4  )
!
!
!   where vi and ti denote an element of the vector v and the scalar tau
!   defining H(i), respectively. xx = machhuge and is used by the subroutine
!   ORTHO_GEN_SYMTRID in order to verify that SYMTRID_CMP has been called
!   before ORTHO_GEN_SYMTRID. Elements yy are not modified by the subroutine.
!
!   This subroutine is adapted from the routine  DSYTD2 in LAPACK. An efficient
!   blocked algorithm is used to reduced the n-by-n symmetric matrix MAT to
!   tridiagonal form T. Furthermore, the computations are parallelized if OPENMP
!   is used.
!
!   In other words, SYMTRID_CMP is much more efficient then SYMTRID_CMP with argument
!   UPPER, which is not blocked and not parallelized.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989:
!           Block reduction of matrices to condensed form for eigenvalue computations. 
!           J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : machhuge
    use Utilities,       only : assert_eq
#ifdef _BLAS
    use BLAS_interfaces, only : axpy, syr2k
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:,:) :: mat
    real(stnd), intent(out),   dimension(:)   :: d, e
!
    logical(lgl), intent(in) :: store_q
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                           :: alpha, tau, scal
    real(stnd), dimension(size(d))       :: v, w, tmp, tmp2
    real(stnd), dimension(size(d),blksz) :: vv, ww
!
    integer(i4b) :: n, k, i, im1, j, p, l
!
#ifdef _OPENMP
!
    integer(i4b) :: num_threads, thread_num, k2
!
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='symtrid_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(d),i4b)     ,   &
                    int(size(e),i4b)     ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!
!   REDUCE THE UPPER TRIANGLE OF mat .
!
    k = n
!
#ifdef _OPENMP
    i = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*n>=omp_limit                      .and.      &
                  n/i>=omp_limit2                     .and.      &
                  i>1_i4b
!
    if ( test_par ) then
!
!$OMP PARALLEL PRIVATE(i,im1,j,p,num_threads,thread_num,tmp,vv,ww), FIRSTPRIVATE(k)
!
        num_threads = omp_get_num_threads()
        thread_num = omp_get_thread_num()
!
        do
!
!            if ( k<2_i4b ) exit
!
            if ( k*k<omp_limit ) exit
!
            i = k - 1_i4b
!
!$OMP SINGLE
!
            d(k) = mat(k,k)
            scal = maxval( abs( mat(1_i4b:i-1_i4b,k) ) )
!
            if ( scal>zero ) then
!
!               GENERATE ELEMENTARY REFLECTOR H(i) = I + tau * v * v'
!               TO ANNIHILATE mat(1:i-1,i+1) STORING v IN v(1:i) .
!
                scal = max( scal, abs(mat(i,k)) )
                v(:i) = mat(:i,k)/scal
                alpha = -sign( sqrt( dot_product( v(:i), v(:i) ) ), v(i) )
                v(i) = v(i) - alpha
                tau = one/( alpha*v(i) )
!
                if ( store_q ) then
                    mat(:i,k) = v(:i)
                    mat(k,k) = tau
                end if
!
                e(i) = alpha*scal
!
                w(:i) = zero
                l = 1_i4b
!
            else
!    
!               H(i) IS THE IDENTITY MATRIX.
!    
                e(i) = mat(i,k)
                mat(k,k) = zero
                l = 0_i4b
!    
            end if
!
!$OMP END SINGLE
!
            if ( l>0_i4b ) then
!
                do p = 1_i4b, blksz
!
                    if ( scal>zero ) then
!
                        vv(:i,l) = v(:i)
!
!                       APPLY H(i) FROM BOTH SIDES TO mat(1:i,1:i) :
!
!                       COMPUTE  x := tau * mat * v  STORING x IN w(1:i) .
!
                        tmp(:i) = zero
!
                        do j = 1_i4b+thread_num, i, num_threads
!
#ifdef _BLAS
                            call axpy( j-1_i4b, v(j), mat(:j-1_i4b,j), 1_i4b, tmp(:j-1_i4b), 1_i4b )
#else
                            tmp(:j-1_i4b) = tmp(:j-1_i4b) + v(j)*mat(:j-1_i4b,j)
#endif
                            tmp(j) = tmp(j) + dot_product( mat(:j,j), v(:j) )
!
                        end do
!
                        do j = 1_i4b+thread_num, l-1_i4b, num_threads
!
                            tmp(:i) = tmp(:i) + dot_product( v(:i), vv(:i,j) )*ww(:i,j) &
                                              + dot_product( v(:i), ww(:i,j) )*vv(:i,j)
!
                        end do
!
!$OMP CRITICAL (updatewb)
!
                        w(:i) = w(:i) + tmp(:i) 
!
!$OMP END CRITICAL (updatewb)
!
!$OMP BARRIER
!
!$OMP SINGLE
!
                        w(:i) = w(:i)*tau
!
!                       COMPUTE  w := x + 1/2 * tau * (x'*v) * v STORING w IN w(1:i) .
!
                        alpha = half*tau*dot_product(w(:i), v(:i) )
#ifdef _BLAS
                        call axpy( i, alpha, v(:i), 1_i4b, w(:i), 1_i4b )
#else
                        w(:i) = w(:i) + alpha*v(:i)
#endif
!
!$OMP END SINGLE
!
                        ww(:i,l) = w(:i)
!
                    end if
!
                    if ( p==blksz .or. i<2_i4b ) exit
!
!                   APPLY THE PREVIOUS TRANSFORMATIONS AS A RANK-2l UPDATE TO mat(1:i,i) :
!                         mat(1:i,i) := mat(1:i,i) + vv * ww' + ww * vv'
!
                    tmp(:i) = zero
!
                    do j = 1_i4b+thread_num, l, num_threads
!
#ifdef _BLAS
                        call axpy( i, vv(i,j), ww(:i,j), 1_i4b, tmp(:i), 1_i4b )
                        call axpy( i, ww(i,j), vv(:i,j), 1_i4b, tmp(:i), 1_i4b )
#else
                        tmp(:i) = tmp(:i) + vv(i,j)*ww(:i,j) + ww(i,j)*vv(:i,j)
#endif
!
                    end do
!
!$OMP CRITICAL (updatewb)
!
                    mat(:i,i) = mat(:i,i) + tmp(:i) 
!
!$OMP END CRITICAL (updatewb)
!
!$OMP BARRIER
!
!$OMP SINGLE
!
                    d(i) = mat(i,i)
                    scal = maxval( abs( mat(1_i4b:i-2_i4b,i) ) )
!
                    im1 = i - 1_i4b
!
                    if ( scal>zero ) then
!
!                       GENERATE ELEMENTARY REFLECTOR H(i-1) = I + tau * v * v'
!                       TO ANNIHILATE mat(1:i-2,i) STORING v IN v(1:i-1) .
!
                        scal = max( scal, abs(mat(im1,i)) )
                        v(:im1) = mat(:im1,i)/scal
                        alpha = -sign( sqrt( dot_product( v(:im1), v(:im1) ) ), v(im1) )
                        v(im1) = v(im1) - alpha
                        tau = one/( alpha*v(im1) )
!
                        if ( store_q ) then
                            mat(:im1,i) = v(:im1)
                            mat(i,i) = tau
                        end if
!
                        e(im1) = alpha*scal
!
                        l = l + 1_i4b
                        w(:im1) = zero
!
                    else
!    
!                       H(i-1) IS THE IDENTITY MATRIX.
!    
                        e(im1) = mat(im1,i)
                        mat(i,i) = zero
!    
                    end if
!
!$OMP END SINGLE
!
                    i = i - 1_i4b
!
!$OMP BARRIER
!
                end do
!
!               APPLY THE PREVIOUS TRANSFORMATIONS AS A RANK-2l UPDATE TO mat(1:i,1:i) :
!                     mat(1:i,1:i) := mat(1:i,1:i) + vv * ww' + ww * vv'
!
                do j = 1_i4b+thread_num, i, num_threads
!
                    do p = 1_i4b, l
!
#ifdef _BLAS
                        call axpy( j, vv(j,p), ww(:j,p), 1_i4b, mat(:j,j), 1_i4b )
                        call axpy( j, ww(j,p), vv(:j,p), 1_i4b, mat(:j,j), 1_i4b )
#else
                        mat(:j,j) = mat(:j,j) + vv(j,p)*ww(:j,p) + ww(j,p)*vv(:j,p)
#endif
!
                    end do
!
                end do
!
            end if
!
            k = i
!
!$OMP BARRIER
!    
        end do
!
!$OMP MASTER
!
        k2 = k
!
!$OMP END MASTER
!
!$OMP END PARALLEL
!
        k = k2
!
    end if
#endif
!
    do
!
        if ( k<2_i4b ) exit
!
        i = k - 1_i4b
!
        d(k) = mat(k,k)
        scal = maxval( abs( mat(1_i4b:i-1_i4b,k) ) )
!
        if ( scal>zero ) then
!
!           GENERATE ELEMENTARY REFLECTOR H(i) = I + tau * v * v'
!           TO ANNIHILATE mat(1:i-1,i+1) STORING v IN v(1:i) .
!
            scal = max( scal, abs(mat(i,k)) )
            v(:i) = mat(:i,k)/scal
            alpha = -sign( sqrt( dot_product( v(:i), v(:i) ) ), v(i) )
            v(i) = v(i) - alpha
            tau = one/( alpha*v(i) )
!
            if ( store_q ) then
                mat(:i,k) = v(:i)
                mat(k,k) = tau
            end if
!
            e(i) = alpha*scal
!
            l = 1_i4b
!
            do p = 1_i4b, blksz
!
                if ( scal>zero ) then
!
!                   APPLY H(i) FROM BOTH SIDES TO mat(1:i,1:i) :
!
!                   COMPUTE  x := tau * mat * v  STORING x IN w(1:i) .
!
                    w(:i) = zero
!
                    do j = 1_i4b, i
!
#ifdef _BLAS
                        call axpy( j-1_i4b, v(j), mat(:j-1_i4b,j), 1_i4b, w(:j-1_i4b), 1_i4b )
#else
                        w(:j-1_i4b) = w(:j-1_i4b) + v(j)*mat(:j-1_i4b,j)
#endif
                        w(j) = w(j) + dot_product( mat(:j,j), v(:j) )
!
                    end do
!
                    tmp(1_i4b:l-1_i4b)  = matmul( v(:i), vv(:i,1_i4b:l-1_i4b) )
                    tmp2(1_i4b:l-1_i4b) = matmul( v(:i), ww(:i,1_i4b:l-1_i4b) )
!
                    do j = 1_i4b, l-1_i4b
!
#ifdef _BLAS
                        call axpy( i, tmp(j), ww(:i,j), 1_i4b, w(:i), 1_i4b )
                        call axpy( i, tmp2(j), vv(:i,j), 1_i4b, w(:i), 1_i4b )
#else
                        w(:i) = w(:i) + tmp(j)*ww(:i,j) +  tmp2(j)*vv(:i,j)
#endif
!   
                    end do
!
                    w(:i) = w(:i)*tau
!
!                   COMPUTE  w := x + 1/2 * tau * (x'*v) * v STORING w IN w(1:i) .
!
                    alpha = half*tau*dot_product(w(:i), v(:i) )
                    ww(:i,l) = w(:i) + alpha*v(:i)
!
                    vv(:i,l) = v(:i)
!
                end if
!
                if ( p==blksz .or. i<2_i4b ) exit
!
!               APPLY THE PREVIOUS TRANSFORMATIONS AS A RANK-2l UPDATE TO mat(1:i,i) :
!                    mat(1:i,i) := mat(1:i,i) + vv * ww' + ww * vv'
!
                do j = 1_i4b, l
!
#ifdef _BLAS
                    call axpy( i, vv(i,j), ww(:i,j), 1_i4b, mat(:i,i), 1_i4b )
                    call axpy( i, ww(i,j), vv(:i,j), 1_i4b, mat(:i,i), 1_i4b )
#else
                    mat(:i,i) = mat(:i,i) + vv(i,j)*ww(:i,j) + ww(i,j)*vv(:i,j)
#endif
!
                end do
!
                d(i) = mat(i,i)
                scal = maxval( abs( mat(1_i4b:i-2_i4b,i) ) )
!
                im1 = i - 1_i4b
!
                if ( scal>zero ) then
!
!                   GENERATE ELEMENTARY REFLECTOR H(i-1) = I + tau * v * v'
!                   TO ANNIHILATE mat(1:i-2,i) STORING v IN v(1:i-1) .
!
                    scal = max( scal, abs(mat(im1,i)) )
                    v(:im1) = mat(:im1,i)/scal
                    alpha = -sign( sqrt( dot_product( v(:im1), v(:im1) ) ), v(im1) )
                    v(im1) = v(im1) - alpha
                    tau = one/( alpha*v(im1) )
!
                    if ( store_q ) then
                        mat(:im1,i) = v(:im1)
                        mat(i,i) = tau
                    end if
!
                    e(im1) = alpha*scal
!
                    l = l + 1_i4b
                    w(:im1) = zero
!
                else
!    
!                   H(i-1) IS THE IDENTITY MATRIX.
!    
                    e(im1) = mat(im1,i)
                    mat(i,i) = zero
!    
                end if
!
                i = i - 1_i4b
!
            end do
!
!
!           APPLY THE PREVIOUS TRANSFORMATIONS AS A RANK-2l UPDATE TO mat(1:i,1:i) :
!                 mat(1:i,1:i) := mat(1:i,1:i) + vv * ww' + ww * vv'
!
#ifdef _BLAS
            call syr2k( 'U', 'N', i, l, one, ww, n, vv, n, one, mat, n )
#else
            do j = 1_i4b, i
!
                do p = 1_i4b, l
!
                    mat(:j,j) = mat(:j,j) + vv(j,p)*ww(:j,p) + ww(j,p)*vv(:j,p)
!
                end do
!
            end do
#endif
!
        else
!    
!           H(i) IS THE IDENTITY MATRIX.
!    
            e(i) = mat(i,k)
            mat(k,k) = zero
!    
        end if
!
        k = i
!    
    end do
!
    d(1_i4b) = mat(1_i4b,1_i4b)
    if ( store_q ) mat(1_i4b,1_i4b) = machhuge
    e(n) = zero
!
!
! END OF SUBROUTINE symtrid_cmp_b
! _______________________________
!
    end subroutine symtrid_cmp_b
!
! =========================================================================================
!
    subroutine symtrid_cmp_c( matp, d, e, store_q, upper )
!
! Purpose
! _______
!
!   SYMTRID_CMP reduces a real n-by-n symmetric matrix MAT stored in packed form
!   to symmetric tridiagonal form T by an orthogonal similarity transformation:
!
!                        Q' * MAT * Q = T
!
!
! Arguments
! _________
!
!   MATP     (INPUT/OUTPUT) real(stnd), dimension(:)
!            On entry, the upper or lower triangle of the symmetric matrix
!            MAT, packed column-wise in a linear array.  The j-th column of MAT
!            is stored in the array MATP as follows:
!
!            - if UPPER = true,  MATP(i + (j-1) * j/2) = MAT(i,j)       for 1<=i<=j;
!            - if UPPER = false, MATP(i + (j-1) * (2*n-j)/2) = MAT(i,j) for j<=i<=n.
!
!            On exit:
!
!            - If UPPER = true and STORE_Q = true :
!              The leading n-by-n upper triangular part of MAT is overwritten
!              by the matrix Q as a product of elementary reflectors.
!            - If UPPER = false and STORE_Q = true :
!              The leading n-by-n lower triangular part of MAT is overwritten
!              by the matrix Q as a product of elementary reflectors.
!            - If UPPER = true and STORE_Q = false :
!              The leading n-by-n upper triangular part of MAT is destroyed.
!            - If UPPER = false and STORE_Q = false :
!              The leading n-by-n lower triangular part of MAT is destroyed.
!
!            The size of MATP must verify: size( MATP ) = (n * (n+1)/2)
!
!   D        (OUTPUT) real(stnd), dimension(:)
!            The diagonal elements of the tridiagonal matrix T:
!
!            - D(i) = T(i,i).
!
!            The size of D must verify:  size( D ) = n .
!
!   E        (OUTPUT) real(stnd), dimension(:)
!            The off-diagonal elements of the tridiagonal matrix T:
!
!            -  E(i) = T(i,i+1) = T(i+1,i)
!            -  E(n) is arbitrary.
!
!            The size of E must verify:  size( E ) = n .
!
!   STORE_Q  (INPUT) logical(lgl)
!            On exit:
!
!            - If UPPER = true and STORE_Q = true :
!              The leading n-by-n upper triangular part of MAT is overwritten
!              by the matrix Q as a product of elementary reflectors.
!            - If UPPER = false and STORE_Q = true :
!              The leading n-by-n lower triangular part of MAT is overwritten
!              by the matrix Q as a product of elementary reflectors.
!            - If UPPER = true and STORE_Q = false :
!              The leading n-by-n upper triangular part of MAT is destroyed.
!            - If UPPER = false and STORE_Q = false :
!              The leading n-by-n lower triangular part of MAT is destroyed.
!
!   UPPER    (INPUT) logical(lgl)
!            Specifies whether the upper or lower triangular part of the
!            symmetric matrix MAT is stored. If:
!
!            - UPPER = true :  Upper triangle of MAT is stored;
!            - UPPER = false:  Lower triangle of MAT is stored.
!
!
! Further Details
! _______________
!
!   If UPPER = true and STORE_Q = true, the matrix Q is represented
!   as a product of elementary reflectors
!
!        Q = H(n-1) * . . . * H(2) * H(1).
!
!   Each H(i) has the form
!
!        H(i) = I + tau * v * v'
!
!   where tau is a real scalar, and v is a real vector with
!   v(i+1:n) = 0 ; v(1:i) and tau are stored on exit in MATP,
!   overwriting MAT(1:i,i+1) and MAT(i+1,i+1), respectively.
!
!   If UPPER = false and STORE_Q = true, the matrix Q is represented
!   as a product of elementary reflectors
!
!        Q = H(1) * H(2) * . . . * H(n-1).
!
!   Each H(i) has the form
!
!        H(i) = I + tau * v * v'
!
!   where tau is a real scalar, and v is a real vector with
!   v(1:i) = 0 ; v(i+1:n) and tau are stored on exit in MATP,
!   overwriting MAT(i+1:n,i) and  MAT(i,i), respectively.
!
!   The contents of MATP on exit are illustrated by the following examples
!   with n = 5:
!
!   if UPPER = true and STORE_Q = true, MAT is equal to :
!
!    (  xx  v1  v2  v3  v4  )
!
!    (  yy  t1  v2  v3  v4  )
!
!    (  yy  yy  t2  v3  v4  )
!
!    (  yy  yy  yy  t3  v4  )
!
!    (  yy  yy  yy  yy  t4  )
!
!
!   if UPPER = false and STORE_Q = true, MAT is equal to :
!
!    (  t1  yy  yy  yy  yy  )
!
!    (  v1  t2  yy  yy  yy  )
!
!    (  v1  v2  t3  yy  yy  )
!
!    (  v1  v2  v3  t4  yy  )
!
!    (  v1  v2  v3  v4  xx  )
!
!
!   where vi and ti denote an element of the vector v and the scalar tau
!   defining H(i), respectively. Elements yy are not used and not stored in MATP.
!   xx = machhuge and is used by other subroutines in order to verify that SYMTRID_CMP
!   has been called.
!
!   This subroutine is adapted from the routine  DSPTRD in LAPACK. Note that
!   this subroutine is not blocked and not parallelized.
!
!   For more details on the reduction algorithm used in SYMTRID_CMP, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : machhuge
    use Utilities,       only : assert_eq
#ifdef _BLAS
    use BLAS_interfaces, only : axpy, spr2
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:) :: matp
    real(stnd), intent(out), dimension(:)   :: d, e
!
    logical(lgl), intent(in) :: store_q, upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: alpha, gam, tau, scal
    real(stnd), dimension(size(d)) :: v, w
!
    integer(i4b) :: n, k, i, j, l, ii, ni, nmp
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='symtrid_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(d),i4b)     ,   &
                    int(size(e),i4b)     ,   &
                    name_proc )
!
    nmp = assert_eq( int(size(matp),i4b)   ,   &
                     n*(n+1_i4b)/2_i4b     ,   &
                     name_proc )
!
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!
    if ( upper ) then
!
!       REDUCE THE UPPER TRIANGLE OF mat .
!
        k = nmp - n
!
        do i = n-1_i4b, 1_i4b, -1_i4b
!
            d(i+1_i4b) = matp(k+i+1_i4b)
            scal = maxval( abs( matp(k+1_i4b:k+i-1_i4b) ) )
!
            if ( scal>zero ) then
!
!               GENERATE ELEMENTARY REFLECTOR H(i) = I + tau * v * v'
!               TO ANNIHILATE mat(1:i-1,i+1) STORING v IN v(1:i) .
!
                scal = max( scal, abs(matp(k+i)) )
                v(:i) = matp(k+1_i4b:k+i)/scal
                alpha = -sign( sqrt( dot_product( v(:i), v(:i) ) ), v(i) )
                v(i) = v(i) - alpha
                tau = one/( alpha*v(i) )
!
!               APPLY H(i) FROM BOTH SIDES TO mat(1:i,1:i) :
!
!               COMPUTE  x := tau * mat * v  STORING x IN w(1:i) .
!
                w(:i) = zero
                l = 1_i4b
!
                do j = 1_i4b, i
!
#ifdef _BLAS
                    call axpy( j-1_i4b, v(j), matp(l:l+j-2_i4b), 1_i4b, w(:j-1_i4b), 1_i4b )
#else
                    w(:j-1_i4b) = w(:j-1_i4b) + v(j)*matp(l:l+j-2_i4b)
#endif
                    w(j) = w(j) + dot_product( matp(l:l+j-1_i4b), v(:j) )
!
                    l = l + j
!
                end do
!
                w(:i) = w(:i)*tau
!
!               COMPUTE  w := x + 1/2 * tau * (x'*v) * v STORING w IN w(1:i) .
!
                gam = half*tau*dot_product(w(:i), v(:i) )
#ifdef _BLAS
                call axpy( i, gam, v(:i), 1_i4b, w(:i), 1_i4b )
#else
                w(:i) = w(:i) + gam*v(:i)
#endif
!
!               APPLY THE TRANSFORMATION AS A RANK-2 UPDATE :
!                     mat := mat + v * w' + w * v'
!
#ifdef _BLAS
                call spr2( 'U', i, one, w, 1_i4b, v, 1_i4b, matp  )
#else
                l = 1_i4b
!
                do j = 1_i4b, i
!
                    matp(l:l+j-1_i4b) = matp(l:l+j-1_i4b) + v(j)*w(:j) + w(j)*v(:j)
!
                    l = l + j
!
                end do
#endif
!
!                l = 1_i4b
!
!                do j = 1_i4b, i
!
!#ifdef _BLAS
!                    call axpy( j, v(j), w(:j), 1_i4b, matp(l:l+j-1_i4b), 1_i4b )
!                    call axpy( j, w(j), v(:j), 1_i4b, matp(l:l+j-1_i4b), 1_i4b )
!#else
!                    matp(l:l+j-1_i4b) = matp(l:l+j-1_i4b) + v(j)*w(:j) + w(j)*v(:j)
!#endif
!
!                    l = l + j
!
!                end do
!
                if ( store_q ) then
                    matp(k+1_i4b:k+i) = v(:i)
                    matp(k+i+1_i4b) = tau
                end if
!
                e(i) = alpha*scal
!
            else
!    
!               H IS THE IDENTITY MATRIX.
!    
                e(i) = matp(k+i)
                matp(k+i+1_i4b) = zero
!
            end if
!
            k = k - i
!
        end do
!
        d(1_i4b) = matp(1_i4b)
        if ( store_q ) matp(1_i4b) = machhuge
!
    else
!
!       REDUCE THE LOWER TRIANGLE OF mat .
!
!       ii IS THE INDEX IN matp OF mat(i,i).
!
        ii = 1_i4b
!
        do i = 1_i4b, n-1_i4b
!
!           ni IS THE INDEX IN matp OF mat(n,i).
!
            ni = ii + n - i
!
            k = i + 1_i4b
!
            d(i) = matp(ii)
            scal = maxval( abs( matp(ii+2_i4b:ni) ) )
!
            if ( scal>zero ) then
!
!               GENERATE ELEMENTARY REFLECTOR H(i) = I + tau * v * v'
!               TO ANNIHILATE mat(i+2:n,i) STORING v IN v(i+1:) .
!
                scal = max( scal, abs(matp(ii+1_i4b)) )
                v(k:n) = matp(ii+1_i4b:ni)/scal
                alpha = -sign( sqrt( dot_product( v(k:n), v(k:n) ) ), v(k) )
                v(k) = v(k) - alpha
                tau = one/( alpha*v(k) )
!
!               APPLY H(i) FROM BOTH SIDES TO mat(i+1:,i+1:) :
!
!               COMPUTE  x := tau * mat * v  STORING x IN w(i+1:) .
!
                w(k:n) = zero
                l = ni + 1_i4b
!
                do j = k, n
!
#ifdef _BLAS
                    call axpy( n-j, v(j), matp(l+1_i4b:l+n-j), 1_i4b, w(j+1_i4b:n), 1_i4b )
#else
                    w(j+1_i4b:n) = w(j+1_i4b:n) + v(j)*matp(l+1_i4b:l+n-j)
#endif
                    w(j) = w(j) + dot_product( matp(l:l+n-j), v(j:n) )
!
                    l = l + n - j + 1_i4b
!
                end do
!
                w(k:n) = w(k:n)*tau
!
!               COMPUTE  w := x + 1/2 * tau * (x'*v) * v STORING w IN w(i+1:) .
!
                gam = half*tau*dot_product(w(k:n), v(k:n) )
#ifdef _BLAS
                call axpy( n-i, gam, v(k:n), 1_i4b, w(k:n), 1_i4b )
#else
                w(k:n) = w(k:n) + gam*v(k:n)
#endif
!
!               APPLY THE TRANSFORMATION AS A RANK-2 UPDATE :
!                     mat := mat + v * w' + w * v'
!
                l = ni + 1_i4b
!
                do j = k, n
!
#ifdef _BLAS
                    call axpy( n-j+1_i4b, v(j), w(j:n), 1_i4b, matp(l:l+n-j), 1_i4b )
                    call axpy( n-j+1_i4b, w(j), v(j:n), 1_i4b, matp(l:l+n-j), 1_i4b )
#else
                    matp(l:l+n-j) = matp(l:l+n-j) + v(j)*w(j:n) + w(j)*v(j:n)
#endif
!
                    l = l + n - j + 1_i4b
!
                end do
!
                if ( store_q ) then
                    matp(ii+1_i4b:ni) = v(k:n)
                    matp(ii) = tau
                end if
!
                e(i) = alpha*scal
!
            else
!    
!               H IS THE IDENTITY MATRIX.
!    
                e(i) = matp(ii+1_i4b)
                matp(ii) = zero
!
            end if
!
            ii = ni + 1_i4b
!
        end do
!
        d(n) = matp(ii)
        if ( store_q ) matp(ii) = machhuge
!
    end if
!
    e(n) = zero
!
!
! END OF SUBROUTINE symtrid_cmp_c
! _______________________________
!
    end subroutine symtrid_cmp_c
!
! =========================================================================================
!
    subroutine symtrid_cmp_d( matp, d, e, store_q )
!
! Purpose
! _______
!
!   SYMTRID_CMP reduces a real n-by-n symmetric matrix MAT stored in packed form
!   to symmetric tridiagonal form T by an orthogonal similarity transformation:
!
!                        Q' * MAT * Q = T
!
!
! Arguments
! _________
!
!   MATP     (INPUT/OUTPUT) real(stnd), dimension(:)
!            On entry, the upper triangle of the symmetric matrix MAT, 
!            packed column-wise in a linear array.  The j-th column of MAT
!            is stored in the array MATP as follows:
!
!                  MATP(i + (j-1) * j/2) = MAT(i,j) for 1<=i<=j;
!
!            On exit:
!
!            - If STORE_Q = true :
!              The leading n-by-n upper triangular part of MAT is overwritten
!              by the matrix Q as a product of elementary reflectors.
!            - If STORE_Q = false :
!              The leading n-by-n upper triangular part of MAT is destroyed.
!
!            The size of MATP must verify: size( MATP ) = (n * (n+1)/2)
!
!   D        (OUTPUT) real(stnd), dimension(:)
!            The diagonal elements of the tridiagonal matrix T:
!
!            - D(i) = T(i,i).
!
!            The size of D must verify:  size( D ) = n .
!
!   E        (OUTPUT) real(stnd), dimension(:)
!            The off-diagonal elements of the tridiagonal matrix T:
!
!            -  E(i) = T(i,i+1) = T(i+1,i)
!            -  E(n) is arbitrary.
!
!            The size of E must verify:  size( E ) = n .
!
!   STORE_Q  (INPUT) logical(lgl)
!            On exit:
!
!            - If STORE_Q = true :
!              The leading n-by-n upper triangular part of MAT is overwritten
!              by the matrix Q as a product of elementary reflectors.
!            - If STORE_Q = false :
!              The leading n-by-n upper triangular part of MAT is destroyed.
!
!
! Further Details
! _______________
!
!   The matrix Q is represented as a product of elementary reflectors
!
!        Q = H(n-1) * . . . * H(2) * H(1).
!
!   Each H(i) has the form
!
!        H(i) = I + tau * v * v'
!
!   where tau is a real scalar, and v is a real vector with
!   v(i+1:n) = 0 ; v(1:i) and tau are stored on exit in MATP,
!   overwriting MAT(1:i,i+1) and MAT(i+1,i+1), respectively,
!   if STORE_Q = true.
!
!   The contents of MATP (if STORE_Q = true) on exit are illustrated
!   by the following example with n = 5 (giving the contents of MAT):
!
!    (  xx  v1  v2  v3  v4  )
!
!    (  yy  t1  v2  v3  v4  )
!
!    (  yy  yy  t2  v3  v4  )
!
!    (  yy  yy  yy  t3  v4  )
!
!    (  yy  yy  yy  yy  t4  )
!
!
!   where vi and ti denote an element of the vector v and the scalar tau
!   defining H(i), respectively. Elements yy are not used and not stored in MATP.
!   xx = machhuge and is used by other subroutines in order to verify that SYMTRID_CMP
!   has been called.
!
!   This subroutine is adapted from the routine  DSPTRD in LAPACK. An efficient
!   blocked algorithm is used to reduced the n-by-n symmetric matrix MAT to
!   tridiagonal form T. Furthermore, the computations are parallelized if OPENMP
!   is used.
!
!   In other words, SYMTRID_CMP is much more efficient then SYMTRID_CMP with argument
!   UPPER which is not blocked and not parallelized.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989:
!           Block reduction of matrices to condensed form for eigenvalue computations. 
!           J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : machhuge
    use Utilities,       only : assert_eq
#ifdef _OPENMP
    use Utilities,       only : cumsum, arth
#endif
#ifdef _BLAS
    use BLAS_interfaces, only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:) :: matp
    real(stnd), intent(out),   dimension(:) :: d, e
!
    logical(lgl), intent(in) :: store_q
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                           :: alpha, tau, scal
    real(stnd), dimension(size(d))       :: v, w, tmp, tmp2
    real(stnd), dimension(size(d),blksz) :: vv, ww
!
    integer(i4b) :: n, k, km1, i, im1, j, j2, l, p, nmp
!
#ifdef _OPENMP
    real(stnd) :: beta1, beta2
!
    integer(i4b)                     :: num_threads, thread_num, k2, i2
    integer(i4b), dimension(size(d)) :: il
!
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='symtrid_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(d),i4b)     ,   &
                    int(size(e),i4b)     ,   &
                    name_proc )
!
    nmp = assert_eq( int(size(matp),i4b)   ,   &
                     n*(n+1_i4b)/2_i4b     ,   &
                     name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!
!   REDUCE THE UPPER TRIANGLE OF mat .
!
    k = nmp - n
    i = n - 1_i4b
!
#ifdef _OPENMP
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*n>=omp_limit                      .and.      &
                  n/j>=omp_limit2                     .and.      &
                  j>1_i4b
!
    if ( test_par ) then
!
        il(:n-1_i4b) = cumsum( arth(0_i4b,1_i4b,n-1_i4b), seed=1_i4b )
!
!$OMP PARALLEL PRIVATE(im1,j,j2,p,km1,num_threads,thread_num,tmp,vv,ww,beta1,beta2), FIRSTPRIVATE(i,k)
!
        num_threads = omp_get_num_threads()
        thread_num = omp_get_thread_num()
!
        do
!
            if ( 2_i4b*k<omp_limit ) exit
!
            im1 = i - 1_i4b
            km1 = k - i
!
!$OMP SINGLE
!
            d(i+1_i4b) = matp(k+i+1_i4b)
            scal = maxval( abs( matp(k+1_i4b:k+im1) ) )
!
            if ( scal>zero ) then
!
!               GENERATE ELEMENTARY REFLECTOR H(i) = I + tau * v * v'
!               TO ANNIHILATE mat(1:i-1,i+1) STORING v IN v(1:i) .
!
                scal = max( scal, abs(matp(k+i)) )
                v(:i) = matp(k+1_i4b:k+i)/scal
                alpha = -sign( sqrt( dot_product( v(:i), v(:i) ) ), v(i) )
                v(i) = v(i) - alpha
                tau = one/( alpha*v(i) )
!
                if ( store_q ) then
                    matp(k+1_i4b:k+i) = v(:i)
                    matp(k+i+1_i4b) = tau
                end if
!
                e(i) = alpha*scal
!
                w(:i) = zero
                l = 1_i4b
!
            else
!    
!               H(i) IS THE IDENTITY MATRIX.
!    
                e(i) = matp(k+i)
                matp(k+i+1_i4b) = zero
                l = 0_i4b
!
            end if
!
!$OMP END SINGLE
!
            if ( l>0_i4b ) then
!
                do p = 1_i4b, blksz
!
                    if ( scal>zero ) then
!
                        vv(:i,l) = v(:i)
!
!                       APPLY H(i) FROM BOTH SIDES TO mat(1:i,1:i) :
!
!                       COMPUTE  x := tau * mat * v  STORING x IN w(1:i) .
!
                        tmp(:i) = zero
!
                        do j = 1_i4b+thread_num, i, num_threads
!
                            j2 = il(j)
!
#ifdef _BLAS
                            call axpy( j-1_i4b, v(j), matp(j2:j2+j-2_i4b), 1_i4b, tmp(:j-1_i4b), 1_i4b )
#else
                            tmp(:j-1_i4b) = tmp(:j-1_i4b) + v(j)*matp(j2:j2+j-2_i4b)
#endif
                            tmp(j) = tmp(j) + dot_product( matp(j2:j2+j-1_i4b), v(:j) )
!
                        end do
!
                        do j = 1_i4b+thread_num, l-1_i4b, num_threads
!
                            beta1 = dot_product( v(:i), vv(:i,j) )
                            beta2 = dot_product( v(:i), ww(:i,j) )
!
#ifdef _BLAS
                            call axpy( i, beta1, ww(:i,j), 1_i4b, tmp(:i), 1_i4b )
                            call axpy( i, beta2, vv(:i,j), 1_i4b, tmp(:i), 1_i4b )
#else
                            tmp(:i) = tmp(:i) + beta1*ww(:i,j) + beta2*vv(:i,j)
#endif
!
                        end do
!
!$OMP CRITICAL (updatewd)
!
                        w(:i) = w(:i) + tmp(:i) 
!
!$OMP END CRITICAL (updatewd)
!
!$OMP BARRIER
!
!$OMP SINGLE
!
                        w(:i) = w(:i)*tau
!
!                       COMPUTE  w := x + 1/2 * tau * (x'*v) * v STORING w IN w(1:i) .
!
                        alpha = half*tau*dot_product(w(:i), v(:i) )
#ifdef _BLAS
                        call axpy( i, alpha, v(:i), 1_i4b, w(:i), 1_i4b )
#else
                        w(:i) = w(:i) + alpha*v(:i)
#endif
!
!$OMP END SINGLE
!
                        ww(:i,l) = w(:i)
!
                    end if
!
                    if ( p==blksz .or. i<2_i4b ) exit
!
!                   APPLY THE PREVIOUS TRANSFORMATIONS AS A RANK-2l UPDATE TO mat(1:i,i) :
!                         mat(1:i,i) := mat(1:i,i) + vv * ww' + ww * vv'
!
                    tmp(:i) = zero
!
                    do j = 1_i4b+thread_num, l, num_threads
!
#ifdef _BLAS
                        call axpy( i, vv(i,j), ww(:i,j), 1_i4b, tmp(:i), 1_i4b )
                        call axpy( i, ww(i,j), vv(:i,j), 1_i4b, tmp(:i), 1_i4b )
#else
                        tmp(:i) = tmp(:i) + vv(i,j)*ww(:i,j) + ww(i,j)*vv(:i,j)
#endif
!
                    end do
!
!$OMP CRITICAL (updatewb)
!
                    matp(km1+1_i4b:km1+i) = matp(km1+1_i4b:km1+i) + tmp(:i) 
!
!$OMP END CRITICAL (updatewb)
!
!$OMP BARRIER
!
!$OMP SINGLE
!
                    d(i) = matp(km1+i)
                    scal = maxval( abs( matp(km1+1_i4b:km1+i-2_i4b) ) )
!
                    if ( scal>zero ) then
!
!                       GENERATE ELEMENTARY REFLECTOR H(i-1) = I + tau * v * v'
!                       TO ANNIHILATE mat(1:i-2,i) STORING v IN v(1:i-1) .
!
                        scal = max( scal, abs(matp(km1+im1)) )
                        v(:im1) = matp(km1+1_i4b:km1+im1)/scal
                        alpha = -sign( sqrt( dot_product( v(:im1), v(:im1) ) ), v(im1) )
                        v(im1) = v(im1) - alpha
                        tau = one/( alpha*v(im1) )
!
                        if ( store_q ) then
                            matp(km1+1_i4b:km1+im1) = v(:im1)
                            matp(km1+i) = tau
                        end if
!
                        e(im1) = alpha*scal
!
                        l = l + 1_i4b
                        w(:im1) = zero
!
                    else
!    
!                       H(i-1) IS THE IDENTITY MATRIX.
!    
                        e(im1) = matp(km1+im1)
                        matp(km1+i) = zero
!    
                    end if
!
!$OMP END SINGLE
!
                    k   = km1
                    i   = im1
                    km1 = k - i
                    im1 = i - 1_i4b
!
!$OMP BARRIER
!
                end do
!
!               APPLY THE PREVIOUS TRANSFORMATIONS AS A RANK-2l UPDATE TO mat(1:i,1:i) :
!                     mat(1:i,1:i) := mat(1:i,1:i) + vv * wv' + ww * vv'
!
                do j = 1_i4b+thread_num, i, num_threads
!
                    j2 = il(j)
!
                    do p = 1_i4b, l
!
#ifdef _BLAS
                        call axpy( j, vv(j,p), ww(:j,p), 1_i4b, matp(j2:j2+j-1_i4b), 1_i4b )
                        call axpy( j, ww(j,p), vv(:j,p), 1_i4b, matp(j2:j2+j-1_i4b), 1_i4b )
#else
                        matp(j2:j2+j-1_i4b) = matp(j2:j2+j-1_i4b) + vv(j,p)*ww(:j,p)   &
                                                                  + ww(j,p)*vv(:j,p)
#endif
!
                    end do
!
                end do
!
            end if
!
            k = km1
            i = im1
!
!$OMP BARRIER
!
        end do
!
!$OMP MASTER
!
        k2 = k
        i2 = i
!
!$OMP END MASTER
!
!$OMP END PARALLEL
!
        k = k2
        i = i2
!
    end if
#endif
!
    do
!
        if ( i<1_i4b ) exit
!
        im1 = i - 1_i4b
        km1 = k - i
!
        d(i+1_i4b) = matp(k+i+1_i4b)
        scal = maxval( abs( matp(k+1_i4b:k+im1) ) )
!
        if ( scal>zero ) then
!
!           GENERATE ELEMENTARY REFLECTOR H(i) = I + tau * v * v'
!           TO ANNIHILATE mat(1:i-1,i+1) STORING v IN v(1:i) .
!
            scal = max( scal, abs(matp(k+i)) )
            v(:i) = matp(k+1_i4b:k+i)/scal
            alpha = -sign( sqrt( dot_product( v(:i), v(:i) ) ), v(i) )
            v(i) = v(i) - alpha
            tau = one/( alpha*v(i) )
!
            if ( store_q ) then
                matp(k+1_i4b:k+i) = v(:i)
                matp(k+i+1_i4b) = tau
            end if
!
            e(i) = alpha*scal
!
            l = 1_i4b
!
            do p = 1_i4b, blksz
!
                if ( scal>zero ) then
!
!                   APPLY H(i) FROM BOTH SIDES TO mat(1:i,1:i) :
!
!                   COMPUTE  x := tau * mat * v  STORING x IN w(1:i) .
!
                    w(:i) = zero
                    j2 = 1_i4b
!
                    do j = 1_i4b, i
!
#ifdef _BLAS
                        call axpy( j-1_i4b, v(j), matp(j2:j2+j-2_i4b), 1_i4b, w(:j-1_i4b), 1_i4b )
#else
                        w(:j-1_i4b) = w(:j-1_i4b) + v(j)*matp(j2:j2+j-2_i4b)
#endif
                        w(j) = w(j) + dot_product( matp(j2:j2+j-1_i4b), v(:j) )
!
                        j2 = j2 + j
!
                    end do
!
                    tmp(1_i4b:l-1_i4b)  = matmul( v(:i), vv(:i,1_i4b:l-1_i4b) )
                    tmp2(1_i4b:l-1_i4b) = matmul( v(:i), ww(:i,1_i4b:l-1_i4b) )
!
                    do j = 1_i4b, l-1_i4b
!
#ifdef _BLAS
                        call axpy( i, tmp(j), ww(:i,j), 1_i4b, w(:i), 1_i4b )
                        call axpy( i, tmp2(j), vv(:i,j), 1_i4b, w(:i), 1_i4b )
#else
                        w(:i) = w(:i) + tmp(j)*ww(:i,j) +  tmp2(j)*vv(:i,j)
#endif
!   
                    end do
!
                    w(:i) = w(:i)*tau
!
!                   COMPUTE  w := x + 1/2 * tau * (x'*v) * v STORING w IN w(1:i) .
!
                    alpha = half*tau*dot_product(w(:i), v(:i) )
                    ww(:i,l) = w(:i) + alpha*v(:i)
!
                    vv(:i,l) = v(:i)
!
                end if
!
                if ( p==blksz .or. i<2_i4b ) exit
!
!               APPLY THE PREVIOUS TRANSFORMATIONS AS A RANK-2l UPDATE TO mat(1:i,i) :
!                      mat(1:i,i) := mat(1:i,i) + vv * ww' + ww * vv'
!
                do j = 1_i4b, l
!
#ifdef _BLAS
                    call axpy( i, vv(i,j), ww(:i,j), 1_i4b, matp(km1+1_i4b:km1+i), 1_i4b )
                    call axpy( i, ww(i,j), vv(:i,j), 1_i4b, matp(km1+1_i4b:km1+i), 1_i4b )
#else
                    matp(km1+1_i4b:km1+i) = matp(km1+1_i4b:km1+i) + vv(i,j)*ww(:i,j) + ww(i,j)*vv(:i,j)
#endif
!
                end do
!
                d(i) = matp(km1+i)
                scal = maxval( abs( matp(km1+1_i4b:km1+i-2_i4b) ) )
!
                if ( scal>zero ) then
!
!                   GENERATE ELEMENTARY REFLECTOR H(i-1) = I + tau * v * v'
!                   TO ANNIHILATE mat(1:i-2,i) STORING v IN v(1:i-1) .
!
                    scal = max( scal, abs(matp(km1+im1)) )
                    v(:im1) = matp(km1+1_i4b:km1+im1)/scal
                    alpha = -sign( sqrt( dot_product( v(:im1), v(:im1) ) ), v(im1) )
                    v(im1) = v(im1) - alpha
                    tau = one/( alpha*v(im1) )
!
                    if ( store_q ) then
                        matp(km1+1_i4b:km1+im1) = v(:im1)
                        matp(km1+i) = tau
                    end if
!
                    e(im1) = alpha*scal
!
                    l = l + 1_i4b
!
                else
!    
!                   H(i-1) IS THE IDENTITY MATRIX.
!    
                    e(im1) = matp(km1+im1)
                    matp(km1+i) = zero
!    
                end if
!
                k   = km1
                i   = im1
                km1 = k - i
                im1 = i - 1_i4b
!
            end do
!
!           APPLY THE PREVIOUS TRANSFORMATIONS AS A RANK-2l UPDATE TO mat(1:i,1:i) :
!                       mat(1:i,1:i) := mat(1:i,1:i) + vv * wv' + ww * vv'
!
            j2 = 1_i4b
!
            do j = 1_i4b, i
!
                do p = 1_i4b, l
!
#ifdef _BLAS
                    call axpy( j, vv(j,p), ww(:j,p), 1_i4b, matp(j2:j2+j-1_i4b), 1_i4b )
                    call axpy( j, ww(j,p), vv(:j,p), 1_i4b, matp(j2:j2+j-1_i4b), 1_i4b )
#else
                    matp(j2:j2+j-1_i4b) = matp(j2:j2+j-1_i4b) + vv(j,p)*ww(:j,p)  &
                                                              + ww(j,p)*vv(:j,p)
#endif
!
                end do
!
                j2 = j2 + j
!
            end do
!
        else
!    
!           H(i) IS THE IDENTITY MATRIX.
!    
            e(i) = matp(k+i)
            matp(k+i+1_i4b) = zero
!
        end if
!
        k = km1
        i = im1
!    
    end do
!
    d(1_i4b) = matp(1_i4b)
    if ( store_q ) matp(1_i4b) = machhuge
    e(n) = zero
!
!
! END OF SUBROUTINE symtrid_cmp_d
! _______________________________
!
    end subroutine symtrid_cmp_d
!
! =========================================================================================
!
    subroutine symtrid_cmp2( mat, d, e, store_q )
!
! Purpose
! _______
!
!   SYMTRID_CMP2 reduces a real n-by-n symmetric matrix product
!
!                      MAT' * MAT
!
!   , where MAT is a m-by-n matrix with m >= n, to symmetric tridiagonal
!   form T by an orthogonal similarity transformation:
!
!                        Q' * MAT' * MAT * Q = T
!
!   where Q is orthogonal.
!
!   SYMTRID_CMP2 computes T and Q, using the one-sided Ralha
!   tridiagonal reduction algorithm without explicitely forming
!   the matrix product MAT' * MAT .
!
!
! Arguments
! _________
!
!   MAT      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, the general m-by-n matrix to be reduced.
!
!            On exit:
!
!            - If STORE_Q = true :
!              The leading n-by-n lower triangular part of MAT is overwritten
!              by the matrix Q as a product of elementary reflectors. The other
!              part of MAT is also destroyed.
!            - If STORE_Q = false :
!              The m-by-n matrix MAT is destroyed.
!
!            The shape of MAT must verify: size( MAT, 1 ) >= size( MAT, 2 ) = n .
!
!   D        (OUTPUT) real(stnd), dimension(:)
!            The diagonal elements of the tridiagonal matrix T:
!
!            - D(i) = T(i,i).
!
!            The size of D must verify:  size( D ) = size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   E        (OUTPUT) real(stnd), dimension(:)
!            The off-diagonal elements of the tridiagonal matrix T:
!
!            -  E(i) = T(i,i+1) = T(i+1,i)
!            -  E(n) is arbitrary.
!
!            The size of E must verify:  size( E ) = size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   STORE_Q  (INPUT) logical(lgl)
!            On exit:
!
!            - If STORE_Q = true :
!              The leading n-by-n lower triangular part of MAT is overwritten
!              by the matrix Q as a product of elementary reflectors. See Further
!              details.
!            - If STORE_Q = false :
!              The m-by-n matrix MAT is destroyed.
!
!
! Further Details
! _______________
!
!   This subroutine is an implementation of the Ralha one-sided method
!   to reduce implicitely a matrix product MAT' * MAT to tridiagonal form T.
!   Q is computed as a product of n-1 elementary reflectors (e.g. Householder
!   transformations):
!
!       Q = H(1) * H(2) * . . . * H(n-1)
!
!   Each H(i) has the form:
!
!       H(i) = I + tau * v * v'
!
!   where tau is a real scalar, and v is a real vector. IF STORE_Q is set to true,
!   the n-1 H(i) elementary reflectors are stored in the leading lower triangle of the
!   array MAT.
!   For the H(i) reflector, tau is stored in MAT(i,i) and v is stored in MATi+1:n,i).
!   Note that if n is equal to 1, no elementary reflectors are needed.
!
!   This is the blocked version of the algorithm. See the references (1), (2) and (3)
!   for further details. Furthermore the algorithm is parallelized if OPENMP is used.
!
!   The contents of MAT on exit are illustrated by the following example
!   with n = 5 and m = 6:
!
!   if STORE_Q = true :
!
!    (  t1  yy  yy  yy  yy  )
!
!    (  v1  t2  yy  yy  yy  )
!
!    (  v1  v2  t3  yy  yy  )
!
!    (  v1  v2  v3  t4  yy  )
!
!    (  v1  v2  v3  v4  xx  )
!
!    (  yy  yy  yy  yy  yy  )
!
!
!   where vi and ti denote an element of the vector v and the scalar tau
!   defining H(i), respectively. xx = machhuge and is used by the subroutine
!   ORTHO_GEN_SYMTRID in order to verify that SYMTRID_CMP2 has been called
!   before ORTHO_GEN_SYMTRID. Elements yy are destroyed by the subroutine.
!
!   Subroutines ORTHO_GEN_SYMTRID or APPLY_Q_SYMTRID, with logical argument UPPER
!   set to false, can be used to generate the orthogonal matrix Q or to apply it
!   to another matrix. See descriptions of these subroutines for further details.
!
!   For further details, see:
!
!   (1) Hegland, M., Kahn, M., and Osborn, M., 1999:
!           A parallel algorithm for the reduction to tridiagonal form for eigendecomposition. 
!           SIAM Journal on Scientific Computing, 21:3, pp. 987-1005.
!
!   (2) Ralha, R.M.S., 2003:
!           One-sided reduction to bidiagonal form.
!           Linear Algebra Appl., No 358, pp. 219-238.
!
!   (3) Barlow, J.L., Bosner, N., and Drmac, Z., 2005:
!           A new stable bidiagonal reduction algorithm.
!           Linear Algebra Appl., No 397, pp. 35-84.
!
!   (4) Bosner, N., and Barlow, J.L., 2007:
!           Block and Parallel versions of one-sided bidiagonalization.
!           SIAM J. Matrix Anal. Appl., Volume 29, No 3, pp. 927-953.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : machhuge
    use Char_Constants,  only : svd_error5
    use Utilities,       only : assert_eq, merror
    use Hous_Procedures, only : hous1
#ifdef _BLAS
    use BLAS_interfaces, only : axpy, gemv, ger, gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:) :: mat
    real(stnd),   intent(out),   dimension(:)   :: d, e
!
    logical(lgl), intent(in)  :: store_q
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                :: tau, beta, alpha
    real(stnd), dimension(size(mat,1))        :: u, u_tmp
    real(stnd), dimension(size(mat,2))        :: v
    real(stnd), dimension(blksz2)             :: prodvec
    real(stnd), dimension(size(mat,1),blksz2) :: umat
    real(stnd), dimension(blksz2,size(mat,2)) :: vmat
!
    integer(i4b) :: m, n, i, i1, i2, i3, l, j, k, blksz
!
#ifdef _OPENMP
    real(stnd), dimension(size(mat,1)) :: tmp
!
    logical(lgl) :: apply_hous
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='symtrid_cmp2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( mat, 1 )
!
    n =  assert_eq( (/ int(size(mat,2),i4b),     &
                       int(size(d),i4b)      ,   &
                       int(size(e),i4b)   /) ,   &
                       name_proc )
!
!   CHECK DIMENSIONS OF THE INPUT MATRIX.
!
    if ( m<n ) then
!
        call merror( name_proc//svd_error5 )
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!
!   REDUCE THE MATRIX PRODUCT mat**(t).mat TO TRIDIAGONAL FORM.
!
!   COMPUTE FIRST DIAGONAL ELEMENT OF THE TRIDIAGONAL MATRIX.
!
    d(1_i4b) = dot_product( mat(1_i4b:m,1_i4b), mat(1_i4b:m,1_i4b) )
!
!   SPECIAL CASE IF mat(:,:) IS A COLUMN VECTOR.
!
    if ( n==1_i4b ) then
!
        e(1_i4b) = zero
!
        mat(1_i4b,1_i4b) = machhuge
!
        return
!
    end if
!
!   ADJUST BLOCK SIZE IF NUMBER OF COLUMNS IS TOO SMALL.
!
    blksz = min( blksz2, n )
!
    u_tmp(1_i4b:m) = zero
!
#ifdef _OPENMP
!
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  m*n>=omp_limit                      .and.      &
                  i>1_i4b
!
    if ( test_par ) then
!
!$OMP PARALLEL PRIVATE(i,i1,i2,i3,l,j,k,alpha,beta,tau,u,tmp)
!
!       PROCESS THE FIRST BLOCK OF COLUMNS OF mat .
!
!       START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE H(1)
!       HOUSEHOLDER TRANSFORMATION.
!
        u(1_i4b:m)   = mat(1_i4b:m,1_i4b)
        tmp(1_i4b:m) = zero
!
!$OMP DO SCHEDULE(STATIC)
!
        do i = 2_i4b, n
!
            v(i) = dot_product( u(1_i4b:m), mat(1_i4b:m,i) )
#ifdef _BLAS
            call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, tmp(1_i4b:m), 1_i4b )
#else
            tmp(1_i4b:m) = tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
        end do
!
!$OMP END DO NOWAIT
!
!$OMP CRITICAL (compu0)
!
        u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + tmp(1_i4b:m)  
!
!$OMP END CRITICAL (compu0)
!
!$OMP BARRIER
!
!$OMP SINGLE
!
!       COMPUTE Q IN FACTORED FORM, STORING THE H(1) HOUSEHOLDER TRANSFORMATION
!       IN mat(1,1) AND mat(2:n,1).
!
        alpha = v(2_i4b)        
!
        call hous1( v(2_i4b:n), tau, beta )         
!
        if ( store_q ) then
!
!           SAVE THE H(1) HOUSEHOLDER TRANSFORMATION IN mat(1:n,1) .
!
            mat(1_i4b,1_i4b)   = tau
            mat(2_i4b:n,1_i4b) = v(2_i4b:n)
!
        end if
!
!       INITIALIZE ARRAYS FOR LATER BLOCK UPDATE TRANSFORMATION.
!
        if ( tau/=zero ) then
            umat(1_i4b:m,1_i4b) = tau*( ( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,2_i4b) )/( alpha - beta ) )
        else
            umat(1_i4b:m,1_i4b) = zero
        end if
!
        umat(1_i4b:m,blksz) = zero
!
        vmat(1_i4b:blksz,1_i4b:n) = zero
        vmat(1_i4b,2_i4b:n)       = v(2_i4b:n)
!
!       UPDATE mat(1:m,2) WITH H(1).
!
        mat(1_i4b:m,2_i4b) = mat(1_i4b:m,2_i4b) + umat(1_i4b:m,1_i4b)
!
!       COMPUTE SECOND ROW OF THE TRIDIAGONAL MATRIX.
!
        d(2_i4b) = dot_product( mat(1_i4b:m,2_i4b), mat(1_i4b:m,2_i4b) )
        e(1_i4b) = beta
!
        u_tmp(1_i4b:m) = zero  
!
!$OMP END SINGLE
!
        do j = 2_i4b, blksz-1_i4b
!
!           START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE H(j)
!           HOUSEHOLDER TRANSFORMATION.
!
            u(1_i4b:m) = mat(1_i4b:m,j)
!
!$OMP DO SCHEDULE(STATIC)
!
            do i = 1_i4b, j-1_i4b
!
                prodvec(i) = dot_product( u(1_i4b:m), umat(1_i4b:m,i) )
!
            end do
!
!$OMP END DO
!
            tmp(1_i4b:m) = zero
!
!$OMP DO SCHEDULE(STATIC)
!
            do i = j+1_i4b, n
!
                v(i) = dot_product( u(1_i4b:m), mat(1_i4b:m,i) ) +              &
                       dot_product(prodvec(1_i4b:j-1_i4b),vmat(1_i4b:j-1_i4b,i) )
#ifdef _BLAS
                call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, tmp(1_i4b:m), 1_i4b )
#else
                tmp(1_i4b:m) = tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
            end do
!
!$OMP END DO NOWAIT
!
!$OMP CRITICAL (compu1)
!
            u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + tmp(1_i4b:m)  
!
!$OMP END CRITICAL (compu1)
!
!$OMP BARRIER
!
!$OMP SINGLE
!
!           COMPUTE Q IN FACTORED FORM, STORING THE H(j) HOUSEHOLDER TRANSFORMATION
!           IN mat(j,j) AND mat(j+1:n,j).
!
            alpha = v(j+1_i4b)        
!
            call hous1( v(j+1_i4b:n), tau, beta )
!
            if ( store_q ) then
!
!               SAVE THE H(j) HOUSEHOLDER TRANSFORMATION IN mat(j:n,j).
!
                mat(j,j)         = tau
                mat(j+1_i4b:n,j) = v(j+1_i4b:n)
!
            end if
!
!           SAVE H(j) HOUSEHOLDER TRANSFORMATION FOR LATER BLOCK UPDATE.
!
            vmat(j,j+1_i4b:n) = v(j+1_i4b:n)
!
            prodvec(1_i4b:j-1_i4b) = matmul( vmat(1_i4b:j-1_i4b,j+1_i4b:n), v(j+1_i4b:n) )
!
            if ( tau/=zero ) then
                umat(1_i4b:m,j) = ( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,j+1_i4b) )/( alpha - beta )
            else
                umat(1_i4b:m,j) = zero
            end if
!
#ifdef _BLAS
            call gemv( 'N', m, j-1_i4b, tau, umat(1_i4b:m,1_i4b:j-1_i4b), m, prodvec(1_i4b:j-1_i4b), &
                        1_i4b, tau, umat(1_i4b:m,j), 1_i4b )
#else
            umat(1_i4b:m,j) = tau*( umat(1_i4b:m,j) + matmul( umat(1_i4b:m,1_i4b:j-1_i4b), prodvec(1_i4b:j-1_i4b) ) )
#endif
!
!           UPDATE mat(1:m,j+1) WITH H(1)* ... *H(j) AND ORTHOGONALIZE mat(1:m,j+1) AGAINST THE j COLUMN OF Q.
!
#ifdef _BLAS
            call gemv( 'N', m, j, one, umat(1_i4b:m,1_i4b:j), m, vmat(1_i4b:j,j+1_i4b), 1_i4b,     &
                       one, mat(1_i4b:m,j+1_i4b), 1_i4b )
#else
            mat(1_i4b:m,j+1_i4b) = mat(1_i4b:m,j+1_i4b) + matmul( umat(1_i4b:m,1_i4b:j), vmat(1_i4b:j,j+1_i4b) )
#endif
!
!           COMPUTE THE j+1 ROW OF THE TRIDIAGONAL MATRIX.
!
            d(j+1_i4b) = dot_product( mat(1_i4b:m,j+1_i4b), mat(1_i4b:m,j+1_i4b) )
            e(j) = beta
!
            u_tmp(1_i4b:m) = zero
!
!$OMP END SINGLE
!
        end do
!
!       PROCESS THE SECOND AND FOLLOWING BLOCKS OF COLUMNS OF mat .
!
        do l = blksz, n-blksz, blksz
!
!           START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE H(l)
!           HOUSEHOLDER TRANSFORMATION.
!
            tmp(1_i4b:m) = zero
!
!$OMP DO SCHEDULE(STATIC)
!
            do i = l, n, blksz
!
!               COMPUTE mat(1:m,l+1:n)*H(l-blksz)* ... *H(l-2)*H(l-1) AS A BLOCK UPDATE.
!
                i1 = i + 1_i4b
                i2 = min( i+blksz, n)
                i3 = i2 - i
!
                if ( i3>0_i4b ) then
!
#ifdef _BLAS
                    call gemm( 'N', 'N', m, i3, blksz, one, umat(1_i4b:m,1_i4b:blksz), m,   &
                               vmat(1_i4b:blksz,i1:i2), blksz, one, mat(1_i4b:m,i1:i2), m   )
!
                    v(i1:i2) = matmul( mat(1_i4b:m,l), mat(1_i4b:m,i1:i2) )
!
                    call gemv( 'N', m, i3, one,  mat(1_i4b:m,i1:i2), m, v(i1:i2), &
                               1_i4b, one, tmp(1_i4b:m), 1_i4b )
!
#else
                    mat(1_i4b:m,i1:i2) = mat(1_i4b:m,i1:i2) +                                         &
                                         matmul( umat(1_i4b:m,1_i4b:blksz), vmat(1_i4b:blksz,i1:i2) )
!
                    v(i1:i2) = matmul( mat(1_i4b:m,l), mat(1_i4b:m,i1:i2) )
!
                    tmp(1_i4b:m) = tmp(1_i4b:m) +  matmul( mat(1_i4b:m,i1:i2), v(i1:i2) )
#endif
!
                end if
!
            end do
!
!$OMP END DO NOWAIT
!
!$OMP CRITICAL (compu2)
!
            u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + tmp(1_i4b:m)  
!
!$OMP END CRITICAL (compu2)
!
!$OMP BARRIER
!
!$OMP SINGLE
!
!           COMPUTE Q IN FACTORED FORM, STORING THE H(l) HOUSEHOLDER TRANSFORMATION
!           IN mat(l,l) AND mat(l+1:n,l).
!
            alpha = v(l+1_i4b)        
!
            call hous1( v(l+1_i4b:n), tau, beta )
!
            if ( store_q ) then
!
!               SAVE THE H(l) HOUSEHOLDER TRANSFORMATION IN mat(l:n,l) .
!
                mat(l,l) = tau
                mat(l+1_i4b:n,l) = v(l+1_i4b:n)
!
            end if
!
!           INITIALIZE ARRAYS FOR LATER BLOCK UPDATE TRANSFORMATION.
!
            if ( tau/=zero ) then
                umat(1_i4b:m,1_i4b) = tau*(( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,l+1_i4b) )/( alpha - beta ))
            else
                umat(1_i4b:m,1_i4b) = zero
            end if
!
            vmat(:blksz,:n)       = zero
            vmat(1_i4b,l+1_i4b:n) = v(l+1_i4b:n)
!
!           UPDATE mat(1:m,l+1) WITH H(l).
!
            mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b) + umat(1_i4b:m,1_i4b)
!
!           COMPUTE THE l+1 ROW OF THE TRIDIAGONAL MATRIX.
!
            d(l+1_i4b) = dot_product( mat(1_i4b:m,l+1_i4b), mat(1_i4b:m,l+1_i4b) )
            e(l)       = beta
!
            u_tmp(1_i4b:m) = zero
!
!$OMP END SINGLE
!
            do j = 2_i4b, blksz
!
!               START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE H(l+j-1)
!               HOUSEHOLDER TRANSFORMATION.
!
                u(1_i4b:m) = mat(1_i4b:m,l+j-1_i4b)
!
!$OMP DO SCHEDULE(STATIC)
!
                do i = 1_i4b, j-1_i4b
!
                    prodvec(i) = dot_product( u(1_i4b:m), umat(1_i4b:m,i) )
!
                end do
!
!$OMP END DO
!
                tmp(1_i4b:m) = zero
!
!$OMP DO SCHEDULE(STATIC)
!
                do i = l+j, n
!
                    v(i) = dot_product( u(1_i4b:m), mat(1_i4b:m,i) ) +  &
                           dot_product(prodvec(:j-1),vmat(:j-1_i4b,i) )
#ifdef _BLAS
                    call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, tmp(1_i4b:m), 1_i4b )
#else
                    tmp(1_i4b:m) = tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
                end do
!
!$OMP END DO NOWAIT
!
!$OMP CRITICAL (compu3)
!
                u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + tmp(1_i4b:m)  
!
!$OMP END CRITICAL (compu3)
!
!$OMP BARRIER
!
!$OMP SINGLE
!
!               COMPUTE Q IN FACTORED FORM, STORING THE H(l+j-1) HOUSEHOLDER TRANSFORMATION
!               IN mat(l+j-1,l+j-1) AND mat(l+j:n,l+j-1).
!
                alpha = v(l+j)        
!
                call hous1( v(l+j:n), tau, beta )
!
                if ( store_q ) then
!
!                   SAVE THE H(l+j-1) HOUSEHOLDER TRANSFORMATION IN mat(l+1:n,l+j-1) .
!
                    mat(l+j-1_i4b,l+j-1_i4b) = tau
                    mat(l+j:n,l+j-1_i4b)     = v(l+j:n)
!
                end if
!
!               SAVE H(l+j-1) HOUSEHOLDER TRANSFORMATION FOR LATER BLOCK UPDATE.
!
                vmat(j,l+j:n) = v(l+j:n)
!
                prodvec(:j-1_i4b) = matmul( vmat(:j-1_i4b,l+j:n), v(l+j:n) )
!
                if ( tau/=zero ) then
                    umat(1_i4b:m,j) = ( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,l+j) )/( alpha - beta )
                else
                    umat(1_i4b:m,j) = zero
                end if
!
#ifdef _BLAS
                call gemv( 'N', m, j-1_i4b, tau, umat(1_i4b:m,1_i4b:j-1_i4b), m, prodvec(1_i4b:j-1_i4b), &
                           1_i4b, tau, umat(1_i4b:m,j), 1_i4b )
#else
                umat(1_i4b:m,j) = tau*( umat(1_i4b:m,j) + matmul( umat(1_i4b:m,:j-1_i4b), prodvec(1_i4b:j-1_i4b) ) )
#endif
!
!               UPDATE mat(1:m,l+j) WITH H(l)* ... *H(l+j-1).
!
#ifdef _BLAS
                call gemv( 'N', m, j, one, umat(1_i4b:m,1_i4b:j), m, vmat(1_i4b:j,l+j), 1_i4b, &
                           one, mat(1_i4b:m,l+j), 1_i4b )
#else
                mat(1_i4b:m,l+j) = mat(1_i4b:m,l+j) + matmul( umat(1_i4b:m,1_i4b:j), vmat(1_i4b:j,l+j) )
#endif
!
!               COMPUTE THE l+J ROW OF THE TRIDIAGONAL MATRIX.
!
                d(l+j)       = dot_product( mat(1_i4b:m,l+j), mat(1_i4b:m,l+j) )
                e(l+j-1_i4b) = beta
!
                u_tmp(1_i4b:m) = zero
!
!$OMP END SINGLE
!
            end do
!
        end do
!
        k = blksz*(n/blksz)
!
        if ( n/=k ) then
!
!           PROCESS THE LAST BLOCK OF COLUMNS OF mat .
!
            do l = k, n-1_i4b
!
!               START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE H(l)
!               HOUSEHOLDER TRANSFORMATION.
!
                tmp(1_i4b:m) = zero
!
                if ( l==k ) then
!
!$OMP DO SCHEDULE(STATIC)
!
                    do i = l+1_i4b, n
!
!                       COMPUTE mat(1:m,l+1:n)*H(l-blksz)* ... *H(l-2)*H(l-1) WITH A LAST BLOCK UPDATE.
!
                        u(1_i4b:m) = mat(1_i4b:m,i)
!
#ifdef _BLAS
                        call gemv( 'N', m, blksz, one, umat(1_i4b:m,1_i4b:blksz), m, vmat(1_i4b:blksz,i), &
                                   1_i4b, one, u(1_i4b:m), 1_i4b )
#else
                        u(1_i4b:m) = u(1_i4b:m) + matmul( umat(1_i4b:m,1_i4b:blksz), vmat(1_i4b:blksz,i) )
#endif
!
                        v(i) = dot_product( mat(1_i4b:m,l), u(1_i4b:m) )
#ifdef _BLAS
                        call axpy( m, v(i), u(1_i4b:m), 1_i4b, tmp(1_i4b:m), 1_i4b )
#else
                        tmp(1_i4b:m) = tmp(1_i4b:m) + v(i)*u(1_i4b:m)
#endif
!
                        mat(1_i4b:m,i) = u(1_i4b:m)
!
                    end do
!
!$OMP END DO NOWAIT
!
                else
!
!$OMP DO SCHEDULE(STATIC)
!
                    do i = l+1_i4b, n
!
                        v(i) = dot_product( mat(1_i4b:m,l), mat(1_i4b:m,i) )
#ifdef _BLAS
                        call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, tmp(1_i4b:m), 1_i4b )
#else
                        tmp(1_i4b:m) = tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
                    end do
!
!$OMP END DO NOWAIT
!
                end if
!
!$OMP CRITICAL (compu4)
!
                u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + tmp(1_i4b:m)  
!
!$OMP END CRITICAL (compu4)
!
!$OMP BARRIER
!
!$OMP SINGLE
!
!               COMPUTE Q IN FACTORED FORM, STORING THE H(l) HOUSEHOLDER TRANSFORMATION
!               IN mat(l,l) AND mat(l+1:n,l).
!
                alpha = v(l+1_i4b)        
!
                call hous1( v(l+1_i4b:n), tau, beta )
!
                if ( store_q ) then
!
!                   SAVE THE H(l) HOUSEHOLDER TRANSFORMATION IN mat(l:n,l) .
!
                    mat(l,l) = tau
                    mat(l+1_i4b:n,l) = v(l+1_i4b:n)
!
                end if
!
                apply_hous = tau/=zero
!
                if ( apply_hous ) then
!
                    umat(1_i4b:m,1_i4b) = tau*(( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,l+1_i4b) )/( alpha - beta ))
!
!                   APPLY THE H(l) HOUSEHOLDER TRANSFORMATION TO mat(:m,l+1) .
!
                    mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b) + umat(1_i4b:m,1_i4b)
!
                end if
!
!               COMPUTE THE l+1 ROW OF THE TRIDIAGONAL MATRIX.
!
                e(l) = beta
!
!$OMP END SINGLE
!
                if ( apply_hous ) then
!
!                   APPLY THE H(l) HOUSEHOLDER TRANSFORMATION TO mat(:m,l+2:n) .
!
!$OMP DO SCHEDULE(STATIC)
!
                    do i = l+2_i4b, n
!
#ifdef _BLAS
                        call axpy( m, v(i), umat(1_i4b:m,1_i4b), 1_i4b, mat(1_i4b:m,i), 1_i4b )
#else
                        mat(1_i4b:m,i) = mat(1_i4b:m,i) + umat(1_i4b:m,1_i4b)*v(i)
#endif
!
                    end do
!
!$OMP END DO NOWAIT
!
                end if
!
!$OMP SINGLE
!
!               COMPUTE THE l+1 ROW OF THE TRIDIAGONAL MATRIX.
!
                d(l+1_i4b) = dot_product( mat(1_i4b:m,l+1_i4b), mat(1_i4b:m,l+1_i4b) )
!
                u_tmp(1_i4b:m) = zero
!
!$OMP END SINGLE
!
            end do
!
        end if
!
!$OMP END PARALLEL
!
    else
#endif
!
!       PROCESS THE FIRST BLOCK OF COLUMNS OF mat .
!
!       START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE H(1)
!       HOUSEHOLDER TRANSFORMATION.
!
        u(1_i4b:m) = mat(1_i4b:m,1_i4b)
!
        do i = 2_i4b, n
!
            v(i) = dot_product( u(1_i4b:m), mat(1_i4b:m,i) )
#ifdef _BLAS
            call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
            u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
        end do
!
!       COMPUTE Q IN FACTORED FORM, STORING THE H(1) HOUSEHOLDER TRANSFORMATION
!       IN mat(1,1) AND mat(2:n,1).
!
        alpha = v(2_i4b)        
!
        call hous1( v(2_i4b:n), tau, beta )         
!
        if ( store_q ) then
!
!           SAVE THE H(1) HOUSEHOLDER TRANSFORMATION IN mat(1:n,1) .
!
            mat(1_i4b,1_i4b)   = tau
            mat(2_i4b:n,1_i4b) = v(2_i4b:n)
!
        end if
!
!       INITIALIZE ARRAYS FOR LATER BLOCK UPDATE TRANSFORMATION.
!
        if ( tau/=zero ) then
            umat(1_i4b:m,1_i4b) = tau*( ( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,2_i4b) )/( alpha - beta ) )
        else
            umat(1_i4b:m,1_i4b) = zero
        end if
!
        umat(1_i4b:m,blksz) = zero
!
        vmat(1_i4b:blksz,1_i4b:n) = zero
        vmat(1_i4b,2_i4b:n)       = v(2_i4b:n)
!
!       UPDATE mat(1:m,2) WITH H(1).
!
        mat(1_i4b:m,2_i4b) = mat(1_i4b:m,2_i4b) + umat(1_i4b:m,1_i4b)
!
!       COMPUTE SECOND ROW OF THE TRIDIAGONAL MATRIX.
!
        d(2_i4b) = dot_product( mat(1_i4b:m,2_i4b), mat(1_i4b:m,2_i4b) )
        e(1_i4b) = beta
!
        do j = 2_i4b, blksz-1_i4b
!
!           START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE H(j)
!           HOUSEHOLDER TRANSFORMATION.
!
            u(1_i4b:m) = mat(1_i4b:m,j)
!
            prodvec(1_i4b:j-1_i4b) = matmul( u(1_i4b:m), umat(1_i4b:m,1_i4b:j-1_i4b) )
!
            u_tmp(1_i4b:m) = zero
!
            do i = j+1_i4b, n
!
                v(i) = dot_product( u(1_i4b:m), mat(1_i4b:m,i) ) +              &
                       dot_product(prodvec(1_i4b:j-1_i4b),vmat(1_i4b:j-1_i4b,i) )
#ifdef _BLAS
                call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
                u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
            end do
!
!           COMPUTE Q IN FACTORED FORM, STORING THE H(j) HOUSEHOLDER TRANSFORMATION
!           IN mat(j,j) AND mat(j+1:n,j).
!
            alpha = v(j+1_i4b)        
!
            call hous1( v(j+1_i4b:n), tau, beta )
!
            if ( store_q ) then
!
!               SAVE THE H(j) HOUSEHOLDER TRANSFORMATION IN mat(j:n,j).
!
                mat(j,j)         = tau
                mat(j+1_i4b:n,j) = v(j+1_i4b:n)
!
            end if
!
!           SAVE H(j) HOUSEHOLDER TRANSFORMATION FOR LATER BLOCK UPDATE.
!
            vmat(j,j+1_i4b:n) = v(j+1_i4b:n)
!
            prodvec(1_i4b:j-1_i4b) = matmul( vmat(1_i4b:j-1_i4b,j+1_i4b:n), v(j+1_i4b:n) )
!
            if ( tau/=zero ) then
                umat(1_i4b:m,j) = ( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,j+1_i4b) )/( alpha - beta )
            else
                umat(1_i4b:m,j) = zero
            end if
!
#ifdef _BLAS
            call gemv( 'N', m, j-1_i4b, tau, umat(1_i4b:m,1_i4b:j-1_i4b), m, prodvec(1_i4b:j-1_i4b), &
                        1_i4b, tau, umat(1_i4b:m,j), 1_i4b )
#else
            umat(1_i4b:m,j) = tau*( umat(1_i4b:m,j) + matmul( umat(1_i4b:m,1_i4b:j-1_i4b), prodvec(1_i4b:j-1_i4b) ) )
#endif
!
!           UPDATE mat(1:m,j+1) WITH H(1)* ... *H(j).
!
#ifdef _BLAS
            call gemv( 'N', m, j, one, umat(1_i4b:m,1_i4b:j), m, vmat(1_i4b:j,j+1_i4b), 1_i4b,     &
                       one, mat(1_i4b:m,j+1_i4b), 1_i4b )
#else
            mat(1_i4b:m,j+1_i4b) = mat(1_i4b:m,j+1_i4b) + matmul( umat(1_i4b:m,1_i4b:j), vmat(1_i4b:j,j+1_i4b) )
#endif
!
!           COMPUTE THE j+1 ROW OF THE TRIDIAGONAL MATRIX.
!
            d(j+1_i4b) = dot_product( mat(1_i4b:m,j+1_i4b), mat(1_i4b:m,j+1_i4b) )
            e(j) = beta
!
        end do
!
!       PROCESS THE SECOND AND FOLLOWING BLOCKS OF COLUMNS OF mat .
!
        do l = blksz, n-blksz, blksz
!
!           START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE H(l)
!           HOUSEHOLDER TRANSFORMATION.
!
            u_tmp(1_i4b:m) = zero
!
            do i = l, n, blksz
!
!               COMPUTE mat(1:m,l+1:n)*G(l-blksz)* ... *G(l-2)*G(l-1) AS A BLOCK UPDATE.
!
                i1 = i + 1_i4b
                i2 = min( i+blksz, n)
                i3 = i2 - i
!
                if ( i3>0_i4b ) then
!
#ifdef _BLAS
                    call gemm( 'N', 'N', m, i3, blksz, one, umat(1_i4b:m,1_i4b:blksz), m,  &
                               vmat(1_i4b:blksz,i1:i2), blksz, one, mat(1_i4b:m,i1:i2), m  )
!
                    v(i1:i2) = matmul( mat(1_i4b:m,l), mat(1_i4b:m,i1:i2) )
!
                    call gemv( 'N', m, i3, one,  mat(1_i4b:m,i1:i2), m, v(i1:i2), &
                               1_i4b, one, u_tmp(1_i4b:m), 1_i4b )
!
#else
                    mat(1_i4b:m,i1:i2) = mat(1_i4b:m,i1:i2) +                                         &
                                         matmul( umat(1_i4b:m,1_i4b:blksz), vmat(1_i4b:blksz,i1:i2) )
!
                    v(i1:i2) = matmul( mat(1_i4b:m,l), mat(1_i4b:m,i1:i2) )
!
                    u_tmp(1_i4b:m) = u_tmp(1_i4b:m) +  matmul( mat(1_i4b:m,i1:i2), v(i1:i2) )
#endif
!
                end if
!
            end do
!
!           COMPUTE Q IN FACTORED FORM, STORING THE H(l) HOUSEHOLDER TRANSFORMATION
!           IN mat(l,l) AND mat(l+1:n,l).
!
            alpha = v(l+1_i4b)        
!
            call hous1( v(l+1_i4b:n), tau, beta )
!
            if ( store_q ) then
!
!               SAVE THE H(l) HOUSEHOLDER TRANSFORMATION IN mat(l:n,l) .
!
                mat(l,l) = tau
                mat(l+1_i4b:n,l) = v(l+1_i4b:n)
!
            end if
!
!           INITIALIZE ARRAYS FOR LATER BLOCK UPDATE TRANSFORMATION.
!
            if ( tau/=zero ) then
                umat(1_i4b:m,1_i4b) = tau*(( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,l+1_i4b) )/( alpha - beta ))
            else
                umat(1_i4b:m,1_i4b) = zero
            end if
!
            vmat(:blksz,:n)       = zero
            vmat(1_i4b,l+1_i4b:n) = v(l+1_i4b:n)
!
!           UPDATE mat(1:m,l+1) WITH H(l).
!
            mat(1_i4b:m,l+1_i4b) = mat(1_i4b:m,l+1_i4b) + umat(1_i4b:m,1_i4b)
!
!           COMPUTE THE l+1 ROW OF THE TRIDIAGONAL MATRIX.
!
            d(l+1_i4b) = dot_product( mat(1_i4b:m,l+1_i4b), mat(1_i4b:m,l+1_i4b) )
            e(l)       = beta
!
            do j = 2_i4b, blksz
!
!               START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE H(l+j-1)
!               HOUSEHOLDER TRANSFORMATION.
!
                u(1_i4b:m) = mat(1_i4b:m,l+j-1_i4b)
!
                prodvec(:j-1_i4b) = matmul( u(1_i4b:m), umat(1_i4b:m,:j-1_i4b) )
!
                u_tmp(1_i4b:m) = zero
!
                do i = l+j, n
!
                    v(i) = dot_product( u(1_i4b:m), mat(1_i4b:m,i) ) +  &
                           dot_product(prodvec(:j-1),vmat(:j-1_i4b,i) )
#ifdef _BLAS
                    call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
                    u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
                end do
!
!               COMPUTE Q IN FACTORED FORM, STORING THE H(l+j-1) HOUSEHOLDER TRANSFORMATION
!               IN mat(l+j-1,l+j-1) AND mat(l+j:n,l+j-1).
!
                alpha = v(l+j)        
!
                call hous1( v(l+j:n), tau, beta )
!
                if ( store_q ) then
!
!                   SAVE THE H(l+j-1) HOUSEHOLDER TRANSFORMATION IN mat(l+1:n,l+j-1) .
!
                    mat(l+j-1_i4b,l+j-1_i4b) = tau
                    mat(l+j:n,l+j-1_i4b)     = v(l+j:n)
!
                end if
!
!               SAVE H(l+j-1) HOUSEHOLDER TRANSFORMATION FOR LATER BLOCK UPDATE.
!
                vmat(j,l+j:n) = v(l+j:n)
!
                prodvec(:j-1_i4b) = matmul( vmat(:j-1_i4b,l+j:n), v(l+j:n) )
!
                if ( tau/=zero ) then
                    umat(1_i4b:m,j) = ( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,l+j) )/( alpha - beta )
                else
                    umat(1_i4b:m,j) = zero
                end if
!
#ifdef _BLAS
                call gemv( 'N', m, j-1_i4b, tau, umat(1_i4b:m,1_i4b:j-1_i4b), m, prodvec(1_i4b:j-1_i4b), &
                           1_i4b, tau, umat(1_i4b:m,j), 1_i4b )
#else
                umat(1_i4b:m,j) = tau*( umat(1_i4b:m,j) + matmul( umat(1_i4b:m,:j-1_i4b), prodvec(1_i4b:j-1_i4b) ) )
#endif
!
!               UPDATE mat(1:m,l+j) WITH H(l)* ... *H(l+j-1).
!
#ifdef _BLAS
                call gemv( 'N', m, j, one, umat(1_i4b:m,1_i4b:j), m, vmat(1_i4b:j,l+j), 1_i4b, &
                           one, mat(1_i4b:m,l+j), 1_i4b )
#else
                mat(1_i4b:m,l+j) = mat(1_i4b:m,l+j) + matmul( umat(1_i4b:m,1_i4b:j), vmat(1_i4b:j,l+j) )
#endif
!
!               COMPUTE THE l+J ROW OF THE TRIDIAGONAL MATRIX.
!
                d(l+j)       = dot_product( mat(1_i4b:m,l+j), mat(1_i4b:m,l+j) )
                e(l+j-1_i4b) = beta
!
            end do
!
        end do
!
        k = blksz*(n/blksz)
!
        if ( n/=k ) then
!
!           PROCESS THE LAST BLOCK OF COLUMNS OF mat .
!
            do l = k, n-1_i4b
!
!               START COMPUTATIONS FOR GENERATION AND APPLICATION OF THE H(l)
!               HOUSEHOLDER TRANSFORMATION.
!
                u_tmp(1_i4b:m) = zero
!
                if ( l==k ) then
!
                    do i = l+1_i4b, n
!
!                       COMPUTE mat(1:m,l+1:n)*H(l-blksz)* ... *H(l-2)*H(l-1) WITH A LAST BLOCK UPDATE.
!
                        u(1_i4b:m) = mat(1_i4b:m,i)
!
#ifdef _BLAS
                        call gemv( 'N', m, blksz, one, umat(1_i4b:m,1_i4b:blksz), m, vmat(1_i4b:blksz,i), &
                                   1_i4b, one, u(1_i4b:m), 1_i4b )
#else
                        u(1_i4b:m) = u(1_i4b:m) + matmul( umat(1_i4b:m,1_i4b:blksz), vmat(1_i4b:blksz,i) )
#endif
!
                        v(i) = dot_product( mat(1_i4b:m,l), u(1_i4b:m) )
#ifdef _BLAS
                        call axpy( m, v(i), u(1_i4b:m), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
                        u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + v(i)*u(1_i4b:m)
#endif
!
                        mat(1_i4b:m,i) = u(1_i4b:m)
!
                    end do
!
                else
!
                    do i = l+1_i4b, n
!
                        v(i) = dot_product( mat(1_i4b:m,l), mat(1_i4b:m,i) )
#ifdef _BLAS
                        call axpy( m, v(i), mat(1_i4b:m,i), 1_i4b, u_tmp(1_i4b:m), 1_i4b )
#else
                        u_tmp(1_i4b:m) = u_tmp(1_i4b:m) + v(i)*mat(1_i4b:m,i)
#endif
!
                    end do
!
                end if
!
!               COMPUTE Q IN FACTORED FORM, STORING THE H(l) HOUSEHOLDER TRANSFORMATION
!               IN mat(l,l) AND mat(l+1:n,l).
!
                alpha = v(l+1_i4b)        
!
                call hous1( v(l+1_i4b:n), tau, beta )
!
                if ( store_q ) then
!
!                   SAVE THE H(l) HOUSEHOLDER TRANSFORMATION IN mat(l:n,l) .
!
                    mat(l,l) = tau
                    mat(l+1_i4b:n,l) = v(l+1_i4b:n)
!
                end if
!
                if ( tau/=zero ) then
!
!                   APPLY THE H(l) HOUSEHOLDER TRANSFORMATION TO mat(:m,l+1:n) .
!
                    u(1_i4b:m) = tau*(( u_tmp(1_i4b:m) - beta*mat(1_i4b:m,l+1_i4b) )/( alpha - beta ))
!
!                   APPLY THE H(l) HOUSEHOLDER TRANSFORMATION TO mat(:m,l+1:n) .
!
#ifdef _BLAS
                    call ger( m, n-l, one, u(1_i4b:m), 1_i4b, v(l+1_i4b:n), 1_i4b, mat(1_i4b:m,l+1_i4b:n), m )
#else
                    do i = l+1_i4b, n
!
                        mat(1_i4b:m,i) = mat(1_i4b:m,i) + u(1_i4b:m)*v(i)
!
                    end do
#endif
!
                end if
!
!               COMPUTE THE l+1 ROW OF THE TRIDIAGONAL MATRIX.
!
                d(l+1_i4b) = dot_product( mat(1_i4b:m,l+1_i4b), mat(1_i4b:m,l+1_i4b) )
                e(l)       = beta
!
!
            end do
!
        end if
!
#ifdef _OPENMP
    end if
#endif
!
    if ( store_q ) then
!
        mat(n,n) = machhuge
!
    end if
!
    e(n) = zero
!
!
! END OF SUBROUTINE symtrid_cmp2
! ______________________________
!
    end subroutine symtrid_cmp2
!
! =========================================================================================
!
    subroutine ortho_gen_symtrid_a( mat, upper )
!
! Purpose
! _______
!
!   ORTHO_GEN_SYMTRID generates a real orthogonal matrix Q, which is defined
!   as the product of n-1 elementary reflectors of order n, as returned by
!   SYMTRID_CMP with STORE_Q = true:
!
!      - if UPPER = true,  Q = H(n-1) * . . . * H(2) * H(1),
!
!      - if UPPER = false, Q = H(1) * H(2) * . . . * H(n-1).
!
!
! Arguments
! _________
!
!   MAT      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, the vectors and the scalars, which define the elementary
!            reflectors, as returned by SYMTRID_CMP with STORE_Q = true.
!
!            On exit, the n-by-n orthogonal matrix Q.
!
!   UPPER    (INPUT) logical(lgl)
!            If:
!
!            - UPPER= true : Upper triangle of MAT contains elementary reflectors
!              from SYMTRID_CMP;
!            - UPPER = false: Lower triangle of MAT contains elementary reflectors
!              from SYMTRID_CMP.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the routine  SORGTR in LAPACK. A blocked algorithm
!   is used for agregating the Householder transformations (e.g. the elementary reflectors)
!   stored in the upper or lower triangle of MAT and generating the orthogonal matrix Q.
!   Furthermore, the computations are parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989:
!           Block reduction of matrices to condensed form for eigenvalue computations. 
!           J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!   (3) Walker, H.F., 1988:
!           Implementation of the GMRES method using Householder transformations.
!           Siam J. Sci. Stat. Comput., Vol. 9, No 1, pp. 152-163.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : machhuge
    use Char_Constants,  only : eig_error1
    use Utilities,       only : assert_eq, merror
#ifdef _BLAS
    use BLAS_interfaces, only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:,:) :: mat
!
    logical(lgl), intent(in)  :: upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                  :: prod, tau
    real(stnd), dimension( blksz )              :: x
    real(stnd), dimension( size(mat,1) )        :: tmp
    real(stnd), dimension( size(mat,1), blksz ) :: umat
    real(stnd), dimension( blksz, blksz )       :: lmat
!
    integer(i4b) :: n, i, i0, i1, i2, l, l2, j, blksz2
!
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='ortho_gen_symtrid'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    name_proc )
!
    if ( upper ) then
        tau = mat(1_i4b,1_i4b)
    else
        tau = mat(n,n)
    end if
!
    if ( tau/=machhuge ) then
        call merror( name_proc//eig_error1 )
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) return
!
#ifdef _OPENMP
    i = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*n>=omp_limit                      .and.      &
                  n/i>=omp_limit2                     .and.      &
                  i>1_i4b
#endif
!
!   INITIALIZE THE UNIT LOWER OR UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!   USED IN THE BLOCKED ALGORITHM.
!
    lmat(1_i4b:blksz,1_i4b:blksz) = zero
!
    do i = 1_i4b, blksz
        lmat(i,i) = one
    end do
!
    if ( upper ) then
!
!       Q WAS DETERMINED BY A CALL TO symtrid_cmp_a WITH upper = true .
!
        mat(1_i4b,1_i4b)   = mat(2_i4b,2_i4b)*( mat(1_i4b,2_i4b)**2 ) + one
        mat(2_i4b:n,1_i4b) = zero
!
!$OMP PARALLEL IF(test_par)                                &
!$OMP         ,PRIVATE(i,i1,i2,l,l2,j,blksz2,tau,prod,x,tmp)
!
        do i = 2_i4b, n - 1_i4b, blksz
!    
            blksz2 = min( blksz, n - i )
!    
!           APPLY H(i+blksz2-1)* ...*H(i) TO mat(1:i+blksz2-1,1:i-1) FROM THE LEFT
!           WITH A BLOCKED ALGORITHM AS
!
!          ( I + U(:n,:blksz2)*inv( lmat(:blksz2,:blksz2) )*U(:n,:blksz2)' )*mat(:n,1:i-1) .
!
!           SAVE H(i) ...H(i+blksz2-1) FOR LATER USE AND COMPUTE THE blksz2 BY blksz2 UNIT
!           LOWER TRIANGULAR MATRIX lmat(:blksz2,:blksz2) USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
            do l = 1_i4b, blksz2
!
                i1 = i + l
!
                umat(1_i4b:i1,l) = mat(1_i4b:i1,i1)
!
                i1 = i1 - 1_i4b
!
                do l2 = l + 1_i4b, blksz2
!
                    i2 = i + l2
!
                    tau = mat(i2,i2)
!
                    lmat(l2,l) = -tau*dot_product( umat(1_i4b:i1,l), mat(1_i4b:i1,i2) )
!
                end do
!                    
            end do
!
!$OMP END DO
!
!$OMP DO SCHEDULE(STATIC,1)
!
            do j = 1_i4b, i-1_i4b
!
!               COMPUTE U(:n,:blksz2)'*mat(:n,j) .
!
                tmp(1_i4b:i+blksz2-1_i4b) = mat(1_i4b:i+blksz2-1_i4b,j)
!
                do l = 1_i4b, blksz2
!
                    i1 = i + l
!
                    tau = umat(i1,l)
!
                    x(l) = tau*dot_product( umat(1_i4b:i1-1_i4b,l), tmp(1_i4b:i1-1_i4b) )
!                    
                end do
!
!               SOLVE lmat(:blksz2,:blksz2)*x(:blksz2) = U(:n,:blksz2)'*mat(:n,j) .
!
!               COMPUTE mat(:n,j) = mat(:n,j) + U(:n,:blksz2)*x(:blksz2) .
!
                do l = 1_i4b, blksz2
!
                    x(l+1_i4b:blksz2) = x(l+1_i4b:blksz2) - x(l)*lmat(l+1_i4b:blksz2,l)
!
                    i1 = i + l - 1_i4b
!
#ifdef _BLAS
                    call axpy( i1, x(l), umat(1_i4b:i1,l), 1_i4b, tmp(1_i4b:i1), 1_i4b )
#else
                    tmp(1_i4b:i1) = tmp(1_i4b:i1) + x(l)*umat(1_i4b:i1,l)
#endif
!
                end do
!
                mat(1_i4b:i+blksz2-1_i4b,j) = tmp(1_i4b:i+blksz2-1_i4b)
!                    
            end do
!
!$OMP END DO NOWAIT
!    
!$OMP SINGLE
!
            do l = 1_i4b, blksz2
!
                i1 = i + l
!
                tau = umat(i1,l)
!
!               APPLY H(i+l-1) TO mat(1:i+l-1,i+1:i+l-2) FROM THE LEFT.
!
                i1 = i1 - 1_i4b
!
                do i2 = i, i + l - 2_i4b
!
                    prod = tau*dot_product( umat(1_i4b:i1-1_i4b,l), mat(1_i4b:i1-1_i4b,i2) )
!
#ifdef _BLAS
                    call axpy( i1, prod, umat(1_i4b:i1,l), 1_i4b, mat(1_i4b:i1,i2), 1_i4b )
#else
                    mat(1_i4b:i1,i2) = mat(1_i4b:i1,i2) + prod*umat(1_i4b:i1,l)
#endif
!    
                end do
!
!               APPLY H(i+l-1) TO mat(1:i+l-1,i+l-1) FROM THE LEFT.
!
                prod               = tau*umat(i1,l)
                mat(1_i4b:i1,i1)   = prod*umat(1_i4b:i1,l)
                mat(i1,i1)         = one + mat(i1,i1)
!
!               SET mat(i+l:n,i+l-1) TO zero.
!
                mat(i1+1_i4b:n,i1) = zero
!    
            end do
!    
!$OMP END SINGLE
!
        end do
!
!$OMP END PARALLEL
!
!       SET THE LAST COLUMN OF Q TO THAT OF THE UNIT MATRIX.
!
        mat(:n-1_i4b,n) = zero
        mat(n,n) = one
!
    else
!
!       Q WAS DETERMINED BY A CALL TO symtrid_cmp_a WITH upper = false .
!
        mat(n,n) = mat(n-1_i4b,n-1_i4b)*( mat(n,n-1_i4b)**2 ) + one
        mat(:n-1_i4b,n) = zero
!
!$OMP PARALLEL IF(test_par)                                &
!$OMP         ,PRIVATE(i,i0,i1,i2,l,l2,j,blksz2,tau,prod,x,tmp)
!
        do i = n - blksz - 2_i4b, -blksz+1_i4b, -blksz
!    
!           APPLY H(i+1)* ...*H(i+blksz) TO mat(i+2:n,i+blksz+2:n) FROM THE LEFT
!           WITH A BLOCKED ALGORITHM AS
!
!          ( I + U(:n,:blksz)*inv( lmat(:blksz,:blksz) )*U(:n,:blksz)' )*mat(:n,i+blksz+2:n) .
!
            blksz2 = min( blksz, i+blksz )
            i0     = max( 0_i4b, i )
!
!           SAVE H(i+1) ...H(i+blksz) FOR LATER USE AND COMPUTE THE blksz BY blksz UNIT UPPER
!           TRIANGULAR MATRIX lmat(:blksz,:blksz) USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
            do l = 1_i4b, blksz2
!
                i1 = i0 + l
!
                umat(i1:n,l) = mat(i1:n,i1)
!
                i1 = i1 + 1_i4b
!
                do l2 = 1_i4b, l - 1_i4b
!
                    i2 = i0 + l2
!
                    tau = mat(i2,i2)
!
                    lmat(l2,l) = -tau*dot_product( umat(i1:n,l), mat(i1:n,i2) )
!
                end do
!                    
            end do
!
!$OMP END DO
!
!$OMP DO SCHEDULE(STATIC,1)
!
            do j = i0 + blksz2 + 2_i4b, n
!
!               COMPUTE U(:n,:blksz)'*mat(:n,j) .
!
                tmp(i0+2_i4b:n) = mat(i0+2_i4b:n,j)
!
                do l = 1_i4b, blksz2
!
                    i1 = i0 + l
!
                    tau = umat(i1,l)
!
                    x(l) = tau*dot_product( umat(i1+1_i4b:n,l), tmp(i1+1_i4b:n) )
!                    
                end do
!
!               SOLVE lmat(:blksz,:blksz)*x(:blksz) = U(:n,:blksz)'*mat(:n,j) .
!
!               COMPUTE mat(:n,j) = mat(:n,j) + U(:n,:blksz)*x(:blksz) .
!
                do l = blksz2, 1_i4b, -1_i4b
!
                    x(1_i4b:l-1_i4b) = x(1_i4b:l-1_i4b) - x(l)*lmat(1_i4b:l-1_i4b,l)
!
                    i1 = i0 + l + 1_i4b
!
#ifdef _BLAS
                    call axpy( n-i1+1_i4b, x(l), umat(i1:n,l), 1_i4b,  tmp(i1:n), 1_i4b )
#else
                    tmp(i1:n) = tmp(i1:n) + x(l)*umat(i1:n,l)
#endif
!
                end do
!
                mat(i0+2_i4b:n,j) = tmp(i0+2_i4b:n)
!                    
            end do
!
!$OMP END DO NOWAIT
!    
!$OMP SINGLE
!
            do l = blksz2, 1_i4b, -1_i4b
!
                i1 = i0 + l
!
                tau = umat(i1,l)
!
                i1 = i1 + 1_i4b
!
!               APPLY H(i+l) TO mat(i+l+1:n,i+l+2:i+blksz+1) FROM THE LEFT.
!
                do l2 = l + 2_i4b, blksz2 + 1_i4b
!
                    i2 = i0 + l2
!
                    prod = tau*dot_product( umat(i1+1_i4b:n,l), mat(i1+1_i4b:n,i2) )
#ifdef _BLAS
                    call axpy( n-i1+1_i4b, prod, umat(i1:n,l), 1_i4b,  mat(i1:n,i2), 1_i4b )
#else
                    mat(i1:n,i2) = mat(i1:n,i2) + prod*umat(i1:n,l)
#endif
!    
                end do
!
!               SET mat(1:i+l,i+l+1) TO zero.
!
                mat(1_i4b:i1-1_i4b,i1) = zero
!
!               APPLY H(i+l) TO mat(i+l+1:n,i+l+1) FROM THE LEFT.
!
                prod         = tau*umat(i1,l)
                mat(i1:n,i1) = prod*umat(i1:n,l)
                mat(i1,i1)   = mat(i1,i1) + one
!    
            end do
!    
!$OMP END SINGLE
!
        end do
!
!$OMP END PARALLEL
!
!       SET THE FIRST COLUMN OF Q TO THAT OF THE UNIT MATRIX.
!
        mat(2_i4b:n,1_i4b) = zero
        mat(1_i4b,1_i4b) = one
!
    end if
!
!
! END OF SUBROUTINE ortho_gen_symtrid_a
! _____________________________________
!
    end subroutine ortho_gen_symtrid_a
!
! =========================================================================================
!
    subroutine ortho_gen_symtrid_b( mat )
!
! Purpose
! _______
!
!   ORTHO_GEN_SYMTRID generates a real orthogonal matrix Q, which is defined
!   as the product of n-1 elementary reflectors of order n, as returned by
!   SYMTRID_CMP with STORE_Q = true:
!
!               Q = H(n-1) * . . . * H(2) * H(1)
!
!
! Arguments
! _________
!
!   MAT      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, the vectors and the scalars, which define the elementary
!            reflectors, as returned by SYMTRID_CMP with STORE_Q = true.
!
!            On exit, the n-by-n orthogonal matrix Q.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the routine  SORGTR in LAPACK. A blocked algorithm
!   is used for agregating the Householder transformations (e.g. the elementary reflectors)
!   stored in the upper triangle of MAT and generating the orthogonal matrix Q.
!   Furthermore, the computations are parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989:
!           Block reduction of matrices to condensed form for eigenvalue computations. 
!           J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!   (3) Walker, H.F., 1988:
!           Implementation of the GMRES method using Householder transformations.
!           Siam J. Sci. Stat. Comput., Vol. 9, No 1, pp. 152-163.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,  only : machhuge
    use Char_Constants, only : eig_error1
    use Utilities,      only : assert_eq, merror
#ifdef _BLAS
    use BLAS_interfaces, only : axpy, trsm, gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:,:) :: mat
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                  :: prod, tau
    real(stnd), dimension( blksz )              :: beta
    real(stnd), dimension( blksz, blksz )       :: lmat
    real(stnd), dimension( size(mat,1), blksz ) :: umat
#ifdef _BLAS
    real(stnd), dimension( blksz, blksz )       :: x
#else
    real(stnd), dimension( blksz )              :: x
#endif
!
    integer(i4b) :: n, i, i1, i2, i3, l, l2, j, blksz2, p
!
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='ortho_gen_symtrid'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    name_proc )
!
    if ( mat(1_i4b,1_i4b)/=machhuge ) then
        call merror( name_proc//eig_error1 )
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) return
!
!   Q WAS DETERMINED BY A CALL TO symtrid_cmp_b .
!
    mat(1_i4b,1_i4b)   = mat(2_i4b,2_i4b)*( mat(1_i4b,2_i4b)**2 ) + one
    mat(2_i4b:n,1_i4b) = zero
!
#ifdef _OPENMP
    i = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*n>=omp_limit                      .and.      &
                  n/i>=omp_limit2                     .and.      &
                  i>1_i4b
#endif
!
!   INITIALIZE THE UNIT LOWER OR UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!   USED IN THE BLOCKED ALGORITHM.
!
    lmat(1_i4b:blksz,1_i4b:blksz) = zero
!
    do i = 1_i4b, blksz
        lmat(i,i) = one
    end do
!
!$OMP PARALLEL IF(test_par)                                 &
!$OMP         ,PRIVATE(i,i1,i2,i3,l,l2,j,blksz2,p,tau,prod,x)
!
    do i = 2_i4b, n - 1_i4b, blksz
!    
        blksz2 = min( blksz, n - i )
!    
!       APPLY H(i+blksz2-1)* ...*H(i) TO mat(1:i+blksz2-1,1:i-1) FROM THE LEFT
!       WITH A BLOCKED ALGORITHM AS
!
!          ( I + U(:n,:blksz2)*inv( lmat(:blksz2,:blksz2) )*U(:n,:blksz2)' )*mat(:n,1:i-1) .
!
!       SAVE H(i) ...H(i+blksz2-1) FOR LATER USE AND COMPUTE THE blksz2 BY blksz2 UNIT
!       LOWER TRIANGULAR MATRIX lmat(:blksz2,:blksz2) USED IN THE BLOCKED ALGORITHM.
!
        p = i + blksz2 - 1_i4b
!
!$OMP DO SCHEDULE(STATIC,1)
!
        do l = 1_i4b, blksz2
!
            i2 = i + l
            i1 = i2 - 1_i4b
!
!           SAVE SHIFTED HOUSEHOLDER VECTOR mat(:i1,i2) .
!
            umat(1_i4b:i1,l) = mat(1_i4b:i1,i2)
            umat(i2:p,l)     = zero
!
            beta(l) = mat(i2,i2)
!
            do l2 = l + 1_i4b, blksz2
!
                i2 = i + l2
!
                tau = mat(i2,i2)
!
                lmat(l2,l) = -tau*dot_product( umat(1_i4b:i1,l), mat(1_i4b:i1,i2) )
!
            end do
!                    
        end do
!
!$OMP END DO
!
!       BLOCK UPDATE OF THE MATRIX mat(1:i+blksz2-1,1:i-1) .
!
!$OMP DO SCHEDULE(STATIC,1)
!
        do j = 0_i4b, i-1_i4b, blksz2
!
            i1 = j + 1_i4b
            i2 = min( j+blksz2, i-1_i4b )
            i3 = max( i2-j, 0_i4b )
!
#ifdef _BLAS
!
!           COMPUTE U(:n,:blksz2)'*mat(:n,j+1:j+blksz2) .
!
            call gemm( 'T', 'N', blksz2, i3, p, one, umat, n, mat(:n,i1:i2), n, zero, x, blksz  )
!
            do l = 1_i4b, i3
!
                x(:blksz2,l) = beta(:blksz2)*x(:blksz2,l)
!                    
            end do
!
!           SOLVE lmat(:blksz2,:blksz2)*x(:blksz2,:blksz2) = U(:n,:blksz2)'*mat(:n,j+1:j+blksz2).
!
            call trsm( 'L', 'L', 'N', 'U', blksz2, i3, one, lmat, blksz, x, blksz )
!
!           COMPUTE mat(:n,j+1:j+blksz2) = mat(:n,j+1:j+blksz2) + U(:n,:blksz2)*x(:blksz2,:blksz2) .
!
            call gemm( 'N', 'N', p, i3, blksz2, one, umat, n, x, blksz, one, mat(:n,i1:i2), n )
#else
!
            do l = 1_i4b, i3
!
                i2 = j + l
!
!               COMPUTE U(:n,:blksz2)'*mat(:n,j+1:j+blksz2) .
!
                do l2 = 1_i4b, blksz2
!
                    i1 = i + l2 - 1_i4b
!
                    x(l2) = beta(l2)*dot_product( umat(1_i4b:i1,l2), mat(1_i4b:i1,i2) )
!                    
                end do
!
!               SOLVE lmat(:blksz2,:blksz2)*x(:blksz2,:blksz2) = U(:n,:blksz2)'*mat(:n,j+1:j+blksz2) .
!               COMPUTE mat(:n,j+1:j+blksz2) = mat(:n,j+1:j+blksz2) + U(:n,:blksz2)*x(:blksz2,:blksz2) .
!                    
                do l2 = 1_i4b, blksz2
!
                    x(l2+1_i4b:blksz2) = x(l2+1_i4b:blksz2) - x(l2)*lmat(l2+1_i4b:blksz2,l2)
!
                    i1 = i + l2 - 1_i4b
!
                    mat(1_i4b:i1,i2) = mat(1_i4b:i1,i2) + x(l2)*umat(1_i4b:i1,l2)
!
                end do
!
            end do
#endif
!                    
        end do
!
!$OMP END DO NOWAIT
!    
!$OMP SINGLE
!
        do l = 1_i4b, blksz2
!
            tau = beta(l)
!
!           APPLY H(i+l-1) TO mat(1:i+l-1,i+1:i+l-2) FROM THE LEFT.
!
            i1 = i + l - 1_i4b
!
            do i2 = i, i + l - 2_i4b
!
                prod = tau*dot_product( umat(1_i4b:i1-1_i4b,l), mat(1_i4b:i1-1_i4b,i2) )
#ifdef _BLAS
                call axpy( i1, prod, umat(1_i4b:i1,l), 1_i4b, mat(1_i4b:i1,i2), 1_i4b )
#else
                mat(1_i4b:i1,i2) = mat(1_i4b:i1,i2) + prod*umat(1_i4b:i1,l)
#endif
!    
            end do
!
!           APPLY H(i+l-1) TO mat(1:i+l-1,i+l-1) FROM THE LEFT.
!
            prod             = tau*umat(i1,l)
            mat(1_i4b:i1,i1) = prod*umat(1_i4b:i1,l)
            mat(i1,i1)       = one + mat(i1,i1)
!
!           SET mat(i+l:n,i+l-1) TO zero.
!
            mat(i1+1_i4b:n,i1) = zero
!    
        end do
!    
!$OMP END SINGLE
!
    end do
!
!$OMP END PARALLEL
!
!   SET THE LAST COLUMN OF Q TO THAT OF THE UNIT MATRIX.
!
    mat(:n-1_i4b,n) = zero
    mat(n,n) = one
!
!
! END OF SUBROUTINE ortho_gen_symtrid_b
! _____________________________________
!
    end subroutine ortho_gen_symtrid_b
!
! =========================================================================================
!
    subroutine apply_q_symtrid_a( mat, c, left, trans, upper )
!
! Purpose
! _______
!
!   APPLY_Q_SYMTRID overwrites the general real m-by-n matrix C with:
!
!           - Q * C  if LEFT = true  and TRANS = false, or
!
!           - Q'* C  if LEFT = true  and TRANS = true,  or
!
!           - C * Q  if LEFT = false and TRANS = false, or
!
!           - C * Q' if LEFT = false and TRANS = true,
!
!   where Q is a real orthogonal matrix of order nq and is defined as the product
!   of nq-1 elementary reflectors:
!
!           - Q = H(nq-1) * . . . * H(2) * H(1), if UPPER = true 
!
!           - Q = H(1) * H(2) * . . . * H(nq-1), if UPPER = false 
!
!   as returned by SYMTRID_CMP with STORE_Q = true.
!
!   Q is of order m (=nq) and is the product of m-1 reflectors if LEFT = true ;
!   Q is of order n (=nq) and is the product of n-1 reflectors if LEFT = false.
!
!
! Arguments
! _________
!
!   MAT          (INPUT) real(stnd), dimension(:,:)
!                On entry, the vectors and the scalars, which define
!                the elementary reflectors, as returned by SYMTRID_CMP with STORE_Q = true.
!                MAT is not modified by the routine.
!
!                The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = nq.
!
!   C            (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the m by n matrix C.
!
!                On exit, C is overwritten by Q * C or Q' * C or C * Q' or C * Q.
!
!                The shape of C must verify:
!
!                - if LEFT = true,  size( C, 1 ) = nq ;
!                - if LEFT = false, size( C, 2 ) = nq .
!
!   LEFT         (INPUT) logical(lgl)
!                If:
!
!                - LEFT = true  : apply Q or Q' from the left  ;
!                - LEFT = false : apply Q or Q' from the right .
!
!   TRANS        (INPUT) logical(lgl)
!                If:
!
!                - TRANS = false : apply Q  (no transpose) ;
!                - TRANS = true  : apply Q' (transpose)    .
!
!   UPPER        (INPUT) logical(lgl)
!                If:
!
!                - UPPER = true : The upper triangle of MAT contains elementary reflectors
!                  generated by SYMTRID_CMP;
!                - UPPER = false: The lower triangle of MAT contains elementary reflectors
!                  generated by SYMTRID_CMP.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the routine SORMTR in LAPACK. A blocked algorithm
!   is used to apply the Householder transformations (e.g. the elementary reflectors)
!   stored in the upper or lower triangle of MAT (see the references (2) and (3) below).
!
!   Furthermore, the subroutine is parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989:
!           Block reduction of matrices to condensed form for eigenvalue computations. 
!           J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!   (3) Walker, H.F., 1988:
!           Implementation of the GMRES method using Householder transformations.
!           Siam J. Sci. Stat. Comput., Vol. 9, No 1, pp. 152-163.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : machhuge
    use Char_Constants,  only : eig_error1
    use Utilities,       only : assert, merror
#ifdef _BLAS
    use BLAS_interfaces, only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in),    dimension(:,:) :: mat
    real(stnd),   intent(inout), dimension(:,:) :: c
!
    logical(lgl), intent(in)    :: left, trans, upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension( blksz, blksz ) :: lmat
    real(stnd), dimension( blksz )        :: x
    real(stnd), dimension( size(mat,1) )  :: tmp
    real(stnd)                            :: tau, prod
!
    integer(i4b) :: m, n, nm, nq, i, i1, i2, j, j1, j2, l, l2, p
!
    logical(lgl) :: descending
!
#ifdef _OPENMP
    integer(i4b) :: num_threads, thread_num, jinc, jrem, joffset
!
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='apply_q_symtrid'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( c, 1 )
    n = size( c, 2 )
    nm = n*m
!
    if ( left ) then
        nq = m
        p  = n
    else
        nq = n
        p  = m
    end if
!
    call assert( logical( nq==size(mat,1), lgl ),    &
                 logical( nq==size(mat,2), lgl ),    &
                 name_proc )
!
    if ( upper ) then
        tau = mat(1_i4b,1_i4b)
    else
        tau = mat(nq,nq)
    end if
!
!   Q WAS DETERMINED BY A CALL TO symtrid_cmp_a .
!
    if ( tau/=machhuge ) then
        call merror( name_proc//eig_error1 )
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( nq<=1_i4b .or.  nm<=0_i4b ) return
!
!   INITIALIZE THE UNIT LOWER OR UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!   USED IN THE BLOCKED ALGORITHM.
!
    lmat(1_i4b:blksz,1_i4b:blksz) = zero
!
    do i = 1_i4b, blksz
        lmat(i,i) = one
    end do
!
#ifdef _OPENMP
    i = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  nm>=omp_limit                       .and.      &
                  (nm/nq)/i>=omp_limit2               .and.      &
                  i>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)                                                             &
!$OMP         ,PRIVATE(descending,i,i1,i2,l,l2,j,j1,j2,jrem,jinc,joffset,num_threads,thread_num,tau,prod,tmp,x)
!
#ifdef _OPENMP
    if ( test_par ) then
!
!       DISTRIBUTE THE WORK AMONG THE THREADS.
!
        num_threads = omp_get_num_threads()
        thread_num  = omp_get_thread_num()
!
        jrem = mod( p, num_threads )
        jinc = ( p - jrem )/num_threads
!
        if ( thread_num<jrem ) then
            jinc    = jinc + 1_i4b
            joffset = 1_i4b
        else
            joffset = jrem + 1_i4b
        end if
!
        j1 = thread_num*jinc + joffset
        j2 = j1 + jinc - 1_i4b
!
    else
#endif
        j1 = 1_i4b
        j2 = p
#ifdef _OPENMP
    end if
#endif
!
    if ( upper ) then
!
!       Q WAS DETERMINED BY A CALL TO symtrid_cmp_a WITH upper = true .
!
!       DETERMINE THE ORDER FOR APPLYING THE HOUSEHOLDER TRANSFORMATIONS.
!
        descending = ( trans .and. left ) .or. ( .not. trans .and. .not. left )
!
        if ( left ) then
!
            if ( descending ) then
!
!               COMPUTE H(1)*H(2)* . . .*H(nq-1)*c(:m,:n) .
!
                do i = nq - blksz, 1_i4b, -blksz
!
!                   APPLY H(i)*H(i+1)* . . . *H(i+blksz-1) TO c(:m,:n) FROM THE LEFT AS
!
!                  ( I + U(:m,:blksz)*inv( lmat(:blksz,:blksz) )*U(:m,:blksz)' )*c(:m,:n) .
!
!                   COMPUTE THE blksz BY blksz UNIT UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!                   USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
                    do l = 2_i4b, blksz
!
                        i1 = i + l
!
                        do l2 = 1_i4b, l - 1_i4b
!
                            i2 = i + l2
!
                            tau = mat(i2,i2)
!
                            lmat(l2,l) = -tau*dot_product( mat(1_i4b:i2-1_i4b,i1), mat(1_i4b:i2-1_i4b,i2) )
!
                        end do
!                    
                    end do
!
!$OMP END DO
!
                    do j = j1, j2
!
!                       COMPUTE U(:m,:blksz)'*c(:m,j) .
!
                        tmp(1_i4b:i+blksz-1_i4b) = c(1_i4b:i+blksz-1_i4b,j)
!
                        do l = 1_i4b, blksz
!
                            i1 = i + l
!
                            tau = mat(i1,i1)
!
                            x(l) = tau*dot_product( mat(1_i4b:i1-1_i4b,i1), tmp(1_i4b:i1-1_i4b) )
!                    
                        end do
!
!                       SOLVE lmat(:blksz,:blksz)*x(:blksz) = U(:m,:blksz)'*c(:m,j)  AND
!
!                       COMPUTE c(:m,j) = c(:m,j) + U(:m,:blksz)*x(:blksz) .
!
                        do l = blksz, 1_i4b, -1_i4b
!
                            x(1_i4b:l-1_i4b) = x(1_i4b:l-1_i4b) - x(l)*lmat(1_i4b:l-1_i4b,l)
!
                            i1 = i + l
!
#ifdef _BLAS
                            call axpy( i1-1_i4b, x(l), mat(1_i4b:i1-1_i4b,i1), 1_i4b, tmp(1_i4b:i1-1_i4b), 1_i4b )
#else
                            tmp(1_i4b:i1-1_i4b) = tmp(1_i4b:i1-1_i4b) + x(l)*mat(1_i4b:i1-1_i4b,i1)
#endif
!
                        end do
!
                        c(1_i4b:i+blksz-1_i4b,j) = tmp(1_i4b:i+blksz-1_i4b)
!                    
                    end do
!
!$OMP BARRIER
!                    
                end do
!
!               APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
                i1 = i + blksz
!
                do i = i1, 2_i4b, -1_i4b
!
                    tau = mat(i,i)
!
                    if ( tau==zero ) cycle
!
!                   APPLY H(i) TO c(:m,:n) FROM THE LEFT.
!
                    tmp(1_i4b:i-1_i4b) = mat(1_i4b:i-1_i4b,i)
!
                    do j = j1, j2
!
                        prod = dot_product( tmp(1_i4b:i-1_i4b), c(1_i4b:i-1_i4b,j) )*tau
#ifdef _BLAS
                        call axpy( i-1_i4b, prod, tmp(1_i4b:i-1_i4b), 1_i4b, c(1_i4b:i-1_i4b,j), 1_i4b )
#else
                        c(1_i4b:i-1_i4b,j) = c(1_i4b:i-1_i4b,j) + prod*tmp(1_i4b:i-1_i4b)                
#endif
!
                    end do            
!
                end do
!
            else
!
!               COMPUTE H(nq-1)* . . .*H(2)*H(1)*c(:m,:n) .
!
                do i = 1_i4b, nq - blksz, blksz
!
!                   APPLY H(i+blksz-1)* . . . *H(i+1)*H(i) TO c(:m,:n) FROM THE LEFT AS
!
!                 ( I + U(:m,:blksz)*inv( lmat(:blksz,:blksz) )*U(:m,:blksz)' )*c(:m,:n) .
!
!                   COMPUTE THE blksz BY blksz UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!                   USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
                    do l = 1_i4b, blksz - 1_i4b
!
                        i1 = i + l
!
                        do l2 = l + 1_i4b, blksz
!
                            i2 = i + l2
!
                            tau = mat(i2,i2)
!
                            lmat(l2,l) = -tau*dot_product( mat(1_i4b:i1-1_i4b,i1), mat(1_i4b:i1-1_i4b,i2) )
!
                        end do
!                    
                    end do
!
!$OMP END DO
!
                    do j = j1, j2
!
!                       COMPUTE U(:m,:blksz)'*c(:m,j) .
!
                        tmp(1_i4b:i+blksz-1_i4b) = c(1_i4b:i+blksz-1_i4b,j)
!
                        do l = 1_i4b, blksz
!
                            i1 = i + l
!
                            tau = mat(i1,i1)
!
                            x(l) = tau*dot_product( mat(1_i4b:i1-1_i4b,i1), tmp(1_i4b:i1-1_i4b) )
!                    
                        end do
!
!                       SOLVE lmat(:blksz,:blksz)*x(:blksz) = U(:m,:blksz)'*c(:m,j)  AND
!
!                       COMPUTE c(:m,j) = c(:m,j) + U(:m,:blksz)*x(:blksz) .
!
                        do l = 1_i4b, blksz
!
                            x(l+1_i4b:blksz) = x(l+1_i4b:blksz) - x(l)*lmat(l+1_i4b:blksz,l)
!
                            i1 = i + l
!
#ifdef _BLAS
                            call axpy( i1-1_i4b, x(l), mat(1_i4b:i1-1_i4b,i1), 1_i4b, tmp(1_i4b:i1-1_i4b), 1_i4b )
#else
                            tmp(1_i4b:i1-1_i4b) = tmp(1_i4b:i1-1_i4b) + x(l)*mat(1_i4b:i1-1_i4b,i1)
#endif
!
                        end do
!
                        c(1_i4b:i+blksz-1_i4b,j) = tmp(1_i4b:i+blksz-1_i4b)
!                    
                    end do
!
!$OMP BARRIER
!
                end do
!
!               APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
                i1 = i + 1_i4b
!
                do i = i1, nq
!
                    tau = mat(i,i)
!
                    if ( tau==zero ) cycle
!
!                   APPLY H(i) TO c(:m,:n) FROM THE LEFT.
!
                    tmp(1_i4b:i-1_i4b) = mat(1_i4b:i-1_i4b,i)
!
                    do j = j1, j2
!
                        prod = dot_product( tmp(1_i4b:i-1_i4b), c(1_i4b:i-1_i4b,j) )*tau
#ifdef _BLAS
                        call axpy( i-1_i4b, prod, tmp(1_i4b:i-1_i4b), 1_i4b, c(1_i4b:i-1_i4b,j), 1_i4b )
#else
                        c(1_i4b:i-1_i4b,j) = c(1_i4b:i-1_i4b,j) + prod*tmp(1_i4b:i-1_i4b)              
#endif
!
                    end do            
!
                end do
!
            end if
!
        else
!
            if ( descending ) then
!
!               COMPUTE c(:m,:n)*H(nq-1)* . . .*H(2)*H(1) .
!
                do i = nq - blksz, 1_i4b, -blksz
!
!                   APPLY H(i+blksz-1)* . . .*H(i+1)*H(i) TO c(:m,:n) FROM THE RIGHT AS
!
!                 c(:m,:n)*( I + U(:n,:blksz)*inv( lmat(:blksz,:blksz) )*U(:n,:blksz)' ) .
!
!                   COMPUTE THE blksz BY blksz UNIT UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!                   USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
                    do l = 2_i4b, blksz
!
                        i1 = i + l
!
                        do l2 = 1_i4b, l - 1_i4b
!
                            i2 = i + l2
!
                            tau = mat(i2,i2)
!
                            lmat(l2,l) = -tau*dot_product( mat(1_i4b:i2-1_i4b,i1), mat(1_i4b:i2-1_i4b,i2) )
!
                        end do
!                    
                    end do
!
!$OMP END DO
!
                    do j = j1, j2
!
!                       COMPUTE c(j,:n)*U(:n,:blksz) .
!
                        tmp(1_i4b:i+blksz-1_i4b) = c(j,1_i4b:i+blksz-1_i4b)
!
                        do l = 1_i4b, blksz
!
                            i1 = i + l
!
                            tau = mat(i1,i1)
!
                            x(l) = tau*dot_product( mat(1_i4b:i1-1_i4b,i1), tmp(1_i4b:i1-1_i4b) )
!                    
                        end do
!
!                       SOLVE lmat(:blksz,:blksz)*x(:blksz) = c(j,:n)*U(:n,:blksz) AND
!
!                       COMPUTE c(j,:n) = c(j,:n) + x(:blksz)*U(:n,:blksz)' .
!
                        do l = blksz, 1_i4b, -1_i4b
!
                            x(1_i4b:l-1_i4b) = x(1_i4b:l-1_i4b) - x(l)*lmat(1_i4b:l-1_i4b,l)
!
                            i1 = i + l
!
#ifdef _BLAS
                            call axpy( i1-1_i4b, x(l), mat(1_i4b:i1-1_i4b,i1), 1_i4b, tmp(1_i4b:i1-1_i4b), 1_i4b )
#else
                            tmp(1_i4b:i1-1_i4b) = tmp(1_i4b:i1-1_i4b) + x(l)*mat(1_i4b:i1-1_i4b,i1)
#endif
!
                        end do
!
                        c(j,1_i4b:i+blksz-1_i4b) = tmp(1_i4b:i+blksz-1_i4b)
!                    
                    end do
!
!$OMP BARRIER
!                    
                end do
!
!               APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
                i1 = i + blksz
!
                do i = i1, 2_i4b, -1_i4b
!
                    tau = mat(i,i)
!
                    if ( tau==zero ) cycle
!
!                   APPLY H(i) TO c(:m,:n) FROM THE RIGHT.
!
                    tmp(1_i4b:i-1_i4b) = mat(1_i4b:i-1_i4b,i)
!
                    do j = j1, j2
!
                        prod = dot_product( tmp(1_i4b:i-1_i4b), c(j,1_i4b:i-1_i4b) )*tau
!#ifdef _BLAS
!                        call axpy( i-1_i4b, prod, tmp(1_i4b:i-1_i4b), 1_i4b, c(j,1_i4b:i-1_i4b), 1_i4b )
!#else
                        c(j,1_i4b:i-1_i4b) = c(j,1_i4b:i-1_i4b) + prod*tmp(1_i4b:i-1_i4b)                
!#endif
!
                    end do            
!
                end do
!
            else
!
!               COMPUTE c(:m,:n)*H(1)*H(2)* . . .*H(nq-1) .
!
                do i = 1_i4b, nq - blksz, blksz
!
!                   APPLY H(i)*H(i+1)* . . .*H(i+blksz-1) TO c(:m,:n) FROM THE RIGHT AS
!
!                 c(:m,:n)*( I + U(:n,:blksz)*inv( lmat(:blksz,:blksz) )*U(:n,:blksz)' ) .
!
!                   COMPUTE THE blksz BY blksz UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!                   USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
                    do l = 1_i4b, blksz - 1_i4b
!
                        i1 = i + l
!
                        do l2 = l + 1_i4b, blksz
!
                            i2 = i + l2
!
                            tau = mat(i2,i2)
!
                            lmat(l2,l) = -tau*dot_product( mat(1_i4b:i1-1_i4b,i1), mat(1_i4b:i1-1_i4b,i2) )
!
                        end do
!                    
                    end do
!
!$OMP END DO
!
                    do j = j1, j2
!
!                       COMPUTE c(j,:n)*U(:n,:blksz) .
!
                        tmp(1_i4b:i+blksz-1_i4b) = c(j,1_i4b:i+blksz-1_i4b)
!
                        do l = 1_i4b, blksz
!
                            i1 = i + l
!
                            tau = mat(i1,i1)
!
                            x(l) = tau*dot_product( mat(1_i4b:i1-1_i4b,i1), tmp(1_i4b:i1-1_i4b) )
!                    
                        end do
!
!                       SOLVE lmat(:blksz,:blksz)*x(:blksz) = c(j,:n)*U(:n,:blksz) AND
!
!                       COMPUTE c(j,:n) = c(j,:n) + x(:blksz)*U(:n,:blksz)' .
!
                        do l = 1_i4b, blksz
!
                            x(l+1_i4b:blksz) = x(l+1_i4b:blksz) - x(l)*lmat(l+1_i4b:blksz,l)
!
                            i1 = i + l
!
#ifdef _BLAS
                            call axpy( i1-1_i4b, x(l), mat(1_i4b:i1-1_i4b,i1), 1_i4b, tmp(1_i4b:i1-1_i4b), 1_i4b )
#else
                            tmp(1_i4b:i1-1_i4b) = tmp(1_i4b:i1-1_i4b) + x(l)*mat(1_i4b:i1-1_i4b,i1)
#endif
!
                        end do
!
                        c(j,1_i4b:i+blksz-1_i4b) = tmp(1_i4b:i+blksz-1_i4b)
!                    
                    end do
!
!$OMP BARRIER
!
                end do
!
!               APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
                i1 = i + 1_i4b
!
                do i = i1, nq
!
                    tau = mat(i,i)
!
                    if ( tau==zero ) cycle
!
!                   APPLY H(i) TO c(:m,:n) FROM THE RIGHT.
!
                    tmp(1_i4b:i-1_i4b) = mat(1_i4b:i-1_i4b,i)
!
                    do j = j1, j2
!
                        prod = dot_product( tmp(1_i4b:i-1_i4b), c(j,1_i4b:i-1_i4b) )*tau
!#ifdef _BLAS
!                        call axpy( i-1_i4b, prod, tmp(1_i4b:i-1_i4b), 1_i4b, c(j,1_i4b:i-1_i4b), 1_i4b )
!#else
                        c(j,1_i4b:i-1_i4b) = c(j,1_i4b:i-1_i4b) + prod*tmp(1_i4b:i-1_i4b)           
!#endif
!
                    end do            
!
                end do
!
            end if
!
        end if
!
    else
!
!       Q WAS DETERMINED BY A CALL TO symtrid_cmp_a WITH upper = false .
!
        descending = .not.( ( trans .and. left ) .or. ( .not. trans .and. .not. left ) )
!
        if ( left ) then
!
            if ( descending ) then
!
!               COMPUTE H(1)*H(2)* . . .*H(nq-1)*c(:m,:n) .
!
                do i = nq - blksz - 1_i4b, 0_i4b, -blksz
!
!                   APPLY H(i+1)*H(i+2)* . . . *H(i+blksz) TO c(:m,:n) FROM THE LEFT AS
!
!                  ( I + U(:m,:blksz)*inv( lmat(:blksz,:blksz) )*U(:m,:blksz)' )*c(:m,:n) .
!
!                   COMPUTE THE blksz BY blksz UNIT UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!                   USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
                    do l = 2_i4b, blksz
!
                        i1 = i + l
!
                        do l2 = 1_i4b, l - 1_i4b
!
                            i2 = i + l2
!
                            tau = mat(i2,i2)
!
                            lmat(l2,l) = -tau*dot_product( mat(i1+1_i4b:m,i1), mat(i1+1_i4b:m,i2) )
!
                        end do
!                    
                    end do
!
!$OMP END DO
!
                    do j = j1, j2
!
!                       COMPUTE U(:m,:blksz)'*c(:m,j) .
!
                        tmp(i+2_i4b:m) = c(i+2_i4b:m,j)
!
                        do l = 1_i4b, blksz
!
                            i1 = i + l
!
                            tau = mat(i1,i1)
!
                            x(l) = tau*dot_product( mat(i1+1_i4b:m,i1), tmp(i1+1_i4b:m) )
!                    
                        end do
!
!                       SOLVE lmat(:blksz,:blksz)*x(:blksz) = U(:m,:blksz)'*c(:m,j)  AND
!
!                       COMPUTE c(:m,j) = c(:m,j) + U(:m,:blksz)*x(:blksz) .
!
                        do l = blksz, 1_i4b, -1_i4b
!
                            x(1_i4b:l-1_i4b) = x(1_i4b:l-1_i4b) - x(l)*lmat(1_i4b:l-1_i4b,l)
!
                            i1 = i + l
!
#ifdef _BLAS
                            call axpy( m-i1, x(l), mat(i1+1_i4b:m,i1), 1_i4b, tmp(i1+1_i4b:m), 1_i4b )
#else
                            tmp(i1+1_i4b:m) = tmp(i1+1_i4b:m) + x(l)*mat(i1+1_i4b:m,i1)
#endif
!
                        end do
!
                        c(i+2_i4b:m,j) = tmp(i+2_i4b:m)
!                    
                    end do
!
!$OMP BARRIER
!
                end do
!
!               APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
                i1 = i + blksz
!
                do i = i1, 1_i4b, -1_i4b
!
                    tau = mat(i,i)
!
                    if ( tau==zero ) cycle
!
!                   APPLY H(i) TO c(:m,:n) FROM THE LEFT.
!
                    tmp(i+1_i4b:m) = mat(i+1_i4b:m,i)
!
                    do j = j1, j2
!
                        prod = dot_product( tmp(i+1_i4b:m), c(i+1_i4b:m,j) )*tau
#ifdef _BLAS
                        call axpy( m-i, prod, tmp(i+1_i4b:m), 1_i4b, c(i+1_i4b:m,j), 1_i4b )
#else
                        c(i+1_i4b:m,j) = c(i+1_i4b:m,j) + prod*tmp(i+1_i4b:m)
#endif
!
                    end do            
!
                end do
!
            else
!
!               COMPUTE H(nq-1)* . . .*H(2)*H(1)*c(:m,:n) .
!
                do i = 0_i4b, nq - blksz - 1_i4b, blksz
!
!                   APPLY H(i+blksz)* . . .*H(i+2)*H(i+1) FROM THE LEFT TO c(:m,:n) AS
!
!                 ( I + U(:m,:blksz)*inv( lmat(:blksz,:blksz) )*U(:m,:blksz)' )*c(:m,:n) .
!
!                   COMPUTE THE blksz BY blksz UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!                   USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
                    do l = 1_i4b, blksz - 1_i4b
!
                        i1 = i + l
!
                        do l2 = l + 1_i4b, blksz
!
                            i2 = i + l2
!
                            tau = mat(i2,i2)
!
                            lmat(l2,l) = -tau*dot_product( mat(i2+1_i4b:m,i1), mat(i2+1_i4b:m,i2) )
!
                        end do
!                    
                    end do
!
!$OMP END DO
!
                    do j = j1, j2
!
!                       COMPUTE U(:m,:blksz)'*c(:m,j) .
!
                        tmp(i+2_i4b:m) = c(i+2_i4b:m,j)
!
                        do l = 1_i4b, blksz
!
                            i1 = i + l
!
                            tau = mat(i1,i1)
!
                            x(l) = tau*dot_product( mat(i1+1_i4b:m,i1), tmp(i1+1_i4b:m) )
!                    
                        end do
!
!                       SOLVE lmat(:blksz,:blksz)*x(:blksz) = U(:m,:blksz)'*c(:m,j)  AND
!
!                       COMPUTE c(:m,j) = c(:m,j) + U(:m,:blksz)*x(:blksz) .
!
                        do l = 1_i4b, blksz
!
                            x(l+1_i4b:blksz) = x(l+1_i4b:blksz) - x(l)*lmat(l+1_i4b:blksz,l)
!
                            i1 = i + l
!
#ifdef _BLAS
                            call axpy( m-i1, x(l), mat(i1+1_i4b:m,i1), 1_i4b, tmp(i1+1_i4b:m), 1_i4b )
#else
                            tmp(i1+1_i4b:m) = tmp(i1+1_i4b:m) + x(l)*mat(i1+1_i4b:m,i1)
#endif
!
                        end do
!
                        c(i+2_i4b:m,j) = tmp(i+2_i4b:m)
!                    
                    end do
!
!$OMP BARRIER
!
                end do
!
!               APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
                i1 = i + 1_i4b
!
                do i = i1, nq - 1_i4b
!
                    tau = mat(i,i)
!
                    if ( tau==zero ) cycle
!
!                   APPLY H(i) TO c(:m,:n) FROM THE LEFT.
!
                    tmp(i+1_i4b:m) = mat(i+1_i4b:m,i)
!
                    do j = j1, j2
!
                        prod = dot_product( tmp(i+1_i4b:m), c(i+1_i4b:m,j) )*tau
#ifdef _BLAS
                        call axpy( m-i, prod, tmp(i+1_i4b:m), 1_i4b, c(i+1_i4b:m,j), 1_i4b )
#else
                        c(i+1_i4b:m,j) = c(i+1_i4b:m,j) + prod*tmp(i+1_i4b:m)              
#endif
!
                    end do            
!
                end do
!
            end if
!
        else
!
            if ( descending ) then
!
!               COMPUTE c(:m,:n)*H(nq-1)* . . .*H(2)*H(1) .
!
                do i = nq - blksz - 1_i4b, 0_i4b, -blksz
!
!                   APPLY H(i+blksz)* . . .*H(i+2)*H(i+1) TO c(:m,:n) FROM THE RIGHT AS
!
!                  c(:m,:n)*( I + U(:n,:blksz)*inv( lmat(:blksz,:blksz) )*U(:n,:blksz)' ) .
!
!                   COMPUTE THE blksz BY blksz UNIT UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!                   USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
                    do l = 2_i4b, blksz
!
                        i1 = i + l
!
                        do l2 = 1_i4b, l - 1_i4b
!
                            i2 = i + l2
!
                            tau = mat(i2,i2)
!
                            lmat(l2,l) = -tau*dot_product( mat(i1+1_i4b:n,i1), mat(i1+1_i4b:n,i2) )
!
                        end do
!                    
                    end do
!
!$OMP END DO
!
                    do j = j1, j2
!
!                       COMPUTE c(j,:n)*U(:n,:blksz) .
!
                        tmp(i+2_i4b:n) = c(j,i+2_i4b:n)
!
                        do l = 1_i4b, blksz
!
                            i1 = i + l
!
                            tau = mat(i1,i1)
!
                            x(l) = tau*dot_product( mat(i1+1_i4b:n,i1), tmp(i1+1_i4b:n) )
!                    
                        end do
!
!                       SOLVE x(:blksz)*lmat(:blksz,:blksz) = c(j,:n)*U(:n,:blksz) AND
!
!                       COMPUTE c(j,:n) = c(j,:n) + x(:blksz)*U(:n,:blksz)' .
!
                        do l = blksz, 1_i4b, -1_i4b
!
                            x(1_i4b:l-1_i4b) = x(1_i4b:l-1_i4b) - x(l)*lmat(1_i4b:l-1_i4b,l)
!
                            i1 = i + l
!
#ifdef _BLAS
                            call axpy( n-i1, x(l), mat(i1+1_i4b:n,i1), 1_i4b, tmp(i1+1_i4b:n), 1_i4b )
#else
                            tmp(i1+1_i4b:n) = tmp(i1+1_i4b:n) + x(l)*mat(i1+1_i4b:n,i1)
#endif
!
                        end do
!
                        c(j,i+2_i4b:n) = tmp(i+2_i4b:n)
!                    
                    end do
!
!$OMP BARRIER
!
                end do
!
!               APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
                i1 = i + blksz
!
                do i = i1, 1_i4b, -1_i4b
!
                    tau = mat(i,i)
!
                    if ( tau==zero ) cycle
!
!                   APPLY H(i) TO c(:m,:n) FROM THE RIGHT.
!
                    tmp(i+1_i4b:n) = mat(i+1_i4b:n,i)
!
                    do j = j1, j2
!
                        prod = dot_product( tmp(i+1_i4b:n), c(j,i+1_i4b:n) )*tau
!#ifdef _BLAS
!                        call axpy( n-i, prod, tmp(i+1_i4b:n), 1_i4b, c(j,i+1_i4b:n), 1_i4b )
!#else
                        c(j,i+1_i4b:n) = c(j,i+1_i4b:n) + prod*tmp(i+1_i4b:n)
!#endif
!
                    end do            
!
                end do
!
            else
!
!               COMPUTE c(:m,:n)*H(1)*H(2)* . . .*H(nq-1) .
!
                do i = 0_i4b, nq - blksz - 1_i4b, blksz
!
!                   APPLY H(i+1)*H(i+2)* . . .*H(i+blksz) TO c(:m,:n) FROM THE RIGHT AS
!
!                 c(:m,:n)*( I + U(:n,:blksz)*inv( lmat(:blksz,:blksz) )*U(:n,:blksz)' ) .
!
!                   COMPUTE THE blksz BY blksz UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!                   USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
                    do l = 1_i4b, blksz - 1_i4b
!
                        i1 = i + l
!
                        do l2 = l + 1_i4b, blksz
!
                            i2 = i + l2
!
                            tau = mat(i2,i2)
!
                            lmat(l2,l) = -tau*dot_product( mat(i2+1_i4b:n,i1), mat(i2+1_i4b:n,i2) )
!
                        end do
!                    
                    end do
!
!$OMP END DO
!
                    do j = j1, j2
!
!                       COMPUTE c(j,:n)*U(:n,:blksz) .
!
                        tmp(i+2_i4b:n) = c(j,i+2_i4b:n)
!
                        do l = 1_i4b, blksz
!
                            i1 = i + l
!
                            tau = mat(i1,i1)
!
                            x(l) = tau*dot_product( mat(i1+1_i4b:n,i1), tmp(i1+1_i4b:n) )
!                    
                        end do
!
!                       SOLVE x(:blksz)*lmat(:blksz,:blksz) = c(j,:n)*U(:n,:blksz)  AND
!
!                       COMPUTE c(j,:n) = c(j,:n) + x(:blksz)*U(:n,:blksz)' .
!
                        do l = 1_i4b, blksz
!
                            x(l+1_i4b:blksz) = x(l+1_i4b:blksz) - x(l)*lmat(l+1_i4b:blksz,l)
!
                            i1 = i + l
!
#ifdef _BLAS
                            call axpy( n-i1, x(l), mat(i1+1_i4b:n,i1), 1_i4b, tmp(i1+1_i4b:n), 1_i4b )
#else
                            tmp(i1+1_i4b:n) = tmp(i1+1_i4b:n) + x(l)*mat(i1+1_i4b:n,i1)
#endif
!
                        end do
!
                        c(j,i+2_i4b:n) = tmp(i+2_i4b:n)
!                    
                    end do
!
!$OMP BARRIER
!
                end do
!
!               APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
                i1 = i + 1_i4b
!
                do i = i1, nq - 1_i4b
!
                    tau = mat(i,i)
!
                    if ( tau==zero ) cycle
!
!                   APPLY H(i) TO c(:m,:n) FROM THE RIGHT.
!
                    tmp(i+1_i4b:n) = mat(i+1_i4b:n,i)
!
                    do j = j1, j2
!
                        prod = dot_product( tmp(i+1_i4b:n), c(j,i+1_i4b:n) )*tau
!#ifdef _BLAS
!                        call axpy( n-i, prod, tmp(i+1_i4b:n), 1_i4b, c(j,i+1_i4b:n), 1_i4b )
!#else
                        c(j,i+1_i4b:n) = c(j,i+1_i4b:n) + prod*tmp(i+1_i4b:n)              
!#endif
!
                    end do            
!
                end do
!
            end if
!
        end if
!
    end if
!
!$OMP END PARALLEL
!
!
! END OF SUBROUTINE apply_q_symtrid_a
! ___________________________________
!
    end subroutine apply_q_symtrid_a
!
! =========================================================================================
!
    subroutine apply_q_symtrid_b( mat, c, left, trans )
!
! Purpose
! _______
!
!   APPLY_Q_SYMTRID overwrites the general real m-by-n matrix C with
!
!            - Q * C  if LEFT = true  and TRANS = false, or
!
!            - Q'* C  if LEFT = true  and TRANS = true,  or
!
!            - C * Q  if LEFT = false and TRANS = false, or
!
!            - C * Q' if LEFT = false and TRANS = true,
!
!   where Q is a real orthogonal matrix of order nq and is defined as
!   the product of nq-1 elementary reflectors
!
!           Q = H(nq-1) * . . . * H(2) * H(1) 
!
!   as returned by SYMTRID_CMP (with UPPER = true or without this argument)
!   and with STORE_Q = true.
!
!   Q is of order m (=nq) and is the product of m-1 reflectors if LEFT = true ;
!   Q is of order n (=nq) and is the product of n-1 reflectors if LEFT = false.
!
!
! Arguments
! _________
!
!   MAT          (INPUT) real(stnd), dimension(:,:)
!                On entry, the vectors and the scalars, which define
!                the elementary reflectors, as returned by SYMTRID_CMP
!                (with UPPER = true or SYMTRID_CMP without argument UPPER)
!                and with STORE_Q = true.
!                MAT is not modified by the routine.
!
!                The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = nq.
!
!   C            (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the m by n matrix C.
!
!                On exit, C is overwritten by Q * C or Q' * C or C * Q' or C * Q.
!
!                The shape of C must verify:
!
!                - if LEFT = true,  size( C, 1 ) = nq ;
!                - if LEFT = false, size( C, 2 ) = nq .
!
!   LEFT         (INPUT) logical(lgl)
!                If:
!
!                - LEFT = true  : apply Q or Q' from the left  ;
!                - LEFT = false : apply Q or Q' from the right .
!
!   TRANS        (INPUT) logical(lgl)
!                If:
!
!                - TRANS = false : apply Q  (no transpose) ;
!                - TRANS = true  : apply Q' (transpose)    .
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the routine  SORMTR in LAPACK. A blocked algorithm
!   is used to apply the Householder transformations (e.g. the elementary reflectors)
!   stored in the upper triangle of MAT (see the references (2) and (3) below).
!
!   Furthermore, the subroutine is parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989:
!           Block reduction of matrices to condensed form for eigenvalue computations. 
!           J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!   (3) Walker, H.F., 1988:
!           Implementation of the GMRES method using Householder transformations.
!           Siam J. Sci. Stat. Comput., Vol. 9, No 1, pp. 152-163.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : machhuge
    use Char_Constants,  only : eig_error1, allocate_error
    use Utilities,       only : assert, merror
#ifdef _BLAS
    use BLAS_interfaces, only : gemv, ger, trsm, gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in),    dimension(:,:) :: mat
    real(stnd),   intent(inout), dimension(:,:) :: c
!
    logical(lgl), intent(in)    :: left, trans
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                  :: tau, prod
    real(stnd), dimension( blksz )              :: beta
    real(stnd), dimension( size(mat,1) )        :: tmp
    real(stnd), dimension( blksz, blksz )       :: lmat
    real(stnd), dimension(:,:), allocatable     :: umat, c2, x
!
    integer(i4b) :: m, n, nm, nq, i, i1, i2, j, j1, j2, j3, l, l2, p, p2, q, nblk, rem
    integer      :: iok
!
    logical(lgl) :: descending
!
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='apply_q_symtrid'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( c, 1 )
    n = size( c, 2 )
    nm = n*m
!
    if ( left ) then
        nq = m
        p  = n
    else
        nq = n
        p  = m
    end if
!
    call assert( logical( nq==size(mat,1), lgl ),    &
                 logical( nq==size(mat,2), lgl ),    &
                 name_proc )
!
!   Q WAS DETERMINED BY A CALL TO symtrid_cmp_b OR symtrid_cmp_a
!   WITH upper = true .
!
    if ( mat(1_i4b,1_i4b)/=machhuge ) then
        call merror( name_proc//eig_error1 )
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( nq<=1_i4b .or.  nm<=0_i4b ) return
!
!   DETERMINE THE ORDER FOR APPLYING THE HOUSEHOLDER TRANSFORMATIONS.
!
    descending = ( trans .and. left ) .or. ( .not. trans .and. .not. left )
!
!   COMPUTE THE NUMBER OF COLUMN BLOCKS FOR THE INPUT MATRIX.
!
    nblk = p/blksz
    rem  = mod( p, blksz )
!
    if ( rem/=0_i4b ) then
!
        nblk = nblk + 1_i4b
        p2   = p - rem
!
    else
!
        p2   = p - blksz
!
    end if
!
!   INITIALIZE THE UNIT LOWER OR UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!   USED IN THE BLOCKED ALGORITHM.
!
    lmat(1_i4b:blksz,1_i4b:blksz) = zero
!
    do i = 1_i4b, blksz
        lmat(i,i) = one
    end do
!
!   ALLOCATE AND INITIALIZE (SHARED) WORK ARRAY FOR STORING HOUSEHOLDER
!   VECTORS .
!
    allocate( umat(nq,blksz), stat = iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    umat(:nq,:blksz) = zero
!
#ifdef _OPENMP
!
    i = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  nm>=omp_limit                       .and.      &
                  nblk>1_i4b                          .and.      &
                  i>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)                                           &
!$OMP         ,PRIVATE(iok,i,i1,i2,l,l2,j,j1,j2,j3,q,tau,prod,tmp,x,c2)
!
!   Q WAS DETERMINED BY A CALL TO symtrid_cmp_b OR symtrid_cmp_a WITH upper = true.
!
    if ( left ) then
#ifdef _BLAS
!
!       ALLOCATE (PRIVATE) WORK ARRAY FOR A FAST BLOCKED ALGORITHM.
!
        allocate( x(blksz,blksz), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
#endif
!
        if ( descending ) then
!
!           COMPUTE H(1)*H(2)* . . .*H(nq-1)*c(:m,:n) .
!
            do i = nq - blksz, 1_i4b, -blksz
!
!               APPLY H(i)*H(i+1)* . . . *H(i+blksz-1) TO c(:m,:n) FROM THE LEFT AS
!
!             ( I + U(:m,:blksz)*inv( lmat(:blksz,:blksz) )*U(:m,:blksz)' )*c(:m,:n) .
!
                q = i + blksz - 1_i4b
!
!               COMPUTE THE blksz BY blksz UNIT UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!               USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do l = 1_i4b, blksz
!
                    i1 = i + l
!
!                   SAVE HOUSEHOLDER VECTOR mat(:i1-1,i1) .
!
                    umat(1_i4b:i1-1_i4b,l) = mat(1_i4b:i1-1_i4b,i1)
                    umat(i1:q,l)           = zero
!
                    beta(l) = mat(i1,i1)
!
                    do l2 = 1_i4b, l - 1_i4b
!
                        i2 = i + l2
!
                        tau = mat(i2,i2)
!
                        lmat(l2,l) = -tau*dot_product( mat(1_i4b:i2-1_i4b,i1), mat(1_i4b:i2-1_i4b,i2) )
!
                    end do
!                    
                end do
!
!$OMP END DO
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do j = 0_i4b, p2, blksz
!
                    j1 = j + 1_i4b
                    j2 = min( j + blksz, p )
!
#ifdef _BLAS
                    j3 = j2 - j
!
!                   COMPUTE  U(:m,:blksz)'*c(:m,j1:j2) .
!
                    call gemm( 'T', 'N', blksz, j3, q, one, umat, m, c(1_i4b:m,j1:j2), m, zero, x, blksz  )
!
                    do l = 1_i4b, j3
!
                        x(:blksz,l) = beta(:blksz)*x(:blksz,l)
!                    
                    end do
!
!                   SOLVE lmat(:blksz,:blksz)*x(:blksz,:j3) = U(:m,:blksz)'*c(:m,j1:j2) .
!
                    call trsm( 'L', 'U', 'N', 'U', blksz, j3, one, lmat, blksz, x, blksz )
!
!                   COMPUTE c(:m,j1:j2) = c(:m,j1:j2) + U(:m,:blksz)*x(:blksz,:j3) .
!
                    call gemm( 'N', 'N', q, j3, blksz, one, umat, m, x, blksz, one, c(1_i4b:m,j1:j2), m   )
#else
                    do l = j1, j2
!
!                       COMPUTE U(:m,:blksz)'*c(:m,j1:j2) .
!
                        do l2 = 1_i4b, blksz
!
                            i1 = i + l2 - 1_i4b
!
                            tmp(l2) = beta(l2)*dot_product( umat(1_i4b:i1,l2), c(1_i4b:i1,l) )
!                    
                        end do
!
!                       SOLVE lmat(:blksz,:blksz)*x(:blksz,:j3) = U(:m,:blksz)'*c(:m,j1:j2) .
!                       COMPUTE c(:m,j1:j2) = c(:m,j1:j2) + U(:m,:blksz)*x(:blksz,:j3) .
!
                        do l2 = blksz, 1_i4b, -1_i4b
!
                            tmp(1_i4b:l2-1_i4b) = tmp(1_i4b:l2-1_i4b) - tmp(l2)*lmat(1_i4b:l2-1_i4b,l2)
!
                            i1 = i + l2 - 1_i4b
!
                            c(1_i4b:i1,l) = c(1_i4b:i1,l) + tmp(l2)*umat(1_i4b:i1,l2)
!
                        end do
!
                    end do
#endif
!
                end do
!
!$OMP END DO
!                    
            end do
!
!           APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
            i1 = i + blksz
!
            do i = i1, 2_i4b, -1_i4b
!
                tau = mat(i,i)
!
                if ( tau==zero ) cycle
!
!               APPLY H(i) TO c(:m,:n) FROM THE LEFT.
!
                q = i - 1_i4b
!
                tmp(1_i4b:q) = mat(1_i4b:q,i)
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do j = 0_i4b, p2, blksz
!
                    j1 = j + 1_i4b
                    j2 = min( j + blksz, p )
!
#ifdef _BLAS
                    j3 = j2 - j
!
                    call gemv( 'T', q, j3, tau, c(1_i4b:m,j1:j2), m, tmp(1_i4b:q), 1_i4b, zero, x(1_i4b:j3,1_i4b), 1_i4b )
!
                    call ger( q, j3, one, tmp(1_i4b:q), 1_i4b, x(1_i4b:j3,1_i4b), 1_i4b, c(1_i4b:m,j1:j2), m )
#else
                    do l = j1, j2
!
                        prod = dot_product( tmp(1_i4b:q), c(1_i4b:q,l) )*tau
!
                        c(1_i4b:q,l) = c(1_i4b:q,l) + prod*tmp(1_i4b:q)                
!
                    end do            
#endif
!
                end do
!
!$OMP END DO
!
            end do
!
        else
!
!           COMPUTE H(nq-1)* . . .*H(2)*H(1)*c(:m,:n) .
!
            do i = 1_i4b, nq - blksz, blksz
!
!               APPLY H(i+blksz-1)* . . . *H(i+1)*H(i) TO c(:m,:n) FROM THE LEFT AS
!
!             ( I + U(:m,:blksz)*inv( lmat(:blksz,:blksz) )*U(:m,:blksz)' )*c(:m,:n) .
!
                q = i + blksz -1_i4b
!
!               COMPUTE THE blksz BY blksz UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!               USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do l = 1_i4b, blksz
!
                    i1 = i + l
!
!                   SAVE HOUSEHOLDER VECTOR mat(:i1-1,i1) .
!
                    umat(1_i4b:i1-1_i4b,l) = mat(1_i4b:i1-1_i4b,i1)
                    umat(i1:q,l)           = zero
!
                    beta(l) = mat(i1,i1)
!
                    do l2 = l + 1_i4b, blksz
!
                        i2 = i + l2
!
                        tau = mat(i2,i2)
!
                        lmat(l2,l) = -tau*dot_product( mat(1_i4b:i1-1_i4b,i1), mat(1_i4b:i1-1_i4b,i2) )
!
                    end do
!                    
                end do
!
!$OMP END DO
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do j = 0_i4b, p2, blksz
!
                    j1 = j + 1_i4b
                    j2 = min( j + blksz, p )
!
#ifdef _BLAS
                    j3 = j2 - j
!
!                   COMPUTE U(:m,:blksz)'*c(:m,j1:j2) .
!
                    call gemm( 'T', 'N', blksz, j3, q, one, umat, m, c(1_i4b:m,j1:j2), m, zero, x, blksz  )
!
                    do l = 1_i4b, j3
!
                        x(:blksz,l) = beta(:blksz)*x(:blksz,l)
!                    
                    end do
!
!                   SOLVE lmat(:blksz,:blksz)*x(:blksz,:j3) = U(:m,:blksz)'*c(:m,j1:j2) .
!
                    call trsm( 'L', 'L', 'N', 'U', blksz, j3, one, lmat, blksz, x, blksz )
!
!                   COMPUTE c(:m,j1:j2) = c(:m,j1:j2) + U(:m,:blksz)*x(:blksz,:j3) .
!
                    call gemm( 'N', 'N', q, j3, blksz, one, umat, m, x, blksz, one, c(1_i4b:m,j1:j2), m   )
#else
                    do l = j1, j2
!
!                       COMPUTE U(:m,:blksz)'*c(:m,j1:j2) .
!
                        do l2 = 1_i4b, blksz
!
                            i1 = i + l2 - 1_i4b
!
                            tmp(l2) = beta(l2)*dot_product( umat(1_i4b:i1,l2), c(1_i4b:i1,l) )
!                    
                        end do
!
!                       SOLVE lmat(:blksz,:blksz)*x(:blksz,:j3) = U(:m,:blksz)'*c(:m,j1:j2) .
!                       COMPUTE c(:m,j1:j2) = c(:m,j1:j2) + U(:m,:blksz)*x(:blksz,:j3) .
!
                        do l2 = 1_i4b, blksz
!
                            tmp(l2+1_i4b:blksz) = tmp(l2+1_i4b:blksz) - tmp(l2)*lmat(l2+1_i4b:blksz,l2)
!
                            i1 = i + l2 - 1_i4b
!
                            c(1_i4b:i1,l) = c(1_i4b:i1,l) + tmp(l2)*umat(1_i4b:i1,l2)
!
                        end do
!
                    end do
#endif
!
                end do
!
!$OMP END DO
!
            end do
!
!           APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
            i1 = i + 1_i4b
!
            do i = i1, nq
!
                tau = mat(i,i)
!
                if ( tau==zero ) cycle
!
!               APPLY H(i) TO c(:m,:n) FROM THE LEFT.
!
                q = i - 1_i4b
!
                tmp(1_i4b:q) = mat(1_i4b:q,i)
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do j = 0_i4b, p2, blksz
!
                    j1 = j + 1_i4b
                    j2 = min( j + blksz, p )
!
#ifdef _BLAS
                    j3 = j2 - j
!
                    call gemv( 'T', q, j3, tau, c(1_i4b:m,j1:j2), m, tmp(1_i4b:q), 1_i4b, zero, x(1_i4b:j3,1_i4b), 1_i4b )
!
                    call ger( q, j3, one, tmp(1_i4b:q), 1_i4b, x(1_i4b:j3,1_i4b), 1_i4b, c(1_i4b:m,j1:j2), m )
#else
                    do l = j1, j2
!
                        prod = dot_product( tmp(1_i4b:q), c(1_i4b:q,l) )*tau
!
                        c(1_i4b:q,l) = c(1_i4b:q,l) + prod*tmp(1_i4b:q)              
!
                    end do            
#endif
!
                end do
!
!$OMP END DO
!
            end do
!
        end if
#ifdef _BLAS
!
!       DEALLOCATE (PRIVATE) WORK ARRAY.
!
        deallocate( x )
#endif
!
    else
!
!       ALLOCATE (PRIVATE) WORK ARRAYS FOR A FAST BLOCKED ALGORITHM.
!
        allocate( x(blksz,blksz), c2(blksz,n), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        if ( descending ) then
!
!           COMPUTE c(:m,:n)*H(nq-1)* . . .*H(2)*H(1) .
!
            do i = nq - blksz, 1_i4b, -blksz
!
!               APPLY H(i+blksz-1)* . . .*H(i+1)*H(i) TO c(:m,:n) FROM THE RIGHT AS
!
!             c(:m,:n)*( I + U(:n,:blksz)*inv( lmat(:blksz,:blksz) )*U(:n,:blksz)' ) .
!
                q = i + blksz - 1_i4b
!
!               COMPUTE THE blksz BY blksz UNIT UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!               USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do l = 1_i4b, blksz
!
                    i1 = i + l
!
!                   SAVE HOUSEHOLDER VECTOR mat(:i1-1,i1) .
!
                    umat(1_i4b:i1-1_i4b,l) = mat(1_i4b:i1-1_i4b,i1)
                    umat(i1:q,l)           = zero
!
                    beta(l) = mat(i1,i1)
!
                    do l2 = 1_i4b, l - 1_i4b
!
                        i2 = i + l2
!
                        tau = mat(i2,i2)
!
                        lmat(l2,l) = -tau*dot_product( mat(1_i4b:i2-1_i4b,i1), mat(1_i4b:i2-1_i4b,i2) )
!
                    end do
!                    
                end do
!
!$OMP END DO
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do j = 0_i4b, p2, blksz
!
                    j1 = j + 1_i4b
                    j2 = min( j + blksz, p )
                    j3 = j2 - j
!
                    c2(1_i4b:j3,1_i4b:q) = c(j1:j2,1_i4b:q)
!
#ifdef _BLAS
!                   COMPUTE c(j1:j2,:n)*U(:n,:blksz) .
!
                    call gemm( 'N', 'N', j3, blksz, q, one, c2, blksz, umat, n, zero, x, blksz  )
!
                    do l = 1_i4b, blksz
!
                        x(1_i4b:j3,l) = beta(l)*x(1_i4b:j3,l)
!                    
                    end do
!
!                   SOLVE x(j1:j2,:blksz)*lmat(:blksz,:blksz)' = c(j1:j2,:n)*U(:n,:blksz) .
!
                    call trsm( 'R', 'U', 'T', 'U', j3, blksz, one, lmat, blksz, x, blksz )
!
!                   COMPUTE c(j1:j2,:n) = c(j1:j2,:n) + x(j1:j2,:blksz)*U(:n,:blksz)' .
!
                    call gemm( 'N', 'T', j3, q,  blksz, one, x, blksz, umat, n, one, c2, blksz )
#else
!                   COMPUTE c(j1:j2,:n)*U(:n,:blksz) .
!
                    do l = 1_i4b, blksz
!
                        i1 = i + l - 1_i4b
!
                        x(1_i4b:j3,l) = umat(1_i4b,l)*c2(1_i4b:j3,1_i4b)
!
                        do l2 = 2_i4b, i1
!
                            x(1_i4b:j3,l) = x(1_i4b:j3,l) +  umat(l2,l)*c2(1_i4b:j3,l2)
!                    
                        end do
!
                        x(1_i4b:j3,l) = beta(l)*x(1_i4b:j3,l)
!                    
                    end do
!
!                   SOLVE x(j1:j2,:blksz)*lmat(:blksz,:blksz)' = c(j1:j2,:n)*U(:n,:blksz) AND
!                   COMPUTE c(j1:j2,i+1:i+blksz-1) = c(j1:j2,i+1:i+blksz-1) + x(j1:j2,:blksz)*U(i+1:i+blksz-1,:blksz)' .
!
                    do l = blksz, 1_i4b, -1_i4b
!
                        do l2 = 1_i4b, l-1_i4b
!
                            x(1_i4b:j3,l2) = x(1_i4b:j3,l2) - x(1_i4b:j3,l)*lmat(l2,l)
!
                        end do
!
                        do l2 = 1_i4b, l-1_i4b
!
                            i2 = i + l2
!
                            c2(1_i4b:j3,i2) = c2(1_i4b:j3,i2) + x(1_i4b:j3,l)*umat(i2,l)
!
                        end do
!
                    end do
!
!                   COMPUTE c(j1:j2,:i) = c(j1:j2,:i) + x(j1:j2,:blksz)*U(:i,:blksz)' .
!
                    do l = 1_i4b, i
!
                        do l2 = 1_i4b, blksz
!
                            c2(1_i4b:j3,l) = c2(1_i4b:j3,l) + x(1_i4b:j3,l2)*umat(l,l2)
!                    
                        end do
!                    
                    end do
#endif
!
                    c(j1:j2,1_i4b:q) = c2(1_i4b:j3,1_i4b:q)
!
                end do
!
!$OMP END DO
!                    
            end do
!
!           APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
            i1 = i + blksz
!
            do i = i1, 2_i4b, -1_i4b
!
                tau = mat(i,i)
!
                if ( tau==zero ) cycle
!
!               APPLY H(i) TO c(:m,:n) FROM THE RIGHT.
!
                q = i - 1_i4b
!
                tmp(1_i4b:q) = mat(1_i4b:q,i)
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do j = 0_i4b, p2, blksz
!
                    j1 = j + 1_i4b
                    j2 = min( j + blksz, p )
                    j3 = j2 - j
!
                    c2(1_i4b:j3,1_i4b:q) = c(j1:j2,1_i4b:q)
!
#ifdef _BLAS
                    call gemv( 'N', j3, q, tau, c2, blksz, tmp(1_i4b:q), 1_i4b, zero, x(1_i4b:j3,1_i4b), 1_i4b )
!
                    call ger( j3, q, one, x(1_i4b:j3,1_i4b), 1_i4b, tmp(1_i4b:q), 1_i4b, c2, blksz )
#else
                    x(1_i4b:j3,1_i4b) = matmul( c2(1_i4b:j3,:q), tmp(1_i4b:q) )

                    do l = 1_i4b, q
!
                        prod = tmp(l)*tau
!
                        c2(1_i4b:j3,l) = c2(1_i4b:j3,l) + prod*x(1_i4b:j3,1_i4b)     
!
                    end do            
#endif
!
                    c(j1:j2,1_i4b:q) = c2(1_i4b:j3,1_i4b:q)
!
                end do
!
!$OMP END DO
!
            end do
!
        else
!
!           COMPUTE c(:m,:n)*H(1)*H(2)* . . .*H(nq-1) .
!
            do i = 1_i4b, nq - blksz, blksz
!
!               APPLY H(i)*H(i+1)* . . .*H(i+blksz-1) TO c(:m,:n) FROM THE RIGHT AS
!
!             c(:m,:n)*( I + U(:n,:blksz)*inv( lmat(:blksz,:blksz) )*U(:n,:blksz)' ) .
!
                q = i + blksz - 1_i4b
!
!               COMPUTE THE blksz BY blksz UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!               USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do l = 1_i4b, blksz
!
                    i1 = i + l
!
!                   SAVE HOUSEHOLDER VECTOR mat(:i1-1,i1) .
!
                    umat(1_i4b:i1-1_i4b,l) = mat(1_i4b:i1-1_i4b,i1)
                    umat(i1:q,l)           = zero
!
                    beta(l) = mat(i1,i1)
!
                    do l2 = l + 1_i4b, blksz
!
                        i2 = i + l2
!
                        tau = mat(i2,i2)
!
                        lmat(l2,l) = -tau*dot_product( mat(1_i4b:i1-1_i4b,i1), mat(1_i4b:i1-1_i4b,i2) )
!
                    end do
!                    
                end do
!
!$OMP END DO
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do j = 0_i4b, p2, blksz
!
                    j1 = j + 1_i4b
                    j2 = min( j + blksz, p )
                    j3 = j2 - j
!
                    c2(1_i4b:j3,1_i4b:q) = c(j1:j2,1_i4b:q)
!
#ifdef _BLAS
!                   COMPUTE c(j1:j2,:n)*U(:n,:blksz) .
!
                    call gemm( 'N', 'N', j3, blksz, q, one, c2, blksz, umat, n, zero, x, blksz  )
!
                    do l = 1_i4b, blksz
!
                        x(1_i4b:j3,l) = beta(l)*x(1_i4b:j3,l)
!                    
                    end do
!
!                   SOLVE x(j1:j2,:blksz)*lmat(:blksz,:blksz)' = c(j1:j2,:n)*U(:n,:blksz) .
!
                    call trsm( 'R', 'L', 'T', 'U', j3, blksz, one, lmat, blksz, x, blksz )
!
!                   COMPUTE c(j1:j2,:n) = c(j1:j2,:n) + x(j1:j2,:blksz)*U(:n,:blksz)' .
!
                    call gemm( 'N', 'T', j3, q,  blksz, one, x, blksz, umat, n, one, c2, blksz )
#else
!                   COMPUTE c(j1:j2,:n)*U(:n,:blksz) .
!
                    do l = 1_i4b, blksz
!
                        i1 = i + l - 1_i4b
!
                        x(1_i4b:j3,l) = umat(1_i4b,l)*c2(1_i4b:j3,1_i4b)
!
                        do l2 = 2_i4b, i1
!
                            x(1_i4b:j3,l) = x(1_i4b:j3,l) +  umat(l2,l)*c2(1_i4b:j3,l2)
!                    
                        end do
!
                        x(1_i4b:j3,l) = beta(l)*x(1_i4b:j3,l)
!                    
                    end do
!
!                   SOLVE x(j1:j2,:blksz)*lmat(:blksz,:blksz)' = c(j1:j2,:n)*U(:n,:blksz) AND
!                   COMPUTE c(j1:j2,i+1:i+blksz-1) = c(j1:j2,i+1:i+blksz-1) + x(j1:j2,:blksz)*U(i+1:i+blksz-1,:blksz)' .
!
                    do l = 1_i4b, blksz
!
                        do l2 = l + 1_i4b, blksz
!
                            x(1_i4b:j3,l2) = x(1_i4b:j3,l2) - x(1_i4b:j3,l)*lmat(l2,l)
!
                        end do
!
                        do l2 = 1_i4b, l-1_i4b
!
                            i2 = i + l2
!
                            c2(1_i4b:j3,i2) = c2(1_i4b:j3,i2) + x(1_i4b:j3,l)*umat(i2,l)
!
                        end do
!
                    end do
!
!                   COMPUTE c(j1:j2,:i) = c(j1:j2,:i) + x(j1:j2,:blksz)*U(:i,:blksz)' .
!
                    do l = 1_i4b, i
!
                        do l2 = 1_i4b, blksz
!
                            c2(1_i4b:j3,l) = c2(1_i4b:j3,l) + x(1_i4b:j3,l2)*umat(l,l2)
!                    
                        end do
!                    
                    end do
#endif
!
                    c(j1:j2,1_i4b:q) = c2(1_i4b:j3,1_i4b:q)
!
                end do
!
!$OMP END DO
!
            end do
!
!           APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
            i1 = i + 1_i4b
!
            do i = i1, nq
!
                tau = mat(i,i)
!
                if ( tau==zero ) cycle
!
!               APPLY H(i) TO c(:m,:n) FROM THE RIGHT.
!
                q = i - 1_i4b
!
                tmp(1_i4b:q) = mat(1_i4b:q,i)
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do j = 0_i4b, p2, blksz
!
                    j1 = j + 1_i4b
                    j2 = min( j + blksz, p )
                    j3 = j2 - j
!
                    c2(1_i4b:j3,1_i4b:q) = c(j1:j2,1_i4b:q)
!
#ifdef _BLAS
                    call gemv( 'N', j3, q, tau, c2, blksz, tmp(1_i4b:q), 1_i4b, zero, x(1_i4b:j3,1_i4b), 1_i4b )
!
                    call ger( j3, q, one, x(1_i4b:j3,1_i4b), 1_i4b, tmp(1_i4b:q), 1_i4b, c2, blksz )
#else
                    x(1_i4b:j3,1_i4b) = matmul( c2(1_i4b:j3,:q), tmp(1_i4b:q) )

                    do l = 1_i4b, q
!
                        prod = tmp(l)*tau
!
                        c2(1_i4b:j3,l) = c2(1_i4b:j3,l) + prod*x(1_i4b:j3,1_i4b)     
!
                    end do            
#endif
!
                    c(j1:j2,1_i4b:q) = c2(1_i4b:j3,1_i4b:q)
!
                end do
!
!$OMP END DO
!
            end do
!
        end if
!
!       DEALLOCATE (PRIVATE) WORK ARRAYS.
!
        deallocate( x, c2 )
!
    end if
!
!$OMP END PARALLEL
!
!   DEALLOCATE (SHARED) WORK ARRAY.
!
    deallocate( umat )
!
!
! END OF SUBROUTINE apply_q_symtrid_b
! ___________________________________
!
    end subroutine apply_q_symtrid_b
!
! =========================================================================================
!
    subroutine apply_q_symtrid_c( matp, c, left, trans, upper )
!
! Purpose
! _______
!
!   APPLY_Q_SYMTRID overwrites the general real m-by-n matrix C with
!
!            - Q * C  if LEFT = true  and TRANS = false, or
!
!            - Q'* C  if LEFT = true  and TRANS = true,  or
!
!            - C * Q  if LEFT = false and TRANS = false, or
!
!            - C * Q' if LEFT = false and TRANS = true,
!
!   where Q is a real orthogonal matrix of order nq and is defined as the product
!   of nq-1 elementary reflectors
!
!           - Q = H(nq-1) * . . . * H(2) * H(1), if UPPER = true 
!
!           - Q = H(1) * H(2) * . . . * H(nq-1),  if UPPER = false 
!
!   as returned by SYMTRID_CMP with STORE_Q = true.
!
!   Q is of order m (=nq) and is the product of m-1 reflectors if LEFT = true ;
!   Q is of order n (=nq) and is the product of n-1 reflectors if LEFT = false.
!
!
! Arguments
! _________
!
!   MATP         (INPUT) real(stnd), dimension(:)
!                On entry, the vectors and the scalars, which define
!                the elementary reflectors, as returned by SYMTRID_CMP
!                in argument MATP. MATP is not modified by the routine.
!
!                The size of MATP must verify size( MATP ) = (nq * (nq+1)/2)
!
!   C            (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the m by n matrix C.
!
!                On exit, C is overwritten by Q * C or Q' * C or C * Q' or C * Q.
!
!                The shape of C must verify:
!
!                - if LEFT = true,  size( C, 1 ) = nq ;
!                - if LEFT = false, size( C, 2 ) = nq .
!
!   LEFT         (INPUT) logical(lgl)
!                If:
!
!                - LEFT = true  : apply Q or Q' from the left  ;
!                - LEFT = false : apply Q or Q' from the right .
!
!   TRANS        (INPUT) logical(lgl)
!                If:
!
!                - TRANS = false : apply Q  (no transpose) ;
!                - TRANS = true  : apply Q' (transpose)    .
!
!   UPPER        (INPUT) logical(lgl)
!                Specifies whether the upper or lower triangular part of the
!                original symmetric matrix MAT was stored in packed form in MATP
!                before the reduction by SYMTRID_CMP. If:
!
!                - UPPER = true :  Upper triangle of MAT was stored;
!                - UPPER = false:  Lower triangle of MAT was stored.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the routine  DORMTR in LAPACK and uses a blocked
!   algorithm to apply the Householder transformations (e.g. the elementary reflectors)
!   stored in packed form in the vector MATP.
!
!   Furthermore, the subroutine is parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989:
!           Block reduction of matrices to condensed form for eigenvalue computations. 
!           J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!   (3) Walker, H.F., 1988:
!           Implementation of the GMRES method using Householder transformations.
!           Siam J. Sci. Stat. Comput., Vol. 9, No 1, pp. 152-163.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : machhuge
    use Char_Constants,  only : eig_error1
    use Utilities,       only : assert, merror, arth, cumsum
#ifdef _BLAS
    use BLAS_interfaces, only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in)   , dimension(:)   :: matp
    real(stnd),   intent(inout), dimension(:,:) :: c
!
    logical(lgl), intent(in) :: left, trans, upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                        :: tau, prod
    real(stnd), dimension( blksz )                    :: x
    real(stnd), dimension( max(size(c,1),size(c,2)) ) :: tmp
    real(stnd), dimension( blksz, blksz )             :: lmat
!
    integer(i4b)                                        :: m, n, nm, nq, i, i1, i2, j, j1, j2, &
                                                           k, k1, k2, l, l2, p
    integer(i4b), dimension( max(size(c,1),size(c,2)) ) :: il
!
    logical(lgl) :: descending
!
#ifdef _OPENMP
    integer(i4b) :: num_threads, thread_num, jinc, jrem, joffset
!
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='apply_q_symtrid'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = size( c, 1 )
    n = size( c, 2 )
    nm = n*m
!
    if ( left ) then
        nq = m
        p  = n
    else
        nq = n
        p  = m
    end if
!
    k = nq*(nq+1_i4b)/2_i4b
!
    call assert( logical(size(matp)==k,lgl)  ,   &
                 name_proc )
!
    if ( upper ) then
        tau = matp(1_i4b)
    else
        tau = matp(k)
    end if
!
    if ( tau/=machhuge ) then
        call merror( name_proc//eig_error1 )
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( nq<=1_i4b .or. nm<=0_i4b ) return
!
!   INITIALIZE THE UNIT LOWER OR UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!   USED IN THE BLOCKED ALGORITHM.
!
    lmat(1_i4b:blksz,1_i4b:blksz) = zero
!
    do i = 1_i4b, blksz
        lmat(i,i) = one
    end do
!
    if ( upper ) then
!
!       COMPUTE INDICES OF THE FIRST ELEMENT OF EACH COLUMN OF THE ORIGINAL MATRIX.
!
        il(:nq) = cumsum( arth(0_i4b,1_i4b,nq), seed=1_i4b )
!
    else
!
!       COMPUTE INDICES OF THE DIAGONAL ELEMENT OF EACH COLUMN OF THE ORIGINAL MATRIX.
!
        il(1_i4b)    = 1_i4b
        il(2_i4b:nq) = cumsum( arth(nq,-1_i4b,nq-1_i4b), seed=1_i4b )
!
    end if
!
#ifdef _OPENMP
    i = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  nm>=omp_limit                       .and.      &
                  (nm/nq)/i>=omp_limit2               .and.      &
                  i>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)                                                        &
!$OMP         ,PRIVATE(descending,i,i1,i2,k1,k2,l,l2,j,j1,j2,jrem,jinc,joffset,    &
!$OMP                  num_threads,thread_num,tau,prod,tmp,x)
!
#ifdef _OPENMP
    if ( test_par ) then
!
!       DISTRIBUTE THE WORK AMONG THE THREADS.
!
        num_threads = omp_get_num_threads()
        thread_num  = omp_get_thread_num()
!
        jrem = mod( p, num_threads )
        jinc = ( p - jrem )/num_threads
!
        if ( thread_num<jrem ) then
            jinc    = jinc + 1_i4b
            joffset = 1_i4b
        else
            joffset = jrem + 1_i4b
        end if
!
        j1 = thread_num*jinc + joffset
        j2 = j1 + jinc - 1_i4b
!
    else
#endif
        j1 = 1_i4b
        j2 = p
#ifdef _OPENMP
    end if
#endif
!
    if ( upper ) then
!
!       Q WAS DETERMINED BY A CALL TO symtrid_cmp_d OR symtrid_cmp_c WITH upper = true.
!
!       DETERMINE THE ORDER FOR APPLYING THE HOUSEHOLDER TRANSFORMATIONS.
!
        descending = ( trans .and. left ) .or. ( .not. trans .and. .not. left )
!
        if ( left ) then
!
            if ( descending ) then
!
!               COMPUTE H(1)*H(2)* . . .*H(nq-1)*c(:m,:n) .
!
                do i = nq - blksz, 1_i4b, -blksz
!
!                   APPLY H(i)*H(i+1)* . . . *H(i+blksz-1) TO c(:m,:n) FROM THE LEFT AS
!
!                  ( I + U(:m,:blksz)*inv( lmat(:blksz,:blksz) )*U(:m,:blksz)' )*c(:m,:n) .
!
!                   COMPUTE THE blksz BY blksz UNIT UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!                   USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
                    do l = 2_i4b, blksz
!
                        i1 = i + l
                        k1 = il(i1)
!
                        do l2 = 1_i4b, l - 1_i4b
!
                            i2 = i + l2
                            k2 = il(i2)
!
                            tau = matp(k2+i2-1_i4b)
!
                            lmat(l2,l) = -tau*dot_product( matp(k1:k1+i2-2_i4b), matp(k2:k2+i2-2_i4b) )
!
                        end do
!                    
                    end do
!
!$OMP END DO
!
                    do j = j1, j2
!
!                       COMPUTE U(:m,:blksz)'*c(:m,j) .
!
                        tmp(1_i4b:i+blksz-1_i4b) = c(1_i4b:i+blksz-1_i4b,j)
!
                        do l = 1_i4b, blksz
!
                            i1 = i + l
                            k1 = il(i1)
!
                            tau = matp(k1+i1-1_i4b)
!
                            x(l) = tau*dot_product( matp(k1:k1+i1-2_i4b), tmp(1_i4b:i1-1_i4b) )
!                    
                        end do
!
!                       SOLVE lmat(:blksz,:blksz)*x(:blksz) = U(:m,:blksz)'*c(:m,j)  AND
!
!                       COMPUTE c(:m,j) = c(:m,j) + U(:m,:blksz)*x(:blksz) .
!
                        do l = blksz, 1_i4b, -1_i4b
!
                            x(1_i4b:l-1_i4b) = x(1_i4b:l-1_i4b) - x(l)*lmat(1_i4b:l-1_i4b,l)
!
                            i1 = i + l
                            k1 = il(i1)
!
#ifdef _BLAS
                            call axpy( i1-1_i4b, x(l), matp(k1:k1+i1-2_i4b), 1_i4b, tmp(1_i4b:i1-1_i4b), 1_i4b )
#else
                            tmp(1_i4b:i1-1_i4b) = tmp(1_i4b:i1-1_i4b) + x(l)*matp(k1:k1+i1-2_i4b)
#endif
!
                        end do
!
                        c(1_i4b:i+blksz-1_i4b,j) = tmp(1_i4b:i+blksz-1_i4b)
!                    
                    end do
!
!$OMP BARRIER
!                    
                end do
!
!               APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
                i1 = i + blksz
!
                do i = i1, 2_i4b, -1_i4b
!
                    k1 = il(i)
!
                    tau = matp(k1+i-1_i4b)
!
                    if ( tau==zero ) cycle
!
!                   APPLY H(i) TO c(:m,:n) FROM THE LEFT.
!
                    tmp(1_i4b:i-1_i4b) = matp(k1:k1+i-2_i4b)
!
                    do j = j1, j2
!
                        prod = dot_product( tmp(1_i4b:i-1_i4b), c(1_i4b:i-1_i4b,j) )*tau
#ifdef _BLAS
                        call axpy( i-1_i4b, prod, tmp(1_i4b:i-1_i4b), 1_i4b, c(1_i4b:i-1_i4b,j), 1_i4b )
#else
                        c(1_i4b:i-1_i4b,j) = c(1_i4b:i-1_i4b,j) + prod*tmp(1_i4b:i-1_i4b)                
#endif
!
                    end do            
!
                end do
!
            else
!
!               COMPUTE H(nq-1)* . . .*H(2)*H(1)*c(:m,:n) .
!
                do i = 1_i4b, nq - blksz, blksz
!
!                   APPLY H(i+blksz-1)* . . . *H(i+1)*H(i) TO c(:m,:n) FROM THE LEFT AS
!
!                 ( I + U(:m,:blksz)*inv( lmat(:blksz,:blksz) )*U(:m,:blksz)' )*c(:m,:n) .
!
!                   COMPUTE THE blksz BY blksz UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!                   USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
                    do l = 1_i4b, blksz - 1_i4b
!
                        i1 = i + l
                        k1 = il(i1)
!
                        do l2 = l + 1_i4b, blksz
!
                            i2 = i + l2
                            k2 = il(i2)
!
                            tau = matp(k2+i2-1_i4b)
!
                            lmat(l2,l) = -tau*dot_product( matp(k1:k1+i1-2_i4b), matp(k2:k2+i1-2_i4b) )
!
                        end do
!                    
                    end do
!
!$OMP END DO
!
                    do j = j1, j2
!
!                       COMPUTE U(:m,:blksz)'*c(:m,j) .
!
                        tmp(1_i4b:i+blksz-1_i4b) = c(1_i4b:i+blksz-1_i4b,j)
!
                        do l = 1_i4b, blksz
!
                            i1 = i + l
                            k1 = il(i1)
!
                            tau = matp(k1+i1-1_i4b)
!
                            x(l) = tau*dot_product( matp(k1:k1+i1-2_i4b), tmp(1_i4b:i1-1_i4b) )
!                    
                        end do
!
!                       SOLVE lmat(:blksz,:blksz)*x(:blksz) = U(:m,:blksz)'*c(:m,j)  AND
!
!                       COMPUTE c(:m,j) = c(:m,j) + U(:m,:blksz)*x(:blksz) .
!
                        do l = 1_i4b, blksz
!
                            x(l+1_i4b:blksz) = x(l+1_i4b:blksz) - x(l)*lmat(l+1_i4b:blksz,l)
!
                            i1 = i + l
                            k1 = il(i1)
!
#ifdef _BLAS
                            call axpy( i1-1_i4b, x(l), matp(k1:k1+i1-2_i4b), 1_i4b, tmp(1_i4b:i1-1_i4b), 1_i4b )
#else
                            tmp(1_i4b:i1-1_i4b) = tmp(1_i4b:i1-1_i4b) + x(l)*matp(k1:k1+i1-2_i4b)
#endif
!
                        end do
!
                        c(1_i4b:i+blksz-1_i4b,j) = tmp(1_i4b:i+blksz-1_i4b)
!                    
                    end do
!
!$OMP BARRIER
!
                end do
!
!               APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
                i1 = i + 1_i4b
!
                do i = i1, nq
!
                    k1 = il(i)
!
                    tau = matp(k1+i-1_i4b)
!
                    if ( tau==zero ) cycle
!
!                   APPLY H(i) TO c(:m,:n) FROM THE LEFT.
!
                    tmp(1_i4b:i-1_i4b) = matp(k1:k1+i-2_i4b)
!
                    do j = j1, j2
!
                        prod = dot_product( tmp(1_i4b:i-1_i4b), c(1_i4b:i-1_i4b,j) )*tau
#ifdef _BLAS
                        call axpy( i-1_i4b, prod, tmp(1_i4b:i-1_i4b), 1_i4b, c(1_i4b:i-1_i4b,j), 1_i4b )
#else
                        c(1_i4b:i-1_i4b,j) = c(1_i4b:i-1_i4b,j) + prod*tmp(1_i4b:i-1_i4b)              
#endif
!
                    end do            
!
                end do
!
            end if
!
        else
!
            if ( descending ) then
!
!               COMPUTE c(:m,:n)*H(nq-1)* . . .*H(2)*H(1) .
!
                do i = nq - blksz, 1_i4b, -blksz
!
!                   APPLY H(i+blksz-1)* . . .*H(i+1)*H(i) TO c(:m,:n) FROM THE RIGHT AS
!
!                 c(:m,:n)*( I + U(:n,:blksz)*inv( lmat(:blksz,:blksz) )*U(:n,:blksz)' ) .
!
!                   COMPUTE THE blksz BY blksz UNIT UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!                   USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
                    do l = 2_i4b, blksz
!
                        i1 = i + l
                        k1 = il(i1)
!
                        do l2 = 1_i4b, l - 1_i4b
!
                            i2 = i + l2
                            k2 = il(i2)
!
                            tau = matp(k2+i2-1_i4b)
!
                            lmat(l2,l) = -tau*dot_product( matp(k1:k1+i2-2_i4b), matp(k2:k2+i2-2_i4b) )
!
                        end do
!                    
                    end do
!
!$OMP END DO
!
                    do j = j1, j2
!
!                       COMPUTE c(j,:n)*U(:n,:blksz) .
!
                        tmp(1_i4b:i+blksz-1_i4b) = c(j,1_i4b:i+blksz-1_i4b)
!
                        do l = 1_i4b, blksz
!
                            i1 = i + l
                            k1 = il(i1)
!
                            tau = matp(k1+i1-1_i4b)
!
                            x(l) = tau*dot_product( matp(k1:k1+i1-2_i4b), tmp(1_i4b:i1-1_i4b) )
!                    
                        end do
!
!                       SOLVE lmat(:blksz,:blksz)*x(:blksz) = c(j,:n)*U(:n,:blksz) AND
!
!                       COMPUTE c(j,:n) = c(j,:n) + x(:blksz)*U(:n,:blksz)' .
!
                        do l = blksz, 1_i4b, -1_i4b
!
                            x(1_i4b:l-1_i4b) = x(1_i4b:l-1_i4b) - x(l)*lmat(1_i4b:l-1_i4b,l)
!
                            i1 = i + l
                            k1 = il(i1)
!
#ifdef _BLAS
                            call axpy( i1-1_i4b, x(l), matp(k1:k1+i1-2_i4b), 1_i4b, tmp(1_i4b:i1-1_i4b), 1_i4b )
#else
                            tmp(1_i4b:i1-1_i4b) = tmp(1_i4b:i1-1_i4b) + x(l)*matp(k1:k1+i1-2_i4b)
#endif
!
                        end do
!
                        c(j,1_i4b:i+blksz-1_i4b) = tmp(1_i4b:i+blksz-1_i4b)
!                    
                    end do
!
!$OMP BARRIER
!                    
                end do
!
!               APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
                i1 = i + blksz
!
                do i = i1, 2_i4b, -1_i4b
!
                    k1 = il(i)
!
                    tau = matp(k1+i-1_i4b)
!
                    if ( tau==zero ) cycle
!
!                   APPLY H(i) TO c(:m,:n) FROM THE RIGHT.
!
                    tmp(1_i4b:i-1_i4b) = matp(k1:k1+i-2_i4b)
!
                    do j = j1, j2
!
                        prod = dot_product( tmp(1_i4b:i-1_i4b), c(j,1_i4b:i-1_i4b) )*tau
!#ifdef _BLAS
!                        call axpy( i-1_i4b, prod, tmp(1_i4b:i-1_i4b), 1_i4b, c(j,1_i4b:i-1_i4b), 1_i4b )
!#else
                        c(j,1_i4b:i-1_i4b) = c(j,1_i4b:i-1_i4b) + prod*tmp(1_i4b:i-1_i4b)                
!#endif
!
                    end do            
!
                end do
!
            else
!
!               COMPUTE c(:m,:n)*H(1)*H(2)* . . .*H(nq-1) .
!
                do i = 1_i4b, nq - blksz, blksz
!
!                   APPLY H(i)*H(i+1)* . . .*H(i+blksz-1) TO c(:m,:n) FROM THE RIGHT AS
!
!                 c(:m,:n)*( I + U(:n,:blksz)*inv( lmat(:blksz,:blksz) )*U(:n,:blksz)' ) .
!
!                   COMPUTE THE blksz BY blksz UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!                   USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
                    do l = 1_i4b, blksz - 1_i4b
!
                        i1 = i + l
                        k1 = il(i1)
!
                        do l2 = l + 1_i4b, blksz
!
                            i2 = i + l2
                            k2 = il(i2)
!
                            tau = matp(k2+i2-1_i4b)
!
                            lmat(l2,l) = -tau*dot_product( matp(k1:k1+i1-2_i4b), matp(k2:k2+i1-2_i4b) )
!
                        end do
!                    
                    end do
!
!$OMP END DO
!
                    do j = j1, j2
!
!                       COMPUTE c(j,:n)*U(:n,:blksz) .
!
                        tmp(1_i4b:i+blksz-1_i4b) = c(j,1_i4b:i+blksz-1_i4b)
!
                        do l = 1_i4b, blksz
!
                            i1 = i + l
                            k1 = il(i1)
!
                            tau = matp(k1+i1-1_i4b)
!
                            x(l) = tau*dot_product( matp(k1:k1+i1-2_i4b), tmp(1_i4b:i1-1_i4b) )
!                    
                        end do
!
!                       SOLVE lmat(:blksz,:blksz)*x(:blksz) = c(j,:n)*U(:n,:blksz) AND
!
!                       COMPUTE c(j,:n) = c(j,:n) + x(:blksz)*U(:n,:blksz)' .
!
                        do l = 1_i4b, blksz
!
                            x(l+1_i4b:blksz) = x(l+1_i4b:blksz) - x(l)*lmat(l+1_i4b:blksz,l)
!
                            i1 = i + l
                            k1 = il(i1)
!
#ifdef _BLAS
                            call axpy( i1-1_i4b, x(l), matp(k1:k1+i1-2_i4b), 1_i4b, tmp(1_i4b:i1-1_i4b), 1_i4b )
#else
                            tmp(1_i4b:i1-1_i4b) = tmp(1_i4b:i1-1_i4b) + x(l)*matp(k1:k1+i1-2_i4b)
#endif
!
                        end do
!
                        c(j,1_i4b:i+blksz-1_i4b) = tmp(1_i4b:i+blksz-1_i4b)
!                    
                    end do
!
!$OMP BARRIER
!
                end do
!
!               APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
                i1 = i + 1_i4b
!
                do i = i1, nq
!
                    k1 = il(i)
!
                    tau = matp(k1+i-1_i4b)
!
                    if ( tau==zero ) cycle
!
!                   APPLY H(i) TO c(:m,:n) FROM THE RIGHT.
!
                    tmp(1_i4b:i-1_i4b) = matp(k1:k1+i-2_i4b)
!
                    do j = j1, j2
!
                        prod = dot_product( tmp(1_i4b:i-1_i4b), c(j,1_i4b:i-1_i4b) )*tau
!#ifdef _BLAS
!                        call axpy( i-1_i4b, prod, tmp(1_i4b:i-1_i4b), 1_i4b, c(j,1_i4b:i-1_i4b), 1_i4b )
!#else
                        c(j,1_i4b:i-1_i4b) = c(j,1_i4b:i-1_i4b) + prod*tmp(1_i4b:i-1_i4b)           
!#endif
!
                    end do            
!
                end do
!
            end if
!
        end if
!
    else
!
!       Q WAS DETERMINED BY A CALL TO symtrid_cmp_c WITH upper = false .
!
        descending = .not.( ( trans .and. left ) .or. ( .not. trans .and. .not. left ) )
!
        if ( left ) then
!
            if ( descending ) then
!
!               COMPUTE H(1)*H(2)* . . .*H(nq-1)*c(:m,:n) .
!
                do i = nq - blksz - 1_i4b, 0_i4b, -blksz
!
!                   APPLY H(i+1)*H(i+2)* . . . *H(i+blksz) TO c(:m,:n) FROM THE LEFT AS
!
!                  ( I + U(:m,:blksz)*inv( lmat(:blksz,:blksz) )*U(:m,:blksz)' )*c(:m,:n) .
!
!                   COMPUTE THE blksz BY blksz UNIT UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!                   USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
                    do l = 2_i4b, blksz
!
                        i1 = i + l
                        k1 = il(i1)
!
                        do l2 = 1_i4b, l - 1_i4b
!
                            i2 = i + l2
                            k2 = il(i2)
!
                            tau = matp(k2)
!
                            lmat(l2,l) = -tau*dot_product( matp(k1+1_i4b:k1+m-i1), matp(k2+i1-i2+1_i4b:k2+m-i2) )
!
                        end do
!                    
                    end do
!
!$OMP END DO
!
                    do j = j1, j2
!
!                       COMPUTE U(:m,:blksz)'*c(:m,j) .
!
                        tmp(i+2_i4b:m) = c(i+2_i4b:m,j)
!
                        do l = 1_i4b, blksz
!
                            i1 = i + l
                            k1 = il(i1)
!
                            tau = matp(k1)
!
                            x(l) = tau*dot_product( matp(k1+1_i4b:k1+m-i1), tmp(i1+1_i4b:m) )
!                    
                        end do
!
!                       SOLVE lmat(:blksz,:blksz)*x(:blksz) = U(:m,:blksz)'*c(:m,j)  AND
!
!                       COMPUTE c(:m,j) = c(:m,j) + U(:m,:blksz)*x(:blksz) .
!
                        do l = blksz, 1_i4b, -1_i4b
!
                            x(1_i4b:l-1_i4b) = x(1_i4b:l-1_i4b) - x(l)*lmat(1_i4b:l-1_i4b,l)
!
                            i1 = i + l
                            k1 = il(i1)
!
#ifdef _BLAS
                            call axpy( m-i1, x(l), matp(k1+1_i4b:k1+m-i1), 1_i4b, tmp(i1+1_i4b:m), 1_i4b )
#else
                            tmp(i1+1_i4b:m) = tmp(i1+1_i4b:m) + x(l)*matp(k1+1_i4b:k1+m-i1)
#endif
!
                        end do
!
                        c(i+2_i4b:m,j) = tmp(i+2_i4b:m)
!                    
                    end do
!
!$OMP BARRIER
!
                end do
!
!               APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
                i1 = i + blksz
!
                do i = i1, 1_i4b, -1_i4b
!
                    k1 = il(i)
!
                    tau = matp(k1)
!
                    if ( tau==zero ) cycle
!
!                   APPLY H(i) TO c(:m,:n) FROM THE LEFT.
!
                    tmp(i+1_i4b:m) = matp(k1+1_i4b:k1+m-i)
!
                    do j = j1, j2
!
                        prod = dot_product( tmp(i+1_i4b:m), c(i+1_i4b:m,j) )*tau
#ifdef _BLAS
                        call axpy( m-i, prod, tmp(i+1_i4b:m), 1_i4b, c(i+1_i4b:m,j), 1_i4b )
#else
                        c(i+1_i4b:m,j) = c(i+1_i4b:m,j) + prod*tmp(i+1_i4b:m)
#endif
!
                    end do            
!
                end do
!
            else
!
!               COMPUTE H(nq-1)* . . .*H(2)*H(1)*c(:m,:n) .
!
                do i = 0_i4b, nq - blksz - 1_i4b, blksz
!
!                   APPLY H(i+blksz)* . . .*H(i+2)*H(i+1) FROM THE LEFT TO c(:m,:n) AS
!
!                 ( I + U(:m,:blksz)*inv( lmat(:blksz,:blksz) )*U(:m,:blksz)' )*c(:m,:n) .
!
!                   COMPUTE THE blksz BY blksz UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!                   USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
                    do l = 1_i4b, blksz - 1_i4b
!
                        i1 = i + l
                        k1 = il(i1)
!
                        do l2 = l + 1_i4b, blksz
!
                            i2 = i + l2
                            k2 = il(i2)
!
                            tau = matp(k2)
!
                            lmat(l2,l) = -tau*dot_product( matp(k1+i2-i1+1_i4b:k1+m-i1), matp(k2+1_i4b:k2+m-i2) )
!
                        end do
!                    
                    end do
!
!$OMP END DO
!
                    do j = j1, j2
!
!                       COMPUTE U(:m,:blksz)'*c(:m,j) .
!
                        tmp(i+2_i4b:m) = c(i+2_i4b:m,j)
!
                        do l = 1_i4b, blksz
!
                            i1 = i + l
                            k1 = il(i1)
!
                            tau = matp(k1)
!
                            x(l) = tau*dot_product( matp(k1+1_i4b:k1+m-i1), tmp(i1+1_i4b:m) )
!                    
                        end do
!
!                       SOLVE lmat(:blksz,:blksz)*x(:blksz) = U(:m,:blksz)'*c(:m,j)  AND
!
!                       COMPUTE c(:m,j) = c(:m,j) + U(:m,:blksz)*x(:blksz) .
!
                        do l = 1_i4b, blksz
!
                            x(l+1_i4b:blksz) = x(l+1_i4b:blksz) - x(l)*lmat(l+1_i4b:blksz,l)
!
                            i1 = i + l
                            k1 = il(i1)
!
#ifdef _BLAS
                            call axpy( m-i1, x(l), matp(k1+1_i4b:k1+m-i1), 1_i4b, tmp(i1+1_i4b:m), 1_i4b )
#else
                            tmp(i1+1_i4b:m) = tmp(i1+1_i4b:m) + x(l)*matp(k1+1_i4b:k1+m-i1)
#endif
!
                        end do
!
                        c(i+2_i4b:m,j) = tmp(i+2_i4b:m)
!                    
                    end do
!
!$OMP BARRIER
!
                end do
!
!               APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
                i1 = i + 1_i4b
!
                do i = i1, nq - 1_i4b
!
                    k1 = il(i)
!
                    tau = matp(k1)
!
                    if ( tau==zero ) cycle
!
!                   APPLY H(i) TO c(:m,:n) FROM THE LEFT.
!
                    tmp(i+1_i4b:m) = matp(k1+1_i4b:k1+m-i)
!
                    do j = j1, j2
!
                        prod = dot_product( tmp(i+1_i4b:m), c(i+1_i4b:m,j) )*tau
#ifdef _BLAS
                        call axpy( m-i, prod, tmp(i+1_i4b:m), 1_i4b, c(i+1_i4b:m,j), 1_i4b )
#else
                        c(i+1_i4b:m,j) = c(i+1_i4b:m,j) + prod*tmp(i+1_i4b:m)              
#endif
!
                    end do            
!
                end do
!
            end if
!
        else
!
            if ( descending ) then
!
!               COMPUTE c(:m,:n)*H(nq-1)* . . .*H(2)*H(1) .
!
                do i = nq - blksz - 1_i4b, 0_i4b, -blksz
!
!                   APPLY H(i+blksz)* . . .*H(i+2)*H(i+1) TO c(:m,:n) FROM THE RIGHT AS
!
!                  c(:m,:n)*( I + U(:n,:blksz)*inv( lmat(:blksz,:blksz) )*U(:n,:blksz)' ) .
!
!                   COMPUTE THE blksz BY blksz UNIT UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!                   USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
                    do l = 2_i4b, blksz
!
                        i1 = i + l
                        k1 = il(i1)
!
                        do l2 = 1_i4b, l - 1_i4b
!
                            i2 = i + l2
                            k2 = il(i2)
!
                            tau = matp(k2)
!
                            lmat(l2,l) = -tau*dot_product( matp(k1+1_i4b:k1+n-i1), matp(k2+i1-i2+1_i4b:k2+n-i2) )
!
                        end do
!                    
                    end do
!
!$OMP END DO
!
                    do j = j1, j2
!
!                       COMPUTE c(j,:n)*U(:n,:blksz) .
!
                        tmp(i+2_i4b:n) = c(j,i+2_i4b:n)
!
                        do l = 1_i4b, blksz
!
                            i1 = i + l
                            k1 = il(i1)
!
                            tau = matp(k1)
!
                            x(l) = tau*dot_product( matp(k1+1_i4b:k1+n-i1), tmp(i1+1_i4b:n) )
!                    
                        end do
!
!                       SOLVE x(:blksz)*lmat(:blksz,:blksz) = c(j,:n)*U(:n,:blksz) AND
!
!                       COMPUTE c(j,:n) = c(j,:n) + x(:blksz)*U(:n,:blksz)' .
!
                        do l = blksz, 1_i4b, -1_i4b
!
                            x(1_i4b:l-1_i4b) = x(1_i4b:l-1_i4b) - x(l)*lmat(1_i4b:l-1_i4b,l)
!
                            i1 = i + l
                            k1 = il(i1)
!
#ifdef _BLAS
                            call axpy( n-i1, x(l), matp(k1+1_i4b:k1+n-i1), 1_i4b, tmp(i1+1_i4b:n), 1_i4b )
#else
                            tmp(i1+1_i4b:n) = tmp(i1+1_i4b:n) + x(l)*matp(k1+1_i4b:k1+n-i1)
#endif
!
                        end do
!
                        c(j,i+2_i4b:n) = tmp(i+2_i4b:n)
!                    
                    end do
!
!$OMP BARRIER
!
                end do
!
!               APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
                i1 = i + blksz
!
                do i = i1, 1_i4b, -1_i4b
!
                    k1 = il(i)
!
                    tau = matp(k1)
!
                    if ( tau==zero ) cycle
!
!                   APPLY H(i) TO c(:m,:n) FROM THE RIGHT.
!
                    tmp(i+1_i4b:n) = matp(k1+1_i4b:k1+n-i)
!
                    do j = j1, j2
!
                        prod = dot_product( tmp(i+1_i4b:n), c(j,i+1_i4b:n) )*tau
!#ifdef _BLAS
!                        call axpy( n-i, prod, tmp(i+1_i4b:n), 1_i4b, c(j,i+1_i4b:n), 1_i4b )
!#else
                        c(j,i+1_i4b:n) = c(j,i+1_i4b:n) + prod*tmp(i+1_i4b:n)
!#endif
!
                    end do            
!
                end do
!
            else
!
!               COMPUTE c(:m,:n)*H(1)*H(2)* . . .*H(nq-1) .
!
                do i = 0_i4b, nq - blksz - 1_i4b, blksz
!
!                   APPLY H(i+1)*H(i+2)* . . .*H(i+blksz) TO c(:m,:n) FROM THE RIGHT AS
!
!                 c(:m,:n)*( I + U(:n,:blksz)*inv( lmat(:blksz,:blksz) )*U(:n,:blksz)' ) .
!
!                   COMPUTE THE blksz BY blksz UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!                   USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
                    do l = 1_i4b, blksz - 1_i4b
!
                        i1 = i + l
                        k1 = il(i1)
!
                        do l2 = l + 1_i4b, blksz
!
                            i2 = i + l2
                            k2 = il(i2)
!
                            tau = matp(k2)
!
                            lmat(l2,l) = -tau*dot_product( matp(k1+i2-i1+1_i4b:k1+n-i1), matp(k2+1_i4b:k2+n-i2) )
!
                        end do
!                    
                    end do
!
!$OMP END DO
!
                    do j = j1, j2
!
!                       COMPUTE c(j,:n)*U(:n,:blksz) .
!
                        tmp(i+2_i4b:n) = c(j,i+2_i4b:n)
!
                        do l = 1_i4b, blksz
!
                            i1 = i + l
                            k1 = il(i1)
!
                            tau = matp(k1)
!
                            x(l) = tau*dot_product( matp(k1+1_i4b:k1+n-i1), tmp(i1+1_i4b:n) )
!                    
                        end do
!
!                       SOLVE x(:blksz)*lmat(:blksz,:blksz) = c(j,:n)*U(:n,:blksz)  AND
!
!                       COMPUTE c(j,:n) = c(j,:n) + x(:blksz)*U(:n,:blksz)' .
!
                        do l = 1_i4b, blksz
!
                            x(l+1_i4b:blksz) = x(l+1_i4b:blksz) - x(l)*lmat(l+1_i4b:blksz,l)
!
                            i1 = i + l
                            k1 = il(i1)
!
#ifdef _BLAS
                            call axpy( n-i1, x(l), matp(k1+1_i4b:k1+n-i1), 1_i4b, tmp(i1+1_i4b:n), 1_i4b )
#else
                            tmp(i1+1_i4b:n) = tmp(i1+1_i4b:n) + x(l)*matp(k1+1_i4b:k1+n-i1)
#endif
!
                        end do
!
                        c(j,i+2_i4b:n) = tmp(i+2_i4b:n)
!                    
                    end do
!
!$OMP BARRIER
!
                end do
!
!               APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
                i1 = i + 1_i4b
!
                do i = i1, nq - 1_i4b
!
                    k1 = il(i)
!
                    tau = matp(k1)
!
                    if ( tau==zero ) cycle
!
!                   APPLY H(i) TO c(:m,:n) FROM THE RIGHT.
!
                    tmp(i+1_i4b:n) = matp(k1+1_i4b:k1+n-i)
!
                    do j = j1, j2
!
                        prod = dot_product( tmp(i+1_i4b:n), c(j,i+1_i4b:n) )*tau
!#ifdef _BLAS
!                        call axpy( n-i, prod, tmp(i+1_i4b:n), 1_i4b, c(j,i+1_i4b:n), 1_i4b )
!#else
                        c(j,i+1_i4b:n) = c(j,i+1_i4b:n) + prod*tmp(i+1_i4b:n)              
!#endif
!
                    end do            
!
                end do
!
            end if
!
        end if
!
    end if
!
!$OMP END PARALLEL
!
!
! END OF SUBROUTINE apply_q_symtrid_c
! ___________________________________
!
    end subroutine apply_q_symtrid_c
!
! =========================================================================================
!
    subroutine apply_q_symtrid_d( matp, c, left, trans )
!
! Purpose
! _______
!
!   APPLY_Q_SYMTRID overwrites the general real m-by-n matrix C with
!
!            - Q * C  if LEFT = true  and TRANS = false, or
!
!            - Q'* C  if LEFT = true  and TRANS = true,  or
!
!            - C * Q  if LEFT = false and TRANS = false, or
!
!            - C * Q' if LEFT = false and TRANS = true,
!
!   where Q is a real orthogonal matrix of order nq and is defined as the product
!   of nq-1 elementary reflectors
!
!           - Q = H(nq-1) * . . . * H(2) * H(1)
!
!   as returned by SYMTRID_CMP (with UPPER = true or without this argument)
!   and with STORE_Q = true.
!
!   Q is of order m (=nq) and is the product of m-1 reflectors if LEFT = true ;
!   Q is of order n (=nq) and is the product of n-1 reflectors if LEFT = false.
!
!
! Arguments
! _________
!
!   MATP         (INPUT) real(stnd), dimension(:)
!                On entry, the vectors and the scalars, which define
!                the elementary reflectors, as returned by SYMTRID_CMP
!                in argument MATP. MATP is not modified by the routine.
!
!                The size of MATP must verify size( MATP ) = (nq * (nq+1)/2)
!
!   C            (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the m by n matrix C.
!
!                On exit, C is overwritten by Q * C or Q' * C or C * Q' or C * Q.
!
!                The shape of C must verify:
!
!                - if LEFT = true,  size( C, 1 ) = nq ;
!                - if LEFT = false, size( C, 2 ) = nq .
!
!   LEFT         (INPUT) logical(lgl)
!                If:
!
!                - LEFT = true  : apply Q or Q' from the left  ;
!                - LEFT = false : apply Q or Q' from the right .
!
!   TRANS        (INPUT) logical(lgl)
!                If:
!
!                - TRANS = false : apply Q  (no transpose) ;
!                - TRANS = true  : apply Q' (transpose)    .
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the routine  DORMTR in LAPACK and uses a blocked
!   algorithm to apply the Householder transformations (e.g. the elementary reflectors)
!   stored in packed form in the vector MATP.
!
!   Furthermore, the subroutine is parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989:
!           Block reduction of matrices to condensed form for eigenvalue computations. 
!           J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!   (3) Walker, H.F., 1988:
!           Implementation of the GMRES method using Householder transformations.
!           Siam J. Sci. Stat. Comput., Vol. 9, No 1, pp. 152-163.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : machhuge
    use Char_Constants,  only : eig_error1, allocate_error
    use Utilities,       only : assert, merror, arth, cumsum
#ifdef _BLAS
    use BLAS_interfaces, only : gemv, ger, trsm, gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in)   , dimension(:)   :: matp
    real(stnd),   intent(inout), dimension(:,:) :: c
!
    logical(lgl), intent(in) :: left, trans
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                        :: tau, prod
    real(stnd), dimension( blksz )                    :: beta
    real(stnd), dimension( max(size(c,1),size(c,2)) ) :: tmp
    real(stnd), dimension( blksz, blksz )             :: lmat
    real(stnd), dimension(:,:), allocatable           :: umat, c2, x
!
    integer(i4b)                                        :: m, n, nm, nq, i, i1, i2, j, j1, j2, j3, &
                                                           k, k1, k2, l, l2, p, p2, q, nblk, rem
    integer(i4b), dimension( max(size(c,1),size(c,2)) ) :: il
    integer                                             :: iok
!
    logical(lgl) :: descending
!
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='apply_q_symtrid'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m  = size( c, 1 )
    n  = size( c, 2 )
    nm = n*m
!
    if ( left ) then
        nq = m
        p  = n
    else
        nq = n
        p  = m
    end if
!
    k = nq*(nq+1_i4b)/2_i4b
!
    call assert( logical(size(matp)==k,lgl)  ,   &
                 name_proc )
!
    tau = matp(1_i4b)
!
    if ( tau/=machhuge ) then
        call merror( name_proc//eig_error1 )
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( nq<=1_i4b .or. nm<=0_i4b ) return
!
!   DETERMINE THE ORDER FOR APPLYING THE HOUSEHOLDER TRANSFORMATIONS.
!
    descending = ( trans .and. left ) .or. ( .not. trans .and. .not. left )
!
!   COMPUTE THE NUMBER OF COLUMN BLOCKS FOR THE INPUT MATRIX.
!
    nblk = p/blksz
    rem  = mod( p, blksz )
!
    if ( rem/=0_i4b ) then
!
        nblk = nblk + 1_i4b
        p2   = p - rem
!
    else
!
        p2   = p - blksz
!
    end if
!
!   INITIALIZE THE UNIT LOWER OR UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!   USED IN THE BLOCKED ALGORITHM.
!
    lmat(1_i4b:blksz,1_i4b:blksz) = zero
!
    do i = 1_i4b, blksz
        lmat(i,i) = one
    end do
!
!   ALLOCATE AND INITIALIZE (SHARED) WORK ARRAY FOR STORING HOUSEHOLDER
!   VECTORS .
!
    allocate( umat(nq,blksz), stat = iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    umat(:nq,:blksz) = zero
!
!   COMPUTE INDICES OF THE FIRST ELEMENT OF EACH COLUMN OF THE ORIGINAL MATRIX.
!
    il(:nq) = cumsum( arth(0_i4b,1_i4b,nq), seed=1_i4b )
!
#ifdef _OPENMP
    i = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  nm>=omp_limit                       .and.      &
                  nblk>1_i4b                          .and.      &
                  i>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)                                                &
!$OMP         ,PRIVATE(iok,i,i1,i2,k1,k2,l,l2,j,j1,j2,j3,q,tau,prod,tmp,x,c2)
!
!   Q WAS DETERMINED BY A CALL TO symtrid_cmp_d OR symtrid_cmp_c WITH upper = true.
!
    if ( left ) then
#ifdef _BLAS
!
!       ALLOCATE (PRIVATE) WORK ARRAY FOR A FAST BLOCKED ALGORITHM.
!
        allocate( x(blksz,blksz), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
#endif
!
        if ( descending ) then
!
!           COMPUTE H(1)*H(2)* . . .*H(nq-1)*c(:m,:n) .
!
            do i = nq - blksz, 1_i4b, -blksz
!
!               APPLY H(i)*H(i+1)* . . . *H(i+blksz-1) TO c(:m,:n) FROM THE LEFT AS
!
!               ( I + U(:m,:blksz)*inv( lmat(:blksz,:blksz) )*U(:m,:blksz)' )*c(:m,:n) .
!
                q = i + blksz - 1_i4b
!
!               COMPUTE THE blksz BY blksz UNIT UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!               USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do l = 1_i4b, blksz
!
                    i1 = i + l
                    k1 = il(i1)
!
!                   SAVE HOUSEHOLDER VECTOR mat(:i1-1,i1) .
!
                    umat(1_i4b:i1-1_i4b,l) = matp(k1:k1+i1-2_i4b)
                    umat(i1:q,l)           = zero
!
                    beta(l) = matp(k1+i1-1_i4b)
!
                    do l2 = 1_i4b, l - 1_i4b
!
                        i2 = i + l2
                        k2 = il(i2)
!
                        tau = matp(k2+i2-1_i4b)
!
                        lmat(l2,l) = -tau*dot_product( matp(k1:k1+i2-2_i4b), matp(k2:k2+i2-2_i4b) )
!
                    end do
!                    
                end do
!
!$OMP END DO
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do j = 0_i4b, p2, blksz
!
                    j1 = j + 1_i4b
                    j2 = min( j + blksz, p )
!
#ifdef _BLAS
                    j3 = j2 - j
!
!                   COMPUTE  U(:m,:blksz)'*c(:m,j1:j2) .
!
                    call gemm( 'T', 'N', blksz, j3, q, one, umat, m, c(1_i4b:m,j1:j2), m, zero, x, blksz  )
!
                    do l = 1_i4b, j3
!
                        x(:blksz,l) = beta(:blksz)*x(:blksz,l)
!                    
                    end do
!
!                   SOLVE lmat(:blksz,:blksz)*x(:blksz,:j3) = U(:m,:blksz)'*c(:m,j1:j2) .
!
                    call trsm( 'L', 'U', 'N', 'U', blksz, j3, one, lmat, blksz, x, blksz )
!
!                   COMPUTE c(:m,j1:j2) = c(:m,j1:j2) + U(:m,:blksz)*x(:blksz,:j3) .
!
                    call gemm( 'N', 'N', q, j3, blksz, one, umat, m, x, blksz, one, c(1_i4b:m,j1:j2), m   )
#else
                    do l = j1, j2
!
!                       COMPUTE U(:m,:blksz)'*c(:m,j1:j2) .
!
                        do l2 = 1_i4b, blksz
!
                            i1 = i + l2 - 1_i4b
!
                            tmp(l2) = beta(l2)*dot_product( umat(1_i4b:i1,l2), c(1_i4b:i1,l) )
!                    
                        end do
!
!                       SOLVE lmat(:blksz,:blksz)*x(:blksz,:j3) = U(:m,:blksz)'*c(:m,j1:j2) .
!                       COMPUTE c(:m,j1:j2) = c(:m,j1:j2) + U(:m,:blksz)*x(:blksz,:j3) .
!
                        do l2 = blksz, 1_i4b, -1_i4b
!
                            tmp(1_i4b:l2-1_i4b) = tmp(1_i4b:l2-1_i4b) - tmp(l2)*lmat(1_i4b:l2-1_i4b,l2)
!
                            i1 = i + l2 - 1_i4b
!
                            c(1_i4b:i1,l) = c(1_i4b:i1,l) + tmp(l2)*umat(1_i4b:i1,l2)
!
                        end do
!
                    end do
#endif
!
                end do
!
!$OMP END DO
!                    
            end do
!
!           APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
            i1 = i + blksz
!
            do i = i1, 2_i4b, -1_i4b
!
                k1 = il(i)
!
                tau = matp(k1+i-1_i4b)
!
                if ( tau==zero ) cycle
!
!               APPLY H(i) TO c(:m,:n) FROM THE LEFT.
!
                q = i - 1_i4b
!
                tmp(1_i4b:q) = matp(k1:k1+i-2_i4b)
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do j = 0_i4b, p2, blksz
!
                    j1 = j + 1_i4b
                    j2 = min( j + blksz, p )
!
#ifdef _BLAS
                    j3 = j2 - j
!
                    call gemv( 'T', q, j3, tau, c(1_i4b:m,j1:j2), m, tmp(1_i4b:q), 1_i4b, zero, x(1_i4b:j3,1_i4b), 1_i4b )
!
                    call ger( q, j3, one, tmp(1_i4b:q), 1_i4b, x(1_i4b:j3,1_i4b), 1_i4b, c(1_i4b:m,j1:j2), m )
#else
                    do l = j1, j2
!
                        prod = dot_product( tmp(1_i4b:q), c(1_i4b:q,l) )*tau
!
                        c(1_i4b:q,l) = c(1_i4b:q,l) + prod*tmp(1_i4b:q)                
!
                    end do            
#endif
!
                end do
!
!$OMP END DO
!
            end do
!
        else
!
!           COMPUTE H(nq-1)* . . .*H(2)*H(1)*c(:m,:n) .
!
            do i = 1_i4b, nq - blksz, blksz
!
!               APPLY H(i+blksz-1)* . . . *H(i+1)*H(i) TO c(:m,:n) FROM THE LEFT AS
!
!               ( I + U(:m,:blksz)*inv( lmat(:blksz,:blksz) )*U(:m,:blksz)' )*c(:m,:n) .
!
                q = i + blksz - 1_i4b
!
!               COMPUTE THE blksz BY blksz UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!               USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do l = 1_i4b, blksz
!
                    i1 = i + l
                    k1 = il(i1)
!
!                   SAVE HOUSEHOLDER VECTOR mat(:i1-1,i1) .
!
                    umat(1_i4b:i1-1_i4b,l) = matp(k1:k1+i1-2_i4b)
                    umat(i1:q,l)           = zero
!
                    beta(l) = matp(k1+i1-1_i4b)
!
                    do l2 = l + 1_i4b, blksz
!
                        i2 = i + l2
                        k2 = il(i2)
!
                        tau = matp(k2+i2-1_i4b)
!
                        lmat(l2,l) = -tau*dot_product( matp(k1:k1+i1-2_i4b), matp(k2:k2+i1-2_i4b) )
!
                    end do
!                    
                end do
!
!$OMP END DO
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do j = 0_i4b, p2, blksz
!
                    j1 = j + 1_i4b
                    j2 = min( j + blksz, p )
!
#ifdef _BLAS
                    j3 = j2 - j
!
!                   COMPUTE U(:m,:blksz)'*c(:m,j1:j2) .
!
                    call gemm( 'T', 'N', blksz, j3, q, one, umat, m, c(1_i4b:m,j1:j2), m, zero, x, blksz  )
!
                    do l = 1_i4b, j3
!
                        x(:blksz,l) = beta(:blksz)*x(:blksz,l)
!                    
                    end do
!
!                   SOLVE lmat(:blksz,:blksz)*x(:blksz,:j3) = U(:m,:blksz)'*c(:m,j1:j2) .
!
                    call trsm( 'L', 'L', 'N', 'U', blksz, j3, one, lmat, blksz, x, blksz )
!
!                   COMPUTE c(:m,j1:j2) = c(:m,j1:j2) + U(:m,:blksz)*x(:blksz,:j3) .
!
                    call gemm( 'N', 'N', q, j3, blksz, one, umat, m, x, blksz, one, c(1_i4b:m,j1:j2), m   )
#else
                    do l = j1, j2
!
!                       COMPUTE U(:m,:blksz)'*c(:m,j1:j2) .
!
                        do l2 = 1_i4b, blksz
!
                            i1 = i + l2 - 1_i4b
!
                            tmp(l2) = beta(l2)*dot_product( umat(1_i4b:i1,l2), c(1_i4b:i1,l) )
!                    
                        end do
!
!                       SOLVE lmat(:blksz,:blksz)*x(:blksz,:j3) = U(:m,:blksz)'*c(:m,j1:j2) .
!                       COMPUTE c(:m,j1:j2) = c(:m,j1:j2) + U(:m,:blksz)*x(:blksz,:j3) .
!
                        do l2 = 1_i4b, blksz
!
                            tmp(l2+1_i4b:blksz) = tmp(l2+1_i4b:blksz) - tmp(l2)*lmat(l2+1_i4b:blksz,l2)
!
                            i1 = i + l2 - 1_i4b
!
                            c(1_i4b:i1,l) = c(1_i4b:i1,l) + tmp(l2)*umat(1_i4b:i1,l2)
!
                        end do
!
                    end do
#endif
!
                end do
!
!$OMP END DO
!
            end do
!
!           APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
            i1 = i + 1_i4b
!
            do i = i1, nq
!
                k1 = il(i)
!
                tau = matp(k1+i-1_i4b)
!
                if ( tau==zero ) cycle
!
!               APPLY H(i) TO c(:m,:n) FROM THE LEFT.
!
                q = i - 1_i4b
!
                tmp(1_i4b:q) = matp(k1:k1+i-2_i4b)
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do j = 0_i4b, p2, blksz
!
                    j1 = j + 1_i4b
                    j2 = min( j + blksz, p )
!
#ifdef _BLAS
                    j3 = j2 - j
!
                    call gemv( 'T', q, j3, tau, c(1_i4b:m,j1:j2), m, tmp(1_i4b:q), 1_i4b, zero, x(1_i4b:j3,1_i4b), 1_i4b )
!
                    call ger( q, j3, one, tmp(1_i4b:q), 1_i4b, x(1_i4b:j3,1_i4b), 1_i4b, c(1_i4b:m,j1:j2), m )
#else
                    do l = j1, j2
!
                        prod = dot_product( tmp(1_i4b:q), c(1_i4b:q,l) )*tau
!
                        c(1_i4b:q,l) = c(1_i4b:q,l) + prod*tmp(1_i4b:q)              
!
                    end do            
#endif
!
                end do
!
!$OMP END DO
!
            end do
!
        end if
#ifdef _BLAS
!
!       DEALLOCATE (PRIVATE) WORK ARRAY.
!
        deallocate( x )
#endif
!
    else
!
!       ALLOCATE (PRIVATE) WORK ARRAYS FOR A FAST BLOCKED ALGORITHM.
!
        allocate( x(blksz,blksz), c2(blksz,n), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        if ( descending ) then
!
!           COMPUTE c(:m,:n)*H(nq-1)* . . .*H(2)*H(1) .
!
            do i = nq - blksz, 1_i4b, -blksz
!
!               APPLY H(i+blksz-1)* . . .*H(i+1)*H(i) TO c(:m,:n) FROM THE RIGHT AS
!
!             c(:m,:n)*( I + U(:n,:blksz)*inv( lmat(:blksz,:blksz) )*U(:n,:blksz)' ) .
!
                q = i + blksz - 1_i4b
!
!               COMPUTE THE blksz BY blksz UNIT UPPER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!               USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do l = 1_i4b, blksz
!
                    i1 = i + l
                    k1 = il(i1)
!
!                   SAVE SHIFTED HOUSEHOLDER VECTOR mat(:i1-1,i1) .
!
                    umat(1_i4b:i1-1_i4b,l) = matp(k1:k1+i1-2_i4b)
                    umat(i1:q,l)           = zero
!
                    beta(l) = matp(k1+i1-1_i4b)
!
                    do l2 = 1_i4b, l - 1_i4b
!
                        i2 = i + l2
                        k2 = il(i2)
!
                        tau = matp(k2+i2-1_i4b)
!
                        lmat(l2,l) = -tau*dot_product( matp(k1:k1+i2-2_i4b), matp(k2:k2+i2-2_i4b) )
!
                    end do
!                    
                end do
!
!$OMP END DO
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do j = 0_i4b, p2, blksz
!
                    j1 = j + 1_i4b
                    j2 = min( j + blksz, p )
                    j3 = j2 - j
!
                    c2(1_i4b:j3,1_i4b:q) = c(j1:j2,1_i4b:q)
!
#ifdef _BLAS
!                   COMPUTE c(j1:j2,:n)*U(:n,:blksz) .
!
                    call gemm( 'N', 'N', j3, blksz, q, one, c2, blksz, umat, n, zero, x, blksz  )
!
                    do l = 1_i4b, blksz
!
                        x(1_i4b:j3,l) = beta(l)*x(1_i4b:j3,l)
!                    
                    end do
!
!                   SOLVE x(j1:j2,:blksz)*lmat(:blksz,:blksz)' = c(j1:j2,:n)*U(:n,:blksz) .
!
                    call trsm( 'R', 'U', 'T', 'U', j3, blksz, one, lmat, blksz, x, blksz )
!
!                   COMPUTE c(j1:j2,:n) = c(j1:j2,:n) + x(j1:j2,:blksz)*U(:n,:blksz)' .
!
                    call gemm( 'N', 'T', j3, q,  blksz, one, x, blksz, umat, n, one, c2, blksz )
#else
!                   COMPUTE c(j1:j2,:n)*U(:n,:blksz) .
!
                    do l = 1_i4b, blksz
!
                        i1 = i + l - 1_i4b
!
                        x(1_i4b:j3,l) = umat(1_i4b,l)*c2(1_i4b:j3,1_i4b)
!
                        do l2 = 2_i4b, i1
!
                            x(1_i4b:j3,l) = x(1_i4b:j3,l) +  umat(l2,l)*c2(1_i4b:j3,l2)
!                    
                        end do
!
                        x(1_i4b:j3,l) = beta(l)*x(1_i4b:j3,l)
!                    
                    end do
!
!                   SOLVE x(j1:j2,:blksz)*lmat(:blksz,:blksz)' = c(j1:j2,:n)*U(:n,:blksz) AND
!                   COMPUTE c(j1:j2,i+1:i+blksz-1) = c(j1:j2,i+1:i+blksz-1) + x(j1:j2,:blksz)*U(i+1:i+blksz-1,:blksz)' .
!
                    do l = blksz, 1_i4b, -1_i4b
!
                        do l2 = 1_i4b, l-1_i4b
!
                            x(1_i4b:j3,l2) = x(1_i4b:j3,l2) - x(1_i4b:j3,l)*lmat(l2,l)
!
                        end do
!
                        do l2 = 1_i4b, l-1_i4b
!
                            i2 = i + l2
!
                            c2(1_i4b:j3,i2) = c2(1_i4b:j3,i2) + x(1_i4b:j3,l)*umat(i2,l)
!
                        end do
!
                    end do
!
!                   COMPUTE c(j1:j2,:i) = c(j1:j2,:i) + x(j1:j2,:blksz)*U(:i,:blksz)' .
!
                    do l = 1_i4b, i
!
                        do l2 = 1_i4b, blksz
!
                            c2(1_i4b:j3,l) = c2(1_i4b:j3,l) + x(1_i4b:j3,l2)*umat(l,l2)
!                    
                        end do
!                    
                    end do
#endif
!
                    c(j1:j2,1_i4b:q) = c2(1_i4b:j3,1_i4b:q)
!
                end do
!
!$OMP END DO
!                    
            end do
!
!           APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
            i1 = i + blksz
!
            do i = i1, 2_i4b, -1_i4b
!
                k1 = il(i)
!
                tau = matp(k1+i-1_i4b)
!
                if ( tau==zero ) cycle
!
!               APPLY H(i) TO c(:m,:n) FROM THE RIGHT.
!
                q = i - 1_i4b
!
                tmp(1_i4b:q) = matp(k1:k1+i-2_i4b)
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do j = 0_i4b, p2, blksz
!
                    j1 = j + 1_i4b
                    j2 = min( j + blksz, p )
                    j3 = j2 - j
!
                    c2(1_i4b:j3,1_i4b:q) = c(j1:j2,1_i4b:q)
!
#ifdef _BLAS
                    call gemv( 'N', j3, q, tau, c2, blksz, tmp(1_i4b:q), 1_i4b, zero, x(1_i4b:j3,1_i4b), 1_i4b )
!
                    call ger( j3, q, one, x(1_i4b:j3,1_i4b), 1_i4b, tmp(1_i4b:q), 1_i4b, c2, blksz )
#else
                    x(1_i4b:j3,1_i4b) = matmul( c2(1_i4b:j3,1_i4b:q), tmp(1_i4b:q) )

                    do l = 1_i4b, q
!
                        prod = tmp(l)*tau
!
                        c2(1_i4b:j3,l) = c2(1_i4b:j3,l) + prod*x(1_i4b:j3,1_i4b)     
!
                    end do            
#endif
!
                    c(j1:j2,1_i4b:q) = c2(1_i4b:j3,1_i4b:q)
!
                end do
!
!$OMP END DO
!
            end do
!
        else
!
!           COMPUTE c(:m,:n)*H(1)*H(2)* . . .*H(nq-1) .
!
            do i = 1_i4b, nq - blksz, blksz
!
!               APPLY H(i)*H(i+1)* . . .*H(i+blksz-1) TO c(:m,:n) FROM THE RIGHT AS
!
!             c(:m,:n)*( I + U(:n,:blksz)*inv( lmat(:blksz,:blksz) )*U(:n,:blksz)' ) .
!
                q = i + blksz - 1_i4b
!
!               COMPUTE THE blksz BY blksz UNIT LOWER TRIANGULAR MATRIX lmat(:blksz,:blksz)
!               USED IN THE BLOCKED ALGORITHM.
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do l = 1_i4b, blksz
!
                    i1 = i + l
                    k1 = il(i1)
!
!                   SAVE HOUSEHOLDER VECTOR mat(:i1-1,i1) .
!
                    umat(1_i4b:i1-1_i4b,l) = matp(k1:k1+i1-2_i4b)
                    umat(i1:q,l)           = zero
!
                    beta(l) = matp(k1+i1-1_i4b)
!
                    do l2 = l + 1_i4b, blksz
!
                        i2 = i + l2
                        k2 = il(i2)
!
                        tau = matp(k2+i2-1_i4b)
!
                        lmat(l2,l) = -tau*dot_product( matp(k1:k1+i1-2_i4b), matp(k2:k2+i1-2_i4b) )
!
                    end do
!                    
                end do
!
!$OMP END DO
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do j = 0_i4b, p2, blksz
!
                    j1 = j + 1_i4b
                    j2 = min( j + blksz, p )
                    j3 = j2 - j
!
                    c2(1_i4b:j3,1_i4b:q) = c(j1:j2,1_i4b:q)
!
#ifdef _BLAS
!                   COMPUTE c(j1:j2,:n)*U(:n,:blksz) .
!
                    call gemm( 'N', 'N', j3, blksz, q, one, c2, blksz, umat, n, zero, x, blksz  )
!
                    do l = 1_i4b, blksz
!
                        x(1_i4b:j3,l) = beta(l)*x(1_i4b:j3,l)
!                    
                    end do
!
!                   SOLVE x(j1:j2,:blksz)*lmat(:blksz,:blksz)' = c(j1:j2,:n)*U(:n,:blksz) .
!
                    call trsm( 'R', 'L', 'T', 'U', j3, blksz, one, lmat, blksz, x, blksz )
!
!                   COMPUTE c(j1:j2,:n) = c(j1:j2,:n) + x(j1:j2,:blksz)*U(:n,:blksz)' .
!
                    call gemm( 'N', 'T', j3, q,  blksz, one, x, blksz, umat, n, one, c2, blksz )
#else
!                   COMPUTE c(j1:j2,:n)*U(:n,:blksz) .
!
                    do l = 1_i4b, blksz
!
                        i1 = i + l - 1_i4b
!
                        x(1_i4b:j3,l) = umat(1_i4b,l)*c2(1_i4b:j3,1_i4b)
!
                        do l2 = 2_i4b, i1
!
                            x(1_i4b:j3,l) = x(1_i4b:j3,l) +  umat(l2,l)*c2(1_i4b:j3,l2)
!                    
                        end do
!
                        x(1_i4b:j3,l) = beta(l)*x(1_i4b:j3,l)
!                    
                    end do
!
!                   SOLVE x(j1:j2,:blksz)*lmat(:blksz,:blksz)' = c(j1:j2,:n)*U(:n,:blksz) AND
!                   COMPUTE c(j1:j2,i+1:i+blksz-1) = c(j1:j2,i+1:i+blksz-1) + x(j1:j2,:blksz)*U(i+1:i+blksz-1,:blksz)' .
!
                    do l = 1_i4b, blksz
!
                        do l2 = l + 1_i4b, blksz
!
                            x(1_i4b:j3,l2) = x(1_i4b:j3,l2) - x(1_i4b:j3,l)*lmat(l2,l)
!
                        end do
!
                        do l2 = 1_i4b, l-1_i4b
!
                            i2 = i + l2
!
                            c2(1_i4b:j3,i2) = c2(1_i4b:j3,i2) + x(1_i4b:j3,l)*umat(i2,l)
!
                        end do
!
                    end do
!
!                   COMPUTE c(j1:j2,:i) = c(j1:j2,:i) + x(j1:j2,:blksz)*U(:i,:blksz)' .
!
                    do l = 1_i4b, i
!
                        do l2 = 1_i4b, blksz
!
                            c2(1_i4b:j3,l) = c2(1_i4b:j3,l) + x(1_i4b:j3,l2)*umat(l,l2)
!                    
                        end do
!                    
                    end do
#endif
!
                    c(j1:j2,1_i4b:q) = c2(1_i4b:j3,1_i4b:q)
!
                end do
!
!$OMP END DO
!
            end do
!
!           APPLY THE REMINDING HOUSEHOLDER TRANSFORMATIONS.
!
            i1 = i + 1_i4b
!
            do i = i1, nq
!
                k1 = il(i)
!
                tau = matp(k1+i-1_i4b)
!
                if ( tau==zero ) cycle
!
!               APPLY H(i) TO c(:m,:n) FROM THE RIGHT.
!
                q = i - 1_i4b
!
                tmp(1_i4b:q) = matp(k1:k1+i-2_i4b)
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do j = 0_i4b, p2, blksz
!
                    j1 = j + 1_i4b
                    j2 = min( j + blksz, p )
                    j3 = j2 - j
!
                    c2(1_i4b:j3,1_i4b:q) = c(j1:j2,1_i4b:q)
!
#ifdef _BLAS
                    call gemv( 'N', j3, q, tau, c2, blksz, tmp(1_i4b:q), 1_i4b, zero, x(1_i4b:j3,1_i4b), 1_i4b )
!
                    call ger( j3, q, one, x(1_i4b:j3,1_i4b), 1_i4b, tmp(1_i4b:q), 1_i4b, c2, blksz )
#else
                    x(1_i4b:j3,1_i4b) = matmul( c2(1_i4b:j3,1_i4b:q), tmp(1_i4b:q) )

                    do l = 1_i4b, q
!
                        prod = tmp(l)*tau
!
                        c2(1_i4b:j3,l) = c2(1_i4b:j3,l) + prod*x(1_i4b:j3,1_i4b)     
!
                    end do            
#endif
!
                    c(j1:j2,1_i4b:q) = c2(1_i4b:j3,1_i4b:q)
!
                end do
!
!$OMP END DO
!
            end do
!
        end if
!
!       DEALLOCATE (PRIVATE) WORK ARRAYS.
!
        deallocate( x, c2 )
!
    end if
!
!$OMP END PARALLEL
!
!   DEALLOCATE (SHARED) WORK ARRAY.
!
    deallocate( umat )
!
!
! END OF SUBROUTINE apply_q_symtrid_d
! ___________________________________
!
    end subroutine apply_q_symtrid_d
!
!
! =========================================================================================
!                                SORTING ROUTINES
! =========================================================================================
!
!
    subroutine eig_sort_a( sort, d, u )
!
! Purpose
! _______
!
!   Given the eigenvalues D and eigenvectors U as output from EIG_CMP, EIG_CMP2
!   EIG_CMP3 or SYMTRID_QRI, SYMTRID_QRI2 and SYMTRID_QRI3, this routine sorts
!   the eigenvalues into ascending or descending order, and, rearranges the
!   columns of U correspondingly.
!
!
! Arguments
! _________
!
!   SORT         (INPUT) character
!                Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!                The eigenvectors are reordered accordingly.
!
!   D            (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the eigenvalues.
!
!                On exit, the eigenvalues in ascending or decreasing order.
!
!   U            (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the columns of U are the eigenvectors. 
!
!                On exit, U contains the reordered eigenvectors.
!
!                The shape of U must verify:  size( U, 2 ) = size( D ) = m .
!
!
! Further Details
! _______________
!
!   The method is straight insertion.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert_eq, iminloc
#ifdef _BLAS
    use BLAS_interfaces, only : swap
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout) :: d(:), u(:,:)
!
    character , intent(in)  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
#ifdef _BLAS
    real(stnd) :: tmp
#else
    real(stnd) :: tmp_vec(size(u,1)), tmp
#endif
!
    integer(i4b) :: i, l, n, m
!   
!   
! PARAMETERS 
! __________
!
    character , parameter ::  a_maj=achar(65),  a_min=achar(97),       &
                              d_maj=achar(68),  d_min=achar(100)
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   CHECK THE ARGUMENTS.
!
    n = size( u, 1 )
    m = assert_eq(    int( size(d), i4b ),    &
                      int( size(u,2), i4b ),  &
                      'eig_sort' )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( m<=1_i4b ) return
!
!   SORT THE EIGENVALUES.
!   INSERTION SORT ON EIGENVALUES, BUT ONLY ONE
!   TRANSPOSITION PER EIGENVECTOR.
!
    select case( sort )
!
        case( a_maj, a_min ) 
!
!           SORT IN ASCENDING ORDER.
!
            do l = 1, m - 1_i4b
!
!               SCAN FOR SMALLEST d(l) .
!
                i = iminloc( d(l:) ) + l - 1_i4b
!
                if ( i==l ) cycle
!
!               SWAP EIGENVALUES AND EIGENVECTORS.
!
                tmp  = d(i)
                d(i) = d(l)
                d(l) = tmp
!
#ifdef _BLAS
                call swap( n, u(:n,l), 1_i4b, u(:n,i), 1_i4b )
#else
                tmp_vec(:n) = u(:n,i)
                u(:n,i) = u(:n,l)
                u(:n,l) = tmp_vec(:n)
#endif
!
            end do
!
        case( d_maj, d_min ) 
!
!           SORT IN DESCENDING ORDER.
!
            do l = m, 2_i4b, -1_i4b
!
!               SCAN FOR SMALLEST d(l) .
!
                i = iminloc( d(:l) )
!
                if ( i==l ) cycle
!
!               SWAP EIGENVALUES AND EIGENVECTORS.
!
                tmp  = d(i)
                d(i) = d(l)
                d(l) = tmp
!
#ifdef _BLAS
                call swap( n, u(:n,l), 1_i4b, u(:n,i), 1_i4b )
#else
                tmp_vec(:n) = u(:n,i)
                u(:n,i) = u(:n,l)
                u(:n,l) = tmp_vec(:n)
#endif
!
            end do
!
    end select
!
!
! END OF SUBROUTINE eig_sort_a
! ____________________________
!
    end subroutine eig_sort_a
!
! =========================================================================================
!
    subroutine eig_abs_sort_a( sort, d, u )
!
! Purpose
! _______
!
!   Given the eigenvalues D and eigenvectors U as output from EIG_CMP, EIG_CMP2
!   EIG_CMP3 or SYMTRID_QRI, SYMTRID_QRI2 and SYMTRID_QRI3, this routine sorts
!   the eigenvalues into ascending or descending order of absolute magnitude, and,
!   rearranges the columns of U correspondingly.
!
!
! Arguments
! _________
!
!   SORT         (INPUT) character
!                Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd' of absolute magnitude.
!                The eigenvectors are reordered accordingly.
!
!   D            (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the eigenvalues.
!
!                On exit, the eigenvalues in ascending or decreasing order of
!                absolute magnitude.
!
!   U            (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the columns of U are the eigenvectors. 
!
!                On exit, U contains the reordered eigenvectors.
!
!                The shape of U must verify:  size( U, 2 ) = size( D ) = m .
!
!
! Further Details
! _______________
!
!   The method is straight insertion.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert_eq, iminloc
#ifdef _BLAS
    use BLAS_interfaces, only : swap
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout) :: d(:), u(:,:)
!
    character , intent(in)  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
#ifdef _BLAS
    real(stnd) :: tmp
#else
    real(stnd) :: tmp_vec(size(u,1)), tmp
#endif
!
    integer(i4b) :: i, l, n, m
!   
!   
! PARAMETERS 
! __________
!
    character , parameter ::  a_maj=achar(65),  a_min=achar(97),       &
                              d_maj=achar(68),  d_min=achar(100)
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   CHECK THE ARGUMENTS.
!
    n = size( u, 1 )
    m = assert_eq(    int( size(d), i4b ),    &
                      int( size(u,2), i4b ),  &
                      'eig_abs_sort' )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( m<=1_i4b ) return
!
!   SORT THE EIGENVALUES.
!   INSERTION SORT ON EIGENVALUES, BUT ONLY ONE
!   TRANSPOSITION PER EIGENVECTOR.
!
    select case( sort )
!
        case( a_maj, a_min ) 
!
!           SORT IN ASCENDING ORDER.
!
            do l = 1, m - 1_i4b
!
!               SCAN FOR SMALLEST abs( d(l) ) .
!
                i = iminloc( abs( d(l:) ) ) + l - 1_i4b
!
                if ( i==l ) cycle
!
!               SWAP EIGENVALUES AND EIGENVECTORS.
!
                tmp  = d(i)
                d(i) = d(l)
                d(l) = tmp
!
#ifdef _BLAS
                call swap( n, u(:n,l), 1_i4b, u(:n,i), 1_i4b )
#else
                tmp_vec(:n) = u(:n,i)
                u(:n,i) = u(:n,l)
                u(:n,l) = tmp_vec(:n)
#endif
!
            end do
!
        case( d_maj, d_min ) 
!
!           SORT IN DESCENDING ORDER.
!
            do l = m, 2_i4b, -1_i4b
!
!               SCAN FOR SMALLEST abs( d(l) ) .
!
                i = iminloc( abs( d(:l) ) )
!
                if ( i==l ) cycle
!
!               SWAP EIGENVALUES AND EIGENVECTORS.
!
                tmp  = d(i)
                d(i) = d(l)
                d(l) = tmp
!
#ifdef _BLAS
                call swap( n, u(:n,l), 1_i4b, u(:n,i), 1_i4b )
#else
                tmp_vec(:n) = u(:n,i)
                u(:n,i) = u(:n,l)
                u(:n,l) = tmp_vec(:n)
#endif
!
            end do
!
    end select
!
!
! END OF SUBROUTINE eig_abs_sort_a
! ________________________________
!
    end subroutine eig_abs_sort_a
!
! =========================================================================================
!
    subroutine eigval_sort( sort, d )
!
! Purpose
! _______
!
!   Given the eigenvalues D as output from EIGVAL_CMP, EIGVAL_CMP2, EIGVAL_CMP3 or
!   SYMTRID_QRI, SYMTRID_QRI2 and SYMTRID_QRI3, this routine sorts the
!   eigenvalues into ascending or descending order.
!
!
! Arguments
! _________
!
!   SORT         (INPUT) character
!                Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!   D            (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the eigenvalues.
!
!                On exit, the eigenvalues in ascending or decreasing order.
!
!
! Further Details
! _______________
!
!   The method is quick sort.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Sort_Procedures, only : quick_sort
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout) :: d(:)
!
    character , intent(in)  :: sort
!   
!   
! PARAMETERS 
! __________
!
    character , parameter ::  a_maj=achar(65),  a_min=achar(97),       &
                              d_maj=achar(68),  d_min=achar(100)
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   SORT EIGENVALUES.
!
    select case( sort )
!
        case( a_maj, a_min ) 
!
!           SORT IN ASCENDING ORDER.
!
            call quick_sort( d, ascending=true )
!
        case( d_maj, d_min ) 
!
!           SORT IN DESCENDING ORDER.
!
            call quick_sort( d, ascending=false )
!
    end select
!
!
! END OF SUBROUTINE eigval_sort
! _____________________________
!
    end subroutine eigval_sort
!
! =========================================================================================
!
    subroutine eigval_abs_sort( sort, d )
!
! Purpose
! _______
!
!   Given the eigenvalues D as output from EIGVAL_CMP, EIGVAL_CMP2, EIGVAL_CMP3 or
!   SYMTRID_QRI, SYMTRID_QRI2 and SYMTRID_QRI3, this routine sorts the
!   eigenvalues into ascending or descending order of absolute magnitude.
!
!
! Arguments
! _________
!
!   SORT         (INPUT) character
!                Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd' of  absolute magnitude.
!
!   D            (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the eigenvalues.
!
!                On exit, the eigenvalues in ascending or decreasing order
!                of absolute magnitude.
!
!
! Further Details
! _______________
!
!   The method is straight insertion.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : iminloc
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout) :: d(:)
!
    character , intent(in)  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: tmp
!
    integer(i4b) :: i, l, m
!   
!   
! PARAMETERS 
! __________
!
    character , parameter ::  a_maj=achar(65),  a_min=achar(97),       &
                              d_maj=achar(68),  d_min=achar(100)
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   CHECK THE ARGUMENTS.
!
    m = size( d )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( m<=1_i4b ) return
!
!   SORT THE EIGENVALUES.
!
    select case( sort )
!
        case( a_maj, a_min ) 
!
!           SORT IN ASCENDING ORDER.
!
            do l = 1, m - 1_i4b
!
!               SCAN FOR SMALLEST abs( d(l) ) .
!
                i = iminloc( abs( d(l:) ) ) + l - 1_i4b
!
                if ( i==l ) cycle
!
!               SWAP EIGENVALUES.
!
                tmp  = d(i)
                d(i) = d(l)
                d(l) = tmp
!
            end do
!
        case( d_maj, d_min ) 
!
!           SORT IN DESCENDING ORDER.
!
            do l = m, 2_i4b, -1_i4b
!
!               SCAN FOR SMALLEST abs( d(l) ) .
!
                i = iminloc( abs( d(:l) ) )
!
                if ( i==l ) cycle
!
!               SWAP EIGENVALUES.
!
                tmp  = d(i)
                d(i) = d(l)
                d(l) = tmp
!
            end do
!
    end select
!
!
! END OF SUBROUTINE eigval_abs_sort
! _________________________________
!
    end subroutine eigval_abs_sort
!
!
! =========================================================================================
!                         SYMMETRIC TRIDIAGONAL EIGENVALUES DECOMPOSITIONS
! =========================================================================================
!
!
    subroutine symtrid_qri_a( d, e, failure, mat, init_mat, sort, maxiter )
!
! Purpose
! _______
!
!   SYMTRID_QRI computes all eigenvalues and eigenvectors of a
!   symmetric n-by-n tridiagonal matrix using the implicit QR method.
!
!   The eigenvalues and eigenvectors of a full symmetric matrix can 
!   also be found if SYMTRID_CMP and ORTHO_GEN_SYMTRID have been used 
!   to reduce this matrix to tridiagonal form before calling SYMTRID_QRI.
!
!
! Arguments
! _________
!
!   D           (INPUT/OUTPUT) real(stnd), dimension(:)
!               On entry, the diagonal elements of the tridiagonal matrix.
!
!               On exit, the eigenvalues.
!
!   E           (INPUT/OUTPUT) real(stnd), dimension(:)
!               On entry, the n-1 subdiagonal elements of the tridiagonal matrix.
!               E(n) is arbitrary and is used as workspace.
!
!               On exit, E has been destroyed.
!
!               The size of E must verify:  size( E ) = size( D ) = n .
!
!   FAILURE     (OUTPUT) logical(lgl)
!               On exit:
!
!               - FAILURE = false :  indicates successful exit.
!               - FAILURE = true  :  indicates that the algorithm did not converge and
!                 that full accuracy was not attained in the Schur decomposition
!                 of the tridiagonal matrix.
!
!   MAT         (INPUT/OUTPUT) real(stnd), dimension(:,:)
!               On entry, if:
!
!               - INIT_MAT is absent or if INIT_MAT = false, then MAT
!                 contains the orthogonal matrix used in the reduction 
!                 to tridiagonal form as returned by ORTHO_GEN_SYMTRID.
!               - INIT_MAT is present and INIT_MAT = true, MAT is set to the
!                 identity matrix of order n.
!
!               On exit, if FAILURE = false:
!
!               - MAT contains the orthonormal eigenvectors of the
!                 original symmetric matrix if INIT_MAT is absent
!                 or if INIT_MAT = false,
!               - MAT contains the orthonormal eigenvectors of the
!                 symmetric tridiagonal matrix if INIT_MAT is present
!                 and INIT_MAT = true.
!
!               The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = size( D ) = n .
!
!   INIT_MAT    (INPUT, OPTIONAL) logical(lgl)
!               If:
!
!               - INIT_MAT = false: The subroutine computes eigenvalues and eigenvectors
!                 of the original symmetric matrix.  On entry, MAT must contain the
!                 orthogonal matrix used to reduce the original matrix
!                 to tridiagonal form.
!               - INIT_MAT = true: The subroutine computes eigenvalues and eigenvectors of the
!                 tridiagonal matrix.  MAT is initialized to the identity matrix of order n.
!
!               The default is false.
!
!   SORT        (INPUT, OPTIONAL) character
!               Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!               or in descending order if SORT = 'D' or 'd'.
!               The eigenvectors are reordered accordingly.
!
!               By default, the eigenvalues are not sorted.
!
!   MAXITER     (INPUT, OPTIONAL) integer(i4b)
!               MAXITER controls the maximum number of QR sweeps in the
!               Schur decomposition of the tridiagonal matrix.
!               The algorithm fails to converge if the number of QR sweeps
!               exceeds MAXITER * size(D). Convergence usually occurs in 
!               about 2 * size(D) QR sweeps.
!
!               The default is 30.
!
!
! Further Details
! _______________
!
!   The eigenvalues and eigenvectors are computed by the implicit tridiagonal QR algorithm
!   described in the reference (1) with modifications suggested in the reference (2).
!
!   This subroutine is adapted from the routine DSTEQR in LAPACK. Note that this subroutine
!   is parallelized with OPENMP using the method described in the reference (3).
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Greenbaum, A., and J. Dongarra, J., 1989:
!           Experiments with QR/QL Methods for the Symmetric Tridiagonal Eigenproblem.
!           LAPACK Working Note No 17, November 1989.
!
!   (3) Demmel, J., Heath, M.T., and Van Der Vorst, H., 1993:
!           Parallel numerical linear algebra.
!           Acta Numerica 2, 111-197.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,    only : lamch
#ifdef _OPENMP2
    use Char_Constants,   only : allocate_error
#endif
    use Utilities,        only : assert_eq, merror, iminloc, lascl, unit_matrix
#ifdef _BLAS
    use BLAS_interfaces,  only : rot
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:), intent(inout) :: mat
    real(stnd), dimension(:),   intent(inout) :: d, e
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(in), optional :: init_mat
    logical(lgl), intent(out)          :: failure
!
    character ,   intent(in), optional :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: eps, eps2, safmin, safmax, ssfmax, ssfmin,    &
                  tridnorm, tst, r, g, f, b, s, c, p, rt1, rt2
    real(stnd), dimension(size(d)) :: temp
!
    integer(i4b) :: n, m, l, i, ip1, iscale, nqrs, maxqrs
!
#ifdef _OPENMP2
    real(stnd), dimension(size(d))          :: d_tmp, e_tmp
    real(stnd), dimension(:,:), allocatable :: mat2
!
    integer(i4b) :: n2, j1, j2, num_threads, thread_num, jrem, joffset
    integer      :: iok
!
    logical :: test_nopar, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='symtrid_qri'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(d),i4b)     ,   &
                    int(size(e),i4b)     ,   &
                    name_proc )
!
    if ( present(init_mat) ) then
!
!       COMPUTE THE EIGENVECTORS AND EIGENVALUES OF THE TRIDIAGONAL MATRIX.
!
        if ( init_mat ) then
!
            call unit_matrix( mat )
!
        end if
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) return
!
!   DETERMINE THE UNIT ROUNDOFF AND OVER/UNDERFLOW THRESHOLDS.
!
    eps  = fudge_eps*lamch( 'u' )
    eps2 = eps*eps
    safmin = lamch( 's' )
    safmax = one/safmin
    ssfmax = sqrt( safmax )/three
    ssfmin = sqrt( safmin )/eps2
!
!   COMPUTE MAX ELEMENT OF TRIDIAGONAL MATRIX.
!
    e(n) = zero
    tridnorm = abs( d(1_i4b) ) + abs( e(1_i4b) )
!
    do i = 2_i4b, n
        tridnorm = max( tridnorm, abs(d(i))+abs(e(i))+abs(e(i-1_i4b)) )
    end do
!
!   SCALE TRIDIAGONAL MATRIX IF NECESSARY.
!
    iscale = 0_i4b
!
    if ( tridnorm==zero ) then
!
        return
!
    else if ( tridnorm>ssfmax ) then
!
        iscale = 1_i4b
        call lascl( d(:n), tridnorm, ssfmax )
        call lascl( e(:n-1_i4b), tridnorm, ssfmax )
!
    else if ( tridnorm<ssfmin ) then
!
        iscale = 2_i4b
        call lascl( d(:n), tridnorm, ssfmin )
        call lascl( e(:n-1_i4b), tridnorm, ssfmin )
!
    end if
!
!   CONVENIENT TO RENUMBER THE ELEMENTS OF e(:) .
!
    temp(:n) = eoshift( e(:n), -1 )
    e(:n) = temp(:n)
!
!   INITIALIZE COUNTERS FOR QR ITERATIONS.
!
    if ( present(maxiter) ) then
        maxqrs = n*merge( maxiter, 30_i4b , maxiter>=1_i4b )
    else
        maxqrs = n*30_i4b
    end if
!
#ifdef _OPENMP2
    i = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested()
    test_nopar  = ( ompparallel .and. .not.ompnested ) .or.      &
                  n*n<omp_limit                        .or.      &
                  n/i<omp_limit2                       .or.      &
                  i==1_i4b
!
    if ( test_nopar ) then
#endif
!
!       QR ITERATIONS.
!
        nqrs = 0_i4b
        l = n
!
        do
!
!           TEST FOR SPLITTING : LOOK FOR SMALL SUBDIAGONAL ELEMENT.
!
            do m = l, 2_i4b, -1_i4b
                tst = e(m)**2
                if ( tst<=eps2*abs( d(m)*d(m-1_i4b) ) + safmin ) exit
            end do            
!
            if ( m==l ) then
!
!               EIGENVALUE FOUND.
!
                l = l - 1_i4b
!
                if ( l>1_i4b ) cycle
!
                exit
!
            else if ( m==l-1_i4b ) then
!
!               IF REMAINING MATRIX IS 2-BY-2, USE LAEV2
!               TO COMPUTE ITS EIGENSYSTEM.
!
                call laev2( d(m), e(l), d(l), rt1, rt2, c, s )
!
!               APPLY ROTATION.
!
#ifdef _BLAS
                call rot( n, mat(:n,m), 1_i4b, mat(:n,l), 1_i4b, c, s )
#else
                temp(:n) = mat(:n,m)
                mat(:n,m) = c*temp(:n) + s*mat(:n,l)
                mat(:n,l) = -s*temp(:n) + c*mat(:n,l)
#endif
!
                d(m) = rt1
                d(l) = rt2
!
                l = l - 2_i4b
!
                if ( l>1_i4b ) cycle
!
                exit
!
            end if
!
!           CUTOFF FOR CONVERGENCE FAILURE.
!
            if ( nqrs>=maxqrs ) then
!
!               SET FAILURE INDICATOR.
!
                failure = true
!
!               IF FAILURE TO CONVERGE SET SMALLEST MAGNITUDE TERM
!               IN OFF-DIAGONAL TO ZERO AND CONTINUE ON.
!
                i = iminloc( abs(e(m+1_i4b:l)) ) + m
                e(i) = zero
                nqrs = 0_i4b
!
                cycle
!
            end if
!
            nqrs = nqrs + 1_i4b
!
!           FORM Wilkinson SHIFT.
!
            g = ( d(l-1_i4b) - d(l) )/( two*e(l) )
!
            if ( abs(g)>one ) then
                p = one/g
                r = g*sqrt( one + p*p )
            else
                r = sign( sqrt( one + g*g ), g )
            end if
!
            g = d(m) - d(l) + ( e(l)/( g + r ) )
!
            s = one
            c = one
            p = zero
!
#ifndef _BLAS
            temp(:n) = mat(:n,m)
#endif
!
!           INNER LOOP.
!
            do i = m, l-1_i4b
!
                ip1 = i + 1_i4b
!
                f = s*e(ip1)
!
                if ( f==zero ) exit
!
                b = c*e(ip1)
!
                if ( abs(g)>=abs(f) ) then
                    s = f/g
                    r = sign( sqrt( one + s*s ), g )
                    c = one/r
                    s = c*s
                    e(i) = g*r
                else
                    c = g/f
                    r = sign( sqrt( one + c*c ), f )
                    s = one/r
                    c = s*c
                    e(i) = f*r
                end if
!
!               APPLY ROTATION.
!
#ifdef _BLAS
                call rot( n, mat(:n,i), 1_i4b, mat(:n,ip1), 1_i4b, c, s )
#else
                mat(:n,i) = c*temp(:n) + s*mat(:n,ip1)
                temp(:n) = -s*temp(:n) + c*mat(:n,ip1)
#endif
!
                g = d(i) - p
                r = ( d(ip1) - g )*s + two*c*b
                p = s*r
                d(i) = g + p
                g = c*r - b
!
            end do
!
#ifndef _BLAS
            mat(:n,i) = temp(:n)
#endif
!
            d(i) = d(i) - p
            e(i) = g
            e(m) = zero
!
        end do
!
#ifdef _OPENMP2
    else
!
!$OMP  PARALLEL PRIVATE(l,i,ip1,j1,j2,m,n2,num_threads,thread_num,          &
!$OMP                   jrem,joffset,iok,nqrs,                              &
!$OMP                   tst,rt1,rt2,c,s,p,r,g,f,b,temp,d_tmp,e_tmp,mat2),   &
!$OMP           SHARED(maxqrs,failure)
!
!       MAKE A LOCAL COPY OF THE TRIDIAGONAL MATRIX.
!
        d_tmp(1_i4b:n) = d(1_i4b:n)
        e_tmp(1_i4b:n) = e(1_i4b:n)
!
!       DISTRIBUTE THE WORK AMONG THE THREADS.
!
        num_threads = omp_get_num_threads()
        thread_num  = omp_get_thread_num()
!
        jrem = mod( n, num_threads )
        n2   = ( n - jrem )/num_threads
!
        if ( thread_num<jrem ) then
            n2      = n2 + 1_i4b
            joffset = 1_i4b
        else
            joffset = jrem + 1_i4b
        end if
!
        j1 = thread_num*n2 + joffset
        j2 = j1 + n2 - 1_i4b
!
!       ALLOCATE WORK VARIABLE TO STORE A LOCAL COPY OF A
!       PORTION OF THE INPUT MATRIX.
!
        allocate( mat2(n2,n), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!$OMP BARRIER
!
!       MAKE A LOCAL COPY OF A PORTION OF THE INPUT MATRIX.
!
        mat2(1_i4b:n2,1_i4b:n) = mat(j1:j2,1_i4b:n)
!
!       QR ITERATIONS.
!
        nqrs = 0_i4b
        l = n
!
        do
!
!           TEST FOR SPLITTING : LOOK FOR SMALL SUBDIAGONAL ELEMENT.
!
            do m = l, 2_i4b, -1_i4b
                tst = e_tmp(m)**2
                if ( tst<=eps2*abs( d_tmp(m)*d_tmp(m-1_i4b) ) + safmin ) exit
            end do            
!
            if ( m==l ) then
!
!               EIGENVALUE FOUND.
!
                l = l - 1_i4b
!
                if ( l>1_i4b ) cycle
!
                exit
!
            else if ( m==l-1_i4b ) then
!
!               IF REMAINING MATRIX IS 2-BY-2, USE LAEV2
!               TO COMPUTE ITS EIGENSYSTEM.
!
                call laev2( d_tmp(m), e_tmp(l), d_tmp(l), rt1, rt2, c, s )
!
!               APPLY ROTATION.
!
#ifdef _BLAS
                call rot( n2, mat2(1_i4b:n2,m), 1_i4b, mat2(1_i4b:n2,l), 1_i4b, c, s )
#else
                temp(1_i4b:n2)   = mat2(1_i4b:n2,m)
                mat2(1_i4b:n2,m) = c*temp(1_i4b:n2) + s*mat2(1_i4b:n2,l)
                mat2(1_i4b:n2,l) = -s*temp(1_i4b:n2) + c*mat2(1_i4b:n2,l)
#endif
!
                d_tmp(m) = rt1
                d_tmp(l) = rt2
!
                l = l - 2_i4b
!
                if ( l>1_i4b ) cycle
!
                exit
!
            end if
!
!           CUTOFF FOR CONVERGENCE FAILURE.
!
            if ( nqrs>=maxqrs ) then
!
!               SET FAILURE INDICATOR.
!
                if ( thread_num==0_i4b ) then
                    failure = true
                end if
!
!               IF FAILURE TO CONVERGE SET SMALLEST MAGNITUDE TERM
!               IN OFF-DIAGONAL TO ZERO AND CONTINUE ON.
!
                i        = iminloc( abs(e_tmp(m+1_i4b:l)) ) + m
                e_tmp(i) = zero
                nqrs     = 0_i4b
!
                cycle
!
            end if
!
            nqrs = nqrs + 1_i4b
!
!           FORM Wilkinson SHIFT.
!
            g = ( d_tmp(l-1_i4b) - d_tmp(l) )/( two*e_tmp(l) )
!
            if ( abs(g)>one ) then
                p = one/g
                r = g*sqrt( one + p*p )
            else
                r = sign( sqrt( one + g*g ), g )
            end if
!
            g = d_tmp(m) - d_tmp(l) + ( e_tmp(l)/( g + r ) )
!
            s = one
            c = one
            p = zero
!
#ifndef _BLAS
            temp(1_i4b:n2) = mat2(1_i4b:n2,m)
#endif
!
!           INNER LOOP.
!
            do i = m, l-1_i4b
!
                ip1 = i + 1_i4b
!
                f = s*e_tmp(ip1)
!
                if ( f==zero ) exit
!
                b = c*e_tmp(ip1)
!
                if ( abs(g)>=abs(f) ) then
                    s = f/g
                    r = sign( sqrt( one + s*s ), g )
                    c = one/r
                    s = c*s
                    e_tmp(i) = g*r
                else
                    c = g/f
                    r = sign( sqrt( one + c*c ), f )
                    s = one/r
                    c = s*c
                    e_tmp(i) = f*r
                end if
!
!               APPLY ROTATION.
!
#ifdef _BLAS
                call rot( n2, mat2(:n2,i), 1_i4b, mat2(:n2,ip1), 1_i4b, c, s )
#else
                mat2(1_i4b:n2,i) = c*temp(1_i4b:n2)  + s*mat2(1_i4b:n2,ip1)
                temp(1_i4b:n2)   = -s*temp(1_i4b:n2) + c*mat2(1_i4b:n2,ip1)
#endif
!
                g = d_tmp(i) - p
                r = ( d_tmp(ip1) - g )*s + two*c*b
                p = s*r
                d_tmp(i) = g + p
                g = c*r - b
!
            end do
!
#ifndef _BLAS
            mat2(1_i4b:n2,i) = temp(1_i4b:n2)
#endif
!
            d_tmp(i) = d_tmp(i) - p
            e_tmp(i) = g
            e_tmp(m) = zero
!
        end do
!
        if ( thread_num==0_i4b ) then
            d(1_i4b:n) = d_tmp(1_i4b:n)
            e(1_i4b:n) = e_tmp(1_i4b:n)
        end if
!
        mat(j1:j2,1_i4b:n) = mat2(1_i4b:n2,1_i4b:n)
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( mat2 )
!
!$OMP END PARALLEL
!
    end if
#endif
!
!   UNDO SCALING IF NECESSARY.
!
    if ( iscale>0_i4b ) then
        c = merge( ssfmax, ssfmin, iscale==1_i4b )
        call lascl( d, c, tridnorm )
    end if
!
!   ORDER EIGENVALUES AND EIGENVECTORS IF REQUIRED.
!
    if ( present(sort) )  call eig_sort_a( sort, d, mat )
!
!
! END OF SUBROUTINE symtrid_qri_a
! _______________________________
!
    end subroutine symtrid_qri_a
!
! =========================================================================================
!
    subroutine symtrid_qri_b( d, e, failure, sort, maxiter )
!
! Purpose
! _______
!
!   SYMTRID_QRI computes all eigenvalues of a symmetric tridiagonal matrix
!   using the Pal-Walker-Kahan variant of the QR algorithm.
!
!   The eigenvalues of a full symmetric matrix can also be found
!   if SYMTRID_CMP has been used to reduce this matrix to tridiagonal form
!   before calling SYMTRID_QRI.
!
!
! Arguments
! _________
!
!   D            (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the diagonal elements of the tridiagonal matrix.
!
!                On exit, the eigenvalues.
!
!   E            (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the n-1 subdiagonal elements of the tridiagonal matrix.
!                E(n) is arbitrary and is used as workspace.
!
!                On exit, E has been destroyed.
!
!                The size of E must verify:  size( E ) = size( D ) .
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  this indicates successful exit.
!                - FAILURE = true  :  this indicates that the algorithm did not converge and
!                  that full accuracy was not attained in the Schur decomposition
!                  of the tridiagonal matrix.
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!                By default, the eigenvalues are not sorted.
!
!   MAXITER      (INPUT, OPTIONAL) integer(i4b)
!                MAXITER controls the maximum number of QR sweeps in the
!                Schur decomposition of the tridiagonal matrix.
!                The algorithm fails to converge if the number of QR sweeps
!                exceeds MAXITER * size(D). Convergence usually occurs in 
!                about 2 * size(D) QR sweeps.
!
!                The default is 30.
!
!
! Further Details
! _______________
!
!   The eigenvalues are computed by the Pal-Walker-Kahan variant of the implicit tridiagonal
!   QR algorithm described in the reference (1) with modifications suggested in the reference (2).
!
!   This subroutine is adapted from the routine DSTERF in LAPACK. This subroutine is not parallelized.
!
!   For further details, see:
!
!   (1) Parlett, B.N., 1998:
!           The Symmetric Eigenvalue Problem.
!           Revised edition, SIAM, Philadelphia.
!
!   (2) Greenbaum, A., and J. Dongarra, J., 1989:
!           Experiments with QR/QL Methods for the Symmetric Tridiagonal Eigenproblem.
!           LAPACK Working Note No 17, November 1989.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,    only : lamch
    use Utilities,        only : assert_eq, iminloc, lascl
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:), intent(inout) :: d, e
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(out)          :: failure
!
    character ,   intent(in), optional :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: eps, eps2, safmin, safmax, ssfmax, ssfmin, tridnorm,         &
                  tst, r, bb, s, c, p, rte, rt1, rt2, sigma, gamma,            &
                  oldgam, oldc, alpha
!
    integer(i4b) :: n, m, l, i, iscale, nqrs, maxqrs
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='symtrid_qri'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(d),i4b)     ,   &
                    int(size(e),i4b)     ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) return
!
!   DETERMINE THE UNIT ROUNDOFF AND OVER/UNDERFLOW THRESHOLDS.
!
    eps  = fudge_eps*lamch( 'u' )
    eps2 = eps*eps
    safmin = lamch( 's' )
    safmax = one/safmin
    ssfmax = sqrt( safmax )/three
    ssfmin = sqrt( safmin )/eps2
!
!   COMPUTE MAX ELEMENT OF TRIDIAGONAL MATRIX.
!
    e(n) = zero
    tridnorm = abs( d(1_i4b) ) + abs( e(1_i4b) )
!
    do i = 2_i4b, n
        tridnorm = max( tridnorm, abs(d(i))+abs(e(i))+abs(e(i-1_i4b)) )
    end do
!
!   SCALE TRIDIAGONAL MATRIX IF NECESSARY.
!
    iscale = 0_i4b
!
    if ( tridnorm==zero ) then
!
        return
!
    else if ( tridnorm>ssfmax ) then
!
        iscale = 1_i4b
        call lascl( d(:), tridnorm, ssfmax )
        call lascl( e(:n-1_i4b), tridnorm, ssfmax )
!
    else if ( tridnorm<ssfmin ) then
!
        iscale = 2_i4b
        call lascl( d(:), tridnorm, ssfmin )
        call lascl( e(:n-1_i4b), tridnorm, ssfmin )
!
    end if
!
!   STORE THE ELEMENTS OF THE TRIDIAGONAL MATRIX IN REVERSED ORDER IF
!   THE DIAGONAL ELEMENTS ARE INCREASING IN MAGNITUDE ALONG THE DIAGONAL
!   IN ORDER TO IMPROVE THE ACCURACY OF THE QR ITERATIONS FOR GRADED MATRICES.
!
    if ( reverse(d(:)) ) then
        d(1_i4b:n)       = d(n:1_i4b:-1_i4b)
        e(1_i4b:n-1_i4b) = e(n-1_i4b:1_i4b:-1_i4b)
    end if
!
!   RENUMBER AND SQUARE THE ELEMENTS OF e(:) .
!
    e(:n) = ( eoshift( e(:n), -1 ) )**2
!
    nqrs = 0_i4b
    if ( present(maxiter) ) then
        maxqrs = n*merge( maxiter, 30_i4b , maxiter>=1_i4b )
    else
        maxqrs = n*30_i4b
    end if
!
!   QR ITERATIONS.
!
    l = n
!
    do
!
!       TEST FOR SPLITTING : LOOK FOR SMALL SUBDIAGONAL ELEMENT.
!
        do m = l, 2_i4b, -1_i4b
            tst = abs( e(m) )
            if ( tst<=eps2*abs( d(m)*d(m-1_i4b) ) + safmin ) exit
        end do            
!
        if ( m==l ) then
!
!           EIGENVALUE FOUND.
!
            l = l - 1_i4b
!
            if ( l>1_i4b ) cycle
!
            exit
!
        else if ( m==l-1_i4b ) then
!
!           IF REMAINING MATRIX IS 2-BY-2, USE LAE2
!           TO COMPUTE ITS EIGENVALUES.
!
            rte = sqrt( e(l) )
            call lae2( d(l), rte, d(m), rt1, rt2 )
!
            d(l) = rt1
            d(m) = rt2
            l = l - 2_i4b
!
            if ( l>1_i4b ) cycle
!
            exit
!
        end if
!
!       CUTOFF FOR CONVERGENCE FAILURE.
!
        if ( nqrs>=maxqrs ) then
!
!           SET FAILURE INDICATOR.
!
            failure = true
!
!           IF FAILURE TO CONVERGE SET SMALLEST MAGNITUDE TERM
!           IN OFF-DIAGONAL TO ZERO AND CONTINUE ON.
!
            i = iminloc( e(m+1_i4b:l) ) + m
            e(i) = zero
            nqrs = 0_i4b
!
            cycle
!
        end if
!
        nqrs = nqrs + 1_i4b
!
!       FORM Wilkinson SHIFT.
!
        rte = sqrt( e(l) )
        sigma = ( d(l-1_i4b) - d(l) )/( two*rte )
!
        if ( abs(sigma)>one ) then
!
            p = one/sigma
            r = sigma*sqrt( one + p*p )
!
        else
!
            r = sign( sqrt( one + sigma*sigma ), sigma )
!
        end if
!
        sigma = d(l) - ( rte/( sigma + r ) )
!
        s = zero
        c = one
        gamma = d(m) - sigma
        p = gamma*gamma
!
!       INNER LOOP.
!
        do i = m, l-1_i4b
!
            bb = e(i+1_i4b)
            r = p + bb
            e(i) = s*r
            oldc = c
            c = p/r
            s = bb/r
            oldgam = gamma
            alpha = d(i+1_i4b)
            gamma = c*( alpha - sigma ) - s*oldgam
            d(i) = oldgam + ( alpha - gamma )
!
            if ( c/=zero ) then
                p = ( gamma*gamma )/c
            else
                p = oldc*bb
            end if
!
        end do
!
        e(l) = s*p
        d(l) = sigma + gamma
        e(m) = zero
!
    end do
!
!   UNDO SCALING IF NECESSARY.
!
    if ( iscale>0_i4b ) then
        c = merge( ssfmax, ssfmin, iscale==1_i4b )
        call lascl( d, c, tridnorm )
    end if
!
!   SORT EIGENVALUES IF REQUIRED.
!
    if ( present(sort) ) call eigval_sort( sort, d )
!
!
! END OF SUBROUTINE symtrid_qri_b
! _______________________________
!
    end subroutine symtrid_qri_b
!
! =========================================================================================
!
    subroutine symtrid_qri2_a( d, e, failure, mat, init_mat, sort, maxiter, max_francis_steps )
!
! Purpose
! _______
!
!   SYMTRID_QRI2 computes all eigenvalues and eigenvectors of a
!   symmetric n-by-n tridiagonal matrix with a perfect shift strategy
!   for the eigenvectors.
!
!   The eigenvalues and eigenvectors of a full symmetric matrix can
!   also be found if SYMTRID_CMP and ORTHO_GEN_SYMTRID have been used
!   to reduce this matrix to tridiagonal form before calling SYMTRID_QRI2.
!
!
! Arguments
! _________
!
!   D                   (INPUT/OUTPUT) real(stnd), dimension(:)
!                       On entry, the diagonal elements of the tridiagonal matrix.
!
!                       On exit, the eigenvalues.
!
!   E                   (INPUT/OUTPUT) real(stnd), dimension(:)
!                       On entry, the n-1 subdiagonal elements of the tridiagonal matrix.
!                       E(n) is arbitrary and is used as workspace.
!
!                       On exit, E has been destroyed.
!
!                       The size of E must verify:  size( E ) = size( D ) = n .
!
!   FAILURE             (OUTPUT) logical(lgl)
!                       On exit:
!
!                       - FAILURE = false :  indicates successful exit.
!                       - FAILURE = true  :  indicates that the algorithm did not converge and
!                         that full accuracy was not attained in the Schur decomposition
!                         of the tridiagonal matrix.
!
!   MAT                 (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                       On entry, if:
!
!                       - INIT_MAT is absent or if INIT_MAT = false, then MAT
!                         contains the orthogonal matrix used in the reduction 
!                         to tridiagonal form as returned by ORTHO_GEN_SYMTRID.
!                       - INIT_MAT is present and INIT_MAT = true, MAT is set to the
!                         identity matrix of order n.
!
!                       On exit, if FAILURE = false:
!
!                       - MAT contains the orthonormal eigenvectors of the
!                         original symmetric matrix if INIT_MAT is absent
!                         or if INIT_MAT = false,
!                       - MAT contains the orthonormal eigenvectors of the
!                         symmetric tridiagonal matrix if INIT_MAT is present
!                         and INIT_MAT = true.
!
!                       The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = size( D ) = n .
!
!   INIT_MAT            (INPUT, OPTIONAL) logical(lgl)
!                       If:
!
!                       - INIT_MAT = false: The subroutine computes eigenvalues and eigenvectors
!                         of the original symmetric matrix.  On entry, MAT must contain the
!                         orthogonal matrix used to reduce the original matrix
!                         to tridiagonal form.
!                       - INIT_MAT = true: The subroutine computes eigenvalues and eigenvectors of the
!                         tridiagonal matrix.  MAT is initialized to the identity matrix of order n.
!
!                       The default is false.
!
!   SORT                (INPUT, OPTIONAL) character
!                       Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                       or in descending order if SORT = 'D' or 'd'.
!                       The eigenvectors are reordered accordingly.
!
!                       By default, the eigenvalues are not sorted.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the
!                       Schur decomposition of the tridiagonal matrix.
!                       The algorithm fails to converge if the number of QR sweeps
!                       exceeds MAXITER * size(D). Convergence usually occurs in 
!                       about 2 * size(D) QR sweeps.
!
!                       The default is 30.
!
!   MAX_FRANCIS_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                       MAX_FRANCIS_STEPS controls the maximum number of Francis sets
!                       (e.g. QR sweeps) of Givens rotations which must be saved
!                       before applying them with a wavefront algorithm to accumulate
!                       the eigenvectors in the QR algorithm. MAX_FRANCIS_STEPS is a
!                       strictly positive integer, otherwise the default value is used.
!
!                       The default is 10.
!
!
! Further Details
! _______________
!
!   The eigenvalues are computed by the Pal-Walker-Kahan variant of the implicit tridiagonal
!   QR algorithm described in the reference (1).
!
!   The eigenvectors are computed with a perfect shift strategy (see the references (1) and (2))
!   with modifications suggested in the references (3) and (4), for applying a set of Givens
!   rotations and deflating a given eigenvalue from the tridiagonal matrix.
!
!   Furthermore, the computation of the eigenvectors is parallelized if OPENMP is used.
!
!   With all these changes, SYMTRID_QRI2 is much more efficient than SYMTRID_QRI for
!   computing the full set of eigenvectors of a real n-by-n symmetric tridiagonal matrix.
!
!   For further details, see:
!
!   (1) Parlett, B.N., 1998:
!           The Symmetric Eigenvalue Problem.
!           Revised edition, SIAM, Philadelphia.
!
!   (2) Greenbaum, A., and J. Dongarra, J., 1989:
!           Experiments with QR/QL Methods for the Symmetric Tridiagonal Eigenproblem.
!           LAPACK Working Note No 17, November 1989.
!
!   (3) Van Zee, F.G., Van de Geijn, R., and Quintana-Orti, G., 2011:
!           Restructuring the QR Algorithm for High-Performance Application of Givens Rotations.
!           FLAME Working Note 60. The University of Texas at Austin, Department of Computer
!           Sciences. Technical Report TR-11-36.
!
!   (4) Malyshev, A.N., 2000:
!           On deflation for symmetric tridiagonal matrices.
!           Report 182 of the Department of Informatics, University of Bergen, Norway.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,         only : lamch
    use Char_Constants,        only : allocate_error
    use Utilities,             only : assert_eq, merror, iminloc, lascl, unit_matrix
#ifdef _F2003
    use ieee_arithmetic,       only : ieee_support_datatype, ieee_is_normal
    use Utilities,             only : realloc => mvalloc
#else
    use Num_Constants,         only : is_nan
    use Utilities_With_Pnter,  only : realloc
#endif
#ifdef _BLAS
    use BLAS_interfaces,       only : rot
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:), intent(inout) :: mat
    real(stnd), dimension(:),   intent(inout) :: d, e
!
    integer(i4b), intent(in), optional :: maxiter, max_francis_steps
!
    logical(lgl), intent(in), optional :: init_mat
    logical(lgl), intent(out)          :: failure
!
    character ,   intent(in), optional :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: eps, eps2, safmin, safmax, ssfmax, ssfmin, tridnorm,             &
                  t, r, g, f, b, s, c, p, oldg, oldc, alpha, beta, lambda
    real(stnd), dimension(size(d))          :: e2, temp
#ifdef _F2003
    real(stnd), dimension(:,:), allocatable :: csmat, snmat
#else
    real(stnd), dimension(:,:), pointer     :: csmat, snmat
#endif
!
    integer(i4b) :: n, m, l, i, ip1, i1, i2, j, k, iscale, nqrs, maxqrs,   &
                    nqrs2, maxqrs2, niter, o, q, qstart, qend, nfrancis
#ifdef _F2003
    integer(i4b), dimension(:), allocatable :: deb, fin
#else
    integer(i4b), dimension(:), pointer     :: deb, fin
#endif
    integer :: iok
!
    logical(lgl) :: use_perfect_shift, deflate
!
#ifdef _OPENMP2
    real(stnd), dimension(:,:), allocatable :: mat2
!
    integer(i4b) :: n2, mm, j1, j2, num_threads, thread_num, jrem, joffset
!
    logical      :: test_nopar, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='symtrid_qri2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(d),i4b)     ,   &
                    int(size(e),i4b)     ,   &
                    name_proc )
!
    if ( present(init_mat) ) then
!
!       COMPUTE THE EIGENVECTORS AND EIGENVALUES OF THE TRIDIAGONAL MATRIX.
!
        if ( init_mat ) then
!
            call unit_matrix( mat )
!
        end if
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) return
!
!   DETERMINE THE UNIT ROUNDOFF AND OVER/UNDERFLOW THRESHOLDS.
!
    eps    = fudge_eps*lamch( 'u' )
    eps2   = eps*eps
    safmin = lamch( 's' )
    safmax = one/safmin
    ssfmax = sqrt( safmax )/three
    ssfmin = sqrt( safmin )/eps2
!
!   COMPUTE MAX ELEMENT OF TRIDIAGONAL MATRIX.
!
    e(n) = zero
    tridnorm = abs( d(1_i4b) ) + abs( e(1_i4b) )
!
    do i = 2_i4b, n
        tridnorm = max( tridnorm, abs(d(i))+abs(e(i))+abs(e(i-1_i4b)) )
    end do
!
!   SCALE TRIDIAGONAL MATRIX IF NECESSARY.
!
    iscale = 0_i4b
!
    if ( tridnorm==zero ) then
!
        return
!
    else if ( tridnorm>ssfmax ) then
!
        iscale = 1_i4b
        call lascl( d(:n), tridnorm, ssfmax )
        call lascl( e(:n-1_i4b), tridnorm, ssfmax )
!
    else if ( tridnorm<ssfmin ) then
!
        iscale = 2_i4b
        call lascl( d(:n), tridnorm, ssfmin )
        call lascl( e(:n-1_i4b), tridnorm, ssfmin )
!
    end if
!
!   CONVENIENT TO RENUMBER THE ELEMENTS OF e(:) .
!
    e2(:n) = eoshift( e(:n), -1 )
    e(:n)  = e2(:n)
!
!   INITIALIZE COUNTERS FOR QR ITERATIONS.
!
    nqrs  = 0_i4b
    niter = 0_i4b
!
    if ( present(maxiter) ) then
        maxqrs = n*merge( maxiter, 30_i4b , maxiter>=1_i4b )
    else
        maxqrs = n*30_i4b
    end if
!    
    if ( present(max_francis_steps) ) then
        nfrancis = merge( max_francis_steps,                                  &
                          max_francis_steps_default,                          &
                          max_francis_steps>=1_i4b .and. max_francis_steps<=n )
    else
        nfrancis = max_francis_steps_default
    end if
!
    nqrs2   = 0_i4b
    maxqrs2 = nfrancis + 5_i4b
!
!   ALLOCATE WORK VARIABLES NEEDED TO STORE THE GIVENS ROTATIONS.
!
    allocate( csmat(n,maxqrs2),     &
              snmat(n,maxqrs2),     &
              deb(maxqrs2),         &
              fin(maxqrs2),         &
              stat = iok      )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    deb(:) = 1_i4b
    fin(:) = -1_i4b
!
#ifdef _OPENMP2
    i = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested()
    test_nopar  = ( ompparallel .and. .not.ompnested ) .or.      &
                  n*n<omp_limit                        .or.      &
                  n/i<omp_limit2                       .or.      &
                  i==1_i4b
!
    if ( test_nopar ) then
#endif
!
!       QR ITERATIONS.
!
        do l = n, 2_i4b, -1_i4b
!
!           TEST FOR SPLITTING : LOOK FOR SMALL SUBDIAGONAL ELEMENT.
!           MAKE A COPY OF SQUARES OF THE OFF-DIAGONAL ELEMENTS.
!
            do m = l, 2_i4b, -1_i4b
!
                t = e(m)**2
!
                if ( t <= eps2*abs( d(m)*d(m-1_i4b) ) + safmin ) exit
!
                e2(m) = t
!
            end do
!
            e2(m) = zero
!
            if ( m/=l ) then
!
                k = m        
!
                deflate = true
                use_perfect_shift = true
!
!               MAKE A COPY OF THE DIAGONAL ELEMENTS OF THE TRIDIAGONAL MATRIX.
!
                i = max( 1_i4b, m-1_i4b )
                temp(i:l) = d(i:l)
!
!               STORE THE ELEMENTS OF THE TRIDIAGONAL MATRIX IN REVERSED ORDER IF
!               THE DIAGONAL ELEMENTS ARE INCREASING ALONG THE DIAGONAL IN ORDER
!               TO MAINTAIN ACCURACY OF THE QR ITERATIONS.
!
!                if ( reverse( d(m:l) ) ) then
!
!                    temp(m:l)     = d(l:m:-1_i4b)
!                    e2(m+1_i4b:l) = e2(l:m+1_i4b:-1_i4b)
!
!                else
!
!                    temp(m:l) = d(m:l)
!
!                end if
!
!               COMPUTE THE NEXT EIGENVALUE.
!
                do
!
!                   CUTOFF FOR CONVERGENCE FAILURE.
!
                    if ( niter>=maxqrs ) then
!
!                       IF FAILURE TO CONVERGE, COMPUTE EIGENVECTOR
!                       WITH Wilkinson SHIFT.
!
                        if ( m/=l ) then
!
                            deflate           = false
                            use_perfect_shift = false
!
                            m     = l
                            niter = 0_i4b
!
                        end if
!
                    end if
!
!                   EIGENVALUE FOUND, COMPUTE EIGENVECTOR.
!
                    if ( m==l ) exit
!
!                   COMPUTE EIGENVALUE WITH Pal-Walker-Kahan ALGORITHM.
!
                    niter = niter + 1_i4b
!
!                   FORM Wilkinson SHIFT.
!
                    g = sqrt( e2(l) )
                    beta = ( temp(l-1_i4b) - temp(l) )/( two*g )
!
                    if ( abs(beta)>one ) then
!
                        p = one/beta
                        r = beta*sqrt( one + p*p )
!
                    else
!
                        r = sign( sqrt( one + beta*beta ), beta )
!
                    end if
!
                    beta = temp(l) - ( g/( beta + r ) )
!
                    s = zero
                    c = one
                    g = temp(m) - beta
                    p = g*g
!
!                   INNER LOOP.
!
                    do i = m, l-1_i4b
!
                        b       = e2(i+1_i4b)
                        r       = p + b
                        e2(i)   = s*r
                        oldc    = c
                        c       = p/r
                        s       = b/r
                        oldg    = g
                        alpha   = temp(i+1_i4b)
                        g       = c*( alpha-beta ) - s*oldg
                        temp(i) = oldg + ( alpha-g )
!
                        if ( c/=zero ) then
                            p = ( g*g ) / c
                        else
                            p = oldc*b
                        end if
!
                    end do
!
                    e2(l)   = s*p
                    temp(l) = real( beta, extd ) + real( g, extd )
                    e2(m)   = zero
!
!                   TEST FOR SPLITTING : LOOK FOR SMALL SUBDIAGONAL ELEMENT.
!
                    do m = l, k + 1_i4b, -1_i4b
!
                        t = abs( e2(m) )
!
                        if ( t <= eps2*abs( temp(m)*temp(m-1_i4b) ) + safmin ) exit
!
                    end do
!       
                end do
!
                if ( deflate ) then
!
                    lambda = temp(l)
!
#ifdef _F2003
                    if ( ieee_support_datatype( lambda ) ) then
!
                        if ( .not.ieee_is_normal( lambda ) ) then
!
                            deflate           = false
                            use_perfect_shift = false
!
                        end if
!
                    end if
#else
                    if ( is_nan( lambda ) ) then
!
                        deflate           = false
                        use_perfect_shift = false
!
                    end if
#endif
!
                end if
!
!               EIGENVALUE FOUND, COMPUTE EIGENVECTOR IMPLICITLY.
!
                do
!
!                   CUTOFF FOR CONVERGENCE FAILURE.
!
                    if ( nqrs>=maxqrs ) then
!
!                       SET FAILURE INDICATOR.
!
                        failure = true
!
!                       IF FAILURE TO CONVERGE SET SMALLEST MAGNITUDE TERM
!                       IN OFF-DIAGONAL TO ZERO AND CONTINUE ON.
!
                        k    = iminloc( abs( e(k+1_i4b:l))) + k
                        nqrs = 0_i4b
!
                        if ( k==l ) exit
!
                    end if
!
                    nqrs  = nqrs + 1_i4b
                    nqrs2 = nqrs2 + 1_i4b
!
                    if ( nqrs2>maxqrs2 ) then
!
                        maxqrs2 = maxqrs2 + 5_i4b
!
                        call realloc( deb, maxqrs2, iok )
!
                        if ( iok==0 ) then
!
                            call realloc( fin, maxqrs2, iok )
!
                            if ( iok==0 ) then
!
                                call realloc( csmat, n, maxqrs2, iok )
!
                                if ( iok==0 ) then
!
                                    call realloc( snmat, n, maxqrs2, iok )
!
                                end if
!
                            end if
!
                        end if
!
                        if ( iok/=0 ) then
                            call merror( name_proc//allocate_error )
                        end if
!
                    end if
!
                    if ( deflate ) then
!
!                       TRY FERNANDO METHOD COMBINED WITH GODUNOV DEFLATION FIRST.
!
                        call dflgen2( d(k:l), e(k+1_i4b:l), lambda,                                &
                                      csmat(k:l-1_i4b,nqrs2), snmat(k:l-1_i4b,nqrs2), deflate )
!
                        if ( deflate ) then
!
                            e(k) = zero
!
                            deb(nqrs2) = k
                            fin(nqrs2) = l - 1_i4b
!
                            exit
!
                        end if
!
                    end if
!
                    if ( use_perfect_shift ) then
!
!                       USE PERFECT SHIFT.
!
                        g = real( d(k), extd ) - real( lambda, extd )
!
                        use_perfect_shift = false
!
                    else
!
!                       USE Wilkinson SHIFT.
!
                        g = ( d(l-1_i4b) - d(l) )/( two*e(l) )
!
                        if ( abs(g)>one ) then
                            p = one/g
                            r = g*sqrt( one + p*p )
                        else
                            r = sign( sqrt( one + g*g ), g )
                        end if
!
                        g = d(k) - d(l) + ( e(l)/( g + r ) )
!
                    end if
!
                    s = one
                    c = one
                    p = zero
!
!                   INNER LOOP.
!
                    do i = k, l-1_i4b
!
                        ip1 = i + 1_i4b
!
                        f = s*e(ip1)
!
                        if ( f==zero ) exit
!
                        b = c*e(ip1)
!
                        if ( abs(g)>=abs(f) ) then
!
                            s    = f/g
                            r    = sign( sqrt( one + s*s ), g )
                            c    = one/r
                            s    = c*s
                            e(i) = g*r
!
                        else
!
                            c    = g/f
                            r    = sign( sqrt( one + c*c ), f )
                            s    = one/r
                            c    = s*c
                            e(i) = f*r
!
                        end if
!
!                       STORE ROTATION.
!
                        csmat(i,nqrs2) = c
                        snmat(i,nqrs2) = s
!
!                       APPLY ROTATION.
!
                        g    = d(i) - p
                        r    = ( d(ip1) - g )*s + two*c*b
                        p    = s*r
                        d(i) = g + p
                        g    = c*r - b
!
                    end do
!
                    d(i) = d(i) - p
                    e(i) = g
                    e(k) = zero
!
                    deb(nqrs2) = k
                    fin(nqrs2) = i - 1_i4b
!
!                   TEST FOR SPLITTING : LOOK FOR SMALL SUBDIAGONAL ELEMENT.
!
                    do k = l, 2_i4b, -1_i4b
!
                        t = e(k)**2
!
                        if ( t<=eps2*abs( d(k)*d(k-1_i4b) ) + safmin ) exit
!
                    end do
!
!                   EIGENVECTOR FOUND.
!
                    if ( k==l ) exit
!
                end do
!
            end if
!
            if ( nqrs2>=nfrancis .or. l==2_i4b  ) then
!
!               APPLY THE PREVIOUS FRANCIS SETS OF GIVENS ROTATIONS.
!
                qstart = maxval( deb(:nqrs2) ) + nqrs2 - 1_i4b
                qend   = minval( fin(:nqrs2) )
!
!               DETERMINE WHAT ALGORITHM TO USE.
!
                if ( qstart>=qend .or. nqrs2==1_i4b ) then
!
!                   APPLY THE GIVENS ROTATIONS WITH A SEQUENTIAL ALGORITHM.
!
                    do j = 1_i4b, nqrs2
!
                        i1 = deb(j)
                        i2 = fin(j)
!
#ifdef _BLAS
                        do i = i1, i2
!
                            call rot( n, mat(:n,i), 1_i4b, mat(:n,i+1_i4b), 1_i4b, csmat(i,j), snmat(i,j) )
!
                        end do
#else
                        temp(:n) = mat(:n,i1)
!
                        do i = i1, i2
!
                            ip1 = i + 1_i4b
!
                            c = csmat(i,j)
                            s = snmat(i,j)
                            mat(:n,i) =  c*temp(:n) + s*mat(:n,ip1)
                            temp(:n)  = -s*temp(:n) + c*mat(:n,ip1)
!
                        end do
!
                        mat(:n,i) = temp(:n)
#endif
!
                    end do
!
                else if ( nqrs2>1_i4b ) then
!
!                   APPLY THE GIVENS ROTATIONS WITH A WAVEFRONT ALGORITHM
!                   IN ORDER TO MINIMIZE CACHE MISSES.
!
!                   STARTUP PHASE.
!
                    o  = 0_i4b
                    i2 = qstart - nqrs2 + 1_i4b
!
                    do q = i2, qstart - 1_i4b
!
                        o = o + 1_i4b
!
                        i1 = deb(o)
!
                        if ( i1<i2 ) then
!
#ifdef _BLAS
                            do i = i1, i2 - 1_i4b
!
                                call rot( n, mat(:n,i), 1_i4b, mat(:n,i+1_i4b), 1_i4b, csmat(i,o), snmat(i,o) )
!
                            end do
#else
                            temp(:n) = mat(:n,i1)
!
                            do i = i1, i2 - 1_i4b
!
                                ip1 = i + 1_i4b
!
                                c = csmat(i,o)
                                s = snmat(i,o)
                                mat(:n,i) =  c*temp(:n) + s*mat(:n,ip1)
                                temp(:n)  = -s*temp(:n) + c*mat(:n,ip1)
!
                            end do
!
                            mat(:n,i) = temp(:n)
#endif
!
                        end if
!
                        i = q
!
#ifdef _BLAS
                        do j = 1_i4b, o
!
                            call rot( n, mat(:n,i+1_i4b), 1_i4b, mat(:n,i), 1_i4b, csmat(i,j), -snmat(i,j) )
!
                            i = i - 1_i4b
!
                        end do
#else
                        temp(:n) = mat(:n,i+1_i4b)
!
                        do j = 1_i4b, o
!
                            c = csmat(i,j)
                            s = snmat(i,j)
                            mat(:n,i+1_i4b) = -s*mat(:n,i) + c*temp(:n)
                            temp(:n)        =  c*mat(:n,i) + s*temp(:n)
!
                            i = i - 1_i4b
!
                        end do
!
                        mat(:n,i+1_i4b) = temp(:n)
#endif
!
                    end do
!
!                    i2 = qstart
!
!                    do j = 1_i4b, nqrs2
!
!                        i1 = deb(j)
!                        i2 = i2 - 1_i4b
!
!#ifdef _BLAS
!                        do i = i1, i2
!
!                            call rot( n, mat(:n,i), 1_i4b, mat(:n,i+1_i4b), 1_i4b, csmat(i,j), snmat(i,j) )
!
!                        end do
!#else
!                        temp(:n) = mat(:n,i1)
!
!                        do i = i1, i2
!
!                            ip1 = i + 1_i4b
!
!                            c = csmat(i,j)
!                            s = snmat(i,j)
!                            mat(:n,i) =  c*temp(:n) + s*mat(:n,ip1)
!                            temp(:n)  = -s*temp(:n) + c*mat(:n,ip1)
!
!                        end do
!
!                        mat(:n,i) = temp(:n)
!#endif
!
!                    end do
!
!                   PIPELINE PHASE.
!
                    do q = qstart, qend
!
                        i = q
!
#ifdef _BLAS
                        do j = 1_i4b, nqrs2
!
                            call rot( n, mat(:n,i), 1_i4b, mat(:n,i+1_i4b), 1_i4b, csmat(i,j), snmat(i,j) )
!
                            i = i - 1_i4b
!
                        end do
#else
                        temp(:n) = mat(:n,i+1_i4b)
!
                        do j = 1_i4b, nqrs2
!
                            c = csmat(i,j)
                            s = snmat(i,j)
                            mat(:n,i+1_i4b) = -s*mat(:n,i) + c*temp(:n)
                            temp(:n)        =  c*mat(:n,i) + s*temp(:n)
!
                            i = i - 1_i4b
!
                        end do
!
                        mat(:n,i+1_i4b) = temp(:n)
#endif
!
                    end do
!
!                   SHUTDOWN PHASE.
!
                    i1 = qend + 1_i4b
!
                    do j = 1_i4b, nqrs2
!
                        i2 = fin(j)
!
                        if ( i1<=i2 ) then
!
#ifdef _BLAS
                            do i = i1, i2
!
                                call rot( n, mat(:n,i), 1_i4b, mat(:n,i+1_i4b), 1_i4b, csmat(i,j), snmat(i,j) )
!
                            end do
#else
                            temp(:n) = mat(:n,i1)
!
                            do i = i1, i2
!
                                ip1 = i + 1_i4b
!
                                c = csmat(i,j)
                                s = snmat(i,j)
                                mat(:n,i) =  c*temp(:n) + s*mat(:n,ip1)
                                temp(:n)  = -s*temp(:n) + c*mat(:n,ip1)
!
                            end do
!
                            mat(:n,i) = temp(:n)
#endif
!
                        end if
!
                        i1 = i1 - 1_i4b
!
                    end do
!
                end if
!
                nqrs2 = 0_i4b
!
            end if
!
        end do
!
#ifdef _OPENMP2
    else
!
!$OMP  PARALLEL PRIVATE(i,ip1,i1,i2,l,m,j,j1,j2,k,n2,num_threads,thread_num,jrem,joffset,iok,  &
!$OMP                   o,q,qstart,qend,b,c,s,t,g,p,r,beta,alpha,lambda,oldc,oldg,temp,mat2)
!
!       DISTRIBUTE THE WORK AMONG THE THREADS.
!
        num_threads = omp_get_num_threads()
        thread_num  = omp_get_thread_num()
!
        jrem = mod( n, num_threads )
        n2   = ( n - jrem )/num_threads
!
        if ( thread_num<jrem ) then
            n2      = n2 + 1_i4b
            joffset = 1_i4b
        else
            joffset = jrem + 1_i4b
        end if
!
        j1 = thread_num*n2 + joffset
        j2 = j1 + n2 - 1_i4b
!
!       ALLOCATE WORK VARIABLE TO STORE A LOCAL COPY OF A
!       PORTION OF THE INPUT MATRIX.
!
        allocate( mat2(n2,n), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A LOCAL COPY OF A PORTION OF THE INPUT MATRIX.
!
        mat2(:n2,:n) = mat(j1:j2,:n)
!
!       QR ITERATIONS.
!
        do l = n, 2_i4b, -1_i4b
!
!$OMP SINGLE
!
!           TEST FOR SPLITTING : LOOK FOR SMALL SUBDIAGONAL ELEMENT.
!           MAKE A COPY OF SQUARES OF THE OFF-DIAGONAL ELEMENTS.
!
            do m = l, 2_i4b, -1_i4b
!
                t = e(m)**2
!
                if ( t <= eps2*abs( d(m)*d(m-1_i4b) ) + safmin ) exit
!
                e2(m) = t
!
            end do
!
            e2(m) = zero
            mm = m        
!
!$OMP END SINGLE
!
            if ( mm/=l ) then
!
                if ( thread_num==0_i4b ) then
!
                    deflate           = true
                    use_perfect_shift = true
                    m                 = mm
!
!                   MAKE A COPY OF THE DIAGONAL ELEMENTS OF THE TRIDIAGONAL MATRIX.
!
                    i = max( 1_i4b, m-1_i4b )
                    temp(i:l) = d(i:l)
!
!                   STORE THE ELEMENTS OF THE TRIDIAGONAL MATRIX IN REVERSED ORDER IF
!                   THE DIAGONAL ELEMENTS ARE INCREASING ALONG THE DIAGONAL IN ORDER
!                   TO MAINTAIN ACCURACY OF THE QR ITERATIONS.
!
!                    if ( reverse( d(m:l) ) ) then
!
!                        temp(m:l)     = d(l:m:-1_i4b)
!                        e2(m+1_i4b:l) = e2(l:m+1_i4b:-1_i4b)
!
!                    else
!
!                        temp(m:l) = d(m:l)
!
!                    end if
!
!                   COMPUTE THE NEXT EIGENVALUE.
!
                    do
!
!                       CUTOFF FOR CONVERGENCE FAILURE.
!
                        if ( niter>=maxqrs ) then
!
!                           IF FAILURE TO CONVERGE, COMPUTE EIGENVECTOR
!                           WITH Wilkinson SHIFT.
!
                            if ( m/=l ) then
!
                                deflate           = false
                                use_perfect_shift = false
!
                                m     = l
                                niter = 0_i4b
!
                            end if
!
                        end if
!
!                       EIGENVALUE FOUND, COMPUTE EIGENVECTOR.
!
                        if ( m==l ) exit
!
!                       COMPUTE EIGENVALUE WITH Pal-Walker-Kahan ALGORITHM.
!
                        niter = niter + 1_i4b
!
!                       FORM Wilkinson SHIFT.
!
                        g = sqrt( e2(l) )
                        beta = ( temp(l-1_i4b) - temp(l) )/( two*g )
!
                        if ( abs(beta)>one ) then
!
                            p = one/beta
                            r = beta*sqrt( one + p*p )
!
                        else
!
                            r = sign( sqrt( one + beta*beta ), beta )
!
                        end if
!
                        beta = temp(l) - ( g/( beta + r ) )
!
                        s = zero
                        c = one
                        g = temp(m) - beta
                        p = g*g
!
!                       INNER LOOP.
!
                        do i = m, l-1_i4b
!
                            b       = e2(i+1_i4b)
                            r       = p + b
                            e2(i)   = s*r
                            oldc    = c
                            c       = p/r
                            s       = b/r
                            oldg    = g
                            alpha   = temp(i+1_i4b)
                            g       = c*( alpha-beta ) - s*oldg
                            temp(i) = oldg + ( alpha-g )
!
                            if ( c/=zero ) then
                                p = ( g*g ) / c
                            else
                                p = oldc*b
                            end if
!
                        end do
!
                        e2(l)   = s*p
                        temp(l) = real( beta, extd ) + real( g, extd )
                        e2(m)   = zero
!
!                       TEST FOR SPLITTING : LOOK FOR SMALL SUBDIAGONAL ELEMENT.
!
                        do m = l, mm + 1_i4b, -1_i4b
!
                            t = abs( e2(m) )
!
                            if ( t <= eps2*abs( temp(m)*temp(m-1_i4b) ) + safmin ) exit
!
                        end do            
!
                    end do
!
                    if ( deflate ) then
!
                        lambda = temp(l)
!
#ifdef _F2003
                        if ( ieee_support_datatype( lambda ) ) then
!
                            if ( .not.ieee_is_normal( lambda ) ) then
!
                                deflate           = false
                                use_perfect_shift = false
!
                            end if
!
                        end if
#else
                        if ( is_nan( lambda ) ) then
!
                            deflate           = false
                            use_perfect_shift = false
!
                        end if
#endif
!
                    end if
!
!                   EIGENVALUE FOUND, COMPUTE EIGENVECTOR IMPLICITLY.
!
                    k = nqrs2
                    m = mm
!
                    do
!
!                       CUTOFF FOR CONVERGENCE FAILURE.
!
                        if ( nqrs>=maxqrs ) then
!
!                           SET FAILURE INDICATOR.
!
                            failure = true
!
!                           IF FAILURE TO CONVERGE SET SMALLEST MAGNITUDE TERM
!                           IN OFF-DIAGONAL TO ZERO AND CONTINUE ON.
!
                            m    = iminloc( abs( e(m+1_i4b:l))) + m
                            nqrs = 0_i4b
!
                            if ( m==l ) exit
!
                        end if
!
                        nqrs = nqrs + 1_i4b
!
                        k = k + 1_i4b
!
                        if ( k>maxqrs2 ) then
!
                            maxqrs2 = maxqrs2 + 5_i4b
!
                            call realloc( deb, maxqrs2, iok )
!
                            if ( iok==0 ) then
!
                                call realloc( fin, maxqrs2, iok )
!
                                if ( iok==0 ) then
!
                                    call realloc( csmat, n, maxqrs2, iok )
!
                                    if ( iok==0 ) then
!
                                        call realloc( snmat, n, maxqrs2, iok )
!
                                    end if
!
                                end if
!
                            end if
!
                            if ( iok/=0 ) then
                                call merror( name_proc//allocate_error )
                            end if
!
                        end if
!
                        if ( deflate ) then
!
!                           TRY FERNANDO METHOD COMBINED WITH GODUNOV DEFLATION FIRST.
!
                            call dflgen2( d(m:l), e(m+1_i4b:l), lambda,                        &
                                          csmat(m:l-1_i4b,k), snmat(m:l-1_i4b,k), deflate )
!
                            if ( deflate ) then
!
                                e(m) = zero
!
                                deb(k) = m
                                fin(k) = l - 1_i4b
!
                                exit
!
                            end if
!
                        end if
!
                        if ( use_perfect_shift ) then
!
!                           USE PERFECT SHIFT.
!
                            g = real( d(m), extd ) - real( lambda, extd )
!
                            use_perfect_shift = false
!
                        else
!
!                           USE Wilkinson SHIFT.
!
                            g = ( d(l-1_i4b) - d(l) )/( two*e(l) )
!
                            if ( abs(g)>one ) then
                                p = one/g
                                r = g*sqrt( one + p*p )
                            else
                                r = sign( sqrt( one + g*g ), g )
                            end if
!
                            g = d(m) - d(l) + ( e(l)/( g + r ) )
!
                        end if
!
                        s = one
                        c = one
                        p = zero
!
!                       INNER LOOP.
!
                        do i = m, l-1_i4b
!
                            ip1 = i + 1_i4b
!
                            f = s*e(ip1)
!
                            if ( f==zero ) exit
!
                            b = c*e(ip1)
!
                            if ( abs(g)>=abs(f) ) then
!
                                s    = f/g
                                r    = sign( sqrt( one + s*s ), g )
                                c    = one/r
                                s    = c*s
                                e(i) = g*r
!
                            else
!
                                c    = g/f
                                r    = sign( sqrt( one + c*c ), f )
                                s    = one/r
                                c    = s*c
                                e(i) = f*r
!
                            end if
!
!                           STORE ROTATION.
!
                            csmat(i,k) = c
                            snmat(i,k) = s
!
!                           APPLY ROTATION.
!
                            g    = d(i) - p
                            r    = ( d(ip1) - g )*s + two*c*b
                            p    = s*r
                            d(i) = g + p
                            g    = c*r - b
!
                        end do
!
                        d(i) = d(i) - p
                        e(i) = g
                        e(m) = zero
!
                        deb(k) = m
                        fin(k) = i - 1_i4b
!
!                       TEST FOR SPLITTING : LOOK FOR SMALL SUBDIAGONAL ELEMENT.
!
                        do m = l, 2_i4b, -1_i4b
!
                            t = e(m)**2
!
                            if ( t<=eps2*abs( d(m)*d(m-1_i4b) ) + safmin ) exit
!
                        end do
!
!                       EIGENVECTOR FOUND.
!
                        if ( m==l ) exit
!
                    end do
!
                    nqrs2 = k
!
                end if
!
            end if
!
!$OMP BARRIER
!
            if ( nqrs2>=nfrancis .or. l==2_i4b  ) then
!
!               APPLY THE PREVIOUS FRANCIS SETS OF GIVENS ROTATIONS.
!
                qstart = maxval( deb(:nqrs2) ) + nqrs2 - 1_i4b
                qend   = minval( fin(:nqrs2) )
!
!               DETERMINE WHAT ALGORITHM TO USE.
!
                if ( qstart>=qend .or. nqrs2==1_i4b ) then
!
!                   APPLY THE GIVENS ROTATIONS WITH A SEQUENTIAL ALGORITHM.
!
                    do j = 1_i4b, nqrs2
!
                        i1 = deb(j)
                        i2 = fin(j)
!
#ifdef _BLAS
                        do i = i1, i2
!
                            call rot( n2, mat2(:n2,i), 1_i4b, mat2(:n2,i+1_i4b), 1_i4b, csmat(i,j), snmat(i,j) )
!
                        end do
#else
                        temp(:n2) = mat2(:n2,i1)
!
                        do i = i1, i2
!
                            ip1 = i + 1_i4b
!
                            c = csmat(i,j)
                            s = snmat(i,j)
                            mat2(:n2,i) =  c*temp(:n2) + s*mat2(:n2,ip1)
                            temp(:n2)   = -s*temp(:n2) + c*mat2(:n2,ip1)
!
                        end do
!
                        mat2(:n2,i) = temp(:n2)
#endif
!
                    end do
!
                else if ( nqrs2>1_i4b ) then
!
!                   APPLY THE GIVENS ROTATIONS WITH A WAVEFRONT ALGORITHM
!                   IN ORDER TO MINIMIZE CACHE MISSES.
!
!                   STARTUP PHASE.
!
                    o  = 0_i4b
                    i2 = qstart - nqrs2 + 1_i4b
!
                    do q = i2, qstart - 1_i4b
!
                        o = o + 1_i4b
!
                        i1 = deb(o)
!
                        if ( i1<i2 ) then
!
#ifdef _BLAS
                            do i = i1, i2 - 1_i4b
!
                                call rot( n2, mat2(:n2,i), 1_i4b, mat2(:n2,i+1_i4b), 1_i4b, csmat(i,o), snmat(i,o) )
!
                            end do
#else
                            temp(:n2) = mat2(:n2,i1)
!
                            do i = i1, i2 - 1_i4b
!
                                ip1 = i + 1_i4b
!
                                c = csmat(i,o)
                                s = snmat(i,o)
                                mat2(:n2,i) =  c*temp(:n2) + s*mat2(:n2,ip1)
                                temp(:n2)   = -s*temp(:n2) + c*mat2(:n2,ip1)
!
                            end do
!
                            mat2(:n2,i) = temp(:n2)
#endif
!
                        end if
!
                        i = q
!
#ifdef _BLAS
                        do j = 1_i4b, o
!
                            call rot( n2, mat2(:n2,i+1_i4b), 1_i4b, mat2(:n2,i), 1_i4b, csmat(i,j), -snmat(i,j) )
!
                            i = i - 1_i4b
!
                        end do
#else
                        temp(:n2) = mat2(:n2,i+1_i4b)
!
                        do j = 1_i4b, o
!
                            c = csmat(i,j)
                            s = snmat(i,j)
                            mat2(:n2,i+1_i4b) = -s*mat2(:n2,i) + c*temp(:n2)
                            temp(:n2)         =  c*mat2(:n2,i) + s*temp(:n2)
!
                            i = i - 1_i4b
!
                        end do
!
                        mat2(:n2,i+1_i4b) = temp(:n2)
#endif
!
                    end do
!
!                    i2 = qstart
!
!                    do j = 1_i4b, nqrs2
!
!                        i1 = deb(j)
!                        i2 = i2 - 1_i4b
!
!#ifdef _BLAS
!                        do i = i1, i2
!
!                            call rot( n2, mat2(:n2,i), 1_i4b, mat2(:n2,i+1_i4b), 1_i4b, csmat(i,j), snmat(i,j) )
!
!                        end do
!#else
!                        temp(:n2) = mat2(:n2,i1)
!
!                        do i = i1, i2
!
!                            ip1 = i + 1_i4b
!
!                            c = csmat(i,j)
!                            s = snmat(i,j)
!                            mat2(:n2,i) =  c*temp(:n2) + s*mat2(:n2,ip1)
!                            temp(:n2)   = -s*temp(:n2) + c*mat2(:n2,ip1)
!
!                        end do
!
!                        mat2(:n2,i) = temp(:n2)
!#endif
!
!                    end do
!
!                   PIPELINE PHASE.
!
                    do q = qstart, qend
!
                        i = q
!
#ifdef _BLAS
                        do j = 1_i4b, nqrs2
!
                            call rot( n2, mat2(:n2,i), 1_i4b, mat2(:n2,i+1_i4b), 1_i4b, csmat(i,j), snmat(i,j) )
!
                            i = i - 1_i4b
!
                        end do
#else
                        temp(:n2) = mat2(:n2,i+1_i4b)
!
                        do j = 1_i4b, nqrs2
!
                            c = csmat(i,j)
                            s = snmat(i,j)
                            mat2(:n2,i+1_i4b) = -s*mat2(:n2,i) + c*temp(:n2)
                            temp(:n2)         =  c*mat2(:n2,i) + s*temp(:n2)
!
                            i = i - 1_i4b
!
                        end do
!
                        mat2(:n2,i+1_i4b) = temp(:n2)
#endif
!
                    end do
!
!                   SHUTDOWN PHASE.
!
                    i1 = qend + 1_i4b
!
                    do j = 1_i4b, nqrs2
!
                        i2 = fin(j)
!
                        if ( i1<=i2 ) then
!
#ifdef _BLAS
                            do i = i1, i2
!
                                call rot( n2, mat2(:n2,i), 1_i4b, mat2(:n2,i+1_i4b), 1_i4b, csmat(i,j), snmat(i,j) )
!
                            end do
#else
                            temp(:n2) = mat2(:n2,i1)
!
                            do i = i1, i2
!
                                ip1 = i + 1_i4b
!
                                c = csmat(i,j)
                                s = snmat(i,j)
                                mat2(:n2,i) =  c*temp(:n2) + s*mat2(:n2,ip1)
                                temp(:n2)   = -s*temp(:n2) + c*mat2(:n2,ip1)
!
                            end do
!
                            mat2(:n2,i) = temp(:n2)
#endif
!
                        end if
!
                        i1 = i1 - 1_i4b
!
                    end do
!
                end if
!
!$OMP BARRIER
!
!$OMP MASTER
                nqrs2 = 0_i4b
!$OMP END MASTER
!
            end if
!
!$OMP BARRIER
!
        end do
!
        mat(j1:j2,:n) = mat2(:n2,:n)
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( mat2 )
!
!$OMP END PARALLEL
!
    end if
#endif
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( csmat, snmat, deb, fin )
!
!   UNDO SCALING IF NECESSARY.
!
    if ( iscale>0_i4b ) then
        c = merge( ssfmax, ssfmin, iscale==1_i4b )
        call lascl( d, c, tridnorm )
    end if
!
!   ORDER EIGENVALUES AND EIGENVECTORS IF REQUIRED.
!
    if ( present(sort) )  call eig_sort_a( sort, d, mat )
!
!
! END OF SUBROUTINE symtrid_qri2_a
! ________________________________
!
    end subroutine symtrid_qri2_a
!
! =========================================================================================
!
    subroutine symtrid_qri2_b( d, e, failure, sort, maxiter )
!
! Purpose
! _______
!
!   SYMTRID_QRI2 computes all eigenvalues of a symmetric tridiagonal matrix
!   using the Pal-Walker-Kahan variant of the QR algorithm.
!
!   The eigenvalues of a full symmetric matrix can also be found
!   if SYMTRID_CMP has been used to reduce this matrix to
!   tridiagonal form before calling SYMTRID_QRI2.
!
!
! Arguments
! _________
!
!   D            (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the diagonal elements of the tridiagonal matrix.
!
!                On exit, the eigenvalues.
!
!   E            (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the n-1 subdiagonal elements of the tridiagonal matrix.
!                E(n) is arbitrary and is used as workspace.
!
!                On exit, E has been destroyed
!
!                The size of E must verify:  size( E ) = size( D ) = n .
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit.
!                - FAILURE = true  :  indicates that the algorithm did not converge and
!                  that full accuracy was not attained in the Schur decomposition
!                  of the tridiagonal matrix.
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!                By default, the eigenvalues are not sorted.
!
!   MAXITER      (INPUT, OPTIONAL) integer(i4b)
!                MAXITER controls the maximum number of QR sweeps in the
!                Schur decomposition of the tridiagonal matrix.
!                The algorithm fails to converge if the number of QR sweeps
!                exceeds MAXITER * size(D). Convergence usually occurs in 
!                about 2 * size(D) QR sweeps.
!
!                The default is 30.
!
!
! Further Details
! _______________
!
!   The eigenvalues are computed by the Pal-Walker-Kahan variant of the implicit tridiagonal
!   QR algorithm described in the reference (1).
!
!   This subroutine is adapted from the routine  DSTERF in LAPACK. This subroutine is
!   not parallelized.
!
!   For further details, see:
!
!   (1) Parlett, B.N., 1998:
!           The Symmetric Eigenvalue Problem.
!           revised edition, SIAM, Philadelphia.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,    only : lamch
    use Utilities,        only : assert_eq, iminloc, lascl
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:), intent(inout) :: d, e
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(out)          :: failure
!
    character ,   intent(in), optional :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: eps, eps2, safmin, safmax, ssfmax, ssfmin, tridnorm,          &
                  tst, r, bb, s, c, p, rte, sigma, gamma, oldgam, oldc, alpha
!
    integer(i4b) :: n, m, l, i, iscale, nqrs, maxqrs
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='symtrid_qri2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(d),i4b)     ,   &
                    int(size(e),i4b)     ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) return
!
!   DETERMINE THE UNIT ROUNDOFF AND OVER/UNDERFLOW THRESHOLDS.
!
    eps    = fudge_eps*lamch( 'u' )
    eps2   = eps*eps
    safmin = lamch( 's' )
    safmax = one/safmin
    ssfmax = sqrt( safmax )/three
    ssfmin = sqrt( safmin )/eps2
!
!   COMPUTE MAX ELEMENT OF TRIDIAGONAL MATRIX.
!
    e(n) = zero
    tridnorm = abs( d(1_i4b) ) + abs( e(1_i4b) )
!
    do i = 2_i4b, n
        tridnorm = max( tridnorm, abs(d(i))+abs(e(i))+abs(e(i-1_i4b)) )
    end do
!
!   SCALE TRIDIAGONAL MATRIX IF NECESSARY.
!
    iscale = 0_i4b
!
    if ( tridnorm==zero ) then
!
        return
!
    else if ( tridnorm>ssfmax ) then
!
        iscale = 1_i4b
        call lascl( d(:), tridnorm, ssfmax )
        call lascl( e(:n-1_i4b), tridnorm, ssfmax )
!
    else if ( tridnorm<ssfmin ) then
!
        iscale = 2_i4b
        call lascl( d(:), tridnorm, ssfmin )
        call lascl( e(:n-1_i4b), tridnorm, ssfmin )
!
    end if
!
!   STORE THE ELEMENTS OF THE TRIDIAGONAL MATRIX IN REVERSED ORDER IF
!   THE DIAGONAL ELEMENTS ARE INCREASING IN MAGNITUDE ALONG THE DIAGONAL
!   IN ORDER TO IMPROVE THE ACCURACY OF THE QR ITERATIONS FOR GRADED MATRICES.
!
    if ( reverse(d(:)) ) then
        d(1_i4b:n)       = d(n:1_i4b:-1_i4b)
        e(1_i4b:n-1_i4b) = e(n-1_i4b:1_i4b:-1_i4b)
    end if
!
!   RENUMBER AND SQUARE THE ELEMENTS OF e(:) .
!
    e(:n) = ( eoshift( e(:n), -1 ) )**2
!
    nqrs = 0_i4b
    if ( present(maxiter) ) then
        maxqrs = n*merge( maxiter, 30_i4b , maxiter>=1_i4b )
    else
        maxqrs = n*30_i4b
    end if
!
!   QR ITERATIONS.
!
    l = n
!
    do
!
!       TEST FOR SPLITTING : LOOK FOR SMALL SUBDIAGONAL ELEMENT.
!
        do m = l, 2_i4b, -1_i4b
!
            tst = abs( e(m) )
!
            if ( tst<=eps2*abs( d(m)*d(m-1_i4b) ) + safmin ) exit
!
        end do            
!
        if ( m==l ) then
!
!           EIGENVALUE FOUND.
!
            l = l - 1_i4b
!
            if ( l>1_i4b ) cycle
!
            exit
!
        end if
!
!       CUTOFF FOR CONVERGENCE FAILURE.
!
        if ( nqrs>=maxqrs ) then
!
!           SET FAILURE INDICATOR.
!
            failure = true
!
!           IF FAILURE TO CONVERGE SET SMALLEST MAGNITUDE TERM
!           IN OFF-DIAGONAL TO ZERO AND CONTINUE ON.
!
            i = iminloc( e(m+1_i4b:l) ) + m
            e(i) = zero
            nqrs = 0_i4b
!
            cycle
!
        end if
!
        nqrs = nqrs + 1_i4b
!
!       FORM Wilkinson SHIFT.
!
        rte = sqrt( e(l) )
        sigma = ( d(l-1_i4b) - d(l) )/( two*rte )
!
        if ( abs(sigma)>one ) then
!
            p = one/sigma
            r = sigma*sqrt( one + p*p )
!
        else
!
            r = sign( sqrt( one + sigma*sigma ), sigma )
!
        end if
!
        sigma = d(l) - ( rte/( sigma + r ) )
!
        s = zero
        c = one
        gamma = d(m) - sigma
        p = gamma*gamma
!
!       INNER LOOP.
!
        do i = m, l-1_i4b
!
            bb = e(i+1_i4b)
            r = p + bb
            e(i) = s*r
            oldc = c
            c = p/r
            s = bb/r
            oldgam = gamma
            alpha = d(i+1_i4b)
            gamma = c*( alpha - sigma ) - s*oldgam
            d(i) = oldgam + ( alpha - gamma )
!
            if ( c/=zero ) then
                p = ( gamma*gamma )/c
            else
                p = oldc*bb
            end if
!
        end do
!
        e(l) = s*p
        d(l) = sigma + gamma
        e(m) = zero
!
    end do
!
!   UNDO SCALING IF NECESSARY.
!
    if ( iscale>0_i4b ) then
        c = merge( ssfmax, ssfmin, iscale==1_i4b )
        call lascl( d, c, tridnorm )
    end if
!
!   SORT EIGENVALUES IF REQUIRED.
!
    if ( present(sort) ) call eigval_sort( sort, d )
!
!
! END OF SUBROUTINE symtrid_qri2_b
! ________________________________
!
    end subroutine symtrid_qri2_b
!
! =========================================================================================
!
    subroutine symtrid_qri3_a( d, e, failure, mat, init_mat, sort, maxiter, max_francis_steps )
!
! Purpose
! _______
!
!   SYMTRID_QRI3 computes all eigenvalues and eigenvectors of a
!   symmetric n-by-n tridiagonal matrix using the implicit QR method.
!
!   The eigenvalues and eigenvectors of a full symmetric matrix can
!   also be found if SYMTRID_CMP and ORTHO_GEN_SYMTRID have been used 
!   to reduce this matrix to tridiagonal form before calling SYMTRID_QRI3.
!
!
! Arguments
! _________
!
!   D                   (INPUT/OUTPUT) real(stnd), dimension(:)
!                       On entry, the diagonal elements of the tridiagonal matrix.
!
!                       On exit, the eigenvalues.
!
!   E                   (INPUT/OUTPUT) real(stnd), dimension(:)
!                       On entry, the n-1 subdiagonal elements of the tridiagonal matrix.
!                       E(n) is arbitrary and is used as workspace.
!
!                       On exit, E has been destroyed.
!
!                       The size of E must verify:  size( E ) = size( D ) = n .
!
!   FAILURE             (OUTPUT) logical(lgl)
!                       On exit:
!
!                       - FAILURE = false :  indicates successful exit.
!                       - FAILURE = true  :  indicates that the algorithm did not converge and
!                         that full accuracy was not attained in the Schur decomposition
!                         of the tridiagonal matrix.
!
!   MAT                 (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                       On entry, if:
!
!                       - INIT_MAT is absent or if INIT_MAT = false, then MAT
!                         contains the orthogonal matrix used in the reduction 
!                         to tridiagonal form as returned by ORTHO_GEN_SYMTRID.
!                       - INIT_MAT is present and INIT_MAT = true, MAT is set to the
!                         identity matrix of order n.
!
!                       On exit, if FAILURE = false:
!
!                       - MAT contains the orthonormal eigenvectors of the
!                         original symmetric matrix if INIT_MAT is absent
!                         or if INIT_MAT = false,
!                       - MAT contains the orthonormal eigenvectors of the
!                         symmetric tridiagonal matrix if INIT_MAT is present
!                         and INIT_MAT = true.
!
!                       The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = size( D ) = n .
!
!   INIT_MAT            (INPUT, OPTIONAL) logical(lgl)
!                       If:
!
!                       - INIT_MAT = false: The subroutine computes eigenvalues and eigenvectors
!                         of the original symmetric matrix.  On entry, MAT must contain the
!                         orthogonal matrix used to reduce the original matrix
!                         to tridiagonal form.
!                       - INIT_MAT = true: The subroutine computes eigenvalues and eigenvectors of the
!                         tridiagonal matrix.  MAT is initialized to the identity matrix of order n.
!
!                       The default is false.
!
!   SORT                (INPUT, OPTIONAL) character
!                       Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                       or in descending order if SORT = 'D' or 'd'.
!                       The eigenvectors are reordered accordingly.
!
!                       By default, the eigenvalues are not sorted.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the
!                       Schur decomposition of the tridiagonal matrix.
!                       The algorithm fails to converge if the number of QR sweeps
!                       exceeds MAXITER * size(D). Convergence usually occurs in 
!                       about 2 * size(D) QR sweeps.
!
!                       The default is 30.
!
!   MAX_FRANCIS_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                       MAX_FRANCIS_STEPS controls the maximum number of Francis sets
!                       (e.g. QR sweeps) of Givens rotations which must be saved
!                       before applying them with a wavefront algorithm to accumulate
!                       the eigenvectors in the QR algorithm. MAX_FRANCIS_STEPS is a
!                       strictly positive integer, otherwise the default value is used.
!
!                       The default is 10.
!
!
! Further Details
! _______________
!
!   The eigenvalues and eigenvectors are computed by the implicit tridiagonal QR algorithm
!   described in the reference (1).
!
!   This subroutine is adapted from the routine DSTEQR in LAPACK with modifications
!   suggested in the reference (2). Furthermore, the computations of the eigenvctors
!   are parallelized if OPENMP is used.
!
!   SYMTRID_QRI3 is much more efficient than SYMTRID_QRI and only slightly less efficient
!   than SYMTRID_QRI2 (but more robust) for computing the full set of eigenvectors of a real
!   n-by-n symmetric tridiagonal matrix.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Van Zee, F.G., Van de Geijn, R., and Quintana-Orti, G., 2011:
!           Restructuring the QR Algorithm for High-Performance Application of Givens Rotations.
!           FLAME Working Note 60. The University of Texas at Austin, Department of Computer
!           Sciences. Technical Report TR-11-36.
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,         only : lamch
    use Char_Constants,        only : allocate_error
    use Utilities,             only : assert_eq, merror, iminloc, lascl, unit_matrix
#ifdef _F2003
    use Utilities,             only : realloc => mvalloc
#else
    use Utilities_With_Pnter,  only : realloc
#endif
#ifdef _BLAS
    use BLAS_interfaces,       only : rot
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:) :: mat
    real(stnd),   intent(inout), dimension(:)   :: d, e
!
    integer(i4b), intent(in), optional :: maxiter, max_francis_steps
!
    logical(lgl), intent(in), optional :: init_mat
    logical(lgl), intent(out)          :: failure
!
    character ,   intent(in), optional :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
#ifdef _F2003
    real(stnd), dimension(:,:), allocatable :: csmat, snmat
#else
    real(stnd), dimension(:,:), pointer     :: csmat, snmat
#endif
    real(stnd), dimension(size(d))          :: temp
    real(stnd)                              :: eps, eps2, safmin, safmax, ssfmax, ssfmin,        &
                                               tridnorm, r, g, f, b, s, c, p
!
#ifdef _F2003
    integer(i4b), dimension(:), allocatable :: deb, fin
#else
    integer(i4b), dimension(:), pointer     :: deb, fin
#endif
    integer(i4b)                            :: n, m, l, i, ip1, i1, i2, j, iscale, nqrs, maxqrs,    &
                                               nqrs2, maxqrs2, o, q, qstart, qend, nfrancis
    integer                                 :: iok
!
#ifdef _OPENMP2
    real(stnd), dimension(:,:), allocatable :: mat2
!
    integer(i4b) :: k, n2, j1, j2, num_threads, thread_num, jrem, joffset
!
    logical :: test_nopar, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='symtrid_qri3'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(d),i4b)     ,   &
                    int(size(e),i4b)     ,   &
                    name_proc )
!
    if ( present(init_mat) ) then
!
!       COMPUTE THE EIGENVECTORS AND EIGENVALUES OF THE TRIDIAGONAL MATRIX.
!
        if ( init_mat ) then
!
            call unit_matrix( mat )
!
        end if
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) return
!
!   DETERMINE THE UNIT ROUNDOFF AND OVER/UNDERFLOW THRESHOLDS.
!
    eps    = fudge_eps*lamch( 'u' )
    eps2   = eps*eps
    safmin = lamch( 's' )
    safmax = one/safmin
    ssfmax = sqrt( safmax )/three
    ssfmin = sqrt( safmin )/eps2
!
!   COMPUTE MAX ELEMENT OF TRIDIAGONAL MATRIX.
!
    e(n) = zero
    tridnorm = abs( d(1_i4b) ) + abs( e(1_i4b) )
!
    do i = 2_i4b, n
        tridnorm = max( tridnorm, abs(d(i))+abs(e(i))+abs(e(i-1_i4b)) )
    end do
!
!   SCALE TRIDIAGONAL MATRIX IF NECESSARY.
!
    iscale = 0_i4b
!
    if ( tridnorm==zero ) then
!
        return
!
    else if ( tridnorm>ssfmax ) then
!
        iscale = 1_i4b
        call lascl( d(:n), tridnorm, ssfmax )
        call lascl( e(:n-1_i4b), tridnorm, ssfmax )
!
    else if ( tridnorm<ssfmin ) then
!
        iscale = 2_i4b
        call lascl( d(:n), tridnorm, ssfmin )
        call lascl( e(:n-1_i4b), tridnorm, ssfmin )
!
    end if
!
!   CONVENIENT TO RENUMBER THE ELEMENTS OF e(:) .
!
    temp(:n) = eoshift( e(:n), -1 )
    e(:n) = temp(:n)
!
    nqrs = 0_i4b
    if ( present(maxiter) ) then
        maxqrs = n*merge( maxiter, 30_i4b , maxiter>=1_i4b )
    else
        maxqrs = n*30_i4b
    end if
!    
    if ( present(max_francis_steps) ) then
        nfrancis = merge( max_francis_steps,                                  &
                          max_francis_steps_default,                          &
                          max_francis_steps>=1_i4b .and. max_francis_steps<=n )
    else
        nfrancis = max_francis_steps_default
    end if
!    
    nqrs2   = 0_i4b
    maxqrs2 = nfrancis + 5_i4b
!
!   ALLOCATE WORK VARIABLES NEEDED TO STORE THE GIVENS ROTATIONS .
!
    allocate( csmat(n,maxqrs2),     &
              snmat(n,maxqrs2),     &
              deb(maxqrs2),         &
              fin(maxqrs2),         &
              stat = iok      )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    deb(:) = 1_i4b
    fin(:) = -1_i4b
!
#ifdef _OPENMP2
    i = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested()
!
    test_nopar  = ( ompparallel .and. .not.ompnested ) .or.      &
                  n*n<omp_limit                        .or.      &
                  n/i<omp_limit2                       .or.      &
                  i==1_i4b
!
    if ( test_nopar ) then
#endif
!
!       QR ITERATIONS.
!
        do l = n, 2_i4b, -1_i4b
!
            do
!
!               TEST FOR SPLITTING : LOOK FOR SMALL SUBDIAGONAL ELEMENT.
!
                do m = l, 2_i4b, -1_i4b
!
                    p = e(m)**2
!
                    if ( p<=eps2*abs( d(m)*d(m-1_i4b) ) + safmin ) exit
!
                end do            
!
!               CUTOFF FOR CONVERGENCE FAILURE.
!
                if ( nqrs>=maxqrs ) then
!
                    if ( m/=l ) then
!
!                       SET FAILURE INDICATOR.
!
                        failure = true
!
!                       IF FAILURE TO CONVERGE SET SMALLEST MAGNITUDE TERM
!                       IN OFF-DIAGONAL TO ZERO AND CONTINUE ON.
!
                        m = iminloc( abs( e(m+1_i4b:l))) + m
                        nqrs = 0_i4b
!
                    end if
!
                end if
!
!               EIGENVALUE FOUND.
!
                if ( m==l ) exit
!
                nqrs  = nqrs + 1_i4b
                nqrs2 = nqrs2 + 1_i4b
!
                if ( nqrs2>maxqrs2 ) then
!
                    maxqrs2 = maxqrs2 + 5_i4b
!
                    call realloc( deb, maxqrs2, iok )
!
                    if ( iok==0 ) then
                        call realloc( fin, maxqrs2, iok  )
                        if ( iok==0 ) then
                            call realloc( csmat, n, maxqrs2, iok  )
                            if ( iok==0 ) then
                                call realloc( snmat, n, maxqrs2, iok  )
                            end if
                        end if
                    end if
!
                    if ( iok/=0 ) then
                        call merror( name_proc//allocate_error )
                    end if
!
                end if
!
!               FORM Wilkinson SHIFT.
!
                g = ( d(l-1_i4b) - d(l) )/( two*e(l) )
!
                if ( abs(g)>one ) then
!
                    p = one/g
                    r = g*sqrt( one + p*p )
!
                else
!
                    r = sign( sqrt( one + g*g ), g )
!
                end if
!
                g = d(m) - d(l) + ( e(l)/( g + r ) )
!
                s = one
                c = one
                p = zero
!
!               INNER LOOP.
!
                do i = m, l-1_i4b
!
                    ip1 = i + 1_i4b
!
                    f = s*e(ip1)
!
                    if ( f==zero ) exit
!
                    b = c*e(ip1)
!
                    if ( abs(g)>=abs(f) ) then
!
                        s = f/g
                        r = sign( sqrt( one + s*s ), g )
                        c = one/r
                        s = c*s
                        e(i) = g*r
!
                    else
!
                        c = g/f
                        r = sign( sqrt( one + c*c ), f )
                        s = one/r
                        c = s*c
                        e(i) = f*r
!
                    end if
!
!                   STORE ROTATION.
!
                    csmat(i,nqrs2) = c
                    snmat(i,nqrs2) = s
!
!                   APPLY ROTATION.
!
                    g = d(i) - p
                    r = ( d(ip1) - g )*s + two*c*b
                    p = s*r
                    d(i) = g + p
                    g = c*r - b
!
                end do
!
                d(i) = d(i) - p
                e(i) = g
                e(m) = zero
!
                deb(nqrs2) = m
                fin(nqrs2) = i - 1_i4b
!
            end do
!
            if ( nqrs2>=nfrancis .or. l==2_i4b  ) then
!
!               APPLY THE PREVIOUS FRANCIS SETS OF GIVENS ROTATIONS.
!
                qstart = maxval( deb(:nqrs2) ) + nqrs2 - 1_i4b
                qend   = minval( fin(:nqrs2) )
!
!               DETERMINE WHAT ALGORITHM TO USE.
!
                if ( qstart>=qend .or. nqrs2==1_i4b ) then
!
!                   APPLY THE GIVENS ROTATIONS WITH A SEQUENTIAL ALGORITHM.
!
                    do j = 1_i4b, nqrs2
!
                        i1 = deb(j)
                        i2 = fin(j)
!
#ifdef _BLAS
                        do i = i1, i2
!
                            call rot( n, mat(:n,i), 1_i4b, mat(:n,i+1_i4b), 1_i4b, csmat(i,j), snmat(i,j) )
!
                        end do
#else
                        temp(:n) = mat(:n,i1)
!
                        do i = i1, i2
!
                            ip1 = i + 1_i4b
!
                            c = csmat(i,j)
                            s = snmat(i,j)
                            mat(:n,i) =  c*temp(:n) + s*mat(:n,ip1)
                            temp(:n)  = -s*temp(:n) + c*mat(:n,ip1)
!
                        end do
!
                        mat(:n,i) = temp(:n)
#endif
!
                    end do
!
                else if ( nqrs2>1_i4b ) then
!
!                   APPLY THE GIVENS ROTATIONS WITH A WAVEFRONT ALGORITHM
!                   IN ORDER TO MINIMIZE CACHE MISSES.
!
!                   STARTUP PHASE.
!
                    o  = 0_i4b
                    i2 = qstart - nqrs2 + 1_i4b
!
                    do q = i2, qstart - 1_i4b
!
                        o = o + 1_i4b
!
                        i1 = deb(o)
!
                        if ( i1<i2 ) then
!
#ifdef _BLAS
                            do i = i1, i2 - 1_i4b
!
                                call rot( n, mat(:n,i), 1_i4b, mat(:n,i+1_i4b), 1_i4b, csmat(i,o), snmat(i,o) )
!
                            end do
#else
                            temp(:n) = mat(:n,i1)
!
                            do i = i1, i2 - 1_i4b
!
                                ip1 = i + 1_i4b
!
                                c = csmat(i,o)
                                s = snmat(i,o)
                                mat(:n,i) =  c*temp(:n) + s*mat(:n,ip1)
                                temp(:n)  = -s*temp(:n) + c*mat(:n,ip1)
!
                            end do
!
                            mat(:n,i) = temp(:n)
#endif
!
                        end if
!
                        i = q
!
#ifdef _BLAS
                        do j = 1_i4b, o
!
                            call rot( n, mat(:n,i+1_i4b), 1_i4b, mat(:n,i), 1_i4b, csmat(i,j), -snmat(i,j) )
!
                            i = i - 1_i4b
!
                        end do
#else
                        temp(:n) = mat(:n,i+1_i4b)
!
                        do j = 1_i4b, o
!
                            c = csmat(i,j)
                            s = snmat(i,j)
                            mat(:n,i+1_i4b) = -s*mat(:n,i) + c*temp(:n)
                            temp(:n)        =  c*mat(:n,i) + s*temp(:n)
!
                            i = i - 1_i4b
!
                        end do
!
                        mat(:n,i+1_i4b) = temp(:n)
#endif
!
                    end do
!
!                    i2 = qstart
!
!                    do j = 1_i4b, nqrs2
!
!                        i1 = deb(j)
!                        i2 = i2 - 1_i4b
!
!#ifdef _BLAS
!                        do i = i1, i2
!
!                            call rot( n, mat(:n,i), 1_i4b, mat(:n,i+1_i4b), 1_i4b, csmat(i,j), snmat(i,j) )
!
!                        end do
!#else
!                        temp(:n) = mat(:n,i1)
!
!                        do i = i1, i2
!
!                            ip1 = i + 1_i4b
!
!                            c = csmat(i,j)
!                            s = snmat(i,j)
!                            mat(:n,i) =  c*temp(:n) + s*mat(:n,ip1)
!                            temp(:n)  = -s*temp(:n) + c*mat(:n,ip1)
!
!                        end do
!
!                        mat(:n,i) = temp(:n)
!#endif
!
!                    end do
!
!                   PIPELINE PHASE.
!
                    do q = qstart, qend
!
                        i = q
!
#ifdef _BLAS
                        do j = 1_i4b, nqrs2
!
                            call rot( n, mat(:n,i), 1_i4b, mat(:n,i+1_i4b), 1_i4b, csmat(i,j), snmat(i,j) )
!
                            i = i - 1_i4b
!
                        end do
#else
                        temp(:n) = mat(:n,i+1_i4b)
!
                        do j = 1_i4b, nqrs2
!
                            c = csmat(i,j)
                            s = snmat(i,j)
                            mat(:n,i+1_i4b) = -s*mat(:n,i) + c*temp(:n)
                            temp(:n)        =  c*mat(:n,i) + s*temp(:n)
!
                            i = i - 1_i4b
!
                        end do
!
                        mat(:n,i+1_i4b) = temp(:n)
#endif
!
                    end do
!
!                   SHUTDOWN PHASE.
!
                    i1 = qend + 1_i4b
!
                    do j = 1_i4b, nqrs2
!
                        i2 = fin(j)
!
                        if ( i1<=i2 ) then
!
#ifdef _BLAS
                            do i = i1, i2
!
                                call rot( n, mat(:n,i), 1_i4b, mat(:n,i+1_i4b), 1_i4b, csmat(i,j), snmat(i,j) )
!
                            end do
#else
                            temp(:n) = mat(:n,i1)
!
                            do i = i1, i2
!
                                ip1 = i + 1_i4b
!
                                c = csmat(i,j)
                                s = snmat(i,j)
                                mat(:n,i) =  c*temp(:n) + s*mat(:n,ip1)
                                temp(:n)  = -s*temp(:n) + c*mat(:n,ip1)
!
                            end do
!
                            mat(:n,i) = temp(:n)
#endif
!
                        end if
!
                        i1 = i1 - 1_i4b
!
                    end do
!
                end if
!
                nqrs2 = 0_i4b
!
            end if
!
        end do
!
#ifdef _OPENMP2
    else
!
!$OMP  PARALLEL PRIVATE(l,i,ip1,i1,i2,j,j1,j2,k,n2,num_threads,thread_num,            &
!$OMP                   jrem,joffset,o,q,qstart,qend,c,s,temp,mat2,iok,p,r,g,f,b),    &
!$OMP           SHARED(failure,nqrs,nqrs2,maxqrs2)
!
!       DISTRIBUTE THE WORK AMONG THE THREADS.
!
        num_threads = omp_get_num_threads()
        thread_num  = omp_get_thread_num()
!
        jrem = mod( n, num_threads )
        n2   = ( n - jrem )/num_threads
!
        if ( thread_num<jrem ) then
            n2      = n2 + 1_i4b
            joffset = 1_i4b
        else
            joffset = jrem + 1_i4b
        end if
!
        j1 = thread_num*n2 + joffset
        j2 = j1 + n2 - 1_i4b
!
!       ALLOCATE WORK VARIABLE TO STORE A LOCAL COPY OF A
!       PORTION OF THE INPUT MATRIX.
!
        allocate( mat2(n2,n), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       MAKE A LOCAL COPY OF A PORTION OF THE INPUT MATRIX.
!
        mat2(:n2,:n) = mat(j1:j2,:n)
!
!       QR ITERATIONS.
!
        do l = n, 2_i4b, -1_i4b
!
            if ( thread_num==0_i4b ) then
!
                k = nqrs2
!
                do
!
!                   TEST FOR SPLITTING : LOOK FOR SMALL SUBDIAGONAL ELEMENT.
!
                    do m = l, 2_i4b, -1_i4b
!
                        p = e(m)**2
!
                        if ( p<=eps2*abs( d(m)*d(m-1_i4b) ) + safmin ) exit
!
                    end do            
!
!                   CUTOFF FOR CONVERGENCE FAILURE.
!
                    if ( nqrs>=maxqrs ) then
!
                        if ( m/=l ) then
!
!                           SET FAILURE INDICATOR.
!
                            failure = true
!
!                           IF FAILURE TO CONVERGE SET SMALLEST MAGNITUDE TERM
!                           IN OFF-DIAGONAL TO ZERO AND CONTINUE ON.
!
                            m = iminloc( abs( e(m+1_i4b:l))) + m
                            nqrs = 0_i4b
!
                        end if
!
                    end if
!
!                   EIGENVALUE FOUND.
!
                    if ( m==l ) exit
!
                    nqrs = nqrs + 1_i4b
                    k = k + 1_i4b
!
                    if ( k>maxqrs2 ) then
!
                        maxqrs2 = maxqrs2 + 5_i4b
!
                        call realloc( deb, maxqrs2, iok )
!
                        if ( iok==0 ) then
                            call realloc( fin, maxqrs2, iok  )
                            if ( iok==0 ) then
                                call realloc( csmat, n, maxqrs2, iok  )
                                if ( iok==0 ) then
                                    call realloc( snmat, n, maxqrs2, iok  )
                                end if
                            end if
                        end if
!
                        if ( iok/=0 ) then
                            call merror( name_proc//allocate_error )
                        end if
!
                    end if
!
!                   FORM Wilkinson SHIFT.
!
                    g = ( d(l-1_i4b) - d(l) )/( two*e(l) )
!
                    if ( abs(g)>one ) then
!
                        p = one/g
                        r = g*sqrt( one + p*p )
!
                    else
!
                        r = sign( sqrt( one + g*g ), g )
!
                    end if
!
                    g = d(m) - d(l) + ( e(l)/( g + r ) )
!
                    s = one
                    c = one
                    p = zero
!
!                   INNER LOOP.
!
                    do i = m, l-1_i4b
!
                        ip1 = i + 1_i4b
!
                        f = s*e(ip1)
!
                        if ( f==zero ) exit
!
                        b = c*e(ip1)
!
                        if ( abs(g)>=abs(f) ) then
!
                            s    = f/g
                            r    = sign( sqrt( one + s*s ), g )
                            c    = one/r
                            s    = c*s
                            e(i) = g*r
!
                        else
!
                            c    = g/f
                            r    = sign( sqrt( one + c*c ), f )
                            s    = one/r
                            c    = s*c
                            e(i) = f*r
!
                        end if
!
!                       STORE ROTATION.
!
                        csmat(i,k) = c
                        snmat(i,k) = s
!
!                       APPLY ROTATION.
!
                        g    = d(i) - p
                        r    = ( d(ip1) - g )*s + two*c*b
                        p    = s*r
                        d(i) = g + p
                        g    = c*r - b
!
                    end do
!
                    d(i) = d(i) - p
                    e(i) = g
                    e(m) = zero
!
                    deb(k) = m
                    fin(k) = i - 1_i4b
!
                end do
!
                nqrs2 = k
!
            end if
!
!$OMP BARRIER
!
            if ( nqrs2>=nfrancis .or. l==2_i4b  ) then
!
!               APPLY THE PREVIOUS FRANCIS SETS OF GIVENS ROTATIONS.
!
                qstart = maxval( deb(:nqrs2) ) + nqrs2 - 1_i4b
                qend   = minval( fin(:nqrs2) )
!
!               DETERMINE WHAT ALGORITHM TO USE.
!
                if ( qstart>=qend .or. nqrs2==1_i4b ) then
!
!                   APPLY THE GIVENS ROTATIONS WITH A SEQUENTIAL ALGORITHM.
!
                    do j = 1_i4b, nqrs2
!
                        i1 = deb(j)
                        i2 = fin(j)
!
#ifdef _BLAS
                        do i = i1, i2
!
                            call rot( n2, mat2(:n2,i), 1_i4b, mat2(:n2,i+1_i4b), 1_i4b, csmat(i,j), snmat(i,j) )
!
                        end do
#else
                        temp(:n2) = mat2(:n2,i1)
!
                        do i = i1, i2
!
                            ip1 = i + 1_i4b
!
                            c = csmat(i,j)
                            s = snmat(i,j)
                            mat2(:n2,i) = c*temp(:n2) + s*mat2(:n2,ip1)
                            temp(:n2) = -s*temp(:n2) + c*mat2(:n2,ip1)
!
                        end do
!
                        mat2(:n2,i) = temp(:n2)
#endif
!
                    end do
!
                else if ( nqrs2>1_i4b ) then
!
!                   APPLY THE GIVENS ROTATIONS WITH A WAVEFRONT ALGORITHM
!                   IN ORDER TO MINIMIZE CACHE MISSES.
!
!                   STARTUP PHASE.
!
                    o  = 0_i4b
                    i2 = qstart - nqrs2 + 1_i4b
!
                    do q = i2, qstart - 1_i4b
!
                        o = o + 1_i4b
!
                        i1 = deb(o)
!
                        if ( i1<i2 ) then
!
#ifdef _BLAS
                            do i = i1, i2 - 1_i4b
!
                                call rot( n2, mat2(:n2,i), 1_i4b, mat2(:n2,i+1_i4b), 1_i4b, csmat(i,o), snmat(i,o) )
!
                            end do
#else
                            temp(:n2) = mat2(:n2,i1)
!
                            do i = i1, i2 - 1_i4b
!
                                ip1 = i + 1_i4b
!
                                c = csmat(i,o)
                                s = snmat(i,o)
                                mat2(:n2,i) =  c*temp(:n2) + s*mat2(:n2,ip1)
                                temp(:n2)   = -s*temp(:n2) + c*mat2(:n2,ip1)
!
                            end do
!
                            mat2(:n2,i) = temp(:n2)
#endif
!
                        end if
!
                        i = q
!
#ifdef _BLAS
                        do j = 1_i4b, o
!
                            call rot( n2, mat2(:n2,i+1_i4b), 1_i4b, mat2(:n2,i), 1_i4b, csmat(i,j), -snmat(i,j) )
!
                            i = i - 1_i4b
!
                        end do
#else
                        temp(:n2) = mat2(:n2,i+1_i4b)
!
                        do j = 1_i4b, o
!
                            c = csmat(i,j)
                            s = snmat(i,j)
                            mat2(:n2,i+1_i4b) = -s*mat2(:n2,i) + c*temp(:n2)
                            temp(:n2)         =  c*mat2(:n2,i) + s*temp(:n2)
!
                            i = i - 1_i4b
!
                        end do
!
                        mat2(:n2,i+1_i4b) = temp(:n2)
#endif
!
                    end do
!
!                    i2 = qstart
!
!                    do j = 1_i4b, nqrs2
!
!                        i1 = deb(j)
!                        i2 = i2 - 1_i4b
!
!#ifdef _BLAS
!                        do i = i1, i2
!
!                            call rot( n2, mat2(:n2,i), 1_i4b, mat2(:n2,i+1_i4b), 1_i4b, csmat(i,j), snmat(i,j) )
!
!                        end do
!#else
!                        temp(:n2) = mat2(:n2,i1)
!
!                        do i = i1, i2
!
!                            ip1 = i + 1_i4b
!
!                            c = csmat(i,j)
!                            s = snmat(i,j)
!                            mat2(:n2,i) = c*temp(:n2) + s*mat2(:n2,ip1)
!                            temp(:n2) = -s*temp(:n2) + c*mat2(:n2,ip1)
!
!                        end do
!
!                        mat2(:n2,i) = temp(:n2)
!#endif
!
!                    end do
!
!                   PIPELINE PHASE.
!
                    do q = qstart, qend
!
                        i = q
!
#ifdef _BLAS
                        do j = 1_i4b, nqrs2
!
                            call rot( n2, mat2(:n2,i), 1_i4b, mat2(:n2,i+1_i4b), 1_i4b, csmat(i,j), snmat(i,j) )
!
                            i = i - 1_i4b
!
                        end do
#else
                        temp(:n2) = mat2(:n2,i+1_i4b)
!
                        do j = 1_i4b, nqrs2
!
                            c = csmat(i,j)
                            s = snmat(i,j)
                            mat2(:n2,i+1_i4b) = -s*mat2(:n2,i) + c*temp(:n2)
                            temp(:n2)         = c*mat2(:n2,i)  + s*temp(:n2)
!
                            i = i - 1_i4b
!
                        end do
!
                        mat2(:n2,i+1_i4b) = temp(:n2)
#endif
!
                    end do
!
!                   SHUTDOWN PHASE.
!
                    i1 = qend + 1_i4b
!
                    do j = 1_i4b, nqrs2
!
                        i2 = fin(j)
!
                        if ( i1<=i2 ) then
!
#ifdef _BLAS
                            do i = i1, i2
!
                                call rot( n2, mat2(:n2,i), 1_i4b, mat2(:n2,i+1_i4b), 1_i4b, csmat(i,j), snmat(i,j) )
!
                            end do
#else
                            temp(:n2) = mat2(:n2,i1)
!
                            do i = i1, i2
!
                                ip1 = i + 1_i4b
!
                                c = csmat(i,j)
                                s = snmat(i,j)
                                mat2(:n2,i) = c*temp(:n2) + s*mat2(:n2,ip1)
                                temp(:n2) = -s*temp(:n2) + c*mat2(:n2,ip1)
!
                            end do
!
                            mat2(:n2,i) = temp(:n2)
#endif
!
                        end if
!
                        i1 = i1 - 1_i4b
!
                    end do
!
                end if
!
!$OMP BARRIER
!
!$OMP MASTER
                nqrs2 = 0_i4b
!$OMP END MASTER
!
            end if
!
!$OMP BARRIER
!
       end do
!
       mat(j1:j2,:n) = mat2(:n2,:n)
!
!      DEALLOCATE WORK ARRAY.
!
       deallocate( mat2 )
!
!$OMP END PARALLEL
!
    end if
#endif
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( csmat, snmat, deb, fin )
!
!   UNDO SCALING IF NECESSARY.
!
    if ( iscale>0_i4b ) then
        c = merge( ssfmax, ssfmin, iscale==1_i4b )
        call lascl( d, c, tridnorm )
    end if
!
!   ORDER EIGENVALUES AND EIGENVECTORS IF REQUIRED.
!
    if ( present(sort) )  call eig_sort_a( sort, d, mat )
!
!
! END OF SUBROUTINE symtrid_qri3_a
! ________________________________
!
    end subroutine symtrid_qri3_a
!
! =========================================================================================
!
    subroutine symtrid_qri3_b( d, e, failure, sort, maxiter )
!
! Purpose
! _______
!
!   SYMTRID_QRI3 computes all eigenvalues of a symmetric n-by-n tridiagonal
!   matrix using the implicit QR method.
!   The eigenvalues of a full symmetric matrix can also be found if SYMTRID_CMP
!   has been used to reduce this matrix to tridiagonal form before calling
!   SYMTRID_QRI3.
!
!
! Arguments
! _________
!
!   D           (INPUT/OUTPUT) real(stnd), dimension(:)
!               On entry, the diagonal elements of the tridiagonal matrix.
!
!               On exit, the eigenvalues.
!
!   E           (INPUT/OUTPUT) real(stnd), dimension(:)
!               On entry, the n-1 subdiagonal elements of the tridiagonal matrix.
!               E(n) is arbitrary and is used as workspace.
!
!               On exit, E has been destroyed.
!
!               The size of E must verify:  size( E ) = size( D ) = n .
!
!   FAILURE     (OUTPUT) logical(lgl)
!               On exit:
!
!               - FAILURE = false :  indicates successful exit.
!               - FAILURE = true  :  indicates that the algorithm did not converge and
!                 that full accuracy was not attained in the Schur decomposition
!                 of the tridiagonal matrix.
!
!   SORT        (INPUT, OPTIONAL) character
!               Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!               or in descending order if SORT = 'D' or 'd'.
!
!               By default, the eigenvalues are not sorted.
!
!   MAXITER     (INPUT, OPTIONAL) integer(i4b)
!               MAXITER controls the maximum number of QR sweeps in the
!               Schur decomposition of the tridiagonal matrix.
!               The algorithm fails to converge if the number of QR sweeps
!               exceeds MAXITER * size(D). Convergence usually occurs in 
!               about 2 * size(D) QR sweeps.
!
!               The default is 30.
!
!
! Further Details
! _______________
!
!   The eigenvalues are computed by the implicit tridiagonal QR algorithm
!   described in the reference (1).
!
!   This subroutine is adapted from the routine DSTEQR in LAPACK. This subroutine
!   is not parallelized.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,    only : lamch
    use Utilities,        only : assert_eq, iminloc, lascl
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:), intent(inout) :: d, e
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(out)          :: failure
!
    character ,   intent(in), optional :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: eps, eps2, safmin, safmax, ssfmax, ssfmin,    &
                  tridnorm, tst, r, g, f, b, s, c, p, rt1, rt2
!
    integer(i4b) :: n, m, l, i, ip1, iscale, nqrs, maxqrs
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='symtrid_qri3'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(d),i4b)     ,   &
                    int(size(e),i4b)     ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) return
!
!   DETERMINE THE UNIT ROUNDOFF AND OVER/UNDERFLOW THRESHOLDS.
!
    eps    = fudge_eps*lamch( 'u' )
    eps2   = eps*eps
    safmin = lamch( 's' )
    safmax = one/safmin
    ssfmax = sqrt( safmax )/three
    ssfmin = sqrt( safmin )/eps2
!
!   COMPUTE MAX ELEMENT OF TRIDIAGONAL MATRIX.
!
    e(n) = zero
    tridnorm = abs( d(1_i4b) ) + abs( e(1_i4b) )
!
    do i = 2_i4b, n
        tridnorm = max( tridnorm, abs(d(i))+abs(e(i))+abs(e(i-1_i4b)) )
    end do
!
!   SCALE TRIDIAGONAL MATRIX IF NECESSARY.
!
    iscale = 0_i4b
!
    if ( tridnorm==zero ) then
!
        return
!
    else if ( tridnorm>ssfmax ) then
!
        iscale = 1_i4b
        call lascl( d(:n), tridnorm, ssfmax )
        call lascl( e(:n-1_i4b), tridnorm, ssfmax )
!
    else if ( tridnorm<ssfmin ) then
!
        iscale = 2_i4b
        call lascl( d(:n), tridnorm, ssfmin )
        call lascl( e(:n-1_i4b), tridnorm, ssfmin )
!
    end if
!
!   STORE THE ELEMENTS OF THE TRIDIAGONAL MATRIX IN REVERSED ORDER IF
!   THE DIAGONAL ELEMENTS ARE INCREASING IN MAGNITUDE ALONG THE DIAGONAL
!   IN ORDER TO IMPROVE THE ACCURACY OF THE QR ITERATIONS FOR GRADED MATRICES.
!
    if ( reverse(d(:)) ) then
        d(1_i4b:n)       = d(n:1_i4b:-1_i4b)
        e(1_i4b:n-1_i4b) = e(n-1_i4b:1_i4b:-1_i4b)
    end if
!
!   CONVENIENT TO RENUMBER THE ELEMENTS OF e(:) .
!
    e(:n) = eoshift( e(:n), -1 )
!
    nqrs = 0_i4b
    if ( present(maxiter) ) then
        maxqrs = n*merge( maxiter, 30_i4b , maxiter>=1_i4b )
    else
        maxqrs = n*30_i4b
    end if
!
!   QR ITERATIONS.
!
    l = n
!
    do
!
!       TEST FOR SPLITTING : LOOK FOR SMALL SUBDIAGONAL ELEMENT.
!
        do m = l, 2_i4b, -1_i4b
!
            tst = e(m)**2
!
            if ( tst<=eps2*abs( d(m)*d(m-1_i4b) ) + safmin ) exit
!
        end do            
!
        if ( m==l ) then
!
!           EIGENVALUE FOUND.
!
            l = l - 1_i4b
!
            if ( l>1_i4b ) cycle
!
            exit
!
        else if ( m==l-1_i4b ) then
!
!           IF REMAINING MATRIX IS 2-BY-2, USE LAE2
!           TO COMPUTE ITS EIGENVALUES.
!
            call lae2( d(m), e(l), d(l), rt1, rt2 )
!
            d(m) = rt1
            d(l) = rt2
            l = l - 2_i4b
!
            if ( l>1_i4b ) cycle
!
            exit
!
        end if
!
!       CUTOFF FOR CONVERGENCE FAILURE.
!
        if ( nqrs>=maxqrs ) then
!
!           SET FAILURE INDICATOR.
!
            failure = true
!
!           IF FAILURE TO CONVERGE SET SMALLEST MAGNITUDE TERM
!           IN OFF-DIAGONAL TO ZERO AND CONTINUE ON.
!
            i = iminloc( abs(e(m+1_i4b:l)) ) + m
            e(i) = zero
            nqrs = 0_i4b
!
            cycle
!
        end if
!
        nqrs = nqrs + 1_i4b
!
!       FORM Wilkinson SHIFT.
!
        g = ( d(l-1_i4b) - d(l) )/( two*e(l) )
!
        if ( abs(g)>one ) then
!
            p = one/g
            r = g*sqrt( one + p*p )
!
        else
!
            r = sign( sqrt( one + g*g ), g )
!
        end if
!
        g = d(m) - d(l) + ( e(l)/( g + r ) )
!
        s = one
        c = one
        p = zero
!
!       INNER LOOP.
!
        do i = m, l-1_i4b
!
            ip1 = i + 1_i4b
!
            f = s*e(ip1)
!
            if ( f==zero ) exit
!
            b = c*e(ip1)
!
            if ( abs(g)>=abs(f) ) then
!
                s = f/g
                r = sign( sqrt( one + s*s ), g )
                c = one/r
                s = c*s
                e(i) = g*r
!
            else
!
                c = g/f
                r = sign( sqrt( one + c*c ), f )
                s = one/r
                c = s*c
                e(i) = f*r
!
            end if
!
!           APPLY ROTATION.
!
            g = d(i) - p
            r = ( d(ip1) - g )*s + two*c*b
            p = s*r
            d(i) = g + p
            g = c*r - b
!
        end do
!
        d(i) = d(i) - p
        e(i) = g
        e(m) = zero
!
    end do
!
!   UNDO SCALING IF NECESSARY.
!
    if ( iscale>0_i4b ) then
        c = merge( ssfmax, ssfmin, iscale==1_i4b )
        call lascl( d, c, tridnorm )
    end if
!
!   SORT EIGENVALUES IF REQUIRED.
!
    if ( present(sort) ) call eigval_sort( sort, d )
!
!
! END OF SUBROUTINE symtrid_qri3_b
! ________________________________
!
    end subroutine symtrid_qri3_b
!
! =========================================================================================
!
    subroutine lae2( a, b, c, rt1, rt2 )
!
! Purpose
! _______
!
!   LAE2  computes the eigenvalues of a 2-by-2 symmetric matrix
!
!                     [  A   B  ]
!
!                     [  B   C  ]
!
!   On return, RT1 is the eigenvalue of larger absolute value, and RT2
!   is the eigenvalue of smaller absolute value.
!
!
! Arguments
! _________
!
!   A       (INPUT) real(stnd)
!           The (1,1) element of the 2-by-2 matrix.
!
!   B       (INPUT) real(stnd)
!           The (1,2) and (2,1) elements of the 2-by-2 matrix.
!
!   C       (INPUT) real(stnd)
!           The (2,2) element of the 2-by-2 matrix.
!
!   RT1     (OUTPUT) real(stnd)
!           The eigenvalue of larger absolute value.
!
!   RT2     (OUTPUT) real(stnd)
!           The eigenvalue of smaller absolute value.
!
!
! Further Details
! _______________
!
!   RT1 is accurate to a few ulps barring over/underflow.
!
!   RT2 may be inaccurate if there is massive cancellation in the
!   determinant A * C - B * B; higher precision or correctly rounded or
!   correctly truncated arithmetic would be needed to compute RT2
!   accurately in all cases.
!
!   Overflow is possible only if RT1 is within a factor of 5 of overflow.
!   Underflow is harmless if the input data is 0 or exceeds
!   underflow_threshold / macheps.
!
!   This subroutine is translated from the routine  DLAE2 in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)  :: a, b, c
    real(stnd), intent(out) :: rt1, rt2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: ab, acmn, acmx, adf, df, rt, sm, tb
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   COMPUTE THE EIGENVALUES
!
    sm = a + c
    df = a - c
    adf = abs( df )
    tb = b + b
    ab = abs( tb )
!
    if ( abs(a)>abs(c) ) then
!
        acmx = a
        acmn = c
!
    else
!
        acmx = c
        acmn = a
!
    end if
!
    if ( adf>ab ) then
!
        rt = adf*sqrt( one + ( ab/adf )**2 )
!
    else if ( adf<ab ) then
!
        rt = ab*sqrt( one + ( adf/ab )**2 )
!
    else
!
!       INCLUDES CASE ab=adf=0 .
!
        rt = ab*sqrt( two )
!
    end if
!
    if ( sm<zero ) then
!
        rt1 = half*( sm - rt )
!
!       ORDER OF EXECUTION IMPORTANT.
!       TO GET FULLY ACCURATE SMALLER EIGENVALUE,
!       NEXT LINE NEEDS TO BE EXECUTED IN HIGHER PRECISION.
!
!       rt2 = ( acmx/rt1 )*acmn - ( b/rt1 )*b
        rt2 = real( acmx/rt1, extd )*real( acmn, extd ) - real( b/rt1, extd )*real( b, extd )
!
    else if ( sm>zero ) then
!
        rt1 = half*( sm + rt )
!
!       ORDER OF EXECUTION IMPORTANT.
!       TO GET FULLY ACCURATE SMALLER EIGENVALUE,
!       NEXT LINE NEEDS TO BE EXECUTED IN HIGHER PRECISION.
!
!       rt2 = ( acmx/rt1 )*acmn - ( b/rt1 )*b
        rt2 = real( acmx/rt1, extd )*real( acmn, extd ) - real( b/rt1, extd )*real( b, extd )
!
    else
!
!       INCLUDES CASE rt1 = rt2 = 0 .
!
        rt1 = half*rt
        rt2 = -half*rt
!
    end if
!
!
! END OF SUBROUTINE lae2
! ______________________
!
    end subroutine lae2
!
! =========================================================================================
!
    subroutine laev2( a, b, c, rt1, rt2, cs1, sn1 )
!
! Purpose
! _______
!
!   LAEV2 computes the eigendecomposition of a 2-by-2 symmetric matrix
!
!                     [  A   B  ]
!
!                     [  B   C  ]
!
!   On return, RT1 is the eigenvalue of larger absolute value, RT2 is the
!   eigenvalue of smaller absolute value, and (CS1,SN1) is the unit right
!   eigenvector for RT1, giving the decomposition
!
!        [+CS1  +SN1 ] [  A   B  ] [ +CS1 -SN1 ]  =  [ RT1  000 ]
!
!        [-SN1  +CS1 ] [  B   C  ] [ +SN1 +CS1 ]  _  [ 000  RT2 ].
!
!
! Arguments
! _________
!
!   A       (INPUT) real(stnd)
!           The (1,1) element of the 2-by-2 matrix.
!
!   B       (INPUT) real(stnd)
!           The (1,2) and (2,1) elements of the 2-by-2 matrix.
!
!   C       (INPUT) real(stnd)
!           The (2,2) element of the 2-by-2 matrix.
!
!   RT1     (OUTPUT) real(stnd)
!           The eigenvalue of larger absolute value.
!
!   RT2     (OUTPUT) real(stnd)
!           The eigenvalue of smaller absolute value.
!
!   CS1     (OUTPUT) real(stnd)
!
!   SN1     (OUTPUT) real(stnd)
!           The vector (CS1, SN1) is a unit right eigenvector for RT1.
!
!
! Further Details
! _______________
!
!   RT1 is accurate to a few ulps barring over/underflow.
!
!   RT2 may be inaccurate if there is massive cancellation in the
!   determinant A * C - B * B; higher precision or correctly rounded or
!   correctly truncated arithmetic would be needed to compute RT2
!   accurately in all cases.
!
!   CS1 and SN1 are accurate to a few ulps barring over/underflow.
!
!   Overflow is possible only if RT1 is within a factor of 5 of overflow.
!   Underflow is harmless if the input data is 0 or exceeds
!   underflow_threshold / macheps.
!
!   This subroutine is translated from the routine  DLAE2 in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)  :: a, b, c
    real(stnd), intent(out) :: rt1, rt2, cs1, sn1
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: ab, acmn, acmx, acs, adf, cs, ct, df, rt, sm, tb, tn
!
    integer :: sgn1, sgn2
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   COMPUTE THE EIGENVALUES.
!
    sm = a + c
    df = a - c
    adf = abs( df )
    tb = b + b
    ab = abs( tb )
!
    if ( abs(a)>abs(c) ) then
!
        acmx = a
        acmn = c
!
    else
!
        acmx = c
        acmn = a
!
    end if
!
    if ( adf>ab ) then
!
        rt = adf*sqrt( one + ( ab/adf )**2 )
!
    else if ( adf<ab ) then
!
        rt = ab*sqrt( one + ( adf/ab )**2 )
!
    else
!
!       INCLUDES CASE ab = adf = 0 .
!
        rt = ab*sqrt( two )
!
    end if
!
    if ( sm<zero ) then
!
        rt1 = half*( sm - rt )
        sgn1 = -1
!
!       ORDER OF EXECUTION IMPORTANT.
!       TO GET FULLY ACCURATE SMALLER EIGENVALUE,
!       NEXT LINE NEEDS TO BE EXECUTED IN HIGHER PRECISION.
!
!       rt2 = ( acmx/rt1 )*acmn - ( b/rt1 )*b
        rt2 = real( acmx/rt1, extd )*real( acmn, extd ) - real( b/rt1, extd )*real( b, extd )
!
    else if ( sm>zero ) then
!
        rt1 = half*( sm + rt )
        sgn1 = 1
!
!       ORDER OF EXECUTION IMPORTANT.
!       TO GET FULLY ACCURATE SMALLER EIGENVALUE,
!       NEXT LINE NEEDS TO BE EXECUTED IN HIGHER PRECISION.
!
!       rt2 = ( acmx/rt1 )*acmn - ( b/rt1 )*b
        rt2 = real( acmx/rt1, extd )*real( acmn, extd ) - real( b/rt1, extd )*real( b, extd )
!
    else
!
!       INCLUDES CASE rt1 = rt2 = 0 .
!
        rt1 = half*rt
        rt2 = -half*rt
        sgn1 = 1
!
    end if
!
!   COMPUTE THE EIGENVECTOR.
!
    if ( df>=zero ) then
!
        cs = df + rt
        sgn2 = 1
!
    else
!
        cs = df - rt
        sgn2 = -1
!
    end if
!
    acs = abs( cs )
!
    if ( acs>ab ) then
!
        ct = -tb/cs
        sn1 = one/sqrt( one + ct*ct )
        cs1 = ct*sn1
!
    else
!
        if ( ab==zero ) then
            cs1 = one
            sn1 = zero
        else
            tn = -cs/tb
            cs1 = one/sqrt( one + tn*tn )
            sn1 = tn*cs1
        end if
!
    end if
!
    if ( sgn1==sgn2 ) then
!
        tn = cs1
        cs1 = -sn1
        sn1 = tn
!
    end if
!
!
! END OF SUBROUTINE laev2
! _______________________
!
    end subroutine laev2
!
! =========================================================================================
!
    subroutine symtrid_ratqri( d, e, m, failure, small, tol )
!
! Purpose
! _______
!
!   SYMTRID_RATQRI computes the m largest or smallest eigenvalues of a symmetric n-by-n
!   tridiagonal matrix using a rational QR method.
!
!   The m largest or smallest eigenvalues of a full symmetric matrix can also be found
!   if SYMTRID_CMP has been used to reduce this matrix to tridiagonal form.
!
!
! Arguments
! _________
!
!   D           (INPUT/OUTPUT) real(stnd), dimension(:)
!               On entry, the diagonal elements of the tridiagonal matrix.
!
!               On exit, the computed eigenvalues replace the first m elements
!               of D in decreasing sequence. The other elements are lost.
!
!   E           (INPUT/OUTPUT) real(stnd), dimension(:)
!               On entry, the n-1 subdiagonal elements of the tridiagonal matrix.
!               E(n) is arbitrary and is used as workspace.
!
!               On exit, E has been destroyed.
!
!               The size of E must verify:  size( E ) = size( D ) = n .
!
!   M           (INPUT) integer(i4b)
!               On entry, the number of smallest or largest eigenvalues wanted.
!               M must be less than or equal to size( E ) = size( D ) = n.
!
!   FAILURE     (OUTPUT) logical(lgl)
!               On exit:
!
!               - FAILURE = false :  indicates successful exit.
!               - FAILURE = true  :  indicates that the algorithm did not converge and
!                 that full accuracy was not attained in the rational QR iterations
!                 for some eigenvalues.
!
!   SMALL       (INPUT, OPTIONAL) logical(lgl)
!               On entry:
!
!               - SMALL = false :  indicates that the M largest eigenvalues are desired.
!               - SMALL = true  :  indicates that the M smallest eigenvalues are desired.
!
!               The default is false.
!
!   TOL         (INPUT, OPTIONAL) real(stnd)
!               On entry, TOL specifies a tolerance for the theoretical error of the
!               computed eigenvalues. The theoretical error of the k-th eigenvalue 
!               is usually not greater than k * TOL.
!
!               The default is zero.
!
!
! Further Details
! _______________
!
!   This subroutine is not parallelized.
!
!   For further details, see:
!
!   (1) Reinsch, C., and  Bauer, F.L., 1968:
!           Rational QR transformation with Newton shift for
!           symmetric tridiagonal matrices. Numerische Mathematik 11, 264-272.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,    only : lamch
    use Utilities,        only : assert, assert_eq, iminloc, lascl
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:), intent(inout)           :: d, e
    real(stnd),               intent(in)  , optional  :: tol
!
    integer(i4b), intent(in) :: m
!
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in), optional  :: small
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: dlam, eps, eps2, safmin, safmax, ssfmax, ssfmin,    &
                  tridnorm, tot, s, err, delta, f, qp, p, q, r, ep
    real(stnd), dimension(size(d)) :: d2, e2
!
    integer(i4b) :: n, i, j, k, iscale
!
    logical(lgl) :: comp_largest
!
!
! PARAMETERS
! __________
!
!   A "FUDGE FACTOR" TO RELAX OR STRENGTHEN THE CONVERGENCE CRITERION FOR THE RATIONAL QR ITERATIONS.
!   IDEALLY, A VALUE OF ONE CAN BE USED, BUT EXPERIENCE SHOWS THAT THIS CAN BE TOO STRICT.
!
    real(stnd),  parameter :: fudge=c400
!
    character(len=*),  parameter :: name_proc='symtrid_ratqri'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(d),i4b)     ,   &
                    int(size(e),i4b)     ,   &
                    name_proc )
!
    call assert( logical( m>0_i4b, lgl ),      &
                 logical( n>=m, lgl ),         &
                 name_proc )
!
    dlam = zero
    if ( present(tol) ) then
        dlam = max( tol, zero )
    end if
!
    comp_largest = true
    if ( present(small) ) then
        comp_largest = .not. small
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) return
!
!   DETERMINE THE UNIT ROUNDOFF.
!
    eps = lamch( 'u' )
    eps2 = eps*eps
    safmin = lamch( 's' )
    safmax = one/safmin
    ssfmax = sqrt( safmax )/three
    ssfmin = sqrt( safmin )/eps2
!
    if ( comp_largest ) then
        d2(:n) = -d(:n)
    else
        d2(:n) = d(:n)
    end if
    e2(:n) = e(:n)
!
!   COMPUTE MAX ELEMENT OF TRIDIAGONAL MATRIX.
!
    e2(n) = zero
    tridnorm = abs( d2(1_i4b) ) + abs( e2(1_i4b) )
!
    do i = 2_i4b, n
        tridnorm = max( tridnorm, abs(d2(i))+abs(e2(i))+abs(e2(i-1_i4b)) )
    end do
!
!   SCALE TRIDIAGONAL MATRIX IF NECESSARY.
!
    iscale = 0_i4b
!
    if ( tridnorm==zero ) then
!
        return
!
    else if ( tridnorm>ssfmax ) then
!
        iscale = 1_i4b
        call lascl( d2(:n), tridnorm, ssfmax )
        call lascl( e2(:n-1_i4b), tridnorm, ssfmax )
!
    else if ( tridnorm<ssfmin ) then
!
        iscale = 2_i4b
        call lascl( d2(:n), tridnorm, ssfmin )
        call lascl( e2(:n-1_i4b), tridnorm, ssfmin )
!
    end if
!
!   DEFINE INITIAL SHIFT FROM GERSHGORIN LOWER BOUND FOR EIGENVALUES.
!
    tot = d2(1_i4b) - abs( e2(1_i4b) )
!
    do i = 2_i4b, n
        tot = min( tot, d2(i) - abs(e2(i)) - abs(e2(i-1_i4b)) )
    end do
!
!   SQUARE THE OFF-DIAGONAL ELEMENTS.
!
    e2(:n-1_i4b) = e2(:n-1_i4b)**2
!
!   CONVENIENT TO RENUMBER THE ELEMENTS OF e(:) .
!
    e2(:n) = eoshift( e2(:n), shift=-1 )
!
!   LOOK FOR SMALL SQUARES SUBDIAGONAL ELEMENTS AND REPLACE THEM
!   BY ZERO TO AVOID UNDERFLOWS.
!
    where ( e2(2_i4b:n)<=eps2*abs( d2(2_i4b:n)*d2(1:n-1_i4b) ) + safmin )   &
    e2(2_i4b:n) = zero
!
!   APPLY INITIAL SHIFT.
!
    d2(:n) = d2(:n) - tot
!
    s = zero
    err = zero
!
    do k = 1_i4b, m
!
!       RATIONAL QR ITERATIONS.
!
        qr_iter : do
!
            tot = tot + s
            delta = d2(n) - s
            i = n
            f = abs( fudge*eps*tot )
            if ( dlam<f ) dlam = f
!
!           EIGENVALUE FOUND.
!
            if ( abs(delta)<=dlam ) exit qr_iter
!
!           LOOK FOR SMALL SQUARES SUBDIAGONAL ELEMENTS AND REPLACE THEM
!           BY ZERO TO AVOID UNDERFLOWS.
!
            where ( e2(k+1_i4b:n)<=eps2*abs( d2(k+1_i4b:n)*d2(k:n-1_i4b) ) + safmin )   &
            e2(k+1_i4b:n) = zero
!
            f = e2(n)/delta
            qp = delta + f
            p = one
!
            do i = n-1_i4b, k, -1_i4b
!
                q = d2(i) - s - f
                r = q/qp
                p = p*r + one
                ep = f*r
                d2(i+1_i4b) = qp + ep
                delta = q - ep
!
!               EIGENVALUE FOUND.
!
                if ( abs(delta)<=dlam ) exit qr_iter
!
                f = e2(i)/q
                qp = delta + f
                e2(i+1_i4b) = qp*ep
!
            end do
!
            d2(k) = qp
            s = qp/p
!
            if ( tot+s<=tot ) then
!
!               IF THE SEQUENCE OF TOTAL SHIFTS DOES NO LONGER INCREASE 
!               DEFLATE MINIMUM DIAGONAL ELEMENT AND CONTINUE ON.
!
                s = zero
                i = iminloc( d2(k:n) ) + k - 1_i4b
                delta = d2(i)
!
!               SET FAILURE INDICATOR.
!
                failure = true
!
                exit qr_iter
!
            end if
!
        end do qr_iter
!
!       CONVERGENCE.
!
        if ( i<n ) e2(i+1_i4b) = e2(i)*f/qp
        err = err + abs( delta )
!
!BUG:   SOME COMPILERS (e.g. PGF90) DO NOT IMPLEMENT THE EOSHIFT SUBROUTINE
!       CORRECTLY. THUS, IT IS SAFER TO USE A SIMPLE LOOP.
!
!        d2(k:i) = eoshift( d2(k:i), shift=-1, boundary=tot )
!        e2(k:i) = eoshift( e2(k:i), shift=-1, boundary=err )
!
        do j = i-1_i4b, k, -1_i4b
!
            d2(j+1) = d2(j) - s
            e2(j+1) = e2(j)
!
        end do
!
        d2(k) = tot
        e2(k) = err
!
    end do
!
!   UNDO SCALING IF NECESSARY.
!
    if ( iscale>0_i4b ) then
        s = merge( ssfmax, ssfmin, iscale==1_i4b )
        call lascl( d2(:m), s, tridnorm )
    end if
!
    if ( comp_largest ) then
        d(:m) = -d2(:m)
    else
        d(:m) = d2(m:1_i4b:-1_i4b)
    end if
!
    d(m+1_i4b:n) = d2(m+1_i4b:n)
    e(:n) = e2(:n)
!
!
! END OF SUBROUTINE symtrid_ratqri
! ________________________________
!
    end subroutine symtrid_ratqri
!
! =========================================================================================
!
    subroutine symtrid_ratqri2( d, e, val, failure, m, small, tol )
!
! Purpose
! _______
!
!   SYMTRID_RATQRI2 computes the largest or smallest eigenvalues of a symmetric n-by-n
!   tridiagonal matrix whose sum in algebraic value exceeds a given value. A rational
!   QR method is used.
!
!   The largest or smallest eigenvalues of a full symmetric matrix whose sum exceeds
!   a given treshold in algebraic value can also be found, if SYMTRID_CMP has been used 
!   to reduce this matrix to tridiagonal form.
!
!
! Arguments
! _________
!
!   D           (INPUT/OUTPUT) real(stnd), dimension(:)
!               On entry, the diagonal elements of the tridiagonal matrix.
!
!               On exit, the computed eigenvalues replace the first M elements
!               of D in decreasing sequence. The other elements are lost.
!
!   E           (INPUT/OUTPUT) real(stnd), dimension(:)
!               On entry, the n-1 subdiagonal elements of the tridiagonal matrix.
!               E(n) is arbitrary and is used as workspace.
!
!               On exit, E has been destroyed.
!
!               The size of E must verify:  size( E ) = size( D ) = n .
!
!   VAL         (INPUT) real(stnd)
!               On entry, the sum of the M eigenvalues found will exceed abs(VAL) or 
!               M is equal to n.
!
!   FAILURE     (OUTPUT) logical(lgl)
!               On exit:
!
!               - FAILURE = false :  indicates successful exit.
!               - FAILURE = true  :  indicates that the algorithm did not converge and
!                 that full accuracy was not attained in the rational QR iterations
!                 for some eigenvalues.
!
!   M           (OUTPUT) integer(i4b)
!               On exit, the number of eigenvalues found.
!
!   SMALL       (INPUT, OPTIONAL) logical(lgl)
!               On entry:
!
!               - SMALL = false :  indicates that the M largest eigenvalues are desired.
!               - SMALL = true  :  indicates that the M smallest eigenvalues are desired.
!
!               The default is false.
!
!   TOL         (INPUT, OPTIONAL) real(stnd)
!               On entry, TOL specifies a tolerance for the theoretical error of the
!               computed eigenvalues. The theoretical error of the k-th eigenvalue 
!               is usually not greater than k * TOL.
!
!               The default is zero.
!
!
! Further Details
! _______________
!
!   This subroutine is not parallelized.
!
!   For further details, see:
!
!   (1) Reinsch, C., and  Bauer, F.L., 1968:
!           Rational QR transformation with Newton shift for
!           symmetric tridiagonal matrices. Numerische Mathematik 11, 264-272.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,    only : lamch
    use Utilities,        only : assert_eq, iminloc, lascl
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:), intent(inout)           :: d, e
    real(stnd),               intent(in)              :: val
    real(stnd),               intent(in)  , optional  :: tol
!
    integer(i4b), intent(out) :: m
!
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in), optional  :: small
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: dlam, eps, eps2, safmin, safmax, ssfmax, ssfmin,       &
                  tridnorm, tot, s, err, delta, f, qp, p, q, r, ep, som
    real(stnd), dimension(size(d)) :: d2, e2
!
    integer(i4b) :: n, i, j, k, iscale
!
    logical(lgl) :: comp_largest
!
!
! PARAMETERS
! __________
!
!   A "FUDGE FACTOR" TO RELAX OR STRENGTHEN THE CONVERGENCE CRITERION FOR THE RATIONAL QR ITERATIONS.
!   IDEALLY, A VALUE OF ONE CAN BE USED, BUT EXPERIENCE SHOWS THAT THIS CAN BE TOO STRICT.
!
    real(stnd),  parameter :: fudge=c400
!
    character(len=*),  parameter :: name_proc='symtrid_ratqri2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(d),i4b)     ,   &
                    int(size(e),i4b)     ,   &
                    name_proc )
!
    dlam = zero
    if ( present(tol) ) then
        dlam = max( tol, zero )
    end if
!
    comp_largest = true
    if ( present(small) ) then
        comp_largest = .not. small
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
            m = 1_i4b
        else
            m = 0_i4b
        end if
!
        return
!
    end if
!
!   DETERMINE THE UNIT ROUNDOFF.
!
    eps = lamch( 'u' )
    eps2 = eps*eps
    safmin = lamch( 's' )
    safmax = one/safmin
    ssfmax = sqrt( safmax )/three
    ssfmin = sqrt( safmin )/eps2
!
    if ( comp_largest ) then
        d2(:n) = -d(:n)
    else
        d2(:n) = d(:n)
    end if
!
    e2(:n) = e(:n)
!
!   COMPUTE MAX ELEMENT OF TRIDIAGONAL MATRIX.
!
    e2(n) = zero
    tridnorm = abs( d2(1_i4b) ) + abs( e2(1_i4b) )
!
    do i = 2_i4b, n
        tridnorm = max( tridnorm, abs(d2(i))+abs(e2(i))+abs(e2(i-1_i4b)) )
    end do
!
!   SCALE TRIDIAGONAL MATRIX IF NECESSARY.
!
    iscale = 0_i4b
!
    if ( tridnorm==zero ) then
!
        m = n
!
        return
!
    else if ( tridnorm>ssfmax ) then
!
        iscale = 1_i4b
        call lascl( d2(:n), tridnorm, ssfmax )
        call lascl( e2(:n-1_i4b), tridnorm, ssfmax )
!
    else if ( tridnorm<ssfmin ) then
!
        iscale = 2_i4b
        call lascl( d2(:n), tridnorm, ssfmin )
        call lascl( e2(:n-1_i4b), tridnorm, ssfmin )
!
    end if
!
!   DEFINE INITIAL SHIFT FROM GERSHGORIN LOWER BOUND FOR EIGENVALUES.
!
    tot = d2(1_i4b) - abs( e2(1_i4b) )
!
    do i = 2_i4b, n
        tot = min( tot, d2(i) - abs(e2(i)) - abs(e2(i-1_i4b)) )
    end do
!
!   SQUARE THE OFF-DIAGONAL ELEMENTS.
!
    e2(:n-1_i4b) = e2(:n-1_i4b)**2
!
!   CONVENIENT TO RENUMBER THE ELEMENTS OF e(:) .
!
    e2(:n) = eoshift( e2(:n), shift=-1 )
!
!   LOOK FOR SMALL SQUARES SUBDIAGONAL ELEMENTS AND REPLACE THEM
!   BY ZERO TO AVOID UNDERFLOWS.
!
    where ( e2(2_i4b:n)<=eps2*abs( d2(2_i4b:n)*d2(1:n-1_i4b) ) + safmin )   &
    e2(2_i4b:n) = zero
!
!   APPLY INITIAL SHIFT.
!
    d2(:n) = d2(:n) - tot
!
    som = zero
    s = zero
    err = zero
!
    do k = 1_i4b, n
!
!       RATIONAL QR ITERATIONS.
!
        qr_iter : do
!
            tot = tot + s
            delta = d2(n) - s
            i = n
            f = abs( fudge*eps*tot )
            if ( dlam<f ) dlam = f
!
!           EIGENVALUE FOUND.
!
            if ( abs(delta)<=dlam ) exit qr_iter
!
!           LOOK FOR SMALL SQUARES SUBDIAGONAL ELEMENTS AND REPLACE THEM
!           BY ZERO TO AVOID UNDERFLOWS.
!
            where ( e2(k+1_i4b:n)<=eps2*abs( d2(k+1_i4b:n)*d2(k:n-1_i4b) ) + safmin )   &
            e2(k+1_i4b:n) = zero
!
            f = e2(n)/delta
            qp = delta + f
            p = one
!
            do i = n-1_i4b, k, -1_i4b
!
                q = d2(i) - s - f
                r = q/qp
                p = p*r + one
                ep = f*r
                d2(i+1_i4b) = qp + ep
                delta = q - ep
!
!               EIGENVALUE FOUND.
!
                if ( abs(delta)<=dlam ) exit qr_iter
!
                f = e2(i)/q
                qp = delta + f
                e2(i+1_i4b) = qp*ep
!
            end do
!
            d2(k) = qp
            s = qp/p
!
            if ( tot+s<=tot ) then
!
!               IF THE SEQUENCE OF TOTAL SHIFTS DOES NO LONGER INCREASE 
!               DEFLATE MINIMUM DIAGONAL ELEMENT AND CONTINUE ON.
!
                s = zero
                i = iminloc( d2(k:n) ) + k - 1_i4b
                delta = d2(i)
!
!               SET FAILURE INDICATOR.
!
                failure = true
!
                exit qr_iter
!
            end if
!
        end do qr_iter
!
!       CONVERGENCE.
!
        if ( i<n ) e2(i+1_i4b) = e2(i)*f/qp
        err = err + abs( delta )
!
!BUG:   SOME COMPILERS (e.g. PGF90) DO NOT IMPLEMENT THE EOSHIFT SUBROUTINE
!       CORRECTLY. THUS, IT IS SAFER TO USE A SIMPLE LOOP.
!
!        d2(k:i) = eoshift( d2(k:i), shift=-1, boundary=tot )
!        e2(k:i) = eoshift( e2(k:i), shift=-1, boundary=err )
!
        do j = i-1_i4b, k, -1_i4b
            d2(j+1) = d2(j) - s
            e2(j+1) = e2(j)
        end do
        d2(k) = tot
        e2(k) = err
!
        som = som + abs( tot )
        if ( som>abs(val) ) exit
!
    end do
!
    if ( k<=n ) then
        m = k
    else
        m = n
    end if
!
!   UNDO SCALING IF NECESSARY.
!
    if ( iscale>0_i4b ) then
        s = merge( ssfmax, ssfmin, iscale==1_i4b )
        call lascl( d2(:m), s, tridnorm )
    end if
!
    if ( comp_largest ) then
        d(:m) = -d2(:m)
    else
        d(:m) = d2(m:1_i4b:-1_i4b)
    end if
!
    d(m+1_i4b:n) = d2(m+1_i4b:n)
    e(:n) = e2(:n)
!
!
! END OF SUBROUTINE symtrid_ratqri2
! _________________________________
!
    end subroutine symtrid_ratqri2
!
! =========================================================================================
!
    subroutine symtrid_bisect( d, e, neig, w, failure, small, sort, vector, abstol, &
                               le, theta, scaling, init                             )
!
! Purpose
! _______
!
!   SYMTRID_BISECT computes all or some of the largest or smallest eigenvalues of a real n-by-n
!   symmetric tridiagonal matrix T using a bisection method.
!
!   The largest or smallest eigenvalues of a full symmetric matrix can also be found
!   if SYMTRID_CMP has been used to reduce this matrix to tridiagonal form.
!
!
! Arguments
! _________
!
!   D            (INPUT) real(stnd), dimension(:)
!                On entry, D contains the diagonal elements of the tridiagonal
!                matrix T.
!
!   E            (INPUT) real(stnd), dimension(:)
!                On entry, E contains the n-1 off-diagonal elements of the tridiagonal
!                matrix T whose eigenvalues are desired. E(n) is arbitrary.
!
!                The size of E must verify: size( E ) = size( D ) = n. 
!
!   NEIG         (OUTPUT) integer(i4b)
!                On output, NEIG specifies the number of eigenvalues which have
!                been computed. Note that NEIG may be greater than the optional argument
!                LE, if multiple eigenvalues at index LE make unique selection
!                impossible.
!
!                If none of the optional arguments LE and THETA are used, NEIG is set to size(D)
!                and all the eigenvalues of T are computed.
!
!   W            (OUTPUT) real(stnd), dimension(:)
!                On exit, W(1:NEIG)  contains the first NEIG largest (or smallest) eigenvalues of T.
!                The other values in W (e.g.  W(NEIG+1:size(D)) ) are flagged  by a quiet NAN.
!
!                The size of W must verify: size( W ) = size( D ) = n . 
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit and the bisection algorithm converged for all
!                  the computed eigenvalues to the desired accuracy ;
!                - FAILURE = true  :  indicates that some or all of the eigenvalues failed to converge
!                  or were not computed. This is generally caused by unexpectedly inaccurate
!                  arithmetic.
!
!   SMALL        (INPUT, OPTIONAL) logical(lgl)
!                On entry:
!
!                - SMALL = false :  indicates that the largest eigenvalues are desired.
!                - SMALL = true  :  indicates that the smallest eigenvalues are desired.
!
!                The default is false.
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'. For other values of SORT
!                nothing is done and W(:NEIG) may not be sorted.
!
!   VECTOR       (INPUT, OPTIONAL) logical(lgl)
!                On entry, if VECTOR is set to true, a vectorized version of the bisection algorithm
!                is used.
!
!                The default is VECTOR=false.
!
!   ABSTOL       (INPUT, OPTIONAL) real(stnd)
!                On entry, the absolute tolerance for the eigenvalues. An eigenvalue
!                (or cluster) is considered to be located if it has been determined to lie in
!                an interval whose width is ABSTOL or less.
!
!                If ABSTOL is less than or equal to zero, or is not specified, then ULP * | T | will be used,
!                where | T | means the 1-norm of T and ULP is the machine precision (distance from 1 to the
!                next larger floating point number).
!
!                Eigenvalues will be computed most accurately when ABSTOL is set to the square root of
!                the underflow threshold, sqrt(LAMCH('S')), not zero. 
!
!   LE           (INPUT, OPTIONAL) integer(i4b)
!                On entry, LE specifies the number of eigenvalues which must be
!                computed by the subroutine. On output, NEIG may be different than LE
!                if multiple eigenvalues at index LE make unique selection
!                impossible.
!
!                If:
!
!                - SMALL=false, the subroutine computes the LE largest eigenvalues of T,
!                - SMALL=true, the subroutine computes the LE smallest eigenvalues of T.
!
!                Only one of the optional arguments LE and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!                LE must be greater than 0 and less or equal to size( D ) .
!
!                The default is LE = size( D ).
!
!   THETA        (INPUT, OPTIONAL) real(stnd)
!                On entry:
!
!                - if SMALL=false, THETA specifies that the eigenvalues which are greater or equal to
!                  THETA must be computed. If none of the eigenvalues are greater or equal to
!                  THETA, NEIG is set to zero and W(:) to a quiet NAN.
!                - if SMALL=true, THETA specifies that the eigenvalues which are less or equal to
!                  THETA must be computed. If none of the eigenvalues are smaller or equal to
!                  THETA, NEIG is set to zero and W(:) to a quiet NAN.
!
!                Only one of the optional arguments LE and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!   SCALING      (INPUT, OPTIONAL) logical(lgl)
!                On entry, if SCALING=true the tridiagonal matrix T is scaled before computing
!                the eigenvalues.
!
!                The default is to scale the tridiagonal matrix.
!
!   INIT         (INPUT, OPTIONAL) logical(lgl)
!                On entry, if INIT=true the initial intervals for the bisection steps are
!                computed from estimates of the eigenvalues of the tridiagonal matrix
!                obtained from the Pal-Walker-Kahan algorithm.
!
!                The default is not to use the Pal-Walker-Kahan algorithm.
!
!
! Further Details
! _______________
!
!   Let S(i), i=1,...,N=size( D ), be the N eigenvalues of the symmetric tridiagonal matrix T in
!   decreasing order of magnitude. SYMTRID_BISECT then computes the LE largest or smallest eigenvalues
!   ( or the eigenvalues which are greater/smaller or equal to THETA) of T by a bisection method
!   (see the reference (1) below, Sec.8.5 ).
!
!   This subroutine is parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants, only : allocate_error, eig_error3
    use Num_Constants,  only : lamch, nan, machbase, maxexp, machhuge
    use Utilities,      only : assert_eq, assert, merror, arth
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),  dimension(:) :: d, e
    real(stnd), intent(out), dimension(:) :: w
    real(stnd), intent(in), optional      :: abstol, theta
!
    integer(i4b), intent(out)          :: neig
    integer(i4b), intent(in), optional :: le
!
    logical(lgl), intent(in), optional :: vector, small, scaling, init
    logical(lgl), intent(out)          :: failure
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: eps, safmin, gamma, omega, tau, scalfac, maxe, tnorm, atol1, rtol1,   &
                  tmp1, tmp2, y, z, th, gu, gl
    real(stnd),  dimension(size(d))               :: d2, e2
    real(stnd),  dimension(:),        allocatable :: lb, ub, yvec, zvec, thvec, tmpvec
!
    integer(i4b)                             :: n, num, nth, nz, ny, ngu, ngl, i, j, k
    integer(i4b),  dimension(:), allocatable :: nlb, nub, nyvec, nzvec, nthvec, nevec
!
    integer  :: iok
!
    logical(lgl) :: comp_smallest, use_theta, use_vectorization, do_scaling, use_init
!
#ifdef _OPENMP
!
    integer(i4b) :: num_threads, thread_num, jinc, jrem, joffset, j1, j2
!
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
!   relfac IS THE RELATIVE TOLERANCE.  AN INTERVAL (a,b] LIES WITHIN "RELATIVE TOLERANCE" IF
!   b-a < relfac*ulp*max(|a|,|b|), WHERE "ulp" IS THE MACHINE PRECISION.
!
    real(stnd),  parameter :: relfac=two
!
!   A "FUDGE FACTOR" TO WIDEN THE UPPER BOUND OF THE GERSHGORIN INTERVAL. IDEALLY, A VALUE OF 1 SHOULD WORK,
!   BUT ON MACHINES WITH SLOPPY ARITHMETIC, THIS NEEDS TO BE LARGER. 
!   THE DEFAULT FOR PUBLICLY RELEASED VERSIONS SHOULD BE LARGE ENOUGH TO HANDLE THE WORST MACHINE AROUND.
!   NOTE THAT THIS PARAMETER HAS NO EFFECT ON ACCURACY OF THE SOLUTION.
!
    real(stnd),  parameter :: fudge=two
!
    character(len=*),  parameter :: name_proc='symtrid_bisect'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure           = false
    comp_smallest     = false
    use_vectorization = false
    do_scaling        = true
    use_init          = false
    th                = -machhuge
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq( int(size(d),i4b),    &
                   int(size(e),i4b),    & 
                   int(size(w),i4b),    & 
                   name_proc )
!
    neig = n
!
    if ( present(le) ) then
!
        call assert( logical( le<=n, lgl ),     &
                     logical( 1_i4b<=le, lgl ), &
                     name_proc )
        neig = le
!
    end if
!
    if ( present(small) ) then
        comp_smallest = small
    end if
!
    use_theta = present( theta )
!
    if ( use_theta ) then
!
        if ( present(le)  )     &
        call merror( name_proc//eig_error3 )
!
        if ( comp_smallest ) then
            th = -theta
        else
            th = theta
        end if
!
    end if
!
    if ( present(abstol) ) then
        atol1 = abstol
    else
        atol1 = -one
    end if
!
    if ( present(vector) ) then
        use_vectorization = vector
    end if
!
    if ( present(scaling) ) then
        do_scaling = scaling
    end if
!
    if ( present(init) ) then
        use_init = init
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            if ( comp_smallest ) then
                tmp1 = -d(1_i4b)
            else
                tmp1 = d(1_i4b)
            end if
!
            if ( th<=tmp1 ) then 
                neig     = 1_i4b
                w(1_i4b) = d(1_i4b)
            else
                neig     = 0_i4b
                w(1_i4b) = nan()
            end if
!
        else
!
            neig = 0_i4b
!
        end if
!
        return
!
    end if
!
    maxe = maxval( abs(e(1_i4b:n-1_i4b)) )
    tnorm = max( maxval( abs(d(1_i4b:n))), maxe )
!
    if ( tnorm<=zero ) then
!
        if ( th<=zero ) then 
            neig       = n
            w(1_i4b:n) = zero
        else
            neig       = 0_i4b
            w(1_i4b:n) = nan()
        end if
!
        return
!
    end if
!
!   GET MACHINE PARAMETERS.
!
    eps   = lamch( 'p' )
    rtol1 = eps*relfac
!
    safmin = lamch( 's' )
    gamma  = sqrt( safmin )
!
    if ( do_scaling ) then
!
        omega  = machbase**(maxexp-1_i4b)
        tau    = sqrt( gamma*omega )
!
!       COMPUTE SCALING FACTOR TO AVOID OVERFLOW.
!
!        j       = ( exponent( tau ) - exponent( maxe ) - 1_i4b )/2_i4b
        j       = exponent( tau ) - exponent( maxe ) - 1_i4b
        scalfac = machbase**j
!
    else
!
        scalfac = one
!
    end if
!
!   SCALE THE ELEMENTS OF THE TRIDIAGONAL MATRIX AND SQUARE THE OFF DIAGONAL ELEMENTS.
!
    d2(1_i4b:n) = scalfac*d(1_i4b:n)
!
    if ( comp_smallest ) then
        d2(1_i4b:n) = -d2(1_i4b:n)
    end if
!
    e2(1_i4b) = zero
!
    tmp1 = gamma/scalfac
!
    where( abs(e(1_i4b:n-1_i4b))>=tmp1 )
        e2(2_i4b:n) = (scalfac*e(1_i4b:n-1_i4b))**2
    elsewhere
        e2(2_i4b:n) = safmin
!        e2(2_i4b:n) = merge( zero, safmin, e2(2_i4b:n)==zero )
    end where
!
!   COMPUTE THE UPPER and LOWER GERSCHGORIN BOUNDS FOR THE TRIDIAGONAL MATRIX.
!
    gu   = d2(1_i4b)
    gl   = d2(1_i4b)
    tmp1 = zero
!
    do i=1_i4b, n-1_i4b
!
        tmp2 = sqrt( e2(i+1_i4b) )
        gu   = max( gu, d2(i)+tmp1+tmp2 )
        gl   = min( gl, d2(i)-tmp1-tmp2 )
        tmp1 = tmp2
!
    end do
!
    gu   = max( gu, d2(n)+tmp1 )
    gl   = min( gl, d2(n)-tmp1 )
!
    tnorm = max( abs(gu), abs(gl) )
    gu    = gu + (real( n, stnd)*fudge + eight)*tnorm*eps
    gl    = gl - (real( n, stnd)*fudge + eight)*tnorm*eps
    ngu   = 0_i4b
    ngl   = n
!
!   CHECK ABSOLUTE TOLERANCE AND REPLACE IT BY DEFAULT VALUE IF NECESSARY.
!
    if ( atol1<=zero ) then
        atol1 = eps*tnorm
    end if
    atol1 = max( gamma, atol1 )
!
    if ( use_theta ) then
!
        gl = th*scalfac
!
!       DETERMINE neig THE NUMBER OF EIGENVALUES WHICH ARE EQUAL
!       OR GREATER THAN th BY STURM SEQUENCE.
!
        ngl  = 0_i4b
        tmp2 = one
!
        do j=1_i4b, n
!
            tmp2 = ( d2(j) - (e2(j)/tmp2) ) - gl
!
            if ( tmp2>=gamma ) then
                ngl = ngl + 1_i4b
            else
                tmp2 = min( tmp2, -gamma )
            end if
!
        end do
!
        neig = ngl
!
    else if ( neig/=n ) then
!
!       DETERMINE AN INTERVAL CONTAINING EXACTLY THE DESIRED EIGENVALUES.
!
        y  = gl
        ny = n
        z  = gu
        nz = 0_i4b
!
!       BISECTION METHOD FOR FINDING THE LOWER BOUND gl ON THE neig
!       LARGEST EIGENVALUES OF THE TRIDIAGONAL MATRIX.
!
        do
!
            tmp1 = abs( z - y )
            tmp2 = max( abs(z), abs(y) )
!
            if ( ngl==neig .or. tmp1<max(gamma,rtol1*tmp2) .or. nz>=ny ) exit
!
            gl = ( z + y )*half
!
            num  = 0_i4b
            tmp2 = one
!
            do j=1_i4b, n
!
                tmp2 = ( d2(j) - (e2(j)/tmp2) ) - gl
!
                if ( tmp2>=gamma ) then
                    num = num + 1_i4b
                else
                    tmp2 = min( tmp2, -gamma )
                end if
!
            end do
!
            ngl = min( max(num,nz), ny ) 
!
            if ( ngl<neig ) then
                z  = gl
                nz = ngl
            else
                y  = gl
                ny = ngl
            end if
!
        end do
!
        if ( ngl/=neig ) then
!
            gl  = y
            ngl = ny
!
            neig = ny
!
        end if
!
    end if
!
    if ( neig>0_i4b ) then
!
!       ALLOCATE WORK VARIABLES NEEDED TO STORE THE UPPER AND LOWER BOUNDS
!       AND EIGENVALUES COUNTS ASSOCIATED WITH THE BOUNDS.
!
        allocate( lb(neig), nlb(neig), ub(neig), nub(neig), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        if ( use_init ) then
!
!           COMPUTE THE INITIAL INTERVALS FROM THE EIGENVALUES OF THE TRIDIAGONAL
!           MATRIX ESTIMATED FROM THE PAL-WALKER-KAHAN ALGORITHM.
!
            allocate( yvec(n), zvec(n), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           COMPUTE EIGENVALUES OF THE TRIDIAGONAL MATRIX T.
!
            if ( comp_smallest ) then
                yvec(1_i4b:n) = -d(1_i4b:n)
                zvec(1_i4b:n) = -e(1_i4b:n)
            else
                yvec(1_i4b:n) = d(1_i4b:n)
                zvec(1_i4b:n) = e(1_i4b:n)
            end if
!
            call symtrid_qri( yvec(:n), zvec(:n), failure, sort='d' )
!
!           COMPUTE LOWER AND UPPER BOUNDS FOR THE EIGENVALUES OF THE TRIDIAGONAL MATRIX.
!
            if (failure ) then
!
                lb(:neig) = gl
                ub(:neig) = gu
!
            else
!
                yvec(1_i4b:neig) = scalfac*yvec(1_i4b:neig)
!
                tmp1 = two*tnorm*eps
!
                lb(:neig) = yvec(:neig) - tmp1
                ub(:neig) = yvec(:neig) + tmp1
!
                where( lb(:neig)<gl )
                    ub(:neig) = min( ub(:neig) + gl - lb(:neig), gu )
                    lb(:neig) = gl
                end where
!
                where( ub(:neig)>gu )
                    lb(:neig) = max( lb(:neig) + gu - ub(:neig), gl )
                    ub(:neig) = gu
                end where
!
            end if
!
            deallocate( yvec, zvec )
!
        else
!
            lb(:neig) = gl
            ub(:neig) = gu
!
        end if
!
        if ( use_vectorization ) then  
!
!           VECTORIZED VERSION OF THE BISECTION ALGORITHM.
!
!           ALLOCATE WORK VARIABLES NEEDED TO STORE THE UPPER AND LOWER BOUNDS
!           FOR THE EIGENVALUES.
!
            allocate( zvec(neig), nzvec(neig), yvec(neig), nyvec(neig),       &
                      thvec(neig), nthvec(neig), tmpvec(neig), nevec(neig),   &
                      stat = iok    )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           INITIALIZE THE LOWER AND THE UPPER BOUNDS FOR THE REQUESTED EIGENVALUES.
!
            zvec(:neig)  = ub(:neig)
            nzvec(:neig) = ngu
!
            yvec(:neig)  = lb(:neig)
            nyvec(:neig) = ngl
!
            k    = 0_i4b
            num  = neig
            nevec(:neig) = arth( 1_i4b, 1_i4b, neig )
!
            do
                i = 0_i4b
!
                do j=1_i4b, num
!
                    y    = yvec(j)
                    ny   = nyvec(j)
                    z    = zvec(j)
                    nz   = nzvec(j)
!
                    tmp1 = abs( z - y )
                    tmp2 = max( abs(z), abs(y) )
!
!                   CHECK CONVERGENCE OF THE nevec(j) EIGENVALUE.
!
                    if ( tmp1<max(rtol1*tmp2,atol1) .or. nz>=ny ) then
!
                        k = k + 1_i4b
!
                        w(k) = min( max(( y + z )*half,y), z )
!
                        lb(k)  = y
                        nlb(k) = ny
                        ub(k)  = z
                        nub(k) = nz
!
                    else
!
                        i = i + 1_i4b
!
                        if ( i/=j ) then
!
                            yvec(i)   = y
                            nyvec(i)  = ny
                            zvec(i)   = z
                            nzvec(i)  = nz
!
                            nevec(i)  = nevec(j)
!
                        end if
!
                    end if
!
                end do
!
!               CHECK IF ALL THE REQUESTED EIGENVALUES HAVE CONVERGED.
!
                if ( k==neig ) exit
!
                num = i
!
!               NEXT BISECTION STEP FOR THE REMAINING i EIGENVALUES.
!
                thvec(:i) = ( yvec(:i) + zvec(:i) )*half
!
!               COMPUTE nthvec(:i) = COUNT(thvec(:i))  THE NUMBER OF EIGENVALUES GREATER
!               OR EQUAL TO thvec(:i) BY STURM SEQUENCE.
!
                nthvec(:i) = 0_i4b
                tmpvec(:i) = one
!
                do j=1_i4b, n
!
                    tmpvec(:i) = ( d2(j) -  (e2(j)/tmpvec(:i)) ) - thvec(:i)
!
                    where ( tmpvec(:i)>=gamma )
                        nthvec(:i) = nthvec(:i) + 1_i4b
                    elsewhere
                        tmpvec(:i) = min( tmpvec(:i), -gamma )
                    end where
!
                end do
!
!               INSURE THAT COUNT(thvec(:i)) IS MONOTONE.
!
                nthvec(:i) = min( max(nthvec(:i),nzvec(:i)),nyvec(:i) ) 
!
!               KEEP ONLY THE INTERVAL CONTAINING w(:i) S.T. COUNT(w(:i))=nevec(:i)
!
                where ( nthvec(:i)<nevec(:i) )
                    zvec(:i)  = thvec(:i)
                    nzvec(:i) = nthvec(:i)
                elsewhere
                    yvec(:i)  = thvec(:i)
                    nyvec(:i) = nthvec(:i)
                end where
!
            end do        
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( yvec, nyvec, zvec, nzvec, thvec, nthvec, tmpvec, nevec )
!
        else
!
!           SCALAR VERSION OF THE BISECTION ALGORITHM, BUT MAY BE PARALLELIZED
!           if OPENMP IS USED.
!
#ifdef _OPENMP
!
            i           = omp_get_max_threads()
            ompparallel = omp_in_parallel()
            ompnested   = omp_get_nested() .and. ompparallel
            test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                          neig>1_i4b                          .and.      &
                          i>1_i4b
!
            if ( test_par ) then
!
!$OMP PARALLEL PRIVATE(num_threads,thread_num,jinc,jrem,joffset,k,i,j,j1,j2,iok,num,tmp1,tmp2,z,nz,y,ny,th,nth)
!
                num_threads = omp_get_num_threads()
                thread_num  = omp_get_thread_num()
!
!               DISTRIBUTE THE WORK AMONG THE THREADS.
!
                jrem = mod( neig, num_threads )
                jinc = ( neig - jrem )/num_threads
!
                if ( thread_num<jrem ) then
                    jinc    = jinc + 1_i4b
                    joffset = 1_i4b
                else
                    joffset = jrem + 1_i4b
                end if
!
                j1 = thread_num*jinc + joffset
                j2 = j1 + jinc - 1_i4b
!
                if ( j1<=j2 ) then
!
!                   INITIALIZE THE UPPER BOUNDS FOR THE REQUESTED EIGENVALUES VALUES.
!
                    z  = gu
                    nz = ngu
!
                    do k=j1, j2
!
                        y  = gl
                        ny = ngl
!
!                       UPDATE THE LOWER AND UPPER BOUNDS FOR THE kTH EIGENVALUE.
!
                        do i=k, j2
!
                            if ( lb(i)>gl ) then
!
                                y  =  lb(i)
!
                                exit
!
                            end if
!
                        end do
!
                        tmp1 = min( z, ub(k) )
!
                        if ( tmp1<y ) then
                            y  = gl
                            z  = gu
                        else
                            z  = tmp1
                        end if
!
                        do
!
                            tmp1 = abs( z - y )
                            tmp2 = max( abs(z), abs(y) )
!
!                           CHECK FOR CONVERGENCE.
!
                            if ( tmp1<max(rtol1*tmp2,atol1) .or. nz>=ny ) exit
!
!                           NEXT BISECTION STEP.
!
                            th = ( y + z )*half
!
!                           COMPUTE nth=COUNT(th) THE NUMBER OF EIGENVALUES GREATER OR EQUAL TO th
!                           BY STURM SEQUENCE.
!
                            num  = 0_i4b
                            tmp2 = one
!
                            do j=1_i4b, n
!
                                tmp2 = ( d2(j) - (e2(j)/tmp2) ) - th
!
                                if ( tmp2>=gamma ) then
                                    num = num + 1_i4b
!                                    if ( (num>=j2) .or. (n-j+num<k) ) exit
                                else
                                    tmp2 = min( tmp2, -gamma )
                                end if
!
                            end do
!
!                           INSURE THAT COUNT(th) IS MONOTONE.
!
                            nth = min( max(num,nz),ny ) 
!
!                           KEEP ONLY THE INTERVAL CONTAINING w S.T. COUNT(w)=k
!
                            if ( nth<k ) then
!
                                z  = th
                                nz = nth
!
                            else
!
                                y  = th
                                ny = nth
!
!                               REFINE THE REMINDING INTERVALS.
!
                                if ( j2<=nth ) then
!
!                                    if ( th>lb(j2) ) then
!                                        if ( th>ub(j2) )  ub(j2) = ub(j2) + th - lb(j2)
!                                        lb(j2) = th
!                                    end if
                                    lb(j2)  = max( th, lb(j2) )
!
                                else
!
!                                    if  ( th>lb(nth) ) then
!                                        if ( th>ub(nth) )  ub(nth) = ub(nth) + th - lb(nth)
!                                        lb(nth) = th
!                                    end if
                                    lb(nth)  = max( th, lb(nth) )
!                                    if ( th<ub(nth+1_i4b) ) then
!                                        if ( th<lb(nth+1_i4b) ) lb(nth+1_i4b) = max( lb(nth+1_i4b) + th - ub(nth+1_i4b), gl )
!                                        ub(nth+1_i4b) = max( th, gl )
!                                    end if
                                    ub(nth+1_i4b)  = min( th, ub(nth+1_i4b) )
!
                                end if
!
                            end if
!
                        end do        
!
!                       EIGENVALUE FOUND.
!
                        w(k) = min( max(( y + z )*half,y), z )
!
                        lb(k)  = y
                        nlb(k) = ny
!
                        ub(k)  = z
                        nub(k) = nz
!
                    end do
!
                end if
!
!$OMP END PARALLEL
!
            else
!
#endif
!
!               INITIALIZE THE UPPER BOUNDS FOR THE REQUESTED EIGENVALUES.
!
                z  = gu
                nz = ngu
!
                do k=1_i4b, neig
!
!                   UPDATE THE LOWER AND UPPER BOUNDS FOR THE kTH EIGENVALUE.
!
                    y  = gl
                    ny = ngl
!
                    do i=k, neig
!
                        if ( lb(i)>gl ) then
!
                            y  =  lb(i)
!
                            exit
!
                        end if
!
                    end do
!
                    tmp1 = min( z, ub(k) )
!
                    if ( tmp1<y ) then
                        y  = gl
                        z  = gu
                    else
                        z  = tmp1
                    end if
!
                    do
!
                        tmp1 = abs( z - y )
                        tmp2 = max( abs(z), abs(y) )
!
!                       CHECK FOR CONVERGENCE.
!
                        if ( tmp1<max(rtol1*tmp2,atol1) .or. nz>=ny ) exit
!
!                       NEXT BISECTION STEP.
!
                        th = ( y + z )*half
!
!                       COMPUTE nth=COUNT(th)  THE NUMBER OF EIGENVALUES GREATER OR EQUAL TO th
!                       BY STURM SEQUENCE.
!
                        num  = 0_i4b
                        tmp2 = one
!
                        do j=1_i4b, n
!
                            tmp2 = ( d2(j) - (e2(j)/tmp2) ) - th
!
                            if ( tmp2>=gamma ) then
                                num = num + 1_i4b
!                                if ( (num>=neig) .or. (n-j+num<k) ) exit
                            else
                                tmp2 = min( tmp2, -gamma )
                            end if
!
                        end do
!
!                       INSURE THAT COUNT(th) IS MONOTONE.
!
                        nth = min( max(num,nz),ny ) 
!
!                       KEEP ONLY THE INTERVAL CONTAINING w S.T. COUNT(w)=k
!
                        if ( nth<k ) then
!
                            z  = th
                            nz = nth
!
                        else
!
                            y  = th
                            ny = nth
!
!                           REFINE THE REMINDING INTERVALS.
!
                            if ( neig<=nth ) then
!
!                                if ( th>lb(neig) ) then
!                                    if ( th>ub(neig) )  ub(neig) = ub(neig) + th - lb(neig)
!                                    lb(neig) = th
!                                end if
                                lb(neig)  = max( th, lb(neig) )
!
                            else
!
!                                if  ( th>lb(nth) ) then
!                                    if ( th>ub(nth) )  ub(nth) = ub(nth) + th - lb(nth)
!                                    lb(nth) = th
!                                end if
                                lb(nth)  = max( th, lb(nth) )
!                                if ( th<ub(nth+1_i4b) ) then
!                                    if ( th<lb(nth+1_i4b) ) lb(nth+1_i4b) = max( lb(nth+1_i4b) + th - ub(nth+1_i4b), gl )
!                                    ub(nth+1_i4b) = max( th, gl )
!                                end if
                                ub(nth+1_i4b)  = min( th, ub(nth+1_i4b) )
!
                            end if
!
                        end if
!
                    end do        
!
!                   EIGENVALUE FOUND.
!
                    w(k) = min( max(( y + z )*half,y), z )
!
                    lb(k)  = y
                    nlb(k) = ny
!
                    ub(k)  = z
                    nub(k) = nz
!
                end do
!
#ifdef _OPENMP
!
            end if
!
#endif
!
        end if
!
!       CHECK FOR CONVERGENCE.
!
        th = one/scalfac
        if ( comp_smallest ) then
            th = -th
        end if
!
        do k=1_i4b, neig
!
            tmp1 = w(k)
!
            if ( lb(k)>ub(k)    .or.  &
                 nub(k)>=nlb(k) .or.  & 
                 tmp1>=gu       .or.  & 
                 tmp1<gl  )   then
!
!               SET FAILURE FLAG.
!
                failure = true
!
            end if
!
!           UNSCALED THE EIGENVALUES.
!
            w(k) = th*tmp1
!
        end do
!
!       ORDER THE neig EIGENVALUES.
!
        if ( present(sort) ) then
            call eigval_sort( sort, w(:neig) )
        end if
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( lb, nlb, ub, nub )
!
    end if
!
!   FLAG THE OTHER EIGENVALUES IF NECESSARY.
!
    if ( neig/=n ) then
!
        th = nan()
        w(neig+1_i4b:n) = th
!
    end if
!
!
! END OF SUBROUTINE symtrid_bisect
! ________________________________
!
    end subroutine symtrid_bisect
!
!
! =========================================================================================
!                         DEFLATION PROCEDURES FOR SYMMETRIC TRIDIAGONAL MATRICES
! =========================================================================================
!
!
    subroutine dflgen( d, e, lambda, cs, sn )
!
! Purpose
! _______
!
!   DFLGEN computes deflation parameters (e.g. a chain of Givens rotations) for
!   a symmetric unreduced n-by-n tridiagonal matrix T and a given eigenvalue of T.
!
!   On output, the arguments CS and SN contain, respectively, the vectors of the
!   cosines and sines coefficients of the chain of n-1 planar rotations that
!   deflates the real n-by-n symmetric tridiagonal matrix T corresponding to
!   an eigenvalue LAMBDA.
!
!
! Arguments
! _________
!
!   D           (INPUT) real(stnd), dimension(:)
!               On entry, the diagonal elements of the tridiagonal matrix.
!
!   E           (INPUT) real(stnd), dimension(:)
!               On entry, the n-1 subdiagonal elements of the tridiagonal matrix.
!
!               The size of E must verify:  size( E ) = size( D ) - 1.
!
!   LAMBDA      (INPUT) real(stnd)
!               On entry, an eigenvalue of the tridiagonal matrix.
!
!   CS          (OUTPUT) real(stnd), dimension(:)
!               On exit, the vector of the cosines coefficients of the chain of n-1 Givens
!               rotations that deflates the symmetric tridiagonal matrix.
!
!               The size of CS must verify:  size( CS ) = size( E ) = size( D ) - 1.
!
!   SN          (OUTPUT) real(stnd), dimension(:)
!               On exit, the vector of the sines coefficients of the chain of n-1 Givens
!               rotations that deflates the symmetric tridiagonal matrix.
!
!               The size of SN must verify:  size( SN ) = size( E ) = size( D ) - 1.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the matlab routine DFLGEN given in the reference (1).
!   No check is done in the subroutine to verify that the input tridiagonal matrix is
!   unreduced.
!
!   For further details, see:
!
!   (1) Malyshev, A.N., 2000:
!           On deflation for symmetric tridiagonal matrices.
!           Report 182 of the Department of Informatics, University of Bergen, Norway.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,    only : lamch, machtiny, machbase
    use Utilities,        only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),                 intent(in)  :: lambda
    real(stnd), dimension(:),   intent(in)  :: d, e
    real(stnd), dimension(:),   intent(out) :: cs, sn
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: eps, safmin, lambda2, tnorm, xim, dmin, tau, f, cf, sfi
    real(stnd), dimension(size(d)) :: diag, d1, d2, diaginv
    real(stnd), dimension(size(e)) :: dl, u1, l2, sf
!
    integer(i4b)                     :: n, i, k, ei, ce, sei
    integer(i4b), dimension(size(e)) :: se
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='dflgen'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(d),i4b)         ,   &
                    int(size(e),i4b)  + 1_i4b,   &
                    int(size(cs),i4b) + 1_i4b,   &
                    int(size(sn),i4b) + 1_i4b,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) return
!
!   DETERMINE THE UNIT ROUNDOFF AND UNDERFLOW THRESHOLD.
!
    eps    = lamch( 'u' )
    safmin = lamch( 's' )
!
    tnorm = max( maxval( abs(d(1_i4b:n))), maxval( abs(e(1_i4b:n-1_i4b))) )
!
!   SCALE THE ELEMENTS OF THE TRIDIAGONAL MATRIX AND THE EIGENVALUE
!   BY A POWER OF THE BASE OF THE SYSTEM.
!
    if ( ( tnorm<safmin/eps ) .or. ( tnorm>eps/safmin ) ) then
!
        f  = fraction( tnorm )
        ei = exponent( tnorm )
        tnorm = f
!
!BUG:   FOR COMPILING STATPACK WITH XLF COMPILER, DON'T USE THE INTRINSIC SCALE FUNCTION
!       WHICH SEEMS TO NOT WORK CORRECTLY.
!
!        diag(:n)     = scale( d(:n), -ei )
!        dl(:n-1_i4b) = scale( e(:n-1_i4b), -ei )
!        lambda2      = scale( lambda, -ei )
!
        xim = machbase**(-ei)
        diag(:n)     = d(:n)*xim
        dl(:n-1_i4b) = e(:n-1_i4b)*xim
        lambda2      = lambda*xim
!
    else
!
        diag(:n)     = d(:n)
        dl(:n-1_i4b) = e(:n-1_i4b)
        lambda2      = lambda
!
    end if
!
!   REPLACE SMALL SUBDIAGONALS BY safmin.
!
    where ( abs(dl(1_i4b:n-1_i4b))<=safmin )                      &
          dl(1_i4b:n-1_i4b) = sign( safmin, dl(1_i4b:n-1_i4b) )
!
!   REPLACEMENT FOR VERY SMALL PIVOTS IN LDU AND UDL FACTORIZATIONS.
!
    dmin = max( safmin, (safmin*tnorm)*tnorm ) 
!    
!   FERNANDO METHOD.  
!
!   L1*D1*U1 FACTORIZATION OF T - lambda*I .
!
    xim            = diag(1_i4b) - lambda2
    diaginv(1_i4b) = zero
!
    if ( abs(xim)<dmin ) then
        xim = sign( dmin, xim )
    end if
    d1(1_i4b) = xim
!
    do i=2_i4b, n
!
        u1(i-1_i4b) = dl(i-1_i4b)/xim
        diaginv(i)  = - dl(i-1_i4b)*u1(i-1_i4b)
        xim         = ( diag(i) + diaginv(i) ) - lambda2
!
        if ( abs(xim)<dmin ) then
            xim = sign( dmin, xim )
        end if
        d1(i) = xim
!
    end do
!
!   U2*D2*L2 FACTORIZATION OF T - lambda*I .
!
!   FIND OPTIMAL k AND tau .
!
    xim = diag(n) - lambda2
    tau = abs( diaginv(n) + xim )
    k   = n
!
    if ( abs(xim)<dmin ) then
        xim = sign( dmin, xim )
    end if
    d2(n) = xim
!
    do i=n-1_i4b, 1_i4b, -1_i4b
!
        l2(i) = dl(i)/xim
        xim   = (diag(i) - dl(i)*l2(i)) - lambda2
        cf    = abs( diaginv(i) + xim )
!
        if ( abs(xim)<dmin ) then
            xim = sign( dmin, xim )
        end if
        d2(i) = xim
!
        if ( cf>=tau ) cycle
!
        tau = cf
        k   = i
!
    end do
!
!   FROM HERE USE FLOATING POINT NUMBERS WITH EXTENDED
!   EXPONENTS FOR THE COMPUTATIONS.
!
    where ( abs(u1(1_i4b:k-1_i4b))>machtiny )
!
!       REPRESENT u1(:k-1) IN THE FORM WITH EXTENDED EXPONENTS.
!
        sf(1_i4b:k-1_i4b) = fraction( u1(1_i4b:k-1_i4b) )
        se(1_i4b:k-1_i4b) = exponent( u1(1_i4b:k-1_i4b) )
!
    elsewhere
!
!       RECOMPUTE u1(:k-1) IN THE FORM WITH EXTENDED EXPONENTS.
!
        sf(1_i4b:k-1_i4b) = fraction( dl(1_i4b:k-1_i4b) )/fraction( d1(1_i4b:k-1_i4b) )
        se(1_i4b:k-1_i4b) = exponent( dl(1_i4b:k-1_i4b) ) - exponent( d1(1_i4b:k-1_i4b) )
!
    end where
!
    where ( abs(l2(k:n-1_i4b))>machtiny )
!
!       REPRESENT 1/l2(k:) IN THE FORM WITH EXTENDED EXPONENTS.
!
        sf(k:n-1_i4b) = one/fraction( l2(k:n-1_i4b) )
        se(k:n-1_i4b) = -exponent( l2(k:n-1_i4b) )                
!
    elsewhere
!
!       RECOMPUTE 1/l2(k:) IN THE FORM WITH EXTENDED EXPONENTS.
!
        sf(k:n-1_i4b) = fraction( d2(k+1_i4b:n) )/fraction( dl(k:n-1_i4b) )
        se(k:n-1_i4b) = exponent( d2(k+1_i4b:n) ) - exponent( dl(k:n-1_i4b) )
!
    end where
!
    cf = fraction( one )
    ce = exponent( one )
!
    do i=1_i4b, n-1_i4b
!
!       CONVERT cf*(base**ce) TO cs(i) AND sf(i)*(base**se(i)) TO sn(i).
!
        sfi = sf(i)
        sei = se(i)
        ei  = max( ce, sei )
        ce  = ce - ei
        sei = sei - ei
!
!BUG:   FOR COMPILING STATPACK WITH XLF COMPILER, DON'T USE THE INTRINSIC SCALE FUNCTION
!       WHICH SEEMS TO NOT WORK CORRECTLY.
!
!        f   = sqrt( (scale(cf,ce))**2 + (scale(sfi,sei))**2 )
!
        f   = sqrt( (cf*(machbase**ce))**2 + (sfi*(machbase**sei))**2 )
!
!       SAVE cs(i) IN THE FORM WITH EXTENDED EXPONENTS.
!
        xim = cf/f
        cf  = fraction( xim )
        ce  = ce + exponent( xim )
!
!       SAVE COSINES AND SINES OF GIVENS ROTATIONS.
!
!
!BUG:   FOR COMPILING STATPACK WITH XLF COMPILER, DON'T USE THE INTRINSIC SCALE FUNCTION
!       WHICH SEEMS TO NOT WORK CORRECTLY.
!
!        cs(i) = scale( cf, ce )
!        sn(i) = scale( sfi/f, sei )
!
        cs(i) = cf*(machbase**ce)
        sn(i) = (sfi/f)*(machbase**sei)
!
    end do
!
!
! END OF SUBROUTINE dflgen
! ________________________
!
    end subroutine dflgen
!
! =========================================================================================
!
    subroutine dflgen2( d, e, lambda, cs, sn, deflate )
!
! Purpose
! _______
!
!   DFLGEN2 computes and applies deflation parameters (e.g. a chain of Givens rotations)
!   for a symmetric unreduced n-by-n tridiagonal matrix T and a given eigenvalue of T.
!
!   On output:
!
!       the arguments D and E contain, respectively, the new main diagonal
!       and subdiagonal of the deflated symmetric tridiagonal matrix if
!       DEFLATE is set to true.
!
!       the arguments CS and SN contain, respectively, the vectors of the
!       cosines and sines coefficients of the chain of n-1 planar rotations that
!       deflates the real n-by-n symmetric tridiagonal matrix T corresponding to
!       the eigenvalue LAMBDA.
!
!
! Arguments
! _________
!
!   D           (INPUT/OUTPUT) real(stnd), dimension(:)
!               On entry, the diagonal elements of the tridiagonal matrix.
!
!               On exit, the new main diagonal of the symmetric tridiagonal matrix
!               if DEFLATE=true. Otherwise, D is not changed.
!
!   E           (INPUT/OUTPUT) real(stnd), dimension(:)
!               On entry, the n-1 subdiagonal elements of the tridiagonal matrix.
!
!               On exit, the new subdiagonal of the symmetric tridiagonal matrix
!               if DEFLATE=true. Otherwise, E is not changed.
!
!               The size of E must verify:  size( E ) = size( D ) - 1.
!
!   LAMBDA      (INPUT) real(stnd)
!               On entry, an eigenvalue of the tridiagonal matrix.
!
!   CS          (OUTPUT) real(stnd), dimension(:)
!               On exit, the vector of the cosines coefficients of the chain of n-1 Givens
!               rotations that deflates the symmetric tridiagonal matrix.
!
!               The size of CS must verify:  size( CS ) = size( E ) = size( D ) - 1.
!
!   SN          (OUTPUT) real(stnd), dimension(:)
!               On exit, the vector of the sines coefficients of the chain of n-1 Givens
!               rotations that deflates the symmetric tridiagonal matrix.
!
!               The size of SN must verify:  size( SN ) = size( E ) = size( D ) - 1.
!
!   DEFLATE     (OUTPUT) logical(lgl)
!               On exit:
!
!               - DEFLATE = true :  indicates successful exit.
!               - DEFLATE = false:  indicates that full accuracy was not attained
!                 in the deflation of the tridiagonal matrix.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the matlab routine DFLGEN given in the reference (1).
!   No check is done in the subroutine to verify that the input tridiagonal matrix is
!   unreduced.
!
!   For further details, see:
!
!   (1) Malyshev, A.N., 2000:
!           On deflation for symmetric tridiagonal matrices.
!           Report 182 of the Department of Informatics, University of Bergen, Norway.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,    only : lamch, machtiny, machbase
    use Utilities,        only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),                 intent(in)    :: lambda
    real(stnd), dimension(:),   intent(inout) :: d, e
    real(stnd), dimension(:),   intent(out)   :: cs, sn
!
    logical(lgl),               intent(out)   :: deflate
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: eps, safmin, lambda2, tnorm, xim, dmin, tau,  &
                                      f, cf, sfi, beta, a, b, g, r, t, c, s
!    real(stnd)                     :: p
    real(stnd), dimension(size(d)) :: diag, d1, d2, diaginv
    real(stnd), dimension(size(e)) :: dl, u1, l2, sf
!
    integer(i4b)                     :: n, i, k, ei, ce, sei
    integer(i4b), dimension(size(e)) :: se
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='dflgen2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    deflate = true
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(d),i4b)         ,   &
                    int(size(e),i4b)  + 1_i4b,   &
                    int(size(cs),i4b) + 1_i4b,   &
                    int(size(sn),i4b) + 1_i4b,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) return
!
!   DETERMINE THE UNIT ROUNDOFF AND UNDERFLOW THRESHOLD.
!
    eps    = lamch( 'u' )
    safmin = lamch( 's' )
!
    tnorm = max( maxval( abs(d(1_i4b:n))), maxval( abs(e(1_i4b:n-1_i4b))) )
!
!   SCALE THE ELEMENTS OF THE TRIDIAGONAL MATRIX AND THE EIGENVALUE
!   BY A POWER OF THE BASE OF THE SYSTEM.
!
    if ( ( tnorm<safmin/eps ) .or. ( tnorm>eps/safmin ) ) then
!
        f  = fraction( tnorm )
        ei = exponent( tnorm )
        tnorm = f
!
!BUG:   FOR COMPILING STATPACK WITH XLF COMPILER, DON'T USE THE INTRINSIC SCALE FUNCTION
!       WHICH SEEMS TO NOT WORK CORRECTLY.
!
!        diag(:n)     = scale( d(:n), -ei )
!        dl(:n-1_i4b) = scale( e(:n-1_i4b), -ei )
!        lambda2      = scale( lambda, -ei )
!
        xim = machbase**(-ei)
        diag(:n)     = d(:n)*xim
        dl(:n-1_i4b) = e(:n-1_i4b)*xim
        lambda2      = lambda*xim
!
    else
!
        diag(:n)     = d(:n)
        dl(:n-1_i4b) = e(:n-1_i4b)
        lambda2      = lambda
!
    end if
!
!   REPLACE SMALL SUBDIAGONALS BY safmin.
!
    where ( abs(dl(1_i4b:n-1_i4b))<=safmin ) dl(1_i4b:n-1_i4b) = sign( safmin, dl(1_i4b:n-1_i4b) )
!
!   REPLACEMENT FOR VERY SMALL PIVOTS IN LDU AND UDL FACTORIZATIONS.
!
    dmin = max( safmin, (safmin*tnorm)*tnorm ) 
!    
!   FERNANDO METHOD.  
!
!   L1*D1*U1 FACTORIZATION OF T - lambda*I .
!
    xim            = diag(1_i4b) - lambda2
    diaginv(1_i4b) = zero
!
    if ( abs(xim)<dmin ) then
        xim = sign( dmin, xim )
    end if
!
    d1(1_i4b) = xim
!
    do i=2_i4b, n
!
        u1(i-1_i4b) = dl(i-1_i4b)/xim
        diaginv(i)  = - dl(i-1_i4b)*u1(i-1_i4b)
        xim         = ( diag(i) + diaginv(i) ) - lambda2
!
        if ( abs(xim)<dmin ) then
            xim = sign( dmin, xim )
        end if
!
        d1(i) = xim
!
    end do
!
!   U2*D2*L2 FACTORIZATION OF T - lambda*I .
!
!   FIND OPTIMAL k AND tau .
!
    xim = diag(n) - lambda2
    tau = abs( diaginv(n) + xim )
    k   = n
!
    if ( abs(xim)<dmin ) then
        xim = sign( dmin, xim )
    end if
    d2(n) = xim
!
    do i=n-1_i4b, 1_i4b, -1_i4b
!
        l2(i) = dl(i)/xim
        xim   = (diag(i) - dl(i)*l2(i)) - lambda2
        cf    = abs( diaginv(i) + xim )
!
        if ( abs(xim)<dmin ) then
            xim = sign( dmin, xim )
        end if
        d2(i) = xim
!
        if ( cf>=tau ) cycle
!
        tau = cf
        k   = i
!
    end do
!
!    k = maxdiag_tinv_qr( diag(:n), dl(:n-1_i4b), lambda2 )
!
!
!   FROM HERE USE FLOATING POINT NUMBERS WITH EXTENDED
!   EXPONENTS FOR THE COMPUTATIONS.
!
    where ( abs(u1(1_i4b:k-1_i4b))>machtiny )
!
!       REPRESENT u1(:k-1) IN THE FORM WITH EXTENDED EXPONENTS.
!
        sf(1_i4b:k-1_i4b) = fraction( u1(1_i4b:k-1_i4b) )
        se(1_i4b:k-1_i4b) = exponent( u1(1_i4b:k-1_i4b) )
!
    elsewhere
!
!       RECOMPUTE u1(:k-1) IN THE FORM WITH EXTENDED EXPONENTS.
!
        sf(1_i4b:k-1_i4b) = fraction( dl(1_i4b:k-1_i4b) )/fraction( d1(1_i4b:k-1_i4b) )
        se(1_i4b:k-1_i4b) = exponent( dl(1_i4b:k-1_i4b) ) - exponent( d1(1_i4b:k-1_i4b) )
!
    end where
!
    where ( abs(l2(k:n-1_i4b))>machtiny )
!
!       REPRESENT 1/l2(k:) IN THE FORM WITH EXTENDED EXPONENTS.
!
        sf(k:n-1_i4b) = one/fraction( l2(k:n-1_i4b) )
        se(k:n-1_i4b) = -exponent( l2(k:n-1_i4b) )                
!
    elsewhere
!
!       RECOMPUTE 1/l2(k:) IN THE FORM WITH EXTENDED EXPONENTS.
!
        sf(k:n-1_i4b) = fraction( d2(k+1_i4b:n) )/fraction( dl(k:n-1_i4b) )
        se(k:n-1_i4b) = exponent( d2(k+1_i4b:n) ) - exponent( dl(k:n-1_i4b) )
!
    end where
!
    cf = fraction( one )
    ce = exponent( one )
!
    tau  = d(1_i4b)
    beta = e(1_i4b)
!
    r = zero
    t = zero
!    p = zero
!
!   COMPUTE AND APPLY DEFLATION PARAMETERS.
!
    do i=1_i4b, n-1_i4b
!
        sfi = sf(i)
        sei = se(i)
!
        ei  = max( ce, sei )
        ce  = ce - ei
        sei = sei - ei
!
!BUG:   FOR COMPILING STATPACK WITH XLF COMPILER, DON'T USE THE INTRINSIC SCALE FUNCTION
!       WHICH SEEMS TO NOT WORK CORRECTLY.
!
!        f   = sqrt( (scale(cf,ce))**2 + (scale(sfi,sei))**2 )
!
        f   = sqrt( (cf*(machbase**ce))**2 + (sfi*(machbase**sei))**2 )
!
!       SAVE cs(i) IN THE FORM WITH EXTENDED EXPONENTS.
!
        xim = cf/f
        cf  = fraction( xim )
        ce  = ce + exponent( xim )
!
!BUG:   FOR COMPILING STATPACK WITH XLF COMPILER, DON'T USE THE INTRINSIC SCALE FUNCTION
!       WHICH SEEMS TO NOT WORK CORRECTLY.
!
!        c = scale( cf, ce )
!        s = scale( sfi/f, sei )
!
        c = cf*(machbase**ce)
        s = (sfi/f)*(machbase**sei)
!
!       SAVE COSINES AND SINES OF THE GIVENS ROTATION.
!
        cs(i) = c
        sn(i) = s
!
!       APPLY THE GIVENS ROTATION TO THE TRIDIAGONAL MATRIX.
!
        d2(i) = c*r + s*t
!
!        xim   = abs( c*t - s*r )
!        p     = max( p, xim )
!
        xim = s*beta
        a   = c*tau + xim
        g   = c*d(i+1_i4b) - xim
!
        xim = c*beta
        b   = xim + s*d(i+1_i4b)
        f   = xim - s*tau
!
        diag(i) = c*a + s*b
        r       = c*b - s*a
        tau     = c*g - s*f
!
        if ( i==n-1_i4b ) exit
!
        t    = s*e(i+1_i4b)
        beta = c*e(i+1_i4b)
!
    end do
!
    diag(n) = tau
    d2(n)   = r
!
!   CHECK ACCURACY OF THE DEFLATION.
!
!    p = max( p, abs(r) )
!    t = p*p
!
    t = r*r
!
    deflate = t<=( fudge_deflate*eps*abs( diag(n)*diag(n-1_i4b) ) + safmin )
!
!   OUTPUT THE NEW TRIDIAGONAL MATRIX IF DEFLATION IS SUCCESSFUL.
!
    if ( deflate ) then
        d(1_i4b:n)       = diag(1_i4b:n)
        e(1_i4b:n-1_i4b) = d2(2_i4b:n)
    end if
!
!
! END OF SUBROUTINE dflgen2
! _________________________
!
    end subroutine dflgen2
!
! =========================================================================================
!
    subroutine dflapp( d, e, cs, sn, deflate )
!
! Purpose
! _______
!
!   DFLAPP deflates a real symmetric n-by-n tridiagonal matrix T by a chain
!   of planar rotations produced by DFLGEN.
!
!   On output, the arguments D and E contain, respectively, the new main diagonal
!   and subdiagonal of the deflated symmetric tridiagonal matrix if DEFLATE is set
!   to true.
!
!
! Arguments
! _________
!
!   D           (INPUT/OUTPUT) real(stnd), dimension(:)
!               On entry, the diagonal elements of the tridiagonal matrix.
!
!               On exit, the new main diagonal of the symmetric tridiagonal matrix
!               if DEFLATE=true. Otherwise, D is not changed.
!
!   E           (INPUT/OUTPUT) real(stnd), dimension(:)
!               On entry, the n-1 subdiagonal elements of the tridiagonal matrix.
!
!               On exit, the new subdiagonal of the symmetric tridiagonal matrix
!               if DEFLATE=true. Otherwise, E is not changed.
!
!               The size of E must verify:  size( E ) = size( D ) - 1.
!
!   CS          (INPUT) real(stnd), dimension(:)
!               On entry, the vector of the cosines coefficients of the chain of n-1 Givens
!               rotations that deflates the symmetric tridiagonal matrix as computed by
!               DFLGEN subroutine.
!
!               The size of CS must verify:  size( CS ) = size( E ) = size( D ) - 1.
!
!   SN          (INPUT) real(stnd), dimension(:)
!               On entry, the vector of the sines coefficients of the chain of n-1 Givens
!               rotations that deflates the symmetric tridiagonal matrix as computed by
!               DFLGEN subroutine.
!
!               The size of SN must verify:  size( SN ) = size( E ) = size( D ) - 1.
!
!   DEFLATE     (OUTPUT) logical(lgl)
!               On exit:
!
!               - DEFLATE = true :  indicates successful exit.
!               - DEFLATE = false:  indicates that full accuracy was not attained
!                 in the deflation of the tridiagonal matrix.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the matlab routine DFLAPP given in the reference (1).
!
!   For further details, see:
!
!   (1) Malyshev, A.N., 2000:
!           On deflation for symmetric tridiagonal matrices.
!           Report 182 of the Department of Informatics, University of Bergen, Norway.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,    only : lamch
    use Utilities,        only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:),   intent(inout)  :: d, e
    real(stnd), dimension(:),   intent(in)     :: cs, sn
!
    logical(lgl),   intent(out)                :: deflate

!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: alpha, beta, a, b, f, g, r, t, c, s,  &
                                      tmp, eps, safmin
!    real(stnd)                     :: p
    real(stnd), dimension(size(d)) :: d2, e2
!
    integer(i4b)                   :: n, i
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='dflapp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    deflate = true
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(d),i4b)         ,   &
                    int(size(e),i4b)  + 1_i4b,   &
                    int(size(cs),i4b) + 1_i4b,   &
                    int(size(sn),i4b) + 1_i4b,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) return
!
!   DETERMINE THE UNIT ROUNDOFF AND UNDERFLOW THRESHOLDS.
!
    eps    = lamch( 'u' )
    safmin = lamch( 's' )
!
    alpha = d(1_i4b)
    beta  = e(1_i4b)
!
    r = zero
    t = zero
!    p = zero
!
!   DEFLATE THE INPUT TRIDIAGONAL MATRIX.
!
    do i=1_i4b, n-1_i4b
!
        c = cs(i)
        s = sn(i)
!
!       APPLY THE GIVENS ROTATION TO THE TRIDIAGONAL MATRIX.
!
        e2(i) = c*r + s*t
!
!        tmp   = abs( c*t - s*r )
!        p     = max( p, tmp )
!
        tmp = s*beta
        a   = c*alpha + tmp
        g   = c*d(i+1_i4b) - tmp
!
        tmp = c*beta
        b   = tmp + s*d(i+1_i4b)
        f   = tmp - s*alpha
!
        d2(i) = c*a + s*b
        r     = c*b - s*a
        alpha = c*g - s*f
!
        if ( i==n-1_i4b ) exit
!
        t    = s*e(i+1_i4b)
        beta = c*e(i+1_i4b)
!
    end do
!
    d2(n) = alpha
    e2(n) = r
!
!   CHECK ACCURACY OF THE DEFLATION.
!
!    p = max( p, abs(r) )
!    t = p*p
!
    t = r*r
!
    deflate = t<=(fudge_deflate*eps*abs( d2(n)*d2(n-1_i4b) ) + safmin )
!
!   OUTPUT THE NEW TRIDIAGONAL MATRIX IF DEFLATION IS SUCCESSFUL.
!
    if ( deflate ) then
!
        d(1_i4b:n)       = d2(1_i4b:n)
        e(1_i4b:n-1_i4b) = e2(2_i4b:n)
!
    end if
!
!
! END OF SUBROUTINE dflapp
! ________________________
!
    end subroutine dflapp
!
! =========================================================================================
!
    subroutine qrstep( d, e, lambda, cs, sn, deflate )
!
! Purpose
! _______
!
!   QRSTEP performs one QR step with a given shift LAMBDA on a n-by-n real symmetric
!   unreduced tridiagonal matrix T.
!
!   On output, the arguments D and E contain, respectively, the new main diagonal
!   and subdiagonal of the deflated symmetric tridiagonal. The chain of n-1 planar
!   rotations produced during the QR step are saved in the arguments CS and SN.
!
!
! Arguments
! _________
!
!   D           (INPUT/OUTPUT) real(stnd), dimension(:)
!               On entry, the diagonal elements of the tridiagonal matrix.
!
!               On exit, the new main diagonal of the symmetric tridiagonal matrix
!               after the QR step.
!
!   E           (INPUT/OUTPUT) real(stnd), dimension(:)
!               On entry, the n-1 subdiagonal elements of the tridiagonal matrix.
!
!               On exit, the new subdiagonal of the symmetric tridiagonal matrix
!               after the QR step.
!
!               The size of E must verify:  size( E ) = size( D ) - 1.
!
!   LAMBDA      (INPUT) real(stnd)
!               On entry, the shift used in the current QR step.
!
!   CS          (OUTPUT) real(stnd), dimension(:)
!               On exit, the vector of the cosines coefficients of the chain of n-1 Givens
!               rotations for the current QR step.
!
!               The size of CS must verify:  size( CS ) = size( E ) = size( D ) - 1.
!
!   SN          (OUTPUT) real(stnd), dimension(:)
!               On exit, the vector of the sines coefficients of the chain of n-1 Givens
!               rotations for the current QR step.
!
!               The size of SN must verify:  size( SN ) = size( E ) = size( D ) - 1.
!
!   DEFLATE     (OUTPUT) logical(lgl)
!               On exit:
!
!               - DEFLATE = true :  indicates that deflation occured at the end of the step.
!               - DEFLATE = false:  indicates that the last subdiagonal element of the
!                 tridiagonal matrix is not small.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the matlab routine QRSTEP given in the reference (1).
!   No check is done in the subroutine to verify that the input tridiagonal matrix is
!   unreduced.
!
!   For further details, see:
!
!   (1) Mastronardi, M., Van Barel, M., Van Camp, E., and Vandebril, R., 2006:
!           On computing the eigenvectors of a class of structured matrices.
!           Journal of Computational and Applied Mathematics, 189, 580-591.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,    only : lamch
    use Utilities,        only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),                 intent(in)    :: lambda
    real(stnd), dimension(:),   intent(inout) :: d, e
    real(stnd), dimension(:),   intent(out)   :: cs, sn
!
    logical(lgl),               intent(out)   :: deflate
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: eps, safmin, f, b, g, r, s, c, p
!
    integer(i4b) :: n, i
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='qrstep'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    deflate = true
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(d),i4b)         ,   &
                    int(size(e),i4b)  + 1_i4b,   &
                    int(size(cs),i4b) + 1_i4b,   &
                    int(size(sn),i4b) + 1_i4b,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) return
!
!   DETERMINE THE UNIT ROUNDOFF AND UNDERFLOW THRESHOLD.
!
    eps    = lamch( 'u' )
    safmin = lamch( 's' )
!
    g = d(1_i4b) - lambda
!
    s = one
    c = one
    p = zero
!
!   INNER LOOP.
!
    do i = 1_i4b, n-1_i4b
!
        f = s*e(i)
        b = c*e(i)
!
!       DEFINE THE GIVENS ROTATION.
!
!        if ( abs(g)>=abs(f) ) then
        if ( abs(g)>abs(f) ) then
!
            s = f/g
            r = sign( sqrt( one + s*s ), g )
            c = one/r
            s = c*s
!
            if ( i/=1_i4b ) e(i-1_i4b) = g*r
!
        else if ( f/=zero ) then
!
            c = g/f
            r = sign( sqrt( one + c*c ), f )
            s = one/r
            c = s*c
!
            if ( i/=1_i4b ) e(i-1_i4b) = f*r
!
        else
!
            c = zero
            s = one
            if ( i/=1_i4b ) e(i-1_i4b) = zero
!
        end if
!
!       STORE ROTATION.
!
        cs(i) = c
        sn(i) = s
!
!       APPLY ROTATION.
!
        g    = d(i) - p
        r    = ( d(i+1_i4b) - g )*s + two*c*b
        p    = s*r
        d(i) = g + p
        g    = c*r - b
!
    end do
!
    d(n)       = d(n) - p
    e(n-1_i4b) = g
!
!   CHECK IF DEFLATION OCCURED.
!
    deflate = g*g<=( fudge_deflate*eps*abs( d(n)*d(n-1_i4b) ) + safmin )
!
!
! END OF SUBROUTINE qrstep
! ________________________
!
    end subroutine qrstep
!
! =========================================================================================
!
    subroutine prodgiv( cs, sn, x )
!
! Purpose
! _______
!
!   PRODGIV applies a chain of n-1 planar rotations produced by DFLGEN, DFLGEN2 or QRSTEP
!   to a vector of length n.
!
!
! Arguments
! _________
!
!   X           (INPUT/OUTPUT) real(stnd), dimension(:)
!               On entry, the input vector of length n.
!
!               On exit, the product of the chain of the n-1 planar rotations
!               by the input vector.
!
!   CS          (INPUT) real(stnd), dimension(:)
!               On entry, the vector of the cosines coefficients of the chain of n-1 Givens
!               rotations as computed by DFLGEN or DFLGEN2 subroutines.
!
!               The size of CS must verify:  size( CS ) = size( X ) - 1.
!
!   SN          (INPUT) real(stnd), dimension(:)
!               On entry, the vector of the sines coefficients of the chain of n-1 Givens
!               rotations as computed by DFLGEN or DFLGEN2 subroutines.
!
!               The size of SN must verify:  size( SN ) = size( CS ) = size( X ) - 1.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the matlab routine PRODGIV given in the reference (1).
!
!   For further details, see:
!
!   (1) Mastronardi, M., Van Barel, M., Van Camp, E., and Vandebril, R., 2006:
!           On computing the eigenvectors of a class of structured matrices.
!           Journal of Computational and Applied Mathematics, 189, 580-591.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,        only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:),   intent(inout)  :: x
    real(stnd), dimension(:),   intent(in)     :: cs, sn
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: c, s, tmp
!
    integer(i4b)                   :: n, i
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='prodgiv'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(x),i4b)         ,   &
                    int(size(cs),i4b) + 1_i4b,   &
                    int(size(sn),i4b) + 1_i4b,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) return
!
!   COMPUTE THE PRODUCT OF THE GIVENS ROTATIONS TIMES THE VECTOR x(:).
!
    tmp = x(n)
!
    do i = n-1_i4b, 1_i4b, -1_i4b
!
        c = cs(i)
        s = sn(i)
!
!       APPLY THE GIVENS ROTATION TO THE VECTOR x(:).
!
        x(i+1_i4b) = s*x(i) + c*tmp
        tmp        = c*x(i) - s*tmp
!
    end do
!
    x(1_i4b) = tmp
!
!
! END OF SUBROUTINE prodgiv
! _________________________
!
    end subroutine prodgiv
!
! =========================================================================================
!
    function prodgiv_eigvec( cs, sn ) result( x )
!
! Purpose
! _______
!
!   PRODGIV_EIGVEC computes an eigenvector of a n-by-n symmetric tridiagonal matrix from
!   a chain of n-1 planar rotations produced by DFLGEN, DFLGEN2 or QRSTEP.
!
!
! Arguments
! _________
!
!   CS          (INPUT) real(stnd), dimension(:)
!               On entry, the vector of the cosines coefficients of the chain of n-1 Givens
!               rotations as computed by DFLGEN or DFLGEN2 subroutines.
!
!   SN          (INPUT) real(stnd), dimension(:)
!               On entry, the vector of the sines coefficients of the chain of n-1 Givens
!               rotations as computed by DFLGEN or DFLGEN2 subroutines.
!
!               The size of SN must verify:  size( SN ) = size( CS ) = n - 1.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the matlab routine PRODGIV given in the reference (1).
!
!   For further details, see:
!
!   (1) Mastronardi, M., Van Barel, M., Van Camp, E., and Vandebril, R., 2006:
!           On computing the eigenvectors of a class of structured matrices.
!           Journal of Computational and Applied Mathematics, 189, 580-591.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:),   intent(in) :: cs, sn
!
    real(stnd)  :: x(size(cs)+1)
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: tmp
!
    integer(i4b) :: nm1, i
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='prodgiv_eigvec'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    nm1 =  assert_eq( int(size(cs),i4b),   &
                      int(size(sn),i4b),   &
                      name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( nm1<=0_i4b ) then
!
        if ( nm1==0_i4b ) x(1_i4b) = one
!
        return
!
    end if
!
!   COMPUTE THE PRODUCT OF THE GIVENS ROTATIONS TIMES THE VECTOR En.
!
    tmp = one
!
    do i = nm1, 1_i4b, -1_i4b
!
!       APPLY THE GIVENS ROTATION TO THE VECTOR En.
!
        x(i+1_i4b) = cs(i)*tmp
        tmp        = -sn(i)*tmp
!
    end do
!
    x(1_i4b) = tmp
!
!
! END OF FUNCTION prodgiv_eigvec
! ______________________________
!
    end function prodgiv_eigvec
!
! =========================================================================================
!
    subroutine symtrid_deflate_rv( d, e, eigval, eigvec, failure, max_qr_steps )
!
! Purpose
! _______
!
!   SYMTRID_DEFLATE computes an eigenvector of a real symmetric tridiagonal
!   matrix T corresponding to a specified eigenvalue, using a deflation technique.
!
!
! Arguments
! _________
!
!   D              (INPUT) real(stnd), dimension(:)
!                  On entry, the diagonal elements of the symmetric tridiagonal matrix T.
!
!   E              (INPUT) real(stnd), dimension(:)
!                  On entry, the n-1 subdiagonal elements of the symmetric tridiagonal
!                  matrix T.
!
!                  The size of E must verify: size( E ) = size( D ) - 1 = n - 1.
!
!   EIGVAL         (INPUT) real(stnd)
!                  On entry, an eigenvalue of the symmetric tridiagonal matrix.
!
!   EIGVEC         (OUTPUT) real(stnd), dimension(:)
!                  On exit, the computed eigenvector associated with the
!                  eigenvalue EIGVAL.
!
!                  The sise of EIGVEC must verify: size( EIGVEC ) = size( D ) = n .
!
!   FAILURE        (OUTPUT) logical(lgl)
!                  On exit:
!
!                  - FAILURE = false :  indicates successful exit.
!                  - FAILURE = true  :  indicates that the algorithm did not converge and
!                    that full accuracy was not attained in the deflation procedure
!                    of the tridiagonal matrix.
!
!   MAX_QR_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                  MAX_QR_STEPS controls the maximum number of QR sweeps for deflating
!                  the tridiagonal matrix for a given eigenvalue.
!                  The algorithm fails to converge if the total number of QR sweeps
!                  exceeds MAX_QR_STEPS.
!
!                  The default is 4.
!
!
! Further Details
! _______________
!
!   SYMTRID_DEFLATE may fail for some zero-diagonal tridiagonal matrices.
!
!   For further details, see:
!
!   (1) Malyshev, A.N., 2000:
!           On deflation for symmetric tridiagonal matrices.
!           Report 182 of the Department of Informatics, University of Bergen, Norway.
!
!   (2) Mastronardi, M., Van Barel, M., Van Camp, E., and Vandebril, R., 2006:
!           On computing the eigenvectors of a class of structured matrices.
!           Journal of Computational and Applied Mathematics, 189, 580-591.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in)                :: eigval
    real(stnd),  intent(in),  dimension(:) :: d, e
    real(stnd),  intent(out), dimension(:) :: eigvec
!
    integer(i4b), intent(in), optional :: max_qr_steps
!
    logical(lgl), intent(out) :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                  :: safmin, gamma, eps, prec, g, xi, xim, tridnorm
    real(stnd),  dimension(size(d))             :: d2, e2
    real(stnd),  dimension(:,:),    allocatable :: csmat, snmat
!
    integer(i4b) :: n, i, i1, i2, i3, k, maxqrs
    integer      :: iok
!
    logical(lgl) :: deflate
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='symtrid_deflate'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(d),i4b)         ,   &
                    int(size(e),i4b) + 1_i4b ,   &
                    int(size(eigvec),i4b)    ,   &
                    name_proc )
!
    failure = false
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
            eigvec(1_i4b) = one
!
        end if
!
        return
!
    end if
!
!   INITIALIZE COUNTERS FOR QR ITERATIONS.
!
    if ( present(max_qr_steps) ) then
        maxqrs = merge( max_qr_steps, 4_i4b , max_qr_steps>=1_i4b )
    else
        maxqrs = 4_i4b
    end if
!
!   DETERMINE THE UNIT ROUNDOFF AND UNDERFLOW THRESHOLD.
!
    eps    = lamch( 'u' )
    prec   = fudge_prec*lamch( 'p' )
    safmin = lamch( 's' )
    gamma  = sqrt( safmin )
!
!   MAKE A COPY OF THE TRIDIAGONAL MATRIX.
!
    d2(1_i4b:n)       = d(1_i4b:n)
    e2(1_i4b:n-1_i4b) = e(1_i4b:n-1_i4b)
    e2(n)             = zero
!
!   COMPUTE THE NORM OF THE TRIDIAGONAL MATRIX.
!
    tridnorm = zero
    xi       = zero
!
    do i=1_i4b, n
!
        xim       = abs( e2(i) )
        g         = abs( d2(i) ) + xi + xim
        tridnorm  = tridnorm + g
!        tridnorm  = max( tridnorm, g )
        xi        = xim
!
    end do
!
    g = real( n, stnd)*prec*tridnorm
!
!   TEST IF THE UPPER PART OF THE TRIDIAGONAL MATRIX IS NOT DIAGONAL OR ZERO.
!
    do i1=1_i4b, n-1_i4b
!
!        if ( abs( e2(i1) )>gamma .or. abs( d2(i1) )>gamma ) exit
!
        if ( abs( e2(i1) )>gamma ) exit
        if ( abs( d2(i1) )>gamma .and. abs( d2(i1) - eigval )<=g ) exit
!
    end do
!
!   TEST IF THE LOWER PART OF THE TRIDIAGONAL MATRIX IS NOT DIAGONAL OR ZERO.
!
    do i2=n, 2_i4b, -1_i4b
!
!        if ( abs( e2(i2-1_i4b) )>gamma .or. abs( d2(i2) )>gamma ) exit
!
        if ( abs( e2(i2-1_i4b) )>gamma ) exit
        if ( abs( d2(i2) )>gamma .and. abs( d2(i2) - eigval )<=g ) exit
!
    end do
!
    if ( ( abs( eigval )<=gamma+g .and. ( i1/=1_i4b .or. i2/=n ) )   &
          .or. ( i2-i1<0_i4b ) ) then
!
!       IF eigval IS VERY SMALL, WE CANNOT WORK ON THE REDUCED PORTION
!       OF THE TRIDIAGONAL MATRIX.
!
        i1 = 1_i4b
        i2 = n
!
    end if
!
    i3 = i2 - i1
!
!   ALLOCATE WORK VARIABLES NEEDED TO STORE THE GIVENS ROTATIONS.
!
    allocate( csmat(i3,maxqrs), snmat(i3,maxqrs), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    csmat(1_i4b:i3,1_i4b:maxqrs) = one
    snmat(1_i4b:i3,1_i4b:maxqrs) = zero
!
!   COMPUTE DEFLATION PARAMETERS.
!
!   TRY FIRST FERNANDO METHOD COMBINED WITH GODUNOV DEFLATION.
!
    call dflgen2( d2(i1:i2), e2(i1:i2-1_i4b), eigval,                   &
                  csmat(1_i4b:i3,1_i4b), snmat(1_i4b:i3,1_i4b), deflate )
!
    if ( .not.deflate ) then
!
!       USE QR STEPS IF DEFLATION FAILED WITH GODUNOV DEFLATION.
!
        do k=1_i4b, maxqrs
!
            g = sqrt( eps*abs( d2(i2)*d2(i2-1_i4b) ) ) + safmin
!
!           REPLACE SMALL SUBDIAGONALS.
!
            where ( abs(e2(i1:i2-2_i4b))<g )                   &
                    e2(i1:i2-2_i4b) = sign( g, e2(i1:i2-2_i4b) )
!
            call qrstep( d2(i1:i2), e2(i1:i2-1_i4b), eigval,           &
                         csmat(1_i4b:i3,k), snmat(1_i4b:i3,k), deflate )
!
            if ( deflate ) exit
!
        end do
!
        k = min( k, maxqrs )
!
    else
!
        k = 1_i4b
!
    end if
!
    failure = .not.deflate
!
!   NOW COMPUTE THE EIGENVECTOR.
!
    eigvec(1_i4b:i1-1_i4b) = zero
    eigvec(i1:i2)          = prodgiv_eigvec( csmat(1_i4b:i3,k), snmat(1_i4b:i3,k) )
    eigvec(i2+1_i4b:n)     = zero
!
    do i=k-1_i4b, 1_i4b, -1_i4b
        call prodgiv( csmat(1_i4b:i3,i), snmat(1_i4b:i3,i), eigvec(i1:i2) )
    end do
!    
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( csmat, snmat )
!
!
! END OF SUBROUTINE symtrid_deflate_rv
! ____________________________________
!
    end subroutine symtrid_deflate_rv
!
! =========================================================================================
!
    subroutine symtrid_deflate_rm( d, e, eigval, eigvec, failure, max_qr_steps )
!
! Purpose
! _______
!
!   SYMTRID_DEFLATE computes eigenvectors of a real symmetric tridiagonal
!   matrix T corresponding to specified eigenvalues, using a deflation technique.
!
!
! Arguments
! _________
!
!   D              (INPUT) real(stnd), dimension(:)
!                  On entry, the diagonal elements of the symmetric tridiagonal matrix T.
!
!   E              (INPUT) real(stnd), dimension(:)
!                  On entry, the n-1 subdiagonal elements of the symmetric tridiagonal
!                  matrix T.
!
!                  The size of E must verify:  size( E ) = size( D ) - 1 = n - 1.
!
!   EIGVAL         (INPUT) real(stnd), dimension(:)
!                  On entry, selected eigenvalues of the symmetric tridiagonal matrix.
!                  The eigenvalues can be given in any order.
!
!                  The size of EIGVAL must verify:  size( EIGVAL ) <= size( D ) = n .
!
!   EIGVEC         (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed eigenvectors. The eigenvector associated
!                  with the eigenvalue EIGVAL(j) is stored in the j-th column of EIGVEC.
!
!                  The shape of EIGVEC must verify:
!
!                  - size( EIGVEC, 1 ) = size( D ) = n ;
!                  - size( EIGVEC, 2 ) = size( EIGVAL ).
!
!   FAILURE        (OUTPUT) logical(lgl), dimension(:)
!                  On exit:
!
!                  - FAILURE(j) = false : indicates successful exit for the jth eigenvector.
!                  - FAILURE(j) = true  : indicates that the algorithm did not converge and
!                    full accuracy was not attained in the deflation procedure
!                    of the tridiagonal matrix for the jth eigenvector.
!
!                  The size of FAILURE must verify:  size( FAILURE ) = size( EIGVAL ) .
!
!   MAX_QR_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                  MAX_QR_STEPS controls the maximum number of QR sweeps for deflating
!                  the tridiagonal matrix for a given eigenvalue.
!                  The algorithm fails to converge if the total number of QR sweeps for
!                  all eigenvalues exceeds MAX_QR_STEPS * size(EIGVAL).
!
!                  The default is 4.
!
!
! Further Details
! _______________
!
!   SYMTRID_DEFLATE may fail if some the eigenvalues specified in parameter EIGVAL are nearly
!   identical or for clusters of small eigenvalues or for some zero-diagonal tridiagonal matrices.
!
!   For further details, see:
!
!   (1) Malyshev, A.N., 2000:
!           On deflation for symmetric tridiagonal matrices.
!           Report 182 of the Department of Informatics, University of Bergen, Norway.
!
!   (2) Mastronardi, M., Van Barel, M., Van Camp, E., and Vandebril, R., 2006:
!           On computing the eigenvectors of a class of structured matrices.
!           Journal of Computational and Applied Mathematics, 189, 580-591.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert_eq, assert, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),  dimension(:)   :: d, e, eigval
    real(stnd),  intent(out), dimension(:,:) :: eigvec
!
    integer(i4b), intent(in), optional :: max_qr_steps
!
    logical(lgl), intent(out),  dimension(:) :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                  :: safmin, gamma, eps, prec, g, tmp, c, s, xi, xim, tridnorm
    real(stnd),  dimension(size(d))             :: d2, e2, eigvec2
    real(stnd),  dimension(:,:),    allocatable :: csmat, snmat
!
    integer(i4b)                          :: n, m, i, i1, i2, j, l, k, p, q, maxqrs, maxqrs2, ndeflate
    integer(i4b), dimension(size(eigval)) :: jstart, jend, ind
    integer                               :: iok
!
    logical(lgl)                          :: done
    logical(lgl), dimension(size(eigval)) :: deflate
!
#ifdef _OPENMP
!
    logical :: test_par, ompparallel
!
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='symtrid_deflate'
!
! NUMBER OF ITERATIONS FOR DEFLATING THE TRIDIAGONAL MATRIX WITH FERNANDO-GODUNOV METHOD.
!
    integer(i4b), parameter :: maxiter=2_i4b
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(d),i4b)         ,   &
                    int(size(e),i4b) + 1_i4b ,   &
                    int(size(eigvec,1),i4b)  ,   &
                    name_proc )
!
    m =  assert_eq( int(size(eigval),i4b)    ,   &
                    int(size(failure),i4b)   ,   &
                    int(size(eigvec,2),i4b)  ,   &
                    name_proc )
!
    call assert( logical( m<=n, lgl ), name_proc )
!
    failure(1_i4b:m) = false
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b .or. m<=0_i4b) then
!
        if ( n==1_i4b .and. m==1_i4b ) then
!
            eigvec(1_i4b,1_i4b) = one
!
        end if
!
        return
!
    end if
!
!   INITIALIZE COUNTERS FOR QR ITERATIONS.
!
    if ( present(max_qr_steps) ) then
        maxqrs = merge( max_qr_steps, 4_i4b , max_qr_steps>=1_i4b )
    else
        maxqrs = 4_i4b
    end if
!
    jstart(1_i4b:m) = 0_i4b
    jend(1_i4b:m)   = -1_i4b
!
!   DETERMINE THE UNIT ROUNDOFF AND UNDERFLOW THRESHOLD.
!
    eps    = lamch( 'u' )
    prec   = fudge_prec*lamch( 'p' )
    safmin = lamch( 's' )
    gamma  = sqrt( safmin )
!
!   MAKE A LOCAL COPY OF THE TRIDIAGONAL MATRIX.
!
    d2(1_i4b:n)       = d(1_i4b:n)
    e2(1_i4b:n-1_i4b) = e(1_i4b:n-1_i4b)
    e2(n)             = zero
!
!   COMPUTE THE NORM OF THE TRIDIAGONAL MATRIX.
!
    tridnorm = zero
    xi       = zero
!
    do i=1_i4b, n
!
        xim       = abs( e2(i) )
        g         = abs( d2(i) ) + xi + xim
        tridnorm  = tridnorm + g
!        tridnorm  = max( tridnorm, g )
        xi        = xim
!
    end do
!
    g = real( n, stnd)*prec*tridnorm
!
!   TEST IF THE UPPER PART OF THE TRIDIAGONAL MATRIX IS NOT DIAGONAL OR ZERO.
!
    do i1=1_i4b, n-1_i4b
!
!        if ( abs( e2(i1) )>gamma .or. abs( d2(i1) )>gamma ) exit
!
        if ( abs( e2(i1) )>gamma ) exit
        if ( abs( d2(i1) )>gamma .and. any( abs( d2(i1) - eigval(1_i4b:m) )<=g )  ) exit
!
    end do
!
!   TEST IF THE LOWER PART OF THE TRIDIAGONAL MATRIX IS NOT DIAGONAL OR ZERO.
!
    do i2=n, 2_i4b, -1_i4b
!
!        if ( abs( e2(i2-1_i4b) )>gamma .or. abs( d2(i2) )>gamma ) exit
!
        if ( abs( e2(i2-1_i4b) )>gamma ) exit
        if ( abs( d2(i2) )>gamma .and. any( abs( d2(i2) - eigval(1_i4b:m) )<=g ) ) exit
!
    end do
!
    if ( ( any( abs( eigval(1_i4b:m) )<=gamma+g ) .and. ( i1/=1_i4b .or. i2/=n ) )  &
          .or. ( m>i2-i1+1_i4b ) ) then
!
!       IF SOME EIGENVALUES ARE VERY SMALL OR IF WE MISS SOME EIGENVALUES
!       WHEN COMPUTING i1 and i2 ABOVE, WE CANNOT WORK ON THE REDUCED PORTION OF 
!       THE TRIDIAGONAL MATRIX.
!
        i1 = 1_i4b
        i2 = n
!
    end if
!
!   ALLOCATE WORK VARIABLES NEEDED TO STORE THE GIVENS ROTATIONS.
!
    k = maxqrs*m
!
    allocate( csmat(n-1_i4b,k), snmat(n-1_i4b,k), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
    csmat(1_i4b:n-1_i4b,1_i4b:k) = one
    snmat(1_i4b:n-1_i4b,1_i4b:k) = zero
!
    l        = i2
    k        = 1_i4b
    ndeflate = 0_i4b
!
    done        = false
    deflate(:m) = false
!
!   COMPUTE DEFLATION PARAMETERS.
!
    do i=1_i4b, maxiter
!
        do j=1_i4b, m
!
            if ( deflate(j) ) cycle
!
!           TRY FIRST FERNANDO METHOD COMBINED WITH GODUNOV DEFLATION.
!
            call dflgen2( d2(i1:l), e2(i1:l-1_i4b), eigval(j),                 &
                          csmat(i1:l-1_i4b,k), snmat(i1:l-1_i4b,k), deflate(j) )
!
            if ( .not.deflate(j) ) cycle
!
            ndeflate = ndeflate + 1_i4b
            ind(j)   = ndeflate
!
            jstart(ndeflate) = k
            jend(ndeflate)   = k
!
            k = k + 1_i4b
            l = l - 1_i4b
!
        end do
!
        done = all( deflate(:m) )
!
        if ( done ) exit
!
    end do
!
    if ( .not.done ) then
!
!       RECOMPUTE DEFLATION PARAMETERS.
!
        do j=1_i4b, m
!
            if ( deflate(j) ) cycle
!
            ndeflate         = ndeflate + 1_i4b
            ind(j)           = ndeflate
            jstart(ndeflate) = k
!
!           TRY AGAIN FERNANDO METHOD COMBINED WITH GODUNOV DEFLATION.
!
            call dflgen2( d2(i1:l), e2(i1:l-1_i4b), eigval(j),                 &
                          csmat(i1:l-1_i4b,k), snmat(i1:l-1_i4b,k), deflate(j) )
!
            if ( .not.deflate(j) ) then
!
!               USE QR STEPS IF DEFLATION FAILED WITH FERNANDO-GODUNOV DEFLATION.
!
                maxqrs2 = maxqrs*(ndeflate - 1_i4b ) - k + 1_i4b + maxqrs
!
                do i=1_i4b, maxqrs2
!
                    g = sqrt( eps*abs( d2(l)*d2(l-1_i4b) ) ) + safmin
!
!                   REPLACE SMALL SUBDIAGONALS.
!
                    where ( abs(e2(i1:l-2_i4b))<g )                   &
                          e2(i1:l-2_i4b) = sign( g, e2(i1:l-2_i4b) )
!
                    call qrstep( d2(i1:l), e2(i1:l-1_i4b), eigval(j),                 &
                                 csmat(i1:l-1_i4b,k), snmat(i1:l-1_i4b,k), deflate(j) )
!
                    if ( deflate(j) ) exit
!
                    k = k + 1_i4b
!
                end do
!
                if ( .not.deflate(j) ) then
!
                    k = k - 1_i4b
!
                end if
!
            end if
!
            jend(ndeflate) = k
!
            k = k + 1_i4b
            l = l - 1_i4b
!
        end do
!
    end if
!
    failure(1_i4b:m) = .not.deflate(1_i4b:m)
!
!   NOW COMPUTE THE EIGENVECTORS.
!
#ifdef _OPENMP
    i = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    test_par    = .not.ompparallel .and.      &
                  m>1_i4b          .and.      &
                  i>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par)                &
!$OMP            ,SCHEDULE(STATIC,1)          &
!$OMP            ,PRIVATE(j,l,i,k,p,q,ndeflate,tmp,c,s,eigvec2)
!
    do j=1_i4b, m
!
        ndeflate = ind(j)
!
        l = i2 - ndeflate + 1_i4b
!
        p = jstart(ndeflate)
        k = jend(ndeflate)
!
!BUG:   SOME VERSIONS OF PGF95 WITH OPENMP HAVE TROUBLES WITH THE FOLLOWING CALL.
!       THUS, IT IS SAFER TO INLINE THE FUNCTION.
!
!        eigvec2(i1:l) = prodgiv_eigvec( csmat(i1:l-1_i4b,k), snmat(i1:l-1_i4b,k) )
!
!       COMPUTE THE PRODUCT OF THE GIVENS ROTATIONS TIMES THE VECTOR El.
!
        tmp = one
!
        do i =l-1_i4b, i1, -1_i4b
!
!           APPLY THE GIVENS ROTATION TO THE VECTOR El.
!
            eigvec2(i+1_i4b) = csmat(i,k)*tmp
            tmp              = -snmat(i,k)*tmp
!
        end do
!
        eigvec2(i1) = tmp
!
        do i=k-1_i4b, p, -1_i4b
!
!BUG:       SOME VERSIONS OF PGF95 WITH OPENMP HAVE TROUBLES WITH THE FOLLOWING CALL.
!           THUS, IT IS SAFER TO INLINE THE SUBROUTINE.
!
!            call prodgiv( csmat(i1:l-1_i4b,i), snmat(i1:l-1_i4b,i), eigvec2(i1:l) )
!
!           COMPUTE THE PRODUCT OF THE GIVENS ROTATIONS TIMES THE VECTOR eigvec(i1:l,j).
!
            tmp = eigvec2(l)
!
            do q =l-1_i4b, i1, -1_i4b
!
                c = csmat(q,i)
                s = snmat(q,i)
!
!               APPLY THE GIVENS ROTATION TO THE VECTOR eigvec(i1:l,j).
!
                eigvec2(q+1_i4b) = s*eigvec2(q) + c*tmp
                tmp              = c*eigvec2(q) - s*tmp
!
            end do
!
            eigvec2(i1) = tmp
!
        end do
!
        eigvec2(l+1_i4b:n) = zero
!
        do i=ndeflate-1_i4b, 1_i4b, -1_i4b
!
            do k=jend(i), jstart(i), -1_i4b
!
!BUG:           SOME VERSIONS OF PGF95 WITH OPENMP HAVE TROUBLES WITH THE FOLLOWING CALL.
!               THUS, IT IS SAFER TO INLINE THE SUBROUTINE.
!
!                call prodgiv( csmat(i1:l,k), snmat(i1:l,k), eigvec2(i1:l+1_i4b) )
!
!               COMPUTE THE PRODUCT OF THE GIVENS ROTATIONS TIMES THE VECTOR eigvec(i1:l+1,j).
!
                tmp = eigvec2(l+1_i4b)
!
                do q =l, i1, -1_i4b
!
                    c = csmat(q,k)
                    s = snmat(q,k)
!
!                   APPLY THE GIVENS ROTATION TO THE VECTOR eigvec(i1:l+1,j).
!
                    eigvec2(q+1_i4b) = s*eigvec2(q) + c*tmp
                    tmp              = c*eigvec2(q) - s*tmp
!
                end do
!
                eigvec2(i1) = tmp
!
            end do
!
            l = l + 1_i4b
!
        end do
!
        eigvec2(1_i4b:i1-1_i4b) = zero
!
        eigvec(1_i4b:n,j) = eigvec2(1_i4b:n)
!
    end do
!
!$OMP END PARALLEL DO
!    
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( csmat, snmat )
!
!
! END OF SUBROUTINE symtrid_deflate_rm
! ____________________________________
!
    end subroutine symtrid_deflate_rm
!
! =========================================================================================
!
    subroutine trid_deflate_a( d, e, eigval, eigvec, failure, max_qr_steps, ortho, inviter )
!
! Purpose
! _______
!
!   TRID_DEFLATE computes the eigenvectors of a real symmetric tridiagonal
!   matrix T corresponding to specified eigenvalues, using deflation techniques.
!
!
! Arguments
! _________
!
!   D              (INPUT) real(stnd), dimension(:)
!                  On entry, the diagonal elements of the symmetric tridiagonal matrix T.
!
!   E              (INPUT) real(stnd), dimension(:)
!                  On entry, the n-1 subdiagonal elements of the symmetric tridiagonal
!                  matrix T, E(n) is arbitrary and is not used .
!
!                  The size of E must verify:  size( E ) = size( D ) = n.
!
!   EIGVAL         (INPUT) real(stnd), dimension(:)
!                  On entry, selected eigenvalues of the symmetric tridiagonal matrix.
!                  The eigenvalues must be given in decreasing order.
!
!                  The size of EIGVAL must verify:  size( EIGVAL ) <= size( D ) = n .
!
!   EIGVEC         (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed eigenvectors. The eigenvector associated
!                  with the eigenvalue EIGVAL(j) is stored in the j-th column of EIGVEC.
!
!                  The shape of EIGVEC must verify: 
!
!                  - size( EIGVEC, 1 ) = size( D ) = n ;
!                  - size( EIGVEC, 2 ) = size( EIGVAL ) .
!
!   FAILURE        (OUTPUT) logical(lgl)
!                  On exit:
!
!                  - FAILURE = false :  indicates successful exit.
!                  - FAILURE = true  :  indicates that the algorithm did not converge and
!                    that full accuracy was not attained in the deflation procedure
!                    of the tridiagonal matrix.
!
!   MAX_QR_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                  MAX_QR_STEPS controls the maximum number of QR sweeps for deflating
!                  the tridiagonal matrix for a given eigenvalue.
!                  The algorithm fails to converge if the total number of QR sweeps for
!                  all eigenvalues exceeds MAX_QR_STEPS * size(EIGVAL).
!
!                  The default is 4.
!
!   ORTHO          (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - ORTHO=true, the tridiagonal matrix is deflated sequentially
!                    for all the specified eigenvalues; this implies that the eigenvectors will
!                    be automatically orthogonal on exit.
!                  - ORTHO=false, the tridiagonal matrix is deflated in parallel for the
!                    different clusters of eigenvalues or isolated eigenvalues; this implies that
!                    orthogonality is preserved inside each cluster, but not automatically between
!                    clusters.
!
!                  The default is ORTHO=false.
!
!   INVITER        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if INVITER=true eigenvectors corresponding to isolated eigenvalues
!                  are computed by inverse iteration instead of deflation.
!
!                  The default is INVITER=true.
!
!
! Further Details
! _______________
!
!   TRID_DEFLATE uses an efficient and robust approach for the computation of (selected)
!   eigenvectors of a tridiagonal matrix corresponding to (selected) eigenvalues by combining
!   Fernando's method for the computation of eigenvectors with deflation procedures by Givens
!   rotations (see the references (1), (2) and (3) below). QR iterations are also used as
!   a back-up procedure if the deflation technique fails (see the reference (4)).
!
!   Optionally, eigenvectors corresponding to isolated eigenvalues may be also
!   computed by inverse iteration on the tridiagonal matrix T.
!   This is the default for eigenvectors associated with isolated eigenvalues
!   since in this case inverse iteration is safe and faster than the deflation algorithms.
!
!   The computation of the eigenvectors is parallelized if OPENMP is used.
!
!   It is essential that eigenvalues given on entry of TRID_DEFLATE are computed to
!   high relative accuracy. Subroutine SYMTRID_BISECT may be used for this purpose.
!
!   TRID_DEFLATE may fail if some the eigenvalues specified in parameter EIGVAL are nearly
!   identical or for clusters of small eigenvalues or for some zero-diagonal matrices.
!
!   The deflation algorithms used in TRID_DEFLATE are competitive with the inverse iteration
!   procedure implemented in TRID_INVITER.
!
!   For further details, see:
!
!   (1) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, pp. 1013-1034.
!
!   (2) Parlett, B.N., and Dhillon, I.S., 1997:
!           Fernando's solution to Wilkinsin's problem: An application of double factorization.
!           Linear Algebra and its Appl., 267, pp.247-279.
!
!   (3) Malyshev, A.N., 2000:
!           On deflation for symmetric tridiagonal matrices.
!           Report 182 of the Department of Informatics, University of Bergen, Norway.
!
!   (4) Mastronardi, M., Van Barel, M., Van Camp, E., and Vandebril, R., 2006:
!           On computing the eigenvectors of a class of structured matrices.
!           Journal of Computational and Applied Mathematics, 189, 580-591.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch
    use Char_Constants,  only : eig_error2, allocate_error
    use Utilities,       only : assert_eq, assert, merror, lascl, unit_matrix
#ifdef _BLAS
    use BLAS_interfaces, only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),  dimension(:)   :: d, e, eigval
    real(stnd),  intent(out), dimension(:,:) :: eigvec
!
    integer(i4b), intent(in), optional :: max_qr_steps
!
    logical(lgl), intent(out)          :: failure
    logical(lgl), intent(in), optional :: ortho, inviter
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                              :: eps, eps2, safmin, ssfmin,  safmax, ssfmax,  &
                                               prec, g, xj, xjm, normt, tridnorm, lim
    real(stnd),  dimension(size(d))         :: d2, e2, temp, temp2
    real(stnd),  dimension(size(eigval))    :: lambda, rayleigh
!
    integer(i4b)                            :: n, n1, n2, m, m0, j, j1, j2, k, l
    integer(i4b), dimension(:), allocatable :: jstart, jend
    integer                                 :: iok
!
    logical(lgl)                          :: do_ortho, do_inviter
    logical(lgl), dimension(size(eigval)) :: failure2
!
#ifdef _OPENMP
!
    logical :: test_par, ompparallel, ompnested
!
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='trid_deflate'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(d),i4b)         ,   &
                    int(size(e),i4b)         ,   &
                    int(size(eigvec,1),i4b)  ,   &
                    name_proc )
!
    m =  assert_eq( int(size(eigval),i4b)    ,   &
                    int(size(eigvec,2),i4b)  ,   &
                    name_proc )
!
    call assert( logical( m<=n, lgl ), name_proc )
!
    failure = false
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b .or. m<=0_i4b) then
!
        if ( n==1_i4b .and. m==1_i4b ) then
!
            eigvec(1_i4b,1_i4b) = one
!
        end if
!
        return
!
    end if
!
!   CHECK OPTIONAL PARAMETERS.
!
    do_ortho = false
!
    if ( present(ortho) ) then
!
        if ( m>1_i4b ) then
!
            do_ortho = ortho
!
        end if
!
    end if
!
    do_inviter = true
!
    if ( present(inviter) ) then
!
        do_inviter = inviter
!
    end if
!
!   TEST IF EIGENVALUES ARE IN DECREASING ORDER.
!
    if ( m/=1_i4b ) then
!
        temp(1_i4b:m-1_i4b) = eigval(2_i4b:m) - eigval(1_i4b:m-1_i4b)
!
        if ( any( temp(1_i4b:m-1_i4b)>zero ) ) then
!
            call merror( name_proc//eig_error2 )
!
        end if
!
    end if
!
!   DETERMINE MACHINE PARAMETERS.
!
    eps    = lamch( 'u' )
    eps2   = eps*eps
    safmin = lamch( 's' )
    safmax = one/safmin
    ssfmax = sqrt( safmax )/three
    ssfmin = sqrt( safmin )/eps2
!
    prec   = fudge_prec*lamch( 'p' )
!
!   MAKE LOCAL COPIES OF THE TRIDIAGONAL MATRIX AND EIGENVALUES.
!
    d2(1_i4b:n)       = d(1_i4b:n)
    e2(1_i4b:n-1_i4b) = e(1_i4b:n-1_i4b)
    e2(n)             = zero
!
    lambda(1_i4b:m)  = eigval(1_i4b:m)
!
!   COMPUTE THE NORMS OF THE TRIDIAGONAL MATRIX.
!
    normt    = zero
    tridnorm = zero
    xj       = zero
!
    do j=1_i4b, n
!
        xjm      = abs( e2(j) )
        g        = abs( d2(j) ) + xj + xjm
        normt    = normt + g
        tridnorm = max( tridnorm, g )
        xj       = xjm
!
    end do
!
!   SCALE TRIDIAGONAL MATRIX IF NECESSARY.
!
    if ( tridnorm==zero ) then
!
!       THE TRIDIAGONAL MATRIX IS THE NULL MATRIX, RETURN.
!
        call unit_matrix( eigvec )
!
        return
!
    else if ( tridnorm>ssfmax ) then
!
        call lascl( d2(1_i4b:n), tridnorm, ssfmax )
        call lascl( e2(1_i4b:n-1_i4b), tridnorm, ssfmax )
        call lascl( lambda(1_i4b:m), tridnorm, ssfmax )
!
        tridnorm = ssfmax
!
    else if ( tridnorm<ssfmin ) then
!
        call lascl( d2(1_i4b:n), tridnorm, ssfmin )
        call lascl( e2(1_i4b:n-1_i4b), tridnorm, ssfmin )
        call lascl( lambda(1_i4b:m), tridnorm, ssfmin )
!
        tridnorm = ssfmin
!
    end if
!    
#ifdef _OPENMP
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  m>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!   COMPUTE THE EIGENVECTORS.
!
    if ( m==1_i4b ) then
!
!       ONLY ONE EIGENVECTOR IS REQUIRED.
!
        if ( do_inviter ) then
!
!           USE INVERSE ITERATION.
!
            call symtrid_inviter( d2(1_i4b:n), e2(1_i4b:n), lambda(1_i4b),         &
                                  eigvec(:n,1_i4b), failure2(1_i4b), initvec=false )
!
        else
!
!           USE DEFLATION.
!
            call symtrid_deflate_rv( d2(1_i4b:n), e2(1_i4b:n-1_i4b), lambda(1_i4b), eigvec(:n,1_i4b),  &
                                     failure2(1_i4b), max_qr_steps=max_qr_steps )
!
        end if
!
    else
!
        if ( m==n .and. all( lambda(1_i4b:m) >= zero ) ) then
!
!           CHECK THE FIRST AND LAST PARTS OF THE TRIDIAGONAL MATRIX FOR ZEROS
!           AND REDUCE ITS SIZE IF POSSIBLE.
!
            m0 = n
!
            do j = n, 2_i4b, -1_i4b
!
                if ( abs(d2(j))>=safmin .or. abs(e2(j-1_i4b))>=safmin ) exit
!
                eigvec(:n,m0) = zero
                eigvec(j,m0)  = one
!
                m0 = m0 - 1_i4b
!
            end do
!
            if ( j==1_i4b .and. abs(d2(1_i4b))<safmin ) then
!
                eigvec(1_i4b,1_i4b)   = one
                eigvec(2_i4b:n,1_i4b) = zero                
!
!               THE TRIDIAGONAL MATRIX IS THE NULL MATRIX, RETURN.
!
                return
!
            end if
!
            n2 = j
!
            do j = 1_i4b, n-1_i4b
!
                if ( abs(d2(j))>=safmin .or. abs(e2(j))>=safmin ) exit
!
                eigvec(:n,m0) = zero
                eigvec(j,m0)  = one
!
                m0 = m0 - 1_i4b
!
            end do
!
            n1 = j
!
            eigvec(1_i4b:n1-1_i4b,1_i4b:m0) = zero
            eigvec(n2+1_i4b:n,1_i4b:m0)     = zero
!
            failure2(m0+1_i4b:m) = false
!
        else
!
            n1 = 1_i4b
            n2 = n
!
            m0 = m
!
        end if
!
        if ( do_ortho ) then
!
!           SEVERAL EIGENVECTORS WITH GUARANTEED ORTHOGONALITY ARE REQUIRED.
!           USE SEQUENTIAL DEFLATION ON ALL EIGENVALUES.
!
            call symtrid_deflate_rm( d2(n1:n2), e2(n1:n2-1_i4b), lambda(1_i4b:m0),    &
                                     eigvec(n1:n2,1_i4b:m0), failure2(1_i4b:m0),      &
                                     max_qr_steps=max_qr_steps )
!
        else
!
!           SEVERAL EIGENVECTORS WITH GUARANTEED ORTHOGONALITY INSIDE EACH CLUSTER OF
!           EIGENVALUES ARE REQUIRED.
!
!           FIRST DETERMINE CLUSTERS OF EIGENVALUES.
!
!           ALLOCATE WORK VARIABLES.
!
            allocate( jstart(m0), jend(m0), stat = iok  )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           DETERMINE THRESHOLD FOR SEPARATION OF EIGENVALUES.
!
!            lim = c1_m3*tridnorm
            lim = fudge_ortho*sqrt( eps )*tridnorm
!
!           DETERMINE THE NUMBER OF CLUSTERS OF EIGENVALUES.
!           k IS THE NUMBER OF CLUSTERS OF EIGENVALUES
!           jstart(k) IS THE INDICE OF THE FIRST EIGENVALUE IN CLUSTER k.
!           jend(k)   IS THE INDICE OF THE LAST EIGENVALUE IN CLUSTER k.
!
            k = 1_i4b
            jstart(1_i4b) = 1_i4b
!
            xj = lambda(1_i4b)
!
            do j=2_i4b, m0
!
                xjm = lambda(j)
                g   = xj - xjm
                xj  = xjm
!
                if ( g<lim ) cycle
!
                jend(k)   = j - 1_i4b
                k         = k + 1_i4b
                jstart(k) = j
!
            end do
!
            jend(k) = m0
!
!           MERGE SOME CLUSTERS FOR SPECIAL CASES.
!
            if ( k==2_i4b ) then
!
!               IF THERE IS ONLY TWO CLUSTERS AND ONE OF THEM IS AN ISOLATED EIGENVALUE
!               MERGE THE TWO CLUSTERS.
!
                if ( ( jstart(1_i4b)==jend(1_i4b) ) .or. ( jstart(2_i4b)==jend(2_i4b) ) ) then
!
                    k = 1_i4b
!
                    jstart(1_i4b) = 1_i4b
                    jend(1_i4b)   = m0
!
                end if
!
            else if ( k>2_i4b ) then
!
!               IF ONE CLUSTER IS AN ISOLATED AND SMALL
!               EIGENVALUE MERGE IT WITH THE PREVIOUS CLUSTER.
!
                j = 1_i4b
!
                do
!
                    j = j + 1_i4b
!
                    if ( j>k ) exit
!
                    j1 = jstart(j)
                    j2 = jend(j)
!
                    if ( j1/=j2 ) cycle
!
                    if (  abs(lambda(j1))<=lim ) then
!
                        jend(j-1_i4b) = j2                
!
                        jstart(j:k-1_i4b) = jstart(j+1_i4b:k)
                        jend(j:k-1_i4b)   = jend(j+1_i4b:k)
!
                        k = k - 1_i4b
!
                    end if
!
                end do
!
            end if
!
#ifdef _OPENMP
!
            if ( test_par .and. k>1_i4b ) then
!
!               OPENMP PARALLEL PATH.
!
!               FIRST, COMPUTE EIGENVECTORS CORRESPONDING TO ISOLATED
!               EIGENVALUES AND SMALL CLUSTERS OF ALMOST-MULTIPLE EIGENVALUES.
!               EACH THREAD TAKES CARE OF ONE ISOLATED EIGENVALUE OR SMALL CLUSTER
!               AND PARALLELIZATION IS OVER THE NUMBER OF SMALL CLUSTERS AND ISOLATED
!               EIGENVALUES.
!
!$OMP PARALLEL DO SCHEDULE(DYNAMIC,1)     &
!$OMP            ,PRIVATE(j,j1,j2)
!
                do j=1_i4b, k
!
                    j1 = jstart(j)
                    j2 = jend(j)
!
                    if ( j2-j1<neig_cluster ) then
!
                        if ( j1==j2 ) then
!
                            if ( do_inviter ) then
!
!                               USE INVERSE ITERATION.
!
                                call symtrid_inviter( d2(n1:n2), e2(n1:n2), lambda(j1),             &
                                                      eigvec(n1:n2,j1), failure2(j1), initvec=false )
!
                            else
!
!                               USE DEFLATION.
!
                                call symtrid_deflate_rv( d2(n1:n2), e2(n1:n2-1_i4b), lambda(j1), &
                                                         eigvec(n1:n2,j1), failure2(j1),         &
                                                         max_qr_steps=max_qr_steps )
!
                            end if
!
                        else
!
!                           USE SEQUENTIAL DEFLATION ON SMALL CLUSTERS.
!
                            call symtrid_deflate_rm( d2(n1:n2), e2(n1:n2-1_i4b), lambda(j1:j2),  &
                                                     eigvec(n1:n2,j1:j2), failure2(j1:j2),       &
                                                     max_qr_steps=max_qr_steps )
!
                        end if
!
                    end if
!
                end do
!
!$OMP END PARALLEL DO
!
!               SECOND, COMPUTE EIGENVECTORS CORRESPONDING TO BIG CLUSTERS 
!               OF ALMOST-MULTIPLE EIGENVALUES. PARALLELIZATION IS OVER
!               THE NUMBER OF EIGENVECTORS IN EACH CLUSTER.
!
                do j=1_i4b, k
!
                    j1 = jstart(j)
                    j2 = jend(j)
!
                    if ( j2-j1<neig_cluster ) cycle
!
!                   USE SEQUENTIAL DEFLATION ON ALL SINGULAR VALUES IN THE CLUSTER.
!
                    call symtrid_deflate_rm( d2(n1:n2), e2(n1:n2-1_i4b), lambda(j1:j2),  &
                                             eigvec(n1:n2,j1:j2), failure2(j1:j2),       &
                                             max_qr_steps=max_qr_steps )
!
                end do
!
            else
#endif
!
!               SERIAL PATH.
!
                do j=1_i4b, k
!
                    j1 = jstart(j)
                    j2 = jend(j)
!
                    if ( j1==j2 ) then
!
!                       COMPUTE EIGENVECTORS ASSOCIATED WITH AN ISOLATED EIGENVALUE.
!
                        if ( do_inviter ) then
!
!                           USE INVERSE ITERATION.
!
                            call symtrid_inviter( d2(n1:n2), e2(n1:n2), lambda(j1),             &
                                                  eigvec(n1:n2,j1), failure2(j1), initvec=false )
!
                        else
!
!                           USE DEFLATION.
!
                            call symtrid_deflate_rv( d2(n1:n2), e2(n1:n2-1_i4b), lambda(j1), &
                                                     eigvec(n1:n2,j1), failure2(j1),         &
                                                     max_qr_steps=max_qr_steps )
!
                        end if
!
                    else
!
!                       COMPUTE EIGENVECTORS ASSOCIATED WITH EIGENVALUES BELONGING
!                       TO A CLUSTER OF EIGENVALUES.
!                       USE SEQUENTIAL DEFLATION ON ALL EIGENVALUES IN THE CLUSTER.
!
                        call symtrid_deflate_rm( d2(n1:n2), e2(n1:n2-1_i4b), lambda(j1:j2),  &
                                                 eigvec(n1:n2,j1:j2), failure2(j1:j2),       &
                                                 max_qr_steps=max_qr_steps )
!
                    end if
!
                end do
!
#ifdef _OPENMP
            end if
#endif
!
        end if
!
    end if
!
!   CHECK ACCURACY OF THE COMPUTED EIGENVECTORS.
!
!$OMP PARALLEL  IF(test_par)                         &
!$OMP          ,PRIVATE(j,g,d2,e2,temp,temp2)
!
    d2(1_i4b:n)       = d(1_i4b:n)
    e2(1_i4b:n-1_i4b) = e(1_i4b:n-1_i4b)
    e2(n)             = zero
!
!$OMP DO SCHEDULE(STATIC,1)
!
    do j=1_i4b, m
!
        temp(1_i4b:n) = eigvec(1_i4b:n,j)
!
!       COMPUTE RAYLEIGH QUOTIENT.
!
        temp2(1_i4b)         = d2(1_i4b)*temp(1_i4b) + e2(1_i4b)*temp(2_i4b)
        temp2(2_i4b:n-1_i4b) = e2(1_i4b:n-2_i4b)*temp(1_i4b:n-2_i4b)           +   &
                               d2(2_i4b:n-1_i4b)*temp(2_i4b:n-1_i4b)           +   &
                               e2(2_i4b:n-1_i4b)*temp(3_i4b:n)
        temp2(n)             = e2(n-1_i4b)*temp(n-1_i4b) + d2(n)*temp(n)
!
        g  = dot_product( temp2(1_i4b:n), temp(1_i4b:n) )
        rayleigh(j) = g
!
!       COMPUTE THE RESIDUAL NORM OF THE JTH EIGENVECTOR
!       WITH RESPECT TO ITS RAYLEIGH QUOTIENT.
!
#ifdef _BLAS
        call axpy( n, -g, temp(1_i4b:n), 1_i4b, temp2(1_i4b:n), 1_i4b )
#else
        temp2(1_i4b:n) = temp2(1_i4b:n) - g*temp(1_i4b:n)
#endif
!
        lambda(j) = sum( abs( temp2(:n) ) )
!
    end do
!
!$OMP END DO
!
!$OMP END PARALLEL
!
    lim = normt*real(n,stnd)*prec
!
    failure = any( lambda(1_i4b:m)>lim ) .or. any( failure2(1_i4b:m) )
!
!   CHECK THE PAIRING OF EIGENVALUES AND EIGENVECTORS.
!
    if ( m/=1_i4b .and. .not.do_ortho ) then
!
        do l=1_i4b, k
!
            j1 = jstart(l)
            j2 = jend(l)
!
            do j=j1, j2-1_i4b
!
                if ( rayleigh(j)<rayleigh(j+1_i4b) ) exit
!
            end do
!
            if ( j<j2 ) then
!
                call eig_sort_a( 'd', rayleigh(j1:j2), eigvec(:n,j1:j2) )
!
            end if
!
            if ( l==1_i4b ) cycle
!
            if ( .not.failure ) then
!
                j2 = jend(l-1_i4b)
!
                failure = rayleigh(j2)<rayleigh(j1)
!
            end if
!
        end do
!
!       DEALLOCATE WORK ARRAYS
!
        deallocate( jstart, jend )
!
    else
!
        do j=1_i4b, m-1_i4b
!
            if ( rayleigh(j)<rayleigh(j+1_i4b) ) exit
!
        end do
!
!        failure = failure .or. j<m
!
        if ( j<m ) then
!
            call eig_sort_a( 'd', rayleigh(1_i4b:m), eigvec(1_i4b:n,1_i4b:m) )
!
        end if
!
    end if
!
!   CHECK FINALLY IF THE RAYLEIGH QUOTIENTS AND THE EIGENVALUES AGREE.
!
    if ( .not.failure ) then
!
        xj  = maxval( abs( rayleigh(1_i4b:m) - eigval(1_i4b:m) ) )
        failure = xj>lim
!
    end if
!
!
! END OF SUBROUTINE trid_deflate_a
! ________________________________
!
    end subroutine trid_deflate_a
!
! =========================================================================================
!
    subroutine trid_deflate_b( d, e, eigval, eigvec, failure, mat, max_qr_steps, ortho, inviter )
!
! Purpose
! _______
!
!   TRID_DEFLATE computes the eigenvectors of a full real n-by-n symmetric matrix MAT
!   corresponding to specified eigenvalues, using deflation techniques applied to a
!   symmetric tridiagonal matrix T followed by a back-transformation procedure.
!
!   It is required that the original symmetric matrix MAT has been reduced to
!   symmetric tridiagonal form T by an orthogonal similarity transformation:
!
!                        Q' * MAT * Q = T
!
!   with a call to SYMTRID_CMP with parameter STORE_Q set to true, before
!   calling TRID_DEFLATE.
!
!
! Arguments
! _________
!
!   D              (INPUT) real(stnd), dimension(:)
!                  On entry, the diagonal elements of the symmetric tridiagonal form T
!                  of MAT.
!
!   E              (INPUT) real(stnd), dimension(:)
!                  On entry, the n-1 subdiagonal elements of the symmetric tridiagonal
!                  form T of MAT. E(n) is arbitrary .
!
!                  The size of E must verify:  size( E ) = size( D ) = n .
!
!   EIGVAL         (INPUT) real(stnd), dimension(:)
!                  On entry, selected eigenvalues of the symmetric matrix MAT.
!                  The eigenvalues must be given in decreasing order.
!
!                  The size of EIGVAL must verify:  size( EIGVAL ) <= size( D ) = n .
!
!   EIGVEC         (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed eigenvectors. The eigenvector associated
!                  with the eigenvalue EIGVAL(j) is stored in the j-th column of EIGVEC.
!
!                  The shape of EIGVEC must verify:
!
!                  - size( EIGVEC, 1 ) = size( D ) = n ;
!                  - size( EIGVEC, 2 ) = size( EIGVAL ).
!
!   FAILURE        (OUTPUT) logical(lgl)
!                  On exit:
!
!                  - FAILURE = false :  indicates successful exit.
!                  - FAILURE = true  :  indicates that the algorithm did not converge and
!                    that full accuracy was not attained in the deflation procedure
!                    of the tridiagonal matrix.
!
!   MAT            (INPUT) real(stnd), dimension(:,:)
!                  On entry, the vectors and the scalars which define the elementary
!                  reflectors used to reduce the full real n-by-n symmetric matrix MAT to
!                  symmetric tridiagonal form T, as returned by SYMTRID_CMP with STORE_Q=true,
!                  in its argument MAT. MAT is not modified by the routine.
!
!                  Back-transformation is used to find the selected eigenvectors of the original
!                  matrix MAT and these eigenvectors are stored in argument EIGVEC.
!
!                  The shape of MAT must verify:
!
!                  - size( MAT, 1 ) = size( D ) = n ;
!                  - size( MAT, 2 ) = size( D ) = n .
!
!   MAX_QR_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                  MAX_QR_STEPS controls the maximum number of QR sweeps for deflating
!                  the tridiagonal matrix for a given eigenvalue.
!                  The algorithm fails to converge if the total number of QR sweeps for
!                  all eigenvalues exceeds MAX_QR_STEPS * size(EIGVAL).
!
!                  The default is 4.
!
!   ORTHO          (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - ORTHO=true, the tridiagonal matrix is deflated sequentially
!                    for all the specified eigenvalues; this implies that the eigenvectors will
!                    be automatically orthogonal on exit.
!                  - ORTHO=false, the tridiagonal matrix is deflated in parallel for the
!                    different clusters of eigenvalues or isolated eigenvalues; this implies that
!                    orthogonality is preserved inside each cluster, but not automatically between
!                    clusters.
!
!                  The default is ORTHO=false.
!
!   INVITER        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if INVITER=true eigenvectors corresponding to isolated eigenvalues
!                  are computed by inverse iteration instead of deflation.
!
!                  The default is INVITER=true.
!
!
! Further Details
! _______________
!
!   TRID_DEFLATE uses an efficient and robust approach for the computation of (selected)
!   eigenvectors of a tridiagonal matrix corresponding to (selected) eigenvalues by combining
!   Fernando's method for the computation of eigenvectors with deflation procedures by Givens
!   rotations (see the references (1), (2) and (3) below). QR iterations are also used as
!   a back-up procedure if the deflation technique fails (see the reference (4)).
!
!   Optionally, eigenvectors corresponding to isolated eigenvalues may be also
!   computed by inverse iteration on the tridiagonal matrix T.
!   This is the default for eigenvectors associated with isolated eigenvalues
!   since in this case inverse iteration is safe and faster than the deflation algorithms.
!
!   In a second step, the corresponding (selected) eigenvectors of the full real n-by-n symmetric
!   matrix MAT are computed by a blocked back-transformation algorithm using the Householder
!   transformations used to reduce the full real n-by-n symmetric matrix MAT to symmetric
!   tridiagonal form T (see the references (5) and (6)).
!
!   The computation of the eigenvectors is parallelized if OPENMP is used.
!
!   It is essential that eigenvalues given on entry of TRID_DEFLATE are computed to
!   high relative accuracy. Subroutine SYMTRID_BISECT may be used for this purpose.
!
!   TRID_DEFLATE may fail if some the eigenvalues specified in parameter EIGVAL are nearly
!   identical or for clusters of small eigenvalues or for some zero-diagonal matrices.
!
!   The deflation algorithms used in TRID_DEFLATE are competitive with the inverse iteration
!   procedure implemented in TRID_INVITER.
!
!   For further details on the deflation algorithm and the blocked backed-transformation
!   algorithm, see:
!
!   (1) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, 1013-1034.
!
!   (2) Parlett, B.N., and Dhillon, I.S., 1997:
!           Fernando's solution to Wilkinsin's problem: An application of double factorization.
!           Linear Algebra and its Appl., 267, pp.247-279.
!
!   (3) Malyshev, A.N., 2000:
!           On deflation for symmetric tridiagonal matrices.
!           Report 182 of the Department of Informatics, University of Bergen, Norway.
!
!   (4) Mastronardi, M., Van Barel, M., Van Camp, E., and Vandebril, R., 2006:
!           On computing the eigenvectors of a class of structured matrices.
!           Journal of Computational and Applied Mathematics, 189, 580-591.
!
!   (5) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989:
!           Block reduction of matrices to condensed form for eigenvalue computations. 
!           J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!   (6) Walker, H.F., 1988:
!           Implementation of the GMRES method using Householder transformations.
!           Siam J. Sci. Stat. Comput., Vol. 9, No 1, pp. 152-163.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,    only : machhuge
    use Char_Constants,   only : eig_error1
    use Utilities,        only : assert_eq, assert, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),  dimension(:)     :: d, e, eigval
    real(stnd),  intent(in),  dimension(:,:)   :: mat
    real(stnd),  intent(out), dimension(:,:)   :: eigvec
!
    integer(i4b), intent(in),  optional :: max_qr_steps
!
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in),  optional :: ortho, inviter
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: n, m
!
    logical(lgl) :: upper
!
#ifdef _OPENMP
!
    integer(i4b) :: i, num_threads, thread_num, jinc, jrem, joffset, j1, j2
!
    logical :: test_par, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='trid_deflate'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(d),i4b)      ,   &
                    int(size(e),i4b)  ,       &
                    int(size(mat,1),i4b)  ,   &
                    int(size(mat,2),i4b)  ,   &
                    name_proc )
!
    call assert( logical(size(eigvec,1)==n,lgl) ,  &
                 name_proc )
!
    m =  assert_eq( int(size(eigval),i4b)   ,   &
                    int(size(eigvec,2),i4b) ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. m<=0_i4b ) then
!
        return
!
    end if
!
    if ( max( mat(1_i4b,1_i4b), mat(n,n) )==machhuge ) then
!
        upper = mat(1_i4b,1_i4b)==machhuge
!
    else
!
        call merror( name_proc//eig_error1 )
!
    end if
!
!   COMPUTE EIGENVECTORS OF THE SYMMETRIC TRIDIAGONAL FORM T OF mat.
!
    call trid_deflate_a( d, e, eigval, eigvec, failure, max_qr_steps=max_qr_steps,  &
                         ortho=ortho, inviter=inviter                               )
!
!   COMPUTE EIGENVECTORS OF THE FULL SYMMETRIC MATRIX mat BY BACK-TRANSFORMATION.
!
#ifdef _OPENMP
    i = omp_get_max_threads()
    ompnested   = omp_get_nested()
    test_par    = ompnested             .and.      &
                  n*m>=omp_limit        .and.      &
                  m/2_i4b>=omp_limit2   .and.      &
                  i>1_i4b
!
    if ( test_par ) then
!
!$OMP PARALLEL NUM_THREADS( 2 )                                      &
!$OMP         ,PRIVATE(j1,j2,jrem,jinc,joffset,num_threads,thread_num)
!
!       DISTRIBUTE THE WORK AMONG THE THREADS.
!
        num_threads = omp_get_num_threads()
        thread_num  = omp_get_thread_num()
!
        jrem = mod( m, num_threads )
        jinc = ( m - jrem )/num_threads
!
        if ( thread_num<jrem ) then
            jinc    = jinc + 1_i4b
            joffset = 1_i4b
        else
            joffset = jrem + 1_i4b
        end if
!
        j1 = thread_num*jinc + joffset
        j2 = j1 + jinc - 1_i4b
!
!       GENERATE EIGENVECTORS OF mat IN eigvec .
!
        if ( upper ) then
!
            call apply_q_symtrid_b( mat, eigvec(:n,j1:j2), left=true, trans=false )
!
        else
!
            call apply_q_symtrid_a( mat, eigvec(:n,j1:j2), left=true, trans=false, upper=false )
!
        end if
!
!$OMP END PARALLEL
!
    else
#endif
!
!       GENERATE EIGENVECTORS OF mat IN eigvec .
!
        if ( upper ) then
!
            call apply_q_symtrid_b( mat, eigvec(:n,:m), left=true, trans=false )
!
        else
!
            call apply_q_symtrid_a( mat, eigvec(:n,:m), left=true, trans=false, upper=false )
!
        end if
!
#ifdef _OPENMP
    end if
#endif
!
!
! END OF SUBROUTINE trid_deflate_b
! ________________________________
!
    end subroutine trid_deflate_b
!
! =========================================================================================
!
    subroutine trid_deflate_c( d, e, eigval, eigvec, failure, matp, max_qr_steps, ortho, inviter )
!
! Purpose
! _______
!
!   TRID_DEFLATE computes the eigenvectors of a full real n-by-n symmetric matrix MAT,
!   packed column-wise in a linear array  MATP, corresponding to specified eigenvalues,
!   using deflation techniques applied to a symmetric tridiagonal matrix T followed by
!   a back-transformation procedure.
!
!   It is required that the original packed symmetric matrix mat has been reduced to symmetric
!   tridiagonal form T by an orthogonal similarity transformation:
!
!                        Q' * MAT * Q = T
!
!   with a call to SYMTRID_CMP with parameter STORE_Q set to true, before
!   calling TRID_DEFLATE.
!
!
! Arguments
! _________
!
!   D              (INPUT) real(stnd), dimension(:)
!                  On entry, the diagonal elements of the symmetric tridiagonal form T
!                  of MAT.
!
!   E              (INPUT) real(stnd), dimension(:)
!                  On entry, the n-1 subdiagonal elements of the symmetric tridiagonal
!                  form T of MAT. E(n) is arbitrary .
!
!                  The size of E must verify:  size( E ) = size( D ) = n .
!
!   EIGVAL         (INPUT) real(stnd), dimension(:)
!                  On entry, selected eigenvalues of the symmetric matrix MAT.
!                  The eigenvalues must be given in decreasing order.
!
!                  The size of EIGVAL must verify:  size( EIGVAL ) <= size( D ) = n .
!
!   EIGVEC         (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed eigenvectors. The eigenvector associated
!                  with the eigenvalue EIGVAL(j) is stored in the j-th column of EIGVEC.
!
!                  The shape of EIGVEC must verify:
!
!                  - size( EIGVEC, 1 ) = size( D ) = n  ;
!                  - size( EIGVEC, 2 ) = size( EIGVAL ) .
!
!   FAILURE        (OUTPUT) logical(lgl)
!                  On exit:
!
!                  - FAILURE = false :  indicates successful exit.
!                  - FAILURE = true  :  indicates that the algorithm did not converge and
!                    that full accuracy was not attained in the deflation procedure
!                    of the tridiagonal matrix.
!
!   MATP           (INPUT) real(stnd), dimension(:)
!                  On entry, the vectors and the scalars which define the elementary
!                  reflectors used to reduce the packed real n-by-n symmetric matrix MAT to
!                  symmetric tridiagonal form T, as returned by SYMTRID_CMP with STORE_Q=true,
!                  in its argument MATP. MATP is not modified by the routine.
!
!                  Back-transformation is used to find the selected eigenvectors of the original
!                  matrix MAT and these eigenvectors are stored in argument EIGVEC.
!
!                  The size of MATP must verify: size( MATP ) = (n * (n+1)/2)
!
!   MAX_QR_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                  MAX_QR_STEPS controls the maximum number of QR sweeps for deflating
!                  the tridiagonal matrix for a given eigenvalue.
!                  The algorithm fails to converge if the total number of QR sweeps for
!                  all eigenvalues exceeds MAX_QR_STEPS * size(EIGVAL).
!
!                  The default is 4.
!
!   ORTHO          (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - ORTHO=true, the tridiagonal matrix is deflated sequentially
!                    for all the specified eigenvalues; this implies that the eigenvectors will
!                    be automatically orthogonal on exit.
!                  - ORTHO=false, the tridiagonal matrix is deflated in parallel for the
!                    different clusters of eigenvalues or isolated eigenvalues; this implies that
!                    orthogonality is preserved inside each cluster, but not automatically between
!                    clusters.
!
!                  The default is ORTHO=false.
!
!   INVITER        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if INVITER=true eigenvectors corresponding to isolated eigenvalues
!                  are computed by inverse iteration instead of deflation.
!
!                  The default is INVITER=true.
!
!
! Further Details
! _______________
!
!   TRID_DEFLATE uses an efficient and robust approach for the computation of (selected)
!   eigenvectors of a tridiagonal matrix corresponding to (selected) eigenvalues by combining
!   Fernando's method for the computation of eigenvectors with deflation procedures by Givens
!   rotations (see the references (1), (2) and (3) below). QR iterations are also used as
!   a back-up procedure if the deflation technique fails (see the reference (4)).
!
!   Optionally, eigenvectors corresponding to isolated eigenvalues may be also
!   computed by inverse iteration on the tridiagonal matrix T.
!   This is the default for eigenvectors associated with isolated eigenvalues
!   since in this case inverse iteration is safe and faster than the deflation algorithms.
!
!   In a second step, the corresponding (selected) eigenvectors of the full real n-by-n symmetric
!   matrix MAT are computed by a blocked back-transformation algorithm with the Householder
!   transformations used to reduce the full real n-by-n symmetric matrix MAT to symmetric
!   tridiagonal form T (see the references (5) and (6)). These Householder transformations
!   must be packed in the linear array MATP (as returned by SYMTRID_CMP)
!   on entry of TRID_DEFLATE.
!
!   The computation of the eigenvectors is parallelized if OPENMP is used.
!
!   It is essential that eigenvalues given on entry of TRID_DEFLATE are computed to
!   high relative accuracy. Subroutine SYMTRID_BISECT may be used for this purpose.
!
!   TRID_DEFLATE may fail if some the eigenvalues specified in parameter EIGVAL are nearly
!   identical or for clusters of small eigenvalues or for some zero-diagonal matrices.
!
!   The deflation algorithms used in TRID_DEFLATE are competitive with the inverse iteration
!   procedure implemented in TRID_INVITER.
!
!   For further details, on the deflation algorithm and the blocked backed-transformation
!   algorithm, see:
!
!   (1) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, 1013-1034.
!
!   (2) Parlett, B.N., and Dhillon, I.S., 1997:
!           Fernando's solution to Wilkinsin's problem: An application of double factorization.
!           Linear Algebra and its Appl., 267, pp.247-279.
!
!   (3) Malyshev, A.N., 2000:
!           On deflation for symmetric tridiagonal matrices.
!           Report 182 of the Department of Informatics, University of Bergen, Norway.
!
!   (4) Mastronardi, M., Van Barel, M., Van Camp, E., and Vandebril, R., 2006:
!           On computing the eigenvectors of a class of structured matrices.
!           Journal of Computational and Applied Mathematics, 189, 580-591.
!
!   (5) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989:
!           Block reduction of matrices to condensed form for eigenvalue computations. 
!           J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!   (6) Walker, H.F., 1988:
!           Implementation of the GMRES method using Householder transformations.
!           Siam J. Sci. Stat. Comput., Vol. 9, No 1, pp. 152-163.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,    only : machhuge
    use Char_Constants,   only : eig_error1
    use Utilities,        only : assert, assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),  dimension(:)     :: d, e, eigval, matp
    real(stnd),  intent(out), dimension(:,:)   :: eigvec
!
    integer(i4b), intent(in),  optional :: max_qr_steps
!
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in),  optional :: ortho, inviter
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: n, m, i
!
    logical(lgl) :: upper
!
#ifdef _OPENMP
!
    integer(i4b) :: num_threads, thread_num, jinc, jrem, joffset, j1, j2
!
    logical :: test_par, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='trid_deflate'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(d),i4b)        , &
                    int(size(e),i4b)        , &
                    int(size(eigvec,1),i4b) , &
                    name_proc )
!
    m =  assert_eq( int(size(eigval),i4b)   ,   &
                    int(size(eigvec,2),i4b) ,   &
                    name_proc )
!
    i = n*(n+1_i4b)/2_i4b
!
    call assert( logical(size(matp)==i,lgl)  ,   &
                 name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. m<=0_i4b ) then
!
        return
!
    end if
!
    if ( max( matp(1_i4b), matp(i) )==machhuge ) then
!
        upper = matp(1_i4b)==machhuge
!
    else
!
        call merror( name_proc//eig_error1 )
!
    end if
!
!   COMPUTE EIGENVECTORS OF THE SYMMETRIC TRIDIAGONAL FORM T OF mat.
!
    call trid_deflate_a( d, e, eigval, eigvec, failure, max_qr_steps=max_qr_steps,  &
                         ortho=ortho, inviter=inviter                               )
!
!   COMPUTE EIGENVECTORS OF THE FULL SYMMETRIC MATRIX mat BY BACK-TRANSFORMATION.
!
#ifdef _OPENMP
    i = omp_get_max_threads()
    ompnested   = omp_get_nested()
    test_par    = ompnested             .and.      &
                  n*m>=omp_limit        .and.      &
                  m/2_i4b>=omp_limit2   .and.      &
                  i>1_i4b
!
    if ( test_par ) then
!
!$OMP PARALLEL NUM_THREADS( 2 )                                      &
!$OMP         ,PRIVATE(j1,j2,jrem,jinc,joffset,num_threads,thread_num)
!
!       DISTRIBUTE THE WORK AMONG THE THREADS.
!
        num_threads = omp_get_num_threads()
        thread_num  = omp_get_thread_num()
!
        jrem = mod( m, num_threads )
        jinc = ( m - jrem )/num_threads
!
        if ( thread_num<jrem ) then
            jinc    = jinc + 1_i4b
            joffset = 1_i4b
        else
            joffset = jrem + 1_i4b
        end if
!
        j1 = thread_num*jinc + joffset
        j2 = j1 + jinc - 1_i4b
!
!       GENERATE EIGENVECTORS OF mat IN eigvec .
!
        if ( upper ) then
!
            call apply_q_symtrid_d( matp, eigvec(:n,j1:j2), left=true, trans=false )
!
        else
!
            call apply_q_symtrid_c( matp, eigvec(:n,j1:j2), left=true, trans=false, upper=false )
!
        end if
!
!$OMP END PARALLEL
!
    else
#endif
!
!       GENERATE EIGENVECTORS OF mat IN eigvec .
!
        if ( upper ) then
!
            call apply_q_symtrid_d( matp, eigvec(:n,:m), left=true, trans=false )
!
        else
!
            call apply_q_symtrid_c( matp, eigvec(:n,:m), left=true, trans=false, upper=false )
!
        end if
!
#ifdef _OPENMP
    end if
#endif
!
!
! END OF SUBROUTINE trid_deflate_c
! ________________________________
!
    end subroutine trid_deflate_c
!
!
! =========================================================================================
!                         FULL SYMMETRIC EIGENVALUES DECOMPOSITIONS
! =========================================================================================
!
!
    subroutine eig_cmp_a( mat, eigval, failure, upper, sort, maxiter )
!
! Purpose
! _______
!
!   EIG_CMP computes all eigenvalues and eigenvectors of a
!   n-by-n real symmetric matrix MAT.
!
!
! Arguments
! _________
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the symmetric matrix MAT.
!
!                If:
!
!                - UPPER = true:
!                  the leading n-by-n upper triangular part of MAT contains the
!                  upper triangular part of the matrix MAT.
!                - UPPER = false:
!                  the leading n-by-n lower triangular part of MAT contains
!                  the lower triangular part of the matrix MAT.
!
!                On exit, MAT contains the orthonormal eigenvectors of
!                the matrix MAT.
!
!                The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   EIGVAL       (OUTPUT) real(stnd), dimension(:)
!                On exit, the eigenvalues.
!
!                The size of EIGVAL must verify:  size( EIGVAL ) = n .
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit.
!                - FAILURE = true  :  indicates that the algorithm did not converge and
!                  that full accuracy was not attained in the Schur decomposition
!                  of an intermediate tridiagonal form T of MAT .
!
!   UPPER        (INPUT) logical(lgl)
!                Specifies whether the upper or lower triangular part of the
!                symmetric matrix MAT is stored. If:
!
!                - UPPER= true :  Upper triangular is stored ;
!                - UPPER= false:  Lower triangular is stored .
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!                The eigenvectors are reordered accordingly.
!
!   MAXITER      (INPUT, OPTIONAL) integer(i4b)
!                MAXITER controls the maximum number of QR sweeps in the Schur 
!                decomposition of an intermediate tridiagonal form T of MAT.
!
!                The algorithm fails to converge if the number of QR sweeps
!                exceeds MAXITER * size(EIGVAL). Convergence usually occurs in 
!                about 2 * size(EIGVAL) QR sweeps.
!
!                The default is 30.
!
!
! Further Details
! _______________
!
!   The matrix MAT is first transformed to tridiagonal form T, then the eigenvalues
!   and the eigenvectors are computed by the QR implicit algorithm.
!
!   For further details, see:
!
!   (1) Parlett, B.N., 1998:
!           The Symmetric Eigenvalue Problem.
!           Revised edition, SIAM, Philadelphia.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,    only : lamch
    use Utilities,        only : assert_eq, lascl
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:,:) :: mat
    real(stnd), intent(out),   dimension(:)   :: eigval
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(in)            :: upper
    logical(lgl), intent(out)           :: failure
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)  :: safmin, eps, smlnum, bignum, rmin, rmax, e(size(eigval)),      &
                   matnorm, sigma
!
    integer(i4b) :: n, j
!
    logical(lgl) :: mat_is_scaled
!
    character :: chr
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='eig_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(eigval),i4b),   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!
    if ( n==1_i4b ) then
!
        eigval(1_i4b)    = mat(1_i4b,1_i4b)
        mat(1_i4b,1_i4b) = one
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = zero
!
    if ( upper ) then
!
        chr = "u"
!
!$OMP PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)   &
!$OMP            ,SCHEDULE(STATIC,1)                   &
!$OMP            ,PRIVATE(j,sigma)                     &
!$OMP            ,REDUCTION(max:matnorm)
!
        do j = 1_i4b, n
!
            sigma = maxval(abs(mat(:j,j)))
            matnorm = max( matnorm, sigma  )
!
        end do
!
!$OMP END PARALLEL DO
!
    else
!
        chr = "l"
!
!$OMP PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)   &
!$OMP            ,SCHEDULE(STATIC,1)                   &
!$OMP            ,PRIVATE(j,sigma)                     &
!$OMP            ,REDUCTION(max:matnorm)
!
        do j = 1_i4b, n
!
            sigma = maxval(abs(mat(j:,j)))
            matnorm = max( matnorm, sigma )
!
        end do
!
!$OMP END PARALLEL DO
!
    end if
!
!   SCALE MATRIX TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl( mat, one, sigma, chr )
!
!   CALL symtrid_cmp_a TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_a( mat, eigval, e, true, upper )
!
!   FIRST CALL ortho_gen_symtrid_a TO GENERATE THE ORTHOGONAL MATRIX,
!   THEN CALL symtrid_qri_a TO COMPUTE EIGENVECTORS AND EIGENVALUES.
!
    call ortho_gen_symtrid_a( mat, upper )
!
    call symtrid_qri_a( eigval, e, failure, mat, sort=sort, maxiter=maxiter )
!
!   IF MATRIX mat WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) eigval(:n) = ( one/sigma )*eigval(:n)
!
!
! END OF SUBROUTINE eig_cmp_a
! ___________________________
!
    end subroutine eig_cmp_a
!
! =========================================================================================
!
    subroutine eig_cmp_b( mat, eigval, failure, sort, maxiter )
!
! Purpose
! _______
!
!   EIG_CMP computes all eigenvalues and eigenvectors of a
!   n-by-n real symmetric matrix MAT.
!
!
! Arguments
! _________
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the symmetric matrix MAT. The leading 
!                n-by-n upper triangular part of MAT contains the
!                upper triangular part of the matrix MAT. The strictly
!                n-by-n lower triangular part of MAT is not referenced.
!
!                On exit, MAT contains the orthonormal eigenvectors of
!                the matrix MAT.
!
!                The shape of MAT must verify: size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   EIGVAL       (OUTPUT) real(stnd), dimension(:)
!                On exit, the eigenvalues.
!
!                The size of EIGVAL must verify: size( EIGVAL ) = n .
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit.
!                - FAILURE = true  :  indicates that the algorithm did not converge and
!                  that full accuracy was not attained in the Schur decomposition
!                  of an intermediate tridiagonal form T of MAT .
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!                The eigenvectors are reordered accordingly.
!
!   MAXITER      (INPUT, OPTIONAL) integer(i4b)
!                MAXITER controls the maximum number of QR sweeps in the Schur 
!                decomposition of an intermediate tridiagonal form T of MAT.
!
!                The algorithm fails to converge if the number of QR sweeps
!                exceeds MAXITER * size(EIGVAL). Convergence usually occurs in 
!                about 2 * size(EIGVAL) QR sweeps.
!
!                The default is 30.
!
!
! Further Details
! _______________
!
!   The matrix MAT is first transformed to tridiagonal form T, then the eigenvalues
!   and the eigenvectors are computed by the QR implicit algorithm.
!
!   For further details, see
!
!   (1) Parlett, B.N., 1998:
!           The Symmetric Eigenvalue Problem.
!           Revised edition, SIAM, Philadelphia.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,    only : lamch
    use Utilities,        only : assert_eq, lascl
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:,:) :: mat
    real(stnd), intent(out),   dimension(:)   :: eigval
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(out)           :: failure
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)  :: safmin, eps, smlnum, bignum, rmin, rmax, e(size(eigval)),      &
                   matnorm, sigma
!
    integer(i4b) :: n, j
!
    logical(lgl) :: mat_is_scaled
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='eig_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(eigval),i4b),   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!
    if ( n==1_i4b ) then
!
        eigval(1_i4b)    = mat(1_i4b,1_i4b)
        mat(1_i4b,1_i4b) = one
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = zero
!
!$OMP PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)   &
!$OMP            ,SCHEDULE(STATIC,1)                   &
!$OMP            ,PRIVATE(j,sigma)                     &
!$OMP            ,REDUCTION(max:matnorm)
!
    do j = 1_i4b, n
!
        sigma =  maxval( abs(mat(:j,j)) )
        matnorm = max( matnorm, sigma )
!
    end do
!
!$OMP END PARALLEL DO
!
!
!   SCALE MATRIX TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl( mat, one, sigma )
!
!   CALL symtrid_cmp_b TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_b( mat, eigval, e, true )
!
!   FIRST CALL ortho_gen_symtrid_b TO GENERATE THE ORTHOGONAL MATRIX,
!   THEN CALL symtrid_qri_a TO COMPUTE EIGENVECTORS AND EIGENVALUES.
!
    call ortho_gen_symtrid_b( mat )
!
    call symtrid_qri_a( eigval, e, failure, mat, sort=sort, maxiter=maxiter )
!
!   IF MATRIX mat WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) eigval(:n) = ( one/sigma )*eigval(:n)
!
!
! END OF SUBROUTINE eig_cmp_b
! ___________________________
!
    end subroutine eig_cmp_b
!
! =========================================================================================
!
    subroutine eig_cmp2_a( mat, eigval, failure, upper, sort, maxiter, max_francis_steps )
!
! Purpose
! _______
!
!   EIG_CMP2 computes all eigenvalues and eigenvectors of a
!   n-by-n real symmetric matrix MAT.
!
!
! Arguments
! _________
!
!   MAT                 (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                       On entry, the symmetric matrix MAT.
!
!                       If:
!
!                       - UPPER = true:
!                         the leading n-by-n upper triangular part of MAT contains the
!                         upper triangular part of the matrix MAT.
!                       - UPPER = false:
!                         the leading n-by-n lower triangular part of MAT contains
!                         the lower triangular part of the matrix MAT.
!
!                       On exit, MAT contains the orthonormal eigenvectors of
!                       the matrix MAT.
!
!                       The shape of MAT must verify: size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   EIGVAL              (OUTPUT) real(stnd), dimension(:)
!                       On exit, the eigenvalues.
!
!                       The size of EIGVAL must verify: size( EIGVAL ) = n .
!
!   FAILURE             (OUTPUT) logical(lgl)
!                       On exit:
!
!                       - FAILURE = false :  indicates successful exit.
!                       - FAILURE = true  :  indicates that the algorithm did not converge and
!                         that full accuracy was not attained in the Schur decomposition
!                         of an intermediate tridiagonal form T of MAT .
!
!   UPPER               (INPUT) logical(lgl)
!                       Specifies whether the upper or lower triangular part of the
!                       symmetric matrix MAT is stored. If:
!
!                       - UPPER = true :  Upper triangular is stored ;
!                       - UPPER = false:  Lower triangular is stored .
!
!   SORT                (INPUT, OPTIONAL) character
!                       Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                       or in descending order if SORT = 'D' or 'd'.
!                       The eigenvectors are reordered accordingly.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the Schur 
!                       decomposition of an intermediate tridiagonal form T of MAT .
!
!                       The algorithm fails to converge if the number of QR sweeps
!                       exceeds MAXITER * size(EIGVAL). Convergence usually occurs in 
!                       about 2 * size(EIGVAL) QR sweeps.
!
!                       The default is 30.
!
!   MAX_FRANCIS_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                       MAX_FRANCIS_STEPS controls the maximum number of Francis sets
!                       (e.g. QR sweeps) of Givens rotations which must be saved
!                       before applying them with a wavefront algorithm to accumulate
!                       the eigenvectors in the QR algorithm. MAX_FRANCIS_STEPS is a
!                       strictly positive integer, otherwise the default value is used.
!
!                       The default is 10.
!
!
! Further Details
! _______________
!
!   This driver subroutine is adapted from the routine  DSYEV in LAPACK.
!
!   The matrix MAT is first transformed to tridiagonal form T, then the eigenvalues
!   are computed by the Pal-Walker-Kahan variant of the QR algorithm and
!   the eigenvectors are computed with a perfect shift strategy.
!
!   For further details, see:
!
!   (1) Parlett, B.N., 1998:
!           The Symmetric Eigenvalue Problem.
!           Revised edition, SIAM, Philadelphia.
!
!   (2) Greenbaum, A., and J. Dongarra, J., 1989:
!           Experiments with QR/QL Methods for the Symmetric Tridiagonal Eigenproblem.
!           LAPACK Working Note No 17, November 1989.
!
!   (3) Van Zee, F.G., van de Geijn, R., and Quintana-Orti, G., 2011:
!           Restructuring the QR Algorithm for High-Performance Application of Givens Rotations.
!           FLAME Working Note 60. The University of Texas at Austin, Department of Computer
!           Sciences. Technical Report TR-11-36.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,    only : lamch
    use Utilities,        only : assert_eq, lascl
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:,:) :: mat
    real(stnd), intent(out),   dimension(:)   :: eigval
!
    integer(i4b), intent(in), optional :: maxiter, max_francis_steps
!
    logical(lgl), intent(in)            :: upper
    logical(lgl), intent(out)           :: failure
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)  :: safmin, eps, smlnum, bignum, rmin, rmax, e(size(eigval)),      &
                   matnorm, sigma
!
    integer(i4b) :: n, j
!
    logical(lgl) :: mat_is_scaled
!
    character :: chr
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='eig_cmp2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(eigval),i4b),   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!
    if ( n==1_i4b ) then
!
        eigval(1_i4b)    = mat(1_i4b,1_i4b)
        mat(1_i4b,1_i4b) = one
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = zero
!
    if ( upper ) then
!
        chr = "u"
!
!$OMP PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)   &
!$OMP            ,SCHEDULE(STATIC,1)                   &
!$OMP            ,PRIVATE(j,sigma)                     &
!$OMP            ,REDUCTION(max:matnorm)
!
        do j = 1_i4b, n
!
            sigma = maxval( abs(mat(:j,j)) )
            matnorm = max( matnorm, sigma )
!
        end do
!
!$OMP END PARALLEL DO
!
    else
!
        chr = "l"
!
!$OMP PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)    &
!$OMP            ,SCHEDULE(STATIC,1)                    &
!$OMP            ,PRIVATE(j,sigma)                      &
!$OMP            ,REDUCTION(max:matnorm)
!
        do j = 1_i4b, n
!
            sigma = maxval( abs(mat(j:,j)) )
            matnorm = max( matnorm, sigma )
!
        end do
!
!$OMP END PARALLEL DO
!
    end if
!
!   SCALE MATRIX TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl( mat, one, sigma, chr )
!
!   CALL symtrid_cmp_a TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_a( mat, eigval, e, true, upper )
!
!   FIRST CALL ortho_gen_symtrid_a TO GENERATE THE ORTHOGONAL MATRIX,
!   THEN CALL symtrid_qri2_a TO COMPUTE EIGENVECTORS AND EIGENVALUES.
!
    call ortho_gen_symtrid_a( mat, upper )
!
    call symtrid_qri2_a( eigval, e, failure, mat, sort=sort,                  &
                         maxiter=maxiter, max_francis_steps=max_francis_steps )
!
!   IF MATRIX mat WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) eigval(:n) = ( one/sigma )*eigval(:n)
!
!
! END OF SUBROUTINE eig_cmp2_a
! ____________________________
!
    end subroutine eig_cmp2_a
!
! =========================================================================================
!
    subroutine eig_cmp2_b( mat, eigval, failure, sort, maxiter, max_francis_steps )
!
! Purpose
! _______
!
!   EIG_CMP2 computes all eigenvalues and eigenvectors of a
!   n-by-n real symmetric matrix MAT.
!
!
! Arguments
! _________
!
!   MAT                 (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                       On entry, the symmetric matrix MAT. The leading 
!                       n-by-n upper triangular part of MAT contains the
!                       upper triangular part of the matrix MAT. The strictly
!                       n-by-n lower triangular part of MAT is not referenced.
!
!                       On exit, MAT contains the orthonormal eigenvectors of
!                       the matrix MAT.
!
!                       The shape of MAT must verify: size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   EIGVAL              (OUTPUT) real(stnd), dimension(:)
!                       On exit, the eigenvalues.
!
!                       The size of EIGVAL must verify: size( EIGVAL ) = n .
!
!   FAILURE             (OUTPUT) logical(lgl)
!                       On exit:
!
!                       - FAILURE = false :  indicates successful exit.
!                       - FAILURE = true  :  indicates that the algorithm did not converge and
!                         that full accuracy was not attained in the Schur decomposition
!                         of an intermediate tridiagonal form T of MAT .
!
!   SORT                (INPUT, OPTIONAL) character
!                       Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                       or in descending order if SORT = 'D' or 'd'.
!                       The eigenvectors are reordered accordingly.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the Schur 
!                       decomposition of an intermediate tridiagonal form T of MAT .
!
!                       The algorithm fails to converge if the number of QR sweeps
!                       exceeds MAXITER * size(EIGVAL). Convergence usually occurs in 
!                       about 2 * size(EIGVAL) QR sweeps.
!
!                       The default is 30.
!
!   MAX_FRANCIS_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                       MAX_FRANCIS_STEPS controls the maximum number of Francis sets
!                       (e.g. QR sweeps) of Givens rotations which must be saved
!                       before applying them with a wavefront algorithm to accumulate
!                       the eigenvectors in the QR algorithm. MAX_FRANCIS_STEPS is a
!                       strictly positive integer, otherwise the default value is used.
!
!                       The default is 10.
!
!
! Further Details
! _______________
!
!   This driver subroutine is adapted from the routine  DSYEV in LAPACK.
!
!   The matrix MAT is first transformed to tridiagonal form T, then the eigenvalues
!   are computed by the Pal-Walker-Kahan variant of the QR algorithm and
!   the eigenvectors are computed with a perfect shift strategy.
!
!   For further details, see:
!
!   (1) Parlett, B.N., 1998:
!           The Symmetric Eigenvalue Problem.
!           Revised edition, SIAM, Philadelphia.
!
!   (2) Greenbaum, A., and J. Dongarra, J., 1989:
!           Experiments with QR/QL Methods for the Symmetric Tridiagonal Eigenproblem.
!           LAPACK Working Note No 17, November 1989.
!
!   (3) Van Zee, F.G., van de Geijn, R., and Quintana-Orti, G., 2011:
!           Restructuring the QR Algorithm for High-Performance Application of Givens Rotations.
!           FLAME Working Note 60. The University of Texas at Austin, Department of Computer
!           Sciences. Technical Report TR-11-36.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,    only : lamch
    use Utilities,        only : assert_eq, lascl
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:,:) :: mat
    real(stnd), intent(out),   dimension(:)   :: eigval
!
    integer(i4b), intent(in), optional :: maxiter, max_francis_steps
!
    logical(lgl), intent(out)           :: failure
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)  :: safmin, eps, smlnum, bignum, rmin, rmax, e(size(eigval)),      &
                   matnorm, sigma
!
    integer(i4b) :: n, j
!
    logical(lgl) :: mat_is_scaled
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='eig_cmp2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(eigval),i4b),   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!
    if ( n==1_i4b ) then
!
        eigval(1_i4b)    = mat(1_i4b,1_i4b)
        mat(1_i4b,1_i4b) = one
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = zero
!
!$OMP PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)    &
!$OMP            ,SCHEDULE(STATIC,1)                    &
!$OMP            ,PRIVATE(j,sigma)                      &
!$OMP            ,REDUCTION(max:matnorm)
!
    do j = 1_i4b, n
!
        sigma = maxval( abs(mat(:j,j)) )
        matnorm = max( matnorm, sigma )
!
    end do
!
!$OMP END PARALLEL DO
!
!
!   SCALE MATRIX TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl( mat, one, sigma )
!
!   CALL symtrid_cmp_b TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_b( mat, eigval, e, true )
!
!   FIRST CALL ortho_gen_symtrid_b TO GENERATE THE ORTHOGONAL MATRIX,
!   THEN CALL symtrid_qri2_a TO COMPUTE EIGENVECTORS AND EIGENVALUES.
!
    call ortho_gen_symtrid_b( mat )
!
    call symtrid_qri2_a( eigval, e, failure, mat, sort=sort,                  &
                         maxiter=maxiter, max_francis_steps=max_francis_steps )
!
!   IF MATRIX mat WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) eigval(:n) = ( one/sigma )*eigval(:n)
!
!
! END OF SUBROUTINE eig_cmp2_b
! ____________________________
!
    end subroutine eig_cmp2_b
!
! =========================================================================================
!
    subroutine eig_cmp3_a( mat, eigval, failure, upper, sort, maxiter, max_francis_steps )
!
! Purpose
! _______
!
!   EIG_CMP3 computes all eigenvalues and eigenvectors of a
!   n-by-n real symmetric matrix MAT.
!
!
! Arguments
! _________
!
!   MAT                 (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                       On entry, the symmetric matrix MAT.
!
!                       If:
!
!                       - UPPER = true:
!                         the leading n-by-n upper triangular part of MAT contains the
!                         upper triangular part of the matrix MAT.
!                       - UPPER = false:
!                         the leading n-by-n lower triangular part of MAT contains
!                         the lower triangular part of the matrix MAT.
!
!                       On exit, MAT contains the orthonormal eigenvectors of
!                       the matrix MAT.
!
!                       The shape of MAT must verify: size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   EIGVAL              (OUTPUT) real(stnd), dimension(:)
!                       On exit, the eigenvalues.
!
!                       The size of EIGVAL must verify: size( EIGVAL ) = n .
!
!   FAILURE             (OUTPUT) logical(lgl)
!                       On exit:
!
!                       - FAILURE = false :  indicates successful exit.
!                       - FAILURE = true  :  indicates that the algorithm did not converge and
!                         that full accuracy was not attained in the Schur decomposition
!                         of an intermediate tridiagonal form T of MAT .
!
!   UPPER               (INPUT) logical(lgl)
!                       Specifies whether the upper or lower triangular part of the
!                       symmetric matrix MAT is stored. If:
!
!                       - UPPER = true :  Upper triangular is stored ;
!                       - UPPER = false:  Lower triangular is stored .
!
!   SORT                (INPUT, OPTIONAL) character
!                       Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                       or in descending order if SORT = 'D' or 'd'.
!                       The eigenvectors are reordered accordingly.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the Schur 
!                       decomposition of an intermediate tridiagonal form T of MAT .
!
!                       The algorithm fails to converge if the number of QR sweeps
!                       exceeds MAXITER * size(EIGVAL). Convergence usually occurs in 
!                       about 2 * size(EIGVAL) QR sweeps.
!
!                       The default is 30.
!
!   MAX_FRANCIS_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                       MAX_FRANCIS_STEPS controls the maximum number of Francis sets
!                       (e.g. QR sweeps) of Givens rotations which must be saved
!                       before applying them with a wavefront algorithm to accumulate
!                       the eigenvectors in the QR algorithm. MAX_FRANCIS_STEPS is a
!                       strictly positive integer, otherwise the default value is used.
!
!                       The default is 10.
!
!
! Further Details
! _______________
!
!   The matrix MAT is first transformed to tridiagonal form T, then the eigenvalues
!   and the eigenvectors are computed by the QR implicit algorithm.
!
!   For further details, see:
!
!   (1) Parlett, B.N., 1998:
!           The Symmetric Eigenvalue Problem.
!           Revised edition, SIAM, Philadelphia.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,    only : lamch
    use Utilities,        only : assert_eq, lascl
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:,:) :: mat
    real(stnd), intent(out),   dimension(:)   :: eigval
!
    integer(i4b), intent(in), optional :: maxiter, max_francis_steps
!
    logical(lgl), intent(in)            :: upper
    logical(lgl), intent(out)           :: failure
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)  :: safmin, eps, smlnum, bignum, rmin, rmax, e(size(eigval)),      &
                   matnorm, sigma
!
    integer(i4b) :: n, j
!
    logical(lgl) :: mat_is_scaled
!
    character :: chr
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='eig_cmp3'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(eigval),i4b),   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!
    if ( n==1_i4b ) then
!
        eigval(1_i4b)    = mat(1_i4b,1_i4b)
        mat(1_i4b,1_i4b) = one
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = zero
!
    if ( upper ) then
!
        chr = "u"
!
!$OMP PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)   &
!$OMP            ,SCHEDULE(STATIC,1)                   &
!$OMP            ,PRIVATE(j,sigma)                     &
!$OMP            ,REDUCTION(max:matnorm)
!
        do j = 1_i4b, n
!
            sigma = maxval( abs(mat(:j,j)) )
            matnorm = max( matnorm, sigma )
!
        end do
!
!$OMP END PARALLEL DO
!
    else
!
        chr = "l"
!
!$OMP PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)   &
!$OMP            ,SCHEDULE(STATIC,1)                   &
!$OMP            ,PRIVATE(j,sigma)                     &
!$OMP            ,REDUCTION(max:matnorm)
!
        do j = 1_i4b, n
!
            sigma = maxval( abs(mat(j:,j)) )
            matnorm = max( matnorm, sigma )
!
        end do
!
!$OMP END PARALLEL DO
!
    end if
!
!   SCALE MATRIX TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl( mat, one, sigma, chr )
!
!   CALL symtrid_cmp_a TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_a( mat, eigval, e, true, upper )
!
!   FIRST CALL ortho_gen_symtrid_a TO GENERATE THE ORTHOGONAL MATRIX,
!   THEN CALL symtrid_qri3_a TO COMPUTE EIGENVECTORS AND EIGENVALUES.
!
    call ortho_gen_symtrid_a( mat, upper )
!
    call symtrid_qri3_a( eigval, e, failure, mat, sort=sort,                  &
                         maxiter=maxiter, max_francis_steps=max_francis_steps )
!
!   IF MATRIX mat WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) eigval(:n) = ( one/sigma )*eigval(:n)
!
!
! END OF SUBROUTINE eig_cmp3_a
! ____________________________
!
    end subroutine eig_cmp3_a
!
! =========================================================================================
!
    subroutine eig_cmp3_b( mat, eigval, failure, sort, maxiter, max_francis_steps )
!
! Purpose
! _______
!
!   EIG_CMP3 computes all eigenvalues and eigenvectors of a
!   n-by-n real symmetric matrix MAT.
!
!
! Arguments
! _________
!
!   MAT                 (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                       On entry, the symmetric matrix MAT. The leading 
!                       n-by-n upper triangular part of MAT contains the
!                       upper triangular part of the matrix MAT. The strictly
!                       n-by-n lower triangular part of MAT is not referenced.
!
!                       On exit, MAT contains the orthonormal eigenvectors of
!                       the matrix MAT.
!
!                       The shape of MAT must verify: size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   EIGVAL              (OUTPUT) real(stnd), dimension(:)
!                       On exit, the eigenvalues.
!
!                       The size of EIGVAL must verify: size( EIGVAL ) = n .
!
!   FAILURE             (OUTPUT) logical(lgl)
!                       On exit:
!
!                       - FAILURE = false :  indicates successful exit.
!                       - FAILURE = true  :  indicates that the algorithm did not converge and
!                         that full accuracy was not attained in the Schur decomposition
!                         of an intermediate tridiagonal form T of MAT .
!
!   SORT                (INPUT, OPTIONAL) character
!                       Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                       or in descending order if SORT = 'D' or 'd'.
!                       The eigenvectors are reordered accordingly.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the Schur 
!                       decomposition of an intermediate tridiagonal form T of MAT .
!
!                       The algorithm fails to converge if the number of QR sweeps
!                       exceeds MAXITER * size(EIGVAL). Convergence usually occurs in 
!                       about 2 * size(EIGVAL) QR sweeps.
!
!                       The default is 30.
!
!   MAX_FRANCIS_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                       MAX_FRANCIS_STEPS controls the maximum number of Francis sets
!                       (e.g. QR sweeps) of Givens rotations which must be saved
!                       before applying them with a wavefront algorithm to accumulate
!                       the eigenvectors in the QR algorithm. MAX_FRANCIS_STEPS is a
!                       strictly positive integer, otherwise the default value is used.
!
!                       The default is 10.
!
!
! Further Details
! _______________
!
!   The matrix MAT is first transformed to tridiagonal form T, then the eigenvalues
!   and the eigenvectors are computed by the QR implicit algorithm.
!
!   For further details, see:
!
!   (1) Parlett, B.N., 1998:
!           The Symmetric Eigenvalue Problem.
!           Revised edition, SIAM, Philadelphia.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,    only : lamch
    use Utilities,        only : assert_eq, lascl
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:,:) :: mat
    real(stnd), intent(out),   dimension(:)   :: eigval
!
    integer(i4b), intent(in), optional :: maxiter, max_francis_steps
!
    logical(lgl), intent(out)           :: failure
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)  :: safmin, eps, smlnum, bignum, rmin, rmax, e(size(eigval)),      &
                   matnorm, sigma
!
    integer(i4b) :: n, j
!
    logical(lgl) :: mat_is_scaled
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='eig_cmp3'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(eigval),i4b),   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!
    if ( n==1_i4b ) then
!
        eigval(1_i4b)    = mat(1_i4b,1_i4b)
        mat(1_i4b,1_i4b) = one
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = zero
!
!$OMP PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)   &
!$OMP            ,SCHEDULE(STATIC,1)                   &
!$OMP            ,PRIVATE(j,sigma)                     &
!$OMP            ,REDUCTION(max:matnorm)
!
    do j = 1_i4b, n
!
        sigma = maxval( abs(mat(:j,j)) )
        matnorm = max( matnorm, sigma )
!
    end do
!
!$OMP END PARALLEL DO
!
!
!   SCALE MATRIX TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl( mat, one, sigma )
!
!   CALL symtrid_cmp_b TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_b( mat, eigval, e, true )
!
!   FIRST CALL ortho_gen_symtrid_b TO GENERATE THE ORTHOGONAL MATRIX,
!   THEN CALL symtrid_qri3_a TO COMPUTE EIGENVECTORS AND EIGENVALUES.
!
    call ortho_gen_symtrid_b( mat )
!
    call symtrid_qri3_a( eigval, e, failure, mat, sort=sort,                  &
                         maxiter=maxiter, max_francis_steps=max_francis_steps )
!
!   IF MATRIX mat WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) eigval(:n) = ( one/sigma )*eigval(:n)
!
!
! END OF SUBROUTINE eig_cmp3_b
! ____________________________
!
    end subroutine eig_cmp3_b
!
! =========================================================================================
!
    function eigvalues_c( mat, upper, sort, maxiter ) result( d )
!
! Purpose
! _______
!
!   Function EIGVALUES computes all eigenvalues of a n-by-n real symmetric matrix MAT.
!
!
! Arguments
! _________
!
!   MAT          (INPUT) real(stnd), dimension(:,:)
!                On entry, the symmetric matrix MAT.
!
!                If:
!
!                - UPPER = true:
!                  The leading n-by-n upper triangular part of MAT contains the
!                  upper triangular part of the matrix MAT.
!                - UPPER = false:
!                  The leading n-by-n lower triangular part of MAT contains
!                  the lower triangular part of the matrix MAT.
!
!                The shape of MAT must verify: size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   UPPER        (INPUT) logical(lgl)
!                Specifies whether the upper or lower triangular part of the
!                symmetric matrix MAT is stored. If:
!
!                - UPPER = true :  Upper triangular is stored ;
!                - UPPER = false:  Lower triangular is stored .
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!   MAXITER      (INPUT, OPTIONAL) integer(i4b)
!                MAXITER controls the maximum number of QR sweeps in the Schur 
!                decomposition of an intermediate tridiagonal form T of MAT .
!
!                The algorithm fails to converge if the number of QR sweeps
!                exceeds MAXITER * size(MAT,1). Convergence usually occurs in 
!                about 2 * size(MAT,1) QR sweeps.
!
!                The default is 30.
!
!
! Further Details
! _______________
!
!   This function is adapted from the routine  DSYEV in LAPACK77 (version 3).
!
!   The matrix MAT is first transformed to tridiagonal form T, then the eigenvalues
!   are computed by the Pal-Walker-Kahan variant of the QR algorithm. If the QR algorithm
!   fails to converge EIGVALUES returns a n-vector filled with NAN() function.
!
!   For further details, see:
!
!   (1) Parlett, B.N., 1998:
!           The Symmetric Eigenvalue Problem.
!           Revised edition, SIAM, Philadelphia.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch, nan
    use Utilities,       only : assert_eq, lascl, triangle
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in), dimension(:,:) :: mat
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(in) :: upper
!
    character ,   intent(in), optional  :: sort
!
    real(stnd)  :: d(size(mat,1))
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)  :: safmin, eps, smlnum, bignum, rmin, rmax,                             &
                   matnorm, sigma, e(size(mat,1)), matp(size(mat,1)*(size(mat,1)+1)/2)
!
    integer(i4b) :: n
!
    logical(lgl) :: mat_is_scaled, failure
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='eigvalues'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!
    if ( n==1_i4b ) then
!
        d(1_i4b) = mat(1_i4b,1_i4b)
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   MAKE A COPY OF mat IN PACKED FORM IN matp .
!
    matp = pack( mat, mask=triangle( upper, n, n, extra=1_i4b) )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = maxval( abs(matp) )
!
!   SCALE PACKED MATRIX matp TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl(  matp, one, sigma )
!
!   CALL symtrid_cmp_c TO REDUCE PACKED SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_c( matp, d, e, store_q=false, upper=upper )
!
!   CALL symtrid_qri_b TO COMPUTE EIGENVALUES.
!
    call symtrid_qri_b( d, e, failure, sort=sort, maxiter=maxiter )
!    
!   VERIFY FAILURE INDICATOR.
!    
    if ( failure ) then
!
        d(:n) = nan()
!
    else if ( mat_is_scaled ) then
!
!       IF MATRIX mat WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
        d(:n) = ( one/sigma )*d(:n)
!
    end if
!
!
! END OF FUNCTION eigvalues_c
! ___________________________
!
    end function eigvalues_c
!
! =========================================================================================
!
    function eigvalues_d( mat, sort, maxiter ) result( d )
!
! Purpose
! _______
!
!   Function EIGVALUES computes all eigenvalues of a n-by-n real symmetric matrix MAT.
!
!
! Arguments
! _________
!
!   MAT          (INPUT) real(stnd), dimension(:,:)
!                On entry, the symmetric matrix MAT. 
!                The leading n-by-n upper triangular part of MAT contains the
!                upper triangular part of the matrix MAT.
!                The strictly n-by-n lower triangular part of MAT is not used.
!
!                The shape of MAT must verify: size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!   MAXITER      (INPUT, OPTIONAL) integer(i4b)
!                MAXITER controls the maximum number of QR sweeps in the Schur 
!                decomposition of an intermediate tridiagonal form T of MAT .
!
!                The algorithm fails to converge if the number of QR sweeps
!                exceeds MAXITER * size(MAT,1). Convergence usually occurs in 
!                about 2 * size(MAT,1) QR sweeps.
!
!                The default is 30.
!
!
! Further Details
! _______________
!
!   This function is adapted from the routine  DSYEV in LAPACK77 (version 3).
!
!   The matrix MAT is first transformed to tridiagonal form T, then the eigenvalues
!   are computed by the Pal-Walker-Kahan variant of the QR algorithm. If the QR algorithm
!   fails to converge EIGVALUES returns a n-vector filled with NAN() function.
!
!   For further details, see:
!
!   (1) Parlett, B.N., 1998:
!           The Symmetric Eigenvalue Problem.
!           Revised edition, SIAM, Philadelphia.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch, nan
    use Utilities,       only : assert_eq, lascl, triangle
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in), dimension(:,:) :: mat
!
    integer(i4b), intent(in), optional :: maxiter
!
    character ,   intent(in), optional  :: sort
!
    real(stnd)  :: d(size(mat,1))
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)  :: safmin, eps, smlnum, bignum, rmin, rmax,                             &
                   matnorm, sigma, e(size(mat,1)), matp(size(mat,1)*(size(mat,1)+1)/2)
!
    integer(i4b) :: n
!
    logical(lgl) :: mat_is_scaled, failure
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='eigvalues'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!
    if ( n==1_i4b ) then
!
        d(1_i4b) = mat(1_i4b,1_i4b)
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   MAKE A COPY OF mat IN PACKED FORM IN matp .
!
    matp = pack( mat, mask=triangle( true, n, n, extra=1_i4b) )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = maxval( abs(matp) )
!
!   SCALE PACKED MATRIX matp TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl(  matp, one, sigma )
!
!   CALL symtrid_cmp_d TO REDUCE PACKED SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_d( matp, d, e, store_q=false )
!
!   CALL symtrid_qri_b TO COMPUTE EIGENVALUES.
!
    call symtrid_qri_b( d, e, failure, sort=sort, maxiter=maxiter )
!    
!   VERIFY FAILURE INDICATOR.
!    
    if ( failure ) then
!
        d(:n) = nan()
!
    else if ( mat_is_scaled ) then
!
!       IF MATRIX mat WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
        d(:n) = ( one/sigma )*d(:n)
!
    end if
!
!
! END OF FUNCTION eigvalues_d
! ___________________________
!
    end function eigvalues_d
!
! =========================================================================================
!
    subroutine eigval_cmp_a( mat, eigval, failure, upper, sort, maxiter, d_e )
!
! Purpose
! _______
!
!   EIGVAL_CMP computes all eigenvalues of a n-by-n real symmetric matrix MAT.
!
!   The matrix MAT is first transformed to symmetric tridiagonal form T
!   by an orthogonal similarity transformation:
!
!                        Q' * MAT * Q = T
!
!   ,then the eigenvalues are computed by the Pal-Walker-Kahan variant of the QR algorithm.
!   
!
! Arguments
! _________
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the symmetric matrix MAT.
!
!                If:
!
!                - UPPER = true:
!                  the leading n-by-n upper triangular part of MAT contains the
!                  upper triangular part of the matrix MAT.
!                - UPPER = false:
!                  the leading n-by-n lower triangular part of MAT contains
!                  the lower triangular part of the matrix MAT.
!
!                On exit:
!
!                - If UPPER = true and D_E is present :
!                  The leading n-by-n upper triangular part of MAT is overwritten
!                  by the matrix Q as a product of elementary reflectors.
!                - If UPPER = false and D_E is present :
!                  The leading n-by-n lower triangular part of MAT is overwritten
!                  by the matrix Q as a product of elementary reflectors.
!                - If UPPER = true and D_E is absent :
!                  The leading n-by-n upper triangular part of MAT is destroyed.
!                - If UPPER = false and D_E is absent :
!                  The leading n-by-n lower triangular part of MAT is destroyed.
!
!                The shape of MAT must verify: size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   EIGVAL       (OUTPUT) real(stnd), dimension(:)
!                On exit, the eigenvalues.
!
!                The size of EIGVAL must verify: size( EIGVAL ) = n .
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit.
!                - FAILURE = true  :  indicates that the algorithm did not converge and
!                  that full accuracy was not attained in the Schur decomposition
!                  of the intermediate tridiagonal form T of MAT .
!
!   UPPER        (INPUT) logical(lgl)
!                Specifies whether the upper or lower triangular part of the
!                symmetric matrix MAT is stored. If:
!
!                - UPPER = true :  Upper triangular is stored ;
!                - UPPER = false:  Lower triangular is stored .
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!   MAXITER      (INPUT, OPTIONAL) integer(i4b)
!                MAXITER controls the maximum number of QR sweeps in the Schur 
!                decomposition of the intermediate tridiagonal form T of MAT .
!
!                The algorithm fails to converge if the number of QR sweeps
!                exceeds MAXITER * size(EIGVAL). Convergence usually occurs in 
!                about 2 * size(EIGVAL) QR sweeps.
!
!                The default is 30.
!
!   D_E          (OUTPUT, OPTIONAL) real(stnd), dimension(:,2)
!                On exit, the first column of D_E contains the n diagonal elements
!                of the intermediate tridiagonal form T of MAT. The n-1 first
!                elements of the second column of D_E contains the n-1 subdiagonal
!                elements of the intermediate tridiagonal form T of MAT. D_E(n,2)
!                is arbitrary.
!
!                The shape of D_E must verify: size( D_E, 1 ) = n and size( D_E, 2 ) = 2
!
!
! Further Details
! _______________
!
!   This driver subroutine is adapted from the routine  DSYEV in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch, machhuge
    use Utilities,       only : assert, assert_eq, lascl
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout)          :: mat(:,:)
    real(stnd), intent(out)            :: eigval(:)
    real(stnd), intent(out), optional  :: d_e(:,:)
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(in)            :: upper
    logical(lgl), intent(out)           :: failure
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)  :: safmin, eps, smlnum, bignum, rmin, rmax,      &
                   sup(size(eigval)), matnorm, sigma
!
    integer(i4b) :: n, j
!
    logical(lgl) :: mat_is_scaled, store_q
!
    character :: chr
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='eigval_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(eigval),i4b),   &
                    name_proc )
!
    store_q = present( d_e )
!
    if ( store_q ) then
!
        call assert( logical( n==size(d_e,1), lgl ),         &
                     logical( 2==size(d_e,2), lgl ),         &
                     name_proc )
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            if ( store_q ) then
                d_e(1_i4b,1_i4b) = mat(1_i4b,1_i4b)
            end if
!
            eigval(1_i4b) = mat(1_i4b,1_i4b)
            mat(1_i4b,1_i4b)  = machhuge
!
        end if
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm =zero
!
    if ( upper ) then
!
        chr = "u"
!
!$OMP PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)   &
!$OMP            ,SCHEDULE(STATIC,1)                   &
!$OMP            ,PRIVATE(j,sigma)                     &
!$OMP            ,REDUCTION(max:matnorm)
!
        do j = 1_i4b, n
!
            sigma = maxval( abs(mat(:j,j)) )
            matnorm = max( matnorm, sigma )
!
        end do
!
!$OMP END PARALLEL DO
!
    else
!
        chr = "l"
!
!$OMP PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)   &
!$OMP            ,SCHEDULE(STATIC,1)                   &
!$OMP            ,PRIVATE(j,sigma)                     &
!$OMP            ,REDUCTION(max:matnorm)
!
        do j = 1_i4b, n
!
            sigma = maxval( abs(mat(j:,j)) )
            matnorm = max( matnorm, sigma )
!
        end do
!
!$OMP END PARALLEL DO
!
    end if
!
!   SCALE MATRIX TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl(  mat, one, sigma, chr )
!
!   CALL symtrid_cmp_a TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_a( mat, eigval, sup, store_q, upper )
!
    if ( store_q ) then
!
!       STORE INTERMEDIATE TRIDIAGONAL FORM OF mat , IF REQUIRED.
!
        d_e(:n,1_i4b) = eigval(:n)
        d_e(:n,2_i4b) = sup(:n)
!
    end if
!
!   CALL symtrid_qri_b TO COMPUTE EIGENVALUES.
!
    call symtrid_qri_b( eigval, sup, failure, sort=sort, maxiter=maxiter )
!
!   IF MATRIX WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) eigval(:n) = ( one/sigma )*eigval(:n)
!
!
! END OF SUBROUTINE eigval_cmp_a
! ______________________________
!
    end subroutine eigval_cmp_a
!
! =========================================================================================
!
    subroutine eigval_cmp_b( mat, eigval, failure, sort, maxiter, d_e )
!
! Purpose
! _______
!
!   EIGVAL_CMP computes all eigenvalues of a n-by-n real symmetric matrix MAT.
!
!   The matrix MAT is first transformed to symmetric tridiagonal form T
!   by an orthogonal similarity transformation:
!
!                        Q' * MAT * Q = T
!
!   ,then the eigenvalues are computed by the Pal-Walker-Kahan variant of the QR algorithm.
!   
!
! Arguments
! _________
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the symmetric matrix MAT. The leading 
!                n-by-n upper triangular part of MAT contains the
!                upper triangular part of the matrix MAT. The strictly
!                n-by-n lower triangular part of MAT is not referenced.
!
!                On exit:
!
!                - If D_E is present:
!                  The leading n-by-n upper triangular part of MAT is overwritten
!                  by the matrix Q as a product of elementary reflectors.
!                - If D_E is absent :
!                  The leading n-by-n upper triangular part of MAT is destroyed.
!
!                The shape of MAT must verify: size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   EIGVAL       (OUTPUT) real(stnd), dimension(:)
!                On exit, the eigenvalues.
!
!                The size of EIGVAL must verify: size( EIGVAL ) = n .
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit.
!                - FAILURE = true  :  indicates that the algorithm did not converge and
!                  that full accuracy was not attained in the Schur decomposition
!                  of an intermediate tridiagonal form T of MAT .
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!   MAXITER      (INPUT, OPTIONAL) integer(i4b)
!                MAXITER controls the maximum number of QR sweeps in the Schur 
!                decomposition of an intermediate tridiagonal form T of MAT .
!
!                The algorithm fails to converge if the number of QR sweeps
!                exceeds MAXITER * size(EIGVAL). Convergence usually occurs in 
!                about 2 * size(EIGVAL) QR sweeps.
!
!                The default is 30.
!
!   D_E          (OUTPUT, OPTIONAL) real(stnd), dimension(:,2)
!                On exit, the first column of D_E contains the n diagonal elements
!                of the intermediate tridiagonal form T of MAT. The n-1 first
!                elements of the second column of D_E contains the n-1 subdiagonal
!                elements of the intermediate tridiagonal form T of MAT. D_E(n,2)
!                is arbitrary.
!
!                The shape of D_E must verify: size( D_E, 1 ) = n and size( D_E, 2 ) = 2
!
!
! Further Details
! _______________
!
!   This driver subroutine is adapted from the routine  DSYEV in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch, machhuge
    use Utilities,       only : assert, assert_eq, lascl
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout)          :: mat(:,:)
    real(stnd), intent(out)            :: eigval(:)
    real(stnd), intent(out), optional  :: d_e(:,:)
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(out)          :: failure
!
    character ,   intent(in), optional :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)  :: safmin, eps, smlnum, bignum, rmin, rmax,      &
                   sup(size(eigval)), matnorm, sigma
!
    integer(i4b) :: n, j
!
    logical(lgl) :: mat_is_scaled, store_q
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='eigval_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(eigval),i4b),   &
                    name_proc )
!
    store_q = present( d_e )
!
    if ( store_q ) then
!
        call assert( logical( n==size(d_e,1), lgl ),         &
                     logical( 2==size(d_e,2), lgl ),         &
                     name_proc )
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            if ( store_q ) then
                d_e(1_i4b,1_i4b) = mat(1_i4b,1_i4b)
            end if
!
            eigval(1_i4b) = mat(1_i4b,1_i4b)
            mat(1_i4b,1_i4b)  = machhuge
!
        end if
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = zero
!
!$OMP PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)   &
!$OMP            ,SCHEDULE(STATIC,1)                   &
!$OMP            ,PRIVATE(j,sigma)                     &
!$OMP            ,REDUCTION(max:matnorm)
!
    do j = 1_i4b, n
!
        sigma = maxval( abs(mat(:j,j)) )
        matnorm = max( matnorm, sigma )
!
    end do
!
!$OMP END PARALLEL DO
!
!
!   SCALE MATRIX TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl( mat, one, sigma )
!
!   CALL symtrid_cmp_b TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_b( mat, eigval, sup, store_q )
!
    if ( store_q ) then
!
!       STORE INTERMEDIATE TRIDIAGONAL FORM OF mat , IF REQUIRED.
!
        d_e(:n,1_i4b) = eigval(:n)
        d_e(:n,2_i4b) = sup(:n)
!
    end if
!
!   CALL symtrid_qri_b TO COMPUTE EIGENVALUES.
!
    call symtrid_qri_b( eigval, sup, failure, sort=sort, maxiter=maxiter )
!
!   IF MATRIX WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) eigval(:n) = ( one/sigma )*eigval(:n)
!
!
! END OF SUBROUTINE eigval_cmp_b
! ______________________________
!
    end subroutine eigval_cmp_b
!
! =========================================================================================
!
    subroutine eigval_cmp_c( matp, eigval, failure, upper, sort, maxiter, d_e )
!
! Purpose
! _______
!
!   EIGVAL_CMP computes all eigenvalues of a n-by-n real symmetric matrix mat
!   stored in packed form in a linear array MATP.
!
!   The matrix mat is first transformed to symmetric tridiagonal form T
!   by an orthogonal similarity transformation:
!
!                        Q' * mat * Q = T
!
!   ,then the eigenvalues are computed by the Pal-Walker-Kahan variant of the QR algorithm.
!   
!
! Arguments
! _________
!
!   MATP         (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the upper or lower triangle of the symmetric matrix
!                mat, packed column-wise in a linear array. The j-th column of mat
!                is stored in the array MATP as follows:
!
!                - if UPPER = true,  MATP(i + (j-1) * j/2) = mat(i,j) for 1<=i<=j;
!                - if UPPER = false, MATP(i + (j-1) * (2 * n-j)/2) = mat(i,j) for j<=i<=n.
!
!                On exit:
!
!                - If D_E is present :
!                  MATP is overwritten by the matrix Q as a product
!                  of elementary reflectors.
!                - If D_E is absent :
!                  MATP is destroyed.
!
!                The size of MATP must verify:   size( MATP ) = (n * (n+1)/2)
!
!   EIGVAL       (OUTPUT) real(stnd), dimension(:)
!                On exit, the eigenvalues.
!
!                The size of EIGVAL must verify: size( EIGVAL ) = n .
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit.
!                - FAILURE = true  :  indicates that the algorithm did not converge and
!                  that full accuracy was not attained in the Schur decomposition
!                  of an intermediate tridiagonal form T of mat .
!
!   UPPER        (INPUT) logical(lgl)
!                Specifies whether the upper or lower triangular part of the
!                symmetric matrix mat is stored in the linear array MATP. If:
!
!                - UPPER = true :  Upper triangle of mat is stored;
!                - UPPER = false:  Lower triangle of mat is stored.
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!   MAXITER      (INPUT, OPTIONAL) integer(i4b)
!                MAXITER controls the maximum number of QR sweeps in the Schur 
!                decomposition of an intermediate tridiagonal form T of mat .
!
!                The algorithm fails to converge if the number of QR sweeps
!                exceeds MAXITER * size(EIGVAL). Convergence usually occurs in 
!                about 2 * size(EIGVAL) QR sweeps.
!
!                The default is 30.
!
!   D_E          (OUTPUT, OPTIONAL) real(stnd), dimension(:,2)
!                On exit, the first column of D_E contains the n diagonal elements
!                of the intermediate tridiagonal form T of mat. The n-1 first
!                elements of the second column of D_E contains the n-1 subdiagonal
!                elements of the intermediate tridiagonal form T of mat. D_E(n,2)
!                is arbitrary.
!
!                The shape of D_E must verify: size( D_E, 1 ) = n and size( D_E, 2 ) = 2
!
!
! Further Details
! _______________
!
!   This driver subroutine is adapted from the routine  DSYEV in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch, machhuge
    use Utilities,       only : assert, lascl
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout)          :: matp(:)
    real(stnd), intent(out)            :: eigval(:)
    real(stnd), intent(out), optional  :: d_e(:,:)
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in)            :: upper
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)  :: safmin, eps, smlnum, bignum, rmin, rmax,      &
                   sup(size(eigval)), matnorm, sigma
!
    integer(i4b) :: n, j
!
    logical(lgl) :: mat_is_scaled, store_q
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='eigval_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  size( eigval )
!
    j = n*(n+1_i4b)/2_i4b
    call assert( logical(size(matp)==j,lgl)  ,   &
                 name_proc )
!
    store_q = present( d_e )
!
    if ( store_q ) then
!
        call assert( logical( n==size(d_e,1), lgl ),         &
                     logical( 2==size(d_e,2), lgl ),         &
                     name_proc )
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            if ( store_q ) then
                d_e(1_i4b,1_i4b) = matp(1_i4b)
            end if
!
            eigval(1_i4b) = matp(1_i4b)
            matp(1_i4b)  = machhuge
!
        end if
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = maxval( abs(matp) )
!
!   SCALE MATRIX PACKED IN LINEAR ARRAY matp TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl(  matp, one, sigma )
!
!   CALL symtrid_cmp_c TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_c( matp, eigval, sup, store_q, upper )
!
    if ( store_q ) then
!
!       STORE INTERMEDIATE TRIDIAGONAL FORM OF mat , IF REQUIRED.
!
        d_e(:n,1_i4b) = eigval(:n)
        d_e(:n,2_i4b) = sup(:n)
!
    end if
!
!   CALL symtrid_qri_b TO COMPUTE EIGENVALUES.
!
    call symtrid_qri_b( eigval, sup, failure, sort=sort, maxiter=maxiter )
!
!   IF MATRIX WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) eigval(:n) = ( one/sigma )*eigval(:n)
!
!
! END OF SUBROUTINE eigval_cmp_c
! ______________________________
!
    end subroutine eigval_cmp_c
!
! =========================================================================================
!
    subroutine eigval_cmp_d( matp, eigval, failure, sort, maxiter, d_e )
!
! Purpose
! _______
!
!   EIGVAL_CMP computes all eigenvalues of a n-by-n real symmetric matrix mat
!   stored in packed form in a linear array MATP.
!
!   The matrix mat is first transformed to symmetric tridiagonal form T
!   by an orthogonal similarity transformation:
!
!                        Q' * mat * Q = T
!
!   ,then the eigenvalues are computed by the Pal-Walker-Kahan variant of the QR algorithm.
!   
!
! Arguments
! _________
!
!   MATP         (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the upper triangle of the symmetric matrix mat,
!                packed column-wise in a linear array. The j-th column of mat
!                is stored in the array MATP as follows:
!
!                      MATP(i + (j-1) * j/2) = mat(i,j) for 1<=i<=j;
!
!                On exit:
!
!                - If D_E is present :
!                  MATP is overwritten by the matrix Q as a product
!                  of elementary reflectors.
!                - If D_E is absent :
!                  MATP is destroyed.
!
!                The size of MATP must verify:  size( MATP ) = (n * (n+1)/2)
!
!   EIGVAL       (OUTPUT) real(stnd), dimension(:)
!                On exit, the eigenvalues.
!
!                The size of EIGVAL must verify: size( EIGVAL ) = n .
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit.
!                - FAILURE = true  :  indicates that the algorithm did not converge and
!                  that full accuracy was not attained in the Schur decomposition
!                  of an intermediate tridiagonal form T of mat .
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!   MAXITER      (INPUT, OPTIONAL) integer(i4b)
!                MAXITER controls the maximum number of QR sweeps in the Schur 
!                decomposition of an intermediate tridiagonal form T of mat .
!
!                The algorithm fails to converge if the number of QR sweeps
!                exceeds MAXITER * size(EIGVAL). Convergence usually occurs in 
!                about 2 * size(EIGVAL) QR sweeps.
!
!                The default is 30.
!
!   D_E          (OUTPUT, OPTIONAL) real(stnd), dimension(:,2)
!                On exit, the first column of D_E contains the n diagonal elements
!                of the intermediate tridiagonal form T of mat. The n-1 first
!                elements of the second column of D_E contains the n-1 subdiagonal
!                elements of the intermediate tridiagonal form T of mat. D_E(n,2)
!                is arbitrary.
!
!                The shape of D_E must verify: size( D_E, 1 ) = n and size( D_E, 2 ) = 2
!
!
! Further Details
! _______________
!
!   This driver subroutine is adapted from the routine  DSYEV in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch, machhuge
    use Utilities,       only : assert, lascl
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout)          :: matp(:)
    real(stnd), intent(out)            :: eigval(:)
    real(stnd), intent(out), optional  :: d_e(:,:)
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(out)           :: failure
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)  :: safmin, eps, smlnum, bignum, rmin, rmax,      &
                   sup(size(eigval)), matnorm, sigma
!
    integer(i4b) :: n, j
!
    logical(lgl) :: mat_is_scaled, store_q
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='eigval_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  size( eigval )
!
    j = n*(n+1_i4b)/2_i4b
    call assert( logical(size(matp)==j,lgl)  ,   &
                 name_proc )
!
    store_q = present( d_e )
!
    if ( store_q ) then
!
        call assert( logical( n==size(d_e,1), lgl ),         &
                     logical( 2==size(d_e,2), lgl ),         &
                     name_proc )
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            if ( store_q ) then
                d_e(1_i4b,1_i4b) = matp(1_i4b)
            end if
!
            eigval(1_i4b) = matp(1_i4b)
            matp(1_i4b)  = machhuge
!
        end if
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = maxval( abs(matp) )
!
!   SCALE MATRIX PACKED IN LINEAR ARRAY matp TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl(  matp, one, sigma )
!
!   CALL symtrid_cmp_d TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_d( matp, eigval, sup, store_q )
!
    if ( store_q ) then
!
!       STORE INTERMEDIATE TRIDIAGONAL FORM OF mat , IF REQUIRED.
!
        d_e(:n,1_i4b) = eigval(:n)
        d_e(:n,2_i4b) = sup(:n)
!
    end if
!
!   CALL symtrid_qri_b TO COMPUTE EIGENVALUES.
!
    call symtrid_qri_b( eigval, sup, failure, sort=sort, maxiter=maxiter )
!
!   IF MATRIX WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) eigval(:n) = ( one/sigma )*eigval(:n)
!
!
! END OF SUBROUTINE eigval_cmp_d
! ______________________________
!
    end subroutine eigval_cmp_d
!
! =========================================================================================
!
    subroutine eigval_cmp2_a( mat, eigval, failure, upper, sort, maxiter, d_e )
!
! Purpose
! _______
!
!   EIGVAL_CMP2 computes all eigenvalues of a n-by-n real symmetric matrix MAT.
!
!   The matrix MAT is first transformed to symmetric tridiagonal form T
!   by an orthogonal similarity transformation:
!
!                        Q' * MAT * Q = T
!
!   ,then the eigenvalues are computed by the Pal-Walker-Kahan variant of the QR algorithm.
!
!
! Arguments
! _________
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the symmetric matrix MAT.
!
!                If:
!
!                - UPPER = true:
!                  the leading n-by-n upper triangular part of MAT contains the
!                  upper triangular part of the matrix MAT.
!                - UPPER = false:
!                  the leading n-by-n lower triangular part of MAT contains
!                  the lower triangular part of the matrix MAT.
!
!                On exit:
!
!                - If UPPER = true and D_E is present :
!                  The leading n-by-n upper triangular part of MAT is overwritten
!                  by the matrix Q as a product of elementary reflectors.
!                - If UPPER = false and D_E is present :
!                  The leading n-by-n lower triangular part of MAT is overwritten
!                  by the matrix Q as a product of elementary reflectors.
!                - If UPPER = true and D_E is absent :
!                  The leading n-by-n upper triangular part of MAT is destroyed.
!                - If UPPER = false and D_E is absent :
!                  The leading n-by-n lower triangular part of MAT is destroyed.
!
!                The shape of MAT must verify: size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   EIGVAL       (OUTPUT) real(stnd), dimension(:)
!                On exit, the eigenvalues.
!
!                The size of EIGVAL must verify: size( EIGVAL ) = n .
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit.
!                - FAILURE = true  :  indicates that the algorithm did not converge and
!                  that full accuracy was not attained in the Schur decomposition
!                  of an intermediate tridiagonal form T of MAT .
!
!   UPPER        (INPUT) logical(lgl)
!                Specifies whether the upper or lower triangular part of the
!                symmetric matrix MAT is stored. If:
!
!                - UPPER = true :  Upper triangular is stored  ;
!                - UPPER = false:  Lower triangular  is stored .
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!   MAXITER      (INPUT, OPTIONAL) integer(i4b)
!                MAXITER controls the maximum number of QR sweeps in the Schur 
!                decomposition of an intermediate tridiagonal form T of MAT .
!
!                The algorithm fails to converge if the number of QR sweeps
!                exceeds MAXITER * size(EIGVAL). Convergence usually occurs in 
!                about 2 * size(EIGVAL) QR sweeps.
!
!                The default is 30.
!
!   D_E          (OUTPUT, OPTIONAL) real(stnd), dimension(:,2)
!                On exit, the first column of D_E contains the n diagonal elements
!                of the intermediate tridiagonal form T of MAT. The n-1 first
!                elements of the second column of D_E contains the n-1 subdiagonal
!                elements of the intermediate tridiagonal form T of MAT. D_E(n,2)
!                is arbitrary.
!
!                The shape of D_E must verify: size( D_E, 1 ) = n and size( D_E, 2 ) = 2
!
!
! Further Details
! _______________
!
!   This driver subroutine is adapted from the routine  DSYEV in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch, machhuge
    use Utilities,       only : assert, assert_eq, lascl
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout)          :: mat(:,:)
    real(stnd), intent(out)            :: eigval(:)
    real(stnd), intent(out), optional  :: d_e(:,:)
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(in)            :: upper
    logical(lgl), intent(out)           :: failure
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)  :: safmin, eps, smlnum, bignum, rmin, rmax,      &
                   sup(size(eigval)), matnorm, sigma
!
    integer(i4b) :: n, j
!
    logical(lgl) :: mat_is_scaled, store_q
!
    character :: chr
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='eigval_cmp2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(eigval),i4b),   &
                    name_proc )
!
    store_q = present( d_e )
!
    if ( store_q ) then
!
        call assert( logical( n==size(d_e,1), lgl ),         &
                     logical( 2==size(d_e,2), lgl ),         &
                     name_proc )
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            if ( store_q ) then
                d_e(1_i4b,1_i4b) = mat(1_i4b,1_i4b)
            end if
!
            eigval(1_i4b)    = mat(1_i4b,1_i4b)
            mat(1_i4b,1_i4b) = machhuge
!
        end if
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm =zero
    if ( upper ) then
!
        chr = "u"
!
!$OMP PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)   &
!$OMP            ,SCHEDULE(STATIC,1)                   &
!$OMP            ,PRIVATE(j,sigma)                     &
!$OMP            ,REDUCTION(max:matnorm)
!
        do j = 1_i4b, n
!
            sigma = maxval( abs(mat(:j,j)) )
            matnorm = max( matnorm, sigma )
!
        end do
!
!$OMP END PARALLEL DO
!
    else
!
        chr = "l"
!
!$OMP PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)   &
!$OMP            ,SCHEDULE(STATIC,1)                   &
!$OMP            ,PRIVATE(j,sigma)                     &
!$OMP            ,REDUCTION(max:matnorm)
!
        do j = 1_i4b, n
!
            sigma = maxval( abs(mat(j:,j)) )
            matnorm = max( matnorm, sigma )
!
        end do
!
!$OMP END PARALLEL DO
!
    end if
!
!   SCALE MATRIX TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl(  mat, one, sigma, chr )
!
!   CALL symtrid_cmp_a TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_a( mat, eigval, sup, store_q, upper )
!
    if ( store_q ) then
!
!       STORE INTERMEDIATE TRIDIAGONAL FORM OF mat , IF REQUIRED.
!
        d_e(:n,1_i4b) = eigval(:n)
        d_e(:n,2_i4b) = sup(:n)
!
    end if
!
!   CALL symtrid_qri2_b TO COMPUTE EIGENVALUES.
!
    call symtrid_qri2_b( eigval, sup, failure, sort=sort, maxiter=maxiter )
!
!   IF MATRIX WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) eigval(:n) = ( one/sigma )*eigval(:n)
!
!
! END OF SUBROUTINE eigval_cmp2_a
! _______________________________
!
    end subroutine eigval_cmp2_a
!
! =========================================================================================
!
    subroutine eigval_cmp2_b( mat, eigval, failure, sort, maxiter, d_e )
!
! Purpose
! _______
!
!   EIGVAL_CMP2 computes all eigenvalues of a n-by-n real symmetric matrix MAT.
!
!   The matrix MAT is first transformed to symmetric tridiagonal form T
!   by an orthogonal similarity transformation:
!
!                        Q' * MAT * Q = T
!
!   ,then the eigenvalues are computed by the Pal-Walker-Kahan variant of the QR algorithm.
!
!
! Arguments
! _________
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the symmetric matrix MAT. The leading 
!                n-by-n upper triangular part of MAT contains the
!                upper triangular part of the matrix MAT. The strictly
!                n-by-n lower triangular part of MAT is not referenced.
!
!                On exit:
!
!                - If D_E is present:
!                  The leading n-by-n upper triangular part of MAT is overwritten
!                  by the matrix Q as a product of elementary reflectors.
!                - If D_E is absent :
!                  The leading n-by-n upper triangular part of MAT is destroyed.
!
!                The shape of MAT must verify: size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   EIGVAL       (OUTPUT) real(stnd), dimension(:)
!                On exit, the eigenvalues.
!
!                The size of EIGVAL must verify: size( EIGVAL ) = n .
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit.
!                - FAILURE = true  :  indicates that the algorithm did not converge and
!                  that full accuracy was not attained in the Schur decomposition
!                  of an intermediate tridiagonal form T of MAT .
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!   MAXITER      (INPUT, OPTIONAL) integer(i4b)
!                MAXITER controls the maximum number of QR sweeps in the Schur 
!                decomposition of an intermediate tridiagonal form T of MAT .
!
!                The algorithm fails to converge if the number of QR sweeps
!                exceeds MAXITER * size(EIGVAL). Convergence usually occurs in 
!                about 2 * size(EIGVAL) QR sweeps.
!
!                The default is 30.
!
!   D_E          (OUTPUT, OPTIONAL) real(stnd), dimension(:,2)
!                On exit, the first column of D_E contains the n diagonal elements
!                of the intermediate tridiagonal form T of MAT. The n-1 first
!                elements of the second column of D_E contains the n-1 subdiagonal
!                elements of the intermediate tridiagonal form T of MAT. D_E(n,2)
!                is arbitrary.
!
!                The shape of D_E must verify: size( D_E, 1 ) = n and size( D_E, 2 ) = 2
!
!
! Further Details
! _______________
!
!   This driver subroutine is adapted from the routine  DSYEV in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch, machhuge
    use Utilities,       only : assert, assert_eq, lascl
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout)          :: mat(:,:)
    real(stnd), intent(out)            :: eigval(:)
    real(stnd), intent(out), optional  :: d_e(:,:)
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(out)          :: failure
!
    character ,   intent(in), optional :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)  :: safmin, eps, smlnum, bignum, rmin, rmax,      &
                   sup(size(eigval)), matnorm, sigma
!
    integer(i4b) :: n, j
!
    logical(lgl) :: mat_is_scaled, store_q
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='eigval_cmp2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(eigval),i4b),   &
                    name_proc )
!
    store_q = present( d_e )
!
    if ( store_q ) then
!
        call assert( logical( n==size(d_e,1), lgl ),         &
                     logical( 2==size(d_e,2), lgl ),         &
                     name_proc )
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            if ( store_q ) then
                d_e(1_i4b,1_i4b) = mat(1_i4b,1_i4b)
            end if
!
            eigval(1_i4b)    = mat(1_i4b,1_i4b)
            mat(1_i4b,1_i4b) = machhuge
!
        end if
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = zero
!
!$OMP PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)   &
!$OMP            ,SCHEDULE(STATIC,1)                   &
!$OMP            ,PRIVATE(j,sigma)                     &
!$OMP            ,REDUCTION(max:matnorm)
!
    do j = 1_i4b, n
!
        sigma = maxval( abs(mat(:j,j)) )
        matnorm = max( matnorm, sigma )
!
    end do
!
!$OMP END PARALLEL DO
!
!
!   SCALE MATRIX TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl( mat, one, sigma )
!
!   CALL symtrid_cmp_b TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_b( mat, eigval, sup, store_q )
!
    if ( store_q ) then
!
!       STORE INTERMEDIATE TRIDIAGONAL FORM OF mat , IF REQUIRED.
!
        d_e(:n,1_i4b) = eigval(:n)
        d_e(:n,2_i4b) = sup(:n)
!
    end if
!
!   CALL symtrid_qri2_b TO COMPUTE EIGENVALUES.
!
    call symtrid_qri2_b( eigval, sup, failure, sort=sort, maxiter=maxiter )
!
!   IF MATRIX WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) eigval(:n) = ( one/sigma )*eigval(:n)
!
!
! END OF SUBROUTINE eigval_cmp2_b
! _______________________________
!
    end subroutine eigval_cmp2_b
!
! =========================================================================================
!
    subroutine eigval_cmp2_c( matp, eigval, failure, upper, sort, maxiter, d_e )
!
! Purpose
! _______
!
!   EIGVAL_CMP2 computes all eigenvalues of a n-by-n real symmetric matrix mat
!   stored in packed form in a linear array MATP.
!
!   The matrix mat is first transformed to symmetric tridiagonal form T
!   by an orthogonal similarity transformation:
!
!                        Q' * mat * Q = T
!
!   ,then the eigenvalues are computed by the Pal-Walker-Kahan variant of the QR algorithm.
!
!
! Arguments
! _________
!
!   MATP         (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the upper or lower triangle of the symmetric matrix
!                mat , packed column-wise in a linear array.  The j-th column of mat
!                is stored in the array MATP as follows:
!
!                - if UPPER = true,  MATP(i + (j-1) * j/2) = mat(i,j) for 1<=i<=j;
!                - if UPPER = false, MATP(i + (j-1) * (2 * n-j)/2) = mat(i,j) for j<=i<=n.
!
!                On exit:
!
!                - If D_E is present :
!                  MATP is overwritten by the matrix Q as a product
!                  of elementary reflectors.
!                - If D_E is absent :
!                  MATP is destroyed.
!
!                The size of MATP must verify:   size( MATP ) = (n * (n+1)/2)
!
!   EIGVAL       (OUTPUT) real(stnd), dimension(:)
!                On exit, the eigenvalues.
!
!                The size of EIGVAL must verify: size( EIGVAL ) = n .
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit.
!                - FAILURE = true  :  indicates that the algorithm did not converge and
!                  that full accuracy was not attained in the Schur decomposition
!                  of an intermediate tridiagonal form T of mat .
!
!   UPPER        (INPUT) logical(lgl)
!                Specifies whether the upper or lower triangular part of the
!                symmetric matrix mat is stored in the linear array MATP. If:
!
!                - UPPER = true :  Upper triangle of mat is stored;
!                - UPPER = false:  Lower triangle of mat is stored.
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!   MAXITER      (INPUT, OPTIONAL) integer(i4b)
!                MAXITER controls the maximum number of QR sweeps in the Schur 
!                decomposition of an intermediate tridiagonal form T of mat .
!
!                The algorithm fails to converge if the number of QR sweeps
!                exceeds MAXITER * size(EIGVAL). Convergence usually occurs in 
!                about 2 * size(EIGVAL) QR sweeps.
!
!                The default is 30.
!
!   D_E         (OUTPUT, OPTIONAL) real(stnd), dimension(:,2)
!               On exit, the first column of D_E contains the n diagonal elements
!               of the intermediate tridiagonal form T of mat. The n-1 first
!               elements of the second column of D_E contains the n-1 subdiagonal
!               elements of the intermediate tridiagonal form T of mat. D_E(n,2)
!               is arbitrary.
!
!               The shape of D_E must verify: size( D_E, 1 ) = n and size( D_E, 2 ) = 2
!
!
! Further Details
! _______________
!
!   This driver subroutine is adapted from the routine  DSYEV in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch, machhuge
    use Utilities,       only : assert, lascl
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout)          :: matp(:)
    real(stnd), intent(out)            :: eigval(:)
    real(stnd), intent(out), optional  :: d_e(:,:)
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(out)  :: failure
    logical(lgl), intent(in)   :: upper
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)  :: safmin, eps, smlnum, bignum, rmin, rmax,      &
                   sup(size(eigval)), matnorm, sigma
!
    integer(i4b) :: n, j
!
    logical(lgl) :: mat_is_scaled, store_q
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='eigval_cmp2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  size(eigval)
!
    j = n*(n+1_i4b)/2_i4b
!
    call assert( logical(size(matp)==j,lgl)  ,   &
                 name_proc )
!
    store_q = present( d_e )
!
    if ( store_q ) then
!
        call assert( logical( n==size(d_e,1), lgl ),         &
                     logical( 2==size(d_e,2), lgl ),         &
                     name_proc )
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            if ( store_q ) then
                d_e(1_i4b,1_i4b) = matp(1_i4b)
            end if
!
            eigval(1_i4b) = matp(1_i4b)
            matp(1_i4b)  = machhuge
!
        end if
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = maxval( abs(matp) )
!
!   SCALE MATRIX PACKED IN LINEAR ARRAY matp TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl(  matp, one, sigma )
!
!   CALL symtrid_cmp_c TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_c( matp, eigval, sup, store_q, upper )
!
    if ( store_q ) then
!
!       STORE INTERMEDIATE TRIDIAGONAL FORM OF mat , IF REQUIRED.
!
        d_e(:n,1_i4b) = eigval(:n)
        d_e(:n,2_i4b) = sup(:n)
!
    end if
!
!   CALL symtrid_qri2_b TO COMPUTE EIGENVALUES.
!
    call symtrid_qri2_b( eigval, sup, failure, sort=sort, maxiter=maxiter )
!
!   IF MATRIX WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) eigval(:n) = ( one/sigma )*eigval(:n)
!
!
! END OF SUBROUTINE eigval_cmp2_c
! _______________________________
!
    end subroutine eigval_cmp2_c
!
! =========================================================================================
!
    subroutine eigval_cmp2_d( matp, eigval, failure, sort, maxiter, d_e )
!
! Purpose
! _______
!
!   EIGVAL_CMP2 computes all eigenvalues of a n-by-n real symmetric matrix mat
!   stored in packed form in a linear array MATP.
!
!   The matrix mat is first transformed to symmetric tridiagonal form T
!   by an orthogonal similarity transformation:
!
!                        Q' * mat * Q = T
!
!   ,then the eigenvalues are computed by the Pal-Walker-Kahan variant of the QR algorithm.
!
!
! Arguments
! _________
!
!   MATP         (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the upper triangle of the symmetric matrix
!                mat, packed column-wise in a linear array.  The j-th column of mat
!                is stored in the array MATP as follows:
!
!                       MATP(i + (j-1) * j/2) = mat(i,j) for 1<=i<=j;
!
!                On exit:
!
!                - If D_E is present :
!                  MATP is overwritten by the matrix Q as a product
!                  of elementary reflectors.
!                - If D_E is absent :
!                  MATP is destroyed.
!
!                The size of MATP must verify: size( MATP ) = (n * (n+1)/2)
!
!   EIGVAL       (OUTPUT) real(stnd), dimension(:)
!                On exit, the eigenvalues.
!
!                The size of EIGVAL must verify: size( EIGVAL ) = n .
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit.
!                - FAILURE = true  :  indicates that the algorithm did not converge and
!                  that full accuracy was not attained in the Schur decomposition
!                  of an intermediate tridiagonal form T of mat .
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!   MAXITER      (INPUT, OPTIONAL) integer(i4b)
!                MAXITER controls the maximum number of QR sweeps in the Schur 
!                decomposition of an intermediate tridiagonal form T of mat .
!
!                The algorithm fails to converge if the number of QR sweeps
!                exceeds MAXITER * size(EIGVAL). Convergence usually occurs in 
!                about 2 * size(EIGVAL) QR sweeps.
!
!                The default is 30.
!
!   D_E         (OUTPUT, OPTIONAL) real(stnd), dimension(:,2)
!               On exit, the first column of D_E contains the n diagonal elements
!               of the intermediate tridiagonal form T of mat. The n-1 first
!               elements of the second column of D_E contains the n-1 subdiagonal
!               elements of the intermediate tridiagonal form T of mat. D_E(n,2)
!               is arbitrary.
!
!               The shape of D_E must verify: size( D_E, 1 ) = n and size( D_E, 2 ) = 2
!
!
! Further Details
! _______________
!
!   This driver subroutine is adapted from the routine  DSYEV in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch, machhuge
    use Utilities,       only : assert, lascl
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout)          :: matp(:)
    real(stnd), intent(out)            :: eigval(:)
    real(stnd), intent(out), optional  :: d_e(:,:)
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(out)           :: failure
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)  :: safmin, eps, smlnum, bignum, rmin, rmax,      &
                   sup(size(eigval)), matnorm, sigma
!
    integer(i4b) :: n, j
!
    logical(lgl) :: mat_is_scaled, store_q
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='eigval_cmp2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  size(eigval)
!
    j = n*(n+1_i4b)/2_i4b
!
    call assert( logical(size(matp)==j,lgl)  ,   &
                 name_proc )
!
    store_q = present( d_e )
!
    if ( store_q ) then
!
        call assert( logical( n==size(d_e,1), lgl ),         &
                     logical( 2==size(d_e,2), lgl ),         &
                     name_proc )
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            if ( store_q ) then
                d_e(1_i4b,1_i4b) = matp(1_i4b)
            end if
!
            eigval(1_i4b) = matp(1_i4b)
            matp(1_i4b)   = machhuge
!
        end if
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = maxval( abs(matp) )
!
!   SCALE MATRIX PACKED IN LINEAR ARRAY matp TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl(  matp, one, sigma )
!
!   CALL symtrid_cmp_d TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_d( matp, eigval, sup, store_q )
!
    if ( store_q ) then
!
!       STORE INTERMEDIATE TRIDIAGONAL FORM OF mat , IF REQUIRED.
!
        d_e(:n,1_i4b) = eigval(:n)
        d_e(:n,2_i4b) = sup(:n)
!
    end if
!
!   CALL symtrid_qri2_b TO COMPUTE EIGENVALUES.
!
    call symtrid_qri2_b( eigval, sup, failure, sort=sort, maxiter=maxiter )
!
!   IF MATRIX WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) eigval(:n) = ( one/sigma )*eigval(:n)
!
!
! END OF SUBROUTINE eigval_cmp2_d
! _______________________________
!
    end subroutine eigval_cmp2_d
!
! =========================================================================================
!
    subroutine eigval_cmp3_a( mat, eigval, failure, upper, sort, maxiter, d_e )
!
! Purpose
! _______
!
!   EIGVAL_CMP3 computes all eigenvalues of a n-by-n real symmetric matrix MAT.
!
!   The matrix MAT is first transformed to symmetric tridiagonal form T
!   by an orthogonal similarity transformation:
!
!                        Q' * MAT * Q = T
!
!   ,then the eigenvalues are computed by the implicit QR algorithm.
!
!
! Arguments
! _________
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the symmetric matrix MAT.
!
!                If:
!
!                - UPPER = true:
!                  the leading n-by-n upper triangular part of MAT contains the
!                  upper triangular part of the matrix MAT.
!                - UPPER = false:
!                  the leading n-by-n lower triangular part of MAT contains
!                  the lower triangular part of the matrix MAT.
!
!                On exit:
!
!                - If UPPER = true and D_E is present :
!                  The leading n-by-n upper triangular part of MAT is overwritten
!                  by the matrix Q as a product of elementary reflectors.
!                - If UPPER = false and D_E is present :
!                  The leading n-by-n lower triangular part of MAT is overwritten
!                  by the matrix Q as a product of elementary reflectors.
!                - If UPPER = true and D_E is absent :
!                  The leading n-by-n upper triangular part of MAT is destroyed.
!                - If UPPER = false and D_E is absent :
!                  The leading n-by-n lower triangular part of MAT is destroyed.
!
!                The shape of MAT must verify: size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   EIGVAL       (OUTPUT) real(stnd), dimension(:)
!                On exit, the eigenvalues.
!
!                The size of EIGVAL must verify: size( EIGVAL ) = n .
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit.
!                - FAILURE = true  :  indicates that the algorithm did not converge and
!                  that full accuracy was not attained in the Schur decomposition
!                  of an intermediate tridiagonal form T of MAT .
!
!   UPPER        (INPUT) logical(lgl)
!                Specifies whether the upper or lower triangular part of the
!                symmetric matrix MAT is stored. If:
!
!                - UPPER = true :  Upper triangular is stored  ;
!                - UPPER = false:  Lower triangular is stored  .
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!   MAXITER      (INPUT, OPTIONAL) integer(i4b)
!                MAXITER controls the maximum number of QR sweeps in the Schur 
!                decomposition of an intermediate tridiagonal form T of MAT .
!
!                The algorithm fails to converge if the number of QR sweeps
!                exceeds MAXITER * size(EIGVAL). Convergence usually occurs in 
!                about 2 * size(EIGVAL) QR sweeps.
!
!                The default is 30.
!
!   D_E          (OUTPUT, OPTIONAL) real(stnd), dimension(:,2)
!                On exit, the first column of D_E contains the n diagonal elements
!                of the intermediate tridiagonal form T of MAT. The n-1 first
!                elements of the second column of D_E contains the n-1 subdiagonal
!                elements of the intermediate tridiagonal form T of MAT. D_E(n,2)
!                is arbitrary.
!
!                The shape of D_E must verify: size( D_E, 1 ) = n and size( D_E, 2 ) = 2
!
!
! Further Details
! _______________
!
!   This driver subroutine is adapted from the routine  DSYEV in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch, machhuge
    use Utilities,       only : assert, assert_eq, lascl
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout)          :: mat(:,:)
    real(stnd), intent(out)            :: eigval(:)
    real(stnd), intent(out), optional  :: d_e(:,:)
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(in)            :: upper
    logical(lgl), intent(out)           :: failure
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)  :: safmin, eps, smlnum, bignum, rmin, rmax,      &
                   sup(size(eigval)), matnorm, sigma
!
    integer(i4b) :: n, j
!
    logical(lgl) :: mat_is_scaled, store_q
!
    character :: chr
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='eigval_cmp3'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(eigval),i4b),   &
                    name_proc )
!
    store_q = present( d_e )
!
    if ( store_q ) then
!
        call assert( logical( n==size(d_e,1), lgl ),         &
                     logical( 2==size(d_e,2), lgl ),         &
                     name_proc )
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            if ( store_q ) then
                d_e(1_i4b,1_i4b) = mat(1_i4b,1_i4b)
            end if
!
            eigval(1_i4b) = mat(1_i4b,1_i4b)
            mat(1_i4b,1_i4b)  = machhuge
!
        end if
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm =zero
    if ( upper ) then
!
        chr = "u"
!
!$OMP PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)   &
!$OMP            ,SCHEDULE(STATIC,1)                   &
!$OMP            ,PRIVATE(j,sigma)                     &
!$OMP            ,REDUCTION(max:matnorm)
!
        do j = 1_i4b, n
!
            sigma = maxval( abs(mat(:j,j)) )
            matnorm = max( matnorm, sigma )
!
        end do
!
!$OMP END PARALLEL DO
!
    else
!
        chr = "l"
!
!$OMP PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)   &
!$OMP            ,SCHEDULE(STATIC,1)                   &
!$OMP            ,PRIVATE(j,sigma)                     &
!$OMP            ,REDUCTION(max:matnorm)
!
        do j = 1_i4b, n
            sigma = maxval( abs(mat(j:,j)) )
            matnorm = max( matnorm,  sigma )
        end do
!
!$OMP END PARALLEL DO
!
    end if
!
!   SCALE MATRIX TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl(  mat, one, sigma, chr )
!
!   CALL symtrid_cmp_a TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_a( mat, eigval, sup, store_q, upper )
!
    if ( store_q ) then
!
!       STORE INTERMEDIATE TRIDIAGONAL FORM OF mat , IF REQUIRED.
!
        d_e(:n,1_i4b) = eigval(:n)
        d_e(:n,2_i4b) = sup(:n)
!
    end if
!
!   CALL symtrid_qri3_b TO COMPUTE EIGENVALUES.
!
    call symtrid_qri3_b( eigval, sup, failure, sort=sort, maxiter=maxiter )
!
!   IF MATRIX WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) eigval(:n) = ( one/sigma )*eigval(:n)
!
!
! END OF SUBROUTINE eigval_cmp3_a
! _______________________________
!
    end subroutine eigval_cmp3_a
!
! =========================================================================================
!
    subroutine eigval_cmp3_b( mat, eigval, failure, sort, maxiter, d_e )
!
! Purpose
! _______
!
!   EIGVAL_CMP3 computes all eigenvalues of a n-by-n real symmetric matrix MAT.
!
!   The matrix MAT is first transformed to symmetric tridiagonal form T
!   by an orthogonal similarity transformation:
!
!                        Q' * MAT * Q = T
!
!   ,then the eigenvalues are computed by the implicit QR algorithm.
!
!
! Arguments
! _________
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the symmetric matrix MAT. The leading 
!                n-by-n upper triangular part of MAT contains the
!                upper triangular part of the matrix MAT. The strictly
!                n-by-n lower triangular part of MAT is not referenced.
!
!                On exit:
!
!                - If D_E is present:
!                  The leading n-by-n upper triangular part of MAT is overwritten
!                  by the matrix Q as a product of elementary reflectors.
!                - If D_E is absent :
!                  The leading n-by-n upper triangular part of MAT is destroyed.
!
!                The shape of MAT must verify: size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   EIGVAL       (OUTPUT) real(stnd), dimension(:)
!                On exit, the eigenvalues.
!
!                The size of EIGVAL must verify: size( EIGVAL ) = n .
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit.
!                - FAILURE = true  :  indicates that the algorithm did not converge and
!                  that full accuracy was not attained in the Schur decomposition
!                  of an intermediate tridiagonal form T of MAT .
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!   MAXITER      (INPUT, OPTIONAL) integer(i4b)
!                MAXITER controls the maximum number of QR sweeps in the Schur 
!                decomposition of an intermediate tridiagonal form T of MAT .
!
!                The algorithm fails to converge if the number of QR sweeps
!                exceeds MAXITER * size(EIGVAL). Convergence usually occurs in 
!                about 2 * size(EIGVAL) QR sweeps.
!
!                The default is 30.
!
!   D_E          (OUTPUT, OPTIONAL) real(stnd), dimension(:,2)
!                On exit, the first column of D_E contains the n diagonal elements
!                of the intermediate tridiagonal form T of MAT. The n-1 first
!                elements of the second column of D_E contains the n-1 subdiagonal
!                elements of the intermediate tridiagonal form T of MAT. D_E(n,2)
!                is arbitrary.
!
!                The shape of D_E must verify: size( D_E, 1 ) = n and size( D_E, 2 ) = 2
!
!
! Further Details
! _______________
!
!   This driver subroutine is adapted from the routine  DSYEV in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch, machhuge
    use Utilities,       only : assert, assert_eq, lascl
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout)          :: mat(:,:)
    real(stnd), intent(out)            :: eigval(:)
    real(stnd), intent(out), optional  :: d_e(:,:)
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(out)           :: failure
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)  :: safmin, eps, smlnum, bignum, rmin, rmax,      &
                   sup(size(eigval)), matnorm, sigma
!
    integer(i4b) :: n, j
!
    logical(lgl) :: mat_is_scaled, store_q
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='eigval_cmp3'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(eigval),i4b),   &
                    name_proc )
!
    store_q = present( d_e )
!
    if ( store_q ) then
!
        call assert( logical( n==size(d_e,1), lgl ),         &
                     logical( 2==size(d_e,2), lgl ),         &
                     name_proc )
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            if ( store_q ) then
                d_e(1_i4b,1_i4b) = mat(1_i4b,1_i4b)
            end if
!
            eigval(1_i4b) = mat(1_i4b,1_i4b)
            mat(1_i4b,1_i4b)  = machhuge
!
        end if
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = zero
!
!$OMP PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)   &
!$OMP            ,SCHEDULE(STATIC,1)                   &
!$OMP            ,PRIVATE(j,sigma)                     &
!$OMP            ,REDUCTION(max:matnorm)
!
    do j = 1_i4b, n
!
        sigma = maxval( abs(mat(:j,j)) )
        matnorm = max( matnorm, sigma )
!
    end do
!
!$OMP END PARALLEL DO
!
!
!   SCALE MATRIX TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl( mat, one, sigma )
!
!   CALL symtrid_cmp_b TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_b( mat, eigval, sup, store_q )
!
    if ( store_q ) then
!
!       STORE INTERMEDIATE TRIDIAGONAL FORM OF mat , IF REQUIRED.
!
        d_e(:n,1_i4b) = eigval(:n)
        d_e(:n,2_i4b) = sup(:n)
    end if
!
!   CALL symtrid_qri3_b TO COMPUTE EIGENVALUES.
!
    call symtrid_qri3_b( eigval, sup, failure, sort=sort, maxiter=maxiter )
!
!   IF MATRIX WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) eigval(:n) = ( one/sigma )*eigval(:n)
!
!
! END OF SUBROUTINE eigval_cmp3_b
! _______________________________
!
    end subroutine eigval_cmp3_b
!
! =========================================================================================
!
    subroutine eigval_cmp3_c( matp, eigval, failure, upper, sort, maxiter, d_e )
!
! Purpose
! _______
!
!   EIGVAL_CMP3 computes all eigenvalues of a n-by-n real symmetric matrix mat
!   stored in packed form in a linear array MATP.
!
!   The matrix mat is first transformed to symmetric tridiagonal form T
!   by an orthogonal similarity transformation:
!
!                        Q' * mat * Q = T
!
!   ,then the eigenvalues are computed by the implicit QR algorithm.
!
!
! Arguments
! _________
!
!   MATP         (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the upper or lower triangle of the symmetric matrix
!                mat , packed column-wise in a linear array.  The j-th column of mat
!                is stored in the array MATP as follows:
!
!                - if UPPER = true,  MATP(i + (j-1) * j/2) = mat(i,j) for 1<=i<=j;
!                - if UPPER = false, MATP(i + (j-1) * (2 * n-j)/2) = mat(i,j) for j<=i<=n.
!
!                On exit:
!
!                - If D_E is present :
!                  MATP is overwritten by the matrix Q as a product
!                  of elementary reflectors.
!                - If D_E is absent :
!                  MATP is destroyed.
!
!                The size of MATP must verify:   size( MATP ) = (n * (n+1)/2)
!
!   EIGVAL       (OUTPUT) real(stnd), dimension(:)
!                On exit, the eigenvalues.
!
!                The size of EIGVAL must verify: size( EIGVAL ) = n .
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit.
!                - FAILURE = true  :  indicates that the algorithm did not converge and
!                  that full accuracy was not attained in the Schur decomposition
!                  of an intermediate tridiagonal form T of mat .
!
!   UPPER        (INPUT) logical(lgl)
!                Specifies whether the upper or lower triangular part of the
!                symmetric matrix mat is stored in the linear array MATP. If:
!
!                - UPPER = true :  Upper triangle of mat is stored ;
!                - UPPER = false:  Lower triangle of mat is stored .
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!   MAXITER      (INPUT, OPTIONAL) integer(i4b)
!                MAXITER controls the maximum number of QR sweeps in the Schur 
!                decomposition of an intermediate tridiagonal form T of mat .
!
!                The algorithm fails to converge if the number of QR sweeps
!                exceeds MAXITER * size(EIGVAL). Convergence usually occurs in 
!                about 2 * size(EIGVAL) QR sweeps.
!
!                The default is 30.
!
!   D_E          (OUTPUT, OPTIONAL) real(stnd), dimension(:,2)
!                On exit, the first column of D_E contains the n diagonal elements
!                of the intermediate tridiagonal form T of mat. The n-1 first
!                elements of the second column of D_E contains the n-1 subdiagonal
!                elements of the intermediate tridiagonal form T of mat. D_E(n,2)
!                is arbitrary.
!
!                The shape of D_E must verify: size( D_E, 1 ) = n and size( D_E, 2 ) = 2
!
!
! Further Details
! _______________
!
!   This driver subroutine is adapted from the routine  DSYEV in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch, machhuge
    use Utilities,       only : assert, lascl
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout)          :: matp(:)
    real(stnd), intent(out)            :: eigval(:)
    real(stnd), intent(out), optional  :: d_e(:,:)
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(out) :: failure
    logical(lgl), intent(in)  :: upper
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)  :: safmin, eps, smlnum, bignum, rmin, rmax,      &
                   sup(size(eigval)), matnorm, sigma
!
    integer(i4b) :: n, j
!
    logical(lgl) :: mat_is_scaled, store_q
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='eigval_cmp3'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  size(eigval)
!
    j = n*(n+1_i4b)/2_i4b
!
    call assert( logical(size(matp)==j,lgl)  ,   &
                 name_proc )
!
    store_q = present( d_e )
!
    if ( store_q ) then
!
        call assert( logical( n==size(d_e,1), lgl ),         &
                     logical( 2==size(d_e,2), lgl ),         &
                     name_proc )
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            if ( store_q ) then
                d_e(1_i4b,1_i4b) = matp(1_i4b)
            end if
!
            eigval(1_i4b) = matp(1_i4b)
            matp(1_i4b)  = machhuge
!
        end if
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = maxval( abs(matp) )
!
!   SCALE MATRIX PACKED IN LINEAR ARRAY matp TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl(  matp, one, sigma )
!
!   CALL symtrid_cmp_c TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_c( matp, eigval, sup, store_q, upper )
!
    if ( store_q ) then
!
!       STORE INTERMEDIATE TRIDIAGONAL FORM OF mat , IF REQUIRED.
!
        d_e(:n,1_i4b) = eigval(:n)
        d_e(:n,2_i4b) = sup(:n)
!
    end if
!
!   CALL symtrid_qri3_b TO COMPUTE EIGENVALUES.
!
    call symtrid_qri3_b( eigval, sup, failure, sort=sort, maxiter=maxiter )
!
!   IF MATRIX WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) eigval(:n) = ( one/sigma )*eigval(:n)
!
!
! END OF SUBROUTINE eigval_cmp3_c
! _______________________________
!
    end subroutine eigval_cmp3_c
!
! =========================================================================================
!
    subroutine eigval_cmp3_d( matp, eigval, failure, sort, maxiter, d_e )
!
! Purpose
! _______
!
!   EIGVAL_CMP3 computes all eigenvalues of a n-by-n real symmetric matrix mat
!   stored in packed form in a linear array MATP.
!
!   The matrix mat is first transformed to symmetric tridiagonal form T
!   by an orthogonal similarity transformation:
!
!                        Q' * mat * Q = T
!
!   ,then the eigenvalues are computed by the implicit QR algorithm.
!
!
! Arguments
! _________
!
!   MATP         (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the upper triangle of the symmetric matrix mat ,
!                packed column-wise in a linear array.  The j-th column of mat
!                is stored in the array MATP as follows:
!
!                    MATP(i + (j-1) * j/2) = mat(i,j) for 1<=i<=j;
!
!                On exit:
!
!                - If D_E is present :
!                  MATP is overwritten by the matrix Q as a product
!                  of elementary reflectors.
!                - If D_E is absent :
!                  MATP is destroyed.
!
!                The size of MATP must verify:   size( MATP ) = (n * (n+1)/2)
!
!   EIGVAL       (OUTPUT) real(stnd), dimension(:)
!                On exit, the eigenvalues.
!
!                The size of EIGVAL must verify: size( EIGVAL ) = n .
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit.
!                - FAILURE = true  :  indicates that the algorithm did not converge and
!                  that full accuracy was not attained in the Schur decomposition
!                  of an intermediate tridiagonal form T of mat .
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'.
!
!   MAXITER      (INPUT, OPTIONAL) integer(i4b)
!                MAXITER controls the maximum number of QR sweeps in the Schur 
!                decomposition of an intermediate tridiagonal form T of mat .
!
!                The algorithm fails to converge if the number of QR sweeps
!                exceeds MAXITER * size(EIGVAL). Convergence usually occurs in 
!                about 2 * size(EIGVAL) QR sweeps.
!
!                The default is 30.
!
!   D_E          (OUTPUT, OPTIONAL) real(stnd), dimension(:,2)
!                On exit, the first column of D_E contains the n diagonal elements
!                of the intermediate tridiagonal form T of mat. The n-1 first
!                elements of the second column of D_E contains the n-1 subdiagonal
!                elements of the intermediate tridiagonal form T of mat. D_E(n,2)
!                is arbitrary.
!
!               The shape of D_E must verify: size( D_E, 1 ) = n and size( D_E, 2 ) = 2
!
!
! Further Details
! _______________
!
!   This driver subroutine is adapted from the routine  DSYEV in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch, machhuge
    use Utilities,       only : assert, lascl
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout)          :: matp(:)
    real(stnd), intent(out)            :: eigval(:)
    real(stnd), intent(out), optional  :: d_e(:,:)
!
    integer(i4b), intent(in), optional :: maxiter
!
    logical(lgl), intent(out)           :: failure
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)  :: safmin, eps, smlnum, bignum, rmin, rmax,      &
                   sup(size(eigval)), matnorm, sigma
!
    integer(i4b) :: n, j
!
    logical(lgl) :: mat_is_scaled, store_q
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='eigval_cmp3'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  size(eigval)
!
    j = n*(n+1_i4b)/2_i4b
!
    call assert( logical(size(matp)==j,lgl)  ,   &
                 name_proc )
!
    store_q = present( d_e )
!
    if ( store_q ) then
!
        call assert( logical( n==size(d_e,1), lgl ),         &
                     logical( 2==size(d_e,2), lgl ),         &
                     name_proc )
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            if ( store_q ) then
                d_e(1_i4b,1_i4b) = matp(1_i4b)
            end if
!
            eigval(1_i4b) = matp(1_i4b)
            matp(1_i4b)  = machhuge
        end if
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin = sqrt( smlnum )
    rmax = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = maxval( abs(matp) )
!
!   SCALE MATRIX PACKED IN LINEAR ARRAY matp TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl(  matp, one, sigma )
!
!   CALL symtrid_cmp_d TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_d( matp, eigval, sup, store_q )
!
    if ( store_q ) then
!
!       STORE INTERMEDIATE TRIDIAGONAL FORM OF mat , IF REQUIRED.
!
        d_e(:n,1_i4b) = eigval(:n)
        d_e(:n,2_i4b) = sup(:n)
!
    end if
!
!   CALL symtrid_qri3_b TO COMPUTE EIGENVALUES.
!
    call symtrid_qri3_b( eigval, sup, failure, sort=sort, maxiter=maxiter )
!
!   IF MATRIX WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) eigval(:n) = ( one/sigma )*eigval(:n)
!
!
! END OF SUBROUTINE eigval_cmp3_d
! _______________________________
!
    end subroutine eigval_cmp3_d
!
! =========================================================================================
!
    subroutine select_eigval_cmp_a( mat, eigval, small, failure, upper, d_e )
!
! Purpose
! _______
!
!   SELECT_EIGVAL_CMP computes the m=size( EIGVAL ) largest or smallest eigenvalues of
!   a n-by-n real symmetric matrix MAT.
!
!   The matrix MAT is first transformed to symmetric tridiagonal form T
!   by an orthogonal similarity transformation:
!
!                        Q' * MAT * Q = T
!
!   ,then the eigenvalues are computed by a rational QR method.
!   
!
! Arguments
! _________
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the symmetric matrix MAT.
!
!                If:
!
!                - UPPER = true:
!                  the leading n-by-n upper triangular part of MAT contains the
!                  upper triangular part of the matrix MAT.
!                - UPPER = false:
!                  the leading n-by-n lower triangular part of MAT contains
!                  the lower triangular part of the matrix MAT.
!
!                On exit:
!
!                - If UPPER = true and D_E is present :
!                  The leading n-by-n upper triangular part of MAT is overwritten
!                  by the matrix Q as a product of elementary reflectors.
!                - If UPPER = false and D_E is present :
!                  The leading n-by-n lower triangular part of MAT is overwritten
!                  by the matrix Q as a product of elementary reflectors.
!                - If UPPER = true and D_E is absent :
!                  The leading n-by-n upper triangular part of MAT is destroyed.
!                - If UPPER = false and D_E is absent :
!                  The leading n-by-n lower triangular part of MAT is destroyed.
!
!                The shape of MAT must verify: size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   EIGVAL       (OUTPUT) real(stnd), dimension(:)
!                On exit, the m=size( EIGVAL ) largest or smallest eigenvalues
!                of MAT in decreasing sequence.
!
!                The size of EIGVAL must verify: size( EIGVAL )<= size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   SMALL        (INPUT) logical(lgl)
!                On entry:
!
!                - SMALL = false :  indicates that the m largest eigenvalues are desired.
!                - SMALL = true  :  indicates that the m smallest eigenvalues are desired.
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit.
!                - FAILURE = true  :  indicates that the algorithm did not converge and
!                  that full accuracy was not attained in the rational QR iterations
!                  for some eigenvalues of the intermediate tridiagonal form T of MAT .
!
!   UPPER        (INPUT) logical(lgl)
!                Specifies whether the upper or lower triangular part of the
!                symmetric matrix MAT is stored. If:
!
!                - UPPER = true :  Upper triangular is stored ;
!                - UPPER = false:  Lower triangular is stored .
!
!   D_E          (OUTPUT, OPTIONAL) real(stnd), dimension(:,2)
!                On exit, the first column of D_E contains the n diagonal elements
!                of the intermediate tridiagonal form T of MAT. The n-1 first
!                elements of the second column of D_E contains the n-1 subdiagonal
!                elements of the intermediate tridiagonal form T of MAT. D_E(n,2)
!                is arbitrary.
!
!                The shape of D_E must verify: size( D_E, 1 ) = n and size( D_E, 2 ) = 2
!
!
! Further Details
! _______________
!
!   This driver subroutine is adapted from the routine  DSYEV in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch, machhuge
    use Utilities,       only : assert, assert_eq, lascl
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout)          :: mat(:,:)
    real(stnd), intent(out)            :: eigval(:)
    real(stnd), intent(out), optional  :: d_e(:,:)
!
    logical(lgl), intent(in)            :: small, upper
    logical(lgl), intent(out)           :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: safmin, eps, smlnum, bignum, rmin, rmax,   &
                  matnorm, sigma
    real(stnd), dimension(size(mat,1))  :: d, e
!
    integer(i4b) :: m, n, j
!
    logical(lgl) :: mat_is_scaled, store_q
!
    character :: chr
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='select_eigval_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    name_proc )
!
    m = size( eigval )
!
    call assert( logical( m<=n, lgl ), name_proc )
!
    store_q = present( d_e )
!
    if ( store_q ) then
!
        call assert( logical( n==size(d_e,1), lgl ),         &
                     logical( 2==size(d_e,2), lgl ),         &
                     name_proc )
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( m==1_i4b ) then
!
            if ( store_q ) then
                d_e(1_i4b,1_i4b) = mat(1_i4b,1_i4b)
            end if
!
            eigval(1_i4b)     = mat(1_i4b,1_i4b)
            mat(1_i4b,1_i4b)  = machhuge
!
        end if
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm =zero
!
    if ( upper ) then
!
        chr = "u"
!
!$OMP PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)   &
!$OMP            ,SCHEDULE(STATIC,1)                   &
!$OMP            ,PRIVATE(j,sigma)                     &
!$OMP            ,REDUCTION(max:matnorm)
!
        do j = 1_i4b, n
!
            sigma = maxval( abs(mat(:j,j)) )
            matnorm = max( matnorm, sigma )
!
        end do
!
!$OMP END PARALLEL DO
!
    else
!
        chr = "l"
!
!$OMP PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)   &
!$OMP            ,SCHEDULE(STATIC,1)                   &
!$OMP            ,PRIVATE(j,sigma)                     &
!$OMP            ,REDUCTION(max:matnorm)
!
        do j = 1_i4b, n
!
            sigma = maxval( abs(mat(j:,j)) )
            matnorm = max( matnorm, sigma )
!
        end do
!
!$OMP END PARALLEL DO
!
    end if
!
!   SCALE MATRIX TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl(  mat, one, sigma, chr )
!
!   CALL symtrid_cmp_a TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_a( mat, d, e, store_q, upper )
!
    if ( store_q ) then
!
!       STORE INTERMEDIATE TRIDIAGONAL FORM OF mat , IF REQUIRED.
!
        d_e(:n,1) = d(:n)
        d_e(:n,2) = e(:n)
!
    end if
!
!   STORE THE ELEMENTS OF THE TRIDIAGONAL MATRIX IN REVERSED ORDER IF
!   THE DIAGONAL ELEMENTS ARE INCREASING ALONG THE DIAGONAL.
!
    if ( reverse(d) ) then
!
        d(1_i4b:n) = d(n:1_i4b:-1_i4b)
        e(1_i4b:n-1_i4b) = e(n-1_i4b:1_i4b:-1_i4b)
!
    end if
!
!   CALL symtrid_ratqri TO COMPUTE SELECTED EIGENVALUES.
!
    call symtrid_ratqri( d, e, m, failure, small )
!
!   IF MATRIX WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) then 
!
        eigval(:m) = ( one/sigma )*d(:m)
!
    else
!
        eigval(:m) = d(:m)
!
    end if
!
!
! END OF SUBROUTINE select_eigval_cmp_a
! _____________________________________
!
    end subroutine select_eigval_cmp_a
!
! =========================================================================================
!
    subroutine select_eigval_cmp_b( mat, eigval, small, failure, d_e )
!
! Purpose
! _______
!
!   SELECT_EIGVAL_CMP computes the m=size( EIGVAL ) largest or smallest eigenvalues of
!   a n-by-n real symmetric matrix MAT.
!
!   The matrix MAT is first transformed to symmetric tridiagonal form T
!   by an orthogonal similarity transformation:
!
!                        Q' * MAT * Q = T
!
!   ,then the eigenvalues are computed by a rational QR method.
!   
!
! Arguments
! _________
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the symmetric matrix MAT. The leading 
!                n-by-n upper triangular part of MAT contains the
!                upper triangular part of the matrix MAT. The strictly
!                n-by-n lower triangular part of MAT is not referenced.
!
!                On exit:
!
!                - If D_E is present:
!                  The leading n-by-n upper triangular part of MAT is overwritten
!                  by the matrix Q as a product of elementary reflectors.
!                - If D_E is absent :
!                  The leading n-by-n upper triangular part of MAT is destroyed.
!
!                The shape of MAT must verify: size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   EIGVAL       (OUTPUT) real(stnd), dimension(:)
!                On exit, the m=size( EIGVAL ) largest or smallest eigenvalues
!                of MAT in decreasing sequence.
!
!                The size of EIGVAL must verify: size( EIGVAL )<= size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   SMALL        (INPUT) logical(lgl)
!                On entry:
!
!                - SMALL = false :  indicates that the m largest eigenvalues are desired.
!                - SMALL = true  :  indicates that the m smallest eigenvalues are desired.
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit.
!                - FAILURE = true  :  indicates that the algorithm did not converge and
!                  that full accuracy was not attained in the rational QR iterations
!                  for some eigenvalues of the intermediate tridiagonal form T of MAT .
!
!   D_E          (OUTPUT, OPTIONAL) real(stnd), dimension(:,2)
!                On exit, the first column of D_E contains the n diagonal elements
!                of the intermediate tridiagonal form T of MAT. The n-1 first
!                elements of the second column of D_E contains the n-1 subdiagonal
!                elements of the intermediate tridiagonal form T of MAT. D_E(n,2)
!                is arbitrary.
!
!                The shape of D_E must verify: size( D_E, 1 ) = n and size( D_E, 2 ) = 2
!
!
! Further Details
! _______________
!
!   This driver subroutine is adapted from the routine  DSYEV in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch, machhuge
    use Utilities,       only : assert, assert_eq, lascl
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout)          :: mat(:,:)
    real(stnd), intent(out)            :: eigval(:)
    real(stnd), intent(out), optional  :: d_e(:,:)
!
    logical(lgl), intent(in)            :: small
    logical(lgl), intent(out)           :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: safmin, eps, smlnum, bignum, rmin, rmax,   &
                  matnorm, sigma
    real(stnd), dimension(size(mat,1))  :: d, e
!
    integer(i4b) :: m, n, j
!
    logical(lgl) :: mat_is_scaled, store_q
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='select_eigval_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    name_proc )
!
    m = size( eigval )
    call assert( logical( m<=n, lgl ),         &
                          name_proc )
!
    store_q = present( d_e )
!
    if ( store_q ) then
!
        call assert( logical( n==size(d_e,1), lgl ),         &
                     logical( 2==size(d_e,2), lgl ),         &
                     name_proc )
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( m==1_i4b ) then
!
            if ( store_q ) then
                d_e(1_i4b,1_i4b) = mat(1_i4b,1_i4b)
            end if
!
            eigval(1_i4b)     = mat(1_i4b,1_i4b)
            mat(1_i4b,1_i4b)  = machhuge
!
        end if
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = zero
!
!$OMP PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)   &
!$OMP            ,SCHEDULE(STATIC,1)                   &
!$OMP            ,PRIVATE(j,sigma)                     &
!$OMP            ,REDUCTION(max:matnorm)
!
    do j = 1_i4b, n
!
        sigma = maxval( abs(mat(:j,j)) )
        matnorm = max( matnorm, sigma )
!
    end do
!
!$OMP END PARALLEL DO
!
!
!   SCALE MATRIX TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl( mat, one, sigma )
!
!   CALL symtrid_cmp_b TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_b( mat, d, e, store_q )
!
    if ( store_q ) then
!
!       STORE INTERMEDIATE TRIDIAGONAL FORM OF mat , IF REQUIRED.
!
        d_e(:n,1) = d(:n)
        d_e(:n,2) = e(:n)
!
    end if
!
!   STORE THE ELEMENTS OF THE TRIDIAGONAL MATRIX IN REVERSED ORDER IF
!   THE DIAGONAL ELEMENTS ARE INCREASING ALONG THE DIAGONAL.
!
    if ( reverse(d) ) then
!
        d(1_i4b:n) = d(n:1_i4b:-1_i4b)
        e(1_i4b:n-1_i4b) = e(n-1_i4b:1_i4b:-1_i4b)
!
    end if
!
!   CALL symtrid_ratqri TO COMPUTE SELECTED EIGENVALUES.
!
    call symtrid_ratqri( d, e, m, failure, small )
!
!   IF MATRIX WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) then 
!
        eigval(:m) = ( one/sigma )*d(:m)
!
    else
!
        eigval(:m) = d(:m)
!
    end if
!
!
! END OF SUBROUTINE select_eigval_cmp_b
! _____________________________________
!
    end subroutine select_eigval_cmp_b
!
! =========================================================================================
!
    subroutine select_eigval_cmp_c( matp, eigval, small, failure, upper, d_e )
!
! Purpose
! _______
!
!   SELECT_EIGVAL_CMP computes the m=size( EIGVAL ) largest or smallest eigenvalues of
!   a n-by-n real symmetric matrix mat stored in packed form in a linear array MATP.
!
!   The matrix mat is first transformed to symmetric tridiagonal form T
!   by an orthogonal similarity transformation:
!
!                        Q' * mat * Q = T
!
!   ,then the eigenvalues are computed by a rational QR method.
!   
!
! Arguments
! _________
!
!   MATP         (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the upper or lower triangle of the symmetric matrix
!                mat, packed column-wise in a linear array.  The j-th column of mat
!                is stored in the array MATP as follows:
!
!                - if UPPER = true,  MATP(i + (j-1) * j/2) = mat(i,j) for 1<=i<=j;
!                - if UPPER = false, MATP(i + (j-1) * (2 * n-j)/2) = mat(i,j) for j<=i<=n.
!
!                On exit:
!
!                - If D_E is present :
!                  MATP is overwritten by the matrix Q as a product
!                  of elementary reflectors.
!                - If D_E is absent :
!                  MATP is destroyed.
!
!                The size of MATP must verify:  size( MATP ) = (n * (n+1)/2).
!
!   EIGVAL       (OUTPUT) real(stnd), dimension(:)
!                On exit, the m=size( EIGVAL ) largest or smallest eigenvalues
!                of MAT in decreasing sequence.
!
!                The size of EIGVAL must verify: (m * (m+1))/2 <= size( MATP ) = (n * (n+1)/2) .
!
!   SMALL        (INPUT) logical(lgl)
!                On entry:
!
!                - SMALL = false :  indicates that the m largest eigenvalues are desired.
!                - SMALL = true  :  indicates that the m smallest eigenvalues are desired.
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit.
!                - FAILURE = true  :  indicates that the algorithm did not converge and
!                  that full accuracy was not attained in the rational QR iterations
!                  for some eigenvalues of the intermediate tridiagonal form T of MAT .
!
!   UPPER        (INPUT) logical(lgl)
!                Specifies whether the upper or lower triangular part of the
!                symmetric matrix mat is stored in the linear array MATP. If:
!
!                - UPPER = true :  Upper triangle of mat is stored;
!                - UPPER = false:  Lower triangle of mat is stored.
!
!   D_E          (OUTPUT, OPTIONAL) real(stnd), dimension(:,2)
!                On exit, the first column of D_E contains the n diagonal elements
!                of the intermediate tridiagonal form T of mat. The n-1 first
!                elements of the second column of D_E contains the n-1 subdiagonal
!                elements of the intermediate tridiagonal form T of mat. D_E(n,2)
!                is arbitrary.
!
!                The shape of D_E must verify: size( D_E, 1 ) = n and size( D_E, 2 ) = 2
!
!
! Further Details
! _______________
!
!   This driver subroutine is adapted from the routine  DSYEV in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch, machhuge
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert, lascl, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout)          :: matp(:)
    real(stnd), intent(out)            :: eigval(:)
    real(stnd), intent(out), optional  :: d_e(:,:)
!
    logical(lgl), intent(in)            :: small, upper
    logical(lgl), intent(out)           :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: safmin, eps, smlnum, bignum, rmin, rmax,   &
                  matnorm, sigma
    real(stnd), dimension(:), allocatable  :: d, e
!
    integer(i4b) :: m, n, j, p
    integer      :: iok
!
    logical(lgl) :: mat_is_scaled, store_q
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='select_eigval_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    m =  size( eigval )
    p =  size( matp )
!
    j = m*(m+1_i4b)/2_i4b
!
    call assert( logical(j<=p,lgl), name_proc )
!
!   COMPUTE THE SIZE OF THE ORIGINAL MATRIX mat .
!
    do n = m, p
!
        j = n*(n+1_i4b)/2_i4b
!
        if ( j==p ) exit
!
    end do
!
    store_q = present( d_e )
!
    if ( store_q ) then
!
        call assert( logical( n==size(d_e,1), lgl ),         &
                     logical( 2==size(d_e,2), lgl ),         &
                     name_proc )
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( m==1_i4b ) then
!
            if ( store_q ) then
                d_e(1_i4b,1_i4b) = matp(1_i4b)
            end if
!
            eigval(1_i4b) = matp(1_i4b)
            matp(1_i4b)   = machhuge
!
        end if
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = maxval( abs(matp) )
!
!   SCALE MATRIX PACKED IN LINEAR ARRAY matp TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl(  matp, one, sigma )
!
!   ALLOCATE WORK VARIABLES.
!
    allocate( d(n), e(n), stat = iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   CALL symtrid_cmp_c TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_c( matp, d, e, store_q, upper )
!
    if ( store_q ) then
!
!       STORE INTERMEDIATE TRIDIAGONAL FORM OF mat , IF REQUIRED.
!
        d_e(:n,1) = d(:n)
        d_e(:n,2) = e(:n)
!
    end if
!
!   STORE THE ELEMENTS OF THE TRIDIAGONAL MATRIX IN REVERSED ORDER IF
!   THE DIAGONAL ELEMENTS ARE INCREASING ALONG THE DIAGONAL.
!
    if ( reverse(d) ) then
!
        d(1_i4b:n) = d(n:1_i4b:-1_i4b)
        e(1_i4b:n-1_i4b) = e(n-1_i4b:1_i4b:-1_i4b)
!
    end if
!
!   CALL symtrid_ratqri TO COMPUTE SELECTED EIGENVALUES.
!
    call symtrid_ratqri( d, e, m, failure, small )
!
!   IF MATRIX WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) then 
!
        eigval(:m) = ( one/sigma )*d(:m)
!
    else
!
        eigval(:m) = d(:m)
!
    end if
!
!   DEALLOCATE WORK VARIABLES.
!
    deallocate( d, e )
!
!
! END OF SUBROUTINE select_eigval_cmp_c
! _____________________________________
!
    end subroutine select_eigval_cmp_c
!
! =========================================================================================
!
    subroutine select_eigval_cmp_d( matp, eigval, small, failure, d_e )
!
! Purpose
! _______
!
!   SELECT_EIGVAL_CMP computes the m=size( EIGVAL ) largest or smallest eigenvalues of
!   a n-by-n real symmetric matrix mat stored in packed form in a linear array MATP.
!
!   The matrix mat is first transformed to symmetric tridiagonal form T
!   by an orthogonal similarity transformation:
!
!                        Q' * mat * Q = T
!
!   ,then the eigenvalues are computed by a rational QR method.
!   
!
! Arguments
! _________
!
!   MATP         (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the upper triangle of the symmetric matrix mat,
!                packed column-wise in a linear array.  The j-th column of mat
!                is stored in the array MATP as follows:
!
!                         MATP(i + (j-1) * j/2) = mat(i,j) for 1<=i<=j;
!
!                On exit:
!
!                - If D_E is present :
!                  MATP is overwritten by the matrix Q as a product
!                  of elementary reflectors.
!                - If D_E is absent :
!                  MATP is destroyed.
!
!                The size of MATP must verify:   size( MATP ) = (n * (n+1)/2).
!
!   EIGVAL       (OUTPUT) real(stnd), dimension(:)
!                On exit, the m=size( EIGVAL ) largest or smallest eigenvalues
!                of MAT in decreasing sequence.
!
!                The size of EIGVAL must verify: (m * (m+1))/2 <= size( MATP ) = (n * (n+1)/2) .
!
!   SMALL        (INPUT) logical(lgl)
!                On entry:
!
!                - SMALL = false :  indicates that the m largest eigenvalues are desired.
!                - SMALL = true  :  indicates that the m smallest eigenvalues are desired.
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit.
!                - FAILURE = true  :  indicates that the algorithm did not converge and
!                  that full accuracy was not attained in the rational QR iterations
!                  for some eigenvalues of the intermediate tridiagonal form T of MAT .
!
!   D_E          (OUTPUT, OPTIONAL) real(stnd), dimension(:,2)
!                On exit, the first column of D_E contains the n diagonal elements
!                of the intermediate tridiagonal form T of mat. The n-1 first
!                elements of the second column of D_E contains the n-1 subdiagonal
!                elements of the intermediate tridiagonal form T of mat. D_E(n,2)
!                is arbitrary.
!
!                The shape of D_E must verify: size( D_E, 1 ) = n and size( D_E, 2 ) = 2
!
!
! Further Details
! _______________
!
!   This driver subroutine is adapted from the routine  DSYEV in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch, machhuge
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert, lascl, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout)          :: matp(:)
    real(stnd), intent(out)            :: eigval(:)
    real(stnd), intent(out), optional  :: d_e(:,:)
!
    logical(lgl), intent(in)            :: small
    logical(lgl), intent(out)           :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: safmin, eps, smlnum, bignum, rmin, rmax,   &
                  matnorm, sigma
    real(stnd), dimension(:), allocatable  :: d, e
!
    integer(i4b) :: m, n, j, p
    integer      :: iok
!
    logical(lgl) :: mat_is_scaled, store_q
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='select_eigval_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    m =  size( eigval )
    p =  size( matp )
!
    j = m*(m+1_i4b)/2_i4b
!
    call assert( logical(j<=p,lgl), name_proc )
!
!   COMPUTE THE SIZE OF THE ORIGINAL MATRIX mat .
!
    do n = m, p
!
        j = n*(n+1_i4b)/2_i4b
        if ( j==p ) exit
!
    end do
!
    store_q = present( d_e )
!
    if ( store_q ) then
!
        call assert( logical( n==size(d_e,1), lgl ),         &
                     logical( 2==size(d_e,2), lgl ),         &
                     name_proc )
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( m==1_i4b ) then
!
            if ( store_q ) then
                d_e(1_i4b,1_i4b) = matp(1_i4b)
            end if
!
            eigval(1_i4b) = matp(1_i4b)
            matp(1_i4b)   = machhuge
        end if
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = maxval( abs(matp) )
!
!   SCALE MATRIX PACKED IN LINEAR ARRAY matp TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl(  matp, one, sigma )
!
!   ALLOCATE WORK VARIABLES.
!
    allocate( d(n), e(n), stat = iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   CALL symtrid_cmp_d TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_d( matp, d, e, store_q )
!
    if ( store_q ) then
!
!       STORE INTERMEDIATE TRIDIAGONAL FORM OF mat , IF REQUIRED.
!
        d_e(:n,1) = d(:n)
        d_e(:n,2) = e(:n)
!
    end if
!
!   STORE THE ELEMENTS OF THE TRIDIAGONAL MATRIX IN REVERSED ORDER IF
!   THE DIAGONAL ELEMENTS ARE INCREASING ALONG THE DIAGONAL.
!
    if ( reverse(d) ) then
!
        d(1_i4b:n) = d(n:1_i4b:-1_i4b)
        e(1_i4b:n-1_i4b) = e(n-1_i4b:1_i4b:-1_i4b)
!
    end if
!
!   CALL symtrid_ratqri TO COMPUTE SELECTED EIGENVALUES.
!
    call symtrid_ratqri( d, e, m, failure, small )
!
!   IF MATRIX WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) then 
!
        eigval(:m) = ( one/sigma )*d(:m)
!
    else
!
        eigval(:m) = d(:m)
!
    end if
!
!   DEALLOCATE WORK VARIABLES.
!
    deallocate( d, e )
!
!
! END OF SUBROUTINE select_eigval_cmp_d
! _____________________________________
!
    end subroutine select_eigval_cmp_d
!
! =========================================================================================
!
    subroutine select_eigval_cmp2_a( mat, eigval, small, val, failure, upper, d_e )
!
! Purpose
! _______
!
!   SELECT_EIGVAL_CMP2 computes the largest or smallest eigenvalues of
!   a n-by-n real symmetric matrix MAT whose sum in algebraic value exceeds a given value.
!
!   The matrix MAT is first transformed to symmetric tridiagonal form T
!   by an orthogonal similarity transformation:
!
!                        Q' * MAT * Q = T
!
!   ,then the eigenvalues are computed by a rational QR method.
!   
!
! Arguments
! _________
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the symmetric matrix MAT.
!
!                If:
!
!                - UPPER = true:
!                  the leading n-by-n upper triangular part of MAT contains the
!                  upper triangular part of the matrix MAT.
!                - UPPER = false:
!                  the leading n-by-n lower triangular part of MAT contains
!                  the lower triangular part of the matrix MAT.
!
!                On exit:
!
!                - If UPPER = true and D_E is present :
!                  The leading n-by-n upper triangular part of MAT is overwritten
!                  by the matrix Q as a product of elementary reflectors.
!                - If UPPER = false and D_E is present :
!                  The leading n-by-n lower triangular part of MAT is overwritten
!                  by the matrix Q as a product of elementary reflectors.
!                - If UPPER = true and D_E is absent :
!                  The leading n-by-n upper triangular part of MAT is destroyed.
!                - If UPPER = false and D_E is absent :
!                  The leading n-by-n lower triangular part of MAT is destroyed.
!
!                The shape of MAT must verify: size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   EIGVAL       (OUTPUT) real(stnd), pointer, dimension(:)
!                On exit, the computed largest or smallest eigenvalues
!                of MAT in decreasing sequence.
!
!   SMALL        (INPUT) logical(lgl)
!                On entry:
!
!                - SMALL = false :  indicates that the largest eigenvalues are desired.
!                - SMALL = true  :  indicates that the smallest eigenvalues are desired.
!
!   VAL          (INPUT) real(stnd)
!                On entry, the sum of the m eigenvalues found will exceed abs(VAL) or 
!                m is equal to n.
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit.
!                - FAILURE = true  :  indicates that the algorithm did not converge and
!                  that full accuracy was not attained in the rational QR iterations
!                  for some eigenvalues of the intermediate tridiagonal form T of MAT .
!
!   UPPER        (INPUT) logical(lgl)
!                Specifies whether the upper or lower triangular part of the
!                symmetric matrix MAT is stored. If:
!
!                - UPPER = true :  Upper triangular is stored ;
!                - UPPER = false:  Lower triangular is stored .
!
!   D_E          (OUTPUT, OPTIONAL) real(stnd), dimension(:,2)
!                On exit, the first column of D_E contains the n diagonal elements
!                of the intermediate tridiagonal form T of MAT. The n-1 first
!                elements of the second column of D_E contains the n-1 subdiagonal
!                elements of the intermediate tridiagonal form T of MAT. D_E(n,2)
!                is arbitrary.
!
!                The shape of D_E must verify: size( D_E, 1 ) = n and size( D_E, 2 ) = 2
!
!
! Further Details
! _______________
!
!   This driver subroutine is adapted from the routine DSYEV in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch, machhuge
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert, assert_eq, lascl, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout)          :: mat(:,:)
    real(stnd), intent(in)             :: val
    real(stnd), pointer                :: eigval(:)
    real(stnd), intent(out), optional  :: d_e(:,:)
!
    logical(lgl), intent(in)            :: small, upper
    logical(lgl), intent(out)           :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: safmin, eps, smlnum, bignum, rmin, rmax,   &
                  matnorm, sigma
    real(stnd), dimension(size(mat,1))  :: d, e
!
    integer(i4b) :: m, n, j
    integer      :: iok
!
    logical(lgl) :: mat_is_scaled, store_q
!
    character :: chr
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='select_eigval_cmp2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    name_proc )
!
    store_q = present( d_e )
!
    if ( store_q ) then
!
        call assert( logical( n==size(d_e,1), lgl ),         &
                     logical( 2==size(d_e,2), lgl ),         &
                     name_proc )
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            if ( store_q ) then
                d_e(1_i4b,1_i4b) = mat(1_i4b,1_i4b)
            end if
!
!           ALLOCATE POINTER PARAMETER.
!
            allocate( eigval(1_i4b), stat = iok )
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
            eigval(1_i4b)     = mat(1_i4b,1_i4b)
            mat(1_i4b,1_i4b)  = machhuge
!
        end if
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm =zero
    if ( upper ) then
!
        chr = "u"
!
!$OMP PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)   &
!$OMP            ,SCHEDULE(STATIC,1)                   &
!$OMP            ,PRIVATE(j,sigma)                     &
!$OMP            ,REDUCTION(max:matnorm)
!
        do j = 1_i4b, n
!
            sigma = maxval( abs(mat(:j,j)) )
            matnorm = max( matnorm, sigma )
!
        end do
!
!$OMP END PARALLEL DO
!
    else
!
        chr = "l"
!
!$OMP PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)   &
!$OMP            ,SCHEDULE(STATIC,1)                   &
!$OMP            ,PRIVATE(j,sigma)                     &
!$OMP            ,REDUCTION(max:matnorm)
!
        do j = 1_i4b, n
!
            sigma = maxval( abs(mat(j:,j)) )
            matnorm = max( matnorm, sigma )
!
        end do
!
!$OMP END PARALLEL DO
!
    end if
!
!   SCALE MATRIX TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl(  mat, one, sigma, chr )
!
!   CALL symtrid_cmp_a TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_a( mat, d, e, store_q, upper )
!
    if ( store_q ) then
!
!       STORE INTERMEDIATE TRIDIAGONAL FORM OF mat , IF REQUIRED.
!
        d_e(:n,1) = d(:n)
        d_e(:n,2) = e(:n)
!
    end if
!
!   STORE THE ELEMENTS OF THE TRIDIAGONAL MATRIX IN REVERSED ORDER IF
!   THE DIAGONAL ELEMENTS ARE INCREASING ALONG THE DIAGONAL.
!
    if ( reverse(d) ) then
!
        d(1_i4b:n) = d(n:1_i4b:-1_i4b)
        e(1_i4b:n-1_i4b) = e(n-1_i4b:1_i4b:-1_i4b)
!
    end if
!
!   CALL symtrid_ratqri2 TO COMPUTE SELECTED EIGENVALUES.
!
    call symtrid_ratqri2( d, e, val, failure, m, small )
!
!   ALLOCATE POINTER PARAMETER.
!
    allocate( eigval(m), stat = iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   IF MATRIX WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) then 
!
        eigval(:m) = ( one/sigma )*d(:m)
!
    else
!
        eigval(:m) = d(:m)
!
    end if
!
!
! END OF SUBROUTINE select_eigval_cmp2_a
! ______________________________________
!
    end subroutine select_eigval_cmp2_a
!
! =========================================================================================
!
    subroutine select_eigval_cmp2_b( mat, eigval, small, val, failure, d_e )
!
! Purpose
! _______
!
!   SELECT_EIGVAL_CMP2 computes the largest or smallest eigenvalues of
!   a n-by-n real symmetric matrix MAT whose sum in algebraic value exceeds a given value.
!
!   The matrix MAT is first transformed to symmetric tridiagonal form T
!   by an orthogonal similarity transformation:
!
!                        Q' * MAT * Q = T
!
!   ,then the eigenvalues are computed by a rational QR method.
!   
!
! Arguments
! _________
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the symmetric matrix MAT. The leading 
!                n-by-n upper triangular part of MAT contains the
!                upper triangular part of the matrix MAT. The strictly
!                n-by-n lower triangular part of MAT is not referenced.
!
!                On exit:
!
!                - If D_E is present:
!                  The leading n-by-n upper triangular part of MAT is overwritten
!                  by the matrix Q as a product of elementary reflectors.
!                - If D_E is absent :
!                  The leading n-by-n upper triangular part of MAT is destroyed.
!
!                The shape of MAT must verify: size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   EIGVAL       (OUTPUT) real(stnd), pointer, dimension(:)
!                On exit, the computed largest or smallest eigenvalues
!                of MAT in decreasing sequence.
!
!   SMALL        (INPUT) logical(lgl)
!                On entry:
!
!                - SMALL = false :  indicates that the largest eigenvalues are desired.
!                - SMALL = true  :  indicates that the smallest eigenvalues are desired.
!
!   VAL          (INPUT) real(stnd)
!                On entry, the sum of the m eigenvalues found will exceed abs(VAL) or 
!                m is equal to n.
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit.
!                - FAILURE = true  :  indicates that the algorithm did not converge and
!                  that full accuracy was not attained in the rational QR iterations
!                  for some eigenvalues of the intermediate tridiagonal form T of MAT .
!
!   D_E          (OUTPUT, OPTIONAL) real(stnd), dimension(:,2)
!                On exit, the first column of D_E contains the n diagonal elements
!                of the intermediate tridiagonal form T of MAT. The n-1 first
!                elements of the second column of D_E contains the n-1 subdiagonal
!                elements of the intermediate tridiagonal form T of MAT. D_E(n,2)
!                is arbitrary.
!
!                The shape of D_E must verify: size( D_E, 1 ) = n and size( D_E, 2 ) = 2
!
!
! Further Details
! _______________
!
!   This driver subroutine is adapted from the routine  DSYEV in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch, machhuge
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert, assert_eq, lascl, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout)          :: mat(:,:)
    real(stnd), intent(in)             :: val
    real(stnd), pointer                :: eigval(:)
    real(stnd), intent(out), optional  :: d_e(:,:)
!
    logical(lgl), intent(in)            :: small
    logical(lgl), intent(out)           :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: safmin, eps, smlnum, bignum, rmin, rmax,   &
                  matnorm, sigma
    real(stnd), dimension(size(mat,1))  :: d, e
!
    integer(i4b) :: m, n, j
    integer      :: iok
!
    logical(lgl) :: mat_is_scaled, store_q
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='select_eigval_cmp2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    name_proc )
!
    store_q = present( d_e )
    if ( store_q ) then
        call assert( logical( n==size(d_e,1), lgl ),         &
                     logical( 2==size(d_e,2), lgl ),         &
                     name_proc )
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            if ( store_q ) then
                d_e(1_i4b,1_i4b) = mat(1_i4b,1_i4b)
            end if
!
!           ALLOCATE POINTER PARAMETER.
!
            allocate( eigval(1_i4b), stat = iok )
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
            eigval(1_i4b)     = mat(1_i4b,1_i4b)
            mat(1_i4b,1_i4b)  = machhuge
        end if
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = zero
!
!$OMP PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)   &
!$OMP            ,SCHEDULE(STATIC,1)                   &
!$OMP            ,PRIVATE(j,sigma)                     &
!$OMP            ,REDUCTION(max:matnorm)
!
    do j = 1_i4b, n
!
        sigma = maxval( abs(mat(:j,j)) )
        matnorm = max( matnorm, sigma )
!
    end do
!
!$OMP END PARALLEL DO
!
!
!   SCALE MATRIX TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl( mat, one, sigma )
!
!   CALL symtrid_cmp_b TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_b( mat, d, e, store_q )
!
    if ( store_q ) then
!
!       STORE INTERMEDIATE TRIDIAGONAL FORM OF mat , IF REQUIRED.
!
        d_e(:n,1) = d(:n)
        d_e(:n,2) = e(:n)
!
    end if
!
!   STORE THE ELEMENTS OF THE TRIDIAGONAL MATRIX IN REVERSED ORDER IF
!   THE DIAGONAL ELEMENTS ARE INCREASING ALONG THE DIAGONAL.
!
    if ( reverse(d) ) then
!
        d(1_i4b:n) = d(n:1_i4b:-1_i4b)
        e(1_i4b:n-1_i4b) = e(n-1_i4b:1_i4b:-1_i4b)
!
    end if
!
!   CALL symtrid_ratqri2 TO COMPUTE SELECTED EIGENVALUES.
!
    call symtrid_ratqri2( d, e, val, failure, m, small )
!
!   ALLOCATE POINTER PARAMETER.
!
    allocate( eigval(m), stat = iok )
!
    if ( iok/=0 ) then
!
        call merror( name_proc//allocate_error )
!
    end if
!
!   IF MATRIX WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) then 
!
        eigval(:m) = ( one/sigma )*d(:m)
!
    else
!
        eigval(:m) = d(:m)
!
    end if
!
!
! END OF SUBROUTINE select_eigval_cmp2_b
! ______________________________________
!
    end subroutine select_eigval_cmp2_b
!
! =========================================================================================
!
    subroutine select_eigval_cmp2_c( matp, eigval, small, val, failure, upper, d_e )
!
! Purpose
! _______
!
!   SELECT_EIGVAL_CMP2 computes the largest or smallest eigenvalues of
!   a n-by-n real symmetric matrix mat, stored in packed form in a linear array MATP,
!   whose sum in algebraic value exceeds a given value.
!
!   The matrix mat is first transformed to symmetric tridiagonal form T
!   by an orthogonal similarity transformation:
!
!                        Q' * mat * Q = T
!
!   ,then the eigenvalues are computed by a rational QR method.
!   
!
! Arguments
! _________
!
!   MATP         (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the upper or lower triangle of the symmetric matrix
!                mat, packed column-wise in a linear array.  The j-th column of mat
!                is stored in the array MATP as follows:
!
!                - if UPPER = true,  MATP(i + (j-1) * j/2) = mat(i,j) for 1<=i<=j;
!                - if UPPER = false, MATP(i + (j-1) * (2 * n-j)/2) = mat(i,j) for j<=i<=n.
!
!                On exit:
!
!                - If D_E is present :
!                  MATP is overwritten by the matrix Q as a product
!                  of elementary reflectors.
!                - If D_E is absent :
!                  MATP is destroyed.
!
!                The size of MATP must verify:    size( MATP ) = (n * (n+1)/2).
!
!   EIGVAL       (OUTPUT) real(stnd), pointer, dimension(:)
!                On exit, the computed largest or smallest eigenvalues
!                of MAT in decreasing sequence.
!
!   SMALL        (INPUT) logical(lgl)
!                On entry:
!
!                - SMALL = false :  indicates that the largest eigenvalues are desired.
!                - SMALL = true  :  indicates that the smallest eigenvalues are desired.
!
!   VAL          (INPUT) real(stnd)
!                On entry, the sum of the m eigenvalues found will exceed abs(VAL) or 
!                m is equal to n.
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit.
!                - FAILURE = true  :  indicates that the algorithm did not converge and
!                  that full accuracy was not attained in the rational QR iterations
!                  for some eigenvalues of the intermediate tridiagonal form T of MAT .
!
!   UPPER        (INPUT) logical(lgl)
!                Specifies whether the upper or lower triangular part of the
!                symmetric matrix mat is stored in the linear array MATP. If:
!
!                - UPPER = true :  Upper triangle of mat is stored;
!                - UPPER = false:  Lower triangle of mat is stored.
!
!   D_E          (OUTPUT, OPTIONAL) real(stnd), dimension(:,2)
!                On exit, the first column of D_E contains the n diagonal elements
!                of the intermediate tridiagonal form T of mat. The n-1 first
!                elements of the second column of D_E contains the n-1 subdiagonal
!                elements of the intermediate tridiagonal form T of mat. D_E(n,2)
!                is arbitrary.
!
!                The shape of D_E must verify: size( D_E, 1 ) = n and size( D_E, 2 ) = 2
!
!
! Further Details
! _______________
!
!   This driver subroutine is adapted from the routine  DSYEV in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch, machhuge
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert, lascl, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout)          :: matp(:)
    real(stnd), intent(in)             :: val
    real(stnd), pointer                :: eigval(:)
    real(stnd), intent(out), optional  :: d_e(:,:)
!
    logical(lgl), intent(in)            :: small, upper
    logical(lgl), intent(out)           :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: safmin, eps, smlnum, bignum, rmin, rmax,   &
                  matnorm, sigma
    real(stnd), dimension(:), allocatable  :: d, e
!
    integer(i4b) :: m, n, j, p
    integer      :: iok
!
    logical(lgl) :: mat_is_scaled, store_q
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='select_eigval_cmp2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
    p =  size( matp )
!
!   COMPUTE THE SIZE OF THE ORIGINAL MATRIX mat .
!
    do n = 1_i4b, p
!
        j = n*(n+1_i4b)/2_i4b
!
        if ( j==p ) exit
!
    end do
!
    store_q = present( d_e )
!
    if ( store_q ) then
!
        call assert( logical( n==size(d_e,1), lgl ),         &
                     logical( 2==size(d_e,2), lgl ),         &
                     name_proc )
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            if ( store_q ) then
                d_e(1_i4b,1_i4b) = matp(1_i4b)
            end if
!
!           ALLOCATE POINTER PARAMETER.
!
            allocate( eigval(1_i4b), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
            eigval(1_i4b) = matp(1_i4b)
            matp(1_i4b)   = machhuge
!
        end if
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = maxval( abs(matp) )
!
!   SCALE MATRIX PACKED IN LINEAR ARRAY matp TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl(  matp, one, sigma )
!
!   ALLOCATE WORK VARIABLES.
!
    allocate( d(n), e(n), stat = iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   CALL symtrid_cmp_c TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_c( matp, d, e, store_q, upper )
!
    if ( store_q ) then
!
!       STORE INTERMEDIATE TRIDIAGONAL FORM OF mat , IF REQUIRED.
!
        d_e(:n,1) = d(:n)
        d_e(:n,2) = e(:n)
!
    end if
!
!   STORE THE ELEMENTS OF THE TRIDIAGONAL MATRIX IN REVERSED ORDER IF
!   THE DIAGONAL ELEMENTS ARE INCREASING ALONG THE DIAGONAL.
!
    if ( reverse(d) ) then
!
        d(1_i4b:n) = d(n:1_i4b:-1_i4b)
        e(1_i4b:n-1_i4b) = e(n-1_i4b:1_i4b:-1_i4b)
!
    end if
!
!   CALL symtrid_ratqri2 TO COMPUTE SELECTED EIGENVALUES.
!
    call symtrid_ratqri2( d, e, val, failure, m, small )
!
!   ALLOCATE POINTER PARAMETER.
!
    allocate( eigval(m), stat = iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   IF MATRIX WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) then 
!
        eigval(:m) = ( one/sigma )*d(:m)
!
    else
!
        eigval(:m) = d(:m)
!
    end if
!
!   DEALLOCATE WORK VARIABLES.
!
    deallocate( d, e )
!
!
! END OF SUBROUTINE select_eigval_cmp2_c
! ______________________________________
!
    end subroutine select_eigval_cmp2_c
!
! =========================================================================================
!
    subroutine select_eigval_cmp2_d( matp, eigval, small, val, failure, d_e )
!
! Purpose
! _______
!
!   SELECT_EIGVAL_CMP2 computes the largest or smallest eigenvalues of
!   a n-by-n real symmetric matrix mat, stored in packed form in a linear array MATP,
!   whose sum in algebraic value exceeds a given value.
!
!   The matrix mat is first transformed to symmetric tridiagonal form T
!   by an orthogonal similarity transformation:
!
!                        Q' * mat * Q = T
!
!   ,then the eigenvalues are computed by a rational QR method.
!   
!
! Arguments
! _________
!
!   MATP         (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the upper triangle of the symmetric matrix mat ,
!                packed column-wise in a linear array.  The j-th column of mat
!                is stored in the array MATP as follows:
!
!                       MATP(i + (j-1) * j/2) = mat(i,j) for 1<=i<=j;
!
!                On exit:
!
!                - If D_E is present :
!                  MATP is overwritten by the matrix Q as a product
!                  of elementary reflectors.
!                - If D_E is absent :
!                  MATP is destroyed.
!
!                The size of MATP must verify:   size( MATP ) = (n * (n+1)/2).
!
!   EIGVAL       (OUTPUT) real(stnd), pointer, dimension(:)
!                On exit, the computed largest or smallest eigenvalues
!                of MAT in decreasing sequence.
!
!   SMALL        (INPUT) logical(lgl)
!                On entry:
!
!                - SMALL = false :  indicates that the largest eigenvalues are desired.
!                - SMALL = true  :  indicates that the smallest eigenvalues are desired.
!
!   VAL          (INPUT) real(stnd)
!                On entry, the sum of the m eigenvalues found will exceed abs(VAL) or 
!                m is equal to n.
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit.
!                - FAILURE = true  :  indicates that the algorithm did not converge and
!                  that full accuracy was not attained in the rational QR iterations
!                  for some eigenvalues of the intermediate tridiagonal form T of MAT .
!
!   D_E          (OUTPUT, OPTIONAL) real(stnd), dimension(:,2)
!                On exit, the first column of D_E contains the n diagonal elements
!                of the intermediate tridiagonal form T of mat. The n-1 first
!                elements of the second column of D_E contains the n-1 subdiagonal
!                elements of the intermediate tridiagonal form T of mat. D_E(n,2)
!                is arbitrary.
!
!                The shape of D_E must verify: size( D_E, 1 ) = n and size( D_E, 2 ) = 2
!
!
! Further Details
! _______________
!
!   This driver subroutine is adapted from the routine DSYEV in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch, machhuge
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert, lascl, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout)          :: matp(:)
    real(stnd), intent(in)             :: val
    real(stnd), pointer                :: eigval(:)
    real(stnd), intent(out), optional  :: d_e(:,:)
!
    logical(lgl), intent(in)            :: small
    logical(lgl), intent(out)           :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: safmin, eps, smlnum, bignum, rmin, rmax,   &
                  matnorm, sigma
    real(stnd), dimension(:), allocatable  :: d, e
!
    integer(i4b) :: m, n, j, p
    integer      :: iok
!
    logical(lgl) :: mat_is_scaled, store_q
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='select_eigval_cmp2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
    p =  size( matp )
!
!   COMPUTE THE SIZE OF THE ORIGINAL MATRIX mat .
!
    do n = 1_i4b, p
!
        j = n*(n+1_i4b)/2_i4b
!
        if ( j==p ) exit
!
    end do
!
    store_q = present( d_e )
!
    if ( store_q ) then
!
        call assert( logical( n==size(d_e,1), lgl ),         &
                     logical( 2==size(d_e,2), lgl ),         &
                     name_proc )
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            if ( store_q ) then
                d_e(1_i4b,1_i4b) = matp(1_i4b)
            end if
!
!           ALLOCATE POINTER PARAMETER.
!
            allocate( eigval(1_i4b), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
            eigval(1_i4b) = matp(1_i4b)
            matp(1_i4b)   = machhuge
!
        end if
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = maxval( abs(matp) )
!
!   SCALE MATRIX PACKED IN LINEAR ARRAY matp TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl(  matp, one, sigma )
!
!   ALLOCATE WORK VARIABLES.
!
    allocate( d(n), e(n), stat = iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   CALL symtrid_cmp_d TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_d( matp, d, e, store_q )
!
    if ( store_q ) then
!
!       STORE INTERMEDIATE TRIDIAGONAL FORM OF mat , IF REQUIRED.
!
        d_e(:n,1) = d(:n)
        d_e(:n,2) = e(:n)
!
    end if
!
!   STORE THE ELEMENTS OF THE TRIDIAGONAL MATRIX IN REVERSED ORDER IF
!   THE DIAGONAL ELEMENTS ARE INCREASING ALONG THE DIAGONAL.
!
    if ( reverse(d) ) then
!
        d(1_i4b:n) = d(n:1_i4b:-1_i4b)
        e(1_i4b:n-1_i4b) = e(n-1_i4b:1_i4b:-1_i4b)
!
    end if
!
!   CALL symtrid_ratqri2 TO COMPUTE SELECTED EIGENVALUES.
!
    call symtrid_ratqri2( d, e, val, failure, m, small )
!
!   ALLOCATE POINTER PARAMETER.
!
    allocate( eigval(m), stat = iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   IF MATRIX WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) then 
!
        eigval(:m) = ( one/sigma )*d(:m)
!
    else
!
        eigval(:m) = d(:m)
!
    end if
!
!   DEALLOCATE WORK VARIABLES.
!
    deallocate( d, e )
!
!
! END OF SUBROUTINE select_eigval_cmp2_d
! ______________________________________
!
    end subroutine select_eigval_cmp2_d
!
! =========================================================================================
!
    subroutine select_eigval_cmp3_a( mat, neig, eigval, small, failure, upper,           &
                                     sort, vector, scaling, init, abstol, le, theta, d_e )
!
! Purpose
! _______
!
!   SELECT_EIGVAL_CMP3 computes the largest or smallest eigenvalues of
!   a n-by-n real symmetric matrix MAT.
!
!   The matrix MAT is first transformed to symmetric tridiagonal form T
!   by an orthogonal similarity transformation:
!
!                        Q' * MAT * Q = T
!
!   ,then the eigenvalues are computed by a bisection method.
!   
!
! Arguments
! _________
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the symmetric matrix MAT.
!
!                If:
!
!                - UPPER = true:
!                  the leading n-by-n upper triangular part of MAT contains the
!                  upper triangular part of the matrix MAT.
!                - UPPER = false:
!                  the leading n-by-n lower triangular part of MAT contains
!                  the lower triangular part of the matrix MAT.
!
!                On exit:
!
!                - If UPPER = true and D_E is present :
!                  The leading n-by-n upper triangular part of MAT is overwritten
!                  by the matrix Q as a product of elementary reflectors.
!                - If UPPER = false and D_E is present :
!                  The leading n-by-n lower triangular part of MAT is overwritten
!                  by the matrix Q as a product of elementary reflectors.
!                - If UPPER = true and D_E is absent :
!                  The leading n-by-n upper triangular part of MAT is destroyed.
!                - If UPPER = false and D_E is absent :
!                  The leading n-by-n lower triangular part of MAT is destroyed.
!
!                The shape of MAT must verify: size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   NEIG         (OUTPUT) integer(i4b)
!                On output, NEIG specifies the number of eigenvalues which have
!                been computed. Note that NEIG may be greater than the optional argument
!                LE, if multiple eigenvalues at index LE make unique selection
!                impossible.
!
!                If none of the optional arguments LE and THETA are used, NEIG is set to n
!                and all the eigenvalues of MAT are computed.
!
!   EIGVAL       (OUTPUT) real(stnd), dimension(:)
!                On exit, EIGVAL(1:NEIG) contains the first NEIG largest or smallest eigenvalues
!                of MAT.
!                The other values in EIGVAL (e.g.  EIGVAL(NEIG+1:) ) are flagged  by a quiet NAN.
!
!                The size of EIGVAL must verify: size( EIGVAL ) = size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   SMALL        (INPUT) logical(lgl)
!                On entry:
!
!                - SMALL = false :  indicates that the largest eigenvalues are desired.
!                - SMALL = true  :  indicates that the smallest eigenvalues are desired.
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit and the bisection algorithm converged for all
!                  the computed eigenvalues to the desired accuracy ;
!                - FAILURE = true  :  indicates that some or all of the eigenvalues failed to converge
!                  or were not computed. This is generally caused by unexpectedly inaccurate
!                  arithmetic.
!
!   UPPER        (INPUT) logical(lgl)
!                Specifies whether the upper or lower triangular part of the
!                symmetric matrix MAT is stored. If:
!
!                - UPPER = true :  Upper triangular is stored ;
!                - UPPER = false:  Lower triangular is stored .
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'. For other values of SORT
!                nothing is done and EIGVAL(:NEIG) may not be sorted.
!
!   VECTOR       (INPUT, OPTIONAL) logical(lgl)
!                On entry, if VECTOR is set to true, a vectorized version of the bisection algorithm
!                is used to find the eigenvalues of the intermediate tridiagonal form T of MAT.
!
!                The default is VECTOR=false.
!
!   SCALING      (INPUT, OPTIONAL) logical(lgl)
!                On entry, if SCALING=true the intermediate tridiagonal matrix T is scaled
!                before computing the eigenvalues.
!
!                The default is to scale the tridiagonal matrix.
!
!   INIT         (INPUT, OPTIONAL) logical(lgl)
!                On entry, if INIT=true the initial intervals for the bisection steps for computing
!                the eigenvalues of the intermediate tridiagonal matrix T are estimated
!                from the eigenvalues of the intermediate tridiagonal matrix obtained from 
!                the Pal-Walker-Kahan algorithm.
!
!                The default is not to use the Pal-Walker-Kahan algorithm.
!
!   ABSTOL       (INPUT, OPTIONAL) real(stnd)
!                On entry, the absolute tolerance for the eigenvalues. An eigenvalue
!                (or cluster) is considered to be located if it has been determined to lie in
!                an interval whose width is ABSTOL or less.
!
!                If ABSTOL is less than or equal to zero, or is not specified, then ULP * | T | will be used,
!                where | T | means the 1-norm of T (T is the intermediate tridiagonal form of MAT) and ULP is
!                the machine precision (distance from 1 to the next larger floating point number).
!
!                Eigenvalues will be computed most accurately when ABSTOL is set to the square root of
!                the underflow threshold, sqrt(LAMCH('S')), not zero. 
!
!   LE           (INPUT, OPTIONAL) integer(i4b)
!                On entry, LE specifies the number of eigenvalues which must be
!                computed by the subroutine. However, on output, NEIG may be different than LE
!                if multiple eigenvalues at index LE make unique selection
!                impossible.
!
!                If:
!
!                - SMALL=false, the subroutine computes the LE largest eigenvalues of MAT,
!                - SMALL=true, the subroutine computes the LE smallest eigenvalues of MAT.
!
!                Only one of the optional arguments LE and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!                LE must be greater than 0 and less or equal to size( EIGVAL ) .
!
!                The default is LE = size( EIGVAL ), e.g. all the eigenvalues are computed.
!
!   THETA        (INPUT, OPTIONAL) real(stnd)
!                On entry:
!
!                - if SMALL=false, THETA specifies that the eigenvalues which are greater or equal to
!                  THETA must be computed. If none of the eigenvalues are greater or equal to
!                  THETA, NEIG is set to zero and EIGVAL(:) to a quiet NAN.
!                - if SMALL=true, THETA specifies that the eigenvalues which are less or equal to
!                  THETA must be computed. If none of the eigenvalues are smaller or equal to
!                  THETA, NEIG is set to zero and EIGVAL(:) to a quiet NAN.
!
!                Only one of the optional arguments LS and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!   D_E          (OUTPUT, OPTIONAL) real(stnd), dimension(:,2)
!                On exit, the first column of D_E contains the n diagonal elements
!                of the intermediate tridiagonal form T of MAT. The n-1 first
!                elements of the second column of D_E contains the n-1 subdiagonal
!                elements of the intermediate tridiagonal form T of MAT. D_E(n,2)
!                is arbitrary.
!
!                The shape of D_E must verify: size( D_E, 1 ) = n and size( D_E, 2 ) = 2
!
!
! Further Details
! _______________
!
!   This driver subroutine is adapted from the routine  DSYEV in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : eig_error3
    use Num_Constants,   only : lamch, machhuge, nan
    use Utilities,       only : assert, assert_eq, lascl, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout)          :: mat(:,:)
    real(stnd), intent(out)            :: eigval(:)
    real(stnd), intent(in), optional   :: abstol, theta
    real(stnd), intent(out), optional  :: d_e(:,:)
!
    integer(i4b), intent(out)          :: neig
    integer(i4b), intent(in), optional :: le
!
    logical(lgl), intent(in)            :: small, upper
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in), optional  :: vector, scaling, init
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: safmin, eps, smlnum, bignum, rmin, rmax,   &
                  matnorm, sigma
    real(stnd), dimension(size(mat,1))  :: d, e
!
    integer(i4b) :: n, j
!
    logical(lgl) :: mat_is_scaled, store_q
!
    character :: chr
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='select_eigval_cmp3'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
    sigma   = -machhuge
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(eigval),i4b) ,  &
                    name_proc )
!
    store_q = present( d_e )
    if ( store_q ) then
        call assert( logical( n==size(d_e,1), lgl ),         &
                     logical( 2==size(d_e,2), lgl ),         &
                     name_proc )
    end if
!
    if ( present(le) ) then
!
        call assert( logical( le<=n, lgl ),     &
                     logical( 1_i4b<=le, lgl ), &
                     name_proc )
!
        if ( present(theta) ) then
            call merror( name_proc//eig_error3 )
        end if
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            if ( store_q ) then
                d_e(1_i4b,1_i4b) = mat(1_i4b,1_i4b)
            end if
!
            if ( small ) then
                eps = -mat(1_i4b,1_i4b)
                if ( present( theta ) ) sigma = -theta
            else
                eps = mat(1_i4b,1_i4b)
                if ( present( theta ) ) sigma = theta
            end if
!
            if ( sigma<=eps ) then 
                neig          = 1_i4b
                eigval(1_i4b) = mat(1_i4b,1_i4b)
            else
                neig          = 0_i4b
                eigval(1_i4b) = nan()
            end if
!
            mat(1_i4b,1_i4b)  = machhuge
!
        else
            neig      = 0_i4b
        end if
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm =zero
!
    if ( upper ) then
!
        chr = "u"
!
!$OMP PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)   &
!$OMP            ,SCHEDULE(STATIC,1)                   &
!$OMP            ,PRIVATE(j,sigma)                     &
!$OMP            ,REDUCTION(max:matnorm)
!
        do j = 1_i4b, n
!
            sigma = maxval( abs(mat(:j,j)) )
            matnorm = max( matnorm, sigma )
!
        end do
!
!$OMP END PARALLEL DO
!
    else
!
        chr = "l"
!
!$OMP PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)   &
!$OMP            ,SCHEDULE(STATIC,1)                   &
!$OMP            ,PRIVATE(j,sigma)                     &
!$OMP            ,REDUCTION(max:matnorm)
!
        do j = 1_i4b, n
!
            sigma = maxval( abs(mat(j:,j)) )
            matnorm = max( matnorm, sigma )
!
        end do
!
!$OMP END PARALLEL DO
!
    end if
!
!   SCALE MATRIX TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl(  mat, one, sigma, chr )
!
!   CALL symtrid_cmp_a TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_a( mat, d, e, store_q, upper )
!
    if ( store_q ) then
!
!       STORE INTERMEDIATE TRIDIAGONAL FORM OF mat , IF REQUIRED.
!
        d_e(:n,1) = d(:n)
        d_e(:n,2) = e(:n)
!
    end if
!
!   CALL symtrid_bisect TO COMPUTE SELECTED EIGENVALUES.
!
    call symtrid_bisect( d, e, neig, eigval, failure, small=small, sort=sort, &
                         vector=vector, abstol=abstol, le=le, theta=theta,    &
                         scaling=scaling, init=init                           )
!
!   IF MATRIX WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) then 
!
        eigval(:neig) = ( one/sigma )*eigval(:neig)
!
    else
!
        eigval(:neig) = eigval(:neig)
!
    end if
!
!
! END OF SUBROUTINE select_eigval_cmp3_a
! ______________________________________
!
    end subroutine select_eigval_cmp3_a
!
! =========================================================================================
!
    subroutine select_eigval_cmp3_b( mat, neig, eigval, small, failure,                  &
                                     sort, vector, scaling, init, abstol, le, theta, d_e )
!
! Purpose
! _______
!
!   SELECT_EIGVAL_CMP3 computes the largest or smallest eigenvalues of
!   a n-by-n real symmetric matrix MAT.
!
!   The matrix MAT is first transformed to symmetric tridiagonal form T
!   by an orthogonal similarity transformation:
!
!                        Q' * MAT * Q = T
!
!   ,then the eigenvalues are computed by a bisection method.
!   
!
! Arguments
! _________
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the symmetric matrix MAT. The leading 
!                n-by-n upper triangular part of MAT contains the
!                upper triangular part of the matrix MAT. The strictly
!                n-by-n lower triangular part of MAT is not referenced.
!
!                On exit:
!
!                - If D_E is present:
!                  The leading n-by-n upper triangular part of MAT is overwritten
!                  by the matrix Q as a product of elementary reflectors.
!                - If D_E is absent :
!                  The leading n-by-n upper triangular part of MAT is destroyed.
!
!                The shape of MAT must verify: size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   NEIG         (OUTPUT) integer(i4b)
!                On output, NEIG specifies the number of eigenvalues which have
!                been computed. Note that NEIG may be greater than the optional argument
!                LE, if multiple eigenvalues at index LE make unique selection
!                impossible.
!
!                If none of the optional arguments LE and THETA are used, NEIG is set to n
!                and all the eigenvalues of MAT are computed.
!
!   EIGVAL       (OUTPUT) real(stnd), dimension(:)
!                On exit, EIGVAL(1:NEIG) contains the first NEIG largest or smallest eigenvalues
!                of MAT.
!                The other values in EIGVAL (e.g.  EIGVAL(NEIG+1:) ) are flagged  by a quiet NAN.
!
!                The size of EIGVAL must verify: size( EIGVAL ) = size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   SMALL        (INPUT) logical(lgl)
!                On entry:
!
!                - SMALL = false :  indicates that the m largest eigenvalues are desired.
!                - SMALL = true  :  indicates that the m smallest eigenvalues are desired.
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit and the bisection algorithm converged for all
!                  the computed eigenvalues to the desired accuracy ;
!                - FAILURE = true  :  indicates that some or all of the eigenvalues failed to converge
!                  or were not computed. This is generally caused by unexpectedly inaccurate
!                  arithmetic.
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'. For other values of SORT
!                nothing is done and EIGVAL(:NEIG) may not be sorted.
!
!   VECTOR       (INPUT, OPTIONAL) logical(lgl)
!                On entry, if VECTOR is set to true, a vectorized version of the bisection algorithm
!                is used to find the eigenvalues of the intermediate tridiagonal form T of MAT.
!
!                The default is VECTOR=false.
!
!   SCALING      (INPUT, OPTIONAL) logical(lgl)
!                On entry, if SCALING=true the intermediate tridiagonal matrix T is scaled
!                before computing the eigenvalues.
!
!                The default is to scale the tridiagonal matrix.
!
!   INIT         (INPUT, OPTIONAL) logical(lgl)
!                On entry, if INIT=true the initial intervals for the bisection steps for computing
!                the eigenvalues of the intermediate tridiagonal matrix T are estimated
!                from the eigenvalues of the intermediate tridiagonal matrix obtained from 
!                the Pal-Walker-Kahan algorithm.
!
!                The default is not to use the Pal-Walker-Kahan algorithm.
!
!   ABSTOL       (INPUT, OPTIONAL) real(stnd)
!                On entry, the absolute tolerance for the eigenvalues. An eigenvalue
!                (or cluster) is considered to be located if it has been determined to lie in
!                an interval whose width is ABSTOL or less.
!
!                If ABSTOL is less than or equal to zero, or is not specified, then ULP * | T | will be used,
!                where | T | means the 1-norm of T (T is the intermediate tridiagonal form of MAT) and ULP is
!                the machine precision (distance from 1 to the next larger floating point number).
!
!                Eigenvalues will be computed most accurately when ABSTOL is set to the square root of
!                the underflow threshold, sqrt(LAMCH('S')), not zero. 
!
!   LE           (INPUT, OPTIONAL) integer(i4b)
!                On entry, LE specifies the number of eigenvalues which must be
!                computed by the subroutine. However, on output, NEIG may be different than LE
!                if multiple eigenvalues at index LE make unique selection
!                impossible.
!
!                If:
!
!                - SMALL=false, the subroutine computes the LE largest eigenvalues of MAT,
!                - SMALL=true, the subroutine computes the LE smallest eigenvalues of MAT.
!
!                Only one of the optional arguments LE and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!                LE must be greater than 0 and less or equal to size( EIGVAL ) .
!
!                The default is LE = size( EIGVAL ), e.g. all the eigenvalues are computed.
!
!   THETA        (INPUT, OPTIONAL) real(stnd)
!                On entry,
!
!                - if SMALL=false, THETA specifies that the eigenvalues which are greater or equal to
!                  THETA must be computed. If none of the eigenvalues are greater or equal to
!                  THETA, NEIG is set to zero and EIGVAL(:) to a quiet NAN.
!                - if SMALL=true, THETA specifies that the eigenvalues which are less or equal to
!                  THETA must be computed. If none of the eigenvalues are smaller or equal to
!                  THETA, NEIG is set to zero and EIGVAL(:) to a quiet NAN.
!
!                Only one of the optional arguments LS and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!   D_E          (OUTPUT, OPTIONAL) real(stnd), dimension(:,2)
!                On exit, the first column of D_E contains the n diagonal elements
!                of the intermediate tridiagonal form T of MAT. The n-1 first
!                elements of the second column of D_E contains the n-1 subdiagonal
!                elements of the intermediate tridiagonal form T of MAT. D_E(n,2)
!                is arbitrary.
!
!                The shape of D_E must verify: size( D_E, 1 ) = n and size( D_E, 2 ) = 2
!
!
! Further Details
! _______________
!
!   This driver subroutine is adapted from the routine  DSYEV in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : eig_error3
    use Num_Constants,   only : lamch, machhuge, nan
    use Utilities,       only : assert, assert_eq, lascl, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout)          :: mat(:,:)
    real(stnd), intent(out)            :: eigval(:)
    real(stnd), intent(in), optional   :: abstol, theta
    real(stnd), intent(out), optional  :: d_e(:,:)
!
    integer(i4b), intent(out)          :: neig
    integer(i4b), intent(in), optional :: le
!
    logical(lgl), intent(in)            :: small
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in), optional  :: vector, scaling, init
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: safmin, eps, smlnum, bignum, rmin, rmax,   &
                  matnorm, sigma
    real(stnd), dimension(size(mat,1))  :: d, e
!
    integer(i4b) :: n, j
!
    logical(lgl) :: mat_is_scaled, store_q
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='select_eigval_cmp3'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
    sigma   = -machhuge
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(eigval),i4b) ,  &
                    name_proc )
!
    store_q = present( d_e )
    if ( store_q ) then
        call assert( logical( n==size(d_e,1), lgl ),         &
                     logical( 2==size(d_e,2), lgl ),         &
                     name_proc )
    end if
!
    if ( present(le) ) then
!
        call assert( logical( le<=n, lgl ),     &
                     logical( 1_i4b<=le, lgl ), &
                     name_proc )
!
        if ( present(theta) ) then
            call merror( name_proc//eig_error3 )
        end if
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            if ( store_q ) then
                d_e(1_i4b,1_i4b) = mat(1_i4b,1_i4b)
            end if
!
            if ( small ) then
                eps = -mat(1_i4b,1_i4b)
                if ( present( theta ) ) sigma = -theta
            else
                eps = mat(1_i4b,1_i4b)
                if ( present( theta ) ) sigma = theta
            end if
!
            if ( sigma<=eps ) then 
                neig          = 1_i4b
                eigval(1_i4b) = mat(1_i4b,1_i4b)
            else
                neig          = 0_i4b
                eigval(1_i4b) = nan()
            end if
!
            mat(1_i4b,1_i4b)  = machhuge
!
        else
            neig      = 0_i4b
        end if
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = zero
!
!$OMP PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit)   &
!$OMP            ,SCHEDULE(STATIC,1)                   &
!$OMP            ,PRIVATE(j,sigma)                     &
!$OMP            ,REDUCTION(max:matnorm)
!
    do j = 1_i4b, n
!
        sigma = maxval( abs(mat(:j,j)) )
        matnorm = max( matnorm, sigma )
!
    end do
!
!$OMP END PARALLEL DO
!
!
!   SCALE MATRIX TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl( mat, one, sigma )
!
!   CALL symtrid_cmp_b TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_b( mat, d, e, store_q )
!
    if ( store_q ) then
!
!       STORE INTERMEDIATE TRIDIAGONAL FORM OF mat , IF REQUIRED.
!
        d_e(:n,1) = d(:n)
        d_e(:n,2) = e(:n)
!
    end if
!
!   CALL symtrid_bisect TO COMPUTE SELECTED EIGENVALUES.
!
    call symtrid_bisect( d, e, neig, eigval, failure, small=small, sort=sort, &
                         vector=vector, abstol=abstol, le=le, theta=theta,    &
                         scaling=scaling, init=init                           )
!
!   IF MATRIX WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) then 
!
        eigval(:neig) = ( one/sigma )*eigval(:neig)
!
    else
!
        eigval(:neig) = eigval(:neig)
!
    end if
!
!
! END OF SUBROUTINE select_eigval_cmp3_b
! ______________________________________
!
    end subroutine select_eigval_cmp3_b
!
! =========================================================================================
!
    subroutine select_eigval_cmp3_c( matp, neig, eigval, small, failure, upper,           &
                                     sort, vector, scaling, init, abstol, le, theta, d_e  )
!
! Purpose
! _______
!
!   SELECT_EIGVAL_CMP3 computes the largest or smallest eigenvalues of
!   a n-by-n real symmetric matrix mat stored in packed form in a linear array MATP.
!
!   The matrix mat is first transformed to symmetric tridiagonal form T
!   by an orthogonal similarity transformation:
!
!                        Q' * mat * Q = T
!
!   ,then the eigenvalues are computed by a bisection method.
!   
!
! Arguments
! _________
!
!   MATP         (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the upper or lower triangle of the symmetric matrix
!                mat, packed column-wise in a linear array.  The j-th column of mat
!                is stored in the array MATP as follows:
!
!                - if UPPER = true,  MATP(i + (j-1) * j/2) = mat(i,j) for 1<=i<=j;
!                - if UPPER = false, MATP(i + (j-1) * (2 * n-j)/2) = mat(i,j) for j<=i<=n.
!
!                On exit:
!
!                - If D_E is present :
!                  MATP is overwritten by the matrix Q as a product
!                  of elementary reflectors.
!                - If D_E is absent :
!                  MATP is destroyed.
!
!                The size of MATP must verify:   size( MATP ) = (n * (n+1)/2).
!
!   NEIG         (OUTPUT) integer(i4b)
!                On output, NEIG specifies the number of eigenvalues which have
!                been computed. Note that NEIG may be greater than the optional argument
!                LE, if multiple eigenvalues at index LE make unique selection
!                impossible.
!
!                If none of the optional arguments LE and THETA are used, NEIG is set to n
!                and all the eigenvalues of MAT are computed.
!
!   EIGVAL       (OUTPUT) real(stnd), dimension(:)
!                On exit, EIGVAL(1:NEIG) contains the first NEIG largest or smallest eigenvalues
!                of MAT.
!                The other values in EIGVAL (e.g.  EIGVAL(NEIG+1:) ) are flagged  by a quiet NAN.
!
!                The size of EIGVAL must verify: size( EIGVAL ) =  n .
!
!   SMALL        (INPUT) logical(lgl)
!                On entry:
!
!                - SMALL = false :  indicates that the m largest eigenvalues are desired.
!                - SMALL = true  :  indicates that the m smallest eigenvalues are desired.
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit and the bisection algorithm converged for all
!                  the computed eigenvalues to the desired accuracy ;
!                - FAILURE = true  :  indicates that some or all of the eigenvalues failed to converge
!                  or were not computed. This is generally caused by unexpectedly inaccurate
!                  arithmetic.
!
!   UPPER        (INPUT) logical(lgl)
!                Specifies whether the upper or lower triangular part of the
!                symmetric matrix mat is stored in the linear array MATP. If:
!
!                - UPPER = true :  Upper triangle of mat is stored;
!                - UPPER = false:  Lower triangle of mat is stored.
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'. For other values of SORT
!                nothing is done and EIGVAL(:NEIG) may not be sorted.
!
!   VECTOR       (INPUT, OPTIONAL) logical(lgl)
!                On entry, if VECTOR is set to true, a vectorized version of the bisection algorithm
!                is used to find the eigenvalues of the intermediate tridiagonal form T of MAT.
!
!                The default is VECTOR=false.
!
!   SCALING      (INPUT, OPTIONAL) logical(lgl)
!                On entry, if SCALING=true the intermediate tridiagonal matrix T is scaled
!                before computing the eigenvalues.
!
!                The default is to scale the tridiagonal matrix.
!
!   INIT         (INPUT, OPTIONAL) logical(lgl)
!                On entry, if INIT=true the initial intervals for the bisection steps for computing
!                the eigenvalues of the intermediate tridiagonal matrix T are estimated
!                from the eigenvalues of the intermediate tridiagonal matrix obtained from 
!                the Pal-Walker-Kahan algorithm.
!
!                The default is not to use the Pal-Walker-Kahan algorithm.
!
!   ABSTOL       (INPUT, OPTIONAL) real(stnd)
!                On entry, the absolute tolerance for the eigenvalues. An eigenvalue
!                (or cluster) is considered to be located if it has been determined to lie in
!                an interval whose width is ABSTOL or less.
!
!                If ABSTOL is less than or equal to zero, or is not specified, then ULP * | T | will be used,
!                where | T | means the 1-norm of T (T is the intermediate tridiagonal form of mat) and ULP is
!                the machine precision (distance from 1 to the next larger floating point number).
!
!                Eigenvalues will be computed most accurately when ABSTOL is set to the square root of
!                the underflow threshold, sqrt(LAMCH('S')), not zero. 
!
!   LE           (INPUT, OPTIONAL) integer(i4b)
!                On entry, LE specifies the number of eigenvalues which must be
!                computed by the subroutine. However, on output, NEIG may be different than LE
!                if multiple eigenvalues at index LE make unique selection
!                impossible.
!
!                If:
!
!                - SMALL=false, the subroutine computes the LE largest eigenvalues of MAT,
!                - SMALL=true, the subroutine computes the LE smallest eigenvalues of MAT.
!
!                Only one of the optional arguments LE and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!                LE must be greater than 0 and less or equal to size( EIGVAL ) .
!
!                The default is LE = size( EIGVAL ), e.g. all the eigenvalues are computed.
!
!   THETA        (INPUT, OPTIONAL) real(stnd)
!                On entry:
!
!                - if SMALL=false, THETA specifies that the eigenvalues which are greater or equal to
!                  THETA must be computed. If none of the eigenvalues are greater or equal to
!                  THETA, NEIG is set to zero and EIGVAL(:) to a quiet NAN.
!                - if SMALL=true, THETA specifies that the eigenvalues which are less or equal to
!                  THETA must be computed. If none of the eigenvalues are smaller or equal to
!                  THETA, NEIG is set to zero and EIGVAL(:) to a quiet NAN.
!
!                Only one of the optional arguments LS and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!   D_E          (OUTPUT, OPTIONAL) real(stnd), dimension(:,2)
!                On exit, the first column of D_E contains the n diagonal elements
!                of the intermediate tridiagonal form T of mat. The n-1 first
!                elements of the second column of D_E contains the n-1 subdiagonal
!                elements of the intermediate tridiagonal form T of mat. D_E(n,2)
!                is arbitrary.
!
!                The shape of D_E must verify: size( D_E, 1 ) = n and size( D_E, 2 ) = 2
!
!
! Further Details
! _______________
!
!   This driver subroutine is adapted from the routine  DSYEV in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch, machhuge, nan
    use Char_Constants,  only : eig_error3
    use Utilities,       only : assert, lascl, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout)          :: matp(:)
    real(stnd), intent(out)            :: eigval(:)
    real(stnd), intent(in), optional   :: abstol, theta
    real(stnd), intent(out), optional  :: d_e(:,:)
!
    integer(i4b), intent(out)          :: neig
    integer(i4b), intent(in), optional :: le
!
    logical(lgl), intent(in)            :: small, upper
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in), optional  :: vector, scaling, init
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: safmin, eps, smlnum, bignum, rmin, rmax,   &
                  matnorm, sigma
    real(stnd), dimension(size(eigval))  :: d, e
!
    integer(i4b) :: n, j, p
!
    logical(lgl) :: mat_is_scaled, store_q
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='select_eigval_cmp3'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
    sigma   = -machhuge
!
!   TEST THE ARGUMENTS.
!
    n =  size( eigval )
    p =  size( matp )
!
    j = n*(n+1_i4b)/2_i4b
!
    call assert( logical(j==p,lgl), name_proc )
!
    store_q = present( d_e )
!
    if ( store_q ) then
!
        call assert( logical( n==size(d_e,1), lgl ),  &
                     logical( 2==size(d_e,2), lgl ),  &
                     name_proc )
!
    end if
!
    if ( present(le) ) then
!
        call assert( logical( le<=n, lgl ),     &
                     logical( 1_i4b<=le, lgl ), &
                     name_proc )
!
        if ( present(theta) ) then
            call merror( name_proc//eig_error3 )
        end if
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            if ( store_q ) then
                d_e(1_i4b,1_i4b) = matp(1_i4b)
            end if
!
            if ( small ) then
                eps = -matp(1_i4b)
                if ( present( theta ) ) sigma = -theta
            else
                eps = matp(1_i4b)
                if ( present( theta ) ) sigma = theta
            end if
!
            if ( sigma<=eps ) then 
                neig          = 1_i4b
                eigval(1_i4b) = matp(1_i4b)
            else
                neig          = 0_i4b
                eigval(1_i4b) = nan()
            end if
!
            matp(1_i4b)  = machhuge
!
        else
            neig      = 0_i4b
        end if
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = maxval( abs(matp) )
!
!   SCALE MATRIX PACKED IN LINEAR ARRAY matp TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl(  matp, one, sigma )
!
!   CALL symtrid_cmp_c TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_c( matp, d, e, store_q, upper )
!
    if ( store_q ) then
!
!       STORE INTERMEDIATE TRIDIAGONAL FORM OF mat , IF REQUIRED.
!
        d_e(:n,1) = d(:n)
        d_e(:n,2) = e(:n)
!
    end if
!
!   CALL symtrid_bisect TO COMPUTE SELECTED EIGENVALUES.
!
    call symtrid_bisect( d, e, neig, eigval, failure, small=small, sort=sort, &
                         vector=vector, abstol=abstol, le=le, theta=theta,    &
                         scaling=scaling, init=init                           )
!
!   IF MATRIX WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) then 
!
        eigval(:neig) = ( one/sigma )*eigval(:neig)
!
    else
!
        eigval(:neig) = eigval(:neig)
!
    end if
!
!
! END OF SUBROUTINE select_eigval_cmp3_c
! ______________________________________
!
    end subroutine select_eigval_cmp3_c
!
! =========================================================================================
!
    subroutine select_eigval_cmp3_d( matp, neig, eigval, small, failure,                  &
                                     sort, vector, scaling, init, abstol, le, theta, d_e  )
!
! Purpose
! _______
!
!   SELECT_EIGVAL_CMP3 computes the largest or smallest eigenvalues of
!   a n-by-n real symmetric matrix mat stored in packed form in a linear array MATP.
!
!   The matrix mat is first transformed to symmetric tridiagonal form T
!   by an orthogonal similarity transformation:
!
!                        Q' * mat * Q = T
!
!   ,then the eigenvalues are computed by a bisection method.
!   
!
! Arguments
! _________
!
!   MATP         (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the upper triangle of the symmetric matrix mat,
!                packed column-wise in a linear array.  The j-th column of mat
!                is stored in the array MATP as follows:
!
!                         MATP(i + (j-1) * j/2) = mat(i,j) for 1<=i<=j;
!
!                On exit:
!
!                - If D_E is present :
!                  MATP is overwritten by the matrix Q as a product
!                  of elementary reflectors.
!                - If D_E is absent :
!                  MATP is destroyed.
!
!                The size of MATP must verify:   size( MATP ) = (n * (n+1)/2).
!
!   NEIG         (OUTPUT) integer(i4b)
!                On output, NEIG specifies the number of eigenvalues which have
!                been computed. Note that NEIG may be greater than the optional argument
!                LE, if multiple eigenvalues at index LE make unique selection
!                impossible.
!
!                If none of the optional arguments LE and THETA are used, NEIG is set to n
!                and all the eigenvalues of MAT are computed.
!
!   EIGVAL       (OUTPUT) real(stnd), dimension(:)
!                On exit, EIGVAL(1:NEIG) contains the first NEIG largest or smallest eigenvalues
!                of MAT.
!                The other values in EIGVAL (e.g. EIGVAL(NEIG+1:) ) are flagged  by a quiet NAN.
!
!                The size of EIGVAL must verify: size( EIGVAL ) =  n .
!
!   SMALL        (INPUT) logical(lgl)
!                On entry:
!
!                - SMALL = false :  indicates that the m largest eigenvalues are desired.
!                - SMALL = true  :  indicates that the m smallest eigenvalues are desired.
!
!   FAILURE      (OUTPUT) logical(lgl)
!                On exit:
!
!                - FAILURE = false :  indicates successful exit and the bisection algorithm converged for all
!                  the computed eigenvalues to the desired accuracy ;
!                - FAILURE = true  :  indicates that some or all of the eigenvalues failed to converge
!                  or were not computed. This is generally caused by unexpectedly inaccurate
!                  arithmetic.
!
!   SORT         (INPUT, OPTIONAL) character
!                Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                or in descending order if SORT = 'D' or 'd'. For other values of SORT
!                nothing is done and EIGVAL(:NEIG) may not be sorted.
!
!   VECTOR       (INPUT, OPTIONAL) logical(lgl)
!                On entry, if VECTOR is set to true, a vectorized version of the bisection algorithm
!                is used to find the eigenvalues of the intermediate tridiagonal form T of MAT.
!
!                The default is VECTOR=false.
!
!   SCALING      (INPUT, OPTIONAL) logical(lgl)
!                On entry, if SCALING=true the intermediate tridiagonal matrix T is scaled
!                before computing the eigenvalues.
!
!                The default is to scale the tridiagonal matrix.
!
!   INIT         (INPUT, OPTIONAL) logical(lgl)
!                On entry, if INIT=true the initial intervals for the bisection steps for computing
!                the eigenvalues of the intermediate tridiagonal matrix T are estimated
!                from the eigenvalues of the intermediate tridiagonal matrix obtained from 
!                the Pal-Walker-Kahan algorithm.
!
!                The default is not to use the Pal-Walker-Kahan algorithm.
!
!   ABSTOL       (INPUT, OPTIONAL) real(stnd)
!                On entry, the absolute tolerance for the eigenvalues. An eigenvalue
!                (or cluster) is considered to be located if it has been determined to lie in
!                an interval whose width is ABSTOL or less.
!
!                If ABSTOL is less than or equal to zero, or is not specified, then ULP * | T | will be used,
!                where | T | means the 1-norm of T (T is the intermediate tridiagonal form of mat) and ULP is
!                the machine precision (distance from 1 to the next larger floating point number).
!
!                Eigenvalues will be computed most accurately when ABSTOL is set to the square root of
!                the underflow threshold, sqrt(LAMCH('S')), not zero. 
!
!   LE           (INPUT, OPTIONAL) integer(i4b)
!                On entry, LE specifies the number of eigenvalues which must be
!                computed by the subroutine. However, on output, NEIG may be different than LE
!                if multiple eigenvalues at index LE make unique selection
!                impossible.
!
!                If:
!
!                - SMALL=false, the subroutine computes the LE largest eigenvalues of MAT,
!                - SMALL=true, the subroutine computes the LE smallest eigenvalues of MAT.
!
!                Only one of the optional arguments LE and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!                LE must be greater than 0 and less or equal to size( EIGVAL ) .
!
!                The default is LE = size( EIGVAL ), e.g. all the eigenvalues are computed.
!
!   THETA        (INPUT, OPTIONAL) real(stnd)
!                On entry:
!
!                - if SMALL=false, THETA specifies that the eigenvalues which are greater or equal to
!                  THETA must be computed. If none of the eigenvalues are greater or equal to
!                  THETA, NEIG is set to zero and EIGVAL(:) to a quiet NAN.
!                - if SMALL=true, THETA specifies that the eigenvalues which are less or equal to
!                  THETA must be computed. If none of the eigenvalues are smaller or equal to
!                  THETA, NEIG is set to zero and EIGVAL(:) to a quiet NAN.
!
!                Only one of the optional arguments LS and THETA must be specified, otherwise
!                the subroutine will stop with an error message.
!
!   D_E          (OUTPUT, OPTIONAL) real(stnd), dimension(:,2)
!                On exit, the first column of D_E contains the n diagonal elements
!                of the intermediate tridiagonal form T of mat. The n-1 first
!                elements of the second column of D_E contains the n-1 subdiagonal
!                elements of the intermediate tridiagonal form T of mat. D_E(n,2)
!                is arbitrary.
!
!                The shape of D_E must verify: size( D_E, 1 ) = n and size( D_E, 2 ) = 2
!
!
! Further Details
! _______________
!
!   This driver subroutine is adapted from the routine  DSYEV in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch, machhuge, nan
    use Char_Constants,  only : eig_error3
    use Utilities,       only : assert, lascl, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout)          :: matp(:)
    real(stnd), intent(out)            :: eigval(:)
    real(stnd), intent(in), optional   :: abstol, theta
    real(stnd), intent(out), optional  :: d_e(:,:)
!
    integer(i4b), intent(out)          :: neig
    integer(i4b), intent(in), optional :: le
!
    logical(lgl), intent(in)            :: small
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in), optional  :: vector, scaling, init
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: safmin, eps, smlnum, bignum, rmin, rmax,   &
                  matnorm, sigma
    real(stnd), dimension(size(eigval))  :: d, e
!
    integer(i4b) :: n, j, p
!
    logical(lgl) :: mat_is_scaled, store_q
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='select_eigval_cmp3'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
    sigma   = -machhuge
!
!   TEST THE ARGUMENTS.
!
    n =  size( eigval )
    p =  size( matp )
!
    j = n*(n+1_i4b)/2_i4b
!
    call assert( logical(j==p,lgl), name_proc )
!
    store_q = present( d_e )
!
    if ( store_q ) then
!
        call assert( logical( n==size(d_e,1), lgl ),   &
                     logical( 2==size(d_e,2), lgl ),   &
                     name_proc )
!
    end if
!
    if ( present(le) ) then
!
        call assert( logical( le<=n, lgl ),     &
                     logical( 1_i4b<=le, lgl ), &
                     name_proc )
!
        if ( present(theta) ) then
            call merror( name_proc//eig_error3 )
        end if
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            if ( store_q ) then
                d_e(1_i4b,1_i4b) = matp(1_i4b)
            end if
!
            if ( small ) then
                eps = -matp(1_i4b)
                if ( present( theta ) ) sigma = -theta
            else
                eps = matp(1_i4b)
                if ( present( theta ) ) sigma = theta
            end if
!
            if ( sigma<=eps ) then 
                neig          = 1_i4b
                eigval(1_i4b) = matp(1_i4b)
            else
                neig          = 0_i4b
                eigval(1_i4b) = nan()
            end if
!
            matp(1_i4b)  = machhuge
!
        else
            neig      = 0_i4b
        end if
!
        return
!
    end if
!
!   GET MACHINE CONSTANTS.
!
    safmin = lamch( 's' )
    eps    = lamch( 'p' )
    smlnum = safmin/eps
    bignum = one/smlnum
    rmin   = sqrt( smlnum )
    rmax   = sqrt( bignum )
!
!   COMPUTE MAX ELEMENT OF mat .
!
    matnorm = maxval( abs(matp) )
!
!   SCALE MATRIX PACKED IN LINEAR ARRAY matp TO ALLOWABLE RANGE, IF NECESSARY.
!
    mat_is_scaled = false
!
    if ( matnorm>zero .and. matnorm<rmin ) then
!
        mat_is_scaled = true
        sigma = rmin / matnorm
!
    else if ( matnorm>rmax ) then
!
        mat_is_scaled = true
        sigma = rmax / matnorm
!
    end if
!
    if ( mat_is_scaled ) call lascl(  matp, one, sigma )
!
!   CALL symtrid_cmp_d TO REDUCE SYMMETRIC MATRIX TO TRIDIAGONAL FORM.
!
    call symtrid_cmp_d( matp, d, e, store_q )
!
    if ( store_q ) then
!
!       STORE INTERMEDIATE TRIDIAGONAL FORM OF mat , IF REQUIRED.
!
        d_e(:n,1) = d(:n)
        d_e(:n,2) = e(:n)
!
    end if
!
!   CALL symtrid_bisect TO COMPUTE SELECTED EIGENVALUES.
!
    call symtrid_bisect( d, e, neig, eigval, failure, small=small, sort=sort, &
                         vector=vector, abstol=abstol, le=le, theta=theta,    &
                         scaling=scaling, init=init                           )
!
!   IF MATRIX WAS SCALED, THEN RESCALE EIGENVALUES APPROPRIATELY.
!
    if ( mat_is_scaled ) then 
!
        eigval(:neig) = ( one/sigma )*eigval(:neig)
!
    else
!
        eigval(:neig) = eigval(:neig)
!
    end if
!
!
! END OF SUBROUTINE select_eigval_cmp3_d
! ______________________________________
!
    end subroutine select_eigval_cmp3_d
!
!
! =========================================================================================
!                         RANDOMIZED PARTIAL SYMMETRIC EIGENVALUES DECOMPOSITION
! =========================================================================================
!
!
    subroutine reig_cmp( mat, eigval, eigvec, failure, niter, nover, ortho,  &
                         extd_samp, rng_alg, maxiter                         )
!
! Purpose
! _______
!
!   REIG_CMP computes approximations of the neig largest eigenvalues (in absolute magnitude) and
!   associated eigenvectors of a full n-by-n real symmetric matrix MAT using randomized power,
!   subspace or block Krylov iterations.
!
!   neig is the target rank of the partial EigenValue Decomposition (EVD), which is sought,
!   and is equal to the size of the output real vector argument EIGVAL, i.e., neig = size( EIGVAL ).
!
!   
! Arguments
! _________
!
!   MAT                 (INPUT) real(stnd), dimension(:,:)
!                       On entry, the n-by-n symmetric matrix MAT.
!
!                       MAT is not modified by the routine.
!
!   EIGVAL              (OUTPUT) real(stnd), dimension(:)
!                       On exit, EIGVAL(:) contains the first top neig eigenvalues of MAT.
!                       The eigenvalues are given in decreasing order of absolute magnitude.
!
!                       The size of EIGVAL must verify:
!
!                       - size( EIGVAL ) = neig <= size( MAT, 1 ) = size( MAT, 2 ) = n.
!
!   EIGVEC              (OUTPUT) real(stnd), dimension(:,:)
!                       On exit, the computed neig top eigenvectors. The eigenvector
!                       associated with the eigenvalue EIGVAL(j) is stored in the j-th column
!                       of EIGVEC.
!
!                       The shape of EIGVEC must verify:
!
!                       - size( EIGVEC, 1 ) = size( MAT, 1 ) = size( MAT, 2 ) = n,
!                       - size( EIGVEC, 2 ) = size( EIGVEC ) = neig .
!
!   FAILURE             (OUTPUT, OPTIONAL) logical(lgl)
!                       On exit, if the optional logical argument FAILURE is present, a test of the
!                       accuracy of the computed partial EVD is performed and in that
!                       case:
!
!                       - FAILURE = false : indicates successful exit;
!                       - FAILURE = true  : indicates that some of the computed eigenvalues
!                         and eigenvectors of MAT failed to converge in NITER iterations.
!
!                       If FAILURE = true on exit, results are still useful, but some of the approximated
!                       eigen couplets have a poor accuracy.
!
!   NITER               (INPUT, OPTIONAL) integer(i4b)
!                       The number of randomized power, subspace or block Krylov iterations performed
!                       in the subroutine for computing the top neig eigen triplets. NITER must be
!                       positive or null.
!
!                       By default, 10 randomized power, subspace or block Krylov iterations are performed.
!
!   NOVER               (INPUT, OPTIONAL) integer(i4b)
!                       The oversampling size used in the randomized power subspace or block Krylov
!                       iterations for computing the top neig eigen triplets.
!
!                       NOVER must be positive or null and verifies the relationship:
!
!                       - NOVER + size( EIGVAL ) <= size( MAT, 1 ) = size( MAT, 2 ) = n
!
!                       and is adjusted if necessary to verify this relationship in all cases.
!
!                       See Further Details and the cited references for the meaning and usefulness
!                       of the oversampling size in the randomized power, subspace or block Krylov iterations.
!
!                       By default, the oversampling size is set to 10.
!
!   ORTHO               (INPUT, OPTIONAL) logical(lgl)
!                       On entry, if:
!
!                       - ORTHO=true, orthonormalization is carried out between each step of
!                         the power or block Krylov iterations to avoid loss of accuracy due
!                         to rounding errors. This means that subspace iterations are used
!                         instead of power iterations;
!                       - ORTHO=false, orthonormalization is not performed. 
!
!                       The default is to use orthonormalization, e.g., ORTHO=true.
!
!   EXTD_SAMP           (INPUT, OPTIONAL) logical(lgl)
!                       The optional argument EXTD_SAMP determines if extended sampling
!                       (e.g., block Krylov iterations) is used or not for computing
!                       the top neig eigen triplets.
!
!                       On entry, if:
!
!                       - EXTD_SAMP=true, block Krylov iterations are used;
!                       - EXTD_SAMP=false, power or subspace iterations are used. 
!
!                       The default is to use power or subspace iterations, e.g., EXTD_SAMP=false.
!
!   RNG_ALG             (INPUT, OPTIONAL) integer(i4b)
!                       On entry, a scalar integer to select the random (uniform) number generator used to build
!                       the random gaussian test matrix in the randomized EVD algorithm.
!
!                       The possible values are:
!
!                       - ALG=1  : selects the Marsaglia's KISS random number generator;
!                       - ALG=2  : selects the fast Marsaglia's KISS random number generator;
!                       - ALG=3  : selects the L'Ecuyer's LFSR113 random number generator;
!                       - ALG=4  : selects the Mersenne Twister random number generator;
!                       - ALG=5  : selects the maximally equidistributed Mersenne Twister random number generator;
!                       - ALG=6  : selects the extended precision of the Marsaglia's KISS random number generator;
!                       - ALG=7  : selects the extended precision of the fast Marsaglia's KISS random number generator;
!                       - ALG=8  : selects the extended precision of the L'Ecuyer's LFSR113 random number generator.
!                       - ALG=9  : selects the extended precision of Mersenne Twister random number generator;
!                       - ALG=10 : selects the extended precision of maximally equidistributed Mersenne Twister
!                         random number generator;
!
!                       For other values, the current random number generator and its current state are not changed.
!                       Note further, that, on exit, the current random number generator is not reset to its previous
!                       value before the call to REIG_CMP.
!
!                       See the documentation of subroutine RANDOM_SEED\ _ in module Random for further information.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the bidiagonal SVD phase
!                       of the SVD algorithm or in the QR phase of the EVD algorithm, which
!                       are used in the last phase of the randomized algorithm.
!
!                       See description of suboutines SVD_CMP and EIG_CMP for further details
!                       about this optional argument.
!
!
! Further Details
! _______________
!
!   For a good introduction to randomized linear algebra, see the references (1) and (2).
!
!   The randomized subspace iteration was proposed in (3; see Algorithm 4.4) to compute
!   an orthonormal matrix whose range approximates the range of MAT. An approximate
!   partial EVD decomposition can then be computed using the aforementioned orthonormal
!   matrix, see Algorithm 5.3 in (3).
!
!   The randomized block Krylov iterations for computing an approximate partial EVD
!   was proposed in (4; see Algorithm 2). See also the reference (1).
!
!   For further details on randomized linear algebra, computing a partial EVD decomposition
!   using randomized power, subspace or block Krylov iterations, see:
!
!   (1) Martinsson, P.G., 2019:
!            Randomized methods for matrix computations.
!            arXiv.1607.01649
!
!   (2) Erichson, N.B., Voronin, S., Brunton, S.L., and Kutz, J.N., 2019:
!            Randomized matrix decompositions using R.
!            arXiv.1608.02148
!
!   (3) Halko, N., Martinsson, P.G., and Tropp, J.A., 2011:
!            Finding structure with randomness:
!            probabilistic algorithms for constructing approximate matrix decompositions.
!            SIAM Rev., 53, 217-288.
!
!   (4) Musco, C., and Musco, C., 2015:
!            Randomized block krylov methods for stronger and faster approximate singular value decomposition.
!            In Proceedings of the 28th International Conference on Neural Information Processing Systems,
!            NIPS 15, pages 1396-1404, Cambridge, MA, USA, 2015. MIT Press.
!
!   (5) Li, H.,Linderman, G.C., Szlam, A., Stanton, K.P., Kluger, Y., and Tygert, M., 2017:
!            Algorithm 971: An implementation of a randomized algorithm for principal component analysis.
!            ACM Trans. Math. Softw. 43, 3, Article 28 (January 2017).
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert, assert_eq, merror, norm
    use Random,          only : random_seed_, normal_random_number3_
    use QR_Procedures,   only : qr_cmp, ortho_gen_qr
#ifdef _BLAS
    use BLAS_interfaces, only : symm, gemm
!    use BLAS_interfaces, only : gemm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),  dimension(:,:) :: mat
    real(stnd),  intent(out), dimension(:)   :: eigval
    real(stnd),  intent(out), dimension(:,:) :: eigvec
!
    integer(i4b), intent(in), optional :: niter, nover, rng_alg, maxiter
!
    logical(lgl), intent(in),  optional :: ortho, extd_samp
    logical(lgl), intent(out), optional :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                              :: err, tmp
    real(stnd), dimension(:,:), allocatable :: q, b, v, k
    real(stnd), dimension(:),   allocatable :: diagr, beta
!
    integer      :: iok, alg
    integer(i4b) :: n, neig, np, npe, i, i1, i2, niter2, nover2
!
    logical(lgl) :: do_test, do_ortho, do_extd_samp, failure2
!   
    character    :: sort='d'
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc ='reig_cmp'
!
!   niter0 IS THE DEFAULT NUMBER OF RANDOMIZED POWER, SUBSPACE OR BLOCK KRYLOV ITERATIONS PERFORMED.
!   nover0 IS THE DEFAULT OVERSAMPLING SIZE.
!
    integer(i4b),  parameter :: niter0=10_i4b, nover0=10_i4b
!   
!   eps0 IS THE THRESHOLD FOR THE ACCURACY TEST OF THE EIGEN COUPLETS.
!   
    real(stnd), parameter  :: eps0=0.001_stnd
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
!
    n = assert_eq( int(size( mat,    1 ),i4b)  ,  &
                   int(size( mat,    2 ),i4b)  ,  &
                   int(size( eigvec, 1 ),i4b)  ,  &
                   name_proc )
!
    neig = assert_eq( int(size( eigval ),i4b)     ,  &
                      int(size( eigvec, 2 ),i4b)  ,  &
                      name_proc )
!
!   DETERMINE IF A TEST OF EIGEN COUPLETS IS REQUIRED BEFORE EXIT.
!
    do_test = present( failure )
!
    if ( do_test ) then
        failure = false
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. neig<=0_i4b ) then
!
        return
!
    end if
!
    call assert( logical( neig>0_i4b .and. neig<=n, lgl ), name_proc )
!
!   SET OPTIONAL ARGUMENTS.
!
!   DETERMINE THE NUMBER OF POWER, SUBSPACE OR BLOCK KRYLOV ITERATIONS
!   niter2 TO BE PERFORMED.
!
    if ( present(niter) ) then
        niter2 = max( niter, 0_i4b )
    else
        niter2 = niter0
    end if
!
!   DETERMINE THE OVERSAMPLING SIZE nover2 .
!
    if ( present(nover) ) then
        nover2 = max( nover, 0_i4b )
    else
        nover2 = nover0
    end if
!
!   SPECIFY IF ORTHONORMALIZATION IS CARRIED OUT BETWEEN EACH STEP
!   OF THE (POWER) SUBSPACE OR BLOCK KRYLOV ITERATIONS, TO AVOID LOSS OF ACCURACY
!   DUE TO ROUNDING ERRORS. THIS IS NORMALLY NOT NEEDED FOR POSITIVE MATRIX.
!
    if ( present(ortho) ) then
        do_ortho = ortho
    else
        do_ortho = true
    end if
!
!   DETERMINE IF POWER, SUBSPACE OR BLOCK KRYLOV ITERATIONS ARE USED.
!
    if ( present(extd_samp) .and. niter2>0_i4b ) then
        do_extd_samp = extd_samp
    else
        do_extd_samp = false
    end if
!
!   CHECK VALIDITY OF THE OVERSAMPLING SIZE PARAMETER
!   AND OF THE NUMBER OF ITERATIONS FOR BLOCK KRYLOV METHOD.
!
    np = min( nover2 + neig, n ) 
!
    if ( do_extd_samp ) then
!
       npe = ( niter2 + 1_i4b )*np
!
       if ( npe>n ) then
!
           npe = np
!
           do_extd_samp = false
!
       end if
!
   else
!
       npe = np
!
    end if
!
!   SELECT AND SET THE RANDOM NUMBER GENERATOR TO BE USED IF REQUIRED.
!
    if ( present(rng_alg) ) then
!
        if ( rng_alg>=1_i4b .and. rng_alg<=10_i4b ) then
!
            alg = rng_alg
!
!           SELECT THE UNIFORM RANDOM GENERATOR TO BE USED.
!
            call random_seed_( alg=alg )
!
!           RESET THE SEEDS USED BY THE UNIFORM RANDOM GENERATOR.
!
            call random_seed_()
!
        end if
!
    end if
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( q(n,npe), b(n,npe), v(npe,npe), &
              diagr(npe), beta(npe), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A n-BY-np IID GAUSSIAN MATRIX.
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING MATRIX
!   FORM OF SUBROUTINE normal_random_number3_.
!
    call normal_random_number3_( b(:n,:np) )
!
!   GENERATE A NORMAL RANDOM REAL MATRIX USING VECTOR
!   FORM OF SUBROUTINE normal_random_number_.
!
!    do i = 1_i4b, np
!
!        call normal_random_number3_( b(:n,i) )
!
!    end do
!
!   COMPUTE RANDOM SAMPLE MATRIX.
!
#ifdef _BLAS
!    call gemm( 'N', 'N', n, np, n, one, mat(1_i4b:n,1_i4b:n), n,     &
!               b(1_i4b:n,1_i4b:np), n, zero, q(1_i4b:n,1_i4b:np), n  )
    call symm( 'L', 'U', n, np, one, mat(1_i4b:n,1_i4b:n), n,        &
               b(1_i4b:n,1_i4b:np), n, zero, q(1_i4b:n,1_i4b:np), n  )
#else
    q(:n,:np) = matmul( mat(:n,:n), b(:n,:np) )
#endif
!
    if ( do_extd_samp ) then
!
!       USE BLOCK KRYLOV ITERATIONS.
!
!       ALLOCATE WORK ARRAY.
!
        allocate( k(n,np), stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SAVE FIRST KRYLOV BLOCK.
!
        k(:n,:np) = q(:n,:np)
!
        i1 = 1_i4b
        i2 = np
!
    end if
!
!   DO POWER OR SUBSPACE ITERATIONS.
!
    do i = 1_i4b, niter2
!
        if ( do_ortho ) then
!
!           COMPUTE QR DECOMPOSITION OF RANDOM MATRIX.
!
            call qr_cmp( q(:n,:np), diagr(:np), beta(:np) )
!
!           GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM MATRIX
!           TO OBTAIN AN ORTHONORMAL BASIS OF RANDOM COLUMN SUBSPACE.
!           q IS ASSUMED OF FULL RANK.
!
            call ortho_gen_qr( q(:n,:np), beta(:np) )
!
        end if
!
!       COMPUTE RANDOM COLUMN SUBSPACE PROJECTION.
!
#ifdef _BLAS
!        call gemm( 'N', 'N', n, np, n, one, mat(1_i4b:n,1_i4b:n), n,     &
!                   q(1_i4b:n,1_i4b:np), n, zero, b(1_i4b:n,1_i4b:np), n  )
        call symm( 'L', 'U', n, np, one, mat(1_i4b:n,1_i4b:n), n,        &
                   q(1_i4b:n,1_i4b:np), n, zero, b(1_i4b:n,1_i4b:np), n  )
#else
        b(:n,:np) = matmul( mat(:n,:n), q(:n,:np)  )
#endif
!
        q(:n,:np) = b(:n,:np)
!
        if ( do_extd_samp ) then
!
!           USE BLOCK KRYLOV ITERATIONS.
!
            i1 = i1 + np
            i2 = i2 + np
!
!           SAVE i+1 KRYLOV BLOCK.
!
            q(:n,i1:i2) = b(:n,:np)
!
        end if
!
    end do
!
    if ( do_extd_samp ) then
!
!       USE BLOCK KRYLOV ITERATIONS.
!
!       RESTORE FIRST KRYLOV BLOCK.
!
        q(:n,:np) = k(:n,:np)
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( k )
!
    end if
!
!   COMPUTE QR DECOMPOSITION OF RANDOM SAMPLE MATRIX.
!
    call qr_cmp( q(:n,:npe), diagr(:npe), beta(:npe) )
!
!   GENERATE ORTHOGONAL MATRIX q OF QR DECOMPOSITION OF RANDOM MATRIX
!   TO OBTAIN AN ORTHONORMAL BASIS OF RANDOM COLUMN SUBSPACE.
!   q IS ASSUMED OF FULL RANK.
!
    call ortho_gen_qr( q(:n,:npe), beta(:npe) )
!
!   COMPUTE FINAL RANDOM SUBSPACE PROJECTION.
!
#ifdef _BLAS
!    call gemm( 'N', 'N', n, npe, n, one, mat(1_i4b:n,1_i4b:n), n,      &
!               q(1_i4b:n,1_i4b:npe), n, zero, b(1_i4b:n,1_i4b:npe), n  )
    call symm( 'L', 'U', n, npe, one, mat(1_i4b:n,1_i4b:n), n,         &
               q(1_i4b:n,1_i4b:npe), n, zero, b(1_i4b:n,1_i4b:npe), n  )
!
!   COMPUTE v = q**(t)*b = q**(t)*mat*q .
!
    call gemm( 'T', 'N', npe, npe, n, one, q(1_i4b:n,1_i4b:npe), n,        &
               b(1_i4b:n,1_i4b:npe), n, zero, v(1_i4b:npe,1_i4b:npe), npe  )
#else
    b(:n,:npe) = matmul( mat(:n,:n), q(:n,:npe)  )
!
!   ALLOCATE WORK ARRAY.
!
    allocate( k(npe,n),  stat=iok  )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   COMPUTE v = q**(t)*b = q**(t)*mat*q .
!
    k(:npe,:n) = transpose( q(:n,:npe) )
!
    v(:npe,:npe) = matmul( k(:npe,:n), b(:n,:npe) )
!
!   DEALLOCATE WORK ARRAY.
!
    deallocate( k )
#endif
!
!   COMPUTE EVD DECOMPOSITION OF REDUCED MATRIX.
!
    call eig_cmp_b( v(:npe,:npe), beta(:npe), failure2, maxiter=maxiter )
!
    call eig_abs_sort_a( sort, beta(:npe), v(:npe,:npe) )
!
!   COMPUTE THE APPROXIMATE TOP neig EIGENVECTORS OF mat .
!
#ifdef _BLAS
    call gemm( 'N', 'N', n, neig, npe, one, q(1_i4b:n,1_i4b:npe), n,              &
               v(1_i4b:npe,1_i4b:neig), npe, zero, eigvec(1_i4b:n,1_i4b:neig), n  )
#else
    eigvec(:n,:neig) = matmul( q(:n,:npe), v(:npe,:neig) )
#endif
!
!   EXTRACT THE APPROXIMATE TOP neig EIGENVALUES (IN ABSOLUTE MAGNITUDE) OF mat .
!
    eigval(:neig) = beta(:neig)
!
!   TEST ACCURACY OF THE APPROXIMATE EVD IF REQUIRED.
!
    if ( do_test ) then
!
!       CHECK FOR SMALL RESIDUALS OF THE EXPRESSION mat*u(:n,:neig) - u(:n,:neig)*diag(eigval(:neig)),
!       WHERE u ARE THE EIGENVECTORS OF mat.
!
        q(:n,:neig) = eigvec(:n,:neig)*spread(eigval(:neig),dim=1,ncopies=n)
!
#ifdef _BLAS
!        call gemm( 'N', 'N', n, neig, n, one, mat(1_i4b:n,1_i4b:n), n,            &
!                   eigvec(1_i4b:n,1_i4b:neig), n, -one, q(1_i4b:n,1_i4b:neig), n  )
        call symm( 'L', 'U', n, neig, one, mat(1_i4b:n,1_i4b:n), n,        &
                   eigvec(1_i4b:n,1_i4b:neig), n, -one, q(1_i4b:n,1_i4b:neig), n  )
#else
        q(:n,:neig) = matmul( mat(:n,:n), eigvec(:n,:neig)  ) - q(:n,:neig)
#endif
!
        beta(:neig) = norm( q(:n,:neig), dim=2_i4b )
!
        err = maxval( beta(:neig) )/( norm( mat )*real(n,stnd) )
!
        failure = err>eps0 .or. failure2
!
    end if
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( q, b, v, diagr, beta )
!
!
! END OF SUBROUTINE reig_cmp
! __________________________
!
    end subroutine reig_cmp
!
!
! =========================================================================================
!                         INVERSE ITERATION PROCEDURES
! =========================================================================================
!
!
    function maxdiag_tinv_qr( d, e, lambda )
!
! Purpose
! _______
!
!   This function computes the index of the element of maximum absolute value in the
!   diagonal entries of ( T - LAMBDA * I )**(-1) where T is a symmetric tridiagonal matrix,
!   I is the identity matrix and LAMBDA is a scalar.
!
! Arguments
! _________
!
!   D           (INPUT) real(stnd), dimension(:)
!               On entry, the diagonal elements of the tridiagonal matrix.
!
!   E           (INPUT) real(stnd), dimension(:)
!               On entry, the n-1 subdiagonal elements of the tridiagonal matrix.
!
!               The size of E must verify:  size( E ) = size( D ) - 1.
!
!   LAMBDA      (INPUT) real(stnd)
!               On entry, the eigenvalue or shift used in the QR factorization.
!
!
! Further Details
! _______________
!
!   The diagonal entries of ( T - LAMBDA * I )**(-1) are computed by means of the QR
!   factorization of ( T - LAMBDA * I ). For the latter computation, the semiseparable
!   structure of ( T - LAMBDA * I )**(-1) is used, see the reference (1). Moreover, it
!   is assumed that T is unreduced, but no check is done in the subroutine to verify
!   this assumption.
!
!   This subroutine is adapted from the pseudo-code trace_Tinv given in the reference (1).
!
!   For further details, see:
!
!   (1) Bini, D.A., Gemignani, L., and Tisseur, F., 2005:
!           The Ehrlich-Aberth method for the nonsymmetric tridiagonal eigenvalue problem.
!           SIAM J. Matrix Anal. Appl., 27, 153-175.
!
!   (2) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, 1013-1034.
!
!   (3) Parlett, B.N., and Dhillon, I.S., 1997:
!           Fernando's solution to Wilkinson's problem: An application of double factorization.
!           Linear Algebra and its Appl., 267, pp.247-279.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,  only : lamch
    use Utilities,      only : assert_eq
    use Giv_Procedures, only : define_rot_givens
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),                 intent(in) :: lambda
    real(stnd), dimension(:),   intent(in) :: d, e
!
    integer(i4b) :: maxdiag_tinv_qr
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: safmin, dmin, a, b, g, cs, sn, tau
    real(stnd), dimension(size(d)) :: r, s, t, v, w
!
    integer(i4b) :: n, i, k
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='maxdiag_tinv_qr'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(d),i4b)         ,   &
                    int(size(e),i4b)  + 1_i4b,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        maxdiag_tinv_qr = n
!
        return
!
    end if
!
!   DETERMINE THE UNDERFLOW THRESHOLD.
!
    safmin = lamch( 's' )
!
!   REPLACEMENT FOR VERY SMALL PIVOTS IN QR FACTORIZATION.
!
!    g = max( maxval( abs(d(1_i4b:n))), maxval( abs(e(1_i4b:n-1_i4b))) )
!    dmin = max( safmin, (safmin*g)*g ) 
!
!    dmin = sqrt( safmin )
!
    dmin = safmin
!
!   MAKE A LOCAL COPY OF THE SHIFTED TRIDIAGONAL MATRIX.
!
    r(:n)       = d(:n) - lambda
    s(:n-1_i4b) = e(:n-1_i4b)
    s(n)        = zero
!
!   REPLACE SMALL SUBDIAGONALS BY safmin.
!
    where ( abs(s(1_i4b:n-1_i4b))<=safmin )                      &
          s(1_i4b:n-1_i4b) = sign( safmin, s(1_i4b:n-1_i4b) )
!
!   COMPUTE THE QR FACTORIZATION AND THE ENTRIES OF THE SCALED
!   FORM OF R IN THE VECTORS r(:), s(:) AND t(:).
!
    a = r(1_i4b)
    g = s(1_i4b)
    b = g
!
    do i = 1_i4b, n-1_i4b
!
        call define_rot_givens( a, b, cs, sn )
!
        r(i) = cs*a + sn*b
        s(i) = -sn*( cs*g + sn*r(i+1_i4b) )
        a    = -sn*g + cs*r(i+1_i4b)
!
!       SAVE COSINUS OF GIVENS ROTATION FOR LATER USE.
!
        v(i) = cs
!
        b    = s(i+1_i4b)
        t(i) = sn*sn*b
        g    = cs*b
!
        if ( i==1_i4b ) cycle
!
        t(i-1_i4b) = sn*t(i-1_i4b)
!
    end do
!
    r(n) = a
    v(n) = one
!
!   REPLACE SMALL DIAGONALS BY dmin.
!
    where ( abs(r(:n))<=dmin ) r(:n) = sign( dmin, r(:n) )
!
!   SOLVE THE TRIANGULAR LINEAR SYSTEM R.w = v .
!
!    w(n)       = v(n)/r(n)
!    w(n-1_i4b) = ( v(n-1_i4b) - w(n)*s(n-1_i4b) )/r(n-1_i4b)
!
!    do i = n-2_i4b, 1_i4b, -1_i4b
!
!       w(i) = ( v(i) - w(i+1_i4b)*s(i) - w(i+2_i4b)*t(i) )/r(i)
!
!    end do
!
!   SOLVE THE TRIANGULAR LINEAR SYSTEM R.w = v WITH SCALING
!   TO AVOID OVERFLOWS.
!
    w(:n) = v(:n)
!
    if ( abs(w(n))>abs(r(n)) ) then
!
        a           = r(n)/w(n)
        w(:n-1_i4b) = a*w(:n-1_i4b)
        w(n)        = one
!
    else
!
        w(n) = w(n)/r(n)
!
    end if
!
    a = w(n-1_i4b) - s(n-1_i4b)*w(n)
!
    if ( abs(a)>abs(r(n-1_i4b)) ) then
!
        a          = r(n-1_i4b)/a
        w(:n)      = a*w(:n)
        w(n-1_i4b) = one
!
    else
!
        w(n-1_i4b) = a/r(n-1_i4b)
!
    end if
!    
    do i = n-2_i4b, 1_i4b, -1_i4b
!
        a = w(i) - s(i)*w(i+1_i4b) - t(i)*w(i+2_i4b)
!
        if ( abs(a)>abs(r(i)) ) then
!
            a     = r(i)/a
            w(:n) = a*w(:n)
            w(i)  = one
!
        else
!
            w(i) = a/r(i)
!
        end if
!
    end do
!
!   FIND OPTIMAL k AND tau .
!
    tau = abs( w(1_i4b) )
    k   = 1_i4b
!
    do i = 2_i4b, n
!
        a = abs( v(i-1_i4b)*w(i) )
!
        if ( a<tau ) cycle
!
        tau = a
        k   = i
!
    end do
!
    maxdiag_tinv_qr = k
!
!
! END OF function maxdiag_tinv_qr
! _______________________________
!
    end function maxdiag_tinv_qr
!
! =========================================================================================
!
    function maxdiag_tinv_ldu( d, e, lambda )
!
! Purpose
! _______
!
!   This function computes the index of the element of maximum absolute value in the
!   diagonal entries of ( T - LAMBDA * I )**(-1) where T is a symmetric tridiagonal matrix,
!   I is the identity matrix and LAMBDA is a scalar.
!
! Arguments
! _________
!
!   D           (INPUT) real(stnd), dimension(:)
!               On entry, the diagonal elements of the tridiagonal matrix.
!
!   E           (INPUT) real(stnd), dimension(:)
!               On entry, the n-1 subdiagonal elements of the tridiagonal matrix.
!
!               The size of E must verify:  size( E ) = size( D ) - 1.
!
!   LAMBDA      (INPUT) real(stnd)
!               On entry, the eigenvalue or shift used.
!
!
! Further Details
! _______________
!
!   The diagonal entries of ( T - LAMBDA * I )**(-1) are computed by means of two
!   triangular factorizations of ( T - LAMBDA * I ) of the forms L(+)D(+)U(+) and
!   U(-)D(-)L(-) where L(+) and L(-) are unit lower bidiagonal, U(+) and U(-) are unit
!   upper bidiagonal, and D(+) and D(-) are diagonal.
!
!   It is assumed that T is unreduced, but no check is done in the subroutine to verify
!   this assumption.
!
!   This subroutine is adapted from the references (1) and (2).
!
!   For further details, on Fernando's method for computing eigenvectors of tridiagonal
!   matrices, see:
!
!   (1) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, pp. 1013-1034.
!
!   (2) Parlett, B.N., and Dhillon, I.S., 1997:
!           Fernando's solution to Wilkinson's problem: An application of double factorization.
!           Linear Algebra and its Appl., 267, pp.247-279.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,  only : lamch
    use Utilities,      only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),                 intent(in) :: lambda
    real(stnd), dimension(:),   intent(in) :: d, e
!
    integer(i4b) :: maxdiag_tinv_ldu
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: safmin, gamma, dmin, tau, tmp, x
    real(stnd), dimension(size(d)) :: diag, e2
!
    integer(i4b) :: n, i, k
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='maxdiag_tinv_ldu'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(d),i4b)         ,   &
                    int(size(e),i4b)  + 1_i4b,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        maxdiag_tinv_ldu = n
!
        return
!
    end if
!
!   DETERMINE THE UNDERFLOW THRESHOLD.
!
    safmin = lamch( 's' )
    gamma  = sqrt( safmin )
!
!   REPLACEMENT FOR VERY SMALL PIVOTS IN LDU AND UDL FACTORIZATIONS.
!
!    x    = max( maxval( abs(d(1_i4b:n))), maxval( abs(e(1_i4b:n-1_i4b))) )
!    dmin = max( safmin, (safmin*x)*x ) 
!
    dmin = gamma 
!
!   MAKE A LOCAL COPY OF THE OFF DIAGONAL ELEMENTS OF THE TRIDIAGONAL MATRIX T.
!
    e2(:n-1_i4b) = e(:n-1_i4b)
    e2(n)        = zero
!
!   REPLACE SMALL SUBDIAGONALS BY gamma.
!
    where ( abs(e2(1_i4b:n-1_i4b))<=gamma )                      &
          e2(1_i4b:n-1_i4b) = sign( gamma, e2(1_i4b:n-1_i4b) )
!
!   SQUARE THE OFF DIAGONAL ELEMENTS OF THE TRIDIAGONAL MATRIX T.
!
    e2(:n-1_i4b) = e2(:n-1_i4b)*e2(:n-1_i4b)
!    
!   FERNANDO METHOD.
!
!   LDU FACTORIZATION OF T - lambda*I .
!
    x = d(1_i4b) - lambda
    diag(1_i4b) = zero
!
    do i = 2_i4b, n
!
        if ( abs(x)<dmin ) then
            x = -dmin
        end if
!
        diag(i) = - e2(i-1_i4b)/x
        x       = ( d(i) + diag(i) ) - lambda
!
    end do
!
!   UDL FACTORIZATION OF T - lambda*I .
!
!   FIND OPTIMAL k AND tau.
!
    x   = d(n) - lambda
    tau = abs( diag(n) + x )
    k   = n
!
    do i = n-1_i4b, 1_i4b, -1_i4b
!
        if ( abs(x)<dmin ) then
             x = -dmin
        end if
!
        x   = ( d(i) - e2(i)/x ) - lambda
        tmp = abs( diag(i) + x )
!
        if ( tmp>=tau ) cycle
!
        tau = tmp
        k   = i
!
    end do
!
    maxdiag_tinv_ldu = k
!
!
! END OF function maxdiag_tinv_ldu
! ________________________________
!
    end function maxdiag_tinv_ldu
!
! =========================================================================================
!
    subroutine trid_qr_cmp_rv( d, e, lambda, cs, sn, diag, sup1, sup2, maxdiag_tinv )
!
! Purpose
! _______
!
!  TRID_QR_CMP factorizes the symmetric matrix T - LAMBDA * I, where 
!  T is an n by n symmetric tridiagonal matrix, I is the identity matrix
!  and LAMBDA is a scalar, as
!
!                         T - LAMBDA * I = Q * R
!
!  where Q is an orthogonal matrix represented as the product of n-1 Givens
!  rotations and R is an upper triangular matrix with at most two non-zero
!  super-diagonal elements per column.
!
!  The parameter LAMBDA is included in the routine so that TRID_QR_CMP may
!  be used to obtain eigenvectors of T by inverse iteration.
!
!  The subroutine also computes the index of the entry of maximum absolute value
!  in the diagonal of ( T - LAMBDA * I )**(-1), which provides a good initial
!  approximation to start the inverse iteration process for computing the eigenvector
!  assoociated with the eigenvalue LAMBDA, see the references (1), (2) and (3) for
!  further details.
!
! Arguments
! _________
!
!   D            (INPUT) real(stnd), dimension(:)
!                On entry, the diagonal elements of the tridiagonal matrix.
!
!   E            (INPUT) real(stnd), dimension(:)
!                On entry, the n-1 subdiagonal elements of the tridiagonal matrix.
!
!                The size of E must verify:  size( E ) = size( D ) - 1.
!
!   LAMBDA       (INPUT) real(stnd)
!                On entry, the eigenvalue or shift used in the QR factorization.
!
!   CS           (OUTPUT) real(stnd), dimension(:)
!                On exit, the vector of the cosines coefficients of the chain of n-1 Givens
!                rotations for the QR factorization of T - LAMBDA * I.
!
!                The size of CS must verify:  size( CS ) = size( E ) = size( D ) - 1.
!
!   SN           (OUTPUT) real(stnd), dimension(:)
!                On exit, the vector of the sines coefficients of the chain of n-1 Givens
!                rotations for the QR factorization of T - LAMBDA * I.
!
!                The size of SN must verify:  size( SN ) = size( E ) = size( D ) - 1.
!
!   DIAG         (OUTPUT) real(stnd), dimension(:)
!                On exit, DIAG(:) contains the n diagonal elements of the
!                upper triangular matrix R of the QR factorization of T - LAMBDA * I.
!               
!                The size of DIAG must verify: size( DIAG ) = size( D ) = n .
!
!   SUP1         (OUTPUT) real(stnd), dimension(:)
!                On exit, SUP1(:n-1) contains the n-1 superdiagonal elements of the
!                upper triangular matrix R of the QR factorization of T - LAMBDA * I,
!                SUP1(n) is arbitrary.
!
!                The size of SUP1 must verify: size( SUP1 ) = size( D ) = n .
!
!   SUP2         (OUTPUT) real(stnd), dimension(:)
!                On exit, SUP2(:n-2) contains the n-2 second superdiagonal elements of
!                the upper triangular matrix R of the QR factorization of T - LAMBDA * I,
!                SUP2(n-1:n) is arbitrary.
!
!                The size of SUP2 must verify: size( SUP2 ) = size( D ) = n .
!
!   MAXDIAG_TINV (OUPTPUT) integer(i4b)
!                On exit, MAXDIAG_TINV is the index of the entry of maximum modulus
!                in the main diagonal of ( T - LAMBDA * I )**(-1).
!
!
! Further Details
! _______________
!
!   The QR factorization of ( T - LAMBDA * I ) is obtained by means of n-1 unitary
!   Givens rotations.
!
!   The diagonal entries of ( T - LAMBDA * I )**(-1) are computed by means of this QR
!   factorization of ( T - LAMBDA * I ). For the latter computation, the semiseparable
!   structure of ( T - LAMBDA * I )**(-1) is used, see the reference (1). Moreover, it
!   is assumed that T is unreduced for computing the index of the entry of maximum
!   absolute value in the diagonal of ( T - LAMBDA * I )**(-1), but no check is done
!   in the subroutine to verify this assumption.
!
!   For further details, see:
!
!   (1) Bini, D.A., Gemignani, L., and Tisseur, F., 2005:
!           The Ehrlich-Aberth method for the nonsymmetric tridiagonal eigenvalue problem.
!           SIAM J. Matrix Anal. Appl., 27, 153-175.
!
!   (2) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, pp. 1013-1034.
!
!   (3) Parlett, B.N., and Dhillon, I.S., 1997:
!           Fernando's solution to Wilkinson's problem: An application of double factorization.
!           Linear Algebra and its Appl., 267, pp.247-279.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,  only : lamch
    use Utilities,      only : assert_eq
    use Giv_Procedures, only : define_rot_givens
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),                intent(in)  :: lambda
    real(stnd), dimension(:),  intent(in)  :: d, e
    real(stnd), dimension(:),  intent(out) :: cs, sn, diag, sup1, sup2
!
    integer(i4b), intent(out) :: maxdiag_tinv
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: safmin, dmin, a, b, g, csi, sni, tau
    real(stnd), dimension(size(d)) :: s, t, w
!
    integer(i4b) :: n, i, k
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='trid_qr_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( (/ int(size(d),i4b)          ,   &
                       int(size(e),i4b)  + 1_i4b ,   &
                       int(size(cs),i4b) + 1_i4b ,   &
                       int(size(sn),i4b) + 1_i4b ,   &
                       int(size(diag),i4b)       ,   &
                       int(size(sup1),i4b)       ,   &
                       int(size(sup2),i4b) /)    ,   &
                       name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            diag(1_i4b) = d(1_i4b) - lambda
            sup1(1_i4b) = zero
            sup2(1_i4b) = zero
!
        end if
!
        maxdiag_tinv = n
!
        return
!
    end if
!
!   DETERMINE THE UNDERFLOW THRESHOLD.
!
    safmin = lamch( 's' )
!
!   REPLACEMENT FOR VERY SMALL PIVOTS IN QR FACTORIZATION.
!
!    g = max( maxval( abs(d(1_i4b:n))), maxval( abs(e(1_i4b:n-1_i4b))) )
!    dmin = max( safmin, (safmin*g)*g ) 
!
!    dmin = sqrt( safmin )
!
    dmin = safmin
!
!   MAKE A COPY OF THE SHIFTED TRIDIAGONAL MATRIX.
!
    diag(:n)       = d(:n) - lambda
    sup1(:n-1_i4b) = e(:n-1_i4b)
    sup1(n)        = zero
!
!   REPLACE SMALL SUBDIAGONALS BY safmin.
!
    where ( abs(sup1(1_i4b:n-1_i4b))<=safmin )                      &
          sup1(1_i4b:n-1_i4b) = sign( safmin, sup1(1_i4b:n-1_i4b) )
!
!   COMPUTE THE QR FACTORIZATION SAVING THE ENTRIES OF R IN THE VECTORS
!   diag(:), sup1(:) AND sup2(:).
!
    a = diag(1_i4b)
    g = sup1(1_i4b)
    b = g
!
    do i = 1_i4b, n-1_i4b
!
        call define_rot_givens( a, b, csi, sni )
!
        diag(i) =  csi*a + sni*b
        sup1(i) =  csi*g + sni*diag(i+1_i4b)
        a       = -sni*g + csi*diag(i+1_i4b)
!
        b       = sup1(i+1_i4b)
        sup2(i) = sni*b
        g       = csi*b
!
!       SAVE THE COSINES AND SINES COEFFICIENTS OF THE CHAIN
!       OF n-1 GIVENS ROTATIONS DEFINING Q.
!
        cs(i) = csi
        sn(i) = sni
!
    end do
!
    diag(n)         = a
    sup2(n-1_i4b:n) = zero
!
!   REPLACE SMALL DIAGONALS BY dmin.
!
    where ( abs(diag(:n))<=dmin ) diag(:n) = sign( dmin, diag(:n) )
!
!   SCALE THE UPPER TRIANGULAR MATRIX R FOR COMPUTING maxdiag_tinv.
!
    s(:n-1_i4b) = -sn(:n-1_i4b)*sup1(:n-1_i4b)
    s(n)        = zero
!
    t(:n-2_i4b)  = sn(2_i4b:n-1_i4b)*sn(:n-2_i4b)*sup2(:n-2_i4b)
    t(n-1_i4b:n) = zero
!
    w(:n-1_i4b) = cs(:n-1_i4b)
!
!   SOLVE THE TRIANGULAR LINEAR SYSTEM R.w = v WITH SCALING
!   TO AVOID OVERFLOWS.
!
    if ( one>abs(diag(n)) ) then
!
        a           = diag(n)
        w(:n-1_i4b) = a*w(:n-1_i4b)
        w(n)        = one
!
    else
!
        w(n) = one/diag(n)
!
    end if
!
    a = w(n-1_i4b) - s(n-1_i4b)*w(n)
!
    if ( abs(a)>abs(diag(n-1_i4b)) ) then
!
        a          = diag(n-1_i4b)/a
        w(:n)      = a*w(:n)
        w(n-1_i4b) = one
!
    else
!
        w(n-1_i4b) = a/diag(n-1_i4b)
!
    end if
!    
    do i = n-2_i4b, 1_i4b, -1_i4b
!
        a = w(i) - s(i)*w(i+1_i4b) - t(i)*w(i+2_i4b)
!
        if ( abs(a)>abs(diag(i)) ) then
!
            a     = diag(i)/a
            w(:n) = a*w(:n)
            w(i)  = one
!
        else
!
            w(i) = a/diag(i)
!
        end if
!
    end do
!
!   SOLVE THE TRIANGULAR LINEAR SYSTEM R.w = v .
!
!    w(n)       = one/diag(n)
!    w(n-1_i4b) = ( w(n-1_i4b) - w(n)*s(n-1_i4b) )/diag(n-1_i4b)
!
!    do i = n-2_i4b, 1_i4b, -1_i4b
!
!       w(i) = ( w(i) - w(i+1_i4b)*s(i) - w(i+2_i4b)*t(i) )/diag(i)
!
!    end do
!
!   FIND OPTIMAL k AND tau .
!
    tau = abs( w(1_i4b) )
    k   = 1_i4b
!
    do i = 2_i4b, n
!
        a = abs( cs(i-1_i4b)*w(i) )
!
        if ( a<tau ) cycle
!
        tau = a
        k   = i
!
    end do
!
    maxdiag_tinv = k
!
!
! END OF subroutine trid_qr_cmp_rv
! ________________________________
!
    end subroutine trid_qr_cmp_rv
!
! =========================================================================================
!
    subroutine trid_qr_solve_rv( cs, sn, diag, sup1, sup2, y )
!
! Purpose
! _______
!
!   TRID_QR_SOLVE may be used to solve for x(:) the system of equations
!
!                x(:) * (T - LAMBDA * I) = scale * y(:)
!
!   , where T is an n-by-n symmetric tridiagonal matrix, I is the n-by-n identity
!   matrix, LAMBDA and scale are scalars, following the factorization of
!   (T - LAMBDA * I) by TRID_QR_CMP or GK_QR_CMP, as
!
!                         T - LAMBDA * I = Q * R
!
!   where Q is an orthogonal matrix represented as the product of n-1 Givens
!   rotations and R is an upper triangular matrix with at most two non-zero
!   super-diagonal elements per column.
!
!   The matrix (T - LAMBDA * I) is assumed to be ill-conditioned, and frequent rescalings
!   are carried out in order to avoid overflow. However, no test for singularity
!   or near-singularity is included in this routine. Such tests must be performed
!   before calling this routine. The scalar, scale, is not output by this routine 
!   since this routine being intended for use in applications such as inverse iteration.
!
!
! Arguments
! _________
!
!    CS          (INPUT) real(stnd), dimension(:)
!                On entry, the vector of the cosines coefficients of the chain of n-1 Givens
!                rotations for the QR factorization of T - LAMBDA * I as computed by
!                TRID_QR_CMP or GK_QR_CMP.
!
!                The size of CS must verify:  size( CS ) = size( Y ) - 1.
!
!    SN          (INPUT) real(stnd), dimension(:)
!                On entry, the vector of the sines coefficients of the chain of n-1 Givens
!                rotations for the QR factorization of GK - LAMBDA * I as computed by
!                TRID_QR_CMP or GK_QR_CMP.
!
!                The size of SN must verify:  size( SN ) = size( Y ) - 1.
!
!    DIAG        (INPUT) real(stnd), dimension(:)
!                On entry, DIAG(:) contains the n diagonal elements of the
!                upper triangular matrix R of the QR factorization of T - LAMBDA * I.
!               
!                The size of DIAG must verify: size( DIAG ) = size( Y ) = n .
!
!    SUP1        (INPUT) real(stnd), dimension(:)
!                On entry, SUP1(:n-1) contains the n-1 superdiagonal elements of the
!                upper triangular matrix R of the QR factorization of T - LAMBDA * I,
!                SUP1(n) is arbitrary.
!
!                The size of SUP1 must verify: size( SUP1 ) = size( Y ) = n .
!
!    SUP2        (INPUT) real(stnd), dimension(:)
!                On entry, SUP2(:n-2) contains the n-2 second superdiagonal elements of
!                the upper triangular matrix R of the QR factorization of T - LAMBDA * I,
!                SUP2(n-1:n) is arbitrary.
!
!                The size of SUP2 must verify: size( SUP2 ) = size( Y ) = n .
!
!    Y           (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the right hand side vector y.
!                On exit, Y is overwritten the solution vector x.
!
!                The size of Y must verify: size( Y ) = n .
!
!
! Further Details
! _______________
!
!   For further details, see:
!
!   (1) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, pp. 1013-1034.
!
!   (2) Bini, D.A., Gemignani, L., and Tisseur, F., 2005:
!           The Ehrlich-Aberth method for the nonsymmetric tridiagonal eigenvalue problem.
!           SIAM J. Matrix Anal. Appl., 27, 153-175.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),    dimension(:) :: cs, sn, diag, sup1, sup2
    real(stnd),  intent(inout), dimension(:) :: y
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: tmp, c, s
!
    integer(i4b) :: n, i
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='trid_qr_solve'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( (/ int(size(y),i4b)             ,   &
                       int(size(diag),i4b)          ,   &
                       int(size(sup1),i4b)          ,   &
                       int(size(sup2),i4b)          ,   &
                       int(size(cs),i4b) + 1_i4b    ,   &
                       int(size(sn),i4b) + 1_i4b /) ,   &
                       name_proc )
!
!    QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!    
!   COMPUTE Q**(t)*y = b AS THE PRODUCT OF THE GIVENS ROTATIONS
!   TIMES THE VECTOR y(:).
!    
    tmp = y(1_i4b)
!
    do i = 1_i4b, n-1_i4b
!
        c = cs(i)
        s = sn(i)
!
!       APPLY THE GIVENS ROTATION TO THE VECTOR y(:).
!
        y(i) = c*tmp + s*y(i+1_i4b)
        tmp  = c*y(i+1_i4b) - s*tmp
!
    end do
!   
!   SOLVE R*x = b WITH SCALING TO AVOID OVERFLOWS.
!    
    if ( abs(tmp)>abs(diag(n)) ) then
!
        tmp         = diag(n)/tmp
        y(:n-1_i4b) = tmp*y(:n-1_i4b)
        y(n)        = one
!
    else
!
        y(n) = tmp/diag(n)
!
    end if
!    
!    y(n) = tmp/diag(n)
!    
    if ( n>1_i4b ) then
!
        tmp = y(n-1_i4b) - sup1(n-1_i4b)*y(n)
!
        if ( abs(tmp)>abs(diag(n-1_i4b)) ) then
!
            tmp        = diag(n-1_i4b)/tmp
            y(:n)      = tmp*y(:n)
            y(n-1_i4b) = one
!
        else
!
            y(n-1_i4b) = tmp/diag(n-1_i4b)
!
        end if
!
!        if ( n>1_i4b )                      &
!        y(n-1_i4b) = ( y(n-1_i4b) - sup1(n-1_i4b)*y(n) )/diag(n-1_i4b)
!    
        do i = n-2_i4b, 1_i4b, -1_i4b
!
            tmp = y(i) - sup1(i)*y(i+1_i4b) - sup2(i)*y(i+2_i4b)
!
            if ( abs(tmp)>abs(diag(i)) ) then
!
                tmp   = diag(i)/tmp
                y(:n) = tmp*y(:n)
                y(i)  = one
!
            else
!
                y(i) = tmp/diag(i)
!
            end if
!
        end do
!
!        do i = n-2_i4b, 1_i4b, -1_i4b
!            y(i) = ( y(i) - sup1(i)*y(i+1_i4b) - sup2(i)*y(i+2_i4b))/diag(i)
!        end do
!
    end if
!
!
! END OF SUBROUTINE trid_qr_solve_rv
! __________________________________
!
    end subroutine trid_qr_solve_rv
!
! =========================================================================================
!
    subroutine trid_cmp_rm( d, e, eigval, sub, diag, sup1, sup2, perm, tol )
!
! Purpose
! _______
!
!   TRID_CMP factorizes symmetric matrices of the form (T - EIGVAL(j) * I), where 
!   T is an n-by-n symmetric tridiagonal matrix and EIGVAL(j) is a scalar, as
!
!     T - EIGVAL(j) * I = P(j) * L(j) * U(j),    for j=1, SIZE(EIGVAL)
!
!   where P(j) is a permutation matrix, L(j) is a unit lower tridiagonal matrix
!   with at most one non-zero sub-diagonal elements per column and U(j) is
!   an upper triangular matrix with at most two non-zero super-diagonal
!   elements per column.
!
!   The factorizations, for j=1, SIZE(EIGVAL), are obtained by Gaussian elimination
!   with partial pivoting and implicit row scaling.
!
!   The parameters EIGVAL are included in the routine so that TRID_CMP may
!   be used to obtain eigenvectors of T by inverse iteration.
!
!
! Arguments
! _________
!
!    D           (INPUT) real(stnd), dimension(:)
!                On entry, the diagonal elements of the symmetric tridiagonal matrix T.
!
!    E           (INPUT) real(stnd), dimension(:)
!                On entry, the n-1 subdiagonal elements of the symmetric tridiagonal
!                matrix T and E(n) is arbitrary .
!
!                The size of E must verify:  size( E ) = size( D ) = n .
!
!    EIGVAL      (INPUT) real(stnd), dimension(:)
!                On entry, selected eigenvalues of the symmetric tridiagonal matrix.
!
!                The size of EIGVAL must verify: size( EIGVAL ) <= size( D ) = n .
!
!    SUB         (OUTPUT) real(stnd), dimension(:,:)
!                On exit, SUB(j,:n-1) contains the n-1 subdiagonal elements of the
!                lower triangular matrix L(j) of the factorization of T - EIGVAL(j) * I,
!                for j=1, SIZE(EIGVAL). SUB(:,n) is arbitrary .
!
!                The shape of SUB must verify: 
!
!                - size( SUB, 1 ) = size( EIGVAL ) ;
!                - size( SUB, 2 ) = size( D ) = n  .
!
!    DIAG        (OUTPUT) real(stnd), dimension(:,:)
!                On exit, DIAG(j,:) contains the n diagonal elements of the
!                upper triangular matrix U(j) of the factorization of T - EIGVAL(j) * I,
!                for j=1, SIZE(EIGVAL).
!               
!                The shape of DIAG must verify: 
!
!                - size( DIAG, 1 ) = size( EIGVAL ) ;
!                - size( DIAG, 2 ) = size( D ) = n  .
!
!    SUP1        (OUTPUT) real(stnd), dimension(:,:)
!                On exit, SUP1(j,:n-1) contains the n-1 superdiagonal elements of the
!                upper triangular matrix U(j) of the factorization of T - EIGVAL(j) * I,
!                for j=1, SIZE(EIGVAL). SUP1(:,n) is arbitrary .
!
!                The shape of SUP1 must verify: 
!
!                - size( SUP1, 1 ) = size( EIGVAL ) ;
!                - size( SUP1, 2 ) = size( D ) = n  .
!
!    SUP2        (OUTPUT) real(stnd), dimension(:,:)
!                On exit, SUP2(j,:n-2) contains the n-2 second superdiagonal elements of
!                the upper triangular matrix U(j) of the factorization of T - EIGVAL(j) * I,
!                for j=1, SIZE(EIGVAL). SUP2(:,n-1:n) is arbitrary .
!
!                The shape of SUP2 must verify: 
!
!                - size( SUP2, 1 ) = size( EIGVAL ) ;
!                - size( SUP2, 2 ) = size( D ) = n  .
!
!    PERM        (OUTPUT) logical(lgl), dimension(:,:)
!                On exit, PERM(j,:n-1)  contains details of the permutation matrix P(j).
!                If an interchange occurred at the kth step of the elimination in the
!                factorization of (T - EIGVAL(j) * I), then PERM(j,k) = true, otherwise
!                PERM(j,k) = false.
!                The element PERM(j,n) is set to true if there is an integer l such that
!
!                  abs( u(j)(l,l) ).le. norm( (T - EIGVAL(j) * I)(l) ) * TOL,
!
!                where norm( A(l) ) denotes the sum of the absolute values of
!                the lth row of the matrix A. If no such l exists then PERM(j,n)
!                is returned as false. If PERM(j,n) is returned as true, then a
!                diagonal element of U(j) is small, indicating that
!                (T - EIGVAL(j) * I) is singular or nearly singular.
!
!                The shape of PERM must verify: 
!
!                - size( PERM, 1 ) = size( EIGVAL ) ;
!                - size( PERM, 2 ) = size( D ) = n  .
!
!    TOL         (INPUT, OPTIONAL)  real(stnd)
!                On entry, a relative tolerance used to indicate whether or
!                not the matrices (T - EIGVAL(j) * I) are nearly singular. TOL should
!                normally be choose as approximately the largest relative error
!                in the elements of T. For example, if the elements of T are
!                correct to about 4 significant figures, then TOL should be
!                set to about 5 * 10**(-4).
!
!                If TOL is supplied as less than eps, where eps is the 
!                relative machine precision, then the value eps is used
!                in place of TOL.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the routine  DLAGTF in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,    only : lamch
    use Utilities,        only : assert_eq, assert
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),  dimension(:)    :: d, e, eigval
    real(stnd),  intent(out), dimension(:,:)  :: sub, diag, sup1, sup2
    real(stnd),  intent(in),  optional        :: tol
!
    logical(lgl), intent(out), dimension(:,:) :: perm
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: tl, ek, ekp1, scal(size(eigval),size(d))
!
    integer(i4b) :: n, m, k
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='trid_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( (/ int(size(d),i4b)         ,   &
                       int(size(e),i4b)         ,   &
                       int(size(sub,2),i4b)     ,   &
                       int(size(diag,2),i4b)    ,   &
                       int(size(sup1,2),i4b)    ,   &
                       int(size(sup2,2),i4b)    ,   &
                       int(size(perm,2),i4b) /) ,   &
                       name_proc )
!
    m =  assert_eq( (/ int(size(eigval),i4b)    ,   &
                       int(size(sub,1),i4b)     ,   &
                       int(size(diag,1),i4b)    ,   &
                       int(size(sup1,1),i4b)    ,   &
                       int(size(sup2,1),i4b)    ,   &
                       int(size(perm,1),i4b) /) ,   &
                       name_proc )
!
    call assert( logical( m<=n, lgl ), name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b .or. m<=0_i4b) then
!
        if ( n==1_i4b .and. m==1_i4b ) then
!
            diag(1_i4b,1_i4b) = d(1_i4b) - eigval(1_i4b)
            perm(1_i4b,1_i4b) = diag(1_i4b,1_i4b)==zero
            sub(1_i4b,1_i4b)  = zero
            sup1(1_i4b,1_i4b) = zero
            sup2(1_i4b,1_i4b) = zero
!
        end if
!
        return
!
    end if
!    
!   INITIALIZE TOLERANCE, IMPLICIT SCALING VECTOR AND WORK VARIABLES.
!    
    tl = lamch( 'u' )
    if ( present(tol) ) tl = max( tl, tol )
!    
    diag(:m,:n) = spread(       d(:n), dim=1, ncopies=m ) +     &
                  spread( -eigval(:m), dim=2, ncopies=n )
!
    sup1(:m,:n) = spread( e(:n), dim=1, ncopies=m )
    sup2(:m,:n) = zero
!    
    sub(1_i4b,1_i4b) = abs( e(1_i4b) )
    sub(1_i4b,2_i4b:n-1_i4b) = abs(e(2_i4b:n-1_i4b)) + abs(e(:n-2_i4b))
    sub(1_i4b,n) = abs( e(n-1_i4b) )
!
    scal(:m,:n) = abs( diag(:m,:n) ) + spread( sub(1_i4b,:n), dim=1, ncopies=m )
!    
    perm(:m,:n) = false
!   
!   TRIANGULATION.  
!    
    ekp1 = e(1_i4b)
!
    do k = 1_i4b, n-1_i4b
!
        where( diag(:m,k)/=zero )
            sup2(:m,n) = abs( diag(:m,k) )/scal(:m,k)
        elsewhere
            sup2(:m,n) = zero
        end where
!
        ek = ekp1
        ekp1 = e(k+1_i4b)
!
        if ( ek/=zero ) then
!
            sup1(:m,n) = abs( ek )/scal(:m,k+1_i4b)
!
            where( sup1(:m,n)<=sup2(:m,n) )
!   
!               NO COLUMNS INTERCHANGE REQUIRED, ELIMINATE sub(:m,k)=e(k) .  
!    
                sub(:m,k) = ek/diag(:m,k)
                diag(:m,k+1_i4b) = diag(:m,k+1_i4b) - sub(:m,k)*sup1(:m,k)
!
            elsewhere
!   
!               INTERCHANGE COLUMNS k AND k+1, ELIMINATE sub(:m,k)=diag(:m,k) .  
!    
                perm(:m,k) = true
                sup2(:m,n) = sup1(:m,n)
                scal(:m,k+1_i4b) = scal(:m,k)
                sub(:m,k) = diag(:m,k)/ek
                diag(:m,k) = ek
                sub(:m,n) = diag(:m,k+1_i4b)
                diag(:m,k+1_i4b) = sup1(:m,k) - sub(:m,k)*sub(:m,n)
                sup2(:m,k) = ekp1
                sup1(:m,k+1_i4b) = -sub(:m,k)*ekp1
                sup1(:m,k) = sub(:m,n)
!
            end where
!
        end if
!
        where( sup2(:m,n)<=tl .and. .not. perm(:m,n) ) perm(:m,n) = true
!
    end do
!
    where( abs( diag(:m,n) )<=scal(:m,n)*tl .and. .not. perm(:m,n) ) perm(:m,n) = true
!
    sup1(:m,n) = zero
    sup2(:m,n-1_i4b:n) = zero
    sub(:m,n) = zero
!
!
! END OF SUBROUTINE trid_cmp_rm
! _____________________________
!
    end subroutine trid_cmp_rm
!
! =========================================================================================
!
    subroutine trid_cmp_rv( d, e, eigval, sub, diag, sup1, sup2, perm, tol )
!
! Purpose
! _______
!
!  TRID_CMP factorizes the symmetric matrix (T - EIGVAL * I), where 
!  T is an n-by-n symmetric tridiagonal matrix and EIGVAL is a scalar, as
!
!                         T - EIGVAL * I = P * L * U
!
!  where P is a permutation matrix, L is a unit lower tridiagonal matrix
!  with at most one non-zero sub-diagonal elements per column and U is
!  an upper triangular matrix with at most two non-zero super-diagonal
!  elements per column.
!
!  The factorization is obtained by Gaussian elimination with partial pivoting
!  and implicit row scaling.
!
!  The parameter EIGVAL is included in the routine so that TRID_CMP may
!  be used to obtain eigenvectors of T by inverse iteration.
!
!
! Arguments
! _________
!
!    D           (INPUT) real(stnd), dimension(:)
!                On entry, the diagonal elements of the symmetric tridiagonal matrix T.
!
!    E           (INPUT) real(stnd), dimension(:)
!                On entry, the n-1 subdiagonal elements of the symmetric tridiagonal
!                matrix T and E(n) is arbitrary.
!
!                The size of E must verify:  size( E ) = size( D ) = n .
!
!    EIGVAL      (INPUT) real(stnd)
!                On entry, an eigenvalue of the symmetric tridiagonal matrix.
!
!    SUB         (OUTPUT) real(stnd), dimension(:)
!                On exit, SUB(:n-1) contains the n-1 subdiagonal elements of the
!                lower triangular matrix L of the factorization of T - EIGVAL * I,
!                SUB(n) is arbitrary .
!
!                The size of SUB must verify: size( SUB ) = size( D ) = n .
!
!    DIAG        (OUTPUT) real(stnd), dimension(:)
!                On exit, DIAG(:) contains the n diagonal elements of the
!                upper triangular matrix U of the factorization of T - EIGVAL * I.
!               
!                 The size of DIAG must verify: size( DIAG ) = size( D ) = n .
!
!    SUP1        (OUTPUT) real(stnd), dimension(:)
!                On exit, SUP1(:n-1) contains the n-1 superdiagonal elements of the
!                upper triangular matrix U of the factorization of T - EIGVAL * I,
!                SUP1(n) is arbitrary.
!
!                The size of SUP1 must verify: size( SUP1 ) = size( D ) = n .
!
!    SUP2        (OUTPUT) real(stnd), dimension(:)
!                On exit, SUP2(:n-2) contains the n-2 second superdiagonal elements of
!                the upper triangular matrix U of the factorization of T - EIGVAL * I,
!                SUP2(n-1:n) is arbitrary .
!
!                The size of SUP2 must verify: size( SUP2 ) = size( D ) = n .
!
!    PERM        (OUTPUT) logical(lgl), dimension(:)
!                On exit, PERM(:n-1)  contains details of the permutation matrix P(j).
!                If an interchange occurred at the kth step of the elimination in the
!                factorization of (T - EIGVAL(j) * I), then PERM(k) = true, otherwise
!                PERM(k) = false.
!                The element PERM(n) is set to true if there is an integer l such that
!
!                  abs( u(l,l) ).le. norm( (T - EIGVAL * I)(l) ) * TOL,
!
!                where norm( A(l) ) denotes the sum of the absolute values of
!                the lth row of the matrix A. If no such l exists then PERM(n)
!                is returned as false. If PERM(n) is returned as true, then a
!                diagonal element of U is small, indicating that
!                (T - EIGVAL * I) is singular or nearly singular.
!
!                The size of PERM must verify: size( PERM ) = size( D ) = n .
!
!    TOL         (INPUT, OPTIONAL)  real(stnd)
!                On entry, a relative tolerance used to indicate whether or
!                not the matrix (T - EIGVAL * I) is nearly singular. TOL should
!                normally be choose as approximately the largest relative error
!                in the elements of T. For example, if the elements of T are
!                correct to about 4 significant figures, then TOL should be
!                set to about 5 * 10**(-4).
!
!                If TOL is supplied as less than eps, where eps is the 
!                relative machine precision, then the value eps is used
!                in place of TOL.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the routine  DLAGTF in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,    only : lamch
    use Utilities,        only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),  dimension(:)    :: d, e
    real(stnd),  intent(in)                   :: eigval
    real(stnd),  intent(out), dimension(:)    :: sub, diag, sup1, sup2
    real(stnd),  intent(in),  optional        :: tol
!
    logical(lgl), intent(out), dimension(:) :: perm
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: tl, ek, ekp1, tmp1, tmp2, scal(size(d))
!
    integer(i4b) :: n, k
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='trid_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( (/ int(size(d),i4b)       ,   &
                       int(size(e),i4b)       ,   &
                       int(size(sub),i4b)     ,   &
                       int(size(diag),i4b)    ,   &
                       int(size(sup1),i4b)    ,   &
                       int(size(sup2),i4b)    ,   &
                       int(size(perm),i4b) /) ,   &
                       name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            diag(1_i4b) = d(1_i4b) - eigval
            perm(1_i4b) = diag(1_i4b)==zero
            sub(1_i4b)  = zero
            sup1(1_i4b) = zero
            sup2(1_i4b) = zero
!
        end if
!
        return
!
    end if
!    
!   INITIALIZE TOLERANCE, IMPLICIT SCALING VECTOR AND WORK VARIABLES.
!    
    tl = lamch( 'u' )
    if ( present(tol) ) tl = max( tl, tol )
!    
    diag(:n) = d(:n) - eigval
!
    sup1(:n) = e(:n)
    sup2(:n) = zero
!    
    sub(1_i4b) = abs( e(1_i4b) )
    sub(2_i4b:n-1_i4b) = abs(e(2_i4b:n-1_i4b)) + abs(e(:n-2_i4b))
    sub(n) = abs( e(n-1_i4b) )
!
    scal(:n) = abs( diag(:n) ) + sub(:n)
!    
    perm(:n) = false
!   
!   TRIANGULATION.  
!    
    ekp1 = e(1_i4b)
!
    do k = 1_i4b, n-1_i4b
!
        if ( diag(k)/=zero ) then
            tmp1 = abs( diag(k) )/scal(k)
        else
            tmp1 = zero
        end if
!
        ek = ekp1
        ekp1 = e(k+1_i4b)
!
        if ( ek/=zero ) then
!
            tmp2 = abs( ek )/scal(k+1_i4b)
!
            if ( tmp2<=tmp1 ) then
!   
!               NO INTERCHANGE REQUIRED, ELIMINATE sub(k)=e(k) .  
!    
                sub(k) = ek/diag(k)
                diag(k+1_i4b) = diag(k+1_i4b) - sub(k)*sup1(k)
!
            else
!   
!               INTERCHANGE ELEMENTS k AND k+1, ELIMINATE sub(k)=diag(k) .  
!    
                perm(k) = true
                tmp1 = tmp2
                scal(k+1_i4b) = scal(k)
                sub(k) = diag(k)/ek
                diag(k) = ek
                tmp2 = diag(k+1_i4b)

                diag(k+1_i4b) = sup1(k) - sub(k)*tmp2
                sup2(k) = ekp1
                sup1(k+1_i4b) = -sub(k)*ekp1
                sup1(k) = tmp2
!
            end if
!
        end if
!
        if ( tmp1<=tl .and. .not. perm(n) ) perm(n) = true
!
    end do
!
    if ( abs( diag(n) )<=scal(n)*tl .and. .not. perm(n) ) perm(n) = true
!
    sup1(n) = zero
    sup2(n-1_i4b:n) = zero
    sub(n) = zero
!
!
! END OF SUBROUTINE trid_cmp_rv
! _____________________________
!
    end subroutine trid_cmp_rv
!
! =========================================================================================
!
    subroutine trid_cmp2_rm( d, e, eigval, sub, diag, sup1, sup2, perm )
!
! Purpose
! _______
!
!   TRID_CMP2 factorizes symmetric matrices of the form (T - EIGVAL(j) * I), where 
!   T is an n-by-n symmetric tridiagonal matrix and EIGVAL(j) is a scalar, as
!
!      T - EIGVAL(j) * I = P(j) * L(j) * U(j),    for j=1, SIZE(EIGVAL)
!
!   where P(j) is a permutation matrix, L(j) is a unit lower tridiagonal matrix
!   with at most one non-zero sub-diagonal elements per column and U(j) is
!   an upper triangular matrix with at most two non-zero super-diagonal
!   elements per column.
!
!   The factorizations, for j=1, SIZE(EIGVAL), are obtained by Gaussian elimination
!   with partial pivoting and row interchanges.
!
!   The parameters EIGVAL are included in the routine so that TRID_CMP2 may
!   be used to obtain eigenvectors of T by inverse iteration.
!
!
! Arguments
! _________
!
!    D           (INPUT) real(stnd), dimension(:)
!                On entry, the diagonal elements of the symmetric tridiagonal matrix T.
!
!    E           (INPUT) real(stnd), dimension(:)
!                On entry, the n-1 subdiagonal elements of the symmetric tridiagonal
!                matrix T and E(n) is arbitrary .
!
!                The size of E must verify:  size( E ) = size( D ) = n .
!
!    EIGVAL      (INPUT) real(stnd), dimension(:)
!                On entry, selected eigenvalues of the symmetric tridiagonal matrix.
!
!                The size of EIGVAL must verify: size( EIGVAL ) <= size( D ) = n .
!
!    SUB         (OUTPUT) real(stnd), dimension(:,:)
!                On exit, SUB(j,:n-1) contains the n-1 subdiagonal elements of the
!                lower triangular matrix L(j) of the factorization of T - EIGVAL(j) * I,
!                for j=1, SIZE(EIGVAL). SUB(:,n) is arbitrary .
!
!                The shape of SUB must verify: 
!
!                - size( SUB, 1 ) = size( EIGVAL ) ;
!                - size( SUB, 2 ) = size( D ) = n  .
!
!    DIAG        (OUTPUT) real(stnd), dimension(:,:)
!                On exit, DIAG(j,:) contains the n diagonal elements of the
!                upper triangular matrix U(j) of the factorization of T - EIGVAL(j) * I,
!                for j=1, SIZE(EIGVAL).
!               
!                The shape of DIAG must verify: 
!
!                - size( DIAG, 1 ) = size( EIGVAL ) ;
!                - size( DIAG, 2 ) = size( D ) = n  .
!
!    SUP1        (OUTPUT) real(stnd), dimension(:,:)
!                On exit, SUP1(j,:n-1) contains the n-1 superdiagonal elements of the
!                upper triangular matrix U(j) of the factorization of T - EIGVAL(j) * I,
!                for j=1, SIZE(EIGVAL). SUP1(:,n) is arbitrary .
!
!                The shape of SUP1 must verify: 
!
!                - size( SUP1, 1 ) = size( EIGVAL ) ;
!                - size( SUP1, 2 ) = size( D ) = n  .
!
!    SUP2        (OUTPUT) real(stnd), dimension(:,:)
!                On exit, SUP2(j,:n-2) contains the n-2 second superdiagonal elements of
!                the upper triangular matrix U(j) of the factorization of T - EIGVAL(j) * I,
!                for j=1, SIZE(EIGVAL). SUP2(:,n-1:n) is arbitrary .
!
!                The shape of SUP2 must verify: 
!
!                - size( SUP2, 1 ) = size( EIGVAL ) ;
!                - size( SUP2, 2 ) = size( D ) = n  .
!
!    PERM        (OUTPUT) logical(lgl), dimension(:,:)
!                On exit, PERM(j,:n-1)  contains details of the permutation matrix P(j).
!                If an interchange occurred at the kth step of the elimination in the
!                factorization of (T - EIGVAL(j) * I), then PERM(j,k) = true, otherwise
!                PERM(j,k) = false. PERM(:,n) is arbitrary .
!
!                The shape of PERM must verify: 
!
!                - size( PERM, 1 ) = size( EIGVAL ) ;
!                - size( PERM, 2 ) = size( D ) = n  .
!
!
! Further Details
! _______________
!
!   TRID_CMP2 is a simplified version of TRID_CMP.
!   This subroutine is adapted from the routine  DGTTRF in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq, assert
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),  dimension(:)    :: d, e, eigval
    real(stnd),  intent(out), dimension(:,:)  :: sub, diag, sup1, sup2
!
    logical(lgl), intent(out), dimension(:,:) :: perm
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: ek, ekp1
!
    integer(i4b) :: n, m, k
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='trid_cmp2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( (/ int(size(d),i4b)         ,   &
                       int(size(e),i4b)         ,   &
                       int(size(sub,2),i4b)     ,   &
                       int(size(diag,2),i4b)    ,   &
                       int(size(sup1,2),i4b)    ,   &
                       int(size(sup2,2),i4b)    ,   &
                       int(size(perm,2),i4b) /) ,   &
                       name_proc )
!
    m =  assert_eq( (/ int(size(eigval),i4b)    ,   &
                       int(size(sub,1),i4b)     ,   &
                       int(size(diag,1),i4b)    ,   &
                       int(size(sup1,1),i4b)    ,   &
                       int(size(sup2,1),i4b)    ,   &
                       int(size(perm,1),i4b) /) ,   &
                       name_proc )
!
    call assert( logical( m<=n, lgl ), name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b .or. m<=0_i4b) then
!
        if ( n==1_i4b .and. m==1_i4b ) then
!
            diag(1_i4b,1_i4b) = d(1_i4b) - eigval(1_i4b)
            perm(1_i4b,1_i4b) = diag(1_i4b,1_i4b)==zero
            sub(1_i4b,1_i4b)  = zero
            sup1(1_i4b,1_i4b) = zero
            sup2(1_i4b,1_i4b) = zero
!
        end if
!
        return
!
    end if
!    
!   INITIALIZE WORK VARIABLES.
!    
    diag(:m,:n) = spread(       d(:n), dim=1, ncopies=m ) +     &
                  spread( -eigval(:m), dim=2, ncopies=n )
!
    sup1(:m,:n) = spread( e(:n), dim=1, ncopies=m )
    sup2(:m,:n) = zero
!
    perm(:m,:n) = false
!   
!   TRIANGULATION.  
!    
    ekp1 = e(1_i4b)
!
    do k = 1_i4b, n-1_i4b
!
        ek = ekp1
        ekp1 = e(k+1_i4b)
!
        if ( ek==zero ) cycle
!   
        where( abs( ek )<=abs( diag(:m,k) ) )
!   
!           NO COLUMNS INTERCHANGE REQUIRED, ELIMINATE sub(:m,k)=ek .  
!    
            sub(:m,k) = ek/diag(:m,k)
            diag(:m,k+1_i4b) = diag(:m,k+1_i4b) - sub(:m,k)*sup1(:m,k)
!
        elsewhere
!   
!           INTERCHANGE COLUMNS k AND k+1, ELIMINATE sub(:m,k)=diag(:m,k) .  
!    
            perm(:m,k) = true
            sub(:m,k) = diag(:m,k)/ek
            diag(:m,k) = ek
            sub(:m,n) = diag(:m,k+1_i4b)
            diag(:m,k+1_i4b) = sup1(:m,k) - sub(:m,k)*sub(:m,n)
            sup2(:m,k) = ekp1
            sup1(:m,k+1_i4b) = -sub(:m,k)*ekp1
            sup1(:m,k) = sub(:m,n)
!
        end where
!
    end do
!
    sup1(:m,n) = zero
    sup2(:m,n-1_i4b:n) = zero
    sub(:m,n) = zero
!
!
! END OF SUBROUTINE trid_cmp2_rm
! ______________________________
!
    end subroutine trid_cmp2_rm
!
! =========================================================================================
!
    subroutine trid_cmp2_rv( d, e, eigval, sub, diag, sup1, sup2, perm )
!
! Purpose
! _______
!
!   TRID_CMP2 factorizes the symmetric matrix (T - EIGVAL * I), where 
!   T is an n by n symmetric tridiagonal matrix and EIGVAL is a scalar, as
!
!                         T - EIGVAL * I = P * L * U
!
!   where P is a permutation matrix, L is a unit lower tridiagonal matrix
!   with at most one non-zero sub-diagonal elements per column and U is
!   an upper triangular matrix with at most two non-zero super-diagonal
!   elements per column.
!
!   The factorization is obtained by Gaussian elimination with partial pivoting
!   and row interchanges.
!
!   The parameter EIGVAL is included in the routine so that TRID_CMP2 may
!   be used to obtain eigenvectors of T by inverse iteration.
!
!
! Arguments
! _________
!
!    D           (INPUT) real(stnd), dimension(:)
!                On entry, the diagonal elements of the symmetric tridiagonal matrix T.
!
!    E           (INPUT) real(stnd), dimension(:)
!                On entry, the n-1 subdiagonal elements of the symmetric tridiagonal
!                matrix T and E(n) is arbitrary .
!
!                The size of E must verify:  size( E ) = size( D ) = n .
!
!    EIGVAL      (INPUT) real(stnd)
!                On entry, an eigenvalue of the symmetric tridiagonal matrix.
!
!    SUB         (OUTPUT) real(stnd), dimension(:)
!                On exit, SUB(:n-1) contains the n-1 subdiagonal elements of the
!                lower triangular matrix L of the factorization of T - EIGVAL * I,
!                SUB(n) is arbitrary .
!
!                The size of SUB must verify: size( SUB ) = size( D ) = n .
!
!    DIAG        (OUTPUT) real(stnd), dimension(:)
!                On exit, DIAG(:) contains the n diagonal elements of the
!                upper triangular matrix U of the factorization of T - EIGVAL * I.
!               
!                The size of DIAG must verify: size( DIAG ) = size( D ) = n .
!
!    SUP1        (OUTPUT) real(stnd), dimension(:)
!                On exit, SUP1(:n-1) contains the n-1 superdiagonal elements of the
!                upper triangular matrix U of the factorization of T - EIGVAL * I,
!                SUP1(n) is arbitrary .
!
!                The size of SUP1 must verify: size( SUP1 ) = size( D ) = n .
!
!    SUP2        (OUTPUT) real(stnd), dimension(:)
!                On exit, SUP2(:n-2) contains the n-2 second superdiagonal elements of
!                the upper triangular matrix U of the factorization of T - EIGVAL * I,
!                SUP2(n-1:n) is arbitrary .
!
!                The size of SUP2 must verify: size( SUP2 ) = size( D ) = n .
!
!    PERM        (OUTPUT) logical(lgl), dimension(:)
!                On exit, PERM(:n-1)  contains details of the permutation matrix P(j).
!                If an interchange occurred at the kth step of the elimination in the
!                factorization of (T - EIGVAL(j) * I), then PERM(k) = true, otherwise
!                PERM(k) = false. PERM(n) is arbitrary .
!
!                The size of PERM must verify: size( PERM ) = size( D ) = n .
!
!
! Further Details
! _______________
!
!   TRID_CMP2 is a simplified version of TRID_CMP.
!   This subroutine is adapted from the routine  DGTTRF in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),  dimension(:)  :: d, e
    real(stnd),  intent(in)                 :: eigval
    real(stnd),  intent(out), dimension(:)  :: sub, diag, sup1, sup2
!
    logical(lgl), intent(out), dimension(:) :: perm
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: ek, ekp1, tmp
!
    integer(i4b) :: n, k
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='trid_cmp2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( (/ int(size(d),i4b)       ,   &
                       int(size(e),i4b)       ,   &
                       int(size(sub),i4b)     ,   &
                       int(size(diag),i4b)    ,   &
                       int(size(sup1),i4b)    ,   &
                       int(size(sup2),i4b)    ,   &
                       int(size(perm),i4b) /) ,   &
                       name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            diag(1_i4b) = d(1_i4b) - eigval
            perm(1_i4b) = diag(1_i4b)==zero
            sub(1_i4b)  = zero
            sup1(1_i4b) = zero
            sup2(1_i4b) = zero
!
        end if
!
        return
!
    end if
!    
!   INITIALIZE WORK VARIABLES.
!    
    diag(:n) = d(:n) - eigval
    sup1(:n) = e(:n)
    sup2(:n) = zero
    perm(:n) = false
!   
!   TRIANGULATION.  
!    
    ekp1 = e(1_i4b)
!
    do k = 1_i4b, n-1_i4b
!
        ek = ekp1
        ekp1 = e(k+1_i4b)
!
        if ( ek==zero ) cycle
!   
        if ( abs( ek )<=abs( diag(k) ) ) then
!   
!           NO INTERCHANGE REQUIRED, ELIMINATE sub(k)=ek .  
!    
            sub(k) = ek/diag(k)
            diag(k+1_i4b) = diag(k+1_i4b) - sub(k)*sup1(k)
!
        else
!   
!           INTERCHANGE ELEMENTS k AND k+1, ELIMINATE sub(k)=diag(k) .  
!    
            perm(k) = true
            sub(k) = diag(k)/ek
            diag(k) = ek
            tmp = diag(k+1_i4b)
            diag(k+1_i4b) = sup1(k) - sub(k)*tmp
            sup2(k) = ekp1
            sup1(k+1_i4b) = -sub(k)*ekp1
            sup1(k) = tmp
!
        end if
!
    end do
!
    sup1(n) = zero
    sup2(n-1_i4b:n) = zero
    sub(n) = zero
!
!
! END OF SUBROUTINE trid_cmp2_rv
! ______________________________
!
    end subroutine trid_cmp2_rv
!
! =========================================================================================
!
    subroutine trid_solve_rm( sub, diag, sup1, sup2, perm, y )
!
! Purpose
! _______
!
!   TRID_SOLVE may be used to solve systems of equations of the form
!
!         x(j,:) * (T - EIGVAL(j) * I) = scale(j) * y(j,:),    for j=1, SIZE(EIGVAL)
!
!   , where T is an n by n symmetric tridiagonal matrix, EIGVAL(j) and scale(j) are scalars,
!   for x(j,:) for j=1, SIZE(EIGVAL), following the factorization of (T - EIGVAL(j) * I) by 
!   TRID_CMP or TRID_CMP2 as
!
!        T - EIGVAL(j) * I = P(j) * L(j) * U(j),    for j=1, SIZE(EIGVAL)
!
!   where P(j) is a permutation matrix, L(j) is a unit lower tridiagonal matrix
!   with at most one non-zero sub-diagonal elements per column and U(j) is
!   an upper triangular matrix with at most two non-zero super-diagonal
!   elements per column.
!
!   The matrices (T - EIGVAL(j) * I) are assumed to be ill-conditioned, and frequent rescalings
!   are carried out in order to avoid overflow. However, no test for singularity
!   or near-singularity is included in this routine. Such tests must be performed
!   before calling this routine. The scalars, scale(j) for j=1, SIZE(EIGVAL), are not
!   output by this routine since this routine being intended for use in applications such
!   as inverse iteration.
!
!
! Arguments
! _________
!
!    SUB         (INPUT) real(stnd), dimension(:,:)
!                On entry, SUB(j,:n-1) contains the n-1 subdiagonal elements of the
!                lower triangular matrix L(j) of the factorization of T - EIGVAL(j) * I,
!                for j=1, SIZE(EIGVAL). SUB(:,n) is arbitrary .
!
!                The shape of SUB must verify: 
!
!                - size( SUB, 1 ) = size( Y, 1 ) = size( EIGVAL ) ;
!                - size( SUB, 2 ) = size( Y, 2 ) = n .
!
!    DIAG        (INPUT) real(stnd), dimension(:,:)
!                On entry, DIAG(j,:) contains the n diagonal elements of the
!                upper triangular matrix U(j) of the factorization of T - EIGVAL(j) * I,
!                for j=1, SIZE(EIGVAL).
!               
!                The shape of DIAG must verify: 
!
!                - size( DIAG, 1 ) = size( Y, 1 ) = size( EIGVAL ) ;
!                - size( DIAG, 2 ) = size( Y, 2 ) = n .
!
!    SUP1        (INPUT) real(stnd), dimension(:,:)
!                On entry, SUP1(j,:n-1) contains the n-1 superdiagonal elements of the
!                upper triangular matrix U(j) of the factorization of T - EIGVAL(j) * I,
!                for j=1, SIZE(EIGVAL). SUP1(:,n) is arbitrary.
!
!                The shape of SUP1 must verify: 
!
!                - size( SUP1, 1 ) = size( Y, 1 ) = size( EIGVAL ) ;
!                - size( SUP1, 2 ) = size( Y, 2 ) = n .
!
!    SUP2        (INPUT) real(stnd), dimension(:,:)
!                On entry, SUP2(j,:n-2) contains the n-2 second superdiagonal elements of
!                the upper triangular matrix U(j) of the factorization of T - EIGVAL(j) * I,
!                for j=1, SIZE(EIGVAL). SUP2(:,n-1:n) is arbitrary.
!
!                The shape of SUP2 must verify: 
!
!                - size( SUP2, 1 ) = size( Y, 1 ) = size( EIGVAL ) ;
!                - size( SUP2, 2 ) = size( Y, 2 ) = n .
!
!    PERM        (INPUT) logical(lgl), dimension(:,:)
!                On entry, PERM(j,:n-1)  contains details of the permutation matrix P(j).
!                If an interchange occurred at the kth step of the elimination in the
!                factorization of (T - EIGVAL(j) * I), then PERM(j,k) = true, otherwise
!                PERM(j,k) = false. PERM(:,n) is arbitrary .
!
!                The shape of PERM must verify: 
!
!                - size( PERM, 1 ) = size( Y, 1 ) = size( EIGVAL ) ;
!                - size( PERM, 2 ) = size( Y, 2 ) = n .
!
!    Y           (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the right hand side matrix y.
!                On exit, Y is overwritten the solution matrix x.
!
!                The shape of Y must verify: 
!
!                - size( Y, 1 ) = size( EIGVAL ) ;
!                - size( Y, 2 ) = n .
!
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the routine  DLAGTS in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq, assert
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),    dimension(:,:) :: sub, diag, sup1, sup2
    real(stnd),  intent(inout), dimension(:,:) :: y
!
    logical(lgl), intent(in), dimension(:,:) :: perm
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(sub,1))   :: tmp1, tmp2
!
    integer(i4b) :: n, m, k
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='trid_solve'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!    TEST THE ARGUMENTS.
!
    n =  assert_eq( (/ int(size(y,2),i4b)       ,   &
                       int(size(sub,2),i4b)     ,   &
                       int(size(diag,2),i4b)    ,   &
                       int(size(sup1,2),i4b)    ,   &
                       int(size(sup2,2),i4b)    ,   &
                       int(size(perm,2),i4b) /) ,   &
                       name_proc )
!
    m =  assert_eq( (/ int(size(y,1),i4b)       ,   &
                       int(size(sub,1),i4b)     ,   &
                       int(size(diag,1),i4b)    ,   &
                       int(size(sup1,1),i4b)    ,   &
                       int(size(sup2,1),i4b)    ,   &
                       int(size(perm,1),i4b) /) ,   &
                       name_proc )
!
    call assert( logical( m<=n, lgl ), name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. m<=0_i4b) return
!    
!   SOLVE L*x = b
!    
    do k = 2_i4b, n
!
        where( perm(:m,k-1_i4b) )
!
            tmp1(:m) = y(:m,k-1_i4b)
            y(:m,k-1_i4b) = y(:m,k)
            y(:m,k) = tmp1(:m) - sub(:m,k-1_i4b)*y(:m,k)
!
        elsewhere
!
            y(:m,k) = y(:m,k) - sub(:m,k-1_i4b)*y(:m,k-1_i4b)
!
        end where
!
    end do
!   
!   SOLVE U*x = b WITH SCALING.
!    
    where( abs(y(:m,n))>abs(diag(:m,n)) )
!
        tmp1(:m) = diag(:m,n)/y(:m,n)
        y(:m,n) = one
!
    elsewhere
!
        tmp1(:m) = one
        y(:m,n) = y(:m,n)/diag(:m,n)
!
    end where
!
    y(:m,:n-1_i4b) = spread( tmp1(:m), dim=2, ncopies=n-1_i4b)*y(:m,:n-1_i4b)
!    
!    y(:m,n) = y(:m,n)/diag(:m,n)
!    
    if ( n>1_i4b ) then
!
        tmp1(:m) = y(:m,n-1_i4b) - sup1(:m,n-1_i4b)*y(:m,n)
!
        where( abs(tmp1(:m))>abs(diag(:m,n-1_i4b)) )
!
            tmp1(:m) = diag(:m,n-1_i4b)/tmp1(:m)
            y(:m,n-1_i4b) = one
!
        elsewhere
!
            y(:m,n-1_i4b) = tmp1(:m)/diag(:m,n-1_i4b)
            tmp1(:m) = one
!
        end where
!
        y(:m,1_i4b:n-2_i4b) = spread( tmp1(:m), dim=2, ncopies=n-2_i4b)*y(:m,1_i4b:n-2_i4b)
        y(:m,n) = tmp1(:m)*y(:m,n)
!
!        if ( n>1_i4b )                      &
!        y(:m,n-1_i4b) = ( y(:m,n-1_i4b) - sup1(:m,n-1_i4b)*y(:m,n) )/diag(:m,n-1_i4b)
!    
        do k = n-2_i4b, 1_i4b, -1_i4b
!
            tmp1(:m) = y(:m,k) - sup1(:m,k)*y(:m,k+1_i4b) - sup2(:m,k)*y(:m,k+2_i4b)
!
            where( abs(tmp1(:m))>abs(diag(:m,k)) )
!
                tmp1(:m) = diag(:m,k)/tmp1(:m)
                tmp2(:m) = one
!
            elsewhere
!
                tmp2(:m) = tmp1(:m)/diag(:m,k)
                tmp1(:m) = one
!
            end where
!
            y(:m,:n) = spread( tmp1(:m), dim=2, ncopies=n)*y(:m,:n)
            y(:m,k) = tmp2(:m)
!
        end do
!
!       do k = n-2_i4b, 1_i4b, -1_i4b
!            y(:m,k) = ( y(:m,k) - sup1(:m,k)*y(:m,k+1_i4b) - sup2(:m,k)*y(:m,k+2_i4b))/diag(:m,k)
!       end do
!
    end if
!
!
! END OF SUBROUTINE trid_solve_rm
! _______________________________
!
    end subroutine trid_solve_rm
!
! =========================================================================================
!
    subroutine trid_solve_rv( sub, diag, sup1, sup2, perm, y )
!
! Purpose
! _______
!
!   TRID_SOLVE may be used to solve the system of equations
!
!                x(:) * (T - EIGVAL * I) = scale * y(:)
!
!   , where T is an n by n symmetric tridiagonal matrix, EIGVAL and scale are scalars,
!   for x(:), following the factorization of (T - EIGVAL * I) by TRID_CMP or TRID_CMP2 as
!
!                 T - EIGVAL * I = P * L * U
!
!   where P is a permutation matrix, L is a unit lower tridiagonal matrix
!   with at most one non-zero sub-diagonal elements per column and U is
!   an upper triangular matrix with at most two non-zero super-diagonal
!   elements per column.
!
!   The matrix (T - EIGVAL * I) is assumed to be ill-conditioned, and frequent rescalings
!   are carried out in order to avoid overflow. However, no test for singularity
!   or near-singularity is included in this routine. Such tests must be performed
!   before calling this routine. The scalar, scale, is not output by this routine 
!   since this routine being intended for use in applications such
!   as inverse iteration.
!
!
! Arguments
! _________
!
!    SUB         (INPUT) real(stnd), dimension(:)
!                On entry, SUB(:n-1) contains the n-1 subdiagonal elements of the
!                lower triangular matrix L of the factorization of T - EIGVAL * I,
!                SUB(n) is arbitrary .
!
!                The size of SUB must verify:  size( SUB ) = size( Y ) = n .
!                              
!
!    DIAG        (INPUT) real(stnd), dimension(:)
!                On entry, DIAG(:) contains the n diagonal elements of the
!                upper triangular matrix U of the factorization of T - EIGVAL * I.
!               
!                The shape of DIAG must verify: size( DIAG ) = size( Y ) = n .
!
!    SUP1        (INPUT) real(stnd), dimension(:)
!                On entry, SUP1(:n-1) contains the n-1 superdiagonal elements of the
!                upper triangular matrix U of the factorization of T - EIGVAL * I,
!                SUP1(n) is arbitrary.
!
!                The shape of SUP1 must verify: size( SUP1 ) = size( Y ) = n .
!
!    SUP2        (INPUT) real(stnd), dimension(:)
!                On entry, SUP2(:n-2) contains the n-2 second superdiagonal elements of
!                the upper triangular matrix U of the factorization of T - EIGVAL * I,
!                SUP2(n-1:n) is arbitrary.
!
!                The shape of SUP2 must verify: size( SUP2 ) = size( Y ) = n .
!
!    PERM        (INPUT) logical(lgl), dimension(:)
!                On entry, PERM(:n-1)  contains details of the permutation matrix P.
!                If an interchange occurred at the kth step of the elimination in the
!                factorization of (T - EIGVAL * I), then PERM(k) = true, otherwise
!                PERM(k) = false. PERM(n) is arbitrary .
!
!                The shape of PERM must verify:  size( PERM ) = size( Y ) = n .
!
!    Y           (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the right hand side vector y.
!                On exit, Y is overwritten the solution vector x.
!
!                The shape of Y must verify: size( Y ) = n .
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the routine  DLAGTS in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),    dimension(:) :: sub, diag, sup1, sup2
    real(stnd),  intent(inout), dimension(:) :: y
!
    logical(lgl), intent(in), dimension(:) :: perm
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: tmp
!
    integer(i4b) :: n, k
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='trid_solve'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( (/ int(size(y),i4b)       ,   &
                       int(size(sub),i4b)     ,   &
                       int(size(diag),i4b)    ,   &
                       int(size(sup1),i4b)    ,   &
                       int(size(sup2),i4b)    ,   &
                       int(size(perm),i4b) /) ,   &
                       name_proc )
!
!    QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!    
!   SOLVE L*x = b
!    
    do k = 2_i4b, n
!
        if ( perm(k-1_i4b) ) then
!
            tmp = y(k-1_i4b)
            y(k-1_i4b) = y(k)
            y(k) = tmp - sub(k-1_i4b)*y(k)
!
        else
!
            y(k) = y(k) - sub(k-1_i4b)*y(k-1_i4b)
!
        end if
!
    end do
!   
!   SOLVE U*x = b WITH SCALING.
!    
    if ( abs(y(n))>abs(diag(n)) ) then
!
        tmp = diag(n)/y(n)
        y(:n-1_i4b) = tmp*y(:n-1_i4b)
        y(n) = one
!
    else
!
        y(n) = y(n)/diag(n)
!
    end if
!    
!    y(n) = y(n)/diag(n)
!    
    if ( n>1_i4b ) then
!
        tmp = y(n-1_i4b) - sup1(n-1_i4b)*y(n)
!
        if ( abs(tmp)>abs(diag(n-1_i4b)) ) then
!
            tmp = diag(n-1_i4b)/tmp
            y(:n) = tmp*y(:n)
            y(n-1_i4b) = one
!
        else
!
            y(n-1_i4b) = tmp/diag(n-1_i4b)
!
        end if
!
!        if ( n>1_i4b )                      &
!        y(n-1_i4b) = ( y(n-1_i4b) - sup1(n-1_i4b)*y(n) )/diag(n-1_i4b)
!    
        do k = n-2_i4b, 1_i4b, -1_i4b
!
            tmp = y(k) - sup1(k)*y(k+1_i4b) - sup2(k)*y(k+2_i4b)
!
            if ( abs(tmp)>abs(diag(k)) ) then
!
                tmp = diag(k)/tmp
                y(:n) = tmp*y(:n)
                y(k) = one
!
            else
!
                y(k) = tmp/diag(k)
!
            end if
!
        end do
!
!        do k = n-2_i4b, 1_i4b, -1_i4b
!            y(k) = ( y(k) - sup1(k)*y(k+1_i4b) - sup2(k)*y(k+2_i4b))/diag(k)
!        end do
!
    end if
!
!
! END OF SUBROUTINE trid_solve_rv
! _______________________________
!
    end subroutine trid_solve_rv
!
! =========================================================================================
!
    subroutine trid_inviter_rv( d, e, eigval, eigvec, failure, maxiter, scaling, initvec )
!
! Purpose
! _______
!
!   TRID_INVITER computes an eigenvector of a real n-by-n symmetric tridiagonal
!   matrix T corresponding to a specified eigenvalue, by combining Fernando's
!   method for computing an eigenvector of a real n-by-n symmetric tridiagonal
!   matrix and an inverse iteration algorithm.
!
!
! Arguments
! _________
!
!   D              (INPUT) real(stnd), dimension(:)
!                  On entry, the diagonal elements of the symmetric tridiagonal matrix T.
!
!   E              (INPUT) real(stnd), dimension(:)
!                  On entry, the n-1 subdiagonal elements of the symmetric tridiagonal
!                  matrix T and E(n) is arbitrary .
!
!                  The size of E must verify:  size( E ) = size( D ) = n .
!
!   EIGVAL         (INPUT) real(stnd)
!                  On entry, an eigenvalue of the symmetric tridiagonal matrix.
!
!   EIGVEC         (OUTPUT) real(stnd), dimension(:)
!                  On exit, the computed eigenvector.
!
!                  The shape of EIGVEC must verify: size( EIGVEC ) = size( D ) = n .
!
!   FAILURE        (OUTPUT) logical(lgl)
!                  On exit:
!
!                  - FAILURE = false indicates successful exit.
!                  - FAILURE = true indicates that the eigenvector
!                    failed to converge in MAXITER iterations.
!
!   MAXITER        (INPUT, OPTIONAL) integer(i4b)
!                  The number of inverse iterations performed in the subroutine.
!
!                  By default, 2 inverse iterations are performed.
!
!   SCALING        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if SCALING=true the tridiagonal matrix T is scaled before computing
!                  the eigenvector.
!
!                  The default is to scale the tridiagonal matrix.
!
!   INITVEC        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if INITVEC=true a Fernando vector is used to start the inverse
!                  iteration process; if INITVEC=false a random uniform starting vector is used.
!
!                  For unreduced tridiagonal matrices, the default is to use a Fernando starting vector.
!                  For reduced tridiagonal matrices, the default is to use a random uniform starting vector.
!
!
! Further Details
! _______________
!
!   TRID_INVITER uses Fernando's method for computing a first estimate of an
!   eigenvector corresponding to an approximate eigenvalue of a real n-by-n
!   symmetric tridiagonal matrix T (by default, only if the input tridiagonal
!   matrix T is unreduced).
!
!   This approximate eigenvector is then refined (or computed if Fernando's method is
!   not used) using an inverse iteration algorithm.
!
!   For further details, on Fernando's method for computing eigenvectors of tridiagonal
!   matrices or inverse iteration, see:
!
!   (1) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, pp. 1013-1034.
!
!   (2) Parlett, B.N., and Dhillon, I.S., 1997:
!           Fernando's solution to Wilkinson's problem: An application of double factorization.
!           Linear Algebra and its Appl., 267, pp.247-279.
!
!   (3) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (4) Bini, D.A., Gemignani, L., and Tisseur, F., 2005:
!           The Ehrlich-Aberth method for the nonsymmetric tridiagonal eigenvalue problem.
!           SIAM J. Matrix Anal. Appl., 27, 153-175.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,    only : lamch, machbase, maxexp
    use Utilities,        only : assert_eq, norm
    use Random,           only : random_number_
#ifdef _BLAS
    use BLAS_interfaces,  only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in)                :: eigval
    real(stnd),  intent(in),  dimension(:) :: d, e
    real(stnd),  intent(out), dimension(:) :: eigvec
!
    integer(i4b), intent(in),  optional :: maxiter
!
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in),  optional :: scaling, initvec
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd),  dimension(size(d)) :: sub, diag, sup1, sup2, temp, e2, d2, cs, sn
    real(stnd)                      :: lambda, xj, xjm, safmin, prec, gamma,  &
                                       omega, normt, lim, scalfac, tau, tmp
!
    integer(i4b) :: n, mm, maxit, i, i1, i2, j
!
    logical(lgl)                      :: do_scaling, fernando_init
    logical(lgl),  dimension(size(d)) :: perm
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='trid_inviter'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(d),i4b)       ,   &
                    int(size(e),i4b)       ,   &
                    int(size(eigvec),i4b)  ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            eigvec(1_i4b) = one
!
        end if
!
        return
!
    end if
!
!   SET OPTIONAL ARGUMENTS.
!
    if ( present(maxiter) ) then
        maxit = max( maxiter, 1_i4b )
    else
        maxit = 2_i4b
    end if
!
    if ( present(scaling) ) then
        do_scaling = scaling
    else
        do_scaling = true
    end if
!    
!   GET MACHINES CONSTANTS.
!
    safmin = lamch( 's' )
    gamma  = sqrt( safmin )
    prec   = fudge_prec*lamch( 'p' )
!
!   MAKE A LOCAL COPY OF THE TRIDIAGONAL MATRIX.
!    
    d2(1_i4b:n)       = d(1_i4b:n)
    e2(1_i4b:n-1_i4b) = e(1_i4b:n-1_i4b)
    e2(n)             = zero
!
!   CHECK IF THE TRIDIAGONAL MATRIX IS THE ZERO MATRIX.
!
    do j = n, 2_i4b, -1_i4b
!
        if ( abs(d2(j))>=safmin .or. abs(e2(j-1_i4b))>=safmin ) exit
!
        if ( j/=n ) cycle
!
        eigvec(:n) = zero
        eigvec(j)  = one
!
    end do
!
    if ( j==1_i4b .and. abs(d2(1_i4b))<safmin ) then
!
!       THE BIDIAGONAL MATRIX IS THE NULL MATRIX, RETURN.
!
        return
!
    end if
!    
!   SCALE THE SYMMETRIC TRIDIAGONAL SYSTEM IF REQUIRED.    
!    
    if ( do_scaling ) then
!    
        normt   = maxval( abs(e2(1_i4b:n-1_i4b) ) )
        omega   = machbase**(maxexp-1_i4b)
        tau     = sqrt( gamma*omega )
!        j       = ( exponent( tau ) - exponent( normt ) - 1_i4b )/2_i4b
        j       = exponent( tau ) - exponent( normt ) - 1_i4b
        scalfac = machbase**j
!    
        tmp = gamma/scalfac
!
        where( abs(e2(1_i4b:n-1_i4b))>=tmp )
            e2(1_i4b:n-1_i4b) = scalfac*e2(1_i4b:n-1_i4b)
        elsewhere
            e2(1_i4b:n-1_i4b) = sign( gamma, e2(1_i4b:n-1_i4b) )
        end where
!
        d2(1_i4b:n) = scalfac*d2(1_i4b:n)
!    
        lambda = scalfac*eigval
!    
    else
!    
        scalfac = one
        lambda  = eigval
!    
    end if
!
!   COMPUTE THE NORM OF THE TRIDIAGONAL MATRIX.
!
    normt = zero
    xj    = zero
!
    do j=1_i4b, n
!
        xjm   = abs( e2(j) )
        normt = normt +  abs( d2(j) ) + xj + xjm
        xj    = xjm
!
    end do
!
!   DETERMINE THRESHOLD FOR TINY EIGENVALUES.
!
    lim = normt*real( n, stnd )*prec
!
!   TEST IF THE UPPER PART OF THE TRIDIAGONAL MATRIX IS NOT DIAGONAL OR ZERO.
!
    do i1=1_i4b, n-1_i4b
!
!        if ( abs( e2(i1) )>gamma .or. abs( d2(i1) )>gamma ) exit
!
        if ( abs( e2(i1) )>gamma ) exit
        if ( abs( d2(i1) )>gamma .and. abs( d2(i1) - lambda )<=lim  ) exit
!
    end do
!
!   TEST IF THE LOWER PART OF THE TRIDIAGONAL MATRIX IS NOT DIAGONAL OR ZERO.
!
    do i2=n, 2_i4b, -1_i4b
!
!        if ( abs( e2(i2-1_i4b) )>gamma .or. abs( d2(i2) )>gamma ) exit
!
        if ( abs( e2(i2-1_i4b) )>gamma ) exit
        if ( abs( d2(i2) )>gamma .and. abs( d2(i2) - lambda )<=lim ) exit
!
    end do
!
    if ( ( abs( lambda )<=gamma+lim .and. ( i1/=1_i4b .or. i2/=n ) ) &
          .or. ( i2<i1 ) ) then
!
!       IF THE EIGENVALUE IS VERY SMALL, WE CANNOT WORK ON THE REDUCED PORTION OF 
!       THE TRIDIAGONAL MATRIX.
!
        i1 = 1_i4b
        i2 = n
!
    end if
!
    if ( present( initvec ) ) then
!
        fernando_init = initvec
!
    else
!
!       TEST IF THE INPUT TRIDIAGONAL MATRIX IS UNREDUCED.
!
        fernando_init = all( abs( e2(i1:i2-1_i4b) )>gamma )
!
    end if
!
!   COMPUTE EIGENVECTOR BY INVERSE ITERATION.
!    
    if ( fernando_init ) then
!    
!       FERNANDO-BINI METHOD.  
!
        call trid_qr_cmp( d2(i1:i2), e2(i1:i2-1_i4b), lambda, cs(i1:i2-1_i4b), sn(i1:i2-1_i4b),    &
                          diag(i1:i2), sup1(i1:i2), sup2(i1:i2), mm )
!    
        mm = mm + i1 - 1_i4b
!    
        temp(i1:i2) = zero
!
        temp(mm) = one
!
!       DO maxit ITERATIONS FOR THE EIGENVECTOR OF THE TRIDIAGONAL MATRIX.
!
        do i=1_i4b, maxit
!
!           SOLVE THE SYSTEM QRx = b WITH RESCALING TO AVOID OVERFLOW.
!
            call trid_qr_solve( cs(i1:i2-1_i4b), sn(i1:i2-1_i4b), diag(i1:i2), sup1(i1:i2),  &
                                sup2(i1:i2), temp(i1:i2)  )
!
!           NORMALIZE THE EIGENVECTOR.
!
            tau = norm( temp(i1:i2) )
!
            if ( tau/=zero ) then
!
                temp(i1:i2) = temp(i1:i2)/tau
!
            else if ( i/=maxit ) then
!
                call random_number_( temp(i1:i2) )
!
                temp(i1:i2) = temp(i1:i2)*two - one
!
            end if
!
        end do
!
    else
!
!       STANDART INVERSE ITERATION METHOD.
!
!       COMPUTE LU FACTORS WITH PARTIAL PIVOTING (PT=LU) OF THE TRIDIAGONAL MATRIX.
!
        call trid_cmp_rv( d2(i1:i2), e2(i1:i2), lambda, sub(i1:i2), diag(i1:i2),  &
                          sup1(i1:i2), sup2(i1:i2), perm(i1:i2) )
!    
!       COMPUTE A RANDOM INITIAL VECTOR.  
!
        call random_number_( temp(i1:i2) )
!
        temp(i1:i2) = temp(i1:i2)*two - one
!
!       REPLACE SMALL DIAGONAL ELEMENTS BY safmin.
!
        where( abs(diag(i1:i2))<safmin ) diag(i1:i2) = sign( safmin, diag(i1:i2) )
!
!       DO maxit ITERATIONS FOR THE EIGENVECTOR OF THE TRIDIAGONAL MATRIX.
!
        do i=1_i4b, maxit
!
!           SOLVE THE SYSTEM LU = Pb WITH RESCALING TO AVOID OVERFLOW.
!
            call trid_solve_rv( sub(i1:i2), diag(i1:i2), sup1(i1:i2),   &
                                sup2(i1:i2), perm(i1:i2), temp(i1:i2) )
!
!           NORMALIZE THE EIGENVECTOR.
!
            tau = norm( temp(i1:i2) )
!
            if ( tau/=zero ) then
!
                temp(i1:i2) = temp(i1:i2)/tau
!
            else if ( i/=maxit ) then
!
                call random_number_( temp(i1:i2) )
!
                temp(i1:i2) = temp(i1:i2)*two - one
!
            end if
!
        end do
!
    end if
!
!   SAVE THE EIGENVECTOR.
!
    temp(1_i4b:i1-1_i4b) = zero
    temp(i2+1_i4b:n)     = zero
!
    eigvec(1_i4b:n) = temp(1_i4b:n)
!
    if ( tau/=zero ) then
!
!       CHECK ACCURACY OF THE EIGENVECTOR.
!
        diag(1_i4b:n)       = d(1_i4b:n)
        sup1(1_i4b:n-1_i4b) = e(1_i4b:n-1_i4b)
        sup1(n)             = zero
!
        sub(1_i4b)         = diag(1_i4b)*temp(1_i4b) + sup1(1_i4b)*temp(2_i4b)
        sub(2_i4b:n-1_i4b) = sup1(1_i4b:n-2_i4b)*temp(1_i4b:n-2_i4b)           +   &
                             diag(2_i4b:n-1_i4b)*temp(2_i4b:n-1_i4b)           +   &
                             sup1(2_i4b:n-1_i4b)*temp(3_i4b:n)
        sub(n)             = sup1(n-1_i4b)*temp(n-1_i4b) + diag(n)*temp(n)
!
!       COMPUTE RAYLEIGH QUOTIENT.
!
        tau = dot_product( sub(1_i4b:n), temp(1_i4b:n) )
!
!       COMPUTE THE RESIDUAL NORM OF THE EIGENVECTOR
!       WITH RESPECT TO ITS RAYLEIGH QUOTIENT.
!
#ifdef _BLAS
        call axpy( n, -tau, temp(1_i4b:n), 1_i4b, sub(1_i4b:n), 1_i4b )
#else
        sub(1_i4b:n) = sub(1_i4b:n) - tau*temp(1_i4b:n)
#endif
        xj  = sum( abs( sub(:n) ) )
        xjm = abs( tau - eigval )
!
!       DETERMINE THRESHOLD FOR THE TESTS OF ACCURACY OF
!       THE COMPUTED EIGENVECTOR.
!
        lim = ( normt*real(n,stnd)*prec )/scalfac
!
        failure = max( xj, xjm )>lim
!
    else
!
        failure = true
!
    end if
!
!
! END OF SUBROUTINE trid_inviter_rv
! _________________________________
!
    end subroutine trid_inviter_rv
!
! =========================================================================================
!
    subroutine trid_inviter_a( d, e, eigval, eigvec, failure, maxiter, ortho, &
                               backward_sweep, scaling, initvec               )
!
! Purpose
! _______
!
!   TRID_INVITER computes the eigenvectors of a real n-by-n symmetric tridiagonal
!   matrix T corresponding to specified eigenvalues, by combining Fernando's
!   method for computing (selected) eigenvectors of a real n-by-n symmetric
!   tridiagonal matrix and an inverse iteration algorithm.
!
!
! Arguments
! _________
!
!   D              (INPUT) real(stnd), dimension(:)
!                  On entry, the diagonal elements of the symmetric tridiagonal matrix T.
!
!   E              (INPUT) real(stnd), dimension(:)
!                  On entry, the n-1 subdiagonal elements of the symmetric tridiagonal
!                  matrix T and E(n) is arbitrary .
!
!                  The size of E must verify:  size( E ) = size( D ) = n .
!
!   EIGVAL         (INPUT) real(stnd), dimension(:)
!                  On entry, selected eigenvalues of the symmetric tridiagonal matrix.
!                  The eigenvalues must be given in decreasing order.
!
!                  The size of EIGVAL must verify: size( EIGVAL ) <= size( D ) = n .
!
!   EIGVEC         (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed eigenvectors. The eigenvector associated
!                  with the eigenvalue EIGVAL(j) is stored in the j-th column of EIGVEC.
!
!                  The shape of EIGVEC must verify: 
!
!                  - size( EIGVEC, 1 ) = size( D ) = n ,
!                  - size( EIGVEC, 2 ) = size( EIGVAL ) .
!
!   FAILURE        (OUTPUT) logical(lgl)
!                  On exit:
!
!                  - FAILURE = false indicates successful exit.
!                  - FAILURE = true indicates that some eigenvectors
!                    failed to converge in MAXITER iterations.
!
!   MAXITER        (INPUT, OPTIONAL) integer(i4b)
!                  The number of inverse iterations performed in the subroutine. By default,
!                  2 inverse iterations are performed for all the eigenvectors.
!
!   ORTHO          (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - ORTHO=true, all the eigenvectors are orthogonalized by the
!                    Modified Gram-Schmidt algorithm;
!                  - ORTHO=false, the eigenvectors are not orthogonalized by 
!                    the Modified Gram-Schmidt algorithm.
!
!                  The default is to orthogonalize the eigenvectors only for the eigenvalues,
!                  which are not well-separated.
!
!   BACKWARD_SWEEP (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - BACKWARD_SWEEP=true and the eigenvectors are orthogonalized by
!                    the modified Gram-Schmidt algorithm, a backward sweep of the modified Gram-Schmidt
!                    algorithm is also performed;
!                  - BACKWARD_SWEEP=false a backward sweep is not performed.
!
!                  The default is not to perform a backward sweep of the modified Gram-Schmidt
!                  algorithm.
!
!   SCALING        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - SCALING=true, the tridiagonal matrix T is scaled before computing
!                    the eigenvectors;
!                  - SCALING=false, the tridiagonal matrix T is not scaled.
!
!                  The default is to scale the tridiagonal matrix.
!
!   INITVEC        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - INITVEC=true, Fernando vectors are used to start the inverse
!                    iteration process;
!                  - INITVEC=false, random uniform starting vectors are used.
!
!                  For unreduced tridiagonal matrices, the default is to use Fernando starting vectors
!                  if the eigenvalues are well-separated and random uniform starting vectors otherwise.
!                  For reduced tridiagonal matrices, the default is to use random uniform starting vectors.
!
!
! Further Details
! _______________
!
!   TRID_INVITER uses Fernando's method for computing a first estimate of (selected)
!   eigenvectors corresponding to (selected) approximate eigenvalues of a real n-by-n
!   symmetric tridiagonal matrix T (by default, only for the eigenvalues which are
!   well separated and if the input tridiagonal matrix T is unreduced).
!
!   These approximate eigenvectors are then refined (or computed if Fernando's method is
!   not used) using an inverse iteration algorithm for all the eigenvalues at one step.
!   The eigenvectors are then orthogonalized by the Modified Gram-Schmidt algorithm if
!   the eigenvalues are not well-separated.
!
!   TRID_INVITER may fail if some the eigenvalues specified in parameter EIGVAL are nearly
!   identical.
!
!   For further details, on Fernando's method for computing eigenvectors of tridiagonal
!   matrices or inverse iteration, see:
!
!   (1) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, pp. 1013-1034.
!
!   (2) Parlett, B.N., and Dhillon, I.S., 1997:
!           Fernando's solution to Wilkinson's problem: An application of double factorization.
!           Linear Algebra and its Appl., 267, pp.247-279.
!
!   (3) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (4) Bini, D.A., Gemignani, L., and Tisseur, F., 2005:
!           The Ehrlich-Aberth method for the nonsymmetric tridiagonal eigenvalue problem.
!           SIAM J. Matrix Anal. Appl., 27, 153-175.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,    only : lamch, machbase, maxexp
    use Char_Constants,   only : eig_error2, allocate_error
    use Utilities,        only : assert_eq, assert, merror, norm
    use Random,           only : random_number_
#ifdef _BLAS
    use BLAS_interfaces,  only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),  dimension(:)   :: d, e, eigval
    real(stnd),  intent(out), dimension(:,:) :: eigvec
!
    integer(i4b), intent(in),  optional :: maxiter
!
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in),  optional :: ortho, backward_sweep, scaling, initvec
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd),  dimension(size(eigval))   :: lambda
    real(stnd),  dimension(size(d))        :: sub, diag, sup1, sup2, temp, e2, d2, cs, sn
    real(stnd)                             :: xj, xjm, safmin, prec, eps, pertol, sep,      &
                                              normt, lim, lim_eig, lim_ortho, scalfac, tau, &
                                              gamma, omega, tmp
!
    integer(i4b)                             :: n, m, mm, maxit, i, i1, i2, i3, j, j1, j2, k,  &
                                                l, nn1, nn2, m0
    integer(i4b),  dimension(:), allocatable :: jstart, jend
    integer                                  :: iok
!
    logical(lgl)                             :: test_ortho, do_ortho, do_backward_sweep,   &
                                                do_scaling, test_initvec, do_initvec,      &
                                                zero_vec
    logical(lgl),  dimension(size(eigval))   :: fernando_init, comp_norm
    logical(lgl),  dimension(size(d))        :: perm
!
#ifdef _OPENMP
!
    integer(i4b) :: num_threads, thread_num, jinc, jrem, joffset
!
    logical      :: test_par, ompparallel, ompnested
!
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='trid_inviter'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure  = false
    zero_vec = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(d),i4b)         ,   &
                    int(size(e),i4b)         ,   &
                    int(size(eigvec,1),i4b)  ,   &
                    name_proc )
!
    m =  assert_eq( int(size(eigval),i4b)    ,   &
                    int(size(eigvec,2),i4b)  ,   &
                    name_proc )
!
    call assert( logical( m<=n, lgl ), name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b .or. m<=0_i4b) then
!
        if ( n==1_i4b .and. m==1_i4b ) then
!
            eigvec(1_i4b,1_i4b) = one
!
        end if
!
        return
!
    end if
!
!   TEST IF EIGENVALUES ARE IN DECREASING ORDER.
!
    if ( m/=1_i4b ) then
!
        temp(1_i4b:m-1_i4b) = eigval(2_i4b:m) - eigval(1_i4b:m-1_i4b)
!
        if ( any( temp(1_i4b:m-1_i4b)>zero ) ) then
!
            call merror( name_proc//eig_error2 )
!
        end if
!
    end if
!
!   SET OPTIONAL ARGUMENTS.
!
    if ( present(maxiter) ) then
        maxit = max( maxiter, 1_i4b )
    else
        maxit = 2_i4b
    end if
!
    test_ortho = present( ortho )
    do_ortho   = false
!
    if ( present(backward_sweep) ) then
        do_backward_sweep = backward_sweep
    else
        do_backward_sweep = false
    end if
!
    test_initvec = present( initvec )
!
    if ( present(scaling) ) then
        do_scaling = scaling
    else
        do_scaling = true
    end if
!    
!   GET MACHINES CONSTANTS.
!
    safmin = lamch( 's' )
    gamma  = sqrt( safmin )
    prec   = fudge_prec*lamch( 'p' )
    eps    = lamch( 'u' )
!
!   MAKE A LOCAL COPY OF THE TRIDIAGONAL MATRIX.
!    
    d2(1_i4b:n)       = d(1_i4b:n)
    e2(1_i4b:n-1_i4b) = e(1_i4b:n-1_i4b)
    e2(n)             = zero
!
!   CHECK IF THE TRIDIAGONAL MATRIX IS THE ZERO MATRIX.
!
    m0 = m
!
    do j = n, 2_i4b, -1_i4b
!
        if ( abs(d2(j))>=safmin .or. abs(e2(j-1_i4b))>=safmin ) exit
!
        if ( m0<=0_i4b ) cycle
!
        eigvec(:n,m0) = zero
        eigvec(j,m0)  = one
!
        m0 = m0 - 1_i4b
!
    end do
!
    if ( j==1_i4b .and. abs(d2(1_i4b))<safmin ) then
!
        if ( m==n ) then
!
            eigvec(1_i4b,1_i4b)   = one
            eigvec(2_i4b:n,1_i4b) = zero                
!
        end if
!
!       THE BIDIAGONAL MATRIX IS THE NULL MATRIX, RETURN.
!
        return
!
    else if ( m==n .and. all( eigval(1_i4b:m)>=zero )  ) then
!
!       CHECK THE FIRST PART OF THE TRIDIAGONAL MATRIX FOR ZEROS
!       AND REDUCE ITS SIZE IF POSSIBLE IF THE MATRIX IS SEMI_DEFINITE
!       POSITIVE.
!
        nn2 = j
!
        do j = 1_i4b, n-1_i4b
!
            if ( abs(d2(j))>=safmin .or. abs(e2(j))>=safmin ) exit
!
            eigvec(:n,m0) = zero
            eigvec(j,m0)  = one
!
            m0 = m0 - 1_i4b
!
        end do
!
        nn1 = j
!
        lambda(m0+1_i4b:m) = one
!
    else
!
        nn1 = 1_i4b
        nn2 = n
!
        m0 = m
!
    end if
!    
!   SCALE THE SYMMETRIC TRIDIAGONAL SYSTEM IF REQUIRED.    
!
    if ( do_scaling ) then
!    
        normt   = maxval( abs(e2(1_i4b:n-1_i4b) ) )
        omega   = machbase**(maxexp-1_i4b)
        tau     = sqrt( gamma*omega )
!        j       = ( exponent( tau ) - exponent( normt ) - 1_i4b )/2_i4b
        j       = exponent( tau ) - exponent( normt ) - 1_i4b
        scalfac = machbase**j
!    
        tmp = gamma/scalfac
!
        where( abs(e2(1_i4b:n-1_i4b))>=tmp )
            e2(1_i4b:n-1_i4b) = scalfac*e2(1_i4b:n-1_i4b)
        elsewhere
            e2(1_i4b:n-1_i4b) = sign( gamma, e2(1_i4b:n-1_i4b) )
        end where
!
        d2(1_i4b:n) = scalfac*d2(1_i4b:n)
!    
        lambda(:m0) = scalfac*eigval(:m0)
!    
    else
!    
        scalfac     = one
        lambda(:m0) = eigval(:m0)
!    
    end if
!
!   COMPUTE THE UPPER GERSCHGORIN BOUND AND THE NORM OF
!   THE TRIDIAGONAL MATRIX.
!
    normt = zero
    tau   = zero
    xj    = zero
!
    do j=1_i4b, n
!
        xjm   = abs( e2(j) )
        sep   = abs( d2(j) ) + xj + xjm
        normt = normt + sep
        tau   = max( tau, sep )
        xj    = xjm
!
    end do
!
!   DETERMINE THRESHOLD FOR TINY EIGENVALUES.
!
    lim_eig = normt*real( n, stnd )*prec
!    lim_eig = tau*real( n, stnd )*prec
!
!   TEST IF THE UPPER PART OF THE TRIDIAGONAL MATRIX IS NOT DIAGONAL OR ZERO.
!
    do i1=nn1, n-1_i4b
!
!        if ( abs( e2(i1) )>gamma .or. abs( d2(i1) )>gamma ) exit
!
        if ( abs( e2(i1) )>gamma ) exit
        if ( abs( d2(i1) )>gamma .and. any( abs( d2(i1) - lambda(1_i4b:m0) )<=lim_eig )  ) exit
!
    end do
!
!   TEST IF THE LOWER PART OF THE TRIDIAGONAL MATRIX IS NOT DIAGONAL OR ZERO.
!
    do i2=nn2, 2_i4b, -1_i4b
!
!        if ( abs( e2(i2-1_i4b) )>gamma .or. abs( d2(i2) )>gamma ) exit
!
        if ( abs( e2(i2-1_i4b) )>gamma ) exit
        if ( abs( d2(i2) )>gamma .and. any( abs( d2(i2) - lambda(1_i4b:m0) )<=lim_eig ) ) exit
!
    end do
!
    i3 = i2 - i1 + 1_i4b
!
    if ( ( any( abs( lambda(1_i4b:m0) )<=gamma+lim_eig ) .and. ( i1/=nn1 .or. i2/=nn2 ) ) &
          .or. ( m0>i3 ) ) then
!
!       IF SOME EIGENVALUES ARE VERY SMALL OR IF WE MISS SOME EIGENVALUES
!       WHEN COMPUTING i1 and i2 ABOVE, WE CANNOT WORK ON THE REDUCED PORTION OF 
!       THE TRIDIAGONAL MATRIX.
!
        i1 = nn1
        i2 = nn2
        i3 = i2 - i1 + 1_i4b
!
    end if
!
!   TEST IF THE INPUT TRIDIAGONAL MATRIX IS UNREDUCED.
!
    do_initvec = all( abs( e2(i1:i2-1_i4b) )>gamma )
!
!   DETERMINE THRESHOLDS FOR SEPARATION OF EIGENVALUES AND
!   REORTHOGONALIZATION OF EIGENVECTORS.
!
    lim_ortho = fudge_ortho*sqrt( eps )*tau
!
    omega = lim_ortho
!    omega = eps*normt
!    omega = real( n, stnd )*sqrt( eps )*normt
!
!   IF EIGENVALUES ARE TOO CLOSE, ADD A RELATIVELY SMALL PERTURBATION
!   TO AVOID PARALLEL ITERATES.
!   IF EIGENVALUES ARE TOO CLOSE, REORTHOGONALIZE EIGENVECTORS BY
!   MODIFIED GRAM-SCHMIDT.
!
    xjm = lambda(m0)
!
    do j=m0-1_i4b, 1_i4b, -1_i4b
!
        xj     = lambda(j)
        pertol = abs( prec*xj )
        sep    = xj - xjm
!
        comp_norm(j+1_i4b) = sep<lim_ortho
!
        fernando_init(j+1_i4b) = sep>omega
!
!BUG:   THE NAGFOR COMPILER DO NOT IMPLEMENT THE MERGE FUNCTION
!       CORRECTLY IN EXTENDED PRECISION. THUS, IT IS SAFER TO USE
!       A SIMPLE IF THEN ELSE INSTRUCTION.
!
!        xjm = merge( xj, xjm+pertol, sep>=pertol )
!
        if ( sep>=pertol ) then
!
            xjm = xj
!
        else
!
            xjm       = xjm + pertol
            lambda(j) = xjm
!
        end if
!
    end do
!
!   SET INDICATOR VARIABLES FOR ORTHOGONALIZATION.
!
    comp_norm(1_i4b) = false
!
    if ( test_ortho ) then
!
        if ( m0/=1_i4b ) then
!
            do_ortho = ortho
!
        end if
!
        if ( do_ortho ) then
!
            comp_norm(2_i4b:m0) = true
!
        end if
!
    else
!
        do_ortho = any( comp_norm(2_i4b:m0) )
!
    end if
!
    if ( do_ortho ) then
!
!       DETERMINE THE NUMBER OF CLUSTERS OF EIGENVALUES.
!
        i = count( comp_norm(2_i4b:m0) )
!
        if ( i==m0-1_i4b ) then
            k = 1_i4b
        else
            k = i
        end if
!
!       ALLOCATE WORK VARIABLES.
!
        allocate( jstart(k), jend(k), stat = iok  )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       k IS THE NUMBER OF CLUSTERS OF EIGENVALUES.
!       jstart(k) IS THE INDICE OF THE FIRST EIGENVALUE IN CLUSTER k.
!       jend(k)+1 IS THE INDICE OF THE LAST EIGENVALUE IN CLUSTER k.
!
        if ( i==m0-1_i4b ) then
!
!           ALL THE EIGENVALUES ARE CLUSTERED TOGETHER.
!
            jstart(1_i4b) = 1_i4b
            jend(1_i4b)   = m0 - 1_i4b
!
        else
!
!           UNKNOWN NUMBER OF CLUSTERS, BUT AT LEAST ONE
!           CLUSTER OF EIGENVALUES.
!
            j = 2_i4b
            k = 0_i4b
!
            do
!
                if ( comp_norm(j) ) then
!
                    k = k + 1_i4b
!
                    jstart(k) = j - 1_i4b
!
                    do i=j+1_i4b, m0
                        if ( .not.comp_norm(i) ) exit
                    end do
!
                    jend(k) = i - 2_i4b
!
                    j = i + 1_i4b
!
                else
!
                    j = j + 1_i4b
!
                end if
!
                if ( j>m0 ) exit
!
            end do
!
        end if
!
    end if
!
!   SET INDICATOR VARIABLE FOR INITIALIZATION OF THE
!   INVERSE ITERATION PROCESS FOR EACH EIGENVECTOR.
!
    if ( test_initvec ) then
        do_initvec = initvec
    end if
!    
    if ( do_initvec ) then
!
        if ( test_initvec ) then
!
!           START THE RIGHT HAND SIDES WITH FERNANDO VECTORS.
!
            fernando_init(1_i4b:m0) = true
!
        else
!
!           START THE RIGHT HAND SIDES WITH FERNANDO VECTORS,
!           ONLY IF THE EIGENVALUES ARE WELL SEPARATED.
!
            fernando_init(1_i4b) = true
!
        end if
!
    else
!
!       START THE RIGHT HAND SIDES AT RANDOM VALUES.
!
        fernando_init(1_i4b:m0) = false
!
    end if
!
!   DETERMINE THRESHOLD FOR THE TEST OF ACCURACY
!   OF THE COMPUTED EIGENVECTORS.
!
    lim = ( normt*real(n,stnd)*prec )/scalfac
!
#ifdef _OPENMP
    i = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  m0>1_i4b                            .and.      &
                  i>1_i4b
!
    if ( test_par ) then
!
!       COMPUTE EIGENVECTORS BY INVERSE ITERATION.
!       OPENMP PARALLEL PATH.
!
!$OMP PARALLEL PRIVATE(i,mm,num_threads,thread_num,jrem,jinc,joffset,j,j1,j2,l,  &
!$OMP                  sub,diag,sup1,sup2,perm,temp,cs,sn,xj,xjm,tau,tmp)
!
!$OMP DO SCHEDULE(STATIC)
!
        do j = 1_i4b, m0
!
            xj = lambda(j)
!    
            if ( fernando_init(j) ) then
!    
!               FERNANDO-BINI METHOD.  
!
                call trid_qr_cmp( d2(i1:i2), e2(i1:i2-1_i4b), xj, cs(i1:i2-1_i4b), sn(i1:i2-1_i4b),    &
                                  diag(i1:i2), sup1(i1:i2), sup2(i1:i2), mm )
!    
                mm = mm + i1 - 1_i4b
!
                if ( comp_norm(j) ) then
!
!                   ADD A SMALL PERTURBATION TO ALL COMPONENTS OF THE
!                   INITIAL VECTOR TO PROTECT AGAINST FAILURE OF FERNANDO
!                   METHOD WHICH MAY PRODUCE PARALLEL INITIAL VECTORS IF
!                   EIGENVALUES ARE TOO CLOSED.
!
                    call random_number_( temp(i1:i2) )
!
                    temp(i1:i2) = sqrt( eps )*( temp(i1:i2)*two - one )
!
                else
!    
                    temp(i1:i2) = zero
!
                end if
!
                temp(mm) = one
!
!               DO maxit ITERATIONS FOR THE EIGENVECTORS.
!
                do i=1_i4b, maxit
!
!                   SOLVE THE SYSTEM QRx = b WITH RESCALING TO AVOID OVERFLOW.
!
                    call trid_qr_solve( cs(i1:i2-1_i4b), sn(i1:i2-1_i4b), diag(i1:i2), sup1(i1:i2),  &
                                        sup2(i1:i2), temp(i1:i2)  )
!
!                   NORMALIZE THE EIGENVECTORS.
!
                    xj = norm( temp(i1:i2) )
!
                    if ( xj/=zero ) then
!
                        temp(i1:i2) = temp(i1:i2)/xj
!
                    else if ( i/=maxit ) then
!
                        call random_number_( temp(i1:i2) )
!
                        temp(i1:i2) = temp(i1:i2)*two - one
!
                    end if
!
                end do
!
            else
!
!               STANDART INVERSE ITERATION METHOD.
!
!               COMPUTE LU FACTORS WITH PARTIAL PIVOTING (PT=LU).
!
                call trid_cmp_rv( d2(i1:i2), e2(i1:i2), xj, sub(i1:i2),              &
                                  diag(i1:i2), sup1(i1:i2), sup2(i1:i2), perm(i1:i2) )
!    
!               COMPUTE A RANDOM INITIAL VECTOR.  
!
                call random_number_( temp(i1:i2) )
!
                temp(i1:i2) = temp(i1:i2)*two - one
!
!               REPLACE SMALL DIAGONAL ELEMENTS BY safmin.
!
                where( abs(diag(i1:i2))<safmin ) diag(i1:i2) = sign( safmin, diag(i1:i2) )
!
!               DO maxit ITERATIONS FOR THE EIGENVECTORS.
!
                do i=1_i4b, maxit
!
!                   SOLVE THE SYSTEM LU = Pb WITH RESCALING TO AVOID OVERFLOW.
!
                    call trid_solve_rv( sub(i1:i2), diag(i1:i2), sup1(i1:i2),  &
                                        sup2(i1:i2), perm(i1:i2), temp(i1:i2)  )
!
!                   NORMALIZE THE EIGENVECTORS.
!
                    xj = norm( temp(i1:i2) )
!
                    if ( xj/=zero ) then
!
                        temp(i1:i2) = temp(i1:i2)/xj
!
                    else if ( i/=maxit ) then
!
                        call random_number_( temp(i1:i2) )
!
                        temp(i1:i2) = temp(i1:i2)*two - one
!
                    end if
!
                end do
!
            end if
!
!           SAVE THE EIGENVECTORS.
!
            if ( xj/=zero ) then
!
                eigvec(1_i4b:i1-1_i4b,j) = zero
                eigvec(i1:i2,j)          = temp(i1:i2)
                eigvec(i2+1_i4b:n,j)     = zero
!
            else
!
                eigvec(1_i4b:n,j) = zero
!
            end if
!
            lambda(j) = xj
!
        end do
!
!$OMP END DO
!
        if ( do_ortho ) then
!
!           REORTHOGONALIZE EIGENVECTORS BY MODIFIED GRAM-SCHMIDT.
!
!           FIRST, REORTHOGONALIZE EIGENVECTORS CORRESPONDING TO SMALL CLUSTERS
!           OF ALMOST-MULTIPLE EIGENVALUES. EACH THREAD TAKES CARE OF ONE CLUSTER
!           AND PARALLELIZATION IS OVER THE NUMBER OF SMALL CLUSTERS.
!
!$OMP DO SCHEDULE(DYNAMIC,1)
!
            do l=1_i4b, k
!
                j1 = jstart(l)
                j2 = jend(l)
!
                if ( j2-j1<neig_cluster ) then
!
                    mm = j2 + 1_i4b
!
                    do j=j1, j2 
!
!                       MAKE A LOCAL COPY OF THE j EIGENVECTOR.
!
                        sup1(i1:i2) = eigvec(i1:i2,j)
!
!                       FORWARD SWEEP OF MGS ORTHOGONALIZATION.
!
                        do i=j+1_i4b, mm
!
                            xj = -dot_product( sup1(i1:i2), eigvec(i1:i2,i) )
#ifdef _BLAS
                            call axpy( i3, xj, sup1(i1:i2), 1_i4b, eigvec(i1:i2,i), 1_i4b )
#else
                            eigvec(i1:i2,i) = eigvec(i1:i2,i) + xj*sup1(i1:i2)
#endif
!
                        end do
!
!                       NORMALIZE THE NEXT j EIGENVECTOR.
!
                        sup1(i1:i2) = eigvec(i1:i2,j+1_i4b)
!
                        xj = norm( sup1(i1:i2) )
!
                        if ( xj<=fudge_ortho2 ) then
!
!                           REORTHOGONALIZATION IS REQUIRED.
!
                            call ortho_vec(eigvec(i1:i2,j1:j), sup1(i1:i2), xj )
!
                        else
!
!                          NORMALIZATION IS ENOUGH.
!
                           sup1(i1:i2) = (one/xj)*sup1(i1:i2)
!
                        end if
!
                        eigvec(i1:i2,j+1_i4b) = sup1(i1:i2)
!
                        lambda(j+1_i4b) = xj
!
                    end do
!
                    if ( do_backward_sweep ) then
!
!                       BACKWARD SWEEP OF MGS ORTHOGONALIZATION.
!
                        do j=j2, j1, -1_i4b
!
                            sup1(i1:i2) = eigvec(i1:i2,j)
!
                            do i=j+1_i4b, mm
!
                                xj = -dot_product( sup1(i1:i2), eigvec(i1:i2,i) )
!
#ifdef _BLAS
                                call axpy( i3, xj, sup1(i1:i2), 1_i4b, eigvec(i1:i2,i), 1_i4b )
#else
                                eigvec(i1:i2,i) = eigvec(i1:i2,i) + xj*sup1(i1:i2)
#endif
!
                            end do
!
                        end do
!
                    end if
!
                end if
!
            end do
!
!$OMP END DO
!
!           SECOND, REORTHOGONALIZE EIGENVECTORS CORRESPONDING TO BIG CLUSTERS
!           OF ALMOST-MULTIPLE EIGENVALUES. PARALLELIZATION IS OVER THE
!           NUMBER OF VECTORS IN EACH CLUSTER.
!
            num_threads = omp_get_num_threads()
            thread_num  = omp_get_thread_num()
!
            do l=1_i4b, k
!
                if ( jend(l)-jstart(l)>=neig_cluster ) then
!
                    mm = jend(l) - jstart(l) + 2_i4b
!
!                   FORWARD SWEEP OF MGS ORTHOGONALIZATION.
!
                    do j=jstart(l), jend(l) 
!
!                       DISTRIBUTE THE WORK AMONG THE THREADS.
!
                        mm   = mm - 1_i4b
                        jrem = mod( mm, num_threads )
                        jinc = ( mm - jrem )/num_threads
!
                        if ( thread_num<jrem ) then
                            jinc    = jinc + 1_i4b
                            joffset = j + 1_i4b
                        else
                            joffset = j + jrem + 1_i4b
                        end if
!
                        j1 = thread_num*jinc + joffset
                        j2 = j1 + jinc - 1_i4b
!
                        if ( j1<=j2 ) then
!
!                           MAKE LOCAL COPIES OF THE j EIGENVECTOR.
!
                            sup1(i1:i2) = eigvec(i1:i2,j)
!
                            do i = j1, j2
!
                                xj = -dot_product( sup1(i1:i2), eigvec(i1:i2,i) )
#ifdef _BLAS
                                call axpy( i3, xj, sup1(i1:i2), 1_i4b, eigvec(i1:i2,i), 1_i4b )
#else
                                eigvec(i1:i2,i) = eigvec(i1:i2,i) + xj*sup1(i1:i2)
#endif
!
                            end do
!
                        end if
!
                        if ( thread_num==0_i4b ) then
!
!                           NORMALIZE THE NEXT j EIGENVECTOR.
!
                            sup1(i1:i2) = eigvec(i1:i2,j+1_i4b)
!
                            xj = norm( sup1(i1:i2) )
!
                            if ( xj<=fudge_ortho2 ) then
!
!                               REORTHOGONALIZATION IS REQUIRED.
!
                                i = max( jstart(l), 1_i4b)
!
                                call ortho_vec(eigvec(i1:i2,i:j), sup1(i1:i2), xj )
!
                            else
!
!                               NORMALIZATION IS ENOUGH.
!
                                sup1(i1:i2) = (one/xj)*sup1(i1:i2)
!
                            end if
!
                            eigvec(i1:i2,j+1_i4b) = sup1(i1:i2)
!
                            lambda(j+1_i4b) = xj
!
                        end if
!
!$OMP BARRIER
!
                    end do
!
                    if ( do_backward_sweep ) then
!
!                       BACKWARD SWEEP OF MGS ORTHOGONALIZATION.
!
                        mm = 0_i4b
!
                        do j=jend(l), jstart(l), -1_i4b
!
!                           DISTRIBUTE THE WORK AMONG THE THREADS.
!
                            mm   = mm + 1_i4b
                            jrem = mod( mm, num_threads )
                            jinc = ( mm - jrem )/num_threads
!
                            if ( thread_num<jrem ) then
                                jinc    = jinc + 1_i4b
                                joffset = j + 1_i4b
                            else
                                joffset = j + jrem + 1_i4b
                            end if
!
                            j1 = thread_num*jinc + joffset
                            j2 = j1 + jinc - 1_i4b
!
                            if ( j1<=j2 ) then
!
                                sup1(i1:i2) = eigvec(i1:i2,j)
!
                                do i = j1, j2
!
                                    xj = -dot_product( sup1(i1:i2), eigvec(i1:i2,i) )
#ifdef _BLAS
                                    call axpy( i3, xj, sup1(i1:i2), 1_i4b, eigvec(i1:i2,i), 1_i4b )
#else
                                    eigvec(i1:i2,i) = eigvec(i1:i2,i) + xj*sup1(i1:i2)
#endif
!
                                end do
!
                            end if
!
!$OMP BARRIER
!
                        end do
!
                    end if
!
                end if
!
            end do
!
        end if
!
!       CHECK ACCURACY OF THE EIGENVECTORS.
!
        diag(1_i4b:n)       = d(1_i4b:n)
        sup1(1_i4b:n-1_i4b) = e(1_i4b:n-1_i4b)
        sup1(n)             = zero
!
!$OMP DO SCHEDULE(STATIC)
!
        do j = 1_i4b, m
!
            if ( lambda(j)/=zero ) then
!
                temp(1_i4b:n) = eigvec(1_i4b:n,j)
!
                sub(1_i4b)         = diag(1_i4b)*temp(1_i4b) + sup1(1_i4b)*temp(2_i4b)
                sub(2_i4b:n-1_i4b) = sup1(1_i4b:n-2_i4b)*temp(1_i4b:n-2_i4b)           +   &
                                     diag(2_i4b:n-1_i4b)*temp(2_i4b:n-1_i4b)           +   &
                                     sup1(2_i4b:n-1_i4b)*temp(3_i4b:n)
                sub(n)             = sup1(n-1_i4b)*temp(n-1_i4b) + diag(n)*temp(n)
!
!               COMPUTE RAYLEIGH QUOTIENT.
!
                xj    = dot_product( sub(1_i4b:n), temp(1_i4b:n) )
                e2(j) = xj
!
!               COMPUTE THE RESIDUAL NORM OF THE JTH EIGENVECTOR
!               WITH RESPECT TO ITS RAYLEIGH QUOTIENT.
!
#ifdef _BLAS
                call axpy( n, -xj, temp(1_i4b:n), 1_i4b, sub(1_i4b:n), 1_i4b )
#else
                sub(1_i4b:n) = sub(1_i4b:n) - xj*temp(1_i4b:n)
#endif
                xj = sum( abs( sub(:n) ) )
!
                comp_norm(j) = xj>lim
!
            else
!
                comp_norm(j) = true
!
!$OMP CRITICAL (zerovec)
                zero_vec = true
!$OMP END CRITICAL (zerovec)
!
            end if
!
        end do
!
!$OMP END DO
!
!$OMP END PARALLEL
!
!       SET OUTPUT FAILURE PARAMETER.
!
        failure = any( comp_norm(:m) )
!
    else
#endif
!
!       COMPUTE EIGENVECTORS BY INVERSE ITERATION.
!       SERIAL PATH.
!
        do j = 1_i4b, m0
!
            xj = lambda(j)
!    
            if ( fernando_init(j) ) then
!    
!               FERNANDO-BINI METHOD.  
!
                call trid_qr_cmp( d2(i1:i2), e2(i1:i2-1_i4b), xj, cs(i1:i2-1_i4b), sn(i1:i2-1_i4b),    &
                                  diag(i1:i2), sup1(i1:i2), sup2(i1:i2), mm )
!    
                mm = mm + i1 - 1_i4b
!
                if ( comp_norm(j) ) then
!
!                   ADD A SMALL PERTURBATION TO ALL COMPONENTS OF THE
!                   INITIAL VECTOR TO PROTECT AGAINST FAILURE OF FERNANDO
!                   METHOD WHICH MAY PRODUCE PARALLEL INITIAL VECTORS IF
!                   EIGENVALUES ARE TOO CLOSED.
!
                    call random_number_( temp(i1:i2) )
!
                    temp(i1:i2) = sqrt( eps )*( temp(i1:i2)*two - one )
!
                else
!    
                    temp(i1:i2) = zero
!
                end if
!
                temp(mm) = one
!
!               DO maxit ITERATIONS FOR THE EIGENVECTORS.
!
                do i=1_i4b, maxit
!
!                   SOLVE THE SYSTEM QRx = b WITH RESCALING TO AVOID OVERFLOW.
!
                    call trid_qr_solve( cs(i1:i2-1_i4b), sn(i1:i2-1_i4b), diag(i1:i2), sup1(i1:i2),  &
                                        sup2(i1:i2), temp(i1:i2)  )
!
!                   NORMALIZE THE EIGENVECTORS.
!
                    xj = norm( temp(i1:i2) )
!
                    if ( xj/=zero ) then
!
                        temp(i1:i2) = temp(i1:i2)/xj
!
                    else if ( i/=maxit ) then
!
                        call random_number_( temp(i1:i2) )
!
                        temp(i1:i2) = temp(i1:i2)*two - one
!
                    end if
!
                end do
!
            else
!
!               STANDART INVERSE ITERATION METHOD.
!
!               COMPUTE LU FACTORS WITH PARTIAL PIVOTING (PT=LU).
!
                call trid_cmp_rv( d2(i1:i2), e2(i1:i2), xj, sub(i1:i2), diag(i1:i2),  &
                                  sup1(i1:i2), sup2(i1:i2), perm(i1:i2) )
!    
!               COMPUTE A RANDOM INITIAL VECTOR.  
!
                call random_number_( temp(i1:i2) )
!
                temp(i1:i2) = temp(i1:i2)*two - one
!
!               REPLACE SMALL DIAGONAL ELEMENTS BY safmin.
!
                where( abs(diag(i1:i2))<safmin ) diag(i1:i2) = sign( safmin, diag(i1:i2) )
!
!               DO maxit ITERATIONS FOR THE EIGENVECTORS.
!
                do i=1_i4b, maxit
!
!                   SOLVE THE SYSTEM LU = Pb WITH RESCALING TO AVOID OVERFLOW.
!
                    call trid_solve_rv( sub(i1:i2), diag(i1:i2), sup1(i1:i2),   &
                                        sup2(i1:i2), perm(i1:i2), temp(i1:i2) )
!
!                   NORMALIZE THE EIGENVECTORS.
!
                    xj = norm( temp(i1:i2) )
!
                    if ( xj/=zero ) then
!
                        temp(i1:i2) = temp(i1:i2)/xj
!
                    else if ( i/=maxit ) then
!
                        call random_number_( temp(i1:i2) )
!
                        temp(i1:i2) = temp(i1:i2)*two - one
!
                    end if
!
                end do
!
            end if
!
!           SAVE THE EIGENVECTORS.
!
            if ( xj/=zero ) then
!
                eigvec(1_i4b:i1-1_i4b,j) = zero
                eigvec(i1:i2,j)          = temp(i1:i2)
                eigvec(i2+1_i4b:n,j)     = zero
!
            else
!
                eigvec(:n,j) = zero
!
            end if
!
            lambda(j) = xj
!
        end do
!
        if ( do_ortho ) then
!
!           REORTHOGONALIZE EIGENVECTORS BY MODIFIED GRAM-SCHMIDT.
!
            do l=1_i4b, k
!
                j1 = jstart(l)
                j2 = jend(l)
!
                mm = j2 + 1_i4b
!
!               FORWARD SWEEP OF MGS ORTHOGONALIZATION.
!
                do j=j1, j2 
!
                    sup1(i1:i2) = eigvec(i1:i2,j)
!
                    do i=j+1_i4b, mm
!
                        xj = -dot_product( sup1(i1:i2), eigvec(i1:i2,i) )
!
#ifdef _BLAS
                        call axpy( i3, xj, sup1(i1:i2), 1_i4b, eigvec(i1:i2,i), 1_i4b )
#else
                        eigvec(i1:i2,i) = eigvec(i1:i2,i) + xj*sup1(i1:i2)
#endif
!
                    end do
!
!                   NORMALIZE THE NEXT j EIGENVECTOR.
!
                    sup1(i1:i2) = eigvec(i1:i2,j+1_i4b)
!
                    xj = norm( sup1(i1:i2) )
!
                    if ( xj<=fudge_ortho2 ) then
!
!                       REORTHOGONALIZATION IS REQUIRED.
!
                        call ortho_vec(eigvec(i1:i2,j1:j), sup1(i1:i2), xj )
!
                    else
!
!                       NORMALIZATION IS ENOUGH.
!
                        sup1(i1:i2) = (one/xj)*sup1(i1:i2)
!
                    end if
!
                    eigvec(i1:i2,j+1_i4b) = sup1(i1:i2)
!
                    lambda(j+1_i4b) = xj
!
                end do
!
                if ( do_backward_sweep ) then
!
!                   BACKWARD SWEEP OF MGS ORTHOGONALIZATION.
!
                    do j=j2, j1, -1_i4b
!
                        sup1(i1:i2) = eigvec(i1:i2,j)
!
                        do i=j+1_i4b, mm
!
                            xj = -dot_product( sup1(i1:i2), eigvec(i1:i2,i) )
!
#ifdef _BLAS
                            call axpy( i3, xj, sup1(i1:i2), 1_i4b, eigvec(i1:i2,i), 1_i4b )
#else
                            eigvec(i1:i2,i) = eigvec(i1:i2,i) + xj*sup1(i1:i2)
#endif
!
                        end do
!
                    end do
!
                end if
!
            end do
!
        end if
!
!       CHECK ACCURACY OF THE EIGENVECTORS.
!
        if ( all( lambda(:m)/=zero ) ) then
!
            diag(1_i4b:n)       = d(1_i4b:n)
            sup1(1_i4b:n-1_i4b) = e(1_i4b:n-1_i4b)
            sup1(n)             = zero
!
            do j = 1_i4b, m
!
                temp(1_i4b:n) = eigvec(1_i4b:n,j)
!
                sub(1_i4b)         = diag(1_i4b)*temp(1_i4b) + sup1(1_i4b)*temp(2_i4b)
                sub(2_i4b:n-1_i4b) = sup1(1_i4b:n-2_i4b)*temp(1_i4b:n-2_i4b)           +   &
                                     diag(2_i4b:n-1_i4b)*temp(2_i4b:n-1_i4b)           +   &
                                     sup1(2_i4b:n-1_i4b)*temp(3_i4b:n)
                sub(n)             = sup1(n-1_i4b)*temp(n-1_i4b) + diag(n)*temp(n)
!
!               COMPUTE RAYLEIGH QUOTIENT.
!
                xj    = dot_product( sub(1_i4b:n), temp(1_i4b:n) )
                e2(j) = xj
!
                if ( .not.failure ) then
!
!                   SET OUTPUT FAILURE PARAMETER.
!
!                   COMPUTE THE RESIDUAL NORM OF THE JTH EIGENVECTOR
!                   WITH RESPECT TO ITS RAYLEIGH QUOTIENT.
!
#ifdef _BLAS
                    call axpy( n, -xj, temp(1_i4b:n), 1_i4b, sub(1_i4b:n), 1_i4b )
#else
                    sub(1_i4b:n) = sub(1_i4b:n) - xj*temp(1_i4b:n)
#endif
                    xj = sum( abs( sub(:n) ) )
!
                    failure = xj>lim
!
                end if
!
            end do
!
        else
!
            failure  = true
            zero_vec = true
!
        end if
!
#ifdef _OPENMP
    end if
#endif
!
    if ( .not.zero_vec ) then
!
!       CHECK THE PAIRING OF EIGENVALUES AND EIGENVECTORS.
!
        if ( do_ortho ) then
!
            do l=1_i4b, k
!
                j1 = jstart(l)
                j2 = jend(l)
!
                do j=j1, j2
!
                    if ( e2(j)<e2(j+1_i4b) ) exit
!
                end do
!
                if ( j<=j2 ) then
!
                    call eig_sort_a( 'd', e2(j1:j2+1_i4b), eigvec(:n,j1:j2+1_i4b) )
!
                end if
!
                if ( l==1_i4b ) cycle
!
                if ( .not.failure ) then
!
                    j2 = jend(l-1_i4b) + 1_i4b
!
                    failure = e2(j2)<e2(j1)
!
                end if
!
            end do
!
!           DEALLOCATE WORK ARRAYS.
!
            deallocate( jstart, jend )
!
        else if ( .not.failure ) then
!
            do i=1_i4b, m0-1_i4b
!
                if ( e2(i)<e2(i+1_i4b) ) exit
!
            end do
!
!            failure = i<m0
!
            if ( i<m0 ) then
!
                call eig_sort_a( 'd', e2(:m0), eigvec(:n,:m0) )
!
            end if
!
        end if
!
!       CHECK FINALLY IF THE RAYLEIGH QUOTIENTS AND THE EIGENVALUES AGREE.
!
        if ( .not.failure ) then
!
            xj = maxval( abs( e2(1_i4b:m) - eigval(1_i4b:m) ) )
            failure = xj>lim
!
        end if
!
    else if ( do_ortho ) then
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( jstart, jend )
!
    end if
!
!
! END OF SUBROUTINE trid_inviter_a
! ________________________________
!
    end subroutine trid_inviter_a
!
! =========================================================================================
!
    subroutine trid_inviter_b( d, e, eigval, eigvec, failure, mat, maxiter, ortho, &
                               backward_sweep, scaling, initvec                    )
!
! Purpose
! _______
!
!   TRID_INVITER computes the eigenvectors of a full real n-by-n symmetric matrix MAT
!   corresponding to specified eigenvalues, by combining Fernando's method for computing
!   (selected) eigenvectors of a real n-by-n symmetric tridiagonal matrix and inverse
!   iteration followed by a back-transformation procedure.
!
!   It is required that the original symmetric matrix MAT has been reduced to symmetric
!   tridiagonal form T by an orthogonal similarity transformation:
!
!                        Q' * MAT * Q = T
!
!   with a call to SYMTRID_CMP with parameter STORE_Q set to true, before
!   calling TRID_INVITER.
!
!
! Arguments
! _________
!
!   D              (INPUT) real(stnd), dimension(:)
!                  On entry, the diagonal elements of the symmetric tridiagonal form T
!                  of MAT.
!
!   E              (INPUT) real(stnd), dimension(:)
!                  On entry, the n-1 subdiagonal elements of the symmetric tridiagonal
!                  form T of MAT. E(n) is arbitrary .
!
!                  The size of E must verify:  size( E ) = size( D ) = n .
!
!   EIGVAL         (INPUT) real(stnd), dimension(:)
!                  On entry, selected eigenvalues of the symmetric matrix MAT.
!                  The eigenvalues must be given in decreasing order.
!
!                  The size of EIGVAL must verify: size( EIGVAL ) <= size( D ) = n .
!
!   EIGVEC         (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed eigenvectors. The eigenvector associated
!                  with the eigenvalue EIGVAL(j) is stored in the j-th column of EIGVEC.
!
!                  The shape of EIGVEC must verify: 
!
!                  - size( EIGVEC, 1 ) = size( D ) = n ;
!                  - size( EIGVEC, 2 ) = size( EIGVAL ) .
!
!   FAILURE        (OUTPUT) logical(lgl)
!                  On exit:
!
!                  - FAILURE = false indicates successful exit.
!                  - FAILURE = true indicates that some eigenvectors
!                    failed to converge in MAXITER iterations.
!
!   MAT            (INPUT) real(stnd), dimension(:,:)
!                  On entry, the vectors and the scalars which define the elementary
!                  reflectors used to reduce the full real n-by-n symmetric matrix MAT to
!                  symmetric tridiagonal form T, as returned by SYMTRID_CMP with STORE_Q=true,
!                  in its argument MAT. MAT is not modified by the routine.
!
!                  Back-transformation is used to find the selected eigenvectors of the original
!                  matrix MAT and these eigenvectors are stored in argument EIGVEC.
!
!                  The shape of MAT must verify: 
!
!                  - size( MAT, 1 ) = size( D ) = n ;
!                  - size( MAT, 2 ) = size( D ) = n .
!
!   MAXITER        (INPUT, OPTIONAL) integer(i4b)
!                  The number of inverse iterations performed in the subroutine. By default,
!                  2 inverse iterations are performed for all the eigenvectors of the tridiagonal
!                  matrix T.
!
!   ORTHO          (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - ORTHO=true, all the eigenvectors are orthogonalized by the
!                    Modified Gram-Schmidt algorithm;
!                  - ORTHO=false, the eigenvectors are not orthogonalized by 
!                    the Modified Gram-Schmidt algorithm.
!
!                  The default is to orthogonalize the eigenvectors only if the eigenvalues
!                  are not well-separated.
!
!   BACKWARD_SWEEP (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - BACKWARD_SWEEP=true and the eigenvectors are orthogonalized by
!                    the modified Gram-Schmidt algorithm, a backward sweep of the modified Gram-Schmidt
!                    algorithm is also performed;
!                  - BACKWARD_SWEEP=false a backward sweep is not performed.
!
!                  The default is not to perform a backward sweep of the modified Gram-Schmidt
!                  algorithm.
!
!   SCALING        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - SCALING=true, the tridiagonal matrix T is scaled before computing
!                    the eigenvectors;
!                  - SCALING=false, the tridiagonal matrix T is not scaled.
!
!                  The default is to scale the tridiagonal matrix.
!
!   INITVEC        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - INITVEC=true, Fernando vectors are used to start the inverse
!                    iteration process;
!                  - INITVEC=false, random uniform starting vectors are used.
!
!                  For unreduced tridiagonal matrices, the default is to use Fernando starting vectors
!                  if the eigenvalues are well-separated and random uniform starting vectors otherwise.
!                  For reduced tridiagonal matrices, the default is to use random uniform starting vectors.
!
!
! Further Details
! _______________
!
!   TRID_INVITER uses Fernando's method for computing a first estimate of (selected)
!   eigenvectors corresponding to (selected) approximate eigenvalues of a real n-by-n
!   symmetric tridiagonal matrix T (by default, only for the eigenvalues which are
!   well separated and if the input tridiagonal matrix T is unreduced). See the references
!   (1), (2) and (4) for details.
!
!   These approximate eigenvectors are then refined (or computed if Fernando's method is
!   not used) using an inverse iteration algorithm for all the eigenvalues at one step. See
!   the reference (3) for details.
!
!   The eigenvectors are then orthogonalized by the Modified Gram-Schmidt algorithm if clusters
!   of eigenvalues are present, in a second step.
!
!   In a last step, the corresponding (selected) eigenvectors of the full real n-by-n symmetric
!   matrix MAT are computed by a blocked back-transformation algorithm with the Householder
!   transformations used to reduce the full real n-by-n symmetric matrix MAT to symmetric
!   tridiagonal form T (see the references (5) and (6)).
!
!   Furthermore, the computation of the eigenvectors is parallelized if OPENMP is used.
!
!   TRID_INVITER may fail if some the eigenvalues specified in parameter EIGVAL are nearly
!   identical.
!
!   For further details on Fernando's method or inverse iteration for computing eigenvectors
!   of tridiagonal matrices or the blocked back-transformation algorithm, see:
!
!   (1) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, pp. 1013-1034.
!
!   (2) Parlett, B.N., and Dhillon, I.S., 1997:
!           Fernando's solution to Wilkinson's problem: An application of double factorization.
!           Linear Algebra and its Appl., 267, pp.247-279.
!
!   (3) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (4) Bini, D.A., Gemignani, L., and Tisseur, F., 2005:
!           The Ehrlich-Aberth method for the nonsymmetric tridiagonal eigenvalue problem.
!           SIAM J. Matrix Anal. Appl., 27, 153-175.
!
!   (5) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989:
!           Block reduction of matrices to condensed form for eigenvalue computations. 
!           J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!   (6) Walker, H.F., 1988:
!           Implementation of the GMRES method using Householder transformations.
!           Siam J. Sci. Stat. Comput., Vol. 9, No 1, pp. 152-163.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,    only : machhuge
    use Char_Constants,   only : eig_error1
    use Utilities,        only : assert_eq, assert, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),  dimension(:)     :: d, e, eigval
    real(stnd),  intent(in),  dimension(:,:)   :: mat
    real(stnd),  intent(out), dimension(:,:)   :: eigvec
!
    integer(i4b), intent(in),  optional :: maxiter
!
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in),  optional :: ortho, backward_sweep, scaling, initvec
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: n, m
!
    logical(lgl) :: upper
!
#ifdef _OPENMP
!
    integer(i4b) :: i, num_threads, thread_num, jinc, jrem, joffset, j1, j2
!
    logical :: test_par, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='trid_inviter'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(d),i4b)        ,   &
                    int(size(e),i4b)        ,   &
                    int(size(mat,1),i4b)    ,   &
                    int(size(mat,2),i4b)    ,   &
                    name_proc )
!
    call assert( logical(size(eigvec,1)==n,lgl) ,  &
                 name_proc )
!
    m =  assert_eq( int(size(eigval),i4b)   ,   &
                    int(size(eigvec,2),i4b) ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. m<=0_i4b ) then
!
        return
!
    end if
!
    if ( max( mat(1_i4b,1_i4b), mat(n,n) )==machhuge ) then
!
        upper = mat(1_i4b,1_i4b)==machhuge
!
    else
!
        call merror( name_proc//eig_error1 )
!
    end if
!
!   COMPUTE EIGENVECTORS OF THE SYMMETRIC TRIDIAGONAL FORM T OF mat.
!
    call trid_inviter_a( d, e, eigval, eigvec, failure, maxiter=maxiter,  &
                         ortho=ortho, backward_sweep=backward_sweep,      &
                         scaling=scaling, initvec=initvec                 )
!
!   COMPUTE EIGENVECTORS OF THE FULL SYMMETRIC MATRIX mat BY BACK-TRANSFORMATION.
!
#ifdef _OPENMP
    i = omp_get_max_threads()
    ompnested   = omp_get_nested()
    test_par    = ompnested             .and.      &
                  n*m>=omp_limit        .and.      &
                  m/2_i4b>=omp_limit2   .and.      &
                  i>1_i4b
!
    if ( test_par ) then
!
!$OMP PARALLEL NUM_THREADS( 2 )                                      &
!$OMP         ,PRIVATE(j1,j2,jrem,jinc,joffset,num_threads,thread_num)
!
!       DISTRIBUTE THE WORK AMONG THE THREADS.
!
        num_threads = omp_get_num_threads()
        thread_num  = omp_get_thread_num()
!
        jrem = mod( m, num_threads )
        jinc = ( m - jrem )/num_threads
!
        if ( thread_num<jrem ) then
            jinc    = jinc + 1_i4b
            joffset = 1_i4b
        else
            joffset = jrem + 1_i4b
        end if
!
        j1 = thread_num*jinc + joffset
        j2 = j1 + jinc - 1_i4b
!
!       GENERATE EIGENVECTORS OF mat IN eigvec .
!
        if ( upper ) then
!
            call apply_q_symtrid_b( mat, eigvec(:n,j1:j2), left=true, trans=false )
!
        else
!
            call apply_q_symtrid_a( mat, eigvec(:n,j1:j2), left=true, trans=false, upper=false )
!
        end if
!
!$OMP END PARALLEL
!
    else
#endif
!
!       GENERATE EIGENVECTORS OF mat IN eigvec .
!
        if ( upper ) then
!
            call apply_q_symtrid_b( mat, eigvec(:n,:m), left=true, trans=false )
!
        else
!
            call apply_q_symtrid_a( mat, eigvec(:n,:m), left=true, trans=false, upper=false )
!
        end if
!
#ifdef _OPENMP
    end if
#endif
!
!
! END OF SUBROUTINE trid_inviter_b
! ________________________________
!
    end subroutine trid_inviter_b
!
! =========================================================================================
!
    subroutine trid_inviter_c( d, e, eigval, eigvec, failure, matp,             &
                               maxiter, ortho, backward_sweep, scaling, initvec )
!
! Purpose
! _______
!
!   TRID_INVITER computes the eigenvectors of a full real n-by-n symmetric matrix MAT,
!   packed column-wise in a linear array  MATP, corresponding to specified eigenvalues,
!   using  Fernando's method for computing (selected) eigenvectors of a real n-by-n
!   symmetric tridiagonal matrix and inverse iteration, followed by a back-transformation
!   procedure.
!
!   It is required that the original packed symmetric matrix MAT has been reduced to symmetric
!   tridiagonal form T by an orthogonal similarity transformation:
!
!                        Q' * MAT * Q = T
!
!   with a call to SYMTRID_CMP with parameter STORE_Q set to true, before
!   calling TRID_INVITER.
!
!
! Arguments
! _________
!
!   D              (INPUT) real(stnd), dimension(:)
!                  On entry, the diagonal elements of the symmetric tridiagonal form T
!                  of MAT.
!
!   E              (INPUT) real(stnd), dimension(:)
!                  On entry, the n-1 subdiagonal elements of the symmetric tridiagonal
!                  form T of MAT. E(n) is arbitrary .
!
!                  The size of E must verify:  size( E ) = size( D ) = n .
!
!   EIGVAL         (INPUT) real(stnd), dimension(:)
!                  On entry, selected eigenvalues of the symmetric matrix MAT.
!                  The eigenvalues must be given in decreasing order.
!
!                  The size of EIGVAL must verify: size( EIGVAL ) <= size( D ) = n .
!
!   EIGVEC         (OUTPUT) real(stnd), dimension(:,:)
!                  On exit, the computed eigenvectors. The eigenvector associated
!                  with the eigenvalue EIGVAL(j) is stored in the j-th column of EIGVEC.
!
!                  The shape of EIGVEC must verify: 
!
!                  - size( EIGVEC, 1 ) = size( D ) = n .
!                  - size( EIGVEC, 2 ) = size( EIGVAL ) ,
!
!   FAILURE        (OUTPUT) logical(lgl)
!                  On exit:
!
!                  - FAILURE = false indicates successful exit.
!                  - FAILURE = true indicates that some eigenvectors
!                    failed to converge in MAXITER iterations.
!
!   MATP           (INPUT) real(stnd), dimension(:)
!                  On entry, the vectors and the scalars which define the elementary
!                  reflectors used to reduce the packed real n-by-n symmetric matrix MAT to
!                  symmetric tridiagonal form T, as returned by SYMTRID_CMP with STORE_Q=true,
!                  in its argument MATP. MATP is not modified by the routine.
!
!                  Back-transformation is used to find the selected eigenvectors of the original
!                  matrix MAT and these eigenvectors are stored in argument EIGVEC.
!
!                  The size of MATP must verify: size( MATP ) = (n * (n+1)/2)
!
!   MAXITER        (INPUT, OPTIONAL) integer(i4b)
!                  The number of inverse iterations performed in the subroutine. By default,
!                  2 inverse iterations are performed for all the eigenvectors of the tridiagonal
!                  matrix T.
!
!   ORTHO          (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - ORTHO=true, all the eigenvectors are orthogonalized by the
!                    Modified Gram-Schmidt algorithm;
!                  - ORTHO=false, the eigenvectors are not orthogonalized by 
!                    the Modified Gram-Schmidt algorithm.
!
!                  The default is to orthogonalize the eigenvectors only if the eigenvalues
!                  are not well-separated.
!
!   BACKWARD_SWEEP (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - BACKWARD_SWEEP=true and the eigenvectors are orthogonalized by
!                    the modified Gram-Schmidt algorithm, a backward sweep of the modified Gram-Schmidt
!                    algorithm is also performed;
!                  - BACKWARD_SWEEP=false a backward sweep is not performed.
!
!                  The default is not to perform a backward sweep of the modified Gram-Schmidt
!                  algorithm.
!
!   SCALING        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - SCALING=true, the tridiagonal matrix T is scaled before computing
!                    the eigenvectors;
!                  - SCALING=false, the tridiagonal matrix T is not scaled.
!
!                  The default is to scale the tridiagonal matrix.
!
!   INITVEC        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if:
!
!                  - INITVEC=true, Fernando vectors are used to start the inverse
!                    iteration process;
!                  - INITVEC=false, random uniform starting vectors are used.
!
!                  For unreduced tridiagonal matrices, the default is to use Fernando starting vectors
!                  if the eigenvalues are well-separated and random uniform starting vectors otherwise.
!                  For reduced tridiagonal matrices, the default is to use random uniform starting vectors.
!
!
! Further Details
! _______________
!
!   TRID_INVITER uses Fernando's method for computing a first estimate of (selected)
!   eigenvectors corresponding to (selected) approximate eigenvalues of a real n-by-n
!   symmetric tridiagonal matrix T (by default, only for the eigenvalues which are
!   well separated and if the input tridiagonal matrix T is unreduced). See the references
!   (1), (2) and (4) for details.
!
!   These approximate eigenvectors are then refined (or computed if Fernando's method is
!   not used) using an inverse iteration algorithm for all the eigenvalues at one step. See
!   the reference (3) for details.
!
!   The eigenvectors are then orthogonalized by the Modified Gram-Schmidt algorithm if clusters
!   of eigenvalues are present in a second step.
!
!   In a final step, the corresponding (selected) eigenvectors of the full real n-by-n symmetric
!   matrix MAT are computed by a blocked back-transformation algorithm with the Householder
!   transformations used to reduce the full real n-by-n symmetric matrix MAT to symmetric
!   tridiagonal form T (see the references (5) and (6)). These Householder transformations
!   must be packed in the linear array MATP (as returned by SYMTRID_CMP) on entry of TRID_INVITER.
!
!   Furthermore, the computation of the eigenvectors is parallelized if OPENMP is used.
!
!   TRID_INVITER may fail if some the eigenvalues specified in parameter EIGVAL are nearly
!   identical.
!
!   For further details on Fernando's method or inverse iteration for computing eigenvectors
!   of tridiagonal matrices or the blocked back-transformation algorithm, see:
!
!   (1) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, pp. 1013-1034.
!
!   (2) Parlett, B.N., and Dhillon, I.S., 1997:
!           Fernando's solution to Wilkinson's problem: An application of double factorization.
!           Linear Algebra and its Appl., 267, pp.247-279.
!
!   (3) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (4) Bini, D.A., Gemignani, L., and Tisseur, F., 2005:
!           The Ehrlich-Aberth method for the nonsymmetric tridiagonal eigenvalue problem.
!           SIAM J. Matrix Anal. Appl., 27, 153-175.
!
!   (5) Dongarra, J.J., Sorensen, D.C., and Hammarling, S.J., 1989:
!           Block reduction of matrices to condensed form for eigenvalue computations. 
!           J. of Computational and Applied Mathematics, Vol. 27, pp. 215-227.
!
!   (6) Walker, H.F., 1988:
!           Implementation of the GMRES method using Householder transformations.
!           Siam J. Sci. Stat. Comput., Vol. 9, No 1, pp. 152-163.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,    only : machhuge
    use Char_Constants,   only : eig_error1
    use Utilities,        only : assert, assert_eq, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),  dimension(:)     :: d, e, eigval, matp
    real(stnd),  intent(out), dimension(:,:)   :: eigvec
!
    integer(i4b), intent(in),  optional :: maxiter
!
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in),  optional :: ortho, backward_sweep, scaling, initvec
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: n, m, i
!
    logical(lgl) :: upper
!
#ifdef _OPENMP
!
    integer(i4b) :: num_threads, thread_num, jinc, jrem, joffset, j1, j2
!
    logical :: test_par, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='trid_inviter'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(d),i4b)        ,   &
                    int(size(e),i4b)        ,   &
                    int(size(eigvec,1),i4b) ,   &
                    name_proc )
!
    m =  assert_eq( int(size(eigval),i4b)   ,   &
                    int(size(eigvec,2),i4b) ,   &
                    name_proc )
!
    i = n*(n+1_i4b)/2_i4b
    call assert( logical(size(matp)==i,lgl)  ,   &
                 name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. m<=0_i4b ) then
!
        return
!
    end if
!
    if ( max( matp(1_i4b), matp(i) )==machhuge ) then
!
        upper = matp(1_i4b)==machhuge
!
    else
!
        call merror( name_proc//eig_error1 )
!
    end if
!
!   COMPUTE EIGENVECTORS OF THE SYMMETRIC TRIDIAGONAL FORM T OF mat.
!
    call trid_inviter_a( d, e, eigval, eigvec, failure, maxiter=maxiter,  &
                         ortho=ortho, backward_sweep=backward_sweep,      &
                         scaling=scaling, initvec=initvec                 )
!
!   COMPUTE EIGENVECTORS OF THE FULL SYMMETRIC MATRIX mat BY BACK-TRANSFORMATION.
!
#ifdef _OPENMP
    i = omp_get_max_threads()
    ompnested   = omp_get_nested()
    test_par    = ompnested             .and.      &
                  n*m>=omp_limit        .and.      &
                  m/2_i4b>=omp_limit2   .and.      &
                  i>1_i4b
!
    if ( test_par ) then
!
!$OMP PARALLEL NUM_THREADS( 2 )                                      &
!$OMP         ,PRIVATE(j1,j2,jrem,jinc,joffset,num_threads,thread_num)
!
!       DISTRIBUTE THE WORK AMONG THE THREADS.
!
        num_threads = omp_get_num_threads()
        thread_num  = omp_get_thread_num()
!
        jrem = mod( m, num_threads )
        jinc = ( m - jrem )/num_threads
!
        if ( thread_num<jrem ) then
            jinc    = jinc + 1_i4b
            joffset = 1_i4b
        else
            joffset = jrem + 1_i4b
        end if
!
        j1 = thread_num*jinc + joffset
        j2 = j1 + jinc - 1_i4b
!
!       GENERATE EIGENVECTORS OF mat IN eigvec .
!
        if ( upper ) then
!
            call apply_q_symtrid_d( matp, eigvec(:n,j1:j2), left=true, trans=false )
!
        else
!
            call apply_q_symtrid_c( matp, eigvec(:n,j1:j2), left=true, trans=false, upper=false )
!
        end if
!
!$OMP END PARALLEL
!
    else
#endif
!
!       GENERATE EIGENVECTORS OF mat IN eigvec .
!
        if ( upper ) then
!
            call apply_q_symtrid_d( matp, eigvec(:n,:m), left=true, trans=false )
!
        else
!
            call apply_q_symtrid_c( matp, eigvec(:n,:m), left=true, trans=false, upper=false )
!
        end if
!
#ifdef _OPENMP
    end if
#endif
!
!
! END OF SUBROUTINE trid_inviter_c
! ________________________________
!
    end subroutine trid_inviter_c
!
!
! =========================================================================================
!                   PROCEDURES FOR GENERATION OF SYMMETRIC TRIDIAGONAL MATRICES
! =========================================================================================
!
!
    subroutine gen_symtrid_mat( type, d, e, failure, known_eigval, eigval, sort, val1, val2, l0, glu0 )
!
! Purpose
! _______
!
!   GEN_SYMTRID_MAT generates different types of symmetric tridiagonal matrices with known
!   eigenvalues or specific numerical properties such as clustered eigenvalues for testing
!   purposes of eigensolvers.
!
!   Optionally, the eigenvalues of the selected symmetric tridiagonal matrix can be computed
!   analytically, if possible, or by a bisection algorithm with high absolute and relative accuracies.
!
!
! Arguments
! _________
!
!   TYPE                (INPUT) integer(i4b)
!                       Select the type of symmetric tridiagonal matrix TRID to be generated
!                       by the subroutine.
!
!                       If TYPE is between 1 and 49, the subroutine generates a specific symmetric
!                       tridiagonal matrix as described in the comments inside the code of the subroutine.
!                       For other values of TYPE, all diagonal and off-diagonal elements of the symmetric
!                       tridiagonal matrix are generated from an uniform random numbers distribution between
!                       0 and 1.
!
!                       For TYPE between 1 and 17, the eigenvalues of the tridiagonal symetric matrix are
!                       known analytically. For other values of TYPE, the eigenvalues are estimated by a
!                       bisection algorithm with high accuracy.
!
!                       In all cases, the eigenvalues may be output in the optional parameter EIGVAL.
!
!   D                   (OUTPUT) real(stnd), dimension(:)
!                       On exit, D contains the diagonal elements of the tridiagonal matrix TRID.
!
!                       The size of D must verify: size( D )>=2 . 
!
!   E                   (OUTPUT) real(stnd), dimension(:)
!                       On exit, E contains the off-diagonal elements of the tridiagonal matrix TRID.
!                       E(size(E)) is arbitrary.
!
!                       The size of E must verify: size( E ) = size( D ) . 
!
!   FAILURE             (OUTPUT, OPTIONAL) logical(lgl)
!                       On exit:
!
!                       - FAILURE = false :  indicates that the eigenvalues of TRID are known analytically
!                         or have been computed with high accuracy;
!                       - FAILURE = true  :  indicates that the eigenvalues of TRID are not known analytically
!                         and have not been computed with maximum accuracy with the bisection
!                         algorithm.
!
!   KNOWN_EIGVAL        (OUTPUT, OPTIONAL) logical(lgl)
!                       On exit:
!
!                       - KNOWN_EIGVAL = true  :  indicates that the eigenvalues of TRID are known analytically for
!                         the selected TYPE.
!                       - KNOWN_EIGVAL = false :  indicates that the eigenvalues of TRID are not known analytically for
!                         the selected TYPE.
!
!   EIGVAL              (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                       On exit, the eigenvalues of TRID computed analytically or estimated to high
!                       accuracy with a bisection algorithm.
!
!                       The size of EIGVAL must verify: size( EIGVAL ) = size( D ) . 
!
!   SORT                (INPUT, OPTIONAL) character
!                       Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!                       or in descending order if SORT = 'D' or 'd', if the optional
!                       argument EIGVAL is present. For other values of SORT
!                       nothing is done and EIGVAL(:) may not be sorted.
!
!   VAL1                (INPUT, OPTIONAL) real(stnd)
!                       On entry, specifies the parameter d0 for parametrized symmetric
!                       tridiagonal matrices (e.g. TYPE= 3-4, 6-10, 12, 35-38).
!
!                       If this parameter is changed for TYPE between 35 and 38,
!                       which correspond to graded (or reversely graded) matrices
!                       with an arithmetic or geometric progression, care must be taken
!                       to insure that some elements of the arithmetic or geometric progression
!                       will not underflow or overflow as no checks are done in the subroutine
!                       for such errors.
!
!                       The default is 1. .
!
!   VAL2                (INPUT, OPTIONAL) real(stnd)
!                       On entry, specifies the parameter e0 for parametrized symmetric
!                       tridiagonal matrices  (e.g. TYPE= 3-4, 6-10, 12, 35-38).
!
!                       If this parameter is changed for TYPE between 35 and 38,
!                       which correspond to graded (or reversely graded) matrices
!                       with an arithmetic or geometric progression, care must be taken
!                       to insure that some elements of the arithmetic or geometric progression
!                       will not underflow or overflow as no checks are done in the subroutine
!                       for such errors.
!
!                       The default is 2. .
!
!   L0                  (INPUT, OPTIONAL) integer(i4b)
!                       On entry, specify the radius of the initial matrix for parametrized
!                       form of glued tridiagonal matrices (e.g. TYPE between 45 and 49).
!
!                       L0 must be greater than 0 and preferably less or equal to size( D )/2 .
!                       The default is 5. .
!
!   GLU0                (INPUT, OPTIONAL) real(stnd)
!                       On entry, specify the glue parameter for parametrized
!                       form of glued tridiagonal matrices (e.g. TYPE between 45 and 49).
!
!                       The default is sqrt( epsilon(GLU0) ).
!
!
! Further Details
! _______________
!
!   This subroutine tries to take care of imprecisions in intrinsic subroutines (e.g. like the cos
!   function in the gfortran compiler) when computing eigenvalues by analytic formulae.
!
!   For further details on the tridiagonal matrices used for testing in GEN_SYMTRID_MAT subroutine,
!   see:
!
!   (1) Gladwell, G.M.L., Jones, T.H., Willms N.B., 2014:
!           A test matrix for an inverse eigenvalue problem.
!           Journal of Applied Mathematics, 14, 6 pages, Article ID 515082, DOI 10.1155/2014/515082.
!
!   (2) Clement, P.A., 1959:
!           A class of triple-diagonal matrices for test purposes.
!           SIAM Review, 1(1):50-52, DOI 10.1137/1001006.
!
!   (3) Gregory, R.T., Karney, D.L., 1969:
!           A collection of matrices for testing computational algorithms.
!           New York: Wiley. Reprinted with corrections by Robert E. Krieger,
!           Huntington, New York, 1978.
!
!   (4) Higham, N.J., 1991: Algorithm 694:
!           A collection of test matrices in MATLAB.
!           ACM Transactions on Mathematical Software 17(3):289-305 DOI 10.1145/114697.116805.
!
!   (5) Godunov, S.K., Antonov, A.G., Kirillyuk, O.P., and Kostin, V.I., 1993:
!           Guaranteed Accuracy in numerical linear algebra.
!           Kluwer Academic Publishers.
!
!   (6) Parlett, B.N., and Vomel, C., 2005:
!           How the MRRR algorithm can fail on tight eigenvalue clusters.
!           Lapack Working Note 163.
!
!   (7) Nakatsukasa, Y., Aishima, K., and Yamazaki, I., 2012:
!           dqds with agressive early deflation.
!           SIAM J. Matrix Anal. Appl., 33(1): 22-51.
!
!   (8) Fernando, K.V., and Parlett, B.N., 1994:
!           Accurate singular values and differenial qd algorithms.
!           Numer. Math., 67: 191-229.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,    only : allocate_error
    use Reals_Constants,   only : c16, c200, c900, c1_01, c1_m2, pi
    use Num_Constants,     only : lamch
    use Utilities,         only : assert_eq, assert, merror, arth, geop
    use Random,            only : random_number_
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(out),           dimension(:) :: d, e
    real(stnd), intent(out), optional, dimension(:) :: eigval
    real(stnd), intent(in),  optional               :: val1, val2, glu0
!
    integer(i4b), intent(in)           :: type
    integer(i4b), intent(in), optional :: l0
!
    logical(lgl), intent(out), optional :: failure, known_eigval
!
    character,   intent(in), optional :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                            :: abstol, ulp, safmin, a, b, c, d0, e0, lambda, glu, tmp
    real(stnd), allocatable, dimension(:) :: temp, temp2
!
    integer      :: iok
    integer(i4b) :: n, i, i1, i2, j, l
!
    logical(lgl)  :: test_eigval, fail
!
!
! PARAMETERS
! __________
!
!
    character(len=*),  parameter :: name_proc='gen_symtrid_mat'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq( int(size(d),i4b),       &
                   int(size(e),i4b),       & 
                   name_proc )
!
    if ( present(eigval) ) then
!
        call assert( logical( int(size(eigval),i4b)==n, lgl ),  &
                     logical( 2_i4b<=n, lgl ),                  &
                     name_proc )
!
        test_eigval = true
!
        eigval(:n) = zero
!
    else
!
        call assert( logical( 2_i4b<=n, lgl ),  &
                     name_proc )
!
        test_eigval = false
!
    end if
!
!   SET OPTIONAL ARGUMENTS IF PRESENT.
!
    if ( present(failure) ) then
!
        failure = false
!
    end if
!
    if ( present(known_eigval) ) then
!
        known_eigval = 1_i4b<=type .and. type<=17_i4b
!
    end if
!
    d0 = one
!
    if ( present(val1) ) then
        d0 = val1
    end if
!
    e0 = two
!
    if ( present(val2) ) then
        e0 = val2
    end if
!
!   GET MACHINE PARAMETERS.
!
    ulp    = epsilon( ulp )
    safmin = lamch('s')
    abstol = sqrt( safmin )
!
!   SPECIFY THE RADIUS l FOR PARAMETRIZED FORM OF GLUED
!   TRIDIAGONAL MATRICES.
!
    l = 5_i4b
!
    if ( present(l0) ) then
!
        if ( l0>0_i4b ) then
            l = l0
        end if
!
    end if
!
!   SPECIFY THE GLUE PARAMETER FOR PARAMETRIZED FORM OF GLUED
!   SYMMETRIC TRIDIAGONAL MATRICES.
!
    glu = sqrt( ulp )
!
    if ( present(glu0) ) then
!
        glu = glu0
!
    end if
!
!   ALLOCATE WORK ARRAYS.
!
    i1 = max( n, 21_i4b, (2_i4b*l)+1_i4b )
    i2 = max( n, 42_i4b, (4_i4b*l)-2_i4b )
!
    allocate( temp(i1), temp2(i2), stat=iok )
!
    if ( iok/=0 ) then
        call merror( name_proc//allocate_error )
    end if
!
!   GENERATE A TRIDIAGONAL MATRIX.
!   THE DIAGONAL ELEMENTS ARE STORED IN d .
!   THE OFF-DIAGONAL ELEMENTS ARE STORED IN e .
!   IF THE EIGENVALUES ARE REQUESTED, THEY ARE
!   COMPUTED AND STORED IN eigval(:n).
!
    select case( type )
!
        case( 1_i4b )
!
!           CASE 1: GENERATE A JACOBIAN TRIDIAGONAL MATRIX WITH
!                   EIGENVALUES 0, 2, ..., 2*n-2.
!                   ALL DIAGONAL ELEMENTS ARE EQUAL TO n-1 AND
!                   OFF-DIAGONAL ELEMENT e(j) IS EQUAL 0.5*sqrt( j*(2*n-j-1) )
!                   FOR j=1 to n-2 AND e(n-1) IS EQUAL TO sqrt( 0.5*n*(n-1) ) .
!                   SEE GLADWELL ET AL.
!
            d(:n) = real( n-1_i4b, stnd )
!
            do j = 1_i4b, n-2_i4b
                e(j) = half*sqrt( real( j*(2_i4b*n-j-1_i4b), stnd ) )
            end do
!
            e(n-1_i4b) = sqrt( real( n*(n-1_i4b), stnd )*half )
!
            if ( test_eigval ) then
!
                eigval(:n) = arth( zero, two, n )
!
            end if
!
        case( 2_i4b )
!
!           CASE 2: GENERATE A PSEUDO-JACOBIAN TRIDIAGONAL MATRIX WITH
!                   EIGENVALUES 1, 3, ..., 2*n-3 AND n.
!                   d(1) IS EQUAL TO n AND ALL OTHER DIAGONAL ELEMENTS ARE EQUAL TO n-1 .
!                   e(1) IS EQUAL TO 0 AND OFF-DIAGONAL ELEMENT e(j) IS EQUAL 0.5*sqrt( (j-1)*(2*n-j) )
!                   FOR j=2 to n-1 .
!                   SEE GLADWELL ET AL.
!
            d(1_i4b)   = real( n, stnd )
            d(2_i4b:n) = real( n-1_i4b, stnd )
!
            e(1_i4b) = zero
!
            do j = 2_i4b, n-1_i4b
                e(j) = half*sqrt( real( (j-1_i4b)*(2_i4b*n-j), stnd ) )
            end do
!
            if ( test_eigval ) then
!
                eigval(:n-1_i4b) = arth( one, two, n-1_i4b )
                eigval(n)        = d(1_i4b)
!
            end if
!
        case( 3_i4b )
!
!           CASE 3: GENERATE A JACOBIAN TRIDIAGONAL MATRIX WITH
!                   EIGENVALUES d0, d0+2*e0, d0+4*e0, d0+6*e0, ..., d0+(2*n-2)*e0.
!                   SEE GLADWELL ET AL.
!
            a  = d0 + e0*real( n-1_i4b, stnd ) 
!
            d(:n) = a
!
            do j = 1_i4b, n-2_i4b
                e(j) = -e0*half*sqrt( real( j*(2_i4b*n-j-1_i4b), stnd ) )
            end do
!
            e(n-1_i4b) = -e0*sqrt( real( n*(n-1_i4b), stnd )*half )
!
            if ( test_eigval ) then
!
                eigval(:n) = arth( d0, two*e0, n )
!
            end if
!
        case( 4_i4b )
!
!           CASE 4: GENERATE A PSEUDO JACOBIAN TRIDIAGONAL MATRIX WITH
!                   EIGENVALUES d0+e0, d0+e0+2*e0, d0+e0+4*e0, d0+e0+6*e0, ..., d0+e0+(2*n-4)*e0 AND
!                   lambda=d0+e0+2*e0 .
!                   SEE GLADWELL ET AL.
!
            lambda = d0 + e0 + two*e0
!
            a  = d0 + e0*real( n-1_i4b, stnd ) 
!
            d(1_i4b)   = lambda
            d(2_i4b:n) = a
!
            e(1_i4b) = zero
!
            do j = 2_i4b, n-1_i4b
                e(j) = -e0*half*sqrt( real( (j-1_i4b)*(2_i4b*n-j), stnd ) )
            end do
!
            if ( test_eigval ) then
!
                eigval(:n-1_i4b) = arth( d0+e0, two*e0, n-1_i4b )
                eigval(n)        = lambda
!
            end if
!
        case( 5_i4b )
!
!           CASE 5: GENERATE A CLEMENT TRIDIAGONAL MATRIX WITH EIGENVALUES
!                   PLUS AND MINUS THE NUMBERS n-1, n-3, n-5, ..., (1 or 0).
!
            d(1_i4b:n) = zero
!
            temp(:n-1_i4b)  = arth( one, one, n-1_i4b )
            temp2(:n-1_i4b) = temp(n-1_i4b:1_i4b:-1_i4b)
            e(1:n-1_i4b)    = sqrt( temp(:n-1_i4b)*temp2(:n-1_i4b) )
!
            if ( test_eigval ) then
!
                c = real( n, stnd ) - one
!
                eigval(:n) = arth( c, -two, n )
!
            end if
!
        case( 6_i4b )
!
!           CASE 6: EXAMPLE 7.4 OF GREGORY AND KARNEY. P. 137.
!                   EXAMPLE 2 OF GODUNOV ET AL. p 398.
!                   GENERATE A TOEPLITZ TRIDIAGONAL MATRIX WITH EIGENVALUES
!                   d0 + 2*e0*cos(j*pi/(n+1)), j=1:n, WHERE d0 AND e0
!                   ARE SCALARS, ALL THE DIAGONAL ELEMENTS ARE EQUAL TO d0
!                   AND ALL SUBDIAGONAL AND SUPERDIAGONAL ELEMENTS ARE EQUAL TO e0.
!
            d(1_i4b:n)       = d0     
            e(1_i4b:n-1_i4b) = e0
!
            if ( test_eigval ) then
!
                c = real( n, stnd) + one
!
                do j = 1_i4b, n
!
                    tmp = real( j, stnd)/c
!
                    if ( tmp==half ) then
                        eigval(j) = d0
                    else if ( tmp==(four/six) ) then
                        eigval(j) = d0 - e0
                    else if ( tmp==one ) then
                        eigval(j) = d0 - two*e0
                    else if ( tmp==zero ) then
                        eigval(j) = d0 + two*e0
                    else
                        eigval(j) = d0 + two*e0*cos( tmp*pi )
                    end if
!
                end do
!
            end if
!
        case( 7_i4b )
!
!           CASE 7: EXAMPLE 7.5 OF GREGORY AND KARNEY. P. 137.
!                   GENERATE A TRIDIAGONAL MATRIX WITH EIGENVALUES
!                   d0 + 2*e0*cos(2*j*pi/(2*n+1)), j=1:n, WHERE d0 AND e0
!                   ARE SCALARS, ALL THE DIAGONAL ELEMENTS ARE EQUAL TO d0, EXCEPTED
!                   THE FIRST, WHICH IS EQUAL TO d0-e0, AND ALL SUBDIAGONAL 
!                   AND SUPERDIAGONAL ELEMENTS ARE EQUAL TO e0.
!
            d(1_i4b)         = d0 - e0   
            d(2_i4b:n)       = d0     
            e(1_i4b:n-1_i4b) = e0
!
            if ( test_eigval ) then
!
                c = two*real( n, stnd) + one
!
                do j = 1_i4b, n
!
                    tmp = (two*real( j, stnd))/c
!
                    if ( tmp==half ) then
                        eigval(j) = d0
                    else if ( tmp==(four/six) ) then
                        eigval(j) = d0 - e0
                    else if ( tmp==one ) then
                        eigval(j) = d0 - two*e0
                    else if ( tmp==zero ) then
                        eigval(j) = d0 + two*e0
                    else
                        eigval(j) = d0 + two*e0*cos( tmp*pi )
                    end if
!
                end do
!
            end if
!
        case( 8_i4b )
!
!           CASE 8: EXAMPLE 7.6 OF GREGORY AND KARNEY. P. 138.
!                   GENERATE A TRIDIAGONAL MATRIX WITH EIGENVALUES
!                   d0 + 2*e0*cos((2*j-1)*pi/(2*n)), j=1:n, WHERE d0 AND e0
!                   ARE SCALARS, ALL THE DIAGONAL ELEMENTS ARE EQUAL TO d0, EXCEPTED
!                   THE FIRST, WHICH IS EQUAL TO d0-e0, AND THE LAST, WHICH IS EQUAL TO d0+e0,
!                   AND ALL SUBDIAGONAL AND SUPERDIAGONAL ELEMENTS ARE EQUAL TO e0.
!
            d(1_i4b)         = d0 - e0
            d(n)             = d0 + e0
            d(2_i4b:n-1_i4b) = d0
            e(1_i4b:n-1_i4b) = e0
!
            if ( test_eigval ) then
!
                c = two*real( n, stnd)
!
                do j = 1_i4b, n
!
                    tmp = (two*real( j, stnd) - one)/c
!
                    if ( tmp==half ) then
                        eigval(j) = d0
                    else if ( tmp==(four/six) ) then
                        eigval(j) = d0 - e0
                    else if ( tmp==one ) then
                        eigval(j) = d0 - two*e0
                    else if ( tmp==zero ) then
                        eigval(j) = d0 + two*e0
                    else
                        eigval(j) = d0 + two*e0*cos( tmp*pi )
                    end if
!
                end do
!
            end if
!        
        case( 9_i4b )
!
!           CASE 9: EXAMPLE 7.7 OF GREGORY AND KARNEY. P. 138.
!                   GENERATE A TRIDIAGONAL MATRIX WITH EIGENVALUES
!                   d0 + 2*e0*cos((pi/n)*(j-1)), j=1:n, WHERE d0 AND e0
!                   ARE SCALARS, ALL THE DIAGONAL ELEMENTS ARE EQUAL TO d0, EXCEPTED
!                   THE FIRST AND THE LAST, WHICH ARE EQUAL TO d0+e0, AND
!                   ALL SUBDIAGONAL AND SUPERDIAGONAL ELEMENTS ARE EQUAL TO e0.
!
            d(1_i4b)         = d0 + e0   
            d(n)             = d0 + e0  
            d(2_i4b:n-1_i4b) = d0     
            e(1_i4b:n-1_i4b) = e0
!
            if ( test_eigval ) then
!
                c = real( n, stnd)
!
                do j = 1_i4b, n
!
                    tmp = (real( j, stnd) - one)/c
!
                    if ( tmp==half ) then
                        eigval(j) = d0
                    else if ( tmp==(four/six) ) then
                        eigval(j) = d0 - e0
                    else if ( tmp==one ) then
                        eigval(j) = d0 - two*e0
                    else if ( tmp==zero ) then
                        eigval(j) = d0 + two*e0
                    else
                        eigval(j) = d0 + two*e0*cos( tmp*pi )
                    end if
!
                end do            
!
            end if
!        
        case( 10_i4b )
!
!           CASE 10: EXAMPLE 7.8 AND 7.9 OF GREGORY AND KARNEY. P. 139.
!                    IF N IS EVEN GENERATE A TRIDIAGONAL MATRIX WITH EIGENVALUES:
!                    (d0 + e0 + sqrt( (d0-e0)**2 + 16*( cos((j*pi)/(n+1)) )**2 ) )/2  FOR j=1:n/2
!                    (d0 + e0 - sqrt( (d0-e0)**2 + 16*( cos((j*pi)/(n+1)) )**2 ) )/2  FOR j=1:n/2
!                    AND IF N IS ODD GENERATE A TRIDIAGONAL MATRIX WITH EIGENVALUES:
!                    (d0 + e0 + sqrt( (d0-e0)**2 + 16*( cos((j*pi)/(n+1)) )**2 ) )/2  F0R j=1:n/2
!                    (d0 + e0 - sqrt( (d0-e0)**2 + 16*( cos((j*pi)/(n+1)) )**2 ) )/2  FOR j=1:n/2 AND d0
!                    , WHERE d0 AND e0 ARE SCALARS, ALL SUBDIAGONAL AND SUPERDIAGONAL
!                    ELEMENTS ARE EQUAL TO 1, AND DIAGONAL ELEMENTS ARE ALTERNATIVELY
!                    EQUAL TO d0 AND e0, STARTING BY d0.
!
            d(1_i4b:n:2_i4b) = d0     
            d(2_i4b:n:2_i4b) = e0     
            e(1_i4b:n-1_i4b) = one
!
            if ( test_eigval ) then
!
                a = ( d0 - e0 )**2_i4b
                b = d0 + e0
                c = real( n, stnd) + one
!
                do j = 1_i4b, n/2_i4b
!
                    tmp = real(j,stnd)/c
!
                    if ( tmp==zero  ) then
                        lambda = sqrt( a + c16 )
                    else if ( tmp==quarter  ) then
                        lambda = sqrt( a + eight )
                    else if ( tmp==half  ) then
                        lambda = abs( d0 - e0 )
                    else
                        lambda = sqrt( a + c16*( cos( tmp*pi ) )**2 )
                    end if
!
                    eigval(j)         = ( b + lambda )/two
                    eigval(n/2_i4b+j) = ( b - lambda )/two
!
                end do
!
                if ( 2_i4b*(n/2_i4b)/=n ) eigval(n) = d0
!
            end if
!
        case( 11_i4b )
!
!           CASE 11: EXAMPLE 1 OF GODUNOV ET AL. P. 391. ALL DIAGONAL ELEMENTS ARE EQUAL TO 1 AND
!                    OFF-DIAGONAL ELEMENTS e(j) ARE EQUAL TO 1/(e0**((j+1)/2_i4b)) FOR j=1, n, 2
!                    AND 0 FOR FOR j=2, n, 2 .
!
!                    IF n IS EVEN THE EIGENVALUES ARE :
!                    1 - 1/(e0**j) FOR j=1, ..., n/2 AND 1 + 1/(e0**j) FOR j=1, ..., n/2
!                    IF N IS ODD THE EIGENVALUES ARE :
!                    1 - 1/(e0**j) FOR j=1, ..., (n-1)/2 , 1 + 1/(e0**j) FOR j=1, ..., (n-1)/2 AND 1 .
!
            e0 = quarter
!
            d(1_i4b:n) = one                     
!
            e(2_i4b:n:2_i4b) = zero
!
            do j = 2_i4b, n, 2_i4b
                e(j-1_i4b) = e0**(j/2_i4b)
            end do
!
            if ( test_eigval ) then
!
                do j = 1_i4b, n/2_i4b
                    temp(j) = e0**j
                end do
!
                eigval(:n/2_i4b) = one - temp(:n/2_i4b)
!
                if ( n == 2_i4b*(n/2_i4b) ) then
                    eigval((n/2_i4b)+1_i4b:n) = one + temp(:n/2_i4b)
                else
                    eigval((n/2_i4b)+1_i4b)   = one
                    eigval((n/2_i4b)+2_i4b:n) = one + temp(:n/2_i4b)
                end if
!
            end if
!
        case( 12_i4b )
!
!           CASE 12: PARAMETRIZED FORM OF EXAMPLE 1 OF GODUNOV ET AL. P. 391.
!                    ALL DIAGONAL ELEMENTS ARE EQUAL TO d0 AND OFF-DIAGONAL ELEMENTS e(j)
!                    ARE EQUAL TO d0/(e0**((j+1)/2_i4b)) FOR j=1, n, 2 AND 0 FOR FOR j=2, n, 2 .
!
!                    IF n IS EVEN THE EIGENVALUES ARE :
!                    d0 - d0/(e0**j) FOR j=1, ..., n/2 AND d0 + d0/(e0**j) FOR j=1, ..., n/2
!                    IF N IS ODD THE EIGENVALUES ARE :
!                    d0 - d0/(e0**j) FOR j=1, ..., (n-1)/2 , d0 + d0/(e0**j) FOR j=1, ..., (n-1)/2 AND d0 .
!
            d(1_i4b:n) = d0                     
!
            e(2_i4b:n:2_i4b) = zero
!
            tmp = one/e0
!
            do j = 2_i4b, n, 2_i4b
                e(j-1_i4b) = d0*(tmp**(j/2_i4b))
            end do
!
            if ( test_eigval ) then
!
                do j = 1_i4b, n/2_i4b
                    temp(j) = d0*(tmp**j)
                end do
!
                eigval(:n/2_i4b) = d0 - temp(:n/2_i4b)
!
                if ( n == 2_i4b*(n/2_i4b) ) then
                    eigval((n/2_i4b)+1_i4b:n) = d0 + temp(:n/2_i4b)
                else
                    eigval((n/2_i4b)+1_i4b)   = d0
                    eigval((n/2_i4b)+2_i4b:n) = d0 + temp(:n/2_i4b)
                end if
!
            end if
!
        case( 13_i4b )
!
!           CASE 13: EXAMPLE 2 OF GODUNOV ET AL. P. 400. ALL THE DIAGONAL
!                    ELEMENTS ARE EQUAL TO 0 AND ALL OFF-DIAGONAL
!                    ELEMENTS ARE EQUAL TO 0.5.
!                    EIGENVALUES ARE THE ROOTS OF THE CHEBYSHEV
!                    POLYNOMIALS OF SECOND KIND AND ORDER 2*n, WITH
!                    THE LEADING COEFFICIENT EQUAL TO ONE AND ARE EQUAL TO
!                    cos(j*pi/(n+1)), j=1 TO n .
!
            d(:n)       = zero
            e(:n-1_i4b) = half
!
            if ( test_eigval ) then
!
                i = n + 1_i4b
                c = real( i, stnd )
!
                do j = 1_i4b, n
!
!                    tmp = real( j, stnd)/c
!
!                    if ( tmp==half ) then
                    if ( 2_i4b*j==i ) then
                        eigval(j) = zero
!                    else if ( tmp==(four/six) ) then
                    else if ( 6_i4b*j==4_i4b*i ) then
                        eigval(j) = -half
!                    else if ( tmp==one ) then
                    else if ( j==i ) then
                        eigval(j) = -one
                    else
                        tmp = real( j, stnd)/c
                        eigval(j) = cos( tmp*pi )
                    end if
!
                end do
!
            end if
!
        case( 14_i4b )
!
!           CASE 14: GENERATE A (1-2-1) TRIDIAGONAL MATRIX.
!                    THE EIGENVALUES ARE ALL EQUAL TO
!                    2 + 2*cos(j*pi/(n+1)), j=1 TO n .
!
            d(:n)       = two
            e(:n-1_i4b) = one
!
            if ( test_eigval ) then
!
                i = n + 1_i4b
                c = real( i, stnd )
!
                do j = 1_i4b, n
!
!                    tmp = real( j, stnd)/c
!
!                    if ( tmp==half ) then
                    if ( 2_i4b*j==i ) then
                        eigval(j) = two
!                    else if ( tmp==(four/six) ) then
                    else if ( 6_i4b*j==4_i4b*i ) then
                        eigval(j) = one
!                    else if ( tmp==one ) then
                    else if ( j==i ) then
                        eigval(j) = zero
                    else
                        tmp = real( j, stnd)/c
                        eigval(j) = two + two*cos( tmp*pi )
                    end if
!
                end do
!
            end if
!
        case( 15_i4b )
!
!           CASE 15: GENERATE A CONSTANT TRIDIAGONAL MATRIX WITH ALL
!                    ELEMENTS EQUAL TO 0.5.
!                    THE EIGENVALUES ARE ALL EQUAL TO
!                    0.5 + cos(j*pi/(n+1)), j=1 TO n .
!
            d(:n)       = half
            e(:n-1_i4b) = half
!
            if ( test_eigval ) then
!
                i = n + 1_i4b
                c = real( i, stnd )
!
                do j = 1_i4b, n
!
!                    tmp = real( j, stnd)/c
!
!                    if ( tmp==half ) then
                    if ( 2_i4b*j==i ) then
                        eigval(j) = half
!                    else if ( tmp==(four/six) ) then
                    else if ( 6_i4b*j==4_i4b*i ) then
                        eigval(j) = zero
!                    else if ( tmp==one ) then
                    else if ( j==i ) then
                        eigval(j) = -half
                    else
                        tmp = real( j, stnd)/c
                        eigval(j) = half + cos( tmp*pi )
                    end if
!
                end do
!
            end if
!
        case( 16_i4b )
!
!           CASE 16: GENERATE A ZERO TRIDIAGONAL MATRIX.
!                    THE EIGENVALUES ARE ALL EQUAL TO ZERO.
!
            d(:n) = zero
            e(:n) = zero
!
            if ( test_eigval ) then
!
                eigval(:n) = zero
!
            end if
!
        case( 17_i4b )
!
!           CASE 17: GENERATE A UNIT TRIDIAGONAL MATRIX.
!                    THE EIGENVALUES ARE ALL EQUAL TO ONE.
!
            d(:n) = one
            e(:n) = zero
!
            if ( test_eigval ) then
!
                eigval(:n) = one
!
            end if
!
        case( 18_i4b )
!
!           CASE 18: EXAMPLE 3 OF GODUNOV ET AL. P. 401.
!                    GENERATE A TRIDIAGONAL MATRIX FROM LEGENDRE RECURRENCE .
!                    EIGENVALUES ARE THE ROOTS OF THE ORTHOGONAL
!                    LEGENDRE POLYNOMIALS OF ORDER 2*n.
!        
            d(:n) = zero
!
            do j = 2_i4b, n
                e(j-1_i4b) = real( j, stnd)/sqrt( real((2_i4b*j-1_i4b)*(2_i4b*j+1_i4b)))
            end do
!
        case( 19_i4b )
!
!           CASE 19: GOLUB-KAHAN FORM OF BIDIAGONAL EXAMPLE 1 OF GODUNOV ET AL. P. 417.
!        
            d(:n) = zero
!
            e(1_i4b:n:2_i4b) = one
            e(2_i4b:n:2_i4b) = ten
!
        case( 20_i4b )
!
!           CASE 20: GOLUB-KAHAN FORM OF BIDIAGONAL EXAMPLE 2 OF GODUNOV ET AL. P. 417.
!        
            d(:n) = zero
!
            e(1_i4b:n:2_i4b) = c1_m2
            e(2_i4b:n:2_i4b) = c900
!
        case( 21_i4b )
!
!           CASE 21: GOLUB-KAHAN FORM OF BIDIAGONAL EXAMPLE 3 OF GODUNOV ET AL. P. 419.
!        
            d(:n) = zero
!
            i = n/3_i4b
            j = n - i
!
            e(1_i4b:i:2_i4b) = c1_m2
            e(2_i4b:i:2_i4b) = c900
!
            e(i+1_i4b:j:2_i4b) = c900
            e(i+2_i4b:j:2_i4b) = c1_m2
!
            e(j+1_i4b:n:2_i4b) = c1_m2
            e(j+2_i4b:n:2_i4b) = c900
!
        case( 22_i4b )
!
!           CASE 22: GOLUB-KAHAN FORM OF BIDIAGONAL EXAMPLE 1 OF NAKATSUKASA ET AL. P. 47 TABLE 1.
!        
            d(:n) = zero
!
            i  = n/2_i4b
            d0 = real( i, stnd )
!
            e(1_i4b:n-1_i4b:2_i4b) = arth( d0, -one, i )
            e(2_i4b:n-1_i4b:2_i4b) = one
!
        case( 23_i4b )
!
!           CASE 23: GOLUB-KAHAN FORM OF BIDIAGONAL EXAMPLE 4 OF NAKATSUKASA ET AL. P. 47 TABLE 1.
!        
            d(:n) = zero
!
            i  = n/2_i4b
            d0 = real( i, stnd) + one
!
            do j = 2_i4b, i, 2_i4b
                temp(j)       = real( j/2_i4b, stnd )
                temp(j-1_i4b) = d0 - temp(j)
            end do
!
            if ( i/=2_i4b*(i/2_i4b) ) then
                temp(i) = d0*half
            end if
!
            do j = 1_i4b, i
                temp2(j) = real( i-j, stnd )/five
            end do
!
            e(1_i4b:n-1_i4b:2_i4b) = temp(:i)
            e(2_i4b:n:2_i4b)       = temp2(:i)
!
        case( 24_i4b )
!
!           CASE 24: GOLUB-KAHAN FORM OF BIDIAGONAL EXAMPLE 5 OF NAKATSUKASA ET AL. P. 47 TABLE 1.
!        
            d(:n) = zero
!
            i = n/2_i4b
            j = i/2_i4b
            if ( i/=2_i4b*j ) then
                j = j + 1_i4b
            end if
!
            d0 = real( j, stnd )
            e0 = c1_01
!
            temp(j)                    = one
            temp(j+1_i4b:i)            = geop( e0, e0, i-j)
            temp(j-1_i4b:1_i4b:-1_i4b) = geop( e0, e0, j-1_i4b)
!
            e(1_i4b:n-1_i4b:2_i4b) = temp(:i)
            e(2_i4b:n-1_i4b:2_i4b) = one
!
        case( 25_i4b )
!
!           CASE 25: GOLUB-KAHAN FORM OF BIDIAGONAL EXAMPLE 6 OF FERNANDO AND PARLETT P. 218,
!        
            d(:n) = zero
!
            i1 = n/2_i4b
            i2 = i1/2_i4b
!
            do i = 1_i4b, i1
!
                temp(i) = real( abs( i-i2+1_i4b ), stnd )
!
            end do
!
            e(1_i4b:n-1_i4b:2_i4b) = temp(:i1)
            e(2_i4b:n-1_i4b:2_i4b) = one
!
        case( 26_i4b )
!
!           CASE 26: GOLUB-KAHAN FORM OF BIDIAGONAL EXAMPLE 7 OF FERNANDO AND PARLETT P. 218,
!        
            d(:n) = zero
!
            i  = n/2_i4b
            i1 = i/2_i4b
            i2 = i/4_i4b
!
            do j = 1_i4b, i1
!
                temp(j) = real( abs( j-i2+1_i4b ), stnd )
!
            end do
!
            do j = i1+1_i4b, n
!
                temp(j) = temp(j-i1)
!
            end do
!
            e(1_i4b:n-1_i4b:2_i4b) = temp(:i)
            e(2_i4b:n-1_i4b:2_i4b) = one
!
        case( 27_i4b )
!
!           CASE 27: EXAMPLE 7.2 OF GREGORY AND KARNEY. P. 135.
!                    GENERATE A WILKINSON TRIDIAGONAL MATRIX .
!
            i = n/2_i4b
            d0 = real( i, stnd )
!
            d(1_i4b:i)  = arth( d0, -one, i )
            if ( n/=2_i4b*i ) then
                d(i+1_i4b)          = zero
                d(n:i+2_i4b:-1_i4b) = d(1_i4b:i)
            else
                d(n:i+1_i4b:-1_i4b) = d(1_i4b:i)
            end if          
!
            e(1_i4b:n-1_i4b) = one
!        
        case( 28_i4b )
!
!           CASE 28: EXAMPLE 7.3 OF GREGORY AND KARNEY. P. 136.
!                    GENERATE A WILKINSON TRIDIAGONAL MATRIX .
!
            i = n/2_i4b
            d0 = real( i, stnd )
!
            d(1_i4b:i)  = arth( d0, -one, i )
            if ( n/=2_i4b*i ) then
                d(i+1_i4b)          = zero
                d(n:i+2_i4b:-1_i4b) = -d(1_i4b:i)
            else
                d(n:i+1_i4b:-1_i4b) = -d(1_i4b:i)
            end if          
!
            e(1_i4b:n-1_i4b) = one
!        
        case( 29_i4b )
!
!           CASE 29: GENERATE A WILKINSON DEFINITE-POSITIVE TRIDIAGONAL MATRIX.
!
            i = n/2_i4b
            d0 = real( i, stnd )
!
            d(1_i4b:i)  = arth( d0, -one, i )
!
            if ( n/=2_i4b*i ) then
                d(i+1_i4b)          = zero
                d(n:i+2_i4b:-1_i4b) = d(1_i4b:i)
            else
                d(n:i+1_i4b:-1_i4b) = d(1_i4b:i)
            end if
!       
            d(:n) = d(:n) + two
!
            e(1_i4b:n-1_i4b) = one
!        
        case( 30_i4b )
!
!           CASE 30: GENERATE A WILKINSON DEFINITE-POSITIVE TRIDIAGONAL MATRIX.
!
            i = n/2_i4b
            d0 = real( i, stnd )
!
            d(1_i4b:i)  = arth( d0, -one, i )
!
            if ( n/=2_i4b*i ) then
                d(i+1_i4b)          = zero
                d(n:i+2_i4b:-1_i4b) = -d(1_i4b:i)
            else
                d(n:i+1_i4b:-1_i4b) = -d(1_i4b:i)
            end if
!        
            d(:n) = d(:n) + real( i+1_i4b, stnd )
!
            e(1_i4b:n-1_i4b) = one
!
        case( 31_i4b )
!
!           CASE 31: FROM BIDIAGONAL EXAMPLE 6 OF FERNANDO AND PARLETT P. 218,
!                    GENERATE A WILKINSON-TYPE TRIDIAGONAL MATRIX .
!
            i = n/2_i4b
!
            do j = 1_i4b, n
!
                d(j) = real( abs( j-i+1_i4b ), stnd )
!
            end do
!
            e(1_i4b:n-1_i4b) = one
!
        case( 32_i4b )
!
!           CASE 32: FROM BIDIAGONAL EXAMPLE 7 OF FERNANDO AND PARLETT P. 218,
!                    GENERATE A WILKINSON-TYPE TRIDIAGONAL MATRIX .
!
            i = n/2_i4b
            l = n/4_i4b
!
            do j = 1_i4b, i
!
                d(j) = real( abs( j-l+1_i4b ), stnd )
!
            end do
!
            do j = i+1_i4b, n
!
                d(j) = d(j-i)
!
            end do
!
            e(1_i4b:n-1_i4b) = one
!
        case( 33_i4b )
!
!           CASE 33: GENERATE A TRIDIAGONAL MATRIX FROM LAGUERRE RECURRENCE.
!
            do j = 1_i4b, n
                lambda  =  real( j, stnd)
                e(j) = lambda + one
                lambda  = two*lambda + one
                d(j) = lambda
            end do
!
        case( 34_i4b )
!
!           TEST 34: GENERATE A TRIDIAGONAL MATRIX FROM HERMITE RECURRENCE .
!
            d(:n) = zero
 !
            do j = 1_i4b, n-1_i4b
                lambda =  sqrt( real( j, stnd) )
                e(j) = lambda
            end do
!
        case( 35_i4b )
!
!           CASE 35: GENERATE A GRADED TRIDIAGONAL MATRIX WITH A GEOMETRIC PROGRESSION.
!                    (PARAMETRIZED FORM).
!
            d(:n)       = geop( d0, e0, n )
            e(:n-1_i4b) = d(:n-1_i4b)
!
        case( 36_i4b )
!
!           CASE 36: GENERATE A REVERSELY GRADED TRIDIAGONAL MATRIX WITH A GEOMETRIC PROGRESSION.
!                    (PARAMETRIZED FORM).
!
            d(n:1_i4b:-1_i4b) = geop( d0, e0, n )
            e(:n-1_i4b)       = d(:n-1_i4b)
!
        case( 37_i4b )
!
!           CASE 37: GENERATE A GRADED TRIDIAGONAL MATRIX WITH AN ARITHMETIC PROGRESSION.
!                    (PARAMETRIZED FORM).
!
            d(:n)       = arth( d0, e0, n )
            e(:n-1_i4b) = d(:n-1_i4b)
!
        case( 38_i4b )
!
!           CASE 38: GENERATE A REVERSELY GRADED TRIDIAGONAL MATRIX WITH AN ARITHMETIC PROGRESSION.
!                    (PARAMETRIZED FORM).
!
            d(n:1_i4b:-1_i4b) = arth( d0, e0, n )
            e(:n-1_i4b)       = d(:n-1_i4b)
            e(n)              = zero
!
        case( 39_i4b )
!
!           CASE 39: ADAPTED FROM BIDIAGONAL EXAMPLE 1 OF NAKATSUKASA ET AL. P. 47 TABLE 1.
!
            d0    = real( n, stnd )
            d(:n) = arth( d0, -one, n )
!       
            e(:n-1_i4b) = one
!
        case( 40_i4b )
!
!           CASE 40: ADAPTED FROM BIDIAGONAL EXAMPLE 4 OF NAKATSUKASA ET AL. P. 47 TABLE 1.
!
            d0 = real( n, stnd) + one
!
            do j = 2_i4b, n, 2_i4b
                d(j)       = real( j/2_i4b, stnd )
                d(j-1_i4b) = d0 - d(j)
            end do
!
            if ( n/=2_i4b*(n/2_i4b) ) then
                d(n) = d0*half
            end if
!
            do j = 1_i4b, n-1_i4b
                e(j) = real( n-j, stnd )/five
            end do
!
        case( 41_i4b )
!
!           CASE 41: ADAPTED FROM BIDIAGONAL EXAMPLE 5 OF NAKATSUKASA ET AL. P. 47 TABLE 1.
!
            i = n/2_i4b
            d0 = real( i, stnd )
            e0 = c1_01
!
            d(i)                    = one
            d(i+1_i4b:n)            = geop( e0, e0, n-i)
            d(i-1_i4b:1_i4b:-1_i4b) = geop( e0, e0, i-1_i4b)
!       
            e(:n-1_i4b) = one
!
        case( 42_i4b )
!
!           CASE 42: GENERATE A TRIDIAGONAL MATRIX WITH ONE EIGENVALUE NEAR ONE AND
!                    TWO CLUSTERS OF SMALL POSITIVE AND NEGATIVE EIGENVALUES.
!
            d0 = c200*ulp
!
            d(:n) = zero
!
            e(1_i4b)         = one
            e(2_i4b:n-1_i4b) = d0
!
        case( 43_i4b )
!
!           CASE 43: GENERATE A PSEUDO HILBERT TRIDIAGONAL MATRIX.
!                    THE HILBERT MATRIX HAS (i,j) ELEMENT EQUAL TO
!                    1/(i+j-1). 
!
            do i = 1_i4b, n
                d(i) = one/( real( 2_i4b*i - 1_i4b, stnd ) )
                e(i) = one/( real( 2_i4b*i, stnd ) )
            end do
!
        case( 44_i4b )
!
!           CASE 44: GENERATE A GLUED KIMURA TRIDIAGONAL MATRIX WITH RADIUS EQUAL TO 11
!                    AND GLUED PARAMETER EQUAL TO 10.**(-4).
!
            d0  = real( 11_i4b, stnd )
            glu = 10.**(-4_i4b)
!
            temp(1_i4b:11_i4b)         = arth( d0, -one, 11_i4b )
            temp(21_i4b:12_i4b:-1_i4b) = temp(1_i4b:10_i4b)
!
            temp2(1_i4b:41_i4b:2_i4b) = temp(1_i4b:21_i4b)
            temp2(2_i4b:41_i4b:2_i4b) = one
            temp2(42_i4b)             = glu
!
            do i=0_i4b, n-1_i4b, 42_i4b
!
               i1 = i + 1_i4b
               i2 = min( i + 42_i4b, n)
               j  = i2 - i1 + 1_i4b
!
               e(i1:i2) = temp2(1_i4b:j)
!
            end do
!           
            d(:n) = zero
!
        case( 45_i4b )
!
!           CASE 45: GENERATE A GLUED KIMURA TRIDIAGONAL MATRIX (PARAMETRIZED FORM).
!
            d0 = real( l, stnd )
!
            temp(1_i4b:l)                      = arth( d0, -one, l )
            temp(2_i4b*l-1_i4b:l+1_i4b:-1_i4b) = temp(1_i4b:l-1_i4b)
!
            temp2(1_i4b:4_i4b*l-3_i4b:2_i4b) = temp(1_i4b:2_i4b*l-1_i4b)
            temp2(2_i4b:4_i4b*l-3_i4b:2_i4b) = one
            temp2(4_i4b*l-2_i4b)             = glu
!
            do i=0_i4b, n-1_i4b, 4_i4b*l-2_i4b
!
               i1 = i + 1_i4b
               i2 = min( i +  4_i4b*l-2_i4b, n)
               j  = i2 - i1 + 1_i4b
!
               e(i1:i2) = temp2(1_i4b:j)
!
            end do
!           
            d(:n) = zero
!        
        case( 46_i4b )
!
!           CASE 46: GENERATE A GLUED WILKINSON TRIDIAGONAL MATRIX FROM
!                    EXAMPLE 7.2 OF GREGORY AND KARNEY. p 135 (PARAMETRIZED FORM).
!                    SEE PARLETT ET AL. P. 6 .
!         
            d0 = real( l, stnd )
!
            temp(1_i4b:l)                      = arth( d0, -one, l )
            temp(l+1_i4b)                      = zero
            temp(2_i4b*l+1_i4b:l+2_i4b:-1_i4b) = temp(1_i4b:l)
!
            temp2(1_i4b:2_i4b*l) = one
            temp2(2_i4b*l+1_i4b) = glu    
!
            do i=0_i4b, n-1_i4b, 2_i4b*l+1_i4b
!
               i1 = i + 1_i4b
               i2 = min( i + 2_i4b*l+1_i4b, n)
               j  = i2 - i1 + 1_i4b
!
               d(i1:i2) = temp(1_i4b:j)
               e(i1:i2) = temp2(1_i4b:j)
!
               if ( i/=0_i4b ) then
                   d(i)  = d(i)  + glu
                   d(i1) = d(i1) + glu
               end if
!
            end do
!        
        case( 47_i4b )
!
!           CASE 47: GENERATE A GLUED WILKINSON TRIDIAGONAL MATRIX FROM
!                    EXAMPLE 7.3 OF GREGORY AND KARNEY. P. 136 (PARAMETRIZED FORM).
!                    SEE PARLETT ET AL. P. 6 .
!         
            d0 = real( l, stnd )
!
            temp(1_i4b:l)                      = arth( d0, -one, l )
            temp(l+1_i4b)                      = zero
            temp(2_i4b*l+1_i4b:l+2_i4b:-1_i4b) = -temp(1_i4b:l)
!
            temp2(1_i4b:2_i4b*l) = one
            temp2(2_i4b*l+1_i4b) = glu    
!
            do i=0_i4b, n-1_i4b, 2_i4b*l+1_i4b
!
               i1 = i + 1_i4b
               i2 = min( i + 2_i4b*l+1_i4b, n)
               j  = i2 - i1 + 1_i4b
!
               d(i1:i2) = temp(1_i4b:j)
               e(i1:i2) = temp2(1_i4b:j)
!
               if ( i/=0_i4b ) then
                   d(i)  = d(i)  + glu
                   d(i1) = d(i1) + glu
               end if
!
            end do
!        
        case( 48_i4b )
!
!           CASE 48: GENERATE A GLUED WILKINSON DEFINITE-POSITIVE TRIDIAGONAL MATRIX FROM
!                    EXAMPLE 7.2 OF GREGORY AND KARNEY. P. 135 (PARAMETRIZED FORM).
!                    SEE PARLETT ET AL. P. 6 .
!         
            d0 = real( l, stnd )
!
            temp(1_i4b:l)                      = arth( d0, -one, l )
            temp(l+1_i4b)                      = zero
            temp(2_i4b*l+1_i4b:l+2_i4b:-1_i4b) = temp(1_i4b:l)
!       
            temp(1_i4b:2_i4b*l+1_i4b) = temp(1_i4b:2_i4b*l+1_i4b) + two
!
            temp2(1_i4b:2_i4b*l) = one
            temp2(2_i4b*l+1_i4b) = glu     
!
            do i=0_i4b, n-1_i4b, 2_i4b*l+1_i4b
!
               i1 = i + 1_i4b
               i2 = min( i + 2_i4b*l+1_i4b, n)
               j  = i2 - i1 + 1_i4b
!
               d(i1:i2) = temp(1_i4b:j)
               e(i1:i2) = temp2(1_i4b:j)
!
               if ( i/=0_i4b ) then
                   d(i)  = d(i)  + glu
                   d(i1) = d(i1) + glu
               end if
!
            end do
!        
        case( 49_i4b )
!
!           CASE 49: GENERATE A GLUED DEFINITE-POSITIVE WILKINSON TRIDIAGONAL MATRIX FROM
!                    EXAMPLE 7.3 OF GREGORY AND KARNEY. P. 136 (PARAMETRIZED FORM).
!                    SEE PARLETT ET AL. P. 6 .
!         
            d0 = real( l, stnd )
!
            temp(1_i4b:l)                      = arth( d0, -one, l )
            temp(l+1_i4b)                      = zero
            temp(2_i4b*l+1_i4b:l+2_i4b:-1_i4b) = -temp(1_i4b:l)
!       
            temp(1_i4b:2_i4b*l+1_i4b) = temp(1_i4b:2_i4b*l+1_i4b) + real( l+1_i4b, stnd )
!
            temp2(1_i4b:2_i4b*l) = one
            temp2(2_i4b*l+1_i4b) = glu
!
            do i=0_i4b, n-1_i4b, 2_i4b*l+1_i4b
!
               i1 = i + 1_i4b
               i2 = min( i + 2_i4b*l+1_i4b, n)
               j  = i2 - i1 + 1_i4b
!
               d(i1:i2) = temp(1_i4b:j)
               e(i1:i2) = temp2(1_i4b:j)
!
               if ( i/=0_i4b ) then
                   d(i)  = d(i)  + glu
                   d(i1) = d(i1) + glu
               end if
!
            end do
!
        case default
!
!           DEFAULT CASE: GENERATE A RANDOM SYMMETRIC TRIDIAGONAL MATRIX.
!
            call random_number_( d(:n) )
            call random_number_( e(:n-1_i4b) )
!
    end select
!
    if ( test_eigval ) then
!
        if ( 1_i4b<=type .and. type<=17_i4b ) then
!
            if ( present(sort) ) then
                call eigval_sort( sort, eigval(:n) )
            end if
!
        else
!
!           COMPUTE THE EIGENVALUES OF THE TRIDIAGONAL MATRIX BY A BISECTION METHOD WITH HIGH ACCURACY.
!           THE EIGENVALUES ARE COMPUTED WITH MAXIMUM ACCURACY WHEN OPTIONAL PARAMETER abstol IS SET TO
!           sqrt( lamch('s') ), WHICH IS THE SQUARE ROOT OF THE UNDERFLOW THRESHOLD.
!
            call symtrid_bisect( d(:n), e(:n), i, temp(:n), fail, sort=sort, abstol=abstol )
!
            eigval(:n) = temp(:n)
!
            if ( present(failure) ) then
!
                failure = fail
!
            end if
!
        end if
!
    end if
!
    e(n) = zero
!
!   DEALLOCATE WORK ARRAYS.
!
    deallocate( temp, temp2 )
!
!
! END OF SUBROUTINE gen_symtrid_mat
! _________________________________
!
    end subroutine gen_symtrid_mat
!
!
! =========================================================================================
!                                 PRIVATE SUBROUTINES
! =========================================================================================
!
!
    subroutine symtrid_inviter( d, e, eigval, eigvec, failure,  &
                                maxiter, scaling, initvec       )
!
! Purpose
! _______
!
!   SYMTRID_INVITER computes an eigenvector of a real n-by-n symmetric tridiagonal
!   matrix T corresponding to a specified eigenvalue, by combining Fernando's
!   method for computing an eigenvector of a real n-by-n symmetric tridiagonal
!   matrix and an inverse iteration algorithm.
!
!
! Arguments
! _________
!
!   D              (INPUT) real(stnd), dimension(:)
!                  On entry, the diagonal elements of the symmetric tridiagonal matrix T.
!
!   E              (INPUT) real(stnd), dimension(:)
!                  On entry, the n-1 subdiagonal elements of the symmetric tridiagonal
!                  matrix T and E(n) is arbitrary .
!
!                  The size of E must verify:  size( E ) = size( D ) = n .
!
!   EIGVAL         (INPUT) real(stnd)
!                  On entry, an eigenvalue of the symmetric tridiagonal matrix.
!
!   EIGVEC         (OUTPUT) real(stnd), dimension(:)
!                  On exit, the computed eigenvector.
!
!                  The shape of EIGVEC must verify: size( EIGVEC ) = size( D ) = n .
!
!   FAILURE        (OUTPUT) logical(lgl)
!                  On exit:
!
!                  - FAILURE = false indicates successful exit.
!                  - FAILURE = true indicates that the eigenvector 
!                    failed to converge in MAXITER iterations.
!
!   MAXITER        (INPUT, OPTIONAL) integer(i4b)
!                  The number of inverse iterations performed in the subroutine.
!
!                  By default, 2 inverse iterations are performed.
!
!   SCALING        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if SCALING=true the tridiagonal matrix T is scaled before computing
!                  the eigenvector.
!
!                  The default is to scale the tridiagonal matrix.
!
!   INITVEC        (INPUT, OPTIONAL) logical(lgl)
!                  On entry, if INITVEC=true a Fernando vector is used to start the inverse
!                  iteration process; if INITVEC=false a random uniform starting vector is used.
!
!                  For unreduced tridiagonal matrices, the default is to use a Fernando starting vector.
!                  For reduced tridiagonal matrices, the default is to use a random uniform starting vector.
!
!
! Further Details
! _______________
!
!   SYMTRID_INVITER uses Fernando's method for computing a first estimate of an
!   eigenvector corresponding to an approximate eigenvalue of a real n-by-n
!   symmetric tridiagonal matrix T (by default, only if the input tridiagonal
!   matrix T is unreduced).
!
!   This approximate eigenvector is then refined (or computed if Fernando's method is
!   not used) using an inverse iteration algorithm.
!
!   For further details, on Fernando's method for computing eigenvectors of tridiagonal
!   matrices or inverse iteration, see:
!
!   (1) Fernando, K.V., 1997:
!           On computing an eigenvector of a tridiagonal matrix. Part I: Basic results.
!           Siam J. Matrix Anal. Appl., Vol. 18, 1013-1034.
!
!   (2) Parlett, B.N., and Dhillon, I.S., 1997:
!           Fernando's solution to Wilkinson's problem: An application of double factorization.
!           Linear Algebra and its Appl., 267, pp.247-279.
!
!   (3) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (4) Bini, D.A., Gemignani, L., and Tisseur, F., 2005:
!           The Ehrlich-Aberth method for the nonsymmetric tridiagonal eigenvalue problem.
!           SIAM J. Matrix Anal. Appl., 27, 153-175.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,    only : lamch, machbase, maxexp
    use Utilities,        only : assert_eq, norm
    use Random,           only : random_number_
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in)                :: eigval
    real(stnd),  intent(in),  dimension(:) :: d, e
    real(stnd),  intent(out), dimension(:) :: eigvec
!
    integer(i4b), intent(in),  optional :: maxiter
!
    logical(lgl), intent(out)           :: failure
    logical(lgl), intent(in),  optional :: scaling, initvec
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd),  dimension(size(d)) :: sub, diag, sup1, sup2, temp, e2, d2, cs, sn
    real(stnd)                      :: lambda, xj, xjm, safmin, prec, gamma,  &
                                       omega, normt, lim, scalfac, tau, tmp
!
    integer(i4b) :: n, mm, maxit, i, i1, i2, j
!
    logical(lgl)                      :: do_scaling, fernando_init
    logical(lgl),  dimension(size(d)) :: perm
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='symtrid_inviter'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(d),i4b)       ,   &
                    int(size(e),i4b)       ,   &
                    int(size(eigvec),i4b)  ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            eigvec(1_i4b) = one
!
        end if
!
        return
!
    end if
!
!   SET OPTIONAL ARGUMENTS.
!
    if ( present(maxiter) ) then
        maxit = max( maxiter, 1_i4b )
    else
        maxit = 2_i4b
    end if
!
    if ( present(scaling) ) then
        do_scaling = scaling
    else
        do_scaling = true
    end if
!    
!   GET MACHINES CONSTANTS.
!
    safmin = lamch( 's' )
    gamma  = sqrt( safmin )
    prec   = fudge_prec*lamch( 'p' )
!
!   MAKE A LOCAL COPY OF THE TRIDIAGONAL MATRIX.
!    
    d2(1_i4b:n)       = d(1_i4b:n)
    e2(1_i4b:n-1_i4b) = e(1_i4b:n-1_i4b)
    e2(n)             = zero
!
!   CHECK IF THE TRIDIAGONAL MATRIX IS THE ZERO MATRIX.
!
    do j = n, 2_i4b, -1_i4b
!
        if ( abs(d2(j))>=safmin .or. abs(e2(j-1_i4b))>=safmin ) exit
!
        if ( j/=n ) cycle
!
        eigvec(:n) = zero
        eigvec(j)  = one
!
    end do
!
    if ( j==1_i4b .and. abs(d2(1_i4b))<safmin ) then
!
!       THE BIDIAGONAL MATRIX IS THE NULL MATRIX, RETURN.
!
        return
!
    end if
!    
!   SCALE THE SYMMETRIC TRIDIAGONAL SYSTEM IF REQUIRED.    
!    
    if ( do_scaling ) then
!    
        normt   = maxval( abs(e2(1_i4b:n-1_i4b) ) )
        omega   = machbase**(maxexp-1_i4b)
        tau     = sqrt( gamma*omega )
!        j       = ( exponent( tau ) - exponent( normt ) - 1_i4b )/2_i4b
        j       = exponent( tau ) - exponent( normt ) - 1_i4b
        scalfac = machbase**j
!    
        tmp = gamma/scalfac
!
        where( abs(e2(1_i4b:n-1_i4b))>=tmp )
            e2(1_i4b:n-1_i4b) = scalfac*e2(1_i4b:n-1_i4b)
        elsewhere
            e2(1_i4b:n-1_i4b) = sign( gamma, e2(1_i4b:n-1_i4b) )
        end where
!
        d2(1_i4b:n) = scalfac*d2(1_i4b:n)
!    
        lambda = scalfac*eigval
!    
    else
!    
        scalfac = one
        lambda  = eigval
!    
    end if
!
!   COMPUTE THE NORM OF THE TRIDIAGONAL MATRIX.
!
    normt = zero
    xj    = zero
!
    do j=1_i4b, n
!
        xjm   = abs( e2(j) )
        normt = normt +  abs( d2(j) ) + xj + xjm
        xj    = xjm
!
    end do
!
!   DETERMINE THRESHOLD FOR TINY EIGENVALUES.
!
    lim = normt*real( n, stnd )*prec
!
!   TEST IF THE UPPER PART OF THE TRIDIAGONAL MATRIX IS NOT DIAGONAL OR ZERO.
!
    do i1=1_i4b, n-1_i4b
!
!        if ( abs( e2(i1) )>gamma .or. abs( d2(i1) )>gamma ) exit
!
        if ( abs( e2(i1) )>gamma ) exit
        if ( abs( d2(i1) )>gamma .and. abs( d2(i1) - lambda )<=lim  ) exit
!
    end do
!
!   TEST IF THE LOWER PART OF THE TRIDIAGONAL MATRIX IS NOT DIAGONAL OR ZERO.
!
    do i2=n, 2_i4b, -1_i4b
!
!        if ( abs( e2(i2-1_i4b) )>gamma .or. abs( d2(i2) )>gamma ) exit
!
        if ( abs( e2(i2-1_i4b) )>gamma ) exit
        if ( abs( d2(i2) )>gamma .and. abs( d2(i2) - lambda )<=lim ) exit
!
    end do
!
    if ( ( abs( lambda )<=gamma+lim .and. ( i1/=1_i4b .or. i2/=n ) ) &
          .or. ( i2<i1 ) ) then
!
!       IF THE EIGENVALUE IS VERY SMALL, WE CANNOT WORK ON THE REDUCED PORTION OF 
!       THE TRIDIAGONAL MATRIX.
!
        i1 = 1_i4b
        i2 = n
!
    end if
!
    if ( present( initvec ) ) then
!
        fernando_init = initvec
!
    else
!
!       TEST IF THE INPUT TRIDIAGONAL MATRIX IS UNREDUCED.
!
        fernando_init = all( abs( e2(i1:i2-1_i4b) )>gamma )
!
    end if
!
!   COMPUTE EIGENVECTOR BY INVERSE ITERATION.
!    
    if ( fernando_init ) then
!    
!       FERNANDO-BINI METHOD.  
!
        call trid_qr_cmp( d2(i1:i2), e2(i1:i2-1_i4b), lambda, cs(i1:i2-1_i4b), sn(i1:i2-1_i4b),    &
                          diag(i1:i2), sup1(i1:i2), sup2(i1:i2), mm )
!    
        mm = mm + i1 - 1_i4b
!    
        temp(i1:i2) = zero
!
        temp(mm) = one
!
!       DO maxit ITERATIONS FOR THE EIGENVECTOR OF THE TRIDIAGONAL MATRIX.
!
        do i=1_i4b, maxit
!
!           SOLVE THE SYSTEM QRx = b WITH RESCALING TO AVOID OVERFLOW.
!
            call trid_qr_solve( cs(i1:i2-1_i4b), sn(i1:i2-1_i4b), diag(i1:i2), sup1(i1:i2),  &
                                sup2(i1:i2), temp(i1:i2)  )
!
!           NORMALIZE THE EIGENVECTOR.
!
            tau = norm( temp(i1:i2) )
!
            if ( tau/=zero ) then
!
                temp(i1:i2) = temp(i1:i2)/tau
!
            else if ( i/=maxit ) then
!
                call random_number_( temp(i1:i2) )
!
                temp(i1:i2) = temp(i1:i2)*two - one
!
            end if
!
        end do
!
    else
!
!       STANDART INVERSE ITERATION METHOD.
!
!       COMPUTE LU FACTORS WITH PARTIAL PIVOTING (PT=LU) OF THE TRIDIAGONAL MATRIX.
!
        call trid_cmp_rv( d2(i1:i2), e2(i1:i2), lambda, sub(i1:i2), diag(i1:i2),  &
                          sup1(i1:i2), sup2(i1:i2), perm(i1:i2) )
!    
!       COMPUTE A RANDOM INITIAL VECTOR.  
!
        call random_number_( temp(i1:i2) )
!
        temp(i1:i2) = temp(i1:i2)*two - one
!
!       REPLACE SMALL DIAGONAL ELEMENTS BY safmin.
!
        where( abs(diag(i1:i2))<safmin ) diag(i1:i2) = sign( safmin, diag(i1:i2) )
!
!       DO maxit ITERATIONS FOR THE EIGENVECTOR OF THE TRIDIAGONAL MATRIX.
!
        do i=1_i4b, maxit
!
!           SOLVE THE SYSTEM LU = Pb WITH RESCALING TO AVOID OVERFLOW.
!
            call trid_solve_rv( sub(i1:i2), diag(i1:i2), sup1(i1:i2),   &
                                sup2(i1:i2), perm(i1:i2), temp(i1:i2) )
!
!           NORMALIZE THE EIGENVECTOR.
!
            tau = norm( temp(i1:i2) )
!
            if ( tau/=zero ) then
!
                temp(i1:i2) = temp(i1:i2)/tau
!
            else if ( i/=maxit ) then
!
                call random_number_( temp(i1:i2) )
!
                temp(i1:i2) = temp(i1:i2)*two - one
!
            end if
!
        end do
!
    end if
!
!   SAVE THE EIGENVECTOR.
!
    temp(1_i4b:i1-1_i4b) = zero
    temp(i2+1_i4b:n)     = zero
!
    eigvec(1_i4b:n) = temp(1_i4b:n)
!
    failure = tau==zero
!
!
! END OF SUBROUTINE symtrid_inviter
! _________________________________
!
    end subroutine symtrid_inviter
!
! =========================================================================================
!
    subroutine ortho_vec( q, vec, normvec, normvec_orig )
!
! Purpose
! _______
!
!   ORTHO_VEC is a private subroutine used by Gram-Schmidt algorithms for updating
!   orthogonal vectors when reorthogonalization is required. ORTHO_VEC implements 
!   a reorthogonalization procedure for the modified Gram-Schmidt algorithm used in
!   some inverse iteration or tridiagonal reduction algorithms in this module.
!
!
! Arguments
! _________
!
!   Q            (INPUT) real(stnd), dimension(:,:)
!                On entry, a n-by-nq matrix Q of previous eigenvectors or normalized
!                orthogonal vectors.
!
!   VEC          (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the current estimate of the nq+1 unnormalized eigenvector
!                or orthogonal vector.
!                On exit, the reorthogonalized and normalized nq+1 eigenvector or
!                orthogonal vector.
!
!                The size of VEC must be equal to size(Q,1).
!
!   NORMVEC      (INPUT/OUTPUT) real(stnd)
!                On entry, the current 2-norm of the vector VEC.
!                On exit, the 2-norm of the nq+1 eigenvector or orthogonal vector,
!                before normalization.
!
!   NORMVEC_ORIG (INPUT, OPTIONAL) real(stnd)
!                On entry, the original 2-norm of the vector VEC.
!                IF NORMVEC_ORIG is not specified, the default value is
!                assumed to be 1.
!
!
! Further Details
! _______________
!
!   ORTHO_VEC is an implementation of an orthogonalization procedure using
!   the Gram-Schmidt method described in the reference (1).
!
!   If size( q, 2) <= 0, ORTHO_VEC returns the normalized vector VEC if
!   NORMVEC is not zero and a nonzero normalized random vector VEC otherwise,
!   and NORMVEC is not modified.
!
!   For more details, see:
!
!   (1) Stewart, G.W., 2007: Block Gram-Schmidt Orthogonalization. Report TR-4823,
!           Department of Computer Science, College Park, University of Maryland.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch
    use Utilities,       only : assert_eq, norm
    use Random,          only : random_number_
#ifdef _BLAS
    use BLAS_interfaces, only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),    dimension(:,:) :: q
    real(stnd), intent(in),    optional       :: normvec_orig
    real(stnd), intent(inout), dimension(:)   :: vec
    real(stnd), intent(inout)                 :: normvec
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(vec)) :: vec1
    real(stnd)                       :: tmp, newnorm, oldnorm, normvec0, eps
!
    integer(i4b) :: i, j, n, nq
!   
!   
! PARAMETERS 
! __________
!
    integer(i4b), parameter :: maxi=10
!
    real(stnd), parameter :: fudge=one
!
    character(len=*),  parameter :: name_proc='ortho_vec'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   CHECK THE ARGUMENTS.
!
    n = assert_eq(    int( size(vec), i4b ),   &
                      int( size(q,1), i4b ),   &
                      name_proc )
!
    nq = size( q, 2 ) 
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!
    if ( nq<=0_i4b ) then
!
!       SPECIAL CASE WHEN size( q, 2) = 0.
!
        if ( normvec==zero ) then
!
!           IF THE NORM OF THE CURRENT ESTIMATE OF THE
!           VECTOR vec IS ZERO, SET IT TO A NONZERO
!           NORMALIZED VECTOR.
!
            call random_number_( vec1(:n) ) 
!
            vec1(:n) = vec1(:n) - half
            tmp      = norm( vec1(:n) )
            vec(:n)  = vec1(:n)/tmp
!
        else
!
!           NORMALIZED THE CURRENT ESTIMATE OF THE
!           ORTHOGONAL VECTOR OTHERWISE.
!
            vec(:n) = vec(:n)/normvec
!
        end if
!
        return
!
    end if
!
    if ( present(normvec_orig) ) then
!
!       GET TNE NORM OF THE ORIGINAL VECTOR.
!
        normvec0 = normvec_orig
!
    else
!
!       BY DEFAULT, WE ASSUME THAT THE ORIGINAL
!       VECTOR IS NORMALIZED.
!
        normvec0 = one
!
    end if
!    
!   GET MACHINE CONSTANT.
!
    eps = lamch( 'u' )
!
!    if ( normvec<=fudge*eps*normvec0 ) then
    if ( normvec==zero ) then
!
!       IF THE NORM OF THE CURRENT ESTIMATE OF THE
!       ORTHOGONAL VECTOR IS ZERO, SET IT TO A NONZERO
!       NORMALIZED RANDOM VECTOR AND PROCEED.
!
        call random_number_( vec1(:n) ) 
!
        vec1(:n) = vec1(:n) - half
        tmp      = norm( vec1(:n) )
        vec1(:n) = vec1(:n)/tmp
!
        normvec0 = one
!
    else
!
!       MAKE A LOCAL COPY OF THE CURRENT ESTIMATE OF THE
!       ORTHOGONAL VECTOR AND NORMALIZE IT OTHERWISE.
!
        vec1(:n) = vec(:n)/normvec
!
        normvec0 = normvec0/normvec
!
    end if
!
!   ORTHOGONALIZATION LOOP.
!
!    oldnorm = one
    oldnorm = normvec0
!
    do i=1_i4b, maxi
!
        do j=1_i4b, nq
!
            tmp = -dot_product( vec1(:n), q(:n,j) )
#ifdef _BLAS
            call axpy( n, tmp, q(:n,j), 1_i4b, vec1(:n), 1_i4b )
#else
            vec1(:n) = vec1(:n) + tmp*q(:n,j)
#endif
!
        end do
!
        newnorm = norm( vec1(:n) )
!
!       RETURN IF vec1 IS ORTHOGONAL TO q.
!
        if ( newnorm>fudge_ortho2*oldnorm ) exit
!
        if ( newnorm>fudge*eps*normvec0 ) then
!
!           CONTINUE ORTHOGONALIZING IF newnorm IS NOT TOO SMALL.
!
            oldnorm = newnorm
!
        else
!
!           REPLACE vec1 BY A SCALED RANDOM VECTOR OTHERWISE
!           AND PROCEED.
!
            normvec0 = eps*normvec0
            oldnorm  = normvec0
!
            call random_number_( vec1(:n) ) 
!
            vec1(:n) = vec1(:n) - half
            tmp      = norm( vec1(:n) )
            vec1(:n) = normvec0*( vec1(:n)/tmp )
!
        end if
!
    end do
!
!   RETURN THE CURRENT ESTIMATE OF THE EIGENVECTOR OR
!   ORTHOGONAL VECTOR AND ITS NORM BEFORE NORMALIZATION.
!
    if ( i<=maxi ) then
!
        tmp = newnorm
!
    else
!
        tmp = oldnorm
!
    end if
!
    if ( tmp/=zero ) then
!
        vec(:n) = vec1(:n)/tmp
!
    else
!
        vec(:n) = zero
!
    end if
!
    normvec = normvec*tmp
!
!
! END OF SUBROUTINE ortho_vec
! ___________________________
!
    end subroutine ortho_vec
!
! =========================================================================================
!
    function reverse( diag )
!
! Purpose
! _______
!
!   Private function REVERSE determines if the elements of the vector DIAG are increasing
!   in magnitude by computing the covariance coefficient of abs( DIAG(:) ) with the
!   vector (1:n) where n is the size of DIAG.
!
!   The function returns true if the covariance coefficient is positive
!   and false otherwise.
!
!
! Arguments
! _________
!
!   DIAG    (INPUT) real(stnd), dimension(:)
!           On entry, the vector DIAG . 
!
!
! Further Details
! _______________
!
!   This function is used to determine if the elements of a symmetric tridiagonal matrix
!   must be stored in reversed order before computing eigenvalues of this tridiagonal matrix.
!
!   This is important for maintaining the accuracy of the QR steps for graded matrices.
!
!   For further details, see:
!
!   (1) Greenbaum, A., and J. Dongarra, J., 1989:
!           Experiments with QR/QL Methods for the Symmetric Tridiagonal Eigenproblem.
!           LAPACK Working Note No 17, November 1989.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : arth
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in) :: diag(:)
!
    logical(lgl)  :: reverse
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(diag))  :: xt, yt
    real(stnd)                         :: sxy
!
    integer(i4b) :: n
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n = size( diag )
!
    reverse = false
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) return
!
!   COMPUTE COVARIANCE COEFFICIENT.
!
    xt(:n) = arth( one, one, n ) - real( n+1_i4b, stnd)*half
    yt(:n) = abs( diag(:n) )
    yt(:n) = yt(:n) - sum( yt(:n) )/real( n, stnd)
    sxy    = dot_product( xt, yt )
!
    reverse = sxy>zero
!
!
! END OF FUNCTION reverse
! _______________________
!
    end function reverse
!
! =========================================================================================
!
!
! ******************************
! END OF MODULE Eig_Procedures *
! ******************************
!    
end module Eig_Procedures
