module LLSQ_Procedures
! ##############################################################################
!
! Copyright 2021 IRD
!
! This file is part of statpack.
!
! statpack is free software: you can redistribute it and/or modify
! it under the terms of the GNU Lesser General Public License as
! published by the Free Software Foundation, either version 3 of 
! the License, or (at your option) any later version.
!
! statpack is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU Lesser General Public License for more details.
!
! You can find a copy of the GNU Lesser General Public License
! in the statpack/doc directory.
!
! ##############################################################################
!                                                                              *
! ******************************************************************************
! MODULE EXPORTING SUBROUTINES AND FUNCTIONS FOR SOLVING LINEAR LEAST SQUARES  *
! PROBLEMS.                                                                    *
!                                                                              *
! LATEST REVISION : 01/06/2021                                                 *
!                                                                              *
! ##############################################################################
!                                                                              *
! ******************************************************************************
!                                                                              *
! THE C PROCESSOR MACROS USED IN THIS MODULE ARE:                              *
!                                                                              *
!  _OPENMP      FOR ACTIVATING OPENMP PARALLELIZATION                          *
!  _BLAS        FOR USING BLAS WHEN POSSIBLE                                   *
!  _DOT_PRODUCT FOR REPLACING THE dot_product INTRINSIC FUNCTION WITH STATPACK *
!               dot_product2 FUNCTION                                          *
!  _MATMUL      FOR REPLACING THE matmul INTRINSIC FUNCTION WITH STATPACK      *
!               matmul2 FUNCTION                                               *
!  _TRANSPOSE   FOR REPLACING THE transpose INTRINSIC FUNCTION WITH STATPACK   *
!               transpose2 FUNCTION                                            *
!  _USE_NAGWARE FOR DESACTIVATING SOME OPENMP CONSTRUCTS FOR THE NAG FORTRAN   *
!               COMPILER                                                       *
!                                                                              *
! ******************************************************************************
!
!
! USED MODULES
! ============
!   
    use Select_Parameters,     only : lgl, i4b, stnd
    use Reals_Constants,       only : zero, one, c1_6, c2000
    use Logical_Constants
#ifdef _MATMUL
    use Utilities,             only : matmul=>matmul2
#endif
#ifdef _DOT_PRODUCT
    use Utilities,             only : dot_product=>dot_product2
#endif
#ifdef _TRANSPOSE
    use Utilities,             only : transpose=>transpose2
#endif
#ifdef _OPENMP
    use Select_Parameters,     only : omp_limit
    use omp_lib,               only : omp_get_max_threads, omp_in_parallel, omp_get_nested
#endif
#ifdef _BLAS
    use BLAS_interfaces,       only : axpy, trsv, trsm
#endif
!
! USED MODULES IN SPECIFIC PROCEDURES
! ===================================
!   
!   use Num_Constants
!   use Char_Constants
!   use Utilities
!   use Hous_Procedures
!   use QR_Procedures
!   use SVD_Procedures
!
! STRONG TYPING IMPOSED 
! =====================
!
    implicit none
!
! PUBLIC ENTITIES 
! ===============
!    
! ALL SUBROUTINES, FUNCTIONS, VARIABLES AND PARAMETERS ARE PRIVATE BY DEFAULT.
!
    private
    public ::   solve_llsq, llsq_qr_solve, llsq_qr_solve2,              &
                qr_solve, qr_solve2, llsq_svd_solve
!
! GENERIC INTERFACES FOR ROUTINES WITH OVERLOADED VERSIONS
! ========================================================
!
    interface solve_llsq
        module procedure    solve_llsq_rm_rv, solve_llsq_rm_rm,    &
                            solve_llsq_rv_rv, solve_llsq_rv_rm
    end interface
!
    interface llsq_qr_solve
        module procedure    llsq_qr_solve_rm_rv, llsq_qr_solve_rm_rm,    &
                            llsq_qr_solve_rv_rv, llsq_qr_solve_rv_rm
    end interface
!
    interface llsq_qr_solve2
        module procedure    llsq_qr_solve2_rm_rv, llsq_qr_solve2_rm_rm,    &
                            llsq_qr_solve2_rv_rv, llsq_qr_solve2_rv_rm
    end interface
!
    interface qr_solve
        module procedure    qr_solve_rv, qr_solve_rm
    end interface
!
    interface qr_solve2
        module procedure    qr_solve2_rv, qr_solve2_rm
    end interface
!
    interface llsq_svd_solve
        module procedure    llsq_svd_solve_rv, llsq_svd_solve_rm
    end interface
!
!
! =========================================================================================
!
                                  contains
!                                 ========
!
!
! =========================================================================================
!                           LINEAR LEAST SQUARES PROBLEMS
! =========================================================================================
!
!
    function solve_llsq_rm_rv( a, b, krank, tol, min_norm ) result( x )
!
! Purpose
! _______
!
!   SOLVE_LLSQ computes a solution to a real linear least squares problem:
!
!                       Minimize 2-norm || B - A * X ||
!
!   using an orthogonal factorization with columns pivoting of A. A is an 
!   m-by-n matrix which may be rank-deficient. m>=n or n>m is permitted. Here,
!   B is a m-elements right hand side vector and X is a n-elements solution vector.
!
!   The function returns the n-elements solution vector X.
!
!   A and B are not overwritten by SOLVE_LLSQ.
!
!
! Arguments
! _________
!
!   A            (INPUT) real(stnd), dimension(:,:)
!                On entry, the m-by-n coefficient matrix A.
!
!   B            (INPUT) real(stnd), dimension(:)
!                On entry, the m-elements right hand side vector B.
!
!                The shape of B must verify:
!
!                - size( B ) = size( A, 1 ) = m .
!
!   KRANK        (INPUT, OPTIONAL) integer(i4b)
!                On entry, KRANK=k implies that the first k columns
!                of A are to be forced into the basis, pivoting is 
!                performed on the last n-k columns of A.
!
!                When KRANK >=min(m,n) is used, pivoting is not performed.
!                This is appropriate when A is known to be full rank.
!
!                If KRANK is absent or is <=0, pivoting is done on all 
!                columns of A. This is appropriate if A may not be
!                of full rank (i.e. certain columns of A are linear 
!                combinations of other columns).
!                
!   TOL          (INPUT, OPTIONAL) real(stnd)
!                On entry, TOL is used to determine the effective rank
!                of A, which is then defined as the order of the largest 
!                leading triangular submatrix R11 in the QR factorization
!                (with pivoting) of A whose estimated condition number,
!                in the 1-norm, is less than 1/TOL. TOl must be set to 
!                the relative precision of the elements in A and B.
!                If each element is correct to, say, 5 digits then
!                TOL=0.00001 should be used.
!
!                TOL must not be greater or equal to 1 or less or equal
!                than 0, otherwise the numerical rank of A is determined
!                and the calculations to determine the condition number
!                are not performed. If TOL is absent, the numerical 
!                rank of A is determined.
!
!   MIN_NORM     (INPUT, OPTIONAL) logical(lgl)
!                On entry, if: 
!
!                - MIN_NORM=true, the minimun 2-norm solution
!                  is computed.
!                - MIN_NORM=false or if MIN_NORM is absent, a solution
!                  is computed such that if the j-th column of A
!                  is omitted from the basis, X[j] is set to zero.
!
!
! Further Details
! _______________
!
!   1) The routine first computes a QR factorization with (partial) column pivoting
!      on option (see below):
!
!                       A * P = Q * R
!
!      , here P is n-by-n permutation matrix, R is an upper triangular or trapezoidal
!      (if n>m) matrix and Q is a m-by-m orthogonal matrix.
!
!      R can then be partioned by defining R11 as the largest leading submatrix of R whose
!      estimated condition number, in the 1-norm, is less than 1/TOL (or such that abs(R[j,j])>0
!      if TOL is absent). The order of R11, arank, is the effective rank of A.
!
!      This leads to the following partition of R:
!
!      [  R11  R12  ]
!
!      [  R21  R22  ]
!
!      where R21 is zero by construction (since R is an upper triangular or trapezoidal)
!      and R22 is considered to be negligible.
!
!
!   2) If MIN_NORM is present and has the value true, R12 is annihilated by 
!      orthogonal transformations from the right, arriving at the complete 
!      orthogonal factorization:
!
!                       A * P = Q * T * Z
!
!      , where Z is a n-by-n orthogonal matrix and T has the form:
!
!      [  T11  T12  ]
!
!      [  T21  T22  ]
!
!      , here T21 (=R21), T12 and T22 (=R22) are zero and T11 is a arank-by-arank upper
!      triangular matrix.
!   
!      The minimum 2-norm solution is then
!
!                       X = [ P * Z' ](:,:arank) * [ inv(T11) * Q1' * B ]
!
!      where inv(T11) is the inverse of T11, Z' is the transpose of Z and Q1 consists
!      of the first arank columns of Q.
!
!
!   3) If MIN_NORM is absent or has the value false, a solution is computed as 
!
!                       X = P(:,:arank) * [ inv(R11) * Q1' * B ]
!
!      where inv(R11) is the inverse of R11 and Q1 consists of the first arank columns of Q.
!      In this case, if the j-th column of A is omitted from the basis, X[j] is set to zero.
!
!
!   4) On input, if KRANK is present and KRANK=k, the first k columns of A are to
!      be forced into the basis. Pivoting is performed on the last n-k columns of A.
!
!      When KRANK is present and KRANK>=min(m,n) is used, pivoting is not performed. This
!      is appropriate when A is known to be of full rank.
!   
!      If KRANK is absent or is present with KRANK<=0, pivoting is done on all columns of A.
!
!
!   5) TOL is an optional argument such that 0<TOL<1. If TOL is not specified, 
!      or is outside ]0,1[, the calculations to determine the condition number of A 
!      are not performed and crude tests on R(j,j) are performed to determine the numerical
!      rank of A. If TOL is present and is in ]0,1[, the calculations to determine the  
!      condition number are performed. 
! 
!
!   6) If it is possible that A may not be full rank (i.e., certain columns of A are
!      linear combinations of other columns), then the linearly dependent columns can
!      usually be eliminated by using KRANK=0  and TOL=relative precision of the elements
!      in A and B. If each element is correct to, say, 5 digits then TOL=0.00001 should
!      be used. Also, it may be helpful to scale the columns of A so that all elements 
!      are about the same order of magnitude.
!
!
!   7) On exit, if A or B are empty, the function returns a n-elements vector filled
!      with nan() value.
!
!   For further details on linear least square problems and algorithms to solve them, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!          Matrix Computations.
!          3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (3) Hansen, P.C., Pereyra, V., and Scherer, G., 2012:
!           Least Squares Data Fitting with Applications.
!           Johns Hopkins University Press, 328 pp.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,    only : allocate_error
    use Num_Constants,     only : machhuge, nan
    use Utilities,         only : assert_eq, merror
    use QR_Procedures,     only : qrfac
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in)            :: a(:,:), b(:)
    real(stnd)                          :: x(size(a,2))
    real(stnd),   intent(in), optional  :: tol
!
    integer(i4b), intent(in), optional  :: krank
!
    logical(lgl), intent(in), optional  :: min_norm
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: cl
    real(stnd), dimension(size(b),size(a,2)+1)    :: syst
    real(stnd), dimension(min(size(a,2),size(b))) :: beta, diagr
    real(stnd), dimension(size(a,2))              :: tau
    real(stnd), dimension(:,:),       allocatable :: hmat
!
    integer(i4b) :: n, m, ndiag, np1, i, k, kfix, kp1, l
    integer(i4b), dimension(:), allocatable :: ip
    integer                                 :: iok
!
    logical(lgl) :: comp_min_norm
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='solve_llsq'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n = size( a, 2 )
    m =  assert_eq( int(size(a,1),i4b) ,   &
                    int(size(b),i4b)   ,   &
                    name_proc )
!
    if ( present(tol) ) then
!
        cl = merge( tol, machhuge, tol>zero .and. tol<one )
!
    else
!
        cl = machhuge
!
    end if
!
    if ( present(min_norm) ) then
!
        comp_min_norm = min_norm

    else
!
        comp_min_norm = false
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    ndiag = min( n, m )
!
    if ( ndiag<=0_i4b ) then
!
        x(:n) = nan()
!
        return
!
    else if ( all(a(:m,:n)==zero) ) then
!
        x(:n) = zero
!
        return
!
    end if
!    
!   COMPUTE QR FACTORIZATION WITH COLUMN PIVOTING OF a .
!   APPLY THE HOUSEHOLDER TRANSFORMATIONS TO b .
!    
    np1 = n + 1_i4b
!    
    syst(:m,:n) = a(:m,:n)
    syst(:m,np1) = b(:m)
!    
    if ( present(krank) ) then
!
        kfix = max( krank, 0_i4b )
!
    else
!
        kfix = 0_i4b
!
    end if
!    
    k = n
!    
    if ( kfix<ndiag ) then
!
!       ALLOCATE WORK VARIABLE FOR STORING PERMUTATIONS.
!
        allocate( ip(n), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        call qrfac( name_proc, syst, kfix, k, comp_min_norm, diagr, beta, tau,      &
                    tol=cl, ip=ip )
!
    else
!
        call qrfac( name_proc, syst, kfix, k, comp_min_norm, diagr, beta, tau,      &
                    tol=cl )
!
    end if
!    
!   SOLVE THE k BY k TRIANGULAR SYSTEM.
!    
#ifdef _BLAS
    do i = 1_i4b, k
!
        syst(i,i) = diagr(i)
!
    end do
!
    call trsv( 'U', 'N', 'N', k, syst, m, syst(1_i4b:k,np1), 1_i4b )
!
    x(1_i4b:k) = syst(1_i4b:k,np1)
#else
    do i = k, 1_i4b, -1_i4b
!
        x(i) = syst(i,np1)/diagr(i)
!
        syst(1_i4b:i-1_i4b,np1) = syst(1_i4b:i-1_i4b,np1) - x(i)*syst(1_i4b:i-1_i4b,i)
!
    end do
#endif
!    
!   COMPLETE COMPUTATION OF SOLUTION VECTOR.
!    
    kp1 = k + 1_i4b
!    
    x(kp1:n) = zero
!
!   COMPUTE MINIMAL NORM SOLUTION VECTOR IF REQUIRED.
!
    if ( k/=n .and. comp_min_norm ) then
!
        l = n - k
!
!       ALLOCATE WORK VARIABLES FOR STORING HOUSEHOLDER VECTORS.
!
        allocate( hmat(l,k), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       TRANSPOSE HOUSEHOLDER VECTORS FOR LATER USE.
!
        hmat(:l,:k) = transpose( syst(:k,kp1:n) )
!
        do i = 1_i4b, k
!
            if ( tau(i)==zero ) cycle
!
            cl       = ( x(i) + dot_product(hmat(:l,i),x(kp1:n)) )*tau(i)
            x(i)     = x(i) + cl
            x(kp1:n) = x(kp1:n) + hmat(:l,i)*cl
!
        end do
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( hmat )
!
    end if
!    
!   RE-ORDER THE SOLUTION VECTOR TO COMPENSATE FOR THE COLUMN INTERCHANGES.
!   DEALLOCATE ip ON EXIT, IF NECESSARY.
!    
    if ( kfix<ndiag ) then
!
        tau(:n) = x(:n)
        x(ip)   = tau(:n)
!
!       DEALLOCATE WORK VARIABLE.
!
        deallocate( ip )
!
    end if
!
!
! END OF FUNCTION solve_llsq_rm_rv
! ________________________________
!
    end function solve_llsq_rm_rv
!
! =========================================================================================
!
    function solve_llsq_rm_rm( a, b, krank, tol, min_norm ) result( x )
!
! Purpose
! _______
!
!   SOLVE_LLSQ computes solutions to real linear least squares problems 
!   of the form:
!
!                       Minimize 2-norm|| B - A * X ||
!
!   using an orthogonal factorization with columns pivoting of A. A is an 
!   m-by-n matrix which may be rank-deficient. m>=n or n>m is permitted.
!
!   Several right hand side vectors b can be  handled in a single call; 
!   they are stored as the columns of the m-by-nb right hand side matrix B.
!
!   The function returns the n-by-nb solution matrix X.
!
!   A and B are not overwritten by SOLVE_LLSQ.
!
!
! Arguments
! _________
!
!   A            (INPUT) real(stnd), dimension(:,:)
!                On entry, the m-by-n coefficient matrix A.
!
!   B            (INPUT) real(stnd), dimension(:,:)
!                On entry, the m-by-nb right hand side matrix B.
!
!                The shape of B must verify:
!
!                - size( B, 1 ) = size( A, 1 ) = m .
!
!   KRANK        (INPUT, OPTIONAL) integer(i4b)
!                On entry, KRANK=k implies that the first k columns
!                of A are to be forced into the basis, pivoting is 
!                performed on the last n-k columns of A.
!
!                When KRANK >=min(m,n) is used, pivoting is not performed.
!                This is appropriate when A is known to be full rank.
!
!                If KRANK is absent or is <=0, pivoting is done on all 
!                columns of A. This is appropriate if A may not be
!                of full rank (i.e. certain columns of A are linear 
!                combinations of other columns).
!                
!   TOL          (INPUT, OPTIONAL) real(stnd)
!                On entry, TOL is used to determine the effective rank
!                of A, which is then defined as the order of the largest 
!                leading triangular submatrix R11 in the QR factorization
!                (with pivoting) of A whose estimated condition number,
!                in the 1-norm, is less than 1/TOL. TOl must be set to 
!                the relative precision of the elements in A and B.
!                If each element is correct to, say, 5 digits then
!                TOL=0.00001 should be used.
!
!                TOL must not be greater or equal to 1 or less or equal
!                than 0, otherwise the numerical rank of A is determined
!                and the calculations to determine the condition number
!                are not performed. If TOL is absent, the numerical 
!                rank of A is determined.
!
!   MIN_NORM     (INPUT, OPTIONAL) logical(lgl)
!                On entry, if: 
!
!                - MIN_NORM=true, the minimun 2-norm solutions
!                  are computed.
!                - MIN_NORM=false or if MIN_NORM is absent, solutions
!                  are computed such that if the j-th column of A
!                  is omitted from the basis, X[j,:] is set to zero.
!
!
! Further Details
! _______________
!
!   1) The routine first computes a QR factorization with (partial) column pivoting
!      on option (see below):
!
!                       A * P = Q * R
!
!      , here P is n-by-n permutation matrix, R is an upper triangular or trapezoidal
!      (if n>m) matrix and Q is a m-by-m orthogonal matrix.
!
!      R can then be partioned by defining R11 as the largest leading submatrix of R whose
!      estimated condition number, in the 1-norm, is less than 1/TOL (or such that abs(R[j,j])>0
!      if TOL is absent). The order of R11, arank, is the effective rank of A.
!
!      This leads to the following partition of R:
!
!      [  R11  R12  ]
!
!      [  R21  R22  ]
!
!      where R21 is zero by construction (since R is an upper triangular or trapezoidal)
!      and R22 is considered to be negligible.
!
!
!   2) If MIN_NORM is present and has the value true, R12 is annihilated by 
!      orthogonal transformations from the right, arriving at the complete 
!      orthogonal factorization:
!
!                       A * P = Q * T * Z
!
!      , where Z is a n-by-n orthogonal matrix and T has the form:
!
!      [  T11  T12  ]
!
!      [  T21  T22  ]
!
!      , here T21 (=R21), T12 and T22 (=R22) are zero and T11 is a arank-by-arank upper
!      triangular matrix.
!   
!      The minimum 2-norm solution is then
!
!                       X = [ P * Z' ](:,:arank) * [ inv(T11) * Q1' * B ]
!
!      where inv(T11) is the inverse of T11, Z' is the transpose of Z and Q1 consists
!      of the first arank columns of Q.
!
!
!   3) If MIN_NORM is absent or has the value false, a solution is computed as 
!
!                          X = P(:,:arank) * [ inv(R11) * Q1' * B ]
!
!      where inv(R11) is the inverse of R11 and Q1 consists of the first arank columns of Q.
!      In this case, if the j-th column of A is omitted from the basis, X[j] is set to zero.
!
!
!   4) On input, if KRANK is present and KRANK=k, the first k columns of A are to
!      be forced into the basis. Pivoting is performed on the last n-k columns of A.
!
!      When KRANK is present and KRANK>=min(m,n) is used, pivoting is not performed. This
!      is appropriate when A is known to be full rank.
!   
!      If KRANK is absent or is present with KRANK<=0, pivoting is done on all columns of A.
!
!
!   5) TOL is an optional argument such that 0<TOL<1. If TOL is not specified, 
!      or is outside ]0,1[, the calculations to determine the condition number of A 
!      are not performed and crude tests on R(j,j) are performed to determine the numerical
!      rank of A. If TOL is present and is in ]0,1[, the calculations to determine the  
!      condition number are performed. 
! 
!
!   6) If it is possible that A may not be full rank (i.e., certain columns of A are
!      linear combinations of other columns), then the linearly dependent columns can
!      usually be eliminated by using KRANK=0  and TOL=relative precision of the elements
!      in A and B. If each element is correct to, say, 5 digits then TOL=0.00001 should
!      be used. Also, it may be helpful to scale the columns of A so that all elements 
!      are about the same order of magnitude.
!
!
!   7) On exit, if A or B are empty, the function returns a n-by-nb matrix filled
!      with nan() value.
!
!   For further details on linear least square problems and algorithms to solve them, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!          Matrix Computations.
!          3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (3) Hansen, P.C., Pereyra, V., and Scherer, G., 2012:
!           Least Squares Data Fitting with Applications.
!           Johns Hopkins University Press, 328 pp.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,    only : allocate_error
    use Num_Constants,     only : machhuge, nan
    use Utilities,         only : assert_eq, merror
    use QR_Procedures,     only : qrfac
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in), dimension(:,:) :: a, b
    real(stnd),   intent(in), optional       :: tol
    real(stnd), dimension(size(a,dim=2),size(b,dim=2)) :: x
!
    integer(i4b), intent(in), optional       :: krank
!
    logical(lgl), intent(in), optional       :: min_norm
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)  :: cl
    real(stnd), dimension(size(a,dim=2),size(b,dim=2))               :: x2
    real(stnd), dimension(size(a,dim=1),size(a,dim=2)+size(b,dim=2)) :: syst
    real(stnd), dimension(size(a,dim=2))                             :: tau
    real(stnd), dimension(size(b,dim=2))                             :: tmp
    real(stnd), dimension(min(size(a,dim=1),size(a,dim=2)))          :: beta, diagr
    real(stnd), dimension(:,:),                          allocatable :: hmat, x2t
!
    integer(i4b)  :: n, m, ndiag, nb, np1, npnb, i, k, kfix, kp1, l, j
    integer(i4b), dimension(:), allocatable :: ip
    integer                                 :: iok
!
    logical(lgl) :: comp_min_norm
#ifndef _BLAS
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc ='solve_llsq'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    n = size( a, 2 )
    m =  assert_eq( int(size(a,1),i4b)   ,   &
                    int(size(b,1),i4b)   ,   &
                    name_proc )
    nb =  size( b, 2 )
!
    if ( present(tol) ) then
!
        cl = merge( tol, machhuge, tol>zero .and. tol<one )
!
    else
!
        cl = machhuge
!
    end if
!
    if ( present(min_norm) ) then
!
        comp_min_norm = min_norm
!
    else
!
        comp_min_norm = false
!
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    ndiag = min( n, m )
!
    if ( min(nb,ndiag)<=0_i4b ) then
!
        x(:n,:nb) = nan()
!
        return
!
    else if ( all(a(:,:)==zero) ) then
!
        x(:n,:nb) = zero
!
        return
!
    end if
!    
!   COMPUTE QR FACTORIZATION WITH COLUMN PIVOTING OF a .
!   APPLY THE HOUSEHOLDER TRANSFORMATIONS TO b .
!    
    np1  = n + 1_i4b
    npnb = n + nb
!    
    syst(:m,:n) = a(:m,:n)
    syst(:m,np1:npnb) = b(:m,:nb)
!    
    if ( present(krank) ) then
!
        kfix = max( krank, 0_i4b )
!
    else
!
        kfix = 0_i4b
!
    end if
!    
    k = n
!    
    if ( kfix<ndiag ) then
!
!       ALLOCATE WORK VARIABLE FOR STORING PERMUTATIONS.
!
        allocate( ip(n), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        call qrfac( name_proc, syst, kfix, k, comp_min_norm, diagr, beta, tau,  &
                    tol=cl, ip=ip )
!
    else
!
        call qrfac( name_proc, syst, kfix, k, comp_min_norm, diagr, beta, tau,   &
                    tol=cl )
!
    end if
!    
!   SOLVE THE k BY k TRIANGULAR SYSTEMS.
!    
#ifdef _BLAS
    do i = 1_i4b, k
!
        syst(i,i) = diagr(i)
!
    end do
!
    x2(1_i4b:k,1_i4b:nb) = syst(1_i4b:k,np1:npnb)
!
    call trsm( 'L', 'U', 'N', 'N', k, nb, one, syst, m, x2, n )
#else
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and. nb>=2_i4b .and. i>1_i4b
#endif
!    
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC,1), PRIVATE(i,l,cl)
!
    do l = np1, npnb
!
        do i = k, 1_i4b, -1_i4b
!
            cl = syst(i,l)/diagr(i)
            syst(1_i4b:i-1_i4b,l) = syst(1_i4b:i-1_i4b,l) - cl*syst(1_i4b:i-1_i4b,i)
            syst(i,l) = cl
!
        end do
!    
        x2(:k,l-n) = syst(:k,l)
!
    end do
!    
!$OMP END PARALLEL DO
#endif
!    
!   COMPLETE COMPUTATION OF SOLUTION VECTOR.
!    
    kp1 = k + 1_i4b
!    
    x2(kp1:n,:nb) = zero
!
!   COMPUTE MINIMAL NORM SOLUTION VECTORS IF REQUIRED.
!
    if ( k/=n .and. comp_min_norm ) then
!
        l = n - k
!
!       ALLOCATE WORK VARIABLES FOR STORING HOUSEHOLDER AND SOLUTION VECTORS.
!
        allocate( hmat(l,k), x2t(nb,n), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       TRANSPOSE HOUSEHOLDER AND SOLUTION VECTORS FOR LATER USE.
!
        hmat(:l,:k) = transpose( syst(:k,kp1:n) )
!
        x2t(:nb,:n) = transpose( x2(:n,:nb) )
!
        do i = 1_i4b, k
!
            if ( tau(i)==zero ) cycle
!
            tmp(:nb)   = matmul( x2t(:nb,kp1:n), hmat(:l,i) )
            tmp(:nb)   = ( x2t(:nb,i) + tmp(:nb) )*tau(i)
            x2t(:nb,i) = x2t(:nb,i) + tmp(:nb)
!
            do j=kp1, n
!
                x2t(:nb,j) = x2t(:nb,j) + hmat(j-k,i)*tmp(:nb)
!
            end do
!
        end do
!
        x2(:n,:nb) = transpose( x2t(:nb,:n) )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( hmat, x2t )
!
    end if
!    
!   RE-ORDER THE SOLUTION VECTOR TO COMPENSATE FOR THE COLUMN INTERCHANGES.
!   DEALLOCATE ip ON EXIT, IF NECESSARY.
!    
    if ( kfix<ndiag ) then
!
        do i = 1_i4b, nb
!
            tau(ip)  = x2(:n,i)
            x2(:n,i) = tau(:n)
!
        end do
!
!       DEALLOCATE WORK VARIABLE.
!
        deallocate( ip )
!
    end if
!
!   OUTPUT THE RESULTS.
!
    x(:n,:nb) = x2(:n,:nb)
!
!
! END OF FUNCTION solve_llsq_rm_rm
! ________________________________
!
    end function solve_llsq_rm_rm
!
! =========================================================================================
!
    function solve_llsq_rv_rv( a, b ) result( x )
!
! Purpose
! _______
!
!   SOLVE_LLSQ computes a solution to a real linear least squares problem:
!
!                       Minimize 2-norm|| B - A * X ||
!
!   A is an m vector, B is a m right hand side vector and X is a real scalar. 
!
!   The function returns the solution scalar X.
!
!   A and B are not overwritten by SOLVE_LLSQ.
!
!
! Arguments
! _________
!
!   A       (INPUT) real(stnd), dimension(:)
!           On entry, the m-elements coefficient vector A.
!
!   B       (INPUT) real(stnd), dimension(:)
!           On entry, the m-elements right hand side vector B.
!
!           The shape of B must verify:
!
!           - size( B ) = size( A ) = m .
!
!
! Further Details
! _______________
!
!   1) The routine first generates a real elementary reflector H of order m, such that
!
!           H * A = D , with  H' * H = I and D' = (  d  0  )
!
!      where d is a scalar. H is represented in the form
!
!           H = I + beta * ( v * v' ) ,
!                      
!      where beta is a real scalar and v is a real m-element vector.
!
!
!   2) The solution X is then computed as
!
!                       X = [ H * B ](1) / d
!
!
!   3) On exit, if A or B are empty, the function returns a nan() value.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : nan
    use Utilities,       only : assert_eq
    use Hous_Procedures, only : h1
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in), dimension(:) :: a, b
    real(stnd)                             :: x
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                      :: beta, diagr
    real(stnd), dimension(size(a))  :: a_tmp
!
    integer(i4b)  ::  m           
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc ='solve_llsq'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m =  assert_eq( int(size(a),i4b)     ,   &
                    int(size(b),i4b)     ,   &
                    name_proc )
!
    if ( m<=0_i4b ) then
!
        x = nan()
!
    else 
!
        a_tmp(:m) = a(:m)
!
        call h1( a_tmp, diagr, beta )
!
        if ( beta==zero ) then
!
!           QUICK RETURN IF POSSIBLE.
!
            if ( diagr==zero ) then
!
                x = zero
!
            else
!
                x = b(1_i4b)/diagr
!
            end if
!
        else
!    
!           APPLY THE HOUSEHOLDER TRANSFORMATION TO b(1).
!    
            x = b(1_i4b) + a_tmp(1_i4b)*( dot_product(a_tmp(:m),b(:m))*beta )
!    
!           SOLVE THE 1 BY 1 TRIANGULAR SYSTEM.
!    
            x = x/diagr
!
        end if
!
    end if
!
!
! END OF FUNCTION solve_llsq_rv_rv
! ________________________________
!
    end function solve_llsq_rv_rv
!
! =========================================================================================
!
    function solve_llsq_rv_rm( a, b ) result( x )
!
! Purpose
! _______
!
!   SOLVE_LLSQ computes solutions to real linear least squares problems 
!   of the form:
!
!                       Minimize 2-norm|| B - A * X ||
!
!   A is an m vector and several right hand side vectors b can be  handled in a single call; 
!   they are stored as the columns of the m-by-nb right hand side matrix B.
!
!   The function returns the nb-elements solution vector X.
!
!   A and B are not overwritten by SOLVE_LLSQ.
!
!
! Arguments
! _________
!
!   A       (INPUT) real(stnd), dimension(:)
!           On entry, the m-elements coefficient vector A.
!
!   B       (INPUT) real(stnd), dimension(:,:)
!           On entry, the m-by-nb right hand side matrix B.
!
!           The shape of B must verify:
!
!           - size( B, 1 ) = size( A ) = m .
!
!
! Further Details
! _______________
!
!   1) The routine first generates a real elementary reflector H of order m, such that
!
!           H * A = D , with  H' * H = I and D' = (  d  0  )
!
!      where d is a scalar. H is represented in the form
!
!           H = I + beta * ( v * v' ) ,
!                      
!      where beta is a real scalar and v is a real m-element vector.
!
!
!   2) The solution vector X is then computed as
!
!                       X(:) = [ H * B ](1,:) / d
!
!
!   3) On exit, if A or B are empty, the function returns a nb-vector filled
!      with nan() value.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : nan
    use Utilities,       only : assert_eq
    use Hous_Procedures, only : h1
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)  :: a(:), b(:,:)
    real(stnd)              :: x(size(b,dim=2))
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)  :: a_tmp(size(a)), tmp(size(b,dim=2)), beta, diagr
!
    integer(i4b)  :: nb, n
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc ='solve_llsq'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(a),i4b)   ,   &
                    int(size(b,1),i4b) ,   &
                    name_proc )
    nb =  size(b,2)
!
    if ( min(nb,n)<=0_i4b ) then
!
        x(:nb) = nan()
!
    else
!
        a_tmp(:n) = a(:n)
!
        call h1( a_tmp, diagr, beta )
!
        if ( beta==zero ) then
!
!           QUICK RETURN IF POSSIBLE.
!
            if ( diagr==zero ) then
!
                x(:nb) = zero
!
            else
!
                x(:nb) = b(1_i4b,:nb)/diagr
!
            end if
!
        else
!    
!           APPLY THE HOUSEHOLDER TRANSFORMATION TO b(1,:nb) .
!    
            tmp(:nb) = matmul( a_tmp(:n), b(:n,:nb) )
            tmp(:nb) = tmp(:nb)*beta
            tmp(:nb) = b(1_i4b,:nb) + a_tmp(1_i4b)*tmp(:nb)
!
!           SOLVE THE 1 BY 1 TRIANGULAR SYSTEMS.
!    
            x(:nb) = tmp(:nb)/diagr
!
        end if
!
    end if
!
!
! END OF FUNCTION solve_llsq_rv_rm
! ________________________________
!
    end function solve_llsq_rv_rm
!
! =========================================================================================
!
    subroutine llsq_qr_solve_rm_rv( mat, b, x, rnorm, resid, krank, tol, min_norm )
!
! Purpose
! _______
!
!   LLSQ_QR_SOLVE computes a solution to a real linear least squares problem:
!
!                       Minimize 2-norm|| B - MAT * X ||
!
!   using an orthogonal factorization with columns pivoting of MAT. MAT is an 
!   m-by-n matrix which may be rank-deficient. m>=n or n>m is permitted. Here,
!   B is a m-elements right hand side vector and X is a n-elements solution vector.
!
!   MAT and B are not overwritten by LLSQ_QR_SOLVE.
!
!
! Arguments
! _________
!
!   MAT          (INPUT) real(stnd), dimension(:,:)
!                On entry, the m-by-n coefficient matrix MAT.
!
!   B            (INPUT) real(stnd), dimension(:)
!                On entry, the m-elements right hand side vector B.
!
!                The shape of B must verify:
!
!                - size( B ) = size( MAT, 1 ) = m .
!
!   X            (OUTPUT) real(stnd), dimension(:)
!                On exit, the n-elements solution vector X.
!
!                The shape of X must verify:
!
!                - size( X ) = size( MAT, 2 ) = n .
!
!   RNORM        (OUTPUT, OPTIONAL) real(stnd)
!                On exit, the 2-norm of the residual vector for 
!                the solution vector X.
!
!   RESID        (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                On exit, the m-elements residual vector for the solution 
!                vector X, RESID = B - MAT * X.
!
!                The shape of RESID must verify:
!
!                - size( RESID ) = size( B ) = size( MAT, 1 ) = m .
!
!   KRANK        (INPUT/OUTPUT, OPTIONAL) integer(i4b)
!                On entry, KRANK=k implies that the first k columns
!                of MAT are to be forced into the basis, pivoting is 
!                performed on the last n-k columns of MAT.
!
!                When KRANK >=min(m,n) is used, pivoting is not performed.
!                This is appropriate when MAT is known to be full rank.
!
!                If KRANK is absent or is <=0, pivoting is done on all 
!                columns of MAT. This is appropriate if MAT may not be
!                of full rank (i.e. certain columns of MAT are linear 
!                combinations of other columns).
!                
!                On exit, KRANK contains the effective rank of MAT,
!                i.e., the number of independent columns in matrix MAT.
!
!   TOL          (INPUT/OUTPUT, OPTIONAL) real(stnd)
!                On entry, TOL is used to determine the effective rank
!                of MAT, which is then defined as the order of the largest 
!                leading triangular submatrix R11 in the QR factorization
!                (with pivoting) of MAT whose estimated condition number,
!                in the 1-norm, is less than 1/TOL. TOl must be set to 
!                the relative precision of the elements in MAT and B.
!                If each element is correct to, say, 5 digits then
!                TOL=0.00001 should be used.
!
!                TOL must not be greater or equal to 1 or less than 0,
!                otherwise the numerical rank of MAT is determined and
!                the calculations to determine the condition number
!                are not performed. If TOL=0, the numerical rank of MAT is 
!                determined, but the condition number is calculated.
!
!                On exit, if a condition number is calculated, its reciprocal
!                is returned in TOL. Otherwise, TOL is not changed.
!
!                If TOL is absent, the numerical rank of MAT is used and 
!                is returned in the optional argument KRANK.
!
!   MIN_NORM     (INPUT, OPTIONAL) logical(lgl)
!                On entry, if: 
!
!                - MIN_NORM=true, the minimun 2-norm solution
!                  is computed.
!                - MIN_NORM=false or if MIN_NORM is absent, a solution
!                  is computed such that if the j-th column of A
!                  is omitted from the basis, X[j] is set to zero.
!
!
! Further Details
! _______________
!
!   1) The routine first computes a QR factorization with (partial) column pivoting
!      on option (see below):
!
!                       MAT * P = Q * R
!
!      , here P is n-by-n permutation matrix, R is an upper triangular or trapezoidal
!      (if n>m) matrix and Q is a m-by-m orthogonal matrix.
!
!      R can then be partioned by defining R11 as the largest leading submatrix of R whose
!      estimated condition number, in the 1-norm, is less than 1/TOL (or such that abs(R[j,j])>0
!      if TOL is absent). The order of R11, KRANK, is the effective rank of MAT.
!
!      This leads to the following partition of R:
!
!      [  R11  R12  ]
!
!      [  R21  R22  ]
!
!      where R21 is zero by construction (since R is an upper triangular or trapezoidal)
!      and R22 is considered to be negligible.
!
!
!   2) If MIN_NORM is present and has the value true, R12 is annihilated by 
!      orthogonal transformations from the right, arriving at the complete 
!      orthogonal factorization:
!
!                       MAT * P = Q * T * Z
!
!      , where Z is a n-by-n orthogonal matrix and T has the form:
!
!      [  T11  T12  ]
!
!      [  T21  T22  ]
!
!      , here T21 (=R21), T12 and T22 (=R22) are zero and T11 is a KRANK-by-KRANK upper
!      triangular matrix.
!   
!      The minimum 2-norm solution is then
!
!                       X = [ P * Z' ](:,:KRANK) * [ inv(T11) * Q1' * B ]
!
!      where inv(T11) is the inverse of T11, Z' is the transpose of Z and Q1 consists
!      of the first KRANK columns of Q.
!
!
!   3) If MIN_NORM is absent or has the value false, a solution is computed as 
!
!                       X = P(:,:KRANK) * [ inv(R11) * Q1' * B ]
!
!      where inv(R11) is the inverse of R11 and Q1 consists of the first KRANK columns of Q.
!      In this case, if the j-th column of MAT is omitted from the basis, X[j] is set to zero.
!
!
!   4) In both cases:
!
!      - The 2-norm of the residual vector for the solution X can 
!        be obtained through the optional argument RNORM .
!
!      - The m-elements residual vector, B - MAT * X, can be obtained through the optional
!        argument RESID.
!
!
!   5) On input, if KRANK is present and KRANK=k, the first k columns of MAT are to
!      be forced into the basis. Pivoting is performed on the last n-k columns of MAT.
!
!      When KRANK is present and KRANK>=min(m,n) is used, pivoting is not performed. This
!      is appropriate when MAT is known to be full rank.
!   
!      If KRANK is absent or is present with KRANK<=0, pivoting is done on all columns of MAT.
!
!      On output, if KRANK is present, it contains the effective rank of MAT, i.e., the
!      order of the submatrix R11.
!
!
!   6) TOL is an optional argument such that 0<=TOL<1. If TOL is not specified, 
!      or is outside [0,1[, the calculations to determine the condition number of MAT 
!      are not performed and crude tests on R(j,j) are performed to determine the numerical
!      rank of MAT. If TOL is present and is in [0,1[, the calculations to determine the  
!      condition number are performed and its reciprocal is return in TOL. 
! 
!   7) If it is possible that MAT may not be full rank (i.e., certain columns of MAT are
!      linear combinations of other columns), then the linearly dependent columns can
!      usually be eliminated by using KRANK=0  and TOL=relative precision of the elements
!      in MAT and B. If each element is correct to, say, 5 digits then TOL=0.00001 should
!      be used. Also, it may be helpful to scale the columns of MAT so that all elements 
!      are about the same order of magnitude.
!
!   For further details on linear least square problems and algorithms to solve them, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!          Matrix Computations.
!          3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (3) Hansen, P.C., Pereyra, V., and Scherer, G., 2012:
!           Least Squares Data Fitting with Applications.
!           Johns Hopkins University Press, 328 pp.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,     only : machhuge
    use Char_Constants,    only : llsq_error1, allocate_error
    use Utilities,         only : merror, assert_eq, norm
    use QR_Procedures,     only : qrfac, apply_q_qr
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in)               :: mat(:,:), b(:)
    real(stnd),   intent(out)              :: x(:)
    real(stnd),   intent(inout), optional  :: tol
    real(stnd),   intent(out),   optional  :: rnorm, resid(:)
!
    integer(i4b), intent(inout), optional :: krank
!
    logical(lgl), intent(in),    optional  :: min_norm
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: cl
    real(stnd), dimension(size(b),size(x)+1)    :: syst
    real(stnd), dimension(min(size(x),size(b))) :: beta, diagr
    real(stnd), dimension(size(x))              :: tau
    real(stnd), dimension(:,:),     allocatable :: hmat
!
    integer(i4b) :: n, m, ndiag, np1, i, k, kfix, kp1, l
    integer(i4b), dimension(:), allocatable :: ip
    integer                                 :: iok
!
    logical(lgl) :: comp_min_norm, comp_resid, comp_rnorm
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='llsq_qr_solve'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,2),i4b) ,      &
                       int(size(x),i4b)     ,   &
                       name_proc )
!
    comp_resid = present( resid )
!
    if ( comp_resid ) then
!
        m =  assert_eq( int(size(mat,1),i4b)   ,   &
                        int(size(b),i4b)       ,   &
                        int(size(resid),i4b)   ,   &
                        name_proc )
!
    else
!
        m =  assert_eq( int(size(mat,1),i4b) ,   &
                        int(size(b),i4b)     ,   &
                        name_proc )
!
    end if
!
    if ( present(min_norm) ) then
        comp_min_norm = min_norm
    else
        comp_min_norm = false
    end if
!
    comp_rnorm = present( rnorm )
!
!   QUICK RETURN IF POSSIBLE.
!
    ndiag = min( n, m )
!
    if ( ndiag<=0_i4b ) then
!
        call merror( name_proc//llsq_error1 )
!
    else if ( all(mat(:m,:n)==zero) ) then
!
        x(:n) = zero
!
        if ( present(tol) )    tol = machhuge
        if ( present(krank) )  krank = 0_i4b
        if ( comp_rnorm )      rnorm = norm( b )
        if ( comp_resid )      resid(:m) = b(:m)
!
        return
!
    end if
!    
!   COMPUTE QR FACTORIZATION WITH COLUMN PIVOTING OF mat .
!   APPLY THE HOUSEHOLDER TRANSFORMATIONS TO b .
!    
    np1 = n + 1_i4b
!    
    syst(:m,:n)  = mat(:m,:n)
    syst(:m,np1) = b(:m)
!    
    if ( present(krank) ) then
!
        kfix = max( krank, 0_i4b )
!
    else
!
        kfix = 0_i4b
!
    end if
!    
    k = n
!    
    if ( kfix<ndiag ) then
!
!       ALLOCATE WORK VARIABLE FOR STORING PERMUTATIONS.
!
        allocate( ip(n), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        call qrfac( name_proc, syst, kfix, k, comp_min_norm, diagr, beta, tau,  &
                    tol=tol, ip=ip )
!
    else
!
        call qrfac( name_proc, syst, kfix, k, comp_min_norm, diagr, beta, tau,  &
                    tol=tol )
!
    end if
!
    kp1 = k + 1_i4b
!    
!   COMPUTE resid = b - mat*x IF REQUIRED.
!    
    if ( comp_resid ) then
!
        resid(kp1:m)   = syst(kp1:m,np1)
        resid(1_i4b:k) = zero
!
        call apply_q_qr( syst(:m,:ndiag), beta(:ndiag), resid(:m), trans=false )
!
    end if
!    
!   SOLVE THE k BY k TRIANGULAR SYSTEM.
!    
#ifdef _BLAS
    do i = 1_i4b, k
!
        syst(i,i) = diagr(i)
!
    end do
!
    call trsv( 'U', 'N', 'N', k, syst, m, syst(1_i4b:k,np1), 1_i4b )
!
    x(1_i4b:k) = syst(1_i4b:k,np1)
#else
    do i = k, 1_i4b, -1_i4b
!
        x(i) = syst(i,np1)/diagr(i)
!
        syst(1_i4b:i-1_i4b,np1) = syst(1_i4b:i-1_i4b,np1) - x(i)*syst(1_i4b:i-1_i4b,i)
!
    end do
#endif
!    
!   COMPLETE COMPUTATION OF SOLUTION VECTOR.
!    
    x(kp1:n) = zero
!
!   COMPUTE MINIMAL NORM SOLUTION VECTOR IF REQUIRED.
!
    if ( k/=n .and. comp_min_norm ) then
!
        l = n - k
!
!       ALLOCATE WORK VARIABLES FOR STORING HOUSEHOLDER VECTORS.
!
        allocate( hmat(l,k), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       TRANSPOSE HOUSEHOLDER VECTORS FOR LATER USE.
!
        hmat(:l,:k) = transpose( syst(:k,kp1:n) )
!
        do i = 1_i4b, k
!
            if ( tau(i)==zero ) cycle
!
            cl       = ( x(i) + dot_product(hmat(:l,i),x(kp1:n)) )*tau(i)
            x(i)     = x(i) + cl
            x(kp1:n) = x(kp1:n) + hmat(:l,i)*cl
!
        end do
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( hmat )
!
    end if
!    
!   RE-ORDER THE SOLUTION VECTOR TO COMPENSATE FOR THE COLUMN INTERCHANGES.
!   DEALLOCATE ip ON EXIT, IF NECESSARY.
!    
    if ( kfix<ndiag ) then
!
        tau(:n) = x(:n)
        x(ip)   = tau(:n)
!
!       DEALLOCATE WORK VARIABLE.
!
        deallocate( ip )
!
    end if
!
!   COMPUTE THE NORM OF THE RESIDUAL VECTOR, IF REQUIRED.
!    
    if ( comp_rnorm ) then
        rnorm = norm( syst(kp1:m,np1) )
    end if
!
!   OUTPUT RANK OF mat ON EXIT, IF REQUIRED.
!
    if ( present(krank) ) then
        krank = k
    end if
!
!
! END OF SUBROUTINE llsq_qr_solve_rm_rv
! _____________________________________
!
    end subroutine llsq_qr_solve_rm_rv
!
! =========================================================================================
!
    subroutine llsq_qr_solve_rm_rm( mat, b, x, rnorm, resid, krank, tol, min_norm )
!
! Purpose
! _______
!
!   LLSQ_QR_SOLVE computes solutions to real linear least squares problems 
!   of the form:
!
!                       Minimize 2-norm|| B - MAT * X ||
!
!   using an orthogonal factorization with columns pivoting of MAT. MAT is an 
!   m-by-n matrix which may be rank-deficient. m>=n or n>m is permitted.
!
!   Several right hand side vectors b and solution vectors x can be
!   handled in a single call; they are stored as the columns of the
!   m-by-nb right hand side matrix B and the n-by-nb solution
!   matrix X, respectively.
!
!   MAT and B are not overwritten by LLSQ_QR_SOLVE.
!
!
! Arguments
! _________
!
!   MAT          (INPUT) real(stnd), dimension(:,:)
!                On entry, the m-by-n coefficient matrix MAT.
!
!   B            (INPUT) real(stnd), dimension(:,:)
!                On entry, the m-by-nb right hand side matrix B.
!
!                The shape of B must verify: 
!
!                - size( B, 1 ) = size( MAT, 1 ) = m
!                - size( B, 2 ) = size( X, 2 )   = nb .
!
!   X            (OUTPUT) real(stnd), dimension(:,:)
!                On exit, the n-by-nb solution matrix X.
!
!                The shape of X must verify: 
!
!                - size( X, 1 ) = size( MAT, 2 ) = n
!                - size( X, 2 ) = size( B, 2 )   = nb .
!
!   RNORM        (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                On exit, the 2-norm of the residual vectors for 
!                the solutions stored columnwise in the matrix X.
!
!                The size of RNORM must verify:
!
!                - size( RNORM ) = size( X, 2 ) = size( B, 2 )  = nb .
!
!   RESID        (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!                On exit, the residual vectors for the solutions 
!                stored columnwise in the matrix X, RESID = B - MAT * X.
!
!                The shape of RESID must verify: 
!
!                - size( RESID, 1 ) = size( B, 1 ) = size( MAT, 1 ) = m
!                - size( RESID, 2 ) = size( B, 2 ) = size( X, 2 ) = nb .
!
!   KRANK        (INPUT/OUTPUT, OPTIONAL) integer(i4b)
!                On entry, KRANK=k implies that the first k columns
!                of MAT are to be forced into the basis, pivoting is 
!                performed on the last n-k columns of MAT.
!
!                When KRANK >=min(m,n) is used, pivoting is not performed.
!                This is appropriate when MAT is known to be full rank.
!
!                If KRANK is absent or is <=0, pivoting is done on all 
!                columns of MAT. This is appropriate if MAT may not be
!                of full rank (i.e. certain columns of MAT are linear 
!                combinations of other columns).
!                
!                On exit, KRANK contains the effective rank of MAT,
!                i.e., the number of independent columns in matrix MAT.
!
!   TOL          (INPUT/OUTPUT, OPTIONAL) real(stnd)
!                On entry, TOL is used to determine the effective rank
!                of MAT, which is then defined as the order of the largest 
!                leading triangular submatrix R11 in the QR factorization
!                (with pivoting) of MAT whose estimated condition number,
!                in the 1-norm, is less than 1/TOL. TOl must be set to 
!                the relative precision of the elements in MAT and B.
!                If each element is correct to, say, 5 digits then
!                TOL=0.00001 should be used.
!
!                TOL must not be greater or equal to 1 or less than 0,
!                otherwise the numerical rank of MAT is determined and
!                the calculations to determine the condition number
!                are not performed. If TOL=0, the numerical rank of MAT is 
!                determined, but the condition number is calculated.
!
!                On exit, if a condition number is calculated, its reciprocal
!                is returned in TOL. Otherwise, TOL is not changed.
!
!                If TOL is absent, the numerical rank of MAT is used and 
!                is returned in the optional argument KRANK.
!
!   MIN_NORM     (INPUT, OPTIONAL) logical(lgl)
!                On entry, if: 
!
!                - MIN_NORM=true, the minimun 2-norm solutions
!                  are computed.
!                - MIN_NORM=false or if MIN_NORM is absent, solutions
!                  are computed such that if the j-th column of A
!                  is omitted from the basis, X[j,:] is set to zero.
!
!
! Further Details
! _______________
!
!   1) The routine first computes a QR factorization with (partial) column pivoting
!      on option (see below):
!
!                       MAT * P = Q * R
!
!      , here P is n-by-n permutation matrix, R is an upper triangular or trapezoidal
!      (if n>m) matrix and Q is a m-by-m orthogonal matrix.
!
!      R can then be partioned by defining R11 as the largest leading submatrix of R whose
!      estimated condition number, in the 1-norm, is less than 1/TOL (or such that abs(R[j,j])>0
!      if TOL is absent). The order of R11, KRANK, is the effective rank of MAT.
!
!      This leads to the following partition of R:
!
!      [  R11  R12  ]
!
!      [  R21  R22  ]
!
!      where R21 is zero by construction (since R is an upper triangular or trapezoidal)
!      and R22 is considered to be negligible.
!
!
!   2) If MIN_NORM is present and has the value true, R12 is annihilated by 
!      orthogonal transformations from the right, arriving at the complete 
!      orthogonal factorization:
!
!                       MAT * P = Q * T * Z
!
!      , where Z is a n-by-n orthogonal matrix and T has the form:
!
!      [  T11  T12  ]
!
!      [  T21  T22  ]
!
!      , here T21 (=R21), T12 and T22 (=R22) are zero and T11 is a KRANK-by-KRANK upper
!      triangular matrix.
!   
!      The minimum 2-norm solution is then
!
!                       X = [ P * Z' ](:,:KRANK) * [ inv(T11) * Q1' * B ]
!
!      where inv(T11) is the inverse of T11, Z' is the transpose of Z and Q1 consists
!      of the first KRANK columns of Q.
!
!
!   3) If MIN_NORM is absent or has the value false, a solution is computed as 
!
!                       X = P(:,:KRANK) * [ inv(R11) * Q1' * B ]
!
!      where inv(R11) is the inverse of R11 and Q1 consists of the first KRANK columns of Q.
!      In this case, if the j-th column of MAT is omitted from the basis, X[j] is set to zero.
!
!
!   4) In both cases:
!
!      - The 2-norm of the residual vector for the solution in the
!        j-th column of X is given in RNORM[j] if argument RNORM is present.
!
!      - The residual matrix, B - MAT * X, can be obtained through the optional
!        argument RESID.
!
!
!   5) On input, if KRANK is present and KRANK=k, the first k columns of MAT are to
!      be forced into the basis. Pivoting is performed on the last n-k columns of MAT.
!
!      When KRANK is present and KRANK>=min(m,n) is used, pivoting is not performed. This
!      is appropriate when MAT is known to be full rank.
!   
!      If KRANK is absent or is present with KRANK<=0, pivoting is done on all columns of MAT.
!
!      On output, if KRANK is present, it contains the effective rank of MAT, i.e., the
!      order of the submatrix R11.
!
!
!   6) TOL is an optional argument such that 0<=TOL<1. If TOL is not specified, 
!      or is outside [0,1[, the calculations to determine the condition number of MAT 
!      are not performed and crude tests on R(j,j) are performed to determine the numerical
!      rank of MAT. If TOL is present and is in [0,1[, the calculations to determine the  
!      condition number are performed and its reciprocal is return in TOL. 
! 
! 
!   7) If it is possible that MAT may not be full rank (i.e., certain columns of MAT are
!      linear combinations of other columns), then the linearly dependent columns can
!      usually be eliminated by using KRANK=0  and TOL=relative precision of the elements
!      in MAT and B. If each element is correct to, say, 5 digits then TOL=0.00001 should
!      be used. Also, it may be helpful to scale the columns of MAT so that all elements 
!      are about the same order of magnitude.
!
!   For further details on linear least square problems and algorithms to solve them, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!          Matrix Computations.
!          3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (3) Hansen, P.C., Pereyra, V., and Scherer, G., 2012:
!           Least Squares Data Fitting with Applications.
!           Johns Hopkins University Press, 328 pp.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,     only : machhuge
    use Char_Constants,    only : llsq_error2, allocate_error
    use Utilities,         only : merror, assert_eq, assert, norm
    use QR_Procedures,     only : qrfac, apply_q_qr
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in),  dimension(:,:) :: mat, b
    real(stnd),   intent(out), dimension(:,:) :: x
    real(stnd),   intent(inout), optional     :: tol
    real(stnd),   intent(out),   optional     :: rnorm(:), resid(:,:)
!
    integer(i4b), intent(inout), optional :: krank
!
    logical(lgl), intent(in),    optional  :: min_norm
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(mat,dim=2))                                :: tau
    real(stnd), dimension(size(b,dim=2))                                  :: tmp
    real(stnd), dimension(min(size(mat,dim=1),size(mat,dim=2)))           :: beta, diagr
    real(stnd), dimension(size(mat,dim=1),size(mat,dim=2)+size(b,dim=2))  :: syst
    real(stnd), dimension(:,:), allocatable                               :: hmat, xt
!
    integer(i4b)  :: n, m, ndiag, nb, np1, npnb, i, k, kfix, kp1, j, l
    integer(i4b), dimension(:), allocatable :: ip
    integer                                 :: iok
!
    logical(lgl) :: comp_min_norm
#ifndef _BLAS
!
    real(stnd)   :: cl
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc ='llsq_qr_solve'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,2),i4b) ,   &
                    int(size(x,1),i4b)   ,   &
                    name_proc )
!
    m =  assert_eq( int(size(mat,1),i4b)   ,   &
                    int(size(b,1),i4b)     ,   &
                    name_proc )
!
    nb =  assert_eq( int(size(b,2),i4b)     ,   &
                     int(size(x,2),i4b)     ,   &
                     name_proc )
!
    if ( present(rnorm) )                                      &
        call assert( logical(nb==int(size(rnorm),i4b),lgl) ,   &
                     name_proc )
!
    if ( present(resid) )                                        &
        call assert( logical(m==int(size(resid,1),i4b),lgl)  ,   &
                     logical(nb==int(size(resid,2),i4b),lgl) ,   &
                     name_proc )
!
    if ( present(min_norm) ) then
        comp_min_norm = min_norm
    else
        comp_min_norm = false
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    ndiag = min( n, m )
!
    if ( min(nb,ndiag)<=0_i4b ) then
!
        call merror( name_proc//llsq_error2 )
!
    else if ( all(mat(:,:)==zero) ) then
!
        x(:n,:nb) = zero
!
        if ( present(tol) )     tol = machhuge
        if ( present(krank) )   krank = 0_i4b
        if ( present(resid) )   resid(:m,:nb) = b(:m,:nb)
!
        if ( present(rnorm) )   then
            rnorm = norm( b, dim=2_i4b )
        end if
!
        return
!
    end if
!    
!   COMPUTE QR FACTORIZATION WITH COLUMN PIVOTING OF mat .
!   APPLY THE HOUSEHOLDER TRANSFORMATIONS TO b .
!    
    np1  = n + 1_i4b
    npnb = n + nb
!    
    syst(:m,:n) = mat(:m,:n)
    syst(:m,np1:npnb) = b(:m,:nb)
!    
    if ( present(krank) ) then
!
        kfix = max( krank, 0_i4b )
!
    else
!
        kfix = 0_i4b
!
    end if
!    
    k = n
!    
    if ( kfix<ndiag ) then
!
!       ALLOCATE WORK VARIABLE FOR STORING PERMUTATIONS.
!
        allocate( ip(n), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        call qrfac( name_proc, syst, kfix, k, comp_min_norm, diagr, beta, tau,       &
                    tol=tol, ip=ip )
!
    else
!
        call qrfac( name_proc, syst, kfix, k, comp_min_norm, diagr, beta, tau,       &
                    tol=tol )
!
    end if
!
    kp1 = k + 1_i4b
!
!   COMPUTE resid = b - mat*x IF REQUIRED.
!    
    if ( present(resid) ) then
!
        resid(kp1:m,:nb)   = syst(kp1:m,np1:npnb)
        resid(1_i4b:k,:nb) = zero
!
        call apply_q_qr( syst(:m,:ndiag), beta(:ndiag), resid(:m,:nb), left=true, trans=false )
!
    end if
!
!   SOLVE THE k BY k TRIANGULAR SYSTEMS.
!    
#ifdef _BLAS
    do i = 1_i4b, k
!
        syst(i,i) = diagr(i)
!
    end do
!
    call trsm( 'L', 'U', 'N', 'N', k, nb, one, syst, m, syst(1_i4b:m,np1:npnb), m )
!
    x(1_i4b:k,1_i4b:nb) = syst(1_i4b:k,np1:npnb)
#else
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and. nb>=2_i4b .and. i>1_i4b
#endif
!    
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC,1), PRIVATE(i,l,cl)
!
    do l = np1, npnb
!
        do i = k, 1_i4b, -1_i4b
!
            cl = syst(i,l)/diagr(i)
            syst(1_i4b:i-1_i4b,l) = syst(1_i4b:i-1_i4b,l) - cl*syst(1_i4b:i-1_i4b,i)
            syst(i,l) = cl
!
        end do
!    
        x(:k,l-n) = syst(:k,l)
!
    end do
!    
!$OMP END PARALLEL DO
#endif
!    
!   COMPLETE COMPUTATION OF SOLUTION VECTORS.
!    
    x(kp1:n,:nb) = zero
!
!   COMPUTE MINIMAL NORM SOLUTION VECTORS IF REQUIRED.
!
    if ( k/=n .and. comp_min_norm ) then
!
        l = n - k
!
!       ALLOCATE WORK VARIABLES FOR STORING HOUSEHOLDER AND SOLUTION VECTORS.
!
        allocate( hmat(l,k), xt(nb,n), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       TRANSPOSE HOUSEHOLDER AND SOLUTION VECTORS FOR LATER USE.
!
        hmat(:l,:k) = transpose( syst(:k,kp1:n) )
!
        xt(:nb,:n) = transpose( x(:n,:nb) )
!
        do i = 1_i4b, k
!
            if ( tau(i)==zero ) cycle
!
            tmp(:nb)  = matmul( xt(:nb,kp1:n), hmat(:l,i) )
            tmp(:nb)  = ( xt(:nb,i) + tmp(:nb) )*tau(i)
            xt(:nb,i) = xt(:nb,i) + tmp(:nb)
!
            do j = kp1, n
!
                xt(:nb,j) = xt(:nb,j) + hmat(j-k,i)*tmp(:nb)
!
            end do
!
        end do
!
        x(:n,:nb) = transpose( xt(:nb,:n) )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( hmat, xt )
!
    end if
!    
!   RE-ORDER THE SOLUTION VECTORS TO COMPENSATE FOR THE COLUMN INTERCHANGES.
!   DEALLOCATE ip ON EXIT, IF NECESSARY.
!    
    if ( kfix<ndiag ) then
!
        do i = 1_i4b, nb
!
            tau(ip) = x(:n,i)
            x(:n,i) = tau(:n)
!
        end do
!
!       DEALLOCATE WORK VARIABLES.
!
        deallocate( ip )
!
    end if
!
!   COMPUTE THE NORM OF THE RESIDUAL VECTORS IF REQUIRED.
!    
    if ( present(rnorm) ) then
        rnorm = norm( syst(kp1:m,np1:npnb), dim=2_i4b )
    end if
!
!   OUTPUT RANK OF mat ON EXIT, IF REQUIRED.
!
    if ( present(krank) ) then
        krank = k
    end if
!
!
! END OF SUBROUTINE llsq_qr_solve_rm_rm
! _____________________________________
!
    end subroutine llsq_qr_solve_rm_rm
!
! =========================================================================================
!
    subroutine llsq_qr_solve_rv_rv( vec, b, x, rnorm, resid )
!
! Purpose
! _______
!
!   LLSQ_QR_SOLVE computes a solution to a real linear least squares problem:
!
!                       Minimize 2-norm|| B - VEC * X ||
!
!   VEC is an m vector, B is a m right hand side vector and X is a real scalar. 
!
!   VEC and B are not overwritten by LLSQ_QR_SOLVE.
!
!
! Arguments
! _________
!
!   VEC          (INPUT) real(stnd), dimension(:)
!                On entry, the m-elements coefficient vector VEC.
!
!   B            (INPUT) real(stnd), dimension(:)
!                On entry, the m-elements right hand side vector B.
!
!                The shape of B must verify:
!
!                - size( B ) = size( VEC ) = m .
!
!   X            (OUTPUT) real(stnd)
!                On exit, the real solution X.
!
!   RNORM        (OUTPUT, OPTIONAL) real(stnd)
!                On exit, the 2-norm of the residual vector for 
!                the solution scalar X.
!
!   RESID        (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                On exit, the m-elements residual vector for the solution 
!                X, RESID = B - MAT * X.
!
!                The shape of RESID must verify: 
!
!                - size( RESID ) = size( B ) = size( VEC ) = m .
!
!
! Further Details
! _______________
!
!   1) The routine first generates a real elementary reflector H of order m, such that
!
!           H * VEC = D , with  H' * H = I and D' = (  d  0  )
!
!      where d is a scalar. H is represented in the form
!
!           H = I + beta * ( v * v' ) ,
!                      
!      where beta is a real scalar and v is a real m-element vector.
!
!
!   2) The solution X is then computed as
!
!                       X = [ H * B ](1) / d
!
!
!   3) The 2-norm of the residual vector for the solution X can be obtained through
!      the optional argument RNORM as
!
!                       2-norm|| [ H * B ](2:) ||
!
!
!   4) The residual vector, B - VEC * X, can be obtained through the optional
!      argument RESID.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_constants,  only : llsq_error3, allocate_error
    use Utilities,       only : merror, assert_eq, norm
    use Hous_Procedures, only : h1
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in), dimension(:) :: vec, b
    real(stnd),   intent(out)              :: x
    real(stnd),   intent(out),   optional  :: rnorm, resid(:)
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                            :: beta, diagr
    real(stnd), dimension(size(vec))      :: vec_tmp
    real(stnd), dimension(:), allocatable :: b_tmp
!
    integer(i4b) ::  m           
    integer      ::  iok          
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc ='llsq_qr_solve'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( present(resid) ) then
!
        m =  assert_eq( int(size(vec),i4b)     ,   &
                        int(size(b),i4b)       ,   &
                        int(size(resid),i4b)   ,   &
                        name_proc )
!
    else
!
        m =  assert_eq( int(size(vec),i4b)     ,   &
                        int(size(b),i4b)       ,   &
                        name_proc )
!
    end if
!
    if ( m<=0_i4b ) then
!
        call merror( name_proc//llsq_error3 )
!
    else
!
        vec_tmp(:m) = vec(:m)
!
        call h1( vec_tmp, diagr, beta )
!
        if ( beta==zero ) then
!
!           QUICK RETURN IF POSSIBLE.
!
            if ( diagr==zero ) then
!
                x = zero
!
                if ( present(rnorm) ) then
!
                    rnorm = norm( b )
!
                end if
                if ( present(resid) ) then
!
                    resid(:m) = b(:m)
!
                end if
!
            else
!
                x = b(1_i4b)/diagr
!
                if ( present(rnorm) ) then
!
                    rnorm = norm( b(2_i4b:m) )
!
                end if
!
                if ( present(resid) ) then
!
                    resid(1_i4b)   = zero
                    resid(2_i4b:m) = b(2_i4b:m)
!
                end if
!
            end if
!
        else if ( present(rnorm) .or. present(resid) ) then
!
!           ALLOCATE WORK VARIABLE.
!
            allocate( b_tmp(m), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!    
!           APPLY THE HOUSEHOLDER TRANSFORMATION TO b .
!    
            b_tmp(:m) = b(:m) + vec_tmp(:m)*( dot_product(vec_tmp(:m),b(:m))*beta )
!    
!           SOLVE THE 1 BY 1 TRIANGULAR SYSTEM.
!    
            x = b_tmp(1_i4b)/diagr
!    
!           COMPUTE THE NORM OF THE RESIDUAL VECTOR IF REQUIRED.
!    
            if ( present(rnorm) ) rnorm = norm( b_tmp(2_i4b:m) )
!    
!           COMPUTE resid = b - vec*x IF REQUIRED.
!    
            if ( present(resid) ) then
!
                b_tmp(1_i4b) = zero
                resid(:m)    = b_tmp(:m) +                          &
                               vec_tmp(:m)*( dot_product(vec_tmp(2_i4b:m),b_tmp(2_i4b:m))*beta )
!
            end if
!
!           DEALLOCATE WORK VARIABLE.
!
            deallocate( b_tmp )
!
        else
!
!           APPLY THE HOUSEHOLDER TRANSFORMATION TO b(1).
!    
            x = b(1_i4b) + vec_tmp(1_i4b)*( dot_product(vec_tmp(:m),b(:m))*beta )
!    
!           SOLVE THE 1 BY 1 TRIANGULAR SYSTEM.
!    
            x = x/diagr
!
        end if
!
    end if
!
!
! END OF SUBROUTINE llsq_qr_solve_rv_rv
! _____________________________________
!
    end subroutine llsq_qr_solve_rv_rv
!
! =========================================================================================
!
    subroutine llsq_qr_solve_rv_rm( vec, b, x, rnorm, resid )
!
! Purpose
! _______
!
!   LLSQ_QR_SOLVE computes solutions to real linear least squares problems 
!   of the form:
!
!                       Minimize 2-norm|| B - VEC * X ||
!
!   VEC is an m vector and several right hand side vectors b and solution scalars x
!   can be handled in a single call; they are stored as the columns of the
!   m-by-nb right hand side matrix B and the nb-elements solution
!   vector X, respectively.
!
!   VEC and B are not overwritten by LLSQ_QR_SOLVE.
!
!
! Arguments
! _________
!
!   VEC          (INPUT) real(stnd), dimension(:)
!                On entry, the m-elements coefficient vector VEC.
!
!   B            (INPUT) real(stnd), dimension(:,:)
!                On entry, the m-by-nb right hand side matrix B.
!
!                The shape of B must verify: 
!
!                - size( B, 1 ) = size( VEC ) = m
!                - size( B, 2 ) = size( X )   = nb .
!
!   X            (OUTPUT) real(stnd), dimension(:)
!                On exit, the nb-elements solution vector X.
!
!                The shape of X must verify: size( X ) = size( B, 2 )   = nb .
!
!   RNORM        (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                On exit, the 2-norms of the residual vectors for 
!                the solutions stored in the vector X.
!
!                The size of RNORM must verify: 
!
!                - size( RNORM ) = size( X ) = size( B, 2 )  = nb .
!
!   RESID        (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!                On exit, the residual vectors for the solutions 
!                stored in the vector X, RESID = B - VEC * X.
!
!                The shape of RESID must verify: 
!
!                - size( RESID, 1 ) = size( B, 1 ) = size( VEC ) = m
!                - size( RESID, 2 ) = size( B, 2 ) = size( X ) = nb .
!
!
! Further Details
! _______________
!
!   1) The routine first generates a real elementary reflector H of order m, such that
!
!           H * A = D , with  H' * H = I and D' = (  d  0  )
!
!      where d is a scalar. H is represented in the form
!
!           H = I + beta * ( v * v' ) ,
!                      
!      where beta is a real scalar and v is a real m-element vector.
!
!
!   2) The solution vector X is then computed as
!
!                       X(:) = [ H * B ](1,:) / d
!
!
!   3) The 2-norm of the residual vector for the solution X[j] 
!      is given in RNORM[j] if argument RNORM is present.
!
!
!   4) The residual matrix, B - VEC * X, can be obtained through the optional
!      argument RESID.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : llsq_error4, allocate_error
    use Utilities,       only : merror, assert_eq, assert, norm
    use Hous_Procedures, only : h1
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),  dimension(:)   :: vec
    real(stnd), intent(in),  dimension(:,:) :: b
    real(stnd), intent(out), dimension(:)   :: x
    real(stnd), intent(out), dimension(:),   optional  :: rnorm
    real(stnd), intent(out), dimension(:,:), optional ::  resid
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                            :: beta, diagr
    real(stnd), dimension(size(vec))      :: vec2
    real(stnd), dimension(size(b,dim=2))  :: tmp
    real(stnd), dimension(:), allocatable :: b2
!
    integer(i4b)  :: nb, n, i
    integer       :: iok          
!
    logical(lgl)  :: comp_resid, comp_rnorm
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc ='llsq_qr_solve'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(vec),i4b)   ,   &
                    int(size(b,1),i4b)   ,   &
                    name_proc )
!
    nb =  assert_eq( int(size(b,2),i4b)     ,   &
                     int(size(x),i4b)       ,   &
                     name_proc )
!
    comp_rnorm = present( rnorm )
!
    if ( comp_rnorm ) then
!
        call assert( logical(nb==int(size(rnorm),i4b),lgl), name_proc )
!
    end if
!
    comp_resid = present( resid )
!
    if ( comp_resid ) then
!
        call assert( logical(n==int(size(resid,1),i4b),lgl)  ,   &
                     logical(nb==int(size(resid,2),i4b),lgl) ,   &
                     name_proc )
!
    end if
!
    if ( min(nb,n)<=0_i4b ) then
!
        call merror( name_proc//llsq_error4 )
!
    else
!
        vec2(:n) = vec(:n)
!
        call h1( vec2, diagr, beta )
!
        if ( beta==zero ) then
!
!           QUICK RETURN IF POSSIBLE.
!
            if ( diagr==zero ) then
!
                x(:nb) = zero
!
                if ( comp_rnorm ) then
!
                    rnorm = norm( b, dim=2_i4b )
!
                end if
!
                if ( comp_resid ) then
!
                    resid(:n,:nb) = b(:n,:nb)
!
                end if
!
            else
!
                x(:nb) = b(1_i4b,:nb)/diagr
!
                if ( comp_rnorm ) then
!
                    rnorm = norm( b(2_i4b:n,:nb), dim=2_i4b )
!
                end if
!
                if ( comp_resid ) then
!
                    do i = 1_i4b, nb
!
                        resid(1_i4b,i)   = zero
                        resid(2_i4b:n,i) = b(2_i4b:n,i)
!
                    end do
!
                end if
!
            end if
!
        else
!    
            tmp(:nb) = matmul( vec2(:n), b(:n,:nb) )
            tmp(:nb) = tmp(:nb)*beta
!
            if ( comp_rnorm .or. comp_resid ) then
!
!               ALLOCATE WORK VARIABLE.
!
                allocate( b2(n), stat = iok )
!
                if ( iok/=0 ) then
                    call merror( name_proc//allocate_error )
                end if
!    
!               APPLY THE HOUSEHOLDER TRANSFORMATION TO b .
!
                do i = 1_i4b, nb
!
                    b2(:n) = b(:n,i) + tmp(i)*vec2(:n)
!    
!                   SOLVE THE 1 BY 1 TRIANGULAR SYSTEM.
!    
                    x(i) = b2(1_i4b)/diagr
!    
!                   COMPUTE THE NORM OF THE RESIDUAL VECTOR IF REQUIRED.
!    
                    if ( comp_rnorm )  then
!    
                        rnorm(i) = norm( b2(2_i4b:n) )
!
                    end if
!    
!                   COMPUTE resid = b - vec*x IF REQUIRED.
!    
                    if ( comp_resid ) then
!
                        b2(1_i4b) = zero
                        tmp(i) = dot_product( vec2(2_i4b:n), b2(2_i4b:n) )
                        tmp(i) = tmp(i)*beta
!
                        resid(:n,i) = b2(:n) + tmp(i)*vec2(:n)
!
                    end if
!
                end do
!
!               DEALLOCATE WORK VARIABLE.
!
                deallocate( b2 )
!
            else
!    
!               APPLY THE HOUSEHOLDER TRANSFORMATION TO b(1,:nb) .
!    
                tmp(:nb) = b(1_i4b,:nb) + vec2(1_i4b)*tmp(:nb)
!    
!               SOLVE THE 1 BY 1 TRIANGULAR SYSTEMS.
!    
                x(:nb) = tmp(:nb)/diagr
!
            end if
!
        end if
!
    end if
!
!
! END OF SUBROUTINE llsq_qr_solve_rv_rm
! _____________________________________
!
    end subroutine llsq_qr_solve_rv_rm
!
! =========================================================================================
!
    subroutine llsq_qr_solve2_rm_rv( mat, b, x, rnorm,        &
                                     comp_resid, krank, tol, min_norm, diagr, beta, ip, tau )
!
! Purpose
! _______
!
!   LLSQ_QR_SOLVE2 computes a solution to a real linear least squares problem:
!
!                       Minimize 2-norm|| B - MAT * X ||
!
!   using a (complete) orthogonal factorization of MAT. MAT is an m-by-n matrix
!   which may be rank-deficient. m>=n or n>m is permitted. Here, B is a m-elements
!   right hand side vector and X is a n-elements solution vector.
!
!   MAT and B are overwritten with information generated by LLSQ_QR_SOLVE2.
!   The (complete) orthogonal factorization of MAT is saved in arguments MAT,
!   DIAGR, BETA, IP and TAU on output.
!
!
! Arguments
! _________
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the m-by-n coefficient matrix MAT.
!
!                On exit, MAT has been overwritten by details of its 
!                (complete) orthogonal factorization. Other parts of
!                the factorization can be obtained if the optional
!                arguments DIAGR, BETA, IP and TAU are present.
!
!                See Further Details.
!
!   B            (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the m-elements right hand side vector B.
!
!                On exit, if COMP_RESID is present and is equal true,
!                the residual vector B - MAT * X overwrites B on output.
!
!                The shape of B must verify:
!
!                - size( B ) = size( MAT, 1 ) = m .
!
!   X            (OUTPUT) real(stnd), dimension(:)
!                On exit, the n-elements solution vector X.
!
!                The shape of X must verify:
!
!                - size( X ) = size( MAT, 2 ) = n .
!
!   RNORM        (OUTPUT, OPTIONAL) real(stnd)
!                On exit, the 2-norm of the residual vector for 
!                the solution vector X.
!
!   COMP_RESID   (INPUT, OPTIONAL) logical(lgl)
!                On entry, if COMP_RESID is present and is equal true,
!                the residual vector B - MAT * X overwrites B on exit.
!
!   KRANK        (INPUT/OUTPUT, OPTIONAL) integer(i4b)
!                On entry, KRANK=k implies that the first k columns
!                of MAT are to be forced into the basis, pivoting is 
!                performed on the last n-k columns of MAT.
!
!                When KRANK >=min(m,n) is used, pivoting is not performed.
!                This is appropriate when MAT is known to be full rank.
!
!                If KRANK is absent or is <=0, pivoting is done on all 
!                columns of MAT. This is appropriate if MAT may not be
!                of full rank (i.e. certain columns of MAT are linear 
!                combinations of other columns).
!                
!                On exit, KRANK contains the effective rank of MAT,
!                i.e., the number of independent columns in matrix MAT.
!
!   TOL          (INPUT/OUTPUT, OPTIONAL) real(stnd)
!                On entry, TOL is used to determine the effective rank
!                of MAT, which is then defined as the order of the largest 
!                leading triangular submatrix R11 in the QR factorization
!                (with pivoting) of MAT whose estimated condition number,
!                in the 1-norm, is less than 1/TOL. TOL must be set to 
!                the relative precision of the elements in MAT and B.
!                If each element is correct to, say, 5 digits then
!                TOL=0.00001 should be used.
!
!                TOL must not be greater or equal to 1 or less than 0,
!                otherwise the numerical rank of MAT is determined and
!                the calculations to determine the condition number
!                are not performed. If TOL=0, the numerical rank of MAT is 
!                determined, but the condition number is calculated.
!
!                On exit, if a condition number is calculated, its reciprocal
!                is returned in TOL. Otherwise, TOL is not changed.
!
!                If TOL is absent, the numerical rank of MAT is used and 
!                is returned in the optional argument KRANK.
!
!   MIN_NORM     (INPUT, OPTIONAL) logical(lgl)
!                On entry, if: 
!
!                - MIN_NORM=true, a complete orthogonal factorization
!                  of MAT and the minimun 2-norm solution is computed.
!                - MIN_NORM=false or if MIN_NORM is absent, a QR factorization
!                  with column pivoting of MAT and a solution is computed
!                  such that if the j-th column of MAT is omitted from the basis,
!                  X[j] is set to zero.
!
!   DIAGR        (OUTPUT, OPTIONAL) real(stnd), dimension(:) 
!                On entry, if: 
!
!                - MIN_NORM=false or is absent, the diagonal elements of the matrix R.
!                - MIN_NORM=true, the diagonal elements of the matrix T11.
!                  The diagonal elements of T11 are stored in DIAGR(1:KRANK).
!
!                See Further Details.
!
!                The size of DIAGR must be min( size(MAT,1) , size(MAT,2) ).
!
!   BETA         (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                On exit, the scalar factors of the elementary reflectors defining Q.
!
!                See Further Details.
!
!                The size of BETA must be min( size(MAT,1) , size(MAT,2) ).
!
!   IP           (OUTPUT, OPTIONAL) integer(i4b), dimension(:) 
!                On exit, if IP(j)=k, then the j-th column of MAT*P was 
!                the k-th column of MAT.
!
!                See Further Details.
!
!                The size of IP must be size(MAT,2) = n.
!
!   TAU          (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                On exit, the scalars factors of the elementary reflectors defining Z
!                in the complete orthogonal factorization of MAT if MIN_NORM is present
!                and is equal to true. Otherwise, TAU is set to 0.
!
!                See Further Details.
!
!                The size of TAU must be min( size(MAT,1) , size(MAT,2) ).
!
!
! Further Details
! _______________
!
!   1) The routine first computes a QR factorization with (partial) column pivoting
!      on option (see below):
!
!                       MAT * P = Q * R
!
!      , here P is n-by-n permutation matrix, R is an upper triangular or trapezoidal
!      (if n>m) matrix and Q is a m-by-m orthogonal matrix.
!
!      The matrix Q is represented as a product of elementary reflectors
!
!            Q = H(1) * H(2) * ... * H(k), where k = min( size(MAT,1) , size(MAT,2) )
!
!      Each H(i) has the form
!
!            H(i) = I + beta * ( v * v' ) ,
!                      
!      where beta is a real scalar and v is a real m-element vector with v(1:i-1) = 0.
!      v(i:m) is stored on exit in MAT(i:m,i) and beta in BETA(i).
!
!      The matrix P is represented in the array IP as follows:
!      If IP(j) = i then the jth column of P is the ith canonical unit vector.
!
!      The elements above the diagonal of the array MAT contain the 
!      corresponding elements of the triangular matrix R. The elements on the diagonal
!      of R are stored in the array DIAGR. 
!
!      R can then be partioned by defining R11 as the largest leading submatrix of R whose
!      estimated condition number, in the 1-norm, is less than 1/TOL (or such that abs(R[j,j])>0
!      if TOL is absent). The order of R11, KRANK, is the effective rank of MAT.
!
!      This leads to the following partition of R:
!
!      [  R11  R12  ]
!
!      [  R21  R22  ]
!
!      where R21 is zero by construction (since R is an upper triangular or trapezoidal)
!      and R22 is considered to be negligible.
!
!
!   2) If MIN_NORM is present and has the value true, R12 is annihilated by 
!      orthogonal transformations from the right, arriving at the complete 
!      orthogonal factorization:
!
!                       MAT * P = Q * T * Z
!
!      , where Z is a n-by-n orthogonal matrix and T is a m-by-n matrix has the form:
!
!      [  T11  T12  ]
!
!      [  T21  T22  ]
!
!      , here T21 (=R21), T12 and T22 (=R22) are zero and T11 is a KRANK-by-KRANK upper
!      triangular matrix.
!
!      The factorization is obtained by Householder's method. The kth transformation
!      matrix, Z(k), which is used to introduce zeros into the kth row of R,
!      is given in the form
!
!      [  I  0    ]
!
!      [  0  T(k) ]
!
!      where
!
!            T(k) = I + tau * ( u(k) * u(k)' ) and u(k)' = (  1  0  z(k)  )
!
!      tau is a scalar, u(k) is a n-k+1 vector and z(k) is an (n-KRANK) element vector.
!      tau and z(k) are chosen to annihilate the elements of the kth row of R12.
!
!      The Z n-by-n orthogonal matrix is given by
!
!            Z = Z(1) * Z(2) * ... * Z(KRANK)
!
!      On exit, the scalar tau is returned in the kth element of TAU and the vector u(k)
!      in the kth row of MAT, such that the elements of z(k) are in MAT(k,KRANK+1:n).
!
!      On exit, the elements above the diagonal of the array section MAT(1:KRANK,1:KRANK)
!      contain the corresponding elements of the triangular matrix T11. The elements of
!      the diagonal of T11 are stored in the array section DIAGR(1:KRANK). The last part
!      of DIAGR is set to zero. In other words, T11 overwrites R11 and Z overwrites R12
!      on exit.
!   
!      The minimum 2-norm solution is then
!
!                       X = [ P * Z' ](:,:KRANK) * [ inv(T11) * Q1' * B ]
!
!      where inv(T11) is the inverse of T11, Z' is the transpose of Z and Q1 consists
!      of the first KRANK columns of Q.
!
!
!   3) If MIN_NORM is absent or has the value false, a solution is computed as 
!
!                       X = P(:,:KRANK) * [ inv(R11) * Q1' * B ]
!
!      where inv(R11) is the inverse of R11 and Q1 consists of the first KRANK columns of Q.
!      In this case, if the j-th column of MAT is omitted from the basis, X[j] is set to zero
!      and R is not destroyed in MAT.
!
!
!   4) In both cases:
!
!      - The 2-norm of the residual vector for the solution X can 
!        be obtained through the optional argument RNORM .
!
!      - If COMP_RESID=true, The m-elements residual vector B - MAT * X overwrites B on exit.
! 
!
!   5) On input, if KRANK is present and KRANK=k, the first k columns of MAT are to
!      be forced into the basis. Pivoting is performed on the last n-k columns of MAT.
!
!      When KRANK is present and KRANK>=min(m,n) is used, pivoting is not performed. This
!      is appropriate when MAT is known to be full rank.
!   
!      If KRANK is absent or is present with KRANK<=0, pivoting is done on all columns of MAT.
!
!      On output, if KRANK is present, it contains the rank of MAT, i.e., the order 
!      of the submatrix R11. This is the same as the order of the submatrix T11
!      in the complete orthogonal factorization of MAT.
! 
!
!   6) TOL is an optional argument such that 0<=TOL<1. If TOL is not specified, 
!      or is outside [0,1[, the calculations to determine the condition number of MAT 
!      are not performed and crude tests on R(j,j) are performed to determine the numerical
!      rank of MAT. If TOL is present and is in [0,1[, the calculations to determine the  
!      condition number are performed, the effective rank of MAT is determined and the
!      reciprocal of the condition number is returned in TOL. 
! 
! 
!   7) If it is possible that MAT may not be full rank (i.e., certain columns of MAT are
!      linear combinations of other columns), then the linearly dependent columns can
!      usually be eliminated by using KRANK=0  and TOL=relative precision of the elements
!      in MAT and B. If each element is correct to, say, 5 digits then TOL=0.00001 should
!      be used. Also, it may be helpful to scale the columns of MAT so that all elements 
!      are about the same order of magnitude.
!
!   For further details on linear least square problems and algorithms to solve them, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!          Matrix Computations.
!          3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (3) Hansen, P.C., Pereyra, V., and Scherer, G., 2012:
!           Least Squares Data Fitting with Applications.
!           Johns Hopkins University Press, 328 pp.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,     only : machhuge
    use Char_Constants,    only : llsq_error1, allocate_error
    use Utilities,         only : merror, assert_eq, assert, arth, norm
    use QR_Procedures,     only : qrfac, apply_q_qr
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout)                          :: mat(:,:), b(:)
    real(stnd),   intent(out)                            :: x(:) 
    real(stnd),   intent(inout), optional                :: tol
    real(stnd),   intent(out),   optional                :: rnorm
    real(stnd),   intent(out),   optional , dimension(:) :: diagr, beta, tau
!
    integer(i4b), intent(inout), optional :: krank
    integer(i4b), intent(out),   optional :: ip(:)
!
    logical(lgl), intent(in), optional  :: comp_resid, min_norm
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: cl
    real(stnd), dimension(size(x))              :: tau_tmp
    real(stnd), dimension(min(size(x),size(b))) :: diagr_tmp, beta_tmp
    real(stnd), dimension(:,:),     allocatable :: hmat
!#ifdef _BLAS
!    real(stnd), dimension(min(size(x)) :: h1_tmp
!#endif
!
    integer(i4b) :: m, n, ndiag, i, k, kfix, kp1, l
    integer(i4b), dimension(:), allocatable :: ip_tmp
    integer                                 :: iok
!
    logical(lgl) :: comp_min_norm, test_comp_resid, comp_tau
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc ='llsq_qr_solve2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(b),i4b)     ,   &
                    name_proc )
!
    n =  assert_eq( int(size(mat,2),i4b) ,   &
                    int(size(x),i4b)     ,   &
                    name_proc )
!
    ndiag = min( n, m )
!
    if ( present(diagr) ) call assert( logical(size(diagr)==ndiag,lgl), name_proc )
    if ( present(beta)  ) call assert( logical(size(beta)==ndiag,lgl), name_proc )
    if ( present(ip) )    call assert( logical(size(ip)==n,lgl), name_proc )
!
    comp_tau = present( tau )
!
    if ( comp_tau ) then
!
        call assert( logical(size(tau)==ndiag,lgl), name_proc )
        tau(:ndiag) = zero
!
    end if
!
    if ( present(min_norm) ) then
        comp_min_norm = min_norm
    else
        comp_min_norm = false
    end if
!
    if ( present(comp_resid) ) then
        test_comp_resid = comp_resid
    else
        test_comp_resid = false
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( ndiag<=0_i4b ) then
!
        call merror( name_proc//llsq_error1 )
!
    else if ( all(mat(:,:)==zero) ) then
!
        x(:n) = zero
!
        if ( present(rnorm) )   rnorm = norm( b )
        if ( present(krank) )   krank = 0_i4b
        if ( present(tol) )     tol = machhuge
        if ( present(diagr) )   diagr(:ndiag) = zero
        if ( present(beta) )    beta(:ndiag) = zero
        if ( present(ip) )      ip(:n) = arth( 1_i4b, 1_i4b, n )
!
        return
!
    end if
!    
!   COMPUTE QR FACTORIZATION WITH COLUMN PIVOTING OF mat .
!    
    if ( present(krank) ) then
        kfix = max( krank, 0_i4b )
    else
        kfix = 0_i4b
    end if
!    
    k = n
!    
    if ( present(ip) ) then
!
        call qrfac( name_proc, mat, kfix, k, comp_min_norm, diagr_tmp, beta_tmp, tau_tmp,      &
                    tol=tol, ip=ip )
!
    else if ( kfix<ndiag ) then
!
!       ALLOCATE WORK VARIABLE FOR STORING PERMUTATIONS.
!
        allocate( ip_tmp(n), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        call qrfac( name_proc, mat, kfix, k, comp_min_norm, diagr_tmp, beta_tmp, tau_tmp,      &
                    tol=tol, ip=ip_tmp )
!
    else
!
        call qrfac( name_proc, mat, kfix, k, comp_min_norm, diagr_tmp, beta_tmp, tau_tmp,      &
                    tol=tol )
!
    end if
!
!   APPLY THE HOUSEHOLDER TRANSFORMATIONS TO b .
!
    call apply_q_qr( mat(:m,:k), beta_tmp(:k), b(:m), trans=true )
!
    x(:k) = b(:k)
!
    kp1 = k + 1_i4b
!
!   OUTPUT OPTIONAL ARGUMENTS, IF REQUIRED.
!
    if ( present(rnorm) )  rnorm = norm( b(kp1:m) )
    if ( present(krank) )  krank = k
    if ( present(diagr) )  diagr(:ndiag) = diagr_tmp(:ndiag)
    if ( present(beta) )   beta(:ndiag) = beta_tmp(:ndiag)
!
!   SOLVE THE k BY k TRIANGULAR SYSTEM.
!
#ifdef _BLAS
    do i = 1_i4b, k
!
!        h1_tmp(i) = mat(i,i)
        mat(i,i) = diagr_tmp(i)
!
    end do
!
    call trsv( 'U', 'N', 'N', k, mat, m, x, 1_i4b )
!
    do i = 1_i4b, k
!
!        mat(i,i) = h1_tmp(i)
        mat(i,i) = one
!
    end do
#else
    do i = k, 1_i4b, -1_i4b
!
        x(i) = x(i)/diagr_tmp(i)
        x(1_i4b:i-1_i4b) = x(1_i4b:i-1_i4b) - x(i)*mat(1_i4b:i-1_i4b,i)
!
    end do
#endif
!
!   COMPLETE COMPUTATION OF SOLUTION VECTOR.
!
    x(kp1:n) = zero
!
!   COMPUTE MINIMAL NORM SOLUTION VECTOR IF REQUIRED.
!
    if ( k/=n .and. comp_min_norm ) then
!
        l = n - k
!
!       ALLOCATE WORK VARIABLES FOR STORING HOUSEHOLDER VECTORS.
!
        allocate( hmat(l,k), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       TRANSPOSE HOUSEHOLDER VECTORS FOR LATER USE.
!
        hmat(:l,:k) = transpose( mat(:k,kp1:n) )
!
        do i = 1_i4b, k
!
            if ( tau_tmp(i)==zero ) cycle
!
            cl       = ( x(i) + dot_product(hmat(:l,i),x(kp1:n)) )*tau_tmp(i)
            x(i)     = x(i) + cl
            x(kp1:n) = x(kp1:n) + hmat(:l,i)*cl
!
        end do
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( hmat )
!
!       OUTPUT OPTIONAL ARGUMENT tau ON EXIT, IF REQUIRED.
!
        if ( comp_tau ) then
            tau(1_i4b:k) = tau_tmp(1_i4b:k)
        end if
!
    end if
!
!   COMPUTE b - mat*x IF REQUIRED.
!
    if ( test_comp_resid ) then
!
        b(1_i4b:k) = zero
!
        call apply_q_qr( mat(:m,:k), beta_tmp(:k), b(:m), trans=false )
!
    end if
!    
!   RE-ORDER THE SOLUTION VECTOR TO COMPENSATE FOR THE COLUMN INTERCHANGES.
!   DEALLOCATE ip_tmp ON EXIT, IF NECESSARY.
!    
    if ( kfix<ndiag ) then
!
        tau_tmp(:n) = x(:n)
!
        if ( present(ip) ) then
!
            x(ip) = tau_tmp(:n)
!
        else
!
            x(ip_tmp) = tau_tmp(:n)
!
!           DEALLOCATE WORK VARIABLE.
!
            deallocate( ip_tmp )
!
        end if
!
    end if
!
!
! END OF SUBROUTINE llsq_qr_solve2_rm_rv
! ______________________________________
!
    end subroutine llsq_qr_solve2_rm_rv
!
! =========================================================================================
!
    subroutine llsq_qr_solve2_rm_rm( mat, b, x, rnorm,            &
                                     comp_resid, krank, tol, min_norm, diagr, beta, ip, tau )
!
! Purpose
! _______
!
!   LLSQ_QR_SOLVE2 computes solutions to real linear least squares problems 
!   of the form:
!
!                       Minimize 2-norm|| B - MAT * X ||
!
!   using an orthogonal factorization with columns pivoting of MAT. MAT is an 
!   m-by-n matrix which may be rank-deficient. m>=n or n>m is permitted.
!
!   Several right hand side vectors b and solution vectors x can be
!   handled in a single call; they are stored as the columns of the
!   m-by-nb right hand side matrix B and the n-by-nb solution
!   matrix X, respectively.
!
!   MAT and B are overwritten with information generated by LLSQ_QR_SOLVE2.
!   The (complete) orthogonal factorization of MAT is saved in arguments MAT,
!   DIAGR, BETA, IP and TAU on output.
!
!
! Arguments
! _________
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the m-by-n coefficient matrix MAT.
!
!                On exit, MAT has been overwritten by details of its 
!                (complete) orthogonal factorization. Other parts of
!                the factorization can be obtained if the optional
!                arguments DIAGR, BETA, IP and TAU are present.
!
!                See Further Details.
!
!   B            (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the m-by-nb right hand side matrix B.
!
!                On exit, if COMP_RESID is present and is equal true,
!                the residual matrix B - MAT * X overwrites B on output.
!
!                The shape of B must verify: 
!
!                - size( B, 1 ) = size( MAT, 1 ) = m
!                - size( B, 2 ) = size( X, 2 )   = nb .
!
!   X            (OUTPUT) real(stnd), dimension(:,:)
!                On exit, the n-by-nb solution matrix X.
!
!                The shape of X must verify: 
!
!                - size( X, 1 ) = size( MAT, 2 ) = n
!                - size( X, 2 ) = size( B, 2 )   = nb .
!
!   RNORM        (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                On exit, the 2-norm of the residual vectors for 
!                the solutions stored columnwise in the matrix X.
!
!                The size of RNORM must verify: 
!
!                - size( RNORM ) = size( X, 2 ) = size( B, 2 )  = nb .
!
!   COMP_RESID   (INPUT, OPTIONAL) logical(lgl)
!                On entry, if COMP_RESID is present and is equal true,
!                the residual matrix B - MAT * X overwrites B on exit.
!
!   KRANK        (INPUT/OUTPUT, OPTIONAL) integer(i4b)
!                On entry, KRANK=k implies that the first k columns
!                of MAT are to be forced into the basis, pivoting is 
!                performed on the last n-k columns of MAT.
!
!                When KRANK >=min(m,n) is used, pivoting is not performed.
!                This is appropriate when MAT is known to be full rank.
!
!                If KRANK is absent or is <=0, pivoting is done on all 
!                columns of MAT. This is appropriate if MAT may not be
!                of full rank (i.e. certain columns of MAT are linear 
!                combinations of other columns).
!                
!                On exit, KRANK contains the effective rank of MAT,
!                i.e., the number of independent columns in matrix MAT.
!
!   TOL          (INPUT/OUTPUT, OPTIONAL) real(stnd)
!                On entry, TOL is used to determine the effective rank
!                of MAT, which is then defined as the order of the largest 
!                leading triangular submatrix R11 in the QR factorization
!                (with pivoting) of MAT whose estimated condition number,
!                in the 1-norm, is less than 1/TOL. TOL must be set to 
!                the relative precision of the elements in MAT and B.
!                If each element is correct to, say, 5 digits then
!                TOL=0.00001 should be used.
!
!                TOL must not be greater or equal to 1 or less than 0,
!                otherwise the numerical rank of MAT is determined and
!                the calculations to determine the condition number
!                are not performed. If TOL=0, the numerical rank of MAT is 
!                determined, but the condition number is calculated.
!
!                On exit, if a condition number is calculated, its reciprocal
!                is returned in TOL. Otherwise, TOL is not changed.
!
!                If TOL is absent, the numerical rank of MAT is used and 
!                is returned in the optional argument KRANK.
!
!   MIN_NORM     (INPUT, OPTIONAL) logical(lgl)
!                On entry, if: 
!
!                - MIN_NORM=true, a complete orthogonal factorization
!                  of MAT and the minimun 2-norm solutions are computed.
!                - MIN_NORM=false or if MIN_NORM is absent, a QR factorization
!                  with column pivoting of MAT and solutions are computed
!                  such that if the j-th column of MAT is omitted from the basis,
!                  X[j,:] is set to zero.
!
!   DIAGR        (OUTPUT, OPTIONAL) real(stnd), dimension(:) 
!                On exit, the diagonal elements of the matrix R if MIN_NORM=false 
!                or is absent, or the diagonal elements of the matrix T11
!                if MIN_NORM is present and is equal to true. The diagonal
!                elements of T11 are stored in DIAGR(1:KRANK).
!
!                See Further Details.
!
!                The size of DIAGR must be min( size(MAT,1) , size(MAT,2) ).
!
!   BETA         (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                On exit, the scalars factors of the elementary reflectors defining Q.
!
!                See Further Details.
!
!                The size of BETA must be min( size(MAT,1) , size(MAT,2) ).
!
!   IP           (OUTPUT, OPTIONAL) integer(i4b), dimension(:) 
!                On exit, if IP(j)=k, then the j-th column of MAT * P was 
!                the k-th column of MAT.
!
!                See Further Details.
!
!                The size of IP must be size(MAT,2) = n.
!
!   TAU          (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                On exit, the scalars factors of the elementary reflectors defining Z
!                in the complete orthogonal factorization of MAT if MIN_NORM is present
!                and is equal to true. Otherwise, TAU is set to 0.
!
!                See Further Details.
!
!                The size of TAU must be min( size(MAT,1) , size(MAT,2) ).
!
!
! Further Details
! _______________
!
!   1) The routine first computes a QR factorization with (partial) column pivoting
!      on option (see below):
!
!                       MAT * P = Q * R
!
!      , here P is n-by-n permutation matrix, R is an upper triangular or trapezoidal
!      (if n>m) matrix and Q is a m-by-m orthogonal matrix.
!
!      The matrix Q is represented as a product of elementary reflectors
!
!            Q = H(1) * H(2) * ... * H(k), where k = min( size(MAT,1) , size(MAT,2) )
!
!      Each H(i) has the form
!
!            H(i) = I + beta * ( v * v' ) ,
!                      
!      where beta is a real scalar and v is a real m-element vector with v(1:i-1) = 0.
!      v(i:m) is stored on exit in MAT(i:m,i) and beta in BETA(i).
!
!      The matrix P is represented in the array IP as follows:
!      If IP(j) = i then the jth column of P is the ith canonical unit vector.
!
!      The elements above the diagonal of the array MAT contain the 
!      corresponding elements of the triangular matrix R. The elements on the diagonal
!      of R are stored in the array DIAGR. 
!
!      R can then be partioned by defining R11 as the largest leading submatrix of R whose
!      estimated condition number, in the 1-norm, is less than 1/TOL (or such that abs(R[j,j])>0
!      if TOL is absent). The order of R11, KRANK, is the effective rank of MAT.
!
!      This leads to the following partition of R:
!
!      [  R11  R12  ]
!
!      [  R21  R22  ]
!
!      where R21 is zero by construction (since R is an upper triangular or trapezoidal)
!      and R22 is considered to be negligible.
!
!
!   2) If MIN_NORM is present and has the value true, R12 is annihilated by 
!      orthogonal transformations from the right, arriving at the complete 
!      orthogonal factorization:
!
!                       MAT * P = Q * T * Z
!
!      , where Z is a n-by-n orthogonal matrix and T has the form:
!
!      [  T11  T12  ]
!
!      [  T21  T22  ]
!
!      , here T21 (=R21), T12 and T22 (=R22) are zero and T11 is a KRANK-by-KRANK upper
!      triangular matrix.
!
!      The factorization is obtained by Householder's method. The kth transformation
!      matrix, Z(k), which is used to introduce zeros into the kth row of R,
!      is given in the form
!
!      [  I  0    ]
!
!      [  0  T(k) ]
!
!      where
!
!            T(k) = I + tau * ( u(k) * u(k)' ) and u(k)' = (  1  0  z(k)  )
!
!      tau is a scalar, u(k) is a n-k+1 vector and z(k) is an (n-KRANK) element vector.
!      tau and z(k) are chosen to annihilate the elements of the kth row of R12.
!
!      The Z n-by-n orthogonal matrix is given by
!
!            Z = Z(1) * Z(2) * ... * Z(KRANK)
!
!      On exit, the scalar tau is returned in the kth element of TAU and the vector u(k)
!      in the kth row of MAT, such that the elements of z(k) are in MAT(k,KRANK+1:n).
!
!      On exit, the elements above the diagonal of the array section MAT(1:KRANK,1:KRANK)
!      contain the corresponding elements of the triangular matrix T11. The elements of
!      the diagonal of T11 are stored in the array section DIAGR(1:KRANK). The last part
!      of DIAGR is set to zero. In other words, T11 overwrites R11 and Z overwrites R12
!      on exit.
!   
!      The minimum 2-norm solution is then
!
!                       X = [ P * Z' ](:,:KRANK) * [ inv(T11) * Q1' * B ]
!
!      where inv(T11) is the inverse of T11, Z' is the transpose of Z and Q1 consists
!      of the first KRANK columns of Q.
!
!
!   3) If MIN_NORM is absent or has the value false, a solution is computed as 
!
!                       X = P(:,:KRANK) * [ inv(R11) * Q1' * B ]
!
!      where inv(R11) is the inverse of R11 and Q1 consists of the first KRANK columns of Q.
!      In this case, if the j-th column of MAT is omitted from the basis, X[j] is set to zero
!      and R is not destroyed in MAT.
!
!
!   4) In both cases:
!
!      - The 2-norm of the residual vector for the solution in the
!        j-th column of X is given in RNORM[j] if argument RNORM is present.
!
!      - If COMP_RESID=true, The residual matrix B - MAT * X overwrites B on exit.
! 
!
!   5) On input, if KRANK is present and KRANK=k, the first k columns of MAT are to
!      be forced into the basis. Pivoting is performed on the last n-k columns of MAT.
!
!      When KRANK is present and KRANK>=min(m,n) is used, pivoting is not performed. This
!      is appropriate when MAT is known to be full rank.
!   
!      If KRANK is absent or is present with KRANK<=0, pivoting is done on all columns of MAT.
!
!      On output, if KRANK is present, it contains the rank of MAT, i.e., the order 
!      of the submatrix R11. This is the same as the order of the submatrix T11
!      in the complete orthogonal factorization of MAT.
! 
!
!   6) TOL is an optional argument such that 0<=TOL<1. If TOL is not specified, 
!      or is outside [0,1[, the calculations to determine the condition number of MAT 
!      are not performed and crude tests on R(j,j) are performed to determine the numerical
!      rank of MAT. If TOL is present and is in [0,1[, the calculations to determine the  
!      condition number are performed, the effective rank of MAT is determined and the
!      reciprocal of the condition number is returned in TOL. 
! 
! 
!   7) If it is possible that MAT may not be full rank (i.e., certain columns of MAT are
!      linear combinations of other columns), then the linearly dependent columns can
!      usually be eliminated by using KRANK=0  and TOL=relative precision of the elements
!      in MAT and B. If each element is correct to, say, 5 digits then TOL=0.00001 should
!      be used. Also, it may be helpful to scale the columns of MAT so that all elements 
!      are about the same order of magnitude.
!
!   For further details on linear least square problems and algorithms to solve them, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!          Matrix Computations.
!          3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (3) Hansen, P.C., Pereyra, V., and Scherer, G., 2012:
!           Least Squares Data Fitting with Applications.
!           Johns Hopkins University Press, 328 pp.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,     only : machhuge
    use Char_Constants,    only : llsq_error2, allocate_error
    use Utilities,         only : merror, assert_eq, assert, arth, norm
    use QR_Procedures,     only : qrfac, apply_q_qr
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout),           dimension(:,:) :: mat, b
    real(stnd),   intent(out),             dimension(:,:) :: x
    real(stnd),   intent(inout), optional                 :: tol
    real(stnd),   intent(out),   optional , dimension(:)  :: rnorm, diagr, beta, tau
!
    integer(i4b), intent(inout), optional :: krank
    integer(i4b), intent(out),   optional :: ip(:)
!
    logical(lgl), intent(in), optional  :: comp_resid, min_norm
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(mat,dim=2))                        :: tau_tmp
    real(stnd), dimension(size(b,dim=2))                          :: tmp
    real(stnd), dimension(min(size(mat,dim=1),size(mat,dim=2)))   :: diagr_tmp, beta_tmp
    real(stnd), dimension(:,:), allocatable                       :: hmat, xt
!#ifdef _BLAS
!    real(stnd), dimension(min(size(mat,dim=2))                    :: h1_tmp
!#endif
!
    integer(i4b) :: m, n, ndiag, nb, i, k, kfix, kp1, j, l
    integer(i4b), dimension(:), allocatable :: ip_tmp
    integer                                 :: iok
!
    logical(lgl) :: comp_min_norm, test_comp_resid
#ifndef _BLAS
!
    real(stnd)   :: cl
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc ='llsq_qr_solve2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(b,1),i4b)   ,   &
                    name_proc )
!
    n =  assert_eq( int(size(mat,2),i4b) ,   &
                    int(size(x,1),i4b)   ,   &
                    name_proc )
!
    ndiag = min( n, m )
!
    nb =  assert_eq( int(size(x,2),i4b)   ,   &
                     int(size(b,2),i4b)   ,   &
                     name_proc )
!
    if ( present(rnorm) )                                      &
        call assert( logical(nb==int(size(rnorm),i4b),lgl) ,   &
                     name_proc )
!
    if ( present(diagr) )                                      &
        call assert( logical(size(diagr)==ndiag,lgl),          &
                     name_proc )
!
    if ( present(beta)  )                                      &
        call assert( logical(size(beta)==ndiag,lgl),           &
                     name_proc )
!
    if ( present(ip) )                                         &
        call assert( logical(size(ip)==n,lgl),                 &
                     name_proc )
!
    if ( present(tau) ) then
        call assert( logical(size(tau)==ndiag,lgl), name_proc )
        tau(:ndiag) = zero
    end if
!
    if ( present(min_norm) ) then
        comp_min_norm = min_norm
    else
        comp_min_norm = false
    end if
!
    if ( present(comp_resid) ) then
        test_comp_resid = comp_resid
    else
        test_comp_resid = false
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( min(nb,ndiag)<=0_i4b ) then
!
        call merror( name_proc//llsq_error2 )
!
    else if ( all(mat(:,:)==zero) ) then
!
        x(:n,:nb) = zero
!
        if ( present(rnorm) ) then
            rnorm = norm( b, dim=2_i4b )
        end if
!
        if ( present(krank) ) krank = 0_i4b
        if ( present(tol) )   tol = machhuge
        if ( present(diagr) ) diagr(:ndiag) = zero
        if ( present(beta) )  beta(:ndiag) = zero
        if ( present(ip) )    ip(:n) = arth( 1_i4b, 1_i4b, n )
!
        return
!
    end if
!    
!   COMPUTE QR FACTORIZATION WITH COLUMN PIVOTING OF mat .
!    
    if ( present(krank) ) then
        kfix = max( krank, 0_i4b )
    else
        kfix = 0_i4b
    end if
!    
    k = n
!    
    if ( present(ip) ) then
!
        call qrfac( name_proc, mat, kfix, k, comp_min_norm, diagr_tmp, beta_tmp, tau_tmp,      &
                    tol=tol, ip=ip )
!
    else if ( kfix<ndiag ) then
!
!       ALLOCATE WORK VARIABLE FOR STORING PERMUTATIONS.
!
        allocate( ip_tmp(n), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
        call qrfac( name_proc, mat, kfix, k, comp_min_norm, diagr_tmp, beta_tmp, tau_tmp,      &
                    tol=tol, ip=ip_tmp )
!
    else
!
        call qrfac( name_proc, mat, kfix, k, comp_min_norm, diagr_tmp, beta_tmp, tau_tmp,      &
                    tol=tol )
!
    end if
!
!   APPLY THE HOUSEHOLDER TRANSFORMATIONS TO b .
!
    call apply_q_qr( mat(:m,:k), beta_tmp(:k), b(:m,:nb), left=true, trans=true )
!
    x(:k,:nb) = b(:k,:nb)
!
    kp1 = k + 1_i4b
!
!   SOLVE THE k BY k TRIANGULAR SYSTEMS.
!
#ifdef _BLAS
    do i = 1_i4b, k
!
!        h1_tmp(i) = mat(i,i)
        mat(i,i) = diagr_tmp(i)
!
    end do
!
    call trsm( 'L', 'U', 'N', 'N', k, nb, one, mat, m, x, n )
!
    do i = 1_i4b, k
!
!        mat(i,i) = h1_tmp(i)
        mat(i,i) = one
!
    end do
#else
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and. &
                  nb>=2_i4b .and. i>1_i4b
#endif
!    
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC,1), PRIVATE(i,l,cl)
!
    do l = 1_i4b, nb
!
        do i = k, 1_i4b, -1_i4b
!
            cl = x(i,l)/diagr_tmp(i)
            x(1_i4b:i-1_i4b,l) = x(1_i4b:i-1_i4b,l) - cl*mat(1_i4b:i-1_i4b,i)
            x(i,l) = cl
!
        end do
!
    end do
!    
!$OMP END PARALLEL DO
#endif
!
!   COMPLETE COMPUTATION OF SOLUTION VECTORS.
!
    x(kp1:n,:nb) = zero
!
!   COMPUTE MINIMAL NORM SOLUTION VECTORS IF REQUIRED.
!
    if ( k/=n .and. comp_min_norm ) then
!
        l = n - k
!
!       ALLOCATE WORK VARIABLES FOR STORING HOUSEHOLDER AND SOLUTION VECTORS.
!
        allocate( hmat(l,k), xt(nb,n), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       TRANSPOSE HOUSEHOLDER AND SOLUTION VECTORS FOR LATER USE.
!
        hmat(:l,:k) = transpose( mat(:k,kp1:n) )
!
        xt(:nb,:n) = transpose( x(:n,:nb) )
!
        do i = 1_i4b, k
!
            if ( tau_tmp(i)==zero ) cycle
!
            tmp(:nb)  = matmul( xt(:nb,kp1:n), hmat(:l,i) )
            tmp(:nb)  = ( xt(:nb,i) + tmp(:nb) )*tau_tmp(i)
            xt(:nb,i) = xt(:nb,i) + tmp(:nb)
!
            do j = kp1, n
!
                xt(:nb,j) = xt(:nb,j) + hmat(j-k,i)*tmp(:nb)
!
            end do
!
        end do
!
        x(:n,:nb) = transpose( xt(:nb,:n) )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( hmat, xt )
!
!       OUTPUT OPTIONAL ARGUMENT tau ON EXIT, IF REQUIRED.
!
        if ( present(tau) ) then
            tau(1_i4b:k) = tau_tmp(1_i4b:k)
        end if
!
    end if
!    
!   RE-ORDER THE SOLUTION VECTORS TO COMPENSATE FOR THE COLUMN INTERCHANGES.
!   DEALLOCATE ip_tmp ON EXIT, IF NECESSARY.
!    
    if ( kfix<ndiag ) then
!
        if ( present(ip) ) then
!
            do i = 1_i4b, nb
!
                tau_tmp(ip) = x(:n,i)
                x(:n,i)     = tau_tmp(:n)
!
            end do
!
        else
!
            do i = 1_i4b, nb
!
                tau_tmp(ip_tmp) = x(:n,i)
                x(:n,i)         = tau_tmp(:n)
!
            end do
!
!           DEALLOCATE WORK VARIABLE.
!
            deallocate( ip_tmp )
!
        end if
!
    end if
!
!   OUTPUT OPTIONAL ARGUMENTS, IF REQUIRED.
!
    if ( present(rnorm) )  then
        rnorm = norm( b(kp1:m,:nb) , dim=2_i4b )
    end if
!
    if ( present(krank) )  krank = k
    if ( present(diagr) )  diagr(:ndiag) = diagr_tmp(:ndiag)
    if ( present(beta) )   beta(:ndiag)  = beta_tmp(:ndiag)
!
!   COMPUTE b - mat*x IF REQUIRED.
!
    if ( test_comp_resid ) then
!
        b(1_i4b:k,:nb) = zero
!
        call apply_q_qr( mat(:m,:k), beta_tmp(:k), b(:m,:nb), left=true, trans=false )
!
    end if
!
!
! END OF SUBROUTINE llsq_qr_solve2_rm_rm
! ______________________________________
!
    end subroutine llsq_qr_solve2_rm_rm
!
! =========================================================================================
!
    subroutine llsq_qr_solve2_rv_rv( vec, b, x, rnorm, comp_resid, diagr, beta )
!
! Purpose
! _______
!
!   LLSQ_QR_SOLVE2 computes a solution to a real linear least squares problem:
!
!                       Minimize 2-norm|| B - VEC * X ||
!
!   VEC is an m vector, B is a m right hand side vector and X is a real scalar. 
!
!   VEC and B are overwritten with information generated by LLSQ_QR_SOLVE2.
!
!
! Arguments
! _________
!
!   VEC          (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the m-elements coefficient vector VEC.
!
!                On exit, VEC contains the vector v of the 
!                Householder reflector H.
!
!                See Further Details.
!
!   B            (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the m-elements right hand side vector B.
!
!                On exit, if COMP_RESID is present and is equal true,
!                the residual vector B - VEC * X overwrites B on output.
!
!                The shape of B must verify:
!
!                - size( B ) = size( VEC ) = m .
!
!   X            (OUTPUT) real(stnd)
!                On exit, the real solution X.
!
!   RNORM        (OUTPUT, OPTIONAL) real(stnd)
!                On exit, the 2-norm of the residual vector for 
!                the solution scalar X.
!
!   COMP_RESID   (INPUT, OPTIONAL) logical(lgl)
!                On entry, if COMP_RESID is present and is equal true,
!                the residual vector B - VEC * X overwrites B on exit.
!
!   DIAGR        (OUTPUT, OPTIONAL) real(stnd) 
!                On exit, the scalar DIAGR. 
!
!                See Further Details.
!
!   BETA         (OUTPUT, OPTIONAL) real(stnd)
!                On exit, the scalar factor BETA of the elementary reflector defining H.
!
!                See Further Details.
!
!
!
! Further Details
! _______________
!
!   1) The routine first generates a real elementary reflector H of order m, such that
!
!           H * VEC = D , with  H' * H = I and D' = (  DIAGR  0  )
!
!      where DIAGR is scalar. H is represented in the form

!           H = I + BETA * ( v * v' ) ,
!                      
!      where BETA is a real scalar and v is a real m-element vector.
!
!
!   2) The solution X is then computed as
!
!                       X = [ H * B ](1) / DIAGR
!
!
!   3) The 2-norm of the residual vector for the solution X can be obtained through
!      the optional argument RNORM as
!
!                       2-norm|| [ H * B ](2:) ||
!
!
!   4) If COMP_RESID=true, The residual vector B - VEC * X overwrites B on exit.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : llsq_error3
    use Utilities,       only : merror, assert_eq, norm
    use Hous_Procedures, only : h1
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:) :: vec, b
    real(stnd),   intent(out)                 :: x 
    real(stnd),   intent(out),   optional     :: rnorm, diagr, beta
!
    logical(lgl), intent(in), optional  :: comp_resid
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)    :: beta_tmp, diagr_tmp
!
    integer(i4b)  ::  n           
!
    logical(lgl)  :: comp_resid2
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc ='llsq_qr_solve2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(vec),i4b)     ,   &
                    int(size(b),i4b)       ,   &
                    name_proc )
!
    comp_resid2 = false
!
    if ( present(comp_resid) ) then
!
        comp_resid2 = comp_resid
!
    end if
!
    if ( n<=0_i4b ) then
!
        call merror( name_proc//llsq_error3 )
!
    else
!
        call h1( vec, diagr_tmp, beta_tmp )
!
        if ( beta_tmp==zero ) then
!
!           QUICK RETURN IF POSSIBLE.
!
            if ( diagr_tmp==zero ) then
!
                x = zero
!
                if ( present(rnorm) ) then
                    rnorm = norm( b )
                end if
!
            else
!
                x = b(1_i4b)/diagr_tmp
!
                if ( present(rnorm) ) then
                    rnorm = norm( b(2_i4b:n) )
                end if
!
                if ( comp_resid2 ) then
!
                    b(1_i4b) = zero
!
                end if
!
            end if
!
        else if ( present(rnorm) .or. comp_resid2 ) then
!    
!           APPLY THE HOUSEHOLDER TRANSFORMATION TO vec AND b .
!    
            b(:n) = b(:n) + vec(:n)*( dot_product(vec(:n),b(:n))*beta_tmp )
!    
!           SOLVE THE 1 BY 1 TRIANGULAR SYSTEM.
!    
            x = b(1_i4b)/diagr_tmp
!    
!           COMPUTE THE NORM OF THE RESIDUAL VECTOR IF REQUIRED.
!    
            if ( present(rnorm) )  rnorm = norm( b(2_i4b:n) )
!    
!           COMPUTE b - vec*x IF REQUIRED.
!    
            if ( comp_resid2 ) then
!
                b(1_i4b) = zero
                b(:n) = b(:n) + ( vec(:n)*( dot_product(vec(2_i4b:n),b(2_i4b:n))*beta_tmp ) )
!
            end if
!
        else
!    
!           APPLY THE HOUSEHOLDER TRANSFORMATION TO vec AND b(1) .
!    
            x = b(1_i4b) + vec(1_i4b)*( dot_product(vec(:n),b(:n))*beta_tmp )
!    
!           SOLVE THE 1 BY 1 TRIANGULAR SYSTEM.
!    
            x = x/diagr_tmp
!
        end if
!
!       OUTPUT OPTIONAL ARGUMENTS ON EXIT, IF REQUIRED.
!
        if ( present(diagr) ) diagr = diagr_tmp
        if ( present(beta) )  beta = beta_tmp
!
    end if
!
!
! END OF SUBROUTINE llsq_qr_solve2_rv_rv
! ______________________________________
!
    end subroutine llsq_qr_solve2_rv_rv
!
! =========================================================================================
!
    subroutine llsq_qr_solve2_rv_rm( vec, b, x, rnorm, comp_resid, diagr, beta )
!
! Purpose
! _______
!
!   LLSQ_QR_SOLVE2 computes solutions to real linear least squares problems 
!   of the form:
!
!                       Minimize 2-norm|| B - VEC * X ||
!
!   here VEC is an m vector and several right hand side vectors b and solution
!   scalars x can be handled in a single call; they are stored as the columns
!   of the m-by-nb right hand side matrix B and the nb-elements solution vector X, respectively.
!
!   VEC and B are overwritten with information generated by LLSQ_QR_SOLVE2.
!
!
! Arguments
! _________
!
!   VEC          (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the m-elements coefficient vector VEC.
!
!                On exit, VEC contains the vector v of the 
!                Householder reflector H.
!
!                See Further Details.
!
!   B            (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the m-by-nb right hand side matrix B.
!
!                On exit, if COMP_RESID is present and is equal true,
!                the residual matrix B - VEC * X overwrites B on output.
!
!                The shape of B must verify: 
!
!                - size( B, 1 ) = size( VEC ) = m
!                - size( B, 2 ) = size( X )   = nb .
!
!   X            (OUTPUT) real(stnd), dimension(:)
!                On exit, the nb-elements solution vector X.
!
!                The shape of X must verify:
!
!                - size( X ) = size( B, 2 ) = nb .
!
!   RNORM        (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                On exit, the 2-norms of the residual vectors for 
!                the solutions stored in the vector X.
!
!                The size of RNORM must verify: 
!
!                - size( RNORM ) = size( X ) = size( B, 2 ) = nb .
!
!   COMP_RESID   (INPUT, OPTIONAL) logical(lgl)
!                On entry, if COMP_RESID is present and is equal true,
!                the residual matrix  B - VEC * X  overwrites B on exit.
!
!   DIAGR        (OUTPUT, OPTIONAL) real(stnd) 
!                On exit, the scalar DIAGR. 
!
!                See Further Details.
!
!   BETA         (OUTPUT, OPTIONAL) real(stnd)
!                On exit, the scalar factor BETA of the elementary reflector defining H.
!
!                See Further Details.
!
!
!
! Further Details
! _______________
!
!   1) The routine first generates a real elementary reflector H of order m, such that
!
!           H * VEC = D , with  H' * H = I and D' = (  DIAGR  0  )
!
!      where DIAGR is scalar. H is represented in the form

!           H = I + BETA * ( v * v' ) ,
!                      
!      where BETA is a real scalar and v is a real m-element vector.
!
!
!   2) The solution vector X is then computed as
!
!                       X(:) = [ H * B ](1,:) / DIAGR
!
!
!   3) The 2-norm of the residual vector for the solution X[j] is given 
!      in RNORM[j] if argument RNORM is present.
!
!
!   4) If COMP_RESID=true, The residual matrix  B - VEC * X  overwrites B on exit.
!
!
! _________________________________________________________________________________________
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : llsq_error4
    use Utilities,       only : merror, assert_eq, assert, norm
    use Hous_Procedures, only : h1
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:)   :: vec
    real(stnd),   intent(inout), dimension(:,:) :: b
    real(stnd),   intent(out),   dimension(:)   :: x
    real(stnd),   intent(out),   dimension(:), optional :: rnorm
    real(stnd),   intent(out),                 optional :: diagr, beta
!
    logical(lgl), intent(in), optional  :: comp_resid
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)  :: beta2, diagr2
    real(stnd), dimension(size(x)) :: tmp
!
    integer(i4b)  :: nb, m, j
!
    logical(lgl)  :: comp_resid2, comp_rnorm
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc ='llsq_qr_solve2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    m =  assert_eq( int(size(vec),i4b)   ,   &
                    int(size(b,1),i4b)   ,   &
                    name_proc )
!
    nb =  assert_eq( int(size(b,2),i4b)     ,   &
                     int(size(x),i4b)       ,   &
                     int(size(rnorm),i4b)   ,   &
                     name_proc )
!
    comp_rnorm = present( rnorm )
!
    if ( comp_rnorm ) then
!
        call assert( logical(nb==int(size(rnorm),i4b),lgl), name_proc )
!
    end if
!
    comp_resid2 = false
!
    if ( present(comp_resid) ) then
!
        comp_resid2 = comp_resid
!
    end if
!
    if ( min(nb,m)<=0_i4b ) then
!
        call merror( name_proc//llsq_error4 )
!
    else
!
        call h1( vec(:m), diagr2, beta2 )
!
        if ( beta2==zero ) then
!
!           QUICK RETURN IF POSSIBLE.
!
            if ( diagr2==zero ) then
!
                x(:nb) = zero
!
                if ( comp_rnorm ) then
!
                    rnorm = norm( b, dim=2_i4b )
!
                end if
!
            else
!
                x(:nb) = b(1_i4b,:nb)/diagr2
!
                if ( comp_rnorm ) then
!
                    rnorm = norm( b(2_i4b:m,:nb), dim=2_i4b )
!
                end if
!
                if ( comp_resid2 ) then
!
                    b(1_i4b,:nb) = zero
!
                end if
!
            end if
!
        else
!    
!           APPLY THE HOUSEHOLDER TRANSFORMATION TO b .
!    
            tmp(:nb) = matmul( vec(:m), b(:m,:nb) )
            tmp(:nb) = tmp(:nb)*beta2
!
            if ( comp_rnorm .or. comp_resid2 ) then
!
                do j = 1_i4b, nb
!
#ifdef _BLAS
                    call axpy( m, tmp(j), vec(:m), 1_i4b, b(:m,j), 1_i4b )
#else
                    b(:m,j) = b(:m,j) + tmp(j)*vec(:m)
#endif
!    
!                   SOLVE THE 1 BY 1 TRIANGULAR SYSTEM.
!    
                    x(j) = b(1_i4b,j)/diagr2
!    
!                   COMPUTE THE NORM OF THE RESIDUAL VECTOR.
!    
                    if ( comp_rnorm )  then
!    
                        rnorm(j) = norm( b(2_i4b:m,j) )
!
                    end if
!    
!                   COMPUTE b - vec*x IF REQUIRED.
!    
                    if ( comp_resid2 ) then
!
                        b(1_i4b,j) = zero
                        tmp(j) = dot_product( vec(2_i4b:m), b(2_i4b:m,j) )
                        tmp(j) = tmp(j)*beta2
!
#ifdef _BLAS
                        call axpy( m, tmp(j), vec(:m), 1_i4b, b(:m,j), 1_i4b )
#else
                        b(:m,j) = b(:m,j) + tmp(j)*vec(:m)
#endif
!
                    end if
!
                end do
!
            else
!
                tmp(:nb) = b(1_i4b,:nb) + vec(1_i4b)*tmp(:nb)
!    
!               SOLVE THE 1 BY 1 TRIANGULAR SYSTEMS.
!    
                x(:nb) = tmp(:nb)/diagr2
!
            end if
!
        end if
!
!       OUTPUT OPTIONAL ARGUMENTS ON EXIT, IF REQUIRED.
!
        if ( present(diagr) ) diagr = diagr2
        if ( present(beta) )  beta  = beta2
!
    end if
!
!
! END OF SUBROUTINE llsq_qr_solve2_rv_rm
! ______________________________________
!
    end subroutine llsq_qr_solve2_rv_rm
!
! =========================================================================================
!
    subroutine qr_solve_rv( mat, diagr, beta, b, x, rnorm, comp_resid )
!
! Purpose
! _______
!
!   QR_SOLVE solves overdetermined or underdetermined real linear systems
!
!                               MAT * X = B          
!
!   with an m-by-n matrix MAT, using a QR factorization of MAT as computed by QR_CMP.
!   m>=n or n>m is permitted, but it is assumed that MAT has full rank. B is a m-elements
!   right hand side vector and X is a n-elements solution vector.
!   
!   It is assumed that QR_CMP has been used to compute the QR factorization of MAT 
!   before QR_SOLVE.
!
!
! Arguments
! _________
!
!   MAT          (INPUT) real(stnd), dimension(:,:)
!                On entry, the QR factorization of the real coefficient matrix
!                MAT as returned by QR_CMP.
!
!                The elements above the diagonal of the array contain the   
!                corresponding elements of R. The elements on and below the 
!                diagonal, with the array BETA, represent the orthogonal matrix
!                Q in the QR decomposition of MAT, as a product of elementary 
!                reflectors, as returned by QR_CMP.
!
!   DIAGR        (INPUT) real(stnd), dimension(:) 
!                On entry, the diagonal elements of the matrix R in the
!                QR decomposition of MAT.
!
!                The size of DIAGR must be min( size(MAT,1) , size(MAT,2) ).
!
!   BETA         (INPUT) real(stnd), dimension(:)
!                On entry, the scalars factors of the elementary reflectors
!                defining Q, as returned by QR_CMP.
!
!                The size of BETA must be min( size(MAT,1) , size(MAT,2) ),
!
!   B            (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the m-elements right hand side vector B.
!
!                On exit, if COMP_RESID is present and is equal true,
!                the residual vector B - MAT * X overwrites B.
!
!                The size of B must verify:
!
!                - size( B ) = size( MAT, 1 ) = m .
!
!   X            (OUTPUT) real(stnd), dimension(:)
!                On exit, the n-elements solution vector X.
!
!                The size of X must verify:
!
!                size( X ) = size( MAT, 2 ) = n .
!
!   RNORM        (OUTPUT, OPTIONAL) real(stnd)
!                On exit, the 2-norm of the residual vector for 
!                the solution vector X.
!
!   COMP_RESID   (INPUT, OPTIONAL) logical(lgl)
!                On entry, if COMP_RESID is present and is equal true,
!                the residual vector B - MAT * X overwrites B on exit.
!
!
! Further Details
! _______________
!
!   1) It is assumed that QR_CMP has been used to compute the QR 
!      factorization of MAT before calling QR_SOLVE.
!
!
!   2) If m>=n: the subroutine finds the least squares solution of an
!      overdetermined system, i.e., solves the least squares problem
!
!                       Minimize 2-norm|| B - MAT * X ||
!
!      If m<n:  the subroutine finds a solution of an underdetermined system
!
!                       MAT * X = B          
!
!
!   3) The 2-norm of the residual vector for the solution X can 
!      be obtained through the optional argument RNORM .
!
!
!   4) If COMP_RESID=true, The m-elements residual vector B - MAT * X overwrites B on exit.
! 
!
!   5) MAT, DIAGR, BETA are not modified by this routine and can be left in place
!      for successive calls with different right-hand side vectors B.
!
!   For further details on linear least square problems and algorithms to solve them, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!          Matrix Computations.
!          3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (3) Hansen, P.C., Pereyra, V., and Scherer, G., 2012:
!           Least Squares Data Fitting with Applications.
!           Johns Hopkins University Press, 328 pp.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert_eq, norm
    use QR_Procedures,   only : apply_q_qr
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in),  dimension(:) :: mat(:,:), diagr, beta
    real(stnd),   intent(inout)             :: b(:)
    real(stnd),   intent(out)               :: x(:)
    real(stnd),   intent(out),  optional    :: rnorm
!
    logical(lgl), intent(in), optional :: comp_resid
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: m, n, ndiag, i, kp1 
!
    logical(lgl) :: test_comp_resid
#ifndef _USE_NAGWARE
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='qr_solve'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(b),i4b)     ,   &
                    name_proc )
!
    n =  assert_eq( int(size(mat,2),i4b) ,   &
                    int(size(x),i4b)     ,   &
                    name_proc )
!
    ndiag =  assert_eq( min( n, m )    ,         &
                        int(size(diagr),i4b) ,   &
                        int(size(beta),i4b)  ,   &
                        name_proc )
!
    if ( present(comp_resid) ) then
        test_comp_resid = comp_resid
    else
        test_comp_resid = false
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( ndiag<=0_i4b ) then
!
        x(:n) = zero
!
        if ( present(rnorm) ) rnorm = norm( b )
!
        return
!
    end if
!
!   APPLY THE HOUSEHOLDER TRANSFORMATIONS TO b .
!
    call apply_q_qr( mat(:m,:ndiag), beta(:ndiag), b(:m), trans=true )
!
    x(:ndiag) = b(:ndiag)
!
    kp1 = ndiag + 1_i4b
!
!   COMPUTE THE NORM OF THE RESIDUAL VECTOR IF REQUIRED.
!
    if ( present(rnorm) ) then
        rnorm = norm( b(kp1:m) )
    end if
!
#ifndef _USE_NAGWARE
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and. &
                  test_comp_resid .and. i>1_i4b
#endif
!
!$OMP PARALLEL  IF( test_par ), PRIVATE(i), NUM_THREADS( 2 )
!
!$OMP SECTIONS
!    
!$OMP SECTION
#endif
!    
!   SOLVE THE ndiag BY ndiag TRIANGULAR SYSTEM.
!
    do i = ndiag, 1_i4b, -1_i4b
!
        x(i) = x(i)/diagr(i)
!
#ifdef _BLAS
        call axpy( i-1_i4b, -x(i), mat(1_i4b:i-1_i4b,i), 1_i4b, x(1_i4b:i-1_i4b), 1_i4b )
#else
        x(1_i4b:i-1_i4b) = x(1_i4b:i-1_i4b) - x(i)*mat(1_i4b:i-1_i4b,i)
#endif
!
    end do
!
!   COMPLETE COMPUTATION OF SOLUTION VECTOR.
!
    x(kp1:n) = zero
!
#ifndef _USE_NAGWARE
!$OMP SECTION
#endif
!
!   COMPUTE b - mat*x IF REQUIRED.
!
    if ( test_comp_resid ) then
!
        b(1_i4b:ndiag) = zero
!
        call apply_q_qr( mat(:m,:ndiag), beta(:ndiag), b(:m), trans=false )
!
    end if
!
#ifndef _USE_NAGWARE
!$OMP END SECTIONS
!
!$OMP END PARALLEL
#endif
!
!
!
! END OF SUBROUTINE qr_solve_rv
! _____________________________
!
    end subroutine qr_solve_rv
!
! =========================================================================================
!
    subroutine qr_solve_rm( mat, diagr, beta, b, x, rnorm, comp_resid )
!
! Purpose
! _______
!
!   QR_SOLVE solves overdetermined or underdetermined real linear systems
!   of the form:
!
!                               MAT * X = B          
!
!   with an m-by-n matrix MAT, using a QR factorization of MAT as computed by
!   QR_CMP. m>=n or n>m is permitted, but it is assumed that MAT has full rank.
!
!   Several right hand side vectors b and solution vectors x can be
!   handled in a single call; they are stored as the columns of the
!   m-by-nb right hand side matrix B and the n-by-nb solution
!   matrix X, respectively.
!   
!   It is assumed that QR_CMP has been used to compute the QR factorization 
!   of MAT before QR_SOLVE.
!
!
! Arguments
! _________
!
!   MAT          (INPUT) real(stnd), dimension(:,:)
!                On entry, the QR factorization of the real coefficient matrix
!                MAT as returned by QR_CMP.
!
!                The elements above the diagonal of the array contain the   
!                corresponding elements of R. The elements on and below the 
!                diagonal, with the array BETA, represent the orthogonal matrix
!                Q in the QR decomposition of MAT, as a product of elementary 
!                reflectors, as returned by QR_CMP.
!
!   DIAGR        (INPUT) real(stnd), dimension(:) 
!                On entry, the diagonal elements of the matrix R in the
!                QR decomposition of MAT.
!
!                The size of DIAGR must be min( size(MAT,1) , size(MAT,2) ).
!
!   BETA         (INPUT) real(stnd), dimension(:)
!                On entry, the scalars factors of the elementary reflectors
!                defining Q, as returned by QR_CMP.
!
!                The size of BETA must be min( size(MAT,1) , size(MAT,2) ),
!
!   B            (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the m-by-nb right hand side matrix B.
!
!                On exit, if COMP_RESID is present and is equal true,
!                the residual matrix B - MAT * X overwrites B on output.
!
!                The shape of B must verify: 
!
!                - size( B, 1 ) = size( MAT, 1 ) = m
!                - size( B, 2 ) = size( X, 2 )   = nb .
!
!   X            (OUTPUT) real(stnd), dimension(:,:)
!                On exit, the n-by-nb solution matrix X.
!
!                The shape of X must verify: 
!
!                - size( X, 1 ) = size( MAT, 2 ) = n
!                - size( X, 2 ) = size( B, 2 )   = nb .
!
!   RNORM        (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                On exit, the 2-norm of the residual vectors for 
!                the solutions stored columnwise in the matrix X.
!
!                The size of RNORM must verify: 
!
!                - size( RNORM ) = size( X, 2 ) = size( B, 2 )  = nb .
!
!   COMP_RESID   (INPUT, OPTIONAL) logical(lgl)
!                On entry, if COMP_RESID is present and is equal true,
!                the residual matrix B - MAT * X overwrites B on exit.
!
!
! Further Details
! _______________
!
!   1) It is assumed that QR_CMP has been used to compute the QR 
!      factorization of MAT before calling QR_SOLVE.
!
!
!   2) If m>=n: the subroutine finds the least squares solutions of 
!      overdetermined systems, i.e., solves least squares problems of the
!      form
!
!                       Minimize 2-norm|| B - MAT * X ||
!
!      If m<n:  the subroutine finds solutions of underdetermined systems
!      of the form
!
!                       MAT * X = B          
!
!      In both cases, several right hand side vectors b and solution vectors
!      x can be handled in a single call; they are stored as the columns of
!      the m-by-nb right hand side matrix B and the n-by-nb solution matrix X,
!      respectively.
!
!
!   3) The 2-norm of the residual vector for the solution in the 
!      j-th column of X is given in RNORM[j] if argument RNORM is present.
!
!
!   4) If COMP_RESID=true, The residual matrix B - MAT * X overwrites B on exit.
! 
!
!   5) MAT, DIAGR, BETA are not modified by this routine and can be left in
!      place for successive calls with different right-hand side matrices B.
!
!   For further details on linear least square problems and algorithms to solve them, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!          Matrix Computations.
!          3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (3) Hansen, P.C., Pereyra, V., and Scherer, G., 2012:
!           Least Squares Data Fitting with Applications.
!           Johns Hopkins University Press, 328 pp.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert_eq, assert, norm
    use QR_Procedures,   only : apply_q_qr
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in),    dimension(:)   :: diagr, beta
    real(stnd),   intent(in),    dimension(:,:) :: mat
    real(stnd),   intent(inout), dimension(:,:) :: b
    real(stnd),   intent(out),   dimension(:,:) :: x
    real(stnd),   intent(out), optional         :: rnorm(:) 
!
    logical(lgl), intent(in), optional  :: comp_resid
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: cl
!
    integer(i4b) :: m, n, ndiag, nb, i, kp1, l
!
    logical(lgl) :: test_comp_resid
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc ='qr_solve'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(b,1),i4b)   ,   &
                    name_proc )
!
    n =  assert_eq( int(size(mat,2),i4b) ,   &
                    int(size(x,1),i4b)   ,   &
                    name_proc )
!
    ndiag =  assert_eq( min( n, m )    ,         &
                        int(size(diagr),i4b) ,   &
                        int(size(beta),i4b)  ,   &
                        name_proc )
!
    nb =  assert_eq( int(size(x,2),i4b)   ,   &
                     int(size(b,2),i4b)   ,   &
                     name_proc )
!
    if ( present(rnorm) )  call assert( logical(size(rnorm)==nb,lgl), name_proc )
!
    if ( present(comp_resid) ) then
        test_comp_resid = comp_resid
    else
        test_comp_resid = false
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( min(nb,ndiag)<=0_i4b ) then
!
        x(:n,:nb) = zero
!
        if ( present(rnorm) ) then
            rnorm(:nb) = norm( b, dim=2_i4b )
        end if
!
        return
!
    end if
!
!   APPLY THE HOUSEHOLDER TRANSFORMATIONS TO b .
!
    call apply_q_qr( mat(:m,:ndiag), beta(:ndiag), b(:m,:nb), left=true, trans=true )
!
    x(:ndiag,:nb) = b(:ndiag,:nb)
!    
    kp1 = ndiag + 1_i4b
!    
!   SOLVE THE ndiag BY ndiag TRIANGULAR SYSTEMS.
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and. &
                  nb>=2_i4b .and. i>1_i4b
#endif
!    
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC,1), PRIVATE(i,l,cl)
!
    do l = 1_i4b, nb

        do i = ndiag, 1_i4b, -1_i4b
!
            cl = x(i,l)/diagr(i)
!
!
#ifdef _BLAS
            call axpy( i-1_i4b, -cl, mat(1_i4b:i-1_i4b,i), 1_i4b, x(1_i4b:i-1_i4b,l), 1_i4b )
#else
            x(1_i4b:i-1_i4b,l) = x(1_i4b:i-1_i4b,l) - cl*mat(1_i4b:i-1_i4b,i)
#endif
!
            x(i,l) = cl
!
        end do
!
    end do
!    
!$OMP END PARALLEL DO
!
!   COMPLETE COMPUTATION OF SOLUTION VECTORS.
!
    x(kp1:n,:nb) = zero
!    
!   COMPUTE THE NORM OF THE RESIDUAL VECTORS IF REQUIRED.
!    
    if ( present(rnorm) ) then
        rnorm(:nb) = norm( b(kp1:m,:nb), dim=2_i4b )
    end if
!
!   COMPUTE b - mat*x IF REQUIRED.
!
    if ( test_comp_resid ) then
!
        b(1_i4b:ndiag,:nb) = zero
!
        call apply_q_qr( mat(:m,:ndiag), beta(:ndiag), b(:m,:nb), left=true, trans=false )
!
    end if
!
!
! END OF SUBROUTINE qr_solve_rm
! _____________________________
!
    end subroutine qr_solve_rm
!
! =========================================================================================
!
    subroutine qr_solve2_rv( mat, diagr, beta, ip, krank, b, x, rnorm, comp_resid, tau )
!
! Purpose
! _______
!
!   QR_SOLVE2 solves overdetermined or underdetermined real linear systems
!
!                               MAT * X = B          
!
!   with an m-by-n matrix MAT, using a QR or (complete) orthogonal factorization of
!   MAT as computed by QR_CMP2. m>=n or n>m is permitted and MAT may be rank-deficient.
!   B is a m-elements right hand side vector and X is a n-elements solution vector.
!   
!   It is assumed that QR_CMP2 has been used to compute the (complete) orthogonal 
!   factorization of MAT before QR_SOLVE2.
!
!
! Arguments
! _________
!
!   MAT          (INPUT) real(stnd), dimension(:,:)
!                On entry, details of the QR or (complete) orthogonal factorization
!                of the real coefficient matrix MAT as returned by QR_CMP2.
!
!                See description of QR_CMP2 subroutine for further details.
!
!   DIAGR        (INPUT) real(stnd), dimension(:) 
!                On entry, the diagonal elements of the matrix R in the QR
!                decomposition with column pivoting of MAT if TAU is absent or
!                the diagonal elements of the matrix T11 in the complete orthogonal 
!                factorization of MAT if TAU is present, as computed by QR_CMP2. If a
!                complete orthogonal factorization has been computed, the
!                diagonal elements of T11 are stored in DIAGR(1:KRANK).
!
!                See description of QR_CMP2 subroutine for further details.
!
!                The size of DIAGR must be min( size(MAT,1) , size(MAT,2) ).
!
!   BETA         (INPUT) real(stnd), dimension(:)
!                On entry, the scalars factors of the elementary reflectors
!                defining Q in the QR or orthogonal factorization of MAT,
!                as returned by QR_CMP2.
!
!                See description of QR_CMP2 subroutine for further details.
!
!                The size of BETA must be min( size(MAT,1) , size(MAT,2) ),
!
!   IP           (INPUT) integer(i4b), dimension(:) 
!                On entry, the permutation P in the QR or (complete) orthogonal
!                factorization of MAT, as returned by QR_CMP2. 
!
!                See description of QR_CMP2 subroutine for further details.
!
!                The size of IP must be size( MAT, 2 ) = n.
!
!   KRANK        (INPUT) integer(i4b)
!                On entry, KRANK contains the effective rank of MAT,
!                as returned by QR_CMP2.
!
!                See description of QR_CMP2 subroutine for further details.
!
!   B            (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the m-elements right hand side vector B .
!
!                On exit, if COMP_RESID is present and is equal to true,
!                the residual vector B - MAT * X overwrites B .
!
!                The size of B must verify:
!
!                - size( B ) = size( MAT, 1 ) = m .
!
!   X            (OUTPUT) real(stnd), dimension(:)
!                On exit, the n-elements solution vector X.
!
!                The size of X must verify:
!
!                - size( X ) = size( MAT, 2 ) = n .
!
!   RNORM        (OUTPUT, OPTIONAL) real(stnd)
!                On exit, the 2-norm of the residual vector for 
!                the solution vector X.
!
!   COMP_RESID   (INPUT, OPTIONAL) logical(lgl)
!                On entry, if COMP_RESID is present and is equal to true,
!                the residual vector B - MAT * X overwrites B on exit.
!
!   TAU          (INPUT, OPTIONAL) real(stnd), dimension(:)
!                On entry, if TAU is present, a complete orthogonal factorization of
!                MAT has been computed by QR_CMP2 and TAU contains the scalars 
!                factors of the elementary reflectors defining Z in this decomposition.
!                Otherwise, only a QR factorization with column pivoting of MAT has
!                been computed  by QR_CMP2.
!
!                See description of QR_CMP2 subroutine for further details.
!
!                The size of TAU must be min( size(MAT,1) , size(MAT,2) ).
!
!
! Further Details
! _______________
!
!   1) It is assumed that QR_CMP2 has been used to compute the (complete) 
!      orthogonal factorization (if TAU is present) or the QR factorization with
!      column pivoting (if TAU is absent) of MAT before calling QR_SOLVE2.
!
!
!   2) If m>=n: the subroutine finds the least squares solution of an
!      overdetermined system, i.e., solves the least squares problem
!
!                       Minimize 2-norm|| B - MAT * X ||
!
!      If m<n:  the subroutine finds a solution of an underdetermined system
!
!                       MAT * X = B          
!
!      In both cases, the minimun 2-norm solution is computed if TAU is present.
!      Otherwise, a solution is computed such that if the j-th column of MAT
!      is omitted from the basis, X[j] is set to zero.
!
!
!   3) The 2-norm of the residual vector for the solution X can 
!      be obtained through the optional argument RNORM .
!
!
!   4) If COMP_RESID=true, The m-elements residual vector B - MAT * X overwrites B on exit.
! 
!
!   5) MAT, DIAGR, BETA, IP, KRANK and TAU are not modified by this routine and can
!      be left in place for successive calls with different right-hand side vectors B.
!
!   For further details on linear least square problems and algorithms to solve them, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!          Matrix Computations.
!          3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (3) Hansen, P.C., Pereyra, V., and Scherer, G., 2012:
!           Least Squares Data Fitting with Applications.
!           Johns Hopkins University Press, 328 pp.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert_eq, assert, merror, norm
    use QR_Procedures,   only : apply_q_qr
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in),    dimension(:)            :: diagr, beta
    real(stnd),   intent(in),    dimension(:,:)          :: mat
    real(stnd),   intent(in),    dimension(:),  optional :: tau
    real(stnd),   intent(inout), dimension(:)            :: b
    real(stnd),   intent(out),   dimension(:)            :: x
    real(stnd),   intent(out),                  optional :: rnorm
!
    integer(i4b), intent(in)               :: krank
    integer(i4b), intent(in), dimension(:) :: ip
!
    logical(lgl), intent(in), optional :: comp_resid
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: cl
    real(stnd), dimension(size(x))          :: x_tmp
    real(stnd), dimension(:,:), allocatable :: hmat
!
    integer      :: iok
    integer(i4b) :: m, n, ndiag, i, kp1, l
!
    logical(lgl) :: test_comp_resid, comp_min_norm
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc ='qr_solve2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(b),i4b)     ,   &
                    name_proc )
!
    n =  assert_eq( int(size(mat,2),i4b) ,   &
                    int(size(x),i4b)     ,   &
                    int(size(ip),i4b)    ,   &
                    name_proc )
!
    ndiag =  assert_eq( min( n, m )    ,         &
                        int(size(diagr),i4b) ,   &
                        int(size(beta),i4b)  ,   &
                        name_proc )
!
    comp_min_norm = present( tau )
!
    if ( comp_min_norm ) then
        call assert( logical(size(tau)==ndiag,lgl), name_proc )
    end if
!
    if ( present(comp_resid) ) then
        test_comp_resid = comp_resid
    else
        test_comp_resid = false
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( krank<=0_i4b ) then
!
        x(:n) = zero
!
        if ( present(rnorm) ) then
            rnorm = norm( b )
        end if
!
        return
!
    end if
!
!   APPLY THE HOUSEHOLDER TRANSFORMATIONS TO b .
!
    call apply_q_qr( mat(:m,:krank), beta(:krank), b(:m), trans=true )
!
    x_tmp(:krank) = b(:krank)
!    
    kp1 = krank + 1_i4b
!    
!   COMPUTE THE NORM OF THE RESIDUAL VECTOR IF REQUIRED.
!    
    if ( present(rnorm) ) then
        rnorm = norm( b(kp1:m) )
    end if
!    
!   SOLVE THE krank BY krank TRIANGULAR SYSTEM.
!
    do i = krank, 1_i4b, -1_i4b
!
        x_tmp(i) = x_tmp(i)/diagr(i)
#ifdef _BLAS
        call axpy( i-1_i4b, -x_tmp(i), mat(1_i4b:i-1_i4b,i), 1_i4b, x_tmp(1_i4b:i-1_i4b), 1_i4b )
#else
        x_tmp(1_i4b:i-1_i4b) = x_tmp(1_i4b:i-1_i4b) - x_tmp(i)*mat(1_i4b:i-1_i4b,i)
#endif
!
    end do
!
!   COMPLETE COMPUTATION OF SOLUTION VECTOR.
!
    x_tmp(kp1:n) = zero
!
!   COMPUTE MINIMAL NORM SOLUTION VECTOR IF REQUIRED.
!
    if ( krank/=n .and. comp_min_norm ) then
!
        l = n - krank
!
!       ALLOCATE WORK VARIABLES FOR STORING HOUSEHOLDER VECTORS.
!
        allocate( hmat(l,krank), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       TRANSPOSE HOUSEHOLDER VECTORS FOR LATER USE.
!
        hmat(:l,:krank) = transpose( mat(:krank,kp1:n) )
!
        do i = 1_i4b, krank
!
            if ( tau(i)==zero ) cycle
!
            cl           = ( x_tmp(i) + dot_product(hmat(:l,i),x_tmp(kp1:n)) )*tau(i)
            x_tmp(i)     = x_tmp(i) + cl
            x_tmp(kp1:n) = x_tmp(kp1:n) + hmat(:l,i)*cl
!
        end do
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( hmat )
!
    end if
!    
!   RE-ORDER THE SOLUTION VECTOR TO COMPENSATE FOR THE COLUMN INTERCHANGES.
!    
    x(ip) = x_tmp(:n)
!
!   COMPUTE b - mat*x IF REQUIRED.
!
    if ( test_comp_resid ) then
!
        b(1_i4b:krank) = zero
!
        call apply_q_qr( mat(:m,:krank), beta(:krank), b(:m), trans=false )
!
    end if
!
!
! END OF SUBROUTINE qr_solve2_rv
! ______________________________
!
    end subroutine qr_solve2_rv
!
! =========================================================================================
!
    subroutine qr_solve2_rm( mat, diagr, beta, ip, krank, b, x, rnorm, comp_resid, tau )
!
! Purpose
! _______
!
!   QR_SOLVE2 solves overdetermined or underdetermined real linear systems
!   of the form:
!
!                               MAT * X = B          
!
!   with an m-by-n matrix MAT, using a QR or (complete) orthogonal factorization of MAT as 
!   computed by QR_CMP2. m>=n or n>m is permitted and MAT may be rank-deficient.
!
!   Several right hand side vectors b and solution vectors x can be
!   handled in a single call; they are stored as the columns of the
!   m-by-nb right hand side matrix B and the n-by-nb solution
!   matrix X, respectively.
!   
!   It is assumed that QR_CMP2 has been used to compute the QR or (complete) orthogonal 
!   factorization of MAT before QR_SOLVE2.
!
!
! Arguments
! _________
!
!   MAT          (INPUT) real(stnd), dimension(:,:)
!                On entry, details of the QR or (complete) orthogonal factorization
!                of the real coefficient matrix MAT as returned by QR_CMP2.
!
!                See description of QR_CMP2 subroutine for further details.
!
!   DIAGR        (INPUT) real(stnd), dimension(:) 
!                On entry, the diagonal elements of the matrix R in the QR
!                decomposition with column pivoting of MAT if TAU is absent or
!                the diagonal elements of the matrix T11 in the complete orthogonal 
!                factorization of MAT if TAU is present, as computed by QR_CMP2. If a
!                complete orthogonal factorization has been computed, the
!                diagonal elements of T11 are stored in DIAGR(1:KRANK).
!
!                See description of QR_CMP2 subroutine for further details.
!
!                The size of DIAGR must be min( size(MAT,1) , size(MAT,2) ).
!
!   BETA         (INPUT) real(stnd), dimension(:)
!                On entry, the scalars factors of the elementary reflectors
!                defining Q in the QR or orthogonal factorization of MAT,
!                as returned by QR_CMP2.
!
!                See description of QR_CMP2 subroutine for further details.
!
!                The size of BETA must be min( size(MAT,1) , size(MAT,2) ),
!
!   IP           (INPUT) integer(i4b), dimension(:) 
!                On entry, the permutation P in the QR or (complete) orthogonal
!                factorization of MAT, as returned by QR_CMP2. 
!
!                See description of QR_CMP2 subroutine for further details.
!
!                The size of IP must be size(MAT,2) = n.
!
!   KRANK        (INPUT) integer(i4b)
!                On entry, KRANK contains the effective rank of MAT,
!                as returned by QR_CMP2.
!
!                See description of QR_CMP2 subroutine for further details.
!
!   B            (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the m-by-nb right hand side matrix B .
!
!                On exit, if COMP_RESID is present and is equal to true,
!                the residual matrix B - MAT * X overwrites B .
!
!                The shape of B must verify: 
!
!                - size( B, 1 ) = size( MAT, 1 ) = m
!                - size( B, 2 ) = size( X, 2 )   = nb .
!
!   X            (OUTPUT) real(stnd), dimension(:,:)
!                On exit, the n-by-nb solution matrix X.
!
!                The shape of X must verify: 
!
!                - size( X, 1 ) = size( MAT, 2 ) = n
!                - size( X, 2 ) = size( B, 2 )   = nb .
!
!   RNORM        (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                On exit, the 2-norm of the residual vectors for 
!                the solutions stored columnwise in the matrix X.
!
!                The size of RNORM must verify: 
!
!                - size( RNORM ) = size( X, 2 ) = size( B, 2 )  = nb .
!
!   COMP_RESID   (INPUT, OPTIONAL) logical(lgl)
!                On entry, if COMP_RESID is present and is equal to true,
!                the residual matrix B - MAT * X overwrites B on exit.
!
!   TAU          (INPUT, OPTIONAL) real(stnd), dimension(:)
!                On entry, if TAU is present, a complete orthogonal factorization of
!                MAT has been computed by QR_CMP2 and TAU contains the scalars 
!                factors of the elementary reflectors defining Z in this decomposition.
!                Otherwise, only a QR factorization with column pivoting of MAT has
!                been computed  by QR_CMP2.
!
!                See description of QR_CMP2 subroutine further details.
!
!                The size of TAU must be min( size(MAT,1) , size(MAT,2) ).
!
!
! Further Details
! _______________
!
!   1) It is assumed that QR_CMP2 has been used to compute the (complete) 
!      orthogonal factorization (if TAU is present) or the QR factorization with
!      column pivoting (if TAU is absent) of MAT before calling QR_SOLVE2.
!
!
!   2) If m>=n: the subroutine finds the least squares solutions of 
!      overdetermined systems, i.e., solves least squares problems of the
!      form
!
!                       Minimize 2-norm|| B - MAT * X ||
!
!      If m<n:  the subroutine finds solutions of underdetermined systems
!      of the form
!
!                       MAT * X = B          
!
!      In both cases, several right hand side vectors b and solution vectors
!      x can be handled in a single call; they are stored as the columns of
!      the m-by-nb right hand side matrix B and the n-by-nb solution matrix X,
!      respectively.
!
!      In both cases, the minimun 2-norm solutions are computed if TAU is present.
!      Otherwise,  solutions are computed such that if the j-th column of MAT
!      is omitted from the basis, X[j,:] is set to zero.
!
!
!   3) The 2-norm of the residual vector for the solution in the 
!      j-th column of X is given in RNORM[j] if argument RNORM is present.
!
!
!   4) If COMP_RESID=true, The residual matrix B - MAT * X overwrites B on exit.
! 
!
!   5) MAT, DIAGR, BETA, IP, KRANK and TAU are not modified by this routine and can
!      be left in place for successive calls with different right-hand side matrices B.
!
!   For further details on linear least square problems and algorithms to solve them, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!          Matrix Computations.
!          3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (3) Hansen, P.C., Pereyra, V., and Scherer, G., 2012:
!           Least Squares Data Fitting with Applications.
!           Johns Hopkins University Press, 328 pp.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert_eq, assert, merror, norm
    use QR_Procedures,   only : apply_q_qr
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in),    dimension(:)            :: diagr, beta
    real(stnd),   intent(in),    dimension(:,:)          :: mat
    real(stnd),   intent(in),    dimension(:),  optional :: tau
    real(stnd),   intent(inout), dimension(:,:)          :: b
    real(stnd),   intent(out),   dimension(:,:)          :: x
    real(stnd),   intent(out),   dimension(:),  optional :: rnorm
!
    integer(i4b), intent(in)               :: krank
    integer(i4b), intent(in), dimension(:) :: ip
!
    logical(lgl), intent(in), optional  :: comp_resid
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: cl
    real(stnd), dimension(max(size(x,1),size(b,2))) :: tmp
    real(stnd), dimension(:,:), allocatable         :: hmat, xt
!
    integer      :: iok
    integer(i4b) :: m, n, ndiag, nb, i, kp1, j, l
!
    logical(lgl) :: test_comp_resid, comp_min_norm
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc ='qr_solve2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(b,1),i4b)   ,   &
                    name_proc )
!
    n =  assert_eq( int(size(mat,2),i4b) ,   &
                    int(size(x,1),i4b)   ,   &
                    int(size(ip),i4b)    ,   &
                    name_proc )
!
    ndiag =  assert_eq( min( n, m )    ,         &
                        int(size(diagr),i4b) ,   &
                        int(size(beta),i4b)  ,   &
                        name_proc )
!
    comp_min_norm = present( tau )
!
    if ( comp_min_norm ) then
        call assert( logical(size(tau)==ndiag,lgl), name_proc )
    end if
!
    nb =  assert_eq( int(size(x,2),i4b)   ,   &
                     int(size(b,2),i4b)   ,   &
                     name_proc )
!
    if ( present(rnorm) ) then
        call assert( logical(size(rnorm)==nb,lgl), name_proc )
    end if
!
    if ( present(comp_resid) ) then
        test_comp_resid = comp_resid
    else
        test_comp_resid = false
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( min(nb,krank)<=0_i4b ) then
!
        x(:n,:nb) = zero
!
        if ( present(rnorm) ) then
            rnorm(:nb) = norm( b, dim=2_i4b)
        end if
!
        return
!
    end if
!
!   APPLY THE HOUSEHOLDER TRANSFORMATIONS TO b .
!
    call apply_q_qr( mat(:m,:krank), beta(:krank), b(:m,:nb), left=true, trans=true )
!
    x(:krank,:nb) = b(:krank,:nb)
!    
    kp1 = krank + 1_i4b
!    
!   SOLVE THE krank BY krank TRIANGULAR SYSTEMS.
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and. &
                  nb>=2_i4b .and. i>1_i4b
#endif
!    
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC,1), PRIVATE(i,l,cl)
!
    do l = 1_i4b, nb
!
        do i = krank, 1_i4b, -1_i4b
!
            cl = x(i,l)/diagr(i)
#ifdef _BLAS
            call axpy( i-1_i4b, -cl, mat(1_i4b:i-1_i4b,i), 1_i4b, x(1_i4b:i-1_i4b,l), 1_i4b )
#else
            x(1_i4b:i-1_i4b,l) = x(1_i4b:i-1_i4b,l) - cl*mat(1_i4b:i-1_i4b,i)
#endif
            x(i,l) = cl
!
        end do
!
    end do
!    
!$OMP END PARALLEL DO
!
!   COMPLETE COMPUTATION OF SOLUTION VECTORS.
!
    x(kp1:n,:nb) = zero
!
!   COMPUTE MINIMAL NORM SOLUTION VECTORS IF REQUIRED.
!
    if ( krank/=n .and. comp_min_norm ) then
!
        l = n - krank
!
!       ALLOCATE WORK VARIABLES FOR STORING HOUSEHOLDER AND SOLUTION VECTORS.
!
        allocate( hmat(l,krank), xt(nb,n), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       TRANSPOSE HOUSEHOLDER AND SOLUTION VECTORS FOR LATER USE.
!
        hmat(:l,:krank) = transpose( mat(:krank,kp1:n) )
!
        xt(:nb,:n) = transpose( x(:n,:nb) )
!
        do i = 1_i4b, krank
!
            if ( tau(i)==zero ) cycle
!
            tmp(:nb)  = matmul( xt(:nb,kp1:n), hmat(:l,i) )
            tmp(:nb)  = ( xt(:nb,i) + tmp(:nb) )*tau(i)
            xt(:nb,i) = xt(:nb,i) + tmp(:nb)
!
            do j = kp1, n
!
                xt(:nb,j) = xt(:nb,j) + hmat(j-krank,i)*tmp(:nb)
!
            end do
!
        end do
!
        x(:n,:nb) = transpose( xt(:nb,:n) )
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( hmat, xt )
!
    end if
!    
!   RE-ORDER THE SOLUTION VECTORS TO COMPENSATE FOR THE COLUMN INTERCHANGES.
!    
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC,1), PRIVATE(i,tmp)
!    
    do i = 1_i4b, nb
!
        tmp(ip) = x(:n,i)
        x(:n,i) = tmp(:n)
!
    end do
!    
!$OMP END PARALLEL DO
!    
!   COMPUTE THE NORM OF THE RESIDUAL VECTORS IF REQUIRED.
!    
    if ( present(rnorm) ) then
        rnorm(:nb) = norm( b(kp1:m,:nb), dim=2_i4b )
    end if
!
!   COMPUTE b - mat*x IF REQUIRED.
!
    if ( test_comp_resid ) then
!
        b(1_i4b:krank,:nb) = zero
!
        call apply_q_qr( mat(:m,:krank), beta(:krank), b(:m,:nb), left=true, trans=false )
!
    end if
!
!
! END OF SUBROUTINE qr_solve2_rm
! ______________________________
!
    end subroutine qr_solve2_rm
!
! =========================================================================================
!
    subroutine llsq_svd_solve_rm( mat, b, failure, x,                   &
                                  singvalues, krank, rnorm, tol,        &
                                  mul_size, maxiter, max_francis_steps, &
                                  perfect_shift                         )
!
! Purpose
! _______
!
!   LLSQ_SVD_SOLVE computes the minimum norm solution to a real linear least
!   squares problem:
!
!                       Minimize 2-norm|| B - MAT * X ||
!
!   using the singular value decomposition (SVD) of MAT. MAT is an m-by-n
!   matrix which may be rank-deficient.
!
!   Several right hand side vectors b and solution vectors x can be
!   handled in a single call; they are stored as the columns of the
!   m-by-nrhs right hand side matrix B and the n-by-nrhs solution
!   matrix X, respectively.
!
!   The effective rank of MAT, KRANK, is determined by treating as zero those
!   singular values which are less than TOL times the largest singular
!   value.
!
!
! Arguments
! _________
!
!   MAT                 (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                       On entry, the m-by-n matrix MAT.
!
!                       On exit, MAT is destroyed. If m>=n, MAT(:n,:n)
!                       is overwritten with the right singular vectors
!                       of MAT, stored columnwise.
!
!   B                   (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                       On entry, the m-by-nrhs right hand side matrix B.
!
!                       On exit, B is destroyed. If m>KRANK, the residual 
!                       sum-of-squares for the solution in the i-th column
!                       is given by the sum of squares of elements KRANK+1:m 
!                       in that column.
!
!                       The shape of B must verify: 
!
!                       - size( B, 1 ) = size( MAT, 1 ) = m
!                       - size( B, 2 ) = size( X, 2 )   = nrhs .
!
!   FAILURE             (OUTPUT) logical(lgl)
!                       If:
!
!                       - FAILURE= false :  indicates successful exit
!                       - FAILURE= true  :  indicates that the SVD algorithm did not converge and
!                         that full accuracy was not attained in the bidiagonal
!                         SVD of an intermediate bidiagonal form BD of MAT.
!
!   X                   (OUTPUT) real(stnd), dimension(:,:)
!                       On exit, the n-by-nrhs solution matrix X.
!
!                       The shape of X must verify: 
!
!                       - size( X, 1 ) = size( MAT, 2 ) = n
!                       - size( X, 2 ) = size( B, 2 )   = nrhs .
!
!   SINGVALUES          (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                       The singular values of MAT in decreasing order.
!                       The condition number of MAT in the 2-norm is
!
!                             SINGVALUES(1)/SINGVALUES(min(m,n)).
!
!                       The size of SINGVALUES must verify: size( SINGVALUES ) = min(m,n) .
!
!   KRANK               (OUTPUT, OPTIONAL) integer(i4b)
!                       On exit, the effective rank of MAT, i.e., the number
!                       of singular values which are greater than TOL * SINGVALUES(1).
!
!   RNORM               (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                       On exit, the 2-norms of the residual vectors for 
!                       the solutions stored columnwise in the matrix X.
!
!                       The size of RNORM must verify: 
!
!                       - size( RNORM ) = size( X, 2 ) = size( B, 2 ) = nrhs .
!
!   TOL                 (INPUT, OPTIONAL) real(stnd)
!                       On entry, TOL is used to determine the effective rank of MAT.
!                       Singular values SINGVALUES(i) <= TOL * SINGVALUES(1) are treated 
!                       as zero. If TOL is absent, machine precision is used instead.
!
!   MUL_SIZE            (INPUT, OPTIONAL) integer(i4b)
!                       Internal parameter. MUL_SIZE must verify 1 <= MUL_SIZE <= max(m,n),
!                       otherwise a default value is used. MUL_SIZE can be increased or decreased
!                       to improve the performance of the algorithm used in LLSQ_SVD_SOLVE.
!                       Maximum performance will be obtained when a real matrix of size
!                       MUL_SIZE * max(m,n) and kind stnd fits in the cache of the processors.
!
!                       The default value is 32.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the
!                       bidiagonal SVD phase of the SVD algorithm.
!                       The bidiagonal SVD algorithm of an intermediate bidiagonal 
!                       form B of MAT fails to converge if the number of QR sweeps 
!                       exceeds MAXITER * min(m,n). Convergence usually occurs in 
!                       about 2 * min(m,n) QR sweeps.
!
!                       The default is 10.
!
!   MAX_FRANCIS_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                       MAX_FRANCIS_STEPS controls the maximum number of Francis sets
!                       (eg QR sweeps) of Givens rotations which must be saved
!                       before applying them with a wavefront algorithm to accumulate
!                       the singular vectors in the bidiagonal SVD algorithm.
!                       MAX_FRANCIS_STEPS is a strictly positive integer, otherwise 
!                       the default value is used.
!
!                       The default is 10.
!
!   PERFECT_SHIFT       (INPUT, OPTIONAL) logical(lgl)
!                       PERFECT_SHIFT determines if a perfect shift strategy is used in the
!                       implicit QR algorithm in order to minimize the number of QR sweeps
!                       in the bidiagonal SVD algorithm.
!
!                       The default is true.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the routine DGELSS in LAPACK. If OPENMP
!   is used, the algorithm is parallelized.
!
!   For further details, on using the SVD for solving a least square problem,
!   see the references (1) or (2).
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!          Matrix Computations.
!          3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (3) Hansen, P.C., Pereyra, V., and Scherer, G., 2012:
!           Least Squares Data Fitting with Applications.
!           Johns Hopkins University Press, 328 pp.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch
    use Char_Constants,  only : allocate_error
    use Utilities,       only : assert_eq, assert, merror, lascl, norm, ifirstloc
    use QR_Procedures,   only : qr_cmp, apply_q_qr, ortho_gen_qr, lq_cmp, apply_q_lq, ortho_gen_lq
    use SVD_Procedures,  only : bd_cmp, bd_svd2
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:,:)         :: mat, b
    real(stnd), intent(out),   dimension(:,:)         :: x
    real(stnd), intent(in),                  optional :: tol
    real(stnd), intent(out),   dimension(:), optional :: rnorm, singvalues 
!
    integer(i4b), intent(in),  optional :: mul_size, maxiter, max_francis_steps
    integer(i4b), intent(out), optional :: krank
!
    logical(lgl), intent(in), optional :: perfect_shift
    logical(lgl), intent(out)          :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension( min(size(mat,1),size(mat,2)) ) :: s, e, tauq, taup
    real(stnd)                    :: eps, sfmin, smlnum, bignum, matnorm, bnorm, thr
    real(stnd), allocatable       :: work(:,:), tau(:)
!
    integer(i4b) :: i, j, k, l, m, mm, n, nrhs, inc, mnthr, imatscl, ibscl, i1, i2
!
    integer :: iok
!
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='llsq_svd_solve'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq( int(size(mat,1),i4b), &
                   int(size(b,1),i4b),   &
                   name_proc )
!
    n = assert_eq( int(size(mat,2),i4b), &
                   int(size(x,1),i4b),   &
                   name_proc )
!
    nrhs = assert_eq( int(size(b,2),i4b),   &
                      int(size(x,2),i4b),   &
                      name_proc )
!
    if ( present(rnorm) ) then
        call assert( logical(size(rnorm)==nrhs,lgl), name_proc )
    end if
!
    k = min( m, n )
!
    if ( present(singvalues) ) then
        call assert( logical(size(singvalues)==k,lgl), name_proc )
    end if
!
    if ( present(mul_size) ) then
        inc = merge( mul_size, 32_i4b, 1_i4b<=mul_size .and. mul_size<=max(m,n) )
    else
        inc = 32_i4b
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( k<=0_i4b ) then
!
        x = zero
!
        if ( present(krank) ) krank = 0_i4b
        if ( present(rnorm) ) rnorm = norm( b, dim=2_i4b)
!
        return
!
    end if
!    
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and. i>1_i4b
#endif
!
!   GET MACHINE CONSTANTS.
!
    eps    = lamch( 'p' )
    sfmin  = lamch( 's' )
    smlnum = sfmin/eps
    bignum = one/smlnum
!
!   IF IT LOOKS LIKE WE'RE ON A CRAY, TAKE THE SQUARE ROOT OF
!   smlnum AND bignum TO AVOID OVERFLOW AND UNDERFLOW PROBLEM.
!
    if ( log10( bignum )>c2000 ) then
         smlnum = sqrt( smlnum )
         bignum = sqrt( bignum )
    end if
!
    j = 1_i4b - inc
!
!   COMPUTE MAX ELEMENT OF mat .
!
    l = inc*( n/inc )
    if ( l/=n)  l = l + inc
    matnorm = zero
!
!$OMP PARALLEL DO IF(n*m>=omp_limit .and. inc+inc<l .and. test_par), SCHEDULE(STATIC)             &
!$OMP            ,PRIVATE(i,i1,i2,thr), REDUCTION(max:matnorm)
!
    do i = inc, l, inc
!
        i1 = i + j
        i2 = min( i, n )
        thr = maxval( abs( mat(:m,i1:i2) ) )
        matnorm = max( matnorm, thr )
!
    end do
!
!$OMP END PARALLEL DO
!
!   SCALE mat IF MAX ELEMENT OUTSIDE RANGE [smlnum,bignum].
!
    imatscl = 0_i4b
!
    if ( matnorm==zero ) then
!
!       MATRIX ALL ZERO, RETURN ZERO SOLUTION.
!
        x = zero
!
        if ( present(krank) ) krank = 0_i4b
        if ( present(singvalues) ) singvalues = zero
 !
        return
!
    else if ( matnorm<smlnum ) then
!
!       SCALE MATRIX NORM UP TO smlnum.
!
        imatscl = 1_i4b
        call lascl( mat, matnorm, smlnum )
!
    else if ( matnorm>bignum ) then
!
!       SCALE MATRIX NORM DOWN TO bignum.
!
        imatscl = 2_i4b
        call lascl( mat, matnorm, bignum )
!
    end if
!
!   COMPUTE MAX ELEMENT OF b .
!
    l = inc*( nrhs/inc )
    if ( l/=nrhs)  l = l + inc
    bnorm = zero
!
!$OMP PARALLEL DO IF(nrhs*m>=omp_limit .and. inc+inc<l .and. test_par), SCHEDULE(STATIC)  &
!$OMP            ,PRIVATE(i,i1,i2,thr), REDUCTION(max:bnorm)
!
    do i = inc, l, inc
!
        i1 = i + j
        i2 = min( i, nrhs )
        thr = maxval( abs( b(:m,i1:i2) ) )
        bnorm = max( bnorm, thr )
!
    end do
!
!$OMP END PARALLEL DO
!
!
!   SCALE b IF MAX ELEMENT OUTSIDE RANGE [smlnum,bignum].
!
    ibscl = 0_i4b
!
    if ( bnorm>zero .and. bnorm<smlnum ) then
!
!       SCALE MATRIX NORM UP TO smlnum.
!
        ibscl = 1_i4b
        call lascl( b, bnorm, smlnum )
!
    else if ( bnorm>bignum ) then
!
!       SCALE MATRIX NORM DOWN TO bignum.
!
        ibscl = 2_i4b
        call lascl( b, bnorm, bignum )
!
    end if
!
!   DETERMINE CROSSOVER POINT FOR THE SVD WHEN REDUCING
!   mat  TO BIDIAGONAL FORM. IF max(m,n) EXCEEDS mnthr ,
!   A QR OR LQ FACTORIZATION IS USED FIRST TO REDUCE THE
!   MATRIX mat TO A TRIANGULAR FORM.
!    
    mnthr = int( real( k, stnd )*c1_6, i4b )
!
    if ( m>=n ) then
!
!       PATH 1 : m LARGER THAN n .
!
        mm = m
!
        if (  m>=mnthr  ) then
!
!           PATH 1a : m MUCH LARGER THAN n .
!
            mm = n
!
!           FIRST REDUCE mat USING THE QR DECOMPOSITION.
!
!           COMPUTE mat = Q * R .
!
            call qr_cmp( mat, s, tauq )
!
!           MULTIPLY b BY Q**t .
!
            call apply_q_qr( mat, tauq, b, left=true, trans=true )
!
!
!           ZERO OUT BELOW R IN mat .
!
!$OMP PARALLEL DO IF((n*(n+1_i4b))/2_i4b>=omp_limit .and. test_par)     &
!$OMP            ,SCHEDULE(STATIC,1)                                    &
!$OMP            ,PRIVATE(i)
!
            do i = 1_i4b, n
                mat(i,i) = s(i)
                mat(i+1_i4b:n,i) = zero
            end do
!
!$OMP END PARALLEL DO
!
        end if
!
!       BIDIAGONALIZE R OR mat IN mat .
!
        call bd_cmp( mat(:mm,:), s, e, tauq, taup )
!
!       MULTIPLY b BY TRANSPOSE OF LEFT BIDIAGONALIZING VECTORS OF R OR mat .
!
        call apply_q_qr( mat(:mm,:), tauq, b(:mm,:), left=true, trans=true )
!
!       GENERATE RIGHT VECTORS BIDIAGONALIZING mat IN mat .
!
!$OMP PARALLEL DO IF(n*n>=omp_limit .and. test_par), SCHEDULE(STATIC,1), PRIVATE(i)
!
        do i = 2_i4b, n
            mat(i,2_i4b:i) = mat(:i-1_i4b,i)
        end do
!
!$OMP END PARALLEL DO
!
        mat(1_i4b,:n) = zero
        mat(2_i4b:n,1_i4b) = zero
        mat(1_i4b,1_i4b) = one
!
        call ortho_gen_qr( mat(2_i4b:n,2_i4b:n), taup(:n-1_i4b) )
!
!       PERFORM BIDIAGONAL QR ITERATION, MULTIPLY b BY TRANSPOSE
!       OF LEFT SINGULAR VECTORS OF R OR mat AND COMPUTE RIGHT SINGULAR
!       VECTORS OF mat IN mat(:n,:n) .
!
        call bd_svd2( false, s, e, failure, mat(:n,:n), b(:n,:), sort='d',  &
                      maxiter=maxiter, max_francis_steps=max_francis_steps, &
                      perfect_shift=perfect_shift  )
!
!       DETERMINE THE EFFECTIVE RANK OF mat .
!
        if ( present(tol) ) then
            thr = max( tol*s(1_i4b), sfmin )
        else
            thr = max( eps*s(1_i4b), sfmin )
        end if
!
        k = ifirstloc( logical( s<=thr, lgl ) ) - 1_i4b
!
!       COMPUTE RECIPROCALS OF SINGULAR VALUES.
!
        e(:k) = one/s(:k)
!
!       MULTIPLY b BY RECIPROCALS OF SINGULAR VALUES.
!
        b(:k,:nrhs) = b(:k,:nrhs)*spread(e(:k), dim=2_i4b, ncopies=nrhs )
!
!        do i=1_i4b, k
!            call lascl( b(i,:), s(i), one )
!        end do
!
!       MULTIPLY b BY RIGHT SINGULAR VECTORS OF mat IN mat .
!
        x(:n,:nrhs) = matmul( mat(:n,:k), b(:k,:nrhs) )
!
    else
!
!       PATH 2 : n LARGER THAN m.
!
        iok = 1_i4b
!
        if (  n>=mnthr  ) then
!
            allocate( work(m,m), tau(m), stat=iok )
!
        end if
!
        if ( iok==0_i4b ) then
!
!           PATH 2a : n MUCH LARGER THAN m AND SUFFICIENT
!                     WORKSPACE FOR A FAST ALGORITHM.
!
!           FIRST REDUCE mat USING THE LQ DECOMPOSITION.
!
!           COMPUTE mat = L * Q .
!
            call lq_cmp( mat, s, tau )
!
!           COPY L**t TO work, ZEROING OUT BELOW IT.
!
!$OMP PARALLEL DO IF(m*m>=omp_limit .and. test_par)    &
!$OMP            ,SCHEDULE(STATIC)                     &
!$OMP            ,PRIVATE(i)
!
            do i = 1_i4b, m
                work(i,:i-1_i4b) = zero
                work(i,i) = s(i)
                work(i,i+1_i4b:m) = mat(i+1_i4b:m,i)
            end do
!
!$OMP END PARALLEL DO
!
!           BIDIAGONALIZE L**t IN work .
!
            call bd_cmp( work, s, e, tauq, taup )
!
!           MULTIPLY b BY TRANSPOSE OF RIGHT BIDIAGONALIZING VECTORS OF L**t .
!
            call apply_q_lq( work(:,2_i4b:), taup(:m-1_i4b), b(2_i4b:,:),    &
                            left=true, trans=false )
!
!           GENERATE LEFT VECTORS BIDIAGONALIZING L**t IN work .
!
            call ortho_gen_qr( work, tauq )
!
!           PERFORM BIDIAGONAL QR ITERATION, MULTIPLY B BY TRANSPOSE
!           OF RIGHT SINGULAR VECTORS OF L**t AND COMPUTE LEFT SINGULAR
!           VECTORS OF L**t IN work .
!
            call bd_svd2( true, s, e, failure, work, b, sort='d',               &
                          maxiter=maxiter, max_francis_steps=max_francis_steps, &
                          perfect_shift=perfect_shift )
!
!           DETERMINE THE EFFECTIVE RANK OF mat .
!
            if ( present(tol) ) then
                thr = max( tol*s(1_i4b), sfmin )
            else
                thr = max( eps*s(1_i4b), sfmin )
            end if
!
            k = ifirstloc( logical( s<=thr, lgl ) ) - 1_i4b
!
!           COMPUTE RECIPROCALS OF SINGULAR VALUES.
!
            e(:k) = one/s(:k)
!
!           MULTIPLY LEFT SINGULAR VECTORS OF L**t IN work
!           BY RECIPROCALS OF SINGULAR VALUES.
!
            work(:m,:k) = work(:m,:k)*spread(e(:k), dim=1_i4b, ncopies=m )
!
!           MULTIPLY b BY SCALED LEFT SINGULAR VECTORS OF L**t IN work .
!
            x(:m,:nrhs) = matmul( work(:m,:k), b(:k,:nrhs) )
!
            x(m+1_i4b:n,:nrhs) = zero
!
!           MULTIPLY x BY Q**t IN mat .
!
            call apply_q_lq( mat, tau, x, left=true, trans=true )
!
!           DEALLOCATE WORKSPACE .
!
            deallocate( work, tau )
!
        else
!
!           PATH 2b : n NOT MUCH LARGER THAN m OR INSUFFICIENT
!                     WORKSPACE FOR A FAST ALGORITHM.
!
            allocate( work(nrhs,m), stat=iok )
!
            if ( iok/=0_i4b ) then
                call merror( name_proc//allocate_error )
            end if
!
!           BIDIAGONALIZE mat .
!
            call bd_cmp( mat, s, e, tauq, taup )
!
!           MULTIPLY b BY TRANSPOSE OF LEFT BIDIAGONALIZING VECTORS OF mat .
!
            call apply_q_qr( mat(2_i4b:,:), tauq(:m-1_i4b), b(2_i4b:,:),    &
                            left=true, trans=true )
!
!           COPY b**t TO work .
!
!$OMP PARALLEL DO IF(m*nrhs>=omp_limit .and. nrhs>=2 .and. test_par), PRIVATE(i), SCHEDULE(STATIC)
!
            do i = 1_i4b, nrhs
                work(i,:m) = b(:m,i)
            end do
!
!$OMP END PARALLEL DO
!
!           GENERATE RIGHT BIDIAGONALIZING VECTORS OF mat IN mat .
!
            call ortho_gen_lq( mat, taup )        
!
!           PERFORM BIDIAGONAL QR ITERATION, MULTIPLY b**t IN work BY
!           LEFT SINGULAR VECTORS OF BIDIAGONAL MATRIX AND COMPUTE RIGHT
!           SINGULAR VECTORS OF mat IN mat .
!
            call bd_svd2( false, s, e, failure, work, mat, sort='d',            &
                          maxiter=maxiter, max_francis_steps=max_francis_steps, &
                          perfect_shift=perfect_shift  )
!
!           DETERMINE THE EFFECTIVE RANK OF mat .
!
            if ( present(tol) ) then
                thr = max( tol*s(1_i4b), sfmin )
            else
                thr = max( eps*s(1_i4b), sfmin )
            end if
!
            k = ifirstloc( logical( s<=thr, lgl ) ) - 1_i4b
!
!           COMPUTE RECIPROCALS OF SINGULAR VALUES.
!
            e(:k) = one/s(:k)
!
!           MULTIPLY RIGHT SINGULAR VECTORS OF mat STORED ROWWISE IN mat
!           BY RECIPROCALS OF SINGULAR VALUES.
!
            mat(:k,:n) = mat(:k,:n)*spread(e(:k), dim=2_i4b, ncopies=n )
!
            do i = 1_i4b, nrhs
!
!               MULTIPLY b BY SCALED RIGHT SINGULAR VECTORS OF mat .
!
                x(:n,i) = matmul( work(i,:k), mat(:k,:n) )
!
            end do
!
!           COPY RESIDUALS IN work(:nrhs,k+1_i4b:m)**t TO b(k+1_i4b:m,:nrhs) .
!
            b(k+1_i4b:m,1_i4b:nrhs) = transpose( work(1_i4b:nrhs,k+1_i4b:m) )
!
!           DEALLOCATE WORKSPACE .
!
            deallocate( work )
!
        end if
!
    end if
!
!   UNDO SCALING IF NECESSARY.
!
    if ( imatscl==1_i4b ) then
       call lascl( x, matnorm, smlnum  )
       call lascl( s, smlnum , matnorm )
    else if ( imatscl==2_i4b ) then
       call lascl( x, matnorm, bignum  )
       call lascl( s, bignum , matnorm )
    end if
!
    if ( ibscl==1_i4b ) then
       call lascl( x, smlnum , bnorm )
    else if ( ibscl==2_i4b ) then
       call lascl( x, bignum , bnorm )
    end if
!    
!   COMPUTE THE NORM OF THE RESIDUAL VECTORS IF REQUIRED.
!    
    if ( present(rnorm) ) then
        rnorm = norm( b(k+1_i4b:m,:nrhs), dim=2_i4b )
    end if
!    
!   OUTPUT SINGULAR VALUES OF mat, IF REQUIRED.
!    
    if ( present(singvalues) )  singvalues = s
!    
!   OUTPUT RANK OF mat, IF REQUIRED.
!    
    if ( present(krank) )  krank = k
!
!
! END OF SUBROUTINE llsq_svd_solve_rm
! ___________________________________
!
    end subroutine llsq_svd_solve_rm
!
! =========================================================================================
!
    subroutine llsq_svd_solve_rv( mat, b, failure, x,                   &
                                  singvalues, krank, rnorm, tol,        &
                                  mul_size, maxiter, max_francis_steps, &
                                  perfect_shift  )
!
! Purpose
! _______
!
!   LLSQ_SVD_SOLVE computes the minimum norm solution to a real linear least
!   squares problem:
!
!                       Minimize 2-norm|| B - MAT * X ||
!
!   using the singular value decomposition (SVD) of MAT. MAT is an m-by-n
!   matrix which may be rank-deficient, B is a m right hand side vector
!   and X is a n solution vector.
!
!   The effective rank of MAT, KRANK, is determined by treating as zero those
!   singular values which are less than TOL times the largest singular
!   value.
!
!
! Arguments
! _________
!
!   MAT                 (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                       On entry, the m-by-n matrix MAT.
!
!                       On exit, MAT is destroyed. If m>=n, MAT(:n,:n)
!                       is overwritten with the right singular vectors
!                       of MAT, stored columnwise.
!
!   B                   (INPUT/OUTPUT) real(stnd), dimension(:)
!                       On entry, the m-elements right hand side vector B.
!
!                       On exit, B is destroyed. If m>KRANK, the residual 
!                       sum-of-squares for the solution X is given by the sum 
!                       of squares of elements KRANK+1:m of B . 
!
!                       The size of B must verify: size( B ) = size( MAT, 1 ) = m .
!
!   FAILURE             (OUTPUT) logical(lgl)
!                       If:
!
!                       - FAILURE = false :  indicates successful exit
!                       - FAILURE = true  :  indicates that the SVD algorithm did not converge and
!                         that full accuracy was not attained in the bidiagonal
!                         SVD of an intermediate bidiagonal form BD of MAT.
!
!   X                   (OUTPUT) real(stnd), dimension(:)
!                       On exit, the n-elements solution vector X.
!
!                       The size of X must verify: size( X ) = size( MAT, 2 ) = n .
!
!   SINGVALUES          (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                       The singular values of MAT in decreasing order.
!                       The condition number of MAT in the 2-norm is
!
!                             SINGVALUES(1)/SINGVALUES(min(m,n)).
!
!                       The size of SINGVALUES must verify: size( SINGVALUES ) = min(m,n) .
!
!   KRANK               (OUTPUT, OPTIONAL) integer(i4b)
!                       On exit, the effective rank of MAT, i.e., the number
!                       of singular values which are greater than TOL * SINGVALUES(1).
!
!   RNORM               (OUTPUT, OPTIONAL) real(stnd)
!                       On exit, the 2-norm of the residual vector for 
!                       the solution vector X.
!
!   TOL                 (INPUT, OPTIONAL) real(stnd)
!                       On entry, TOL is used to determine the effective rank of MAT.
!                       Singular values SINGVALUES(i) <= TOL * SINGVALUES(1) are treated 
!                       as zero. If TOL is absent, machine precision is used instead.
!
!   MUL_SIZE            (INPUT, OPTIONAL) integer(i4b)
!                       Internal parameter. MUL_SIZE must verify 1 <= MUL_SIZE <= max(m,n),
!                       otherwise a default value is used. MUL_SIZE can be increased or decreased
!                       to improve the performance of the algorithm used in LLSQ_SVD_SOLVE.
!                       Maximum performance will be obtained when a real matrix of size
!                       MUL_SIZE * max(m,n) and kind stnd fits in the cache of the processors.
!
!                       The default value is 32.
!
!   MAXITER             (INPUT, OPTIONAL) integer(i4b)
!                       MAXITER controls the maximum number of QR sweeps in the
!                       bidiagonal SVD phase of the SVD algorithm.
!                       The bidiagonal SVD algorithm of an intermediate bidiagonal 
!                       form B of MAT fails to converge if the number of QR sweeps 
!                       exceeds MAXITER * min(m,n). Convergence usually occurs in 
!                       about 2 * min(m,n) QR sweeps.
!
!                       The default value is 10.
!
!   MAX_FRANCIS_STEPS   (INPUT, OPTIONAL) integer(i4b)
!                       MAX_FRANCIS_STEPS controls the maximum number of Francis sets
!                       (eg QR sweeps) of Givens rotations which must be saved
!                       before applying them with a wavefront algorithm to accumulate
!                       the singular vectors in the bidiagonal SVD algorithm.
!                       MAX_FRANCIS_STEPS is a strictly positive integer, otherwise 
!                       the default value is used.
!
!                       The default value is 10.
!
!   PERFECT_SHIFT       (INPUT, OPTIONAL) logical(lgl)
!                       PERFECT_SHIFT determines if a perfect shift strategy is used in the
!                       implicit QR algorithm in order to minimize the number of QR sweeps
!                       in the bidiagonal SVD algorithm.
!
!                       The default is true.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the routine DGELSS in LAPACK. If OPENMP
!   is used, the algorithm is parallelized.
!
!   For further details on using the SVD for solving a least square problem,
!   see the references (1) or (2).
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Lawson, C.L., and Hanson, R.J., 1974:
!           Solving least square problems.
!           Prentice-Hall.
!
!   (3) Hansen, P.C., Pereyra, V., and Scherer, G., 2012:
!           Least Squares Data Fitting with Applications.
!           Johns Hopkins University Press, 328 pp.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : norm
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout)           :: mat(:,:), b(:)
    real(stnd), intent(out)             :: x(:)
    real(stnd), intent(in),  optional   :: tol
    real(stnd), intent(out), optional   :: rnorm, singvalues(:) 
!
    integer(i4b), intent(in),  optional :: mul_size, maxiter, max_francis_steps
    integer(i4b), intent(out), optional :: krank
!
    logical(lgl), intent(in), optional :: perfect_shift
    logical(lgl), intent(out)          :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: x2(size(x),1), b2(size(b),1)
!
    integer(i4b) :: k
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   COPY RIGHT HAND SIDE b TO WORK ARRAY b2.
!
    b2(:,1_i4b) = b
!
    call llsq_svd_solve_rm( mat, b2, failure, x2,                       &
                            singvalues=singvalues, krank=k, tol=tol,    &
                            mul_size=mul_size, maxiter=maxiter,         &
                            max_francis_steps=max_francis_steps,        &
                            perfect_shift=perfect_shift                 )
!
    x = x2(:,1_i4b)
    b = b2(:,1_i4b)
!    
!   COMPUTE THE NORM OF THE RESIDUAL VECTOR IF REQUIRED.
!    
    if ( present(rnorm) ) rnorm = norm( b(k+1_i4b:) )
!    
!   OUTPUT RANK OF mat, IF REQUIRED.
!    
    if ( present(krank) )  krank = k
!
!
! END OF SUBROUTINE llsq_svd_solve_rv
! ___________________________________
!
    end subroutine llsq_svd_solve_rv
!
!
! =========================================================================================
!
!
! *******************************
! END OF MODULE LLSQ_Procedures *
! *******************************
!    
end module LLSQ_Procedures
