module Lin_Procedures
! #############################################################################################
!
! Copyright 2020 IRD
!
! This file is part of statpack.
!
! statpack is free software: you can redistribute it and/or modify
! it under the terms of the GNU Lesser General Public License as
! published by the Free Software Foundation, either version 3 of 
! the License, or (at your option) any later version.
!
! statpack is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU Lesser General Public License for more details.
!
! You can find a copy of the GNU Lesser General Public License
! in the statpack/doc directory.
!
! #############################################################################################
!                                                                                             *
! *********************************************************************************************
! MODULE EXPORTING SUBROUTINES AND FUNCTIONS FOR THE SOLUTION OF SYSTEMS OF LINEAR EQUATIONS, *
! COMPUTING A TRIANGULAR FACTORIZATION (LU, CHOLESKY), COMPUTING THE INVERSE OF A MATRIX AND  *
! COMPUTING THE DETERMINANT OF A MATRIX.                                                      *
!                                                                                             *
! LATEST REVISION : 17/09/2020                                                                *
!                                                                                             *
! #############################################################################################
!                                                                                             *
! *********************************************************************************************
!                                                                                             *
! THE C PROCESSOR MACROS USED IN THIS MODULE ARE:                                             *
!                                                                                             *
!  _OPENMP      FOR ACTIVATING OPENMP PARALLELIZATION                                         *
!  _BLAS        FOR USING BLAS WHEN POSSIBLE                                                  *
!  _DOT_PRODUCT FOR REPLACING THE dot_product INTRINSIC FUNCTION WITH STATPACK                *
!               dot_product2 FUNCTION                                                         *
!  _MATMUL      FOR REPLACING THE matmul INTRINSIC FUNCTION WITH STATPACK                     *
!               matmul2 FUNCTION                                                              *
!  _ORDERED     FOR DESACTIVATING OPENMP PARALLELIZATION IN comp_uut_ltl_a SUBROUTINE, WHICH  *
!               USES AN ORDERED CLAUSE IN A PARALLEL DO DIRECTIVE                             *
!  _USE_NAGWARE FOR DESACTIVATING SOME OPENMP CONSTRUCTS FOR THE NAG FORTRAN                  *
!               COMPILER. HERE, THIS INCLUDES ONLY THE ACTIVATION OF THE _ORDERED MACRO       *
!  _USE_PGI     FOR DESACTIVATING SOME OPENMP CONSTRUCTS FOR THE PORTLAND PGFORTRAN           *
!               COMPILER                                                                      *
!                                                                                             *
! *********************************************************************************************
!
#ifdef _USE_NAGWARE
#define _ORDERED
#endif
!
! USED MODULES
! ============
!   
    use Select_Parameters,     only : lgl, i4b, stnd, blksz_lin
    use Reals_Constants,       only : zero, one
    use Logical_Constants
#ifdef _MATMUL
    use Utilities,             only : matmul=>matmul2
#endif
#ifdef _DOT_PRODUCT
    use Utilities,             only : dot_product=>dot_product2
#endif
#ifdef _OPENMP
    use Select_Parameters,     only : omp_limit
    use omp_lib,               only : omp_lock_kind, omp_get_max_threads, omp_in_parallel, omp_get_nested,  &
                                      omp_get_num_threads, omp_get_thread_num, omp_init_lock, omp_set_lock, &
                                      omp_unset_lock, omp_destroy_lock
#endif
!
! USED MODULES IN SPECIFIC PROCEDURES
! ===================================
!   
!   use Num_Constants
!   use Char_Constants
!   use Utilities
!   use BLAS_interfaces
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!
! PUBLIC ENTITIES 
! ===============
!    
! ALL SUBROUTINES, FUNCTIONS, VARIABLES AND PARAMETERS ARE PRIVATE BY DEFAULT.
!
    private
    public ::   lu_cmp, lu_cmp2,                            &
                chol_cmp, chol_cmp2,                        &
                gchol_cmp, gchol_cmp2,                      &
                solve_lin, lin_lu_solve,                    &
                lu_solve, lu_solve2,                        &
                chol_solve, triang_solve,                   &
                inv, comp_inv,                              &
                sym_inv, comp_sym_inv,                      &
                comp_sym_ginv,                              &
                comp_triang_inv, comp_uut_ltl,              &
                det, comp_det,                              &
                sym_trid_cmp, sym_trid_cmp2, sym_trid_solve
!
! GENERIC INTERFACES FOR ROUTINES WITH OVERLOADED VERSIONS
! ========================================================
!
!
    interface solve_lin
        module procedure    solve_lin_rv, solve_lin_rm
    end interface
!
    interface lin_lu_solve
        module procedure    lin_lu_solve_rv,   lin_lu_solve_rm,    &
                            lin_lu_solve_b_rv, lin_lu_solve_b_rm
    end interface
!
    interface lu_solve
        module procedure    lu_solve_rv, lu_solve_rm
    end interface
!
    interface lu_solve2
        module procedure    lu_solve2_rv, lu_solve2_rm
    end interface
!
    interface chol_solve
        module procedure    chol_solve_rv, chol_solve_rm
    end interface
!
    interface triang_solve
        module procedure    triang_solve_rv,   triang_solve_rm,    &
                            triang_solve_b_rv, triang_solve_b_rm
    end interface
!
    interface comp_inv
        module procedure    comp_inv_a, comp_inv_b
    end interface
!
    interface comp_sym_inv
        module procedure    comp_sym_inv_a, comp_sym_inv_b
    end interface
!
    interface comp_sym_ginv
        module procedure    comp_sym_ginv_a, comp_sym_ginv_b
    end interface
!
    interface comp_triang_inv
        module procedure    comp_triang_inv_a, comp_triang_inv_b
    end interface
!
    interface comp_uut_ltl
        module procedure    comp_uut_ltl_a, comp_uut_ltl_b
    end interface
!
    interface sym_trid_solve
        module procedure    sym_trid_solve_a, sym_trid_solve_b
    end interface
!
!   
! MODULE PARAMETERS 
! =================
!
! BLOCK SIZE FOR THE LU AND CHOLESKY ALGORITHMS IN lu_cmp, lu_solve_rv, lu_solve_rm, lin_lu_solve_rv,
! lin_lu_solve_rm, lin_lu_solve_b_rv, lin_lu_solve_b_rm, chol_cmp, gchol_cmp, comp_det SUBROUTINES AND
! det FUNCTION.
!
    integer(i4b), parameter :: blksz = blksz_lin
!    
!
! =========================================================================================
!
                                  contains
!                                 ========
!
! =========================================================================================
!                         LU, CHOLESKY DECOMPOSITIONS
! =========================================================================================
!
!
    subroutine lu_cmp( mat, ip, d1, d2, tol, small )
!
! Purpose
! _______
!
!   LU_CMP computes the LU decomposition with partial pivoting and implicit
!   row scaling of a given n-by-n real matrix MAT
!
!                           P * MAT = L * U
!
!   where P is a permutation matrix, L is a n-by-n unit lower triangular matrix and
!   U is a n-by-n upper triangular matrix. P is a permutation matrix, stored in
!   argument IP, such that
!
!                        P = P(n) * ... * P(1)
!
!   with  P(i) is the identity with row i and IP(i) interchanged.
!
!
! Arguments
! _________
!
!   MAT     (INPUT/OUTPUT) real(stnd), dimension(:,:)
!           On entry, the coefficient matrix MAT.
!
!           On exit, MAT is replaced by the LU decomposition of
!           a rowwise permutation of MAT. The unit diagonal of L
!           is not stored. For solving efficiency, the diagonal 
!           reciprocals of the matrix U are saved in the diagonal
!           entries of MAT.
!
!           The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   IP      (OUTPUT) integer(i4b), dimension(:)
!           On exit, IP records the permutations effected by the partial
!           pivoting.
!
!           The size of IP must verify:  size( IP ) = n .
!
!   D1      (OUTPUT) real(stnd)
!           On exit, if D2 is absent:
!
!           - D1 = +1, if an even number of interchanges was carried out
!           - D1 = -1, if an odd number of interchanges was carried out
!           - D1 = 0,  if MAT is algorithmically singular.
!
!           On exit, if D2 is present, D1 is the first component of the determinant
!           of MAT (mantissa of determinant).
!
!   D2      (OUTPUT, OPTIONAL) integer(i4b)
!           If D2 is present, the two components of the determinant of 
!           MAT are computed.
!
!           On exit, D2 is the second component of the determinant of MAT
!           (exponent of determinant):
!
!                      determinant(MAT) = scale( D1, D2 )
!
!   TOL     (INPUT, OPTIONAL) real(stnd)
!           On entry, a relative tolerance used to indicate whether or not MAT
!           is nearly singular. Tol should normally be choose as approximately
!           the largest relative error in the elements of MAT. For example, 
!           if the elements of MAT are correct to about 4 significant figures,
!           then TOL should be set to about 5 * 10**(-4).
!
!           If TOL is supplied as less than EPS, where EPS is the relative machine
!           precision, then the value EPS is used in place of TOL.
!
!           Default: EPS, the relative machine precision.
!
!   SMALL   (INPUT, OPTIONAL) real(stnd)
!           On entry, if the system is singular, replaces a diagonal term of the 
!           matrix U if it is smaller in magnitude than the value SMALL using 
!           the same sign as the diagonal term. An approximate solution based 
!           on this replacement can be obtained if no overflow results.
!
!           If SMALL is supplied as less than SAFMIN, the smallest number that can be 
!           reciprocated safely, then the value SAFMIN is used in place of SMALL.
!
!           Default: SAFMIN, the smallest number that can be reciprocated safely.
!
!
! Further Details
! _______________
!
!   MAT is declared singular if a diagonal element of U is such that
!
!             abs( U(j,j) ) <= n * norm( MAT(j,:) ) * TOL
!
!   where norm( MAT(j,:) ) denotes the maximum of the absolute values of 
!   the jth row of the matrix MAT. In this case, a diagonal element of U is
!   small, indicating that MAT is singular or nearly singular and D1 is set
!   to zero.
!
!   If D1/=0 then the linear system MAT * X = B can be solved with subroutines
!   LU_SOLVE or LU_SOLVE2.
!
!   If MAT is algorithmically singular (D1=0), the diagonal terms of U smaller 
!   in magnitude than the value SMALL have been replaced by SMALL, using the
!   same sign as the diagonal terms and the decomposition has been completed. 
!   An approximate solution based on this replacement can be obtained if
!   no overflow results.
!
!   A blocked algorithm is used to compute the factorization. Furthermore,
!   the computations are parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Higham, N.J., 2011:
!         Gaussian elimination.
!         Wiley Interdisciplinary Reviews: Computational Statistics,
!         Vol. 3, Issue 3, pp 230-238.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
!    
#ifdef _OPENMP
    use Char_Constants, only : allocate_error
#endif
    use Num_Constants,  only : lamch, machbase
    use Char_Constants, only : lin_error1
    use Utilities,      only : assert_eq, merror, imaxloc
#ifdef _BLAS
    use BLAS_interfaces, only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout)         :: mat(:,:)
    real(stnd),   intent(out)           :: d1
    real(stnd),   intent(in), optional  :: tol, small
!
    integer(i4b), intent(out)           :: ip(:)
    integer(i4b), intent(out), optional :: d2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                       :: piv, tol_piv, safmin
    real(stnd), dimension(size(ip))  :: tmp_row, equil
!
    integer(i4b)  :: n, i, imax, j, j1, j2, k, m, nblk, rem, n2
!
    logical(lgl)  :: comp_det
!
#ifdef _OPENMP
    integer      :: iok
    integer(i4b) :: l, l1, l2, iblk, num_threads, thread_num
    integer(kind=omp_lock_kind), dimension(:), allocatable :: lck
!
    logical(lgl) :: do_lock
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    real(stnd), parameter :: invmachbase=one/machbase
!
    character(len=*),  parameter :: name_proc='lu_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(ip),i4b)    ,   &
                    name_proc )
!
    d1 = one
    if ( present( d2 ) ) then
        comp_det = true
        d2 = 0_i4b
    else
        comp_det = false
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!    
!   INITIALIZE TOLERANCE, SAFE MINIMUM AND IMPLICIT SCALING VECTOR.
!    
    tol_piv = lamch( 'u' )
    if ( present( tol ) ) tol_piv = max( tol, tol_piv )
    tol_piv = tol_piv*real( n, stnd )
!    
    safmin = lamch( 's' )
!    
    tmp_row(:n) = maxval( abs(mat(:n,:n)), dim=2 )
    if ( any( tmp_row(:n)<safmin ) ) call merror( name_proc//lin_error1 )
    equil(:n) = one/tmp_row(:n)
!    
    if ( present( small ) ) safmin = max( small, safmin )
!
!   DETERMINE THE NUMBER OF BLOCKS.
!
    nblk = n/blksz
    rem  = mod( n, blksz )
!
    if ( rem/=0_i4b ) then
!
        nblk = nblk + 1_i4b
        n2   = n - rem
!
    else
!
        n2   = n - blksz
!
    end if
!   
!   TRIANGULATION.  
!    
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*n>=omp_limit                      .and.      &
                  nblk>1_i4b                          .and.      &
                  i>1_i4b
!
    if ( test_par ) then
!
!       PARALLEL AND BLOCKED LU DECOMPOSITION OF mat.
!
!       ALLOCATE WORK ARRAY FOR STORING THE LOCKS FOR EACH BLOCK.
!
        allocate( lck(nblk), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SET UP LOCKS FOR EACH BLOCK.
!
        do i = 1_i4b, nblk
            call omp_init_lock( lck(i) )
        end do
!
!       COMPUTE LU FACTORIZATION OF THE FIRST BLOCK IN PLACE.
!
        do i = 1_i4b, blksz
!
!           FIND PIVOT.
!
            tmp_row(i:n) = equil(i:n)*abs( mat(i:n,i) )
            imax         = i - 1_i4b + imaxloc( tmp_row(i:n) )
            piv          = mat(imax,i)
!
!           TEST FOR SINGULARITY.
!
            if ( tmp_row(imax)<=tol_piv ) then
!
!               mat IS SINGULAR.
!
                d1 = zero
!
                if ( comp_det ) then
                    d2       = 0_i4b
                    comp_det = false
                end if
!
!               REPLACE PIVOT ELEMENT BY SAFE MINIMUM.
!
                piv = sign( max( safmin, abs(piv) ), piv )
!
            end if
!
!           IF NECESSARY, APPLY ROW INTERCHANGE.
!
            if ( imax/=i ) then
!
                d1 = -d1
!
                tmp_row(1_i4b:blksz)  = mat(imax,1_i4b:blksz)
                mat(imax,1_i4b:blksz) = mat(i,1_i4b:blksz)
                mat(i,1_i4b:blksz)    = tmp_row(1_i4b:blksz)
!
                equil(imax) = equil(i)
!
            else
!
                tmp_row(i+1_i4b:blksz) = mat(i,i+1_i4b:blksz)
!
            end if
!
!           STORE PERMUTATION.
!
            ip(i) = imax
!
!           COMPUTE DETERMINANT IF REQUIRED.
!
            if ( comp_det ) then
!
                d1 = d1*piv
!
                do
!
                    if ( abs(d1)<=one ) exit
!
                    d1 = d1*invmachbase
                    d2 = d2 + 1_i4b
!
                end do
!
                do
!
                    if ( abs(d1)>=invmachbase ) exit
!
                    d1 = d1*machbase
                    d2 = d2 - 1_i4b
!
                end do
!
            end if
!
            piv              = one/piv
            mat(i,i)         = piv
            mat(i+1_i4b:n,i) = mat(i+1_i4b:n,i)*piv
!
!           REDUCE REMAINING MATRIX.
!
            do k = i+1_i4b, blksz
!
#ifdef _BLAS
                call axpy( n-i, -tmp_row(k), mat(i+1_i4b:n,i), 1_i4b, mat(i+1_i4b:n,k), 1_i4b )
#else
                mat(i+1_i4b:n,k) = mat(i+1_i4b:n,k) - tmp_row(k)*mat(i+1_i4b:n,i)
#endif
!
            end do
!
        end do
!
!$OMP PARALLEL PRIVATE(i,imax,j,j1,j2,k,l,l1,l2,m,iblk,thread_num,num_threads,piv,do_lock)
!
        num_threads = omp_get_num_threads()
        thread_num  = omp_get_thread_num()
!
!       SET THE LOCKS FOR THE BLOCKS.
!
!$OMP DO SCHEDULE(STATIC,1)
!
        do i = 1_i4b, nblk
            call omp_set_lock( lck(i) )
        end do
!
!$OMP END DO
!
        if ( thread_num==0_i4b ) then
!
!           RELEASE THE FIRST LOCK.
!
            call omp_unset_lock( lck(1) )     
!
            do_lock = false
!
        else
!
            do_lock = true
!
        end if
!
!       START THE BLOCKED LU FACTORIZATION OF mat.
!
        do j = 0_i4b, n2, blksz
!
            iblk = j/blksz + 1_i4b
!
            if ( do_lock ) then
!
                call omp_set_lock( lck(iblk) )
!
! $OMP FLUSH (mat,d1,d2,comp_det,equil,ip)
!
                call omp_unset_lock( lck(iblk) )
!
            else
!
                do_lock = true
!
            end if
!
            j1 = j + 1_i4b
            j2 = min( j + blksz, n )
!
!           APPLY TRANSFORMATIONS IN THE iblk BLOCK TO LATER COLUMNS, EG TO mat(1:n,j2+1:n).
!
            k = ( (iblk/num_threads)*num_threads ) + thread_num
            if ( thread_num<mod(iblk,num_threads) ) k = k + num_threads
!
            do l = k, nblk-1_i4b, num_threads
!
!               COMPUTE INDICES FOR THE FIRST AND LAST COLUMNS OF THE l+1 BLOCK.
!
                l1 = l*blksz + 1_i4b
                l2 = min( (l+1_i4b)*blksz, n )
!
                do i = l1, l2
!
!                   APPLY PERMUTATIONS TO LATER COLUMNS, EG TO mat(j1:n,j2+1:n).
!
                    do m = j1, j2
!
                        k        = ip(m)
                        piv      = mat(k,i)
                        mat(k,i) = mat(m,i)
                        mat(m,i) = piv
!
                    end do
!    
!                   APPLY TRANSFORMATIONS TO LATER COLUMNS, EG TO mat(j1:n,j2+1:n).
!    
                    do m = j1, j2
!
#ifdef _BLAS
                        call axpy( n-m, -mat(m,i), mat(m+1_i4b:n,m), 1_i4b, mat(m+1_i4b:n,i), 1_i4b )
#else
                        mat(m+1_i4b:n,i) = mat(m+1_i4b:n,i) - mat(m,i)*mat(m+1_i4b:n,m)
#endif
!
                    end do
!
                end do
!
                if ( l==iblk ) then
!
!                   COMPUTE THE LU FACTORIZATION OF THE iblk+1 BLOCK, EG OF mat(j2+1:n,j2+1:j2+blksz), IN PLACE.
!
                    do i = l1, l2
!
!                       FIND PIVOT.
!
                        tmp_row(i:n) = equil(i:n)*abs( mat(i:n,i) )
                        imax         = i - 1_i4b + imaxloc( tmp_row(i:n) )
                        piv          = mat(imax,i)
!
!                       TEST FOR SINGULARITY.
!
                        if ( tmp_row(imax)<=tol_piv ) then
!
!                           mat IS SINGULAR.
!
                            d1 = zero
!
                            if ( comp_det ) then
                                d2       = 0_i4b
                                comp_det = false
                            end if
!
!                           REPLACE PIVOT ELEMENT BY SAFE MINIMUM.
!
                            piv = sign( max( safmin, abs(piv) ), piv )
!
                        end if
!
!                       IF NECESSARY, APPLY ROW INTERCHANGE.
!
                        if ( imax/=i ) then
!
                            d1 = -d1
!
                            tmp_row(l1:l2)  = mat(imax,l1:l2)
                            mat(imax,l1:l2) = mat(i,l1:l2)
                            mat(i,l1:l2)    = tmp_row(l1:l2)
!
                            equil(imax) = equil(i)
!
                        else
!
                            tmp_row(i+1_i4b:l2) = mat(i,i+1_i4b:l2)
!
                        end if
!
!                       STORE PERMUTATION.
!
                        ip(i) = imax
!
!                       COMPUTE DETERMINANT IF REQUIRED.
!
                        if ( comp_det ) then
!
                            d1 = d1*piv
!
                            do
!
                                if ( abs(d1)<=one ) exit
!
                                d1 = d1*invmachbase
                                d2 = d2 + 1_i4b
!
                            end do
!
                            do
!
                                if ( abs(d1)>=invmachbase ) exit
!
                                d1 = d1*machbase
                                d2 = d2 - 1_i4b
!
                            end do
!
                        end if
!
                        piv              = one/piv
                        mat(i,i)         = piv
                        mat(i+1_i4b:n,i) = mat(i+1_i4b:n,i)*piv
!
!                       REDUCE REMAINING COLUMNS IN THE iblk+1 BLOCK.
!
                        do k = i+1_i4b, l2
!
#ifdef _BLAS
                            call axpy( n-i, -tmp_row(k), mat(i+1_i4b:n,i), 1_i4b, mat(i+1_i4b:n,k), 1_i4b )
#else
                            mat(i+1_i4b:n,k) = mat(i+1_i4b:n,k) - tmp_row(k)*mat(i+1_i4b:n,i)
#endif
!
                        end do
!
                    end do
!
!                   RELEASE THE iblk+1_i4b LOCK.
!
! $OMP FLUSH (mat,d1,d2,comp_det,equil,ip)
!
                    call omp_unset_lock( lck(iblk+1_i4b) )                 
!
                    do_lock = false
!
                end if
!
            end do
!
        end do
!
!       APPLY PENDING PERMUTATIONS TO THE BLOCKS IN THE LOWER TRIANGLE OF mat.
!
!$OMP DO SCHEDULE(STATIC,1)
!
!       APPLY PENDING PERMUTATIONS TO THE BLOCKS IN THE LOWER TRIANGLE OF mat.
!
        do j = 0_i4b, n2-blksz, blksz
!
            j1 = j + 1_i4b
            j2 = j + blksz 
!
!           APPLY PERMUTATIONS TO mat(j2+1:n,j1:j2).
!
            do i = j1, j2
!
                do m = j2+1_i4b, n
!
                    k        = ip(m)
                    piv      = mat(k,i)
                    mat(k,i) = mat(m,i)
                    mat(m,i) = piv
!
                end do
!
            end do
!
!            do m = j2+1_i4b, n
!
!                k              = ip(m)
!                tmp_row(j1:j2) = mat(k,j1:j2)
!                mat(k,j1:j2)   = mat(m,j1:j2)
!                mat(m,j1:j2)   = tmp_row(j1:j2)
!
!            end do
!
        end do
!
!$OMP END DO NOWAIT
!
!$OMP END PARALLEL
!
!       DESTROY THE LOCKS.
!
        do i = 1_i4b, nblk
            call omp_destroy_lock( lck(i) )
        end do
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( lck )
!
    else
!
#endif
!
!       BLOCKED LU DECOMPOSITION OF mat.
!
        do j = 0_i4b, n2, blksz
!
            j1 = j + 1_i4b
            j2 = min( j + blksz, n )
!
!           COMPUTE THE LU FACTORIZATION OF mat(j1:n,j1:j2) IN PLACE.
!
            do i = j1, j2
!
!               FIND PIVOT.
!
                tmp_row(i:n) = equil(i:n)*abs( mat(i:n,i) )
                imax         = i - 1_i4b + imaxloc( tmp_row(i:n) )
                piv          = mat(imax,i)
!
!               TEST FOR SINGULARITY.
!
                if ( tmp_row(imax)<=tol_piv ) then
!
!                   mat IS SINGULAR.
!
                    d1 = zero
!
                    if ( comp_det ) then
                        d2       = 0_i4b
                        comp_det = false
                    end if
!
!                   REPLACE PIVOT ELEMENT BY SAFE MINIMUM.
!
                    piv = sign( max( safmin, abs(piv) ), piv )
!
                end if
!
!               IF NECESSARY, APPLY ROW INTERCHANGE.
!
                if ( imax/=i ) then
!
                    d1 = -d1
!
                    tmp_row(j1:j2)  = mat(imax,j1:j2)
                    mat(imax,j1:j2) = mat(i,j1:j2)
                    mat(i,j1:j2)    = tmp_row(j1:j2)
!
                    equil(imax) = equil(i)
!
                else
!
                    tmp_row(i+1_i4b:j2) = mat(i,i+1_i4b:j2)
!
                end if
!
!               STORE PERMUTATION.
!
                ip(i) = imax
!
!               COMPUTE DETERMINANT IF REQUIRED.
!
                if ( comp_det ) then
!
                    d1 = d1*piv
!
                    do
!
                        if ( abs(d1)<=one ) exit
!
                        d1 = d1*invmachbase
                        d2 = d2 + 1_i4b
!
                    end do
!
                    do
!
                        if ( abs(d1)>=invmachbase ) exit
!
                        d1 = d1*machbase
                        d2 = d2 - 1_i4b
!
                    end do
!
                end if
!
                piv              = one/piv
                mat(i,i)         = piv
                mat(i+1_i4b:n,i) = mat(i+1_i4b:n,i)*piv
!
!               REDUCE REMAINING MATRIX.
!
                do k = i+1_i4b, j2
!
#ifdef _BLAS
                    call axpy( n-i, -tmp_row(k), mat(i+1_i4b:n,i), 1_i4b, mat(i+1_i4b:n,k), 1_i4b )
#else
                    mat(i+1_i4b:n,k) = mat(i+1_i4b:n,k) - tmp_row(k)*mat(i+1_i4b:n,i)
#endif
!
                end do
!
            end do
!
!           APPLY PERMUTATION TO PRIOR COLUMNS, EG TO mat(1:n,1:j).
!
!            do l = 0_i4b, j-blksz, blksz
!
!                l1 = l + 1_i4b
!                l2 = l + blksz
!
!                do i = j1, j2
!
!                    k              = ip(i)
!                    tmp_row(l1:l2) = mat(k,l1:l2)
!                    mat(k,l1:l2)   = mat(i,l1:l2)
!                    mat(i,l1:l2)   = tmp_row(l1:l2)
!
!                end do
!
!            end do
!
!           APPLY TRANSFORMATIONS TO LATER COLUMNS, EG TO mat(j1:n,j2+1:n).
!
            do i = j2+1_i4b, n
!
!               APPLY PERMUTATION TO LATER COLUMNS, EG TO mat(j1:n,j2+1:n).
!
                do m = j1, j2
!
                    k        = ip(m)
                    piv      = mat(k,i)
                    mat(k,i) = mat(m,i)
                    mat(m,i) = piv
!
                end do
!    
!               APPLY TRANSFORMATIONS TO LATER COLUMNS, EG TO mat(j1:n,j2+1:n).
!    
                do m = j1, j2
!
#ifdef _BLAS
                    call axpy( n-m, -mat(m,i), mat(m+1_i4b:n,m), 1_i4b, mat(m+1_i4b:n,i), 1_i4b )
#else
                    mat(m+1_i4b:n,i) = mat(m+1_i4b:n,i) - mat(m,i)*mat(m+1_i4b:n,m)
#endif
!
                end do
!
            end do
!          
        end do
!
!       APPLY PENDING PERMUTATIONS TO THE BLOCKS IN THE LOWER TRIANGLE OF mat.
!
        do j = 0_i4b, n2-blksz, blksz
!
            j1 = j + 1_i4b
            j2 = j + blksz 
!
!           APPLY PERMUTATIONS TO mat(j2+1:n,j1:j2).
!
            do i = j1, j2
!
                do m = j2+1_i4b, n
!
                    k        = ip(m)
                    piv      = mat(k,i)
                    mat(k,i) = mat(m,i)
                    mat(m,i) = piv
!
                end do
!
            end do
!
!            do m = j2+1_i4b, n
!
!                k              = ip(m)
!                tmp_row(j1:j2) = mat(k,j1:j2)
!                mat(k,j1:j2)   = mat(m,j1:j2)
!                mat(m,j1:j2)   = tmp_row(j1:j2)
!
!            end do
!
        end do
!
#ifdef _OPENMP
    end if
#endif
!
!
! END OF SUBROUTINE lu_cmp
! ________________________
!
    end subroutine lu_cmp
!
! =========================================================================================
!
    subroutine lu_cmp2( mat, ip, d1, d2, b, matinv, tol, small )
!
! Purpose
! _______
!
!   LU_CMP2 computes the LU decomposition with partial pivoting and implicit
!   row scaling of a given n-by-n real matrix MAT
!
!                           P * MAT = L * U
!
!   where P is a permutation matrix, L is a n-by-n unit lower triangular matrix and
!   U is a n-by-n upper triangular matrix. P is a permutation matrix, stored in
!   argument IP, such that
!
!                        P = P(n) * ... * P(1)
!
!   with  P(i) is the identity with row i and IP(i) interchanged.
!
!
!   If D2 is present, LU_CMP2 computes the determinant of MAT as
!
!                      determinant(MAT) = scale( D1, D2 )
!
!
!   If B is present, LU_CMP2 solves the system of linear equations
!
!                       MAT * X = B
!
!   using the LU factorization with scaled partial pivoting of MAT.
!   Here B is a n-vector.
!
!
!   If MATINV is present, LU_CMP2 computes the inverse of MAT
!
!                       MATINV = MAT**(-1)
!
!
! Arguments
! _________
!
!   MAT     (INPUT/OUTPUT) real(stnd), dimension(:,:)
!           On entry, the coefficient matrix MAT.
!
!           On exit, MAT is replaced by the LU decomposition of
!           a rowwise permutation of MAT. The unit diagonal of L
!           is not stored. For solving efficiency, the diagonal 
!           reciprocals of the matrix U are saved in the diagonal
!           entries of MAT.
!
!           The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   IP      (OUTPUT) integer(i4b), dimension(:)
!           On exit, IP records the permutations effected by the partial
!           pivoting.
!
!           The size of IP must verify:  size( IP ) = n .
!
!   D1      (OUTPUT) real(stnd)
!           On exit, if D2 is absent:
!
!           - D1 = +1, if an even number of interchanges was carried out
!           - D1 = -1, if an odd number of interchanges was carried out
!           - D1 = 0,  if MAT is algorithmically singular.
!
!           On exit, if D2 is present, D1 is the first component of the determinant
!           of MAT (mantissa of determinant).
!
!   D2      (OUTPUT, OPTIONAL) integer(i4b)
!           If D2 is present, the components of the determinant of 
!           MAT are computed.
!
!           On exit, D2 is the second component of the determinant of MAT
!           (exponent of determinant) :
!
!                      determinant(MAT) = scale( D1, D2 )
!
!   B       (INPUT/OUTPUT, OPTIONAL) real(stnd), dimension(:)
!           On entry, the right hand side vector B.
!
!           On exit, the solution vector X.
!
!           The shape of B must verify:  size( B ) = n .
!
!   MATINV  (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!           On exit, if MAT is not singular, MATINV contains the inverse of MAT.
!
!           The shape of MATINV must verify:  size( MATINV, 1 ) = size( MATINV, 2 ) = n .
!
!   TOL     (INPUT, OPTIONAL) real(stnd)
!           On entry, a relative tolerance used to indicate whether or not MAT
!           is nearly singular. Tol should normally be choose as approximately
!           the largest relative error in the elements of MAT. For example, 
!           if the elements of MAT are correct to about 4 significant figures,
!           then TOL should be set to about 5 * 10**(-4). If TOL is supplied as
!           less than EPS, where EPS is the relative machine precision, then
!           the value EPS is used in place of TOL.
!
!           Default: EPS, the relative machine precision.
!
!   SMALL   (INPUT, OPTIONAL) real(stnd)
!           On entry, if the system is singular, replaces a diagonal term of the 
!           matrix U if it is smaller in magnitude than the value SMALL using 
!           the same sign as the diagonal term. An approximate solution for X based 
!           on this replacement can be obtained if no overflow results. If SMALL
!           is supplied as less than SAFMIN, the smallest number that can be 
!           reciprocated safely, then the value SAFMIN is used in place of SMALL.
!
!           Default: SAFMIN, the smallest number that can be reciprocated safely.
!
!
! Further Details
! _______________
!
!   MAT is declared singular if a diagonal element of U is such that
!
!             abs( U(j,j) ) <= n * norm( MAT(j,:) ) * TOL
!
!   where norm( MAT(j,:) ) denotes the maximum of the absolute values of 
!   the jth row of the matrix MAT. In this case, a diagonal element of U is
!   small, indicating that MAT is singular or nearly singular and D1 is set
!   to zero.
!
!   If MAT is algorithmically singular (D1=0), the diagonal terms of U 
!   smaller in magnitude than the value SMALL have been replaced by SMALL, using
!   the same sign as the diagonal terms, and the decomposition has been completed. 
!   An approximate solution for X based on this replacement is then obtained if 
!   no overflow results and MATINV is filled with nan() value.
!
!   If D1/=0 then the linear system MAT * Z = D can be solved with subroutines
!   LU_SOLVE or LU_SOLVE2.
!
!   The computations are parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Higham, N.J., 2011:
!         Gaussian elimination.
!         Wiley Interdisciplinary Reviews: Computational Statistics,
!         Vol. 3, Issue 3, pp 230-238.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants, only : nan
    use Utilities,     only : assert_eq, assert, unit_matrix
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout)            :: mat(:,:)
    real(stnd),   intent(out)              :: d1
    real(stnd),   intent(inout), optional  :: b(:)
    real(stnd),   intent(out),   optional  :: matinv(:,:)
    real(stnd),   intent(in),    optional  :: tol, small
!
    integer(i4b), intent(out)             :: ip(:)
    integer(i4b), intent(out),   optional :: d2
!
    logical(lgl) :: comp_b, comp_matinv
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: n
!
#ifdef _OPENMP
    integer(i4b) :: i
!
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='lu_cmp2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(ip),i4b)    ,   &
                    name_proc )
!
    comp_b      = present( b )
    if ( comp_b ) call assert( logical( n==size(b), lgl ),        &
                               name_proc )
!
    comp_matinv = present( matinv )
    if ( comp_matinv ) call assert( logical( n==size(matinv,1), lgl ),   &
                                    logical( n==size(matinv,2), lgl ),   &
                                    name_proc )
!    
!   FIRST DECOMPOSE mat .
!    
    call lu_cmp( mat, ip, d1, d2=d2, tol=tol, small=small )
!    
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.     &
                  comp_b                              .and.     &
                  comp_matinv                         .and.     &
                  i>1_i4b
#endif
!    
!$OMP PARALLEL IF( test_par ), NUM_THREADS( 2 )
!    
!$OMP SECTIONS
!    
!$OMP SECTION
!    
    if ( comp_b ) then
!    
!       SOLVE mat*X = b, OVERWRITING b WITH X .
!    
        call lu_solve_rv( mat, ip, b )
!
    end if
!    
!$OMP SECTION
!    
    if ( comp_matinv ) then
!
        if ( d1==zero ) then
!    
!           mat IS SINGULAR.
!
            matinv(:n,:n) = nan()
!
        else
!    
!           mat IS REGULAR, COMPUTE INVERSE OF mat.
!    
!           FIRST SET matinv TO IDENTITY MATRIX.
!
            call unit_matrix( matinv )
!    
!           CALL ROUTINE lu_solve TO SOLVE MAT*MATINV = I .
!    
            call lu_solve_rm( mat, ip, matinv )
!            call lu_solve2_rm( mat, ip, matinv )
!
        end if
!
    end if
!    
!$OMP END SECTIONS
!
!$OMP END PARALLEL
!
!
! END OF SUBROUTINE lu_cmp2
! _________________________
!
    end subroutine lu_cmp2
!
! =========================================================================================
!
    subroutine chol_cmp( mat, invdiag, d1, d2, upper, tol )
!
! Purpose
! _______
!
!   CHOL_CMP computes the Cholesky factorization of a n-by-n real symmetric
!   positive definite matrix MAT. The factorization has the form
!
!         MAT = U' * U , if UPPER=true or is absent,
!
!   and
!
!         MAT = L * L' , if UPPER=false,
!
!   where U is an upper triangular matrix and L is a lower triangular matrix.
!
!
! Arguments
! _________
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the symmetric positive definite matrix MAT. If:
!
!                - UPPER = true or is absent:
!                  The leading n-by-n upper triangular part of MAT contains the
!                  upper triangular part of the matrix and the strictly
!                  lower triangular part of MAT is not referenced.
!                - UPPER = false:
!                  The leading n-by-n lower triangular part of MAT contains the
!                  lower triangular part of the matrix and the strictly
!                  upper triangular part of MAT is not referenced.
!
!                On exit, if D1/=0, the factor U or L from the Cholesky
!                factorization  MAT = U' * U  or  MAT = L * L', except for
!                the main diagonal elements which are stored in reciprocal
!                form in INVDIAG. The main diagonal elements of MAT are not
!                modified.
!
!                The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   INVDIAG      (OUTPUT) real(stnd), dimension(:)
!                On exit, INVDIAG contains the reciprocals of the actual
!                diagonal elements of L or U.
!
!                The size of INVDIAG must verify:  size( INVDIAG ) = n .
!
!   D1           (OUTPUT) real(stnd)
!                On exit, D1 = zero indicates that the matrix MAT is algorithmically
!                not positive definite and that the factorization can not be
!                completed. Any other value indicates successful exit.
!
!                On exit, if D2 is present, D1 is the first component of the 
!                determinant (mantissa of determinant) of MAT.
!
!   D2           (OUTPUT, OPTIONAL) integer(i4b)
!                If D2 is present, the components of the determinant 
!                of MAT are computed.
!
!                On exit, D2 is the second component of the determinant of MAT
!                (exponent of determinant):
!
!                           determinant(MAT) = scale( D1, D2 )
!
!   UPPER        (INPUT, OPTIONAL) logical(lgl)
!                Specifies whether the upper or lower triangular part of the
!                symmetric matrix MAT is stored. If:
!
!                - UPPER = true :  Upper triangular is stored
!                - UPPER = false:  Lower triangular is stored.
!
!                The default is true.
!
!   TOL          (INPUT, OPTIONAL) real(stnd)
!                Tolerance used to test the matrix for positive-definiteness.
!                TOL is used as a multiplying factor for determining effective
!                zero for pivots. TOL must be greater or equal to zero, otherwise
!                the default value is used.
!
!                The default value is the machine precision multiplied by n.
!
!
! Further Details
! _______________
!
!   MAT is declared not positive definite if during the j-th stage of the
!   factorization of MAT, a pivot, PIV(j), is such that
!
!             PIV(j) <= MAT(j,j) * TOL
!
!   In this case, the leading minor of order j of MAT is declared not
!   positive definite and on exit of CHOL_CMP:
!
!   - D1 is set to zero,  
!   - INVDIAG(j) = PIV(j),
!   - INVDIAG(j+1_i4b:n) are set to nan() value,
!
!   and the Cholesky factorization is not completed.
!
!   On the other hand, if MAT is positive definite then
!
!        U(j,j) = sqrt(PIV(j)) (if UPPER=true), for j=1 to n,
!
!   or
!
!        L(j,j) = sqrt(PIV(j)) (if UPPER=false), for j=1 to n,
!
!   and on exit of CHOL_CMP:
!
!   - D1/=0,
!   - INVDIAG(j)=1/sqrt(PIV(j)) for j=1 to n,
!
!   and the linear system MAT * Z = D can be solved with 
!   subroutine CHOL_SOLVE.
!
!   This is a GAxpy version of the Cholesky algorithm, for more details
!   see the reference (1).
!
!   A blocked algorithm is used to compute the factorization. Furthermore,
!   the computations are parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Higham, N.J., 2009:
!         Cholesky factorization.
!         Wiley Interdisciplinary Reviews: Computational Statistics,
!         Vol. 1, pp 251-254.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error
    use Num_Constants,   only : lamch, nan, machbase
    use Utilities,       only : assert_eq, merror
#ifdef _BLAS
    use BLAS_interfaces, only : dot
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:), intent(inout) :: mat
    real(stnd), dimension(:),   intent(out)   :: invdiag
    real(stnd),                 intent(out)   :: d1
    real(stnd), optional,       intent(in)    :: tol
!
    integer(i4b),  intent(out), optional :: d2
!
    logical(lgl), intent(in), optional :: upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                            :: piv, tol_piv
    real(stnd), dimension(:), allocatable :: vec
!
    integer(i4b) :: n, i, j, j1, j2, k, nblk, rem, n2, jfail
    integer      :: iok
!
    logical(lgl) :: comp_det, mat_is_upper, failure
!
#ifdef _OPENMP
    integer(i4b) :: l, l1, l2, m, iblk, num_threads, thread_num
    integer(kind=omp_lock_kind), dimension(:), allocatable :: lck
!
    logical(lgl) :: do_lock
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='chol_cmp'
!
    real(stnd), parameter :: invmachbase=one/machbase
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,    &
                    int(size(mat,2),i4b) ,    &
                    int(size(invdiag),i4b),   &
                    name_proc )
!    
    d1 = one
!
    if ( present( d2 ) ) then
        comp_det = true
        d2 = 0_i4b
    else
        comp_det = false
    end if
!
    mat_is_upper = true
    if ( present(upper) ) mat_is_upper = upper
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!    
!   INITIALIZE TOLERANCE.
!    
    tol_piv = lamch( 'p' )*real( n, stnd )
!
    if ( present( tol ) ) then
        tol_piv = merge( tol, tol_piv, tol>=zero )
    end if
!   
!   CHECK MAGNITUDE OF THE FIRST PIVOT.
!   
    if ( mat(1_i4b,1_i4b)<=lamch( 's' ) ) then
!
        invdiag(1_i4b)   = mat(1_i4b,1_i4b)
        invdiag(2_i4b:n) = nan()
!
        d1 = zero
        if ( comp_det ) d2 = 0_i4b
!
        failure = true
!
        return
!
    else
!    
        failure = false
!
    end if
!
!   COMPUTE THE NUMBER OF BLOCKS.
!
    nblk = n/blksz
    rem  = mod( n, blksz )
!
    if ( rem/=0_i4b ) then
!
        nblk = nblk + 1_i4b
        n2   = n - rem
!
    else
!
        n2   = n - blksz
!
    end if
!
!   COMPUTE THE CHOLESKY FACTORIZATION mat = U'*U OR mat = L*L'.
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*n>=omp_limit                      .and.      &
                  nblk>1_i4b                          .and.      &
                  i>1_i4b
!
    if ( test_par ) then
!
!       PARALLEL AND BLOCKED CHOLESKY DECOMPOSITION OF mat.
!
!       ALLOCATE WORK ARRAY FOR STORING THE LOCKS FOR EACH BLOCK.
!
        allocate( lck(nblk), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SET UP LOCKS FOR EACH BLOCK.
!
        do i = 1_i4b, nblk
            call omp_init_lock( lck(i) )
        end do
!
!$OMP PARALLEL PRIVATE(i,j,j1,j2,k,l,l1,l2,m,iblk,iok,piv,vec,num_threads,thread_num,do_lock)
!
        num_threads = omp_get_num_threads()
        thread_num  = omp_get_thread_num()
!
!       SET THE LOCKS FOR THE BLOCKS.
!
!$OMP DO SCHEDULE(STATIC,1)
!
        do i = 1_i4b, nblk
            call omp_set_lock( lck(i) )
        end do
!
!$OMP END DO
!
        if ( .not.mat_is_upper ) then
!
!           ALLOCATE WORK SPACE FOR A FAST FACTORIZATION
!           IF THE LOWER TRIANGULAR PART OF mat IS STORED.
!
            allocate( vec(n), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           TRANSPOSE THE INPUT MATRIX.
!
!$OMP DO SCHEDULE(STATIC)
!
            do j = 1_i4b, n-1_i4b
                vec(j+1_i4b:n)   = mat(j,j+1_i4b:n)
                mat(j,j+1_i4b:n) = mat(j+1_i4b:n,j)
                mat(j+1_i4b:n,j) = vec(j+1_i4b:n)
            end do
!
!$OMP END DO
!
        end if
!
        if ( thread_num==0_i4b ) then
!
!           COMPUTE CHOLESKY FACTORIZATION OF THE FIRST BLOCK IN PLACE.
!
            do k = 1_i4b, blksz
!
!               START COMPUTATION OF ELEMENTS k TO blksz OF ROW k OF U.
!
                do i = k, blksz       
#ifdef _BLAS
                    invdiag(i) = mat(k,i) - dot( k-1_i4b, mat(1_i4b:k-1_i4b,k), 1_i4b, mat(1_i4b:k-1_i4b,i), 1_i4b )
#else
                    invdiag(i) = mat(k,i) - dot_product( mat(1_i4b:k-1_i4b,k), mat(1_i4b:k-1_i4b,i) )
#endif
                end do
!
                piv = invdiag(k)
!
!               TEST FOR POSITIVE-DEFINITENESS.
!
                if ( piv<=tol_piv*mat(k,k) ) then
!
!                   mat IS NOT POSITIVE-DEFINITE.
!
                    failure = true
!
                    jfail = k
!
                    exit
!
                end if
!
!               COMPUTE DETERMINANT IF REQUIRED.
!
                if ( comp_det ) then
!
                    d1 = d1*piv
!
                    do
                        if ( abs(d1)<= one ) exit
                        d1 = d1*invmachbase
                        d2 = d2 + 1_i4b
                    end do
!
                    do
                        if ( abs(d1)>= invmachbase ) exit
                        d1 = d1*machbase
                        d2 = d2 - 1_i4b
                    end do
!
                end if
!
!               COMPUTE 1/U(k,k).
!
                piv        = one/sqrt( piv )
                invdiag(k) = piv
!
!               COMPUTE ELEMENTS k+1 TO blksz OF ROW k OF U.
!
                mat(k,k+1_i4b:blksz) = invdiag(k+1_i4b:blksz)*piv
!
            end do
!
!           RELEASE THE FIRST LOCK.
!
! $OMP FLUSH (mat,invdiag,d1,d2,comp_det,failure,jfail)
!
            call omp_unset_lock( lck(1) )     
!
            do_lock = false
!
        else
!
            do_lock = true
!
        end if
!
!       START THE BLOCKED CHOLESKY FACTORIZATION.
!
        do j = 0_i4b, n2, blksz
!
            iblk = j/blksz + 1_i4b
!
            if ( do_lock ) then
!
                call omp_set_lock( lck(iblk) )
!
! $OMP FLUSH (mat,invdiag,d1,d2,comp_det,failure,jfail)
!
                call omp_unset_lock( lck(iblk) )
!
            else
!
                do_lock = true
!
            end if
!
            if ( failure ) exit
!
            j1 = j + 1_i4b
            j2 = min( j + blksz, n )
!
!           APPLY TRANSFORMATIONS IN THE iblk BLOCK TO LATER COLUMNS, EG TO mat(j1:j2,j2+1:n).
!
            k = ( (iblk/num_threads)*num_threads ) + thread_num
            if ( thread_num<mod(iblk,num_threads) ) k = k + num_threads
!
            do l = k, nblk-1_i4b, num_threads
!
!               COMPUTE INDICES FOR THE FIRST AND LAST COLUMNS OF THE l+1 BLOCK.
!
                l1 = l*blksz + 1_i4b
                l2 = min( (l+1_i4b)*blksz, n )
!
                do i = l1, l2
!
                    do m = j1, j2
!
#ifdef _BLAS
                        piv = mat(m,i) - dot( m-1_i4b, mat(1_i4b:m-1_i4b,m), 1_i4b, mat(1_i4b:m-1_i4b,i), 1_i4b )
#else
                        piv = mat(m,i) - dot_product( mat(1_i4b:m-1_i4b,m), mat(1_i4b:m-1_i4b,i) )
#endif
!
!                       COMPUTE ELEMENTS l1 to l2 OF ROW m OF U.
!
                        mat(m,i) = piv*invdiag(m)
!
                    end do
!
                end do
!
                if ( l==iblk ) then
!
!                   COMPUTE THE CHOLESKY FACTORIZATION OF THE iblk+1 BLOCK, EG OF mat(j2+1:j2+blksz,j2+1:j2+blksz), IN PLACE.
!
                    do i = l1, l2
!
!                       START COMPUTATION OF ELEMENTS i TO l2 OF ROW i OF U.
!
                        do m = i, l2         
#ifdef _BLAS
                            invdiag(m) = mat(i,m) - dot( i-1_i4b, mat(1_i4b:i-1_i4b,i), 1_i4b, mat(1_i4b:i-1_i4b,m), 1_i4b )
#else
                            invdiag(m) = mat(i,m) - dot_product( mat(1_i4b:i-1_i4b,i), mat(1_i4b:i-1_i4b,m) )
#endif
                        end do
!
                        piv = invdiag(i)
!
!                       TEST FOR POSITIVE-DEFINITENESS.
!
                        if ( piv<=tol_piv*mat(i,i) ) then
!
!                           mat IS NOT POSITIVE-DEFINITE.
!
                            failure = true
!
                            jfail = i
!
                            exit
!
                        end if
!
!                       COMPUTE DETERMINANT IF REQUIRED.
!
                        if ( comp_det ) then
!
                            d1 = d1*piv
!
                            do
                                if ( abs(d1)<= one ) exit
                                d1 = d1*invmachbase
                                d2 = d2 + 1_i4b
                            end do
!
                            do
                                if ( abs(d1)>= invmachbase ) exit
                                d1 = d1*machbase
                                d2 = d2 - 1_i4b
                            end do
!
                        end if
!
!                       COMPUTE 1/U(i,i).
!
                        piv        = one/sqrt( piv )
                        invdiag(i) = piv
!
!                       COMPUTE ELEMENTS i+1 TO l2 OF ROW i OF U.
!
                        mat(i,i+1_i4b:l2) = invdiag(i+1_i4b:l2)*piv
!
                    end do
!
!                   RELEASE THE iblk+1_i4b LOCK.
!
! $OMP FLUSH (mat,invdiag,d1,d2,comp_det,failure,jfail)
!
                    call omp_unset_lock( lck(iblk+1_i4b) )                 
!
                    do_lock = false
!
                end if
!
            end do
!
        end do
!
        if ( failure ) then
!
!           UNSET THE LOCKS FOR THE REMAINING BLOCKS IF REQUIRED.
!
!$OMP DO SCHEDULE(STATIC,1)
!
            do i = 1_i4b, nblk
!
                if ( i>iblk ) then
                    call omp_unset_lock( lck(i) )
                end if
!
            end do
!
!$OMP END DO
!
        end if
!
!$OMP BARRIER
!
        if ( .not.mat_is_upper ) then
!
!           TRANSPOSE THE OUTPUT MATRIX.
!
!$OMP DO SCHEDULE(STATIC)
!
            do j = 1_i4b, n-1_i4b
                vec(j+1_i4b:n)   = mat(j,j+1_i4b:n)
                mat(j,j+1_i4b:n) = mat(j+1_i4b:n,j)
                mat(j+1_i4b:n,j) = vec(j+1_i4b:n)
            end do
!
!$OMP END DO
!
!           DEALLOCATE WORK SPACE.
!
            deallocate( vec )
!
        end if
!
!$OMP END PARALLEL
!
!       DESTROY THE LOCKS.
!
        do i = 1_i4b, nblk
            call omp_destroy_lock( lck(i) )
        end do
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( lck )
!
    else
!
#endif
!
!       BLOCKED CHOLESKY DECOMPOSITION OF mat.
!
        if ( .not.mat_is_upper ) then
!
!           ALLOCATE WORK SPACE FOR A FAST FACTORIZATION
!           IF THE LOWER TRIANGULAR PART OF mat IS STORED.
!
            allocate( vec(n), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           TRANSPOSE THE INPUT MATRIX.
!
            do j = 1_i4b, n-1_i4b
                vec(j+1_i4b:n)   = mat(j,j+1_i4b:n)
                mat(j,j+1_i4b:n) = mat(j+1_i4b:n,j)
                mat(j+1_i4b:n,j) = vec(j+1_i4b:n)
            end do
!
        end if
!
!       START THE BLOCKED CHOLESKY FACTORIZATION.
!
        j_loop : do j = 0_i4b, n2, blksz
!
            j1 = j + 1_i4b
            j2 = min( j + blksz, n )
!
!           COMPUTE THE CHOLESKY FACTORIZATION OF THE (j*blksz)+1 BLOCK, EG OF mat(j+1:j+blksz,j+1:j+blksz), IN PLACE.
!
            do k = j1, j2
!
!               START COMPUTATION OF ELEMENTS k TO j+blksz OF ROW k OF U.
!
                do i = k, j2         
#ifdef _BLAS
                    invdiag(i) = mat(k,i) - dot( k-1_i4b, mat(1_i4b:k-1_i4b,k), 1_i4b, mat(1_i4b:k-1_i4b,i), 1_i4b )
#else
                    invdiag(i) = mat(k,i) - dot_product( mat(1_i4b:k-1_i4b,k), mat(1_i4b:k-1_i4b,i) )
#endif
                end do
!
                piv = invdiag(k)
!
!               TEST FOR POSITIVE-DEFINITENESS.
!
                if ( piv<=tol_piv*mat(k,k) ) then
!
!                   mat IS NOT POSITIVE-DEFINITE.
!
                    failure = true
!
                    jfail = k
!
                    exit j_loop
!
                end if
!
!               COMPUTE DETERMINANT IF REQUIRED.
!
                if ( comp_det ) then
!
                    d1 = d1*piv
!
                    do
                        if ( abs(d1)<= one ) exit
                        d1 = d1*invmachbase
                        d2 = d2 + 1_i4b
                    end do
!
                    do
                        if ( abs(d1)>= invmachbase ) exit
                        d1 = d1*machbase
                        d2 = d2 - 1_i4b
                    end do
!
                end if
!
!               COMPUTE 1/U(k,k).
!
                piv = one/sqrt( piv )
                invdiag(k) = piv
!
!               COMPUTE ELEMENTS k+1 TO j+blksz OF ROW k OF U.
!
                mat(k,k+1_i4b:j2) = invdiag(k+1_i4b:j2)*piv
!
            end do
!
!           APPLY TRANSFORMATIONS IN THE (j/blksz)+1 BLOCK TO LATER COLUMNS, EG TO mat(j1:j2,j2+1:n).
!
            do i = j2+1_i4b, n
!
                do k = j1, j2
!
#ifdef _BLAS
                    piv = mat(k,i) - dot( k-1_i4b, mat(1_i4b:k-1_i4b,k), 1_i4b, mat(1_i4b:k-1_i4b,i), 1_i4b )
#else
                    piv = mat(k,i) - dot_product( mat(1_i4b:k-1_i4b,k), mat(1_i4b:k-1_i4b,i) )
#endif
!
!                   COMPUTE ELEMENTS j+blksz+1 to n OF ROW k OF U.
!
                    mat(k,i) = piv*invdiag(k)
!
                end do
!          
            end do
!
        end do j_loop
!
        if ( .not.mat_is_upper ) then
!
!           TRANSPOSE THE OUTPUT MATRIX.
!
            do j = 1_i4b, n-1_i4b
                vec(j+1_i4b:n)   = mat(j,j+1_i4b:n)
                mat(j,j+1_i4b:n) = mat(j+1_i4b:n,j)
                mat(j+1_i4b:n,j) = vec(j+1_i4b:n)
            end do
!
!           DEALLOCATE WORK SPACE.
!
            deallocate( vec )
!
        end if
!
#ifdef _OPENMP
    end if
#endif
!
    if ( failure ) then
!
        invdiag(jfail+1_i4b:n) = nan()
        d1 = zero
        if ( comp_det ) d2 = 0_i4b
!
    end if
!
!
! END OF SUBROUTINE chol_cmp
! __________________________
!
    end subroutine chol_cmp
!
! =========================================================================================
!
    subroutine chol_cmp2( mat, invdiag, d1, d2, b, matinv, upper, fill, tol )
!
! Purpose
! _______
!
!   CHOL_CMP2 computes the Cholesky factorization of a n-by-n real symmetric
!   positive definite matrix MAT. The factorization has the form
!
!         MAT = U' * U , if UPPER=true or is absent,
!
!   and
!
!         MAT = L * L' , if UPPER=false,
!
!   where U is an upper triangular matrix and L is a lower triangular matrix.
!
!   If D2 is present, CHOL_CMP2 computes the determinant of MAT as
!
!                      determinant(MAT) = scale( D1, D2 )
!
!   If B is present, CHOL_CMP2 solves the system of linear equations
!
!                       MAT * X = B
!
!   using the Cholesky factorization of MAT. Here B is a n-vector.
!
!
!   If MATINV is present, CHOL_CMP2 computes the inverse of MAT
!
!                       MATINV = MAT**(-1) 
!
!
! Arguments
! _________
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the symmetric matrix MAT. If:
!
!                - UPPER = true or is absent:
!                  The leading n-by-n upper triangular part of MAT contains the
!                  upper triangular part of the matrix and the strictly
!                  lower triangular part of MAT is not referenced.
!                - UPPER = false:
!                  The leading n-by-n lower triangular part of MAT contains the
!                  lower triangular part of the matrix and the strictly
!                  upper triangular part of MAT is not referenced.
!
!                On exit, if D1/=0, the factor U or L from the Cholesky
!                factorization  MAT = U' * U  or  MAT = L * L', except for
!                the main diagonal elements which are stored in reciprocal
!                form in INVDIAG. The main diagonal elements of MAT are not
!                modified.
!
!                The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   INVDIAG      (OUTPUT) real(stnd), dimension(:)
!                On exit, INVDIAG contains the reciprocals of the actual
!                diagonal elements of L or U.
!
!                The size of INVDIAG must verify:  size( INVDIAG ) = n .
!
!   D1           (OUTPUT) real(stnd)
!                On exit, D1 = zero indicates that the matrix MAT is algorithmically
!                not positive definite and that the factorization can not be
!                completed. Any other value indicates successful exit.
!
!                On exit, if D2 is present, D1 is the first component of the 
!                determinant (mantissa of determinant) of MAT.
!
!   D2           (OUTPUT, OPTIONAL) integer(i4b)
!                If D2 is present, the components of the determinant 
!                of MAT are computed.
!
!                On exit, D2 is the second component of the determinant of MAT
!                (exponent of determinant)
!
!                           determinant(MAT) = scale( D1, D2 )
!
!   B            (INPUT/OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                On entry, the right hand side vector B.
!                On exit, the solution vector X.
!
!                The shape of B must verify:  size( B ) = n .
!
!   MATINV       (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!                On exit, if:
!
!                - FILL = true or is absent: The (symmetric) inverse of MAT.
!                - FILL = false: The upper (if UPPER=true) or lower (if UPPER=false)
!                  triangle of the (symmetric) inverse of MAT, is stored in the upper
!                  or lower triangular part of the matrix MATINV and the other part of
!                  MATINV is not referenced.
!
!                The shape of MATINV must verify:  size( MATINV, 1 ) = size( MATINV, 2 ) = n .
!
!   UPPER        (INPUT, OPTIONAL) logical(lgl)
!                Specifies whether the upper or lower triangular part of the
!                symmetric matrix MAT is stored. If:
!
!                - UPPER = true :  Upper triangular is stored
!                - UPPER = false:  Lower triangular is stored.
!
!                The default is true.
!
!   FILL         (INPUT, OPTIONAL) logical(lgl)
!                On entry, when argument FILL is present, FILL is used as follows. If:
!
!                - FILL= true and UPPER= true,  the lower triangle of MATINV is filled on exit
!                - FILL= true and UPPER= false,  the upper triangle of MATINV is filled on exit
!                - FILL= false, the lower (UPPER= true) or upper (UPPER= false) triangle of
!                  MATINV is not filled on exit.
!
!                The default is true.
!
!   TOL          (INPUT, OPTIONAL) real(stnd)
!                Tolerance used to test the matrix for positive-definiteness.
!                TOL is used as a multiplying factor for determining effective
!                zero for pivots. TOL must be greater or equal to zero, otherwise
!                the default value is used.
!
!                The default value is the machine precision multiplied by n.
!
!
! Further Details
! _______________
!
!   MAT is declared not positive definite if during the j-th stage of the
!   factorization of MAT, a pivot, PIV(j), is such that
!
!             PIV(j) <= MAT(j,j) * TOL
!
!   In this case, the leading minor of order j of MAT is declared not
!   positive definite and on exit of CHOL_CMP2:
!
!   - D1 is set to zero,  
!   - INVDIAG(j) = PIV(j),
!   - INVDIAG(j+1_i4b:n) are set to nan() value,
!   - B is filled with nan() value,
!   - the upper or lower triangle of MATINV is filled with nan() value if FILL=false,
!   - the matrix MATINV is filled with nan() value if FILL=true,
!
!   and the Cholesky factorization is not completed.
!
!   On the other hand, if MAT is positive definite then
!
!        U(j,j) = sqrt(PIV(j)) (if UPPER=true), for j=1 to n,
!
!   or
!
!        L(j,j) = sqrt(PIV(j)) (if UPPER=false), for j=1 to n,
!
!   and on exit of CHOL_CMP2:
!
!   - D1/=0,
!   - INVDIAG(j)=1/sqrt(PIV(j)) for j=1 to n,
!   - B and MATINV are computed, if these arguments are present,
!
!   and the linear system MAT * Z = D can be solved with 
!   subroutine CHOL_SOLVE.
!
!   This is a GAxpy version of the Cholesky algorithm, for more details
!   see the reference (1).
!
!   A blocked algorithm is used to compute the factorization. Furthermore,
!   the computations are parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Higham, N.J., 2009:
!         Cholesky factorization.
!         Wiley Interdisciplinary Reviews: Computational Statistics,
!         Vol. 1, pp 251-254.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : nan
    use Utilities,       only : assert, assert_eq
#ifdef _BLAS
    use BLAS_interfaces, only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout)            :: mat(:,:)
    real(stnd),   intent(out)              :: invdiag(:), d1
    real(stnd),   intent(inout), optional  :: b(:)
    real(stnd),   intent(out),   optional  :: matinv(:,:)
    real(stnd),   intent(in),    optional  :: tol
!
    integer(i4b),  intent(out), optional :: d2
!
    logical(lgl), intent(in), optional :: upper, fill
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: tmp
!
    integer(i4b) :: i, j, n
!
    logical(lgl) :: mat_is_upper, fill_triang, comp_b, comp_matinv
!
#ifdef _OPENMP
    real(stnd)   :: piv
!
    logical      :: test_par2, ompparallel, ompnested
#ifndef _USE_PGI
    logical      :: test_par
#endif
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='chol_cmp2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,    &
                    int(size(mat,2),i4b) ,    &
                    int(size(invdiag),i4b),   &
                    name_proc )
!
    comp_b      = present( b )
    if ( comp_b ) call assert( logical( n==size(b), lgl ),        &
                               name_proc )
!
    comp_matinv = present( matinv )
    if ( comp_matinv ) call assert( logical( n==size(matinv,1), lgl ),   &
                                    logical( n==size(matinv,2), lgl ),   &
                                    name_proc )
!    
    if ( present(upper) ) then
        mat_is_upper = upper
    else
        mat_is_upper = true
    end if
!
    if ( present(fill) ) then
        fill_triang = fill
    else
        fill_triang = true
    end if
!    
!   DECOMPOSE mat .
!    
    call chol_cmp( mat, invdiag, d1, d2, upper=upper, tol=tol )
!    
    if ( d1==zero ) tmp = nan()
!    
#ifndef _USE_PGI
!    
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  comp_b                              .and.      &
                  comp_matinv                         .and.      &
                  i>1_i4b
#endif
!    
!$OMP PARALLEL IF( test_par ), NUM_THREADS( 2 ),                 &
!$OMP          PRIVATE(test_par2,ompparallel,ompnested,i),       &
!$OMP          SHARED(mat_is_upper,fill_triang,comp_b,comp_matinv)
!    
!$OMP SECTIONS
!    
!$OMP SECTION
!    
#endif
!    
    if ( comp_b ) then
!
        if ( d1==zero ) then
!    
!           mat IS ALGORITHMICALLY NOT POSITIVE DEFINITE.
!    
            b = tmp
!
        else
!    
!           SOLVE mat*X = b, OVERWRITING b WITH X .
!    
            call chol_solve_rv( mat, invdiag, b, upper=mat_is_upper )
!
        end if
!
    end if
!    
#ifndef _USE_PGI
!$OMP SECTION
#endif
!    
    if ( comp_matinv ) then
!
        if ( d1==zero ) then
!    
!           mat IS ALGORITHMICALLY NOT POSITIVE DEFINITE.
!    
            if ( fill_triang ) then
!
                matinv(:n,:n) = tmp
!
            else if ( mat_is_upper ) then
!
                do i = 1_i4b, n
                    matinv(1_i4b:i,i) = tmp
                end do
!
            else
!
                do i = 1_i4b, n
                    matinv(i:n,i) = tmp
                end do
!
            end if
!
        else
!
#ifdef _OPENMP
            i           = omp_get_max_threads()
            ompparallel = omp_in_parallel()
            ompnested   = omp_get_nested() .and. ompparallel
            test_par2   = ( .not.ompparallel .or. ompnested ) .and.      &
                          n*n>=omp_limit                      .and.      &
                          i>1_i4b
!    
            if ( test_par2 ) then
!
!$OMP PARALLEL PRIVATE(i,j,piv)
!    
!               COMPUTE INVERSE OF mat.
!    
                if ( mat_is_upper ) then
!
!                   INVERT THE TRIANGULAR CHOLESKY FACTOR U.
!
!$OMP DO SCHEDULE(STATIC,1)
!
                    do i = 1_i4b, n
!
                        matinv(1_i4b:i-1_i4b,i) = -invdiag(i)*mat(1_i4b:i-1_i4b,i)
                        matinv(i,i)             = invdiag(i)
!
                        do j = i-1_i4b, 1_i4b, -1_i4b
!
                            piv                     = matinv(j,i)*invdiag(j)
#ifdef _BLAS
                            call axpy( j-1_i4b, -piv, mat(1_i4b:j-1_i4b,j), 1_i4b, matinv(1_i4b:j-1_i4b,i), 1_i4b )
#else
                            matinv(1_i4b:j-1_i4b,i) = matinv(1_i4b:j-1_i4b,i) - piv*mat(1_i4b:j-1_i4b,j)
#endif
                            matinv(j,i)             = piv
!
                        end do
!
                    end do
!
!$OMP END DO
!
                else
!
!                   INVERT THE TRIANGULAR CHOLESKY FACTOR L.
!
!$OMP DO SCHEDULE(STATIC,1)
!
                    do i = 1_i4b, n
!
                        matinv(i,i)         = invdiag(i)
                        matinv(i+1_i4b:n,i) = -invdiag(i)*mat(i+1_i4b:n,i)
!
                        do j = i+1_i4b, n
!
                            piv                 = matinv(j,i)*invdiag(j)
#ifdef _BLAS
                            call axpy( n-j, -piv, mat(j+1_i4b:n,j), 1_i4b, matinv(j+1_i4b:n,i), 1_i4b )
#else
                            matinv(j+1_i4b:n,i) = matinv(j+1_i4b:n,i) - piv*mat(j+1_i4b:n,j)
#endif
                            matinv(j,i)         = piv
!
                        end do
!
                    end do
!
!$OMP END DO
!
                end if
!
!$OMP END PARALLEL
!    
            else
#endif
!    
!               COMPUTE INVERSE OF mat.
!    
                if ( mat_is_upper ) then
!
!                   INVERT THE TRIANGULAR CHOLESKY FACTOR U.
!
                    do j = 1_i4b, n
!
                        matinv(1_i4b:j-1_i4b,j) = mat(1_i4b:j-1_i4b,j)
                        matinv(j,j)             = invdiag(j)
!
!                       COMPUTE ELEMENTS 1:j-1 OF COLUMN j.
!
                        do i = 1_i4b, j-1_i4b
!
                            if ( matinv(i,j)==zero ) cycle
!
#ifdef _BLAS
                            call axpy( i-1_i4b, matinv(i,j), matinv(1_i4b:i-1_i4b,i), 1_i4b, matinv(1_i4b:i-1_i4b,j), 1_i4b )
#else
                            matinv(1_i4b:i-1_i4b,j) = matinv(1_i4b:i-1_i4b,j) + matinv(i,j)*matinv(1_i4b:i-1_i4b,i)
#endif
                            matinv(i,j)             = matinv(i,j)*invdiag(i)
!
                        end do
!
                        matinv(1_i4b:j-1_i4b,j) = -invdiag(j)*matinv(1_i4b:j-1_i4b,j)
!
                    end do
!
                else
!
!                   INVERT THE TRIANGULAR CHOLESKY FACTOR L.
!
                    do j = n, 1_i4b, -1_i4b
!
                        matinv(j,j)         = invdiag(j)
                        matinv(j+1_i4b:n,j) = mat(j+1_i4b:n,j)
!
!                       COMPUTE ELEMENTS j+1:n OF COLUMN j.
!
                        do i = n, j+1_i4b, -1_i4b
!
                            if ( matinv(i,j)==zero ) cycle
!
#ifdef _BLAS
                            call axpy( n-i, matinv(i,j), matinv(i+1_i4b:n,i), 1_i4b, matinv(i+1_i4b:n,j), 1_i4b )
#else
                            matinv(i+1_i4b:n,j) = matinv(i+1_i4b:n,j) + matinv(i,j)*matinv(i+1_i4b:n,i)
#endif
                            matinv(i,j)         = matinv(i,j)*invdiag(i)
!
                        end do
!
                        matinv(j+1_i4b:n,j) = -invdiag(j)*matinv(j+1_i4b:n,j)
!
                    end do
!
                end if
!
#ifdef _OPENMP
            end if
#endif
!    
!           FORM inv(U)*inv(U)' OR inv(L)'*inv(L).
!
            call comp_uut_ltl_a( matinv, upper=mat_is_upper, fill=fill_triang )
!
        end if
!
    end if
!    
#ifndef _USE_PGI
!$OMP END SECTIONS
!    
!$OMP END PARALLEL
#endif
!
!
! END OF SUBROUTINE chol_cmp2
! ___________________________
!
    end subroutine chol_cmp2
!
! =========================================================================================
!
    subroutine gchol_cmp( mat, invdiag, krank, d1, d2, upper, tol )
!
! Purpose
! _______
!
!   GCHOL_CMP computes the Cholesky factorization of a n-by-n real symmetric
!   positive semidefinite matrix MAT. The factorization has the form
!
!         MAT = U' * U , if UPPER=true or is absent,
!
!   and
!
!         MAT = L * L' , if UPPER=false,
!
!   where U is an upper triangular matrix and L is a lower triangular matrix.
!
!
! Arguments
! _________
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the symmetric matrix MAT. If:
!
!                - UPPER = true or is absent:
!                  The leading n-by-n upper triangular part of MAT contains the
!                  upper triangular part of the matrix and the strictly
!                  lower triangular part of MAT is not referenced.
!                - UPPER = false:
!                  The leading n-by-n lower triangular part of MAT contains the
!                  lower triangular part of the matrix and the strictly
!                  upper triangular part of MAT is not referenced.
!
!                On exit, if D1>=0, the factor U or L from the Cholesky
!                factorization  MAT = U' * U  or  MAT = L * L', except for
!                the main diagonal elements which are stored in reciprocal
!                form in INVDIAG. The main diagonal elements of MAT are not
!                modified.
!
!                The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   INVDIAG      (OUTPUT) real(stnd), dimension(:)
!                On exit, INVDIAG contains the reciprocals of the actual
!                diagonal elements of L or U, excepted for zeroed elements
!                if MAT is not positive definite.
!
!                The shape of INVDIAG must verify:  size( INVDIAG ) = n .
!
!   KRANK        (OUTPUT) integer(i4b)
!                On exit, KRANK contains the effective rank of MAT, which is defined
!                as the number of nonzero elements of the diagonal of L or U. Note
!                that KRANK may be different from the true rank of MAT. See the
!                reference (2) for details.
!
!   D1           (OUTPUT) real(stnd)
!                On exit, D1 < zero indicates that the matrix MAT is algorithmically
!                not positive semidefinite and that the factorization can not be
!                completed. Any other value indicates successful exit.
!
!                On exit, if D2 is present, D1 is the first component of the 
!                determinant (mantissa of determinant) of MAT.
!
!   D2           (OUTPUT, OPTIONAL) integer(i4b)
!                If D2 is present, the components of the determinant 
!                of MAT are computed.
!
!                On exit, D2 is the second component of the determinant of MAT
!                (exponent of determinant)
!
!                           determinant(MAT) = scale( D1, D2 )
!
!   UPPER        (INPUT, OPTIONAL) logical(lgl)
!                Specifies whether the upper or lower triangular part of the
!                symmetric matrix MAT is stored. If:
!
!                - UPPER = true :  Upper triangular is stored
!                - UPPER = false:  Lower triangular is stored.
!
!                The default is true.
!
!   TOL          (INPUT, OPTIONAL) real(stnd)
!                Tolerance used to test MAT for positive-semidefiniteness.
!                TOL is used as a multiplying factor for determining effective
!                zero for pivots. TOL must be greater or equal to zero, otherwise
!                the default value is used.
!
!                The default value is the machine precision multiplied by n.
!
!
! Further Details
! _______________
!
!   MAT is declared not positive semidefinite if during the j-th stage of the
!   factorization of MAT, a pivot, PIV(j), is such that
!
!             PIV(j) <= - abs( MAT(j,j) * TOL )
!
!   In this case, the leading minor of order j of MAT is declared not
!   positive semidefinite and on exit of GCHOL_CMP:
!
!   - D1 is set to -1,  
!   - KRANK is set to -1,  
!   - INVDIAG(j) = PIV(j),
!   - INVDIAG(j+1_i4b:n) are set to nan() value,
!
!   and the Cholesky factorization is not completed.
!
!   On the other hand, if MAT is positive semidefinite (e.g. D1>=0),
!   KRANK is computed as follows:
!
!   KRANK is initially set to n. if, during the factorization,
!   a pivot, PIV(j), is such that
!
!              abs( PIV(j) ) <= abs( MAT(j,j) * TOL )
!
!   KRANK is decreased by 1 and U(j,j:n) (if UPPER=true) or L(j:n,j)
!   (if UPPER=false) is set to zero. Note that KRANK may be different
!   from the true rank of MAT. See the reference (2) for details.
!
!   IF PIV(j) does not satisfy this condition then
!
!   - U(j,j) = sqrt(PIV(j)) (if UPPER=true),
!   - L(j,j) = sqrt(PIV(j)) (if UPPER=false).
!
!   On exit of GCHOL_CMP, if MAT is positive semidefinite, INVDIAG
!   contains the reciprocals of the diagonal elements of U or L,
!   excepted for zeroed elements during the factorization as described
!   above.
!
!   If MAT is positive semidefinite (D1>=0), the linear system MAT * Z = D
!   can also be solved with subroutine CHOL_SOLVE.
!
!   This is a GAxpy version of the Cholesky algorithm, for more details
!   see the reference (1).
!
!   A blocked algorithm is used to compute the factorization. Furthermore,
!   the computations are parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Higham, N.J., 2009:
!         Cholesky factorization.
!         Wiley Interdisciplinary Reviews: Computational Statistics,
!         Vol. 1, pp 251-254.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error
    use Num_Constants,   only : lamch, nan, machbase
    use Utilities,       only : assert_eq, merror
#ifdef _BLAS
    use BLAS_interfaces, only : dot
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), dimension(:,:), intent(inout) :: mat
    real(stnd), dimension(:),   intent(out)   :: invdiag
    real(stnd),                 intent(out)   :: d1
    real(stnd), optional,       intent(in)    :: tol
!
    integer(i4b),  intent(out)           :: krank
    integer(i4b),  intent(out), optional :: d2
!
    logical(lgl), intent(in), optional :: upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                            :: piv, tol_piv
    real(stnd), dimension(:), allocatable :: vec
!
    integer(i4b) :: n, i, j, j1, j2, k, nblk, rem, n2, jfail
    integer      :: iok
!
    logical(lgl) :: comp_det, mat_is_upper, failure
!
#ifdef _OPENMP
    integer(i4b) :: l, l1, l2, m, iblk, num_threads, thread_num
    integer(kind=omp_lock_kind), dimension(:), allocatable :: lck
!
    logical(lgl) :: do_lock
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='gchol_cmp'
!
    real(stnd), parameter :: invmachbase=one/machbase
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,    &
                    int(size(mat,2),i4b) ,    &
                    int(size(invdiag),i4b),   &
                    name_proc )
!    
    krank = n
!
    d1 = one
!
    if ( present( d2 ) ) then
        comp_det = true
        d2 = 0_i4b
    else
        comp_det = false
    end if
!
    mat_is_upper = true
    if ( present(upper) ) mat_is_upper = upper
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!    
!   INITIALIZE TOLERANCE.
!    
    tol_piv = lamch( 'p' )*real( n, stnd )
!
    if ( present( tol ) ) then
        tol_piv = merge( tol, tol_piv, tol>=zero )
    end if
!
!   CHECK MAGNITUDE OF THE FIRST PIVOT.
!
    piv = mat(1_i4b,1_i4b)
!
    if ( abs(piv)<=lamch( 's') ) then
!
        if ( mat_is_upper ) then
!
!           ZERO ELEMENTS 2:n OF ROW 1 OF U.
!
            mat(1_i4b,2_i4b:n) = zero
!
        else
!
!           ZERO ELEMENTS 2:n OF COLUMN 1 OF L.
!
            mat(2_i4b:n,1_i4b) = zero
!
        end if
!
        invdiag(1_i4b) = zero
!
        krank = krank - 1_i4b
!
        d1 = zero
!
        if ( comp_det ) then
            d2 = 0_i4b
            comp_det = false
        end if
!
    else if ( piv>zero ) then
!
!       TEST FOR POSITIVE-DEFINITENESS.
!
!       COMPUTE DETERMINANT IF REQUIRED.
!
        if ( comp_det ) then
!
            d1 = d1*piv
!
            do
                if ( abs(d1)<= one ) exit
                d1 = d1*invmachbase
                d2 = d2 + 1_i4b
            end do
!
            do
                if ( abs(d1)>= invmachbase ) exit
                d1 = d1*machbase
                d2 = d2 - 1_i4b
            end do
!
        end if
!
!       COMPUTE 1/U(1,1) OR 1/L(1,1).
!
        piv            = one/sqrt( piv )
        invdiag(1_i4b) = piv
!
        if ( mat_is_upper ) then
!
!           COMPUTE ELEMENTS 2:n OF ROW 1 OF U.
!
            mat(1_i4b,2_i4b:n) = mat(1_i4b,2_i4b:n)*piv
!
        else
!
!           COMPUTE ELEMENTS 2:n OF COLUMN 1 OF L.
!
            mat(2_i4b:n,1_i4b) = mat(2_i4b:n,1_i4b)*piv
!
        end if
!
    else
!
!       mat IS NOT POSITIVE SEMIDEFINITE,
!       QUICK RETURN.
!
        invdiag(2_i4b:n) = nan()
!
        d1 = -one
        if ( comp_det ) d2 = 0_i4b
!
        krank = -1_i4b
!
        return
!
    end if
!    
    failure = false
!
!   COMPUTE THE NUMBER OF BLOCKS.
!
    nblk = (n-1_i4b)/blksz
    rem  = mod( n-1_i4b, blksz )
!
    if ( rem/=0_i4b ) then
!
        nblk = nblk + 1_i4b
        n2   = n - rem
!
    else
!
        n2   = n - blksz
!
    end if
!
!   COMPUTE THE CHOLESKY FACTORIZATION mat = U'*U OR mat = L*L'.
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*n>=omp_limit                      .and.      &
                  nblk>1_i4b                          .and.      &
                  i>1_i4b
!
    if ( test_par ) then
!
!       PARALLEL AND BLOCKED CHOLESKY DECOMPOSITION OF mat.
!
!       ALLOCATE WORK ARRAY FOR STORING THE LOCKS FOR EACH BLOCK.
!
        allocate( lck(nblk), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SET UP LOCKS FOR EACH BLOCK.
!
        do i = 1_i4b, nblk
            call omp_init_lock( lck(i) )
        end do
!
!$OMP PARALLEL PRIVATE(i,j,j1,j2,k,l,l1,l2,m,iblk,iok,piv,vec,num_threads,thread_num,do_lock)
!
        num_threads = omp_get_num_threads()
        thread_num  = omp_get_thread_num()
!
!       SET THE LOCKS FOR THE BLOCKS.
!
!$OMP DO SCHEDULE(STATIC,1)
!
        do i = 1_i4b, nblk
            call omp_set_lock( lck(i) )
        end do
!
!$OMP END DO
!
        if ( .not.mat_is_upper ) then
!
!           ALLOCATE WORK SPACE FOR A FAST FACTORIZATION
!           IF THE LOWER TRIANGULAR PART OF mat IS STORED.
!
            allocate( vec(n), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           TRANSPOSE THE INPUT MATRIX.
!
!$OMP DO SCHEDULE(STATIC)
!
            do j = 1_i4b, n-1_i4b
                vec(j+1_i4b:n)   = mat(j,j+1_i4b:n)
                mat(j,j+1_i4b:n) = mat(j+1_i4b:n,j)
                mat(j+1_i4b:n,j) = vec(j+1_i4b:n)
            end do
!
!$OMP END DO
!
        end if
!
        if ( thread_num==0_i4b ) then
!
!           COMPUTE CHOLESKY FACTORIZATION OF THE FIRST BLOCK IN PLACE.
!
            do k = 2_i4b, blksz+1_i4b
!
!               START COMPUTATION OF ELEMENTS k TO blksz+1 OF ROW k OF U.
!
                do i = k, blksz+1_i4b       
#ifdef _BLAS
                    invdiag(i) = mat(k,i) - dot( k-1_i4b, mat(1_i4b:k-1_i4b,k), 1_i4b, mat(1_i4b:k-1_i4b,i), 1_i4b )
#else
                    invdiag(i) = mat(k,i) - dot_product( mat(1_i4b:k-1_i4b,k), mat(1_i4b:k-1_i4b,i) )
#endif
                end do
!
                piv = invdiag(k)
!
!               TEST FOR SINGULARITY.
!
                if ( abs(piv)<=abs(tol_piv*mat(k,k)) ) then
!
                    krank = krank - 1_i4b
!
                    d1 = zero
!
                    if ( comp_det ) then
                        d2 = 0_i4b
                        comp_det = false
                    end if
!
!                   ZERO ELEMENTS k TO blksz+1 OF ROW k OF U.
!
                    invdiag(k)                 = zero
                    mat(k,k+1_i4b:blksz+1_i4b) = zero
!
                else if ( piv>zero ) then
!
!                   TEST FOR POSITIVE-DEFINITENESS.
!
!                   COMPUTE DETERMINANT IF REQUIRED.
!
                    if ( comp_det ) then
!
                        d1 = d1*piv
!
                        do
                            if ( abs(d1)<= one ) exit
                            d1 = d1*invmachbase
                            d2 = d2 + 1_i4b
                        end do
!
                        do
                            if ( abs(d1)>= invmachbase ) exit
                            d1 = d1*machbase
                            d2 = d2 - 1_i4b
                        end do
!
                    end if
!
!                   COMPUTE 1/U(k,k).
!
                    piv        = one/sqrt( piv )
                    invdiag(k) = piv
!
!                   COMPUTE ELEMENTS k+1 TO blksz+1 OF ROW k OF U.
!
                    mat(k,k+1_i4b:blksz+1_i4b) = invdiag(k+1_i4b:blksz+1_i4b)*piv
!
                else
!
!                   mat IS NOT POSITIVE SEMIDEFINITE,
!                   QUICK RETURN.
!
                    failure = true
                    jfail   = k
!
                    exit
!
                end if
!
            end do
!
!           RELEASE THE FIRST LOCK.
!
! $OMP FLUSH (mat,invdiag,d1,d2,comp_det,failure,jfail,krank)
!
            call omp_unset_lock( lck(1) )     
!
            do_lock = false
!
        else
!
            do_lock = true
!
        end if
!
!       START THE BLOCKED CHOLESKY FACTORIZATION.
!
        do j = 1_i4b, n2, blksz
!
            iblk = (j-1_i4b)/blksz + 1_i4b
!
            if ( do_lock ) then
!
                call omp_set_lock( lck(iblk) )
!
! $OMP FLUSH (mat,invdiag,d1,d2,comp_det,failure,jfail,krank)
!
                call omp_unset_lock( lck(iblk) )
!
            else
!
                do_lock = true
!
            end if
!
            if ( failure ) exit
!
            j1 = j + 1_i4b
            j2 = min( j + blksz, n )
!
!           APPLY TRANSFORMATIONS IN THE iblk BLOCK TO LATER COLUMNS, EG TO mat(j1:j2,j2+1:n).
!
            k = ( (iblk/num_threads)*num_threads ) + thread_num
            if ( thread_num<mod(iblk,num_threads) ) k = k + num_threads
!
            do l = k, nblk-1_i4b, num_threads
!
!               COMPUTE INDICES FOR THE FIRST AND LAST COLUMNS OF THE l+1 BLOCK.
!
                l1 = l*blksz + 2_i4b
                l2 = min( (l+1_i4b)*blksz + 1_i4b, n )
!
                do i = l1, l2
!
                    do m = j1, j2
!
#ifdef _BLAS
                        piv = mat(m,i) - dot( m-1_i4b, mat(1_i4b:m-1_i4b,m), 1_i4b, mat(1_i4b:m-1_i4b,i), 1_i4b )
#else
                        piv = mat(m,i) - dot_product( mat(1_i4b:m-1_i4b,m), mat(1_i4b:m-1_i4b,i) )
#endif
!
!                       COMPUTE ELEMENTS l1 to l2 OF ROW m OF U.
!
                        mat(m,i) = piv*invdiag(m)
!
                    end do
!
                end do
!
                if ( l==iblk ) then
!
!                   COMPUTE THE CHOLESKY FACTORIZATION OF THE iblk+1 BLOCK, EG OF mat(j2+1:j2+blksz,j2+1:j2+blksz), IN PLACE.
!
                    do i = l1, l2
!
!                       START COMPUTATION OF ELEMENTS i:l2 OF ROW i OF U.
!
                        do m = i, l2         
#ifdef _BLAS
                            invdiag(m) = mat(i,m) - dot( i-1_i4b, mat(1_i4b:i-1_i4b,i), 1_i4b, mat(1_i4b:i-1_i4b,m), 1_i4b )
#else
                            invdiag(m) = mat(i,m) - dot_product( mat(1_i4b:i-1_i4b,i), mat(1_i4b:i-1_i4b,m) )
#endif
                        end do
!
                        piv = invdiag(i)
!
!                       TEST FOR SINGULARITY.
!
                        if ( abs(piv)<=abs(tol_piv*mat(i,i)) ) then
!
                            krank = krank - 1_i4b
!
                            d1 = zero
!
                            if ( comp_det ) then
                                d2 = 0_i4b
                                comp_det = false
                            end if
!
!                           ZERO ELEMENTS i TO l2 OF ROW i OF U.
!
                            invdiag(i)        = zero
                            mat(i,i+1_i4b:l2) = zero
!
                        else if ( piv>zero ) then
!
!                           TEST FOR POSITIVE-DEFINITENESS.
!
!                           COMPUTE DETERMINANT IF REQUIRED.
!
                            if ( comp_det ) then
!
                                d1 = d1*piv
!
                                do
                                    if ( abs(d1)<= one ) exit
                                    d1 = d1*invmachbase
                                    d2 = d2 + 1_i4b
                                end do
!
                                do
                                    if ( abs(d1)>= invmachbase ) exit
                                    d1 = d1*machbase
                                    d2 = d2 - 1_i4b
                                end do
!
                            end if
!
!                           COMPUTE 1/U(i,i).
!
                            piv        = one/sqrt( piv )
                            invdiag(i) = piv
!
!                           COMPUTE ELEMENTS i+1 TO l2 OF ROW i OF U.
!
                            mat(i,i+1_i4b:l2) = invdiag(i+1_i4b:l2)*piv
!
                        else
!
!                           mat IS NOT POSITIVE SEMIDEFINITE,
!                           QUICK RETURN.
!
                            failure = true
                            jfail   = i
!
                            exit
!
                        end if
!
                    end do
!
!                   RELEASE THE iblk+1_i4b LOCK.
!
! $OMP FLUSH (mat,invdiag,d1,d2,comp_det,failure,jfail,krank)
!
                    call omp_unset_lock( lck(iblk+1_i4b) )                 
!
                    do_lock = false
!
                end if
!
            end do
!
        end do
!
        if ( failure ) then
!
!           UNSET THE LOCKS FOR THE REMAINING BLOCKS IF REQUIRED.
!
!$OMP DO SCHEDULE(STATIC,1)
!
            do i = 1_i4b, nblk
!
                if ( i>iblk ) then
                    call omp_unset_lock( lck(i) )
                end if
!
            end do
!
!$OMP END DO
!
        end if
!
!$OMP BARRIER
!
        if ( .not.mat_is_upper ) then
!
!           TRANSPOSE THE OUTPUT MATRIX.
!
!$OMP DO SCHEDULE(STATIC)
!
            do j = 1_i4b, n-1_i4b
                vec(j+1_i4b:n)   = mat(j,j+1_i4b:n)
                mat(j,j+1_i4b:n) = mat(j+1_i4b:n,j)
                mat(j+1_i4b:n,j) = vec(j+1_i4b:n)
            end do
!
!$OMP END DO
!
!           DEALLOCATE WORK SPACE.
!
            deallocate( vec )
!
        end if
!
!$OMP END PARALLEL
!
!       DESTROY THE LOCKS.
!
        do i = 1_i4b, nblk
            call omp_destroy_lock( lck(i) )
        end do
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( lck )
!
    else
!
#endif
!
!       BLOCKED CHOLESKY DECOMPOSITION OF mat.
!
        if ( .not.mat_is_upper ) then
!
!           ALLOCATE WORK SPACE FOR A FAST FACTORIZATION
!           IF THE LOWER TRIANGULAR PART OF mat IS STORED.
!
            allocate( vec(n), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!           TRANSPOSE THE INPUT MATRIX.
!
            do j = 1_i4b, n-1_i4b
                vec(j+1_i4b:n)   = mat(j,j+1_i4b:n)
                mat(j,j+1_i4b:n) = mat(j+1_i4b:n,j)
                mat(j+1_i4b:n,j) = vec(j+1_i4b:n)
            end do
!
        end if
!
!       START THE BLOCKED CHOLESKY FACTORIZATION.
!
        j_loop : do j = 1_i4b, n2, blksz
!
            j1 = j + 1_i4b
            j2 = min( j + blksz, n )
!
!           COMPUTE THE CHOLESKY FACTORIZATION OF THE ((j-1)*blksz)+1 BLOCK, EG OF mat(j+1:j+blksz,j+1:j+blksz), IN PLACE.
!
            do k = j1, j2
!
!               START COMPUTATION OF ELEMENTS k TO j+blksz OF ROW k OF U.
!
                do i = k, j2         
#ifdef _BLAS
                    invdiag(i) = mat(k,i) - dot( k-1_i4b, mat(1_i4b:k-1_i4b,k), 1_i4b, mat(1_i4b:k-1_i4b,i), 1_i4b )
#else
                    invdiag(i) = mat(k,i) - dot_product( mat(1_i4b:k-1_i4b,k), mat(1_i4b:k-1_i4b,i) )
#endif
                end do
!
                piv = invdiag(k)
!
!               TEST FOR SINGULARITY.
!
                if ( abs(piv)<=abs(tol_piv*mat(k,k)) ) then
!
                    krank = krank - 1_i4b
!
                    d1 = zero
!
                    if ( comp_det ) then
                        d2 = 0_i4b
                        comp_det = false
                    end if
!
!                   ZERO ELEMENTS k TO j+blksz OF ROW k OF U.
!
                    invdiag(k)        = zero
                    mat(k,k+1_i4b:j2) = zero
!
                else if ( piv>zero ) then
!
!                   TEST FOR POSITIVE-DEFINITENESS.
!
!                   COMPUTE DETERMINANT IF REQUIRED.
!
                    if ( comp_det ) then
!
                        d1 = d1*piv
!
                        do
                            if ( abs(d1)<= one ) exit
                            d1 = d1*invmachbase
                            d2 = d2 + 1_i4b
                        end do
!
                        do
                            if ( abs(d1)>= invmachbase ) exit
                            d1 = d1*machbase
                            d2 = d2 - 1_i4b
                        end do
!
                    end if
!
!                   COMPUTE 1/U(k,k).
!
                    piv        = one/sqrt( piv )
                    invdiag(k) = piv
!
!                   COMPUTE ELEMENTS k+1 TO j+blksz OF ROW k OF U.
!
                    mat(k,k+1_i4b:j2) = invdiag(k+1_i4b:j2)*piv
!
                else
!
!                   mat IS NOT POSITIVE SEMIDEFINITE,
!                   QUICK RETURN.
!
                    failure = true
                    jfail   = k
!
                    exit j_loop
!
                end if
!
            end do
!
!           APPLY TRANSFORMATIONS IN THE ((j-1)/blksz)+1 BLOCK TO LATER COLUMNS, EG TO mat(j1:j2,j2+1:n).
!
            do i = j2+1_i4b, n
!
                do k = j1, j2
!
#ifdef _BLAS
                    piv = mat(k,i) - dot( k-1_i4b, mat(1_i4b:k-1_i4b,k), 1_i4b, mat(1_i4b:k-1_i4b,i), 1_i4b )
#else
                    piv = mat(k,i) - dot_product( mat(1_i4b:k-1_i4b,k), mat(1_i4b:k-1_i4b,i) )
#endif
!
!                   COMPUTE ELEMENTS j+blksz+1 to n OF ROW k OF U.
!
                    mat(k,i) = piv*invdiag(k)
!
                end do
!          
            end do
!
        end do j_loop
!
        if ( .not.mat_is_upper ) then
!
!           TRANSPOSE THE OUTPUT MATRIX.
!
            do j = 1_i4b, n-1_i4b
                vec(j+1_i4b:n)   = mat(j,j+1_i4b:n)
                mat(j,j+1_i4b:n) = mat(j+1_i4b:n,j)
                mat(j+1_i4b:n,j) = vec(j+1_i4b:n)
            end do
!
!           DEALLOCATE WORK SPACE.
!
            deallocate( vec )
!
        end if
!
#ifdef _OPENMP
    end if
#endif
!
    if ( failure ) then
!
        invdiag(jfail+1_i4b:n) = nan()
!
        d1 = -one
        if ( comp_det ) d2 = 0_i4b
!
        krank = -1_i4b
!
    end if
!
!
! END OF SUBROUTINE gchol_cmp
! ___________________________
!
    end subroutine gchol_cmp
!
! =========================================================================================
!
    subroutine gchol_cmp2( mat, invdiag, krank, d1, d2, b, matinv, upper, fill, tol )
!
! Purpose
! _______
!
!   GCHOL_CMP2 computes the Cholesky factorization of a n-by-n real symmetric positive
!   semidefinite matrix MAT. The factorization has the form
!
!         MAT = U' * U , if UPPER=true or is absent,
!
!   and
!
!         MAT = L * L' , if UPPER=false,
!
!   where U is an upper triangular matrix and L is a lower triangular matrix.
!
!   If D2 is present, GCHOL_CMP2 computes the determinant of MAT as
!
!                      determinant(MAT) = scale( D1, D2 )
!
!   If B is present, GCHOL_CMP2 solves the system of linear equations
!
!                       MAT * X = B
!
!   using the Cholesky factorization of MAT if B belongs to the range of MAT.
!   Here B is a n-vector. IF B does not belongs to the range of MAT, an approximate
!   solution is computed as
!
!                       X = MATINV * B
!
!   where MATINV is a (generalized) inverse of MAT.
!
!   If MATINV is present, GCHOL_CMP2 computes a (generalized) inverse of MAT.
!
!
! Arguments
! _________
!
!   MAT          (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the symmetric matrix MAT. If:
!
!                - UPPER = true or is absent:
!                  The leading n-by-n upper triangular part of MAT contains the
!                  upper triangular part of the matrix and the strictly
!                  lower triangular part of MAT is not referenced.
!                - UPPER = false:
!                  The leading n-by-n lower triangular part of MAT contains the
!                  lower triangular part of the matrix and the strictly
!                  upper triangular part of MAT is not referenced.
!
!                On exit, if D1>=0, the factor U or L from the Cholesky
!                factorization  MAT = U' * U  or  MAT = L * L', except for
!                the main diagonal elements which are stored in reciprocal
!                form in INVDIAG. The main diagonal elements of MAT are not
!                modified.
!
!                The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   INVDIAG      (OUTPUT) real(stnd), dimension(:)
!                On exit, INVDIAG contains the reciprocals of the actual
!                diagonal elements of L or U, excepted for zeroed elements
!                if MAT is not positive definite.
!
!                The shape of INVDIAG must verify:  size( INVDIAG ) = n .
!
!   KRANK        (OUTPUT) integer(i4b)
!                On exit, KRANK contains the effective rank of MAT, which is defined
!                as the number of nonzero elements of the diagonal of L or U. Note
!                that KRANK may be different from the true rank of MAT. See the
!                reference (2) for details.
!
!   D1           (OUTPUT) real(stnd)
!                On exit, D1 < zero indicates that the matrix MAT is algorithmically
!                not positive semidefinite and that the factorization can not be
!                completed. Any other value indicates successful exit.
!
!                On exit, if D2 is present, D1 is the first component of the 
!                determinant (mantissa of determinant) of MAT.
!
!   D2           (OUTPUT, OPTIONAL) integer(i4b)
!                If D2 is present, the components of the determinant 
!                of MAT are computed.
!
!                On exit, D2 is the second component of the determinant of MAT
!                (exponent of determinant)
!
!                           determinant(MAT) = scale( D1, D2 )
!
!   B            (INPUT/OUTPUT, OPTIONAL) real(stnd), dimension(:)
!                On entry, the right hand side vector B.
!
!                On exit, one solution vector X if B belongs to the range
!                of MAT, otherwise an approximate solution computed with
!                the help of a symmetric generalized inverse of MAT.
!
!                The shape of B must verify:  size( B ) = n .
!
!   MATINV       (OUTPUT) real(stnd), dimension(:,:)
!                On exit, if:
!
!                - FILL = true or is absent:
!                  The (symmetric) (generalized) inverse of MAT.
!                - FILL = false:
!                  The upper (if UPPER=true) or lower (if UPPER=false) triangle
!                  of the (symmetric) (generalized) inverse of MAT, is stored in 
!                  the upper or lower triangular part of the matrix MATINV and the
!                  other part of MATINV is not referenced.
!
!                The shape of MATINV must verify:  size( MATINV, 1 ) = size( MATINV, 2 ) = n .
!
!   UPPER        (INPUT, OPTIONAL) logical(lgl)
!                Specifies whether the upper or lower triangular part of the
!                symmetric matrix MAT is stored. If:
!
!                - UPPER = true :  Upper triangular is stored
!                - UPPER = false:  Lower triangular is stored.
!
!                The default is true.
!
!   FILL         (INPUT, OPTIONAL) logical(lgl)
!                On entry, when argument FILL is present, FILL is used as follows. If:
!
!                - FILL= true and UPPER= true,  the lower triangle of MATINV is filled on exit.
!                - FILL= true and UPPER= false,  the upper triangle of MATINV is filled on exit.
!                - FILL= false, the lower (UPPER= true) or upper (UPPER= false) triangle of
!                  MATINV is not filled on exit.
!
!                The default is true.
!
!   TOL          (INPUT, OPTIONAL) real(stnd)
!                Tolerance used to test MAT for positive-semidefiniteness. TOL is
!                used as a multiplying factor for determining effective zero
!                for pivots. TOL must be greater or equal to zero, otherwise
!                the default value is used.
!
!                The default is the machine precision multiplied by n.
!
!
! Further Details
! _______________
!
!   MAT is declared not positive semidefinite if during the j-th stage of the
!   factorization of MAT, a pivot, PIV(j), is such that
!
!             PIV(j) <= - abs( MAT(j,j) * TOL )
!
!   In this case, the leading minor of order j of MAT is declared not
!   positive semidefinite and on exit of GCHOL_CMP2:
!
!   - D1 is set to -1,  
!   - KRANK is set to -1,  
!   - INVDIAG(j) = PIV(j),
!   - INVDIAG(j+1_i4b:n) are set to nan() value,
!   - B is filled with nan() value,
!   - the upper or lower triangle of MATINV is filled with nan() value if FILL=false,
!   - the matrix MATINV is filled with nan() value if FILL=true,
!
!   and the Cholesky factorization is not completed.
!
!   On the other hand, if MAT is positive semidefinite (D1>=0),
!   KRANK is computed as follows:
!
!   KRANK is initially set to n. if, during the factorization,
!   a pivot, PIV(j), is such that
!
!              abs( PIV(j) ) <= abs( MAT(j,j) * TOL )
!
!   KRANK is decreased by 1 and U(j,j:n) (if UPPER=true) or L(j:n,j)
!   (if UPPER=false) is set to zero. Note that KRANK may be different
!   from the true rank of MAT. See the reference (2) for details.
!
!   IF PIV(j) does not satisfy this condition then
!
!   - U(j,j) = sqrt(PIV(j)) (if UPPER=true),
!   - L(j,j) = sqrt(PIV(j)) (if UPPER=false).
!
!   On exit of GCHOL_CMP2, if MAT is positive semidefinite, INVDIAG
!   contains the reciprocals of the diagonal elements of U or L,
!   excepted for zeroed elements during the factorization as described
!   above.
!
!   If MAT is positive semidefinite (D1>=0), MATINV is computed as follows. If:
!
!   - KRANK=n, MATINV is just the inverse of MAT, MATINV = MAT**(-1) ,
!   - KRANK<n, MATINV is a generalized inverse of MAT.
!
!   MATINV is a generalized inverse of MAT if
!
!            MAT * MATINV * MAT = MAT  and  MATINV * MAT * MATINV = MATINV
!
!   If MAT is positive semidefinite (D1>=0), the linear system MAT * Z = D can also be
!   solved with subroutine CHOL_SOLVE if D belongs to the range of MAT.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Higham, N.J., 2009:
!         Cholesky factorization.
!         Wiley Interdisciplinary Reviews: Computational Statistics,
!         Vol. 1, pp 251-254.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : nan
    use Utilities,       only : assert, assert_eq
#ifdef _BLAS
    use BLAS_interfaces, only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout)            :: mat(:,:)
    real(stnd),   intent(out)              :: invdiag(:), d1
    real(stnd),   intent(inout), optional  :: b(:)
    real(stnd),   intent(out),   optional  :: matinv(:,:)
    real(stnd),   intent(in),    optional  :: tol
!
    integer(i4b),  intent(out)           :: krank
    integer(i4b),  intent(out), optional :: d2
!
    logical(lgl), intent(in), optional :: upper, fill
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: tmp
!
    integer(i4b) :: i, j, n
!
    logical(lgl) :: mat_is_upper, fill_triang, comp_b, comp_matinv
!
#ifdef _OPENMP
    real(stnd)   :: piv
!
    logical      :: test_par2, ompparallel, ompnested
#ifndef _USE_PGI
    logical      :: test_par
#endif
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='gchol_cmp2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,    &
                    int(size(mat,2),i4b) ,    &
                    int(size(invdiag),i4b),   &
                    name_proc )
!
    comp_b      = present( b )
    if ( comp_b ) call assert( logical( n==size(b), lgl ),        &
                               name_proc )
!
    comp_matinv = present( matinv )
    if ( comp_matinv ) call assert( logical( n==size(matinv,1), lgl ),   &
                                    logical( n==size(matinv,2), lgl ),   &
                                    name_proc )
!
    if ( present(upper) ) then
        mat_is_upper = upper
    else
        mat_is_upper = true
    end if
!
    if ( present(fill) ) then
        fill_triang = fill
    else
        fill_triang = true
    end if
!    
!   DECOMPOSE mat .
!    
    call gchol_cmp( mat, invdiag, krank, d1, d2, upper=upper, tol=tol )
!    
    if ( d1<zero ) tmp = nan()
!    
#ifndef _USE_PGI
!    
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  comp_b                              .and.      &
                  comp_matinv                         .and.      &
                  i>1_i4b
#endif
!    
!$OMP PARALLEL IF( test_par ), NUM_THREADS( 2 ),                 &
!$OMP          PRIVATE(test_par2,ompparallel,ompnested,i),       &
!$OMP          SHARED(mat_is_upper,fill_triang,comp_b,comp_matinv)
!    
!$OMP SECTIONS
!    
!$OMP SECTION
!    
#endif
!    
    if ( comp_b ) then
!
        if ( d1<zero ) then
!    
!           mat IS ALGORITHMICALLY NOT POSITIVE SEMIDEFINITE.
!    
            b = tmp
!
        else
!    
!           SOLVE mat*X = b, OVERWRITING b WITH X .
!    
            call chol_solve_rv( mat, invdiag, b, upper=mat_is_upper )
!
        end if
!
    end if
!    
#ifndef _USE_PGI
!$OMP SECTION
#endif
!    
    if ( comp_matinv ) then
!
        if ( d1<zero ) then
!    
!           mat IS ALGORITHMICALLY NOT POSITIVE SEMIDEFINITE.
!    
            if ( fill_triang ) then
!
                matinv(:n,:n) = tmp
!
            else if ( mat_is_upper ) then
!
                do i = 1_i4b, n
                  matinv(1_i4b:i,i) = tmp
                end do
!
            else
!
                do i = 1_i4b, n
                  matinv(i:n,i) = tmp
                end do
!
            end if
!
        else
!
#ifdef _OPENMP
            i           = omp_get_max_threads()
            ompparallel = omp_in_parallel()
            ompnested   = omp_get_nested() .and. ompparallel
            test_par2   = ( .not.ompparallel .or. ompnested ) .and.      &
                          n*n>=omp_limit                      .and.      &
                          i>1_i4b
!    
            if ( test_par2 ) then
!
!$OMP PARALLEL PRIVATE(i,j,piv)
!    
!               COMPUTE INVERSE OF mat.
!    
                if ( mat_is_upper ) then
!
!                   COMPUTE GENERALIZED INVERSE OF THE TRIANGULAR CHOLESKY FACTOR U : ginv(U) .
!
!$OMP DO SCHEDULE(STATIC,1)
!
                    do i = 1_i4b, n
!
                        if ( invdiag(i)==zero ) then
!
                            matinv(1_i4b:i,i) = zero
!
                        else
!
                            matinv(1_i4b:i-1_i4b,i) = -invdiag(i)*mat(1_i4b:i-1_i4b,i)
                            matinv(i,i)             = invdiag(i)
!
                            do j = i-1_i4b, 1_i4b, -1_i4b
!
                                if ( invdiag(j)==zero ) cycle
!
                                piv                     = matinv(j,i)*invdiag(j)
#ifdef _BLAS
                                call axpy( j-1_i4b, -piv, mat(1_i4b:j-1_i4b,j), 1_i4b, matinv(1_i4b:j-1_i4b,i), 1_i4b )
#else
                                matinv(1_i4b:j-1_i4b,i) = matinv(1_i4b:j-1_i4b,i) - piv*mat(1_i4b:j-1_i4b,j)
#endif
                                matinv(j,i)             = piv
!
                            end do
!
                        end if
!
                    end do
!
!$OMP END DO
!
                else
!
!                   COMPUTE GENERALIZED INVERSE OF THE TRIANGULAR CHOLESKY FACTOR L : ginv(L) .
!
!$OMP DO SCHEDULE(STATIC,1)
!
                    do i = 1_i4b, n
!
                        if ( invdiag(i)==zero ) then
!
                            matinv(i:n,i) = zero
!
                        else
!
                            matinv(i,i)         = invdiag(i)
                            matinv(i+1_i4b:n,i) = -invdiag(i)*mat(i+1_i4b:n,i)
!
                            do j = i+1_i4b, n
!
                                if ( invdiag(j)==zero ) then
!
                                    matinv(j,i) = zero
!
                                else
!
                                    piv                 = matinv(j,i)*invdiag(j)
                                    matinv(j,i)         = piv
#ifdef _BLAS
                                    call axpy( n-j, -piv, mat(j+1_i4b:n,j), 1_i4b, matinv(j+1_i4b:n,i), 1_i4b )
#else
                                    matinv(j+1_i4b:n,i) = matinv(j+1_i4b:n,i) - piv*mat(j+1_i4b:n,j)
#endif
!
                                end if
!
                            end do
!
                        end if
!
                    end do
!
!$OMP END DO
!
                end if
!
!$OMP END PARALLEL
!    
            else
#endif
!    
!               FORM GENERALIZED-INVERSE OF mat.
!    
                if ( mat_is_upper ) then
!
!                   COMPUTE GENERALIZED INVERSE OF THE TRIANGULAR CHOLESKY FACTOR U : ginv(U) .
!
                    do j = 1_i4b, n
!
                        if ( invdiag(j)==zero ) then
!
                            matinv(1_i4b:j,j) = zero
!
                        else
!
                            matinv(1_i4b:j-1_i4b,j) = mat(1_i4b:j-1_i4b,j)
                            matinv(j,j)             = invdiag(j)
!
!                           COMPUTE ELEMENTS 1:j-1 OF COLUMN j.
!
                            do i = 1_i4b, j-1_i4b
!
                                if ( invdiag(i)==zero ) cycle
!
#ifdef _BLAS
                                call axpy( i-1_i4b, matinv(i,j), matinv(1_i4b:i-1_i4b,i), 1_i4b, matinv(1_i4b:i-1_i4b,j), 1_i4b )
#else
                                matinv(1_i4b:i-1_i4b,j) = matinv(1_i4b:i-1_i4b,j) + matinv(i,j)*matinv(1_i4b:i-1_i4b,i)
#endif
                                matinv(i,j)             = matinv(i,j)*invdiag(i)
!
                            end do
!
                            matinv(1_i4b:j-1_i4b,j) = -invdiag(j)*matinv(1_i4b:j-1_i4b,j)
!
                        end if
!
                    end do
!
                else
!
!                   COMPUTE GENERALIZED INVERSE OF THE TRIANGULAR CHOLESKY FACTOR L : ginv(L) .
!
                    do j = n, 1_i4b, -1_i4b
!
                        if ( invdiag(j)==zero ) then
!
                            matinv(j,1_i4b:j) = zero
!
                        else
!
                            matinv(j,1_i4b:j-1_i4b) = mat(j,1_i4b:j-1_i4b)
                            matinv(j,j)             = invdiag(j)
!
!                           COMPUTE ELEMENTS j+1:n OF COLUMN j.
!
                            do i = n, j+1_i4b, -1_i4b
!
                                if ( invdiag(i)==zero ) cycle
!
#ifdef _BLAS
                                call axpy( n-i, matinv(i,j), matinv(i+1_i4b:n,i), 1_i4b, matinv(i+1_i4b:n,j), 1_i4b )
#else
                                matinv(i+1_i4b:n,j) = matinv(i+1_i4b:n,j) + matinv(i,j)*matinv(i+1_i4b:n,i)
#endif
                                matinv(i,j)         = matinv(i,j)*invdiag(i)
!
                            end do
!
                            matinv(j+1_i4b:n,j) = -invdiag(j)*matinv(j+1_i4b:n,j)
!
                        end if
!
                    end do
!
                end if
!
#ifdef _OPENMP
            end if
#endif
!    
!           FORM ginv(U)*ginv(U)' OR ginv(L)'*ginv(L).
!
            call comp_uut_ltl_a( matinv, upper=mat_is_upper, fill=fill_triang )
!
        end if
!
    end if
!    
#ifndef _USE_PGI
!$OMP END SECTIONS
!    
!$OMP END PARALLEL
#endif
!    
!
! END OF SUBROUTINE gchol_cmp2
! ____________________________
!
    end subroutine gchol_cmp2
!
!
! =========================================================================================
!                                   LINEAR PROBLEMS
! =========================================================================================
!
!
    subroutine lu_solve_rv( mat, ip, b )
!
! Purpose
! _______
!
!   LU_SOLVE solves a system of linear equations
!
!                       MAT * X = B
!
!  where MAT is a n-by-n coefficient matrix and B is a n-vector, using the LU
!  factorization with scaled partial pivoting of MAT, P * MAT = L * U,
!  as computed by LU_CMP or LU_CMP2.
!
! Arguments
! _________
!
!   MAT     (INPUT) real(stnd), dimension(:,:)
!           MAT contains the LU factorization of P * MAT for some permutation matrix
!           P specified by argument IP. It is assumed that MAT is as generated by LU_CMP
!           or LU_CMP2.
!
!           The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   IP      (INPUT) integer(i4b), dimension(:)
!           The permutation matrix P as generated by LU_CMP or LU_CMP2.
!
!           The shape of IP must verify:  size( IP ) = n .
!
!   B       (INPUT/OUTPUT) real(stnd), dimension(:)
!           On entry, the right hand side vector B.
!
!           On exit, the solution vector X.
!
!           The shape of B must verify:  size( B ) = n .
!
!
! Further Details
! _______________
!
!   It is assumed that LU_CMP or LU_CMP2 has been used to compute the LU 
!   factorization of MAT before LU_SOLVE.
!
!   MAT and IP are not modified by this routine and can be left in place for
!   successive calls with different right-hand sides B.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Higham, N.J., 2011:
!           Gaussian elimination.
!           Wiley Interdisciplinary Reviews: Computational Statistics,
!           Vol. 3, Issue 3, pp 230-238.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert_eq
#ifdef _BLAS
    use BLAS_interfaces, only : axpy, trsv
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in)    :: mat(:,:)
    real(stnd),   intent(inout) :: b(:)
!
    integer(i4b), intent(in) :: ip(:)
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)    :: tmp
!
    integer(i4b)  :: n, i, j
#ifndef _BLAS
    integer(i4b)  :: i1, i2, rem, n2
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc ='lu_solve'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(ip),i4b)    ,   &
                    int(size(b),i4b)     ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!    
!   APPLY PERMUTATIONS.
!    
    do i = 1_i4b, n
!
        j    = ip(i)
!
        if ( i==j ) cycle
!
        tmp  = b(j)
        b(j) = b(i)
        b(i) = tmp
!
    end do
!    
!   FORWARD SUBSTITUTION.
!    
#ifdef _BLAS
    call trsv( 'L', 'N', 'U', n, mat, n, b, 1_i4b )
#else
    rem = mod( n, blksz )
    n2  = n - rem - blksz
!
    do j = 0_i4b, n2, blksz
!
        i1 = j + 1_i4b
        i2 = j + blksz
!
        do i = i1, i2-1_i4b
!
            b(i+1_i4b:i2) = b(i+1_i4b:i2) - b(i)*mat(i+1_i4b:i2,i)
!
        end do
!
        b(i2+1_i4b:n) = b(i2+1_i4b:n) - matmul( mat(i2+1_i4b:n,i1:i2) , b(i1:i2) )
!          
    end do
!
    if ( rem/=0_i4b ) then
!
        do i = n-rem+1_i4b, n-1_i4b
!
!           APPLY THE TRANSFORMATION TO b .
!
            b(i+1_i4b:n) = b(i+1_i4b:n) - b(i)*mat(i+1_i4b:n,i)
!
        end do
!
    end if
#endif
!    
!   BACK SUBSTITUTION.
!    
!    b(n) = b(n)*mat(n,n)
!    do i = n-1_i4b, 1_i4b, -1_i4b
!        b(1_i4b:i) = b(1_i4b:i) - b(i+1_i4b)*mat(1_i4b:i,i)
!        b(i) = b(i)*mat(i,i)
!    end do

    do i = n, 1_i4b, -1_i4b
!
        b(i) = b(i)*mat(i,i)
#ifdef _BLAS
        call axpy( i-1_i4b, -b(i), mat(1_i4b:i-1_i4b,i), 1_i4b, b(1_i4b:i-1_i4b), 1_i4b )
#else
        b(1_i4b:i-1_i4b) = b(1_i4b:i-1_i4b) - b(i)*mat(1_i4b:i-1_i4b,i)
#endif
!
    end do
!
!
! END OF SUBROUTINE lu_solve_rv
! _____________________________
!
    end subroutine lu_solve_rv
!
! =========================================================================================
!
    subroutine lu_solve_rm( mat, ip, b )
!
! Purpose
! _______
!
!   LU_SOLVE solves a system of linear equations with several right hand sides
!
!                       MAT * X = B
!
!   where MAT is a n-by-n coefficient matrix and B is a n-by-nb matrix, using the
!   LU factorization with scaled partial pivoting of MAT, P * MAT = L * U, as
!   computed by LU_CMP or LU_CMP2.
!
!
! Arguments
! _________
!
!   MAT     (INPUT) real(stnd), dimension(:,:)
!           MAT contains the LU factorization of P * MAT for some permutation
!           matrix P specified by argument IP. It is assumed that MAT is as 
!           generated by LU_CMP or LU_CMP2.
!
!           The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   IP      (INPUT) integer(i4b), dimension(:)
!           The permutation matrix P as generated by LU_CMP or LU_CMP2.
!
!           The shape of IP must verify:  size( IP ) = n .
!
!   B       (INPUT/OUTPUT) real(stnd), dimension(:,:)
!           On entry, the right hand side matrix B.
!
!           On exit, the solution matrix X.
!
!           The shape of B must verify:  size( B, 1 ) = n .
!
!
! Further Details
! _______________
!
!   It is assumed that LU_CMP or LU_CMP2 has been used to compute the LU
!   factorization of MAT before LU_SOLVE.
!
!   MAT and IP are not modified by this routine and can be left in place for
!   successive calls with different right-hand sides B.
!
!   The computations are parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert_eq
#ifdef _BLAS
    use BLAS_interfaces, only : axpy, trsm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in)    :: mat(:,:)
    real(stnd),   intent(inout) :: b(:,:)
!
    integer(i4b), intent(in) :: ip(:)
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: tmp
!
    integer(i4b) :: n, nb, i, j, k
#ifndef _BLAS
    integer(i4b) :: i1, i2, rem, n2
#endif
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc ='lu_solve'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(ip),i4b)    ,   &
                    int(size(b,1),i4b)   ,   &
                    name_proc )
!
    nb =  size( b, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. nb<=0_i4b ) return
!    
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.     &
                  nb>=2_i4b                           .and.     &
                  i>1_i4b
#endif
!    
#ifdef _BLAS
!$OMP PARALLEL IF(test_par), PRIVATE(i,j,k,tmp)
!    
!$OMP DO SCHEDULE(STATIC)
!
    do k = 1_i4b, nb
!    
!       APPLY PERMUTATIONS.
!    
        do i = 1_i4b, n
!
            j      = ip(i)
!
            if ( i==j ) cycle
!
            tmp    = b(j,k)
            b(j,k) = b(i,k)
            b(i,k) = tmp
!
        end do
!
    end do
!    
!$OMP END DO
!    
!$OMP END PARALLEL
!    
!   FORWARD SUBSTITUTION.
!    
    call trsm( 'L', 'L', 'N', 'U', n, nb, one, mat, n, b, n )
!    
!$OMP PARALLEL IF(test_par), PRIVATE(i,k)
!    
!$OMP DO SCHEDULE(STATIC)
!
    do k = 1_i4b, nb
!    
!       BACK SUBSTITUTION.
!
        do i = n, 1_i4b, -1_i4b
!
            b(i,k) = b(i,k)*mat(i,i)
!
            call axpy( i-1_i4b, -b(i,k), mat(1_i4b:i-1_i4b,i), 1_i4b, b(1_i4b:i-1_i4b,k), 1_i4b )
!
        end do
!
    end do
!    
!$OMP END DO
!    
!$OMP END PARALLEL
#else
    rem = mod( n, blksz )
    n2  = n - rem - blksz
!    
!$OMP PARALLEL IF(test_par), PRIVATE(i,i1,i2,j,k,tmp)
!    
!$OMP DO SCHEDULE(STATIC)
!
    do k = 1_i4b, nb
!    
!       APPLY PERMUTATIONS.
!    
        do i = 1_i4b, n
!
            j      = ip(i)
!
            if ( i==j ) cycle
!
            tmp    = b(j,k)
            b(j,k) = b(i,k)
            b(i,k) = tmp
!
        end do
!    
!       FORWARD SUBSTITUTION.
!    
        do j = 0_i4b, n2, blksz
!
            i1 = j + 1_i4b
            i2 = j + blksz
!
            do i = i1, i2-1_i4b
!
                b(i+1_i4b:i2,k) = b(i+1_i4b:i2,k) - b(i,k)*mat(i+1_i4b:i2,i)
!
            end do
!
            b(i2+1_i4b:n,k) = b(i2+1_i4b:n,k) - matmul( mat(i2+1_i4b:n,i1:i2) , b(i1:i2,k) )
!          
        end do
!
        if ( rem/=0_i4b ) then
!
            do i = n-rem+1_i4b, n-1_i4b
!
!               APPLY THE TRANSFORMATION TO b .
!
                b(i+1_i4b:n,k) = b(i+1_i4b:n,k) - b(i,k)*mat(i+1_i4b:n,i)
!
            end do
!
        end if
!    
!       BACK SUBSTITUTION.
!
        do i = n, 1_i4b, -1_i4b
!
            b(i,k) = b(i,k)*mat(i,i)
!
            b(1_i4b:i-1_i4b,k) = b(1_i4b:i-1_i4b,k) - b(i,k)*mat(1_i4b:i-1_i4b,i)
!
        end do
!
    end do
!    
!$OMP END DO
!    
!$OMP END PARALLEL
#endif
!
!
! END OF SUBROUTINE lu_solve_rm
! _____________________________
!
    end subroutine lu_solve_rm
!
! =========================================================================================
!
    subroutine lu_solve2_rv( mat, ip, b )
!
! Purpose
! _______
!
!   LU_SOLVE2 solves a system of linear equations
!
!                       MAT * X = B
!
!   where MAT is a n-by-n coefficient matrix and B is a n-vector, using the LU
!   factorization with scaled partial pivoting of MAT, P * MAT = L * U, as
!   computed by LU_CMP or LU_CMP2.
!
!
! Arguments
! _________
!
!   MAT     (INPUT) real(stnd), dimension(:,:)
!           MAT contains the LU factorization of P * MAT for some permutation
!           matrix P specified by argument IP. It is assumed that MAT is as 
!           generated by LU_CMP or LU_CMP2.
!
!           The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   IP      (INPUT) integer(i4b), dimension(:)
!           The permutation matrix P as generated by LU_CMP or LU_CMP2.
!
!           The shape of IP must verify:  size( IP ) = n .
!
!   B       (INPUT/OUTPUT) real(stnd), dimension(:)
!           On entry, the right hand side vector B.
!
!           On exit, the solution vector X.
!
!           The shape of B must verify:  size( B ) = n .
!
!
! Further Details
! _______________
!
!   It is assumed that LU_CMP or LU_CMP2 has been used to factor MAT before LU_SOLVE2.
!
!   MAT and IP are not modified by this routine and can be left in place for
!   successive calls with different right-hand sides B.
!
!   This subroutines takes into account the possibility that B will begin with
!   many zero elements, so it is efficient for use in matrix inversion.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Higham, N.J., 2011:
!           Gaussian elimination.
!           Wiley Interdisciplinary Reviews: Computational Statistics,
!           Vol. 3, Issue 3, pp 230-238.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert_eq
#ifdef _BLAS
    use BLAS_interfaces, only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in)    :: mat(:,:)
    real(stnd),   intent(inout) :: b(:)
!
    integer(i4b), intent(in) :: ip(:)
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: tmp
!
    integer(i4b)  :: n, i, j, istart
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='lu_solve2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(ip),i4b)    ,   &
                    int(size(b),i4b)     ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!    
!   APPLY PERMUTATIONS AND DO FORWARD SUBSTITUTION.
!    
    istart = 0_i4b
!
    do i = 1_i4b, n
!
        j    = ip(i)
        tmp  = b(j)
        b(j) = b(i)
!
        if ( istart/=0_i4b ) then
            tmp = tmp - dot_product( mat(i,istart:i-1_i4b), b(istart:i-1_i4b) )
        else if ( tmp/=zero ) then
            istart = i
        end if
!
        b(i) = tmp
!
    end do
!    
!   BACK SUBSTITUTION.
!    
    do i = n, 1_i4b, -1_i4b
!
        b(i) = b(i)*mat(i,i)
#ifdef _BLAS
        call axpy( i-1_i4b, -b(i), mat(1_i4b:i-1_i4b,i), 1_i4b, b(1_i4b:i-1_i4b), 1_i4b )
#else
        b(1_i4b:i-1_i4b) = b(1_i4b:i-1_i4b) - b(i)*mat(1_i4b:i-1_i4b,i)
#endif
!
    end do
!
!
! END OF SUBROUTINE lu_solve2_rv
! ______________________________
!
    end subroutine lu_solve2_rv
!
! =========================================================================================
!
    subroutine lu_solve2_rm( mat, ip, b )
!
! Purpose
! _______
!
!   LU_SOLVE2 solves a system of linear equations with several right hand sides
!
!                       MAT * X = B
!
!   where MAT is a n-by-n coefficient matrix and B is a n-by-nb matrix, using the LU
!   factorization with scaled partial pivoting of MAT, P * MAT = L * U, as computed
!   by LU_CMP or LU_CMP2.
!
!
! Arguments
! _________
!
!   MAT     (INPUT) real(stnd), dimension(:,:)
!           MAT contains the LU factorization of P * MAT for some permutation
!           matrix P specified by argument IP. It is assumed that MAT is as
!           generated by LU_CMP or LU_CMP2.
!
!           The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   IP      (INPUT) integer(i4b), dimension(:)
!           The permutation matrix P as generated by LU_CMP or LU_CMP2.
!
!           The shape of IP must verify:  size( IP ) = n .
!
!   B       (INPUT/OUTPUT) real(stnd), dimension(:,:)
!           On entry, the right hand side matrix B.
!
!           On exit, the solution matrix X.
!
!           The shape of B must verify:  size( B, 1 ) = n .
!
!
! Further Details
! _______________
!
!   It is assumed that LU_CMP or LU_CMP2 has been used to factor MAT before LU_SOLVE2.
!
!   MAT and IP are not modified by this routine and can be left in place for
!   successive calls with different right-hand sides B.
!
!   This subroutines takes into account the possibility that each column of B will
!   begin with many zero elements, so it is efficient for use in matrix inversion.
!
!   The computations are parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert_eq
#ifdef _BLAS
    use BLAS_interfaces, only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in)    :: mat(:,:)
    real(stnd),   intent(inout) :: b(:,:)
!
    integer(i4b), intent(in) :: ip(:)
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd) :: tmp
!
    integer(i4b) :: n, nb, i, j, k, istart
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='lu_solve2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(ip),i4b)    ,   &
                    int(size(b,1),i4b)   ,   &
                    name_proc )
!
    nb =  size( b, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. nb<=0_i4b ) return
!    
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.     &
                  nb>=2_i4b                           .and.     &
                  i>1_i4b
#endif
!    
!$OMP PARALLEL IF(test_par), PRIVATE(i,j,k,istart,tmp)
!    
!$OMP DO SCHEDULE(STATIC)
!
    do k = 1_i4b, nb
!    
!       APPLY PERMUTATIONS AND DO FORWARD SUBSTITUTION.
!    
        istart = 0_i4b
!
        do i = 1_i4b, n
!
            j      = ip(i)
            tmp    = b(j,k)
            b(j,k) = b(i,k)
!
            if ( istart/=0_i4b ) then
                tmp = tmp - dot_product( mat(i,istart:i-1_i4b), b(istart:i-1_i4b,k) )
            else if ( tmp/=zero ) then
                istart = i
            end if
!
            b(i,k) = tmp
!
        end do
!    
!       BACK SUBSTITUTION.
!    
        do i = n, 1_i4b, -1_i4b
!
            b(i,k) = b(i,k)*mat(i,i)
#ifdef _BLAS
            call axpy( i-1_i4b, -b(i,k), mat(1_i4b:i-1_i4b,i), 1_i4b, b(1_i4b:i-1_i4b,k), 1_i4b )
#else
            b(1_i4b:i-1_i4b,k) = b(1_i4b:i-1_i4b,k) - b(i,k)*mat(1_i4b:i-1_i4b,i)
#endif
!
        end do
!
    end do
!    
!$OMP END DO
!    
!$OMP END PARALLEL
!
!
! END OF SUBROUTINE lu_solve2_rm
! _____________________________
!
    end subroutine lu_solve2_rm
!
! =========================================================================================
!
    function solve_lin_rv( mat, b, tol ) result( x )
!
! Purpose
! _______
!
!   SOLVE_LIN solves a system of linear equations
!
!                       MAT * X = B
!
!   with a n-by-n coefficient matrix MAT. B is a n-vector.
!
!   The function returns the solution vector X, if the
!   matrix MAT is not singular.
!
!
! Arguments
! _________
!
!   MAT     (INPUT) real(stnd), dimension(:,:)
!           MAT contains the coefficient matrix of the equation
!
!                             MAT * X = B
!
!           MAT is not modified by the subroutine.
!
!           The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   B       (INPUT) real(stnd), dimension(:)
!           On entry, the right hand side vector B.
!           B is not modified by the subroutine.
!
!           The shape of B must verify:  size( B ) = n .
!
!   TOL     (INPUT, OPTIONAL) real(stnd)
!           On entry, a relative tolerance used to indicate whether or not MAT
!           is nearly singular. Tol should normally be choose as approximately
!           the largest relative error in the elements of MAT. For example, 
!           if the elements of MAT are correct to about 4 significant figures,
!           then TOL should be set to about 5 * 10**(-4). If TOL is supplied as
!           less than EPS, where EPS is the relative machine precision, then
!           the value EPS is used in place of TOL.
!
!           Default: EPS, the relative machine precision.
!
!
! Further Details
! _______________
!
!   The LU decomposition with partial pivoting and implicit row
!   scaling of the matrix MAT is used to solve the linear system.
!
!   MAT is declared singular if a diagonal element of U is such that
!
!             abs( U(j,j) ) <= n * norm( MAT(j,:) ) * TOL
!
!   where norm( MAT(j,:) ) denotes the maximum of the absolute values of 
!   the jth row of the matrix MAT. In this case, a diagonal element of U is
!   small, indicating that MAT is singular or nearly singular.
!
!   On exit, if MAT is singular, the function returns a n-vector filled
!   with nan() value.
!
!   A blocked algorithm is used to compute the factorization and solve the
!   triangular systems. Furthermore, the computations are parallelized if OPENMP
!   is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Higham, N.J., 2011:
!           Gaussian elimination.
!           Wiley Interdisciplinary Reviews: Computational Statistics,
!           Vol. 3, Issue 3, pp 230-238.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : nan, lamch
    use Utilities,       only : assert_eq, imaxloc
#ifdef _BLAS
    use BLAS_interfaces, only : axpy, trsv
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in)            :: mat(:,:), b(:)
    real(stnd),   intent(in), optional  :: tol
!
    real(stnd) :: x(size(b))
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                :: piv, tol_piv, safmin
    real(stnd), dimension(size(b))            :: tmp_row, equil
#ifndef _BLAS
    real(stnd), dimension(size(b))            :: diag
#endif
    real(stnd), dimension(size(b), size(b)+1) :: syst
!
    integer(i4b)                     :: n, np1, i, i1, i2, imax, j, k, l, rem, n2
    integer(i4b), dimension(size(b)) :: ip
!
    logical(lgl) :: failure
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='solve_lin'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(b),i4b)     ,   &
                    name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!    
!   INITIALIZE TOLERANCE AND IMPLICIT SCALING VECTOR.
!    
    tol_piv = lamch( 'u' )
    if ( present( tol ) ) tol_piv = max( tol, tol_piv )
    tol_piv = tol_piv*real( n, stnd )
!    
    safmin = lamch( 's' )
!    
    tmp_row(:n) = maxval( abs(mat(:,:)), dim=2 )
!
    if ( any( tmp_row(:n)<safmin ) ) then
!
!       mat IS SINGULAR, EXIT.
!
        x(:n) = nan()
!
        return
!
    end if
!
    equil(:n) = one/tmp_row(:n)
!    
!   INITIALIZE WORK VARIABLES.
!    
    np1          = n + 1_i4b
!
    syst(:n,:n)  = mat(:n,:n)
    syst(:n,np1) = b(:n)
!
    failure = false
!   
!   TRIANGULATION.  
!    
    rem = mod( n, blksz )
!
    if ( rem/=0_i4b ) then
!
        n2   = n - rem
!
    else
!
        n2   = n - blksz
!
    end if
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*np1>=omp_limit                    .and.      &
                  n2/blksz>=1_i4b                     .and.      &
                  i>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par),PRIVATE(i,i1,i2,j,k,l,piv)
!
!   START THE BLOCKED LU FACTORIZATION OF mat.
!
    do j = 0_i4b, n2, blksz
!
        i1 = j + 1_i4b
        i2 = min( j + blksz, n )
!
!$OMP SINGLE
!
!       COMPUTE THE LU FACTORIZATION OF mat(i1:n,i1:i2).
!
        do i = i1, i2
!
!           FIND PIVOT.
!
            tmp_row(i:n) = equil(i:n)*abs( syst(i:n,i) )
            imax         = i - 1_i4b + imaxloc( tmp_row(i:n) )
            piv          = syst(imax,i)
!
!           TEST FOR SINGULARITY.
!
            if ( tmp_row(imax)<=tol_piv ) then
!
!               mat IS SINGULAR.
!
                failure = true
!
                exit
!
            end if
!
!           IF NECESSARY, APPLY ROW INTERCHANGE TO mat(:,j+1:j+blksz).
!
            if ( imax/=i ) then
!
                tmp_row(i1:i2)   = syst(imax,i1:i2)
                syst(imax,i1:i2) = syst(i,i1:i2)
                syst(i,i1:i2)    = tmp_row(i1:i2)
!
                equil(imax) = equil(i)
!
            else
!
                tmp_row(i+1_i4b:i2) = syst(i,i+1_i4b:i2)
!
            end if
!
!           STORE PERMUTATION.
!
            ip(i) = imax
!
#ifdef _BLAS
            syst(i,i) = piv
            piv       = one/piv
#else
            piv     = one/piv
            diag(i) = piv
#endif
            syst(i+1_i4b:n,i) = syst(i+1_i4b:n,i)*piv
!
!           REDUCE REMAINING MATRIX.
!
            do k = i+1_i4b, i2
!
#ifdef _BLAS
                call axpy( n-i, -tmp_row(k), syst(i+1_i4b:n,i), 1_i4b, syst(i+1_i4b:n,k), 1_i4b )
#else
                syst(i+1_i4b:n,k) = syst(i+1_i4b:n,k) - tmp_row(k)*syst(i+1_i4b:n,i)
#endif
!
            end do
!
        end do
!
!$OMP END SINGLE
!
        if ( failure ) exit
!
!$OMP DO SCHEDULE(STATIC)
!
        do l = i2+1_i4b, np1
!
!           APPLY PERMUTATIONS TO LATER COLUMNS, EG TO syst(i1:n,i2+1:np1).
!
            do i = i1, i2
!
                k         = ip(i)
                piv       = syst(k,l)
                syst(k,l) = syst(i,l)
                syst(i,l) = piv
!
            end do
!    
!           APPLY TRANSFORMATIONS TO LATER COLUMNS, EG TO syst(i1:n,i2+1:np1).
!    
            do i = i1, i2
!
#ifdef _BLAS
                call axpy( n-i, -syst(i,l), syst(i+1_i4b:n,i), 1_i4b, syst(i+1_i4b:n,l), 1_i4b )
#else
                syst(i+1_i4b:n,l) = syst(i+1_i4b:n,l) - syst(i,l)*syst(i+1_i4b:n,i)
#endif
!
            end do
!
        end do
!
!$OMP END DO
!          
    end do
!
!$OMP END PARALLEL
!    
    if ( .not.failure ) then
!    
!       BACK SUBSTITUTION.
!
#ifdef _BLAS
        call trsv( 'U', 'N', 'N', n, syst(1_i4b:n,1_i4b:n), n, syst(1_i4b:n,np1), 1_i4b )
!
        x(1_i4b:n) = syst(1_i4b:n,np1)
#else
        do i = n, 1_i4b, -1_i4b
!
            piv = syst(i,np1)*diag(i)
!
            syst(1_i4b:i-1_i4b,np1) = syst(1_i4b:i-1_i4b,np1) - piv*syst(1_i4b:i-1_i4b,i)
!
            x(i) = piv
!
        end do
#endif
!
    else
!
        x(:n) = nan()
!
    end if
!
!
! END OF FUNCTION solve_lin_rv
! ____________________________
!
    end function solve_lin_rv
!
! =========================================================================================
!
    function solve_lin_rm( mat, b, tol ) result( x )
!
! Purpose
! _______
!
!   SOLVE_LIN solves a system of linear equations with several right hand sides
!
!                       MAT * X = B
!
!   with a n-by-n coefficient matrix MAT. B is a n-by-nb matrix.
!
!   The function returns the n-by-nb solution matrix X, if MAT is
!   not singular.
!
!
! Arguments
! _________
!
!   MAT     (INPUT) real(stnd), dimension(:,:)
!           MAT contains the coefficient matrix of the equation
!
!                             MAT * X = B
!
!           MAT is not modified by the subroutine.
!
!           The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   B       (INPUT) real(stnd), dimension(:,:)
!           On entry, the right hand side matrix B.
!           B is not modified by the subroutine.
!
!           The shape of B must verify:  size( B, 1 ) = n .
!
!   TOL     (INPUT, OPTIONAL) real(stnd)
!           On entry, a relative tolerance used to indicate whether or not MAT
!           is nearly singular. Tol should normally be choose as approximately
!           the largest relative error in the elements of MAT. For example, 
!           if the elements of MAT are correct to about 4 significant figures,
!           then TOL should be set to about 5 * 10**(-4). If TOL is supplied as
!           less than EPS, where EPS is the relative machine precision, then
!           the value EPS is used in place of TOL.
!
!           Default: EPS, the relative machine precision.
!
!
!
! Further Details
! _______________
!
!   The LU decomposition with partial pivoting and implicit row
!   scaling of the matrix MAT is used to solve the linear system.
!
!   MAT is declared singular if a diagonal element of U is such that
!
!             abs( U(j,j) ) <= n * norm( MAT(j,:) ) * TOL
!
!   where norm( MAT(j,:) ) denotes the maximum of the absolute values of 
!   the jth row of the matrix MAT. In this case, a diagonal element of U is
!   small, indicating that MAT is singular or nearly singular.
!
!   On exit, if MAT is algorithmically singular, the function returns a n-by-nb matrix
!   filled with nan() value.
!
!   A blocked algorithm is used to compute the factorization and solve the triangular
!   systems. Furthermore, the computations are parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Higham, N.J., 2011:
!           Gaussian elimination.
!           Wiley Interdisciplinary Reviews: Computational Statistics,
!           Vol. 3, Issue 3, pp 230-238.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : nan, lamch
    use Utilities,       only : assert_eq, imaxloc
#ifdef _BLAS
    use BLAS_interfaces, only : axpy, trsm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in), dimension(:,:) :: mat, b
    real(stnd),   intent(in),    optional    :: tol
!
    real(stnd) :: x(size(b,1),size(b,2))
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                       :: piv, tol_piv, safmin
    real(stnd), dimension(size(b,1)) :: tmp_row, equil
#ifndef _BLAS
    real(stnd), dimension(size(b,1)) :: diag
#endif
    real(stnd), dimension(size(b,1), size(b,1)+size(b,2)) :: syst
!
    integer(i4b)  :: n, np1, nb, npb, i, i1, i2, imax, j, k, l, rem, n2
    integer(i4b), dimension(size(b,1)) :: ip
!
    logical(lgl) :: failure
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='solve_lin'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(b,1),i4b)   ,   &
                    name_proc )
!
    nb =  size( b, 2 )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. nb<=0_i4b ) return
!    
!   INITIALIZE TOLERANCE AND IMPLICIT SCALING VECTOR.
!    
    tol_piv = lamch( 'u' )
    if ( present( tol ) ) tol_piv = max( tol, tol_piv )
    tol_piv = tol_piv*real( n, stnd )
!    
    safmin = lamch( 's' )
!    
    tmp_row(:n) = maxval( abs(mat(:n,:n)), dim=2 )
!
    if ( any( tmp_row(:n)<safmin ) ) then
!
!       mat IS SINGULAR, EXIT.
!
        x(:n,:nb) = nan()
!
        return
!
    end if
!
    equil(:n) = one/tmp_row(:n)
!    
!   INITIALIZE WORK VARIABLES.
!    
    np1 = n + 1_i4b
    npb = n + nb
!
    syst(:n,:n) = mat(:n,:n)
    syst(:n,np1:npb) = b(:n,:nb)
!
    failure = false
!   
!   TRIANGULATION.  
!    
    rem = mod( n, blksz )
!
    if ( rem/=0_i4b ) then
!
        n2   = n - rem
!
    else
!
        n2   = n - blksz
!
    end if
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*npb>=omp_limit                    .and.      &
                  i>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par),PRIVATE(i,i1,i2,j,k,l,piv)
!
!   START THE BLOCKED LU FACTORIZATION OF mat.
!
    do j = 0_i4b, n2, blksz
!
        i1 = j + 1_i4b
        i2 = min( j + blksz, n )
!
!$OMP SINGLE
!
!       COMPUTE THE LU FACTORIZATION OF mat(i1:n,i1:i2).
!
        do i = i1, i2
!
!           FIND PIVOT.
!
            tmp_row(i:n) = equil(i:n)*abs( syst(i:n,i) )
            imax         = i - 1_i4b + imaxloc( tmp_row(i:n) )
            piv          = syst(imax,i)
!
!           TEST FOR SINGULARITY.
!
            if ( tmp_row(imax)<=tol_piv ) then
!
!               mat IS SINGULAR.
!
                failure = true
!
                exit
!
            end if
!
!           IF NECESSARY, APPLY ROW INTERCHANGE TO mat(i1:n,i1:i2).
!
            if ( imax/=i ) then
!
                tmp_row(i1:i2)   = syst(imax,i1:i2)
                syst(imax,i1:i2) = syst(i,i1:i2)
                syst(i,i1:i2)    = tmp_row(i1:i2)
!
                equil(imax) = equil(i)
!
            else
!
                tmp_row(i+1_i4b:i2) = syst(i,i+1_i4b:i2)
!
            end if
!
!           STORE PERMUTATION.
!
            ip(i) = imax
!
#ifdef _BLAS
            syst(i,i) = piv
            piv       = one/piv
#else
            piv     = one/piv
            diag(i) = piv
#endif
            syst(i+1_i4b:n,i) = syst(i+1_i4b:n,i)*piv
!
!           REDUCE REMAINING MATRIX.
!
            do k = i+1_i4b, i2
!
#ifdef _BLAS
                call axpy( n-i, -tmp_row(k), syst(i+1_i4b:n,i), 1_i4b, syst(i+1_i4b:n,k), 1_i4b )
#else
                syst(i+1_i4b:n,k) = syst(i+1_i4b:n,k) - tmp_row(k)*syst(i+1_i4b:n,i)
#endif
!
            end do
!
        end do
!
!$OMP END SINGLE
!
        if ( failure ) exit
!
!$OMP DO SCHEDULE(STATIC)
!
        do l = i2+1_i4b, npb
!
!           APPLY PERMUTATIONS TO LATER COLUMNS, EG TO syst(i1:n,i2+1:npb).
!
            do i = i1, i2
!
                k         = ip(i)
                piv       = syst(k,l)
                syst(k,l) = syst(i,l)
                syst(i,l) = piv
!
            end do
!    
!           APPLY TRANSFORMATIONS TO LATER COLUMNS, EG TO syst(i1:n,i2+1:npb).
!    
            do i = i1, i2
!
#ifdef _BLAS
                call axpy( n-i, -syst(i,l), syst(i+1_i4b:n,i), 1_i4b, syst(i+1_i4b:n,l), 1_i4b )
#else
                syst(i+1_i4b:n,l) = syst(i+1_i4b:n,l) - syst(i,l)*syst(i+1_i4b:n,i)
#endif
!
            end do
!
        end do
!
!$OMP END DO
!          
    end do
!    
#ifndef _BLAS
    if ( .not.failure ) then
!    
!       BACK SUBSTITUTION.
!
!$OMP DO SCHEDULE(STATIC)
!
        do l = np1, npb
!
            do i = n, 1_i4b, -1_i4b
!
                piv = syst(i,l)*diag(i)
!
                syst(1_i4b:i-1_i4b,l) = syst(1_i4b:i-1_i4b,l) - piv*syst(1_i4b:i-1_i4b,i)
!
                syst(i,l) = piv
!
            end do
!    
            x(:n,l-n) = syst(1_i4b:n,l)
!
        end do
!
!$OMP END DO
!
    end if
#endif
!
!$OMP END PARALLEL
!
#ifdef _BLAS
    if ( .not.failure ) then
!    
!       BACK SUBSTITUTION.
!
        call trsm( 'L', 'U', 'N', 'N', n, nb, one, syst(1_i4b:n,1_i4b:n), n, syst(1_i4b:n,np1:npb), n )
!
        x(1_i4b:n,1_i4b:nb) = syst(1_i4b:n,np1:npb)
!
    else
!
!       mat IS SINGULAR.
!    
        x(:n,:nb) = nan()
!
    end if
#else
    if ( failure ) then
!
!       mat IS SINGULAR.
!    
        x(:n,:nb) = nan()
!
    end if
#endif
!
!
! END OF FUNCTION solve_lin_rm
! ____________________________
!
    end function solve_lin_rm
!
! =========================================================================================
!
    subroutine lin_lu_solve_rv( mat, b, failure, tol, small )
!
! Purpose
! _______
!
!   LIN_LU_SOLVE solves a system of linear equations
!
!                       MAT * X = B
!
!   with a n-by-n coefficient matrix MAT. B is a n-vector.
!
!   The LU decomposition with partial pivoting and implicit row scaling
!   of the matrix MAT
!
!                           P * MAT = L * U
!
!   where P is a permutation matrix, L is a n-by-n unit lower triangular matrix and
!   U is a n-by-n upper triangular matrix, is used to solve the linear system.
!
!
! Arguments
! _________
!
!   MAT         (INPUT/OUTPUT) real(stnd), dimension(:,:)
!               On entry, MAT contains the coefficient matrix of the equation
!
!                             MAT * X = B
!
!               On exit, MAT is destroyed.
!
!               The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   B           (INPUT/OUTPUT) real(stnd), dimension(:)
!               On entry, the right hand side vector B.
!
!               On exit, the solution vector X.
!
!               The shape of B must verify:  size( B ) = n .
!
!   FAILURE     (OUTPUT) logical(lgl)
!               On exit, if:
!
!               - FAILURE = true :  MAT is algorithmically singular
!               - FAILURE = false:  MAT is not singular.
!
!   TOL         (INPUT, OPTIONAL) real(stnd)
!               On entry, a relative tolerance used to indicate whether or not MAT
!               is nearly singular. Tol should normally be choose as approximately
!               the largest relative error in the elements of MAT. For example, 
!               if the elements of MAT are correct to about 4 significant figures,
!               then TOL should be set to about 5 * 10**(-4). If TOL is supplied as
!               less than EPS, where EPS is the relative machine precision, then
!               the value EPS is used in place of TOL.
!
!               Default: EPS, the relative machine precision.
!
!   SMALL       (INPUT, OPTIONAL) real(stnd)
!               On entry, if the system is singular, replaces a diagonal term of the 
!               matrix U if it is smaller in magnitude than the value SMALL using 
!               the same sign as the diagonal term. An approximate solution based 
!               on this replacement is obtained if no overflow results. If SMALL
!               is supplied as less than SAFMIN, the smallest number that can be 
!               reciprocated safely, then the value SAFMIN is used in place of SMALL.
!
!               Default: SAFMIN, the smallest number that can be reciprocated safely.
!
!
! Further Details
! _______________
!
!   MAT is declared singular if a diagonal element of U is such that
!
!             abs( U(j,j) ) <= n * norm( MAT(j,:) ) * TOL
!
!   where norm( MAT(j,:) ) denotes the maximum of the absolute values of 
!   the jth row of the matrix MAT. In this case, a diagonal element of U is
!   small, indicating that MAT is singular or nearly singular and FAILURE is set
!   to true. Otherwise, FAILURE is set to false.
!
!   If MAT is algorithmically singular (FAILURE=true), the diagonal terms of U 
!   smaller in magnitude than the value SMALL have been replaced by SMALL, using
!   the same sign as the diagonal terms, and the decomposition has been completed. 
!   An approximate solution based on this replacement is obtained if no overflow results.
!
!   A blocked algorithm is used to compute the factorization and solve the triangular systems.
!   Furthermore, the computations are parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Higham, N.J., 2011:
!           Gaussian elimination.
!           Wiley Interdisciplinary Reviews: Computational Statistics,
!           Vol. 3, Issue 3, pp 230-238.
!
!
! _________________________________________________________________________________________
!
! USED MODULES
! ____________
!
#ifdef _OPENMP
    use Char_Constants,  only : allocate_error
#endif
    use Num_Constants,   only : lamch
    use Char_Constants,  only : lin_error1
    use Utilities,       only : assert_eq, merror, imaxloc
#ifdef _BLAS
    use BLAS_interfaces, only : axpy, trsv
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout)         :: mat(:,:), b(:)
    real(stnd),   intent(in), optional  :: tol, small
!
    logical(lgl), intent(out)  :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                     :: tmp, piv, tol_piv, safmin
    real(stnd), dimension(size(b)) :: tmp_row, equil
!
    integer(i4b)                     :: n, i, imax, j, j1, j2, k, m, nblk, rem, n2
    integer(i4b), dimension(size(b)) :: ip
!
#ifdef _OPENMP
    integer      :: iok
    integer(i4b) :: l, l1, l2, iblk, num_threads, thread_num
    integer(kind=omp_lock_kind), dimension(:), allocatable :: lck
!
    logical(lgl) :: do_lock
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='lin_lu_solve'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(b),i4b)     ,   &
                    name_proc )
!
    failure = false
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!    
!   INITIALIZE TOLERANCE, SAFE MINIMUM AND IMPLICIT SCALING VECTOR.
!    
    tol_piv = lamch( 'u' )
    if ( present( tol ) ) tol_piv = max( tol, tol_piv )
    tol_piv = tol_piv*real( n, stnd )
!    
    safmin = lamch( 's' )
!    
    tmp_row(:n) = maxval( abs(mat(:n,:n)), dim=2 )
    if ( any( tmp_row(:n)<safmin ) ) call merror( name_proc//lin_error1 )
    equil(:n) = one/tmp_row(:n)
!    
    if ( present( small ) ) safmin = max( small, safmin )
!
!   DETERMINE THE NUMBER OF BLOCKS.
!
    nblk = n/blksz
    rem  = mod( n, blksz )
!
    if ( rem/=0_i4b ) then
!
        nblk = nblk + 1_i4b
        n2   = n - rem
!
    else
!
        n2   = n - blksz
!
    end if
!   
!   TRIANGULATION.  
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*(n+1_i4b)>=omp_limit              .and.      &
                  nblk>1_i4b                          .and.      &
                  i>1_i4b
!
    if ( test_par ) then
!
!       PARALLEL AND BLOCKED LU DECOMPOSITION OF mat.
!
!       ALLOCATE WORK ARRAY FOR STORING THE LOCKS FOR EACH BLOCK.
!
        allocate( lck(nblk), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SET UP LOCKS FOR EACH BLOCK.
!
        do i = 1_i4b, nblk
            call omp_init_lock( lck(i) )
        end do
!
!       COMPUTE LU FACTORIZATION OF THE FIRST BLOCK IN PLACE.
!
        do i = 1_i4b, blksz
!
!           FIND PIVOT.
!
            tmp_row(i:n) = equil(i:n)*abs( mat(i:n,i) )
            imax         = i - 1_i4b + imaxloc( tmp_row(i:n) )
            piv          = mat(imax,i)
!
!           TEST FOR SINGULARITY.
!
            if ( tmp_row(imax)<=tol_piv ) then
!
!               mat IS SINGULAR.
!
                failure = true
!
!               REPLACE PIVOT ELEMENT BY SAFE MINIMUM.
!
                piv = sign( max( safmin, abs(piv) ), piv )
!
            end if
!
!           IF NECESSARY, APPLY ROW INTERCHANGE.
!
            if ( imax/=i ) then
!
                tmp_row(1_i4b:blksz)  = mat(imax,1_i4b:blksz)
                mat(imax,1_i4b:blksz) = mat(i,1_i4b:blksz)
                mat(i,1_i4b:blksz)    = tmp_row(1_i4b:blksz)
!
                equil(imax) = equil(i)
!
                tmp     = b(imax)
                b(imax) = b(i)
                b(i)    = tmp
!
            else
!
                tmp_row(i+1_i4b:blksz) = mat(i,i+1_i4b:blksz)
!
            end if
!
!           STORE PERMUTATION.
!
            ip(i) = imax
!
#ifdef _BLAS
            mat(i,i) = piv
            piv      = one/piv
#else
            piv      = one/piv
            mat(i,i) = piv
#endif
            mat(i+1_i4b:n,i) = mat(i+1_i4b:n,i)*piv
!
!           REDUCE REMAINING MATRIX.
!
            do k = i+1_i4b, blksz
!
#ifdef _BLAS
                call axpy( n-i, -tmp_row(k), mat(i+1_i4b:n,i), 1_i4b, mat(i+1_i4b:n,k), 1_i4b )
#else
                mat(i+1_i4b:n,k) = mat(i+1_i4b:n,k) - tmp_row(k)*mat(i+1_i4b:n,i)
#endif
!
            end do
!
!           APPLY THE TRANSFORMATION TO b .
!
#ifdef _BLAS
            call axpy( n-i, -b(i), mat(i+1_i4b:n,i), 1_i4b, b(i+1_i4b:n), 1_i4b )
#else
            b(i+1_i4b:n) = b(i+1_i4b:n) - b(i)*mat(i+1_i4b:n,i)
#endif
!
        end do
!
!$OMP PARALLEL PRIVATE(i,imax,j,j1,j2,k,l,l1,l2,m,iblk,thread_num,num_threads,piv,tmp,do_lock)
!
        num_threads = omp_get_num_threads()
        thread_num  = omp_get_thread_num()
!
!       SET THE LOCKS FOR THE BLOCKS.
!
!$OMP DO SCHEDULE(STATIC,1)
!
        do i = 1_i4b, nblk
            call omp_set_lock( lck(i) )
        end do
!
!$OMP END DO
!
        if ( thread_num==0_i4b ) then
!
!           RELEASE THE FIRST LOCK.
!
            call omp_unset_lock( lck(1) )     
!
            do_lock = false
!
        else
!
            do_lock = true
!
        end if
!
!       START THE BLOCKED LU FACTORIZATION OF mat.
!
        do j = 0_i4b, n2, blksz
!
            iblk = j/blksz + 1_i4b
!
            if ( do_lock ) then
!
                call omp_set_lock( lck(iblk) )
!
! $OMP FLUSH (mat,b,equil,ip,failure)
!
                call omp_unset_lock( lck(iblk) )
!
            else
!
                do_lock = true
!
            end if
!
            j1 = j + 1_i4b
            j2 = min( j + blksz, n )
!
!           APPLY TRANSFORMATIONS IN THE iblk BLOCK TO LATER COLUMNS, EG TO mat(1:n,j2+1:n).
!
            k = ( (iblk/num_threads)*num_threads ) + thread_num
            if ( thread_num<mod(iblk,num_threads) ) k = k + num_threads
!
            do l = k, nblk-1_i4b, num_threads
!
!               COMPUTE INDICES FOR THE FIRST AND LAST COLUMNS OF THE l+1 BLOCK.
!
                l1 = l*blksz + 1_i4b
                l2 = min( (l+1_i4b)*blksz, n )
!
                do i = l1, l2
!
!                   APPLY PERMUTATIONS TO LATER COLUMNS, EG TO mat(j1:n,j2+1:n).
!
                    do m = j1, j2
!
                        k        = ip(m)
                        piv      = mat(k,i)
                        mat(k,i) = mat(m,i)
                        mat(m,i) = piv
!
                    end do
!    
!                   APPLY TRANSFORMATIONS TO LATER COLUMNS, EG TO mat(j1:n,j2+1:n).
!    
                    do m = j1, j2
!
#ifdef _BLAS
                        call axpy( n-m, -mat(m,i), mat(m+1_i4b:n,m), 1_i4b, mat(m+1_i4b:n,i), 1_i4b )
#else
                        mat(m+1_i4b:n,i) = mat(m+1_i4b:n,i) - mat(m,i)*mat(m+1_i4b:n,m)
#endif
!
                    end do
!
                end do
!
!               mat(j2+1_i4b:n,l1:l2) = mat(j2+1_i4b:n,l1:l2) - matmul( mat(j2+1_i4b:n,j1:j2) , mat(j1:j2,l1:l2) )
!
                if ( l==iblk ) then
!
!                   COMPUTE THE LU FACTORIZATION OF THE iblk+1 BLOCK, EG OF mat(j2+1:n,j2+1:j2+blksz), IN PLACE.
!
                    do i = l1, l2
!
!                       FIND PIVOT.
!
                        tmp_row(i:n) = equil(i:n)*abs( mat(i:n,i) )
                        imax         = i - 1_i4b + imaxloc( tmp_row(i:n) )
                        piv          = mat(imax,i)
!
!                       TEST FOR SINGULARITY.
!
                        if ( tmp_row(imax)<=tol_piv ) then
!
!                           mat IS SINGULAR.
!
                            failure = true
!
!                           REPLACE PIVOT ELEMENT BY SAFE MINIMUM.
!
                            piv = sign( max( safmin, abs(piv) ), piv )
!
                        end if
!
!                       IF NECESSARY, APPLY ROW INTERCHANGE.
!
                        if ( imax/=i ) then
!
                            tmp_row(l1:l2)  = mat(imax,l1:l2)
                            mat(imax,l1:l2) = mat(i,l1:l2)
                            mat(i,l1:l2)    = tmp_row(l1:l2)
!
                            equil(imax) = equil(i)
!
                            tmp     = b(imax)
                            b(imax) = b(i)
                            b(i)    = tmp
!
                        else
!
                            tmp_row(i+1_i4b:l2) = mat(i,i+1_i4b:l2)
!
                        end if
!
!                       STORE PERMUTATION.
!
                        ip(i) = imax
!
#ifdef _BLAS
                        mat(i,i)         = piv
                        piv              = one/piv
#else
                        piv              = one/piv
                        mat(i,i)         = piv
#endif
                        mat(i+1_i4b:n,i) = mat(i+1_i4b:n,i)*piv
!
!                       REDUCE REMAINING COLUMNS IN THE iblk+1 BLOCK.
!
                        do k = i+1_i4b, l2
!
#ifdef _BLAS
                            call axpy( n-i, -tmp_row(k), mat(i+1_i4b:n,i), 1_i4b, mat(i+1_i4b:n,k), 1_i4b )
#else
                            mat(i+1_i4b:n,k) = mat(i+1_i4b:n,k) - tmp_row(k)*mat(i+1_i4b:n,i)
#endif
!
                        end do
!
!                       APPLY THE TRANSFORMATION TO b .
!
#ifdef _BLAS
                        call axpy( n-i, -b(i), mat(i+1_i4b:n,i), 1_i4b, b(i+1_i4b:n), 1_i4b )
#else
                        b(i+1_i4b:n) = b(i+1_i4b:n) - b(i)*mat(i+1_i4b:n,i)
#endif
!
                    end do
!
!                   RELEASE THE iblk+1_i4b LOCK.
!
! $OMP FLUSH (mat,b,equil,ip,failure)
!
                    call omp_unset_lock( lck(iblk+1_i4b) )                 
!
                    do_lock = false
!
                end if
!
            end do
!
        end do
!
!$OMP END PARALLEL
!
!       DESTROY THE LOCKS.
!
        do i = 1_i4b, nblk
            call omp_destroy_lock( lck(i) )
        end do
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( lck )
!
    else
!
#endif
!
!       BLOCKED LU DECOMPOSITION OF mat.
!
        do j = 0_i4b, n2, blksz
!
            j1 = j + 1_i4b
            j2 = min( j + blksz, n )
!
!           COMPUTE THE LU FACTORIZATION OF mat(j1:n,j1:j2) IN PLACE.
!
            do i = j1, j2
!
!               FIND PIVOT.
!
                tmp_row(i:n) = equil(i:n)*abs( mat(i:n,i) )
                imax         = i - 1_i4b + imaxloc( tmp_row(i:n) )
                piv          = mat(imax,i)
!
!               TEST FOR SINGULARITY.
!
                if ( tmp_row(imax)<=tol_piv ) then
!
!                   mat IS SINGULAR.
!
                    failure = true
!
!                   REPLACE PIVOT ELEMENT BY SAFE MINIMUM.
!
                    piv = sign( max( safmin, abs(piv) ), piv )
!
                end if
!
!               IF NECESSARY, APPLY ROW INTERCHANGE.
!
                if ( imax/=i ) then
!
                    tmp_row(j1:j2)  = mat(imax,j1:j2)
                    mat(imax,j1:j2) = mat(i,j1:j2)
                    mat(i,j1:j2)    = tmp_row(j1:j2)
!
                    equil(imax) = equil(i)
!
                    tmp     = b(imax)
                    b(imax) = b(i)
                    b(i)    = tmp
!
                else
!
                    tmp_row(i+1_i4b:j2) = mat(i,i+1_i4b:j2)
!
                end if
!
!               STORE PERMUTATION.
!
                ip(i) = imax
!
#ifdef _BLAS
                mat(i,i) = piv
                piv      = one/piv
#else
                piv      = one/piv
                mat(i,i) = piv
#endif
                mat(i+1_i4b:n,i) = mat(i+1_i4b:n,i)*piv
!
!               REDUCE REMAINING MATRIX.
!
                do k = i+1_i4b, j2
!
#ifdef _BLAS
                    call axpy( n-i, -tmp_row(k), mat(i+1_i4b:n,i), 1_i4b, mat(i+1_i4b:n,k), 1_i4b )
#else
                    mat(i+1_i4b:n,k) = mat(i+1_i4b:n,k) - tmp_row(k)*mat(i+1_i4b:n,i)
#endif
!
                end do
!
!               APPLY THE TRANSFORMATION TO b .
!
#ifdef _BLAS
                call axpy( n-i, -b(i), mat(i+1_i4b:n,i), 1_i4b, b(i+1_i4b:n), 1_i4b )
#else
                b(i+1_i4b:n) = b(i+1_i4b:n) - b(i)*mat(i+1_i4b:n,i)
#endif
!
            end do
!
!           APPLY TRANSFORMATIONS TO LATER COLUMNS, EG TO mat(j1:n,j2+1:n).
!
            do i = j2+1_i4b, n
!
!               APPLY PERMUTATION TO LATER COLUMNS, EG TO mat(j1:n,j2+1:n).
!
                do m = j1, j2
!
                    k        = ip(m)
                    piv      = mat(k,i)
                    mat(k,i) = mat(m,i)
                    mat(m,i) = piv
!
                end do
!    
!               APPLY TRANSFORMATIONS TO LATER COLUMNS, EG TO mat(j1:n,j2+1:n).
!    
                do m = j1, j2
!
#ifdef _BLAS
                    call axpy( n-m, -mat(m,i), mat(m+1_i4b:n,m), 1_i4b, mat(m+1_i4b:n,i), 1_i4b )
#else
                    mat(m+1_i4b:n,i) = mat(m+1_i4b:n,i) - mat(m,i)*mat(m+1_i4b:n,m)
#endif
!
                end do
!
            end do
!          
        end do
!
#ifdef _OPENMP
    end if
#endif
!    
!   BACK SUBSTITUTION.
!
#ifdef _BLAS
    call trsv( 'U', 'N', 'N', n, mat(1_i4b:n,1_i4b:n), n, b(1_i4b:n), 1_i4b )
#else
    do i = n, 1_i4b, -1_i4b
!
        b(i) = b(i)*mat(i,i)
        b(1_i4b:i-1_i4b) = b(1_i4b:i-1_i4b) - b(i)*mat(1_i4b:i-1_i4b,i)
!
    end do
#endif
!
!
! END OF SUBROUTINE lin_lu_solve_rv
! _________________________________
!
    end subroutine lin_lu_solve_rv
!
! =========================================================================================
!
    subroutine lin_lu_solve_rm( mat, b, failure, tol, small )
!
! Purpose
! _______
!
!   LIN_LU_SOLVE solves a system of linear equations with several right hand sides
!
!                       MAT * X = B
!
!   with a n-by-n coefficient matrix MAT. B is a n-by-nb matrix.
!
!   The LU decomposition with partial pivoting and implicit row scaling
!   of the matrix MAT
!
!                           P * MAT = L * U
!
!   where P is a permutation matrix, L is a n-by-n unit lower triangular matrix and
!   U is a n-by-n upper triangular matrix, is used to solve the linear systems.
!
!
! Arguments
! _________
!
!   MAT         (INPUT/OUTPUT) real(stnd), dimension(:,:)
!               On entry, MAT contains the coefficient matrix of the equation
!
!                             MAT * X = B
!
!               On exit, MAT is destroyed.
!
!               The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   B           (INPUT/OUTPUT) real(stnd), dimension(:,:)
!               On entry, the right hand side matrix B.
!
!               On exit, the solution matrix X.
!
!               The shape of B must verify:  size( B, 1 ) = n .
!
!   FAILURE     (OUTPUT) logical(lgl)
!               On exit, if:
!
!               - FAILURE = true :  MAT is algorithmically singular
!               - FAILURE = false:  MAT is not singular.
!
!   TOL         (INPUT, OPTIONAL) real(stnd)
!               On entry, a relative tolerance used to indicate whether or not MAT
!               is nearly singular. Tol should normally be choose as approximately
!               the largest relative error in the elements of MAT. For example, 
!               if the elements of MAT are correct to about 4 significant figures,
!               then TOL should be set to about 5 * 10**(-4). If TOL is supplied as
!               less than EPS, where EPS is the relative machine precision, then
!               the value EPS is used in place of TOL.
!
!               Default: EPS, the relative machine precision.
!
!   SMALL       (INPUT, OPTIONAL) real(stnd)
!               On entry, if the system is singular, replaces a diagonal term of the 
!               matrix U if it is smaller in magnitude than the value SMALL using 
!               the same sign as the diagonal term. Approximate solutions based 
!               on this replacement are obtained if no overflow results. If SMALL
!               is supplied as less than SAFMIN, the smallest number that can be 
!               reciprocated safely, then the value SAFMIN is used in place of SMALL.
!
!               Default: SAFMIN, the smallest number that can be reciprocated safely.
!
!
! Further Details
! _______________
!
!   MAT is declared singular if a diagonal element of U is such that
!
!             abs( U(j,j) ) <= n * norm( MAT(j,:) ) * TOL
!
!   where norm( MAT(j,:) ) denotes the maximum of the absolute values of 
!   the jth row of the matrix MAT. In this case, a diagonal element of U is
!   small, indicating that MAT is singular or nearly singular and FAILURE is set
!   to true. Otherwise, FAILURE is set to false.
!
!   If MAT is algorithmically singular (FAILURE=true), the diagonal terms of U 
!   smaller in magnitude than the value SMALL have been replaced by SMALL, using
!   the same sign as the diagonal terms, and the decomposition has been completed. 
!   Approximate solutions based on this replacement are obtained if no overflow results.
!
!   A blocked algorithm is used to compute the factorization and solve the triangular systems.
!   Furthermore, the computations are parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Higham, N.J., 2011:
!           Gaussian elimination.
!           Wiley Interdisciplinary Reviews: Computational Statistics,
!           Vol. 3, Issue 3, pp 230-238.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
#ifdef _OPENMP
    use Char_Constants,  only : allocate_error
#endif
    use Num_Constants,   only : lamch
    use Char_Constants,  only : lin_error1
    use Utilities,       only : assert_eq, merror, imaxloc
#ifdef _BLAS
    use BLAS_interfaces, only : axpy, trsm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout), dimension(:,:) :: mat, b
    real(stnd),   intent(in), optional          :: tol, small
!
    logical(lgl), intent(out)  :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                       :: piv, tol_piv, safmin
    real(stnd), dimension(size(b,1)) :: tmp_row, equil
!
    integer(i4b)                       :: n, nb, i, imax, j, j1, j2, k, m, nblk, rem, n2
    integer(i4b), dimension(size(b,1)) :: ip
!
#ifdef _OPENMP
    integer      :: iok
    integer(i4b) :: l, l1, l2, iblk, num_threads, thread_num
    integer(kind=omp_lock_kind), dimension(:), allocatable :: lck
!
    logical(lgl) :: do_lock
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='lin_lu_solve'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(b,1),i4b)   ,   &
                    name_proc )
!
    nb =  size( b, 2 )
!
    failure = false
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. nb<=0_i4b ) return
!    
!   INITIALIZE TOLERANCE, SAFE MINIMUM AND IMPLICIT SCALING VECTOR.
!    
    tol_piv = lamch( 'u' )
    if ( present( tol ) ) tol_piv = max( tol, tol_piv )
    tol_piv = tol_piv*real( n, stnd )
!    
    safmin = lamch( 's' )
!    
    tmp_row(:n) = maxval( abs(mat(:n,:n)), dim=2 )
    if ( any( tmp_row(:n)<safmin ) ) call merror( name_proc//lin_error1 )
    equil(:n) = one/tmp_row(:n)
!    
    if ( present( small ) ) safmin = max( small, safmin )
!
!   DETERMINE THE NUMBER OF BLOCKS.
!
    nblk = n/blksz
    rem  = mod( n, blksz )
!
    if ( rem/=0_i4b ) then
!
        nblk = nblk + 1_i4b
        n2   = n - rem
!
    else
!
        n2   = n - blksz
!
    end if
!   
!   TRIANGULATION.  
!    
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*(n+nb)>=omp_limit                 .and.      &
                  nblk>1_i4b                          .and.      &
                  i>1_i4b
!
    if ( test_par ) then
!
!       PARALLEL AND BLOCKED LU DECOMPOSITION OF mat.
!
!       ALLOCATE WORK ARRAY FOR STORING THE LOCKS FOR EACH BLOCK.
!
        allocate( lck(nblk), stat = iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!
!       SET UP LOCKS FOR EACH BLOCK.
!
        do i = 1_i4b, nblk
            call omp_init_lock( lck(i) )
        end do
!
!       COMPUTE LU FACTORIZATION OF THE FIRST BLOCK IN PLACE.
!
        do i = 1_i4b, blksz
!
!           FIND PIVOT.
!
            tmp_row(i:n) = equil(i:n)*abs( mat(i:n,i) )
            imax         = i - 1_i4b + imaxloc( tmp_row(i:n) )
            piv          = mat(imax,i)
!
!           TEST FOR SINGULARITY.
!
            if ( tmp_row(imax)<=tol_piv ) then
!
!               mat IS SINGULAR.
!
                failure = true
!
!               REPLACE PIVOT ELEMENT BY SAFE MINIMUM.
!
                piv = sign( max( safmin, abs(piv) ), piv )
!
            end if
!
!           IF NECESSARY, APPLY ROW INTERCHANGE.
!
            if ( imax/=i ) then
!
                tmp_row(1_i4b:blksz)  = mat(imax,1_i4b:blksz)
                mat(imax,1_i4b:blksz) = mat(i,1_i4b:blksz)
                mat(i,1_i4b:blksz)    = tmp_row(1_i4b:blksz)
!
                equil(imax) = equil(i)
!
            else
!
                tmp_row(i+1_i4b:blksz) = mat(i,i+1_i4b:blksz)
!
            end if
!
!           STORE PERMUTATION.
!
            ip(i) = imax
!
#ifdef _BLAS
            mat(i,i) = piv
            piv      = one/piv
#else
            piv      = one/piv
            mat(i,i) = piv
#endif
            mat(i+1_i4b:n,i) = mat(i+1_i4b:n,i)*piv
!
!           REDUCE REMAINING MATRIX.
!
            do k = i+1_i4b, blksz
!
#ifdef _BLAS
                call axpy( n-i, -tmp_row(k), mat(i+1_i4b:n,i), 1_i4b, mat(i+1_i4b:n,k), 1_i4b )
#else
                mat(i+1_i4b:n,k) = mat(i+1_i4b:n,k) - tmp_row(k)*mat(i+1_i4b:n,i)
#endif
!
            end do
!
        end do
!
!$OMP PARALLEL PRIVATE(i,imax,j,j1,j2,k,l,l1,l2,m,iblk,thread_num,num_threads,piv,do_lock)
!
        num_threads = omp_get_num_threads()
        thread_num  = omp_get_thread_num()
!
!       SET THE LOCKS FOR THE BLOCKS.
!
!$OMP DO SCHEDULE(STATIC,1)
!
        do i = 1_i4b, nblk
            call omp_set_lock( lck(i) )
        end do
!
!$OMP END DO
!
        if ( thread_num==0_i4b ) then
!
!           RELEASE THE FIRST LOCK.
!
            call omp_unset_lock( lck(1) )     
!
            do_lock = false
!
        else
!
            do_lock = true
!
        end if
!
!       START THE BLOCKED LU FACTORIZATION OF mat.
!
        do j = 0_i4b, n2, blksz
!
            iblk = j/blksz + 1_i4b
!
            if ( do_lock ) then
!
                call omp_set_lock( lck(iblk) )
!
! $OMP FLUSH (mat,equil,ip,failure)
!
                call omp_unset_lock( lck(iblk) )
!
            else
!
                do_lock = true
!
            end if
!
            j1 = j + 1_i4b
            j2 = min( j + blksz, n )
!
!           APPLY TRANSFORMATIONS IN THE iblk BLOCK TO LATER COLUMNS, EG TO mat(1:n,j2+1:n).
!
            k = ( (iblk/num_threads)*num_threads ) + thread_num
            if ( thread_num<mod(iblk,num_threads) ) k = k + num_threads
!
            do l = k, nblk-1_i4b, num_threads
!
!               COMPUTE INDICES FOR THE FIRST AND LAST COLUMNS OF THE l+1 BLOCK.
!
                l1 = l*blksz + 1_i4b
                l2 = min( (l+1_i4b)*blksz, n )
!
                do i = l1, l2
!
!                   APPLY PERMUTATIONS TO LATER COLUMNS, EG TO mat(j1:n,j2+1:n).
!
                    do m = j1, j2
!
                        k        = ip(m)
                        piv      = mat(k,i)
                        mat(k,i) = mat(m,i)
                        mat(m,i) = piv
!
                    end do
!    
!                   APPLY TRANSFORMATIONS TO LATER COLUMNS, EG TO mat(j1:n,j2+1:n).
!    
                    do m = j1, j2
!
#ifdef _BLAS
                        call axpy( n-m, -mat(m,i), mat(m+1_i4b:n,m), 1_i4b, mat(m+1_i4b:n,i), 1_i4b )
#else
                        mat(m+1_i4b:n,i) = mat(m+1_i4b:n,i) - mat(m,i)*mat(m+1_i4b:n,m)
#endif
!
                    end do
!
                end do
!
!               mat(j2+1_i4b:n,l1:l2) = mat(j2+1_i4b:n,l1:l2) - matmul( mat(j2+1_i4b:n,j1:j2) , mat(j1:j2,l1:l2) )
!
                if ( l==iblk ) then
!
!                   COMPUTE THE LU FACTORIZATION OF THE iblk+1 BLOCK, EG OF mat(j2+1:n,j2+1:j2+blksz), IN PLACE.
!
                    do i = l1, l2
!
!                       FIND PIVOT.
!
                        tmp_row(i:n) = equil(i:n)*abs( mat(i:n,i) )
                        imax         = i - 1_i4b + imaxloc( tmp_row(i:n) )
                        piv          = mat(imax,i)
!
!                       TEST FOR SINGULARITY.
!
                        if ( tmp_row(imax)<=tol_piv ) then
!
!                           mat IS SINGULAR.
!
                            failure = true
!
!                           REPLACE PIVOT ELEMENT BY SAFE MINIMUM.
!
                            piv = sign( max( safmin, abs(piv) ), piv )
!
                        end if
!
!                       IF NECESSARY, APPLY ROW INTERCHANGE.
!
                        if ( imax/=i ) then
!
                            tmp_row(l1:l2)  = mat(imax,l1:l2)
                            mat(imax,l1:l2) = mat(i,l1:l2)
                            mat(i,l1:l2)    = tmp_row(l1:l2)
!
                            equil(imax) = equil(i)
!
                        else
!
                            tmp_row(i+1_i4b:l2) = mat(i,i+1_i4b:l2)
!
                        end if
!
!                       STORE PERMUTATION.
!
                        ip(i) = imax
!
#ifdef _BLAS
                        mat(i,i)         = piv
                        piv              = one/piv
#else
                        piv              = one/piv
                        mat(i,i)         = piv
#endif
                        mat(i+1_i4b:n,i) = mat(i+1_i4b:n,i)*piv
!
!                       REDUCE REMAINING COLUMNS IN THE iblk+1 BLOCK.
!
                        do k = i+1_i4b, l2
!
#ifdef _BLAS
                            call axpy( n-i, -tmp_row(k), mat(i+1_i4b:n,i), 1_i4b, mat(i+1_i4b:n,k), 1_i4b )
#else
                            mat(i+1_i4b:n,k) = mat(i+1_i4b:n,k) - tmp_row(k)*mat(i+1_i4b:n,i)
#endif
!
                        end do
!
                    end do
!
!                   RELEASE THE iblk+1_i4b LOCK.
!
! $OMP FLUSH (mat,equil,ip,failure)
!
                    call omp_unset_lock( lck(iblk+1_i4b) )                 
!
                    do_lock = false
!
                end if
!
            end do
!
!           APPLY TRANSFORMATIONS TO b.
!
!$OMP DO SCHEDULE(STATIC,1)
!          
            do i = 1_i4b, nb
!
!               APPLY PERMUTATIONS TO b.
!
                do m = j1, j2
!
                    k      = ip(m)
                    piv    = b(k,i)
                    b(k,i) = b(m,i)
                    b(m,i) = piv
!
                end do
!
!               APPLY TRANSFORMATIONS TO b.
!
                do m = j1, j2
!
#ifdef _BLAS
                    call axpy( n-m, -b(m,i), mat(m+1_i4b:n,m), 1_i4b, b(m+1_i4b:n,i), 1_i4b )
#else
                    b(m+1_i4b:n,i) = b(m+1_i4b:n,i) - b(m,i)*mat(m+1_i4b:n,m)
#endif
!
                end do
!
            end do
!
!$OMP END DO NOWAIT
!
        end do
!
#ifndef _BLAS
!       COMPUTE THE SOLUTION VECTORS BY BACK SUBSTITUTION.
!       BACK SUBSTITUTION.
!
!$OMP DO SCHEDULE(STATIC,1)
!          
        do j = 1_i4b, nb
!
            do i = n, 1_i4b, -1_i4b
!
                b(i,j) = b(i,j)*mat(i,i)
                b(1_i4b:i-1_i4b,j) = b(1_i4b:i-1_i4b,j) - b(i,j)*mat(1_i4b:i-1_i4b,i)
!
            end do
!
        end do
!
!$OMP END DO NOWAIT
#endif
!
!$OMP END PARALLEL
!
!       DESTROY THE LOCKS.
!
        do i = 1_i4b, nblk
            call omp_destroy_lock( lck(i) )
        end do
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( lck )
!
    else
!
#endif
!
!       BLOCKED LU DECOMPOSITION OF mat.
!
        do j = 0_i4b, n2, blksz
!
            j1 = j + 1_i4b
            j2 = min( j + blksz, n )
!
!           COMPUTE THE LU FACTORIZATION OF mat(j1:n,j1:j2) IN PLACE.
!
            do i = j1, j2
!
!               FIND PIVOT.
!
                tmp_row(i:n) = equil(i:n)*abs( mat(i:n,i) )
                imax         = i - 1_i4b + imaxloc( tmp_row(i:n) )
                piv          = mat(imax,i)
!
!               TEST FOR SINGULARITY.
!
                if ( tmp_row(imax)<=tol_piv ) then
!
!                   mat IS SINGULAR.
!
                    failure = true
!
!                   REPLACE PIVOT ELEMENT BY SAFE MINIMUM.
!
                    piv = sign( max( safmin, abs(piv) ), piv )
!
                end if
!
!               IF NECESSARY, APPLY ROW INTERCHANGE.
!
                if ( imax/=i ) then
!
                    tmp_row(j1:j2)  = mat(imax,j1:j2)
                    mat(imax,j1:j2) = mat(i,j1:j2)
                    mat(i,j1:j2)    = tmp_row(j1:j2)
!
                    equil(imax) = equil(i)
!
                else
!
                    tmp_row(i+1_i4b:j2) = mat(i,i+1_i4b:j2)
!
                end if
!
!               STORE PERMUTATION.
!
                ip(i) = imax
!
#ifdef _BLAS
                mat(i,i) = piv
                piv      = one/piv
#else
                piv      = one/piv
                mat(i,i) = piv
#endif
                mat(i+1_i4b:n,i) = mat(i+1_i4b:n,i)*piv
!
!               REDUCE REMAINING MATRIX.
!
                do k = i+1_i4b, j2
!
#ifdef _BLAS
                    call axpy( n-i, -tmp_row(k), mat(i+1_i4b:n,i), 1_i4b, mat(i+1_i4b:n,k), 1_i4b )
#else
                    mat(i+1_i4b:n,k) = mat(i+1_i4b:n,k) - tmp_row(k)*mat(i+1_i4b:n,i)
#endif
!
                end do
!
            end do
!
!           APPLY TRANSFORMATIONS TO LATER COLUMNS, EG TO mat(j1:n,j2+1:n).
!
            do i = j2+1_i4b, n
!
!               APPLY PERMUTATION TO LATER COLUMNS, EG TO mat(j1:n,j2+1:n).
!
                do m = j1, j2
!
                    k        = ip(m)
                    piv      = mat(k,i)
                    mat(k,i) = mat(m,i)
                    mat(m,i) = piv
!
                end do
!    
!               APPLY TRANSFORMATIONS TO LATER COLUMNS, EG TO mat(j1:n,j2+1:n).
!    
                do m = j1, j2
!
#ifdef _BLAS
                    call axpy( n-m, -mat(m,i), mat(m+1_i4b:n,m), 1_i4b, mat(m+1_i4b:n,i), 1_i4b )
#else
                    mat(m+1_i4b:n,i) = mat(m+1_i4b:n,i) - mat(m,i)*mat(m+1_i4b:n,m)
#endif
!
                end do
!
            end do
!
!           APPLY TRANSFORMATIONS TO b.
!          
            do i = 1_i4b, nb
!
!               APPLY PERMUTATIONS TO b.
!
                do m = j1, j2
!
                    k      = ip(m)
                    piv    = b(k,i)
                    b(k,i) = b(m,i)
                    b(m,i) = piv
!
                end do
!
!               APPLY TRANSFORMATIONS TO b.
!
                do m = j1, j2
!
#ifdef _BLAS
                    call axpy( n-m, -b(m,i), mat(m+1_i4b:n,m), 1_i4b, b(m+1_i4b:n,i), 1_i4b )
#else
                    b(m+1_i4b:n,i) = b(m+1_i4b:n,i) - b(m,i)*mat(m+1_i4b:n,m)
#endif
!
                end do
!
            end do
!          
        end do
!    
#ifndef _BLAS
!    
!       BACK SUBSTITUTION.
!
        do j = 1_i4b, nb
!
            do i = n, 1_i4b, -1_i4b
!
                b(i,j) = b(i,j)*mat(i,i)
                b(1_i4b:i-1_i4b,j) = b(1_i4b:i-1_i4b,j) - b(i,j)*mat(1_i4b:i-1_i4b,i)
!
            end do
!
        end do
#endif
!
#ifdef _OPENMP
    end if
#endif
#ifdef _BLAS
!    
!   BACK SUBSTITUTION.
!
    call trsm( 'L', 'U', 'N', 'N', n, nb, one, mat(1_i4b:n,1_i4b:n), n, b(1_i4b:n,1_i4b:nb), n )
!
#endif
!    
!
! END OF SUBROUTINE lin_lu_solve_rm
! _________________________________
!
    end subroutine lin_lu_solve_rm
!
! =========================================================================================
!
    subroutine lin_lu_solve_b_rv( mat, b, failure, x, tol, small )
!
! Purpose
! _______
!
!   LIN_LU_SOLVE solves a system of linear equations
!
!                       MAT * X = B
!
!   with a n-by-n coefficient matrix MAT. B and X are n-vectors.
!
!   The LU decomposition with partial pivoting and implicit row scaling
!   of the matrix MAT
!
!                           P * MAT = L * U
!
!   where P is a permutation matrix, L is a n-by-n unit lower triangular matrix and
!   U is a n-by-n upper triangular matrix, is used to solve the linear system.
!
!
! Arguments
! _________
!
!   MAT         (INPUT) real(stnd), dimension(:,:)
!               MAT contains the coefficient matrix of the equation
!
!                             MAT * X = B
!
!               MAT is not modified by the subroutine.
!
!               The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   B           (INPUT) real(stnd), dimension(:)
!               On entry, the right hand side vector B.
!               B is not modified by the subroutine.
!
!               The shape of B must verify:  size( B ) = n .
!
!   FAILURE     (OUTPUT) logical(lgl)
!               On exit, if:
!
!               - FAILURE = true :  MAT is algorithmically singular
!               - FAILURE = false:  MAT is not singular.
!
!   X           (OUTPUT) real(stnd), dimension(:)
!               On exit, the solution vector X.
!
!               The shape of X must verify:  size( X ) = n .
!
!   TOL         (INPUT, OPTIONAL) real(stnd)
!               On entry, a relative tolerance used to indicate whether or not MAT
!               is nearly singular. Tol should normally be choose as approximately
!               the largest relative error in the elements of MAT. For example, 
!               if the elements of MAT are correct to about 4 significant figures,
!               then TOL should be set to about 5 * 10**(-4). If TOL is supplied as
!               less than EPS, where EPS is the relative machine precision, then
!               the value EPS is used in place of TOL.
!
!               Default: EPS, the relative machine precision.
!
!   SMALL       (INPUT, OPTIONAL) real(stnd)
!               On entry, if the system is singular, replaces a diagonal term of the 
!               matrix U if it is smaller in magnitude than the value SMALL using 
!               the same sign as the diagonal term. An approximate solution based 
!               on this replacement is obtained if no overflow results. If SMALL
!               is supplied as less than SAFMIN, the smallest number that can be 
!               reciprocated safely, then the value SAFMIN is used in place of SMALL.
!
!               Default: SAFMIN, the smallest number that can be reciprocated safely.
!
!
! Further Details
! _______________
!
!   MAT is declared singular if a diagonal element of U is such that
!
!             abs( U(j,j) ) <= n * norm( MAT(j,:) ) * TOL
!
!   where norm( MAT(j,:) ) denotes the maximum of the absolute values of 
!   the jth row of the matrix MAT. In this case, a diagonal element of U is
!   small, indicating that MAT is singular or nearly singular and FAILURE is set
!   to true. Otherwise, FAILURE is set to false.
!
!   If MAT is algorithmically singular (FAILURE=true), the diagonal terms of U 
!   smaller in magnitude than the value SMALL have been replaced by SMALL, using
!   the same sign as the diagonal terms, and the decomposition has been completed. 
!   An approximate solution based on this replacement is obtained if no overflow results.
!
!   A blocked algorithm is used to compute the factorization and solve the triangular systems.
!   Furthermore, the computations are parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Higham, N.J., 2011:
!           Gaussian elimination.
!           Wiley Interdisciplinary Reviews: Computational Statistics,
!           Vol. 3, Issue 3, pp 230-238.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in)            :: mat(:,:), b(:)
    real(stnd),   intent(out)           :: x(:)
    real(stnd),   intent(in), optional  :: tol, small
!
    logical(lgl), intent(out)  :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(mat,1),size(mat,2)) :: mat2
!
    integer(i4b) :: n
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='lin_lu_solve'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(b),i4b)     ,   &
                    int(size(x),i4b)     ,   &
                    name_proc )
!
    failure = false
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!    
!   MAKE COPIES OF INPUT MATRIX AND RIGHT HAND SIDE VECTOR. 
!    
    mat2(:n,:n) = mat(:n,:n)
!
    x(:n) = b(:n)
!
    call lin_lu_solve_rv( mat2, x, failure, tol=tol, small=small )
!
!
! END OF SUBROUTINE lin_lu_solve_b_rv
! ___________________________________
!
    end subroutine lin_lu_solve_b_rv
!
! =========================================================================================
!
    subroutine lin_lu_solve_b_rm( mat, b, failure, x, tol, small )
!
! Purpose
! _______
!
!   LIN_LU_SOLVE solves a system of linear equations with several right hand sides
!
!                       MAT * X = B
!
!   with a n-by-n coefficient matrix MAT. B and X are n-by-nb matrices.
!
!   The LU decomposition with partial pivoting and implicit row scaling
!   of the matrix MAT
!
!                           P * MAT = L * U
!
!   where P is a permutation matrix, L is a n-by-n unit lower triangular matrix and
!   U is a n-by-n upper triangular matrix, is used to solve the linear systems.
!
!
! Arguments
! _________
!
!   MAT         (INPUT) real(stnd), dimension(:,:)
!               MAT contains the coefficient matrix of the equation
!
!                             MAT * X = B
!
!               MAT is not modified by the subroutine.
!
!               The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   B           (INPUT) real(stnd), dimension(:,:)
!               On entry, the right hand side matrix B.
!               B is not modified by the subroutine.
!
!               The shape of B must verify:  size( B, 1 ) = n .
!
!   FAILURE     (OUTPUT) logical(lgl)
!               On exit, if:
!
!               - FAILURE = true :  MAT is algorithmically singular
!               - FAILURE = false:  MAT is not singular.
!
!   X           (OUTPUT) real(stnd), dimension(:,:)
!               On exit, the solution matrix X.
!
!               The shape of X must verify:
!
!               - size( X, 1 ) = n ;
!               - size( X, 2 ) = size( B, 2 ) = nb .
!
!   TOL         (INPUT, OPTIONAL) real(stnd)
!               On entry, a relative tolerance used to indicate whether or not MAT
!               is nearly singular. Tol should normally be choose as approximately
!               the largest relative error in the elements of MAT. For example, 
!               if the elements of MAT are correct to about 4 significant figures,
!               then TOL should be set to about 5 * 10**(-4). If TOL is supplied as
!               less than EPS, where EPS is the relative machine precision, then
!               the value EPS is used in place of TOL.
!
!               Default: EPS, the relative machine precision.
!
!   SMALL       (INPUT, OPTIONAL) real(stnd)
!               On entry, if the system is singular, replaces a diagonal term of the 
!               matrix U if it is smaller in magnitude than the value SMALL using 
!               the same sign as the diagonal term. Approximate solutions based 
!               on this replacement are obtained if no overflow results. If SMALL
!               is supplied as less than SAFMIN, the smallest number that can be 
!               reciprocated safely, then the value SAFMIN is used in place of SMALL.
!
!               Default: SAFMIN, the smallest number that can be reciprocated safely.
!
!
! Further Details
! _______________
!
!   MAT is declared singular if a diagonal element of U is such that
!
!             abs( U(j,j) ) <= n * norm( MAT(j,:) ) * TOL
!
!   where norm( MAT(j,:) ) denotes the maximum of the absolute values of 
!   the jth row of the matrix MAT. In this case, a diagonal element of U is
!   small, indicating that MAT is singular or nearly singular and FAILURE is set
!   to true. Otherwise, FAILURE is set to false.
!
!   If MAT is algorithmically singular (FAILURE=true), the diagonal terms of U 
!   smaller in magnitude than the value SMALL have been replaced by SMALL, using
!   the same sign as the diagonal terms, and the decomposition has been completed. 
!   Approximate solutions based on this replacement are obtained if no overflow results.
!
!   A blocked algorithm is used to compute the factorization and solve the triangular systems.
!   Furthermore, the computations are parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Higham, N.J., 2011:
!           Gaussian elimination.
!           Wiley Interdisciplinary Reviews: Computational Statistics,
!           Vol. 3, Issue 3, pp 230-238.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in),  dimension(:,:) :: mat, b
    real(stnd),   intent(out), dimension(:,:) :: x
    real(stnd),   intent(in), optional        :: tol, small
!
    logical(lgl), intent(out)  :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(mat,1),size(mat,2)) :: mat2
!
    integer(i4b) :: n, nb
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='lin_lu_solve'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    int(size(b,1),i4b)   ,   &
                    int(size(x,1),i4b)   ,   &
                    name_proc )
!
    nb =  assert_eq( int(size(x,2),i4b)   ,   &
                     int(size(b,2),i4b)   ,   &
                     name_proc )
!
    failure = false
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. nb<=0_i4b ) return
!    
!   MAKE COPIES OF INPUT AND RIGHT HAND SIDE MATRICES. 
!    
    mat2(:n,:n) = mat(:n,:n)
!
    x(:n,:nb) = b(:n,:nb)
!
    call lin_lu_solve_rm( mat2, x, failure, tol=tol, small=small )
!
!
! END OF SUBROUTINE lin_lu_solve_b_rm
! ___________________________________
!
    end subroutine lin_lu_solve_b_rm
!
! =========================================================================================
!
    subroutine chol_solve_rv( mat, invdiag, b, upper )
!
! Purpose
! _______
!
!   CHOL_SOLVE solves a system of linear equations
!
!                       MAT * X = B
!
!   where MAT is a n-by-n symmetric positive definite matrix and B is a n-vector, using
!   the CHOLESKY factorization MAT = U' * U or MAT = L * L', as computed by CHOL_CMP or
!   GCHOL_CMP.
!
!
! Arguments
! _________
!
!   MAT          (INPUT) real(stnd), dimension(:,:)
!                On entry, the triangular factor U or L from the Cholesky factorisation,
!                as computed by CHOL_CMP, except for the main diagonal elements which
!                are stored in reciprocal form in INVDIAG.
!
!                The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   INVDIAG      (INPUT) real(stnd), dimension(:)
!                On entry, INVDIAG contains the reciprocals of the actual
!                diagonal elements of L or U, as computed by CHOL_CMP.
!
!                The shape of INVDIAG must verify:  size( INVDIAG ) = n .
!
!   B            (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the right hand side vector B.
!
!                On exit, the solution vector X.
!
!                The shape of B must verify:  size( B ) = n .
!
!   UPPER        (INPUT, OPTIONAL) logical(lgl)
!                Specifies whether the upper or lower triangular part of the
!                symmetric matrix MAT is stored. If:
!
!                - UPPER = true :  Upper triangular is stored
!                - UPPER = false:  Lower triangular is stored.
!
!                This argument must have the same value as used in CHOL_CMP
!                or GCHOL_CMP subroutines for computing the Cholesky factorisation.
!
!                The default is true.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert_eq
#ifdef _BLAS
    use BLAS_interfaces, only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in)    :: mat(:,:), invdiag(:)
    real(stnd),   intent(inout) :: b(:)
!
    logical(lgl), intent(in), optional :: upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: n, j
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='chol_solve'
!
    logical(lgl) :: mat_is_upper
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,    &
                    int(size(mat,2),i4b) ,    &
                    int(size(invdiag),i4b),   &
                    int(size(b),i4b),         &
                    name_proc )
!
    mat_is_upper = true
    if ( present(upper) ) mat_is_upper = upper
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!   
    if ( mat_is_upper ) then
!
!       SOLVE  mat*X = b  WHERE  mat = U'*U .
!
!
!       SOLVE  U'*X = b  OVERWRITING b WITH X .
!
        do j = 1_i4b, n
!
            b(j) = ( b(j) - dot_product( mat(1_i4b:j-1_i4b,j), b(1_i4b:j-1_i4b) ) )*invdiag(j)
!
        end do
!
!       SOLVE  U*X = b  OVERWRITING b WITH X .
!
        do j = n, 1_i4b, -1_i4b
!
            b(j) = b(j)*invdiag(j)
#ifdef _BLAS
            call axpy( j-1_i4b, -b(j), mat(1_i4b:j-1_i4b,j), 1_i4b, b(1_i4b:j-1_i4b), 1_i4b )
#else
            b(1_i4b:j-1_i4b) = b(1_i4b:j-1_i4b) - b(j)*mat(1_i4b:j-1_i4b,j)
#endif
!
        end do
!
    else
!
!       SOLVE  mat*X = b  WHERE  mat = L*L'.
!
!
!       SOLVE  L*X = b  OVERWRITING b WITH X .
!
        do j = 1_i4b, n
!
            b(j) = b(j)*invdiag(j)
#ifdef _BLAS
            call axpy( n-j, -b(j), mat(j+1_i4b:n,j), 1_i4b, b(j+1_i4b:n), 1_i4b )
#else
            b(j+1_i4b:n) = b(j+1_i4b:n) - b(j)*mat(j+1_i4b:n,j)
#endif
!
        end do
!
!       SOLVE  L'*X = b  OVERWRITING b WITH X .
!
        do j = n, 1_i4b, -1_i4b
!
            b(j) = ( b(j) - dot_product( mat(j+1_i4b:n,j), b(j+1_i4b:n) ) )*invdiag(j)
!
        end do
!
    end if
!
!
! END OF SUBROUTINE chol_solve_rv
! _______________________________
!
    end subroutine chol_solve_rv
!
! =========================================================================================
!
    subroutine chol_solve_rm( mat, invdiag, b, upper )
!
! Purpose
! _______
!
!   CHOL_SOLVE solves a system of linear equations with several right hand sides
!
!                       MAT * X = B
!
!   where MAT is a n-by-n symmetric positive (semi)-definite matrix and B is a n-by-nb matrix,
!   using the CHOLESKY factorization MAT = U' * U or MAT = L * L' as computed by CHOL_CMP or
!   GCHOL_CMP.
!
!
! Arguments
! _________
!
!   MAT          (INPUT) real(stnd), dimension(:,:)
!                On entry, the triangular factor U or L from the Cholesky factorisation,
!                as computed by CHOL_CMP, except for the main diagonal elements which
!                are stored in reciprocal form in INVDIAG.
!
!                The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   INVDIAG      (INPUT) real(stnd), dimension(:)
!                On entry, INVDIAG contains the reciprocals of the actual
!                diagonal elements of L or U, as computed by CHOL_CMP.
!
!                The shape of INVDIAG must verify:  size( INVDIAG ) = n .
!
!   B            (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the right hand side matrix B.
!
!                On exit, the solution matrix X.
!
!                The shape of B must verify:  size( B, 1 ) = n .
!
!   UPPER        (INPUT, OPTIONAL) logical(lgl)
!                Specifies whether the upper or lower triangular part of the
!                symmetric matrix MAT is stored. If:
!
!                - UPPER = true :  Upper triangular is stored
!                - UPPER = false:  Lower triangular is stored.
!
!                This argument must have the same value as used in CHOL_CMP
!                or GCHOL_CMP subroutines for computing the Cholesky factorisation.
!
!                The default is true.
!
!
! Further Details
! _______________
!
!   The computations are parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Higham, N.J., 2009:
!         Cholesky factorization.
!         Wiley Interdisciplinary Reviews: Computational Statistics,
!         Vol. 1, pp 251-254.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert_eq
#ifdef _BLAS
    use BLAS_interfaces, only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in)    :: mat(:,:), invdiag(:)
    real(stnd),   intent(inout) :: b(:,:)
!
    logical(lgl), intent(in), optional :: upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: n, nb, i, j
!
    logical(lgl) :: mat_is_upper
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='chol_solve'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,    &
                    int(size(mat,2),i4b) ,    &
                    int(size(invdiag),i4b),   &
                    int(size(b,1),i4b),       &
                    name_proc )
!
    nb = size( b, 2 )
!
    if ( present(upper) ) then
        mat_is_upper = upper
    else
        mat_is_upper = true
    end if
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. nb<=0_i4b ) return
!    
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.     &
                  nb>=2_i4b                           .and.     &
                  i>1_i4b
#endif
!    
!$OMP PARALLEL IF(test_par), PRIVATE(i,j),        &
!$OMP          SHARED(mat_is_upper)
!    
    if ( mat_is_upper ) then
!
!       SOLVE  mat*X = b  WHERE  mat = U'*U .
!    
!$OMP DO SCHEDULE(STATIC)
!
        do i = 1_i4b, nb
!
!           SOLVE  U'*X = b  OVERWRITING b WITH X .
!
            do j = 1_i4b, n
!
                b(j,i) = ( b(j,i) - dot_product( mat(1_i4b:j-1_i4b,j), b(1_i4b:j-1_i4b,i) ) )*invdiag(j)
!
            end do
!
!           SOLVE  U*X = b  OVERWRITING b WITH X .
!
            do j = n, 1_i4b, -1_i4b
!
                b(j,i) = b(j,i)*invdiag(j)
#ifdef _BLAS
                call axpy( j-1_i4b, -b(j,i), mat(1_i4b:j-1_i4b,j), 1_i4b, b(1_i4b:j-1_i4b,i), 1_i4b )
#else
                b(1_i4b:j-1_i4b,i) = b(1_i4b:j-1_i4b,i) - b(j,i)*mat(1_i4b:j-1_i4b,j)
#endif
!
            end do
!
        end do
!    
!$OMP END DO
!
    else
!
!       SOLVE  mat*X = b  WHERE  mat = L*L'.
!    
!$OMP DO SCHEDULE(STATIC)
!
        do i = 1_i4b, nb
!
!           SOLVE  L*X = b  OVERWRITING b WITH X .
!
            do j = 1_i4b, n
!
                b(j,i) = b(j,i)*invdiag(j)
#ifdef _BLAS
                call axpy( n-j, -b(j,i), mat(j+1_i4b:n,j), 1_i4b, b(j+1_i4b:n,i), 1_i4b )
#else
                b(j+1_i4b:n,i) = b(j+1_i4b:n,i) - b(j,i)*mat(j+1_i4b:n,j)
#endif
!
            end do
!
!           SOLVE  L'*X = b  OVERWRITING b WITH X .
!
            do j = n, 1_i4b, -1_i4b
!
                b(j,i) = ( b(j,i) - dot_product( mat(j+1_i4b:n,j), b(j+1_i4b:n,i) ) )*invdiag(j)
!
            end do
!
        end do
!    
!$OMP END DO
!
    end if
!    
!$OMP END PARALLEL
!
!   
! END OF SUBROUTINE chol_solve_rm
! _______________________________
!
    end subroutine chol_solve_rm
!
! =========================================================================================
!
    subroutine triang_solve_rv( mat, b, upper, trans )
!
! Purpose
! _______
!
!   TRIANG_SOLVE solves a triangular system of the form
!
!                 MAT * X = B   or   MAT' * X = B,
!
!   where MAT is a triangular matrix of order n, and B is an n-vector.
!
!   No test for singularity or near-singularity is included in this
!   routine. Such tests must be performed before calling this routine.
!
!
! Arguments
! _________
!
!   MAT          (INPUT) real(stnd), dimension(:,:)
!                The triangular matrix MAT. If:
!
!                - UPPER = true or is absent:
!                  The leading n-by-n upper triangular part of the array
!                  MAT contains the upper triangular matrix, and the strictly
!                  lower triangular part of MAT is not referenced.
!                - UPPER = false:
!                  The leading n-by-n lower triangular part of the array
!                  MAT contains the lower triangular matrix, and the strictly
!                  upper triangular part of MAT is not referenced.
!
!                The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   B            (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the right hand side vector B.
!
!                On exit, the solution vector X.
!
!                The shape of B must verify:  size( B ) = n .
!
!   UPPER        (INPUT, OPTIONAL) logical(lgl)
!                Specifies whether MAT is upper or lower triangular. If:
!
!                - UPPER = true :  MAT is upper triangular
!                - UPPER = false:  MAT is lower triangular.
!
!                The default is true.
!
!   TRANS        (INPUT, OPTIONAL) logical(lgl)
!                Specifies the form of the system of equations. If:
!
!                - TRANS = true :  MAT' * X = B    (Transpose)
!                - TRANS = false:  MAT * X = B     (No transpose)
!
!                The default is false.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert_eq
#ifdef _BLAS
    use BLAS_interfaces, only : trsv
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in)    :: mat(:,:)
    real(stnd),   intent(inout) :: b(:)
!
    logical(lgl), intent(in), optional :: upper, trans
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: n
#ifndef _BLAS
    integer(i4b) :: j
#endif
!
    logical(lgl) :: mat_is_upper, op_is_transpose
!
#ifdef _BLAS
    character    :: uplo, tran
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='triang_solve'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,    &
                    int(size(mat,2),i4b) ,    &
                    int(size(b),i4b),         &
                    name_proc )
!
    mat_is_upper = true
    if ( present(upper) ) mat_is_upper = upper
!
    op_is_transpose = false
    if ( present(trans) ) op_is_transpose = trans
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!   
#ifdef _BLAS
    if ( mat_is_upper ) then
        uplo = 'U'
    else
        uplo = 'L'
    end if
!
    if ( op_is_transpose ) then
        tran = 'T'
    else
        tran = 'N'
    end if
!
    call trsv( uplo, tran, 'N', n, mat, n, b, 1_i4b )    
!
#else
    if ( mat_is_upper ) then
!
!       mat IS UPPER TRIANGULAR.
!
        if ( op_is_transpose ) then
!
!           SOLVE  mat'*X = b  OVERWRITING b WITH X .
!
            do j = 1_i4b, n
!
                b(j) = ( b(j) - dot_product( mat(1_i4b:j-1_i4b,j), b(1_i4b:j-1_i4b) ) )/mat(j,j)
!
            end do
!
        else
!
!           SOLVE  mat*X = b  OVERWRITING b WITH X .
!
            do j = n, 1_i4b, -1_i4b
!
                b(j) = b(j)/mat(j,j)
                b(1_i4b:j-1_i4b) = b(1_i4b:j-1_i4b) - b(j)*mat(1_i4b:j-1_i4b,j)
!
            end do
!
        end if
!
    else
!
!       mat IS LOWER TRIANGULAR.
!
        if ( op_is_transpose ) then
!
!           SOLVE  mat'*X = b  OVERWRITING b WITH X .
!
            do j = n, 1_i4b, -1_i4b
!
                b(j) = ( b(j) - dot_product( mat(j+1_i4b:n,j), b(j+1_i4b:n) ) )/mat(j,j)
!
            end do
!
        else
!
!           SOLVE  mat*X = b  OVERWRITING b WITH X .
!
            do j = 1_i4b, n
!
                b(j) = b(j)/mat(j,j)
                b(j+1_i4b:n) = b(j+1_i4b:n) - b(j)*mat(j+1_i4b:n,j)
!
            end do
!
        end if
!
    end if
#endif
!
!
! END OF SUBROUTINE triang_solve_rv
! _________________________________
!
    end subroutine triang_solve_rv
!
! =========================================================================================
!
    subroutine triang_solve_b_rv( mat, b, scal, upper, trans )
!
! Purpose
! _______
!
!   TRIANG_SOLVE solves a nonsingular triangular linear system of the form
!
!                 MAT * X = B   or   MAT' * X = B,
!
!   where MAT is a triangular matrix of order n, and B is an n-vector.
!
!   The matrix MAT is assumed to be ill-conditioned, and frequent rescalings
!   are carried out in order to avoid overflow. However, no test for singularity
!   or near-singularity is included in this routine. Such tests must be performed
!   before calling this routine.
!
!
! Arguments
! _________
!
!   MAT          (INPUT) real(stnd), dimension(:,:)
!                The nonsingular triangular matrix MAT. If:  
!
!                - UPPER = true or is absent:
!                  The leading n-by-n upper triangular part of the array
!                  MAT contains the upper triangular matrix, and the strictly
!                  lower triangular part of MAT is not referenced.
!                - UPPER = false:
!                  The leading n-by-n lower triangular part of the array
!                  MAT contains the lower triangular matrix, and the strictly
!                  upper triangular part of MAT is not referenced.
!
!                The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   B            (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the right hand side vector B.
!
!                On exit, the scaled solution vector X.
!
!                The shape of B must verify:  size( B ) = n .
!
!   SCAL         (OUTPUT) real(stnd)
!                On exit, SCAL is a scaling factor introduced in order 
!                to avoid overflow. The solution of the given system of
!                equations is B(:)/SCAL . Note that SCAL may be negative.
!
!   UPPER        (INPUT, OPTIONAL) logical(lgl)
!                Specifies whether MAT is upper or lower triangular. If:
!
!                - UPPER = true :  MAT is upper triangular
!                - UPPER = false:  MAT is lower triangular.
!
!                The default is true.
!
!   TRANS        (INPUT, OPTIONAL) logical(lgl)
!                Specifies the form of the system of equations. If:
!
!                - TRANS = true :  MAT' * X = B    (Transpose)
!                - TRANS = false:  MAT * X = B     (No transpose)
!
!                The default is false.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert_eq
#ifdef _BLAS
    use BLAS_interfaces, only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in)    :: mat(:,:)
    real(stnd),   intent(inout) :: b(:)
    real(stnd),   intent(out)   :: scal
!
    logical(lgl), intent(in), optional :: upper, trans
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: tmp
!
    integer(i4b) :: n, j
!
    logical(lgl) :: mat_is_upper, op_is_transpose
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='triang_solve'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,    &
                    int(size(mat,2),i4b) ,    &
                    int(size(b),i4b),         &
                    name_proc )
!
    mat_is_upper = true
    if ( present(upper) ) mat_is_upper = upper
!
    op_is_transpose = false
    if ( present(trans) ) op_is_transpose = trans
!
!   QUICK RETURN IF POSSIBLE.
!
    scal = one
    if ( n<=0_i4b ) return
!   
    if ( mat_is_upper ) then
!
!       mat IS UPPER TRIANGULAR.
!
        if ( op_is_transpose ) then
!
!           SOLVE  mat'*X = b  OVERWRITING b WITH X .
!
            do j = 1_i4b, n
!
                tmp = b(j) - dot_product( mat(1_i4b:j-1_i4b,j), b(1_i4b:j-1_i4b) )
!
                if ( abs(tmp)>abs(mat(j,j)) ) then
                    tmp = mat(j,j)/tmp
                    b = b*tmp
                    b(j) = one
                    scal = tmp*scal
                else
                    b(j) = tmp/mat(j,j)
                end if
!
            end do
!
        else
!
!           SOLVE  mat*X = b  OVERWRITING b WITH X .
!
            do j = n, 1_i4b, -1_i4b
!
                if ( abs(b(j))>abs(mat(j,j)) ) then
                    tmp = mat(j,j)/b(j)
                    b = b*tmp
                    b(j) = one
                    scal = tmp*scal
                else
                    b(j) = b(j)/mat(j,j)
                end if
!
#ifdef _BLAS
                call axpy( j-1_i4b, -b(j), mat(1_i4b:j-1_i4b,j), 1_i4b, b(1_i4b:j-1_i4b), 1_i4b )
#else
                b(1_i4b:j-1_i4b) = b(1_i4b:j-1_i4b) - b(j)*mat(1_i4b:j-1_i4b,j)
#endif
!
            end do
!
        end if
!
    else
!
!       mat IS LOWER TRIANGULAR.
!
        if ( op_is_transpose ) then
!
!           SOLVE  mat'*X = b  OVERWRITING b WITH X .
!
            do j = n, 1_i4b, -1_i4b
!
                tmp = b(j) - dot_product( mat(j+1_i4b:n,j), b(j+1_i4b:n) )
!
                if ( abs(tmp)>abs(mat(j,j)) ) then
                    tmp = mat(j,j)/tmp
                    b = b*tmp
                    b(j) = one
                    scal = tmp*scal
                else
                    b(j) = tmp/mat(j,j)
                end if
!
            end do
!
        else
!
!           SOLVE  mat*X = b  OVERWRITING b WITH X .
!
            do j = 1_i4b, n
!
                if ( abs(b(j))>abs(mat(j,j)) ) then
                    tmp = mat(j,j)/b(j)
                    b = b*tmp
                    b(j) = one
                    scal = tmp*scal
                else
                    b(j) = b(j)/mat(j,j)
                end if
!
#ifdef _BLAS
                call axpy( n-j, -b(j), mat(j+1_i4b:n,j), 1_i4b, b(j+1_i4b:n), 1_i4b )
#else
                b(j+1_i4b:n) = b(j+1_i4b:n) - b(j)*mat(j+1_i4b:n,j)
#endif
!
            end do
!
        end if
!
    end if
!
!
! END OF SUBROUTINE triang_solve_b_rv
! ___________________________________
!
    end subroutine triang_solve_b_rv
!
! =========================================================================================
!
    subroutine triang_solve_rm( mat, b, upper, trans )
!
! Purpose
! _______
!
!   TRIANG_SOLVE solves a triangular system of the form
!
!                 MAT * X = B   or   MAT' * X = B,
!
!   where MAT is a triangular matrix of order n, and B is an n-by-nb matrix.
!
!   No test for singularity or near-singularity is included in this
!   routine. Such tests must be performed before calling this routine.
!
!
! Arguments
! _________
!
!   MAT          (INPUT) real(stnd), dimension(:,:)
!                The triangular matrix MAT. If:
!
!                - UPPER = true or is absent:
!                  The leading n-by-n upper triangular part of the array
!                  MAT contains the upper triangular matrix, and the strictly
!                  lower triangular part of MAT is not referenced.
!                - UPPER = false:
!                  The leading n-by-n lower triangular part of the array
!                  MAT contains the lower triangular matrix, and the strictly
!                  upper triangular part of MAT is not referenced.
!
!                The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   B            (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the right hand side matrix B.
!
!                On exit, the solution matrix X.
!
!                The shape of B must verify:  size( B, 1 ) = n .
!
!   UPPER        (INPUT, OPTIONAL) logical(lgl)
!                Specifies whether MAT is upper or lower triangular. If:
!
!                - UPPER = true :  MAT is upper triangular
!                - UPPER = false:  MAT is lower triangular.
!
!                The default is true.
!
!   TRANS        (INPUT, OPTIONAL) logical(lgl)
!                Specifies the form of the system of equations. If:
!
!                - TRANS = true :  MAT' * X = B    (Transpose)
!                - TRANS = false:  MAT * X = B     (No transpose)
!
!                The default is false.
!
!
! Further Details
! _______________
!
!   The computations are parallelized if OPENMP is used.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert_eq
#ifdef _BLAS
    use BLAS_interfaces, only : trsm
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in)    :: mat(:,:)
    real(stnd),   intent(inout) :: b(:,:)
!
    logical(lgl), intent(in), optional :: upper, trans
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: n, nb
!
    logical(lgl) :: mat_is_upper, op_is_transpose
!
#ifdef _BLAS
    character :: uplo, tran
#else
#ifdef _OPENMP
    integer(i4b) :: j
!
    logical :: test_par, ompparallel, ompnested
#endif
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='triang_solve'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,    &
                    int(size(mat,2),i4b) ,    &
                    int(size(b,1),i4b),       &
                    name_proc )
!
    nb = size( b, 2 )
!
    mat_is_upper = true
    if ( present(upper) ) mat_is_upper = upper
!
    op_is_transpose = false
    if ( present(trans) ) op_is_transpose = trans
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b .or. nb<=0_i4b ) return
!    
#ifdef _BLAS
    if ( mat_is_upper ) then
        uplo = 'U'
    else
        uplo = 'L'
    end if
!
    if ( op_is_transpose ) then
        tran = 'T'
    else
        tran = 'N'
    end if
!
    call trsm( 'L', uplo, tran, 'N', n, nb, one, mat, n, b, n )    
#else
!    
#ifdef _OPENMP
    j           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.     &
                  nb>=2_i4b                           .and.     &
                  j>1_i4b
#endif
!    
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC,1), PRIVATE(j),           &
!$OMP             SHARED(mat_is_upper,op_is_transpose)
!
    do j = 1_i4b, nb
        call triang_solve_rv( mat, b(:,j), mat_is_upper, op_is_transpose )
    end do
!    
!$OMP END PARALLEL DO
!
#endif
!
!
! END OF SUBROUTINE triang_solve_rm
! _________________________________
!
    end subroutine triang_solve_rm
!
! =========================================================================================
!
    subroutine triang_solve_b_rm( mat, b, scal, upper, trans )
!
! Purpose
! _______
!
!   TRIANG_SOLVE solves a nonsingular triangular linear system of the form
!
!                 MAT * X = B   or   MAT' * X = B,
!
!   where MAT is a triangular matrix of order n, and B is an n-by-nb matrix.
!
!   The matrix MAT is assumed to be ill-conditioned, and frequent rescalings
!   are carried out in order to avoid overflow. However, no test for singularity
!   or near-singularity is included in this routine. Such tests must be performed
!   before calling this routine.
!
!
! Arguments
! _________
!
!   MAT          (INPUT) real(stnd), dimension(:,:)
!                The nonsingular triangular matrix MAT. If: 
!
!                - UPPER = true or is absent:
!                  The leading n-by-n upper triangular part of the array
!                  MAT contains the upper triangular matrix, and the strictly
!                  lower triangular part of MAT is not referenced.
!                - UPPER = false:
!                  The leading n-by-n lower triangular part of the array
!                  MAT contains the lower triangular matrix, and the strictly
!                  upper triangular part of MAT is not referenced.
!
!                The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   B            (INPUT/OUTPUT) real(stnd), dimension(:,:)
!                On entry, the right hand side matrix B.
!
!                On exit, the scaled solution matrix X.
!
!                The shape of B must verify:
!
!                - size( B, 1 ) = n ,
!                - size( B, 2 ) = size( SCAL ) = nb .
!
!   SCAL         (OUTPUT) real(stnd), dimension(:)
!                On exit, SCAL is a scaling vector introduced in order 
!                to avoid overflow. The solution of the given system of
!                equations is B/spread(SCAL,dim=1, ncopies=n) .
!                Note that elements of SCAL may be negative.
!
!                The size of SCAL must verify:  size( SCAL ) = size( B, 2 ) = nb .
!
!   UPPER        (INPUT, OPTIONAL) logical(lgl)
!                Specifies whether MAT is upper or lower triangular. If:
!
!                - UPPER = true :  MAT is upper triangular
!                - UPPER = false:  MAT is lower triangular.
!
!                The default is true.
!
!   TRANS        (INPUT, OPTIONAL) logical(lgl)
!                Specifies the form of the system of equations. If:
!
!                - TRANS = true :  MAT' * X = B    (Transpose)
!                - TRANS = false:  MAT * X = B     (No transpose)
!
!                The default is false.
!
!
! Further Details
! _______________
!
!   The computations are parallelized if OPENMP is used.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,     only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in)    :: mat(:,:)
    real(stnd),   intent(inout) :: b(:,:)
    real(stnd),   intent(out)   :: scal(:)
!
    logical(lgl), intent(in), optional :: upper, trans
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: n, nb, j
!
    logical(lgl) :: mat_is_upper, op_is_transpose
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='triang_solve'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,    &
                    int(size(mat,2),i4b) ,    &
                    int(size(b,1),i4b),       &
                    name_proc )
!
    nb =  assert_eq( int(size(b,2),i4b),       &
                     int(size(scal),i4b),      &
                     name_proc )
!
    mat_is_upper = true
    if ( present(upper) ) mat_is_upper = upper
!
    op_is_transpose = false
    if ( present(trans) ) op_is_transpose = trans
!
!   QUICK RETURN IF POSSIBLE.
!
    scal(:nb) = one
    if ( n<=0_i4b .or. nb<=0_i4b ) return
!    
#ifdef _OPENMP
    j           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.     &
                  nb>=2_i4b                           .and.     &
                  j>1_i4b
#endif
!    
!$OMP PARALLEL DO IF(test_par), SCHEDULE(STATIC,1), PRIVATE(j),      &
!$OMP             SHARED(mat_is_upper,op_is_transpose)
!
    do j = 1_i4b, nb
        call triang_solve_b_rv( mat, b(:n,j), scal(j), mat_is_upper, op_is_transpose )
    end do
!    
!$OMP END PARALLEL DO
!
!
! END OF SUBROUTINE triang_solve_b_rm
! ___________________________________
!
    end subroutine triang_solve_b_rm
!
!
! =========================================================================================
!                            COMPUTING THE INVERSE OF A MATRIX
! =========================================================================================
!
!
    subroutine comp_inv_a( mat, failure, tol )
!
! Purpose
! _______
!
!   COMP_INV computes, in place, the inverse of a matrix MAT.
!
!
! Arguments
! _________
!
!   MAT         (INPUT/OUTPUT) real(stnd), dimension(:,:)
!               On entry, MAT contains the matrix to be inverted.
!
!               On exit, MAT is replaced by its inverse.
!
!               The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   FAILURE     (OUTPUT) logical(lgl)
!               On exit, if:
!
!               - FAILURE = true :  MAT is algorithmically singular.
!               - FAILURE = false:  MAT has been inverted.
!
!   TOL         (INPUT, OPTIONAL) real(stnd)
!               On entry, a relative tolerance used to indicate whether or not MAT
!               is nearly singular. Tol should normally be choose as approximately
!               the largest relative error in the elements of MAT. For example, 
!               if the elements of MAT are correct to about 4 significant figures,
!               then TOL should be set to about 5 * 10**(-4). If TOL is supplied as
!               less than EPS, where EPS is the relative machine precision, then
!               the value EPS is used in place of TOL.
!
!               Default: EPS, the relative machine precision.
!
!
! Further Details
! _______________
!
!   The LU decomposition with partial pivoting and implicit row
!   scaling of the matrix MAT is used to compute the inverse.
!
!   MAT is declared singular if a diagonal element of U is such that
!
!             abs( U(j,j) ) <= n * norm( MAT(j,:) ) * TOL
!
!   where norm( MAT(j,:) ) denotes the maximum of the absolute values of 
!   the jth row of the matrix MAT. In this case, a diagonal element of U is
!   small, indicating that MAT is singular or nearly singular and FAILURE is
!   set to true. Otherwise, FAILURE is set to false.
!
!   On exit, if FAILURE=true, MAT is filled with nan() value.
!
!   The computations are parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Higham, N.J., 2011:
!         Gaussian elimination.
!         Wiley Interdisciplinary Reviews: Computational Statistics,
!         Vol. 3, Issue 3, pp 230-238.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : nan
    use Utilities,       only : assert_eq
#ifdef _BLAS
    use BLAS_interfaces, only : axpy, swap
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout)         :: mat(:,:)
    real(stnd),   intent(in), optional  :: tol
!
    logical(lgl), intent(out) :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                         :: d1
    real(stnd), dimension(size(mat,1)) :: tmp_vec, tmp_vec2
!
    integer(i4b)                         :: n, i, j, n2
    integer(i4b), dimension(size(mat,1)) :: ip
!
#ifdef _OPENMP
!
    real(stnd)                         :: tmp, matii
    real(stnd), dimension(size(mat,1)) :: tmp_vec3
!
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_inv'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq( int(size(mat,1),i4b) ,      &
                   int(size(mat,2),i4b) ,      &
                   name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    failure = false
    if ( n<=0_i4b ) return
!    
!   DECOMPOSE mat .
!    
    call lu_cmp( mat, ip, d1, tol=tol )
!    
!   SET FAILURE INDICATOR.
!    
    failure = d1==zero
!
    if ( failure ) then
!    
!       mat IS SINGULAR .
!    
        mat(:,:) = nan()
!
    else
!    
!       MATRIX INVERSION.
!
#ifdef _OPENMP
        i           = omp_get_max_threads()
        ompparallel = omp_in_parallel()
        ompnested   = omp_get_nested() .and. ompparallel
        test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                      n*n>=omp_limit                      .and.      &
                      i>1_i4b
!    
        if ( test_par ) then
!
            tmp = sqrt( real( omp_limit, stnd ) )
            n2  = max( n - int( tmp, i4b ), 1_i4b )
!
        else
#endif
!
            n2 = 1_i4b
!
#ifdef _OPENMP
        end if
#endif
!
        do i = n-1_i4b, n2, -1_i4b
!
            tmp_vec(i+1_i4b:n)  = -mat(i+1_i4b:n,i)
            tmp_vec2(i+1_i4b:n) = -mat(i,i+1_i4b:n)
!
!            mat(i+1_i4b:n,i)   = matmul( mat(i+1_i4b:n,i+1_i4b:n), tmp_vec(i+1_i4b:n) )
!            tmp_vec(i+1_i4b:n) = matmul( tmp_vec2(i+1_i4b:n), mat(i+1_i4b:n,i+1_i4b:n) )
!
            mat(i+1_i4b:n,i)   = tmp_vec(i+1_i4b)*mat(i+1_i4b:n,i+1_i4b)
            tmp_vec(i+1_i4b)   = dot_product( tmp_vec2(i+1_i4b:n), mat(i+1_i4b:n,i+1_i4b) )
!
            do j = i+2_i4b, n
!
#ifdef _BLAS
                call axpy( n-i, tmp_vec(j), mat(i+1_i4b:n,j), 1_i4b, mat(i+1_i4b:n,i), 1_i4b )
#else
                mat(i+1_i4b:n,i) = mat(i+1_i4b:n,i) + tmp_vec(j)*mat(i+1_i4b:n,j)
#endif
                tmp_vec(j)       = dot_product( tmp_vec2(i+1_i4b:n), mat(i+1_i4b:n,j) )
!
            end do
!
            tmp_vec(i) = one + dot_product( tmp_vec2(i+1_i4b:n), mat(i+1_i4b:n,i) )
            mat(i,i:n) = tmp_vec(i:n)*mat(i,i)
!
        end do
!
#ifdef _OPENMP
!
        if ( n2>1_i4b ) then
!
!$OMP PARALLEL PRIVATE(i,j,tmp,matii,tmp_vec3)
!
            do i = n2-1_i4b, 1_i4b, -1_i4b
!
                tmp_vec3(i+1_i4b:n) = zero
                matii               = mat(i,i)
!
!$OMP SINGLE
!
                tmp_vec(i+1_i4b:n)  = -mat(i+1_i4b:n,i)
                tmp_vec2(i+1_i4b:n) = -mat(i,i+1_i4b:n)
!
                mat(i+1_i4b:n,i) = tmp_vec(i+1_i4b)*mat(i+1_i4b:n,i+1_i4b)
                tmp              = dot_product( tmp_vec2(i+1_i4b:n), mat(i+1_i4b:n,i+1_i4b) )
                mat(i,i+1_i4b)   = tmp*matii
!
!$OMP END SINGLE
!
!$OMP DO SCHEDULE(STATIC)
!
                do j = i+2_i4b, n
!
#ifdef _BLAS
                    call axpy( n-i, tmp_vec(j), mat(i+1_i4b:n,j), 1_i4b, tmp_vec3(i+1_i4b:n), 1_i4b )
#else
                    tmp_vec3(i+1_i4b:n) = tmp_vec3(i+1_i4b:n) + tmp_vec(j)*mat(i+1_i4b:n,j)
#endif
                    tmp                 = dot_product( tmp_vec2(i+1_i4b:n), mat(i+1_i4b:n,j) )
                    mat(i,j)            = tmp*matii
!
                end do
!
!$OMP END DO NOWAIT
!
!$OMP CRITICAL (matinv)
!
                mat(i+1_i4b:n,i) = mat(i+1_i4b:n,i) + tmp_vec3(i+1_i4b:n)
!
!$OMP END CRITICAL (matinv)
!
!$OMP BARRIER
!
!$OMP SINGLE
!
                tmp      = one + dot_product( tmp_vec2(i+1_i4b:n), mat(i+1_i4b:n,i) )
                mat(i,i) = tmp*matii
!
!$OMP END SINGLE
!
            end do
!
!$OMP END PARALLEL
!
        end if
#endif
!    
!       PERMUTE COLUMNS TO GET INVERSE OF mat .
!    
        do i = n, 1_i4b, -1_i4b
!
            j = ip(i)
!
            if ( i==j ) cycle
!
#ifdef _BLAS
            call swap( n, mat(:n,j), 1_i4b, mat(:n,i), 1_i4b )
#else
            tmp_vec(:n) = mat(:n,j)
            mat(:n,j)   = mat(:n,i)
            mat(:n,i)   = tmp_vec(:n)
#endif
!
        end do
!
    end if
!
!
! END OF SUBROUTINE comp_inv_a
! ____________________________
!
    end subroutine comp_inv_a
!
! =========================================================================================
!
    subroutine comp_inv_b( mat, failure, matinv, tol )
!
!
! Purpose
! _______
!
!   COMP_INV computes the inverse of a matrix MAT.
!
!
! Arguments
! _________
!
!   MAT         (INPUT/OUTPUT) real(stnd), dimension(:,:)
!               On entry, MAT contains the matrix to be inverted.
!
!               On exit, MAT is replaced by the LU decomposition of
!               a rowwise permutation of MAT.
!
!               The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   FAILURE     (OUTPUT) logical(lgl)
!               On exit, if:
!
!               - FAILURE = true :  MAT is algorithmically singular.
!               - FAILURE = false:  MAT has been inverted.
!
!   MATINV      (OUTPUT) real(stnd), dimension(:,:)
!                On exit, if MAT is not singular, MATINV contains the inverse of MAT.
!
!               The shape of MATINV must verify:  size( MATINV, 1 ) = size( MATINV, 2 ) = n .
!
!   TOL         (INPUT, OPTIONAL) real(stnd)
!               On entry, a relative tolerance used to indicate whether or not MAT
!               is nearly singular. Tol should normally be choose as approximately
!               the largest relative error in the elements of MAT. For example, 
!               if the elements of MAT are correct to about 4 significant figures,
!               then TOL should be set to about 5 * 10**(-4). If TOL is supplied as
!               less than EPS, where EPS is the relative machine precision, then
!               the value EPS is used in place of TOL.
!
!               Default: EPS, the relative machine precision.
!
!
! Further Details
! _______________
!
!   MAT is modified by COMP_INV.
!
!   The LU decomposition with partial pivoting and implicit row
!   scaling of the matrix MAT is used to compute the inverse.
!
!   MAT is declared singular if a diagonal element of U is such that
!
!             abs( U(j,j) ) <= n * norm( MAT(j,:) ) * TOL
!
!   where norm( MAT(j,:) ) denotes the maximum of the absolute values of 
!   the jth row of the matrix MAT. In this case, a diagonal element of U is
!   small, indicating that MAT is singular or nearly singular and FAILURE is
!   set to true. Otherwise, FAILURE is set to false.
!
!   On exit, if FAILURE=true, MATINV is filled with nan() value.
!
!   The computations are parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Higham, N.J., 2011:
!         Gaussian elimination.
!         Wiley Interdisciplinary Reviews: Computational Statistics,
!         Vol. 3, Issue 3, pp 230-238.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants, only : nan
    use Utilities,     only : assert_eq, unit_matrix
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout)         :: mat(:,:)
    real(stnd),   intent(out)           :: matinv(:,:)
    real(stnd),   intent(in), optional  :: tol
!
    logical(lgl), intent(out) :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)  :: d1
!
    integer(i4b) :: ip(size(mat,1)), n
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_inv'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq( int(size(mat,1),i4b) ,      &
                   int(size(mat,2),i4b) ,      &
                   int(size(matinv,1),i4b) ,   &
                   int(size(matinv,2),i4b) ,   &
                   name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    failure = false
    if ( n<=0_i4b ) return
!    
!   DECOMPOSE mat .
!    
    call lu_cmp( mat, ip, d1, tol=tol )
!    
!   SET FAILURE INDICATOR.
!    
    failure = d1==zero
!
    if ( failure ) then
!    
!       mat IS SINGULAR .
!    
        matinv(:,:) = nan()
!
    else
!    
!       mat IS REGULAR, COMPUTE INVERSE OF mat.
!    
!       FIRST SET matinv TO IDENTITY MATRIX.
!
        call unit_matrix( matinv )
!    
!       CALL ROUTINE lu_solve TO SOLVE MAT*MATINV = I .
!    
        call lu_solve_rm( mat, ip, matinv )
!        call lu_solve2_rm( mat, ip, matinv )
!
    end if
!
!
! END OF SUBROUTINE comp_inv_b
! ____________________________
!
    end subroutine comp_inv_b
!
! =========================================================================================
!
    function inv( mat, tol ) result( matinv )
!
!
! Purpose
! _______
!
!   INV computes the inverse of a real matrix MAT,
!
!                MAT * INV(MAT) = I
!
!
! Arguments
! _________
!
!   MAT     (INPUT) real(stnd), dimension(:,:)
!           On entry, MAT contains the matrix to be inverted.
!           MAT is not modified by the function.
!
!           The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   TOL     (INPUT, OPTIONAL) real(stnd)
!           On entry, a relative tolerance used to indicate whether or not MAT
!           is nearly singular. Tol should normally be choose as approximately
!           the largest relative error in the elements of MAT. For example, 
!           if the elements of MAT are correct to about 4 significant figures,
!           then TOL should be set to about 5 * 10**(-4). If TOL is supplied as
!           less than EPS, where EPS is the relative machine precision, then
!           the value EPS is used in place of TOL.
!
!           Default: EPS, the relative machine precision.
!
!
! Further Details
! _______________
!
!   MAT is not modified by function INV.
!
!   The LU decomposition with partial pivoting and implicit row
!   scaling of the matrix MAT is used to compute the inverse.
!
!   MAT is declared singular if a diagonal element of U is such that
!
!             abs( U(j,j) ) <= n * norm( MAT(j,:) ) * TOL
!
!   where norm( MAT(j,:) ) denotes the maximum of the absolute values of 
!   the jth row of the matrix MAT. In this case, a diagonal element of U is
!   small, indicating that MAT is singular or nearly singular.
!
!   On exit, if MAT is algorithmically singular, the function INV returns 
!   a n-by-n matrix filled with nan() value.
!
!   The computations are parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Higham, N.J., 2011:
!         Gaussian elimination.
!         Wiley Interdisciplinary Reviews: Computational Statistics,
!         Vol. 3, Issue 3, pp 230-238.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,     only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in)            :: mat(:,:)
    real(stnd),   intent(in), optional  :: tol
!
    real(stnd) :: matinv(size(mat,1),size(mat,1))
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: n
!
    logical(lgl) :: failure
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='inv'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq( int(size(mat,1),i4b) ,      &
                   int(size(mat,2),i4b) ,      &
                   name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!    
!   INVERSE mat IN matinv .
!    
    matinv(:n,:n) = mat(:n,:n)
!
    call comp_inv_a( matinv, failure, tol )
!
!
! END OF FUNCTION inv
! ___________________
!
    end function inv
!
! =========================================================================================
!
    subroutine comp_sym_inv_a( mat, failure, upper, fill, tol )
!
! Purpose
! _______
!
!   COMP_SYM_INV computes, in place, the inverse of a real symmetric positive definite
!   matrix MAT using the Cholesky factorization MAT = U' * U or MAT = L * L'.
!
!
! Arguments
! _________
!
!   MAT         (INPUT/OUTPUT) real(stnd), dimension(:,:)
!               On entry, the symmetric matrix to be inverted:
!
!               - If UPPER = true or is absent:
!                 The leading n-by-n upper triangular part of MAT contains the
!                 upper triangular part of the matrix and the strictly
!                 lower triangular part of MAT is not referenced.
!
!               - If UPPER = false:
!                 The leading n-by-n lower triangular part of MAT contains the
!                 lower triangular part of the matrix and the strictly
!                 upper triangular part of MAT is not referenced.
!
!               On exit:
!
!               - If FILL = true or is absent:
!                 The (symmetric) inverse of MAT overwrites MAT.
!               - If FILL = false:
!                 The upper (if UPPER= true) or lower (if UPPER= false)
!                 triangle of the (symmetric) inverse of MAT, overwrites the 
!                 input upper or lower triangular part of MAT and the other
!                 part of MAT is not modified.
!
!               The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   FAILURE     (OUTPUT) logical(lgl)
!               On exit, if:
!
!               - FAILURE = true :  MAT is algorithmically not positive definite.
!               - FAILURE = false:  MAT has been inverted.
!
!   UPPER       (INPUT, OPTIONAL) logical(lgl)
!               Specifies whether the upper or lower triangular part of the
!               symmetric matrix MAT is stored. If:
!
!               - UPPER = true :  Upper triangular is stored
!               - UPPER = false:  Lower triangular is stored.
!
!                The default is true.
!
!   FILL        (INPUT, OPTIONAL) logical(lgl)
!               On entry, when argument FILL is present, FILL is used as follows. If:
!
!               - FILL= true and UPPER= true,  the lower triangle of MAT is filled on exit.
!               - FILL= true and UPPER= false,  the upper triangle of MAT is filled on exit.
!               - FILL= false, the lower (UPPER= true) or upper (UPPER= false) triangle of MAT
!                 is not filled and not modified on exit.
!
!               The default is true.
!
!   TOL         (INPUT, OPTIONAL) real(stnd)
!               Tolerance used to test the matrix for positive-definiteness.
!               TOL is used as a multiplying factor for determining effective
!               zero for pivots. TOL must be greater or equal to zero, otherwise
!               the default value is used.
!
!               The default value is the machine precision multiplied by n.
!
!
! Further Details
! _______________
!
!   MAT is declared not positive definite if during the j-th stage of the
!   factorization of MAT, a pivot, PIV(j), is such that
!
!             PIV(j) <= MAT(j,j) * TOL
!
!   In this case, the leading minor of order j of MAT is declared not
!   positive definite, the Cholesky factorization is not completed and,
!   on exit of COMP_SYM_INV, FAILURE is set to true.
!
!   On exit, if FAILURE=true: 
!
!   - The upper or lower triangle of MAT is filled with nan() value if FILL=false.
!   - The matrix MAT is filled with nan() value if FILL=true.
!
!   A blocked algorithm is used to compute the Cholesky factorization. Furthermore,
!   the computations are parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error
    use Num_Constants,   only : nan
    use Utilities,       only : assert_eq, merror
#ifdef _BLAS
    use BLAS_interfaces, only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout)         :: mat(:,:)
    real(stnd),   intent(in), optional  :: tol
!
    logical(lgl), intent(in), optional :: upper, fill
    logical(lgl), intent(out)          :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(mat,1))  :: invdiag
    real(stnd)                          :: d1
!
    integer(i4b) :: n, i, j
!
    logical(lgl) :: mat_is_upper, fill_triang
!
#ifdef _OPENMP
    real(stnd), dimension(:,:), allocatable :: mat2
!
    integer :: iok
!
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_sym_inv'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq( int(size(mat,1),i4b) ,      &
                   int(size(mat,2),i4b) ,      &
                   name_proc )
!
    mat_is_upper = true
    if ( present(upper) ) mat_is_upper = upper
!
    fill_triang = true
    if ( present(fill) ) fill_triang = fill
!
!   QUICK RETURN IF POSSIBLE.
!
    failure = false
!
    if ( n<=0_i4b ) return
!    
!   CHOLESKY DECOMPOSITION OF mat .
!    
    call chol_cmp( mat, invdiag, d1, upper=upper, tol=tol )
!    
!   SET FAILURE INDICATOR.
!    
    failure = d1==zero
!
    if ( failure ) then
!    
!       mat IS NOT POSITIVE DEFINITE.
!
        d1 = nan()
!
        if ( fill_triang ) then
!
            mat(:n,:n) = d1
!
        else if ( mat_is_upper ) then
!
            do i = 1_i4b, n
                mat(1_i4b:i,i) = d1
            end do
!
        else
!
            do i = 1_i4b, n
                mat(i:n,i) = d1
            end do
!
        end if
!
    else
!    
#ifdef _OPENMP
        i           = omp_get_max_threads()
        ompparallel = omp_in_parallel()
        ompnested   = omp_get_nested() .and. ompparallel
        test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                      n*n>=omp_limit                      .and.      &
                      i>1_i4b
!
        if ( test_par ) then
!
!           ALLOCATE WORK SPACE FOR A FAST AND PARALLEL
!           MATRIX INVERSION IF POSSIBLE.
!
            allocate( mat2(n,n), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
        else
!
            iok = -1
!
        end if
!
        if ( iok==0 ) then
!    
!           ENOUGH WORKSPACE FOR A PARALLEL ALGORITHM, FORM mat-INVERSE.
!    
!$OMP PARALLEL PRIVATE(i,j),        &
!$OMP          SHARED(mat_is_upper)
!
            if ( mat_is_upper ) then
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do i = 1_i4b, n
!
                    mat2(i,i) = invdiag(i)
!
!                   SOLVE  U'*X = b  WHERE b IS ONE COLUMN OF THE IDENTITY MATRIX OF ORDER n .
!
                    do j = i+1_i4b, n
!
                        mat2(j,i) = -invdiag(j)*dot_product( mat(i:j-1_i4b,j), mat2(i:j-1_i4b,i) )
!
                    end do
!
!                   SOLVE  U*X = b  OVERWRITING b WITH X .
!
!
                    do j = n, i, -1_i4b
!
                        mat2(j,i) = mat2(j,i)*invdiag(j)
#ifdef _BLAS
                        call axpy( j-i, -mat2(j,i), mat(i:j-1_i4b,j), 1_i4b, mat2(i:j-1_i4b,i), 1_i4b )
#else
                        mat2(i:j-1_i4b,i) = mat2(i:j-1_i4b,i) - mat2(j,i)*mat(i:j-1_i4b,j)
#endif
!
                    end do
!
                end do
!
!$OMP END DO
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do i = 1_i4b, n
!
                    mat(i,i:n) = mat2(i:n,i)
!
                    if ( fill_triang ) then
                        mat(i+1_i4b:n,i) = mat2(i+1_i4b:n,i)
                    end if
!
                end do
!
!$OMP END DO
!
            else
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do i = 1_i4b, n
!
                    mat2(i,i)         = invdiag(i)
                    mat2(i+1_i4b:n,i) = -invdiag(i)*mat(i+1_i4b:n,i)
!
!                   SOLVE  L*X = b  WHERE b IS ONE COLUMN OF THE IDENTITY MATRIX OF ORDER n .
!
                    do j = i+1_i4b, n
!
                        mat2(j,i) = mat2(j,i)*invdiag(j)
#ifdef _BLAS
                        call axpy( n-j, -mat2(j,i), mat(j+1_i4b:n,j), 1_i4b, mat2(j+1_i4b:n,i), 1_i4b )
#else
                        mat2(j+1_i4b:n,i) = mat2(j+1_i4b:n,i) - mat2(j,i)*mat(j+1_i4b:n,j)
#endif
!
                    end do
!
!                   SOLVE  L'*X = b  OVERWRITING b WITH X .
!
                    do j = n, i, -1_i4b
!
                        mat2(j,i) = ( mat2(j,i) - dot_product( mat(j+1_i4b:n,j), mat2(j+1_i4b:n,i) ) )*invdiag(j)
!
                    end do
!
                end do
!
!$OMP END DO
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do i = 1_i4b, n
!
                    mat(i:n,i) = mat2(i:n,i)
!
                    if ( fill_triang ) then
                        mat(i,i+1_i4b:n) = mat2(i+1_i4b:n,i)
                    end if
!
                end do
!
!$OMP END DO
!
            end if
!
!$OMP END PARALLEL
!
!           DEALLOCATE WORK ARRAY.
!
            deallocate( mat2 )
!
        else
!
#endif
!    
!           FORM mat-INVERSE IN PLACE.
!    
            if ( mat_is_upper ) then
!
!               INVERT THE TRIANGULAR CHOLESKY FACTOR U IN PLACE.
!
                do j = 1_i4b, n
!
                    mat(j,j) = invdiag(j)
!
!                   COMPUTE ELEMENTS 1:j-1 OF COLUMN j.
!
                    do i = 1_i4b, j-1_i4b
!
                        if ( mat(i,j)==zero ) cycle
!
#ifdef _BLAS
                        call axpy( i-1_i4b, mat(i,j), mat(1_i4b:i-1_i4b,i), 1_i4b, mat(1_i4b:i-1_i4b,j), 1_i4b )
#else
                        mat(1_i4b:i-1_i4b,j) = mat(1_i4b:i-1_i4b,j) + mat(i,j)*mat(1_i4b:i-1_i4b,i)
#endif
                        mat(i,j)             = mat(i,j)*invdiag(i)
!
                    end do
!
                    mat(1_i4b:j-1_i4b,j) = -invdiag(j)*mat(1_i4b:j-1_i4b,j)
!
                end do
!
            else
!
!               INVERT THE TRIANGULAR CHOLESKY FACTOR L IN PLACE.
!
                do j = n, 1_i4b, -1_i4b
!
                    mat(j,j) = invdiag(j)
!
!                   COMPUTE ELEMENTS j+1:n OF COLUMN j.
!
                    do i = n, j+1_i4b, -1_i4b
!
                        if ( mat(i,j)==zero ) cycle
!
#ifdef _BLAS
                        call axpy( n-i, mat(i,j), mat(i+1_i4b:n,i), 1_i4b, mat(i+1_i4b:n,j), 1_i4b )
#else
                        mat(i+1_i4b:n,j) = mat(i+1_i4b:n,j) + mat(i,j)*mat(i+1_i4b:n,i)
#endif
                        mat(i,j)         = mat(i,j)*invdiag(i)
!
                    end do
!
                    mat(j+1_i4b:n,j) = -invdiag(j)*mat(j+1_i4b:n,j)
!
                end do
!
            end if
!    
!           FORM inv(U)*inv(U)' OR inv(L)'*inv(L).
!
            call comp_uut_ltl_a( mat, upper=upper, fill=fill )
!
#ifdef _OPENMP
        end if
#endif
!
    end if
!
!
! END OF SUBROUTINE comp_sym_inv_a
! ________________________________
!
    end subroutine comp_sym_inv_a
!
! =========================================================================================
!
    subroutine comp_sym_inv_b( mat, failure, matinv, upper, fill, tol )
!
! Purpose
! _______
!
!   COMP_SYM_INV computes the inverse of a real symmetric positive definite
!   matrix MAT using the Cholesky factorization MAT = U' * U or MAT = L * L'.
!
!
! Arguments
! _________
!
!   MAT         (INPUT) real(stnd), dimension(:,:)
!               On entry, the symmetric matrix to be inverted.
!
!               - If UPPER = true or is absent:
!                 The leading n-by-n upper triangular part of MAT contains the
!                 upper triangular part of the matrix and the strictly
!                 lower triangular part of MAT is not referenced.
!               - If UPPER = false:
!                 The leading n-by-n lower triangular part of MAT contains the
!                 lower triangular part of the matrix and the strictly
!                 upper triangular part of MAT is not referenced.
!
!               MAT is not modified by the subroutine.
!
!               The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   FAILURE     (OUTPUT) logical(lgl)
!               On exit, if:
!
!               - FAILURE = true :  MAT is algorithmically not positive definite.
!               - FAILURE = false:  MAT has been inverted.
!
!   MATINV      (OUTPUT) real(stnd), dimension(:,:)
!               On exit:
!
!               - If FILL = true or is absent:
!                 The (symmetric) inverse of MAT.
!               - If FILL = false:
!                 The upper (if UPPER=true) or lower (if UPPER=false) triangle
!                 of the (symmetric) inverse of MAT, is stored in the upper or lower 
!                 triangular part of the matrix MATINV and the other part of MATINV
!                 is not modified.
!
!               The shape of MATINV must verify:  size( MATINV, 1 ) = size( MATINV, 2 ) = n .
!
!   UPPER       (INPUT, OPTIONAL) logical(lgl)
!               Specifies whether the upper or lower triangular part of the
!               symmetric matrix MAT is stored. If:
!
!               - UPPER = true :  Upper triangular is stored
!               - UPPER = false:  Lower triangular is stored.
!
!               The default is true.
!
!   FILL        (INPUT, OPTIONAL) logical(lgl)
!               On entry, when argument FILL is present, FILL is used as follows. If:
!
!               - FILL= true and UPPER= true, the lower triangle of MATINV is filled on exit.
!               - FILL= true and UPPER= false, the upper triangle of MATINV is filled on exit.
!               - FILL= false, the lower (UPPER= true) or upper (UPPER= false) triangle of
!                 MATINV is not filled and not modified on exit.
!
!               The default is true.
!
!   TOL         (INPUT, OPTIONAL) real(stnd)
!               Tolerance used to test the matrix for positive-definiteness.
!               TOL is used as a multiplying factor for determining effective
!               zero for pivots. TOL must be greater or equal to zero, otherwise
!               the default value is used.
!
!               The default value is the machine precision multiplied by n.
!
!
! Further Details
! _______________
!
!   MAT is not modified by COMP_SYM_INV.
!
!   MAT is declared not positive definite if during the j-th stage of the
!   factorization of MAT, a pivot, PIV(j), is such that
!
!             PIV(j) <= MAT(j,j) * TOL
!
!   In this case, the leading minor of order j of MAT is declared not
!   positive definite, the Cholesky factorization is not completed and,
!   on exit of COMP_SYM_INV, FAILURE is set to true.
!
!   On exit, if FAILURE=true: 
!
!   - The upper or lower triangle of MATINV is filled with nan() value if FILL=false.
!   - The matrix MATINV is filled with nan() value if FILL=true.
!
!   A blocked algorithm is used to compute the factorization. Furthermore,
!   the computations are parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,     only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in)           :: mat(:,:)
    real(stnd),   intent(in), optional :: tol
    real(stnd),   intent(out)          :: matinv(:,:)
!
    logical(lgl), intent(in), optional :: upper, fill
    logical(lgl), intent(out)          :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: n, i
!
    logical(lgl) :: mat_is_upper, fill_triang
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_sym_inv'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq( int(size(mat,1),i4b) ,      &
                   int(size(mat,2),i4b) ,      &
                   int(size(matinv,1),i4b) ,   &
                   int(size(matinv,2),i4b) ,   &
                   name_proc )
!
    mat_is_upper = true
    if ( present(upper) ) mat_is_upper = upper
!
    fill_triang = true
    if ( present(fill) ) fill_triang = fill
!
!   QUICK RETURN IF POSSIBLE.
!
    failure = false
!
    if ( n<=0_i4b ) return
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.     &
                  (n*(n+1_i4b))/2_i4b>=omp_limit      .and.     &
                  i>1_i4b
#endif
!    
    if ( mat_is_upper ) then
!
!$OMP PARALLEL DO IF(test_par)                       &
!$OMP            ,SCHEDULE(STATIC,1)                 &
!$OMP            ,PRIVATE(i)
!
        do i = 1_i4b, n
            matinv(1_i4b:i,i) = mat(1_i4b:i,i)
        end do
!
!$OMP END PARALLEL DO
!
    else if ( fill_triang ) then
!
!$OMP PARALLEL DO IF(test_par)                           &
!$OMP            ,SCHEDULE(STATIC,1)                     &
!$OMP            ,PRIVATE(i)
!
        do i = 1_i4b, n
            matinv(i,i:n) = mat(i:n,i)
        end do
!
!$OMP END PARALLEL DO
!
        mat_is_upper = true
!
    else
!
!$OMP PARALLEL DO IF(test_par)                           &
!$OMP            ,SCHEDULE(STATIC,1)                     &
!$OMP            ,PRIVATE(i)
!
        do i = 1_i4b, n
            matinv(i:n,i) = mat(i:n,i)
        end do
!
!$OMP END PARALLEL DO
!
    end if
!    
!   FORM mat-INVERSE IN matinv.
!    
    call comp_sym_inv_a( matinv, failure, upper=mat_is_upper, fill=fill, tol=tol )
!
!
! END OF SUBROUTINE comp_sym_inv_b
! ________________________________
!
    end subroutine comp_sym_inv_b
!
! =========================================================================================
!
    function sym_inv( mat, upper, tol ) result( matinv )
!
! Purpose
! _______
!
!   SYM_INV computes the inverse of a real symmetric positive definite
!   matrix MAT using the Cholesky factorization MAT = U' * U or MAT = L * L'.
!
!
! Arguments
! _________
!
!   MAT         (INPUT) real(stnd), dimension(:,:)
!               On entry, the symmetric matrix to be inverted. If:
!
!               - If UPPER = true or is absent:
!                 The leading n-by-n upper triangular part of MAT contains the
!                 upper triangular part of the matrix and the strictly
!                 lower triangular part of MAT is not referenced.
!               - If UPPER = false:
!                 The leading n-by-n lower triangular part of MAT contains the
!                 lower triangular part of the matrix and the strictly
!                 upper triangular part of MAT is not referenced.
!
!               MAT is not modified by the function.
!
!               The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   UPPER       (INPUT, OPTIONAL) logical(lgl)
!               Specifies whether the upper or lower triangular part of the
!               symmetric matrix MAT is stored. If:
!
!               - UPPER = true :  Upper triangular
!               - UPPER = false:  Lower triangular.
!
!               The default is true.
!
!   TOL         (INPUT, OPTIONAL) real(stnd)
!               Tolerance used to test the matrix for positive-definiteness.
!               TOL is used as a multiplying factor for determining effective
!               zero for pivots. TOL must be greater or equal to zero, otherwise
!               the default value is used.
!
!               The default value is the machine precision multiplied by n.
!
!
! Further Details
! _______________
!
!   MAT is not modified by function SYM_INV.
!
!   The (symmetric) inverse of MAT is returned if MAT is positive definite. 
!
!   MAT is declared not positive definite if during the j-th stage of the
!   Cholesky factorization of MAT, a pivot, PIV(j), is such that
!
!             PIV(j) <= MAT(j,j) * TOL
!
!   In this case, the leading minor of order j of MAT is declared not
!   positive definite.
!
!   On exit, if MAT is algorithmically not positive definite, SYM_INV returns a matrix
!   filled with nan() value.
!
!   A blocked algorithm is used to compute the factorization. Furthermore,
!   the computations are parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : nan
    use Utilities,       only : assert_eq
#ifdef _BLAS
    use BLAS_interfaces, only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in)           :: mat(:,:)
    real(stnd),   intent(in), optional :: tol
!
    logical(lgl), intent(in), optional :: upper
!
    real(stnd), dimension(size(mat,1),size(mat,1)) :: matinv
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(mat,1),size(mat,1)) :: matinv2
    real(stnd), dimension(size(mat,1))             :: invdiag
    real(stnd)                                     :: d1
!
    integer(i4b) :: n, i, j
!
    logical(lgl) :: mat_is_upper
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='sym_inv'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq( int(size(mat,1),i4b) ,      &
                   int(size(mat,2),i4b) ,      &
                   name_proc )
!
    mat_is_upper = true
    if ( present(upper) ) mat_is_upper = upper
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!
#ifdef _OPENMP
    j           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.     &
                  n*n>=omp_limit                      .and.     &
                  j>1_i4b
#endif
!
!   FIRST COPY UPPER OR LOWER TRIANGLE OF mat IN UPPER TRIANGLE OF matinv2 .
!
    if ( mat_is_upper ) then
!
!$OMP PARALLEL DO IF(test_par)                     &
!$OMP            ,SCHEDULE(STATIC,1)               &
!$OMP            ,PRIVATE(i)
!
        do i = 1_i4b, n
            matinv2(1_i4b:i,i) = mat(1_i4b:i,i)
        end do
!
!$OMP END PARALLEL DO
!
    else
!
!$OMP PARALLEL DO IF(test_par)                     &
!$OMP            ,SCHEDULE(STATIC,1)               &
!$OMP            ,PRIVATE(i)
!
        do i = 1_i4b, n
            matinv2(i,i:n) = mat(i:n,i)
        end do
!
!$OMP END PARALLEL DO
!
    end if
!    
!   CHOLESKY DECOMPOSITION OF mat .
!    
    call chol_cmp( matinv2, invdiag, d1, tol=tol )
!    
    if ( d1==zero ) then
!    
!       mat IS NOT POSITIVE DEFINITE.
!
        d1 = nan()
        matinv(:n,:n) = d1
!
    else 
!    
!       FORM mat-INVERSE IN PLACE IN matinv2.
!
!       INVERT THE TRIANGULAR CHOLESKY FACTOR U.
!
        do j = 1_i4b, n
!
            matinv2(j,j) = invdiag(j)
!
!           COMPUTE ELEMENTS 1:j-1 OF COLUMN j.
!
            do i = 1_i4b, j-1_i4b
!
                if ( matinv2(i,j)==zero ) cycle
!
#ifdef _BLAS
                call axpy( i-1_i4b, matinv2(i,j), matinv2(1_i4b:i-1_i4b,i), 1_i4b, matinv2(1_i4b:i-1_i4b,j), 1_i4b )
#else
                matinv2(1_i4b:i-1_i4b,j) = matinv2(1_i4b:i-1_i4b,j) + matinv2(i,j)*matinv2(1_i4b:i-1_i4b,i)
#endif
                matinv2(i,j)             = matinv2(i,j)*invdiag(i)
!
            end do
!
            matinv2(1_i4b:j-1_i4b,j) = -invdiag(j)*matinv2(1_i4b:j-1_i4b,j)
!
        end do
!    
!       FORM inv(U)*inv(U)' OR inv(L)'*inv(L).
!
#ifdef _USE_PGI
        call comp_uut_ltl_a( matinv2 )
!
!       COPY THE RESULT.
!
        matinv(:n,:n) = matinv2(:n,:n)
!
#else
        call comp_uut_ltl_b( matinv2, matinv )
#endif
!
    end if
!
!
! END OF FUNCTION sym_inv
! _______________________
!
    end function sym_inv
!
! =========================================================================================
!
    subroutine comp_sym_ginv_a( mat, failure, krank, upper, fill, tol )
!
! Purpose
! _______
!
!   COMP_SYM_GINV computes, in place, the (generalized) inverse of a real symmetric positive
!   semidefinite matrix MAT using the Cholesky factorization MAT = U' * U or MAT = L * L'.
!
!
! Arguments
! _________
!
!   MAT         (INPUT/OUTPUT) real(stnd), dimension(:,:)
!               On entry, the symmetric matrix to be inverted. If:
!
!               - UPPER = true or is absent:
!                 The leading n-by-n upper triangular part of MAT contains the
!                 upper triangular part of the matrix and the strictly
!                 lower triangular part of MAT is not referenced.
!               - UPPER = false:
!                 The leading n-by-n lower triangular part of MAT contains the
!                 lower triangular part of the matrix and the strictly
!                 upper triangular part of MAT is not referenced.
!
!               On exit, if:
!
!               - FILL = true or is absent:
!                 The (symmetric) inverse of MAT overwrites MAT.
!               - FILL = false:
!                 The upper or lower triangle of the (symmetric)
!                 inverse of MAT, overwrites the input upper or lower 
!                 triangular part of the matrix MAT and the other part
!                 of MAT is not modified.
!
!               The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   FAILURE     (OUTPUT) logical(lgl)
!                On exit, if:
!
!                - FAILURE = true :  MAT is algorithmically not positive semidefinite.
!                - FAILURE = false:  MAT has been inverted.
!
!   KRANK        (OUTPUT) integer(i4b)
!                On exit, KRANK contains the effective rank of MAT, which is defined
!                as the number of nonzero elements of the diagonal of L or U. Note
!                that KRANK may be different from the true rank of MAT. See the
!                reference (2) for details.
!
!   UPPER        (INPUT, OPTIONAL) logical(lgl)
!                Specifies whether the upper or lower triangular part of the
!                symmetric matrix MAT is stored. If:
!
!                - UPPER  = true :  Upper triangular is stored
!                - UPPER  = false:  Lower triangular is stored.
!
!                The default is true.
!
!   FILL         (INPUT, OPTIONAL) logical(lgl)
!                On entry, when argument FILL is present, FILL is used as follows. If:
!
!                - FILL= true and UPPER= true,  the lower triangle of MAT is filled on exit.
!                - FILL= true and UPPER= false,  the upper triangle of MAT is filled on exit.
!                - FILL= false, the lower (UPPER= true) or upper (UPPER= false) triangle of MAT
!                  is not filled on exit.
!
!                The default is true.
!
!   TOL          (INPUT, OPTIONAL) real(stnd)
!                Tolerance used to test MAT for positive-semidefiniteness.
!                TOL is used as a multiplying factor for determining effective
!                zero for pivots. TOL must be greater or equal to zero, otherwise
!                the default value is used.
!
!                The default value is the machine precision multiplied by n.
!
!
! Further Details
! _______________
!
!   If MAT is positive semidefinite, the subroutine computes a generalized inverse of MAT.
!   GMAT is a generalized inverse of MAT if
!
!                    MAT * GMAT * MAT = MAT  and  GMAT * MAT * GMAT = GMAT
!
!   See description of subroutine GCHOL_CMP2 for more details.
!   The subroutine also computes and returns an estimate of the effective rank of MAT
!   in the argument RANK. Note that KRANK may be different
!   from the true rank of MAT. See the reference (2) for details.
!
!   MAT is declared not positive semidefinite if during the j-th stage of the
!   Cholesky factorization of MAT, a pivot, PIV(j), is such that
!
!             PIV(j) <= - abs( MAT(j,j) * TOL )
!
!   In this case, the leading minor of order j of MAT is declared not
!   positive semidefinite, the Cholesky factorization is not completed
!   and on exit of COMP_SYM_GINV, FAILURE is set to true.
!
!   On exit, if FAILURE=true: 
!
!   - KRANK is set to -1,  
!   - The upper or lower triangle of MAT is filled with nan() value if FILL=false.
!   - The matrix MAT is filled with nan() value if FILL=true.
!
!   A blocked algorithm is used to compute the Cholesky factorization. Furthermore,
!   the computations are parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Higham, N.J., 2009:
!           Cholesky factorization.
!           Wiley Interdisciplinary Reviews: Computational Statistics,
!           Vol. 1, pp 251-254.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,  only : allocate_error
    use Num_Constants,   only : nan
    use Utilities,       only : assert_eq, merror
#ifdef _BLAS
    use BLAS_interfaces, only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout)         :: mat(:,:)
    real(stnd),   intent(in), optional  :: tol
!
    integer(i4b),  intent(out)          :: krank
!
    logical(lgl), intent(in), optional :: upper, fill
    logical(lgl), intent(out)          :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(mat,1))  :: invdiag
    real(stnd)                          :: d1
!
    integer(i4b) :: n, i, j
!
    logical(lgl) :: mat_is_upper, fill_triang
!
#ifdef _OPENMP
    real(stnd)                              :: piv
    real(stnd), dimension(:,:), allocatable :: mat2
!
    integer :: iok
!
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_sym_ginv'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq( int(size(mat,1),i4b) ,      &
                   int(size(mat,2),i4b) ,      &
                   name_proc )
!
    mat_is_upper = true
    if ( present(upper) ) mat_is_upper = upper
!
    fill_triang = true
    if ( present(fill) ) fill_triang = fill
!
    krank = n
!
    failure = false
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!    
!   CHOLESKY DECOMPOSITION OF mat .
!    
    call gchol_cmp( mat, invdiag, krank, d1, upper=upper, tol=tol )
!    
!   SET FAILURE INDICATOR.
!    
    failure = d1<zero
!
    if ( failure ) then
!    
!       mat IS NOT POSITIVE SEMIDEFINITE.
!
        d1 = nan()
!
        if ( fill_triang ) then
!
            mat(:n,:n) = d1
!
        else if ( mat_is_upper ) then
!
            do i = 1_i4b, n
                mat(1_i4b:i,i) = d1
            end do
!
        else
!
            do i = 1_i4b, n
                mat(i:n,i) = d1
            end do
!
        end if
!
    else
!    
#ifdef _OPENMP
        i           = omp_get_max_threads()
        ompparallel = omp_in_parallel()
        ompnested   = omp_get_nested() .and. ompparallel
        test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                      n*n>=omp_limit                      .and.      &
                      i>1_i4b
!
        if ( test_par ) then
!
!           ALLOCATE WORK SPACE FOR A FAST AND PARALLEL
!           MATRIX INVERSION IF POSSIBLE.
!
            allocate( mat2(n,n), stat = iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
        else
!
            iok = -1
!
        end if
!
        if ( iok==0 ) then
!    
!           ENOUGH WORKSPACE FOR A PARALLEL ALGORITHM, FORM GENERALIZED-INVERSE OF mat.
!    
!$OMP PARALLEL PRIVATE(i,j,piv),        &
!$OMP          SHARED(mat_is_upper)
!
            if ( mat_is_upper ) then
!
!               COMPUTE GENERALIZED INVERSE OF THE TRIANGULAR CHOLESKY FACTOR U : ginv(U) .
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do i = 1_i4b, n
!
                    if ( invdiag(i)==zero ) then
!
                        mat2(1_i4b:i,i) = zero
!
                    else
!
                        mat2(1_i4b:i-1_i4b,i) = -invdiag(i)*mat(1_i4b:i-1_i4b,i)
                        mat2(i,i)             = invdiag(i)
!
                        do j = i-1_i4b, 1_i4b, -1_i4b
!
                            if ( invdiag(j)==zero ) cycle
!
                            piv = mat2(j,i)*invdiag(j)
#ifdef _BLAS
                            call axpy( j-1_i4b, -piv, mat(1_i4b:j-1_i4b,j), 1_i4b, mat2(1_i4b:j-1_i4b,i), 1_i4b )
#else
                            mat2(1_i4b:j-1_i4b,i) = mat2(1_i4b:j-1_i4b,i) - piv*mat(1_i4b:j-1_i4b,j)
#endif
                            mat2(j,i) = piv
!
                        end do
!
                    end if
!
                end do
!
!$OMP END DO
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do i = 1_i4b, n
!
                    mat(1_i4b:i,i) = mat2(1_i4b:i,i)
!
                end do
!
!$OMP END DO
!
            else
!
!               COMPUTE GENERALIZED INVERSE OF THE TRIANGULAR CHOLESKY FACTOR L : ginv(L) .
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do i = 1_i4b, n
!
                    if ( invdiag(i)==zero ) then
!
                        mat2(i:n,i) = zero
!
                    else
!
                        mat2(i,i)         = invdiag(i)
                        mat2(i+1_i4b:n,i) = -invdiag(i)*mat(i+1_i4b:n,i)
!
                        do j = i+1_i4b, n
!
                            if ( invdiag(j)==zero ) then
!
                                mat2(j,i) = zero
!
                            else
!
                                piv       = mat2(j,i)*invdiag(j)
                                mat2(j,i) = piv
#ifdef _BLAS
                                call axpy( n-j, -piv, mat(j+1_i4b:n,j), 1_i4b, mat2(j+1_i4b:n,i), 1_i4b )
#else
                                mat2(j+1_i4b:n,i) = mat2(j+1_i4b:n,i) - piv*mat(j+1_i4b:n,j)
#endif
!
                            end if
!
                        end do
!
                    end if
!
                end do
!
!$OMP END DO
!
!$OMP DO SCHEDULE(STATIC,1)
!
                do i = 1_i4b, n
!
                    mat(i:n,i) = mat2(i:n,i)
!
                end do
!
!$OMP END DO
!
            end if
!
!$OMP END PARALLEL
!
!           DEALLOCATE WORK ARRAY.
!
            deallocate( mat2 )
!
        else
!
#endif
!    
!           FORM GENERALIZED-INVERSE OF mat IN PLACE.
!
            if ( mat_is_upper ) then
!
!               COMPUTE GENERALIZED INVERSE OF THE TRIANGULAR CHOLESKY FACTOR U : ginv(U) .
!
                do j = 1_i4b, n
!
                    if ( invdiag(j)==zero ) then
!
                        mat(1_i4b:j,j) = zero
!
                    else
!
                        mat(j,j) = invdiag(j)
!
!                       COMPUTE ELEMENTS 1:j-1 OF COLUMN j.
!
                        do i = 1_i4b, j-1_i4b
!
                            if ( invdiag(i)==zero ) cycle
!
#ifdef _BLAS
                            call axpy( i-1_i4b, mat(i,j), mat(1_i4b:i-1_i4b,i), 1_i4b, mat(1_i4b:i-1_i4b,j), 1_i4b )
#else
                            mat(1_i4b:i-1_i4b,j) = mat(1_i4b:i-1_i4b,j) + mat(i,j)*mat(1_i4b:i-1_i4b,i)
#endif
                            mat(i,j)             = mat(i,j)*invdiag(i)
!
                        end do
!
                        mat(1_i4b:j-1_i4b,j) = -invdiag(j)*mat(1_i4b:j-1_i4b,j)
!
                    end if
!
                end do
!
            else
!
!               COMPUTE GENERALIZED INVERSE OF THE TRIANGULAR CHOLESKY FACTOR L : ginv(L) .
!
                do j = n, 1_i4b, -1_i4b
!
                    if ( invdiag(j)==zero ) then
!
                        mat(j,1:j) = zero
!
                    else
!
                        mat(j,j) = invdiag(j)
!
!                       COMPUTE ELEMENTS j+1:n OF COLUMN j.
!
                        do i = n, j+1_i4b, -1_i4b
!
                            if ( invdiag(i)==zero ) cycle
!
#ifdef _BLAS
                            call axpy( n-i, mat(i,j), mat(i+1_i4b:n,i), 1_i4b, mat(i+1_i4b:n,j), 1_i4b )
#else
                            mat(i+1_i4b:n,j) = mat(i+1_i4b:n,j) + mat(i,j)*mat(i+1_i4b:n,i)
#endif
                            mat(i,j)         = mat(i,j)*invdiag(i)
!
                        end do
!
                        mat(j+1_i4b:n,j) = -invdiag(j)*mat(j+1_i4b:n,j)
!
                    end if
!
                end do
!
            end if
!
#ifdef _OPENMP
        end if
#endif
!    
!       FORM ginv(U)*ginv(U)' OR ginv(L)'*ginv(L).
!
        call comp_uut_ltl_a( mat, upper=upper, fill=fill )
!
    end if
!
!
! END OF SUBROUTINE comp_sym_ginv_a
! _________________________________
!
    end subroutine comp_sym_ginv_a
!
! =========================================================================================
!
    subroutine comp_sym_ginv_b( mat, failure, krank, matinv, upper, fill, tol )
!
! Purpose
! _______
!
!   COMP_SYM_GINV computes the (generalized) inverse of a real symmetric positive
!   semidefinite matrix MAT using the Cholesky factorization MAT = U' * U or MAT = L * L'.
!
!
! Arguments
! _________
!
!   MAT         (INPUT) real(stnd), dimension(:,:)
!               On entry, the symmetric matrix to be inverted. If:
!
!               - UPPER = true or is absent:
!                 The leading n-by-n upper triangular part of MAT contains the
!                 upper triangular part of the matrix and the strictly
!                 lower triangular part of MAT is not referenced.
!               - If UPPER = false:
!                 The leading n-by-n lower triangular part of MAT contains the
!                 lower triangular part of the matrix and the strictly
!                 upper triangular part of MAT is not referenced.
!
!               MAT is not modified by the subroutine.
!
!               The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   FAILURE     (OUTPUT) logical(lgl)
!               On exit, if:
!
!               -FAILURE = true :  MAT is algorithmically not positive semidefinite.
!               -FAILURE = false:  MAT has been inverted.
!
!   KRANK       (OUTPUT) integer(i4b)
!               On exit, KRANK contains the effective rank of MAT, which is defined
!               as the number of nonzero elements of the diagonal of L or U. Note
!               that KRANK may be different from the true rank of MAT. See the
!               reference (2) for details.
!
!   MATINV      (OUTPUT) real(stnd), dimension(:,:)
!               On exit, if:
!
!               - FILL = true or is absent:
!                 The (symmetric) (generalized) inverse of MAT.
!               - If FILL = false:
!                 The upper (if UPPER=true) or lower (if UPPER=false) triangle
!                 of the (symmetric) (generalized) inverse of MAT, is stored in 
!                 the upper or lower triangular part of the matrix MATINV and the
!                 other part of MATINV is not modified.
!
!               The shape of MATINV must verify:  size( MATINV, 1 ) = size( MATINV, 2 ) = n .
!
!   UPPER       (INPUT, OPTIONAL) logical(lgl)
!               Specifies whether the upper or lower triangular part of the
!               symmetric matrix MAT is stored. If:
!
!               - UPPER = true :  Upper triangular is stored
!               - UPPER = false:  Lower triangular is stored.
!
!               The default is true.
!
!   FILL        (INPUT, OPTIONAL) logical(lgl)
!               On entry, when argument FILL is present, FILL is used as follows. If:
!
!               - FILL= true and UPPER= true,  the lower triangle of MATINV is filled on exit.
!               - FILL= true and UPPER= false,  the upper triangle of MATINV is filled on exit.
!               - FILL= false, the lower (UPPER= true) or upper (UPPER= false) triangle of
!                 MATINV is not filled on exit.
!
!               The default is true.
!
!   TOL         (INPUT, OPTIONAL) real(stnd)
!               Tolerance used to test MAT for positive-semidefiniteness.
!               TOL is used as a multiplying factor for determining effective
!               zero for pivots. TOL must be greater or equal to zero, otherwise
!               the default value is used.
!
!               The default value is the machine precision multiplied by n.
!
!
! Further Details
! _______________
!
!   MAT is not modified by COMP_SYM_GINV.
!
!   If MAT is positive semidefinite, the subroutine computes a generalized inverse of MAT.
!   MATINV is a generalized inverse of MAT if
!
!                    MAT * MATINV * MAT = MAT  and  MATINV * MAT * MATINV = MATINV
!
!   See description of subroutine GCHOL_CMP2 for more details.
!   The subroutine also computes and returns an estimate of the rank of MAT
!   in the argument RANK. Note that KRANK may be different
!   from the true rank of MAT. See the reference (2) for details.
!
!   MAT is declared not positive semidefinite if during the j-th stage of the
!   Cholesky factorization of MAT, a pivot, PIV(j), is such that
!
!             PIV(j) <= - abs( MAT(j,j) * TOL )
!
!   In this case, the leading minor of order j of MAT is declared not
!   positive semidefinite and on exit of COMP_SYM_GINV, FAILURE is set to true.
!
!   On exit, if FAILURE=true: 
!
!   - KRANK is set to -1,  
!   - The upper or lower triangle of MATINV is filled with nan() value if FILL=false,
!   - The matrix MATINV is filled with nan() value if FILL=true.
!
!   A blocked algorithm is used to compute the Cholesky factorization. Furthermore,
!   the computations are parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Higham, N.J., 2009:
!           Cholesky factorization.
!           Wiley Interdisciplinary Reviews: Computational Statistics,
!           Vol. 1, pp 251-254.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in)           :: mat(:,:)
    real(stnd),   intent(in), optional :: tol
    real(stnd),   intent(out)          :: matinv(:,:)
!
    integer(i4b),  intent(out)         :: krank
!
    logical(lgl), intent(in), optional :: upper, fill
    logical(lgl), intent(out)          :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    integer(i4b) :: n, i
!
    logical(lgl) :: mat_is_upper, fill_triang
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_sym_ginv'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq( int(size(mat,1),i4b) ,      &
                   int(size(mat,2),i4b) ,      &
                   int(size(matinv,1),i4b) ,   &
                   int(size(matinv,2),i4b) ,   &
                   name_proc )
!
    mat_is_upper = true
    if ( present(upper) ) mat_is_upper = upper
!
    fill_triang = true
    if ( present(fill) ) fill_triang = fill
!
    failure = false
!    
    krank = n
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.     &
                  (n*(n+1_i4b))/2_i4b>=omp_limit      .and.     &
                  i>1_i4b
#endif
!
    if ( mat_is_upper ) then
!
!$OMP PARALLEL DO IF(test_par)                       &
!$OMP            ,SCHEDULE(STATIC,1)                 &
!$OMP            ,PRIVATE(i)
!
        do i = 1_i4b, n
            matinv(1_i4b:i,i) = mat(1_i4b:i,i)
        end do
!
!$OMP END PARALLEL DO
!
    else if ( fill_triang ) then
!
!$OMP PARALLEL DO IF(test_par)                           &
!$OMP            ,SCHEDULE(STATIC,1)                     &
!$OMP            ,PRIVATE(i)
!
        do i = 1_i4b, n
            matinv(i,i:n) = mat(i:n,i)
        end do
!
!$OMP END PARALLEL DO
!
        mat_is_upper = true
!
    else
!
!$OMP PARALLEL DO IF(test_par)                           &
!$OMP            ,SCHEDULE(STATIC,1)                     &
!$OMP            ,PRIVATE(i)
!
        do i = 1_i4b, n
            matinv(i:n,i) = mat(i:n,i)
        end do
!
!$OMP END PARALLEL DO
!
    end if
!    
!   FORM mat-INVERSE IN matinv.
!    
    call comp_sym_ginv_a( matinv, failure, krank, upper=mat_is_upper, fill=fill, tol=tol )
!
!
! END OF SUBROUTINE comp_sym_ginv_b
! _________________________________
!
    end subroutine comp_sym_ginv_b
!
! =========================================================================================
!
    subroutine comp_triang_inv_a( mat, upper )
!
! Purpose
! _______
!
!   COMP_TRIANG_INV computes, in place, the inverse of a real upper or lower triangular
!   matrix MAT.
!
!   On entry, if MAT is algorithmically singular, diagonal terms  smaller in magnitude 
!   than the value SAFMIN (the smallest number that can be reciprocated safely) are
!   replaced by SAFMIN using the same sign as the diagonal term. An approximate solution 
!   based on this replacement is then obtained.
!
!
! Arguments
! _________
!
!   MAT         (INPUT/OUTPUT) real(stnd), dimension(:,:)
!               On entry, the triangular matrix to be inverted.
!
!               - If UPPER = true or is absent:
!                 The leading n-by-n upper triangular part of MAT contains 
!                 the upper triangular matrix and the strictly
!                 lower triangular part of MAT is not referenced.
!               - If UPPER = false:
!                 The leading n-by-n lower triangular part of MAT contains
!                 the lower triangular matrix and the strictly
!                 upper triangular part of MAT is not referenced.
!
!               On exit, the (triangular) inverse of the original matrix
!               in the same storage format. 
!
!               The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   UPPER       (INPUT, OPTIONAL) logical(lgl)
!               Specifies whether the matrix MAT is upper or lower triangular. If:
!
!               - UPPER = true :  MAT is  Upper triangular,
!               - UPPER = false:  MAT is  Lower triangular.
!
!               The default is true.
!
!
! Further Details
! _______________
!
!   The computations are not parallelized even if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch
    use Utilities,       only : assert_eq
#ifdef _BLAS
    use BLAS_interfaces, only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout) :: mat(:,:)
!
    logical(lgl), intent(in), optional :: upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: matjj, safmin
!
    integer(i4b) :: n, i, j
!
    logical(lgl) :: mat_is_upper
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_triang_inv'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq( int(size(mat,1),i4b) ,      &
                   int(size(mat,2),i4b) ,      &
                   name_proc )
!
    mat_is_upper = true
    if ( present(upper) ) mat_is_upper = upper
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!    
!   COMPUTE SAFE MINIMUM.
!    
    safmin = lamch( 's' )
!
    if ( mat_is_upper ) then
!
!       COMPUTE INVERSE OF UPPER TRIANGULAR MATRIX.
!
        do j = 1_i4b, n
!
            matjj    = one/sign( max(abs(mat(j,j)),safmin), mat(j,j) )
            mat(j,j) = matjj
!
!           COMPUTE ELEMENTS 1:j-1 OF COLUMN j.
!
            do i = 1_i4b, j-1_i4b
!
                if ( mat(i,j)==zero ) cycle
!
#ifdef _BLAS
                call axpy( i-1_i4b, mat(i,j), mat(1_i4b:i-1_i4b,i), 1_i4b, mat(1_i4b:i-1_i4b,j), 1_i4b )
#else
                mat(1_i4b:i-1_i4b,j) = mat(1_i4b:i-1_i4b,j) + mat(i,j)*mat(1_i4b:i-1_i4b,i)
#endif
                mat(i,j)             = mat(i,j)*mat(i,i)
!
            end do
!
            mat(1_i4b:j-1_i4b,j) = -matjj*mat(1_i4b:j-1_i4b,j)
!
        end do
!
    else
!
!       COMPUTE INVERSE OF LOWER TRIANGULAR MATRIX.
!
        do j = n, 1_i4b, -1_i4b
!
            matjj    = one/sign( max(abs(mat(j,j)),safmin), mat(j,j) )
            mat(j,j) = matjj
!
!           COMPUTE ELEMENTS j+1:n OF COLUMN j.
!
            do i = n, j+1_i4b, -1_i4b
!
                if ( mat(i,j)==zero ) cycle
!
#ifdef _BLAS
                call axpy( n-i, mat(i,j), mat(i+1_i4b:n,i), 1_i4b, mat(i+1_i4b:n,j), 1_i4b )
#else
                mat(i+1_i4b:n,j) = mat(i+1_i4b:n,j) + mat(i,j)*mat(i+1_i4b:n,i)
#endif
                mat(i,j)         = mat(i,j)*mat(i,i)
!
            end do
!
            mat(j+1_i4b:n,j) = -matjj*mat(j+1_i4b:n,j)
!
        end do
!
    end if
!
!
! END OF SUBROUTINE comp_triang_inv_a
! ___________________________________
!
    end subroutine comp_triang_inv_a
!
! =========================================================================================
!
    subroutine comp_triang_inv_b( mat, matinv, upper )
!
! Purpose
! _______
!
!   COMP_TRIANG_INV computes the inverse of a real upper or lower triangular
!   matrix MAT.
!
!   On entry, if MAT is algorithmically singular, diagonal terms  smaller in magnitude 
!   than the value SAFMIN (the smallest number that can be reciprocated safely) are
!   replaced by SAFMIN using the same sign as the diagonal term. An approximate solution 
!   based on this replacement is then obtained.
!
!
! Arguments
! _________
!
!   MAT         (INPUT) real(stnd), dimension(:,:)
!               On entry, the triangular matrix to be inverted. If:
!
!               - UPPER = true or is absent:
!                 The leading n-by-n upper triangular part of MAT contains 
!                 the upper triangular matrix and the strictly
!                 lower triangular part of MAT is not referenced.
!               - UPPER = false:
!                 The leading n-by-n lower triangular part of MAT contains
!                 the lower triangular matrix and the strictly
!                 upper triangular part of MAT is not referenced.
!
!               MAT is not modified by the subroutine.
!
!               The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   MATINV      (OUTPUT) real(stnd), dimension(:,:)
!               On exit, the (triangular) inverse of the original matrix
!               in the same storage format. 
!
!               The shape of MATINV must verify:  size( MATINV, 1 ) = size( MATINV, 2 ) = n .
!
!   UPPER       (INPUT, OPTIONAL) logical(lgl)
!               Specifies whether the matrix MAT is upper or lower triangular. If:
!
!               - UPPER = true :   MAT is upper triangular,
!               - UPPER = false:   MAT is lower triangular.
!
!               The default is true.
!
!
! Further Details
! _______________
!
!   The computations are parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch
    use Utilities,       only : assert_eq
#ifdef _BLAS
    use BLAS_interfaces, only : axpy, trsv
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in)   :: mat(:,:)
    real(stnd),   intent(out)  :: matinv(:,:)
!
    logical(lgl), intent(in), optional :: upper
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: matjj, safmin
!
    integer(i4b) :: n, i, j
!
    logical(lgl) :: mat_is_upper
!
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_triang_inv'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq( int(size(mat,1),i4b) ,      &
                   int(size(mat,2),i4b) ,      &
                   int(size(matinv,1),i4b) ,   &
                   int(size(matinv,2),i4b) ,   &
                   name_proc )
!
    mat_is_upper = true
    if ( present(upper) ) mat_is_upper = upper
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!    
!   COMPUTE SAFE MINIMUM.
!    
    safmin = lamch( 's' )
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*n>=omp_limit                      .and.      &
                  i>1_i4b
!    
!   COMPUTE INVERSE OF mat.
!
    if ( test_par ) then
!
!$OMP PARALLEL PRIVATE(i,j,matjj)
!
!       FIRST, COMPUTE DIAGONAL OF INVERSE OF THE TRIANGULAR MATRIX.
!
!$OMP DO SCHEDULE(STATIC)
!
        do j = 1_i4b, n
!
            matjj       = mat(j,j)
            matinv(j,j) = one/sign( max(abs(matjj),safmin), matjj )
!
        end do
!
!$OMP END DO
!    
        if ( mat_is_upper ) then
!
!           COMPUTE INVERSE OF UPPER TRIANGULAR MATRIX.
!
!$OMP DO SCHEDULE(STATIC,1)
!
            do i = 1_i4b, n
!
                matinv(1_i4b:i-1_i4b,i) = -matinv(i,i)*mat(1_i4b:i-1_i4b,i)
!
#ifdef _BLAS
                call trsv( 'U', 'N', 'N', i-1_i4b, mat(1_i4b:n,1_i4b:i-1_i4b), n, matinv(1_i4b:i-1_i4b,i), 1_i4b )
#else
                do j = i-1_i4b, 1_i4b, -1_i4b
!
                    matjj                   = matinv(j,i)*matinv(j,j)
                    matinv(1_i4b:j-1_i4b,i) = matinv(1_i4b:j-1_i4b,i) - matjj*mat(1_i4b:j-1_i4b,j)
                    matinv(j,i)             = matjj
!
                end do
#endif
!
            end do
!
!$OMP END DO
!
        else
!
!
!           COMPUTE INVERSE OF LOWER TRIANGULAR MATRIX.
!
!$OMP DO SCHEDULE(STATIC,1)
!
            do i = 1_i4b, n
!
                matinv(i+1_i4b:n,i) = -matinv(i,i)*mat(i+1_i4b:n,i)
!
                do j = i+1_i4b, n
!
                    matjj               = matinv(j,i)*matinv(j,j)
#ifdef _BLAS
                    call axpy( n-j, -matjj, mat(j+1_i4b:n,j), 1_i4b, matinv(j+1_i4b:n,i), 1_i4b )
#else
                    matinv(j+1_i4b:n,i) = matinv(j+1_i4b:n,i) - matjj*mat(j+1_i4b:n,j)
#endif
                    matinv(j,i)         =  matjj
!
                end do
!
            end do
!
!$OMP END DO
!
        end if
!
!$OMP END PARALLEL
!    
    else
!
#endif

        if ( mat_is_upper ) then
!
!           COMPUTE INVERSE OF UPPER TRIANGULAR MATRIX.
!
            do j = 1_i4b, n
!
                matinv(1_i4b:j-1_i4b,j) = mat(1_i4b:j-1_i4b,j)
                matjj                   = one/sign( max(abs(mat(j,j)),safmin), mat(j,j) )
                matinv(j,j)             = matjj
!
!               COMPUTE ELEMENTS 1:j-1 OF COLUMN j.
!
                do i = 1_i4b, j-1_i4b
!
                    if ( matinv(i,j)==zero ) cycle
!
#ifdef _BLAS
                    call axpy( i-1_i4b, matinv(i,j), matinv(1_i4b:i-1_i4b,i), 1_i4b, matinv(1_i4b:i-1_i4b,j), 1_i4b )
#else
                    matinv(1_i4b:i-1_i4b,j) = matinv(1_i4b:i-1_i4b,j) + matinv(i,j)*matinv(1_i4b:i-1_i4b,i)
#endif
                    matinv(i,j)             = matinv(i,j)*matinv(i,i)
!
                end do
!
                matinv(1_i4b:j-1_i4b,j) = -matjj*matinv(1_i4b:j-1_i4b,j)
!
            end do
!
        else
!
!           COMPUTE INVERSE OF LOWER TRIANGULAR MATRIX.
!
            do j = n, 1_i4b, -1_i4b
!
                matjj               = one/sign( max(abs(mat(j,j)),safmin), mat(j,j) )
                matinv(j,j)         = matjj
                matinv(j+1_i4b:n,j) = mat(j+1_i4b:n,j)
!
!               COMPUTE ELEMENTS j+1:n OF COLUMN j.
!
                do i = n, j+1_i4b, -1_i4b
!
                    if ( matinv(i,j)==zero ) cycle
!
#ifdef _BLAS
                    call axpy( n-i, matinv(i,j), matinv(i+1_i4b:n,i), 1_i4b, matinv(i+1_i4b:n,j), 1_i4b )
#else
                    matinv(i+1_i4b:n,j) = matinv(i+1_i4b:n,j) + matinv(i,j)*matinv(i+1_i4b:n,i)
#endif
                    matinv(i,j)         = matinv(i,j)*matinv(i,i)
!
                end do
!
                matinv(j+1_i4b:n,j) = -matjj*matinv(j+1_i4b:n,j)
!
            end do
!
        end if
!
#ifdef _OPENMP
    end if
#endif
!
!
! END OF SUBROUTINE comp_triang_inv_b
! ___________________________________
!
    end subroutine comp_triang_inv_b
!
!
! =========================================================================================
!                   COMPUTING THE PRODUCT U*U' OR L'*L WHERE U OR L IS TRIANGULAR
! =========================================================================================
!
!
    subroutine comp_uut_ltl_a( mat, upper, fill )
!
! Purpose
! _______
!
!   COMP_UUT_LTL computes, in place, the product U * U' or L' * L, where the triangular
!   factor U or L is stored in the upper or lower triangular part of MAT.
!
!
! Arguments
! _________
!
!   MAT         (INPUT/OUTPUT) real(stnd), dimension(:,:)
!               On entry, if:
!
!               - UPPER = true or is absent:
!                 The leading n-by-n upper triangular part of MAT contains the
!                 upper triangular factor U and the strictly lower triangular
!                 part of MAT is not referenced.
!               - UPPER = false:
!                 The leading n-by-n lower triangular part of MAT contains the
!                 lower triangular factor L and the strictly upper triangular
!                 part of MAT is not referenced.
!
!               On exit, if:
!
!               - If FILL = true or is absent:
!                 The product U * U' or L' * L overwrites MAT.
!               - If FILL = false and UPPER = true or is absent:
!                 The leading n-by-n upper triangular part of MAT is overwritten 
!                 with the upper triangular part of the product U * U' and the strictly
!                 lower triangular part of MAT is not referenced.
!               - If FILL = false and UPPER = false:
!                 The leading n-by-n lower triangular part of MAT is overwritten
!                 with the lower triangular part of the product L' * L and the strictly
!                 upper triangular part of MAT is not referenced.
!
!               The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   UPPER       (INPUT, OPTIONAL) logical(lgl)
!               Specifies whether the triangular factor stored in matrix MAT 
!               is upper or lower triangular. If:
!
!               - UPPER = true :  Upper triangular is stored
!               - UPPER = false:  Lower triangularis stored.
!
!               The default is true.
!
!   FILL        (INPUT, OPTIONAL) logical(lgl)
!               On entry, when argument FILL is present, FILL is used as follows. If:
!
!               - FILL= true and UPPER= true,  the lower triangle of MAT is filled on exit.
!               - FILL= true and UPPER= false,  the upper triangle of MAT is filled on exit.
!               - FILL= false, the lower (UPPER= true) or upper (UPPER= false) triangle of MAT
!                 is not filled on exit.
!
!               The default is true.
!
!
! Further Details
! _______________
!
!   The computations are parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert_eq
#ifdef _BLAS
    use BLAS_interfaces, only : gemv
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout)         :: mat(:,:)
!
    logical(lgl), intent(in), optional :: upper, fill
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(mat,1)) :: tmp, tmp2
!
    integer(i4b) :: n, i
!
    logical(lgl) :: mat_is_upper, fill_triang
!
#ifdef _OPENMP
#ifndef _ORDERED
    logical      :: test_par, ompparallel, ompnested
#endif
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_uut_ltl'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq( int(size(mat,1),i4b) ,      &
                   int(size(mat,2),i4b) ,      &
                   name_proc )
!
    mat_is_upper = true
    if ( present(upper) ) mat_is_upper = upper
!
    fill_triang = true
    if ( present(fill) ) fill_triang = fill
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!
#ifdef _OPENMP
#ifndef _ORDERED
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.     &
                  n*n>=omp_limit                      .and.     &
                  i>1_i4b
!
    if ( test_par ) then
!
        if ( mat_is_upper ) then
!    
!           FORM THE PRODUCT U*U'.
!
!$OMP PARALLEL DO SCHEDULE(STATIC,1)  &
!$OMP            ,PRIVATE(tmp,tmp2,i) &
!$OMP            ,ORDERED
!
            do i = 1_i4b, n
!
                tmp2(i:n) = mat(i,i:n)
!
#ifdef _USE_PGI
                tmp(1_i4b:i) = matmul( mat(1_i4b:i,i:n), tmp2(i:n) )
#else
#ifdef _BLAS
                call gemv( 'N', i, n-i+1_i4b, one, mat(1_i4b:n,i:n), n, tmp2(i:n), 1_i4b, zero, tmp(1_i4b:i), 1_i4b )
#else
                tmp(1_i4b:i) = matmul( mat(1_i4b:i,i:n), tmp2(i:n) )
#endif
#endif
!
                if ( fill_triang ) then
                    mat(i,1_i4b:i-1_i4b) = tmp(1_i4b:i-1_i4b)
                end if
!
!$OMP ORDERED
                mat(1_i4b:i,i) = tmp(1_i4b:i)
!$OMP END ORDERED
!
            end do
!
!$OMP END PARALLEL DO
!
        else
!    
!           FORM THE PRODUCT L'*L.
!
!$OMP PARALLEL DO SCHEDULE(STATIC,1)     &
!$OMP            ,PRIVATE(tmp,i)         &
!$OMP            ,ORDERED
!
            do i = 1_i4b, n
!
                tmp(1_i4b:i) = matmul( mat(i:n,i), mat(i:n,1_i4b:i) )
!
                if ( fill_triang ) then
                    mat(1_i4b:i-1_i4b,i) = tmp(1_i4b:i-1_i4b)
                end if
!
!$OMP ORDERED
                mat(i,1_i4b:i) = tmp(1_i4b:i)
!$OMP END ORDERED
!
            end do
!
!$OMP END PARALLEL DO
!
        end if
!
    else
!
#endif
#endif
        if ( mat_is_upper ) then
!    
!           FORM THE PRODUCT U*U'.
!
            do i = 1_i4b, n
!
                tmp2(i:n) = mat(i,i:n)
!
#ifdef _BLAS
                call gemv( 'N', i, n-i+1_i4b, one, mat(1_i4b:n,i:n), n, tmp2(i:n), 1_i4b, zero, tmp(1_i4b:i), 1_i4b )
#else
                tmp(1_i4b:i) = matmul( mat(1_i4b:i,i:n), tmp2(i:n) )
#endif
!
                if ( fill_triang ) then
                    mat(i,1_i4b:i-1_i4b) = tmp(1_i4b:i-1_i4b)
                end if
!
                mat(1_i4b:i,i) = tmp(1_i4b:i)
!
            end do
!
        else
!    
!           FORM THE PRODUCT L'*L.
!
            do i = 1_i4b, n
!
                tmp(1_i4b:i) = matmul( mat(i:n,i), mat(i:n,1_i4b:i) )
!
                if ( fill_triang ) then
                    mat(1_i4b:i-1_i4b,i) = tmp(1_i4b:i-1_i4b)
                end if
!
                mat(i,1_i4b:i) = tmp(1_i4b:i)
!
            end do
!
        end if
!
#ifdef _OPENMP
#ifndef _ORDERED
    end if    
#endif
#endif
!
! END OF SUBROUTINE comp_uut_ltl_a
! ________________________________
!
    end subroutine comp_uut_ltl_a
!
! =========================================================================================
!
    subroutine comp_uut_ltl_b( mat, prod, upper, fill )
!
! Purpose
! _______
!
!   COMP_UUT_LTL computes the product U * U' or L' * L, where the triangular
!   factor U or L is stored in the upper or lower triangular part of MAT.
!
!
! Arguments
! _________
!
!   MAT         (INPUT) real(stnd), dimension(:,:)
!               On entry, the triangular factor U or L.
!
!               The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   PROD        (OUTPUT) real(stnd), dimension(:,:)
!               On exit, if:
!
!               - FILL = true or is absent:
!                 The product U * U' or L' * L .
!               - FILL = false and UPPER = true or is absent:
!                 The leading n-by-n upper triangular part of PROD contains
!                 the upper triangular part of the product U*Ut and the strictly
!                 lower triangular part of PROD is not referenced.
!               - FILL = false and UPPER = false:
!                 The leading n-by-n lower triangular part of PROD contains
!                 the lower triangular part of the product Lt*L and the strictly
!                 upper triangular part of PROD is not referenced.
!
!               The shape of PROD must verify:  size( PROD, 1 ) = size( PROD, 2 ) = n .
!
!   UPPER       (INPUT, OPTIONAL) logical(lgl)
!               Specifies whether the triangular factor stored in matrix MAT is
!               the upper or lower triangle. If:
!
!               - UPPER = true :  Upper triangular is stored
!               - UPPER = false:  Lower triangular is stored.
!
!               The default is true.
!
!   FILL        (INPUT, OPTIONAL) logical(lgl)
!               On entry, when argument FILL is present, FILL is used as follows. If:
!
!               - FILL= true and UPPER= true,  the lower triangle of PROD is filled on exit.
!               - FILL= true and UPPER= false,  the upper triangle of PROD is filled on exit.
!               - FILL= false, the lower (UPPER= true) or upper (UPPER= false) triangle of PROD
!                 is not filled on exit.
!
!               The default is true.
!
!
! Further Details
! _______________
!
!   The computations are parallelized if OPENMP is used.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities,       only : assert_eq
#ifdef _BLAS
    use BLAS_interfaces, only : gemv
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in)   :: mat(:,:)
    real(stnd),   intent(out)  :: prod(:,:)
!
    logical(lgl), intent(in), optional :: upper, fill
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(mat,1)) :: tmp
!
    integer(i4b) :: n, i
!
    logical(lgl) :: mat_is_upper, fill_triang
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_uut_ltl'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq( int(size(mat,1),i4b) ,    &
                   int(size(mat,2),i4b) ,    &
                   int(size(prod,1),i4b) ,   &
                   int(size(prod,2),i4b) ,   &
                   name_proc )
!
    mat_is_upper = true
    if ( present(upper) ) mat_is_upper = upper
!
    fill_triang = true
    if ( present(fill) ) fill_triang = fill
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.     &
                  n*n>=omp_limit                      .and.     &
                  i>1_i4b
#endif
!
    if ( mat_is_upper ) then
!    
!       FORM THE PRODUCT U*U'.
!
!$OMP PARALLEL DO IF (test_par)          &
!$OMP            ,SCHEDULE(STATIC,1)     &
!$OMP            ,PRIVATE(i,tmp)
!
        do i = 1_i4b, n
!
            tmp(i:n) = mat(i,i:n)
!
#ifdef _BLAS
            call gemv( 'N', i, n-i+1_i4b, one, mat(1_i4b:n,i:n), n, tmp(i:n), 1_i4b, zero, prod(1_i4b:i,i), 1_i4b )
#else
            prod(1_i4b:i,i) = matmul( mat(1_i4b:i,i:n), tmp(i:n) )
#endif
!
            if ( fill_triang ) then
                prod(i,1_i4b:i-1_i4b) = prod(1_i4b:i-1_i4b,i)
            end if
!
        end do
!
!$OMP END PARALLEL DO
!
    else
!    
!       FORM THE PRODUCT L'*L.
!
!$OMP PARALLEL DO IF (test_par)          &
!$OMP            ,SCHEDULE(STATIC,1)     &
!$OMP            ,PRIVATE(i)
!
        do i = 1_i4b, n
!
            prod(i,1_i4b:i) = matmul( mat(i:n,i), mat(i:n,1_i4b:i) )
!
            if ( fill_triang ) then
                prod(1_i4b:i-1_i4b,i) = prod(i,1_i4b:i-1_i4b)
            end if
!
        end do
!
!$OMP END PARALLEL DO
!
    end if
!
!
! END OF SUBROUTINE comp_uut_ltl_b
! ________________________________
!
    end subroutine comp_uut_ltl_b
!
!
! =========================================================================================
!                            COMPUTING THE DETERMINANT OF A MATRIX
! =========================================================================================
!
!
    subroutine comp_det( mat, det, tol, man_det, exp_det )
!
! Purpose
! _______
!
!   COMP_DET computes the determinant of a real matrix MAT
!
!                   DET = determinant( MAT )
!
!
! Arguments
! _________
!
!   MAT         (INPUT/OUTPUT) real(stnd), dimension(:,:)
!               On entry, the real matrix MAT.
!
!               On exit, MAT is destroyed. 
!
!               The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   DET         (OUTPUT) real(stnd)
!               On exit, the determinant of MAT.
!
!   TOL         (INPUT, OPTIONAL) real(stnd)
!               On entry, a relative tolerance used to indicate whether or not MAT
!               is nearly singular. Tol should normally be choose as approximately
!               the largest relative error in the elements of MAT. For example, 
!               if the elements of MAT are correct to about 4 significant figures,
!               then TOL should be set to about 5 * 10**(-4). If TOL is supplied as
!               less than EPS, where EPS is the relative machine precision, then
!               the value EPS is used in place of TOL.
!
!               Default: EPS, the relative machine precision.
!
!   MAN_DET     (OUTPUT, OPTIONAL) real(stnd)
!               On exit, the mantissa of the determinant of MAT.
!
!   EXP_DET     (OUTPUT, OPTIONAL) integer(i4b)
!               On exit, the exponent of the determinant of MAT.
!
!
! Further Details
! _______________
!
!   The LU decomposition with partial pivoting and implicit row
!   scaling of the matrix MAT is used to compute the determinant.
!
!   MAT is declared singular if a diagonal element of U is such that
!
!             abs( U(j,j) ) <= n * norm( MAT(j,:) ) * TOL
!
!   where norm( MAT(j,:) ) denotes the maximum of the absolute values of 
!   the jth row of the matrix MAT. In this case, a diagonal element of U is
!   small, indicating that MAT is singular or nearly singular and DET is set
!   to zero.
!
!   On exit:
!
!   - DET = nan()                       if MAT is a zero sized matrix.
!   - DET = scale( MAN_DET, EXP_DET )   if minexponent(DET) <= EXP_DET <= maxexponent(DET)
!   - DET = sign( 0, MAN_DET )          if EXP_DET < minexponent(DET)
!   - DET = sign( huge(DET), MAN_DET )  if maxexponent(DET) < EXP_DET
!
!   A blocked algorithm is used to compute the LU factorization. Furthermore,
!   the computations are parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Higham, N.J., 2011:
!         Gaussian elimination.
!         Wiley Interdisciplinary Reviews: Computational Statistics,
!         Vol. 3, Issue 3, pp 230-238.
!   
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch, machbase, machhuge, maxexp, minexp, nan
    use Utilities,       only : assert_eq, imaxloc
#ifdef _BLAS
    use BLAS_interfaces, only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(inout)         :: mat(:,:)
    real(stnd),   intent(out)           :: det
    real(stnd),   intent(in), optional  :: tol
    real(stnd),   intent(out), optional :: man_det 
!
    integer(i4b),   intent(out), optional :: exp_det 
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                          :: piv, tol_piv, d1, safmin
    real(stnd), dimension(size(mat,1))  :: tmp_row, equil
!
    integer(i4b)                         :: n, i, i1, i2, imax, j, k, l, rem, n2, d2
    integer(i4b), dimension(size(mat,1)) :: ip
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    real(stnd), parameter :: invmachbase=one/machbase
!
    character(len=*),  parameter :: name_proc='comp_det'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    name_proc )
!
    det = nan()
!
    if ( present( man_det ) ) man_det = nan()
    if ( present( exp_det ) ) exp_det = 0_i4b
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!    
    safmin = lamch( 's' )
!    
    tmp_row(:n) = maxval( abs(mat(:n,:n)), dim=2 )
!
    if ( any( tmp_row(:n)<safmin ) ) then
!
        det = zero
!
        if ( present( man_det ) ) man_det = zero
!
        return
!
    end if
!    
!   INITIALIZE TOLERANCE AND IMPLICIT SCALING VECTOR.
!    
    tol_piv = lamch( 'u' )
    if ( present( tol ) ) tol_piv = max( tol, tol_piv )
    tol_piv = tol_piv*real( n, stnd )
!    
    equil(:n) = one/tmp_row(:n)
!   
!   TRIANGULATION.  
!    
    d1 = one
    d2 = 0_i4b
!    
    rem = mod( n, blksz )
!
    if ( rem/=0_i4b ) then
!
        n2   = n - rem
!
    else
!
        n2   = n - blksz
!
    end if
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*n>=omp_limit                      .and.      &
                  n2/blksz>=1_i4b                     .and.      &
                  i>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par),PRIVATE(i,i1,i2,j,k,l,piv)
!
!   START THE BLOCKED LU FACTORIZATION OF mat.
!
    do j = 0_i4b, n2, blksz
!
        i1 = j + 1_i4b
        i2 = min( j + blksz, n )
!
!$OMP SINGLE
!
!       COMPUTE THE LU FACTORIZATION OF mat(j+1:n,j+1:j+blksz) IN PLACE.
!
        do i = i1, i2
!
!           FIND PIVOT.
!
            tmp_row(i:n) = equil(i:n)*abs( mat(i:n,i) )
            imax         = i - 1_i4b + imaxloc( tmp_row(i:n) )
            piv          = mat(imax,i)
!
!           TEST FOR SINGULARITY.
!
            if ( tmp_row(imax)<=tol_piv ) then
!
!               mat IS SINGULAR.
!
                d1 = zero
                d2 = 0_i4b
!
                exit
!
            end if
!
!           IF NECESSARY, APPLY ROW INTERCHANGE TO mat(1:n,j+1:j+blksz).
!
            if ( imax/=i ) then
!
                d1 = -d1
!
                tmp_row(i1:i2)  = mat(imax,i1:i2)
                mat(imax,i1:i2) = mat(i,i1:i2)
                mat(i,i1:i2)    = tmp_row(i1:i2)
!
                equil(imax) = equil(i)
!
            else
!
                tmp_row(i+1_i4b:i2) = mat(i,i+1_i4b:i2)
!
            end if
!
!           STORE PERMUTATION.
!
            ip(i) = imax
!
!           COMPUTE DETERMINANT.
!
            d1 = d1*piv
!
            do
!
                if ( abs(d1)<=one ) exit
!
                d1 = d1*invmachbase
                d2 = d2 + 1_i4b
!
            end do
!
            do
!
                if ( abs(d1)>=invmachbase ) exit
!
                d1 = d1*machbase
                d2 = d2 - 1_i4b
!
            end do
!
            piv              = one/piv
            mat(i+1_i4b:n,i) = mat(i+1_i4b:n,i)*piv
!
!           REDUCE REMAINING MATRIX.
!
            do k = i+1_i4b, i2
!
#ifdef _BLAS
                call axpy( n-i, -tmp_row(k), mat(i+1_i4b:n,i), 1_i4b, mat(i+1_i4b:n,k), 1_i4b )
#else
                mat(i+1_i4b:n,k) = mat(i+1_i4b:n,k) - tmp_row(k)*mat(i+1_i4b:n,i)
#endif
!
            end do
!
        end do
!
!$OMP END SINGLE
!
!       TEST IF mat IS SINGULAR.
!
        if ( d1==zero ) exit
!
!$OMP DO SCHEDULE(STATIC)
!
        do l = i2+1_i4b, n
!
!           APPLY PERMUTATIONS TO LATER COLUMNS, EG TO mat(1:n,j+blksz+1:n).
!
            do i = i1, i2
!
                k        = ip(i)
                piv      = mat(k,l)
                mat(k,l) = mat(i,l)
                mat(i,l) = piv
!
            end do
!    
!           APPLY TRANSFORMATIONS TO LATER COLUMNS, EG TO mat(j+1:n,j+blksz+1:n).
!    
            do i = i1, i2
!
#ifdef _BLAS
                call axpy( n-i, -mat(i,l), mat(i+1_i4b:n,i), 1_i4b, mat(i+1_i4b:n,l), 1_i4b )
#else
                mat(i+1_i4b:n,l) = mat(i+1_i4b:n,l) - mat(i,l)*mat(i+1_i4b:n,i)
#endif
!
            end do
!
        end do
!
!$OMP END DO
!          
    end do
!
!$OMP END PARALLEL
!
!   COMPUTE DETERMINANT.
!
    select case( d2 )
!
        case ( :minexp-1_i4b )
!
            det = sign( zero, d1 )
!
        case ( maxexp+1_i4b: )
!
            det = sign( machhuge, d1 )
!
        case default
!
!            det = scale( d1, d2 )
            det = d1*( machbase**d2 )
!
    end select
!
    if ( present( man_det ) ) man_det = d1
    if ( present( exp_det ) ) exp_det = d2
!
!
! END OF SUBROUTINE comp_det
! __________________________
!
    end subroutine comp_det
!
! =========================================================================================
!
    function det( mat, tol )
!
! Purpose
! _______
!
!   DET computes the determinant of a real matrix MAT
!
!                   determinant( MAT ) = DET( MAT )
!
!
! Arguments
! _________
!
!   MAT     (INPUT) real(stnd), dimension(:,:)
!           On entry, the real matrix MAT.
!           MAT is not modified by the routine.
!
!           The shape of MAT must verify:  size( MAT, 1 ) = size( MAT, 2 ) = n .
!
!   TOL     (INPUT, OPTIONAL) real(stnd)
!           On entry, a relative tolerance used to indicate whether or not MAT
!           is nearly singular. Tol should normally be choose as approximately
!           the largest relative error in the elements of MAT. For example, 
!           if the elements of MAT are correct to about 4 significant figures,
!           then TOL should be set to about 5 * 10**(-4). If TOL is supplied as
!           less than EPS, where EPS is the relative machine precision, then
!           the value EPS is used in place of TOL.
!
!           Default: EPS, the relative machine precision.
!
!
! Further Details
! _______________
!
!   The LU decomposition with partial pivoting and implicit row
!   scaling of the matrix MAT is used to compute the determinant.
!
!   MAT is declared singular if a diagonal element of U is such that
!
!             abs( U(j,j) ) <= n * norm( MAT(j,:) ) * TOL
!
!   where norm( MAT(j,:) ) denotes the maximum of the absolute values of 
!   the jth row of the matrix MAT. In this case, a diagonal element of U is
!   small, indicating that MAT is singular or nearly singular and DET returns
!   the value zero.
!
!   If MAT is a zero sized matrix, DET( MAT ) = nan().
!
!   A blocked algorithm is used to compute the LU factorization. Furthermore,
!   the computations are parallelized if OPENMP is used.
!
!   For further details, see:
!
!   (1) Golub, G.H., and Van Loan, C.F., 1996:
!           Matrix Computations.
!           3rd ed. The Johns Hopkins University Press, Baltimore.
!
!   (2) Higham, N.J., 2011:
!         Gaussian elimination.
!         Wiley Interdisciplinary Reviews: Computational Statistics,
!         Vol. 3, Issue 3, pp 230-238.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,   only : lamch, machbase, machhuge, maxexp, minexp, nan
    use Utilities,       only : assert_eq, imaxloc
#ifdef _BLAS
    use BLAS_interfaces, only : axpy
#endif
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),   intent(in)            :: mat(:,:)
    real(stnd),   intent(in), optional  :: tol
!
    real(stnd) :: det
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                      :: piv, tol_piv, d1, safmin
    real(stnd), dimension(size(mat,1))              :: tmp_row, equil
    real(stnd), dimension(size(mat,1),size(mat,2))  :: mat2
!
    integer(i4b)                         :: n, i, i1, i2, imax, j, k, l, rem, n2, d2
    integer(i4b), dimension(size(mat,1)) :: ip
!
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    real(stnd), parameter :: invmachbase=one/machbase
!
    character(len=*),  parameter :: name_proc='det'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( int(size(mat,1),i4b) ,   &
                    int(size(mat,2),i4b) ,   &
                    name_proc )
!
    det = nan()
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!    
    safmin = lamch( 's' )
!    
    tmp_row(:n) = maxval( abs(mat(:n,:n)), dim=2 )
!
    if ( any( tmp_row(:n)<safmin ) ) then
!
        det = zero
!
        return
!
    end if
!    
!   INITIALIZE TOLERANCE AND IMPLICIT SCALING VECTOR.
!    
    tol_piv = lamch( 'u' )
    if ( present( tol ) ) tol_piv = max( tol, tol_piv )
    tol_piv = tol_piv*real( n, stnd )
!    
    equil(:n) = one/tmp_row(:n)
!
!   MAKE A COPY OF mat TO COMPUTE THE LU DECOMPOSITION WITHOUT SIDE EFFECTS.
!
    mat2(:n,:n) = mat(:n,:n)
!   
!   TRIANGULATION.  
!    
    d1 = one
    d2 = 0_i4b
!    
    rem = mod( n, blksz )
!
    if ( rem/=0_i4b ) then
!
        n2   = n - rem
!
    else
!
        n2   = n - blksz
!
    end if
!
#ifdef _OPENMP
    i           = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  n*n>=omp_limit                      .and.      &
                  n2/blksz>=1_i4b                     .and.      &
                  i>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par),PRIVATE(i,i1,i2,j,k,l,piv)
!
!   START THE BLOCKED LU FACTORIZATION OF mat.
!
    do j = 0_i4b, n2, blksz
!
        i1 = j + 1_i4b
        i2 = min( j + blksz, n )
!
!$OMP SINGLE
!
!       COMPUTE THE LU FACTORIZATION OF mat(j+1:n,j+1:j+blksz) IN PLACE.
!
        do i = i1, i2
!
!           FIND PIVOT.
!
            tmp_row(i:n) = equil(i:n)*abs( mat2(i:n,i) )
            imax         = i - 1_i4b + imaxloc( tmp_row(i:n) )
            piv          = mat2(imax,i)
!
!           TEST FOR SINGULARITY.
!
            if ( tmp_row(imax)<=tol_piv ) then
!
!               mat IS SINGULAR.
!
                d1 = zero
                d2 = 0_i4b
!
                exit
!
            end if
!
!           IF NECESSARY, APPLY ROW INTERCHANGE TO mat(1:n,j+1:j+blksz).
!
            if ( imax/=i ) then
!
                d1 = -d1
!
                tmp_row(i1:i2)   = mat2(imax,i1:i2)
                mat2(imax,i1:i2) = mat2(i,i1:i2)
                mat2(i,i1:i2)    = tmp_row(i1:i2)
!
                equil(imax) = equil(i)
!
            else
!
                tmp_row(i+1_i4b:i2) = mat2(i,i+1_i4b:i2)
!
            end if
!
!           STORE PERMUTATION.
!
            ip(i) = imax
!
!           COMPUTE DETERMINANT.
!
            d1 = d1*piv
!
            do
!
                if ( abs(d1)<=one ) exit
!
                d1 = d1*invmachbase
                d2 = d2 + 1_i4b
!
            end do
!
            do
!
                if ( abs(d1)>=invmachbase ) exit
!
                d1 = d1*machbase
                d2 = d2 - 1_i4b
!
            end do
!
            piv               = one/piv
            mat2(i+1_i4b:n,i) = mat2(i+1_i4b:n,i)*piv
!
!           REDUCE REMAINING MATRIX.
!
            do k = i+1_i4b, i2
!
#ifdef _BLAS
                call axpy( n-i, -tmp_row(k), mat2(i+1_i4b:n,i), 1_i4b, mat2(i+1_i4b:n,k), 1_i4b )
#else
                mat2(i+1_i4b:n,k) = mat2(i+1_i4b:n,k) - tmp_row(k)*mat2(i+1_i4b:n,i)
#endif
!
            end do
!
        end do
!
!$OMP END SINGLE
!
!       TEST IF mat IS SINGULAR.
!
        if ( d1==zero ) exit
!
!$OMP DO SCHEDULE(STATIC)
!
        do l = i2+1_i4b, n
!
!           APPLY PERMUTATIONS TO LATER COLUMNS, EG TO mat(1:n,j+blksz+1:n).
!
            do i = i1, i2
!
                k         = ip(i)
                piv       = mat2(k,l)
                mat2(k,l) = mat2(i,l)
                mat2(i,l) = piv
!
            end do
!    
!           APPLY TRANSFORMATIONS TO LATER COLUMNS, EG TO mat(j+1:n,j+blksz+1:n).
!    
            do i = i1, i2
!
#ifdef _BLAS
                call axpy( n-i, -mat2(i,l), mat2(i+1_i4b:n,i), 1_i4b, mat2(i+1_i4b:n,l), 1_i4b )
#else
                mat2(i+1_i4b:n,l) = mat2(i+1_i4b:n,l) - mat2(i,l)*mat2(i+1_i4b:n,i)
#endif
!
            end do
!
        end do
!
!$OMP END DO
!          
    end do
!
!$OMP END PARALLEL
!
!   COMPUTE DETERMINANT.
!
    select case( d2 )
!
        case ( :minexp-1_i4b )
!
            det = sign( zero, d1 )
!
        case ( maxexp+1_i4b: )
!
            det = sign( machhuge, d1 )
!
        case default
!
!            det = scale( d1, d2 )
            det = d1*( machbase**d2 )
!
    end select
!
!
! END OF FUNCTION det
! ___________________
!
    end function det
!
!
! =========================================================================================
!                            SYMMETRIC TRIDIAGONAL SYSTEMS
! =========================================================================================
!
!
    subroutine sym_trid_cmp( d, e, sub, diag, sup1, sup2, perm, tol )
!
! Purpose
! _______
!
!  SYM_TRID_CMP factorizes an n by n symmetric tridiagonal matrix T as
!
!                         T = P * L * U
!
!  where P is a permutation matrix, L is a unit lower tridiagonal matrix
!  with at most one non-zero sub-diagonal elements per column and U is
!  an upper triangular matrix with at most two non-zero super-diagonal
!  elements per column.
!
!  The factorization is obtained by Gaussian elimination with
!  partial pivoting and implicit row scaling.
!
!
! Arguments
! _________
!
!    D           (INPUT) real(stnd), dimension(:)
!                On entry, the diagonal elements of the symmetric tridiagonal matrix T.
!
!    E           (INPUT) real(stnd), dimension(:)
!                On entry, the n-1 subdiagonal elements of the symmetric tridiagonal
!                matrix T and E(n) is arbitrary .
!
!                The size of E must be  size( E ) = size( D ) = n .
!
!    SUB         (OUTPUT) real(stnd), dimension(:)
!                On exit, SUB(:n-1) contains the n-1 subdiagonal elements of the
!                lower triangular matrix L of the factorization of T,
!                SUB(n) is arbitrary .
!
!                The size of SUB must verify:  size( SUB ) = size( D ) = n .
!
!    DIAG        (OUTPUT) real(stnd), dimension(:)
!                On exit, DIAG(:) contains the n diagonal elements of the
!                upper triangular matrix U of the factorization of T.
!               
!                The size of DIAG must verify:  size( DIAG ) = size( D ) = n .
!
!    SUP1        (OUTPUT) real(stnd), dimension(:)
!                On exit, SUP1(:n-1) contains the n-1 superdiagonal elements of the
!                upper triangular matrix U of the factorization of T,
!                SUP1(n) is arbitrary .
!
!                The size of SUP1 must verify:  size( SUP1 ) = size( D ) = n .
!
!    SUP2        (OUTPUT) real(stnd), dimension(:)
!                On exit, SUP2(:n-2) contains the n-2 second superdiagonal elements of
!                the upper triangular matrix U of the factorization of T,
!                SUP2(n-1:n) is arbitrary .
!
!                The size of SUP2 must verify:  size( SUP2 ) = size( D ) = n .
!
!    PERM        (OUTPUT) logical(lgl), dimension(:)
!                On exit, PERM(:n-1)  contains details of the permutation matrix P(j):
!
!                - if, an interchange occurred at the kth step of the elimination in the
!                  factorization of T, then PERM(k) = TRUE
!                - otherwise, PERM(k) = FALSE.
!
!                The element PERM(n) is set to TRUE, if there is an integer l such that
!
!                  abs( U(l,l) ) <= norm( T(l) ) * TOL,
!
!                where norm( T(l) ) denotes the sum of the absolute values of
!                the lth row of the matrix T. If no such l exists then PERM(n)
!                is returned as FALSE.
!
!                If PERM(n) is returned as TRUE, then a diagonal element of U is
!                small, indicating that T is singular or nearly singular.
!
!                The size of PERM must verify:  size( PERM ) = size( D ) = n .
!
!    TOL         (INPUT, OPTIONAL)  real(stnd)
!                On entry, a relative tolerance used to indicate whether or
!                not the matrix T is nearly singular. TOL should normally
!                be choose as approximately the largest relative error
!                in the elements of T. For example, if the elements of T are
!                correct to about 4 significant figures, then TOL should be
!                set to about 5 * 10**(-4). If TOL is supplied as less than EPS,
!                where EPS is the relative machine precision, then the value
!                EPS is used in place of TOL.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the subroutine  DLAGTF in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Num_Constants,    only : lamch
    use Utilities,        only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),  dimension(:)    :: d, e
    real(stnd),  intent(out), dimension(:)    :: sub, diag, sup1, sup2
    real(stnd),  intent(in),  optional        :: tol
!
    logical(lgl), intent(out), dimension(:) :: perm
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: tl, ek, ekp1, tmp1, tmp2, scal(size(d))
!
    integer(i4b) :: n, k
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='sym_trid_cmp'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( (/ int(size(d),i4b)       ,   &
                       int(size(e),i4b)       ,   &
                       int(size(sub),i4b)     ,   &
                       int(size(diag),i4b)    ,   &
                       int(size(sup1),i4b)    ,   &
                       int(size(sup2),i4b)    ,   &
                       int(size(perm),i4b) /) ,   &
                       name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            diag(1_i4b) = d(1_i4b)
            perm(1_i4b) = diag(1_i4b)==zero
            sub(1_i4b)  = zero
            sup1(1_i4b) = zero
            sup2(1_i4b) = zero
!
        end if
!
        return
!
    end if
!    
!   INITIALIZE TOLERANCE, IMPLICIT SCALING VECTOR AND WORK VARIABLES.
!    
    tl = lamch( 'u' )
    if ( present(tol) ) tl = max( tl, tol )
!    
    diag(:n) = d(:n)
!
    sup1(:n) = e(:n)
    sup2(:n) = zero
!    
    sub(1_i4b) = abs( e(1_i4b) )
    sub(2_i4b:n-1_i4b) = abs(e(2_i4b:n-1_i4b)) + abs(e(:n-2_i4b))
    sub(n) = abs( e(n-1_i4b) )
!
    scal(:n) = abs( diag(:n) ) + sub(:n)
!    
    perm(:n) = false
!   
!   TRIANGULATION.  
!    
    ekp1 = e(1_i4b)
!
    do k = 1_i4b, n-1_i4b
!
        if ( diag(k)/=zero ) then
            tmp1 = abs( diag(k) )/scal(k)
        else
            tmp1 = zero
        end if
!
        ek = ekp1
        ekp1 = e(k+1_i4b)
!
        if ( ek/=zero ) then
!
            tmp2 = abs( ek )/scal(k+1_i4b)
!
            if ( tmp2<=tmp1 ) then
!   
!               NO INTERCHANGE REQUIRED, ELIMINATE sub(k)=e(k) .  
!    
                sub(k) = ek/diag(k)
                diag(k+1_i4b) = diag(k+1_i4b) - sub(k)*sup1(k)
!
            else
!   
!               INTERCHANGE ELEMENTS k AND k+1, ELIMINATE sub(k)=diag(k) .  
!    
                perm(k) = true
                tmp1 = tmp2
                scal(k+1_i4b) = scal(k)
                sub(k) = diag(k)/ek
                diag(k) = ek
                tmp2 = diag(k+1_i4b)

                diag(k+1_i4b) = sup1(k) - sub(k)*tmp2
                sup2(k) = ekp1
                sup1(k+1_i4b) = -sub(k)*ekp1
                sup1(k) = tmp2
!
            end if
!
        end if
!
        if ( tmp1<=tl .and. .not. perm(n) ) perm(n) = true
!
    end do
!
    if ( abs( diag(n) )<=scal(n)*tl .and. .not. perm(n) ) perm(n) = true
!
    sup1(n) = zero
    sup2(n-1_i4b:n) = zero
    sub(n) = zero
!
!
! END OF SUBROUTINE sym_trid_cmp
! ______________________________
!
    end subroutine sym_trid_cmp
!
! =========================================================================================
!
    subroutine sym_trid_cmp2( d, e, sub, diag, sup1, sup2, perm )
!
! Purpose
! _______
!
!   SYM_TRID_CMP2 factorizes an n by n symmetric tridiagonal matrix T, as
!
!                         T = P * L * U
!
!   where P is a permutation matrix, L is a unit lower tridiagonal matrix
!   with at most one non-zero sub-diagonal elements per column and U is
!   an upper triangular matrix with at most two non-zero super-diagonal
!   elements per column.
!
!   The factorization is obtained by Gaussian elimination with partial pivoting
!   and row interchanges.
!
!
! Arguments
! _________
!
!    D           (INPUT) real(stnd), dimension(:)
!                On entry, the diagonal elements of the symmetric tridiagonal matrix T.
!
!    E           (INPUT) real(stnd), dimension(:)
!                On entry, the n-1 subdiagonal elements of the symmetric tridiagonal
!                matrix T and E(n) is arbitrary .
!
!                The size of E must be  size( E ) = size( D ) = n .
!
!    SUB         (OUTPUT) real(stnd), dimension(:)
!                On exit, SUB(:n-1) contains the n-1 subdiagonal elements of the
!                lower triangular matrix L of the factorization of T,
!                SUB(n) is arbitrary .
!
!                The size of SUB must verify:  size( SUB ) = size( D ) = n .
!
!    DIAG        (OUTPUT) real(stnd), dimension(:)
!                On exit, DIAG(:) contains the n diagonal elements of the
!                upper triangular matrix U of the factorization of T.
!               
!                The size of DIAG must verify:  size( DIAG ) = size( D ) = n .
!
!    SUP1        (OUTPUT) real(stnd), dimension(:)
!                On exit, SUP1(:n-1) contains the n-1 superdiagonal elements of the
!                upper triangular matrix U of the factorization of T,
!                SUP1(n) is arbitrary .
!
!                The size of SUP1 must verify:  size( SUP1 ) = size( D ) = n .
!
!    SUP2        (OUTPUT) real(stnd), dimension(:)
!                On exit, SUP2(:n-2) contains the n-2 second superdiagonal elements of
!                the upper triangular matrix U of the factorization of T,
!                SUP2(n-1:n) is arbitrary .
!
!                The size of SUP2 must verify:  size( SUP2 ) = size( D ) = n .
!
!    PERM        (OUTPUT) logical(lgl), dimension(:)
!                On exit, PERM(:n-1)  contains details of the permutation matrix P(j):
!
!                - If an interchange occurred at the kth step of the elimination in the
!                  factorization of T, then PERM(k) = TRUE
!                - Otherwise PERM(k) = FALSE.
!
!                PERM(n) is arbitrary .
!
!                The size of PERM must verify:  size( PERM ) = size( D ) = n .
!
!
! Further Details
! _______________
!
!   SYM_TRID_CMP2 is a simplified version of SYM_TRID_CMP.
!   This subroutine is adapted from the subroutine  DGTTRF in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),  dimension(:)  :: d, e
    real(stnd),  intent(out), dimension(:)  :: sub, diag, sup1, sup2
!
    logical(lgl), intent(out), dimension(:) :: perm
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: ek, ekp1, tmp
!
    integer(i4b) :: n, k
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='sym_trid_cmp2'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( (/ int(size(d),i4b)       ,   &
                       int(size(e),i4b)       ,   &
                       int(size(sub),i4b)     ,   &
                       int(size(diag),i4b)    ,   &
                       int(size(sup1),i4b)    ,   &
                       int(size(sup2),i4b)    ,   &
                       int(size(perm),i4b) /) ,   &
                       name_proc )
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=1_i4b ) then
!
        if ( n==1_i4b ) then
!
            diag(1_i4b) = d(1_i4b)
            perm(1_i4b) = diag(1_i4b)==zero
            sub(1_i4b)  = zero
            sup1(1_i4b) = zero
            sup2(1_i4b) = zero
!
        end if
!
        return
!
    end if
!    
!   INITIALIZE WORK VARIABLES.
!    
    diag(:n) = d(:n)
    sup1(:n) = e(:n)
    sup2(:n) = zero
    perm(:n) = false
!   
!   TRIANGULATION.  
!    
    ekp1 = e(1_i4b)
!
    do k = 1_i4b, n-1_i4b
!
        ek = ekp1
        ekp1 = e(k+1_i4b)
!
        if ( ek==zero ) cycle
!   
        if ( abs( ek )<=abs( diag(k) ) ) then
!   
!           NO INTERCHANGE REQUIRED, ELIMINATE sub(k)=ek .  
!    
            sub(k) = ek/diag(k)
            diag(k+1_i4b) = diag(k+1_i4b) - sub(k)*sup1(k)
!
        else
!   
!           INTERCHANGE ELEMENTS k AND k+1, ELIMINATE sub(k)=diag(k) .  
!    
            perm(k) = true
            sub(k) = diag(k)/ek
            diag(k) = ek
            tmp = diag(k+1_i4b)
            diag(k+1_i4b) = sup1(k) - sub(k)*tmp
            sup2(k) = ekp1
            sup1(k+1_i4b) = -sub(k)*ekp1
            sup1(k) = tmp
!
        end if
!
    end do
!
    sup1(n) = zero
    sup2(n-1_i4b:n) = zero
    sub(n) = zero
!
!
! END OF SUBROUTINE sym_trid_cmp2
! _______________________________
!
    end subroutine sym_trid_cmp2
!
! =========================================================================================
!
    subroutine sym_trid_solve_a( sub, diag, sup1, sup2, perm, y, scale )
!
! Purpose
! _______
!
!   SYM_TRID_SOLVE may be used to solve the system of equations
!
!                         x(:) * T = scale * y(:)
!
!   , where T is an n by n symmetric tridiagonal matrix and
!   scale is a scalar for x(:), following the factorization
!   of T by SYM_TRID_CMP or SYM_TRID_CMP2 as
!
!                                T = P * L * U
!
!   where P is a permutation matrix, L is a unit lower tridiagonal matrix
!   with at most one non-zero sub-diagonal elements per column and U is
!   an upper triangular matrix with at most two non-zero super-diagonal
!   elements per column.
!
!   The matrix T is assumed to be ill-conditioned, and frequent rescalings
!   are carried out in order to avoid overflow. However, no test for singularity
!   or near-singularity is included in this routine. Such tests must be performed
!   before calling this routine.
!
!
! Arguments
! _________
!
!    SUB         (INPUT) real(stnd), dimension(:)
!                On entry, SUB(:n-1) contains the n-1 subdiagonal elements of the
!                lower triangular matrix L of the factorization of T,
!
!                SUB(n) is arbitrary.
!
!                The size of SUB must verify:   size( SUB ) = size( Y ) = n .
!                              
!
!    DIAG        (INPUT) real(stnd), dimension(:)
!                On entry, DIAG(:) contains the n diagonal elements of the
!                upper triangular matrix U of the factorization of T.
!               
!                The shape of DIAG must verify:  size( DIAG ) = size( Y ) = n .
!
!    SUP1        (INPUT) real(stnd), dimension(:)
!                On entry, SUP1(:n-1) contains the n-1 superdiagonal elements of the
!                upper triangular matrix U of the factorization of T,
!
!                SUP1(n) is arbitrary.
!
!                The shape of SUP1 must verify:  size( SUP1 ) = size( Y ) = n .
!
!    SUP2        (INPUT) real(stnd), dimension(:)
!                On entry, SUP2(:n-2) contains the n-2 second superdiagonal elements of
!                the upper triangular matrix U of the factorization of T,
!                SUP2(n-1:n) is arbitrary.
!
!                The shape of SUP2 must verify:  size( SUP2 ) = size( Y ) = n .
!
!    PERM        (INPUT) logical(lgl), dimension(:)
!                On entry, PERM(:n-1)  contains details of the permutation matrix P:
!
!                - if an interchange occurred at the kth step of the elimination in the
!                  factorization of T, then PERM(k) = TRUE,
!                - otherwise PERM(k) = FALSE.
!
!                PERM(n) is arbitrary .
!
!                The shape of PERM must verify:   size( PERM ) = size( Y ) = n .
!
!    Y           (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the right hand side vector y.
!
!                On exit, Y is overwritten the solution vector x.
!
!                The shape of Y must verify:  size( Y ) = n .
!
!    SCALE       (OUTPUT) real(stnd)
!                On exit, the scalar SCALE.
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the subroutine  DLAGTS in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),    dimension(:) :: sub, diag, sup1, sup2
    real(stnd),  intent(inout), dimension(:) :: y
    real(stnd),  intent(out)                 :: scale
!
    logical(lgl), intent(in), dimension(:) :: perm
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: tmp
!
    integer(i4b) :: n, k
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='sym_trid_solve'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( (/ int(size(y),i4b)       ,   &
                       int(size(sub),i4b)     ,   &
                       int(size(diag),i4b)    ,   &
                       int(size(sup1),i4b)    ,   &
                       int(size(sup2),i4b)    ,   &
                       int(size(perm),i4b) /) ,   &
                       name_proc )
!
    scale = one
!
!   QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!    
!   SOLVE L*x = b .
!    
    do k = 2_i4b, n
!
        if ( perm(k-1_i4b) ) then
!
            tmp = y(k-1_i4b)
            y(k-1_i4b) = y(k)
            y(k) = tmp - sub(k-1_i4b)*y(k)
!
        else
!
            y(k) = y(k) - sub(k-1_i4b)*y(k-1_i4b)
!
        end if
!
    end do
!   
!   SOLVE U*x = b WITH SCALING.
!    
    if ( abs(y(n))>abs(diag(n)) ) then
!
        tmp = diag(n)/y(n)
        y(:n-1_i4b) = tmp*y(:n-1_i4b)
        y(n) = one
        scale = tmp
!
    else
!
        y(n) = y(n)/diag(n)
!
    end if
!    
    if ( n>1_i4b ) then
!
        tmp = y(n-1_i4b) - sup1(n-1_i4b)*y(n)
!
        if ( abs(tmp)>abs(diag(n-1_i4b)) ) then
!
            tmp = diag(n-1_i4b)/tmp
            y(:n) = tmp*y(:n)
            y(n-1_i4b) = one
            scale = tmp*scale
!
        else
!
            y(n-1_i4b) = tmp/diag(n-1_i4b)
!
        end if
!    
        do k = n-2_i4b, 1_i4b, -1_i4b
!
            tmp = y(k) - sup1(k)*y(k+1_i4b) - sup2(k)*y(k+2_i4b)
!
            if ( abs(tmp)>abs(diag(k)) ) then
!
                tmp = diag(k)/tmp
                y(:n) = tmp*y(:n)
                y(k) = one
                scale = tmp*scale
!
            else
!
                y(k) = tmp/diag(k)
!
            end if
!
        end do
!
    end if
!
!
! END OF SUBROUTINE sym_trid_solve_a
! __________________________________
!
    end subroutine sym_trid_solve_a
!
! =========================================================================================
!
    subroutine sym_trid_solve_b( sub, diag, sup1, sup2, perm, y )
!
! Purpose
! _______
!
!   SYM_TRID_SOLVE may be used to solve the system of equations
!
!                         x(:) * T = y(:)
!
!   , where T is an n by n symmetric tridiagonal matrix, following
!   the factorization of T by SYM_TRID_CMP or SYM_TRID_CMP2 as
!
!                                T = P * L * U
!
!   where P is a permutation matrix, L is a unit lower tridiagonal matrix
!   with at most one non-zero sub-diagonal elements per column and U is
!   an upper triangular matrix with at most two non-zero super-diagonal
!   elements per column.
!
!   The matrix T is assumed to be no singular and well-conditioned.
!
!
! Arguments
! _________
!
!    SUB         (INPUT) real(stnd), dimension(:)
!                On entry, SUB(:n-1) contains the n-1 subdiagonal elements of the
!                lower triangular matrix L of the factorization of T,
!                SUB(n) is arbitrary .
!
!                The size of SUB must verify:   size( SUB ) = size( Y ) = n .
!                              
!
!    DIAG        (INPUT) real(stnd), dimension(:)
!                On entry, DIAG(:) contains the n diagonal elements of the
!                upper triangular matrix U of the factorization of T.
!               
!                The shape of DIAG must verify:  size( DIAG ) = size( Y ) = n .
!
!    SUP1        (INPUT) real(stnd), dimension(:)
!                On entry, SUP1(:n-1) contains the n-1 superdiagonal elements of the
!                upper triangular matrix U of the factorization of T,
!                SUP1(n) is arbitrary.
!
!                The shape of SUP1 must verify:  size( SUP1 ) = size( Y ) = n .
!
!    SUP2        (INPUT) real(stnd), dimension(:)
!                On entry, SUP2(:n-2) contains the n-2 second superdiagonal elements of
!                the upper triangular matrix U of the factorization of T,
!                SUP2(n-1:n) is arbitrary.
!
!                The shape of SUP2 must verify:  size( SUP2 ) = size( Y ) = n .
!
!    PERM        (INPUT) logical(lgl), dimension(:)
!                On entry, PERM(:n-1)  contains details of the permutation matrix P:
!
!                - if an interchange occurred at the kth step of the elimination in the
!                  factorization of T, then PERM(k) = TRUE,
!                - otherwise PERM(k) = FALSE.
!
!                PERM(n) is arbitrary .
!
!                The shape of PERM must verify:   size( PERM ) = size( Y ) = n .
!
!    Y           (INPUT/OUTPUT) real(stnd), dimension(:)
!                On entry, the right hand side vector y.
!
!                On exit, Y is overwritten the solution vector x.
!
!                The shape of Y must verify:  size( Y ) = n .
!
!
! Further Details
! _______________
!
!   This subroutine is adapted from the routine  DLAGTS in LAPACK.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert_eq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd),  intent(in),    dimension(:) :: sub, diag, sup1, sup2
    real(stnd),  intent(inout), dimension(:) :: y
!
    logical(lgl), intent(in), dimension(:) :: perm
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: tmp
!
    integer(i4b) :: n, k
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='sym_trid_solve'
!
!
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n =  assert_eq( (/ int(size(y),i4b)       ,   &
                       int(size(sub),i4b)     ,   &
                       int(size(diag),i4b)    ,   &
                       int(size(sup1),i4b)    ,   &
                       int(size(sup2),i4b)    ,   &
                       int(size(perm),i4b) /) ,   &
                       name_proc )
!
!    QUICK RETURN IF POSSIBLE.
!
    if ( n<=0_i4b ) return
!    
!   SOLVE L*x = b .
!    
    do k = 2_i4b, n
!
        if ( perm(k-1_i4b) ) then
!
            tmp = y(k-1_i4b)
            y(k-1_i4b) = y(k)
            y(k) = tmp - sub(k-1_i4b)*y(k)
!
        else
!
            y(k) = y(k) - sub(k-1_i4b)*y(k-1_i4b)
!
        end if
!
    end do
!   
!   SOLVE U*x = b .
!    
    y(n) = y(n)/diag(n)
!
    if ( n>1_i4b ) then
!
        y(n-1_i4b) = ( y(n-1_i4b) - sup1(n-1_i4b)*y(n) )/diag(n-1_i4b)
!
        do k = n-2_i4b, 1_i4b, -1_i4b
!
            y(k) = ( y(k) - sup1(k)*y(k+1_i4b) - sup2(k)*y(k+2_i4b))/diag(k)
!
        end do
!
    end if
!
!
! END OF SUBROUTINE sym_trid_solve_b
! __________________________________
!
    end subroutine sym_trid_solve_b
!
! =========================================================================================
!
!
! ******************************
! END OF MODULE Lin_Procedures *
! ******************************
!    
end module Lin_Procedures
