module Mul_Stat_Procedures
! ##############################################################################
!
! Copyright 2020 IRD
!
! This file is part of statpack.
!
! statpack is free software: you can redistribute it and/or modify
! it under the terms of the GNU Lesser General Public License as
! published by the Free Software Foundation, either version 3 of
! the License, or (at your option) any later version.
!
! statpack is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU Lesser General Public License for more details.
!
! You can find a copy of the GNU Lesser General Public License
! in the statpack/doc directory.
!
! ######################################################################################
!                                                                                      *
! **************************************************************************************
! MODULE EXPORTING SUBROUTINES AND FUNCTIONS FOR MULTIVARIATE STATISTICAL COMPUTATIONS *
!                                                                                      *
! LATEST REVISION : 23/09/2020                                                         *
!                                                                                      *
! ######################################################################################
!                                                                                      *
! **************************************************************************************
!                                                                                      *
! THE C PROCESSOR MACROS USED IN THIS MODULE ARE:                                      *
!                                                                                      *
!  _OPENMP      FOR ACTIVATING OPENMP PARALLELIZATION                                  *
!  _DOT_PRODUCT FOR REPLACING THE dot_product INTRINSIC FUNCTION WITH STATPACK         *
!               dot_product2 FUNCTION                                                  *
!  _MATMUL      FOR REPLACING THE matmul INTRINSIC FUNCTION WITH STATPACK              *
!               matmul2 FUNCTION                                                       *
!  _WHERE       FOR REPLACING SOME where CONSTRUCTS BY do LOOPS INSIDE SOME OPENMP     *
!               DIRECTIVES (USEFUL FOR SOME COMPILERS LIKE IFORT WHEN OPENMP IS        *
!               USED)                                                                  *
!  _BLAS        FOR USING BLAS WHEN POSSIBLE                                           *
!  _USE_INTEL   FOR SIGNALING THAT THE INTEL IFORT COMPILER IS USED.                   *
!               HERE, THIS INCLUDES ONLY THE ACTIVATION OF THE _WHERE MACRO            *
!                                                                                      *
! **************************************************************************************
!
#ifdef _USE_INTEL
#define _WHERE
#endif
!
!
! USED MODULES
! ============
!   
    use Select_Parameters,  only : lgl, i4b, stnd
#ifdef _MATMUL
    use Utilities,          only : matmul=>matmul2
#endif
#ifdef _DOT_PRODUCT
    use Utilities,          only : dot_product=>dot_product2
#endif
#ifdef _OPENMP
    use Select_Parameters,  only : omp_limit, omp_chunk
    use omp_lib,            only : omp_get_num_procs, omp_get_max_threads,   &
                                   omp_in_parallel, omp_get_nested
#endif
#ifdef _BLAS
    use BLAS_interfaces,    only : axpy
#endif
!
! USED MODULES IN SPECIFIC PROCEDURES
! ===================================
!
!   use Logical_Constants
!   use Reals_Constants
!   use Num_Constants
!   use Char_Constants
!   use Utilities
!   use Sort_Procedures
!   use Random
!   use FFT_Procedures
!   use Eig_Procedures
!   use SVD_Procedures
!   use LLSQ_Procedures
!   use Stat_Procedures
!   use Prob_Procedures
!
! STRONG TYPING IMPOSED 
! =====================
!    
    implicit none
!
! PUBLIC ENTITIES 
! ===============
!    
! ALL SUBROUTINES, FUNCTIONS, VARIABLES AND PARAMETERS ARE PRIVATE BY DEFAULT.
!
    private
    public ::   comp_cor,      comp_cor_miss,      comp_cor_miss2,                &
                permute_cor,   phase_scramble_cor, bootstrap_cor,                 &
                comp_cormat,   comp_cormat_miss,                                  &
                comp_eof,      comp_eof2,          comp_eof3,      comp_pc_eof,   &
                comp_eof_miss, comp_eof_miss2,     comp_eof_miss3,                &
                comp_mca,      comp_mca2,          comp_pc_mca,                   &
                comp_mca_miss, comp_mca_miss2,                                    &
                comp_pc,       comp_pc_miss,                                      &
                update_cor,    update_cor_miss2
!
! GENERIC INTERFACES FOR ROUTINES WITH OVERLOADED VERSIONS
! ========================================================
!
    interface comp_cor
        module procedure    comp_cor_rv, comp_cor_rm, comp_cor_rt, comp_cor_rm_rm
    end interface
!
    interface comp_cor_miss
        module procedure    comp_cor_miss_rv, comp_cor_miss_rm, comp_cor_miss_rt, comp_cor_miss_rm_rm
    end interface
!
    interface comp_cor_miss2
        module procedure    comp_cor_miss2_rv, comp_cor_miss2_rm, comp_cor_miss2_rt
    end interface
!
    interface update_cor
        module procedure    update_cor_rv, update_cor_rm, update_cor_rt
    end interface
!
    interface update_cor_miss2
        module procedure    update_cor_rv, update_cor_miss2_rm, update_cor_miss2_rt
    end interface
!
    interface permute_cor
        module procedure  permute_cor_rv, permute_cor_rm
    end interface
!
    interface phase_scramble_cor
        module procedure  phase_scramble_cor_rv, phase_scramble_cor_rm
    end interface
!
    interface bootstrap_cor
        module procedure  bootstrap_cor_rv, bootstrap_cor_rm
    end interface
!
    interface comp_cormat
        module procedure    comp_cormat_a, comp_cormat_b
    end interface
!
    interface comp_cormat_miss
        module procedure    comp_cormat_miss_a, comp_cormat_miss_b
    end interface
!
    interface comp_pc
        module procedure    comp_pc_rv, comp_pc_rm
    end interface
!
    interface comp_pc_miss
        module procedure    comp_pc_miss_rv, comp_pc_miss_rm
    end interface
!
!
! =========================================================================================
!
                                  contains
!                                 ========
!
! =========================================================================================
!                               BIVARIATE STATISTICS
! =========================================================================================
!
!
    subroutine comp_cor_rv( x, y, first, last, xstat, ystat, xycor, xyn, z,   &
                            prob, ndf_max, cortest, cov )
!
! Purpose
! _______
!
!   COMP_COR computes estimates of mean, variance and correlation
!   coefficient from two data vectors XX and YY.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:)
!            On entry, input subvector containing size(X) observations
!            from the vector of data XX for which basic univariate and bivariate
!            statistics are desired. If all the data are available at once, X
!            can be the full data vector XX.
!
!   Y        (INPUT) real(stnd), dimension(:)
!            On entry, input subvector containing size(X) observations
!            from another vector of data YY for which correlation coefficient
!            with XX is desired. If all the data are available at once, YY
!            can be the full data vector.
!
!            The size of Y must verify: size(Y) = size(X) .
!
!   FIRST    (INPUT) logical(lgl)
!            On entry, if:
!
!            - FIRST = true  the current subvector is the first subvector 
!              of the data vector XX (or YY).
!            - FIRST = false the current subvector is not the first subvector
!              of the data vector XX (or YY).
!
!   LAST     (INPUT) logical(lgl)
!            On entry, if:
!
!            - LAST = true  the current subvector is the last subvector 
!              of the data vector XX (or YY).
!            - LAST = false the current subvector is not the last subvector
!              of the data vector XX (or YY).
!
!   XSTAT    (INPUT/OUTPUT) real(stnd), dimension(2)
!            On entry, after the first call to COMP_COR (e.g. when FIRST=true),
!            XSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_COR. XSTAT should not be changed between calls to
!            COMP_COR.
!
!            On exit, when LAST=true, XSTAT contains the following statistics:
!
!            - XSTAT(1) contains the mean value of the data vector XX.
!            - XSTAT(2) contains the variance of the data vector XX.
!
!            The size of XSTAT must verify: size(XSTAT) = 2.
!
!   YSTAT    (INPUT/OUTPUT) real(stnd), dimension(2)
!            On entry, after the first call to COMP_COR (e.g. when FIRST=true),
!            YSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_COR. YSTAT should not be changed between calls to
!            COMP_COR.
!
!            On exit, when LAST=true, YSTAT contains the following statistics:
!
!            - YSTAT(1) contains the mean value of the data vector YY.
!            - YSTAT(2) contains the variance of the data vector YY.
!
!            The size of YSTAT must verify: size(YSTAT) = 2.
!
!   XYCOR    (INPUT/OUTPUT) real(stnd)
!            On entry, after the first call to COMP_COR (e.g. when FIRST=true),
!            XYCOR is used as workspace to accumulate quantities on previous calls
!            to COMP_COR. XYCOR should not be changed between calls to
!            COMP_COR.
!
!            On exit, when LAST=true, XYCOR contains the correlation coefficient
!            between XX(:) and YY(:).
!
!   XYN      (INPUT/OUTPUT) real(stnd)
!            On entry, after the first call to COMP_COR (e.g. when FIRST=true),
!            XYN contains count of observations from previous calls
!            to COMP_COR. XYN should not be changed between calls to
!            COMP_COR.
!
!            On exit, XYN contains the number of observations in the data vectors XX
!            and YY.
!
!   Z        (OUTPUT, OPTIONAL) real(stnd)
!            On exit, when LAST=true, Z contains the Fisher's Z transformation of
!            XYCOR.
!
!            Z needs to be specified only on the last call to COMP_COR (e.g. when LAST=true).
!          
!   PROB     (OUTPUT, OPTIONAL) real(stnd)
!            On exit, when LAST=true, PROB gives the probability that the random sample
!            of XYN observation pairs YY(:) and XX(:) came from a bivariate normal
!            population with a correlation coefficient equal to zero.
!
!            PROB needs to be specified only on the last call to COMP_COR (e.g. when LAST=true).
!          
!   NDF_MAX  (INPUT, OPTIONAL) integer(i4b)
!            On entry, when argument PROB is present, NDF_MAX is used as follows:
!          
!            - If XYN-2 is lower or equal to NDF_MAX, the t_density is integrated
!              for computing PROB.
!            - If XYN-2 is greater than NDF_MAX, an asymptotic series is used
!              for computing PROB.
!          
!            The default is 20.
!
!            NDF_MAX needs to be specified only on the last call to COMP_COR (e.g. when LAST=true).
!          
!   CORTEST  (INPUT/OUTPUT, OPTIONAL) real(stnd)
!            On entry, a probability. CORTEST is the sum of the areas (equal) 
!            in both tails of the Student's t distribution with XYN-2 degrees
!            of freedom.
!
!            CORTEST must verify:   0. < P < 1.
!
!            On exit, the two-tail CORTEST quantile of the sample correlation coefficient,
!            that is a value R such that the probability of the absolute value of a sample
!            correlation coefficient computed from XYN observation pairs being
!            greater than R is CORTEST under the null hypothesis of no correlation in the
!            parent normal population.
!
!            CORTEST needs to be specified only on the last call to COMP_COR (e.g. when LAST=true).
!          
!   COV      (INPUT, OPTIONAL) logical(lgl)
!            On entry, if COV=true, covariance coefficient between the data vectors XX and
!            YY is computed instead of correlation coefficient. If COV=true, Z and PROB are
!            set to Nan code.
!
!            COV needs to be specified only on the last call to COMP_COR (e.g. when LAST=true).
!
!
! Further Details
! _______________
!
!   The subroutine computes the basic univariate statistics and the correlation
!   coefficient with only one pass through the data.
!
!   If fewer than two valid observations were present, the pertinent statistics are set
!   to Nan code.
!
!   For more details on correlation and regression analysis, see:
!
!   (1) von Storch, H., and Zwiers, F.W., 2002:
!            Statistical Analysis in Climate Research.
!            Cambridge, UK, Chapter 8, 484 pp., ISBN:9780521012300
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, half
    use Num_Constants,     only : nan, lamch
    use Char_Constants,    only : stat_error2
    use Utilities,         only : assert_eq, assert, merror
    use Prob_Procedures,   only : probt, pinvstudent
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),    dimension(:) :: x, y
    real(stnd), intent(inout), dimension(:) :: xstat, ystat
    real(stnd), intent(inout)               :: xycor, xyn
    real(stnd), intent(out),       optional :: prob, z
    real(stnd), intent(inout),     optional :: cortest
!
    integer(i4b), intent(in), optional :: ndf_max
!
    logical(lgl), intent(in)            :: first, last
    logical(lgl), intent(in),  optional :: cov
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: dxi, dyi, a, b, xynm1
!
    integer(i4b) :: i, n
!
    logical(lgl) :: comp_cor
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_cor'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq(  int(size(x),i4b),       &
                    int(size(y),i4b),       &
                    name_proc )
!
    call assert( logical(size(xstat)==2,lgl),    &
                 logical(size(ystat)==2,lgl),    &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBVECTOR IS THE FIRST.
!
!       INITIALIZE xstat, ystat, xycor AND xyn .
!
        xstat(1_i4b:2_i4b) = zero       
        ystat(1_i4b:2_i4b) = zero
        xycor = zero
        xyn   = zero
!
    end if
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS AND CORRECTED SUMS OF SQUARES AND CROSS-PRODUCT, 
!       HANDLING THE DATA ONCE.
!
        xynm1 = xyn
        xyn = xynm1 + one
        a = one/xyn
        b = xynm1*a
!
        dxi = x(i) - xstat(1_i4b)
        dyi = y(i) - ystat(1_i4b)
!
        xstat(1_i4b) = xstat(1_i4b) + a*dxi
        xstat(2_i4b) = xstat(2_i4b) + b*dxi*dxi
!
        ystat(1_i4b) = ystat(1_i4b) + a*dyi
        ystat(2_i4b) = ystat(2_i4b) + b*dyi*dyi
!
        xycor = xycor + b*dxi*dyi
!
    end do
!
    if ( last ) then
!
        comp_cor = true
        if ( present(cov) ) then
            comp_cor = .not.cov
        end if
!
        b = nan()
!
!       THE CURRENT SUBVECTOR IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
        if ( xyn>one ) then
!
            a = one/xyn
            xynm1 = lamch('s')
!
!           COMPUTE VARIANCE OF XX.
!
            dxi = xstat(2_i4b)
            xstat(2_i4b) = dxi*a
!
!           COMPUTE VARIANCE OF YY.
!
            dyi = ystat(2_i4b)
            ystat(2_i4b) = dyi*a
!
            if ( comp_cor ) then
!
!               COMPUTE CORRELATION COEFFICIENT.
!
                dxi = dxi*dyi
                if ( dxi>xynm1 ) then
                    xycor = xycor/sqrt( dxi )
                else
                    xycor = b
                end if
!
!               OUTPUT FISHER'S z TRANSFORMATION, IF REQUIRED.
!
                if ( present(z) ) then
!
                    if ( dxi>xynm1 ) then
                        z = half*log(( (one+xycor) + xynm1 )/( (one-xycor) + xynm1 ))
                    else
                        z = b
                    end if
!
                end if
!
!               OUTPUT SIGNIFICANCE PROBABILITY IF REQUIRED.
!
                if ( present(prob) ) then
!
                    if ( xyn>two .and. dxi>xynm1 ) then
!
                        dxi = abs( xycor )
!
                        if ( dxi/=one ) then
                            a = xyn - two
                            i = int( a, i4b )
                            dyi = sqrt( a )*dxi/sqrt( one - dxi*dxi )
                            prob = two*probt( dyi, i, true, ndf_max=ndf_max )
                        else
                            prob = zero
                        end if
!
                    else
!
                        prob = b
!
                    end if
!
                end if
!
            else
!
!               COMPUTE COVARIANCE COEFFICIENT BETWEEN XX AND YY.
!
                xycor = xycor*a
!
                if ( present(prob) ) prob = b
                if ( present(z) )    z = b
!
            end if
!
        else
!
!           FEWER THAN TWO VALID OBSERVATIONS, SET THE PERTINENT STATISTICS TO NaN CODE.
!
            if ( xyn<one ) then
                xstat(1_i4b:2_i4b) = b
                ystat(1_i4b:2_i4b) = b
            end if
!
            xycor = b
!
            if ( present(prob) ) prob = b
!
            if ( present(z) ) z = b
!
        end if
!
        if ( present(cortest) ) then
!
!           COMPUTE TWO-TAIL QUANTILE OF STUDENT-T DISTRIBUTION.
!
            if ( zero>=cortest .or. cortest>=one ) then
                call merror( name_proc//stat_error2 )
            else if ( xyn>two ) then
                a = xyn - two
                dyi = pinvstudent( cortest, a )
                dyi = dyi*dyi
                cortest = sqrt( dyi/( dyi + a ) )
            else
                cortest = b
            end if
!
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_cor_rv
! _____________________________
!
    end subroutine comp_cor_rv
!
! =========================================================================================
!
    subroutine comp_cor_rm( x, y, first, last, xstat, ystat, xycor, xyn, dimvar, z,  &
                            prob, ndf_max, cortest, cov )
!
! Purpose
! _______
!
!   COMP_COR computes estimates of means, variances and correlation
!   coefficients (with a data vector YY) from a data matrix XX.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:,:)
!            On entry, input submatrix containing size(X,3-DIMVAR)
!            observations on size(X,DIMVAR) variables from the matrix of data XX
!            for which basic univariate  and bivariate statistics are desired. By
!            default, DIMVAR is equal to 1. See description of optional DIMVAR
!            argument for details.
!            If all the data are available at once, X can be the full 
!            data matrix XX.
!
!   Y        (INPUT) real(stnd), dimension(:)
!            On entry, input subvector containing size(X,3-DIMVAR) observations
!            from a vector of data YY for which correlation coefficient
!            with XX is desired. If all the data are available at once, YY
!            can be the full data vector.
!
!            The size of Y must verify: size(Y) = SIZE(X,3-DIMVAR) .
!
!   FIRST    (INPUT) logical(lgl)
!            On entry, if:
!
!            - FIRST = true  the current submatrix is the first submatrix 
!              of the data matrix XX.
!            - FIRST = false the current submatrix is not the first submatrix
!              of the data matrix XX.
!
!   LAST     (INPUT) logical(lgl)
!            On entry, if:
!
!            - LAST = true  the current submatrix is the last submatrix 
!              of the data matrix XX.
!            - LAST = false the current submatrix is not the last submatrix
!              of the data matrix XX.
!
!   XSTAT    (INPUT/OUTPUT) real(stnd), dimension(:,2)
!            On entry, after the first call to COMP_COR (e.g. when FIRST=true),
!            XSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_COR. XSTAT should not be changed between calls to
!            COMP_COR.
!
!            On exit, when LAST=true, XSTAT contains the following statistics
!            on all variables:
!
!            - XSTAT(:,1) contains the mean values of the data matrix XX.
!            - XSTAT(:,2) contains the variances of the data matrix XX.
!
!            The shape of XSTAT must verify:
!
!            - size(XSTAT,1) = size(X,DIMVAR),
!            - size(XSTAT,2) = 2 .
!
!   YSTAT    (INPUT/OUTPUT) real(stnd), dimension(2)
!            On entry, after the first call to COMP_COR (e.g. when FIRST=true),
!            YSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_COR. YSTAT should not be changed between calls to
!            COMP_COR.
!
!            On exit, when LAST=true, YSTAT contains the following statistics:
!
!            - YSTAT(1) contains the mean value of the data vector YY.
!            - YSTAT(2) contains the variance of the data vector YY.
!
!            The size of YSTAT must verify: size(YSTAT) = 2.
!
!   XYCOR    (INPUT/OUTPUT) real(stnd), dimension(:)
!            On entry, after the first call to COMP_COR (e.g. when FIRST=true),
!            XYCOR is used as workspace to accumulate quantities on previous calls
!            to COMP_COR. XYCOR should not be changed between calls to
!            COMP_COR.
!
!            On exit, when LAST=true, XYCOR(i) contains the correlation coefficient
!            between XX(i,:) ( XX(:,i) if DIMVAR=2 ) and YY(:).
!
!            The size of XYCOR must verify: size(XYCOR) = size(X,DIMVAR).
!
!   XYN      (INPUT/OUTPUT) real(stnd)
!            On entry, after the first call to COMP_COR (e.g. when FIRST=true),
!            XYN contains count of observations from previous calls
!            to COMP_COR. XYN should not be changed between calls to
!            COMP_COR.
!
!            On exit, XYN contains the number of observations in the data matrix XX
!            and the data vector YY.
!
!   DIMVAR   (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVAR is present, DIMVAR is used as follows:
!
!            - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!              on size(X,1) variables.
!            - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!              on size(X,2) variables.
!
!            The default is DIMVAR = 1.
!
!   Z        (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            On exit, when LAST=true, Z contains the Fisher's Z transformation of
!            XYCOR.
!
!            Z needs to be specified only on the last call to COMP_COR (e.g. when LAST=true).
!          
!            The size of Z must verify: size(Z) = size(X,DIMVAR).
!
!   PROB     (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            On exit, when LAST=true, PROB(i) gives the probability that the random sample
!            of XYN observation pairs YY(:) and XX(i,:) ( XX(:,i) if DIMVAR=2 ) came from
!            a bivariate normal population with a correlation coefficient equal to zero.
!
!            PROB needs to be specified only on the last call to COMP_COR (e.g. when LAST=true).
!          
!            The size of PROB must verify: size(PROB) = size(X,DIMVAR).
!
!   NDF_MAX  (INPUT, OPTIONAL) integer(i4b)
!            On entry, when argument PROB is present, NDF_MAX is used as follows:
!
!            - If XYN-2 is lower or equal to NDF_MAX, the t_density is integrated
!              for computing PROB(i).
!            - If XYN-2 is greater than NDF_MAX, an asymptotic series is used
!              for computing PROB(i).
!
!            The default is 20.
!
!            NDF_MAX needs to be specified only on the last call to COMP_COR (e.g. when LAST=true).
!
!   CORTEST  (INPUT/OUTPUT, OPTIONAL) real(stnd)
!            On entry, a probability. CORTEST is the sum of the areas (equal) 
!            in both tails of the Student's t distribution with XYN-2 degrees
!            of freedom.
!
!            CORTEST must verify:   0. < P < 1.
!
!            On exit, the two-tail CORTEST quantile of the sample correlation coefficient,
!            that is a value R such that the probability of the absolute value of a sample
!            correlation coefficient computed from XYN observation pairs being
!            greater than R is CORTEST under the null hypothesis of no correlation in the
!            parent normal population.
!
!            CORTEST needs to be specified only on the last call to COMP_COR (e.g. when LAST=true).
!          
!   COV      (INPUT, OPTIONAL) logical(lgl)
!            On entry, if COV=true, covariance coefficients between the data matrix XX and data
!            vector YY are computed instead of correlation coefficients. If COV=true, Z and PROB
!            are set to Nan code.
!            COV needs to be specified only on the last call to COMP_COR (e.g. when LAST=true).
!
!
! Further Details
! _______________
!
!   The subroutine computes the basic univariate statistics and the correlation
!   coefficients with only one pass through the data.
!
!   If fewer than two valid observations were present, the pertinent statistics are set
!   to Nan code.
!
!   For more details on correlation and regression analysis, see:
!
!   (1) von Storch, H., and Zwiers, F.W., 2002:
!            Statistical Analysis in Climate Research.
!            Cambridge, UK, Chapter 8, 484 pp., ISBN:9780521012300
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, half
    use Num_Constants,     only : nan, lamch
    use Char_Constants,    only : stat_error1, stat_error2, stat_error5
    use Utilities,         only : assert_eq, assert, merror
    use Prob_Procedures,   only : probt, pinvstudent
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),    dimension(:,:)            :: x
    real(stnd), intent(in),    dimension(:)              :: y
    real(stnd), intent(inout), dimension(:,:)            :: xstat
    real(stnd), intent(inout), dimension(:)              :: ystat, xycor
    real(stnd), intent(inout)                            :: xyn
    real(stnd), intent(out),   dimension(:),    optional :: prob, z
    real(stnd), intent(inout),                  optional :: cortest
!
    integer(i4b), intent(in), optional :: dimvar, ndf_max
!
    logical(lgl), intent(in)            :: first, last
    logical(lgl), intent(in),  optional :: cov
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                         :: a, b, xynm1, dyi, tmp
    real(stnd), dimension(size(xycor)) :: dxi
!
    integer(i4b) :: i, n, m
!
    logical(lgl) :: dim_is_one, comp_cor
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_cor'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
    n = assert_eq(  int(size(x,3-int(i)),i4b),  &
                    int(size(y),i4b),           &
                    name_proc )
!
    m = assert_eq(  int(size(x,int(i)),i4b),      &
                    int(size(xstat,1),i4b),       &
                    int(size(xycor),i4b),         &
                    name_proc )
!
    if ( m<=0_i4b ) call merror( name_proc//stat_error1 )
!
    call assert( logical(size(xstat,2)==2,lgl),    &
                 logical(size(ystat)==2,lgl),      &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBMATRIX IS THE FIRST.
!
!       INITIALIZE xstat, ystat, xycor AND xyn .
!
        xstat(:m,1_i4b:2_i4b) = zero       
        ystat(1_i4b:2_i4b)    = zero
        xycor(:m) = zero
        xyn       = zero
!
    end if
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS AND CORRECTED SUMS OF SQUARES AND CROSS-PRODUCTS, 
!       HANDLING THE DATA ONCE.
!
        xynm1 = xyn
        xyn = xynm1 + one
        a = one/xyn
        b = xynm1*a
!
        if ( dim_is_one ) then
            dxi(:m) = x(:m,i) - xstat(:m,1_i4b)
        else
            dxi(:m) = x(i,:m) - xstat(:m,1_i4b)
        end if
        dyi = y(i) - ystat(1_i4b)
!
        ystat(1_i4b) = ystat(1_i4b) + a*dyi
        ystat(2_i4b) = ystat(2_i4b) + b*( dyi*dyi )
!
        tmp = b*dyi
!
#ifdef _BLAS
        call axpy( m, a, dxi(:m), 1_i4b, xstat(:m,1_i4b), 1_i4b )
        call axpy( m, tmp, dxi(:m), 1_i4b, xycor(:m), 1_i4b )
!
        dxi(:m) = dxi(:m)*dxi(:m)
        call axpy( m, b, dxi(:m), 1_i4b, xstat(:m,2_i4b), 1_i4b )
#else
        xstat(:m,1_i4b) = xstat(:m,1_i4b) + a*dxi(:m)
        xstat(:m,2_i4b) = xstat(:m,2_i4b) + b*( dxi(:m)*dxi(:m) )
!
        xycor(:m) = xycor(:m) + tmp*dxi(:m)
#endif
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBMATRIX IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
!
!       TEST OPTIONAL ARGUMENTS.
!
        if ( present(prob) ) call assert( logical(size(prob)==m,lgl), name_proc )
!
        if ( present(z) ) call assert( logical(size(z)==m,lgl), name_proc )
!
        comp_cor = true
        if ( present(cov) ) then
            comp_cor = .not.cov
        end if
!
        b = nan()
!
        if ( xyn>one ) then
!
            a = one/xyn
            xynm1 = lamch('s')
!
!           COMPUTE VARIANCES OF XX.
!
            dxi(:m) = xstat(:m,2_i4b)
            xstat(:m,2_i4b) = dxi(:m)*a
!
!           COMPUTE VARIANCE OF YY.
!
            dyi = ystat(2_i4b)
            ystat(2_i4b) = dyi*a
!
            if ( comp_cor ) then
!
!               COMPUTE CORRELATION COEFFICIENTS BETWEEN XX AND YY.
!
                dxi(:m) = dxi(:m)*dyi
                where ( dxi(:m)>xynm1 )
                    xycor(:m) = xycor(:m)/sqrt( dxi(:m) )
                elsewhere
                    xycor(:m) = b
                end where
!
!               OUTPUT FISHER'S z TRANSFORMATION, IF REQUIRED.
!
                if ( present(z) ) then
!
                    where ( dxi(:m)>xynm1 )
                        z(:m) = half*log(( (one+xycor(:m)) + xynm1 )/( (one-xycor(:m)) + xynm1 ))
                    elsewhere
                        z(:m) = b
                    end where
!
                end if
!
!               OUTPUT SIGNIFICANCE PROBABILITY, IF REQUIRED.
!
                if ( present(prob) ) then
!
                    if ( xyn>two ) then
!
                        a = xyn - two
                        i = int( a, i4b )
                        a = sqrt( a )
!
                        where ( dxi(:m)>xynm1 .and. abs(xycor(:m))/=one )
                            dxi(:m) = a*abs(xycor(:m))/sqrt( one - xycor(:m)**2 )
                        elsewhere
                           dxi(:m) = -one
                        end where
!
                        prob(:m) = two*probt( dxi(:m), i, true, ndf_max=ndf_max )
                        where ( dxi(:m)==-one )  prob(:m) = merge( zero, b, abs(xycor(:m))==one )
!
                    else
!
                        prob(:m) = b
!
                    end if
!
                end if
!
            else
!
!               COMPUTE COVARIANCE COEFFICIENTS BETWEEN XX AND YY.
!
                xycor(:m) = xycor(:m)*a
!
                if ( present(prob) ) prob(:m) = b
                if ( present(z) )    z(:m) = b
!
            end if
!
        else
!
!           FEWER THAN TWO VALID OBSERVATIONS, SET THE PERTINENT STATISTICS TO NaN CODE.
!
            if ( xyn<one ) then
                xstat(:m,1_i4b:2_i4b) = b
                ystat(1_i4b:2_i4b) = b
            end if
!
            xycor(:m) = b
!
            if ( present(prob) ) prob(:m) = b
            if ( present(z) ) z(:m) = b
!
        end if
!
        if ( present(cortest) ) then
!
!           COMPUTE TWO-TAIL QUANTILE OF STUDENT-T DISTRIBUTION.
!
            if ( zero>=cortest .or. cortest>=one ) then
                call merror( name_proc//stat_error2 )
            else if ( xyn>two ) then
                a = xyn - two
                dyi = pinvstudent( cortest, a )
                dyi = dyi*dyi
                cortest = sqrt( dyi/( dyi + a ) )
            else
                cortest = b
            end if
!
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_cor_rm
! _____________________________
!
    end subroutine comp_cor_rm
!
! =========================================================================================
!
    subroutine comp_cor_rt( x, y, first, last, xstat, ystat, xycor, xyn, z,   &
                            prob, ndf_max, cortest, cov )
!
! Purpose
! _______
!
!   COMP_COR computes estimates of means, variances and correlation
!   coefficients (with a data vector YY) from a data tridimensional array XX.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:,:,:)
!            On entry, input subarray containing size(X,3) observations on size(X,1)
!            by size(X,2) variables from the tridimensional array of data XX for which 
!            basic univariate and bivariate statistics are desired.
!            If all the data are available at once, X can be the full data array XX.
!
!   Y        (INPUT) real(stnd), dimension(:)
!            On entry, input subvector containing size(X,3) observations
!            from a vector of data YY for which correlation coefficients
!            with XX is desired. If all the data are available at once, YY
!            can be the full data vector.
!
!            The size of Y must verify: size(Y) = SIZE(X,3) .
!
!   FIRST    (INPUT) logical(lgl)
!            On entry, if:
!
!            - FIRST = true  the current subarray is the first subarray 
!              of the data array XX.
!            - FIRST = false the current subarray is not the first subarray
!              of the data array XX.
!
!   LAST     (INPUT) logical(lgl)
!            On entry, if:
!
!            - LAST = true  the current subarray is the last subarray 
!              of the data array XX.
!            - LAST = false the current subarray is not the last subarray
!              of the data array XX.
!
!   XSTAT    (INPUT/OUTPUT) real(stnd), dimension(:,:,2)
!            On entry, after the first call to COMP_COR (e.g. when FIRST=true),
!            XSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_COR. XSTAT should not be changed between calls to
!            COMP_COR.
!
!            On exit, when LAST=true, XSTAT contains the following statistics
!            on all variables:
!
!            - XSTAT(:,:,1) contains the mean values of the data array XX.
!            - XSTAT(:,:,2) contains the variances of the data array XX.
!
!            The shape of XSTAT must verify:
!
!            - size(XSTAT,1) = size(X,1) ,
!            - size(XSTAT,2) = size(X,2) ,
!            - size(XSTAT,3) = 2 .
!
!   YSTAT    (INPUT/OUTPUT) real(stnd), dimension(2)
!            On entry, after the first call to COMP_COR (e.g. when FIRST=true),
!            YSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_COR. YSTAT should not be changed between calls to
!            COMP_COR.
!
!            On exit, when LAST=true, YSTAT contains the following statistics:
!
!            - YSTAT(1) contains the mean value of the data vector YY.
!            - YSTAT(2) contains the variance of the data vector YY.
!
!            The size of YSTAT must verify: size(YSTAT) = 2.
!
!   XYCOR    (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, after the first call to COMP_COR (e.g. when FIRST=true),
!            XYCOR is used as workspace to accumulate quantities on previous calls
!            to COMP_COR. XYCOR should not be changed between calls to
!            COMP_COR.
!
!            On exit, when LAST=true, XYCOR(i,j) contains the correlation coefficient
!            between XX(i,j,:) and YY(:).
!
!            The shape of XYCOR must verify:
!
!            - size(XYCOR,1) = size(X,1) ,
!            - size(XYCOR,2) = size(X,2) .
!
!   XYN      (INPUT/OUTPUT) real(stnd)
!            On entry, after the first call to COMP_COR (e.g. when FIRST=true),
!            XYN contains count of observations from previous calls
!            to COMP_COR. XYN should not be changed between calls to
!            COMP_COR.
!
!            On exit, XYN contains the number of observations in the data array XX
!            and the data vector YY.
!
!   Z        (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!            On exit, when LAST=true, Z contains the Fisher's Z transformation of
!            XYCOR.
!
!            Z needs to be specified only on the last call to COMP_COR (e.g. when LAST=true).
!          
!            The shape of Z must verify:
!          
!            - size(Z,1) = size(X,1) ,
!            - size(Z,2) = size(X,2) .
!
!   PROB     (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!            On exit, when LAST=true, PROB(i,j) gives the probability that the random sample
!            of XYN observation pairs YY(:) and XX(i,j,:) came from a bivariate normal
!            population with a correlation coefficient equal to zero.
!
!            PROB needs to be specified only on the last call to COMP_COR (e.g. when LAST=true).
!          
!            The shape of PROB must verify:
!          
!            - size(PROB,1) = size(X,1) ,
!            - size(PROB,2) = size(X,2) .
!
!   NDF_MAX  (INPUT, OPTIONAL) integer(i4b)
!            On entry, when argument PROB is present, NDF_MAX is used as follows:
!
!            - If XYN-2 is lower or equal to NDF_MAX, the t_density is integrated
!              for computing PROB(i,j).
!            - If XYN-2 is greater than NDF_MAX, an asymptotic series is used
!              for computing PROB(i,j).
!
!            The default is 20.
!
!            NDF_MAX needs to be specified only on the last call to COMP_COR (e.g. when LAST=true).
!
!   CORTEST  (INPUT/OUTPUT, OPTIONAL) real(stnd)
!            On entry, a probability. CORTEST is the sum of the areas (equal) 
!            in both tails of the Student's t distribution with XYN-2 degrees
!            of freedom.
!
!            CORTEST must verify:   0. < P < 1.
!
!            On exit, the two-tail CORTEST quantile of the sample correlation coefficient,
!            that is a value R such that the probability of the absolute value of a sample
!            correlation coefficient computed from XYN observation pairs being
!            greater than R is CORTEST under the null hypothesis of no correlation in the
!            parent normal population.
!
!            CORTEST needs to be specified only on the last call to COMP_COR (e.g. when LAST=true).
!          
!   COV      (INPUT, OPTIONAL) logical(lgl)
!            On entry, if COV=true, covariance coefficients between the data matrices XX and
!            YY are computed instead of correlation coefficients. If COV=true, Z and PROB are
!            set to Nan code.
!            COV needs to be specified only on the last call to COMP_COR (e.g. when LAST=true).
!          
!
! Further Details
! _______________
!
!   The subroutine computes the basic univariate statistics and the correlation
!   coefficients with only one pass through the data.
!
!   If fewer than two valid observations were present, the pertinent statistics are set
!   to Nan code.
!
!   For more details on correlation and regression analysis, see:
!
!   (1) von Storch, H., and Zwiers, F.W., 2002:
!            Statistical Analysis in Climate Research.
!            Cambridge, UK, Chapter 8, 484 pp., ISBN:9780521012300
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, half
    use Num_Constants,     only : nan, lamch
    use Char_Constants,    only : stat_error1, stat_error2
    use Utilities,         only : assert_eq, assert, merror
    use Prob_Procedures,   only : probt, pinvstudent
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),    dimension(:,:,:)            :: x
    real(stnd), intent(in),    dimension(:)                :: y
    real(stnd), intent(inout), dimension(:,:,:)            :: xstat
    real(stnd), intent(inout), dimension(:)                :: ystat
    real(stnd), intent(inout), dimension(:,:)              :: xycor
    real(stnd), intent(inout)                              :: xyn
    real(stnd), intent(out),   dimension(:,:),    optional :: prob, z
    real(stnd), intent(inout),                    optional :: cortest
!
    integer(i4b), intent(in), optional :: ndf_max
!
    logical(lgl), intent(in)            :: first, last
    logical(lgl), intent(in),  optional :: cov
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                       :: a, b, xynm1, dyi, tmp
    real(stnd), dimension(size(x,1)) :: dxi
!
    integer(i4b) :: i, j, n, m, p
!
    logical(lgl) :: comp_cor
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_cor'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq(  int(size(x,3),i4b),     &
                    int(size(y),i4b),       &
                    name_proc )
!
    m = assert_eq(  int(size(x,1),i4b),           &
                    int(size(xstat,1),i4b),       &
                    int(size(xycor,1),i4b),       &
                    name_proc )
!
    p = assert_eq(  int(size(x,2),i4b),           &
                    int(size(xstat,2),i4b),       &
                    int(size(xycor,2),i4b),       &
                    name_proc )
!
    if ( min(m,p)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    call assert( logical(size(xstat,3)==2,lgl),    &
                 logical(size(ystat)==2,lgl),      &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBARRAY IS THE FIRST.
!
!       INITIALIZE xstat, ystat, xycor AND xyn .
!
        xstat(:m,:p,1_i4b:2_i4b) = zero       
        ystat(1_i4b:2_i4b)       = zero
        xycor(:m,:p) = zero
        xyn          = zero
!
    end if
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS AND CORRECTED SUMS OF SQUARES AND CROSS-PRODUCTS, 
!       HANDLING THE DATA ONCE.
!
        xynm1 = xyn
        xyn = xynm1 + one
        a = one/xyn
        b = xynm1*a
!
        dyi = y(i) - ystat(1_i4b)
!
        ystat(1_i4b) = ystat(1_i4b) + a*dyi
        ystat(2_i4b) = ystat(2_i4b) + b*( dyi*dyi )
!
        tmp = b*dyi
!
        do j = 1_i4b, p
!
            dxi(:m) = x(:m,j,i) - xstat(:m,j,1_i4b)
!
#ifdef _BLAS
            call axpy( m, a, dxi(:m), 1_i4b, xstat(:m,j,1_i4b), 1_i4b )
            call axpy( m, tmp, dxi(:m), 1_i4b, xycor(:m,j), 1_i4b )
!
            dxi(:m) = dxi(:m)*dxi(:m)
            call axpy( m, b, dxi(:m), 1_i4b, xstat(:m,j,2_i4b), 1_i4b )
#else
            xstat(:m,j,1_i4b) = xstat(:m,j,1_i4b) + a*dxi(:m)
            xstat(:m,j,2_i4b) = xstat(:m,j,2_i4b) + b*( dxi(:m)*dxi(:m) )
!
            xycor(:m,j) = xycor(:m,j) + tmp*dxi(:m)
#endif
!
        end do
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBARRAY IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
!
!       TEST OPTIONAL ARGUMENTS.
!
        if ( present(prob) ) call assert( logical(size(prob,1)==m,lgl),     &
                                          logical(size(prob,2)==p,lgl),     &
                                          name_proc )
!
        if ( present(z) ) call assert( logical(size(z,1)==m,lgl),     &
                                       logical(size(z,2)==p,lgl),     &
                                       name_proc )
!
        comp_cor = true
        if ( present(cov) ) then
            comp_cor = .not.cov
        end if
!
        b = nan()
!
        if ( xyn>one ) then
!
            xynm1 = lamch('s')
!
!           COMPUTE VARIANCE OF YY.
!
            dyi = ystat(2_i4b)
            ystat(2_i4b) = dyi/xyn
!
            if ( comp_cor ) then
!
#ifdef _OPENMP
                i = omp_get_num_procs()
                j = omp_get_max_threads()
                ompparallel = omp_in_parallel()
                ompnested   = omp_get_nested() .and. ompparallel
                test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                              (m*p)>=omp_limit                    .and.      &
                              i>1_i4b                             .and.      &
                              j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)        &
!$OMP         ,PRIVATE(i,j,dxi,a)
!
!               COMPUTE CORRELATION COEFFICIENTS BETWEEN XX AND YY.
!
!$OMP DO SCHEDULE(STATIC) 
!
                do j = 1_i4b, p
!
                    dxi(:m) = xstat(:m,j,2_i4b)*dyi
!
                    where ( dxi(:m)>xynm1 )
                         xycor(:m,j) = xycor(:m,j)/sqrt( dxi(:m) )
                    elsewhere
                        xycor(:m,j) = b
                    end where
!
                end do
!
!$OMP END DO
!
!               OUTPUT FISHER'S z TRANSFORMATION, IF REQUIRED.
!
                if ( present(z) ) then
!
!$OMP DO SCHEDULE(STATIC) 
!
                    do j = 1_i4b, p
!
                        dxi(:m) = xstat(:m,j,2_i4b)*dyi
!
                        where ( dxi(:m)>xynm1 )
                            z(:m,j) = half*log(( (one+xycor(:m,j)) + xynm1 )/( (one-xycor(:m,j)) + xynm1 ))
                        elsewhere
                            z(:m,j) = b
                        end where
!
                    end do
!
!$OMP END DO
!
                end if
!
!               OUTPUT SIGNIFICANCE PROBABILITY IF REQUIRED.
!
                if ( present(prob) ) then
!
                    if ( xyn>two ) then
!
                        a = xyn - two
                        i = int( a, i4b )
                        a = sqrt( a )
!
!$OMP DO SCHEDULE(STATIC) 
!
                        do j = 1_i4b, p
!
                            dxi(:m) = xstat(:m,j,2_i4b)*dyi
!
                            where ( dxi(:m)>xynm1 .and. abs(xycor(:m,j))/=one )
                                dxi(:m) = a*abs(xycor(:m,j))/sqrt( one - xycor(:m,j)**2 )
                            elsewhere
                                dxi(:m) = -one
                            end where
!
                            prob(:m,j) = two*probt( dxi(:m), i, true, ndf_max=ndf_max )
!
                            where ( dxi(:m)==-one )  &
                            prob(:m,j) = merge( zero, b, abs(xycor(:m,j))==one )
!
                        end do
!
!$OMP END DO
!
                    else
!
!$OMP SINGLE
!
                        prob(:m,:p) = b
!
!$OMP END SINGLE
!
                    end if
!
                end if
!
!$OMP END PARALLEL
!
            else
!
!               COMPUTE COVARIANCE COEFFICIENTS BETWEEN XX AND YY.
!
                xycor(:m,:p) = xycor(:m,:p)/xyn
!
                if ( present(prob) ) prob(:m,:p) = b
                if ( present(z) )    z(:m,:p)    = b
!
            end if
!
!           COMPUTE VARIANCES OF XX.
!
            xstat(:m,:p,2_i4b) = xstat(:m,:p,2_i4b)/xyn
!
        else
!
!           FEWER THAN TWO VALID OBSERVATIONS, SET THE PERTINENT STATISTICS TO NaN CODE.
!
            if ( xyn<one ) then
                xstat(:m,:p,1_i4b:2_i4b) = b
                ystat(1_i4b:2_i4b) = b
            end if
!
            xycor(:m,:p) = b
!
            if ( present(prob) ) prob(:m,:p) = b
            if ( present(z) )    z(:m,:p) = b
!
        end if
!
        if ( present(cortest) ) then
!
!           COMPUTE TWO-TAIL QUANTILE OF STUDENT-T DISTRIBUTION.
!
            if ( zero>=cortest .or. cortest>=one ) then
                call merror( name_proc//stat_error2 )
            else if ( xyn>two ) then
                a = xyn - two
                dyi = pinvstudent( cortest, a )
                dyi = dyi*dyi
                cortest = sqrt( dyi/( dyi + a ) )
            else
                cortest = b
            end if
!
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_cor_rt
! _____________________________
!
    end subroutine comp_cor_rt
!
! =========================================================================================
!
    subroutine comp_cor_rm_rm( x, y, first, last, xstat, ystat, xycor, xyn,    &
                               dimvar, dimvary, z, prob, ndf_max, cortest, cov )
!
! Purpose
! _______
!
!   COMP_COR computes estimates of means, variances and correlation
!   coefficients between two data matrices YY and XX.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:,:)
!            On entry, input submatrix containing size(X,3-DIMVAR)
!            observations on size(X,DIMVAR) variables from the matrix of data XX
!            for which basic univariate  and bivariate statistics are desired. By
!            default, DIMVAR is equal to 1. See description of optional DIMVAR
!            argument for details.
!            If all the data are available at once, X can be the full 
!            data matrix XX.
!
!   Y        (INPUT) real(stnd), dimension(:,:)
!            On entry, input submatrix containing size(Y,3-DIMVARY)
!            observations on size(Y,DIMVARY) variables from the matrix of data YY
!            for which basic univariate  and bivariate statistics are desired. By
!            default, DIMVARY is equal to 1. See description of optional DIMVARY
!            argument for details.
!            If all the data are available at once, Y can be the full 
!            data matrix YY.
!
!            The shape of Y must verify: size(Y,3-DIMVARY)  = size(X,3-DIMVAR) .
!
!   FIRST    (INPUT) logical(lgl)
!            On entry, if:
!
!            - FIRST = true  the current submatrices are the first submatrices 
!              of the data matrices XX and YY.
!            - FIRST = false the current submatrices are not the first submatrices
!              of the data matrices XX and YY.
!
!   LAST     (INPUT) logical(lgl)
!            On entry, if:
!
!            - LAST = true  the current submatrices are the last submatrices 
!              of the data matrices XX and YY.
!            - LAST = false the current submatrix are not the last submatrices
!              of the data matrices XX and YY.
!
!   XSTAT    (INPUT/OUTPUT) real(stnd), dimension(:,2)
!            On entry, after the first call to COMP_COR (e.g. when FIRST=true),
!            XSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_COR. XSTAT should not be changed between calls to
!            COMP_COR.
!
!            On exit, when LAST=true, XSTAT contains the following statistics
!            on all variables from the XX matrix:
!
!            - XSTAT(:,1) contains the mean values of the data matrix XX.
!            - XSTAT(:,2) contains the variances of the data matrix XX.
!
!            The shape of XSTAT must verify:
!
!            - size(XSTAT,1) = size(X,DIMVAR) ,
!            - size(XSTAT,2) = 2 .
!
!   YSTAT    (INPUT/OUTPUT) real(stnd), dimension(:,2)
!            On entry, after the first call to COMP_COR (e.g. when FIRST=true),
!            YSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_COR. YSTAT should not be changed between calls to
!            COMP_COR.
!
!            On exit, when LAST=true, YSTAT contains the following statistics
!            on all variables from the YY matrix:
!
!            - YSTAT(:,1) contains the mean values of the data matrix YY.
!            - YSTAT(:,2) contains the variances of the data matrix YY.
!
!            The shape of YSTAT must verify:
!
!            - size(YSTAT,1) = size(Y,DIMVARY) ,
!            - size(YSTAT,2) = 2 .
!
!   XYCOR    (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, after the first call to COMP_COR (e.g. when FIRST=true),
!            XYCOR is used as workspace to accumulate quantities on previous calls
!            to COMP_COR. XYCOR should not be changed between calls to
!            COMP_COR.
!
!            On exit, when LAST=true, XYCOR(i,j) contains the correlation coefficient
!            between XX(i,:) and YY(j,:) ( XX(:,i) and YY(:,j) if DIMVAR=2 and DIMVARY=2 ).
!
!            The shape of XYCOR must verify:
!
!            - size(XYCOR,1) = size(X,DIMVAR) ,
!            - size(XYCOR,2) = size(Y,DIMVARY) .
!
!   XYN      (INPUT/OUTPUT) real(stnd)
!            On entry, after the first call to COMP_COR (e.g. when FIRST=true),
!            XYN contains count of observations from previous calls
!            to COMP_COR. XYN should not be changed between calls to
!            COMP_COR.
!
!            On exit, XYN contains the number of observations in the data matrices XX
!            and YY.
!
!   DIMVAR   (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVAR is present, DIMVAR is used as follows:
!
!            - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!              on size(X,1) variables.
!            - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!              on size(X,2) variables, respectively.
!
!            The default is DIMVAR = 1.
!
!   DIMVARY  (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVARY is present, DIMVARY is used as follows:
!
!            - DIMVARY = 1, the input submatrix Y contains size(Y,2) observations
!              on size(Y,1) variables.
!            - DIMVARY = 2, the input submatrix Y contains size(Y,1) observations
!              on size(Y,2) variables, respectively.
!            
!            The default is DIMVARY = 1.
!
!   Z        (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!            On exit, when LAST=true, Z contains the Fisher's Z transformation of
!            XYCOR.
!
!            Z needs to be specified only on the last call to COMP_COR (e.g. when LAST=true).
!          
!            The shape of Z must verify:
!          
!            - size(Z,1) = size(X,DIMVAR) ,
!            - size(Z,2) = size(Y,DIMVARY) .
!
!   PROB     (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!            On exit, when LAST=true, PROB(i,j) gives the probability that the random sample
!            of XYN observation pairs XX(i,:) and YY(j,:) ( XX(:,i) and YY(:,j) if DIMVAR=2 
!            and DIMVARY=2 ) came from a bivariate normal population with a correlation
!            coefficient equal to zero.
!
!            PROB needs to be specified only on the last call to COMP_COR (e.g. when LAST=true).
!          
!            The shape of PROB must verify:
!          
!            - size(PROB,1) = size(X,DIMVAR) ,
!            - size(PROB,2) = size(Y,DIMVARY) .
!
!   NDF_MAX  (INPUT, OPTIONAL) integer(i4b)
!            On entry, when argument PROB is present, NDF_MAX is used as follows:
!
!            - If XYN-2 is lower or equal to NDF_MAX, the t_density is integrated
!              for computing PROB(i,j).
!            - If XYN-2 is greater than NDF_MAX, an asymptotic series is used
!              for computing PROB(i,j).
!
!            The default is 20.
!
!            NDF_MAX needs to be specified only on the last call to COMP_COR (e.g. when LAST=true).
!
!   CORTEST  (INPUT/OUTPUT, OPTIONAL) real(stnd)
!            On entry, a probability. CORTEST is the sum of the areas (equal) 
!            in both tails of the Student's t distribution with XYN-2 degrees
!            of freedom.
!
!            CORTEST must verify:   0. < P < 1.
!
!            On exit, the two-tail CORTEST quantile of the sample correlation coefficient,
!            that is a value R such that the probability of the absolute value of a sample
!            correlation coefficient computed from XYN observation pairs being
!            greater than R is CORTEST under the null hypothesis of no correlation in the
!            parent normal population.
!
!            CORTEST needs to be specified only on the last call to COMP_COR (e.g. when LAST=true).
!          
!   COV      (INPUT, OPTIONAL) logical(lgl)
!            On entry, if COV=true, covariance coefficients between the data matrices XX and
!            YY are computed instead of correlation coefficients. If COV=true, Z and PROB are
!            set to Nan code.
!            COV needs to be specified only on the last call to COMP_COR (e.g. when LAST=true).
!          
!
! Further Details
! _______________
!
!   The subroutine computes the basic univariate statistics and the correlation
!   coefficients with only one pass through the data.
!
!   If fewer than two valid observations were present, the pertinent statistics are set
!   to Nan code.
!
!   For more details on correlation and regression analysis, see:
!
!   (1) von Storch, H., and Zwiers, F.W., 2002:
!            Statistical Analysis in Climate Research.
!            Cambridge, UK, Chapter 8, 484 pp., ISBN:9780521012300
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true, false
    use Reals_Constants,   only : zero, one, two, half
    use Num_Constants,     only : nan, lamch
    use Char_Constants,    only : stat_error1, stat_error2, stat_error5
    use Utilities,         only : assert_eq, assert, merror
    use Prob_Procedures,   only : probt, pinvstudent
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),    dimension(:,:)            :: x, y
    real(stnd), intent(inout), dimension(:,:)            :: xstat, ystat, xycor
    real(stnd), intent(inout)                            :: xyn
    real(stnd), intent(out),   dimension(:,:),  optional :: prob, z
    real(stnd), intent(inout),                  optional :: cortest
!
    integer(i4b), intent(in), optional :: dimvar, dimvary, ndf_max
!
    logical(lgl), intent(in)            :: first, last
    logical(lgl), intent(in),  optional :: cov
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                           :: a, b, xynm1
    real(stnd), dimension(size(xycor,1)) :: dxi, tmp
    real(stnd), dimension(size(xycor,2)) :: dyi
!
    integer(i4b) :: i, j, n, m, p
!
    logical(lgl) :: dimx_is_one, dimy_is_one, comp_cor, comp_prob, comp_z
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_cor'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dimx_is_one = i==1_i4b
!
    if ( present(dimvary) ) then
!
        if ( dimvary==1_i4b .or. dimvary==2_i4b ) then
            j = dimvary
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        j = 1_i4b
!
    end if
!
    dimy_is_one = j==1_i4b
!
    n = assert_eq(  int(size(x,3-int(i)),i4b),  &
                    int(size(y,3-int(j)),i4b),  &
                    name_proc )
!
    m = assert_eq(  int(size(x,int(i)),i4b),      &
                    int(size(xstat,1),i4b),       &
                    int(size(xycor,1),i4b),       &
                    name_proc )
!
    p = assert_eq(  int(size(y,int(j)),i4b),      &
                    int(size(ystat,1),i4b),       &
                    int(size(xycor,2),i4b),       &
                    name_proc )
!
    if ( m<=0_i4b .or. p<=0_i4b ) call merror( name_proc//stat_error1 )
!
    call assert( logical(size(xstat,2)==2,lgl),    &
                 logical(size(ystat,2)==2,lgl),    &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBMATRICES ARE THE FIRST.
!
!       INITIALIZE xstat, ystat, xycor AND xyn .
!
        xstat(:m,1_i4b:2_i4b) = zero       
        ystat(:p,1_i4b:2_i4b) = zero
        xycor(:m,:p) = zero
        xyn = zero
!
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  (m*p)>=omp_limit                    .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)      &
!$OMP         ,PRIVATE(i,j)
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS AND CORRECTED SUMS OF SQUARES AND CROSS-PRODUCTS, 
!       HANDLING THE DATA ONCE.
!
!$OMP SINGLE
!
        xynm1 = xyn
        xyn = xynm1 + one
        a = one/xyn
        b = xynm1*a
!
        if ( dimx_is_one ) then
            dxi(:m) = x(:m,i) - xstat(:m,1_i4b)
        else
            dxi(:m) = x(i,:m) - xstat(:m,1_i4b)
        end if
!
        tmp(:m) = b*dxi(:m)
!
        xstat(:m,1_i4b) = xstat(:m,1_i4b) + a*dxi(:m)
        xstat(:m,2_i4b) = xstat(:m,2_i4b) + tmp(:m)*dxi(:m)
!
        if ( dimy_is_one ) then
            dyi(:p) = y(:p,i) - ystat(:p,1_i4b)
        else
            dyi(:p) = y(i,:p) - ystat(:p,1_i4b)
        end if
!
        ystat(:p,1_i4b) = ystat(:p,1_i4b) + a*dyi(:p)
        ystat(:p,2_i4b) = ystat(:p,2_i4b) + b*( dyi(:p)*dyi(:p) )
!
!$OMP END SINGLE
!
!$OMP DO SCHEDULE(STATIC) 
!
        do j = 1_i4b, p
!
#ifdef _BLAS
            call axpy( m, dyi(j), tmp(:m), 1_i4b, xycor(:m,j), 1_i4b )
#else
            xycor(:m,j) = xycor(:m,j) + dyi(j)*tmp(:m)
#endif
!
        end do
!
!$OMP END DO
!
    end do
!
!$OMP END PARALLEL
!
    if ( last ) then
!
!       THE CURRENT SUBMATRICES ARE THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
!       TEST OPTIONAL ARGUMENTS.
!        
        if ( present(prob) ) then
!
            call assert( logical(size(prob,1)==m,lgl),   &
                         logical(size(prob,2)==p,lgl),   &
                         name_proc )
            comp_prob = true
!
        else
!
            comp_prob = false
!
        end if
!
        if ( present(z) ) then
!
            call assert( logical(size(z,1)==m,lgl),   &
                         logical(size(z,2)==p,lgl),   &
                         name_proc )
            comp_z = true
!
        else
!
            comp_z = false
!
        end if
!
        comp_cor = true
!
        if ( present(cov) ) then
!
            comp_cor = .not.cov
!
        end if
!
        b = nan()
!
        if ( xyn>one ) then
!
            a = one/xyn
            xynm1 = lamch('s')
!
!           COMPUTE VARIANCES OF XX.
!
            dxi(:m) = xstat(:m,2_i4b)
            xstat(:m,2_i4b) = dxi(:m)*a
!
!           COMPUTE VARIANCE OF YY.
!
            dyi(:p) = ystat(:p,2_i4b)
            ystat(:p,2_i4b) = dyi(:p)*a
!
            if ( comp_cor ) then
!
!$OMP PARALLEL IF(test_par)        &
!$OMP         ,PRIVATE(i,j,a,tmp)
!
!               COMPUTE CORRELATION COEFFICIENTS BETWEEN XX AND YY.
!
!$OMP DO SCHEDULE(STATIC) 
!
                do j = 1_i4b, p
!
                    tmp(:m) = dxi(:m)*dyi(j)
!
                    where ( tmp(:m)>xynm1 )
                        xycor(:m,j) = xycor(:m,j)/sqrt( tmp(:m) )
                    elsewhere
                        xycor(:m,j) = b
                    end where
!
                end do
!
!$OMP END DO
!
!               OUTPUT FISHER'S z TRANSFORMATION, IF REQUIRED.
!
                if ( comp_z ) then
!
!$OMP DO SCHEDULE(STATIC) 
!
                    do j = 1_i4b, p
!
                        tmp(:m) = dxi(:m)*dyi(j)
!
                        where ( tmp(:m)>xynm1 )
                            z(:m,j) = half*log(( (one+xycor(:m,j)) + xynm1 )/( (one-xycor(:m,j)) + xynm1 ))
                        elsewhere
                            z(:m,j) = b
                        end where
!
                    end do
!
!$OMP END DO
!
                end if
!
!               OUTPUT SIGNIFICANCE PROBABILITY, IF REQUIRED.
!
                if ( comp_prob ) then
!
                    if ( xyn>two ) then
!
                        a = xyn - two
                        i = int( a, i4b )
                        a = sqrt( a )
!
!$OMP DO SCHEDULE(STATIC) 
!
                        do j = 1_i4b, p
!
                            tmp(:m) = dxi(:m)*dyi(j)
!
                            where ( tmp(:m)>xynm1 .and. abs(xycor(:m,j))/=one )
                                tmp(:m) = a*abs(xycor(:m,j))/sqrt( one - xycor(:m,j)**2 )
                            elsewhere
                               tmp(:m) = -one
                            end where
!
                            prob(:m,j) = two*probt( tmp(:m), i, true, ndf_max=ndf_max )
                            where ( tmp(:m)==-one )  prob(:m,j) = merge( zero, b, abs(xycor(:m,j))==one )
!
                        end do
!
!$OMP END DO
!
                    else
!
!$OMP SINGLE
!
                        prob(:m,:p) = b
!
!$OMP END SINGLE
!
                    end if
!
                end if
!
!$OMP END PARALLEL
!
            else
!
!               COMPUTE COVARIANCE COEFFICIENTS BETWEEN XX AND YY.
!
                xycor(:m,:p) = xycor(:m,:p)*a
!
                if ( comp_prob ) prob(:m,:p) = b
                if ( comp_z )    z(:m,:p) = b
!
            end if
!
        else
!
!           FEWER THAN TWO VALID OBSERVATIONS, SET THE PERTINENT STATISTICS TO NaN CODE.
!
            if ( xyn<one ) then
                xstat(:m,1_i4b:2_i4b) = b
                ystat(:p,1_i4b:2_i4b) = b
            end if
!
            xycor(:m,:p) = b
!
            if ( comp_prob ) prob(:m,:p) = b
            if ( comp_z )    z(:m,:p) = b
!
        end if
!
        if ( present(cortest) ) then
!
!           COMPUTE TWO-TAIL QUANTILE OF STUDENT-T DISTRIBUTION.
!
            if ( zero>=cortest .or. cortest>=one ) then
                call merror( name_proc//stat_error2 )
            else if ( xyn>two ) then
                a = xyn - two
                xynm1 = pinvstudent( cortest, a )
                xynm1 = xynm1*xynm1
                cortest = sqrt( xynm1/( xynm1 + a ) )
            else
                cortest = b
            end if
!
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_cor_rm_rm
! ________________________________
!
    end subroutine comp_cor_rm_rm
!
! =========================================================================================
!
    subroutine comp_cor_miss_rv( x, y, first, last, xstat, ystat, xycor, xymiss,   &
                                 z, prob, ndf_max, cov )
!
! Purpose
! _______
!
!   COMP_COR_MISS computes estimates of mean, variance and correlation
!   coefficient from two data vectors XX and YY possibly containing
!   missing values.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:)
!            On entry, input subvector containing size(X) observations
!            from the vector of data XX for which basic univariate and bivariate
!            statistics are desired. If all the data are available at once, X
!            can be the full data vector XX.
!
!   Y        (INPUT) real(stnd), dimension(:)
!            On entry, input subvector containing size(X) observations
!            from another vector of data YY for which correlation coefficient
!            with XX is desired. If all the data are available at once, YY
!            can be the full data vector.
!
!            The size of Y must verify: size(Y) = size(X) .
!
!   FIRST    (INPUT) logical(lgl)
!            On entry, if:
!
!            - FIRST = true  the current subvector is the first subvector 
!              of the data vector XX (or YY).
!            - FIRST = false the current subvector is not the first subvector
!              of the data vector XX (or YY).
!
!   LAST     (INPUT) logical(lgl)
!            On entry, if:
!
!            - LAST = true  the current subvector is the last subvector 
!              of the data vector XX (or YY).
!            - LAST = false the current subvector is not the last subvector
!              of the data vector XX (or YY).
!
!   XSTAT    (INPUT/OUTPUT) real(stnd), dimension(4)
!            On entry, after the first call to COMP_COR_MISS (e.g. when FIRST=true),
!            XSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_COR_MISS. XSTAT should not be changed between calls to
!            COMP_COR_MISS.
!
!            On exit, when LAST=true, XSTAT contains the following statistics:
!
!            - XSTAT(1) contains the mean value of the data vector XX.
!            - XSTAT(2) contains the variance of the data vector XX.
!            - XSTAT(3) contains the the number of non-missing observations in
!              the data vector XX.
!            - XSTAT(4) is used as workspace.
!
!            The size of XSTAT must verify: size(XSTAT) = 4.
!
!   YSTAT    (INPUT/OUTPUT) real(stnd), dimension(4)
!            On entry, after the first call to COMP_COR_MISS (e.g. when FIRST=true),
!            YSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_COR_MISS. YSTAT should not be changed between calls to
!            COMP_COR_MISS.
!
!            On exit, when LAST=true, YSTAT contains the following statistics:
!
!            - YSTAT(1) contains the mean value of the data vector YY.
!            - YSTAT(2) contains the variance of the data vector YY.
!            - YSTAT(3) contains the the number of non-missing observations
!              in the data vector YY.
!            - YSTAT(4) is used as workspace.
!
!            The size of YSTAT must verify: size(YSTAT) = 4.
!
!   XYCOR    (INPUT/OUTPUT) real(stnd), dimension(4)
!            On entry, after the first call to COMP_COR_MISS (e.g. when FIRST=true),
!            XYCOR is used as workspace to accumulate quantities on previous calls
!            to COMP_COR_MISS. XYCOR should not be changed between calls to
!            COMP_COR_MISS.
!
!            On exit, when LAST=true, XYCOR contains the following statistics:
!
!            - XYCOR(1) contains the correlation coefficient between XX(:) and YY(:).
!            - XYCOR(2) contains the incidence value between XX(:) and YY(:). XYCOR(2)
!              indicates the number of non-missing pairs of observations which
!              were used in the calculation of XYCOR(1).
!            - XYCOR(3:4) is used as workspace.
!
!            The size of XYCOR must verify: size(XYCOR) = 4.
!
!   XYMISS   (INPUT) real(stnd)
!            On entry, the missing value indicator. Any value in
!            X or Y which is equal to XYMISS is assumed to be missing
!            or invalid.
!            The basic univariate and bivariate statistics are computed
!            on all the observations where XX and YY are not missing (see Further Details).
!
!   Z        (OUTPUT, OPTIONAL) real(stnd)
!            On exit, when LAST=true, Z contains the Fisher's Z transformation of
!            XYCOR(1).
!
!            Z needs to be specified only on the last call to COMP_COR_MISS (e.g. when LAST=true).
!          
!   PROB     (OUTPUT, OPTIONAL) real(stnd)
!            On exit, when LAST=true, PROB gives the probability that the random sample
!            of XYCOR(2) observation pairs YY(:) and XX(:) came from a bivariate normal
!            population with a correlation coefficient equal to zero.
!
!            PROB needs to be specified only on the last call to COMP_COR_MISS (e.g. when LAST=true).
!          
!   NDF_MAX  (INPUT, OPTIONAL) integer(i4b)
!            On entry, when argument PROB is present, NDF_MAX is used as follows, if:
!
!            - XYCOR(2)-2 is lower or equal to NDF_MAX, the t_density is integrated
!              for computing PROB.
!            - XYCOR(2)-2 is greater than NDF_MAX, an asymptotic series is used
!              for computing PROB.
!
!            The default is 20.
!
!            NDF_MAX needs to be specified only on the last call to COMP_COR_MISS (e.g. when LAST=true).
!          
!   COV      (INPUT, OPTIONAL) logical(lgl)
!            On entry, if COV=true, covariance coefficient between the data vectors XX and
!            YY are computed instead of correlation coefficient. If COV=true, Z and PROB are
!            set to XYMISS .
!
!            COV needs to be specified only on the last call to COMP_COR_MISS (e.g. when LAST=true).
!
!
! Further Details
! _______________
!
!   The subroutine computes the basic univariate statistics and the correlation
!   coefficient with only one pass through the data.
!
!   If fewer than two valid observations were present, the pertinent statistics are set to XYMISS.
!
!   The means and standard-deviations of XX and YY are computed from all valid data. The
!   correlation coefficients are based on these univariate statistics and on all valid pairs
!   of observations.
!
!   For more details on correlation and regression analysis, see:
!
!   (1) von Storch, H., and Zwiers, F.W., 2002:
!            Statistical Analysis in Climate Research.
!            Cambridge, UK, Chapter 8, 484 pp., ISBN:9780521012300
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, half
    use Num_Constants,     only : lamch
    use Utilities,         only : assert_eq, assert
    use Prob_Procedures,   only : probt
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                  :: xymiss
    real(stnd), intent(in),    dimension(:) :: x, y
    real(stnd), intent(inout), dimension(:) :: xstat, ystat, xycor
    real(stnd), intent(out),    optional    :: prob, z
!
    integer(i4b), intent(in), optional :: ndf_max
!
    logical(lgl), intent(in)            :: first, last
    logical(lgl), intent(in),  optional :: cov
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: dxi, dfxi, dyi, dfyi, cntx, cnty, cntxy
!
    integer(i4b) :: i, n
!
    logical(lgl) :: mask, comp_cor
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_cor_miss'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq(  int(size(x),i4b),       &
                    int(size(y),i4b),       &
                    name_proc )
!
    call assert( logical(size(xstat)==4,lgl),    &
                 logical(size(ystat)==4,lgl),    &
                 logical(size(xycor)==4,lgl),    &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBVECTOR IS THE FIRST.
!
!       INITIALIZE xstat, ystat AND xycor.
!
        xstat(1_i4b:3_i4b) = zero       
        ystat(1_i4b:3_i4b) = zero
        xycor(1_i4b:4_i4b) = zero
!
        xstat(4_i4b) = xymiss
        ystat(4_i4b) = xymiss
!
    end if
!
    cntx = xstat(3_i4b)
    cnty = ystat(3_i4b)
    cntxy = xycor(2_i4b)
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS AND CORRECTED SUMS OF SQUARES AND CROSS-PRODUCT, 
!       HANDLING THE DATA ONCE.
!
        mask = x(i)/=xymiss
!
        if ( mask ) then
!
            if ( xstat(4_i4b)==xymiss ) xstat(4_i4b) = x(i)
!
            cntx = cntx + one
            dxi = x(i) - xstat(1_i4b)
            dfxi = dxi/cntx
            xstat(1_i4b) = xstat(1_i4b) + dfxi
            xstat(2_i4b) = xstat(2_i4b) + dxi*dxi - dxi*dfxi
!
        end if
!
        if ( y(i)/=xymiss ) then
!
            if ( ystat(4_i4b)==xymiss ) ystat(4_i4b) = y(i)
!
            cnty = cnty + one
            dyi = y(i) - ystat(1_i4b)
            dfyi = dyi/cnty
            ystat(1_i4b) = ystat(1_i4b) + dfyi
            ystat(2_i4b) = ystat(2_i4b) + dyi*dyi - dyi*dfyi
!
            if ( mask ) then
!
                cntxy = cntxy + one
                dxi = x(i) - xstat(4_i4b)
                dyi = y(i) - ystat(4_i4b)
                xycor(1_i4b) = xycor(1_i4b) + dxi*dyi
                xycor(3_i4b) = xycor(3_i4b) + dxi
                xycor(4_i4b) = xycor(4_i4b) + dyi
!
            end if
!
        end if
!
    end do
!
!   SAVE COUNTS OF OBSERVATIONS FOR THE NEXT CALL.
!
    xstat(3_i4b) = cntx
    ystat(3_i4b) = cnty
    xycor(2_i4b) = cntxy
!
    if ( last ) then
!
        comp_cor = true
!
        if ( present(cov) ) then
            comp_cor = .not.cov
        end if
!
!       THE CURRENT SUBVECTOR IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
        if ( cntx>one ) then
!
!           COMPUTE VARIANCE OF XX.
!
            dfxi = xstat(2_i4b)/cntx
            xstat(2_i4b) = dfxi
!
        else if ( cntx<one ) then
!
!           FEWER THAN ONE VALID OBSERVATION, SET THE STATISTICS TO xymiss .
!
            xstat(1_i4b:2_i4b) = xymiss
!
        end if
!
        if ( cnty>one ) then
!
!           COMPUTE VARIANCE OF YY.
!
            dfyi = ystat(2_i4b)/cnty
            ystat(2_i4b) = dfyi
!
        else if ( cnty<one ) then
!
!           FEWER THAN ONE VALID OBSERVATION, SET THE STATISTICS TO xymiss .
!
            ystat(1_i4b:2_i4b) = xymiss
!
        end if
!
        if ( cntxy>one ) then
!
            dyi = lamch('s')
!
            if ( comp_cor ) then
!
                dxi = dfxi*dfyi
!
                if ( dxi>dyi ) then
!
!                   COMPUTE CORRELATION COEFFICIENT.
!
                    dfxi = xstat(4_i4b) - xstat(1_i4b)
                    dfyi = ystat(4_i4b) - ystat(1_i4b)
                    xycor(1_i4b) = xycor(1_i4b) + dfxi*xycor(4_i4b) +       &
                                   dfyi*xycor(3_i4b) + dfxi*dfyi*cntxy
                    xycor(1_i4b) = xycor(1_i4b)/( cntxy*sqrt( dxi ) )
!
                    if ( xycor(1_i4b)>one ) xycor(1_i4b) = one
                    if ( xycor(1_i4b)<-one ) xycor(1_i4b) = -one
!
                else
!
!                   THE OBSERVATIONS ON XX or YY WERE CONSTANT, SET THE STATISTIC TO xymiss .
!
                    xycor(1_i4b) = xymiss
!
                end if
!
!               OUTPUT FISHER'S z TRANSFORMATION, IF REQUIRED.
!
                if ( present(z) ) then
!
                    if ( dxi>dyi ) then
                        z = half*log(( (one+xycor(1_i4b)) + dyi )/( (one-xycor(1_i4b)) + dyi ))
                    else
                        z = xymiss
                    end if
!
                end if
!
!               OUTPUT SIGNIFICANCE PROBABILITY IF REQUIRED.
!
                if ( present(prob) ) then
!
                    if ( cntxy>two .and. dxi>dyi ) then
!
                        dxi = abs( xycor(1_i4b) )
!
                        if ( dxi/=one ) then
                            dfxi = cntxy - two
                            i = int( dfxi, i4b )
                            dyi = sqrt( dfxi )*dxi/sqrt( one - dxi*dxi )
                            prob = two*probt( dyi, i, true, ndf_max=ndf_max )
                        else
                            prob = zero
                        end if
!
                    else
!
                        prob = xymiss
!
                    end if
!
                end if
!
            else
!
!               COMPUTE COVARIANCE COEFFICIENTS BETWEEN XX AND YY.
!
                dfxi = xstat(4_i4b) - xstat(1_i4b)
                dfyi = ystat(4_i4b) - ystat(1_i4b)
                xycor(1_i4b) = xycor(1_i4b) + dfxi*xycor(4_i4b) +       &
                               dfyi*xycor(3_i4b) + dfxi*dfyi*cntxy
                xycor(1_i4b) = xycor(1_i4b)/cntxy
!
!
                if ( present(prob) ) prob = xymiss
                if ( present(z) )    z    = xymiss
!
            end if
!
        else
!
!           FEWER THAN TWO VALID OBSERVATIONS, SET THE STATISTICS TO xymiss .
!
            xycor(1_i4b) = xymiss
!
            if ( present(prob) ) prob = xymiss
            if ( present(z) )    z    = xymiss
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_cor_miss_rv
! __________________________________
!
    end subroutine comp_cor_miss_rv
!
! =========================================================================================
!
    subroutine comp_cor_miss_rm( x, y, first, last, xstat, ystat, xycor, xymiss, &
                                 dimvar, z, prob, ndf_max, cov )
!
! Purpose
! _______
!
!   COMP_COR_MISS computes estimates of means, variances and correlation
!   coefficients (with another data vector YY) from a data matrix XX  possibly
!   containing missing values.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:,:)
!            On entry, input submatrix containing size(X,3-DIMVAR)
!            observations on size(X,DIMVAR) variables from the matrix of data XX
!            for which basic univariate  and bivariate statistics are desired. By
!            default, DIMVAR is equal to 1. See description of optional DIMVAR
!            argument for details.
!            If all the data are available at once, X can be the full 
!            data matrix XX.
!
!   Y        (INPUT) real(stnd), dimension(:)
!            On entry, input subvector containing size(X,3-DIMVAR) observations
!            from a vector of data YY for which correlation coefficient
!            with XX is desired. If all the data are available at once, YY
!            can be the full data vector.
!
!            The size of Y must verify: size(Y) = SIZE(X,3-DIMVAR) .
!
!   FIRST    (INPUT) logical(lgl)
!            On entry, if:
!
!            - FIRST = true  the current submatrix is the first submatrix 
!              of the data matrix XX.
!            - FIRST = false the current submatrix is not the first submatrix
!              of the data matrix XX.
!
!   LAST     (INPUT) logical(lgl)
!            On entry, if:
!
!            - LAST = true  the current submatrix is the last submatrix 
!              of the data matrix XX.
!            - LAST = false the current submatrix is not the last submatrix
!              of the data matrix XX.
!
!   XSTAT    (INPUT/OUTPUT) real(stnd), dimension(:,4)
!            On entry, after the first call to COMP_COR_MISS (e.g. when FIRST=true),
!            XSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_COR_MISS. XSTAT should not be changed between calls to
!            COMP_COR_MISS.
!
!            On exit, when LAST=true, XSTAT contains the following statistics
!            on all variables:
!
!            - XSTAT(:,1) contains the mean values of the data matrix XX.
!            - XSTAT(:,2) contains the variances of the data matrix XX.
!            - XSTAT(:,3) contains the the numbers of non-missing observations
!              in the data matrix XX.
!            - XSTAT(:,4) is used as workspace.
!
!            The shape of XSTAT must verify:
!
!            - size(XSTAT,1) = size(X,DIMVAR) ,
!            - size(XSTAT,2) = 4 .
!
!   YSTAT    (INPUT/OUTPUT) real(stnd), dimension(4)
!            On entry, after the first call to COMP_COR_MISS (e.g. when FIRST=true),
!            YSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_COR_MISS. YSTAT should not be changed between calls to
!            COMP_COR_MISS.
!
!            On exit, when LAST=true, YSTAT contains the following statistics:
!
!            - YSTAT(1) contains the mean value of the data vector YY.
!            - YSTAT(2) contains the variance of the data vector YY.
!            - YSTAT(3) contains the the number of non-missing observations
!              in the data vector YY.
!            - YSTAT(4) is used as workspace.
!
!            The size of YSTAT must verify: size(YSTAT) = 4.
!
!   XYCOR    (INPUT/OUTPUT) real(stnd), dimension(:,4)
!            On entry, after the first call to COMP_COR_MISS (e.g. when FIRST=true),
!            XYCOR is used as workspace to accumulate quantities on previous calls
!            to COMP_COR_MISS. XYCOR should not be changed between calls to
!            COMP_COR_MISS.
!
!            On exit, when LAST=true, XYCOR contains the following statistics:
!
!            - XYCOR(i,1) contains the correlation coefficients between XX(i,:) ( XX(:,i) if
!              DIMVAR=2 ) and YY(:).
!            - XYCOR(i,2) contains the incidence values between XX(i,:) ( XX(:,i) if DIMVAR=2 )
!              and YY(:). XYCOR(i,2) indicates the numbers of non-missing pairs of 
!              observations which were used in the calculation of XYCOR(i,1).
!            - XYCOR(:,3:4) is used as workspace.
!
!            The shape of XYCOR must verify:
!
!            - size(XYCOR,1) = size(X,DIMVAR) ,
!            - size(XYCOR,2) = 4 .
!
!   XYMISS   (INPUT) real(stnd)
!            On entry, the missing value indicator. Any value in
!            X or Y which is equal to XYMISS is assumed to be missing
!            or invalid.
!            The basic univariate and bivariate statistics are computed
!            on all the observations where X and Y are not missing (see Further Details).
!
!   DIMVAR   (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVAR is present, DIMVAR is used as follows:
!
!            - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!              on size(X,1) variables.
!            - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!              on size(X,2) variables.
!
!            The default is DIMVAR = 1.
!
!   Z        (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            On exit, when LAST=true, Z contains the Fisher's Z transformation of
!            XYCOR(:,1).
!
!            Z needs to be specified only on the last call to COMP_COR_MISS (e.g. when LAST=true).
!          
!            The size of Z must verify:  size(Z) = size(X,DIMVAR).
!
!   PROB     (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            On exit, when LAST=true, PROB(i) gives the probability that the random sample
!            of XYCOR(i,2) observation pairs YY(:) and XX(i,:) ( XX(:,i) if DIMVAR=2 ) 
!            came from a bivariate normal population with a correlation coefficient
!            equal to zero.
!
!            PROB needs to be specified only on the last call to COMP_COR_MISS (e.g. when LAST=true).
!          
!            The size of PROB must verify:  size(PROB) = size(X,DIMVAR).
!
!   NDF_MAX  (INPUT, OPTIONAL) integer(i4b)
!            On entry, when argument PROB is present, NDF_MAX is used as follows:
!
!            - If XYCOR(i,2)-2 is lower or equal to NDF_MAX, the t_density is integrated
!              for computing PROB(i).
!            - If XYCOR(i,2)-2 is greater than NDF_MAX, an asymptotic series is used
!              for computing PROB(i).
!
!            The default is 20.
!
!            NDF_MAX needs to be specified only on the last call to COMP_COR_MISS (e.g. when LAST=true).
!          
!   COV      (INPUT, OPTIONAL) logical(lgl)
!            On entry, if COV=true, covariance coefficients between the data matrix XX and data
!            vector YY are computed instead of correlation coefficients. If COV=true, Z and PROB are
!            set to XYMISS .
!
!            COV needs to be specified only on the last call to COMP_COR_MISS (e.g. when LAST=true).
!
!
! Further Details
! _______________
!
!   The subroutine computes the basic univariate statistics and the correlation
!   coefficients with only one pass through the data.
!
!   If fewer than two valid observations were present, the statistics are set to XYMISS .
!
!   The means and standard-deviations of XX and YY are computed from all valid data. The
!   correlation coefficients are based on these univariate statistics and on all valid pairs
!   of observations.
!
!   For more details on correlation and regression analysis, see:
!
!   (1) von Storch, H., and Zwiers, F.W., 2002:
!            Statistical Analysis in Climate Research.
!            Cambridge, UK, Chapter 8, 484 pp., ISBN:9780521012300
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, half
    use Num_Constants,     only : lamch
    use Char_Constants,    only : stat_error1, stat_error5
    use Utilities,         only : assert_eq, assert, merror
    use Prob_Procedures,   only : probt
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                               :: xymiss
    real(stnd), intent(in),    dimension(:,:)            :: x
    real(stnd), intent(in),    dimension(:)              :: y
    real(stnd), intent(inout), dimension(:,:)            :: xstat, xycor
    real(stnd), intent(inout), dimension(:)              :: ystat
    real(stnd), intent(out),   dimension(:),    optional :: prob, z
!
    integer(i4b), intent(in), optional :: dimvar, ndf_max
!
    logical(lgl), intent(in)            :: first, last
    logical(lgl), intent(in),  optional :: cov
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                           :: dyi, dfyi, cnty
    real(stnd), dimension(size(xycor,1)) :: xi, dxi, dfxi
!
    integer(i4b) :: i, n, m
!
    logical(lgl)                            :: dim_is_one, comp_cor
    logical(lgl), dimension(size(xycor,1))  :: mask
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_cor_miss'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
    n = assert_eq(  int(size(x,3-int(i)),i4b),   &
                    int(size(y),i4b),            &
                    name_proc )
!
    m = assert_eq(  int(size(x,int(i)),i4b),      &
                    int(size(xstat,1),i4b),       &
                    int(size(xycor,1),i4b),       &
                    name_proc )
!
    if ( m<=0_i4b ) call merror( name_proc//stat_error1 )
!
    call assert( logical(size(xstat,2)==4,lgl),    &
                 logical(size(xycor,2)==4,lgl),    &
                 logical(size(ystat)==4,lgl),      &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBMATRIX IS THE FIRST.
!
!       INITIALIZE xstat, ystat AND xycor.
!
        xstat(:m,1_i4b:3_i4b) = zero       
        ystat(1_i4b:3_i4b) = zero
        xycor(:m,1_i4b:4_i4b) = zero
!
        xstat(:m,4_i4b) = xymiss
        ystat(4_i4b) = xymiss
!
    end if
!
!   GET COUNT OF OBSERVATIONS FOR THE PREVIOUS CALL.
!
    cnty = ystat(3_i4b)
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS AND CORRECTED SUMS OF SQUARES AND CROSS-PRODUCTS, 
!       HANDLING THE DATA ONCE.
!
        if ( dim_is_one ) then
            xi(:m) = x(:m,i)
        else
            xi(:m) = x(i,:m)
        end if
!
        mask(:m) = xi(:m)/=xymiss
!
        where ( mask(:m) )
!
            where ( xstat(:m,4_i4b)==xymiss ) xstat(:m,4_i4b) = xi(:m)
!
!            xstat(:m,4_i4b) = merge( xi(:m), xstat(:m,4_i4b), xstat(:m,4_i4b)==xymiss )
!
            xstat(:m,3_i4b) = xstat(:m,3_i4b) + one
            dxi(:m) = xi(:m) - xstat(:m,1_i4b)
            dfxi(:m) = dxi(:m)/xstat(:m,3_i4b)
            xstat(:m,1_i4b) = xstat(:m,1_i4b) + dfxi(:m)
            xstat(:m,2_i4b) = xstat(:m,2_i4b) + dxi(:m)*dxi(:m) - dxi(:m)*dfxi(:m)
!
        end where
!
        if ( y(i)/=xymiss ) then
!
            if ( ystat(4_i4b)==xymiss ) ystat(4_i4b) = y(i)
!
            cnty = cnty + one
            dyi = y(i) - ystat(1_i4b)
            dfyi = dyi/cnty
            ystat(1_i4b) = ystat(1_i4b) + dfyi
            ystat(2_i4b) = ystat(2_i4b) + dyi*dyi - dyi*dfyi
!
            dyi = y(i) - ystat(4_i4b)
!
            where ( mask(:m) )
!
                dxi(:m) = xi(:m) - xstat(:m,4_i4b)
                xycor(:m,1_i4b) = xycor(:m,1_i4b) + dxi(:m)*dyi
                xycor(:m,2_i4b) = xycor(:m,2_i4b) + one
                xycor(:m,3_i4b) = xycor(:m,3_i4b) + dxi(:m)
                xycor(:m,4_i4b) = xycor(:m,4_i4b) + dyi
!
            end where
!
        end if
!
    end do
!
!   SAVE COUNT OF OBSERVATIONS FOR THE NEXT CALL.
!
    ystat(3_i4b) = cnty
!
    if ( last ) then
!
!       THE CURRENT SUBMATRIX IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
!       TEST OPTIONAL ARGUMENTS.
!
        if ( present(prob) ) call assert( logical(size(prob)==m,lgl), name_proc )
!
        if ( present(z) ) call assert( logical(size(z)==m,lgl), name_proc )
!
        comp_cor = true
        if ( present(cov) ) then
            comp_cor = .not.cov
        end if
!
        mask(:m) = xstat(:m,3_i4b)>one
!
        where ( mask(:m) )
!
!           COMPUTE VARIANCES OF XX.
!
            dfxi(:m) = xstat(:m,2_i4b)/xstat(:m,3_i4b)
            xstat(:m,2_i4b) = dfxi(:m)
!
        elsewhere
!
!           FEWER THAN TWO VALID OBSERVATIONS, SET THE PERTINENT STATISTICS TO xymiss .
!
            dfxi(:m) = zero
!
            where ( xstat(:m,3_i4b)<one )
               xstat(:m,1_i4b) = xymiss
               xstat(:m,2_i4b) = xymiss
            elsewhere
               xstat(:m,2_i4b) = zero
            end where
!
!            xstat(:m,1_i4b) = merge( xymiss, xstat(:m,1_i4b), xstat(:m,3_i4b)<one )
!            xstat(:m,2_i4b) = merge( xymiss, zero, xstat(:m,3_i4b)<one )
!
        end where
!
        if ( cnty>one ) then
!
!           COMPUTE VARIANCE OF YY.
!
            dfyi = ystat(2_i4b)/cnty
            ystat(2_i4b) = dfyi
!
        else
!
!           FEWER THAN TWO VALID OBSERVATIONS, SET THE PERTINENT STATISTICS TO xymiss .
!
            dfyi = zero
            if ( cnty<one ) ystat(1_i4b:2_i4b) = xymiss
!
        end if
!
        if ( comp_cor ) then
!
            dyi = lamch( 's' )
!
            dxi(:m) = dfxi(:m)*dfyi
            mask(:m) = xycor(:m,2_i4b)>one .and. dxi(:m)>dyi
            dfyi = ystat(4_i4b) - ystat(1_i4b)
!
            where ( mask(:m) )
!
!               COMPUTE CORRELATION COEFFICIENTS BETWEEN XX AND YY.
!
                dfxi(:m) = xstat(:m,4_i4b) - xstat(:m,1_i4b)
                xycor(:m,1_i4b) = xycor(:m,1_i4b) + dfxi(:m)*xycor(:m,4_i4b) +          &
                                  dfyi*xycor(:m,3_i4b) + (dfxi(:m)*dfyi)*xycor(:m,2_i4b)
                dfxi(:m) = xycor(:m,1_i4b)/( xycor(:m,2_i4b)*sqrt( dxi(:m) ) )
!
                where ( dfxi(:m)>one )
                    dxi(:m) = one
                elsewhere
                    dxi(:m) = dfxi(:m)
                end where
!
!                dxi(:m) = merge( one, dfxi(:m), dfxi(:m)>one )
!
                where ( dxi(:m)<-one )
                    xycor(:m,1_i4b) = -one
                elsewhere
                    xycor(:m,1_i4b) = dxi(:m)
                end where
!
!                xycor(:m,1_i4b) = merge( -one, dxi(:m), dxi(:m)<-one )
!
            elsewhere
!
!               THE OBSERVATIONS ON XX or YY WERE CONSTANT OR FEWER THAN
!               TWO VALID PAIRS WERE PRESENT, SET THE CORRELATION TO xymiss .
!
                xycor(:m,1_i4b) = xymiss
!
            end where
!
!           OUTPUT FISHER'S z TRANSFORMATION, IF REQUIRED.
!
            if ( present(z) ) then
!
                where ( mask(:m) )
                    z(:m) = half*log(( (one+xycor(:m,1_i4b)) + dyi )/( (one-xycor(:m,1_i4b)) + dyi ))
                elsewhere
                    z(:m) = xymiss
                end where
!
            end if
!
            if ( present(prob) ) then
!
!               OUTPUT SIGNIFICANCE PROBABILITIES.
!
                mask(:m) = mask(:m) .and. xycor(:m,2_i4b)>two
!
                where ( mask(:m) )
                    mask(:m) = abs( xycor(:m,1_i4b) )/=one
                    prob(:m) = zero
                elsewhere
                    prob(:m) = xymiss
                end where
!
                where ( mask(:m) )
                    dfxi(:m) = xycor(:m,2_i4b) - two
                    dxi(:m) = sqrt( dfxi(:m) )*abs( xycor(:m,1_i4b) )/sqrt( one - xycor(:m,1_i4b)**2 )
                end where
!
!$OMP PARALLEL DO IF(m>omp_chunk)                &
!$OMP            ,SCHEDULE(DYNAMIC,omp_chunk)    &
!$OMP            ,PRIVATE(i), SHARED(ndf_max)
!
                do i = 1_i4b, m
!
                    if ( mask(i) ) then
                        prob(i) = two*probt( dxi(i), int( dfxi(i), i4b ), true, ndf_max=ndf_max )
                    end if
!
                end do
!
!$OMP END PARALLEL DO
!
            end if
!
        else
!
            mask(:m) = xycor(:m,2_i4b)>one
            dfyi = ystat(4_i4b) - ystat(1_i4b)
!
            where ( mask(:m) )
!
!               COMPUTE COVARIANCE COEFFICIENTS BETWEEN XX AND YY.
!
                dfxi(:m) = xstat(:m,4_i4b) - xstat(:m,1_i4b)
                xycor(:m,1_i4b) = xycor(:m,1_i4b) + dfxi(:m)*xycor(:m,4_i4b) +          &
                                  dfyi*xycor(:m,3_i4b) + (dfxi(:m)*dfyi)*xycor(:m,2_i4b)
                xycor(:m,1_i4b) = xycor(:m,1_i4b)/xycor(:m,2_i4b)
!
            elsewhere
!
!               FEWER THAN TWO VALID PAIRS WERE PRESENT,
!               SET THE COVARIANCE TO xymiss .
!
                xycor(:m,1_i4b) = xymiss
!
            end where
!
            if ( present(prob) ) prob(:m) = xymiss
            if ( present(z) )    z(:m)    = xymiss
!
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_cor_miss_rm
! __________________________________
!
    end subroutine comp_cor_miss_rm
!
! =========================================================================================
!
    subroutine comp_cor_miss_rt( x, y, first, last, xstat, ystat, xycor, xymiss,   &
                                 z, prob, ndf_max, cov )
!
! Purpose
! _______
!
!   COMP_COR_MISS computes estimates of means, variances and correlation
!   coefficients (with another data vector YY) from a data tridimensional
!   array XX  possibly containing missing values.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:,:,:)
!            On entry, input subarray containing size(X,3) observations
!            on size(X,1) by size(X,2) variables from the array of data XX for which 
!            basic univariate and bivariate statistics are desired.
!            If all the data are available at once, X can be the full data array XX.
!
!   Y        (INPUT) real(stnd), dimension(:)
!            On entry, input subvector containing size(X,3) observations
!            from a vector of data YY for which correlation coefficient
!            with XX is desired. If all the data are available at once, YY
!            can be the full data vector.
!
!            The size of Y must verify: size(Y) = SIZE(X,3) .
!
!   FIRST    (INPUT) logical(lgl)
!            On entry, if
!
!            - FIRST = true  the current subarray is the first subarray 
!              of the data array XX.
!            - FIRST = false the current subarray is not the first subarray
!              of the data array XX.
!
!   LAST     (INPUT) logical(lgl)
!            On entry, if
!
!            - LAST = true  the current subarray is the last subarray 
!              of the data array XX.
!            - LAST = false the current subarray is not the last subarray
!              of the data array XX.
!
!   XSTAT    (INPUT/OUTPUT) real(stnd), dimension(:,:,4)
!            On entry, after the first call to COMP_COR_MISS (e.g. when FIRST=true),
!            XSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_COR_MISS. XSTAT should not be changed between calls to
!            COMP_COR_MISS.
!
!            On exit, when LAST=true, XSTAT contains the following statistics
!            on all variables:
!
!            - XSTAT(:,:,1) contains the mean values of the data array XX.
!            - XSTAT(:,:,2) contains the variances of the data array XX.
!            - XSTAT(:,:,3) contains the numbers of non-missing observations
!              in the data array XX.
!            - XSTAT(:,:,4) is used as workspace.
!
!            The shape of XSTAT must verify:
!
!            - size(XSTAT,1) = size(X,1) ,
!            - size(XSTAT,2) = size(X,2) ,
!            - size(XSTAT,3) = 4 .
!
!   YSTAT    (INPUT/OUTPUT) real(stnd), dimension(4)
!            On entry, after the first call to COMP_COR_MISS (e.g. when FIRST=true),
!            YSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_COR_MISS. YSTAT should not be changed between calls to
!            COMP_COR_MISS.
!
!            On exit, when LAST=true, YSTAT contains the following statistics:
!
!            - YSTAT(1) contains the mean value of the data vector YY.
!            - YSTAT(2) contains the variance of the data vector YY.
!            - YSTAT(3) contains the the number of non-missing observations
!              in the data vector YY.
!            - YSTAT(4) is used as workspace.
!
!            The size of YSTAT must verify: size(YSTAT) = 4.
!
!   XYCOR    (INPUT/OUTPUT) real(stnd), dimension(:,:,4)
!            On entry, after the first call to COMP_COR_MISS (e.g. when FIRST=true),
!            XYCOR is used as workspace to accumulate quantities on previous calls
!            to COMP_COR_MISS. XYCOR should not be changed between calls to
!            COMP_COR_MISS.
!
!            On exit, when LAST=true, XYCOR contains the following statistics:
!
!            - XYCOR(i,j,1) contains the correlation coefficients between XX(i,j,:) and YY(:).
!            - XYCOR(i,j,2) contains the incidence values between XX(i,j,:) and YY(:). XYCOR(i,j,2)
!              indicates the numbers of valid pairs of observations which were
!              used in the calculation of XYCOR(:,:,1).
!            - XYCOR(:,:,3:4) is used as workspace.
!
!            The shape of XYCOR must verify:
!
!            - size(XYCOR,1) = size(X,1) ,
!            - size(XYCOR,2) = size(X,2) ,
!            - size(XYCOR,3) = 4 .
!
!   XYMISS   (INPUT) real(stnd)
!            On entry, the missing value indicator. Any value in
!            X or Y which is equal to XYMISS is assumed to be missing
!            or invalid.
!            The basic univariate and bivariate statistics are computed
!            on all the observations where X and Y are not missing (see Further Details).
!
!   Z        (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!            On exit, when LAST=true, Z contains the Fisher's Z transformation of
!            XYCOR(:,:,1).
!
!            Z needs to be specified only on the last call to COMP_COR_MISS (e.g. when LAST=true).
!          
!            The shape of Z must verify:
!          
!            - size(Z,1) = size(X,1) ,
!            - size(Z,2) = size(X,2) .
!          
!   PROB     (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!            On exit, when LAST=true, PROB(i,j) gives the probability that the random sample
!            of XYCOR(i,j,2) observation pairs YY(:) and XX(i,j,:) came from a bivariate normal
!            population with a correlation coefficient equal to zero.
!
!            PROB needs to be specified only on the last call to COMP_COR_MISS (e.g. when LAST=true).
!
!            The shape of PROB must verify:
!
!            - size(PROB,1) = size(X,1) ,
!            - size(PROB,2) = size(X,2) .
!          
!   NDF_MAX  (INPUT, OPTIONAL) integer(i4b)
!            On entry, when argument PROB is present, NDF_MAX is used as follows:
!          
!            - If XYCOR(i,j,2)-2 is lower or equal to NDF_MAX, the t_density is integrated
!              for computing PROB(i,j).
!            - If XYCOR(i,j,2)-2 is greater than NDF_MAX, an asymptotic series is used
!              for computing PROB(i,j).
!          
!            The default is 20.
!
!            NDF_MAX needs to be specified only on the last call to COMP_COR_MISS (e.g. when LAST=true).
!          
!   COV      (INPUT, OPTIONAL) logical(lgl)
!            On entry, if COV=true, covariance coefficients between the data array XX and the data
!            vector YY are computed instead of correlation coefficients. If COV=true, Z and PROB are
!            set to XYMISS .
!
!            COV needs to be specified only on the last call to COMP_COR_MISS (e.g. when LAST=true).
!
!
! Further Details
! _______________
!
!   The subroutine computes the basic univariate statistics and the correlation
!   coefficients with only one pass through the data.
!
!   If fewer than two valid observations were present, the statistics are set to XYMISS .
!
!   The means and standard-deviations of XX and YY are computed from all valid data. The
!   correlation coefficients are based on these univariate statistics and on all valid pairs
!   of observations.
!
!   For more details on correlation and regression analysis, see:
!
!   (1) von Storch, H., and Zwiers, F.W., 2002:
!            Statistical Analysis in Climate Research.
!            Cambridge, UK, Chapter 8, 484 pp., ISBN:9780521012300
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, half
    use Char_Constants,    only : stat_error1
    use Num_Constants,     only : lamch
    use Utilities,         only : assert_eq, assert, merror
    use Prob_Procedures,   only : probt
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                                 :: xymiss
    real(stnd), intent(in),    dimension(:)                :: y
    real(stnd), intent(in),    dimension(:,:,:)            :: x
    real(stnd), intent(inout), dimension(:,:,:)            :: xstat, xycor
    real(stnd), intent(inout), dimension(:)                :: ystat
    real(stnd), intent(out),   dimension(:,:),    optional :: prob, z
!
    integer(i4b), intent(in), optional :: ndf_max
!
    logical(lgl), intent(in)            :: first, last
    logical(lgl), intent(in),  optional :: cov
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                       :: dyi, dfyi, cnty, safmin
    real(stnd), dimension(size(x,1)) :: dxi, dfxi
!
    integer(i4b) :: i, j, n, m, p
!
    logical(lgl), dimension(size(x,1))  :: mask
    logical(lgl)                        :: comp_cor, comp_z, comp_prob
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_cor_miss'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq(  int(size(x,3),i4b),     &
                    int(size(y),i4b),       &
                    name_proc )
!
    m = assert_eq(  int(size(x,1),i4b),           &
                    int(size(xstat,1),i4b),       &
                    int(size(xycor,1),i4b),       &
                    name_proc )
!
    p = assert_eq(  int(size(x,2),i4b),           &
                    int(size(xstat,2),i4b),       &
                    int(size(xycor,2),i4b),       &
                    name_proc )
!
    if ( min(m,p)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    call assert( logical(size(xstat,3)==4,lgl),    &
                 logical(size(xycor,3)==4,lgl),    &
                 logical(size(ystat)==4,lgl),      &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBARRAY IS THE FIRST.
!
!       INITIALIZE xstat, ystat AND xycor.
!
        xstat(:m,:p,1_i4b:3_i4b) = zero       
        ystat(1_i4b:3_i4b) = zero
        xycor(:m,:p,1_i4b:4_i4b) = zero
!
        xstat(:m,:p,4_i4b) = xymiss
        ystat(4_i4b) = xymiss
!
    end if
!
!   GET COUNT OF OBSERVATIONS FOR THE PREVIOUS CALL.
!
    cnty = ystat(3_i4b)
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS AND CORRECTED SUMS OF SQUARES AND CROSS-PRODUCTS, 
!       HANDLING THE DATA ONCE.
!
        if ( y(i)/=xymiss ) then
!
            if ( ystat(4_i4b)==xymiss ) ystat(4_i4b) = y(i)
!
            cnty = cnty + one
            dyi  = y(i) - ystat(1_i4b)
            dfyi = dyi/cnty
!
            ystat(1_i4b) = ystat(1_i4b) + dfyi
            ystat(2_i4b) = ystat(2_i4b) + dyi*dyi - dyi*dfyi
!
            dyi = y(i) - ystat(4_i4b)
!
            do j = 1_i4b, p
!
                mask(:m) = x(:m,j,i)/=xymiss
!
                where ( mask(:m) )
!
                    where ( xstat(:m,j,4_i4b)==xymiss ) xstat(:m,j,4_i4b) = x(:m,j,i)
!
!                    xstat(:m,j,4_i4b) = merge( x(:m,j,i), xstat(:m,j,4_i4b), xstat(:m,j,4_i4b)==xymiss )
!
                    xstat(:m,j,3_i4b) = xstat(:m,j,3_i4b) + one
                    dxi(:m) = x(:m,j,i) - xstat(:m,j,1_i4b)
                    dfxi(:m) = dxi(:m)/xstat(:m,j,3_i4b)
                    xstat(:m,j,1_i4b) = xstat(:m,j,1_i4b) + dfxi(:m)
                    xstat(:m,j,2_i4b) = xstat(:m,j,2_i4b) + dxi(:m)*dxi(:m) - dxi(:m)*dfxi(:m)
!
                    dxi(:m) = x(:m,j,i) - xstat(:m,j,4_i4b)
                    xycor(:m,j,1_i4b) = xycor(:m,j,1_i4b) + dxi(:m)*dyi
                    xycor(:m,j,2_i4b) = xycor(:m,j,2_i4b) + one
                    xycor(:m,j,3_i4b) = xycor(:m,j,3_i4b) + dxi(:m)
                    xycor(:m,j,4_i4b) = xycor(:m,j,4_i4b) + dyi
!
                end where
!
            end do
!
        else
!
            do j = 1_i4b, p
!
                mask(:m) = x(:m,j,i)/=xymiss
!
                where ( mask(:m) )
!
                    where ( xstat(:m,j,4_i4b)==xymiss ) xstat(:m,j,4_i4b) = x(:m,j,i)
!
!                    xstat(:m,j,4_i4b) = merge( x(:m,j,i), xstat(:m,j,4_i4b), xstat(:m,j,4_i4b)==xymiss )
!
                    xstat(:m,j,3_i4b) = xstat(:m,j,3_i4b) + one
                    dxi(:m) = x(:m,j,i) - xstat(:m,j,1_i4b)
                    dfxi(:m) = dxi(:m)/xstat(:m,j,3_i4b)
                    xstat(:m,j,1_i4b) = xstat(:m,j,1_i4b) + dfxi(:m)
                    xstat(:m,j,2_i4b) = xstat(:m,j,2_i4b) + dxi(:m)*dxi(:m) - dxi(:m)*dfxi(:m)
!
                end where
!
            end do
!
        end if
!
    end do
!
!   SAVE COUNT OF OBSERVATIONS FOR THE NEXT CALL.
!
    ystat(3_i4b) = cnty
!
    if ( last ) then
!
!       THE CURRENT SUBARRAY IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
!       TEST OPTIONAL ARGUMENTS.
!
        comp_prob = present( prob )
!
        if ( comp_prob ) call assert( logical(size(prob,1)==m,lgl),     &
                                      logical(size(prob,2)==p,lgl),     &
                                      name_proc )
!
        comp_z = present( z )
!
        if ( comp_z ) call assert( logical(size(z,1)==m,lgl),     &
                                   logical(size(z,2)==p,lgl),     &
                                   name_proc )
!
        comp_cor = true
!
        if ( present(cov) ) then
            comp_cor = .not.cov
        end if
!
        if ( cnty>one ) then
!
!           COMPUTE VARIANCE OF YY.
!
            dfyi = ystat(2_i4b)/cnty
            ystat(2_i4b) = dfyi
!
            dyi = ystat(4_i4b) - ystat(1_i4b)
!
        else
!
!           FEWER THAN TWO VALID OBSERVATIONS, SET THE PERTINENT STATISTICS TO xymiss .
!
            dfyi = zero
            if ( cnty<one ) ystat(1_i4b:2_i4b) = xymiss
!
            dyi = zero
!
        end if
!
        if ( comp_cor ) then
!
            safmin = lamch( 's' )
!
        end if
!
#ifdef _OPENMP
        i = omp_get_num_procs()
        j = omp_get_max_threads()
        ompparallel = omp_in_parallel()
        ompnested   = omp_get_nested() .and. ompparallel
        test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                      (m*p)>=omp_limit                    .and.      &
                      i>1_i4b                             .and.      &
                      j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(DYNAMIC,1)  &
!$OMP            ,PRIVATE(j,i,mask,dfxi,dxi)         &
!$OMP            ,SHARED(safmin,dfyi,dyi,ndf_max,m,p,comp_cor,comp_z,comp_prob)
!
        do j = 1_i4b, p
!
            mask(:m) = xstat(:m,j,3_i4b)>one
!
            where ( mask(:m) )
!
!               COMPUTE VARIANCES OF XX.
!
                dfxi(:m) = xstat(:m,j,2_i4b)/xstat(:m,j,3_i4b)
                xstat(:m,j,2_i4b) = dfxi(:m)
!
            elsewhere
!
!               FEWER THAN TWO VALID OBSERVATIONS, SET THE PERTINENT STATISTICS TO xymiss .
!
                dfxi(:m) = zero
!
                where ( xstat(:m,j,3_i4b)<one )
                    xstat(:m,j,1_i4b) = xymiss
                    xstat(:m,j,2_i4b) = xymiss
                elsewhere
                    xstat(:m,j,2_i4b) = zero
                end where
!
!                xstat(:m,j,1_i4b) = merge( xymiss, xstat(:m,j,1_i4b), xstat(:m,j,3_i4b)<one )
!                xstat(:m,j,2_i4b) = merge( xymiss, zero, xstat(:m,j,3_i4b)<one )
!
            end where
!
            if ( comp_cor ) then
!
                dxi(:m) = dfxi(:m)*dfyi
                mask(:m) = xycor(:m,j,2_i4b)>one .and. dxi(:m)>safmin
!
                where ( mask(:m) )
!
!                   COMPUTE CORRELATION COEFFICIENTS BETWEEN XX AND YY.
!
                    dfxi(:m) = xstat(:m,j,4_i4b) - xstat(:m,j,1_i4b)
                    xycor(:m,j,1_i4b) = xycor(:m,j,1_i4b) + dfxi(:m)*xycor(:m,j,4_i4b) +           &
                                        dyi*xycor(:m,j,3_i4b) + (dfxi(:m)*dyi)*xycor(:m,j,2_i4b)
                    dfxi(:m) = xycor(:m,j,1_i4b)/( xycor(:m,j,2_i4b)*sqrt( dxi(:m) ) )
!
!                   CHECK IF CORRELATION IS BOUNDED BY -1 AND 1.
!
                    where ( dfxi(:m)>one )
                        dxi(:m) = one
                    elsewhere
                        dxi(:m) = dfxi(:m)
                    end where
!
!                    dxi(:m) = merge( one, dfxi(:m), dfxi(:m)>one )
!
                    where ( dxi(:m)<-one )
                        xycor(:m,j,1_i4b) = -one
                    elsewhere
                        xycor(:m,j,1_i4b) = dxi(:m)
                    end where
!
!                    xycor(:m,j,1_i4b) = merge( -one, dxi(:m), dxi(:m)<-one )
!
                elsewhere
!
!                   THE OBSERVATIONS ON XX OR YY WERE CONSTANT OR FEWER THAN
!                   TWO VALID PAIRS WERE PRESENT, SET THE CORRELATION TO xymiss .
!
                    xycor(:m,j,1_i4b) = xymiss
!
                end where
!
!               OUTPUT FISHER'S z TRANSFORMATION, IF REQUIRED.
!
                if ( comp_z ) then
!
                    where ( mask(:m) )
                        z(:m,j) = half*log(( (one+xycor(:m,j,1_i4b)) + safmin )/( (one-xycor(:m,j,1_i4b)) + safmin ))
                    elsewhere
                        z(:m,j) = xymiss
                    end where
!
                end if
!
                if ( comp_prob ) then
!
!                   OUTPUT SIGNIFICANCE PROBABILITIES.
!
                    mask(:m) = mask(:m) .and. xycor(:m,j,2_i4b)>two
!
                    where ( mask(:m) )
                        mask(:m) = abs( xycor(:m,j,1_i4b) )/=one
                        prob(:m,j) = zero
                    elsewhere
                        prob(:m,j) = xymiss
                    end where
!
!                   COMPUTE STUDENT STATISTIC.
!
                    where ( mask(:m) )
                        dfxi(:m) = xycor(:m,j,2_i4b) - two
                        dxi(:m) = sqrt( dfxi(:m) )*abs( xycor(:m,j,1_i4b) )/sqrt( one - xycor(:m,j,1_i4b)**2 )
                    end where
!
                    do i = 1_i4b, m
!
                        if ( mask(i) ) then
                            prob(i,j) = two*probt( dxi(i), int( dfxi(i), i4b ), true, ndf_max=ndf_max )
                        end if
!
                    end do
!
                end if
!
            else
!
                mask(:m) = xycor(:m,j,2_i4b)>one
!
                where ( mask(:m) )
!
!                   COMPUTE COVARIANCE COEFFICIENTS BETWEEN XX AND YY.
!
                    dfxi(:m) = xstat(:m,j,4_i4b) - xstat(:m,j,1_i4b)
                    xycor(:m,j,1_i4b) = xycor(:m,j,1_i4b) + dfxi(:m)*xycor(:m,j,4_i4b) +           &
                                        dyi*xycor(:m,j,3_i4b) + (dfxi(:m)*dyi)*xycor(:m,j,2_i4b)
                    xycor(:m,j,1_i4b) = xycor(:m,j,1_i4b)/xycor(:m,j,2_i4b)
!
                elsewhere
!
!                   FEWER THAN TWO VALID PAIRS WERE PRESENT, 
!                   SET THE COVARIANCE TO xymiss .
!
                    xycor(:m,j,1_i4b) = xymiss
!
                end where
!
                if ( comp_prob ) prob(:m,j) = xymiss
                if ( comp_z )    z(:m,j)    = xymiss
!
            end if
!
        end do
!
!$OMP END PARALLEL DO
!
    end if
!
!
! END OF SUBROUTINE comp_cor_miss_rt
! __________________________________
!
    end subroutine comp_cor_miss_rt
!
! =========================================================================================
!
    subroutine comp_cor_miss_rm_rm( x, y, first, last, xstat, ystat, xycor, xymiss, &
                                    dimvar, dimvary, z, prob, ndf_max, cov )
!
! Purpose
! _______
!
!   COMP_COR_MISS computes estimates of means, variances and correlation
!   coefficients between two data matrices YY and XX possibly containing
!   missing values.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:,:)
!            On entry, input submatrix containing size(X,3-DIMVAR)
!            observations on size(X,DIMVAR) variables from the matrix of data XX
!            for which basic univariate  and bivariate statistics are desired. By
!            default, DIMVAR is equal to 1. See description of optional DIMVAR
!            argument for details.
!            If all the data are available at once, X can be the full 
!            data matrix XX.
!
!   Y        (INPUT) real(stnd), dimension(:,:)
!            On entry, input submatrix containing size(Y,3-DIMVARY)
!            observations on size(Y,DIMVARY) variables from the matrix of data YY
!            for which basic univariate  and bivariate statistics are desired. By
!            default, DIMVARY is equal to 1. See description of optional DIMVARY
!            argument for details.
!            If all the data are available at once, Y can be the full 
!            data matrix YY.
!
!            The shape of Y must verify: size(Y,3-DIMVARY)  = size(X,3-DIMVAR) .
!
!   FIRST    (INPUT) logical(lgl)
!            On entry, if:
!
!            - FIRST = true  the current submatrices are the first submatrices 
!              of the data matrices XX and YY.
!            - FIRST = false the current submatrices are not the first submatrices
!              of the data matrices XX and YY.
!
!   LAST     (INPUT) logical(lgl)
!            On entry, if:
!
!            - LAST = true  the current submatrices are the last submatrices 
!              of the data matrices XX and YY.
!            - LAST = false the current submatrix are not the last submatrices
!              of the data matrices XX and YY.
!
!   XSTAT    (INPUT/OUTPUT) real(stnd), dimension(:,4)
!            On entry, after the first call to COMP_COR_MISS (e.g. when FIRST=true),
!            XSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_COR_MISS. XSTAT should not be changed between calls to
!            COMP_COR_MISS.
!
!            On exit, when LAST=true, XSTAT contains the following statistics
!            on all variables from the XX matrix:
!
!            - XSTAT(:,1) contains the mean values of the data matrix XX.
!            - XSTAT(:,2) contains the variances of the data matrix XX.
!            - XSTAT(:,3) contains the the numbers of non-missing observations
!              in the data matrix XX.
!            - XSTAT(:,4) is used as workspace.
!
!            The shape of XSTAT must verify:
!
!            - size(XSTAT,1) = size(X,DIMVAR) ,
!            - size(XSTAT,2) = 4 .
!
!   YSTAT    (INPUT/OUTPUT) real(stnd), dimension(:,4)
!            On entry, after the first call to COMP_COR_MISS (e.g. when FIRST=true),
!            YSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_COR_MISS. YSTAT should not be changed between calls to
!            COMP_COR_MISS.
!
!            On exit, when LAST=true, YSTAT contains the following statistics
!            on all variables from the YY matrix:
!
!            - YSTAT(:,1) contains the mean values of the data matrix YY.
!            - YSTAT(:,2) contains the variances of the data matrix YY.
!            - YSTAT(:,3) contains the the numbers of non-missing observations
!              in the data matrix YY.
!            - YSTAT(:,4) is used as workspace.
!
!            The shape of YSTAT must verify:
!
!            - size(YSTAT,1) = size(Y,DIMVARY) ,
!            - size(YSTAT,2) = 4 .
!
!   XYCOR    (INPUT/OUTPUT) real(stnd), dimension(:,:,4)
!            On entry, after the first call to COMP_COR_MISS (e.g. when FIRST=true),
!            XYCOR is used as workspace to accumulate quantities on previous calls
!            to COMP_COR_MISS. XYCOR should not be changed between calls to
!            COMP_COR_MISS.
!
!            On exit, when LAST=true, XYCOR contains the following statistics:
!
!            - XYCOR(i,j,1)   contains the correlation coefficients between XX(i,:) and YY(j,:)
!              ( XX(:,i) and YY(:,j) if DIMVAR=2 and DIMVARY=2 ).
!            - XYCOR(i,j,2)   contains the incidence values between XX(i,:) and YY(j,:) ( XX(:,i) and 
!              YY(:,j) if DIMVAR=2  and DIMVARY=2). XYCOR(i,j,2) indicates the numbers
!              of non-missing pairs of  observations which were used in the calculation
!              of XYCOR(i,j,1).
!            - XYCOR(:,:,3:4) is used as workspace.
!
!            The shape of XYCOR must verify:
!
!            - size(XYCOR,1) = size(X,DIMVAR) , 
!            - size(XYCOR,2) = size(Y,DIMVARY) ,
!            - size(XYCOR,3) = 4 .
!
!   XYMISS   (INPUT) real(stnd)
!            On entry, the missing value indicator. Any value in
!            X or Y which is equal to XYMISS is assumed to be missing
!            or invalid.
!            The basic univariate and bivariate statistics are computed
!            on all the observations where X and Y are not missing (see Further Details).
!
!   DIMVAR   (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVAR is present, DIMVAR is used as follows:
!
!            - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!              on size(X,1) variables.
!            - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!              on size(X,2) variables, respectively.
!
!            The default is DIMVAR = 1.
!
!   DIMVARY  (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVARY is present, DIMVARY is used as follows:
!
!            - DIMVARY = 1, the input submatrix Y contains size(Y,2) observations
!              on size(Y,1) variables.
!            - DIMVARY = 2, the input submatrix Y contains size(Y,1) observations
!              on size(Y,2) variables, respectively.
!
!            The default is DIMVARY = 1.
!            
!   Z        (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!            On exit, when LAST=true, Z contains the Fisher's Z transformation of
!            XYCOR.
!
!            Z needs to be specified only on the last call to COMP_COR_MISS (e.g. when LAST=true).
!          
!            The shape of Z must verify:
!          
!            - size(Z,1) = size(X,DIMVAR) ,
!            - size(Z,2) = size(Y,DIMVARY) .
!
!   PROB     (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!            On exit, when LAST=true, PROB(i,j) gives the probability that the random sample
!            of XYCOR(i,j,2) observation pairs XX(i,:) and YY(j,:) ( XX(:,i) and YY(:,j) if DIMVAR=2
!            and DIMVARY=2 ) came from a bivariate normal population with a correlation
!            coefficient equal to zero.
!
!            PROB needs to be specified only on the last call to COMP_COR_MISS (e.g. when LAST=true).
!          
!            The shape of PROB must verify:
!          
!            - size(PROB,1) = size(X,DIMVAR) ,
!            - size(PROB,2) = size(Y,DIMVARY) .
!
!   NDF_MAX  (INPUT, OPTIONAL) integer(i4b)
!            On entry, when argument PROB is present, NDF_MAX is used as follows:
!
!            - If XYCOR(i,j,2)-2 is lower or equal to NDF_MAX, the t_density is integrated
!              for computing PROB(i,j).
!            - If XYCOR(i,j,2)-2 is greater than NDF_MAX, an asymptotic series is used
!              for computing PROB(i,j).
!
!            The default is 20.
!
!            NDF_MAX needs to be specified only on the last call to COMP_COR_MISS (e.g. when LAST=true).
!          
!   COV      (INPUT, OPTIONAL) logical(lgl)
!            On entry, if COV=true, covariance coefficients between the data matrices XX and
!            YY are computed instead of correlation coefficients. If COV=true, Z and PROB are
!            set to XYMISS .
!
!            COV needs to be specified only on the last call to COMP_COR_MISS (e.g. when LAST=true).
!
!
! Further Details
! _______________
!
!   The subroutine computes the basic univariate statistics and the correlation
!   coefficients with only one pass through the data.
!
!   If fewer than two valid observations were present, the statistics are set to XYMISS .
!
!   The means and standard-deviations of XX and YY are computed from all valid data. The
!   correlation coefficients are based on these univariate statistics and on all valid pairs
!   of observations.
!
!   For more details on correlation and regression analysis, see
!
!   (1) von Storch, H., and Zwiers, F.W., 2002:
!            Statistical Analysis in Climate Research.
!            Cambridge, UK, Chapter 8, 484 pp., ISBN:9780521012300
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true, false
    use Reals_Constants,   only : zero, one, two, half
    use Num_Constants,     only : lamch
    use Char_Constants,    only : stat_error1, stat_error5
    use Utilities,         only : assert_eq, assert, merror
    use Prob_Procedures,   only : probt
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                               :: xymiss
    real(stnd), intent(in),    dimension(:,:)            :: x, y
    real(stnd), intent(inout), dimension(:,:)            :: xstat, ystat
    real(stnd), intent(inout), dimension(:,:,:)          :: xycor
    real(stnd), intent(out),   dimension(:,:),  optional :: prob, z
!
    integer(i4b), intent(in), optional :: dimvar, dimvary, ndf_max
!
    logical(lgl), intent(in)            :: first, last
    logical(lgl), intent(in),  optional :: cov
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                              :: safmin
    real(stnd), dimension(size(xycor,1))                    :: xi, dxi, dfxi
    real(stnd), dimension(size(xycor,2))                    :: yi, dfyi
    real(stnd), dimension(max(size(xycor,1),size(xycor,2))) :: dyi
!
    integer(i4b) :: i, j, n, m, p
!
    logical(lgl)                            :: dimx_is_one, dimy_is_one, comp_cor, comp_prob, comp_z
    logical(lgl), dimension(size(xycor,1))  :: maskx
    logical(lgl), dimension(size(xycor,2))  :: masky
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_cor_miss'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dimx_is_one = i==1_i4b
!
    if ( present(dimvary) ) then
!
        if ( dimvary==1_i4b .or. dimvary==2_i4b ) then
            j = dimvary
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        j = 1_i4b
!
    end if
!
    dimy_is_one = j==1_i4b
!
    n = assert_eq(  int(size(x,3-int(i)),i4b),  &
                    int(size(y,3-int(j)),i4b),  &
                    name_proc )
!
    m = assert_eq(  int(size(x,int(i)),i4b),      &
                    int(size(xstat,1),i4b),       &
                    int(size(xycor,1),i4b),       &
                    name_proc )
!
    p = assert_eq(  int(size(y,int(j)),i4b),      &
                    int(size(ystat,1),i4b),       &
                    int(size(xycor,2),i4b),       &
                    name_proc )
!
    if ( m<=0_i4b .or. p<=0_i4b ) call merror( name_proc//stat_error1 )
!
    call assert( logical(size(xstat,2)==4,lgl),    &
                 logical(size(ystat,2)==4,lgl),    &
                 logical(size(xycor,3)==4,lgl),    &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBMATRICES ARE THE FIRST.
!
!       INITIALIZE xstat, ystat AND xycor.
!
        xstat(:m,1_i4b:3_i4b) = zero       
        ystat(:p,1_i4b:3_i4b) = zero
        xycor(:m,:p,1_i4b:4_i4b) = zero
!
        xstat(:m,4_i4b) = xymiss
        ystat(:p,4_i4b) = xymiss
!
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  (m*p)>=omp_limit                    .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)      &
!$OMP         ,PRIVATE(i,j)
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS AND CORRECTED SUMS OF SQUARES AND CROSS-PRODUCTS, 
!       HANDLING THE DATA ONCE.
!
!$OMP SINGLE
!
        if ( dimx_is_one ) then
            xi(:m) = x(:m,i)
        else
            xi(:m) = x(i,:m)
        end if
!
        maskx(:m) = xi(:m)/=xymiss
#ifdef _WHERE
        do j = 1_i4b, m
!
            if ( maskx(j) ) then
!
                if ( xstat(j,4_i4b)==xymiss ) then
                    xstat(j,4_i4b) = xi(j)
                end if
!
!                xstat(j,4_i4b) = merge( xi(j), xstat(j,4_i4b), xstat(j,4_i4b)==xymiss )
!
                xstat(j,3_i4b) = xstat(j,3_i4b) + one
                dxi(j) = xi(j) - xstat(j,1_i4b)
                dfxi(j) = dxi(j)/xstat(j,3_i4b)
                xstat(j,1_i4b) = xstat(j,1_i4b) + dfxi(j)
                xstat(j,2_i4b) = xstat(j,2_i4b) + dxi(j)*dxi(j) - dxi(j)*dfxi(j)
                dxi(j) = xi(j) - xstat(j,4_i4b)
!
            end if
!
        end do
#else
        where ( maskx(:m) )
!
            where ( xstat(:m,4_i4b)==xymiss ) xstat(:m,4_i4b) = xi(:m)
!
!            xstat(:m,4_i4b) = merge( xi(:m), xstat(:m,4_i4b), xstat(:m,4_i4b)==xymiss )
!
            xstat(:m,3_i4b) = xstat(:m,3_i4b) + one
            dxi(:m) = xi(:m) - xstat(:m,1_i4b)
            dfxi(:m) = dxi(:m)/xstat(:m,3_i4b)
            xstat(:m,1_i4b) = xstat(:m,1_i4b) + dfxi(:m)
            xstat(:m,2_i4b) = xstat(:m,2_i4b) + dxi(:m)*dxi(:m) - dxi(:m)*dfxi(:m)
            dxi(:m) = xi(:m) - xstat(:m,4_i4b)
!
        end where
#endif
!
        if ( dimy_is_one ) then
            yi(:p) = y(:p,i)
        else
            yi(:p) = y(i,:p)
        end if
!
        masky(:p) = yi(:p)/=xymiss
!#ifdef _WHERE
!        do j = 1_i4b, p
!            if ( masky(j) ) then
!
!                ystat(j,4_i4b) = merge( yi(j), ystat(j,4_i4b), ystat(j,4_i4b)==xymiss )
!
!                ystat(j,3_i4b) = ystat(j,3_i4b) + one
!                dyi(j) = yi(j) - ystat(j,1_i4b)
!                dfyi(j) = dyi(j)/ystat(j,3_i4b)
!                ystat(j,1_i4b) = ystat(j,1_i4b) + dfyi(j)
!                ystat(j,2_i4b) = ystat(j,2_i4b) + dyi(j)*dyi(j) - dyi(j)*dfyi(j)
!                dyi(j) = yi(j) - ystat(j,4_i4b)
!            end if
!        end do
!#else
!        where ( masky(:p) )
!
!            ystat(:p,4_i4b) = merge( yi(:p), ystat(:p,4_i4b), ystat(:p,4_i4b)==xymiss )
!
!            ystat(:p,3_i4b) = ystat(:p,3_i4b) + one
!            dyi(:p) = yi(:p) - ystat(:p,1_i4b)
!            dfyi(:p) = dyi(:p)/ystat(:p,3_i4b)
!            ystat(:p,1_i4b) = ystat(:p,1_i4b) + dfyi(:p)
!            ystat(:p,2_i4b) = ystat(:p,2_i4b) + dyi(:p)*dyi(:p) - dyi(:p)*dfyi(:p)
!            dyi(:p) = yi(:p) - ystat(:p,4_i4b)
!        end where
!#endif
!
!$OMP END SINGLE
!
!$OMP DO SCHEDULE(STATIC) 
!
        do j = 1_i4b, p
!
            if ( masky(j) ) then
!
                if ( ystat(j,4_i4b)==xymiss ) then
                    ystat(j,4_i4b) = yi(j)
                end if
!
!                ystat(j,4_i4b) = merge( yi(j), ystat(j,4_i4b), ystat(j,4_i4b)==xymiss )
!
                ystat(j,3_i4b) = ystat(j,3_i4b) + one
                dyi(j)         = yi(j) - ystat(j,1_i4b)
                dfyi(j)        = dyi(j)/ystat(j,3_i4b)
                ystat(j,1_i4b) = ystat(j,1_i4b) + dfyi(j)
                ystat(j,2_i4b) = ystat(j,2_i4b) + dyi(j)*dyi(j) - dyi(j)*dfyi(j)
                dyi(j)         = yi(j) - ystat(j,4_i4b)
!
                where ( maskx(:m) )
!
                    xycor(:m,j,1_i4b) = xycor(:m,j,1_i4b) + dxi(:m)*dyi(j)
                    xycor(:m,j,2_i4b) = xycor(:m,j,2_i4b) + one
                    xycor(:m,j,3_i4b) = xycor(:m,j,3_i4b) + dxi(:m)
                    xycor(:m,j,4_i4b) = xycor(:m,j,4_i4b) + dyi(j)
!
                end where
!
            end if
!
        end do
!
!$OMP END DO
!
    end do
!
!$OMP END PARALLEL
!
    if ( last ) then
!
!       THE CURRENT SUBMATRICES ARE THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
!       TEST OPTIONAL ARGUMENTS.
!
        if ( present( prob ) ) then
!
            call assert( logical(size(prob,1)==m,lgl),   &
                         logical(size(prob,2)==p,lgl),   &
                         name_proc )
            comp_prob = true
!
        else
!
            comp_prob = false
!
        end if
!
        if ( present( z ) ) then
!
            call assert( logical(size(z,1)==m,lgl),   &
                         logical(size(z,2)==p,lgl),   &
                         name_proc )
            comp_z = true
!
        else
!
            comp_z = false
!
        end if
!
        comp_cor = true
!
        if ( present(cov) ) then
!
            comp_cor = .not.cov
!
        end if
!
        if ( comp_cor ) then
!
            safmin = lamch( 's' )
!
        end if
!
!$OMP PARALLEL IF(test_par)               &
!$OMP         ,PRIVATE(i,j,dxi,dyi,maskx)
!
!
!$OMP SECTIONS
!
!$OMP SECTION
!
        maskx(:m) = xstat(:m,3_i4b)>one
#ifdef _WHERE
        do j = 1_i4b, m
!
            if ( maskx(j) ) then
!
!               COMPUTE VARIANCES OF XX.
!
                dfxi(j) = xstat(j,2_i4b)/xstat(j,3_i4b)
                xstat(j,2_i4b) = dfxi(j)
                xi(j) = xstat(j,4_i4b) - xstat(j,1_i4b)
!
            else
!
!               FEWER THAN TWO VALID OBSERVATIONS, SET THE PERTINENT STATISTICS TO xymiss .
!
                dfxi(j) = zero
!
                if ( xstat(j,3_i4b)<one ) then
                    xstat(j,1_i4b:2_i4b) = xymiss
                else
                    xstat(j,2_i4b) = zero
                end if
!
!                xstat(j,1_i4b) = merge( xymiss, xstat(j,1_i4b), xstat(j,3_i4b)<one )
!                xstat(j,2_i4b) = merge( xymiss, zero, xstat(j,3_i4b)<one )
!
            end if
!
        end do
#else
        where ( maskx(:m) )
!
!           COMPUTE VARIANCES OF XX.
!
            dfxi(:m) = xstat(:m,2_i4b)/xstat(:m,3_i4b)
            xstat(:m,2_i4b) = dfxi(:m)
            xi(:m) = xstat(:m,4_i4b) - xstat(:m,1_i4b)
!
        elsewhere
!
!           FEWER THAN TWO VALID OBSERVATIONS, SET THE PERTINENT STATISTICS TO xymiss .
!
            dfxi(:m) = zero
!
            where ( xstat(:m,3_i4b)<one )
                xstat(:m,1_i4b) = xymiss
                xstat(:m,2_i4b) = xymiss
            elsewhere
                xstat(:m,2_i4b) = zero
            end where
!
!            xstat(:m,1_i4b) = merge( xymiss, xstat(:m,1_i4b), xstat(:m,3_i4b)<one )
!            xstat(:m,2_i4b) = merge( xymiss, zero, xstat(:m,3_i4b)<one )
!
        end where
#endif
!
!$OMP SECTION
!
        masky(:p) = ystat(:p,3_i4b)>one
#ifdef _WHERE
        do j = 1_i4b, p
!
            if ( masky(j) ) then
!
!               COMPUTE VARIANCES OF YY.
!
                dfyi(j) = ystat(j,2_i4b)/ystat(j,3_i4b)
                ystat(j,2_i4b) = dfyi(j)
                yi(j) = ystat(j,4_i4b) - ystat(j,1_i4b)
!
            else
!
!               FEWER THAN TWO VALID OBSERVATIONS, SET THE PERTINENT STATISTICS TO xymiss .
!
                dfyi(j) = zero
                if ( ystat(j,3_i4b)<one ) then
                    ystat(j,1_i4b:2_i4b) = xymiss
                else
                    ystat(j,2_i4b) = zero
                end if
!
!                ystat(j,1_i4b) = merge( xymiss, ystat(j,1_i4b), ystat(j,3_i4b)<one )
!                ystat(j,2_i4b) = merge( xymiss, zero, ystat(j,3_i4b)<one )
!
             end if
!
        end do
#else
        where ( masky(:p) )
!
!           COMPUTE VARIANCES OF YY.
!
            dfyi(:p) = ystat(:p,2_i4b)/ystat(:p,3_i4b)
            ystat(:p,2_i4b) = dfyi(:p)
            yi(:p) = ystat(:p,4_i4b) - ystat(:p,1_i4b)
!
        elsewhere
!
!           FEWER THAN TWO VALID OBSERVATIONS, SET THE PERTINENT STATISTICS TO xymiss .
!
            dfyi(:p) = zero
            where ( ystat(:p,3_i4b)<one )
                ystat(:p,1_i4b) = xymiss
                ystat(:p,2_i4b) = xymiss
            elsewhere
                ystat(:p,2_i4b) = zero
            end where
!
!            ystat(:p,1_i4b) = merge( xymiss, ystat(:p,1_i4b), ystat(:p,3_i4b)<one )
!            ystat(:p,2_i4b) = merge( xymiss, zero, ystat(:p,3_i4b)<one )
!
        end where
#endif
!
!$OMP END SECTIONS
!
        if ( comp_cor ) then
!
!$OMP DO SCHEDULE(STATIC) 
!
            do j = 1_i4b, p
!
                if ( masky(j) ) then
!
                    dxi(:m) = dfxi(:m)*dfyi(j)
                    maskx(:m) = xycor(:m,j,2_i4b)>one .and. dxi(:m)>safmin
!
                    where ( maskx(:m) )
!
!                       COMPUTE CORRELATION COEFFICIENTS BETWEEN XX AND YY.
!
                        xycor(:m,j,1_i4b) = xycor(:m,j,1_i4b) + xi(:m)*xycor(:m,j,4_i4b) +               &
                                          yi(j)*xycor(:m,j,3_i4b) + (xi(:m)*yi(j))*xycor(:m,j,2_i4b)
                        dyi(:m) = xycor(:m,j,1_i4b)/( xycor(:m,j,2_i4b)*sqrt( dxi(:m) ) )
!
                        where ( dyi(:m)>one )
                            dxi(:m) = one
                        elsewhere
                            dxi(:m) = dyi(:m)
                        end where
!
!                        dxi(:m) = merge( one, dyi(:m), dyi(:m)>one )
!
                        where ( dxi(:m)<-one )
                            xycor(:m,j,1_i4b) = -one
                        elsewhere
                            xycor(:m,j,1_i4b) = dxi(:m)
                        end where
!
!                        xycor(:m,j,1_i4b) = merge( -one, dxi(:m), dxi(:m)<-one )
!
                    elsewhere
!
!                       THE OBSERVATIONS ON XX or YY WERE CONSTANT OR FEWER THAN
!                       TWO VALID PAIRS WERE PRESENT, SET THE CORRELATION TO xymiss .
!
                        xycor(:m,j,1_i4b) = xymiss
!
                    end where
!
                else
!
!                   FEWER THAN TWO VALID OBSERVATIONS WERE PRESENT ON YY,
!                   SET THE CORRELATION TO xymiss .
!
                    xycor(:m,j,1_i4b) = xymiss
!
                end if
!
            end do
!
!$OMP END DO
!
            if ( comp_z ) then
!
!               OUTPUT FISHER'S z TRANSFORMATION, IF REQUIRED.
!
!$OMP DO SCHEDULE(STATIC)
!
                do j = 1_i4b, p
!
                    maskx(:m) = xycor(:m,j,1_i4b)/=xymiss
!
                    where ( maskx(:m) )
                        z(:m,j) = half*log(( (one+xycor(:m,j,1_i4b)) + safmin )/( (one-xycor(:m,j,1_i4b)) + safmin ))
                    elsewhere
                        z(:m,j) = xymiss
                    end where
!
                end do
!
!$OMP END DO
!
            end if
!
            if ( comp_prob ) then
!
!               OUTPUT SIGNIFICANCE PROBABILITIES.
!
!$OMP DO SCHEDULE(STATIC)
!
                do j = 1_i4b, p
!
                    if ( masky(j) ) then
!
                        maskx(:m) = xycor(:m,j,1_i4b)/=xymiss .and. xycor(:m,j,2_i4b)>two
!
                        where ( maskx(:m) )
                            maskx(:m) = abs( xycor(:m,j,1_i4b) )/=one
                            prob(:m,j) = zero
                        elsewhere
                            prob(:m,j) = xymiss
                        end where
!
                        where ( maskx(:m) )
                            dyi(:m) = xycor(:m,j,2_i4b) - two
                            dxi(:m) = sqrt( dyi(:m) )*abs( xycor(:m,j,1_i4b) )/sqrt( one - xycor(:m,j,1_i4b)**2 )
                        end where
!
                        do i = 1_i4b, m
!
                            if ( maskx(i) ) then
                                prob(i,j) = two*probt( dxi(i), int( dyi(i), i4b ), true, ndf_max=ndf_max )
                            end if
!
                        end do
!
                    else
!
                        prob(:m,j) = xymiss
!
                    end if
!
                end do
!
!$OMP END DO
!
            end if
!
        else
!
!$OMP DO SCHEDULE(STATIC) 
!
            do j = 1_i4b, p
!
                if ( masky(j) ) then
!
                    maskx(:m) = xycor(:m,j,2_i4b)>one
!
                    where ( maskx(:m) )
!
!                       COMPUTE COVARIANCE COEFFICIENTS BETWEEN XX AND YY.
!
                        xycor(:m,j,1_i4b) = xycor(:m,j,1_i4b) + xi(:m)*xycor(:m,j,4_i4b) +               &
                                          yi(j)*xycor(:m,j,3_i4b) + (xi(:m)*yi(j))*xycor(:m,j,2_i4b)
                        dxi(:m) = xycor(:m,j,1_i4b)/xycor(:m,j,2_i4b)
                        xycor(:m,j,1_i4b) = dxi(:m)
!
                    elsewhere
!
!                       THE OBSERVATIONS ON XX or YY WERE CONSTANT OR FEWER THAN
!                       TWO VALID PAIRS WERE PRESENT, SET THE COVARIANCE TO xymiss .
!
                        xycor(:m,j,1_i4b) = xymiss
!
                    end where
!
                else
!
!                   FEWER THAN TWO VALID OBSERVATIONS WERE PRESENT ON YY,
!                   SET THE COVARIANCE TO xymiss .
!
                    xycor(:m,j,1_i4b) = xymiss
!
                end if
!
            end do
!
!$OMP END DO NOWAIT
!
!$OMP SINGLE
!
            if ( comp_prob ) prob(:m,:p) = xymiss
            if ( comp_z )    z(:m,:p)    = xymiss
!
!$OMP END SINGLE
!
        end if
!
!$OMP END PARALLEL
!
    end if
!
!
! END OF SUBROUTINE comp_cor_miss_rm_rm
! _____________________________________
!
    end subroutine comp_cor_miss_rm_rm
!
! =========================================================================================
!
    subroutine comp_cor_miss2_rv( x, y, first, last, xstat, ystat, xycor, xyn, xymiss,  &
                                  z, prob, ndf_max )
!
! Purpose
! _______
!
!   COMP_COR_MISS2 computes estimates of mean, variance and correlation
!   coefficient from two data vectors XX and YY possibly containing
!   missing values.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:)
!            On entry, input subvector containing size(X) observations
!            from the vector of data XX for which basic univariate and bivariate
!            statistics are desired. If all the data are available at once, X
!            can be the full data vector XX.
!
!   Y        (INPUT) real(stnd), dimension(:)
!            On entry, input subvector containing size(X) observations
!            from another vector of data YY for which correlation coefficient
!            with XX is desired. If all the data are available at once, YY
!            can be the full data vector.
!
!            The size of Y must verify: size(Y) = size(X) .
!
!   FIRST    (INPUT) logical(lgl)
!            On entry, if
!
!            - FIRST = true  the current subvector is the first subvector 
!              of the data vector XX (or YY).
!            - FIRST = false the current subvector is not the first subvector
!              of the data vector XX (or YY).
!
!   LAST     (INPUT) logical(lgl)
!            On entry, if
!
!            - LAST = true  the current subvector is the last subvector 
!              of the data vector XX (or YY).
!            - LAST = false the current subvector is not the last subvector
!              of the data vector XX (or YY).
!
!   XSTAT    (INPUT/OUTPUT) real(stnd), dimension(2)
!            On entry, after the first call to COMP_COR_MISS2 (e.g. when FIRST=true),
!            XSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_COR_MISS2. XSTAT should not be changed between calls to
!            COMP_COR_MISS2.
!
!            On exit, when LAST=true, XSTAT contains the following statistics:
!
!            - XSTAT(1) contains the mean value of the data vector XX.
!            - XSTAT(2) contains the variance of the data vector XX.
!
!            The size of XSTAT must verify: size(XSTAT) = 2.
!
!   YSTAT    (INPUT/OUTPUT) real(stnd), dimension(2)
!            On entry, after the first call to COMP_COR_MISS2 (e.g. when FIRST=true),
!            YSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_COR_MISS2. YSTAT should not be changed between calls to
!            COMP_COR_MISS2.
!
!            On exit, when LAST=true, YSTAT contains the following statistics:
!
!            - YSTAT(1) contains the mean value of the data vector YY.
!            - YSTAT(2) contains the variance of the data vector YY.
!
!            The size of YSTAT must verify: size(YSTAT) = 2.
!
!   XYCOR    (INPUT/OUTPUT) real(stnd)
!            On entry, after the first call to COMP_COR_MISS2 (e.g. when FIRST=true),
!            XYCOR is used as workspace to accumulate quantities on previous calls
!            to COMP_COR_MISS2. XYCOR should not be changed between calls to
!            COMP_COR_MISS2.
!
!            On exit, when LAST=true, XYCOR contains the correlation coefficient
!            between XX(:) and YY(:).
!
!   XYN      (INPUT/OUTPUT) real(stnd)
!            On entry, after the first call to COMP_COR_MISS2 (e.g. when FIRST=true),
!            XYN contains count of valid pairs of observations from previous calls
!            to COMP_COR_MISS2. XYN should not be changed between calls to
!            COMP_COR_MISS2.
!
!            On exit, XYN contains the incidence value between XX(:) and YY(:). XYN
!            indicates the number of non-missing pairs of observations which were
!            used in the calculation of XSTAT, YSTAT and XYCOR.
!
!   XYMISS   (INPUT) real(stnd)
!            On entry, the missing value indicator. Any value in
!            X or Y which is equal to XYMISS is assumed to be missing
!            or invalid.
!            The basic univariate and bivariate statistics are computed
!            on all valid pairs of observations where XX and YY are not 
!            missing (see Further Details).
!
!   Z        (OUTPUT, OPTIONAL) real(stnd)
!            On exit, when LAST=true, Z contains the Fisher's Z transformation of
!            XYCOR.
!
!            Z needs to be specified only on the last call to COMP_COR_MISS2 (e.g. when LAST=true).
!          
!   PROB     (OUTPUT, OPTIONAL) real(stnd)
!            On exit, when LAST=true, PROB gives the probability that the random sample
!            of XYN observation pairs YY(:) and XX(:) came from a bivariate normal
!            population with a correlation coefficient equal to zero.
!
!            PROB needs to be specified only on the last call to COMP_COR_MISS2 (e.g. when LAST=true).
!          
!   NDF_MAX  (INPUT, OPTIONAL) integer(i4b)
!            On entry, when argument PROB is present, NDF_MAX is used as follows:
!
!            - If XYN-2 is lower or equal to NDF_MAX, the t_density is integrated
!              for computing PROB.
!            - If XYN-2 is greater than NDF_MAX, an asymptotic series is used
!              for computing PROB.
!
!            The default is 20.
!
!            NDF_MAX needs to be specified only on the last call to COMP_COR_MISS2 (e.g. when LAST=true).
!
!
! Further Details
! _______________
!
!   The subroutine computes the basic univariate statistics and the correlation
!   coefficient with only one pass through the data.
!
!   If fewer than two valid observations were present, the pertinent statistics are set to XYMISS.
!
!   The univariate and bivariate statistics are computed from all valid pairs of observations.
!
!   For more details on correlation and regression analysis, see:
!
!   (1) von Storch, H., and Zwiers, F.W., 2002:
!            Statistical Analysis in Climate Research.
!            Cambridge, UK, Chapter 8, 484 pp., ISBN:9780521012300
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, half
    use Num_Constants,     only : lamch
    use Utilities,         only : assert_eq, assert
    use Prob_Procedures,   only : probt
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                  :: xymiss
    real(stnd), intent(in),    dimension(:) :: x, y
    real(stnd), intent(inout), dimension(:) :: xstat, ystat
    real(stnd), intent(inout)               :: xycor, xyn
    real(stnd), intent(out),    optional    :: prob, z
!
    integer(i4b), intent(in), optional :: ndf_max
!
    logical(lgl), intent(in) :: first, last
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: dxi, dyi, a, b, safmin
!
    integer(i4b) :: i, n
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_cor_miss2'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq(  int(size(x),i4b),       &
                    int(size(y),i4b),       &
                    name_proc )
!
    call assert( logical(size(xstat)==2,lgl),    &
                 logical(size(ystat)==2,lgl),    &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBVECTOR IS THE FIRST.
!
!       INITIALIZE xstat, ystat, xycor AND xyn .
!
        xstat(1_i4b:2_i4b) = zero       
        ystat(1_i4b:2_i4b) = zero
        xycor = zero
        xyn   = zero
!
    end if
!
    do i = 1_i4b, n
!
        if ( x(i)==xymiss .or. y(i)==xymiss ) cycle
!
!       CALCULATE PARTIAL MEANS AND CORRECTED SUMS OF SQUARES AND CROSS-PRODUCT, 
!       HANDLING THE DATA ONCE.
!
        dxi = xyn
        xyn = dxi + one
        a = one/xyn
        b = dxi*a
!
        dxi = x(i) - xstat(1_i4b)
        dyi = y(i) - ystat(1_i4b)
!
        xstat(1_i4b) = xstat(1_i4b) + a*dxi
        xstat(2_i4b) = xstat(2_i4b) + b*dxi*dxi
!
        ystat(1_i4b) = ystat(1_i4b) + a*dyi
        ystat(2_i4b) = ystat(2_i4b) + b*dyi*dyi
!
        xycor = xycor + b*dxi*dyi
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBVECTOR IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
        if ( xyn>one ) then
!
            safmin = lamch('s')
!
            a = one/xyn
            b = xstat(2_i4b)*ystat(2_i4b)
!
!           COMPUTE VARIANCE OF XX.
!
            xstat(2_i4b) = xstat(2_i4b)*a
!
!           COMPUTE VARIANCE OF YY.
!
            ystat(2_i4b) = ystat(2_i4b)*a
!
!           COMPUTE CORRELATION COEFFICIENT.
!
            if ( b>safmin ) then
                xycor = xycor/sqrt( b )
            else
                xycor = xymiss
            end if
!
!           OUTPUT FISHER'S z TRANSFORMATION, IF REQUIRED.
!
            if ( present(z) ) then
!
                if ( b>safmin ) then
                    z = half*log(( (one+xycor) + safmin )/( (one-xycor) + safmin ))
                else
                    z = xymiss
                end if
!
            end if
!
!           OUTPUT SIGNIFICANCE PROBABILITY IF REQUIRED.
!
            if ( present(prob) ) then
!
                if ( xyn>two .and. b>safmin ) then
!
                    a = abs( xycor )
!
                    if ( a/=one ) then
                        b = xyn - two
                        i = int( b, i4b )
                        dxi = sqrt( b )*a/sqrt( one - a*a )
                        prob = two*probt( dxi, i, true, ndf_max=ndf_max )
                    else
                        prob = zero
                    end if
!
                else
!
                    prob = xymiss
!
                end if
!
            end if
!
        else
!
!           FEWER THAN TWO VALID OBSERVATIONS, SET THE PERTINENT STATISTICS TO MISSING.
!
            if ( xyn<one ) then
                xstat(1_i4b:2_i4b) = xymiss
                ystat(1_i4b:2_i4b) = xymiss
            end if
!
            xycor = xymiss
!
            if ( present(prob) ) prob = xymiss
            if ( present(z) )    z    = xymiss
!
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_cor_miss2_rv
! ___________________________________
!
    end subroutine comp_cor_miss2_rv
!
! =========================================================================================
!
    subroutine comp_cor_miss2_rm( x, y, first, last, xstat, ystat, xycor, xyn, xymiss, &
                                  dimvar, z, prob, ndf_max )
!
! Purpose
! _______
!
!   COMP_COR_MISS2 computes estimates of means, variances and correlation
!   coefficients (with another data vector YY) from a data matrix XX  possibly
!   containing missing values.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:,:)
!            On entry, input submatrix containing size(X,3-DIMVAR)
!            observations on size(X,DIMVAR) variables from the matrix of data XX
!            for which basic univariate  and bivariate statistics are desired. By
!            default, DIMVAR is equal to 1. See description of optional DIMVAR
!            argument for details.
!            If all the data are available at once, X can be the full 
!            data matrix XX.
!
!   Y        (INPUT) real(stnd), dimension(:)
!            On entry, input subvector containing size(X,3-DIMVAR) observations
!            from a vector of data YY for which correlation coefficient
!            with XX is desired. If all the data are available at once, YY
!            can be the full data vector.
!
!            The size of Y must verify: size(Y) = SIZE(X,3-DIMVAR) .
!
!   FIRST    (INPUT) logical(lgl)
!            On entry, if
!
!            - FIRST = true  the current submatrix is the first submatrix 
!              of the data matrix XX.
!            - FIRST = false the current submatrix is not the first submatrix
!              of the data matrix XX.
!
!   LAST     (INPUT) logical(lgl)
!            On entry, if
!
!            - LAST = true  the current submatrix is the last submatrix 
!              of the data matrix XX.
!            - LAST = false the current submatrix is not the last submatrix
!              of the data matrix XX.
!
!   XSTAT    (INPUT/OUTPUT) real(stnd), dimension(:,2)
!            On entry, after the first call to COMP_COR_MISS2 (e.g. when FIRST=true),
!            XSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_COR_MISS2. XSTAT should not be changed between calls to
!            COMP_COR_MISS2.
!
!            On exit, when LAST=true, XSTAT contains the following statistics
!            on all variables:
!
!            - XSTAT(:,1) contains the mean values of the data matrix XX.
!            - XSTAT(:,2) contains the variances of the data matrix XX.
!
!            The shape of XSTAT must verify:
!
!            - size(XSTAT,1) = size(X,DIMVAR) ,
!            - size(XSTAT,2) = 2.
!
!   YSTAT    (INPUT/OUTPUT) real(stnd), dimension(:,2)
!            On entry, after the first call to COMP_COR_MISS2 (e.g. when FIRST=true),
!            YSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_COR_MISS2. YSTAT should not be changed between calls to
!            COMP_COR_MISS2.
!
!            The shape of YSTAT must verify:
!
!            - size(YSTAT,1) = size(X,DIMVAR) ,
!            - size(YSTAT,2) = 2.
!
!   XYCOR    (INPUT/OUTPUT) real(stnd), dimension(:)
!            On entry, after the first call to COMP_COR_MISS2 (e.g. when FIRST=true),
!            XYCOR is used as workspace to accumulate quantities on previous calls
!            to COMP_COR_MISS2. XYCOR should not be changed between calls to
!            COMP_COR_MISS2.
!
!            On exit, when LAST=true, XYCOR(i) contains the correlation coefficient
!            between XX(i,:) ( XX(:,i) if DIMVAR=2 ) and YY(:).
!
!            The size of XYCOR must verify: size(XYCOR) = size(X,DIMVAR).
!
!   XYN      (INPUT/OUTPUT) real(stnd), dimension(:)
!            On entry, after the first call to COMP_COR_MISS2 (e.g. when FIRST=true),
!            XYN contains counts of valid pairs of observations from previous calls
!            to COMP_COR_MISS2. XYN should not be changed between calls to
!            COMP_COR_MISS2.
!
!            On exit, XYN(i) contains the incidence value between XX(i,:) ( XX(:,i) if DIMVAR=2 )
!            and YY(:). XYN(i) indicates the number of non-missing pairs of observations
!            which were used in the calculation of XSTAT and XYCOR.
!
!            The size of XYN must verify: size(XYN) = size(X,DIMVAR).
!
!   XYMISS   (INPUT) real(stnd)
!            On entry, the missing value indicator. Any value in
!            X or Y which is equal to XYMISS is assumed to be missing
!            or invalid.
!            The basic univariate and bivariate statistics for variable XX(i,:)
!            ( XX(:,i) if DIMVAR=2 ) are computed on all valid pairs of observations 
!            where XX(i,:) ( XX(:,i) if DIMVAR=2 ) and YY(:) are not missing
!            (see Further Details).
!
!   DIMVAR   (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVAR is present, DIMVAR is used as follows:
!
!            - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!              on size(X,1) variables.
!            - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!              on size(X,2) variables.
!
!            The default is DIMVAR = 1.
!
!   Z        (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            On exit, when LAST=true, Z contains the Fisher's Z transformation of
!            XYCOR(:).
!
!            Z needs to be specified only on the last call to COMP_COR_MISS2 (e.g. when LAST=true).
!          
!            The size of Z must verify:  size(Z) = size(X,DIMVAR).
!
!   PROB     (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            On exit, when LAST=true, PROB(i) gives the probability that the random sample
!            of XYN(i) observation pairs YY(:) and XX(i,:) ( XX(:,i) if DIMVAR=2 ) came from
!            a bivariate normal population with a correlation coefficient equal to zero.
!
!            PROB needs to be specified only on the last call to COMP_COR_MISS2 (e.g. when LAST=true).
!          
!            The size of PROB must verify:  size(PROB) = size(X,DIMVAR).
!
!   NDF_MAX  (INPUT, OPTIONAL) integer(i4b)
!            On entry, when argument PROB is present, NDF_MAX is used as follows:
!
!            - If XYN(i)-2 is lower or equal to NDF_MAX, the t_density is integrated
!              for computing PROB(i).
!            - If XYN(i)-2 is greater than NDF_MAX, an asymptotic series is used
!              for computing PROB(i).
!
!            The default is 20.
!
!            NDF_MAX needs to be specified only on the last call to COMP_COR_MISS2 (e.g. when LAST=true).
!
!
! Further Details
! _______________
!
!   The subroutine computes the basic univariate statistics and the correlation
!   coefficients with only one pass through the data.
!
!   If fewer than two valid observations were present, the statistics are set to XYMISS .
!
!   The univariate and bivariate statistics are computed from all valid pairs of observations.
!
!   For more details on correlation and regression analysis, see:
!
!   (1) von Storch, H., and Zwiers, F.W., 2002:
!            Statistical Analysis in Climate Research.
!            Cambridge, UK, Chapter 8, 484 pp., ISBN:9780521012300
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, half
    use Num_Constants,     only : lamch
    use Char_Constants,    only : stat_error1, stat_error5
    use Utilities,         only : assert_eq, assert, merror
    use Prob_Procedures,   only : probt
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                               :: xymiss
    real(stnd), intent(in),    dimension(:,:)            :: x
    real(stnd), intent(in),    dimension(:)              :: y
    real(stnd), intent(inout), dimension(:,:)            :: xstat, ystat
    real(stnd), intent(inout), dimension(:)              :: xycor, xyn
    real(stnd), intent(out),   dimension(:),    optional :: prob, z
!
    integer(i4b), intent(in), optional :: dimvar, ndf_max
!
    logical(lgl), intent(in) :: first, last
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                         :: safmin
    real(stnd), dimension(size(xycor)) :: dxi, dyi, a, b
!
    integer(i4b) :: i, n, m
!
    logical(lgl)                          :: dim_is_one
    logical(lgl), dimension(size(xycor))  :: mask
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_cor_miss2'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
    n = assert_eq(  int(size(x,3-int(i)),i4b),    &
                    int(size(y),i4b),             &
                    name_proc )
!
    m = assert_eq(  int(  (/ size(x,int(i)),      &
                             size(xstat,1),       &
                             size(ystat,1),       &
                             size(xycor),         &
                             size(xyn) /) ,i4b),  &
                    name_proc )
!
    if ( m<=0_i4b ) call merror( name_proc//stat_error1 )
!
    call assert( logical(size(xstat,2)==2,lgl),    &
                 logical(size(ystat,2)==2,lgl),    &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBMATRIX IS THE FIRST.
!
!       INITIALIZE xstat, ystat, xycor AND xyn .
!
        xstat(:m,1_i4b:2_i4b) = zero       
        ystat(:m,1_i4b:2_i4b) = zero
        xycor(:m) = zero
        xyn(:m)   = zero
!
    end if
!
    do i = 1_i4b, n
!
        if ( y(i)==xymiss ) cycle
!
!       CALCULATE PARTIAL MEANS AND CORRECTED SUMS OF SQUARES AND CROSS-PRODUCTS, 
!       HANDLING THE DATA ONCE.
!
        if ( dim_is_one ) then
            dxi(:m) = x(:m,i)
        else
            dxi(:m) = x(i,:m)
        end if
!
        mask(:m) = dxi(:m)/=xymiss
!
        where ( mask(:m) )
!
            dyi(:m) = xyn(:m)
            xyn(:m) = dyi(:m) + one
            a(:m) = one/xyn(:m)
            b(:m) = dyi(:m)*a(:m)
!
            dxi(:m) = dxi(:m) - xstat(:m,1_i4b)
            dyi(:m) = y(i) - ystat(:m,1_i4b)
!
            xstat(:m,1_i4b) = xstat(:m,1_i4b) + a(:m)*dxi(:m)
            xstat(:m,2_i4b) = xstat(:m,2_i4b) + b(:m)*( dxi(:m)*dxi(:m) )
!
            ystat(:m,1_i4b) = ystat(:m,1_i4b) + a(:m)*dyi(:m)
            ystat(:m,2_i4b) = ystat(:m,2_i4b) + b(:m)*( dyi(:m)*dyi(:m) )
!
            xycor(:m) = xycor(:m) + b(:m)*dyi(:m)*dxi(:m)
!
        end where
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBMATRIX IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
!
!       TEST OPTIONAL ARGUMENTS.
!
        if ( present(prob) ) call assert( logical(size(prob)==m,lgl), name_proc )
!
        if ( present(z) ) call assert( logical(size(z)==m,lgl), name_proc )
!
        safmin = lamch( 's' )
!
        mask(:m) = xyn(:m)>one
!
        where ( mask(:m) )
!
            a(:m) = one/xyn(:m)
            b(:m) = xstat(:m,2_i4b)*ystat(:m,2_i4b)
!
            mask(:m) = b(:m)>safmin
!
!           COMPUTE VARIANCES OF XX.
!
            xstat(:m,2_i4b) = xstat(:m,2_i4b)*a(:m)
!
!           COMPUTE VARIANCES OF YY.
!
            ystat(:m,2_i4b) = ystat(:m,2_i4b)*a(:m)
!
        elsewhere
!
!           FEWER THAN TWO VALID OBSERVATIONS, SET THE PERTINENT STATISTICS TO xymiss .
!
            where ( xyn(:m)<one )
                xstat(:m,1_i4b) = xymiss
                xstat(:m,2_i4b) = xymiss
                ystat(:m,1_i4b) = xymiss
                ystat(:m,2_i4b) = xymiss
            elsewhere
                xstat(:m,2_i4b) = zero
                ystat(:m,2_i4b) = zero
            end where
!                
!            xstat(:m,1_i4b) = merge( xymiss, xstat(:m,1_i4b), xyn(:m)<one )
!            xstat(:m,2_i4b) = merge( xymiss, zero, xyn(:m)<one )
!            ystat(:m,1_i4b) = merge( xymiss, ystat(:m,1_i4b), xyn(:m)<one )
!            ystat(:m,2_i4b) = merge( xymiss, zero, xyn(:m)<one )
!
        end where
!
!       COMPUTE CORRELATION COEFFICIENTS BETWEEN XX AND YY.
!
        where ( mask(:m) )
            xycor(:m) = xycor(:m)/sqrt( b(:m) )
        elsewhere
            xycor(:m) = xymiss
        end where
!
!       OUTPUT FISHER'S z TRANSFORMATION, IF REQUIRED.
!
        if ( present(z) ) then
!
            where ( mask(:m) )
                z(:m) = half*log(( (one+xycor(:m)) + safmin )    /    &
                                 ( (one-xycor(:m)) + safmin ))
            elsewhere
                z(:m) = xymiss
            end where
!
        end if
!
        if ( present(prob) ) then
!
!           OUTPUT SIGNIFICANCE PROBABILITIES.
!
            mask(:m) = mask(:m) .and. xyn(:m)>two
!
            where ( mask(:m) )
                a(:m) = abs( xycor(:m) )
                mask(:m) = a(:m)/=one
                prob(:m) = zero
            elsewhere
                prob(:m) = xymiss
            end where
!
            where ( mask(:m) )
                b(:m) = xyn(:m) - two
                dxi(:m) = sqrt( b(:m) )*a(:m)/sqrt( one - a(:m)**2 )
            end where
!
!$OMP PARALLEL DO IF(m>omp_chunk)                &
!$OMP            ,SCHEDULE(DYNAMIC,omp_chunk)    &
!$OMP            ,PRIVATE(i), SHARED(ndf_max)
!
            do i = 1_i4b, m
!
                if ( mask(i) ) then
                    prob(i) = two*probt( dxi(i), int( b(i), i4b ), true, ndf_max=ndf_max )
                end if
!
            end do
!
!$OMP END PARALLEL DO
!
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_cor_miss2_rm
! ___________________________________
!
    end subroutine comp_cor_miss2_rm
!
! =========================================================================================
!
    subroutine comp_cor_miss2_rt( x, y, first, last, xstat, ystat, xycor, xyn, xymiss, &
                                  z, prob, ndf_max )
!
! Purpose
! _______
!
!   COMP_COR_MISS2 computes estimates of means, variances and correlation
!   coefficients (with another data vector YY) from a data tridimensional
!   array XX  possibly containing missing values.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:,:,:)
!            On entry, input subarray containing size(X,3) observations
!            on size(X,1) by size(X,2) variables from the array of data XX for which 
!            basic univariate and bivariate statistics are desired.
!            If all the data are available at once, X can be the full data array XX.
!
!   Y        (INPUT) real(stnd), dimension(:)
!            On entry, input subvector containing size(X,3) observations
!            from a vector of data YY for which correlation coefficient
!            with XX is desired. If all the data are available at once, YY
!            can be the full data vector.
!
!            The size of Y must verify: size(Y) = SIZE(X,3) .
!
!   FIRST    (INPUT) logical(lgl)
!            On entry, if
!
!            - FIRST = true  the current subarray is the first subarray 
!              of the data array XX.
!            - FIRST = false the current subarray is not the first subarray
!              of the data array XX.
!
!   LAST     (INPUT) logical(lgl)
!            On entry, if
!
!            - LAST = true  the current subarray is the last subarray 
!              of the data array XX.
!            - LAST = false the current subarray is not the last subarray
!              of the data array XX.
!
!   XSTAT    (INPUT/OUTPUT) real(stnd), dimension(:,:,2)
!            On entry, after the first call to COMP_COR_MISS2 (e.g. when FIRST=true),
!            XSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_COR_MISS2. XSTAT should not be changed between calls to
!            COMP_COR_MISS2.
!
!            On exit, when LAST=true, XSTAT contains the following statistics
!            on all variables:
!
!            - XSTAT(:,:,1) contains the mean values of the data array XX.
!            - XSTAT(:,:,2) contains the variances of the data array XX.
!
!            The shape of XSTAT must verify:
!
!            - size(XSTAT,1) = size(X,1) ,
!            - size(XSTAT,2) = size(X,2) ,
!            - size(XSTAT,3) = 2 .
!
!   YSTAT    (INPUT/OUTPUT) real(stnd), dimension(:,:,2)
!            On entry, after the first call to COMP_COR_MISS2 (e.g. when FIRST=true),
!            YSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_COR_MISS2. YSTAT should not be changed between calls to
!            COMP_COR_MISS2.
!
!            The shape of XSTAT must verify:
!
!            - size(YSTAT,1) = size(X,1) ,
!            - size(YSTAT,2) = size(X,2) ,
!            - size(YSTAT,3) = 2.
!
!   XYCOR    (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, after the first call to COMP_COR_MISS2 (e.g. when FIRST=true),
!            XYCOR is used as workspace to accumulate quantities on previous calls
!            to COMP_COR_MISS2. XYCOR should not be changed between calls to
!            COMP_COR_MISS2.
!
!            On exit, when LAST=true, XYCOR(i,j) contains the correlation coefficient
!            between XX(i,j,:) and YY(:).
!
!            The shape of XYCOR must verify:
!
!            - size(XYCOR,1) = size(X,1) ,
!            - size(XYCOR,2) = size(X,2) .
!
!   XYN      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, after the first call to COMP_COR_MISS2 (e.g. when FIRST=true),
!            XYN contains counts of valid pairs of observations from previous calls
!            to COMP_COR_MISS2. XYN should not be changed between calls to
!            COMP_COR_MISS2.
!            On exit, XYN(i,j) contains the incidence value between XX(i,j,:) and YY(:). XYN(i,j)
!            indicates the number of non-missing pairs of observations which were
!            used in the calculation of XSTAT and XYCOR.
!
!            The shape of XYN must verify:
!
!            - size(XYN,1) = size(X,1) ,
!            - size(XYN,2) = size(X,2) .
!
!   XYMISS   (INPUT) real(stnd)
!            On entry, the missing value indicator. Any value in
!            X or Y which is equal to XYMISS is assumed to be missing
!            or invalid.
!            The basic univariate and bivariate statistics for variable XX(i,j,:) are
!            computed on all valid pairs of observations where XX(i,j,:) and YY(:) are not 
!            missing (see Further Details).
!
!   Z        (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!            On exit, when LAST=true, Z contains the Fisher's Z transformation of
!            XYCOR(:,:).
!
!            Z needs to be specified only on the last call to COMP_COR_MISS2 (e.g. when LAST=true).
!          
!            The shape of Z must verify:
!          
!            - size(Z,1) = size(X,1) ,
!            - size(Z,2) = size(X,2) .
!
!   PROB     (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!            On exit, when LAST=true, PROB(i,j) gives the probability that the random sample
!            of XYN(i,j) observation pairs YY(:) and XX(i,j,:) came from a bivariate normal
!            population with a correlation coefficient equal to zero.
!
!            PROB needs to be specified only on the last call to COMP_COR_MISS2 (e.g. when LAST=true).
!
!            The shape of PROB must verify:
!
!            - size(PROB,1) = size(X,1) ,
!            - size(PROB,2) = size(X,2) .
!          
!   NDF_MAX  (INPUT, OPTIONAL) integer(i4b)
!            On entry, when argument PROB is present, NDF_MAX is used as follows:
!
!            - If XYN(i,j)-2 is lower or equal to NDF_MAX, the t_density is integrated
!              for computing PROB(i,j).
!            - If XYN(i,j)-2 is greater than NDF_MAX, an asymptotic series is used
!              for computing PROB(i,j).
!
!            The default is 20.
!
!            NDF_MAX needs to be specified only on the last call to COMP_COR_MISS2 (e.g. when LAST=true).
!
!
! Further Details
! _______________
!
!   The subroutine computes the basic univariate statistics and the correlation
!   coefficients with only one pass through the data.
!
!   If fewer than two valid observations were present, the statistics are set to XYMISS .
!
!   The univariate and bivariate statistics are computed from all valid pairs of observations.
!
!   For more details on correlation and regression analysis, see:
!
!   (1) von Storch, H., and Zwiers, F.W., 2002:
!            Statistical Analysis in Climate Research.
!            Cambridge, UK, Chapter 8, 484 pp., ISBN:9780521012300
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true
    use Reals_Constants,   only : zero, one, two, half
    use Num_Constants,     only : lamch
    use Char_Constants,    only : stat_error1
    use Utilities,         only : assert_eq, assert, merror
    use Prob_Procedures,   only : probt
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                                 :: xymiss
    real(stnd), intent(in),    dimension(:,:,:)            :: x
    real(stnd), intent(in),    dimension(:)                :: y
    real(stnd), intent(inout), dimension(:,:,:)            :: xstat, ystat
    real(stnd), intent(inout), dimension(:,:)              :: xycor, xyn
    real(stnd), intent(out),   dimension(:,:),    optional :: prob, z
!
    integer(i4b), intent(in), optional :: ndf_max
!
    logical(lgl), intent(in) :: first, last
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                       :: safmin
    real(stnd), dimension(size(x,1)) :: dxi, dyi, a, b
!
    integer(i4b) :: i, j, n, m, p
!
    logical(lgl), dimension(size(x,1)) :: mask
    logical(lgl)                       :: comp_z, comp_prob
#ifdef _OPENMP
    logical      :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_cor_miss2'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq(  int(size(x,3),i4b),     &
                    int(size(y),i4b),       &
                    name_proc )
!
    m = assert_eq(  int(  (/ size(x,1),             &
                             size(xstat,1),         &
                             size(ystat,1),         &
                             size(xycor,1),         &
                             size(xyn,1) /) ,i4b),  &
                    name_proc )
!
!
    p = assert_eq(  int(  (/ size(x,2),             &
                             size(xstat,2),         &
                             size(ystat,2),         &
                             size(xycor,2),         &
                             size(xyn,2) /) ,i4b),  &
                    name_proc )
!
    if ( min(m,p)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    call assert( logical(size(xstat,3)==2,lgl),    &
                 logical(size(ystat,3)==2,lgl),    &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBMATRIX IS THE FIRST.
!
!       INITIALIZE xstat, ystat, xycor AND xyn.
!
        xstat(:m,:p,1_i4b:2_i4b) = zero       
        ystat(:m,:p,1_i4b:2_i4b) = zero
        xycor(:m,:p) = zero
        xyn(:m,:p)   = zero
!
    end if
!
    do i = 1_i4b, n
!
        if ( y(i)==xymiss ) cycle
!
!       CALCULATE PARTIAL MEANS AND CORRECTED SUMS OF SQUARES AND CROSS-PRODUCTS, 
!       HANDLING THE DATA ONCE.
!
        do j = 1_i4b, p
!
            mask(:m) = x(:m,j,i)/=xymiss
!
            where ( mask(:m) )
!
                dxi(:m)   = xyn(:m,j)
                xyn(:m,j) = dxi(:m) + one
                a(:m)     = one/xyn(:m,j)
                b(:m)     = dxi(:m)*a(:m)
!
                dxi(:m) = x(:m,j,i) - xstat(:m,j,1_i4b)
                dyi(:m) = y(i) - ystat(:m,j,1_i4b)
!
                xstat(:m,j,1_i4b) = xstat(:m,j,1_i4b) + a(:m)*dxi(:m)
                xstat(:m,j,2_i4b) = xstat(:m,j,2_i4b) + b(:m)*( dxi(:m)*dxi(:m) )
!
                ystat(:m,j,1_i4b) = ystat(:m,j,1_i4b) + a(:m)*dyi(:m)
                ystat(:m,j,2_i4b) = ystat(:m,j,2_i4b) + b(:m)*( dyi(:m)*dyi(:m) )
!
                xycor(:m,j) = xycor(:m,j) + b(:m)*dyi(:m)*dxi(:m)
!
            end where
!
        end do
!
    end do
!
    if ( last ) then
!
!       THE CURRENT SUBARRAY IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
!       TEST OPTIONAL ARGUMENTS.
!
        comp_prob = present( prob )
!
        if ( comp_prob ) call assert( logical(size(prob,1)==m,lgl),     &
                                      logical(size(prob,2)==p,lgl),     &
                                      name_proc )
!
        comp_z = present( z )
!
        if ( comp_z ) call assert( logical(size(z,1)==m,lgl),     &
                                   logical(size(z,2)==p,lgl),     &
                                   name_proc )
!
        safmin = lamch( 's' )
!
#ifdef _OPENMP
        i = omp_get_num_procs()
        j = omp_get_max_threads()
        ompparallel = omp_in_parallel()
        ompnested   = omp_get_nested() .and. ompparallel
        test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                      (m*p)>=omp_limit                    .and.      &
                      i>1_i4b                             .and.      &
                      j>1_i4b
#endif
!
!$OMP PARALLEL DO IF(test_par), SCHEDULE(DYNAMIC,1)         &
!$OMP            ,PRIVATE(j,i,mask,a,b,dxi)                 &
!$OMP            ,SHARED(safmin,ndf_max,m,p,comp_z,comp_prob)
!
        do j = 1_i4b, p
!
            mask(:m) = xyn(:m,j)>one
!
            where ( mask(:m) )
!
                a(:m)    = one/xyn(:m,j)
                b(:m)    = xstat(:m,j,2_i4b)*ystat(:m,j,2_i4b)
                mask(:m) = b(:m)>safmin
!
!               COMPUTE VARIANCES OF XX.
!
                xstat(:m,j,2_i4b) = xstat(:m,j,2_i4b)*a(:m)
!
!               COMPUTE VARIANCES OF YY.
!
                ystat(:m,j,2_i4b) = ystat(:m,j,2_i4b)*a(:m)
!
            elsewhere
!
!               FEWER THAN TWO VALID OBSERVATIONS, SET THE PERTINENT STATISTICS TO xymiss .
!
                where ( xyn(:m,j)<one )
                    xstat(:m,j,1_i4b) = xymiss
                    xstat(:m,j,2_i4b) = xymiss
                    ystat(:m,j,1_i4b) = xymiss
                    ystat(:m,j,2_i4b) = xymiss
                elsewhere
                    xstat(:m,j,2_i4b) = zero
                    ystat(:m,j,2_i4b) = zero
                end where
!
!                xstat(:m,j,1_i4b) = merge( xymiss, xstat(:m,j,1_i4b), xyn(:m,j)<one )
!                xstat(:m,j,2_i4b) = merge( xymiss, zero, xyn(:m,j)<one )
!                ystat(:m,j,1_i4b) = merge( xymiss, ystat(:m,j,1_i4b), xyn(:m,j)<one )
!                ystat(:m,j,2_i4b) = merge( xymiss, zero, xyn(:m,j)<one )
!
            end where
!
!           COMPUTE CORRELATION COEFFICIENTS BETWEEN XX AND YY.
!
            where ( mask(:m) )
                xycor(:m,j) = xycor(:m,j)/sqrt( b(:m) )
            elsewhere
                xycor(:m,j) = xymiss
            end where
!
!           OUTPUT FISHER'S z TRANSFORMATION, IF REQUIRED.
!
            if ( comp_z ) then
!
                where ( mask(:m) )
                    z(:m,j) = half*log(( (one+xycor(:m,j)) + safmin )   /      &
                                        ( (one-xycor(:m,j)) + safmin ))
                elsewhere
                    z(:m,j) = xymiss
                end where
!
            end if
!
            if ( comp_prob ) then
!
!               OUTPUT SIGNIFICANCE PROBABILITIES.
!
                mask(:m) = mask(:m) .and. xyn(:m,j)>two
!
                where ( mask(:m) )
                    a(:m) = abs( xycor(:m,j) )
                    mask(:m) = a(:m)/=one
                    prob(:m,j) = zero
                elsewhere
                    prob(:m,j) = xymiss
                end where
!
                where ( mask(:m) )
                    b(:m) = xyn(:m,j) - two
                    dxi(:m) = sqrt( b(:m) )*a(:m)/sqrt( one - a(:m)**2 )
                end where
!
                do i = 1_i4b, m
!
                    if ( mask(i) ) then
                        prob(i,j) = two*probt( dxi(i), int( b(i), i4b ), true, ndf_max=ndf_max )
                    end if
!
                end do
!
            end if
!
        end do
!
!$OMP END PARALLEL DO
!
    end if
!
!
! END OF SUBROUTINE comp_cor_miss2_rt
! ___________________________________
!
    end subroutine comp_cor_miss2_rt
!
!
! =========================================================================================
!                  UPDATING COVARIANCE AND CORRELATION COEFFICIENTS
! =========================================================================================
!
!
    subroutine update_cor_rv( xstat, ystat, xycor, xyn, xstat2, ystat2, xycor2, xyn2 )
!
! Purpose
! _______
!
!   UPDATE_COR computes sample means and corrected sums of squares and cross-products
!   for a sample of size XYN+XYN2 given the means and corrected sum of squares and
!   cross-products for two subsamples of size XYN and XYN2 as output by a call to COMP_COR
!   when LAST=false on the two subsamples separetely.
!
!   The sample means, variances and coefficient correlation for the sample of size XYN+XYN2
!   may be obtained by a call to COMP_COR with LAST=true.
!
! Arguments
! _________
!
!   XSTAT    (INPUT/OUTPUT) real(stnd), dimension(2)
!            On entry, the XSTAT argument of COMP_COR for the first subsample.
!            On exit, the XSTAT argument of the combined sample.
!
!   YSTAT    (INPUT/OUTPUT) real(stnd), dimension(2)
!            On entry, the YSTAT argument of COMP_COR for the first subsample.
!            On exit, the YSTAT argument of the combined sample.
!
!   XYCOR    (INPUT/OUTPUT) real(stnd)
!            On entry, the XYCOR argument of COMP_COR for the first subsample.
!            On exit, the XYCOR argument of the combined sample.
!
!   XYN      (INPUT/OUTPUT) real(stnd)
!            On entry, the XYN argument of COMP_COR for the first subsample.
!            On exit, the XYN argument of the combined sample.
!
!   XSTAT2   (INPUT) real(stnd), dimension(2)
!            On entry, the XSTAT argument of COMP_COR for the second subsample.
!
!   YSTAT2   (INPUT) real(stnd), dimension(2)
!            On entry, the YSTAT argument of COMP_COR for the second subsample.
!
!   XYCOR2   (INPUT) real(stnd)
!            On entry, the XYCOR argument of COMP_COR for the second subsample.
!
!   XYN2     (INPUT) real(stnd)
!            On entry, the XYN argument of COMP_COR for the second subsample.
!
!
! Further Details
! _______________
!
!   One possible application of this subroutine is to parallel processing. If one has two
!   or more processors available, the sample can be split up into smaller subsamples,
!   and the means and corrected sums of squares and cross-products computed for each
!   subsample independently using COMP_COR. The means and corrected sums of squares
!   and cross-products for the original sample can then be calculated using UPDATE_COR.
!   The means, variances and correlation coefficient for the original sample can 
!   be computed by a final call to COMP_COR with LAST=true.
!
!   This subroutine is adapted from:
!
!   (1) Chan, T.F., Golub, G.H, and Leveque, R.J., 1979:
!           Updating formulae and a pairwise algorithm for computing sample variances.
!           STAN-CS-79-773.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Utilities, only : assert
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:)    :: xstat, ystat, xycor
    real(stnd), intent(inout)                  :: xyn
!
    real(stnd), intent(in), dimension(:)   :: xstat2, ystat2, xycor2
    real(stnd), intent(in)                 :: xyn2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)   :: xyn_tmp, xmx2, ymy2, tmp, tmp2
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='update_cor'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    call assert( logical(size(xstat)==2,lgl),    &
                 logical(size(ystat)==2,lgl),    &
                 logical(size(xstat2)==2,lgl),   &
                 logical(size(ystat2)==2,lgl),   &
                 name_proc )
!
    xyn_tmp = xyn + xyn2
    ymy2    = ystat(1) - ystat2(1)
    xmx2    = xstat(1) - xstat2(1)
!
    tmp     = ( xyn*xyn2 )/xyn_tmp
    tmp2    = tmp*ymy2
!
    xycor = xycor + xycor2 + tmp2*xmx2
!
    xstat(2) = xstat(2) +  xstat2(2) + tmp*(xmx2**2)
!
    ystat(2) = ystat(2) +  ystat2(2) + tmp2*ymy2
!
    tmp  = xyn/xyn_tmp
    tmp2 = xyn2/xyn_tmp
!
    xstat(1) = tmp*xstat(1) + tmp2*xstat2(1)
!
    ystat(1) = tmp*ystat(1) + tmp2*ystat2(1)
!
    xyn = xyn_tmp
!
!
! END OF SUBROUTINE update_cor_rv
! _______________________________
!
    end subroutine update_cor_rv
!
! =========================================================================================
!
    subroutine update_cor_rm( xstat, ystat, xycor, xyn, xstat2, ystat2, xycor2, xyn2 )
!
! Purpose
! _______
!
!   UPDATE_COR computes sample means and corrected sums of squares and cross-products
!   for a sample of size XYN+XYN2 given the means and corrected sums of squares and
!   cross-products for two subsamples of size XYN and XYN2 as output by a call to COMP_COR
!   when LAST=false on the two subsamples separetely.
!
!   The sample means, variances and coefficient correlations for the sample of size XYN+XYN2
!   may be obtained by a call to COMP_COR with LAST=true.
!
!
! Arguments
! _________
!
!   XSTAT    (INPUT/OUTPUT) real(stnd), dimension(:,2)
!            On entry, the XSTAT argument of COMP_COR for the first subsample.
!            On exit, the XSTAT argument of the combined sample.
!
!   YSTAT    (INPUT/OUTPUT) real(stnd), dimension(2)
!            On entry, the YSTAT argument of COMP_COR for the first subsample.
!            On exit, the YSTAT argument of the combined sample.
!
!   XYCOR    (INPUT/OUTPUT) real(stnd), dimension(:)
!            On entry, the XYCOR argument of COMP_COR for the first subsample.
!            On exit, the XYCOR argument of the combined sample.
!
!   XYN      (INPUT/OUTPUT) real(stnd)
!            On entry, the XYN argument of COMP_COR for the first subsample.
!            On exit, the XYN argument of the combined sample.
!
!   XSTAT2   (INPUT) real(stnd), dimension(:,2)
!            On entry, the XSTAT argument of COMP_COR for the second subsample.
!
!   YSTAT2   (INPUT) real(stnd), dimension(2)
!            On entry, the YSTAT argument of COMP_COR for the second subsample.
!
!   XYCOR2   (INPUT) real(stnd), dimension(:)
!            On entry, the XYCOR argument of COMP_COR for the second subsample.
!
!   XYN2     (INPUT) real(stnd)
!            On entry, the XYN argument of COMP_COR for the second subsample.
!
!
! Further Details
! _______________
!
!   One possible application of this subroutine is to parallel processing. If one has two
!   or more processors available, the sample can be split up into smaller subsamples,
!   and the means and corrected sums of squares and cross-products computed for each
!   subsample independently using COMP_COR. The means and corrected sums of squares
!   and cross-products for the original sample can then be calculated using UPDATE_COR.
!   The means, variances and correlation coefficient for the original sample can 
!   be computed by a final call to COMP_COR with LAST=true.
!
!   This subroutine is adapted from:
!
!   (1) Chan, T.F., Golub, G.H, and Leveque, R.J., 1979:
!           Updating formulae and a pairwise algorithm for computing sample variances.
!           STAN-CS-79-773.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,    only : stat_error1
    use Utilities,         only : assert_eq, assert, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:,:)  :: xstat
    real(stnd), intent(inout), dimension(:)    :: ystat, xycor
    real(stnd), intent(inout)                  :: xyn
!
    real(stnd), intent(in), dimension(:,:) :: xstat2
    real(stnd), intent(in), dimension(:)   :: ystat2, xycor2
    real(stnd), intent(in)                 :: xyn2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                         :: xyn_tmp, ymy2, tmp, tmp2
    real(stnd), dimension(size(xycor)) :: xmx2
!
    integer(i4b) :: m
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='update_cor'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq(  int(size(xstat,1),i4b),       &
                    int(size(xstat2,1),i4b),      &
                    int(size(xycor),i4b),         &
                    int(size(xycor2),i4b),        &
                    name_proc )
!
    if ( m<=0_i4b ) call merror( name_proc//stat_error1 )
!
    call assert( logical(size(xstat,2)==2,lgl),    &
                 logical(size(ystat)==2,lgl),      &
                 logical(size(xstat2,2)==2,lgl),   &
                 logical(size(ystat2)==2,lgl),     &
                 name_proc )
!
    xyn_tmp  = xyn + xyn2
    ymy2     = ystat(1) - ystat2(1)
    xmx2(:m) = xstat(:m,1) - xstat2(:m,1)
!
    tmp     = ( xyn*xyn2 )/xyn_tmp
    tmp2    = tmp*ymy2
!
    xycor(:m) = xycor(:m) + xycor2(:m) + tmp2*xmx2(:m)
!
    xstat(:m,2) = xstat(:m,2) +  xstat2(:m,2) + tmp*(xmx2(:m)**2)
!
    ystat(2) = ystat(2) +  ystat2(2) + tmp2*ymy2
!
    tmp  = xyn/xyn_tmp
    tmp2 = xyn2/xyn_tmp
!
    xmx2(:m)   = tmp*xstat(:m,1) + tmp2*xstat2(:m,1)
    xstat(:m,1) = xmx2(:m)
!
    ystat(1) = tmp*ystat(1) + tmp2*ystat2(1)
!
    xyn = xyn_tmp
!
!
! END OF SUBROUTINE update_cor_rm
! _______________________________
!
    end subroutine update_cor_rm
!
! =========================================================================================
!
    subroutine update_cor_rt( xstat, ystat, xycor, xyn, xstat2, ystat2, xycor2, xyn2 )
!
! Purpose
! _______
!
!   UPDATE_COR computes sample means and corrected sums of squares and cross-products
!   for a sample of size XYN+XYN2 given the means and corrected sums of squares and
!   cross-products for two subsamples of size XYN and XYN2 as output by a call to COMP_COR
!   when LAST=false on the two subsamples separetely.
!
!   The sample means, variances and coefficient correlations for the sample of size XYN+XYN2
!   may be obtained by a call to COMP_COR with LAST=true.
!
!
! Arguments
! _________
!
!   XSTAT    (INPUT/OUTPUT) real(stnd), dimension(:,:,2)
!            On entry, the XSTAT argument of COMP_COR for the first subsample.
!            On exit, the XSTAT argument of the combined sample.
!
!   YSTAT    (INPUT/OUTPUT) real(stnd), dimension(2)
!            On entry, the YSTAT argument of COMP_COR for the first subsample.
!            On exit, the YSTAT argument of the combined sample.
!
!   XYCOR    (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, the XYCOR argument of COMP_COR for the first subsample.
!            On exit, the XYCOR argument of the combined sample.
!
!   XYN      (INPUT/OUTPUT) real(stnd)
!            On entry, the XYN argument of COMP_COR for the first subsample.
!            On exit, the XYN argument of the combined sample.
!
!   XSTAT2   (INPUT) real(stnd), dimension(:,:,2)
!            On entry, the XSTAT argument of COMP_COR for the second subsample.
!
!   YSTAT2   (INPUT) real(stnd), dimension(2)
!            On entry, the YSTAT argument of COMP_COR for the second subsample.
!
!   XYCOR2   (INPUT) real(stnd), dimension(:,:)
!            On entry, the XYCOR argument of COMP_COR for the second subsample.
!
!   XYN2     (INPUT) real(stnd)
!            On entry, the XYN argument of COMP_COR for the second subsample.
!
!
! Further Details
! _______________
!
!   One possible application of this subroutine is to parallel processing. If one has two
!   or more processors available, the sample can be split up into smaller subsamples,
!   and the means and corrected sums of squares and cross-products computed for each
!   subsample independently using COMP_COR. The means and corrected sums of squares
!   and cross-products for the original sample can then be calculated using UPDATE_COR.
!   The means, variances and correlation coefficient for the original sample can 
!   be computed by a final call to COMP_COR with LAST=true.
!
!   This subroutine is adapted from:
!
!   (1) Chan, T.F., Golub, G.H, and Leveque, R.J., 1979:
!           Updating formulae and a pairwise algorithm for computing sample variances.
!           STAN-CS-79-773.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,    only : stat_error1
    use Utilities,         only : assert_eq, assert, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:,:,:) :: xstat
    real(stnd), intent(inout), dimension(:,:)   :: xycor
    real(stnd), intent(inout), dimension(:)     :: ystat
    real(stnd), intent(inout)                   :: xyn
!
    real(stnd), intent(in),    dimension(:,:,:) :: xstat2
    real(stnd), intent(inout), dimension(:,:)   :: xycor2
    real(stnd), intent(inout), dimension(:)     :: ystat2
    real(stnd), intent(in)                      :: xyn2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                         :: xyn_tmp, ymy2, tmp, tmp2
    real(stnd), dimension(size(xycor,1),size(xycor,2)) :: xmx2
!
    integer(i4b) :: m, n
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='update_cor'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq(  int(size(xstat,1),i4b),       &
                    int(size(xstat2,1),i4b),      &
                    int(size(xycor,1),i4b),       &
                    int(size(xycor2,1),i4b),      &
                    name_proc )
!
    n = assert_eq(  int(size(xstat,2),i4b),       &
                    int(size(xstat2,2),i4b),      &
                    int(size(xycor,2),i4b),       &
                    int(size(xycor2,2),i4b),      &
                    name_proc )
!
    if ( min(m,n)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    call assert( logical(size(xstat,3)==2,lgl),    &
                 logical(size(ystat)==2,lgl),      &
                 logical(size(xstat2,3)==2,lgl),   &
                 logical(size(ystat2)==2,lgl),     &
                 name_proc )
!
    xyn_tmp     = xyn + xyn2
    ymy2        = ystat(1) - ystat2(1)
    xmx2(:m,:n) = xstat(:m,:n,1) - xstat2(:m,:n,1)
!
    tmp     = ( xyn*xyn2 )/xyn_tmp
    tmp2    = tmp*ymy2
!
    xycor(:m,:n) = xycor(:m,:n) + xycor2(:m,:n) + tmp2*xmx2(:m,:n)
!
    xstat(:m,:n,2) = xstat(:m,:n,2) +  xstat2(:m,:n,2) + tmp*(xmx2(:m,:n)**2)
!
    ystat(2) = ystat(2) +  ystat2(2) + tmp2*ymy2
!
    tmp  = xyn/xyn_tmp
    tmp2 = xyn2/xyn_tmp
!
    xmx2(:m,:n)   = tmp*xstat(:m,:n,1) + tmp2*xstat2(:m,:n,1)
    xstat(:m,:n,1) = xmx2(:m,:n)
!
    ystat(1) = tmp*ystat(1) + tmp2*ystat2(1)
!
    xyn = xyn_tmp
!
!
! END OF SUBROUTINE update_cor_rt
! _______________________________
!
    end subroutine update_cor_rt
!
! =========================================================================================
!
    subroutine update_cor_miss2_rm( xstat, ystat, xycor, xyn, xstat2, ystat2, xycor2, xyn2 )
!
! Purpose
! _______
!
!   UPDATE_COR_MISS2 computes sample means and corrected sums of squares and cross-products
!   for a sample of size XYN+XYN2, possibly containing missing values, given the means and
!   corrected sums of squares and cross-products for two subsamples of size XYN and XYN2 
!   as output by a call to COMP_COR_MISS2 when LAST=false on the two subsamples separetely.
!
!   The sample means, variances and coefficient correlations for the sample of size XYN+XYN2
!   may be obtained by a call to COMP_COR_MISS2 with LAST=true.
!
!
! Arguments
! _________
!
!   XSTAT    (INPUT/OUTPUT) real(stnd), dimension(:,2)
!            On entry, the XSTAT argument of COMP_COR_MISS2 for the first subsample.
!            On exit, the XSTAT argument of the combined sample.
!
!   YSTAT    (INPUT/OUTPUT) real(stnd), dimension(:,2)
!            On entry, the YSTAT argument of COMP_COR_MISS2 for the first subsample.
!            On exit, the YSTAT argument of the combined sample.
!
!   XYCOR    (INPUT/OUTPUT) real(stnd), dimension(:)
!            On entry, the XYCOR argument of COMP_COR_MISS2 for the first subsample.
!            On exit, the XYCOR argument of the combined sample.
!
!   XYN      (INPUT/OUTPUT) real(stnd), dimension(:)
!            On entry, the XYN argument of COMP_COR_MISS2 for the first subsample.
!            On exit, the XYN argument of the combined sample.
!
!   XSTAT2   (INPUT) real(stnd), dimension(:,2)
!            On entry, the XSTAT argument of COMP_COR_MISS2 for the second subsample.
!
!   YSTAT2   (INPUT) real(stnd), dimension(:,2)
!            On entry, the YSTAT argument of COMP_COR_MISS2 for the second subsample.
!
!   XYCOR2   (INPUT) real(stnd), dimension(:)
!            On entry, the XYCOR argument of COMP_COR_MISS2 for the second subsample.
!
!   XYN2     (INPUT) real(stnd), dimension(:)
!            On entry, the XYN argument of COMP_COR_MISS2 for the second subsample.
!
!
! Further Details
! _______________
!
!   One possible application of this subroutine is to parallel processing. If one has two
!   or more processors available, the sample can be split up into smaller subsamples,
!   and the means and corrected sums of squares and cross-products computed for each
!   subsample independently using COMP_COR_MISS2. The means and corrected sums of squares
!   and cross-products for the original sample can then be calculated using UPDATE_COR_MISS2.
!   The means, variances and correlation coefficient for the original sample can 
!   be computed by a final call to COMP_COR_MISS2 with LAST=true.
!
!   This subroutine is adapted from:
!
!   (1) Chan, T.F., Golub, G.H, and Leveque, R.J., 1979:
!           Updating formulae and a pairwise algorithm for computing sample variances.
!           STAN-CS-79-773.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,    only : stat_error1
    use Utilities,         only : assert_eq, assert, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:,:)  :: xstat, ystat
    real(stnd), intent(inout), dimension(:)    :: xycor, xyn
!
    real(stnd), intent(in), dimension(:,:) :: xstat2, ystat2
    real(stnd), intent(in), dimension(:)   :: xycor2, xyn2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(xycor)) :: xyn_tmp, ymy2, xmx2, tmp, tmp2
!
    integer(i4b) :: m
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='update_cor_miss2'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq(  (/ int(size(xstat,1),i4b),       &
                       int(size(xstat2,1),i4b),      &
                       int(size(ystat,1),i4b),       &
                       int(size(ystat2,1),i4b),      &
                       int(size(xycor),i4b),         &
                       int(size(xycor2),i4b),        &
                       int(size(xyn),i4b),           &
                       int(size(xyn2),i4b)  /) ,     &
                       name_proc )
!
    if ( m<=0_i4b ) call merror( name_proc//stat_error1 )
!
    call assert( logical(size(xstat,2)==2,lgl),    &
                 logical(size(ystat,2)==2,lgl),    &
                 logical(size(xstat2,2)==2,lgl),   &
                 logical(size(ystat2,2)==2,lgl),   &
                 name_proc )
!
    xyn_tmp(:m)  = xyn(:m) + xyn2(:m)
    ymy2(:m)     = ystat(:m,1) - ystat2(:m,1)
    xmx2(:m)     = xstat(:m,1) - xstat2(:m,1)
!
    tmp(:m)     = ( xyn(:m)*xyn2(:m) )/xyn_tmp(:m)
    tmp2(:m)     = tmp(:m)*ymy2(:m)
!
    xycor(:m) = xycor(:m) + xycor2(:m) + tmp2(:m)*xmx2(:m)
!
    xstat(:m,2) = xstat(:m,2) +  xstat2(:m,2) + tmp(:m)*(xmx2(:m)**2)
!
    ystat(:m,2) = ystat(:m,2) +  ystat2(:m,2) + tmp2(:m)*ymy2(:m)
!
    tmp(:m)  = xyn(:m)/xyn_tmp(:m)
    tmp2(:m) = xyn2(:m)/xyn_tmp(:m)
!
    xstat(:m,1) = tmp(:m)*xstat(:m,1) + tmp2(:m)*xstat2(:m,1)
!
    ystat(:m,1) = tmp(:m)*ystat(:m,1) + tmp2(:m)*ystat2(:m,1)
!
    xyn(:m) = xyn_tmp(:m)
!
!
! END OF SUBROUTINE update_cor_miss2_rm
! _____________________________________
!
    end subroutine update_cor_miss2_rm
!
! =========================================================================================
!
    subroutine update_cor_miss2_rt( xstat, ystat, xycor, xyn, xstat2, ystat2, xycor2, xyn2 )
!
! Purpose
! _______
!
!   UPDATE_COR_MISS2 computes sample means and corrected sums of squares and cross-products
!   for a sample of size XYN+XYN2, possibly containing missing values, given the means and
!   corrected sums of squares and cross-products for two subsamples of size XYN and XYN2 
!   as output by a call to COMP_COR_MISS2 when LAST=false on the two subsamples separetely.
!
!   The sample means, variances and coefficient correlations for the sample of size XYN+XYN2
!   may be obtained by a call to COMP_COR_MISS2 with LAST=true.
!
!
! Arguments
! _________
!
!   XSTAT    (INPUT/OUTPUT) real(stnd), dimension(:,:,2)
!            On entry, the XSTAT argument of COMP_COR_MISS2 for the first subsample.
!            On exit, the XSTAT argument of the combined sample.
!
!   YSTAT    (INPUT/OUTPUT) real(stnd), dimension(:,:,2)
!            On entry, the YSTAT argument of COMP_COR_MISS2 for the first subsample.
!            On exit, the YSTAT argument of the combined sample.
!
!   XYCOR    (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, the XYCOR argument of COMP_COR_MISS2 for the first subsample.
!            On exit, the XYCOR argument of the combined sample.
!
!   XYN      (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, the XYN argument of COMP_COR_MISS2 for the first subsample.
!            On exit, the XYN argument of the combined sample.
!
!   XSTAT2   (INPUT) real(stnd), dimension(:,:,2)
!            On entry, the XSTAT argument of COMP_COR_MISS2 for the second subsample.
!
!   YSTAT2   (INPUT) real(stnd), dimension(:,:,2)
!            On entry, the YSTAT argument of COMP_COR_MISS2 for the second subsample.
!
!   XYCOR2   (INPUT) real(stnd), dimension(:,:)
!            On entry, the XYCOR argument of COMP_COR_MISS2 for the second subsample.
!
!   XYN2     (INPUT) real(stnd), dimension(:,:)
!            On entry, the XYN argument of COMP_COR_MISS2 for the second subsample.
!
!
! Further Details
! _______________
!
!   One possible application of this subroutine is to parallel processing. If one has two
!   or more processors available, the sample can be split up into smaller subsamples,
!   and the means and corrected sums of squares and cross-products computed for each
!   subsample independently using COMP_COR_MISS2. The means and corrected sums of squares
!   and cross-products for the original sample can then be calculated using UPDATE_COR_MISS2.
!   The means, variances and correlation coefficient for the original sample can 
!   be computed by a final call to COMP_COR_MISS2 with LAST=true.
!
!   This subroutine is adapted from:
!
!   (1) Chan, T.F., Golub, G.H, and Leveque, R.J., 1979:
!           Updating formulae and a pairwise algorithm for computing sample variances.
!           STAN-CS-79-773.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Char_Constants,    only : stat_error1
    use Utilities,         only : assert_eq, assert, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:,:,:)  :: xstat, ystat
    real(stnd), intent(inout), dimension(:,:)    :: xycor, xyn
!
    real(stnd), intent(in), dimension(:,:,:) :: xstat2, ystat2
    real(stnd), intent(in), dimension(:,:)   :: xycor2, xyn2
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(xycor,1),size(xycor,2)) :: xyn_tmp, ymy2, xmx2, tmp, tmp2
!
    integer(i4b) :: m, n
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='update_cor_miss2'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    m = assert_eq(  (/ int(size(xstat,1),i4b),       &
                       int(size(xstat2,1),i4b),      &
                       int(size(ystat,1),i4b),       &
                       int(size(ystat2,1),i4b),      &
                       int(size(xycor,1),i4b),       &
                       int(size(xycor2,1),i4b),      &
                       int(size(xyn,1),i4b),         &
                       int(size(xyn2,1),i4b)  /) ,   &
                       name_proc )
!
    n = assert_eq(  (/ int(size(xstat,2),i4b),       &
                       int(size(xstat2,2),i4b),      &
                       int(size(ystat,2),i4b),       &
                       int(size(ystat2,2),i4b),      &
                       int(size(xycor,2),i4b),       &
                       int(size(xycor2,2),i4b),      &
                       int(size(xyn,2),i4b),         &
                       int(size(xyn2,2),i4b)  /) ,   &
                       name_proc )
!
    if ( min(m,n)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    call assert( logical(size(xstat,3)==2,lgl),    &
                 logical(size(ystat,3)==2,lgl),    &
                 logical(size(xstat2,3)==2,lgl),   &
                 logical(size(ystat2,3)==2,lgl),   &
                 name_proc )
!
    xyn_tmp(:m,:n)  = xyn(:m,:n) + xyn2(:m,:n)
    ymy2(:m,:n)     = ystat(:m,:n,1) - ystat2(:m,:n,1)
    xmx2(:m,:n)     = xstat(:m,:n,1) - xstat2(:m,:n,1)
!
    tmp(:m,:n)     = ( xyn(:m,:n)*xyn2(:m,:n) )/xyn_tmp(:m,:n)
    tmp2(:m,:n)     = tmp(:m,:n)*ymy2(:m,:n)
!
    xycor(:m,:n) = xycor(:m,:n) + xycor2(:m,:n) + tmp2(:m,:n)*xmx2(:m,:n)
!
    xstat(:m,:n,2) = xstat(:m,:n,2) +  xstat2(:m,:n,2) + tmp(:m,:n)*(xmx2(:m,:n)**2)
!
    ystat(:m,:n,2) = ystat(:m,:n,2) +  ystat2(:m,:n,2) + tmp2(:m,:n)*ymy2(:m,:n)
!
    tmp(:m,:n)  = xyn(:m,:n)/xyn_tmp(:m,:n)
    tmp2(:m,:n) = xyn2(:m,:n)/xyn_tmp(:m,:n)
!
    xstat(:m,:n,1) = tmp(:m,:n)*xstat(:m,:n,1) + tmp2(:m,:n)*xstat2(:m,:n,1)
!
    ystat(:m,:n,1) = tmp(:m,:n)*ystat(:m,:n,1) + tmp2(:m,:n)*ystat2(:m,:n,1)
!
    xyn(:m,:n) = xyn_tmp(:m,:n)
!
!
! END OF SUBROUTINE update_cor_miss2_rt
! _____________________________________
!
    end subroutine update_cor_miss2_rt
!
!
! =========================================================================================
!                  PERMUTATION TESTS FOR CORRELATION COEFFICIENTS
! =========================================================================================
!
!
    subroutine permute_cor_rv( x, y, xstat, ystat, xycor, prob, nrep, initseed )
!
! Purpose
! _______
!
!   PERMUTE_COR performs a permutation test of a correlation
!   coefficient between two data vectors Y and X.
!
!
! Arguments
! _________
!
!   X        (INPUT/OUTPUT) real(stnd), dimension(:)
!            On entry, the input data vector X.
!            On exit, the data are standardized with the univariate statistics
!            stored in the XSTAT argument.
!
!   Y        (INPUT) real(stnd), dimension(:)
!            On entry, the input data vector Y.
!
!            The size of Y must verify: SIZE(Y) = SIZE(X) .
!
!   XSTAT    (INPUT) real(stnd), dimension(2)
!            On entry, XSTAT must contain the following statistics
!            as output by COMP_COR subroutine in argument XSTAT:
!
!            - XSTAT(1) contains the mean value of the data vector X.
!            - XSTAT(2) contains the variance of the data vector X.
!
!            The size of XSTAT must verify: size(XSTAT) = 2.
!
!   YSTAT    (INPUT) real(stnd), dimension(2)
!            On entry, YSTAT must contain the following statistics,
!            as output by COMP_COR subroutine in argument
!            YSTAT:
!
!            - YSTAT(1) contains the mean value of the data vector Y.
!            - YSTAT(2) contains the variance of the data vector Y.
!
!            The size of YSTAT must verify: size(YSTAT) = 2.
!
!   XYCOR    (INPUT) real(stnd)
!            On entry, XYCOR contains the correlation coefficient
!            between X(:) and Y(:). XYCOR must be specified 
!            as output by COMP_COR subroutine.
!
!   PROB     (OUTPUT) real(stnd)
!            On exit, PROB gives the critical probability associated
!            with XYCOR, as computed by a permutation test with NREP shuffles.
!
!   NREP    (INPUT, OPTIONAL) integer(i4b)
!            On entry, when argument NREP is present, NREP specifies the number of
!            shuffles for the permutation test of the correlation
!            coefficient. The default is 99.
!
!   INITSEED (INPUT, OPTIONAL) logical(lgl)
!            On entry, if INITSEED=true, a call to RANDOM_SEED_() without arguments
!            is done in the subroutine, in order to initiates a non-repeatable reset
!            of the seed used by the STATPACK random generator.
!            The default is INITSEED=false.
!
!
! Further Details
! _______________
!
!   This subroutine is parallelized if OPENMP is used.
!
!   For more details and algorithm, see:
!
!   (1) von Storch, H., and Zwiers, F.W., 2002:
!            Statistical Analysis in Climate Research.
!            Cambridge, UK, Chapter 8, 484 pp., ISBN:9780521012300
!
!   (2) Noreen, E.W., 1989:
!             Computer-intensive methods for testing hypotheses: an introduction.
!             Wiley and Sons, New York, USA, ISBN:978-0-471-61136-3
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : one
    use Char_Constants,    only : stat_error7, stat_error8
    use Utilities,         only : assert_eq, assert, merror
    use Random,            only : simple_shuffle, random_seed_
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:) :: x
    real(stnd), intent(in),    dimension(:) :: y, xstat, ystat
    real(stnd), intent(in)                  :: xycor
    real(stnd), intent(out)                 :: prob
!
    integer(i4b), intent(in), optional :: nrep
!
    logical(lgl), intent(in),  optional :: initseed
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                            :: cor, std
    real(stnd), dimension(size(y))        :: y2
!
    integer(i4b)                          :: nge, i, n, nrep2
!
    logical(lgl) :: init_seed
!
#ifdef _OPENMP
    integer(i4b) :: nge2, j
!
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='permute_cor'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq(  int(size(x),i4b),  &
                    int(size(y),i4b),  &
                    name_proc )
!
    if ( n<=0_i4b ) call merror( name_proc//stat_error7 )
!
    call assert( logical(size(xstat)==2,lgl),    &
                 logical(size(ystat)==2,lgl),    &
                 name_proc )
!
    init_seed = false
    if ( present(initseed) ) then
        init_seed = initseed
    end if
!
    nrep2 = 99_i4b
    if ( present(nrep) ) then
        if ( nrep<=0_i4b )        &
        call merror( name_proc//stat_error8 )
        nrep2 = nrep
    end if
!
!   STANDARDIZE THE ORIGINAL DATA VECTORS X AND Y.
!
    std = one/sqrt( xstat(2_i4b)*real(n,stnd) )
    x(:n) = ( x(:n) - xstat(1_i4b) )*std
!
    std = one/sqrt( ystat(2_i4b)*real(n,stnd) )
    y2(:n) = ( y(:n) - ystat(1_i4b) )*std
!
!   INITIALIZE THE RANDOM GENERATOR IF REQUIRED.
!
    if ( init_seed ) then
!
!$OMP SINGLE
        call random_seed_()
!$OMP END SINGLE
!
    end if
!
!   A TWO-SIDED TEST IS PERFORMED.
!
    prob = abs( xycor )
!
    nge  = 0_i4b
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and. i>1 .and. j>1
!
    if ( test_par ) then
!
!$OMP PARALLEL IF(test_par)          &
!$OMP         ,PRIVATE(cor,i,nge2)   &
!$OMP         ,FIRSTPRIVATE(y2)
!
        nge2 = 0_i4b
!
!$OMP DO SCHEDULE(STATIC) 
!
        do i=1_i4b, nrep2
!
!           PERMUTE DATA VECTOR Y.
!
            call simple_shuffle( y2(:n) )
!
!           COMPUTE CORRELATION.
!
            cor = dot_product( x(:n), y2(:n) )
!
            if ( abs( cor )>= prob ) nge2 = nge2 + 1_i4b
!
        end do
!
!$OMP END DO NOWAIT
!
!$OMP ATOMIC
        nge = nge + nge2
!
!$OMP END PARALLEL
!
    else
!
#endif
!
        do i=1_i4b, nrep2
!
!           PERMUTE DATA VECTOR Y.
!
            call simple_shuffle( y2(:n) )
!
!           COMPUTE CORRELATION.
!
            cor = dot_product( x(:n), y2(:n) )
!
            if ( abs( cor )>= prob ) nge = nge + 1_i4b
!
        end do
!
#ifdef _OPENMP
    end if
#endif
!
!   COMPUTE THE SIGNIFICANCE LEVEL.
!
    prob = real( nge+1_i4b, stnd )/real( nrep2+1_i4b, stnd )
!
!
! END OF SUBROUTINE permute_cor_rv
! ________________________________
!
    end subroutine permute_cor_rv
!
! =========================================================================================
!
    subroutine permute_cor_rm( x, y, xstat, ystat, xycor, prob, dimvar, nrep, initseed )
!
! Purpose
! _______
!
!   PERMUTE_COR performs permutation tests of correlation
!   coefficients between a data vector Y and a data matrix X.
!
!
! Arguments
! _________
!
!   X        (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, input matrix containing size(X,3-DIMVAR) observations on 
!            size(X,DIMVAR) variables for which permutation tests
!            are desired.
!            By default, DIMVAR is equal to 1. See description of optional
!            DIMVAR argument for details.
!
!            On exit, the data are standardized with the univariate statistics
!            stored in the XSTAT argument.
!
!   Y        (INPUT) real(stnd), dimension(:)
!            On entry, input vector containing size(X,3-DIMVAR) observations
!            for which permutation tests are desired.
!
!            The size of Y must verify: size(Y) = SIZE(X,3-DIMVAR) .
!
!   XSTAT    (INPUT) real(stnd), dimension(:,2)
!            On entry, XSTAT must contain the following statistics
!            on all variables, as output by COMP_COR subroutine in argument
!            XSTAT:
!
!            - XSTAT(:,1) contains the mean values of the data matrix X.
!            - XSTAT(:,2) contains the variances of the data matrix X.
!
!            The shape of XSTAT must verify:
!
!            - size(XSTAT,1) = size(X,DIMVAR) ,
!            - size(XSTAT,2) = 2 .
!
!   YSTAT    (INPUT) real(stnd), dimension(2)
!            On entry, YSTAT must contain the following statistics,
!            as output by COMP_COR subroutine in argument
!            YSTAT:
!
!            - YSTAT(1) contains the mean value of the data vector Y.
!            - YSTAT(2) contains the variance of the data vector Y.
!
!            The size of YSTAT must verify: size(YSTAT) = 2.
!
!   XYCOR    (INPUT) real(stnd), dimension(:)
!            On entry, XYCOR(i) contains the correlation coefficient
!            between XX(i,:) ( XX(:,i) if DIMVAR=2 ) and YY(:). XYCOR
!            must be specified as output by COMP_COR subroutine.
!
!            The size of XYCOR must verify: size(XYCOR) = size(X,DIMVAR).
!
!   PROB     (OUTPUT) real(stnd), dimension(:)
!            On exit, PROB(i) gives the critical probability associated
!            with XYCOR(i), as computed by a permutation test
!            with NREP shuffles.
!          
!            The size of PROB must verify: size(PROB) = size(X,DIMVAR).
!
!   DIMVAR   (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVAR is present, DIMVAR is used as follows:
!
!            - DIMVAR = 1, the input matrix X contains size(X,2) observations
!              on size(X,1) variables.
!            - DIMVAR = 2, the input matrix X contains size(X,1) observations
!              on size(X,2) variables.
!
!            The default is DIMVAR = 1.
!
!   NREP    (INPUT, OPTIONAL) integer(i4b)
!            On entry, when argument NREP is present, NREP specifies the number of
!            shuffles for the permutation test of the correlation coefficients.
!            The default is 99.
!
!   INITSEED (INPUT, OPTIONAL) logical(lgl)
!            On entry, if INITSEED=true, a call to RANDOM_SEED_() without arguments
!            is done in the subroutine, in order to initiates a non-repeatable reset
!            of the seed used by the STATPACK random generator.
!            The default is INITSEED=false.
!
!
! Further Details
! _______________
!
!   This subroutine is parallelized if OPENMP is used.
!
!   For more details and algorithm, see:
!
!   (1) von Storch, H., and Zwiers, F.W., 2002:
!            Statistical Analysis in Climate Research.
!            Cambridge, UK, Chapter 8, 484 pp., ISBN:9780521012300
!
!   (2) Noreen, E.W., 1989:
!             Computer-intensive methods for testing hypotheses: an introduction.
!             Wiley and Sons, New York, USA, ISBN:978-0-471-61136-3
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : one
    use Char_Constants,    only : stat_error1, stat_error5, stat_error7, stat_error8
    use Utilities,         only : assert_eq, assert, merror
    use Random,            only : simple_shuffle, random_seed_
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:,:)      :: x
    real(stnd), intent(in),    dimension(:)        :: y
    real(stnd), intent(in),    dimension(:,:)      :: xstat
    real(stnd), intent(in),    dimension(:)        :: ystat, xycor
    real(stnd), intent(out),   dimension(:)        :: prob
!
    integer(i4b), intent(in), optional :: dimvar, nrep
!
    logical(lgl), intent(in),  optional :: initseed
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                            :: std
    real(stnd), dimension(size(xycor))    :: cor
    real(stnd), dimension(size(y))        :: y2
!
    integer(i4b), dimension(size(xycor)) :: nge
    integer(i4b)                         :: i, n, m, nrep2
!
    logical(lgl) :: dim_is_one, init_seed
!
#ifdef _OPENMP
    integer(i4b), dimension(size(xycor)) :: nge2
    integer(i4b)                         :: j
!
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='permute_cor'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
    n = assert_eq(  int(size(x,3-int(i)),i4b),  &
                    int(size(y),i4b),           &
                    name_proc )
!
    m = assert_eq(  int(size(x,int(i)),i4b),      &
                    int(size(xstat,1),i4b),       &
                    int(size(xycor),i4b),         &
                    int(size(prob),i4b),          &
                    name_proc )
!
    if ( m<=0_i4b ) call merror( name_proc//stat_error1 )
    if ( n<=0_i4b ) call merror( name_proc//stat_error7 )
!
    call assert( logical(size(xstat,2)==2,lgl),    &
                 logical(size(ystat)==2,lgl),      &
                 name_proc )
!
    init_seed = false
    if ( present(initseed) ) then
        init_seed = initseed
    end if
!
    nrep2 = 99_i4b
    if ( present(nrep) ) then
        if ( nrep<=0_i4b )        &
        call merror( name_proc//stat_error8 )
        nrep2 = nrep
    end if
!
!   STANDARDIZE THE ORIGINAL DATA VECTOR Y.
!
    std = one/sqrt( ystat(2_i4b)*real(n,stnd) )
    y2(:n) = ( y(:n) - ystat(1_i4b) )*std
!
    cor(:m) = one/sqrt( xstat(:m,2_i4b)*real(n,stnd) )
!
!   INITIALIZE THE RANDOM GENERATOR.
!
    if ( init_seed ) then
!
!$OMP SINGLE
        call random_seed_()
!$OMP END SINGLE
!
    end if
!
!   A TWO-SIDED TEST IS PERFORMED.
!
    prob(:m)  = abs( xycor(:m) )
!
    nge(:m) = 0_i4b
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and. i>1 .and. j>1
!
    if ( test_par ) then
!
!$OMP PARALLEL IF(test_par)             &
!$OMP         ,PRIVATE(i,nge2)          &
!$OMP         ,FIRSTPRIVATE(y2,cor)
!
!
        nge2(:m) = 0_i4b
!
!       STANDARDIZE THE ORIGINAL DATA X.
!
        if ( dim_is_one ) then
!
!$OMP DO SCHEDULE(STATIC) 
!
            do i=1_i4b, n
                x(:m,i) = ( x(:m,i) - xstat(:m,1_i4b) )*cor(:m)
            end do
!
!$OMP END DO
!
        else
!
!$OMP DO SCHEDULE(STATIC) 
!
            do i=1_i4b, m
                x(:n,i) = ( x(:n,i) - xstat(i,1_i4b) )*cor(i)
            end do
!
!$OMP END DO
!
        end if
!
!$OMP DO SCHEDULE(STATIC) 
!
        do i=1_i4b, nrep2
!
!           PERMUTE DATA VECTOR Y.
!
            call simple_shuffle( y2(:n) )
!
!           COMPUTE CORRELATIONS.
!
            if ( dim_is_one ) then
!
                cor(:m) = matmul( x(:m,:n), y2(:n) )
!
            else
!
                cor(:m) = matmul( y2(:n), x(:n,:m) )
!
            end if
!
            where( abs( cor(:m) )>= prob(:m) ) nge2(:m) = nge2(:m) + 1_i4b
!
        end do
!
!$OMP END DO NOWAIT
!
!$OMP CRITICAL (permute)
!
        nge(:m) = nge(:m) + nge2(:m)
!
!$OMP END CRITICAL (permute)
!
!$OMP END PARALLEL
!
    else
!
#endif
!
!       STANDARDIZE THE ORIGINAL DATA X.
!
        if ( dim_is_one ) then
!
            do i=1_i4b, n
                x(:m,i) = ( x(:m,i) - xstat(:m,1_i4b) )*cor(:m)
            end do
!
        else
!
            do i=1_i4b, m
                x(:n,i) = ( x(:n,i) - xstat(i,1_i4b) )*cor(i)
            end do
!
        end if
!
        do i=1_i4b, nrep2
!
!           PERMUTE DATA VECTOR Y.
!
            call simple_shuffle( y2(:n) )
!
!           COMPUTE CORRELATIONS.
!
            if ( dim_is_one ) then
!
                cor(:m) = matmul( x(:m,:n), y2(:n) )
!
            else
!
                cor(:m) = matmul( y2(:n), x(:n,:m) )
!
            end if
!
            where( abs( cor(:m) )>= prob(:m) ) nge(:m) = nge(:m) + 1_i4b
!
        end do
!
#ifdef _OPENMP
    end if
#endif
!
!   COMPUTE THE SIGNIFICANCE LEVELS.
!
    std      = one/real( nrep2+1_i4b, stnd )
    prob(:m) = real( nge(:m)+1_i4b, stnd )*std
!
!
! END OF SUBROUTINE permute_cor_rm
! ________________________________
!
    end subroutine permute_cor_rm
!
!
! =========================================================================================
!                  PHASE SCRAMBLE BOOTSTRAP TESTS FOR CORRELATION COEFFICIENTS
! =========================================================================================
!
!
    subroutine phase_scramble_cor_rv( x, y, xstat, ystat, xycor, prob, nrep, method, &
                                      norm, initseed )
!
! Purpose
! _______
!
!   PHASE_SCRAMBLE_COR performs phase-scrambled bootstrap tests of a correlation
!   coefficient between two data vectors Y and X.
!
!
! Arguments
! _________
!
!   X        (INPUT/OUTPUT) real(stnd), dimension(:)
!            On entry, the input data vector X.
!            On exit, the data are standardized with the univariate statistics
!            stored in the XSTAT argument.
!
!   Y        (INPUT) real(stnd), dimension(:)
!            On entry, the input data vector Y.
!
!            The size of Y must verify: SIZE(Y) = SIZE(X) .
!
!   XSTAT    (INPUT) real(stnd), dimension(2)
!            On entry, XSTAT must contain the following statistics
!            as output by COMP_COR subroutine in argument XSTAT:
!
!            - XSTAT(1) contains the mean value of the data vector X.
!            - XSTAT(2) contains the variance of the data vector X.
!
!            The size of XSTAT must verify: size(XSTAT) = 2.
!
!   YSTAT    (INPUT) real(stnd), dimension(2)
!            On entry, YSTAT must contain the following statistics,
!            as output by COMP_COR subroutine in argument
!            YSTAT:
!
!            - YSTAT(1) contains the mean value of the data vector Y.
!            - YSTAT(2) contains the variance of the data vector Y.
!
!            The size of YSTAT must verify: size(YSTAT) = 2.
!
!   XYCOR    (INPUT) real(stnd)
!            On entry, XYCOR contains the correlation coefficient
!            between X(:) and Y(:). XYCOR must be specified 
!            as output by COMP_COR subroutine.
!
!   PROB     (OUTPUT) real(stnd)
!            On exit, PROB gives the critical probability associated
!            with XYCOR, as computed by a phase-scrambled bootstrap test
!            with NREP shuffles.
!
!   NREP     (INPUT, OPTIONAL) integer(i4b)
!            On entry, when argument NREP is present, NREP specifies the number of
!            shuffles for the phase-scrambled bootstrap test of the correlation
!            coefficient. The default is 99.
!
!   METHOD   (INPUT, OPTIONAL) integer(i4b)
!            On entry, determine the phase randomisation algorithm used to generate
!            surrogate series.
!
!            On entry, if
!
!            - METHOD = 1 : the phase randomisation algorithm of Theiler is used;
!            - METHOD = 2 : the phase randomisation algorithm of Davison and Hinkley is used.
!
!            The default is METHOD = 1.
!
!   NORM     (INPUT, OPTIONAL) logical(lgl)
!            On entry, if NORM=true, then normal margins are used in the phase-scrambled
!            algorithm, otherwise exact empirical margins are used.
!            The default is NORM=true.
!
!   INITSEED (INPUT, OPTIONAL) logical(lgl)
!            On entry, if INITSEED=true, a call to RANDOM_SEED_() without arguments
!            is done in the subroutine, in order to initiates a non-repeatable reset
!            of the seed used by the STATPACK random generator.
!            The default is INITSEED=false.
!
!
! Further Details
! _______________
!
!   This subroutine is parallelized if OPENMP is used.
!
!   The tests are adapted from:
!
!   (1) Ebisuzaki, W., 1997:
!            A method to estimate the statistical significance of a correlation
!            when the data are serially correlated.
!            Journal of climate, vol. 10, 2147-2153.
!
!   (2) Davison, A.C., and Hinkley, D.V., 1997:
!            Bootstrap methods and their application.
!            Cambridge University Press, Cambridge, UK. doi:10.1017/CBO9780511802843
!
!   (3) Theiler, J., Eubank, S., Longtin, A., Galdrikian, B., and Farmer, J.D., 1992:
!            Testing for nonlinearity in time series: the method of surrogate data.
!            Physica D, vol. 58, 77-94, doi:10.1016/0167-2789(92)90102-s
!
!   (4) Braun, W.J., and Kulperger, R.J., 1997:
!            Properties of a fourier bootstrap method for time series.
!            Communications in Statistics - Theory and Methods, vol 26, 1329-1336,
!            doi:10.1080/03610929708831985
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true, false
    use Reals_Constants,   only : zero, one, two, pi, twopi
    use Char_Constants,    only : stat_error7, stat_error8, stat_error9, allocate_error
    use Utilities,         only : assert_eq, assert, merror, arth
    use Sort_Procedures,   only : do_index
    use Random,            only : random_seed_, random_number_
    use Prob_Procedures,   only : pinvn
    use FFT_Procedures,    only : init_fft, fft, end_fft
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:) :: x
    real(stnd), intent(in),    dimension(:) :: y, xstat, ystat
    real(stnd), intent(in)                  :: xycor
    real(stnd), intent(out)                 :: prob
!
    integer(i4b), intent(in), optional :: nrep, method
!
    logical(lgl), intent(in),  optional :: norm, initseed
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                            :: cor, std, invsqrttwo
    real(stnd), dimension(size(y))        :: e, e2
    real(stnd), dimension(:), allocatable :: y2
!
    complex(stnd), dimension(size(y))         :: ffte, cf
    complex(stnd), dimension(:), allocatable  :: f
!
    integer(i4b), dimension(:), allocatable :: inde, ranke, indi
    integer(i4b)                            :: nge, i, k, n, nd2, nrep2
    integer                                 :: iok
!
    logical(lgl) :: use_empir, init_seed, use_davison, even
!
#ifdef _OPENMP
    integer(i4b) :: nge2, j
!
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='phase_scramble_cor'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq(  int(size(x),i4b),  &
                    int(size(y),i4b),  &
                    name_proc )
!
    if ( n<=0_i4b ) call merror( name_proc//stat_error7 )
!
    call assert( logical(size(xstat)==2,lgl),    &
                 logical(size(ystat)==2,lgl),    &
                 name_proc )
!
    use_empir = false
    if ( present(norm) ) then
        use_empir = .not. norm
    end if
!
    init_seed = false
    if ( present(initseed) ) then
        init_seed = initseed
    end if
!
    use_davison = false
    if ( present(method) ) then
!
        if ( method/=1_i4b .and. method/=2_i4b )      &
               call merror( name_proc//stat_error9 )
!
        use_davison = method==2_i4b
!
    end if
!
    nrep2 = 99_i4b
    if ( present(nrep) ) then
!
        if ( nrep<=0_i4b )        &
        call merror( name_proc//stat_error8 )
        nrep2 = nrep
!
    end if
!
!   STANDARDIZE THE ORIGINAL DATA SERIES.
!
    std = one/sqrt( xstat(2_i4b)*real(n,stnd) )
    x(:n) = ( x(:n) - xstat(1_i4b) )*std
!
    std = one/sqrt(ystat(2_i4b)*real(n,stnd))
    e(:n) = (y(:n) - ystat(1_i4b))*std
!
    if ( use_empir ) then
!
!       ALLOCATE WORK ARRAYS .
!
        allocate( y2(n), indi(n), inde(n), ranke(n), stat=iok )
!
        if ( iok/=0 ) call merror( name_proc//allocate_error )
!
!       SORT THE SERIES Y.
!
        call do_index( e(:n), inde(:n) )
        y2(:n) = e( inde(:n) )
!
!       COMPUTE THE RANK OF THE SERIES Y.
!
        indi(:n) = arth( 1_i4b, 1_i4b, n )
        ranke(inde(:n)) = indi(:n)
!
!       COMPUTE THE RESCALED SERIES.
!
        std = one/real( n+1_i4b, stnd )
        e2(:n) = real( ranke(:n), stnd )*std
        e(:n) = pinvn( e2(:n) )
!
    end if
!
!   INITIALIZE THE FFT SUBROUTINE.
!
    call init_fft( n )
!
!   TRANSFORM THE ORIGINAL OR RESCALED SEQUENCE Y.
!
    ffte(:n) = cmplx( e(:n), zero, kind=stnd )
!
    call fft( ffte(:n), forward=true  )
!
    if ( use_davison ) then
!
!       SET UP CONSTANTS FOR THE Davison-Hinkley ALGORITHM.
!
        invsqrttwo = one/sqrt(two)
!
        k  = n
!
    else
!
!       SET UP CONSTANTS FOR THE Theiler ALGORITHM.
!
        even = n==(n/2_i4b)*2_i4b
!
        if ( even ) then
            nd2 = n/2_i4b
            k   = nd2 + 1_i4b
        else
            nd2 = (n+1_i4b)/2_i4b
            k   = nd2
        end if
!
    end if
!
!   INITIALIZE THE RANDOM GENERATOR IF REQUIRED.
!
    if ( init_seed ) then
!
!$OMP SINGLE
        call random_seed_()
!$OMP END SINGLE
!
    end if
!
!   A TWO-SIDED TEST IS PERFORMED.
!
    prob = abs( xycor )
!
    nge = 0_i4b
!
#ifdef _OPENMP
!
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and. i>1 .and. j>1
!
    if ( test_par ) then
!
        if ( use_empir ) then
!
!           DEALLOCATE WORK ARRAYS.
!   
            deallocate( inde, ranke )
!
        end if
!
!$OMP PARALLEL PRIVATE(e, e2, f, cf, std, cor, inde, ranke, i, iok, nge2)
!
        nge2 = 0_i4b
!
        if ( use_empir ) then
!
!           ALLOCATE WORK ARRAYS IF NECESSARY.
!
            allocate( inde(n), ranke(n), stat=iok )
!
            if ( iok/=0 ) call merror( name_proc//allocate_error )
!
        end if
!
        if ( use_davison ) then
!
!           ALLOCATE WORK ARRAY IF NECESSARY.
!
            allocate( f(n), stat=iok )
!
            if ( iok/=0 ) call merror( name_proc//allocate_error )
!
        end if
!
        e(1_i4b) = zero
!
!$OMP DO SCHEDULE(STATIC) 
!
        do i=1_i4b, nrep2
!
!           GENERATE A RANDOM NUMBER SEQUENCE.
!
            call random_number_(e(2_i4b:k))
!
            e(2_i4b:k) = twopi*e(2_i4b:k)
!
!           PHASE SCRAMBLING ALGORITHM.
!
            if ( use_davison ) then
!
!               Davison-Hinkley METHOD.
!
                f(2_i4b:n)  = ffte(2_i4b:n)*cmplx( cos(e(2_i4b:n)), sin(e(2_i4b:n)), kind=stnd)
!                f(2_i4b:n)  = abs(ffte(2_i4b:n))*cmplx( cos(e(2_i4b:n)), sin(e(2_i4b:n)), kind=stnd)
                cf(1_i4b)   = ffte(1_i4b)
                cf(2_i4b:n) = (conjg(f(n:2_i4b:-1_i4b))+f(2_i4b:n))*invsqrttwo
!
            else
!
!               Theiler METHOD.
!
                e2(2_i4b:nd2) = e(2_i4b:nd2)
                e(k+1_i4b:n)  = -e2(nd2:2_i4b:-1_i4b)
!
                if ( even ) then
                    e(k) = merge( zero, pi, e(k)<=pi )
                end if
!
                cf(:n) = ffte(:n)*cmplx( cos(e(:n)), sin(e(:n)), kind=stnd)
!                cf(:n) = abs(ffte(:n))*cmplx( cos(e(:n)), sin(e(:n)), kind=stnd)
!
            end if
!
!           INVERT THE SEQUENCE BACK.
!
            call fft( cf(:n), forward=false  )
            e2(:n) = real( cf(:n), kind=stnd )
!
            if ( use_empir ) then
!
!               RESCALING SURROGATE STEP.
!
                call do_index( e2(:n), inde(:n) )
                ranke(inde(:n)) = indi(:n)
                e2(:n) = y2(ranke(:n))
!
            else if ( use_davison ) then
!
!               STANDARDIZE THE TIME SERIES SINCE THE Davison-Hinkley ALGORITHM
!               DOES NOT PRESERVE THE VARIANCE OF THE TIME SERIES.
!
                std = one/sqrt( sum( e2(:n)**2 ) )
                e2(:n) = e2(:n)*std
!
            end if
!
!           COMPUTE CORRELATION.
!
            cor = dot_product( x(:n), e2(:n) )
!
            if ( abs( cor )>=prob ) nge2 = nge2 + 1_i4b
!
        end do
!
!$OMP END DO NOWAIT
!
!$OMP ATOMIC
        nge = nge + nge2
!
        if ( use_empir ) then
!
!           DEALLOCATE WORK ARRAYS IF NECESSARY.
!   
            deallocate( inde, ranke )
!   
        end if
!
        if ( use_davison ) then
!
!           DEALLOCATE WORK ARRAYS IF NECESSARY.
!   
            deallocate( f )
!   
        end if
!
!$OMP END PARALLEL
!
    else
!
#endif
!
        if ( use_davison ) then
!
!           ALLOCATE WORK ARRAY IF NECESSARY.
!
            allocate( f(n), stat=iok )
            if ( iok/=0 ) call merror( name_proc//allocate_error )
!
        end if
!
        e(1_i4b) = zero
!
        do i=1_i4b, nrep2
!
!           GENERATE A RANDOM NUMBER SEQUENCE.
!
            call random_number_(e(2:k))
!
            e(2_i4b:k) = twopi*e(2_i4b:k)
!
!           PHASE SCRAMBLING ALGORITHM.
!
            if ( use_davison ) then
!
!               Davison-Hinkley METHOD.
!
                f(2_i4b:n)  = ffte(2_i4b:n)*cmplx( cos(e(2_i4b:n)), sin(e(2_i4b:n)), kind=stnd)
!                f(2_i4b:n)  = abs(ffte(2_i4b:n))*cmplx( cos(e(2_i4b:n)), sin(e(2_i4b:n)), kind=stnd)
                cf(1_i4b)   = ffte(1_i4b)
                cf(2_i4b:n) = (conjg(f(n:2_i4b:-1_i4b))+f(2_i4b:n))*invsqrttwo
!
            else
!
!               Theiler METHOD.
!
                e2(2_i4b:nd2)  = e(2_i4b:nd2)
                e(k+1_i4b:n)   = -e2(nd2:2_i4b:-1_i4b)
!
                if ( even ) then
                    e(k) = merge( zero, pi, e(k)<=pi )
                end if
!
                cf(:n) = ffte(:n)*cmplx( cos(e(:n)), sin(e(:n)), kind=stnd)
!                cf(:n) = abs(ffte(:n))*cmplx( cos(e(:n)), sin(e(:n)), kind=stnd)
!
            end if
!
!           INVERT THE SEQUENCE BACK.
!
            call fft( cf(:n), forward=false  )
            e2(:n) = real( cf(:n), kind=stnd )
!
            if ( use_empir ) then
!
!               RESCALING SURROGATE STEP.
!
                call do_index( e2(:n), inde(:n) )
                ranke(inde(:n)) = indi(:n)
                e2(:n) = y2(ranke(:n))
!
            else if ( use_davison ) then
!
!               STANDARDIZE THE TIME SERIES SINCE THE Davison-Hinkley ALGORITHM
!               DOES NOT PRESERVE THE VARIANCE OF THE TIME SERIES.
!
                std = one/sqrt( sum( e2(:n)**2 ) )
                e2(:n) = e2(:n)*std
!
            end if
!
!           COMPUTE CORRELATION.
!
            cor = dot_product( x(:n), e2(:n) )
!
            if ( abs( cor )>=prob ) nge = nge + 1_i4b
!
        end do
!
        if ( use_empir ) then
!
!           DEALLOCATE WORK ARRAYS IF NECESSARY.
!   
            deallocate( inde, ranke )
!   
        end if
!
        if ( use_davison ) then
!
!           DEALLOCATE WORK ARRAYS IF NECESSARY.
!   
            deallocate( f )
!   
        end if
!
#ifdef _OPENMP
    end if
#endif
!
!   COMPUTE THE SIGNIFICANCE LEVEL.
!
    prob = real( nge+1_i4b, stnd )/real( nrep2+1_i4b, stnd )
!
!   DEALLOCATE WORK ARRAYS.
!   
    call end_fft()
!
    if ( use_empir ) then
        deallocate( y2, indi )
    end if
!
!
! END OF SUBROUTINE phase_scramble_cor_rv
! _______________________________________
!
    end subroutine phase_scramble_cor_rv
!
! =========================================================================================
!
    subroutine phase_scramble_cor_rm( x, y, xstat, ystat, xycor, prob, dimvar, nrep, &
                                      method, norm, initseed )
!
! Purpose
! _______
!
!   PHASE_SCRAMBLE_COR performs phase-scrambled bootstrap tests of correlation
!   coefficients between a data vector Y and a data matrix X.
!
!
! Arguments
! _________
!
!   X        (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, input matrix containing size(X,3-DIMVAR) observations on 
!            size(X,DIMVAR) variables for which phase-scrambled bootstrap tests
!            are desired.
!            By default, DIMVAR is equal to 1. See description of optional
!            DIMVAR argument for details.
!
!            On exit, the data are standardized with the univariate statistics
!            stored in the XSTAT argument.
!
!   Y        (INPUT) real(stnd), dimension(:)
!            On entry, input vector containing size(X,3-DIMVAR) observations
!            for which phase-scrambled bootstrap tests are desired.
!
!            The size of Y must verify: size(Y) = SIZE(X,3-DIMVAR) .
!
!   XSTAT    (INPUT) real(stnd), dimension(:,2)
!            On entry, XSTAT must contain the following statistics
!            on all variables, as output by COMP_COR subroutine in argument
!            XSTAT:
!
!            - XSTAT(:,1) contains the mean values of the data matrix X.
!            - XSTAT(:,2) contains the variances of the data matrix X.
!
!            The shape of XSTAT must verify:
!
!            - size(XSTAT,1) = size(X,DIMVAR) ,
!            - size(XSTAT,2) = 2 .
!
!   YSTAT    (INPUT) real(stnd), dimension(2)
!            On entry, YSTAT must contain the following statistics,
!            as output by COMP_COR subroutine in argument
!            YSTAT:
!
!            - YSTAT(1) contains the mean value of the data vector Y.
!            - YSTAT(2) contains the variance of the data vector Y.
!
!            The size of YSTAT must verify: size(YSTAT) = 2.
!
!   XYCOR    (INPUT) real(stnd), dimension(:)
!            On entry, XYCOR(i) contains the correlation coefficient
!            between XX(i,:) ( XX(:,i) if DIMVAR=2 ) and YY(:). XYCOR
!            must be specified as output by COMP_COR subroutine.
!
!            The size of XYCOR must verify: size(XYCOR) = size(X,DIMVAR).
!
!   PROB     (OUTPUT) real(stnd), dimension(:)
!            On exit, PROB(i) gives the critical probability associated
!            with XYCOR(i), as computed by a phase-scrambled bootstrap test
!            with NREP shuffles.
!          
!            The size of PROB must verify: size(PROB) = size(X,DIMVAR).
!
!   DIMVAR   (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVAR is present, DIMVAR is used as follows, if:
!
!            - DIMVAR = 1, the input matrix X contains size(X,2) observations
!              on size(X,1) variables.
!            - DIMVAR = 2, the input matrix X contains size(X,1) observations
!              on size(X,2) variables.
!
!            The default is DIMVAR = 1.
!
!   NREP    (INPUT, OPTIONAL) integer(i4b)
!            On entry, when argument NREP is present, NREP specifies the number of
!            shuffles for the phase-scrambled bootstrap test of the correlation
!            coefficients. The default is 99.
!
!   METHOD   (INPUT, OPTIONAL) integer(i4b)
!            On entry, determine the phase randomisation algorithm used to generate
!            surrogate series.
!
!            On entry, if
!
!            - METHOD = 1 : the phase randomisation algorithm of Theiler is used;
!            - METHOD = 2 : the phase randomisation algorithm of Davison and Hinkley is used.
!
!            The default is METHOD = 1.
!
!   NORM     (INPUT, OPTIONAL) logical(lgl)
!            On entry, if NORM=true, then normal margins are used in the phase-scrambled
!            algorithm, otherwise exact empirical margins are used.
!            The default is NORM=true.
!
!   INITSEED (INPUT, OPTIONAL) logical(lgl)
!            On entry, if INITSEED=true, a call to RANDOM_SEED_() without arguments
!            is done in the subroutine, in order to initiates a non-repeatable reset
!            of the seed used by the STATPACK random generator.
!            The default is INITSEED=false.
!
!
! Further Details
! _______________
!
!   This subroutine is parallelized if OPENMP is used.
!
!   The tests are adapted from:
!
!   (1) Ebisuzaki, W., 1997:
!            A method to estimate the statistical significance of a correlation
!            when the data are serially correlated.
!            Journal of climate, vol. 10, 2147-2153.
!
!   (2) Davison, A.C., and Hinkley, D.V., 1997:
!            Bootstrap methods and their application.
!            Cambridge University Press, Cambridge, UK. doi:10.1017/CBO9780511802843
!
!   (3) Theiler, J., Eubank, S., Longtin, A., Galdrikian, B., and Farmer, J.D., 1992:
!            Testing for nonlinearity in time series: the method of surrogate data.
!            Physica D, vol. 58, 77-94, doi:10.1016/0167-2789(92)90102-s
!
!   (4) Braun, W.J., and Kulperger, R.J., 1997:
!            Properties of a fourier bootstrap method for time series.
!            Communications in Statistics - Theory and Methods, vol 26, 1329-1336,
!            doi:10.1080/03610929708831985
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true, false
    use Reals_Constants,   only : zero, one, two, pi, twopi
    use Char_Constants,    only : stat_error1, stat_error5, stat_error7, stat_error8, stat_error9, allocate_error
    use Utilities,         only : assert_eq, assert, merror, arth
    use Sort_Procedures,   only : do_index
    use Random,            only : random_number_, random_seed_
    use Prob_Procedures,   only : pinvn
    use FFT_Procedures,    only : init_fft, fft, end_fft
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:,:)      :: x
    real(stnd), intent(in),    dimension(:)        :: y
    real(stnd), intent(in),    dimension(:,:)      :: xstat
    real(stnd), intent(in),    dimension(:)        :: ystat, xycor
    real(stnd), intent(out),   dimension(:)        :: prob
!
    integer(i4b), intent(in), optional :: dimvar, nrep, method
!
    logical(lgl), intent(in),  optional :: norm, initseed
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                            :: std, invsqrttwo
    real(stnd), dimension(size(xycor))    :: cor
    real(stnd), dimension(size(y))        :: e, e2
    real(stnd), dimension(:), allocatable :: y2
!
    complex(stnd), dimension(size(y))        :: ffte, cf
    complex(stnd), dimension(:), allocatable :: f
!
    integer(i4b), dimension(:), allocatable :: inde, ranke, indi
    integer(i4b), dimension(size(xycor))    :: nge
    integer(i4b)                            :: i, k, n, nd2, m, nrep2
    integer                                 :: iok
!
    logical(lgl) :: dim_is_one, use_empir, init_seed, use_davison, even
!
#ifdef _OPENMP
    integer(i4b), dimension(size(xycor))    :: nge2
    integer(i4b)                            :: j
!
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='phase_scramble_cor'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
    n = assert_eq(  int(size(x,3-int(i)),i4b),  &
                    int(size(y),i4b),           &
                    name_proc )
!
    m = assert_eq(  int(size(x,int(i)),i4b),      &
                    int(size(xstat,1),i4b),       &
                    int(size(xycor),i4b),         &
                    int(size(prob),i4b),          &
                    name_proc )
!
    if ( m<=0_i4b ) call merror( name_proc//stat_error1 )
    if ( n<=0_i4b ) call merror( name_proc//stat_error7 )
!
    call assert( logical(size(xstat,2)==2,lgl),    &
                 logical(size(ystat)==2,lgl),      &
                 name_proc )
!
    use_empir = false
    if ( present(norm) ) then
        use_empir = .not. norm
    end if
!
    init_seed = false
    if ( present(initseed) ) then
        init_seed = initseed
    end if
!
    use_davison = false
    if ( present(method) ) then
!
        if ( method/=1_i4b .and. method/=2_i4b )      &
               call merror( name_proc//stat_error9 )
!
        use_davison = method==2_i4b
!
    end if
!
    nrep2 = 99_i4b
    if ( present(nrep) ) then
!
        if ( nrep<=0_i4b )        &
        call merror( name_proc//stat_error8 )
        nrep2 = nrep
!
    end if
!
!   STANDARDIZE THE ORIGINAL DATA SERIES.
!
    std = one/sqrt(ystat(2_i4b)*real(n,stnd))
    e(:n) = (y(:n) - ystat(1_i4b))*std
!
    if ( use_empir ) then
!
!       ALLOCATE WORK ARRAYS .
!
        allocate( y2(n), indi(n), inde(n), ranke(n), stat=iok )
        if ( iok/=0 ) call merror( name_proc//allocate_error )
!
!       SORT THE SERIES Y.
!
        call do_index( e(:n), inde(:n) )
        y2(:n) = e( inde(:n) )
!
!       COMPUTE THE RANK OF THE SERIES Y.
!
        indi(:n) = arth( 1_i4b, 1_i4b, n )
        ranke(inde(:n)) = indi(:n)
!
!       COMPUTE THE RESCALED SERIES Y.
!
        std = one/real( n+1_i4b, stnd )
        e2(:n) = real( ranke(:n), stnd )*std
        e(:n) = pinvn( e2(:n) )
!
    end if
!
!   INITIALIZE THE FFT SUBROUTINE.
!
    call init_fft( n )
!
!   TRANSFORM THE ORIGINAL OR RESCALED SEQUENCE Y.
!
    ffte(:n) = cmplx( e(:n), zero, kind=stnd )
!
    call fft( ffte(:n), forward=true  )
!
    if ( use_davison ) then
!
!       SET UP CONSTANTS FOR THE Davison-Hinkley ALGORITHM.
!
        invsqrttwo = one/sqrt(two)
!
        k  = n
!
    else
!
!       SET UP CONSTANTS FOR THE Theiler ALGORITHM.
!
        even = n==(n/2_i4b)*2_i4b
!
        if ( even ) then
            nd2 = n/2_i4b
            k   = nd2 + 1_i4b
        else
            nd2 = (n+1_i4b)/2_i4b
            k   = nd2
        end if
!
    end if
!
!   INITIALIZE THE RANDOM GENERATOR IF REQUIRED.
!
    if ( init_seed ) then
!
!$OMP SINGLE
        call random_seed_()
!$OMP END SINGLE
!
    end if
!
!   A TWO-SIDED TEST IS PERFORMED.
!
    prob(:m) = abs( xycor(:m) )
!
    cor(:m) = one/sqrt( xstat(:m,2_i4b)*real(n,stnd) )
!
    nge(:m)   = 0_i4b
!
#ifdef _OPENMP
!
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and. i>1 .and. j>1
!
    if ( test_par ) then
!
        if ( use_empir ) then
!
!           DEALLOCATE WORK ARRAYS.
!   
            deallocate( inde, ranke )
!
        end if
!
!$OMP PARALLEL PRIVATE(e, e2, f, cf, std, inde, ranke, i, iok, nge2)   &
!$OMP         ,FIRSTPRIVATE(cor)
!
        nge2(:m)   = 0_i4b
!
!       STANDARDIZE THE ORIGINAL DATA X.
!
        if ( dim_is_one ) then
!
!$OMP DO SCHEDULE(STATIC) 
!
            do i=1_i4b, n
                x(:m,i) = ( x(:m,i) - xstat(:m,1_i4b) )*cor(:m)
            end do
!
!$OMP END DO
!
        else
!
!$OMP DO SCHEDULE(STATIC) 
!
            do i=1_i4b, m
                x(:n,i) = ( x(:n,i) - xstat(i,1_i4b) )*cor(i)
            end do
!
!$OMP END DO
!
        end if
!
        if ( use_empir ) then
!
!           ALLOCATE WORK ARRAYS IF NECESSARY.
!
            allocate( inde(n), ranke(n), stat=iok )
!
            if ( iok/=0 ) call merror( name_proc//allocate_error )
!
        end if
!
        if ( use_davison ) then
!
!           ALLOCATE WORK ARRAY IF NECESSARY.
!
            allocate( f(n), stat=iok )
!
            if ( iok/=0 ) call merror( name_proc//allocate_error )
!
        end if
!
        e(1_i4b) = zero
!
!$OMP DO SCHEDULE(STATIC) 
!
        do i=1_i4b, nrep2
!
!           GENERATE A RANDOM NUMBER SEQUENCE.
!
            call random_number_(e(2_i4b:k))
!
            e(2_i4b:k) = twopi*e(2_i4b:k)
!
!           PHASE SCRAMBLING ALGORITHM.
!
            if ( use_davison ) then
!
!               Davison-Hinkley METHOD.
!
                f(2_i4b:n)  = ffte(2_i4b:n)*cmplx( cos(e(2_i4b:n)), sin(e(2_i4b:n)), kind=stnd)
!                f(2_i4b:n)  = abs(ffte(2_i4b:n))*cmplx( cos(e(2_i4b:n)), sin(e(2_i4b:n)), kind=stnd)
                cf(1_i4b)   = ffte(1_i4b)
                cf(2_i4b:n) = (conjg(f(n:2_i4b:-1_i4b))+f(2_i4b:n))*invsqrttwo
!
            else
!
!               Theiler METHOD.
!
                e2(2_i4b:nd2) = e(2_i4b:nd2)
                e(k+1_i4b:n)  = -e2(nd2:2_i4b:-1_i4b)
!
                if ( even ) then
                    e(k) = merge( zero, pi, e(k)<=pi )
                end if
!
                cf(:n) = ffte(:n)*cmplx( cos(e(:n)), sin(e(:n)), kind=stnd)
!                cf(:n) = abs(ffte(:n))*cmplx( cos(e(:n)), sin(e(:n)), kind=stnd)
!
            end if
!
!           INVERT THE SEQUENCE BACK.
!
            call fft( cf(:n), forward=false  )
            e2(:n) = real( cf(:n), kind=stnd )
!
            if ( use_empir ) then
!
!               RESCALING SURROGATE STEP.
!
                call do_index( e2(:n), inde(:n) )
                ranke(inde(:n)) = indi(:n)
                e2(:n) = y2(ranke(:n))
!
            else if ( use_davison ) then
!
!               STANDARDIZE THE TIME SERIES SINCE THE Davison-Hinkley ALGORITHM
!               DOES NOT PRESERVE THE VARIANCE OF THE TIME SERIES.
!
                std = one/sqrt( sum( e2(:n)**2 ) )
                e2(:n) = e2(:n)*std
!
            end if
!
!           COMPUTE CORRELATIONS.
!
            if ( dim_is_one ) then
!
                cor(:m) = matmul( x(:m,:n), e2(:n) )
!
            else
!
                cor(:m) = matmul( e2(:n), x(:n,:m) )
!
            end if
!
            where( abs( cor(:m) )>= prob(:m) ) nge2(:m) = nge2(:m) + 1_i4b
!
        end do
!
!$OMP END DO NOWAIT
!
!$OMP CRITICAL  (scramble)
!
        nge(:m) = nge(:m) + nge2(:m)
!
!$OMP END CRITICAL  (scramble)
!
!
        if ( use_empir ) then
!
!           DEALLOCATE WORK ARRAYS IF NECESSARY.
!   
            deallocate( inde, ranke )
!   
        end if
!
        if ( use_davison ) then
!
!           DEALLOCATE WORK ARRAYS IF NECESSARY.
!   
            deallocate( f )
!   
        end if
!
!$OMP END PARALLEL
!
    else
!
#endif
!
!       STANDARDIZE THE ORIGINAL DATA X.
!
        if ( dim_is_one ) then
!
            do i=1_i4b, n
                x(:m,i) = ( x(:m,i) - xstat(:m,1_i4b) )*cor(:m)
            end do
!
        else
!
            do i=1_i4b, m
                x(:n,i) = ( x(:n,i) - xstat(i,1_i4b) )*cor(i)
            end do
!
        end if
!
        if ( use_davison ) then
!
!           ALLOCATE WORK ARRAY IF NECESSARY.
!
            allocate( f(n), stat=iok )
!
            if ( iok/=0 ) call merror( name_proc//allocate_error )
!
        end if
!
        e(1_i4b) = zero
!
        do i=1_i4b, nrep2
!
!           GENERATE A RANDOM NUMBER SEQUENCE.
!
            call random_number_(e(2:k))
!
            e(2_i4b:k) = twopi*e(2_i4b:k)
!
!           PHASE SCRAMBLING ALGORITHM.
!
            if ( use_davison ) then
!
!               Davison-Hinkley METHOD.
!
                f(2_i4b:n)  = ffte(2_i4b:n)*cmplx( cos(e(2_i4b:n)), sin(e(2_i4b:n)), kind=stnd)
!                f(2_i4b:n)  = abs(ffte(2_i4b:n))*cmplx( cos(e(2_i4b:n)), sin(e(2_i4b:n)), kind=stnd)
                cf(1_i4b)   = ffte(1_i4b)
                cf(2_i4b:n) = (conjg(f(n:2_i4b:-1_i4b))+f(2_i4b:n))*invsqrttwo
!
            else
!
!               Theiler METHOD.
!
                e2(2_i4b:nd2)  = e(2_i4b:nd2)
                e(k+1_i4b:n)   = -e2(nd2:2_i4b:-1_i4b)
!
                if ( even ) then
                    e(k) = merge( zero, pi, e(k)<=pi )
                end if
!
                cf(:n) = ffte(:n)*cmplx( cos(e(:n)), sin(e(:n)), kind=stnd)
!                cf(:n) = abs(ffte(:n))*cmplx( cos(e(:n)), sin(e(:n)), kind=stnd)
!
            end if
!
!           INVERT THE SEQUENCE BACK.
!
            call fft( cf(:n), forward=false  )
            e2(:n) = real( cf(:n), kind=stnd )
!
            if ( use_empir ) then
!
!               RESCALING SURROGATE STEP.
!
                call do_index( e2(:n), inde(:n) )
                ranke(inde(:n)) = indi(:n)
                e2(:n) = y2(ranke(:n))
!
            else if ( use_davison ) then
!
!               STANDARDIZE THE TIME SERIES SINCE THE Davison-Hinkley ALGORITHM
!               DOES NOT PRESERVE THE VARIANCE OF THE TIME SERIES.
!
                std = one/sqrt( sum( e2(:n)**2 ) )
                e2(:n) = e2(:n)*std
!
            end if
!
!           COMPUTE CORRELATIONS.
!
            if ( dim_is_one ) then
!
                cor(:m) = matmul( x(:m,:n), e2(:n) )
!
            else
!
                cor(:m) = matmul( e2(:n), x(:n,:m) )
!
            end if
!
            where( abs( cor(:m) )>= prob(:m) ) nge(:m) = nge(:m) + 1_i4b
!
        end do
!
        if ( use_empir ) then
!
!           DEALLOCATE WORK ARRAYS IF NECESSARY.
!   
            deallocate( inde, ranke )
!   
        end if
!
        if ( use_davison ) then
!
!           DEALLOCATE WORK ARRAYS IF NECESSARY.
!   
            deallocate( f )
!   
        end if
!
#ifdef _OPENMP
    end if
#endif
!
!   COMPUTE THE SIGNIFICANCE LEVELS.
!
    std      = one/real( nrep2+1_i4b, stnd )
    prob(:m) = real( nge(:m)+1_i4b, stnd )*std
!
!   DEALLOCATE WORK ARRAYS.
!
    call end_fft()
!
    if ( use_empir ) then
!
!       DEALLOCATE WORK ARRAYS IF NECESSARY.
!   
        deallocate( y2, indi )
!
    end if
!
!
! END OF SUBROUTINE phase_scramble_cor_rm
! _______________________________________
!
    end subroutine phase_scramble_cor_rm
!
! =========================================================================================
!
    subroutine bootstrap_cor_rv( x, y, xstat, xycor, prob, nrep, initseed, periodicity, &
                                 season, block_size )
!
! Purpose
! _______
!
!   BOOTSTRAP_COR performs a moving block bootstrap test of a correlation
!   coefficient between two data vectors X and Y.
!
!
! Arguments
! _________
!
!   X           (INPUT/OUTPUT) real(stnd), dimension(:)
!               On entry, the input data vector X.
!
!               On exit, the data are standardized with the univariate statistics
!               stored in the XSTAT argument.
!
!   Y           (INPUT) real(stnd), dimension(:)
!               On entry, the input data vector Y.
!
!               The size of Y must verify: SIZE(Y) = SIZE(X) .
!
!   XSTAT       (INPUT) real(stnd), dimension(2)
!               On entry, XSTAT must contain the following statistics
!               as output by COMP_COR subroutine in argument XSTAT:
!
!               - XSTAT(1) contains the mean value of the data vector X.
!               - XSTAT(2) contains the variance of the data vector X.
!
!               The size of XSTAT must verify: size(XSTAT) = 2.
!
!   XYCOR       (INPUT) real(stnd)
!               On entry, XYCOR contains the correlation coefficient
!               between X(:) and Y(:). XYCOR must be specified 
!               as output by COMP_COR subroutine.
!
!   PROB        (OUTPUT) real(stnd)
!               On exit, PROB gives the critical probability associated
!               with XYCOR, as computed by a moving block bootstrap test with NREP shuffles.
!
!   NREP        (INPUT, OPTIONAL) integer(i4b)
!               On entry, when argument NREP is present, NREP specifies the number of
!               shuffles for the moving block bootstrap test of the correlation
!               coefficient. The default is 99.
!
!   INITSEED    (INPUT, OPTIONAL) logical(lgl)
!               On entry, if INITSEED=true, a call to RANDOM_SEED_() without arguments
!               is done in the subroutine, in order to initiates a non-repeatable reset
!               of the seed used by the STATPACK random generator.
!               The default is INITSEED=false.
!
!   PERIODICITY (INPUT, OPTIONAL) integer(i4b)
!               On entry, argument PERIODICITY specifies that the indice, i, of the first observation
!               of each selected block in the moving block bootstrap algorithm verifies the condition
!               i=1+(PERIODICITY * j) where j is a random positive integer.
!               PERIODICITY must be greater than zero and less than size(X).
!               By default, PERIODICITY is set to 1.
!
!   SEASON      (INPUT, OPTIONAL) integer(i4b)
!               On entry, argument SEASON specifies that the input time series is a repetition of the
!               same season for different years and SEASON specifies the length of the season.
!               SEASON must be greater than zero and size(X) must be a multiple of SEASON. If the optional
!               argument PERIODICITY is used, SEASON must also be greater or equal to PERIODICITY.
!               By default, SEASON is set to size(X).
!
!   BLOCK_SIZE  (INPUT, OPTIONAL) integer(i4b)
!               On entry, argument BLOCK_SIZE specifies the size of the block in the moving block bootstrap.
!               BLOCK_SIZE must be greater than zero and less than size(X). If the optional argument
!               PERIODICITY is used, BLOCK_SIZE must also be greater or equal to PERIODICITY.
!               Moreover, if the optional argument SEASON is used, BLOCK_SIZE must also be less
!               than SEASON.
!               By default, BLOCK_SIZE is set to 1 or to PERIODICITY if this optional argument is used.
!
!
! Further Details
! _______________
!
!   This subroutine is parallelized if OPENMP is used.
!
!   The test is adapted from:
!
!   (1) Davison, A.C., and Hinkley, D.V., 1997:
!           Bootstrap methods and their application.
!           Cambridge University Press, Cambridge, UK. doi:10.1017/CBO9780511802843
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : one
    use Char_Constants,    only : stat_error7, stat_error8, stat_error10, stat_error11,    &
                                  stat_error12, stat_error13, stat_error14, stat_error15,  &
                                  stat_error16, allocate_error
    use Utilities,         only : assert_eq, assert, merror
    use Random,            only : random_seed_, random_number_
    use Stat_Procedures,   only : comp_mvs
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:) :: x
    real(stnd), intent(in),    dimension(:) :: y, xstat
    real(stnd), intent(in)                  :: xycor
    real(stnd), intent(out)                 :: prob
!
    integer(i4b), intent(in), optional :: nrep, block_size, periodicity, season
    
!
    logical(lgl), intent(in),  optional :: initseed
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                            :: cor, ystd, ymean, tmp, nblock_max
    real(stnd), dimension(size(y))        :: y2
    real(stnd), dimension(:), allocatable :: randvec, randvec_seas
!
    integer(i4b)                            :: nge, i, j, j1, j2, k, k1, k2, l, l1, p, n, nrep2,  &
                                               block_size2, nblock, nblock2, per, seas, n_seas
    integer(i4b), dimension(:), allocatable :: index
    integer                                 :: iok
!
    logical(lgl) :: init_seed
!
#ifdef _OPENMP
    integer(i4b) :: nge2
!
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='bootstrap_cor'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    n = assert_eq(  int(size(x),i4b),  &
                    int(size(y),i4b),  &
                    name_proc )
!
    if ( n<=0_i4b ) call merror( name_proc//stat_error7 )
!
    call assert( logical(size(xstat)==2,lgl),    &
                 name_proc )
!
    init_seed = false
    if ( present(initseed) ) then
        init_seed = initseed
    end if
!
    nrep2 = 99_i4b
    if ( present(nrep) ) then
        if ( nrep<=0_i4b )        &
        call merror( name_proc//stat_error8 )
        nrep2 = nrep
    end if
!
    if ( present( periodicity ) ) then
!
        if ( periodicity<=0_i4b .or. periodicity>=n )        &
        call merror( name_proc//stat_error10 )        
!
        per = periodicity
!
    else
!
        per = 1_i4b
!
    end if
!
    if ( present(block_size) ) then
!
        if ( block_size<=0_i4b .or. block_size>=n )        &
        call merror( name_proc//stat_error11 )
!
!       THE BLOCK SIZE MUST BE GREATER THAN THE PERIODICITY.
!
        if ( block_size < per )        &
        call merror( name_proc//stat_error12 ) 
!
        block_size2 = block_size
!
    else
!
        block_size2 = per
!
    end if
!
    if ( present( season ) ) then
!
        if ( season<=0_i4b )        &
        call merror( name_proc//stat_error13 )        
!
!       THE LENGTH OF THE TIME SERIES MUST BE A MULTIPLE OF THE SEASON.
!
        if ( (n/season)*season/=n )        &
        call merror( name_proc//stat_error14 )
!
!       THE BLOCK SIZE MUST BE LESS THAN THE SEASON.
!
        if ( block_size2>season )        &
        call merror( name_proc//stat_error15 ) 
!
!       THE PERIODICITY MUST BE LESS THAN THE SEASON.
!
        if ( per>season )        &
        call merror( name_proc//stat_error16 ) 
!
        seas   = season
        n_seas = n/seas
!
    else
!
        seas=n
        n_seas=1_i4b
!
    end if
!
!   CALCULATE MAXIMUM NUMBER OF BLOCKS AND REAL NUMBER OF BLOCKS.
!   
    nblock_max = real( (seas - block_size2 + 1_i4b)/per, stnd )
!
    nblock2 = seas/block_size2
    if ( nblock2*block_size2/=seas ) nblock2 = nblock2 + 1_i4b
!
    nblock = n_seas*nblock2
!
!   INITIALIZE THE RANDOM GENERATOR IF REQUIRED.
!
    if ( init_seed ) then
!
!$OMP SINGLE
        call random_seed_()
!$OMP END SINGLE
!
    end if
!
!   A TWO-SIDED TEST IS PERFORMED.
!
    prob = abs( xycor )
!   
!   INITIALIZE SIGNIFICANCE LEVEL.
!
    nge  = 0_i4b
!
!   STANDARDIZE THE ORIGINAL DATA VECTOR X.
!
    tmp = one/( sqrt(xstat(2_i4b))*real(n,stnd) )
    x(:n) = ( x(:n) - xstat(1_i4b) )*tmp
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and. i>1 .and. j>1
!
    if ( test_par ) then
!
!$OMP PARALLEL IF(test_par)          &
!$OMP         ,PRIVATE(i,j,j1,j2,k,k1,k2,l,l1,p,nge2,iok,tmp,ymean,ystd,cor,y2,randvec,randvec_seas,index)
!
        nge2 = 0_i4b
!    
!       ALLOCATE WORK ARRAYS. 
!    
        allocate( randvec(nblock), index(nblock) , stat=iok )
!
        if ( iok/=0 ) then
!
            call merror( name_proc//allocate_error )
!
        end if
!    
        if ( n_seas==1_i4b ) then
!
!$OMP DO SCHEDULE(STATIC) 
!    
            do i=1_i4b, nrep2
!    
                call random_number_( randvec(:nblock) )
!
                index(:nblock) = int( randvec(:nblock)*nblock_max, i4b )
!
                if ( per/=1_i4b) index(:nblock) = index(:nblock)*per
!    
                do j=1_i4b, nblock
!    
!                   j1 IS THE START OF THE CURRENT BLOCK.  
!                   j2 IS THE END OF THE CURRENT BLOCK.  
!    
                    j1 = (j-1_i4b)*block_size2 + 1_i4b
                    j2 = min( j*block_size2, n)
!    
!                   k1 IS THE START OF THE CURRENT RANDOM BLOCK.  
!                   k2 IS THE END OF THE CURRENT RANDOM BLOCK.  
!    
                    k1 = index(j) + 1_i4b
                    k2 = k1 + j2 - j1
!    
                    y2(j1:j2) = y(k1:k2)
!    
                end do
!
!               COMPUTE UNIVARIATE STATISTICS FOR THE SYNTHETIC TIME SERIES.
!
                call comp_mvs(  y2(:n), true, true, ymean, tmp, ystd )
!
!               STANDARDIZE THE SYNTHETIC TIME SERIES.
!
                tmp = one/ystd
                y2(:n) = ( y2(:n) - ymean )*tmp
!
!               COMPUTE CORRELATION.
!
                cor = dot_product( x(:n), y2(:n) )
!
                if ( abs( cor )>=prob ) nge2 = nge2 + 1_i4b
!    
            end do
!
!$OMP END DO NOWAIT
!    
        else
!    
!           ALLOCATE WORK ARRAY. 
!    
            allocate( randvec_seas(nblock), stat=iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!$OMP DO SCHEDULE(STATIC) 
!    
            do i=1_i4b, nrep2
!    
                call random_number_( randvec_seas(:nblock) )
                call random_number_( randvec(:nblock) )
!
                index(:nblock) = int( randvec(:nblock)*nblock_max, i4b )
!
                if ( per/=1_i4b) index(:nblock) = index(:nblock)*per
!    
!               j IS THE NUMBER OF THE CURRENT BLOCK.  
!    
                j = 0_i4b
!
                do l=1_i4b, n_seas
!    
                    j2 = (l-1_i4b)*seas
                    l1 = l*seas
!    
                    do k=1_i4b, nblock2
!    
!                       j1 IS THE START OF THE CURRENT BLOCK.  
!                       j2 IS THE END OF THE CURRENT BLOCK.  
!    
                        j1 = j2 + 1_i4b
                        j2 = min( j2+block_size2, l1 )
!    
                        j   = j + 1_i4b
                        tmp = randvec_seas(j)*real( n_seas, stnd )
                        p   = int( tmp, i4b )*seas
!    
!                       k1 IS THE START OF THE CURRENT RANDOM BLOCK.  
!                       k2 IS THE END OF THE CURRENT RANDOM BLOCK.  
!                  
                        k1  = p + index(j) + 1_i4b
                        k2  = k1 + j2 - j1
!    
                        y2(j1:j2) = y(k1:k2)
!    
                    end do
!    
                end do
!
!               COMPUTE UNIVARIATE STATISTICS FOR THE SYNTHETIC TIME SERIES.
!
                call comp_mvs(  y2(:n), true, true, ymean, tmp, ystd )
!
!               STANDARDIZE THE SYNTHETIC TIME SERIES.
!
                tmp = one/ystd
                y2(:n) = ( y2(:n) - ymean )*tmp
!
!               COMPUTE CORRELATION.
!
                cor = dot_product( x(:n), y2(:n) )
!
                if ( abs( cor )>=prob ) nge2 = nge2 + 1_i4b
!    
            end do
!
!$OMP END DO NOWAIT
!
!           DEALLOCATE WORK ARRAY.
!    
            deallocate( randvec_seas )
!    
        end if
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( randvec, index )
!
!$OMP ATOMIC
        nge = nge + nge2
!
!$OMP END PARALLEL
!
    else
!
#endif
!    
!       ALLOCATE WORK ARRAYS. 
!    
        allocate( randvec(nblock), index(nblock) , stat=iok )
!
        if ( iok/=0 ) then
           call merror( name_proc//allocate_error )
        end if
!    
        if ( n_seas==1_i4b ) then
!    
            do i=1_i4b, nrep2
!    
                call random_number_( randvec(:nblock) )
!
                index(:nblock) = int( randvec(:nblock)*nblock_max, i4b )
!
                if ( per/=1_i4b) index(:nblock) = index(:nblock)*per
!    
                do j=1_i4b, nblock
!    
!                   j1 IS THE START OF THE CURRENT BLOCK.  
!                   j2 IS THE END OF THE CURRENT BLOCK.  
!    
                    j1 = (j-1_i4b)*block_size2 + 1_i4b
                    j2 = min( j*block_size2, n)
!    
!                   k1 IS THE START OF THE CURRENT RANDOM BLOCK.  
!                   k2 IS THE END OF THE CURRENT RANDOM BLOCK.  
!    
                    k1 = index(j) + 1_i4b
                    k2 = k1 + j2 - j1
!    
                    y2(j1:j2) = y(k1:k2)
!    
                end do
!
!               COMPUTE UNIVARIATE STATISTICS FOR THE SYNTHETIC TIME SERIES.
!
                call comp_mvs( y2(:n), true, true, ymean, tmp, ystd )
!
!               STANDARDIZE THE SYNTHETIC TIME SERIES.
!
                tmp = one/ystd
                y2(:n) = ( y2(:n) - ymean )*tmp
!
!               COMPUTE CORRELATION.
!
                cor = dot_product( x(:n), y2(:n) )
!
                if ( abs( cor )>=prob ) nge = nge + 1_i4b
!    
            end do
!    
        else
!    
!           ALLOCATE WORK ARRAY. 
!    
            allocate( randvec_seas(nblock), stat=iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!    
            do i=1_i4b, nrep2
!    
                call random_number_( randvec_seas(:nblock) )
                call random_number_( randvec(:nblock) )
!    
                index(:nblock) = int( randvec(:nblock)*nblock_max, i4b )
!
                if ( per/=1_i4b) index(:nblock) = index(:nblock)*per
!    
!               j IS THE NUMBER OF THE CURRENT BLOCK.  
!    
                j = 0_i4b
!
                do l=1_i4b, n_seas
!    
                    j2 = (l-1_i4b)*seas
                    l1 = l*seas
!    
                    do k=1_i4b, nblock2
!    
!                       j1 IS THE START OF THE CURRENT BLOCK.  
!                       j2 IS THE END OF THE CURRENT BLOCK.  
!    
                        j1 = j2 + 1_i4b
                        j2 = min( j2+block_size2, l1 )
!    
                        j   = j + 1_i4b
                        tmp = randvec_seas(j)*real( n_seas, stnd )
                        p   = int( tmp, i4b )*seas
!    
!                       k1 IS THE START OF THE CURRENT RANDOM BLOCK.  
!                       k2 IS THE END OF THE CURRENT RANDOM BLOCK.  
!                  
                        k1  = p + index(j) + 1_i4b
                        k2  = k1 + j2 - j1
!    
                        y2(j1:j2) = y(k1:k2)
!    
                    end do
!    
                end do
!
!               COMPUTE UNIVARIATE STATISTICS FOR THE SYNTHETIC TIME SERIES.
!
                call comp_mvs(  y2(:n), true, true, ymean, tmp, ystd )
!
!               STANDARDIZE THE SYNTHETIC TIME SERIES.
!
                tmp = one/ystd
                y2(:n) = ( y2(:n) - ymean )*tmp
!
!               COMPUTE CORRELATION.
!
                cor = dot_product( x(:n), y2(:n) )
!
                if ( abs( cor )>=prob ) nge = nge + 1_i4b
!    
            end do
!
!           DEALLOCATE WORK ARRAY.
!    
            deallocate( randvec_seas )
!    
        end if
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( randvec, index )
!
#ifdef _OPENMP
    end if
#endif
!
!   COMPUTE THE SIGNIFICANCE LEVEL.
!
    prob = real( nge+1_i4b, stnd )/real( nrep2+1_i4b, stnd )
!
!
! END OF SUBROUTINE bootstrap_cor_rv
! __________________________________
!
    end subroutine bootstrap_cor_rv
!
! =========================================================================================
!
    subroutine bootstrap_cor_rm( x, y, xstat, xycor, prob, dimvar, nrep, initseed, &
                                 periodicity, season, block_size )
!
! Purpose
! _______
!
!   BOOTSTRAP_COR performs a moving block bootstrap test of a correlation
!   coefficients between a data vector Y and a data matrix X.
!
!
! Arguments
! _________
!
!   X           (INPUT/OUTPUT) real(stnd), dimension(:,:)
!               On entry, input matrix containing size(X,3-DIMVAR) observations on 
!               size(X,DIMVAR) variables for which moving block bootstrap tests
!               are desired.
!               By default, DIMVAR is equal to 1. See description of optional
!               DIMVAR argument for details.
!
!               On exit, the data are standardized with the univariate statistics
!               stored in the XSTAT argument.
!
!   Y           (INPUT) real(stnd), dimension(:)
!               On entry, input vector containing size(X,3-DIMVAR) observations
!               for which moving block bootstrap tests are desired.
!
!               The size of Y must verify: size(Y) = size(X,3-DIMVAR) .
!
!   XSTAT       (INPUT) real(stnd), dimension(:,2)
!               On entry, XSTAT must contain the following statistics
!               on all variables, as output by COMP_COR subroutine in argument
!               XSTAT:
!
!               - XSTAT(:,1) contains the mean values of the data matrix X.
!               - XSTAT(:,2) contains the variances of the data matrix X.
!
!               The shape of XSTAT must verify:
!
!               - size(XSTAT,1) = size(X,DIMVAR) ,
!               - size(XSTAT,2) = 2 .
!
!   XYCOR       (INPUT) real(stnd), dimension(:)
!               On entry, XYCOR(i) contains the correlation coefficient
!               between XX(i,:) ( XX(:,i) if DIMVAR=2 ) and YY(:). XYCOR
!               must be specified as output by COMP_COR subroutine.
!
!               The size of XYCOR must verify: size(XYCOR) = size(X,DIMVAR).
!
!   PROB        (OUTPUT) real(stnd), dimension(:)
!               On exit, PROB(i) gives the critical probability associated
!               with XYCOR(i), as computed by a moving block bootstrap test
!               with NREP shuffles.
!          
!               The size of PROB must verify: size(PROB) = size(X,DIMVAR).
!
!   DIMVAR      (INPUT, OPTIONAL) integer(i4b)
!               On entry, if DIMVAR is present, DIMVAR is used as follows, if:
!
!               - DIMVAR = 1, the input matrix X contains size(X,2) observations
!                 on size(X,1) variables.
!               - DIMVAR = 2, the input matrix X contains size(X,1) observations
!                 on size(X,2) variables.
!
!               The default is DIMVAR = 1.
!
!   NREP        (INPUT, OPTIONAL) integer(i4b)
!               On entry, when argument NREP is present, NREP specifies the number of
!               shuffles for the moving block bootstrap test of the correlation
!               coefficient. The default is 99.
!
!   INITSEED    (INPUT, OPTIONAL) logical(lgl)
!               On entry, if INITSEED=true, a call to RANDOM_SEED_() without arguments
!               is done in the subroutine, in order to initiates a non-repeatable reset
!               of the seed used by the STATPACK random generator.
!               The default is INITSEED=false.
!
!   PERIODICITY (INPUT, OPTIONAL) integer(i4b)
!               On entry, argument PERIODICITY specifies that the indice, i, of the first observation
!               of each selected block in the moving block bootstrap algorithm verifies the condition
!               i=1+(PERIODICITY * j) where j is a random positive integer.
!               PERIODICITY must be greater than zero and less than size(X,3-DIMVAR).
!               By default, PERIODICITY is set to 1.
!
!   SEASON      (INPUT, OPTIONAL) integer(i4b)
!               On entry, argument SEASON specifies that the input time series is a repetition of the
!               same season for different years and SEASON specifies the length of the season.
!               SEASON must be greater than zero and size(X) must be a multiple of SEASON. If the optional
!               argument PERIODICITY is used, SEASON must also be greater or equal to PERIODICITY.
!               By default, SEASON is set to size(X,3-DIMVAR).
!
!   BLOCK_SIZE  (INPUT, OPTIONAL) integer(i4b)
!               On entry, argument BLOCK_SIZE specifies the size of the block in the moving block bootstrap.
!               BLOCK_SIZE must be greater than zero and less than size(X). If the optional argument
!               PERIODICITY is used, BLOCK_SIZE must also be greater or equal to PERIODICITY.
!               Moreover, if the optional argument SEASON is used, BLOCK_SIZE must also be less
!               than SEASON.
!               By default, BLOCK_SIZE is set to 1 or to PERIODICITY if this optional argument is used.
!
!
! Further Details
! _______________
!
!   This subroutine is parallelized if OPENMP is used.
!
!   The test is adapted from:
!
!   (1) Davison, A.C., and Hinkley, D.V., 1997:
!           Bootstrap methods and their application.
!           Cambridge University Press, Cambridge, UK. doi:10.1017/CBO9780511802843
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : one
    use Char_Constants,    only : stat_error1, stat_error5, stat_error7, stat_error8,      &
                                  stat_error10, stat_error11, stat_error12, stat_error13,  &
                                  stat_error14, stat_error15, stat_error16, allocate_error
    use Utilities,         only : assert_eq, assert, merror
    use Random,            only : random_seed_, random_number_
    use Stat_Procedures,   only : comp_mvs
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(inout), dimension(:,:)      :: x
    real(stnd), intent(in),    dimension(:)        :: y
    real(stnd), intent(in),    dimension(:,:)      :: xstat
    real(stnd), intent(in),    dimension(:)        :: xycor
    real(stnd), intent(out),   dimension(:)        :: prob
!
    integer(i4b), intent(in), optional :: dimvar, nrep, block_size, periodicity, season
    
!
    logical(lgl), intent(in),  optional :: initseed
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                            :: ystd, ymean, tmp, nblock_max
    real(stnd), dimension(size(xycor))    :: cor, tmpvec
    real(stnd), dimension(size(y))        :: y2
    real(stnd), dimension(:), allocatable :: randvec, randvec_seas
!
    integer(i4b)                            ::  i, j, j1, j2, k, k1, k2, l, l1, m, n, p, nrep2,  &
                                                block_size2, nblock, nblock2, per, seas, n_seas
    integer(i4b), dimension(size(xycor))    :: nge
    integer(i4b), dimension(:), allocatable :: index
    integer                                 :: iok
!
    logical(lgl) :: dim_is_one, init_seed
!
#ifdef _OPENMP
    integer(i4b), dimension(size(xycor)) :: nge2
!
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='bootstrap_cor'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
    n = assert_eq(  int(size(x,3-int(i)),i4b),  &
                    int(size(y),i4b),           &
                    name_proc )
!
    m = assert_eq(  int(size(x,int(i)),i4b),      &
                    int(size(xstat,1),i4b),       &
                    int(size(xycor),i4b),         &
                    int(size(prob),i4b),          &
                    name_proc )
!
    if ( m<=0_i4b ) call merror( name_proc//stat_error1 )
    if ( n<=0_i4b ) call merror( name_proc//stat_error7 )
!
    call assert( logical(size(xstat,2)==2,lgl),    &
                 name_proc )
!
    init_seed = false
    if ( present(initseed) ) then
        init_seed = initseed
    end if
!
    nrep2 = 99_i4b
    if ( present(nrep) ) then
        if ( nrep<=0_i4b )        &
        call merror( name_proc//stat_error8 )
        nrep2 = nrep
    end if
!
    if ( present( periodicity ) ) then
!
!       THE LENGTH OF THE TIME SERIES MUST BE GREATER THAN THE PERIODICITY.
!
        if ( periodicity<=0_i4b .or. periodicity>=n )        &
        call merror( name_proc//stat_error10 )        
!
        per = periodicity
!
    else
!
        per = 1_i4b
!
    end if
!
    if ( present(block_size) ) then
!
        if ( block_size<=0_i4b .or. block_size>=n )        &
        call merror( name_proc//stat_error11 )
!
!       THE BLOCK SIZE MUST BE GREATER THAN THE PERIODICITY.
!
        if ( block_size < per )        &
        call merror( name_proc//stat_error12 ) 
!
        block_size2 = block_size
!
    else
!
        block_size2 = per
!
    end if
!
    if ( present( season ) ) then
!
        if ( season<=0_i4b )        &
        call merror( name_proc//stat_error13 )        
!
!       THE LENGTH OF THE TIME SERIES MUST BE A MULTIPLE OF THE SEASON.
!
        if ( (n/season)*season/=n )        &
        call merror( name_proc//stat_error14 )
!
!       THE BLOCK SIZE MUST BE LESS THAN THE SEASON.
!
        if ( block_size2>season )        &
        call merror( name_proc//stat_error15 ) 
!
!       THE PERIODICITY MUST BE LESS THAN THE SEASON.
!
        if ( per>season )        &
        call merror( name_proc//stat_error16 ) 
!
        seas   = season
        n_seas = n/seas
!
    else
!
        seas=n
        n_seas=1_i4b
!
    end if
!
!   CALCULATE MAXIMUM NUMBER OF BLOCKS AND REAL NUMBER OF BLOCKS.
!   
    nblock_max = real( (seas - block_size2 + 1_i4b)/per, stnd )
!
    nblock2 = seas/block_size2
    if ( nblock2*block_size2/=seas ) nblock2 = nblock2 + 1_i4b
!
    nblock = n_seas*nblock2
!
!   INITIALIZE THE RANDOM GENERATOR IF REQUIRED.
!
    if ( init_seed ) then
!
!$OMP SINGLE
        call random_seed_()
!$OMP END SINGLE
!
    end if
!
!   A TWO-SIDED TEST IS PERFORMED.
!
    tmpvec(:m) = one/(sqrt( xstat(:m,2_i4b))*real(n,stnd) )
!
    prob(:m)  = abs( xycor(:m) )
!   
!   INITIALIZE SIGNIFICANCE LEVEL.
!
    nge(:m) = 0_i4b
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and. i>1 .and. j>1
!
    if ( test_par ) then
!
!$OMP PARALLEL IF(test_par)          &
!$OMP         ,PRIVATE(i,j,j1,j2,k,k1,k2,l,l1,p,nge2,iok,tmp,ymean,ystd,cor,y2,randvec,randvec_seas,index)
!
        nge2(:m) = 0_i4b
!
!       STANDARDIZE THE ORIGINAL DATA X.
!
        if ( dim_is_one ) then
!
!$OMP DO SCHEDULE(STATIC) 
!
            do i=1_i4b, n
                x(:m,i) = ( x(:m,i) - xstat(:m,1_i4b) )*tmpvec(:m)
            end do
!
!$OMP END DO
!
        else
!
!$OMP DO SCHEDULE(STATIC) 
!
            do i=1_i4b, m
                x(:n,i) = ( x(:n,i) - xstat(i,1_i4b) )*tmpvec(i)
            end do
!
!$OMP END DO
!
        end if
!    
!       ALLOCATE WORK ARRAYS. 
!    
        allocate( randvec(nblock), index(nblock) , stat=iok )
!
        if ( iok/=0 ) then
            call merror( name_proc//allocate_error )
        end if
!    
        if ( n_seas==1_i4b ) then
!
!$OMP DO SCHEDULE(STATIC) 
!    
            do i=1_i4b, nrep2
!    
                call random_number_( randvec(:nblock) )
!    
                index(:nblock) = int( randvec(:nblock)*nblock_max, i4b )
                if ( per/=1_i4b) index(:nblock) = index(:nblock)*per
!    
                do j=1_i4b, nblock
!    
!                  j1 IS THE START OF THE CURRENT BLOCK.  
!                  j2 IS THE END OF THE CURRENT BLOCK.  
!    
                   j1 = (j-1_i4b)*block_size2 + 1_i4b
                   j2 = min( j*block_size2, n)
!    
!                  K1 IS THE START OF THE CURRENT RANDOM BLOCK.  
!                  K2 IS THE END OF THE CURRENT RANDOM BLOCK.  
!    
                   k1 = index(j) + 1_i4b
                   k2 = k1 + j2 - j1
!    
                   y2(j1:j2) = y(k1:k2)
!    
                end do
!
!               COMPUTE UNIVARIATE STATISTICS FOR THE SYNTHETIC TIME SERIES.
!
                call comp_mvs(  y2(:n), true, true, ymean, tmp, ystd )
!
!               STANDARDIZE THE SYNTHETIC TIME SERIES.
!
                tmp = one/ystd
                y2(:n) = ( y2(:n) - ymean )*tmp
!
!               COMPUTE CORRELATION.
!
                if ( dim_is_one ) then
!
                    cor(:m) = matmul( x(:m,:n), y2(:n) )
!
                else
!
                    cor(:m) = matmul( y2(:n), x(:n,:m) )
!
                end if
!
                where( abs( cor(:m) )>=prob(:m) ) nge2(:m) = nge2(:m) + 1_i4b
!    
            end do
!
!$OMP END DO NOWAIT
!    
        else
!    
!           ALLOCATE WORK ARRAY. 
!    
            allocate( randvec_seas(nblock), stat=iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!
!$OMP DO SCHEDULE(STATIC) 
!    
            do i=1_i4b, nrep2
!    
                call random_number_( randvec_seas(:nblock) )
                call random_number_( randvec(:nblock) )
!    
                index(:nblock) = int( randvec(:nblock)*nblock_max, i4b )
                if ( per/=1_i4b) index(:nblock) = index(:nblock)*per
!    
!               j IS THE NUMBER OF THE CURRENT BLOCK.  
!    
                j = 0_i4b
!
                do l=1_i4b, n_seas
!    
                    j2 = (l-1_i4b)*seas
                    l1 = l*seas
!    
                    do k=1_i4b, nblock2
!    
!                       j1 IS THE START OF THE CURRENT BLOCK.  
!                       j2 IS THE END OF THE CURRENT BLOCK.  
!    
                        j1 = j2 + 1_i4b
                        j2 = min( j2+block_size2, l1 )
!    
                        j   = j + 1_i4b
                        tmp = randvec_seas(j)*real( n_seas, stnd )
                        p   = int( tmp, i4b )*seas
!    
!                       K1 IS THE START OF THE CURRENT RANDOM BLOCK.  
!                       K2 IS THE END OF THE CURRENT RANDOM BLOCK.  
!                  
                        k1  = p + index(j) + 1_i4b
                        k2  = k1 + j2 - j1
!    
                        y2(j1:j2) = y(k1:k2)
!    
                    end do
!    
                end do    
!
!               COMPUTE UNIVARIATE STATISTICS FOR THE SYNTHETIC TIME SERIES.
!
                call comp_mvs( y2(:n), true, true, ymean, tmp, ystd )
!
!               STANDARDIZE THE SYNTHETIC TIME SERIES.
!
                tmp = one/ystd
                y2(:n) = ( y2(:n) - ymean )*tmp
!
!               COMPUTE CORRELATION.
!
                if ( dim_is_one ) then
!
                    cor(:m) = matmul( x(:m,:n), y2(:n) )
!
                else
!
                    cor(:m) = matmul( y2(:n), x(:n,:m) )
!
                end if
!
                where( abs( cor(:m) )>=prob(:m) ) nge2(:m) = nge2(:m) + 1_i4b
!
             end do
!
!$OMP END DO NOWAIT
!
!            DEALLOCATE WORK ARRAY.
!    
             deallocate( randvec_seas )
!    
        end if
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( randvec, index )
!
!$OMP CRITICAL (bootstrap)
!    
        nge(:m) = nge(:m) + nge2(:m)
!
!$OMP END CRITICAL (bootstrap)
!
!$OMP END PARALLEL
!
!
    else
!
#endif
!   
!       STANDARDIZE THE ORIGINAL DATA X.
!
        if ( dim_is_one ) then
!
           do i=1_i4b, n
              x(:m,i) = ( x(:m,i) - xstat(:m,1_i4b) )*tmpvec(:m)
           end do
!
        else
!
           do i=1_i4b, m
              x(:n,i) = ( x(:n,i) - xstat(i,1_i4b) )*tmpvec(i)
           end do
!
        end if
!    
!       ALLOCATE WORK ARRAYS. 
!    
        allocate( randvec(nblock), index(nblock) , stat=iok )
!
        if ( iok/=0 ) then
           call merror( name_proc//allocate_error )
        end if
!    
        if ( n_seas==1_i4b ) then
!    
            do i=1_i4b, nrep2
!    
                call random_number_( randvec(:nblock) )
!
                index(:nblock) = int( randvec(:nblock)*nblock_max, i4b )
                if ( per/=1_i4b) index(:nblock) = index(:nblock)*per
!    
                do j=1_i4b, nblock
!    
!                   j1 IS THE START OF THE CURRENT BLOCK.  
!                   j2 IS THE END OF THE CURRENT BLOCK.  
!    
                    j1 = (j-1_i4b)*block_size2 + 1_i4b
                    j2 = min( j*block_size2, n)
!    
!                   K1 IS THE START OF THE CURRENT RANDOM BLOCK.  
!                   K2 IS THE END OF THE CURRENT RANDOM BLOCK.  
!    
                    k1 = index(j) + 1_i4b
                    k2 = k1 + j2 - j1
!    
                    y2(j1:j2) = y(k1:k2)
!    
                end do
!
!               COMPUTE UNIVARIATE STATISTICS FOR THE SYNTHETIC TIME SERIES.
!
                call comp_mvs(  y2(:n), true, true, ymean, tmp, ystd )
!
!               STANDARDIZE THE SYNTHETIC TIME SERIES.
!
                tmp = one/ystd
                y2(:n) = ( y2(:n) - ymean )*tmp
!
!               COMPUTE CORRELATION.
!
                if ( dim_is_one ) then
!
                    cor(:m) = matmul( x(:m,:n), y2(:n) )
!
                else
!
                    cor(:m) = matmul( y2(:n), x(:n,:m) )
!
                end if
!
                where( abs( cor(:m) )>= prob(:m) ) nge(:m) = nge(:m) + 1_i4b
!    
            end do
!    
        else
!    
!           ALLOCATE WORK ARRAY. 
!    
            allocate( randvec_seas(nblock), stat=iok )
!
            if ( iok/=0 ) then
                call merror( name_proc//allocate_error )
            end if
!    
            do i=1_i4b, nrep2
!    
                call random_number_( randvec_seas(:nblock) )
                call random_number_( randvec(:nblock) )
!    
                index(:nblock) = int( randvec(:nblock)*nblock_max, i4b )
                if ( per/=1_i4b) index(:nblock) = index(:nblock)*per
!    
!               j IS THE NUMBER OF THE CURRENT BLOCK.  
!    
                j = 0_i4b
!
                do l=1_i4b, n_seas
!    
                    j2 = (l-1_i4b)*seas
                    l1 = l*seas
!    
                    do k=1_i4b, nblock2
!    
!                       j1 IS THE START OF THE CURRENT BLOCK.  
!                       j2 IS THE END OF THE CURRENT BLOCK.  
!    
                        j1 = j2 + 1_i4b
                        j2 = min( j2+block_size2, l1 )
!    
                        j   = j + 1_i4b
                        tmp = randvec_seas(j)*real( n_seas, stnd )
                        p   = int( tmp, i4b )*seas
!    
!                       K1 IS THE START OF THE CURRENT RANDOM BLOCK.  
!                       K2 IS THE END OF THE CURRENT RANDOM BLOCK.  
!                  
                        k1  = p + index(j) + 1_i4b
                        k2  = k1 + j2 - j1
!    
                        y2(j1:j2) = y(k1:k2)
!    
                    end do
!    
                end do    
!
!               COMPUTE UNIVARIATE STATISTICS FOR THE SYNTHETIC TIME SERIES.
!
                call comp_mvs(  y2(:n), true, true, ymean, tmp, ystd )
!
!               STANDARDIZE THE SYNTHETIC TIME SERIES.
!
                tmp = one/ystd
                y2(:n) = ( y2(:n) - ymean )*tmp
!
!               COMPUTE CORRELATION.
!
                if ( dim_is_one ) then
!
                    cor(:m) = matmul( x(:m,:n), y2(:n) )
!
                else
!
                    cor(:m) = matmul( y2(:n), x(:n,:m) )
!
                end if
!
                where( abs( cor(:m) )>= prob(:m) ) nge(:m) = nge(:m) + 1_i4b
!
             end do
!
!            DEALLOCATE WORK ARRAY.
!    
             deallocate( randvec_seas )
!    
        end if
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( randvec, index )
!
#ifdef _OPENMP
    end if
#endif
!
!   COMPUTE THE SIGNIFICANCE LEVEL.
!
    prob(:m) = real( nge(:m)+1_i4b, stnd )/real( nrep2+1_i4b, stnd )
!
!
! END OF SUBROUTINE bootstrap_cor_rm
! __________________________________
!
    end subroutine bootstrap_cor_rm
!
!
! =========================================================================================
!                               CORRELATION ANALYSIS
! =========================================================================================
!
!
    subroutine comp_cormat_a( x, first, last, xmean, xcor, xn, dimvar, xstd, cov, fill, failure )
!
! Purpose
! _______
!
!   COMP_CORMAT computes estimates of means and variance-covariance or correlation
!   matrix from a data matrix.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:,:)
!            On entry, input submatrix containing size(X,3-DIMVAR)
!            observations on size(X,DIMVAR) variables from the matrix of data
!            for which means, variances and covariances are desired. By default,
!            DIMVAR is equal to 1. See description of optional DIMVAR argument 
!            for details.
!            If all the data are available at once, X can be the full 
!            data matrix.
!
!   FIRST    (INPUT) logical(lgl)
!            On entry, if:
!
!            - FIRST = true  the current submatrix is the first submatrix 
!              of the data matrix.
!            - FIRST = false the current submatrix is not the first submatrix
!              of the data matrix.
!
!   LAST     (INPUT) logical(lgl)
!            On entry, if:
!
!            - LAST = true  the current submatrix is the last submatrix 
!              of the data matrix.
!            - LAST = false the current submatrix is not the last submatrix
!              of the data matrix.
!
!   XMEAN    (INPUT/OUTPUT) real(stnd), dimension(:)
!            On entry, after the first call to COMP_CORMAT (e.g. when e.g. when FIRST=true),
!            XMEAN contains the variable means from  previous calls
!            to COMP_CORMAT. XMEAN should not be changed between calls to
!            COMP_CORMAT.
!
!            On exit, when LAST=true, XMEAN contains the variable means
!
!            The size of XMEAN must verify: size(XMEAN) = size(X,DIMVAR).
!
!   XCOR     (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, after the first call to COMP_CORMAT (e.g. when e.g. when FIRST=true),
!            the matrix XCOR contains the upper triangle of the corrected
!            sums of squared and cross-products matrix computed from  
!            previous calls to COMP_CORMAT. XCOR should not be changed
!            between calls to COMP_CORMAT.
!
!            On exit, when LAST=true, XCOR contains the upper triangle of the
!            symetric correlation or variance-covariance matrix as controlled by 
!            the COV argument. If the optional argument FILL is present and equal
!            to true, the lower triangle of XCOR is also filled.
!
!            The shape of XCOR must verify:
!
!            - size(XCOR,1) = size(X,DIMVAR) ,
!            - size(XCOR,2) = size(X,DIMVAR) .
!
!   XN       (INPUT/OUTPUT) real(stnd)
!            On entry, after the first call to COMP_CORMAT (e.g. when e.g. when FIRST=true),
!            XN contains count of observations from previous calls
!            to COMP_CORMAT. XN should not be changed between calls to
!            COMP_CORMAT.
!
!            On exit, XN contains the number of observations in the data matrix XX.
!
!   DIMVAR   (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVAR is present, DIMVAR is used as follows, if:
!
!            - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!              on size(X,1) variables.
!            - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!              on size(X,2) variables.
!
!            The default is DIMVAR = 1.
!
!   XSTD     (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            On exit, when LAST=true and XSTD is present, XSTD contains the variable 
!            standard-deviations.
!
!            The size of XSTD must verify: size(XSTD) = size(X,DIMVAR).
!
!            XSTD needs to be specified only on the last call to COMP_CORMAT (e.g. when LAST=true).
!          
!   COV      (INPUT, OPTIONAL) logical(lgl)
!            On entry, when argument COV is present, COV is used as follows, if:
!
!            - COV= true,   XCOR contains the variances-covariances matrix, when LAST=true.
!            - COV= false,  XCOR contains the correlation matrix, when LAST=true.
!
!            By default, the correlation matrix is output.
!
!            COV needs to be specified only on the last call to COMP_CORMAT (e.g. when LAST=true).
!          
!   FILL     (INPUT, OPTIONAL) logical(lgl)
!            On entry, when argument FILL is present, FILL is used as follows, if:
!
!            - FILL= true,  the lower triangle of XCOR is filled, when LAST=true.
!            - FILL= false, the lower triangle of XCOR is not filled, when LAST=true.
!
!            By default, the lower triangle of XCOR is not filled.
!
!            FILL needs to be specified only on the last call to COMP_CORMAT (e.g. when LAST=true).
!          
!   FAILURE  (OUTPUT, OPTIONAL) logical(lgl)
!            On exit, when argument FAILURE is present:
!
!            - FAILURE = false:   indicates successful exit.
!            - FAILURE = true:    indicates that fewer than two valid observations
!              were present or that the observations on some
!              variable were constant and the correlations
!              were requested.
!
!
! Further Details
! _______________
!
!   The subroutine computes the means and correlation matrix with only one pass
!   through the data.
!
!   If the observations on some variable were constant, the pertinent correlations 
!   are set to nan() code .
!
!   If fewer than two valid observations were present, the correlations are set to 
!   nan() code .
!
!   If fewer than one valid observation is present, the means are also set to 
!   nan() code .
!
!   For more details on correlation analysis, see
!
!   (1) von Storch, H., and Zwiers, F.W., 2002:
!            Statistical Analysis in Climate Research.
!            Cambridge, UK, Chapter 8, 484 pp., ISBN:9780521012300
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : zero, one
    use Num_Constants,     only : nan, lamch
    use Char_Constants,    only : stat_error1, stat_error5
    use Utilities,         only : assert_eq, assert, merror, get_diag
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),             dimension(:,:) :: x
    real(stnd), intent(inout),          dimension(:)   :: xmean
    real(stnd), intent(inout),          dimension(:,:) :: xcor
    real(stnd), intent(inout)                          :: xn
    real(stnd), intent(out),  optional , dimension(:)  :: xstd
!
    integer(i4b), intent(in),  optional :: dimvar
!
    logical(lgl), intent(in)            :: first, last
    logical(lgl), intent(in),  optional :: cov, fill
    logical(lgl), intent(out), optional :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                         :: a, b, xnm1, tmp
    real(stnd), dimension(size(xmean)) :: dxi
!
    integer(i4b) :: i, j,  n, m
!
    logical(lgl) :: cor, low_triang, dim_is_one
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_cormat'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
    n = size( x, 3-int(i) )
!
    m = assert_eq(  int(size(x,int(i)),i4b),    &
                    int(size(xmean),i4b),       &
                    int(size(xcor,1),i4b),      &
                    int(size(xcor,2),i4b),      &
                    name_proc )
!
    if ( m<=0_i4b ) call merror( name_proc//stat_error1 )
!
    if ( first ) then
!
!       THE CURRENT SUBMATRIX IS THE FIRST.
!
!       INITIALIZE xmean, xcor AND xn .
!
        xmean(1_i4b:m) = zero       
        xcor(1_i4b:m,1_i4b:m) = zero
        xn = zero
!
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  (m*(m+1_i4b))/2_i4b>=omp_limit      .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)      &
!$OMP         ,PRIVATE(i,j,tmp)
!
    do i = 1_i4b, n
!
!$OMP SINGLE
!
        xnm1 = xn
        xn = xnm1 + one
        a = one/xn
        b = xnm1*a
!
!       CALCULATE PARTIAL MEANS AND CORRECTED SUMS OF SQUARES AND CROSS-PRODUCTS, 
!       HANDLING THE DATA ONCE.
!
        if ( dim_is_one ) then
            dxi(1_i4b:m) = x(1_i4b:m,i) - xmean(1_i4b:m)
        else
            dxi(1_i4b:m) = x(i,1_i4b:m) - xmean(1_i4b:m)
        end if
!
        xmean(1_i4b:m) = xmean(1_i4b:m) + a*dxi(1_i4b:m)
!
!$OMP END SINGLE
!
!$OMP DO SCHEDULE(STATIC,1) 
!
        do j = 1_i4b, m
!
            tmp = b*dxi(j)
!
#ifdef _BLAS
            call axpy( j, tmp, dxi(1_i4b:j), 1_i4b, xcor(1_i4b:j,j), 1_i4b )
#else
            xcor(1_i4b:j,j) = xcor(1_i4b:j,j) + tmp*dxi(1_i4b:j)
#endif
!
        end do
!
!$OMP END DO
!
    end do
!
!$OMP END PARALLEL
!
    if ( last ) then
!
!       THE CURRENT SUBMATRIX IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
!       TEST OPTIONAL ARGUMENTS.
!
        if ( present(cov) ) then
           cor = .not.cov
        else
           cor = true
        end if
!
        if ( present(fill) ) then
           low_triang = fill
        else
           low_triang = false
        end if
!
        if ( present(xstd) ) then
            call assert( logical(size(xstd)==m,lgl),    &
                         name_proc )
        end if
!
        b = nan()
!
        if ( xn>one ) then
!
!           MORE THAN TWO VALID OBSERVATIONS, COMPUTE THE STATISTICS.
!
            a = one/xn    
!
            if ( cor ) then
!
!               GET DIAGONAL ELEMENTS OF CORRECTED SUMS OF SQUARES 
!               AND CROSS-PRODUCTS MATRIX.
!
                dxi(1_i4b:m) = sqrt( get_diag( xcor ) )
!
                if ( present(xstd) ) then
!
!                   COMPUTE STANDARD-DEVIATIONS, IF REQUIRED.
!
                    xstd(1_i4b:m) = dxi(1_i4b:m)*sqrt( a )
!
                end if
!
                xnm1 = lamch( 's' )
!
                where ( dxi(1_i4b:m)>= xnm1 )
                   dxi(1_i4b:m) = one/dxi(1_i4b:m)
                elsewhere
                   dxi(1_i4b:m) = zero
                end where
!
!               COMPUTE THE CORRELATION MATRIX.
!
!$OMP PARALLEL DO IF(test_par)                    &
!$OMP            ,SCHEDULE(STATIC,1)              &
!$OMP            ,PRIVATE(j)
!
                do j = 1_i4b, m
!
                    if ( dxi(j)/=zero ) then
!
                        where ( dxi(1_i4b:j-1_i4b)/=zero )
                            xcor(1_i4b:j-1_i4b,j) = xcor(1_i4b:j-1_i4b,j)*( dxi(1_i4b:j-1_i4b)*dxi(j) )
                        elsewhere
                            xcor(1_i4b:j-1_i4b,j) = b
                        end where
!
                        xcor(j,j) = one
!
                    else
!
                        xcor(1_i4b:j,j) = b
!
                    end if
!
                end do
!
!$OMP END PARALLEL DO
!
!               SET MISSING INDICATOR IF REQUIRED.
!
                if ( present( failure ) ) failure = any( dxi(1_i4b:m)==zero )
!
           else
!
!               COMPUTE THE VARIANCE-COVARIANCE MATRIX.
!
!$OMP PARALLEL DO IF(test_par)                    &
!$OMP            ,SCHEDULE(STATIC,1)              &
!$OMP            ,PRIVATE(j)
!
                do j = 1_i4b, m
!
                    xcor(1_i4b:j,j) = xcor(1_i4b:j,j)*a
!
                end do
!
!$OMP END PARALLEL DO
!
                if ( present(xstd) ) then
!
!                   COMPUTE STANDARD-DEVIATIONS, IF REQUIRED.
!
                    dxi(1_i4b:m) = get_diag( xcor )
                    xstd(1_i4b:m) = sqrt( dxi(1_i4b:m) )
!
                end if
!
!               SET MISSING INDICATOR IF REQUIRED.
!
                if ( present( failure ) ) failure = false
!
            end if
!
        else
!
!           FEWER THAN TWO VALID OBSERVATIONS, SET THE PERTINENT STATISTICS TO NaN CODE.
!
            if ( xn<one ) xmean(1_i4b:m) = b
!
            if ( present(xstd) ) then
                xstd(1_i4b:m) = b
            end if
!
!$OMP PARALLEL DO IF(test_par)                    &
!$OMP            ,SCHEDULE(STATIC,1)              &
!$OMP            ,PRIVATE(j)
!
            do j = 1_i4b, m
!
                xcor(1_i4b:j,j) = b
!
            end do
!
!$OMP END PARALLEL DO
!
!           SET MISSING INDICATOR, IF REQUIRED.
!
            if ( present( failure ) ) failure = true
!
        end if
!
        if ( low_triang ) then
!
!           FILL LOWER TRIANGLE OF SYMETRIC MATRIX xcor, IF REQUIRED.
!
!$OMP PARALLEL DO IF(test_par)                    &
!$OMP            ,SCHEDULE(STATIC,1)              &
!$OMP            ,PRIVATE(j)
!
            do j = 2_i4b, m
!
                xcor(j,1_i4b:j-1_i4b) = xcor(1_i4b:j-1_i4b,j)
!
            end do
!
!$OMP END PARALLEL DO
!
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_cormat_a
! _______________________________
!
    end subroutine comp_cormat_a
!
! =========================================================================================
!
    subroutine comp_cormat_b( x, first, last, xmean, xcorp, xn, dimvar, xstd, cov, failure )
!
! Purpose
! _______
!
!   COMP_CORMAT computes estimates of means and variance-covariance or correlation
!   matrix from a data matrix.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:,:)
!            On entry, input submatrix containing size(X,3-DIMVAR)
!            observations on size(X,DIMVAR) variables from the matrix of data
!            for which means, variances and covariances are desired. By default,
!            DIMVAR is equal to 1. See description of optional DIMVAR argument 
!            for details.
!            If all the data are available at once, X can be the full 
!            data matrix.
!
!   FIRST    (INPUT) logical(lgl)
!            On entry, if:
!
!            - FIRST = true  the current submatrix is the first submatrix 
!              of the data matrix.
!            - FIRST = false the current submatrix is not the first submatrix
!              of the data matrix.
!
!   LAST     (INPUT) logical(lgl)
!            On entry, if:
!
!            - LAST = true  the current submatrix is the last submatrix 
!              of the data matrix.
!            - LAST = false the current submatrix is not the last submatrix
!              of the data matrix.
!
!   XMEAN    (INPUT/OUTPUT) real(stnd), dimension(:)
!            On entry, after the first call to COMP_CORMAT (e.g. when e.g. when FIRST=true),
!            XMEAN contains the variable means from  previous calls
!            to COMP_CORMAT. XMEAN should not be changed between calls to
!            COMP_CORMAT.
!            On exit, when LAST=true, XMEAN contains the variable means
!
!            The size of XMEAN must verify: size(XMEAN) = size(X,DIMVAR).
!
!   XCORP    (INPUT/OUTPUT) real(stnd), dimension(:)
!            On entry, after the first call to COMP_CORMAT (e.g. when e.g. when FIRST=true),
!            the linear array XCORP contains the upper triangle of the corrected
!            sums of squared and cross-products matrix, packed columnwise, computed from  
!            previous calls to COMP_CORMAT. XCORP should not be changed
!            between calls to COMP_CORMAT.
!
!            On exit, when LAST=true, XCORP contains the correlation or
!            variance-covariance matrix as controlled by the COV argument.
!            XCORP is stored in symmetric storage mode (see further details).
!
!            The size of XCORP must verify: size(XCORP) = (size(X,DIMVAR) * (size(X,DIMVAR)+1))/2.
!
!   XN       (INPUT/OUTPUT) real(stnd)
!            On entry, after the first call to COMP_CORMAT (e.g. when e.g. when FIRST=true),
!            XN contains count of observations from previous calls
!            to COMP_CORMAT. XN should not be changed between calls to
!            COMP_CORMAT.
!
!            On exit, XN contains the number of observations in the data matrix XX.
!
!   DIMVAR   (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVAR is present, DIMVAR is used as follows, if:
!
!            - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!              on size(X,1) variables.
!            - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!              on size(X,2) variables.
!
!            The default is DIMVAR = 1.
!
!   XSTD     (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            On exit, when LAST=true and XSTD is present, XSTD contains the variable 
!            standard-deviations.
!
!            The size of XSTD must verify: size(XSTD) = size(X,DIMVAR).
!
!            XSTD needs to be specified only on the last call to COMP_CORMAT (e.g. when LAST=true).
!          
!   COV      (INPUT, OPTIONAL) logical(lgl)
!            On entry, when argument COV is present, COV is used as follows, if:
!
!            - COV= true,   XCORP contains the variances-covariances matrix, when LAST=true.
!            - COV= false,  XCORP contains the correlation matrix, when LAST=true.
!
!            By default, the correlation matrix is output.
!
!            COV needs to be specified only on the last call to COMP_CORMAT (e.g. when LAST=true).
!          
!   FAILURE  (OUTPUT, OPTIONAL) logical(lgl)
!            On exit, when argument FAILURE is present:
!
!            - FAILURE = false:   indicates successful exit.
!            - FAILURE = true:    indicates that fewer than two valid observations
!              were present or that the observations on some
!              variable were constant and the correlations
!              were requested.
!
!
! Further Details
! _______________
!
!   The subroutine computes the means and the correlation matrix with only one pass
!   through the data.
!
!   On exit, the upper triangle of the symmetric correlation or variance-covariance matrix
!   XCOR is packed columnwise in the linear array XCORP.  More precisely, the j-th column
!   of XCOR is stored in the array XCORP as follows:
!
!            XCORP(i + (j-1) * j/2) = XCOR(i,j) for 1<=i<=j;
!
!   If the observations on some variable were constant, the pertinent correlations 
!   are set to nan() code .
!
!   If fewer than two valid observations were present, the correlations are set to 
!   nan() code .
!
!   If fewer than one valid observation is present, the means are also set to 
!   nan() code .
!
!   For more details on correlation analysis, see:
!
!   (1) von Storch, H., and Zwiers, F.W., 2002:
!            Statistical Analysis in Climate Research.
!            Cambridge, UK, Chapter 8, 484 pp., ISBN:9780521012300
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : zero, one
    use Num_Constants,     only : nan, lamch
    use Char_Constants,    only : stat_error1, stat_error5
    use Utilities,         only : assert_eq, assert, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),              dimension(:,:) :: x
    real(stnd), intent(inout),           dimension(:)   :: xmean, xcorp
    real(stnd), intent(inout)                           :: xn
    real(stnd), intent(out),  optional , dimension(:)   :: xstd
!
    integer(i4b), intent(in), optional :: dimvar
!
    logical(lgl), intent(in)            :: first, last
    logical(lgl), intent(in),  optional :: cov
    logical(lgl), intent(out), optional :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                         :: a, b, xnm1, tmp
    real(stnd), dimension(size(xmean)) :: dxi
!
    integer(i4b) :: i, j, l, l1, l2,  n, m, p
!
    logical(lgl) :: cor, dim_is_one
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_cormat'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
    n = size( x, 3-int(i) )
!
    m = assert_eq(  int(size(x,int(i)),i4b),   &
                    int(size(xmean),i4b),      &
                    name_proc )
!
    if ( m<=0_i4b ) call merror( name_proc//stat_error1 )
!
    p = ( m*(m+1_i4b) )/2_i4b
!
    call assert( logical(size(xcorp)==p,lgl),    &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBMATRIX IS THE FIRST.
!
!       INITIALIZE xmean, xcorp AND xn .
!
        xmean(1_i4b:m) = zero       
        xcorp(1_i4b:p) = zero
        xn             = zero
!
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  p>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)           &
!$OMP         ,PRIVATE(i,j,l,l1,l2,tmp)
!
    do i = 1_i4b, n
!
!$OMP SINGLE
!
        xnm1 = xn
        xn = xnm1 + one
        a = one/xn
        b = xnm1*a
!
!       CALCULATE PARTIAL MEANS AND CORRECTED SUMS OF SQUARES AND CROSS-PRODUCTS, 
!       HANDLING THE DATA ONCE.
!
        if ( dim_is_one ) then
            dxi(1_i4b:m) = x(1_i4b:m,i) - xmean(1_i4b:m)
        else
            dxi(1_i4b:m) = x(i,1_i4b:m) - xmean(1_i4b:m)
        end if
!
        xmean(1_i4b:m) = xmean(1_i4b:m) + a*dxi(1_i4b:m)
!
!$OMP END SINGLE
!
!$OMP DO SCHEDULE(STATIC,1) 
!
        do j = 1_i4b, m
!
            l = (j-1_i4b)*j/2_i4b
            l1 = l + 1_i4b
            l2 = l + j
!
            tmp = b*dxi(j)
!
#ifdef _BLAS
            call axpy( j, tmp, dxi(1_i4b:j), 1_i4b, xcorp(l1:l2), 1_i4b )
#else
            xcorp(l1:l2) = xcorp(l1:l2) + tmp*dxi(1_i4b:j)
#endif
!
        end do
!
!$OMP END DO
!
    end do
!
!$OMP END PARALLEL
!
    if ( last ) then
!
!       THE CURRENT SUBMATRIX IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
!
!       TEST OPTIONAL ARGUMENTS.
!
        if ( present(cov) ) then
           cor = .not.cov
        else
           cor = true
        end if
!
        if ( present(xstd) ) then
            call assert( logical(size(xstd)==m,lgl),    &
                         name_proc )
        end if
!
        b = nan()
!
        if ( xn>one ) then
!
!           MORE THAN TWO VALID OBSERVATIONS, COMPUTE THE STATISTICS.
!
            a = one/xn
!
            if ( cor ) then
!
!               COMPUTE THE CORRELATION MATRIX.
!
!
!               GET DIAGONAL ELEMENTS OF CORRECTED SUMS OF SQUARES 
!               AND CROSS-PRODUCTS.
!
!$OMP PARALLEL DO IF(test_par)               &
!$OMP            ,SCHEDULE(STATIC,1)         &
!$OMP            ,PRIVATE(j, l)
!
                do j = 1_i4b, m
!
                    l = (j-1_i4b)*j/2_i4b + j
                    dxi(j) = sqrt( xcorp(l) )
!
                end do
!
!$OMP END PARALLEL DO
!
                if ( present(xstd) ) then
!
!                   COMPUTE STANDARD-DEVIATIONS, IF REQUIRED.
!
                    xstd(:m) = dxi(:m)*sqrt( a )
!
                end if
!
                xnm1 = lamch( 's' )
!
                where ( dxi(1_i4b:m)>= xnm1 )
                   dxi(1_i4b:m) = one/dxi(1_i4b:m)
                elsewhere
                   dxi(1_i4b:m) = zero
                end where
!
!               COMPUTE THE CORRELATION MATRIX.
!
!$OMP PARALLEL DO IF(test_par)               &
!$OMP            ,SCHEDULE(STATIC,1)         &
!$OMP            ,PRIVATE(j, l, l1, l2)
!
                do j = 1_i4b, m
!
                    l = (j-1_i4b)*j/2_i4b
                    l1 = l + 1_i4b
                    l2 = l + j
!
                    if ( dxi(j)/=zero ) then
!
                        where ( dxi(1_i4b:j-1_i4b)/=zero )
                            xcorp(l1:l2-1_i4b) = xcorp(l1:l2-1_i4b)*( dxi(1_i4b:j-1_i4b)*dxi(j) )
                        elsewhere
                            xcorp(l1:l2-1_i4b) = b
                        end where
                        xcorp(l2) = one
!
                    else
!
                        xcorp(l1:l2) = b
!
                    end if
!
                end do
!
!$OMP END PARALLEL DO
!
!               SET MISSING INDICATOR, IF REQUIRED.
!
                if ( present( failure ) ) failure = any( dxi(1_i4b:m)==zero )
!
            else
!
!               COMPUTE THE VARIANCE-COVARIANCE MATRIX.
!
                xcorp(1_i4b:p) = xcorp(1_i4b:p)*a
!
                if ( present(xstd) ) then
!
!                   COMPUTE STANDARD-DEVIATIONS, IF REQUIRED.
!
!$OMP PARALLEL DO IF(test_par)               &
!$OMP            ,SCHEDULE(STATIC,1)         &
!$OMP            ,PRIVATE(j, l)
!
                    do j = 1_i4b, m
!
                        l = (j-1_i4b)*j/2_i4b + j
                        xstd(j) = sqrt( xcorp(l) )
!
                    end do
!
!$OMP END PARALLEL DO
!
                end if
!
!               SET MISSING INDICATOR, IF REQUIRED.
!
                if ( present( failure ) ) failure = false
!
            end if
!
        else
!
!           FEWER THAN TWO VALID OBSERVATIONS, SET THE PERTINENT STATISTICS TO NaN CODE.
!
            if ( xn<one ) xmean(1_i4b:m) = b
!
            if ( present(xstd) ) then
                xstd(1_i4b:m) = b
            end if
!
            xcorp(1_i4b:p) = b
!
!           SET MISSING INDICATOR IF REQUIRED.
!
            if ( present( failure ) ) failure = true
!
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_cormat_b
! _______________________________
!
    end subroutine comp_cormat_b
!
! =========================================================================================
!
    subroutine comp_cormat_miss_a( x, first, last, xmean, xcor, xn, xmiss, dimvar, xstd, &
                                   cov, fill, failure )
!
! Purpose
! _______ 
!
!   COMP_CORMAT_MISS computes estimates of means and variance-covariance or correlation
!   matrix from a data matrix possibly containing missing values.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:,:)
!            On entry, input submatrix containing size(X,3-DIMVAR)
!            observations on size(X,DIMVAR) variables from the matrix of data
!            for which means, variances and covariances are desired. By default,
!            DIMVAR is equal to 1. See description of optional DIMVAR argument 
!            for details.
!            If all the data are available at once, X can be the full 
!            data matrix.
!
!   FIRST    (INPUT) logical(lgl)
!            On entry, if:
!
!             - FIRST = true  the current submatrix is the first submatrix 
!               of the data matrix.
!             - FIRST = false the current submatrix is not the first submatrix
!               of the data matrix.
!
!   LAST     (INPUT) logical(lgl)
!            On entry, if:
!
!             - LAST = true  the current submatrix is the last submatrix 
!               of the data matrix.
!             - LAST = false the current submatrix is not the last submatrix
!               of the data matrix.
!
!   XMEAN    (INPUT/OUTPUT) real(stnd), dimension(:,2)
!            On entry, after the first call to COMP_CORMAT_MISS (e.g. when FIRST=true),
!            XMEAN(:,1) contains the variable means from  previous calls
!            to COMP_CORMAT_MISS. XMEAN should not be changed between calls to
!            COMP_CORMAT_MISS.
!
!            On exit, when LAST=true, XMEAN(:,1) contains the variable means
!            computed from all non-missing observations in the data matrix.
!            XMEAN(:,2) is used as workspace.
!
!            The shape of XMEAN must verify:
!
!            - size(XMEAN,1) = size(X,DIMVAR) ,
!            - size(MEAN,2) = 2 .
!
!   XCOR     (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, after the first call to COMP_CORMAT_MISS (e.g. when FIRST=true),
!            the matrix XCOR contains the upper triangle of the corrected
!            sums of squared and cross-products matrix computed from  
!            previous calls to COMP_CORMAT_MISS. XCOR should not be changed
!            between calls to COMP_CORMAT_MISS.
!
!            On exit, when LAST=true, XCOR contains the upper triangle of the
!            symetric correlation or variance-covariance matrix as controlled by 
!            the COV argument. If the optional argument FILL is present and equal
!            to true, the lower triangle of XCOR is also filled.
!
!            The shape of XCOR must verify:
!
!            - size(XCOR,1) = size(X,DIMVAR) ,
!            - size(XCOR,2) = size(X,DIMVAR) .
!
!   XN       (INPUT/OUTPUT) real(stnd), dimension(:,3)
!            On entry, after the first call to COMP_CORMAT_MISS (e.g. when FIRST=true),
!            XN is used as workspace to accumulate quantities from previous calls
!            to COMP_CORMAT_MISS. XN should not be changed between calls to
!            COMP_CORMAT_MISS.
!
!            On exit, XN(:,1) contains the upper triangle of the matrix of the incidence values
!            between each pair of variables, packed columnwise, in a linear array. 
!            XN(i + (j-1) * j/2,1) indicates the numbers of non-missing pairs which were 
!            used in the calculation of XCOR(i,j) for 1<=i<=j .
!            XN(:,2:3) is used as workspace.
!
!            The shape of XN must verify:
!
!            - size(XN,1) = (size(X,DIMVAR) * (size(X,DIMVAR)+1))/2 ,
!            - size(XN,2) = 3 .
!
!   XMISS    (INPUT) real(stnd)
!            On entry, the missing value indicator. Any value in
!            X which is equal to XMISS is assumed to be missing
!            or invalid.
!            The means and the correlations are computed
!            on all the observations where X are not missing (see Further Details).
!
!   DIMVAR   (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVAR is present, DIMVAR is used as follows, if:
!
!            - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!              on size(X,1) variables.
!            - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!              on size(X,2) variables.
!
!            The default is DIMVAR = 1.
!
!   XSTD     (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            On exit, when LAST=true and XSTD is present, XSTD contains the variable 
!            standard-deviations.
!
!            The size of XSTD must verify: size(XSTD) = size(X,DIMVAR).
!
!            XSTD needs to be specified only on the last call to COMP_CORMAT_MISS (e.g. when LAST=true).
!          
!   COV      (INPUT, OPTIONAL) logical(lgl)
!            On entry, when argument COV is present, COV is used as follows, if:
!
!            - COV= true,   XCORP contains the variances-covariances matrix, when LAST=true.
!            - COV= false,  XCORP contains the correlation matrix, when LAST=true.
!
!            By default, the correlation matrix is output.
!
!            COV needs to be specified only on the last call to COMP_CORMAT_MISS (e.g. when LAST=true).
!          
!   FILL     (INPUT, OPTIONAL) logical(lgl)
!            On entry, when argument FILL is present, FILL is used as follows, if:
!          
!            - FILL= true,  the lower triangle of XCOR is filled, when LAST=true.
!            - FILL= false, the lower triangle of XCOR is not filled, when LAST=true.
!
!            By default, the lower triangle of XCOR is not filled.
!
!            FILL  needs to be specified only on the last call to COMP_CORMAT_MISS (e.g. when LAST=true).
!
!   FAILURE  (OUTPUT, OPTIONAL) logical(lgl)
!            On exit, when argument FAILURE is present:
!
!            - FAILURE = false:   indicates successful exit.
!            - FAILURE = true:    indicates that fewer than two valid observations
!              were present for some pair of variables or that the
!              observations on some variable were constant and the
!              correlations were requested.
!
!
! Further Details
! _______________
!
!   The subroutine computes the means and the correlation matrix with only one pass
!   through the data.
!
!   If the observations on some variable were constant, the pertinent correlations 
!   are set to XMISS.
!
!   If fewer than two valid observations were present for some pair of variables,
!   the pertinent correlations are set to XMISS.
!
!   If fewer than one valid observation is present for some variables, the pertinent
!   means are also set to XMISS.
!
!   The means and standard-deviations of the data matrix are computed from all valid data.
!   The correlation coefficients are based on these univariate statistics and on all valid
!   pairs of observations.
!
!   For more details on correlation analysis, see:
!
!   (1) von Storch, H., and Zwiers, F.W., 2002:
!            Statistical Analysis in Climate Research.
!            Cambridge, UK, Chapter 8, 484 pp., ISBN:9780521012300
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : zero, one
    use Num_Constants,     only : lamch
    use Char_Constants,    only : stat_error1, stat_error5
    use Utilities,         only : assert_eq, assert, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                             :: xmiss
    real(stnd), intent(in),             dimension(:,:) :: x
    real(stnd), intent(inout),          dimension(:,:) :: xmean, xn, xcor
    real(stnd), intent(out),  optional , dimension(:)  :: xstd
!
    integer(i4b), intent(in),  optional :: dimvar
!
    logical(lgl), intent(in)            :: first, last
    logical(lgl), intent(in),  optional :: cov, fill
    logical(lgl), intent(out), optional :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                           :: tmp
    real(stnd), dimension(size(xmean,1)) :: dxi, a, b
!
    integer(i4b) :: i, j, l, l1, l2,  n, m, p
!
    logical(lgl)                           :: cor, low_triang, dim_is_one
    logical(lgl), dimension(size(xmean,1)) :: mask
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_cormat_miss'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
!
            i = dimvar
!
        else
!
           call merror( name_proc//stat_error5 )
!
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
    n = size( x, 3-int(i) )
!
    m = assert_eq(  int(size(x,int(i)),i4b),    &
                    int(size(xmean,1),i4b),     &
                    int(size(xcor,1),i4b),      &
                    int(size(xcor,2),i4b),      &
                    name_proc )
!
    if ( m<=0_i4b ) call merror( name_proc//stat_error1 )
!
    p = ( m*(m+1_i4b) )/2_i4b
!
    call assert( logical(int(size(xn,1),i4b)==p,lgl),      &
                 logical(size(xn,2)==3,lgl),               &
                 logical(size(xmean,2)==2,lgl),            &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBMATRIX IS THE FIRST.
!
!       INITIALIZE xmean, xcor AND xn .
!
        xmean(1_i4b:m,1_i4b)    = zero       
        xcor(1_i4b:m,1_i4b:m)   = zero
        xn(1_i4b:p,1_i4b:3_i4b) = zero
!
        xmean(1_i4b:m,2_i4b) = xmiss
!
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  p>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)                &
!$OMP         ,PRIVATE(i,j,l,l1,l2,b,tmp)
!
    do i = 1_i4b, n
!
!$OMP SINGLE
!
!       CALCULATE PARTIAL MEANS AND CORRECTED SUMS OF SQUARES AND CROSS-PRODUCTS, 
!       HANDLING THE DATA ONCE.
!
        if ( dim_is_one ) then
            dxi(1_i4b:m) = x(1_i4b:m,i)
        else
            dxi(1_i4b:m) = x(i,1_i4b:m)
        end if
!
        mask(1_i4b:m) = dxi(1_i4b:m)/=xmiss
!
#ifdef _WHERE
        do j = 1_i4b, m
!
            if ( xmean(j,2_i4b)==xmiss ) xmean(j,2_i4b) = dxi(j)
!
        end do
#else
        where ( xmean(1_i4b:m,2_i4b)==xmiss ) xmean(1_i4b:m,2_i4b) = dxi(1_i4b:m)
#endif
!
!$OMP END SINGLE
!
!
!$OMP DO SCHEDULE(STATIC,1) 
!
        do j = 1_i4b, m
!
            if ( mask(j) ) then
!
                l = (j-1_i4b)*j/2_i4b
                l1 = l + 1_i4b
                l2 = l + j
!
                where ( mask(1_i4b:j) )
!
                    b(1_i4b:j) = dxi(1_i4b:j) - xmean(1_i4b:j,2_i4b)
                    xcor(1_i4b:j,j) = xcor(1_i4b:j,j) + b(1_i4b:j)*b(j)
!
                    xn(l1:l2,1_i4b) = xn(l1:l2,1_i4b) + one
                    xn(l1:l2,2_i4b) = xn(l1:l2,2_i4b) + b(1_i4b:j)
                    xn(l1:l2,3_i4b) = xn(l1:l2,3_i4b) + b(j)
!
                end where
!
                tmp = dxi(j) - xmean(j,1_i4b)
                xmean(j,1_i4b) = xmean(j,1_i4b) + tmp/xn(l2,1_i4b)
!
            end if
!
        end do
!
!$OMP END DO
!
    end do
!
!$OMP END PARALLEL
!
    if ( last ) then
!
!       THE CURRENT SUBMATRIX IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
!       TEST OPTIONAL ARGUMENTS.
!
        if ( present(cov) ) then
           cor = .not.cov
        else
           cor = true
        end if
!
        if ( present(fill) ) then
           low_triang = fill
        else
           low_triang = false
        end if
!
        if ( present(xstd) ) then
            call assert( logical(size(xstd)==m,lgl),    &
                         name_proc )
        end if
!
        where ( xmean(1_i4b:m,2_i4b)/=xmiss )
!
!           COMPUTE THE VARIANCE-COVARIANCE MATRIX.
!
            dxi(1_i4b:m) = xmean(1_i4b:m,2_i4b) - xmean(1_i4b:m,1_i4b)
!
        elsewhere
!
!            FEWER THAN ONE VALID OBSERVATION, SET THE MEANS TO MISSING CODE.
!
             xmean(1_i4b:m,1_i4b) = xmiss
!
        end where
!
!       COMPUTE THE COVARIANCE MATRIX.
!
!$OMP PARALLEL DO IF(test_par)              &
!$OMP            ,SCHEDULE(STATIC,1)        &
!$OMP            ,PRIVATE(j, l, l1, l2)
!
        do j = 1_i4b, m
!
            l = (j-1_i4b)*j/2_i4b
            l1 = l + 1_i4b
            l2 = l + j
!
            where ( xn(l1:l2,1_i4b)>one )
!
                xcor(1_i4b:j,j) = xcor(1_i4b:j,j) + dxi(1_i4b:j)*xn(l1:l2,3_i4b) +                    &
                                  dxi(j)*xn(l1:l2,2_i4b) + (dxi(j)*dxi(1_i4b:j))*xn(l1:l2,1_i4b)
                xcor(1_i4b:j,j) = xcor(1_i4b:j,j)/xn(l1:l2,1_i4b)
!
            elsewhere
!
                xcor(1_i4b:j,j) = xmiss
!
            end where
!
            a(j) = xcor(j,j)
!
        end do
!
!$OMP END PARALLEL DO
!
        if ( cor .or. present(xstd) ) then
!
!           COMPUTE STANDARD-DEVIATIONS, IF REQUIRED.
!
            mask(1_i4b:m) = a(1_i4b:m)/=xmiss
!
            where ( mask(1_i4b:m) )
                dxi(1_i4b:m) = sqrt( a(1_i4b:m) )
            elsewhere
                dxi(1_i4b:m) = xmiss
            end where
!
            if ( present(xstd) ) xstd(1_i4b:m) = dxi(1_i4b:m) 
!
            if ( cor ) then
!
!               COMPUTE THE CORRELATION MATRIX, IF REQUIRED.
!
                b(1_i4b) = lamch( 's' )
!
                where ( dxi(1_i4b:m)>=b(1_i4b) .and. mask(1_i4b:m) )
!
                    a(1_i4b:m) = one/dxi(1_i4b:m)
                    mask(1_i4b:m) = false
!
                end where
!
!$OMP PARALLEL DO IF(test_par)                    &
!$OMP            ,SCHEDULE(STATIC,1)              &
!$OMP            ,PRIVATE(j, b, dxi )
!
                do j = 1_i4b, m
!
                    if ( xcor(j,j)/=xmiss ) then
!
                        if ( .not.mask(j) ) then
!
                            where( mask(1_i4b:j-1_i4b) ) xcor(1_i4b:j-1_i4b,j) = xmiss
!
                            where ( xcor(1_i4b:j-1_i4b,j)/=xmiss  )
!
                                b(1_i4b:j-1_i4b) = xcor(1_i4b:j-1_i4b,j)*( a(1_i4b:j-1_i4b)*a(j) )
!
                                where ( b(1_i4b:j-1_i4b)<one )
                                    dxi(1_i4b:j-1_i4b) = b(1_i4b:j-1_i4b)
                                elsewhere
                                    dxi(1_i4b:j-1_i4b) = one
                                end where
!                                dxi(1_i4b:j-1_i4b) = merge( b(1_i4b:j-1_i4b), one, b(1_i4b:j-1_i4b)<one )
!
                                where ( dxi(1_i4b:j-1_i4b)>-one )
                                    xcor(1_i4b:j-1_i4b,j) = dxi(1_i4b:j-1_i4b)
                                elsewhere
                                    xcor(1_i4b:j-1_i4b,j) = -one
                                end where
!                                xcor(1_i4b:j-1_i4b,j) = merge( dxi(1_i4b:j-1_i4b), -one, dxi(1_i4b:j-1_i4b)>-one )
!
                            end where
!
                            xcor(j,j) = one
!
                        else
!
                            xcor(1_i4b:j,j) = xmiss
!
                        end if
!
                    end if
!
                end do
!
!$OMP END PARALLEL DO
!
            end if
!
        end if
!
        if ( present( failure ) ) then
!
!           SET MISSING INDICATOR, IF REQUIRED.
!
            failure = any(  xn(1_i4b:p,1_i4b)<=one )
!
            if ( cor ) failure = any( mask(1_i4b:m) ) .or. failure
!
        end if
!
        if ( low_triang ) then
!
!           FILL LOWER TRIANGLE OF SYMETRIC MATRIX xcor, IF REQUIRED.
!
!$OMP PARALLEL DO IF(test_par)          &
!$OMP            ,SCHEDULE(STATIC,1)    &
!$OMP            ,PRIVATE(j)
!
            do j = 2_i4b, m
!
              xcor(j,1_i4b:j-1_i4b) = xcor(1_i4b:j-1_i4b,j)
!
            end do
!
!$OMP END PARALLEL DO
!
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_cormat_miss_a
! ____________________________________
!
    end subroutine comp_cormat_miss_a
!
! =========================================================================================
!
    subroutine comp_cormat_miss_b( x, first, last, xmean, xcorp, xn, xmiss, dimvar, &
                                   xstd, cov, failure )
!
! Purpose
! _______ 
!
!   COMP_CORMAT_MISS computes estimates of means and variance-covariance or correlation
!   matrix from a data matrix possibly containing missing values.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:,:)
!            On entry, input submatrix containing size(X,3-DIMVAR)
!            observations on size(X,DIMVAR) variables from the matrix of data
!            for which means, variances and covariances are desired. By default,
!            DIMVAR is equal to 1. See description of optional DIMVAR argument 
!            for details.
!            If all the data are available at once, X can be the full 
!            data matrix.
!
!   FIRST    (INPUT) logical(lgl)
!            On entry, if:
!
!            - FIRST = true  the current submatrix is the first submatrix 
!              of the data matrix.
!            - FIRST = false the current submatrix is not the first submatrix
!              of the data matrix.
!
!   LAST     (INPUT) logical(lgl)
!            On entry, if:
!
!            - LAST = true  the current submatrix is the last submatrix 
!              of the data matrix.
!            - LAST = false the current submatrix is not the last submatrix
!              of the data matrix.
!
!   XMEAN    (INPUT/OUTPUT) real(stnd), dimension(:,2)
!            On entry, after the first call to COMP_CORMAT_MISS (e.g. when FIRST=true),
!            XMEAN(:,1) contains the variable means from  previous calls
!            to COMP_CORMAT_MISS. XMEAN should not be changed between calls to
!            COMP_CORMAT_MISS.
!
!            On exit, when LAST=true, XMEAN(:,1) contains the variable means
!            computed from all non-missing observations in the data matrix.
!            XMEAN(:,2) is used as workspace.
!
!            The shape of XMEAN must verify:
!
!            - size(XMEAN,1) = size(X,DIMVAR) ,
!            - size(MEAN,2) = 2 .
!
!   XCORP    (INPUT/OUTPUT) real(stnd), dimension(:)
!            On entry, after the first call to COMP_CORMAT_MISS (e.g. when FIRST=true),
!            the linear array XCORP contains the upper triangle of the corrected
!            sums of squared and cross-products matrix, packed columnwise, computed from  
!            previous calls to COMP_CORMAT_MISS. XCORP should not be changed
!            between calls to COMP_CORMAT_MISS.
!
!            On exit, when LAST=true, XCORP contains the correlation or
!            variance-covariance matrix as controlled by the COV argument.
!            XCORP is stored in symmetric storage mode (see further details).
!
!            The size of XCORP must verify: size(XCORP) = (size(X,DIMVAR) * (size(X,DIMVAR)+1))/2.
!
!   XN       (INPUT/OUTPUT) real(stnd), dimension(:,3)
!            On entry, after the first call to COMP_CORMAT_MISS (e.g. when FIRST=true),
!            XN is used as workspace to accumulate quantities from previous calls
!            to COMP_CORMAT_MISS. XN should not be changed between calls to
!            COMP_CORMAT_MISS.
!
!            On exit, XN(:,1) contains the incidence values between each pair of variables.
!            XN(i,1) indicates the numbers of non-missing pairs of observations which were
!            used in the calculation of XCORP(i).
!
!            The shape of XN must verify:
!
!            - size(XN,1) = (size(X,DIMVAR) * (size(X,DIMVAR)+1))/2 ,
!            - size(XN,2) = 3 .
!
!   XMISS    (INPUT) real(stnd)
!            On entry, the missing value indicator. Any value in
!            X which is equal to XMISS is assumed to be missing
!            or invalid.
!            The means and the correlations are computed
!            on all the observations where X are not missing (see Further Details).
!
!   DIMVAR   (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVAR is present, DIMVAR is used as follows, if:
!
!            - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!              on size(X,1) variables.
!            - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!              on size(X,2) variables.
!
!            The default is DIMVAR = 1.
!
!   XSTD     (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            On exit, when LAST=true and XSTD is present, XSTD contains the variable 
!            standard-deviations.
!
!            The size of XSTD must verify: size(XSTD) = size(X,DIMVAR).
!
!            XSTD needs to be specified only on the last call to COMP_CORMAT_MISS (e.g. when LAST=true).
!          
!   COV      (INPUT, OPTIONAL) logical(lgl)
!            On entry, when argument COV is present, COV is used as follows, if:
!          
!            - COV= true,   XCORP contains the variances-covariances matrix, when LAST=true.
!            - COV= false,  XCORP contains the correlation matrix, when LAST=true.
!
!            By default, the correlation matrix is output.
!
!            COV needs to be specified only on the last call to COMP_CORMAT_MISS (e.g. when LAST=true).
!          
!   FAILURE  (OUTPUT, OPTIONAL) logical(lgl)
!            On exit, when argument FAILURE is present:
!          
!            - FAILURE = false:   indicates successful exit.
!            - FAILURE = true:  indicates that fewer than two valid observations
!              were present for some variable or that the observations
!              on some variable were constant and the correlations
!              were requested.
!
!
! Further Details
! _______________
!
!   The subroutine computes the means and the correlation matrix with only one pass
!   through the data.
!
!   On exit, the upper triangle of the symmetric correlation or variance-covariance matrix
!   XCOR is packed columnwise in the linear array XCORP.  More precisely, the j-th column
!   of XCOR is stored in the array XCORP as follows:
!
!            XCORP(i + (j-1) * j/2) = XCOR(i,j) for 1<=i<=j;
!
!
!   If the observations on some variable were constant, the pertinent correlations 
!   are set to XMISS .
!
!   If fewer than two valid observations were present on some variable, the pertinent
!   correlations are set to XMISS.
!
!   If fewer than one valid observation is present for some variable, the pertinent
!   means are also set to XMISS.
!
!   The means and standard-deviations of the data matrix are computed from all valid data. The
!   correlation coefficients are based on these univariate statistics and on all valid pairs
!   of observations.
!
!   For more details on correlation analysis, see:
!
!   (1) von Storch, H., and Zwiers, F.W., 2002:
!            Statistical Analysis in Climate Research.
!            Cambridge, UK, Chapter 8, 484 pp., ISBN:9780521012300
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : zero, one
    use Num_Constants,     only : lamch
    use Char_Constants,    only : stat_error1, stat_error5
    use Utilities,         only : assert_eq, assert, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                              :: xmiss
    real(stnd), intent(in),              dimension(:,:) :: x
    real(stnd), intent(inout),           dimension(:)   :: xcorp
    real(stnd), intent(inout),           dimension(:,:) :: xmean, xn
    real(stnd), intent(out),  optional , dimension(:)   :: xstd
!
    integer(i4b), intent(in), optional :: dimvar
!
    logical(lgl), intent(in)            :: first, last
    logical(lgl), intent(in),  optional :: cov
    logical(lgl), intent(out), optional :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                           :: tmp
    real(stnd), dimension(size(xmean,1)) :: dxi, a, b
!
    integer(i4b) :: i, j, l, l1, l2,  n, m, p
!
    logical(lgl)                           :: cor, dim_is_one
    logical(lgl), dimension(size(xmean,1)) :: mask
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_cormat_miss'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
    n = size( x, 3-int(i) )
!
    m = assert_eq(  int(size(x,int(i)),i4b),    &
                    int(size(xmean,1),i4b),     &
                    name_proc )
!
    if ( m<=0_i4b ) call merror( name_proc//stat_error1 )
!
    p = ( m*(m+1_i4b) )/2_i4b
!
    call assert( logical(int(size(xcorp),i4b)==p,lgl),    &
                 logical(int(size(xn,1),i4b)==p,lgl),     &
                 logical(size(xn,2)==3,lgl),              &
                 logical(size(xmean,2)==2,lgl),           &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBMATRIX IS THE FIRST.
!
!       INITIALIZE xmean, xcorp AND xn .
!
        xmean(1_i4b:m,1_i4b) = zero       
        xcorp(1_i4b:p) = zero
        xn(1_i4b:p,1_i4b:3_i4b) = zero
!
        xmean(1_i4b:m,2_i4b) = xmiss
!
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  p>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)                &
!$OMP         ,PRIVATE(i,j,l,l1,l2,b,tmp)
!
    do i = 1_i4b, n
!
!$OMP SINGLE
!
!       CALCULATE PARTIAL MEANS AND CORRECTED SUMS OF SQUARES AND CROSS-PRODUCTS, 
!       HANDLING THE DATA ONCE.
!
        if ( dim_is_one ) then
            dxi(1_i4b:m) = x(1_i4b:m,i)
        else
            dxi(1_i4b:m) = x(i,1_i4b:m)
        end if
!
        mask(1_i4b:m) = dxi(1_i4b:m)/=xmiss
!
#ifdef _WHERE
        do j = 1_i4b, m
            if ( xmean(j,2_i4b)==xmiss ) xmean(j,2_i4b) = dxi(j)
        end do
#else
        where ( xmean(1_i4b:m,2_i4b)==xmiss ) xmean(1_i4b:m,2_i4b) = dxi(1_i4b:m)
#endif
!
!$OMP END SINGLE
!
!
!$OMP DO SCHEDULE(STATIC,1) 
!
        do j = 1_i4b, m
!
            if ( mask(j) ) then
!
                l = (j-1_i4b)*j/2_i4b
                l1 = l + 1_i4b
                l2 = l + j
!
                where ( mask(1_i4b:j) )
!
                    b(1_i4b:j) = dxi(1_i4b:j) - xmean(1_i4b:j,2_i4b)
                    xcorp(l1:l2) = xcorp(l1:l2) + b(:j)*b(j)
                    xn(l1:l2,1_i4b) = xn(l1:l2,1_i4b) + one
                    xn(l1:l2,2_i4b) = xn(l1:l2,2_i4b) + b(1_i4b:j)
                    xn(l1:l2,3_i4b) = xn(l1:l2,3_i4b) + b(j)
!
                end where
!
                tmp = dxi(j) - xmean(j,1_i4b)
                xmean(j,1_i4b) = xmean(j,1_i4b) + tmp/xn(l2,1_i4b)
!
            end if
!
        end do
!
!$OMP END DO
!
    end do
!
!$OMP END PARALLEL
!
    if ( last ) then
!
!       THE CURRENT SUBMATRIX IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
!       TEST OPTIONAL ARGUMENTS.
!
        if ( present(cov) ) then
           cor = .not.cov
        else
           cor = true
        end if
!
        if ( present(xstd) ) then
            call assert( logical(size(xstd)==m,lgl),    &
                         name_proc )
        end if
!
        where ( xmean(1_i4b:m,2_i4b)/=xmiss )
!
            dxi(1_i4b:m) = xmean(1_i4b:m,2_i4b) - xmean(1_i4b:m,1_i4b)
!
        elsewhere
!
!           FEWER THAN ONE VALID OBSERVATION, SET THE MEANS TO MISSING CODE.
!
            xmean(1_i4b:m,1_i4b) = xmiss
!
        end where
!
!       COMPUTE THE VARIANCE-COVARIANCE MATRIX.
!
!$OMP PARALLEL DO IF(test_par)              &
!$OMP            ,SCHEDULE(STATIC,1)        &
!$OMP            ,PRIVATE(j, l, l1, l2)
!
        do j = 1_i4b, m
!
            l = (j-1_i4b)*j/2_i4b
            l1 = l + 1_i4b
            l2 = l + j
!
            where ( xn(l1:l2,1_i4b)>one )
!
                xcorp(l1:l2) = xcorp(l1:l2) + dxi(1_i4b:j)*xn(l1:l2,3_i4b) +                    &
                               dxi(j)*xn(l1:l2,2_i4b) + (dxi(j)*dxi(1_i4b:j))*xn(l1:l2,1_i4b)
                xcorp(l1:l2) = xcorp(l1:l2)/xn(l1:l2,1_i4b)
!
            elsewhere
!
                xcorp(l1:l2) = xmiss
!
            end where
!
            a(j) = xcorp(l2)
!
        end do
!
!$OMP END PARALLEL DO
!
        if ( cor .or. present(xstd) ) then
!
!           OUTPUT STANDARD-DEVIATIONS, IF REQUIRED.
!
            mask(1_i4b:m) = a(1_i4b:m)/=xmiss
!
            where ( mask(1_i4b:m) )
                dxi(1_i4b:m) = sqrt( a(1_i4b:m) )
            elsewhere
                dxi(1_i4b:m) = xmiss
            end where
!
            if ( present(xstd) ) xstd(1_i4b:m) = dxi(1_i4b:m) 
!
            if ( cor ) then
!
!               COMPUTE THE CORRELATION MATRIX, IF REQUIRED.
!
                b(1_i4b) = lamch( 's' )
!
                where ( dxi(1_i4b:m)>=b(1_i4b) .and. mask(1_i4b:m) )
                    a(1_i4b:m) = one/dxi(1_i4b:m)
                    mask(1_i4b:m) = false
                end where
!
!$OMP PARALLEL DO IF(test_par)                    &
!$OMP            ,SCHEDULE(STATIC,1)              &
!$OMP            ,PRIVATE(j, l, l1, l2, b, dxi )
!
                do j = 1_i4b, m
!
                    l = (j-1_i4b)*j/2_i4b
                    l1 = l + 1_i4b
                    l2 = l + j
!
                    if ( xcorp(l2)/=xmiss ) then
!
                        if ( .not.mask(j) ) then
!
                            where( mask(1_i4b:j-1_i4b) ) xcorp(l1:l2-1_i4b) = xmiss
!
                            where ( xcorp(l1:l2-1_i4b)/=xmiss )
!
                                b(1_i4b:j-1_i4b) = xcorp(l1:l2-1_i4b)*( a(1_i4b:j-1_i4b)*a(j) )
!
                                where ( b(1_i4b:j-1_i4b)<one )
                                    dxi(1_i4b:j-1_i4b) = b(1_i4b:j-1_i4b)
                                elsewhere
                                    dxi(1_i4b:j-1_i4b) = one
                                end where
!                                dxi(1_i4b:j-1_i4b) = merge( b(1_i4b:j-1_i4b), one, b(1_i4b:j-1_i4b)<one )
!
                                where ( dxi(1_i4b:j-1_i4b)>-one )
                                    xcorp(l1:l2-1_i4b) = dxi(1_i4b:j-1_i4b)
                                elsewhere
                                    xcorp(l1:l2-1_i4b) = -one
                                end where
!                                xcorp(l1:l2-1_i4b) = merge( dxi(1_i4b:j-1_i4b), -one, dxi(1_i4b:j-1_i4b)>-one )
!
                            end where
!
                            xcorp(l2) = one
!
                        else
!
                            xcorp(l1:l2) = xmiss
!
                        end if
!
                    end if
!
                end do
!
!$OMP END PARALLEL DO
!
            end if
!
        end if
!
        if ( present( failure ) ) then
!
!           SET MISSING INDICATOR IF REQUIRED.
!
            failure = any( xcorp(1_i4b:p)==xmiss )
!
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_cormat_miss_b
! ____________________________________
!
    end subroutine comp_cormat_miss_b
!
!
! =========================================================================================
!                               EOF ANALYSIS
! =========================================================================================
!
!
    subroutine comp_eof( x, first, last, xeigval, xeigvec, xn, failure, dimvar, cov, &
                         sort, maxiter, xmean, xstd, xeigvar, xcorp )
!
! Purpose
! _______
!
!   COMP_EOF computes estimates of Empirical Orthogonal Functions (EOF; also known
!   as Principal Component Analysis) from a data matrix.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:,:)
!            On entry, input submatrix containing size(X,3-DIMVAR)
!            observations on size(X,DIMVAR) variables from the matrix of data
!            for which Empirical Orthogonal Functions are desired. By default,
!            DIMVAR is equal to 1. See description of optional DIMVAR argument 
!            for details.
!            If all the data are available at once, X can be the full 
!            data matrix.
!
!   FIRST    (INPUT) logical(lgl)
!            On entry, if:
!
!            - FIRST = true  the current submatrix is the first submatrix 
!              of the data matrix.
!            - FIRST = false the current submatrix is not the first submatrix
!              of the data matrix.
!
!   LAST     (INPUT) logical(lgl)
!            On entry, if:
!
!            - LAST = true  the current submatrix is the last submatrix 
!              of the data matrix.
!            - LAST = false the current submatrix is not the last submatrix
!              of the data matrix.
!
!   XEIGVAL  (INPUT/OUTPUT) real(stnd), dimension(:)
!            On entry, after the first call to COMP_EOF (e.g. when FIRST=true),
!            XEIGVAL contains temporary results from  previous calls
!            to COMP_EOF. XEIGVAL should not be changed between calls to
!            COMP_EOF.
!
!            On exit, when LAST=true, XEIGVAL contains the eigenvalues of
!            th variance-covariance (or correlation) matrix from the data matrix.
!            The near zero eigenvalues made negative by round off errors are set 
!            to zero.
!
!            The size of XEIGVAL must verify: size(XEIGVAL) = size(X,DIMVAR).
!
!   XEIGVEC  (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, after the first call to COMP_EOF (e.g. when FIRST=true),
!            the matrix XEIGVEC contains temporary results from  
!            previous calls to COMP_EOF. XEIGVEC should not be changed
!            between calls to COMP_EOF.
!
!            On exit, when LAST=true, XEIGVEC contains the eigenvectors of
!            th variance-covariance (or correlation) matrix from the data matrix.
!
!            The shape of XEIGVEC must verify:
!
!            - size(XEIGVEC,1) = size(X,DIMVAR) ,
!            - size(XEIGVEC,2) = size(X,DIMVAR) .
!
!   XN       (INPUT/OUTPUT) real(stnd)
!            On entry, after the first call to COMP_EOF (e.g. when FIRST=true),
!            XN contains count of observations from previous calls
!            to COMP_EOF. XN should not be changed between calls to
!            COMP_EOF.
!
!            On exit, XN contains the number of observations in the data matrix.
!
!   FAILURE  (OUTPUT) logical(lgl)
!            On exit when LAST=true:
!
!            - FAILURE = false:   indicates successful exit.
!            - FAILURE = true:    indicates that fewer than two valid observations
!              were present or that the observations on some
!              variable were constant and the correlations
!              were requested or that maximum accuracy was not
!              achieved when computing the eigenvectors and the
!              eigenvalues.
!
!            On exit when LAST=false, FAILURE is always set to false.
!          
!   DIMVAR   (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVAR is present, DIMVAR is used as follows, if:
!
!            - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!              on size(X,1) variables.
!            - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!              on size(X,2) variables.
!
!            The default is DIMVAR = 1.
!
!   COV      (INPUT, OPTIONAL) logical(lgl)
!            On entry, when argument COV is present, COV is used as follows, if:
!
!            - COV= true,   XEIGVEC contains the eigenvectors of the variances-covariances 
!              matrix, when LAST=true.
!            - COV= false,  XEIGVEC contains the eigenvectors of the correlation matrix, 
!              when LAST=true.
!
!            By default, the eigenvectors of the correlation matrix are output.
!
!            COV needs to be specified only on the last call to COMP_EOF (e.g. when LAST=true).
!
!   SORT     (INPUT, OPTIONAL) character
!            Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!            or in descending order if SORT = 'D' or 'd'.
!            The eigenvectors are reordered accordingly.
!
!            SORT needs to be specified only on the last call to COMP_EOF (e.g. when LAST=true).
!
!   MAXITER  (INPUT, OPTIONAL) integer(i4b)
!            MAXITER controls the maximum number of QR sweeps in the Schur 
!            decomposition of an intermediate tridiagonal form T of the covariance matrix .
!            The algorithm fails to converge if the number of QR sweeps
!            exceeds MAXITER * size(XEIGVAL). Convergence usually occurs in 
!            about 2 * size(XEIGVAL) QR sweeps. The default is 30.
!
!            MAXITER needs to be specified only on the last call to COMP_EOF (e.g. when LAST=true).
!
!   XMEAN    (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            On exit, when LAST=true and XMEAN is present, XMEAN contains the variable 
!            means.
!
!            XMEAN needs to be specified only on the last call to COMP_EOF (e.g. when LAST=true).
!
!            The size of XMEAN must verify: size(XMEAN) = size(X,DIMVAR).
!          
!   XSTD     (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            On exit, when LAST=true and XSTD is present, XSTD contains the variable 
!            standard-deviations.
!
!            XSTD needs to be specified only on the last call to COMP_EOF (e.g. when LAST=true).
!
!            The size of XSTD must verify: size(XSTD) = size(X,DIMVAR).
!          
!   XEIGVAR  (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            On exit, when LAST=true and XEIGVAR is present, XEIGVAR contains
!            percentages of total variance associated with the eigenvectors in the
!            order of the eigenvalues stored in XEIGVAL.
!
!            XEIGVAR needs to be specified only on the last call to COMP_EOF (e.g. when LAST=true).
!
!            The size of XEIGVAR must verify: size(XEIGVAR) = size(X,DIMVAR).
!          
!   XCORP    (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            On exit, when LAST=true and XCORP is present, XCORP contains the upper 
!            triangle of the correlation or variance-covariance matrix, as controlled 
!            by the COV argument, stored in symmetric storage mode.
!            The upper triangle of the symmetric correlation or variance-covariance
!            matrix is packed columnwise in the linear array XCORP.  More precisely,
!            the j-th column of this matrix is stored in the array XCORP as follows:
!
!                     XCORP(i + (j-1) * j/2,2) = XCOR(i,j) for 1<=i<=j;
!
!            XCORP needs to be specified only on the last call to COMP_EOF (e.g. when LAST=true).
!
!            The size of XCORP must verify: 
!          
!            - size(XCORP) = (size(X,DIMVAR) * (size(X,DIMVAR)+1))/2
!          
!          
! Further Details
! _______________
!
!   The subroutine computes the Empirical Orthogonal Functions with only one pass
!   through the data.
!
!   This subroutine may be used in a call with no observations (e.g. size(X,3-DIMVAR) = 0)
!   in order to finish the computations with LAST=true when the total number of observations
!   is unknown at the beginning of the computations.
!
!   If fewer than two valid observations were present for some pair of variables or if 
!   the observations on some variable were constants, the statistics XEIGVAL, XEIGVEC,
!   XEIGVAR and XCORP are globally set to NaN code.
!
!   For more details on EOF or PCA analysis, see:
!
!   (1) von Storch, H., and Zwiers, F.W., 2002:
!            Statistical Analysis in Climate Research.
!            Cambridge, UK, Chapter 8, 484 pp., ISBN:9780521012300
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : zero, one, c100
    use Num_Constants,     only : nan, lamch
    use Char_Constants,    only : stat_error1, stat_error5
    use Utilities,         only : assert_eq, assert, merror, get_diag
    use Eig_Procedures,    only : eig_cmp2
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),             dimension(:,:) :: x
    real(stnd), intent(inout),          dimension(:)   :: xeigval
    real(stnd), intent(inout),          dimension(:,:) :: xeigvec
    real(stnd), intent(inout)                          :: xn
    real(stnd), intent(out),  optional , dimension(:)  :: xstd, xmean, xeigvar, xcorp
!
    integer(i4b), intent(in),  optional :: dimvar, maxiter
!
    logical(lgl), intent(in)            :: first, last
    logical(lgl), intent(in),  optional :: cov
    logical(lgl), intent(out)           :: failure
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                           :: a, b, xnm1, tmp
    real(stnd), dimension(size(xeigval)) :: dxi
!
    integer(i4b) :: i, i1, i2, j,  n, m
!
    logical(lgl) :: cor, dim_is_one
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_eof'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
    n = size( x, 3-int(i) )
!
    m = assert_eq(  int(size(x,int(i)),i4b),    &
                    int(size(xeigval),i4b),     &
                    int(size(xeigvec,1),i4b),   &
                    int(size(xeigvec,2),i4b),   &
                    name_proc )
!
    if ( m<=0_i4b ) call merror( name_proc//stat_error1 )
!
    if ( first ) then
!
!       THE CURRENT SUBMATRIX IS THE FIRST.
!
!       INITIALIZE xeigval, xeigvec AND xn .
!
        xeigval(1_i4b:m) = zero       
        xeigvec(1_i4b:m,1_i4b:m) = zero
        xn = zero
!
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  (m*(m+1_i4b))/2_i4b>=omp_limit      .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)      &
!$OMP         ,PRIVATE(i,j,tmp)
!
    do i = 1_i4b, n
!
!$OMP SINGLE
!
        xnm1 = xn
        xn = xnm1 + one
        a = one/xn
        b = xnm1*a
!
!       CALCULATE PARTIAL MEANS AND CORRECTED SUMS OF SQUARES AND CROSS-PRODUCTS, 
!       HANDLING THE DATA ONCE.
!
        if ( dim_is_one ) then
            dxi(1_i4b:m) = x(1_i4b:m,i) - xeigval(1_i4b:m)
        else
            dxi(1_i4b:m) = x(i,1_i4b:m) - xeigval(1_i4b:m)
        end if
!
        xeigval(1_i4b:m) = xeigval(1_i4b:m) + a*dxi(1_i4b:m)
!
!$OMP END SINGLE
!
!$OMP DO SCHEDULE(STATIC,1) 
!
        do j = 1_i4b, m
!
            tmp = b*dxi(j)
!
#ifdef _BLAS
            call axpy( j, tmp, dxi(1_i4b:j), 1_i4b, xeigvec(1_i4b:j,j), 1_i4b )
#else
            xeigvec(1_i4b:j,j) = xeigvec(1_i4b:j,j) + tmp*dxi(1_i4b:j)
#endif
!
        end do
!
!$OMP END DO
!
    end do
!
!$OMP END PARALLEL
!
    if ( last ) then
!
!       THE CURRENT SUBMATRIX IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
        b = nan()
        failure = true
!
!       TEST OPTIONAL ARGUMENTS.
!
        if ( present(cov) ) then
           cor = .not.cov
        else
           cor = true
        end if
!
        if ( present(xmean) ) then
!
!           OUTPUT THE MEANS IF REQUIRED.
!
            call assert( logical(size(xmean)==m,lgl),    &
                         name_proc )
!
            if ( xn<one ) then
                xmean(1_i4b:m) = b
            else
                xmean(1_i4b:m) = xeigval(1_i4b:m)
            end if
!
        end if
!
        if ( present(xstd) ) then
!
            call assert( logical(size(xstd)==m,lgl),    &
                         name_proc )
!
            if ( xn<=one ) then
                xstd(1_i4b:m) = b
            end if
!
        end if
!
        if ( present(xeigvar) ) then
!
            call assert( logical(size(xeigvar)==m,lgl),    &
                         name_proc )
!
        end if
!
        if ( present(xcorp) ) then
!
            n = ( m*(m+1_i4b) )/2_i4b
            call assert( logical(size(xcorp)==n,lgl),    &
                         name_proc )
!
        end if
!
        if ( xn>one ) then
!
!           MORE THAN TWO VALID OBSERVATIONS, COMPUTE THE STATISTICS.
!
            a = one/xn    
!
            if ( cor ) then
!
!               GET DIAGONAL ELEMENTS OF CORRECTED SUMS OF SQUARES 
!               AND CROSS-PRODUCTS MATRIX.
!
                dxi(1_i4b:m) = sqrt( get_diag( xeigvec ) )
!
                if ( present(xstd) ) then
!
!                   COMPUTE STANDARD-DEVIATIONS, IF REQUIRED.
!
                    xstd(1_i4b:m) = dxi(1_i4b:m)*sqrt( a )
                end if
!
!               SET FAILURE INDICATOR.
!
                xnm1 = lamch( 's' )
                failure = any( dxi(1_i4b:m)< xnm1 )
!
                if ( .not. failure ) then
!
!                   COMPUTE THE CORRELATION MATRIX.
!
                    dxi(1_i4b:m) = one/dxi(1_i4b:m)
!
!$OMP PARALLEL DO IF(test_par)                    &
!$OMP            ,SCHEDULE(STATIC,1)              &
!$OMP            ,PRIVATE(j)
!
                    do j = 1_i4b, m
!
                        xeigvec(1_i4b:j-1_i4b,j) = xeigvec(1_i4b:j-1_i4b,j)*( dxi(1_i4b:j-1_i4b)*dxi(j) )
                        xeigvec(j,j) = one
!
                    end do
!
!$OMP END PARALLEL DO
!
                end if
!
            else
!
!               COMPUTE THE VARIANCE-COVARIANCE MATRIX.
!
!$OMP PARALLEL DO IF(test_par)                    &
!$OMP            ,SCHEDULE(STATIC,1)              &
!$OMP            ,PRIVATE(j)
!
                do j = 1_i4b, m
!
                    xeigvec(1_i4b:j,j) = xeigvec(1_i4b:j,j)*a
!
                end do
!
!$OMP END PARALLEL DO
!
                if ( present(xstd) ) then
!
!                   COMPUTE STANDARD-DEVIATIONS, IF REQUIRED.
!
                    dxi(1_i4b:m) = get_diag( xeigvec )
                    xstd(1_i4b:m) = sqrt( dxi(1_i4b:m) )
!
                end if
!
!               SET FAILURE INDICATOR .
!
                failure = false
!
            end if
!
        end if
!
        if ( failure ) then
!
!           FEWER THAN TWO VALID OBSERVATIONS OR CONSTANT VARIABLES,
!           SET THE PERTINENT STATISTICS TO NaN CODE.
!
            xeigval(1_i4b:m) =  b
!
            xeigvec(1_i4b:m,1_i4b:m) = b
!
            if ( present(xeigvar) ) then
                xeigvar(1_i4b:m) = b
            end if
!
            if ( present(xcorp) ) then
                xcorp(1_i4b:n) = b
            end if
!
        else
!
            if ( present(xcorp) ) then
!
!               SAVE COVARIANCE OR CORRELATION MATRIX, IF REQUIRED.
!
!$OMP PARALLEL DO IF(test_par)              &
!$OMP            ,SCHEDULE(STATIC,1)        &
!$OMP            ,PRIVATE(j, i, i1, i2)
!
                do j = 1_i4b, m
!
                    i = (j-1_i4b)*j/2_i4b
                    i1 = i + 1_i4b
                    i2 = i + j
!
                    xcorp(i1:i2) = xeigvec(1_i4b:j,j)
!
                end do
!
!$OMP END PARALLEL DO
!
            end if
!
!           COMPUTE EIGENVALUES AND EIGENVECTORS.
!
            call eig_cmp2( xeigvec, xeigval, failure, sort=sort, maxiter=maxiter )
!
!           SET NEGATIVE EIGENVALUES TO ZERO .
!
            where ( xeigval(1_i4b:m) < zero )  xeigval(1_i4b:m) = zero
!
            if ( present(xeigvar) ) then
!
!               COMPUTE PERCENTAGES OF VARIANCE, IF REQUIRED.
!
                a = c100 / sum( xeigval(1_i4b:m) )
                xeigvar(1_i4b:m) = xeigval(1_i4b:m) * a
!
            end if
!
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_eof
! __________________________
!
    end subroutine comp_eof
!
! =========================================================================================
!
    subroutine comp_eof2( x, first, last, xeigval, xcorp, xn, failure, dimvar, cov, &
                          savecor, maxiter, ortho, xmean, xstd, xeigvar, xeigvec )
!
! Purpose
! _______
!
!   COMP_EOF2 computes estimates of Empirical Orthogonal Functions (EOF; also known
!   as Principal Component Analysis) from a data matrix.
!
!   COMP_EOF2 computes all the eigenvalues, and optionally selected eigenvectors
!   (by inverse iteration), of the covariance (or correlation matrix) from a data matrix.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:,:)
!            On entry, input submatrix containing size(X,3-DIMVAR)
!            observations on size(X,DIMVAR) variables from the matrix of data
!            for which Empirical Orthogonal Functions are desired. By default,
!            DIMVAR is equal to 1. See description of optional DIMVAR argument 
!            for details.
!            If all the data are available at once, X can be the full 
!            data matrix.
!
!   FIRST    (INPUT) logical(lgl)
!            On entry, if:
!
!            - FIRST = true  the current submatrix is the first submatrix 
!              of the data matrix.
!            - FIRST = false the current submatrix is not the first submatrix
!              of the data matrix.
!
!   LAST     (INPUT) logical(lgl)
!            On entry, if:
!
!            - LAST = true  the current submatrix is the last submatrix 
!              of the data matrix.
!            - LAST = false the current submatrix is not the last submatrix
!              of the data matrix.
!
!   XEIGVAL  (INPUT/OUTPUT) real(stnd), dimension(:)
!            On entry, after the first call to COMP_EOF2 (e.g. when FIRST=true),
!            XEIGVAL contains temporary results from  previous calls
!            to COMP_EOF2. XEIGVAL should not be changed between calls to
!            COMP_EOF2.
!
!            On exit, when LAST=true, XEIGVAL contains the eigenvalues of
!            th variance-covariance (or correlation) matrix from the data matrix.
!            The near zero eigenvalues made negative by round off errors are set 
!            to zero. The eigenvalues are sorted in descending order.
!            The eigenvectors in XEIGVEC are reordered accordingly.
!
!            The size of XEIGVAL must verify: size(XEIGVAL) = size(X,DIMVAR).
!
!
!   XCORP    (INPUT/OUTPUT) real(stnd), dimension(:)
!            On entry, after the first call to COMP_EOF2 (e.g. when FIRST=true),
!            the linear array XCORP contains the upper triangle of the corrected
!            sums of squared and cross-products matrix, packed columnwise, computed from  
!            previous calls to COMP_EOF2. XCORP should not be changed
!            between calls to COMP_EOF2.
!
!            On exit, when LAST=true and SAVECOR=true, XCORP contains the correlation
!            or variance-covariance matrix as controlled by the COV argument.
!            XCORP is stored in symmetric storage mode (see further details).
!            If SAVECOR=false, the correlation matrix is not saved on exit. In this
!            case XCORP does not contain useful information.
!
!            The size of XCORP must verify: size(XCORP) = (size(X,DIMVAR) * (size(X,DIMVAR)+1))/2.
!
!   XN       (INPUT/OUTPUT) real(stnd)
!            On entry, after the first call to COMP_EOF2 (e.g. when FIRST=true),
!            XN contains count of observations from previous calls
!            to COMP_EOF2. XN should not be changed between calls to
!            COMP_EOF2.
!
!            On exit, XN contains the number of observations in the data matrix.
!
!   FAILURE  (OUTPUT) logical(lgl)
!            On exit when LAST=true:
!
!            - FAILURE = false:   indicates successful exit.
!            - FAILURE = true:    indicates that fewer than two valid observations
!              were present or that the observations on some
!              variable were constant and the correlations
!              were requested or that maximum accuracy was not
!              achieved when computing the eigenvalues or that
!              some eigenvectors failed to converge with MAXITER
!              inverse iterations.
!
!            On exit when LAST=false, FAILURE is always set to false.
!          
!   DIMVAR   (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVAR is present, DIMVAR is used as follows, if:
!
!            - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!              on size(X,1) variables.
!            - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!              on size(X,2) variables.
!
!            The default is DIMVAR = 1.
!
!   COV      (INPUT, OPTIONAL) logical(lgl)
!            On entry, when argument COV is present, COV is used as follows, if:
!
!            - COV= true,   the eigenvalues and the eigenvectors are computed from 
!              the variances-covariances matrix, when LAST=true.
!            - COV= false,  the eigenvalues and the eigenvectors are computed from  
!              the correlation matrix, when LAST=true.
!
!            By default, the eigenvalues and eigenvectors of the correlation matrix are computed.
!
!            COV needs to be specified only on the last call to COMP_EOF2 (e.g. when LAST=true).
!
!   SAVECOR  (INPUT, OPTIONAL) logical(lgl)
!            On exit, when argument SAVECOR is present and LAST=true, SAVECOR is used as follows, if:
!
!            - SAVECOR= true, the correlation (or covariance) matrix is saved in packed form
!              in argument XCORP.
!            - SAVECOR= false, the correlation (or covariance) matrix is destroyed.
!
!            By default, the correlation (or covariance) matrix is destroyed.
!
!            SAVECOR needs to be specified only on the last call to COMP_EOF2 (e.g. when LAST=true).
!
!   MAXITER  (INPUT, OPTIONAL) integer(i4b)
!            The number of inverse iterations performed in the subroutine for computing
!            the eigenvectors. By default, 2 inverse iterations are performed 
!            for all the eigenvectors. This optional argument is used only
!            if the optional argument XEIGVEC is present.
!
!            MAXITER needs to be specified only on the last call to COMP_EOF2 (e.g. when LAST=true).
!
!   ORTHO    (INPUT, OPTIONAL) logical(lgl)
!            If ORTHO=true the computed eigenvectors are orthogonalized by the
!            Modified Gram-Schmidt algorithm. This optional argument is used only
!            if the optional argument XEIGVEC is present.
!            The default is FALSE.
!
!            ORTHO needs to be specified only on the last call to COMP_EOF2 (e.g. when LAST=true).
!
!   XMEAN    (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            On exit, when LAST=true and XMEAN is present, XMEAN contains the variable 
!            means.
!
!            XMEAN needs to be specified only on the last call to COMP_EOF2 (e.g. when LAST=true).
!
!            The size of XMEAN must verify: size(XMEAN) = size(X,DIMVAR).
!          
!   XSTD     (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            On exit, when LAST=true and XSTD is present, XSTD contains the variable 
!            standard-deviations.
!
!            XSTD needs to be specified only on the last call to COMP_EOF2 (e.g. when LAST=true).
!
!            The size of XSTD must verify: size(XSTD) = size(X,DIMVAR).
!          
!   XEIGVAR  (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            On exit, when LAST=true and XEIGVAR is present, XEIGVAR contains
!            percentages of total variance associated with the eigenvectors in the
!            order of the eigenvalues stored in XEIGVAL.
!
!            XEIGVAR needs to be specified only on the last call to COMP_EOF2 (e.g. when LAST=true).
!
!            The size of XEIGVAR must verify: size(XEIGVAR) = size(X,DIMVAR).
!          
!   XEIGVEC  (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!            On exit, when LAST=true, XEIGVEC contains the first size(XEIGVEC,2) eigenvectors
!            of the variance-covariance (or correlation) matrix from the data matrix.
!
!            XEIGVEC needs to be specified only on the last call to COMP_EOF2 (e.g. when LAST=true).
!
!            The shape of XEIGVEC must verify:
!
!            - size(XEIGVEC,1) = size(X,DIMVAR)  ,
!            - size(XEIGVEC,2) <= size(X,DIMVAR) .
!
!          
! Further Details
! _______________
!
!   The subroutine computes the means and the covariance (or correlation) matrix with only one pass
!   through the data.
!
!   On exit, if SAVECOR= true, the upper triangle of the symmetric correlation or variance-covariance
!   matrix XCOR is packed columnwise in the linear array XCORP.  More precisely, the j-th column
!   of XCOR is stored in the array XCORP as follows:
!
!            XCORP(i + (j-1) * j/2) = XCOR(i,j) for 1<=i<=j;
!
!   Eigenvalues and selected eigenvectors are computed from the packed correlation matrix when LAST=true.
!
!   This subroutine may be used in a call with no observations (e.g. size(X,3-DIMVAR) = 0),
!   in order to finish the computations with LAST=true when the total number of observations
!   is unknown at the beginning of the computations.
!
!   If fewer than two valid observations were present for some pair of variables or if 
!   the observations on some variable were constants, the statistics XEIGVAL, XEIGVEC,
!   XEIGVAR and XCORP are globally set to NaN code.
!
!   For more details on EOF or PCA analysis, see:
!
!   (1) von Storch, H., and Zwiers, F.W., 2002:
!            Statistical Analysis in Climate Research.
!            Cambridge, UK, Chapter 8, 484 pp., ISBN:9780521012300
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : zero, one, c100
    use Num_Constants,     only : nan, lamch
    use Char_Constants,    only : stat_error1, stat_error5, allocate_error
    use Utilities,         only : assert_eq, assert, merror
    use Eig_Procedures,    only : trid_inviter, eigval_cmp
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),             dimension(:,:)  :: x
    real(stnd), intent(inout),          dimension(:)    :: xeigval, xcorp
    real(stnd), intent(inout)                           :: xn
    real(stnd), intent(out),  optional , dimension(:)   :: xstd, xmean, xeigvar
    real(stnd), intent(out),  optional , dimension(:,:) :: xeigvec
!
    integer(i4b), intent(in),  optional :: dimvar, maxiter
!
    logical(lgl), intent(in)            :: first, last
    logical(lgl), intent(in),  optional :: cov, savecor, ortho
    logical(lgl), intent(out)           :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                              :: a, b, xnm1, tmp
    real(stnd), dimension(size(xeigval))    :: dxi
    real(stnd), allocatable, dimension(:)   :: work
    real(stnd), allocatable, dimension(:,:) :: d_e
!
    integer(i4b) :: i, l, l1, l2, j,  n, m, p
    integer      :: iok
!
    logical(lgl) :: cor, dim_is_one, save_cor
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_eof2'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
    n = size( x, 3-int(i) )
!
    m = assert_eq(  int(size(x,int(i)),i4b),    &
                    int(size(xeigval),i4b),     &
                    name_proc )
!
    if ( m<=0_i4b ) call merror( name_proc//stat_error1 )
!
    p = ( m*(m+1_i4b) )/2_i4b
!
    call assert( logical(size(xcorp)==p,lgl),    &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBMATRIX IS THE FIRST.
!
!       INITIALIZE xeigval, xcorp AND xn .
!
        xeigval(1_i4b:m) = zero       
        xcorp(1_i4b:p)   = zero
        xn               = zero
!
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  p>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)           &
!$OMP         ,PRIVATE(i,j,l,l1,l2,tmp)
!
    do i = 1_i4b, n
!
!$OMP SINGLE
!
        xnm1 = xn
        xn = xnm1 + one
        a = one/xn
        b = xnm1*a
!
!       CALCULATE PARTIAL MEANS AND CORRECTED SUMS OF SQUARES AND CROSS-PRODUCTS, 
!       HANDLING THE DATA ONCE.
!
        if ( dim_is_one ) then
            dxi(1_i4b:m) = x(1_i4b:m,i) - xeigval(1_i4b:m)
        else
            dxi(1_i4b:m) = x(i,1_i4b:m) - xeigval(1_i4b:m)
        end if
!
        xeigval(1_i4b:m) = xeigval(1_i4b:m) + a*dxi(1_i4b:m)
!
!$OMP END SINGLE
!
!$OMP DO SCHEDULE(STATIC,1) 
!
        do j = 1_i4b, m
!
            l = (j-1_i4b)*j/2_i4b
            l1 = l + 1_i4b
            l2 = l + j
!
            tmp = b*dxi(j)
!
#ifdef _BLAS
            call axpy( j, tmp, dxi(1_i4b:j), 1_i4b, xcorp(l1:l2), 1_i4b )
#else
            xcorp(l1:l2) = xcorp(l1:l2) + tmp*dxi(1_i4b:j)
#endif
!
        end do
!
!$OMP END DO
!
    end do
!
!$OMP END PARALLEL
!
    if ( last ) then
!
!       THE CURRENT SUBMATRIX IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
        b = nan()
        failure = true
!
!       TEST OPTIONAL ARGUMENTS.
!
        if ( present(cov) ) then
           cor = .not.cov
        else
           cor = true
        end if
!
        if ( present(savecor) ) then
           save_cor = savecor
        else
           save_cor = false
        end if
!
        if ( present(xmean) ) then
!
!           OUTPUT THE MEANS IF REQUIRED.
!
            call assert( logical(size(xmean)==m,lgl),    &
                         name_proc )
!
            if ( xn<one ) then
                xmean(1_i4b:m) = b
            else
                xmean(1_i4b:m) = xeigval(1_i4b:m)
            end if
!
        end if
!
        if ( present(xstd) ) then
!
            call assert( logical(size(xstd)==m,lgl),    &
                         name_proc )
!
            if ( xn<=one ) then
                xstd(1_i4b:m) = b
            end if
!
        end if
!
        if ( present(xeigvar) ) then
!
            call assert( logical(size(xeigvar)==m,lgl),    &
                         name_proc )
!
        end if
!
        if ( present(xeigvec) ) then
!
            n = size( xeigvec, 2 )
            call assert( logical(size(xeigvec,1)==m,lgl),    &
                         logical(n<=m,lgl),                  &
                         logical(1_i4b<=n,lgl),              &
                         name_proc )
!
        end if
!
        if ( xn>one ) then
!
!           MORE THAN TWO VALID OBSERVATIONS, COMPUTE THE STATISTICS.
!
            a = one/xn    
!
            if ( cor ) then
!
!               COMPUTE THE CORRELATION MATRIX.
!
!
!               GET DIAGONAL ELEMENTS OF CORRECTED SUMS OF SQUARES 
!               AND CROSS-PRODUCTS.
!
!$OMP PARALLEL DO IF(test_par)               &
!$OMP            ,SCHEDULE(STATIC,1)         &
!$OMP            ,PRIVATE(j, l)
!
                do j = 1_i4b, m
!
                    l = (j-1_i4b)*j/2_i4b + j
!
                    dxi(j) = sqrt( xcorp(l) )
!
                end do
!
!$OMP END PARALLEL DO
!
                if ( present(xstd) ) then
!
!                   COMPUTE STANDARD-DEVIATIONS, IF REQUIRED.
!
                    xstd(1_i4b:m) = dxi(1_i4b:m)*sqrt( a )
!
                end if
!
!               SET FAILURE INDICATOR.
!
                xnm1 = lamch( 's' )
                failure = any( dxi(1_i4b:m)< xnm1 )
!
                if ( .not. failure ) then
!
!                   COMPUTE THE CORRELATION MATRIX.
!
                    dxi(1_i4b:m) = one/dxi(1_i4b:m)
!
!$OMP PARALLEL DO IF(test_par)               &
!$OMP            ,SCHEDULE(STATIC,1)         &
!$OMP            ,PRIVATE(j, l, l1, l2)
!
                    do j = 1_i4b, m
!
                        l = (j-1_i4b)*j/2_i4b
                        l1 = l + 1_i4b
                        l2 = l + j
!
                        xcorp(l1:l2-1_i4b) = xcorp(l1:l2-1_i4b)*( dxi(1_i4b:j-1_i4b)*dxi(j) )
                        xcorp(l2) = one
!
                    end do
!
!$OMP END PARALLEL DO
!
                end if
!
            else
!
!               COMPUTE THE VARIANCE-COVARIANCE MATRIX.
!
                xcorp(1_i4b:p) = xcorp(1_i4b:p)*a
!
                if ( present(xstd) ) then
!
!                   COMPUTE STANDARD-DEVIATIONS, IF REQUIRED.
!
!$OMP PARALLEL DO IF(test_par)               &
!$OMP            ,SCHEDULE(STATIC,1)         &
!$OMP            ,PRIVATE(j, l)
!
                    do j = 1_i4b, m
!
                        l = (j-1_i4b)*j/2_i4b + j
!
                        xstd(j) = sqrt( xcorp(l) )
!
                    end do
!
!$OMP END PARALLEL DO
!
                end if
!
!               SET FAILURE INDICATOR .
!
                failure = false
!
            end if
!
        end if
!
        if ( failure ) then
!
!           FEWER THAN TWO VALID OBSERVATIONS OR CONSTANT VARIABLES,
!           SET THE PERTINENT STATISTICS TO NaN CODE.
!
            xcorp(1_i4b:n) = b
!
            xeigval(1_i4b:m) =  b
!
            if ( present(xeigvar) ) then
                xeigvar(1_i4b:m) = b
            end if
!
            if ( present(xeigvec) ) then
                xeigvec(1_i4b:m,1_i4b:n) = b
            end if
!
        else
!
            if ( save_cor ) then
!
!               ALLOCATE NEEDED WORKSPACE.
!
                allocate( work(p), stat=iok )
!
                if ( iok/=0 ) then
                    call merror( name_proc//allocate_error )
                end if
!
!               MAKE A COPY OF THE CORRELATION MATRIX.
!
                work(1_i4b:p) = xcorp(1_i4b:p)
!
                if ( present(xeigvec) ) then
!
!                   ALLOCATE NEEDED WORKSPACE.
!
                    allocate( d_e(m,2), stat=iok )
!
                    if ( iok/=0 ) then
                        call merror( name_proc//allocate_error )
                    end if
!
!                   COMPUTE ALL EIGENVALUES AND SAVE THE INTERMEDIATE 
!                   TRIDIAGONAL MATRIX IN PARAMETER d_e .
!
                    call eigval_cmp( work(:p), xeigval(:m), failure, sort='d', d_e=d_e(:m,:2) )
!
!                   COMPUTE THE FIRST n EIGENVECTORS OF THE CORRELATION MATRIX
!                   BY maxiter INVERSE ITERATIONS ON THE INTERMEDIATE TRIDIAGONAL
!                   MATRIX d_e AND BACK-TRANSFORMATION.
!
                    call trid_inviter( d_e(:m,1), d_e(:m,2), xeigval(:n), xeigvec(:m,:n), dim_is_one,     &
                                       matp=work(:p), maxiter=maxiter, ortho=ortho )
!
                    failure = failure .or. dim_is_one
!
!                   DEALLOCATE WORKSPACE.
!
                    deallocate( d_e )
!
                else
!
!                   COMPUTE ALL EIGENVALUES. 
!
                    call eigval_cmp( work(:p), xeigval(:m), failure, sort='d' )
!
                end if
!
!               DEALLOCATE WORKSPACE.
!
                deallocate( work )
!
            else
!
                if ( present(xeigvec) ) then
!
!                   ALLOCATE NEEDED WORKSPACE.
!
                    allocate( d_e(m,2), stat=iok )
!
                    if ( iok/=0 ) then
                        call merror( name_proc//allocate_error )
                    end if
!
!                   COMPUTE ALL EIGENVALUES AND SAVE THE INTERMEDIATE 
!                   TRIDIAGONAL MATRIX IN PARAMETER d_e .
!
                    call eigval_cmp( xcorp(1_i4b:p), xeigval(1_i4b:m), failure, sort='d', d_e=d_e(:m,:2) )
!
!                   COMPUTE THE FIRST n EIGENVECTORS OF THE SELF-ADJOINT MATRIX
!                   BY maxiter INVERSE ITERATIONS ON THE INTERMEDIATE TRIDIAGONAL MATRIX
!                   d_e AND BACK-TRANSFORMATION.
!
                    call trid_inviter( d_e(:m,1), d_e(:m,2), xeigval(:n), xeigvec(:m,:n), dim_is_one,   &
                                       matp=xcorp(:p), maxiter=maxiter, ortho=ortho )
!
                    failure = failure .or. dim_is_one
!
!                   DEALLOCATE WORKSPACE.
!
                    deallocate( d_e )
!
                else
!
!                   COMPUTE ALL EIGENVALUES. 
!
                    call eigval_cmp( xcorp(:p), xeigval(:m), failure, sort='d' )
!
                end if
!
            end if
!
!           SET NEGATIVE EIGENVALUES TO ZERO .
!
            where ( xeigval(1_i4b:m) < zero )  xeigval(1_i4b:m) = zero
!
            if ( present(xeigvar) ) then
!
!               COMPUTE PERCENTAGES OF VARIANCE, IF REQUIRED.
!
                a = c100 / sum( xeigval(1_i4b:m) )
                xeigvar(1_i4b:m) = xeigval(1_i4b:m) * a
!
            end if
!
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_eof2
! ___________________________
!
    end subroutine comp_eof2
!
! =========================================================================================
!
    subroutine comp_eof3( x, dimvar, failure, xcorp, xeigval, xeigvec, maxiter, ortho )
!
! Purpose
! _______ 
!
!   COMP_EOF3 computes estimates of Empirical Orthogonal Functions (EOF; also known
!   as Principal Component Analysis) from a data matrix X with n observations.
!
!   COMP_EOF3 computes the matrix product (1/n) (X' * X) or (1/n) (X * X') from
!   the data matrix X, all the eigenvalues, and selected eigenvectors 
!   (by inverse iteration), of this matrix product.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:,:)
!            On entry, input data matrix containing size(X,3-DIMVAR)
!            observations on size(X,DIMVAR) variables for which Empirical Orthogonal
!            Functions or Principal Components are desired.
!
!   DIMVAR   (INPUT) integer(i4b)
!            On entry, DIMVAR is used as follows, if:
!
!            - DIMVAR = 1, the input matrix X contains size(X,2) observations
!              on size(X,1) variables and the matrix product
!              (1/n) (X * X')  is computed.
!            - DIMVAR = 2, the input matrix X contains size(X,1) observations
!              on size(X,2) variables and the matrix product
!              (1/n) (X' * X) is computed.
!
!   FAILURE  (OUTPUT) logical(lgl)
!            On exit:
!
!            - FAILURE = false:   indicates successful exit.
!            - FAILURE = true:    indicates that fewer than two valid observations
!              were present or that maximum accuracy was not
!              achieved when computing the eigenvalues or that
!              some eigenvectors failed to converge with MAXITER
!              inverse iterations.
!
!   XCORP    (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            On exit, XCORP(:) contains the matrix product (1/n) (X' * X) or
!            (1/n) (X * X'), stored in symmetric storage mode.
!
!            The upper triangle of the symmetric matrix product matrix is packed columnwise 
!            in the linear array XCORP(:).  More precisely, the j-th column of this matrix
!            is stored in the array XCORP(:) as follows:
!
!                              XCORP(i + (j-1) * j/2,1) = XCOR(i,j) for 1<=i<=j;
!
!            The size of XCORP must verify: size(XCORP) = (size(X,DIMVAR) * (size(X,DIMVAR)+1))/2
!
!   XEIGVAL  (OUTPUT, OPTIONAL) real(stnd), dimension(:,2)
!            On exit:
!
!            - XEIGVAL(:,1) contains the eigenvalues in decreasing order of the matrix 
!              product (1/n) (X' * X) or (1/n) (X * X') from the data matrix X.
!              The near zero eigenvalues made negative by round off errors are set to zero.
!            - XEIGVAL(:,2) contains percentages of total variance associated with the
!              eigenvectors in the order of the eigenvalues stored in XEIGVAL(:,1).
!
!            The shape of XEIGVAL must verify:
!
!            - size(XEIGVAL,1) = size(X,DIMVAR) ,
!            - size(XEIGVAL,2) = 2 .
!          
!   XEIGVEC  (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!            On exit, XEIGVEC contains the first size(XEIGVEC,2) eigenvectors
!            of the matrix product (1/n) (X' * X) or (1/n) (X * X') from
!            the data matrix X.
!
!            The shape of XEIGVEC must verify:
!
!            - size(XEIGVEC,1) = size(X,DIMVAR) ,
!            - size(XEIGVEC,2) <= size(X,DIMVAR) .
!
!   MAXITER  (INPUT, OPTIONAL) integer(i4b)
!            The number of inverse iterations performed in the subroutine for computing
!            the eigenvectors. By default, 2 inverse iterations are performed 
!            for all the eigenvectors. This optional argument is used only if the XEIGVEC
!            is present.
!
!   ORTHO    (INPUT, OPTIONAL) logical(lgl)
!            If ORTHO=true the computed eigenvectors are orthogonalized by the
!            Modified Gram-Schmidt algorithm. This optional argument is used only if the XEIGVEC
!            is present.
!
!            The default is FALSE.
!
!
! Further Details
! _______________
!
!   The subroutine computes the Empirical Orthogonal Functions or the Principal Components
!   with only one pass through the data.
!
!   If size(X,3-DIMVAR)<=0, the subroutine set FAILURE to true and returns without doing
!   any computations.
!
!   For more details on EOF or PCA analysis, see:
!
!   (1) von Storch, H., and Zwiers, F.W., 2002:
!            Statistical Analysis in Climate Research.
!            Cambridge, UK, Chapter 8, 484 pp., ISBN:9780521012300
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : zero, c100
    use Char_Constants,    only : allocate_error, stat_error1, stat_error5
    use Utilities,         only : assert, merror
    use Eig_Procedures,    only : eigval_cmp, trid_inviter
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),              dimension(:,:) :: x
    real(stnd), intent(out),  optional , dimension(:)   :: xcorp
    real(stnd), intent(out),  optional , dimension(:,:) :: xeigval, xeigvec
!
    integer(i4b), intent(in)                            :: dimvar
    integer(i4b), intent(in),   optional                :: maxiter
!
    logical(lgl), intent(in),  optional :: ortho
    logical(lgl), intent(out)           :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                 :: tmp
    real(stnd), dimension(size(x,int(dimvar))) :: dxi
    real(stnd), dimension(:),   allocatable    :: xcorp2
    real(stnd), dimension(:,:), allocatable    :: d_e
!
    integer(i4b)                             :: i, j, l, l1, l2,  n, m, p, neig
    integer                                  :: iok
!
    logical(lgl)                             :: dim_is_one
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_eof3'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
!
        dim_is_one = dimvar==1_i4b
!
    else
!
        call merror( name_proc//stat_error5 )
!
    end if
!
    m = size( x, int(dimvar) )
!
    if ( m<=0_i4b ) call merror( name_proc//stat_error1 )
!
    n = size( x, 3-int(dimvar) )
!
    if ( n<=0_i4b ) then
!
        failure = true
!
        return
!
    end if
!
    p = ( m*(m+1_i4b) )/2_i4b
!
    if ( present( xcorp ) ) then
        call assert( logical(int(size(xcorp),i4b)==p,lgl),       &
                     name_proc )
    end if
!
    if ( present( xeigval ) ) then
        call assert( logical(int(size(xeigval,1),i4b)==m,lgl),   &
                     logical(size(xeigval,2)==2,lgl),            &
                     name_proc )
    end if
!
    if ( present( xeigvec ) ) then
        neig = size( xeigvec,2 )
        call assert( logical(int(size(xeigvec,1),i4b)==m,lgl),   &
                     logical(neig<=m,lgl),                       &
                     logical(1_i4b<=neig,lgl),                   &
                     name_proc )
    end if
!
!   ALLOCATE WORK ARRAY.
!
    allocate( xcorp2(p), stat=iok )
!
    if ( iok/=0 ) call merror( name_proc//allocate_error )
!
!   INITIALIZE xcorp2 AND xincp2 .
!
    xcorp2(1_i4b:p)  = zero
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  p>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)                &
!$OMP         ,PRIVATE(i,j,l,l1,l2)
!
    do i = 1_i4b, n
!
!$OMP SINGLE
!
!       CALCULATE SUMS OF SQUARES AND CROSS-PRODUCTS, 
!       HANDLING THE DATA ONCE.
!
        if ( dim_is_one ) then
            dxi(1_i4b:m) = x(1_i4b:m,i)
        else
            dxi(1_i4b:m) = x(i,1_i4b:m)
        end if
!
!$OMP END SINGLE
!
!
!$OMP DO SCHEDULE(STATIC,1) 
!
        do j = 1_i4b, m
!
            l = (j-1_i4b)*j/2_i4b
            l1 = l + 1_i4b
            l2 = l + j
!
#ifdef _BLAS
            call axpy( j, dxi(j), dxi(1_i4b:j), 1_i4b, xcorp2(l1:l2), 1_i4b )
#else
            xcorp2(l1:l2) = xcorp2(l1:l2) + dxi(1_i4b:j)*dxi(j)
#endif
!
        end do
!
!$OMP END DO
!
    end do
!
!$OMP END PARALLEL
!
!   POSTPROCESSING BEGIN HERE.
!
!   COMPUTE THE MATRIX PRODUCT.
!
    xcorp2(1_i4b:p) = xcorp2(1_i4b:p)/real( n, stnd )
!
    if ( present(xcorp) ) then
        xcorp(1_i4b:p) = xcorp2(1_i4b:p)
    end if
!
    if ( present(xeigvec) ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( d_e(m,2), stat=iok )
!
        if ( iok/=0 ) call merror( name_proc//allocate_error )
!
!       COMPUTE ALL EIGENVALUES AND SAVE THE INTERMEDIATE 
!       TRIDIAGONAL MATRIX IN PARAMETER d_e(:m,:2) .
!
        call eigval_cmp( xcorp2(:p), dxi(:m), failure, sort='d', d_e=d_e(:m,:2) )
!
!       COMPUTE THE FIRST n EIGENVECTORS OF THE SELF-ADJOINT MATRIX
!       BY maxiter INVERSE ITERATIONS ON THE INTERMEDIATE TRIDIAGONAL MATRIX
!       AND BACK-TRANSFORMATION.
!
        call trid_inviter( d_e(:m,1), d_e(:m,2), dxi(:neig), xeigvec(:m,:neig), dim_is_one,     &
                           matp=xcorp2(:p), maxiter=maxiter, ortho=ortho )
!
        failure = failure .or. dim_is_one
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( xcorp2, d_e )
!
    else if ( present(xeigval) ) then
!
!       COMPUTE ALL EIGENVALUES. 
!
        call eigval_cmp( xcorp2(:p), dxi(:m), failure, sort='d' )
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( xcorp2 )
!
    end if
!
    if ( present(xeigval) ) then
!
!       SET NEGATIVE EIGENVALUES TO ZERO.
!
        where ( dxi(1_i4b:m) < zero )
            xeigval(1_i4b:m,1_i4b) = zero
        elsewhere
            xeigval(1_i4b:m,1_i4b) = dxi(1_i4b:m)
        end where
!
!       COMPUTE PERCENTAGES OF VARIANCE.
!
        if ( any( xeigval(1_i4b:m,1_i4b) > zero ) ) then
            tmp = c100 / sum( xeigval(1_i4b:m,1_i4b) )
            xeigval(1_i4b:m,2_i4b) = xeigval(1_i4b:m,1_i4b) * tmp
        else
            xeigval(1_i4b:m,2_i4b) = zero
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_eof3
! ___________________________
!
    end subroutine comp_eof3
!
! =========================================================================================
!
    subroutine comp_eof_miss( x, first, last, xeigval, xeigvec, xcorp, xmiss, failure, &
                              dimvar, cov, sort, maxiter, xmean, xstd )
!
! Purpose
! _______ 
!
!   COMP_EOF_MISS computes estimates of Empirical Orthogonal Functions (EOF; also known as
!   Principal Component Analysis) from a data matrix possibly containing missing values.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:,:)
!            On entry, input submatrix containing size(X,3-DIMVAR)
!            observations on size(X,DIMVAR) variables from the matrix of data
!            for which Empirical Orthogonal Functions are desired. By default,
!            DIMVAR is equal to 1. See description of optional DIMVAR argument 
!            for details.
!            If all the data are available at once, X can be the full 
!            data matrix.
!
!   FIRST    (INPUT) logical(lgl)
!            On entry, if:
!
!            - FIRST = true  the current submatrix is the first submatrix 
!              of the data matrix.
!            - FIRST = false the current submatrix is not the first submatrix
!              of the data matrix.
!
!   LAST     (INPUT) logical(lgl)
!            On entry, if:
!
!            - LAST = true  the current submatrix is the last submatrix 
!              of the data matrix.
!            - LAST = false the current submatrix is not the last submatrix
!              of the data matrix.
!
!   XEIGVAL  (INPUT/OUTPUT) real(stnd), dimension(:,2)
!            On entry, after the first call to COMP_EOF_MISS (e.g. when FIRST=true),
!            XEIGVAL contains temporary results from  previous calls
!            to COMP_EOF_MISS. XEIGVAL should not be changed between calls to
!            COMP_EOF_MISS.
!            On exit, when LAST=true:
!
!            - XEIGVAL(:,1) contains the eigenvalues of the variance-covariance 
!              (or correlation) matrix from the data matrix.
!              The near zero eigenvalues made negative by round off errors or because
!              the variance-covariance (or correlation) matrix from the data matrix
!              with missing values is not positive definite are set to zero.
!            - XEIGVAL(:,2) contains percentages of total variance associated with the
!              eigenvectors in the order of the eigenvalues stored in XEIGVAL(:,1).
!
!            The shape of XEIGVAL must verify:
!
!            - size(XEIGVAL,1) = size(X,DIMVAR) ,
!            - size(XEIGVAL,2) = 2 .
!
!
!   XEIGVEC  (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, after the first call to COMP_EOF_MISS (e.g. when FIRST=true),
!            the matrix XEIGVEC contains temporary results from  
!            previous calls to COMP_EOF_MISS. XEIGVEC should not be changed
!            between calls to COMP_EOF_MISS.
!
!            On exit, when LAST=true, XEIGVEC contains the eigenvectors of
!            th variance-covariance (or correlation) matrix from the data matrix.
!
!            The shape of XEIGVEC must verify:
!
!            - size(XEIGVEC,1) = size(X,DIMVAR) ,
!            - size(XEIGVEC,2) = size(X,DIMVAR) .
!
!   XCORP    (INPUT/OUTPUT) real(stnd), dimension(:,3)
!            On entry, after the first call to COMP_EOF_MISS (e.g. when FIRST=true),
!            XCORP is used as workspace to accumulate quantities from previous calls
!            to COMP_EOF_MISS. XCORP should not be changed between calls to
!            COMP_EOF_MISS.
!
!            On exit, when LAST=true:
!
!            - XCORP(:,1) contains the correlation or variance-covariance matrix, as
!              controlled by the COV argument, stored in symmetric storage mode.
!              The upper triangle of the symmetric correlation or variance-covariance
!              matrix is packed columnwise in the linear array XCORP(:,1).  More precisely,
!              the j-th column of this matrix is stored in the array XCORP(:,1) as follows:
!
!                              XCORP(i + (j-1) * j/2,1) = XCOR(i,j) for 1<=i<=j;
!
!            - XCORP(:,2) contains the upper triangle of the matrix of the incidence values
!              between each pair of variables, packed columnwise, in a linear array. 
!              XCORP(i + (j-1) * j/2,2) indicates the numbers of non-missing pairs which were 
!              used in the calculation of the covariance (or correlation) between variables
!              i and j, for 1<=i<=j .
!            - XCORP(:,3) is used as workspace and contains no useful information.
!
!            The shape of XCORP must verify:
!
!            - size(XCORP,1) = (size(X,DIMVAR) * (size(X,DIMVAR)+1))/2 ,
!            - size(XCORP,2) = 3 .
!
!   XMISS    (INPUT) real(stnd)
!            On entry, the missing value indicator. Any value in
!            X which is equal to XMISS is assumed to be missing
!            or invalid.
!            The means, standard-deviations and the correlations are computed
!            on all the observations where X are not missing (see Further Details).
!
!   FAILURE  (OUTPUT) logical(lgl)
!            On exit when LAST=true:
!
!            - FAILURE = false:   indicates successful exit.
!            - FAILURE = true:    indicates that fewer than two valid observations
!              were present for some pair of variables
!              or that the observations on some
!              variable were constant and the correlations
!              were requested or that maximum accuracy was not
!              achieved when computing the eigenvectors and the
!              eigenvalues.
!
!            On exit when LAST=false, FAILURE is always set to false.
!
!   DIMVAR   (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVAR is present, DIMVAR is used as follows, if:
!
!            - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!              on size(X,1) variables.
!            - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!              on size(X,2) variables.
!
!            The default is DIMVAR = 1.
!          
!   COV      (INPUT, OPTIONAL) logical(lgl)
!            On entry, when argument COV is present, COV is used as follows, if:
!          
!            - COV= true,   XEIGVEC contains the eigenvectors of the variances-covariances 
!              matrix, when LAST=true.
!            - COV= false,  XEIGVEC contains the eigenvectors of the correlation matrix, 
!              when LAST=true.
!
!            By default, the  eigenvectors of the correlation matrix are output.
!
!            COV needs to be specified only on the last call to COMP_EOF_MISS (e.g. when LAST=true).
!
!   SORT     (INPUT, OPTIONAL) character
!            Sort the eigenvalues into ascending order if SORT = 'A' or 'a',
!            or in descending order if SORT = 'D' or 'd'.
!            The eigenvectors are reordered accordingly.
!
!            SORT needs to be specified only on the last call to COMP_EOF_MISS (e.g. when LAST=true).
!
!   MAXITER  (INPUT, OPTIONAL) integer(i4b)
!            MAXITER controls the maximum number of QR sweeps in the Schur 
!            decomposition of an intermediate tridiagonal form T of the covariance matrix.
!            The algorithm fails to converge if the number of QR sweeps
!            exceeds MAXITER * size(XEIGVAL). Convergence usually occurs in 
!            about 2 * size(XEIGVAL) QR sweeps. The default is 30.
!
!            MAXITER needs to be specified only on the last call to COMP_EOF_MISS (e.g. when LAST=true).
!
!   XMEAN    (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            On exit, when LAST=true and XMEAN is present, XMEAN contains the variable 
!            means computed from all non-missing observations in the data matrix.
!
!            XMEAN needs to be specified only on the last call to COMP_EOF_MISS (e.g. when LAST=true).
!
!            The size of XMEAN must verify: size(XMEAN) = size(X,DIMVAR).
!
!   XSTD     (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            On exit, when LAST=true and XSTD is present, XSTD contains the variable 
!            standard-deviations.
!
!            XSTD needs to be specified only on the last call to COMP_EOF_MISS (e.g. when LAST=true).
!
!            The size of XSTD must verify: size(XSTD) = size(X,DIMVAR).
!
!
! Further Details
! _______________
!
!   The subroutine computes the Empirical Orthogonal Functions with only one pass
!   through the data.
!
!   If fewer than two valid observations were present for some pair of variables or if 
!   the observations on some variable were constants, the statistics XEIGVAL, XEIGVEC,
!   and XCORP(:,1) are globally set to XMISS.
!
!   The means and standard-deviations of the data matrix are computed from all valid data. The
!   correlation coefficients are based on these univariate statistics and on all valid pairs
!   of observations. The eigenvectors and eigenvalues are computed from these bivariate statistics.
!
!   This subroutine may be used in a call with no observations (e.g. size(X,3-DIMVAR) = 0)
!   in order to finish the computations with LAST=true when the total number of observations
!   is unknown at the beginning of the computations.
!
!   For more details on EOF or PCA analysis, see:
!
!   (1) von Storch, H., and Zwiers, F.W., 2002:
!            Statistical Analysis in Climate Research.
!            Cambridge, UK, Chapter 8, 484 pp., ISBN:9780521012300
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : zero, one, c100
    use Num_Constants,     only : lamch
    use Char_Constants,    only : stat_error1, stat_error5
    use Utilities,         only : assert_eq, assert, merror
    use Eig_Procedures,    only : eig_cmp2
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                              :: xmiss
    real(stnd), intent(in),              dimension(:,:) :: x
    real(stnd), intent(inout),           dimension(:,:) :: xeigval, xcorp, xeigvec
    real(stnd), intent(out),  optional , dimension(:)   :: xstd, xmean
!
    integer(i4b), intent(in),  optional :: dimvar, maxiter
!
    logical(lgl), intent(in)            :: first, last
    logical(lgl), intent(in),  optional :: cov
    logical(lgl), intent(out)           :: failure
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                             :: tmp
    real(stnd), dimension(size(xeigval,1)) :: dxi, a, b
!
    integer(i4b) :: i, j, l, l1, l2,  n, m, p
!
    logical(lgl)                             :: cor, dim_is_one
    logical(lgl), dimension(size(xeigval,1)) :: mask
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_eof_miss'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
    n = size( x, 3-int(i) )
!
    m = assert_eq(  int(size(x,int(i)),i4b),    &
                    int(size(xeigval,1),i4b),   &
                    int(size(xeigvec,1),i4b),   &
                    int(size(xeigvec,2),i4b),   &
                    name_proc )
!
    if ( m<=0_i4b ) call merror( name_proc//stat_error1 )
!
    p = ( m*(m+1_i4b) )/2_i4b
!
    call assert( logical(int(size(xcorp,1),i4b)==p,lgl),   &
                 logical(size(xcorp,2)==3,lgl),            &
                 logical(size(xeigval,2)==2,lgl),          &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBMATRIX IS THE FIRST.
!
!       INITIALIZE xeigval, xeigvec AND xcorp .
!
        xeigval(1_i4b:m,1_i4b)      = zero       
        xeigvec(1_i4b:m,1_i4b:m)    = zero
        xcorp(1_i4b:p,1_i4b:3_i4b)  = zero
!
        xeigval(1_i4b:m,2_i4b) = xmiss
!
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  p>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)                &
!$OMP         ,PRIVATE(i,j,l,l1,l2,b,tmp)
!
    do i = 1_i4b, n
!
!$OMP SINGLE
!
!       CALCULATE PARTIAL MEANS AND CORRECTED SUMS OF SQUARES AND CROSS-PRODUCTS, 
!       HANDLING THE DATA ONCE.
!
        if ( dim_is_one ) then
            dxi(1_i4b:m) = x(1_i4b:m,i)
        else
            dxi(1_i4b:m) = x(i,1_i4b:m)
        end if
!
        mask(1_i4b:m) = dxi(1_i4b:m)/=xmiss
!
#ifdef _WHERE
        do j = 1_i4b, m
            if ( xeigval(j,2_i4b)==xmiss ) xeigval(j,2_i4b) = dxi(j)
        end do
#else
        where ( xeigval(1_i4b:m,2_i4b)==xmiss ) xeigval(1_i4b:m,2_i4b) = dxi(1_i4b:m)
#endif
!
!$OMP END SINGLE
!
!$OMP DO SCHEDULE(STATIC,1) 
!
        do j = 1_i4b, m
!
            if ( mask(j) ) then
!
                l = (j-1_i4b)*j/2_i4b
                l1 = l + 1_i4b
                l2 = l + j
!
                where ( mask(1_i4b:j) )
!
                    b(1_i4b:j) = dxi(1_i4b:j) - xeigval(1_i4b:j,2_i4b)
                    xeigvec(1_i4b:j,j) = xeigvec(1_i4b:j,j) + b(1_i4b:j)*b(j)
!
                    xcorp(l1:l2,1_i4b) = xcorp(l1:l2,1_i4b) + b(1_i4b:j)
                    xcorp(l1:l2,2_i4b) = xcorp(l1:l2,2_i4b) + one
                    xcorp(l1:l2,3_i4b) = xcorp(l1:l2,3_i4b) + b(j)
!
                end where
!
                tmp = dxi(j) - xeigval(j,1_i4b)
                xeigval(j,1_i4b) = xeigval(j,1_i4b) + tmp/xcorp(l2,2_i4b)
!
            end if
!
        end do
!
!$OMP END DO
!
    end do
!
!$OMP END PARALLEL
!
    if ( last ) then
!
!       THE CURRENT SUBMATRIX IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
        failure = any(  xcorp(1_i4b:p,2_i4b)<=one )
!
!       TEST OPTIONAL ARGUMENTS.
!
        if ( present(cov) ) then
           cor = .not.cov
        else
           cor = true
        end if
!
        where ( xeigval(1_i4b:m,2_i4b)/=xmiss )
!
!            COMPUTE DIFFERENCES BETWEEN TEMPORARY MEANS AND MEANS.
!
            dxi(1_i4b:m) = xeigval(1_i4b:m,2_i4b) - xeigval(1_i4b:m,1_i4b)
!
        elsewhere
!
!            FEWER THAN ONE VALID OBSERVATION, SET THE MEANS TO MISSING CODE.
!
             xeigval(1_i4b:m,1_i4b) = xmiss
!
        end where
!
        if ( present(xmean) ) then
!
!           OUTPUT THE MEANS IF REQUIRED.
!
            call assert( logical(size(xmean)==m,lgl),    &
                         name_proc )
!
            xmean(1_i4b:m) = xeigval(1_i4b:m,1_i4b)
!
        end if
!
        if ( present(xstd) ) then
            call assert( logical(size(xstd)==m,lgl),    &
                         name_proc )
        end if
!
        if ( failure ) then
!
            if ( present(xstd) ) then
!
!               COMPUTE STANDARD-DEVIATIONS, IF REQUIRED.
!
                do j = 1_i4b, m
!
                    l = (j-1_i4b)*j/2_i4b
                    l2 = l + j
!
                    if ( xcorp(l2,2_i4b)>one ) then
                        a(j) = xeigvec(j,j) + dxi(j)*xcorp(l2,3_i4b) +                    &
                               dxi(j)*xcorp(l2,1_i4b) + (dxi(j)*dxi(j))*xcorp(l2,2_i4b)
                        xstd(j) = sqrt( a(j)/xcorp(l2,2_i4b) ) 
                    else
                        xstd(j) = xmiss 
                    end if
!
                end do
!
            end if
!
        else
!
!           COMPUTE THE VARIANCE-COVARIANCE MATRIX.
!
!$OMP PARALLEL DO IF(test_par)              &
!$OMP            ,SCHEDULE(STATIC,1)        &
!$OMP            ,PRIVATE(j, l, l1, l2)
!
            do j = 1_i4b, m
!
                l = (j-1_i4b)*j/2_i4b
                l1 = l + 1_i4b
                l2 = l + j
!
                xeigvec(1_i4b:j,j) = xeigvec(1_i4b:j,j) + dxi(1_i4b:j)*xcorp(l1:l2,3_i4b) +                    &
                                     dxi(j)*xcorp(l1:l2,1_i4b) + (dxi(j)*dxi(1_i4b:j))*xcorp(l1:l2,2_i4b)
                xeigvec(1_i4b:j,j) = xeigvec(1_i4b:j,j)/xcorp(l1:l2,2_i4b)
!
                a(j) = xeigvec(j,j)
!
            end do
!
!$OMP END PARALLEL DO
!
            if ( cor .or. present(xstd) ) then
!
!               COMPUTE STANDARD-DEVIATIONS, IF REQUIRED.
!
                dxi(1_i4b:m) = sqrt( a(1_i4b:m) )
!
                if ( present(xstd) ) then
                    xstd(1_i4b:m) = dxi(1_i4b:m)
                end if
!
                if ( cor ) then
!
!                   COMPUTE THE CORRELATION MATRIX, IF REQUIRED.
!
                    tmp = lamch( 's' )
                    mask(1_i4b:m) = dxi(1_i4b:m) < tmp
!
!                   SET FAILURE INDICATOR.
!
                    failure = any( mask(1_i4b:m) )
!
                    if ( .not. failure ) then
!
                        a(1_i4b:m) = one/dxi(1_i4b:m)
!
!$OMP PARALLEL DO IF(test_par)              &
!$OMP            ,SCHEDULE(STATIC,1)        &
!$OMP            ,PRIVATE(j, b, dxi )
!
                        do j = 1_i4b, m
!
                            b(1_i4b:j-1_i4b) = xeigvec(1_i4b:j-1_i4b,j)*( a(1_i4b:j-1_i4b)*a(j) )
!
                            where( b(1_i4b:j-1_i4b)<one )
                                dxi(1_i4b:j-1_i4b) = b(1_i4b:j-1_i4b)
                            elsewhere
                                dxi(1_i4b:j-1_i4b) = one
                            end where
!                            dxi(1_i4b:j-1_i4b) = merge( b(1_i4b:j-1_i4b), one, b(1_i4b:j-1_i4b)<one )
!
                            where( dxi(1_i4b:j-1_i4b)>-one )
                                xeigvec(1_i4b:j-1_i4b,j) = dxi(1_i4b:j-1_i4b)
                            elsewhere
                                xeigvec(1_i4b:j-1_i4b,j) = -one
                            end where
!                            xeigvec(1_i4b:j-1_i4b,j) = merge( dxi(1_i4b:j-1_i4b), -one, dxi(1_i4b:j-1_i4b)>-one )
!
                            xeigvec(j,j) = one
!
                        end do
!
!$OMP END PARALLEL DO
!
                    end if
!
                end if
!
            end if
!
        end if
!
        if ( failure ) then
!
!           FEWER THAN TWO VALID OBSERVATIONS OR CONSTANT VARIABLES,
!           SET THE PERTINENT STATISTICS TO MISSING CODE.
!
            xeigval(1_i4b:m,1_i4b:2_i4b) =  xmiss
!
            xeigvec(1_i4b:m,1_i4b:m) = xmiss
!
            xcorp(1_i4b:p,1_i4b) = xmiss
!
        else
!
!           SAVE COVARIANCE OR CORRELATION MATRIX.
!
!$OMP PARALLEL DO IF(test_par)              &
!$OMP            ,SCHEDULE(STATIC,1)        &
!$OMP            ,PRIVATE(j, l, l1, l2)
!
            do j = 1_i4b, m
!
                l = (j-1_i4b)*j/2_i4b
                l1 = l + 1_i4b
                l2 = l + j
!
                xcorp(l1:l2,1_i4b) = xeigvec(1_i4b:j,j)
!
            end do
!
!$OMP END PARALLEL DO
!
!           COMPUTE EIGENVALUES AND EIGENVECTORS.
!
            call eig_cmp2( xeigvec, xeigval(1_i4b:m,1_i4b), failure, sort=sort, maxiter=maxiter )
!
!           SET NEGATIVE EIGENVALUES TO ZERO.
!
            where ( xeigval(1_i4b:m,1_i4b) < zero )  xeigval(1_i4b:m,1_i4b) = zero
!
!           COMPUTE PERCENTAGES OF VARIANCE.
!
            if ( any( xeigval(1_i4b:m,1_i4b) > zero ) ) then
                tmp = c100 / sum( xeigval(1_i4b:m,1_i4b) )
                xeigval(1_i4b:m,2_i4b) = xeigval(1_i4b:m,1_i4b) * tmp
            else
                xeigval(1_i4b:m,2_i4b) = zero
            end if
!
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_eof_miss
! _______________________________
!
    end subroutine comp_eof_miss
!
! =========================================================================================
!
    subroutine comp_eof_miss2( x, first, last, xeigval, xcorp, xmiss, failure, dimvar, &
                               cov, maxiter, ortho, xmean, xstd, xeigvec )
!
! Purpose
! _______ 
!
!   COMP_EOF_MISS2 computes estimates of Empirical Orthogonal Functions (EOF; also known as
!   Principal Component Analysis) from a data matrix possibly containing missing values.
!
!   COMP_EOF_MISS2 computes all the eigenvalues, and optionally selected eigenvectors
!   (by inverse iteration), of the covariance (or correlation matrix) from a data matrix.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:,:)
!            On entry, input submatrix containing size(X,3-DIMVAR)
!            observations on size(X,DIMVAR) variables from the matrix of data
!            for which Empirical Orthogonal Functions are desired. By default,
!            DIMVAR is equal to 1. See description of optional DIMVAR argument 
!            for details.
!            If all the data are available at once, X can be the full 
!            data matrix.
!
!   FIRST    (INPUT) logical(lgl)
!            On entry, if:
!
!            - FIRST = true  the current submatrix is the first submatrix 
!              of the data matrix.
!            - FIRST = false the current submatrix is not the first submatrix
!              of the data matrix.
!
!   LAST     (INPUT) logical(lgl)
!            On entry, if:
!
!            - LAST = true  the current submatrix is the last submatrix 
!              of the data matrix.
!            - LAST = false the current submatrix is not the last submatrix
!              of the data matrix.
!
!   XEIGVAL  (INPUT/OUTPUT) real(stnd), dimension(:,2)
!            On entry, after the first call to COMP_EOF_MISS2 (e.g. when FIRST=true),
!            XEIGVAL contains temporary results from  previous calls
!            to COMP_EOF_MISS2. XEIGVAL should not be changed between calls to
!            COMP_EOF_MISS2.
!
!            On exit, when LAST=true:
!
!            - XEIGVAL(:,1) contains the eigenvalues of the variance-covariance 
!              (or correlation) matrix from the data matrix.
!              The near zero eigenvalues made negative by round off errors or because
!              the variance-covariance (or correlation) matrix from the data matrix
!              with missing values is not positive definite are set to zero.
!            - XEIGVAL(:,2) contains percentages of total variance associated with the
!              eigenvectors in the order of the eigenvalues stored in XEIGVAL(:,1).
!
!            The shape of XEIGVAL must verify:
!
!            - size(XEIGVAL,1) = size(X,DIMVAR) ,
!            - size(XEIGVAL,2) = 2 .
!
!   XCORP    (INPUT/OUTPUT) real(stnd), dimension(:,4)
!            On entry, after the first call to COMP_EOF_MISS2 (e.g. when FIRST=true),
!            XCORP is used as workspace to accumulate quantities from previous calls
!            to COMP_EOF_MISS2. XCORP should not be changed between calls to
!            COMP_EOF_MISS2.
!
!            On exit:
!
!            - XCORP(:,1) contains the correlation or variance-covariance matrix, as 
!              controlled by the COV argument, stored in symmetric storage mode.
!              The upper triangle of the symmetric correlation or variance-covariance
!              matrix is packed columnwise in the linear array XCORP(:,1).  More precisely,
!              the j-th column of this matrix is stored in the array XCORP(:,1) as follows:
!
!                              XCORP(i + (j-1) * j/2,1) = XCOR(i,j) for 1<=i<=j;
!
!            - XCORP(:,2) contains the upper triangle of the matrix of the incidence values
!              between each pair of variables, packed columnwise, in a linear array. 
!              XCORP(i + (j-1) * j/2,2) indicates the numbers of non-missing pairs which were 
!              used in the calculation of the covariance (or correlation) between variables
!              i and j, for 1<=i<=j .
!            - XCORP(:,3:4) is used as workspace and contains no useful informations.
!
!            The shape of XCORP must verify:
!
!            - size(XCORP,1) = (size(X,DIMVAR) * (size(X,DIMVAR)+1))/2 ,
!            - size(XCORP,2) = 4 .
!
!   XMISS    (INPUT) real(stnd)
!            On entry, the missing value indicator. Any value in
!            X which is equal to XMISS is assumed to be missing
!            or invalid.
!            The means, standard-deviations and the correlations are computed
!            on all the observations where X are not missing (see Further Details).
!
!   FAILURE  (OUTPUT) logical(lgl)
!            On exit when LAST=true:
!
!            - FAILURE = false:   indicates successful exit.
!            - FAILURE = true:    indicates that fewer than two valid observations
!              were present for some pair of variables
!              or that the observations on some
!              variable were constant and the correlations
!              were requested or that maximum accuracy was not
!              achieved when computing the eigenvalues or that
!              some eigenvectors failed to converge with MAXITER
!              inverse iterations.
!
!            On exit when LAST=false, FAILURE is always set to false.
!
!   DIMVAR   (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVAR is present, DIMVAR is used as follows, if:
!
!            - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!              on size(X,1) variables.
!            - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!              on size(X,2) variables.
!
!            The default is DIMVAR = 1.
!          
!   COV      (INPUT, OPTIONAL) logical(lgl)
!            On entry, when argument COV is present, COV is used as follows, if:
!
!            - COV= true,   the eigenvalues and the eigenvectors are computed from 
!              the variances-covariances matrix, when LAST=true.
!            - COV= false,  the eigenvalues and the eigenvectors are computed from  
!              the correlation matrix, when LAST=true.
!
!            By default, the eigenvalues and eigenvectors of the correlation matrix are computed.
!
!            COV needs to be specified only on the last call to COMP_EOF_MISS2 (e.g. when LAST=true).
!
!   MAXITER  (INPUT, OPTIONAL) integer(i4b)
!            The number of inverse iterations performed in the subroutine for computing
!            the eigenvectors. By default, 2 inverse iterations are performed 
!            for all the eigenvectors. This optional argument is used only if the XEIGVEC
!            is present.
!
!            MAXITER needs to be specified only on the last call to COMP_EOF_MISS2 (e.g. when LAST=true).
!
!   ORTHO    (INPUT, OPTIONAL) logical(lgl)
!            If ORTHO=true the computed eigenvectors are orthogonalized by the
!            Modified Gram-Schmidt algorithm. This optional argument is used only if the XEIGVEC
!            is present.
!            The default is FALSE.
!
!            ORTHO needs to be specified only on the last call to COMP_EOF_MISS2 (e.g. when LAST=true).
!
!   XMEAN    (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            On exit, when LAST=true and XMEAN is present, XMEAN contains the variable 
!            means computed from all non-missing observations in the data matrix.
!
!            XMEAN needs to be specified only on the last call to COMP_EOF_MISS2 (e.g. when LAST=true).
!
!            The size of XMEAN must verify: size(XMEAN) = size(X,DIMVAR).
!
!   XSTD     (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            On exit, when LAST=true and XSTD is present, XSTD contains the variable 
!            standard-deviations.
!
!            XSTD needs to be specified only on the last call to COMP_EOF_MISS2 (e.g. when LAST=true).
!
!            The size of XSTD must verify: size(XSTD) = size(X,DIMVAR).
!
!   XEIGVEC  (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!            On exit, when LAST=true, XEIGVEC contains the first size(XEIGVEC,2) eigenvectors
!            of the variance-covariance (or correlation) matrix from the data matrix.
!
!            XEIGVEC needs to be specified only on the last call to COMP_EOF_MISS2 (e.g. when LAST=true).
!
!            The shape of XEIGVEC must verify:
!
!            - size(XEIGVEC,1) = size(X,DIMVAR) ,
!            - size(XEIGVEC,2) <= size(X,DIMVAR) .
!
!
! Further Details
! _______________
!
!   The subroutine computes the Empirical Orthogonal Functions with only one pass
!   through the data.
!
!   If fewer than two valid observations were present for some pair of variables or if 
!   the observations on some variable were constants, the statistics XEIGVAL, XEIGVEC,
!   and XCORP(:,1) are globally set to XMISS .
!
!   The means and standard-deviations of the data matrix are computed from all valid data. The
!   correlation coefficients are based on these univariate statistics and on all valid pairs
!   of observations. The eigenvectors and eigenvalues are computed from these bivariate statistics.
!
!   This subroutine may be used in a call with no observations (e.g. size(X,3-DIMVAR) = 0)
!   in order to finish the computations with LAST=true when the total number of observations
!   is unknown at the beginning of the computations.
!
!   For more details on EOF or PCA analysis, see:
!
!   (1) von Storch, H., and Zwiers, F.W., 2002:
!            Statistical Analysis in Climate Research.
!            Cambridge, UK, Chapter 8, 484 pp., ISBN:9780521012300
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : zero, one, c100
    use Num_Constants,     only : lamch
    use Char_Constants,    only : stat_error1, stat_error5
    use Utilities,         only : assert_eq, assert, merror
    use Eig_Procedures,    only : eigval_cmp, trid_inviter
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                              :: xmiss
    real(stnd), intent(in),              dimension(:,:) :: x
    real(stnd), intent(inout),           dimension(:,:) :: xeigval, xcorp
    real(stnd), intent(out),  optional , dimension(:)   :: xmean, xstd
    real(stnd), intent(out),  optional , dimension(:,:) :: xeigvec
!
    integer(i4b), intent(in),  optional :: dimvar, maxiter
!
    logical(lgl), intent(in)            :: first, last
    logical(lgl), intent(in),  optional :: cov, ortho
    logical(lgl), intent(out)           :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                             :: tmp
    real(stnd), dimension(size(xeigval,1)) :: dxi, a, b
!
    integer(i4b) :: i, j, l, l1, l2,  n, m, p
!
    logical(lgl)                             :: cor, dim_is_one
    logical(lgl), dimension(size(xeigval,1)) :: mask
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_eof_miss2'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
    n = size( x, 3-int(i) )
!
    m = assert_eq(  int(size(x,int(i)),i4b),    &
                    int(size(xeigval,1),i4b),   &
                    name_proc )
!
    if ( m<=0_i4b ) call merror( name_proc//stat_error1 )
!
    p = ( m*(m+1_i4b) )/2_i4b
!
    call assert( logical(int(size(xcorp,1),i4b)==p,lgl),   &
                 logical(size(xcorp,2)==4,lgl),            &
                 logical(size(xeigval,2)==2,lgl),          &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBMATRIX IS THE FIRST.
!
!       INITIALIZE xeigval, xcorp .
!
        xeigval(1_i4b:m,1_i4b)      = zero       
        xcorp(1_i4b:p,1_i4b:4_i4b)  = zero
!
        xeigval(1_i4b:m,2_i4b) = xmiss
!
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  p>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)                &
!$OMP         ,PRIVATE(i,j,l,l1,l2,b,tmp)
!
    do i = 1_i4b, n
!
!$OMP SINGLE
!
!       CALCULATE PARTIAL MEANS AND CORRECTED SUMS OF SQUARES AND CROSS-PRODUCTS, 
!       HANDLING THE DATA ONCE.
!
        if ( dim_is_one ) then
            dxi(1_i4b:m) = x(1_i4b:m,i)
        else
            dxi(1_i4b:m) = x(i,1_i4b:m)
        end if
!
        mask(1_i4b:m) = dxi(1_i4b:m)/=xmiss
!
#ifdef _WHERE
        do j = 1_i4b, m
            if ( xeigval(j,2_i4b)==xmiss ) xeigval(j,2_i4b) = dxi(j)
        end do
#else
        where ( xeigval(1_i4b:m,2_i4b)==xmiss ) xeigval(1_i4b:m,2_i4b) = dxi(1_i4b:m)
#endif
!
!$OMP END SINGLE
!
!
!$OMP DO SCHEDULE(STATIC,1) 
!
        do j = 1_i4b, m
!
            if ( mask(j) ) then
!
                l = (j-1_i4b)*j/2_i4b
                l1 = l + 1_i4b
                l2 = l + j
!
                where ( mask(1_i4b:j) )
!
                    b(1_i4b:j) = dxi(1_i4b:j) - xeigval(1_i4b:j,2_i4b)
!
                    xcorp(l1:l2,1_i4b) = xcorp(l1:l2,1_i4b) + b(1_i4b:j)
                    xcorp(l1:l2,2_i4b) = xcorp(l1:l2,2_i4b) + one
                    xcorp(l1:l2,3_i4b) = xcorp(l1:l2,3_i4b) + b(j)
                    xcorp(l1:l2,4_i4b) = xcorp(l1:l2,4_i4b) + b(1_i4b:j)*b(j)
!
                end where
!
                tmp = dxi(j) - xeigval(j,1_i4b)
                xeigval(j,1_i4b) = xeigval(j,1_i4b) + tmp/xcorp(l2,2_i4b)
!
            end if
!
        end do
!
!$OMP END DO
!
    end do
!
!$OMP END PARALLEL
!
    if ( last ) then
!
!       THE CURRENT SUBMATRIX IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
        failure = any(  xcorp(1_i4b:p,2_i4b)<=one )
!
!       TEST OPTIONAL ARGUMENTS.
!
        if ( present(cov) ) then
           cor = .not.cov
        else
           cor = true
        end if
!
        where ( xeigval(1_i4b:m,2_i4b)/=xmiss )
!
!            COMPUTE DIFFERENCES BETWEEN TEMPORARY MEANS AND MEANS.
!
            dxi(1_i4b:m) = xeigval(1_i4b:m,2_i4b) - xeigval(1_i4b:m,1_i4b)
!
        elsewhere
!
!            FEWER THAN ONE VALID OBSERVATION, SET THE MEANS TO MISSING CODE.
!
             xeigval(1_i4b:m,1_i4b) = xmiss
!
        end where
!
        if ( present(xmean) ) then
!
!           OUTPUT THE MEANS IF REQUIRED.
!
            call assert( logical(size(xmean)==m,lgl),    &
                         name_proc )
!
            xmean(1_i4b:m) = xeigval(1_i4b:m,1_i4b)
!
        end if
!
        if ( present(xstd) ) then
            call assert( logical(size(xstd)==m,lgl),    &
                         name_proc )
        end if
!
        if ( present(xeigvec) ) then
!
            n = size( xeigvec, 2 )
!
            call assert( logical(size(xeigvec,1)==m,lgl),    &
                         logical(n<=m,lgl),                  &
                         logical(1_i4b<=n,lgl),              &
                         name_proc )
!
        end if
!
        if ( failure ) then
!
            if ( present(xstd) ) then
!
!               COMPUTE STANDARD-DEVIATIONS, IF REQUIRED.
!
                do j = 1_i4b, m
!
                    l = (j-1_i4b)*j/2_i4b
                    l2 = l + j
!
                    if ( xcorp(l2,2_i4b)>one ) then
!
                        a(j) = xcorp(l2,4_i4b) + dxi(j)*xcorp(l2,3_i4b) +                    &
                               dxi(j)*xcorp(l2,1_i4b) + (dxi(j)*dxi(j))*xcorp(l2,2_i4b)
                        xstd(j) = sqrt( a(j)/xcorp(l2,2_i4b) ) 
!
                    else
!
                        xstd(j) = xmiss 
!
                    end if
!
                end do
!
            end if
!
        else
!
!           COMPUTE THE VARIANCE-COVARIANCE MATRIX.
!
!$OMP PARALLEL DO IF(test_par)              &
!$OMP            ,SCHEDULE(STATIC,1)        &
!$OMP            ,PRIVATE(j, l, l1, l2)
!
            do j = 1_i4b, m
!
                l = (j-1_i4b)*j/2_i4b
                l1 = l + 1_i4b
                l2 = l + j
!
                xcorp(l1:l2,4_i4b) = xcorp(l1:l2,4_i4b) + dxi(1_i4b:j)*xcorp(l1:l2,3_i4b) +                    &
                                     dxi(j)*xcorp(l1:l2,1_i4b) + (dxi(j)*dxi(1_i4b:j))*xcorp(l1:l2,2_i4b)
                xcorp(l1:l2,4_i4b) = xcorp(l1:l2,4_i4b)/xcorp(l1:l2,2_i4b)
!
                a(j) = xcorp(l2,4_i4b)
!
            end do
!
!$OMP END PARALLEL DO
!
            if ( cor .or. present(xstd) ) then
!
!               COMPUTE STANDARD-DEVIATIONS, IF REQUIRED.
!
                dxi(1_i4b:m) = sqrt( a(1_i4b:m) )
!
                if ( present(xstd) ) then
                    xstd(1_i4b:m) = dxi(1_i4b:m)
                end if
!
                if ( cor ) then
!
!                   COMPUTE THE CORRELATION MATRIX, IF REQUIRED.
!
                    tmp = lamch( 's' )
                    mask(1_i4b:m) = dxi(1_i4b:m) < tmp
!
!                   SET FAILURE INDICATOR.
!
                    failure = any( mask(1_i4b:m) )
!
                    if ( .not. failure ) then
!
                        a(1_i4b:m) = one/dxi(1_i4b:m)
!
!$OMP PARALLEL DO IF(test_par)              &
!$OMP            ,SCHEDULE(STATIC,1)        &
!$OMP            ,PRIVATE(j, l, l1, l2, b, dxi )
!
                        do j = 1_i4b, m
!
                            l = (j-1_i4b)*j/2_i4b
                            l1 = l + 1_i4b
                            l2 = l + j
!
                            b(1_i4b:j-1_i4b) = xcorp(l1:l2-1_i4b,4_i4b)*( a(1_i4b:j-1_i4b)*a(j) )
!
                            where ( b(1_i4b:j-1_i4b)<one )
                                dxi(1_i4b:j-1_i4b) = b(1_i4b:j-1_i4b)
                            elsewhere
                                dxi(1_i4b:j-1_i4b) = one
                            end where
!                            dxi(1_i4b:j-1_i4b) = merge( b(1_i4b:j-1_i4b), one, b(1_i4b:j-1_i4b)<one )
!
                            where ( dxi(1_i4b:j-1_i4b)>-one )
                                xcorp(l1:l2-1_i4b,4_i4b) = dxi(1_i4b:j-1_i4b)
                            elsewhere
                                xcorp(l1:l2-1_i4b,4_i4b) = -one
                            end where
!                            xcorp(l1:l2-1_i4b,4_i4b) = merge( dxi(1_i4b:j-1_i4b), -one, dxi(1_i4b:j-1_i4b)>-one )
!
                            xcorp(l2,4_i4b) = one
!
                        end do
!
!$OMP END PARALLEL DO
!
                    end if
!
                end if
!
            end if
!
        end if
!
        if ( failure ) then
!
!           FEWER THAN TWO VALID OBSERVATIONS OR CONSTANT VARIABLES,
!           SET THE PERTINENT STATISTICS TO MISSING CODE.
!
            xeigval(1_i4b:m,1_i4b:2_i4b) =  xmiss
!
            if ( present(xeigvec) ) then
                xeigvec(1_i4b:m,1_i4b:n) = xmiss
            end if
!
            xcorp(1_i4b:p,1_i4b) = xmiss
!
        else
!
!           SAVE COVARIANCE OR CORRELATION MATRIX.
!
            xcorp(1_i4b:p,1_i4b) = xcorp(1_i4b:p,4_i4b)
!
            if ( present(xeigvec) ) then
!
!               COMPUTE ALL EIGENVALUES AND SAVE THE INTERMEDIATE 
!               TRIDIAGONAL MATRIX IN PARAMETER xeigval(:m,:2) .
!
                call eigval_cmp( xcorp(:p,4_i4b), a(:m), failure, sort='d', d_e=xeigval(:m,:2) )
!
!               COMPUTE THE FIRST n EIGENVECTORS OF THE SELF-ADJOINT MATRIX
!               BY maxiter INVERSE ITERATIONS ON THE INTERMEDIATE TRIDIAGONAL MATRIX
!               AND BACK-TRANSFORMATION.
!
                call trid_inviter( xeigval(:m,1), xeigval(:m,2), a(:n), xeigvec(:m,:n), dim_is_one,     &
                                   matp=xcorp(:p,4_i4b), maxiter=maxiter, ortho=ortho )
!
                failure = failure .or. dim_is_one
!
            else
!
!               COMPUTE ALL EIGENVALUES. 
!
                call eigval_cmp( xcorp(:p,4_i4b), a(:m), failure, sort='d' )
!
            end if
!
!           SET NEGATIVE EIGENVALUES TO ZERO.
!
            where ( a(1_i4b:m) < zero )
                xeigval(1_i4b:m,1_i4b) = zero
            elsewhere
                xeigval(1_i4b:m,1_i4b) = a(1_i4b:m)
            end where
!
!           COMPUTE PERCENTAGES OF VARIANCE.
!
            if ( any( xeigval(1_i4b:m,1_i4b) > zero ) ) then
                tmp = c100 / sum( xeigval(1_i4b:m,1_i4b) )
                xeigval(1_i4b:m,2_i4b) = xeigval(1_i4b:m,1_i4b) * tmp
            else
                xeigval(1_i4b:m,2_i4b) = zero
            end if
!
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_eof_miss2
! ________________________________
!
    end subroutine comp_eof_miss2
!
! =========================================================================================
!
    subroutine comp_eof_miss3( x, xmiss, dimvar, failure, xcorp, xincp, xeigval, &
                               xeigvec, maxiter, ortho )
!
! Purpose
! _______ 
!
!   COMP_EOF_MISS3 computes estimates of Empirical Orthogonal Functions (EOF) 
!   or Principal Components (PC) from a data matrix X with n observations possibly
!   containing missing values.
!
!   COMP_EOF_MISS3 computes an estimate of the matrix product (1/n) (X' * X) or
!   (1/n) (X * X') from the data matrix X, the associated matrix of incidence values,
!   all the eigenvalues, and selected eigenvectors (by inverse iteration), of this estimate 
!   of the matrix product.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:,:)
!            On entry, input data matrix containing size(X,3-DIMVAR)
!            observations on size(X,DIMVAR) variables for which Empirical Orthogonal
!            Functions or Principal Components are desired.
!
!   XMISS    (INPUT) real(stnd)
!            On entry, the missing value indicator. Any value in
!            X which is equal to XMISS is assumed to be missing
!            or invalid.
!            The estimate of the matrix product (1/n) (X' * X) or (1/n) (X * X')
!            is computed on all the observations where X are not missing (see Further Details).
!
!   DIMVAR   (INPUT) integer(i4b)
!            On entry, DIMVAR is used as follows, if:
!
!            - DIMVAR = 1, the input matrix X contains size(X,2) observations
!              on size(X,1) variables and the matrix product
!              (1/n) (X * X')  is computed.
!            - DIMVAR = 2, the input matrix X contains size(X,1) observations
!              on size(X,2) variables and the matrix product
!              (1/n) (X' * X) is computed.
!
!   FAILURE  (OUTPUT) logical(lgl)
!            On exit:
!
!            - FAILURE = false:   indicates successful exit.
!            - FAILURE = true:    indicates that fewer than one valid observation
!              were present for some pair of variables or that
!              maximum accuracy was not achieved when computing
!              the eigenvalues or that some eigenvectors failed 
!              to converge with MAXITER inverse iterations.
!
!   XCORP    (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            On exit, XCORP(:) contains the estimate of the matrix product (1/n) (X' * X) or
!            (1/n) (X * X'), stored in symmetric storage mode.
!
!            The upper triangle of the symmetric matrix product matrix is packed columnwise 
!            in the linear array XCORP(:).  More precisely, the j-th column of this matrix
!            is stored in the array XCORP(:) as follows:
!
!                              XCORP(i + (j-1) * j/2,1) = XCOR(i,j) for 1<=i<=j;
!
!            The size of XCORP must verify: size(XCORP) = (size(X,DIMVAR) * (size(X,DIMVAR)+1))/2
!
!   XINCP    (OUTPUT, OPTIONAL) integer(i4b), dimension(:)
!            On exit, XINCP(:) contains the upper triangle of the matrix of the incidence values
!            between each pair of variables, packed columnwise, in a linear array. 
!            XINCP(i + (j-1) * j/2) indicates the numbers of non-missing pairs which were 
!            used in the calculation of the scalar product between variables
!            i and j, for 1<=i<=j .
!
!            The size of XINCP must verify:  size(XINCP) = (size(X,DIMVAR) * (size(X,DIMVAR)+1))/2
!
!   XEIGVAL  (OUTPUT, OPTIONAL) real(stnd), dimension(:,2)
!            On exit:
!
!            - XEIGVAL(:,1) contains the eigenvalues in decreasing order of the estimate of the
!              matrix product (1/n) (X' * X) or (1/n) (X * X') from the data matrix X.
!              The near zero eigenvalues made negative by round off errors or because
!              the matrix product from the data matrix X with missing values 
!              is not positive definite are set to zero.
!            - XEIGVAL(:,2) contains percentages of total variance associated with the
!              eigenvectors in the order of the eigenvalues stored in XEIGVAL(:,1).
!
!            The shape of XEIGVAL must verify:
!
!            - size(XEIGVAL,1) = size(X,DIMVAR) ,
!            - size(XEIGVAL,2) = 2 .
!          
!   XEIGVEC  (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!            On exit, XEIGVEC contains the first size(XEIGVEC,2) eigenvectors
!            of the estimate of the matrix product (1/n) (X' * X) or (1/n) (X * X')
!            from the data matrix X.
!
!            The shape of XEIGVEC must verify:
!
!            - size(XEIGVEC,1) = size(X,DIMVAR) ,
!            - size(XEIGVEC,2) <= size(X,DIMVAR) .
!
!   MAXITER  (INPUT, OPTIONAL) integer(i4b)
!            The number of inverse iterations performed in the subroutine for computing
!            the eigenvectors. By default, 2 inverse iterations are performed 
!            for all the eigenvectors. This optional argument is used only if the XEIGVEC
!            is present.
!
!   ORTHO    (INPUT, OPTIONAL) logical(lgl)
!            If ORTHO=true the computed eigenvectors are orthogonalized by the
!            Modified Gram-Schmidt algorithm. This optional argument is used only if the XEIGVEC
!            is present.
!
!            The default is FALSE.
!
!
! Further Details
! _______________
!
!   The subroutine computes the Empirical Orthogonal Functions or the Principal Components
!   with only one pass through the data.
!
!   The estimate of the matrix product (1/n) (X' * X) or (1/n) (X * X') is computed
!   from all valid pairs of observations. The eigenvectors and eigenvalues are computed from
!   these bivariate statistics.
!
!   If fewer than one valid observation is present for some pair of variables, 
!   the scalar product between this pair of variables is set to zero for computing the
!   eigenvectors and eigenvalues of the estimated matrix product.
!   
!   If size(X,3-DIMVAR)<=0, the subroutine set FAILURE to true and returns without doing
!   any computations.
!
!   For more details on EOF or PCA analysis, see:
!
!   (1) von Storch, H., and Zwiers, F.W., 2002:
!            Statistical Analysis in Climate Research.
!            Cambridge, UK, Chapter 8, 484 pp., ISBN:9780521012300
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : zero, c100
    use Char_Constants,    only : allocate_error, stat_error1, stat_error5
    use Utilities,         only : assert, merror
    use Eig_Procedures,    only : eigval_cmp, trid_inviter
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                              :: xmiss
    real(stnd), intent(in),              dimension(:,:) :: x
    real(stnd), intent(out),  optional , dimension(:)   :: xcorp
    real(stnd), intent(out),  optional , dimension(:,:) :: xeigval, xeigvec
!
    integer(i4b), intent(in)                            :: dimvar
    integer(i4b), intent(in),   optional                :: maxiter
    integer(i4b), intent(out),  optional , dimension(:) :: xincp
!
    logical(lgl), intent(in),  optional :: ortho
    logical(lgl), intent(out)           :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                 :: tmp
    real(stnd), dimension(size(x,int(dimvar))) :: dxi
    real(stnd), dimension(:),   allocatable    :: xcorp2
    real(stnd), dimension(:,:), allocatable    :: d_e
!
    integer(i4b)                             :: i, j, l, l1, l2,  n, m, p, neig
    integer(i4b), dimension(:), allocatable  :: xincp2
    integer                                  :: iok
!
    logical(lgl)                                  :: dim_is_one
    logical(lgl), dimension(size(x,int(dimvar)))  :: mask
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_eof_miss3'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
!
        dim_is_one = dimvar==1_i4b
!
    else
!
        call merror( name_proc//stat_error5 )
!
    end if
!
    m = size( x, int(dimvar) )
!
    if ( m<=0_i4b ) call merror( name_proc//stat_error1 )
!
    n = size( x, 3-int(dimvar) )
!
    if ( n<=0_i4b ) then
!
        failure = true
!
        return
!
    end if
!
    p = ( m*(m+1_i4b) )/2_i4b
!
    if ( present( xcorp ) ) then
        call assert( logical(int(size(xcorp),i4b)==p,lgl),       &
                     name_proc )
    end if
!
    if ( present( xincp ) ) then
        call assert( logical(int(size(xincp),i4b)==p,lgl),       &
                     name_proc )
    end if
!
    if ( present( xeigval ) ) then
        call assert( logical(int(size(xeigval,1),i4b)==m,lgl),   &
                     logical(size(xeigval,2)==2,lgl),            &
                     name_proc )
    end if
!
    if ( present( xeigvec ) ) then
!
        neig = size( xeigvec,2 )
!
        call assert( logical(int(size(xeigvec,1),i4b)==m,lgl),   &
                     logical(neig<=m,lgl),                       &
                     logical(1_i4b<=neig,lgl),                   &
                     name_proc )
!
    end if
!
!   ALLOCATE WORK ARRAYS.
!
    allocate( xcorp2(p), xincp2(p), stat=iok )
!
    if ( iok/=0 ) call merror( name_proc//allocate_error )
!
!   INITIALIZE xcorp2 AND xincp2.
!
    xcorp2(1_i4b:p)  = zero
    xincp2(1_i4b:p)  = 0_i4b
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  p>=omp_limit                        .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)                &
!$OMP         ,PRIVATE(i,j,l,l1,l2)
!
    do i = 1_i4b, n
!
!$OMP SINGLE
!
!       CALCULATE SUMS OF SQUARES AND CROSS-PRODUCTS, 
!       HANDLING THE DATA ONCE.
!
        if ( dim_is_one ) then
            dxi(1_i4b:m) = x(1_i4b:m,i)
        else
            dxi(1_i4b:m) = x(i,1_i4b:m)
        end if
!
        mask(1_i4b:m) = dxi(1_i4b:m)/=xmiss
!
!$OMP END SINGLE
!
!
!$OMP DO SCHEDULE(STATIC,1) 
!
        do j = 1_i4b, m
!
            if ( mask(j) ) then
!
                l = (j-1_i4b)*j/2_i4b
                l1 = l + 1_i4b
                l2 = l + j
!
                where ( mask(1_i4b:j) )
                    xcorp2(l1:l2) = xcorp2(l1:l2) + dxi(1_i4b:j)*dxi(j)
                    xincp2(l1:l2) = xincp2(l1:l2) + 1_i4b
                end where
!
            end if
!
        end do
!
!$OMP END DO
!
    end do
!
!$OMP END PARALLEL
!
!   POSTPROCESSING BEGIN HERE.
!
    failure = any(  xincp2(1_i4b:p)==0_i4b )
!
    if ( failure ) then
!
!       COMPUTE THE MATRIX PRODUCT.
!
        where( xincp2(1_i4b:p)/=0_i4b ) xcorp2(1_i4b:p) = xcorp2(1_i4b:p)/real( xincp2(1_i4b:p), stnd )
!
        if ( present(xcorp) ) then
            where( xincp2(1_i4b:p)==0_i4b )
                xcorp(1_i4b:p) = xmiss
            elsewhere
                xcorp(1_i4b:p) = xcorp2(1_i4b:p)
            end where
        end if
!
    else
!
!       COMPUTE THE MATRIX PRODUCT.
!
        xcorp2(1_i4b:p) = xcorp2(1_i4b:p)/real( xincp2(1_i4b:p), stnd )
!
        if ( present(xcorp) ) then
            xcorp(1_i4b:p) = xcorp2(1_i4b:p)
        end if
!
    end if
!
    if ( present(xincp) ) then
        xincp(1_i4b:p) = xincp2(1_i4b:p)
    end if
!
!   DEALLOCATE WORK ARRAY.
!
    deallocate( xincp2 )
!
    if ( present(xeigvec) ) then
!
!       ALLOCATE WORK ARRAY.
!
        allocate( d_e(m,2), stat=iok )
!
        if ( iok/=0 ) call merror( name_proc//allocate_error )
!
!       COMPUTE ALL EIGENVALUES AND SAVE THE INTERMEDIATE 
!       TRIDIAGONAL MATRIX IN PARAMETER d_e(:m,:2) .
!
        call eigval_cmp( xcorp2(:p), dxi(:m), dim_is_one, sort='d', d_e=d_e(:m,:2) )
!
        failure = failure .or. dim_is_one
!
!       COMPUTE THE FIRST n EIGENVECTORS OF THE SELF-ADJOINT MATRIX
!       BY maxiter INVERSE ITERATIONS ON THE INTERMEDIATE TRIDIAGONAL MATRIX
!       AND BACK-TRANSFORMATION.
!
        call trid_inviter( d_e(:m,1), d_e(:m,2), dxi(:neig), xeigvec(:m,:neig), dim_is_one,     &
                           matp=xcorp2(:p), maxiter=maxiter, ortho=ortho )
!
        failure = failure .or. dim_is_one
!
!       DEALLOCATE WORK ARRAYS.
!
        deallocate( xcorp2, d_e )
!
    else if ( present(xeigval) ) then
!
!       COMPUTE ALL EIGENVALUES. 
!
        call eigval_cmp( xcorp2(:p), dxi(:m), dim_is_one, sort='d' )
!
        failure = failure .or. dim_is_one
!
!       DEALLOCATE WORK ARRAY.
!
        deallocate( xcorp2 )
!
    end if
!
    if ( present(xeigval) ) then
!
!       SET NEGATIVE EIGENVALUES TO ZERO.
!
        where ( dxi(1_i4b:m) < zero )
            xeigval(1_i4b:m,1_i4b) = zero
        elsewhere
            xeigval(1_i4b:m,1_i4b) = dxi(1_i4b:m)
        end where
!
!       COMPUTE PERCENTAGES OF VARIANCE.
!
        if ( any( xeigval(1_i4b:m,1_i4b) > zero ) ) then
            tmp = c100 / sum( xeigval(1_i4b:m,1_i4b) )
            xeigval(1_i4b:m,2_i4b) = xeigval(1_i4b:m,1_i4b) * tmp
        else
            xeigval(1_i4b:m,2_i4b) = zero
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_eof_miss3
! ________________________________
!
    end subroutine comp_eof_miss3
!
! =========================================================================================
!
    subroutine comp_pc_eof( x, xeigvec, xsingval, xpc, dimvar, xmean, xstd, xpccor  )
!
! Purpose
! _______
!
!   COMP_PC_EOF computes estimates of Principal Components (PC) 
!   from a data matrix and a set of eigenvectors derived
!   from an EOF or PCA analysis.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:,:)
!            On entry, input submatrix containing size(X,3-DIMVAR)
!            observations on size(X,DIMVAR) variables from the matrix of data
!            for which Principal Components are desired. By default,
!            DIMVAR is equal to 1. See description of optional DIMVAR argument 
!            for details.
!            If all the data are available at once, X can be the full 
!            data matrix.
!
!   XEIGVEC  (INPUT) real(stnd), dimension(:,:)
!            On entry, XEIGVEC contains selected eigenvectors of
!            th variance-covariance (or correlation) matrix from the data matrix.
!
!            The shape of XEIGVEC must verify: size(XEIGVEC,1) = size(X,DIMVAR).
!
!
!   XSINGVAL (INPUT) real(stnd), dimension(:)
!            On entry, XSINGVAL must contain the singular values of the covariance
!            (or correlation) matrix from the data matrix associated with the
!            eigenvectors in XEIGVEC array. The Principal Components are normalized 
!            by XSINGVAL on output (the variances of the Principal Components are equal
!            to one).
!
!            The size of XSINGVAL must verify: size(XSINGVAL) = size(XEIGVEC,2).
!
!   XPC      (OUTPUT) real(stnd), dimension(:,:)
!            On exit, XPC contains the normalized Principal Components derived
!            from X and XEIGVEC.
!
!            The shape of XPC must verify:
!
!            - size(XPC,1) = size(X,3-DIMVAR) ,
!            - size(XPC,2) = size(XEIGVEC,2) .
!
!   DIMVAR   (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVAR is present, DIMVAR is used as follows, if:
!
!            - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!              on size(X,1) variables.
!            - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!              on size(X,2) variables.
!
!            The default is DIMVAR = 1.
!
!   XMEAN    (INPUT, OPTIONAL) real(stnd), dimension(:)
!            On entry, if XMEAN is present, XMEAN contains the variable 
!            means and the Principal Components are computed from the
!            centered data matrix X.
!
!            The size of XMEAN must verify: size(XMEAN) = size(X,DIMVAR).
!          
!   XSTD     (INPUT, OPTIONAL) real(stnd), dimension(:)
!            On entry, if XSTD is present, XSTD contains the variable 
!            standard-deviations and the Principal Components are computed from 
!            the normalized data matrix X.
!
!            The size of XSTD must verify: size(XSTD) = size(X,DIMVAR).
!          
!   XPCCOR   (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!            On exit, when XPCCOR is present, XPCCOR contains the correlations (or covariances)
!            between the data matrix XX and the principal components (factor loadings matrix).
!
!            This optional argument may be specified in a call to COMP_PC_EOF with
!            size(X,3-DIMVAR) = size(XPC,1) = 0 (e.g. with no observations) such as:
!
!                  call comp_pc_eof( x(:nvar,1:0), xeigvec(:nvar,:neigvec), xsingval(:neigvec),  &
!                                    xpc(1:0,:neigvec), xpccor=xpccor(:nvar,:neigvec)            )
!
!            The shape of XPCCOR must verify:
!
!            - size(XPCCOR,1) = size(X,DIMVAR) ,
!            - size(XPCCOR,2) = size(XEIGVEC,2) .
!
!
! Further Details
! _______________
!
!   The subroutine computes the Principal Components with only one pass
!   through the data.
!
!   If unnormalized PCs are desired, use argument XSINGVAL with all values set to one,
!   however in this case, do not use argument XPCCOR.
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : zero
    use Char_Constants,    only : stat_error1, stat_error5, stat_error6
    use Utilities,         only : assert_eq, assert, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),             dimension(:)   :: xsingval
    real(stnd), intent(in),             dimension(:,:) :: x, xeigvec
    real(stnd), intent(in),  optional,  dimension(:)   :: xstd, xmean
    real(stnd), intent(out),            dimension(:,:) :: xpc
    real(stnd), intent(out), optional,  dimension(:,:) :: xpccor
!
    integer(i4b), intent(in),  optional :: dimvar
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(xeigvec,1)) :: dxi
    real(stnd), dimension(size(xeigvec,2)) :: dxpci
!
    integer(i4b) :: i, n, m, p
!
    logical(lgl) :: dim_is_one, test_mean, test_std
!
#ifdef _OPENMP
    integer(i4b) :: j
!
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_pc_eof'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    test_mean     = false
    test_std      = false
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
    n = assert_eq(  int(size(x,3-int(i)),i4b),     &
                    int(size(xpc,1),i4b),          &
                    name_proc )
!
    m = assert_eq(  int(size(x,int(i)),i4b),       &
                    int(size(xeigvec,1),i4b),      &
                    name_proc )
!
    p = assert_eq(  int(size(xeigvec,2),i4b),      &
                    int(size(xpc,2),i4b),          &
                    int(size(xsingval),i4b),       &
                    name_proc )
!
    if ( min(m,p)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    if ( any(xsingval<=zero) ) call merror( name_proc//stat_error6 )
!
    if ( present(xmean) ) then
!
        call assert( logical(size(xmean)==m,lgl),    &
                      name_proc )
!
        test_mean = true
!
    end if
!
    if ( present(xstd) ) then
!
        call assert( logical(size(xstd)==m,lgl),    &
                      name_proc )
!
        test_std = true
!
    end if
!
    if ( present(xpccor) ) then
!
        call assert( logical(size(xpccor,1)==m,lgl),    &
                     logical(size(xpccor,2)==p,lgl),    &
                      name_proc )
!
        xpccor(:m,:p) = xeigvec(:m,:p)*spread( xsingval(:p), dim=1, ncopies=m )
!
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  (n*p)>=omp_limit                    .and.      &
                  n>1_i4b                             .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)        &
!$OMP         ,PRIVATE(i,dxi,dxpci)
!
!$OMP DO SCHEDULE(STATIC) 
!
    do i = 1_i4b, n
!
        if ( dim_is_one ) then
            dxi(1_i4b:m) = x(1_i4b:m,i)
        else
            dxi(1_i4b:m) = x(i,1_i4b:m)
        end if
!
        if ( test_mean ) then
            dxi(1_i4b:m) = dxi(1_i4b:m) - xmean(1_i4b:m)
        end if
!
        if ( test_std ) then
            dxi(1_i4b:m) = dxi(1_i4b:m) / xstd(1_i4b:m)
        end if
!
        dxpci(1_i4b:p) = matmul( dxi(1_i4b:m), xeigvec(1_i4b:m,1_i4b:p) )
!
        xpc(i,1_i4b:p) = dxpci(1_i4b:p) / xsingval(1_i4b:p)
!
    end do
!
!$OMP END DO
!
!$OMP END PARALLEL
!
!
! END OF SUBROUTINE comp_pc_eof
! _____________________________
!
    end subroutine comp_pc_eof
!
!
! =========================================================================================
!                               MAXIMUM COVARIANCE ANALYSIS
! =========================================================================================
!
!
    subroutine comp_mca( x, y, first, last, xstat, ystat, xysingval, xsingvec, failure,    &
                         dimvarx, dimvary, cov, sort, maxiter, ysingvec, xysingvar, xycor )
!
! Purpose
! _______
!
!   COMP_MCA performs Maximum Covariance Analysis (MCA) or canonical covariance analysis
!   between two data matrices XX and YY.
!
!   COMP_MCA computes the singular value decomposition (SVD) of the correlation 
!   (or covariance) matrix XYCOR between two data matrices XX and YY. This SVD is written
!
!                       XYCOR = U * SIGMA * V'
!
!   where SIGMA is a m-by-n matrix which is zero except for its
!   min(m,n) diagonal elements, U is a m-by-m orthogonal matrix, and
!   V is a n-by-n orthogonal matrix.  The diagonal elements of SIGMA
!   are the singular values of XYCOR; they are real and non-negative.
!   The first min(m,n) columns of U and V are the left and right
!   singular vectors of XYCOR.
!
!   The routine returns the singular values, the left and, optionally, the right
!   singular vectors of the correlation (or covariance) matrix XYCOR between two
!   data matrices XX and YY.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:,:)
!            On entry, input submatrix containing size(X,3-DIMVARX)
!            observations on size(X,DIMVARX) variables from the "left" matrix of data XX.
!            By default, DIMVARX is equal to 1. See description of optional DIMVARX
!            argument for details.
!            If all the data are available at once, X can be the full 
!            data matrix XX.
!
!            The shape of X must verify: size(X,3-DIMVARX) = size(Y,3-DIMVARY) .
!                                         
!   Y        (INPUT) real(stnd), dimension(:,:)
!            On entry, input submatrix containing size(Y,3-DIMVARY)
!            observations on size(Y,DIMVARY) variables from the "right" matrix of data YY.
!            By default, DIMVARY is equal to 1. See description of optional DIMVARY
!            argument for details.
!            If all the data are available at once, Y can be the full 
!            data matrix YY.
!
!            The shape of Y must verify: size(Y,3-DIMVARY) = size(X,3-DIMVARX) .
!
!   FIRST    (INPUT) logical(lgl)
!            On entry, if:
!
!            - FIRST = true  the current submatrices are the first submatrices 
!              of the data matrices XX and YY.
!            - FIRST = false the current submatrices are not the first submatrices
!              of the data matrices XX and YY.
!
!   LAST     (INPUT) logical(lgl)
!            On entry, if:
!
!            - LAST = true  the current submatrices are the last submatrices 
!              of the data matrices XX and YY.
!            - LAST = false the current submatrix are not the last submatrices
!              of the data matrices XX and YY.
!
!   XSTAT    (INPUT/OUTPUT) real(stnd), dimension(:,2)
!            On entry, after the first call to COMP_MCA (e.g. when FIRST=true),
!            XSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_MCA. XSTAT should not be changed between calls to
!            COMP_MCA.
!
!            On exit, when LAST=true, XSTAT contains the following statistics
!            on all variables from the XX matrix:
!
!            - XSTAT(:,1) contains the mean values of the "left" data matrix XX.
!            - XSTAT(:,2) contains the standard-deviations of the "left" data matrix XX.
!
!            The shape of XSTAT must verify:
!
!            - size(XSTAT,1) = size(X,DIMVARX) ,
!            - size(XSTAT,2) = 2 .
!
!   YSTAT    (INPUT/OUTPUT) real(stnd), dimension(:,2)
!            On entry, after the first call to COMP_MCA (e.g. when FIRST=true),
!            YSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_MCA. YSTAT should not be changed between calls to
!            COMP_MCA.
!
!            On exit, when LAST=true, YSTAT contains the following statistics
!            on all variables from the YY matrix:
!
!            - YSTAT(:,1) contains the mean values of the "right" data matrix YY.
!            - YSTAT(:,2) contains the the standard-deviations of the "right" data matrix YY.
!
!            The shape of YSTAT must verify:
!
!            - size(YSTAT,1) = size(Y,DIMVARY) ,
!            - size(YSTAT,2) = 2 .
!
!   XYSINGVAL (INPUT/OUTPUT) real(stnd), dimension(:)
!             On entry, after the first call to COMP_MCA (e.g. when FIRST=true),
!             XYSINGVAL(1) contains count of observations from previous calls
!             to COMP_MCA. XYSINGVAL(1) should not be changed between calls to
!             COMP_MCA.
!
!             On exit, XYSINGVAL contains the singular values of the correlation
!             (or covariance) matrix XYCOR between the data matrices XX and YY.
!
!             The size of XYSINGVAL must verify: size(XYSINGVAL) = min(size(X,DIMVARX),size(Y,DIMVARY)).
!
!   XSINGVEC (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, after the first call to COMP_MCA (e.g. when FIRST=true),
!            XSINGVEC is used as workspace to accumulate quantities on previous calls
!            to COMP_MCA. XSINGVEC should not be changed between calls to
!            COMP_MCA.
!
!            On exit, when LAST=true, XSINGVEC is overwritten with the first 
!            min(size(X,DIMVARX),size(Y,DIMVARY)) columns of U, the left singular
!            vectors of the correlation (or covariance) matrix XYCOV between XX and YY.
!
!            The shape of XSINGVEC must verify:
!
!            - size(XSINGVEC,1) = size(X,DIMVARX) ,
!            - size(XSINGVEC,2) = size(Y,DIMVARY).
!
!   FAILURE  (OUTPUT) logical(lgl)
!            On exit when LAST=true:
!
!            - FAILURE = false:   indicates successful exit.
!            - FAILURE = true:    indicates that fewer than two valid observations
!              were present or that maximum accuracy was not
!              achieved when computing the SVD of the covariance
!              (or correlation) matrix between the data matrices XX and YY .
!
!            On exit when LAST=false, FAILURE is always set to false.
!          
!   DIMVARX  (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVARX is present, DIMVARX is used as follows, if:
!          
!            - DIMVARX = 1, the input submatrix X contains size(X,2) observations
!              on size(X,1) variables.
!            - DIMVARX = 2, the input submatrix X contains size(X,1) observations
!              on size(X,2) variables, respectively.
!          
!            The default is DIMVARX = 1.
!
!   DIMVARY  (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVARY is present, DIMVARY is used as follows, if:
!          
!            - DIMVARY = 1, the input submatrix Y contains size(Y,2) observations
!              on size(Y,1) variables.
!            - DIMVARY = 2, the input submatrix Y contains size(Y,1) observations
!              on size(Y,2) variables, respectively.
!          
!            The default is DIMVARY = 1.
!          
!   COV      (INPUT, OPTIONAL) logical(lgl)
!            On entry, if COV is present and COV=true, a covariance matrix between
!            the data matrices XX and YY is computed instead of a correlation matrix.
!
!            By default, a correlation matrix is computed.
!
!            COV needs to be specified only on the last call to COMP_MCA (e.g. when LAST=true).
!
!   SORT     (INPUT, OPTIONAL) character
!            Sort the singular values into ascending order if SORT = 'A' or 'a',
!            or in descending order if SORT = 'D' or 'd'.
!            The singular vectors are rearranged accordingly.
!
!            SORT needs to be specified only on the last call to COMP_MCA (e.g. when LAST=true).
!
!   MAXITER  (INPUT, OPTIONAL) integer(i4b)
!            MAXITER controls the maximum number of QR sweeps in the
!            bidiagonal SVD phase of the SVD algorithm.
!            The bidiagonal SVD algorithm of an intermediate bidiagonal 
!            form B of XYCOR fails to converge if the number of QR sweeps 
!            exceeds MAXITER * min(m,n). Convergence usually occurs in 
!            about 2 * min(m,n) QR sweeps. The default is 10.
!
!            MAXITER needs to be specified only on the last call to COMP_MCA (e.g. when LAST=true).
!
!   YSINGVEC (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!            On exit, when LAST=true and YSINGVEC is present, YSINGVEC contains the first
!            min(size(X,DIMVARX),size(Y,DIMVARY))  columns of V, the right singular vectors
!            of the correlation (or covariance) matrix XYCOR between XX and YY.
!
!            YSINGVEC needs to be specified only on the last call to COMP_MCA (e.g. when LAST=true).
!
!            The shape of YSINGVEC must verify:
!
!            - size(YSINGVEC,1) = size(Y,DIMVARY) ,
!            - size(YSINGVEC,2) = min( size(X,DIMVARX) ,size(Y,DIMVARY) ) .
!
!   XYSINGVAR (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!             On exit, when LAST=true and XYSINGVAR is present, XYSINGVAR contains
!             percentages of total squared covariance associated with the left and right
!             singular vectors in order of the singular values stored in XYSINGVAL.
!
!             XYSINGVAR needs to be specified only on the last call to COMP_MCA (e.g. when LAST=true).
!
!             The size of XYSINGVAR must verify: size(XYSINGVAR) = min( size(X,DIMVARX),size(Y,DIMVARY) ).
!          
!   XYCOR    (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!            On exit, when LAST=true and XYCOR is present, XYCOR contains the correlation
!            or variance-covariance matrix between data arrays XX and YY, as controlled 
!            by the COV argument.
!
!            XYCOR needs to be specified only on the last call to COMP_MCA (e.g. when LAST=true).
!          
!            The shape of XYCOR must verify:
!          
!            - size(XYCOR,1) = size(X,DIMVARX) ,
!            - size(XYCOR,2) = size(Y,DIMVARY) .
!
!          
! Further Details
! _______________
!
!   The subroutine computes the basic univariate statistics and the correlation (or covariance)
!   matrix with only one pass through the data.
!
!   If fewer than two valid observations were present, the statistics XSTAT, YSTAT, XYSINGVAL
!   XSINGVEC, YSINGVEC, XYSINGVAR and XYCOR are set to Nan code.
!
!   This subroutine may be used in a call with no observations (e.g. size(X,3-DIMVARX) = 0)
!   in order to finish the computations with LAST=true when the total number of observations
!   is unknown at the beginning of the computations.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true, false
    use Reals_Constants,   only : zero, one, c100
    use Num_Constants,     only : nan, lamch
    use Char_Constants,    only : stat_error1, stat_error5
    use Utilities,         only : assert_eq, assert, merror
    use SVD_Procedures,    only : bd_cmp, ortho_gen_bd, ortho_gen_q_bd, bd_svd
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),    dimension(:,:)           :: x, y
    real(stnd), intent(inout), dimension(:,:)           :: xstat, ystat, xsingvec
    real(stnd), intent(inout), dimension(:)             :: xysingval
    real(stnd), intent(out),   dimension(:),   optional :: xysingvar
    real(stnd), intent(out),   dimension(:,:), optional :: ysingvec, xycor
!
    integer(i4b), intent(in), optional :: dimvarx, dimvary, maxiter
!
    logical(lgl), intent(in)            :: first, last
    logical(lgl), intent(in),  optional :: cov
    logical(lgl), intent(out)           :: failure
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                              :: a, b, xyn, xynm1
    real(stnd), dimension(size(xsingvec,1)) :: dxi, tmp
    real(stnd), dimension(size(xsingvec,2)) :: dyi
!
    integer(i4b) :: i, j, n, m, p
!
    logical(lgl) :: dimx_is_one, dimy_is_one, comp_cor
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_mca'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvarx) ) then
!
        if ( dimvarx==1_i4b .or. dimvarx==2_i4b ) then
            i = dimvarx
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dimx_is_one = i==1_i4b
!
    if ( present(dimvary) ) then
!
        if ( dimvary==1_i4b .or. dimvary==2_i4b ) then
            j = dimvary
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        j = 1_i4b
!
    end if
!
    dimy_is_one = j==1_i4b
!
    n = assert_eq(  int(size(x,3-int(i)),i4b),  &
                    int(size(y,3-int(j)),i4b),  &
                    name_proc )
!
    m = assert_eq(  int(size(x,int(i)),i4b),      &
                    int(size(xstat,1),i4b),       &
                    int(size(xsingvec,1),i4b),    &
                    name_proc )
!
    p = assert_eq(  int(size(y,int(j)),i4b),      &
                    int(size(ystat,1),i4b),       &
                    int(size(xsingvec,2),i4b),    &
                    name_proc )
!
    if ( m<=0_i4b .or. p<=0_i4b ) call merror( name_proc//stat_error1 )
!
    i = min( m, p )
!
    call assert( logical(size(xstat,2)==2,lgl),           &
                 logical(size(ystat,2)==2,lgl),           &
                 logical(size(xysingval)==int(i),lgl),    &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBMATRICES ARE THE FIRST.
!
!       INITIALIZE xstat, ystat, xsingvec AND xyn .
!
        xstat(:m,1_i4b:2_i4b) = zero       
        ystat(:p,1_i4b:2_i4b) = zero
        xsingvec(:m,:p)       = zero
        xysingval(:i)         = zero
!
    end if
!
    xyn = xysingval(1_i4b)
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  (m*p)>=omp_limit                    .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)      &
!$OMP         ,PRIVATE(i,j)
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS AND CORRECTED SUMS OF SQUARES AND CROSS-PRODUCTS, 
!       HANDLING THE DATA ONCE.
!
!$OMP SINGLE
!
        xynm1 = xyn
        xyn = xynm1 + one
        a = one/xyn
        b = xynm1*a
!
        if ( dimx_is_one ) then
            dxi(:m) = x(:m,i) - xstat(:m,1_i4b)
        else
            dxi(:m) = x(i,:m) - xstat(:m,1_i4b)
        end if
!
        tmp(:m) = b*dxi(:m)
!
        xstat(:m,1_i4b) = xstat(:m,1_i4b) + a*dxi(:m)
        xstat(:m,2_i4b) = xstat(:m,2_i4b) + tmp(:m)*dxi(:m)
!
        if ( dimy_is_one ) then
            dyi(:p) = y(:p,i) - ystat(:p,1_i4b)
        else
            dyi(:p) = y(i,:p) - ystat(:p,1_i4b)
        end if
!
        ystat(:p,1_i4b) = ystat(:p,1_i4b) + a*dyi(:p)
        ystat(:p,2_i4b) = ystat(:p,2_i4b) + b*( dyi(:p)*dyi(:p) )
!
!$OMP END SINGLE
!
!$OMP DO SCHEDULE(STATIC) 
!
        do j = 1_i4b, p
!
#ifdef _BLAS
            call axpy( m, dyi(j), tmp(:m), 1_i4b, xsingvec(:m,j), 1_i4b )
#else
            xsingvec(:m,j) = xsingvec(:m,j) + dyi(j)*tmp(:m)
#endif
!
        end do
!
!$OMP END DO
!
    end do
!
!$OMP END PARALLEL
!
    if ( last ) then
!
!       THE CURRENT SUBMATRICES ARE THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
        n = min( m, p )
!
!       TEST OPTIONAL ARGUMENTS.
!
        comp_cor = true
!
        if ( present(cov) ) then
            comp_cor = .not. cov
        end if
!
        if ( present(ysingvec) ) then
            call assert( logical(size(ysingvec, 1)==int(p) ,lgl),    &
                         logical(size(ysingvec, 2)==int(n) ,lgl),    &
                         name_proc )
        end if
!
        if ( present(xysingvar) ) then
            call assert( logical(size(xysingvar)==int(n) ,lgl),     &
                         name_proc )
        end if
!
        if ( present(xycor) ) then
            call assert( logical(size(xycor, 1)==int(m) ,lgl),    &
                         logical(size(xycor, 2)==int(p) ,lgl),    &
                         name_proc )
        end if
!
        if ( xyn>one ) then
!
            a = one/xyn
            b = sqrt( a )
!
!           COMPUTE STANDARD-DEVIATIONS OF XX.
!
            dxi(:m) = sqrt( xstat(:m,2_i4b) )
            xstat(:m,2_i4b) = dxi(:m)*b
!
!           COMPUTE STANDARD-DEVIATIONS OF YY.
!
            dyi(:p) = sqrt( ystat(:p,2_i4b) )
            ystat(:p,2_i4b) = dyi(:p)*b
!
            if ( comp_cor ) then
!
                b = lamch( 's' )
!
!$OMP PARALLEL IF(test_par)        &
!$OMP         ,PRIVATE(j,tmp)
!
!               COMPUTE CORRELATION MATRIX BETWEEN XX AND YY.
!
!$OMP DO SCHEDULE(STATIC) 
!
                do j = 1_i4b, p
!
                    tmp(:m) = dxi(:m)*dyi(j)
                    xsingvec(:m,j) = xsingvec(:m,j)/( tmp(:m) + b )
!
                end do
!
!$OMP END DO
!
!$OMP END PARALLEL
!
            else
!
!               COMPUTE COVARIANCE MATRIX BETWEEN XX AND YY.
!
                xsingvec(:m,:p) = xsingvec(:m,:p)*a
!
            end if
!
            if ( present(xycor) ) then
!
!               SAVE THE COVARIANCE OR CORRELATION MATRIX XYCOR, IF REQUIRED.
!
                xycor(:m,:p) = xsingvec(:m,:p)
!
            end if
!
!           COMPUTE SVD OF THE COVARIANCE OR CORRELATION MATRIX XYCOR.
!
!           BIDIAGONALIZE COVARIANCE OR CORRELATION MATRIX XYCOR.
!
            call bd_cmp( xsingvec(:m,:p), xysingval(:n), tmp(:n), dxi(:n), dyi(:n) )
!
            if ( present(ysingvec) ) then
!
!               GENERATE RIGHT VECTORS BIDIAGONALIZING XYCOV IN ysingvec .
!               GENERATE LEFT VECTORS BIDIAGONALIZING XYCOV IN xsingvec .
!
                call ortho_gen_bd( xsingvec(:m,:p), dxi(:n), dyi(:n), ysingvec(:p,:n) )
!
!               PERFORM BIDIAGONAL QR ITERATION, COMPUTING LEFT
!               SINGULAR VECTORS OF XYCOV IN xsingvec AND COMPUTING RIGHT
!               SINGULAR VECTORS OF XYCOV IN ysingvec .
!
                call bd_svd( merge(true,false,m>=p), xysingval(:n), tmp(:n), failure, xsingvec(:m,:n), ysingvec(:p,:n),     &
                             sort=sort, maxiter=maxiter )
!
            else
!
!               GENERATE LEFT VECTORS BIDIAGONALIZING XYCOV IN xsingvec .
!
                call ortho_gen_q_bd( xsingvec(:m,:p), dxi(:n) )
!
!               PERFORM BIDIAGONAL QR ITERATION, COMPUTING LEFT
!               SINGULAR VECTORS OF XYCOV IN xsingvec.
!
                call bd_svd( merge(true,false,m>=p), xysingval(:n), tmp(:n), failure, xsingvec(:m,:n),     &
                             sort=sort, maxiter=maxiter )
!
            end if
!
!           SET UNUSED PART OF xsingvec TO ZERO.
!
            xsingvec(:m,n+1_i4b:p) = zero
!
            if ( present(xysingvar) ) then
!
!               COMPUTE PERCENTAGES OF SQUARED COVARIANCE, IF REQUIRED.
!
                xysingvar(1_i4b:n) = xysingval(1_i4b:n)**2
                a = c100 / sum( xysingvar(1_i4b:n) )
                xysingvar(1_i4b:n) = xysingvar(1_i4b:n) * a
!
            end if
!
        else
!
!           SET FAILURE INDICATOR.
!
            failure = true
!
            b = nan()
!
!           FEWER THAN TWO VALID OBSERVATIONS, SET THE PERTINENT STATISTICS TO NaN CODE.
!
            if ( xyn<one ) then
!
                xstat(:m,1_i4b:2_i4b) = b
                ystat(:p,1_i4b:2_i4b) = b
!
            end if
!
            xysingval(:n)    = b
            xsingvec(:m,:p)  = b
!
            if ( present(ysingvec) ) then
                ysingvec(:p,:n) = b
            end if
!
            if ( present(xysingvar) ) then
                xysingvar(:n) = b
            end if
!
            if ( present(xycor) ) then
                xycor(:m,:p) = b
            end if
!
        end if
!
    else
!
        failure = false
!
!       SAVE COUNT OF OBSERVATIONS FOR THE NEXT CALL.
!
        xysingval(1_i4b) = xyn
!
    end if
!
!
! END OF SUBROUTINE comp_mca
! __________________________
!
    end subroutine comp_mca
!
! =========================================================================================
!
    subroutine comp_mca2( x, y, first, last, xstat, ystat, xysingval, xycor, failure,          &
                          dimvarx, dimvary, cov, savecor, maxiter, ortho, xysingvar, xysingvec )
!
! Purpose
! _______
!
!   COMP_MCA2 performs Maximum Covariance Analysis (MCA) or canonical covariance analysis
!   between two data matrices XX and YY.
!
!   COMP_MCA2 computes a partial singular value decomposition (SVD) of the correlation 
!   (or covariance) matrix XYCOR between two data matrices XX and YY. This partial SVD is
!   written
!
!                       U(:m,:k) * SIGMA(:k,:k) * V(:n,:k)'
!
!   where SIGMA is a k-by-k matrix which is zero except for its
!   k diagonal elements, U is a m-by-k orthogonal matrix, and
!   V is a n-by-k orthogonal matrix.  The diagonal elements of SIGMA
!   are the first k singular values of XYCOR; they are real and non-negative.
!   The k columns of U and V are the first k left and right singular 
!   vectors of XYCOR.
!
!   COMP_MCA2 computes all the singular values, and, optionally, selected left and right
!   singular vectors (by inverse iteration), of the covariance (or correlation matrix) XYCOR
!   between two data matrices XX and YY.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:,:)
!            On entry, input submatrix containing size(X,3-DIMVARX)
!            observations on size(X,DIMVARX) variables from the "left" matrix of data XX.
!            By default, DIMVARX is equal to 1. See description of optional DIMVARX
!            argument for details.
!            If all the data are available at once, X can be the full 
!            data matrix XX.
!
!            The shape of X must verify: size(X,3-DIMVARX)  = size(Y,3-DIMVARY) .
!                                         
!   Y        (INPUT) real(stnd), dimension(:,:)
!            On entry, input submatrix containing size(Y,3-DIMVARY)
!            observations on size(Y,DIMVARY) variables from the "right" matrix of data YY.
!            By default, DIMVARY is equal to 1. See description of optional DIMVARY
!            argument for details.
!            If all the data are available at once, Y can be the full 
!            data matrix YY.
!
!            The shape of Y must verify: size(Y,3-DIMVARY)  = size(X,3-DIMVARX) .
!
!   FIRST    (INPUT) logical(lgl)
!            On entry, if:
!
!            - FIRST = true  the current submatrices are the first submatrices 
!              of the data matrices XX and YY.
!            - FIRST = false the current submatrices are not the first submatrices
!              of the data matrices XX and YY.
!
!   LAST     (INPUT) logical(lgl)
!            On entry, if:
!
!            - LAST = true  the current submatrices are the last submatrices 
!              of the data matrices XX and YY.
!            - LAST = false the current submatrix are not the last submatrices
!              of the data matrices XX and YY.
!
!   XSTAT    (INPUT/OUTPUT) real(stnd), dimension(:,2)
!            On entry, after the first call to COMP_MCA (e.g. when FIRST=true),
!            XSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_MCA. XSTAT should not be changed between calls to
!            COMP_MCA.
!
!            On exit, when LAST=true, XSTAT contains the following statistics
!            on all variables from the XX matrix:
!
!            - XSTAT(:,1) contains the mean values of the "left" data matrix XX.
!            - XSTAT(:,2) contains the standard-deviations of the "left" data matrix XX.
!
!            The shape of XSTAT must verify:
!
!            - size(XSTAT,1) = size(X,DIMVARX) ,
!            - size(XSTAT,2) = 2 .
!
!   YSTAT    (INPUT/OUTPUT) real(stnd), dimension(:,2)
!            On entry, after the first call to COMP_MCA (e.g. when FIRST=true),
!            YSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_MCA. YSTAT should not be changed between calls to
!            COMP_MCA.
!
!            On exit, when LAST=true, YSTAT contains the following statistics
!            on all variables from the YY matrix:
!
!            - YSTAT(:,1) contains the mean values of the "right" data matrix YY.
!            - YSTAT(:,2) contains the the standard-deviations of the "right" data matrix YY.
!
!            The shape of YSTAT must verify:
!
!            - size(YSTAT,1) = size(Y,DIMVARY) ,
!            - size(YSTAT,2) = 2 .
!
!   XYSINGVAL (INPUT/OUTPUT) real(stnd), dimension(:)
!             On entry, after the first call to COMP_MCA2 (e.g. when FIRST=true),
!             XYSINGVAL(1) contains count of observations from previous calls
!             to COMP_MCA2. XYSINGVAL(1) should not be changed between calls to
!             COMP_MCA2.
!
!             On exit, XYSINGVAL contains the singular values of the correlation
!             (or covariance) matrix XYCOR between the data matrices XX and YY.
!
!             The size of XYSINGVAL must verify: size(XYSINGVAL) = min(size(X,DIMVARX),size(Y,DIMVARY)).
!
!   XYCOR    (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, after the first call to COMP_MCA2 (e.g. when FIRST=true),
!            XYCOR is used as workspace to accumulate quantities on previous calls
!            to COMP_MCA2. XYCOR should not be changed between calls to
!            COMP_MCA2.
!
!            On exit, when LAST=true and SAVECOR=true, XYCOR contains the correlation
!            or variance-covariance matrix as controlled by the COV argument. In this
!            case XYCOR(i,j) contains the correlation (or covariance) coefficient
!            between XX(i,:) and YY(j,:) ( XX(:,i) and YY(:,j) if DIMVARX=2 and
!            DIMVARY=2 ).
!
!            If SAVECOR=false, the correlation (or covariance) matrix is not saved on exit.
!            In this case, XYCOR does not contain useful information.
!
!            The shape of XYCOR must verify:
!
!            - size(XYCOR,1) = size(X,DIMVARX) ,
!            - size(XYCOR,2) = size(Y,DIMVARY) .
!
!   FAILURE  (OUTPUT) logical(lgl)
!            On exit when LAST=true:
!
!            - FAILURE = false:   indicates successful exit.
!            - FAILURE = true:    indicates that fewer than two valid observations
!              were present or that maximum accuracy was not
!              achieved when computing the singular values or that
!              some singular vectors failed to converge with MAXITER
!              inverse iterations.
!
!            On exit when LAST=false, FAILURE is always set to false.
!          
!   DIMVARX  (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVARX is present, DIMVARX is used as follows, if:
!
!            - DIMVARX = 1, the input submatrix X contains size(X,2) observations
!              on size(X,1) variables.
!            - DIMVARX = 2, the input submatrix X contains size(X,1) observations
!              on size(X,2) variables, respectively.
!
!            The default is DIMVARX = 1.
!
!   DIMVARY  (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVARY is present, DIMVARY is used as follows, if:
!
!            - DIMVARY = 1, the input submatrix Y contains size(Y,2) observations
!              on size(Y,1) variables.
!            - DIMVARY = 2, the input submatrix Y contains size(Y,1) observations
!              on size(Y,2) variables, respectively.
!
!            The default is DIMVARY = 1.
!          
!   COV      (INPUT, OPTIONAL) logical(lgl)
!            On entry, if COV is present and COV=true, a covariance matrix between
!            the data matrices XX and YY is computed instead of a correlation matrix.
!
!            By default, a correlation matrix is computed.
!
!            COV needs to be specified only on the last call to COMP_MCA2 (e.g. when LAST=true).
!
!   SAVECOR  (INPUT, OPTIONAL) logical(lgl)
!            On exit, when argument SAVECOR is present and LAST=true, SAVECOR is used as follows, if:
!
!            - SAVECOR= true, the correlation (or covariance) matrix is saved
!              in argument XYCOR.
!            - SAVECOR= false, the correlation (or covariance) matrix is destroyed.
!
!            By default, the correlation (or covariance) matrix is destroyed.
!
!            SAVECOR needs to be specified only on the last call to COMP_MCA2 (e.g. when LAST=true).
!
!   MAXITER  (INPUT, OPTIONAL) integer(i4b)
!            The number of inverse iterations performed in the subroutine for computing
!            the singular vectors. By default, 2 inverse iterations are performed 
!            for all the singular vectors. This optional argument is used only if
!            the XYSINGVEC argument is present.
!
!            MAXITER needs to be specified only on the last call to COMP_MCA2 (e.g. when LAST=true).
!
!   ORTHO    (INPUT, OPTIONAL) logical(lgl)
!            If ORTHO=true the computed singular vectors are orthogonalized by the
!            Modified Gram-Schmidt algorithm. This optional argument is used only if
!            the XYSINGVEC argument is present.
!            The default is FALSE.
!
!            ORTHO needs to be specified only on the last call to COMP_MCA2 (e.g. when LAST=true).
!
!   XYSINGVAR (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!             On exit, when LAST=true and XYSINGVAR is present, XYSINGVAR contains
!             percentages of total squared covariance associated with the left and right
!             singular vectors in order of the singular values stored in XYSINGVAL.
!
!             XYSINGVAR needs to be specified only on the last call to COMP_MCA2 (e.g. when LAST=true).
!
!             The size of XYSINGVAR must verify: size(XYSINGVAR) = min( size(X,DIMVARX),size(Y,DIMVARY) ).
!          
!   XYSINGVEC (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!             On exit, when LAST=true and XYSINGVEC is present, XYSINGVEC contains the
!             first columns of U and V, the first k left and right singular vectors
!             of the correlation (or covariance) matrix XYCOR between XX and YY.
!          
!             The first k left singular vectors are stored in XYSINGVEC(:size(X,DIMVARX),:)
!             The first k right singular vectors are stored in XYSINGVEC(size(X,DIMVARX)+1:,:)
!
!             XYSINGVEC needs to be specified only on the last call to COMP_MCA2 (e.g. when LAST=true).
!
!             The shape of XYSINGVEC must verify:
!
!             - size(XYSINGVEC,1)  = size(X,DIMVARX) + size(Y,DIMVARY) ,
!             - size(XYSINGVEC,2) <= min( size(X,DIMVARX) ,size(Y,DIMVARY) ).
!
!          
! Further Details
! _______________
!
!   The subroutine computes the basic univariate statistics and the correlation (or covariance)
!   matrix with only one pass through the data.
!
!   If fewer than two valid observations were present, the statistics XSTAT, YSTAT, XYSINGVAL
!   XYCOR , XYSINGVAR and XYSINGVEC are set to Nan code.
!
!   This subroutine may be used in a call with no observations (e.g. size(X,3-DIMVARX) = 0)
!   in order to finish the computations with LAST=true when the total number of observations
!   is unknown at the beginning of the computations.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true, false
    use Reals_Constants,   only : zero, one, c100
    use Num_Constants,     only : nan, lamch
    use Char_Constants,    only : stat_error1, stat_error5, allocate_error
    use Utilities,         only : assert_eq, assert, merror
    use SVD_Procedures,    only : bd_inviter2, svd_cmp
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),    dimension(:,:)           :: x, y
    real(stnd), intent(inout), dimension(:,:)           :: xstat, ystat, xycor
    real(stnd), intent(inout), dimension(:)             :: xysingval
    real(stnd), intent(out),   dimension(:),   optional :: xysingvar
    real(stnd), intent(out),   dimension(:,:), optional :: xysingvec 
!
    integer(i4b), intent(in), optional :: dimvarx, dimvary, maxiter
!
    logical(lgl), intent(in)            :: first, last
    logical(lgl), intent(in),  optional :: cov, savecor, ortho
    logical(lgl), intent(out)           :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                        :: a, b, xyn, xynm1
    real(stnd),              dimension(size(xycor,1)) :: dxi, tmp
    real(stnd),              dimension(size(xycor,2)) :: dyi
    real(stnd), allocatable, dimension(:)             :: d
    real(stnd), allocatable, dimension(:,:)           :: work
!
    integer(i4b) :: i, j, k, n, m, p, mp
    integer      :: iok
!
    logical(lgl) :: dimx_is_one, dimy_is_one, comp_cor, save_cor
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_mca2'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvarx) ) then
!
        if ( dimvarx==1_i4b .or. dimvarx==2_i4b ) then
            i = dimvarx
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dimx_is_one = i==1_i4b
!
    if ( present(dimvary) ) then
!
        if ( dimvary==1_i4b .or. dimvary==2_i4b ) then
            j = dimvary
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        j = 1_i4b
!
    end if
!
    dimy_is_one = j==1_i4b
!
    n = assert_eq(  int(size(x,3-int(i)),i4b),  &
                    int(size(y,3-int(j)),i4b),  &
                    name_proc )
!
    m = assert_eq(  int(size(x,int(i)),i4b),      &
                    int(size(xstat,1),i4b),       &
                    int(size(xycor,1),i4b),       &
                    name_proc )
!
    p = assert_eq(  int(size(y,int(j)),i4b),      &
                    int(size(ystat,1),i4b),       &
                    int(size(xycor,2),i4b),       &
                    name_proc )
!
    if ( m<=0_i4b .or. p<=0_i4b ) call merror( name_proc//stat_error1 )
!
    i = min( m, p )
!
    call assert( logical(size(xstat,2)==2,lgl),           &
                 logical(size(ystat,2)==2,lgl),           &
                 logical(size(xysingval)==int(i),lgl),    &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBMATRICES ARE THE FIRST.
!
!       INITIALIZE xstat, ystat, xycor AND xyn .
!
        xstat(:m,1_i4b:2_i4b) = zero       
        ystat(:p,1_i4b:2_i4b) = zero
        xycor(:m,:p)          = zero
        xysingval(:i)         = zero
!
    end if
!
    xyn = xysingval(1_i4b)
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  (m*p)>=omp_limit                    .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)      &
!$OMP         ,PRIVATE(i,j)
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS AND CORRECTED SUMS OF SQUARES AND CROSS-PRODUCTS, 
!       HANDLING THE DATA ONCE.
!
!$OMP SINGLE
!
        xynm1 = xyn
        xyn = xynm1 + one
        a = one/xyn
        b = xynm1*a
!
        if ( dimx_is_one ) then
            dxi(:m) = x(:m,i) - xstat(:m,1_i4b)
        else
            dxi(:m) = x(i,:m) - xstat(:m,1_i4b)
        end if
!
        tmp(:m) = b*dxi(:m)
!
        xstat(:m,1_i4b) = xstat(:m,1_i4b) + a*dxi(:m)
        xstat(:m,2_i4b) = xstat(:m,2_i4b) + tmp(:m)*dxi(:m)
!
        if ( dimy_is_one ) then
            dyi(:p) = y(:p,i) - ystat(:p,1_i4b)
        else
            dyi(:p) = y(i,:p) - ystat(:p,1_i4b)
        end if
!
        ystat(:p,1_i4b) = ystat(:p,1_i4b) + a*dyi(:p)
        ystat(:p,2_i4b) = ystat(:p,2_i4b) + b*( dyi(:p)*dyi(:p) )
!
!$OMP END SINGLE
!
!$OMP DO SCHEDULE(STATIC) 
!
        do j = 1_i4b, p
!
#ifdef _BLAS
            call axpy( m, dyi(j), tmp(:m), 1_i4b, xycor(:m,j), 1_i4b )
#else
            xycor(:m,j) = xycor(:m,j) + dyi(j)*tmp(:m)
#endif
!
        end do
!
!$OMP END DO
!
    end do
!
!$OMP END PARALLEL
!
    if ( last ) then
!
!       THE CURRENT SUBMATRICES ARE THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
        n = min( m, p )
!
!       TEST OPTIONAL ARGUMENTS.
!
        comp_cor = true
        if ( present(cov) ) then
            comp_cor = .not. cov
        end if
!
        if ( present(savecor) ) then
           save_cor = savecor
        else
           save_cor = false
        end if
!
        if ( present(xysingvec) ) then
!
            mp = m + p 
            k = size( xysingvec, 2 )
!
            call assert( logical(size(xysingvec, 1)==int(mp) ,lgl),   &
                         logical(k<=n ,lgl),                          &
                         name_proc )
!
        end if
!
        if ( present(xysingvar) ) then
            call assert( logical(size(xysingvar)==int(n) ,lgl),     &
                         name_proc )
        end if
!
        if ( xyn>one ) then
!
            a = one/xyn
            b = sqrt( a )
!
!           COMPUTE STANDARD-DEVIATIONS OF XX.
!
            dxi(:m) = sqrt( xstat(:m,2_i4b) )
            xstat(:m,2_i4b) = dxi(:m)*b
!
!           COMPUTE STANDARD-DEVIATIONS OF YY.
!
            dyi(:p) = sqrt( ystat(:p,2_i4b) )
            ystat(:p,2_i4b) = dyi(:p)*b
!
            if ( comp_cor ) then
!
                b = lamch( 's' )
!
!$OMP PARALLEL IF(test_par)        &
!$OMP         ,PRIVATE(j,tmp)
!
!               COMPUTE CORRELATION MATRIX BETWEEN XX AND YY.
!
!$OMP DO SCHEDULE(STATIC) 
!
                do j = 1_i4b, p
!
                    tmp(:m) = dxi(:m)*dyi(j)
                    xycor(:m,j) = xycor(:m,j)/( tmp(:m) + b )
!
                end do
!
!$OMP END DO
!
!$OMP END PARALLEL
!
            else
!
!               COMPUTE COVARIANCE MATRIX BETWEEN XX AND YY.
!
                xycor(:m,:p) = xycor(:m,:p)*a
!
            end if
!
            if ( save_cor ) then
!
!               ALLOCATE NEEDED WORKSPACE.
!
                allocate( work(m,p), stat=iok )
!
                if ( iok/=0 ) then
                    call merror( name_proc//allocate_error )
                end if
!
!               MAKE A COPY OF THE CORRELATION MATRIX.
!
                work(1_i4b:m,1_i4b:p) = xycor(1_i4b:m,1_i4b:p)
!
                if ( present(xysingvec) ) then
!
!                   ALLOCATE NEEDED WORKSPACE.
!
                    allocate( d(n), stat=iok )
!
                    if ( iok/=0 ) then
                        call merror( name_proc//allocate_error )
                    end if
!
!                   COMPUTE ALL SINGULAR VALUES AND SAVE THE INTERMEDIATE 
!                   BIDIAGONAL MATRIX .
!
                    call svd_cmp( work(:m,:p), xysingval(:n), failure, sort='d',    &
                                  d=d(:n), e=tmp(:n), tauq=dxi(:n), taup=dyi(:n)  )
!
!                   COMPUTE THE FIRST k SINGULAR VECTORS OF THE CORRELATION MATRIX
!                   BY maxiter INVERSE ITERATIONS ON THE INTERMEDIATE BIDIAGONAL
!                   MATRIX AND BACK-TRANSFORMATION.
!
                    call bd_inviter2( work(:m,:p), dxi(:n), dyi(:n), d(:n), tmp(:n), xysingval(:k),    &
                                      xysingvec(:m,:k), xysingvec(m+1_i4b:,:k), dimx_is_one,           &
                                      maxiter=maxiter, ortho=ortho )
!
                    failure = failure .or. dimx_is_one
!
!                   DEALLOCATE WORKSPACE.
!
                    deallocate( d )
!
                else
!
!                   COMPUTE ALL SINGULAR VALUES. 
!
                    call svd_cmp( work(:m,:p), xysingval(:n), failure, sort='d' )
!
                end if
!
!               DEALLOCATE WORKSPACE.
!
                deallocate( work )
!
            else
!
                if ( present(xysingvec) ) then
!
!                   ALLOCATE NEEDED WORKSPACE.
!
                    allocate( d(n), stat=iok )
!
                    if ( iok/=0 ) then
                        call merror( name_proc//allocate_error )
                    end if
!
!                   COMPUTE ALL SINGULAR VALUES AND SAVE THE INTERMEDIATE 
!                   BIDIAGONAL MATRIX.
!
                    call svd_cmp( xycor(:m,:p), xysingval(:n), failure, sort='d', &
                                  d=d(:n), e=tmp(:n), tauq=dxi(:n), taup=dyi(:n)  )
!
!                   COMPUTE THE FIRST k SINGULAR VECTORS OF THE CORRELATION MATRIX
!                   BY maxiter INVERSE ITERATIONS ON THE INTERMEDIATE BIDIAGONAL
!                   MATRIX AND BACK-TRANSFORMATION.
!
                    call bd_inviter2( xycor(:m,:p), dxi(:n), dyi(:n), d(:n), tmp(:n), xysingval(:k),   &
                                      xysingvec(:m,:k), xysingvec(m+1_i4b:,:k), dimx_is_one,           &
                                      maxiter=maxiter, ortho=ortho )
!
                    failure = failure .or. dimx_is_one
!
!                   DEALLOCATE WORKSPACE.
!
                    deallocate( d )
!
                else
!
!                   COMPUTE ALL SINGULAR VALUES. 
!
                    call svd_cmp( xycor(:m,:p), xysingval(:n), failure, sort='d' )
!
                end if
!
            end if
!
            if ( present(xysingvar) ) then
!
!               COMPUTE PERCENTAGES OF SQUARED COVARIANCE, IF REQUIRED.
!
                xysingvar(1_i4b:n) = xysingval(1_i4b:n)**2
                a = c100 / sum( xysingvar(1_i4b:n) )
                xysingvar(1_i4b:n) = xysingvar(1_i4b:n) * a
!
            end if
!
        else
!
!           SET FAILURE INDICATOR.
!
            failure = true
!
            b = nan()
!
!           FEWER THAN TWO VALID OBSERVATIONS, SET THE PERTINENT STATISTICS TO NaN CODE.
!
            if ( xyn<one ) then
                xstat(:m,1_i4b:2_i4b) = b
                ystat(:p,1_i4b:2_i4b) = b
            end if
!
            xysingval(:n) = b
!
            xycor(:m,:p) = b
!
            if ( present(xysingvec) ) then
                xysingvec(:mp,:k) = b
            end if
!
            if ( present(xysingvar) ) then
                xysingvar(:n) = b
            end if
!
        end if
!
    else
!
        failure = false
!
!       SAVE COUNT OF OBSERVATIONS FOR THE NEXT CALL.
!
        xysingval(1_i4b) = xyn
!
    end if
!
!
! END OF SUBROUTINE comp_mca2
! ___________________________
!
    end subroutine comp_mca2
!
! =========================================================================================
!
    subroutine comp_mca_miss( x, y, first, last, xstat, ystat, xycor, xymiss, failure,             &
                              dimvarx, dimvary, cov, sort, maxiter, xysingval, xysingvar, ysingvec )
!
! Purpose
! _______
!
!   COMP_MCA_MISS performs Maximum Covariance Analysis (MCA) or canonical covariance analysis
!   between two data matrices XX and YY possibly containing missing values.
!
!   COMP_MCA_MISS computes the singular value decomposition (SVD) of the correlation 
!   (or covariance) matrix XYCOR between two data matrices XX and YY. This SVD is written
!
!                       XYCOR = U * SIGMA * V'
!
!   where SIGMA is a m-by-n matrix which is zero except for its
!   min(m,n) diagonal elements, U is a m-by-m orthogonal matrix, and
!   V is a n-by-n orthogonal matrix.  The diagonal elements of SIGMA
!   are the singular values of XYCOR; they are real and non-negative.
!   The first min(m,n) columns of U and V are the left and right
!   singular vectors of XYCOR.
!
!   The routine returns the singular values, the left and, optionally, the right
!   singular vectors of the correlation (or covariance) matrix XYCOR between two
!   data matrices XX and YY.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:,:)
!            On entry, input submatrix containing size(X,3-DIMVARX)
!            observations on size(X,DIMVARX) variables from the "left" matrix of data XX.
!            By default, DIMVARX is equal to 1. See description of optional DIMVARX
!            argument for details.
!            If all the data are available at once, X can be the full 
!            data matrix XX.
!
!            The shape of X must verify: size(X,3-DIMVARX)  = size(Y,3-DIMVARY) .
!                                         
!   Y        (INPUT) real(stnd), dimension(:,:)
!            On entry, input submatrix containing size(Y,3-DIMVARY)
!            observations on size(Y,DIMVARY) variables from the "right" matrix of data YY.
!            By default, DIMVARY is equal to 1. See description of optional DIMVARY
!            argument for details.
!            If all the data are available at once, Y can be the full 
!            data matrix YY.
!
!            The shape of Y must verify: size(Y,3-DIMVARY)  = size(X,3-DIMVARX) .
!
!   FIRST    (INPUT) logical(lgl)
!            On entry, if:
!
!            - FIRST = true  the current submatrices are the first submatrices 
!              of the data matrices XX and YY.
!            - FIRST = false the current submatrices are not the first submatrices
!              of the data matrices XX and YY.
!
!   LAST     (INPUT) logical(lgl)
!            On entry, if:
!
!            - LAST = true  the current submatrices are the last submatrices 
!              of the data matrices XX and YY.
!            - LAST = false the current submatrix are not the last submatrices
!              of the data matrices XX and YY.
!
!   XSTAT    (INPUT/OUTPUT) real(stnd), dimension(:,4)
!            On entry, after the first call to COMP_MCA_MISS (e.g. when FIRST=true),
!            XSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_MCA_MISS. XSTAT should not be changed between calls to
!            COMP_MCA_MISS.
!
!            On exit, when LAST=true, XSTAT contains the following statistics
!            on all variables from the XX matrix:
!
!            - XSTAT(:,1) contains the mean values of the "left" data matrix XX.
!            - XSTAT(:,2) contains the standard-deviations of the "left" data matrix XX.
!            - XSTAT(:,3) contains the the numbers of non-missing observations
!              in the "left" data matrix XX.
!            - XSTAT(:,4) is used as workspace.
!
!            The shape of XSTAT must verify:
!
!            - size(XSTAT,1) = size(X,DIMVARX) ,
!            - size(XSTAT,2) = 4 .
!
!   YSTAT    (INPUT/OUTPUT) real(stnd), dimension(:,4)
!            On entry, after the first call to COMP_MCA_MISS (e.g. when FIRST=true),
!            YSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_MCA_MISS. YSTAT should not be changed between calls to
!            COMP_MCA_MISS.
!
!            On exit, when LAST=true, YSTAT contains the following statistics
!            on all variables from the YY matrix:
!
!            - YSTAT(:,1) contains the mean values of the "right" data matrix YY.
!            - YSTAT(:,2) contains the the standard-deviations of the "right" data matrix YY.
!            - YSTAT(:,3) contains the the numbers of non-missing observations
!              in the "right" data matrix YY.
!            - YSTAT(:,4) is used as workspace.
!
!            The shape of YSTAT must verify:
!
!            - size(YSTAT,1) = size(Y,DIMVARY) ,
!            - size(YSTAT,2) = 4 .
!
!   XYCOR    (INPUT/OUTPUT) real(stnd), dimension(:,:,4)
!            On entry, after the first call to COMP_MCA_MISS (e.g. when FIRST=true),
!            XYCOR is used as workspace to accumulate quantities on previous calls
!            to COMP_MCA_MISS. XYCOR should not be changed between calls to
!            COMP_MCA_MISS.
!
!            On exit, when LAST=true, XYCOR contains the following statistics:
!
!            - XYCOR(i,j,1)   contains the correlation coefficients between XX(i,:) and YY(j,:)
!              ( XX(:,i) and YY(:,j) if DIMVARX=2 and DIMVARY=2 ).
!            - XYCOR(i,j,2)   contains the incidence values between XX(i,:) and YY(j,:) ( XX(:,i) and 
!              YY(:,j) if DIMVARX=2  and DIMVARY=2). XYCOR(i,j,2) indicates the numbers
!              of non-missing pairs of  observations which were used in the calculation
!              of XYCOR(i,j,1).
!            - XYCOR(:,:,3)   contains the first min(size(X,DIMVARX),size(Y,DIMVARY)) columns of U,
!              the left singular vectors of the correlation (or covariance) matrix
!              XYCOV between XX and YY.
!            - XYCOR(:,:,4) is used as workspace.
!
!            The shape of XYCOR must verify:
!
!            - size(XYCOR,1) = size(X,DIMVARX) ,
!            - size(XYCOR,2) = size(Y,DIMVARY) ,
!            - size(XYCOR,3) = 4 .
!
!   XYMISS   (INPUT) real(stnd)
!            On entry, the missing value indicator. Any value in
!            X or Y which is equal to XYMISS is assumed to be missing
!            or invalid.
!            The basic univariate statistics and correlation (or covariance) matrix are computed
!            on all the observations where X and Y are not missing (see Further Details).
!
!   FAILURE  (OUTPUT) logical(lgl)
!            On exit when LAST=true:
!
!            - FAILURE = false:   indicates successful exit.
!            - FAILURE = true:    indicates that fewer than two valid observations
!              were present for some pair of variables or
!              that the observations on some
!              variable were constant and the correlations
!              were requested or that maximum accuracy was not
!              achieved when computing the SVD of the covariance
!              (or correlation) matrix between the data matrices XX and YY .
!
!            On exit when LAST=false, FAILURE is always set to false.
!
!   DIMVARX  (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVARX is present, DIMVARX is used as follows, if:
!
!            - DIMVARX = 1, the input submatrix X contains size(X,2) observations
!              on size(X,1) variables.
!            - DIMVARX = 2, the input submatrix X contains size(X,1) observations
!              on size(X,2) variables, respectively.
!
!            The default is DIMVARX = 1.
!
!   DIMVARY  (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVARY is present, DIMVARY is used as follows, if:
!
!            - DIMVARY = 1, the input submatrix Y contains size(Y,2) observations
!              on size(Y,1) variables.
!            - DIMVARY = 2, the input submatrix Y contains size(Y,1) observations
!              on size(Y,2) variables, respectively.
!
!            The default is DIMVARY = 1.
!          
!   COV      (INPUT, OPTIONAL) logical(lgl)
!            On entry, if COV is present and COV=true, a covariance matrix between
!            the data matrices XX and YY is computed instead of a correlation matrix.
!
!            By default, a correlation matrix is computed.
!
!            COV needs to be specified only on the last call to COMP_MCA_MISS (e.g. when LAST=true).
!
!   SORT     (INPUT, OPTIONAL) character
!            Sort the singular values into ascending order if SORT = 'A' or 'a',
!            or in descending order if SORT = 'D' or 'd'.
!            The singular vectors are rearranged accordingly.
!
!            SORT needs to be specified only on the last call to COMP_MCA (e.g. when LAST=true).
!
!   MAXITER  (INPUT, OPTIONAL) integer(i4b)
!            MAXITER controls the maximum number of QR sweeps in the
!            bidiagonal SVD phase of the SVD algorithm.
!            The bidiagonal SVD algorithm of an intermediate bidiagonal 
!            form B of XYCOR fails to converge if the number of QR sweeps 
!            exceeds MAXITER * min(m,n). Convergence usually occurs in 
!            about 2 * min(m,n) QR sweeps. The default is 10.
!
!            MAXITER needs to be specified only on the last call to COMP_MCA (e.g. when LAST=true).
!
!   XYSINGVAL (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!             On exit, XYSINGVAL contains the singular values of the correlation
!             (or covariance) matrix XYCOR between the data matrices XX and YY.
!             If this optional argument is absent, the singular values are stored
!             in XSTAT(:,4), when LAST=true.
!
!             The size of XYSINGVAL must verify: size(XYSINGVAL) = min(size(X,DIMVARX),size(Y,DIMVARY)).
!          
!   YSINGVEC (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!            On exit, when LAST=true and YSINGVEC is present, YSINGVEC contains the first
!            min(size(X,DIMVARX),size(Y,DIMVARY))  columns of V, the right singular vectors
!            of the correlation (or covariance) matrix XYCOR between XX and YY.
!
!            YSINGVEC needs to be specified only on the last call to COMP_MCA_MISS (e.g. when LAST=true).
!
!            The shape of YSINGVEC must verify:
!
!            - size(YSINGVEC,1) = size(Y,DIMVARY) ,
!            - size(YSINGVEC,2) = min( size(X,DIMVARX) ,size(Y,DIMVARY) ) .
!
!   XYSINGVAR (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!             On exit, when LAST=true and XYSINGVAR is present, XYSINGVAR contains
!             percentages of total squared covariance associated with the left and right
!             singular vectors in order of the singular values stored in XYSINGVAL.
!
!             XYSINGVAR needs to be specified only on the last call to COMP_MCA_MISS (e.g. when LAST=true).
!
!             The size of XYSINGVAR must verify: size(XYSINGVAR) = min( size(X,DIMVARX),size(Y,DIMVARY) ).
!          
!          
! Further Details
! _______________
!
!   The subroutine computes the basic univariate statistics and the correlation (or covariance)
!   matrix with only one pass through the data.
!
!   If fewer than two valid observations were present for some pair of variables or if 
!   the observations on some variable were constants, the statistics XYSINGVAL, XYSINGVEC,
!   XYSINGVAR and XYCOR(:,:,1) are globally set to XMISS.
!
!   The means and standard-deviations of XX and YY are computed from all valid data. The
!   correlation coefficients are based on these univariate statistics and on all valid pairs
!   of observations. The singular vectors and singular values are computed from these 
!   bivariate statistics.
!
!   This subroutine may be used in a call with no observations (e.g. size(X,3-DIMVARX) = 0)
!   in order to finish the computations with LAST=true when the total number of observations
!   is unknown at the beginning of the computations.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true, false
    use Reals_Constants,   only : zero, one, c100
    use Num_Constants,     only : lamch
    use Char_Constants,    only : stat_error1, stat_error5
    use Utilities,         only : assert_eq, assert, merror
    use SVD_Procedures,    only : bd_cmp, ortho_gen_bd, ortho_gen_q_bd, bd_svd
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                              :: xymiss
    real(stnd), intent(in),    dimension(:,:)           :: x, y
    real(stnd), intent(inout), dimension(:,:)           :: xstat, ystat
    real(stnd), intent(inout), dimension(:,:,:)         :: xycor
    real(stnd), intent(out),   dimension(:),   optional :: xysingval, xysingvar
    real(stnd), intent(out),   dimension(:,:), optional :: ysingvec 
!
    integer(i4b), intent(in), optional :: dimvarx, dimvary, maxiter
!
    logical(lgl), intent(in)            :: first, last
    logical(lgl), intent(in),  optional :: cov
    logical(lgl), intent(out)           :: failure
!
    character ,   intent(in), optional  :: sort
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                              :: safmin
    real(stnd), dimension(size(xycor,1))                    :: xi, dxi, dfxi
    real(stnd), dimension(size(xycor,2))                    :: yi, dfyi
    real(stnd), dimension(max(size(xycor,1),size(xycor,2))) :: dyi
!
    integer(i4b) :: i, j, n, m, p
!
    logical(lgl)                            :: dimx_is_one, dimy_is_one, comp_cor
    logical(lgl), dimension(size(xycor,1))  :: maskx
    logical(lgl), dimension(size(xycor,2))  :: masky
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_mca_miss'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvarx) ) then
!
        if ( dimvarx==1_i4b .or. dimvarx==2_i4b ) then
            i = dimvarx
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dimx_is_one = i==1_i4b
!
    if ( present(dimvary) ) then
!
        if ( dimvary==1_i4b .or. dimvary==2_i4b ) then
            j = dimvary
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        j = 1_i4b
!
    end if
!
    dimy_is_one = j==1_i4b
!
    n = assert_eq(  int(size(x,3-int(i)),i4b),  &
                    int(size(y,3-int(j)),i4b),  &
                    name_proc )
!
    m = assert_eq(  int(size(x,int(i)),i4b),      &
                    int(size(xstat,1),i4b),       &
                    int(size(xycor,1),i4b),       &
                    name_proc )
!
    p = assert_eq(  int(size(y,int(j)),i4b),      &
                    int(size(ystat,1),i4b),       &
                    int(size(xycor,2),i4b),       &
                    name_proc )
!
    if ( m<=0_i4b .or. p<=0_i4b ) call merror( name_proc//stat_error1 )
!
    call assert( logical(size(xstat,2)==4,lgl),    &
                 logical(size(ystat,2)==4,lgl),    &
                 logical(size(xycor,3)==4,lgl),    &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBMATRICES ARE THE FIRST.
!
!       INITIALIZE xstat, ystat AND xycor.
!
        xstat(:m,1_i4b:3_i4b) = zero       
        ystat(:p,1_i4b:3_i4b) = zero
        xycor(:m,:p,1_i4b:4_i4b) = zero
!
        xstat(:m,4_i4b) = xymiss
        ystat(:p,4_i4b) = xymiss
!
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  (m*p)>=omp_limit                    .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)      &
!$OMP         ,PRIVATE(i,j)
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS AND CORRECTED SUMS OF SQUARES AND CROSS-PRODUCTS, 
!       HANDLING THE DATA ONCE.
!
!$OMP SINGLE
!
        if ( dimx_is_one ) then
            xi(:m) = x(:m,i)
        else
            xi(:m) = x(i,:m)
        end if
!
        maskx(:m) = xi(:m)/=xymiss
!
#ifdef _WHERE
        do j = 1_i4b, m
!
            if ( maskx(j) ) then
!
                if ( xstat(j,4_i4b)==xymiss ) then
                    xstat(j,4_i4b) = xi(j)
                end if
!                xstat(j,4_i4b) = merge( xi(j), xstat(j,4_i4b), xstat(j,4_i4b)==xymiss )
!
                xstat(j,3_i4b) = xstat(j,3_i4b) + one
                dxi(j) = xi(j) - xstat(j,1_i4b)
                dfxi(j) = dxi(j)/xstat(j,3_i4b)
                xstat(j,1_i4b) = xstat(j,1_i4b) + dfxi(j)
                xstat(j,2_i4b) = xstat(j,2_i4b) + dxi(j)*dxi(j) - dxi(j)*dfxi(j)
                dxi(j) = xi(j) - xstat(j,4_i4b)
!
            end if
!
        end do
#else
        where ( maskx(:m) )
!
            where ( xstat(:m,4_i4b)==xymiss ) xstat(:m,4_i4b) = xi(:m)
!            xstat(:m,4_i4b) = merge( xi(:m), xstat(:m,4_i4b), xstat(:m,4_i4b)==xymiss )
!
            xstat(:m,3_i4b) = xstat(:m,3_i4b) + one
            dxi(:m) = xi(:m) - xstat(:m,1_i4b)
            dfxi(:m) = dxi(:m)/xstat(:m,3_i4b)
            xstat(:m,1_i4b) = xstat(:m,1_i4b) + dfxi(:m)
            xstat(:m,2_i4b) = xstat(:m,2_i4b) + dxi(:m)*dxi(:m) - dxi(:m)*dfxi(:m)
            dxi(:m) = xi(:m) - xstat(:m,4_i4b)
!
        end where
#endif
!
        if ( dimy_is_one ) then
            yi(:p) = y(:p,i)
        else
            yi(:p) = y(i,:p)
        end if
!
        masky(:p) = yi(:p)/=xymiss
!#ifdef _WHERE
!        do j = 1_i4b, p
!            if ( masky(j) ) then
!
!                ystat(j,4_i4b) = merge( yi(j), ystat(j,4_i4b), ystat(j,4_i4b)==xymiss )
!
!                ystat(j,3_i4b) = ystat(j,3_i4b) + one
!                dyi(j) = yi(j) - ystat(j,1_i4b)
!                dfyi(j) = dyi(j)/ystat(j,3_i4b)
!                ystat(j,1_i4b) = ystat(j,1_i4b) + dfyi(j)
!                ystat(j,2_i4b) = ystat(j,2_i4b) + dyi(j)*dyi(j) - dyi(j)*dfyi(j)
!                dyi(j) = yi(j) - ystat(j,4_i4b)
!            end if
!        end do
!#else
!        where ( masky(:p) )
!
!            ystat(:p,4_i4b) = merge( yi(:p), ystat(:p,4_i4b), ystat(:p,4_i4b)==xymiss )
!
!            ystat(:p,3_i4b) = ystat(:p,3_i4b) + one
!            dyi(:p) = yi(:p) - ystat(:p,1_i4b)
!            dfyi(:p) = dyi(:p)/ystat(:p,3_i4b)
!            ystat(:p,1_i4b) = ystat(:p,1_i4b) + dfyi(:p)
!            ystat(:p,2_i4b) = ystat(:p,2_i4b) + dyi(:p)*dyi(:p) - dyi(:p)*dfyi(:p)
!            dyi(:p) = yi(:p) - ystat(:p,4_i4b)
!        end where
!#endif
!
!$OMP END SINGLE
!
!$OMP DO SCHEDULE(STATIC) 
!
        do j = 1_i4b, p
!
            if ( masky(j) ) then
!
                if ( ystat(j,4_i4b)==xymiss ) then
                    ystat(j,4_i4b) = yi(j)
                end if
!                ystat(j,4_i4b) = merge( yi(j), ystat(j,4_i4b), ystat(j,4_i4b)==xymiss )
!
                ystat(j,3_i4b) = ystat(j,3_i4b) + one
                dyi(j)         = yi(j) - ystat(j,1_i4b)
                dfyi(j)        = dyi(j)/ystat(j,3_i4b)
                ystat(j,1_i4b) = ystat(j,1_i4b) + dfyi(j)
                ystat(j,2_i4b) = ystat(j,2_i4b) + dyi(j)*dyi(j) - dyi(j)*dfyi(j)
                dyi(j)         = yi(j) - ystat(j,4_i4b)
!
                where ( maskx(:m) )
                    xycor(:m,j,1_i4b) = xycor(:m,j,1_i4b) + dxi(:m)*dyi(j)
                    xycor(:m,j,2_i4b) = xycor(:m,j,2_i4b) + one
                    xycor(:m,j,3_i4b) = xycor(:m,j,3_i4b) + dxi(:m)
                    xycor(:m,j,4_i4b) = xycor(:m,j,4_i4b) + dyi(j)
                end where
!
            end if
!
        end do
!
!$OMP END DO
!
    end do
!
!$OMP END PARALLEL
!
    if ( last ) then
!
!       THE CURRENT SUBMATRICES ARE THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
        n = min( m, p )
!
!       TEST OPTIONAL ARGUMENTS.
!
        comp_cor = true
        if ( present(cov) ) then
            comp_cor = .not. cov
        end if
!
        if ( comp_cor ) then
            safmin = lamch( 's' )
        end if
!
        if ( present(ysingvec) ) then
            call assert( logical(size(ysingvec, 1)==int(p) ,lgl),    &
                         logical(size(ysingvec, 2)==int(n) ,lgl),    &
                         name_proc )
        end if
!
        if ( present(xysingval) ) then
            call assert( logical(size(xysingval)==int(n) ,lgl),     &
                         name_proc )
        end if
!
        if ( present(xysingvar) ) then
            call assert( logical(size(xysingvar)==int(n) ,lgl),     &
                         name_proc )
        end if
!
!       SET FAILURE INDICATOR.
!
        failure = any( xycor(:m,:p,2_i4b)<=one ) 
!
        if ( .not. failure ) then
!
!           COMPUTE STANDARD-DEVIATIONS OF XX.
!
            dfxi(:m) = sqrt( xstat(:m,2_i4b)/xstat(:m,3_i4b) )
            xstat(:m,2_i4b) = dfxi(:m)
            xi(:m) = xstat(:m,4_i4b) - xstat(:m,1_i4b)
!
!           COMPUTE STANDARD-DEVIATIONS OF YY.
!
            dfyi(:p) = sqrt( ystat(:p,2_i4b)/ystat(:p,3_i4b) )
            ystat(:p,2_i4b) = dfyi(:p)
            yi(:p) = ystat(:p,4_i4b) - ystat(:p,1_i4b)
!
            if ( comp_cor ) then
!
!$OMP PARALLEL IF(test_par)                 &
!$OMP         ,PRIVATE(j,dxi,dyi, maskx)
!
!$OMP DO SCHEDULE(STATIC) 
!
                do j = 1_i4b, p
!
                    dxi(:m) = dfxi(:m)*dfyi(j)
                    maskx(:m) = dxi(:m)>safmin
!
                    where ( maskx(:m) )
!
!                       COMPUTE CORRELATION COEFFICIENTS BETWEEN XX AND YY.
!
                        xycor(:m,j,1_i4b) = xycor(:m,j,1_i4b) + xi(:m)*xycor(:m,j,4_i4b) +               &
                                            yi(j)*xycor(:m,j,3_i4b) + (xi(:m)*yi(j))*xycor(:m,j,2_i4b)
                        dyi(:m) = xycor(:m,j,1_i4b)/( xycor(:m,j,2_i4b)*dxi(:m) )
!
                        where ( dyi(:m)>one )
                            dxi(:m) = one
                        elsewhere
                            dxi(:m) = dyi(:m)
                        end where
!
!                        dxi(:m) = merge( one, dyi(:m), dyi(:m)>one )
!
                        where ( dxi(:m)<-one )
                            xycor(:m,j,1_i4b) = -one
                        elsewhere
                            xycor(:m,j,1_i4b) = dxi(:m)
                        end where
!
!                        xycor(:m,j,1_i4b) = merge( -one, dxi(:m), dxi(:m)<-one )
!
                    elsewhere
!
!                       THE OBSERVATIONS ON XX or YY WERE CONSTANT OR FEWER THAN
!                       TWO VALID PAIRS WERE PRESENT, SET THE CORRELATION TO xymiss .
!
                        xycor(:m,j,1_i4b) = xymiss
!
                    end where
!
                end do
!
!$OMP END DO
!
!$OMP END PARALLEL
!
                failure = any( xycor(1_i4b,:p,1_i4b) == xymiss )
!
            else
!
!$OMP PARALLEL IF(test_par)                 &
!$OMP         ,PRIVATE(j,dxi)
!
!$OMP DO SCHEDULE(STATIC) 
!
                do j = 1_i4b, p
!
!                   COMPUTE COVARIANCE COEFFICIENTS BETWEEN XX AND YY.
!
                    xycor(:m,j,1_i4b) = xycor(:m,j,1_i4b) + xi(:m)*xycor(:m,j,4_i4b) +               &
                                      yi(j)*xycor(:m,j,3_i4b) + (xi(:m)*yi(j))*xycor(:m,j,2_i4b)
                    dxi(:m) = xycor(:m,j,1_i4b)/xycor(:m,j,2_i4b)
                    xycor(:m,j,1_i4b) = dxi(:m)
!
                end do
!
!$OMP END DO
!
!$OMP END PARALLEL
!
            end if
!
        end if
!
        if ( .not. failure ) then
!
!           MAKE A COPY OF THE CORRELATION MATRIX.
!
            xycor(1_i4b:m,1_i4b:p,3_i4b) = xycor(1_i4b:m,1_i4b:p,1_i4b)
!
!           COMPUTE SVD OF THE COVARIANCE OR CORRELATION MATRIX XYCOR.
!
!           BIDIAGONALIZE COVARIANCE OR CORRELATION MATRIX XYCOR .
!
            call bd_cmp( xycor(:m,:p,3_i4b), xi(:n), yi(:n), dxi(:n), dyi(:n) )
!
            if ( present(ysingvec) ) then
!
!               GENERATE RIGHT VECTORS BIDIAGONALIZING XYCOV IN ysingvec .
!               GENERATE LEFT VECTORS BIDIAGONALIZING XYCOV IN xsingvec .
!
                call ortho_gen_bd( xycor(:m,:p,3_i4b), dxi(:n), dyi(:n), ysingvec(:p,:n) )
!
!               PERFORM BIDIAGONAL QR ITERATION, COMPUTING LEFT
!               SINGULAR VECTORS OF XYCOV IN xsingvec AND COMPUTING RIGHT
!               SINGULAR VECTORS OF XYCOV IN ysingvec .
!
                call bd_svd( merge(true,false,m>=p), xi(:n), yi(:n), failure, xycor(:m,:n,3_i4b), ysingvec(:p,:n),     &
                             sort=sort, maxiter=maxiter )
!
            else
!
!               GENERATE LEFT VECTORS BIDIAGONALIZING XYCOV IN xsingvec .
!
                call ortho_gen_q_bd( xycor(:m,:p,3_i4b), dxi(:n) )
!
!               PERFORM BIDIAGONAL QR ITERATION, COMPUTING LEFT
!               SINGULAR VECTORS OF XYCOV IN xsingvec.
!
                call bd_svd( merge(true,false,m>=p), xi(:n), yi(:n), failure, xycor(:m,:n,3_i4b),     &
                             sort=sort, maxiter=maxiter )
!
            end if
!
!           SET UNUSED PART OF xsingvec TO ZERO.
!
            xycor(:m,n+1_i4b:p,3_i4b) = zero
!
!           OUTPUT SINGULAR VALUES.
!
            if ( present(xysingval) ) then
!
                xysingval(1_i4b:n) = xi(1_i4b:n)
!
            else
!
                xstat(1_i4b:n,4_i4b)   = xi(1_i4b:n)
                xstat(n+1_i4b:m,4_i4b) = zero
!
            end if
!
            if ( present(xysingvar) ) then
!
!               COMPUTE PERCENTAGES OF SQUARED COVARIANCE, IF REQUIRED.
!
                xysingvar(1_i4b:n) = xi(1_i4b:n)**2
                safmin = c100 / sum( xysingvar(1_i4b:n) )
                xysingvar(1_i4b:n) = xysingvar(1_i4b:n) * safmin
!
            end if
!
        else
!
!           SET FAILURE INDICATOR.
!
            failure = true
!
            maskx(:m) = xstat(:m,3_i4b)>one
!
            where ( maskx(:m) )
!
!               COMPUTE STANDARD-DEVIATIONS OF XX.
!
                dfxi(:m) = xstat(:m,2_i4b)/xstat(:m,3_i4b)
                xstat(:m,2_i4b) = sqrt( dfxi(:m) )
!
            elsewhere
!
!               FEWER THAN TWO VALID OBSERVATIONS, SET THE PERTINENT STATISTICS TO xymiss .
!
                where ( xstat(:m,3_i4b)<one )
                    xstat(:m,1_i4b) = xymiss
                    xstat(:m,2_i4b) = xymiss
                elsewhere
                    xstat(:m,2_i4b) = zero
                end where
!
!                xstat(:m,1_i4b) = merge( xymiss, xstat(:m,1_i4b), xstat(:m,3_i4b)<one )
!                xstat(:m,2_i4b) = merge( xymiss, zero, xstat(:m,3_i4b)<one )
!
            end where
!
            masky(:p) = ystat(:p,3_i4b)>one
!
            where ( masky(:p) )
!
!               COMPUTE STANDARD-DEVIATIONS OF YY.
!
                dfyi(:p) = ystat(:p,2_i4b)/ystat(:p,3_i4b)
                ystat(:p,2_i4b) = sqrt( dfyi(:p) )
!
            elsewhere
!
!               FEWER THAN TWO VALID OBSERVATIONS, SET THE PERTINENT STATISTICS TO xymiss .
!
                where ( ystat(:p,3_i4b)<one )
                    ystat(:p,1_i4b) = xymiss
                    ystat(:p,2_i4b) = xymiss
                elsewhere
                    xstat(:m,2_i4b) = zero
                end where
!
!                ystat(:p,1_i4b) = merge( xymiss, ystat(:p,1_i4b), ystat(:p,3_i4b)<one )
!                ystat(:p,2_i4b) = merge( xymiss, zero, ystat(:p,3_i4b)<one )
!
            end where
!
!           FEWER THAN TWO VALID OBSERVATIONS, SET THE PERTINENT STATISTICS TO xymiss CODE.
!
            xycor(:m,:p,1_i4b) = xymiss
            xycor(:m,:p,3_i4b) = xymiss
!
            if ( present(ysingvec) ) then
                ysingvec(:p,:n) = xymiss
            end if
!
            if ( present(xysingval) ) then
                xysingval(:n) = xymiss
            else
                xstat(:m,4_i4b) = xymiss
            end if
!
            if ( present(xysingvar) ) then
                xysingvar(:n) = xymiss
            end if
!
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_mca_miss
! _______________________________
!
    end subroutine comp_mca_miss
!
! =========================================================================================
!
    subroutine comp_mca_miss2( x, y, first, last, xstat, ystat, xycor, xymiss, failure,               &
                               dimvarx, dimvary, cov, xysingval, maxiter, ortho, xysingvar, xysingvec )
!
! Purpose
! _______
!
!   COMP_MCA_MISS2 performs Maximum Covariance Analysis (MCA) or canonical covariance analysis
!   between two data matrices XX and YY possibly containing missing values.
!
!   COMP_MCA_MISS2 computes a partial singular value decomposition (SVD) of the correlation 
!   (or covariance) matrix XYCOR between two data matrices XX and YY. This partial SVD is
!   written
!
!                       U(:m,:k) * SIGMA(:k,:k) * V(:n,:k)'
!
!   where SIGMA is a k-by-k matrix which is zero except for its
!   k diagonal elements, U is a m-by-k orthogonal matrix, and
!   V is a n-by-k orthogonal matrix.  The diagonal elements of SIGMA
!   are the first k singular values of XYCOR; they are real and non-negative.
!   The k columns of U and V are the first k left and right singular 
!   vectors of XYCOR.
!
!
!   COMP_MCA_MISS2 computes all the singular values, and, optionally, selected left and right
!   singular vectors (by inverse iteration), of the covariance (or correlation matrix) XYCOR
!   between two data matrices XX and YY.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:,:)
!            On entry, input submatrix containing size(X,3-DIMVARX)
!            observations on size(X,DIMVARX) variables from the "left" matrix of data XX.
!            By default, DIMVARX is equal to 1. See description of optional DIMVARX
!            argument for details.
!            If all the data are available at once, X can be the full 
!            data matrix XX.
!
!            The shape of X must verify: size(X,3-DIMVARX)  = size(Y,3-DIMVARY) .
!                                         
!   Y        (INPUT) real(stnd), dimension(:,:)
!            On entry, input submatrix containing size(Y,3-DIMVARY)
!            observations on size(Y,DIMVARY) variables from the "right" matrix of data YY.
!            By default, DIMVARY is equal to 1. See description of optional DIMVARY
!            argument for details.
!            If all the data are available at once, Y can be the full 
!            data matrix YY.
!
!            The shape of Y must verify: size(Y,3-DIMVARY)  = size(X,3-DIMVARX) .
!
!   FIRST    (INPUT) logical(lgl)
!            On entry, if:
!
!            - FIRST = true  the current submatrices are the first submatrices 
!              of the data matrices XX and YY.
!            - FIRST = false the current submatrices are not the first submatrices
!              of the data matrices XX and YY.
!
!   LAST     (INPUT) logical(lgl)
!            On entry, if:
!
!            - LAST = true  the current submatrices are the last submatrices 
!              of the data matrices XX and YY.
!            - LAST = false the current submatrix are not the last submatrices
!              of the data matrices XX and YY.
!
!   XSTAT    (INPUT/OUTPUT) real(stnd), dimension(:,4)
!            On entry, after the first call to COMP_MCA_MISS2 (e.g. when FIRST=true),
!            XSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_MCA_MISS2. XSTAT should not be changed between calls to
!            COMP_MCA_MISS2.
!
!            On exit, when LAST=true, XSTAT contains the following statistics
!            on all variables from the XX matrix:
!
!            - XSTAT(:,1) contains the mean values of the "left" data matrix XX.
!            - XSTAT(:,2) contains the standard-deviations of the "left" data matrix XX.
!            - XSTAT(:,3) contains the the numbers of non-missing observations
!              in the "left" data matrix XX.
!            - XSTAT(:,4) is used as workspace.
!
!            The shape of XSTAT must verify:
!
!            - size(XSTAT,1) = size(X,DIMVARX) ,
!            - size(XSTAT,2) = 4 .
!
!   YSTAT    (INPUT/OUTPUT) real(stnd), dimension(:,4)
!            On entry, after the first call to COMP_MCA_MISS2 (e.g. when FIRST=true),
!            YSTAT is used as workspace to accumulate quantities on previous calls
!            to COMP_MCA_MISS2. YSTAT should not be changed between calls to
!            COMP_MCA_MISS2.
!
!            On exit, when LAST=true, YSTAT contains the following statistics
!            on all variables from the YY matrix:
!
!            - YSTAT(:,1) contains the mean values of the "right" data matrix YY.
!            - YSTAT(:,2) contains the the standard-deviations of the "right" data matrix YY.
!            - YSTAT(:,3) contains the the numbers of non-missing observations
!              in the "right" data matrix YY.
!            - YSTAT(:,4) is used as workspace.
!
!            The shape of YSTAT must verify:
!
!            - size(YSTAT,1) = size(Y,DIMVARY) ,
!            - size(YSTAT,2) = 4 .
!
!   XYCOR    (INPUT/OUTPUT) real(stnd), dimension(:,:,4)
!            On entry, after the first call to COMP_MCA_MISS2 (e.g. when FIRST=true),
!            XYCOR is used as workspace to accumulate quantities on previous calls
!            to COMP_MCA_MISS2. XYCOR should not be changed between calls to
!            COMP_MCA_MISS2.
!
!            On exit, when LAST=true, XYCOR contains the following statistics:
!
!            - XYCOR(i,j,1)   contains the correlation coefficients between XX(i,:) and YY(j,:)
!              ( XX(:,i) and YY(:,j) if DIMVARX=2 and DIMVARY=2 ).
!            - XYCOR(i,j,2)   contains the incidence values between XX(i,:) and YY(j,:) ( XX(:,i) and 
!              YY(:,j) if DIMVARX=2  and DIMVARY=2). XYCOR(i,j,2) indicates the numbers
!              of non-missing pairs of  observations which were used in the calculation
!              of XYCOR(i,j,1).
!            - XYCOR(:,:,3:4) is used as workspace.
!
!            The shape of XYCOR must verify:
!
!            - size(XYCOR,1) = size(X,DIMVARX) ,
!            - size(XYCOR,2) = size(Y,DIMVARY) ,
!            - size(XYCOR,3) = 4 .
!
!   XYMISS   (INPUT) real(stnd)
!            On entry, the missing value indicator. Any value in
!            X or Y which is equal to XYMISS is assumed to be missing
!            or invalid.
!            The basic univariate statistics and correlation (or covariance) matrix are computed
!            on all the observations where X and Y are not missing (see Further Details).
!
!   FAILURE  (OUTPUT) logical(lgl)
!            On exit when LAST=true:
!
!            - FAILURE = false:   indicates successful exit.
!            - FAILURE = true:    indicates that fewer than two valid observations
!              were present for some pair of variables or
!              that the observations on some
!              variable were constant and the correlations
!              were requested or that maximum accuracy was not
!              achieved when computing the eigenvalues or that
!              some eigenvectors failed to converge with MAXITER
!              inverse iterations.
!
!            On exit when LAST=false, FAILURE is always set to false.
!
!   DIMVARX  (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVARX is present, DIMVARX is used as follows, if:
!
!            - DIMVARX = 1, the input submatrix X contains size(X,2) observations
!              on size(X,1) variables.
!            - DIMVARX = 2, the input submatrix X contains size(X,1) observations
!              on size(X,2) variables, respectively.
!
!            The default is DIMVARX = 1.
!
!   DIMVARY  (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVARY is present, DIMVARY is used as follows, if:
!
!            - DIMVARY = 1, the input submatrix Y contains size(Y,2) observations
!              on size(Y,1) variables.
!            - DIMVARY = 2, the input submatrix Y contains size(Y,1) observations
!              on size(Y,2) variables, respectively.
!
!            The default is DIMVARY = 1.
!          
!   COV      (INPUT, OPTIONAL) logical(lgl)
!            On entry, if COV is present and COV=true, a covariance matrix between
!            the data matrices XX and YY is computed instead of a correlation matrix.
!
!            By default, a correlation matrix is computed.
!
!            COV needs to be specified only on the last call to COMP_MCA_MISS2 (e.g. when LAST=true).
!
!   XYSINGVAL (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!             On exit, XYSINGVAL contains the singular values of the correlation
!             (or covariance) matrix XYCOR between the data matrices XX and YY.
!             If this optional argument is absent, the singular values are stored
!             in XSTAT(:,4), when LAST=true.
!
!             The size of XYSINGVAL must verify: size(XYSINGVAL) = min(size(X,DIMVARX),size(Y,DIMVARY)) .
!          
!   MAXITER  (INPUT, OPTIONAL) integer(i4b)
!            The number of inverse iterations performed in the subroutine for computing
!            the singular vectors. By default, 2 inverse iterations are performed 
!            for all the singular vectors. This optional argument is used only if
!            the XYSINGVEC argument is present.
!
!            MAXITER needs to be specified only on the last call to COMP_MCA2 (e.g. when LAST=true).
!
!   ORTHO    (INPUT, OPTIONAL) logical(lgl)
!            If ORTHO=true the computed singular vectors are orthogonalized by the
!            Modified Gram-Schmidt algorithm. This optional argument is used only if
!            the XYSINGVEC argument is present.
!            The default is FALSE.
!
!            ORTHO needs to be specified only on the last call to COMP_MCA2 (e.g. when LAST=true).
!
!   XYSINGVAR (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!             On exit, when LAST=true and XYSINGVAR is present, XYSINGVAR contains
!             percentages of total squared covariance associated with the left and right
!             singular vectors in order of the singular values stored in XYSINGVAL.
!
!             XYSINGVAR needs to be specified only on the last call to COMP_MCA_MISS2 (e.g. when LAST=true).
!
!             The size of XYSINGVAR must verify: size(XYSINGVAR) = min( size(X,DIMVARX),size(Y,DIMVARY) ).
!          
!   XYSINGVEC (OUTPUT, OPTIONAL) real(stnd), dimension(:,:)
!             On exit, when LAST=true and XYSINGVEC is present, XYSINGVEC contains the
!             first columns of U and V, the first k left and right singular vectors
!             of the correlation (or covariance) matrix XYCOR between XX and YY.
!
!             The first k left singular vectors are stored in XYSINGVEC(:size(X,DIMVARX),:)
!             The first k right singular vectors are stored in XYSINGVEC(size(X,DIMVARX)+1:,:)
!
!             XYSINGVEC needs to be specified only on the last call to COMP_MCA2 (e.g. when LAST=true).
!
!             The shape of XYSINGVEC must verify:
!
!             - size(XYSINGVEC,1)  = size(X,DIMVARX) + size(Y,DIMVARY) ,
!             - size(XYSINGVEC,2) <= min( size(X,DIMVARX) ,size(Y,DIMVARY) ) .
!
!          
! Further Details
! _______________
!
!   The subroutine computes the basic univariate statistics and the correlation (or covariance)
!   matrix with only one pass through the data.
!
!   If fewer than two valid observations were present for some pair of variables or if 
!   the observations on some variable were constants, the statistics XYSINGVAL, XYSINGVEC,
!   XYSINGVAR and XYCOR(:,:,1) are globally set to XMISS.
!
!   The means and standard-deviations of XX and YY are computed from all valid data. The
!   correlation coefficients are based on these univariate statistics and on all valid pairs
!   of observations. The singular vectors and singular values are computed from these 
!   bivariate statistics.
!
!   This subroutine may be used in a call with no observations (e.g. size(X,3-DIMVARX) = 0)
!   in order to finish the computations with LAST=true when the total number of observations
!   is unknown at the beginning of the computations.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants, only : true, false
    use Reals_Constants,   only : zero, one, c100
    use Num_Constants,     only : lamch
    use Char_Constants,    only : stat_error1, stat_error5
    use Utilities,         only : assert_eq, assert, merror
    use SVD_Procedures,    only : bd_inviter2, svd_cmp
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                              :: xymiss
    real(stnd), intent(in),    dimension(:,:)           :: x, y
    real(stnd), intent(inout), dimension(:,:)           :: xstat, ystat
    real(stnd), intent(inout), dimension(:,:,:)         :: xycor
    real(stnd), intent(out),   dimension(:),   optional :: xysingval, xysingvar
    real(stnd), intent(out),   dimension(:,:), optional :: xysingvec 
!
    integer(i4b), intent(in), optional :: dimvarx, dimvary, maxiter
!
    logical(lgl), intent(in)            :: first, last
    logical(lgl), intent(in),  optional :: cov, ortho
    logical(lgl), intent(out)           :: failure
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                              :: safmin
    real(stnd), dimension(size(xycor,1))                    :: xi, dxi, dfxi
    real(stnd), dimension(size(xycor,2))                    :: yi, dfyi
    real(stnd), dimension(max(size(xycor,1),size(xycor,2))) :: dyi
!
    integer(i4b) :: i, j, k, n, m, p, mp
!
    logical(lgl)                            :: dimx_is_one, dimy_is_one, comp_cor
    logical(lgl), dimension(size(xycor,1))  :: maskx
    logical(lgl), dimension(size(xycor,2))  :: masky
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_mca_miss2'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    failure = false
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvarx) ) then
!
        if ( dimvarx==1_i4b .or. dimvarx==2_i4b ) then
            i = dimvarx
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dimx_is_one = i==1_i4b
!
    if ( present(dimvary) ) then
!
        if ( dimvary==1_i4b .or. dimvary==2_i4b ) then
            j = dimvary
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        j = 1_i4b
!
    end if
!
    dimy_is_one = j==1_i4b
!
    n = assert_eq(  int(size(x,3-int(i)),i4b),  &
                    int(size(y,3-int(j)),i4b),  &
                    name_proc )
!
    m = assert_eq(  int(size(x,int(i)),i4b),           &
                    int(size(xstat,1),i4b),            &
                    int(size(xycor,1),i4b),            &
                    name_proc )
!
    p = assert_eq(  int(size(y,int(j)),i4b),           &
                    int(size(ystat,1),i4b),            &
                    int(size(xycor,2),i4b),            &
                    name_proc )
!
    if ( m<=0_i4b .or. p<=0_i4b ) call merror( name_proc//stat_error1 )
!
    call assert( logical(size(xstat,2)==4,lgl),    &
                 logical(size(ystat,2)==4,lgl),    &
                 logical(size(xycor,3)==4,lgl),    &
                 name_proc )
!
    if ( first ) then
!
!       THE CURRENT SUBMATRICES ARE THE FIRST.
!
!       INITIALIZE xstat, ystat AND xycor.
!
        xstat(:m,1_i4b:3_i4b) = zero       
        ystat(:p,1_i4b:3_i4b) = zero
        xycor(:m,:p,1_i4b:4_i4b) = zero
!
        xstat(:m,4_i4b) = xymiss
        ystat(:p,4_i4b) = xymiss
!
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  (m*p)>=omp_limit                    .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)      &
!$OMP         ,PRIVATE(i,j)
!
    do i = 1_i4b, n
!
!       CALCULATE PARTIAL MEANS AND CORRECTED SUMS OF SQUARES AND CROSS-PRODUCTS, 
!       HANDLING THE DATA ONCE.
!
!$OMP SINGLE
!
        if ( dimx_is_one ) then
            xi(:m) = x(:m,i)
        else
            xi(:m) = x(i,:m)
        end if
!
        maskx(:m) = xi(:m)/=xymiss
!
#ifdef _WHERE
        do j = 1_i4b, m
!
            if ( maskx(j) ) then
!
                if ( xstat(j,4_i4b)==xymiss ) then
                    xstat(j,4_i4b) = xi(j)
                end if
!                xstat(j,4_i4b) = merge( xi(j), xstat(j,4_i4b), xstat(j,4_i4b)==xymiss )
!
                xstat(j,3_i4b) = xstat(j,3_i4b) + one
                dxi(j) = xi(j) - xstat(j,1_i4b)
                dfxi(j) = dxi(j)/xstat(j,3_i4b)
                xstat(j,1_i4b) = xstat(j,1_i4b) + dfxi(j)
                xstat(j,2_i4b) = xstat(j,2_i4b) + dxi(j)*dxi(j) - dxi(j)*dfxi(j)
                dxi(j) = xi(j) - xstat(j,4_i4b)
!
            end if
!
        end do
#else
        where ( maskx(:m) )
!
            where ( xstat(:m,4_i4b)==xymiss ) xstat(:m,4_i4b) = xi(:m)
!            xstat(:m,4_i4b) = merge( xi(:m), xstat(:m,4_i4b), xstat(:m,4_i4b)==xymiss )
!
            xstat(:m,3_i4b) = xstat(:m,3_i4b) + one
            dxi(:m) = xi(:m) - xstat(:m,1_i4b)
            dfxi(:m) = dxi(:m)/xstat(:m,3_i4b)
            xstat(:m,1_i4b) = xstat(:m,1_i4b) + dfxi(:m)
            xstat(:m,2_i4b) = xstat(:m,2_i4b) + dxi(:m)*dxi(:m) - dxi(:m)*dfxi(:m)
            dxi(:m) = xi(:m) - xstat(:m,4_i4b)
!
        end where
#endif
!
        if ( dimy_is_one ) then
            yi(:p) = y(:p,i)
        else
            yi(:p) = y(i,:p)
        end if
!
        masky(:p) = yi(:p)/=xymiss
!#ifdef _WHERE
!        do j = 1_i4b, p
!            if ( masky(j) ) then
!
!                ystat(j,4_i4b) = merge( yi(j), ystat(j,4_i4b), ystat(j,4_i4b)==xymiss )
!
!                ystat(j,3_i4b) = ystat(j,3_i4b) + one
!                dyi(j) = yi(j) - ystat(j,1_i4b)
!                dfyi(j) = dyi(j)/ystat(j,3_i4b)
!                ystat(j,1_i4b) = ystat(j,1_i4b) + dfyi(j)
!                ystat(j,2_i4b) = ystat(j,2_i4b) + dyi(j)*dyi(j) - dyi(j)*dfyi(j)
!                dyi(j) = yi(j) - ystat(j,4_i4b)
!            end if
!        end do
!#else
!        where ( masky(:p) )
!
!            ystat(:p,4_i4b) = merge( yi(:p), ystat(:p,4_i4b), ystat(:p,4_i4b)==xymiss )
!
!            ystat(:p,3_i4b) = ystat(:p,3_i4b) + one
!            dyi(:p) = yi(:p) - ystat(:p,1_i4b)
!            dfyi(:p) = dyi(:p)/ystat(:p,3_i4b)
!            ystat(:p,1_i4b) = ystat(:p,1_i4b) + dfyi(:p)
!            ystat(:p,2_i4b) = ystat(:p,2_i4b) + dyi(:p)*dyi(:p) - dyi(:p)*dfyi(:p)
!            dyi(:p) = yi(:p) - ystat(:p,4_i4b)
!        end where
!#endif
!
!$OMP END SINGLE
!
!$OMP DO SCHEDULE(STATIC) 
!
        do j = 1_i4b, p
!
            if ( masky(j) ) then
!
                if ( ystat(j,4_i4b)==xymiss ) then
                    ystat(j,4_i4b) = yi(j)
                end if
!                ystat(j,4_i4b) = merge( yi(j), ystat(j,4_i4b), ystat(j,4_i4b)==xymiss )
!
                ystat(j,3_i4b) = ystat(j,3_i4b) + one
                dyi(j)         = yi(j) - ystat(j,1_i4b)
                dfyi(j)        = dyi(j)/ystat(j,3_i4b)
                ystat(j,1_i4b) = ystat(j,1_i4b) + dfyi(j)
                ystat(j,2_i4b) = ystat(j,2_i4b) + dyi(j)*dyi(j) - dyi(j)*dfyi(j)
                dyi(j)         = yi(j) - ystat(j,4_i4b)
!
                where ( maskx(:m) )
!
                    xycor(:m,j,1_i4b) = xycor(:m,j,1_i4b) + dxi(:m)*dyi(j)
                    xycor(:m,j,2_i4b) = xycor(:m,j,2_i4b) + one
                    xycor(:m,j,3_i4b) = xycor(:m,j,3_i4b) + dxi(:m)
                    xycor(:m,j,4_i4b) = xycor(:m,j,4_i4b) + dyi(j)
!
                end where
!
            end if
!
        end do
!
!$OMP END DO
!
    end do
!
!$OMP END PARALLEL
!
    if ( last ) then
!
!       THE CURRENT SUBMATRICES ARE THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
        n = min( m, p )
!
!       TEST OPTIONAL ARGUMENTS.
!
        comp_cor = true
        if ( present(cov) ) then
            comp_cor = .not. cov
        end if
!
        if ( comp_cor ) then
            safmin = lamch( 's' )
        end if
!
        if ( present(xysingvec) ) then
!
            mp = m + p 
            k = size( xysingvec, 2 )
!
            call assert( logical(size(xysingvec, 1)==int(mp) ,lgl),   &
                         logical(k<=n ,lgl),                          &
                         name_proc )
!
        end if
!
        if ( present(xysingval) ) then
            call assert( logical(size(xysingval)==int(n) ,lgl),     &
                         name_proc )
        end if
!
        if ( present(xysingvar) ) then
            call assert( logical(size(xysingvar)==int(n) ,lgl),     &
                         name_proc )
        end if
!
!       SET FAILURE INDICATOR.
!
        failure = any( xycor(:m,:p,2_i4b)<=one ) 
!
        if ( .not. failure ) then
!
!           COMPUTE STANDARD-DEVIATIONS OF XX.
!
            dfxi(:m) = sqrt( xstat(:m,2_i4b)/xstat(:m,3_i4b) )
            xstat(:m,2_i4b) = dfxi(:m)
            xi(:m) = xstat(:m,4_i4b) - xstat(:m,1_i4b)
!
!           COMPUTE STANDARD-DEVIATIONS OF YY.
!
            dfyi(:p) = sqrt( ystat(:p,2_i4b)/ystat(:p,3_i4b) )
            ystat(:p,2_i4b) = dfyi(:p)
            yi(:p) = ystat(:p,4_i4b) - ystat(:p,1_i4b)
!
            if ( comp_cor ) then
!
!$OMP PARALLEL IF(test_par)                 &
!$OMP         ,PRIVATE(j,dxi,dyi, maskx)
!
!$OMP DO SCHEDULE(STATIC) 
!
                do j = 1_i4b, p
!
                    dxi(:m) = dfxi(:m)*dfyi(j)
                    maskx(:m) = dxi(:m)>safmin
!
                    where ( maskx(:m) )
!
!                       COMPUTE CORRELATION COEFFICIENTS BETWEEN XX AND YY.
!
                        xycor(:m,j,1_i4b) = xycor(:m,j,1_i4b) + xi(:m)*xycor(:m,j,4_i4b) +               &
                                            yi(j)*xycor(:m,j,3_i4b) + (xi(:m)*yi(j))*xycor(:m,j,2_i4b)
                        dyi(:m) = xycor(:m,j,1_i4b)/( xycor(:m,j,2_i4b)*dxi(:m) )
!
                        where ( dyi(:m)>one )
                            dxi(:m) = one
                        elsewhere
                            dxi(:m) = dyi(:m)
                        end where
!
!                        dxi(:m) = merge( one, dyi(:m), dyi(:m)>one )
!
                        where ( dxi(:m)<-one )
                            xycor(:m,j,1_i4b) = -one
                        elsewhere
                            xycor(:m,j,1_i4b) = dxi(:m)
                        end where
!
!                        xycor(:m,j,1_i4b) = merge( -one, dxi(:m), dxi(:m)<-one )
!
                    elsewhere
!
!                       THE OBSERVATIONS ON XX or YY WERE CONSTANT OR FEWER THAN
!                       TWO VALID PAIRS WERE PRESENT, SET THE CORRELATION TO xymiss .
!
                        xycor(:m,j,1_i4b) = xymiss
!
                    end where
!
                end do
!
!$OMP END DO
!
!$OMP END PARALLEL
!
                failure = any( xycor(1_i4b,:p,1_i4b) == xymiss )
!
            else
!
!$OMP PARALLEL IF(test_par)                 &
!$OMP         ,PRIVATE(j,dxi)
!
!$OMP DO SCHEDULE(STATIC) 
!
                do j = 1_i4b, p
!
!                   COMPUTE COVARIANCE COEFFICIENTS BETWEEN XX AND YY.
!
                    xycor(:m,j,1_i4b) = xycor(:m,j,1_i4b) + xi(:m)*xycor(:m,j,4_i4b) +               &
                                      yi(j)*xycor(:m,j,3_i4b) + (xi(:m)*yi(j))*xycor(:m,j,2_i4b)
                    dxi(:m) = xycor(:m,j,1_i4b)/xycor(:m,j,2_i4b)
                    xycor(:m,j,1_i4b) = dxi(:m)
!
                end do
!
!$OMP END DO
!
!$OMP END PARALLEL
!
            end if
!
        end if
!
        if ( .not. failure ) then
!
!           MAKE A COPY OF THE CORRELATION MATRIX.
!
            xycor(1_i4b:m,1_i4b:p,3_i4b) = xycor(1_i4b:m,1_i4b:p,1_i4b)
!
            if ( present(xysingvec) ) then
!
!               COMPUTE ALL SINGULAR VALUES AND SAVE THE INTERMEDIATE 
!               BIDIAGONAL MATRIX.
!
                call svd_cmp( xycor(1_i4b:m,1_i4b:p,3_i4b), xi(:n), failure, sort='d',  &
                              d=dfxi(:n), e=dfyi(:n), tauq=dxi(:n), taup=dyi(:n)        )
!
!               COMPUTE THE FIRST k SINGULAR VECTORS OF THE CORRELATION MATRIX
!               BY maxiter INVERSE ITERATIONS ON THE INTERMEDIATE BIDIAGONAL
!               MATRIX AND BACK-TRANSFORMATION.
!
                call bd_inviter2( xycor(1_i4b:m,1_i4b:p,3_i4b), dxi(:n), dyi(:n), dfxi(:n), dfyi(:n), xi(:k),   &
                                  xysingvec(:m,:k), xysingvec(m+1_i4b:mp,:k), dimx_is_one,                      &
                                  maxiter=maxiter, ortho=ortho )
!
                failure = failure .or. dimx_is_one
!
            else
!
!               COMPUTE ALL SINGULAR VALUES. 
!
                call svd_cmp( xycor(1_i4b:m,1_i4b:p,3_i4b), xi(:n), failure, sort='d' )
!
            end if
!
!           OUTPUT SINGULAR VALUES.
!
            if ( present(xysingval) ) then
!
                xysingval(1_i4b:n) = xi(1_i4b:n)
!
            else
!
                xstat(1_i4b:n,4_i4b)   = xi(1_i4b:n)
                xstat(n+1_i4b:m,4_i4b) = zero
!
            end if
!
            if ( present(xysingvar) ) then
!
!               COMPUTE PERCENTAGES OF SQUARED COVARIANCE, IF REQUIRED.
!
                xysingvar(1_i4b:n) = xi(1_i4b:n)**2
                safmin = c100 / sum( xysingvar(1_i4b:n) )
                xysingvar(1_i4b:n) = xysingvar(1_i4b:n) * safmin
!
            end if
!
        else
!
!           SET FAILURE INDICATOR.
!
            failure = true
!
            maskx(:m) = xstat(:m,3_i4b)>one
!
            where ( maskx(:m) )
!
!               COMPUTE STANDARD_DEVIATIONS OF XX.
!
                dfxi(:m) = xstat(:m,2_i4b)/xstat(:m,3_i4b)
                xstat(:m,2_i4b) = sqrt( dfxi(:m) )
!
            elsewhere
!
!               FEWER THAN TWO VALID OBSERVATIONS, SET THE PERTINENT STATISTICS TO xymiss .
!
                where ( xstat(:m,3_i4b)<one )
                    xstat(:m,1_i4b) = xymiss
                    xstat(:m,2_i4b) = xymiss
                elsewhere
                    xstat(:m,2_i4b) = zero
                end where
!
!                xstat(:m,1_i4b) = merge( xymiss, xstat(:m,1_i4b), xstat(:m,3_i4b)<one )
!                xstat(:m,2_i4b) = merge( xymiss, zero, xstat(:m,3_i4b)<one )
!
            end where
!
            masky(:p) = ystat(:p,3_i4b)>one
!
            where ( masky(:p) )
!
!               COMPUTE STANDARD_DEVIATIONS OF YY.
!
                dfyi(:p) = ystat(:p,2_i4b)/ystat(:p,3_i4b)
                ystat(:p,2_i4b) = sqrt( dfyi(:p) )
!
            elsewhere
!
!               FEWER THAN TWO VALID OBSERVATIONS, SET THE PERTINENT STATISTICS TO xymiss .
!
                where ( xstat(:m,3_i4b)<one )
                    ystat(:p,1_i4b) = xymiss
                    ystat(:p,2_i4b) = xymiss
                elsewhere
                    ystat(:p,2_i4b) = zero
                end where
!
!                ystat(:p,1_i4b) = merge( xymiss, ystat(:p,1_i4b), ystat(:p,3_i4b)<one )
!                ystat(:p,2_i4b) = merge( xymiss, zero, ystat(:p,3_i4b)<one )
!
            end where
!
!           FEWER THAN TWO VALID OBSERVATIONS, SET THE PERTINENT STATISTICS TO xymiss CODE.
!
            xysingval(:n) = xymiss
!
            xycor(:m,:p,1_i4b) = xymiss
!
            if ( present(xysingval) ) then
                xysingval(:n) = xymiss
            else
                xstat(:m,4_i4b) = xymiss
            end if
!
            if ( present(xysingvec) ) then
                xysingvec(:mp,:k) = xymiss
            end if
!
            if ( present(xysingvar) ) then
                xysingvar(:n) = xymiss
            end if
!
        end if
!
    end if
!
!
! END OF SUBROUTINE comp_mca_miss2
! ________________________________
!
    end subroutine comp_mca_miss2
!
! =========================================================================================
!
    subroutine comp_pc_mca( x, xsingvec, first, last, xpccor, pccorp, xpc, xn, dimvar, &
                            xmean, xstd, xpcvar  )
!
! Purpose
! _______
!
!   COMP_PC_MCA computes estimates of Singular Variables (SV) and correlation (or covariance)
!   fields from a data matrix XX and a set of singular vectors derived from MCA analysis of XX with
!   another matrix YY.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:,:)
!            On entry, input submatrix containing size(X,3-DIMVAR)
!            observations on size(X,DIMVAR) variables from the matrix of data
!            for which Singular Variables are desired. By default,
!            DIMVAR is equal to 1. See description of optional DIMVAR argument 
!            for details.
!            If all the data are available at once, X can be the full 
!            data matrix.
!
!   XSINGVEC (INPUT) real(stnd), dimension(:,:)
!            On entry, XSINGVEC contains selected left (or right) singular vectors 
!            of the SVD of the covariance (or correlation) matrix
!            between the data matrix XX and another data matrix YY.
!
!            The shape of XSINGVEC must verify: size(XSINGVEC,1) = size(X,DIMVAR).
!
!   FIRST    (INPUT) logical(lgl)
!            On entry, if:
!
!            - FIRST = true  the current submatrix is the first submatrix 
!              of the data matrix.
!            - FIRST = false the current submatrix is not the first submatrix
!              of the data matrix.
!
!   LAST     (INPUT) logical(lgl)
!            On entry, if:
!
!            - LAST = true  the current submatrix is the last submatrix 
!              of the data matrix.
!            - LAST = false the current submatrix is not the last submatrix
!              of the data matrix.
!
!   XPCCOR   (INPUT/OUTPUT) real(stnd), dimension(:,:)
!            On entry, after the first call to COMP_PC_MCA (e.g. when FIRST=true),
!            XPCCOR is used as workspace to accumulate quantities on previous calls
!            to COMP_PC_MCA. XPCCOR should not be changed between calls to
!            COMP_PC_MCA.
!
!            On exit, when LAST=true, XPCCOR contains:
!
!            - the correlations between the data matrix
!              and the Singular Variables if the optional
!              arguments XMEAN and XSTD are present.
!            - the covariances between the data matrix
!              and the normalized Singular Variables if only
!              the optional argument XMEAN is present.
!
!            The shape of XPCCOR must verify:
!
!            - size(XPCCOR,1) = size(X,DIMVAR) ,
!            - size(XPCCOR,2) = size(XSINGVEC,2) .
!
!   PCCORP   (INPUT/OUTPUT) real(stnd), dimension(:)
!            On entry, after the first call to COMP_PC_MCA (e.g. when FIRST=true),
!            PCCORP is used as workspace to accumulate quantities on previous calls
!            to COMP_PC_MCA. PCCORP should not be changed between calls to
!            COMP_PC_MCA.
!
!            On exit, when LAST=true, PCCORP contains the correlation matrix
!            between the Singular Variables stored in argument XPC.
!            PCCORP is stored in symmetric storage mode (see further details).
!
!            The size of PCCORP must verify: size(PCCORP) = (size(XSINGVEC,2) * (size(XSINGVEC,2)+1))/2.
!
!   XPC      (OUTPUT) real(stnd), dimension(:,:)
!            On exit, XPC contains the unnormalized Singular Variables derived
!            from X and XSINGVEC.
!
!            The shape of XPC must verify:
!
!            - size(XPC,1) = size(X,3-DIMVAR) ,
!            - size(XPC,2) = size(XSINGVEC,2) .
!
!   XN       (INPUT/OUTPUT) real(stnd)
!            On entry, after the first call to COMP_PC_MCA (e.g. when FIRST=true),
!            XN contains count of observations from previous calls
!            to COMP_PC_MCA. XN should not be changed between calls to
!            COMP_PC_MCA.
!
!            On exit, XN contains the number of observations in the data matrix XX.
!
!   DIMVAR   (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVAR is present, DIMVAR is used as follows, if:
!
!            - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!              on size(X,1) variables.
!            - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!              on size(X,2) variables.
!
!            The default is DIMVAR = 1.
!
!   XMEAN    (INPUT, OPTIONAL) real(stnd), dimension(:)
!            On entry, if XMEAN is present, XMEAN contains the variable 
!            means and the Singular Variables are computed from the
!            centered data matrix X.
!
!            The size of XMEAN must verify: size(XMEAN) = size(X,DIMVAR).
!          
!   XSTD     (INPUT, OPTIONAL) real(stnd), dimension(:)
!            On entry, if XSTD is present, XSTD contains the variable 
!            standard-deviations and the Singular Variables are computed from 
!            the normalized data matrix X.
!
!            The size of XSTD must verify: size(XSTD) = size(X,DIMVAR).
!
!   XPCVAR   (OUTPUT, OPTIONAL) real(stnd), dimension(:)
!            On exit, when LAST=true and XPCVAR is present, XPCVAR contains
!            the variances of the Singular Variables stored in argument XPC.
!
!            The size of XPCVAR must verify: size(XPCVAR) = size(XSINGVEC,2)
!
!            XPCVAR needs to be specified only on the last call to COMP_PC_MCA (e.g. when LAST=true).
!
!          
! Further Details
! _______________
!
!   The subroutine computes the Singular Variables and the correlations matrices with only
!   one pass through the data.
!
!   On exit, the upper triangle of the symmetric correlation matrix COR between the Singular
!   Variables is packed columnwise in the linear array PCCORP.  More precisely, the j-th column
!   of this matrix COR is stored in the array PCCORP as follows:
!
!            PCCORP(i + (j-1) * j/2) = COR(i,j) for 1<=i<=j;
!
!   If fewer than two valid observations were present, the arguments XPCVAR, XPCCOR and PCCORP
!   are set to nan() code.
!
!   This subroutine may be used in a call with no observations (e.g. size(X,3-DIMVAR) = size(XPC,1) = 0)
!   in order to finish the computations with LAST=true when the total number of observations is unknown
!   at the beginning of the computations.
!
!   The correlations (or covariance) between the Singular variables XPC and the data array YY
!   may be computed easily from the outputs of COMP_MCA and COMP_PC_MCA. If YSINGVEC(:p,:k) 
!   are the k singular vectors (derived from data array YY) associated with the k singular
!   values XYSINGVAL(:k) and the k singular vectors XSINGVEC(:m,:k) (derived from data array XX).
!   Then, the correlations (or covariance) between the Singular variables XPC and the data array YY
!   are equal to
!
!      spread( XYSINGVAL(:k)/SQRT(XPCVAR(:k)), dim=1, ncopies=p) * YSINGVEC(:p,:k)
!
!   This matrix is a covariance matrix between data array YY and the normalized Singular Variables
!   XPC if a covariance matrix between data arrays XX and YY is analysed or a correlation
!   matrix between data array YY and the Singular Variables XPC if a correlation matrix between data
!   arrays XX and YY is analysed.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : zero, one
    use Num_Constants,     only : nan, lamch
    use Char_Constants,    only : stat_error1, stat_error5
    use Utilities,         only : assert_eq, assert, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),                dimension(:,:) :: x, xsingvec
    real(stnd), intent(in),     optional , dimension(:)   :: xstd, xmean
    real(stnd), intent(inout)                             :: xn
    real(stnd), intent(inout),             dimension(:)   :: pccorp
    real(stnd), intent(inout),             dimension(:,:) :: xpccor
    real(stnd), intent(out),               dimension(:,:) :: xpc
    real(stnd), intent(out),    optional , dimension(:)   :: xpcvar
!
    integer(i4b), intent(in),  optional :: dimvar
!
    logical(lgl), intent(in)            :: first, last
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd)                                               :: a, b
    real(stnd), dimension(size(xsingvec,1))                  :: dxi
    real(stnd), dimension(size(xsingvec,2))                  :: dxpci
    real(stnd), dimension(size(pccorp))                      :: tmpcorp
    real(stnd), dimension(size(xsingvec,1),size(xsingvec,2)) :: tmpcor
!
    integer(i4b) :: i, j, n, m, k, nk, l, l1, l2
!
    logical(lgl) :: dim_is_one, test_mean, test_std
#ifdef _OPENMP
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_pc_mca'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    test_mean     = false
    test_std      = false
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
    j = assert_eq(  int(size(x,3-int(i)),i4b),      &
                    int(size(xpc,1),i4b),           &
                    name_proc )
!
    n = max( j, 0_i4b )
!
    m = assert_eq(  int(size(x,int(i)),i4b),       &
                    int(size(xsingvec,1),i4b),     &
                    int(size(xpccor,1),i4b),       &
                    name_proc )
!
    k = assert_eq(  int(size(xsingvec,2),i4b),     &
                    int(size(xpc,2),i4b),          &
                    int(size(xpccor,2),i4b),       &
                    name_proc )
!
    if ( min(m,k)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    nk = ( k*(k+1_i4b) )/2_i4b
!
    call assert( logical(size(pccorp)==nk,lgl),    &
                 name_proc )
!
    if ( present(xmean) ) then
!
        call assert( logical(size(xmean)==m,lgl),    &
                      name_proc )
!
        test_mean = true
!
    end if
!
    if ( present(xstd) ) then
!
        call assert( logical(size(xstd)==m,lgl),    &
                      name_proc )
!
        test_std = true
!
    end if
!
    if ( first ) then
!
!       THE CURRENT SUBMATRIX IS THE FIRST.
!
!       INITIALIZE pccorp, xpccor AND xn .
!
        xpccor(1_i4b:m,1_i4b:k) = zero
        pccorp(1_i4b:nk)        = zero
        xn = zero
!
    end if
!
    xn = xn + real( n, stnd )
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  ((n+m)*k)>=omp_limit                .and.      &
                  n>1_i4b                             .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)                                &
!$OMP         ,PRIVATE(i,j,l,l1,l2,dxi,dxpci,tmpcor,tmpcorp)
!
    tmpcorp(1_i4b:nk)       = zero
    tmpcor(1_i4b:m,1_i4b:k) = zero
!
!$OMP DO SCHEDULE(STATIC) 
!
    do i = 1_i4b, n
!
        if ( dim_is_one ) then
            dxi(1_i4b:m) = x(1_i4b:m,i)
        else
            dxi(1_i4b:m) = x(i,1_i4b:m)
        end if
!
        if ( test_mean ) then
            dxi(1_i4b:m) = dxi(1_i4b:m) - xmean(1_i4b:m)
        end if
!
        if ( test_std ) then
            dxi(1_i4b:m) = dxi(1_i4b:m) / xstd(1_i4b:m)
        end if
!
        dxpci(1_i4b:k) = matmul( dxi(1_i4b:m), xsingvec(1_i4b:m,1_i4b:k) )
!
        xpc(i,1_i4b:k) = dxpci(1_i4b:k)
!
        do j = 1_i4b, k
!
            l = (j-1_i4b)*j/2_i4b
            l1 = l + 1_i4b
            l2 = l + j
!
            tmpcorp(l1:l2) = tmpcorp(l1:l2) + dxpci(j)*dxpci(1_i4b:j)
            tmpcor(1_i4b:m,j) = tmpcor(1_i4b:m,j) + dxpci(j)*dxi(1_i4b:m)
!
        end do
!
    end do
!
!$OMP END DO
!
!$OMP CRITICAL (updatepc)
!
    pccorp(1_i4b:nk)        = pccorp(1_i4b:nk)        + tmpcorp(1_i4b:nk)
    xpccor(1_i4b:m,1_i4b:k) = xpccor(1_i4b:m,1_i4b:k) + tmpcor(1_i4b:m,1_i4b:k)
!
!$OMP END CRITICAL (updatepc)
!
!$OMP END PARALLEL
!
    if ( last ) then
!
!       THE CURRENT SUBMATRIX IS THE LAST.
!       POSTPROCESSING BEGIN HERE.
!
!
!       TEST OPTIONAL ARGUMENTS.
!
        if ( present(xpcvar) ) then
            call assert( logical(size(xpcvar)==k,lgl),    &
                         name_proc )
        end if
!
        if ( xn>one ) then
!
!           MORE THAN TWO VALID OBSERVATIONS, COMPUTE THE STATISTICS.
!
            a = one / xn
            b = lamch( 's' )
!
!           GET DIAGONAL ELEMENTS OF CORRECTED SUMS OF SQUARES 
!           AND CROSS-PRODUCTS BETWEEN PRINCIPAL COMPONENTS.
!
            do j = 1_i4b, k
!
                l = (j-1_i4b)*j/2_i4b + j
                dxpci(j) = pccorp(l)
!
            end do
!
            if ( present(xpcvar) ) then
!
!               COMPUTE VARIANCES OF PRINCIPAL COMPONENTS, IF REQUIRED.
!
                xpcvar(1_i4b:k) = dxpci(1_i4b:k) * a
!
            end if
!
            dxi(1_i4b:k) = one / ( sqrt( dxpci(1_i4b:k ) ) + b )
            dxpci(1_i4b:k) = sqrt( a ) * dxi(1_i4b:k)
!
!           COMPUTE THE VARIANCE-COVARIANCE (OR CORRELATION) MATRIX
!           BETWEEN THE PRINCIPAL COMPONENTS AND THE DATA ARRAY XX;
!           COMPUTE THE CORRELATION MATRIX BETWEEN THE PRINCIPAL COMPONENTS.
!
!$OMP PARALLEL DO IF(test_par)                    &
!$OMP            ,SCHEDULE(STATIC)                &
!$OMP            ,PRIVATE(j,l,l1,l2)
!
            do j = 1_i4b, k
!
                l = (j-1_i4b)*j/2_i4b
                l1 = l + 1_i4b
                l2 = l + j
!
                pccorp(l1:l2-1_i4b) = pccorp(l1:l2-1_i4b)*( dxi(1_i4b:j-1_i4b)*dxi(j) )
                pccorp(l2) = one
!
                xpccor(1_i4b:m,j) = xpccor(1_i4b:m,j)*dxpci(j)
!
            end do
!
!$OMP END PARALLEL DO
!
        else
!
!           FEWER THAN TWO VALID OBSERVATIONS
!           SET THE PERTINENT STATISTICS TO NaN CODE.
!
            b = nan()
!
            xpccor(1_i4b:m,1_i4b:k) = b
!
            pccorp(1_i4b:nk)        = b
!
           if ( present(xpcvar) ) then
               xpcvar(1_i4b:k) =  b
           end if
!
        end if
!
    end if
!
! END OF SUBROUTINE comp_pc_mca
! _____________________________
!
    end subroutine comp_pc_mca
!
!
! =========================================================================================
!                PRINCIPAL COMPONENTS AND SINGULAR VARIABLES COMPUTATIONS
! =========================================================================================
!
!
    subroutine comp_pc_rv( x, xeigvec, xpc, dimvar, xmean, xstd, xsingval  )
!
! Purpose
! _______
!
!   COMP_PC computes estimates of a Principal Component (PC) 
!   from a data matrix XX and an eigenvector or singular vector derived
!   from EOF or MCA analysis.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:,:)
!            On entry, input submatrix containing size(X,3-DIMVAR)
!            observations on size(X,DIMVAR) variables from the matrix of data XX
!            for which the Principal Component is desired. By default,
!            DIMVAR is equal to 1. See description of optional DIMVAR argument 
!            for details.
!            If all the data are available at once, X can be the full 
!            data matrix XX.
!
!   XEIGVEC  (INPUT) real(stnd), dimension(:)
!            On entry, XEIGVEC contains an eigenvector of
!            the variance-covariance (or correlation) matrix from the data matrix XX
!            or a selected left (or right) singular vector of the SVD of the covariance
!            matrix between the data matrix XX and another data matrix YY.
!
!            The shape of XEIGVEC must verify: size(XEIGVEC) = size(X,DIMVAR).
!
!   XPC      (OUTPUT) real(stnd), dimension(:)
!            On exit, XPC contains the Principal Component derived from X and XEIGVEC.
!
!            The size of XPC must verify: size(XPC) = size(X,3-DIMVAR).
!
!   DIMVAR   (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVAR is present, DIMVAR is used as follows, if:
!
!            - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!              on size(X,1) variables.
!            - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!              on size(X,2) variables.
!
!            The default is DIMVAR = 1.
!
!   XMEAN    (INPUT, OPTIONAL) real(stnd), dimension(:)
!            On entry, if XMEAN is present, XMEAN contains the variable 
!            means and the Principal Component is computed from the
!            centered data matrix X.
!
!            The size of XMEAN must verify: size(XMEAN) = size(X,DIMVAR).
!          
!   XSTD     (INPUT, OPTIONAL) real(stnd), dimension(:)
!            On entry, if XSTD is present, XSTD contains the variable 
!            standard-deviations and the Principal Component is computed from 
!            the normalized data matrix X.
!
!            The size of XSTD must verify: size(XSTD) = size(X,DIMVAR).
!
!   XSINGVAL (INPUT, OPTIONAL) real(stnd)
!            On entry, XSINGVAL must contain the singular value of the covariance
!            (or correlation) matrix from the data matrix XX associated with the
!            eigenvector in XEIGVEC array. If SINGVAL is present, the Principal
!            Component is normalized by XSINGVAL on output (the variance of the
!            Principal Component is equal to one). This optional argument is useful
!            only if XEIGVEC contains an eigenvector derived from an EOF analysis.
!
!          
! Further Details
! _______________
!
!   The subroutine computes the Principal Component with only one pass
!   through the data.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : zero
    use Char_Constants,    only : stat_error1, stat_error5, stat_error6
    use Utilities,         only : assert_eq, assert, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),             dimension(:)   :: xeigvec
    real(stnd), intent(in),             dimension(:,:) :: x
    real(stnd), intent(in),  optional                  :: xsingval
    real(stnd), intent(in),  optional , dimension(:)   :: xstd, xmean
    real(stnd), intent(out),            dimension(:)   :: xpc
!
    integer(i4b), intent(in),  optional :: dimvar
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(xeigvec)) :: dxi
!
    integer(i4b) :: i, n, m
!
    logical(lgl) :: dim_is_one, test_mean, test_std, test_singval
!
#ifdef _OPENMP
    integer(i4b) :: j
!
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_pc'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    test_mean     = false
    test_std      = false
    test_singval  = false
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
    n = assert_eq(  int(size(x,3-int(i)),i4b),      &
                    int(size(xpc),i4b),             &
                    name_proc )
!
    m = assert_eq(  int(size(x,int(i)),i4b),       &
                    int(size(xeigvec),i4b),        &
                    name_proc )
!
    if ( min(n,m)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    if ( present(xmean) ) then
!
        call assert( logical(size(xmean)==m,lgl),    &
                      name_proc )
!
        test_mean = true
!
    end if
!
    if ( present(xstd) ) then
!
        call assert( logical(size(xstd)==m,lgl),    &
                      name_proc )
!
        test_std = true
!
    end if
!
    if ( present(xsingval) ) then
!
        if ( xsingval<=zero ) call merror( name_proc//stat_error6 )
!
        test_singval = true
!
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  (n*m)>=omp_limit                    .and.      &
                  n>1_i4b                             .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)        &
!$OMP         ,PRIVATE(i,dxi)
!
!$OMP DO SCHEDULE(STATIC) 
!
    do i = 1_i4b, n
!
        if ( dim_is_one ) then
            dxi(1_i4b:m) = x(1_i4b:m,i)
        else
            dxi(1_i4b:m) = x(i,1_i4b:m)
        end if
!
        if ( test_mean ) then
            dxi(1_i4b:m) = dxi(1_i4b:m) - xmean(1_i4b:m)
        end if
!
        if ( test_std ) then
            dxi(1_i4b:m) = dxi(1_i4b:m) / xstd(1_i4b:m)
        end if
!
        xpc(i) = dot_product( dxi(1_i4b:m), xeigvec(1_i4b:m) )
!
    end do
!
!$OMP END DO
!
!$OMP END PARALLEL
!
    if ( test_singval ) then
        xpc(1_i4b:n) = xpc(1_i4b:n) / xsingval
    end if
!
!
! END OF SUBROUTINE comp_pc_rv
! ____________________________
!
    end subroutine comp_pc_rv
!
! =========================================================================================
!
    subroutine comp_pc_rm( x, xeigvec, xpc, dimvar, xmean, xstd, xsingval  )
!
! Purpose
! _______
!
!   COMP_PC computes estimates of Principal Components (PC) 
!   from a data matrix XX and a set of eigenvectors or singular vectors derived
!   from EOF or MCA analysis.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:,:)
!            On entry, input submatrix containing size(X,3-DIMVAR)
!            observations on size(X,DIMVAR) variables from the matrix of data XX
!            for which Principal Components are desired. By default,
!            DIMVAR is equal to 1. See description of optional DIMVAR argument 
!            for details.
!            If all the data are available at once, X can be the full 
!            data matrix XX.
!
!   XEIGVEC  (INPUT) real(stnd), dimension(:,:)
!            On entry, XEIGVEC contains selected eigenvectors of
!            the variance-covariance (or correlation) matrix from the data matrix XX
!            or selected left (or right) singular vectors of the SVD of the covariance matrix
!            between the data matrix XX and another data matrix YY.
!
!            The shape of XEIGVEC must verify:  size(XEIGVEC,1) = size(X,DIMVAR).
!
!   XPC      (OUTPUT) real(stnd), dimension(:,:)
!            On exit, XPC contains the Principal Components derived from X and XEIGVEC.
!
!            The shape of XPC must verify:
!
!            - size(XPC,1) = size(X,3-DIMVAR) ,
!            - size(XPC,2) = size(XEIGVEC,2)  .
!
!   DIMVAR   (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVAR is present, DIMVAR is used as follows, if:
!
!            - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!              on size(X,1) variables.
!            - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!              on size(X,2) variables.
!
!            The default is DIMVAR = 1.
!
!   XMEAN    (INPUT, OPTIONAL) real(stnd), dimension(:)
!            On entry, if XMEAN is present, XMEAN contains the variable 
!            means and the Principal Components are computed from the
!            centered data matrix X.
!
!            The size of XMEAN must verify: size(XMEAN) = size(X,DIMVAR).
!          
!   XSTD     (INPUT, OPTIONAL) real(stnd), dimension(:)
!            On entry, if XSTD is present, XSTD contains the variable 
!            standard-deviations and the Principal Components are computed from 
!            the normalized data matrix X.
!
!            The size of XSTD must verify: size(XSTD) = size(X,DIMVAR).
!
!   XSINGVAL  (INPUT, OPTIONAL) real(stnd), dimension(:)
!             On entry, XSINGVAL must contain the singular values of the covariance
!             (or correlation) matrix from the data matrix XX associated with the
!             eigenvectors in XEIGVEC array. If SINGVAL is present, the Principal
!             Components are normalized by XSINGVAL on output (the variances of the
!             Principal Components are equal to one). This optional argument is useful
!             only if XEIGVEC contains eigenvectors derived from an EOF analysis.
!
!             The size of XSINGVAL must verify: size(XSINGVAL) = size(XEIGVEC,2).
!          
!          
! Further Details
! _______________
!
!   The subroutine computes the Principal Components with only one pass
!   through the data.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : zero
    use Char_Constants,    only : stat_error1, stat_error5, stat_error6
    use Utilities,         only : assert_eq, assert, merror
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in),             dimension(:,:) :: x, xeigvec
    real(stnd), intent(in),  optional,  dimension(:)   :: xstd, xmean, xsingval
    real(stnd), intent(out),            dimension(:,:) :: xpc
!
    integer(i4b), intent(in),  optional :: dimvar
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(xeigvec,1)) :: dxi
    real(stnd), dimension(size(xeigvec,2)) :: dxpci
!
    integer(i4b) :: i, n, m, p
!
    logical(lgl) :: dim_is_one, test_mean, test_std, test_singval
!
#ifdef _OPENMP
    integer(i4b) :: j
!
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_pc'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    test_mean     = false
    test_std      = false
    test_singval  = false
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
    n = assert_eq(  int(size(x,3-int(i)),i4b),      &
                    int(size(xpc,1),i4b),           &
                    name_proc )
!
    m = assert_eq(  int(size(x,int(i)),i4b),       &
                    int(size(xeigvec,1),i4b),      &
                    name_proc )
!
    p = assert_eq(  int(size(xeigvec,2),i4b),      &
                    int(size(xpc,2),i4b),          &
                    name_proc )
!
    if ( min(n,m,p)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    if ( present(xmean) ) then
!
        call assert( logical(size(xmean)==m,lgl),    &
                      name_proc )
!
        test_mean = true
!
    end if
!
    if ( present(xstd) ) then
!
        call assert( logical(size(xstd)==m,lgl),    &
                      name_proc )
!
        test_std = true
!
    end if
!
    if ( present(xsingval) ) then
!
        call assert( logical(size(xsingval)==p,lgl),    &
                      name_proc )
!
        if ( any(xsingval<=zero) ) call merror( name_proc//stat_error6 )
!
        test_singval = true
!
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.      &
                  (n*p*m)>=omp_limit                  .and.      &
                  n>1_i4b                             .and.      &
                  i>1_i4b                             .and.      &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)        &
!$OMP         ,PRIVATE(i,dxi,dxpci)
!
!$OMP DO SCHEDULE(STATIC) 
!
    do i = 1_i4b, n
!
        if ( dim_is_one ) then
            dxi(1_i4b:m) = x(1_i4b:m,i)
        else
            dxi(1_i4b:m) = x(i,1_i4b:m)
        end if
!
        if ( test_mean ) then
            dxi(1_i4b:m) = dxi(1_i4b:m) - xmean(1_i4b:m)
        end if
!
        if ( test_std ) then
            dxi(1_i4b:m) = dxi(1_i4b:m) / xstd(1_i4b:m)
        end if
!
        dxpci(1_i4b:p) = matmul( dxi(1_i4b:m), xeigvec(1_i4b:m,1_i4b:p) )
!
        if ( test_singval ) then
            xpc(i,1_i4b:p) = dxpci(1_i4b:p) / xsingval(1_i4b:p)
        else
            xpc(i,1_i4b:p) = dxpci(1_i4b:p)
        end if
!
    end do
!
!$OMP END DO
!
!$OMP END PARALLEL
!
!
! END OF SUBROUTINE comp_pc_rm
! ____________________________
!
    end subroutine comp_pc_rm
!
! =========================================================================================
!
    subroutine comp_pc_miss_rv( x, xeigvec, xpc, xmiss, dimvar, xmean, xstd, xsingval   )
!
! Purpose
! _______
!
!   COMP_PC_MISS computes estimates of a Principal Component (PC) 
!   from a data matrix XX and an eigenvector or singular vector derived
!   from EOF or MCA analysis when XX contains missing values.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:,:)
!            On entry, input submatrix containing size(X,3-DIMVAR)
!            observations on size(X,DIMVAR) variables from the matrix of data XX
!            for which a Principal Components is desired. By default,
!            DIMVAR is equal to 1. See description of optional DIMVAR argument 
!            for details.
!            If all the data are available at once, X can be the full 
!            data matrix XX.
!
!   XEIGVEC  (INPUT) real(stnd), dimension(:)
!            On entry, XEIGVEC contains a selected eigenvector of
!            the variance-covariance (or correlation) matrix from the data matrix XX
!            or a selected left (or right) singular vector of the SVD of the covariance
!            matrix between the data matrix XX and another data matrix YY.
!
!            The shape of XEIGVEC must verify: size(XEIGVEC) = size(X,DIMVAR).
!
!   XPC      (OUTPUT) real(stnd), dimension(:)
!            On exit, XPC contains the Principal Component derived from X and XEIGVEC.
!
!            The shape of XPC must verify: size(XPC) = size(X,3-DIMVAR).
!
!   XMISS    (INPUT) real(stnd)
!            On entry, the missing value indicator. Any value in
!            X which is equal to XMISS is assumed to be missing
!            or invalid.
!            The Principal Components are computed from all the observations where X
!            X are not missing by regression (see Further Details).
!
!   DIMVAR   (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVAR is present, DIMVAR is used as follows, if:
!
!            - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!              on size(X,1) variables.
!            - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!              on size(X,2) variables.
!
!            The default is DIMVAR = 1.
!
!   XMEAN    (INPUT, OPTIONAL) real(stnd), dimension(:)
!            On entry, if XMEAN is present, XMEAN contains the variable 
!            means and the Principal Component is computed from the
!            centered data matrix X.
!
!            The size of XMEAN must verify: size(XMEAN) = size(X,DIMVAR).
!          
!   XSTD     (INPUT, OPTIONAL) real(stnd), dimension(:)
!            On entry, if XSTD is present, XSTD contains the variable 
!            standard-deviations and the Principal Component is computed from 
!            the normalized data matrix X.
!
!            The size of XSTD must verify: size(XSTD) = size(X,DIMVAR).
!
!   XSINGVAL (INPUT, OPTIONAL) real(stnd)
!            On entry, XSINGVAL must contain the singular value of the covariance
!            (or correlation) matrix from the data matrix XX associated with the
!            eigenvector in XEIGVEC array. If SINGVAL is present, the Principal
!            Component is normalized by XSINGVAL on output.
!            This optional argument is useful only if XEIGVEC contains 
!            an eigenvector derived from an EOF analysis.
!
!          
! Further Details
! _______________
!
!   The subroutine computes the Principal Component with only one pass
!   through the data by regressing the observations onto the eigenvector
!   of the correlation or covariance matrix of X.
!
!   If for some observations in X there is no available data, the corresponding element
!   in XPC is filled with XMISS.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : zero
    use Char_Constants,    only : stat_error1, stat_error5, stat_error6
    use Utilities,         only : assert_eq, assert, merror
    use LLSQ_Procedures,   only : solve_llsq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                             :: xmiss
    real(stnd), intent(in),             dimension(:)   :: xeigvec
    real(stnd), intent(in),             dimension(:,:) :: x
    real(stnd), intent(in),  optional                  :: xsingval
    real(stnd), intent(in),  optional , dimension(:)   :: xstd, xmean
    real(stnd), intent(out),            dimension(:)   :: xpc
!
    integer(i4b), intent(in),  optional :: dimvar
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(xeigvec)) :: dxi
!
    integer(i4b) :: i, n, m
!
    logical(lgl)                           :: dim_is_one, test_mean, test_std, test_singval
    logical(lgl), dimension(size(xeigvec)) :: mask
!
#ifdef _OPENMP
    integer(i4b) :: j
!
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_pc_miss'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    test_mean     = false
    test_std      = false
    test_singval  = false
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
    n = assert_eq(  int(size(x,3-int(i)),i4b),     &
                    int(size(xpc),i4b),            &
                    name_proc )
!
    m = assert_eq(  int(size(x,int(i)),i4b),       &
                    int(size(xeigvec),i4b),        &
                    name_proc )
!
    if ( min(n,m)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    if ( present(xmean) ) then
!
        call assert( logical(size(xmean)==m,lgl),    &
                      name_proc )
!
        test_mean = true
!
    end if
!
    if ( present(xstd) ) then
!
        call assert( logical(size(xstd)==m,lgl),    &
                      name_proc )
!
        test_std = true
!
    end if
!
    if ( present(xsingval) ) then
!
        if ( xsingval<=zero ) call merror( name_proc//stat_error6 )
!
        test_singval = true
!
    end if
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.    &
                  (n*m)>=omp_limit                    .and.    &
                  n>1_i4b                             .and.    &
                  i>1_i4b                             .and.    &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)        &
!$OMP         ,PRIVATE(i,dxi,mask)
!
!$OMP DO SCHEDULE(STATIC) 
!
    do i = 1_i4b, n
!
        if ( dim_is_one ) then
            dxi(1_i4b:m) = x(1_i4b:m,i)
        else
            dxi(1_i4b:m) = x(i,1_i4b:m)
        end if
!
        mask(1_i4b:m) = dxi(1_i4b:m) /= xmiss
!
        if ( any(mask(1_i4b:m)) ) then
!
            if ( test_mean ) then
                where ( mask(1_i4b:m) ) dxi(1_i4b:m) = dxi(1_i4b:m) - xmean(1_i4b:m)
            end if
!
            if ( test_std ) then
                where ( mask(1_i4b:m) ) dxi(1_i4b:m) = dxi(1_i4b:m) / xstd(1_i4b:m)
            end if
!
!           SOLVE THE REGRESSION PROBLEM.
!
            xpc(i) = solve_llsq( pack(xeigvec(1_i4b:m),mask(1_i4b:m)), pack(dxi(1_i4b:m),mask(1_i4b:m)) )
!
            if ( test_singval ) then
                xpc(i) = xpc(i) / xsingval
            end if
!
        else
!
            xpc(i) = xmiss
!
        end if
!
    end do
!
!$OMP END DO
!
!$OMP END PARALLEL
!
!
! END OF SUBROUTINE comp_pc_miss_rv
! _________________________________
!
    end subroutine comp_pc_miss_rv
!
! =========================================================================================
!
    subroutine comp_pc_miss_rm( x, xeigvec, xpc, xmiss, dimvar, xmean, xstd, xsingval, tol, min_norm )
!
! Purpose
! _______
!
!   COMP_PC_MISS computes estimates of Principal Components (PC) 
!   from a data matrix XX and a set of eigenvectors or singular vectors derived
!   from EOF or MCA analysis when XX contains missing values.
!
!
! Arguments
! _________
!
!   X        (INPUT) real(stnd), dimension(:,:)
!            On entry, input submatrix containing size(X,3-DIMVAR)
!            observations on size(X,DIMVAR) variables from the matrix of data XX
!            for which Principal Components are desired. By default,
!            DIMVAR is equal to 1. See description of optional DIMVAR argument 
!            for details.
!            If all the data are available at once, X can be the full 
!            data matrix XX.
!
!   XEIGVEC  (INPUT) real(stnd), dimension(:,:)
!            On entry, XEIGVEC contains selected eigenvectors of
!            the variance-covariance (or correlation) matrix from the data matrix XX
!            or selected left (or right) singular vectors of the SVD of the covariance matrix
!            between the data matrix XX and another data matrix YY.
!
!            The shape of XEIGVEC must verify: size(XEIGVEC,1) = size(X,DIMVAR).
!
!   XPC      (OUTPUT) real(stnd), dimension(:,:)
!            On exit, XPC contains the Principal Components derived from X and XEIGVEC.
!
!            The shape of XPC must verify:
!
!            - size(XPC,1) = size(X,3-DIMVAR) ,
!            - size(XPC,2) = size(XEIGVEC,2)  .
!
!   XMISS    (INPUT) real(stnd)
!            On entry, the missing value indicator. Any value in
!            X which is equal to XMISS is assumed to be missing
!            or invalid.
!            The Principal Components are computed from all the observations where X
!            X are not missing by regression (see Further Details).
!
!   DIMVAR   (INPUT, OPTIONAL) integer(i4b)
!            On entry, if DIMVAR is present, DIMVAR is used as follows, if:
!
!            - DIMVAR = 1, the input submatrix X contains size(X,2) observations
!              on size(X,1) variables.
!            - DIMVAR = 2, the input submatrix X contains size(X,1) observations
!              on size(X,2) variables.
!
!            The default is DIMVAR = 1.
!
!   XMEAN    (INPUT, OPTIONAL) real(stnd), dimension(:)
!            On entry, if XMEAN is present, XMEAN contains the variable 
!            means and the Principal Components are computed from the
!            centered data matrix X.
!
!            The size of XMEAN must verify: size(XMEAN) = size(X,DIMVAR).
!          
!   XSTD     (INPUT, OPTIONAL) real(stnd), dimension(:)
!            On entry, if XSTD is present, XSTD contains the variable 
!            standard-deviations and the Principal Components are computed from 
!            the normalized data matrix X.
!
!            The size of XSTD must verify: size(XSTD) = size(X,DIMVAR).
!
!   XSINGVAL (INPUT, OPTIONAL) real(stnd), dimension(:)
!            On entry, XSINGVAL must contain the singular values of the covariance
!            (or correlation) matrix from the data matrix XX associated with the
!            eigenvectors in XEIGVEC array. If SINGVAL is present, the Principal
!            Components are normalized by XSINGVAL on output.
!            This optional argument is useful only if XEIGVEC contains 
!            eigenvectors derived from an EOF analysis.
!
!            The size of XSINGVAL must verify: size(XSINGVAL) = size(XEIGVEC,2).
!
!   TOL      (INPUT, OPTIONAL) real(stnd)
!            On entry, TOL is used to determine the effective rank
!            of the coefficient matrix A for each regression problem,
!            which is then defined as the order of the largest 
!            leading triangular submatrix R11 in the QR factorization
!            (with pivoting) of A whose estimated condition number,
!            in the 1-norm, is less than 1/TOL. TOl must be set to 
!            the relative precision of the elements in A and B.
!            If each element is correct to, say, 5 digits then
!            TOL=0.00001 should be used.
!
!            TOL must not be greater or equal to 1 or less or equal
!            than 0, otherwise the numerical rank of A is determined
!            and the calculations to determine the condition number
!            are not performed. If TOL is absent, the numerical 
!            rank of A is determined for each regression problem.
!
!   MIN_NORM (INPUT, OPTIONAL) logical(lgl)
!            On entry, If MIN_NORM=true,  minimun 2-norm solutions
!            are computed. If MIN_NORM=false or if MIN_NORM is absent,
!            solutions are computed such that if the j-th column of XEIGVEC
!            is omitted from the basis for the ith observation (regression problem),
!            X[i,j] is set to zero.
!
!          
! Further Details
! _______________
!
!   The subroutine computes the Principal Components with only one pass
!   through the data by regressing the observations onto the eigenvectors
!   of the correlation or covariance matrix of X.
!
!   If for some observations in X there is no available data, the corresponding line in XPC
!   is filled with XMISS.
!
!
! _________________________________________________________________________________________
!
!
! USED MODULES
! ____________
!
    use Logical_Constants
    use Reals_Constants,   only : zero
    use Char_Constants,    only : stat_error1, stat_error5, stat_error6
    use Utilities,         only : assert_eq, assert, merror, arth
    use LLSQ_Procedures,   only : solve_llsq
!
!
! SPECIFICATIONS FOR ARGUMENTS
! ____________________________
!
    real(stnd), intent(in)                             :: xmiss
    real(stnd), intent(in),             dimension(:,:) :: x, xeigvec
    real(stnd), intent(in),  optional                  :: tol
    real(stnd), intent(in),  optional , dimension(:)   :: xstd, xmean, xsingval
    real(stnd), intent(out),            dimension(:,:) :: xpc
!
    integer(i4b), intent(in),  optional :: dimvar
!
    logical(lgl), intent(in), optional  :: min_norm
!
!
! SPECIFICATIONS FOR LOCAL VARIABLES
! __________________________________
!
    real(stnd), dimension(size(xeigvec,1)) :: dxi
!
    integer(i4b)                             :: i, n, n2, m, p
    integer(i4b), dimension(size(xeigvec,1)) :: indice, indice2
!
    logical(lgl)                             :: dim_is_one, test_mean, test_std, test_singval
    logical(lgl), dimension(size(xeigvec,1)) :: mask
!
#ifdef _OPENMP
    integer(i4b) :: j
!
    logical :: test_par, ompparallel, ompnested
#endif
!
!
! PARAMETERS
! __________
!
    character(len=*),  parameter :: name_proc='comp_pc_miss'
!   
!   
! EXECUTABLE STATEMENTS
! _____________________
!
    test_mean     = false
    test_std      = false
    test_singval  = false
!
!   TEST THE ARGUMENTS.
!
    if ( present(dimvar) ) then
!
        if ( dimvar==1_i4b .or. dimvar==2_i4b ) then
            i = dimvar
        else
           call merror( name_proc//stat_error5 )
        end if
!
    else
!
        i = 1_i4b
!
    end if
!
    dim_is_one = i==1_i4b
!
    n = assert_eq(  int(size(x,3-int(i)),i4b),      &
                    int(size(xpc,1),i4b),           &
                    name_proc )
!
    m = assert_eq(  int(size(x,int(i)),i4b),       &
                    int(size(xeigvec,1),i4b),      &
                    name_proc )
!
    p = assert_eq(  int(size(xeigvec,2),i4b),      &
                    int(size(xpc,2),i4b),          &
                    name_proc )
!
    if ( min(n,m,p)<=0_i4b ) call merror( name_proc//stat_error1 )
!
    if ( present(xmean) ) then
!
        call assert( logical(size(xmean)==m,lgl),    &
                      name_proc )
!
        test_mean = true
!
    end if
!
    if ( present(xstd) ) then
!
        call assert( logical(size(xstd)==m,lgl),    &
                      name_proc )
!
        test_std = true
!
    end if
!
    if ( present(xsingval) ) then
!
        call assert( logical(size(xsingval)==p,lgl),    &
                      name_proc )
!
        if ( any(xsingval<=zero) ) call merror( name_proc//stat_error6 )
!
        test_singval = true
!
    end if
!
    indice=arth(1_i4b,1_i4b,m)
!
#ifdef _OPENMP
    i = omp_get_num_procs()
    j = omp_get_max_threads()
    ompparallel = omp_in_parallel()
    ompnested   = omp_get_nested() .and. ompparallel
    test_par    = ( .not.ompparallel .or. ompnested ) .and.    &
                  (n*p*m)>=omp_limit                  .and.    &
                  n>1_i4b                             .and.    &
                  i>1_i4b                             .and.    &
                  j>1_i4b
#endif
!
!$OMP PARALLEL IF(test_par)        &
!$OMP         ,PRIVATE(i,n2,dxi,mask,indice2)
!
!$OMP DO SCHEDULE(STATIC) 
!
    do i = 1_i4b, n
!
        if ( dim_is_one ) then
            dxi(1_i4b:m) = x(1_i4b:m,i)
        else
            dxi(1_i4b:m) = x(i,1_i4b:m)
        end if
!
        mask(1_i4b:m) = dxi(1_i4b:m) /= xmiss
!
        n2 = count( mask )
!
        if ( n2/=0_i4b ) then
!
            if ( test_mean ) then
                where ( mask(1_i4b:m) ) dxi(1_i4b:m) = dxi(1_i4b:m) - xmean(1_i4b:m)
            end if
!
            if ( test_std ) then
                where ( mask(1_i4b:m) ) dxi(1_i4b:m) = dxi(1_i4b:m) / xstd(1_i4b:m)
            end if
!
!           SOLVE THE REGRESSION PROBLEM.
!
            indice2(1_i4b:n2) = pack( indice(1_i4b:m), mask(1_i4b:m) )
            xpc(i,1_i4b:p) = solve_llsq( xeigvec(indice2(1_i4b:n2),1_i4b:p) ,      &
                                         dxi(indice2(1_i4b:n2)),                   &
                                         tol=tol, min_norm=min_norm                )
!
            if ( test_singval ) then
                xpc(i,1_i4b:p) = xpc(i,1_i4b:p) / xsingval(1_i4b:p)
            end if
!
        else
!
            xpc(i,1_i4b:p) = xmiss
!
        end if
!
    end do
!
!$OMP END DO
!
!$OMP END PARALLEL
!
!                                                                              
! END OF SUBROUTINE comp_pc_miss_rm
! _________________________________
!
    end subroutine comp_pc_miss_rm
!
! =========================================================================================
!
!
! ***********************************
! END OF MODULE Mul_Stat_Procedures *
! ***********************************
!    
end module Mul_Stat_Procedures
